/*
 *-----------------------------------------------------------------
 *
 * Name:	cache.h
 *
 * RCS:  $Id: cache.h,v 1.17 2002/03/09 10:54:36 frank Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Cache related macros and definitions
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/14/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _SYS_CACHE_H
#define _SYS_CACHE_H

#define CACHELINE_SIZE	0x20
#define CACHELINE_MASK	0x1F
#define CACHELINE_SHIFT	5

#if defined(LINUX_TEST) || defined(SIBYTE)

#define CACHEOP(_op, _addr)
#define CACHE_FLUSH(_sa, _ea)
#define CACHE_INVALIDATE_STRUCT(_sa, _nbytes)
#define CACHE_FLUSH_PACKET(_edesc)
#define CACHE_FLUSH_STRUCT(_sa, _nbytes)

#else

#include <sys/mips.h>

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *				Generic cacheop routine.  Caller specifies address and
 *				cacheop to perform at that address.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/5/01
 *
 *-----------------------------------------------------------------
 */
#define CACHEOP(_op, _addr) {					\
	uint32	_r;									\
	__asm volatile (".set noreorder\n\tand %0,%2,~0x1F\n\tand %0,%0,%0\n\tcache %1,0(%0)\n\t.set reorder"						\
	: "=r" (_r) : "i" (_op), "r" (_addr));		\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *				Flush (writeback to memory and invalidate the caches)
 *				an area of memory.  The addresses do not need to be
 *				aligned on cacheline boundaries, although the caller
 *				should be aware of what's getting flushed if the
 *				addresses are not justified.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/5/01
 *
 *-----------------------------------------------------------------
 */
#define CACHE_FLUSH(_sa, _ea) {								\
	uint32	_s, _e;											\
	uint32	_len;											\
															\
	_s = (uint32)(_sa) & ~CACHELINE_MASK;					\
	_e = ((uint32)(_ea) + CACHELINE_MASK) & ~CACHELINE_MASK;\
	_len = _e - _s;											\
	CACHE_FLUSH_STRUCT(_s, _len);							\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *				Assumes a cacheline aligned start address.
 *				The number of bytes is from the cacheline base,
 *					so if someone calls this routine, but before they do,
 *					they justify the start address, the number of
 *					bytes must also be adjusted.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/5/01
 *
 *-----------------------------------------------------------------
 */
#define CACHE_INVALIDATE_STRUCT(_sa, _nbytes) {							\
	uint32	_i, _clines, _addr;											\
																		\
	if ((_sa) & CACHELINE_MASK)											\
		panic("unaligned invalidate command, 0x%x\n", (_sa));			\
																		\
	_clines = ((_nbytes) + (CACHELINE_SIZE - 1)) >> CACHELINE_SHIFT;	\
																		\
	for (_i = 0, _addr = (uint32)(_sa); _i < _clines; _i++) {			\
		CACHEOP(Hit_Invalidate_S, _addr);								\
		_addr += CACHELINE_SIZE;										\
	}																	\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *				Assumes a cacheline aligned start address.
 *				The number of bytes is from the cacheline base,
 *					so if someone calls this routine, but before they do,
 *					they justify the start address, the number of
 *					bytes must also be adjusted.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/5/01
 *
 *-----------------------------------------------------------------
 */
#define CACHE_FLUSH_STRUCT(_sa, _nbytes) {								\
	uint32	_i, _clines, _addr;											\
																		\
	if ((uint32)(_sa) & CACHELINE_MASK)									\
		panic("unaligned invalidate command, 0x%x\n", (_sa));			\
																		\
	_clines = ((_nbytes) + (CACHELINE_SIZE - 1)) >> CACHELINE_SHIFT;	\
																		\
	for (_i = 0, _addr = (uint32)(_sa); _i < _clines; _i++) {			\
		CACHEOP(Hit_Writeback_Inv_S, _addr);							\
		_addr += CACHELINE_SIZE;										\
	}																	\
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *				Flush (Writeback to memory and invalidate the caches)
 *				the contents of an EEE packet (buffers and descriptor).
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/5/01
 *
 *-----------------------------------------------------------------
 */
#if 1
#define CACHE_FLUSH_PACKET(_edesc) {										\
	eee_descBuf_t	* _bd;													\
	eee_descHdr_t	* _hdr;													\
	uchar8			* _buf, * _j;											\
	int32			_num_bd, _i;											\
	int32			_nbytes;												\
																			\
	_hdr = (eee_descHdr_t *)(_edesc);										\
	_num_bd = (_hdr->control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;	\
	if (_hdr->control & EDESC_COMMAND_USE_FIRST_BUFFER)						\
		_bd = &((_edesc)->bd[0]);											\
	else																	\
		_bd = &((_edesc)->bd[1]);											\
																			\
	for (_i=0; _i<_num_bd; _i++, _bd++) {									\
		_buf = (uchar8 *)eee_getBufBase(_bd->buf);							\
		_nbytes = eee_getBufSize(eee_getBufType(_buf));						\
		for (_j=_buf; _j<_buf +_nbytes; _j+=CACHELINE_SIZE) {				\
			CACHEOP(Hit_Writeback_Inv_S, _j);								\
		}																	\
	}																		\
	CACHE_FLUSH_STRUCT ((_edesc), sizeof(eee_desc_t));						\
}
#else
#define CACHE_FLUSH_PACKET(_edesc) {										\
	eee_descBuf_t	* _bd;													\
	eee_descHdr_t	* _hdr;													\
	uchar8			* _buf, * _j;											\
	int32			_num_bd, _i;											\
																			\
	_hdr = (eee_descHdr_t *)(_edesc);										\
	_num_bd = (_hdr->control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;	\
	if (_hdr->control & EDESC_COMMAND_USE_FIRST_BUFFER)						\
		_bd = &((_edesc)->bd[0]);											\
	else																	\
		_bd = &((_edesc)->bd[1]);											\
																			\
	for (_i=0; _i<_num_bd; _i++, _bd++) {									\
		_buf = (uchar8 *)_bd->buf;											\
		for (_j=_buf; _j<_buf+2048; _j+=0x20) {								\
			CACHEOP(Hit_Writeback_Inv_S, _j);								\
		}																	\
	}																		\
	CACHE_FLUSH_STRUCT ((_edesc), sizeof(eee_desc_t));						\
}
#endif

#endif	/* LINUX_TEST */

#endif	/* _SYS_CACHE_H */

