/* vi:ts=4:sw=4
 *-----------------------------------------------------------------
 *
 * Name:	agent-control.c
 *
 * RCS:  $Id: agent-control.c,v 1.16 2002/10/02 01:10:16 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	AM control module
 *
 * Created by:	Pankaj Gupta 
 *
 * Date Created:	11/05/01
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()

#include <sys/time.h>
#include <sys/socket.h>
#include <signal.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <errno.h>

#include "exports/anpssc-api.h"
#include "exports/ncm-api.h"
#include "exports/evm-api.h"
#include "exports/fs-api.h"
#include "agent-api.h"


#define CRON_PID              "/var/run/cron.pid"

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

int32 am_writeConf(void);
int32 am_readConf(void);
void am_addCrontabEntry(am_srcConfig_t *srcCfgPtr);
void am_deleteCrontabEntry(am_srcConfig_t *srcCfgPtr);
int32 am_getVolId(char *volName, evm_volId_t *volId);
int32 am_startMirror(am_srcConfig_t *srcCfgPtr);
int32 am_stopMirror(am_srcConfig_t *srcCfgPtr);
void handleAmMessages(int sockFd);

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* Basic message structure. */
typedef struct
{   
    char    buf[2000];  // optimize!!!
	 uint32  len;
} am_msg_t;

/* Wait 10s for response (10s).
 */
static struct timeval waitEvmTo = {10, 0};
static struct timeval waitAmTo = {60, 0};


#define		MAX_AM_MSG_SIZE		10000

#define AM_CONF 			"/usr/local/agile/conf/am.conf"
#define AM_CRON_PROG		"/usr/local/agile/bin/amcron"
#define NFX_CRONTAB_FILE	"/var/cron/tabs/root"
#define NFX_CRONTAB_TMPFILE	"/var/cron/tabs/.root.tmp"

int32    amAppId;
int amAppFd;
am_tgtConfig_t tgtConfigBlk;
am_srcConfig_t *srcConfigTbl[MAX_SRC_MIRROR_ENTRY];

/*------------------------------------------------------------------------------
 * Name         : am_writeConf()
 *
 * Description  : Write AM configuration
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
int32
am_writeConf()
{
	int32 idx;
	FILE *fd;
	int32 numSrcEntry, numTgtEntry;
	am_srcConfig_t *srcCfgPtr;

	fd = fopen (AM_CONF, "w");
	if (!fd)
	{
		AM_ERR_PRINT(("am_writeConf: Could not open %s for write config", \
					AM_CONF));
		return -1;
	}

	numSrcEntry = 0;
	for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
	{
		if (srcConfigTbl[idx])
			numSrcEntry++;
	}

	numTgtEntry = 0;
	for (idx = 0; idx < MAX_TGT_HOST_ENTRY; idx++)
	{
		if (tgtConfigBlk.allowedHost[idx])
			numTgtEntry++;
	}

	fprintf(fd, "%d\n", numSrcEntry);
	fprintf(fd, "%d\n\n", numTgtEntry);

	for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
	{
		srcCfgPtr = srcConfigTbl[idx];

		if (srcCfgPtr == NULL)
			continue;

		fprintf(fd, "%d\n", idx);
		fprintf(fd, "%s\n", srcCfgPtr->amName);
		fprintf(fd, "%d\n", srcCfgPtr->adminState);
		fprintf(fd, "%d\n", srcCfgPtr->numMirrorStarted);
		fprintf(fd, "%d\n", srcCfgPtr->numMirrorCompleted);
		fprintf(fd, "%d\n", srcCfgPtr->numMirrorAborted);
		fprintf(fd, "%llx\n", srcCfgPtr->minBitMap);
		fprintf(fd, "%x\n", srcCfgPtr->hourBitMap);
		fprintf(fd, "%x\n", srcCfgPtr->dateBitMap);
		fprintf(fd, "%x\n", srcCfgPtr->monthBitMap);
		fprintf(fd, "%x\n", srcCfgPtr->dayBitMap);
		fprintf(fd, "%d\n", srcCfgPtr->maxDataRate);
		fprintf(fd, "%x\n", srcCfgPtr->srcAddr);
		fprintf(fd, "%x\n", srcCfgPtr->tgtAddr);

		fprintf(fd, "%s\n", srcCfgPtr->srcFsysName);
		fprintf(fd, "%s\n\n", srcCfgPtr->tgtFsysName);
	}

	for (idx = 0; idx < MAX_TGT_HOST_ENTRY; idx++)
	{
		if (tgtConfigBlk.allowedHost[idx] == 0)
			continue;

		fprintf(fd, "%x\n", tgtConfigBlk.allowedHost[idx]);
	}

	fclose(fd);

	return 0;
}

/*------------------------------------------------------------------------------
 * Name         : am_readConf()
 *
 * Description  : Read AM configuration
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
int32
am_readConf()
{
	int32 	idx;
	uint32	ctrlHdl = 0;
	FILE 	*fd;
	int32 	numSrcEntry, numTgtEntry;

	am_srcConfig_t *srcCfgPtr;

	fd = fopen (AM_CONF, "r");
	if (!fd)
	{
		AM_ERR_PRINT(("am_readConf: Could not open %s for read config", \
					AM_CONF));
		return -1;
	}

	fscanf(fd, "%d", &numSrcEntry);
	fscanf(fd, "%d", &numTgtEntry);

	for (idx = 0; idx < numSrcEntry; idx++)
	{
		fscanf(fd, "%d", &ctrlHdl);

		if ((ctrlHdl < 0) || (ctrlHdl >= MAX_SRC_MIRROR_ENTRY) ||
				(srcConfigTbl[ctrlHdl] != NULL))
		{
			AM_ERR_PRINT(("Error encountered in AM source config block"));
			break;
		}

		srcConfigTbl[ctrlHdl] = malloc(sizeof(am_srcConfig_t));
		memset(srcConfigTbl[ctrlHdl], 0, sizeof(am_srcConfig_t));
		srcCfgPtr = srcConfigTbl[ctrlHdl];

		srcCfgPtr->ctrlHdl = ctrlHdl;
		fscanf(fd, "%s", srcCfgPtr->amName);
		fscanf(fd, "%d", &srcCfgPtr->adminState);
		fscanf(fd, "%d", &srcCfgPtr->numMirrorStarted);
		fscanf(fd, "%d", &srcCfgPtr->numMirrorCompleted);
		fscanf(fd, "%d", &srcCfgPtr->numMirrorAborted);
		fscanf(fd, "%llx", &srcCfgPtr->minBitMap);
		fscanf(fd, "%x", &srcCfgPtr->hourBitMap);
		fscanf(fd, "%x", &srcCfgPtr->dateBitMap);
		fscanf(fd, "%x", &srcCfgPtr->monthBitMap);
		fscanf(fd, "%x", &srcCfgPtr->dayBitMap);
		fscanf(fd, "%d", &srcCfgPtr->maxDataRate);
		fscanf(fd, "%x", &srcCfgPtr->srcAddr);
		fscanf(fd, "%x", &srcCfgPtr->tgtAddr);

		fscanf(fd, "%s", srcCfgPtr->srcFsysName);
		fscanf(fd, "%s", srcCfgPtr->tgtFsysName);
	}

	for (idx = 0; idx < numTgtEntry; idx++)
	{
		fscanf(fd, "%x", &tgtConfigBlk.allowedHost[idx]);
	}

	fclose(fd);

	return 0;
}

/*------------------------------------------------------------------------------
 * Name         : am_addCrontabEntry()
 *
 * Description  : Add an entry in the cron tab file
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void
am_addCrontabEntry(am_srcConfig_t *srcCfgPtr)
{
	char cronCmdLine[1024];
	char lclBuf[256];
	int32 needComma = 0;
	int32 idx;
	FILE *fd;

	if (!srcCfgPtr->minBitMap && !srcCfgPtr->hourBitMap && 
			!srcCfgPtr->dateBitMap && !srcCfgPtr->monthBitMap &&
				!srcCfgPtr->dayBitMap)
	{
		return;
	}
	
	/*
	 * Open the crontab file in append mode
	 */
	fd = fopen (NFX_CRONTAB_FILE, "a");
	if (!fd)
	{
		AM_ERR_PRINT((
			"am_addCrontabEntry: Could not open %s for cron config", \
					NFX_CRONTAB_FILE));
		return;
	}

	/*
	 * Prepare the cron tab entry in correct format
	 */

	memset(cronCmdLine, 0, sizeof(cronCmdLine));
	memset(lclBuf, 0, sizeof(lclBuf));

	/* 
	 * Minute field 
	 */
	needComma = 0;
	if (srcCfgPtr->minBitMap)
	{
		for (idx = 0; idx <= 59; idx++)
		{
			if (GET_BIT(srcCfgPtr->minBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Hour field 
	 */
	needComma = 0;
	if (srcCfgPtr->hourBitMap)
	{
		for (idx = 0; idx <= 23; idx++)
		{
			if (GET_BIT(srcCfgPtr->hourBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Date field 
	 */
	needComma = 0;
	if (srcCfgPtr->dateBitMap)
	{
		for (idx = 1; idx <= 31; idx++)
		{
			if (GET_BIT(srcCfgPtr->dateBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Month field 
	 */
	needComma = 0;
	if (srcCfgPtr->monthBitMap)
	{
		for (idx = 1; idx <= 12; idx++)
		{
			if (GET_BIT(srcCfgPtr->monthBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Day field 
	 */
	needComma = 0;
	if (srcCfgPtr->dayBitMap)
	{
		for (idx = 0; idx <= 7; idx++)
		{
			if (GET_BIT(srcCfgPtr->dayBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/*
	 * Put program name
	 */
	strcat(cronCmdLine, AM_CRON_PROG);
	strcat(cronCmdLine, " ");

	/*
	 * Put mirror name as argument
	 */
	sprintf(lclBuf, "%s", srcCfgPtr->amName);
	strcat(cronCmdLine, lclBuf);

	/*
	 * Write entry in the cron tab file
	 */
	fprintf(fd, "%s\n", cronCmdLine);
	fclose(fd);
}

/*------------------------------------------------------------------------------
 * Name         : am_deleteCrontabEntry()
 *
 * Description  : Update the cron tab file
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void
am_deleteCrontabEntry(am_srcConfig_t *srcCfgPtr)
{
	char cronCmdLine[1024];
	char patternBuf[256];
	char lclBuf[256];
	char cmdBuf[256];
	FILE *fd;
	FILE *fdTmp;

	AM_DBG_PRINT(("Enter am_deleteCrontabEntry"));
	fd = fopen (NFX_CRONTAB_FILE, "r");
	if (!fd)
	{
		AM_ERR_PRINT((
			"am_deleteCrontabEntry: Could not open %s for cron config",
					NFX_CRONTAB_FILE));
		return;
	}

	fdTmp = fopen (NFX_CRONTAB_TMPFILE, "w");
	if (!fdTmp)
	{
		AM_ERR_PRINT((
			"am_deleteCrontabEntry: Could not open %s for cron config",
					NFX_CRONTAB_TMPFILE));
		return;
	}

	memset(cronCmdLine, 0, sizeof(cronCmdLine));
	memset(patternBuf, 0, sizeof(patternBuf));
	memset(lclBuf, 0, sizeof(lclBuf));
	memset(cmdBuf, 0, sizeof(cmdBuf));

	/*
	 * Put program name
	 */
	strcat(patternBuf, AM_CRON_PROG);
	strcat(patternBuf, " ");

	/*
	 * Put mirror name as argument
	 */
	sprintf(lclBuf, "%s", srcCfgPtr->amName);
	strcat(patternBuf, lclBuf);

	/*
	 * Keep reading each line and writing it back
	 * Remove the line which has pattern matching for 
	 * mirror entry to be deleted
	 */
	 while(fgets(cronCmdLine, 1024, fd) != NULL)
	 {
		if (strstr(cronCmdLine, patternBuf) != NULL)
			continue;
		fputs(cronCmdLine, fdTmp);
		memset(cronCmdLine, 0, sizeof(cronCmdLine));
	 }

	 fclose(fd);
	 fclose(fdTmp);

	 sprintf(cmdBuf, "mv %s %s", NFX_CRONTAB_TMPFILE, NFX_CRONTAB_FILE);
	 system(cmdBuf);
}

/*------------------------------------------------------------------------------
 * Name         : am_getVolId()
 *
 * Description  : Get identifier of a volume knowing its name.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
am_getVolId(char *volName, evm_volId_t *volId)
{
    evm_MsgGetId_t      evmReq;
    am_msg_t            evmMsg;
    evm_MsgStat_t *     evmRsp = (evm_MsgStat_t *) evmMsg.buf;

    /* Reset memory for request and response.
     */
    bzero((void *) &evmReq, sizeof(evmReq));
    bzero((void *) &evmMsg, sizeof(evmMsg));

    /* Fill request.
     */
    evmReq.hdr.msgType = EVM_VOL_GET_ID;
    evmReq.hdr.clientPrivate = 0;   // not used
    strcpy(evmReq.name, volName);

    /* Send request to EVM.
     */
    evmMsg.len = sendAgileMsg(-1,
                              &evmReq, sizeof(evmReq), 
                              "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                              &waitEvmTo,
                              evmMsg.buf, sizeof(evmMsg.buf));

    /* Receive response from EVM.
     * Copy the returned volume id in control block.
     */
	if (evmMsg.len <  sizeof(evm_msgHdr_t))
		return NFX_ERR;

	if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		return NFX_ERR;

    if (evmRsp->stat != NFX_OK)
		return NFX_ERR;

    memcpy((void *) volId, (const void *) &(evmRsp->payload.lVolId), 
					sizeof(evm_volId_t));
    
    return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : am_startMirror
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 02/12/02
 *
 *------------------------------------------------------------------------------
 */
int32 am_startMirror(am_srcConfig_t *srcCfgPtr)
{
	int32 rspLen = 0;
	am_startMirrorReq_t startReq;
	am_startMirrorRsp_t startRsp;

	/*
	 * Mirror operation is triggered by the PCC of the cluster
	 * that owns the source node. So we can obtain source vol id
	 * For target vol id DM message exchange will use Tgt Vol
	 * Name and volid will be obtained later 
	 */
#ifdef IN_CLUSTER_ONLY
	if (am_getVolId(srcCfgPtr->tgtFsysName, &srcCfgPtr->tgtVolId) != NFX_OK)
	{
		AM_ERR_PRINT(("am_startMirror did not get Tgt volid for %s", 
						srcCfgPtr->tgtFsysName));
		return NFX_ERR;
	}
#endif

	if (am_getVolId(srcCfgPtr->srcFsysName, &srcCfgPtr->srcVolId) != NFX_OK)
	{
		AM_ERR_PRINT(("am_startMirror did not get Src volid for %s", 
						srcCfgPtr->srcFsysName));
		return NFX_ERR;
	}

	/*
	 * Send out request to source for start mirroring
	 */
	memset(&startReq, 0, sizeof(am_startMirrorReq_t));
	memset(&startRsp, 0, sizeof(am_startMirrorRsp_t));
        
	startReq.hdr.msgType = AM_START_MIRROR_REQ;
	startReq.hdr.ctrlHdl = srcCfgPtr->ctrlHdl; 
            
	strcpy(startReq.amName, srcCfgPtr->amName);
	strcpy(startReq.tgtFsysName, srcCfgPtr->tgtFsysName);
	startReq.amVolId = srcCfgPtr->srcVolId;
	startReq.lclAddr = srcCfgPtr->srcAddr;		// local IP @ (source)
	startReq.rmtAddr = srcCfgPtr->tgtAddr;      // local IP @ (target)
	startReq.rmtPort = htons(11000);			// local port (target)
	startReq.maxDataRate = srcCfgPtr->maxDataRate;

	/*
	 * For next mirror operation let the index be 
	 * current mirror plus 1. Only use the last 3 digits
	 */
	startReq.snapSeqSuffix = 
			(srcCfgPtr->numMirrorStarted + 1) % 999;

	AM_FCT_PRINT(("Sending AM_START_MIRROR_REQ for %s", srcCfgPtr->amName));

	rspLen = sendAgileMsg(-1,
						&startReq, sizeof(am_startMirrorReq_t),
						"am", NFX_NFP_FP2, 1, MGMTPLANE,
						&waitAmTo, &startRsp, sizeof(am_startMirrorRsp_t));

	AM_FCT_PRINT(("Got AM_START_MIRROR_RSP Result %s",
					(startRsp.result == NFX_OK) ? "OK" : "NOT_OK"));

	/*
	 * Process response from AM.
	 */
	if ((rspLen == sizeof(am_startMirrorRsp_t)) &&
			(startRsp.hdr.msgType == AM_START_MIRROR_RSP) &&
				(startRsp.result == NFX_OK))
	{
		srcCfgPtr->operState = 1;
		srcCfgPtr->amApplHdl = startRsp.hdr.amApplHdl;
		return NFX_OK;
	}
	else
	{
		return NFX_ERR;
	}
}

/*------------------------------------------------------------------------------
 * Name         : am_stopMirror
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 02/12/02
 *
 *------------------------------------------------------------------------------
 */
int32 am_stopMirror(am_srcConfig_t *srcCfgPtr)
{
	int32 rspLen = 0;
	am_abortAgentReq_t abortReq;
	am_abortAgentRsp_t abortRsp;

	AM_FCT_PRINT(("Sending stop mirror request for %s",
			srcCfgPtr->amName));

	/*
	 * Send out request to source to stop mirroring
	 */
	memset(&abortReq, 0, sizeof(am_abortAgentReq_t));
	memset(&abortRsp, 0, sizeof(am_abortAgentRsp_t));
        
	abortReq.hdr.msgType = AM_AGENT_ABORT_REQ;
	abortReq.hdr.amApplHdl = srcCfgPtr->amApplHdl; 
	abortReq.hdr.ctrlHdl = srcCfgPtr->ctrlHdl; 
	strcpy(abortReq.amName, srcCfgPtr->amName);
	abortReq.amVolId = srcCfgPtr->srcVolId;

	AM_DBG_PRINT(("Sending AM_AGENT_ABORT_REQ"));

	rspLen = sendAgileMsg(-1, &abortReq, sizeof(am_abortAgentReq_t),
				"am", NFX_NFP_FP2, 1, MGMTPLANE,
				&waitAmTo, &abortRsp, sizeof(am_abortAgentRsp_t));

	AM_FCT_PRINT((
		"Got AM_AGENT_ABORT_RSP rspLen = %d exp %d, type %d, result %s",
					rspLen, sizeof(am_abortAgentRsp_t),
					abortRsp.hdr.msgType,
					(abortRsp.result == NFX_OK) ? "OK" : "NOT_OK"));
				
	/*
	 * Process response from AM.
	 */
	if ((rspLen == sizeof(am_abortAgentRsp_t)) &&
			(abortRsp.hdr.msgType == AM_AGENT_ABORT_RSP) &&
				(abortRsp.result == NFX_OK))
	{
		srcCfgPtr->operState = 0;
		return NFX_OK;
	}
	else
	{
		return NFX_ERR;
	}
}

/*------------------------------------------------------------------------------
 * Name         : cron_restart()
 *
 * Description  : Restart the cron daemon
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 13-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void cron_restart(void)
{
	FILE *pidfile = fopen (CRON_PID, "r");
	pid_t cron_pid = 0;

	if (pidfile) 
	{
		fscanf (pidfile, "%d", &cron_pid);
		fclose (pidfile);
	}

	if (cron_pid) 
	{
		kill (cron_pid, SIGHUP);
	}
}


/*------------------------------------------------------------------------------
 * Name         : handleAmMessages
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 02/12/02
 *
 *------------------------------------------------------------------------------
 */

void handleAmMessages(int sockFd)
{
	int32			idx;
	am_msg_cmd_t	*amMsg;
	am_srcConfig_t	*srcCfgPtr;
	am_tgtConfig_t	*tgtCfgPtr;
    int             rcvdLen;
	char 			rcvMsgBuf[MAX_AM_MSG_SIZE];
    char            srcApp[256];
	uint32          srcCpu;
	uint32          srcSlot;
	int				*msgType;
	int32			changedConfigFlag = 0;

	am_msg_cmd_t	*amRspMsg = NULL;
	int32			rspLen = sizeof(am_msg_cmd_t);
	int				status = NFX_OK;

    /* 
	 * Receive and process event requests
     */
	memset(rcvMsgBuf, 0, MAX_AM_MSG_SIZE);

    rcvdLen = recvAgileMsg (sockFd, rcvMsgBuf, MAX_AM_MSG_SIZE, NULL,
                            srcApp, &srcCpu, &srcSlot, 0);
        
    if (rcvdLen <= 0)
    {
        perror("handleAmMessages: Error in rcv\n");
        return;
    }
      
	msgType = (int *)rcvMsgBuf;

    switch(*msgType)
    {
		/*
		 * Msgs from CLI/Operator
		 */

        case AM_CREATE_MIRROR_REQ:
		{
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];

			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_CREATE_MIRROR_RSP;
	
			/*
			 * Check for duplicate entry
			 */
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
				{
					status = NFX_ERR;
					sprintf(amRspMsg->output, "Mirror %s already configured\n", 
							srcCfgPtr->amName);
					break;
				}
			}

			if (status == NFX_ERR)
				break;

			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
				{
					srcConfigTbl[idx] = malloc(sizeof(am_srcConfig_t));
					memcpy(srcConfigTbl[idx], srcCfgPtr,
								sizeof(am_srcConfig_t));
					srcConfigTbl[idx]->ctrlHdl = idx;
					am_addCrontabEntry(srcCfgPtr);
					cron_restart();
					changedConfigFlag = 1;
					break;
				}
			}
	
			if (idx >= MAX_SRC_MIRROR_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "Max mirror %d already configured\n", 
							MAX_SRC_MIRROR_ENTRY);
			}
	
			break;
		}

		case AM_DELETE_MIRROR_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_DELETE_MIRROR_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];
	
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
				{
					am_deleteCrontabEntry(srcCfgPtr);
					cron_restart();
					free(srcConfigTbl[idx]);
					srcConfigTbl[idx] = NULL;
					changedConfigFlag = 1;
					break;
				}
			}
	
			if (idx >= MAX_SRC_MIRROR_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "%s mirror not found\n", 
						srcCfgPtr->amName);
			}
	
			break;
		}

		case AM_ENABLE_MIRROR_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_ENABLE_MIRROR_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];
	
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
				{
					srcConfigTbl[idx]->adminState = 1;
					changedConfigFlag = 1;
					break;
				}
			}
	
			if (idx >= MAX_SRC_MIRROR_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "%s mirror not found\n", 
						srcCfgPtr->amName);
			}
	
			break;
		}

		case AM_DISABLE_MIRROR_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_DISABLE_MIRROR_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];
	
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
				{
					srcConfigTbl[idx]->adminState = 0;
					changedConfigFlag = 1;
					break;
				}
			}
	
			if (idx >= MAX_SRC_MIRROR_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "%s mirror not found\n", 
						srcCfgPtr->amName);
			}
	
			break;
		}

		case AM_SHOW_MIRROR_REQ:
		{
			int32 tgtIdx, tgtRspIdx;
			int32 numSrcEntry = 0;
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];

			/*
		 	 * Prepare the response
		 	 */
			if (strlen(srcCfgPtr->amName) > 0)
			{
				rspLen = sizeof(am_msg_cmd_t);
				numSrcEntry++;
			}
			else
			{
				for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
				{
					if (srcConfigTbl[idx])
						numSrcEntry++;
				}
			}

			if (numSrcEntry > 1)
			{
				rspLen = sizeof(am_msg_cmd_t) + 
							(numSrcEntry - 1)*sizeof(am_srcConfig_t);
			}
			else
			{
				rspLen = sizeof(am_msg_cmd_t);
			}
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_SHOW_MIRROR_RSP;
	
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strlen(srcCfgPtr->amName) > 0)
				{
					if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
					{
						memcpy(&amRspMsg->srcCfg[0], srcConfigTbl[idx],
									sizeof(am_srcConfig_t));
						amRspMsg->numBlk = 1;
						break;
					}
				}
				else
				{
					memcpy(&amRspMsg->srcCfg[amRspMsg->numBlk], 
								srcConfigTbl[idx], sizeof(am_srcConfig_t));
					amRspMsg->numBlk++;
				}
			}

			tgtRspIdx = 0;
			for (tgtIdx = 0; tgtIdx < MAX_TGT_HOST_ENTRY; tgtIdx++)
			{
				if (tgtConfigBlk.allowedHost[tgtIdx] == 0)
					continue;

				amRspMsg->tgtCfg.allowedHost[tgtRspIdx] =
							tgtConfigBlk.allowedHost[tgtIdx];
				tgtRspIdx++;
			}

	
			if ((strlen(srcCfgPtr->amName) > 0) && 
								(idx >= MAX_SRC_MIRROR_ENTRY))
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "%s mirror not found\n", 
						srcCfgPtr->amName);
			}
	
			break;
		}

		case AM_ADD_HOST_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_ADD_HOST_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			tgtCfgPtr = &amMsg->tgtCfg;
	
			for (idx = 0; idx < MAX_TGT_HOST_ENTRY; idx++)
			{
				if (tgtConfigBlk.allowedHost[idx] == 0)
				{
					tgtConfigBlk.allowedHost[idx] = tgtCfgPtr->allowedHost[0];
					changedConfigFlag = 1;
					break;
				}
			}
	
			if (idx >= MAX_TGT_HOST_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, 
					"Max allowed host already configured\n");
			}
	
			break;
		}

		case AM_DELETE_HOST_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_DELETE_HOST_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			tgtCfgPtr = &amMsg->tgtCfg;
	
			for (idx = 0; idx < MAX_TGT_HOST_ENTRY; idx++)
			{
				if (tgtConfigBlk.allowedHost[idx] == tgtCfgPtr->allowedHost[0])
				{
					tgtConfigBlk.allowedHost[idx] = 0;
					changedConfigFlag = 1;
					break;
				}
			}
	
			if (idx >= MAX_TGT_HOST_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "Host not on allowed node list\n");
			}
	
			break;
		}

		case AM_USER_START_MIRROR_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_USER_START_MIRROR_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];
	
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
				{
					if (srcConfigTbl[idx]->adminState == 0)
					{
						status = NFX_ERR;
						sprintf(amRspMsg->output, "%s mirroring disabled\n", 
							srcCfgPtr->amName);
						break;
					}

					if (srcConfigTbl[idx]->operState)
					{
						status = NFX_ERR;
						sprintf(amRspMsg->output, "%s mirror in progress\n", 
							srcCfgPtr->amName);
						break;
					}
							
					status = am_startMirror(srcConfigTbl[idx]);
					if (status == NFX_OK)
					{
						srcConfigTbl[idx]->numMirrorStarted++;
						changedConfigFlag = 1;
					}
					break;
				}
			}

			if (status == NFX_ERR)
				break;
	
			if (idx >= MAX_SRC_MIRROR_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "%s mirror not found\n", 
						srcCfgPtr->amName);
			}
	
			break;
		}

		case AM_USER_STOP_MIRROR_REQ:
		{
			/*
		 	 * Prepare the response
		 	 */
			rspLen = sizeof(am_msg_cmd_t);
			amRspMsg = malloc(rspLen);
			memset((char *)amRspMsg, 0, rspLen);
			amRspMsg->msgType = AM_USER_STOP_MIRROR_RSP;
	
			amMsg = (am_msg_cmd_t *) &rcvMsgBuf[0];
			srcCfgPtr = &amMsg->srcCfg[0];
	
			for (idx = 0; idx < MAX_SRC_MIRROR_ENTRY; idx++)
			{
				if (srcConfigTbl[idx] == NULL)
					continue;

				if (strcmp(srcConfigTbl[idx]->amName, srcCfgPtr->amName) == 0)
				{
					if (srcConfigTbl[idx]->operState == 0)
					{
						status = NFX_ERR;
						sprintf(amRspMsg->output, "%s no mirror in progress\n", 
							srcCfgPtr->amName);
						break;
					}

					status = am_stopMirror(srcConfigTbl[idx]);
					if (status == NFX_OK)
					{
						srcConfigTbl[idx]->numMirrorAborted++;
						changedConfigFlag = 1;
					}
					break;
				}
			}

			if (status == NFX_ERR)
				break;
	
			if (idx >= MAX_SRC_MIRROR_ENTRY)
			{
				status = NFX_ERR;
				sprintf(amRspMsg->output, "%s mirror not found\n", 
						srcCfgPtr->amName);
			}
	
			break;
		}

		/*
		 * Msg from AM agent (source/target) to Controller
		 */
		case AM_MIRROR_STATUS_IND:
		{
			am_mirrorStatusInd_t *amStatusInd = 
						(am_mirrorStatusInd_t *) &rcvMsgBuf[0];

			AM_FCT_PRINT(("Rcvd AM_MIRROR_STATUS_IND idx %d status %lld",
					amStatusInd->hdr.ctrlHdl,
						amStatusInd->status));

			if ((amStatusInd->hdr.ctrlHdl < 0) ||
				(amStatusInd->hdr.ctrlHdl >= MAX_SRC_MIRROR_ENTRY))
			{
				break;
			}

			srcCfgPtr = srcConfigTbl[amStatusInd->hdr.ctrlHdl];

			if ((amStatusInd->status == AM_MIRROR_COMPLETED) &&
					(srcCfgPtr->operState))
			{
				AM_FCT_PRINT(("Mirror %s operation completed",
							srcCfgPtr->amName));
				srcCfgPtr->operState = 0;
				srcCfgPtr->numMirrorCompleted++;
				changedConfigFlag = 1;
			}
			else if ((amStatusInd->status == AM_MIRROR_ABORTED) &&
					(srcCfgPtr->operState))
			{
				AM_FCT_PRINT(("Mirror %s operation aborted",
							srcCfgPtr->amName));
				srcCfgPtr->operState = 0;
				srcCfgPtr->numMirrorAborted++;
				changedConfigFlag = 1;
			}

			break;
		}

		case AM_AGENT_LISTEN_RSP:
		{
			am_listenRsp_t *amListenRsp = (am_listenRsp_t *) &rcvMsgBuf[0];

			if (amListenRsp->result == NFX_OK)
				AM_FCT_PRINT((
						"AM_AGENT_LISTEN_RSP: Target started listening"));
			else
				AM_FCT_PRINT((
						"AM_AGENT_LISTEN_RSP: Target listen attempt failed"));

			break;
		}

       	default:
		{
			status = NFX_ERR;
			AM_ERR_PRINT(("AM Rcvd unhandled msg %d", *msgType));
		}
    }

	/*
	 * Return the response if there is one
	 */
	if (amRspMsg)
	{
		amRspMsg->result = status;

		sendAgileMsg(-1, amRspMsg, rspLen,
	 				srcApp, srcCpu, srcSlot,
					MGMTPLANE, 0, 0, 0);

		free(amRspMsg);
	}

	/*
	 * Make sure the persistent configuration is uptodate
	 */
	if (changedConfigFlag)
		am_writeConf();
}

/*------------------------------------------------------------------------------
 * Name         : nfx_am_stop()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
static void nfx_am_stop(int sig)
{
	unRegAgileApp(amAppFd);
    exit(0);
}

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
int main(int argc, char **argv)
{
	/*
	 * Sets the procedure to follow in case of request for 
	 * program termination or CTRL-C interrupt.
	 */
	signal(SIGTERM, nfx_am_stop);
	signal(SIGINT, nfx_am_stop);

	 /*
      * Initialize the AM controller
      */
	amAppId = eee_getApplicationId("am");

	/*
	 * Register to receive messages for AM module
	 */
	amAppFd = regAgileApp("am");
	if (amAppFd < 0)
	{
		perror("Could not register AM app with Agile\n");
		return 0;
	}

	am_readConf();
/*
 * Tell mcp we're done with our init and then get to work
 */
	kill (getppid(), SIGUSR1);

	while (1)
		handleAmMessages(amAppFd);

	return 0;
}
