/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : agent.h
 *
 * Description  : Main header file of the Agile Mirror (AM) agent.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-OCT-2001
 * Re-designed  : 25-JAN-2002     Nitin Jain
 *------------------------------------------------------------------------------
 */
#ifndef _AM_AGENT_H_
#define _AM_AGENT_H_

/*
 * Include files
 */

#include "nfx-incl.h"

#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-netinet/in.h"

#include "exports/eee-api.h"
#include "exports/esm-api.h"
#include "exports/malloc-api.h"
#include "exports/evm-api.h"
#include "exports/fs-api.h"
#include "exports/fs-am-api.h"
#ifndef SSC
#include "exports/tpl-api.h"
#endif

#include "exports/datamover-api.h"
#include "agent-api.h"
#include "source-sm.h"
#include "target-sm.h"

/*
 * Constants and macros.
 */

#define M_AM_SEND_RSP_EVT_OK(_c) \
       ( (_c->cbType == AM_SOURCE_CB) ? SOURCE_EVENT_SEND_DATA_OK \
                                      : TARGET_EVENT_SEND_DATA_OK )

#define M_AM_SEND_RSP_EVT_NOK(_c) \
       ( (_c->cbType == AM_SOURCE_CB) ? SOURCE_EVENT_SEND_DATA_NOK \
                                      : TARGET_EVENT_SEND_DATA_NOK )

#define M_AM_SEND_DM_CLOSE_EVT(_c) \
       ( (_c->cbType == AM_SOURCE_CB) ? SOURCE_EVENT_KILL_DM_OK \
                                      : TARGET_EVENT_DM_CLOSED )

#define M_AM_MIRROR_ABORT_EVT(_c) \
       ( (_c->cbType == AM_SOURCE_CB) ? SOURCE_EVENT_MIRROR_ABORT \
                                      : TARGET_EVENT_MIRROR_ABORT )

#define M_POST_ABORT_EVENT(_c, _s, _rc) \
       { if (_rc != NFX_OK) am_post_event_abort(_c, _s); }

#define M_GET_CB_TYPE_STRING(_c) \
       ( (_c->cbType == AM_SOURCE_CB) ? "Source" \
                                      : "Target" )

#ifdef EFENCE
#define am_eee_ramAlloc      ef_eee_ramAlloc
#define am_eee_ramDealloc    ef_eee_ramDealloc
#else
#define am_eee_ramAlloc      eee_ramAlloc
#define am_eee_ramDealloc    eee_ramDealloc
#endif


/*
 * Global variables.
 */

/*
 * My own application id and EEE address port.
 */
extern int32    amAppId;
extern uint32   amAddr;

/*
 * EEE address port of the AM controler.
 */
extern uint32   ctlAddr;

/*
 * EEE address port of the data mover.
 */
extern uint32   datamoverAddr;

/*
 * EEE address port of the EFS.
 */
extern uint32   efsAddr;

/*
 * EEE address port of the EVM (FP).
 */
extern uint32   evmAddr;

/*
 * EEE address port of the EVM SSC.
 */
extern uint32   evmSscAddr;

#define AM_MAX_HASH_INDEXS      64

/*
 * Hash table for known DM handles
 */
extern address_t     dmHashID;

/*
 * Naming convention for SnapShot names
 */
extern char  amNewSnapName[];
extern uint8 amSnapNum;

/*
 * Type definitions.
 */

typedef struct _am_blk_list
{
    uint64                 blkOffSet;
    uint32                 blkCnt;
    struct _am_blk_list    *next;
} am_blk_list_t;

typedef struct _am_rcv_msg
{
    char                   *rcvMsg;
    struct _am_rcv_msg     *next;
} am_rcv_msg_t;

typedef char     am_cbType_t;
#define AM_SOURCE_CB 1
#define AM_TARGET_CB 2

/*
 * Control block for source AM agent.
 */
typedef struct
{
    am_cbType_t     cbType;       // Source or target
    esm_handle_t    amSmHdl;      // handle to AM agent (itself)
    uint32          ctrlHdl;      // handle to AM controller
	uint32			pad1;		
    uint64          dmHdl;        // handle to data mover
    uint64          efsHdl;       // handle to EFS (efs cookie)
    uint64          evmHdl;       // handle to EVM
#ifndef INCLUDE_FS_ON_SOURCE
    boolean         nextExt;      // tmp here. To avoid crash
#endif
    uint64          lclAddr;      // local IP @ (source)
    uint64        	lclPort;      // local port (source)
    uint64          rmtAddr;      // remote IP @ (target)
    uint64        	rmtPort;      // remote port (target)
    
    evm_volId_t     amVolId;      // source volume id
    evm_volId_t     tgtVolId;     // target volume id
	char			tgtFsysName[VOL_MAX_NAME]; // target volume name
    am_snapName_t   lastSnap;     // last snapshot on target
    am_snapName_t   newSnap;      // new snapshot on source

    /*
     * Field used by source only
     */
    uint32          maxDataRate;  // max data to transfer at the time
    am_blk_list_t   *lastBlk;     // last block in the list of extents
    am_blk_list_t   *blks;        // link list of blocks to be transfered
    am_blk_list_t   *currBlk;     // pointer to the current sent block
	int32			tmrHandle;

    /*
     * Field used by target only
     */
    am_rcv_msg_t    *msgQ;        // Queue of messages received from the source
                                  // the first elemet in the queue will always
                                  // point to the current request. procQReq func
                                  // will adjust the queue.
    am_rcv_msg_t    *lastMsg;     // pointer to the last entry in the queue
                                  // this is where new requested with be queued
} am_agCb_t;


typedef struct _dm_hdl_key
{
#define AM_HDL_KEY_DM_HDL   1
#define AM_HDL_KEY_AM_CB    2
    int32     hdlType;
    union
    {
        uint64    hdl;
        am_agCb_t *agt;
    } u;
} dm_hdl_key_t;

/*
 * Prototypes of global functions.
 */

// agent.c

void am_free_control_block(am_agCb_t *agtCb);

uint32 am_getHashKey(void *key);

int32 am_compareHashKey(void *ht_data, void *search_data);

void am_releaseHashObj(void *data);

void am_ag_init(void);

void am_ag_cleanup(void);

void am_ag_rcvMsg(eee_descPtr_t  edesc);

void am_ag_rcvMcastMsg(eee_descPtr_t  edesc);

void source_evmIOStat(evm_IOStat_t * stat);

void target_evmIOStat(evm_IOStat_t * stat);

// source.c

void am_free_block_list(am_agCb_t *agtCb);

// target.c

void am_free_queued_requests(am_agCb_t *agtCb);

#endif // _AM_AGENT_H_
