/*
 *=============================================================================
 * File Name   :    bcm5600-cmic.c
 *
 *  RCS: $Id: bcm5600-cmic.c,v 1.5 2002/10/23 00:12:31 tonyw Exp $
 *
 *  Description:	Copied and modified from Broadcom
 *
 *  Created    :    09-04-2001
 *
 *  Author     :    Tony Wong
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */

#include <nfx-incl.h>

#include "bcm5600-soctypes.h"
#include "bcm5600-cmic.h"
#include "bcm5600-schanmsg.h"
#include "bcm5600-error.h"

#define DBG_PRINTF_5600 if (0) printf 

extern void delay_us(uint32);

/*
 * Enable (unmask) or disable (mask) a set of CMIC interrupts.  These
 * routines should be used instead of manipulating CMIC_IRQ_MASK
 * directly, since a read-modify-write is required.  The return value is
 * the previous mask (can pass mask of 0 to just get the current mask).
 */

uint32 cmic_irq_enable(int unit, uint32 mask)
{
    uint32      oldMask;

    oldMask = cmic_pci_getreg(unit, CMIC_IRQ_MASK);
    cmic_pci_setreg(unit, CMIC_IRQ_MASK, oldMask | mask);

    return oldMask;
}

uint32 cmic_irq_disable(int unit, uint32 mask)
{
    uint32      oldMask;

    oldMask = cmic_pci_getreg(unit, CMIC_IRQ_MASK);
    cmic_pci_setreg(unit, CMIC_IRQ_MASK, oldMask & ~mask);

    return oldMask;
}

/*
 * arl_insert_5600
 *
 *   An ARL insert message is sent over the S-Channel.
 */

int arl_insert_5600(int unit, arl_entry_t *entry)
{
    schan_msg_t         schan_msg;

    schan_msg_clear(&schan_msg);
    schan_msg.arlins.header.opcode = ARL_INSERT_CMD_MSG;
    schan_msg.arlins.header.srcblk = CMIC_SCH_BLK_NUM;

    /* Fill in ARL packet data */

    memcpy(schan_msg.arlins.data, entry, sizeof (*entry));

    /* Write onto S-Channel (header word + 3 DWORDs) */

    return cmic_schan_op(unit, &schan_msg, 4, 0);
}

/*
 * Initializes the MMU with S-Channel Init SFAP (Initialize Slot Free
 * Address Pool) command.
 */

int cmic_init_sfap(int unit)
{
    int rv;
    schan_msg_t schan_msg;
    schan_msg_clear(&schan_msg);
    schan_msg.header.opcode = INIT_SFAP_MSG;
    schan_msg.header.dstblk = MMU_SCH_BLK_NUM;
    schan_msg.header.srcblk = CMIC_SCH_BLK_NUM;

    /*
     * Send init SFAP message for MMU GBP.
     */

    rv = cmic_schan_op(unit, &schan_msg, 1, 0);

    /*
     * An INIT_SFAP command starts a process in the chip that takes a
     * while to complete.  This delay should be enough to ensure it's done.
     */

	delay_us(10000);

    return rv;
}

#if defined(PCI_SWAP_REQUIRED)
uint32 pci_swap_5600(uint32 val)
{
    val = val << 16 | val >> 16;
    return (val & 0xff00ffff) >> 8 | (val & 0xffff00ff) << 8;
}
#endif 

/*
 * Get a CMIC register in PCI space.
 * Input address is relative to the base of CMIC registers.
 */

uint32 cmic_pci_getreg(int unit, uint32 addr)
{
    uint32 pci_addr, data;

    pci_addr = BCM5600_PCI_BASE_ADDR + addr;  

# if defined(PCI_SWAP_REQUIRED)
    data = pci_swap_5600(*(uint32 *) pci_addr);
# else /* !PCI_SWAP_REQUIRED */
    data = *(uint32 *) pci_addr;
# endif /* !PCI_SWAP_REQUIRED */

    DBG_PRINTF_5600("PCI%d memR(0x%x)=0x%x\n", unit, pci_addr, data);

	delay_us(50);

    return data;
}

/*
 * Set a CMIC register in PCI space.
 * Input address is relative to the base of CMIC registers.
 */

int cmic_pci_setreg(int unit, uint32 addr, uint32 data)
{
    uint32 pci_addr;

    pci_addr = BCM5600_PCI_BASE_ADDR + addr;  

    DBG_PRINTF_5600("PCI%d memW(0x%x)=0x%x\n", unit, pci_addr, data);

	delay_us(50);

# if defined(PCI_SWAP_REQUIRED)
    *(uint32 *) pci_addr = pci_swap_5600(data);
# else /* !PCI_SWAP_REQUIRED */
    *(uint32 *) pci_addr = data;
# endif /* !PCI_SWAP_REQUIRED */

    return 0;
}

/*
 * Resets the CMIC S-Channel interface. This is required when we sent
 * a message and did not receive a response after the poll count was
 * exceeded.
 */

/* private */
void cmic_schan_reset(int unit)
{
    uint32 cmicConfig = cmic_pci_getreg(unit, CMIC_CONFIG);

    /* Toggle S-Channel abort bit in CMIC_CONFIG register */

    cmic_pci_setreg(unit, CMIC_CONFIG, cmicConfig | CC_SCHAN_ABORT);
    cmic_pci_setreg(unit, CMIC_CONFIG, cmicConfig);
}
/*
 * Dump an S-Channel message for debugging
 */

void cmic_dump_schan_msg(schan_msg_t *msg, int dwc)
{
    int i;


    printf("  HEADER <CPU=%d COS=%d ECODE=%d EBIT=%d ",
           msg->header.cpu, msg->header.cos,
           msg->header.ecode, msg->header.ebit);
    printf("LEN=%d SRC=%d DST=%d OP=%d>",
           msg->header.datalen, msg->header.srcblk,
           msg->header.dstblk, msg->header.opcode);

    for (i = 0; i < dwc; i++) {
        if (i % 4 == 0)
            printf("\n ");
        printf(" DW[%2d]=0x%08x", i, msg->dwords[i]);
    }

    printf("\n");
}

/*
 * cmic_schan_op
 *
 * Writes a message of dwc_write DWORDs from msg to the S-Channel, waits
 * for the operation to complete, then reads dwc_read DWORDs from the
 * channel into msg.  If there is no return data, use dwc_read = 0.
 *
 * On platforms where it is appropriate, the S-Channel is locked during
 * the operation to prevent multiple tasks trying to use the S-Channel
 * simultaneously.
 *
 * Return value is negative on error, 0 on success.
 *
 * If INTERRUPT_DRIVEN_SCHAN is used, this routine goes to sleep until
 * an interrupt wakes up a semaphore.  Otherwise, it polls for
 * completion.  Polling is actually around 3x faster in the normal case
 * and only takes a few microseconds.  Unfortunately, the ARL insert and
 * delete response time can be unbounded due to the chip design.  In
 * this case, the polling may perform badly.  The worst case we have
 * seen is 1 msec when the switch is passing max traffic with random
 * addresses at min packet size.
 */

int cmic_schan_op(int unit, schan_msg_t *msg, int dwc_write, int dwc_read)
{
    int i, rv = SOC_E_SCHAN_TIMEOUT;
    int polls = 0;

    /* Write raw S-Channel Data: dwc_write words */
    for (i = 0; i < dwc_write; i++) {
	cmic_pci_setreg(unit, CMIC_SCHAN_MESSAGE(i), msg->dwords[i]);
    }

    /*
     * Tell CMIC to start
     */

    cmic_pci_setreg(unit, CMIC_SCHAN_CTRL, SC_MSG_START_SET);

    /* Poll for completion of S-Channel operation */

    while (!(cmic_pci_getreg(unit, CMIC_SCHAN_CTRL) & SC_MSG_DONE_TST))
	if (++polls >= SCHAN_TIMEOUT) {
	    printf("cmic_schan_op ERR: operation timed out(polls=%d)\n", polls);
	    cmic_schan_reset(unit);
	    goto done;
	}

    /* Read in data from S-Channel buffer space, if any */

    for (i = 0; i < dwc_read; i++) {
	msg->dwords[i] = cmic_pci_getreg(unit, CMIC_SCHAN_MESSAGE(i));
    }
    rv = SOC_E_NONE;

 done:
    return rv;
}

/************************************************************************/
/* Routines for reading/writing SOC internal registers			*/
/************************************************************************/

/*
 * Read an internal SOC register through S-Channel messaging buffer.
 */

int cmic_read_reg(int unit, uint32 addr, uint32 *data)
{
    schan_msg_t schan_msg;

    /*
     * Write message to S-Channel.
     */
    schan_msg_clear(&schan_msg);
    schan_msg.readcmd.header.opcode = READ_REGISTER_CMD_MSG;
    schan_msg.readcmd.header.srcblk = CMIC_SCH_BLK_NUM;
    schan_msg.readcmd.address = addr;

    /* Write header word + address DWORD, read header word + data DWORD */
    SOC_IF_ERROR_RETURN(cmic_schan_op(unit, &schan_msg, 2, 2));

    /* Check result */
    if (schan_msg.readresp.header.opcode != READ_REGISTER_ACK_MSG) {
	printf("cmic_read_reg ERROR: "
	       "invalid S-Channel reply, expected READ_REGISTER_ACK:\n");
	printf("got %x\n", schan_msg.readresp.header.opcode);
	return SOC_E_INTERNAL;
    }

    *data = schan_msg.readresp.data[0];

    return SOC_E_NONE;
}

/*
 * Write an internal SOC register through S-Channel messaging buffer.
 */

int cmic_write_reg(int unit, uint32 addr, uint32 data)
{
    schan_msg_t schan_msg;

    /*
     * Setup S-Channel command packet
     *
     * NOTE: the datalen field matters only for the Write Memory and
     * Write Register commands, where it is used only by the CMIC to
     * determine how much data to send, and is in units of bytes.
     */

    schan_msg_clear(&schan_msg);
    schan_msg.writecmd.header.opcode = WRITE_REGISTER_CMD_MSG;
    schan_msg.writecmd.header.srcblk = CMIC_SCH_BLK_NUM;
    schan_msg.writecmd.header.datalen = 4;
    schan_msg.writecmd.address = addr;
    schan_msg.writecmd.data[0] = data;

    /* Write header word + address + data DWORD */
    /* Note: The hardware does not send WRITE_REGISTER_ACK_MSG. */

    return cmic_schan_op(unit, &schan_msg, 3, 0);
}

/*
 * Write internal register for a group of ports.
 * The specified register must be a port reg (type soc_portreg).
 */

int cmic_write_reg_ports(int unit, soc_reg_t reg, pbmp_t pbmp,
			 uint32 value)
{
    soc_port_t		port;
    soc_block_t		block;

    block = soc_registers[reg].block;

    if (isPIC(block)) {
	for (port = minport(block); port <= maxport(block); port++)
	    if (pbmp & PBMP_PORT(port))
		SOC_IF_ERROR_RETURN(cmic_write_reg(unit, 
						   reg_addr(reg, port),
						   value));
    } else {	/* MMU */
	for (port = MIN_PORT_5600; port <= MAX_PORT_5600; port++)
	    if ((pbmp & PBMP_PORT(port)) && soc_ports[port].block != INVALIDb)
		SOC_IF_ERROR_RETURN(cmic_write_reg(unit,
						   reg_addr(reg, port),
						   value));
    }

    return SOC_E_NONE;
}

/*
 * Write internal register for a block or group of blocks.
 * The specified register must be a generic reg (type soc_genreg).
 * The block given may be a specific block, ANY_PIC, ANY_EPIC, or ANY_GPIC.
 */

int cmic_write_reg_blocks(int unit, soc_reg_t reg, soc_block_t block,
			  uint32 value)
{
    soc_block_t		block_min, block_max;

    if (isPIC(block)) {
	block_min = minblock(block);
	block_max = maxblock(block);
    } else
	block_min = block_max = block;

    for (block = block_min; block <= block_max; block++)
	SOC_IF_ERROR_RETURN(cmic_write_reg(unit,
					   reg_addr(reg, block),
					   value));

    return SOC_E_NONE;
}

/************************************************************************/
/* Routines for reading/writing all tables (except no write to ARL)	*/
/************************************************************************/

/*
 * Read a memory internal to the SOC.
 * NOTE: GBP/CBP memory is only accessible when in DEBUG mode.
 */

int cmic_read_memory(int unit,
		     soc_mem_t mem,
		     int copyno,
		     int index,
		     void *entry_data)
{
    schan_msg_t schan_msg; 

    int entry_dw = BYTES2WORDS(get_mem_numbytes(mem));

    /* Setup S-Channel command packet */
    schan_msg_clear(&schan_msg);
    schan_msg.readcmd.header.opcode = READ_MEMORY_CMD_MSG;
    schan_msg.readcmd.header.srcblk = CMIC_SCH_BLK_NUM;
    if (mem == GBPm)
		schan_msg.readcmd.address = mem_addr(mem, index);
    else
		schan_msg.readcmd.address = mem_addr(mem, copyno, index);
    /*
     * Write onto S-Channel "memory read" command packet consisting of header
     * word + address word, and read back header word + entry_dw data words.
     */

    SOC_IF_ERROR_RETURN(cmic_schan_op(unit, &schan_msg, 2, 1 + entry_dw));
    /* Check result */
    if (schan_msg.readresp.header.opcode != READ_MEMORY_ACK_MSG) {
	printf("cmic_read_memory ERROR: "
	       "invalid S-Channel reply, expected READ_MEMORY_ACK:\n");
	return SOC_E_INTERNAL;
    }
    memcpy(entry_data,
	   schan_msg.readresp.data,
	   entry_dw * sizeof (uint32));
	
    return SOC_E_NONE;
}

/*
 * Write a memory internal to the SOC.
 * NOTE: GBP/CBP memory is only accessible when in DEBUG mode.
 */

int cmic_write_memory(int unit,
		      soc_mem_t mem,
		      int copyno,	/* Use COPYNO_ALL for all */
		      int index,
		      void *entry_data)
{
    int entry_dw = BYTES2WORDS(get_mem_numbytes(mem));
    schan_msg_t schan_msg;
    int copyno_start, copyno_end;

    /*
     * Setup S-Channel command packet
     *
     * NOTE: the datalen field matters only for the Write Memory and
     * Write Register commands, where it is used only by the CMIC to
     * determine how much data to send, and is in units of bytes.
     */

    schan_msg_clear(&schan_msg);
    schan_msg.writecmd.header.opcode = WRITE_MEMORY_CMD_MSG;
    schan_msg.writecmd.header.srcblk = CMIC_SCH_BLK_NUM;
    schan_msg.writecmd.header.datalen = entry_dw * sizeof(uint32);

    memcpy(schan_msg.writecmd.data, entry_data, entry_dw * sizeof (uint32));

	if (copyno == COPYNO_ALL) {
		copyno_start = 0;
		copyno_end = soc_memories[mem].nBlocks - 1;
	}
	else 
		copyno_start = copyno_end = copyno;

    /* Write to one or all copies of the memory */

    for (copyno = copyno_start; copyno <= copyno_end; copyno++) {
	if (mem == GBPm)
	    schan_msg.writecmd.address = mem_addr(mem, index);
	else
	    schan_msg.writecmd.address = mem_addr(mem, copyno, index);

	/* Write header + address + entry_dw data DWORDs */
	/* Note: The hardware does not send WRITE_MEMORY_ACK_MSG. */

	SOC_IF_ERROR_RETURN(cmic_schan_op(unit, &schan_msg, 2 + entry_dw, 0));
    }

    return SOC_E_NONE;
}

/************************************************************************
 * Routines for reading/writing sorted tables except ARL,		*
 * and searching sorted tables including the ARL.			*
 * Sorted tables are:  IRULE, L3, MCAST, PVLAN				*
 ************************************************************************/

/*
 * cmic_insert_memory
 *
 *   Only used for ARL Insert 
 *
 *   If COPYNO_ALL is specified, the insertion is repeated for each copy
 *	of the table.
 *
 *   Returns  -- index of insertion point
 * 		 if -1 on error
*/

int cmic_insert_memory(int unit,
		       soc_mem_t mem,
		       int copyno,
		       void *entry_data)
{
    int rv = SOC_E_NOT_IMPL;

    if (copyno == COPYNO_ALL) {
	if (mem == ARL_m) {
	    /*
	     * Let the hardware do ARL inserts.
	     */
	    return arl_insert_5600(unit, entry_data);
	}
    }
    return rv;
}

/*
 * Enter or leave DEBUG mode by sending appropriate S-Channel message.
 * Returns previous enable state, or -1 on error.
 */

int cmic_debug_memory(int unit, int enable)
{
    schan_msg_t schan_msg;
    int	msg, rv = 0;

    if (enable)
	msg = ENTER_DEBUG_MODE_MSG;
    else 
	msg = EXIT_DEBUG_MODE_MSG;

    if (msg >= 0) {
	schan_msg_clear(&schan_msg);
	schan_msg.header.opcode = msg;
	schan_msg.header.dstblk = MMU_SCH_BLK_NUM;
	schan_msg.header.srcblk = CMIC_SCH_BLK_NUM;

	if ((rv = cmic_schan_op(unit, &schan_msg, 1, 0)) >= 0)
	    return rv;
	else
	    return (-1);
    }

    return (rv < 0) ? rv : enable;
}

/*
 * Function:
 *	cmic_write_miim
 * Purpose:
 *	Write a value to a MIIM register.
 * Parameters:
 *	unit - StrataSwitch Unit #.
 *	phy_id - Phy ID to write (MIIM address)
 *	phy_reg_addr - PHY register to write
 *	phy_wr_date - Data to write.
 * Returns:
 *	SOC_E_XXX
 * Notes:
 * 	Temporarily disables auto link scan if it was enabled.
 *
 * 	On platforms where it is appropriate, the MIIM registers are locked
 * 	during the operation to prevent multiple tasks trying to access PHY
 * 	registers simultaneously.
 *
 * 	See comment in schan_op about interrupt driven vs. polled mode.
 */
int 
cmic_write_miim(int unit, uint8 phy_id,
		uint8 phy_reg_addr, uint16 phy_wr_data)
{

    int rv;
    int polls = 0;

    DBG_PRINTF_5600("cmic_write_miim: addr=0x%02x id=0x%02x data=0x%04x\n",
					phy_reg_addr, phy_id, phy_wr_data);
#if 0
	cmic_link_scan_pause(unit);
#endif
    /* Write parameter register and tell CMIC to start */

    cmic_pci_setreg(unit, CMIC_MIIM_PARAM,
		    (uint32) phy_reg_addr << 24 |
		    (uint32) phy_id << 16 |
		    (uint32) phy_wr_data);

    cmic_pci_setreg(unit, CMIC_SCHAN_CTRL, SC_MIIM_WR_START_SET);

    /* Poll for completion of MIIM operation */

    while (!(cmic_pci_getreg(unit, CMIC_SCHAN_CTRL) & SC_MIIM_OP_DONE_TST))
	if (++polls >= MIIM_TIMEOUT) {
	    printf("cmic_write_miim ERROR: timeout (addr=0x%02x id=0x%02x data=0x%04x)\n",
		   phy_reg_addr, phy_id, phy_wr_data);
	    rv = SOC_E_MII_TIMEOUT;
	    goto done;
	}

    rv = SOC_E_NONE;

done:
#if 0
	cmic_link_scan_enable(unit);
#endif
    return rv;

}

/*
 * cmic_read_miim
 *
 * Reads a value from a PHY register.
 * Stores the 16-bit read data in the phy_rd_data pointer.
 * Temporarily disables auto link scan if it was enabled.
 *
 * On platforms where it is appropriate, the MIIM registers are locked
 * during the operation to prevent multiple tasks trying to access PHY
 * registers simultaneously.
 *
 * Return value is negative on error, 0 on success.
 */
int cmic_read_miim(int unit, uint8 phy_id,
		   uint8 phy_reg_addr, uint16 *phy_rd_data)
{
    int rv;
    int polls = 0;

    DBG_PRINTF_5600("cmic_read_miim: addr=0x%02x id=0x%02x\n", 
					phy_reg_addr, phy_id);
#if 0
	cmic_link_scan_pause(unit);
#endif
    /* Write parameter register and tell CMIC to start */

    cmic_pci_setreg(unit, CMIC_MIIM_PARAM,
		    (uint32) phy_reg_addr << 24 |
		    (uint32) phy_id << 16);

    cmic_pci_setreg(unit, CMIC_SCHAN_CTRL, SC_MIIM_RD_START_SET);

    /* Poll for completion of MIIM operation */

    while (!(cmic_pci_getreg(unit, CMIC_SCHAN_CTRL) & SC_MIIM_OP_DONE_TST))
	if (++polls >= MIIM_TIMEOUT) {
	    printf("cmic_read_miim ERROR: timeout (addr=0x%02x id=0x%02x)\n",
		   phy_reg_addr, phy_id);
	    rv = SOC_E_MII_TIMEOUT;
	    goto done;
	}

    *phy_rd_data = (uint16) cmic_pci_getreg(unit, CMIC_MIIM_READ_DATA);

    rv = SOC_E_NONE;

done:
#if 0
	cmic_link_scan_enable(unit);
#endif
    return rv;
}

/*
 * Function:
 *  cmic_link_scan_pause
 * Purpose:
 *  Pauses link scanning, without disabling it.  This call is
 *  used to temporarily pause scanning.
 * Parameters:
 *  unit - StrataSwitch unit #.
 * Returns:
 *  Nothing
 * Notes:
 *  Tracks nests calls, so upper layer software may simple call
 *  cmic_link_scan_pause whenever it needs to be sure link scanning
 *  is not running as long as it is paired with a call to
 *  cmic_link_scan_continue.
 */
void
cmic_link_scan_pause(int unit)
{
    cmic_pci_setreg(unit, CMIC_SCHAN_CTRL, SC_MIIM_LINK_SCAN_EN_CLR);
    while (cmic_pci_getreg(unit, CMIC_SCHAN_CTRL) & SC_MIIM_SCAN_BUSY_TST)
        ;
}

/*
 * Function:
 *  cmic_link_scan_continue
 * Purpose:
 *  Continued link scanning after it has been paused.
 * Parameters:
 *  unit - StrataSwitch unit #.
 * Returns:
 *  Nothing
 * Notes:
 *  Must always be paired with a call to cmic_link_scan_pause.
 */
void
cmic_link_scan_enable(int unit)
{
    cmic_pci_setreg(unit, CMIC_MIIM_PARAM, (uint32) 0x01 << 24);
    cmic_pci_setreg(unit, CMIC_SCHAN_CTRL, SC_MIIM_LINK_SCAN_EN_SET);
}

/*
 * Function:
 *      cmic_link_scan
 * Purpose:
 *      Set ports to scan in CMIC.
 *
 * Parameters:
 *      unit - StrataSwich Unit #
 *      mii_pbm - Port bit map of ports to scan with MIIM registers
 *      direct_pbm - Port bit map of ports to scan using NON MII.
 *
 * Returns:
 *      SOC_E_XXX
 */
int
cmic_link_scan(int unit, pbmp_t mii_pbm, pbmp_t direct_pbm)
{
    uint32      cmic_config;

    /*
     * Non MIIM scanning is only supported on 10/100 ports.
     */
    if (direct_pbm & PBMP_FE_ALL) {
        return(SOC_E_NOT_IMPL);
    }

    /*
     * The LINK_SCAN_GIG control affects ALL ports. Thus, all ports
     * being scanned by H/W must be either MIIM scanned or scanned
     * using the direct connection.
     */
    if ((mii_pbm & PBMP_GE_ALL) && (direct_pbm & PBMP_GE_ALL)) {
        return(SOC_E_NOT_IMPL);
    }

    /* Check if disabling port scanning */

    cmic_config = cmic_pci_getreg(unit, CMIC_CONFIG);
    cmic_config &= ~(CC_LINK_STAT_EN|CC_LINK_SCAN_GIG);

    if (mii_pbm || direct_pbm) {
        cmic_config    |= CC_LINK_STAT_EN;
    }

    /*
     * If using MIIM scanning for GIGs (NOT using NON-MIIM) and enabling
     * scanning, then enable link scan GIGs.
     */

    if (mii_pbm & PBMP_GE_ALL) {
        /*
         * If scanning GIG ports using the MIIM interface, then LINK_SCAN_GIG
         * must be disabled.
         */
        cmic_config &= ~CC_LINK_SCAN_GIG;
    } else if (direct_pbm & PBMP_GE_ALL) {
        cmic_config |= CC_LINK_SCAN_GIG;
    }

    cmic_pci_setreg(unit, CMIC_SCAN_PORTS, mii_pbm);
    cmic_pci_setreg(unit, CMIC_CONFIG, cmic_config);

    return(SOC_E_NONE);
}
