/*
 *=============================================================================
 * File Name   :    bcm5600-trunk.c
 *
 *  RCS: $Id: bcm5600-trunk.c,v 1.1.1.1 2001/12/20 04:53:01 tonyw Exp $		Copied and modified from Broadcom
 *
 *  Description:
 *
 *  Created    :    09-04-2001
 *
 *  Author     :    Tony Wong
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */

#include <nfx-incl.h>

#include "bcm5600-soctypes.h"
#include "bcm5600-cmic.h"
#include "bcm5600-schanmsg.h"
#include "bcm5600-error.h"
#include "bcm5600-port.h"

#define BCM_RTAG_SRCMAC         0x01
#define BCM_RTAG_DSTMAC         0x02
#define BCM_RTAG_SRCDSTMAC      0x03
#define BCM_RTAG_SRCIP          0x04
#define BCM_RTAG_DSTIP          0x05
#define BCM_RTAG_SRCDSTIP       0x06

extern int bcm_trunk_init(int unit);

extern int bcm_trunk_get(int unit, trunk_id_t tid, pbmp_t *pbmp);
extern int bcm_trunk_get_peers(int unit, soc_port_t port, pbmp_t *pbmp);

extern int bcm_trunk_bitmap_reduce(int unit, pbmp_t *pbmp_ptr);
extern int bcm_trunk_bitmap_expand(int unit, pbmp_t *pbmp_ptr);

static uint32 bcm_zeroes[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

/*
 * Function: bcm_trunk_init
 *
 * Description:
 *	Initializes the trunk tables
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *	None.
 */
int 
bcm_trunk_init(int unit)
{
    trunk_id_t		tid;

    for (tid = TRUNK_ID_MIN; tid <= TRUNK_ID_MAX; tid++) {
	if (cmic_write_memory(unit, TTABLE_TBMAPm, COPYNO_ALL, tid,
			      bcm_zeroes) < 0)
	    return BCM_E_INTERNAL;

	if (cmic_write_memory(unit, TTABLE_TTRm, COPYNO_ALL, tid,
			      bcm_zeroes) < 0)
	    return BCM_E_INTERNAL;
    }

    return BCM_E_NONE;
}

/*
 * Function: _bcm_ttr_port_find (Internal Function)
 *
 * Description:
 *	Returns the index of a port in a ttr_entry_t,
 *	-1 if not found.
 */

static int
_bcm_ttr_port_find(ttr_entry_t *ttrentry, soc_port_t port)
{
    if (ttrentry->tp0 == port) return 0;
    if (ttrentry->tp1 == port) return 1;
    if (ttrentry->tp2 == port) return 2;
    if (ttrentry->tp3 == port) return 3;
    if (ttrentry->tp4 == port) return 4;
    if (ttrentry->tp5 == port) return 5;
    if (ttrentry->tp6 == port) return 6;
    if (ttrentry->tp7 == port) return 7;
    return -1;
}

/*
 * Function: _bcm_ttr_port_add (Internal Function)
 *
 * Description:
 *	Adds a port to a ttr_entry_t and updates the tg_size field.
 */

static void
_bcm_ttr_port_add(ttr_entry_t *ttrentry, soc_port_t port)
{
    switch (ttrentry->tg_size) {
    case 0: ttrentry->tp0 = port; break;
    case 1: ttrentry->tp1 = port; break;
    case 2: ttrentry->tp2 = port; break;
    case 3: ttrentry->tp3 = port; break;
    case 4: ttrentry->tp4 = port; break;
    case 5: ttrentry->tp5 = port; break;
    case 6: ttrentry->tp6 = port; break;
    case 7: ttrentry->tp7 = port; break;
    }

    ttrentry->tg_size++;
}

/*
 * Function: _bcm_ttr_port_remove (Internal Function)
 *
 * Description:
 *	Removes a port from a ttr_entry_t by index, repacking
 *	the remaining ports, and updating the tg_size field.
 */

static void
_bcm_ttr_port_remove(ttr_entry_t *ttrentry, int ttr_index)
{
    switch (ttr_index) {
    case 0: ttrentry->tp0 = ttrentry->tp1; /* FALLTHROUGH */
    case 1: ttrentry->tp1 = ttrentry->tp2; /* FALLTHROUGH */
    case 2: ttrentry->tp2 = ttrentry->tp3; /* FALLTHROUGH */
    case 3: ttrentry->tp3 = ttrentry->tp4; /* FALLTHROUGH */
    case 4: ttrentry->tp4 = ttrentry->tp5; /* FALLTHROUGH */
    case 5: ttrentry->tp5 = ttrentry->tp6; /* FALLTHROUGH */
    case 6: ttrentry->tp6 = ttrentry->tp7;
    }

    ttrentry->tp7 = 0;
    ttrentry->tg_size--;
}

/*
 * Function: bcm_trunk_get
 *
 * Description:
 *	Returns a bitmap of the members of the trunk
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *      tid - Trunk Id must be 0 to 7
 *	pbmp - (Output) Bitmap of ports that are members of the trunk
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 */
int
bcm_trunk_get(int unit, trunk_id_t tid, pbmp_t *pbmp)
{
    tbmap_entry_t		tbmapentry;

    if (cmic_read_memory(unit, TTABLE_TBMAPm, 0, tid, &tbmapentry) < 0)
	return (BCM_E_INTERNAL);

    *pbmp = tbmapentry.tbmap;

    return (BCM_E_NONE);
}

/*
 * Function: bcm_trunk_get_peers
 *
 * Description:
 *	Return a bitmap of all ports aggregated with the specified port.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	port - Port to operate on
 *	pbmp - Place to store bitmap of returned ports
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *	The bitmap returned always includes the port itself.
 */

int
bcm_trunk_get_peers(int unit, soc_port_t port, pbmp_t *pbmp)
{
    tbmap_entry_t		tbmapentry;
    ptable_entry_t		entry;

    if (cmic_read_memory(unit, PTABLE_m, 0, port, &entry) < 0)
	return (BCM_E_INTERNAL);

    if (! entry.trunk)
	*pbmp = PBMP_PORT(port);
    else if (cmic_read_memory(unit, TTABLE_TBMAPm, 0,
			      entry.tgid, &tbmapentry) < 0)
	return (BCM_E_INTERNAL);
    else
	*pbmp = tbmapentry.tbmap;

    return (BCM_E_NONE);
}

/*
 * Function: bcm_trunk_bitmap_reduce
 *
 * Description:
 *	Given a port bitmap, wherever multiple ports are trunked
 *	together, remove all but the first trunked port from the bitmap.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	pbmp_ptr - Input/output port bitmap
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *	Port bitmaps in the VTABLE and PTABLE must contain only one of
 *	the trunk ports when multiple ports are aggregated in a trunk.
 *	This is so DLF flooding, etc. goes out only one trunk port.
 */

int
bcm_trunk_bitmap_reduce(int unit, pbmp_t *pbmp_ptr)
{
    tbmap_entry_t		tbmapentry;
    trunk_id_t			tid;
    pbmp_t			tports;

    for (tid = TRUNK_ID_MIN; tid <= TRUNK_ID_MAX; tid++) {
	if (cmic_read_memory(unit, TTABLE_TBMAPm, 0, tid, &tbmapentry) < 0)
	    return (BCM_E_INTERNAL);

	/*
	 * If any trunk ports are in the bitmap, remove all but the
	 * lowest-numbered port in the trunk.
	 */

	tports = tbmapentry.tbmap;

	if ((*pbmp_ptr & tports) != PBMP_NONE)
	    *pbmp_ptr &= ~(tports & (tports - 1));
    }

    return (BCM_E_NONE);
}

/*
 * Function: bcm_trunk_bitmap_expand
 *
 * Description:
 *	Given a port bitmap, if any of the ports are in a trunk,
 *	add ALL of the ports in that trunk to the bitmap.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	pbmp_ptr - Input/output port bitmap
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *	Port bitmaps in the VTABLE and PTABLE must contain only one of
 *	the trunk ports when multiple ports are aggregated in a trunk.
 *	This is so DLF flooding, etc. goes out only one trunk port.
 */

int
bcm_trunk_bitmap_expand(int unit, pbmp_t *pbmp_ptr)
{
    tbmap_entry_t		tbmapentry;
    trunk_id_t			tid;
    pbmp_t			tports;

    for (tid = TRUNK_ID_MIN; tid <= TRUNK_ID_MAX; tid++) {
	if (cmic_read_memory(unit, TTABLE_TBMAPm, 0, tid, &tbmapentry) < 0)
	    return (BCM_E_INTERNAL);

	/*
	 * If any trunk ports are in the bitmap, add all trunk peers to
	 * the bitmap.
	 */

	tports = tbmapentry.tbmap;

	if ((*pbmp_ptr & tports) != PBMP_NONE)
	    *pbmp_ptr |= tports;
    }

    return (BCM_E_NONE);
}
