/*
 *=============================================================================
 * File Name   :    bcm5632-phy.c
 *
 *  RCS: $Id: bcm5632-phy.c,v 1.2 2002/10/09 20:24:31 tonyw Exp $
 *
 *  Description:	Copied and modified from Broadcom
 *
 *  Created    :    09-04-2001
 *
 *  Author     :    Tony Wong
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */
#include <nfx-incl.h>

#include "bcm5632-soctypes.h"
#include "bcm5632-cmic.h"
#include "bcm5632-error.h"
#include "bcm5632-port.h"
#include "bcm5632-phy.h"

/* phy mapping for port 0-11 */
uint8 PORT_TO_PHY_ID[BCM5632_NUM_DEV][NUM_GE_PORT_5632] = 
{ 
  {	0x00, 0x01, 0x02, 0x03, 0x06, 0x07, 0x04, 0x05, 0x1c, 0x1d, 0x1e, 0x1f },
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
}; 

/*
 * Function: 	
 *	phy_ge_enable_set5632
 *
 * Purpose:	
 *	Enable or disable the physical interface.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	enable - Boolean, true = enable PHY, false = disable.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_enable_set5632(int unit, uint32 port, int enable)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG, 
											&mii_ctrl));
    mii_ctrl = (mii_ctrl & ~MII_CTRL_IP) | (enable ? 0 : MII_CTRL_IP);
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG, 
											 mii_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_enable_get5632
 *
 * Purpose:	
 *	Enable or disable the physical interface.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	enable - (OUT) Boolean, true = enable PHY, false = disable.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_enable_get5632(int unit, uint32 port, int *enable)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG, 
											&mii_ctrl));
    *enable = (mii_ctrl & MII_CTRL_IP) ? FALSE : TRUE;
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_init5632
 *
 * Purpose:	
 *	Initialize the PHY (MII mode) to a known good state.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *
 * Returns:	
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level.
 */
int
phy_ge_init5632(int unit, uint32 port)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl, mii_gb_ctrl;

    mii_ctrl = MII_CTRL_FD | MII_CTRL_SS_1000 | MII_CTRL_AE | MII_CTRL_RAN;
    mii_gb_ctrl = MII_GB_CTRL_ADV_1000FD | MII_GB_CTRL_PT;

    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_GB_CTRL_REG, 
					mii_gb_ctrl));
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG,
                                        mii_ctrl));

    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_link_get5632
 *
 * Purpose:	
 *	Determine the current link up/down status
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	link - (OUT) Boolean, true indicates link established.
 *
 * Returns:
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level.
 */
int
phy_ge_link_get5632(int unit, uint32 port, int *link)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_stat;
    
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG,
						&mii_stat));
    *link = (mii_stat & MII_STAT_LA) ? TRUE : FALSE;
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	_phy_auto_negotiate_gcd5632 (greatest common denominator).
 *
 * Purpose:	
 *	Determine the current greatest common denominator between 
 *	two ends of a link
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	speed- (OUT) greatest common speed.
 *	duplex-(OUT) greatest common duplex.
 *	link - (OUT) Boolean, true indicates link established.
 *
 * Returns:	
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level.
 */
static int
_phy_auto_negotiate_gcd5632(int unit, uint32 port, int *speed, int *duplex)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    int		t_speed = 0;
	int		t_duplex = 0;
    uint16	mii_ana, mii_anp, mii_stat;
    uint16	mii_gb_stat, mii_esr, mii_gb_ctrl;

    mii_gb_stat = 0;			/* Start off 0 */
    mii_gb_ctrl = 0;			/* Start off 0 */

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ANA_REG, 
											&mii_ana));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ANP_REG, 
											&mii_anp));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG, 
											&mii_stat));
    if (mii_stat & MII_STAT_ES) {	/* Supports extended status */
		/*
	 	 * If the PHY supports extended status, check if it is 1000MB capable.
	 	 * If it is, check the 1000Base status register to see if 1000MB
	 	 * negotiated.
	 	 */
		SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ESR_REG, 
					   						    &mii_esr));
		if (mii_esr & (MII_ESR_1000_X_FD | MII_ESR_1000_X_HD | 
		       		   MII_ESR_1000_T_FD | MII_ESR_1000_T_HD)) {
	    	SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, 
									MII_GB_STAT_REG, &mii_gb_stat));
	    	SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, 
									MII_GB_CTRL_REG, &mii_gb_ctrl));
		}
    }
    /*
     * At this point, if we did not see Gig status, one of mii_gb_stat or 
     * mii_gb_ctrl will be 0. This will cause the first 2 cases below to 
     * fail and fall into the default 10/100 cases.
     */
    mii_ana &= mii_anp;

    if ((mii_gb_ctrl & MII_GB_CTRL_ADV_1000FD) &&
		(mii_gb_stat & MII_GB_STAT_LP_1000FD)) {
		t_speed  = 1000;
		t_duplex = 1;
    } 
	else 
	if ((mii_gb_ctrl & MII_GB_CTRL_ADV_1000HD) &&
        (mii_gb_stat & MII_GB_STAT_LP_1000HD)) {
		t_speed  = 1000;
		t_duplex = 0;
    } 
    if (speed)	*speed = t_speed;
    if (duplex)	*duplex = t_duplex;
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_duplex_set5632
 *
 * Purpose:	
 *	Set the current duplex mode (forced).
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	duplex -Boolean, true indicates full duplex, false indicates half.
 *
 * Returns:	
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level. Autonegotiation is 
 *	not manipulated. 
 */
int
phy_ge_duplex_set5632(int unit, uint32 port, int duplex)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    if (duplex) {
		mii_ctrl |= MII_CTRL_FD;
    } 
	else {
		mii_ctrl &= ~MII_CTRL_FD;
    }
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG,
											 mii_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_duplex_get5632
 *
 * Purpose:	
 *	Get the current operating duplex mode. If autoneg is enabled, 
 *	then operating mode is returned, otherwise forced mode is returned.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	duplex - (OUT) Boolean, true indicates full duplex, false 
 *		indicates half.
 *
 * Returns:	
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level. Autonegotiation is 
 *	not manipulated. 
 */
int
phy_ge_duplex_get5632(int unit, uint32 port, int *duplex)
{
    int		rv;
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl, mii_stat;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG,
											&mii_stat));
    if (mii_ctrl & MII_CTRL_AE) {	/* Auto-negotiation enabled */
		if (!(mii_stat & MII_STAT_AN_DONE)) { /* Auto-neg NOT complete */
	    	*duplex = 0;
	    	return(SOC_E_NONE);
		}
		if ((rv = _phy_auto_negotiate_gcd5632(unit, port, NULL, duplex)) < 0) {
	    	return(rv);
		}
    } 
	else {	/* Auto-negotiation disabled */
		*duplex = (mii_ctrl & MII_CTRL_FD) ? 1 : 0;
    }
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_speed_set5632
 *
 * Purpose:	
 *	Set the current operating speed (forced).
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	duplex - (OUT) Boolean, true indicates full duplex, false 
 *		indicates half.
 *
 * Returns:	
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level. Autonegotiation is 
 *	not manipulated. 
 */
int
phy_ge_speed_set5632(int unit, uint32 port, int speed)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    mii_ctrl &= ~(MII_CTRL_SS_LSB | MII_CTRL_SS_MSB);
    switch(speed) {
    	case 10:
			mii_ctrl |= MII_CTRL_SS_10;
			break;
    	case 100:
			mii_ctrl |= MII_CTRL_SS_100;
			break;
    	case 1000:	
			mii_ctrl |= MII_CTRL_SS_1000;
			break;
    	default:
			return(SOC_E_NOT_IMPL);
    }
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG,
											 mii_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_speed_get5632
 *
 * Purpose:	
 *	Get the current operating speed. If autoneg is enabled, 
 *	then operating mode is returned, otherwise forced mode is returned.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	duplex - (OUT) Boolean, true indicates full duplex, false 
 *		indicates half.
 *
 * Returns:	
 *	SOC_E_XXXX
 *
 * Notes: 
 *	No synchronization performed at this level. Autonegotiation is 
 *	not manipulated. 
 */
int
phy_ge_speed_get5632(int unit, uint32 port, int *speed)
{
    int		rv;
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl, mii_stat;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG,
											&mii_stat));

    if (mii_ctrl & MII_CTRL_AE) {	/* Auto-negotiation enabled */
		if (!(mii_stat & MII_STAT_AN_DONE)) { /* Auto-neg NOT complete */
	    	*speed = 0;
	    	rv = SOC_E_AUTONEG_BUSY;
		} 
		else {
	    	rv = _phy_auto_negotiate_gcd5632(unit, port, speed, NULL);
		}
    } 
	else {	/* Auto-negotiation disabled */
		/*
	 	 * Simply pick up the values we force in CTRL register.
	 	 */
		switch(MII_CTRL_SS(mii_ctrl)) {
			case MII_CTRL_SS_10:
	    		*speed = 10;
	    		break;
			case MII_CTRL_SS_100:
	    		*speed = 100;
	    		break;
			case MII_CTRL_SS_1000:
	    		*speed = 1000;
	    		break;
			default:			/* Just pass error back */
	    		return(SOC_E_NOT_IMPL);
		}
		rv = SOC_E_NONE;
    }
    return(rv);
}

/*
 * Function: 	
 *	phy_ge_an_set5632
 *
 * Purpose:	
 *	Enable or disabled auto-negotiation on the specified port.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	an   - Boolean, if true, auto-negotiation is enabled 
 *		(and/or restarted). If false, autonegotiation is disabled.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_an_set5632(int unit, uint32 port, int an)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    if (an) {
		mii_ctrl |= MII_CTRL_AE | MII_CTRL_RAN;
    } 
	else {
		mii_ctrl &= ~(MII_CTRL_AE | MII_CTRL_RAN);
    }
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG,
											 mii_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_an_get5632
 *
 * Purpose:	
 *	Get the current auto-negotiation status (enabled/busy)
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	an   - (OUT) if true, auto-negotiation is enabled.
 *	an_done - (OUT) if true, auto-negotioation is complete. This
 *	     	value is undefined if an == false.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_an_get5632(int unit, uint32 port, int *an, int *an_done)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl, mii_stat;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
						&mii_ctrl));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG,
						&mii_stat));
    *an = (mii_ctrl & MII_CTRL_AE) ? TRUE : FALSE;
    *an_done = (mii_stat & MII_STAT_AN_DONE) ? TRUE : FALSE;
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_adv_local_set5632
 *
 * Purpose:	
 *	Set the current advertisement for auto-negotiation.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	mode - Port mode mask indicating supported options/speeds.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int 
phy_ge_adv_local_set5632(int unit, uint32 port, uint32 mode)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_adv, mii_gb_ctrl;

    mii_adv	= MII_ANA_ASF_802_3;
    mii_gb_ctrl = 0;

    if (mode & SOC_PM_1000MB_HD) mii_gb_ctrl |= MII_GB_CTRL_ADV_1000HD;
    if (mode & SOC_PM_1000MB_FD) mii_gb_ctrl |= MII_GB_CTRL_ADV_1000FD;

    if (mode & SOC_PM_PAUSE) {
		mii_adv |= MII_ANA_PAUSE;
    }
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_ANA_REG, 
						mii_adv));
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_GB_CTRL_REG, 
						mii_gb_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_adv_local_get5632
 *
 * Purpose:	
 *	Get the current advertisement for auto-negotiation.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	mode - (OUT) Port mode mask indicating supported options/speeds.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_adv_local_get5632(int unit, uint32 port, uint32 *mode)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ana, mii_gb_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ANA_REG, 
						&mii_ana));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_GB_CTRL_REG, 
				        &mii_gb_ctrl));
    *mode = 0;

    if (mii_ana & MII_ANA_PAUSE) {
		*mode |= SOC_PM_PAUSE_TX;
        *mode |= SOC_PM_PAUSE_RX;
    }
    /* GE Specific values */

    if (mii_gb_ctrl & MII_GB_CTRL_ADV_1000FD)
	 	*mode |= SOC_PM_1000MB_FD;
    if (mii_gb_ctrl & MII_GB_CTRL_ADV_1000HD)
		*mode |= SOC_PM_1000MB_HD;

    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_adv_remote_get5632
 *
 * Purpose:	
 *	Get partnets current advertisement for auto-negotiation.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	mode - (OUT) Port mode mask indicating supported options/speeds.
 *
 * Returns:
 *	SOC_E_XXXX
 */
int
phy_ge_adv_remote_get5632(int unit, uint32 port, uint32 *mode)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl, mii_stat, mii_gb_stat, mii_anp;

    *mode = 0;

    /* ORDER: must read control, then stat, then anp */

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG, 
						&mii_ctrl));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG, 
						&mii_stat));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ANP_REG, 
						&mii_anp));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, port, MII_GB_STAT_REG,
				        &mii_gb_stat));

    if (!(mii_ctrl & MII_CTRL_AE)) {
		return(SOC_E_AUTONEG_DISABLED);
    } 
	else 
	if (!(mii_stat & MII_STAT_AN_DONE)) {
		return(SOC_E_AUTONEG_BUSY);
    }
    if (mii_gb_stat & MII_GB_STAT_LP_1000FD)
		*mode |= SOC_PM_1000MB_FD;
    if (mii_gb_stat & MII_GB_STAT_LP_1000HD)
		*mode |= SOC_PM_1000MB_HD;
    if (mii_anp & MII_ANA_PAUSE) 
        *mode |= SOC_PM_PAUSE;

    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_lb_set5632
 *
 * Purpose:	
 *	Set the local PHY loopback mode.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	loopback - Boolean: true = enable loopback, false = disable.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_lb_set5632(int unit, uint32 port, int enable)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
						&mii_ctrl));

    mii_ctrl &= ~MII_CTRL_LE;
    mii_ctrl |= enable ? MII_CTRL_LE : 0;

    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG,
						mii_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_lb_get5632
 *
 * Purpose:	
 *	Get the local PHY loopback mode.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	loopback - (OUT) Boolean: true = enable loopback, false = disable.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_lb_get5632(int unit, uint32 port, int *enable)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
						&mii_ctrl));
    *enable = (mii_ctrl & MII_CTRL_LE) ? 1 : 0;
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_ge_interface_set5632
 *
 * Purpose:
 *  Set the current operating mode of the PHY (10B or MII/GMII).
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mii - TRUE --> MII/GMII, FALSE --> 10B.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_ge_interface_set5632(int unit, uint32 port, int mii)
{
    uint8   phy_id = PORT_TO_PHY_ID[unit][port];
    uint16  mii_ecr;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ECR_REG, 
											&mii_ecr));
    mii_ecr &= ~MII_ECR_10B;
    mii_ecr |= mii ? 0 : MII_ECR_10B;
    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_ECR_REG, 
											 mii_ecr));
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_ge_interface_get5632
 *
 * Purpose:
 *  Get the current operating mode of the PHY (10B or MII/GMII).
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mii - (OUT) TRUE --> MII/GMII, FALSE --> 10B.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_ge_interface_get5632(int unit, uint32 port, int *mii)
{
    uint8   phy_id = PORT_TO_PHY_ID[unit][port];
    uint16  mii_ecr;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ECR_REG, 
											&mii_ecr));
    *mii = (mii_ecr & MII_ECR_10B) ? FALSE : TRUE;
    return(SOC_E_NONE);
}

/*
 * Function: 	
 *	phy_ge_ability5632
 *
 * Purpose:	
 *	Get the abilities of the local gigabit phy.
 *
 * Parameters:
 *	unit - StrataSwitch unit #.
 *	port - StrataSwitch port #. 
 *	mode - (OUT) Port mode mask indicating supported options/speeds.
 *
 * Returns:	
 *	SOC_E_XXXX
 */
int
phy_ge_ability5632(int unit, uint32 port, uint32 *mode)
{
    uint8	phy_id = PORT_TO_PHY_ID[unit][port];
    uint16	mii_esr, mii_status;

    *mode = SOC_PM_LB_PHY | SOC_PM_MII;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG,
						&mii_status));
    if (mii_status & MII_STAT_100_T4)           *mode |= SOC_PM_100MB_FD;
    if (mii_status & MII_STAT_FD_100_T2)        *mode |= SOC_PM_100MB_FD;
    if (mii_status & MII_STAT_HD_100_T2)        *mode |= SOC_PM_100MB_HD;
    if (mii_status & MII_STAT_FD_10)            *mode |= SOC_PM_10MB_FD;
    if (mii_status & MII_STAT_HD_10)            *mode |= SOC_PM_10MB_HD;
    if (mii_status & MII_STAT_FD_100)           *mode |= SOC_PM_100MB_FD;
    if (mii_status & MII_STAT_HD_100)           *mode |= SOC_PM_100MB_HD;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit,phy_id,MII_ESR_REG,&mii_esr));

    if (mii_esr & MII_ESR_1000_T_FD)		*mode |= SOC_PM_1000MB_FD;
    if (mii_esr & MII_ESR_1000_T_HD)		*mode |= SOC_PM_1000MB_HD;
    if (mii_esr & MII_ESR_1000_X_FD)		*mode |= SOC_PM_1000MB_FD;
    if (mii_esr & MII_ESR_1000_X_HD)		*mode |= SOC_PM_1000MB_HD;

    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_init
 *
 * Purpose:
 *  Initialize the PHY (MII mode) to a known good state.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *
 * Returns:
 *  SOC_E_XXXX
 *
 * Notes:
 *  No synchronization performed at this level.
 */
int
phy_xge_init(int unit, soc_port_t port)
{
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_link_get
 *
 * Purpose:
 *  Determine the current link up/down status
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  link - (OUT) Boolean, true indicates link established.
 *
 * Returns:
 *  SOC_E_XXXX
 *
 * Notes:
 *  No synchronization performed at this level.
 */
int
phy_xge_link_get(int unit, soc_port_t port, int *link)
{
    uint8   phy_id = PORT_TO_XPHY_ID(unit, port);
    uint16  mii_stat;

    SOC_IF_ERROR_RETURN(cmic_read_xmiim_5632(unit, phy_id, 0, MII_STAT_REG, 
											 &mii_stat));
    *link = (mii_stat & XMII_STAT_LA) ? TRUE : FALSE;
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_enable_set
 *
 * Purpose:
 *  Enable or disable the physical interface.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  enable - Boolean, true = enable PHY, false = disable.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_enable_set(int unit, soc_port_t port, int enable)
{
    /* BCM8011 can not be set to electrically isolate PHY from xMII */
    if (enable == TRUE) {
        return(SOC_E_NOT_IMPL);
    } else {
        return(SOC_E_NONE);
    }
}

/*
 * Function:
 *  phy_xge_enable_get
 *
 * Purpose:
 *  Enable or disable the physical interface.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  enable - (OUT) Boolean, true = enable PHY, false = disable.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_enable_get(int unit, soc_port_t port, int *enable)
{
    /* BCM8011 can not be set to electrically isolate PHY from xMII */
    *enable = FALSE;
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_duplex_set
 *
 * Purpose:
 *  Set the current duplex mode (forced).
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  duplex -Boolean, true indicates full duplex, false indicates half.
 *
 * Returns:
 *  SOC_E_XXXX
 *
 * Notes:
 *  No synchronization performed at this level. Autonegotiation is
 *  not manipulated.
 */
int
phy_xge_duplex_set(int unit, soc_port_t port, int duplex)
{
    /* BCM8011 does not support half duplex */
    if (duplex) {
        return(SOC_E_NOT_IMPL);
    } else {
        return(SOC_E_NONE);
    }
}

/*
 * Function:
 *  phy_xge_duplex_get
 *
 * Purpose:
 *  Get the current operating duplex mode. If autoneg is enabled,
 *  then operating mode is returned, otherwise forced mode is returned.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  duplex - (OUT) Boolean, true indicates full duplex, false
 *      indicates half.
 *
 * Returns:
 *  SOC_E_XXXX
 *
 * Notes:
 *  No synchronization performed at this level. Autonegotiation is
 *  not manipulated.
 */
int
phy_xge_duplex_get(int unit, soc_port_t port, int *duplex)
{
    /* BCM8011 only supports full duplex mode */
    *duplex = 1;
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_speed_set
 *
 * Purpose:
 *  Set the current operating speed (forced).
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  duplex - (OUT) Boolean, true indicates full duplex, false
 *      indicates half.
 *
 * Returns:
 *  SOC_E_XXXX
 *
 * Notes:
 *  No synchronization performed at this level. Autonegotiation is
 *  not manipulated.
 */
int
phy_xge_speed_set(int unit, soc_port_t port, int speed)
{
    /* BCM8011 only supports 10G speed */
    if (speed != 10000) {
        return(SOC_E_NOT_IMPL);
    } else {
        return(SOC_E_NONE);
    }
}

/*
 * Function:
 *  phy_xge_speed_get
 *
 * Purpose:
 *  Get the current operating speed. If autoneg is enabled,
 *  then operating mode is returned, otherwise forced mode is returned.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  duplex - (OUT) Boolean, true indicates full duplex, false
 *      indicates half.
 *
 * Returns:
 *  SOC_E_XXXX
 *
 * Notes:
 *  No synchronization performed at this level. Autonegotiation is
 *  not manipulated.
 */
int
phy_xge_speed_get(int unit, soc_port_t port, int *speed)
{
    /* BCM8011 only supports 10G speed */
    *speed = 10000;
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_an_set
 *
 * Purpose:
 *  Enable or disabled auto-negotiation on the specified port.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  an   - Boolean, if true, auto-negotiation is enabled
 *      (and/or restarted). If false, autonegotiation is disabled.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_an_set(int unit, soc_port_t port, int an)
{
    /* BCM8011 does not support auto negotiation */
    if (an) {
        return(SOC_E_NOT_IMPL);
    } else {
        return(SOC_E_NONE);
    }

    return(SOC_E_NONE);
}

/*
 * Function:
 *  phyxge_an_get
 *
 * Purpose:
 *  Get the current auto-negotiation status (enabled/busy)
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  an   - (OUT) if true, auto-negotiation is enabled.
 *  an_done - (OUT) if true, auto-negotioation is complete. This
 *          value is undefined if an == false.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_an_get(int unit, soc_port_t port, int *an, int *an_done)
{
    /* BCM8011 does not support auto negotiation */
    *an = FALSE;
    *an_done = FALSE;

    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_adv_local_set
 *
 * Purpose:
 *  Set the current advertisement for auto-negotiation.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mode - Port mode mask indicating supported options/speeds.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_adv_local_set(int unit, soc_port_t port, port_mode_t mode)
{
    /* BCM8011 only supports 10G full duplex, and no pause support */
    if ((mode & SOC_PM_10MB_HD)   ||
        (mode & SOC_PM_10MB_FD)   ||
        (mode & SOC_PM_100MB_HD)  ||
        (mode & SOC_PM_100MB_FD)  ||
        (mode & SOC_PM_1000MB_HD) ||
        (mode & SOC_PM_1000MB_FD))
       return(SOC_E_NOT_IMPL);

    if (mode & SOC_PM_PAUSE) {
        return(SOC_E_NOT_IMPL);
    }
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_adv_local_get
 *
 * Purpose:
 *  Get the current advertisement for auto-negotiation.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mode - (OUT) Port mode mask indicating supported options/speeds.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_adv_local_get(int unit, soc_port_t port, port_mode_t *mode)
{
    /* BCM8011 only supports 10G full duplex, and no pause support */

    *mode |= SOC_PM_10000MB_FD;

    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_interface_set
 *
 * Purpose:
 *  Set the current operating mode of the PHY (10B or MII/GMII).
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mii - TRUE --> MII/GMII, FALSE --> 10B.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_interface_set(int unit, uint32 port, int mii)
{
    /* need to double check this area */
    return(SOC_E_NOT_IMPL);
}

/*
 * Function:
 *  phy_xge_interface_get
 *
 * Purpose:
 *  Get the current operating mode of the PHY (10B or MII/GMII).
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mii - (OUT) TRUE --> MII/GMII, FALSE --> 10B.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_interface_get(int unit, uint32 port, int *mii)
{
    /* need to double check this area */
    *mii = TRUE;
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_ability
 *
 * Purpose:
 *  Get the abilities of the local gigabit phy.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mode - (OUT) Port mode mask indicating supported options/speeds.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_ability(int unit, soc_port_t port, port_mode_t *mode)
{
    /* BCM8011 only supports 10G full duplex, and no pause support */
    *mode = 0;
    *mode |= SOC_PM_10000MB_FD;

    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_adv_remote_get
 *
 * Purpose:
 *  Get partnets current advertisement for auto-negotiation.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  mode - (OUT) Port mode mask indicating supported options/speeds.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_adv_remote_get(int unit, soc_port_t port, port_mode_t *mode)
{
    uint8   phy_id = PORT_TO_PHY_ID[unit][port];
    uint16  mii_ctrl, mii_stat, mii_gb_stat, mii_anp;

    *mode = 0;

    /* ORDER: must read control, then stat, then anp */

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG, 
											&mii_ctrl));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_STAT_REG, 
											&mii_stat));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_ANP_REG, 
											&mii_anp));
    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, port, MII_GB_STAT_REG,
                       						&mii_gb_stat));
    if (!(mii_ctrl & MII_CTRL_AE)) {
    	return(SOC_E_AUTONEG_DISABLED);
    } 
	else 
	if (!(mii_stat & MII_STAT_AN_DONE)) {
    	return(SOC_E_AUTONEG_BUSY);
    }
    if (mii_gb_stat & MII_GB_STAT_LP_1000FD)    *mode |= SOC_PM_1000MB_FD;
    if (mii_gb_stat & MII_GB_STAT_LP_1000HD)    *mode |= SOC_PM_1000MB_HD;

    if (mii_anp & MII_ANA_PAUSE) {
    	if (mii_anp & MII_ANA_ASYM_PAUSE) {
        	*mode |= SOC_PM_PAUSE_TX;
    	} 
		else {
        	*mode |= SOC_PM_PAUSE;
    	}
    }
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_lb_set
 *
 * Purpose:
 *  Set the local PHY loopback mode.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  loopback - Boolean: true = enable loopback, false = disable.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_lb_set(int unit, soc_port_t port, int enable)
{
    uint8   phy_id = PORT_TO_XPHY_ID(unit, port);
    uint16  mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    mii_ctrl &= ~MII_CTRL_LE;
    mii_ctrl |= enable ? MII_CTRL_LE : 0;

    SOC_IF_ERROR_RETURN(cmic_write_miim_5632(unit, phy_id, MII_CTRL_REG,
											 mii_ctrl));
    return(SOC_E_NONE);
}

/*
 * Function:
 *  phy_xge_lb_get
 *
 * Purpose:
 *  Get the local PHY loopback mode.
 *
 * Parameters:
 *  unit - StrataSwitch unit #.
 *  port - StrataSwitch port #.
 *  loopback - (OUT) Boolean: true = enable loopback, false = disable.
 *
 * Returns:
 *  SOC_E_XXXX
 */
int
phy_xge_lb_get(int unit, soc_port_t port, int *enable)
{
    uint8   phy_id = PORT_TO_XPHY_ID(unit, port);
    uint16  mii_ctrl;

    SOC_IF_ERROR_RETURN(cmic_read_miim_5632(unit, phy_id, MII_CTRL_REG,
											&mii_ctrl));
    *enable = (mii_ctrl & MII_CTRL_LE) ? 1 : 0;
    return(SOC_E_NONE);
}

