/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	bsdintf-api.c
 *
 * RCS:  $Id: bsdintf-api.c,v 1.21 2002/10/30 19:45:07 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions manage IP interface.
 *
 * Created by:	Su Wang
 *
 * Date Created:	7/25/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/errno.h"
#include "../sm-bsd-sys/macros.h"
#include "../sm-bsd-sys/mbuf.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/socketvar.h"
#include "../sm-bsd-sys/sockio.h"
#include "../sm-bsd-sys/protosw.h"

#include "../sm-bsd-net/if.h"
#include "../sm-bsd-net/if_types.h"
#include "../sm-bsd-net/route.h"
#include "../sm-bsd-net/if_dl.h"
#include "../sm-bsd-net/if_arp.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/in_var.h"
#include "../sm-bsd-netinet/in_pcb.h"
#include "../sm-bsd-netinet/ip_var.h"

#include "exports/cm-api.h"
#include "exports/bsd-api.h"
#include "exports/bsdnet-api.h"
#include "exports/ipm-api.h"
#include "../sm-ipm/ipm.h"
#include "bsd-vif.h"
#if defined(SIBYTE)
#include "exports/bmc12500-eth-api.h"
#endif /* SIBYTE */

char* bsd_ifnamebuf[IPM_INTF_TYPE_MAX+1] = {"dummy", "fp", "sc"};

/* hack to make rx_rcv work */
extern uint32   myIpaddr;

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: return first IP interface that subsumes or being
 *              subsumed by given network range.
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
struct in_ifaddr* findFirstMatchIntf(uint32 addr, uint32 in_mask)
{
  
  register struct in_ifaddr *ia;
  uint32   mask;

  for (ia = in_ifaddrhead.tqh_first; ia;
       ia = ia->ia_link.tqe_next)
    {
      mask = ia->ia_sockmask.sin_addr.s_addr & in_mask;
      if ((ia->ia_addr.sin_addr.s_addr & mask) == (addr & mask))
	return ia;
    }

  return 0;
}

/*-----------------------------------------------------------------
 * Name: bsd_FillIpmMsg
 *
 * Description: fill in Ipm msg for give interface
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void
bsd_FillIpmMsg(int		mtype,
			   void		*msg,
			   void		*i_ifp,
			   void		*parm)
{
  struct ifnet	*ifp = (struct ifnet*)i_ifp;
  ipm_mintf_t	*ipm_intfp;

  switch (mtype)
	{
	case IPM_MSG_INTF_NEW:
	case IPM_MSG_INTF_RESP:
	  ipm_intfp = (ipm_mintf_t*)msg;
	  ipm_intfp->if_dtype = bsdDTypeStoN(ifp->if_name);
	  ipm_intfp->if_slot = ifp->if_slot;
	  ipm_intfp->if_cpu  = ifp->if_cpu;
	  ipm_intfp->if_link = ifp->if_unit;
	  bcopy(((struct arpcom*)ifp)->ac_enaddr, 
			ipm_intfp->if_mac, 6); /* gw has mac addr */
	  ipm_intfp->if_flags = ifp->if_flags;
	  ipm_intfp->if_dhdrlen = ifp->if_hdrlen;
	  ipm_intfp->if_daddrlen = ifp->if_addrlen;
	  ipm_intfp->if_mtusize = ifp->if_mtu;

#if defined(SIBYTE)
	  /* get auto negociate info */
	  if (bmc12500Eth_get_autoneg(ifp->if_unit))
		  ipm_intfp->if_flags |= IFF_LINK0;
#endif /* SIBYTE */
	  break;
	}
  return;
}


/*-----------------------------------------------------------------
 * Name: bsd_FindIntfByName
 *
 * Description: return interface ptr by given interface name
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void*
bsd_FindIntfByName(char  *name)
{
  return (void*)ifunit(name);
}

/*-----------------------------------------------------------------
 * Name: bsd_FindIntfByIp
 *
 * Description: find interface by IP
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void*
bsd_FindIntfByIp(uint32 ipAddr)
{
	struct ifnet		*ifp;
	struct ifaddr		*ifa;
	struct sockaddr_in	*ifIpAddr;
		
	
	for (ifp = ifnet.tqh_first; ifp; ifp = ifp->if_link.tqe_next) {
		ifa = ifp->if_addrhead.tqh_first;
		
		/* delete IP address associated with each intf */
		while (ifa) {
			ifIpAddr = (struct sockaddr_in*)ifa->ifa_addr;
			if (ifIpAddr->sin_family != AF_INET) {
				ifa = ifa->ifa_link.tqe_next;
				continue;
			}
			
			if (ifIpAddr->sin_addr.s_addr == ipAddr)
				return (void*)ifp;
			else
				ifa = ifa->ifa_link.tqe_next;
		}
	}
	return NULL;
}

/*-----------------------------------------------------------------
 * Name: bsd_FindIpByIntf
 *
 * Description: find IP that associated with an interface
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void*
bsd_FindIpByIntf(void*	in_ifp)
{
	struct ifnet		*ifp = (struct ifnet*)in_ifp;
	struct ifaddr		*ifa;
	struct sockaddr_in	*ip_addr;

	if (!ifp)
		return NULL;
	
	for (ifa = ifp->if_addrhead.tqh_first; ifa; 
		 ifa = ifa->ifa_link.tqe_next) {
		ip_addr = (struct sockaddr_in*)ifa->ifa_addr;
		if (ip_addr->sin_family == AF_INET) 
			return (void*)ifa;
	}

	return NULL;
}


/*-----------------------------------------------------------------
 * Name: bsd_AttachIf
 *
 * Description: adding a device interface
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_AttachIf(void *i_ipm_mp)
{
  ipm_mintf_t		*ipm_intfp = (ipm_mintf_t*)i_ipm_mp;
  int			dtype = ipm_intfp->if_dtype;
  int			slot  = ipm_intfp->if_slot;
  int			cpu   = ipm_intfp->if_cpu;
  int			link  = ipm_intfp->if_link;
  struct ifnet	*ifp;
  struct ifaddr	*ifa;
  struct sockaddr_dl	*sdl;
  char			*device;
  

  device = (char*)bsdPrintPort(slot, cpu, link, dtype);
  
  
  /* interface exist ? */
  ifp = (struct ifnet*)bsd_FindIntfByName(device); 

  if (ifp)
	return NFX_OK;

  /* only attach if interface is from another card */
  if ((slot == nfxMySlotId) && (cpu == nfxMyCPUId))
	  return NFX_ERR;

  /* alloc arpcom instead, */
  if ((ifp = (struct ifnet*)malloc(sizeof(struct arpcom), 0,0)) 
	  == NULL)
	return NFX_ERR;

  bzero(ifp, sizeof(struct arpcom));

  /* fill in ifnet */
  ifp->if_softc	   = ifp;
  ifp->if_type     = IFT_ETHER;
  ifp->if_addrlen  = ipm_intfp->if_daddrlen;
  ifp->if_hdrlen   = ipm_intfp->if_dhdrlen;
  ifp->if_mtu	   = ipm_intfp->if_mtusize;
  ifp->if_flags    = ipm_intfp->if_flags;
  ifp->if_name     = bsd_ifnamebuf[dtype];
  ifp->if_slot	   = slot;
  ifp->if_cpu      = cpu;
  ifp->if_unit     = link;
  ifp->if_output   = bsd_XmtPkt;
  
  
  /* get mac address */
  bcopy(ipm_intfp->if_mac, ((struct arpcom*)ifp)->ac_enaddr, 6);

  /* link to system etc */
  if_attach(ifp);

  ifa = ifnet_addrs[ifp->if_index - 1];
  if (ifa == 0) {
	bsd_log(BSD_ERR_INTF, "ether_ifattach: no lladdr!\n");
	return NFX_ERR;
  }

  sdl = (struct sockaddr_dl *)ifa->ifa_addr;
  sdl->sdl_type = IFT_ETHER;
  sdl->sdl_alen = ifp->if_addrlen;
  bcopy(((struct arpcom *)ifp)->ac_enaddr, LLADDR(sdl), ifp->if_addrlen);

  return NFX_OK;
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_CreateIpIntf
 *
 * Description: create an IP interface 
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_CreateIpIntf(uint32 ipaddr,  /* in network order */ 
		 uint32 ipmask,  /* in network order */
		 void* device,   /* media port id */
		 void* param1,
		 void* param2
		 )
{
  struct in_aliasreq in_data;
  uint32 h_ipaddr = ntohl(ipaddr);
  uint32 h_ipmask = ntohl(ipmask);
  int error;

  bzero(&in_data, sizeof(struct in_aliasreq));
  /* sanity check for input */
  if (MASK_HAS_HOLE(h_ipmask))
    {
      uiPrintf("mask %x has holes ~mask+1 %x %x\n", 
	       h_ipmask, ~h_ipmask + 1,  (h_ipmask | (~h_ipmask + 1)));
       return NFX_ERR;
    }

  if (IN_EXPERIMENTAL(h_ipaddr) || 
      IN_BADCLASS(h_ipaddr) || IN_CLASSD(h_ipaddr))
    {
      uiPrintf("bad ip address %x\n", h_ipaddr);
      return NFX_ERR;
    }
  
  
  if (bsd_IsSupernet(h_ipaddr, h_ipmask))
    {
      uiPrintf("address is a supernet ip %x mask %x\n", h_ipaddr, h_ipmask);
      return NFX_ERR;
    }
  
#if 0 
  /* make sure added interface is not overlapping with 
     exist interfaces */
  if (intf = findFirstMatchIntf(ipaddr, ipmask))
   {
   	uiPrintf("can not add, overlapping with intf %x %x\n", 
       	intf->ia_addr.sin_addr.s_addr, intf->ia_sockmask.sin_addr.s_addr); 
   	return NFX_OK;
   }
#endif  
  
  bsdPortToString(in_data.ifra_name, device);

  in_data.ifra_addr.sin_len = 16;
  in_data.ifra_addr.sin_family = AF_INET;
  in_data.ifra_addr.sin_addr.s_addr = ipaddr;

  in_data.ifra_mask.sin_len = 8;
  in_data.ifra_mask.sin_family = AF_INET;
  in_data.ifra_mask.sin_addr.s_addr = ipmask;

  in_data.ifra_broadaddr.sin_len = 16;
  in_data.ifra_broadaddr.sin_family = AF_INET;
  in_data.ifra_broadaddr.sin_addr.s_addr = (ipaddr & ipmask) | ~ipmask;

  error=ifioctl(0, SIOCAIFADDR, (caddr_t)&in_data, 0);
  
  if (error && (error != EEXIST))	  
    {
      bsd_log(BSD_WARN_INTF, "bsd did not create interface, errno %d \n", error);
      return NFX_ERR;
    }
  else {
    if (!myIpaddr)
	    myIpaddr = ntohl(ipaddr);
    return NFX_OK;
  }
      
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_AddIpByIntf
 *
 * Description: add an IP interface 
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_AddIpByIntf(uint32 ipAddr,  /* in network order */
				uint32	mask,
				void*   in_ifp)
{
	struct ifnet		*ifp = (struct ifnet*)in_ifp;
	char				device[32];
	
	if (!ifp)
		return NFX_ERR;
	
	sprintf(device, "%s%d", ifp->if_name, ifp->if_unit);
	return bsd_CreateIpIntf(ipAddr, mask, device, NULL, NULL);
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_SetMtuIntf
 *
 * Description: change interface MTU size
 *
 * Created by: Su Wang
 *
 * Data Created: 02/28/02
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_SetMtuIntf(void* device, 
					int	  mtu)
{
	struct ifnet	*ifp;
	
	ifp = (struct ifnet*)bsd_FindIntfByName((char*)device);
	
	if (!ifp)
		return NFX_ERR;
	
	ifp->if_mtu = mtu;
	return NFX_OK;
}


/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_ChangeStateIntf
 *
 * Description: change the administrative state of the device.
 *              all ip interfaces associated with this device
 *              will be affected.    
 *
 *				Note: Unlike in bsd code IFF_UP change does not affect
 *					  IFF_RUNNING. The driver must set IFF_RUNNING 
 *					  in order to indicate operational state of 
 *					  interface. 
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_ChangeStateIntf(void* device,  /* in network order */ 
		    int32 state,   /* new state */
		    void* param1,
		    void* param2
		 )
{
  struct ifreq in_data;
  int error;

  
  bzero(&in_data, sizeof(struct ifreq));
  bsdPortToString(in_data.ifr_name, device);
  
 
  if ((error=ifioctl(0, SIOCGIFFLAGS, (caddr_t)&in_data, 0)))
    {
      bsd_log(BSD_WARN_INTF,
			  "bsd did not get interface flags, errno %d \n", error);
      return NFX_ERR;
    }

  if (state == BSD_UP)
    {
      if (in_data.ifr_flags & IFF_UP)
		return NFX_OK;

      in_data.ifr_flags |= IFF_UP;
    }
  else if (state == BSD_DOWN)
    {
      if ((in_data.ifr_flags & IFF_UP) == 0)
		return NFX_OK;      
      
	  in_data.ifr_flags &= ~IFF_UP;
    }
  else
	return NFX_ERR;

  if ((error=ifioctl(0, SIOCSIFFLAGS, (caddr_t)&in_data, 0)))
    {
      bsd_log(BSD_WARN_INTF, 
			  "bsd did not set interface flags, errno %d \n", error);
      return NFX_ERR;
    }   
  else {
#if defined(LINK_AGG)
	  struct ifnet *ifp;
	  
	  ifp = bsd_FindIntfByName(in_data.ifr_name);
	  bsd_VifLinkChange(ifp);
	  
#endif /* LINK_AGG */	  
	  return NFX_OK;
  }
  
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_ChangeStateIntfAll
 *
 * Description: change state of all interface on the card.
 *
 * Created by: Su Wang
 *
 * Data created : 03/22/02
 *-----------------------------------------------------------------
 */
void bsd_ChangeStateIntfAll(int32 state) 
{
	struct ifnet	*ifp=NULL;
	char			device[IFNAMSIZ];

	while ((ifp = (struct ifnet*)bsd_GetNextIntf(ifp))) {
		sprintf(device, "%s:%d:%d:%d",
				ifp->if_name, ifp->if_slot, 
				ifp->if_cpu, ifp->if_unit);
		bsd_ChangeStateIntf(device, state, NULL, NULL);
	}
	
	return;
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_ChangeLinkStatus
 *
 * Description: change the operational state of the device. Called 
 *				driver.
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_ChangeLinkStatus(int link, int state )
{
	  struct ifnet	*ifp;

	  ifp = (struct ifnet*)bsd_FindIntfByName((char*)bsdPrintPort(nfxMySlotId,
																  nfxMyCPUId,
																  link,
																  IPM_INTF_TYPE_ETH));

	  if (!ifp)
		  return NFX_ERR;
	  
	  if (state == BSD_UP)
		  ifp->if_flags |= IFF_RUNNING;
	  else
		  ifp->if_flags &= ~IFF_RUNNING;


#if defined(LINK_AGG)
	  bsd_VifLinkChange(ifp);
#endif /* LINK_AGG */	  

	  return NFX_OK;
}



/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_DeleteIpIntf
 *
 * Description: delete an IP interface 
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_DeleteIpIntf(uint32 ipaddr,  /* in network order */
		 void* device,  /* device to which IP interface associates with */
		 void* param1,
		 void* param2
		 )
{
  
  struct in_aliasreq in_data;
  int error;
  

  bzero(&in_data, sizeof(struct in_aliasreq));
  bsdPortToString(in_data.ifra_name, device);

  in_data.ifra_addr.sin_len = 16;
  in_data.ifra_addr.sin_family = AF_INET;
  in_data.ifra_addr.sin_addr.s_addr = ipaddr;

  if ((error=ifioctl(0, SIOCDIFADDR, (caddr_t)&in_data, 0)))
    {
      bsd_log(BSD_WARN_INTF, 
			  "bsd did not delete interface, errno %d \n", error);
      return NFX_ERR;
    }
  else
    return NFX_OK;  
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_DeleteIpByIntf
 *
 * Description: delete an IP interface 
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *-----------------------------------------------------------------
 */
int32  /* NFX_ERR fail, NFX_OK ok */
bsd_DeleteIpByIntf(uint32 ipAddr,  /* in network order */
				   void*   in_ifp)
{
	struct ifnet		*ifp = (struct ifnet*)in_ifp;
	char				device[32];
	
	if (!ifp)
		return NFX_ERR;
	
	sprintf(device, "%s%d", ifp->if_name, ifp->if_unit);
	return bsd_DeleteIpIntf(ipAddr, device, NULL, NULL);
}

/*
 *-----------------------------------------------------------------
 *
 * Name : bsd_DeleteIpAll
 *
 * Description: delete all ip address 
 *
 * Created by: Su Wang
 *
 * Data Created: 08/19/02
 *
 *-----------------------------------------------------------------
 */
void
bsd_DeleteIpAll(void	*in_ifp)
{
	struct ifnet		*ifp = (struct ifnet*)in_ifp;
	struct ifaddr		*ifa, *ifaNxt;
	struct sockaddr_in	*ip_addr;
	char				device[32];
	int					single = 1;

	/* delete all ip address on the txrx */
	if (!ifp) {
		single = 0;
		ifp = ifnet.tqh_first;
	}
	
	for (; ifp; ifp = ifp->if_link.tqe_next) {
		if (strcmp(ifp->if_name, "lo") == 0)
			continue;
		
		sprintf(device, "%s%d", ifp->if_name, ifp->if_unit);
	
		// printf("device is %s\n", device);	
		ifa = ifp->if_addrhead.tqh_first;
		
		/* delete IP address associated with each intf */
		while (ifa) {
			ifaNxt = ifa->ifa_link.tqe_next;
			ip_addr = (struct sockaddr_in*)ifa->ifa_addr;
			if (ip_addr->sin_family != AF_INET) {
				ifa = ifaNxt;
				continue;
			}
			
			// printf("delete ip %x from intf %s\n",
			//	   ip_addr->sin_addr.s_addr, device);
			bsd_DeleteIpIntf(ip_addr->sin_addr.s_addr, device, NULL, NULL);
			ifa = ifaNxt;
		}

		/* already deleted ip addresses on a single interface */
		if (single)
			break;
	}

	return;
}

/*
 *----------------------------------------------------------------
 * 
 * Name : bsd_GetNextIntf
 *
 * Description: get next local intf, lo excluded.
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *---------------------------------------------------------------
 */
void*
bsd_GetNextIntf(void *ifp)
{
  struct ifnet	*next_ifp;
 
next: 
  if (!ifp)
	next_ifp =  ifnet.tqh_first;
  else
	next_ifp = (((struct ifnet*)ifp)->if_link.tqe_next);

  /* reach last one */
  if (!next_ifp)
	return NULL;

  /* ignore remote and loopback */
  if ((strcmp(next_ifp->if_name, "lo") == 0) ||
	  next_ifp->if_slot != nfxMySlotId ||
	  next_ifp->if_cpu  != nfxMyCPUId)
	{
	  ifp = (void*)next_ifp;
	  goto next;
	}

  return next_ifp;
}

extern byte *seepGetMyExtMacAddr(int32 link);

/*
 *----------------------------------------------------------------
 * 
 * Name : bsd_GetIntfMacVif
 *
 * Description: if link belongs to link aggregation vif, return
 *				vif mac, else return link mac.
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *---------------------------------------------------------------
 */
byte*
bsd_GetIntfMacVif(void	*in_ifp)
{
	struct ifnet*	ifp = (struct ifnet*)in_ifp;
	
	
	if (!ifp)
		return NULL;

#if defined(LINK_AGG)
	if ((ifp->if_vif) && 
		(((bsd_vif_t*)ifp->if_vif)->bsd_vifFlags & IPM_VIF_MODE_LA))
		return ((bsd_vif_t*)ifp->if_vif)->bsd_vifMac;
	else
#endif /* LINK_AGG */
		return seepGetMyExtMacAddr(ifp->if_unit);
}

/*
 *----------------------------------------------------------------
 * 
 * Name : bsd_GetIntfMac
 *
 * Description: get mac address from seep
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *---------------------------------------------------------------
 */
byte*
bsd_GetIntfMac(void	*in_ifp)
{
	struct ifnet*	ifp = (struct ifnet*)in_ifp;
	
	
	if (!ifp)
		return NULL;

	return seepGetMyExtMacAddr(ifp->if_unit);
}

/*
 *----------------------------------------------------------------
 * 
 * Name : bsd_ShowIntf
 *
 * Description: display a summary of all IP interfaces or one IP 
 *              interface in detail.
 *
 * Created by: Su Wang
 *
 * Data Created: 7/25/01
 *
 *---------------------------------------------------------------
 */
void
bsd_ShowIntf(int32 how,   /* summary or a single interface */
	     void* device, /* IP address if single interface */
	     void* param1,
	     void* param2
	     )
{
  /* can use standard ifioctl to retrive IP interface info,
     but its too much hassle. Instead, we iterate IP interface 
     directly in this function */

  register struct ifnet *ifp;
  register struct ifaddr *ifa;  
  char name_buf[IFNAMSIZ+1];

  switch (how)
    {
    case BSD_ALL:
#if defined(LINK_AGG)
	{
		bsd_vif_t	*tmpVif;
		for (tmpVif = VifHead; tmpVif; tmpVif = tmpVif->bsd_vifNxt)
			bsd_PrintVif(tmpVif);
	}
#endif /* LINK_AGG */

      uiPrintf("Interface Table\n");
      uiPrintf("--------------------------------------------------------------\n");
      uiPrintf("Name                Mac-Addr                Op-St   Adm-St      IP/Mask  \n");
      uiPrintf("--------------------------------------------------------------\n");
      for (ifp = ifnet.tqh_first; ifp; ifp = ifp->if_link.tqe_next)
		{ 
		  struct sockaddr_dl *link_addr;
		  struct sockaddr_in *ip_addr;
		  uint32 mask;

		  /* first addr struct should be MAC addr */
		  ifa = ifp->if_addrhead.tqh_first;
	  
		  if (ifa->ifa_addr->sa_family != AF_LINK)
			{
			  debugPrintf(("bsd_showIntf, incorrect struct %s\n", 
						   bsdPrintfIntf(ifp)));
			  continue;
			}
	  
		  /* remember the mac address */
		  link_addr = (struct sockaddr_dl*)ifa->ifa_addr;
		  bsdMacNtoS(LLADDR(link_addr), prnt_mac, sizeof(prnt_mac));
		  uiPrintf("%-20s%-24s%-8s%-11s",
				   bsdPrintIntfWithVif(ifp),
				   prnt_mac,
				   (ifp->if_flags & IFF_RUNNING)?"UP":"DOWN",
				   (ifp->if_flags & IFF_UP)?"UP":"DOWN");

		  /* IP interfaces */
		  ifa = ifa->ifa_link.tqe_next;
		  for (; ifa; ifa = ifa->ifa_link.tqe_next) 
			{
			  ip_addr = (struct sockaddr_in*)ifa->ifa_addr;

			  if (ip_addr->sin_family != AF_INET)
				continue;
	      
			  /* rember IP address */
			  bsdIpNtoS(&(ip_addr->sin_addr.s_addr), prnt_ip, sizeof(prnt_ip));
			  mask = ((struct sockaddr_in*)ifa->ifa_netmask)->sin_addr.s_addr;
			  uiPrintf("%s/%d\n",
					   prnt_ip,
					   bsdMasklen(&mask));
			 uiPrintf("%-57s", " ");	
			}   
 			printf("\n"); 
		}
      break;

    case BSD_SINGLE:
      bsdPortToString(name_buf, device);
      if (!(ifp = ifunit(name_buf)))
		{
		  uiPrintf("device %s not found\n", name_buf);
		  return;
		}

      uiPrintf("Detailed Info for Device %s\n", name_buf);
      uiPrintf("-------------------------------------------------------------\n");
      uiPrintf("Device Name:      %s\n", bsdPrintIntfWithVif(ifp));
      uiPrintf("Device Uint:      %d\n", ifp->if_unit);
      uiPrintf("Device Inx:       %d\n", ifp->if_index);
      uiPrintf("Device Flags:     %x\n", ifp->if_flags);
      uiPrintf("Device MTU:       %d\n", ifp->if_mtu);
      uiPrintf("Device Type:      %d\n", ifp->if_type);

      
      for (ifa = ifp->if_addrhead.tqh_first; ifa;
		   ifa = ifa->ifa_link.tqe_next) 
		{
		  if (ifa->ifa_addr->sa_family == AF_LINK)
			{
			  struct sockaddr_dl *dladdr, *dlmask;

			  uiPrintf("\nLink Layer Interface\n");
			  dladdr = (struct sockaddr_dl*)ifa->ifa_addr;
			  dlmask = (struct sockaddr_dl*)ifa->ifa_netmask;
	      
			  bsdMacNtoS(LLADDR(dladdr), prnt_mac, 20);
			  uiPrintf("Mac Host Addr:    %s\n", prnt_mac);
			  bsdMacNtoS(LLADDR(dlmask), prnt_mac, 20);
			  uiPrintf("Mac Bcast Addr:   %s\n", prnt_mac);
			}
		  else if (ifa->ifa_addr->sa_family == AF_INET)
			{
			  struct sockaddr_in* in_addr;
			  struct in_ifaddr* ia = (struct in_ifaddr*)ifa;

			  uiPrintf("\nIP Layer Interface\n");
	      
			  in_addr = (struct sockaddr_in*)ifa->ifa_addr;
			  bsdIpNtoS(&(in_addr->sin_addr.s_addr), prnt_ip, 20);
			  uiPrintf("IP Addr:     %s\n", prnt_ip);
	      
			  in_addr = (struct sockaddr_in*)ifa->ifa_broadaddr;
			  bsdIpNtoS(&(in_addr->sin_addr.s_addr), prnt_ip, 20);
			  uiPrintf("IP Bcast Addr:     %s\n", prnt_ip);

			  in_addr = (struct sockaddr_in*)ifa->ifa_netmask;
			  bsdIpNtoS(&(in_addr->sin_addr.s_addr), prnt_ip, 20);
			  uiPrintf("IP Mask:     %s\n", prnt_ip);

			  /* print rest of stuff in hex */
			  uiPrintf("Network Num:    %08x\n", ia->ia_net);
			  uiPrintf("Network Mask:   %08x\n", ia->ia_netmask);
			  uiPrintf("Subnet Num:     %08x\n", ia->ia_subnet);
			  uiPrintf("Subnet Mask:    %08x\n", ia->ia_subnetmask);
			  uiPrintf("Network Bcast:  %08x\n", ia->ia_netbroadcast.s_addr);

			}
		  else
			uiPrintf("\nbad address type %d detected\n", ifa->ifa_addr->sa_family);

		  uiPrintf("Metric  :     %d\n", ifa->ifa_metric);
		  uiPrintf("Flags   :     %x\n", ifa->ifa_flags);
		  uiPrintf("Ref Cnt :     %d\n", ifa->ifa_refcnt);
		   
		}     
      break;
    } /* switch */

  return;
}

/*
 *----------------------------------------------------------------
 * 
 * Name : bsd_ShowMbufStats
 *
 * Description: Display mbuff stats
 *             
 *
 * Created by: Su Wang
 *
 * Data Created: 8/16/01
 *
 *---------------------------------------------------------------
 */
void
bsd_ShowMbufStats()
{
  int i, count=0;

  uiPrintf("Summary of mbuffer statistatics\n");
  uiPrintf("-----------------------------------------------\n");
  uiPrintf("mbuf size:                 %u\n", mbstat.m_msize);
  uiPrintf("cluster size:              %u\n", mbstat.m_mclbytes);
  uiPrintf("min clust data size:       %u\n", mbstat.m_minclsize);
  uiPrintf("mbuf data size:            %u\n", mbstat.m_mlen);
  uiPrintf("pkt mbuf data size:        %u\n", mbstat.m_mhlen);
  uiPrintf("\n\n");
  
  uiPrintf("total mbuf:                %u\n", mbstat.m_mbufs);
  for (i = 1; i<256; i++)
    count += mbstat.m_mtypes[i];
  uiPrintf("mbuf in use:               %u\n", count);
  uiPrintf("total cluster:             %u\n", mbstat.m_clusters);
  uiPrintf("cluster in use:            %u\n", mbstat.m_clusters - mbstat.m_clfree);
  uiPrintf("request delayed:           %u\n", mbstat.m_wait);
  uiPrintf("request dropped:           %u\n", mbstat.m_drops);
  uiPrintf("drain routing called:      %u\n", mbstat.m_drain);

  return;
 }

