/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: bsdrelay.c
 *
 * RCS:  $Id: bsdrelay.c,v 1.25 2002/10/30 22:18:28 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: contains IP pkt relay functions
 *
 * Created by:   Su Wang
 *
 * Date Created: 10/19/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/errno.h"
#include "../sm-bsd-sys/macros.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/sockio.h"
#include "../sm-bsd-sys/protosw.h"

#include "../sm-bsd-net/if.h"
#include "../sm-bsd-net/if_types.h"
#include "../sm-bsd-net/ethernet.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/in_var.h"
#include "../sm-bsd-netinet/ip.h"
#include "../sm-bsd-netinet/ip_var.h"

#include "exports/eee-api.h"
#include "exports/cm-api.h"
#include "exports/pkt-api.h"
#include "exports/luc-api.h"
#include "exports/bsdnet-api.h"
#include "bsdrelay.h"
#include "exports/tpl-api.h"
#include "bsd-vif.h"

#if !(defined(LINUX_TEST) || defined(PROM))
#define ENABLE_RELAY
#endif /* LINUX_TEST || PROM */

int						bsdrl_appid;/* appl id to eee */
int						bsdeee_src;	/* eee src addr */
rl_rq_t					bsdrl_ReassemQ[NFX_NUM_SLOTS_PER_CHASSIS];



/* 
 *-----------------------------------------------------------------
 *
 * Name: bsdrl_Init
 *
 * Description: initialize relay layer.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/22/2001
 *
 *-----------------------------------------------------------------
 */
void
bsdrl_Init()
{
  /* register with EEE */
  bsdrl_appid = eee_allocateApplicationId("bsdrl");

  if (bsdrl_appid == EEE_ERR_INVALID_APP_NAME)
	panic("bsdrl_Init: invalid app name \n");

  bsdeee_src = PACK_UNICAST_PORT_ADDRESS(
									nfxMySlotId,
									nfxMyCPUId,
									bsdrl_appid,
									EEE_USER_DATA,
									NFX_COS_HIGH,
									0);
  
  if (eee_registerReceiveFunc(bsdrl_appid, bsdrl_RcvPkt, bsdrl_RcvPkt)
	  != NFX_OK)
	panic("bsdrl_Init: duplicated registeration with EEE\n"); 

  bzero(bsdrl_ReassemQ, sizeof(rl_rq_t)*NFX_NUM_SLOTS_PER_CHASSIS);
}


/*-----------------------------------------------------------------
 * Name:        bsdrl_DstPkt
 *
 * Description: determine the pkt destination, if pkt is to be 
 *				forwarded to another card, the rl_adddr_t struct is
 *				filled to specify destination.
 *
 * Created by : Su Wang
 *
 * Date Created: 10/19/01
 *-----------------------------------------------------------------
 */
int32							/* BSD_RELAY, pkt relayed via eee */
								/* BSD_NREPLAY, pkt is for loc card*/
bsdrl_DstPkt(pkt_desc_t	*pdesc, 
			rl_addr_t	*rla,
			void		*i_ifp)
{
  struct ip				*iphdr;
  uchar8				proto;
  ushort16				dport;
  struct ifnet			*ifp = (struct ifnet*)i_ifp;
  struct in_ifaddr		*ia;
  ushort16				dslot = 0;
  ushort16				dcpu = NFX_SSC_UNIX;
  

#if defined(PROM) || !defined(ENABLE_RELAY) 
  /* process all received pkts in prom */
  return BSD_NRELAY;
#endif /* PROM */  

  /* sanity check */
  rla->rl_ifp = (void*)ifp;
  if (ifp->if_slot != nfxMySlotId ||
	  ifp->if_cpu != nfxMyCPUId)
	{
	  bsd_log(BSD_ERR_RL_PKT, "bad pkt rcvd at intf %lx, myslot %d mycpu %d\n",
			 ifp, nfxMySlotId, nfxMyCPUId);
	  return BSD_RELAY_ERR;
	}


  /* assuming pdesc->pIpHdr is 32 bits aligned */
  iphdr = (struct ip*)pdesc->pIpHdr;
 
#if defined(ENABLE_RELAY) 
  /* always process ip fragment, funct. will be called again to */
  /* determine the destination of the resulting ip pkt */
  if (pdesc->flags & PKT_IP_FRAGMENT)
	  return BSD_NRELAY;
#endif /* ENABLE_RELAY */
  
  /* check for protocol, txrx only takes tcp/udp pkt */
  proto = iphdr->ip_p;

#if defined(LINUX_TEST)
  if (proto == BSD_IPPROTO_ICMP)
	return BSD_NRELAY;
#endif /* LINUX_TEST */ 


  if (!(proto == BSD_IPPROTO_TCP || proto == BSD_IPPROTO_UDP))
	goto to_ssc;

  /* check for udp/tcp range. */
  /* noting tcp/udp port start at */
  /* the same offset from end of ip header */
  dport = *(ushort16*)(pdesc->pTplHdr + 2);
  NTOHS(dport);
  if ((dport < BSD_PORT_TXRX_FIRST || dport > BSD_PORT_TXRX_LAST) &&
	  (!BSD_WELLKNOWN_PORTS(dport)))
	  goto to_ssc;
 
 
  /* determine pkt dest by examing receiving IP interface */
  for (ia = TAILQ_FIRST(&in_ifaddrhead); ia;
	   ia = TAILQ_NEXT(ia, ia_link)) 
	{
	  /* use bcmp to avoid misalignment for ip addr */
	  if (bcmp((uchar8*)&(IA_SIN(ia)->sin_addr.s_addr), 
			   (uchar8*)&(iphdr->ip_dst.s_addr), 4) == 0)
		break;
	}
  
  /* pkt to one of IP interface */ 
  if (ia)
	{
	  /* ok to reuse ifp, rla has a record */
	  ifp = ia->ia_ifa.ifa_ifp;
	  
	  /* pkt is to local txrx */
	  if ((ifp->if_slot == nfxMySlotId) &&
		  (ifp->if_cpu == nfxMyCPUId))
		return BSD_NRELAY;
	  else
		{
		  dslot = ifp->if_slot;
		  dcpu  = ifp->if_cpu;
		  
		  /* little misleading, but we have reset dslot and dcpu */
		  /* which enables pkt to go to the correct dest  */
		  goto to_ssc;
		}
	}
  else
	{
	  /* pkt not to IP interface, continue to process by txrx bsd stack */
	  return BSD_NRELAY;
	}

to_ssc:
  rla->rl_slot = dslot;
  rla->rl_cpu  = dcpu;
  rla->rl_ctrl = 0;
  return BSD_RELAY;
}

/*-----------------------------------------------------------------
 * Name:        bsd_RelayPkt
 *
 * Description: relay an pkt to destination specified by rl_addr_t 
 *				struct..
 *
 * Created by : Su Wang
 *
 * Date Created: 10/19/01
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsdrl_RelayPkt(pkt_desc_t	*pdesc,
			   rl_addr_t	*rla,  
			   void			*parm)
{
#if defined (PROM) || !defined(ENABLE_RELAY)
  /* disable relay, */
  return NFX_ERR;
#else
  
  eee_desc_t	*eDescp;
  rl_hdr_t		*rl_hdr;
  struct ifnet	*ifp = (struct ifnet*)rla->rl_ifp;
  uint32		e_src;
  uint32		e_dest;
  int			bcast = 0;
  int			rc;
  int			pktsize;
  pkt_desc_t	*old_pdesc;
  int			frags;			/* # of fragments for cur pkts */
  int			frag_flag = 0;		/* fragmentation needed */
  ushort16		offset=0, offset_size = 0;
  ushort16		lport;
  


	/*-------------------------------
 	 * Make sure pkt desc is valid.
	 *-------------------------------
	 */
	if ((rc = pkt_validateDesc( pdesc , 1)) != NFX_OK)
	{
		PKT_LOGERR( rc, pdesc );
		return NFX_ERR;
	}


	/* Su, 09/23/02
	 * forward selective broadcast IP pkts 
	 */
	lport = ntohs(pdesc->lport);
	if (*(ushort16*)pdesc->pMac == 0xffff) 
		if ( !((pdesc->tplProt == TPL_PROT_UDP) && (BSD_BCAST_TO_SSC(lport))))
			return NFX_ERR;

   pktsize = pdesc->pktSize;


   /* fragmentation */
   frags = (pktsize+BSD_BP_MTU-1)/BSD_BP_MTU;/* roundup */
   if (frags > 1)
	 {
	   ipstat.rl_fragedfwd++;
	   frag_flag = 1;
	 }
   
   /* save the income pdesc */
   old_pdesc = pdesc;
   
next_frag:
   if (frag_flag)			/* fragmentation */
	 {
	   char		*ebufp;


	   /* allocate new pkt descriptor */
	   if ((pdesc = PKT_ALLOC_PACKET(PKT_USR_PKT))
		   == NULL)
		 {
		   ipstat.rl_fragsfwderr++;
		   return NFX_ERR;
		 }

	   /* setup various pdesc ptrs */
	   ebufp = pdesc->pHdr - PKT_BUFF_INDENT;/* buf base addr */
	   pdesc->pMac = ebufp + sizeof(rl_hdr_t); /* leave just enough room for the header */

	   /* setup pkt size */
//	   printf("max size is %d\n", BSD_BP_MTU);
	   offset_size = pdesc->pktSize = (pktsize>BSD_BP_MTU)?BSD_BP_MTU:pktsize;
//	   printf("pkt size is %d\n", pdesc->pktSize);
	   

	   /* cpy to new pkt */
	   memcpy(pdesc->pMac, old_pdesc->pMac, pdesc->pktSize);
	   old_pdesc->pMac += pdesc->pktSize; /* adv. ptr, reading for next cpy */

	   frags--;
	   pktsize -= pdesc->pktSize;

	   if (!frags)				/* done fragmentation, free original pkt */
		   {
			   rla->rl_ctrl &= ~BSD_RL_PKT_MORE; 
			   PKT_FREE_PACKET(old_pdesc);
		   }
	   else {
		   rla->rl_ctrl |= BSD_RL_PKT_MORE; /* more to come */
	   }
	   
	   
	   ipstat.rl_fragsfwd++;
	 }
   else							/* single pkts */
   /* bad temparory fix, shift pMac such that there is sizeof(rl_hdr_t) space */
   /* at begining of the buffer. Note that this packet must be starting at */
   /* start of buffer, because this will become 2nd buffer. */
	{
	  char	*intp;

//	 printf("pkt relay: mac ptr %x; ip ptr %x\n", pdesc->pMac, pdesc->pIpHdr);
//	bsd_ShowPkt(pdesc->pMac, pktsize);

	  intp = (char*)eee_ramAlloc(pktsize);
	  if (!intp)
		  return NFX_ERR;
	  
	  memcpy(intp,  pdesc->pMac, pktsize);
     
	  /* shift corresponding ptr, must be start of buffer. 2 is */
	  /* byte left initially */
	  pdesc->pMac +=sizeof(*rl_hdr) -2; 
	  pdesc->pIpHdr += sizeof(*rl_hdr) -2;

      memcpy(pdesc->pMac, intp, pktsize);
      eee_ramDealloc(intp);
 
  //    printf("pkt relay: mac ptr %x; ip ptr %x\n", pdesc->pMac, pdesc->pIpHdr);
  //    bsd_ShowPkt(pdesc->pMac, pktsize);
	}




	
	/* fill in the pkt header */
	/* cpu, slot here is from ipf, which are not */
	/* destination slot, cpu */
	rl_hdr = (rl_hdr_t*)(((char*)pdesc->pMac) - sizeof(rl_hdr_t));
	rl_hdr->rh_dtype = bsdDTypeStoN(ifp->if_name);
	rl_hdr->rh_slot = ifp->if_slot;
	rl_hdr->rh_cpu  = ifp->if_cpu;
	rl_hdr->rh_link = ifp->if_unit;
	rl_hdr->rh_ctrl = rla->rl_ctrl;
	rl_hdr->rh_offset = offset;

	/* update pkt size and hdr size to reflect added relay hdr */
	pdesc->hdrSize += sizeof(*rl_hdr);
	pdesc->pktSize += sizeof(*rl_hdr);

	/* calculate eee-layer src, dest address */
	if (rla->rl_ctrl & BSD_RL_PKT_BCAST)
	  bcast = 1;
	
	/* since broadcast not supported now, send only to ssc */
	/* take this out when bcast supported */
	if (bcast)
	  {
		rla->rl_slot = 0;		/* SSC_SLOT */
		rla->rl_cpu  = NFX_SSC_UNIX;
		bcast = 0;
	  }

	e_dest = PACK_UNICAST_PORT_ADDRESS(
									   rla->rl_slot,
									   rla->rl_cpu,
									   bsdrl_appid,
									   EEE_MGMT,
									   NFX_COS_HIGH,
									   0 );

	e_src = bsdeee_src;

	
	/*-------------------------------
 	 * Setup buf desc. len & buff ptr.
	 *-------------------------------
	 */
	eDescp= pdesc->eDescp;
	eDescp->bd[1].len = pdesc->pktSize;
	eDescp->bd[1].buf = pdesc->pMac - sizeof(*rl_hdr);

	/*-------------------------------
 	 * Save pkt desc in attr.
	 *-------------------------------
	 */
	eDescp->hdr.attr = pdesc;
	
	/*-------------------------------
 	 * Setup dest port to Data link driver.
	 *-------------------------------
	 */
	eDescp->hdr.dest_port = e_dest;
	eDescp->hdr.src_port  = e_src;
    eDescp->hdr.control   = (pdesc->pktSize & EDESC_TOTAL_LEN) |
                            (( 1 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

	if (pdesc->dataSize)
	  {
		/* we should not relay any packet with reference to gmc */
		bsd_log(BSD_ERR_RL_PKT, "bsdrl_RelayPkt: relay pkt reference to gmc\n");
		return NFX_ERR;
	  }

	
	/* forward via eee, pkt ownership given to eee */
	(void) eee_forwardPacket(eDescp);

	/* pdesc now, because backplane driver does not recognize it */
	PKT_FREE_DESC(pdesc);
	
	/* is there next fragments */
	if (frag_flag && frags)
	  {
		offset += offset_size;
		goto next_frag;
	  }


	return NFX_OK;
#endif /* PROM || !ENABLE_RELAY */
}

void
bsdrl_ClrReassemQ(rl_rq_t  *q)
{
  int	i, pos;
  
  pos = q->position;

  for (i=0; i<  pos; i++)
	{
	  if (q->edescp[i])
		eee_freePacket(q->edescp[i]);
	}

  bzero(q,sizeof(rl_rq_t));

  return;
}


/*-----------------------------------------------------------------
 * Name:        bsdrl_RcvPkt
 *
 * Description: rcv pkt at  back plane.  
 *
 *
 * Created by : Su Wang
 *
 * Date Created: 10/22/01
 *-----------------------------------------------------------------
 */
void
bsdrl_RcvPkt(eee_descPtr_t edescp)
{
	uint32				bufcnt;
	uint32				pktlen;  
	uchar8				*pbuf;
	struct ether_header	*eh;				
	rl_hdr_t			*rl_hdr;
	pkt_desc_t			*pdesc;
	struct ifnet		*ifp;
	int					slot, cpu, link, dtype, ctrl;
	ushort16			offset;
	int					*position, i;
	eee_descBuf_t		*bd;
	eee_descPtr_t		tmp_edescp;
	int					start;
	int					fromSlot;
	

	pktlen = (edescp->hdr.control & EDESC_TOTAL_LEN);
    bufcnt = (edescp->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	bd = (edescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
	  &(edescp->bd[0]) :&(edescp->bd[1]);
	pbuf = bd->buf;
//	printf("pkt @ %x recv pdesc %x, flag %x, buflen %d\n", pbuf, edescp, edescp->hdr.control, bd->len);
//	bsd_ShowPkt(pbuf, 128);

	
	/* pbuff 16 bits aligned */
	if ((address_t)pbuf & 1) {
		ipstat.rl_fragsrcverr++;
		eee_freePacket(edescp);
		return;
	}
	

	rl_hdr = (rl_hdr_t*)pbuf;
	slot = rl_hdr->rh_slot;
	cpu = rl_hdr->rh_cpu;
	link = rl_hdr->rh_link;
	dtype = rl_hdr->rh_dtype;
	ctrl = rl_hdr->rh_ctrl;
	offset = rl_hdr->rh_offset;

#if defined(LINK_AGG)
	/* for link aggregation vif, convert active link
	 * to up link
	 */
	link = luc_getUpLALink(link);
#endif /* LINK_AGG */

	/* take away bck plane header */
	pktlen -= sizeof(rl_hdr_t);
	pbuf   += sizeof(rl_hdr_t);
	bd->len -= sizeof(rl_hdr_t);
	bd->buf += sizeof(rl_hdr_t);


	/* dont process any pkt not generated locally */
	if (!(ctrl & BSD_RL_PKT_OWN)) {
		ipstat.rl_fragsrcverr++;
		eee_freePacket(edescp);
		return;
	}
	
		
	if (!(offset) && !(ctrl & BSD_RL_PKT_MORE))
		goto process;			/* single pkt, no need to queue */

	ipstat.rl_fragsrcv++;

	/* where from */
	fromSlot = (edescp->hdr.src_port >> EDESC2_SLOT_SHIFT) & EDESC2_SLOT_MASK;
	
	if ((fromSlot == nfxMySlotId) || 
		(fromSlot >= NFX_NUM_SLOTS_PER_CHASSIS)) {
		ipstat.rl_fragsrcverr++;
		eee_freePacket(edescp);
		return;
	}

	/* put to reassemble queue */
	position = &(bsdrl_ReassemQ[fromSlot].position);

	/* set queue time to live */
	if (*position == 0)
		bsdrl_ReassemQ[fromSlot].ttl = 3;
	
	/* too many buffers queued already, clear queue */
	if (*position >= (EEE_MAX_BUFS - 1))
	  bsdrl_ClrReassemQ(&bsdrl_ReassemQ[fromSlot]);
						
	/* offset not maching, clear queue and return */
	if (offset != bsdrl_ReassemQ[fromSlot].offset)
	  {
		bsdrl_ClrReassemQ(&bsdrl_ReassemQ[fromSlot]);
	
//	    printf("bad offset: slot %d,  offset %d, pkt offset %d\n", 
//				fromSlot, bsdrl_ReassemQ[fromSlot].offset,
//				offset);	
		ipstat.rl_fragsrcverr++;
		eee_freePacket(edescp);
		return;
	  }

	/* OK, offset matches and there is more space, queue frames */
	bsdrl_ReassemQ[fromSlot].offset += pktlen;
//	printf("offset updated slot %d, offset %d, pktlen %d\n", 
//			fromSlot, bsdrl_ReassemQ[fromSlot].offset, pktlen);
	bsdrl_ReassemQ[fromSlot].edescp[*position] = edescp;
	(*position)++;

	
	/* is this last fragment */
	if (ctrl & BSD_RL_PKT_MORE)
	  return;


	/* arranges all frames into first edescp in queue*/
	/* total pkt len */
	pktlen = bsdrl_ReassemQ[fromSlot].offset;
	edescp = bsdrl_ReassemQ[fromSlot].edescp[0];

	/* calcualte total buffer in chain */
	bufcnt = (edescp->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	for (i = 1; i<*position; i++)
	  {
		tmp_edescp = bsdrl_ReassemQ[fromSlot].edescp[i];
		bufcnt += (tmp_edescp->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	  }
	
	if (bufcnt >= (EEE_MAX_BUFS - 1))
	  {
		printf("bufcnt 1\n");
		bsdrl_ClrReassemQ(&bsdrl_ReassemQ[fromSlot]);
		ipstat.rl_fragsrcverr++;
		return;
	  }

	/* ready to reassign buf ptrs */
	start = (edescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)?0:1;
	start += (edescp->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	for (i = 1; i<*position; i++)
	  {
		int	tmpBufCnt;
		
		
		tmp_edescp =  bsdrl_ReassemQ[fromSlot].edescp[i];
		bd = (tmp_edescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)?
		  &(tmp_edescp->bd[0]):&(tmp_edescp->bd[1]);
		tmpBufCnt = (tmp_edescp->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;


		while (tmpBufCnt--) {
			/* too many buffers queued */
			if (start >=  EEE_MAX_BUFS) {
				printf("bad buffer count: #desc %d, bufcnt %d, i %d\n",
					   *position, bufcnt, i);
				
				bsdrl_ClrReassemQ(&bsdrl_ReassemQ[fromSlot]);
				ipstat.rl_fragsrcverr++;
				return;
			}
			
			edescp->bd[start].buf = bd->buf;
			edescp->bd[start].len = bd->len;
			//		printf("buf %x iwthlen %d appended at pos %d\n", bd->buf,bd->len, start);
			bd->buf = NULL;
			bd++;
			start++;
		}

		tmp_edescp->hdr.control &=  ~EDESC_BUFCNT_MASK;
	  }

	/* reset resembly queue, keep the first ddescriptor */
	bsdrl_ReassemQ[fromSlot].edescp[0] = 0;
	bsdrl_ClrReassemQ(&bsdrl_ReassemQ[fromSlot]);

	
	/* set pktlen and bufcnt */
	edescp->hdr.control &= ~EDESC_BUFCNT_MASK;
	edescp->hdr.control |= (bufcnt << EDESC_BUFCNT_SHIFT) &  EDESC_BUFCNT_MASK;
	edescp->hdr.control &= EDESC_TOTAL_LEN;
	edescp->hdr.control |= pktlen &  EDESC_TOTAL_LEN;

	/* need to rearrange buffers, such that each frame start at begining of */
	/* the buffer, execept for the first frame */
	
	/* for now, we take care the simple case only, i.e. 1500 ethernet */
	if ((pktlen > ETHER_MAX_LEN- ETHER_CRC_LEN) ||
		bufcnt > 2)
	  {
//		printf("buf cnt %d, pktlen %d\n", bufcnt, pktlen) ;
		ipstat.rl_fragsrcverr++;
		eee_freePacket(edescp);
		return;
	  }

	bd = (edescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)? &(edescp->bd[0]) : 
		&(edescp->bd[1]);

	
	/* cpy second buffer to tail of the first one */
//	printf("cpy secnd buf %x for len %d\n", (bd+1)->buf, (bd+1)->len);
	
	memcpy((bd->buf + bd->len), (bd+1)->buf, (bd+1)->len);

	/* setup ptr to start of buffer chain */
	pbuf = bd->buf;

	/* remove the second buffer */
	eee_deallocateBuffer((bd+1)->buf);
	(bd+1)->buf = 0;
	edescp->hdr.control &= ~EDESC_BUFCNT_MASK;
	edescp->hdr.control |= (1 << EDESC_BUFCNT_SHIFT) &  EDESC_BUFCNT_MASK;
	ipstat.rl_reasrcv++;

 process:
	/*-------------------------------
     * Allocate pkt desc. If fail
	 * free EEE desc & pkt.
     *-------------------------------
	 */
	if (!(pdesc = PKT_ALLOC_DESC()))
	{
	    eee_freePacket(edescp);
		return;
	}

	pdesc->pMac = pbuf;
	pdesc->pIpHdr = pbuf + ETHER_HDR_LEN;
#if defined(LINUX_TEST)
	/* luc use pHdr in simulated envir */
	pdesc->pHdr = pbuf;
#endif /* LINUX_TEST */

	pdesc->pktSize = pktlen;
	pdesc->hdrSize = pktlen;
	pdesc->dataSize= 0;
	pdesc->cod	   = 0;
	pdesc->eDescp  = edescp;
	pdesc->pNext   = NULL;

	/* determine protocol type */
	eh = (struct ether_header*)pbuf;
	if (ntohs(eh->ether_type) == ETHERTYPE_ARP)
	  {
		/* arp received from bckpln, */
		/* need to send to bsd stack */

		/* 1) case. if chassis self generated, needs to go out */
		if (ctrl & BSD_RL_PKT_OWN)
		  {
			if (slot == nfxMySlotId && 
				cpu  == nfxMyCPUId)
			  {
				ipstat.rl_arprcv++;
				/* expect luc to take over pdesc ownership */
				pdesc->link = link;
				(void)luc_xmtPktReq(link, pdesc); 
				return;
			  }
			else
			  {
				PKT_FREE_PACKET(pdesc);
				ipstat.rl_iprcverr++;
				return;
			  }			  
		  }

		/* 2) case. arp must be received from a remote interface */
		/* give to bsd stack */

		/* sanity check on intf locality */
		if (slot == nfxMySlotId && 
			cpu  == nfxMyCPUId)
		  {
			PKT_FREE_PACKET(pdesc);
			ipstat.rl_iprcverr++;
			return;
		  }

		/* locate the interface */
		ifp = (struct ifnet*)bsd_FindIntfByName
		  ((char*)bsdPrintPort(slot,
							   cpu,
							   link,
							   dtype));
		
		if (!ifp) 
		  {
			ipstat.rl_arprcverr++;
			PKT_FREE_PACKET(pdesc);
			return;
		  }

		/* give to bsd stack */
		ipstat.rl_arprcv++;
		if (bsd_RcvArpPkt(pdesc, ifp, 0, 0) == PKT_FREE_PKT)
			PKT_FREE_PACKET(pdesc);

		return;
		  
	  }
	else if (ntohs(eh->ether_type) == ETHERTYPE_IP)
	  {
		/* ip recived from bckpln */
		/* for now, we don't expect IP pkt to be forwarded to */
		/* bsd stack, this can be changed in the future */

		/* pkt must be on its way out, not to bsd stack */
		if (slot == nfxMySlotId && 
			cpu  == nfxMyCPUId)
		  {
			/* send to luc driver */
			ipstat.rl_iprcv++;
			/* expect luc to take over pdesc ownership */
			pdesc->link = link;
			(void)luc_xmtPktReq(link, pdesc);
			return;
		  }
		else
		  {
			/* drop pkt because interface requested interface */
			/* is not local */
			ipstat.rl_iprcverr++;
			PKT_FREE_PACKET(pdesc);
			return;
		  }
	  }
	else
	  {
		/* unrecoginized protocol, drop pkt */
		ipstat.rl_noproto++;
		PKT_FREE_PACKET(pdesc);
		return;
	  }
}


#if defined(ETH_ARE_BP)
static char	eth_src[]={00, 01, 02, 03, 04, 05};
static char	eth_dest[]={00, 01, 02, 03, 04, 05};

static void bsd_AddBPHdr(char *mac, uint32 dest, uint32 src)
{
  agile_hdr_t	*eee = (agile_hdr_t*)(mac-sizeof(agile_hdr_t));
  struct ether_header *eh = (struct ether_header*)
	(((char*)eee)-sizeof(struct ether_header));

  eee->dest_port = dest;
  eee->src_port  = src;

  memcpy(eh->ether_dhost, eth_dest, 6);
  memcpy(eh->ether_shost, eth_src, 6);
  eh->ether_type = htons(0x700); /* AF_EEE */
  
  return;
}

int32 
bsd_FwdBP(eee_descPtr_t edesc)
{
  int			size;
  eee_descBuf_t *ebuf;

  if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
	ebuf = &(edesc->bd[0]);
  else
	ebuf = &(edesc->bd[1]);

  /* adding ethernet and eee header */
  bsd_AddBPHdr(ebuf->buf, edesc->hdr.dest_port,
			   edesc->hdr.src_port);

  ebuf->buf -= sizeof(agile_hdr_t) + sizeof(struct ether_header);
  ebuf->len += sizeof(agile_hdr_t) + sizeof(struct ether_header);
  size = edesc->hdr.control & EDESC_TOTAL_LEN;
  size += sizeof(agile_hdr_t) + sizeof(struct ether_header);
  edesc->hdr.control  &= ~EDESC_TOTAL_LEN;
  edesc->hdr.control |= (size & EDESC_TOTAL_LEN);

  edesc->hdr.dest_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
												   nfxMyCPUId,
												   bsdrl_appid,
												   EEE_USER_DATA,
												   NFX_COS_HIGH,
												   0);

  /* give to driver directly */
  
  bmc12500Eth_txUnicast( edesc->hdr.dest_port, edesc ); 

  return EEE_FORWARDED_PACKET;
}

void
bsd_RcvBP(eee_descPtr_t eDescp)
{
  eee_descBuf_t		*ebuf;
  agile_hdr_t			*eee;
  int					size;
  
  
  /* change descriptor to reflect real content */
  if (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
	ebuf = &(eDescp->bd[0]);
  else
	ebuf = &(eDescp->bd[1]);		  
  
  eee = (agile_hdr_t*)(ebuf->buf + sizeof(struct ether_header));
  
  
  eDescp->hdr.dest_port = eee->dest_port;
  eDescp->hdr.src_port  = eee->src_port;
  size = eDescp->hdr.control & EDESC_TOTAL_LEN;
  size -= sizeof(agile_hdr_t) + sizeof(struct ether_header);
  eDescp->hdr.control  &= ~EDESC_TOTAL_LEN;
  eDescp->hdr.control |= (size & EDESC_TOTAL_LEN);
  
  /* strip header */
  ebuf->buf += sizeof(agile_hdr_t) + sizeof(struct ether_header);
  ebuf->len -= sizeof(agile_hdr_t) + sizeof(struct ether_header);
  
  /* deliver de-capsulated packet to eee again */
  eee_receivePacket(eDescp);
  return;
}

#endif /* ETH_ARE_BP */
