/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	bsdroute-api.c
 *
 * RCS:  $Id: bsdroute-api.c,v 1.11 2002/10/24 17:46:56 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions to manage route/arp table.
 *
 * Created by:	Su Wang
 *
 * Date Created:	7/25/01
 *
 *-----------------------------------------------------------------
 */
 
#include <nfx-incl.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/errno.h"
#include "../sm-bsd-sys/macros.h"
#include "../sm-bsd-sys/mbuf.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/socketvar.h"
#include "../sm-bsd-sys/sockio.h"
#include "../sm-bsd-sys/protosw.h"

#include "../sm-bsd-net/if.h"
#include "../sm-bsd-net/route.h"
#include "../sm-bsd-net/if_dl.h"
#include "../sm-bsd-net/if_var.h"
#include "../sm-bsd-net/if_arp.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/in_var.h"
#include "../sm-bsd-netinet/in_pcb.h"
#include "../sm-bsd-netinet/ip_var.h"
#include "../sm-bsd-netinet/if_ether.h"

#include "exports/bsd-api.h"
#include "exports/bsdnet-api.h"

//#define MAC_TEST

static char* default_gw = "0.0.0.0";
static char* default_mask = "255.255.255.255";

/* globally defined static route/arp list */
SLIST_HEAD(srt_head, srtentry) bsd_static_rt_list = {0},
  bsd_static_arp_list = {0};

int route_output(struct mbuf *m, struct socket *so);

extern	struct	inpcbhead ripcb;
extern  struct	inpcbhead udb;
extern  struct	inpcbhead tcb;


/*
 *------------------------------------------------------------------
 *
 * Name : bsd_ModRoute
 *
 * Description: Add a route to bsd route domain. The type arg is 
 *				not used for now, since we use this function only
 *				to add static route. In future when more route types
 *				are used, this function needs to be modified. 
 *
 * Created by : Su Wang
 *
 * Date Created: 7/27/01
 *-----------------------------------------------------------------
 */	     
int32 /* NFX_OK, NFX_ERR */
bsd_ModRoute(uint32 ipaddr,  /* in network order */
			 uint32 ipmask,
			 uint32 gateway, /* gateway of this destination */
			 uint32 type,   /* see route types definition */
			 int    action, 
			 void*  misc,
			 void*  param2
			 )
{
  struct rt_msghdr		*rt_msg;
  struct sockaddr_in	*in_addr;
  int					len = sizeof(*rt_msg) + 3*sizeof(*in_addr);
  int					error;


  /* need some sanity check on ip/mask */
  if (MASK_HAS_HOLE(ntohl(ipmask)))
    {
      uiPrintf("IP mask %x has hole \n", ipmask);
      return NFX_ERR;
    }
  
  if (!(rt_msg = (struct rt_msghdr*)malloc(len, MT_DATA, M_DONTWAIT)))
    return NFX_ERR;

  bzero((char*)rt_msg, len);
  
  /* fill in the header */
  rt_msg->rtm_msglen = len;

  if (action == BSD_ADD)
	rt_msg->rtm_type = RTM_ADD;
  else if (action == BSD_DEL)
	rt_msg->rtm_type = RTM_DELETE;
  else
	return NFX_ERR;

  rt_msg->rtm_version = RTM_VERSION;
  rt_msg->rtm_flags = RTF_GATEWAY | RTF_STATIC;
  if (ipmask == 0xffffffff)
    rt_msg->rtm_flags |= RTF_HOST;
  rt_msg->rtm_addrs = RTA_DST | RTA_GATEWAY;

  /* fill in dst, gateway, and mask */
  in_addr = (struct sockaddr_in*)(rt_msg + 1);

  in_addr->sin_len = sizeof(*in_addr);
  in_addr->sin_family = AF_INET;
  in_addr->sin_addr.s_addr = ipaddr;

  in_addr++;
  in_addr->sin_len = sizeof(*in_addr);
  in_addr->sin_family = AF_INET;
  in_addr->sin_addr.s_addr = gateway;

  if (ipmask != 0xffffffff)
    { 
      in_addr++;
      in_addr->sin_len = 8;
      in_addr->sin_family = AF_INET;
      in_addr->sin_addr.s_addr = ipmask;
      rt_msg->rtm_addrs |= RTA_NETMASK;
    }  

  error = route_output((struct mbuf*)rt_msg, 0);

  free(rt_msg, MT_DATA);

#if defined(BSD_SAVE_STATIC)
  /* add/del static routes. Adding if */
  /* 1) route added to route table successfully */
  /* 2) BSD_RTMISC_NO_STATIC is not set */
  /* del is performed with no restriction.*/
  if ((action == BSD_ADD) && (error == 0) &&  
	  ((uint32)misc & BSD_RTMISC_NO_STATIC) == 0)
	{
	  if (bsd_ChgeStaticRt(BSD_ADD, &bsd_static_rt_list, ipaddr, ipmask,
							  &gateway, AF_INET) != NFX_OK)
		bsd_log(BSD_WARN_IPRT, "add route %x to static list failed\n", 
					 ipaddr);	  
	}
  else if (action == BSD_DEL)
	{
	  if (bsd_ChgeStaticRt(BSD_DEL, &bsd_static_rt_list, ipaddr, ipmask,
							  &gateway, AF_INET) != NFX_OK)
		bsd_log(BSD_WARN_IPRT, "del route %x from static list failed\n", 
				ipaddr);		  
	}
#endif /* BSD_SAVE_STATIC */

  if (error)
    {
     bsd_log(BSD_WARN_IPRT, 
			 "route dst %x mask %x gateway %x fail to be added/deleted\n",
			 ipaddr, ipmask, gateway);
      return NFX_ERR;
    }
  else
	  return NFX_OK;

}


/*
 *------------------------------------------------------------------
 *
 * Name : bsd_FindRt
 *
 * Description: find a route of given type
 *
 * Created by : Su Wang
 *
 * Date Created: 7/27/01
 *-----------------------------------------------------------------
 */	  
void*
bsd_FindRt(	uint32 ipaddr, uint32 ipmask, uint32 gw, uint32 flags)
{
  struct sockaddr_in in_addr, in_mask;
  struct rtentry *rt;
  struct radix_node_head *rnh;

  if ((rnh = rt_tables[AF_INET]) == 0) 
    {
      bsd_log(BSD_ERR_IPRT, "rt_table not initialized \n");
      return NULL;
    };
  
  bzero((char*)&in_addr, sizeof(in_addr));
  bzero((char*)&in_mask, sizeof(in_mask));

  in_addr.sin_len = sizeof(in_addr);
  in_addr.sin_family = AF_INET;
  in_addr.sin_addr.s_addr = ipaddr & ipmask;

  in_mask.sin_len = 8;
  in_mask.sin_addr.s_addr = ipmask;

  rt = (struct rtentry*)rnh->rnh_lookup(&in_addr, 
						 (ntohl(ipmask) == 0xffffffff)? 0 : &in_mask, rnh);

  if (rt && rt->rt_flags & flags)
	  return rt;
  else
	  return NULL;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
static int 
bsdPrintRouteDetail(struct rtentry *rt)
{
  struct sockaddr_in*  in_addr;
  struct sockaddr_dl*  dl_addr;
  char* gw = default_gw;
  char* mask = default_mask;


  in_addr = (struct sockaddr_in*)rt_key(rt);
  bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
  uiPrintf("Dest Address    :       %s\n", prnt_ip);

  in_addr = (struct sockaddr_in*)rt_mask(rt);
  if (in_addr)
    {
      bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
      mask = prnt_ip;
    }
  uiPrintf("Dest Mask       :       %s\n", mask);
  
  if ((rt->rt_gateway))
    {
      if (rt->rt_gateway->sa_family == AF_INET)
		{
		  in_addr = (struct sockaddr_in*)rt->rt_gateway;
		  bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
		  gw = prnt_ip;
		}
      else if (rt->rt_gateway->sa_family == AF_LINK)
		{
		  dl_addr = (struct sockaddr_dl*)rt->rt_gateway;
		  bsdMacNtoS(LLADDR(dl_addr), prnt_mac, sizeof(prnt_mac));
		  gw = prnt_mac;
		}
    }

  uiPrintf("Gateway         :       %s\n",gw);

  
  uiPrintf("Reference Count :        %d\n", rt->rt_refcnt);
  uiPrintf("Packet Send     :        %d\n", rt->rt_use);
  uiPrintf("Dev. Interface  :        %s(%lx)\n", 
		   bsdPrintIntfWithVif(rt->rt_ifp), (address_t)rt->rt_ifp);

  in_addr = (struct sockaddr_in*)rt->rt_ifa->ifa_addr;
  bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
  uiPrintf("IP Interface    :        %s(%lx)\n", prnt_ip, (address_t)rt->rt_ifa);

  if (rt->rt_genmask)
    {
      in_addr = (struct sockaddr_in*)rt->rt_genmask;
      bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
      uiPrintf("Gen Mask        :        %s\n", prnt_ip);
    }
  else
    uiPrintf("Gen Mask        :        \n");
  
  uiPrintf("Link Cache      :        %lx\n", (address_t)rt->rt_llinfo);
  uiPrintf("Gateway Rt.     :        %lx\n", (address_t)rt->rt_gwroute);
  
  return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
static int 
bsdPrintRouteSummary(struct radix_node *r_node, void* inflags)
{
  uint32 flags = *(uint32*)inflags;
  struct rtentry* rt = (struct rtentry*)r_node;
  struct sockaddr_in*  in_addr;
  struct sockaddr_dl*  dl_addr;
  char*  gw = default_gw; 
  int masklen;

  if (rt->rt_flags & flags)
    {
      
      in_addr = (struct sockaddr_in*)rt_key(rt);
      bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
      
      in_addr = (struct sockaddr_in*)rt_mask(rt);
      if (in_addr)
		{
		  masklen = bsdMasklen(&(in_addr->sin_addr));
		}
      else
		masklen = 32;

      uiPrintf("%-15s/%-5d", prnt_ip, masklen);

      if ((rt->rt_gateway))
		{
		  if (rt->rt_gateway->sa_family == AF_INET)
			{
			  in_addr = (struct sockaddr_in*)rt->rt_gateway;
			  bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
			  gw = prnt_ip;
			}
		  else if (rt->rt_gateway->sa_family == AF_LINK)
			{
			  dl_addr = (struct sockaddr_dl*)rt->rt_gateway;
			  bsdMacNtoS(LLADDR(dl_addr), prnt_mac, sizeof(prnt_mac));
			  gw = prnt_mac;
			}
		}

      uiPrintf("%-22s0x%-8x%-20s%u(%lx)\n",
			   gw,
			   rt->rt_flags,    
			   bsdPrintIntfWithVif(rt->rt_ifp),
			   rt->rt_refcnt,
			   (address_t)rt);
    }
  return NFX_OK;
}  


/*
 *------------------------------------------------------------------
 *
 * Name : bsd_ShowRoute
 *
 * Description: Add a route to bsd route domain 
 *
 * Created by : Su Wang
 *
 * Date Created: 7/27/01
 *-----------------------------------------------------------------
 */	  
void
bsd_ShowRoute(
			  int    how,
			  uint32 type,  /* see flag definition */ 
			  uint32 ipaddr, /* display a single route, 0xffffffff all routes */
			  uint32 ipmask, /* given if want to display a single route */
			  void*  param1,
			  void*  param2
			  )
{
  struct sockaddr_in in_addr, in_mask;
  struct rtentry *rt;
  struct radix_node_head *rnh;
  uint32 flags = 0;


  if ((rnh = rt_tables[AF_INET]) == 0) 
    {
      bsd_log(BSD_ERR_IPRT, "rt_table not initialized \n");
      return;
    };
  
  bzero((char*)&in_addr, sizeof(in_addr));
  bzero((char*)&in_mask, sizeof(in_mask));

  if (how == BSD_SINGLE) /* single route case */ 
    {
      in_addr.sin_len = sizeof(in_addr);
      in_addr.sin_family = AF_INET;
      in_addr.sin_addr.s_addr = ipaddr & ipmask;

      in_mask.sin_len = 8;
      in_mask.sin_addr.s_addr = ipmask;

      if ((rt = (struct rtentry *)
		   rnh->rnh_lookup(&in_addr, 
						   (ntohl(ipmask) == 0xffffffff)? 0 : &in_mask, rnh)) == 0)
		return;


      bsdPrintRouteDetail(rt);
    }
  else /* multiple route case */
    {
      uiPrintf("Route Table\n");
      uiPrintf("----------------------------------------------------------------------------------\n");
      uiPrintf("Dest                 Gateway               Flags    Intf                  refc(id)   \n");
      uiPrintf("----------------------------------------------------------------------------------\n");

	  if (type & IPRT_LOCAL)
		flags |= RTF_LOCAL;
	  
	  if (type & IPRT_HOST)
		flags |= RTF_HOST;
	  
	  if (type & IPRT_STATIC)
		flags |= RTF_STATIC;

	  if (type == 0)
		flags = 0xffffffff;

      rnh->rnh_walktree(rnh, bsdPrintRouteSummary, &flags);

	}
  return;
}


/*
 *------------------------------------------------------------------
 *
 * Name : bsd_ModARP
 *
 * Description: Add/Delete an ARP entry to bsd route domain 
 *
 * Created by : Su Wang
 *
 * Date Created: 7/27/01
 *-----------------------------------------------------------------
 */	     
int32 /* NFX_OK, NFX_ERR */
bsd_ModARP (uint32 ipaddr,  /* in network order */
			char*  mac, /* mac address in network order */
			int    action,
			void*  misc,
			void*  param2
			)
{
  struct rt_msghdr		*rt_msg;
  struct sockaddr_in	*in_addr;
  struct sockaddr_dl	*dl_addr;
  int					len =sizeof(*rt_msg)+sizeof(*in_addr)+sizeof(*dl_addr);  
  int					error;

  if (!(rt_msg = (struct rt_msghdr*)malloc(len, MT_DATA, M_DONTWAIT)))
    return NFX_ERR;

  bzero((char*)rt_msg, len);

  /* fill in the header */
  rt_msg->rtm_msglen = len;

  if (action == BSD_ADD)
	rt_msg->rtm_type = RTM_ADD;
  else if (action == BSD_DEL)
	rt_msg->rtm_type = RTM_DELETE;
  else
	return NFX_ERR;

  rt_msg->rtm_version = RTM_VERSION;
  rt_msg->rtm_flags = RTF_STATIC | RTF_HOST;
  rt_msg->rtm_addrs = RTA_DST | RTA_GATEWAY;

  in_addr = (struct sockaddr_in*)(rt_msg + 1);
  
  /* ip address */
  in_addr->sin_len = sizeof(*in_addr);
  in_addr->sin_family = AF_INET;
  in_addr->sin_addr.s_addr = ipaddr;  

  /* mac addr */
  dl_addr = (struct sockaddr_dl*)(in_addr + 1);
  dl_addr->sdl_len = sizeof(*dl_addr);
  dl_addr->sdl_family = AF_LINK;
  dl_addr->sdl_nlen = 0           ;
  dl_addr->sdl_slen = 0;
  dl_addr->sdl_alen = 6;
  memcpy(dl_addr->sdl_data, mac, 6);


  error = route_output((struct mbuf*)rt_msg, 0);

  free(rt_msg, MT_DATA);

#if defined(BSD_SAVE_STATIC)
  /* add/del static routes. Adding if */
  /* 1) route added to route table successfully */
  /* 2) BSD_RTMISC_NO_STATIC is not set */
  /* del is performed with no restriction.*/
  if ((action == BSD_ADD) && (error == 0) && 
	  ((uint32)misc & BSD_RTMISC_NO_STATIC) == 0)
	{
	  if (bsd_ChgeStaticRt(BSD_ADD, &bsd_static_arp_list, ipaddr, 0xffffffff,
							  mac, AF_LINK) != NFX_OK)
		bsd_log(BSD_WARN_IPRT, "add arp %x to static list failed\n", 
					 ipaddr);	  
	}
  else if (action == BSD_DEL)
	{
	  if (bsd_ChgeStaticRt(BSD_DEL, &bsd_static_arp_list, ipaddr, 0xffffffff,
							  mac, AF_LINK) != NFX_OK)
		bsd_log(BSD_WARN_IPRT, "del arp %x from static list failed\n", 
				ipaddr);		  
	}
#endif /* BSD_SAVE_STATIC */
  
  if (error)
    {
      bsd_log(BSD_WARN_IPRT, "ModARP, arp %lx fail to be added/removed\n",
				   ipaddr);
      return NFX_ERR;
    }
  else
	  return NFX_OK;

}

struct llinfo_arp {
  LIST_ENTRY(llinfo_arp) la_le;
  struct	rtentry *la_rt;
  struct	mbuf *la_hold;		/* last packet until resolved/timeout */
  long	la_asked;		/* last time we QUERIED for this addr */
};

extern 	LIST_HEAD(, llinfo_arp) llinfo_arp;
/*
 *------------------------------------------------------------------
 *
 * Name : bsd_ShowARP
 *
 * Description: Show ARP table. 
 *
 * Created by : Su Wang
 *
 * Date Created: 7/27/01
 *-----------------------------------------------------------------
 */	  
void
bsd_ShowARP (  uint32 ip_ran, /* network range  */
			   uint32 mask_ran, /* network range */
			   void*  param1,
			   void*  param2
			   )
{
  struct llinfo_arp *la;
  struct rtentry *rt;
  struct ifnet   *ifp;
  struct sockaddr_in*  in_addr;
  struct sockaddr_dl*  dl_addr;
  
  uiPrintf("Summary of ARP Table\n");
  uiPrintf("-----------------------------------------------------------------\n");
  uiPrintf("IP address              MAC                  Interface               Age \n");
  uiPrintf("-----------------------------------------------------------------\n");
  for (la = llinfo_arp.lh_first; la ; la = la->la_le.le_next)
    {
      if (!(rt = la->la_rt))
		{
		  debugPrintf(("showarp,arp  bad structure1\n"));
		  return;
		}

      /* the correspond route entry must be a 
		 local route, must point back to current la,
		 must has a valid pointer to a interface */
      if ((rt->rt_flags & RTF_GATEWAY) || 
		  (rt->rt_llinfo != (char*)la) ||
		  !(ifp = rt->rt_ifp))
		{
		  debugPrintf(("showarp,route bad structure1\n"));
		  return;	
		}

      
      in_addr = (struct sockaddr_in*)rt_key(rt);
      bsdIpNtoS((uint32*)&(in_addr->sin_addr), prnt_ip, sizeof(prnt_ip));
      
      dl_addr = (struct sockaddr_dl*)rt->rt_gateway;
      bsdMacNtoS(LLADDR(dl_addr), prnt_mac, sizeof(prnt_mac));      
      
      uiPrintf("%-24s%-21s%-20s%d\n",
			   prnt_ip, prnt_mac, bsdPrintIntfWithVif(ifp), rt->rt_rmx.rmx_expire);
	  
    }

  return;
}

/*
 *------------------------------------------------------------------------------
 * 
 * Name: bsd_UpdtPcbRt
 *
 * Descriptions: called when route info. in each PCB need to updated. 
 *               This happens when ARP changes, and/or route changes, or
 *				 interface state changes.
 *
 * Created By : Su Wang
 * 
 * Date Created: 8/22/01
 *------------------------------------------------------------------------------
 */
void
bsd_UpdtPcbRt(
			  int		type,	/* types of info changed */
			  int		action,	/* what change? add/del/mod etc */
			  void		*info,	/* ptr to info, *rt or *ifaddr */
			  void*		param1,
			  void*		param2
			  )
{
  int					again = 0;
  struct inpcbhead		head;
  struct inpcb			*inp, *nxt_inp;
  struct ifaddr			*ifa = 0;
  struct rtentry		*rt0 = 0,  *rt = 0;
  int					ret;

  /* make sure info type and action matchs */
  if (type == BSD_UPDT_INTF)
	{
	  if (action != BSD_DOWN)
		return;
	  ifa = (struct ifaddr*)info;
	}
  else if (type == BSD_UPDT_ARP)
	{
	  if (action != BSD_MOD)
		return;
	  rt0 = (struct rtentry*)info;
	}
  else if (type == BSD_UPDT_RT)
	{
	  if (action != BSD_DEL)
		return;
	  rt0 = (struct rtentry*)info;
	}
  else
	return;


  /* go through udp, tcp, and raw-ip pcb lists */ 
again:
  if(again == 0)
	head = udb;
  else if (again == 1)
	head = tcb;
  else if (again == 2)
	head = ripcb;
  else
	return;
  
  inp = head.lh_first;
  while(inp)
	{
	  /* need to save next inp, because */
	  /* current inp may be deleted in this loop */
	  nxt_inp = inp->inp_list.le_next;	
	  
	  if (!(inp->inp_socket->so_state & SS_ISCONNECTED))
		goto loop;

	  if (!(rt = inp->inp_route.ro_rt))
		goto loop;;


	  if (type == BSD_UPDT_INTF)
		{
		  /* currently, this case never got called */
		  if (rt->rt_ifa == ifa)
			{
			  if (inp->inp_socket->so_connevent)
				{
				  inp->inp_socket->so_connevent(inp->inp_socket, inp, 
												SO_CONN_DOWN_IND, NULL);
				}
			}
		}
	  else if (type == BSD_UPDT_ARP)
		{
		  /* arp route must be a direct route */
		  if (rt->rt_flags & RTF_GATEWAY)
			rt = rt->rt_gwroute;
		  
		  if (rt == rt0)
			{
#if !defined(MAC_TEST)
			ret = bsd_ChkPcbRt(inp);
#else
		      memcpy((char*)inp->inp_nhmac,
					 LLADDR((struct sockaddr_dl*)rt->rt_gateway), 6);
				ret = NFX_OK;
#endif /* MAC_TEST */
			  if (inp->inp_socket->so_connevent)
				{
					if (ret == NFX_OK) {
						inp->inp_flags &= ~INP_NOMAC;
						inp->inp_socket->so_connevent(inp->inp_socket, inp, 
													  SO_CONN_RTCHG_IND, 
													  &(inp->inp_socket->so_handle));
					}
					
			//	  else
			//		inp->inp_socket->so_connevent(inp->inp_socket, inp, 
			//									  SO_CONN_DOWN_IND, 
			//									  &(inp->inp_socket->so_handle));
				}
			}
		}
	  else if (type == BSD_UPDT_RT)
		{
		  int updt = 0;
		  
		  /* pcb need to be updated, if rt or gw-rt */
		  /* has been deleted */
		  if (rt == rt0)
			updt = 1;
		  else if (rt->rt_flags & RTF_GATEWAY)
			if (rt0 == rt->rt_gwroute)
			  updt = 1;

		  /* if new route found, update. Otherwise */
		  /* close the connection */
		  if (updt == 1)
			{
			  ret = bsd_ChkPcbRt(inp);
			  if (inp->inp_socket->so_connevent)
				{
				  if (ret == NFX_OK)
					inp->inp_socket->so_connevent(inp->inp_socket, inp, 
												  SO_CONN_RTCHG_IND, 
												  &(inp->inp_socket->so_handle));
				 // else
				//	inp->inp_socket->so_connevent(inp->inp_socket, inp, 
				//								  SO_CONN_DOWN_IND, 
				//								  &(inp->inp_socket->so_handle));
				}
			}
		}

loop:
	  inp = nxt_inp;
	}
  
  again++;
  goto again;
}

#if defined(MAC_TEST)
int	first = 4;
#endif /* MAC_TEST */

/*
 *------------------------------------------------------------------------------
 * 
 * Name: bsd_ChkPcbRt
 *
 * Descriptions: called to retrive next hop infor when a socket is first
 *               connected. This function will try to establish a route 
 *				 to destination specified. 1) it check to see if 
 *				 the route is no longer up. If so, free the route, and then
 *				 follows, 2) if the route is empty, if so type to allocate
 *				 an route. If route not found, exit; if found, 3) if the 
 *				 the route is gateway route, if so, 4) if the gateway route
 *				 is up, if not, allocate a new gateway route.5) if 		 
 *
 *
 * Created By : Su Wang
 * 
 * Date Created: 8/15/01
 *------------------------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_ChkPcbRt(struct inpcb* inp)
{
  struct route		*ro = &(inp->inp_route);
  struct rtentry*	rt = ro->ro_rt;

  if (rt && (rt->rt_flags & RTF_UP) == 0)
	{
	  RTFREE(rt);
	  ro->ro_rt = rt = 0;
	}
   
  if (!rt)
    {
      /* no route setup yet, try to aquire one */
	  memset(&ro->ro_dst, 0, sizeof(struct sockaddr));
      ro->ro_dst.sa_family = AF_INET;
      ro->ro_dst.sa_len = sizeof(struct sockaddr_in);
      ((struct sockaddr_in *) &ro->ro_dst)->sin_addr = inp->inp_faddr;
      rtalloc_ign(ro, RTF_PRCLONING);
      if ((rt = ro->ro_rt) == 0)
		{
		  inp->inp_nhip = 0;
		  debugPrintf(("bsd_ChkPcbRt: route not found\n"));
		  return NFX_ERR;
		}
    }

  if (rt->rt_flags & RTF_GATEWAY)
	{
	  if (rt->rt_gwroute == 0)
		goto lookup;
	  if (((rt = rt->rt_gwroute)->rt_flags & RTF_UP) == 0) 
		{
		  rtfree(rt);
		  rt = ro->ro_rt;
		  
lookup:
		  rt->rt_gwroute = rtalloc1(rt->rt_gateway, 1, 0UL);
		  if ((rt = rt->rt_gwroute) == 0)
			{
			  inp->inp_nhip = 0;
			  debugPrintf(("bsd_ChkPcbRt: gw route not found\n"));
			  return NFX_ERR;			  
			}
		}
	}

  inp->inp_nhip = ((struct sockaddr_in*)rt_key(rt))->sin_addr.s_addr;

#if defined(MAC_TEST)
      if (first &&
          inp->inp_lport == htons(2049)) {
          first--;
          ((struct sockaddr_dl*)rt->rt_gateway)->sdl_alen = 0;
		}
#endif /* MAC_TEST */
	
  if (((struct sockaddr_dl*)rt->rt_gateway)->sdl_alen) 
	{
	  /* valid mac address, otherwise use broadcast for now */
	  memcpy((char*)inp->inp_nhmac, 
			 LLADDR((struct sockaddr_dl*)rt->rt_gateway), 6);
	}
  else
    { 
	  /* we need to send out an arp request */
	  /* the problem is if this arp or its reply */
	  /* is lost somethere, the application is stucked */
	  /* with broadcast mac */
	  bsd_SendArp((void*)inp);

	  /* potentially, there can be two arp send in one sec interval. */
	  /* One here, the other one at bsd_XmtPkt. This is ok for now, */
	  /* the correct implementation would have been just allow one */
	  /* arp request in a 1-sec interval by set rt_expire and */
	  /* la_asked */

	  /* pcb does not have mac */
	  inp->inp_flags |= INP_NOMAC;
	  memset(inp->inp_nhmac, 0xff, 6);
	  if (rt->rt_ifp != loif)
		  rt->rt_flags |= RTF_WANT_MAC; /* flag to update pcbs when rcv reply */
    }

  return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 *
 * Name: bsd_SendArp
 *
 * Descriptions: allow application to send arp.
 *
 *
 * Created By : Su Wang
 *
 * Date Created: 04/25/02
 *------------------------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
bsd_SendArp(void *in_inp)
{
	struct inpcb	*inp = (struct inpcb*)in_inp;
	struct route	*ro;
	struct rtentry* rt;
	char			data[6];
	struct sockaddr_in inAddr;

	if (!inp)
		return NFX_ERR;
	
	ro = &(inp->inp_route);
	rt = ro->ro_rt;
	
	if (!rt)
		return NFX_ERR;

	inAddr.sin_len = 16;
	inAddr.sin_family = AF_INET;
	inAddr.sin_addr.s_addr = inp->inp_nhip;
	

	arpresolve((struct arpcom*)rt->rt_ifp,
			   rt, 0, (struct sockaddr*)&inAddr, data, NULL);
	
	return NFX_OK;
}


#if defined(BSD_SAVE_STATIC)
/*
 *------------------------------------------------------------------------------
 * 
 * Name: bsd_ChgeStaticRt
 *
 * Descriptions: add or delete an entry on static rt/arp list
 *
 *
 * Created By : Su Wang
 * 
 * Date Created: 8/22/01
 *------------------------------------------------------------------------------
 */
static int32					/* NFX_OK, NFX_ERR */
bsd_ChgeStaticRt(
					int					action,
					struct srt_head		*head, 
					uint32				ipaddr,
					uint32				ipmask,
					char				*gw,
					int					gw_type
					)
{
  srt_ptr_t				srt;
  struct sockaddr_in	in_addr = {sizeof(in_addr), AF_INET};
  int					gw_size = 0;
  uint32				taddr = 0;

  if (gw_type == AF_LINK)
	{
	  gw_size = 6;
	  taddr = ipaddr;
	}
  else if (gw_type == AF_INET) 
	{ 
	  gw_size = 4;
	  taddr = *(uint32*)gw;
	}
  else
	return NFX_ERR;
		   

  /* see if there already is a entry */
  for (srt = head->slh_first; 
	   srt; srt = srt->next.sle_next)
	{
	  if ((srt->ipaddr == ipaddr) && (srt->ipmask == ipmask) &&
		  (bcmp(gw, srt->gw, gw_size) == 0))
		break;
	}

  
  if (action == BSD_ADD)
	{
	  if (!srt)
		srt = (srt_ptr_t)
		  malloc(sizeof(srt_t),MT_DATA,M_DONTWAIT);
	  
	  /* overwrite the old entry */
	  if (srt)
		{ 
		  srt->next.sle_next = 0;
		  srt->ipaddr = ipaddr;
		  srt->ipmask = ipmask;
		  memcpy(srt->gw, gw, gw_size);
		  in_addr.sin_addr.s_addr = taddr;
		  /* need to know which intf this route belongs to */
		  srt->ifa = ifa_ifwithnet((struct sockaddr*)&in_addr);
		  
		  SLIST_INSERT_HEAD(head, srt, next);
		  return NFX_OK;
		}
	}
  else if (action == BSD_DEL)
	{
	  if (srt)
		{
		  SLIST_REMOVE(head, srt, srtentry, next);
		  free(srt, MT_DATA);
		  return NFX_OK;
		}
	}
  return NFX_ERR;
  
}

/*
 *------------------------------------------------------------------------------
 * 
 * Name: bsd_RestoreStaticRoute
 *
 * Descriptions: called when an interface just come up or during config.
 *				 restore process to add static route back to route table.
 *
 *
 * Created By : Su Wang
 * 
 * Date Created: 8/22/01
 *------------------------------------------------------------------------------
 */
void
bsd_RestoreStaticRoute(void* ifa)
{
  
  srt_ptr_t		srt;

  for (srt = bsd_static_arp_list.slh_first; srt; srt = srt->next.sle_next)
	{
	  if ((ifa != 0) && (ifa != srt->ifa))
		  continue;

	  bsd_ModARP(srt->ipaddr, srt->gw, BSD_ADD, (void*)BSD_RTMISC_NO_STATIC, NULL);
	}

  for (srt = bsd_static_rt_list.slh_first; srt; srt = srt->next.sle_next)
	{
	  if ((ifa != 0) && (ifa != srt->ifa))
		  continue;

	  bsd_ModRoute(srt->ipaddr, srt->ipmask, *(uint32*)(srt->gw), IPRT_STATIC,
				   BSD_ADD, (void*)BSD_RTMISC_NO_STATIC, NULL);
	}  
  
  return;
}


/*
 *------------------------------------------------------------------------------
 * 
 * Name: bsd_PurgeStaticRoute
 *
 * Descriptions: called when an IP interface is to be removed 
 *
 *
 * Created By : Su Wang
 * 
 * Date Created: 8/22/01
 *------------------------------------------------------------------------------
 */
void
bsd_PurgeStaticRoute(
					 void		*ifa /* ptr to intf addr, 0 restore all routes */
					 )
{
  srt_ptr_t		srt, nxt_srt;


  srt = bsd_static_arp_list.slh_first;
  while (srt)
	{
	  nxt_srt = srt->next.sle_next;
	  if ((ifa == 0) || (ifa == srt->ifa))
		{
		  SLIST_REMOVE(&bsd_static_arp_list, srt, srtentry, next);
		  free(srt, MT_DATA);
		}
	  srt = nxt_srt;
	}

  srt = bsd_static_rt_list.slh_first;
  while (srt)
	{
	  nxt_srt = srt->next.sle_next;
	  if ((ifa == 0) || (ifa == srt->ifa))
		{
		  SLIST_REMOVE(&bsd_static_rt_list, srt, srtentry, next);
		  free(srt, MT_DATA);
		}
	  srt = nxt_srt;
	}  

  return;  
}


/*
 *------------------------------------------------------------------------------
 * 
 * Name: bsd_ShowStaticRoute
 *
 * Descriptions: show current static route and static arp.
 *
 *
 * Created By : Su Wang
 * 
 * Date Created: 8/23/01
 *------------------------------------------------------------------------------
 */
void
bsd_ShowStaticRoute()
{
  
  srt_ptr_t srt;

  uiPrintf("Summary of static route\n");
  uiPrintf("---------------------------------------------------------------\n");
  uiPrintf("Dest              Mask              gateway           ifa\n");
  for (srt = bsd_static_arp_list.slh_first; srt; srt = srt->next.sle_next)
	{
	  uiPrintf("%-18s", bsdPrintIpa(srt->ipaddr));
	  uiPrintf("%-18s", bsdPrintIpa(srt->ipmask));
	  uiPrintf("%-18s%lx\n", bsdPrintMac(srt->gw), srt->ifa);
	}

  for (srt = bsd_static_rt_list.slh_first; srt; srt = srt->next.sle_next)
	{
	  uiPrintf("%-18s", bsdPrintIpa(srt->ipaddr));
	  uiPrintf("%-18s", bsdPrintIpa(srt->ipmask));
	  uiPrintf("%-18s%lx\n", bsdPrintIpa(*(uint32*)srt->gw), srt->ifa);
	} 

  return;
  
}
#endif /* BSD_SAVE_STATIC */
