/*  File Name  :    agilefsetdquo.c
 *
 *  Description:    HTTPD action functions for display/modify default fileset quota.
 *
 *  Created    :    3/20/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>
#include <evm_com_defs.h>
#include <evm-api.h>
#include <fs-api.h>

#include "agilepost.h"
#include "agileutil.h"


#define INVALID_TAG			0
#define VOL_ID				1
#define VOL_NAME			2
#define PATH_NAME			3
#define PATH_FISOFT			4
#define PATH_FIHARD			5
#define PATH_FITIME			6
#define PATH_FBSOFT			7
#define PATH_FBHARD			8
#define PATH_FBTIME			9
#define PATH_GISOFT			10
#define PATH_GIHARD			11
#define PATH_GITIME			12
#define PATH_GBSOFT			13
#define PATH_GBHARD			14
#define PATH_GBTIME			15
#define PATH_UISOFT			16
#define PATH_UIHARD			17
#define PATH_UITIME			18
#define PATH_UBSOFT			19
#define PATH_UBHARD			20
#define PATH_UBTIME			21
#define NUM_OF_TAGS			21

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{VOL_ID,		"args.vol_id"},
	{VOL_NAME,		"args.vol_name"},
	{PATH_NAME,		"args.path"},
	{PATH_FISOFT,	"fileset_entry.fs.inodeSoftLimit"},
	{PATH_FIHARD,	"fileset_entry.fs.inodeHardLimit"},
	{PATH_FITIME,	"fileset_entry.fs.inodeTimeLimit"},
	{PATH_FBSOFT,	"fileset_entry.fs.blockSoftLimit"},
	{PATH_FBHARD,	"fileset_entry.fs.blockHardLimit"},
	{PATH_FBTIME,	"fileset_entry.fs.blockTimeLimit"},
	{PATH_GISOFT,	"fileset_entry.grp.inodeSoftLimit"},
	{PATH_GIHARD,	"fileset_entry.grp.inodeHardLimit"},
	{PATH_GITIME,	"fileset_entry.grp.inodeTimeLimit"},
	{PATH_GBSOFT,	"fileset_entry.grp.blockSoftLimit"},
	{PATH_GBHARD,	"fileset_entry.grp.blockHardLimit"},
	{PATH_GBTIME,	"fileset_entry.grp.blockTimeLimit"},
	{PATH_UISOFT,	"fileset_entry.usr.inodeSoftLimit"},
	{PATH_UIHARD,	"fileset_entry.usr.inodeHardLimit"},
	{PATH_UITIME,	"fileset_entry.usr.inodeTimeLimit"},
	{PATH_UBSOFT,	"fileset_entry.usr.blockSoftLimit"},
	{PATH_UBHARD,	"fileset_entry.usr.blockHardLimit"},
	{PATH_UBTIME,	"fileset_entry.usr.blockTimeLimit"},
};


#define PARAM_VOLID			1
#define PARAM_VOLNAME		2
#define PARAM_PATH			3
#define NUM_OF_PARAMS		3

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_VOLID,	"id"},
	{PARAM_VOLNAME,	"vn"},
	{PARAM_PATH,	"path"},
};


/* HERE ARE VARIABLES FOR FILESET INFORMATION */
static char bufpath[FS_PATH_MAX+1];
static char bufvol[VOL_MAX_NAME+1];
static uint64 inId;
static fs_QuotaBlk_t defdata;
static fs_QuotaBlk_t grpdata;
static fs_QuotaBlk_t usrdata;


/***********************************************************
 * routines to retrieve fileset information 
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_VOLID:
			sscanf(argval[argind], "%lld", &inId);
			break;
		case PARAM_VOLNAME:
			strcpy(bufvol, argval[argind]);
			break;
		case PARAM_PATH:
			strcpy(bufpath, argval[argind]);
			break;
		}
	}

	return 0;
}

static void getFileSetInfo()
{
    int						stat;
	fs_MsgGetQuotaReq_t		req;
	fs_MsgGetQuotaResp_t	rsp;
	struct timeval			timeout = {60, 0};

	defdata.inodeSoftLimit = 0;
	defdata.inodeHardLimit = 0;
	defdata.inodeTimeLimit = 0;
	defdata.blockSoftLimit = 0;
	defdata.blockHardLimit = 0;
	defdata.blockTimeLimit = 0;
	grpdata.inodeSoftLimit = 0;
	grpdata.inodeHardLimit = 0;
	grpdata.inodeTimeLimit = 0;
	grpdata.blockSoftLimit = 0;
	grpdata.blockHardLimit = 0;
	grpdata.blockTimeLimit = 0;
	usrdata.inodeSoftLimit = 0;
	usrdata.inodeHardLimit = 0;
	usrdata.inodeTimeLimit = 0;
	usrdata.blockSoftLimit = 0;
	usrdata.blockHardLimit = 0;
	usrdata.blockTimeLimit = 0;

	/*------------------ default fileset/aggregate quota ------------------*/
    /* Fill the request. */
    bzero((void *) &req, sizeof(fs_MsgGetQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_GET_QUOTA;
	req.fsid.whole_thing = inId;
	strcpy(req.path, bufpath);
	req.numberEntities = 1;
	req.entities[0].id = 0;
	req.entities[0].flag = 0;
	req.entities[0].flag |= FS_QUOTA_DEFAULT | FS_QUOTA_AGGREGATE;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat < 0) 
		return;
	else if (!stat)
		return;
    if (rsp.msgHdr.requestType != FS_REQ_GET_QUOTA)
		return;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return;

	defdata.inodeSoftLimit = rsp.responses[0].inodeSoftLimit;
	defdata.inodeHardLimit = rsp.responses[0].inodeHardLimit;
	defdata.inodeTimeLimit = rsp.responses[0].inodeTimeLimit;
	defdata.blockSoftLimit = rsp.responses[0].blockSoftLimit;
	defdata.blockHardLimit = rsp.responses[0].blockHardLimit;
	defdata.blockTimeLimit = rsp.responses[0].blockTimeLimit;

	/*------------------ default Group quota ------------------------------*/
    /* Fill the request. */
    bzero((void *) &req, sizeof(fs_MsgGetQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_GET_QUOTA;
	req.fsid.whole_thing = inId;
	strcpy(req.path, bufpath);
	req.numberEntities = 1;
	req.entities[0].id = 0;
	req.entities[0].flag = 0;
	req.entities[0].flag |= FS_QUOTA_DEFAULT | FS_QUOTA_GROUP;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat < 0) 
		return;
	else if (!stat)
		return;
    if (rsp.msgHdr.requestType != FS_REQ_GET_QUOTA)
		return;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return;

	grpdata.inodeSoftLimit = rsp.responses[0].inodeSoftLimit;
	grpdata.inodeHardLimit = rsp.responses[0].inodeHardLimit;
	grpdata.inodeTimeLimit = rsp.responses[0].inodeTimeLimit;
	grpdata.blockSoftLimit = rsp.responses[0].blockSoftLimit;
	grpdata.blockHardLimit = rsp.responses[0].blockHardLimit;
	grpdata.blockTimeLimit = rsp.responses[0].blockTimeLimit;

	/*------------------ default User quota -------------------------------*/
    /* Fill the request. */
    bzero((void *) &req, sizeof(fs_MsgGetQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_GET_QUOTA;
	req.fsid.whole_thing = inId;
	strcpy(req.path, bufpath);
	req.numberEntities = 1;
	req.entities[0].id = 0;
	req.entities[0].flag = 0;
	req.entities[0].flag |= FS_QUOTA_DEFAULT | FS_QUOTA_USER;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat < 0) 
		return;
	else if (!stat)
		return;
    if (rsp.msgHdr.requestType != FS_REQ_GET_QUOTA)
		return;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return;

	usrdata.inodeSoftLimit = rsp.responses[0].inodeSoftLimit;
	usrdata.inodeHardLimit = rsp.responses[0].inodeHardLimit;
	usrdata.inodeTimeLimit = rsp.responses[0].inodeTimeLimit;
	usrdata.blockSoftLimit = rsp.responses[0].blockSoftLimit;
	usrdata.blockHardLimit = rsp.responses[0].blockHardLimit;
	usrdata.blockTimeLimit = rsp.responses[0].blockTimeLimit;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case VOL_ID:
		sprintf(value, "%lld", inId);
		break;
	case VOL_NAME:
		strcpy(value, bufvol);
		break;
	case PATH_NAME:
		strcpy(value, bufpath);
		break;
	case PATH_FISOFT:
		sprintf(value, "%lld", defdata.inodeSoftLimit);
		break;
	case PATH_FIHARD:
		sprintf(value, "%lld", defdata.inodeHardLimit);
		break;
	case PATH_FITIME:
		sprintf(value, "%d", defdata.inodeTimeLimit);
		break;
	case PATH_FBSOFT:
		sprintf(value, "%lld", defdata.blockSoftLimit);
		break;
	case PATH_FBHARD:
		sprintf(value, "%lld", defdata.blockHardLimit);
		break;
	case PATH_FBTIME:
		sprintf(value, "%d", defdata.blockTimeLimit);
		break;
	case PATH_GISOFT:
		sprintf(value, "%lld", grpdata.inodeSoftLimit);
		break;
	case PATH_GIHARD:
		sprintf(value, "%lld", grpdata.inodeHardLimit);
		break;
	case PATH_GITIME:
		sprintf(value, "%d", grpdata.inodeTimeLimit);
		break;
	case PATH_GBSOFT:
		sprintf(value, "%lld", grpdata.blockSoftLimit);
		break;
	case PATH_GBHARD:
		sprintf(value, "%lld", grpdata.blockHardLimit);
		break;
	case PATH_GBTIME:
		sprintf(value, "%d", grpdata.blockTimeLimit);
		break;
	case PATH_UISOFT:
		sprintf(value, "%lld", usrdata.inodeSoftLimit);
		break;
	case PATH_UIHARD:
		sprintf(value, "%lld", usrdata.inodeHardLimit);
		break;
	case PATH_UITIME:
		sprintf(value, "%d", usrdata.inodeTimeLimit);
		break;
	case PATH_UBSOFT:
		sprintf(value, "%lld", usrdata.blockSoftLimit);
		break;
	case PATH_UBHARD:
		sprintf(value, "%lld", usrdata.blockHardLimit);
		break;
	case PATH_UBTIME:
		sprintf(value, "%d", usrdata.blockTimeLimit);
		break;
	default:
		return -1;
	}

	return 1;
}

/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionFsetDQuo(void *r, int count, char **argname, char **argval)
{
	/* initialize fileset information variables */
	getInputParams(count, argname, argval);
	getFileSetInfo();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
