/*  File Name  :    agilefsetfusg.c
 *
 *  Description:    HTTPD action functions for display fileset usage.
 *
 *  Created    :    3/20/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>
#include <evm_com_defs.h>
#include <evm-api.h>
#include <fs-api.h>

#include "agilepost.h"
#include "agileutil.h"


#define INVALID_TAG			0
#define VOL_ID				1
#define VOL_NAME			2
#define PATH_NAME			3
#define PATH_INODE			4
#define PATH_BLOCK			5
#define PATH_INODELMT		6
#define PATH_BLOCKLMT		7
#define NUM_OF_TAGS			7

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{VOL_ID,		"args.vol_id"},
	{VOL_NAME,		"args.vol_name"},
	{PATH_NAME,		"args.path"},
	{PATH_INODE,	"fileset_entry.inodesConsumed"},
	{PATH_BLOCK,	"fileset_entry.blocksConsumed"},
	{PATH_INODELMT,	"fileset_entry.inodeSoftLimitTime"},
	{PATH_BLOCKLMT,	"fileset_entry.blockSoftLimitTime"},
};


#define PARAM_VOLID			1
#define PARAM_VOLNAME		2
#define PARAM_PATH			3
#define NUM_OF_PARAMS		3

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_VOLID,	"id"},
	{PARAM_VOLNAME,	"vn"},
	{PARAM_PATH,	"path"},
};


/* HERE ARE VARIABLES FOR FILESET INFORMATION */
static char bufpath[FS_PATH_MAX+1];
static char bufvol[VOL_MAX_NAME+1];
static uint64 inId;
static fs_QuotaBlk_t data;


/***********************************************************
 * routines to retrieve fileset information 
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_VOLID:
			sscanf(argval[argind], "%lld", &inId);
			break;
		case PARAM_VOLNAME:
			strcpy(bufvol, argval[argind]);
			break;
		case PARAM_PATH:
			strcpy(bufpath, argval[argind]);
			break;
		}
	}

	return 0;
}

static void getFileSetInfo()
{
    int						stat;
	fs_MsgGetQuotaReq_t		req;
	fs_MsgGetQuotaResp_t	rsp;
	struct timeval			timeout = {60, 0};

	data.inodesConsumed = 0;
	data.blocksConsumed = 0;
	data.inodeSoftLimitTime = 0;
	data.blockSoftLimitTime = 0;

    /* Fill the request. */
    memset(&req, 0, sizeof(fs_MsgGetQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_GET_QUOTA;
	req.fsid.whole_thing = inId;
	strcpy(req.path, bufpath);
	req.numberEntities = 1;
	req.entities[0].id = 0;
	req.entities[0].flag = FS_QUOTA_AGGREGATE;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat <= 0) 
		return;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return;

	data.inodesConsumed = rsp.responses[0].inodesConsumed;
	data.blocksConsumed = rsp.responses[0].blocksConsumed;
	data.inodeSoftLimitTime = rsp.responses[0].inodeSoftLimitTime;
	data.blockSoftLimitTime = rsp.responses[0].blockSoftLimitTime;

}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case VOL_ID:
		sprintf(value, "%lld", inId);
		break;
	case VOL_NAME:
		strcpy(value, bufvol);
		break;
	case PATH_NAME:
		strcpy(value, bufpath);
		break;
	case PATH_INODE:
		sprintf(value, "%lld", data.inodesConsumed);
		break;
	case PATH_BLOCK:
		sprintf(value, "%lld", data.blocksConsumed);
		break;
	case PATH_INODELMT:
		sprintf(value, "%d", data.inodeSoftLimitTime);
		break;
	case PATH_BLOCKLMT:
		sprintf(value, "%d", data.blockSoftLimitTime);
		break;
	default:
		return -1;
	}

	return 1;
}

/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionFsetFUsg(void *r, int count, char **argname, char **argval)
{
	/* initialize fileset information variables */
	getInputParams(count, argname, argval);
	getFileSetInfo();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
