/*  File Name  :    agilefsetugquomod.c
 *
 *  Description:    HTTPD action functions for modifing fileset group/user quota.
 *
 *  Created    :    3/21/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>
#include <evm_com_defs.h>
#include <evm-api.h>
#include <fs-api.h>

#include "agilepost.h"
#include "agileutil.h"


#define INVALID_TAG			0
#define VOL_ID				1
#define VOL_NAME			2
#define PATH_NAME			3
#define PATH_TYPE			4
#define PATH_TYPESTR		5
#define PATH_UGID			6	
#define PATH_ISOFT			7
#define PATH_IHARD			8
#define PATH_ITIME			9
#define PATH_BSOFT			10
#define PATH_BHARD			11
#define PATH_BTIME			12
#define NUM_OF_TAGS			12

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{VOL_ID,		"args.vol_id"},
	{VOL_NAME,		"args.vol_name"},
	{PATH_NAME,		"args.path"},
	{PATH_TYPE,		"args.type"},
	{PATH_TYPESTR,	"args.typestr"},
	{PATH_UGID,		"args.ugid"},
	{PATH_ISOFT,	"fileset_entry.inodeSoftLimit"},
	{PATH_IHARD,	"fileset_entry.inodeHardLimit"},
	{PATH_ITIME,	"fileset_entry.inodeTimeLimit"},
	{PATH_BSOFT,	"fileset_entry.blockSoftLimit"},
	{PATH_BHARD,	"fileset_entry.blockHardLimit"},
	{PATH_BTIME,	"fileset_entry.blockTimeLimit"},
};


#define PARAM_VOLID			1
#define PARAM_VOLNAME		2
#define PARAM_PATH			3
#define PARAM_TYPE			4
#define PARAM_UGID			5
#define NUM_OF_PARAMS		5

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_VOLID,	"id"},
	{PARAM_VOLNAME,	"vn"},
	{PARAM_PATH,	"path"},
	{PARAM_TYPE,	"type"},
	{PARAM_UGID,	"ugid"},
};


/* HERE ARE VARIABLES FOR FILESET INFORMATION */
static char bufpath[FS_PATH_MAX+1];
static char bufvol[VOL_MAX_NAME+1];
static char inType[5];
static uint64 inId;
static fs_QuotaBlk_t data;


/***********************************************************
 * routines to retrieve fileset information 
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_VOLID:
			sscanf(argval[argind], "%lld", &inId);
			break;
		case PARAM_VOLNAME:
			strcpy(bufvol, argval[argind]);
			break;
		case PARAM_PATH:
			strcpy(bufpath, argval[argind]);
			break;
		case PARAM_TYPE:
			strcpy(inType, argval[argind]);
			break;
		case PARAM_UGID:
			sscanf(argval[argind], "%lld", &data.id);
			break;
		}
	}

	return 0;
}

static void getFileSetInfo()
{
    int						stat;
	fs_MsgGetQuotaReq_t		req;
	fs_MsgGetQuotaResp_t	rsp;
	struct timeval			timeout = {60, 0};

	data.inodeSoftLimit = 0;
	data.inodeHardLimit = 0;
	data.inodeTimeLimit = 0;
	data.blockSoftLimit = 0;
	data.blockHardLimit = 0;
	data.blockTimeLimit = 0;

    /* Fill the request. */
    bzero((void *) &req, sizeof(fs_MsgGetQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_GET_QUOTA;
	req.fsid.whole_thing = inId;
	strcpy(req.path, bufpath);
	req.numberEntities = 1;
	req.entities[0].id = data.id;
	req.entities[0].flag = 0;
	if (strcmp(inType, "u") == 0)
		req.entities[0].flag |= FS_QUOTA_USER;
	else
		req.entities[0].flag |= FS_QUOTA_GROUP;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat < 0) 
		return;
	else if (!stat)
		return;
    if (rsp.msgHdr.requestType != FS_REQ_GET_QUOTA)
		return;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return;

	data.inodeSoftLimit = rsp.responses[0].inodeSoftLimit;
	data.inodeHardLimit = rsp.responses[0].inodeHardLimit;
	data.inodeTimeLimit = rsp.responses[0].inodeTimeLimit;
	data.blockSoftLimit = rsp.responses[0].blockSoftLimit;
	data.blockHardLimit = rsp.responses[0].blockHardLimit;
	data.blockTimeLimit = rsp.responses[0].blockTimeLimit;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case VOL_ID:
		sprintf(value, "%lld", inId);
		break;
	case VOL_NAME:
		strcpy(value, bufvol);
		break;
	case PATH_NAME:
		strcpy(value, bufpath);
		break;
	case PATH_TYPE:
		strcpy(value, inType);
		break;
	case PATH_TYPESTR:
		if (strcmp(inType, "u") == 0)
			strcpy(value, "User");
		else
			strcpy(value, "Group");
		break;
	case PATH_UGID:
		sprintf(value, "%lld", data.id);
		break;
	case PATH_ISOFT:
		sprintf(value, "%lld", data.inodeSoftLimit);
		break;
	case PATH_IHARD:
		sprintf(value, "%lld", data.inodeHardLimit);
		break;
	case PATH_ITIME:
		sprintf(value, "%d", data.inodeTimeLimit);
		break;
	case PATH_BSOFT:
		sprintf(value, "%lld", data.blockSoftLimit);
		break;
	case PATH_BHARD:
		sprintf(value, "%lld", data.blockHardLimit);
		break;
	case PATH_BTIME:
		sprintf(value, "%d", data.blockTimeLimit);
		break;
	default:
		return -1;
	}

	if (value && !value[0])
		strcpy(value, "&nbsp");

	return 1;
}

/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionFsetUGQuoMod(void *r, int count, char **argname, char **argval)
{
	/* initialize fileset information variables */
	getInputParams(count, argname, argval);
	getFileSetInfo();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
