/*  File Name  :    agilefsetugusg.c 
 *
 *  Description:    HTTPD action functions for displaying FileSet group/user usage.
 *
 *  Created    :    3/20/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <anpssc-api.h>

#include <evm_com_defs.h>
#include <evm-api.h>
#include <fs-api.h>

#include "agilepost.h"
#include "agileutil.h"

#define INVALID_TAG			0
#define FS_VOLID			1
#define FS_VOLNAME			2
#define FS_PATH				3
#define FS_TYPE				4
#define FS_TYPESTR			5
#define FS_GUID				6
#define FS_INODE			7
#define FS_BLOCK			8
#define FS_INODELMT			9
#define FS_BLOCKLMT			10
#define NUM_OF_TAGS			10

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{FS_VOLID,		"args.vol_id"},
	{FS_VOLNAME,	"args.vol_name"},
	{FS_PATH,		"args.path"},
	{FS_TYPE,		"args.type"},
	{FS_TYPESTR,	"args.typestr"},
	{FS_GUID,		"fileset_entry.grpusr_id"},
	{FS_INODE,		"fileset_entry.inodesConsumed"},
	{FS_BLOCK,		"fileset_entry.blocksConsumed"},
	{FS_INODELMT,	"fileset_entry.inodeSoftLimitTime"},
	{FS_BLOCKLMT,	"fileset_entry.blockSoftLimitTime"},
};


#define PARAM_TYPE			1
#define PARAM_VOLID			2
#define PARAM_VOLNAME		3
#define PARAM_FSPATH		4
#define NUM_OF_PARAMS		4

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_TYPE,	"type"},
	{PARAM_VOLID,	"id"},
	{PARAM_VOLNAME,	"vn"},
	{PARAM_FSPATH,	"path"},
};


/* HERE ARE VARIABLES FOR FILESET INFORMATION */
#define MAX_NUM_OF_USG		10

static char inType[5];
static uint64 inId;
static char inName[VOL_MAX_NAME+1];
static char inPath[FS_PATH_MAX+1];

static int dataCount;
static int dataCurr;
static fs_QuotaBlk_t stData;
static fs_QuotaBlk_t *usgList;


/***********************************************************
 *  Main routines to handle Add/Delete/Qouta On/Off actions
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_TYPE:
			strcpy(inType, argval[argind]);
			break;
		case PARAM_VOLID:
			sscanf(argval[argind], "%lld", &inId);
			break;
		case PARAM_VOLNAME:
			strcpy(inName, argval[argind]);
			break;
		case PARAM_FSPATH:
			strcpy(inPath, argval[argind]);
			break;
		}
	}

	return 0;
}


/***********************************************************
 * routines to get fileset information thru anpssc
 **********************************************************/

static int getTableRowCount(void)
{
    int						stat, i;
    char					respbuff[4096];
	fs_MsgReportQuotaReq_t	req;
	fs_MsgReportQuotaResp_t	*rsp;
	struct timeval			timeout = {60, 0};
	int						sindex = 0;

    /* Fill the request. */
    memset(&req, 0, sizeof(fs_MsgReportQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_REPORT_QUOTA;
	req.fsid.whole_thing = inId;
	strcpy(req.path, inPath);
	if (strcmp(inType, "u") == 0)
		req.flag = FS_QUOTA_USER;
	else
		req.flag = FS_QUOTA_GROUP;
	req.numberEntries = MAX_NUM_OF_USG;

	while (1)
	{
		req.startIndex = sindex;

		/* Send message to FS. */
		stat = sendAgileMsg(-1, &req, sizeof(req),
							"efs", NFX_NFP_FP2, 1, MGMTPLANE,
							&timeout, respbuff, sizeof(respbuff));

		/* Check response from FS. */
		if (stat <= 0) 
			break;
		else if (!stat)
			break;

		rsp = (fs_MsgReportQuotaResp_t *)respbuff;

		if (rsp->msgHdr.fsStatus != FS_ERR_NONE)
			break;
		if (rsp->numberResponses <= 0)
			break;

		if (usgList == NULL)
		{
			usgList = malloc(rsp->numberResponses * sizeof(fs_QuotaBlk_t));
			if (usgList == NULL)
				break;
		}
		else
		{
			fs_QuotaBlk_t *p;

			p = realloc(usgList,
				(sindex+rsp->numberResponses) * sizeof(fs_QuotaBlk_t));
			if (p)
				usgList = p;
			else
				break;
		}

		for (i = 0; i < rsp->numberResponses; i++)
		{
			int j = sindex + i;
			usgList[j].id = rsp->responses[i].id;
			usgList[j].inodesConsumed = rsp->responses[i].inodesConsumed;
			usgList[j].blocksConsumed = rsp->responses[i].blocksConsumed;
			usgList[j].inodeSoftLimitTime = rsp->responses[i].inodeSoftLimitTime;
			usgList[j].blockSoftLimitTime = rsp->responses[i].blockSoftLimitTime;
		}

		sindex += rsp->numberResponses;
	}

	return sindex;
}

static int getTableRowData(int *index, fs_QuotaBlk_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (usgList == NULL)
		return -1;

	memcpy(data, &usgList[(*index)-1], sizeof(fs_QuotaBlk_t));
	return 0;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	/* first for those tags don't need to have fset data */
	switch (tagType) 
	{
	case FS_PATH:
		strcpy(value, inPath);
		return 1;
	case FS_TYPE:
		strcpy(value, inType);
		return 1;
	case FS_TYPESTR:
		if (strcmp(inType, "u") == 0)
			strcpy(value, "User");
		else
			strcpy(value, "Group");
		return 1;
	}

	/* for those tags required to have data */
	if ((dataCurr <= 0) || (dataCurr > dataCount))
		return -1;

	switch (tagType) 
	{
	case FS_VOLID:
		sprintf(value, "%lld", inId);
		break;
	case FS_VOLNAME:
		strcpy(value, inName);
		break;
	case FS_GUID:
		sprintf(value, "%lld", stData.id);
		break;
	case FS_INODE:
		sprintf(value, "%lld", stData.inodesConsumed);
		break;
	case FS_BLOCK:
		sprintf(value, "%lld", stData.blocksConsumed);
		break;
	case FS_INODELMT:
		sprintf(value, "%d", stData.inodeSoftLimitTime);
		break;
	case FS_BLOCKLMT:
		sprintf(value, "%d", stData.blockSoftLimitTime);
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else {
			return 1;
		}
	}
	else {
		/* other tags */
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionFsetUGUsg(void *r, int count, char **argname, char **argval)
{
	getInputParams(count, argname, argval);

	prepare_agile_template(r, NULL);

	/* initialize fileset information variables */
	dataCount = 0;
	dataCurr = 0;
	usgList = NULL;

	agile_scan_and_send(r, actRptTagFunc);

	if (usgList != NULL)
	{
		free(usgList);
		usgList = NULL;
	}

	return 1;
}
