/*  File Name  :    agileifdetail.c
 *
 *  Description:    HTTPD action functions for displaying interface detail/stats.
 *
 *  Created    :    3/3/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <sys/socket.h>
#include <net/if.h>
#include <ipmd-lib.h>

#include "agilepost.h"
#include "agileutil.h"
#include "../agent/httpdagent.h"

#define INVALID_TAG			0
#define NWIF_TYPE			1
#define NWIF_SLOT			2
#define NWIF_LINK			3
#define NWIF_NAME			4
#define NWIF_NTYPE			5
#define NWIF_SPEED			6
#define NWIF_QLEN			7
#define NWIF_PHYS			8
#define NWIF_ADDRL			9
#define NWIF_HEADL			10
#define NWIF_INQUO			11
#define NWIF_OUTQUO			12
#define NWIF_MTU			13
#define NWIF_BRATE			14
#define NWIF_INPKT			15
#define NWIF_INERR			16
#define NWIF_OUTPKT			17
#define NWIF_OUTERR			18
#define NWIF_COLL			19
#define NWIF_INOCT			20
#define NWIF_OUTOCT			21
#define NWIF_INMCST			22
#define NWIF_OUTMCST		23
#define NWIF_INDROP			24
#define NWIF_UNSUPP			25
#define NWIF_INTIME			26
#define NWIF_OUTTIME		27
#define NWIF_MODTIME		28
#define NWIF_MODMTIME		29
#define NWIF_IFNAME			30
#define NWIF_POS			31
#define NWIF_STATUS			32
#define NWIF_IFMTU			33
#define NWIF_MAC			34
#define NUM_OF_TAGS			34

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{NWIF_TYPE,		"args.if_type"},
	{NWIF_SLOT,		"args.if_slot"},
	{NWIF_LINK,		"args.if_link"},
	{NWIF_IFNAME,	"ipm_if_t.if_name"},
	{NWIF_POS,		"ipm_if_t.if_position"},
	{NWIF_STATUS,	"ipm_if_t.if_status"},
	{NWIF_IFMTU,	"ipm_if_t.if_mtusize"},
	{NWIF_MAC,		"ipm_if_t.if_mac"},
	{NWIF_NAME,		"ipm_ifstat_t.ifi_ifname"},
	{NWIF_NTYPE,	"ipm_ifstat_t.ifi_type"},
	{NWIF_SPEED,	"ipm_ifstat_t.ifi_speed"},
	{NWIF_QLEN,		"ipm_ifstat_t.ifi_oQlen"},
	{NWIF_PHYS,		"ipm_ifstat_t.ifi_physical"},
	{NWIF_ADDRL,	"ipm_ifstat_t.ifi_addrlen"},
	{NWIF_HEADL,	"ipm_ifstat_t.ifi_hdrlen"},
	{NWIF_INQUO,	"ipm_ifstat_t.ifi_recvquota"},
	{NWIF_OUTQUO,	"ipm_ifstat_t.ifi_xmitquota"},
	{NWIF_MTU,		"ipm_ifstat_t.ifi_mtu"},
	{NWIF_BRATE,	"ipm_ifstat_t.ifi_baudrate"},
	{NWIF_INPKT,	"ipm_ifstat_t.ifi_ipackets"},
	{NWIF_INERR,	"ipm_ifstat_t.ifi_ierrors"},
	{NWIF_OUTPKT,	"ipm_ifstat_t.ifi_opackets"},
	{NWIF_OUTERR,	"ipm_ifstat_t.ifi_oerrors"},
	{NWIF_COLL,		"ipm_ifstat_t.ifi_collisions"},
	{NWIF_INOCT,	"ipm_ifstat_t.ifi_ibytes"},
	{NWIF_OUTOCT,	"ipm_ifstat_t.ifi_obytes"},
	{NWIF_INMCST,	"ipm_ifstat_t.ifi_imcasts"},
	{NWIF_OUTMCST,	"ipm_ifstat_t.ifi_omcasts"},
	{NWIF_INDROP,	"ipm_ifstat_t.ifi_iqdrops"},
	{NWIF_UNSUPP,	"ipm_ifstat_t.ifi_noproto"},
	{NWIF_INTIME,	"ipm_ifstat_t.ifi_recvtiming"},
	{NWIF_OUTTIME,	"ipm_ifstat_t.ifi_xmittiming"},
	{NWIF_MODTIME,	"ipm_ifstat_t.ifi_lastchange_sec"},
	{NWIF_MODMTIME,	"ipm_ifstat_t.ifi_lastchange_usec"},
};


#define PARAM_TYPE			1
#define PARAM_SLOT			2
#define PARAM_LINK			3
#define PARAM_NAME          4
#define NUM_OF_PARAMS		4

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_TYPE,		"type"},
	{PARAM_SLOT,		"slot"},
	{PARAM_LINK,		"link"},
	{PARAM_NAME,		"name"}
};

static char *strUp   = "Up";
static char *strDown = "Down";


/* HERE ARE VARIABLES FOR NETWORK INFORMATION */
static char bufname[IPM_STR_SIZE];	/* interface name */
static char buftype[10];
static char bufslot[10];
static char buflink[10];
static ipm_if_t stIfData;
static ipm_ifstat_t stData;
static int getres;


static void GetStatData(void);
static void GetDetailData(void);


/***********************************************************
 * routines to retrieve network information 
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_NAME:
			strcpy(bufname, argval[argind]);
			break;
		case PARAM_TYPE:
			strcpy(buftype, argval[argind]);
			break;
		case PARAM_SLOT:
			strcpy(bufslot, argval[argind]);
			break;
		case PARAM_LINK:
			strcpy(buflink, argval[argind]);
			break;
		}
	}
#if 0
	GetInterfaceNameStr(bufname, buftype, bufslot, buflink);
#endif
	return 0;
}

static void GetStatData()
{
int n;
haMessage_t *msg;
haIfStats_t *phaifmsg;
char buf[1024];
struct timeval tout;
ipm_ifstat_t *pifstat;

	msg = (haMessage_t *) buf;
	msg->clientData = getpid();
	msg->type = HA_GETIFSTATS;
	msg->rspstatus = 0;
	msg->datalen = sizeof(haIfStats_t);
	strcpy(msg->data, bufname);

	tout.tv_sec = 5;
	tout.tv_usec = 0;

	n = sendAgileMsg(-1, buf, HAMESSAGE_HEADER_SIZE + sizeof(haIfStats_t),
					"httpdagent", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&tout, buf, sizeof(buf));
	if (n < 0) {
		getres = -1;
	}
	else if (!n) {
		getres = -1;
	}
	else {
		/* received response from httpdagent */
		msg = (haMessage_t *) buf;
		if (msg->rspstatus != HARSP_OK)
			return;
		phaifmsg = (haIfStats_t *) msg->data;
		pifstat = (ipm_ifstat_t *) phaifmsg->stats;
		memcpy(&stData, pifstat, sizeof(ipm_ifstat_t));
		getres = 1;
	}

#if 0
/* This involves privileged operation which "nobody" account */
/* can't go through. */

	int count, need;
	int32 res;

	memset(&stData, 0, sizeof(stData));
	getres = 1;
	res = ipmd_GetIfStats(bufname, &stData, sizeof(ipm_ifstat_t),
							&count, &need);
	if ((res == IPM_GET_ERR) ||
		(res == IPM_GET_NO_BUF) ||
		(res == IPM_CMDERR_APPNOBUF)) {
		getres = -1;
	}
	else
	{
	}
#endif
}

static void GetDetailData()
{
	// Get detail data
	int32	res;
	int count, need;

	getres = -1;
	GetInterfaceNameStr(bufname, buftype, bufslot, buflink);
	res = ipmd_GetIfByName(bufname, &stIfData, sizeof(ipm_if_t), 
							&count, &need);
	if ((res == IPM_GET_ERR) ||
		(res == IPM_GET_NO_BUF) ||
		(res == IPM_CMDERR_APPNOBUF))
		return;
	if (count < 0)
		return;

	// Get Stat data
	GetStatData();
	getres = 1;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	if (getres < 0)
		return -1;

	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	// Detail
	case NWIF_IFNAME:
		strcpy(value, stIfData.if_name);
		break;
	case NWIF_POS:
		sprintf(value, "%d/%d/%d", stIfData.if_slot, stIfData.if_cpu, stIfData.if_link);
		break;
	case NWIF_STATUS:
		if ((stIfData.if_flags & IFF_UP) != 0)
			strcpy(value, strUp);
		else
			strcpy(value, strDown);
		break;
	case NWIF_MAC:
		GetMacAddressStr(stIfData.if_mac, value);
		break;
	case NWIF_IFMTU:
		sprintf(value, "%d", stIfData.if_mtusize);
		break;
	//Stat
	case NWIF_TYPE:
		strcpy(value, buftype);
		break;
	case NWIF_SLOT:
		strcpy(value, bufslot);
		break;
	case NWIF_LINK:
		strcpy(value, buflink);
		break;
	case NWIF_NAME:
		strcpy(value, stData.ifi_ifname);
		break;
	case NWIF_NTYPE:
		sprintf(value, "%d", stData.ifi_type);	/* need a GetString fct from ipmd */
		break;
	case NWIF_SPEED:
		sprintf(value, "%d", stData.ifi_speed);
		break;
	case NWIF_QLEN:
		sprintf(value, "%d", stData.ifi_oQlen);
		break;
	case NWIF_PHYS:
		sprintf(value, "%d", stData.ifi_physical);	/* need a GetString fct from ipmd */
		break;
	case NWIF_ADDRL:
		sprintf(value, "%d", stData.ifi_addrlen);
		break;
	case NWIF_HEADL:
		sprintf(value, "%d", stData.ifi_hdrlen);
		break;
	case NWIF_INQUO:
		sprintf(value, "%d", stData.ifi_recvquota);
		break;
	case NWIF_OUTQUO:
		sprintf(value, "%d", stData.ifi_xmitquota);
		break;
	case NWIF_MTU:
		sprintf(value, "%d", stData.ifi_mtu);
		break;
	case NWIF_BRATE:
		sprintf(value, "%d", stData.ifi_baudrate);
		break;
	case NWIF_INPKT:
		sprintf(value, "%d", stData.ifi_ipackets);
		break;
	case NWIF_INERR:
		sprintf(value, "%d", stData.ifi_ierrors);
		break;
	case NWIF_OUTPKT:
		sprintf(value, "%d", stData.ifi_opackets);
		break;
	case NWIF_OUTERR:
		sprintf(value, "%d", stData.ifi_oerrors);
		break;
	case NWIF_COLL:
		sprintf(value, "%d", stData.ifi_collisions);
		break;
	case NWIF_INOCT:
		sprintf(value, "%d", stData.ifi_ibytes);
		break;
	case NWIF_OUTOCT:
		sprintf(value, "%d", stData.ifi_obytes);
		break;
	case NWIF_INMCST:
		sprintf(value, "%d", stData.ifi_imcasts);
		break;
	case NWIF_OUTMCST:
		sprintf(value, "%d", stData.ifi_omcasts);
		break;
	case NWIF_INDROP:
		sprintf(value, "%d", stData.ifi_iqdrops);
		break;
	case NWIF_UNSUPP:
		sprintf(value, "%d", stData.ifi_noproto);
		break;
	case NWIF_INTIME:
		sprintf(value, "%d", stData.ifi_recvtiming);
		break;
	case NWIF_OUTTIME:
		sprintf(value, "%d", stData.ifi_xmittiming);
		break;
	case NWIF_MODTIME:
		sprintf(value, "%d", stData.ifi_lastchange_sec);
		break;
	case NWIF_MODMTIME:
		sprintf(value, "%d", stData.ifi_lastchange_usec);
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionInterfaceDetail(void *r, int count, char **argname, char **argval)
{
	/* initialize network information variables */
	getInputParams(count, argname, argval);
	GetDetailData();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
