/*  File Name  :    agilensmod.c
 *
 *  Description:    HTTPD action functions for modify name service entry.
 *
 *  Created    :    3/29/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <nfxnis-api.h>

#include "agilepost.h"
#include "agileutil.h"


#define INVALID_TAG			0
#define ARG_IP				1
#define ARG_MASK			2
#define NS_DNS				3
#define NS_DNSDM			4
#define NS_DNSSVR1			5
#define NS_DNSSVR2			6
#define NS_DNSSVR3			7
#define NS_NIS				8
#define NS_NISDM			9
#define NS_NISSVR			10
#define NUM_OF_TAGS			10

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{ARG_IP,		"args.ip"},
	{ARG_MASK,		"args.mask"},
	{NS_DNS,		"nfxnis_cfg_t.dns"},
	{NS_DNSDM,		"nfxnis_cfg_t.dnsdm"},
	{NS_DNSSVR1,	"nfxnis_cfg_t.dnssvr1"},
	{NS_DNSSVR2,	"nfxnis_cfg_t.dnssvr2"},
	{NS_DNSSVR3,	"nfxnis_cfg_t.dnssvr3"},
	{NS_NIS,		"nfxnis_cfg_t.nis"},
	{NS_NISDM,		"nfxnis_cfg_t.nisdm"},
	{NS_NISSVR,		"nfxnis_cfg_t.nissvr"},
};


#define PARAM_IP			1
#define PARAM_MASK			2
#define NUM_OF_PARAMS		2

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_IP,		"ip"},
	{PARAM_MASK,	"mask"},
};


/* HERE ARE VARIABLES FOR NIS,DNS INFORMATION */

#define DOMAIN_NAME_LEN		64

typedef struct _export_entry_t
{
    uint32			ip;
    uint32			mask;
	char			dnsDomain[DOMAIN_NAME_LEN+1];
	int				numDnsServers;
	uint32			dnsServers[NFX_MAX_DNS_SERVERS_PER_DOMAIN];
	char			nisDomain[DOMAIN_NAME_LEN+1];
	int				numNisServers;
	uint32			nisServers[1];
} nfxnis_entry_t;

static nfxnis_entry_t stData;
static char bufip[20];
static char bufmask[20];



/***********************************************************
 * routines to retrieve volume information 
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;
	uint32 data;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_IP:
			sscanf(argval[argind], "%u", &data);
			GetIpAddressStr(data, bufip);
			break;
		case PARAM_MASK:
			sscanf(argval[argind], "%u", &data);
			GetIpAddressStr(data, bufmask);
			break;
		}
	}

	return 0;
}

static void AddEntry(uint32 ip, uint32 mask, nfxnis_cfg_t* entry)
{
	int i;

	if (entry->net != ip)
		return;
	stData.mask = entry->mask;

	// Update the entry
	if (entry->type == dns_name)
	{
		strcpy(stData.dnsDomain, entry->domain);
		stData.numDnsServers = entry->num_srvrs;

		for (i = 0; i < entry->num_srvrs; i ++)
			stData.dnsServers[i] = entry->srvrs[i];
		for (i = entry->num_srvrs; i < NFX_MAX_DNS_SERVERS_PER_DOMAIN; i ++)
			stData.dnsServers[i] = 0;
	}
	else if (entry->type == nis_name)
	{
		strcpy(stData.nisDomain, entry->domain);
		stData.numNisServers = entry->num_srvrs;
		stData.nisServers[0] = entry->srvrs[0];
	}
}

static void GetNameServiceInfo()
{
	uint32 ip, mask;
	nfxnis_cfg_t **list;
	nfxnis_cfg_t *entry;
	int i;

	memset(&stData, 0, sizeof(nfxnis_entry_t));
	ip = GetIpAddressFromStr(bufip);
	mask = GetIpAddressFromStr(bufmask);

	// Get dns,nis configurations
	list = nfxnis_get_cfg(ip, mask);
	if (list == NULL || *list == NULL)
		return;

	// Copy the configuration to local memory
	for (i = 0; list[i] != NULL; i ++)
	{
		entry = list[i];
		AddEntry(ip, mask, entry);

		// Free memory allocated by nfxnis_api
		free(entry->domain);
		free(entry);
	}
	free(list);
	return;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case ARG_IP:
		strcpy(value, bufip);
		break;
	case ARG_MASK:
		strcpy(value, bufmask);
		break;
	case NS_DNS:
		if (stData.numDnsServers > 0)
			strcpy(value, "checked");
		else
			strcpy(value, "");
		break;
	case NS_DNSDM:
		strcpy(value, stData.dnsDomain);
		break;
	case NS_DNSSVR1:
		if (stData.dnsServers[0] != 0)
			GetIpAddressStr(stData.dnsServers[0], value);
		else
			strcpy(value, "");
		break;
	case NS_DNSSVR2:
		if (stData.dnsServers[1] != 0)
			GetIpAddressStr(stData.dnsServers[1], value);
		else
			strcpy(value, "");
		break;
	case NS_DNSSVR3:
		if (stData.dnsServers[2] != 0)
			GetIpAddressStr(stData.dnsServers[2], value);
		else
			strcpy(value, "");
		break;
	case NS_NIS:
		if (stData.numNisServers > 0)
			strcpy(value, "checked");
		else
			strcpy(value, "");
		break;
	case NS_NISDM:
		strcpy(value, stData.nisDomain);
		break;
	case NS_NISSVR:
		if (stData.nisServers[0] != 0)
			GetIpAddressStr(stData.nisServers[0], value);
		else
			strcpy(value, "");
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionNsMod(void *r, int count, char **argname, char **argval)
{
	getInputParams(count, argname, argval);
	GetNameServiceInfo();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
