/*  File Name  :    agileportstat.c 
 *
 *  Description:    HTTPD action functions for displaying storage port stats.
 *
 *  Created    :    2/25/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>

#include <nfx-defs.h>
#include <nfx-types.h>
#include <eee-desc.h>
#include <scsi-api.h>
#include <hal.h>
#include <sdm-api.h>

#include "agilepost.h"
#include "agileutil.h"

#define INVALID_TAG			0
#define STGPT_LNKFAIL		1
#define STGPT_LOSSYNC		2
#define STGPT_LOSSIG		3
#define STGPT_INVXMIT		4
#define STGPT_INVCRC		5
#define STGPT_FRMOUT		6
#define STGPT_FRMIN			7
#define STGPT_XPRTERR		8
#define STGPT_UDRUN			9
#define STGPT_READ6			10
#define STGPT_READ10		11
#define STGPT_WRITE6		12
#define STGPT_WRITE10		13
#define STGPT_OTHER			14
#define STGPT_SLOT			15
#define STGPT_PORT			16
#define NUM_OF_TAGS			16

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{STGPT_LNKFAIL,	"hal_fc_link_stats_t.link_failure"},
	{STGPT_LOSSYNC,	"hal_fc_link_stats_t.loss_of_sync"},
	{STGPT_LOSSIG,	"hal_fc_link_stats_t.loss_of_signal"},
	{STGPT_INVXMIT,	"hal_fc_link_stats_t.invalid_xmit_words"},
	{STGPT_INVCRC,	"hal_fc_link_stats_t.invalid_crc"},
	{STGPT_FRMOUT,	"hal_fc_link_stats_t.frames_out"},
	{STGPT_FRMIN,	"hal_fc_link_stats_t.frames_in"},
	{STGPT_XPRTERR,	"hal_fc_link_stats_t.transport_errors"},
	{STGPT_UDRUN,	"hal_fc_link_stats_t.underrun_cnt"},
	{STGPT_READ6,	"hal_fc_link_stats_t.read_6_cmd_cnt"},
	{STGPT_READ10,	"hal_fc_link_stats_t.read_10_cmd_cnt"},
	{STGPT_WRITE6,	"hal_fc_link_stats_t.write_6_cmd_cnt"},
	{STGPT_WRITE10,	"hal_fc_link_stats_t.write_10_cmd_cnt"},
	{STGPT_OTHER,	"hal_fc_link_stats_t.other_cmd_cnt"},
	{STGPT_SLOT,	"args.slot"},
	{STGPT_PORT,	"args.port"},
};


#define PARAM_SLOT			1
#define PARAM_PORT			2
#define NUM_OF_PARAMS		2

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_SLOT,	"slot"},
	{PARAM_PORT,	"port"},
};


/* HERE ARE VARIABLES FOR STORAGE DEVICE DETAIL INFORMATION */
static int32 inSlot;
static int32 inPort;
static int res;
static hal_fc_link_stats_t stData;


/***********************************************************
 * routines to retrieve storage information thru anpssc
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_SLOT:
			inSlot = atoi(argval[argind]);
			break;
		case PARAM_PORT:
			inPort = atoi(argval[argind]);
			break;
		}
	}

	return 0;
}

static void GetPortStatsData()
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	struct timeval			timeout = {5, 0};

	res = -1;
	if (!IsStoragePortAlive(inSlot, inPort))
		return;

	memset(&msg, 0, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS;
	msg.slot = inSlot;
	msg.port = inPort;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", NFX_SSC_UNIX, SSC_SLOT, 0, &timeout,
					  respbuff, sizeof(respbuff));
	if (rc <= 0)
		return;

	rsp = (sdm_ioctlRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP ||
		rsp->eCode != SDM_OK)
		return;
	else if (rsp->dataLen < sizeof(hal_fc_link_stats_t)) 
		return;

	memcpy(&stData, rsp->data, sizeof(hal_fc_link_stats_t));
	res = 1;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	if (res <= 0) {
		strcpy(value, "&nbsp");
		return -1;
	}

	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i ++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case STGPT_LNKFAIL:
		sprintf(value, "%d", stData.link_failure);
		break;
	case STGPT_LOSSYNC:
		sprintf(value, "%d", stData.loss_of_sync);
		break;
	case STGPT_LOSSIG:
		sprintf(value, "%d", stData.loss_of_signal);
		break;
	case STGPT_INVXMIT:
		sprintf(value, "%d", stData.invalid_xmit_words);
		break;
	case STGPT_INVCRC:
		sprintf(value, "%d", stData.invalid_crc);
		break;
	case STGPT_FRMOUT:
		sprintf(value, "%lld", stData.frames_out);
		break;
	case STGPT_FRMIN:
		sprintf(value, "%lld", stData.frames_in);
		break;
	case STGPT_XPRTERR:
		sprintf(value, "%d", stData.transport_errors);
		break;
	case STGPT_UDRUN:
		sprintf(value, "%d", stData.underrun_cnt);
		break;
	case STGPT_READ6:
		sprintf(value, "%d", stData.read_6_cmd_cnt);
		break;
	case STGPT_READ10:
		sprintf(value, "%d", stData.read_10_cmd_cnt);
		break;
	case STGPT_WRITE6:
		sprintf(value, "%d", stData.write_6_cmd_cnt);
		break;
	case STGPT_WRITE10:
		sprintf(value, "%d", stData.write_10_cmd_cnt);
		break;
	case STGPT_OTHER:
		sprintf(value, "%d", stData.other_cmd_cnt);
		break;
	case STGPT_SLOT:
		sprintf(value, "%d", inSlot);
		break;
	case STGPT_PORT:
		sprintf(value, "%d", inPort);
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		return -1;
	}
	else {
		/* other tags */
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionPortStat(void *r, int count, char **argname, char **argval)
{
	/* initialize storage information variables */
	getInputParams(count, argname, argval);
	GetPortStatsData();

	prepare_agile_template(r, NULL);

	agile_scan_and_send(r, actRptTagFunc);

	return 1;
}
