/*  File Name  :    agileroutemain.c
 *
 *  Description:    HTTPD action functions for route main page.
 *
 *  Created    :    2/6/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <anpssc-api.h>

#undef MIN
#undef MAX
#include <sys/socket.h>
#include <net/route.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ipmd-lib.h>
#include "agilepost.h"
#include "agileutil.h"


#define INVALID_TAG			0
#define NWRT_DEST			1
#define NWRT_MASK			2
#define NWRT_NHOP			3
#define NWRT_STATUS			4
#define NWRT_IFNAME			5
#define NUM_OF_TAGS			5

#define NUM_RT_ENTRIES		20

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{NWRT_DEST,		"ipm_rt_t.rt_addr"},
	{NWRT_MASK,		"ipm_rt_t.rt_mask"},
	{NWRT_NHOP,		"ipm_rt_t.rt_nhop"},
	{NWRT_STATUS,	"ipm_rt_t.rt_status"},
	{NWRT_IFNAME,	"ipm_rt_t.rt_ifname"},
};


#define PARAM_ACT			1
#define PARAM_IP			2
#define PARAM_MASK			3
#define PARAM_NHOP			4
#define NUM_OF_PARAMS		4

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_ACT,		"act"},
	{PARAM_IP,		"ip"},
	{PARAM_MASK,	"mask"},
	{PARAM_NHOP,	"nhop"},
};

static char *strUp   = "Usable";
static char *strDown = "Down";

static char bufact[5];
static char bufip[20];
static char bufmask[20];
static char bufnhop[20];


/* HERE ARE VARIABLES FOR STORAGE DEVICE LIST INFORMATION */
static int dataCount;
static int dataCurr;
static ipm_rt_t *stRouteTable;
static ipm_rt_t stData;


/***********************************************************
 * routines to retrieve network information 
 **********************************************************/

static int getTableRowCount(void)
{
	int32	res;
	int count, need;

	/* Try to get Route Table with maxi. 20 entries */
	stRouteTable = (ipm_rt_t *) malloc(NUM_RT_ENTRIES*sizeof(ipm_rt_t));
	if (stRouteTable == NULL)
		return -1;
	res = ipmd_GetRtTbl(stRouteTable, NUM_RT_ENTRIES*sizeof(ipm_rt_t), 
							&count, &need);
	if (res == IPM_GET_ERR)
		return -1;
	else if ((res == IPM_GET_NO_BUF) || (res == IPM_CMDERR_APPNOBUF))
	{
		/* Need a larger buffer */
		free(stRouteTable);
		stRouteTable = (ipm_rt_t *) malloc(need);
		if (stRouteTable == NULL)
			return -1;

		res = ipmd_GetRtTbl(stRouteTable, need, &count, &need);
		if (res == IPM_GET_ERR)
			return -1;
		else if ((res == IPM_GET_NO_BUF) || (res == IPM_CMDERR_APPNOBUF))
			return -1;
	}

	if (count < 0)
		return -1;
	return count;
}

static int getTableRowData(int *index, ipm_rt_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (stRouteTable == NULL)
		return -1;

	memcpy(data, &stRouteTable[(*index)-1], sizeof(ipm_rt_t));
	return 0;
}


/***********************************************************
 *  Main routines to handle Add, Del actions
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_ACT:
			strcpy(bufact, argval[argind]);
			break;
		case PARAM_IP:
			strcpy(bufip, argval[argind]);
			break;
		case PARAM_MASK:
			strcpy(bufmask, argval[argind]);
			break;
		case PARAM_NHOP:
			strcpy(bufnhop, argval[argind]);
			break;
		}
	}

	return 0;
}

static int HandleRtAction(char *newfname, int count, char **argname, char **argval)
{
	int32 res;
	struct in_addr ipaddr, ipmask, ipnhop;

	getInputParams(count, argname, argval);

	if (!strcmp(bufact, strAdd))
	{
		inet_aton(bufip, &ipaddr);
		inet_aton(bufmask, &ipmask);
		inet_aton(bufnhop, &ipnhop);
		res = ipmd_AddRt((uint32) ipaddr.s_addr, (uint32) ipmask.s_addr, 
				(uint32) ipnhop.s_addr, NULL, 0);
		if (res == NFX_ERR)
			strcpy(newfname, "error_addrt.html");
	}
	else if (!strcmp(bufact, strDel))
	{
		inet_aton(bufip, &ipaddr);
		inet_aton(bufmask, &ipmask);
		inet_aton(bufnhop, &ipnhop);
		res = ipmd_DelRt((uint32) ipaddr.s_addr, (uint32) ipmask.s_addr, 
				(uint32) ipnhop.s_addr, NULL, 0);
		if (res == NFX_ERR)
			strcpy(newfname, "error_delrt.html");
	}
	return 1;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType;
	
	if ((dataCurr <= 0) || (dataCurr > dataCount))
		return -1;

	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	switch (tagType) 
	{
	case NWRT_DEST:
		GetIpAddressStr(stData.rt_addr, value);
		break;
	case NWRT_MASK:
		GetIpAddressStr(stData.rt_mask, value);
		break;
	case NWRT_NHOP:
		GetIpAddressStr(stData.rt_nhop, value);
		break;
	case NWRT_STATUS:
		if ((stData.rt_flags & RTF_UP) != 0)
			strcpy(value, strUp);
		else
			strcpy(value, strDown);
		break;
	case NWRT_IFNAME:
		strcpy(value, stData.rt_ifname);
		break;
	default:
		return -1;
	}

	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else {
			return 1;
		}
	}
	else {
		/* other tags */
		if (dataCount <= 0) {
			dataCount = getTableRowCount();
			if (dataCount <= 0)
				return -1;
		}

		if ((dataCurr < 1) || (dataCurr > dataCount)) {
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) < 0) {
				return -1;
			}
		}

		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionRouteMain(void *r, int count, char **argname, char **argval)
{
char newfname[128];

	/* Handle Add, Del actions */
	memset(newfname, 0, sizeof(newfname));

	/* Handle Add, Del actions */
	if (count > 0)
		HandleRtAction(newfname, count, argname, argval);

	prepare_agile_template(r, newfname);

	/* initialize storage information variables */
	dataCount = 0;
	dataCurr = 0;
	stRouteTable = NULL;

	agile_scan_and_send(r, actRptTagFunc);

	if (stRouteTable != NULL)
		free(stRouteTable);
	return 1;
}
