/*
 *  Interfaces MIB group implementation - interfaces.c
 *
 *  Currently, we only handle E10. That is, the interface index
 *  is fixed as following:
 *  1, 2 - Ethernet ports on SSC
 *  3, 4, 5, 6 - GbE ports on FP-NIM
 *  No support for FC-NIM in MIB-2
 *
 *  And for E10, the SSC is on slot 0. FP-NIM and FC-NIM are then
 *  can be on slot 1 and 2.
 *
 */

#include <config.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <netinet/in_var.h>
#include <net/if_types.h>
#include <net/if_dl.h>

#include "mibincl.h"
#include "kernel.h"

#include "system.h"
#include "snmp_logging.h"
#include "read_config.h"

#include "interfaces.h"
#include "struct.h"
#include "util_funcs.h"
#include "auto_nlist.h"
#include "sysORTable.h"

#include <ipmd-lib.h>
#include <snmpmsgapi.h>
#include "ipifhelp.h"

#define IFENTRY_NAME_LENGTH	10

#define AGILE_CARD_SSC		1
#define AGILE_CARD_GBENIM	2
#define AGILE_CARD_FCNIM	3


extern struct timeval starttime;

char *getGbENIMStats(struct variable *vp, size_t *var_len,
			WriteMethod **write_method, int interface);
static int Interface_Get_Count (void);

static void parse_interface_config(const char *, char *);
static void free_interface_config(void);

struct variable4 interfaces_variables[] = {
	{IFNUMBER, ASN_INTEGER, RONLY, var_interfaces, 1, {1}},
	{IFINDEX, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 1}},
	{IFDESCR, ASN_OCTET_STR, RONLY, var_ifEntry, 3, {2, 1, 2}},
	{IFTYPE, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 3}},
	{IFMTU, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 4}},
	{IFSPEED, ASN_GAUGE, RONLY, var_ifEntry, 3, {2, 1, 5}},
	{IFPHYSADDRESS, ASN_OCTET_STR, RONLY, var_ifEntry, 3, {2, 1, 6}},
	{IFADMINSTATUS, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 7}},
	{IFOPERSTATUS, ASN_INTEGER, RONLY, var_ifEntry, 3, {2, 1, 8}},
	{IFLASTCHANGE, ASN_TIMETICKS, RONLY, var_ifEntry, 3, {2, 1, 9}},
	{IFINOCTETS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 10}},
	{IFINUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 11}},
	{IFINNUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 12}},
	{IFINDISCARDS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 13}},
	{IFINERRORS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 14}},
	{IFINUNKNOWNPROTOS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 15}},
	{IFOUTOCTETS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 16}},
	{IFOUTUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 17}},
	{IFOUTNUCASTPKTS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 18}},
	{IFOUTDISCARDS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 19}},
	{IFOUTERRORS, ASN_COUNTER, RONLY, var_ifEntry, 3, {2, 1, 20}},
	{IFOUTQLEN, ASN_GAUGE, RONLY, var_ifEntry, 3, {2, 1, 21}},
	{IFSPECIFIC, ASN_OBJECT_ID, RONLY, var_ifEntry, 3, {2, 1, 22}}
};


// static struct timeval msg_timeout = {3, 0};  // We'll wait 3 sec

/* Define the OID pointer to the top of the mib tree that we're
   registering underneath, and the OID of the MIB module */
oid interfaces_variables_oid[] = { SNMP_OID_MIB2,2 };
oid interfaces_module_oid[]	= { SNMP_OID_MIB2,31 };

void init_interfaces(void)
{
  /* register ourselves with the agent to handle our mib tree */
  REGISTER_MIB("mibII/interfaces", interfaces_variables, variable4, \
			   interfaces_variables_oid);
  REGISTER_SYSOR_ENTRY(interfaces_module_oid,
	"The MIB module to describe generic objects for network interface sub-layers");
  
  snmpd_register_config_handler("interface", parse_interface_config,
					free_interface_config, "name type speed");
}

#if 0
/*
 * if_type_from_name
 * Return interface type using the interface name as a clue.
 * Returns 1 to imply "other" type if name not recognized. 
 */
static int
if_type_from_name( const char *pcch)
{
	typedef struct _match_if {
		int mi_type;
		const char *mi_name;
	} *pmatch_if, match_if;
	
	static match_if lmatch_if[] = {
		{ 117, "e" },
		{  56, "f" },
		{  0, 0 }  /* end of list */
	};

	int ii, len;
	register pmatch_if pm;

	for (ii = 0, pm=lmatch_if; pm->mi_name; pm++) {
		len = strlen(pm->mi_name);
		if (0 == strncmp(pcch, pm->mi_name, len))
		{
			return (pm->mi_type);
		}
	}
	return (1); /* in case search fails */
}
#endif

typedef struct _conf_if_list {
	char *name;
	int type;
	u_long speed;
	struct _conf_if_list *next;
} conf_if_list;

static conf_if_list *conf_list;
// REMOVED static struct ifnet *ifnetaddr_list;

static void parse_interface_config(const char *token, char *cptr)
{
	conf_if_list *if_ptr, *if_new;
	char *name, *type, *speed, *ecp;

	name = strtok(cptr, " \t");
	if (!name) {
		config_perror("Missing NAME parameter");
		return;
	}
	type = strtok(NULL, " \t");
	if (!type) {
		config_perror("Missing TYPE parameter");
		return;
	}
	speed = strtok(NULL, " \t");
	if (!speed) {
		config_perror("Missing SPEED parameter");
		return;
	}
	if_ptr = conf_list;
	while (if_ptr)
		if (strcmp(if_ptr->name, name)) if_ptr = if_ptr->next;
		else break;
	if (if_ptr)
		config_pwarn("Duplicate interface specification");
	if_new = (conf_if_list *)malloc(sizeof(conf_if_list));
	if (!if_new) {
		config_perror("Out of memory");
		return;
	}
	if_new->speed = strtoul(speed, &ecp, 0);
	if (*ecp) {
		config_perror("Bad SPEED value");
		free(if_new);
		return;
	}
	if_new->type = strtol(type, &ecp, 0);
	if (*ecp || if_new->type < 0) {
		config_perror("Bad TYPE");
		free(if_new);
		return;
	}
	if_new->name = strdup(name);
	if (!if_new->name) {
		config_perror("Out of memory");
		free(if_new);
		return;
	}
	if_new->next = conf_list;
	conf_list = if_new;
}

static void free_interface_config(void)
{
	conf_if_list *if_ptr = conf_list, *if_next;
	while (if_ptr) {
	if_next = if_ptr->next;
	free(if_ptr->name);
	free(if_ptr);
	if_ptr = if_next;
	}
	conf_list = NULL;
}


/*
  header_ifEntry(...
  Arguments:
  vp	  IN	  - pointer to variable entry that points here
  name	IN/OUT  - IN/name requested, OUT/name found
  length  IN/OUT  - length of IN/OUT oid's 
  exact   IN	  - TRUE if an exact match was requested
  var_len OUT	 - length of variable or 0 if function returned
  write_method
  
*/
static int
header_ifEntry(struct variable *vp,
		   oid *name,
		   size_t *length,
		   int exact,
		   size_t *var_len,
		   WriteMethod **write_method)
{
	oid newname[MAX_OID_LEN];
	register int	interface;
	int result, count;

	DEBUGMSGTL(("mibII/interfaces", "var_ifEntry: "));
	DEBUGMSGOID(("mibII/interfaces", name, *length));
	DEBUGMSG(("mibII/interfaces"," %d\n", exact));

	memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
	/* find "next" interface */
	count = Interface_Get_Count();

	for(interface = 1; interface <= count; interface++){
		newname[IFENTRY_NAME_LENGTH] = (oid)interface;
		result = snmp_oid_compare(name, *length, newname, (int)vp->namelen + 1);
		if ((exact && (result == 0)) || (!exact && (result < 0)))
			break;
	}
	if (interface > count) {
		DEBUGMSGTL(("mibII/interfaces", "... index out of range\n"));
		return MATCH_FAILED;
	}

	memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));
	*length = vp->namelen + 1;
	*write_method = 0;
	*var_len = sizeof(long);	/* default to 'long' results */

	DEBUGMSGTL(("mibII/interfaces", "... get I/F stats "));
	DEBUGMSGOID(("mibII/interfaces", name, *length));
	DEBUGMSG(("mibII/interfaces","\n"));

	return interface;
}



u_char *
var_interfaces(struct variable *vp,
		   oid *name,
		   size_t *length,
		   int exact,
		   size_t *var_len,
		   WriteMethod **write_method)
{
  if (header_generic(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
	return NULL;

  switch (vp->magic)
	{
	case IFNUMBER:
	  long_return = Interface_Get_Count ();
	  return (u_char *)&long_return;
	default:
	  DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_interfaces\n", vp->magic));
	}
  return NULL;
}

	/*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/


	/*********************
	 *
	 *  System specific implementation functions
	 *
	 *********************/

#if 0
char *getSscMgmtPortStats(struct variable *vp, size_t *var_len,
			WriteMethod **write_method, int interface)
{
char *cp;
static char Name[32];
static struct ifnet ifnet;
static struct in_ifaddr in_ifaddr;
int i, count;
  
	Interface_Scan_Init();
	count = 0;
	while (Interface_Scan_Next(&i, Name, &ifnet, &in_ifaddr)) {
		if (ifnet.if_type == IFT_ETHER) {
			count++;
			if (count == interface)
				break;
		}
	}
	if (count <= 0)
		return NULL;

	switch (vp->magic){
	case IFINDEX:
		long_return = interface;
		return (u_char *) &long_return;
	case IFDESCR:
		cp = Name;
		*var_len = strlen(cp);
		return (u_char *)cp;
	case IFTYPE:
		long_return = ifnet.if_type;
		return (u_char *) &long_return;
	case IFMTU: {
		long_return = (long) ifnet.if_mtu;
		return (u_char *) &long_return;
	}
	case IFSPEED:
		long_return = ifnet.if_baudrate;
		return (u_char *) &long_return;
	case IFPHYSADDRESS:
		Interface_Get_Ether_By_Index(interface, return_buf);
		*var_len = 6;
		if ((return_buf[0] == 0) && (return_buf[1] == 0) &&
		(return_buf[2] == 0) && (return_buf[3] == 0) &&
		(return_buf[4] == 0) && (return_buf[5] == 0))
		*var_len = 0;
		return(u_char *) return_buf;
	case IFADMINSTATUS:
		long_return = ifnet.if_flags & IFF_UP ? 1 : 2;
		return (u_char *) &long_return;
	case IFOPERSTATUS:
		long_return = ifnet.if_flags & IFF_RUNNING ? 1 : 2;
		return (u_char *) &long_return;
	case IFLASTCHANGE:
/* XXX - SNMP's ifLastchange is time when op. status changed
 * FreeBSD's if_lastchange is time when packet was input or output
 * (at least in 2.1.0-RELEASE. Changed in later versions of the kernel?)
 */
/* FreeBSD's if_lastchange before the 2.1.5 release is the time when
 * a packet was last input or output.  In the 2.1.5 and later releases,
 * this is fixed, thus the 199607 comparison.
 */
		  if (ifnet.if_lastchange.tv_sec == 0 &&
			  ifnet.if_lastchange.tv_usec == 0)
			long_return = 0;
	  else if (ifnet.if_lastchange.tv_sec < starttime.tv_sec)
		long_return = 0;
		  else {
			long_return = (u_long)
			  ((ifnet.if_lastchange.tv_sec - starttime.tv_sec) * 100
			   + (ifnet.if_lastchange.tv_usec - starttime.tv_usec) / 10000);
		  }
		  return (u_char *) &long_return;
	case IFINOCTETS:
		  long_return = (u_long)  ifnet.if_ibytes;
		return (u_char *) &long_return;
	case IFINUCASTPKTS:
		{
		long_return = (u_long)  ifnet.if_ipackets;
		long_return -= (u_long) ifnet.if_imcasts;
		}
		return (u_char *) &long_return;
	case IFINNUCASTPKTS:
		long_return = (u_long)  ifnet.if_imcasts;
		return (u_char *) &long_return;
	case IFINDISCARDS:
		long_return = (u_long)  ifnet.if_iqdrops;
		return (u_char *) &long_return;
	case IFINERRORS:
		long_return = (u_long) ifnet.if_ierrors;
		return (u_char *) &long_return;
	case IFINUNKNOWNPROTOS:
		long_return = (u_long)  ifnet.if_noproto;
		return (u_char *) &long_return;
	case IFOUTOCTETS:
		  long_return = (u_long)  ifnet.if_obytes;
		return (u_char *) &long_return;
	case IFOUTUCASTPKTS:
		{
		long_return = (u_long)  ifnet.if_opackets;
		long_return -= (u_long) ifnet.if_omcasts;
		}
		return (u_char *) &long_return;
	case IFOUTNUCASTPKTS:
		long_return = (u_long)  ifnet.if_omcasts;
		return (u_char *) &long_return;
	case IFOUTDISCARDS:
		  long_return = ifnet.if_snd.ifq_drops;
		  return (u_char *) &long_return;
	case IFOUTERRORS:
		  long_return = ifnet.if_oerrors;
		  return (u_char *) &long_return;
	case IFOUTQLEN:
		  long_return = ifnet.if_snd.ifq_len;
		  return (u_char *) &long_return;
	case IFSPECIFIC:
		*var_len = nullOidLen;
		return (u_char *) nullOid;
	default:
		DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_ifEntry\n", vp->magic));
	}
	return NULL;
}
#endif

char *getGbENIMStats(struct variable *vp, size_t *var_len,
			WriteMethod **write_method, int interface)
{
int rval, port, size, count, need;
ipm_if_t *ipmifdata;
ipm_ifstat_t ifstats;
static char Name[32];
char *cp;
int i;

	/* This is a GbE interface			   */
	/* use IPMD API to interact with GbE NIM */

	/* port number is 0 to 3 (mapped to ifIndex 3 - 6) */

	count = 34;
	size = count * sizeof(ipm_if_t);
	ipmifdata = malloc(size);
	if (ipmifdata == NULL)
		return NULL;

	rval = ipmd_GetIfByName(NULL, ipmifdata, size, &count, &need);
	if (rval == IPM_GET_ERR) {
		free(ipmifdata);
		return NULL;
	}
	if ((rval == IPM_GET_NO_BUF) || (rval == IPM_CMDERR_APPNOBUF)) {
		free(ipmifdata);
		count = need / sizeof(ipm_if_t);
		size = need;
		ipmifdata = malloc(size);
		if (ipmifdata == NULL)
			return NULL;
		rval = ipmd_GetIfByName(NULL, ipmifdata, size, &count, &need);
		if (rval != IPM_GET_OK) {
			free(ipmifdata);
			return NULL;
		}
	}

	if (count <= 0) {
		free(ipmifdata);
		return NULL;
	}

	Name[0] = '\0';
	for (i = 0; i < count; i++) {
		if ((interface > 2) && !strncmp(ipmifdata[i].if_name, "fp", 2)) {
			char *p = &ipmifdata[i].if_name[2];
			while (*p && (*p != '.')) p++;
			if (*p) p++;
			port = *p - '0';
#if 0
/* in case number of ports gets more than 9 */
			port = 0;
			while (*p && isdigit((int)*p)) {
				port = (*p - '0') + port * 10;
				p++;
			}
#endif
			if (port == interface - 3) {
				strcpy(Name, ipmifdata[i].if_name);
				break;
			}
		}
		else if (((interface == 1) || (interface == 2)) &&
				 	!strncmp(ipmifdata[i].if_name, "sc", 2)) {
			port = ipmifdata[i].if_name[2] - '0';
			if (port == interface) {
				strcpy(Name, ipmifdata[i].if_name);
				break;
			}
		}
	}

	if (Name[0] == '\0') {
		free(ipmifdata);
		return NULL;
	}

	count = 1;
	rval = ipmd_GetIfStats(Name, &ifstats, sizeof(ipm_ifstat_t), &count, &need);
	if (rval != IPM_GET_OK) {
		return NULL;
	}

	switch (vp->magic){
	case IFINDEX:
		long_return = interface;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFDESCR:
		// strcpy(Name, ifstats.ifi_ifname);
		cp = Name;
		*var_len = strlen(cp);
		free(ipmifdata);
		return (u_char *)cp;
	case IFTYPE:
		long_return = ifstats.ifi_type;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFMTU:
		long_return = ifstats.ifi_mtu;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFSPEED:
		long_return = ifstats.ifi_speed;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFPHYSADDRESS:
#if 0
		*var_len = ifstats.ifi_addrlen;
#endif
		*var_len = 6;
		memcpy(return_buf, ipmifdata[i].if_mac, *var_len);
		free(ipmifdata);
		return (u_char *) return_buf;
	case IFADMINSTATUS:
#if 0
/* disable read-write for now */
		*write_method = write_ifAdminStatus;
#endif
		long_return = ipmifdata[i].if_flags & IFF_UP ? 1 : 2;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOPERSTATUS:
		long_return = ipmifdata[i].if_flags & IFF_RUNNING ? 1 : 2;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFLASTCHANGE:
		// long_return = ifstats.ifi_lastchange_sec;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFINOCTETS:
		long_return = ifstats.ifi_ibytes;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFINUCASTPKTS:
		// ipmd doesn't provide ifi_imcasts
		// long_return = ifstats.ifi_ipackets - ifstats.ifi_imcasts;
		long_return = ifstats.ifi_ipackets;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFINNUCASTPKTS:
		// ipmd doesn't provide ifi_imcasts
		// long_return = ifstats.ifi_imcasts;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFINDISCARDS:
		long_return = ifstats.ifi_iqdrops;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFINERRORS:
		// not supported by ipmd
		// long_return = ifstats.ifi_ierrors;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFINUNKNOWNPROTOS:
		// not supported by ipmd
		// long_return = ifstats.ifi_noproto;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOUTOCTETS:
		long_return = ifstats.ifi_obytes;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOUTUCASTPKTS:
		// ifstats.ifi_omcasts not supported by ipmd
		// long_return = ifstats.ifi_opackets - ifstats.ifi_omcasts;
		long_return = ifstats.ifi_opackets;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOUTNUCASTPKTS:
		// ifstats.ifi_omcasts not supported by ipmd
		// long_return = ifstats.ifi_omcasts;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOUTDISCARDS:
		// not supported by ipmd
		// long_return = ifstats.ifi_odiscards;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOUTERRORS:
		long_return = ifstats.ifi_oerrors;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFOUTQLEN:
		// not supported by ipmd
		// long_return = ifstats.ifi_oQlen;
		long_return = 0;
		free(ipmifdata);
		return (u_char *) &long_return;
	case IFSPECIFIC:
		*var_len = nullOidLen;
		free(ipmifdata);
		return (u_char *) nullOid;
	default:
		DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_ifEntry\n", vp->magic));
		free(ipmifdata);
		return NULL;
	}

	return NULL;
}

u_char *
var_ifEntry(struct variable *vp,
		oid *name,
		size_t *length,
		int exact,
		size_t *var_len,
		WriteMethod **write_method)
{
int interface;

	interface = header_ifEntry(vp, name, length, exact, var_len, write_method);
	if (interface == MATCH_FAILED)
		return NULL;

	return getGbENIMStats(vp, var_len, write_method, interface);
#if 0
	if ((interface == 1) || (interface == 2))
		return getSscMgmtPortStats(vp, var_len, write_method, interface);
	else if ((interface > 2) && (interface <= 6))
		return getGbENIMStats(vp, var_len, write_method, interface);
#endif
	return NULL;
}


int
write_ifAdminStatus(
	int		action,
	u_char	*var_val,
	u_char	var_val_type,
	size_t	var_val_len,
	u_char	*statP,
	oid		*name,
	size_t	name_len)
{
// int size;
uint32_t adminaction;

	switch ( action ) {
	case RESERVE1:
		if (var_val_type != ASN_INTEGER){
			fprintf(stderr, "write to ifAdminStatus not ASN_INTEGER\n");
			return SNMP_ERR_WRONGTYPE;
		}

		if (var_val == NULL)
			return SNMP_ERR_WRONGVALUE;

		/* by MIB-II definition, adminaction can have value	*/
		/* 1: up,  2: down,  or 3: testing.  Value 3 (testing) */
		/* is not supported here......						 */
		adminaction = *((uint32_t *) var_val);
		if ((adminaction > 2) || (adminaction < 1))
			return SNMP_ERR_WRONGVALUE;

		break;

	case RESERVE2:
		break;


	case FREE:
		/* Release any resources that have been allocated */
		break;


	case ACTION:
		break;


	case UNDO:
		/* Back out any changes made in the ACTION case */
		break;


	case COMMIT:
		{
			int interface;
			// int slot;
			// int agile_cardtype;
			// char Name[32];

			interface = name[IFENTRY_NAME_LENGTH];
			if ((interface == 1) || (interface == 2)) {
				/* SSC slot, don't know how to handle yet */
				/* NEED TO ADD SUPPORT!!!!!!!!!!!!! */
				return NULL;
			}
			else if ((interface >=3) || (interface <= 6)) {
				return NULL;
#if 0
				slot = GetSlotNumber(interface);
				sprintf(Name, "fp%d.%d", slot, (interface-2) % 4);
				if (adminaction == 1)
					ipmd_EnableIntf(Name, NULL, -1);
				else if (adminaction == 2)
					ipmd_DisableIntf(Name, NULL, -1);
#endif
			}
		}
		break;
	}
	return SNMP_ERR_NOERROR;
}


// static int Interface_Count=0;

static int Interface_Get_Count (void)
{
int count, size, need, rval, ifcount, i;
ipm_if_t *ipmifdata;

	count = 34;
	size = count * sizeof(ipm_if_t);
	ipmifdata = malloc(size);
	if (ipmifdata == NULL)
		return 0;

	rval = ipmd_GetIfByName(NULL, ipmifdata, size, &count, &need);
	if (rval == IPM_GET_ERR) {
		free(ipmifdata);
		return 0;
	}

	if ((rval == IPM_GET_NO_BUF) || (rval == IPM_CMDERR_APPNOBUF)) {
		free(ipmifdata);
		count = need / sizeof(ipm_if_t);
		size = need;
		ipmifdata = malloc(size);
		if (ipmifdata == NULL)
			return 0;
		rval = ipmd_GetIfByName(NULL, ipmifdata, size, &count, &need);
		if (rval != IPM_GET_OK) {
			free(ipmifdata);
			return 0;
		}
	}

	ifcount = 0;
	for (i = 0; i < count; i++) {
		if (!strncmp(ipmifdata[i].if_name, "fp", 2))
			ifcount++;
		else if (!strncmp(ipmifdata[i].if_name, "sc", 2)) {
			if ((ipmifdata[i].if_name[2] == '1') ||
				(ipmifdata[i].if_name[2] == '2'))
				ifcount++;
		}
	}

	free(ipmifdata);
	return ifcount;
}
