/*
 *  TCP MIB group implementation - tcp.c
 *
 */

#include <config.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/param.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <sys/queue.h>
#include <netinet/ip_var.h>
#include <sys/socketvar.h>
#include <netinet/in_pcb.h>
#include "tools.h"
#include "kernel.h"

#include "system.h"
#include <sys/sysctl.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcp_fsm.h>

#include "mibincl.h"
#include "util_funcs.h"
#include "auto_nlist.h"

#include "tcp.h"
#include "tcpTable.h"
#include "sysORTable.h"

#include <ipmd-lib.h>
#include <snmpmsgapi.h>
#include "ipifhelp.h"

#define TCP_STATS_CACHE_TIMEOUT	MIB_STATS_CACHE_TIMEOUT
marker_t tcp_stats_cache_marker = NULL;

// static struct timeval msg_timeout = {3, 0};  // We'll wait 3 sec

static unsigned int tcpRtoAlgorithm = 1;
static unsigned int tcpRtoMin = 200;
static unsigned int tcpRtoMax = 60000;
static unsigned int tcpMaxConns = (unsigned int) -1;

void tcp_parse_config_algorithm(const char *token, char *cptr);
void tcp_parse_config_rtomin(const char *token, char *cptr);
void tcp_parse_config_rtomax(const char *token, char *cptr);
void tcp_parse_config_maxconn(const char *token, char *cptr);

	/*********************
	 *
	 *  snmpd.conf config parsing
	 *
	 *********************/

void tcp_parse_config_algorithm(const char *token, char *cptr)
{
	char tmpbuf[1024];
	
	if (strcmp(cptr,"\"\"") == 0) {
		tcpRtoAlgorithm = 1;	// default value
	} else if (strcmp(cptr, "other") == 0) {
		tcpRtoAlgorithm = 1;
	} else if (strcmp(cptr, "constant") == 0) {
		tcpRtoAlgorithm = 2;
	} else if (strcmp(cptr, "rsre") == 0) {
		tcpRtoAlgorithm = 3;
	} else if (strcmp(cptr, "vanj") == 0) {
		tcpRtoAlgorithm = 4;
	} else {
		sprintf(tmpbuf, "tcprtoalgorithm token invalid: %s", cptr);
		config_perror(tmpbuf);
	}
}

void tcp_parse_config_rtomin(const char *token, char *cptr)
{
	tcpRtoMin = atoi(cptr);
}

void tcp_parse_config_rtomax(const char *token, char *cptr)
{
	tcpRtoMax = atoi(cptr);
}

void tcp_parse_config_maxconn(const char *token, char *cptr)
{
	tcpMaxConns = atoi(cptr);
}


	/*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/

	/*********************
	 *
	 *  Initialisation & common implementation functions
	 *
	 *********************/

struct variable13 tcp_variables[] = {
    {TCPRTOALGORITHM, ASN_INTEGER, RONLY, var_tcp, 1, {1}},
    {TCPRTOMIN, ASN_INTEGER, RONLY, var_tcp, 1, {2}},
    {TCPRTOMAX, ASN_INTEGER, RONLY, var_tcp, 1, {3}},
    {TCPMAXCONN, ASN_INTEGER, RONLY, var_tcp, 1, {4}},
    {TCPACTIVEOPENS, ASN_COUNTER, RONLY, var_tcp, 1, {5}},
    {TCPPASSIVEOPENS, ASN_COUNTER, RONLY, var_tcp, 1, {6}},
    {TCPATTEMPTFAILS, ASN_COUNTER, RONLY, var_tcp, 1, {7}},
    {TCPESTABRESETS, ASN_COUNTER, RONLY, var_tcp, 1, {8}},
    {TCPCURRESTAB, ASN_GAUGE, RONLY, var_tcp, 1, {9}},
    {TCPINSEGS, ASN_COUNTER, RONLY, var_tcp, 1, {10}},
    {TCPOUTSEGS, ASN_COUNTER, RONLY, var_tcp, 1, {11}},
    {TCPRETRANSSEGS, ASN_COUNTER, RONLY, var_tcp, 1, {12}},
    {TCPCONNSTATE, ASN_INTEGER, RONLY, var_tcpEntry, 3, {13, 1, 1}},
    {TCPCONNLOCALADDRESS, ASN_IPADDRESS, RONLY, var_tcpEntry, 3, {13, 1, 2}},
    {TCPCONNLOCALPORT, ASN_INTEGER, RONLY, var_tcpEntry, 3, {13, 1, 3}},
    {TCPCONNREMADDRESS, ASN_IPADDRESS, RONLY, var_tcpEntry, 3, {13, 1, 4}},
    {TCPCONNREMPORT, ASN_INTEGER, RONLY, var_tcpEntry, 3, {13, 1, 5}},
    {TCPINERRS, ASN_COUNTER, RONLY, var_tcp, 1, {14}},
    {TCPOUTRSTS, ASN_COUNTER, RONLY, var_tcp, 1, {15}}
};

/* Define the OID pointer to the top of the mib tree that we're
   registering underneath, and the OID for the MIB module */
oid tcp_variables_oid[] = { SNMP_OID_MIB2,6 };
oid tcp_module_oid[]    = { SNMP_OID_MIB2,49 };

void init_tcp(void)
{
  /* register ourselves with the agent to handle our mib tree */
	REGISTER_MIB("mibII/tcp", tcp_variables, variable13, tcp_variables_oid);
	REGISTER_SYSOR_ENTRY( tcp_module_oid,
				"The MIB module for managing TCP implementations");

	/* register our config handlers */
	snmpd_register_config_handler("tcprtoalgorithm", tcp_parse_config_algorithm,
								NULL, "other/constant/rsre/vanj");
	snmpd_register_config_handler("tcprtomin", tcp_parse_config_rtomin,
								NULL, "RTO-minimum");
	snmpd_register_config_handler("tcprtomax", tcp_parse_config_rtomax,
								NULL, "RTO-maximum");
	snmpd_register_config_handler("tcpmaxconnects", tcp_parse_config_maxconn,
								NULL, "maximum-connections");
}


	/*********************
	 *
	 *  System specific implementation functions
	 *
	 *********************/


long read_tcp_stat(ipm_tcpstat_t *, int);

u_char *
var_tcp(struct variable *vp,
	oid *name,
	size_t *length,
	int exact,
	size_t *var_len,
	WriteMethod **write_method)
{
    static ipm_tcpstat_t tcpstat;
    static long ret_value;

    if (header_generic(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
		return NULL;

	if (vp->magic == TCPRTOALGORITHM) {
		long_return = tcpRtoAlgorithm;
		return (u_char *) &long_return;
	}
	else if (vp->magic == TCPRTOMIN) {
		long_return = tcpRtoMin;
		return (u_char *) &long_return;
	}
	else if (vp->magic == TCPRTOMAX) {
		long_return = tcpRtoMax;
		return (u_char *) &long_return;
	}
	else if (vp->magic == TCPMAXCONN) {
		long_return = tcpMaxConns;
		return (u_char *) &long_return;
	}

    ret_value = read_tcp_stat(&tcpstat, vp->magic);
    if (ret_value < 0)
		return NULL;

    switch (vp->magic){
	case TCPACTIVEOPENS:
		long_return = tcpstat.tcps_connattempt;
		return (u_char *) &long_return;
	case TCPPASSIVEOPENS:
		long_return = tcpstat.tcps_accepts;
		return (u_char *) &long_return;
	case TCPATTEMPTFAILS:
		long_return = tcpstat.tcps_conndrops;
		return (u_char *) &long_return;
	case TCPESTABRESETS:
		long_return = tcpstat.tcps_drops;
		return (u_char *) &long_return;
	case TCPCURRESTAB:
		long_return = tcpstat.tcps_connects;
		return (u_char *) &long_return;
	case TCPINSEGS:
		long_return = tcpstat.tcps_rcvtotal;
		return (u_char *) &long_return;
	case TCPOUTSEGS:
		long_return = tcpstat.tcps_sndtotal - tcpstat.tcps_sndrexmitpack;
		return (u_char *) &long_return;
	case TCPRETRANSSEGS:
		long_return = tcpstat.tcps_sndrexmitpack;
		return (u_char *) &long_return;
	case TCPINERRS:
		long_return = tcpstat.tcps_rcvbadsum + tcpstat.tcps_rcvbadoff +
						tcpstat.tcps_rcvshort;
		return (u_char *) &long_return;
	case TCPOUTRSTS:
		long_return = tcpstat.tcps_sndctrl - tcpstat.tcps_closed;
		return (u_char *) &long_return;

	default:
		DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_tcp\n", vp->magic));
    }
    return NULL;
}


	/*********************
	 *
	 *  Internal implementation functions
	 *
	 *********************/

long
read_tcp_stat(ipm_tcpstat_t *tcpstat, int magic)
{
	long ret_value = -1;
	int gbeslots[32], numofgbeslots;
	ipm_tcpstat_t tmpstat;
	struct tcpstat ssctcpstat;
	int count, needed;

    if (tcp_stats_cache_marker &&
		(!atime_ready( tcp_stats_cache_marker, TCP_STATS_CACHE_TIMEOUT*1000 )))
		return 0;

    if (tcp_stats_cache_marker )
		atime_setMarker( tcp_stats_cache_marker );
    else
		tcp_stats_cache_marker = atime_newMarker();

	/* get TCP stats on SSC ports */
	if (auto_nlist(TCPSTAT_SYMBOL, (char *)&ssctcpstat, sizeof (struct tcpstat)))
		ret_value = 0;
	else
		goto finished_readtcpstat;

	/* get TCP stats on GbE ports */
	numofgbeslots = 32;
	if (IpInterface_Get_GbE_Slot(gbeslots, &numofgbeslots, NULL) >= 0) {
		int rval = -1;

		/* for now, we assume we have only one GbE slot */
		count = 1;
		needed = 0;
		rval = ipmd_GetTcpStats(gbeslots[0],
				&tmpstat, sizeof(ipm_tcpstat_t), &count, &needed);
		if ((rval == IPM_GET_NO_BUF) || (rval == IPM_CMDERR_APPNOBUF)) {
			/* need to do something here.... */
		}
		else if (rval == IPM_GET_ERR) {
			/* something wrong here... */
		}
		else {
			/* assume there is only one GbE slot */
			/* if more than one, then we need to */
			/* add all the stats together. But   */
			/* we don't support that yet...      */
			memcpy(tcpstat, &tmpstat, sizeof(ipm_tcpstat_t));
		}
	}

	/* now combine tcp stats from both SSC and GbE */
	tcpstat->tcps_connattempt += ssctcpstat.tcps_connattempt;
	tcpstat->tcps_accepts += ssctcpstat.tcps_accepts;
	tcpstat->tcps_connects += ssctcpstat.tcps_connects;
	tcpstat->tcps_drops += ssctcpstat.tcps_drops;
	tcpstat->tcps_conndrops += ssctcpstat.tcps_conndrops;
	tcpstat->tcps_closed += ssctcpstat.tcps_closed;
	tcpstat->tcps_segstimed += ssctcpstat.tcps_segstimed;
	tcpstat->tcps_rttupdated += ssctcpstat.tcps_rttupdated;
	tcpstat->tcps_delack += ssctcpstat.tcps_delack;
	tcpstat->tcps_timeoutdrop += ssctcpstat.tcps_timeoutdrop;
	tcpstat->tcps_rexmttimeo += ssctcpstat.tcps_rexmttimeo;
	tcpstat->tcps_persisttimeo += ssctcpstat.tcps_persisttimeo;
	tcpstat->tcps_keeptimeo += ssctcpstat.tcps_keeptimeo;
	tcpstat->tcps_keepprobe += ssctcpstat.tcps_keepprobe;
	tcpstat->tcps_keepdrops += ssctcpstat.tcps_keepdrops;
	tcpstat->tcps_sndtotal += ssctcpstat.tcps_sndtotal;
	tcpstat->tcps_sndpack += ssctcpstat.tcps_sndpack;
	tcpstat->tcps_sndbyte += ssctcpstat.tcps_sndbyte;
	tcpstat->tcps_sndrexmitpack += ssctcpstat.tcps_sndrexmitpack;
	tcpstat->tcps_sndrexmitbyte += ssctcpstat.tcps_sndrexmitbyte;
	tcpstat->tcps_sndacks += ssctcpstat.tcps_sndacks;
	tcpstat->tcps_sndprobe += ssctcpstat.tcps_sndprobe;
	tcpstat->tcps_sndurg += ssctcpstat.tcps_sndurg;
	tcpstat->tcps_sndwinup += ssctcpstat.tcps_sndwinup;
	tcpstat->tcps_sndctrl += ssctcpstat.tcps_sndctrl;
	tcpstat->tcps_rcvtotal += ssctcpstat.tcps_rcvtotal;
	tcpstat->tcps_rcvpack += ssctcpstat.tcps_rcvpack;
	tcpstat->tcps_rcvbyte += ssctcpstat.tcps_rcvbyte;
	tcpstat->tcps_rcvbadsum += ssctcpstat.tcps_rcvbadsum;
	tcpstat->tcps_rcvbadoff += ssctcpstat.tcps_rcvbadoff;
	tcpstat->tcps_rcvshort += ssctcpstat.tcps_rcvshort;
	tcpstat->tcps_rcvduppack += ssctcpstat.tcps_rcvduppack;
	tcpstat->tcps_rcvdupbyte += ssctcpstat.tcps_rcvdupbyte;
	tcpstat->tcps_rcvpartduppack += ssctcpstat.tcps_rcvpartduppack;
	tcpstat->tcps_rcvpartdupbyte += ssctcpstat.tcps_rcvpartdupbyte;
	tcpstat->tcps_rcvoopack += ssctcpstat.tcps_rcvoopack;
	tcpstat->tcps_rcvoobyte += ssctcpstat.tcps_rcvoobyte;
	tcpstat->tcps_rcvpackafterwin += ssctcpstat.tcps_rcvpackafterwin;
	tcpstat->tcps_rcvbyteafterwin += ssctcpstat.tcps_rcvbyteafterwin;
	tcpstat->tcps_rcvafterclose += ssctcpstat.tcps_rcvafterclose;
	tcpstat->tcps_rcvwinprobe += ssctcpstat.tcps_rcvwinprobe;
	tcpstat->tcps_rcvdupack += ssctcpstat.tcps_rcvdupack;
	tcpstat->tcps_rcvacktoomuch += ssctcpstat.tcps_rcvacktoomuch;
	tcpstat->tcps_rcvackpack += ssctcpstat.tcps_rcvackpack;
	tcpstat->tcps_rcvackbyte += ssctcpstat.tcps_rcvackbyte;
	tcpstat->tcps_rcvwinupd += ssctcpstat.tcps_rcvwinupd;
	tcpstat->tcps_pawsdrop += ssctcpstat.tcps_pawsdrop;
	tcpstat->tcps_predack += ssctcpstat.tcps_predack;
	tcpstat->tcps_preddat += ssctcpstat.tcps_preddat;
	tcpstat->tcps_persistdrop += ssctcpstat.tcps_persistdrop;
	tcpstat->tcps_badsyn += ssctcpstat.tcps_badsyn;

finished_readtcpstat:
    if (ret_value == -1) {
		free(tcp_stats_cache_marker);
		tcp_stats_cache_marker = NULL;
    }
    return ret_value;
}
