/*
 *-----------------------------------------------------------------
 *
 * Name: chassis-ui.c
 *
 * RCS:  $Id: chassis-ui.c,v 1.7 2002/11/19 04:43:32 tonyw Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: user interface functions for chassis manager
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>

#include <nfx-incl.h>
#include <exports/anpssc-api.h>
#include <exports/cm-api.h>
#include "cm.h"

/*
 * UI parsing and structures copied from sm-sdm/storage.c - thanks Frank
 */
#define LEAF	0
#define BRANCH	1

typedef struct ui_node {
	char			* name;
	int32			type;		/* LEAF or BRANCH */
	int32			arg_cnt;	/* minimum argument count for leaf node */
	struct ui_node	* branch;
	int32			(*cmd_func)();
	uchar8			* usage;
} ui_node_t;

// Function prototypes
int chassis_ui_usage(char *, ui_node_t *);
int chassis_ui_parse(char *, ui_node_t *, int, char **);
int chassis_ui_viewSystem(int, char **);
int chassis_ui_viewPower(int, char **);
int chassis_ui_viewTemp(int, char **);
int chassis_ui_viewVersion(int, char **);
int chassis_ui_rebootSlot(int, char **);
int chassis_ui_rebootCPU(int, char **);
int chassis_ui_debug(int, char **);

char * chassis_ui_getModuleString(int);
char * chassis_ui_getCPUString(int, int);
char * chassis_ui_getCPUState(int);

// View Commands
ui_node_t chassis_ui_view[] = {
	{"sys",		LEAF,	0,	0,	chassis_ui_viewSystem,	"view sys"},
	{"temp",	LEAF,	0,	0,	chassis_ui_viewTemp,	"view temp"},
	{"power",	LEAF,	0,	0,	chassis_ui_viewPower,	"view power"},
	{"ver",		LEAF,	0,	0,	chassis_ui_viewVersion,	"view ver"},
	{0, 0, 0, 0, 0, 0},
};

// Reboot Commands
ui_node_t chassis_ui_reboot[] = {
	{"slot",LEAF,	1,	0,	chassis_ui_rebootSlot,	"reboot slot <slot>"},
	{"cpu",	LEAF,	2,	0,	chassis_ui_rebootCPU,	"reboot cpu <slot> <cpu>"},
	{0, 0, 0, 0, 0, 0},
};

// Top Level Command Table
ui_node_t chassis_ui_root[] = {
	{"view",		BRANCH,	2,	chassis_ui_view,	0,	0},
	{"reboot",		BRANCH,	2,	chassis_ui_reboot,	0,	0},
	{"debug",		LEAF,	0,	0,	chassis_ui_debug,	"debug"},
	{0, 0, 0, 0, 0, 0},
};

/*-----------------------------------------------------------------
 * Name:	main()
 *
 * Description: entry point for "chassis" command
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
main(
	int		argc,
	char	* argv[]
)
{
	if (argc == 1) {
		printf("\n");
		chassis_ui_usage(argv[0], &chassis_ui_root[0]);
		printf("\n");
		return 0;
	}
	chassis_ui_parse(argv[0], &chassis_ui_root[0], --argc, &argv[1]);
	return 0;

}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_usage(
	char		* prg_name,
	ui_node_t	*nd
)
{
	ui_node_t *node;

    node = nd;
	while(node->name) {
		if( node->type == LEAF )
			printf("Usage: %s %s\n", prg_name, node->usage);
		else
			chassis_ui_usage(prg_name, node->branch);
		node = node++;
	} // while
	return 0;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_parse(
	char		* prg_name,
	ui_node_t	* nd,
	int			argc,
	char		* argv[]
)
{
	int rc;
	char *tok;
	ui_node_t *node;

	node = nd;
	tok = argv[0];

	if( tok ) {
		while( node->name ) {
			if( !(strcmp(tok, node->name)) ) {
				if( node->type == LEAF ) {
					if( --argc >= node->arg_cnt ) {
						rc = node->cmd_func(argc, &argv[1]);
						if( rc < 0 )
							printf("Usage: %s %s\n", prg_name, node->usage);
					}
					else
						printf("Usage: %s %s\n", prg_name, node->usage);
				}
				else {
					chassis_ui_parse(prg_name, node->branch, --argc, &argv[1]);
				}
				return 0;
			}
			else {
				node++;
			}
		} // while
	}
	printf("Command Not Found.\n");
	return 0;
}

/*-----------------------------------------------------------------
 * Name:	bad_stat
 *
 * Description:	General check of sendAgileMsg() returned status
 *
 * Created by:	Greg Seibert
 *
 * Date Created:	20 June 2001
 *
 *-----------------------------------------------------------------
 */
static int bad_stat (int stat)
{
	if (stat < 0) {
		perror ("Send to chassisd failed");
		return 1;
	}
	else if (!stat) {
		fprintf (stderr, "Timed out waiting for response\n");
		return 1;
	}
	return 0;
}

/*-----------------------------------------------------------------
 * Name: chassis_ui_viewSystem
 *
 * Description:	View board type and cpu status
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_viewSystem(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgInfoReq_t	req;
	chassisdMsgInfoResp_t	resp;
	struct timeval			to;
	int		i, j;

	bzero(&resp, sizeof(resp));

	req.hdr.mid = CHASSISD_GET_CHASSIS_INFO_REQ;

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	printf("\n");
	printf("slot    module     cpu         state\n");
	printf("----------------------------------------------\n");
	for (i=0; i<CM_NUM_SLOTS; i++) {
		if (resp.present[i]) {
			if (resp.num_cpus[i] == 0) {
				printf("%4d    %-8s\n", i, "UNKNOWN");
			}
			else {
				for (j=0; j<resp.num_cpus[i]; j++) {
					if (j == 0)
						printf("%4d    %-8s   ", i,
								chassis_ui_getModuleString(resp.board_type[i]));
					else
						printf("                   ");
					printf("%-8s    %s\n",
								chassis_ui_getCPUString(resp.board_type[i], j),
								chassis_ui_getCPUState(resp.cpu_state[i][j]));
				}
			}
		}
		else
			printf("%4d    %-8s\n", i, "EMPTY");
	}
	printf("----------------------------------------------\n");

	return 0;
}

/*-----------------------------------------------------------------
 * Name: chassis_ui_viewPower
 *
 * Description: View Power usage from each board
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_viewPower(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgPowerReq_t	req;
	chassisdMsgPowerResp_t	resp;
	struct timeval			to;
	int		i;

	bzero(&resp, sizeof(resp));

	req.hdr.mid = CHASSISD_GET_POWER_REQ;

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	printf("\n");
	printf("slot    power\n");
	printf("-------------\n");
	for (i=0; i<CM_NUM_SLOTS; i++) {
		if (resp.present[i]) {
			if (resp.valid[i])
				printf("%4d    %-5d\n", i, resp.power[i]);
			else
				printf("%4d    %-5s\n", i, "???");
		}
		else
			printf("%4d    %-5s\n", i, "EMPTY");
	}
	printf("-------------\n");

	return 0;
}

/*-----------------------------------------------------------------
 * Name: chassis_ui_viewTemp
 *
 * Description: View Temperature on each board
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_viewTemp(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgTempReq_t	req;
	chassisdMsgTempResp_t	resp;
	struct timeval			to;
	int		i;

	bzero(&resp, sizeof(resp));

	req.hdr.mid = CHASSISD_GET_TEMP_REQ;

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	printf("\n");
	printf("slot    temp\n");
	printf("------------\n");
	for (i=0; i<CM_NUM_SLOTS; i++) {
		if (resp.present[i]) {
			if (resp.valid[i])
				printf("%4d    %-4d\n", i, resp.temp[i]);
			else
				printf("%4d    %-4s\n", i, "???");
		}
		else
			printf("%4d    %-4s\n", i, "EMPTY");
	}
	printf("-------------\n");

	return 0;
}

/*-----------------------------------------------------------------
 * Name: chassis_ui_viewVersion
 *
 * Description: View Version strings from each board
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_viewVersion(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgVersionReq_t	req;
	chassisdMsgVersionResp_t	resp;
	struct timeval			to;
	int		i, j;

	bzero(&resp, sizeof(resp));

	req.hdr.mid = CHASSISD_GET_VERSION_REQ;

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	for (i=0; i<CM_NUM_SLOTS; i++) {
		int pad = 0;
		if (resp.present[i]) {
			printf("\n");
			if (strlen(resp.bmfpga_ver[i]) || strlen(resp.luc_ver[i]) ||
				strlen(resp.lmux_ver[i]) || strlen(resp.hw_ver[i])) {
				printf("slot %d\n", i);
				printf("-------\n");
				if (strlen(resp.bmfpga_ver[i])) {
					printf("  %s\n", resp.bmfpga_ver[i]);
				}
				if (strlen(resp.luc_ver[i])) {
					printf("  %s\n", resp.luc_ver[i]);
				}
				if (strlen(resp.lmux_ver[i])) {
					printf("  %s\n", resp.lmux_ver[i]);
				}
				if (strlen(resp.hw_ver[i])) {
					printf("  %s\n", resp.hw_ver[i]);
				}
			}
			for (j=0; j<CM_NUM_CPUS; j++) {
				if (strlen(resp.prom_ver[i][j]) || strlen(resp.rt_ver[i][j])) {
					printf("slot %d, %s\n", i,
								chassis_ui_getCPUString(resp.board_type[i], j));
					printf("----------------------\n");
					if (strlen(resp.prom_ver[i][j])) {
						printf("  %s\n", resp.prom_ver[i][j]);
					}
					if (strlen(resp.rt_ver[i][j])) {
						printf("  %s\n", resp.rt_ver[i][j]);
					}
				}
			}
		}
	}
	printf("\n");

	return 0;
}

/*-----------------------------------------------------------------
 * Name: chassis_ui_rebootSlot
 *
 * Description: Reboot an entine card
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_rebootSlot(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgRebootReq_t	req;
	chassisdMsgRebootResp_t	resp;
	struct timeval			to;
	int		slot;

	if (argc != 1)
		return -1;

	bzero(&resp, sizeof(resp));

	slot = atoi(argv[0]);

	if (slot == 0) {
		printf("Cannot reset slot 0 with this command\n");
		return 1;
	}

	if (slot >= CM_NUM_SLOTS) {
		printf("Invalid slot number, %d\n", slot);
		return 1;
	}

	req.hdr.mid = CHASSISD_REBOOT_REQ;
	req.reboot_type = CM_REBOOT_TYPE_SLOT;
	req.slot = slot;
	req.slice = 0;		/* unused */

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	if (resp.rc == CM_REBOOT_OK)
		printf("Slot %d rebooted\n", slot);
	else
		printf("Error rebooting slot %d, errno 0x%x\n", slot, resp.rc);

	return 0;
}

/*-----------------------------------------------------------------
 * Name: chassis_ui_rebootCPU
 *
 * Description: Reboot a CPU on a multi-slice card
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/24/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_rebootCPU(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgRebootReq_t	req;
	chassisdMsgRebootResp_t	resp;
	struct timeval			to;
	int		slot, slice;

	if (argc != 2)
		return -1;

	bzero(&resp, sizeof(resp));

	slot = atoi(argv[0]);
	slice = atoi(argv[1]);

	if (slot == 0) {
		printf("Cannot reset slot 0 with this command\n");
		return 1;
	}

	if (slot >= CM_NUM_SLOTS) {
		printf("Invalid slot number, %d\n", slot);
		return 1;
	}
	if (slice >= CM_NUM_CPUS) {
		printf("Invalid cpu number, %d\n", slice);
		return 1;
	}

	req.hdr.mid = CHASSISD_REBOOT_REQ;
	req.reboot_type = CM_REBOOT_TYPE_SLICE;
	req.slot = slot;
	req.slice = slice;

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	if (resp.rc == CM_REBOOT_OK)
		printf("Slot %d, cpu %d rebooted\n", slot, slice);
	else
		printf("Error rebooting slot %d, cpu %d, errno 0x%x\n", slot, slice, resp.rc);

	return 0;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/25/02
 *
 *-----------------------------------------------------------------
 */
int
chassis_ui_debug(
	int		argc,
	char	* argv[]
)
{
	int		stat;
	chassisdMsgDebugReq_t	req;
	chassisdMsgDebugResp_t	resp;
	struct timeval			to;
	cm_t	* cm;
	int		slot, cpu;

	bzero(&resp, sizeof(resp));

	req.hdr.mid = CHASSISD_DEBUG_REQ;

	to.tv_sec = 0;
	to.tv_usec = 500;

	stat = sendAgileMsg(-1, &req, sizeof(req),
						"cm", 0, 0, 0,
						&to,
						&resp, sizeof(resp));

	if (bad_stat(stat))
		return stat;

	cm = (cm_t *)&(resp.cm);

	printf("chassisd cm_t structure:\n");
	for (slot=0; slot<CM_NUM_SLOTS; slot++) {
		printf("\tcm.slot[%d].present = 0x%x\n", slot, cm->slot[slot].present);
		printf("\tcm.slot[%d].type = 0x%x\n", slot, cm->slot[slot].type);
		printf("\tcm.slot[%d].power_valid = 0x%x\n", slot, cm->slot[slot].power_valid);
		printf("\tcm.slot[%d].power = 0x%x\n", slot, cm->slot[slot].power);
		printf("\tcm.slot[%d].temp_valid = 0x%x\n", slot, cm->slot[slot].temp_valid);
		printf("\tcm.slot[%d].temp = 0x%x\n", slot, cm->slot[slot].temp);
		printf("\tcm.slot[%d].state = 0x%x\n", slot, cm->slot[slot].state);
        if(strlen(cm->slot[slot].bmfpga_ver))
            printf("\tbmfpga_ver = %s\n", cm->slot[slot].bmfpga_ver);
        if(strlen(cm->slot[slot].luc_ver))
            printf("\tluc_ver    = %s\n", cm->slot[slot].luc_ver);
        if(strlen(cm->slot[slot].lmux_ver))
            printf("\tlmux_ver   = %s\n", cm->slot[slot].lmux_ver);
        if(strlen(cm->slot[slot].hw_ver))
            printf("\thw_ver   = %s\n", cm->slot[slot].hw_ver);
		printf("\tcm.slot[%d].num_cpus = 0x%x\n", slot, cm->slot[slot].num_cpus);
		for (cpu=0; cpu<CM_NUM_CPUS; cpu++) {
			printf("\tcm.slot[%d].cpu[%d].state = 0x%x\n",
                    slot, cpu, cm->slot[slot].cpu[cpu].state);
            if(strlen(cm->slot[slot].cpu[cpu].prom_ver))
                printf("\t\tprom_ver = %s\n", cm->slot[slot].cpu[cpu].prom_ver);
            if(strlen(cm->slot[slot].cpu[cpu].rt_ver))
                printf("\t\trt_ver   = %s\n", cm->slot[slot].cpu[cpu].rt_ver);
        }
		printf("\n");
	}

	return 0;
}
