/*
 *-----------------------------------------------------------------
 *
 * Name: cm-fan.c
 *
 * Copyright (C) 2004, OnStor Inc.
 *
 * Description: Maxim max6550 fan controller routines for Bobcat HW.
 *
 * Created by:  Shaun Hui
 *
 * Date Created:    6/18/04
 *
 *-----------------------------------------------------------------
 */
                                                                                
#ifdef R9000
                                                                                
#include <nfx-incl.h>
#include <exports/hwMap.h>
#include <exports/cm-api.h>
#include <exports/i2c-api.h>
#include <exports/timer-api.h>
#include <sys/bmfpga.h>
#include "cm-fan.h"

typedef struct _cmFanCb_t
{
	uint32	timer_id;
	int32 	failed_count;
	int32 	clear_count;
	int32 	debug;
} cmFanCb_t;

cmFanCb_t	cmFanCb;

void    delay_us(unsigned int);

int32
cmFanMonitor(uint32	tmr,
			 void	*info1,
			 void	*info2,
			 void	*info3,
			 void	*info4 );

int32	fanCtrlTbl[FAN_MAX_CTRL] = {I2C_ADDR_FAN_CTRL_0,I2C_ADDR_FAN_CTRL_1};
int32	fanTachMax[FAN_MAX_CTRL] = {FAN_0_MAX_TACH_REG, FAN_1_MAX_TACH_REG};

/*-----------------------------------------------------------------
 * Name: cmInitFan
 *
 * Description: Initialize fan controller via I2C.
 *
 *-----------------------------------------------------------------
 */
void 
cmInitFan( int32 ctrl_addr )
{
	uchar8	data;
	uchar8  bus=0;

	// Setup fan speed divisor
	data = 1;
	i2c(I2C_WRITE, bus, ctrl_addr, FAN_COUNT_REG, 1, &data);
	delay_us(10000);

	// Setup fan speed divisor
	data = 0x20;
	i2c(I2C_WRITE, bus, ctrl_addr, FAN_SPEED_REG, 1, &data);
	delay_us(10000);
	
	// Read ALARM_REG to clear alarms, ignore data
	data = 0x00;
	i2c(I2C_READ, bus, ctrl_addr, FAN_ALARM_REG, 1, &data);
	delay_us(10000);
	
	// Enable closed-loop control
	data = FAN_CONFIG_CLOSE_LOOP | FAN_CONFIG_12VOLT | FAN_CONFIG_DIV_BY_4;
	i2c(I2C_WRITE, bus, ctrl_addr, FAN_CONFIG_REG, 1, &data);
	delay_us(10000);
	
	// Enable alarms
	data = FAN_ALARM_ENABLE_MAX | FAN_ALARM_ENABLE_MIN | FAN_ALARM_ENABLE_TACH;
	i2c(I2C_WRITE, bus, ctrl_addr, FAN_ALARM_EN_REG, 1, &data);
	delay_us(10000);
	
	// Clear alarms
	data = 0x00;
	i2c(I2C_READ, bus, ctrl_addr, FAN_ALARM_REG, 1, &data);
	delay_us(10000);

	i2c(I2C_READ, bus, ctrl_addr, FAN_ALARM_REG, 1, &data);

	if (data)
	{
		BMFPGA_WRITE(BMFPGA_REG_STATUS_CTRL, BMFPGA_STATUS_CTRL_FAN_CTRL_ON);
	}
	
	// Setup GPIO bits
	data =  FAN_GPIO_DEF_4_LOGIC_HI | FAN_GPIO_DEF_3_LOGIC_HI |
			FAN_GPIO_DEF_2_LOGIC_HI | FAN_GPIO_DEF_1_LOGIC_LO |
			FAN_GPIO_DEF_0_ALERT;
	i2c(I2C_WRITE, bus, ctrl_addr, FAN_GPIO_DEF_REG, 1, &data);
	delay_us(10000);
}

void
cmFanMonitorInit( void)
{
	
	// create & start timer to monitor fan status
	cmFanCb.failed_count = 0;
	cmFanCb.clear_count  = 0;
	cmFanCb.debug  = 0;
	cmFanCb.timer_id = eee_timerCreate(FAN_TIMER_INTERVAL, EEE_TIMER_REPEAT);	
	if (cmFanCb.timer_id == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		printf("%s: failed to create timer!\n", __FUNCTION__);
	}
	else
	{
		eee_timerStart(cmFanCb.timer_id, cmFanMonitor, 0, 0, 0, 0);
	}
}

/*-----------------------------------------------------------------
 * Name: cmTempEna
 *
 * Description: Enable temp. monitor 
 *
 *-----------------------------------------------------------------
 */
void
cmTempEna( int32 ctrl_addr )
{
	uchar8	data;

	data=1;
	i2c(I2C_WRITE, 0, ctrl_addr, 0xee, 1, &data);
}

/*-----------------------------------------------------------------
 * Name: is_cmFanSpeedLow
 *
 * Description: Check if fan speed is within low range.
 *
 *-----------------------------------------------------------------
 */
int
is_cmFanSpeedLow( uchar8 count )
{
	return ((count > FAN_LOW_MIN_REV_COUNT) && 
			(count < FAN_LOW_MAX_REV_COUNT)) ?  TRUE: FALSE;
}

/*-----------------------------------------------------------------
 * Name: is_cmFanSpeedHigh
 *
 * Description: Check if fan speed is within high range.
 *
 *-----------------------------------------------------------------
 */
int
is_cmFanSpeedHigh( uchar8 count )
{
	return ((count > FAN_HIGH_MIN_REV_COUNT) && 
			(count < FAN_HIGH_MAX_REV_COUNT)) ?  TRUE: FALSE;
}

/*-----------------------------------------------------------------
 * Name: is_cmFanFullOn
 *
 * Description: Check if fan full on is set in BMFPGA.
 *
 *-----------------------------------------------------------------
 */
int
is_cmFanFullOn( void )
{
	uint32	status;

	// Read BMFPGA status reg
	status = BMFPGA_READ(BMFPGA_REG_STATUS_CTRL); 

	// Check FAN control ON bit.
	return ((status & BMFPGA_STATUS_CTRL_FAN_CTRL_ON) ?  TRUE : FALSE);
}

/*-----------------------------------------------------------------
 * Name: is_cmFanStatusOk
 *
 * Description: Check if fan status Ok bits are set in BMFPGA.
 *
 *-----------------------------------------------------------------
 */
int
is_cmFanStatusOk( void )
{
	uint32	status;

	// Read BMFPGA status reg
	status = BMFPGA_READ(BMFPGA_REG_STATUS_CTRL); 

	// Both fan status bits must be set. 
	if ((status & BMFPGA_STATUS_CTRL_FAN0_STATUS) &&
		(status & BMFPGA_STATUS_CTRL_FAN1_STATUS)) 
		return TRUE;

	return FALSE;
}

/*-----------------------------------------------------------------
 * Name: cmFanDisFullOn
 *
 * Description: Disable fan full on in BMFPGA.
 *
 *-----------------------------------------------------------------
 */
void
cmFanDisFullOn( void )
{
	uint32	status;

	if (cmFanCb.debug)
		printf("%s\n", __FUNCTION__);

	// Read BMFPGA status reg & clear FAN ON bit
	status  = BMFPGA_READ(BMFPGA_REG_STATUS_CTRL); 
	status &= ~BMFPGA_STATUS_CTRL_FAN_CTRL_ON;

	BMFPGA_WRITE(BMFPGA_REG_STATUS_CTRL, status);
}

/*-----------------------------------------------------------------
 * Name: cmFanEnaFullOn
 *
 * Description: Enable fan full on in BMFPGA.
 *
 *-----------------------------------------------------------------
 */
void
cmFanEnaFullOn( void )
{
	uint32	status;

	if (cmFanCb.debug)
		printf("%s\n", __FUNCTION__);

	// Read BMFPGA status reg & set FAN ON bit
	status = BMFPGA_READ(BMFPGA_REG_STATUS_CTRL); 
	status |= BMFPGA_STATUS_CTRL_FAN_CTRL_ON;

	BMFPGA_WRITE(BMFPGA_REG_STATUS_CTRL, status);
}

/*-----------------------------------------------------------------
 * Name: cmFanClrAlarm
 *
 * Description: Clear fan alarms
 *
 *-----------------------------------------------------------------
 */
void
cmFanClrAlarm( void )
{
	uchar8	data;

	if (cmFanCb.debug)
		printf("%s\n", __FUNCTION__);

	// Read fan ctrl alarm reg to clear alarms.
	i2c(I2C_READ, 0, I2C_ADDR_FAN_CTRL_0, FAN_ALARM_REG, 1, &data);
	delay_us(10000);
	i2c(I2C_READ, 0, I2C_ADDR_FAN_CTRL_1, FAN_ALARM_REG, 1, &data);
	delay_us(10000);
}

/*-----------------------------------------------------------------
 * Name: cmFanEnaCloseLoop
 *
 * Description: Enable fan close loop control
 *
 *-----------------------------------------------------------------
 */
void
cmFanEnaCloseLoop( void )
{
	uchar8	data;

	if (cmFanCb.debug)
		printf("%s\n", __FUNCTION__);

	// Re-enable close loop control 
	data = FAN_CONFIG_CLOSE_LOOP | FAN_CONFIG_12VOLT | FAN_CONFIG_DIV_BY_4;
	i2c(I2C_WRITE, 0, I2C_ADDR_FAN_CTRL_0, FAN_CONFIG_REG, 1, &data);
	delay_us(10000);
	i2c(I2C_WRITE, 0, I2C_ADDR_FAN_CTRL_1, FAN_CONFIG_REG, 1, &data);
	delay_us(10000);
}

/*-----------------------------------------------------------------
 * Name: cmFanFailed
 *
 * Description: Enable fan close loop control
 *
 *-----------------------------------------------------------------
 */
void
cmFanFailed( int32 ctrl, int32 reg, int32 val)
{
	// Report bad fan controller & register.
	if (cmFanCb.debug)
	{
		printf("%s\n", __FUNCTION__);
		printf("Controller [0x%x] failed: reg[0x%x]=0x%x\n", ctrl, reg, val);
		printf("Fan failure count=%d all on=%d\n", 
			cmFanCb.failed_count, is_cmFanFullOn()); 
	}

	// increment failed count
	if (cmFanCb.failed_count++ < FAN_MAX_FAIL_COUNT)
		return;
	
	// nothing to do if fan is full on.
	if (is_cmFanFullOn( )==TRUE)
		return;
	
	cmFanEnaFullOn( );	// Turn on ALL fans
	cmFanClrAlarm( );	// Clear fan alarm conditions
}

/*-----------------------------------------------------------------
 * Name: cmFanMonitor
 *
 * Description: Fan Monitor called by timer.
 *              Go thru TACH register for each fan controller.
 *              If any one of the TACH value is out of range,
 *              or temperture is too high, assume a fan is bad. 
 *              After 3 times, turn all fans to FULL power.
 *
 *              If TACH values are in range & temperture returns 
 *				to normal, disable FULL power and enable close loop 
 *				control.
 *
 *-----------------------------------------------------------------
 */
int32
cmFanMonitor(uint32	tmr,
			 void	*info1,
			 void	*info2,
			 void	*info3,
			 void	*info4 )
{
	int32	i, j;
	int32	ctrl_addr;
	int32	tach_max;
	int32	tach_reg;
	int32	fan;
	uchar8	temp_val;
	uchar8	tach_val[FAN_MAX_CTRL][FAN_1_MAX_TACH_REG];

	fan = 0;

	// Read tach register values from FAN Controllers 0 & 1
	for (i=0; i < FAN_MAX_CTRL; i++)
	{ 
		ctrl_addr = fanCtrlTbl[i];	// set ctrl I2C address
		tach_max  = fanTachMax[i];	// set # of tach reg to read
		tach_reg  = FAN_TACH0_REG;	// set base tach reg addr

		for (j=0; j < tach_max; j++)
		{
			i2c(I2C_READ, 0, ctrl_addr, tach_reg, 1, &tach_val[i][j]);
			delay_us(10000);
			if (cmFanCb.debug)
				printf("fan[%d] ctrl=0x%0x tach=0x%x val=0x%0x\n",
						fan, ctrl_addr, tach_reg, tach_val[i][j]);

			// check if fan speed is within low range
			if (is_cmFanSpeedLow(tach_val[i][j]) != TRUE)
			{
				// check if fan speed is within high range
				if (is_cmFanSpeedHigh(tach_val[i][j]) != TRUE)
				{
					cmFanFailed( ctrl_addr, tach_reg, tach_val[i][j] );
					return NFX_OK;
				}
			}
			tach_reg += FAN_REG_OFFSET;
			fan++;
		}
	}

	// read temperature.
	i2c(I2C_READ, 0, I2C_ADDR_TEMP_SENSOR, 0xAA, 1, &temp_val);

	// temperature too high, treat as fan failure.
	if (temp_val >= FAN_MAX_CHASSIS_TEMP)
	{
		cmFanFailed( I2C_ADDR_TEMP_SENSOR, 0xAA, temp_val);
		return NFX_OK;
	}

	// temperature & TACH values are in range.
	// Disable FULL on and enable close loop control.
	if (is_cmFanFullOn() == TRUE)
	{
		cmFanClrAlarm( );

		// leave all fans on if failed more than max. times.
		if (cmFanCb.clear_count++ > FAN_MAX_FAIL_COUNT)
			return NFX_OK;

		cmFanDisFullOn( );

		delay_us(100);

		// Need to clear alarm before enabling close loop.
		cmFanClrAlarm( );
		cmFanEnaCloseLoop( );
	}
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: cmFanShowStatus
 *
 * Description: Show fan controller status
 *
 *-----------------------------------------------------------------
 */
void
cmFanShowStatus( void )
{
	int32	i, j, fan;
	int32	ctrl_addr;
	int32	tach_max;
	int32	tach_reg;
	uchar8	temp_val;
	uchar8	tach_val[FAN_MAX_CTRL][FAN_1_MAX_TACH_REG];

	fan = 0;

	// Read tach register values from FAN Controllers 0 & 1
	for (i=0; i < FAN_MAX_CTRL; i++)
	{ 
		ctrl_addr = fanCtrlTbl[i];	// set ctrl I2C address
		tach_max  = fanTachMax[i];	// set # of tach reg to read
		tach_reg  = FAN_TACH0_REG;	// set base tach reg addr

		for (j=0; j < tach_max; j++)
		{
			i2c(I2C_READ, 0, ctrl_addr, tach_reg, 1, &tach_val[i][j]);
			delay_us(10000);
			printf("fan[%d] ctrl=0x%0x tach=0x%x val=0x%0x\n",
					fan, ctrl_addr, tach_reg, tach_val[i][j]);

			tach_reg += FAN_REG_OFFSET;
			fan++;
		}
	}

	// read temperature.
	i2c(I2C_READ, 0, I2C_ADDR_TEMP_SENSOR, 0xAA, 1, &temp_val);
	printf("Temperture = %d C\n", temp_val);
	printf("Fan Debug  = %s \n", cmFanCb.debug ? "ON" : "OFF");
	printf("Fail count = %d \n", cmFanCb.failed_count);
	printf("Clear count= %d \n", cmFanCb.clear_count);

}

void gtsdma_test(void);

int
cmFanCmds( int argc, char *argv[] )
{

	if (argc != 2)
		return -1;

	if (strequ(argv[1], "on"))
	{
		cmFanCb.debug = 1;
		return 0;
	}

	if (strequ(argv[1], "off"))
	{
		cmFanCb.debug = 0;
		return 0;
	}

	if (strequ(argv[1], "status"))
	{
		cmFanShowStatus();
		return 0;
	}

	return -1;
}
#endif
