/*
 *-----------------------------------------------------------------
 *
 * Name:	cm-intr.c
 *
 * RCS:  $Id: cm-intr.c,v 1.14 2002/10/08 03:16:17 johnh Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description:	Interrupt handling routines for BMFPGA
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/21/02
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/cm-api.h>
#include <sys/bmfpga.h>
#include "cm.h"
#if !defined(PROM) && !defined(FCNIM_TM)
#include "exports/event-api.h"
#endif  /* !PROM && !FCNIM_TM */

#ifdef SSC_MGMT
extern volatile bmfpga_master_t		* bmfpga;
#else
extern volatile bmfpga_slave_t		* bmfpga;
#endif

#if defined(NFP_TXRX) || defined(FCNIM) || defined(SSC_MGMT)

#ifndef PROM
extern void cmSendCpuEvent(uint32 slot, uint32 cpu, int state);
#endif  /* !PROM */



#ifdef SSC_MGMT
/*-----------------------------------------------------------------
 * Name :	update_field_fn()
 *
 * Description: 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/21/02
 *
 *-----------------------------------------------------------------
 */
static int
update_field_fn(uint32 *field, uint32 value, int *changed)
{
	if (*field != value) {
		*changed = 1;
		*field = value;
		return 1;
	}
	else {
		return 0;
	}
}


#define update(field, value)  \
if (update_field_fn(&(field), (value), &changed)) \
   { CM_DEBUG(("%s changed\n", #field));}

	
/*-----------------------------------------------------------------
 * Name :	cm_poll_module_present()
 *
 * Description: Poll for changes in bmfpgaIsModulePresent
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/27/02
 *
 *-----------------------------------------------------------------
 */
void
cm_poll_module_present(void *cb, uint32 tref)
{
	int changed = 0;
	int i, j;
	for (i = 1; i < CM_NUM_SLOTS; ++i) {
		cm_slot_t *cm_slot = &cm.slot[i];
		int present = bmfpgaIsModulePresent(i);
		if (!present && cm_slot->present) {
			CM_DEBUG(("slot %d changed to not present\n", i));
			for (j = 0; j < cm_slot->num_cpus; ++j) {
				cm_cpu_t *cpu = &cm_slot->cpu[j];
				CM_DEBUG(("cpu %d state = %d\n", j, cpu->state));
#if !defined(PROM)
				if (cpu->state == CM_SLICE_STATE_RUNTIME) {
					cmSendCpuEvent(i, j, EVENT_STATE_DOWN);
				}
#endif				
				cpu->state = CM_SLICE_STATE_NOT_PRESENT;
			}
			cm_slot->present = 0;
			changed = 1;
		}
	}
	if (changed) {
		cmSendInfoResp(0, 0);
	}
}
#endif


/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/21/02
 *
 *-----------------------------------------------------------------
 */
void
cm_intrHandleStatusChange(
	uint32		sr,
	uint32		cause,
	address_t	epc
	)
{
#ifdef SSC_MGMT
	uint32	int_stat;
	uint32	i, j;
	int changed = 0;

	CM_DEBUG(("cm_intrHandleStatusChange\n"));

	int_stat = bmfpga->module_presence;

	CM_DEBUG(("module presence && interrupt status = 0x%x\n", int_stat));

	for (i=0; i<BMFPGA_NIM_SLOTS; i++) {
		if (int_stat & (BMFPGA_STAT_INT_PENDING_MASK << (i*2))) {
			cm_slot_t *cm_slot = &cm.slot[i+1];
			volatile bmfpga_slot_t *fpga_slot = &bmfpga->nim_slot[i];
			
			CM_DEBUG(("slot %d status reg = 0x%x\n", i+1,
												fpga_slot->status));
			CM_DEBUG(("slot %d status exp reg = 0x%x\n", i+1,
												fpga_slot->status_exp));
			CM_DEBUG(("slot %d power in reg = 0x%x\n", i+1,
												fpga_slot->power_in));

			update(cm_slot->present, 1);
			update(cm_slot->type,
				   (fpga_slot->status &
					BMFPGA_SLOT_STATUS_BOARD_TYPE_MASK) >>
				   BMFPGA_SLOT_STATUS_BOARD_TYPE_SHIFT);
			update(cm_slot->num_cpus, cmGetNumCpus(cm_slot->type));
			
			for (j=0; j<cm_slot->num_cpus; j++) {
                uint32 state;
				cm_cpu_t *cpu = &cm_slot->cpu[j];
                uint32 slice;

                /* for fc j is cpuid, need to map it slice id in fpga access*/
                if ((i+1) == NFX_SLOTTYPE_FC)
                    slice = CM_FC_CPUID2SLICE(j); 
                else
                    slice = j;

				state = (fpga_slot->status &
						 (BMFPGA_SLOT_STATUS_SLICE_STATE_MASK << (slice*4))) >> (slice*4);

#ifndef PROM
				if ((cpu->state != CM_SLICE_STATE_RUNTIME) 
							&& (state == CM_SLICE_STATE_RUNTIME))
				{
					/*
					 * Send CPU up event
					 */
					cmSendCpuEvent(i+1, j, EVENT_STATE_UP);
				}
				else if ((cpu->state == CM_SLICE_STATE_RUNTIME) 
							&& (state != CM_SLICE_STATE_RUNTIME))
				{
					/*
					 * Send CPU down event
					 */
					cmSendCpuEvent(i+1, j, EVENT_STATE_DOWN);
				}
#endif  /* !PROM */

				update(cpu->state, state);
                /* check if the cpu is requesting reset/reboot */
                if (state == CM_SLICE_STATE_RESET_WAIT) {
                    CM_DEBUG(("Reset slot %d cpu %d state=%d\n", i+1, j, state));
                    cmRebootCPU(i+1, j);
                    /* the state will change to REBOOT */
                }
            }
			update(cm_slot->temp,
				   (fpga_slot->status & BMFPGA_SLOT_STATUS_TEMP_MASK) >>
				   BMFPGA_SLOT_STATUS_TEMP_SHIFT);
			update(cm_slot->temp_valid, 1);
			update(cm_slot->power, fpga_slot->power_in);
			update(cm_slot->power_valid, 1);
		}
	}
	if (changed) {
		CM_DEBUG(("cm_intrHandleStatusChange: call cmSendInfoResp\n"));
		cmSendInfoResp(0, 0);
	}
#else
	CM_DEBUG(("cm_intrHandleStatusChange\n"));
#endif
}


/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/21/02
 *
 *-----------------------------------------------------------------
 */
void
cm_intrHandleMessageChange(
	uint32		sr,
	uint32		cause,
	address_t	epc
)
{
#ifdef SSC_MGMT
	uint32	int_stat;
	uint32	i;

	CM_DEBUG(("cm_intrHandleMessageChange\n"));

	int_stat = bmfpga->module_presence;

	CM_DEBUG(("module presence && interrupt status = 0x%x\n", int_stat));

	for (i=0; i<BMFPGA_NIM_SLOTS; i++) {
		if (int_stat & (BMFPGA_MSG_INT_PENDING_MASK << (i*2)))
			CM_DEBUG(("slot %d message reg = 0x%x\n", i+1,
												bmfpga->nim_slot[i].message_in));
	}
#else
	CM_DEBUG(("cm_intrHandleMessageChange\n"));
#endif
}

#endif	/* NFP_TXRX || FCNIM || SSC_MGMT */

#if defined(SSC_MGMT)

#include <sys/tx3927.h>

#define IRSSR_INT_0 0x1
#define IRSSR_INT_1 0x2

/*-----------------------------------------------------------------
 * Name :	cm_poll_bmfpga_intr()
 *
 * Description: bmfpga interrupt polling routine
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/05/02
 *
 *-----------------------------------------------------------------
 */
void
cm_poll_bmfpga_intr(void *cb, uint32 t)
{
	volatile tx3927_intcntrl_t *int_cntrl =
		(tx3927_intcntrl_t *)TX3927_INT_CONTROL_BASE;

	while (int_cntrl->ssr & IRSSR_INT_0) {
		CM_DEBUG(("Status change active\n"));
		cm_intrHandleStatusChange(0, 0, 0);
	}

	while (int_cntrl->ssr & IRSSR_INT_1) {
		CM_DEBUG(("Message change active\n"));
		cm_intrHandleMessageChange(0, 0, 0);
	}
}

#endif

