/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-ed-hdr.c,v 1.1.1.1 2002/04/23 17:47:36 laurentb Exp $
 *
 * Name			: cifs-rpc-ed-hdr.c
 *
 * Description	: Contains CIFS DCE/RPC headers encode & decode functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/08/2002
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#else
#include <exports/cifs-api.h>
#endif

#include "cifs-rpc.x"


/*-----------------------------------------------------------------
 * Name         : cifsrpc_decodeHeader
 *
 * Description  : decode RPC header
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 12/11/2001
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_decodeHeader(cifsrpc_pipe *rpip, uchar8 *datp, uint32 bytesToWriteCnt)
{
	cifsrpc_inPdu *inpp = &rpip->inPdu;
	
	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_decodeHeader"));
	
	if (bytesToWriteCnt < CIFSRPC_S_HEADER)
	{
	   	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_decodeHeader: "
					"received too few data (%d)", bytesToWriteCnt));
		return NFX_ERR;
	}
	
	if ((datp[RPCHO_MAJVERS] != CIFSRPC_MAJORVERSION)
	 || (datp[RPCHO_MINVERS] != CIFSRPC_MINORVERSION))
	{
	   	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_decodeHeader: "
					"version %d-%d not supported", datp[0], datp[1]));
		return NFX_ERR;
	}
	inpp->hdr.packetType= datp[RPCHO_PKTYPE];
	inpp->hdr.flags		= datp[RPCHO_FLAGS];
	inpp->hdr.bigEndian	= (datp[RPCHO_ENDIAN] != 0x10);
	if (inpp->hdr.bigEndian)
	{
		inpp->hdr.fragLen = GETBE2(datp, RPCHO_FRAGLEN);
		inpp->hdr.authLen = GETBE2(datp, RPCHO_AUTHLEN);
		inpp->hdr.callId  = GETBE4(datp, RPCHO_CALLID);
	} else {
		inpp->hdr.fragLen = GETLE2(datp, RPCHO_FRAGLEN);
		inpp->hdr.authLen = GETLE2(datp, RPCHO_AUTHLEN);
		inpp->hdr.callId  = GETLE4(datp, RPCHO_CALLID);
	}
	
	//TODO Check if endian is same as in previous fragments ? 
	//See if it works fine like this
	inpp->bigEndian = inpp->hdr.bigEndian;
	
	if ((inpp->hdr.fragLen < CIFSRPC_S_HEADER)
	 || (inpp->hdr.fragLen > CIFSRPC_S_PDUFRAGMENT))
	{
	   	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_decodeHeader: "
					"fragment length screwed (%d)", inpp->hdr.fragLen));
		return NFX_ERR;
	}
	inpp->fragLen = inpp->hdr.fragLen - CIFSRPC_S_HEADER;

	return NFX_OK;
}


/*-----------------------------------------------------------------
 * Name         : cifsrpc_encodeHeader
 *
 * Description  : encode RPC header
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/09/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_encodeHeader(cifsrpc_pipe *rpip, uchar8 packetType, uchar8 flags,
						uint32 fragLen, uint32 authLen, uint32 callId)
{
	cifsrpc_outPdu *outPp = &rpip->outPdu;
	char *datp = outPp->frag;
	
	PCAR(datp,	RPCHO_MAJVERS,	CIFSRPC_MAJORVERSION);
	PCAR(datp,	RPCHO_MINVERS,	CIFSRPC_MINORVERSION);
	PCAR(datp,	RPCHO_PKTYPE,	packetType);
    PCAR(datp,	RPCHO_FLAGS,	flags);
	PCAR(datp,	RPCHO_ENDIAN,	0x10);
	PCAR(datp,	RPCHO_ENDIAN+1,	0);
	PCAR(datp,	RPCHO_ENDIAN+2,	0);
	PCAR(datp,	RPCHO_ENDIAN+3,	0);
	PUTLE2(datp,RPCHO_FRAGLEN,	CIFSRPC_S_HEADER + fragLen);
	PUTLE2(datp,RPCHO_AUTHLEN,	authLen);
	PUTLE4(datp,RPCHO_CALLID,	callId);

	outPp->fragSentLen = 0;
	outPp->fragLen = fragLen + CIFSRPC_S_HEADER;

	return NFX_OK;
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDinterf()
 *
 * Description  : CIFS RPC interface encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDinterf(cifsrpc_ed *edp, cifsrpc_intf *ifp, char *label, uchar8 lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%s%s interface", cifs_spaces(lv), label));
	lv++;
	if (cifsrpcEDalign   (edp)								   != NFX_OK
	 || cifsrpcEDuint32  (edp, &ifp->uuid.d1,		"uuid d1", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &ifp->uuid.d2,		"uuid d2", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &ifp->uuid.d3,		"uuid d3", lv) != NFX_OK
	 || cifsrpcEDuchar8Ar(edp, ifp->uuid.d4, 2, 0,	"uuid d4", lv) != NFX_OK
	 || cifsrpcEDuchar8Ar(edp, ifp->uuid.d5, 6, 0,	"uuid d5", lv) != NFX_OK
	 || cifsrpcEDuint32  (edp, &ifp->version,		"version", lv) != NFX_OK)
	  return (NFX_ERR);

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDbindRequest()
 *
 * Description  : CIFS RPC Bind Request encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDbindRequest(cifsrpc_ed *edp, cifsrpc_bindReq *bp, uchar8 lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%scifsrpcEDbindRequest", cifs_spaces(lv)));
	lv++;
	
	if (cifsrpcEDushort16(edp, &bp->maxTFragSize, "maxTFragS ", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &bp->maxRFragsize, "maxRFragS ", lv) != NFX_OK
	 ||	cifsrpcEDuint32  (edp, &bp->assocGroupId, "assocGrpId", lv) != NFX_OK
	 || cifsrpcEDuint32  (edp, &bp->nbCtxElts,	  "nbCtxElts ", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &bp->contextId,    "contextId ", lv) != NFX_OK
	 || cifsrpcEDuchar8  (edp, &bp->nbSyntaxes,	  "nbSyntaxes", lv) != NFX_OK
	 || cifsrpcEDinterf  (edp, &bp->abstract,     "abstract",   lv) != NFX_OK
	 || cifsrpcEDinterf  (edp, &bp->transfer,     "transfer",   lv) != NFX_OK)
		return (NFX_ERR);
		
	return (NFX_OK);		
					
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDbindAck()
 *
 * Description  : CIFS RPC Bind Ack encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/09/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDbindAck(cifsrpc_ed *edp, cifsrpc_bindAck *bp, uchar8 lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%scifsrpcEDbindAck", cifs_spaces(lv)));
	lv++;
	
	if (cifsrpcEDushort16(edp, &bp->maxTFragSize, "maxTFragS ", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &bp->maxRFragsize, "maxRFragS ", lv) != NFX_OK
	 ||	cifsrpcEDuint32  (edp, &bp->assocGroupId, "assocGrpId", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &bp->secAddrlen,	  "secAddlen ", lv) != NFX_OK
	 || cifsrpcEDuchar8Ar(edp, bp->secAddrPort, bp->secAddrlen, 1,
										  		  "secAddPort", lv)	!= NFX_OK
	 || cifsrpcEDalign   (edp) 										!= NFX_OK
	 || cifsrpcEDuchar8  (edp, &bp->nbResults,	  "nbResults ", lv) != NFX_OK
	 || cifsrpcEDalign   (edp)										!= NFX_OK
	 || cifsrpcEDushort16(edp, &bp->result,		  "result    ", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &bp->reason,		  "reason    ", lv) != NFX_OK
	 || cifsrpcEDinterf  (edp, &bp->transfer,	  "transfer",   lv) != NFX_OK)
		return (NFX_ERR);
		
	return (NFX_OK);		
}					


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDrequestHdr()
 *
 * Description  : CIFS RPC request header encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/10/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDrequestHdr(cifsrpc_ed *edp, cifsrpc_reqHdr *rp, uchar8 lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%scifsrpcEDrequestHdr", cifs_spaces(lv)));
	lv++;
	
	if (cifsrpcEDuint32  (edp, &rp->allocHint, "allocHint", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &rp->contextId, "contextId", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &rp->opNb,	   "opNumber ", lv) != NFX_OK)
		return (NFX_ERR);
		
	return (NFX_OK);		
}					



/*-----------------------------------------------------------------
 * Name         : cifsrpcEDresponseHdr()
 *
 * Description  : CIFS RPC response header encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/10/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDresponseHdr(cifsrpc_ed *edp, cifsrpc_respHdr *rp, uchar8 lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%scifsrpcEDresponseHdr", cifs_spaces(lv)));
	lv++;
	
	if (cifsrpcEDuint32  (edp, &rp->allocHint, "allocHint", lv) != NFX_OK
	 || cifsrpcEDushort16(edp, &rp->contextId, "contextId", lv) != NFX_OK
	 || cifsrpcEDuchar8  (edp, &rp->cancelCnt, "cancelCnt", lv) != NFX_OK
	 || cifsrpcEDuchar8  (edp, &rp->reserved,  "reserved ", lv) != NFX_OK)
		return (NFX_ERR);
		
	return (NFX_OK);		
}					


/*-----------------------------------------------------------------
 * Name         : cifsrpc_encodeResponseHeader()
 *
 * Description  : CIFS RPC response header encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/10/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_encodeResponseHeader(cifsrpc_pipe *rpip, uint32 allocHint)
{
	cifsrpc_ed ed;
	cifsrpc_respHdr respHdr;

	cifsrpcEDinit(&ed, TRUE, 0, &rpip->memList);
	cifsrpcEDassignData(&ed,
						rpip->outPdu.frag + CIFSRPC_S_HEADER,
						CIFSRPC_S_RESP_HDR,
						FALSE);
	memset(&respHdr, 0, sizeof(cifsrpc_respHdr));
	respHdr.allocHint = allocHint;
	return cifsrpcEDresponseHdr(&ed, &respHdr, 0);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDfault()
 *
 * Description  : CIFS RPC fault pdu encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/11/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDfault(cifsrpc_ed *edp, cifsrpc_faultHdr *fp, uchar8 lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%scifsrpcEDfault", cifs_spaces(lv)));
	
	if (cifsrpcEDresponseHdr(edp, &fp->respHdr, lv++)			  != NFX_OK
	 || cifsrpcEDuint32		(edp, &fp->status,	 "status   ", lv) != NFX_OK
	 || cifsrpcEDuint32		(edp, &fp->reserved, "reserved ", lv) != NFX_OK)
		return (NFX_ERR);
		
	return (NFX_OK);		
}



/*-----------------------------------------------------------------
 * Name         : cifsrpcED()
 *
 * Description  : CIFS RPC encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01//2002
 *
 *-----------------------------------------------------------------
 */
//int32	//NFX_OK or NFX_ERR
//cifsrpcED(cifsrpc_ed *edp, 


