/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-ed-sec.c,v 1.3 2002/12/10 18:22:27 laurentb Exp $
 * 
 * Name			: cifs-rpc-ed-sec.c
 *
 * Description	: Contains CIFS DCE/RPC security encode & decode functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/22/2002
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#else
#include <exports/cifs-api.h>
#endif

#include "cifs-rpc.x"


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDsid()
 *
 * Description  : security identifier (SID) encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDsid(cifsrpc_ed *edp, sec_cifsSid *sidp, char *label, uchar8 lv)
{
    CIFS_PRINT(CIFS_DBGLV_DBG3, ("%s%s (sid)", cifs_spaces(lv), label));
	lv++;
	
	if (sidp->nbSubAuthy > CIFSEC_SID_MAX_SUB_AUTHORITIES)
	{
	    CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpcEDsid: "
					"too many subAuthorities : %u in %s sid!)",
					sidp->nbSubAuthy, label));
		return (NFX_ERR);
	}
	
	if (cifsrpcEDalign   (edp)										  != NFX_OK
     || cifsrpcEDuchar8  (edp, &sidp->revision,		"revision  ", lv) != NFX_OK
     || cifsrpcEDuchar8  (edp, &sidp->nbSubAuthy,	"nbSubAuthy", lv) != NFX_OK
     || cifsrpcEDuchar8Ar(edp, sidp->authority.val, 6, 0,
			 										"authority ", lv) != NFX_OK
     || cifsrpcEDuint32Ar(edp, sidp->subAuthy, sidp->nbSubAuthy, 
			 										"subAuthy  ", lv) != NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDrpcSid()
 *
 * Description  : security identifier (SID) encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 04/20/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDrpcSid(cifsrpc_ed *edp, cifsrpc_sid *sidp, char *label, uchar8 lv)
{
    CIFS_PRINT(CIFS_DBGLV_DBG3, ("%srpc sid", cifs_spaces(lv)));
	lv++;
	
	if (cifsrpcEDalign (edp)										!= NFX_OK
     || cifsrpcEDuint32(edp, &sidp->nbSubAuthy, "nbSubAuthy", 	lv)	!= NFX_OK
     || cifsrpcEDsid   (edp, &sidp->sid,		label,			lv) != NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDace()
 *
 * Description  : security access control entry encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDace(cifsrpc_ed *edp, sec_cifsAce *acep, uchar8 index, uchar8 lv)
{
	uint32 i, aceOff, off;
	
    CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%saccess control entry %u", cifs_spaces(lv), index));
	lv++;
	
	if (cifsrpcEDalign(edp)	!= NFX_OK)
		return (NFX_ERR);

	//Remember where the ACE starts to calculate its size
	aceOff = edp->offset;
	
	if (cifsrpcEDuchar8  (edp, &acep->type,		 "type      ", lv)	!= NFX_OK
	 || cifsrpcEDuchar8  (edp, &acep->flags,	 "flags     ", lv)	!= NFX_OK
	 || cifsrpcEDushort16(edp, &acep->size,		 "size      ", lv)	!= NFX_OK
     || cifsrpcEDuint32  (edp, &acep->accessMask,"accessMask", lv)	!= NFX_OK
	 || cifsrpcEDsid	 (edp, &acep->sid,	 	 "user/group", lv)	!= NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_ENCODING(edp))
	{
		//Calculate the size of the ACE
		acep->size = edp->offset - aceOff;
		//Remember the actual offset
		off = edp->offset;
		//Set the offset to set the ACE size
		edp->offset = aceOff + 2 * sizeof(uchar8);  //type & flags fields
		//Set ACE size
		cifsrpcEDushort16(edp, &acep->size,		"realsize  ", lv);
		//Restore the offset to its position
		edp->offset = off;
	}
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDacl()
 *
 * Description  : security access control list encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDacl(cifsrpc_ed *edp, sec_cifsAcl *aclp, char *label, uchar8 lv)
{
	uint32 i, aclOff, off;
	
    CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s access control list", cifs_spaces(lv), label));
	lv++;
	
	if (cifsrpcEDalign(edp)	!= NFX_OK)
		return (NFX_ERR);

	//Remember where the ACL starts to calculate its size
	aclOff = edp->offset;
	
	if (cifsrpcEDuchar8  (edp, &aclp->revision,	"revision", lv)	!= NFX_OK
	 || cifsrpcEDuchar8  (edp, &aclp->sbz1,		"sbz1    ", lv)	!= NFX_OK
	 || cifsrpcEDushort16(edp, &aclp->size,		"size    ", lv)	!= NFX_OK
     || cifsrpcEDuint32  (edp, &aclp->nbAces,	"nb Aces ", lv)	!= NFX_OK)
		return (NFX_ERR);

	if (CIFSRPC_DECODING(edp))
	{
		aclp->acep = (sec_cifsAce*)cifsrpc_allocMemItem(edp->memLp, 
										sizeof(sec_cifsAce)*(aclp->nbAces+1));
		if (aclp->acep == NULL)
			return (NFX_ERR);
	}
	for (i=0; i<aclp->nbAces; i++)
		if (cifsrpcEDace(edp, aclp->acep + i, i, lv) != NFX_OK)
			return (NFX_ERR);
	
	if (cifsrpcEDalign(edp) != NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_ENCODING(edp))
	{
		//Calculate the size of the ACL
		aclp->size = edp->offset - aclOff;
		//Remember the actual offset
		off = edp->offset;
		//Set the offset to set the ACL size
		edp->offset = aclOff + sizeof(ushort16);  //revision field
		//Set ACL size
		cifsrpcEDushort16(edp, &aclp->size,		"realsize", lv);
		//Restore the offset to its position
		edp->offset = off;
	}
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDsecDesc()
 *
 * Description  : Security descriptor encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDsecDesc(cifsrpc_ed *edp, sec_cifsSd *sdp, char *label, uchar8 lv)
{
	uint32 off, maxOff;
	
    CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s security descriptor", cifs_spaces(lv), label));
	lv++;
	
	if (cifsrpcEDalign(edp) != NFX_OK)
		return (NFX_ERR);
	
	//Remember the original offset for offsets calculation
	off = edp->offset;
	
	if (cifsrpcEDuchar8  (edp, &sdp->revision,	 "revision   ", lv)	!= NFX_OK
	 || cifsrpcEDuchar8  (edp, &sdp->sbz1,		 "sbz1       ", lv)	!= NFX_OK
	 || cifsrpcEDushort16(edp, &sdp->control,	 "control    ", lv)	!= NFX_OK
     || cifsrpcEDuint32  (edp, &sdp->ownerSidOff,"ownerSidOff", lv)	!= NFX_OK
     || cifsrpcEDuint32  (edp, &sdp->groupSidOff,"groupSidOff", lv)	!= NFX_OK
     || cifsrpcEDuint32  (edp, &sdp->sAclOff,	 "sAclOff    ", lv)	!= NFX_OK
     || cifsrpcEDuint32  (edp, &sdp->dAclOff,	 "dAclOff    ", lv)	!= NFX_OK)
		return (NFX_ERR);
	
	//Initialize maxOff to this offset
	maxOff = edp->offset;

	if (sdp->ownerSidOff != 0)
	{
		if (cifsrpcEDsetOffset(edp, off + sdp->ownerSidOff)		!= NFX_OK
		 || cifsrpcEDsid(edp, &sdp->ownerSid, "owner sid", lv)	!= NFX_OK
		 || cifsrpcEDalign(edp)									!= NFX_OK)
			return (NFX_ERR);

		maxOff = MAX(maxOff, edp->offset);
	}

	if (sdp->groupSidOff != 0)
	{
		if(cifsrpcEDsetOffset(edp, off + sdp->groupSidOff)		!= NFX_OK
		 || cifsrpcEDsid(edp, &sdp->groupSid, "group sid", lv)	!= NFX_OK
		 || cifsrpcEDalign(edp)									!= NFX_OK)
			return (NFX_ERR);
		
		maxOff = MAX(maxOff, edp->offset);
	}

	if ((sdp->control & CIFSEC_SD_SACL_PRESENT) && sdp->sAclOff != 0)
	{
		if (cifsrpcEDsetOffset(edp, off + sdp->sAclOff)	!= NFX_OK
		 || cifsrpcEDacl(edp, &sdp->sAcl, "system", lv)	!= NFX_OK)
			return (NFX_ERR);

		maxOff = MAX(maxOff, edp->offset);
	}

	if ((sdp->control & CIFSEC_SD_DACL_PRESENT) && sdp->dAclOff != 0)
	{
		if (cifsrpcEDsetOffset(edp, off + sdp->dAclOff)	!= NFX_OK
		 || cifsrpcEDacl(edp, &sdp->dAcl, "discretionary", lv)	!= NFX_OK)
			return (NFX_ERR);

		maxOff = MAX(maxOff, edp->offset);
	}

	//Store SID size
	sdp->size = maxOff - off;

	if (cifsrpcEDsetOffset(edp, maxOff) != NFX_OK)
		return (NFX_ERR);

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDsdBuf()
 *
 * Description  : Security descriptor buffer structure encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDsdBuf(cifsrpc_ed *edp, sec_cifsSdBuf *sdbp, char *label, uchar8 lv)
{

    CIFS_PRINT(CIFS_DBGLV_DBG3, 
				("%s%s security descriptor buffer", cifs_spaces(lv), label));
	lv++;
	
	if (cifsrpcEDalign(edp) != NFX_OK)
		return (NFX_ERR);
	
	if (CIFSRPC_DECODING(edp) || sdbp->sdp == 0)
	{
		if (cifsrpcEDuint32(edp, &sdbp->maxLen, "maxLen", lv)	!= NFX_OK
	     || cifsrpcEDuint32(edp, &sdbp->sdp,	"sdp   ", lv)	!= NFX_OK
	     || cifsrpcEDuint32(edp, &sdbp->len,	"length", lv)	!= NFX_OK
		 || (sdbp->sdp != 0
		  && cifsrpcEDsecDesc(edp, &sdbp->sd.sd, label, lv) != NFX_OK))
	        return (NFX_ERR);
	} else {	//Encoding a non empty security descriptor
		//Postpone the encoding of maxLen and length as we don't know them
		uint32 off1, off2;
		off1 = edp->offset;
		if (cifsrpcEDsetOffset(edp, off1 + 3 * sizeof(uint32))	!= NFX_OK)
			return (NFX_ERR);
		off2 = edp->offset;
		if (cifsrpcEDsecDesc(edp, sdbp->sd.sdp, label, lv) != NFX_OK)
            return (NFX_ERR);
		//Calculate the security descriptor size
		sdbp->len = edp->offset - off2;
		if (sdbp->len != 0)
			sdbp->maxLen = sdbp->len;
		//Remember where the offset is
		off2 = edp->offset;
		//Set the offset back to maxLen, to encode maxLen, unknown and length
		edp->offset = off1;
		//We know there is enough memory, don't have to check
		cifsrpcEDuint32(edp, &sdbp->maxLen, "maxLen", lv);
	    cifsrpcEDuint32(edp, &sdbp->sdp,	"sdp   ", lv);
	    cifsrpcEDuint32(edp, &sdbp->len,	"length", lv);
		//Restore the offset
		edp->offset = off2;
	}
	
	return (NFX_OK);
}


