/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-encdec.h,v 1.1.1.1 2002/04/23 17:47:36 laurentb Exp $
 * 
 * Name			: cifs-rpc-encdec.h
 *
 * Description	: Contains CIFS DCE/RPC encode & decode definitions.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/04/2002
 *
 *-----------------------------------------------------------------
 */
#ifndef _CIFS_RPC_ENCDEC_H
#define _CIFS_RPC_ENCDEC_H


#define		CIFSRPC_ALIGNT	4


//Encoding / decoding macros
#define		CIFSRPC_ENCODING(edp)	((edp)->enc)
#define		CIFSRPC_DECODING(edp)	(!((edp)->enc))

#define		CIFSRPC_ED_SETENDIAN(edp, endian)	((edp)->bigEndian = endian)
#define		CIFSRPC_ED_SETOENCODING(edp)		((edp)->enc = TRUE)
#define		CIFSRPC_ED_SETODECODING(edp)		((edp)->enc = FALSE)

// Linked definitions pre-definitions
struct cifsrpc_uniStruc;
struct cifsrpc_enumHandle;


typedef struct cifsrpc_ed	//CIFS RPC Encoding Decoding
{
  uchar8	*bufp;			//buffer containing data being processed
  uint32	size;			//buffer size
  uint32	offset;			//current data offset in the buffer
  boolean	bigEndian;		//TRUE if decoding has to be done in big endian
  boolean	enc;			//TRUE is encoding, FALSE is decoding
  boolean	freeMem;		//TRUE if buffer memory has to be freed
  uchar8	align;			//alignement: 1, 2, 4
  cifsrpc_memList *memLp;	//list of the memory items used by the processing
	
}cifsrpc_ed;




/*-----------------------------------------------------------------
 * Name         : cifsrpcEDinit()
 *
 * Description  : CIFS RPC encode / decode init
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/04/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDinit(cifsrpc_ed *edp, 
				boolean encoding,
				uint32 size,
				cifsrpc_memList *mlp);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDassignData()
 *
 * Description  : Assign data to an RPC encode / decode structure
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
void cifsrpcEDassignData(	cifsrpc_ed	*edp, 
							uchar8		*datp,
							uint32		size,
							boolean		freeMem);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDreleaseData()
 *
 * Description  : Release buffer data
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
void cifsrpcEDreleaseData(cifsrpc_ed *edp);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDcheckMemAvailability()
 *
 * Description  : CIFS RPC check whether memory is available to encode
 * 					or decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDcheckMemAvailability(cifsrpc_ed *edp, uint32 size);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDalign()
 *
 * Description  : move the data offset to the next alignment boundary
 * 				  if possible
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDalign(cifsrpc_ed *edp);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDaddData()
 *
 * Description  : Add data to an encode / decode structure
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/10/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDaddData(cifsrpc_ed *edp, char *datp, uint32 size);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDsetOffset()
 *
 * Description  : set the offset of an encode / decode structure
 * 				  check if possible
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDsetOffset(cifsrpc_ed *edp, uint32 offset);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuchar8()
 *
 * Description  : CIFS RPC uchar8 encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuchar8(cifsrpc_ed *edp, uchar8 *uchar8p, char *label, uchar8 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuchar8Ar()
 *
 * Description  : CIFS RPC uchar8 array encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuchar8Ar(cifsrpc_ed *edp, uchar8 *uchar8p, uint32 uchar8n, 
					uint32 string, char *label, uchar8 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDushort16()
 *
 * Description  : CIFS RPC ushort16 encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDushort16(cifsrpc_ed *edp, ushort16 *ushort16p, char*label, uchar8 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDushort16Ar()
 *
 * Description  : CIFS RPC ushort16 array encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDushort16Ar(cifsrpc_ed *edp, ushort16 *ushort16p, uint32 ushort16n,
					uint32 string, char *label, ushort16 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuint32()
 *
 * Description  : CIFS RPC uint32 encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuint32(cifsrpc_ed *edp, uint32 *uint32p, char *label, uchar8 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDuint32Ar()
 *
 * Description  : CIFS RPC uint32 array encoding / decoding
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/22/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuint32Ar(cifsrpc_ed *edp, uint32 *uint32p, uint32 uint32n,	
					char *label, uint32 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDunistruc()
 *
 * Description  : CIFS RPC unicode structure encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/14/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDuniStruc(	cifsrpc_ed *edp, 
					struct cifsrpc_uniStruc *sp,
					char *label,
					uchar8 lv);


/*-----------------------------------------------------------------
 * Name         : cifsrpcEDenumHd()
 *
 * Description  : CIFS RPC enum handle encode / decode 
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/15/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpcEDenumHd(cifsrpc_ed *edp, 
				struct cifsrpc_enumHandle *ehp,
				char *label,
				uchar8 lv);




#endif // #ifndef _CIFS_RPC_ENCDEC_H 

