/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-srvsvc.c,v 1.2 2002/12/10 18:22:27 laurentb Exp $
 *
 * Name			: cifs-rpc-srvsvc.c
 *
 * Description	: Contains CIFS DCE/RPC server services api functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#include "../sm-cifs/cifs-unicode.h"
#include "../sm-cifs/cifs-utils.h"
#include "../sm-cifs/cifs-err.h"
#else
#include <exports/cifs-api.h>
#include <exports/cifs-utils.h>
#include <exports/cifs-unicode.h>
#include <exports/cifs-err.h>
#endif

#include "cifs-rpc-utils.h"
#include "cifs-rpc.x"
#include "cifs-rpc-utils.h"
#include "cifs-rpc-srvsvc.h"
#include "cifs-rpc-ed-srvsvc.h"


//Unicode strings
uni16 unistr_cColSlash[] = { 'C', ':', '\\' };



/*-----------------------------------------------------------------
 * Name			: isHiddenShare
 *
 * Description	: Return TRUE if the associated share is hidden
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
static boolean
isHiddenShare(cifs_share *shp)
{
	return (shp->cfg.name[unistrlen(shp->cfg.name)-1] == '$');
}


/*-----------------------------------------------------------------
 * Name			: getShareType
 *
 * Description	: Return TRUE if the associated share is hidden
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//share type
getShareType(cifs_share *shp)
{
	uint32 type;
	if (shp->cfg.flags & CIFS_SHCFG_ISADMIN) 
		type = CIFS_SHARE_TYPE_IPC;
	else
		type = CIFS_SHARE_TYPE_DISK;
	if (isHiddenShare(shp))
		type |= CIFS_SHARE_HIDDEN;
	return type;
}

/*-----------------------------------------------------------------
 * Name			: fillShareInfoLv1
 *
 * Description	: Fill the share information level 1 structure
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
static void
fillShareInfoLv1(cifsrpc_pipe *rpip, cifs_share *shp, shareInfoLv1 *i1p)
{
	i1p->netNamep = RPC_DUMMY_PTR;
	i1p->remarkp  = RPC_DUMMY_PTR;	
	i1p->type 	  = getShareType(shp);
	fillUniStrucWithoutAlloc(&i1p->netName, shp->cfg.name);
	fillUniStrucWithoutAlloc(&i1p->remark, shp->cfg.comment);
}


/*-----------------------------------------------------------------
 * Name			: fillShareInfoLv2
 *
 * Description	: Fill the share information level 2 structure
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static void
fillShareInfoLv2(cifsrpc_pipe *rpip, cifs_share *shp, shareInfoLv2 *i2p)
{
	i2p->netNamep	= RPC_DUMMY_PTR;
	i2p->remarkp  	= RPC_DUMMY_PTR;	
	i2p->pathp		= RPC_DUMMY_PTR;	
	i2p->passwordp	= RPC_DUMMY_PTR;	
	i2p->type		= getShareType(shp);
	i2p->perms		= 0;
	i2p->nbUsesMax	= 0xFFFFFFFF;
	i2p->nbUses		= shp->useCnt;
	fillUniStrucWithoutAlloc(&i2p->netName, shp->cfg.name);
	fillUniStrucWithoutAlloc(&i2p->remark, shp->cfg.comment);
	fillUniStrucWithoutAlloc(&i2p->path, shp->cfg.pathName);
	fillUniStrucWithoutAlloc(&i2p->password, (uni16*)"\0");
}


/*-----------------------------------------------------------------
 * Name			: fillShareInfoLv502
 *
 * Description	: Fill the share information level 502 structure
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static void
fillShareInfoLv502(cifsrpc_pipe *rpip, cifs_share *shp, shareInfoLv502 *i502p)
{
	i502p->netNamep		= RPC_DUMMY_PTR;
	i502p->remarkp  	= RPC_DUMMY_PTR;	
	i502p->pathp		= RPC_DUMMY_PTR;	
	i502p->passwordp	= RPC_DUMMY_PTR;	
	i502p->type			= getShareType(shp);
	i502p->perms		= 0;
	i502p->nbUsesMax	= 0xFFFFFFFF;
	i502p->nbUses		= shp->useCnt;
	fillUniStrucWithoutAlloc(&i502p->netName, shp->cfg.name);
	fillUniStrucWithoutAlloc(&i502p->remark, shp->cfg.comment);
	fillUniStrucWithoutAlloc(&i502p->path, shp->cfg.pathName);
	fillUniStrucWithoutAlloc(&i502p->password, (uni16*)"\0");
	if (shp->cfg.secDescp == NULL)
	{
		i502p->secDescSize	= 0;
		i502p->secDescp		= 0;
		i502p->sdSize2		= 0;
	} else {
		i502p->secDescSize	= shp->cfg.secDescp->size;
		i502p->secDescp		= RPC_DUMMY_PTR;
		i502p->sdSize2		= shp->cfg.secDescp->size;
		i502p->secDesc.sdp	= shp->cfg.secDescp;
	}
}


/*-----------------------------------------------------------------
 * Name			: fillShareEnumResp
 *
 * Description	: Fill the share enum response structure
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/15/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
fillShareEnumResp(cifsrpc_pipe		*rpip,
				  uint32			infoLevel,
				  boolean			showHidden,
				  netShareEnumResp	*shrp,
				  uint32			*resumeHandlep)
{
	cifs_server *svp = rpip->clp->serverp;	
	cifs_share  *shp;
	uint32 n, i;
		
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("fillShareEnumResp"));
		
	*resumeHandlep = 0;
	
	shrp->shiContr.level = shrp->shiContr.switchValue = infoLevel;
	
	//Count the shares
	if (showHidden)
		n = svp->shareCnt;
	else 
	{
		n = 0;
		DLL_ITERATE(shp, &svp->shares, entry)
		{
			if (!isHiddenShare(shp))
				n++;
		}
	}

	shrp->shiContr.shareInfop = shrp->shiContr.entriesp = RPC_DUMMY_PTR;
	shrp->totalEntries = n;
	shrp->shiContr.nbEntries = shrp->shiContr.entriesRead = n;

	if (n == 0)
		return (NFX_OK);

	i = 0;
	switch (infoLevel) {
	case 1:
	{
		shareInfoLv1 *i1p;
		i1p = (shareInfoLv1*)cifsrpc_allocMemItem(&rpip->memList,
												  n * sizeof(shareInfoLv1));
		if (i1p == NULL)
			return (NFX_ERR);
		DLL_ITERATE(shp, &svp->shares, entry)
		{
			if (showHidden || !isHiddenShare(shp))
			{
				fillShareInfoLv1(rpip, shp, &i1p[i++]);
				if (i == n) break;
			}
		}
		shrp->shiContr.info.lv1p = i1p;
		break;
	}
	
	case 2:
	{
		shareInfoLv2 *i2p;
		i2p = (shareInfoLv2*)cifsrpc_allocMemItem(&rpip->memList,
												  n * sizeof(shareInfoLv2));
		if (i2p == NULL)
			return (NFX_ERR);
		DLL_ITERATE(shp, &svp->shares, entry)
		{
			if (showHidden || !isHiddenShare(shp))
			{
				fillShareInfoLv2(rpip, shp, &i2p[i++]);
				if (i == n) break;
			}
		}
		shrp->shiContr.info.lv2p = i2p;
		break;
	}
	
	case 502:
	{
		shareInfoLv502 *i502p;
		i502p = (shareInfoLv502*)cifsrpc_allocMemItem(&rpip->memList,
												  n * sizeof(shareInfoLv502));
		if (i502p == NULL)
			return (NFX_ERR);
		DLL_ITERATE(shp, &svp->shares, entry)
		{
			if (showHidden || !isHiddenShare(shp))
			{
				fillShareInfoLv502(rpip, shp, &i502p[i++]);
				if (i == n) break;
			}
		}
		shrp->shiContr.info.lv502p = i502p;
		break;
	}

	default:		
		return (NFX_ERR);
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name			: netShareEnum
 *
 * Description	: Enumerate the shares of the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/15/2002
 *
 *-----------------------------------------------------------------
 */
static void	
netShareEnum(cifsrpc_pipe *rpip, netShareEnumCall *shcp, netShareEnumResp *shrp)
{
	uint32 resumeHandle;	
		
	resumeHandle = GETENUMHANDLE(&shcp->resumeHandle);

	if (fillShareEnumResp(	rpip, 
							shcp->shiContr.level,
							shcp->showHidden,
							shrp,
							&resumeHandle) == NFX_OK)
		shrp->returnValue = CIFSTATUS_OK;
	else
		shrp->returnValue = CIFSTATUS_INVALID_INFO_CLASS;
	
	SETENUMHANDLE(&shrp->resumeHandle, resumeHandle);
}	



/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareEnumAndShareEnumSticky
 *
 * Description	: SRVSVC Enumerate the shares of the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareEnumAndShareEnumSticky(cifsrpc_pipe *rpip, boolean showHidden)
{
	netShareEnumCall shEnumCall;
	netShareEnumResp shEnumResp;

	memset(&shEnumCall, 0, sizeof(netShareEnumCall));
	memset(&shEnumResp, 0, sizeof(netShareEnumResp));

	//Decode netShareEnum function call parameters
	if (cifsrpcEDnetShareEnumCall(&rpip->inPdu.dec, &shEnumCall, 1) != NFX_OK)
		return (NFX_ERR);

	//Call netShareEnum
	shEnumCall.showHidden = showHidden;
	netShareEnum(rpip, &shEnumCall, &shEnumResp);
	
	//Encode netShareEnum function response parameters
	if (cifsrpcEDnetShareEnumResp(&rpip->outPdu.enc, &shEnumResp, 1) != NFX_OK)
		return (NFX_ERR);

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareEnum
 *
 * Description	: SRVSVC Enumerate all the shares of the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareEnum(cifsrpc_pipe *rpip)
{
	return srvsvc_netShareEnumAndShareEnumSticky(rpip, TRUE);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareEnumSticky
 *
 * Description	: SRVSVC Enumerate the normal shares of the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareEnumSticky(cifsrpc_pipe *rpip)
{
	return srvsvc_netShareEnumAndShareEnumSticky(rpip, FALSE);
}


/*-----------------------------------------------------------------
 * Name			: netShareGetInfo
 *
 * Description	: get information on a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
netShareGetInfo(	cifsrpc_pipe *rpip, 
					netShareGetInfoCall *shcp,
					netShareGetInfoResp *shrp)
{
	shareInfo *sip = &shrp->shareInfo;
	cifs_share *shp;	
	
	//Look for the share in the server's configuration
	shp = cifs_findServerShareByName(rpip->clp->serverp, shcp->shareName.str);
	if (shp == NULL)
		return (CIFSTATUS_BAD_NETWORK_NAME);
	
	switch(shcp->level) {
		case 1:
			fillShareInfoLv1(rpip, shp, &sip->info.lv1);
			break;
		case 2:
			fillShareInfoLv2(rpip, shp, &sip->info.lv2);
			break;
		case 502:
			fillShareInfoLv502(rpip, shp, &sip->info.lv502);
			break;
		default:
			return (CIFSTATUS_INVALID_INFO_CLASS);
	}			
	
	sip->switchValue = shcp->level;
	sip->shareInfop	 = RPC_DUMMY_PTR;
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareGetInfo
 *
 * Description	: SRVSVC get information on a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareGetInfo(cifsrpc_pipe *rpip)
{
	netShareGetInfoCall shGetInfoCall;
	netShareGetInfoResp shGetInfoResp;

	memset(&shGetInfoCall, 0, sizeof(netShareGetInfoCall));
	memset(&shGetInfoResp, 0, sizeof(netShareGetInfoResp));
	
	//Decode netShareGetInfo function call parameters
	if (cifsrpcEDnetShareGetInfoCall(&rpip->inPdu.dec, &shGetInfoCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call netShareGetInfo
	shGetInfoResp.returnValue = netShareGetInfo(rpip,
												&shGetInfoCall,
												&shGetInfoResp);
	
	//Encode netShareGetInfo function response parameters
	if (cifsrpcEDnetShareGetInfoResp(&rpip->outPdu.enc, &shGetInfoResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: netShareSetInfo
 *
 * Description	: set information on a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
netShareSetInfo(	cifsrpc_pipe *rpip, 
					netShareSetInfoCall *shcp,
					netShareSetInfoResp *shrp)
{
	shareInfo *ship = &shcp->shareInfo;
	cifs_share *shp;	
	boolean commentPresent, PathPresent, shareCfgHasChanged;
	static uni16 comment[CIFS_S_SHARECOMMENT+1];
	static uni16 path[CIFS_S_PATHSTR+1];
	sec_cifsSd *sdp = NULL;
	
	//Check if share info is present
	if (!shcp->shareInfo.shareInfop)
		return (CIFSTATUS_INVALID_PARAMETER);
	
	//Look for the share in the server's configuration
	shp = cifs_findServerShareByName(rpip->clp->serverp, shcp->shareName.str);
	if (shp == NULL)
		return (CIFSTATUS_BAD_NETWORK_NAME);

	if (shp->cfg.flags & CIFS_SHCFG_ISADMIN)
		return (CIFSTATUS_ACCESS_DENIED);

	//TODO check user access rights					
	
	commentPresent = PathPresent = shareCfgHasChanged = FALSE;
	switch(shcp->level) {
		case 1:
			strncpyFromUniStruc(comment, 
								&ship->info.lv1.remark,
								CIFS_S_SHARECOMMENT);
			commentPresent = TRUE;
			break;
		case 2:
			strncpyFromUniStruc(comment,
								&ship->info.lv2.remark,
								CIFS_S_SHARECOMMENT);
			commentPresent = TRUE;
			strncpyFromUniStruc(path,
								&ship->info.lv2.path,
								CIFS_S_PATHSTR);
			PathPresent = TRUE;
			break;
		case 502:
			strncpyFromUniStruc(comment,
								&ship->info.lv502.remark,
								CIFS_S_SHARECOMMENT);
			commentPresent = TRUE;
			strncpyFromUniStruc(path,
								&ship->info.lv502.path,
								CIFS_S_PATHSTR);
			PathPresent = TRUE;
			if (ship->info.lv502.secDescp)
				sdp = &ship->info.lv502.secDesc.sd;
			break;
		case 1005:
			return (CIFSTATUS_ACCESS_DENIED);
		case 1501:
			if (ship->info.lv1501.sdp)
				sdp = &ship->info.lv1501.sd.sd;
			break;
		default:
			return (CIFSTATUS_INVALID_INFO_CLASS);
	}			
	
	if (commentPresent && unistrcmp(shp->cfg.comment, comment))
	{
		unistrcpy(shp->cfg.comment, comment);
		shareCfgHasChanged = TRUE;
	}
	if (PathPresent && unistrcmp(shp->cfg.pathName, path))
	{
		//TODO convert the path according to our internal volume/path scheme
		unistrcpy(shp->cfg.pathName, path);
		shareCfgHasChanged = TRUE;
	}
	
	if (sdp != NULL)
	{
		if (shp->cfg.secDescp == NULL 
		 || cifs_secSecDescCmp(shp->cfg.secDescp, sdp) != 0)
		{	//New security descriptor for the share
			sec_cifsSd *newSdp;
				
			//Convert generic access bits in ACEs to their specific mapping
			cifs_secSDMapGenericAccessBits(sdp, &cifs_fileAccessGenericMap);
			
			//Allocate a new security descriptor and duplicate the temporary one
			newSdp = cifs_secSecDescClone(sdp);
			if (newSdp == NULL)
				return (CIFSTATUS_NO_MEMORY);
			//Free existing security descriptor if any
			if (shp->cfg.secDescp != NULL)
				cifs_secFreeSecDesc(shp->cfg.secDescp, TRUE);
			
			//Assign the new descriptor to the share
			shp->cfg.secDescp = newSdp;

			shareCfgHasChanged = TRUE;
		}
				
					
	}
	
	//TODO update configuration on the SSC
	//if (shareCfgHasChanged)
	//{
	//}
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareSetInfo
 *
 * Description	: SRVSVC set information of a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareSetInfo(cifsrpc_pipe *rpip)
{
	netShareSetInfoCall shSetInfoCall;
	netShareSetInfoResp shSetInfoResp;

	memset(&shSetInfoCall, 0, sizeof(netShareSetInfoCall));
	memset(&shSetInfoResp, 0, sizeof(netShareSetInfoResp));
	
	//Decode netShareSetInfo function call parameters
	if (cifsrpcEDnetShareSetInfoCall(&rpip->inPdu.dec, &shSetInfoCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call netShareSetInfo
	shSetInfoResp.returnValue = netShareSetInfo(rpip,
												&shSetInfoCall,
												&shSetInfoResp);
	
	//Encode netShareSetInfo function response parameters
	if (cifsrpcEDnetShareSetInfoResp(&rpip->outPdu.enc, &shSetInfoResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: netShareAdd
 *
 * Description	: Add a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/18/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
netShareAdd(cifsrpc_pipe *rpip, netShareAddCall *shcp, netShareAddResp *shrp)
{
	shareInfo *ship = &shcp->shareInfo;
	cifs_share *shp;	
	static cifs_shareCfg shCfg;
	uint32 type, perms;
	sec_cifsSd *sdp = NULL;
	
	//Check if share info is present
	if (!shcp->shareInfo.shareInfop)
		return (CIFSTATUS_INVALID_PARAMETER);
	
	//TODO check user access rights					
	
	memset(&shCfg, 0, sizeof(cifs_shareCfg));
	switch(shcp->level) {
		case 2:
			type = ship->info.lv2.type;
			strncpyFromUniStruc(shCfg.name,
								&ship->info.lv2.netName,
								CIFS_S_SHARECOMMENT);
			strncpyFromUniStruc(shCfg.comment,
								&ship->info.lv2.remark,
								CIFS_S_SHARECOMMENT);
			strncpyFromUniStruc(shCfg.pathName,
								&ship->info.lv2.path,
								CIFS_S_PATHSTR);
			break;
		case 502:
			type = ship->info.lv2.type;
			strncpyFromUniStruc(shCfg.name,
								&ship->info.lv2.netName,
								CIFS_S_SHARECOMMENT);
			strncpyFromUniStruc(shCfg.comment,
								&ship->info.lv2.remark,
								CIFS_S_SHARECOMMENT);
			strncpyFromUniStruc(shCfg.pathName,
								&ship->info.lv2.path,
								CIFS_S_PATHSTR);
			if (ship->info.lv502.secDescp)
				sdp = &ship->info.lv502.secDesc.sd;
			break;
		default:
			return (CIFSTATUS_INVALID_INFO_CLASS);
	}			
			
	//Check if the share already exists
	shp = cifs_findServerShareByName(rpip->clp->serverp, shCfg.name);
	if (shp != NULL)
		return (CIFSTATUS_OBJECT_NAME_COLLISION);

	//New share has to be a disk
	if (type != CIFS_SHARE_TYPE_DISK)
		return (CIFSTATUS_ACCESS_DENIED);
	
	//Check the path
	//TODO convert the path according to our internal volume/path scheme
	
	//Create the new share
	shp = cifs_addShare(rpip->clp->serverp, &shCfg);
	if (shp == NULL)
		return (CIFSTATUS_UNSUCCESSFUL);
	
	if (sdp != NULL)
	{	//New share has a security descriptor
			
		//Convert generic access bits in ACEs to their specific mapping
		cifs_secSDMapGenericAccessBits(sdp, &cifs_fileAccessGenericMap);
			
		//Allocate a new security descriptor and duplicate the temporary one
		shp->cfg.secDescp = cifs_secSecDescClone(sdp);
		if (shp->cfg.secDescp == NULL)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("netShareAdd: "
						"failed to store security desc"));
		}
		
	}
	
	//TODO update configuration on the SSC
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareAdd
 *
 * Description	: SRVSVC add a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/18/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareAdd(cifsrpc_pipe *rpip)
{
	netShareAddCall shAddCall;
	netShareAddResp shAddResp; 

	memset(&shAddCall, 0, sizeof(netShareAddCall));
	memset(&shAddResp, 0, sizeof(netShareAddResp));

	//Decode netShareAdd function call parameters
	if (cifsrpcEDnetShareAddCall(&rpip->inPdu.dec, &shAddCall, 1) != NFX_OK)
		return (NFX_ERR);

	//Call netShareAdd
	shAddResp.returnValue = netShareAdd(rpip, &shAddCall, &shAddResp);

	//Encode netShareAdd function response parameters
	if (cifsrpcEDnetShareAddResp(&rpip->outPdu.enc, &shAddResp, 1) != NFX_OK)
		return (NFX_ERR);

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: netShareDel
 *
 * Description	: Delete a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/18/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
netShareDel(cifsrpc_pipe *rpip, netShareDelCall *shcp, netShareDelResp *shrp)
{
	cifs_share *shp;	
	
	//Look for the share in the server's configuration
	shp = cifs_findServerShareByName(rpip->clp->serverp, shcp->shareName.str);
	if (shp == NULL)
		return (CIFSTATUS_BAD_NETWORK_NAME);

	if (shp->cfg.flags & CIFS_SHCFG_ISADMIN)
		return (CIFSTATUS_ACCESS_DENIED);

	//TODO check user access rights					
	
	//Disconnect the shares in use and delete the share
	cifs_deleteShare(rpip->clp->serverp, shp);

	//TODO update configuration on the SSC
	//if (shareCfgHasChanged)
	//{
	//}
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netShareDel
 *
 * Description	: SRVSVC Del a CIFS share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/18/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netShareDel(cifsrpc_pipe *rpip)
{
	netShareDelCall shDelCall;
	netShareDelResp shDelResp;

	memset(&shDelCall, 0, sizeof(netShareDelCall));
	memset(&shDelResp, 0, sizeof(netShareDelResp));
	
	//Decode netShareDel function call parameters
	if (cifsrpcEDnetShareDelCall(&rpip->inPdu.dec, &shDelCall, 1) != NFX_OK)
		return (NFX_ERR);

	//Call netShareDel
	shDelResp.returnValue = netShareDel(rpip, &shDelCall, &shDelResp);
	
	//Encode netShareDel function response parameters
	if (cifsrpcEDnetShareDelResp(&rpip->outPdu.enc, &shDelResp, 1) != NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: netServerGetInfo
 *
 * Description	: get information on the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
netServerGetInfo(	cifsrpc_pipe *rpip, 
					netServerGetInfoCall *svcp,
					netServerGetInfoResp *svrp)
{
	serverInfoContainer *ctp = &svrp->sviContr;
	cifs_server *svp = rpip->clp->serverp;	
	
	switch (svcp->level)
	{
		case 100:
			ctp->info.lv100.platformId	= 500;
			ctp->info.lv100.namep		= RPC_DUMMY_PTR;
			fillUniStrucWithoutAlloc(&ctp->info.lv100.name, svp->cfg.uniName);
			break;
		case 101:
			ctp->info.lv101.platformId	= 500;
			ctp->info.lv101.namep		= RPC_DUMMY_PTR;
			ctp->info.lv101.verMajor	= CIFS_SERVER_VERSION_MAJOR;
			ctp->info.lv101.verMinor	= CIFS_SERVER_VERSION_MINOR;
			ctp->info.lv101.type		= CIFS_SERVER_TYPE;
			ctp->info.lv101.commentp	= RPC_DUMMY_PTR;
			fillUniStrucWithoutAlloc(&ctp->info.lv101.name, svp->cfg.uniName);
			fillUniStrucWithoutAlloc(&ctp->info.lv101.comment,svp->cfg.comment);
			break;
		case 102:
			ctp->info.lv102.platformId	= 500;
			ctp->info.lv102.namep		= RPC_DUMMY_PTR;
			ctp->info.lv102.verMajor	= CIFS_SERVER_VERSION_MAJOR;
			ctp->info.lv102.verMinor	= CIFS_SERVER_VERSION_MINOR;
			ctp->info.lv102.type		= CIFS_SERVER_TYPE;
			ctp->info.lv102.commentp	= RPC_DUMMY_PTR;
			ctp->info.lv102.users		= 0xFFFFFFFF;
			ctp->info.lv102.disc		= 15;
			ctp->info.lv102.hidden		= 0;
			ctp->info.lv102.announce	= 240;
			ctp->info.lv102.annDelta	= 3000;
			ctp->info.lv102.licenses	= 0;
			ctp->info.lv102.usrPathp	= RPC_DUMMY_PTR;
			fillUniStrucWithoutAlloc(&ctp->info.lv102.name, svp->cfg.uniName);
			fillUniStrucWithoutAlloc(&ctp->info.lv102.comment,svp->cfg.comment);
			fillUniStrucWithoutAlloc(&ctp->info.lv102.usrPath,unistr_cColSlash);
			break;
		default:
			return (CIFSTATUS_INVALID_INFO_CLASS);
	}

	ctp->switchValue = svcp->level;
	ctp->serverInfop = RPC_DUMMY_PTR;
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netServerGetInfo
 *
 * Description	: SRVSVC get information on the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netServerGetInfo(cifsrpc_pipe *rpip)
{
	netServerGetInfoCall svGetInfoCall;
	netServerGetInfoResp svGetInfoResp;

	memset(&svGetInfoCall, 0, sizeof(netServerGetInfoCall));
	memset(&svGetInfoResp, 0, sizeof(netServerGetInfoResp));
	
	//Decode netServerGetInfo function call parameters
	if (cifsrpcEDnetServerGetInfoCall(&rpip->inPdu.dec, &svGetInfoCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call netServerGetInfo
	svGetInfoResp.returnValue = netServerGetInfo(rpip, 
												&svGetInfoCall,
												&svGetInfoResp);
	
	//Encode netServerGetInfo function response parameters
	if (cifsrpcEDnetServerGetInfoResp(&rpip->outPdu.enc, &svGetInfoResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: netServerSetInfo
 *
 * Description	: set information on the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
netServerSetInfo(	cifsrpc_pipe *rpip, 
					netServerSetInfoCall *svcp,
					netServerSetInfoResp *svrp)
{
	serverInfoContainer *ctp = &svcp->sviContr;
	cifs_server *svp = rpip->clp->serverp;	
	boolean serverCfgHasChanged = FALSE;
	
	switch (svcp->level)
	{
		case 101:
			if (ctp->info.lv101.commentp)
			{
				strncpyFromUniStruc(svp->cfg.comment,
									&ctp->info.lv101.comment,
									CIFS_S_SERVERCOMMENT);
				serverCfgHasChanged = TRUE;
			}
			break;
		case 102:
			if (ctp->info.lv102.commentp)
			{
				strncpyFromUniStruc(svp->cfg.comment,
									&ctp->info.lv102.comment,
									CIFS_S_SERVERCOMMENT);
				serverCfgHasChanged = TRUE;
			}
			break;
	}
	
	//TODO update configuration on the SSC
	//if (serverCfgHasChanged)
	//{
	//}
	
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: srvsvc_netServerSetInfo
 *
 * Description	: SRVSVC set information of the CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
srvsvc_netServerSetInfo(cifsrpc_pipe *rpip)
{
	netServerSetInfoCall svSetInfoCall;
	netServerSetInfoResp svSetInfoResp;

	memset(&svSetInfoCall, 0, sizeof(netServerSetInfoCall));
	memset(&svSetInfoResp, 0, sizeof(netServerSetInfoResp));
	
	//Decode netServerSetInfo function call parameters
	if (cifsrpcEDnetServerSetInfoCall(&rpip->inPdu.dec, &svSetInfoCall, 1)
			!= NFX_OK)
		return (NFX_ERR);

	//Call netServerSetInfo
	svSetInfoResp.returnValue = netServerSetInfo(rpip,
												&svSetInfoCall,
												&svSetInfoResp);
	
	//Encode netServerSetInfo function response parameters
	if (cifsrpcEDnetServerSetInfoResp(&rpip->outPdu.enc, &svSetInfoResp, 1)
			!= NFX_OK)
		return (NFX_ERR);
	
	return (NFX_OK);
}


static cifsrpc_function srvsvc_functions[] = {
{ SRVSVC_NETSHARENUM,		"NetrShareEnum",		srvsvc_netShareEnum },
{ SRVSVC_NETSHARENUMSTICKY,	"NetrShareEnumSticky",	srvsvc_netShareEnumSticky },
{ SRVSVC_NETSHAREGETINFO,	"NetrShareGetInfo",		srvsvc_netShareGetInfo },
{ SRVSVC_NETSHARESETINFO,	"NetrShareSetInfo",		srvsvc_netShareSetInfo },
{ SRVSVC_NETSHAREADD,		"NetrShareAdd",			srvsvc_netShareAdd },
{ SRVSVC_NETSHAREDEL,		"NetrShareDel",			srvsvc_netShareDel },
{ SRVSVC_NETSRVGETINFO,		"NetrServerGetInfo",	srvsvc_netServerGetInfo },
{ SRVSVC_NETSRVSETINFO,		"NetrServerSetInfo",	srvsvc_netServerSetInfo },
{ 0,						NULL,					NULL }
};


/*-----------------------------------------------------------------
 * Name			: cifsrpc_srvsvc
 *
 * Description	: process server services pipe requests
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_srvsvc(cifsrpc_pipe *rpip)
{
	return (cifsrpc_callFunction(rpip, srvsvc_functions));	
}


