/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc.c,v 1.1.1.1 2002/04/23 17:47:36 laurentb Exp $
 *
 * Name			: cifs-rpc.c
 *
 * Description	: Contains CIFS DCE/RPC api functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/09/2001
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#include "../sm-cifs/cifs-utils.h"
#include "../sm-cifs/cifs-unicode.h"
#else
#include <exports/cifs-api.h>
#include <exports/cifs-utils.h>
#include <exports/cifs-unicode.h>
#endif

#include "cifs-rpc.x"
#include "cifs-rpc-ed-hdr.h"
#include "cifs-rpc-srvsvc.h"
#include "cifs-rpc-wkssvc.h"
#include "cifs-rpc-lsarpc.h"
#include "cifs-rpc-err.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)
#define DBGPRINT    printf


//Ascii strings
char astr_srvsvc[]	= "srvsvc";
char astr_wkssvc[]	= "wkssvc";
char astr_ntsvcs[]	= "ntsvcs";
char atsr_lsarpc[]	= "lsarpc";
char atsr_lsass[]	= "lsass";

//Unicode strings
uni16 unistr_backpipe[] = { '\\', 'P', 'I', 'P', 'E', 0 };

uni16 unistr_srvsvc[]	= { 's','r','v','s','v','c', 0 };
uni16 unistr_samr[]		= { 's','a','m','r', 0 };
uni16 unistr_wkssvc[]	= { 'w','k','s','s','v','c', 0 };
uni16 unistr_netlogon[] = { 'N','E','T','L','O','N', 0 };
uni16 unistr_ntlsa[]	= { 'n','t','l','s','a', 0 };
uni16 unistr_ntsvcs[]	= { 'n','t','s','v','c','s', 0 };
uni16 unistr_lsass[]	= { 'l','s','a','s','s', 0 };
uni16 unistr_lsarpc[]	= { 'l','s','a','r','p','c', 0 };
uni16 unistr_winreg[]	= { 'w','i','n','r','e','g', 0 };

//Syntaxes
static cifsrpc_intf srvsvcV3 = 		//server services abstract interface Vers 3
{{0x4b324fc8, 0x1670, 0x01d3, {0x12,0x78}, {0x5a,0x47,0xbf,0x6e,0xe1,0x88}}, 3};
static cifsrpc_intf wkssvcV1 =		//workstation services abstract intf Vers 1
{{0x6bffd098, 0xa112, 0x3610, {0x98,0x33}, {0x46,0xc3,0xf8,0x7e,0x34,0x5a}}, 1};
static cifsrpc_intf lsarpcV0 =		//lsarpc abstract interface Version 0
{{0x12345678, 0x1234, 0xabcd, {0xef,0x00}, {0x01,0x23,0x45,0x67,0x89,0xab}}, 0};
static cifsrpc_intf netLogonV1 = 	//netLogon abstract interface Version 1
{{0x12345678, 0x1234, 0xabcd, {0xef,0x00}, {0x01,0x23,0x45,0x67,0xcf,0xfb}}, 1};
static cifsrpc_intf trIfV2 =		//transfer interface syntax Version 2 
{{0x8a885d04, 0x1ceb, 0x11c9, {0x9f,0xe8}, {0x08,0x00,0x2b,0x10,0x48,0x60}}, 2};
#define netLogonV2	trIfV2

//Pipes information
typedef struct {
	uni16   		*uCliName;				//unicode client name
	char			*aCliName;				//ascii client name
	uni16			*uSrvName;				//unicode server name
	char			*aSrvName;				//ascii server name
	cifsrpc_intf	*abstract;				//abstract syntax
	cifsrpc_intf	*transfer;				//transfer syntax
    int32   (*pipeFn) (cifsrpc_pipe *rpip);	//processing function
}pipeProperties;
                
#define CIFS_N_pipesProps	(sizeof(pipesProps)/sizeof(pipeProperties))

pipeProperties pipesProps[] = {
  { unistr_srvsvc,	astr_srvsvc,	unistr_ntsvcs,	astr_ntsvcs, 
	&srvsvcV3,		&trIfV2, cifsrpc_srvsvc },
  { unistr_wkssvc,	astr_wkssvc,	unistr_ntsvcs,	astr_ntsvcs,
	&wkssvcV1,		&trIfV2, cifsrpc_wkssvc },
  { unistr_lsarpc,	atsr_lsarpc,	unistr_lsass,	atsr_lsass,
	&lsarpcV0,		&trIfV2, cifsrpc_lsarpc },
/*  { unistr_samr,	unistr_lsass,  	trIfV2, NULL },
  { unistr_wkssvc,	unistr_ntsvcs,	trIfV2, NULL },
  { unistr_netlogon,unistr_lsass,	trIfV2, NULL },
  { unistr_ntlsa,	unistr_lsass,	trIfV2, NULL },
  { unistr_ntsvcs,	unistr_ntsvcs,	trIfV2, NULL },
  { unistr_lsass,	unistr_lsass,	trIfV2, NULL },
  { unistr_winreg,	unistr_winreg,	trIfV2, NULL },
*/  
};


/*-----------------------------------------------------------------
 * Name         : cifsrpc_callFunction
 *
 * Description  : look for the function associated with an operation
 * 				  number and call it.
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/11/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_callFunction(cifsrpc_pipe *rpip, cifsrpc_function *pipeFns)
{
	int32 i;
	ushort16 operationNb = rpip->inPdu.reqHdr.opNb;

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_callFunction: "
				"look for op 0x%x", operationNb));
	
	for (i=0; pipeFns[i].func != NULL; i++)
	{
		if (pipeFns[i].opNb == operationNb)
		{
			CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_callFunction: "
						"call %s", pipeFns[i].name));
			
			//Call the function
			if (pipeFns[i].func(rpip) != NFX_OK)
			{
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_callFunction: "
							"%s returned an error", pipeFns[i].name));
				cifsrpcEDreleaseData(&rpip->outPdu.enc);
				return (NFX_ERR);
			}
			return (NFX_OK);

		}
	}

	//function not supported
	cifsrpc_encFaultPdu(rpip, 0x1c010002);	
	
	return (NFX_OK);
}

			
/*-----------------------------------------------------------------
 * Name         : cifsrpc_pipeIsSupported
 *
 * Description  : check whether a named pipe is supported
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 12/12/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsrpc_pipeIsSupported(uni16 *pipeName)
{
	int32 i;
	
	//Check if pipe is supported
	for (i = 0; i < CIFS_N_pipesProps; i++)
	{
		if (unistrcasecmp(pipeName, pipesProps[i].uCliName) == 0)
		{
			if (pipesProps[i].pipeFn != NULL)
				return i;
			else
				return -1;
		}
	}

	return -1;
}
			

/*-----------------------------------------------------------------
 * Name         : cifsrpc_initPipe
 *
 * Description  : initialize an RPC pipe
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 12/11/2001
 *
 *-----------------------------------------------------------------
 */
int32 //NFX_OK or NFX_ERR
cifsrpc_initPipe(	cifsrpc_pipe *rpip, 
					uchar8 		 pipeIndex,
					cifs_client	 *clp,
					cifs_session *ssp)
{
		
	//Since pipe object is not filled with 0s when allocated, set all the
	//necessary fields
	rpip->fault		= FALSE;
	rpip->rpcBind	= FALSE;
	memset(&rpip->inPdu, 0, sizeof(cifsrpc_hdr));

	//Initialize the memory list associated with the pipe
	cifsrpc_initMemList(&rpip->memList);

	//Initialize the decoding structure
	if (cifsrpcEDinit(&rpip->inPdu.dec, FALSE, 0, &rpip->memList) != NFX_OK)
		return (NFX_ERR);
		
	//Initialize the encoding structure
	if (cifsrpcEDinit(&rpip->outPdu.enc, TRUE, 0, &rpip->memList) != NFX_OK)
		return (NFX_ERR);
	
	//Store which pipe we are initializing, and for which client and user
	rpip->index	= pipeIndex;
	rpip->clp	= clp;
	rpip->ssp	= ssp;
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name         : cifsrpc_closePipe
 *
 * Description  : close an RPC pipe
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/29/2002
 *
 *-----------------------------------------------------------------
 */
int32 //NFX_OK or NFX_ERR
cifsrpc_closePipe(cifsrpc_pipe *rpip) 
{
	
	//Free memory currently used by the incoming and outgoing PDU mgt
	cifsrpcEDreleaseData(&rpip->inPdu.dec);
	cifsrpcEDreleaseData(&rpip->outPdu.enc);
	
	//Free memory currently used on this pipe (There should not be any but...)
	cifsrpc_freeMemList(&rpip->memList);
	
	return (NFX_OK);
}




/*-----------------------------------------------------------------
 * Name         : cifsrpc_initOutPdu
 *
 * Description  : initialize the outgoing PDU
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 01/08/2002
 *
 *-----------------------------------------------------------------
 */
static int32 //NFX_OK or NFX_ERR
cifsrpc_initOutPdu(cifsrpc_pipe *rpip, uint32 size)
{
	cifsrpc_outPdu *outPdu = &rpip->outPdu;

	outPdu->fragLen 	= 0;
	outPdu->fragSentLen = 0;
	outPdu->pduSentLen	= 0;
	
	//TODO: Don't set the Pdufragment to 0 for now.
	//		SEE LATER IF REALLY NECESSARY
	
	//Release data still contained in the encoding structure, if any
	cifsrpcEDreleaseData(&outPdu->enc);
	
	//Allocate a Pdu fragment as the initial memory for the outgoing Pdu
	return cifsrpcEDinit(&outPdu->enc, TRUE, size, &rpip->memList);
}


	
/*-----------------------------------------------------------------
 * Name         : cifsrpc_fault
 *
 * Description  : set pipe's state to fault
 *
 * Created by   : Laurent Brard
 *
 * Date Created : 12/11/2001
 *
 *-----------------------------------------------------------------
 */
void cifsrpc_fault(cifsrpc_pipe *rpip)
{
   	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_fault: "
				"pipe %s in fault !", pipesProps[rpip->index].aCliName));
	rpip->fault = TRUE;
}


/*-----------------------------------------------------------------
 * Name			: cifsrpc_encFaultPdu
 *
 * Description	: encode a fault PDU
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_encFaultPdu(cifsrpc_pipe *rpip, uint32 status)
{
	cifsrpc_faultHdr fault;
	cifsrpc_ed encPdu;

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_encFaultPdu"));
	
	//Since it fits in a PDU fragment, build it in the outgoing
	//Pdu fragment buffer
	
	//Release memory used in outgoing encoding structre if any
	cifsrpcEDreleaseData(&rpip->outPdu.enc);
	
	//We don't use the encoding structure, so initialize it with a size of 0
	cifsrpc_initOutPdu(rpip, 0);
	
	cifsrpcEDinit(&encPdu, TRUE, 0, &rpip->memList);
	cifsrpcEDassignData(&encPdu, rpip->outPdu.frag + CIFSRPC_S_HEADER, 
							CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_HEADER, FALSE);
	
	//Fill the fault structure
	memset(&fault, 0, sizeof(cifsrpc_fault));
	fault.status = status;
	
	//Encode it
	cifsrpcEDfault(&encPdu, &fault, 1);

	//Encode header
	cifsrpc_encodeHeader(rpip,
						CIFSRPC_FAULT, 
						CIFSRPC_FLG_FIRST|CIFSRPC_FLG_LAST|CIFSRPC_FLG_NOCALL,
						encPdu.offset,
						0,
						rpip->inPdu.hdr.callId);

	return (NFX_OK);	
}
	
/*-----------------------------------------------------------------
 * Name			: cifsrpc_validateBind
 *
 * Description	: process a Bind Request PDU
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/09/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
cifsrpc_validateBind(cifsrpc_pipe *rpip, cifsrpc_bindReq *brp, uchar8 lv)
{
	pipeProperties *p = &pipesProps[rpip->index];

    IFDBG(CIFS_DBGLV_DBG3)
	{
		DBGPRINT("%scifsrpc_validateBind: rec av=%u tv=%u,expect av=%u tv=%u",
				cifs_spaces(lv), brp->abstract.version, brp->transfer.version,
				p->abstract->version, p->transfer->version);
		DBGPRINT("%srec auuid   =", cifs_spaces(lv+1));	
		memDumplight(&brp->abstract.uuid, sizeof(cifsrpc_uuid));
		DBGPRINT("%sexpect auuid=", cifs_spaces(lv+1));	
		memDumplight(&p->abstract->uuid, sizeof(cifsrpc_uuid));
		DBGPRINT("%srec tuuid   =", cifs_spaces(lv+1));	
		memDumplight(&brp->transfer.uuid, sizeof(cifsrpc_uuid));
		DBGPRINT("%sexpect tuuid=", cifs_spaces(lv+1));	
		memDumplight(&p->transfer->uuid, sizeof(cifsrpc_uuid));
	}
		
	//verify abstract and tranfer interfaces
	if ((brp->abstract.version != p->abstract->version)
	 || (brp->transfer.version != p->transfer->version)
	 || (memcmp(&brp->abstract.uuid, &p->abstract->uuid, sizeof(cifsrpc_uuid)))
	 || (memcmp(&brp->transfer.uuid, &p->transfer->uuid, sizeof(cifsrpc_uuid))))
		return (NFX_ERR);
	
	return (NFX_OK);
}
	

/*-----------------------------------------------------------------
 * Name			: cifsrpc_recBindRequest
 *
 * Description	: process a received Bind Request PDU
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
cifsrpc_recBindRequest(cifsrpc_pipe *rpip, cifsrpc_ed *decp)
{
	cifsrpc_bindReq bindReq;
	cifsrpc_bindAck	bindAck;
	cifsrpc_ed encPdu;
	boolean bindAccepted;	

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_recBindRequest"));

	//Check if already received
	if (rpip->rpcBind)
		return (NFX_ERR);
	
	//Decode the bind request
	if (cifsrpcEDbindRequest(decp, &bindReq, 1) != NFX_OK)
		return (NFX_ERR);
	
	//Decode the authentication verifier if any
	if (rpip->inPdu.hdr.authLen != 0)
	{
		CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_recBindRequest: "
					"Authentication verifier"));
		//TODO
	}

	//Verify we support that bind
	bindAccepted = (cifsrpc_validateBind(rpip, &bindReq, 1) == NFX_OK);
	
	//Build Answer: since it fits in a PDU fragment, build it in the outgoing
	//Pdu fragment buffer
	
	//We don't use the encoding structure, so initialize it with a size of 0
	cifsrpc_initOutPdu(rpip, 0);
	
	cifsrpcEDinit(&encPdu, TRUE, 0, &rpip->memList);
	cifsrpcEDassignData(&encPdu, rpip->outPdu.frag + CIFSRPC_S_HEADER, 
							CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_HEADER, FALSE);
	
	//Fill the Bind Ack structure
	bindAck.maxTFragSize = CIFSRPC_S_PDUFRAGMENT;
	bindAck.maxRFragsize = CIFSRPC_S_PDUFRAGMENT;
	//TODO check authentication to set assocGroupId accordingly
	if (bindReq.assocGroupId != 0)
		bindAck.assocGroupId = bindReq.assocGroupId;
	else
		bindAck.assocGroupId = 0xDEAD;	//see if it works...
	strcpy(bindAck.secAddrPort, "\\PIPE\\");
	strcat(bindAck.secAddrPort, pipesProps[rpip->index].aSrvName);
	bindAck.secAddrlen			= strlen(bindAck.secAddrPort) + 1;
	bindAck.nbResults			= 1;
	if (bindAccepted)
	{
		bindAck.result			= 0; 
		bindAck.reason			= 0;
		bindAck.transfer.version= bindReq.transfer.version;
		memcpy(&bindAck.transfer.uuid, &bindReq.transfer.uuid,
				sizeof(cifsrpc_uuid));
	} else {
		bindAck.result			= 2; 	//reject bind
		bindAck.reason			= 1;	//abstract syntax not supported
		memset(&bindAck.transfer.uuid, 0, sizeof(cifsrpc_uuid));
	}
	
	//Encode it
	cifsrpcEDbindAck(&encPdu, &bindAck, 1);

	//TODO: Add authentication if any
	
	//Encode header
	cifsrpc_encodeHeader(	rpip,
							CIFSRPC_BINDACK, 
							CIFSRPC_FLG_FIRST | CIFSRPC_FLG_LAST,
							encPdu.offset,
							0/*authLen*/,
							rpip->inPdu.hdr.callId);

	//Remember we received the Bind request
	rpip->rpcBind = TRUE;
	
	return (NFX_OK);	
}
	
		
/*-----------------------------------------------------------------
 * Name			: cifsrpc_recRequest
 *
 * Description	: receive a Request PDU fragment
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/10/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
cifsrpc_recRequest(cifsrpc_pipe *rpip, cifsrpc_inPdu *inpp, cifsrpc_ed *fdecp)
{
	uint32 fragDataSize;
	uchar8 flags;
	uchar8 *fragDatap;
	cifsrpc_ed *pdecp;
	int32 result;

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_recRequest"));

	//Check if pipe bound
	if (!rpip->rpcBind)
		return (NFX_ERR);
	
	//Decode the request header
	if (cifsrpcEDrequestHdr(fdecp, &inpp->reqHdr, 1) != NFX_OK)
		return (NFX_ERR);
	
	//TODO handle authentication
	
	fragDataSize = inpp->fragLen - CIFSRPC_S_REQ_HDR /* - authen stuff */;
	fragDatap	 = fdecp->bufp + fdecp->offset;
	pdecp		 = &inpp->dec;					//PDU decoding struct
	
	//Check PDU size
	if (inpp->dec.size + fragDataSize > CIFSRPC_S_PDU)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_recRequest: PDU too long !"));
		return (NFX_ERR);
	}
	
	if ((inpp->hdr.flags & CIFSRPC_FLG_FIRST) 
	 && (inpp->hdr.flags & CIFSRPC_FLG_LAST))
	{	//PDU is in one piece, analyze it directly in the fragment buffer
		cifsrpcEDassignData(pdecp, fragDatap, fragDataSize, FALSE);

	} else {
		//We have to copy the data in the decoding structure
		
		if ((inpp->hdr.flags & CIFSRPC_FLG_FIRST) && pdecp->offset == 0) 
		{
			//Initialize the decoding structure with the maximum PDU size
			//we support, so we never need to realloc the received data
			if (cifsrpcEDinit(pdecp, FALSE, CIFSRPC_S_PDU, &rpip->memList)
					!= NFX_OK)
				return (NFX_ERR);

			//TODO? we need to get the allocation Hint to know how much we are
			//expecting ?
			//This would be more a sanity check. When we receive the last full
			//fragment, we should have everything.
		}
		
		if (cifsrpcEDaddData(pdecp, fragDatap, fragDataSize) != NFX_OK)
		{	//We don't have enough room to store the PDU
			printf("cifsrpc_recRequest: PDU bigger than supported size (%u)",
					CIFSRPC_S_PDU);			
			return (NFX_ERR);
		}
	}
		
	if (inpp->hdr.flags & CIFSRPC_FLG_LAST)
	{	//I make the bet that the result data will fit in a fragment PDU.
		//If I loose, i will have to copy all the data in a bigger buffer, 
		//but if I win I save a copy.
		//As most of the RPC call results we are supposed to handle are smaller
		//than the PDU fragment size, this should be the optimal option

		//Get ready to decode the incoming PDU
		pdecp->offset = 0;
		
		//We don't use the encoding structure, so initialize it with a size of 0
		cifsrpc_initOutPdu(rpip, 0);
	
		//Assign to the encoding structure the outgoing PDU fragment
		cifsrpcEDinit(&rpip->outPdu.enc, TRUE, 0, &rpip->memList);
		cifsrpcEDassignData(&rpip->outPdu.enc, 
							rpip->outPdu.frag + CIFSRPC_S_RESP_FULLHDR, 
							CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_RESP_FULLHDR,
							FALSE);
		
		//Process the request
		if (pipesProps[rpip->index].pipeFn != NULL)
		{	//a little paranoia
			result = (pipesProps[rpip->index].pipeFn)(rpip);
		}
		else {
			//TODO remove the test when sure we never come here
			NFX_ASSERT(FALSE);
			result = NFX_ERR;
		}

		//Free memory used to process the request
		cifsrpc_freeMemList(&rpip->memList);

		//TODO: handle authentication if any
		
		//Check if the answer finally fits in the PDU fragment
		if (rpip->outPdu.enc.freeMem == FALSE)
		{	//the answer is in the outgoing PDU fragment buffer
			flags = CIFSRPC_FLG_FIRST | CIFSRPC_FLG_LAST;
			fragDataSize = rpip->outPdu.enc.offset;
		}
		else
		{	//the answer is in a malloc'ed buffer
			flags = CIFSRPC_FLG_FIRST;
			//Copy back first fragment data in the outgoing PDU fragment buffer
			memcpy(	rpip->outPdu.frag + CIFSRPC_S_RESP_FULLHDR, 
					rpip->outPdu.enc.bufp,
					CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_RESP_FULLHDR);
			fragDataSize = CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_RESP_FULLHDR;
		}

		//Add headers
		cifsrpc_encodeHeader(rpip,
							CIFSRPC_RESPONSE, 
							flags,
							fragDataSize + CIFSRPC_S_RESP_HDR,
							0/*authLen*/,
							rpip->inPdu.hdr.callId);

		cifsrpc_encodeResponseHeader(rpip, rpip->outPdu.enc.offset);
	}	
	
	//Store what has already been "sent" (put in outgoing PDU fragment buffer)
	rpip->outPdu.pduSentLen = fragDataSize;

	return (NFX_OK);	
}
	


/*-----------------------------------------------------------------
 * Name			: cifsrpc_recPduFrag
 *
 * Description	: process an RPC PDU fragment
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/11/2001
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK / NFX_ERR
cifsrpc_recPduFrag(cifsrpc_pipe *rpip)
{
	cifsrpc_inPdu *inpp = &rpip->inPdu;
	uint32 pduFragSize = inpp->fragLen;
	cifsrpc_ed dec;
	
	int32 result;

	if (rpip->fault)
	{
		CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsrpc_recPduFrag: pipe in fault state"));
		cifsrpc_encFaultPdu(rpip, 0x1c010002);	
		return (NFX_OK);	//we return OK so that fault PDU goes out
	}
	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_recPduFrag: "
				"pduFragSize=%d", pduFragSize));
	IFDBG(CIFS_DBGLV_DBG2)
	{
		//IFDBG(CIFS_DBGLV_DBG3)
		//	memDump(inpp->frag, pduFragSize);
	}

	//Initialize PDU fragment decoding structure
	cifsrpcEDinit(&dec, FALSE, 0, &rpip->memList);
	CIFSRPC_ED_SETENDIAN(&dec, inpp->bigEndian);
	cifsrpcEDassignData(&dec, inpp->frag, pduFragSize, FALSE);
	
	switch (inpp->hdr.packetType)
	{
		case CIFSRPC_BIND:
			result = cifsrpc_recBindRequest(rpip, &dec);
			break;
		
		case CIFSRPC_REQUEST:
			result = cifsrpc_recRequest(rpip, inpp, &dec);
			break;
			
		default:
			result = NFX_ERR;
	}

	if (result != NFX_OK)
	{
		cifsrpc_fault(rpip);		
		cifsrpc_encFaultPdu(rpip, CIFSRPC_ERR_CALL_CANCELED);
		//TODO see the <> with cifsrpc_encFaultPdu(rpip, 0x1c010002);
		//TODO test those status values with sniffer
	}
	
	inpp->fragLen	 = 0;
	inpp->fragRecLen = 0;
	
	return (NFX_OK);
}

	
/*-----------------------------------------------------------------
 * Name			: cifsrpc_writeData
 *
 * Description	: write on a RPC named pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/09/2001
 *
 *-----------------------------------------------------------------
 */
int32	//Number of bytes written
cifsrpc_writeData(	cifsrpc_pipe	*rpip,
					uchar8 			*datp,
					uint32			datCnt)
{
	uint32 bytesToWriteCnt;
	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_writeData: "
				"%d bytes on %s (fragL=%d, fragRecL=%d)", 
				datCnt, pipesProps[rpip->index].aCliName, 
				rpip->inPdu.fragLen, rpip->inPdu.fragRecLen));

	bytesToWriteCnt = MIN(CIFSRPC_S_PDUFRAGMENT - rpip->inPdu.fragRecLen,
							datCnt);
	
	if (bytesToWriteCnt == 0)
	{
    	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_writeData: "
					"curr. fragment size %d exceeds max %d",
					rpip->inPdu.fragRecLen + datCnt, CIFSRPC_S_PDUFRAGMENT));
		
		cifsrpc_fault(rpip);
		return (-1);
	}

	if (rpip->inPdu.fragLen == 0)
	{	//First part of the PDU fragment, analyze header
		//Assume we always receive at least the header size the first time 
		if (cifsrpc_decodeHeader(rpip, datp, bytesToWriteCnt) != NFX_OK)
		{
			cifsrpc_fault(rpip);
			return (-1);
		}
		bytesToWriteCnt -= CIFSRPC_S_HEADER; 
		datp += CIFSRPC_S_HEADER;
	}

	bytesToWriteCnt = MIN(bytesToWriteCnt, rpip->inPdu.fragLen);
	memcpy(&rpip->inPdu.frag[rpip->inPdu.fragRecLen], datp,	bytesToWriteCnt);
	rpip->inPdu.fragRecLen += bytesToWriteCnt;

    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_writeData: "
				"bytesToWC=%d, fragL=%d, fragRecL=%d", 
				bytesToWriteCnt, rpip->inPdu.fragLen, rpip->inPdu.fragRecLen));

	if (rpip->inPdu.fragRecLen == rpip->inPdu.fragLen)
	{
		cifsrpc_recPduFrag(rpip);
		//TODO Check recPdu worked fine
	}
		
	return (datCnt);
}

	
/*-----------------------------------------------------------------
 * Name			: cifsrpc_buildNextPdu
 *
 * Description	: build next PDU
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/16/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK / NFX_ERR
cifsrpc_buildNextPdu(cifsrpc_pipe *rpip, cifsrpc_outPdu *outPp)
{
	uchar8 flags;
	ushort16 fragDataSize;
		
	if (rpip->fault)
	{
		CIFS_PRINT(CIFS_DBGLV_DBG, 
					("cifsrpc_buildNextPdu: pipe in fault state"));
		cifsrpc_encFaultPdu(rpip, 0x1c010002);	
		return (NFX_OK);
	}

	if (outPp->pduSentLen < outPp->enc.offset)
	{	//more data to go
	
		fragDataSize = outPp->enc.offset - outPp->pduSentLen;
		if (fragDataSize > CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_RESP_FULLHDR)
		{
			fragDataSize = CIFSRPC_S_PDUFRAGMENT - CIFSRPC_S_RESP_FULLHDR;
			flags = 0;
		} else 	//this is the last fragment
			flags = CIFSRPC_FLG_LAST;
		
		//copy fragment data in the outgoing PDU fragment buffer
		memcpy( outPp->frag + CIFSRPC_S_RESP_FULLHDR,
				outPp->enc.bufp + outPp->pduSentLen,
				fragDataSize);

		//TODO: handle authentication if any
		
		//Add headers
		cifsrpc_encodeHeader(rpip,
							CIFSRPC_RESPONSE, 
							flags,
							fragDataSize + CIFSRPC_S_RESP_HDR,
							0/*authLen*/,
							rpip->inPdu.hdr.callId);

		cifsrpc_encodeResponseHeader(rpip, outPp->enc.offset-outPp->pduSentLen);
	
		//update what has already been "sent"
		rpip->outPdu.pduSentLen += fragDataSize;
		
	} else {	//client tries to read more data than we have !
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsrpc_buildNextPdu: no more data !"));
		//return (NFX_ERR);
	}
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifsrpc_readData
 *
 * Description	: read from an RPC named pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/09/2001
 *
 *-----------------------------------------------------------------
 */
int32	//Number of bytes read
cifsrpc_readData(cifsrpc_pipe	*rpip,
				 uchar8			**datpp,
				 uint32			dataRequested,
				 boolean		*moreData)
{
	cifsrpc_outPdu *outPp = &rpip->outPdu;
	uint32 dataAvailable, dataRead;
		
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsrpc_readData: "
				"request %d bytes on %s", 
				dataRequested, pipesProps[rpip->index].aCliName));
	
	dataRead = 0;
	
	//We should not send more than a PDU fragment at a time
	if (dataRequested > CIFSRPC_S_PDUFRAGMENT)
	{	
		CIFS_PRINT(CIFS_DBGLV_ERR, 
			("cifsrpc_readData: dataReq=%d > PDU frag size", dataRequested));
		//cifsrpc_fault(rpip);	TODO see if we want to do this
		return (0);
	}
	
	dataAvailable = outPp->fragLen - outPp->fragSentLen; 	
	if (dataAvailable <= 0)
	{	//Current fragment is empty, see if the whole PDU has been sent

		if (cifsrpc_buildNextPdu(rpip, outPp) != NFX_OK)
		{
			//cifsrpc_fault(rpip);  TODO see if we want to do this
			return (0);
		}
		dataAvailable = outPp->fragLen;
	}	
	
	dataRead = MIN(dataRequested, dataAvailable);
	*datpp = outPp->frag + outPp->fragSentLen;
	outPp->fragSentLen += dataRead;
	
	*moreData = (outPp->fragLen > outPp->fragSentLen);
			
	return (dataRead);
}





