/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-dc.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: cifs-dc.c
 *
 * Description	: Contains functions that send requests from CIFS
 * 				  servers to a domain controller (via SSC). 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/24/2002
 *
 *-----------------------------------------------------------------
 */
#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-intmsg.h"


/*-----------------------------------------------------------------
 * Name			: cifs_authReqToSsc
 *
 * Description	: Send a request to cifsauth on SSC for user authentication 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_authReqToSsc(cifs_operation *opP)
{
	int32				rc = NFX_OK;
	cifs_client			*clientP;
	cifs_session		*sessionP;
	cifsauthReqMsg_t	authReq;
	uint32				srcPort, destPort;

	CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_authReqToSsc"));

	clientP = opP->clp;
	sessionP = opP->ssp;

	memset(&authReq, 0, sizeof(cifsauthReqMsg_t));

	//-------------------------------
	// Fill in authentication request 
	//-------------------------------
	authReq.hdr.msgType = CIFSAUTH_REQ;
	authReq.hdr.clientPrivate = opP->threadId;
	
	authReq.fromAddr = clientP->serverp->cfg.ipAddr;
	strUni2Dos(authReq.userName, sessionP->user, CIFS_S_USERNAME);
	strUni2Dos(authReq.domName, sessionP->domain, CIFS_S_DOMAINNAME);
	strncpy(authReq.chal, clientP->encryptKey, CIFS_S_ENCRYPTKEY); 
	// Case insensitive password
	authReq.lm_resp.len = CIFS_S_ENCRYPTPWD;
	memcpy(authReq.lm_resp.blob, sessionP->caseInsenPwd, CIFS_S_ENCRYPTPWD);
	// Case sensitive password
	authReq.nt_resp.len = CIFS_S_ENCRYPTPWD;
	memcpy(authReq.nt_resp.blob, sessionP->caseSenPwd, CIFS_S_ENCRYPTPWD);

	//----------------------------------------------------
	// Send user authentication request to cifsauth on SSC
	//----------------------------------------------------
	rc = cifs_msgSend(CIFS_INTMSG_TO_AUTH,
					(uchar8 *)&authReq,
					sizeof(cifsauthReqMsg_t),
					FALSE);

	if (rc != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_authReqToSsc: "
					"Failure to send user authentication request to SSC"));
	}

	CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_authReqToSsc"));
	return(rc);
} /* cifs_authReqToSsc() */



/*-----------------------------------------------------------------
 * Name			: cifs_showUserCreds
 *
 * Description	: This function displays the user credentials
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/24/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_showUserCreds(cifs_session		*ssP,
					cifsauthCred_t	*c)
{
	char			user[CIFS_S_USERNAME+1];
	char			domain[CIFS_S_DOMAINNAME+1];
	int				i;

	memset(user, 0, CIFS_S_USERNAME+1);
	strUni2Dos(user, ssP->user, CIFS_S_USERNAME);

	memset(domain, 0, CIFS_S_DOMAINNAME+1);
	strUni2Dos(domain, ssP->domain, CIFS_S_DOMAINNAME);

	printf("user: %s, domain: %s\n", user, domain);

    printf ("urid: 0x%x, grid: 0x%x, uflgs: 0x%x\n",
				c->user_rid, c->group_rid, c->user_flgs);

	printf ("numGids: %d\n", c->num_gids);
	printf ("g_rid/attr: ");
	for (i = 0; i < c->num_gids; i++)
	{
		printf ("0x%x/0x%x ", c->gids[i].g_rid, c->gids[i].attr);
	}
	printf ("\n");

	printf ("num_auths: %d, sid_rev: %d\n", c->num_auths, c->sid_rev_num);
	printf ("id_auth: ");
	for (i = 0; i < 6; i++)
	{
		printf ("%x ", c->id_auth[i]);
	}
	printf ("\n");

	printf ("sub_auths: ");
	for (i = 0; i < c->num_auths; i++)
	{
		printf ("0x%x ", c->sub_auths[i]);
	}
	printf ("\n");
} /* cifs_showUserCreds() */
