/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-encdec.c,v 1.3 2002/12/10 18:13:18 laurentb Exp $
 *
 * Name			: cifs-encdec.c
 *
 * Description	: Contains CIFS encoding and decoding functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-err.h"
#include "cifs-msg.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)

static char smb_headerStr[] = { 0xFF, 'S', 'M', 'B' };

#define MF_SESSION		0x01	//Session required
#define MF_SIU			0x03	//ShareInUse required (NB: implies session)
#define MF_WR			0x04	//Write access required
#define MF_WOPLOCK		0x08	//Wait before process when oplock
#define MF_IPCUSE		0x10	//Used in IPC dialog
#define MF_FRAGSUP		0x20	//Support fragmented SMB packet
#define MF_NOTHREAD		0x40	//Does NOT need to run in a thread
#define MF_COD			0x80	//Need an outgoing buffer with a COD

struct cifs_msg_description
{
	char *name;
	int32 	(*processFunction)(	cifs_client *clp,
								cifs_session *ssp,
								cifs_shareInUse *siup,
								cifs_operation *opp,
								uchar8 *icmdp,
								uchar8 *ocmdp);
	uint32	flags;
}cifs_msg_description[256] = {
{ "createDir",	cifsm_createDir,MF_SIU | MF_WR },						//0x00
{ "delDir",		cifsm_delDir,	MF_SIU | MF_WR },
{ "open",		cifsm_open,		MF_SIU | MF_WOPLOCK },
{ "create",		cifsm_create,	MF_SIU },
{ "close",		cifsm_close,	MF_SIU | MF_IPCUSE },
{ "flush",		cifsm_flush,	MF_SIU },
{ "delete",		cifsm_delete,	MF_SIU | MF_WR | MF_WOPLOCK },
{ "rename",		cifsm_rename,	MF_SIU | MF_WR | MF_WOPLOCK },
{ "queryInfo",	cifsm_queryInfo,MF_SIU },
{ "setInfo",	cifsm_setInfo,	MF_SIU | MF_WR },
{ "read",		cifsm_read,		MF_SIU | MF_COD },
{ "write",		cifsm_write,	MF_SIU | MF_WR | MF_IPCUSE | MF_FRAGSUP },
{ "lockByte",	cifsm_lockByte,	MF_SIU },
{ "unlockByte",	cifsm_ulockByte,MF_SIU },
{ "createTmp",	cifsm_createTmp,MF_SIU | MF_WOPLOCK },
{ "createNew",	cifsm_createNew,MF_SIU },
{ "checkPath",	cifsm_checkPath,MF_SIU },								//0x10
{ "exit",		cifsm_exit,		0 },
{ "seek",		cifsm_seek,		MF_SIU },
{ "lockRead",	cifsm_lockRead,	MF_SIU | MF_COD },
{ "wrUnlock",	cifsm_wrUnlock,	MF_SIU | MF_WR },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ "readRaw",	NULL,			0 },
{ "readMpx",	NULL,			0 },
{ "readMpxS",	NULL,			0 },
{ "writeRaw",	NULL,			0 },
{ "writeMpx",	NULL,			0 },
{ "writeMpxS",	NULL,			0 },
{ "writeCmp",	NULL,			0 },									//0x20
{ "queryServer",NULL,			0 },
{ "setInf2",	cifsm_setInf2,	MF_SIU | MF_WR },
{ "queryInf2",	cifsm_queryInf2,MF_SIU },
{ "lockNX",		cifsm_lockNX,	MF_SIU },
{ "transac",	cifsm_transac,	MF_SIU | MF_IPCUSE | MF_WOPLOCK | MF_FRAGSUP },
{ "transacS",	cifsm_transacS,	MF_SIU | MF_IPCUSE },
{ "ioctl",		cifsm_ioctl,	0 /* ? */ },
{ "ioctlS",		NULL,			0 },
{ "copy",		cifsm_copy,		MF_SIU | MF_WR | MF_WOPLOCK },
{ "move",		cifsm_move,		MF_SIU | MF_WR | MF_WOPLOCK },
{ "echo",		cifsm_echo,		0 },
{ "wrClose",	cifsm_wrClose,	MF_SIU | MF_FRAGSUP },
{ "openNX",		cifsm_openNX,	MF_SIU | MF_IPCUSE | MF_WOPLOCK },
{ "readNX",		cifsm_readNX,	MF_SIU | MF_IPCUSE | MF_COD },
{ "writeNX",	cifsm_writeNX,	MF_SIU | MF_IPCUSE | MF_FRAGSUP },
{ "newFileSize",NULL,			0 },									//0x30
{ "closeNTreeD",NULL,			0 },
{ "transac2",	cifsm_transac2,	MF_SIU | MF_WOPLOCK },
{ "transac2S",	cifsm_transac2S,MF_SIU },
{ "findClos2",	cifsm_findClos2,MF_SIU },
{ "findNClose",	cifsm_findNClos,MF_SIU },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0x40
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0x50
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0x60
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ "treeConnect",cifsm_treeCon,	MF_SESSION | MF_NOTHREAD },				//0x70
{ "treeDisc",	cifsm_treeDisc,	MF_SIU },
{ "negociate",	cifsm_negociate,MF_NOTHREAD },
{ "ssSetupNX",	cifsm_ssSetupNX,0 },
{ "logoffNX",	cifsm_logoffNX,	0 },
{ "treeConnNX",	cifsm_treeConNX,MF_SESSION | MF_NOTHREAD },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ "infoDisk",	cifsm_infoDisk,	MF_SIU },								//0x80
{ "search",		cifsm_search,	MF_SIU },
{ "findFirst",	cifsm_search,	MF_SIU },
{ "findUnique",	cifsm_search,	MF_SIU },
{ "findClose",	cifsm_findClose,MF_SIU },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0x90
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0x9F
{ "ntTransac",	cifsm_ntTrans,	MF_SIU | MF_IPCUSE | MF_WOPLOCK/*|MF_FRAGSUP*/},
{ "ntTransacS",	cifsm_ntTransS,	MF_SIU | MF_IPCUSE /*| MF_FRAGSUP ???*/},
{ "ntCreateNX",	cifsm_ntCreatNX,MF_SIU | MF_IPCUSE | MF_WOPLOCK },
{ NULL,			NULL,			0 },
{ "ntCancel",	cifsm_ntCancel,	MF_SIU },
{ "ntRename",	NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0xB0
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0xC0
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0xD0
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0xE0
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },									//0xF0
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 },
{ NULL,			NULL,			0 }										//0xFF
}; 



/*-----------------------------------------------------------------
 * Name			: cifs_msgDataCpy
 *
 * Description	: copy data from a CIFS message
 * 				  supports fragmented message
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/11/2002
 *
 *-----------------------------------------------------------------
 */
uchar8 *
cifs_msgDataCpy(uchar8 *dstp, cifs_cmdDesc *msgDscp, uint32 offset, uint32 len)
{
	uchar8 *datp;
	uint32 fragSize, fragOff, dataToCopySize;
	pkt_desc_t *fragp = msgDscp->tplPktDescp;

	//Check we are looking for data within CIFS message
	if (offset + len > msgDscp->smbLen)
		return (NULL);

	//Calculate first fragment size
	fragSize = cifs_msgFragSize(msgDscp);
	
	if (offset + len <= fragSize)
	{	//Data are in the first fragment
		return ((uchar8*)memcpy(dstp, msgDscp->smbp + offset, len));
	} else {
		fragOff = 0;

		//Look for the fragment where data start
		while (offset > fragOff + fragSize)
		{	
			fragOff	+= fragSize;
			fragp = fragp->pNext;
			if (fragp == NULL)
				return (NULL);
			fragSize = fragp->hdrSize;
		}
		
		//Copy data from fragment where data start
		dataToCopySize = MIN(len, fragSize - (offset - fragOff));
		memcpy(dstp, fragp->pHdr + (offset - fragOff), dataToCopySize);
		datp = dstp + dataToCopySize;
		len  -= dataToCopySize;
		
		//Copy remaining data
		while (len > 0)
		{
			fragp = fragp->pNext;
			if (fragp == NULL)
				return (NULL);
			dataToCopySize = MIN(len, fragp->hdrSize);
			memcpy(datp, fragp->pHdr, dataToCopySize); 
			datp += dataToCopySize;
			len  -= dataToCopySize;
		}
	}
	return dstp;
}
	
/*-----------------------------------------------------------------
 * Name			: cifs_initSMBHeader 
 *
 * Description	: Initialize the header of an SMB packet and the
 * 					associated SMB msg descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/07/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_initSMBHeader(cifs_cmdDesc *cmdp)
{
	//Fill message header
	memcpy(cmdp->smbp, smb_headerStr, SMBO_CMD);
	memset(cmdp->smbp + SMBO_CMD, 0, SMB_S_HEADER - SMBO_CMD);
	
	//Initialize SMB command descriptor
	cmdp->smbLen	= SMB_S_HEADER;	
	cmdp->cmdTypep  = cmdp->smbp + SMBO_CMD;
	cmdp->cmdp		= cmdp->smbp + SMB_S_HEADER;
}


/*-----------------------------------------------------------------
 * Name			: cifs_initSMBHeaderFromIncomingSMB
 *
 * Description	: Initialize the header of an SMB packet and the
 * 					associated SMB msg descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 2/14/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_initSMBHeaderFromIncomingSMB(uchar8 *osmbp, uchar8 *ismbp)
{
	 //Copy request header
	memcpy(osmbp, ismbp, SMB_S_HEADER);

	//Status field = 0 (success)
	memset(osmbp + SMBO_ERCL, 0, 4);

	//Set flags fields
	PCAR(osmbp, SMBO_FLGS, SMB_FLGS_RESPONSE /*TODO see what is needed */);
	PUTLE2(osmbp, SMBO_FLG2, SMB_FLG2_LONG_NAMES  /*TODO see what is needed */);
	
	//Set Word count and byte count to 0
	osmbp[SMBO_WCNT] = 0;
	PUTLE2(osmbp, SMBO_WCNT+1, 0);
}	
	

/*-----------------------------------------------------------------
 * Name			: cifs_setCmdWordCnt 
 *
 * Description	: Set command word count and update msg length
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/27/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_setCmdWordCnt(cifs_cmdDesc *cmdp, uint32 wordCnt, boolean init)
{
	if (init) 
		memset(cmdp->cmdp + 3, 0, wordCnt*2);
	
	cmdp->cmdp[0] = wordCnt;
	cmdp->smbLen += wordCnt*2 + 1;
/*	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_setCmdWordCnt: "
				"wcnt@=%x, wcnt=%d, smbl=%d",
				cmdp->cmdp, wordCnt, cmdp->smbLen));
*/
}


/*-----------------------------------------------------------------
 * Name			: cifs_setCmdByteCnt 
 *
 * Description	: Set command byte count and update msg length
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/27/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_setCmdByteCnt(cifs_cmdDesc *cmdp, uint32 byteCnt, boolean init)
{
	if (init) 
		memset(cmdp->cmdp + 3 + cmdp->cmdp[0]*2, 0, byteCnt);
	
	PUTLE2(cmdp->cmdp, cmdp->cmdp[0]*2 + 1, byteCnt);
	cmdp->smbLen += byteCnt + 2;
/*	
	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_setCmdByteCnt: "
				"bcnt@=%x, bcnt=%d, smbl=%d",
				(cmdp->cmdp + cmdp->cmdp[0]*2 + 1), byteCnt, cmdp->smbLen));
*/				
}


/*-----------------------------------------------------------------
 * Name			: cifs_setCmdWordByteCnt 
 *
 * Description	: Set command word and byte count and update msg length
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_setCmdWordByteCnt(cifs_cmdDesc *cmdp, uint32 wordCnt, uint32 byteCnt,
						boolean init)
{
	if (init) {
		//Skip word count and set words, byte count and bytes to 0
		memset(cmdp->cmdp + 1, 0, wordCnt*2 + byteCnt + 2);
		cmdp->smbLen += 3;					// Word and byte counters
	}
	cmdp->cmdp[0] = wordCnt;
	PUTLE2(cmdp->cmdp, wordCnt*2 + 1, byteCnt);
	cmdp->smbLen += wordCnt*2 + byteCnt;
/*	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_setCmdWordByteCnt: "
				"wcnt@=%x, wcnt=%d, bcnt@=%x, bcnt=%d, smbl=%d",
				cmdp->cmdp, wordCnt, (cmdp->cmdp + wordCnt*2 + 1), byteCnt,
				cmdp->smbLen));
*/
}


/*-----------------------------------------------------------------
 * Name			: cifs_errResp 
 *
 * Description	: fill error field of the response
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_errResp(cifs_operation *opp, uchar8 class, ushort16 err,
				uint32 nterr)
{
	char *smbp = opp->respMsg.smbp;

	cifs_setCmdWordByteCnt(&opp->respMsg, 0, 0, TRUE);
	if (nterr != 0 
	 && opp->clp != NULL
	 && opp->clp->capabilities & CAP_STATUS32)
	{
		PUTLE2(smbp, SMBO_FLG2, GETLE2(smbp, SMBO_FLG2)|SMB_FLG2_ERR_STATUS);
		PUTLE4(smbp, SMBO_ERCL, nterr);
	} else {
		if (class == 0 && nterr != 0)	//Map NT status to Dos error
			cifs_ntStatusToDosErrMapping(nterr, &class, &err);
		PCAR(smbp, SMBO_ERCL, class);
		PUTLE2(smbp, SMBO_ERR, err);
	}
}


/*-----------------------------------------------------------------
 * Name			: cifs_initXmtBuffer
 *
 * Description	: Init a xmt buffer for outgoing message.
 * 				  If the buffer is not allocated, allocate one first.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/15/2002
 *
 *-----------------------------------------------------------------
 */
pkt_desc_t *
cifs_initXmtBuffer(cifs_operation *opp, pkt_desc_t *pktp, uchar8 *osmbp)
{

	if (pktp == NULL)
	{
		pktp = cifs_getClientBuf(opp->clp, &osmbp);
		if (pktp == NULL)
		{
			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_initXmtBuffer: no xmt buffer"));
		
			//Wait for an outgoing buffer to be freed to process the command
			DLL_INSERT_TAIL(&opp->clp->waitOutBufOps, opp, entry);
			return (NULL);
		}
	}

	//Chain outgoing packet descriptors, because there may be
	//multiple packets sent for one response
	pktp->pNext				= opp->respMsg.tplPktDescp;
	opp->respMsg.tplPktDescp = pktp;

	opp->respMsg.smbp		= osmbp;
	opp->respMsg.smbLen		= SMB_S_HEADER;
	opp->respMsg.cmdTypep	= osmbp + SMBO_CMD;
	opp->respMsg.cmdp		= osmbp + SMBO_WCNT;

	return (pktp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_processCmd 
 *
 * Description	: process a CIFS command
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
static void
cifs_processCmdInThread(esm_handle_t handle, void *info)
{
	cifs_operation  *opp = (cifs_operation*)info;
	
	if ((cifs_msg_description[*(opp->reqMsg.cmdTypep)].processFunction)
			(opp->clp, opp->ssp, opp->siup,	opp, opp->reqMsg.cmdp,
			 opp->respMsg.cmdp)	== CIFS_SENDRESPNOW)
		cifs_sendClientMsg(opp->clp, opp);
}

		
/*-----------------------------------------------------------------
 * Name			: cifs_processCmd 
 *
 * Description	: process a CIFS command
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32
cifs_processCmd(cifs_client *clp, cifs_operation *opp, uchar8 type)
{
	uchar8 *ismbp = opp->reqMsg.smbp;
	struct cifs_msg_description *msgDesc = &cifs_msg_description[type];
	cifs_session *ssp = NULL;
	cifs_shareInUse *siup = NULL;
	
    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processCmd: "
				"checking cmd... flags=%x", msgDesc->flags));
	
	if (msgDesc->processFunction == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_processCmd: "
					"unsupported cmd %x", type));
		if (msgDesc->name == NULL)
			cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_smbcmd, 
							CIFSTATUS_NOT_IMPLEMENTED);
		else
			cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_nosupport, 
							CIFSTATUS_NOT_SUPPORTED);
		return CIFS_SENDRESPNOW;
	}
	
	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_processCmd: "
				"%s cmd", msgDesc->name));

	if ((opp->ssp == NULL) && (msgDesc->flags & MF_SESSION))
	{	//Look for the client session linked with the uid
    	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_processCmd: finding uid..."));
		
		ssp = cifs_findClientSession(clp, GETLE2(ismbp, SMBO_UID));
		if (ssp == NULL)
		{
            cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_baduid, 
							CIFSTATUS_UNSUCCESSFUL);
			return CIFS_SENDRESPNOW;
		}
		opp->ssp = ssp;
		//Set unicode flag in SMB header if client supports it
		if (clp->capabilities & CAP_UNICODE)
		{
			ushort16 flags2;
			flags2 = GETLE2(opp->respMsg.smbp, SMBO_FLG2) | SMB_FLG2_UNICODE;
			PUTLE2(opp->respMsg.smbp, SMBO_FLG2, flags2);
		}
	}
	else
		ssp = opp->ssp;

	if ((opp->siup == NULL) && ((msgDesc->flags & MF_SIU) == MF_SIU))
	{	//Look for the share in use linked with the tid
			
	    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processCmd: finding tid..."));
		
		siup = cifs_findClientSessionShareInUse(clp,
												ssp,
												GETLE2(ismbp, SMBO_TID));
		if (siup == NULL)
		{
			// FIXME? Samba maps ERRinvtid to STATUS_NETWORK_NAME_DELETED
            cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_invtid, 
							CIFSTATUS_UNSUCCESSFUL);
			return CIFS_SENDRESPNOW;
		}
		opp->siup = siup;
	}
	else
		siup = opp->siup;

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processCmd: "
				"checking write access..."));
	
	if ((msgDesc->flags & MF_WR) && !(siup->access & SEC_STDR_WRITE_ACCESS))
	{	
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_processCmd: wr access denied"));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_access, CIFSTATUS_ACCESS_DENIED);
		return CIFS_SENDRESPNOW;
	}

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processCmd: checking ipc..."));
	
	if (!(msgDesc->flags & MF_IPCUSE) 
	 && (siup != NULL)
	 && (siup->flags & CIFS_SIU_ISIPC))
	{	
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_processCmd: non IPC cmd %d", type));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_access, CIFSTATUS_ACCESS_DENIED);
		return CIFS_SENDRESPNOW;
	}
	if ((msgDesc->flags & MF_WOPLOCK) /* && ??? */)
	{	//Queue operation to process it after oplock break completion
		opp->status = cifs_opstatus_waitOplock;
		//TODO 
	}
			
	if ((msgDesc->flags & MF_NOTHREAD))
	{
	    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processCmd: "
					"calling processing function"));
		return ((*msgDesc->processFunction)
					(clp, ssp, siup, opp, opp->reqMsg.cmdp, opp->respMsg.cmdp));
	}

    opp->threadId = thread_create(cifs_processCmdInThread);
    thread_start(&(opp->threadId), cifs_processCmdInThread);
    CIFS_PRINT(CIFS_DBGLV_DBG3,	("cifs_processCmd:"
				" thread %u created to process command", opp->threadId));
	return (CIFS_DONTSENDRESP);	
}


/*-----------------------------------------------------------------
 * Name			: cifs_processNextCmd 
 *
 * Description	: process the next batched CIFS command
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *
 *-----------------------------------------------------------------
 */
int32
cifs_processNextCmd(cifs_operation *opp, uchar8 *ocmdp)
{
	uchar8		cmd;
	ushort16	offset;
		
	//TODO: Handle fragmented request message somewhere inside this function.

	//Move command and command type pointers to the next command request	
	opp->reqMsg.cmdTypep = opp->reqMsg.cmdp + 1;
	opp->reqMsg.cmdp	 = opp->reqMsg.smbp
							+ GETLE2(opp->reqMsg.cmdp, SMBCO_WP1);
	cmd = *opp->reqMsg.cmdTypep;
	
	//Calculate the offset of the next command response
	offset  = ocmdp - opp->respMsg.smbp;
	offset += 3 + (GCAR(ocmdp, 0) * 2) + GETLE2(ocmdp, SMBCO_BCNT(ocmdp));
	offset += offset % 2;				//Align on word boundary
			
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_processNextCmd: offset=%x", offset));
	
	PUTLE2(ocmdp, SMBCO_WP1, offset);	//Store it
	
	//Move command and command type pointers to the next command response
	opp->respMsg.cmdTypep	= ocmdp + 1;
	*opp->respMsg.cmdTypep	= cmd;
	opp->respMsg.cmdp		= opp->respMsg.smbp + offset;
	
	cifs_setCmdWordByteCnt(&opp->respMsg, 0, 0, FALSE);

	//Process the command
	return (cifs_processCmd(opp->clp, opp, cmd));
}



/*-----------------------------------------------------------------
 * Name			: cifs_defragMsg 
 *
 * Description	: Defragment a CIFS message
 * 				  The first fragment contains Session Header (Netbios
 * 				  or SMB over TCP header) and first CIFS msg part.
 * 				  Following fragments contain only part of the CIFS msg
 *				  We use the first packet received to defragment.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifs_defragMsg(cifs_client* clp, pkt_desc_t *imsgp, char *ismbp, uint32 ismbl)
{
	pkt_desc_t *currFragp;
	uchar8 *datp;
	int32 len;

	CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_defragMsg: "
				"pHdr=%p, hdrSize=%u, ismbp=%p, ismbl=%u",
				imsgp->pHdr, imsgp->hdrSize, ismbp, ismbl));
	
	//The msg size shall not exceed the maximum size of our internal buffers
	if (ismbl > CIFS_S_BUFFER)
		return (NFX_ERR);

	//assign datp at the end of the first fragment data
	datp  = imsgp->pHdr + imsgp->hdrSize;
	
	//substract the first CIFS fragment size from the total size to get
	len = ismbl - (imsgp->hdrSize - ((address_t)(ismbp - imsgp->pHdr)));
		
	currFragp = imsgp->pNext;
			
	//Add following fragments
	while (currFragp != NULL && len > 0)
	{
		CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_defragMsg: "
					"datp=%p, len=%u, pHdr=%p, hdrSize=%u",
					datp, len, currFragp->pHdr, currFragp->hdrSize));
	
		memcpy(datp, currFragp->pHdr, currFragp->hdrSize);
		datp  += currFragp->hdrSize;
		len -= currFragp->hdrSize;
		currFragp = currFragp->pNext;	
	}
	
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_defragMsg: "
				"currFragp=%p len=%u", currFragp, len));
	
	if (currFragp == NULL && len == 0)
	{	//Fragmented message is OK. Free the fragments except the first one
		//since the message is now in it
		if (imsgp->pNext != NULL)
			cifs_freeInBuf(clp, &imsgp->pNext);
		
		//Update hdrSize
		imsgp->hdrSize = ismbl + ((address_t)(ismbp - imsgp->pHdr));
		
		//TODO Check if we need to set pktSize at its real value
		
		IFDBG(CIFS_DBGLV_DBG3)
		{
			printf("cifs_defragMsg: defragmented msg:\n");
			memDump(ismbp, ismbl);
		}
		
		return (NFX_OK);
	}
	else
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_defragMsg: "
					"Fragmented message corrupted"));
		return (NFX_ERR);
	}
}


/*-----------------------------------------------------------------
 * Name			: cifs_processMsg 
 *
 * Description	: CIFS messages reception 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_processMsg(cifs_client* clp,
				pkt_desc_t *imsgp,
				uchar8 *ismbp,
				uint32 ismbl)
{
	uchar8	type = GCAR(ismbp, SMBO_CMD);
	cifs_operation *opp;
	uchar8 *osmbp;
	pkt_desc_t *pktp;
	
	//Check SMB header
	if (memcmp(ismbp, smb_headerStr, sizeof(smb_headerStr)) != 0) 
	{
		CIFS_PRINT(CIFS_DBGLV_WARN, ("cifs_processMsg: no SMB header"));
		goto MsgError;
	}
	
	//Check if Message is fragmented
	if (ismbl > (imsgp->hdrSize - ((address_t)((char*)ismbp - imsgp->pHdr))))
	{
	    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processMsg: fragmented SMB"));
		if ((cifs_msg_description[ismbp[SMBO_CMD]].flags & MF_FRAGSUP) == 0
		 || imsgp->hdrSize < CIFS_S_MINFRAGMENT)
		{	//SMB command does not support fragmented message, or first 
			//fragment size is too small to decode headers. Defragment.
			if (cifs_defragMsg(clp, imsgp, ismbp, ismbl) != NFX_OK)
				goto MsgError;
		}
	} 
	if (ismbl < CIFS_S_BUFFER - 2)
	{	//Make sure there is at least 2 zeros after the message 
		//(from an NTCreateNX without 0 at the end of the file name)
		ismbp[ismbl]	= 0;
		ismbp[ismbl+1]	= 0;
	}
		
	//Create an operation
	opp = cifs_allocOperation();
	if (opp == NULL) {
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_processMsg: no operation obj"));
		
		//Say to the client we are currently out of resource
		pktp = cifs_getClientBuf(clp, &osmbp);
		if (pktp == NULL)
			goto MsgError;
		cifs_initSMBHeaderFromIncomingSMB(osmbp, ismbp);
		PCAR(osmbp, SMBO_ERCL, CIFS_ERRSRV);
		PUTLE2(osmbp, SMBO_ERR, CIFSERR_noresource);
		if (cifs_sendClientMsgWithoutOp(clp, pktp, SMB_S_HEADER+3) != NFX_OK)
			goto MsgError;
		cifs_freeInBuf(clp, &imsgp);
		return;
	}
	opp->clp				= clp;
	opp->reqMsg.tplPktDescp	= imsgp;
	opp->reqMsg.smbp		= ismbp;
	opp->reqMsg.smbLen		= ismbl;
	opp->reqMsg.cmdTypep	= ismbp + SMBO_CMD;
	opp->reqMsg.cmdp		= ismbp + SMBO_WCNT;
	opp->pid				= GETLE2(ismbp, SMBO_PID);
	opp->pid			   += GETLE2(ismbp, SMBO_PIDH) << 16;
	opp->mid				= GETLE2(ismbp, SMBO_MID);
	
	//Get a buffer to send out
	if (cifs_initXmtBuffer(opp, NULL, NULL) == NULL)
	{
		opp->status = cifs_opstatus_waitOutBuffer;
		return;
	}

	//Fill response message header
	cifs_initSMBHeaderFromIncomingSMB(opp->respMsg.smbp, ismbp);
	
	//We are ready to send any answer back, process the command
	if (cifs_processCmd(clp, opp, type) == CIFS_SENDRESPNOW)
		cifs_sendClientMsg(clp, opp);

	return;
	
MsgError:	
	cifs_freeInBuf(clp, &imsgp);
	cifs_abortClientConnection(clp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_procMsgWithXmtBuf 
 *
 * Description	: Process a CIFS message queued because we had
 * 				  no outgoing buffer when we received it
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/14/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_procMsgWithXmtBuf(	cifs_client* clp,
						cifs_operation *opp,
						pkt_desc_t *pktp,
						uchar8 *osmbp)
{
	uchar8	type = GCAR(opp->reqMsg.smbp, SMBO_CMD);
	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_procMsgWithXmtBuf: opp=%p", opp));
	
	//Init the buffer to send out
	cifs_initXmtBuffer(opp, pktp, osmbp);

	//Fill response message header
	cifs_initSMBHeaderFromIncomingSMB(osmbp, opp->reqMsg.smbp);
	
	//We are ready to send any answer back, process the command
	if (cifs_processCmd(clp, opp, type) == CIFS_SENDRESPNOW)
		cifs_sendClientMsg(clp, opp);
}


