/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-encdec.h,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 * 
 * Name			: cifs-encdec.h
 *
 * Description	: CIFS encoding and decoding functions header
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */
#ifndef _CIFS_ENCDEC_H
#define _CIFS_ENCDEC_H

// Alignments in CIFS response message
#define CIFS_ALIGN_16		2		// Align on 16 bits boundary
#define CIFS_ALIGN_32		4		// Align on 32 bits boundary
#define CIFS_ALIGN_64		8		// Align on 64 bits boundary

#define CIFS_SENDRESPNOW	1
#define	CIFS_DONTSENDRESP	0

#define cifs_initNXCmdHeader(cmdp)	(*(((char*)cmdp)+1) = 0xFF)

#define cifs_msgFragSize(fragp)	((fragp)->tplPktDescp->hdrSize \
		- ((address_t)((char*)((fragp)->smbp) - (fragp)->tplPktDescp->pHdr)))

#define cifs_msgOffsetIsInFirstFragment(fragp, offset)\
	((offset) <= cifs_msgFragSize(fragp))
		
#define cifs_msgIsFragmented(msgp)	\
	((msgp)->smbLen > cifs_msgFragSize(msgp))
		

/*-----------------------------------------------------------------
 * Name			: cifs_initSMBHeader 
 *
 * Description	: Initialize the header of an SMB packet and the
 * 					associated SMB msg descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/07/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_initSMBHeader(cifs_cmdDesc *cmdp);

		
/*-----------------------------------------------------------------
 * Name			: cifs_initSMBHeaderFromIncomingSMB
 *
 * Description	: Initialize the header of an SMB packet and the
 * 					associated SMB msg descriptor
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 2/14/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_initSMBHeaderFromIncomingSMB(uchar8 *osmbp, uchar8 *ismbp);


/*-----------------------------------------------------------------
 * Name			: cifs_setCmdWordCnt 
 *
 * Description	: Set command word count and update msg length
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/27/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_setCmdWordCnt(cifs_cmdDesc *cmdp, uint32 wordCnt, boolean init);


/*-----------------------------------------------------------------
 * Name			: cifs_setCmdByteCnt 
 *
 * Description	: Set command byte count and update msg length
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/27/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_setCmdByteCnt(cifs_cmdDesc *cmdp, uint32 byteCnt, boolean init);


/*-----------------------------------------------------------------
 * Name			: cifs_setCmdWordByteCnt 
 *
 * Description	: Set command word and byte count and update msg length
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_setCmdWordByteCnt(cifs_cmdDesc *cmdp, uint32 wordCnt, uint32 byteCnt,
						boolean init);


/*-----------------------------------------------------------------
 * Name			: cifs_errResp 
 *
 * Description	: fill error field of the response
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_errResp(cifs_operation *opp, uchar8 class, ushort16 err,
				uint32 nterr);


/*-----------------------------------------------------------------
 * Name			: cifs_processCmd 
 *
 * Description	: process a CIFS command
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32	//CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifs_processCmd(cifs_client *clp, cifs_operation *opp, uchar8 type);


/*-----------------------------------------------------------------
 * Name			: cifs_processNextCmd 
 *
 * Description	: process the next batched CIFS command
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *
 *-----------------------------------------------------------------
 */
int32
cifs_processNextCmd(cifs_operation *opp, uchar8 *ocmdp);


/*-----------------------------------------------------------------
 * Name			: cifs_initXmtBuffer
 *
 * Description	: Init a xmt buffer for outgoing message.
 * 				  If the buffer is not allocated, allocate one first.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/15/2002
 *
 *-----------------------------------------------------------------
 */
pkt_desc_t *
cifs_initXmtBuffer(cifs_operation *opp, pkt_desc_t *pktp, uchar8 *osmbp);


/*-----------------------------------------------------------------
 * Name			: cifs_processMsg 
 *
 * Description	: process a CIFS message
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_processMsg(cifs_client* clp,
				pkt_desc_t *imsgp,
				uchar8 *ismbp,
				uint32 ismbl);


/*-----------------------------------------------------------------
 * Name			: cifs_defragMsg 
 *
 * Description	: Defragment a CIFS message
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/08/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifs_defragMsg(cifs_client* clp, pkt_desc_t *imsgp, char *ismbp, uint32 ismbl);


/*-----------------------------------------------------------------
 * Name			: cifs_procMsgWithXmtBuf 
 *
 * Description	: Process a CIFS message queued because we had
 * 				  no outgoing buffer when we received it
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/14/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_procMsgWithXmtBuf(	cifs_client* clp,
						cifs_operation *opp,
						pkt_desc_t *pktp,
						uchar8 *osmbp);
		
#endif

