/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-msglb.c,v 1.2 2002/12/10 18:13:18 laurentb Exp $
 *
 * Name			: cifs-msg.c
 *
 * Description	: Contains CIFS encoding and decoding functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-string.h"
#include "cifs-err.h"
#include "cifs-msg.h"
#include "cifs-ipc.h"
#include "cifs-intmsg.h"
#include "cifs-filetools.h"
#include "cifs-filereqs.h"


#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)

//Unicode strings


/*-----------------------------------------------------------------
 * Name			: cifsm_createDir
 *
 * Description	: create directory (0)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/21/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_createDir(	cifs_client *clp,
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	uni16 dir[FS_NAME_MAX];		//Directory to create
	uint32 rc;
	fs_Status fsStatus;
	fs_Attr	fsAttr;
	fs_MsgMclResp *MCLResp;
	fs_MsgMkdirResp *mkdirResp;
	eee_desc_t *eDescp;

	//check share access
	rc = cifs_secCheckShareAccess(siup, SEC_GENR_FILE_WRITE);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Send MCL request to file system
	rc = cifs_fsMCLReqFromMsg(	opp, 
								NULL,
								opp->reqMsg.smbp,
								SMBCP_BPAR(icmdp) + 1,
								dir,
								FS_POST_ATTRS);//TODO request SD
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MCL, (void**)&MCLResp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	
	fsStatus = MCLResp->hdr.fsStatus;
	if (fsStatus == FS_ERR_NONE)	
		fsStatus = FS_ERR_EXIST;	//last elem already exists, it is an error
	if (fsStatus != FS_ERR_NOENT)	
		cifsm_testFsStatus_ifErrReturn(opp, fsStatus, eDescp);

CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_createDir: numAttrs=%d, handle:",
				MCLResp->numAttrs));
cifs_decodeFileHandle(&MCLResp->handle);
CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_createDir: parentHandle:"));
cifs_decodeFileHandle(&MCLResp->parentHandle);

	//Path is valid and last elem does not exist, check write access on parent
	//TODO check SD

	//Request directory creation
	cifs_initFsAttr(&fsAttr);
	fsAttr.type = FS_FTYPE_DIR;
	fsAttr.mode = FS_IFMT_DIR | FS_IMODE_RWXUGO; 
	rc = cifs_fsMkdirReq(opp, 
						opp->reqMsg.smbp,
						&MCLResp->handle,//handle of the parent directory
						dir,
						&fsAttr,
						0);
	//free MCL response 
	eee_freePacket(eDescp);
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MKDIR, (void**)&mkdirResp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	cifsm_testFsStatus_ifErrReturn(opp, mkdirResp->hdr.fsStatus, eDescp);
	
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_delDir
 *
 * Description	: delete directory (1)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/26/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_delDir(	cifs_client *clp,
					cifs_session *ssp,
					cifs_shareInUse *siup,
					cifs_operation *opp,
					uchar8 *icmdp,
					uchar8 *ocmdp)
{
	uni16 dir[FS_NAME_MAX];		//Directory to delete
	uint32 rc;
	fs_Status fsStatus;
//	fs_Attr	fsAttr;
	fs_MsgMclResp *MCLResp;
	fssf_removeRsp_t *rmResp;
	eee_desc_t *eDescp;

	//check share access
	rc = cifs_secCheckShareAccess(siup, SEC_STDR_DELETE_ACCESS);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Send MCL request to file system
	rc = cifs_fsMCLReqFromMsg(	opp, 
								NULL,
								opp->reqMsg.smbp,
								SMBCP_BPAR(icmdp) + 1,
								dir,
								FS_POST_PAR_ATTRS);//TODO request SD
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MCL, (void**)&MCLResp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	
	fsStatus = MCLResp->hdr.fsStatus;
	cifsm_testFsStatus_ifErrReturn(opp, fsStatus, eDescp);

CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_delDir: numAttrs=%d, handle:",
				MCLResp->numAttrs));
cifs_decodeFileHandle(&MCLResp->handle);
CIFS_PRINT(CIFS_DBGLV_DBG, ("cifs_delDir: parentHandle:"));
cifs_decodeFileHandle(&MCLResp->parentHandle);
	
	//Path is valid and last elem does exist, check if it is a directory
	
	//Check if delete access is allowed
	//TODO check SD

	//Request directory deletion
	rc = cifs_fssfRemoveReq(opp, 
							opp->reqMsg.smbp,
							&MCLResp->parentHandle,
							dir);

	//free MCL response 
	eee_freePacket(eDescp);
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Wait for the answer
	rc = cifs_waitFssfResp(FSSF_REMOVE_RSP, (void**)&rmResp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	cifsm_testFsStatus_ifErrReturn(opp, rmResp->status, eDescp);
	
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_close 
 *
 * Description	: close (0x4) 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/29/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_close(	cifs_client *clp, 
					cifs_session *ssp,
					cifs_shareInUse *siup,
					cifs_operation *opp,
					uchar8 *icmdp,
					uchar8 *ocmdp)
{
	cifs_fileOpen *fop;
	ushort16 fid;
	int32 lastModTime;
	uint32 rc, i;
	eee_desc_t *eDescp;
	fssf_closeRsp_t *cloResp;

	//retrieve file handle
	fid = GETLE2(icmdp, SMBCO_WP0);
	
	//Check if share is an IPC
	if (siup->flags & CIFS_SIU_ISIPC)
		return cifsm_ipcClose(siup, opp, fid);
		
	fop = cifs_findFileOpenByFid(siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_WARN, ("cifsm_close: unknown fid 0x%x", fid));
		cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_badfid, CIFSTATUS_FILE_CLOSED);
		return (CIFS_SENDRESPNOW);
	}

	//Set the last modification time
	lastModTime = GETLE4S(icmdp, SMBCO_WP1);
	if (lastModTime != 0 && lastModTime != -1)
	{
		fssf_attr_t fssfAttr;
		fssf_setattrRsp_t *setattrResp;
		
		CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_close: set last modif time"));
		cifs_initFssfAttr(&fssfAttr);
		//TODO get Local time difference from UTC. Adjust time accordingly
		fssfAttr.fsAttr.mtime.secs	= lastModTime;
		fssfAttr.fsAttr.mtime.usecs	= 0;
		fssfAttr.bitmap = FSSF_ATTR_TIME_MODIFY;
		do {
			rc = cifs_fssfSetattrByIdReq(opp, fop->stateid, &fssfAttr);		
			if (rc == CIFSTATUS_OK)
			{
				//Wait for the answer
				rc = cifs_waitFsResp(FSSF_SETATTR_RSP,
									(void**)&setattrResp,
									&eDescp);
				if (rc == CIFSTATUS_OK)
				{
					//free received message
					eee_freePacket(eDescp);
					if (setattrResp->status == FS_ERR_NONE)
						break;
				}
			}
			CIFS_PRINT(CIFS_DBGLV_ERR, 
						("cifsm_close: failed to set last modif time"));
		} while(0);
	}

	for (i = 0; i < CIFS_N_FS_RETRIES; i++)
	{
		//Send close request to file system statefull
		if (cifs_fssfCloseReq(opp, fop->stateid) != CIFSTATUS_OK)
		{
			CIFS_PRINT(CIFS_DBGLV_WARN, 
						("cifsm_close: failed to send FSSF close"));
			cifs_waitResp(0, (void**)&cloResp, &eDescp, TRUE,CIFS_T_FS_RETRIES);
			continue;
		}

		//Wait for the answer
		rc = cifs_waitFsResp(FSSF_CLOSE_RSP, (void**)&cloResp, &eDescp);
		if (rc == CIFSTATUS_OK)
			break;
	}
	
	//close the file (anyway !)
	cifs_closeShareInUseFileOpen(siup, fop);
	
	//test if we got an answer from FSSF
	cifs_testWaitResp_ifErrReturn(opp, rc);
	
	//test FSSF answer
	cifsm_testFsStatus_ifErrReturn(opp, cloResp->status, eDescp);

	//free received message
	eee_freePacket(eDescp);
		
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_delete
 *
 * Description	: delete file (0x6)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/08/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_delete(	cifs_client *clp,
					cifs_session *ssp,
					cifs_shareInUse *siup,
					cifs_operation *opp,
					uchar8 *icmdp,
					uchar8 *ocmdp)
{
	uni16 file[FS_NAME_MAX];		//file to delete
	uint32 rc;
	fs_Status fsStatus;
	fs_MsgMclResp *MCLResp;
	fssf_removeRsp_t *rmResp;
	eee_desc_t *eDescp;

	//check share access
	rc = cifs_secCheckShareAccess(siup, SEC_STDR_DELETE_ACCESS);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Send MCL request to file system
	rc = cifs_fsMCLReqFromMsg(	opp, 
								NULL,
								opp->reqMsg.smbp,
								SMBCP_BPAR(icmdp) + 1,
								file,
								FS_POST_ATTRS);//TODO request SD
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MCL, (void**)&MCLResp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	
	fsStatus = MCLResp->hdr.fsStatus;
	cifsm_testFsStatus_ifErrReturn(opp, fsStatus, eDescp);
	
	//Path is valid and last elem does exist, check if it is a file and if
	//its attributes match the search attributes
	//TODO
	
	//Check if delete access is allowed
	//TODO check SD

	//Request file deletion
	rc = cifs_fssfRemoveReq(opp, 
							opp->reqMsg.smbp,
							&MCLResp->parentHandle,
							file);

	//free MCL response 
	eee_freePacket(eDescp);
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Wait for the answer
	rc = cifs_waitFssfResp(FSSF_REMOVE_RSP, (void**)&rmResp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	cifsm_testFsStatus_ifErrReturn(opp, rmResp->status, eDescp);
	
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_checkPath
 *
 * Description	: check directory (0x10)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/19/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_checkPath(	cifs_client *clp,
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	int32 rc;
	fs_MsgMclResp *resp;
	eee_desc_t *eDescp;

	//Send MCL request to file system
	rc = cifs_fsMCLReqFromMsg(	opp, 
								NULL,
								opp->reqMsg.smbp,
								SMBCP_BPAR(icmdp) + 1,
								NULL,
								0);
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MCL, (void**)&resp, &eDescp);
	cifs_testWaitResp_ifErrReturn(opp, rc);
	cifsm_testFsStatus_ifErrReturn(opp, resp->hdr.fsStatus, eDescp);
	
	//free received message
	eee_freePacket(eDescp);
	
	return (CIFS_SENDRESPNOW);
}











int32 cifsm_open(	cifs_client		*clp, 
					cifs_session	*ssp,
					cifs_shareInUse *siup,
					cifs_operation	*opp,
					uchar8			*icmdp,
					uchar8			*ocmdp)
{
    cifs_errResp(opp, 0, 0, CIFSTATUS_NOT_IMPLEMENTED);
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_create(	cifs_client		*clp, 
					cifs_session	*ssp,
					cifs_shareInUse *siup,
					cifs_operation	*opp,
					uchar8			*icmdp,
					uchar8			*ocmdp)
{
    cifs_errResp(opp, 0, 0, CIFSTATUS_NOT_IMPLEMENTED);
	return (CIFS_SENDRESPNOW);
}

int32 cifsm_createNew(	cifs_client		*clp, 
						cifs_session	*ssp,
						cifs_shareInUse *siup,
						cifs_operation	*opp,
						uchar8			*icmdp,
						uchar8			*ocmdp)
{
    cifs_errResp(opp, 0, 0, CIFSTATUS_NOT_IMPLEMENTED);
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_openNX 
 *
 * Description	: open and X (0x2D)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_openNX(	cifs_client		*clp, 
					cifs_session	*ssp,
					cifs_shareInUse *siup,
					cifs_operation	*opp,
					uchar8			*icmdp,
					uchar8			*ocmdp)
{
	uni16 path[CIFS_S_PATHSTR+1];
	ushort16 openFlags, openMode, fileAttr, openFunc;
	uint32 fileSize, desiredAccess, shareAccess, fssfOpenFlags, fssfAction;
	uint32 rc;
	fs_Handle *handlep;
	fs_Attr	fsAttr;
	fs_MsgMclResp *mclResp;
	eee_desc_t *eDescp;
	fs_Status fsStatus;
	fssf_action_t fssfOpenAction;
	fssf_reqAccess_t fssfReqAccess;
	cifs_fileOpen *fop;
	
	//Check if share is an IPC
	if (siup->flags & CIFS_SIU_ISIPC)
		return (cifsm_ipcOpenNX(clp, ssp, siup, opp, icmdp, ocmdp));
	
	openFlags	= GETLE2(icmdp, SMBCO_WP2);
	openMode	= GETLE2(icmdp, SMBCO_WP3);
	fileAttr	= GETLE2(icmdp, SMBCO_WP5);
	openFunc	= GETLE2(icmdp, SMBCO_WP8);
	fileSize	= GETLE4(icmdp, SMBCO_WP9);

	CIFS_PRINT(CIFS_DBGLV_DBG2, 
		("cifsm_openNX: oFlags=0x%x oMode=0x%x fAttr=0x%x oFunc=0x%x fSize=%d", 
		 openFlags, openMode, fileAttr, openFunc, fileSize));
	
	//check share access
	desiredAccess = cifs_getDesiredAccessFromOpenMode(openMode);
	rc = cifs_secCheckShareAccess(siup, desiredAccess);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Extract path from CIFS message
	rc = cifs_Smb2UniPath(opp->reqMsg.smbp, SMBCP_BPAR(icmdp), path);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	//Send MCL request to file system
	rc = cifs_fsMCLReq(	opp, 
						NULL,
						opp->reqMsg.smbp,
						path,
						FS_POST_ATTRS);//TODO ask SD
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MCL, (void**)&mclResp, &eDescp);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	fsStatus = mclResp->hdr.fsStatus;
	if (fsStatus == FS_ERR_NONE)		
	{	//file exists
		handlep = &mclResp->parentHandle;//take parent handle
		//TODO check access against file SD
	} else {
		if (fsStatus == FS_ERR_NOENT)
		{	//file does not exist, parent does
			handlep = &mclResp->handle;	//take handle which is the parent handle
			//TODO check access against parent directory SD
		}
		else
			cifsm_testFsStatus_ifErrReturn(opp, fsStatus, eDescp);
	}

	//Convert all attributes, options & flags to FSSF style
	
	//Convert Requested Access
	cifs_buildFssfReqAccessFromDesiredAccess(fssfReqAccess, desiredAccess);
	//Convert share access
	shareAccess = cifs_getFssfShareAccessFromOpenMode(openMode);	
	//Convert File attributes
	cifs_initFsAttr(&fsAttr);
	fsAttr.mode = FS_IMODE_RWXUGO; 
	fsAttr.simpleAttrs = fileAttr & FS_INODE_SATTR_CIFSMASK;
	if (fileSize != 0)
		fsAttr.size = fileSize;
	//Build open action
	fssfOpenAction.ifExist	 = cifs_getOpenActionIfExistsFromOFunc(openFunc);
	fssfOpenAction.ifNotExist= cifs_getOpenActionIfNotExistsFromOFunc(openFunc);
	//Build open flags 
	fssfOpenFlags = 0;
	
	//Open file & directory
	fop = cifs_openFile(opp,
						handlep,
						path,
						fssfOpenAction, 
						fssfReqAccess,
						shareAccess,
						fssfOpenFlags,
						&fsAttr,
						&fssfAction);

	//free MCL response 
	eee_freePacket(eDescp);

	if (fop == NULL)
		return (CIFS_SENDRESPNOW);

	//Build the response
			
	openMode &= 0x7F;
	openFunc = fssfAction2createAction[fssfAction & 0x7];
	
	cifs_setCmdWordByteCnt(&opp->respMsg, 15, 0, TRUE);
	cifs_initNXCmdHeader(ocmdp);
	PUTLE2(ocmdp, SMBCO_WP2, fop->fid);						//file ID
	cifs_fillSmbFileAttr(ocmdp+SMBCO_WP3, &fsAttr, TRUE);	//file attributes
	//TODO add Time...
	PUTLE4(ocmdp, SMBCO_WP6, fsAttr.size);					//file size
	PUTLE2(ocmdp, SMBCO_WP8, openMode);						//openMode
	PUTLE2(ocmdp, SMBCO_WP11, openFunc);					//action taken

	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return (cifs_processNextCmd(opp, ocmdp));
	else
		return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ntCreatNX 
 *
 * Description	: NT Create and X (0xA2)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/29/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ntCreatNX(	cifs_client		*clp, 
						cifs_session	*ssp,
						cifs_shareInUse *siup,
						cifs_operation	*opp,
						uchar8			*icmdp,
						uchar8			*ocmdp)
{
	uni16 path[CIFS_S_PATHSTR+1];
	uchar8 *smbp = opp->reqMsg.smbp;
	ushort16 rootDirFid;
	uint32 createFlags, desiredAccess, fileAttr, shareAccess, createDispo;
	uint32 createOptions, fssfOpenFlags, rc;
	uint64 fileAllocSize;
	fs_Attr	fsAttr;
	fs_Handle *handlep;
	fs_MsgMclResp *mclResp;
	eee_desc_t *eDescp;
	fs_Status fsStatus;
	boolean isDir = FALSE;
	fssf_action_t fssfOpenAction;
	fssf_reqAccess_t fssfReqAccess;
	cifs_fileOpen *fop;
	
	//Check if share is an IPC
	if (siup->flags & CIFS_SIU_ISIPC)
		return cifsm_ipcNtCreateNX(clp, ssp, siup, opp, icmdp, ocmdp);
	
	createFlags		= GETLE4(icmdp, SMBCO_NTCREATE_CREATEFLAGS);
	rootDirFid		= GETLE2(icmdp, SMBCO_NTCREATE_ROOTDIRFID);
	desiredAccess	= GETLE4(icmdp, SMBCO_NTCREATE_DESIREDACC);
//Warning on 32bits	fileAllocSize	= GETLE8(icmdp, SMBCO_NTCREATE_ALLOCSIZE);
fileAllocSize =  GETLE4(icmdp, SMBCO_NTCREATE_ALLOCSIZE);
	fileAttr		= GETLE4(icmdp, SMBCO_NTCREATE_FILEATTR);
	shareAccess		= GETLE4(icmdp, SMBCO_NTCREATE_SHAREACC);
	createDispo		= GETLE4(icmdp, SMBCO_NTCREATE_CREATEDISPO);
	createOptions	= GETLE4(icmdp, SMBCO_NTCREATE_CREATEOPTS);

	CIFS_PRINT(CIFS_DBGLV_DBG2, 
			("cifsm_ntCreatNX: crFlags=0x%x rootFid=0x%x desiredAcc=0x%x\n"
			 "\tfileAlloc=%lld fileAttr=0x%x shAcc=0x%x crDispo=%d crOpts=0x%x",
			 createFlags, rootDirFid, desiredAccess, fileAllocSize, fileAttr,
			 shareAccess, createDispo, createOptions));
	
	//check share access
	rc = cifs_secCheckShareAccess(siup, desiredAccess);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	
	if (rootDirFid)
	{	//concatenate directory path and file relative path
		uint32 len;
		
		fop = cifs_findFileOpenByFid(siup, rootDirFid);
		if (fop == NULL)
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_FILE_CLOSED);
		unistrcpy(path, fop->pathName);
		len = unistrlen(path);
		if (len >= CIFS_S_PATHSTR - 2)
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NAME_TOO_LONG);
		if (path[len-1] != '\\')
			path[len++]	= '\\';
	    if (!cifs_getSmbUniStr(	path + len, 
								SMBCP_BPAR(icmdp),
								CIFS_S_PATHSTR - len,
								smbp))
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NAME_TOO_LONG);
	} 
	else {
		
		//Extract path from CIFS message
		rc = cifs_Smb2UniPath(smbp, SMBCP_BPAR(icmdp), path);
		cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	}

	//Send MCL request to file system
	rc = cifs_fsMCLReq(	opp, 
						NULL,
						opp->reqMsg.smbp,
						path,
						FS_POST_ATTRS);//TODO ask SD
	
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//Wait for the answer
	rc = cifs_waitFsResp(FS_REQ_MCL, (void**)&mclResp, &eDescp);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	fsStatus = mclResp->hdr.fsStatus;
	if (fsStatus == FS_ERR_NONE)		
	{	//file exists
		handlep = &mclResp->parentHandle;//take parent handle
		isDir = (mclResp->attr[0].type == FS_FTYPE_DIR);
		//TODO check access against file SD
	} else {
		if (fsStatus == FS_ERR_NOENT)
		{	//file does not exist, parent does
			handlep = &mclResp->handle;	//take handle which is the parent handle
			//TODO check access against parent directory SD
		}
		else
			cifsm_testFsStatus_ifErrReturn(opp, fsStatus, eDescp);
	}

	//Convert all attributes, options & flags to FSSF style
	//Convert Requested Access
	cifs_buildFssfReqAccessFromDesiredAccess(fssfReqAccess, desiredAccess);
	//Convert File attributes
	cifs_initFsAttr(&fsAttr);
	fsAttr.mode = FS_IMODE_RWXUGO; 
	if (isDir
	 || (fileAttr & SEC_FILE_ATTRIBUTE_DIRECTORY)
	 || (createOptions & SEC_FILE_DIRECTORY_FILE))
	{
		fsAttr.type = FS_FTYPE_DIR;
		fsAttr.mode |= FS_IFMT_DIR;
	} else 
		fsAttr.type = FS_FTYPE_REG;
	fsAttr.simpleAttrs = fileAttr & FS_INODE_SATTR_CIFSMASK;
	if (fileAllocSize != 0)
		fsAttr.size = fileAllocSize;
	//Build open action
	if (createDispo >= SEC_FILE_CREATEDISPO_MAX)
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_PARAMETER);
	fssfOpenAction.ifExist	  = createDispo2fssfAction[createDispo].ifExist;
	fssfOpenAction.ifNotExist = createDispo2fssfAction[createDispo].ifNotExist;
	//Build open flags 
	fssfOpenFlags = 0;
	if (createOptions & SEC_FILE_DELETE_ON_CLOSE)
		fssfOpenFlags |= FSSF_FLAG_DELETE_ON_CLOSE;
	
	//Open file & directory
	fop = cifs_openFile(opp,
						handlep,
						path,
						fssfOpenAction, 
						fssfReqAccess,
						shareAccess,
						fssfOpenFlags,
						&fsAttr,
						&createDispo);	//reuse createDispo to get create action

	//free MCL response 
	eee_freePacket(eDescp);

	if (fop == NULL)
		return (CIFS_SENDRESPNOW);

	//Build the response
	
	createDispo = fssfAction2createAction[createDispo & 0x7];
	isDir		= (fsAttr.type == FS_FTYPE_DIR);
			
	cifs_setCmdWordByteCnt(&opp->respMsg, 34, 0, TRUE);//TODO see what W2K adds
	cifs_initNXCmdHeader(ocmdp);
	PUTLE2(ocmdp, SMBCO_WP2, 0);							//Oplock (NONE)
	PUTLE2(ocmdp, SMBCO_WP2+1, fop->fid);					//file ID
	PUTLE4(ocmdp, SMBCO_WP3+1, createDispo);				//action taken
	PUTLE8(ocmdp, SMBCO_WP5+1, 0);							//creation time
	cifs_fillLongTime(ocmdp+SMBCO_WP9+1,  &fsAttr.atime);	//last access
	cifs_fillLongTime(ocmdp+SMBCO_WP13+1, &fsAttr.mtime);	//last modification 
	cifs_fillLongTime(ocmdp+SMBCO_WP17+1, &fsAttr.ctime);	//last meta modif
	cifs_fillSmbFileAttr(ocmdp+SMBCO_WP21+1,&fsAttr, TRUE);	//file attributes
	PUTLE8(ocmdp, SMBCO_WP23+1, fsAttr.size);				//file allocation
	PUTLE8(ocmdp, SMBCO_WP27+1, fsAttr.size);				//file size
	PUTLE2(ocmdp, SMBCO_WP31+1, SEC_FILE_TYPE_DISK);		//file type
	PUTLE2(ocmdp, SMBCO_WP32+1, 0);							//device state
	PCAR  (ocmdp, SMBCO_WP33+1, isDir);						//isADirectory
	
	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return (cifs_processNextCmd(opp, ocmdp));
	else
		return (CIFS_SENDRESPNOW);
}



