/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-string.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: cifs-string.c
 *
 * Description	: Contains CIFS string tools functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/19/2002
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-err.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)


/*-----------------------------------------------------------------
 * Name			: cifs_spaces
 *
 * Description	: Return a string with spaces according to the level
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/08/2002
 *
 *-----------------------------------------------------------------
 */
char *
cifs_spaces(uchar8 level)
{
static char spaces[] = 
"                                                                          ";

	if (level > 18) 
		level = 18;
	return (spaces + sizeof(spaces)-1 - level * 4);
}


/*-----------------------------------------------------------------
 * Name			: cifs_strcpy
 *
 * Description	: special cifs string copy returning a pointer on
 * 					the address after the ending 0
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/16/2001
 *
 *-----------------------------------------------------------------
 */
char *
cifs_strcpy(char *dstp, const char *srcp)
{
	char8	* dp = dstp;

	if (!dstp)
		return (0);

	*dp = 0;

	if (!srcp)
		return (dstp);

	while ((*dp++ = *srcp++) != 0);

	return (dp);
}

		
/*-----------------------------------------------------------------
 * Name			: cifs_getUniStr
 *
 * Description	: Extract a UNICODE string from smb message
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/01/2002
 *
 *-----------------------------------------------------------------
 */
uchar8	*	//pointer on address after the source string ending 0, NULL if error
cifs_getUniStr(uni16		*dstp,
				uchar8		*srcp,
				ushort16	maxDstLen,	//in unicode characters
				boolean		uniSrc)		//The source string is UNICODE or not
{
	if (uniSrc)
	{
		uni16 *sp;
		uint32 len;
		
		// Source UNICODE string
		srcp += ((address_t)srcp) % 2;	//Unicode string are aligned on word
		sp = (ushort16 *)srcp;
		len = unistrlen(sp);
		if (maxDstLen > len)
		{
			unistrcpy(dstp, sp);
			return ((uchar8*)(sp + len + 1)); 
		} else {
			*dstp = 0;
			return (NULL);
		}
	}
	else 
		// Source is ASCII string
		return (strDos2Uni(dstp, srcp, maxDstLen, TRUE));

}


/*-----------------------------------------------------------------
 * Name			: cifs_putUniStr
 *
 * Description	: Put an UNICODE string into an SMB message 
 * 
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/23/2002
 *
 *-----------------------------------------------------------------
 */
int32			// Space in bytes that destination string including 
				// terminating '\0' occupies in message buffer
cifs_putUniStr(	uchar8	*dstp,	//Destination string in UNICODE or ASCII
				uni16	*srcp,	//Source string in UNICODE
				uchar8	*smbp,	//Destination SMB message pointer
				uint32	*dstStrLenInBytes,
				uint32	maxDestLenInBytes)
{
	int32	dataCnt, pad;
	boolean uniDst, outOfSpace = FALSE;

	uniDst = ((GETLE2(smbp, SMBO_FLG2) & SMB_FLG2_UNICODE) != 0);
	if (uniDst)
	{
		// Put UNICODE string
		pad = ((address_t)dstp) % 2;
		dstp += pad;			//Unicode string are aligned on word

		maxDestLenInBytes -= pad;
		*dstStrLenInBytes = unistrsize(srcp);
		if (*dstStrLenInBytes < maxDestLenInBytes)
		{
			//No enough space to put the whole string
			outOfSpace = TRUE;
		}
		*dstStrLenInBytes = MIN(*dstStrLenInBytes, maxDestLenInBytes);

		//XXX: If no enough space, shoule we copy partial string or nothing?
		dstp = (uchar8 *)unistrncpy((uni16 *)dstp, srcp, *dstStrLenInBytes - 2);
		*dstp = 0;				//Add terminating '\0'
		dataCnt = *dstStrLenInBytes + pad;
	}
	else
	{
		// Put ASCII string
		if (strUni2Dos(dstp, srcp, maxDestLenInBytes) == NULL)
		{
			//No enough space to put the whole string
			outOfSpace = TRUE;
		}
		*dstStrLenInBytes = strlen(dstp) + 1;
		dataCnt = *dstStrLenInBytes;
	}

	if (outOfSpace == TRUE)
		return (-1);
	else
		return (dataCnt);
}


/*-----------------------------------------------------------------
 * Name			: cifs_putStr
 *
 * Description	: Put an ASCII string into an SMB message 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/15/2002
 *
 *-----------------------------------------------------------------
 */
int32			// Space in bytes that destination string including 
				// terminating '\0' occupies in message buffer
cifs_putStr(uchar8	*dstp,	//Destination string in UNICODE or ASCII
			uchar8	*srcp,	//Source string in ASCII
			uchar8	*smbp,	//Destination SMB message pointer
			uint32	*dstStrLenInBytes,
			uint32	maxDestLenInBytes)
{
	int32	len, pad;

	if (GETLE2(smbp, SMBO_FLG2) & SMB_FLG2_UNICODE)
	{			// Put UNICODE string
		pad = ((address_t)dstp) % 2;
		dstp += pad;
		maxDestLenInBytes -= pad;
		if (strDos2Uni((uni16*)dstp, srcp, maxDestLenInBytes, TRUE) == NULL)
		{
			*dstStrLenInBytes = 0;
			return (-1);
		}
		*dstStrLenInBytes = unistrsize((uni16*)dstp);
	} else {	// Put ASCII string
		pad = 0;
		len = strlen(srcp);
		if (len <= maxDestLenInBytes - 1)
		{
			strcpy(dstp, srcp);
			*dstStrLenInBytes = len + 1;
		} else {
			dstp[0] = '\0';
			*dstStrLenInBytes = 0;
			return (-1);
		}
	}

	return (*dstStrLenInBytes + pad);
}


/*-----------------------------------------------------------------
 * Name			: cutString
 *
 * Description	: search and remove special starting and ending motifs
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/19/2002
 *
 *-----------------------------------------------------------------
 */
void
cutString(uni16 *s, uni16 *start, uni16 *end)
{
	uint32 sLen, startLen, endLen;

	if (*s == 0)
		return;
	
	sLen	 = unistrlen (s);
	startLen = start ? unistrlen (start) : 0;
	endLen	 = end   ? unistrlen (end)   : 0;

	if (startLen)
	{
		while (sLen && unistrncmp(s, start, startLen) == 0)
		{
			unistrcpy(s, s + startLen);
			sLen -= startLen;
		}
	}
	if (endLen)
	{
		while (unistrncmp(s+sLen-endLen, end, endLen) == 0)
		{
			s[sLen-endLen] = 0;
			sLen -= endLen;
		}
	}
}


