/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: nfp-objects.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: nfp-objects.c
 *
 * Description	: NFP objects management functions
 * 				  This is the generic version of NFPv1 done by Joseph Vu
 * 
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"


#define	NFP_DBG_PRINTF	printf
//#define	NFP_DBG_PRINTF	if (0) printf	//FOR DEBUGING

#define	NFP_DBG_ON	

//#define NFP_USE_SHMEM
#ifdef  NFP_USE_SHMEM
#define NFP_SHMEM_ALLOC     eee_allocateBlock
#else
#define NFP_SHMEM_ALLOC     eee_ramAlloc
#endif  /*NFP_USE_SHMEM*/

//This module adds an object header to objects managed
typedef struct nfp_objHdr {
  void      *nextFreeObj; 
} nfp_objHdr;



/*-----------------------------------------------------------------
 * Name			: nfp_allocObject 
 *
 * Description	: Allocate NFP object
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *
 *-----------------------------------------------------------------
 */
void *	//Allocated object, NULL if none available
nfp_allocObject(nfp_objList* list, boolean setToZero)
{
	nfp_objHdr	*availObjp;

	//Check for any entries
	if (list->freeObjCnt == 0) {
		printf("nfp_allocObject: no free %s object \n", list->objName);
		return (NULL);
	}

	availObjp = (nfp_objHdr*)list->listFreep;
	//if (availObjp == NULL)	//No object available and list->freeObjCnt != 0
	//	panic("nfp_allocObject: %s object list corrupted\n", list->objName);
	NFX_ASSERT(availObjp != NULL);
	
	//Remove object from the free list
	list->listFreep = availObjp->nextFreeObj;

	list->freeObjCnt--;
	if (setToZero)
		memset(availObjp, 0, list->objSize);
	else 
		availObjp->nextFreeObj = NULL;
	//This way allocated objects have their nextFreeObj equal to NULL

	//Object pointer starts right after object header
	availObjp++;
	
	NFP_DBG_PRINTF("nfp_allocObject: Available %s object at %p\n", 
					list->objName, availObjp);
	
	return ((void*)availObjp);
}


/*-----------------------------------------------------------------
 * Name			: nfp_deallocObject 
 *
 * Description	: Deallocate a NFP object
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR 
nfp_deallocObject(nfp_objList* list, void *object)
{
	nfp_objHdr *objp = (nfp_objHdr*)object;
	
	NFP_DBG_PRINTF("nfp_deallocObject: %s at %p\n",list->objName, objp);
	
	list->freeObjCnt++;
	
	//Object pointer was pointing right after object header
	objp--;
	
	//Insert object at the beginning of the free list
	objp->nextFreeObj = list->listFreep;
	list->listFreep = objp;

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: nfp_validateObject 
 *
 * Description	: Validate a NFP object
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR 
nfp_validateObject(nfp_objList* list, void *object)
{
	char *lastObjectp;
	nfp_objHdr *objp = (nfp_objHdr*)object;

	lastObjectp = (char*)list->listBasep + (list->objSize*(list->nbOfObj-1));
	
	//Object pointer was pointing right after the object header
	objp--;
	
	if ((objp < (nfp_objHdr*)list->listBasep)
	 || (objp > (nfp_objHdr*)lastObjectp)
	 || (objp->nextFreeObj != NULL))
	  return (NFX_ERR);

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name			: nfp_objInit 
 *
 * Description	: Initialize a NFP object list
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *
 *-----------------------------------------------------------------
 */
int32 	//NFX_OK or NFX_ERR
nfp_objInit(nfp_objList* list, char *objName, uint32 objSize, uint32 nbOfObj)
{
	int i;
	char *objp, *nextobjp;
	

	NFP_DBG_PRINTF("nfp_objListInit: %s list: size=%d nb=%d allocsize=%d\n", 
					objName, objSize, nbOfObj, objSize * nbOfObj);
	
	//We add the size of the object header
	objSize += sizeof(nfp_objHdr);
	
	list->listBasep = (void*)NFP_SHMEM_ALLOC(objSize * nbOfObj);
	//if (list->listBasep == NULL)
	//	panic("nfp_objList_init: Failed to alloc %s list\n", objName);
	NFX_ASSERT(list->listBasep != NULL);
	
	//NB: Each object will be initialized to 0 at allocation time
	
	list->listFreep		= list->listBasep;
	list->freeObjCnt	= nbOfObj;
	list->objSize		= objSize;
	list->nbOfObj		= nbOfObj;
	strncpy(list->objName, objName, NFP_S_OBJECTNAME);
		
	// Initialize objects header
	objp = list->listBasep;
	for (i=0; i<nbOfObj-1; i++) {
		nextobjp = objp + objSize;
		((nfp_objHdr*)objp)->nextFreeObj = nextobjp;
		objp = nextobjp;
	}
	// Set the last object's nextFreeObj to NULL
	((nfp_objHdr*)objp)->nextFreeObj = NULL;

	return (NFX_OK);
}



