/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : crypt.c
 *
 * Description  : Encrytion library.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <pwd.h>
#include <unistd.h>

#include "crypt.h"
#include "crypt-api.h"

/*------------------------------------------------------------------------------
 * Constants and macros definition.
 *------------------------------------------------------------------------------
 */

#define key 0x59

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

static void
toHexString(byte *  a ,
            char *  hex,
            int     len);

static void
fromHexString(char *    hex,
              byte *    a);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : crypt_toHexString()
 *
 * Description  : Convert a byte array into a printable string.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
toHexString(byte *  a ,
            char *  hex,
            int     len)

{
    int i;
    int c;

    for (i = 0; i < len; i++)
    {
      
        c = (a[i] & 0xf0) >> 4;
        hex[i*2] = (c <= 9)? '0'+c :'a'+c-10;
        c = a[i] & 0x0f;
        hex[i*2+1] = (c <= 9)? '0'+c:'a'+c-10;        
    }
    hex[2*len] = 0;
}

/*------------------------------------------------------------------------------
 * Name         : crypt_fromHexString()
 *
 * Description  : Convert a printable string into a byte array.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
fromHexString(char *    hex,
              byte *    a)

{
    int i=0;
    int j=0;

    bzero((void *) a, strlen(a)/2);
    while(hex[i] != 0)
    {
        a[j] = (((hex[i]-'0') <=9 )? hex[i]-'0' : hex[i]-'a'+10)<<4;
        a[j] += ((hex[i+1]-'0') <= 9)? hex[i+1]-'0' : hex[i+1]-'a'+10;
        i += 2;
        j++;
    } // while
}
 
/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : crypt_encode()
 *
 * Description  : Encode a data using a 2 way encrytion.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int                                 // NFX_OK, NFX_ERR      - OUPUT
crypt_encode(char * plainStr,       // string to encode     - INPUT
             char * encodStr)       // encoded string       - OUPUT
    
{
    int     i;
    char    e[128];
    
    CRYPT_FCT_PRINT(("> crypt_encode\n"));

    e[0] = plainStr[0] ^ key;
    for (i = 1; i < strlen(plainStr); i++)
    {
        e[i] = plainStr[i] ^e [i-1] ^ key ^ i;
    } // for
    toHexString(e, encodStr, strlen(plainStr));
    
    CRYPT_DBG_PRINT(("%s ->(encode)-> %s\n", plainStr, encodStr));
    
    CRYPT_FCT_PRINT(("< crypt_encode\n"));
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : crypt_decode()
 *
 * Description  : Decode a data using a 2 way encrytion.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int                             // NFX_OK, NFX_ERR      - OUPUT
crypt_decode(char * encodStr,   // string to decode     - INPUT
             char * decodStr)   // decoded string       - OUPUT
    
{
    int     i;
    byte    e[128];
    int     len = strlen(encodStr)/2;
    
    CRYPT_FCT_PRINT(("> crypt_decode\n"));

    fromHexString(encodStr, e);
    decodStr[0] = e[0] ^ key;
    for (i = 1; i < len; i++)
    {
        decodStr[i] = e[i] ^ e[i-1] ^ key ^ i;
    } // for
    decodStr[len]=0;
    
    CRYPT_DBG_PRINT(("%s ->(decode)-> %s\n", encodStr, decodStr));
    
    CRYPT_FCT_PRINT(("< crypt_decode\n"));
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : crypt_digest()
 *
 * Description  : Message digest.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int                                 // NFX_OK, NFX_ERR              - OUPUT
crypt_digest(char *         algo,   // MD5, SHA, SHA1...            - INPUT
             char *         msg,    // message to digest            - INPUT
             size_t         msgLen, // length do message to digest  - INPUT
             crypt_md_t *   md)     // message digest               - OUPUT
    
{
    EVP_MD_CTX      ctxt;
    const EVP_MD *  type;
    int             i;

    CRYPT_FCT_PRINT(("> crypt_digest\n"));

    OpenSSL_add_all_digests();
    
    type = EVP_get_digestbyname(algo);
    
    if(!type)
    {
        CRYPT_ERR_PRINT(("Unknown message digest %s\n", algo));
        return(NFX_ERR);
    }
    
    EVP_DigestInit(&ctxt, type);
    EVP_DigestUpdate(&ctxt, msg, msgLen);
    EVP_DigestFinal(&ctxt, md->value, &(md->len));
    
    CRYPT_DBG_PRINT(("Digest is (%d): ", md->len));
    for(i = 0; i < md->len; i++)
    {
        CRYPT_DBG_PRINT(("%02x", md->value[i]));
    }
    CRYPT_DBG_PRINT(("\n"));
    
    CRYPT_FCT_PRINT(("< crypt_digest\n"));
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : crypt_getPasswd()
 *
 * Description  : Get a new password from the user.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int
crypt_getPasswd(char * newPasswd)   // new password - OUPUT
    
{
    char    pass1[CRYPT_MAX_PASSWD_LEN];
    char    pass2[CRYPT_MAX_PASSWD_LEN];
    
    CRYPT_FCT_PRINT(("> crypt_getPasswd\n"));

	strcpy(pass1, getpass("New Password : "));
    if (strlen(pass1) > 8)
    {
        fprintf(stderr, "Max. password length is 8 characters\n");
        return NFX_ERR;
    }
	strcpy(pass2, getpass("Confirm Password : "));
	if (!strcmp(pass1, pass2))
    {
		strcpy(newPasswd, pass1);
        CRYPT_DBG_PRINT(("newPasswd = %s\n", newPasswd));
		return(NFX_OK);
	}
	fprintf(stderr, "Did not match, password not changed.\n");  

    CRYPT_FCT_PRINT(("< crypt_getPasswd\n"));
	return(NFX_ERR);	
}

/*----------------------------------------------------------------------------*/
