/*
 *=============================================================================
 * File Name   :    dp83816-rx.c
 *
 *  Description:    Receive routines for DP83816 (NS 10/100 Ethernet MAC)
 *                  Ported from sm-dp83820 module.
 *
 *  Created    :    05/24/04
 *
 *  Author     :    Shaun Hui
 *
 *  Copyright(c)  2004, OnStor Inc
 *=============================================================================
 */

#include "dp83816.h"
#include <exports/mem-api.h>

int dp83816_rx_pkt_trace = 0;
int dp_rxHWM = 0;

int dp83816_printBuf(char *, int, int);

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_rxPacket
 *
 * Description: 	Transmits a unicast packet to the backplane.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 30, 2001
 * 					
 * Returns: ?????
 *
 *------------------------------------------------------------------------------
 */
int
dp83816_rxPacket( dp83816_rxCtrl_t *rxc , dp83816_cfg_t *dp, uint32 loopCnt )
{
	int 			i, diff;
	address_t		tmp;
	uint32			fError;
	uint32 			cmdsts;
	uint32 			pktLen;
	uchar8 			*newBuf;
	eee_desc_t     	*eDesc = NULL;
	dp83816_desc_t 	*rxCur;
	int 			hdrLen;
	int				maxloopCnt = loopCnt;

	hdrLen = 0;
	rxCur = rxc->rxCur;
	cmdsts = rxCur->cmdsts;
	while( (cmdsts & DP83816_CMDSTS_OWN) && loopCnt-- ) {

		/*
		 * Allocate a new receive buffer and add to rx queue.
		 */
		newBuf = eee_allocateBuffer(EEE_SHARED_POOL, DP83816_BUFFER_SIZE,0);
		if( !newBuf ) {
			rxc->noBufErr++;
			dp->stats.rxNoBuf++;
			break;			/* Wait until more buffers are available */
		}

		/*
		 * Allocate EEE packet descriptor.
		 */
		if(((cmdsts & DP83816_CMDSTS_MORE)==0)&&
			((cmdsts & DP83816_CMDSTS_RX_ERRBITS) ==0)) {
			eDesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
			if( !eDesc ) {
				eee_deallocateBuffer(newBuf);
				rxc->noDescErr++;
				break;		/* Wait until more descriptors are available */
			}
			// Mark first buffer as unused.
			eDesc->bd[0].control = 0;
		}
		else
			eDesc = NULL;

		/*
		 * Store buffer pointers in buffer cache for now.
		 */
		tmp = phys_to_virt_cached((address_t)rxCur->bufptr);
		rxc->bc[rxc->bufCnt].ptr = (uchar8 *)tmp;
		rxc->bc[rxc->bufCnt].len = cmdsts & DP83816_CMDSTS_LEN;
		rxc->bufCnt++;

#if 0
		if( dp83816_rx_pkt_trace ) {
			printf("dp83816 Rx: Bufptr: %x, BufLen: %d, cmdsts: %x\n",
			tmp, (cmdsts & DP83816_CMDSTS_LEN), cmdsts);
			dp83816_printBuf( tmp, (cmdsts & DP83816_CMDSTS_LEN), 1); 
		}
#endif

		/*
		 * Replenish buffer and transfer descriptor ownership 
		 * back to the chip.
		 */
		rxCur->bufptr = (uchar8 *)virt_to_phys((address_t)newBuf);
		rxCur->cmdsts = (DP83816_RX_BUF_SIZE-1) & DP83816_CMDSTS_LEN;

		/* 
		 * Look for the last buffer in the packet 
		 */
		if( (cmdsts & DP83816_CMDSTS_MORE) == 0 ) {
			if ((cmdsts & DP83816_CMDSTS_RX_ERRBITS) == 0) {

				fError = 0;
				pktLen = 0;
				for(i=1; i <= rxc->bufCnt; i++) {

					if( dp83816_rx_pkt_trace ) {
						printf("dp83816 Rx: Bufptr: %lx, BufLen: %d\n",
							   (address_t)rxc->bc[i-1].ptr, rxc->bc[i-1].len);
					}
					
					eDesc->bd[i].buf = rxc->bc[i-1].ptr;
					eDesc->bd[i].len = rxc->bc[i-1].len;
					eDesc->bd[i].control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
					pktLen += rxc->bc[i-1].len;
				}

				//-- pass it up to the receive function.
				eDesc->hdr.dest_port = dp->ipAppPort;
				eDesc->hdr.src_port  = dp->loopAddr;
				hdrLen = 0;

				if( dp83816_rx_pkt_trace ) {
					printf("RxPkt: Len: %d, buf: %lx, ", pktLen,
						   (address_t)(eDesc->bd[1].buf - hdrLen));
					printf("dest_port: %x, ", eDesc->hdr.dest_port);
					printf("src_port: %x\n", eDesc->hdr.src_port);
					dp83816_printBuf( eDesc->bd[1].buf - hdrLen,
									  eDesc->bd[1].len + hdrLen, 1); 
				}
				eDesc->hdr.control = 
					(pktLen & EDESC_TOTAL_LEN) |
					((rxc->bufCnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

				rxc->pkts++;
				rxc->octets += pktLen;
				dp->stats.rxPkts++;
				dp->stats.rxOctets += pktLen;

				/*
				 * Queue the packet to the EEE layer.
				 */
				eee_receivePacket(eDesc);
				
			}
			else {    /* Errors found */

				printf("Rx Error Detected. Dropping Packet: cmdsts: %x\n", cmdsts);
				fError = 1;
				if( cmdsts & DP83816_CMDSTS_RXA ) 
					rxc->rxaErr++;
				if( cmdsts & DP83816_CMDSTS_RXO )
					rxc->rxoErr++;
				if( cmdsts & DP83816_CMDSTS_LONG ) 
					rxc->tooLongErr++;
				if( cmdsts & DP83816_CMDSTS_RUNT )
					rxc->runtErr++;
				if( cmdsts & DP83816_CMDSTS_ISE )
					rxc->isErr++;
				if( cmdsts & DP83816_CMDSTS_CRCE ) {
					rxc->crcErr++;
					dp->stats.rxCrc++;
				}
				if( cmdsts & DP83816_CMDSTS_FAE )
					rxc->faErr++;
				if( cmdsts & DP83816_CMDSTS_LBP )
					rxc->lbpErr++;
				if( cmdsts & DP83816_CMDSTS_COL )
					rxc->colErr++;

				if (eDesc)
					eee_deallocateDesc(eDesc);
			}

			/*
			 * Clear buffer cache.
			 */
			for(i=0; i < rxc->bufCnt; i++) {
				if(fError && rxc->bc[i].ptr)
					eee_deallocateBuffer(rxc->bc[i].ptr);
				rxc->bc[i].ptr = 0;
				rxc->bc[i].len = 0;
			}
			rxc->bufCnt = 0;
		}
		if( ++rxCur >= rxc->rxqEnd )
			rxCur = rxc->rxqStart;

		cmdsts = rxCur->cmdsts;
	}
	rxc->rxCur = rxCur;
	diff = maxloopCnt - loopCnt;
	if (diff > dp_rxHWM) // high water mark for rx descriptors processed
		dp_rxHWM = diff;
	return(NFX_OK);
} /* dp83816_rxPacket */


/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_rxPktPoll
 *
 * Description: 	Called by the eee layer and in turn calls dp83816_rxPacket
 *					for each receive priority queue to collect packet buffers
 *					and pass them up to the eee layer.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 2, 2001
 * 					
 * Returns: NFX_OK
 *
 *------------------------------------------------------------------------------
 */
void
dp83816_rxPktPoll( void *cb, uint32 tlimit )
{
	int 			i;
	int 			port;
	int32 			loopCnt;
	dp83816_cfg_t	*dpCfg;
	
	loopCnt = 256;
	for(i=0; i < DP83816_RX_Q_CNT; i++) {
		for (port=0, dpCfg=dp83816; port < DP83816_NUM_PORT; port++, dpCfg++)
			dp83816_rxPacket( &dpCfg->rxc[i], dpCfg, loopCnt );
	}
	return;
} /* dp83816_rxPktPoll */


static int
dp83816_decodeHdr( char *data )
{
	uint32 tmp;

	printf("\n");
	printf("MAC DA  : %02x:%02x:%02x:%02x:%02x:%02x\n", 
		   data[0]&0xff, data[1]&0xff, data[2]&0xff, 
		   data[3]&0xff, data[4]&0xff, data[5]&0xff);
	printf("MAC SA  : %02x:%02x:%02x:%02x:%02x:%02x\n", 
		   data[6]&0xff, data[7]&0xff, data[8]&0xff, 
		   data[9]&0xff, data[10]&0xff, data[11]&0xff);

	tmp = (*(uint32 *)(data+12)) & 0xFFFF;
	printf("TYPE    : %04x\n", tmp);

	if((tmp == EEE_ETH_TYPE) || (tmp==EEE_ETH_TYPE_SAC)) {		// Only if Agile Protocol type.
		tmp = *(uint32 *)(data+16);
		printf("EEE DEST: %08x, App: %d:%s, CPU: %d, Slot: %d, Dtype: %d\n", 
			   tmp,
			   EEE_GET_APP_ID(tmp),
			   eee_getAppNameFromId(EEE_GET_APP_ID(tmp)),
			   EEE_GET_DEST_CPU(tmp),
			   EEE_GET_SLOT_NUM(tmp),
			   EEE_GET_DTYPE(tmp));

		tmp = *(uint32 *)(data+20);
		printf("EEE SRC : %08x, App: %d:%s, CPU: %d, Slot: %d, Dtype: %d\n", 
			   tmp,
			   EEE_GET_APP_ID(tmp),
			   eee_getAppNameFromId(EEE_GET_APP_ID(tmp)),
			   EEE_GET_DEST_CPU(tmp),
			   EEE_GET_SLOT_NUM(tmp),
			   EEE_GET_DTYPE(tmp));

	}

	printf("\n............HEX DUMP.................\n");
	return 0;
} // dp83816_decodeHdr


int
dp83816_printBuf(char *buf, int len, int fDecode)
{
	int i, j;
    char ch;
	char *data;
 
	data = buf;
	if( fDecode ) {
		dp83816_decodeHdr(data);
	}

#if 0
	if( len > 64 )
		len = 64;
#endif

    for(i=0; i < len; i+=16) {
        printf("\n%04x:\t", i);
        for( j=i; j < (i+16); j++ )
            printf("%02x ", data[j] & 0xff);
        printf("    ");
        for( j=i; j < (i+16); j++ ) {
            ch = data[j] & 0xff;
            if( (ch > 0x20) && (ch < 0x7F))
                printf("%c", ch);
            else
            if( ch == 0 )
                printf("%c",'.');
            else
                printf("%c",'_');
        }
    }
    printf("\n\n");                      
	return 0;
} // dp83816_printBuf

