/*  
 *=============================================================================
 * File Name   :    dp83816.c
 *
 *  Description:    Init routines for DP83816 (NS 10/100 Ethernet MAC)
 *                  Ported from sm-dp83820 module.
 *
 *  Created    :    05/24/04
 *
 *  Author     :    Shaun Hui
 *
 *  Copyright(c)  2004, OnStor Inc
 *=============================================================================
 */
#include "dp83816.h"
#include <exports/mem-api.h>
#include <exports/malloc-api.h>

dp83816_cfg_t *dp83816;

int32 dp83816_txUnicast(uint32, eee_desc_t *);
int dp83816_setMacSpeed(dp83816_cfg_t *);
int dp83816_readReg(volatile uint32 *); 
int dp83816_fillRxQ(dp83816_desc_t *, int); 
int dp83816_initQ(dp83816_desc_t *, int);
int dp83816_initDevice(int port);
int dp83816_resetDevice(dp83816_cfg_t *); 
int dp83816_initRegs(dp83816_cfg_t *); 

extern int dp83816_diagLbPktReceive(eee_desc_t *);
extern int bp_getMyMacAddr(uint32, uint32, uint32, uint32, char *);
extern byte *seepGetMyExtMacAddr(int32 link);



/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_initModule
 *
 * Description: 	Initializes the DP83816 device driver module.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 * 					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
dp83816_initModule()
{
	int 			rc;
	int 			port;
	int 			dpAppId;
	int 			lucAppId;
	dp83816_cfg_t	*dpCfg;
	
	dp83816 = (dp83816_cfg_t *)eee_ramAlloc(sizeof(dp83816_cfg_t) * 
											DP83816_NUM_PORT);
	if(!dp83816) {
		printf("[dp83816_initModule] Error: Malloc Failed: dp83816_cfg_t.\n");
		return(NFX_ERR);
	}
	memset(dp83816,0,sizeof(dp83816_cfg_t)*DP83816_NUM_PORT);
	
	/*
	 *--------------------------------------
	 * Register with EEE and get app id
	 *--------------------------------------
	 */
	dpAppId = eee_allocateApplicationId("dp83816");
	if( dpAppId == NFX_ERR )
	{
		printf("[dp83816_initModule] Error Allocate app id failed.");
		return NFX_ERR;
	}

	//-- Used only in prom on the SSC board.
	lucAppId = eee_allocateApplicationId("lucRcv");
	if( lucAppId == NFX_ERR ) {
		printf("[dp83816_initModule] Error: Allocate IP app id failed.");
		return NFX_ERR;
	}

	for (port = 0, dpCfg = dp83816; port < DP83816_NUM_PORT; port++, dpCfg++)
	{
		dpCfg->appId = dpAppId;
		dpCfg->ipAppPort = PACK_UNICAST_PORT_ADDRESS( nfxMySlotId,
                                                     nfxMyCPUId,
                                                     lucAppId,
                                                     EEE_USER_DATA,
                                                     NFX_COS_HIGH,
                                                     PROM_PORT_NET0+port);

		dpCfg->loopAddr = PACK_UNICAST_PORT_ADDRESS( nfxMySlotId, 
												   nfxMyCPUId,
												   dpAppId,
												   EEE_USER_DATA,
												   NFX_COS_HIGH, 
												   PROM_PORT_NET0+port);
		memcpy( dpCfg->macAddr, seepGetMyExtMacAddr(port), 6);

#ifdef DIAG_R9000 
		dpCfg->loopAddr = PACK_UNICAST_PORT_ADDRESS( nfxMySlotId, 
												   nfxMyCPUId,
												   dpAappId,
												   EEE_USER_DATA,
												   NFX_COS_HIGH, 
												   PROM_PORT_NET0+port);
		bp_getMyMacAddr(nfxMySlotId, 
						nfxMyCPUId,
						1,
						EEE_USER_DATA,
						dpCfg->macAddr);

#endif	

		//-- Initialize the National DP83816 Mac
		if( dp83816_initDevice( port ) == NFX_ERR ) {
			printf("DP83816 Initialize Module Failed\n");
			return 0;
		}
	}

	rc = eee_registerReceiveFunc( dpAppId,			/* App Id		 */
								  dp83816_rxAppPkt,	/* Ucast Rx Func */
								  NULL);			/* Mcast Rx Func */
	if( rc == NFX_ERR )
	{
		printf("[dp83816_initModule] Error: registerReceiveFunc\n");
		return NFX_ERR;
	}

	eee_registerPollingFunc( EEE_HIGH_PRIORITY, dp83816_rxPktPoll, 0 );
	eee_registerPollingFunc( EEE_HIGH_PRIORITY, dp83816_txCleanPoll, 0 );

	//printf("Calling eee_registerTransmitFunc\n");

#if defined DIAG_R9000 

	rc = eee_registerTransmitFunc( PROM_PORT_NET0,
								   dp83816_txUnicast,
								   dp83816_txUnicast,
								   1500);
	rc = eee_registerTransmitFunc( EEE_USER_DATA,
								   dp83816_txUnicast,
								   dp83816_txUnicast,
								   9*1024);

	eee_registerDefaultTransmitFunc( EEE_USER_DATA,
									 dp83816_txUnicast,
									 9*1024);
#else
	/* SSC PROM */
	rc = eee_registerTransmitFunc( PROM_PORT_MGMT,
								   dp83816_txUnicast,
								   dp83816_txUnicast,
								   1500);

	eee_registerDefaultTransmitFunc( EEE_MGMT,
									 dp83816_txUnicast,
									 1500);
#endif
	if( rc == NFX_ERR )
		printf("Tx Func Register Error: %x\n", rc);

	return(NFX_OK);
} /* dp83816_initModule */


/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_initDevice
 *
 * Description: 	Initializes the DP83816 device.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 * 					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
dp83816_initDevice(int port)
{
	uint32           i;
	uint32           align_diff;
	uint32           size_rx;
	uint32           size_tx;
	uchar8           *qstart;
	dp83816_txCtrl_t *txc;
	dp83816_rxCtrl_t *rxc;
	dp83816_cfg_t    *dpCfg;

	//printf("[dp83816_initDevice] Enter\n");
	dpCfg = dp83816+port;
	dpCfg->baseAddr = DP83816_BASE_ADDR + (port * DP83816_PORT_OFFSET);
	dpCfg->regs 	 = (dp83816_regs_t *)(dpCfg->baseAddr);


	/*
	 * Allocate and align memory on 64-bit boundry
	 * for receive and transmit descriptor rings. 
	 */
	size_tx = sizeof(dp83816_desc_t) * DP83816_TX_Q_LEN;
	size_rx = sizeof(dp83816_desc_t) * DP83816_RX_Q_LEN;


	/* Setup descriptor rings */
	for( i=0; i < DP83816_TX_Q_CNT; i++ )
	{
		txc = &dpCfg->txc[i];

		/* 
		 * Allocate TX Desc from Shared mem.
		 */
		qstart = (uchar8 *)eee_allocateBlock(size_tx + 64);

		//printf("txq start: %x: %d bytes allocated\n", qstart, size_tx+64);
		/* Convert to un-cached memory address */
		qstart = (uchar8 *)cached_to_uncached((address_t)qstart);

		align_diff = (address_t)qstart & 0x00000007;

		if( align_diff )
			align_diff = 8 - align_diff;
		qstart += align_diff;

		memset(qstart, 0, size_tx);
		txc->txqStart = (dp83816_desc_t *)qstart;
		txc->txCur = txc->txCln = txc->txqStart;
		txc->txqEnd = txc->txqStart + DP83816_TX_Q_LEN;
		txc->avail = DP83816_TX_Q_LEN;
		dp83816_initQ( txc->txqStart, DP83816_TX_Q_LEN );
	}

	for( i=0; i < DP83816_RX_Q_CNT; i++ )
	{
		rxc = &dpCfg->rxc[i];

		/* 
		 * Allocate RX Desc from Shared mem.
		 */
		qstart = (uchar8 *)eee_allocateBlock(size_tx + 64);

		//printf("rxq start: %x: %d bytes allocated\n", qstart, size_rx+64);
		/* Convert to un-cached memory address */
		qstart = (uchar8 *)cached_to_uncached((address_t)qstart);

		memset(qstart, 0, size_rx);

		align_diff = (address_t)qstart & 0x00000007;
		if( align_diff )
			align_diff = 8 - align_diff;
		qstart += align_diff;

		rxc->rxqStart = (dp83816_desc_t *)(qstart);
		rxc->rxCur = rxc->rxqStart;
		rxc->rxqEnd = rxc->rxqStart + DP83816_RX_Q_LEN;
		rxc->avail = DP83816_RX_Q_LEN;
		dp83816_initQ( rxc->rxqStart, DP83816_RX_Q_LEN );
		dp83816_fillRxQ( rxc->rxqStart, DP83816_RX_Q_LEN);
	}

	/*
	 * Initialize the device.
	 */
	dp83816_resetDevice( dpCfg );
	dp83816_initRegs( dpCfg ); 

	//-- Put link into normal mode.
	dpCfg->link.loopback = lb_none;
			
	//-- Mark MAC Id: 
	dpCfg->macId = i;

	//printf("[dp83816_initDevice] Leave\n");
	return( NFX_OK ); 
} /* dp83816_initDevice */ 

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_initRegs
 *
 * Description: 	Initializes the DP83816 registers.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 * 					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
dp83816_initRegs(dp83816_cfg_t *dpCfg) 
{
	ushort16  *addr;
	uint32 *reg;
	uint32 val;
	dp83816_regs_t *regs;
	dp83816_cr_t cr;
	dp83816_cfgr_t cfg;
	dp83816_rxcfgr_t rxcfg;
	dp83816_txcfgr_t txcfg;
	dp83816_rfcr_t rfcr;
	dp83816_pcr_t pcr;

	//printf("[dp83816_initRegs] Enter\n");
	regs = dpCfg->regs;

	/* 
	 * Setup Config Register.
	 */
	cfg.wd = 0;

	cfg.bits.pint_acen = 1;
	cfg.bits.pause_adv = 1;
	cfg.bits.aneg_sel  = 0x7;
	cfg.bits.reqalg    = 0;
	cfg.bits.pesel     = 1;
	cfg.bits.brom_dis  = 1;
	cfg.bits.bem       = 0;

	cfg.wd |= 0x00040000;

	dp83816_writeReg(&regs->cfg, cfg.wd);

	/* 
	 * Disable Rx and Tx 
	 */
	cr.wd = 0;
	cr.bits.txd = 1;
	cr.bits.rxd = 1;
	dp83816_writeReg(&regs->cr, cr.wd);

	/*
	 * Setup Receive Config Register.
	 */
	rxcfg.wd = dp83816_readReg(&regs->rxcfg);
	rxcfg.bits.aep      = 0;	// Do not accept errroed pkts
	rxcfg.bits.arp      = 0;	// Do not accept runt pkts
	rxcfg.bits.atx      = 1;	// 
	rxcfg.bits.alp      = 0;	// Do not accept long pkts
	rxcfg.bits.mxdma    = 6; 	// 128 bytes
	rxcfg.bits.drth     = 0x04;	// 32 (4 * 8) bytes.
	dp83816_writeReg(&regs->rxcfg, rxcfg.wd);

	/* 
	 * Setup Unicast Address register.
	 */
	addr = (ushort16 *)dpCfg->macAddr;
	dp83816_writeReg(&regs->rfcr, 0);
	dp83816_writeReg(&regs->rfdr, *addr);
	addr++;
	dp83816_writeReg(&regs->rfcr, 2);
	dp83816_writeReg(&regs->rfdr, *addr);
	addr++;
	dp83816_writeReg(&regs->rfcr, 4);
	dp83816_writeReg(&regs->rfdr, *addr);
	
	//-- Setup Pause Control Register
	pcr.wd = 0;
	pcr.bits.pause_cnt	= 0xfff0;
	dp83816_writeReg(&regs->pcr, pcr.wd);
	pcr.bits.psen		= 1;		// Enable Pause Receive
	pcr.bits.ps_mcast	= 1;		// pause on mcast
	pcr.bits.ps_da		= 1;		// pause on DA
	dp83816_writeReg(&regs->pcr, pcr.wd);

	/*
	 * Setup Receive Filter/Match Register.
	 * EN bit must be 0 before setting AAB & APM.
	 */
	rfcr.wd = 0;
	rfcr.bits.aab  = 1;			// accept all broadcast packets 
	rfcr.bits.apm  = 1;			// accept perfect match 
	dp83816_writeReg(&regs->rfcr, rfcr.wd);

	rfcr.bits.rfen = 1;			// Receive filter enable.
	dp83816_writeReg(&regs->rfcr, rfcr.wd);

	/*
	 * Initialize MIB Control Register.
	 */
	reg = &(regs->mibc);
	val = 0;
	dp83816_writeReg(reg, val);

	/*
	 * Initialize Pointer to Rx Desc Queue
	 */
	reg = &(regs->rxdp);
	val = virt_to_phys((address_t)(dpCfg->rxc[0].rxqStart));
	dp83816_writeReg(reg, val);

	/*
	 * Initialize Pointer to Tx Desc Queue
	 */
	reg = &(regs->txdp);
	val = virt_to_phys((address_t)(dpCfg->txc[0].txqStart));
	dp83816_writeReg(reg, val);


	//-- Initialize the TX CFG register.
	txcfg.wd = 0;
	txcfg.bits.csi   = 1;
	txcfg.bits.hbi   = 1;
	txcfg.bits.atp   = 1;		// Automatic transmit padding.
	txcfg.bits.mxdma = 6;		// 128 bytes
	txcfg.bits.drth  = 0x30;	// 
	txcfg.bits.flth  = 0x10;	// 512 (16 * 32) bytes.

	dp83816_writeReg(&regs->txcfg, txcfg.wd);

	/* 
	 * Enable Rx 
	 */
	cr.wd = 0;
	cr.bits.rxe = 1;
	dp83816_writeReg(&regs->cr, cr.wd);

	//printf("[dp83816_initRegs] Leave\n");
	return(NFX_OK);
} /* dp83816_initRegs */

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_resetDevice
 *
 * Description: 	Resets the device.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 * 					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
dp83816_resetDevice(dp83816_cfg_t *dpCfg) 
{
	int i;
	volatile dp83816_regs_t *regs;
	dp83816_cr_t cr;

	//printf("[dp83816_resetDevice] Enter\n");
	regs = dpCfg->regs;

	cr.wd  = 0;
	cr.bits.rst = 1;
	cr.bits.rxr = 1;
	cr.bits.txr = 1;
	dp83816_writeReg((volatile uint32 *)&regs->cr, cr.wd);
	for(i=0; i < 10000; i++) {
		cr.wd = dp83816_readReg((volatile uint32 *)&regs->cr);
		if( cr.bits.rst == 0 ) {
			//printf("reset done: %08x\n", cr.wd);
			break;
		}
	}
	if (i == 10000) {
		panic("dp83816_resetDevice: chip reset failed!\n");
	}
	//printf("[dp83816_resetDevice] Leave\n");
	return(NFX_OK);
} /* dp83816_resetDevice */

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_resetModule
 *
 * Description: 	Resets the module and devce.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 * 					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
dp83816_resetModule(dp83816_cfg_t *dp83816) 
{
	return 0;
} /* dp83816_resetModule */

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_initQ
 *
 * Description: 	Initializes descriptor queue by carving up the 
 *					descriptor memory into individual descriptors and 
 *					linking them together.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 *------------------------------------------------------------------------------
 */
int
dp83816_initQ(dp83816_desc_t *qstart, int qLen)
{
	int i;
	uint32 val;
	dp83816_desc_t *d;

	d  = qstart;
	/* Convert to physical address space */
	for(i=1; i < qLen; i++, d++) {
		val = virt_to_phys((address_t)(d+1));
		d->link = (dp83816_desc_t *)((address_t)val);
	}
	val = virt_to_phys((address_t)qstart);
	d->link = (dp83816_desc_t *)((address_t)val);
	return(NFX_OK);
} /* dp83816_initQ */

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_fillRxQ
 *
 * Description: 	Fills a receive queue with empty system data uffers.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 26, 2001
 *------------------------------------------------------------------------------
 */
int
dp83816_fillRxQ(dp83816_desc_t *rxQ, int qLen) 
{
	int i;
	char *pBuf;
	dp83816_desc_t *d;

	//printf("[dp83816_fillRxQ] Enter\n");
	d = rxQ;
	for(i=0; i < qLen; i++, d++)
	{
		pBuf = eee_allocateBuffer(EEE_SHARED_POOL, DP83816_BUFFER_SIZE, 0);
		if( !pBuf ) {
			printf("[dp83816_fillRxQ]: Error: Cannot allocate free buffer\n");
			return(NFX_ERR);
		}
		pBuf = (char *)virt_to_phys((address_t)pBuf);
		d->bufptr = (uchar8 *)pBuf;
		d->cmdsts = ((DP83816_RX_BUF_SIZE-1) & DP83816_CMDSTS_LEN);  
	}
	//printf("[dp83816_fillRxQ] Leave\n");
	return(NFX_OK);
} /* dp83816_fillRxQ */


/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_writeReg
 *
 * Description: 	Writes 32-bit value to a device register.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 30, 2001
 *------------------------------------------------------------------------------
 */
int
dp83816_writeReg(volatile uint32 *reg, uint32 val) 
{
	*reg = val;
	return NFX_OK;
} /* dp83816_writeReg */
/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_readReg
 *
 * Description: 	Returns 32-bit value read from device register.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 30, 2001
 *------------------------------------------------------------------------------
 */
int
dp83816_readReg(volatile uint32 *reg) 
{
	uint32 val;
	val = *reg;
	return(val);
} /* dp83816_readReg */

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_rxAppPkt
 *
 * Description: 	Receive packet sent to DP83816 mgmt layer.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 12, 2001
 *------------------------------------------------------------------------------
 */
void
dp83816_rxAppPkt( eee_desc_t *edesc )
{
	dp83816_diagLbPktReceive(edesc);
	return;
} /* dp83816_rxAppPkt */

int
dp83816_wait( int msec )	
{

	return 0;
} /* dp83816_wait */

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83816_setMacSpeed
 *
 * Description: 	Sets up the MAC to the speed auto-detected by the
 *					physical layer.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	October 11, 2001
 * 					
 * Returns: 		NFX_OK
 *------------------------------------------------------------------------------
 */
int 
dp83816_setMacSpeed(dp83816_cfg_t *dp)
{
	dp83816_cfgr_t cfg;
	dp83816_regs_t *r;


	r = dp->regs;

	//-- Read current register values 
    cfg.wd    = dp83816_readReg(&r->cfg);

	//-- Setup appropriate bits in cfg register.
    switch( dp->link.speed ) 
	{
		case speed_10:
			cfg.bits.aneg_sel		= 0x04;	// force 10Mbps, full duplex
			break;

		case speed_100:
			cfg.bits.aneg_sel		= 0x06;	// force 100Mbps, full duplex
			break;

		default:
			return NFX_OK;
    }
	
    //-- Write bits back to Mac registers.
    dp83816_writeReg(&r->cfg, cfg.wd);

	return NFX_OK;
} // dp83816_setMacSpeed

void
dp83816_disablePort(int port)
{
	dp83816_cfg_t	*dpCfg;
	dp83816_cr_t	cr;
	dp83816_regs_t	*regs;

	dpCfg = dp83816+port;
	regs = dpCfg->regs;

	/* 
	 * Disable Rx and Tx 
	 */
	cr.wd = dp83816_readReg((volatile uint32 *)&regs->cr);
	cr.bits.txd = 1;
	cr.bits.rxd = 1;
	dp83816_writeReg(&regs->cr, cr.wd);

}
