/*  
 *=============================================================================
 * File Name   :    dp83820-mii.c
 *
 *  Description:   Media independent interface access routines. 
 *
 *  Created    :    26 June 2001
 *
 *  Author     :    Frank Salzmann
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */

#include "dp83820.h"

#define	PHYIDR1			0x02
#define	PHYIDR2			0x03
#define MIIread 		0x60000000l
#define	MIIwrite		0x50020000l
#define MDDIR           0x00000020      /* MII management data direction    */
#define MDC             0x00000040      /* MII management data clock        */
#define MDIO            0x00000010      /* MII management data input/output */

#define	BMCR	0x00 	/* Basic Mode Control (rw) */
#define	BMSR	0x01	/* Basic Mode Status (ro) */
#define PHYIDR1	0x02
#define PHYIDR2	0x03
#define ANAR	0x04	/* Autonegotiation Advertisement */
#define	ANLPAR	0x05	/* Autonegotiation Link Partner Ability (rw) */
#define ANER	0x06	/* Autonegotiation Expansion */
#define K1CTL	0x09	/* 1000baseT control */
#define K1STSR	0x0A	/* 1K Status Register (ro) */
#define AUXCTL	0x18	/* aux control register */

/* Basic Mode Status register (RO) */
#define BMSR_100BT4		0x8000
#define BMSR_100BT_FDX		0x4000
#define BMSR_100BT_HDX  	0x2000
#define BMSR_10BT_FDX   	0x1000
#define BMSR_10BT_HDX   	0x0800
#define BMSR_100BT2_FDX 	0x0400
#define BMSR_100BT2_HDX 	0x0200
#define BMSR_1000BT_XSR		0x0100
#define BMSR_PRESUP		0x0040
#define BMSR_ANCOMPLETE		0x0020
#define BMSR_REMFAULT		0x0010
#define BMSR_AUTONEG		0x0008
#define BMSR_LINKSTAT		0x0004
#define BMSR_JABDETECT		0x0002
#define BMSR_EXTCAPAB		0x0001

/* Autonegotiation Advertisement register (RW) */
#define ANAR_NP			0x8000
#define ANAR_RF			0x2000
#define ANAR_ASYPAUSE		0x0800
#define ANAR_PAUSE		0x0400
#define ANAR_T4			0x0200
#define ANAR_TXFD		0x0100
#define ANAR_TXHD		0x0080
#define ANAR_10FD		0x0040
#define ANAR_10HD		0x0020
#define ANAR_PSB		0x001F

#define PSB_802_3		0x0001	/* 802.3 */

/* Autonegotiation Link Partner Abilities register (RW) */
#define ANLPAR_NP		0x8000
#define ANLPAR_ACK		0x4000
#define ANLPAR_RF		0x2000
#define ANLPAR_ASYPAUSE		0x0800
#define ANLPAR_PAUSE		0x0400
#define ANLPAR_T4		0x0200
#define ANLPAR_TXFD		0x0100
#define ANLPAR_TXHD		0x0080
#define ANLPAR_10FD		0x0040
#define ANLPAR_10HD		0x0020
#define ANLPAR_PSB		0x001F

#define K1TCR_TESTMODE		0x0000
#define K1TCR_MSMCE		0x1000
#define K1TCR_MSCV		0x0800
#define K1TCR_RPTR		0x0400
#define K1TCR_1000BT_FDX 	0x200
#define K1TCR_1000BT_HDX 	0x100

#define K1STSR_MSMCFLT		0x8000
#define K1STSR_MSCFGRES		0x4000
#define K1STSR_LRSTAT		0x2000
#define K1STSR_RRSTAT		0x1000
#define K1STSR_LP1KFD		0x0800
#define K1STSR_LP1KHD   	0x0400
#define K1STSR_LPASMDIR		0x0200

#define K1SCR_1KX_FDX		0x8000
#define K1SCR_1KX_HDX		0x4000
#define K1SCR_1KT_FDX		0x2000
#define K1SCR_1KT_HDX		0x1000

uint32 dp83820_readMiiReg(dp83820_cfg_t *dp, uint32 phyReg);
void dp83820_writeMiiPreamble(uint32 * pmear);
void dp83820_writeMiiIdle(uint32 *pmear);

extern int dp83820_readReg(volatile uint32 *);

int
dp83820_getPhyAddr(dp83820_cfg_t *dp)
{
	uint32 phyRev;
	uint32 tmp;

	/*
	 *      Getting the PhyAddress of the DP83891 chip.
	 *          Read PHYIDR1 and PHYIDR2 and compare the OUI.
	 */
	for(dp->phyAddr=31; dp->phyAddr != 0; dp->phyAddr--)
	{
		tmp = dp83820_readMiiReg(dp, PHYIDR1);
		if(tmp == 0x2000)
		{
			phyRev = dp83820_readMiiReg(dp, PHYIDR2);
			
			/*
			 *  Break, if this is a DP83891 / DP83861
			 */
			if ( (phyRev & ~(0x000f)) == 0x00005c50 ||
				 (phyRev & ~(0x000f)) == 0x00005c60 )
			{
				break;
			}
		}
	}
	return 0;
} /* dp83820_getPhyAddr */

int
dp83820_phyInit(dp83820_cfg_t *dp)
{
	uint32 phyData;

	// Read Phy Status
	phyData = dp83820_readMiiReg(dp, BMSR);
	// printf("PHY BMSR: %x\n", phyData);
	phyData = dp83820_readMiiReg(dp, BMSR);
	// printf("PHY BMSR: %x\n", phyData);

	// Place Phy into loopback mode.
	phyData = dp83820_readMiiReg(dp, BMCR);
	// printf("PHY BMCR: %x\n", phyData);
	phyData |= 0x4000;
	//dp83820_writeMiiReg(dp, BMCR, phyData);

	return 0;
} /* dp83820_phyInit */

void
dp83820_writeMiiReg(dp83820_cfg_t *dp, uint32 phyReg, uint32 phyData)
{
    uint32 mdio;
    uint32 i = 32;
    uint32 *pmear = &(dp->regs->mear);

    uint32 miiData = MIIwrite | ( dp->phyAddr << 23 ) | 
		                        ( phyReg << 18 )      | phyData;

    dp83820_writeMiiPreamble( pmear );
    dp83820_writeMiiIdle( pmear );

    for ( i = 32; i != 0; i--, miiData <<= 1 )
    {
        mdio = ( miiData & 0x80000000l ) ? MDIO : 0;
        dp83820_writeReg( pmear, MDDIR | mdio );
        dp83820_writeReg( pmear, MDDIR | mdio | MDC );
    }

    dp83820_writeMiiIdle( pmear );
} /* dp83820_writeMiiReg */

void
dp83820_writeMII_common(uint32 unit, uint32 phy_id, uint32 reg, uint32 data)
{
	dp83820_cfg_t *dp;
	uint32 *pmear;
	uint32 miiData, mdio, i;

	dp = dp83820;
	pmear = &(dp->regs->mear);
    miiData = MIIwrite | ( phy_id << 23 ) | ( reg << 18 ) | data;

    dp83820_writeMiiPreamble( pmear );
    dp83820_writeMiiIdle( pmear );

    for ( i = 32; i != 0; i--, miiData <<= 1 )
    {
        mdio = ( miiData & 0x80000000l ) ? MDIO : 0;
        dp83820_writeReg( pmear, MDDIR | mdio );
        dp83820_writeReg( pmear, MDDIR | mdio | MDC );
    }
    dp83820_writeMiiIdle( pmear );
}

uint32
dp83820_readMiiReg(dp83820_cfg_t *dp, uint32 phyReg)
{
    uint32 i;
    uint32 mdio;
    uint32 *pmear = &(dp->regs->mear);
    uint32 data = MIIread | ( dp->phyAddr << 23 ) | ( phyReg << 18 );

    dp83820_writeMiiPreamble( pmear );
    dp83820_writeMiiIdle( pmear );

    for ( i = 14; i != 0; i--, data <<= 1 )
    {
        mdio = ( data & 0x80000000l ) ? MDIO : 0;
        dp83820_writeReg( pmear, MDDIR | mdio );
        dp83820_writeReg(  pmear, MDDIR | mdio | MDC );
    }

    dp83820_writeReg(  pmear, 0 );
    dp83820_writeReg(  pmear, MDC );
    mdio = dp83820_readReg( pmear );

    for ( data = 0, i = 16; i != 0; i-- )
    {
        dp83820_writeReg( pmear, 0 );
        dp83820_writeReg( pmear, MDC );
        mdio = dp83820_readReg( pmear );
        data <<= 1;
        if ( mdio & MDIO )
        {
            data |= 1;
        }
    }

    dp83820_writeReg( pmear, 0 );
    dp83820_writeReg( pmear, MDC );
    mdio = dp83820_readReg( pmear );

    return( data );
} /* dp83820_readMiiReg */

uint32
dp83820_readMII_common(uint32 phy_id, uint32 reg)
{
	dp83820_cfg_t *dp;
	uint32 *pmear;
	uint32 miiData, mdio, i;

	dp = dp83820;
	pmear = &(dp->regs->mear);
    miiData = MIIread | ( phy_id << 23 ) | ( reg << 18 );

    dp83820_writeMiiPreamble( pmear );
    dp83820_writeMiiIdle( pmear );
    for ( i = 14; i != 0; i--, miiData <<= 1 )
    {
        mdio = ( miiData & 0x80000000l ) ? MDIO : 0;
        dp83820_writeReg( pmear, MDDIR | mdio );
        dp83820_writeReg(  pmear, MDDIR | mdio | MDC );
    }

    dp83820_writeReg(  pmear, 0 );
    dp83820_writeReg(  pmear, MDC );
    mdio = dp83820_readReg( pmear );

    for ( miiData = 0, i = 16; i != 0; i-- )
    {
        dp83820_writeReg( pmear, 0 );
        dp83820_writeReg( pmear, MDC );
        mdio = dp83820_readReg( pmear );
        miiData <<= 1;
        if ( mdio & MDIO )
        {
            miiData |= 1;
        }
    }

    dp83820_writeReg( pmear, 0 );
    dp83820_writeReg( pmear, MDC );
    mdio = dp83820_readReg( pmear );

    return( miiData );
}

void
dp83820_writeMiiPreamble(uint32 * pmear)
{
    uint32 i;

    for ( i = 0; i != 32; i++ )
    {
        dp83820_writeReg(pmear, MDDIR | MDIO);
        dp83820_writeReg(pmear, MDDIR | MDIO | MDC);
    }
} /* dp83820_writeMiiPreamble */

void
dp83820_writeMiiIdle(uint32 *pmear)
{
    dp83820_writeReg(pmear, MDDIR | MDIO);
    dp83820_writeReg(pmear, MDDIR | MDIO | MDC);
} // dp83820_writeMiiIdle 

/*
 *------------------------------------------------------------------------------
 * Function Name:	dp83820_mii_poll
 *
 * Description:     Ask the PHY what is going on, and configure speed 
 *					appropriately. For the moment, we only support automatic 
 *					configuration.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	November 09, 2001
 * 					
 * Returns: 		1 if changed else 0
 *------------------------------------------------------------------------------
 */
int 
dp83820_mii_poll(dp83820_cfg_t *dp)
{
    uint16 bmsr, bmcr, k1stsr, anlpar;
    int chg;
    char buffer[100];
    char *p = buffer;
	dp83820_phy_t *phy;

	phy = &dp->phy;
    // Read the mode status and mode control registers. 
	bmsr = dp83820_readMiiReg(dp, BMSR);
    bmcr = dp83820_readMiiReg(dp, BMCR);

    // get the link partner status 
    anlpar = dp83820_readMiiReg(dp, ANLPAR);

    // If supported, read the 1000baseT register 
    if (bmsr & BMSR_1000BT_XSR) {
		k1stsr = dp83820_readMiiReg(dp, K1STSR);
	}
    else {
		k1stsr = 0;
	}

    chg = 0;

    if( (phy->p.bmc5411.bmsr != bmsr)	 ||
		(phy->p.bmc5411.bmcr != bmcr) 	 ||
		(phy->p.bmc5411.anlpar != anlpar) ||
		(phy->p.bmc5411.k1stsr != k1stsr) ) {

		phy->p.bmc5411.bmsr = bmsr;
		phy->p.bmc5411.bmcr = bmcr;
		phy->p.bmc5411.anlpar = anlpar;
		phy->p.bmc5411.k1stsr = k1stsr;
		chg = 1;
	}

    if (chg == 0) 
		return 0;

    p += sprintf(p,"Link speed: ");

    if (k1stsr & K1STSR_LP1KFD) {
		dp->link.speed    = speed_1000;
		dp->link.duplex   = full_duplex;
		dp->link.fc		  = fc_frame; 
		dp->link.linkstat = ETHER_SPEED_1000FDX;
		p += sprintf(p,"1000BaseT FDX");
	}
    else if (k1stsr & K1STSR_LP1KHD) {
		dp->link.speed    = speed_1000;
		dp->link.duplex   = half_duplex;
		dp->link.fc		  = fc_disabled; 
		dp->link.linkstat = ETHER_SPEED_1000HDX;
		p += sprintf(p,"1000BaseT HDX");
	}
    else if (anlpar & ANLPAR_TXFD) {
		dp->link.speed    = speed_100;
		dp->link.duplex   = full_duplex;
		dp->link.fc		  = (anlpar & ANLPAR_PAUSE) ? fc_frame : fc_disabled;
		dp->link.linkstat = ETHER_SPEED_100FDX;
		p += sprintf(p,"100BaseT FDX");
	}
    else if (anlpar & ANLPAR_TXHD) {
		dp->link.speed    = speed_100;
		dp->link.duplex   = half_duplex;
		dp->link.fc		  = fc_disabled; 
		dp->link.linkstat = ETHER_SPEED_100HDX;
		p += sprintf(p,"100BaseT HDX");
	}
    else if (anlpar & ANLPAR_10FD) {
		dp->link.speed    = speed_10;
		dp->link.duplex   = full_duplex;
		dp->link.fc		  = fc_frame; 
		dp->link.linkstat = ETHER_SPEED_10FDX;
		p += sprintf(p,"10BaseT FDX");
	}
    else if (anlpar & ANLPAR_10HD) {
		dp->link.speed    = speed_10;
		dp->link.duplex   = half_duplex;
		dp->link.fc		  = fc_collision; 
		dp->link.linkstat = ETHER_SPEED_10HDX;
		p += sprintf(p,"10BaseT HDX");
	}
    else {
		dp->link.linkstat = ETHER_SPEED_UNKNOWN;
		dp->link.speed    = speed_unknown;
		dp->link.duplex   = unknown_duplex;
		dp->link.fc		  = fc_unknown; 
		p += sprintf(p,"Unknown");
	}
	printf("\n%s: %s\n","DP83820",buffer);

    return 1;
} //dp83820_mii_poll
