/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea-api.h
 *
 * Description  : Defines the interfaces of Eagle Agent (EA).
 *
 * Part I
 * ------
 * An instance of EA (Eagle Agent) runs on each NFX within the cluster. 
 * Upon system startup each EA will login to the configured AV (Agile Vision)
 * system.
 *
 * For any cluster wide configuration commands AV will talk to the
 * EA running on the PCC. For view and status commands AV will talk 
 * directly with the EA on the target NFX system.
 *
 * This file defines the interface between EA and AV/CLI. 
 
 * Since this interface is accesible to AV or CLI to keep AV up to date on
 * any configuration changes EA sends a notification (NTFY) to AV each time 
 * something has changed (creation, modification, deletion...). This 
 * message contains all necessary information AV need to update its
 * database.
 * 
 * This notification is sent to AV only if the change was not triggered by
 * AV itself.
 *
 * For each request (REQ) from AV, EA will return the result (RSP) of
 * the operation (ok/error) as well as the cookie (user info.) specified
 * in the request.
 *
 * All the requests sent by AV to EA can be triggered by CLIs on the NFX itself.
 * 
 *        AV                              EA
 *        |           NTFY                |
 *        |<------------------------------| <-- (Non AV initiated change)
 *        |                               |
 *         
 *        AV                              EA
 *        |           REQ                 |
 *        |------------------------------>|
 *        |                               |
 *        |           RSP(ok/error)       |
 *        |<------------------------------|
 *        |                               |
 *
 *        CLI                             EA
 *        |           REQ                 |
 *        |------------------------------>|
 *        |                               |
 *        |           RSP(ok/error)       |
 *        |<------------------------------|
 *        |                               | NTFY to AV
 *        |                               | ------------------->
 *
 * Created by   : Isabelle Cnudde
 *                Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
#ifndef _EA_API_H_
#define _EA_API_H_

/*------------------------------------------------------------------------------
 * Include files
 *------------------------------------------------------------------------------
 */

#include "exports/evm-api.h"
#include "exports/fs-api.h"
#include "exports/lnm-api.h"
#include "exports/cm-api.h"

/*------------------------------------------------------------------------------
 * Header of the message from/to AV.
 *------------------------------------------------------------------------------
 */

#define SNAP_MAX_NAME 	VOL_MAX_NAME
#define EXP_HOST_MAX	VOL_MAX_NAME
#define SMB_NAME_LEN	256

/*
 * EA return codes
 */

#define EA_NFX_OK		0 		// NFX_OK
#define EA_NFX_ERR		1		// NFX_ERR
#define EA_NFX_PENDING	2

/* 
 * Message type
 */
typedef enum
{
    /*
     * EA admin messages
     */
    EA_LOGIN_REQ        = EA_MSG_BASE, // 8192
    EA_LOGIN_RSP,

	/*
	 * SPM messages
	 *
	 * Req SPM -> AV
	 * Rsp AV -> SPM
	 * Indications AV -> SPM
	 *
	 */
	EA_SPM_GET_TOTAL_POOLS_REQ	= EA_SPM_MSG_BASE, // 8200
	EA_SPM_GET_TOTAL_POOLS_RSP,
	EA_SPM_GET_POOL_LIST_REQ,
	EA_SPM_GET_POOL_LIST_RSP,
	EA_SPM_GET_TOTAL_LUNS_REQ,
	EA_SPM_GET_TOTAL_LUNS_RSP,
	EA_SPM_GET_LUN_LIST_REQ,
	EA_SPM_GET_LUN_LIST_RSP,
	EA_SPM_MARK_LUN_AS_AGILE_REQ,
	EA_SPM_MARK_LUN_AS_AGILE_RSP,
	EA_SPM_MARK_LUN_AS_OC_REQ,
	EA_SPM_MARK_LUN_AS_OC_RSP,
	EA_SPM_UNMARK_LUN_AS_AGILE_REQ,
	EA_SPM_UNMARK_LUN_AS_AGILE_RSP,
	EA_SPM_RESCAN_LUN_REQ,
	EA_SPM_RESCAN_LUN_RSP,
	EA_SPM_GET_STORAGE_POOL_LIST_REQ,
	EA_SPM_GET_STORAGE_POOL_LIST_RSP,
	EA_SPM_GET_STORAGE_LUN_LIST_REQ,
	EA_SPM_GET_STORAGE_LUN_LIST_RSP, 
	
	EA_SPM_SRM_INIT_IND,
	EA_SPM_SRM_NEW_LUN_IND,
	EA_SPM_SRM_DEL_LUN_IND,
	EA_SPM_SRM_LUN_STATE_CHANGE_IND,


    /*
     * Notifications from NFX/EA to AV
     */
    EA_FSYS_NTFY     = EA_NTFY_MSG_BASE, // 8250
    EA_FSET_NTFY,
    EA_QUOTA_NTFY,
    EA_SNAP_NTFY,

    EA_OO_VOL_NTFY,
    EA_DEL_VOL_NTFY,
    EA_CRT_VOL_NTFY,
    EA_GRW_VOL_NTFY,
    EA_MOD_VOL_NTFY,
    EA_FULL_VOL_NTFY,
    EA_FAIL_LUN_NTFY,
    EA_TAKEOVER_VOL_NTFY,

    /*
     * File system related EA operations
     * Requests from AV to NFX/EA
     */
    EA_ADD_FSYS_REQ    = EA_FSYS_MSG_BASE, // 8300
    EA_ADD_FSYS_RSP,
    
    EA_MOD_FSYS_REQ,
    EA_MOD_FSYS_RSP,
    
    EA_MNT_FSYS_REQ,
    EA_MNT_FSYS_RSP,
    
    EA_UMNT_FSYS_REQ,
    EA_UMNT_FSYS_RSP,
    
    EA_DEL_FSYS_REQ,
    EA_DEL_FSYS_RSP,

    EA_NUM_FSYS_REQ,
    EA_NUM_FSYS_RSP,

    EA_GET_FSYS_REQ,
    EA_GET_FSYS_RSP,

    EA_GET_DETAIL_FSYS_REQ,
    EA_GET_DETAIL_FSYS_RSP,

    EA_USG_FSYS_REQ,
    EA_USG_FSYS_RSP,

    EA_MOVIN_FSYS_REQ,
    EA_MOVIN_FSYS_RSP,

    EA_MOVOUT_FSYS_REQ,
    EA_MOVOUT_FSYS_RSP,

    /*
     * File system related EA operations
     * Requests from NFX/EA to AV
     */
    EA_GROW_FSYS_REQ,
    EA_GROW_FSYS_RSP,
    EA_GROW_FSYS_IND,

	EA_FORCE_MOVIN_FSYS_REQ,
	EA_FORCE_MOVIN_FSYS_RSP,
    
    /*
     * File Sets related EA operations
     * Requests from AV to NFX/EA
     */
    EA_ADD_FSET_REQ     = EA_FSET_MSG_BASE, // 8350
    EA_ADD_FSET_RSP,

    EA_MOD_FSET_REQ,
    EA_MOD_FSET_RSP,
    
    EA_DEL_FSET_REQ,
    EA_DEL_FSET_RSP,

    EA_NUM_FSET_REQ,
    EA_NUM_FSET_RSP,

    EA_GET_FSET_REQ,
    EA_GET_FSET_RSP,

    EA_USG_FSET_REQ,
    EA_USG_FSET_RSP,

    /*
     * Export related EA operations
     */
    EA_ADD_EXP_REQ     = EA_EXPORT_MSG_BASE, // 8400
    EA_ADD_EXP_RSP,

    EA_MOD_EXP_REQ,
    EA_MOD_EXP_RSP,
    
    EA_DEL_EXP_REQ,
    EA_DEL_EXP_RSP,

    EA_DEL_PATH_REQ,
    EA_DEL_PATH_RSP,

    EA_GET_EXP_REQ,
    EA_GET_EXP_RSP,

    EA_GET_PATH_EXP_REQ,
    EA_GET_PATH_EXP_RSP,


    /*
     * Snap shot related EA operations
     */
    EA_CRT_SNAP_REQ   = EA_SNAPS_MSG_BASE, // 8450
    EA_CRT_SNAP_RSP,

    EA_DEL_SNAP_REQ,
    EA_DEL_SNAP_RSP,

    EA_SCH_SNAP_REQ,
    EA_SCH_SNAP_RSP,

    EA_GET_USG_SNAP_REQ,
    EA_GET_USG_SNAP_RSP,

    EA_GET_SCH_SNAP_REQ,
    EA_GET_SCH_SNAP_RSP,

    /*
     * Chassis related EA operations
     */
    EA_GET_CHASSIS_INFO_REQ = EA_CHASSIS_MSG_BASE, // 8480
    EA_GET_CHASSIS_INFO_RSP,

    /*
     * SDM related EA operations
     */
	EA_GET_FC_PORT_WWN_REQ = EA_SDM_MSG_BASE, // 8490
	EA_GET_FC_PORT_WWN_RSP,

	EA_LUN_REDISCOVER_REQ,
	EA_LUN_REDISCOVER_RSP,

    EA_GET_PORT_TOPOLOGY_REQ,
	EA_GET_PORT_TOPOLOGY_RSP,
	
    /*
     * RAID Controller related EA operations
     */

	/* RAID config manager miscellaneous messages */
	EA_RAID_MID_GET_LIST_REQ = EA_RAID_MSG_BASE, // 8500
	EA_RAID_MID_GET_LIST_RSP,	
	EA_RAID_MID_OPEN_REQ,
	EA_RAID_MID_OPEN_RSP,
	EA_RAID_MID_CLOSE_REQ,
	EA_RAID_MID_CLOSE_RSP,

	/* RAID Controller Messages */
	EA_RAID_MID_CTRL_INFO_REQ,	
	EA_RAID_MID_CTRL_INFO_RSP,	
	EA_RAID_MID_CTRL_GETPARAM_REQ, 
	EA_RAID_MID_CTRL_GETPARAM_RSP, 

	/* Physical Device Messages */
	EA_RAID_MID_PHYDEV_LIST_REQ, 
	EA_RAID_MID_PHYDEV_LIST_RSP, 
	EA_RAID_MID_PHYDEV_INFO_REQ, 
	EA_RAID_MID_PHYDEV_INFO_RSP, 

	/* Logical Device Messages */
	EA_RAID_MID_LOGDEV_LIST_REQ,
	EA_RAID_MID_LOGDEV_LIST_RSP, 
	EA_RAID_MID_LOGDEV_INFO_REQ, 
	EA_RAID_MID_LOGDEV_INFO_RSP, 

	/* Logical Device/LUN Config. Messages */
	EA_RAID_MID_LUN_CREATE_REQ,
	EA_RAID_MID_LUN_CREATE_RSP, 
	EA_RAID_MID_LUN_CREATE_SIMPLE_REQ,
	EA_RAID_MID_LUN_CREATE_SIMPLE_RSP,
	EA_RAID_MID_LUN_DELETE_REQ, 
	EA_RAID_MID_LUN_DELETE_RSP, 
	EA_RAID_MID_LUN_INIT_REQ,
	EA_RAID_MID_LUN_INIT_RSP,


	/* LUN Masking Messages */
	EA_RAID_MID_LUNMASK_INFO_REQ,
	EA_RAID_MID_LUNMASK_INFO_RSP,
	EA_RAID_MID_LUNMASK_ADD_REQ,
	EA_RAID_MID_LUNMASK_ADD_RSP,
	EA_RAID_MID_LUNMASK_DEL_REQ,
	EA_RAID_MID_LUNMASK_DEL_RSP,

	/* Misc. Messages */
	EA_RAID_MID_LUN_BPLIST_REQ,
	EA_RAID_MID_LUN_BPLIST_RSP,
	EA_RAID_MID_LUN_CANCELBP_REQ,
	EA_RAID_MID_LUN_CANCELBP_RSP

} ea_msgType_t;

/* 
 * Header of the message.
 */
typedef struct
{
    ea_msgType_t    msgType;    // the message type SHALL always be first
	uint32			msgLen;
	uint32			pad1;
    uint32          userInfo;
    char            node[EVM_MAX_NODE_NAME];
} ea_msgHdr_t;

#define EA_OUTPUT_LEN   512


/*------------------------------------------------------------------------------
 * LOGIN Request/response
 *------------------------------------------------------------------------------
 */

typedef struct
{
    ea_msgHdr_t     hdr;           			// header
	char   			name[256];         		// PCC serial Number
	char   			password[256];   		// encripted password
	char			clusterName[LNM_MAX_CLUSTER_NAME];
	uint64			isPCC;					// 1 - PCC, 0 - Non PCC
}loginReq_t;								// EA_LOGIN_REQ


typedef struct
{
    ea_msgHdr_t     hdr;           			// header
	uint32 			result; 				// Login success or fail
	char 			output[EA_OUTPUT_LEN]; 	// here is the reason if fails
}loginRsp_t;								// EA_LOGIN_RSP


/*------------------------------------------------------------------------------
 * Common data structures
 *------------------------------------------------------------------------------
 */

/* 
 * File system identification
 */
typedef struct
{
    evm_volId_t     volId;                      // Volume id
    char            fsysName[VOL_MAX_NAME];     // file system name
} ea_fsysId_t;


/*------------------------------------------------------------------------------
 * FILE SYSTEM OPERATIONS 
 *------------------------------------------------------------------------------
 */

/* 
 * List of LUNs (physical storage).
 */
typedef struct
{
    uint64  wwn; 
    uint64  lun;
} ea_lun_t;

typedef struct
{
    uint64      num;
    ea_lun_t    lun[1];
} ea_lunList_t;

/*
 * Basic file system config
 */
typedef struct
{
    ea_fsysId_t          fsysId;                // file system id

    /*
     * Info to be obtained from EFS
     */
    uint32               fsState;              // from file system
    uint32               mountStatus;           // Mounted/unmounted
                                                // 0 = Unmounted; 1 = Mounted
    uint64               usedSize;              // current used size

    /*
     * Fields we need from evm_VolConfig_t and
     * evm_strg_attributes_t structure
     */
    uint64               maxSize;               // max vol size/hardQuota
    uint64               currSize;              // current vol size
    int32                onlineStatus;          // vol status 
                                                // 0 = Not Online; 1 = Online
    uint32               highWaterMark;         // Get from EFS now

#define DISABLE_HIGHWATERMARK	0
#define MIN_HIGHWATERMARK      25
#define MAX_HIGHWATERMARK      100

    uint64               minAutoGrow;           // Growth increment (blocks)
	uint32				 softQuota;				// soft Quota
	evm_volType_t        type;                  // file system type
    evm_raid_t           raidLevel;             // raid level

	uint32				 numVendors;
    evm_vendor_model_t   vendorInfo[EVM_MAX_VENDOR_LIST];            
                                                // vendor name and model
} ea_fsysCfg_t;

NFX_ASSERT_STATIC((sizeof(ea_fsysCfg_t) % 8) == 0);

typedef struct
{
    uint32            nb;
	uint32			 pad1;
    ea_fsysCfg_t    fsys[1];
} ea_fsysList_t;

/*
 * Detailed file system config
 */
typedef struct
{
    evm_VolConfig_t       volCfg;

    /*
     * Info to be obtained from EFS
     */
    uint32              highWaterMark;       // How full before growth
	uint32				pad1;
    uint32              fsState;             // from file system
    uint32              mountStatus;         // Mounted/unmounted
                                             // 0 = Unmounted; 1 = Mounted
    uint64              usedSize;            // current used size

    /*
     * List of LUNs obtained from EVM
     */
	uint32          	nbStrg;     		 // number of LUNs
    evm_lvStrg_t        strg[1];             // list of LUNs allocated
} ea_fsysDetail_t;


/*------------------------------------------------------------------------------
 * Requests from AV to NFX - File system
 *----------------------------------------------------------------------------*/

/* Add a file system
 * -----------------
 */
typedef struct
{
    ea_msgHdr_t             hdr;           // header
    ea_fsysId_t             fsysId;        // file system id
    uint64                  volSize;       // start vol size
    uint64                  maxSize;       // hard Quota
    uint64                  minAutoGrow;   // min auto grow
	uint32					pad1;			// added after minAutoGrow
											// changed from 32 to 64 bits
    uint32                  highWaterMark; // high water mark
    uint32                  nim;           // N-NIM for the file system
	uint32					softQuota;	   // % when to notify operator
    evm_strg_attributes_t   attr;          // storage pool attributes
} ea_addFsysReq_t;                         // EA_ADD_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_addFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysCfg_t    fsysCfg;                // file system config
} ea_addFsysRsp_t;                          // EA_ADD_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_addFsysRsp_t) % 8) == 0);

/* Modify file system
 * ------------------
 */
#define FSYS_NAME_OFFSET            0
#define FSYS_HQUOTA_OFFSET          1
#define FSYS_SQUOTA_OFFSET          2
#define FSYS_MINAUTOGROW_OFFSET     3
#define FSYS_HIGHWATERMARK_OFFSET   4

#define FSYS_NAME_MASK              (1 << FSYS_NAME_OFFSET)
#define FSYS_HQUOTA_MASK             (1 << FSYS_HQUOTA_OFFSET)
#define FSYS_SQUOTA_MASK             (1 << FSYS_SQUOTA_OFFSET)
#define FSYS_MINAUTOGROW_MASK          (1 << FSYS_MINAUTOGROW_OFFSET)
#define FSYS_HIGHWATERMARK_MASK     (1 << FSYS_HIGHWATERMARK_OFFSET)

#define IS_FSYS_NAME(mask)          ((mask & FSYS_NAME_MASK)? TRUE : FALSE)
#define IS_FSYS_HQUOTA(mask)         ((mask & FSYS_HQUOTA_MASK)? TRUE : FALSE)
#define IS_FSYS_SQUOTA(mask)         ((mask & FSYS_SQUOTA_MASK)? TRUE : FALSE)
#define IS_FSYS_MINAUTOGROW(mask)    ((mask & FSYS_MINAUTOGROW_MASK)? TRUE : FALSE)
#define IS_FSYS_HIGHWATERMARK(mask)    ((mask & FSYS_HIGHWATERMARK_MASK)? TRUE : FALSE)

#define SET_FSYS_NAME(mask)             (mask |= FSYS_NAME_MASK)
#define SET_FSYS_HQUOTA(mask)            (mask |= FSYS_HQUOTA_MASK)
#define SET_FSYS_SQUOTA(mask)            (mask |= FSYS_SQUOTA_MASK)
#define SET_FSYS_MINAUTOGROW(mask)         (mask |= FSYS_MINAUTOGROW_MASK)
#define SET_FSYS_HIGHWATERMARK(mask)       (mask |= FSYS_HIGHWATERMARK_MASK)

#define CLR_FSYS_NAME(mask)             (mask &= ~FSYS_NAME_MASK)
#define CLR_FSYS_HQUOTA(mask)            (mask &= ~FSYS_HQUOTA_MASK)
#define CLR_FSYS_SQUOTA(mask)            (mask &= ~FSYS_SQUOTA_MASK)
#define CLR_FSYS_MINAUTOGROW(mask)        (mask &= ~FSYS_MINAUTOGROW_MASK)
#define CLR_FSYS_HIGHWATERMARK(mask)    (mask &= ~FSYS_HIGHWATERMARK_MASK)

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
    uint32          mask;                   // indicate what to modify
    uint32          highWaterMark;          // high water mark
    uint64          hardQuota;              // hard Quota
    uint64          minAutoGrow;            // min auto grow
	uint32			pad1;					// added after change minautogrow
											// from 32 to 64 bits
	uint32			softQuota;				// % when to notify operator
    char            newFsysName[VOL_MAX_NAME];  // new file system name
    
} ea_modFsysReq_t;                          // EA_MOD_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_modFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysCfg_t    fsysCfg;                // file system config
} ea_modFsysRsp_t;                          // EA_MOD_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_modFsysRsp_t) % 8) == 0);

/* Mount file system
 * -----------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
} ea_mntFsysReq_t;                          // EA_MNT_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_mntFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysId_t     fsysId;                 // file system id
} ea_mntFsysRsp_t;                          // EA_MNT_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_mntFsysRsp_t) % 8) == 0);

/* Unmount file system
 * -------------------
 */
typedef struct
{
    ea_msgHdr_t        hdr;                 // header
    ea_fsysId_t        fsysId;              // file system id
} ea_umntFsysReq_t;                         // EA_UMNT_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_umntFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysId_t     fsysId;                 // file system id
} ea_umntFsysRsp_t;                         // EA_UMNT_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_umntFsysRsp_t) % 8) == 0);

/* Delete file system
 * ------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
} ea_delFsysReq_t;                          // EA_DEL_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_delFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
} ea_delFsysRsp_t;                          // EA_DEL_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_delFsysRsp_t) % 8) == 0);

/* Get num file system
 * -------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header - specify hdr.node
											//          if necessary...
} ea_numFsysReq_t;                          // EA_NUM_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_numFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    int32           totNum;                 // total number of file systems
	uint32			pad2;
} ea_numFsysRsp_t;                          // EA_NUM_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_numFsysRsp_t) % 8) == 0);

/* Get file system(s) configuration
 * --------------------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header, with node specified
											// in hdr.node
    int32           reqType;          
#define EA_REQ_HAS_FSYS_INIT        1       // entry 1 start id of first entry
                                            // id = 0 means use start of table
#define EA_REQ_HAS_FSYS_LIST        2       // num desired entries are listed 
                                            // in array

    int32           num;                    // how many entries to return
                                            // num = 0 means all entries
    ea_fsysId_t     fsysId[1];              // file system id

} ea_getFsysReq_t;                          // EA_GET_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_getFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysList_t   fsysList;               // list of file systems
} ea_getFsysRsp_t;                          // EA_GET_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_getFsysRsp_t) % 8) == 0);

/* Get file system detail
 * ----------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
} ea_getDetailFsysReq_t;                    // EA_GET_DETAIL_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_getDetailFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysDetail_t fsysDetail;             // detail file system info
} ea_getDetailFsysRsp_t;                    // EA_GET_DETAIL_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_getDetailFsysRsp_t) % 8) == 0);

/* Usage file system request
 * -------------------------
 */

typedef   struct 
{
    ea_fsysId_t     fsysId;                 // file system id
    uint64          usedSize;
    uint64          allocatedSize;
} fsysUsage_t;

NFX_ASSERT_STATIC((sizeof(fsysUsage_t) % 8) == 0);
 
typedef struct
{
    ea_msgHdr_t     hdr;					// if request info from another node
											// hdr.node should specify node name
    uint64          num;                    // num ids included
    ea_fsysId_t     fsysList[1];            // list of file system 
} ea_usgFsysReq_t;                          // EA_USG_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_usgFsysReq_t) % 8) == 0);
 
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    uint64          num;                    // num usage records
    fsysUsage_t     usageList[1];           // usage info
} ea_usgFsysRsp_t;                          // EA_USG_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_usgFsysRsp_t) % 8) == 0);

/* Move file system out the NFX
 * ---------------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
} ea_movOutFsysReq_t;                       // EA_MOVOUT_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_movOutFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
} ea_movOutFsysRsp_t;                       // EA_MOVOUT_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_movOutFsysRsp_t) % 8) == 0);

/* Move file system in the NFX
 * ---------------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsysId_t     fsysId;                 // file system id
    uint32          nim;                    // N-NIM serving the file system
	uint32			forceMoveIn;			// Should the movin be forced
	uint32			forceTime;				// how long to try force takeover
} ea_movInFsysReq_t;                        // EA_MOVIN_FSYS_REQ
											// EA_FORCE_MOVIN_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_movInFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsysCfg_t    fsysCfg;                // file system config
} ea_movInFsysRsp_t;                        // EA_MOVIN_FSYS_RSP
											// EA_FORCE_MOVIN_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_movInFsysRsp_t) % 8) == 0);

/*------------------------------------------------------------------------------
 * Requests from NFX to AV - File system
 *----------------------------------------------------------------------------*/

/* Grow file system in the NFX
 * ---------------------------
 */
typedef struct
{
    ea_msgHdr_t     		hdr;                   	// header
    char            		node[EVM_MAX_NODE_NAME];
	uint64					txnId;					// transaction id
    uint64          		neededSize;             // requested space to grow
    evm_strg_attributes_t   attr;          			// storage pool attributes
} ea_growFsysReq_t;                        			// EA_GROW_FSYS_REQ

NFX_ASSERT_STATIC((sizeof(ea_growFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    char            node[EVM_MAX_NODE_NAME];
	uint64			txnId;					// transaction id
} ea_growFsysRsp_t;                         // EA_GROW_FSYS_RSP

NFX_ASSERT_STATIC((sizeof(ea_growFsysReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    char            node[EVM_MAX_NODE_NAME];
	uint64			txnId;					// transaction id
    // ea_lunList_t    lunList;				// LUN list alocated
    uint32          num;                    // number of devices
    uint32          pad2;
    char            devName[1][SDM_DEV_NAME_LEN+8]; // list of devices
} ea_growFsysInd_t;                        	// EA_GROW_FSYS_IND

NFX_ASSERT_STATIC((sizeof(ea_growFsysInd_t) % 8) == 0);

/*------------------------------------------------------------------------------
 * FILE SET OPERATIONS 
 *------------------------------------------------------------------------------
 */

/* 
 * File set configuration.
 */

typedef struct
{
    ea_fsysId_t     fsysId;                     // file system id
    char            path[FS_PATH_MAX];          // Path
} ea_fsetId_t;

typedef struct 
{
	uint64   		inodeSoftLimit;		// Inode soft limit
	uint64  		inodeHardLimit; 	// Inode hard limit
	uint64  		inodeTimeLimit; 	// Inode time limit
	uint64   		blockSoftLimit; 	// block soft limit
	uint64   		blockHardLimit; 	// block hard limit
	uint64   		blockTimeLimit; 	// block time limit
} ea_quotaCfg_t;

typedef struct 
{
	uint64      	inodesConsumed;     // number of inodes consumed
    uint64      	blocksConsumed;     // number of blocks consumed 
	uint32      	inodeSoftLimitTime; // time when inode softlimit was crossed
	uint32      	blockSoftLimitTime; // time when block softlimit was crossed
} ea_quotaUsage_t;

typedef struct
{
	ea_fsetId_t 	fsetId; 			// file system id
	uint32  		quotaOn; 			// quota (on/off)
	uint32			pad1;
	ea_quotaCfg_t	quotaCfg;			// Quota configuration
	ea_quotaUsage_t	quotaUsage;			// Quota usage
} ea_fsetCfg_t;

NFX_ASSERT_STATIC((sizeof(ea_fsetCfg_t) % 8) == 0);

typedef struct
{
    uint32          nb;
	uint32			pad1;
    ea_fsetCfg_t    fset[1];
} ea_fsetList_t;

/*------------------------------------------------------------------------------
 * Requests from AV to NFX - File set
 *----------------------------------------------------------------------------*/

/* Add file set
 * ------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsetId_t     fsetId;                 // file set id
	uint32  		quotaOn; 				// quota (on/off)
	uint32			pad1;
	ea_quotaCfg_t	quotaCfg;				// Quota configuration and usage
} ea_addFsetReq_t;                          // EA_ADD_FSET_REQ;

NFX_ASSERT_STATIC((sizeof(ea_addFsetReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsetCfg_t    fsetCfg;                // file set config
} ea_addFsetRsp_t;                          // EA_ADD_FSET_RSP

NFX_ASSERT_STATIC((sizeof(ea_addFsetRsp_t) % 8) == 0);

/* Modify file set
 * ---------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsetId_t     fsetId;                 // file set id
	uint32  		quotaOn; 				// quota (on/off)
	uint32			pad1;
	ea_quotaCfg_t	quotaCfg;				// Quota configuration and usage
} ea_modFsetReq_t;                          // EA_MOD_FSET_REQ;

NFX_ASSERT_STATIC((sizeof(ea_modFsetReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsetCfg_t    fsetCfg;                // file set config
} ea_modFsetRsp_t;                          // EA_MOD_FSET_RSP

NFX_ASSERT_STATIC((sizeof(ea_modFsetRsp_t) % 8) == 0);

/* Delete file set
 * ---------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_fsetId_t     fsetId;                 // file set id
} ea_delFsetReq_t;                          // EA_DEL_FSET_REQ;

NFX_ASSERT_STATIC((sizeof(ea_delFsetReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsetId_t     fsetId;                 // file set id
} ea_delFsetRsp_t;                          // EA_DEl_FSET_RSP

NFX_ASSERT_STATIC((sizeof(ea_delFsetRsp_t) % 8) == 0);

/* Num file set
 * ----------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header - specify hdr.node if req
											//          is for cluster node
} ea_numFsetReq_t;                          // EA_NUM_FSET_REQ

NFX_ASSERT_STATIC((sizeof(ea_numFsetReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    int32           totNum;                 // total number of file sets
	uint32			pad2;
} ea_numFsetRsp_t;                          // EA_NUM_FSET_RSP

NFX_ASSERT_STATIC((sizeof(ea_numFsetRsp_t) % 8) == 0);

/* Get file set(s) configuration
 * -----------------------------
 */
typedef struct
{
    ea_msgHdr_t     hdr;                    // header

    int32           reqType;          
#define EA_REQ_HAS_FSET_INIT        1       // entry 1 start id of first entry
                                            // id = 0 means use start of table
#define EA_REQ_HAS_FSET_LIST        2       // num desired entries are listed 
                                            // in array

    int32           num;                    // how many entries to return
                                            // num = 0 means all entries
    ea_fsetId_t     fsetIds[1];            // list of file sets
	// IMPORTANT NOTES:
	// 1. If ea_fsetId_t are specified, only fsysId on the same node can be
	// specified at the same time. EA will only talk to the node, on which
	// the fsetIds[0].fsysId.volId (first one) resides, to retrieve the
	// fset info
	// 2. hdr.node is checked first to see if a node name is specified there.
	// If not, fsetIds[0].fsysId.volId is checked. If it's non-zero, then this
	// ID will be used to locate the node that has this volume mounted. If
	// it's 0, then fsetIds[0].fsysId.fsysName is used to get the volume
	// ID and node name for this volume. If fsysName is also null string, then
	// local node is used.
} ea_getFsetReq_t;                          // EA_GET_FSET_REQ

NFX_ASSERT_STATIC((sizeof(ea_getFsetReq_t) % 8) == 0);

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    uint32          result;                 // operation result
	uint32			pad1;
    char            output[EA_OUTPUT_LEN];  // reason
    ea_fsetList_t   fsetList;               // list of file sets
} ea_getFsetRsp_t;                          // EA_GET_FSET_RSP

NFX_ASSERT_STATIC((sizeof(ea_getFsetRsp_t) % 8) == 0);

/* Usage file set request
 * ----------------------
 */

typedef struct
{
	ea_fsetId_t 	fsetId; 		// file system id
	ea_quotaUsage_t	quotaUsage;		// Quota configuration and usage
} fsetUsage_t;

typedef struct
{
	ea_msgHdr_t 	hdr; 					// header
	uint64 			num; 					// num ids included
	ea_fsetId_t 	fsetIds[1]; 			// list of file sets
	// IMPORTANT NOTES:
	// 1. If ea_fsetId_t are specified, only fsysId on the same node can be
	// specified at the same time. EA will only talk to the node, on which
	// the fsetIds[0].fsysId.volId (first one) resides, to retrieve the
	// fset info
	// 2. hdr.node is checked first to see if a node name is specified there.
	// If not, fsetIds[0].fsysId.volId is checked. If it's non-zero, then this
	// ID will be used to locate the node that has this volume mounted. If
	// it's 0, then fsetIds[0].fsysId.fsysName is used to get the volume
	// ID and node name for this volume. If fsysName is also null string, then
	// local node is used.
} ea_usgFsetReq_t; 							// EA_USG_FSET_REQ

typedef struct
{
	ea_msgHdr_t 	hdr; 					// header
	uint32 			result; 				// operation result
	uint32			pad1;
	char 			output[EA_OUTPUT_LEN]; 	// reason
	uint64 			num; 					// num usage records
	fsetUsage_t 	usageList[1]; 			// usage info
} ea_usgFsetRsp_t; 							// EA_USG_FSET_RSP


/*------------------------------------------------------------------------------
 * SNAPHOT OPERATIONS
 *------------------------------------------------------------------------------
 */


/*------------------------------------------------------------------------------
 * Requests from AV to NFX - Snap shot 
 *----------------------------------------------------------------------------*/

 
/* Create snapshot 
 * -----------------
 */
 
typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
	ea_fsysId_t     fsysId;                     // file system id
    char            snapName[SNAP_MAX_NAME];	// snap name
} ea_crtSnapReq_t;  							// EA_CRT_SNAP_REQ                        
 

typedef struct 
{ 
	ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
	ea_fsysId_t     fsysId;                     // file system id
    char            snapName[SNAP_MAX_NAME];	// snap name
} ea_crtSnapRsp_t; 								// EA_CRT_SNAP_RSP
 

/* Delete snapshot 
 * -----------------
 */
 
typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
	ea_fsysId_t     fsysId;                     // file system id
    char            snapName[SNAP_MAX_NAME];	// snap name
} ea_delSnapReq_t;        						// EA_DEL_SNAP_REQ
 

typedef struct 
{ 
	ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
	ea_fsysId_t     fsysId;                     // file system id
    char            snapName[SNAP_MAX_NAME];	// snap name
} ea_delSnapRsp_t;								// EA_DEL_SNAP_RSP 
 
 
/* Schedule snapshot
 * -----------------
 */

typedef struct
{
    ea_fsysId_t    	fsysId;
	uint32			pad1;
    uint32          nbWeekSnap;         // Number of Weekly Snapshots to Keep
    uint32          nbDaySnap;          // Number of Daily Snapshots to Keep
    uint32          nbHourSnap;         // Number of Hourly Snapshots to Keep
    uint32          hourSchedule[FS_SNAP_SCHED_MAX_HOURS];   
										// AM: index 0-11, PM: index 12-23; 
                                        // taking bool value
} ea_snapSch_t;                         // snapshot configuration

 
typedef struct
{
    ea_msgHdr_t     hdr;                   	 	// header
    ea_snapSch_t   	snapSch;  					// schedule config from snap
} ea_schSnapReq_t;                         		// EA_SCH_SNAP_REQ 
 
typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
	ea_snapSch_t	snapSch;
} ea_schSnapRsp_t; 								// EA_SCH_SNAP_RSP
 

/* Get snapshot schedule (for any number of file systems)
 * ---------------------
 */
 
typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    uint32          num;
	uint32			pad1;
    ea_fsysId_t     fsysId[1];    
} ea_getSchSnapReq_t;                        	// EA_GET_SCH_SNAP_REQ 
 

typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
	uint32			pad2;
	uint32     		num;						// num entries in list
    ea_snapSch_t 	snapList[1]; 				// file system id 
} ea_getSchSnapRsp_t; 							// EA_GET_SCH_SNAP_RSP
 

/* Get snapshot usage
 * -----------------
 */

#define DATE_STR_LEN	64

typedef struct
{
    char            name[SNAP_MAX_NAME]; 	// snapshot name 
    uint32          pct;					// snapshot size as % of FS size
	uint32			pad1;
    uint64          blks;					// size in blocks for snapshot
    char            date[DATE_STR_LEN]; 	// time when snapshot was created
} ea_snapUsg_t;                	
 
typedef struct
{
    ea_fsysId_t     fsysId;
	uint32     		num;						// num entries in list
	uint32			pad1;
    ea_snapUsg_t    usage[1]; 
} ea_snapUsgList_t;               	
 
typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    ea_fsysId_t     fsysId;    
} ea_getUsgSnapReq_t;                        	// EA_GET_USG_SNAP_REQ 
 
typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
    ea_snapUsgList_t usageList; 				// snapshot usage list
} ea_getUsgSnapRsp_t; 							// EA_GET_USG_SNAP_RSP
 
 
 
 
/*------------------------------------------------------------------------------
 * EXPORT OPERATIONS
 *------------------------------------------------------------------------------
 */

typedef struct
{
	uint32    		hostIP;
	uint32			subnet;
    uint32    		root; 						// yes, no
    uint32    		access;						// yes, no
    uint32    		rtsquash;					// yes, no
    uint32    		ro_rw; 						// 0-none, 1-ro, 2-rw     
} ea_expHost_t;

typedef struct
{
    char    pathName[FS_PATH_MAX];
} ea_pathId_t;

typedef struct
{
	ea_pathId_t		path;
    uint32    		shareIP;
	uint32			num;
	ea_expHost_t	expHostList[1];
} ea_expCfg_t;  


/* Export
 * ------
 */

typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    ea_fsysId_t 	fsysId;
    ea_expCfg_t     expCfg; 
} ea_addExpReq_t;                         		// EA_ADD_EXP_REQ 
 

typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
    ea_fsysId_t 	fsysId;
    ea_expCfg_t     expCfg; 
} ea_addExpRsp_t;								// EA_ADD_EXP_RSP 
 

/* Modify export
 * -------------
 */

typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    ea_fsysId_t 	fsysId;
    ea_expCfg_t     expCfg; 
} ea_modExpReq_t;                         		// EA_MOD_EXP_REQ 
 

typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
    ea_fsysId_t 	fsysId;
    ea_expCfg_t     expCfg; 
} ea_modExpRsp_t;                         		// EA_MOD_EXP_RSP 
 
 
/* Un-export (Delete export)
 * -------------------------
 */
 
typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    ea_fsysId_t 	fsysId;
    ea_expCfg_t     expCfg; 
} ea_delExpReq_t;    							// EA_DEL_EXP_REQ                      
 
typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
    ea_fsysId_t 	fsysId;
    ea_expCfg_t     expCfg; 
} ea_delExpRsp_t; 								// EA_DEL_EXP_RSP

/* Un-export (Delete path)
 * -------------------------
 */

typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
	ea_fsysId_t     fsysId;
	uint32          num;                        // send multiple deletions
	uint32			pad1;
	ea_pathId_t     pathList[1];
} ea_delPathReq_t;      						// EA_DEL_PATH_REQ                 

typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
	ea_fsysId_t     fsysId;
	uint32          num;                        // path num deleted
	uint32			pad2;
	ea_pathId_t     pathList[1];
} ea_delPathRsp_t; 								// EA_DEL_PATH_RSP
 
/* Get export info on a file system
 * --------------------------------
 */
 
typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    ea_fsysId_t     fsysId;      				// must specify fsysId(name)
} ea_getExpReq_t;                          		// EA_GET_EXP_REQ
 
typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
    ea_fsysId_t 	fsysId;
	uint32          num;                        // path num deleted
	uint32			pad2;
    ea_expCfg_t    	expList[1];
} ea_getExpRsp_t; 								// EA_GET_EXP_RSP

/* Get export info for a given path
 * --------------------------------
 */

typedef struct
{
    ea_msgHdr_t     hdr;                    	// header
    ea_fsysId_t     fsysId;      			
	ea_pathId_t		path;
} ea_getPathExpReq_t;                      		// EA_GET_PATH_EXP_REQ 
 
typedef struct 
{ 
    ea_msgHdr_t 	hdr; 						// header 
    uint32 			result; 					// operation result 
	uint32			pad1;
    char 			output[EA_OUTPUT_LEN]; 		// reason 
    ea_fsysId_t 	fsysId;
    ea_expCfg_t    	expCfg;
} ea_getPathExpRsp_t; 							// EA_GET_PATH_EXP_RSP

/*------------------------------------------------------------------------------
 * CHASSIS/FC-PORT OPERATIONS
 *------------------------------------------------------------------------------
 */

typedef struct
{
	uint32		slotId;
	uint32		portId;
} SlotPort_t;

typedef struct
{
	uint32		slotId;
	uint32 		portId;
	int64		wwn;
} SlotPortWwn_t;

typedef struct
{
	ea_msgHdr_t			hdr; 						// header
} ea_getChassisInfoReq_t;							// EA_GET_CHASSIS_INFO_REQ

typedef struct 
{ 
	ea_msgHdr_t				hdr;					// header 
	uint32					result;					// operation result 
	uint32					pad;
	char					output[EA_OUTPUT_LEN];	// reason 
	chassisdMsgInfoResp_t	chInfo;
} ea_getChassisInfoRsp_t;							// EA_GET_CHASSIS_INFO_RSP

typedef struct
{
	ea_msgHdr_t		hdr;
	uint64			num;					// how many ports
	SlotPort_t		portList[1];			// list of slotId, portId
} ea_listFcPortWwnReq_t;					// EA_GET_FC_PORT_WWN_REQ

typedef struct
{
	ea_msgHdr_t		hdr;					// header 
	uint32			result;					// operation result 
	uint32			pad1;
	char			output[EA_OUTPUT_LEN];	// reason 
	uint64			num;			
	SlotPortWwn_t	portList[1];			// list of port/slot/wwn
} ea_listFcPortWwnRsp_t;					// EA_GET_FC_PORT_WWN_RSP

/*------------------------------------------------------------------------------
 * Raid Control Messages
 *------------------------------------------------------------------------------
 */

/*
 *-----------------------------------------------------------
 * RAID Result Codes
 *-----------------------------------------------------------
 */
#define EA_RAID_RCODE_SUCCESS					0
#define EA_RAID_RCODE_NOT_OPEN					1
#define EA_RAID_RCODE_NO_DEVICE					2
#define EA_RAID_RCODE_BUSY						3
#define EA_RAID_RCODE_IO_ERROR					4
#define EA_RAID_RCODE_INVALID_PARAMETERS		5
#define EA_RAID_RCODE_INVALID_PHYSICAL_DEVICE	6
#define EA_RAID_RCODE_INVALID_LUN				7
#define EA_RAID_RCODE_INSUFFICIENT_SPACE		8
#define EA_RAID_RCODE_TOO_MANY_LUNS				9
#define EA_RAID_RCODE_UNABLE_TO_DELETE_LUN		10
#define EA_RAID_RCODE_LUN_IS_BUSY				11
#define EA_RAID_RCODE_NOT_IMPLEMENTED			20


/*
 *-----------------------------------------------------------
 * RAID Controller Id
 *-----------------------------------------------------------
 */

#define EA_RAID_FIELD_LEN_MAX	64  /*  Any Raid character string len should 
										not exceeed this and should be null
										terminated */

typedef char ea_raid_ctrl_id_t[EA_RAID_FIELD_LEN_MAX];

/*
 ******************************************************************
 *
 * Messages for RAID controller discovery
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Get RAID List Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
} ea_raid_getListReqMsg_t;			// EA_RAID_MID_GET_LIST_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
	uint32				numEntries;	// Number of entries in idList
	ea_raid_ctrl_id_t	idList[1];	// Variable len list
} ea_raid_getListRspMsg_t; 			// EA_RAID_MID_GET_LIST_RSP

/*
 *-----------------------------------------------------------
 * Connection Establishment/Teardown Messages
 *-----------------------------------------------------------
 */

typedef struct 
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
} ea_raid_openReqMsg_t;				// EA_RAID_MID_OPEN_REQ

typedef struct 
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
	uint32				lun;        // LUN of open device
	scsi_id_t			scsi_id;    // 64-bit SCSI ID of open device
	uint64				wwn;        // WWN of open device
} ea_raid_openRspMsg_t;				// EA_RAID_MID_OPEN_RSP

typedef struct 
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
} ea_raid_closeReqMsg_t;			// EA_RAID_MID_CLOSE_REQ
								
								
typedef struct 
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;      // return code
    uint32              pad1;       // padding
} ea_raid_closeRspMsg_t;			// EA_RAID_MID_CLOSE_REQ

/*
 ******************************************************************
 *
 * Messages for getting RAID controller Info./State/Parameters
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Get RAID Controller Info. Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
} ea_raid_ctrlInfoReqMsg_t;			// EA_RAID_MID_CTRL_INFO_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
	uchar8				bbuPresent;
	uchar8				pad1;		// padding
	short16				pad2;		// padding
	char				ctrlType[EA_RAID_FIELD_LEN_MAX];
	char				ctrlName[EA_RAID_FIELD_LEN_MAX];
	char				serialNumber[EA_RAID_FIELD_LEN_MAX];
	char				vendorName[EA_RAID_FIELD_LEN_MAX];
	char				hardwareVer[EA_RAID_FIELD_LEN_MAX];
	char				firmwareVer[EA_RAID_FIELD_LEN_MAX];
	uint32				cacheInMB;
	uint32 				memoryInMB;
	uint32				numDriveChannels;
	uint32				maxDrivesPerChannel;
	uint64				ctrlWwn;
	uint64				dualCtrlWwn;
	uint32				numLogicalDevices;
	uint32				numPhysicalDevices;
} ea_raid_ctrlInfoRspMsg_t; 		// EA_RAID_MID_CTRL_INFO_RSP


/*
 *-----------------------------------------------------------
 * RAID Controller Parameter Structure
 *-----------------------------------------------------------
 */

typedef struct
{
	uchar8			readAhead;		// Is read ahead enabled
	uchar8			reasRestrict;	// Is reassign restricted to 1 block enabled
	uchar8			trueVerify;		// Is true verification of data enabled
	uchar8			writeVerify;	// Is write through verification enabled
	uchar8			bgInit;			// Is background init enabled
	uchar8			superReadAhead;	// Is super read ahead enabled
	uchar8			coalOptimal;	// Is coalescing optimisation enabled
	uchar8			autoRebuild;	// Is auto rebuild enabled
	uchar8			operFault;		// Is operation fault manage enabled
	uchar8			vendorTUR;		// Is vendor unique TUR enabled
	uchar8			disCCInvLun;	// Is disable CC for invalid LUN enabled
	uchar8			noPause;		// Is no pause on ctler nor ready enabled
	uchar8			queueFullBusy;	// Is on queue full give busy enabled
	uchar8			disBusyFailback;// Is disable busy on Failback enabled
	uchar8			nodeNameRet;	// Is node name retention enabled
	uchar8			safteofups;		// Is SAF-TE use of UPS enabled
	uchar8			resetProp;		// Is reset propagation enabled
	uchar8			smartLrgTrans;	// Is smart large transfer enabled
	uchar8			autoFailback;	// Is auto failback enabled
	uchar8			forceSimplex;	// Is force simplex enabled
	uchar8			conserveCache;	// Is conservative cache mode enabled
	uchar8			noRSTCOM;		// Is simplex no RSTCOM enabled
	uchar8			ctrl0Port0;		// Is C0P0 enabled
	uchar8			ctrl0Port1;		// Is C0P1 enabled
	uchar8			ctrl1Port0;		// Is C1P0 enabled
	uchar8			ctrl1Port1;		// Is C1P1 enabled
	uchar8			ctrl0Port0LoopId;
	uchar8			ctrl0Port1LoopId;
	uchar8			ctrl1Port0LoopId;
	uchar8			ctrl1Port1LoopId;
	uchar8			debugDump;		// Is debug dump enabled
    byte            pad1;           // padding

	short16			rebuildRate;	// Default rebuild rate (0 - 50)
	short16			initRate;		// Default init rate (0 - 50)
	short16			consistencyCheckRate;   
									// Default Consistency Check rate (0 - 50)

	short16			startupMode;	// Disk startup mode
#define EA_RAID_STARTUP_AUTO	0	// automatic
#define EA_RAID_STARTUP_POWER	1	// on power
#define EA_RAID_STARTUP_CMD		2	// on command

	short16			startupNumDev;	// Startup number of devices (1 - 8)
	short16			startupDelay;	// Startup delay (0 - 255)
	short16			scsiStartupDelay;	// SCSI startup delay (0 - 255)

	short16			topology;		// Topology
#define EA_RAID_TOPO_INACTPORT	0	// Inactive Port
#define EA_RAID_TOPO_MULTIPORT	1	// Multiport
#define EA_RAID_TOPO_CLUSTER	2	// Clustering
#define EA_RAID_TOPO_MULTITID	3	// MultiTID
#define EA_RAID_TOPO_MASSLAV	4	// Master Slave

	short16			rs232PortType;	// RS232 port type
#define EA_RAID_PORT_DEBUG		0	// debug
#define EA_RAID_PORT_SLPVT100	3	// SLP/VT 100

	short16			rs232BaudRate;	// RS232 baud rate 
									// (2400, 4800, 9600, or 19200)
	short16			frameSize;		// 2048, 1024, or 512
	short16			pciLatCtrl;		// PCI latency control
#define EA_RAID_PLC_LONG		0	// long
#define EA_RAID_PLC_MEDIUM		1	// medium
#define EA_RAID_PLC_SHORT		2	// short

	short16			queueLimit;		// Queue limit (1 - 230)

	short16			rofInterval;	// ROF rearm interval
#define EA_RAID_ROFI_NEVER		0	// never
#define EA_RAID_ROFI_3min		1	// 3 minutes
#define EA_RAID_ROFI_5min		2	// 5 minutes
#define EA_RAID_ROFI_15min		3	// 15 minutes
#define EA_RAID_ROFI_30min		4	// 30 minutes
#define EA_RAID_ROFI_60min		5	// 60 minutes
#define EA_RAID_ROFI_90min		6	// 90 minutes
#define EA_RAID_ROFI_2hour		7	// 2 hours
#define EA_RAID_ROFI_4hour		8	// 4 hours
#define EA_RAID_ROFI_8hour		9	// 8 hours
#define EA_RAID_ROFI_12hour		10	// 12 hours
#define EA_RAID_ROFI_16hour		11	// 16 hours
#define EA_RAID_ROFI_24hour		12	// 24 hours
#define EA_RAID_ROFI_2day		13	// 2 days
#define EA_RAID_ROFI_4day		14	// 4 days
#define EA_RAID_ROFI_7day		15	// 7 days

	short16			rofReboot;		// ROF reboot count 
									// (1 - 15, 0 means disable)
    short16         pad2;			// padding
} ea_raid_ctrlParam_t;


/*
 *-----------------------------------------------------------
 * Get RAID Controller Parameters Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
} ea_raid_getCtrlParamReqMsg_t; 	// EA_RAID_MID_CTRL_GETPARAM_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
    uint32         	 	pad1;       // padding
	ea_raid_ctrlParam_t	params;
} ea_raid_getCtrlParamRspMsg_t; 	// EA_RAID_MID_CTRL_GETPARAM_RSP


/*
 ******************************************************************
 *
 * Messages for getting physical device Info.
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Physical Device Id
 *-----------------------------------------------------------
 */
typedef struct 
{
	short16			channelNumber;
	short16			targetId;
    uint32          pad1;              // padding
} ea_raid_phydev_id_t;


/*
 *-----------------------------------------------------------
 * Get Physical Device List Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
} ea_raid_phyDevListReqMsg_t; 		// EA_RAID_MID_PHYDEV_LIST_REQ

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	uint32				rCode;			// return code
	uint32				numEntries;		// Number of entries in phyDevList
	ea_raid_phydev_id_t	phyDevList[1];	// Variable len list
} ea_raid_phyDevListRspMsg_t; 			// EA_RAID_MID_PHYDEV_LIST_RSP


/*
 *-----------------------------------------------------------
 * Get Physical Device Info. Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	ea_raid_phydev_id_t	pid;
} ea_raid_phyDevInfoReqMsg_t; 		// EA_RAID_MID_PHYDEV_INFO_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
	uint32				pad1;
	uint64				wwn;
	short16				enclosure;
	short16				row;
	short16				column;
	ushort16			rdn;
	uint32 				physicalSizeInMB;
	uint32 				usableSizeInMB;
	uint32 				freeSizeInMB;
	uint32 				sectorSizeInByte;

	short16				deviceState;
#define EA_RAID_PHYDEVSTATE_UNCONFIGURED   	0x00
#define EA_RAID_PHYDEVSTATE_OPTIMAL			0x01
#define EA_RAID_PHYDEVSTATE_REBUILD			0x03
#define EA_RAID_PHYDEVSTATE_FAILED			0x08
#define EA_RAID_PHYDEVSTATE_CRITICAL		0x09
#define EA_RAID_PHYDEVSTATE_REBUILDING     	0x0b
#define EA_RAID_PHYDEVSTATE_GLOBHOTSPARE   	0x21
#define EA_RAID_PHYDEVSTATE_PERMHOTSPARE   	0x61

	short16				portInUse;
	uint32				pad2;
	ea_raid_phydev_id_t	pid;
	char				vendorName[EA_RAID_FIELD_LEN_MAX];
	char				modelName[EA_RAID_FIELD_LEN_MAX];
	char				serialNumber[EA_RAID_FIELD_LEN_MAX];
	char				microcodeVer[EA_RAID_FIELD_LEN_MAX];

} ea_raid_phyDevInfoRspMsg_t; 		// EA_RAID_MID_PHYDEV_INFO_RSP


/*
 ******************************************************************
 *
 * Messages for getting logical device Info.
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Logical Device Id
 *-----------------------------------------------------------
 */

typedef struct 
{
   short16	id;
   byte    	pad1[6];
} ea_raid_logdev_id_t;

/*
 *-----------------------------------------------------------
 * Get Logical Device List Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t					hdr;		// header
	ea_raid_ctrl_id_t			id;			// controller ID
} ea_raid_logDevListReqMsg_t; 				// EA_RAID_MID_LOGDEV_LIST_REQ

typedef struct
{
	ea_msgHdr_t					hdr;				// header
	ea_raid_ctrl_id_t			id;					// controller ID
	uint32						rCode;				// return code
	uint32						numEntries;			
	ea_raid_logdev_id_t			logDevIdList[1]; 	// Variable len list

} ea_raid_logDevListRspMsg_t; 				// EA_RAID_MID_LOGDEV_LIST_RSP


/*
 *-----------------------------------------------------------
 * Get Logical Device Info. Messages
 *-----------------------------------------------------------
 */
/*
 *-----------------------------------------------------------
 * RAID Level Codes
 *-----------------------------------------------------------
 */
#define EA_RAID_LEVEL_RAID_0							0x00
#define EA_RAID_LEVEL_RAID_1							0x01
#define EA_RAID_LEVEL_RAID_3_RIGHT_ASYMETRIC_PARITY		0x03
#define EA_RAID_LEVEL_RAID_5_RIGHT_ASYMETRIC_PARITY		0x05
#define EA_RAID_LEVEL_RAID_0_PLUS_1						0x06
#define EA_RAID_LEVEL_RAID_JBOD							0x07
#define EA_RAID_LEVEL_RAID_3_FIXED_PARITY				0x09
#define EA_RAID_LEVEL_RAID_3_LEFT_ASYMETRIC_PARITY		0x0b
#define EA_RAID_LEVEL_RAID_5_LEFT_ASYMETRIC_PARITY		0x0d

typedef struct 
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	ea_raid_logdev_id_t	logDevId;
} ea_raid_logDevInfoReqMsg_t; 		// EA_RAID_MID_LOGDEV_INFO_REQ

typedef struct
{
	ea_msgHdr_t			hdr;				// header
	ea_raid_ctrl_id_t	id;					// controller ID
	uint32				rCode;				// return code
	uint32				raidLevel;
	uint32 				physicalSizeInMB;
	uint32 				actualSizeInMB;
	uint32 				stripeSizeInKB;
	uint32				cacheLineSizeInKB;
	ea_raid_logdev_id_t	logDevId;

	short16				deviceState;
#define EA_RAID_LOGDEVSTATE_UNCONFIGURED   	0x00
#define EA_RAID_LOGDEVSTATE_OPTIMAL			0x01
#define EA_RAID_LOGDEVSTATE_FAILED			0x08
#define EA_RAID_LOGDEVSTATE_CRITICAL		0x09
#define EA_RAID_LOGDEVSTATE_REBUILDING     	0x0b
	uchar8				readCacheEnabled;	// Is read cache enabled
	uchar8				writeCacheEnabled;	// Is write cache enabled
	uchar8				hasBgTask;			// Has any background tasks
    uchar8              pad1[3];           	// padding
} ea_raid_logDevInfoRspMsg_t; 				// EA_RAID_MID_LOGDEV_INFO_RSP


/*
 *-----------------------------------------------------------
 * Logical Device Physical Drive Info. Structure
 *-----------------------------------------------------------
 */

/*
 ******************************************************************
 *
 * Messages for create/remove/initialize Logical Device (LUN)
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Logical Device Definition Structure
 *-----------------------------------------------------------
 */
typedef struct
{
	uint32				raidLevel;
	uint32 				sizeInMB;
	uint32				stripeSizeInKB;
	uchar8				cacheEnabled;		// Is write cache enabled
	uchar8				autoInit;			// Is auto init enabled
	short16				numOfPhyDevices;	// Number of entries in phyDevList
	ea_raid_phydev_id_t	phyDevList[1];		// Variable len list
} ea_raid_logDev_def_t;


/*
 *-----------------------------------------------------------
 * Create LUN Messages
 *	Create a new LUN using the given Logical Device
 *	Definition. LUN is created w/o LUN Masking.
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t				hdr;		// header
	ea_raid_ctrl_id_t		id;			// controller ID
	ea_raid_logDev_def_t	lundef;
} ea_raid_createLunReqMsg_t; 			// EA_RAID_MID_LUN_CREATE_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
	uint32				pad1;
	ea_raid_logdev_id_t	logDevId;
} ea_raid_createLunRspMsg_t;		// EA_RAID_MID_LUN_CREATE_RSP

/*
 *-----------------------------------------------------------
 * Create Simple LUN Messages
 *  Create n LUNs of the same RAID type.
 *  LUNs are created w/o LUN Masking.
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t				hdr;		// header
	ea_raid_ctrl_id_t		id;			// controller ID
	uint32          		lunCount;  	// Number of LUNs to create
	uint32          		raidLevel;
	uint32          		sizeInMB;
	uint32          		stripeSizeInKB;
} ea_raid_createSimpleLunReqMsg_t;		// EA_RAID_MID_LUN_CREATE_SIMPLE_REQ

typedef struct
{
	ea_msgHdr_t				hdr;		// header
	ea_raid_ctrl_id_t		id;			// controller ID
	uint32					rCode;		// return code
	uint32					pad1;		
} ea_raid_createSimpleLunRspMsg_t;		// EA_RAID_MID_LUN_CREATE_SIMPLE_RSP

/*
 *-----------------------------------------------------------
 * Init LUN Messages
 *-----------------------------------------------------------
 */

typedef struct 
{
	ea_msgHdr_t				hdr;		// header
	ea_raid_ctrl_id_t		id;			// controller ID
	short16         		type;       // Init Type
#define EA_RAID_INITTYPE_ONE   	0       // Init one LUN, defined in id
#define EA_RAID_INITTYPE_ALL   	1       // Init all LUNs
#define EA_RAID_INITTYPE_ALLNOT 2       // Init all uninitialized LUNs

	char					pad1[6];
	ea_raid_logdev_id_t		logDevId;
} ea_raid_initLunReqMsg_t;				// EA_RAID_MID_LUN_INIT_REQ

typedef struct 
{
	ea_msgHdr_t				hdr;		// header
	ea_raid_ctrl_id_t		id;			// controller ID
	uint32          		rCode;      // return code
	uint32          		init_started_count;
} ea_raid_initLunRspMsg_t;				// EA_RAID_MID_LUN_INIT_RSP

/*
 *-----------------------------------------------------------
 * Delete LUN Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	ea_raid_logdev_id_t	logDevId;
} ea_raid_deleteLunReqMsg_t; 		// EA_RAID_MID_LUN_DELETE_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
    uint32              pad1;       // padding
} ea_raid_deleteLunRspMsg_t; 		// EA_RAID_MID_LUN_DELETE_RSP


/*
 ******************************************************************
 *
 * Messages for add/delete LUN Masking.
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Get LUN Masking Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	ea_raid_logdev_id_t	logDevId;
	short16				ctrlNumber;		// Controller number
	short16				portNumber;
    uint32            	pad1;        	// padding
} ea_raid_getLunMaskReqMsg_t; 			// EA_RAID_MID_LUNMASK_INFO_REQ

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	uint32				rCode;			// return code
    uint32            	pad1;        	// padding
	ea_raid_logdev_id_t	logDevId;
	short16				ctrlNumber;		// Controller number
	short16				portNumber;
    short16             pad2;           // padding
	short16				numOfHostWwns;	// Number of entries in hostWwnList
	uint64				hostWwnList[1];	// Variable len list
} ea_raid_getLunMaskRspMsg_t; 			// EA_RAID_MID_LUNMASK_INFO_RSP


/*
 *-----------------------------------------------------------
 * Add LUN Masking Messages
 *	Add the given host wwns to LUN's host wwn table
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	ea_raid_logdev_id_t	logDevId;
	short16				ctrlNumber;		// Controller number
	short16				portNumber;
    short16             pad1;           // padding
	short16				numOfHostWwns;	// Number of entries in hostWwnList
	uint64				hostWwnList[1];	// Variable len list
} ea_raid_addLunMaskReqMsg_t; 			// EA_RAID_MID_LUNMASK_ADD_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
    uint32            	pad1;       // padding
	ea_raid_logdev_id_t	logDevId;
	short16				ctrlNumber;	// Controller number
	short16				portNumber;
    uint32              pad2;    	// padding
} ea_raid_addLunMaskRspMsg_t; 		// EA_RAID_MID_LUNMASK_ADD_RSP


/*
 *-----------------------------------------------------------
 * Delete LUN Masking Messages
 *	Delete the given host wwns from LUN's host wwn table
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	ea_raid_logdev_id_t	logDevId;
	short16				ctrlNumber;		// Controller number
	short16				portNumber;
    short16             pad1;           // padding
	short16				numOfHostWwns;	// Number of entries in hostWwnList
	uint64				hostWwnList[1];	// Variable len list
} ea_raid_delLunMaskReqMsg_t; 			// EA_RAID_MID_LUNMASK_DEL_REQ

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	uint32				rCode;			// return code
    uint32            	pad1;        	// padding
	ea_raid_logdev_id_t	logDevId;
	short16				ctrlNumber;		// Controller number
	short16				portNumber;
    uint32              pad2;    		// padding
} ea_raid_delLunMaskRspMsg_t; 			// EA_RAID_MID_LUNMASK_DEL_RSP

/*
 ******************************************************************
 *
 * Misc. Messages.
 *
 ******************************************************************
 */

/*
 *-----------------------------------------------------------
 * Background Process Structure
 *-----------------------------------------------------------
 */
typedef struct
{
	ea_raid_logdev_id_t	logDevId;
	short16				checkProgress;		// percentage or -1 if not active
	short16				initProgress;		// percentage or -1 if not active
	short16				rebuildProgress;	// percentage or -1 if not active
	short16				expandProgress;		// percentage or -1 if not active
} ea_raid_backProc_t;

/*
 *-----------------------------------------------------------
 * Display Background Process Messages
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
} ea_raid_bpListReqMsg_t;				// EA_RAID_MID_LUN_BPLIST_REQ

typedef struct
{
	ea_msgHdr_t			hdr;			// header
	ea_raid_ctrl_id_t	id;				// controller ID
	uint32				rCode;			// return code 
	short16				numEntries;		// Number of entries in bpList
	short16				pad1;
	ea_raid_backProc_t	bpList[1];		// Variable len list == numEntries
} ea_raid_bpListRspMsg_t;				// EA_RAID_MID_LUN_BPLIST_RSP

/*
 *-----------------------------------------------------------
 * Cancel Background Process Messages
 *	Cancel all background tasks of the given process type
 *	on ALL LUNs.
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID

	short16				type;		// Process Type
#define EA_RAID_BPTYPE_CHECK	0	// Consistency check
#define EA_RAID_BPTYUE_CHECHRES	1	// Check and restore
#define EA_RAID_BPTYPE_INIT		2	// INIT
#define EA_RAID_BPTYPE_BGINIT	3	// BGINIT - background init
#define EA_RAID_BPTYPE_REBUILD	4	// REBUILD
#define EA_RAID_BPTYPE_ALL		5	// ALL - all background process

    short16             pad1;    	// padding
    uint32              pad2;    	// padding
} ea_raid_cancelBpReqMsg_t; 		// EA_RAID_MID_LUN_CANCELBP_REQ

typedef struct
{
	ea_msgHdr_t			hdr;		// header
	ea_raid_ctrl_id_t	id;			// controller ID
	uint32				rCode;		// return code
    uint32              pad1;    	// padding
} ea_raid_cancelBpRspMsg_t; 		// EA_RAID_MID_LUN_CANCELBP_RSP

/*
 *-----------------------------------------------------------
 * SDM Lun Rediscover Req/Rsp
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t			hdr;
	uint64				wwn; 	// lun wwn
} ea_lunRediscoverReq_t; 		// EA_LUN_REDISCOVER_REQ
				    
typedef struct
{
	ea_msgHdr_t    		hdr;
	uint32              result;
	uint32              pad1;
	char                output[EA_OUTPUT_LEN];
} ea_lunRediscoverRsp_t;		// EA_LUN_REDISCOVER_RSP
													  
/*
 *-----------------------------------------------------------
 * SDM Get Port Topology Req/Rsp
 *-----------------------------------------------------------
 */

typedef struct
{
	ea_msgHdr_t		hdr;
	uint32			slotId;
	uint32			portId;
} ea_getPortTopologyReq_t; 		// EA_GET_PORT_TOPOLOGY_REQ

typedef struct
{
	uint64    		portWwn;
	uint32    		portId;
	uint32    		portType;
	uint64   		numAttachedPorts;
	uint64    		attachedPortWwnList[1];
} ea_iePort_t;

typedef struct
{
	uint64      	nodeWwn;
	char        	symbolicName[SMB_NAME_LEN]; 
	uint32       	nodeType; 
	uint32       	numPorts;
	ea_iePort_t    	portList[1]; 
} ea_ieNode_t;

typedef struct
{
	uint32     		slotId;
	uint32     		portId;
	uint64     		portWwn;
	uint64     		numNodes;
	ea_ieNode_t    	ieNodeList[1];
} ea_portTopologySwitch_t;

typedef struct
{
	ea_msgHdr_t    				hdr;
	uint32              		result;
	uint32              		pad1;
	char                		output[EA_OUTPUT_LEN];
	ea_portTopologySwitch_t  	portTopologySwitch;  // through brocade switches
} ea_getPortTopologyRsp_t; 		// EA_GET_PORT_TOPOLOGY_RSP

/*------------------------------------------------------------------------------
 * NOTIFICATIONS
 *----------------------------------------------------------------------------*/

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_msgType_t    msgType;                // message that led to the action
	uint32			pad1;
    ea_fsysCfg_t    fsysCfg;                // file sys configuration
} ea_fsysNtfy_t;        // EA_FSYS_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;                    // header
    ea_msgType_t    msgType;                // message that led to the action
    ea_fsetCfg_t    fsetCfg;                // file set configuration
} ea_fsetNtfy_t;        // EA_FSET_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_VolConfig_t volCfg;
} ea_crtVolNtfy_t;      // EA_CRT_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_volId_t     volId;
} ea_delVolNtfy_t;      // EA_DEL_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_VolConfig_t volCfg;
} ea_modVolNtfy_t;      // EA_MOD_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_volId_t     volId;
    boolean         volOnline;
} ea_ooVolNtfy_t;       // EA_OO_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_VolConfig_t volCfg;
    ea_lunList_t    lunList;
} ea_grwVolNtfy_t;      // EA_GRW_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_volId_t     volId;
} ea_fullVolNtfy_t;     // EA_FULL_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_volId_t     volId;
} ea_takeoverVolNtfy_t; // EA_TAKEOVER_VOL_NTFY

typedef struct
{
    ea_msgHdr_t     hdr;
    evm_volId_t     volId;
    ea_lunList_t    lunList;
} ea_failLunNtfy_t;     // EA_FAIL_LUN_NTFY


#endif // _EA_API_H_

