/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea-common.c
 *
 * Description  : This module contains all the common functions used by EA.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include "ea.h"
#include "exports/ncm-api.h"

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */
const char toAppName[APPNAME_TOTAL][8] =
{
	"ncm",
	"evm",
	"efs",
	"cm",
	"mountd",
};

const char ncmAppName[8] = "ncm";

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_sendMsg()
 *
 * Description  :
 * Send a message to a given application on the SSC.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_sendMsg(void *           msg,
           size_t           len,
           char *           dest)

{
    int32  rc = NFX_OK;
	ea_msgHdr_t *eaMsg = (ea_msgHdr_t *) msg;
        
    EA_FCT_PRINT(("> ea_sendMsg(%d) len %d\n", eaMsg->msgType, eaMsg->msgLen));

	/* 
	 * set the message length
	 */
	eaMsg->msgLen = len;

    /*
     * Send the buffer to the destination located on the SSC.
     */
    rc = sendAgileMsg(-1,
                      msg, len,
                      dest, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);

    EA_FCT_PRINT(("< ea_sendMsg(%d) len %d\n", eaMsg->msgType, eaMsg->msgLen));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_sendEvNtfy()
 *
 * Description  :
 * Send a notification message to EV.
 * No response is expected.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_sendEvNtfy(void *            msg,
              size_t            len)

{
    int32  rc = NFX_OK;
        
    EA_FCT_PRINT(("> ea_sendEvNtfy()\n"));

    /*
     * Send the buffer to the destination located on the SSC.
     */
    rc = sendAgileMsg(-1,
                      msg, len,
                      "ev", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);

    EA_FCT_PRINT(("< ea_sendEvNtfy()\n"));
    return;
}

#ifdef ISABELLE // code to delete
/*------------------------------------------------------------------------------
 * Name         : ea_sendEvRsp()
 *
 * Description  :
 * Send a response to EV.
 * No response is expected.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_sendEvRsp(uint32 result,
             char * output,
             uint32 userInfo)

{
    ea_Rsp_t  evRsp;
        
    EA_FCT_PRINT(("> ea_sendEvRsp()\n"));

    evRsp.hdr.msgType = EA_RSP;
    evRsp.hdr.userInfo = userInfo;
    evRsp.result = result;
    strcpy(evRsp.output, output);

    ea_sendEvNtfy((void *) &evRsp,
                  sizeof(ea_Rsp_t));

    EA_FCT_PRINT(("< ea_sendEvNtfy()\n"));
    return;
}
#endif // ISABELLE


/*------------------------------------------------------------------------------
 * Name         : isLocalNodeName()
 *
 * Description  : Check whether this is a valid local node name 
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 08-OCT-2002
 *
 * Parameters   :
 *   node - IN, node name to be checked
 *              NULL, null string, "local", and hostname are valid
 *              local node name for EA.
 *
 *------------------------------------------------------------------------------
 */

uint32 isLocalNodeName(char *nodeName)
{
 	char myhostname[256];

	if ((nodeName == NULL) ||
		(*nodeName == '\0') ||
		!strcmp(nodeName, "local"))
	{
		return 1;
	}

	gethostname(myhostname, sizeof(myhostname));
	if (!strcmp(nodeName, myhostname))
		return 1;

	return 0;
}

#ifdef REMOVED_BY_SHYHPEI
/*------------------------------------------------------------------------------
 * Name         : ea_prepClusterReq()
 *
 * Description  : Prepare a request based on whether this is a local
 *                request or cluster node request
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 08-OCT-2002
 *
 * Parameters   :
 *   realReq - OUT, point to starting of buffer for real request (not ncm req).
 *   reqLen  - IN/OUT, in with real request size, out with ncm header size
 *                        added in case of remote node
 *   reqbuf  - IN, buffer space for reqeust
 *   node    - IN, remote node name, NULL or null string for local node
 *   toApp   - IN/OUT, real destination IN, and ncm OUT if for remote, otherwise
 *                     it'll remain the same
 *   slot    - IN/OUT, real destination slot IN, local ssc slot OUT if for
 *                     remote node. Otherwise, remain the same
 *   cpu     - IN/OUT, real destination slot IN, local ssc cpu OUT if
 *                     for remote node. Otherwise, remain the same
 *
 *------------------------------------------------------------------------------
 */

void ea_prepClusterReq(void **realReq, uint32 *reqLen,
					char *reqbuf,
					char *node,
					char **toApp,
					uint32 *slot, uint32 *cpu)
{
	if (node && *node)
	{
		/* send to application on remote node */
    	ncmMsg_t *ncmReq = (ncmMsg_t *) reqbuf;

		*reqLen += NCM_MSGHDR_LEN;
		if (*slot == SSC_SLOT)
			ncmReq->msgType = NCM_SEND;
		else
			ncmReq->msgType = NCM_SEND_EMB;
		strcpy(ncmReq->appName, *toApp);
		strcpy(ncmReq->nodeName, node);
		ncmReq->rqResp = 1;
		ncmReq->slot = *slot;
		ncmReq->cpu = *cpu;

		*realReq = (void *) ncmReq->appMsg;
		*toApp = (char *) toAppName[APPNAME_NCM];

		*slot = SSC_SLOT;
		*cpu = NFX_SSC_UNIX;
	}
	else
	{
		/* send to application on local node */
		/* no change on *slot, *cpu, and *toApp */

		*realReq = (void *) reqbuf;
	}

	return;
}


/*------------------------------------------------------------------------------
 * Name         : ea_getClusterRsp()
 *
 * Description  : Check NCM status and retrieve real response.
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 08-OCT-2002
 *
 * Parameters   :
 *   node    - IN, cluster node name if for remote, NULL or null string for local
 *   rspSize - IN/OUT, original response size IN. If this is from ncm, ncm
 *                     header size will be substracted from this OUT, otherwise
 *                     the original size would be OUT (local response).
 *   rspBuf  - IN, the pointer to the receiving buffer
 *   realRsp - OUT, it will point to the real response from application.
 *
 *------------------------------------------------------------------------------
 */
int32 ea_getClusterRsp(char *node, uint32 *rspSize, char *rspBuf, void **realRsp)
{
	if (node && *node)
	{
		ncmMsg_t *ncmRsp = (ncmMsg_t *) rspBuf;

		if (ncmRsp->msgType == NCM_REJECT)
			return NFX_ERR;

		*rspSize -= NCM_MSGHDR_LEN;
		*realRsp = (void *) (ncmRsp->appMsg);
	}
	else
		*realRsp = (void *) rspBuf;

	return NFX_OK;
}
#endif

/*------------------------------------------------------------------------------
 * Name         : ea_sendClusterAgileMsg()
 *
 * Description  : send a message - to a cluster node if "node" is not NULL or
 *                is not a null string, to local node otherwise. The message
 *                is always ncmMsg_t formatted, even if it's for local.
 *                ea_sendClusterAgileMsg() will choose correct starting
 *                buffer location based on "node". The caller should treat
 *                the response the same way (point to correct buffer location
 *                to get the real response from application).
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 10-OCT-2002
 *
 * Parameters   : Pretty much the same as sendAgileMsg except:
 *                req    - includes ncm msg header
 *                reqLen - the real request length, not including ncm msg header
 *                rsp    - includes ncm msg header
 *                rspLen - the real response buffer length, not including ncm
 *                         msg header
 *
 *------------------------------------------------------------------------------
 */
int ea_sendClusterAgileMsg(int sock,
	ea_clusmsg_t *req, int reqLen,
	char *toApp, uint32 cpu, uint32 slot, uint32 plane,
	struct timeval *timeout,
	ea_clusmsg_t *rsp, int rspLen,
	char *node)
{
char *sendToApp = NULL, *reqPtr = NULL, *rspPtr = NULL;
int retSize;

	if (node && *node)
	{
		/* send to application on remote node */
    	ncmMsg_t *ncmReq = (ncmMsg_t *) req->ncm;

		if (slot == SSC_SLOT)
			ncmReq->msgType = NCM_SEND;
		else
			ncmReq->msgType = NCM_SEND_EMB;
		strcpy(ncmReq->appName, toApp);
		strcpy(ncmReq->nodeName, node);
		if (!rsp || !rspLen)
			ncmReq->rqResp = 0;
		else
			ncmReq->rqResp = 1;
		ncmReq->slot = slot;
		ncmReq->cpu = cpu;

		// redirect to local NCM for forwarding
		slot = SSC_SLOT;
		cpu = NFX_SSC_UNIX;

		reqLen += NCM_MSGHDR_LEN;
		rspLen += NCM_MSGHDR_LEN;
		reqPtr = req->ncm;
		rspPtr = rsp->ncm;
		sendToApp = (char *) ncmAppName;
	}
	else
	{
		reqPtr = req->buf;
		rspPtr = rsp->buf;
		sendToApp = toApp;
	}

	memset(rspPtr, 0, rspLen);

	retSize = sendAgileMsg(sock, reqPtr, reqLen, 
								sendToApp, cpu, slot, plane, timeout,
								rspPtr, rspLen);
	if (retSize <= 0)
		return retSize;

	if (node && *node)
	{
		ncmMsg_t *ncmRsp = (ncmMsg_t *) rsp->ncm;

		if ((retSize < NCM_MSGHDR_LEN) ||
			(ncmRsp->msgType == NCM_REJECT))
			return 0;

		retSize -= NCM_MSGHDR_LEN;
	}

	return retSize;
}


/*------------------------------------------------------------------------------
 * Name         : checkClusterAgileDest()
 *
 * Description  : Verify a slot/cpu is available on a specific cluster node.
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 08-OCT-2002
 *
 *------------------------------------------------------------------------------
 */

int checkClusterAgileDest(char *node, uint32 cpu, uint32 slot)
{
	uint32					reqSize;
    ea_msg_t                reqBuf;
    ea_msg_t                rspBuf;
    ncmMsg_t                *ncmReq;
	chassisdMsgInfoReq_t	*cmReq;
	chassisdMsgInfoResp_t	*cmRsp;
    int32               	rc = NFX_OK;
	int32					status = NFX_OK;

	struct timeval timeOut = {10, 0};

	if (!node || !(*node))
		return checkAgileDest(cpu, slot);

    /* 
	 * Initialize the structures.
     */
	bzero((void *)&reqBuf, sizeof(reqBuf));
	bzero((void *)&rspBuf, sizeof(rspBuf));

	reqSize = NCM_MSGHDR_LEN + sizeof(chassisdMsgInfoReq_t);
	ncmReq = (ncmMsg_t *) reqBuf.buf;
	ncmReq->msgType = NCM_SEND;
	strcpy(ncmReq->appName, toAppName[APPNAME_CM]);
	strcpy(ncmReq->nodeName, node);
	ncmReq->rqResp = 1;

	cmReq = (chassisdMsgInfoReq_t *) (ncmReq->appMsg);

	/*
	 * send the request to Chassisd
	 */
	cmReq->hdr.mid = CHASSISD_GET_CHASSIS_INFO_REQ;

	rc = sendAgileMsg(-1, &cmReq, sizeof(chassisdMsgInfoReq_t),
	                      (char *) toAppName[APPNAME_NCM],
						  0, 0, 0,
						  &timeOut,
						  &cmRsp, sizeof(chassisdMsgInfoResp_t));

	if (rc != (NCM_MSGHDR_LEN + sizeof(chassisdMsgInfoResp_t)))
		return NFX_ERR;

	cmRsp = (chassisdMsgInfoResp_t *) (((ncmMsg_t *) rspBuf.buf)->appMsg);
	if (cmRsp->present[slot] == 0)
		status = NFX_ERR;
	else if (cmRsp->cpu_state[slot][cpu] != CM_SLICE_STATE_RUNTIME)
		status = NFX_ERR;
	else
		status = NFX_OK;

	return status;
}


/*----------------------------------------------------------------------------*/
