/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea-processCliMsg.c
 *
 * Description  : This module deals with the storage grades.
 * It processes messages received from CLI.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

#include "exports/evm-api.h"

#include "ea.h"
#include "ea-api.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

typedef struct ea_vendor_model_T ea_vendor_model_t;
struct ea_vendor_model_T
{
    evm_vendor_model_t  info;
    ea_vendor_model_t * next;
};

typedef struct
{
    uint16              idx;
    evm_raid_t          raidLevel;
    uint32              minLunSize; // megabytes
    uint32              maxLunSize; // megabytes
    ea_vendor_model_t * make;
} ea_sgrade_t;

static ea_sgrade_t * sgradeTab[MAX_SGRADE];

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static int
ea_sg_getFreeIdx(void);

static void
ea_processCrtSgradeReq(ea_crtSgradeReq_t * req,
                       ea_senderInfo_t *   sender);

static void
ea_processAddSgradeReq(ea_addSgradeReq_t * req,
                       ea_senderInfo_t *   sender);

static void
ea_processRmvSgradeReq(ea_rmvSgradeReq_t * req,
                       ea_senderInfo_t *   sender);

static void
ea_processDstrSgradeReq(ea_dstrSgradeReq_t * req,
                        ea_senderInfo_t *   sender);

static void
ea_processGetSgradeReq(ea_getSgradeReq_t * req,
                       ea_senderInfo_t *   sender);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_sg_getFreeIdx()
 *
 * Description  : Get free index in the storage grade table..
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static int
ea_sg_getFreeIdx(void)

{
    int i;
    EA_FCT_PRINT(("> ea_sg_getFreeIdx()\n"));

    for (i = 0; i < MAX_SGRADE; i++)
    {
        if (sgradeTab[i] == NULL)
        {
            break;
        }
    } // for
    
    EA_FCT_PRINT(("< ea_sg_getFreeIdx()\n"));
    return(i);
}
       
/*------------------------------------------------------------------------------
 * Name         : ea_processCrtSgradeReq()
 *
 * Description  : Create a storage grade.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processCrtSgradeReq(ea_crtSgradeReq_t * req,
                       ea_senderInfo_t *   sender)

{
    int                 idx;
    int                 rc;
    ea_crtSgradeRsp_t   rsp;
    
    EA_FCT_PRINT(("> ea_processCrtSgradeReq()\n"));
    NFX_ASSERT(req->sgrade.num == 1);
    
    /* Initialize the response.
     */
    bzero((void *) &rsp, sizeof(ea_crtSgradeRsp_t));
    rsp.hdr.msgType = EA_CRT_SGRADE_RSP;
    rsp.hdr.userInfo = req->hdr.userInfo;

    /* Get a new entry in the table.
     */
    idx = ea_sg_getFreeIdx();

    if (idx < MAX_SGRADE)
    {
        NFX_ASSERT(sgradeTab[idx] == NULL);

        /* Add new storage grade into table.
         */
        sgradeTab[idx] = (ea_sgrade_t *) malloc(sizeof(ea_sgrade_t));
        sgradeTab[idx]->make = (ea_vendor_model_t *)
            malloc(sizeof(ea_vendor_model_t));
        
        sgradeTab[idx]->idx = idx;
        sgradeTab[idx]->raidLevel = req->sgrade.raidLevel;
        sgradeTab[idx]->minLunSize = req->sgrade.minLunSize;
        sgradeTab[idx]->maxLunSize = req->sgrade.maxLunSize;
        strcpy(sgradeTab[idx]->make->info.vendor,
               req->sgrade.vendorInfo[0].vendor);
        strcpy(sgradeTab[idx]->make->info.model,
               req->sgrade.vendorInfo[0].model);
        sgradeTab[idx]->make->next = NULL;
        
        rsp.result = NFX_OK;
        strcpy(rsp.output, "Done.");
    }
    else
    {
        rsp.result = NFX_ERR;
        strcpy(rsp.output, "Error: table of storage grade is full!");
    }

    /* Send response back.
     */
    rc = sendAgileMsg(-1,
                      &rsp, sizeof(rsp),
                      sender->name,
                      sender->cpu,
                      sender->slot,
                      sender->plane,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);
    
    EA_DBG_PRINT(("%s (%d)\n", rsp.output, rsp.result));
    EA_FCT_PRINT(("< ea_processCrtSgradeReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processAddSgradeReq()
 *
 * Description  : Add a vendor/make to an existing storage grade.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processAddSgradeReq(ea_addSgradeReq_t * req,
                       ea_senderInfo_t *   sender)

{
    int                 idx;
    int                 rc;
    ea_addSgradeRsp_t   rsp;
    
    EA_FCT_PRINT(("> ea_processAddSgradeReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &rsp, sizeof(ea_addSgradeRsp_t));
    rsp.hdr.msgType = EA_ADD_SGRADE_RSP;
    rsp.hdr.userInfo = req->hdr.userInfo;

    /* Check if index is valid.
     * Get the existing entry.
     */
    idx = req->sgradeId - 1;
    if ((0 <= idx) && (idx <= (MAX_SGRADE - 1)))
    {
        if (sgradeTab[idx] != NULL)
        {
            /* Add vendor/model to the existing entry.
             */
            ea_vendor_model_t * pNew = NULL;

            pNew = (ea_vendor_model_t *) malloc(sizeof(ea_vendor_model_t));
            strcpy(pNew->info.vendor, req->vendorInfo.vendor);
            strcpy(pNew->info.model, req->vendorInfo.model);
            pNew->next = sgradeTab[idx]->make->next;
            sgradeTab[idx]->make->next = pNew;
            
            rsp.result = NFX_OK;
            strcpy(rsp.output, "Done.");
        }
        else
        {
            rsp.result = NFX_ERR;
            strcpy(rsp.output, "Error: Entry not found!");
        }
    }
    else
    {
        rsp.result = NFX_ERR;
        strcpy(rsp.output, "Error: Entry not found!");
    }

    /* Send response back.
     */
    rc = sendAgileMsg(-1,
                      &rsp, sizeof(rsp),
                      sender->name,
                      sender->cpu,
                      sender->slot,
                      sender->plane,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);
    
    EA_DBG_PRINT(("%s (%d)\n", rsp.output, rsp.result));
    EA_FCT_PRINT(("< ea_processAddSgradeReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRmvSgradeReq()
 *
 * Description  : Remove a vendor/make from an existing storage grade.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processRmvSgradeReq(ea_rmvSgradeReq_t * req,
                       ea_senderInfo_t *   sender)

{
    int                 idx;
    int                 rc;
    ea_rmvSgradeRsp_t   rsp;
    
    EA_FCT_PRINT(("> ea_processRmvSgradeReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &rsp, sizeof(ea_rmvSgradeRsp_t));
    rsp.hdr.msgType = EA_RMV_SGRADE_RSP;
    rsp.hdr.userInfo = req->hdr.userInfo;

    /* Check if index is valid.
     * Get the existing entry.
     */
    idx = req->sgradeId - 1;
    if ((0 <= idx) && (idx <= (MAX_SGRADE - 1)))
    {
        if (sgradeTab[idx] != NULL)
        {
            /* Remove vendor/model to the existing entry.
             */
            ea_vendor_model_t * pCurr = sgradeTab[idx]->make;
            ea_vendor_model_t * pPrev = pCurr;
            boolean             found = FALSE;
            
            while (pCurr != NULL)
            {
                if (!strcmp(req->vendorInfo.vendor, pCurr->info.vendor) &&
                    !strcmp(req->vendorInfo.model, pCurr->info.model))
                {
                    found = TRUE;
                    break;
                }
                pPrev = pCurr;
                pCurr = pCurr->next;
            } // while
            
            if (found == TRUE)
            {
                if (pCurr == sgradeTab[idx]->make)
                {
                    sgradeTab[idx]->make = pCurr->next;
                    
                    /* Delete the entry if no more vendor/model.
                     */
                    if (sgradeTab[idx]->make == NULL)
                    {
                        free(sgradeTab[idx]);
                        sgradeTab[idx] = NULL;
                    }
                }
                else
                {
                    pPrev->next = pCurr->next;
                }
                free(pCurr);

                rsp.result = NFX_OK;
                strcpy(rsp.output, "Done.");
            }
            else
            {
                rsp.result = NFX_ERR;
                strcpy(rsp.output, "Error: Entry not found!");
            }
        }
        else
        {
            rsp.result = NFX_ERR;
            strcpy(rsp.output, "Error: Entry not found!");
        }
    }
    else
    {
        rsp.result = NFX_ERR;
        strcpy(rsp.output, "Error: Entry not found!");
    }

    /* Send response back.
     */
    rc = sendAgileMsg(-1,
                      &rsp, sizeof(rsp),
                      sender->name,
                      sender->cpu,
                      sender->slot,
                      sender->plane,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);
    
    EA_DBG_PRINT(("%s (%d)\n", rsp.output, rsp.result));
    EA_FCT_PRINT(("< ea_processRmvSgradeReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDstrSgradeReq()
 *
 * Description  : Destroy a storage grade.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processDstrSgradeReq(ea_dstrSgradeReq_t * req,
                       ea_senderInfo_t *   sender)

{
    int                 idx = req->sgradeId - 1;
    int                 rc;
    ea_dstrSgradeRsp_t  rsp;

    EA_FCT_PRINT(("> ea_processDstrSgradeReq()\n"));

    bzero((void *) &rsp, sizeof(ea_dstrSgradeRsp_t));
    rsp.hdr.msgType = EA_DSTR_SGRADE_RSP;
    rsp.hdr.userInfo = req->hdr.userInfo;
    
    if (idx < MAX_SGRADE)
    {
        if (sgradeTab[idx] != NULL)
        {
            ea_vendor_model_t * pMake = sgradeTab[idx]->make;
            
            while(pMake != NULL)
            {
                ea_vendor_model_t * pDel = pMake;
                
                pMake = pMake->next;
                free(pDel);
            }
            sgradeTab[idx]->make = NULL;
            
            free(sgradeTab[idx]);
            sgradeTab[idx] = NULL;
            
            rsp.result = NFX_OK;
            strcpy(rsp.output, "Done.");
        }
        else
        {
            rsp.result = NFX_ERR;
            strcpy(rsp.output, "Error: entry not found.");
        }
    }
    else
    {
        rsp.result = NFX_ERR;
        sprintf(rsp.output,
                "Error: invalid index in storage grade table (%d + 1).", idx);
    }
    
    /* Send response back.
     */
    rc = sendAgileMsg(-1,
                      &rsp, sizeof(rsp),
                      sender->name,
                      sender->cpu,
                      sender->slot,
                      sender->plane,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);

    EA_DBG_PRINT(("%s (%d)\n", rsp.output, rsp.result));
    EA_FCT_PRINT(("< ea_processDstrSgradeReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetSgradeReq()
 *
 * Description  : Get the existing storage grades.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processGetSgradeReq(ea_getSgradeReq_t * req,
                       ea_senderInfo_t *   sender)

{
    ea_getSgradeRsp_t * rsp;
    int rc;
    int i;
    int nbSgrade = 0;
    int nbVendor = 0;
    int len;
    
    EA_FCT_PRINT(("> ea_processGetSgradeReq()\n"));

    /* Calculate how storage grades and vendors there are in the list.
     */
    EA_DBG_PRINT(("Storage Grade Table\n"));
    EA_DBG_PRINT(("%35s %19s\n", "RAID", "LUN size"));
    EA_DBG_PRINT((" %2s %10s %10s %10s %10s %10s\n",
                  "Id", "Vendor", "Model", "level", "Min", "Max"));
    EA_DBG_PRINT((" %2s %10s %10s %10s %10s %10s\n",
                  "--", "------", "-----", "-----", "---", "---"));
    for (i = 0; i < MAX_SGRADE; i++)
    {
        if (sgradeTab[i] != NULL)
        {
            ea_vendor_model_t * pMake = NULL;
            
            NFX_ASSERT(sgradeTab[i]->make != NULL);
            pMake = sgradeTab[i]->make->next;
            
            EA_DBG_PRINT((" %2d %10s %10s %10d %10d %10d\n",
                          (i + 1),
                          sgradeTab[i]->make->info.vendor,
                          sgradeTab[i]->make->info.model,
                          sgradeTab[i]->raidLevel,
                          sgradeTab[i]->minLunSize,
                          sgradeTab[i]->maxLunSize));
            nbVendor ++;
            
            while (pMake != NULL)
            {
                EA_DBG_PRINT(("%14s %10s\n",
                              pMake->info.vendor,
                              pMake->info.model));
                pMake = pMake->next;
                nbVendor ++;
            } // while
            nbSgrade ++;
        }
    } // for
    EA_DBG_PRINT(("\n"));
    EA_DBG_PRINT(("nbSgrade = %d\n", nbSgrade));
    EA_DBG_PRINT(("nbVendor = %d\n", nbVendor));

    /* Copy the table into the response.
     */
    len = sizeof(ea_getSgradeRsp_t) + 
        ((nbSgrade+nbVendor-1) * sizeof(ea_strg_attributes_t));
    rsp = (ea_getSgradeRsp_t *) malloc(len);
    bzero((void *) rsp, len);
    rsp->hdr.msgType = EA_GET_SGRADE_RSP;
    rsp->hdr.userInfo = req->hdr.userInfo;
    rsp->nbSgrade = 0;

    for (i = 0; i < MAX_SGRADE; i++)
    {
        if (sgradeTab[i] != NULL)
        {
            ea_vendor_model_t * pMake = NULL;

            NFX_ASSERT(sgradeTab[i]->make != NULL);
            pMake = sgradeTab[i]->make->next;
            
            rsp->sgrade[rsp->nbSgrade].idx = i + 1;
            rsp->sgrade[rsp->nbSgrade].raidLevel = sgradeTab[i]->raidLevel;
            rsp->sgrade[rsp->nbSgrade].minLunSize = sgradeTab[i]->minLunSize;
            rsp->sgrade[rsp->nbSgrade].maxLunSize = sgradeTab[i]->maxLunSize;
            strcpy(rsp->sgrade[rsp->nbSgrade].vendorInfo.vendor,
                   sgradeTab[i]->make->info.vendor);
            strcpy(rsp->sgrade[rsp->nbSgrade].vendorInfo.model,
                   sgradeTab[i]->make->info.model);
            rsp->nbSgrade ++;
            
            while (pMake != NULL)
            {
                rsp->sgrade[rsp->nbSgrade].idx = 0;
                rsp->sgrade[rsp->nbSgrade].raidLevel = 0;
                rsp->sgrade[rsp->nbSgrade].minLunSize = 0;
                rsp->sgrade[rsp->nbSgrade].maxLunSize = 0;
                strcpy(rsp->sgrade[rsp->nbSgrade].vendorInfo.vendor,
                       pMake->info.vendor);
                strcpy(rsp->sgrade[rsp->nbSgrade].vendorInfo.model,
                       pMake->info.model);
                rsp->nbSgrade ++;
                pMake = pMake->next;
            } // while
        }
    } // for

#ifdef ISABELLE // test only
    {
        ea_getSgradeRsp_t * eaRsp = rsp;
        
    printf("%35s %19s\n", "RAID", "LUN size");
    printf(" %2s %10s %10s %10s %10s %10s\n",
           "Id", "Vendor", "Model", "level", "Min", "Max");
    printf(" %2s %10s %10s %10s %10s %10s\n",
           "--", "------", "-----", "-----", "---", "---");
    for (i = 0; i < eaRsp->nbSgrade; i++)
    {
        if (eaRsp->sgrade[i].idx != 0)
        {
            printf(" %2d %10s %10s %10d %10d %10d\n",
                   eaRsp->sgrade[i].idx,
                   eaRsp->sgrade[i].vendorInfo.vendor,
                   eaRsp->sgrade[i].vendorInfo.model,
                   eaRsp->sgrade[i].raidLevel,
                   eaRsp->sgrade[i].minLunSize,
                   eaRsp->sgrade[i].maxLunSize);
        }
        else
        {
            printf("%14s %10s\n",
                   eaRsp->sgrade[i].vendorInfo.vendor,
                   eaRsp->sgrade[i].vendorInfo.model);
        }
    } // for
    printf("\n");
    }
#endif // ISABELLE
    
    /* Send response back.
     */
    rc = sendAgileMsg(-1,
                      (void *) rsp, len,
                      sender->name,
                      sender->cpu,
                      sender->slot,
                      sender->plane,
                      NULL, NULL, 0);
    EA_ASSERT(rc == NFX_OK);

    free(rsp);
    
    EA_FCT_PRINT(("< ea_processGetSgradeReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_sg_init()
 *
 * Description  : Initialize the storage grade's table.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_sg_init(void)

{
    int i;
    
    EA_FCT_PRINT(("> ea_sg_init()\n"));

    /* Initialize the table of storage grade.
     * Will be done later reading the config file - TODO!!!
     */
    for (i = 0; i < MAX_SGRADE; i++)
    {
        sgradeTab[i] = NULL;
    } // for
    
    EA_FCT_PRINT(("< ea_sg_init()\n"));
    return;
}
       
/*------------------------------------------------------------------------------
 * Name         : ea_sg_cleanup()
 *
 * Description  : Cleanup the storage grade's table.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_sg_cleanup(void)

{
    int i;
    
    EA_FCT_PRINT(("> ea_sg_cleanup()\n"));

    /* Cleanup the table of storage grade.
     * Wont' be necessary - will be saved in config file - TODO!!!
     */
    for (i = 0; i < MAX_SGRADE; i++)
    {
        if (sgradeTab[i] != NULL)
        {
            ea_vendor_model_t * pMake = sgradeTab[i]->make;
            
            while(pMake != NULL)
            {
                ea_vendor_model_t * pDel = pMake;
                
                pMake = pMake->next;
                free(pDel);
            }
            sgradeTab[i]->make = NULL;
            
            free(sgradeTab[i]);
            sgradeTab[i] = NULL;
        }
    } // for
    
    EA_FCT_PRINT(("< ea_sg_cleanup()\n"));
    return;
}
       
/*------------------------------------------------------------------------------
 * Name         : ea_processCliMsg()
 *
 * Description  : Process all messages received from the CLIs.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_processCliMsg(ea_msg_t *          msg,
                 ea_senderInfo_t *   sender)
    
{
    EA_FCT_PRINT(("> ea_processCliMsg()\n"));
                       
    if (msg->len < sizeof(ea_msgHdr_t))
    {
        printf("Got truncated message of len %d", msg->len);
    }
    else
    {
        ea_msgHdr_t *   hdr;
        
        hdr = (ea_msgHdr_t *) msg->buf;
        EA_DBG_PRINT(("msg type is %d\n", hdr->msgType));
        /*
         * Process the message according to its type.
         */
        switch (hdr->msgType)
        {
            
        case EA_CRT_SGRADE_REQ:
        {
            ea_crtSgradeReq_t * req = (ea_crtSgradeReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_CRT_SGRADE_REQ\n"));
            ea_processCrtSgradeReq(req, sender);
            break;
        }

        case EA_ADD_SGRADE_REQ:
        {
            ea_addSgradeReq_t * req = (ea_addSgradeReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_ADD_SGRADE_REQ\n"));
            ea_processAddSgradeReq(req, sender);
            break;
        }

        case EA_RMV_SGRADE_REQ:
        {
            ea_rmvSgradeReq_t * req = (ea_rmvSgradeReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_RMV_SGRADE_REQ\n"));
            ea_processRmvSgradeReq(req, sender);
            break;
        }

        case EA_DSTR_SGRADE_REQ:
        {
            ea_dstrSgradeReq_t * req = (ea_dstrSgradeReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_DSTR_SGRADE_REQ\n"));
            ea_processDstrSgradeReq(req, sender);
            break;
        }
                
        case EA_GET_SGRADE_REQ:
        {
            ea_getSgradeReq_t * req = (ea_getSgradeReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_SGRADE_REQ\n"));
            ea_processGetSgradeReq(req, sender);
            break;
        }

        default:
        {
            printf("EA-CLI: Unknown msg type (%d)\n", hdr->msgType);
            break;
        }
        } /* switch */
    }
    
    EA_FCT_PRINT(("< ea_processCliMsg()\n"));
    return;
}

/*----------------------------------------------------------------------------*/
