/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea-processEvmMsg.c
 *
 * Description  : This module process the messages received from EVM.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "ea.h"
#include "exports/scsi-api.h"

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static void
ea_processOoVolEvt(evm_volId_t  volId,
                   boolean      online);

static void
ea_processDelVolEvt(evm_volId_t volId);

static void
ea_processCrtVolEvt(evm_VolConfig_t *   volCfg);

static void
ea_processGrwVolEvt(evm_VolConfig_t *   volCfg,
                    evm_strgConfig_t *  strgCfg);

static void
ea_processModVolEvt(evm_VolConfig_t *   volCfg);

static void
ea_processFullVolEvt(evm_volId_t    volId);

static void
ea_processFailLunEvt(evm_volId_t    volId,
                     scsi_id_t      strgId);

/*------------------------------------------------------------------------------
 * Name         : ea_processOoVolEvt()
 *
 * Description  :
 * Process the "volume on/off line" event received from EVM.
 * Notify AV about the change of the volume status.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processOoVolEvt(evm_volId_t   volId,
                   boolean       online)

{
    ea_ooVolNtfy_t  ntfy;
    
    EA_FCT_PRINT(("> ea_processOoVolEvt()\n"));

    ntfy.hdr.msgType = EA_OO_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volId),
           (void *) &volId,
           sizeof(evm_volId_t));
    ntfy.volOnline = online;

    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_ooVolNtfy_t));

    EA_FCT_PRINT(("< ea_processOoVolEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDelVolEvt()
 *
 * Description  :
 * Process the "destroy volume" event received from EVM.
 * Notify AV about the deletion of the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processDelVolEvt(evm_volId_t  volId)

{
    ea_delVolNtfy_t ntfy;
    
    EA_FCT_PRINT(("> ea_processDelVolEvt()\n"));

    ntfy.hdr.msgType = EA_DEL_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volId),
           (void *) &volId,
           sizeof(evm_volId_t));

    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_delVolNtfy_t));

    EA_FCT_PRINT(("< ea_processDelVolEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processCrtVolEvt()
 *
 * Description  :
 * Process the "create volume" event received from EVM.
 * Notify AV about the creation of the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processCrtVolEvt(evm_VolConfig_t * volCfg)

{
    ea_crtVolNtfy_t ntfy;
    
    EA_FCT_PRINT(("> ea_processCrtVolEvt()\n"));

    ntfy.hdr.msgType = EA_CRT_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volCfg),
           (void *) volCfg,
           sizeof(evm_VolConfig_t));

    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_crtVolNtfy_t));

    EA_FCT_PRINT(("< ea_processCrtVolEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGrwVolEvt()
 *
 * Description  :
 * Process the "grow volume/add storage" event received from EVM.
 * Notify AV about the growth of the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processGrwVolEvt(evm_VolConfig_t *   volCfg,
                    evm_strgConfig_t *  strgCfg)

{
    ea_grwVolNtfy_t ntfy;
    
    EA_FCT_PRINT(("> ea_processGrwVolEvt()\n"));

    ntfy.hdr.msgType = EA_GRW_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volCfg),
           (void *) volCfg,
           sizeof(evm_VolConfig_t));

    ntfy.lunList.num = 0;
// shouldn't pass WWN and LUN any more as a device could show
// up with multiple WWN:LUNs
//  ntfy.lunList.lun[0].wwn = SCSI_ID_GET_WWN(strgCfg->strgId);
//  ntfy.lunList.lun[0].lun = SCSI_ID_GET_LUN(strgCfg->strgId);
    ntfy.lunList.num++;

#if 0
// AgileVision is not currently processing this message
    ea_sendEvNtfy(&ntfy, sizeof(ea_grwVolNtfy_t));
#endif

    EA_FCT_PRINT(("< ea_processGrwVolEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processModVolEvt()
 *
 * Description  :
 * Process the "change volume configuration" event received from EVM.
 * Notify AV about the modification of the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processModVolEvt(evm_VolConfig_t *   volCfg)

{
    ea_modVolNtfy_t ntfy;
    
    EA_FCT_PRINT(("> ea_processModVolEvt()\n"));

    ntfy.hdr.msgType = EA_MOD_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volCfg),
           (void *) volCfg,
           sizeof(evm_VolConfig_t));

    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_modVolNtfy_t));

    EA_FCT_PRINT(("< ea_processModVolEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processFullVolEvt()
 *
 * Description  :
 * Process the "volume unable to grow" event received from EVM.
 * Notify AV that the volume is full.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processFullVolEvt(evm_volId_t volId)

{
    ea_fullVolNtfy_t    ntfy;
    
    EA_FCT_PRINT(("> ea_processFullVolEvt()\n"));

    ntfy.hdr.msgType = EA_FULL_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volId),
           (void *) &volId,
           sizeof(evm_volId_t));

    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_fullVolNtfy_t));

    EA_FCT_PRINT(("< ea_processFullVolEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processFailLunEvt()
 *
 * Description  :
 * Process the "volume failure" event received from EVM.
 * Notify AV that the LUN of a volume failed.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processFailLunEvt(evm_volId_t volId,
                     scsi_id_t  strgId)

{
    ea_failLunNtfy_t    ntfy;
    
    EA_FCT_PRINT(("> ea_processFailLunEvt()\n"));

    ntfy.hdr.msgType = EA_FAIL_LUN_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volId),
           (void *) &volId,
           sizeof(evm_volId_t));
    ntfy.lunList.num = 1;
// shouldn't pass WWN and LUN any more as a device could show
// up with multiple WWN:LUNs
//    ntfy.lunList.lun[0].wwn = SCSI_ID_GET_WWN(strgId);
//    ntfy.lunList.lun[0].lun = SCSI_ID_GET_LUN(strgId);    

#if 0
// AgileVision is not currently processing this message
    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_failLunNtfy_t));
#endif

    EA_FCT_PRINT(("< ea_processFailLunEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processTakeoverEvt()
 *
 * Description  :
 * Process the "takeover event" event received from EVM.
 * Notify AV that the LUN of a volume failed.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processTakeoverEvt(evm_volId_t volId)

{
    ea_takeoverVolNtfy_t    ntfy;
    
    EA_FCT_PRINT(("> ea_processTakeoverEvt()\n"));

    ntfy.hdr.msgType = EA_TAKEOVER_VOL_NTFY;
    ntfy.hdr.userInfo = 0;   //not used
    
    memcpy((void *) &(ntfy.volId),
           (void *) &volId,
           sizeof(evm_volId_t));

    ea_sendEvNtfy(&ntfy,
                  sizeof(ea_takeoverVolNtfy_t));

    // TODO!!!
    // probably need to do a bit more than just notifying AV
    // need to bring down the volume
    
    EA_FCT_PRINT(("< ea_processTakeoverEvt()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processEvmMsg()
 *
 * Description  : Process all messages received from EVM.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_processEvmMsg(ea_msg_t *          msg,
                 ea_senderInfo_t *   sender)
    
{
    EA_FCT_PRINT(("> ea_processEvmMsg()\n"));
                       
    if (msg->len < sizeof(evm_msgHdr_t))
    {
        printf("Got truncated message of len %d", msg->len);
    }
    else
    {
        evm_msgHdr_t *  hdr;
        
        hdr = (evm_msgHdr_t *) msg->buf;
        EA_FCT_PRINT(("ea_processEvmMsg msg type is %d\n", hdr->msgType));
        /*                // No more storage pool

         * Process the message according to its type.
         */
        switch (hdr->msgType)
        {
#ifdef ISABELLE // code to delete
        case EVM_LVOL_EVT:
        {
            evm_MsgLVolEvt_t * evt;

            evt = (evm_MsgLVolEvt_t *) msg->buf;
            EA_DBG_PRINT(("EVM_LVOL_EVT from node %s\n", evt->node));

            switch (evt->event)
            {
            case EVM_LVOL_ONLINE_E:     // LV has been put online
            {
                ea_processOoVolEvt(evt->lvid, TRUE);
                break;
            }
            
			case EVM_LVOL_OFFLINE_E:    // LV had been taken offline
            {
                ea_processOoVolEvt(evt->lvid, FALSE);
                break;
            }
                  
			case EVM_LVOL_DESTROY_E:    // No more LV
            {
                ea_processDelVolEvt(evt->lvid);
                break;
            }
            
			case EVM_LVOL_CHANGE_E:     // LV has changed somehow
            {
                switch(evt->chg_type)
                {
                case EVM_LVOL_NEW:      // New LV just created
                {
                    ea_processCrtVolEvt(&(evt->chg_cfg));
                    break;
                }
                
                case EVM_LVOL_ADD:      // Manual add of storage to an LV
                case EVM_LVOL_GROW:     // Automatic growth
                {
                    ea_processGrwVolEvt(&(evt->chg_cfg), &(evt->chg_strg));
                    break;
                }
                
                case EVM_LVOL_MOD:      // LV's parameters have been changed
                {
                    ea_processModVolEvt(&(evt->chg_cfg));
                    break;
                }

                default:
                {
                    EA_ERR_PRINT(("Unknown sub-event (%d)\n", evt->chg_type));
                    break;
                }
                } /* switch */
                break;
            }
            
			case EVM_LVOL_FULL_E:       // LV unable to grow
            {
                ea_processFullVolEvt(evt->lvid);
                break;
            }
            
			case EVM_LVOL_STRG_FAILED_E:    // Something on this LV has failed
            {
                ea_processFailLunEvt(evt->lvid, evt->strgid);
                break;
            }
            
			case EVM_LVOL_TAKEOVER_E:
            {
                // Somebody is willing to takeover the volume
                ea_processTakeoverEvt(evt->lvid);
                break;
            }

            default:
            {
                EA_ERR_PRINT(("Unknown event (%d)\n", evt->event));
                break;
            }
            } /* switch */
            
            break;
        }
#endif // ISABELLE

        default:
        {
            EA_ERR_PRINT(("Unknown msg type (%d)\n", hdr->msgType));
            break;
        }
        } /* switch */
    }
    
    EA_FCT_PRINT(("< ea_processEvmMsg()\n"));
    return;
}

/*----------------------------------------------------------------------------*/
