/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : sim-ev-rx.c
 *
 * Description  :
 * This module simulates the reception part of Eagle Vision (EV).
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <signal.h>

#include "ea.h"
#include "ea-api.h"

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static void
ev_terminate(int sig);

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

/* EV's file descriptor */
static int fd;

/*------------------------------------------------------------------------------
 * Name         : ev_terminate()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ev_terminate(int sig)

{
    /* Unregister EV.
     */
    unRegAgileApp(fd);
    
    exit(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : main
 *
 * Description  : EV's entry point.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int main(int argc, char **argv, char **envp) 
{
    /* Register to receive messages for EV.
     */
	fd = regAgileApp("ev");
	if (fd < 0)
    {
        perror("regAgileApp() failed");
		exit(NFX_ERR);
	}
    
    /* * Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, ev_terminate);
    signal(SIGINT, ev_terminate);

    /* Receive messages and process them.
     */
	while (TRUE)
    {
        ea_senderInfo_t sender; // not used
        ea_msg_t        msg;

        /* Wait for messages...
         */
		msg.len = recvAgileMsg(fd,
                               msg.buf, sizeof(msg.buf),
                               NULL,
                               sender.name,
                               &sender.cpu,
                               &sender.slot,
                               &sender.plane);
        /* Something has been received, check it.
         */
		if (msg.len < 0)
        {
			printf("recvAgileMsg() failed\n");
		} 
		else
        {
            ea_msgHdr_t * hdr;

            hdr = (ea_msgHdr_t *) &(msg.buf);

            printf("A message (%d) has been received from %s!\n",
                   hdr->msgType, sender.name);
            printf("UserInfo (cookie) = %d\n", hdr->userInfo);
#ifdef ISABELLE // for debug pupose only
#endif // ISABELLE

            switch(hdr->msgType)
            {
            case EA_ADD_FSYS_RSP:
            {
                ea_addFsysRsp_t * rsp = (ea_addFsysRsp_t *) &(msg.buf);
                printf("EA_ADD_FSYS_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_MOD_FSYS_RSP:
            {
                ea_modFsysRsp_t * rsp = (ea_modFsysRsp_t *) &(msg.buf);
                printf("EA_MOD_FSYS_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_MNT_FSYS_RSP:
            {
                ea_mntFsysRsp_t * rsp = (ea_mntFsysRsp_t *) &(msg.buf);
                printf("EA_MNT_FSYS_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_UMNT_FSYS_RSP:
            {
                ea_umntFsysRsp_t * rsp = (ea_umntFsysRsp_t *) &(msg.buf);
                printf("EA_UMNT_FSYS_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_DEL_FSYS_RSP:
            {
                ea_delFsysRsp_t * rsp = (ea_delFsysRsp_t *) &(msg.buf);
                printf("EA_DEL_FSYS_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_GET_FSYS_RSP:
            {
                ea_getFsysRsp_t * rsp = (ea_getFsysRsp_t *) &(msg.buf);
                printf("EA_GET_FSYS_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_ADD_FSET_RSP:
            {
                ea_addFsetRsp_t * rsp = (ea_addFsetRsp_t *) &(msg.buf);
                printf("EA_ADD_FSET_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_MOD_FSET_RSP:
            {
                ea_modFsetRsp_t * rsp = (ea_modFsetRsp_t *) &(msg.buf);
                printf("EA_MOD_FSET_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }

            case EA_DEL_FSET_RSP:
            {
                ea_delFsetRsp_t * rsp = (ea_delFsetRsp_t *) &(msg.buf);
                printf("EA_DEL_FSET_RSP (%s)\n",
                       (rsp->result == NFX_OK)? "OK":"ERROR");
                break;
            }
#ifdef ISABELLE // code to delete
            case EA_RSP:
            {
                printf("EA_RSP\n");
                {
                    ea_Rsp_t *  rsp;

                    rsp = (ea_Rsp_t *) &(msg.buf);
                    printf("\tresult = %s\n",
                                 (rsp->result == NFX_OK)? "OK :)":"ERROR :(");
                }
                break;
            }
#endif // ISABELLE

            case EA_CRT_VOL_NTFY:
            {
                printf("EA_CRT_VOL_NTFY\n");
                break;
            }
    
            case EA_DEL_VOL_NTFY:
            {
                printf("EA_DEL_VOL_NTFY\n");
                break;
            }

            case EA_MOD_VOL_NTFY:
            {
                printf("EA_MOD_VOL_NTFY\n");
                break;
            }
    
            case EA_OO_VOL_NTFY:
            {
                printf("EA_OO_VOL_NTFY\n");
                {
                    ea_ooVolNtfy_t *    ntfy;

                    ntfy = (ea_ooVolNtfy_t *) &(msg.buf);
                    printf("\tvolId = %lld\n", ntfy->volId.whole_thing);
                    printf("\tvolOnline = %s\n",
                           (ntfy->volOnline == TRUE)? "OnLine" : "OffLine");
                }
                break;
            }
    
            case EA_GRW_VOL_NTFY:
            {
                printf("EA_GRW_VOL_NTFY\n");
                break;
            }
    
            case EA_FULL_VOL_NTFY:
            {
                printf("EA_FULL_VOL_NTFY\n");
                break;
            }

            case EA_FAIL_LUN_NTFY:
            {
                printf("EA_FAIL_LUN_NTFY\n");
                break;
            }
            
            case EA_TAKEOVER_VOL_NTFY:
            {
                printf("EA_TAKEOVER_VOL_NTFY\n");
                break;
            }
            
            default:
            {
                printf("Unknow message type (%d)\n", hdr->msgType);
                break;
            }
            } // switch
        }
    } // while
    
    printf("< main()\n");
    exit(NFX_ERR);  // should never exit by this way!
}

/*----------------------------------------------------------------------------*/
