/*
 *-----------------------------------------------------------------
 *
 * Name:	eee-app.c
 *
 * RCS:  $Id: eee-app.c,v 1.16 2002/11/07 01:20:03 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Application interface routines
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/eee-api.h>

// This header defines the application ids and names. It's synchronized
// with the agile.services addendum to /etc/services.
#include <exports/app_ids.h>

/*-----------------------------------------------------------------
 * Name:	eee_allocateApplicationId(uchar8  * appName)
 *
 * Description:	An application that wants to receive packets from
 *				the EEE layer must allocate an application Id with
 *				this function and register a receive function to that
 *				appId in the following function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns new applicationId,
							 * EEE_ERR_INVALID_APP_NAME
							 */
eee_allocateApplicationId(
	uchar8	* appName
)
{
	uint32	i;

	for (i=0; ; i++) { 
		if (app_trans_table[i].app_id < 0)
			break;
		if (strcmp(app_trans_table[i].app_name, appName) == 0)
			return (app_trans_table[i].app_id);
	}

	printf("Invalid Application name, %s\n", appName);

	return EEE_ERR_INVALID_APP_NAME;

}	/* eee_allocateApplicationId() */
 
/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32						/* returns applicationId,
							 * EEE_ERR_INVALID_APP_NAME
							 */
eee_getApplicationId(
	uchar8	* appName
)
{
	uint32	i;

	for (i=0; ; i++) { 
		if (app_trans_table[i].app_id < 0)
			break;
		if (strcmp(app_trans_table[i].app_name, appName) == 0)
			return (app_trans_table[i].app_id);
	}

	printf("Invalid Application name, %s\n", appName);

	return EEE_ERR_INVALID_APP_NAME;

}	/* eee_getApplicationId() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
uchar8 *
eee_getAppNameFromId(
	int32	app_id
)
{
	uint32	i;

	for (i=0; ; i++) {
		if (app_trans_table[i].app_id < 0)
			break;
		if (app_trans_table[i].app_id == app_id)
			return (app_trans_table[i].app_name);
	}

	return NULL;

}	/* eee_getAppNameFromId() */

/*-----------------------------------------------------------------
 * Name:	eee_registerReceiveFunc()
 *
 * Description:	To register a receive function which will be called
 *				when the EEE layer receives a packet destined for the
 *				indicated application Id.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */
int32								/* NFX_OK or
									 * EEE_ERR_APP_ALREADY_REGISTERED
									 */
eee_registerReceiveFunc(
	int32	appId,
	void	(*rcv_func)(eee_descPtr_t edesc),
	void	(*mcast_rcv_func)(eee_descPtr_t edesc))
{
	if (0 <= appId && appId < EEE_MAX_APP_CLIENTS) {
#ifndef DIAG_R7000	
		if ((eee.rcv.uni[appId]) || (eee.rcv.multi[appId]))
			return EEE_ERR_APP_ALREADY_REGISTERED;
#endif	
		eee.rcv.uni[appId]= rcv_func;
		eee.rcv.multi[appId] = mcast_rcv_func;
	}
	else {
		panic("eee_registerReceiveFunc: invalid application id %d\n",
			  appId);
	}

	return NFX_OK;

}	/* eee_registerReceiveFunc() */

/*-----------------------------------------------------------------
 * Name:	eee_unregisterReceiveFunc()
 *
 * Description:	Remove function addresses attached to application Id
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/19/01
 *
 *-----------------------------------------------------------------
 */
void
eee_unregisterReceiveFunc(
	int32	appId
)
{
	if (0 <= appId && appId < EEE_MAX_APP_CLIENTS) {
		eee.rcv.uni[appId] = NULL;
		eee.rcv.multi[appId] = NULL;
	}
	else {
		panic("eee_unregisterReceiveFunc: invalid application id %d\n",
			  appId);
	}

}	/* eee_unregisterReceiveFunc() */

/*-----------------------------------------------------------------
 * Name:	eee_registerCleanupRoutine()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	11/12/01
 *
 *-----------------------------------------------------------------
 */
int32
eee_registerCleanupRoutine(
	void	(*function_name)(eee_descPtr_t	edesc)
)
{
	int32	i = 0;

	while (i < EEE_MAX_FREE_ROUTINES-1) {
		if (!eee.free_routine[i]) {
			eee.free_routine[i] = function_name;
			return i+1;
		}
	}

	panic("Too many registered free routines\n");
	return 0;

}	/* eee_registerCleanupRoutine() */

