/*
 *-----------------------------------------------------------------
 *
 * Name:	eee-test.c
 *
 * RCS:  $Id: eee-test.c,v 1.31 2002/12/10 00:37:44 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Test routines for the EEE layer.
 *
 *				There are 2 test options:
 *					1) simple - send a single packet to a node.  The
 *								simple test rcv function should receive
 *								the packet.  Counters will be updated.
 *
 *					2) complex - send packets from each CPU on a module
 *								 to every other CPU on that module.  This
 *								 will occur until something triggers a stop
 *								 condition.
 *								 The 'complex' test runs in 2 modes:
 *										1) allocate/deallocate a
 *											packet each time it gets sent.
 *										2) allocate packets at init time
 *											and have the EEE layer just
 *											spin them around.
 *
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/19/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>

/* #define EEE_TEST_SPIN		0 */
#define EEE_TEST_ACK_SIMPLE	1

eee_testPktSpec_t	eee_test_spec[3][EEE_TEST_NUM_PKT_SPEC] = {
	{	{2,		8192},			/* EEE_USER_DATA */
		{2,		8192},
		{2,		8192},
		{2,		8192},
		{2,		8192},
		{2,		8192},
		{2,		8192},
		{2,		8192},
	},
	{	{0,		1024},			/* EEE_CONTROL */
		{0,		1024},
		{0,		1024},
		{1,		1024},
		{1,		1024},
		{0,		1024},
		{1,		1024},
		{1,		1024},
	},
	{	{0,		1400},			/* EEE_MGMT */
		{0,		1400},
		{0,		1400},
		{1,		1400},
		{1,		1400},
		{0,		1400},
		{1,		1400},
		{1,		1400},
	}
};

//WTG test to see if I can fit more into the Recovery Prom
// #define EEE_TEST_PKT_SIZE_MAX	3200
// uchar8	eee_mgmt_tst_pkt[EEE_TEST_PKT_SIZE_MAX];

uint32	eee_test_spec_index[3];

#ifndef PROM
uint32	next_exp_seq_num[NFX_NUM_SLOTS_PER_CHASSIS][NFX_MAX_CPUS_PER_SLOT][3];
#endif

uint32	simple_exp_next_seq_num;

/* System DVT Variables  */

#ifndef PROM
eee_dvt_test_stats	dvtstats;
#endif
eee_test_hdr_t	* test_hdr;
uint32  incoming_appid;
#ifndef PROM
uint32  sentFromSSC[4][3]; /* [cpu][plane]  Now max cpu on NFP/NFC is 4 per card */

#ifdef FCNIM
int32 ispfc_fcadmin(int32 argc, char **argv);
#endif
#endif

#ifndef PROM
void eee_reinit_validity_table();
void dvtRemake_pkt(uint32 slot, uint32 cpu, uint32 seq, uint32 plane, uint32 spin);
void eee_systemDVTinit();
void eee_systemDVTSend(uint32 slot, uint32 cpu, uchar8 * bfr, uint32 size);
void dvt_front_loopback(eee_test_hdr_t	*test_hdr);
void eee_dvttestFrontEnd(uint32 plane);
void dvt_rcvd_txrx_front_pkt(eee_descPtr_t edesc);

volatile uint32 dvtStop;

extern sysdvt_qlogic_t  *qdvt;
#endif
extern int32  pktSize;

/*-----------------------------------------------------------------
 * Name:	eee_testCreateSimplePacket()
 *
 * Description:	Create an EEE packet of the given size
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/19/01
 *
 *-----------------------------------------------------------------
 */
int32								/*  NFX_OK,
									 *	EEE_ERR_NO_DESC,
									 *	EEE_ERR_NO_BUF,
									 *	EEE_ERR_BAD_PARAM,
									 *	EEE_ERR_PACKET_TOO_LARGE
									 */
eee_testCreateSimplePacket(
	uint32	dest_slot,			/* slot number */
	uint32	dest_cpu,			/* cpu id on that slot */
	uint32	size,				/* number of bytes in test message to send */
	uint32	bufsize,			/* EEE_BUF_SMALL,
									EEE_BUF_LARGE,
									EEE_BUF_ULTRA */
	uint32	plane,				/* 0 = data, 1 = mgmt, 2 = control */
	uint32	seq_num)
{
	uint32			mp;
	eee_desc_t		* edesc;
	eee_descBuf_t	* bd;
	uchar8			* buf[EEE_MAX_BUFS];
	uint32			bd_size[EEE_MAX_BUFS];
	uint32			num_bufs;
	uint32			i, j;
	int32			rc;
	uint32			buf_bytes;
	uint32			remain;
	uint32			dtype;
	uint32			*bufhdr;

	/*
	 * Verify parameters
	 */
	if (bufsize >= EEE_NUM_BUF_SIZES)
		return EEE_ERR_BAD_PARAM;

	if (size > ((eee.buf_size[bufsize]-eee.buf_offset) * EEE_MAX_BUFS))
		return EEE_ERR_BAD_PARAM;

	if (size < 64) {
		printf("nbytes must be >= 64\n");
		return EEE_ERR_BAD_PARAM;
	}

	/*
	 * Allocate descriptor from appropriate memory pool
	 */
	if (dest_slot == nfxMySlotId) {
		if (dest_cpu >= EEE_NUM_CPUS)
			return EEE_ERR_BAD_PARAM;
		mp = GET_MEMPOOL_INDEX_FROM_CPU_ID(dest_cpu);
	}
	else
		mp = eee.my_pool_id;

	edesc = eee_allocateDesc(mp, NFX_COS_HIGH);
	if (!edesc)
		return EEE_ERR_NO_DESC;

	/*
	 * Allocate buffers from appropriate memory pool
	 */
	num_bufs = 0;
	buf_bytes = eee.buf_size[bufsize] - eee.buf_offset;
	remain = size;
	do {
		buf[num_bufs] = eee_allocateBuffer(mp, bufsize, NFX_COS_HIGH);
		if (!buf[num_bufs]) {
			for (j=0; j<num_bufs; j++)
				eee_deallocateBuffer(buf[j]);
			eee_deallocateDesc(edesc);
			return EEE_ERR_NO_BUF;
		}
		if (remain > buf_bytes) {
			bd_size[num_bufs] = buf_bytes;
			remain -= buf_bytes;
		}
		else {
			bd_size[num_bufs] = remain;
			remain = 0;
		}
		num_bufs++;
	} while (remain);

	/*
	 * Fill in buffers with test data
	 */
	for (i=0; i<num_bufs; i++) {
		for (j=0; j<bd_size[i]; j++)
			buf[i][j] = j & 0xff;
	}
	bufhdr = (uint32 *)&(buf[0][0]);
	*bufhdr++ = nfxMySlotId;
	*bufhdr++ = nfxMyCPUId;
	*bufhdr++ = size;

	/*
	 * Create packet
 	 */
	bd = &(edesc->bd[0]);
	edesc->hdr.control = (num_bufs << EDESC_BUFCNT_SHIFT) | size;
	if (num_bufs == EEE_MAX_BUFS) {
		edesc->hdr.control |= EDESC_COMMAND_USE_FIRST_BUFFER;
	}
	else
		bd++;

	if (plane == 0)
		dtype = EEE_USER_DATA;
	else if (plane == 1)
		dtype = EEE_MGMT;
	else if (plane == 2)
		dtype = EEE_CONTROL;
	else if (plane == 3)
		dtype = EEE_SAC;
	else {
		dtype = 0;
		panic("invalid plane, %u\n", plane);
	}

	*bufhdr++ = dtype;
	*bufhdr++ = seq_num;

	edesc->hdr.dest_port = PACK_UNICAST_PORT_ADDRESS(dest_slot, dest_cpu,
													eee.simple_test_app_id,
													dtype, NFX_COS_HIGH, 0);
	edesc->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId,
													eee.simple_test_app_id,
													dtype, NFX_COS_HIGH, 0);
	edesc->hdr.attr = 0;

	for (i=0; i<num_bufs; i++) {
		bd->control = EDESC_BD_CMD_IN_USE | bufsize;
		bd->len = bd_size[i];
		bd->buf = buf[i];
		bd++;
	}

	/*
	 * Send packet
	 */
	if ((rc = eee_forwardPacket(edesc)) != EEE_FORWARDED_PACKET)
		return rc;

	return NFX_OK;

}	/* eee_testCreateSimplePacket() */


#ifndef PROM
/*-----------------------------------------------------------------
 * Name:	eee_testCreateDVTTestPkt()
 *
 * Description:	Create a packet from a known sequence
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/01
 *
 *-----------------------------------------------------------------
 */
eee_descPtr_t							/* returns edesc addr or NULL */
eee_testCreateDVTTestPkt(
	uint32	slot,
	uint32	cpu,
	uint32	plane,
	uint32	spin,
	uint32  seq
)
{
	eee_descPtr_t	edesc;
	uint32			size, bs;
	uint32			remain;
	eee_descBuf_t	* bd;
	uchar8			* buf[EEE_MAX_BUFS];
	uint32			bd_size[EEE_MAX_BUFS];
	uint32			num_bufs;
	uint32			buf_bytes;
	uint32			i;
	eee_test_hdr_t	* test_hdr;

	/*
	 * Get info from test_pkt_spec structure about packet to be created
	 */
	size = eee_test_spec[plane][eee_test_spec_index[plane]].num_bytes;
	bs = eee_test_spec[plane][eee_test_spec_index[plane]].bufsize;
	/*
	 * Allocate descriptor from my pool
	 */
	edesc = eee_allocateDesc(eee.my_pool_id, NFX_COS_HIGH);
	if (!edesc) {
		eee.test_dvt_send_err[plane]++;
		return NULL;
	}

	/*
	 * Allocate buffers from my pool
	 */
	num_bufs = 0;
	buf_bytes = eee.buf_size[bs] - eee.buf_offset;
	remain = size;
	do {
		buf[num_bufs] = eee_allocateBuffer(eee.my_pool_id, bs, NFX_COS_HIGH);
		if (!buf[num_bufs]) {
			for (i=0; i<num_bufs; i++)
				eee_deallocateBuffer(buf[i]);
			eee_deallocateDesc(edesc);
			return NULL;
		}
		if (remain > buf_bytes) {
			bd_size[num_bufs] = buf_bytes;
			remain -= buf_bytes;
		}
		else {
			bd_size[num_bufs] = remain;
			remain = 0;
		}
memset(buf[num_bufs], 0xab, buf_bytes);
		num_bufs++;
	} while (remain);

	/*
	 * Create packet
 	 */
	bd = &(edesc->bd[0]);
	edesc->hdr.control = (num_bufs << EDESC_BUFCNT_SHIFT) | size;
	if (num_bufs == EEE_MAX_BUFS) {
		edesc->hdr.control |= EDESC_COMMAND_USE_FIRST_BUFFER;
	}
	else
		bd++;

	for (i=0; i<num_bufs; i++) {
		bd->control = EDESC_BD_CMD_IN_USE | bs;
		bd->len = bd_size[i];
		bd->buf = buf[i];
		bd++;
	}

	edesc->hdr.dest_port = PACK_UNICAST_PORT_ADDRESS(slot,
													cpu,
													eee.dvt_test_app_id,
													plane,
													NFX_COS_HIGH,
													PORT_BEST_FIT);


	edesc->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
													nfxMyCPUId,
													eee.dvt_test_app_id,
													plane,
													NFX_COS_HIGH,
													PORT_BEST_FIT);
	edesc->hdr.attr = 0;
	test_hdr = (eee_test_hdr_t *)(buf[0]);
	test_hdr->cmd = EEE_MGMT_TEST_CMD_DATA;
	/* For missing pkts we need to create pkt with only the missed seq no.
	 * So, when seq is not given(0xFFFF), base seq no from eee_test_spec_index
	 * otherwise, make pkt with provided seq no.
	 */ 
	if(seq == 0xFFFF)
		test_hdr->seq = eee_test_spec_index[plane];
	else
		test_hdr->seq = seq;
	test_hdr->size = size;
	test_hdr->orig_slot = nfxMySlotId;
	test_hdr->orig_cpu = nfxMyCPUId;
	test_hdr->plane = plane;
	test_hdr->spin = spin;
	eee_test_spec_index[plane] =
					(eee_test_spec_index[plane] + 1) % EEE_TEST_NUM_PKT_SPEC;
	
	return (edesc);

}	/* eee_testCreateDVTTestPkt() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	12/01/01
 *
 *-----------------------------------------------------------------
 */
int32
eee_testSendDVTStartPkt(
	uint32	slot,
	uint32	cpu,
	uint32	plane,
	uint32	num_pkts,
	uint32	spin
)
{
	eee_test_hdr_t	test_msg;
	uint32			dp;
	uint32			sp;

	test_msg.cmd = EEE_MGMT_TEST_CMD_START;
	test_msg.size = num_pkts;
	test_msg.plane = plane;
	test_msg.orig_slot = nfxMySlotId;
	test_msg.orig_cpu = nfxMyCPUId;
	test_msg.spin = spin;

	dp = PACK_UNICAST_PORT_ADDRESS(slot,
									cpu,
									eee.dvt_test_app_id,
									plane,
									NFX_COS_HIGH,
									PORT_BEST_FIT);

	sp = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
									nfxMyCPUId,
									eee.dvt_test_app_id,
									plane,
									NFX_COS_HIGH,
									PORT_BEST_FIT);

	return (eee_sendMessage(&test_msg, sizeof(eee_test_hdr_t), dp, sp));

}	/* eee_testSendDVTStartPkt() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	12/01/01
 *
 *-----------------------------------------------------------------
 */
int32
eee_testSendDVTStopPkt(
	uint32	slot,
	uint32	cpu,
	uint32	plane
)
{
	eee_test_hdr_t	test_msg;
	uint32			dp;
	uint32			sp;

	test_msg.cmd = EEE_MGMT_TEST_CMD_STOP;
	test_msg.orig_slot = nfxMySlotId;
	test_msg.orig_cpu = nfxMyCPUId;
	test_msg.plane = plane;

	dp = PACK_UNICAST_PORT_ADDRESS(slot,
									cpu,
									eee.dvt_test_app_id,
									plane,
									NFX_COS_HIGH,
									PORT_BEST_FIT);

	sp = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
									nfxMyCPUId,
									eee.dvt_test_app_id,
									plane,
									NFX_COS_HIGH,
									PORT_BEST_FIT);

	//printf("slot %d cpu %d, nfxMyCPUId %d, sp %08x dp %08x\n", slot, cpu, nfxMyCPUId,sp, dp);
	return (eee_sendMessage(&test_msg, sizeof(eee_test_hdr_t), dp, sp));

}	/* eee_testSendDVTStopPkt() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
int32
eee_testSendDVTClearPkt(
	uint32	slot,
	uint32	cpu,
	uint32	plane
)
{
	eee_test_hdr_t	test_msg;
	uint32			dp;
	uint32			sp;

	test_msg.cmd = EEE_MGMT_TEST_CMD_CLEAR;
	test_msg.orig_slot = nfxMySlotId;
	test_msg.orig_cpu = nfxMyCPUId;
	test_msg.plane = plane;

	dp = PACK_UNICAST_PORT_ADDRESS(slot,
									cpu,
									eee.dvt_test_app_id,
									plane,
									NFX_COS_HIGH,
									PORT_BEST_FIT);

	sp = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
									nfxMyCPUId,
									eee.dvt_test_app_id,
									plane,
									NFX_COS_HIGH,
									PORT_BEST_FIT);

	return (eee_sendMessage(&test_msg, sizeof(eee_test_hdr_t), dp, sp));

}	/* eee_testSendDVTStopPkt() */

/*-----------------------------------------------------------------
 * Name:	eee_testDVTStart(mode)
 *
 * Description:	Start EEE layer test
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/01
 *
 *-----------------------------------------------------------------
 */
int32
eee_testDVTStart(
	uint32	num_pkts,
	uint32	plane,
	uint32	spin
)
{
	uint32	slot, cpu, i;
	int32	rc;
	eee_descPtr_t	edesc;

	if (!eee.test_dvt_valid[nfxMySlotId][nfxMyCPUId][plane]) {
		printf("%s DVT test not valid on this CPU\n",
						(plane == EEE_MGMT) ? "MGMT" :
						((plane == EEE_CONTROL) ? "CONTROL" : "DATA"));
		return NFX_ERR;
	}

	if (eee.test_dvt_running[plane]) {
		printf("EEE test already running\n");
		return NFX_OK;
	}

	/*
	 * Broadcast start packet
	 */
	for (slot=0; slot<NFX_NUM_SLOTS_PER_CHASSIS; slot++) {
		for (cpu=0; cpu<NFX_MAX_CPUS_PER_SLOT; cpu++) {
			if (eee.test_start_valid[slot][cpu][plane]) {
				if ((slot == nfxMySlotId) && (cpu == nfxMyCPUId))
					continue;
				if (eee_testSendDVTStartPkt(slot, cpu, plane, num_pkts, spin) != NFX_OK)
					eee.test_dvt_send_err[plane]++;
			}
		}
	}

	/*
	 * Send num_pkts to each slot/cpu
	 */

	for (slot=0; slot<NFX_NUM_SLOTS_PER_CHASSIS; slot++) {
		for (cpu=0; cpu<NFX_MAX_CPUS_PER_SLOT; cpu++) {
			if (eee.test_dvt_valid[slot][cpu][plane]) {
				if ((slot == nfxMySlotId) && (cpu == nfxMyCPUId))
					continue;
				next_exp_seq_num[slot][cpu][plane] = 0;
				for (i=0; i<num_pkts; i++) {
					edesc = eee_testCreateDVTTestPkt(slot, cpu, plane, spin, 0xFFFF);
					if (edesc) {
						if ((rc = eee_forwardPacket(edesc)) !=
														EEE_FORWARDED_PACKET)
							eee.test_dvt_send_err[plane]++;
						else
							eee.test_dvt_my_sent[slot][cpu][plane]++;
					}
					else
						eee.test_dvt_send_err[plane]++;
				}
			}
		}
	}

#ifdef USE_SMP						
	ATOMIC_INC_INT(&eee.test_dvt_running[plane], 1);
#else						
	eee.test_dvt_running[plane] = 1;
#endif
	return NFX_OK;

}	/* eee_testDVTStart() */


/*-----------------------------------------------------------------
 * Name:	eee_testDVTEnd()
 *
 * Description:	Stop EEE layer test
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/01
 *
 *-----------------------------------------------------------------
 */
void
eee_testDVTEnd(
	uint32	plane
)
{
	uint32	slot, cpu;

	if (!eee.test_dvt_running[plane]) {
		printf("EEE test not running\n");
		return;
	}

	dvtStop = 1;
	/*
	 * Broadcast start packet
	 */
	if(!sentFromSSC[nfxMyCPUId][plane]) {
	for (slot=0; slot<NFX_NUM_SLOTS_PER_CHASSIS; slot++) {
		for (cpu=0; cpu<NFX_MAX_CPUS_PER_SLOT; cpu++) {
			if (eee.test_start_valid[slot][cpu][plane]) {
				if ((slot == nfxMySlotId) && (cpu == nfxMyCPUId))
					continue;
#ifdef NFP_FP
				/* To prevent a recursive loop */
				if((slot == NFX_SLOTTYPE_NFP) && ((cpu == NFX_NFP_FP1) 
							|| (cpu == NFX_NFP_FP2)))
					continue;
#endif				
				if (eee_testSendDVTStopPkt(slot, cpu, plane) != NFX_OK)
					eee.test_dvt_send_err[plane]++;
			}
		}
	}
	}
#ifdef USE_SMP						
	ATOMIC_AND_INT(&eee.test_dvt_running[plane], 0);
#else						
	eee.test_dvt_running[plane] = 0;
#endif	
	/* If test was initiated from the SSC, set the correct Init table */
	if(sentFromSSC[nfxMyCPUId][plane])
		eee_testInit();
	sentFromSSC[nfxMyCPUId][plane] = 0;

}	/* eee_testDVTEnd() */

/*-----------------------------------------------------------------
 * Name:	eee_testDVTClear()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
void
eee_testDVTClear(
	uint32	plane
)
{
	uint32	slot, cpu;

	/*
	 * Broadcast start packet
	 */
	for (slot=0; slot<NFX_NUM_SLOTS_PER_CHASSIS; slot++) {
		for (cpu=0; cpu<NFX_MAX_CPUS_PER_SLOT; cpu++) {
			if (eee.test_start_valid[slot][cpu][plane]) {
				if ((slot == nfxMySlotId) && (cpu == nfxMyCPUId))
					continue;
				if (eee_testSendDVTClearPkt(slot, cpu, plane) != NFX_OK)
					eee.test_dvt_send_err[plane]++;
			}
		}
	}

#ifdef USE_SMP						
	ATOMIC_AND_INT(&eee.test_dvt_running[plane], 0);
#else						
	eee.test_dvt_running[plane] = 0;
#endif

}	/* eee_testDVTEnd() */

/*-----------------------------------------------------------------
 * Name:	eee_testDVTPrintStats()
 *
 * Description:	Show send/receive stats for EEE test
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/01
 *
 *-----------------------------------------------------------------
 */
void
eee_testDVTPrintStats(
	uint32	plane
)
{
	uint32	i, j;
	uint32	tot_sent=0, tot_rcvd=0;

	printf("\nDVT BP %s Test : %s\n", (plane == EEE_MGMT) ? "MGMT" :
						((plane == EEE_CONTROL) ? "CONTROL" : "DATA"),
						eee.test_dvt_running[plane] ? "RUNNING" : "STOPPED");
	printf("     Local Pkts:\n");
	printf("        slot  cpu       sent       rcv\n");
	printf("        ----  ---   --------  --------\n");
	for (i=0; i<NFX_NUM_SLOTS_PER_CHASSIS; i++)
		for (j=0; j<NFX_MAX_CPUS_PER_SLOT; j++)
			if (eee.test_dvt_valid[i][j][plane]) {
				printf("          %2d   %2d   %8d  %8d\n",
								i, j,
								eee.test_dvt_my_sent[i][j][plane],
								eee.test_dvt_my_rcvd[i][j][plane]);
				tot_sent += eee.test_dvt_my_sent[i][j][plane];
				tot_rcvd += eee.test_dvt_my_rcvd[i][j][plane];
			}
	printf("\n");
	printf("     Remote Pkts:\n");
	printf("        slot  cpu        rcv      sent\n");
	printf("        ----  ---   --------  --------\n");
	for (i=0; i<NFX_NUM_SLOTS_PER_CHASSIS; i++)
		for (j=0; j<NFX_MAX_CPUS_PER_SLOT; j++)
			if (eee.test_dvt_valid[i][j][plane]) {
				printf("          %2d   %2d   %8d  %8d\n",
								i, j,
								eee.test_dvt_other_rcvd[i][j][plane],
								eee.test_dvt_other_sent[i][j][plane]);
				tot_sent += eee.test_dvt_other_sent[i][j][plane];
				tot_rcvd += eee.test_dvt_other_rcvd[i][j][plane];
			}
	printf("\n");
	printf("     Send Errors: %d\n", eee.test_dvt_send_err[plane]);
	printf("     Rcv Errors : %d\n", eee.test_dvt_rcv_err[plane]);
	printf("\n");


	printf("Totals: sent:  %d\n", tot_sent);
	printf("        rcvd:  %d\n", tot_rcvd);

}	/* eee_testDVTPrintStats() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	12/05/01
 *
 *-----------------------------------------------------------------
 */
void
eee_testDVTClearStats(
	uint32	plane
)
{
	uint32	i, j;

	for (i=0; i<NFX_NUM_SLOTS_PER_CHASSIS; i++) {
		for (j=0; j<NFX_MAX_CPUS_PER_SLOT; j++) {
			eee.test_dvt_my_sent[i][j][plane] = 0;
			eee.test_dvt_my_rcvd[i][j][plane] = 0;
			eee.test_dvt_other_sent[i][j][plane] = 0;
			eee.test_dvt_other_rcvd[i][j][plane] = 0;
		}
	}

	eee.test_dvt_send_err[plane] = 0;
	eee.test_dvt_rcv_err[plane] = 0;

}	/* eee_testDVTClearStats() */

#endif  /* not PROM */




static int
eee_is_local_cpu(int cpu)
{
#if defined(NFP_FP)
	return cpu == NFX_NFP_FP1 || cpu == NFX_NFP_FP2
		|| cpu == NFX_NFP_FP_ANY;
#elif defined(NFP_TXRX)
	return cpu == NFX_NFP_TXRX1 || cpu == NFX_NFP_TXRX2
		|| cpu == NFX_NFP_TXRX_ANY;
#else
	return cpu == nfxMyCPUId;
#endif
}

/*-----------------------------------------------------------------
 * Name:	eee_testSimpleRcv()
 *
 * Description:	Receive function for EEE test packets
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/19/01
 *
 *-----------------------------------------------------------------
 */
void
eee_testSimpleRcv(
	eee_descPtr_t	edesc
)
{
	uint32			pkt_len, bufcnt;
	uint32			i;
	eee_descBuf_t	* bd;
	uint32			* buf;
	uint32			src_slot, src_cpu, src_len, src_plane, src_seq_num;
	uchar8			* lbuf;
	int				error;
	int				ncopied;

	bufcnt = EEE_GET_NUM_BUFS(edesc);
	pkt_len = EEE_GET_PKT_LEN(edesc);

#if EEE_TEST_ACK_SIMPLE

//	printf("eee_testSimpleRcv, %p\n", edesc);
//	printf("pkt_len = %d bytes\n", pkt_len);
//	printf("num_bufs = %d\n", bufcnt);

	if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
		i = 0;
	else
		i = 1;
	bd = &(edesc->bd[i]);

	buf = (uint32 *)bd->buf;

	src_slot = *buf++;
	src_cpu = *buf++;
	src_len = *buf++;
	src_plane = *buf++;
	src_seq_num = *buf++;

	if (src_len != pkt_len) {
		printf("Error in simple test pkt_len, src_len = %d, rcv_len = %d\n",
															src_len, pkt_len);
		eee_freePacket(edesc);
		return;
	}
	if (src_plane != EEE_GET_DTYPE(edesc->hdr.dest_port)) {
		printf("Error in simple test dtype = %d, src_plane = %d\n",
								EEE_GET_DTYPE(edesc->hdr.dest_port), src_plane);
		eee_freePacket(edesc);
		return;
	}
#if 0
	if (src_seq_num != 0) {
		if (src_seq_num != simple_exp_next_seq_num) {
			printf("Error in simple test exp_seq_num = %d, sep_num = %d\n",
										simple_exp_next_seq_num, src_seq_num);
			eee_freePacket(edesc);
			return;
		}
		simple_exp_next_seq_num++;
	}
	else
		simple_exp_next_seq_num = 1;
#endif

	lbuf = eee_ramAlloc(pkt_len);
	if (lbuf == NULL) {
		printf("eee_testSimpleRcv: cannot allocate local buffer\n");
		eee_freePacket(edesc);
		return;
	}
	if ((ncopied = eee_copyPacketToLocalBuffer(lbuf, edesc, pkt_len, 0)) != pkt_len) {
		printf("eee_testSimpleRcv: copy failed ncopied=%d pktlen=%d\n",
			   ncopied, pkt_len);
	}
	if (!(src_slot == nfxMySlotId && eee_is_local_cpu(src_cpu))) {
		if ((error = eee_sendMessage(lbuf, pkt_len, edesc->hdr.src_port,
									 edesc->hdr.dest_port)) != NFX_OK) {
			printf("eee_testSimpleRcv: send message failed err=%d\n", error);
		}
	}
	eee_ramDealloc(lbuf);
	eee_freePacket(edesc);
#else
	eee_freePacket(edesc);
#endif

}	/* eee_testSimpleRcv() */



#ifndef PROM
/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void
eee_testDVTRcv(
	eee_descPtr_t	edesc
)
{

	/*
	 * if (packet == START) {
	 *		if (!test_running)
	 *			send START pkt to all slots/cpus
	 *			send N pkts to all slots/cpus
	 *			test_running = 1;
	 *		drop packet
	 * }
	 * if (packet == STOP) {
	 *		if (test_running) {
	 *			test_running = 0;
	 *			send STOP pkt to all slots/cpus
	 *		}
	 *		drop packet
	 * }
	 * if (packet == DATA) {
	 *		if (test_running) {
	 *			spin packet around to source
	 *		}
	 *		else {
	 *			if (orig_slot == myslot) && (orig_cpu == mycpu)
	 *				drop packet
	 *			else
	 *				spin packet around to source
	 *		}
	 * }
	 */

	eee_test_hdr_t	* test_hdr;
	eee_descBuf_t	* bd;
	uint32			cmd;
	int32			rc;
	uint32			src_slot, src_cpu, src_plane;
	uint32 			i,j,k;
	uint32 			plane;
#ifdef FCNIM
	fcstr_set_t		fcstr;
#endif	

    bd = &(edesc->bd[0]);
    if (!(edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER))
        bd++;

	src_slot = EEE_GET_SLOT_NUM(edesc->hdr.src_port);
	src_cpu = EEE_GET_DEST_CPU(edesc->hdr.src_port);
	src_plane = EEE_GET_DTYPE(edesc->hdr.src_port);

	test_hdr = (eee_test_hdr_t *)(bd->buf);

	cmd = test_hdr->cmd;

	switch(cmd) {
		case EEE_MGMT_TEST_CMD_START :
			if((src_slot == 0) && (src_cpu == 0)) {
#ifdef NFP_TXRX					
				if(test_hdr->plane == SYS_DVT_FRONT) {
					if(isTXRX()) {
						if(isCheetah())	
							if (eee.test_dvt_running[EEE_USER_DATA]) {
								printf("EEE test already running\n");
								eee_freePacket(edesc);
								return;
							}	
							dvt_front_loopback(test_hdr);	
					}				
				}	
#elif defined (FCNIM)
				if(test_hdr->plane == SYS_DVT_FRONT) {
					if(isFC()) {
						fcstr.str2 = "mode_diag_elback";
						ispfc_fcadmin(2,(char **)&fcstr);
					}	
				}	
#else
				if (eee.test_dvt_running[test_hdr->plane]) {
					printf("EEE test already running\n");
					eee_freePacket(edesc);
					return;
				}	
#endif				
				eee_reinit_validity_table();
				sentFromSSC[nfxMyCPUId][test_hdr->plane] = 1;
				eee_testSysDVTStart(test_hdr->size, test_hdr->plane, test_hdr->spin);
			}	
			else  /* Not initiated from SSC */
				eee_testDVTStart(test_hdr->size, test_hdr->plane, test_hdr->spin);

			eee_freePacket(edesc);
			break;

		case EEE_MGMT_TEST_CMD_STOP :
			if((src_slot == 0) && (src_cpu == 0)) {
			/* if sent from ssc, stop all traffic : need to modify later to support only 1 plane*/
				for(plane=0;plane<3;plane++) {	
#ifdef NFP_TXRX					
					if((isTXRX()) & (plane == EEE_USER_DATA))	
						eee_dvttestFrontEnd(plane);
					else
#endif						
						eee_testDVTEnd(plane);
				}	
#ifdef FCNIM
				fcstr.str2 = "mode_normal";
				ispfc_fcadmin(2,(char **)&fcstr);
#endif				
			}	
			else	
				eee_testDVTEnd(test_hdr->plane);

			eee_freePacket(edesc);
			break;

		case EEE_MGMT_TEST_CMD_CLEAR :
			eee_testDVTClearStats(test_hdr->plane);
			eee_freePacket(edesc);
			break;

		case EEE_MGMT_TEST_CMD_DATA :

			if (test_hdr->orig_slot == nfxMySlotId)
						   	//&& (test_hdr->orig_cpu == nfxMyCPUId)) /* smp */
			{
#ifdef USE_SMP						
				ATOMIC_INC_INT(&eee.test_dvt_my_rcvd[src_slot][src_cpu][src_plane], 1);
#else						
				eee.test_dvt_my_rcvd[src_slot][src_cpu][src_plane]++;
#endif				
				if (eee.test_dvt_running[test_hdr->plane]) {
					uint32	t;
					uint32	seq;
					seq = test_hdr->seq;

#ifdef NFP_TXRX							
					if((isTXRX()) && (test_hdr->plane == EEE_USER_DATA))
							dvt_rcvd_txrx_front_pkt(edesc);
					else 
#endif						
					{
#ifndef NFP_FP							
					if (seq != next_exp_seq_num[src_slot][src_cpu][src_plane]) {
//t = *(volatile uint32 *)0xA0000090;
						printf("%s test error, expected seq %d, got %d, slot %d, cpu %d\n",
							(src_plane == EEE_MGMT) ? "MGMT" :
							((src_plane == EEE_CONTROL) ? "CONTROL" : "DATA"),
							next_exp_seq_num[src_slot][src_cpu][src_plane], seq,
							src_slot, src_cpu);

						if(sentFromSSC[nfxMyCPUId][src_plane]) {
							/* Since pkt is missing we need to create another one
							 * to fill missing gap and proceed with test
							 */ 	
							dvtRemake_pkt(src_slot, src_cpu, 
										  next_exp_seq_num[src_slot][src_cpu][src_plane],
										  src_plane, test_hdr->spin);
						}	
						next_exp_seq_num[src_slot][src_cpu][src_plane] = seq;
#ifdef USE_SMP						
						ATOMIC_INC_INT(&eee.test_dvt_rcv_err[src_plane], 1);
#else						
						eee.test_dvt_rcv_err[src_plane]++;
#endif						
printf("got test data pkt:\n");
printf("edesc = 0x%08x  0x%08x  0x%08x  %p\n", edesc->hdr.control, edesc->hdr.dest_port, edesc->hdr.src_port, (void *)((uintptr_t)edesc->hdr.attr));
printf("      = 0x%04x%04x  %p  0x%04x%04x  %p\n", edesc->bd[0].control, edesc->bd[0].len, edesc->bd[0].buf, edesc->bd[1].control, edesc->bd[1].len, edesc->bd[1].buf);
printf("test_pkt =  cmd       = 0x%08x\n", test_hdr->cmd);
printf("            seq       = 0x%08x\n", test_hdr->seq);
printf("            size      = 0x%08x\n", test_hdr->size);
printf("            orig_slot = 0x%08x\n", test_hdr->orig_slot);
printf("            orig_cpu  = 0x%08x\n", test_hdr->orig_cpu);
printf("            plane     = 0x%08x\n", test_hdr->plane);
/* DO NOT halt test if started from SSC */
if(!sentFromSSC[nfxMyCPUId][src_plane])
	eee_testDVTEnd(src_plane);

					}
#endif					
					next_exp_seq_num[src_slot][src_cpu][src_plane] = 
							(next_exp_seq_num[src_slot][src_cpu][src_plane] + 1) %
														EEE_TEST_NUM_PKT_SPEC;
					t = edesc->hdr.dest_port;
					edesc->hdr.dest_port = edesc->hdr.src_port;
					edesc->hdr.src_port = t;
						if ((rc = eee_forwardPacket(edesc)) !=
														EEE_FORWARDED_PACKET)
#ifdef USE_SMP						
							ATOMIC_INC_INT(&eee.test_dvt_send_err[src_plane], 1);
#else						
							eee.test_dvt_send_err[src_plane]++;
#endif							
						else
#ifdef USE_SMP						
							ATOMIC_INC_INT(&eee.test_dvt_my_sent[src_slot][src_cpu][src_plane], 1);
#else						
							eee.test_dvt_my_sent[src_slot][src_cpu][src_plane]++;
#endif							
					}	
				}	
				else
					eee_freePacket(edesc);
			}
			else {
#ifdef USE_SMP						
				ATOMIC_INC_INT(&eee.test_dvt_other_rcvd[test_hdr->orig_slot][test_hdr->orig_cpu][test_hdr->plane], 1);
#else						
				eee.test_dvt_other_rcvd[test_hdr->orig_slot]
										[test_hdr->orig_cpu][test_hdr->plane]++;
#endif				
				if (eee.test_dvt_running[test_hdr->plane])
			   	{
					uint32	t;
					t = edesc->hdr.dest_port;
					edesc->hdr.dest_port = edesc->hdr.src_port;
					edesc->hdr.src_port = t;
					if (test_hdr->spin) {
						if ((rc = eee_forwardPacket(edesc)) !=
															EEE_FORWARDED_PACKET)
#ifdef USE_SMP						
							ATOMIC_INC_INT(&eee.test_dvt_send_err[src_plane], 1);
#else						
							eee.test_dvt_send_err[src_plane]++;
#endif							
						else
#ifdef USE_SMP						
							ATOMIC_INC_INT(&eee.test_dvt_other_rcvd[test_hdr->orig_slot][test_hdr->orig_cpu][test_hdr->plane], 1);
#else						
							eee.test_dvt_other_sent[test_hdr->orig_slot]
										[test_hdr->orig_cpu][test_hdr->plane]++;
#endif							
					}
					else
						eee_freePacket(edesc);
				}
				else
					eee_freePacket(edesc);
			}
			break;

		case EEE_MGMT_STATS_REQUEST:
#ifdef SSC_MGMT			
			/* Do not return stats for SSC_MGMT */
			eee_freePacket(edesc);
			return;
#endif			
			src_slot = EEE_GET_SLOT_NUM(edesc->hdr.src_port);
			src_cpu = EEE_GET_DEST_CPU(edesc->hdr.src_port);
			src_plane = EEE_GET_DTYPE(edesc->hdr.src_port);
			incoming_appid = EEE_GET_APP_ID(edesc->hdr.src_port);
			eee_reinit_validity_table();

			for(i=0;i<3;i++) {
				for (j=0; j<NFX_NUM_SLOTS_PER_CHASSIS; j++) {
					for (k=0; k<NFX_MAX_CPUS_PER_SLOT; k++) {
						if (eee.test_dvt_valid[j][k][i] > 0) {
							if((isTXRX()) || (isFP()) )	
							{
								if (eee.test_dvt_running[i]) {
									dvtstats.packets_sent[i] += 
										eee.test_dvt_my_sent[j][k][i];
									dvtstats.bytes_sent[i] += 
										(eee_test_spec[i][eee_test_spec_index[i]].num_bytes) * (eee.test_dvt_my_sent[j][k][i]);
									dvtstats.packets_received[i] += 
										eee.test_dvt_my_rcvd[j][k][i];
									dvtstats.bytes_received[i] += 
										(eee_test_spec[i][eee_test_spec_index[i]].num_bytes) * (eee.test_dvt_my_rcvd[j][k][i]);
								}
							}
							else if(isFC()) {	
								if (eee.test_dvt_running[i]) {
									dvtstats.packets_sent[i] += 
										 eee.test_dvt_other_sent[j][k][i];
									dvtstats.bytes_sent[i] += 
										(eee_test_spec[i][eee_test_spec_index[i]].num_bytes) * (eee.test_dvt_other_sent[j][k][i]);
									dvtstats.packets_received[i] += 
										eee.test_dvt_other_rcvd[j][k][i];
									dvtstats.bytes_received[i] += 
										(eee_test_spec[i][eee_test_spec_index[i]].num_bytes) * (eee.test_dvt_other_rcvd[j][k][i]);
								}	
							}		
							dvtstats.packet_send_errors[i] = eee.test_dvt_send_err[i];
							dvtstats.seq_errors[i] = eee.test_dvt_rcv_err[i];
						}
					}
				}	
			}
			/* For front plane pkts - since DATA is not valid in the validity table */ 
			if((isTXRX()) && (nfxMyCPUId == NFX_NFP_TXRX1)) { 
				dvtstats.packets_sent[EEE_USER_DATA] =
						eee.test_dvt_my_sent[NFX_SLOTTYPE_NFP][NFX_NFP_TXRX1][EEE_USER_DATA];
				dvtstats.bytes_sent[EEE_USER_DATA] = 
						pktSize * eee.test_dvt_my_sent[NFX_SLOTTYPE_NFP][NFX_NFP_TXRX1][EEE_USER_DATA];
				dvtstats.packets_received[EEE_USER_DATA] = 
						eee.test_dvt_my_rcvd[NFX_SLOTTYPE_NFP][NFX_NFP_TXRX1][EEE_USER_DATA];
				dvtstats.bytes_received[EEE_USER_DATA] = 
						pktSize * eee.test_dvt_my_rcvd[NFX_SLOTTYPE_NFP][NFX_NFP_TXRX1][EEE_USER_DATA];
				dvtstats.packet_send_errors[EEE_USER_DATA] = 
						eee.test_dvt_send_err[EEE_USER_DATA];
				dvtstats.seq_errors[EEE_USER_DATA] = 
									eee.test_dvt_rcv_err[EEE_USER_DATA];
			}	
			if(isFC()) {
				dvtstats.packets_sent[3] = qdvt->qlogic_sent;
				dvtstats.bytes_sent[3] = qdvt->qlogic_bytes_sent;
				dvtstats.packets_received[3] = qdvt->qlogic_rcvd; 
				dvtstats.bytes_received[3] = qdvt->qlogic_bytes_rcvd; 
				dvtstats.packet_send_errors[3] = qdvt->qlogic_dispCnt;
				dvtstats.crc_errors[3] = qdvt->qlogic_crcErr; 
				dvtstats.pktComp_errors[3] = qdvt->qlogic_dataErr; 
				dvtstats.fatal_errors = qdvt->qlogic_FatalErr; 
			}		
			dvtstats.command = EEE_MGMT_STATS_REPORT;
			dvtstats.slot = nfxMySlotId;
			dvtstats.cpu = nfxMyCPUId;
			eee_systemDVTSend(src_slot, src_cpu, (uchar8 *)&dvtstats, sizeof(eee_dvt_test_stats));
			eee_freePacket(edesc);
			//eee_testInit();
			break;

		case EEE_MGMT_STATS_CLEAR:
			memset(&dvtstats,0,sizeof(eee_dvt_test_stats));
			memset(qdvt,0,sizeof(sysdvt_qlogic_t));
			for(i=0;i<4;i++) {
				eee_testDVTClearStats(i);
			}	
			eee_freePacket(edesc);
			break;

		default:
			printf("Invalid EEE test cmd, %d\n", cmd);
	}
}	/* eee_testDVTRcv() */

#endif  /* not PROM */




/*-----------------------------------------------------------------
 * Name:	eee_anpTestRcv()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void
eee_anpTestRcv(
	eee_descPtr_t	edesc
)
{
	uint32	dp, sp, len;
	uchar8	* buf;

	printf("eee_anpTestRcv: edesc = %p\n", edesc);

	sp = edesc->hdr.src_port;
	dp = edesc->hdr.dest_port;
	len = edesc->hdr.control & EDESC_TOTAL_LEN;

	printf("    source port = 0x%x\n", sp);
	printf("    dest port   = 0x%x\n", dp);
	printf("    msg len     = %d bytes\n", len);

	buf = eee_ramAlloc(16);
	memcpy(buf, edesc->bd[1].buf, 16);
	eee_freePacket(edesc);

	memcpy (buf, "Response", 8);
	eee_sendMessage(buf, len, sp, dp);
	eee_ramDealloc(buf);

}	/* eee_anpTestRcv() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32
eee_sendMessageTest(
	uint32	slot,
	uint32	cpu,
	uint32	num_bytes
)
{
	uint32	* buf;
	uint32	i;
	uint32	dp, sp;
	int32	rc;

	dp = PACK_UNICAST_PORT_ADDRESS(slot, cpu, eee.simple_test_app_id,
									EEE_MGMT, NFX_COS_HIGH, 0);
	sp = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId,
									eee.simple_test_app_id,
									EEE_MGMT, NFX_COS_HIGH, 0);
	buf = eee_ramAlloc(num_bytes);
	if (!buf) {
		printf("Failed to allocate buffer\n");
		return -1;
	}
	for (i=0; i<num_bytes/4; i++)
		buf[i] = i;

	rc = eee_sendMessage((uchar8 *)buf, num_bytes, dp, sp);

	eee_ramDealloc(buf);

	return rc;
}

/*-----------------------------------------------------------------
 * Name:	eee_testInit()
 *
 * Description:	Initialize receive functions for test application
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/19/01
 *
 *-----------------------------------------------------------------
 */
void
eee_testInit()
{
	static uint32 testInitDone = 0;
	if(!testInitDone) {
	eee_unregisterReceiveFunc(eee.simple_test_app_id);
	eee_registerReceiveFunc(eee.simple_test_app_id, eee_testSimpleRcv,
													eee_testSimpleRcv);

	eee_unregisterReceiveFunc(eee.dvt_test_app_id);
#ifndef PROM
	eee_registerReceiveFunc(eee.dvt_test_app_id, eee_testDVTRcv, eee_testDVTRcv);
#endif  /* not PROM */

	eee_unregisterReceiveFunc(eee.anp_test_app_id);
	eee_registerReceiveFunc(eee.anp_test_app_id, eee_anpTestRcv,
													eee_anpTestRcv);
	testInitDone = 1;
#ifndef PROM
	eee_systemDVTinit();
#endif
	}	
#ifndef PROM
#if 1
	/*
	 * Which CPUs should receive start/stop/clear messages
	 */
	eee.test_start_valid[0][0][EEE_MGMT] = 0;
	eee.test_start_valid[0][1][EEE_MGMT] = 0;
	eee.test_start_valid[1][0][EEE_MGMT] = 1;
	eee.test_start_valid[1][1][EEE_MGMT] = 0;
	eee.test_start_valid[1][2][EEE_MGMT] = 0;
	eee.test_start_valid[1][3][EEE_MGMT] = 0;
	eee.test_start_valid[2][0][EEE_MGMT] = 1;
	eee.test_start_valid[2][1][EEE_MGMT] = 1;
	eee.test_start_valid[2][2][EEE_MGMT] = 1;
	eee.test_start_valid[2][3][EEE_MGMT] = 1;

	eee.test_start_valid[0][0][EEE_CONTROL] = 0;
	eee.test_start_valid[0][1][EEE_CONTROL] = 0;
	eee.test_start_valid[1][0][EEE_CONTROL] = 0;
	eee.test_start_valid[1][1][EEE_CONTROL] = 0;
	eee.test_start_valid[1][2][EEE_CONTROL] = 0;
	eee.test_start_valid[1][3][EEE_CONTROL] = 1;
	eee.test_start_valid[2][0][EEE_CONTROL] = 1;
	eee.test_start_valid[2][1][EEE_CONTROL] = 1;
	eee.test_start_valid[2][2][EEE_CONTROL] = 1;
	eee.test_start_valid[2][3][EEE_CONTROL] = 1;

	eee.test_start_valid[0][0][EEE_USER_DATA] = 0;
	eee.test_start_valid[0][1][EEE_USER_DATA] = 0;
	eee.test_start_valid[1][0][EEE_USER_DATA] = 0;
	eee.test_start_valid[1][1][EEE_USER_DATA] = 0;
	eee.test_start_valid[1][2][EEE_USER_DATA] = 0;
	eee.test_start_valid[1][3][EEE_USER_DATA] = 1;
	eee.test_start_valid[2][0][EEE_USER_DATA] = 1;
	eee.test_start_valid[2][1][EEE_USER_DATA] = 1;
	eee.test_start_valid[2][2][EEE_USER_DATA] = 1;
	eee.test_start_valid[2][3][EEE_USER_DATA] = 1;

	/*
	 * Which CPUs should receive data packets
	 */
	eee.test_dvt_valid[0][0][EEE_MGMT] = 0;
	eee.test_dvt_valid[0][1][EEE_MGMT] = 0;
	eee.test_dvt_valid[1][0][EEE_MGMT] = 1;
	eee.test_dvt_valid[1][1][EEE_MGMT] = 0;
	eee.test_dvt_valid[1][2][EEE_MGMT] = 0;
	eee.test_dvt_valid[1][3][EEE_MGMT] = 0;
	eee.test_dvt_valid[2][0][EEE_MGMT] = 1;
	eee.test_dvt_valid[2][1][EEE_MGMT] = 1;
	eee.test_dvt_valid[2][2][EEE_MGMT] = 1;
	eee.test_dvt_valid[2][3][EEE_MGMT] = 1;

	if (isFP()) {
		eee.test_dvt_valid[0][0][EEE_CONTROL] = 0;
		eee.test_dvt_valid[0][1][EEE_CONTROL] = 0;
		eee.test_dvt_valid[1][0][EEE_CONTROL] = 0;
		eee.test_dvt_valid[1][1][EEE_CONTROL] = 0;
		eee.test_dvt_valid[1][2][EEE_CONTROL] = 0;
		eee.test_dvt_valid[1][3][EEE_CONTROL] = 1;
		eee.test_dvt_valid[2][0][EEE_CONTROL] = 1;
		eee.test_dvt_valid[2][1][EEE_CONTROL] = 0;
		eee.test_dvt_valid[2][2][EEE_CONTROL] = 0;
		eee.test_dvt_valid[2][3][EEE_CONTROL] = 1;

		eee.test_dvt_valid[0][0][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[0][1][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[1][0][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[1][1][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[1][2][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[1][3][EEE_USER_DATA] = 1;
		eee.test_dvt_valid[2][0][EEE_USER_DATA] = 1;
		eee.test_dvt_valid[2][1][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[2][2][EEE_USER_DATA] = 0;
		eee.test_dvt_valid[2][3][EEE_USER_DATA] = 1;
	}
	else if (isFC()) {
		if (nfxMyCPUId == NFX_FCNIM_CPUID3) {
			eee.test_dvt_valid[0][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[0][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][2][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][3][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][0][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][2][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][3][EEE_CONTROL] = 0;

			eee.test_dvt_valid[0][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[0][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][2][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][3][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][0][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][2][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][3][EEE_USER_DATA] = 0;
		}
		else if (nfxMyCPUId == NFX_FCNIM_CPUID2) {
			eee.test_dvt_valid[0][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[0][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][2][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][3][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][1][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][2][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][3][EEE_CONTROL] = 0;

			eee.test_dvt_valid[0][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[0][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][2][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][3][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][1][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][2][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][3][EEE_USER_DATA] = 0;
		}
		else if (nfxMyCPUId == NFX_FCNIM_CPUID1) {
			eee.test_dvt_valid[0][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[0][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][2][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][3][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][1][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][2][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][3][EEE_CONTROL] = 0;

			eee.test_dvt_valid[0][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[0][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][2][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][3][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][1][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][2][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][3][EEE_USER_DATA] = 0;
		}
		else if (nfxMyCPUId == NFX_FCNIM_CPUID0) {
			eee.test_dvt_valid[0][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[0][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][2][EEE_CONTROL] = 0;
			eee.test_dvt_valid[1][3][EEE_CONTROL] = 1;
			eee.test_dvt_valid[2][0][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][1][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][2][EEE_CONTROL] = 0;
			eee.test_dvt_valid[2][3][EEE_CONTROL] = 1;

			eee.test_dvt_valid[0][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[0][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][2][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[1][3][EEE_USER_DATA] = 1;
			eee.test_dvt_valid[2][0][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][1][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][2][EEE_USER_DATA] = 0;
			eee.test_dvt_valid[2][3][EEE_USER_DATA] = 1;
		}
	}
#else
	eee.test_dvt_valid[0][0][EEE_MGMT] = 0;
	eee.test_dvt_valid[0][1][EEE_MGMT] = 0;
	eee.test_dvt_valid[1][0][EEE_MGMT] = 1;
	eee.test_dvt_valid[1][1][EEE_MGMT] = 0;
	eee.test_dvt_valid[1][2][EEE_MGMT] = 0;
	eee.test_dvt_valid[1][3][EEE_MGMT] = 0;
	eee.test_dvt_valid[2][0][EEE_MGMT] = 1;
	eee.test_dvt_valid[2][1][EEE_MGMT] = 1;
	eee.test_dvt_valid[2][2][EEE_MGMT] = 1;
	eee.test_dvt_valid[2][3][EEE_MGMT] = 1;

	eee.test_dvt_valid[0][0][EEE_CONTROL] = 0;
	eee.test_dvt_valid[0][1][EEE_CONTROL] = 0;
	eee.test_dvt_valid[1][0][EEE_CONTROL] = 0;
	eee.test_dvt_valid[1][1][EEE_CONTROL] = 0;
	eee.test_dvt_valid[1][2][EEE_CONTROL] = 0;
	eee.test_dvt_valid[1][3][EEE_CONTROL] = 1;
	eee.test_dvt_valid[2][0][EEE_CONTROL] = 1;
	eee.test_dvt_valid[2][1][EEE_CONTROL] = 1;
	eee.test_dvt_valid[2][2][EEE_CONTROL] = 1;
	eee.test_dvt_valid[2][3][EEE_CONTROL] = 1;

	eee.test_dvt_valid[0][0][EEE_USER_DATA] = 0;
	eee.test_dvt_valid[0][1][EEE_USER_DATA] = 0;
	eee.test_dvt_valid[1][0][EEE_USER_DATA] = 0;
	eee.test_dvt_valid[1][1][EEE_USER_DATA] = 0;
	eee.test_dvt_valid[1][2][EEE_USER_DATA] = 0;
	eee.test_dvt_valid[1][3][EEE_USER_DATA] = 1;
	eee.test_dvt_valid[2][0][EEE_USER_DATA] = 1;
	eee.test_dvt_valid[2][1][EEE_USER_DATA] = 1;
	eee.test_dvt_valid[2][2][EEE_USER_DATA] = 1;
	eee.test_dvt_valid[2][3][EEE_USER_DATA] = 1;
#endif

#endif  /* not PROM */
}	/* eee_testSimple() */

#ifndef PROM

/*-----------------------------------------------------------------
 * Name:	eee_testSysDVTStart(mode)
 *
 * Description:	Start system DVT test 
 *
 * Created by: lipika	
 *
 * Date Created:	6/01/02
 *
 *-----------------------------------------------------------------
 */
int32
eee_testSysDVTStart(
	uint32	size,
	uint32	plane,
	uint32	spin
)
{
	uint32	slot, cpu, i;
	int32	rc;
	eee_descPtr_t	edesc;
	if (!eee.test_dvt_valid[nfxMySlotId][nfxMyCPUId][plane]) {
		printf("%s DVT test not valid on this CPU\n",
						(plane == EEE_MGMT) ? "MGMT" :
						((plane == EEE_CONTROL) ? "CONTROL" : "DATA"));
		return NFX_ERR;
	}

	if (eee.test_dvt_running[plane]) {
		printf("EEE test already running\n");
		return NFX_OK;
	}

	/* We allow the second core to start test as the msg 
	 * from SSC can go to either core
	 */ 
#ifdef NFP_FP
	/* Do not start test from the second core of the FP */
//	if(nfxMyCPUId == NFX_NFP_FP2)
//		return NFX_ERR;
#endif	
	/*
	 * Send num_pkts to each slot/cpu
	 */

	for (slot=0; slot<NFX_NUM_SLOTS_PER_CHASSIS; slot++) {
		for (cpu=0; cpu<NFX_MAX_CPUS_PER_SLOT; cpu++) {

			/* Allowed to create test pkt, and have NOT created pkt on this 
			 * cpu for this data type
			 */ 	
//printf("test_dvt_valid[%d][%d][%d] %d\n",slot, cpu,plane,eee.test_dvt_valid[slot][cpu][plane]);				
			if (eee.test_dvt_valid[slot][cpu][plane] > 0) 
			{
				if ((slot == nfxMySlotId) && (cpu == nfxMyCPUId))
					continue;
				if((isTXRX()) || (isFP()))
			   	{
						next_exp_seq_num[slot][cpu][plane] = 0;
						/* Create 8 packets to be spun around */
						for (i=0; i<EEE_TEST_NUM_PKT_SPEC; i++) {
							edesc = eee_testCreateDVTTestPkt(slot, cpu, plane, spin, 0xFFFF);
							if (edesc) {
								printf("slot %d cpu %d plane %d Created test pkt,\n",slot,cpu,plane);
								if ((rc = eee_forwardPacket(edesc)) !=
																EEE_FORWARDED_PACKET)
									eee.test_dvt_send_err[plane]++;
								else
									eee.test_dvt_my_sent[slot][cpu][plane]++;
							}
							else
								eee.test_dvt_send_err[plane]++;
						}
				}		
			}
		}
	}

#ifdef USE_SMP						
	ATOMIC_INC_INT(&eee.test_dvt_running[plane], 1);
#else						
	eee.test_dvt_running[plane] = 1;
#endif	
	return NFX_OK;

}	/* eee_testSysDVTStart() */

#endif  /* not PROM */

