// vi:ts=4:sw=4
/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : el-tcp-client.c
 *
 * Description  : Eagle Link (EL) on NFX.
 *
 * Eagle Link is the interface between Eagle Vision and the NFX.
 * In a first instance, the interface will be TCP/IP.
 * Eagle Vision will be the TCP server, whereas NFXs will act as TCP client.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>

#include "nfx-incl.h"
#include "exports/anpssc-api.h"
#include "exports/ea-api.h"

#include "el-tcp-common.h"
#include "el-api.h"

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

static void elc_terminate(int sig);
static void el_startCommunication(void);
static int el_getConfig(char *option);
static int el_updateConfig(opt_entry_t *entry);
static void elc_sendLoginReq(void);
static int32 el_connectToAv(void);
static void handleElMsg(el_msg_t *elMsg, char *senderName);
static void el_startCommunication(void);

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* 
 * EV's file descriptor.
 * TCP socket descriptor.
 */
static int appFd;
static int tcpFd;

static struct timeval waitRegmTo = {10, 0};
static struct timeval *selectToPtr;

el_config_t		elConfigBlk;
int32 			evLoginDone = 0;
int32   		evConnected = 0;


/* 
 * Basic message structure.
 */
typedef struct
{
    char    buf[4096];
	uint32  len;
} el_msgBuf_t;

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */


/*------------------------------------------------------------------------------
 * Name         : el_getConfig()
 *
 * Description  : Get configuration from registry manager.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 
 *
 *------------------------------------------------------------------------------
 */

static int el_getConfig(char * option)
{
    opt_msgType_t   reqType = OPT_GET_OPTION_REQ;
    el_msgBuf_t     req;
    el_msgBuf_t     rsp;
    opt_msgHdr_t *  hdr;
    void *          bufp = NULL;
    int             nbEntry = 0;
    int             i;

    /* Fill -a %sthe request.
     */
    bufp = &req.buf;
    bzero(bufp,
          (sizeof(opt_msgHdr_t) +
           sizeof(opt_getOptionReq_t)));
    
    memcpy(bufp, (void *) &reqType, sizeof(opt_msgType_t));
    bufp += sizeof(opt_msgType_t);
    // userInfo is not used
    bufp += sizeof(opt_userInfo_t);
    memcpy(bufp, (void *) option, sizeof(opt_option_t));
    bufp += sizeof(opt_option_t);

    req.len = bufp - (void *) req.buf;

    /* Send the request to the registry manager located on the SSC.
     */
    rsp.len = sendAgileMsg(-1,
                           req.buf, req.len,
                           "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitRegmTo,
                           rsp.buf, sizeof(rsp.buf));
    
    /* Receive response from the registry manager.
     */
    if (rsp.len == -1)
    {
		perror("sendAgileMsg failed");
		return(NFX_ERR);
    }

    /* Read the message header and check if it's expected response.
     */
    bufp = &rsp.buf;
    hdr = (opt_msgHdr_t *) rsp.buf;
    bufp += sizeof(opt_msgHdr_t);
    if (hdr->msgType != OPT_GET_OPTION_RSP)
    {
        printf("Message received from REGM is not the expected one!\n");
		return(NFX_ERR);
    }
    
    /* Read the response and print its content.
     * XXX
     */
    memcpy((void *) &nbEntry, bufp, sizeof(int));
    bufp += sizeof(int);

    for (i = 0; i < nbEntry; i++)
    {
        opt_entry_t * entry;

        entry = (opt_entry_t *) bufp;
        bufp += sizeof(opt_entry_t);

        el_updateConfig(entry);
    } // for
    
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : el_updateConfig()
 *
 * Description  : Update local configuration with info rcvd from registry mgr.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 
 *
 *------------------------------------------------------------------------------
 */

static int el_updateConfig(opt_entry_t * entry)
{
    if (!strcmp(entry->option, OPTIONS_AV_LOGIN))
    {
#ifdef DONT_USE_NODE_NAME
		printf("OPTIONS_AV_LOGIN is %s\n", entry->value);
		strcpy(elConfigBlk.avUserid, entry->value);
#endif /* DONT_USE_NODE_NAME */
    }
    else if (!strcmp(entry->option, OPTIONS_AV_PASSWORD))
    {
		printf("OPTIONS_AV_PASSWORD is %s\n", entry->value);
		strcpy(elConfigBlk.avPassword, entry->value);
    }
    else if (!strcmp(entry->option, OPTIONS_AV_IP))
	{
		printf("OPTIONS_AV_IP is %s\n", entry->value);
		elConfigBlk.avIp = inet_addr(entry->value);
	}
    else
    {
        printf("Invalid ELOG option requested\n");
		return NFX_ERR;
    }

    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : elc_sendLoginReq()
 *
 * Description  : Send login request to EV
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void elc_sendLoginReq(void)
{
	int32 rc;
	loginReq_t	loginReq;

	if (!evConnected)
		return;

    bzero((void *) &loginReq, sizeof(loginReq_t));

	loginReq.hdr.msgType = EA_LOGIN_REQ;
	loginReq.hdr.userInfo = getpid();

	strcpy(loginReq.hdr.node, elConfigBlk.nodeName);
	strcpy(loginReq.clusterName, elConfigBlk.clusterName);
	strcpy(loginReq.name, elConfigBlk.avUserid);
	strcpy(loginReq.password, elConfigBlk.avPassword);
	loginReq.isPCC = elConfigBlk.isPCC;
	
	printf("EL login info Node %s Cluster %s Login %s Passwd %s Type %s\n",
			loginReq.hdr.node, loginReq.clusterName, loginReq.name,
			loginReq.password, loginReq.isPCC ? "PCC":"Non-PCC");

    /* 
	 * Forward the buffer to EV.
     */
    rc = sendto (tcpFd, (void *) &loginReq, sizeof(loginReq_t), 0, NULL, 0);
    EL_ASSERT(rc == sizeof(loginReq_t));
}

/*------------------------------------------------------------------------------
 * Name         : elc_terminate()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void elc_terminate(int sig)
{
    /* 
	 * Unregister EV.
     */
    unRegAgileApp(appFd);

    /* 
	 * Close TCP socket.
     */
    close(tcpFd);
    
    exit(0);
}

/*------------------------------------------------------------------------------
 * Name         : el_connectToAv()
 *
 * Description  : Connect to AV
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int32 el_connectToAv(void)
{
    int                 rc;
    struct sockaddr_in  serv_addr;
	struct in_addr      ip_addr;

	if (elConfigBlk.avIp == 0)
		return NFX_ERR;

    /* 
	 * Fill in the structure "serv_addr" with the address of the
     * server that we want to connect with.
     */
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = elConfigBlk.avIp;
    serv_addr.sin_port = htons(SERV_TCP_PORT);
  
    /* 
	 * Connect to the server (EV).
     */
    rc = connect(tcpFd,
                 (struct sockaddr *) &serv_addr,
                 sizeof(serv_addr));
	if (rc < 0)
		return NFX_ERR;

	evConnected = 1;

	memset((char *)&ip_addr, 0, sizeof(struct in_addr));
	ip_addr.s_addr = elConfigBlk.avIp;

	printf("Connected with %s at port %d\n", inet_ntoa(ip_addr), 
						SERV_TCP_PORT);

	return NFX_OK;
}


void handleElMsg(el_msg_t *elMsg, char* senderName)
{
    /* 
	 * Process the message according to its type.
     */
	switch (elMsg->msgType)
    {
        case EL_CONFIG_NTFY:
        {
			/*
			 * Registry manager handing config update
			 * local config
			 */
            //printf("entry = %s\n", elMsg->entry.option);
            //printf("value = %s\n", elMsg->entry.value);
            el_updateConfig((opt_entry_t*) &(elMsg->entry));
			kill(getpid(), SIGTERM);
            break;
        }
            
		case EL_CONFIG_SET_REQ:
		{
			/*
			 * CLI change on config update local config
			 * and set option correctly in Registry Manager
			 */
			el_updateConfig((opt_entry_t*) &(elMsg->entry));

			/*
			 * Update config in options file with registry manager
			 */
			setRegistryOption(elMsg->entry.option, elMsg->entry.value);
			kill(getpid(), SIGTERM);
			break;
		}

		case EL_CONFIG_VIEW_REQ:
		{
			/*
			 * Send back the config block
			 */

			int32 rc = sendAgileMsg(-1, &elConfigBlk, sizeof(el_config_t),
						senderName, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						NULL, NULL, 0);

			if (rc != NFX_OK)
				printf("Could not return EL config\n");

			break;
		}

        default:
			printf("handleElMsg: unknown message %d\n", elMsg->msgType);

    } // switch
}

/*------------------------------------------------------------------------------
 * Name         : el_startCommunication()
 *
 * Description  :
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void el_startCommunication(void)
{
    int     maxFd;    // maximum file descriptor (select optimization)
    int     stdineof;   // boolean value indicating whether an EOF has occured
    fd_set  rset1, rset2;   // set of file descriptors we're interested in
    char    rcvBuf[MAXBUF];   // data buffers
    int     rcvBufLen;
    int     rc;

	struct timeval selectTo = {5, 0};  // 5 seconds should be enough

	/*
	 * Initialize the timeout pointer 
	 */
	selectToPtr = &selectTo;

    /* 
	 * Initializations
     */
    stdineof = 0;                       // no EOF seen initially
    FD_ZERO(&rset1);                     // reset rset membership
   	FD_SET(tcpFd, &rset1);          		// add tcpFd to the read set
    FD_SET(appFd, &rset1);               // add appFd to the read set
    FD_ZERO(&rset2);                     // reset rset membership
    FD_SET(appFd, &rset2);               // add appFd to the read set
   	maxFd = MAX(tcpFd, appFd) + 1;

    /* 
	 * Loop forever (exit through return statement).
     */
    while(TRUE)
    {
        fd_set rset_tmp;

		if (!evConnected && elConfigBlk.avIp)
		{
			el_connectToAv();
		}

		if (evConnected)
			rset_tmp = rset1;  /* copy flags (structure copy) */
		else
			rset_tmp = rset2;  /* copy flags (structure copy) */

		if (evConnected && !evLoginDone)
		{
			/*
			 * Send login request
			 */
			elc_sendLoginReq();
		}

        /* 
		 * Use valid timeout till we are not logged in to AV. After that
		 * we can block on select and wait for one of the file descriptors 
		 * in rset to be ready for reading. Since we have only two file 
		 * descriptors, we don't care for the return value (we check for 
		 * either 1 or 2 manually).
         */

       	if (select(maxFd, &rset_tmp, NULL, NULL, selectToPtr) < 1)
			continue;

        /* 
		 * Is there data in the TCP socket ?
         */
        if (evConnected && FD_ISSET(tcpFd, &rset_tmp))
        {
			ea_msgHdr_t *eaMsgHdr;
			uint32 readLen1 = 0;
			uint32 readLen2 = 0;


            /* 
			 * Read received message.
             */
            bzero((void *) rcvBuf, sizeof(rcvBuf));

			/*
			 * First read the header only
			 */
			readLen1 = sizeof(ea_msgHdr_t);
            rcvBufLen = recvfrom(tcpFd, (void *) rcvBuf, readLen1,
                                 (MSG_WAITALL | MSG_PEEK), NULL, NULL);
            EL_ASSERT(rcvBufLen == readLen1);      

			/*
			 * Based on msgLen specified in header read the remaining
			 * packet
			 */
			eaMsgHdr = (ea_msgHdr_t *) rcvBuf;
			readLen2 = eaMsgHdr->msgLen;

            rcvBufLen = recvfrom(tcpFd, (void *) rcvBuf , 
									readLen2, MSG_WAITALL, NULL, NULL);
            EL_ASSERT(rcvBufLen == readLen2);

			/*
			 * Do not process any message from AV unless we are
			 * authenticated and connected
			 */
			if (evLoginDone == 0)
			{
				/*
			 	 * Keep looking for EA_LOGIN_RSP from AV
			 	 */
				loginRsp_t *loginRsp = (loginRsp_t *) rcvBuf;

				if ((loginRsp->hdr.msgType == EA_LOGIN_RSP) &&
						(loginRsp->result == NFX_OK))
				{
					evLoginDone = 1;
					selectToPtr = NULL;
				}
				else
				{
					printf("EL not logged in to AV: %s\n",
							loginRsp->output);
				}
				continue;
			}

			// pad1 is used as status
			// since this is from AV to EA, pad status should be
			// 0 (NFX_OK) to indicate received OK

			eaMsgHdr->pad1 = NFX_OK;
           
            /* 
			 * Forward the buffer to EA.
             */
            rc = sendAgileMsg(appFd, rcvBuf, rcvBufLen,
                              "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                              NULL, NULL, 0);
            EL_ASSERT(rc == NFX_OK);

            printf("Message %d (len %d) forwarded from AV --> EA/NFX\n", 
									*((uint32 *) rcvBuf), rcvBufLen);
        }

        /* 
		 * Is there data in the internal socket ?
         */
        if (FD_ISSET(appFd, &rset_tmp))
        {
            /* 
			 * Read received message.
             */
			uint32 *msgType;
			char    senderName[256];
			uint32  cpu;
			uint32  slot;

            bzero((void *) rcvBuf, sizeof(rcvBuf));

			rcvBufLen = recvAgileMsg(appFd, (void *) rcvBuf, MAXBUF,
							 NULL, senderName, &cpu, &slot, NULL);

            EL_ASSERT(rcvBufLen > 0);      

			msgType = (uint32 *) rcvBuf;

            /* 
			 * Forward the buffer to EV.
             */
			if ((*msgType >= EL_MSG_BASE) && (*msgType < EV_MSG_BASE))
			{
				handleElMsg((el_msg_t *) rcvBuf, senderName);
			}
			else if (evLoginDone)
			{
            	rc = sendto (tcpFd, (void *) rcvBuf, rcvBufLen, 0, NULL, 0);
            	EL_ASSERT(rc == rcvBufLen);

            	printf("Message %d forwarded from EA/NFX --> AV\n", *msgType);
			}
			else
			{
				ea_msgHdr_t *eaMsgHdr = (ea_msgHdr_t *) rcvBuf;


				// For EA_GROW_FSYS_REQ:
				// Since we can't talk to AV link, we need to return the
				// whole packet back and indicates status code in "pad1"
				// field of EA msg header
				if (eaMsgHdr->msgType == EA_GROW_FSYS_REQ)
				{
					eaMsgHdr->pad1 = evConnected ? EL_NOTLOGIN : EL_NOTCONNECTED;

            		sendAgileMsg(appFd, rcvBuf, rcvBufLen,
                              "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                              NULL, NULL, 0);
				}
			}

			// SY: Talked to Isabelle and Chuan and decided to remove this line.
            // don't send too fast the message to EV, otherwise they are lost!!!
            // sleep(1);

         }
    } // while (TRUE)

    return;
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int main(int argc, char **argv)
{
    /* 
	 * Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, elc_terminate);
    signal(SIGINT, elc_terminate);

    /* 
	 * Register to receive messages for EV.
     */
	appFd = regAgileApp("ev");
	EL_ASSERT(appFd >= 0);

    /* 
	 * Open a TCP socket (an Internet stream socket).
     */
    tcpFd = socket(AF_INET, SOCK_STREAM, 0);
    EL_ASSERT(tcpFd >= 0);

	/*
	 * Tell mcp we're ready to roll.
	 */
	kill(getppid(), SIGUSR1);

	/*
	 * Initialize config block and read config from reg manager
	 * Get EL config and node and cluster info
	 */
	memset(&elConfigBlk, 0, sizeof(el_config_t));

	gethostname(elConfigBlk.nodeName, LNM_MAX_NODE_NAME);
	sprintf(elConfigBlk.avUserid, "%ld", gethostid());

	getRegistryOption(OPTIONS_CLUSTER_NAME, elConfigBlk.clusterName);
	el_getConfig(OPTIONS_AV_ALL);

    el_startCommunication();

    return(0); // should never exit this way
}

/*----------------------------------------------------------------------------*/
