/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : sim-ev.c
 *
 * Description  : Eagle Vision (EV) simulator.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 23-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>

#include "nfx-incl.h"
#include "exports/ea-api.h"
#include "el-tcp-common.h"
#include "sim-ev.h"

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

extern int tcpFd;

/* The name of this program. */
static char * myName;

/*------------------------------------------------------------------------------
 * Prototype of local functions
 *------------------------------------------------------------------------------
 */

static void
print_usage(void);

/*------------------------------------------------------------------------------
 * Local functions
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : print_usage()
 *
 * Description  : Shows the correct usage of this program.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 23-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
print_usage(void)

{
	fprintf(stderr,"Usage:\n");

	fprintf(stderr," add fsys <name>\n");
	fprintf(stderr," mod fsys <name> <...>\n");
	fprintf(stderr," del fsys <name>\n");
	fprintf(stderr," add fset <name> <path>\n");
	fprintf(stderr," mod fset <name> <path> <...>\n");
	fprintf(stderr," del fset <name> <path>\n");
}

/*------------------------------------------------------------------------------
 * Global functions
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : el_processCli()
 *
 * Description  : Simulate EV's transmission of messages.
 * Process the user input and send a message to NFX/EA.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 23-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void
el_processCli(int argc, char * argv[])

{
	char *      actionStr;
	char *      objectStr;
    char        msgBuf[10000];
    int         msgLen;
    uint32      cookie;
    int         rc;

    /* Initialize my own name.
     * Discard path if any.
     */
	myName = argv[0];
	if (rindex(myName, '/'))
    {
		strcpy(myName, rindex(myName, '/') + 1);
	}
    
    /* Make sure there's at least the action and the object strings.
     * Get the command and the object on which to execute the command.
     */
	if (argc < 3)
    {
		print_usage();
		return;
	}
	actionStr = argv[1];
	objectStr = argv[2];
    
    /* Initialize the cookie.
     */
    cookie = random();
    printf("cookie = %d\n", cookie);
   
    /* Each "command" has it's own action function.
     */
	if (!strcasecmp(actionStr, "add"))
    {
        if (!strcasecmp(objectStr, "fsys"))
        {
            /* Create file system.
             */
            ea_addFsysReq_t * req = (ea_addFsysReq_t *) &(msgBuf);
            
            req->hdr.msgType = EA_ADD_FSYS_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);
            req->volSize = 102396;
            req->maxSize = 0;
            req->minAutoGrow = 0;
            req->highWaterMark = 0;
            req->attr.raidLevel = EVM_RAID_ANY;
            req->attr.minLunSize = EVM_DONT_CARE;
            req->attr.maxLunSize = EVM_DONT_CARE;
            req->attr.num = 1;
            strcpy(req->attr.vendorInfo[0].vendor, "");
            strcpy(req->attr.vendorInfo[0].model, "");
            req->nim = 1;    // N-NIM in simulation

            msgLen = sizeof(ea_addFsysReq_t);
        }
        else if (!strcasecmp(objectStr, "fset"))
        {
            /* Create file set.
             */
            ea_addFsetReq_t * req = (ea_addFsetReq_t *) &(msgBuf);
            
            req->hdr.msgType = EA_ADD_FSET_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsetId.fsysId.fsysName, argv[3]);
            strcpy(req->fsetId.path, argv[4]);
            req->quotaOn = FALSE;

            msgLen = sizeof(ea_addFsetReq_t);
        }
		else
        {
			print_usage();
			return;
		}            
	}
	else if(!strcmp(actionStr, "del"))
    {
        if (!strcasecmp(objectStr, "fsys"))
        {
            /* Delete file system.
             */
            ea_delFsysReq_t * req = (ea_delFsysReq_t *) &(msgBuf);
            
            req->hdr.msgType = EA_DEL_FSYS_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsysId.fsysName, argv[3]);

            msgLen = sizeof(ea_delFsysReq_t);            
        }
        else if (!strcasecmp(objectStr, "fset"))
        {
            /* Delete file set.
             */
            ea_delFsetReq_t * req = (ea_delFsetReq_t *) &(msgBuf);
            
            req->hdr.msgType = EA_DEL_FSET_REQ;
            req->hdr.userInfo = cookie;
            strcpy(req->fsetId.fsysId.fsysName, argv[3]);
            strcpy(req->fsetId.path, argv[4]);

            msgLen = sizeof(ea_delFsetReq_t);
        }
		else
        {
			print_usage();
			return;
		}
	}
	else if(!strcmp(actionStr, "mod"))
    {
        if (!strcasecmp(objectStr, "fsys"))
        {
            // Modify file system
            printf("not yet implemented!\n");
            return;
        }
        else if (!strcasecmp(objectStr, "fset"))
        {
            // Modify file set
            printf("not yet implemented!\n");
            return;
        }
		else
        {
			print_usage();
			return;
		}
	}
	else
    {
		fprintf(stderr, "Invalid command: %s\n",actionStr);
		print_usage();
		return;
	}

    /* Send message to EA.
     */
    EL_ASSERT(tcpFd >= 0);
    printf("Sending a message to EA...(%d)\n", tcpFd);
    rc = sendto(tcpFd, (void *) &msgBuf, msgLen, 0, NULL, 0);
    EL_ASSERT(rc == msgLen);
		
	return;
}

/*------------------------------------------------------------------------------
 * Name         : el_processEaMsg()
 *
 * Description  : Simulate EV's reception of messages.
 * Basically, display their content...
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 23-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void
el_processEaMsg(char *  buf,
                int     len) 

{
    ea_msgHdr_t * hdr = (ea_msgHdr_t *) buf;
            
    switch(hdr->msgType)
    {
    case EA_ADD_FSYS_RSP:
    {
        ea_addFsysRsp_t * rsp = (ea_addFsysRsp_t *) buf;
        printf("EA_ADD_FSYS_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_MOD_FSYS_RSP:
    {
        ea_modFsysRsp_t * rsp = (ea_modFsysRsp_t *) buf;
        printf("EA_MOD_FSYS_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_MNT_FSYS_RSP:
    {
        ea_mntFsysRsp_t * rsp = (ea_mntFsysRsp_t *) buf;
        printf("EA_MNT_FSYS_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_UMNT_FSYS_RSP:
    {
        ea_umntFsysRsp_t * rsp = (ea_umntFsysRsp_t *) buf;
        printf("EA_UMNT_FSYS_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_DEL_FSYS_RSP:
    {
        ea_delFsysRsp_t * rsp = (ea_delFsysRsp_t *) buf;
        printf("EA_DEL_FSYS_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_GET_FSYS_RSP:
    {
        ea_getFsysRsp_t * rsp = (ea_getFsysRsp_t *) buf;
        printf("EA_GET_FSYS_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_ADD_FSET_RSP:
    {
        ea_addFsetRsp_t * rsp = (ea_addFsetRsp_t *) buf;
        printf("EA_ADD_FSET_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_MOD_FSET_RSP:
    {
        ea_modFsetRsp_t * rsp = (ea_modFsetRsp_t *) buf;
        printf("EA_MOD_FSET_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_DEL_FSET_RSP:
    {
        ea_delFsetRsp_t * rsp = (ea_delFsetRsp_t *) buf;
        printf("EA_DEL_FSET_RSP (%s)\n",
               (rsp->result == NFX_OK)? "OK":"ERROR");
        break;
    }

    case EA_CRT_VOL_NTFY:
    {
        printf("EA_CRT_VOL_NTFY\n");
        break;
    }
    
    case EA_DEL_VOL_NTFY:
    {
        printf("EA_DEL_VOL_NTFY\n");
        break;
    }

    case EA_MOD_VOL_NTFY:
    {
        printf("EA_MOD_VOL_NTFY\n");
        break;
    }
    
    case EA_OO_VOL_NTFY:
    {
        printf("EA_OO_VOL_NTFY\n");
        {
            ea_ooVolNtfy_t *    ntfy = (ea_ooVolNtfy_t *) buf;
            
            printf("volId = %lld\n", ntfy->volId.whole_thing);
            printf("volOnline = %d\n", ntfy->volOnline);
        }
        break;
    }
    
    case EA_GRW_VOL_NTFY:
    {
        printf("EA_GRW_VOL_NTFY\n");
        break;
    }
    
    case EA_FULL_VOL_NTFY:
    {
        printf("EA_FULL_VOL_NTFY\n");
        break;
    }

    case EA_FAIL_LUN_NTFY:
    {
        printf("EA_FAIL_LUN_NTFY\n");
        break;
    }
    default:
    {
        printf("Unknow message type (%d)\n", hdr->msgType);
        break;
    }
    } // switch
    
    return;
}

/*----------------------------------------------------------------------------*/
