/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Name:  elog-def.h
 *
 * RCS:  $Id: elog-def.h,v 1.8 2002/08/01 17:56:04 sandrineb Exp $
 *
 * Description:   Defines the global types for the elog logging interface
 *                of the Eagle NFX.
 *
 * Created by:	  Jan Vanhoutte
 *                Pankaj Gupta
 *
 * Date Created:  11 sep 2001
 *
 */

#ifndef _elog_def_h_
#define _elog_def_h_

/*
 * priorities/facilities are encoded into a single 32-bit quantity, where the
 * bottom 3 bits are the priority (0-7) and the top 28 bits are the facility
 * (0-big number).  Both the priorities and the facilities map roughly
 * one-to-one to strings in the syslogd(8) source code.  This mapping is
 * included in this file.
 *
 * priorities (these are ordered)
 */
#define LOG_EMERG   0   /* system is unusable */
#define LOG_ALERT   1   /* action must be taken immediately */
#define LOG_CRIT    2   /* critical conditions */
#define LOG_ERR     3   /* error conditions */
#define LOG_WARNING 4   /* warning conditions */
#define LOG_NOTICE  5   /* normal but significant condition */
#define LOG_INFO    6   /* informational */
#define LOG_DEBUG   7   /* debug-level messages */

#define LOG_PRIMASK 0x07    /* mask to extract priority part (internal) */
                /* extract priority */
#define LOG_PRI(p)  ((p) & LOG_PRIMASK)
#define LOG_MAKEPRI(fac, pri)   (((fac) << 3) | (pri))

#ifdef SYSLOG_NAMES
#define INTERNAL_NOPRI  0x10    /* the "no priority" priority */
                /* mark "facility" */
#define INTERNAL_MARK   LOG_MAKEPRI(LOG_NFACILITIES, 0)
typedef struct _code {
    char    *c_name;
    int c_val;
} CODE;

CODE prioritynames[] =
  {
    { "alert", LOG_ALERT },
    { "crit", LOG_CRIT },
    { "debug", LOG_DEBUG },
    { "emerg", LOG_EMERG },
    { "err", LOG_ERR },
    { "error", LOG_ERR },       /* DEPRECATED */
    { "info", LOG_INFO },
    { "none", INTERNAL_NOPRI },     /* INTERNAL */
    { "notice", LOG_NOTICE },
    { "panic", LOG_EMERG },     /* DEPRECATED */
    { "warn", LOG_WARNING },        /* DEPRECATED */
    { "warning", LOG_WARNING },
    { NULL, -1 }
  };
#endif

/* facility codes */
#define LOG_KERN    (0<<3)  /* kernel messages */
#define LOG_USER    (1<<3)  /* random user-level messages */
#define LOG_MAIL    (2<<3)  /* mail system */
#define LOG_DAEMON  (3<<3)  /* system daemons */
#define LOG_AUTH    (4<<3)  /* security/authorization messages */
#define LOG_SYSLOG  (5<<3)  /* messages generated internally by syslogd */
#define LOG_LPR     (6<<3)  /* line printer subsystem */
#define LOG_NEWS    (7<<3)  /* network news subsystem */
#define LOG_UUCP    (8<<3)  /* UUCP subsystem */
#define LOG_CRON    (9<<3)  /* clock daemon */
#define LOG_AUTHPRIV    (10<<3) /* security/authorization messages (private) */
#define LOG_FTP     (11<<3) /* ftp daemon */

    /* other codes through 15 reserved for system use */
#define LOG_LOCAL0  (16<<3) /* reserved for local use */
#define LOG_LOCAL1  (17<<3) /* reserved for local use */
#define LOG_LOCAL2  (18<<3) /* reserved for local use */
#define LOG_LOCAL3  (19<<3) /* reserved for local use */
#define LOG_LOCAL4  (20<<3) /* reserved for local use */
#define LOG_LOCAL5  (21<<3) /* reserved for local use */
#define LOG_LOCAL6  (22<<3) /* reserved for local use */
#define LOG_LOCAL7  (23<<3) /* reserved for local use */

#define LOG_NFACILITIES 24  /* current number of facilities */
#define LOG_FACMASK 0x03f8  /* mask to extract facility part */
                /* facility of pri */
#define LOG_FAC(p)  (((p) & LOG_FACMASK) >> 3)

// We have 7 classes of log messages at our disposal.
// We should probably assign these to specific uses (TBD) and give
// their data declaration some meaningful name.

typedef enum {
    class_1 = LOG_LOCAL0,	// SNMP trap?
    class_2 = LOG_LOCAL1,	// Ethernet Links?
    class_3 = LOG_LOCAL2,	// FC Links?
    class_4 = LOG_LOCAL3,	// Security?
    class_5 = LOG_LOCAL4,	// Device errors?
    class_6 = LOG_LOCAL5,	// The coffee pot's empty?
    class_7 = LOG_LOCAL6	// We're out of Espresso Beans?
} log_class_t;


// There are seven severity levels available. These correspond to the
// syslog defined levels.

typedef enum {
    emerg_s = LOG_EMERG,     // It's reset time. Call the paramedics.
                             // system is unusable
    alert_s = LOG_ALERT,     // It's really busted 
                             // action must be taken immediately
    crit_s = LOG_CRIT,       // Something's wrong and continuing's
                             // not a viable option for me.
    err_s = LOG_ERR,         // Something's wrong but I can press 
                             // in a degraded mode.
    warning_s = LOG_WARNING, // Something's not quite right
                             // 
    notice_s = LOG_NOTICE,   // You really should want to know 
                             // about this.
    info_s = LOG_INFO,       // Someone might find this interesting.
                             // 
    debug_s = LOG_DEBUG      // Debug, unimportant info.
                             // Also used as TRACE
} log_severity_t;


/*------------------------------------------------------------------------------
 * Header of the message.
 *------------------------------------------------------------------------------
 */

#define ELOG_DBG_PRINT(x) //printf
#define ELOG_FCT_PRINT(x) //printf
#define ELOG_ERR_PRINT(x) //printf


/* Basic message structure.
 */
typedef struct
{
    char    buf[4096];
	uint32  len;
} elog_msg_t;

typedef struct
{
    boolean         enable;
	boolean			display;
	log_severity_t  levelDefault;
	uint32			facility;
	uint32			hostIpa;
} elog_config_t;

/* Message type
 */
typedef enum
{
    ELOG_MSG_NOTIFY      = ELOG_MSG_BASE, // Elog message notification
	ELOG_CONFIG_NTFY,						// elog config from Regm
	ELOG_CONFIG_GET_REQ,					// get current config 		
	ELOG_CONFIG_SET_REQ,					// elog config from CLI
	ELOG_CONFIG_VIEW_REQ					// elog view request from CLI
} elog_msgType_t;

/* Header of the message.
 */
typedef struct
{
    elog_msgType_t   msgType;    // the message type SHALL always be first
} elog_msgHdr_t;

/*------------------------------------------------------------------------------
 * The EVENT NTFY message.
 *------------------------------------------------------------------------------
 */

// This looks amazingly like the arguments to elog() because that's where
// the data came from. The names and meaning are the same. This is the
// format of the messages sent to the "elog" application on the SSC.

typedef struct
{
    elog_msgType_t   msgType;    // the message type SHALL always be first
    log_class_t     class;
    log_severity_t  severity;
    uint32          appId;
    uint32          appType;
    uint32          cookie;
    uint32          code;
    char            amsg[1];
} elog_eventNtfy_t;

/*------------------------------------------------------------------------------
 * The CONFIG NTFY message.
 *------------------------------------------------------------------------------
 */

typedef struct
{ 
    elog_msgType_t   msgType;    // the message type SHALL always be first
    struct
    { 
        char        option[128];
        char        value[256];
    } entry;
} elog_configNtfy_t;

#endif 
