/*
 *-----------------------------------------------------------------
 *
 * Name:	esm-api.h
 *
 * RCS:  $Id: esm-api.h,v 1.28 2002/11/05 00:21:49 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:  Definitions for Eagle State Machine scheduler
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _ESM_API_H
#define _ESM_API_H

#ifdef USE_THREAD_STACKS
/*
 * We normally want to define USE_THREAD_STACKS for debug builds.
 * It provides protection for stack overrun by having an unmapped
 * page between stacks.
 *
 * These are also used in sm-sb1250/pte.c for TLB PTE initializations.
 * NUM_STACKS * STACK_SIZE should be <= 4 GB so that it fits in a single region.
 */
#define ESM_NUM_STACKS	2048
#define ESM_STACK_SIZE	(16 * 1024)
#endif

#ifdef USE_BSS_THREAD_STACKS
#ifdef USE_THREAD_STACKS
#error "Cant define USE_BSS_THREAD_STACKS and USE_THREAD_STACKS"
#endif
#ifdef NFP_MOD_FP
#define ESM_NUM_STACKS	2048
#else
#define ESM_NUM_STACKS	512
#endif
#define ESM_STACK_SIZE	(16 * 1024)

#ifdef LANGUAGE_C
extern uchar8	bss_threadstack[ESM_NUM_STACKS][ESM_STACK_SIZE];
#endif
#endif

#ifdef LANGUAGE_C

typedef uint32	esm_handle_t;

extern nfx_spinlock_t   esm_spinlock;

typedef struct esm_param_t esm_param_t;
struct esm_param_t
{
    union
    {
        uint64   dw;
        uint32   w;
        void     *p;
    } u;
};

typedef struct esm_event_t esm_event_t;
struct esm_event_t
{
	int32		    event;
	esm_param_t		p1;
	esm_param_t		p2;
};

/*
 * Event delivered to threads in esm_yield() if their
 * specified timeout has elapsed.
 */
enum {
	ESM_THREAD_THREADLIB_EVENT_RANGE	=	500,
	ESM_THREAD_EVENT_USER_MAX			= NEXT_MSG_BASE +
											ESM_THREAD_THREADLIB_EVENT_RANGE,
    ESM_THREAD_EVENT_TIMEOUT    		= ESM_THREAD_EVENT_USER_MAX
};

/*-----------------------------------------------------------------
 * Name:    esm_getCurrHandle
 *
 * Description: Return the handle for the currently running state
 *              machine.
 * 
 * Created by:  Bob Fozard
 * 
 * Date Created:    3/20/01
 *
 *-----------------------------------------------------------------
 */
esm_handle_t
esm_getCurrHandle();

/*-----------------------------------------------------------------
 * Name:    esm_getCurrState
 *
 * Description: Return the current state for the specified state
 *              machine.
 * 
 * Created by:  Bob Fozard
 * 
 * Date Created:    3/20/01
 *
 *-----------------------------------------------------------------
 */
int32                       /* currentState
                             * NFX_ERR
                             */
esm_getCurrState(
    esm_handle_t   handle
);

/*-----------------------------------------------------------------
 * Name:    esm_getUserInfo
 *
 * Description: Return the the user info for the specified state machine.
 *              Returns NULL if the state machine handle is invalid.
 * 
 * Created by:  Paul Haahr
 * 
 * Date Created:    9/13/01
 *
 *-----------------------------------------------------------------
 */
void *
esm_getUserInfo(
    esm_handle_t   handle
);

/*-----------------------------------------------------------------
 * Name:	esm_deliverEvent()
 *
 * Description:	Called to send an event to a state machine
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK
							 * ESM_ERR_
							 */
esm_deliverEvent(
	esm_handle_t	handle,		/* handle of instance to deliver event to */
	esm_event_t     *event 		/* state machine defined event */
);

/*-----------------------------------------------------------------
 * Name:	esm_deliverHighPriorityEvent()
 *
 * Description:	Called to send an event to a state machine for
 *              high-priority delivery.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK
							 * ESM_ERR_
							 */
esm_deliverHighPriorityEvent(
	esm_handle_t	handle,		/* handle of instance to deliver event to */
	esm_event_t     *event 		/* state machine defined event */
);

/*------------------------------------------------------------------
 *
 *	Structures for state machine operation
 *
 *	Files that are created by the state machine tool need these
 *	structure definitions, therefore these are in the exported API
 *	header file.
 *
 *-----------------------------------------------------------------*/

typedef struct esm_actionFunc
{
	int32		(*func)(	esm_handle_t	handle,
							int32			nextState,
							esm_event_t 	event,
							void			*userInfo);
	int32		nextState;
} esm_actionFunc_t;

typedef struct esm_table
{
	esm_actionFunc_t	***eventVectors;
	char				**xlateState;
	char				**xlateEvent;
	ushort16			numEvents;
	ushort16			numStates;
	ushort16			maxQueueDepth;
    ushort16            canYield;
    uint32              stackSize;
#define ESM_NAME_MAX    32
    char                name[ESM_NAME_MAX];
} esm_table_t;


extern int32
esm_instantiate(
	esm_table_t  			*table,
	esm_actionFunc_t  		*createAction,
	esm_handle_t			*handle,
	void					*userInfo);

extern int32
esm_destroy(
	esm_handle_t		handle);

extern void
esm_yield(
	esm_event_t			*eventp,		// return my waking event info here
	int32				waitSecs);		// warn if I sleep longer than this

extern void
esm_initModule();

extern void
esm_longJump(
    esm_handle_t        handle);

/**
 * Return the next thread id in sequence after the
 * specified one.  If tid is 0, start at the beginning.
 * 
 * A thread id is the handle for the state machine instance.
 */
extern uint32
esm_gdbGetNextThreadId(
    uint32  tid);

/**
 * Return the thread id if the specified thread id is
 * still alive, otherwise return 0.
 *
 * A thread id is the handle for the state machine instance.
 */
extern uint32
esm_gdbIsThreadAlive(
    uint32  tid);

/**
 * Return the registers for the specified thread id.
 *
 * A thread id is the handle for the state machine instance.
 */
extern uint32
esm_gdbGetRegs(
    void        *ptr,
    int32       num,
    uint32      tid);

#endif	/* LANGUAGE_C */

#endif /* _ESM_API_H */
