/*
 *-----------------------------------------------------------------
 *
 * Name:	esm-linux.c
 *
 * RCS:  $Id: esm-linux.c,v 1.11 2002/03/09 02:09:07 rahulk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	esm stuff when compiled with LINUX_TEST
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>

#if defined(LINUX_TEST)

#include <exports/queue-api.h>
#include <exports/esm-api.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#include <exports/debug.h>

#include <sys/errno.h>
#include <sys/time.h>
#include <pthread.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <setjmp.h>

#include "esm.h"

#define ESM_LINUX_POLL_FD_RX    0
#define ESM_LINUX_POLL_FD_TX    1
#define ESM_LINUX_POLL_FD_MAX   2
int esm_linuxPollFd[ESM_LINUX_POLL_FD_MAX];

// TODO - use elog for tracing.
#define ESM_TRACE

/*-----------------------------------------------------------------
 * The following definitions provide for state machine
 * instances that run as a thread (yield capable) on a
 * UNIX/LINUX host using pthreads.
 */
typedef struct esm_threadArg esm_threadArg_t;
struct esm_threadArg
{
    esm_handle_t        handle;
    esm_event_t         event;
    esm_instance_t      *esm;
    esm_actionFunc_t    *func;
    int32               threadNum;
    DLL_ENTRY(esm_threadArg_t)
                        dllEntry;
};

pthread_mutex_t          esm_execMutex;
pthread_cond_t           esm_execCondVar;
DLL_HDR(esm_threadArg_t) esm_threadJobQueueHdr;
int32                    esm_numThreadsWaiting = 0;
int32                    esm_numThreadsCreated = 0;

typedef struct esm_thread
{
    pthread_t           thread;
    pthread_cond_t      condVar;
    esm_threadArg_t     threadArg;
#define ESM_THREAD_BUSY     0
#define ESM_THREAD_AVAIL    1
    int32               state;
	jmp_buf				jmpBuf;
} esm_thread_t;
#define ESM_THREADS_MAX 16
esm_thread_t            esm_thread[ESM_THREADS_MAX];
pthread_key_t           esm_keyThreadNum;

/*-----------------------------------------------------------------
 * Platform specific init.
 */
void
esm_initPlatform()
{
#if !defined(FCNIM)
    int32   rc;

    DLL_INIT(&esm_threadJobQueueHdr);
    rc = pthread_mutex_init(&esm_execMutex, NULL);
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }
    rc = pthread_cond_init(&esm_execCondVar, NULL);
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }
    rc = pthread_key_create(&esm_keyThreadNum, NULL);
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }

    if (socketpair(AF_UNIX, SOCK_DGRAM, 0, esm_linuxPollFd) < 0) {
        perror("esm socketpair");
    }
    if (fcntl(esm_linuxPollFd[ESM_LINUX_POLL_FD_RX], F_SETFL, O_NONBLOCK))
    {
        perror("esm fcntl");
        panic("esm pollFd rx socket fcntl failed");
    }
    if (fcntl(esm_linuxPollFd[ESM_LINUX_POLL_FD_TX], F_SETFL, O_NONBLOCK))
    {
        perror("esm fcntl");
        panic("esm pollFd tx socket fcntl failed");
    }

#ifdef LINUX_POLL_NONBLOCKING_FD
    eee_addPollIndicationFd(esm_linuxPollFd[ESM_LINUX_POLL_FD_RX]);
#endif
#endif

} /* esm_initPlatform */

/*-----------------------------------------------------------------------------
 * esm_threadMain
 *
 * Description:  Yield capable state machines start the thread here.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
static void *
esm_threadMain(void *arg)
{
    esm_actionFunc_t        ***evFunArr;
    esm_actionFunc_t        **evFun;
    esm_actionFunc_t        *fun;
    esm_threadArg_t         *threadArg;
    int32                   threadNum;
    int                     rc;

    threadArg = (esm_threadArg_t *) arg;
    threadNum = threadArg->threadNum;
    rc = pthread_setspecific(esm_keyThreadNum, (void *) threadNum);
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }

    ESM_TRACE("esm proc %ld thread %ld handle 0x%x created =========\n",
          nfxMyCPUId, threadNum, threadArg->handle);

    for (;;)
    {
        NFX_ASSERT(esm_thread[threadNum].state == ESM_THREAD_BUSY);

        memcpy( &esm_thread[threadNum].threadArg,
                threadArg,
                sizeof(esm_threadArg_t));

        evFunArr = threadArg->esm->table->eventVectors;
        evFun = evFunArr[threadArg->event.event];
        fun = evFun[threadArg->esm->currentState];
    
        threadArg->esm->event = threadArg->event;
        threadArg->esm->threadNum = threadNum;
    
        ESM_TRACE("esm proc %ld thread %ld handle 0x%x calling thread func %p\n", 
              nfxMyCPUId, threadNum, threadArg->handle, fun->func);
    
        if (!setjmp(esm_thread[threadNum].jmpBuf))
        {
        	(void) fun->func(threadArg->handle,
                             threadArg->esm->currentState,
                             threadArg->event,
                             threadArg->esm->userInfo);
		}
        
        // If the function returns, we're done.

        ESM_TRACE("esm proc %ld thread %ld handle 0x%x func returned, signalling exec\n", 
              nfxMyCPUId, threadNum, threadArg->handle);

    	// Thread returned, mark it for destruction.  It doesn't
    	// hurt to do this even if its already been done by the client.
    	esm_destroy_esm(threadArg->esm);

        // This works fine for threads that die by returning to
        // the executive, but for threads that are destroyed before
        // returning, we'll leak these threadArgs!!!  TODO - fix this!
        eee_ramDealloc(threadArg);

        // Wake up the exec.
        rc = pthread_mutex_lock(&esm_execMutex);
        if (rc)
        {
            NFX_ASSERT(FALSE);
        }

        esm_numThreadsWaiting += 1;
        esm_thread[threadNum].state = ESM_THREAD_AVAIL;

        rc = pthread_cond_signal(&esm_execCondVar);
        if (rc)
        {
            NFX_ASSERT(FALSE);
        }

        /* PAUSE */
        rc = pthread_cond_wait( &esm_thread[threadNum].condVar,
                                &esm_execMutex);
        /* RESUME */
        if (rc)
        {
            NFX_ASSERT(FALSE);
        }
    
        DLL_REMOVE_HEAD(&esm_threadJobQueueHdr, threadArg, dllEntry);

        if (threadArg == NULL)
        {
            printf("Failed to get threadArg from job queue!!!\n");
            NFX_ASSERT(FALSE);
        }
        rc = pthread_mutex_unlock(&esm_execMutex);
        if (rc)
        {
            NFX_ASSERT(FALSE);
        }

        ESM_TRACE("esm proc %ld thread %ld handle 0x%x running job from queue ++++\n",
              nfxMyCPUId, threadNum, threadArg->handle);
    }

    /* NOTREACHED */

    // gdb has a problem if we exit the thread!!!!
    //pthread_exit(NFX_OK);

} /* esm_threadMain */

/*-----------------------------------------------------------------
 * Name:	esm_yield()
 *
 * Description:	Used by state machine threads to yield back the cpu.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 */
void
esm_yield(
	esm_event_t			*eventp,
	int32				waitSecs)
{
    esm_instance_t      *esm;
    esm_handle_t        handle;
    int32               rc, threadNum;
    struct timeval      now;
    struct timespec     timeout;

    threadNum = (int32) pthread_getspecific(esm_keyThreadNum);

    handle = esm_thread[threadNum].threadArg.handle;
    esm = esm_getInstFromHandle(handle);

    NFX_ASSERT(esm->threadNum == threadNum);

    if ((esm->flags & ESM_CAN_YIELD) == 0)
    {
        printf("ERR: esm_yield() called from non-yieldable ESM %s\n",
                esm->table->name);
        printf("   currState: %d\n", esm->currentState);
        NFX_ASSERT(FALSE);
        return;
    }

    esm->flags |= ESM_YIELD;

    ESM_TRACE("esm proc %ld handle 0x%x esm_yield signalling exec\n",
      nfxMyCPUId, handle);

    // wake up the exec
    rc = pthread_mutex_lock(&esm_execMutex);
    if (rc)
    {
        NFX_ASSERT(FALSE);
        return;
    }
    rc = pthread_cond_signal(&esm_execCondVar);
    if (rc)
    {
        NFX_ASSERT(FALSE);
        return;
    }

    /* PAUSE */

    ESM_TRACE("esm proc %ld handle 0x%x esm_yield pausing\n", nfxMyCPUId, handle);

	if (waitSecs != 0)
	{
    	gettimeofday(&now, NULL);
    	timeout.tv_sec = now.tv_sec + waitSecs;
    	timeout.tv_nsec = now.tv_usec * 1000;

    	rc = pthread_cond_timedwait(&esm_thread[threadNum].condVar,
                                	&esm_execMutex,
                                	&timeout);
    	if (rc == ETIMEDOUT) {
        	printf("esm proc %ld handle 0x%x esm_yield: waited %d seconds\n",
                	nfxMyCPUId, handle, waitSecs);
        	rc = pthread_cond_wait(&esm_thread[threadNum].condVar,
                               	&esm_execMutex);
    	}
	}
	else
	{
       	rc = pthread_cond_wait(&esm_thread[threadNum].condVar,
                           	   &esm_execMutex);
	}

    if (rc)
    {
        NFX_ASSERT(FALSE);
        return;
    }

    /* RESUME */
    rc = pthread_mutex_unlock(&esm_execMutex);
    if (rc)
    {
        NFX_ASSERT(FALSE);
        return;
    }

    ESM_TRACE("esm proc %ld handle 0x%x esm_yield resumed\n",
      nfxMyCPUId, handle);

	if (eventp != NULL)
	{
    	*eventp = esm->event;
	}

} /* esm_yield */

/*-----------------------------------------------------------------
 * Name:	esm_runThreadEvent()
 *
 * Description:	state machine engine for yielding esm types
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 */
void
esm_runThreadEvent(
	esm_instance_t		*esm,
	esm_event_t			event)
{
	int32				newState;
	esm_actionFunc_t  	***evFunArr;
	esm_actionFunc_t  	**evFun;
	esm_actionFunc_t  	*fun;
	esm_handle_t		handle;
    int32               i, rc;

    rc = pthread_mutex_lock(&esm_execMutex);
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }
    if (esm->flags & ESM_YIELD)
    {
        ESM_TRACE("esm proc %ld handle 0x%x runEvent signalling yielder\n",
          nfxMyCPUId, esm->handle);

        // Signal the condition variable so the
        // thread will resume.
        esm->flags &= ~ESM_YIELD;
        esm->event = event;

        rc = pthread_cond_signal(&(esm_thread[esm->threadNum].condVar));
        if (rc)
        {
            NFX_ASSERT(FALSE);
        }
    }
    else if ((esm->flags & ESM_RUNNING) == 0)
    {
        esm_threadArg_t     *threadArg;
        pthread_attr_t      ptAttr;

        ESM_TRACE("esm proc %ld handle 0x%x runEvent starting new\n",
          nfxMyCPUId, esm->handle);

        evFunArr = esm->table->eventVectors;
        evFun = evFunArr[event.event];

        threadArg = eee_ramAlloc(sizeof(esm_threadArg_t));
        threadArg->handle = esm->handle;
        threadArg->func = evFun[esm->currentState];
        threadArg->event = event;
        threadArg->esm = esm;

        // First event for this yielding instance.  Create a
        // (or grab a waiting) thread and call its action function.

        esm->flags |= ESM_RUNNING;

        if (esm_numThreadsWaiting > 0)
        {
            esm_numThreadsWaiting -= 1;

            // An idle thread is available, use it.
            
            DLL_INSERT_TAIL(&esm_threadJobQueueHdr, threadArg, dllEntry);

            for (i = 0; i < ESM_THREADS_MAX; i++)
            {
                if (esm_thread[i].state == ESM_THREAD_AVAIL)
                {
                    ESM_TRACE("esm proc %ld handle 0x%x using avail thread %d\n",
                      nfxMyCPUId, esm->handle, i);
                    threadArg->threadNum = i;
                    esm_thread[i].state = ESM_THREAD_BUSY;
                    rc = pthread_cond_signal(&esm_thread[i].condVar);
                    if (rc)
                    {
                        NFX_ASSERT(FALSE);
                    }
                    break;
                }
            }
            NFX_ASSERT(i < ESM_THREADS_MAX);
        }
        else
        {
            if (esm_numThreadsCreated >= ESM_THREADS_MAX)
            {
                printf("max req threads (%ld) exceeded\n",
                        esm_numThreadsCreated);
                NFX_ASSERT(FALSE);
            }

            ESM_TRACE("esm proc %ld handle 0x%x no avail threads, creating new thread %ld\n",
                  nfxMyCPUId, esm->handle, esm_numThreadsCreated);

            // No idle threads available, create one.

            rc = pthread_attr_init(&ptAttr);
            if (rc)
            {
                NFX_ASSERT(FALSE);
            }
            rc = pthread_attr_setdetachstate(&ptAttr, PTHREAD_CREATE_DETACHED);
            if (rc)
            {
                NFX_ASSERT(FALSE);
            }
            threadArg->threadNum = esm_numThreadsCreated;

            rc = pthread_cond_init( &esm_thread[threadArg->threadNum].condVar,
                                    NULL);
            if (rc)
            {
                NFX_ASSERT(FALSE);
            }

            rc = pthread_create(&esm_thread[threadArg->threadNum].thread,
                                &ptAttr,
                                &esm_threadMain,
                                threadArg);
            if (rc)
            {
                NFX_ASSERT(FALSE);
            }

            esm_numThreadsCreated += 1;
        }
    }
    else
    {
        // This is not permitted.  Events delivered to yield'able
        // ESM types when they're not in esm_yield() are not handled.

        panic("ERR: esm_runEvent: unexpected event for yielder\n");
        return;
    }

    /* PAUSE */
    ESM_TRACE("esm proc %ld handle 0x%x runEvent pausing\n", nfxMyCPUId, esm->handle);
    rc = pthread_cond_wait( &esm_execCondVar,
                            &esm_execMutex);
    /* RESUME */
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }
    rc = pthread_mutex_unlock(&esm_execMutex);
    if (rc)
    {
        NFX_ASSERT(FALSE);
    }
    ESM_TRACE("esm proc %ld handle 0x%x runEvent resuming\n", nfxMyCPUId, esm->handle);

} /* esm_runThreadEvent */

/*-----------------------------------------------------------------
 * Name:	esm_readPollInd
 *
 * Description:	 read from the poll indicator
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 */
void
esm_readPollInd(
    int32   count)
{
#if !defined(FCNIM)
    int data;

    // Drain our polling indicator socket. Ignore return, there
    // may or may not be something to read.
    read(esm_linuxPollFd[ESM_LINUX_POLL_FD_RX], &data, 1);
#endif
}

/*-----------------------------------------------------------------
 * Name:	esm_writePollInd
 *
 * Description:	 write to the poll indicator
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 */
void
esm_writePollInd(
    int32   count)
{
#if !defined(FCNIM)
    int data;

    // Ignore return, it might WOULD_BLOCK
    write(esm_linuxPollFd[ESM_LINUX_POLL_FD_TX], &data, 1);
#endif
}

/*-----------------------------------------------------------------
 * Name:    esm_allocStack()
 *
 * Description: allocate a stack for the esm
 *
 * Created by:  Bob Fozard
 *
 * Date Created:    3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_allocStack(
    esm_instance_t  *esm)
{
    // Not necessary with pthreads.
    esm->stackBase = (address_t) NULL;
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:    esm_destroyed()
 *
 * Description: Handle destruction notification from upper layer.
 *
 * Created by:  Bob Fozard
 *
 * Date Created:    3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_destroyed(
    esm_instance_t  *esm)
{
    esm->stackBase = (address_t) NULL;
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:    esm_threadLongJump()
 *
 * Description: Perform long jump on the specified thread.
 *
 * Created by:  Bob Fozard
 *
 * Date Created:    3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_threadLongJump(
    esm_instance_t  *esm)
{
    int32   threadNum;

    ESM_TRACE("proc %ld thread handle %x long jumped, switching to exec\n",
          		nfxMyCPUId, esm->handle);

    // Thread returned, mark it for destruction.  It doesn't
    // hurt to do this even if its already been done by the client.
    esm_destroy_esm(esm);

    threadNum = (int32) pthread_getspecific(esm_keyThreadNum);

    // Switch back to the exec.
	longjmp(esm_thread[threadNum].jmpBuf, 1);
}

#endif // LINUX_TEST
