/*
 *-----------------------------------------------------------------
 *
 * Name:	esm-nfx.c
 *
 * RCS:  $Id: esm-nfx.c,v 1.32 2002/11/26 22:27:31 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	esm stuff when compiled for NFX (vs. LINUX_TEST)
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>

#if !defined(LINUX_TEST)

#include <exports/esm-api.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#include <exports/debug.h>
#include <exports/timer-api.h>
#include <exports/queue-api.h>
#include <exports/crashdump-api.h>
#include <exports/except-api.h>

#include "esm.h"
#include "esm-nfx.h"

address_t	esm_execSavedStackPtr CPU_PRIVATE;

#define ESM_STACK_SIZE_MIN       (4 * 1024)
#define ESM_STACK_SIZE_DEFAULT  (16 * 1024)
#define ESM_STACK_SIZE_MAX      (32 * 1024)

// TODO - use elog for tracing.
#define ESM_TRACE

#include <nfx-target.h>

extern address_t	stack_addr;

nfx_spinlock_t		esm_yieldQueueSpin;

/*-----------------------------------------------------------------
 * Code to provide yield timeout mechanism.
 */
DLL_HDR(esm_instance_t)         esm_yieldQueueHdr;
int32 esm_yieldTimerHandle;
#define ESM_YIELD_TIMER_ENQUEUE(_esm)                                       \
    NFX_SPIN_LOCK(&esm_yieldQueueSpin);                                     \
    DLL_INSERT_TAIL(&esm_yieldQueueHdr, (_esm), yieldQueueEntry);           \
    NFX_SPIN_UNLOCK(&esm_yieldQueueSpin);                           
#define ESM_YIELD_TIMER_DEQUEUE(_esm)                                       \
    NFX_SPIN_LOCK(&esm_yieldQueueSpin);                                     \
    DLL_REMOVE(&esm_yieldQueueHdr, (_esm), yieldQueueEntry);                \
    NFX_SPIN_UNLOCK(&esm_yieldQueueSpin);                           

int32
esm_yieldTimer(
    uint32   timerHandle,
    void    *p1,
    void    *p2,
    void    *p3,
    void    *p4)
{
    esm_instance_t  *esm, *esmNext;
	esm_event_t     event;

    NFX_SPIN_LOCK(&esm_yieldQueueSpin);
    DLL_ITERATE_NEXT(esm, esmNext, &esm_yieldQueueHdr, yieldQueueEntry)
    {
        esm->waitSecs -= 1;
        if (esm->waitSecs == 0)
        {
            //printf("ESM: esm %s instance %x timed out in esm_yield\n",
            //       esm->table->name, esm->handle);
            DLL_REMOVE(&esm_yieldQueueHdr, esm, yieldQueueEntry);
            event.event = ESM_THREAD_EVENT_TIMEOUT;

            NFX_SPIN_LOCK(&esm_spinlock);
            if (!(esm->flags & ESM_DESTROY))
            {
                esm_enqueueEvent(esm, event, ESM_QUEUE_PRI_LO);
            }
            NFX_SPIN_UNLOCK(&esm_spinlock);
        }
    }
    NFX_SPIN_UNLOCK(&esm_yieldQueueSpin);

    return 0;

} /* esm_yieldTimer */

static void
esm_yieldTimerInit()
{
    int32   rc;

    DLL_INIT(&esm_yieldQueueHdr);

	NFX_SPIN_INIT(&esm_yieldQueueSpin, SPINLOCK_LEVEL_ESM);

    // Start one-second timer for yield timeout.
    esm_yieldTimerHandle = eee_timerCreate(1000, EEE_TIMER_REPEAT);
    if (esm_yieldTimerHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
    {
        panic("esm_initPlatform: no timer\n");
    }
    rc = eee_timerStart(esm_yieldTimerHandle,
                        esm_yieldTimer,
                        NULL,NULL,NULL,NULL);
    if (rc != NFX_OK)
    {
        panic("esm_initPlatform: failed to start yield timer\n");
    }

} /* esm_yieldTimerInit */

/*-----------------------------------------------------------------
 * Code to do some near-overflow checking on thread stacks.
 */
#if defined(DEBUG)
#if !defined(USE_THREAD_STACKS) && !defined(USE_BSS_THREAD_STACKS)
uint32  esm_stackRedZoneHits = 0;
#define ESM_STACK_RED_ZONE_FLAG 0x55
#define ESM_STACK_RED_ZONE_SIZE (ESM_STACK_SIZE_MIN / 2)

static void
esm_initStackRedZone(
    esm_instance_t  *esm)
{
    uchar8  *redZone;
    int32   i;

    redZone = (uchar8 *) esm->stackBase;
    for (i = 0; i < ESM_STACK_RED_ZONE_SIZE; i++, redZone++)
    {
        *redZone = ESM_STACK_RED_ZONE_FLAG;
    }

} /* esm_initStackRedZone */

static void
esm_checkStackRedZone(
    esm_instance_t  *esm)
{
    uchar8  *redZone;
    int32   i;

    redZone = (uchar8 *) esm->stackBase;
    for (i = 0; i < ESM_STACK_RED_ZONE_SIZE; i++, redZone++)
    {
        if (*redZone != ESM_STACK_RED_ZONE_FLAG)
            break;
    }
    if (i < ESM_STACK_RED_ZONE_SIZE)
    {
        printf(
            "ESM_WARNING: esm %s instance %llx thread stack in red zone (%d)\n",
                esm->table->name, (uint64)((uintptr_t)esm->handle), i);
        esm_stackRedZoneHits += 1;
		panic("thread stack overrun detected\n");
    }

} /* esm_checkStackRedZone */

#endif	/* !USE_THREAD_STACKS && !USE_BSS_THREAD_STACKS */
#endif // DEBUG

/*-----------------------------------------------------------------
 * Platform specific init.
 */
void
esm_initPlatform()
{
    esm_yieldTimerInit();

} /* esm_initPlatform */

/*-----------------------------------------------------------------------------
 * esm_threadMain
 *
 * Description:  Yield capable state machines start the thread here.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void *
esm_threadMain(void *arg)
{
    esm_actionFunc_t        ***evFunArr;
    esm_actionFunc_t        **evFun;
    esm_actionFunc_t        *fun;
    esm_instance_t          *esm;
    address_t               ignoreStack;

    esm = (esm_instance_t *) arg;
    evFunArr = esm->table->eventVectors;
    evFun = evFunArr[esm->event.event];
    fun = evFun[esm->currentState];
    
    //ESM_TRACE("proc %ld esm %x userInfo %x handle 0x%x created thread func %x\n", 
    //      nfxMyCPUId, esm, esm->userInfo, esm->handle, fun->func);
    
    // Call the thread's action function.
    (void) fun->func(   esm->handle,
                        esm->currentState,
                        esm->event,
                        esm->userInfo);
        
    //printf("proc %ld thread handle 0x%x func %x returned, switching to exec\n", 
    //nfxMyCPUId, esm->handle, fun->func);

    // Thread returned, mark it for destruction.  It doesn't
    // hurt to do this even if its already been done by the client.
    esm_destroy_esm(esm);

    // Switch back to the exec.
	exc_crashdump_set_stacktop(stack_addr + STKSIZE);
    _esm_switchThread(&ignoreStack, esm_execSavedStackPtr);
    
    return NULL;

} /* esm_threadMain */

/*-----------------------------------------------------------------
 * Name:	esm_yield()
 *
 * Description:	Used by state machine threads to yield back the cpu.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_yield(
	esm_event_t			*eventp,
	int32				waitSecs)
{
    esm_instance_t      *esm;
    esm_handle_t        handle;

    esm = esm_currInst;
    if (!esm)
    {
        panic("esm_yield from non state machine context");
    }
    handle = esm->handle;

    if ((esm->flags & ESM_CAN_YIELD) == 0)
    {
        printf("ERR: esm_yield() called from non-yieldable ESM %s\n",
                esm->table->name);
        printf("   currState: %d\n", esm->currentState);
        panic("esm_yield from non-yieldable context");
    }

#if defined(NFX_SPINLOCK_DEBUG) && defined(NFX_SMP)
	if (!LIST_EMPTY(&esm->sm_spinlocksHeld)) {
		nfx_spinlock_t *spin;
		printf("esm_yield while holding spinlocks:\n");
		LIST_FOREACH(spin, &esm->sm_spinlocksHeld, sl_entry) {
			spinlock_print(spin);
		}
		panic("esm_yield while holding spinlocks\n");
	}
#endif

    esm->flags |= ESM_YIELD;

    //ESM_TRACE("proc %ld handle 0x%x esm_yield %d secs\n",
    //  nfxMyCPUId, handle, waitSecs);

#if defined(DEBUG) && !defined(USE_THREAD_STACKS) && !defined(USE_BSS_THREAD_STACKS)
    esm_checkStackRedZone(esm);
#endif

    if (waitSecs > 0)
    {
        esm->waitSecs = waitSecs;
        ESM_YIELD_TIMER_ENQUEUE(esm);
    }

    // switch to the exec
	exc_crashdump_set_stacktop(stack_addr + STKSIZE);
    _esm_switchThread(&esm->savedStackPtr, esm_execSavedStackPtr);

    // thread resumed
    //ESM_TRACE("proc %ld handle 0x%x esm_yield resumed\n",
    //    nfxMyCPUId, handle);

    if (waitSecs > 0 && !DLL_IS_ISOLATED(esm, yieldQueueEntry))
    {
        ESM_YIELD_TIMER_DEQUEUE(esm);
    }

	if (eventp != NULL)
	{
    	*eventp = esm->event;
	}

} /* esm_yield */

/*-----------------------------------------------------------------
 * Name:	esm_runThreadEvent()
 *
 * Description:	state machine engine for yielding esm types
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_runThreadEvent(
	esm_instance_t		*esm,
	esm_event_t			event)
{
    if (esm->flags & ESM_YIELD)
    {
        //ESM_TRACE("proc %ld handle 0x%x runEvent signalling yielder\n",
        //    nfxMyCPUId, esm->handle);

        esm->flags &= ~ESM_YIELD;
        esm->event = event;

        // Resume the waiting thread.
		exc_crashdump_set_stacktop(esm->stackBase + esm->stackSize);
        _esm_switchThread(&esm_execSavedStackPtr, esm->savedStackPtr);
    }
    else if ((esm->flags & ESM_RUNNING) == 0)
    {

        //ESM_TRACE("proc %ld handle 0x%x runEvent starting new\n",
        //    nfxMyCPUId, esm->handle);

        esm->event = event;

        // First event for this yielding instance.  Create a
        // thread and call its action function.

        esm->flags |= ESM_RUNNING;

        //ESM_TRACE("proc %ld handle 0x%x creating new thread\n",
        //          nfxMyCPUId, esm->handle);

        // Create and switch to the new thread.
        //ESM_TRACE("_esm_startThread %llx, %x, %x, %x %x\n",
        //    &esm_execSavedStackPtr,
        //    esm->stackBase + esm->stackSize,
        //    esm_threadMain,
        //    esm,
        //    esm->userInfo);

		exc_crashdump_set_stacktop(esm->stackBase + esm->stackSize);
        _esm_startThread(&esm_execSavedStackPtr,
                         esm->stackBase + esm->stackSize,
                         esm_threadMain,
                         esm);
    }
    else
    {
        panic("ERR: esm_runEvent: unexpected event for yielder\n");
    }

} /* esm_runThreadEvent */

/*-----------------------------------------------------------------
 * Name:	esm_allocStack()
 *
 * Description:	allocate a stack for the esm
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_allocStack(
    esm_instance_t  *esm)
{

#if defined(USE_THREAD_STACKS) || defined(USE_BSS_THREAD_STACKS)
    esm->stackBase = esm_stackAlloc();
    if (esm->stackBase == NULL)
    {
        return NFX_ERR;
    }
    esm->stackSize = ESM_STACK_SIZE;
#else
    int32   stackSize = ESM_STACK_SIZE_DEFAULT;
    if (esm->table->stackSize)
    {
        stackSize = esm->table->stackSize;
        if (stackSize < ESM_STACK_SIZE_MIN)
        {
            stackSize = ESM_STACK_SIZE_MIN;
        }
        else if (stackSize > ESM_STACK_SIZE_MAX)
        {
            stackSize = ESM_STACK_SIZE_MAX;
        }
        // Round it up to nearest 64-bit size;
        stackSize += 7;
        stackSize &= ~7;
    }
    
    esm->stackBase = (address_t)eee_ramAlloc(stackSize);
    if (esm->stackBase == NULL)
    {
        return NFX_ERR;
    }
    esm->stackSize = stackSize;

#if defined(DEBUG)
    esm_initStackRedZone(esm);
#endif

#endif

    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:	esm_destroyed()
 *
 * Description:	Handle destruction notification from upper layer.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
int32
esm_destroyed(
    esm_instance_t  *esm)
{
    if (esm->stackBase)
    {
#if defined(USE_THREAD_STACKS) || defined(USE_BSS_THREAD_STACKS)
		esm_stackDealloc(esm->stackBase);
#else
#if defined(DEBUG)
        esm_checkStackRedZone(esm);
#endif
        eee_ramDealloc((void *)esm->stackBase);
#endif
        esm->stackBase = NULL;
    }

    if (!DLL_IS_ISOLATED(esm, yieldQueueEntry))
    {
        ESM_YIELD_TIMER_DEQUEUE(esm);
    }

    return NFX_OK;
}


/*-----------------------------------------------------------------
 * Name:	esm_threadLongJump()
 *
 * Description:	Perform long jump on the specified thread.
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	3/20/01
 *
 *-----------------------------------------------------------------
 */
void
esm_threadLongJump(
    esm_instance_t  *esm)
{
    address_t  ignoreStack;

    //ESM_TRACE("proc %ld thread handle %x long jumped, switching to exec\n", 
    //      	  nfxMyCPUId, esm->handle);

    // Thread returned, mark it for destruction.  It doesn't
    // hurt to do this even if its already been done by the client.
    esm_destroy_esm(esm);

    // Switch back to the exec.
	exc_crashdump_set_stacktop(stack_addr + STKSIZE);
    _esm_switchThread(&ignoreStack, esm_execSavedStackPtr);
}

/**
 * Return the next thread id in sequence after the
 * specified one.  If tid is 0, start at the beginning.
 * If tid is the last thread, return 0.
 * 
 * A thread id is the handle for the state machine instance.
 */
uint32
esm_gdbGetNextThreadId(
    uint32  tid)
{
    esm_handle_t    handle = (esm_handle_t) tid;
    uint32          index = ESM_GET_HANDLE_INDEX(handle);

    // No spinlock here, we're assumed to be in the
    // debugger.

    //printf("esm_gdbGetNextThreadId called for tid 0x%x\n",
    //       tid);

    if (tid != 0)
        index += 1;

    while (index < esm_instance_curr_size)
    {
        esm_instance_t  *esm = *(esm_instance_ptrs + index);

		// Make sure its a valid thread and its not the
		// running thread.
        if (esm != NULL && 
            esm->handle != 0 &&
            (esm->flags & ESM_CAN_YIELD) &&
            (esm->flags & ESM_YIELD))
        {
            //printf("esm_gdbGetNextThreadId returns next tid 0x%x\n",
            //    esm->handle);
            return esm->handle;
        }
        index += 1;
    }

    //printf("esm_gdbGetNextThreadId returns next tid 0\n");
    return 0;
}

/**
 * Return the thread id if the specified thread id is
 * still alive, otherwise return 0.
 *
 * A thread id is the handle for the state machine instance.
 */
uint32
esm_gdbIsThreadAlive(
    uint32  tid)
{
    esm_handle_t    handle = (esm_handle_t) tid;
    esm_instance_t  *esm = esm_getInstFromHandle(handle);

    //printf("esm_gdbIsThreadAlive called for tid 0x%x, esm %p\n",
    //       tid, esm);

    // No spinlock here, we're assumed to be in the
    // debugger.

    if (esm != NULL &&
        (esm->flags & ESM_CAN_YIELD) &&
        (esm->flags & ESM_YIELD))
    {
        //printf("esm_gdbIsThreadAlive returns TRUE\n");
        return tid;
    }
    else
    {
        //printf("esm_gdbIsThreadAlive returns FALSE\n");
        return 0;
    }
}

/**
 * Return the registers for the specified thread id.
 * If the thread is currently "running", or if the thread
 * is not found, return 0, otherwise the registers are
 * updated and the thread id is returned.
 *
 * A thread id is the handle for the state machine instance.
 */
uint32
esm_gdbGetRegs(
    void        *ptr,
    int32       num,
    uint32      tid)
{
    esm_handle_t    handle = (esm_handle_t) tid;
    esm_instance_t  *esm = esm_getInstFromHandle(handle);
    uint64          *stackPtr;
    cpureg_t        *regPtr = ptr;
    int32           i;

    // No spinlock here, we're assumed to be in the
    // debugger.

    //printf("esm_gdbGetRegs called for tid 0x%x, esm %p, stack %p\n",
    //       tid, esm, esm->savedStackPtr);

    if (esm != NULL &&
        (esm->flags & ESM_CAN_YIELD) &&
        (esm->flags & ESM_YIELD))
    {
        bzero(ptr, (sizeof *ptr) * num);

        for (i = 0, stackPtr = (uint64 *) ((uintptr_t)esm->savedStackPtr);
             i < num && i < 8; // up to and including s7
             i++)
        {
            *(regPtr + R_S0 + i) = *(stackPtr++);
        }

        *(regPtr + R_FP) = *(stackPtr++);
        *(regPtr + R_RA) = *(stackPtr++);
        *(regPtr + R_EPC) = *(regPtr + R_RA);
        *(regPtr + R_SP) = esm->savedStackPtr + 80;

        return tid;
    }

    // Its no longer a valid thread id or its the
    // "running" thread.

    //printf("esm_gdbGetRegs returns 0\n");

    return 0;
}

#endif // !LINUX_TEST
