/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	event.c
 *
 * RCS:  $Id: event.c,v 1.41 2002/10/08 03:16:18 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Event Manager initialization and processing
 *
 * Created by:	Pankaj Gupta 
 *
 * Date Created:	11/05/01
 *
 *-----------------------------------------------------------------
 */

#ifdef SSC

#include "nfx-incl.h"

#include <stdio.h>      //
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()

#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <errno.h>

#include "exports/elog-api.h"
#include "exports/anpssc-api.h"
#include "exports/event-api.h"
#include "exports/ncm-api.h"
#include "exports/opt-api.h"
#include "exports/eee-error.h"
#include "event-as.h"



/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

int event_setConfig(char * option, char * value);
int event_getConfig(char * option);
int event_updateConfig(opt_entry_t * entry);
void event_init(void);
void handleEventMessages(int sockFd);

/*------------------------------------------------------------------------------

 * Global variables.
 *------------------------------------------------------------------------------
 */

int eventAppFd;

as_config_t   asCfg;
event_stats_t eventStats;

char *eventTypeStr [] = {
						"Node",
						"PCC",
						"SCC",
						"Card",
						"CPU",
						"Volume",
						"IP i/f",
						"Port",
						"NFS",
						"Route",
						"Quota",
						"Mkfs"
					};

char *eventStateStr [] = {
						"None",
						"Up",
						"Down",
						"Offline",
						"Failed",
						"Accessible",
						"Remove",
						"No Change"
					};

char *volEvtStr [] = {
						"Undefined",
						"Create",
						"Delete",
						"Grow",
						"Softquota",
						"Full",
						"Modify",
						"Broken",
						"IO Error",
						"Conflict",
						"Takeover",
						"Corrupt",
						"Online",
						"Offline",
						"File System Abort"
					};

int myappid;

/*
 * We maintain a list of clients for each event type
 */
nfx_event_client_t	eventClientList[NUM_NFX_EVENTS][MAX_CLIENT_PER_EVENT];

/*------------------------------------------------------------------------------
 * Name         : event_forwardToCPUs
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 02/12/02
 *
 *------------------------------------------------------------------------------
 */
static void 
event_forwardToCPUs(nfx_eventMsg_t *eventMsg, uint32 srcCpu, 
									uint32 srcSlot)
{
	if (srcSlot != SSC_SLOT)
	{
		/* 
		 * Event from FC, FP or TxRx
		 * Forward event to SSC-MGMT 
		 */
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_SSC_MGMT, SSC_SLOT,
							MGMTPLANE, 0, 0, 0);
	}
	else if (srcSlot == SSC_SLOT)
	{
		/* 
		 * Event from SSC Unix or SSC-MGMT
		 * Forward to FC, FP and TxRx
		 */

		/* 
         * FC, cpuid 3, 2, 1, 0 are slice 0 1 2 3 respectively
		 */
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_FCNIM_CPUID3, 2,
							MGMTPLANE, 0, 0, 0);
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_FCNIM_CPUID2, 2,
							MGMTPLANE, 0, 0, 0);
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_FCNIM_CPUID1, 2,
							MGMTPLANE, 0, 0, 0);
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_FCNIM_CPUID0, 2,
							MGMTPLANE, 0, 0, 0);

		/* 
		 * TxRx 
		 */
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_NFP_TXRX1, 1,
							MGMTPLANE, 0, 0, 0);

		/* 
		 * FP 
		 */
 		sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_NFP_FP2, 1,
							MGMTPLANE, 0, 0, 0);


		if (srcCpu == NFX_SSC_UNIX)
		{
			/* 
		 	 * SSC-MGMT 
		 	 */
 			sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
		 				EVENT_SERVICE_NAME, NFX_SSC_MGMT, SSC_SLOT,
							MGMTPLANE, 0, 0, 0);
		}
	}
}

/*------------------------------------------------------------------------------
 * Name         : handleEventMessages
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 02/12/02
 *
 *------------------------------------------------------------------------------
 */

void handleEventMessages(int sockFd)
{
	int				rc;
    int             cnt;
    char            srcApp[256];
	uint32          srcCpu;
	uint32          srcSlot;
	uint32			appId;
	nfx_eventMsg_t	eventRspMsg;
	int				idx, lastUsedIdx, removedIdx, status;

	int				eventId;
	nfx_eventMsg_t	*eventMsg;
	char 			rcvMsgBuf[MAX_EVENT_BUF_SIZE];
	int				*msgType;

	event_configMsg_t *eventCfgReq;
	event_configMsg_t eventCfgRsp;

    /* 
	 * Receive and process event requests
     */
	memset(rcvMsgBuf, 0, MAX_EVENT_BUF_SIZE);

    cnt = recvAgileMsg (sockFd, rcvMsgBuf, MAX_EVENT_BUF_SIZE, NULL,
                            srcApp, &srcCpu, &srcSlot, 0);

	appId = eee_getApplicationId (srcApp);

    if (cnt <= 0)
    {
        perror("handleEventMessages: Error in rcv\n");
        return;
    }
     
	eventMsg = NULL;
	msgType = (int *)rcvMsgBuf;
	status = NFX_EVENT_STATUS_OK; 

    switch(*msgType)
    {
	case NFX_EVENT_REGISTER:
	{
		/*
		 * Register a client for given event
		 */
		eventMsg = (nfx_eventMsg_t *)rcvMsgBuf;
	   	eventId = eventMsg->eventType;

		E_LOG(class_3, info_s, myappid, 0, 0, 0,
	         ("Rcvd register request APP: %s EVENT: %s\n",
					srcApp, eventTypeStr[eventId]));

		if (appId == EEE_ERR_INVALID_APP_NAME)
		{
			status = NFX_EVENT_STATUS_ERR;
			break;
		}

		for (idx = 0; idx < MAX_CLIENT_PER_EVENT; idx++)
		{
			if (eventClientList[eventId][idx].flag == 0)
			{
				eventClientList[eventId][idx].flag = 1;
				eventClientList[eventId][idx].slot = srcSlot;
				eventClientList[eventId][idx].cpu = srcCpu;
				eventClientList[eventId][idx].appId = appId;
				strcpy(eventClientList[eventId][idx].appName,
							srcApp);
				status = NFX_EVENT_STATUS_OK;
				break;
			}
			else if ((appId == eventClientList[eventId][idx].appId) 
			      && (srcSlot == eventClientList[eventId][idx].slot)
			      && (srcCpu == eventClientList[eventId][idx].cpu))
			{
				break;	// App already registered for this event
			}
		}

		if (idx >= MAX_CLIENT_PER_EVENT)
			status = NFX_EVENT_STATUS_ERR;
        break;
	}


	case NFX_EVENT_UNREGISTER:
	{
		/*
		 * Unregister a client
		 */

		status = NFX_EVENT_STATUS_ERR;
		lastUsedIdx = removedIdx = -1;
		eventMsg = (nfx_eventMsg_t *)rcvMsgBuf;
	   	eventId = eventMsg->eventType;

		E_LOG(class_3, info_s, myappid, 0, 0, 0,
	         ("Rcvd unregister request APP: %s EVENT: %s\n",
					srcApp, eventTypeStr[eventId]));

		for (idx = 0; idx < MAX_CLIENT_PER_EVENT; idx++)
		{
			if (eventClientList[eventId][idx].flag == 0)
			{
				lastUsedIdx = idx - 1;
				break;
			}
			else if (idx == (MAX_CLIENT_PER_EVENT - 1))
			{
				lastUsedIdx = idx;
				break;
			}
				
			if ((eventClientList[eventId][idx].appId == appId)
		     && (srcSlot == eventClientList[eventId][idx].slot)
		     && (srcCpu == eventClientList[eventId][idx].cpu))
			{
				memset(&eventClientList[eventId][idx], 0, 
                       sizeof(nfx_event_client_t));
				removedIdx = idx;
				status = NFX_EVENT_STATUS_OK;
			}
		}

		if ((removedIdx >= 0) && (lastUsedIdx >=0) &&
					(removedIdx != lastUsedIdx))
		{
			/*
			 * Move the last used entry to currently
			 * freed location and free the last used location
			 */
			memcpy(&eventClientList[eventId][removedIdx],
					&eventClientList[eventId][lastUsedIdx],
						sizeof(nfx_event_client_t));

			memset(&eventClientList[eventId][lastUsedIdx], 0,
								sizeof(nfx_event_client_t));
		}
        break;
	}


	case NFX_EVENT_POST:
	{
		/*
		 * Distribute event to registered clients
		 */
		eventMsg = (nfx_eventMsg_t *)rcvMsgBuf;
	   	eventId = eventMsg->eventType;
		/*
		 * The posting SSC is responsible for tagging from which node it came
		 * the exception being NODE events since they're generated from LNM
		 * and the nodeName is that of the reported node.
		 */
		switch (eventId) {
			case NFX_EVENT_NODE:
			case NFX_EVENT_PCC:
			case NFX_EVENT_SCC:
				break;
			default:
				gethostname (eventMsg->nodeName, sizeof(eventMsg->nodeName));
		}

		/*
		 * Put the event info in local log
		 */
		ems_logEvent(eventMsg);


		for (idx = 0; idx < MAX_CLIENT_PER_EVENT; idx++)
		{
			if (eventClientList[eventId][idx].flag == 0)
				break;

			/*
			 * Forward event to local registered applications
			 */
			 sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
			 	eventClientList[eventId][idx].appName,
				eventClientList[eventId][idx].cpu,
				eventClientList[eventId][idx].slot,
				MGMTPLANE, 0, 0, 0);
		}

		/*
		 * Forward event to other CPUs
		 * Proxy events for SSC-MGMT CPU
		 */
		event_forwardToCPUs(eventMsg, srcCpu, srcSlot);
		
		/*
		 * Pass on the event to SNMP manager
		 */
		ems_sendTrap(eventMsg);

		/*
		 * Send auto support email
		 */
		if (strlen(eventMsg->reason) > 0) 
		{
			ems_sendAlarmMail(eventMsg, srcApp);
		}
		else if ((eventMsg->eventType == NFX_EVENT_CARD) ||
				(eventMsg->eventType == NFX_EVENT_CPU))
		{
			/*
		 	 * Force card and CPU events to always be sent
		 	 */
			if (eventMsg->eventType == NFX_EVENT_CARD) 
			{
				sprintf (eventMsg->reason, "Card state %s", 
			             eventStateStr[eventMsg->param.cardEvent.state]);
			}
			else if (eventMsg->eventType == NFX_EVENT_CPU) 
			{
				sprintf (eventMsg->reason, "CPU state %s", 
			             eventStateStr[eventMsg->param.cpuEvent.state]);
			}

			ems_sendAlarmMail(eventMsg, srcApp);
		}

		if (eventMsg->eventScope == NFX_EVENT_LOCAL)
			break;

		break;
	}

	case NFX_EVENT_FORWARD:
	{
		/*
		 * This is an event coming from a different node. We must be the PCC.
		 */
		E_LOG(class_3, info_s, myappid, 0, 0, 0,
	         ("Eventd Rcvd forwarded msg\n"));

		eventMsg = (nfx_eventMsg_t *)rcvMsgBuf;
		eventId = eventMsg->eventType;

		E_LOG(class_3, info_s, myappid, 0, 0, 0,
	         ("Rcvd Remote EVENT: %s\n", eventTypeStr[eventId]));

		for (idx = 0; idx < MAX_CLIENT_PER_EVENT; idx++)
		{
			if (eventClientList[eventId][idx].flag == 0)
				break;

			/*
			 * Forward event to local registered applications
			 */
			 sendAgileMsg(-1, eventMsg, sizeof(nfx_eventMsg_t),
			 	eventClientList[eventId][idx].appName,
				eventClientList[eventId][idx].cpu,
				eventClientList[eventId][idx].slot,
				MGMTPLANE, 0, 0, 0);
		}

		break;
	}

	case NFX_EVENT_AS_STATS_VIEW_REQ:
	{
		/*
		 * Send back the config block
		 */
		rc = sendAgileMsg(-1, &eventStats, sizeof(event_stats_t),
							srcApp, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							NULL, NULL, 0);

		if (rc != NFX_OK)
			printf("Could not return event statistics\n");

		break;
	}

	case NFX_EVENT_AS_STATS_CLEAR_REQ:
	{
		/*
		 * Clear the statistics counters
		 */
		memset(&eventStats, 0, sizeof(event_stats_t));

		/*
		 * Send back status ok response
		 */
		memset(&eventCfgRsp, 0, sizeof(event_configMsg_t));
		eventCfgRsp.msgType = NFX_EVENT_STATUS;
		eventCfgRsp.status = NFX_EVENT_STATUS_OK;
		sendAgileMsg(-1, &eventCfgRsp, sizeof(event_configMsg_t),
						srcApp, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						NULL, NULL, 0);
		break;
	}

	case NFX_EVENT_AS_CONFIG_SET_REQ:
	{
		eventCfgReq = (event_configMsg_t *)rcvMsgBuf;
		EVENT_DBG_PRINT(("entry = %s\n", eventCfgReq->entry.option));
		EVENT_DBG_PRINT(("value = %s\n", eventCfgReq->entry.value));
		/*
		 * Update local config
		 */
		event_updateConfig((opt_entry_t*) &(eventCfgReq->entry));

		/*
		 * Update config in options file with registry manager
		 */
		event_setConfig(eventCfgReq->entry.option, eventCfgReq->entry.value);

		/*
		 * Send back status ok response
		 */
		memset(&eventCfgRsp, 0, sizeof(event_configMsg_t));
		eventCfgRsp.msgType = NFX_EVENT_STATUS;
		eventCfgRsp.status = NFX_EVENT_STATUS_OK;

		sendAgileMsg(-1, &eventCfgRsp, sizeof(event_configMsg_t),
						srcApp, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						NULL, NULL, 0);
		break;
	}

	case NFX_EVENT_AS_CONFIG_VIEW_REQ:
	{

		/*
		 * Send back the config block
		 */
		rc = sendAgileMsg(-1, &asCfg, sizeof(as_config_t),
							srcApp, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
							NULL, NULL, 0);

		if (rc != NFX_OK)
			printf("Could not return Auto support config\n");

		break;
	}

	case NFX_EVENT_AS_GENERATE_REPORT:
	{
		/*
		 * Prepare a status message
		 */
		memset(&eventCfgRsp, 0, sizeof(event_configMsg_t));
		eventCfgRsp.msgType = NFX_EVENT_STATUS;

		/*
		 * Generate auto support report
		 */
		if (ems_sendAlarmMail(NULL, NULL) == NFX_OK)
			eventCfgRsp.status = NFX_EVENT_STATUS_OK;
		else
			eventCfgRsp.status = NFX_EVENT_STATUS_ERR;

		/*
		 * Send back status response
		 */
		sendAgileMsg(-1, &eventCfgRsp, sizeof(event_configMsg_t),
						srcApp, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						NULL, NULL, 0);
		break;
	}

    default:
	{
		status = NFX_EVENT_STATUS_ERR;
		E_LOG(class_3, info_s, myappid, 0, 0, 0,
	         ("Eventd Rcvd unhandled msg %d\n", *msgType));
	}
    } // switch(*msgType)

	/*
	 * Return the response
	 */
	if (eventMsg && eventMsg->needAck)
	{
		memset(&eventRspMsg, 0, sizeof(nfx_eventMsg_t));

		eventRspMsg.msgType = NFX_EVENT_STATUS;
		eventRspMsg.status = status;
	
		sendAgileMsg(sockFd, &eventRspMsg, sizeof(nfx_eventMsg_t),
	 					srcApp, srcCpu, srcSlot,
							MGMTPLANE, 0, 0, 0);
	}

}

static void reg_me (int evt, int place, char *app)
{
	eventClientList[evt][place].flag = 1;
	eventClientList[evt][place].slot = SSC_SLOT;
	eventClientList[evt][place].cpu = NFX_SSC_UNIX;
	eventClientList[evt][place].appId = eee_getApplicationId (app);
	strcpy (eventClientList[evt][place].appName, app);
}

/*------------------------------------------------------------------------------
 * Name         : event_init()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void event_init()
{
	 /*
	  * Prepare the event client list array
	  */
	memset(&eventClientList, 0,
			sizeof(nfx_event_client_t)*NUM_NFX_EVENTS*MAX_CLIENT_PER_EVENT);
	/*
	 * Hardwire known event consumers
	 */
	reg_me (NFX_EVENT_VOLUME,	0, "ccm");
	reg_me (NFX_EVENT_VOLUME,	1, "evm");
    reg_me (NFX_EVENT_VOLUME,   2, "ssc_ndmp");
	reg_me (NFX_EVENT_VOLUME,	3, "mountd");

	reg_me (NFX_EVENT_IP_IF,	0, "ccm");

	reg_me (NFX_EVENT_ROUTE,	0, "ccm");

	reg_me (NFX_EVENT_NFS,		0, "ccm");

	reg_me (NFX_EVENT_NODE,		0, "ccm");
	reg_me (NFX_EVENT_NODE,		1, "ncm");

    /* SPM should register for the PCC event before NCM/CCM */
	reg_me (NFX_EVENT_PCC,		0, "spm");
	reg_me (NFX_EVENT_PCC,		1, "ccm");
	reg_me (NFX_EVENT_PCC,		2, "ncm");

	reg_me (NFX_EVENT_SCC,		0, "ccm");
	reg_me (NFX_EVENT_SCC,		1, "ncm");

	reg_me (NFX_EVENT_MKFS,     0, "ssc_ndmp");

	reg_me (NFX_EVENT_CPU,		0, "ipm");
	reg_me (NFX_EVENT_CPU,		1, "evm");
	reg_me (NFX_EVENT_CPU,		2, "sdm");
	reg_me (NFX_EVENT_CPU,		3, "ea");
	reg_me (NFX_EVENT_CPU,      4, "ssc_ndmp");
	reg_me (NFX_EVENT_CPU,      5, "ccm");
	reg_me (NFX_EVENT_CPU,      6, "mountd");

	reg_me (NFX_EVENT_CARD,     0, "ccm");

	/* 
	 * Set the statistics and configuration to default.
	 */
	memset(&eventStats, 0, sizeof(event_stats_t));
	memset(&asCfg, 0, sizeof(as_config_t));

	asCfg.enable = FALSE;
	strcpy(asCfg.to, "");
	strcpy(asCfg.noteto, "");


	/* 
	 * Get the configuration for autosupport.
	 */
	event_getConfig(OPTIONS_AUTOSUPPORT_ALL);

	E_LOG(class_3, info_s, myappid, 0, 0, 0,
	      ("event_init auto support conf is TO %s NOTETO %s enable %d\n",
				asCfg.to, asCfg.noteto, asCfg.enable));

	/* 
	 * Send autosupport mail after reboot.
	 * Get the reason why we rebooted
	 */

}

/*------------------------------------------------------------------------------
 * Name         : nfx_event_stop()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
static void nfx_event_stop(int sig)
{
	unRegAgileApp(eventAppFd);
    exit(0);
}

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
int main(int argc, char **argv)
{
	/*
	 * Sets the procedure to follow in case of request for 
	 * program termination or CTRL-C interrupt.
	 */
	signal(SIGTERM, nfx_event_stop);
	signal(SIGINT, nfx_event_stop);

	 /*
      * Initialize the event manager
      */
	event_init();

	/*
	 * Register to receive messages for cluster module
	 */
	eventAppFd = regAgileApp(EVENT_SERVICE_NAME);
	if (eventAppFd < 0)
	{
		perror("Could not register Event app with Agile\n");
		return 0;
	}

	myappid = eee_getApplicationId (EVENT_SERVICE_NAME);

	kill (getppid(), SIGUSR1);		// Tell MCP we're ready to go

	while (1)
	{
		handleEventMessages(eventAppFd);
	}

	return 0;
}

#endif	/* SSC */

