/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-lv.c,v 1.56 2002/12/10 23:30:39 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-lv.c
 *
 * Description  : Volume Manager (EVM) - Logical Volume (LV) management
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "exports/timer-api.h"

#include "evm-lv.h"
#include "evm-io.h"
#include "evm-debug.h"
#include "evm-hash.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

address_t           evm_volHash;
evm_volErr_t        evmErrors;
nfx_spinlock_t      evmVolLock;
static int32        tkoClock;

/*------------------------------------------------------------------------------
 * Type definitions.
 *------------------------------------------------------------------------------
 */

typedef struct
{
    evm_IOReq_t *   req;
    evm_lun_t *     lun;
    uint64          offset;
} evm_lv_cb_t;

typedef struct evm_ioreq_T evm_ioreq_t;
struct evm_ioreq_T
{
    evm_IOReq_t *           req;
    DLL_ENTRY(evm_ioreq_t)  entry;
};

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

/* Bit operations.
 */
static inline uint64 leastBit64(uint64 n);
static inline boolean isPowerOfTwo64(uint64 n);
static inline uint64 powerOfTwoCeiling64(uint64 n);
static inline int log2_64(uint64 n);

/* Volume operations.
 */
static void
evm_destroyVol(evm_vol_t * lv);

/* Takeover operations.
 */
static void
evm_lv_updateMgrAreaOfAllVol(void);

static void
evm_lv_checkTkoAreaOfAllVol(void);

static void
evm_offlineAllVol(boolean   doLock);

/*------------------------------------------------------------------------------
 * Bit operations.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Return the smallest bit set in a 64-bit word.
 */
static inline uint64 leastBit64(uint64 n)
{
    return n &~ (n - 1);
}

/*------------------------------------------------------------------------------
 * Return true if the specified quantity is a power of two.
 */
static inline boolean isPowerOfTwo64(uint64 n)
{
    return n == leastBit64(n);
}

/*------------------------------------------------------------------------------
 * Round the specified quantity up to the next power of two.
 */
static inline uint64 powerOfTwoCeiling64(uint64 n)
{
    while (TRUE)
    {
        uint64 lowBit = leastBit64(n);
        if (n == lowBit)
            return n;
        n += lowBit;
    }
}

/*------------------------------------------------------------------------------
 * Single iteration step for base 2 logarithm functions.
 */
#define LOG_STEP(log, type, value, bits)        \
    do {                                        \
        type shift = 1 << (bits);               \
        type x = value >> shift;                \
        if (x != 0) {                           \
            value = x;                          \
            log += shift;                       \
        }                                       \
    } while (0)

/*------------------------------------------------------------------------------
 * Return the base two logarithm for a power of two.
 */
static inline int log2_64(uint64 n)
{
    uint64 log2 = 0;
    uint64 m = n;
    
    NFX_ASSERT(n > 0);
    NFX_ASSERT(isPowerOfTwo64(n));
    
    LOG_STEP(log2, uint64, m, 5);
    LOG_STEP(log2, uint64, m, 4);
    LOG_STEP(log2, uint64, m, 3);
    LOG_STEP(log2, uint64, m, 2);
    LOG_STEP(log2, uint64, m, 1);
    LOG_STEP(log2, uint64, m, 0);
    
    return log2;
}

/*------------------------------------------------------------------------------
 * Volume operations.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Online the volume on this NIM.
 * Add this volume to the list of volumes.
 */
int32
evm_createVol(evm_MsgCreateLV_t * cmsg)
    
{
    int32       rc = NFX_OK;
    evm_vol_t * newVol = NULL;
    int32       a;

    evm_trace(EVM_TRACE_DETAIL, "EVM SRVR: "
              "Create LV: VolId 0x%llx nInitSAU %d nfxId = %s/%s/%d",
              cmsg->lVol.whole_thing, cmsg->nInitSAU,
              cmsg->nfxId.cluster, cmsg->nfxId.node, cmsg->nfxId.chassisId);

    /* Create volume.
     */
    EVM_RAM_ALLOC(newVol, evm_vol_t, sizeof(evm_vol_t));
    if (newVol == NULL)
    {
        rc = NFX_ERR;
        goto evm_createVol_exit;
    }

    /* Initialize volume.
     */
    newVol->id = cmsg->lVol;
    newVol->size = 0;
    newVol->totalSize = 0;
    newVol->nbLun = 0;
    DLL_INIT(&(newVol->lunList));
    newVol->lunHash = hash_create(EVM_LUN_BUCKETS + 1,
                                  evm_lun_hash,
                                  evm_lun_compare,
                                  evm_lun_release);
    newVol->lastTakeoverTstamp = -1;
    newVol->nfxId = cmsg->nfxId;
    newVol->nRetries = 0;
    DLL_ISOLATE_ELEM(newVol, entry);

    /* Initialize volume's lun list.
     */
    for (a = 0; a < cmsg->nInitSAU; a++)
    {
        evm_lun_t * newLun = NULL;
        
        EVM_RAM_ALLOC(newLun, evm_lun_t, sizeof(evm_lun_t));
        if (newLun == NULL)
        {
            rc = NFX_ERR;
            goto evm_createVol_exit;
        }

        newLun->idx = newVol->nbLun;
        newLun->id = cmsg->initSAU[a].strgId;
        newLun->openHandle = cmsg->initSAU[a].openHandle;
        newLun->physBlkOffset = cmsg->initSAU[a].startBlk;
        newLun->volOffsetStart = newVol->size;
        newLun->volOffsetEnd = (newVol->size + cmsg->initSAU[a].auSize) - 1;
        newLun->nBlks = cmsg->initSAU[a].auSize;
        newLun->nTotalBlks = cmsg->initSAU[a].totalSize;
        DLL_ISOLATE_ELEM(newLun, entry);
        DLL_INSERT_TAIL(&(newVol->lunList), newLun, entry);
        if (hash_insert(newVol->lunHash, (void *) &(newLun->idx),
                        (void *) newLun) != NFX_OK)
        {
            evm_trace(EVM_TRACE_ERRORS, "hash_insert(newVol->lunHash) failed");
        }
        
        newVol->size += newLun->nBlks;
        newVol->totalSize += newLun->nTotalBlks;
        newVol->nbLun++;
    } // for
    newVol->pwr2AvgLunSize =
        log2_64(powerOfTwoCeiling64(newVol->size / newVol->nbLun));
    newVol->pwr2AvgTotalLunSize =
        log2_64(powerOfTwoCeiling64(newVol->totalSize / newVol->nbLun));
    
    /* Add volume to list of volumes.
     * Insert it in the hash table.
     */
    NFX_SPIN_LOCK(&evmVolLock);
    DLL_INSERT_TAIL(&evm_volList, newVol, entry);
    rc = hash_insert(evm_volHash, (void *) &(newVol->id), (void *) newVol);
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "hash_insert(evm_volHash, 0x%llx) failed",
                  newVol->id.whole_thing);
    }
    NFX_SPIN_UNLOCK(&evmVolLock);
    
    evm_trace(EVM_TRACE_INFO, "Logical volume 0x%llx is online.",
              cmsg->lVol.whole_thing);

  evm_createVol_exit:

    if (rc != NFX_OK)
    {
        if (newVol != NULL)
        {
            evm_lun_t * lun = NULL;
            
            while(!DLL_IS_EMPTY(&(newVol->lunList)))
            {
                DLL_REMOVE_TAIL(&(newVol->lunList), lun, entry);
                EVM_RAM_DEALLOC(lun);
            } // while
        }
        EVM_RAM_DEALLOC(newVol);
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Grow volume.
 * Add a LUN to a logical volume.
 */
int32
evm_addLun(evm_MsgAddStrg_t * amsg)
    
{
    int32       rc = NFX_OK;
    evm_vol_t * vol = NULL;
    evm_lun_t * newLun = NULL;

    NFX_SPIN_LOCK(&evmVolLock);
    
    /* Get volume.
     */
    if (evm_lvExists(amsg->lVol, &vol) == 0)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "cannot add storage to non existent logical vol 0x%llx",
                  amsg->lVol.whole_thing);
        return NFX_ERR;
    }

    /* Add LUN to volume.
     */
    EVM_RAM_ALLOC(newLun, evm_lun_t, sizeof(evm_lun_t));
    if (newLun == NULL)
    {
        rc = NFX_ERR;
        goto evm_addLun_exit;
    }

    newLun->idx = vol->nbLun;
    newLun->id = amsg->strg.strgId;
    newLun->openHandle = amsg->strg.openHandle;
    newLun->physBlkOffset = amsg->strg.startBlk;
    newLun->volOffsetStart = vol->size;
    newLun->volOffsetEnd = (vol->size + amsg->strg.auSize) - 1;
    newLun->nBlks = amsg->strg.auSize;
    newLun->nTotalBlks = amsg->strg.totalSize;
    DLL_ISOLATE_ELEM(newLun, entry);

    /* Add LUN to volume and update volume info.
     */
    DLL_INSERT_TAIL(&(vol->lunList), newLun, entry);
    if (hash_insert(vol->lunHash, (void *) &(newLun->idx),
                    (void *) newLun) != NFX_OK)
    {
        evm_trace(EVM_TRACE_ERRORS, "hash_insert(vol->lunHash) failed");
    }
    
    vol->size += newLun->nBlks;
    vol->totalSize += newLun->nTotalBlks;
    vol->nbLun++;
    vol->pwr2AvgLunSize = log2_64(powerOfTwoCeiling64(vol->size / vol->nbLun));
    vol->pwr2AvgTotalLunSize =
        log2_64(powerOfTwoCeiling64(vol->totalSize / vol->nbLun));
    
    evm_trace(EVM_TRACE_INFO, "Logical volume 0x%llx has grown.",
              vol->id.whole_thing);

  evm_addLun_exit:
    
    NFX_SPIN_UNLOCK(&evmVolLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Takes a logical volume offline.
 */
int32
evm_offlineLV(evm_volId_t   id,
              boolean       doLock)
    
{
    evm_vol_t * lv = NULL;
    int32       rc = NFX_OK;

    if (doLock) NFX_SPIN_LOCK(&evmVolLock);
    
    /* Find the logical volume
     */
    if (evm_lvExists(id, &lv) == 0)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "Volume 0x%llx is not online on this NIM.", id.whole_thing);
        rc = NFX_ERR;
        goto evm_offlineLV_exit;
    }
    
    // TODO!!! need to flush all outstanding IO requests ??!!!
    
    /* Just deallocate and remove lv
     * easier when the same lv comes back online via a CREATE message
     */
    evm_destroyVol(lv);
    evm_trace(EVM_TRACE_DETAIL, "EVM DETAIL: "
              "Volume 0x%llx is now offline.", id.whole_thing);

  evm_offlineLV_exit:
    
    if (doLock) NFX_SPIN_UNLOCK(&evmVolLock);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Offline a volume.
 * Remove a volume from the list of volumes served.
 * Important: volume list shall be previously locked.
 */
static void
evm_destroyVol(evm_vol_t * vol)
    
{
    evm_lun_t * lun = NULL;

    /* Mark the outstanding IOs corresponding to this volume.
     */
    evm_io_markOutstandingIO(&(vol->id));
    
    /* Delete volume's lun list.
     */
    while(!DLL_IS_EMPTY(&(vol->lunList)))
    {
        DLL_REMOVE_TAIL(&(vol->lunList), lun, entry);
        hash_delete(vol->lunHash, (void *) &(lun->idx), (void *) lun);
        EVM_RAM_DEALLOC(lun);
    } // while

    /* Delete volume.
     */
    hash_destroy(vol->lunHash);
    DLL_REMOVE(&evm_volList, vol, entry);
    if (hash_delete(evm_volHash, (void *) &(vol->id), (void *) vol) == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS, "volume 0x%llx not found in evm_volHash",
                  vol->id.whole_thing);
    }
    evm_trace(EVM_TRACE_INFO, "Logical volume 0x%llx is offline.",
              vol->id.whole_thing);    
    EVM_RAM_DEALLOC(vol);
    
    return;
}

/*------------------------------------------------------------------------------
 * Return the list of volumes served on this NIM.
 */
uint32
evm_getListVol(evm_volId_t * ids)
    
{
    uint32      nbVol = 0;
    evm_vol_t * volCurr = NULL;

    NFX_SPIN_LOCK(&evmVolLock);
    
    DLL_ITERATE(volCurr, &evm_volList, entry)
    {
        memcpy((void *) &ids[nbVol], (const void *) &(volCurr->id),
               sizeof(evm_volId_t));
        nbVol++;
    } // DLL_ITERATE

    NFX_SPIN_UNLOCK(&evmVolLock);
    return(nbVol);
}

/*------------------------------------------------------------------------------
 * Determines if a logical volume is already being served.
 * Important: volume list shall be previously locked.
 */
boolean
evm_lvExists(evm_volId_t   id,
             evm_vol_t **  lvp)
    
{
    evm_vol_t * volCurr = NULL;
    
    DLL_ITERATE(volCurr, &evm_volList, entry)
    {
        if (volCurr->id.whole_thing == id.whole_thing)
        {
            if (lvp)
            {
                *lvp = volCurr;
            }
            return(TRUE);
        }
    }
    
    return(FALSE);
}

/*----------------------------------------------------------------------------
 * Return the corresponding volume.
 * Important: volume list shall be previously locked.
 */
evm_vol_t *
evm_getVol(evm_volId_t id)
    
{
    evm_vol_t   searchVol;
    evm_vol_t * vol;

    searchVol.id = id;
    vol = (evm_vol_t *) hash_lookup(evm_volHash,
                                    (void *) &id, (void *) &searchVol);
    return(vol);
}

/*------------------------------------------------------------------------------
 * IO mapping.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Map the logical volume offset/len to a physical LUN offset/len.
 */
evm_lun_t *
evm_getPhys(uint64      lpos,   // logical volume block start
            int         blks,   // number of blocks
            evm_vol_t * vol,    // pointer to logical volume descriptor
            uint32 *    pos,    // return -> physical block start
            uint32 *    nblks,  // return ->  physical number of blocks
            boolean     accessPrivate)

{
    evm_lun_t   searchLun;
    evm_lun_t * lun = NULL;
    uint64      pwr2AvgLunSize;
    int32       lunIdx;
    uint64      start, end, len;
    int32       left, right;

    NFX_SPIN_LOCK(&evmVolLock);
    pwr2AvgLunSize = ((accessPrivate == FALSE)?
                      vol->pwr2AvgLunSize :
                      vol->pwr2AvgTotalLunSize);
    lunIdx = (int32) (lpos >> pwr2AvgLunSize);
    NFX_ASSERT(lunIdx < vol->nbLun);
    left = 0;
    right = vol->nbLun - 1;
    //printf("pwr2AvgLunSize %lld lpos %lld\n", pwr2AvgLunSize, lpos);
    
    while (TRUE)
    {
        searchLun.idx = (uint32) lunIdx;
        //printf("hash_lookup(%lx, %d)\n", vol->lunHash, lunIdx);
        lun = (evm_lun_t *) hash_lookup(vol->lunHash, (void *) &lunIdx,
                                        (void *) &searchLun);
        if (lun == NULL)
        {
            printf("wrong lunIdx %d\n", lunIdx);
            NFX_ASSERT(FALSE);
        }
        
        if (accessPrivate == FALSE)
        {
            start = lun->volOffsetStart;
            end = lun->volOffsetEnd;
            len = lun->nBlks;
        }
        else
        {
            start = lun->volOffsetEnd + 1;
            end = lun->volOffsetEnd + N_BLKS_CFG;
            len = lun->nBlks + N_BLKS_CFG;
        }
        //printf("lunIdx %d start %lld end %lld len %lld\n",
        //lun->idx, start, end, len);
        
        if ((start <= lpos) && (lpos <= end))
        {
            uint64  offset = lpos - lun->volOffsetStart;
            *pos = lun->physBlkOffset + offset;
            if (blks <= (len - offset))
            {
                *nblks = blks;
                //printf("Found offset %lld on lun #%d\n", lpos, lun->idx); 
                //printf("pos %d nblks %d\n", *pos, *nblks); 
                NFX_SPIN_UNLOCK(&evmVolLock);
                return(lun);
            }
            else
            {
                evm_trace(EVM_TRACE_ERRORS, "A logical I/O request "
                          "should not involved more than 1 LUN!");
                break;
            }
        }

        if (lpos > end)
        {
            // cannot go more on the right!
            NFX_ASSERT(lunIdx != (vol->nbLun - 1));

            left = lunIdx + 1;
            lunIdx += ((right - left) / 2) + 1;
            NFX_ASSERT((left <= lunIdx) && (lunIdx <= right));
        }
        else if (start > lpos)
        {
            // cannot go more on the left!
            NFX_ASSERT(lunIdx != 0);
            
            right = lunIdx - 1;
            lunIdx -= ((right - left) / 2) + 1;
            NFX_ASSERT((left <= lunIdx) && (lunIdx <= right));
        }
        else
        {
            NFX_ASSERT(FALSE);
            return(NULL);
        }
    } // while
    
    NFX_SPIN_UNLOCK(&evmVolLock);
    return(NULL);
}

/*------------------------------------------------------------------------------
 * Takeover operations.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Start looking if a node wants to takeover the volumes we are serving.
 */
void
evm_startLookingForTakeover(void)

{
    int32 rc = eee_timerStart(tkoClock, evm_tkoClockTick,
                              NULL, NULL, NULL, NULL);
    NFX_ASSERT((rc == NFX_OK) || (rc == TIMER_ERR_ALREADY_RUNNING));
    evm_trace(EVM_TRACE_INFO, "START looking for possible takeover.");
#ifdef DEBUG
    evmDoTstamp = TRUE;
#endif // DEBUG
    return;
}

/*------------------------------------------------------------------------------
 * Stop looking if a node wants to takeover the volumes we are serving.
 */
void
evm_stopLookingForTakeover(void)

{
    int32 rc = eee_timerStop(tkoClock);
    NFX_ASSERT((rc == NFX_OK) || (rc == TIMER_ERR_NOT_RUNNING));
    evm_trace(EVM_TRACE_INFO, "STOP looking for possible takeover.");
#ifdef DEBUG
    evmDoTstamp = FALSE;
#endif // DEBUG
    return;
}

/*------------------------------------------------------------------------------
 * Takeover clock is ticking.
 * Assert that we still serve the volumes (write mgrArea).
 * Check if somebody wants to takeover our volumes (read tkoArea).
 */
int32
evm_tkoClockTick(uint32  tmr_handle,
                 void *  usr_param1,
                 void *  usr_param2,
                 void *  usr_param3,
                 void *  usr_param4)

{
    evm_lv_updateMgrAreaOfAllVol();
    evm_lv_checkTkoAreaOfAllVol();
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Update the manager private area on all volumes served by this NIM.
 * Go through the list volumes served on this NIM.
 * Timestamp 1st LUN of each volume to show that we still own it.
 */
static void
evm_lv_updateMgrAreaOfAllVol(void)

{
    DLL_HDR(evm_ioreq_t)    reqList;
    evm_vol_t *             vol = NULL;
    eeeGMT_t                now;
    evm_ioreq_t *           currReq = NULL;
    evm_ioreq_t *           nextReq = NULL;
        
    DLL_INIT(&reqList);
    eee_TimeGMT(&now);

    /* Build the list of IO requests.
     */
    NFX_SPIN_LOCK(&evmVolLock);
    DLL_ITERATE(vol, &evm_volList, entry)
    {
        char *          buf = NULL;
        evm_mgrArea_t * data = NULL;
        evm_lv_cb_t *   cb = NULL;
        evm_IOReq_t *   req = NULL;
        evm_ioreq_t *   ioreq = NULL;
        
        evm_lun_t * lun = vol->lunList.head;
        NFX_ASSERT(lun != NULL);       
        
        EVM_RAM_ALLOC_ALIGNED_ZERO(buf, char, EVM_MANAGER_AREA * EVM_LBLK_SIZE);
        NFX_ASSERT(buf != NULL);
        data = (evm_mgrArea_t *) buf;
        data->magic = NFX_STRG_MAGIC;
        data->mgrId = vol->nfxId;
        data->timestamp = now.sec;
        data->cksum = 0;
        data->cksum = uint32_checksum(data, sizeof(evm_mgrArea_t), 0);
        
        EVM_RAM_ALLOC_ZERO(cb, evm_lv_cb_t, sizeof(evm_lv_cb_t));
        NFX_ASSERT(cb != NULL);
        
        EVM_RAM_ALLOC_ZERO(req, evm_IOReq_t, sizeof(evm_IOReq_t));
        NFX_ASSERT(req != NULL);
        
        cb->req = req;
        cb->lun = lun;
        cb->offset = 0;
        
        req->reqType = EVM_WRITE_REQ;
        req->dataType = EVM_META_DATA_REQ;
        req->id = vol->id;
        req->offset = 0 + lun->nBlks + EVM_LUN_AREA;
        req->reqSize = EVM_MANAGER_AREA;
#ifdef ISABELLE // express IO not ready
        req->attributes = SCSI_MSG_ATTR_EXPRESS;
#else // temporary
        req->attributes = SCSI_MSG_ATTR_HI_PRI;
#endif // ISABELLE
        req->wDataHandle.dataPtr = (void *) buf;
        req->clientPrivate = (uint64) (address_t) cb;
        req->codOffset = 0; // not used

        EVM_RAM_ALLOC(ioreq, evm_ioreq_t, sizeof(evm_ioreq_t));
        ioreq->req = req;
        DLL_ISOLATE_ELEM(ioreq, entry);
        DLL_INSERT_TAIL(&reqList, ioreq, entry);
    } // DLL_ITERATE_NEXT
    NFX_SPIN_UNLOCK(&evmVolLock);

    /* Send the IO requests.
     */
    DLL_ITERATE_NEXT(currReq, nextReq, &reqList, entry)
    {
        int32 rc = evm_io_req(currReq->req, evm_lv_writeMgrAreaDone, TRUE);
        if (rc != NFX_OK)
        {
            evm_trace(EVM_TRACE_ERRORS, "[][]: "
                      "evm_io_req() called from evm_lv_writeMgrArea() failed\n"
                      "Turn volume 0x%llx offline.", vol->id.whole_thing);
            EVM_RAM_DEALLOC(currReq->req->wDataHandle.dataPtr);
            EVM_RAM_DEALLOC(currReq->req);
        }
        DLL_REMOVE(&reqList, currReq, entry);        
        EVM_RAM_DEALLOC(currReq);
    } // DLL_ITERATE_NEXT
    NFX_ASSERT(DLL_IS_EMPTY(&reqList));

    return;
}

/*------------------------------------------------------------------------------
 * evm_lv_writeMgrArea() callback.
 */
void
evm_lv_writeMgrAreaDone(evm_IOStat_t * stat)

{
    evm_lv_cb_t *   cb = (evm_lv_cb_t *) (address_t) stat->clientPrivate;
    evm_IOReq_t *   req = cb->req;
    evm_mgrArea_t * data = (evm_mgrArea_t *) req->wDataHandle.dataPtr;
    
    evm_trace(EVM_TRACE_DETAIL, "EVM DETAIL: "
              "evm_lv_writeMgrAreaDone(%d)\n", stat->status);

    /* Check if the LUN has been successfully timestamped in a reasonable time.
     * if (stat->status != EVM_STAT_NO_ERROR) - volume is already offline
     */
    if (stat->status == EVM_STAT_NO_ERROR)
    {
        eeeGMT_t now;
        
        eee_TimeGMT(&now);
        if ((now.sec - data->timestamp) >= (2 * EVM_LV_TICK))
        {
            evm_trace(EVM_TRACE_ERRORS, "[][]: "
                      "evm_io_req() called from evm_lv_writeMgrArea() failed\n"
                      "Turn volume 0x%llx offline.", req->id.whole_thing);
            
            /* Turn immediately the volume offline.
             * And notify the world about it.
             */
            evm_offlineLV(req->id, TRUE);
            evm_postEvt(EVENT_VOL_CONFLICT, req->id, EVENT_STATE_DOWN); // need special event!!!
        }
    }
    
    /* Deallocate the request and the data writen.
     */
    EVM_RAM_DEALLOC(req->wDataHandle.dataPtr);
    EVM_RAM_DEALLOC(req);
    EVM_RAM_DEALLOC(cb);

    return;
}

/*------------------------------------------------------------------------------
 * Check if another node wants to take over any of the volumes
 * served by this NIM. 
 * Go through the list volumes served on this NIM.
 * Read takeover timestamp on the first LUN of the volume.
 */
static void
evm_lv_checkTkoAreaOfAllVol(void)

{
    int32                   rc = NFX_OK;
    evm_vol_t *             pCurr = NULL;
    evm_vol_t *             pNext = NULL;
    DLL_HDR(evm_ioreq_t)    reqList;
    evm_ioreq_t *           currReq = NULL;
    evm_ioreq_t *           nextReq = NULL;
    
    /* Build the list of IO requests.
     */
    NFX_SPIN_LOCK(&evmVolLock);
    DLL_ITERATE_NEXT(pCurr, pNext, &evm_volList, entry)
    {
        evm_IOReq_t *   req = NULL;
        evm_ioreq_t *   ioreq = NULL;

        EVM_RAM_ALLOC_ZERO(req, evm_IOReq_t, sizeof(evm_IOReq_t));
        NFX_ASSERT(req != NULL);
        req->reqType = EVM_READ_REQ;
        req->dataType = EVM_META_DATA_REQ;
        req->id = pCurr->id;
        req->offset = pCurr->lunList.head->volOffsetEnd
            + EVM_LUN_AREA + EVM_MANAGER_AREA + 1;
        req->reqSize = EVM_TAKEOVER_AREA;
        req->attributes = SCSI_MSG_ATTR_HI_PRI;
        req->wDataHandle.dataPtr = NULL;
        req->clientPrivate = (uint64) (address_t) req;
        req->codOffset = 0;                 // not used

        EVM_RAM_ALLOC(ioreq, evm_ioreq_t, sizeof(evm_ioreq_t));
        ioreq->req = req;
        DLL_ISOLATE_ELEM(ioreq, entry);
        DLL_INSERT_TAIL(&reqList, ioreq, entry);
    } // DLL_ITERATE_NEXT
    NFX_SPIN_UNLOCK(&evmVolLock);
    
    /* Send the IO requests.
     */
    DLL_ITERATE_NEXT(currReq, nextReq, &reqList, entry)
    {
        rc = evm_io_req(currReq->req, evm_lv_readTkoAreaDone, TRUE);
        if (rc != NFX_OK)
        {
            evm_trace(EVM_TRACE_ERRORS, "[][]: "
                      "evm_io_req() called from evm_lv_checkTkoAreaOfAllVol() "
                      "failed\n");
            EVM_RAM_DEALLOC(currReq->req);
        }
        DLL_REMOVE(&reqList, currReq, entry);
        EVM_RAM_DEALLOC(currReq);
    } // DLL_ITERATE_NEXT
    NFX_ASSERT(DLL_IS_EMPTY(&reqList));

    return;
}

/*------------------------------------------------------------------------------
 * evm_lv_checkTkoAreaOfAllVol() callback.
 */
void
evm_lv_readTkoAreaDone(evm_IOStat_t * stat)

{
    evm_IOReq_t *   req = (evm_IOReq_t *) (address_t) stat->clientPrivate;
    evm_tkoArea_t * data = (evm_tkoArea_t *) stat->rDataHandle.dataPtr;
    evm_vol_t *     lv = NULL;
    
    evm_trace(EVM_TRACE_DETAIL, "EVM DETAIL: "
              "evm_lv_readTkoAreaDone(%d)\n", stat->status);

    NFX_SPIN_LOCK(&evmVolLock);
    lv = evm_getVol(req->id);
    
    /* Check if the volume is still on-line.
     */
    if (lv == NULL)
    {
        goto evm_lv_readTkoAreaDone_exit;
    }
    NFX_ASSERT(data != NULL);

#ifdef ISABELLE // don't know if we still need that
    /* Check if data not corrupted.
     */
    if (uint32_checksum(data, sizeof(evm_tkoArea_t), 0) != 0)
    {
        eeeGMT_t now = {0, 0};
        eee_TimeGMT(&now);
        
        evm_trace(EVM_TRACE_ERRORS, "%s", eee_humandate(now.sec));
        evm_trace(EVM_TRACE_ERRORS, "Takeover area corrupted!");
        evm_trace(EVM_TRACE_ERRORS, "\tmagic: 0x%x", data->magic);
        evm_trace(EVM_TRACE_ERRORS, "\ttkoId: %s/%s/%d", data->tkoId.cluster,
                  data->tkoId.node, data->tkoId.chassisId);
        evm_trace(EVM_TRACE_ERRORS, "\ttimestamp: %d", data->timestamp);
        evm_trace(EVM_TRACE_ERRORS, "\tcksum: 0x%x", data->cksum);

        if (evmDoAssert) NFX_ASSERT(FALSE);

        /* Turn immediately the volume offline.
         */
        evm_offlineLV(lv->id, FALSE);

        /* Notify EVM and the others about it.
         */
        evm_postEvt(EVENT_VOL_CORRUPT, lv->id, EVENT_STATE_FAILED);
        goto evm_lv_readTkoAreaDone_exit;
    }
#endif // ISABELLE
    
    /* Check if the timestamp has been modified since last time.
     * If yes, notifies EA that somebody is willing to takeover this volume.
     */
    if ((data->timestamp != lv->lastTakeoverTstamp) &&
        (lv->lastTakeoverTstamp != -1))
    {
        evm_trace(EVM_TRACE_ERRORS, "Turn all volumes offline. "
                  "Node %s/%s/%d is going to take them over.\n",
                  data->tkoId.cluster, data->tkoId.node, data->tkoId.chassisId);
        evm_offlineAllVol(FALSE);
    }
    
    /* Save last takeover timestamp in memory (with the volume configuration).
     */
    lv->lastTakeoverTstamp = data->timestamp;

  evm_lv_readTkoAreaDone_exit:
    
    /* Deallocate the request and the data read.
     */
    EVM_DEALLOC_BUF(stat->rDataHandle.dataPtr);
    EVM_RAM_DEALLOC(req);
    NFX_SPIN_UNLOCK(&evmVolLock);
    return;
}

/*------------------------------------------------------------------------------
 * Stop serving all volumes.
 */
static void
evm_offlineAllVol(boolean   doLock)
{
    evm_vol_t * pCurr = NULL;
    evm_vol_t * pNext = NULL;

    /* Offline all volumes and notify the world about it.
     */
    DLL_ITERATE_NEXT(pCurr, pNext, &evm_volList, entry)
    {
        evm_offlineLV(pCurr->id, doLock);
        evm_postEvt(EVENT_VOL_TAKEOVER, pCurr->id, EVENT_STATE_DOWN);        
    } // DLL_ITERATE_NEXT
        
    /* Close all write gates.
     */
    // TODO need SCSI support !!!
    
    /* Stop timestamping.
     */
    evm_stopLookingForTakeover();

    return;
}

/*------------------------------------------------------------------------------
 * Misc. utilities.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Reduce the number of lines printed when having IO retries...
 */
void
evm_printVolErr(void)
    
{
    evm_vol_t *   x;

    NFX_SPIN_LOCK(&evmVolLock);
    
    if (evmErrors.nStaleRsp > 1)
    {
        evm_trace(EVM_TRACE_ERRORS, "[][]: %lld stale I/O responses.",
                  evmErrors.nStaleRsp);
    }

    if (evmErrors.nOffline > 1)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "[][]: %lld I/Os not sent/retried because volume is offline.",
                  evmErrors.nOffline);
    }

    if (evmErrors.nQueueFull > 1)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "[][]: %lld I/Os rejected because queue full.",
                  evmErrors.nQueueFull);
    }

    DLL_ITERATE(x, &evm_volList, entry)
    {
        if (x->nRetries > 1)
        {
            evm_trace(EVM_TRACE_ERRORS,
                      "[][]: %lld I/O retries for volume 0x%llx.",
                      x->nRetries, x->id);
        }
    } // DLL_ITERATE

    NFX_SPIN_UNLOCK(&evmVolLock);
    return;
}

/*------------------------------------------------------------------------------
 * Same as above.
 */

void
evm_resetVolErr(void)
    
{
    evm_vol_t * x = NULL;

    NFX_SPIN_LOCK(&evmVolLock);

    evmErrors.isError = FALSE;
    evmErrors.nStaleRsp = 0;
    evmErrors.nOffline = 0;
    evmErrors.nQueueFull = 0;
    DLL_ITERATE(x, &evm_volList, entry)
    {
        x->nRetries = 0;
    } // DLL_ITERATE

    NFX_SPIN_UNLOCK(&evmVolLock);
    return;
}

/*------------------------------------------------------------------------------
 * Initialization.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_lv_init
 *
 * Description  : Initialize the logical volume module.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_lv_init(void)

{
    int32   rc = NFX_OK;
    
    /* Initialize our list of served logical volumes.
     */
    DLL_INIT(&evm_volList);
    evm_volHash = hash_create(EVM_VOL_BUCKETS + 1,
                              evm_vol_hash,
                              evm_vol_compare,
                              evm_vol_release);
    
    /* Initialize spin locks for the lv configuration list.
     */
    NFX_SPIN_INIT(&evmVolLock, SPINLOCK_LEVEL_UNKNOWN);

    /* Create a timer to periodically timestamp the volume once
     * we are PCC-less.
     */
    tkoClock = eee_timerCreate(EVM_LV_TICK, EEE_TIMER_REPEAT);
    NFX_ASSERT(tkoClock != TIMER_ERR_NO_TIMER_AVAILABLE);

    return(rc);
}

/*----------------------------------------------------------------------------*/
