/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-perfeee.c,v 1.7 2002/08/23 17:29:43 isabellec Exp $
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Name		 : evm-perf.c
 *
 * Description  : 
 *
 * Created by   : Lan Chin
 *				Isabelle Cnudde
 *              Shyh-Pei Yen
 *
 * Date Created : 25-JUN-2002
 *
 *-----------------------------------------------------------------
 */

#include "nfx-incl.h"
#include "exports/hash-api.h"
#include "exports/timer-api.h"
#include "evm-lv.h"
#include "evm-msg.h"
#include "evm-debug.h"

#ifdef SY_IO_PERFORMANCE_TEST

/* this skip the whole performance testing code if not defined... */


/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef struct
{
	int32		key;	// hash key
	int32		vidx;	// volume index for this test (always start from 0)
	int32		id;		// test id
	evm_pReq_t	type;	// request type
	evm_IOReq_t	req;
	char		*buf;	// write data
	int32		len;	// nbytes of data
	ushort16	chksum;	// calculated write checksum
	eeeGMT_t	ts;		// timestamp at request sent
} evm_perf_cb_t;

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

#define EVM_SRVR_TEST_HASH_BUCKETS  255

static address_t	evm_srvr_test_hash;
static int32		evm_srvr_test_seqno = 1;

#define EVM_MAX_NUM_OF_LVS		10
#define EVM_MAX_NUM_OF_TESTS	0x00FFFFFF

#define EVM_PERF_FIXED		0
#define EVM_PERF_SEQ		1
#define EVM_PERF_SKIPSEQ	2
#define EVM_PERF_RAND		3
#define EVM_PERF_LOOPBACK	4


/* command line parameters */
static uint32			totalvols;
static int32			perf_type;		/* m_write/m_read/u_write/u_read */
static evm_volOffset_t	perf_offset;	/* LBA offset */
static int32			perf_nBlks;		/* total blocks to read/write */
static int32			perf_opoffset;	/* fixed|skip|seq|rand|loopback */
static uint32			perf_nreqs;		/* total requests to send out for each volume */
static uint32			perf_Qlen;		/* keep request queue to this size for each vol */
static evm_vol_t	    *perf_lvp[EVM_MAX_NUM_OF_LVS];	/* LUN volume array */
static char 			*perf_mwdatabuf = NULL;

static eee_desc_t *sampleDesc = NULL;
static scsi_writeReqMsg_t *sampleReq = NULL;

/* stats keeper */
static uint32		nextid[EVM_MAX_NUM_OF_LVS];
static eeeGMT_t		perf_before_reqdump, perf_after_reqdump, perf_last_resp;
static uint32		totalrsps[EVM_MAX_NUM_OF_LVS];


/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static int32
evm_doWrite(int32			vidx,
			int32			n,
			evm_pReq_t		type,
			evm_vol_t	    *lv,
			evm_volOffset_t	offset,
			evm_volSize_t	size,
			boolean			doUltra);


/*-----------------------------------------------------------------
 * Name:		evm_perfDumpStats()
 *
 * Description: Dump statistics
 *
 * Created by : Shyh-Pei Yen
 *
 * Date Created: 6/26/02
 *-----------------------------------------------------------------
 */
static void
evm_perfDumpStats(int32 dumpT, int32 dumpQ)
{

	evm_trace(EVM_TRACE_ERRORS,
				"timestamp before queue up requests - %d:%06d",
				perf_before_reqdump.sec, perf_before_reqdump.usec);
	evm_trace(EVM_TRACE_ERRORS,
				"timestamp after queue up requests  - %d:%06d",
				perf_after_reqdump.sec, perf_after_reqdump.usec);
	evm_trace(EVM_TRACE_ERRORS,
				"timestamp last resp received       - %d:%06d",
				perf_last_resp.sec, perf_last_resp.usec);
	return;
} /* evm_perfDumpStats() */


/*------------------------------------------------------------------------------
 * Name         : evm_perf_rcvMsg
 *
 * Description  : Receives unicast messages sent to evm_perf
 *
 * Created by   : Greg Seibert
 *                Lan Chin
 *                Isabelle Cnudde
 *                Shyh-Pei Yen
 *
 * Date Created : 11-JUL-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_perf_rcvMsg(eee_descPtr_t edesc)
{
	uint32          msgLen = 0;
	char *          msgBuf = NULL;
	uint32 *        msgType;
    eee_descBuf_t * bd = NULL;
    int32           rc, numBufs;
    uint32          srcPort, destPort;

    numBufs = EEE_GET_NUM_BUFS(edesc);
    srcPort = edesc->hdr.src_port;
    destPort = edesc->hdr.dest_port;
    if (numBufs == 1)
    {
        // Single buffer message, don't copy to flat buffer.
        if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
            bd = &(edesc->bd[0]);
        else
            bd = &(edesc->bd[1]);
        msgBuf = bd->buf;
        msgLen = bd->len;
        
        // Done with the descriptor.
        eee_deallocateDesc(edesc);
    }
    else
    {
        // Copy the multi-buffer message to a
        // contiguous malloc buffer and free the
        // received packet.
        msgLen = EEE_GET_PKT_LEN(edesc);
        EVM_RAM_ALLOC(msgBuf, char, msgLen);
        if (msgBuf == NULL)
        {
            printf("evm: no mem, dropping packet\n");
            eee_freePacket(edesc);
            return;
        }
        rc = eee_copyPacketToLocalBuffer(msgBuf,
                                         edesc,
                                         msgLen,
                                         0);
        if (rc != msgLen)
        {
            printf("evm: copyPacketToLocal failed\n");
            eee_freePacket(edesc);
            return;
        }

        // Done with the packet.
        eee_freePacket(edesc);
    }

	msgType = (int *)msgBuf;
    
	switch (*msgType)
    {
    case SCSI_MID_WRITE_RSP:
    {
		uint32 vidx, n;

        scsi_writeRspMsg_t *wp = (scsi_writeRspMsg_t *) msgBuf;
		vidx = ((uint32) wp->hdr.cHandle) >> 24;
		n = ((uint32) wp->hdr.cHandle) & 0x00ffffff;
		totalrsps[vidx]++;
		if (totalrsps[vidx] >= perf_nreqs)
		{
			evm_trace(EVM_TRACE_ERRORS, "EVM perf test for vol idx %d is done", vidx);
			eee_TimeGMT(&perf_last_resp);
		}
		else if (nextid[vidx] < perf_nreqs)
		{
			evm_volOffset_t	offset;
			evm_volSize_t size;

			size = perf_nBlks * EVM_LBLK_SIZE;
			if (perf_opoffset == EVM_PERF_SEQ)
				offset = perf_offset + (nextid[vidx] * size);
			else if (perf_opoffset == EVM_PERF_SKIPSEQ)
				offset = perf_offset + (nextid[vidx] * (size+2));

			rc = evm_doWrite(vidx, nextid[vidx]++, perf_type,
						perf_lvp[vidx], offset, perf_nBlks, FALSE);
			if (rc == NFX_ERR)
			{
				evm_trace(EVM_TRACE_ERRORS,
						"ERROR[%d]: cannot do metawr for vol %d offset %llx nBlks %lld",
							nextid[vidx], perf_offset, perf_nBlks);
			}
		}
        // evm_io_scsiWriteRsp(wp);
        break;
    } // case SCSI_MID_WRITE_RSP
    
    case SCSI_MID_ERR_RSP:
    {
        scsi_errRspMsg_t *  ep = (scsi_errRspMsg_t *) msgBuf;
        evm_trace(EVM_TRACE_ERRORS,
                  "EVM ERROR: SCSI_MID_ERR_RSP eCode %d cHandle %d",
                  ep->eCode, ep->hdr.cHandle);
        break;
    } // case SCSI_MID_ERR_RSP
    
    default:
        evm_trace(EVM_TRACE_ERRORS, "EVM ERROR: "
                  "unknown message type %d from 0x%x to 0x%x",
                  *msgType, srcPort, destPort);
	} // switch
    
	if (numBufs == 1)
		{ EVM_DEALLOC_BUF(msgBuf); }
	else
		{ EVM_RAM_DEALLOC(msgBuf); }

	return;
}


/*-----------------------------------------------------------------
 * Name:		evm_doWrite
 *
 * Description: called to unit test EVM srvr meta WRITE
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_doWrite(int32				vidx,
			int32				n,
			evm_pReq_t			type,
			evm_vol_t		    *lv,
			evm_volOffset_t		offset,
			evm_volSize_t		size,
			boolean				doUltra)
{
    int32 rc;
    uint32 totalSize;
    eee_descPtr_t msgp = NULL;
    scsi_writeReqMsg_t *hp = NULL;

    /*
     * allocate eee message descriptor and scsi message
     */
    msgp = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
    if (msgp == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS,
            "EVM ERROR[%d]: cannot allocate EEE descriptor for evm_doWrite");
        goto evm_doWrite_err;
    }
    memset((void *) msgp, 0, sizeof(eee_desc_t));
	//memcpy((void *)msgp, (void *)&sampleDesc, sizeof(eee_desc_t));

#if 0
	datasize = size * EVM_LBLK_SIZE;
	EVM_RAM_ALLOC_ALIGNED(buf, char, datasize);
	if (buf == NULL)
	{
		evm_trace(EVM_TRACE_ERRORS, "EVM PERF: no memory for write buffer\n");
        goto evm_doWrite_err;
	}
	// memset(buf, 0x08, datasize);

    msgp->bd[2].buf = buf;
#endif

    msgp->bd[2].buf = perf_mwdatabuf;
    msgp->bd[2].len = size * EVM_LBLK_SIZE;
    msgp->bd[2].control = EDESC_BD_CMD_IN_USE | EDESC_BD_CMD_NO_DEALLOC;

    hp = (scsi_writeReqMsg_t *)
			eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_SMALL, 0);
    if (hp == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS, "EVM ERR: can't allocate write header buffer");
        goto evm_doWrite_err;
    }

    // Apply an offset into the buffer for transport alignment requirements.
    //(uint32)hp += MALLOC_ALIGNED_SIZE; // is it ok???
    (uint64) (address_t) hp += (MALLOC_ALIGNED_SIZE * 2);
    (char *)hp -= sizeof(scsi_writeReqMsg_t);

    msgp->bd[1].buf = (void *) hp;
    msgp->bd[1].len = sizeof(scsi_writeReqMsg_t);
    msgp->bd[1].control = EDESC_BD_CMD_IN_USE;

	memset((void *)hp, 0, sizeof(scsi_writeReqMsg_t));
    //memcpy((void *)hp, (void *)&sampleReq, sizeof(scsi_writeReqMsg_t));
	hp->hdr.mid = SCSI_MID_WRITE_REQ;
    hp->hdr.cHandle = (vidx << 24) | n;		/* this will come back to us */
	hp->transSeqId = 0;
	if (perf_opoffset == EVM_PERF_LOOPBACK)
   		hp->attributes = SCSI_MSG_ATTR_LOOPBACK;
   	hp->blkOffset = offset;
   	hp->blkCnt = size;
   	hp->checkSum = 0;

	totalSize = sizeof(scsi_writeReqMsg_t) + (size * EVM_LBLK_SIZE);

	rc = sdma_handleToPath(perf_lvp[vidx]->lunList.head->openHandle,
				&(hp->hdr.path), &(msgp->hdr.dest_port));
	if (rc == -1)
	{
		evm_trace(EVM_TRACE_ERRORS, "EVM ERR: sdma_handleToPath() failed");
	}

	msgp->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                  nfxMyCPUId,
                                                  perfAppId,
                                                  EEE_USER_DATA,//EEE_CONTROL,
                                                  NFX_COS_HIGH,
                                                  PORT_BEST_FIT);
	/* buffer count (2) and total data size */
	msgp->hdr.control = (2 << EDESC_BUFCNT_SHIFT) | totalSize;

    rc = eee_forwardPacket(msgp);
	if (rc != EEE_FORWARDED_PACKET)
		goto evm_doWrite_err;

	// evm_trace(EVM_TRACE_ERRORS, "EVM EEE[%d]: SCSI msg forwarded", n);
	return NFX_OK;

evm_doWrite_err:
	evm_trace(EVM_TRACE_ERRORS, "EVM ERR[%d]: evm_doWrite: Error Processing\n");
    if (msgp)
 		eee_freePacket(msgp);
       
    return NFX_ERR;

} /* evm_doWrite */


/*-----------------------------------------------------------------
 * Name:		evm_perf_usage
 *-----------------------------------------------------------------
 */
static void
evm_perf_usage()
{
	evm_trace(EVM_TRACE_ERRORS,
			"USAGE: evmadmin perf_eee {metawr} {fixed|seq|skip|loopback} "
			"<offset> <nsectors> <totalreqs> <num-of-vol> <Qnreqs>");
	evm_trace(EVM_TRACE_ERRORS,
			"       evmadmin perf_eee dump\n");
	return;
}


/*-----------------------------------------------------------------
 * Name:		evm_perfeee
 *
 * Description: called to performance test EVM srvr IO
 * Created by : Lan Chin
 *				Shyh-Pei Yen
 * Date Created: 6/26/02
 *-----------------------------------------------------------------
 */

int32
evm_perfeee(int32 argc, char8 *argv[])
{
	int32			rc, argi;
	int32			i, j;
	int32			lvToIdx;
	evm_vol_t       *lv;
	boolean			doUltra = FALSE;

	/* check number of arguments */
	if (argc < 2)
	{
		evm_perf_usage();
		return NFX_ERR;
	}

	/* NOW PARSE THE PARAMETERS */
	argi = 1;

	/* ==================================== */
	/* what type of operation? meta write   */
	/* or meta read or user write or user   */
	/* read?                                */
	if (strcmp(argv[argi], "dump") == 0)
	{
		int32 dumpTime = 0;
		int32 dumpQdepth = 0;

		for (i = 2; i < argc; i++)
		{
			if (strcmp(argv[i], "time") == 0)
				dumpTime = 1;
			else if (strcmp(argv[i], "qdepth") == 0)
				dumpQdepth = 1;
		}
		evm_perfDumpStats(dumpTime, dumpQdepth);

		return NFX_OK;
	}
	else if (strcmp(argv[argi], "metawr") == 0)
	{
		/* check number of arguments */
		if (argc < 7)
		{
			evm_perf_usage();
			return NFX_ERR;
		}

		perf_type = m_write;
		doUltra = FALSE;
	}
	else
	{
		evm_perf_usage();
		return NFX_ERR;
	}
	argi++;

	/* initialized variables and counters */
	perf_offset = 0;
	perf_nBlks = 0;
	perf_opoffset = 0;
	perf_nreqs = 0;
	perf_Qlen = 0;
	for (i=0; i<EVM_MAX_NUM_OF_LVS; i++)
	{
		perf_lvp[i] = 0;
		nextid[i] = 0;
		totalrsps[i] = 0;
	}
	totalvols = 0;

	/* ==================================== */
	/* is it a fixed or sequential or       */
	/* random offset for each read/write?   */
	if (strcmp(argv[argi], "fixed") == 0)
		perf_opoffset = EVM_PERF_FIXED;
	else if (strcmp(argv[argi], "seq") == 0)
		perf_opoffset = EVM_PERF_SEQ;
#if 0
/* not yet supported */
	else if (strcmp(argv[argi], "rand") == 0)
		perf_opoffset = EVM_PERF_RAND;
#endif
	else if (strcmp(argv[argi], "skip") == 0)
		perf_opoffset = EVM_PERF_SKIPSEQ;
	else if (strcmp(argv[argi], "loopback") == 0)
		perf_opoffset = EVM_PERF_LOOPBACK;
	else
	{
		evm_perf_usage();
		return NFX_ERR;
	}
	argi++;

	/* ==================================== */
	/* get volume offset and number of      */
	/* sectors to be read/write             */
	perf_offset = atol(argv[argi++]);
	perf_nBlks = atol(argv[argi++]);
	if (perf_nBlks % 2)
	{
		evm_trace(EVM_TRACE_ERRORS, "\nERROR: scsi checksum reqires even number block io\n");
		return NFX_ERR;
	}

	/* ==================================== */
	/* total requests to send out           */
	perf_nreqs = atoi(argv[argi++]);
	if (perf_nreqs > EVM_MAX_NUM_OF_TESTS)
	{
		evm_trace(EVM_TRACE_ERRORS,
			"EVMPERF num of reqs specified(%d) exceeds limit(%d) - truncated\n",
			perf_nreqs, EVM_MAX_NUM_OF_TESTS);
		perf_nreqs = EVM_MAX_NUM_OF_TESTS;
	}

	/* ==================================== */
	/* get the volumes (from and to indexes */
	lvToIdx = atoi(argv[argi++]);
	if (lvToIdx <= 0)
	{
		evm_perf_usage();
		return NFX_ERR;
	}

	NFX_SPIN_LOCK(&evm_lv_lock);
    DLL_ITERATE(lv, &evm_volList, entry)
	{
		if (lv == NULL)
			break;
		if (lvToIdx > 0)
			lvToIdx--;
		else
			break;
		perf_lvp[i] = lv;
		totalvols++;
	} // DLL_ITERATE
	NFX_SPIN_UNLOCK(&evm_lv_lock);
	if (lvToIdx > 0)
	{
		evm_trace(EVM_TRACE_ERRORS,
			"No. of volumes specified exceeds max %d volumes this test supported.",
			EVM_MAX_NUM_OF_LVS);
	}

	/* no volume? */
	if (perf_lvp[0] == NULL)
	{
		evm_trace(EVM_TRACE_ERRORS, "No logical volume found...\n");
		return NFX_ERR;
	}

	/* ==================================== */
	/* how many requests to sent to queue   */
	/* (maintain that queue length until    */
	/* all requests are sent)               */
	perf_Qlen = atol(argv[argi++]);

	if (perf_nreqs < perf_Qlen)
		perf_Qlen = perf_nreqs;

	if (perf_type == m_write)
	{
#if 0
/* IT's NOT YET FINISHED */
/* THIS IS SUPPOSED TO BE A REQUEST TEMPLATE WHICH CAN */
/* BE COPIED OVER AND OVER AGAIN TO SAVE TIME. BUT IT  */
/* DOESN'T WORK. SO THESE CODE ARE NOT COMMENTED OUT.  */
		uint32 totalSize;

		if (sampleDesc == NULL)
		{
			sampleDesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
			if (sampleDesc == NULL)
			{
				evm_trace(EVM_TRACE_ERRORS,
					"EVM ERROR[%d]: cannot allocate EEE descriptor for evm_doWrite");
				return NFX_ERR;
			}
		}
		memset((void *) sampleDesc, 0, sizeof(eee_desc_t));

		if (sampleReq == NULL)
		{
			sampleReq = (scsi_writeReqMsg_t *)
						eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_SMALL, 0);
			if (sampleReq == NULL)
			{
				evm_trace(EVM_TRACE_ERRORS, "EVM ERR: can't allocate write header buffer");
				return NFX_ERR;
			}
		}
		memset((void *) sampleReq, 0, sizeof(scsi_writeReqMsg_t));

		sampleReq->hdr.mid = SCSI_MID_WRITE_REQ;
    	//sampleReq->hdr.cHandle = 0;
		//sampleReq->transSeqId = 0;
    	sampleReq->attributes = SCSI_MSG_ATTR_LOOPBACK;
    	sampleReq->blkOffset = perf_offset;
    	sampleReq->blkCnt = perf_nBlks;
    	//sampleReq->checkSum = 0;

		totalSize = sizeof(scsi_writeReqMsg_t) + (perf_nBlks * EVM_LBLK_SIZE);

		rc = sdma_handleToPath(perf_lvp[0]->assets->allocation.phys->handle,
					&(sampleReq->hdr.path), &(sampleDesc->hdr.dest_port));
		if (rc == -1)
		{
			evm_trace(EVM_TRACE_ERRORS, "EVM ERR: sdma_handleToPath() failed");
		}

		sampleDesc->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                   nfxMyCPUId,
                                                   perfAppId,
                                                   EEE_USER_DATA,//EEE_CONTROL,
                                                   NFX_COS_HIGH,
                                                   PORT_BEST_FIT);
		/* buffer count (2) and total data size */
		sampleDesc->hdr.control = (2 << EDESC_BUFCNT_SHIFT) | totalSize;
#endif

		/* free data buffer first if previous data is there */
		if (perf_mwdatabuf != NULL)
			EVM_RAM_DEALLOC(perf_mwdatabuf);

		/* allocate data buffer for repeat use in meta write */
		EVM_RAM_ALLOC_ALIGNED(perf_mwdatabuf, char, (perf_nBlks * EVM_LBLK_SIZE));
		if (perf_mwdatabuf == NULL)
		{
			evm_trace(EVM_TRACE_ERRORS, "EVM PERF: no memory for write buffer\n");
			return NFX_ERR;
		}
		memset(perf_mwdatabuf, 0x88, perf_nBlks * EVM_LBLK_SIZE);


		eee_TimeGMT(&perf_before_reqdump);

		for (i=0; i<totalvols; i++)
		{
			evm_volOffset_t	offset1 = perf_offset;

			evm_trace(EVM_TRACE_ERRORS,
					"evmadmin performance test: volID=%llx nsectors=%d totalreqs=%d offset=%lld Qnreqs=%d",
					perf_lvp[i]->id.whole_thing, perf_nBlks, perf_nreqs, offset1, perf_Qlen);
			nextid[i] = perf_Qlen;
			for (j=0; j<perf_Qlen; j++)
			{
				rc = evm_doWrite(i, j, perf_type, perf_lvp[i], offset1, perf_nBlks, doUltra);
				if (rc == NFX_ERR)
				{
					evm_trace(EVM_TRACE_ERRORS,
						"\nERROR[%d]: can't do metawr for vol %d offset %llx nBlks %lld\n",
						i, offset1, perf_nBlks);
				}
				if (perf_opoffset == EVM_PERF_SEQ)
					offset1 += perf_nBlks;
				if (perf_opoffset == EVM_PERF_SKIPSEQ)
					offset1 += perf_nBlks + 2;
			}
		}

		eee_TimeGMT(&perf_after_reqdump);
	}

	return NFX_OK;

} /* evm_perf */

/*----------------------------------------------------------------------------*/

#endif // SY_IO_PERFORMANCE_TEST
