/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-queue.h
 *
 * Description  : Volume Manager (EVM) - Queue management
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

#ifndef _EVM_QUEUE_H_
#define _EVM_QUEUE_H_

#include "nfx-incl.h"
#include "exports/queue-api.h"  // DLL_xxx macros
#include "evm-io.h"

/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef struct t_evm_qelt evm_qelt_t;
struct t_evm_qelt
{
    evm_lioCb_t             lcb;        // control block of logical I/O
    evm_pioCb_t             pcb;        // control block of physical I/O
    eeeRTC_t                timestamp;  // time the element entered the queue
    evm_qhandle_t           handle;     // handle to find back this elt
    DLL_ENTRY(evm_qelt_t)   pEntry;     // double link list
}; 


// Pio = Physical I/O
// Lio = Logical I/O

typedef struct
{
    nfx_spinlock_t      lock;       // flag to (un)lock access to this queue

    uint32              dataType;   // type of data in the queue (meta/user)
    
    uint32              size;       // queue size (nb of elements in the queue)
    evm_qelt_t *        pMem;       // pointer to memory shared by all queues
    
    uint32              nbFree;     // number of free elements
    DLL_HDR(evm_qelt_t) pFree;      // list of free elements

    uint32              maxPioWait; // max nb of waiting physical IO requests
    uint32              nbPioWait;  // number of waiting physical IO requests
    DLL_HDR(evm_qelt_t) pPioWait;   // list of waiting physical IO requests
    
    uint32              maxPioPend; // max nb of pending physical IO requests
    uint32              nbPioPend;  // number of pending physical IO requests
    DLL_HDR(evm_qelt_t) pPioPend;   // list of pending physical IO requests
} evm_ioQueue_t;

/*------------------------------------------------------------------------------
 * Macros.
 *------------------------------------------------------------------------------
 */

#define EVM_Q_GET_LCB(_qPtr_, _handle_, _lcb_)          \
{                                                       \
    int16 _index_ = (_handle_)->parts.idx;              \
    if ((0 <= _index_) && (_index_ < (_qPtr_)->size))   \
    {                                                   \
        if ((_qPtr_)->pMem[_index_].handle.whole == (_handle_)->whole) \
        {                                               \
            _lcb_ = &((_qPtr_)->pMem[_index_].lcb);     \
        }                                               \
        else                                            \
        {                                               \
            _lcb_ = NULL;                               \
        }                                               \
    }                                                   \
    else                                                \
    {                                                   \
        _lcb_ = NULL;                                   \
    }                                                   \
}

#define EVM_Q_GET_PCB(_qPtr_, _handle_, _pcb_)          \
{                                                       \
    int16 _index_ = (_handle_)->parts.idx;              \
    if ((0 <= _index_) && (_index_ < (_qPtr_)->size))   \
    {                                                   \
        if ((_qPtr_)->pMem[_index_].handle.whole == (_handle_)->whole) \
        {                                               \
            _pcb_ = &((_qPtr_)->pMem[_index_].pcb);     \
        }                                               \
        else                                            \
        {                                               \
            _pcb_ = NULL;                               \
        }                                               \
    }                                                   \
    else                                                \
    {                                                   \
        _pcb_ = NULL;                                   \
    }                                                   \
}

#define EVM_Q_GET_TIMESTAMP(_qPtr_, _handle_, _tstamp_) \
{                                                       \
    int16 _index_ = (_handle_)->parts.idx;              \
    _tstamp_ = (_qPtr_)->pMem[_index_].timestamp;       \
}

#define EVM_Q_GO_THRU_PIOPEND(_ivar_, _nvar_, _hdr_, _handle_)      \
    ASSERT((_hdr_)->pPioPend.head != NULL);                         \
    ASSERT((_hdr_)->pPioPend.tail != NULL);                         \
    for (_ivar_ = (_hdr_)->pPioPend.head;                           \
         (_ivar_) != DLL_NIL ? ((_nvar_ = _ivar_->pEntry.next),     \
         (_handle_ = _ivar_->handle), TRUE) : FALSE;                \
         _ivar_ = _nvar_)

#define EVM_Q_GO_THRU_PIOWAIT(_ivar_, _nvar_, _hdr_, _handle_)      \
    ASSERT((_hdr_)->pPioWait.head != NULL);                         \
    ASSERT((_hdr_)->pPioWait.tail != NULL);                         \
    for (_ivar_ = (_hdr_)->pPioWait.head;                           \
         (_ivar_) != DLL_NIL ? ((_nvar_ = _ivar_->pEntry.next),     \
         (_handle_ = _ivar_->handle), TRUE) : FALSE;                \
         _ivar_ = _nvar_)

#define EVM_Q_INSERT_TO_FREE(_qPtr_, _handle_, _lcb_, _pcb_)    \
    if ((_handle_ != NULL) &&                                   \
        (evm_qInsertToFree(_qPtr_, _handle_) == NFX_OK))        \
    {                                                           \
        if (_lcb_ != NULL) _lcb_ = NULL;                        \
        if (_pcb_ != NULL) _pcb_ = NULL;                        \
    }

/*------------------------------------------------------------------------------
 * Function prototypes.
 *------------------------------------------------------------------------------
 */

int32
evm_qCreate(evm_ioQueue_t * qPtr,
            uint32          dataType,
            uint32          qSize,
            uint32          maxPioWait,
            uint32          maxPioPend);

int32
evm_qDestroy(evm_ioQueue_t *    qPtr);
    
int32
evm_qDisplay(evm_ioQueue_t *    qPtr);
    
int32
evm_qInsertToFree(evm_ioQueue_t *   qPtr,
                  evm_qhandle_t *   handle);
    
int32
evm_qRemoveFromFree(evm_ioQueue_t *     qPtr,
                    evm_qhandle_t **    handle);
    
int32
evm_qInsertToPioWait(evm_ioQueue_t *    qPtr,
                     evm_qhandle_t *    handle);
    
int32
evm_qRemoveFromPioWait(evm_ioQueue_t *  qPtr,
                       evm_qhandle_t ** handle);
    
int32
evm_qInsertToPioPend(evm_ioQueue_t *    qPtr,
                     evm_qhandle_t *    handle);

int32
evm_qRemoveFromPioPend(evm_ioQueue_t *  qPtr,
                       evm_qhandle_t *  handle);

/*----------------------------------------------------------------------------*/

#endif // _EVM_QUEUE_H_
