/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-sdm.c,v 1.12 2002/11/06 18:39:00 sandrineb Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-sdm.c
 *
 * Description  : Volume Manager (EVM) - interface with SDM
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // alarm()
#include <signal.h>     // signal()

#include "exports/anpssc-api.h"
#include "exports/timer-api.h"
#include "exports/eee-desc.h"
#include "exports/scsi-api.h"
#include "exports/sdm-api.h"
#include "evm-api.h"
#include "evm.h"
#include "evm-takeover.h"


/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

// Wait 10 sec for SDM to respond
static struct timeval       sdm_to = { 10, 0 };

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_readLunLabel
 *
 * Description  : Read the LUN label.
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 24-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_readLunLabel(scsi_id_t        id,     // scsi id of LUN to read
                 evm_StrgAttr_t * sAttr,  // contains rest of info necc for read
                 evm_strgCfg_t  * cfgp)   // put it here
{
    int32                   stat;
    sdm_readDataReqMsg_t    sdmReadMsg;
    char                    rBuf[sizeof(sdm_readDataRspMsg_t)
                                 + sizeof(evm_strgCfg_t)];
    sdm_readDataRspMsg_t *  sdmReadRsp = (sdm_readDataRspMsg_t *) rBuf;
    evm_strgCfg_t *         odCfg;
    char                    err[4096];

    /* Send SDM_MID_READ_REQ to read private meta data
     * EVM_LUN_AREA blocks are reserved at the end of lun for the private data
     * evm_strgCfg_t follows the sdmReadRsp message
     */
    memset((void *) &sdmReadMsg, 0, sizeof(sdm_readDataReqMsg_t));
    sdmReadMsg.hdr.mid = SDM_MID_READ_REQ;
    sdmReadMsg.hdr.chandle = getpid();
    sdmReadMsg.handle = sAttr->openHandle;
    sdmReadMsg.blkOffset = sAttr->blkCnt;
    sdmReadMsg.blkCnt = EVM_LUN_AREA;
    sdmReadMsg.blkSize = sAttr->blkSize;
    sdmReadMsg.attributes = SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
    memset((void *) rBuf, 0, sizeof(rBuf));
    stat = sendAgileMsg (-1, &sdmReadMsg, sizeof(sdm_readDataReqMsg_t),
                         "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                         &sdm_to, rBuf, sizeof(rBuf));
    if ((stat <= 0) || (sdmReadRsp->eCode == SDM_FAIL))
    {
        // read didn't work
        sprintf(err, "Cannot read label of LUN 0x%llx / %x (%d)\n",
                id, sAttr->openHandle, sdmReadRsp->eCode);
        EVM_ERR_PRINT(("%s", err));
        return(NFX_ERR);
    }
    
    odCfg = (evm_strgCfg_t *) sdmReadRsp->data;
    EVM_DBG_PRINT(("get_od_cfg(): LUN scsi id %llx magic %x\n",
                   id, odCfg->magic));

    if (odCfg->magic == NFX_STRG_MAGIC)
    {
        NFX_ASSERT(uint32_checksum(odCfg, sizeof(evm_strgCfg_t), 0) == 0);
    }

    memcpy((void *) cfgp, (const void *) odCfg, sizeof(evm_strgCfg_t)); 

    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_writeLunLabel
 *
 * Description  : Write the volume label on a LUN.
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 24-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_writeLunLabel(scsi_id_t        id,     // LUN for write
                  evm_StrgAttr_t * sAttr,  // contains info necc for write
                  evm_strgCfg_t *  odCfg)  // data to write
    
{
    int                     size, stat;
    char                    wBuf[sizeof(sdm_writeDataReqMsg_t) +
                                 sizeof(evm_strgCfg_t)];
    sdm_writeDataReqMsg_t * sdmWriteMsg = (sdm_writeDataReqMsg_t *) wBuf;
    sdm_writeDataRspMsg_t   sdmWriteRsp;
    evm_strgCfg_t *         destCfg;
    char                    err[EVM_MAX_ERR_STR];

    /* Initialize the write request.
     */
    bzero((void *) wBuf, sizeof(wBuf));
    sdmWriteMsg->hdr.mid = SDM_MID_WRITE_REQ;
    sdmWriteMsg->hdr.chandle = getpid();
    sdmWriteMsg->handle = sAttr->openHandle;
    sdmWriteMsg->blkCnt = EVM_LUN_AREA;
    sdmWriteMsg->blkOffset = sAttr->blkCnt;
    sdmWriteMsg->blkSize = sAttr->blkSize;
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_IGNORE_WRITE_GATE;

    /* Write modification time.
     */
    EVM_GET_TIME(odCfg->timestamp);
    
    /* Checksum the entire volume label.
     */
    odCfg->cksum = 0;
    odCfg->cksum = uint32_checksum(odCfg, sizeof(evm_strgCfg_t), 0);
    EVM_DBG_PRINT(("upd_one_od_cfg(): checksum %x for LUN %llx\n",
                   odCfg->cksum, id));

    destCfg = (evm_strgCfg_t *) sdmWriteMsg->data;
    memcpy((void *) destCfg, (void *) odCfg, sizeof(evm_strgCfg_t));
    size = sizeof(sdm_writeDataReqMsg_t) + sizeof(evm_strgCfg_t) -
           sizeof(sdmWriteMsg->data);
    stat = sendAgileMsg (-1, wBuf, size,
                         "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                         &sdm_to, &sdmWriteRsp, sizeof(sdm_writeDataRspMsg_t));
    if ((stat != sizeof(sdm_writeDataRspMsg_t)) ||
        (sdmWriteRsp.eCode == SDM_FAIL))
    {
        sprintf(err, "Cannot write manager area of LUN 0x%llx / %x (%d)\n",
                id, sAttr->openHandle, sdmWriteRsp.eCode);
        goto evm_writeLunLabel_err;
    }
    EVM_DBG_PRINT(("SDM_MID_WRITE_REQ ok -> "
                   "openHandle %d blkOffset %d blkCnt %d blkSize %d\n",
                   sdmWriteMsg->handle, sdmWriteMsg->blkOffset,
                   sdmWriteMsg->blkCnt, sdmWriteMsg->blkSize));
    
    /* Tell SPM that the LUN label has been modified,
     * so it can tell the other nodes to rescan this LUN.
     */
    {
        ssize_t                         ncmReqSize;
        ncmMsg_t *                      ncmReq = NULL;
        spm_MsgUpdateLunLabelInd_t *    spmReq= NULL;
        
        /* Build NCM message.
         */
        ncmReqSize = sizeof(ncmMsg_t) + sizeof(spm_MsgUpdateLunLabelInd_t) - 1;
        EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
        NFX_ASSERT(ncmReq != NULL);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->appName, "spm");
        strcpy(ncmReq->nodeName, "PCC");
        ncmReq->rqResp = 0;    // no response expected
        spmReq = (spm_MsgUpdateLunLabelInd_t *) ncmReq->appMsg;
        
        /* Build SPM message.
         */
        spmReq->hdr.msgType = SPM_UPDATE_LUN_LABEL_IND;
        strcpy(spmReq->hdr.node, myNode);
        spmReq->attr.scsiId = id;
        
        stat = sendAgileMsg(-1, ncmReq, ncmReqSize,
                            "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            NULL, NULL, 0);
        EVM_FREE(ncmReq);
        if (stat != NFX_OK)
        {
            sprintf(err,
                    "Unable to notify SPM about LUN label udpdate (0x%llx).",
                    id);
            goto evm_writeLunLabel_err;
        }
    }
    
    return NFX_OK;

  evm_writeLunLabel_err:

    EVM_ERR_PRINT(("%s", err));
    return NFX_ERR;
}

/*------------------------------------------------------------------------------
 * Name         : evm_readMgrArea
 *
 * Description  : Read the manager area of a LUN.
 *                Caller is responsible to deallocate the read response.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_readMgrArea(evm_StrgAttr_t *        lun,
                sdm_readDataRspMsg_t ** readRsp)

{
    int32                   rc;
    sdm_readDataReqMsg_t    req;
    char *                  buf = NULL;
    uint32                  bufSize = (sizeof(sdm_readDataRspMsg_t) +
                                       sizeof(evm_mgrArea_t));
    
    NFX_ASSERT(lun != NULL);

    /* Allocate space for the response.
     */
    EVM_ALLOC_ZERO(buf, char, bufSize);
    *readRsp = (sdm_readDataRspMsg_t *) buf;

    /* Prepare the request.
     */
    bzero((void *) &req,sizeof(sdm_readDataReqMsg_t));
    req.hdr.mid = SDM_MID_READ_REQ;
    //req.hdr.chandle = getpid();
    req.handle = lun->openHandle;
    req.blkOffset = lun->blkCnt + EVM_LUN_AREA;
    req.blkCnt = EVM_MANAGER_AREA;
    req.blkSize = lun->blkSize;
#ifdef ISABELLE // express IO not ready
    req.attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    req.attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    /* Read the manager area.
     */
    rc = sendAgileMsg(-1, &req, sizeof(sdm_readDataReqMsg_t),
                      "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &sdm_to, buf, bufSize);
    if ((rc >= sizeof(sdm_readDataRspMsg_t)) && ((*readRsp)->eCode == SDM_OK))
    {
        evm_mgrArea_t * mgrArea = (evm_mgrArea_t *)
            ((sdm_readDataRspMsg_t *) buf)->data;
        
        if (uint32_checksum(mgrArea, sizeof(evm_mgrArea_t), 0) != 0)
        {
#ifdef ISABELLE // debug only!!!
            printf("ERROR: manager area corrupted!\n");
#endif // ISABELLE
            EVM_ERR_PRINT(("ERROR: manager area corrupted!"));
            EVM_ERR_PRINT(("\tmagic: 0x%x", mgrArea->magic));
            EVM_ERR_PRINT(("\tmgrId: %s/%s/%d", mgrArea->mgrId.cluster,
                           mgrArea->mgrId.node, mgrArea->mgrId.chassisId));
            EVM_ERR_PRINT(("\ttimestamp: %d", mgrArea->timestamp));
            EVM_ERR_PRINT(("\tcksum: 0x%x", mgrArea->cksum));
            rc = NFX_ERR;
        }
        else
        {
            rc = NFX_OK;
        }
    }
    else
    {
        EVM_FREE(buf);
        *readRsp = NULL;
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot read manager area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, (*readRsp)->eCode));
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_writeMgrArea
 *
 * Description  : Write the manager private area on the LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_writeMgrArea(evm_StrgAttr_t *   lun,
                 evm_mgrArea_t *    mgrArea)
    
{
    int32                   rc = NFX_OK;
    char                    wBuf[sizeof(sdm_writeDataReqMsg_t) +
                                 sizeof(evm_mgrArea_t)];
    sdm_writeDataReqMsg_t * sdmWriteMsg = (sdm_writeDataReqMsg_t *) wBuf;
    evm_mgrArea_t *         data = (evm_mgrArea_t *) sdmWriteMsg->data;
    int                     size;
    sdm_writeDataRspMsg_t   sdmWriteRsp;

    NFX_ASSERT((mgrArea != NULL) && (lun != NULL));
    EVM_FCT_PRINT(("> evm_writeMgrArea()"));
    
    /* Send SDM_MID_WRITE_REQ to write private meta data.
     */   
    bzero((void *) wBuf, sizeof(wBuf));
    
    sdmWriteMsg->hdr.mid = SDM_MID_WRITE_REQ;
    sdmWriteMsg->hdr.chandle = getpid();
    sdmWriteMsg->handle = lun->openHandle;
    sdmWriteMsg->blkCnt = EVM_MANAGER_AREA;
    sdmWriteMsg->blkOffset =  lun->blkCnt + EVM_LUN_AREA;
    sdmWriteMsg->blkSize =  lun->blkSize;
#ifdef ISABELLE // express IO not ready
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    mgrArea->cksum = 0;
    mgrArea->cksum = uint32_checksum(mgrArea, sizeof(evm_mgrArea_t), 0);
#ifdef ISABELLE // debug only!!!
    printf("mgrArea->cksum = 0x%x\n", mgrArea->cksum);
#endif // ISABELLE
    memcpy((void *) data, (void *) mgrArea, sizeof(evm_mgrArea_t));

    size = sizeof(sdm_writeDataReqMsg_t) + sizeof(evm_mgrArea_t) -
        sizeof(sdmWriteMsg->data);
    rc = sendAgileMsg (-1, wBuf, size,
                       "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                       &sdm_to, &sdmWriteRsp, sizeof(sdm_writeDataRspMsg_t));
    if ((rc >= sizeof(sdm_writeDataRspMsg_t)) && (sdmWriteRsp.eCode == SDM_OK))
    {
        rc = NFX_OK;
    }
    else
    {
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot write manager area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, sdmWriteRsp.eCode));
    }
            
    EVM_FCT_PRINT(("< evm_writeMgrArea()"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_readTkoArea
 *
 * Description  : Read the takeover area of a LUN.
 *                Caller is responsible to deallocate the read response.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_readTkoArea(evm_StrgAttr_t *        lun,
                sdm_readDataRspMsg_t ** readRsp)

{
    int32                   rc;
    sdm_readDataReqMsg_t    req;
    char *                  buf = NULL;
    uint32                  bufSize = (sizeof(sdm_readDataRspMsg_t) +
                                       sizeof(evm_tkoArea_t));
    
    NFX_ASSERT(lun != NULL);

    /* Allocate space for the response.
     */
    EVM_ALLOC_ZERO(buf, char, bufSize);
    *readRsp = (sdm_readDataRspMsg_t *) buf;

    /* Prepare the request.
     */
    bzero((void *) &req,sizeof(sdm_readDataReqMsg_t));
    req.hdr.mid = SDM_MID_READ_REQ;
    //req.hdr.chandle = getpid();
    req.handle = lun->openHandle;
    req.blkOffset = lun->blkCnt + EVM_LUN_AREA + EVM_MANAGER_AREA;
    req.blkCnt = EVM_TAKEOVER_AREA;
    req.blkSize = lun->blkSize;
#ifdef ISABELLE // express IO not ready
    req.attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    req.attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    /* Read the manager area.
     */
    rc = sendAgileMsg(-1, &req, sizeof(sdm_readDataReqMsg_t),
                      "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &sdm_to, buf, bufSize);
    if ((rc >= sizeof(sdm_readDataRspMsg_t)) && ((*readRsp)->eCode == SDM_OK))
    {
        evm_tkoArea_t * tkoArea = (evm_tkoArea_t *)
            ((sdm_readDataRspMsg_t *) buf)->data;
        
        if (uint32_checksum(tkoArea, sizeof(evm_tkoArea_t), 0) != 0)
        {
#ifdef ISABELLE // debug only!!!
            printf("ERROR: manager area corrupted!\n");
#endif // ISABELLE
            EVM_ERR_PRINT(("ERROR: manager area corrupted!"));
            EVM_ERR_PRINT(("\tmagic: 0x%x", tkoArea->magic));
            EVM_ERR_PRINT(("\ttkoId: %s/%s/%d", tkoArea->tkoId.cluster,
                           tkoArea->tkoId.node, tkoArea->tkoId.chassisId));
            EVM_ERR_PRINT(("\ttimestamp: %d", tkoArea->timestamp));
            EVM_ERR_PRINT(("\tcksum: 0x%x", tkoArea->cksum));
            rc = NFX_ERR;
        }
        else
        {
            rc = NFX_OK;
        }
    }
    else
    {
        EVM_FREE(buf);
        *readRsp = NULL;
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot read takeover area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, (*readRsp)->eCode));
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_writeTkoArea
 *
 * Description  : Write the takeover private area on the LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 06-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_writeTkoArea(evm_StrgAttr_t *   lun,
                 evm_tkoArea_t *    tkoArea)

{
    int32                   rc = NFX_OK;
    char                    wBuf[sizeof(sdm_writeDataReqMsg_t) +
                                 sizeof(evm_tkoArea_t)];
    sdm_writeDataReqMsg_t * sdmWriteMsg = (sdm_writeDataReqMsg_t *) wBuf;
    evm_tkoArea_t *         data = (evm_tkoArea_t *) sdmWriteMsg->data;
    int                     size;
    sdm_writeDataRspMsg_t   sdmWriteRsp;

    NFX_ASSERT((tkoArea != NULL) && (lun != NULL));
    EVM_FCT_PRINT(("> evm_writeTkoArea()"));
    
    /* Send SDM_MID_WRITE_REQ to write private meta data.
     */   
    bzero((void *) wBuf, sizeof(wBuf));
    
    sdmWriteMsg->hdr.mid = SDM_MID_WRITE_REQ;
    sdmWriteMsg->hdr.chandle = getpid();
    sdmWriteMsg->handle = lun->openHandle;
    sdmWriteMsg->blkCnt = EVM_TAKEOVER_AREA;
    sdmWriteMsg->blkOffset =  lun->blkCnt + EVM_LUN_AREA + EVM_MANAGER_AREA;
    sdmWriteMsg->blkSize =  lun->blkSize;
#ifdef ISABELLE // express IO not ready
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    tkoArea->cksum = 0;
    tkoArea->cksum = uint32_checksum(tkoArea, sizeof(evm_tkoArea_t), 0);
    memcpy((void *) data, (void *) tkoArea, sizeof(evm_tkoArea_t));

    size = sizeof(sdm_writeDataReqMsg_t) + sizeof(evm_tkoArea_t) -
        sizeof(sdmWriteMsg->data);
    rc = sendAgileMsg (-1, wBuf, size,
                       "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                       &sdm_to, &sdmWriteRsp, sizeof(sdm_writeDataRspMsg_t));
    if ((rc >= sizeof(sdm_writeDataRspMsg_t)) && (sdmWriteRsp.eCode == SDM_OK))
    {
        rc = NFX_OK;
    }
    else
    {
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot write takeover area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, sdmWriteRsp.eCode));
    }
    
    EVM_FCT_PRINT(("< evm_writeTkoArea()"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_readQuorumArea
 *
 * Description  : Read the quorum area of a LUN.
 *                Caller is responsible to deallocate the read response.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 28-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_readQuorumArea(evm_StrgAttr_t *        lun,
                   sdm_readDataRspMsg_t ** readRsp)

{
    int32                   rc;
    sdm_readDataReqMsg_t    req;
    char *                  buf = NULL;
    uint32                  bufSize = (sizeof(sdm_readDataRspMsg_t) +
                                       sizeof(evm_qArea_t));
    
    NFX_ASSERT(lun != NULL);

    /* Allocate space for the response.
     */
    EVM_ALLOC_ZERO(buf, char, bufSize);
    *readRsp = (sdm_readDataRspMsg_t *) buf;

    /* Prepare the request.
     */
    bzero((void *) &req,sizeof(sdm_readDataReqMsg_t));
    req.hdr.mid = SDM_MID_READ_REQ;
    //req.hdr.chandle = getpid();
    req.handle = lun->openHandle;
    req.blkOffset = (lun->blkCnt + EVM_LUN_AREA + \
                     EVM_MANAGER_AREA + EVM_TAKEOVER_AREA);
    req.blkCnt = EVM_QUORUM_AREA;
    req.blkSize = lun->blkSize;
#ifdef ISABELLE // express IO not ready
    req.attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    req.attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    /* Read the manager area.
     */
    rc = sendAgileMsg(-1, &req, sizeof(sdm_readDataReqMsg_t),
                      "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &sdm_to, buf, bufSize);
    if ((rc >= sizeof(sdm_readDataRspMsg_t)) && ((*readRsp)->eCode == SDM_OK))
    {
        evm_qArea_t * qArea = (evm_qArea_t *)
            ((sdm_readDataRspMsg_t *) buf)->data;
        
        if (uint32_checksum(qArea, sizeof(evm_qArea_t), 0) != 0)
        {
            EVM_ERR_PRINT(("ERROR: quorum area corrupted!"));
            EVM_ERR_PRINT(("\tmagic: 0x%x", qArea->magic));
            EVM_ERR_PRINT(("\tpccId: %s/%s/%d", qArea->pccId.cluster,
                           qArea->pccId.node, qArea->pccId.chassisId));
            EVM_ERR_PRINT(("\tcksum: 0x%x", qArea->cksum));
            rc = NFX_ERR;
        }
        else
        {
            rc = NFX_OK;
        }
    }
    else
    {
        EVM_FREE(buf);
        *readRsp = NULL;
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot read quorum area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, (*readRsp)->eCode));
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_writeQuorumArea
 *
 * Description  : Write the quorum private area on the LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 28-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_writeQuorumArea(evm_StrgAttr_t *    lun,
                    evm_qArea_t *       qArea)
    
{
    int32                   rc = NFX_OK;
    char                    wBuf[sizeof(sdm_writeDataReqMsg_t) +
                                 sizeof(evm_qArea_t)];
    sdm_writeDataReqMsg_t * sdmWriteMsg = (sdm_writeDataReqMsg_t *) wBuf;
    evm_qArea_t *           data = (evm_qArea_t *) sdmWriteMsg->data;
    int                     size;
    sdm_writeDataRspMsg_t   sdmWriteRsp;

    NFX_ASSERT((qArea != NULL) && (lun != NULL));
    EVM_FCT_PRINT(("> evm_writeQuorumArea()"));
    
    /* Send SDM_MID_WRITE_REQ to write private meta data.
     */   
    bzero((void *) wBuf, sizeof(wBuf));
    
    sdmWriteMsg->hdr.mid = SDM_MID_WRITE_REQ;
    sdmWriteMsg->hdr.chandle = getpid();
    sdmWriteMsg->handle = lun->openHandle;
    sdmWriteMsg->blkCnt = EVM_QUORUM_AREA;
    sdmWriteMsg->blkOffset =  (lun->blkCnt + EVM_LUN_AREA +             \
                               EVM_MANAGER_AREA + EVM_TAKEOVER_AREA);
    sdmWriteMsg->blkSize =  lun->blkSize;
#ifdef ISABELLE // express IO not ready
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    qArea->cksum = 0;
    qArea->cksum = uint32_checksum(qArea, sizeof(evm_qArea_t), 0);
    memcpy((void *) data, (void *) qArea, sizeof(evm_qArea_t));

    size = sizeof(sdm_writeDataReqMsg_t) + sizeof(evm_qArea_t) -
        sizeof(sdmWriteMsg->data);
    rc = sendAgileMsg (-1, wBuf, size,
                       "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                       &sdm_to, &sdmWriteRsp, sizeof(sdm_writeDataRspMsg_t));
    if ((rc >= sizeof(sdm_writeDataRspMsg_t)) && (sdmWriteRsp.eCode == SDM_OK))
    {
        rc = NFX_OK;
    }
    else
    {
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot write quorum area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, sdmWriteRsp.eCode));
    }
    
    EVM_FCT_PRINT(("< evm_writeQuorumArea()"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_readMirrorArea
 *
 * Description  : Read the mirror area of a LUN.
 *                Caller is responsible to deallocate the read response.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 28-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_readMirrorArea(evm_StrgAttr_t *        lun,
                   sdm_readDataRspMsg_t ** readRsp)

{
    int32                   rc;
    sdm_readDataReqMsg_t    req;
    char *                  buf = NULL;
    uint32                  bufSize = (sizeof(sdm_readDataRspMsg_t) +
                                       sizeof(evm_qArea_t));
    
    NFX_ASSERT(lun != NULL);

    /* Allocate space for the response.
     */
    EVM_ALLOC_ZERO(buf, char, bufSize);
    *readRsp = (sdm_readDataRspMsg_t *) buf;

    /* Prepare the request.
     */
    bzero((void *) &req,sizeof(sdm_readDataReqMsg_t));
    req.hdr.mid = SDM_MID_READ_REQ;
    //req.hdr.chandle = getpid();
    req.handle = lun->openHandle;
    req.blkOffset = (lun->blkCnt + EVM_LUN_AREA + \
                     EVM_MANAGER_AREA + EVM_TAKEOVER_AREA);
    req.blkCnt = EVM_MIRROR_AREA;
    req.blkSize = lun->blkSize;
#ifdef ISABELLE // express IO not ready
    req.attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    req.attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    /* Read the manager area.
     */
    rc = sendAgileMsg(-1, &req, sizeof(sdm_readDataReqMsg_t),
                      "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &sdm_to, buf, bufSize);
    if ((rc >= sizeof(sdm_readDataRspMsg_t)) && ((*readRsp)->eCode == SDM_OK))
    {
        evm_qArea_t * qArea = (evm_qArea_t *)
            ((sdm_readDataRspMsg_t *) buf)->data;
        
        if (uint32_checksum(qArea, sizeof(evm_qArea_t), 0) != 0)
        {
            EVM_ERR_PRINT(("ERROR: mirror area corrupted!"));
            EVM_ERR_PRINT(("\tmagic: 0x%x", qArea->magic));
            EVM_ERR_PRINT(("\tpccId: %s/%s/%d", qArea->pccId.cluster,
                           qArea->pccId.node, qArea->pccId.chassisId));
            EVM_ERR_PRINT(("\tcksum: 0x%x", qArea->cksum));
            rc = NFX_ERR;
        }
        else
        {
            rc = NFX_OK;
        }
    }
    else
    {
        EVM_FREE(buf);
        *readRsp = NULL;
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot read mirror area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, (*readRsp)->eCode));
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_writeMirrorArea
 *
 * Description  : Write the mirror private area on the LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 28-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_writeMirrorArea(evm_StrgAttr_t *    lun,
                    evm_mirrorArea_t *  mArea)
    
{
    int32                   rc = NFX_OK;
    char                    wBuf[sizeof(sdm_writeDataReqMsg_t) +
                                 sizeof(evm_mirrorArea_t)];
    sdm_writeDataReqMsg_t * sdmWriteMsg = (sdm_writeDataReqMsg_t *) wBuf;
    evm_mirrorArea_t *      data = (evm_mirrorArea_t *) sdmWriteMsg->data;
    int                     size;
    sdm_writeDataRspMsg_t   sdmWriteRsp;

    NFX_ASSERT((mArea != NULL) && (lun != NULL));
    EVM_FCT_PRINT(("> evm_writeMirrorArea()"));
    
    /* Send SDM_MID_WRITE_REQ to write private meta data.
     */   
    bzero((void *) wBuf, sizeof(wBuf));
    
    sdmWriteMsg->hdr.mid = SDM_MID_WRITE_REQ;
    sdmWriteMsg->hdr.chandle = getpid();
    sdmWriteMsg->handle = lun->openHandle;
    sdmWriteMsg->blkCnt = EVM_MIRROR_AREA;
    sdmWriteMsg->blkOffset =  (lun->blkCnt + EVM_LUN_AREA +             \
                               EVM_MANAGER_AREA + EVM_TAKEOVER_AREA +   \
                               EVM_QUORUM_AREA);
    sdmWriteMsg->blkSize =  lun->blkSize;
#ifdef ISABELLE // express IO not ready
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_EXPRESS | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#else // temporary
    sdmWriteMsg->attributes = SCSI_MSG_ATTR_HI_PRI | SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif // ISABELLE

    mArea->cksum = 0;
    mArea->cksum = uint32_checksum(mArea, sizeof(evm_mirrorArea_t), 0);
    memcpy((void *) data, (void *) mArea, sizeof(evm_mirrorArea_t));

    size = sizeof(sdm_writeDataReqMsg_t) + sizeof(evm_mirrorArea_t) -
        sizeof(sdmWriteMsg->data);
    rc = sendAgileMsg (-1, wBuf, size,
                       "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                       &sdm_to, &sdmWriteRsp, sizeof(sdm_writeDataRspMsg_t));
    if ((rc >= sizeof(sdm_writeDataRspMsg_t)) && (sdmWriteRsp.eCode == SDM_OK))
    {
        rc = NFX_OK;
    }
    else
    {
        rc = NFX_ERR;
        EVM_ERR_PRINT(("Cannot write mirror area of LUN 0x%llx / %x (%d)\n",
                       lun->id, lun->openHandle, sdmWriteRsp.eCode));
    }
    
    EVM_FCT_PRINT(("< evm_writeMirrorArea()"));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_lockLun
 *
 * Description  : Lock a LUN so nobody else can access it till we release it.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_lockLun(evm_StrgAttr_t * lun)

#ifdef ISABELLE // need SDM full support
{
    int32               rc;
    sdm_reserveReqMsg_t req;
    sdm_reserveRspMsg_t rsp;
    
	bzero((void *) &req, sizeof(sdm_reserveReqMsg_t));
	req.hdr.mid = SDM_MID_DEV_RESERVE_REQ;
    //req.hdr.chandle = getpid();
	req.handle = lun->openHandle;
	req.action = SDM_DEV_RSV_ACTION_RESERVE;
	req.key = SDM_DEV_RSV_KEY_DEFAULT; // need my own key???
	
	rc = sendAgileMsg(-1, &req, sizeof(sdm_reserveReqMsg_t),
                      "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &sdm_to, &rsp, sizeof(sdm_reserveReqMsg_t));
    if (rc == sizeof(sdm_reserveRspMsg_t))
    {
        rc = rsp.rCode;
    }
    else
    {
        rc = NFX_ERR;
    }
    
    return(rc);
}
#else
{
    return(NFX_OK);
}
#endif // ISABELLE

/*------------------------------------------------------------------------------
 * Name         : evm_releaseLun
 *
 * Description  : Release LUN we previously locked.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_releaseLun(evm_StrgAttr_t * lun)

#ifdef ISABELLE // need SDM full support
{
    int32               rc;
    sdm_reserveReqMsg_t req;
    sdm_reserveRspMsg_t rsp;
    
	bzero((void *) &req, sizeof(sdm_reserveReqMsg_t));
	req.hdr.mid = SDM_MID_DEV_RESERVE_REQ;
    //req.hdr.chandle = getpid();
	req.handle = lun->openHandle;
	req.action = SDM_DEV_RSV_ACTION_RELEASE;
	req.key = SDM_DEV_RSV_KEY_DEFAULT; // need my own key???
	
	rc = sendAgileMsg(-1, &req, sizeof(sdm_reserveReqMsg_t),
                      "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &sdm_to, &rsp, sizeof(sdm_reserveReqMsg_t));
    if (rc == sizeof(sdm_reserveRspMsg_t))
    {
        rc = rsp.rCode;
    }
    else
    {
        rc = NFX_ERR;
    }
    
    return(rc);
}
#else
{
    return(NFX_OK);
}
#endif // ISABELLE

/*----------------------------------------------------------------------------*/
