/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm.h
 *
 * Description  : Volume Manager (EVM) - common header file
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 24-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

#ifndef _EVM_H_
#define _EVM_H_

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#ifdef EFENCE
#include "ef-fence.h"
#endif // EFENCE

#include "nfx-incl.h"
#include "exports/event-api.h"
#include "exports/eee-desc.h"
#include "exports/scsi-api.h"
#include "exports/sdm-api.h"
#include "exports/elog-api.h"
#include "exports/anpssc-api.h"
#include "exports/ncm-api.h"
#include "exports/spm-api.h"
#include "evm-api.h"
#include "evm_loc_defs.h"
#include "evm-takeover.h"

/*------------------------------------------------------------------------------
 * Type definitions.
 *------------------------------------------------------------------------------
 */

/* Basic message structure.
 */
typedef struct
{
    char    buf[EVM_MAX_BUF_SIZE];
    uint32  len;
} evm_msg_t;

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* EVM application id.
 */
extern int32 evmAppId;

/* Node name.
 */
extern char myNode[EVM_MAX_NODE_NAME];
extern char myCluster[EVM_MAX_NODE_NAME];
extern evm_nfxId_t myNfx;

/* EVM statistics.
 */
extern uint64   evmNbAlloc;
extern uint64   evmNbFree;

/*------------------------------------------------------------------------------
 * Macros.
 *------------------------------------------------------------------------------
 */

#ifdef EFENCE
#define malloc ef_malloc
#define calloc ef_calloc
#define free ef_free
#endif // EFENCE

#ifdef NO_SUCH_DEF
#define EVM_DBG_PRINT(_par_) printf _par_; printf("\n")
#define EVM_ERR_PRINT(_par_) printf _par_; printf("\n")
#define EVM_FCT_PRINT(_par_) printf _par_; printf("\n")
#else
#define EVM_DBG_PRINT(_par_) E_LOG(class_1, LOG_DEBUG, evmAppId, 0, 0, 0, _par_)
#define EVM_ERR_PRINT(_par_) E_LOG(class_1, LOG_ERR  , evmAppId, 0, 0, 0, _par_)
#define EVM_FCT_PRINT(_par_) E_LOG(class_1, LOG_INFO , evmAppId, 0, 0, 0, _par_)
#define EVM_ALR_PRINT(_par_) E_LOG(class_1, LOG_ALERT, evmAppId, 0, 0, 0, _par_)
#endif // NO_SUCH_DEF

#define EVM_FREE(_p_)                               \
    if (_p_ != NULL)                                \
    {                                               \
        free(_p_);                                  \
        _p_ = NULL;                                 \
        evmNbFree++;                                \
    }

#define EVM_ALLOC(_p_, _type_, _size_)              \
    _p_ = (_type_ *) malloc(_size_);                \
    if (_p_ == NULL)                                \
    {                                               \
        EVM_ERR_PRINT(("malloc() failed"));         \
        NFX_ASSERT(TRUE);                           \
    }                                               \
    else                                            \
    {                                               \
        evmNbAlloc++;                               \
    }

#define EVM_ALLOC_ZERO(_p_, _type_, _size_)         \
    _p_ = (_type_ *) calloc(1, _size_);             \
    if (_p_ == NULL)                                \
    {                                               \
        EVM_ERR_PRINT(("calloc() failed"));         \
        NFX_ASSERT(TRUE);                           \
    }                                               \
    else                                            \
    {                                               \
        evmNbAlloc++;                               \
    }

#define EVM_PRINT_STATS(_text_)                                 \
    printf("*** EVM statistics %s.\n", _text_);                 \
    printf("\tevmNbAlloc ............ %10lld\n", evmNbAlloc);   \
    printf("\tevmNbFree ............. %10lld\n", evmNbFree);    \
    printf("\talloc - free .......... %10lld\n",                \
           evmNbAlloc - evmNbFree);

// Initialize the on-disk configuration.
#define EVM_OD_CFG_INIT(_odCfg)         \
    _odCfg.magic = NFX_STRG_MAGIC;      \
    _odCfg.version = EVM_CFG_VERSION;   \
    strcpy(_odCfg.cluster, myCluster);

// Get a new volume id.
#define EVM_GET_NEW_VOLID(_id_)                             \
    _id_.whole_thing = 0;                                   \
    _id_.details.u.parts.clusterUnique = time(NULL) / 5;    \
    _id_.details.worldUnique = gethostid();

// Get the time in seconds.
#define EVM_GET_TIME(_sec_)     \
{                               \
    struct timeval tv;          \
    gettimeofday(&tv, NULL);    \
    _sec_ = tv.tv_sec;          \
}

/*------------------------------------------------------------------------------
 * Global function prototypes.
 *------------------------------------------------------------------------------
 */

/* Defined in evm_cfgd_rx.c
 * ------------------------
 */
int
check_lvol(evm_VolConfig_t *volCfg, evm_volStrg_t *volStrg);

int
create_lvol (evm_VolConfig_t *  cfg,
             boolean            allowDuplicate,
             char *             errstr);

int32
evm_onlineVol(evm_volId_t   volId,
              uint32        nim,
              char *        err);

int
upd_evm_srvr (evm_strgConfig_t *au, evm_VolConfig_t *cfg, char *err);

void
evm_processMsg(char *       msgbuf,
               char *       fromapp,
               uint32       fromcpu,
               uint32       fromslot);

void
evm_tkoCallback(evm_volId_t   volId,
                uint32        status,
                uint64        clientPrivate);

void
evm_onCallback(evm_volId_t  volId,
               uint32       status,
               uint64       clientPrivate);

/* Defined in evm_db.c
 * -------------------
 */
int upd_od_lun_cfg(scsi_id_t      id,
                   evm_StrgAttr_t *sAttr,
                   int            delSpool,
                   int            delVol);

int upd_od_vol_cfg(evm_VolConfig_t *volConfig,
                   evm_volStrg_t   *volStrg);

int upd_db_for_lun(scsi_id_t id, char *err);

int
upd_vol_from_disk(scsi_id_t         id,
                  evm_StrgAttr_t *  sAttr,
                  evm_strgCfg_t *   odCfg,
                  boolean           lunRescan,
                  char *            err);

int init_dbs(void);

void evm_clearLun(void);

int close_dbs(void);

int32
evm_processNewLun(sdm_devInfoRspMsg_t * lunInfo,
                  char *                err);

void 
evm_initLun(sdm_devInfoRspMsg_t * lunInfo,
            evm_StrgAttr_t *      newLun);

int
upd_vol_for_lun(scsi_id_t           id,
                evm_StrgAttr_t *    sAttr,
                evm_VolConfig_t *   cfg);

int32
evm_addLun(evm_StrgAttr_t * newLun,
           boolean          tellSPM,
           char *           err);

void
display_od_cfg(evm_strgCfg_t *  odCfg);

/* Defined in evm_event.c
 * ----------------------
 */

void
evm_sendEvt(uint32      event,
            evm_volId_t volId,
            int			state);

/* Defined in evm_lvdb.c
 * ---------------------
 */

int
init_lvdb(void);

int
close_lvdb(void);

evm_volId_t
next_lvol_id(void);

int
init_lv_cfg(void);

int
upd_lv_cfg(evm_VolConfig_t *cfg, int upd_od, char *err);

int
add_lv_cfg(evm_VolConfig_t *    cfg,
           boolean              allowDuplicate,
           char *               err);

int
get_lv_cfg_byid(evm_volId_t id, evm_VolConfig_t *cfg, char *err);

int
del_lv_cfg(evm_volId_t id, char *err);

int
get_lv_ids(int max, evm_volId_t *ids, char *err);

int
evm_getVolIdsAndNames(int           max,
                      evm_volId_t * ids,
                      char          names[][VOL_MAX_NAME],
                      char *        err);

int
get_lv_id_byname (char *name, evm_volId_t *id, char *err);

int
get_lv_strg(evm_volId_t         id,
            evm_volStrg_t **    strg,
            char *              err);

int
upd_lv_strg (evm_volId_t id, evm_volStrg_t *strg, char *err);

int
get_spool_lvs (int32 spool, evm_volId_t *ids, char *err);

int
look_for_spool (int32 spid, char *err);

/* Defined in evm_pstrgdb.c
 * ------------------------
 */

int
add_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t *  cfg,
              char *            err);

void
upd_pstrgdb(void *  msg,
            boolean add);

int
init_pstrgdb(void);

int
close_pstrgdb(void);

int
get_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t ** cfg,
              char *            err);

int
upd_pstrg_cfg (scsi_id_t id, evm_StrgAttr_t *cfg, int32 upd_od, char *err);

int
get_pstrg_id (char *name, scsi_id_t *id, char *err);

int
get_pstrg_list (evm_MsgStrgList_t *lst, char *err);

int32
evm_openLun(scsi_id_t   id,
            uint32      slot,
            int32 *     handle);

int32
evm_closeLun(int32  handle);

int32
evm_closeAllLun(void);

int
del_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t *  cfg,
              char *            err);

int32
evm_getLunInfo(scsi_id_t        lunId,
               evm_StrgAttr_t * lunInfo,
               char *           err);

int32
evm_labelLun(scsi_id_t  lunId,
             char *     clusterName,
             char *     err);

int32
evm_unlabelLun(scsi_id_t    lunId,
               char *       err);

int32
evm_modLunCluster(scsi_id_t lunId,
                  char *    clusterName,
                  char *    err);

/* Defined in evm_reg.c
 * --------------------
 */

int32
evm_regGetCluster(void);

int32
evm_getNewVolId(evm_volId_t *   newVolId,
                char *          err);

boolean
evm_isVolNameUnique(char *  volName);

/* Defined in evm-sdm.c
 * --------------------
 */

int32
evm_readLunLabel(scsi_id_t        id,
                 evm_StrgAttr_t * sAttr,
                 evm_strgCfg_t  * cfgp);

int32
evm_writeLunLabel(scsi_id_t        id,
                  evm_StrgAttr_t   *sAttr,
                  evm_strgCfg_t    *odCfg);

int32
evm_readMgrArea(evm_StrgAttr_t *        lun,
                sdm_readDataRspMsg_t ** readRsp);

int32
evm_writeMgrArea(evm_StrgAttr_t *   lun,
                 evm_mgrArea_t *    mgrArea);

int32
evm_readTkoArea(evm_StrgAttr_t *        lun,
                sdm_readDataRspMsg_t ** readRsp);

int32
evm_writeTkoArea(evm_StrgAttr_t *   lun,
                 evm_tkoArea_t *    tkoArea);

int32
evm_readQuorumArea(evm_StrgAttr_t *        lun,
                   sdm_readDataRspMsg_t ** readRsp);

int32
evm_writeQuorumArea(evm_StrgAttr_t *    lun,
                    evm_qArea_t *       qArea);

int32
evm_readMirrorArea(evm_StrgAttr_t *        lun,
                   sdm_readDataRspMsg_t ** readRsp);

int32
evm_writeMirrorArea(evm_StrgAttr_t *    lun,
                    evm_mirrorArea_t *  mArea);

int32
evm_lockLun(evm_StrgAttr_t * lun);

int32
evm_releaseLun(evm_StrgAttr_t * lun);

/*----------------------------------------------------------------------------*/
#endif // _EVM_H_
