/*------------------------------------------------------------------------------
 * RCS:  $Id: evm_cfgd.c,v 1.33 2002/09/04 17:53:22 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm_cfgd.c
 *
 * Description  : Controls EVM and maintains its volume configuration
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 30-APR-2001
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "nfx-incl.h"

#include "exports/eee-desc.h"
#include "exports/scsi-api.h"
#include "exports/anpssc-api.h"
#include "exports/sdm-api.h"
#include "exports/elog-api.h"

#include "evm.h"
#include "evm-api.h"
#include "evm_loc_defs.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* EVM application id.
 */
int32 evmAppId;

/* Node and cluster name.
 */
char myNode[EVM_MAX_NODE_NAME];
char myCluster[EVM_MAX_NODE_NAME];
evm_nfxId_t myNfx;

int     evm_debug = 0;

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static void terminate_me(int sig);

/*------------------------------------------------------------------------------
 * Name         : terminate_me
 *
 * Description  : This handles the SIGTERM to ensure the config is
 *                synced other stuff.
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 11-JUL-2001
 *
 *------------------------------------------------------------------------------
 */

static void
terminate_me (int sig)

{
    EVM_DBG_PRINT(("Terminating EVM..."));

    close_dbs();
       
	EVM_DBG_PRINT(("evm_cfgd exits."));
	exit(0);
}

/*------------------------------------------------------------------------------
 * Name         : evm_cfg main
 *
 * Description  : This is where EVM's control processes get fire up.
 *                It will fork() and execve() the rest of the 
 *                processes. 
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 01-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(int        argc,
     char **    argv,
     char **    envp)
    
{
	int evm_in;				/* EVM's input messages */
	sigset_t set;

    EVM_FCT_PRINT(("EVM daemon is (re-)starting..."));

    /* Block all but the signals we'll be using
     */
	sigfillset(&set);
	sigdelset(&set, SIGALRM);
	sigdelset(&set, SIGTERM);
	sigdelset(&set, SIGCHLD);
	sigdelset(&set, SIGSEGV);
	sigdelset(&set, SIGBUS);
	sigdelset(&set, SIGTRAP);
	sigprocmask(SIG_SETMASK, &set, NULL);

    /* Set up our termination handler
     */
	signal (SIGTERM, terminate_me);
    /* Also, let's keep track of lvmgr
     */
    signal (SIGCHLD, terminate_me);
    
    /* Register us to be receiving the messages for EVM.
     */
	evm_in = regAgileApp ("evm");
    evmAppId = eee_getApplicationId("evm");
	if (evm_in < 0)
    {
		EVM_ERR_PRINT(("eee_getApplicationId(evm) failed"));
		EVM_ERR_PRINT(("evm_cfgd exits."));
		exit(1);
	}

    /* Get the name of the node where I am running.
     */
    if (gethostname(myNode, EVM_MAX_NODE_NAME) != 0)
    {
		EVM_ERR_PRINT(("gethostname(%s, %d) failed",
                       myNode, EVM_MAX_NODE_NAME));
		EVM_ERR_PRINT(("evm_cfgd exits."));
		exit(1);
    }

    /* Get cluster name.
     */
    evm_regGetCluster();

    /* Set NFX identifier.
     */
    strcpy(myNfx.cluster, myCluster);
    strcpy(myNfx.node, myNode);
    myNfx.chassisId = gethostid();
    EVM_DBG_PRINT(("EVM: myNfx = %s/%s/%d",
                   myNfx.cluster, myNfx.node, myNfx.chassisId));
        
    /* Reset all LUNs seen by this node if requested.
     */
    if ((argc >= 2) && (!strcmp(argv[1], "letsresetandagilizeall")))
    {
        evm_clearLun();
        printf("Done.\n");
        exit(0);
    }
    
    /* Open up Volume configuration databases
     */
	if (init_dbs())
    {
		EVM_ERR_PRINT(("EVM failed to initialize its databases."));
		EVM_ERR_PRINT(("evm_cfgd exits."));
		exit(1);
    }

    /* Initialize the takeover sub-module.
     */
    evm_tk_init();
    
    /* It's time to daemonize and let the rest of the world continue
     */
	sigemptyset(&set);
	sigaddset(&set, SIGINT);
	sigprocmask(SIG_UNBLOCK, &set, NULL);
	signal(SIGINT, terminate_me);
    
	/* Tell MCP we're done with our init.
	 */
	kill (getppid(), SIGUSR1);

#ifdef ISABELLE // debug only!!!
        EVM_PRINT_STATS("after EVM initialization");
#endif // ISABELLE
    
    /* Read the initial configuration here on the SSC and send the
     * messages to configure its Srvr(s).
     */
	init_lv_cfg();

    {
        time_t          now = time(NULL);
        time_t          next = now + EVM_TK_TICK;
        struct timeval  clockTick = {0, 0};
        int             rc = 0;
        char            msgbuf[EVM_MAX_BUF_SIZE];
        char            fromapp[EVM_MAX_APP_STR];
        uint32          fromcpu;
        uint32          fromslot;
        
        while(1)
        {
            clockTick.tv_sec = next - now;
            //EVM_DBG_PRINT(("call recvAgileMsg() with TO = %d", clockTick.tv_sec));
            rc = recvAgileMsg(evm_in, msgbuf, sizeof(msgbuf), &clockTick,
                              fromapp, &fromcpu, &fromslot, 0);
            if (rc < 0)
            {
                // If this happens, we're screwed
                EVM_ERR_PRINT(("process_messages: rcv error %s", strerror(errno)));
                return 1;
            }
            else if (rc > 0)
            {
                evm_processMsg(msgbuf, fromapp, fromcpu, fromslot);
            }
            now = time(NULL);
            if (now >= next)
            {
                evm_tk_clockTick();
                next = now + EVM_TK_TICK;
            }
        } // while(1)
    }
    
    EVM_ERR_PRINT(("Unexpected return from process_messages"));
	return 1; /* Theory says we shouldn't get here */
}

/*----------------------------------------------------------------------------*/
