/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 * RCS          : $Id: evm_cfgd_rx.c,v 1.140 2002/12/05 18:52:32 isabellec Exp $
 *
 * Name         : evm_cfgd_rx.c
 *
 * Description  : The main action of evm_cfgd.
 *
 * Created by   : Greg Seibert
 * Modified by  : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 01-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

#include "evm.h"
#include "exports/fs-evm.h"
#include "exports/opt-api.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* EVM statistics.
 */
uint64   evmNbAlloc = 0;
uint64   evmNbFree = 0;

static struct timeval  waitSpmTo = {30, 0};

/*------------------------------------------------------------------------------
 * Type definitions.
 *------------------------------------------------------------------------------
 */

typedef struct
{
    uint64          clientPrivate;
    char            fromapp[EVM_MAX_APP_STR];
	uint32          fromcpu;
	uint32          fromslot;
    uint32          nim;
} evm_tkoCb_t;

typedef struct
{
    uint32          action;
    uint64			clientPrivate;
    evm_volId_t     volId;
    boolean         online;
    uint32          nim;
    char            fromapp[EVM_MAX_APP_STR];
    uint32          fromcpu;
    uint32          fromslot;
} evm_growPendingCb_t;

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static int
check_volStrg(evm_VolConfig_t * volCfg,
              evm_volStrg_t *   volStrg);

static boolean
evm_isVolOk(evm_VolConfig_t *   volCfg,
            char *              err);

static int
send_vol_cfg(evm_volId_t    id,
             uint64         priv,
             char *         app, 
             uint32         cpu,
             uint32         slot,
             char *         err);

static int
send_volName_cfg(uint64         priv,
                 char *         app, 
                 uint32         cpu,
                 uint32         slot,
                 char *         err);

static int32
evm_growVol(evm_volId_t     volId,
            uint64          fsysSize,
            uint32          nbLun,
            char            lunList[][SDM_DEV_NAME_LEN+8],
            uint64          userInfo,
            char *          err);

static int32
evm_addLunToVol(evm_VolConfig_t *       volCfg,
                spm_MsgAllocLunRsp_t *  spmRsp,
                char *                  err);

static void
evm_delLun(sdm_devDelReqMsg_t *  sdm_msg,
           char *                err);
static int32
evm_delVol(evm_volId_t  volId,
           char *       err);

static int32
evm_offlineVol(evm_volId_t  volId,
               boolean      tellEvmSrvr,
               char *       err);

static int32
evm_removeLunFromVolume(scsi_id_t   lunId,
                        evm_volId_t volId,
                        char *      err);

static void
evm_rescanLun(scsi_id_t lunId,
              char *    err);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : check_volStrg
 *
 * Description  : perform basic sanity checks for this logical volume
 *                config data from disk before bringing on line
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

static int
check_volStrg(evm_VolConfig_t * volCfg,
              evm_volStrg_t *   volStrg)
    
{
    int32   a;
    uint64  currSize=0;

    /* Verify all the LUN pieces are there
     */
    for (a=0; a<volStrg->num; a++)
    {
        currSize += volStrg->au[a].auSize;
        if (volStrg->au[a].strgId == 0)
        {
            EVM_DBG_PRINT(("check_lvol(): missing LUN[%d] for volId 0x%llx\n",
                           a, volCfg->id.whole_thing));
            return NFX_ERR;
        }
        if (volStrg->au[a].openHandle == -1)
        {
            EVM_DBG_PRINT(("check_lvol(): LUN[%d] 0x%llx "
                           "not open for volId 0x%llx\n",
                           a, volStrg->au[a].strgId, volCfg->id.whole_thing));
            return NFX_ERR;
        }
    } // for
    
    if (currSize != volCfg->currSize)
    {
        EVM_DBG_PRINT(("check_lvol(): currSize %d != volCfg->currSize %d\n",
                       currSize, volCfg->currSize));
        return NFX_ERR;
    }
    
    return NFX_OK;
}

#ifdef ISABELLE // code to delete
/*------------------------------------------------------------------------------
 * Name         : chk_lv_cfg
 *
 * Description  : Checks the configuration record of a logical volume
 *                for reasonableness and also sets defaults for invariants
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 03-JUL-2001
 *
 *------------------------------------------------------------------------------
 */

static int
chk_lv_cfg(evm_VolConfig_t *    cfg,
           char *               errstr)

{
    cfg->numMirrors = cfg->numStripes = 1;
	return NFX_OK;
}
#endif // ISABELLE

/*------------------------------------------------------------------------------
 * Name         : send_vol_cfg
 *
 * Description  : Responds to requests for LV configs
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

static int
send_vol_cfg(evm_volId_t    id,
             uint64         priv,
             char *         app, 
             uint32         cpu,
             uint32         slot,
             char *         err)
    
{
    int             rc = NFX_OK;
    evm_volStrg_t * strg = NULL;
    
    /* The id determines which message is returned.
     */
	if (id.whole_thing == -1)   // Want the list
    {
        evm_MsgVolList_t    vol_list;
        int                 num;
        
		err[0] = 0;
		num = get_lv_ids(MAX_IDS, vol_list.ids, err);
		if (err[0]) // Got error along the way
        {
			rc = NFX_ERR;
            goto send_vol_cfg_exit;
		}
		vol_list.hdr.msgType = EVM_VOL_LIST;
		vol_list.hdr.len = sizeof(vol_list);
		vol_list.hdr.clientPrivate = priv;
		vol_list.num = num;
		
		if (sendAgileMsg(-1, &vol_list, sizeof(vol_list), 
                         app, cpu, slot, MGMTPLANE, NULL, NULL, 0) < 0)
        {
			sprintf (err, "error sending list %s", strerror(errno));
			rc = NFX_ERR;
            goto send_vol_cfg_exit;
		}
	}
	else    // Just this one cfg
    {
        evm_MsgVolCreate_t  cfgmsg;
        
        /* Seems kind of wonky to use the CREATE message to return the config
         * but it's all there. We do define another message type though.
         */
		cfgmsg.hdr.msgType = EVM_VOL_CONFIG;
		cfgmsg.hdr.len = sizeof(cfgmsg);
		cfgmsg.hdr.clientPrivate = priv;

        rc = get_lv_cfg_byid (id, &cfgmsg.config, err);
		if (rc != NFX_OK)
        {
            goto send_vol_cfg_exit;
		}
#ifdef ISABELLE // debug only!!!
        printf("cfgmsg.config.minAutoGrow = %lld\n", cfgmsg.config.minAutoGrow);
#endif // ISABELLE

        /* Check if all the LUN pieces are there for this vol.
         */
        rc = get_lv_strg(cfgmsg.config.id, &strg, err);
		if ((rc != NFX_OK) || (strg == NULL))
        {
            goto send_vol_cfg_exit;
        }

        rc = check_lvol(&cfgmsg.config, strg);
        if (rc == NFX_OK)
        {
            cfgmsg.state = EVM_VOL_READY;
        }
        else
        {
            cfgmsg.state = EVM_VOL_NOT_READY;
        }
		if (sendAgileMsg(-1, &cfgmsg, sizeof(cfgmsg), 
                         app, cpu, slot, MGMTPLANE, NULL, NULL, 0) < 0)
        {
			sprintf (err, "error sending list %s", strerror(errno));
			rc = NFX_ERR;
            goto send_vol_cfg_exit;
		}
	}
  send_vol_cfg_exit:

    EVM_FREE(strg);
	return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : send_volName_cfg
 *
 * Description  : Send the list of volume ids and names.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static int
send_volName_cfg(uint64         priv,
                 char *         app, 
                 uint32         cpu,
                 uint32         slot,
                 char *         err)
    
{
    int                     rc = NFX_OK;
    int                     num = 0, i;
    evm_volId_t             ids[MAX_IDS];
    char                    names[MAX_IDS][VOL_MAX_NAME];
    int                     rspSize;
    evm_MsgVolNameList_t *  rsp = NULL;
        
        
    err[0] = 0;
    num = evm_getVolIdsAndNames(MAX_IDS, ids, names, err);
    if (err[0]) // Got error along the way
    {
        rc = NFX_ERR;
        goto send_volName_cfg_exit;
    }

    rspSize = sizeof(evm_MsgVolNameList_t) +
        ((num - 1) * (sizeof(evm_volId_t) + VOL_MAX_NAME * sizeof(char)));
    EVM_ALLOC_ZERO(rsp, evm_MsgVolNameList_t, rspSize);
#ifdef ISABELLE // debug only!!!
    printf("*** rspSize = %d\n", rspSize);
#endif // ISABELLE

    rsp->hdr.msgType = EVM_VOL_NAME_LIST;
    rsp->hdr.len = rspSize;
    rsp->hdr.clientPrivate = priv;
    rsp->num = num;
    for (i = 0; i < num; i++)
    {
        rsp->s[i].id = ids[i];
        strcpy(rsp->s[i].name, names[i]);
    } // for
		
    if (sendAgileMsg(-1, rsp, rspSize, 
                     app, cpu, slot, MGMTPLANE, NULL, NULL, 0) < 0)
    {
        sprintf (err, "error sending list %s", strerror(errno));
        rc = NFX_ERR;
        goto send_volName_cfg_exit;
    }
	
  send_volName_cfg_exit:
    
    EVM_FREE(rsp);
	return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_growVol
 *
 * Description  : We've been told this logical volume has reached its
 *                high water ratio so we need to allocate more storage out
 *                of its storage pool and assign it.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 20-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_growVol(evm_volId_t     volId,
            uint64          fsysSize,
            uint32          nbLun,
            char            lunList[][SDM_DEV_NAME_LEN+8],
            uint64          userInfo,
            char *          err)
    
{
    int                     rc = NFX_OK;
	evm_VolConfig_t         volCfg;
    evm_msg_t               spmMsg;
    spm_MsgAllocLunRsp_t *  spmRsp = NULL;

    EVM_FCT_PRINT(("evm_growVol(0x%llx, %lld)", volId.whole_thing, fsysSize));
    
    /* Get the volume configuration.
     */
    rc = get_lv_cfg_byid(volId, &volCfg, err);
    if (rc != NFX_OK)
    {
        goto evm_growVol_exit;
    }

    /* Check if the volume really need to grow.
     * It may happened that the filesystem is not using the whole
     * allocated space of the volume.
     */
    if (fsysSize != 0)
    {
        if (volCfg.currSize >= (fsysSize + volCfg.minAutoGrow))
        {
            sprintf(err, "Volume %s (0x%llx) already grown.",
                    volCfg.name, volId.whole_thing);
            EVM_DBG_PRINT(("%s", err));
            rc = NFX_OK;
            goto evm_growVol_exit;            
        }
    }
        
    /* Check if this volume hasn't reached its maximum size.
     */
    if (volCfg.maxSize != 0)
    {
        if (volCfg.maxSize - volCfg.currSize <= 0)
        {
            sprintf(err, "Volume %s (0x%llx) already at maximum allowed size.",
                    volCfg.name, volId.whole_thing);
            rc = NFX_ERR; // Already at maximum growth potential
            goto evm_growVol_exit;
        }
        else if ((volCfg.currSize + volCfg.minAutoGrow) > volCfg.maxSize)
        {
            sprintf(err, "Cannot grow volume %s (0x%llx) by min and "
                    "still not exceed max.", volCfg.name, volId.whole_thing);
            rc = NFX_ERR;	// Would exceed maximum size
            goto evm_growVol_exit;
        }
    }

    {
        /* Get LUNs from SPM.
         */
        ssize_t                 ncmReqSize;
        ncmMsg_t *              ncmReq = NULL;
        ncmMsg_t *              ncmRsp = (ncmMsg_t *) spmMsg.buf;
        spm_MsgAllocLunReq_t *  spmReq = NULL;

        bzero((void *) &spmMsg, sizeof(evm_msg_t));
        spmRsp = (spm_MsgAllocLunRsp_t *) ncmRsp->appMsg;
    
        /* Build NCM message.
         */
        ncmReqSize = sizeof(ncmMsg_t) + sizeof(spm_MsgAllocLunReq_t) - 1;
        EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
        NFX_ASSERT(ncmReq != NULL);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->appName, "spm");
        strcpy(ncmReq->nodeName, "PCC");
        ncmReq->rqResp = 1;    // response expected
        spmReq = (spm_MsgAllocLunReq_t *) ncmReq->appMsg;
        
        /* Build SPM message.
         */
        spmReq->hdr.msgType = SPM_ALLOC_LUN_REQ;
        spmReq->hdr.clientPrivate = userInfo;
        strcpy(spmReq->hdr.node, myNode);
        spmReq->minSize = volCfg.minAutoGrow;
        memcpy((void *) &(spmReq->attr), (const void *) &(volCfg.attrs),
               sizeof(evm_strg_attributes_t));
        spmReq->volId = volId;
        spmReq->nbLun = nbLun;
        if (nbLun != 0)
        {
            int i;
            
            for (i = 0; i < nbLun; i++)
            {
                strcpy(spmReq->lunList[i], lunList[i]);
            } // for
        }    

        spmMsg.len = sendAgileMsg(-1, ncmReq, ncmReqSize,
                                  "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                  &waitSpmTo,
                                  spmMsg.buf, sizeof(spmMsg.buf));
        NFX_ASSERT(spmMsg.len > sizeof(ncmMsg_t));
        NFX_ASSERT(ncmRsp->msgType == NCM_RESP);
        NFX_ASSERT(spmRsp->hdr.msgType == SPM_ALLOC_LUN_RSP);
        EVM_FREE(ncmReq);
    }
    
    if (spmRsp->result == SPM_ALLOC_OK)
    {
        rc = evm_addLunToVol(&volCfg, spmRsp, err);
    }
    else if (spmRsp->result == SPM_ALLOC_PENDING)
    {
        sprintf(err, "Volume %s (0x%llx) cannot grow immediately. "
                "LUN(s) allocation pending...", volCfg.name, volId.whole_thing);
        rc = EVM_VOL_GROW_PENDING;
    }
    else // SPM_ALLOC_FAILED
    {
        sprintf(err, "No more LUNs. Volume %s (0x%llx) cannot grow.",
                volCfg.name, volId.whole_thing);
        rc = NFX_ERR;
    }

  evm_growVol_exit:
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_addLunToVol
 *
 * Description  : Grow a volume by adding LUNs to it.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-AUG-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_addLunToVol(evm_VolConfig_t *       volCfg,
                spm_MsgAllocLunRsp_t *  spmRsp,
                char *                  err)
    
{
    int                 rc = NFX_OK;
    int                 i;
    uint64              sizeBeforeGrowth = volCfg->currSize;
    uint64              threshold;
	evm_volStrg_t *     lvstrg = NULL;
	evm_volStrg_t *     newLvstrg = NULL;
	evm_StrgAttr_t *    pattr = NULL;
    // List of open handles
    typedef struct oh_T oh_t;
    struct oh_T
    {
        int32   handle;     // open handle
        oh_t *  next;
    };
    oh_t * ohList = NULL;
    oh_t * oh = NULL;
   
    EVM_DBG_PRINT(("Got %d LUNs (0x%llx)", spmRsp->nbLun, spmRsp->id[0]));

    /* Go through all the received LUNs.
     */
    for (i = 0; i < spmRsp->nbLun; i++)
    {
        evm_StrgAttr_t      newLun;
        evm_strgConfig_t    au;
        evm_strgCfg_t       odCfg;
        evm_mgrArea_t       mgrArea;
        evm_tkoArea_t       tkoArea;

        /* Get LUNs currently composing the volume.
         */
        rc = get_lv_strg(volCfg->id, &lvstrg, err);
        if ((rc != NFX_OK) || (lvstrg == NULL))
        {
            goto evm_addLunToVol_exit;
        }
            
        /* Get detailed info for this LUN.
         */
        rc = evm_getLunInfo(spmRsp->id[i], &newLun, err);
        if (rc != NFX_OK)
        {
            EVM_ERR_PRINT(("EVM: %s", err));
            goto evm_addLunToVol_exit;
        }
            
        /* Open the LUN.
         */
        rc = evm_openLun(newLun.id, SSC_SLOT, &(newLun.openHandle));
        if ((rc != NFX_OK) || (newLun.openHandle == -1))
        {
            sprintf(err, "cannot open LUN 0x%llx", newLun.id);
            EVM_ERR_PRINT(("EVM: %s", err));
            goto evm_addLunToVol_exit;
        }
        EVM_ALLOC(oh, oh_t, sizeof(oh_t));
        oh->handle = newLun.openHandle;
        oh->next = ohList;
        ohList = oh;

        /* Read LUN label and make sure it's still a free LUN.
         */
        rc = evm_readLunLabel(newLun.id, &newLun, &odCfg);
        if (rc != NFX_OK)
        {
            sprintf(err, "cannot read LUN label (0x%llx)", newLun.id);
            EVM_ERR_PRINT(("EVM: %s", err));
            goto evm_addLunToVol_exit;
        }

        if ((odCfg.magic != NFX_STRG_MAGIC) ||
            (odCfg.id.whole_thing != 0) ||
            (strcmp(odCfg.cluster, myCluster)))
        {
            sprintf(err, "LUN 0x%llx is not free! "
                    "magic = %s, volId = 0x%llx, cluster = %s", newLun.id,
                    (odCfg.magic == NFX_STRG_MAGIC) ? "ok":"wrong",
                    odCfg.id.whole_thing, odCfg.cluster);
            EVM_ERR_PRINT(("EVM: %s", err));
            rc = NFX_ERR;
            goto evm_addLunToVol_exit;
        }

        /* Add LUN to local LUN db.
         */
        newLun.nallocs = 1;
        newLun.allocs[0].lvid = volCfg->id;
        newLun.allocs[0].stBlk = 0;                // ???
        newLun.allocs[0].nBlks = newLun.blkCnt;    // ???
        rc = add_pstrg_cfg(newLun.id, &newLun, err);
        if (rc != NFX_OK)
        {
            EVM_ERR_PRINT(("EVM: %s", err));
            goto evm_addLunToVol_exit;
        }

        /* Initialize the physical storage allocation.
         */
        au.strgId = newLun.id;
        au.openHandle = newLun.openHandle;
        au.startBlk = 0; // ???
        au.auSize = newLun.blkCnt;
        au.totalSize = newLun.totalBlkCnt;

        /* Update the volume in the database.
         */            
        EVM_ALLOC_ZERO(newLvstrg, evm_volStrg_t,
                       sizeof(evm_volStrg_t) +
                       (lvstrg->num * sizeof(evm_strgConfig_t)));
        memcpy((void *) newLvstrg, (const void *) lvstrg,
               sizeof(evm_volStrg_t) +
               (((lvstrg->num==0)?0:(lvstrg->num - 1)) *
                sizeof(evm_strgConfig_t)));
        newLvstrg->au[newLvstrg->num] = au;
        newLvstrg->num++;

        if (volCfg->currSize == 0)
        {
            /* First grow after create.
             * Refine the LUNs attributes.
             */
            if (volCfg->attrs.flags & EVM_SAME_VENDOR)
            {
                /* Grow using LUNs from same vendors.
                 */
                if ((volCfg->attrs.num == 0) ||
                    !strcmp(volCfg->attrs.vendorInfo[0].vendor, ""))
                {
                    volCfg->attrs.num = 1;
                    strcpy(volCfg->attrs.vendorInfo[0].vendor,
                           newLun.vendor);
                    if (volCfg->attrs.flags & EVM_SAME_MODEL)
                    {
                        /* Grow using LUNs from same vendor/model.
                         */
                        strcpy(volCfg->attrs.vendorInfo[0].model,
                               newLun.model);
                    }
                    else
                    {
                        strcpy(volCfg->attrs.vendorInfo[0].model, "");
                    }
                }
                else if (volCfg->attrs.num > 1)
                {
                    volCfg->attrs.num = 1;
                    strcpy(volCfg->attrs.vendorInfo[0].vendor,
                           newLun.vendor);
                    if (volCfg->attrs.flags & EVM_SAME_MODEL)
                    {
                        /* Grow using LUNs from same vendor/model.
                         */
                        strcpy(volCfg->attrs.vendorInfo[0].model,
                               newLun.model);
                    }
                    else
                    {
                        strcpy(volCfg->attrs.vendorInfo[0].model, "");
                    }
                }
            }
                
            /* Grow using LUNs from same RAID controller.
             */
            if (volCfg->attrs.flags & EVM_SAME_RAIDCTRL)
            {
                strcpy(volCfg->attrs.raidCtrl, newLun.name);
                *rindex(volCfg->attrs.raidCtrl, '_') = 0;
                EVM_DBG_PRINT(("RAID controller %s",
                               volCfg->attrs.raidCtrl));
            }
                
            /* Grow using LUNs with same RAID level.
             */
            volCfg->attrs.raidLevel = newLun.raidLevel;
        }
        //volCfg->currSize += (au.auSize * volCfg->numStripes);
        volCfg->currSize += au.auSize;

        /* Clean manager area.
         */
        bzero((void *) &mgrArea, sizeof(evm_mgrArea_t));
        mgrArea.magic = NFX_STRG_MAGIC;
        rc = evm_writeMgrArea(&newLun, &mgrArea);
        if  (rc != NFX_OK)
        {
            goto evm_addLunToVol_exit;
        }
    
        /* Clean takeover area.
         */
        bzero((void *) &tkoArea, sizeof(evm_tkoArea_t));
        tkoArea.magic = NFX_STRG_MAGIC;
        rc = evm_writeTkoArea(&newLun, &tkoArea);
        if  (rc != NFX_OK)
        {
            goto evm_addLunToVol_exit;
        }
                                       
        /* Tell EVM server about the new allocated LUN.
         */
        if (upd_evm_srvr(&au, volCfg, err) != NFX_OK)
        {
            /* Back out failed allocation from storage DB
             */
            rc = get_pstrg_cfg(au.strgId, &pattr, err);
            if ((rc != NFX_OK) || (pattr != NULL))
            {
                goto evm_addLunToVol_exit;
            }
            pattr->nallocs--;
            bzero(&(pattr->allocs[pattr->nallocs - 1]), sizeof(alloc_t));
            rc = upd_pstrg_cfg(au.strgId, pattr, 0, err);
            if (rc != NFX_OK)
            {
                goto evm_addLunToVol_exit;
            }
            rc = NFX_ERR;
            goto evm_addLunToVol_exit;
        }
                
        /* Commit this to the database.
         */
        rc = upd_lv_strg(volCfg->id, newLvstrg, err);
        if (rc != NFX_OK)
        {
            goto evm_addLunToVol_exit;
        }
                
        /* Communicate the fact that we've auto-grown
         */
        evm_sendEvt(EVENT_VOL_GROW, volCfg->id, EVENT_STATE_NOCHANGE);

        /* Update the volume configuration as well as the label
         * of all LUNs composing the volume.
         */
        rc = upd_lv_cfg(volCfg, TRUE, err);
        if (rc != NFX_OK)
        {
            goto evm_addLunToVol_exit;
        }

        /* If the LUN is part of a corevolume, close the LUN.
         */
        if (volCfg->type == EVM_TYPE_COREVOL)
            evm_closeLun(newLun.openHandle);
        
        EVM_FREE(pattr);
        EVM_FREE(lvstrg);
        EVM_FREE(newLvstrg);
    } // for

    if ((!volCfg->onLine) && (evm_isVolOk(volCfg, err) == TRUE))
    {
        evm_sendEvt(EVENT_VOL_CREATE, volCfg->id, EVENT_STATE_DOWN);
    }

    /* Check if softQuota is reached and send an event if it is.
     */
    threshold = volCfg->softQuota * volCfg->maxSize / 100;
    if ((threshold != 0) &&
        (volCfg->currSize >= threshold) &&
        (sizeBeforeGrowth < threshold))
    {
        evm_sendEvt(EVENT_VOL_SOFTQUOTA, volCfg->id, EVENT_STATE_NOCHANGE);
    }

  evm_addLunToVol_exit:

    /* Free the openHandle list.
     * If the grow operation failed, close the opened LUN.
     */
    while (ohList != NULL)
    {
        oh = ohList;
        if (rc != NFX_OK)
            evm_closeLun(ohList->handle);
        ohList = ohList->next;
        EVM_FREE(oh);
    } // while
    
    /* If the LUN have already been allocated and the grow operation failed,
     * return the LUNs to SPM.
     */
    if ((spmRsp != NULL) && (rc != NFX_OK))
    {
        int                     i;
        evm_msg_t               spmMsg;
        ssize_t                 ncmReqSize;
        ncmMsg_t *              ncmReq = NULL;
        ncmMsg_t *              ncmRsp = (ncmMsg_t *) spmMsg.buf;
        uint32                  spmReqSize;
        spm_MsgFreeLunReq_t *   spmReq = NULL;
        spm_MsgFreeLunRsp_t *   spmRsp2 = (spm_MsgFreeLunRsp_t *) ncmRsp->appMsg;
    
        /* Build NCM message.
         */
        spmReqSize = sizeof(spm_MsgFreeLunReq_t) +
            (((spmRsp->nbLun == 0)? 0:(spmRsp->nbLun - 1)) * sizeof(scsi_id_t));

        ncmReqSize = sizeof(ncmMsg_t) + spmReqSize - 1;
        EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
        NFX_ASSERT(ncmReq != NULL);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->appName, "spm");
        strcpy(ncmReq->nodeName, "PCC");
        ncmReq->rqResp = 1;    // response expected
        spmReq = (spm_MsgFreeLunReq_t *) ncmReq->appMsg;
        
        /* Build SPM message.
         */
        spmReq->hdr.msgType = SPM_FREE_LUN_REQ;
        strcpy(spmReq->hdr.node, myNode);
        spmReq->nbLun = spmRsp->nbLun;
        for (i = 0; i < spmReq->nbLun; i++)
        {
            spmReq->id[i] = spmRsp->id[i];
        } // for
        
        spmMsg.len = sendAgileMsg(-1, ncmReq, ncmReqSize,
                                  "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                  &waitSpmTo, spmMsg.buf, sizeof(spmMsg.buf));
        NFX_ASSERT((spmMsg.len > sizeof(ncmMsg_t)) &&
                   (ncmRsp->msgType == NCM_RESP) &&
                   (spmRsp2->hdr.msgType == SPM_FREE_LUN_RSP));
        EVM_FREE(ncmReq);
    }

    EVM_FREE(pattr);
    EVM_FREE(lvstrg);
    EVM_FREE(newLvstrg);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_delLun
 *
 * Description  : A LUN 'disappeared'.
 *                If the LUN was part of a volume, give an alarm.
 *                Remove it from the storage pool.
 *                Close the device.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_delLun(sdm_devDelReqMsg_t * sdm_msg,
           char *               err)

{
    int32               rc = NFX_OK;
    evm_StrgAttr_t *    sattr = NULL;

    /* Notify SPM about this.
     */
    {
        evm_msg_t               spmMsg;
        ssize_t                 ncmReqSize;
        ncmMsg_t *              ncmReq = NULL;
        ncmMsg_t *              ncmRsp = (ncmMsg_t *) spmMsg.buf;
        spm_MsgDelLunReq_t *    spmReq = NULL;
        spm_MsgDelLunRsp_t *    spmRsp = (spm_MsgDelLunRsp_t *) ncmRsp->appMsg;
    
        /* Build NCM message.
         */
        ncmReqSize = sizeof(ncmMsg_t) + sizeof(spm_MsgDelLunReq_t) - 1;
        EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
        NFX_ASSERT(ncmReq != NULL);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->appName, "spm");
        strcpy(ncmReq->nodeName, "PCC");
        ncmReq->rqResp = 1;    // response expected
        spmReq = (spm_MsgDelLunReq_t *) ncmReq->appMsg;
        
        /* Build SPM message.
         */
        spmReq->hdr.msgType = SPM_DEL_LUN_REQ;
        strcpy(spmReq->hdr.node, myNode);
        spmReq->id = sdm_msg->id;
        
        spmMsg.len = sendAgileMsg(-1, ncmReq, ncmReqSize,
                     "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                     &waitSpmTo, spmMsg.buf, sizeof(spmMsg.buf));
        NFX_ASSERT((spmMsg.len > sizeof(ncmMsg_t)) &&
                   (ncmRsp->msgType == NCM_RESP) &&
                   (spmRsp->hdr.msgType == SPM_DEL_LUN_RSP));
        EVM_FREE(ncmReq);
    }

    /* Check if this LUN was part of a volume (<=> in our local LUN db).
     */
    rc = get_pstrg_cfg(sdm_msg->id, &sattr, err);
    if ((rc == NFX_OK) && (sattr != NULL))
    {
        evm_VolConfig_t     volCfg;
    
        NFX_ASSERT(sattr->nallocs == 1);

        /* Check if the volume is already broken.
         * If not, then send an alarm and offline it.
         */
        rc = get_lv_cfg_byid(sattr->allocs[0].lvid, &volCfg, err);
        NFX_ASSERT(rc == NFX_OK);
        EVM_ERR_PRINT(("LUN 0x%llx failed. Volume %s (0x%llx) is broken.",
                       sdm_msg->id, volCfg.name,
                       sattr->allocs[0].lvid.whole_thing));
        if (evm_isVolOk(&volCfg, err) == TRUE)
        {
            /* Send an alarm.
             */
            evm_sendEvt(EVENT_VOL_BROKEN, sattr->allocs[0].lvid,
                        EVENT_STATE_DOWN);
            
            /* Turn off-line a volume.
             */
            evm_offlineVol(sattr->allocs[0].lvid, TRUE, err);
        }
        
        /* Remove LUN from volume.
         */
        evm_removeLunFromVolume(sdm_msg->id, sattr->allocs[0].lvid, err);
    }

    EVM_FREE(sattr);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_offlineVol
 *
 * Description  : Put offline a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_offlineVol(evm_volId_t  volId,
               boolean      tellEvmSrvr,
               char *       err)

{
    int32           rc = NFX_OK;
    evm_VolConfig_t cfg;
            
    rc = get_lv_cfg_byid(volId, &cfg, err);
    if (rc == NFX_OK)
    {
        if (cfg.onLine)
        {
            evm_mgrArea_t   mgrArea;
            evm_tkoArea_t   tkoArea;
            
            cfg.onLine = 0;

            /* Send message to tell EVM server to take it offline.
             * If no response is received, just continue.
             * The FP-NIM is probably down.
             */
            if (tellEvmSrvr == TRUE)
            {
                upd_evm_srvr(NULL, &cfg, err);
            }

            /* Update the configuration DB
             */
            cfg.onNIM = 0;
            rc = upd_lv_cfg(&cfg, FALSE, err);
            if (rc != NFX_OK)
            {
                goto evm_offlineVol_exit;
            }

            /* If the offline request is not coming from EVM server.
             * Reset the manager and takeover areas of each LUN of the volume.
             */
            if (tellEvmSrvr == TRUE)
            {
                evm_volStrg_t *     lvstrg = NULL;
                evm_StrgAttr_t *    attr = NULL;
                int                 i;

                rc = get_lv_strg(cfg.id, &lvstrg, err);
                if ((rc != NFX_OK) || (lvstrg == NULL))
                {
                    goto evm_offlineVol_exit;
                }
                
                for (i = 0; i < lvstrg->num; i++)
                {
                    rc = get_pstrg_cfg(lvstrg->au[i].strgId, &attr, err);
                    if ((rc != NFX_OK) || (attr == NULL))
                    {
                        EVM_FREE(lvstrg);
                        goto evm_offlineVol_exit;
                    }
                    
                    /* Clean manager area.
                     */
                    bzero((void *) &mgrArea, sizeof(evm_mgrArea_t));
                    mgrArea.magic = NFX_STRG_MAGIC;
                    rc = evm_writeMgrArea(attr, &mgrArea);
                    if (rc != NFX_OK)
                    {
                        sprintf(err, "Failed to clean the manager area.");
                        EVM_FREE(lvstrg);
                        EVM_FREE(attr);
                        goto evm_offlineVol_exit;
                    }
            
                    /* Clean takeover area.
                     */
                    bzero((void *) &tkoArea, sizeof(evm_tkoArea_t));
                    tkoArea.magic = NFX_STRG_MAGIC;
                    rc = evm_writeTkoArea(attr, &tkoArea);
                    if (rc != NFX_OK)
                    {
                        sprintf(err, "Failed to clean the takeover area.");
                        EVM_FREE(lvstrg);
                        EVM_FREE(attr);
                        goto evm_offlineVol_exit;
                    }
                } // for
                EVM_FREE(lvstrg);
                EVM_FREE(attr);
            }
        }
        else
        {
            strcpy(err, "Volume is already offline.");
            rc = EVM_VOL_ALREADY_OFFLINE;
        }
    }

  evm_offlineVol_exit:

    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_removeLunFromVolume
 *
 * Description  : Remove a LUN from a volume.
 *                If it's the latest LUN then remove the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 20-JUN-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_removeLunFromVolume(scsi_id_t   lunId,
                        evm_volId_t volId,
                        char *      err)
    
{
    int32           rc = NFX_OK;
    evm_volStrg_t * lvstrg = NULL;
    int32			openHandle = -1;
    
    EVM_DBG_PRINT(("Remove LUN 0x%llx from volume 0x%llx.",
                   lunId, volId.whole_thing));
            
    /* Get list of LUNs composing the volume.
     */
    rc = get_lv_strg(volId, &lvstrg, err);
    if ((rc == NFX_OK) && (lvstrg != NULL))
    {
        int     i;
        boolean volWithNoLun = TRUE;
        
        /* Remove/reset the failed LUN from the volume.
         */
        for (i = 0; i < lvstrg->num; i++)
        {
            if (lvstrg->au[i].strgId == lunId)
            {
                openHandle = lvstrg->au[i].openHandle;
                bzero((void *) &(lvstrg->au[i]), sizeof(evm_strgConfig_t));
                lvstrg->au[i].openHandle = -1;
            }
            if (lvstrg->au[i].strgId != 0)
            {
                volWithNoLun = FALSE;
            }
        } // for
        rc = upd_lv_strg(volId, lvstrg, err);
        if (volWithNoLun == TRUE)
        {
            EVM_DBG_PRINT(("Volume 0x%llx has no more LUN. "
                           "Remove it from our database.", volId.whole_thing));
            rc = del_lv_cfg(volId, err);
        }
    }
    else
    {
        EVM_ERR_PRINT(("Fail finding the list of LUNs composing "
                       "volume 0x%llx.", volId.whole_thing));
    }
    EVM_FREE(lvstrg);
    
    /* Close device.
     */
    rc = evm_closeLun(openHandle);
    
    /* Remove LUN from our local LUN db.
     */
    upd_pstrgdb(&lunId, FALSE);

    return(rc);
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_tkoCallback
 *
 * Description  : Takeover has been performed, put online a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 25-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void evm_tkoCallback(evm_volId_t   volId,
                     uint32        status,
                     uint64        clientPrivate)
{
	char            rtnmsg[EVM_MAX_BUF_SIZE];
    evm_MsgStat_t * statMsg = (evm_MsgStat_t *)rtnmsg;
    char *          rtnstr = &rtnmsg[sizeof(evm_MsgStat_t)];
    int             rtnsize;
    evm_tkoCb_t *   cb = (evm_tkoCb_t *) (address_t) clientPrivate;

    NFX_ASSERT(cb != NULL);

    EVM_FCT_PRINT(("evm_tkoCallback(0x%llx, %d)\n", volId.whole_thing, status));

    bzero((void *) statMsg, sizeof(evm_MsgStat_t));
    statMsg->hdr.msgType = EVM_MSG_STAT;
    statMsg->hdr.clientPrivate = cb->clientPrivate;

    switch(status)
    {
    case EVM_TKO_OK:
    {
        statMsg->stat = evm_onlineVol(volId, cb->nim, rtnstr);
        break;
    } // case EVM_TKO_OK

    case EVM_TKO_GIVEUP:
    {
        statMsg->stat = status;
        sprintf(rtnstr, "Takeover timed out. "
                "Give up trying to takeover volume 0x%llx.",
                volId.whole_thing);
        break;
    } // case EVM_TKO_GIVEUP

    case EVM_TKO_CONFLICT:
    {
        statMsg->stat = status;
        sprintf(rtnstr, "Takeover conflict. "
                "Another node is also trying to takeover volume 0x%llx.",
                volId.whole_thing);
        break;
    } // case EVM_TKO_CONFLICT

    default:
    {
        statMsg->stat = status;
        sprintf(rtnstr, "Cannot takeover volume 0x%llx.", volId.whole_thing);
        break;
    }
    } // switch
    
    statMsg->payload.lVolId = volId;
    
    rtnsize = sizeof(evm_MsgStat_t);
    if (statMsg->stat != NFX_OK)
    {
        // Could be an error message
        rtnsize += strlen(rtnstr) + 1; // message and its NULL
    }
    statMsg->hdr.len = rtnsize;
    if (sendAgileMsg(-1, rtnmsg, rtnsize, 
                     cb->fromapp, cb->fromcpu, cb->fromslot, MGMTPLANE,
                     NULL, NULL, 0) < 0)
    {
        EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
    }

    EVM_FREE(cb);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_onCallback
 *
 * Description  : Check has been performed, put online a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_onCallback(evm_volId_t  volId,
               uint32       status,
               uint64       clientPrivate)

{
	char            rtnmsg[EVM_MAX_BUF_SIZE];
    evm_MsgStat_t * statMsg = (evm_MsgStat_t *)rtnmsg;
    char *          rtnstr = &rtnmsg[sizeof(evm_MsgStat_t)];
    int             rtnsize;
    evm_tkoCb_t *   cb = (evm_tkoCb_t *) (address_t) clientPrivate;

    NFX_ASSERT(cb != NULL);

    EVM_FCT_PRINT(("evm_onCallback(%d)\n", status));

    bzero((void *) statMsg, sizeof(evm_MsgStat_t));
    statMsg->hdr.msgType = EVM_MSG_STAT;
    statMsg->hdr.clientPrivate = cb->clientPrivate;
    if (status == EVM_TKO_OK)
    {
        statMsg->stat = evm_onlineVol(volId, cb->nim, rtnstr);
    }
    else
    {
        statMsg->stat = status;
        sprintf(rtnstr, "Cannot online the volume 0x%llx.", volId.whole_thing);
    }
    statMsg->payload.lVolId = volId;
    
    rtnsize = sizeof(evm_MsgStat_t);
    if (statMsg->stat != NFX_OK)
    {
        // Could be an error message
        rtnsize += strlen(rtnstr) + 1; // message and its NULL
    }
    statMsg->hdr.len = rtnsize;
    if (sendAgileMsg(-1, rtnmsg, rtnsize, 
                     cb->fromapp, cb->fromcpu, cb->fromslot, MGMTPLANE,
                     NULL, NULL, 0) < 0)
    {
        EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
    }

    EVM_FREE(cb);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_onlineVol
 *
 * Description  : Put online a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_onlineVol(evm_volId_t   volId,
              uint32        nim,
              char *        err)

{
    int32           rc = NFX_OK;
    evm_VolConfig_t cfg;
    boolean         online = FALSE;
    
    rc = get_lv_cfg_byid (volId, &cfg, err);
    if (rc != NFX_OK)
    {
        goto evm_onlineVol_exit;
    }

    // TODO!!! when chassis management will offer the interface to do so...
    // check if specified nim corresponds to an FP-NIM
    if (nim != 1)
    {
        sprintf(err, "Slot %d is not an FP-NIM.", nim);
        EVM_ERR_PRINT(("%s", err));
        rc = NFX_ERR;
        goto evm_onlineVol_exit;
    }

    if (!cfg.onLine)
    {
        cfg.onLine = 1;
        cfg.onNIM = nim;
        
        /* We have to make sure that there is some storage
         * allocated to this LV before we attempt to put it online.
         */
        if (!cfg.currSize)
        {
            rc = evm_growVol(cfg.id, 0, 0, NULL, 0, err);
            if (rc != NFX_OK)
            {
                strcat(err, " - Cannot online empty volume.");
                goto evm_onlineVol_exit;
            }
            
            /* LV config was updated from grow_lvol() (size)
             * and update it with online info
             */
            get_lv_cfg_byid(cfg.id, &cfg, err);
            cfg.onLine = 1;
            cfg.onNIM = nim;
        }
        else
        {
            evm_volStrg_t * strg = NULL;
            
            /* make sure all the LUN pieces are there
             */
            rc = get_lv_strg(cfg.id, &strg, err);
            if (rc != NFX_OK)
            {
                // Cannot put LV online
                goto evm_onlineVol_exit;
            }
            NFX_ASSERT(strg != NULL);
            
            rc = check_lvol(&cfg, strg);
            if (rc != NFX_OK)
            {
                // Cannot put LV online
                goto evm_onlineVol_exit;
            }
            EVM_FREE(strg);
        }

        /* Send message to tell evm_srvr about this new LV and
         * also send ALL physical storage assigned.
         */
        if (upd_evm_srvr (NULL, &cfg, err) != NFX_OK)
        {
            rc = NFX_ERR;
        }
        else
        {
            online = TRUE;
            
            /* Indicate on the 1st LUN that we now own it.
             */
            {
                int32               rc = NFX_OK;
                evm_volStrg_t *     lvstrg = NULL;
                evm_StrgAttr_t *    attr = NULL;
                evm_mgrArea_t       data;
                evm_tkoArea_t       tkoArea;

                rc = get_lv_strg(cfg.id, &lvstrg, err);
                if ((rc != NFX_OK) || (lvstrg == NULL))
                {
                    goto evm_onlineVol_exit;
                }
                
                rc = get_pstrg_cfg(lvstrg->au[0].strgId, &attr, err);
                if ((rc != NFX_OK) || (attr == NULL))
                {
                    goto evm_onlineVol_exit;         
                }

                /* Write our name in the manager area.
                 */
                data.magic = NFX_STRG_MAGIC;
                memcpy(&data.mgrId, &myNfx, sizeof(evm_nfxId_t));
                EVM_GET_TIME(data.timestamp);
                rc = evm_writeMgrArea(attr, &data);
                if  (rc != NFX_OK)
                {
                    goto evm_onlineVol_exit;
                }
                
                /* Clean takeover area.
                 */
                bzero((void *) &tkoArea, sizeof(evm_tkoArea_t));
                tkoArea.magic = NFX_STRG_MAGIC;
                rc = evm_writeTkoArea(attr, &tkoArea);
                if  (rc != NFX_OK)
                {
                    goto evm_onlineVol_exit;
                }
        
                EVM_FREE(lvstrg);
                EVM_FREE(attr);
            }

            /* Update the volume configuration.
             */
            rc = upd_lv_cfg(&cfg, FALSE, err);
       }
    }
    else
    {
        strcpy (err, "volume is already online");
        rc = EVM_VOL_ALREADY_ONLINE;
    }
    
  evm_onlineVol_exit:

    /* If a problem occurs after the volume has been put online on the FP,
     * offline it so there is no inconsitency between EVM and EVM server.
     */
    if ((rc != NFX_OK) && (rc != EVM_VOL_ALREADY_ONLINE) && (online == TRUE))
    {
        cfg.onLine = 0;
        upd_evm_srvr(NULL, &cfg, err);
    }
    
    return(rc);
}
    
/*------------------------------------------------------------------------------
 * Name         : check_lvol
 *
 * Description  : perform basic sanity checks for this logical volume
 *                config data from disk before bringing on line
 *
 * Created by   : Lan Chin
 *
 * Date Created : 02-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

int
check_lvol(evm_VolConfig_t *    volCfg,
           evm_volStrg_t *      volStrg)
    
{
    return(check_volStrg(volCfg, volStrg));
}

/*------------------------------------------------------------------------------
 * Name         : evm_isVolOk
 *
 * Description  : Check if a volume is complete (ie. no LUN missing)
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-OCT-2002
 *
 *------------------------------------------------------------------------------
 */

static boolean
evm_isVolOk(evm_VolConfig_t *   volCfg,
            char *              err)
    
{
    boolean         volOk = FALSE;
    int32           rc = NFX_OK;
    evm_volStrg_t * lvstrg = NULL;
    
    /* Check if the volume is complete.
     * If yes, notify the world about it.
     */
    rc = get_lv_strg(volCfg->id, &lvstrg, err);
    NFX_ASSERT((rc == NFX_OK) && (lvstrg != NULL));
    if (check_volStrg(volCfg, lvstrg) == NFX_OK)
    {
        volOk = TRUE;
    }
    EVM_FREE(lvstrg);

    return(volOk);
}

/*------------------------------------------------------------------------------
 * Name         : evm_delVol
 *
 * Description  : Delete a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created: 22-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

static int32
evm_delVol(evm_volId_t  volId,
           char *       err)

{
    int32               rc = NFX_OK;
    evm_VolConfig_t     volCfg;
    evm_volStrg_t *     strg = NULL;
    evm_StrgAttr_t *    sattr = NULL;
    int                 i;
    evm_strgCfg_t       odCfg;
      
    rc = get_lv_cfg_byid(volId, &volCfg, err);
    if (rc != NFX_OK)
    {
        goto evm_delVol_exit;
    }
    
    if (volCfg.onLine == TRUE)
    {
        rc  = NFX_ERR;
        sprintf(err, "Volume is online. Cannot delete it.");
        goto evm_delVol_exit;
    }

    rc = get_lv_strg(volId, &strg, err);
    if ((rc != NFX_OK) || (strg == NULL))
    {
        goto evm_delVol_exit;
    }

    /* Go through all LUNs composing this volume.
     */
    for (i = 0; i < strg->num; i++)
    {
        /* The SCSI id should never be null,
         * but in lab we sometime came to this situation
         * and we cannot remove the volume anymore...
         */
        if (strg->au[i].strgId == 0)
        {
            EVM_ERR_PRINT(("volId = 0x%llx - strg->au[%d].strgId == 0",
                           volId.whole_thing, i));
            continue;
        }
                
        /* Get LUN configuration.
         */
        rc = get_pstrg_cfg(strg->au[i].strgId, &sattr, err);    
        if ((rc != NFX_OK) || (sattr == NULL))
        {
            goto evm_delVol_exit;         
        }

        /* Get an openHandle for the LUNs composing a corevolume.
         */
        if (volCfg.type == EVM_TYPE_COREVOL)
        {
            evm_StrgAttr_t  lun;
            
            rc = evm_getLunInfo(strg->au[i].strgId, &lun, err);
            if (rc != NFX_OK)
            {
                EVM_ERR_PRINT(("EVM: %s", err));
                goto evm_delVol_exit;
            }

            if (lun.openHandle == -1)
            {
                rc = evm_openLun(lun.id, SSC_SLOT, &(lun.openHandle));
                if ((rc != NFX_OK) || (lun.openHandle == -1))
                {
                    sprintf(err, "cannot open LUN 0x%llx", lun.id);
                    EVM_ERR_PRINT(("EVM: %s", err));
                    goto evm_delVol_exit;
                }
            }

            sattr->openHandle = lun.openHandle;
        }
        
        /* Update/Reset LUN's label.
         */
        rc = evm_readLunLabel(strg->au[i].strgId, sattr, &odCfg);
        if (rc != NFX_OK)
        {
            goto evm_delVol_exit;
        }
        bzero((void *) &odCfg, sizeof(evm_strgCfg_t));
        EVM_OD_CFG_INIT(odCfg);
        rc = evm_writeLunLabel(strg->au[i].strgId, sattr, &odCfg);
        if (rc != NFX_OK)
        {
            goto evm_delVol_exit;
        }
        
        /* Remove it from the LUN db.
         */
        rc = del_pstrg_cfg(strg->au[i].strgId, sattr, err);
        if (rc != NFX_OK)
        {
            goto evm_delVol_exit;
        }
        
        /* Close the LUN.
         */
        rc = evm_closeLun(sattr->openHandle);
        if (rc != NFX_OK)
        {
            goto evm_delVol_exit;
        }

        /* Tell SPM that this LUN is now free.
         */
        {
            evm_msg_t               spmMsg;
            ssize_t                 ncmReqSize;
            ncmMsg_t *              ncmReq = NULL;
            ncmMsg_t *              ncmRsp = (ncmMsg_t *) spmMsg.buf;
            spm_MsgFreeLunReq_t *   spmReq = NULL;
            spm_MsgFreeLunRsp_t *   spmRsp = (spm_MsgFreeLunRsp_t *) ncmRsp->appMsg;
    
            /* Build NCM message.
             */
            ncmReqSize = sizeof(ncmMsg_t) + sizeof(spm_MsgFreeLunReq_t) - 1;
            EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
            NFX_ASSERT(ncmReq != NULL);
            ncmReq->msgType = NCM_SEND;
            strcpy(ncmReq->appName, "spm");
            strcpy(ncmReq->nodeName, "PCC");
            ncmReq->rqResp = 1;    // response expected
            spmReq = (spm_MsgFreeLunReq_t *) ncmReq->appMsg;
        
            /* Build SPM message.
             */
            spmReq->hdr.msgType = SPM_FREE_LUN_REQ;
            strcpy(spmReq->hdr.node, myNode);
            spmReq->nbLun = 1;
            spmReq->id[0] = strg->au[i].strgId;
            
            spmMsg.len = sendAgileMsg(-1, ncmReq, ncmReqSize,
                                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                      &waitSpmTo,
                                      spmMsg.buf, sizeof(spmMsg.buf));
#ifdef ISABELLE
            NFX_ASSERT((spmMsg.len > sizeof(ncmMsg_t)) &&
                       (ncmRsp->msgType == NCM_RESP) &&
                       (spmRsp->hdr.msgType == SPM_FREE_LUN_RSP));
#else // to facilitate the debugging
            NFX_ASSERT(spmMsg.len > sizeof(ncmMsg_t));
            NFX_ASSERT(ncmRsp->msgType == NCM_RESP);
            NFX_ASSERT(spmRsp->hdr.msgType == SPM_FREE_LUN_RSP);
#endif // ISABELLE
            EVM_FREE(ncmReq);
        }

        EVM_FREE(sattr);
    } // for
            
    /* Tell all the world that we've destroyed this LV
     */
    evm_sendEvt(EVENT_VOL_DELETE, volId, EVENT_STATE_NONE);
    
    /* Remove all LUNs from the volume.
     * Delete the volume.
     */
    strg->num = 0;
    rc = upd_lv_strg(volId, strg, err);
    rc = del_lv_cfg(volId, err);
    
evm_delVol_exit:
    EVM_FREE(strg);
    EVM_FREE(sattr);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : create_lvol
 *
 * Description  : Create a LV database entry.
 *                Volumes can also be created based on storage attributes
 *
 * Created by   : Greg Seibert
 * Modified by  : Lan Chin
 *
 * Date Created: 19-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

int
create_lvol(evm_VolConfig_t *   cfg,
            boolean             allowDuplicate,
            char *              errstr)
    
{
    EVM_FCT_PRINT(("create_lvol(0x%llx)\n", cfg->id.whole_thing));
    
    /* Get a volume id if not specified.
     * If I am the PCC ok, otherwise ask the PCC.
     */
    if (cfg->id.whole_thing == 0)
    {
        if (node_is_PCC())
        {
            int rc;
            
            rc = evm_getNewVolId(&(cfg->id), errstr);
            if (rc != NFX_OK)
            {
                return(rc);
            }
        }
        else
        {
            evm_msg_t               evmMsg;
            ssize_t                 ncmReqSize;
            ncmMsg_t *              ncmReq = NULL;
            ncmMsg_t *              ncmRsp = (ncmMsg_t *) evmMsg.buf;
            evm_MsgGetNewVolId_t *  evmReq = NULL;
            evm_MsgStat_t *         evmRsp = (evm_MsgStat_t *) ncmRsp->appMsg;
    
            /* Build NCM + EVM message.
             */
            ncmReqSize = sizeof(ncmMsg_t) + sizeof(evm_MsgGetNewVolId_t) - 1;
            EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
            NFX_ASSERT(ncmReq != NULL);
            ncmReq->msgType = NCM_SEND;
            strcpy(ncmReq->appName, "evm");
            strcpy(ncmReq->nodeName, "PCC");
            ncmReq->rqResp = 1;    // response expected
            evmReq = (evm_MsgGetNewVolId_t *) ncmReq->appMsg;
            
            evmReq->hdr.msgType = EVM_VOL_GET_NEWVOLID;
            evmReq->hdr.len = sizeof(evm_MsgGetNewVolId_t);
            evmReq->hdr.clientPrivate = 0;

            /* Send message.
             */
            evmMsg.len = sendAgileMsg(-1, ncmReq, ncmReqSize,
                                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                      &waitSpmTo,
                                      evmMsg.buf, sizeof(evmMsg.buf));
            NFX_ASSERT(evmMsg.len > sizeof(ncmMsg_t));
            NFX_ASSERT(ncmRsp->msgType == NCM_RESP);
            NFX_ASSERT(evmRsp->hdr.msgType == EVM_MSG_STAT);
            EVM_FREE(ncmReq);

            if (evmRsp->stat == NFX_OK)
            {
                cfg->id.whole_thing = evmRsp->payload.lVolId.whole_thing;
            }
            else
            {
                strcpy(errstr, &(evmMsg.buf[sizeof(evm_MsgStat_t)]));
                return(evmRsp->stat);
            }
        }
    }
    
    /* If they didn't specify a name string, make one up
     */
	if (!strlen (cfg->name) ) {
		sprintf (cfg->name, "vol%llx", cfg->id.whole_thing);
	}

    /* Make sure the volume name is unique within the cluster.
     */
    if (allowDuplicate == FALSE)
    {
        boolean nameIsUnique = FALSE;
    
        if (node_is_PCC())
        {
            nameIsUnique = evm_isVolNameUnique(cfg->name);
        }
        else
        {
            evm_msg_t               evmMsg;
            ssize_t                 ncmReqSize;
            ncmMsg_t *              ncmReq = NULL;
            ncmMsg_t *              ncmRsp = (ncmMsg_t *) evmMsg.buf;
            evm_MsgCheckVolName_t * evmReq = NULL;
            evm_MsgStat_t *         evmRsp = (evm_MsgStat_t *) ncmRsp->appMsg;
            
            /* Build NCM + EVM message.
             */
            ncmReqSize = sizeof(ncmMsg_t) + sizeof(evm_MsgCheckVolName_t) - 1;
            EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
            NFX_ASSERT(ncmReq != NULL);
            ncmReq->msgType = NCM_SEND;
            strcpy(ncmReq->appName, "evm");
            strcpy(ncmReq->nodeName, "PCC");
            ncmReq->rqResp = 1;    // response expected
            evmReq = (evm_MsgCheckVolName_t *) ncmReq->appMsg;
            
            evmReq->hdr.msgType = EVM_VOL_CHECK_VOLNAME;
            evmReq->hdr.len = sizeof(evm_MsgCheckVolName_t);
            evmReq->hdr.clientPrivate = 0;
            strcpy(evmReq->name, cfg->name);
            
            /* Send message.
             */
            evmMsg.len = sendAgileMsg(-1, ncmReq, ncmReqSize,
                                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                      &waitSpmTo,
                                      evmMsg.buf, sizeof(evmMsg.buf));
            NFX_ASSERT(evmMsg.len > sizeof(ncmMsg_t));
            NFX_ASSERT(ncmRsp->msgType == NCM_RESP);
            NFX_ASSERT(evmRsp->hdr.msgType == EVM_MSG_STAT);
            EVM_FREE(ncmReq);
            
            nameIsUnique = evmRsp->stat;
        }
        
        if (!nameIsUnique)
        {
            sprintf(errstr, "Duplicate volume name in cluster.");
            return NFX_ERR;
        }
    }

    return(add_lv_cfg(cfg, allowDuplicate, errstr));
}

/*------------------------------------------------------------------------------
 * Name         : evm_getNewOpenHandle
 *
 * Description  : Close the local (SSC) handle and get one for the FP-NIM.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

static int
evm_getNewOpenHandle(evm_strgConfig_t * au,
                     evm_VolConfig_t *  cfg,
                     char *             err)

{
    int                 rc = NFX_OK;
    evm_StrgAttr_t *    pattr = NULL;

    if (SDM_HANDLE_GET_FP_SLOT(au->openHandle) == cfg->onNIM)
    {
        EVM_DBG_PRINT(("Keep handle %x\n", au->openHandle));
        goto evm_getNewOpenHandle_exit;
    }
            
    /* Close current handle.
     */
    if (evm_closeLun(au->openHandle) != NFX_OK)
    {
        sprintf(err, "Cannot close LUN 0x%llx on SSC.", au->strgId);
        rc = NFX_ERR;
        goto evm_getNewOpenHandle_exit;
    }

    /* Open new one on requested FP-NIM.
     */
    if (evm_openLun(au->strgId, cfg->onNIM, &(au->openHandle)) != NFX_OK)
    {
        sprintf(err, "Cannot open LUN 0x%llx on slot %d.",
                au->strgId, cfg->onNIM);
        rc = NFX_ERR;
        goto evm_getNewOpenHandle_exit;
    }

    /* Update openHandle in LUN database.
     */
    rc = get_pstrg_cfg(au->strgId, &pattr, err);
    if ((rc != NFX_OK) || (pattr == NULL))
    {
        goto evm_getNewOpenHandle_exit;
    }
    pattr->openHandle = au->openHandle;
    rc = upd_pstrg_cfg(au->strgId, pattr, 0, err);
    if (rc != NFX_OK)
    {
        goto evm_getNewOpenHandle_exit;
    }

  evm_getNewOpenHandle_exit:
    
    EVM_FREE(pattr);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : upd_evm_srvr
 *
 * Description  : Upon change of a currently in service or
 *                the creation and initial export of logical volumes, this
 *                will send the message to the proper instance of evm_srvr
 *                to accomplish the task.
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 16-JUL-2001
 *
 *------------------------------------------------------------------------------
 */

int
upd_evm_srvr(evm_strgConfig_t * au,
             evm_VolConfig_t *  cfg,
             char *             err)
    
{
    int             rc = NFX_OK;
    char *          msg = NULL;
	int             msgsize;
	int             i;
	int             stat;
	char            xbuf[sizeof(evm_MsgStat_t)  + EVM_MAX_ERR_STR];
	evm_MsgStat_t * resp = (evm_MsgStat_t *) xbuf;
	struct timeval  nimto = {10, 0 };
    evm_volStrg_t * lvstrg = NULL;

    NFX_ASSERT(cfg != NULL);
    
    /* Check if the requested FP-NIM is up and running.
     */
    if ((cfg->onNIM) &&
        (checkAgileDest(NFX_NFP_FP1, cfg->onNIM) != NFX_OK) &&
        (checkAgileDest(NFX_NFP_FP2, cfg->onNIM) != NFX_OK))
    {
        sprintf(err, "Slot %d is not up and running.", cfg->onNIM);
        EVM_ERR_PRINT(("%s", err));
        rc = NFX_ERR;
        goto upd_evm_srvr_exit;
    }
    
	if (cfg->onLine)
    {
        rc = get_lv_strg(cfg->id, &lvstrg, err);
        if ((rc != NFX_OK) || (lvstrg == NULL))
        {
            rc = NFX_ERR;
            goto upd_evm_srvr_exit;
        }

		if (au)     // Only notify it of this one
        {
            evm_MsgAddStrg_t *  addmsg = NULL;

            rc = evm_getNewOpenHandle(au, cfg, err);
            if (rc != NFX_OK)
            {
                EVM_ERR_PRINT(("%s", err));
                goto upd_evm_srvr_exit;
            }                
            
            msgsize = sizeof(evm_MsgAddStrg_t);
            EVM_ALLOC_ZERO(msg, char, msgsize);
            addmsg = (evm_MsgAddStrg_t *) msg;
            
			addmsg->hdr.msgType = EVM_LVACT_ADDSTRG;
			addmsg->hdr.clientPrivate = getpid();
			addmsg->lVol = cfg->id;
			addmsg->strg = *au;

			EVM_DBG_PRINT(("Adding %lld blocks to volume 0x%llx",
                           au->auSize, cfg->id.whole_thing));
		}
		else
        {
            evm_MsgCreateLV_t * creatmsg = NULL;
            int                 rc = NFX_OK;
            
            msgsize = sizeof(evm_MsgCreateLV_t) + 
                ((lvstrg->num - 1) * sizeof(evm_strgConfig_t));
            EVM_ALLOC_ZERO(msg, char, msgsize);
            creatmsg = (evm_MsgCreateLV_t *) msg;

			creatmsg->hdr.msgType = EVM_LVACT_CREATELV;
			creatmsg->hdr.clientPrivate = getpid();

			creatmsg->lVol = cfg->id;
			creatmsg->nfxId = myNfx;
			creatmsg->nInitSAU = lvstrg->num;
			for (i = 0; i < lvstrg->num; i++)
            {
                rc = evm_getNewOpenHandle(&(lvstrg->au[i]), cfg, err);
                if (rc != NFX_OK)
                {
                    EVM_ERR_PRINT(("%s", err));
                    goto upd_evm_srvr_exit;
                }

				creatmsg->initSAU[i] = lvstrg->au[i];
			} // for

			EVM_DBG_PRINT(("Creating volume 0x%llx on slot %d.",
                           cfg->id.whole_thing, cfg->onNIM));
		}

        /* Commit the change (openHandles) to the database.
         */
        rc = upd_lv_strg(cfg->id, lvstrg, err);
        if (rc != NFX_OK)
        {
            EVM_ERR_PRINT(("upd_lv_strg() failed"));
            rc = NFX_ERR;
            goto upd_evm_srvr_exit;
        }
	}
	else
    {
        /* Take the volume offline.
         */
		evm_MsgOffLinLV_t * offmsg = NULL;

        msgsize = sizeof(evm_MsgOffLinLV_t);
        EVM_ALLOC_ZERO(msg, char, msgsize);
        offmsg = (evm_MsgOffLinLV_t *) msg;
        
		offmsg->hdr.msgType = EVM_LVACT_OFFLINELV;
		offmsg->hdr.clientPrivate = getpid();
		offmsg->lVol = cfg->id;

		if (cfg->onNIM)
        {
			EVM_DBG_PRINT(("Placing volume 0x%llx offline",
                           cfg->id.whole_thing));
		}
	}

	stat = 0;
	if (cfg->onNIM)
    {
        ((evm_msgHdr_t *) msg)->len = msgsize;
        stat = sendAgileMsg(-1, msg, msgsize,
                            "evm", NFX_NFP_FP_ANY, cfg->onNIM, MGMTPLANE,
                            &nimto, xbuf, sizeof(xbuf));
		if (!stat)
        {
			sprintf (err, "Timed out trying to send message to evm on NIM %d",
			         cfg->onNIM);
            EVM_DBG_PRINT(("EVM: %s", err));
			rc = NFX_ERR;
            goto upd_evm_srvr_exit;
		}
		else if (stat < 0)
        {
			sprintf (err, "Error sending to NIM %d: %s", cfg->onNIM, 
			         strerror(errno));
            EVM_DBG_PRINT(("EVM: %s", err));
			rc = NFX_ERR;
            goto upd_evm_srvr_exit;
		}
		else if ((resp->stat != NFX_OK) &&
                 (resp->stat != EVM_VOL_ALREADY_EXIST))
        {
			sprintf (err, "evm on NIM %d rejected request because %s",
                     cfg->onNIM, &(xbuf[sizeof(evm_MsgStat_t)]));
            EVM_DBG_PRINT(("EVM: %s", err));
			rc = NFX_ERR;
            goto upd_evm_srvr_exit;
		}
	}
    
upd_evm_srvr_exit:    

    EVM_FREE(lvstrg);
    EVM_FREE(msg);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_rescanLun
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-AUG-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_rescanLun(scsi_id_t lunId,
              char *    err)
    
{
    int32                   rc = NFX_OK;
    evm_StrgAttr_t *        lun = NULL;
    boolean                 lunWasPartOfVol = FALSE;
    evm_strgCfg_t           odCfg;
    boolean                 closeLun = FALSE;

    EVM_DBG_PRINT(("evm_rescanLun(0x%llx)\n", lunId));

    /* Do we know about this LUN?
     */
    rc = get_pstrg_cfg(lunId, &lun, err);
    if ((rc == NFX_OK) && (lun != NULL))
    {
        // it's part of a volume and we have an open handle
        lunWasPartOfVol = TRUE;
    }
    else
    {
        // it isn't part of a volume and we don't have an open handle
        EVM_ALLOC_ZERO(lun, evm_StrgAttr_t, sizeof(evm_StrgAttr_t));
        rc = evm_getLunInfo(lunId, lun, err);
        rc = evm_openLun(lunId, SSC_SLOT, &(lun->openHandle));
        closeLun = TRUE;
    }
    
    /* Read label.
     */
    rc = evm_readLunLabel(lunId, lun, &odCfg);
    
    /* Check that the LUN is an agile one and belongs to my cluster.
     */
    if ((odCfg.magic == NFX_STRG_MAGIC) &&
        (!strcmp(odCfg.cluster, myCluster)))
    {
        if (odCfg.id.whole_thing == 0)
        {
            // free LUN
            if (lunWasPartOfVol)
            {
                // remove LUN from volume
                // remove volume if no more LUNs
                rc = evm_removeLunFromVolume(lunId,
                                             lun->allocs[0].lvid,
                                             err);
                closeLun = FALSE;
            }
            else
            {
                // do nothing
            }
        }
        else
        {
            // used LUN
            if (lunWasPartOfVol)
            {
                // is same volume id?
                if (lun->allocs[0].lvid.whole_thing == odCfg.id.whole_thing)
                {
                    // yes, read label and update volume info
                    rc = upd_vol_from_disk(lunId, lun, &odCfg, TRUE, err);
                }
                else
                {
                    // no, alarm!
                    EVM_ERR_PRINT(("EVM ERROR: "
                                   "Conflict on volume id on LUN 0x%llx. "
                                   "In database: 0x%llx, on disk: 0x%llx.",
                                   lun->allocs[0].lvid.whole_thing,
                                   odCfg.id.whole_thing));
                }
            }
            else
            {
                // add LUN to volume or create vol
                evm_addLun(lun, FALSE, err);
                closeLun = FALSE;
            }
        }
    }
    else
    {
        // free LUN
        if (lunWasPartOfVol)
        {
            // remove LUN from volume
            // remove volume if no more LUNs
            rc = evm_removeLunFromVolume(lunId,
                                         lun->allocs[0].lvid,
                                         err);
            closeLun = FALSE;
        }
        else
        {
            // do nothing
        }
    }
    
    if (closeLun == TRUE)
    {
        evm_closeLun(lun->openHandle);
    }
    if (!lunWasPartOfVol)
    {
        EVM_FREE(lun);
    }
    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_processMsg
 *
 * Description  : Process messages sent to EVM.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_processMsg(char *       msgbuf,
               char *       fromapp,
               uint32       fromcpu,
               uint32       fromslot)
    
{
	char            rtnmsg[EVM_MAX_BUF_SIZE]; // Sometimes it's not a MsgStat_t
    int             rtnsize;
    evm_MsgStat_t * stat_msg = (evm_MsgStat_t *)rtnmsg;
    char *          rtnstr = &rtnmsg[sizeof(evm_MsgStat_t)];

    /* Init a status message buffer since it's the response from most
     * messages.
     */
    memset(rtnmsg, 0, sizeof(rtnmsg) + 1);  // Everyone gets status
    stat_msg->hdr.msgType = EVM_MSG_STAT;
    stat_msg->hdr.clientPrivate =  ((evm_msgHdr_t *)msgbuf)->clientPrivate;
    
    /* These are the messages we expect
     */
    switch(((evm_msgHdr_t *)msgbuf)->msgType)
    {

    /* ------------------------------------
     * Logical Volume Manipulation Messages
     * ------------------------------------
     */
        
    case EVM_VOL_GET_ATTRIB:
    {
        stat_msg->stat =  send_vol_cfg(((evm_MsgVolGetAttrib_t *)msgbuf)->id, 
                                       stat_msg->hdr.clientPrivate,
                                       fromapp, fromcpu, fromslot, rtnstr);
        if (stat_msg->stat == NFX_OK)
        {
            return; // Sent its own response
        }
        else
        {
            break;	// Failure gets status message
        }
    } // case EVM_VOL_GET_ATTRIB

    case EVM_VOL_GET_NAME_LIST:
    {
        stat_msg->stat = send_volName_cfg(stat_msg->hdr.clientPrivate,
                                          fromapp, fromcpu, fromslot,
                                          rtnstr);
        if (stat_msg->stat == NFX_OK)
        {
            return; // Sent its own response
        }
        else
        {
            break;	// Failure gets status message
        }
    } // case EVM_VOL_GET_NAME_LIST
            
    case EVM_VOL_GET_ID:
    {
        stat_msg->stat =  get_lv_id_byname (((evm_MsgGetId_t *)msgbuf)->name,
                                            &stat_msg->payload.lVolId, rtnstr);
        break;
    } // case EVM_VOL_GET_ID

    case EVM_VOL_GET_NEWVOLID:
    {
        if (node_is_PCC())
        {
            stat_msg->stat =  evm_getNewVolId(&stat_msg->payload.lVolId,
                                              rtnstr);
        }
        else
        {
            sprintf(rtnstr, "Cannot generate a new volume id on a non-PCC.");
            stat_msg->stat = NFX_ERR;
        }
        break;
    } // case EVM_VOL_GET_NEWVOLID
    
    case EVM_VOL_CHECK_VOLNAME:
    {
        evm_MsgCheckVolName_t * req = (evm_MsgCheckVolName_t *) msgbuf;
        
        if (node_is_PCC())
        {
            stat_msg->stat =  evm_isVolNameUnique(req->name);
        }
        else
        {
            sprintf(rtnstr,
                    "Cannot check the uniqueness of a volume name on a non-PCC.");
            stat_msg->stat = FALSE;
        }
        break;
    } // case EVM_VOL_CHECK_VOLNAME

    case EVM_VOL_CREATE:
    {
        evm_MsgVolCreate_t *    req = (evm_MsgVolCreate_t *) msgbuf;
        uint32                  onLine = req->config.onLine;
        uint32                  onNIM = req->config.onNIM;

        /* Dirty workaround:
         * Reset the online flag and NIM before creating and growing
         * the volume, otherwise upd_evm_srvr() thinks it on-line and
         * tries to contact the evm_srvr on the NIM!
         */
        req->config.onLine = FALSE;
        req->config.onNIM = 0;
            
        /* Create the volume.
         */
        stat_msg->stat = create_lvol(&(req->config), FALSE, rtnstr);
        stat_msg->payload.lVolId = req->config.id;
        if (stat_msg->stat == NFX_OK)
        {
            /* Allocate LUNs to the new volume.
             */
            evm_growPendingCb_t *       cb = NULL;
            EVM_ALLOC(cb, evm_growPendingCb_t, sizeof(evm_growPendingCb_t));
            
            stat_msg->stat = evm_growVol(req->config.id, 0,
                                         req->nbLun, req->lunList,
                                         (address_t) cb, rtnstr);
            if (stat_msg->stat == NFX_OK)
            {
                /* Check if volume needs to be put on-line.
                 */
                EVM_FREE(cb);
                
                if ((onLine == TRUE) && (req->config.type != EVM_TYPE_COREVOL))
                {
                    stat_msg->stat = evm_onlineVol(req->config.id,
                                                   onNIM, rtnstr);
                }
                else
                {
                    /* Do nothing.
                     */
                }
            }
            else if (stat_msg->stat == EVM_VOL_GROW_PENDING)
            {
                /* Tell CCM about this volume even if empty.
                 */
                evm_sendEvt(EVENT_VOL_CREATE, stat_msg->payload.lVolId,
                            EVENT_STATE_DOWN);

                /* Save context so we know what to do when receiving
                 * the asynchronous LUN allocation response.
                 */
                cb->action = EVM_VOL_CREATE;
                cb->clientPrivate = req->hdr.clientPrivate;

                cb->volId = stat_msg->payload.lVolId;
                cb->online = onLine;
                cb->nim = onNIM;
                strcpy(cb->fromapp, fromapp);
                cb->fromcpu = fromcpu;
                cb->fromslot = fromslot;
            }
            else
            {
                /* EVM wasn't able to allocate LUNs.
                 * Undo the volume creation (destroy it!).
                 */
                int     rc;
                char    str[EVM_MAX_ERR_STR];

                EVM_FREE(cb);
                
                strcpy(str, "");
                rc = evm_delVol(req->config.id, str);
            }
        }
        else
        {
            /* Do nothing.
             */
        }
        break;
    } // case EVM_VOL_CREATE
        
    case EVM_VOL_ONLINE:
    {
        evm_MsgVolOnline_t *    req = (evm_MsgVolOnline_t *) msgbuf;
        evm_VolConfig_t         volCfg;

        EVM_FCT_PRINT(("EVM_VOL_ONLINE 0x%llx on nim %d\n",
                       req->id.whole_thing, req->nim));

        /* Check if the volume exists and can be put online.
         */
        stat_msg->stat = get_lv_cfg_byid(req->id, &volCfg, rtnstr);
        if (stat_msg->stat != NFX_OK)
        {
            break;
        }

        if (evm_isVolOk(&volCfg, rtnstr) == FALSE)
        {
            sprintf(rtnstr, "Volume '%s' is broken. Cannot online it.", 
                    volCfg.name);
            stat_msg->stat = NFX_ERR;
            break;
        }
        
        if (volCfg.type == EVM_TYPE_COREVOL)
        {
            sprintf(rtnstr,
                    "%s (0x%llx) is a corevolume which shall not be online.",
                    volCfg.name, req->id.whole_thing);
            EVM_ERR_PRINT(("%s", rtnstr));
            stat_msg->stat = NFX_ERR;
            break;
        }
        
        /* Online the volume.
         */
        stat_msg->stat = evm_onlineVol(req->id, req->nim, rtnstr);
        break;
    } // case EVM_VOL_ONLINE

    case EVM_VOL_TAKEOVER:
    {
        int32                   rc;
        evm_MsgVolTakeover_t *  req = (evm_MsgVolTakeover_t *) msgbuf;
        evm_tkoCb_t *           cb = NULL;
        evm_VolConfig_t         volCfg;
        
        EVM_FCT_PRINT(("EVM_VOL_TAKEOVER 0x%llx on nim %d\n",
                       req->id.whole_thing, req->nim));
            
        /* Check if the volume exists and can be put online.
         */
        stat_msg->stat = get_lv_cfg_byid(req->id, &volCfg, rtnstr);
        if (stat_msg->stat != NFX_OK)
        {
            break;
        }

        if (volCfg.type == EVM_TYPE_COREVOL)
        {
            sprintf(rtnstr,
                    "%s (0x%llx) is a corevolume which shall not be online.",
                    volCfg.name, req->id.whole_thing);
            EVM_ERR_PRINT(("%s", rtnstr));
            stat_msg->stat = NFX_ERR;
            break;
        }

        /* Start to takeover the volume.
         */
        EVM_ALLOC_ZERO(cb, evm_tkoCb_t, sizeof(evm_tkoCb_t));
        cb->clientPrivate = req->hdr.clientPrivate;
        strcpy(cb->fromapp, fromapp);
        cb->fromcpu = fromcpu;
        cb->fromslot = fromslot;
        cb->nim = req->nim;

        req->hdr.clientPrivate = (uint64) (address_t) cb;
                
        rc = evm_takeoverVol(req, evm_tkoCallback, FALSE, rtnstr);
        if (rc == NFX_OK)
        {
            return;   // wait for callback...
        }
        else
        {
            stat_msg->stat = rc;
            EVM_FREE(cb);
            break;
        }
        break;
    }

    case EVM_VOL_OFFLINE:
    {
        stat_msg->stat = evm_offlineVol(((evm_MsgVolOffline_t *)msgbuf)->id,
                                        TRUE, rtnstr);
        break;
    }

    case EVM_VOL_DESTROY:
    {
        stat_msg->stat = evm_delVol(((evm_MsgVolDestroy_t *)msgbuf)->id,
                                    rtnstr);
        break;
    }
        
    case EVM_VOL_RENAME:
    {
        evm_VolConfig_t cfg;
        evm_volId_t     id;

        if (get_lv_id_byname(((evm_MsgVolRename_t *)msgbuf)->newName,
                             &id, rtnstr) == NFX_OK)
        {
            sprintf(rtnstr, "Volume '%s' already exists.", 
                    ((evm_MsgVolRename_t *)msgbuf)->newName);
            stat_msg->stat =  NFX_ERR;
        }
        else
        {
            stat_msg->stat = get_lv_cfg_byid(((evm_MsgVolRename_t *)msgbuf)->id,
                                             &cfg, rtnstr);
            if (stat_msg->stat != NFX_OK)
                break;

            if (cfg.onLine)
            {
                sprintf(rtnstr, "Volume '%s' is online. Cannot rename it.", 
                        cfg.name);
                stat_msg->stat = NFX_ERR;
                break;
            }
                
            if (evm_isVolOk(&cfg, rtnstr) == FALSE)
            {
                sprintf(rtnstr, "Volume '%s' is broken. Cannot rename it.", 
                        cfg.name);
                stat_msg->stat = NFX_ERR;
                break;
            }
            
            memcpy(cfg.name, ((evm_MsgVolRename_t *)msgbuf)->newName,
                   VOL_MAX_NAME);
            stat_msg->stat = upd_lv_cfg(&cfg, 1, rtnstr);
        }
        break;
    } // case EVM_VOL_RENAME

    case EVM_SET_LV_OPTS:
    {
        evm_VolConfig_t     cfg;
        evm_MsgSetOpt_t *   msg = (evm_MsgSetOpt_t *)msgbuf;
				
        stat_msg->stat = get_lv_cfg_byid (msg->id, &cfg, rtnstr);
        if (stat_msg->stat == NFX_OK)
        {
            if (evm_isVolOk(&cfg, rtnstr) == TRUE)
            {
                cfg.maxSize = msg->maxSize;
                cfg.softQuota = msg->softQuota;
                cfg.minAutoGrow = msg->minAutoGrow;
                cfg.type = msg->type;
                stat_msg->stat = upd_lv_cfg(&cfg, 1, rtnstr);
            }
            else
            {
                sprintf(rtnstr, "Volume '%s' is broken. Cannot modify it.", 
                        cfg.name);
                stat_msg->stat = NFX_ERR;
            }
        }
        break;
    } // case EVM_SET_LV_OPTS

    case EVM_VOL_GROW:
    {
        evm_MsgVolStrg_t *  msg = (evm_MsgVolStrg_t *)msgbuf;
        evm_VolConfig_t     volCfg;

        /* Check if the volume exists and can grow.
         */
        stat_msg->stat = get_lv_cfg_byid(msg->id, &volCfg, rtnstr);
        if (stat_msg->stat != NFX_OK)
        {
            break;
        }
        
        if (evm_isVolOk(&volCfg, rtnstr) == FALSE)
        {
            sprintf(rtnstr, "Volume '%s' is broken. Cannot grow it.",
                    volCfg.name);
            stat_msg->stat = NFX_ERR;
            break;
        }

        if (volCfg.type == EVM_TYPE_COREVOL)
        {
            sprintf(rtnstr, "%s (0x%llx) is a corevolume which shall not grow.",
                    volCfg.name, msg->id.whole_thing);
            EVM_ERR_PRINT(("%s", rtnstr));
            stat_msg->stat = NFX_ERR;
            break;
        }

        /* Grow volume with specified LUNs.
         */
        stat_msg->stat = evm_growVol(msg->id, 0, msg->nbLun,
                                     msg->lunList, 0, rtnstr);
        break;
    } // case EVM_VOL_GROW

    case EVM_GET_LV_STRG:
    {
        evm_MsgGetStrg_t *  req = (evm_MsgGetStrg_t *) msgbuf;
        evm_MsgLVStrg_t *   rsp = (evm_MsgLVStrg_t *) rtnmsg;
        uint32              rspSize;
        evm_volStrg_t *     strg = NULL;
        int                 stat;

        stat = get_lv_strg(req->id, &strg, rtnstr);
        if ((stat == NFX_OK) && (strg != NULL))
        {
            int i;

            rspSize = (sizeof(evm_MsgLVStrg_t) +
                       (strg->num * sizeof(evm_lvStrg_t)));
            EVM_ALLOC_ZERO(rsp, evm_MsgLVStrg_t, rspSize);
            rsp->hdr.msgType = EVM_LV_STRG_CFG;
            rsp->hdr.len = rspSize;
            rsp->hdr.clientPrivate = req->hdr.clientPrivate;
            rsp->id = req->id;
            rsp->nbStrg = strg->num;
            
            for (i = 0; i < strg->num; i++)
            {
                evm_StrgAttr_t * pattr = NULL;
            
                stat = get_pstrg_cfg(strg->au[i].strgId, &pattr, rtnstr);
                if ((stat == NFX_OK) && (pattr != NULL))
                {
                    strcpy(rsp->strg[i].name, pattr->name);
                    rsp->strg[i].wwn = pattr->wwn;
                    rsp->strg[i].lun = pattr->lun;
                    rsp->strg[i].au = strg->au[i];
                }
                else
                {
                    // do nothing
                    // LUN is probably broken
                }
                
                EVM_FREE(pattr);
            } // for
            
            stat = sendAgileMsg(-1, rsp, rspSize,
                                fromapp, fromcpu, fromslot, MGMTPLANE,
                                NULL, NULL, 0);
            EVM_FREE(rsp);
            EVM_FREE(strg);
            return;
        }

        stat_msg->stat = stat;
        EVM_FREE(rsp);
        EVM_FREE(strg);
        break;
        
    } // case EVM_GET_LV_STRG


    /* Physical Storage Manipulation Messages
     * --------------------------------------
     */

    case EVM_STRG_GET_ATTRIB:
    {
        if (((evm_MsgStrg_Get_Attrib_t *)msgbuf)->id == -1)
        {
            evm_MsgStrgList_t       list;
                
            memset(&list, 0, sizeof(list));
            stat_msg->stat = get_pstrg_list(&list, rtnstr);
            if (stat_msg->stat == NFX_OK)
            {
                list.hdr.msgType = EVM_STRG_LIST;
                list.hdr.len = sizeof(list);
                list.hdr.clientPrivate =
                    ((evm_msgHdr_t *)msgbuf)->clientPrivate;

                stat_msg->stat = sendAgileMsg(-1, &list, sizeof(list),
                                              fromapp, fromcpu,
                                              fromslot, MGMTPLANE,
                                              NULL, NULL, 0);

                if (stat_msg->stat == NFX_OK)
                {
                    return;	// All is kosher no need to error
                }
                sprintf(rtnstr, "error sending response %s",
                        strerror(errno));
            }
        }
        else
        {
            evm_MsgStrgVolCfg_t *   msg = NULL;
            int                     msgSize;
            int                     attrSize;
            evm_StrgAttr_t *        sattr = NULL;

            stat_msg->stat =
                get_pstrg_cfg(((evm_MsgStrg_Get_Attrib_t *)msgbuf)->id,
                              &sattr, rtnstr);
            if ((stat_msg->stat == NFX_OK) && (sattr != NULL))
            {
                if (sattr->nallocs <= 1)
                {
                    attrSize = 0;
                }
                else
                {
                    attrSize = (sattr->nallocs - 1) * sizeof(alloc_t);
                }
                msgSize = sizeof(evm_MsgStrgVolCfg_t) + attrSize;
                EVM_ALLOC_ZERO(msg, evm_MsgStrgVolCfg_t, msgSize);
                msg->hdr.msgType = EVM_STRG_VOL_CFG;
                stat_msg->hdr.len = msgSize;
                msg->hdr.clientPrivate =
                    ((evm_msgHdr_t *)msgbuf)->clientPrivate;
                msg->devId = ((evm_MsgStrg_Get_Attrib_t *)msgbuf)->id;
                memcpy((void *) &(msg->volAttr),
                       (const void *) sattr,
                       sizeof(evm_StrgAttr_t) + attrSize);

                stat_msg->stat = sendAgileMsg(-1, msg, msgSize, 
                                              fromapp, fromcpu,
                                              fromslot, MGMTPLANE,
                                              NULL, NULL, 0);
                if (stat_msg->stat == NFX_OK)
                {
                    EVM_FREE(sattr);
                    EVM_FREE(msg);
                    return;	// All is kosher no need to error
                }
                sprintf (rtnstr, "error sending response %s",
                         strerror(errno));
            }
            EVM_FREE(sattr);
            EVM_FREE(msg);
        }
        break;
    } // case EVM_STRG_GET_ATTRIB

    case EVM_STRG_EVT:
        break;

    case EVM_STRG_GET_ID:
        stat_msg->stat = 
            get_pstrg_id (((evm_MsgGetStrgId_t *)msgbuf)->name, 
                          &(stat_msg->payload.sId), rtnstr);
        break;
            
    /* Messages from FS.
     * -----------------
     */
    case EVM_FSYS_FULL:
    {
        fs_EvmMsgGrowFileSystem_t * req = (fs_EvmMsgGrowFileSystem_t *) msgbuf;
        evm_VolConfig_t             volCfg;
        evm_growPendingCb_t *       cb = NULL;
        int32       rc = NFX_OK;
        uint64      fsysSize = req->fsSize * (EFS_BLOCK_SIZE / EVM_LBLK_SIZE);

        EVM_DBG_PRINT(("EVM_FSYS_FULL: filesystem 0x%llx needs to grow"
                       "(fsysSize %lld)", req->volId.whole_thing, fsysSize));
        
        EVM_ALLOC(cb, evm_growPendingCb_t, sizeof(evm_growPendingCb_t));
        
        /* HighWaterMark has been reached.
         * Auto-grow the volume.
         */
        rc = evm_growVol(req->volId, fsysSize, 0, NULL, (address_t) cb, rtnstr);
        if (rc == NFX_OK)
        {
            // grow filesystem
            fs_MsgGrowfsReq     fsReq;

            EVM_FREE(cb);
            get_lv_cfg_byid(req->volId, &volCfg, rtnstr);
                
            bzero((void *) &fsReq, sizeof(fs_MsgGrowfsReq));
            fsReq.hdr.requestType = FS_REQ_GROWFS;
            fsReq.volId = req->volId;
            fsReq.newNumBlks = volCfg.currSize * EVM_LBLK_SIZE / EFS_BLOCK_SIZE;

            if (sendAgileMsg(-1, (void *) &fsReq, sizeof(fs_MsgGrowfsReq), 
                             fromapp, fromcpu, fromslot, MGMTPLANE,
                             NULL, NULL, 0) < 0)
            {
                EVM_ERR_PRINT(("EVM: error sending response %s",
                               strerror(errno)));
            }
            EVM_DBG_PRINT(("growReq sent from EVM to FS (newNumBlks %lld)",
                           fsReq.newNumBlks));
        }
        else if (rc == EVM_VOL_GROW_PENDING)
        {
            /* Fill the control block so we know who to callback
             * when the grow is done.
             */
            cb->action = EVM_FSYS_FULL;
            cb->volId = req->volId;
            strcpy(cb->fromapp, fromapp);
            cb->fromcpu = fromcpu;
            cb->fromslot = fromslot;
        }
        else
        {
            EVM_FREE(cb);
            EVM_ERR_PRINT(("Unable to autogrow volume 0x%llx: %s",
                           req->volId.whole_thing, rtnstr));
                
            /* It's an event when it cannot grow
             */
            if (get_lv_cfg_byid(((evm_MsgLVFull_t *)msgbuf)->lVol,
                                &volCfg, rtnstr) == NFX_OK)
            {
                evm_sendEvt(EVENT_VOL_FULL, req->volId, EVENT_STATE_UP);
            }
        }
        return;    // no Stat response is expected
    } // case FS_REQ_EVM_GROW
        
    /* Messages from event manager.
     * ----------------------------
     */
        
    case NFX_EVENT_POST:
    {
        nfx_eventMsg_t * evt = (nfx_eventMsg_t *) msgbuf;

        switch(evt->eventType)
        {
        case NFX_EVENT_CPU:
        {
            /* FP core 0 (FP1) will always come first.
             * So synchronize EVM and EVM server when it comes up.
             */
            if ((evt->param.cpuEvent.slot == NFX_SLOTTYPE_NFP) &&
                (evt->param.cpuEvent.cpu == NFX_NFP_FP1) &&
                (evt->param.cpuEvent.state == EVENT_STATE_UP))
            {
                int32 rc = NFX_OK;
                
                EVM_FCT_PRINT(("FP is up. Let's synchronize with EVM server!"));
                
                /* Synchronize EVM and EVM server = restart EVM.
                 */
                rc = evm_closeAllLun();
                rc = close_dbs();
                rc = init_dbs();
                
                stat_msg->stat = rc;
            }
            break;
        }
        case NFX_EVENT_VOLUME:
        {
            switch(evt->param.volEvent.event)
            {
            case EVENT_VOL_IOERROR:
            case EVENT_VOL_CONFLICT:
            case EVENT_VOL_TAKEOVER:
            case EVENT_VOL_CORRUPT:
            {
                /* EVM server put a volume offline.
                 * Update EVM database and notify CCM.
                 */
                evm_offlineVol(evt->param.volEvent.volId, FALSE, rtnstr);
            }
            default:
            {
                // do nothing!
            }
            } // switch
            return;   // no response expected
        }
        default:
        {
            EVM_ERR_PRINT(("EVM never subscribed for eventType %d.",
                           evt->eventType));
            NFX_ASSERT(FALSE);
            return;   // no response expected
        }
        } // switch
    } // case NFX_EVENT_POST
    
    /* Status Messages from evm_srvr
     * -----------------------------
     */

    case EVM_MSG_STAT:
        break;

    /* Events from SDM
     * ---------------
     */

    case SDM_MID_DEV_INFO_RSP:
    {
        sdm_devInfoRspMsg_t *sdm_msg=(sdm_devInfoRspMsg_t *) msgbuf;

#ifdef DEBUG
        {
            time_t now = time(NULL);
            printf("SDM_MID_DEV_INFO_RSP for LUN 0x%llx rc %d - %s",
                   sdm_msg->id, sdm_msg->rc, ctime(&now));
        }
#endif // DEBUG
        EVM_DBG_PRINT(("EVM: SDM_MID_DEV_INFO_RSP for LUN 0x%llx rc %d",
                       sdm_msg->id, sdm_msg->rc));
        if (sdm_msg->rc == SDM_OK)
        {
            evm_StrgAttr_t  newLun;

            evm_initLun(sdm_msg, &newLun);
            evm_addLun(&newLun, TRUE, rtnstr);
        }
        return;	// No response expected
    }

    case SDM_MID_DEV_DEL_REQ:
    {
        sdm_devDelReqMsg_t * sdm_msg = (sdm_devDelReqMsg_t *) msgbuf;

#ifdef DEBUG
        {
            time_t now = time(NULL);
            printf("SDM_MID_DEV_DEL_REQ for LUN 0x%llx - %s",
                   sdm_msg->id, ctime(&now));
        }
#endif // DEBUG
        EVM_DBG_PRINT(("EVM: SDM_MID_DEV_DEL_REQ for LUN 0x%llx",
                       sdm_msg->id));
        evm_delLun(sdm_msg, rtnstr);
        return;    // no response is expected
    } // case SDM_MID_DEV_DEL_REQ

    /* Messages from SPM.
     * ------------------
     */
    case SPM_ALLOC_LUN_RSP:
    {
        /* A grow was pending - here are the allocated LUNs...
         */
        spm_MsgAllocLunRsp_t *  spmRsp = (spm_MsgAllocLunRsp_t *) msgbuf;
        evm_growPendingCb_t *   cb = ((evm_growPendingCb_t *) (address_t)
                                      spmRsp->hdr.clientPrivate);
        evm_VolConfig_t         volCfg;

        EVM_FCT_PRINT(("asynchronous SPM_ALLOC_LUN_RSP: %d", spmRsp->result));

        /* SRM allocated more LUNs to grow.
         * Grow the volume with the new LUNs.
         */
        stat_msg->stat = spmRsp->result;
        if (stat_msg->stat == SPM_ALLOC_OK)
        {
            stat_msg->stat = get_lv_cfg_byid(spmRsp->volId, &volCfg, rtnstr);
            if (stat_msg->stat == NFX_OK)
            {
                stat_msg->stat = evm_addLunToVol(&volCfg, spmRsp, rtnstr);
                if (stat_msg->stat == NFX_OK)
                {
                    EVM_DBG_PRINT(("Volume %s (0x%llx) finally grown.",
                                   volCfg.name, spmRsp->volId));
                }
            }
        }
        else
        {
            stat_msg->stat = NFX_ERR;
            sprintf(rtnstr, "Volume 0x%llx cannot grow. No more LUNs.",
                    spmRsp->volId.whole_thing);
            EVM_ERR_PRINT(("%s", rtnstr));
        }
        
        /* Callback whoever need to be notified.
         */
        if (cb != NULL)
        {
            NFX_ASSERT(spmRsp->volId.whole_thing == cb->volId.whole_thing);
            
            switch (cb->action)
            {
            case EVM_VOL_CREATE:
            {
                /* Grow was triggered by a volume creation.
                 * If it failed, delete the empty volume.
                 * If it passed, online the volume if it was requested.
                 */
                if (stat_msg->stat != NFX_OK)
                {
                    evm_delVol(cb->volId, rtnstr);
                }
                else if (cb->online == TRUE)
                {
                    stat_msg->stat = evm_onlineVol(cb->volId, cb->nim, rtnstr);
                }

                if (cb->clientPrivate == NFX_STRG_MAGIC)
                {
                    /* Probably came from "lvol" CLI.
                     */
                    EVM_FREE(cb);
                    return; // No response expected
                }
                else
#ifdef ISABELLE // code to delete
                {
                    stat_msg->hdr.clientPrivate = cb->clientPrivate;
                    stat_msg->payload.lVolId = cb->volId;
                    strcpy(fromapp, cb->fromapp);
                    fromcpu = cb->fromcpu;
                    fromslot = cb->fromslot;
                    EVM_FREE(cb);
                    break;
                }
#endif // ISABELLE
                {
                    
                    ssize_t     ncmReqSize;
                    ncmMsg_t *  ncmReq = NULL;
                                        
                    stat_msg->hdr.clientPrivate = cb->clientPrivate;
                    stat_msg->payload.lVolId = cb->volId;
                    
                    /* Build NCM message.
                     */
                    ncmReqSize = (sizeof(ncmMsg_t) + 1 +
                                  sizeof(evm_MsgStat_t) + strlen(rtnstr) - 1);
                    EVM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
                    NFX_ASSERT(ncmReq != NULL);
                    ncmReq->msgType = NCM_SEND;
                    strcpy(ncmReq->appName, cb->fromapp);
                    strcpy(ncmReq->nodeName, "PCC");
                    ncmReq->rqResp = 0;    // no response expected
                    memcpy(ncmReq->appMsg, stat_msg,
                           sizeof(evm_MsgStat_t) + strlen(rtnstr) - 1);
                    
                    if (sendAgileMsg(-1, ncmReq, ncmReqSize,
                                     "ncm", NFX_SSC_UNIX, SSC_SLOT,
                                     MGMTPLANE, NULL, NULL, 0) < 0)
                    {
                        EVM_ERR_PRINT(("EVM: error sending response %s",
                                       strerror(errno)));
                    }
                    EVM_FREE(ncmReq);                    
                    EVM_FREE(cb);
                    return;
                }
            }
            case EVM_FSYS_FULL:
            {
                /* Grow was triggered by an autogrow from FS.
                 * Tell filesystem that we've grown.
                 */
                fs_MsgGrowfsReq     fsReq;
                
                get_lv_cfg_byid(cb->volId, &volCfg, rtnstr);
                
                bzero((void *) &fsReq, sizeof(fs_MsgGrowfsReq));
                fsReq.hdr.requestType = FS_REQ_GROWFS;
                fsReq.volId = cb->volId;
                fsReq.newNumBlks = volCfg.currSize *
                    EVM_LBLK_SIZE / EFS_BLOCK_SIZE;
                
                if (sendAgileMsg(-1, (void *) &fsReq, sizeof(fs_MsgGrowfsReq), 
                                 cb->fromapp,  cb->fromcpu,  cb->fromslot,
                                 MGMTPLANE, NULL, NULL, 0) < 0)
                {
                    EVM_ERR_PRINT(("EVM: error sending response %s",
                                   strerror(errno)));
                }
                EVM_DBG_PRINT(("growReq sent from EVM to FS (newNumBlks %lld)",
                               fsReq.newNumBlks));
                EVM_FREE(cb);
                return; // No response expected
            }
            default:
            {
                EVM_FREE(cb);
                return; // No response expected
            }
            } // switch
        }
        else
        {
            return; // No response expected
        }
	break;
    } // case SPM_ALLOC_LUN_RSP
    
    case EVM_GET_LUN_REQ:
    {
        evm_MsgGetLunReq_t *    req = (evm_MsgGetLunReq_t *) msgbuf;
        evm_MsgGetLunRsp_t      rsp;
        evm_StrgAttr_t *        lun;
        int32                   rc;

        EVM_DBG_PRINT(("EVM: EVM_GET_LUN_REQ for LUN 0x%llx\n", req->id));

        bzero((void *) &rsp, sizeof(evm_MsgGetLunRsp_t));
        rsp.hdr.msgType = EVM_GET_LUN_RSP;
        rsp.hdr.len = sizeof(evm_MsgGetLunRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
            
        rc = get_pstrg_cfg(req->id, &lun, rtnstr);    
        if ((rc == NFX_OK) && (lun != NULL))
        {
            rsp.state = EVM_USED_LUN;         
            rsp.volId = lun->allocs[0].lvid;
            strcpy(rsp.cluster, myCluster);
            goto evm_get_lun_req_exit;
        }
        else
        {
            evm_strgCfg_t   odCfg;
                
            EVM_ALLOC_ZERO(lun, evm_StrgAttr_t, sizeof(evm_StrgAttr_t));
            
            rc = evm_getLunInfo(req->id, lun, rtnstr);
            if (rc != NFX_OK)
            {
                goto evm_get_lun_req_err;
            }
                
            rc = evm_openLun(req->id, SSC_SLOT, &(lun->openHandle));
            if ((rc != NFX_OK) || (lun->openHandle == -1))
            {
                goto evm_get_lun_req_err;
            }
                
            rc = evm_readLunLabel(req->id, lun, &odCfg);
            if (rc != NFX_OK)
            {
                goto evm_get_lun_req_err;
            }
                
            if ((odCfg.magic == NFX_STRG_MAGIC) &&
                (odCfg.version == EVM_CFG_VERSION))
            {
                if (odCfg.id.whole_thing == 0)
                {
                    if (!strcmp(odCfg.cluster, myCluster))
                    {
                        /* This LUN is a free LUN.
                         */
                        rsp.state = EVM_FREE_LUN;
                        rsp.volId.whole_thing = 0;
                        strcpy(rsp.cluster, myCluster);
                    }
                    else
                    {
                        /* This LUN is a free LUN but out of this cluster.
                         */
                        rsp.state = EVM_OC_FREE_LUN;
                        rsp.volId.whole_thing = 0;
                        strcpy(rsp.cluster, odCfg.cluster);
                    }                
                }
                else
                {
                    if (!strcmp(odCfg.cluster, myCluster))
                    {
                        /* This LUN is part of a volume.
                         */
                        EVM_ERR_PRINT(("LUN 0x%llx is part of a volume, "
                                       "but is not in our LUN db!",
                                       req->id));
                    }
                    else
                    {
                        /* This LUN is a used LUN but out of this cluster.
                         */
                        rsp.state = EVM_OC_USED_LUN;
                        rsp.volId = odCfg.id;
                        strcpy(rsp.cluster, odCfg.cluster);
                    }
                }
            }
            else
            {
                rsp.state = EVM_FOREIGN_LUN;         
                rsp.volId.whole_thing = 0;
                strcpy(rsp.cluster, "");
            }
        }
            
      evm_get_lun_req_err:

        /* Close LUN.
         */
        if ((lun != NULL) && (lun->openHandle != -1))
        {
            evm_closeLun(lun->openHandle);
        }
            
      evm_get_lun_req_exit:

        if (lun != NULL)
        {
            rsp.id = lun->id;
            rsp.wwn = lun->wwn;
            rsp.lun =  lun->lun;
            rsp.raidLevel = lun->raidLevel;
            strcpy(rsp.name, lun->name);
            strcpy(rsp.vendor, lun->vendor);
            strcpy(rsp.model, lun->model);
            rsp.size = (uint64) lun->blkCnt;
        }
        rsp.result = rc;
            
        EVM_FREE(lun);

        if (sendAgileMsg(-1, (void *) &rsp, sizeof(evm_MsgGetLunRsp_t), 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s",
                           strerror(errno)));
        }

        return;    // no response is expected
    } // case EVM_GET_LUN_REQ

    case EVM_LABEL_LUN_REQ:
    {
        evm_MsgLabelLunReq_t *  req = (evm_MsgLabelLunReq_t *) msgbuf;
        evm_MsgLabelLunRsp_t    rsp;
        int32                   cnt;

        EVM_DBG_PRINT(("EVM: EVM_LABEL_LUN_REQ for %d LUN(s)\n", req->idCnt));

        bzero((void *) &rsp, sizeof(evm_MsgLabelLunRsp_t));
        rsp.hdr.msgType = EVM_LABEL_LUN_RSP;
        rsp.hdr.len = sizeof(evm_MsgLabelLunRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        for (cnt = 0; cnt < req->idCnt; cnt++)
        {
            EVM_DBG_PRINT(("EVM: label req for LUN 0x%llx\n", req->id[cnt]));
            rsp.result = evm_labelLun(req->id[cnt], req->cluster, rtnstr);
        }
            
        if (sendAgileMsg(-1, (void *) &rsp, sizeof(evm_MsgLabelLunRsp_t), 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s",
                           strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_LABEL_LUN_REQ

    case EVM_UNLABEL_LUN_REQ:
    {
        evm_MsgUnlabelLunReq_t *    req = (evm_MsgUnlabelLunReq_t *) msgbuf;
        evm_MsgUnlabelLunRsp_t      rsp;
        int32                       cnt;

        EVM_DBG_PRINT(("EVM: EVM_UNLABEL_LUN_REQ for # LUN 0x%llx\n",
                       req->idCnt));

        bzero((void *) &rsp, sizeof(evm_MsgUnlabelLunRsp_t));
        rsp.hdr.msgType = EVM_UNLABEL_LUN_RSP;
        rsp.hdr.len = sizeof(evm_MsgUnlabelLunRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        for (cnt = 0; cnt < req->idCnt; cnt++)
        {
            EVM_DBG_PRINT(("EVM: unlabel req for LUN 0x%llx\n", req->id[cnt]));
            rsp.result = evm_unlabelLun(req->id[cnt], rtnstr);
        }
        if (sendAgileMsg(-1, (void *) &rsp, sizeof(evm_MsgUnlabelLunRsp_t), 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s",
                           strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_UNLABEL_LUN_REQ

    case SPM_ADD_LUN_RSP:
    {
        // ignore response
        return;
    }

    case EVM_RESCAN_LUN_REQ:
    {
        evm_MsgRescanLunReq_t * req = (evm_MsgRescanLunReq_t *) msgbuf;

        evm_rescanLun(req->id, rtnstr);
        return; // no response is expected
    } // EVM_RESCAN_LUN_REQ
        
    case EVM_MOD_LUN_CLUSTER_REQ:
    {
        evm_MsgModLunClusterReq_t * req = (evm_MsgModLunClusterReq_t *) msgbuf;
        evm_MsgModLunClusterRsp_t   rsp;
        int32                       cnt;

        EVM_DBG_PRINT(("EVM: EVM_MOD_LUN_CLUSTER_REQ for %d LUN(s)\n",
                       req->idCnt));

        bzero((void *) &rsp, sizeof(evm_MsgModLunClusterRsp_t));
        rsp.hdr.msgType = EVM_MOD_LUN_CLUSTER_RSP;
        rsp.hdr.len = sizeof(evm_MsgModLunClusterRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        for (cnt = 0; cnt < req->idCnt; cnt++)
        {
            EVM_DBG_PRINT(("EVM: modify cluster of LUN 0x%llx\n",
                           req->id[cnt]));
            rsp.result = evm_modLunCluster(req->id[cnt], req->cluster, rtnstr);
            evm_rescanLun(req->id[cnt], rtnstr);
        }
            
        if (sendAgileMsg(-1, (void *) &rsp, sizeof(evm_MsgModLunClusterRsp_t), 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_MOD_LUN_CLUSTER_REQ

    case EVM_READ_QUORUM_REQ:
    {
        evm_MsgReadQuorumReq_t *    req = (evm_MsgReadQuorumReq_t *) msgbuf;
        evm_MsgReadQuorumRsp_t      rsp;
        evm_StrgAttr_t              lun;
        boolean                     closeLun = FALSE;
        sdm_readDataRspMsg_t *      readRsp = NULL;

        /* Prepare the response.
         */
        rsp.hdr.msgType = EVM_READ_QUORUM_RSP;
        rsp.hdr.len = sizeof(evm_MsgReadQuorumRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;

        /* Get info about the specified quorum LUN.
         */
        rsp.result = evm_getLunInfo(req->lunId, &lun, rtnstr);
        if (rsp.result != NFX_OK)
        {
            goto evm_read_quorum_exit;
        }

        /* Open it if not alread opened (ie. part od a volume).
         */
        if (lun.openHandle == -1)
        {
            rsp.result = evm_openLun(lun.id, SSC_SLOT, &(lun.openHandle));
            if (rsp.result != NFX_OK)
            {
                sprintf(rtnstr, "Cannot open LUN (0x%llx).", lun.id);
                goto evm_read_quorum_exit;
            }
            closeLun = TRUE;
        }

        /* Read the quorum area of the LUN.
         */
        rsp.result = evm_readQuorumArea(&lun, &readRsp);
        if ((rsp.result != NFX_OK) || (readRsp == NULL))
        {
            goto evm_read_quorum_exit;
        }
        memcpy(&(rsp.quorum), readRsp->data, sizeof(evm_qArea_t));
        EVM_FREE(readRsp);

      evm_read_quorum_exit:

        /* Close the LUN if we opened it.
         */
        if (closeLun == TRUE)
        {
            evm_closeLun(lun.openHandle);
        }

        /* Send the response to the caller.
         */
        if (sendAgileMsg(-1, (void *) &rsp, rsp.hdr.len, 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_READ_QUORUM_REQ
    
    case EVM_WRITE_QUORUM_REQ:
    {
        evm_MsgWriteQuorumReq_t *   req = (evm_MsgWriteQuorumReq_t *) msgbuf;
        evm_MsgWriteQuorumRsp_t     rsp;
        evm_StrgAttr_t              lun;
        boolean                     closeLun = FALSE;

        /* Prepare the response.
         */
        rsp.hdr.msgType = EVM_WRITE_QUORUM_RSP;
        rsp.hdr.len = sizeof(evm_MsgWriteQuorumRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;

        /* Get info about the specified quorum LUN.
         */
        rsp.result = evm_getLunInfo(req->lunId, &lun, rtnstr);
        if (rsp.result != NFX_OK)
        {
            goto evm_write_quorum_exit;
        }

        /* Open it if not alread opened (ie. part od a volume).
         */
        if (lun.openHandle == -1)
        {
            rsp.result = evm_openLun(lun.id, SSC_SLOT, &(lun.openHandle));
            if (rsp.result != NFX_OK)
            {
                sprintf(rtnstr, "Cannot open LUN (0x%llx).", lun.id);
                goto evm_write_quorum_exit;
            }
            closeLun = TRUE;
        }

        /* Write the given data to the quorum area of the LUN.
         */
        rsp.result = evm_writeQuorumArea(&lun, &(req->quorum));

      evm_write_quorum_exit:

        /* Close the LUN if we opened it.
         */
        if (closeLun == TRUE)
        {
            evm_closeLun(lun.openHandle);
        }

        /* Send the response to the caller.
         */
        if (sendAgileMsg(-1, (void *) &rsp, rsp.hdr.len, 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_WRITE_QUORUM_REQ

    case EVM_READ_MIRROR_REQ:
    {
        evm_MsgReadMirrorReq_t *    req = (evm_MsgReadMirrorReq_t *) msgbuf;
        evm_MsgReadMirrorRsp_t      rsp;
        evm_volStrg_t *             strg = NULL;
        evm_StrgAttr_t              lun;
        sdm_readDataRspMsg_t *      readRsp = NULL;

        /* Prepare the response.
         */
        rsp.hdr.msgType = EVM_READ_MIRROR_RSP;
        rsp.hdr.len = sizeof(evm_MsgReadMirrorRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;

        /* Get the LUNs composing the specified volume.
         */
        rsp.result = get_lv_strg(req->volId, &strg, rtnstr);
        if ((rsp.result != NFX_OK) || (strg != NULL))
        {
            goto evm_read_mirror_exit;
        }
        NFX_ASSERT(strg->num >= 1);
        
        /* Get info about the first LUN of the volume.
         */
        rsp.result = evm_getLunInfo(strg->au[0].strgId, &lun, rtnstr);
        if (rsp.result != NFX_OK)
        {
            goto evm_read_mirror_exit;
        }
        NFX_ASSERT(lun.openHandle != -1);
        
        /* Read the mirror area of the LUN.
         */
        rsp.result = evm_readMirrorArea(&lun, &readRsp);
        if ((rsp.result != NFX_OK) || (readRsp == NULL))
        {
            goto evm_read_mirror_exit;
        }
        memcpy(&(rsp.mirror), readRsp->data, sizeof(evm_mirrorArea_t));
        EVM_FREE(readRsp);
        
      evm_read_mirror_exit:

        /* Send the response to the caller.
         */
        if (sendAgileMsg(-1, (void *) &rsp, rsp.hdr.len, 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_READ_MIRROR_REQ
    
    case EVM_WRITE_MIRROR_REQ:
    {
        evm_MsgWriteMirrorReq_t *   req = (evm_MsgWriteMirrorReq_t *) msgbuf;
        evm_MsgWriteMirrorRsp_t     rsp;
        evm_volStrg_t *             strg = NULL;
        evm_StrgAttr_t              lun;

        /* Prepare the response.
         */
        rsp.hdr.msgType = EVM_WRITE_MIRROR_RSP;
        rsp.hdr.len = sizeof(evm_MsgWriteMirrorRsp_t);
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;

        /* Get the LUNs composing the specified volume.
         */
        rsp.result = get_lv_strg(req->volId, &strg, rtnstr);
        if ((rsp.result != NFX_OK) || (strg != NULL))
        {
            goto evm_write_mirror_exit;
        }
        NFX_ASSERT(strg->num >= 1);
        
        /* Get info about the first LUN of the volume.
         */
        rsp.result = evm_getLunInfo(strg->au[0].strgId, &lun, rtnstr);
        if (rsp.result != NFX_OK)
        {
            goto evm_write_mirror_exit;
        }
        NFX_ASSERT(lun.openHandle != -1);
        
        /* Write the given data to the mirror area of the LUN.
         */
        rsp.result = evm_writeMirrorArea(&lun, &(req->mirror));
        
      evm_write_mirror_exit:

        /* Send the response to the caller.
         */
        if (sendAgileMsg(-1, (void *) &rsp, rsp.hdr.len, 
                         fromapp, fromcpu, fromslot, MGMTPLANE,
                         NULL, NULL, 0) < 0)
        {
            EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
        }
        
        return;    // no response is expected
    } // case EVM_WRITE_MIRROR_REQ

    default:
    {
        EVM_DBG_PRINT(("EVM: Eeek! Invalid message %d", 
                       ((evm_msgHdr_t *)msgbuf)->msgType));
        stat_msg->stat = NFX_ERR;
        sprintf (rtnstr, "Invalid message %d",
                 ((evm_msgHdr_t *)msgbuf)->msgType);
    } // default
    } // switch

    /* Return message to the sender of this message
     */
    rtnsize = sizeof(evm_MsgStat_t);
    if (stat_msg->stat != NFX_OK)
    {
        // Could be an error message
        rtnsize += strlen(rtnstr) + 1; // message and its NULL
    }
    stat_msg->hdr.len = rtnsize;
    
    if (sendAgileMsg(-1, rtnmsg, rtnsize, 
                     fromapp, fromcpu, fromslot, MGMTPLANE,
                     NULL, NULL, 0) < 0)
    {
        EVM_ERR_PRINT(("EVM: error sending response %s", strerror(errno)));
    }
    return;
}

/*----------------------------------------------------------------------------*/
