/*------------------------------------------------------------------------------
 * RCS:  $Id: evm_com_defs.h,v 1.49 2002/11/07 01:19:33 rick Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm_com_defs.h
 *
 * Description  : Volume Manager (EVM) common definitions.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 27-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

#ifndef _EVM_COM_DEF_H_
#define _EVM_COM_DEF_H_

#include "nfx-incl.h"
#include "exports/eee-api.h"
#include "exports/eee-desc.h"	// For eee_descPtr_t
#include "exports/scsi-api.h"	// For scsi_id_t
#include "exports/sdm-api.h"	// For scsi_id_t
#include "exports/opt-api.h"	// For scsi_id_t

/*------------------------------------------------------------------------------
 * Constants definition. 
 *------------------------------------------------------------------------------
 */

#define EVM_LBLK_SIZE           512
#define EVM_MEGA_BYTES          (1024 * 1024)
#define EVM_BLK_PER_MBYTES      (EVM_MEGA_BYTES / EVM_LBLK_SIZE)
#define DEFAULT_MIN_AUTOGROW	(100 * EVM_BLK_PER_MBYTES)

#define EVM_CFG_VERSION     2           // config version number
#define NFX_STRG_MAGIC      0xbabebeef  // private tag
#define EVM_MAX_VENDOR_LIST 10          // set some limits on vendor/model list

#define EVM_MAX_NODE_NAME     256
#define EVM_MAX_APP_STR       256
#define EVM_MAX_ERR_STR       256
#define EVM_MAX_BUF_SIZE    32768
#define	VOL_MAX_NAME		  256
#define MAX_IDS              1000 // TODO!!! rework
#define MAX_PSTRG             500 // TODO!!! rework

// Support for automatic storage pool creation based on SDM LUN attributes
#define EVM_VENDOR_NAME_LEN 9   // SDM_VENDOR_NAME_LEN+1
#define EVM_MODEL_NAME_LEN  17  // SDM_MODEL_NAME_LEN+1
#define EVM_DONT_CARE       0xffffffff

/*------------------------------------------------------------------------------
 * Types definition.
 *------------------------------------------------------------------------------
 */

/* Volume identifier.
 */
typedef	union
{
    int64   whole_thing;
    struct {
        union {
            int32   id;
            struct {
                unsigned    clusterUnique:26;   // unique number in cluster
                unsigned    reserved:6;         // for filesystem use
            } parts;
        } u;
        int32   worldUnique;                    // unique number in the world
    } details;  
} evm_volId_t;

/* Logical volumes can be really big.
 */
typedef	uint64	evm_volOffset_t;
typedef	uint64	evm_volSize_t;

/* Handle to the data.
 */
typedef	union
{
    void *      dataPtr;    // Pointer to data to read/write
    nfx_cod_t	codHandle;	// Where in the SAC it is
} evm_handle_t;

/* RAID levels.
 */
typedef enum
{
    EVM_RAID0 = 0,      // stripping
    EVM_RAID1 = 1,      // mirror
    EVM_RAID01 = 2,     // stripe + mirror
    EVM_RAID3 = 3,      // raid 3
    EVM_RAID5 = 5,      // raid 5
    EVM_RAID_NONE = 6,  // no raid level, ie plain
    EVM_RAID_ANY = 0xffff
} evm_raid_t;

/* Physical storage allocation.
 */
typedef	struct
{
    scsi_id_t			strgId;		// Phys Storage volume
    int32				openHandle;	// If it's open, this is its handle
    uint32				startBlk;	// starting at this block
    int64				auSize;		// how many blocks this alloc unit
    uint32              totalSize;  // total LUN size
} evm_strgConfig_t;

/*------------------------------------------------------------------------------
 * Logical volume storage record.
 *------------------------------------------------------------------------------
 */
typedef struct
{
    int					num;		// how many 'au's follow?
    evm_strgConfig_t	au[1];
} evm_volStrg_t;
#define VOLSTRG_SIZE(x) (sizeof(evm_volStrg_t) + ((((x)->num == 0)? 0: ((x)->num - 1)) * sizeof(evm_strgConfig_t)))

typedef	struct
{
    scsi_id_t		strgId;		// This physical device
    uint32			startBlk;	// Starting here
    uint32			numBlks;	// and for this many blocks
} evm_PoolStrg_t;

typedef struct
{
    evm_volId_t		lvid;
    uint32			stBlk;
    uint32			nBlks;          // nb blocks available for the filesystem
} alloc_t;

typedef	struct
{
    scsi_id_t       id;                 // SCSI id
	uint64          wwn;                // WWN
	uint16          lun;                // LUN
    int32			openHandle;         // When open, this is its handle
    char			name[VOL_MAX_NAME];	// We call it this
    uint32			blkSize;            // 512 bytes
    uint32          totalBlkCnt;        // total numbers of blocks in LUN
    uint32			blkCnt;             // blocks available for the filesystem
    char            vendor[EVM_VENDOR_NAME_LEN];
    char            model[EVM_MODEL_NAME_LEN];
    evm_raid_t      raidLevel;
    int				onLine;		// Is online yes?
    int				nallocs;	// to logical volume(s)
    alloc_t			allocs[1];
} evm_StrgAttr_t;
#define STRG_ATTR_SIZE(x) (sizeof(evm_StrgAttr_t) + ((((x)->nallocs==0)?0:((x)->nallocs - 1) * sizeof(alloc_t))))

/* Logical volume's storage.
 */
typedef struct
{
	char                name[SDM_DEV_NAME_LEN+1];   // device name
	uint64              wwn;                        // WWN
	uint16              lun;                        // LUN
    evm_strgConfig_t    au;                         // 
} evm_lvStrg_t;

/* LUN states.
 */
typedef enum
{
    EVM_FREE_LUN,       // LUN is in this cluster and is free
    EVM_USED_LUN,       // LUN is in this cluster and is part of a volume
    EVM_OC_FREE_LUN,    // LUN is out of this cluster and is free
    EVM_OC_USED_LUN,    // LUN is out of this cluster and is part of a volume
    EVM_FOREIGN_LUN     // LUN is not an Agile LUN
} evm_lunState_t;

/* EVM server statistics.
 */
typedef struct
{
#define EVM_AMSEC    100    // msec
#define EVM_BMSEC    300    // msec
#define EVM_CMSEC    400    // msec
#define EVM_DMSEC    900    // msec
#define EVM_EMSEC   1350    // msec
#define EVM_FMSEC   3200    // msec
    uint64  moreThanAmsec;
    uint64  moreThanBmsec;
    uint64  moreThanCmsec;
    uint64  moreThanDmsec;
    uint64  moreThanEmsec;
    uint64  moreThanFmsec;
    eeeRTC_t  maxTime;
    eeeRTC_t  sumTime;
    uint64  numTime;
} evm_timeTab_t;

typedef struct
{
    int32   metaFree;     // number of free elements
    int32   metaPioWait;  // number of waiting physical IO requests
    int32   metaPioPend;  // number of pending physical IO requests
    int32   userFree;     // number of free elements
    int32   userPioWait;  // number of waiting physical IO requests
    int32   userPioPend;  // number of pending physical IO requests
} evm_queueCnt_t;

typedef struct
{
    uint64  nIOReqs;            // calls to evm_IOReq
    uint64  nIOCompletes;       // callbacks issued after IOReq completion
    uint64  nIORejected;        // I/O requests rejected
    uint64  nMReads;            // META READ requests
    uint64  nMReadsRetry;       // META READ requests retries
    uint64  nMWrites;           // META WRITE requests
    uint64  nMWritesRetry;      // META WRITE requests retries
    uint64  nUReads;            // USER READ requests
    uint64  nUReadsRetry;       // USER READ requests retries
    uint64  nUWrites;           // USER WRITE requests
    uint64  nUWritesRetry;      // USER WRITE requests retries
    uint64  nIOEvents;          // IO event EEE messages
    uint64  nScsiMReadRsps;     // SCSI Read responses
    uint64  nScsiMWriteRsps;    // SCSI Write responses
    uint64  nScsiUReadRsps;     // SCSI Read responses
    uint64  nScsiUWriteRsps;    // SCSI Write responses
    uint64  nScsiDevErrs;       // SCSI Device error responses
    uint64  nAllocs;            // calls to eee_ramDealloc
    uint64  nDeallocs;          // calls to eee_ramDealloc
    uint64  nTimerCreates;      // calls to eee_timerCreate
    uint64  nTimerDestroys;     // calls to eee_timerDestory
    uint64  nErrors;            // system errors: eg memory, hash, etc
    uint64  nTimeouts;          // IO timeouts
    uint64  nStales;            // stale IO responses
    uint64  nEEEDropPacket;     // EEE_DROP_PACKET
    uint64  nEEETxQueueFull;    // EEE_TX_QUEUE_FULL
    uint64  nBadPacketLength;   // 
    uint64  nBlkMRead;          // nb EVM blocks read (meta data)
    uint64  nBlkMWrite;         // nb EVM blocks written (meta data)
    uint64  nBlkURead;          // nb EVM blocks read (user data)
    uint64  nBlkUWrite;         // nb EVM blocks written (user data)

    struct{
        uint64  metaReadReq;
        uint64  metaReadRsp;
        uint64  metaWriteReq;
        uint64  metaWriteRsp;
        uint64  userReadReq;
        uint64  userReadRsp;
        uint64  userWriteReq;
        uint64  userWriteRsp;
    } ioPerFc[1][4];            // 1 FC-NIM with 4 slices
    
    uint32  nMetaMaxPioWait;    // waiting phys. meta requests' high water mark
    uint32  nMetaMaxPioPend;    // pending phys. meta requests' high water mark
    uint32  nUserMaxPioWait;    // waiting phys. user requests' high water mark
    uint32  nUserMaxPioPend;    // pending phys. user requests' high water mark
    uint32  nMaxPioWait;        // waiting phys. requests' high water mark
    uint32  nMaxPioPend;        // pending phys. requests' high water mark
    
    uint64  maxQueueDepth;      // SCSI queue depth
    uint64  sumQueueDepth;      // SCSI queue depth
    uint64  numQueueDepth;      // SCSI queue depth
	uint64  count_per_usec;     // RTC conversion factor

    evm_timeTab_t   MReadTime;  // meta read response time
    evm_timeTab_t   MWriteTime; // meta write response time
    evm_timeTab_t   UReadTime;  // user read response time
    evm_timeTab_t   UWriteTime; // user write response time
    
    evm_queueCnt_t  qCnt;       // should be the last elt (stat clear)
} evm_stats_t;

/*------------------------------------------------------------------------------
 * Storage Configuration Interface.
 *
 *                          "THE TRUTH IS ON DISK"
 *
 * No configuration data are stored in the NFX, but on the network itself.
 * This simplify the volume management by avoiding inconsistency issues
 * and providing a greater flexibility to the whole system.
 *
 * The configuration data are stored on the disk and are hiden to the file
 * system. The different categories of data are stored on 3 areas on the volume:
 *
 * I. LUN area
 * -----------
 * It contains the minimum necessary configuration data to re-build a volume
 * from its LUNs.
 * This data are stored on each LUN of the volume.
 *
 * evm_strgCfg_t contains the on disk private meta data:
 *
 *  1. include enough of the evm_VolConfig_t struct to reconfigure
 *     the list of configured volumes
 *  2. nLvols will always be one since LUN subdivisions are not supported
 *  3. if LUN is not assigned to any volume, the volume id in
 *     evm_VolConfig_t is set to 0 and lVols is not valid
 *
 * Since we have limited private meta data space on each LUN, certain
 * in-memory (via DB) configuration data cannot be saved.  For example, all
 * LUNs assigned to a logical volume.  As LUNs are discovered via SDM,
 * we can count and compare if this matches the total count saved in 
 * the meta data.
 *
 * When "on disk" meta data is written or modified:
 *
 *  1. when a LUN is allocated to a logical volume
 *  2. when a logical volume is destroyed
 *  3. TBD: when attributes of a logical volume (eg name, size) are modified
 *
 * logical volume configurations are NOT saved when NO LUN are assigned to them
 *
 * II. Manager area
 * ----------------
 * It contains the volume configuration that is needed once the volume is
 * complete (e.g. Agile Mirror configuration).
 * These data are stored on the first LUN of the volume,
 * they do not need to be saved on each LUN composing the volume.
 *
 * This area also contains a timestamp which is periodically updated by the
 * NFX owner to show that a volume is still managed by an NFX (cf. clustering).
 * Data stored in the "volume area" are maintained by the NFX owner.
 * The takeover NFX access this area in READ ONLY.
 *
 * III. Takeover area
 * ------------------
 * As the data of the "volume area", takeover data are stored on the first LUN
 * of the volume.
 *
 * This area contains a timestamp which is updated by a NFX which wants to
 * take over the volume.
 * The owner NFX accesses this area in READ ONLY to see if somebody wants
 * to take over its volume.
 *
 *
 * Note: For the first release, we will not support any LUN subdivisions
 * ie the entire LUN is assigned to a storage pool.
 *
 *------------------------------------------------------------------------------
 */

typedef struct
{
    char    vendor[EVM_VENDOR_NAME_LEN];
    char    model[EVM_MODEL_NAME_LEN];
    byte    pad[6];
} evm_vendor_model_t;
NFX_ASSERT_STATIC(sizeof(evm_vendor_model_t) == 32);

typedef struct
{
    evm_raid_t          raidLevel;
    uint32              minLunSize; // # EVM blocks
    uint32              maxLunSize; // # EVM blocks
#define EVM_SAME_RAIDCTRL   1       // use LUNs of same raid controller
#define EVM_SAME_VENDOR     2       // use LUNs of same vendor only
#define EVM_SAME_MODEL      4       // use LUNs of same model only
    uint32              flags;
    char                raidCtrl[SDM_DEV_NAME_LEN]; // RAID controller
    uint32              num;                        // number of vendor/model
    uint32              pad;
    evm_vendor_model_t  vendorInfo[EVM_MAX_VENDOR_LIST];
} evm_strg_attributes_t;

typedef struct
{
    uint32      allocUnit;  // which alloc unit am I ??
    uint32      pad;
    uint32      mirrorNum;  // which mirror has it ??
    uint32      stripeNum;  // which position in line we are
    uint64      lblk;       // Starts at which logical block
    uint32      stBlk;      // Starts here (physical block)
    uint32      nBlks;      // and goes for this far
} lv_alloc_t;

/* Logical volume type.
 */
typedef enum
{
    EVM_TYPE_FSYS       = 0,
    EVM_TYPE_COREVOL    = 1,
    EVM_TYPE_MIRROR     = 2
} evm_volType_t;

/* Logical Volume Configuration definition. */
typedef struct
{
    char                    name[VOL_MAX_NAME]; // We call it this
    evm_volType_t           type;       // volume type
    uint32                  softQuota;  // % when to notify operator
    uint64                  maxSize;    // How big can it get
    uint64                  minAutoGrow;// Growth increment (blocks)
    evm_volId_t             id;         // How we identify it
    uint64                  currSize;   // How big it is now
    uint32                  onNIM;      // Served on this NIM
    uint32                  onLine;     // Online status
    evm_strg_attributes_t   attrs;      // LUN attributes vol was created with
} evm_VolConfig_t;

/* LUN area = volume configuration stored on each LUN of the volume.
 * Important note:
 * The structure contains 'dummy' fields to reserve some space for future grow.
 * This structure has to be backward compatible and cannot grow (= 2048 bytes)!
 */
typedef struct
{
    uint32          magic;              // Agile label
    uint32          version;            // version number
    char            cluster[256];       // cluster name
    
    // volume info
    evm_volId_t     id;                 // 64-bits volume id (0 = no volume)
    char            name[256];          // volume name
    uint32          creator;            // node which created the volume
    uint32          softQuota;          // soft quota (%)
    uint64          currSize;           // current volume size (in blocks)
    uint64          hardQuota;          // hard quota (in blocks)
    uint64          minAutoGrow;        // growth increment (in blocks)
    
    // storage attributes used to create/grow the volume
    evm_raid_t      raidLevel;          // 32-bits RAID level
    uint32          minLunSize;         // minimum LUN size (in blocks)
    uint32          maxLunSize;         // maximum LUN size (in blocks)
    uint32          flags;              // SAME_WWN/SAME_VENDOR/SAME_MODEL
    uint64          pad1;               // WWN has been removed
    uint32          nbVendor;           // number of vendor/model
    uint32          pad2;
    evm_vendor_model_t  vendorInfo[10]; // list of vendor/model
    
    // LUN info
    uint32          allocUnit;          // position which alloc unit am I?
    evm_volType_t   type;               // volume type
    uint64          lblk;               // starts at which logical block
    uint32          stBlk;              // starts here (physical block)
    uint32          nBlks;              // and goes for this far

    // more storage attributes used to create/grow the volume
    char            raidCtrl[64];       // RAID controller
    
    uint64          empty[130];

    int32           timestamp;          // last time label has been modified
    uint32          cksum;              // checksum
} evm_strgCfg_t;
NFX_ASSERT_STATIC(sizeof(evm_strgCfg_t) == 2048);

typedef struct
{
    char        cluster[256];   // cluster name
    char        node[256];      // host name
    uint32      chassisId;      // host id
    uint32      pad;
} evm_nfxId_t;

/* The manager private area is regurlarly updated by the node managing
 * the volume. It contains the manager unique id and a timestamp to show
 * that the volume is currently managed.
 * This private area will be on the first LUN of the volume.
 */
typedef struct
{
    uint32          magic;      // Agile label
    evm_nfxId_t     mgrId;      // chassis id of the manager
    int32           timestamp;  // tstamp periodically updated by the manager
    uint64          empty[61];
    uint32          pad;    
    uint32          cksum;      // checksum    
} evm_mgrArea_t;
NFX_ASSERT_STATIC(sizeof(evm_mgrArea_t) == 1024);

/* The takeover private area is writen when a node wishes to take over
 * the volume and that the manager node cannot be reached via its management
 * interfaces.
 * The idea is that the manager node will see that another node wants to
 * takeover its volume and stop managing it.
 * It contains the takeover node unique id and a timestamp.
 * This private area will be on the first LUN of the volume.
 */
typedef struct
{
    uint32          magic;      // Agile label
    evm_nfxId_t     tkoId;      // chassis id of the NFX wishing to takeover
    int32           timestamp;  // tstamp periodically updated by the takeover
    uint64          empty[61];
    uint32          pad;
    uint32          cksum;      // checksum    
} evm_tkoArea_t;
NFX_ASSERT_STATIC(sizeof(evm_tkoArea_t) == 1024);

/* Quorum area for further usage.
 */
typedef struct
{
    uint32          magic;
    uint32          version;
    evm_nfxId_t     pccId;
    int32           timestamp;
    uint32          vote;
    uint32          nodeList[16];
    byte            empty[420];
    uint32          cksum;
} evm_qArea_t;
NFX_ASSERT_STATIC(sizeof(evm_qArea_t) == 1024);

/* Mirror area.
 */
typedef struct
{
    uint32          magic;
    byte            empty[1016];
    uint32          cksum;
} evm_mirrorArea_t;
NFX_ASSERT_STATIC(sizeof(evm_mirrorArea_t) == 1024);

/* Reserved area for further usage.
 */
typedef struct
{
    byte    empty[(64*1024) -               \
                  sizeof(evm_strgCfg_t) -   \
                  sizeof(evm_mgrArea_t) -   \
                  sizeof(evm_tkoArea_t) -   \
                  sizeof(evm_qArea_t) -     \
                  sizeof(evm_mirrorArea_t)];
} evm_rsvArea_t;

/* Number of 512 bytes blocks needed for the configuration data.
 * Note: Needs to be an even number because of the checksum stuff
 *       Radek does down deep in SCSI
 */
#define EVM_LUN_AREA (((sizeof(evm_strgCfg_t) / 1024) +                     \
                      ((sizeof(evm_strgCfg_t) % 1024 == 0)? 0 : 1)) * 2)
NFX_ASSERT_STATIC(EVM_LUN_AREA == 4);
                      
#define EVM_MANAGER_AREA (((sizeof(evm_mgrArea_t) / 1024) +                 \
                          ((sizeof(evm_mgrArea_t) % 1024 == 0)? 0 : 1)) * 2)
NFX_ASSERT_STATIC(EVM_MANAGER_AREA == 2);
                      
#define EVM_TAKEOVER_AREA (((sizeof(evm_tkoArea_t) / 1024) +                \
                           ((sizeof(evm_tkoArea_t) % 1024 == 0)? 0 : 1)) * 2)
NFX_ASSERT_STATIC(EVM_TAKEOVER_AREA == 2);

#define EVM_QUORUM_AREA (((sizeof(evm_qArea_t) / 1024) +                    \
                         ((sizeof(evm_qArea_t) % 1024 == 0)? 0 : 1)) * 2)
NFX_ASSERT_STATIC(EVM_QUORUM_AREA == 2);

#define EVM_MIRROR_AREA (((sizeof(evm_mirrorArea_t) / 1024) +               \
                         ((sizeof(evm_mirrorArea_t) % 1024 == 0)? 0 : 1)) * 2)
NFX_ASSERT_STATIC(EVM_MIRROR_AREA == 2);

#define EVM_RESERVED_AREA (((sizeof(evm_rsvArea_t) / 1024) +                \
                           ((sizeof(evm_rsvArea_t) % 1024 == 0)? 0 : 1)) * 2)
NFX_ASSERT_STATIC(EVM_RESERVED_AREA == (128 -                   \
                                        EVM_LUN_AREA -          \
                                        EVM_MANAGER_AREA  -     \
                                        EVM_TAKEOVER_AREA -     \
                                        EVM_QUORUM_AREA -       \
                                        EVM_MIRROR_AREA));

#define N_BLKS_CFG  (EVM_LUN_AREA + EVM_MANAGER_AREA + EVM_TAKEOVER_AREA +  \
                     EVM_QUORUM_AREA + EVM_MIRROR_AREA + EVM_RESERVED_AREA)

/*----------------------------------------------------------------------------*/

#endif // _EVM_COM_DEF_H_
