/*------------------------------------------------------------------------------
 * RCS:  $Id: evm_db.c,v 1.58 2002/12/05 18:52:32 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm_db.c
 *
 * Description  : This file contains the methods used to manipulate
 *                EVM's Volume Configuration database. There is a single
 *                instance of these on each SSC.
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 02-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <db.h>

#include "evm.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

static uint64         evm_txn_seq_id = 1; // transaction id
static struct timeval sdm_to = {5, 0}; // Wait 5 sec for SDM to respond
static struct timeval nimto = {10, 0}; // Wait 10 sec for EVM server to respond

char evm_raid_level_strings[7][32] =
{
    "raid 0",
    "raid 1",
    "raid 2",
    "raid 3",
    "invalid raid level",
    "raid 5",
    "none"};

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

//static void display_od_cfg(evm_strgCfg_t *odCfg);

static uint64 get_od_txn_id();

static int cmp_vol_cfg(scsi_id_t        id,
                       evm_StrgAttr_t   *sAttr,
                       evm_VolConfig_t  *cfg,
                       evm_VolConfig_t  *odCfg,
                       char             *err);

static void
evm_strgCfg2VolConfig(evm_strgCfg_t *    odCfg,
                      evm_VolConfig_t *  volConfig);

static void
evm_VolConfig2strgCfg(evm_VolConfig_t *  volConfig,
                      evm_strgCfg_t *    odCfg);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_strgCfg2VolConfig
 *
 * Description  : evm_strgCfg_t -> evm_VolConfig_t
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 13-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_strgCfg2VolConfig(evm_strgCfg_t *    odCfg,     // input
                      evm_VolConfig_t *  volConfig) // output

{
    NFX_ASSERT((odCfg != NULL) && (volConfig != NULL));
    
    bzero((void *) volConfig, sizeof(evm_VolConfig_t));
    strcpy(volConfig->name, odCfg->name);
    volConfig->type = odCfg->type;
    volConfig->softQuota = odCfg->softQuota;
    volConfig->maxSize = odCfg->hardQuota;
    volConfig->minAutoGrow = odCfg->minAutoGrow;
    volConfig->id = odCfg->id;
    volConfig->currSize = odCfg->currSize;
    //volConfig->onNIM;
    //volConfig->onLine;
    volConfig->attrs.raidLevel = odCfg->raidLevel;
    volConfig->attrs.minLunSize = odCfg->minLunSize;
    volConfig->attrs.maxLunSize = odCfg->maxLunSize;
    volConfig->attrs.flags = odCfg->flags;
    strcpy(volConfig->attrs.raidCtrl, odCfg->raidCtrl);
    volConfig->attrs.num = odCfg->nbVendor;
    memcpy((void *) &(volConfig->attrs.vendorInfo),
           (const void *) (odCfg->vendorInfo),
           EVM_MAX_VENDOR_LIST * sizeof(evm_vendor_model_t));

    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_VolConfig2strgCfg
 *
 * Description  : evm_VolConfig_t -> evm_strgCfg_t
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 13-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void
evm_VolConfig2strgCfg(evm_VolConfig_t *  volConfig, // input
                      evm_strgCfg_t *    odCfg)     // output

{
    NFX_ASSERT((volConfig != NULL) && (odCfg != NULL));
    
    odCfg->id = volConfig->id;
    strcpy(odCfg->name, volConfig->name);
    odCfg->type = volConfig->type;
    odCfg->currSize = volConfig->currSize;
    odCfg->hardQuota = volConfig->maxSize;
    odCfg->softQuota = volConfig->softQuota;
    odCfg->minAutoGrow = volConfig->minAutoGrow;
    odCfg->raidLevel = volConfig->attrs.raidLevel;
    odCfg->minLunSize = volConfig->attrs.minLunSize;
    odCfg->maxLunSize = volConfig->attrs.maxLunSize;
    odCfg->flags = volConfig->attrs.flags;
    strcpy(odCfg->raidCtrl, volConfig->attrs.raidCtrl);
    odCfg->nbVendor = volConfig->attrs.num;
    memcpy((void *) &(odCfg->vendorInfo),
           (const void *) (volConfig->attrs.vendorInfo),
           EVM_MAX_VENDOR_LIST * sizeof(evm_vendor_model_t));

    return;
}

/*------------------------------------------------------------------------------
 * Name         : display_od_cfg
 *
 * Description  : Display our config data
 *
 * Created by   : Lan Chin
 *
 * Date Created : 24-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

//static void
void
display_od_cfg(evm_strgCfg_t *  odCfg)
    
{
    int i;
    
    EVM_DBG_PRINT(("ON DISK config (version %d):\n", odCfg->version));
    
    if (odCfg->version != EVM_CFG_VERSION)
    {
        return;
    }
    
    printf("*** Volume label ***\n");
    printf("\tmagic: 0x%x\n", odCfg->magic);
    printf("\tversion: %d\n", odCfg->version);
    printf("\tcluster: %s\n", odCfg->cluster);
    printf("\tcreator: 0x%x\n", odCfg->creator);
    printf("\tid: 0x%llx\n", odCfg->id.whole_thing);
    printf("\tname: %s\n", odCfg->name);
    printf("\tcurrSize: %lld\n", odCfg->currSize);
    printf("\thardQuota: %lld\n", odCfg->hardQuota);
    printf("\tsoftQuota: %d\n", odCfg->softQuota);
    printf("\tminAutoGrow: %lld\n", odCfg->minAutoGrow);
    
    printf("\traidLevel: %s\n", (odCfg->raidLevel == EVM_RAID_ANY) ?
           "no raid level" : evm_raid_level_strings[odCfg->raidLevel]);
    printf("\tminLunSize: %d\n", odCfg->minLunSize);
    printf("\tmaxLunSize: %d\n", odCfg->maxLunSize);
    //printf("\twwn: 0x%llx\n", odCfg->wwn);
    printf("\traidController: %s\n", odCfg->raidCtrl);
    printf("\tnbVendor: %d\n", odCfg->nbVendor);
    for (i = 0; i < odCfg->nbVendor; i++)
    {
        printf("\tvendor/model %s/%s\n",
               (strlen(odCfg->vendorInfo[i].vendor) ?
                odCfg->vendorInfo[i].vendor : "no vendor name"),
               (strlen(odCfg->vendorInfo[i].model) ?
                odCfg->vendorInfo[i].model : "no model name"));
    } // for

    printf("\tallocUnit: %d\n", odCfg->allocUnit);
    printf("\tlblk: %lld\n", odCfg->lblk);
    printf("\tstBlk: %d\n", odCfg->stBlk);
    printf("\tnBlks: %d\n", odCfg->nBlks);

    printf("\ttimestamp: %d\n", odCfg->timestamp);
    printf("\tcksum: 0x%x\n", odCfg->cksum);
    return;
}

/*-----------------------------------------------------------------
 * Name: get_od_txn_id
 *
 * Description: Get our unique transaction id for this on disk operation
 *
 * This id will be used to resolve partial updates of config data
 * Ids with a greater value (timestamp) will assumed to contain the
 *  most recent update
 *
 * Created by: Lan Chin
 *
 * Date Created: 02 Nov 2001
 *
 *-----------------------------------------------------------------
 */

static uint64 get_od_txn_id()
{
    /*
     * TBD:
     * until we get tod (time of day), just return a unique sequence number
     */

    return evm_txn_seq_id++;
}

/*------------------------------------------------------------------------------
 * Name         : upd_vol_for_lun
 *
 * Description  : Update the LUN's on disk vol config
 *
 * Called during reconfig of the memory vol DB rec at boot time when
 * a conflict is encountered between an on disk version and the one
 * in memory.  This is called when the in memory version is assumed
 * to be the correct one and the one on disk needs to be updated.
 *
 * Created by   : Lan Chin
 *
 * Date Created : 05-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

int
upd_vol_for_lun(scsi_id_t           id,     // lun to be updated
                evm_StrgAttr_t *    sAttr,  // info to do disk write
                evm_VolConfig_t *   cfg)    // update with this
{
    evm_strgCfg_t   odCfg;

    EVM_DBG_PRINT(("upd_vol_for_lun(): LUN %llx", id));

    if (sAttr->openHandle < 0)
    {
        EVM_DBG_PRINT(("ERROR: device not opened"));
        return NFX_ERR;
    }
    
    /* Get a copy of the on disk config data
     * and make sure we still own it
     */
    if (evm_readLunLabel(id, sAttr, &odCfg) == NFX_ERR)
    {
        EVM_DBG_PRINT(("ERROR: cannot read on disk cfg"));
        return NFX_ERR;
    }

    /* Set the creator id if not yet set.
     */
    if (odCfg.creator == 0)
    {
        odCfg.creator = gethostid();
    }
    
    /* Overwrite the appropriate od cfg that was just read
     */
    bzero((void *) &odCfg, sizeof(evm_strgCfg_t));
    EVM_OD_CFG_INIT(odCfg);
    evm_VolConfig2strgCfg(cfg, &odCfg);
    
    if (evm_writeLunLabel(id, sAttr, &odCfg) == NFX_ERR)
    {
        EVM_DBG_PRINT(("ERROR: cannot write od cfg"));
        return NFX_ERR;
    }
    return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : cmp_vol_cf
 *
 * Description  : compare the logical volume's in memory DB rec
 *                with one from the disk
 *
 * Created by   : Lan Chin
 *
 * Date Created : 29-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static int
cmp_vol_cfg(scsi_id_t           id,     // lun containin the on disk config
            evm_StrgAttr_t *    sAttr,  // info to do disk op
            evm_VolConfig_t *   cfg,    // in memory DB
            evm_VolConfig_t *   odCfg,  // on disk version
            char *              err)
{
    EVM_DBG_PRINT(("cmp_vol_cfg(): LUN 0x%llx\n", id));

    if (odCfg->maxSize != cfg->maxSize)
    {
        sprintf(err, "mismatched maxSize (od %lld vs %lld)",
                odCfg->maxSize, cfg->maxSize);
        goto cmp_vol_cfg_err;
    }
    if (odCfg->softQuota != cfg->softQuota)
    {
        sprintf(err, "mismatched softQuota (od %d vs %d)",
                odCfg->softQuota, cfg->softQuota);
        goto cmp_vol_cfg_err;
    }
    if (odCfg->minAutoGrow != cfg->minAutoGrow)
    {
        sprintf(err, "mismatched minAutoGrow (od %lld vs %lld)",
                odCfg->minAutoGrow, cfg->minAutoGrow);
        goto cmp_vol_cfg_err;
    }
    if (odCfg->id.whole_thing != cfg->id.whole_thing)
    {
        sprintf(err, "mismatched vol id (od 0x%llx vs %llx)",
                odCfg->id.whole_thing, cfg->id.whole_thing);
        goto cmp_vol_cfg_err;
    }
    if (odCfg->currSize != cfg->currSize)
    {
        sprintf(err, "mismatched currSize (od %lld vs %lld)",
                odCfg->currSize, cfg->currSize);
        goto cmp_vol_cfg_err;
    }
    if (memcmp((void *) &odCfg->attrs, (void *) &cfg->attrs,
                sizeof(evm_strg_attributes_t)) != 0)
    {
        EVM_DBG_PRINT(("  mismatched attributes\n"));
        // ok, vol cannot grow but can come online, just let someone know
        EVM_DBG_PRINT(("VolId %llx cannot grow: no original storage "
                       "attributes found",
                       odCfg->id.whole_thing));
    }

    return NFX_OK;

cmp_vol_cfg_err:

    EVM_DBG_PRINT(("ERROR: %s\n", err));
    return NFX_ERR;
}

/*------------------------------------------------------------------------------
 * Name         : upd_vol_from_disk
 *
 * Description  : update the in memory vol and vol storage DB from disk
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 30-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int
upd_vol_from_disk(scsi_id_t         id,
                  evm_StrgAttr_t *  sAttr,
                  evm_strgCfg_t *   odCfg,
                  boolean           lunRescan,
                  char *            err)
{
    int             rc = NFX_OK;
    int             a;
    evm_volId_t     volId;
    evm_VolConfig_t vCfg;
    evm_volStrg_t * vscfg = NULL;
    evm_volStrg_t * newVscfg = NULL;

    EVM_FCT_PRINT(("upd_vol_from_disk(): scsi id 0x%llx", id));

    /* create and/or update logical volume config
     * for the first release, no LUN subdivisions are allowed so nLvols
     * should always be one, ie the entire LUN is assigned to a vol
     */
    if (odCfg->id.whole_thing != 0)
    {
        evm_VolConfig_t volConfig;

        evm_strgCfg2VolConfig(odCfg, &volConfig);

        volId = odCfg->id;
        if (volId.whole_thing == 0)
        {
            sprintf(err, "inconsistent on disk data: no volId");
            EVM_ERR_PRINT(("%s", err));
            rc = NFX_ERR;
            goto upd_vol_from_disk_exit;
        }
        EVM_DBG_PRINT(("Vol config found: volId 0x%llx\n", volId.whole_thing));

        /* If already in DB, verify same logical vol config data !!
         * also get the vol's physical storage DB rec
         */
        if (get_lv_cfg_byid(volId, &vCfg, err) == NFX_OK)
        {
            if (lunRescan)
            {
                volConfig.onNIM = vCfg.onNIM;
                volConfig.onLine = vCfg.onLine;
                rc = upd_lv_cfg(&volConfig, FALSE, err);
            }
            else
            {
                rc = cmp_vol_cfg(id, sAttr, &vCfg, &volConfig, err);
                if (rc != NFX_OK)
                {
                    goto upd_vol_from_disk_exit;
                }
            }
            
            if (get_lv_strg(volId, &vscfg, err) == NFX_ERR)
            {
                rc = NFX_ERR;
                goto upd_vol_from_disk_exit;
            }            
            NFX_ASSERT(vscfg != NULL);
        }
        else
        {
            /* Volume doesn't exist, create new logical volume
             * init and create both vol config and vol physical storage DB recs
             * create_lvol() will also add an empty physical storage DB
             */
            if (create_lvol(&volConfig, TRUE, err) == NFX_ERR)
            {
                rc = NFX_ERR;
                goto upd_vol_from_disk_exit;
            }
            
            /* Read back what we just created
             */
            if (get_lv_cfg_byid(volId, &vCfg, err) == NFX_ERR)
            {
                rc = NFX_ERR;
                goto upd_vol_from_disk_exit;
            }
            if (get_lv_strg(volId, &vscfg, err) == NFX_ERR)
            {
                rc = NFX_ERR;
                goto upd_vol_from_disk_exit;
            }
            NFX_ASSERT(vscfg != NULL);
        }

        /* Update the logical vol's physical storage DB rec
         * add this LUN to the vol's physical storage record
         *   a is which allocation unit
         *   s is which stripe number
         *   m is which mirror number
         */
        a = odCfg->allocUnit;

        if (vscfg->num == 0)
        {
            EVM_ALLOC_ZERO(newVscfg, evm_volStrg_t,
                           (sizeof(evm_volStrg_t) +
                            (a * sizeof(evm_strgConfig_t))));
            newVscfg->num = a + 1;
        }
        else if (a >= vscfg->num)
        {
            EVM_ALLOC_ZERO(newVscfg, evm_volStrg_t,
                           (sizeof(evm_volStrg_t) +
                            (a * sizeof(evm_strgConfig_t))));
            memcpy((void *) newVscfg, (const void *) vscfg,
                   VOLSTRG_SIZE(vscfg));
            newVscfg->num = a + 1;
        }
        else if (a < vscfg->num)
        {
            newVscfg = vscfg;
            vscfg = NULL;
        }
        
        if ((newVscfg->au[a].strgId != 0) && (newVscfg->au[a].strgId != id))
        {
            // oops, already occupied !!
            sprintf(err, "LV %llx already has LUN %llx for allocation %d",
                    volId.whole_thing, newVscfg->au[a].strgId, a);
            EVM_ERR_PRINT(("%s", err));
            rc = NFX_ERR;
            goto upd_vol_from_disk_exit;
        }
        newVscfg->au[a].totalSize = sAttr->totalBlkCnt;
        newVscfg->au[a].auSize = sAttr->blkCnt;
        newVscfg->au[a].strgId = id;
        newVscfg->au[a].openHandle = sAttr->openHandle;
        newVscfg->au[a].startBlk = 0;   // 0 since no LUN subdivisions
        if (upd_lv_strg(volId, newVscfg, err) == NFX_ERR)
        {
            rc = NFX_ERR;
            goto upd_vol_from_disk_exit;
        }
        EVM_DBG_PRINT(("new AU[%d] -> totalSize %d auSize %lld startBlk %d "
                       "openHandle %x\n",
                       a,
                       newVscfg->au[a].totalSize,
                       newVscfg->au[a].auSize,
                       newVscfg->au[a].startBlk,
                       newVscfg->au[a].openHandle));

        if (!vCfg.onLine)
        {
            evm_volStrg_t * lvstrg = NULL;
            
            /* Check if the volume is complete.
             * If yes, notify the world about it.
             */
            rc = get_lv_strg(vCfg.id, &lvstrg, err);
            NFX_ASSERT((rc == NFX_OK) && (lvstrg != NULL));
            if (check_lvol(&vCfg, lvstrg) == NFX_OK)
            {
                evm_sendEvt(EVENT_VOL_CREATE, vCfg.id, EVENT_STATE_DOWN);
            }
            EVM_FREE(lvstrg);
        }
        
        /* Check if this logical vol is ready for primetime
         */
        if ((check_lvol(&vCfg, newVscfg) == NFX_OK) &&
            vCfg.onLine && vCfg.onNIM)
        {
            /*
             * notify evm that this logical vol is ready to be served
             */
            if (upd_evm_srvr(NULL, &vCfg, err) == NFX_ERR)
            {
                EVM_ERR_PRINT(("upd_vol_from_disk(): ERROR putting "
                               "volId %llx ONLINE on NIM %d\n",
                               vCfg.id.whole_thing, vCfg.onNIM));
                rc = NFX_ERR;
                goto upd_vol_from_disk_exit;
            }
            EVM_DBG_PRINT(("upd_vol_from_disk(): volId %llx ONLINE on NIM %d\n",
                           vCfg.id.whole_thing, vCfg.onNIM));
        }
        EVM_FREE(vscfg);
        EVM_FREE(newVscfg);
    }
    if (vCfg.id.whole_thing)
    {
        EVM_DBG_PRINT(("LUN %llx: assigned to VolId %llx",
                       id, vCfg.id.whole_thing));
    }

upd_vol_from_disk_exit:
    
    EVM_FREE(vscfg);
    EVM_FREE(newVscfg);
    return(rc);
}

/*-----------------------------------------------------------------
 * Name: upd_od_lun_cfg
 *
 * Description: Update a LUN's on disk config data
 *
 * Called when the LUN's configured volume is destroyed
 * upd_od_vol_cfg() is called when a LUN is assigned to a logical volume
 *
 * Created by: Lan Chin
 *
 * Date Created: 1 Nov 2001
 *
 *-----------------------------------------------------------------
 */
int upd_od_lun_cfg(scsi_id_t      id,       // LUN triggering the update
                   evm_StrgAttr_t *sAttr,   // contains info to access LUN
                   int            delSpool, // delete storage pool
                   int            delVol)   // delete vol config
{
    evm_strgCfg_t   odCfg;
    char            err[4096];

    EVM_DBG_PRINT(("upd_od_lun_cfg(): LUN %llx delSpool %d delVol %d\n",
                   id, delSpool, delVol));

    if (sAttr->openHandle < 0)
    {
        sprintf(err, "device not opened");
        goto upd_od_lun_cfg_err;
    }
    /*
     * get a copy of the on disk config data
     * and make sure we still own it
     */
    if (evm_readLunLabel(id, sAttr, &odCfg) == NFX_ERR)
    {
        sprintf(err, "cannot read from device");
        goto upd_od_lun_cfg_err;
    }
    if (odCfg.creator != gethostid())
    {
        sprintf(err, "not owner of LUN (od owner =%d)", odCfg.creator);
        goto upd_od_lun_cfg_err;
    }
    /*
     * overwrite the appropriate od cfg that was just read
     */
    EVM_OD_CFG_INIT(odCfg);
    /*
     * indicate LUN no longer part of this vol or storage pool
     * zero out all the appropriate config areas
     */
    if (delVol)
    {
        /* Reset the volume label.
         */
        //bzero((void *) &(odCfg.id), sizeof(evm_volId_t));
        bzero((void *) &(odCfg.id),
              (sizeof(odCfg) - sizeof(odCfg.magic) -
               sizeof(odCfg.version) - sizeof(odCfg.cluster)));
    }
    if (evm_writeLunLabel(id, sAttr, &odCfg) == NFX_ERR)
    {
        sprintf(err, "cannot write to device");
        goto upd_od_lun_cfg_err;
    }

    return NFX_OK;

upd_od_lun_cfg_err:

    EVM_ERR_PRINT(("Cannot update on disk data for LUN %llx:%s", id, err));

    return NFX_ERR;
}

/*------------------------------------------------------------------------------
 * Name         : upd_od_vol_cfg
 *
 * Description  : update the vol config for all LUNs allocated to this vol
 *                called by upd_lv_cfg() to update the in memory DB's vol config
 *
 * All the LUN pieces in the configured volume are updated with this
 * config update.
 *
 * A unique transaction id (timestamp) is included for this update
 * transaction.  A partial update can occur when an error is encountered.
 * Upon reboot or reconfigure of this configured volume, the transaction id
 * is used to resolve partial updates.  Configs with a later transaction id is
 * assumed to contain the most recent correct data.
 *
 * Created by   : Lan Chin
 *
 * Date Created : 24-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int
upd_od_vol_cfg(evm_VolConfig_t *    volConfig,     // volume config
               evm_volStrg_t   *    volStrg)       // volume storage config
    
{
    int                 rc = NFX_OK;
    int                 i;
    uint64              lblk;
    uint64              txn;
    int                 nErrs=0;
    char                err[4096];
    evm_strgCfg_t       odCfg;
    scsi_id_t           id;
    evm_StrgAttr_t *    attr = NULL;

    EVM_DBG_PRINT(("upd_od_vol_cfg(): VOL id %llx name %s LUNs assigned %d\n",
                  volConfig->id.whole_thing, volConfig->name, volStrg->num));

    txn = get_od_txn_id();   // get the transaction id for this op
    lblk = 0;                // init logical vol start
    /*
     * go through the list of LUNs that are allocated to
     * this logical volume and update their private meta data
     * to reflect this change in the volume config
     */
    for (i=0; i<volStrg->num; i++)
    {
        /*
         * the storage/scsi ids are saved in the allocation
         * "array" based on stripe and mirror indices
         */
        id = volStrg->au[i].strgId;
        /*
         * get the in memory DB evm_StrgAttr_t rec
         * scsi id is the DB key
         */
        rc = get_pstrg_cfg(id, &attr, err);
        if ((rc != NFX_OK) || (attr == NULL))
        {
            EVM_DBG_PRINT(("upd_od_vol_cfg(): ERR for LUN %llx\n", id));
            nErrs++;
            continue;   // move on to the next
        }
        if (attr->openHandle < 0)
        {
            EVM_DBG_PRINT(("upd_od_vol_cfg(): "
                           "wrong handlefor LUN %llx\n", id));
            nErrs++;
            continue;   // move on to the next
        }
        /* get a copy of the on disk config data
         * make sure we still own it
         */
        if (evm_readLunLabel(id, attr, &odCfg) == NFX_ERR)
        {
            EVM_DBG_PRINT(("upd_od_vol_cfg(): "
                           "cannot read od config for LUN %llx\n", id));
            nErrs++;
            continue;   // move on to the next
        }
        if (odCfg.creator != gethostid())
        {
            //nErrs++;
            //continue;   // move on to the next
        }
        /*
         * overwrite the appropriate od cfg that was just read
         * nLvols always one since we have no LUN subdivisions
         */
        EVM_OD_CFG_INIT(odCfg);
        //odCfg.txnId = txn;
        evm_VolConfig2strgCfg(volConfig, &odCfg);
        /*
         * do not write pool ID back on disk
         * but assemble the LV allocations in the on disk config
         */
        //for (j=0; j < odCfg.nLvols; j++)
        if (odCfg.id.whole_thing != 0)
        {
            odCfg.allocUnit = i;
            odCfg.lblk = lblk;
            odCfg.stBlk = attr->allocs[0].stBlk;
            odCfg.nBlks = attr->allocs[0].nBlks;
            EVM_DBG_PRINT(("  UPDATE [%d] -> VOL config for scsi id %llx:\n",
                           i, id));
            EVM_DBG_PRINT(("    lblk %lld stBlk %d nBlks %d allocUnit %d\n",
                           odCfg.lblk, odCfg.stBlk, odCfg.nBlks,
                           odCfg.allocUnit));
        }
        /*
         * write it back to disk
         */
        if (evm_writeLunLabel(id, attr, &odCfg) == NFX_ERR)
        {
            EVM_DBG_PRINT(("upd_od_vol_cfg(): cannot write "
                           "od config for LUN %llx\n", id));
            nErrs++;
            continue;   // move on to the next
        }
        lblk += volStrg->au[i].auSize;
    } /* for i */
    if (nErrs)
    {
        EVM_ERR_PRINT(("Cannot update disk cfg for VolId %llx",
                       volConfig->id.whole_thing));
        rc = NFX_ERR;
    }
    else
    {
        EVM_DBG_PRINT(("Updated disk cfg for VolId %llx",
                       volConfig->id.whole_thing));
        rc = NFX_OK;
    }

    EVM_FREE(attr);
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_initLun
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 20-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

void 
evm_initLun(sdm_devInfoRspMsg_t * lunInfo,
            evm_StrgAttr_t *      newLun)
    
{
    int n, blkCnt;
    
    bzero((void *) newLun, sizeof(evm_StrgAttr_t));
    newLun->id = lunInfo->id;
    newLun->wwn = lunInfo->wwn;
    // newLun->lun = SCSI_ID_GET_LUN(lunInfo->id);
    newLun->lun = 0; // We should remove LUN from the struct.
                     // It is set to zero because for compilation
                     // reason right now. TBD!!!!!
    EVM_DBG_PRINT(("LUN 0x%llx is in state %d", lunInfo->id, lunInfo->state));
    if (lunInfo->state == SDM_DEV_STATE_OPENED)
    {
        EVM_DBG_PRINT(("LUN 0x%llx already opened (handle = 0x%x)",
                       lunInfo->id, lunInfo->openHandle));
        newLun->openHandle = lunInfo->openHandle;
    }
    else
    {
        newLun->openHandle = -1;
    }
    if (strlen(lunInfo->st.name))
    {
        strncpy(newLun->name, lunInfo->st.name, VOL_MAX_NAME);
    }
    else
    {
        sprintf(newLun->name, "disk_%llx", lunInfo->id);
    }
    newLun->blkSize = lunInfo->blockSize;
    newLun->totalBlkCnt = lunInfo->blockCnt;
    blkCnt = lunInfo->blockCnt - N_BLKS_CFG;
    // make sure that the space for filesystem is %64kbytes
    newLun->blkCnt = (( blkCnt % 128 == 0)? blkCnt : 128 * (blkCnt / 128));
    if ((n = strlen(lunInfo->st.vendor)) && (n <= EVM_VENDOR_NAME_LEN))
    {
        strncpy(newLun->vendor, lunInfo->st.vendor, EVM_VENDOR_NAME_LEN);
    }
    else
    {
        memcpy(newLun->vendor, lunInfo->st.vendor, SDM_VENDOR_NAME_LEN);
        newLun->vendor[SDM_VENDOR_NAME_LEN] = 0;
    }
    if ((n=strlen(lunInfo->st.model)) && (n <= EVM_MODEL_NAME_LEN))
    {
        strncpy(newLun->model, lunInfo->st.model, EVM_MODEL_NAME_LEN);
    }
    newLun->raidLevel = lunInfo->raidLevel;
    newLun->onLine = FALSE;
    newLun->nallocs = 0;

    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_addLun
 *
 * Description  : A new LUN is seen by this node.
 *                . open LUN.
 *                . read label.
 *                . if no Agile label is on the LUN
 *                  . notify SPM about this 'foreign' LUN
 *                  . close LUN
 *                . else if no volume on LUN
 *                  . notify SPM about this 'free' LUN
 *                  . close LUN
 *                . else LUN is part of a volume
 *                  . notify SPM about this 'used' LUN
 *                  . add LUN in local LUN db
 *                  . start building volume (store in volume db)
 *                  . if volume is complete
 *                    . check the owner
 *                    . send volume event to PCC
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 20-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_addLun(evm_StrgAttr_t * newLun,
           boolean          tellSPM,
           char *           err)
    
{
    int32           rc = NFX_OK;
    evm_strgCfg_t   odCfg;
    evm_lunState_t  lunState;
    boolean         closeLun = FALSE;

    EVM_DBG_PRINT(("EVM: evm_addLun(0x%llx).", newLun->id));

    /* Open the LUN if it is not already opened.
     */
    if (newLun->openHandle == -1)
    {
        rc = evm_openLun(newLun->id, SSC_SLOT, &(newLun->openHandle));
        if ((rc != NFX_OK) || (newLun->openHandle == -1))
        {
            sprintf(err, "cannot open LUN 0x%llx", newLun->id);
            EVM_ERR_PRINT(("EVM: %s", err));
            goto evm_addLun_exit;
        }
        closeLun = TRUE;
    }
   
    /* Read the label on LUN.
     */
    rc = evm_readLunLabel(newLun->id, newLun, &odCfg);
    if (rc != NFX_OK)
    {
        sprintf(err, "cannot read label on LUN 0x%llx", newLun->id);
        EVM_ERR_PRINT(("EVM: %s", err));
        goto evm_addLun_exit;
    }

    if ((odCfg.magic == NFX_STRG_MAGIC) && (odCfg.version == EVM_CFG_VERSION))
    {
        /* This LUN is an Agile LUN.
         */
        if (odCfg.id.whole_thing == 0)
        {
            if (!strcmp(odCfg.cluster, myCluster))
            {
                /* This LUN is a free LUN.
                 * Do nothing further with this LUN. Close it.
                 */
                EVM_DBG_PRINT(("EVM: LUN 0x%llx is a free LUN.", newLun->id));
                lunState = EVM_FREE_LUN;
                evm_closeLun(newLun->openHandle);
                newLun->openHandle = -1;
            }
            else
            {
                /* This LUN is a free LUN but out of this cluster.
                 * Do nothing further with this LUN. Close it.
                 */
                EVM_DBG_PRINT(("EVM: LUN 0x%llx is a free out-of-cluster LUN.",
                               newLun->id));
                lunState = EVM_OC_FREE_LUN;
                evm_closeLun(newLun->openHandle);
                newLun->openHandle = -1;
            }                
        }
        else
        {
            if (!strcmp(odCfg.cluster, myCluster))
            {
                /* This LUN is part of a volume.
                 * Add it to our local LUN db.
                 */
                EVM_DBG_PRINT(("EVM: LUN 0x%llx is part of volume 0x%lld.",
                               newLun->id, odCfg.id.whole_thing));
                lunState = EVM_USED_LUN;
                newLun->nallocs = 1;
                newLun->allocs[0].lvid = odCfg.id;
                newLun->allocs[0].stBlk = odCfg.stBlk;
                newLun->allocs[0].nBlks = odCfg.nBlks;
                rc = add_pstrg_cfg(newLun->id, newLun, err);
                if (rc != NFX_OK)
                {
                    EVM_ERR_PRINT(("EVM: %s", err));
                    goto evm_addLun_exit;
                }
                rc = upd_vol_from_disk(newLun->id, newLun, &odCfg, FALSE, err);
                if (rc != NFX_OK)
                {
                    EVM_ERR_PRINT(("EVM: %s", err));
                    goto evm_addLun_exit;
                }
                
                /* If the LUN is part of a corevolume, close the LUN.
                 */
                if (odCfg.type == EVM_TYPE_COREVOL)
                    evm_closeLun(newLun->openHandle);

            }
            else
            {
                /* This LUN is a used LUN but out of this cluster.
                 * Do nothing further with this LUN. Close it.
                 */
                EVM_DBG_PRINT(("EVM: LUN 0x%llx is a used out-of-cluster LUN.",
                               newLun->id));
                lunState = EVM_OC_USED_LUN;
                evm_closeLun(newLun->openHandle);
                newLun->openHandle = -1;
            }
        }
    }
    else
    {
        /* This LUN is not an Agile LUN.
         * Do nothing further with this LUN. Close it.
         */
        EVM_DBG_PRINT(("EVM: LUN 0x%llx is a foreign LUN.", newLun->id));
        lunState = EVM_FOREIGN_LUN;
        strcpy(odCfg.cluster, "");
        evm_closeLun(newLun->openHandle);
        newLun->openHandle = -1;
    }

    /* Tell SPM about this LUN.
     */
    if (tellSPM)
    {
        ssize_t                 ncmMsgSize;
        ncmMsg_t *              ncmMsg = NULL;
        spm_MsgAddLunReq_t *    spmReq = NULL;

        /* Build NCM message.
         */
        ncmMsgSize = sizeof(ncmMsg_t) + sizeof(spm_MsgAddLunReq_t) - 1;
        EVM_ALLOC_ZERO(ncmMsg, ncmMsg_t, ncmMsgSize);
        NFX_ASSERT(ncmMsg != NULL);
        ncmMsg->msgType = NCM_SEND;
        strcpy(ncmMsg->appName, "spm");
        strcpy(ncmMsg->nodeName, "PCC");
        ncmMsg->rqResp = 0;    // no response expected
        spmReq = (spm_MsgAddLunReq_t *) ncmMsg->appMsg;
                
        /* Build SPM message.
         */
        spmReq->hdr.msgType = SPM_ADD_LUN_REQ;
        strcpy(spmReq->hdr.node, myNode);
        spmReq->lunAttr.scsiId = newLun->id;
        spmReq->lunAttr.wwn = newLun->wwn;
        spmReq->lunAttr.lun = newLun->lun;
        strcpy(spmReq->lunAttr.name, newLun->name);
        strcpy(spmReq->lunAttr.vendor, newLun->vendor);
        strcpy(spmReq->lunAttr.model, newLun->model);
        spmReq->lunAttr.raidLevel = newLun->raidLevel;
        spmReq->lunAttr.size = (uint64) newLun->blkCnt;
        spmReq->lunAttr.state = lunState;
        strcpy(spmReq->lunAttr.cluster, odCfg.cluster);
        spmReq->lunAttr.devType = SCSI_DEVTYPE_DISK;
        
        sendAgileMsg(-1, ncmMsg, ncmMsgSize,
                     "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                     NULL, NULL, 0);
        EVM_FREE(ncmMsg);
    }
    
evm_addLun_exit:

    if ((rc != NFX_OK) && (closeLun == TRUE))
    {
        evm_closeLun(newLun->openHandle);
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : init_dbs
 *
 * Description  : Open/Initialize the configuration databases
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 02-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

int
init_dbs(void)

{
    int32                   rc;
    char                    err[EVM_MAX_ERR_STR];
    sdm_devListReqMsg_t     list_req;
    char                    buf[EVM_MAX_BUF_SIZE];
    sdm_devListRspMsg_t *   list_msg = (sdm_devListRspMsg_t *) buf;
    int                     stat;
    int                     i;
    
    /* Initialize the databases.
     */
    rc = init_pstrgdb();
	if (rc != NFX_OK)
    {
        goto init_dbs_exit;
    }
    rc = init_lvdb();
    if (rc != NFX_OK)
    {
        goto init_dbs_exit;
    }

    /* Get the list of LUNs this node sees.
     */
    memset(&list_req, 0, sizeof(list_req));
    list_req.hdr.mid = SDM_MID_DEV_LIST_REQ;
	list_req.port = SDM_DEVLIST_PORTALL;
    list_req.hdr.chandle = getpid();
    
    stat = sendAgileMsg(-1, &list_req, sizeof(list_req),
                        "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &sdm_to, buf, sizeof(buf));
    if (stat > 0)
    {
        /* Go through the list of LUNs.
         */
        for (i = 0; i < list_msg->numEntries; i++)
        {
            evm_StrgAttr_t  lunInfo;
            
            /* Get detailed info for this LUN.
             */
            rc = evm_getLunInfo(list_msg->idList[i], &lunInfo, err);
            if (rc != NFX_OK)
            {
                // should never fail since SDM just tell me about this LUN
                EVM_ERR_PRINT(("EVM: %s", err));
                goto init_dbs_exit;
            }

            /* Examine the LUN and keep it if part of a volume.
             */
            rc = evm_addLun(&lunInfo, TRUE, err);
            if (rc != NFX_OK)
            {
                // could happen in the case a LUN is dying
                // SDM sees it but we can't access it
                EVM_ERR_PRINT(("EVM: %s", err));
                EVM_ALR_PRINT(("LUN 0x%llx cannot be accessed. "
                               "It will be ignored. (%s)",
                               list_msg->idList[i], err));
                //goto init_dbs_exit;
                rc = NFX_OK;
            }                    
        } // for
    }
    else
    {
        rc = NFX_ERR;
        EVM_ERR_PRINT(("EVM: Unable to get LUNs' list from SDM."));
        goto init_dbs_exit;
    }

    /* Get list of volumes that are already on-line on the FP-NIMs.
     * TODO later: multicast the request to all FP-NIMs.
     */
    {
        evm_MsgReqVolList_t     req;
        evm_MsgRspVolList_t *   rsp = (evm_MsgRspVolList_t *) buf;
        uint32                  nim = 1;

        bzero((void *) buf, sizeof(buf));
        bzero((void *) &req, sizeof(evm_MsgReqVolList_t));
        req.hdr.msgType = EVM_LVACT_VOLLIST_REQ;
        req.hdr.len = sizeof(evm_MsgReqVolList_t);
        
        stat = sendAgileMsg(-1, &req, sizeof(evm_MsgReqVolList_t),
                            "evm", NFX_NFP_FP_ANY, nim, MGMTPLANE,
                            &nimto, buf, sizeof(buf));
        if (stat > 0)
        {
            /* Go through the list of on-line volumes.
             */
            NFX_ASSERT(rsp->hdr.msgType == EVM_LVACT_VOLLIST_RSP);
#ifdef ISABELLE // debug only!!!
            printf("***rsp->nbVol = %d\n", rsp->nbVol);
#endif // ISABELLE
            for (i = 0; i < rsp->nbVol; i++)
            {
#ifdef ISABELLE // debug only!!!
                printf("***rsp->ids[%d] = %x\n", i, rsp->ids[i].whole_thing);
#endif // ISABELLE
                evm_onlineVol(rsp->ids[i], nim, err);
            } // for
        }
        else
        {
            /* FP-NIM is not responding. It might be down.
             * Do nothing no volumes are on-line.
             */
        }
    }
    
init_dbs_exit:

	return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_clearLun
 *
 * Description  : Reset the volume label of all LUNs seen by this node.
 *                Very dangerous command since all volumes are lost.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 14-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_clearLun(void)

{
    char                    err[EVM_MAX_ERR_STR];
    sdm_devListReqMsg_t     list_req;
    char                    listbuf[EVM_MAX_BUF_SIZE];
    sdm_devListRspMsg_t *   list_msg = (sdm_devListRspMsg_t *) listbuf;
    int                     stat;
    int                     i;

    /* Close all open LUNs first.
     */
    evm_closeAllLun();
    
    /* Get the list of LUNs this node sees.
     */
    memset(&list_req, 0, sizeof(list_req));
    list_req.hdr.mid = SDM_MID_DEV_LIST_REQ;
	list_req.port = SDM_DEVLIST_PORTALL;
    list_req.hdr.chandle = getpid();
    
    stat = sendAgileMsg(-1, &list_req, sizeof(list_req),
                        "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &sdm_to,
                        listbuf, sizeof(listbuf));
    if (stat > 0)
    {
        /* Go through the list of LUNs.
         */
        for (i = 0; i < list_msg->numEntries; i++)
        {
            evm_labelLun(list_msg->idList[i], "", err);
        } // for
    }
    else
    {
        EVM_ERR_PRINT(("EVM: Unable to get LUNs' list from SDM."));
    }
    
	return;
}

/*------------------------------------------------------------------------------
 * Name         : close_dbs
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int
close_dbs(void)

{
    int32   rc = NFX_OK;
    
    /* Close the databases.
     */
    rc = close_pstrgdb();
	if (rc != NFX_OK)
    {
        goto close_dbs_exit;
    }
    rc = close_lvdb();
    if (rc != NFX_OK)
    {
        goto close_dbs_exit;
    }

close_dbs_exit:
	return(rc);
}

/*----------------------------------------------------------------------------*/
