/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 * RCS          : $Id: evm_pstrgdb.c,v 1.44 2002/09/27 21:14:56 isabellec Exp $
 *
 * Name         : evm_pstrgdb.c
 *
 * Description  : This file contains the methods used to manipulate
 *                EVM's Volume Configuration database. There is a single
 *                instance of these on each SSC.
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 02-JUL-2001
 *
 *------------------------------------------------------------------------------
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <db.h>
#include <string.h>

#include "exports/elog-api.h"
#include "exports/eee-desc.h"
#include "exports/scsi-api.h"
#include "exports/sdm-api.h"
#include "exports/anpssc-api.h"

#include "evm-api.h"
#include "evm_loc_defs.h"
#include "evm.h"

/*
	strgdb uses the scsi_id as key, record is evm_StrgAttr_t
	Like the LV DB, there is an name string index too.
 */
DB * strgdb = NULL;		// Physical storage database
DB * strgidx = NULL;    // Name index

static struct timeval sdm_to = { 10, 0 };	// Wait 10 sec for SDM to respond

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : add_pstrg_cfg
 *
 * Description  : Add this physical storage volume to the DB
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 27-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

int
add_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t *  cfg,
              char *            err)
    
{
	DBT idkey;
	DBT namkey;
	DBT rec;

    EVM_DBG_PRINT(("EVM: add_pstrg_cfg(0x%llx / %s)...", id, cfg->name));

    idkey.data = &id;
	idkey.size = sizeof(id);
	rec.data = cfg;
	rec.size = STRG_ATTR_SIZE (cfg);
	if (strgdb->put(strgdb, &idkey, &rec, 0) == -1)
    {
		sprintf(err, "Error putting strgdb rec: %s", strerror(errno));
		EVM_ERR_PRINT(("EVM: %s", err));
		return(NFX_ERR);
	}

	namkey.data = cfg->name;
	namkey.size = VOL_MAX_NAME;    
	if (strgidx->put(strgidx, &namkey, &idkey, 0) == -1)
    {
		sprintf("Error putting strg name: %s", strerror(errno));
		EVM_ERR_PRINT(("EVM: %s", err));
		return(NFX_ERR);
	}
    
    EVM_DBG_PRINT(("EVM: add_pstrg_cfg(0x%llx / %s) ok.", id, cfg->name));
	return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : del_pstrg_cfg
 *
 * Description  : Delete this physical storage volume to the DB
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int
del_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t *  cfg,
              char *            err)
    
{
	DBT idkey;
	DBT namkey;

    EVM_DBG_PRINT(("del_pstrg_cfg(0x%llx / %s)...", id, cfg->name));

    idkey.data = &id;
	idkey.size = sizeof(scsi_id_t);
	if (strgdb->del(strgdb, &idkey, 0) == -1)
    {
        sprintf(err, "Error deleting strgdb record (%s).", strerror(errno));
		EVM_ERR_PRINT(("%s", err));
		return(NFX_ERR);
	}

    namkey.data = cfg->name;
	namkey.size = VOL_MAX_NAME;
	if (strgidx->del(strgidx, &namkey, 0) == -1)
    {
        sprintf(err, "Error putting strg name (%s).", strerror(errno));
		EVM_ERR_PRINT(("%s", err));
		return(NFX_ERR);
	}

    EVM_DBG_PRINT(("del_pstrg_cfg(0x%llx / %s) ok.", id, cfg->name));
	return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : upd_pstrgdb
 *
 * Description  : When we get a message from SDM, we update our internal
 *                database to track this storage device
 *
 * Created by   : Greg Seibert
 * Modified by  : Lan Chin
 *                Isabelle Cnudde
 *
 *  Support grouping LUNs into storage pool by attributes:
 *  (attributes available from SDM)
 *      1. lun size
 *      2. raid level
 *      3. vendor name
 *      4. model name
 *
 * Date Created : 26 Jul 2001
 *
 *------------------------------------------------------------------------------
 */

void
upd_pstrgdb(void *  msg,
            boolean add)
    
{
    int                 rc = NFX_OK;
    evm_StrgAttr_t *    strgattr = NULL;
    char                errmsg[EVM_MAX_ERR_STR];

    NFX_ASSERT(add == FALSE);

    {
        scsi_id_t * lunId = (scsi_id_t *) msg;

        rc = get_pstrg_cfg(*lunId, &strgattr, errmsg);
        if (rc == NFX_OK)
        {
            rc = del_pstrg_cfg(*lunId, strgattr, errmsg);
            if (rc == NFX_OK)
            {
                EVM_DBG_PRINT(("LUN %llx successfully removed from db.",
                               *lunId));
            }
            else
            {
                EVM_ERR_PRINT(("Cannot remove LUN %llx from db (%s).",
                               *lunId, errmsg));
            }
        }
        else
        {
            /* LUN is not in the database.
             * Do nothing.
             */
            EVM_DBG_PRINT(("LUN %llx is not in strgdb.", *lunId));
        }
        EVM_FREE(strgattr);
    }
    return;
}

/*-----------------------------------------------------------------
 * Name: init_pstrgdb
 *
 * Description: Open/Initialize the physical storage configuration database
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 2 May 2001
 *
 *-----------------------------------------------------------------
 */
int init_pstrgdb(void)
{
    /*
      Likewise for the search databases for the storage pools and physical
      storage volumes.
      Storage pool's assets will be retrieved from the stored configuration
      while the physical storage will be verified against the SCSI's DCS
      database.
      We'll have to make sure the physical storage specified in the saved
      storage pool and logical volume config jives with what SCSI discovered
      on this boot.
      Otherwise it's, "Houston, we have a problem."
    */
	strgdb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
	strgidx = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
	if (!strgdb || !strgidx)
    {
		EVM_ERR_PRINT(("Error opening local Device DB %s", strerror(errno)));
		exit(1);
	}
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: close_pstrgdb
 *
 * Description: 
 *
 * Created by:	
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */

int
close_pstrgdb(void)

{
    int32 rc = NFX_OK;

    if (strgdb != NULL)
        rc = strgdb->close(strgdb);
    if (strgidx != NULL)
        rc = strgidx->close(strgidx);

    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : get_pstrg_cfg
 *
 * Description  : Retrieve a physical storage volume's configuration
 *                given its scsi_id.
 *
 * Created by   : Greg Seibert
 *                Isabelle Cnudde
 *
 * Date Created : 27-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

int
get_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t ** cfg,
              char *            err)
    
{
	DBT             idkey;
	DBT             rec;
	int             stat;
	//evm_StrgAttr_t  crud;

	idkey.data = &id;
	idkey.size = sizeof(id);
	stat = strgdb->get(strgdb, &idkey, &rec, 0);
	if (stat)
    {
		if (stat < 0)
        {
			sprintf (err, "Error getting strg id: %s", strerror(errno));
		}
		else
        {
			sprintf (err, "scsi_id %lld not found", id);
		}
		return NFX_ERR;
	}
	//memcpy(&crud, rec.data, sizeof (evm_StrgAttr_t));
	//memcpy(&crud, rec.data, rec.size);

    //*cfg = (evm_StrgAttr_t *) malloc(rec.size);
    EVM_ALLOC(*cfg, evm_StrgAttr_t, rec.size);
	memcpy((void *) *cfg, (const void *) rec.data, rec.size);

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: upd_pstrg_cfg
 *
 * Description: Update this physical storage volume in the DB
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 27 June 2001
 *
 *-----------------------------------------------------------------
 */

int
upd_pstrg_cfg(scsi_id_t         id,
              evm_StrgAttr_t *  cfg,
              int32             upd_od,
              char *            err)
{
	int stat;
	DBT nkey;
	DBT idkey;
	DBT rec;
	char oldname[VOL_MAX_NAME];
	evm_StrgAttr_t crud;

#ifdef ISABELLE // debug only!!!
    printf("upd_pstrg_cfg(%llx)\n", id);
#endif // ISABELLE
    
    /* Make sure that record really is there
     */
	idkey.data = &id;
	idkey.size = sizeof(id);
	stat = strgdb->get(strgdb, &idkey, &rec, 0);
	if (stat)
    {
		sprintf (err, "%lld: not found", id);
		return NFX_ERR;
	}
    /* Hold on to this stuff so we can determine if we need to open/close
     * this particular handle.
     */
	memcpy (&crud, rec.data, sizeof (evm_StrgAttr_t));

    /* If the name changes we'll have to update the index too
     */
	if (memcmp (cfg->name, ((evm_StrgAttr_t *)rec.data)->name, VOL_MAX_NAME))
    {
        /* Is the new name unique?
         */
		nkey.data = cfg->name;
		nkey.size = VOL_MAX_NAME;
		stat = strgidx->get (strgidx, &nkey, &rec, 0);
		if (!stat)
        {
			sprintf (err, "%s duplicate name", cfg->name);
			return NFX_ERR;
		}
        /*
          Make a copy on our stack (never know what libdb will do to the stuff
          second time around) of the old name and delete it's index record
        */
		memcpy (oldname, ((evm_StrgAttr_t *)rec.data)->name, VOL_MAX_NAME);
		nkey.data = oldname;
		nkey.size = VOL_MAX_NAME;
		stat = strgidx->del (strgidx, &nkey, 0);
		if (stat)
        {
			return NFX_ERR;
		}
	}

    /* Overwrite with the new data
     */
	rec.data = cfg;
	rec.size = STRG_ATTR_SIZE (cfg);
	stat = strgdb->put (strgdb, &idkey, &rec, 0);
	if (stat)
    {
		return NFX_ERR;
	}
    /* Likewise with the new key
     */
	nkey.data = cfg->name;
	nkey.size = VOL_MAX_NAME;
	stat = strgidx->put (strgidx, &nkey, &idkey, 0);

	if (stat)
    {
		return NFX_ERR;
	}
	else
    {
        /* Update on disk config
         */
        if (upd_od) upd_od_lun_cfg(id, cfg, 0, 0);
        EVM_DBG_PRINT(("LUN %s openHandle =%x\n", cfg->name, cfg->openHandle));
		return NFX_OK;
	}
}

/*-----------------------------------------------------------------
 * Name: get_pstrg_id
 *
 * Description: Given its name, return its scsi_id
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 June 2001
 *
 *-----------------------------------------------------------------
 */
int get_pstrg_id (char *name, scsi_id_t *id, char *err)
{
	int stat;
	DBT nkey;
	DBT rec;

	nkey.data = name;
	nkey.size = VOL_MAX_NAME;
	stat = strgidx->get (strgidx, &nkey, &rec, 0);

	if (stat) {
		sprintf (err, "%s: not found", name);
		return NFX_ERR;
	}
	else {
		memcpy(id, rec.data, sizeof(scsi_id_t));
		return NFX_OK;
	}
}

/*-----------------------------------------------------------------
 * Name: get_pstrg_list
 *
 * Description: Returns a list of the physical storage objects known
 *              to EVM.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 25 July 2001
 *
 *-----------------------------------------------------------------
 */
int get_pstrg_list (evm_MsgStrgList_t *lst, char *err)
{
	int stat;
	DBT key;
	DBT data;
	int num = 0;
	int rtn_code = NFX_OK;
/*
	Sequentially go through the pstrg config db recording the strg ids
	found therein
 */
	stat = strgdb->seq (strgdb, &key, &data, R_FIRST);

	while (!stat && (num < MAX_PSTRG)) {
		memcpy (&(lst->ids[num]), key.data, sizeof(scsi_id_t));
		num++;

		stat = strgdb->seq (strgdb, &key, &data, R_NEXT);
	}
	lst->num = num;
/*
	If there was an error or we reached max before coming to the end,
	it is a problem.
 */
	if (stat < 0) {
		if (err) {
			sprintf (err, "Error reading PStrg database %s", strerror(errno));
			EVM_DBG_PRINT(("%s", err));
		}
		rtn_code = NFX_ERR;
	}
	else if (!stat) {
		if (err) {
			sprintf (err, "Did not get all PStrg ids");
            EVM_DBG_PRINT(("%s", err));
		}
		rtn_code = NFX_ERR;
	}
	return rtn_code;
}

/*------------------------------------------------------------------------------
 * Name         : evm_openLun
 *
 * Description  : Open a LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_openLun(scsi_id_t   id,
            uint32      slot,
            int32 *     handle)
    
{
    int32               rc = NFX_OK;
    int                 stat;
    sdm_openReqMsg_t    req;
    sdm_openRspMsg_t    rsp;

    bzero(&req, sizeof(sdm_openReqMsg_t));
    req.hdr.mid = SDM_MID_DEV_OPEN_REQ;
    req.hdr.chandle = getpid();
    req.id = id;
    req.slot = slot;
	//req.flags = 0;
    stat = sendAgileMsg(-1, &req, sizeof(sdm_openReqMsg_t),
                        "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &sdm_to,
                        &rsp, sizeof(sdm_openRspMsg_t));
    if ((stat <= 0) || (rsp.rCode != SDM_OK))
    {
        EVM_ERR_PRINT(("EVM: Unable to open LUN 0x%llx on slot %d.", id, slot));
        *handle = -1;
        rc = NFX_ERR;
    }
    else
    {
        *handle = rsp.handle;
        rc = NFX_OK;
    }
    
    EVM_DBG_PRINT(("EVM: SDM_MID_DEV_OPEN_REQ (0x%llx / %x).\n", id, *handle));
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_closeLun
 *
 * Description  : Close a LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_closeLun(int32  handle)
    
{
    int32               rc = NFX_OK;
    int                 stat;
    sdm_closeReqMsg_t   req;
    sdm_closeRspMsg_t   rsp;

    EVM_DBG_PRINT(("EVM: SDM_MID_DEV_CLOSE_REQ(%x)", handle));
    
    bzero(&req, sizeof(sdm_closeReqMsg_t));
    req.hdr.mid = SDM_MID_DEV_CLOSE_REQ;
    req.hdr.chandle = getpid();
    req.handle = handle;

    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &sdm_to,
                        &rsp, sizeof(sdm_closeRspMsg_t));

    if ((stat <= 0) || (rsp.rCode == SDM_FAIL))
    {
        EVM_ERR_PRINT(("EVM: Unable to close scsi device (handle = %x).",
                       handle));
        rc = NFX_ERR;
    }

    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_getLunInfo
 *
 * Description  : Get detailed info for this LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_getLunInfo(scsi_id_t        lunId,
               evm_StrgAttr_t * lunInfo,
               char *           err)
    
{
    int32                   rc = NFX_OK;
    sdm_devInfoReqMsg_t     info_req;
    sdm_devInfoRspMsg_t     info_msg;
    int                     stat;

    bzero((void *) lunInfo, sizeof(evm_StrgAttr_t));

    /* Request SDM detailed info about this LUN.
     */
    memset(&info_req, 0, sizeof(info_req));
    info_req.hdr.mid = SDM_MID_DEV_INFO_REQ;
    info_req.hdr.chandle = getpid();
    info_req.id = lunId;
    stat = sendAgileMsg(-1, &info_req, sizeof(info_req),
                        "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &sdm_to, &info_msg, sizeof(info_msg));
    
    /* Shall we update our internal database with this info (s'OK?)
     */
    if ((stat > 0) && (info_msg.rc == SDM_OK))
    {
        evm_initLun(&info_msg, lunInfo);
    }
    else
    {
        rc = info_msg.rc;
        sprintf(err, "Unable to get detailed info for LUN 0x%llx.",
                info_msg.id);
        EVM_ERR_PRINT(("EVM: %s", err));
    }

    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_labelLun
 *
 * Description  : Write the Agile label to a LUN.
 *                If a cluster name is specified, write it in the label.
 *                Otherwise, write our own cluster name.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_labelLun(scsi_id_t  lunId,
             char *     clusterName,
             char *     err)
    
{
    int32           rc = NFX_OK;
    evm_StrgAttr_t  lunInfo;
    evm_strgCfg_t   odCfg;
    evm_mgrArea_t   mgrArea;
    evm_tkoArea_t   tkoArea;
    
    rc = evm_getLunInfo(lunId, &lunInfo, err);
    if (rc != NFX_OK)
    {
        goto evm_labelLun_exit;
    }
  
    if (lunInfo.openHandle != -1)
    {
        rc = NFX_ERR;
        sprintf(err, "Cannot label a LUN (0x%llx) that is opened.", lunId);
        lunInfo.openHandle = -1; // to avoid it to be closed
        goto evm_labelLun_exit;
    }
    
    rc = evm_openLun(lunId, SSC_SLOT, &(lunInfo.openHandle));
    if (rc != NFX_OK)
    {
        sprintf(err, "Cannot open LUN (0x%llx).", lunId);
        goto evm_labelLun_exit;
    }
            
    /* Clean LUN label.
     */
    bzero((void *) &odCfg, sizeof(evm_strgCfg_t));
    EVM_OD_CFG_INIT(odCfg);
    if (strcmp(clusterName, ""))
    {
        strcpy(odCfg.cluster, clusterName);
    }

    rc = evm_writeLunLabel(lunId, &lunInfo, &odCfg);
    if (rc != NFX_OK)
    {
        sprintf(err, "Cannot write LUN label (0x%llx).", lunId);
        goto evm_labelLun_exit;
    }

    /* Clean manager area.
     */
    bzero((void *) &mgrArea, sizeof(evm_mgrArea_t));
    mgrArea.magic = NFX_STRG_MAGIC;
    rc = evm_writeMgrArea(&lunInfo, &mgrArea);
    if  (rc != NFX_OK)
    {
        sprintf(err, "Cannot write LUN mgr area (0x%llx).", lunId);
        goto evm_labelLun_exit;
    }
    
    /* Clean takeover area.
     */
    bzero((void *) &tkoArea, sizeof(evm_tkoArea_t));
    tkoArea.magic = NFX_STRG_MAGIC;
    rc = evm_writeTkoArea(&lunInfo, &tkoArea);
    if  (rc != NFX_OK)
    {
        sprintf(err, "Cannot write LUN tko area (0x%llx).", lunId);
        goto evm_labelLun_exit;
    }
    
evm_labelLun_exit:
    
    if (lunInfo.openHandle != -1)
    {
        evm_closeLun(lunInfo.openHandle);
    }
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_unlabelLun
 *
 * Description  : Remove the Agile label from a LUN.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_unlabelLun(scsi_id_t    lunId,
               char *       err)
    
{
    int32           rc = NFX_OK;
    evm_StrgAttr_t  lunInfo;
    evm_strgCfg_t   odCfg;
    evm_mgrArea_t   mgrArea;
    evm_tkoArea_t   tkoArea;
    
    rc = evm_getLunInfo(lunId, &lunInfo, err);
    if (rc != NFX_OK)
    {
        goto evm_unlabelLun_exit;
    }
  
    rc = evm_openLun(lunId, SSC_SLOT, &(lunInfo.openHandle));
    if (rc != NFX_OK)
    {
        goto evm_unlabelLun_exit;
    }

    /* Zero LUN label.
     */
    bzero((void *) &odCfg, sizeof(evm_strgCfg_t));    
    rc = evm_writeLunLabel(lunId, &lunInfo, &odCfg);
    if (rc != NFX_OK)
    {
        goto evm_unlabelLun_exit;
    }

    /* Zero manager area.
     */
    bzero((void *) &mgrArea, sizeof(evm_mgrArea_t));
    rc = evm_writeMgrArea(&lunInfo, &mgrArea);
    if  (rc != NFX_OK)
    {
        goto evm_unlabelLun_exit;
    }

    /* Zero takeover area.
     */
    bzero((void *) &tkoArea, sizeof(evm_tkoArea_t));
    rc = evm_writeTkoArea(&lunInfo, &tkoArea);
    if  (rc != NFX_OK)
    {
        goto evm_unlabelLun_exit;
    }
    
evm_unlabelLun_exit:
    
    if (lunInfo.openHandle != -1)
    {
        rc = evm_closeLun(lunInfo.openHandle);
    }
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_modLunCluster
 *
 * Description  : Modify the cluster to which a LUN belongs.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-JUL-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_modLunCluster(scsi_id_t lunId,
                  char *    clusterName,
                  char *    err)
    
{
    int32           rc = NFX_OK;
    evm_StrgAttr_t  lunInfo;
    evm_strgCfg_t   odCfg;
    boolean         closeLun = FALSE;
    
    rc = evm_getLunInfo(lunId, &lunInfo, err);
    if (rc != NFX_OK)
    {
        goto evm_modLunCluster_exit;
    }

    if (lunInfo.openHandle == -1)
    {
        rc = evm_openLun(lunId, SSC_SLOT, &(lunInfo.openHandle));
        if (rc != NFX_OK)
        {
            sprintf(err, "Failed to open LUN 0x%llx.", lunId);
            goto evm_modLunCluster_exit;
        }
        closeLun = TRUE;
    }

    /* Read current label.
     */
    rc = evm_readLunLabel(lunId, &lunInfo, &odCfg);
    if (rc != NFX_OK)
    {
        sprintf(err, "Failed to read LUN's label 0x%llx.", lunId);
        goto evm_modLunCluster_exit;
    }

    /* Make sure that the LUN is an agile one.
     */
    if (odCfg.magic != NFX_STRG_MAGIC)
    {
        rc = NFX_ERR;
        sprintf(err, "Cannot change the cluster of a foreign LUN 0x%llx.",
                lunId);
        goto evm_modLunCluster_exit;
    }

    /* Write the new cluster name into the LUN's label.
     */
    if (!strcmp(clusterName, ""))
    {
        strcpy(odCfg.cluster, myCluster);
    }
    else
    {
        strcpy(odCfg.cluster, clusterName);
    }
    rc = evm_writeLunLabel(lunId, &lunInfo, &odCfg);
    if (rc != NFX_OK)
    {
        sprintf(err, "Failed to write LUN's label 0x%llx.", lunId);
        goto evm_modLunCluster_exit;
    }
    
evm_modLunCluster_exit:
    
    if ((lunInfo.openHandle != -1) && (closeLun == TRUE))
    {
        rc = evm_closeLun(lunInfo.openHandle);
    }
    return(rc);
}

/*------------------------------------------------------------------------------
 * Name         : evm_closeAllLun
 *
 * Description  : Close all LUNs.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int32
evm_closeAllLun(void)

{
    sdm_closeAllDiskReqMsg_t    req;

    EVM_DBG_PRINT(("EVM: SDM_MID_CLOSE_ALL_DISK_REQ"));

    req.hdr.mid = SDM_MID_CLOSE_ALL_DISK_REQ;
    req.hdr.chandle = getpid();
    sendAgileMsg(-1, &req, sizeof(sdm_closeAllDiskReqMsg_t),
                 "sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                 NULL, NULL, 0);
    
    return(NFX_OK);
}

/*----------------------------------------------------------------------------*/
