/*
 *-----------------------------------------------------------------
 *
 * Name:	exc-interrupt.c
 *
 * RCS:  $Id: exc-interrupt.c,v 1.13 2002/12/02 18:45:07 rick Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	2/21/02
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <sys/mips.h>
#include <sys/tx3927.h>
#include "except-api.h"
#include "except.h"

void (*intr_func[2 + 0x10])();

uchar8	intr_stack[STKSIZE]	CPU_PRIVATE;
uint32	intr_mode	CPU_PRIVATE;

/*-----------------------------------------------------------------
 * Name:	exc_interrupt(sr, cause, epc)
 *
 * Description:	Interrupt handler, calls registered routine
 *
 * Created by:	Rick Lund
 *
 * Date Created:	2/28/02
 *
 *-----------------------------------------------------------------
 */
void
exc_interrupt(
	uint32		sr,
	uint32		cause,
	address_t	epc
)
{
#if defined(R7000) || defined(SB1250)
	uchar8	pending, mask;
	uint32	ipl;

	pending = (cause & CAUSE_IPMASK) >> CAUSE_IPSHIFT;
	mask = (sr & SR_IMASK) >> 8;

	for (ipl = 7; ipl >= 0; ipl--) {
		if (!pending || !mask)
			break;

		if ((pending & 0x80) && (mask & 0x80)) {
			if (intr_func[ipl] == NULL)
				panic("Unexpected interrupt, sr = 0x%x, cause = 0x%x, epc = %p\n",
															sr, cause, epc);
			intr_func[ipl](sr, cause, epc);
		}

		pending <<= 1;
		mask <<= 1;
	}

#elif defined(TX3927)
	uint32	pending, mask;
	uint32	ipl;

	pending = cause & CAUSE_IPMASK;
	mask = sr & SR_IMASK;

	if ((pending & CAUSE_SW1) && (mask & SR_IM0))
		intr_func[0](sr, cause, epc);
	if ((pending & CAUSE_SW2) && (mask & SR_IM1))
		intr_func[1](sr, cause, epc);
	if (pending >= 0x8000) {
		ipl = pending >> (CAUSE_IPSHIFT + 2);
		ipl -= 0x20;
		if (intr_func[2+ipl] == NULL)
			panic("Unexpected interrupt, sr = 0x%x, cause = 0x%x, epc = %p\n",
															sr, cause, epc);
		intr_func[2+ipl](sr, cause, epc);
	}
#endif
}

/*-----------------------------------------------------------------
 * Name:	intr_registerHandler(ipl, func)
 *
 * Description:	register a interrupt handler function for the given level
 *
 * Created by:	Rick Lund
 *
 * Date Created:	2/28/02
 *
 *-----------------------------------------------------------------
 */
int32
intr_registerHandler(
	uint32	ipl,
	void	(*func)(uint32 sr, uint32 cause, address_t epc)
)
{
#if defined(R7000) || defined(SB1250)

	uint32	sr;

	if ((ipl < 0) || (ipl > 7)) {
		printf("Invalid interrupt level, ipl = %d\n", ipl);
		return -1;
	}

	if (intr_func[ipl] != NULL) {
		printf("handler already registered for ipl = %d (%p)\n",
														ipl, intr_func[ipl]);
		return -1;
	}

	intr_func[ipl] = func;

	sr = _mips_mfc0(C0_SR);
	sr |= (SR_IM0 << ipl);
	_mips_mtc0(C0_SR, sr);

#elif defined(TX3927)

	tx3927_intcntrl_t	* int_cntrl = (tx3927_intcntrl_t *)TX3927_INT_CONTROL_BASE;
	uint32				sr;

	sr = _mips_mfc0(C0_SR);
	if (ipl == 0) {
		sr |= SR_IM0;
		intr_func[0] = func;
	}
	else if (ipl == 1) {
		sr |= SR_IM1;
		intr_func[1] = func;
	}
	else {
		int_cntrl->der = 0x1;
		if (ipl == 0x2d)
			int_cntrl->ilr6 = 0x100;
		if (ipl == 0x20)
			int_cntrl->ilr0 |= 0x2;
		if (ipl == 0x21)
			int_cntrl->ilr0 |= 0x200;
		int_cntrl->imr = 0x1;

		if (ipl >= 0x20 && ipl < 0x30) {
			sr |= SR_IM7;
			intr_func[2+ipl-0x20] = func;
		}
	}
	_mips_mtc0(C0_SR, sr);

#endif

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/24/02
 *
 *-----------------------------------------------------------------
 */
int32
intr_unregisterHandler(
	uint32	ipl
)
{
#if defined(R7000) || defined(SB1250)

	uint32	sr;

	if ((ipl < 0) || (ipl > 7)) {
		printf("Invalid interrupt level, ipl = %d\n", ipl);
		return -1;
	}

	intr_func[ipl] = NULL;

	sr = _mips_mfc0(C0_SR);
	sr &= ~(SR_IM0 << ipl);
	_mips_mtc0(C0_SR, sr);

#elif defined(TX3927)

#endif
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:	intr_init()
 *
 * Description:	interrupt initialization function:
 *					zero's registered interrupt handler functions
 *					enables interrupts in CPU core
 *
 * Created by:	Rick Lund
 *
 * Date Created:	2/28/02
 *
 *-----------------------------------------------------------------
 */
void
intr_init()
{
	uint32	ipl;
#if defined(R7000) || defined(SB1250)
	uint32	cause;
#endif
	uint32	sr;
	extern void sw_int_hndlr(uint32 sr, uint32 cause, address_t epc);

	for (ipl=0; ipl<=7; ipl++)
		intr_func[ipl] = NULL;

#if defined(R7000) || defined(SB1250)
	cause = _mips_mfc0(C0_CAUSE);
	cause |= CAUSE_IV;
	_mips_mtc0(C0_CAUSE, cause);
#endif

	sr = _mips_mfc0(C0_SR);
#if defined(R7000) || defined(SB1250)
	sr |= SR_IE;
#elif defined(TX3927)
	sr |= SR_IEC;
#endif
	_mips_mtc0(C0_SR, sr);

	intr_registerHandler(0,sw_int_hndlr);
	intr_registerHandler(1,sw_int_hndlr);

#ifdef FCNIM
	{
	extern void gtint_init();
	gtint_init();
	}
#endif

}

