/*
 * Name:  filesh.c
 *
 * RCS:  $Id: filesh.c,v 1.3 2002/08/07 22:33:15 bob Exp $
 *
 * Description:  Shell-like commands for file access on eagle filesystem.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  11/09/2001
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

#include <nfx-incl.h>
#include "file-api.h"

int cmd_touch(int argc, char *argv[]);
int cmd_echo(int argc, char *argv[]);
int cmd_cp(int argc, char *argv[]);
int cmd_cat(int argc, char *argv[]);
int cmd_rm(int argc, char *argv[]);
int cmd_mkdir(int argc, char *argv[]);
int cmd_rmdir(int argc, char *argv[]);
int cmd_mv(int argc, char *argv[]);
int cmd_help(int argc, char *argv[]);

struct
{
    char    *name;
    int     (*cmd)(int argc, char **argv);
    int     minArgs;
    int     maxArgs;
    char    *usage;
} cmds[] = {
    {"touch",   cmd_touch,  1,  1,  "touch <filename>"},
    {"echo",    cmd_echo,   2,  2,  "echo <string> <toFilename>"},
    {"cp",      cmd_cp,     2,  2,  "cp <fromFilename> <toFilename>"},
    {"cat",     cmd_cat,    1,  1,  "cat <filename>"},
    {"rm",      cmd_rm,     1,  1,  "rm <filename>"},
    {"mkdir",   cmd_mkdir,  1,  1,  "mkdir <dirname>"},
    {"rmdir",   cmd_rmdir,  1,  1,  "rmdir <dirname>"},
    {"mv",      cmd_mv,     2,  2,  "mv <fromname> <toname>"},
    {"help",    cmd_help,   0,  0,  "help"},
    {0,0,0,0,0}
};

#define READ_BLOCK_SIZE 1024
char readBuf[READ_BLOCK_SIZE];
    
/**
 * Create a file.
 */
int
cmd_touch(int argc, char **argv)
{
    int fd;

    fd = file_create(argv[1], 0644);
    if (fd < 0)
    {
        printf("%s: create failed\n", argv[1]);
        return 2;
    }

    return 0;

} /* cmd_touch */

/**
 * Write a string to a file.
 */
int
cmd_echo(int argc, char **argv)
{
    int fd, rc = 0;

    fd = file_create(argv[2], 0644);
    if (fd < 0)
    {
        printf("%s: create failed\n", argv[2]);
        return 2;
    }

    rc = file_write(fd, argv[1], strlen(argv[1]));
    if (rc != strlen(argv[1]))
    {
        printf("%s: write failed\n", argv[2]);
        rc = 2;
    }

    return rc;

} /* cmd_echo */

/**
 * Copy a file to another.  From file can be local (UNIX) or
 * on the NFX.
 */
int
cmd_cp(int argc, char **argv)
{
    int rc = 0, fromFd, toFd, numRead, numWritten;
    int (*fromRead)(int, char *, int);

    if (strchr(argv[1], ':') != NULL)
    {
        // fromFile is on the NFX
        fromFd = file_open(argv[1]);
        fromRead = file_read;
    }
    else
    {
        // fromFile is local
        fromFd = open(argv[1], 0);
        fromRead = read;
    }
    if (fromFd < 0)
    {
        printf("%s: file not found\n", argv[1]);
        return 2;
    }

    toFd = file_create(argv[2], 0644);
    if (toFd < 0)
    {
        printf("%s: create failed\n", argv[2]);
        return 2;
    }

    while ((numRead = fromRead(fromFd, readBuf, READ_BLOCK_SIZE)) > 0)
    {
        numWritten = file_write(toFd, readBuf, numRead);
        if (numWritten != numRead)
        {
            printf("%s: write failed\n", argv[2]);
            rc = 2;
            break;
        }
    }

    return rc;

} /* cmd_cp */

/**
 * Read and display a file.
 */
int
cmd_cat(int argc, char **argv)
{
    int fd, rc = 0, numRead, numWritten;

    fd = file_open(argv[1]);
    if (fd < 0)
    {
        printf("%s: file not found\n", argv[1]);
        return 2;
    }

    while((numRead = file_read(fd, readBuf, READ_BLOCK_SIZE)) > 0)
    {
        numWritten = write(1, readBuf, numRead);
        if (numWritten != numRead)
        {
            perror("write to stdout");
            rc = 2;
            break;
        }
    }

    return rc;

} /* cmd_cat */

/**
 * Remove a file.
 */
int
cmd_rm(int argc, char **argv)
{
    int rc;

    rc = file_unlink(argv[1]);
    if (rc < 0)
    {
        printf("%s: remove failed\n", argv[1]);
        return 2;
    }

    return 0;

} /* cmd_rm */

/**
 * Create a directory.
 */
int
cmd_mkdir(int argc, char **argv)
{
    int rc;

    rc = file_mkdir(argv[1], 0755);
    if (rc < 0)
    {
        printf("%s: mkdir failed\n", argv[1]);
        return 2;
    }

    return 0;

} /* cmd_mkdir */

/**
 * Remove a directory.
 */
int
cmd_rmdir(int argc, char **argv)
{
    int rc;

    rc = file_rmdir(argv[1]);
    if (rc < 0)
    {
        printf("%s: rmdir failed\n", argv[1]);
        return 2;
    }

    return 0;

} /* cmd_rmdir */

/**
 * Move file/dir.
 */
int
cmd_mv(int argc, char **argv)
{
    int rc;

    rc = file_rename(argv[1], argv[2]);
    if (rc < 0)
    {
        printf("%s: rename failed\n", argv[1]);
        return 2;
    }

    return 0;

} /* cmd_mv */


/**
 * Display help text.
 */
int
cmd_help(int argc, char **argv)
{
    int     i;

    for (i = 0; cmds[i].name != NULL; i++)
    {
        printf("%s\n", cmds[i].usage);
    }

    return 0;

} /* cmd_help */

int
main(int argc, char **argv)
{
    int rc, i;

    argc -= 1;
    if (argc == 0)
    {
        cmd_help(argc, argv);
        exit(2);
    }
    for (i = 0; cmds[i].name != NULL; i++)
    {
        if (strcmp(cmds[i].name, argv[1]) == 0)
        {
            if (argc < cmds[i].minArgs + 1 || argc > cmds[i].maxArgs + 1)
            {
                printf("usage: %s\n", cmds[i].usage);
                exit(2);
            }            
            argv++;
            rc = cmds[i].cmd(argc, argv);
            exit(rc);
        }
    }
    printf("%s: command %s not found\n", argv[0], argv[1]);
    exit(2);

} /* main */
