// -*- tab-width: 4 -*-

/*
 * btrace.h -- debugging utility that provides printings of the backtrace of
 * the current function.
 * originally:  btrace.c,v 1.3 2000/10/13 00:00:59
 *
 * Borrowed for sm-fs.
 *
 * $Id: btrace.c,v 1.11 2002/01/29 23:03:06 haahr Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <elf.h>
#include <stab.h>
#include <execinfo.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <assert.h>

#include "btrace.h"

#if defined(linux) && defined(i386)
#define HAVE_UCONTEXT
#ifndef __USE_GNU
#define __USE_GNU
#endif
#endif

#ifdef HAVE_UCONTEXT
#include <ucontext.h>
#endif

#define DEFAULT_BTRACE_LEVEL        30

#define streq(s, t)     (strcmp(s, t) == 0)

#if !defined(offsetof)
# define offsetof(T,F) ((unsigned int)((char *)&((T *)0L)->F - (char *)0L))
#endif

typedef enum { FALSE, TRUE } boolean;


/**
 * Mappings of offsets within function to their source file line numbers.
 * Only correct for functions entirely contained within a single source
 * file, which is by far the common case.
 */
typedef struct {
    u_int offset;               // offset within function
    u_int line;                 // line number within source file
} Posn;


/**
 * The type of symbol table entries.
 */
typedef struct btrace_st_tag {
    u_char *name;               // symbol name
    void *begin, *end;          // start and end addresses respectively
    const char *sourcefile;     // source file name
    Posn *posn;                 // array of (offset, line number) pairs
    int maxposn;                // size of position array.
    int maxoffset;              // maximum offset within function
} btrace_st_t;
typedef btrace_st_t Func;


/**
 * Our global symbol and addresses table.
 */
static Func *btrace_st;


/**
 * Return the name for the executable file being run by this process.
 */
static u_char *
get_exec_name(void)
{
    pid_t pid;
    u_char exe[PATH_MAX], fname[PATH_MAX];

    pid = getpid();

    snprintf(exe, sizeof(exe) - 1, "/proc/%u/exe", (u_int) pid);

    if (readlink(exe, fname, sizeof(fname) - 1) == -1) {
        perror("readlink");
        return NULL;
    }

    return strdup(fname);
}


/**
 * Open a file in read-only mode, printing an error message if we
 * failed to open it.
 */
static int
open_file(const u_char *file_name)
{
    int fd;

    if ((fd = open(file_name, O_RDONLY)) == -1) {
        perror(file_name);
        return -1;
    }

    return fd;
}


/**
 * Ensure that we're looking at a well-formed ELF file.
 */
static int
check_sanity(const Elf32_Ehdr *elf_hdr)
{
    if (memcmp(elf_hdr->e_ident, ELFMAG, SELFMAG) != 0) {
        fprintf(stderr, "not a valid ELF file\n");
        return -1;
    }

    if (elf_hdr->e_ident[EI_CLASS] != ELFCLASS32) {
        fprintf(stderr, "file class not supported\n");
        return -1;
    }

    if (elf_hdr->e_ident[EI_DATA] == ELFDATANONE) {
        fprintf(stderr, "invalid data encoding\n");
        return -1;
    }

    if (elf_hdr->e_ident[EI_VERSION] == EV_NONE) {
        fprintf(stderr, "invalid ELF file version\n");
        return -1;
    }

    if (elf_hdr->e_type != ET_EXEC) {
        fprintf(stderr, "not an executable ELF file\n");
        return -1;
    }

    if (elf_hdr->e_version == EV_NONE) {
        fprintf(stderr, "wrong version number in ELF file\n");
        return -1;
    }

    if (elf_hdr->e_phoff == 0) {
        fprintf(stderr, "program header not found\n");
        return -1;
    }

    if (elf_hdr->e_shoff == 0) {
        fprintf(stderr, "section header not found\n");
        return -1;
    }

    return 0;
}


/**
 * Return the index of the symbol table, using the specified table of
 * section headers.
 */
static int
get_symtab_idx(const Elf32_Shdr *sec_hdr, int num_sects)
{
    int i;
    int symtab_idx = 0;

    for (i = 0; i < num_sects; i++)
        if (sec_hdr[i].sh_type == SHT_SYMTAB) {
            symtab_idx = i;
            break;
        }

    return symtab_idx;
}


/**
 * Return the string table linked to the specified symbol table.
 */
static int
get_strtab_idx(const Elf32_Shdr *symtab)
{
    return symtab->sh_link;
}


/**
 * Return the number of symbols in the given symbol table.
 */
static int
get_num_syms(const Elf32_Shdr *symtab)
{
    return (symtab->sh_size / symtab->sh_entsize);
}


/**
 * Fill in the symbol table using the specifed ELF symbol table and
 * string table.
 */
static void
fill_symtab(const Elf32_Sym *sym_ent, int num_syms, const u_char *strtab)
{
    int i, j;

    /*
     * 1st we need to count how many elements our table will hold and then
     * we'll allocate enough space.
     */
    for (i = 1, j = 0; i < num_syms; i++)
        if (ELF32_ST_TYPE(sym_ent[i].st_info) == STT_FUNC)
            ++j;

    btrace_st = (Func *) calloc(j + 1, sizeof (Func));
    if (btrace_st == NULL) {
           perror("calloc");
           abort();
    }

    /*
     * We start with the symbol table entry number one because number 0
     * (STN_UNDEF) is reserved.
     */
    for (i = 1, j = 0; i < num_syms; i++) {
        if (ELF32_ST_TYPE(sym_ent[i].st_info) != STT_FUNC)
            continue;
        if (sym_ent[i].st_shndx == 0)
            continue;

        btrace_st[j].name = strdup(strtab + sym_ent[i].st_name);
        if (btrace_st[j].name == NULL) {
            perror("strdup");
            abort();
        }
        btrace_st[j].begin = (void *) sym_ent[i].st_value;
        btrace_st[j].end = btrace_st[j].begin + sym_ent[i].st_size;

        ++j;
    }

    // The table is finished by a void entry.
}


/**
 * Return the name of the section of the ELF file with the specified
 * section header index.
 */
static const char *
section_header_name(const char *contents, int index)
{
    Elf32_Ehdr *elf_hdr = (Elf32_Ehdr *) &contents[0];
    Elf32_Shdr *sec_hdr = (Elf32_Shdr *) &contents[elf_hdr->e_shoff];
    const char *nametab = &contents[sec_hdr[elf_hdr->e_shstrndx].sh_offset];
    return &nametab[sec_hdr[index].sh_name];
}


/**
 * Return the section header for the section with the specified.
 */
static Elf32_Shdr *
section_header_for_name(const char *contents, const char *name)
{
    int i;
    Elf32_Ehdr *elf_hdr = (Elf32_Ehdr *) &contents[0];
    Elf32_Shdr *sec_hdr = (Elf32_Shdr *) &contents[elf_hdr->e_shoff];
    for (i = 0; i < elf_hdr->e_shnum; i++) {
        if (streq(section_header_name(contents, i), name))
            return &sec_hdr[i];
    }
    return NULL;
}

/**
 * Return a function descriptor for the given program address.
 *
 * TODO - binary search, if btrace array is ordered
 */
static Func *
addr_to_func(void *addr)
{
    int i;

    btrace_init();

    for (i = 0; btrace_st[i].name != NULL; i++)
        if ((addr >= btrace_st[i].begin) && (addr < btrace_st[i].end))
            return &btrace_st[i];

    return NULL;
}


/**
 * An entry in the "stab" debugging information.
 */
typedef struct 
{
    u_long strx;                // index into string table
    u_char type;                // type of symbol
    u_char other;               // misc info (usually empty)
    u_short desc;               // description field
    u_long value;               // value of symbol
} Stab;
    

/**
 * Load the line numbers for the specified function, starting from
 * the specified stab entry, which should be a pointer to the N_FUN
 * stab for the function.
 */
static void
load_function_lines(Func *func, Stab *stabs, const char *strtab)
{
    int i;
    int max;
    int maxposn = 0;

    assert(stabs[0].type == N_FUN);
    assert(func->begin == (void *) stabs[0].value);

    func->sourcefile = NULL;
    func->posn       = NULL;
    func->maxposn    = 0;
    func->maxoffset  = 0;

    for (i = 1;; i++) {
        Stab *stab = &stabs[i];

        if (stab->type == N_FUN) {
            assert(stab->strx == 0);
            max = i;
            func->maxoffset = stab->value;
            break;
        }

        else if (stab->type == N_SLINE) {
            maxposn++;
        }
    }

    if (maxposn > 0) {
        int j;
        Posn *posn = malloc(maxposn * sizeof (Posn));
        if (posn == NULL)
            return;
        
        for (i = 1, j = 0; i < max; i++) {
            Stab *stab = &stabs[i];
            if (stab->type == N_SLINE) {
                assert(stab->value <= func->maxoffset);
                posn[j].offset = stab->value;
                posn[j].line   = stab->desc;
                j++;
            }
        }
        assert(j == maxposn);
        
        func->posn = posn;
        func->maxposn = maxposn;
    }
}


/**
 * A chain entry for struniq.
 */
typedef struct StringList StringList;
struct StringList {
    StringList *next;
    char string[1];             // variable length
};


/**
 * Make a string list entry.
 */
static StringList *makeStringList(
    const char *string,
    StringList *next)
{
    int len = strlen(string);
    StringList *list = malloc(offsetof(StringList, string[len + 1]));
    list->next = next;
    strcpy(list->string, string);
    return list;
}


/**
 * Number of chains for struniq hash table.
 */
enum { STRUNIQ_HASH_MAX = 1009 };


/**
 * Hash function over strings.
 */
static unsigned int strhash(const char *string)
{
    const unsigned char *s;
    int c;
    int hash = 0;
    for (s = string; (c = *s) != '\0'; s++) {
        hash = (hash << 3) | c;
    }
    return hash;
}


/**
 * Return a unique copy of a string name.
 */
static char *struniq(const char *string)
{
    int n = strhash(string) % STRUNIQ_HASH_MAX;
    StringList *list;

    static StringList *uniq[STRUNIQ_HASH_MAX];

    for (list = uniq[n]; list != NULL; list = list->next) {
        if (streq(string, list->string))
            return list->string;
    }

    uniq[n] = makeStringList(string, uniq[n]);
    return uniq[n]->string;
}


/**
 * Load line numbers into the function symbol table.
 *
 * TODO - do this lazily
 */
static void
load_line_numbers(const char *contents)
{
    int i;
    
    Elf32_Shdr *stab_hdr = section_header_for_name(contents, ".stab");
    Stab *stabs = (Stab *) &contents[stab_hdr->sh_offset];
    int stabmax = stab_hdr->sh_size / stab_hdr->sh_entsize;

    Elf32_Shdr *stabstr_hdr = section_header_for_name(contents, ".stabstr");
    const char *strtab = &contents[stabstr_hdr->sh_offset];

    const char *sourcefile = NULL;

    Func *lastfunc = NULL;

    assert(stab_hdr->sh_entsize == sizeof (Stab));

    for (i = 0; i < stabmax; i++) {
        Stab *stab = &stabs[i];

        //printf("%5d  %#4x  %#4x  %6d  %#9lx  %s\n",
        //       i, stab->type, stab->other, stab->desc, stab->value,
        //       &strtab[stab->strx]);

        if (stab->type == N_FUN && stab->strx != 0) {
            Func *func = addr_to_func((void *) stab->value);
            if (func != NULL) {
                load_function_lines(func, stab, strtab);
                if (sourcefile != NULL)
                    func->sourcefile = sourcefile;
            }
            lastfunc = func;
        }

        else if (stab->type == N_SOL || stab->type == N_SO) {
            sourcefile = struniq(&strtab[stab->strx]);

            if (lastfunc != NULL && sourcefile != NULL)
                lastfunc->sourcefile = sourcefile;
        }
    }
}


/**
 * Find the line number for a specified offset within the specified function.
 * Returns a negative number if a line number could not be found.
 *
 * TODO - binary search?
 */
static int
find_line_number(Func *func, int offset)
{
    int i;
    if (offset < 0 || func->maxoffset < offset)
        return -1;
    for (i = 0; i < func->maxposn; i++) {
        Posn *posn = &func->posn[i];
        if (posn->offset <= offset) {
            int nextoffset = (((i + 1) < func->maxposn)
                              ? posn[1].offset
                              : func->maxoffset);
            if (offset < nextoffset)
                return posn->line;
        }
    }
    return -1;
}


/**
 * Load the symbol table from the specified ELF file.
 */
static int
create_symtab(int fd)
{
    int symtab_idx, strtab_idx, num_syms;
    u_char *contents, *strtab;
    Elf32_Ehdr *elf_hdr;
    Elf32_Shdr *sec_hdr;
    Elf32_Off offset;
    Elf32_Sym *sym_ent;
    struct stat st;

    if (fstat(fd, &st) == -1) {
        perror("fstat");
        return -1;
    }

    contents = (u_char *) mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
    if (contents == MAP_FAILED) {
        perror("mmap");
        return -1;
    }

    elf_hdr = (Elf32_Ehdr *) &contents[0];

    if (check_sanity(elf_hdr) == -1)
        return -1;

    sec_hdr = (Elf32_Shdr *) &contents[elf_hdr->e_shoff];

    if ((symtab_idx = get_symtab_idx(sec_hdr, elf_hdr->e_shnum)) == 0) {
        fprintf(stderr, "no symbols were associated to this file\n");
        return -1;
    }

    if ((strtab_idx = get_strtab_idx(&sec_hdr[symtab_idx])) == 0) {
        fprintf(stderr, "the symbol table didn't have an "
            "associated string table, wtf??\n");
        return -1;
    }

    if (sec_hdr[strtab_idx].sh_type != SHT_STRTAB) {
        fprintf(stderr, "the associated string table for the symbol "
            "table is not valid!\n");
        return -1;
    }

    offset = sec_hdr[strtab_idx].sh_offset;
    strtab = &contents[offset];

    if ((num_syms = get_num_syms(&sec_hdr[symtab_idx])) == 0) {
        fprintf(stderr, "there were no symbols in the symbol table!\n");
        return -1;
    }

    offset = sec_hdr[symtab_idx].sh_offset;
    sym_ent = (Elf32_Sym *) &contents[offset];

    fill_symtab(sym_ent, num_syms, strtab);

    load_line_numbers(contents);

    if (munmap(contents, st.st_size) == -1) {
        perror("munmap");
        return -1;
    }
    return 0;
}


/**
 * Initialize the backtrace library.
 */
void
btrace_init(void)
{
    int exec_fd;
    u_char *number = NULL, *fname = NULL;

    static boolean initialized = FALSE;
    if (initialized)
        return;
    initialized = TRUE;

    if ((number = getenv("BTRACE_LEVEL")) == NULL)
        btrace_level = DEFAULT_BTRACE_LEVEL;
    else
        btrace_level = (u_int) strtoul(number, NULL, 10);

    if ((fname = get_exec_name()) == NULL)
        abort();

    if ((exec_fd = open_file(fname)) == -1)
        abort();

    if ((create_symtab(exec_fd)) == -1)
        abort();

    free(fname);
    close(exec_fd);
}


/**
 * Print a single call line from a backtrace.  The is_first parameter
 * is used to determine whether this is the first line of the traceback.
 * The is_call indicates whether the last instruction executed was likely
 * to be a call instruction.
 */
static void
print_call(Func *func, void *addr, boolean is_first, boolean is_call)
{
    int offset = ((char *) addr) - ((char *) func->begin);
    int line = find_line_number(func, is_call ? (offset - 1) : offset);

    fprintf(stderr, "  %9s %p %s +%d",
            is_first ? "at" : "called by", addr, func->name, offset);

    if (func->sourcefile != NULL || line >= 0) {
        fprintf(stderr, " (");
        if (func->sourcefile != NULL) {
            fprintf(stderr, "%s", func->sourcefile);
        }
        if (line >= 0) {
            fprintf(stderr, "%s%d",
                    ((func->sourcefile == NULL) ? "line " : ":"), line);
        }
        fprintf(stderr, ")");
    }

    fprintf(stderr, "\n");
    fflush(stdout);
}


/**
 * Return the level of backtraces to generate, after initializing the
 * backtrace library.
 */
static int
get_btrace_level()
{
    btrace_init();
    return btrace_level;
}


/**
 * Print a stack backtrace from the current call state.
 */
void
btrace_print()
{
    void *ba[get_btrace_level() + 1];
    int n = backtrace(ba, sizeof (ba) / sizeof (ba[0]));
    int i;
    for (i = 1; i < n; i++) {
        void *addr = ba[i];
        Func *func = addr_to_func(addr);
        if (func != NULL) {
            print_call(func, addr, (i == 1), TRUE);
        }
    }
}


/**
 * Translate a function address to a name.
 */
static u_char *
addr_to_name(void *addr)
{
    Func *func = addr_to_func(addr);
    if (func == NULL)
        return NULL;
    else
        return func->name;
}


/**
 * prints a backtrace of function names and addresses, starting from
 * the point that a signal was delivered.
 */
void
btrace_print_signal(void *context, void *handler)
{
#ifdef HAVE_UCONTEXT
    ucontext_t *ucontext = context;
    void *pc = (void *) ucontext->uc_mcontext.gregs[REG_EIP];
    Func *func;

    btrace_init();

    func = addr_to_func(pc);
    if (func != NULL) {
        int i;
        boolean started = FALSE;

        void *ba[get_btrace_level() + 1];
        int n = backtrace(ba, sizeof (ba) / sizeof (ba[0]));

        print_call(func, pc, TRUE, FALSE);
        for (i = 1; i < n; i++) {
            void *addr = ba[i];
            func = addr_to_func(addr);
            if (func != NULL) {
                if (!started) {
                    started = (func->begin == handler);
                }
                else {
                    print_call(func, addr, FALSE, TRUE);
                }
            }
        }
        if (started)
            return;
    }

#endif /* HAVE_UCONTEXT */

    btrace_print();
}


/**
 * Print the function name or hex address for the specified PC into the
 * specified buffer.  Returns the number of bytes written.
 */
int btrace_sprint_function(char *buf, void *pc)
{
    Func *func = addr_to_func(pc);
    if (func == NULL)
        return sprintf(buf, "%p", pc);
    else {
        int offset = ((char *) pc) - ((char *) func->begin);
        return sprintf(buf, "%s+%d", func->name, offset);
    }
}


/**
 * Print the source file name and line number for the specified PC if
 * known to the specified buffer, or do nothing if the line is not known.
 * Returns the number of bytes written.
 */
int btrace_sprint_source_line(char *buf, void *pc)
{
    Func *func = addr_to_func(pc);
    if (func == NULL)
        return 0;
    else {
        int offset = ((char *) pc) - ((char *) func->begin);
        int line = find_line_number(func, offset);
        int n;

        if (line < 0 && func->sourcefile == NULL)
            return 0;

        n = sprintf(buf, " at ");
        if (func->sourcefile != NULL) {
            n += sprintf(&buf[n], "%s", func->sourcefile);
        }
        if (line >= 0) {
            n += sprintf(&buf[n], "%s%d",
                         ((func->sourcefile == NULL) ? "line " : ":"), line);
        }
        return n;
    }
}
