/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * $Id: btree.c,v 1.65 2002/12/04 20:01:09 bob Exp $
 */

#if !defined(SSC)

#include "btree.h"

static int fs_bt_find_child(fs_Buf *buf, uint64 key);
static uint64 fs_bt_get_child_blknum(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i, boolean do_cow);
// PR #1904: need to take additional do_cow boolean
static fs_Buf *fs_bt_get_child(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i, boolean exclusive, boolean do_cow);
static void fs_bt_insertrecord(fs_Context *context, fs_Buf *buf, int i, uint64 key, uint64 blknum);
static int fs_bt_deleterecord(fs_Context *context, fs_Buf *buf, int i, boolean addLog);
static fs_Bt_split fs_bt_split(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int pivot);
static uint64 *fs_bt_get_child_slot(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i);

/*
 * Initialize a btree inner node to a given level, and set up the index heap to
 * its initial state (index[i] == i).
 */
void
fs_bt_init(fs_Context *context, fs_Buf *buf, int level)
{
	fs_Btree *t = buf->u.btree;
	int i;

	assert(level > 0);

	t->nkeys = 0;
	t->level = level;
	for (i = 0; i < FS_BTREE_MAXKEYS; i++)
		t->index[i] = i;

    FS_BUF_ZERO_ALL(buf);
	TOUCH_ELEM(buf, &t->level);
	TOUCH_ELEM(buf, &t->nkeys);
	TOUCH(buf, t->index, sizeof t->index);
}

/*
 * Initialize a B-tree inner node with a single right pointer, and return the
 * newly allocated block number for this node.
 */
uint64
fs_bt_init_singleton(fs_Context *context, fs_Inode *inode, int level, uint64 right)
{
	fs_Buf *buf = B_CREATE(inode);
	fs_Btree *t = buf->u.btree;

	fs_bt_init(context, buf, level);
	t->right = right;
	TOUCH_ELEM(buf, &t->right);
	return fs_blockNumber(buf->block);
}

/*
 * Find the correct child to descend to in a B-tree inner node.
 * It must satisfy the condition that child[i] has a key such that
 * key <= child[i].key. If the key is greater than all the child[i]
 * keys, then we take the rightmost pointer, t->right.
 */
static int
fs_bt_find_child(fs_Buf *buf, uint64 key)
{
	fs_Btree *t = buf->u.btree;
	int bot, top;
	int i;

	bot = i = 0;
	top = t->nkeys;

	/*
	 * Binary search to a likely starting point.
	 */
	while (bot < top) {
		uint64 k;

		i = (top + bot) / 2;
		k = *get_key(t, i);
		if (key == k)
			break;
		if (key < k)
			top = i-1;
		else
			bot = i+1;
	}

	assert(i <= t->nkeys);
	assert(i >= 0);

	/*
	 * Move forward over small keys.
	 */
	while (i < t->nkeys && key > *get_key(t, i))
		i++;
	/*
	 * Back up over larger keys.
	 */
	while (i > 0 && key <= *get_key(t, i-1))
		--i;

	return i;
}

/*
 * Return TRUE if a broken btree node is disposed of by the user.
 */
static boolean
query_dispose(fs_Context *context, fs_Buf *buf, fs_Bt_eeksearch *r,
                int level, char *fmt, ...)
{
	va_list ap;
	char msg[256];

	va_start(ap, fmt);
	vsprintf(msg, fmt, ap);
	va_end(ap);

    context->eek.errcount++;

    if (r->fixIt) {
	    if (!fs_eek_query(context, "inode %lld btree node %lld: %s.\nDispose?",
		        buf->inum, fs_blockNumber(buf->block), msg)) {
			return FALSE;
        }
	} else {
        fs_eek_print_status(context, FALSE,
                            "%s: (I#%lld) CORRUPTED btree inner node %lld: %s",
                            context->fileSystem->super->mountName,
                            buf->inum, fs_blockNumber(buf->block), msg);
        return FALSE;
    }

    // just return status to avoid redundant fs_bt_init() call
    // before the buffer is free'd

    r->status = FS_BT_EEK_DISPOSE;
	return TRUE;
}

/*
 * Check an inner node for consistency before using it to access the
 * underlying file.
 */
boolean
fs_bt_eek_check(fs_Context *context, fs_Buf *buf, int level, uint64 max_blocknum,
                fs_Bt_eeksearch *r)
{
	uint32 index[FS_BTREE_MAXKEYS];
	fs_Btree *t = buf->u.btree;
	uint32 magic = 0xCAFEBABE; /* illegal index value used as sentinel */
	uint64 current_offset;
	int i;

	/*
	 * Verify that the index maps every entry somewhere.
	 */
	for (i = 0; i < FS_BTREE_MAXKEYS; i++)
		index[i] = magic;

	for (i = 0; i < FS_BTREE_MAXKEYS; i++) {
		if (t->index[i] >= FS_BTREE_MAXKEYS) {
			return query_dispose(context, buf, r, level,
					"index %d out of range: %d", i, t->index[i]);
		}
		if (index[t->index[i]] != magic) {
			return query_dispose(context, buf, r, level,
					"index %d double-mapped: %d", i, t->index[i]);
		}
		index[t->index[i]] = i;
	}

	/*
	 * Should catch any unmapped entries with a corresponding double-map,
	 * but assert to make sure.
     * eek should not assert for filesystem inconsistencies
	 */
	for (i = 0; i < FS_BTREE_MAXKEYS; i++) {
		//assert(index[i] != magic);
        if (index[i] == magic) {
			return query_dispose(context, buf, r, level,
					"unmapped index %d", i);
        }
    }

	/*
	 * Verify that btree offsets increase monotonically
	 */
	current_offset = *get_key(t, 0);
	for (i = 1; i < t->nkeys; i++) {
		if (*get_key(t, i) <= current_offset) {
			return query_dispose(context, buf, r, level,
				"indexes %d, %d have overlapping file offsets: %lld, %lld",
				i, i-1,
				*get_key(t, i), current_offset);
		}
		current_offset = *get_key(t, i);
	}

	/*
	 * Verify that each block number mapped is valid.
	 */
	for (i = 0; i < t->nkeys; i++) {
		if (*get_left(t, i) >= max_blocknum) {
			return query_dispose(context, buf, r, level,
				"index %d has block out of range: %lld",
				i, *get_left(t, i));
		}
	}
	if (t->right >= max_blocknum) {
		return query_dispose(context, buf, r, level,
				"right pointer has block out of range: %lld",
				t->right);
	}

	return TRUE;
}

/*
 * Perform a recursive search through the B-tree. The search results are
 * placed in the "void *data" argument so fs_bt_search itself returns void.
 * 
 * The return value is a key used to guide a sequential traversal of
 * all the leaves. If the return value is zero, then the parent picks
 * the next child branch to visit, by setting the key to "key+1".
 * XXX: explain this better.
 */
uint64
fs_bt_search(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *data,
	     fs_Bt_search_dispatch *search,
         boolean exclusive)
{
	int child_level = level - 1;
	uint64 next_key;
	fs_Buf *child;
	int i;

    assert(exclusive == FALSE);

	/* Traverse B-tree nodes descending to the left on <=. */
	i = fs_bt_find_child(buf, key);
	child = fs_bt_get_child(context, inode, buf, i, exclusive, FALSE);
	assert(child != NULL);
	next_key = search->func[child_level](context, inode, child, child_level, key, data, search, exclusive);

	// XXX: by using the next_key hack even on single probes, we limit
	// concurrency by forcing the bufRelease after the call to the
	// child search->func(). Overall, crab-walking to increase concurrency
	// could probably be revisited in the btree code.
	if (next_key == 0) {
		if (i < buf->u.btree->nkeys)
			next_key = *get_key(buf->u.btree, i) + 1;
	}

    fs_bufRelease(context, buf);
	return next_key;
}

uint64
fs_bt_walk(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *walk0,
	     fs_Bt_search_dispatch *search,
	     boolean exclusive)
{
	int child_level = level - 1;
	fs_Bt_walk *walk = walk0;
	uint64 next_key, *childp;
	fs_Block child_block;
	fs_Buf *child;
	int i;
	fs_Bt_eeksearch r;

	r.key = 0;
	r.status = FS_BT_EEK_OK;
	r.fixIt = FALSE;

	/*
	 * eek uses walk to mark blocks
	 * since eek does not assert for inconsistencies, do some basic
	 * sanity checks of inner node before traversing
	 */
	if (!fs_bt_eek_check(context, buf, level,
			     context->fileSystem->params.totalBlocks, &r)) {
		fs_bufReleaseExclusive(context, buf);
		return 0;
	}

	/*
	 * Traverse B-tree nodes descending to the left on <=.
	 */
	i = fs_bt_find_child(buf, key);
	childp = fs_bt_get_child_slot(context, inode, buf, i);
	child_block = fs_makeBlock(*childp);

	/*
	 * Avoid double-counting inner nodes by keeping track of
	 * the last block we walked at that level.
	 */
	assert(child_level >= 0 && child_level < FS_INOROOT_MAXLEVEL);
	if (walk->last_blk[child_level] != fs_blockNumber(child_block)) {
		fs_WalkResult result;
		result.type = (child_level == 0) ?
		    FS_WALK_LEAF_NODE_TYPE : FS_WALK_INNER_NODE_TYPE;
		result.block = child_block;
		walk->last_blk[child_level] = fs_blockNumber(child_block);
		/*
		 * Don't descend into the child block below if the
		 * worker function fails.
		 */
		if (!walk->worker(context, inode, &result, walk->arg)) {
			next_key = 0;
			goto abort;
		}
	}

	child = B_GET_EXCL(inode, child_block);

	assert(child != NULL);

	next_key = search->func[child_level](context, inode, child, child_level,
				       key, walk0, search, exclusive);

abort:
	if (next_key == 0) {
		if (i < buf->u.btree->nkeys)
			next_key = *get_key(buf->u.btree, i) + 1;
	}

	fs_bufReleaseExclusive(context, buf);
	return next_key;
}

/*
 * Eek an inner btree node
 *
 * If FS_BT_EEK_DISPOSE is returned, the buf is free'd and the bit unmarked
 * Return key zero to indicate no more searches at this node needed
 *
 * If FS_BT_EEK_COW is returned, the buf was cow'd and the cow'd block
 * is returned so the caller can set the key with the new block number
 * This return code implies an error since cow only happens when something
 * is written (ie a correction)
 */
fs_Bt_eeksearch
fs_bt_eek(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *eek0,
	     fs_Bt_eek_dispatch *disp)
{
	int child_level = level - 1;
	fs_Bt_eek *eek = eek0;
	uint8 *bitmap = eek->bitmap;
	uint64 child_blknum;
	fs_Buf *child;
	int i;
	fs_Bt_eeksearch r;

	r.key = 0;
	r.status = FS_BT_EEK_OK;
    r.fixIt = eek->fixIt;

	if (!fs_bt_eek_check(context, buf, level,
                        context->fileSystem->params.totalBlocks, &r)) {
        goto fs_bt_eek_err_exit;
    }

	/*
	 * Buf may need to be free'd; return early w/disposal.
	 */
	if (r.status == FS_BT_EEK_DISPOSE) {
        goto fs_bt_eek_err_exit;
	}

	/*
	 * Traverse B-tree nodes descending to the left on <=.
	 */
	i = fs_bt_find_child(buf, key);
	child_blknum = fs_bt_get_child_blknum(context, inode, buf, i, FALSE);

	/*
	 * Mark all inner nodes as long as eek is in that mode (EEK_MARK_NODES)
	 */
	if (eek->marktype & FS_EEK_MARK_NODES) {
		if (!fs_eek_mark(context, "btree inner node block number",
				        bitmap, child_blknum)) {
            goto fs_bt_eek_err_exit;
		}

		/*
		 * Store the blknum of the current child node so
		 * we don't visit it on a subsequent pass.
	 	 */
		eek->last_blk[child_level] = child_blknum;
	}

	child = B_GET_EXCL(inode, fs_makeBlock(child_blknum));
	if (child == NULL) {
        goto fs_bt_eek_err_exit;
    }

	r = disp->func[child_level](context, inode, child, child_level, key, eek0, disp);

	if (r.key == 0) {
		if (i < buf->u.btree->nkeys)
			r.key = *get_key(buf->u.btree, i) + 1;
	}

	if (r.status != FS_BT_EEK_OK) {
        uint64 block = fs_blockNumber(buf->block);
        boolean cow = FALSE;
        uint64 myCowBlock;

        if (!r.fixIt)
            goto fs_bt_eek_err_exit;

        /* 
         * Take care of my own COW concerns before writing to this buf
         * My COW status should be returned to caller
         */

        if (fs_blockIsCopyOnWrite(context, buf->block)) {
            myCowBlock = fs_make_block_copy(context, inode, block);
            buf = B_GET_EXCL(inode, fs_makeBlock(myCowBlock));
            cow = TRUE;
        }

        /*
         * COW happened lower down which means we'll need to
         * update the new child for this key
         */

        if (r.status == FS_BT_EEK_COW) {
            uint64 *bp;
            bp = fs_bt_get_child_slot(context, inode, buf, i);
            *bp = r.cowBlock;
            TOUCH_ELEM(buf, bp);
        }

	    /*
	     * If the child buffer was corrupted and the operator requested it to
	     * be disposed of, we remove the pointer to it at this level. If this
	     * causes the current btree node to go empty, remove it also so as to
	     * preserve the B-tree invariant (no inner nodes or leaves may be empty).
	     */

		if ((r.status == FS_BT_EEK_DISPOSE) &&
            (fs_bt_deleterecord(context, buf, i, FALSE) == FS_BT_REMOVE_OK_EMPTY)) {
            /* just unmark block from bitmap (buf freed already) */
		    fs_eek_unmark(context, bitmap, block);
            r.key = 0;  /* buf disposed, nothing else at this level to eek */
            return r;
		}

        /*
         * For directory eek, return EEK_FAILED back to the parent since
         * the directory indexer needs to be fixed for any leaf corruption.
         * If COW happened, this status implies failure.
         *
         * We've taken care of removing the corrupted child's key at this
         * level, return EEK_DISPOSE to parent only if this node becomes empty
         */

        if (cow) {
            r.status = FS_BT_EEK_COW;
            r.cowBlock = myCowBlock;
        } else {
            r.status = FS_BT_EEK_FAILED;
        }

	}

	if (!fs_bufWasModified(context, buf))
	    fs_bufReleaseExclusive(context, buf);

    return r;

fs_bt_eek_err_exit:

    r.key = 0;  /* error: nothing else at this level to eek */

    if (r.fixIt) {
        r.status = FS_BT_EEK_DISPOSE;
		fs_eek_unmark(context, bitmap, fs_blockNumber(buf->block));
        fs_bufFree(context, buf, BUF_DO_DEALLOCATE|BUF_DONT_LOG_FLUSH);
		return r;
    }

    /* just walking, cannot make modifications, return failed */

    r.status = FS_BT_EEK_FAILED;
    r.key = 0;
    fs_bufReleaseExclusive(context, buf);

	return r;
}

/*
 * To rebuild indexer: traverse dir btree and scan for key/hash
 */
fs_Bt_eeksearch
fs_bt_eek_scan(fs_Context *context,
	        fs_Inode *inode,
	        fs_Buf *buf,
	        int level,
	        uint64 key,
	        void *eek0,
	        fs_Bt_eek_dispatch *disp)
{
	int child_level = level - 1;
	uint64 child_blknum;
	fs_Buf *child;
	int i;
	fs_Bt_eeksearch r;

    r.key = 0;
    r.status = FS_BT_EEK_OK;

    /*
     * Traverse B-tree nodes descending to the left on <=.
     */
    i = fs_bt_find_child(buf, key);
    child_blknum = fs_bt_get_child_blknum(context, inode, buf, i, FALSE);

    child = B_GET_SHARED(inode, fs_makeBlock(child_blknum));
    r = disp->func[child_level](context, inode, child, child_level, key, eek0, disp);

    fs_bufRelease(context, buf);
    return r;
}

/*
 * Insert into a btree, called from the insert function for the root
 * of the given btree (directory btree, file btree, etc.)
 *
 * The "split_allowed" variable is used to guide optimistic insertions.
 * On the first try from the root, split_allowed is set to false. If
 * split_allowed is false and a split is required somewhere in the tree,
 * the return value FS_BT_NEEDS_SPLIT is bounced back.
 *
 * The root code can then try to reinsert with split_allowed set to true.
 * This implies exclusive locks on all the nodes during the descent,
 * and splits are performed as necessary.
 */
fs_Bt_insert
fs_bt_insert(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *data,
	     boolean split_allowed,
	     fs_Bt_insert_dispatch *insert)
{
	fs_Btree *t = buf->u.btree;
	fs_Bt_insert pivot;
	fs_Bt_insert p;
	fs_Buf *child;
	int child_level;
	boolean exclusive;
	int i;

	pivot.status = FS_BT_INSERT_OK;	/* default status -- assume success */
	pivot.blknum = 0;		/* default split result -- assume no split */

	if (split_allowed && t->nkeys == FS_BTREE_MAXKEYS) {
		fs_Bt_split sp = fs_bt_split(context, inode, buf, t->nkeys / 2);

		/*
		 * pick the appropriate child, i.e., side-step to my left
		 * sibling if the key is less-than-or-equal to the pivot.
		 * NB: an exclusive lock is held on the newly created left
		 * child.
		 */
		if (key <= sp.key) {
			buf = sp.buf;
			t = buf->u.btree;
		}
		pivot.key = sp.key;
		pivot.blknum = fs_blockNumber(sp.buf->block);
	}

	i = fs_bt_find_child(buf, key);

	/*
	 * If I'm about to visit a leaf or a split is possible, acquire an
	 * exclusive lock on the child.
	 */
    exclusive = (split_allowed || t->level == 1);
	child = fs_bt_get_child(context, inode, buf, i, exclusive, TRUE);
    /*
     * If child is NULL, it means we need to COW it, but
     * the inode is not held exclusive, so return NEEDS_SPLIT
     * to force our caller to re-acquire the inode in exclusive
     * mode.
     */
	//assert(child != NULL);
    if (child == NULL)
    {
        assert(!split_allowed);
		fs_bufRelease(context, buf);
        pivot.status = FS_BT_NEEDS_SPLIT;
        return pivot;
    }
        
	child_level = level - 1;

	/*
	 * In the optimistic case, unlock myself assuming no key will be
	 * promoted from below.
	 */
	if (!split_allowed)
		fs_bufRelease(context, buf);

	p = insert->func[child_level](context, inode, child, child_level, key, data, split_allowed, insert);

	if (p.status == FS_BT_NEEDS_SPLIT) {
		if (split_allowed && !fs_bufWasModified(context, buf))
			fs_bufReleaseExclusive(context, buf);
		return p;
	}

	/*
	 * Did a key get promoted from a split down below?
	 */
	if (p.blknum != 0) {
		assert(split_allowed);
		fs_bt_insertrecord(context, buf, i, p.key, p.blknum);
	} else {
		if (split_allowed && !fs_bufWasModified(context, buf))
			fs_bufReleaseExclusive(context, buf); 
	}

	pivot.got_new_leaf = p.got_new_leaf;

	return pivot;
}

/*
 * Insert a key promoted from a split in a child into a B-tree
 * inner node.
 */
static void
fs_bt_insertrecord(fs_Context *context, fs_Buf *buf, int i, uint64 key, uint64 blknum)
{
	fs_Btree *t = buf->u.btree;
	int new = t->index[t->nkeys];

	assert(i <= t->nkeys);

	if (i < t->nkeys) {
		memmove(&t->index[i+1], &t->index[i], (t->nkeys-i) * sizeof *t->index);
		t->index[i] = new;
	}

	t->keys[new] = key;
	t->left[new] = blknum;
	t->nkeys++;

	TOUCH_ELEM(buf, &t->nkeys);
	TOUCH_ELEM(buf, &t->keys[new]);
	TOUCH_ELEM(buf, &t->left[new]);
	TOUCH(buf, &t->index[i], (t->nkeys+1-i) * sizeof *t->index);
}

/*
 * Delete the ith record in the given B-tree inner node. If i == t->nkeys,
 * this refers to the rightmost pointer.
 *
 * One invariant of btrees is that the right pointer is always valid.
 * So, if the right pointer is being deleted (because the node it points
 * to is empty), we evict the rightmost left-pointer's key, and make
 * that the new right pointer:
 *
 *	    ---------
 *	    | A B C |
 *	    |a b c x|
 *	    ---------
 *
 * after the delete of "x", becomes:
 *
 *	    -------
 *	    | A B |
 *	    |a b c|
 *	    -------
 *
 * One special case is when t->nkeys is zero; in this case we invalidate
 * the right pointer and return an EMPTY status.
 */
static int
fs_bt_deleterecord(fs_Context *context, fs_Buf *buf, int i, boolean addLog)
{
	fs_Btree *t = buf->u.btree;

	if (i == t->nkeys) {
		assert(t->right != 0);

		if (t->nkeys == 0) {
            fs_bufFree(context, buf,
                (addLog ? BUF_DO_DEALLOCATE : BUF_DO_DEALLOCATE|BUF_DONT_LOG_FLUSH));
			return FS_BT_REMOVE_OK_EMPTY;
		}

		t->right = *get_left(t, t->nkeys-1);
		--t->nkeys;
		TOUCH_ELEM(buf, &t->right);
		TOUCH_ELEM(buf, &t->nkeys);
	} else {
		int old = t->index[i];

		if (i < t->nkeys - 1) {
			memmove(&t->index[i], &t->index[i+1],
				(t->nkeys - (i+1)) * sizeof *t->index);
			t->index[t->nkeys-1] = old;
			TOUCH(buf, &t->index[i], (t->nkeys-i) * sizeof *t->index);
		}
		--t->nkeys;
		TOUCH_ELEM(buf, &t->nkeys);
	}

	return FS_BT_REMOVE_OK;
}

int bt_splits;

static fs_Bt_split
fs_bt_split(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int pivot)
{
	uint16 swap[FS_BTREE_MAXKEYS];
	fs_Btree *t = buf->u.btree;
	fs_Bt_split p;
	fs_Buf *buf2;
	fs_Btree *t2;
	int i;

	bt_splits++;

	/*
	 * Remove the pivot from the current node, and split the
	 * node into two children.
	 *
	 * This is accomplished by copying any keys to the left
	 * of the pivot into a new Btree node:
	 *
	 *	    -----------------
	 *	    | A B C P D E F |
	 *	    |a b c p d e f g|
	 *	    -----------------
	 *
	 * becomes:
	 *
	 *	         ------
	 *	         | P  |
	 *	         |t2 t|
	 *	         ------
	 *               /    \
	 * 	---------      ---------
	 *   t2:| A B C |   t :| D E F |
	 *	|a b c p|      |d e f g|
	 *	---------      ---------
	 *
	 * When we promote "P" to the parent node, we also promote
	 * a pointer to "t2". This way, the key which is in the parent
	 * which was used to find "t" will still point to "t", and
	 * the new key "P" will point to "t2".
	 */
	buf2 = B_CREATE(inode);
	fs_bt_init(context, buf2, t->level);

	p.key = *get_key(t, pivot);
	p.buf = buf2;

	/*
	 * Copy over records to t2 up to the pivot. The right pointer of "t2" becomes
	 * the pivot record, but the key itself is promoted to the parent.
	 */
	t2 = buf2->u.btree;
	t2->nkeys = pivot;
	for (i = 0; i < pivot; i++) {
		*get_key(t2, i) = *get_key(t, i);
		*get_left(t2, i) = *get_left(t, i);
	}
	t2->right = *get_left(t, pivot);

	TOUCH_ELEM(buf2, &t2->nkeys);
	TOUCH_ELEM(buf2, &t2->right);
	TOUCH(buf2, &t2->keys, pivot * sizeof *t2->keys);
	TOUCH(buf2, &t2->left, pivot * sizeof *t2->left);

	/*
	 * Shift over the remaining records in "t" to the front, and shift
	 * the index entries we would overwrite to the middle of the index.
	 * Even though they are now being freed they have to point to the
	 * correct freed blknums.
	 * XXX: explain this better!
	 */
	t->nkeys -= (pivot+1);
	memcpy(swap, &t->index[0], (pivot+1) * sizeof *t->index);
	memmove(&t->index[0], &t->index[pivot+1], t->nkeys * sizeof *t->index);
	memcpy(&t->index[t->nkeys], swap, (pivot+1) * sizeof *t->index);

	TOUCH_ELEM(buf, &t->nkeys);
	TOUCH(buf, t->index, sizeof t->index);

	return p;
}

/*
 * Perform ultra-conservative deletion of records. This routine works top-down,
 * but does no merges.
 */
fs_Bt_remove
fs_bt_remove(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *data,
	     boolean dispose_allowed,
	     fs_Bt_remove_dispatch *remove)
{
	fs_Btree *t = buf->u.btree;
	int child_level;
	fs_Buf *child;
	fs_Bt_remove r;
	boolean exclusive;
	int i;

	/*
	 * Descend to a leaf node before deleting the record, scanning over
	 * possible dup keys.
	 */
	i = fs_bt_find_child(buf, key);

	/*
	 * If I'm about to visit a leaf or it's possible to free an inner node, acquire an
	 * exclusive lock on the child.
	 */
	exclusive = (dispose_allowed || t->level == 1);
	child = fs_bt_get_child(context, inode, buf, i, exclusive, TRUE);
    /*
     * If child is NULL, it means we need to COW it, but
     * the inode is not held exclusive, so return NEEDS_DISPOSAL
     * to force our caller to re-acquire the inode in exclusive
     * mode.
     */
	//assert(child != NULL);
    if (child == NULL)
    {
        assert(!dispose_allowed);
		fs_bufRelease(context, buf);
        r.status = FS_BT_REMOVE_NEEDS_DISPOSAL;
        return r;
    }

	child_level = level - 1;

	/*
	 * In the optimistic case, unlock myself assuming no page below
	 * will become empty.
	 */
	if (!dispose_allowed)
		fs_bufRelease(context, buf);
		
	r = remove->func[t->level-1](context, inode, child, child_level, key, data, dispose_allowed, remove);

	switch (r.status) {
	default:
		assert(0);
	case FS_BT_REMOVE_OK:
		if (dispose_allowed && !fs_bufWasModified(context, buf))
		    fs_bufReleaseExclusive(context, buf);
		break;
	case FS_BT_REMOVE_OK_EMPTY:
		assert(dispose_allowed);
		r.status = fs_bt_deleterecord(context, buf, i, TRUE);
		break;
	case FS_BT_REMOVE_NEEDS_DISPOSAL:
		assert(!dispose_allowed);
		break;
	}
	
	return r;
}

/*
 * Cleave at the "edge" of an inner node, i.e., when i==0 or i==t->nkeys.
 *
 * What's passed as an argument are pointers to the cleavepair we intend
 * to return, and pointers to the child cleavepair we got back from the
 * lower level cleave.
 *
 * Since this operation is symmetric, one edge_check() routine is used
 * to handle cleaving at i==0 or i==t->nekys. Hence, "a" and "child_a"
 * refer to the "inner" blocknumber of a cleavepair, and "b" and "child_b"
 * refer to the "outer" number of a cleavepair.
 *
 * i.e., if i==0, we are on the left hand side of the inner node, so the
 * "outer" cleavepair member is to the left of the "inner" member, and
 * so cleavepair.left == b, and cleavepair.right == a.
 * 
 * The situation is reversed at i==t->nkeys: cleavepair.left == a, and
 * cleavepair.right == b.
 *
 * The job of edge_check is to take the pair of buffers passed from
 * the child cleave and pass back a pair of buffers at the current
 * level to the parent.
 *
 * The first case is straightforward: if the "outer" buffer is null,
 * i.e., if the cleave for example removes a hole from the file but
 * doesn't produce any trash B-tree, then the current B-tree remains
 * unaltered and we pass back a cleavepair to reflect that:
 * 
 * 	cleavepair.b = 0, and cleavepair.a = buf->block.
 * 
 * The second case is trickier: First the trash in child_b must be
 * threaded into a new buffer, which we then pass back in cleavepair.b.
 * Hence the call to fs_bt_init_singleton.
 *
 * Then, child_a must be handled carefully. 1) If the operation on the
 * child node left the node empty, e.g., if the child cleave operation
 * cleanly removes the entire child node, then this node must be
 * removed from the inner node at this level. 2) If the child node
 * is otherwise unchanged, we can bufRelease the current inner node
 * since no change is required to it at this level either. 3) Depending
 * on the way the child is split into two new nodes, the old child node
 * may have found itself into child_cleavepair.b. In this case, the
 * new value must be swapped in to the current inner node, since otherwise
 * the new blocknumber in child_cleavepair.a will be leaked away.
 * 
 * XXX: I recognize this explanation would benefit greatly from the use of
 * diagrams!
 */
static void
edge_check(
	fs_Context *context,
	fs_Inode *inode,
	fs_Buf *buf,
	int i,
	uint64 key,
	uint64 *a,
	uint64 *b,
	uint64 *child_a,
	uint64 *child_b)
{
	uint64 *old_child = fs_bt_get_child_slot(context, inode, buf, i);
	fs_Btree *t = buf->u.btree;

	*a = fs_blockNumber(buf->block);

	if (*child_b == 0) {	/* no trash produced by child */
		FS_ASSERT(*child_a == *old_child);
		*b = 0;
        // PR #1917: buf may have been COW'd
	    if (!fs_bufWasModified(context, buf))
		    fs_bufReleaseExclusive(context, buf);
	} else {		/* thread trash into new btree node */
		*b = fs_bt_init_singleton(context, inode, t->level, *child_b);
		if (*child_a == 0) {
			if (fs_bt_deleterecord(context, buf, i, TRUE) == FS_BT_REMOVE_OK_EMPTY)
				*a = 0;
		} else if (*child_a == *old_child) {
            // PR #1917: buf may have been COW'd
	        if (!fs_bufWasModified(context, buf))
			    fs_bufReleaseExclusive(context, buf);
		} else {
			FS_ASSERT(*child_b == *old_child);
			*old_child = *child_a;
			TOUCH_ELEM(buf, old_child);
		}
	}
}

fs_Bufpair
fs_bt_cleave(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *data,
	     fs_Bt_cleave_dispatch *cleave)
{
	fs_Btree *t = buf->u.btree;
	int child_level = level - 1;
	fs_Buf *child;
	fs_Bufpair cleavepair, child_cleavepair;
	int i;

	i = fs_bt_find_child(buf, key);
	FS_ASSERT(fs_lockOwnedExclusive(context, &inode->btreeLock));
	child = fs_bt_get_child(context, inode, buf, i, TRUE, TRUE);
	FS_ASSERT(child != NULL);

	child_cleavepair = cleave->func[child_level](context, inode, child, child_level, key, data, cleave);

	/*
	 * Sort out the special cases first:
	 *
	 * If (i == 0 || i == t->nkeys) then we are cleaving at the edge of
	 * the current buffer so no split is necessary -- the new left or right
	 * buffer has one key in it and the old one remains unchanged with
	 * respect to the number of keys.
	 * 
	 * Also, if any of the child cleave blknums are zero, we can elide that
	 * part of the cleave operation.
	 *
	 * In the general case, a cleave is like a split, except that the
	 * pivot which gets extracted from the split operation must be
	 * re-inserted into the right node. XXX: diagram for this.
	 */


	if (i == t->nkeys) {
		edge_check(context, inode, buf, i, key,
			&cleavepair.left, &cleavepair.right,
			&child_cleavepair.left, &child_cleavepair.right);
	} else if (i == 0) {
		edge_check(context, inode, buf, i, key,
			&cleavepair.right, &cleavepair.left,
			&child_cleavepair.right, &child_cleavepair.left);
	} else {
		fs_Buf *buf2;
		fs_Btree *t2;

		buf2 = fs_bt_split(context, inode, buf, i).buf;
		t2 = buf2->u.btree;

		cleavepair.left = fs_blockNumber(buf2->block);
		cleavepair.right = fs_blockNumber(buf->block);

		if (child_cleavepair.right != 0) {
			fs_bt_insertrecord(context, buf, 0, key, child_cleavepair.right);
		}
		if (child_cleavepair.left != 0) {
			FS_ASSERT(t2->right == child_cleavepair.left
			       || t2->right == child_cleavepair.right);
			t2->right = child_cleavepair.left;
			TOUCH_ELEM(buf2, &t2->right);
		} else {
			if (fs_bt_deleterecord(context, buf2, t2->nkeys, TRUE) == FS_BT_REMOVE_OK_EMPTY)
				cleavepair.left = 0;
		}
	}

	return cleavepair;
}

static uint64
fs_bt_get_child_blknum(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i, boolean do_cow)
{
	fs_Btree *t = buf->u.btree;
	uint64 *blockp;

	if (i < t->nkeys)
		blockp = get_left(t, i);
	else
		blockp = &t->right;

	if (do_cow && fs_blockIsCopyOnWrite(context, fs_makeBlock(*blockp))) {
		if (!fs_lockOwnedExclusive(context, &buf->lock))
			return 0;
		*blockp = fs_make_block_copy(context, inode, *blockp);
		TOUCH_ELEM(buf, blockp);
	}

	return *blockp;
}

/*
 * Get a pointer to the ith block number in a buffer; used by b-tree cleave.
 */
static uint64 *
fs_bt_get_child_slot(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i)
{
	fs_Btree *t = buf->u.btree;
	uint64 *blockp;

	if (i < t->nkeys)
		blockp = get_left(t, i);
	else
		blockp = &t->right;

	return blockp;
}

/*
 * Get the i'th child of a Btree inner node.
 * PR #1904: need to take additional do_cow boolean
 */
static fs_Buf *
fs_bt_get_child(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i, boolean exclusive, boolean do_cow)
{
	uint64 blknum;
	fs_Buf *child;

	blknum = fs_bt_get_child_blknum(context, inode, buf, i, do_cow);

	if (blknum == 0)
		return NULL;

	if (exclusive)
		child = B_GET_EXCL(inode, fs_makeBlock(blknum));
	else
		child = B_GET_SHARED(inode, fs_makeBlock(blknum));

                
	return child;
}

#if 0 // not currently used
/*
 * Get the i'th child of a Btree inner node and split it off: i.e., 
 * assign it a new block number. Used by eek when two nodes point to
 * the same block.
 */
static void
fs_bt_split_child(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i)
{
	fs_Btree *t = buf->u.btree;
	uint64 *blockp;

	if (i < t->nkeys)
		blockp = get_left(t, i);
	else
		blockp = &t->right;

	*blockp = fs_make_block_copy(context, inode, *blockp);
	TOUCH_ELEM(buf, blockp);
}
#endif // 0

fs_Bt_dispatch bt_funcs;

void
fs_bt_dispatch_init()
{
	int level;

	for (level = 1; level < FS_BT_MAXLEVEL; level++) {
		bt_funcs.emap.insert.func[level] = fs_bt_insert;
		bt_funcs.emap.remove.func[level] = fs_bt_remove;
		bt_funcs.emap.search.func[level] = fs_bt_search;
		bt_funcs.emap.read.func[level] = fs_bt_search;
		bt_funcs.emap.eek_blocks.func[level] = fs_bt_eek;
		bt_funcs.emap.eek_inodes.func[level] = NULL;
		bt_funcs.emap.eek_scan.func[level] = NULL;
		bt_funcs.emap.shrink.func[level] = fs_bt_remove;
		bt_funcs.emap.cleave.func[level] = fs_bt_cleave;
		bt_funcs.emap.walk.func[level] = fs_bt_walk;

		bt_funcs.dir.insert.func[level] = fs_bt_insert;
		bt_funcs.dir.remove.func[level] = fs_bt_remove;
		bt_funcs.dir.search.func[level] = fs_bt_search;
		bt_funcs.dir.read.func[level] = fs_bt_search;
		bt_funcs.dir.eek_blocks.func[level] = fs_bt_eek;
		bt_funcs.dir.eek_inodes.func[level] = fs_bt_eek;
		bt_funcs.dir.eek_scan.func[level] = fs_bt_eek_scan;
		bt_funcs.dir.shrink.func[level] = NULL;
		bt_funcs.dir.cleave.func[level] = NULL;
		bt_funcs.dir.walk.func[level] = fs_bt_walk;
	}

	bt_funcs.emap.insert.func[0] = fs_emapleaf_insert;
	bt_funcs.emap.remove.func[0] = fs_emapleaf_remove;
	bt_funcs.emap.search.func[0] = fs_emapleaf_search;
	bt_funcs.emap.read.func[0] = NULL;
	bt_funcs.emap.eek_blocks.func[0] = fs_emapleaf_eek_mark;
	bt_funcs.emap.eek_inodes.func[0] = NULL;
	bt_funcs.emap.eek_scan.func[0] = NULL;
	bt_funcs.emap.shrink.func[0] = fs_emapleaf_shrink;
	bt_funcs.emap.cleave.func[0] = fs_emapleaf_cleave;
	bt_funcs.emap.walk.func[0] = fs_emapleaf_walk;
			
	bt_funcs.dir.insert.func[0] = NULL; // fs_dirleaf_insert;
	bt_funcs.dir.remove.func[0] = NULL; // fs_dirleaf_remove;
	bt_funcs.dir.search.func[0] = NULL; // fs_dirleaf_search;
	bt_funcs.dir.read.func[0] = NULL; // fs_dirleaf_read;
	bt_funcs.dir.eek_blocks.func[0] = NULL; // fs_dirleaf_eek_mark_blocks;
	bt_funcs.dir.eek_inodes.func[0] = NULL; // fs_dirleaf_eek_mark_inodes;
	bt_funcs.dir.eek_scan.func[0] = NULL; // fs_dirleaf_eek_scan;
	bt_funcs.dir.shrink.func[0] = NULL;
	bt_funcs.dir.cleave.func[0] = NULL;
	bt_funcs.dir.walk.func[0] = NULL; // fs_dirleaf_walk;
}

/*
 * Btree locking routines; return B-tree root portion of a Dinode.
 * (returned as void * to avoid casts).
 */

void *
fs_getBtreeExclusive(fs_Context *context, fs_Inode *inode)
{
	fs_lockExclusive(context, &inode->btreeLock, NULL);
	return &inode->dinode.u;
}

void *
fs_getBtreeShared(fs_Context *context, fs_Inode *inode)
{
	fs_lockShared(context, &inode->btreeLock, NULL);
	return &inode->dinode.u;
}

void
fs_releaseBtreeShared(fs_Context *context, fs_Inode *inode)
{
	fs_unlockShared(context, &inode->btreeLock);
}

void
fs_releaseBtreeExclusive(fs_Context *context, fs_Inode *inode)
{
	fs_unlockExclusive(context, &inode->btreeLock);
}

void
fs_demoteBtreeLock(fs_Context *context, fs_Inode *inode)
{
	fs_lockDemote(context, &inode->btreeLock, FS_LOCK_MODE_SHARE);
}

#endif // SSC
