/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: emapflatroot.c,v 1.38 2002/11/26 22:54:07 bob Exp $
 *
 */

#if !defined(SSC)

#include "btree.h"

/*
 * Do a compact spill.
 */
static void
fs_emaproot_spill_to_compact(fs_Context *context, fs_Inode *inode, fs_Buf *buf)
{
    fs_EmapleafCompact *l;
	fs_Emapflatroot *fr;
    int i;

	l = buf->u.emapleafCompact;
	fr = &inode->dinode.u.emapflatroot;

	//printf("spilling to compact leaf i#%lld %lld\n", inode->inum, buf->page);

	/*
	 * Initialize the new leaf by copying the extents from the
	 * flat root.
	 */
	for (i = 0; i < FS_INOFLATROOT_MAXKEYS; i++) {
		l->offsets[i] = fr->offsets[i];
		l->extents[i] = fr->extents[i];
	}

	TOUCH(buf, &l->extents[0], i * sizeof *l->extents);
	TOUCH(buf, &l->offsets[0], i * sizeof *l->offsets);

	l->nkeys = FS_INOFLATROOT_MAXKEYS;

	TOUCH_ELEM(buf, &l->nkeys);
}

/*
 * Do a full (ie. non-compact) spill.
 */
static void
fs_emaproot_spill_to_full(fs_Context *context, fs_Inode *inode, fs_Buf *buf)
{
    fs_Emapleaf *l;
    fs_Emapflatroot *fr;
    int i;

    l = buf->u.emapleaf;
    fr = &inode->dinode.u.emapflatroot;

    //printf("spilling to full leaf i#%lld %lld\n", inode->inum, buf->page);

    /*
     * Initialize the new leaf by copying the extents from the
     * flat root.
     */
    for (i = 0; i < FS_INOFLATROOT_MAXKEYS; i++) {
        l->offsets[i] = fr->offsets[i];
        l->extents[i] = fr->extents[i];
    }

    TOUCH(buf, &l->extents[0], i * sizeof *l->extents);
    TOUCH(buf, &l->offsets[0], i * sizeof *l->offsets);

    l->nkeys = FS_INOFLATROOT_MAXKEYS;

    TOUCH_ELEM(buf, &l->nkeys);
}


/*
 * Spill an inode root mini-"leaf page" into a full filesystem
 * buffer, and turn the inode root "leaf page" into a btree
 * root.
 */
void
fs_emaproot_spill(fs_Context *context, fs_Inode *inode)
{
	fs_Emapflatroot *fr;
	fs_Buf *buf = B_CREATE(inode);
	fs_Emaproot *r;

    /*
     * Don't do system files as compacts.
     */
    if (inode->inum >= FS_MAX_RESERVED_INODE) {
	    /*
	     * Repackage an "ultra" buf into a compact one.
	     */
        if (!fs_bufIsCompact(buf)) {
            fs_ramDealloc(FS_RAM_READ_DATA, buf->u.data);
		    buf->u.data = fs_ramAlloc(FS_RAM_COMPACT_LEAF,
                                      FS_COMPACT_PAGE_SIZE);
		    fs_bufLeafShrunk(context, buf, FALSE);
        }
	    fs_emapleaf_compact_init(context, buf);
	    inode->dinode.attrs |= FS_IFLG_COMPACT_EMAP;

        fs_emaproot_spill_to_compact(context, inode, buf);
    }
    else
    {
        assert(!(inode->dinode.attrs & FS_IFLG_COMPACT_EMAP));

	    fs_emapleaf_init(context, buf);
        fs_emaproot_spill_to_full(context, inode, buf);
    }

	fr = &inode->dinode.u.emapflatroot;

	/*
	 * Initialize the new root as keyless with a single right pointer
	 * pointing to the new leaf.
	 */
	r = (fs_Emaproot *)fr;
	r->hdr.nkeys = 0;
	r->hdr.level = 1;
	r->right = fs_blockNumber(buf->block);

	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
}

static void fs_emap_flat_removeent(fs_Emapflatroot *r, int i)
{
	/*
	 * Preserve the sorted order of extents by memmoving the other
	 * extents to fill the vacated entry.
	 */
	if (i < r->hdr.nkeys - 1) {
		memmove(&r->offsets[i], &r->offsets[i+1],
				(r->hdr.nkeys - (i+1)) * sizeof *r->offsets);
		memmove(&r->extents[i], &r->extents[i+1],
				(r->hdr.nkeys - (i+1)) * sizeof *r->extents);
	}

	--r->hdr.nkeys;
}

void
fs_emap_flat_add(fs_Context *context, fs_Inode *inode, uint64 offset, fs_Extent extent)
{
	fs_Emapflatroot *r;
	int i;

	r = &inode->dinode.u.emapflatroot;

	assert(r->hdr.nkeys < FS_INOFLATROOT_MAXKEYS);

	for (i = 0; i < r->hdr.nkeys; i++) {
		assert(offset != r->offsets[i]);
		if (offset < r->offsets[i])
			break;
	}

	/*
	 * Verify that the new offset+extent does not interfere with one
	 * of the neighboring offset+extents.
	 */
	assert(i == 0 || offset >= (r->offsets[i-1] +
                                    fs_extentLength(r->extents[i-1])));
	assert(i >= (r->hdr.nkeys-1)
               || offset + fs_extentLength(extent) <= r->offsets[i+1]);

	/*
	 * First check for merges between this extent and the extents
	 * to the left and right of the insertion point.
	 */

	/*
	 * Check for a merge to the right -- if this is successful, then
	 * check for a merge to the left: this double merge would cause
	 * the leaf to shrink by an element.
	 */
	if (i != r->hdr.nkeys) {
		if (fs_areAdjacent(extent, r->extents[i], offset, r->offsets[i])) {
			// printf("merging on the right\n");
			r->offsets[i] = offset;
			r->extents[i] = fs_mergeExtents(extent, r->extents[i]);

			/*
			 * Now that the new extent has been merged with
			 * the extent on the rhs, check to see whether
			 * it might "fill a hole" and merge with an
			 * extent to the left as well. In that case,
			 * grow the extent and remove the extent to
			 * the left.
			 */

            /*
             * call fs_areAdjacent w/ the new merged extent
             * since we need to check for length overflow in a 3 way merge
             */

			if (i != 0 &&
			    fs_areAdjacent(r->extents[i-1], r->extents[i],
			    		       r->offsets[i-1], r->offsets[i])) {
				r->offsets[i] = r->offsets[i-1];
				r->extents[i] = fs_mergeExtents(r->extents[i-1], r->extents[i]);
				fs_emap_flat_removeent(r, i-1);
			}
			fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
			return;
		}
	}

	/*
	 * A separate check for a merge to the left is performed when
	 * the first check for a merge to the right fails.
	 */
	if (i != 0) {
		fs_Extent extent_left = r->extents[i-1];
		uint64 offset_left = r->offsets[i-1];

		if (fs_areAdjacent(extent_left, extent, offset_left, offset)) {
			//printf("merging on the left\n");
			r->extents[i-1] = fs_mergeExtents(extent_left, extent);
			fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
			return;
		}
	}

	if (i < r->hdr.nkeys) {
		memmove(&r->offsets[i+1], &r->offsets[i], (r->hdr.nkeys - i) * sizeof *r->offsets);
		memmove(&r->extents[i+1], &r->extents[i], (r->hdr.nkeys - i) * sizeof *r->extents);
	}

	r->offsets[i] = offset;
	r->extents[i] = extent;
	r->hdr.nkeys++;

	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
}

uint64
fs_emap_flat_remove(fs_Context *context, fs_Inode *inode, uint64 offset)
{
	fs_Emapflatroot *r;
	uint64 nblocks;
	int i;

	r = &inode->dinode.u.emapflatroot;

	for (i = 0; i < r->hdr.nkeys; i++)
		if (offset == r->offsets[i])
			break;

	/*
	 * Make sure the given offset exists.
	 */
	assert(i < r->hdr.nkeys);

	nblocks = fs_extentLength(r->extents[i]);
	
	fs_emap_flat_removeent(r, i);

	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);

	return nblocks;
}

/*
 * Look up an extent from a given offset, return the null extent {0,0}
 * if it's not found.
 */
fs_Emappair
fs_emap_flat_lookup(fs_Context *context, fs_Emapflatroot *r, uint64 offset)
{
	fs_Emappair p;
	int i;

	p.offset = 0;
	p.extent = fs_makeExtent(0, fs_makeBlock(0));

	for (i = 0; i < r->hdr.nkeys; i++)
		if (offset >= r->offsets[i]
		 && offset < r->offsets[i] + fs_extentLength(r->extents[i]))
			break;

	if (i < r->hdr.nkeys) {
		p.offset = r->offsets[i];
		p.extent = r->extents[i];
	}

	return p;
}

/*
 * De-allocate all the extents in a flat root above a given page offset
 * to shrink an inode. Return number of blocks shrunk from flat root.
 */
uint64
fs_emap_flat_shrink(fs_Context *context, fs_Inode *inode, uint64 offset)
{
	fs_Emapflatroot *r = &inode->dinode.u.emapflatroot;
	int nkeys_remain;
	uint64 nblocks = 0;
	int i = 0;

	/*
	 * Find the first extent in which the offset falls.
	 */
	while (i < r->hdr.nkeys && offset > r->offsets[i])
		i++;

	nkeys_remain = i;

	/*
	 * Deallocate all extents >= i.
	 */
	while (i < r->hdr.nkeys) {
		nblocks += fs_extentLength(r->extents[i]);
		fs_deallocateExtent(context, r->extents[i]);
		i++;
	}

	/*
	 * Re-size the extents counter to the number of remaining
	 * extents.
	 */
	r->hdr.nkeys = nkeys_remain;
	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);

	/*
	 * If the offset falls in the middle of an extent, split
	 * the extent and deallocate the blocks greater than the
	 * offset.
	 */
	if (r->hdr.nkeys == 0)
		return nblocks;

	i = r->hdr.nkeys - 1;
	if (offset > r->offsets[i] &&
	    offset < r->offsets[i] + fs_extentLength(r->extents[i])) {
		fs_Extentpair p;

		p = fs_splitExtent(r->extents[i], offset - r->offsets[i]);
		r->extents[i] = p.left;
		fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
		nblocks += fs_extentLength(p.right);
		fs_deallocateExtent(context, p.right);
	}

	return nblocks;
}


/**
 * Do a forward search for the free tree in a flat inode root.  If a
 * match succeeds, remove the matching extent.
 */
fs_Extent
fs_emap_flat_search_forward(
	fs_Context *context,
	fs_Inode *inode,
	fs_Block start,
	fs_Block end,
	boolean (*match)(fs_Extent extent, void *match_arg),
	void *match_arg)
{
	fs_Emapflatroot *r = &inode->dinode.u.emapflatroot;
	fs_Extent extent;
	int i;

	bzero(&extent, sizeof extent);

	for (i = 0; i < r->hdr.nkeys; i++) {
                //printf("offset: %lld, length: %d, bno: %lld\n",
                //r->offsets[i],
                //fs_extentLength(r->extents[i]),
                //fs_blockNumber(fs_extentStartBlock(r->extents[i]))
                //);

		if (r->offsets[i] >= fs_blockNumber(start)
		 && r->offsets[i] <  fs_blockNumber(end)
		 && match(r->extents[i], match_arg)) {
			extent = r->extents[i];
			fs_emap_flat_remove(context, inode, r->offsets[i]);
			fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
			return extent;
		}
	}

	fs_releaseBtreeExclusive(context, inode);
	return extent;
}

/*
 * Return TRUE if eek of extents in emapflat is okay.
 */
boolean
fs_emapflat_check_extents(fs_Context *context, fs_Inode *inode, boolean walk)
{
	fs_Emapflatroot *r = &inode->dinode.u.emapflatroot;
	int i;

	/*
	 * Verify that each extent does not overlap with its
	 * neighbors.
	 */
	for (i = 0; i < r->hdr.nkeys - 1; i++) {
		if (r->offsets[i] + fs_extentLength(r->extents[i]) > r->offsets[i+1]) {
			fs_emaproot_dispose(context, inode, walk,
				            "extents %d and %d overlap %lld:(%d,%lld), %lld:(%d, %lld)",
				            i, i+1,
				            r->offsets[i],
				            fs_extentLength(r->extents[i]),
				            fs_blockNumber(fs_extentStartBlock(r->extents[i])),
				            r->offsets[i+1],
				            fs_extentLength(r->extents[i+1]),
				            fs_blockNumber(fs_extentStartBlock(r->extents[i+1])));
            return FALSE;
        }
	}

	/*
	 * Verify that each extent has a valid block number
	 */
	for (i = 0; i < r->hdr.nkeys; i++) {
		fs_Extent e = r->extents[i];
		if (fs_extentLength(e) + fs_blockNumber(fs_extentStartBlock(e)) >
		            context->fileSystem->params.totalBlocks) {
            fs_emaproot_dispose(context, inode, walk,
                               "extent %d has invalid blocknum: (%d,%lld)",
				               i,
				               fs_extentLength(e),
				               fs_blockNumber(fs_extentStartBlock(e)));
            return FALSE;
        }
	}

    return TRUE;
}

#endif // SSC
