/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * $Id: emaproot.c,v 1.79 2002/11/26 22:54:07 bob Exp $
 */

#if !defined(SSC)

#include "btree.h"

void
fs_emaproot_init(fs_Context *context, fs_Inode *inode)
{
	fs_Emaproot *r;

	//r = fs_getBtreeExclusive(context, inode);
 	fs_lockExclusive(context, &inode->btreeLock, NULL);
	r = &inode->dinode.u.emaproot;   
	r->hdr.nkeys = 0;
	r->hdr.level = 0;

	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
	fs_inodeWrite(context, inode, INO_WRITE_LOG);
}

boolean
fs_emap_is_empty(fs_Emaproot *r)
{
	return (r->hdr.nkeys == 0 && r->right == 0);
}

static fs_Buf *
get_child_buf(fs_Context *context, fs_Inode *inode, fs_Emaproot *r, uint64 offset, int *indexp, boolean get_exclusive, boolean do_cow)
{
	uint64 *bp;
	int i;

	/*
	 * Find a matching child to search.
	 */
	for (i = 0; i < r->hdr.nkeys; i++) {
		if (offset <= r->offsets[i])
			break;
	}

	if (indexp != NULL)
		*indexp = i;

	if (i < r->hdr.nkeys)
		bp = &r->left[i];
	else
		bp = &r->right;

	if (do_cow && fs_blockIsCopyOnWrite(context, fs_makeBlock(*bp))) {
		if (!fs_lockOwnedExclusive(context, &inode->btreeLock))
			return NULL;
		*bp = fs_make_block_copy(context, inode, *bp);
		fs_inodeWrite(context, inode,
			      INO_WRITE_LOG | INO_WRITE_BTREE);
	}

	if (get_exclusive)
		return B_GET_EXCL(inode, fs_makeBlock(*bp));
	else
		return B_GET_SHARED(inode, fs_makeBlock(*bp));
}

static void
fs_emaproot_split(fs_Context *context, fs_Inode *inode, fs_Emaproot *r)
{
	int pivot;
	fs_Buf *buf, *buf2;
	fs_Btree *t, *t2, *dest;
	int i, j;

	pivot = r->hdr.nkeys / 2;

	buf = B_CREATE(inode);
	buf2 = B_CREATE(inode);

	fs_bt_init(context, buf, r->hdr.level);
	fs_bt_init(context, buf2, r->hdr.level);

	t = buf->u.btree;
	t2 = buf2->u.btree;

	t2->right = r->left[pivot];
	t->right = r->right;

	TOUCH_ELEM(buf, &t->right);
	TOUCH_ELEM(buf2, &t2->right);

	dest = t2;
	j = 0;
	for (i = 0; i < r->hdr.nkeys; i++, j++) {
		if (i == pivot)
			continue;
		if (i > pivot && dest == t2) {
			dest = t;
			j = 0;
		}

		dest->keys[j] = r->offsets[i];
		dest->left[j] = r->left[i];
		dest->nkeys++;
	}

	TOUCH_ELEM(buf2, &t2->nkeys);
	TOUCH(buf2, &t2->keys, pivot * sizeof *t2->keys);
	TOUCH(buf2, &t2->left, pivot * sizeof *t2->left);
	TOUCH_ELEM(buf, &t->nkeys);
	TOUCH(buf, &t->keys, (r->hdr.nkeys - pivot - 1) * sizeof *t->keys);
	TOUCH(buf, &t->left, (r->hdr.nkeys - pivot - 1) * sizeof *t->left);

	r->offsets[0] = r->offsets[pivot];
	r->left[0] = fs_blockNumber(buf2->block);
	r->right = fs_blockNumber(buf->block);
	r->hdr.nkeys = 1;
	r->hdr.level++;

	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
}

void fs_emap_add_block(fs_Context *context, fs_Inode *inode, uint64 offset, fs_Block block)
{
        fs_emap_add(context, inode, offset, fs_makeExtent(1, block));
}

void
fs_emap_add(fs_Context *context, fs_Inode *inode, uint64 offset, fs_Extent extent)
{
	fs_Emaproot *r;
	boolean split_allowed;
	fs_Buf *child;
	fs_Bt_insert p;
	boolean locked_shared;
	boolean did_modify;
	int i;

    //printf("# %lld: emap_add(%lld, %lld+%d)\n",
    //       inode->inum, offset,
    //       fs_blockNumber(fs_extentBlock(extent, 0)),
    //       fs_extentLength(extent));
        
    r = fs_getBtreeShared(context, inode);
	locked_shared = TRUE;
	did_modify = FALSE;

	if (r->hdr.level == 0) {
		fs_releaseBtreeShared(context, inode);
 	    fs_lockExclusive(context, &inode->btreeLock, NULL);
	    r = &inode->dinode.u.emaproot;   
		locked_shared = FALSE;
		if (r->hdr.level == 0) {
			if (r->hdr.nkeys < FS_INOFLATROOT_MAXKEYS) {
				fs_emap_flat_add(context, inode, offset, extent);
				return;
			}
			fs_emaproot_spill(context, inode);
			did_modify = TRUE;
		}
	}

	split_allowed = FALSE;

	do {
		int child_level;

		/*
		 * The second time through the loop, reacquire the
		 * root exclusive so that a split can happen.
		 */
		if (split_allowed && locked_shared) {
 	        fs_lockExclusive(context, &inode->btreeLock, NULL);
	        r = &inode->dinode.u.emaproot;   
			locked_shared = FALSE;
		}
		
		/*
		 * Split the root top-down?
		 */
		if (split_allowed && r->hdr.nkeys == FS_INOROOT_MAXKEYS)
		{
			fs_emaproot_split(context, inode, r);
			did_modify = TRUE;
		}

		/*
		 * Get exclusive when we're about to split or when we're at
		 * level 1, i.e., about to visit a leaf (which will be modified
		 * regardless).
		 */
		child = get_child_buf(context, inode, r, offset, &i, split_allowed
				      || r->hdr.level == 1, TRUE);

		child_level = r->hdr.level - 1;

		if (!split_allowed && locked_shared) {
			fs_releaseBtreeShared(context, inode);
		}

	    // XXX: at this point the btree root is unlocked; could someone
	    // else sneak in inside this loop?

        if (child == NULL) {
            /*
             * Fake out loop termination condition: we want the block
             * copy for COW to happen in get_child_buf, but that can
             * only happen if the btree root is grabbed EXCL, which
             * occurs the second time through the loop.
             */
            p.status = FS_BT_NEEDS_SPLIT;
        } else {

            p = bt_funcs.emap.insert.func[child_level](context, inode, child,
						    child_level,
						    offset, &extent,
						    split_allowed,
						    &bt_funcs.emap.insert);
        }

		assert(!(split_allowed && p.status == FS_BT_NEEDS_SPLIT));
		split_allowed = TRUE;
	} while (p.status == FS_BT_NEEDS_SPLIT);

	/*
	 * Did a key get promoted from a split lower down?
	 */
	if (p.blknum != 0) {
		int j;
		for (j = 0; j < r->hdr.nkeys - 1; j++)
			assert(r->offsets[j] <= r->offsets[j + 1]);

		if (i < r->hdr.nkeys) {
			memmove(&r->left[i + 1], &r->left[i],
				(r->hdr.nkeys - i) * sizeof *r->left);
			memmove(&r->offsets[i + 1], &r->offsets[i],
				(r->hdr.nkeys - i) * sizeof *r->offsets);
		}
		r->offsets[i] = p.key;
		r->left[i] = p.blknum;
		r->hdr.nkeys++;

        fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
	} else {
		if (!locked_shared && !did_modify && !fs_inodeBtreeWasModified(context, inode))
			fs_releaseBtreeExclusive(context, inode);
	}
}

fs_Emappair
fs_emap_lookup(fs_Context *context, fs_Inode *inode, uint64 offset)
{
	fs_Emaproot *r;
	fs_Emapflatroot *fr;
	fs_Buf *child;
	fs_Emappair p;
	int child_level;

	r = fs_getBtreeShared(context, inode);
	fr = (fs_Emapflatroot *)r;

	if (r->hdr.level == 0) {
		p = fs_emap_flat_lookup(context, fr, offset);
		fs_releaseBtreeShared(context, inode);
		return p;
	}

	child = get_child_buf(context, inode, r, offset, NULL, FALSE, FALSE);
	child_level = r->hdr.level - 1;

	fs_releaseBtreeShared(context, inode);

    p.offset = 0;
    p.extent = fs_makeExtent(0, fs_makeBlock(0));

	bt_funcs.emap.search.func[child_level](
                        context,
                        inode,
                        child,
                        child_level,
                        offset,
                        &p,
                        &bt_funcs.emap.search,
                        FALSE);

	return p;
}


/**
 * Look up a single block in an extent map.  Returns FS_INVALID_BLOCK if
 * the block is not mapped.
 */
fs_Block fs_emap_lookup_block(
    fs_Context *context,
    fs_Inode *inode,
    uint64 offset)
{
    fs_Emappair pair = fs_emap_lookup(context, inode, offset);
    fs_Block block;

    if (fs_extentLength(pair.extent) == 0)
        block = FS_INVALID_BLOCK;
    else {
        block = fs_extentBlock(pair.extent, offset - pair.offset);
		if (fs_blockNumber(block) >= context->fileSystem->super->numBlks)
		{
			fs_volumeException(context, context->fileSystem,
                               FS_VOLUME_EXCEPTION);
		}
	}

    //printf("# emap_lookup_block(inode %lld, block %lld) ="
    //       " (%lld, %lld+%d) = %lld\n",
    //       inode->inum, offset, pair.offset,
    //       fs_blockNumber(fs_extentStartBlock(pair.extent)),
    //       fs_extentLength(pair.extent),
    //       fs_blockNumber(block));

    return block;
}


void
fs_emaproot_remove_child(fs_Context *context, fs_Inode *inode, int i)
{
	fs_Emaproot *r = &inode->dinode.u.emaproot;
	int nkeys = r->hdr.nkeys;

	assert(i <= nkeys);

	/*
	 * if (i < nkeys), remove one of the left pointers; shift over the
	 * ones to its right by one so the sort is maintained.
	 *
	 * If we're freeing the right pointer (i == nkeys), then grab the
	 * rightmost left pointer to become the new right pointer (see the
	 * design document for a discussion of B-tree invariants.)
	 *
	 * If the right pointer is the last one left, zero it and set
	 * the B-tree level to 0. This is the only time that the height
	 * of a B-tree decreases.
	 */
	if (i == nkeys - 1) {
		--r->hdr.nkeys;
	} else if (i < nkeys) {
		memmove(&r->left[i], &r->left[i+1], (nkeys-i) * sizeof *r->left);
		memmove(&r->offsets[i], &r->offsets[i+1], (nkeys-i) * sizeof *r->offsets);
		--r->hdr.nkeys;
	} else if (nkeys > 0) {
		r->right = r->left[nkeys-1];
		--r->hdr.nkeys;
	} else {
		assert(r->right != 0);
		r->right = 0;
		r->hdr.level = 0;
	}

	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
}

void
fs_emap_remove(fs_Context *context, fs_Inode *inode, uint64 offset)
{
	fs_Emaproot *r;
	fs_Buf *child;
	char dummy;
	boolean dispose_allowed;
    boolean held_excl = FALSE;
	fs_Bt_remove rs;
	int i;

	r = fs_getBtreeShared(context, inode);

	if (r->hdr.level == 0) {
		uint64 nblocks;

		fs_releaseBtreeShared(context, inode);
		//r = fs_getBtreeExclusive(context, inode);
        fs_lockExclusive(context, &inode->btreeLock, NULL);
        r = &inode->dinode.u.emaproot;   
		if (r->hdr.level == 0) {
			nblocks = fs_emap_flat_remove(context, inode, offset);
			return;
		}
		fs_demoteBtreeLock(context, inode);
	}

    dispose_allowed = FALSE;
	do {
		int child_level;

		/*
		 * On the second time through the loop reacquire
		 * the root exclusive so that a dispose operation
		 * can be performed further down (and possibly
		 * alter the root).
		 */
		if (dispose_allowed) {
			//r = fs_getBtreeExclusive(context, inode);
            fs_lockExclusive(context, &inode->btreeLock, NULL);
            r = &inode->dinode.u.emaproot;   
            held_excl = TRUE;
		}

		/*
		 * Get exclusive when we're about to split or when we're
		 * at level 1, i.e., about to visit a leaf (which will be
		 * modified regardless).
		 */
		child = get_child_buf(context, inode, r, offset, &i,
                                dispose_allowed || r->hdr.level == 1, TRUE);
		child_level = r->hdr.level - 1;

		if (!dispose_allowed) {
			fs_releaseBtreeShared(context, inode);
		}
	
        if (child == NULL) {
            /*
             * Fake out loop termination condition: we want the block
             * copy for COW to happen in get_child_buf, but that can
             * only happen if the btree root is grabbed EXCL, which
             * occurs the second time through the loop.
             */
            rs.status = FS_BT_REMOVE_NEEDS_DISPOSAL;
        }
        else {
            rs = bt_funcs.emap.remove.func[child_level](context, inode, child,
                    child_level, offset, &dummy, dispose_allowed, &bt_funcs.emap.remove);
        }

		assert(!(dispose_allowed && rs.status == FS_BT_REMOVE_NEEDS_DISPOSAL));

		if (rs.status == FS_BT_REMOVE_OK_EMPTY)
			fs_emaproot_remove_child(context, inode, i);

		dispose_allowed = TRUE;
	} while (rs.status == FS_BT_REMOVE_NEEDS_DISPOSAL);

    if (held_excl && !fs_inodeBtreeWasModified(context, inode))
        fs_releaseBtreeExclusive(context, inode);
}

/*
 * Shrink the B-tree by one leaf page. Returns the number of blocks removed
 * on this pass (0 when the file is finally empty).
 */
uint64
fs_emap_shrink(fs_Context *context, fs_Inode *inode)
{

	fs_Emaproot *r;
	fs_Bt_remove rs;
	fs_Buf *child;
	int child_level;
	int i;

	/*
	 * XXX: only used by the trashcan, so we don't try to increase
	 * concurrency with shared acccess for the moment.
	 */
	//r = fs_getBtreeExclusive(context, inode);
 	fs_lockExclusive(context, &inode->btreeLock, NULL);
	r = &inode->dinode.u.emaproot;   

	if (r->hdr.level == 0)
		return fs_emap_flat_shrink(context, inode, 0);

	child = get_child_buf(context, inode, r, 0, &i, TRUE, TRUE);
	assert(child != NULL);
	child_level = r->hdr.level - 1;

	rs = bt_funcs.emap.shrink.func[child_level](context, inode, child, child_level, 0, NULL, TRUE, &bt_funcs.emap.shrink);

	if (rs.status == FS_BT_REMOVE_OK_EMPTY) {
		fs_emaproot_remove_child(context, inode, i);
	} else if (!fs_inodeBtreeWasModified(context, inode)) {
		fs_releaseBtreeExclusive(context, inode);
	}

	return rs.nblocks;
}

fs_Inode *
fs_emap_cleave(fs_Context *context, fs_Inode *trashcan_inode, fs_Inode *inode, uint64 byte_offset)
{

	fs_Emaproot *r, *tr;
	fs_Inode *trash;
	fs_Buf *child;
	int child_level;
	char dummy;
	fs_Bufpair cleavepair;
	uint64 offset;
	int i, j;

	offset = ceilingDiv64(byte_offset, PAGE_SIZE);

	//printf("cleave: %lld\n", offset);

	//r = fs_getBtreeExclusive(context, inode);
 	fs_lockExclusive(context, &inode->btreeLock, NULL);
	r = &inode->dinode.u.emaproot;   

	if (r->hdr.level == 0) {
		fs_emap_flat_shrink(context, inode, offset);
		return NULL;
	}

	child = get_child_buf(context, inode, r, offset, &i, TRUE, TRUE);
	assert(child != NULL);
	child_level = r->hdr.level - 1;
	cleavepair = bt_funcs.emap.cleave.func[child_level](context, inode, child, child_level, offset, &dummy, &bt_funcs.emap.cleave);

	/*
	 * Return NULL if the trash is a trunc of a hole to the right of the
	 * last extent.
	 */
	if (cleavepair.right == 0 && i == r->hdr.nkeys) {
		if (!fs_inodeBtreeWasModified(context, inode))
			fs_releaseBtreeExclusive(context, inode);
		return NULL;
	}

	/*
	 * Create the trash inode, copy over any extents from the original
	 * inode.
	 */
	trash = fs_allocateFileInode(context, trashcan_inode, 0700, 0, 0, 0);
	tr = fs_getBtreeExclusive(context, trash);
 	fs_lockExclusive(context, &trash->btreeLock, NULL);
	r = &inode->dinode.u.emaproot;   
	fs_emaproot_init(context, trash);
	tr->hdr.level = r->hdr.level;
	tr->hdr.nkeys = r->hdr.nkeys - i;
	for (j = i; j < r->hdr.nkeys; j++) {
		tr->left[j - i]    = r->left[j];
		tr->offsets[j - i] = r->offsets[j];
	}
	if (j - i > 0) {
	}

	tr->right = r->right;

	/*
	 * Remove extents from the original inode to the right of the
	 * offset. XXX: this happens iteratively but in fact it should
	 * be just one assignment.
	 */
	while (r->hdr.nkeys > i)
		fs_emaproot_remove_child(context, inode, r->hdr.nkeys);

	/*	
	 * Now assign the new blocks from the cleavepair at the insertion
	 * points r->hdr.nkeys on the left and zero on the right.
	 */
	if (cleavepair.left != 0) {
		r->right = cleavepair.left;
		fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
	} else {
		fs_emaproot_remove_child(context, inode, r->hdr.nkeys);
	}

	if (cleavepair.right != 0) {
		if (tr->hdr.nkeys == 0) {
			tr->right = cleavepair.right;
		} else {
			tr->offsets[0] = offset;
			tr->left[0] = cleavepair.right;
		}
	} else {
		fs_emaproot_remove_child(context, trash, 0);
	}

	fs_inodeWrite(context, trash, INO_WRITE_LOG|INO_WRITE_BTREE);
	fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);

	return trash;
}

void
fs_emaproot_dispose(fs_Context *context, fs_Inode *inode, boolean fixIt, char *fmt, ...)
{
	va_list ap;
	char msg[256];

	va_start(ap, fmt);
	vsprintf(msg, fmt, ap);
	va_end(ap);

    context->eek.errcount++;
	fs_eek_print_status(context, FALSE, "%s: (I#%lld) CORRUPTED emaproot: %s",
                        context->fileSystem->super->mountName,
                        inode->inum, msg);
    /* 
     * modifications of btree structures only allowed when the
     * the buddy has been reconciled with the refrence byte
     * during walk, the inode has been been acquired exclusive
     */
    if (fixIt) {
	    fs_emaproot_init(context, inode);
        context->eek.fixcount++;
    }
}

/*
 * Return TRUE if eek of emap hdr is okay.
 * if walk, no modifications are done to correct corrupted emap blocks
 */
static boolean
fs_emap_eek_hdr(fs_Context *context, fs_Inode *inode, boolean fixIt)
{
    fs_Emapflatroot *r = &inode->dinode.u.emapflatroot;
    uint32 maxkeys;
  
    maxkeys = (r->hdr.level == 0) ? FS_INOFLATROOT_MAXKEYS :
            FS_INOROOT_MAXKEYS;
  
    if ((unsigned)r->hdr.nkeys > maxkeys) {
            fs_emaproot_dispose(context, inode, fixIt, "nkeys %lld > %d",
                                    (unsigned)r->hdr.nkeys, maxkeys);
        return FALSE;
    }
	if ((unsigned)r->hdr.level > FS_INOROOT_MAXLEVEL) {
		fs_emaproot_dispose(context, inode, fixIt, "level %lld > %d",
			                (unsigned)r->hdr.nkeys, FS_INOROOT_MAXLEVEL);
        return FALSE;
	}
	return TRUE;
}

/*
 * Return TRUE if eek of flat extent is okay
 * and bitmap is marked with blocks in the extent
 */
boolean
fs_emap_flat_mark_used_blocks(fs_Context *context, fs_Inode *inode, uint8 *bitmap, boolean fixIt)
{
	fs_Emapflatroot *r = &inode->dinode.u.emapflatroot;
	int i;

    if (!fs_emapflat_check_extents(context, inode, FALSE)) {
        return FALSE;
    }

	for (i = 0; i < r->hdr.nkeys; i++) {
		fs_Extent extent = r->extents[i];
		int nblocks = fs_extentLength(extent);
		int j;

		for (j = 0; j < nblocks; j++) {
			fs_Block block = fs_extentBlock(extent, j);
			uint64 blocknum = fs_blockNumber(block);

			if (!fs_eek_mark(context, "(I#%lld) emapflatroot block number", bitmap, blocknum, inode->inum)) {
                if (fixIt) fs_emaproot_init(context, inode);
                context->eek.fixcount++;
                return FALSE;
			}
		}
	}

    return TRUE;
}

/*
 * Mark in eek the blocks used by a file.
 *
 * The "type" argument is used to guide the marking: if EEK_MARK_EXTENTS is
 * set, then mark the extents which the extent map allocates. If EEK_MARK_NODES
 * is set, then mark the inner nodes which the extent map itself owns.
 * 
 * This type field is used to provide special marking for the free tree inode.
 * In general, one wants to map both the extents and the inner nodes of any
 * file. However, the free tree is special: we want to mark its inner nodes
 * as blocks used, but the extents which it maps as free blocks.
 *
 * Return TRUE if inode ok
 * FALSE otherwise
 */
boolean
fs_emap_fix_used_blocks(fs_Context *context, uint8 *bitmap, fs_Inode *inode, fs_Eekmark type)
{
	fs_Emaproot *r = fs_getBtreeExclusive(context, inode);
	fs_Bt_eek eek;
	int nkeys, level;
	uint64 offset;
	fs_Bt_eeksearch rc;
	int i;
    boolean ok = TRUE;

	if (!fs_emap_eek_hdr(context, inode, TRUE)) {
        ok = FALSE;
        goto fs_emap_mark_used_blocks_dispose;
	}

	if (r->hdr.level == 0) {
		if ((type & FS_EEK_MARK_EXTENTS) &&
		    (!fs_emap_flat_mark_used_blocks(context, inode, bitmap, TRUE))) {
            ok = FALSE;
        }
        goto fs_emap_mark_used_blocks_dispose;
	}

	offset = 0;

	nkeys = r->hdr.nkeys;
	level = r->hdr.level;

	eek.bitmap = bitmap;
	eek.marktype = type;
	memset(&eek.last_blk[0], 0xfe, sizeof eek.last_blk); // init to garbage
    eek.fixIt = TRUE;

	for (i = 0; i <= nkeys; i++) {
		fs_Buf *child;
		int child_level;
		uint64 *bp;

		if (i < nkeys)
			bp = &r->left[i];
		else
			bp = &r->right;

		if (type & FS_EEK_MARK_NODES) {
			if (!fs_blockIsValid(fs_makeBlock(*bp)) ||
			    !fs_eek_mark(context, "(I#%lld) emaproot block number", bitmap, *bp, inode->inum)) {
                /*
                 * get rid of all the bad blocks so shrink will work
                 */
			    fs_emaproot_remove_child(context, inode, i);
			    --i;
			    --nkeys;
                ok = FALSE;
                context->eek.fixcount++;
                continue;
			}
		}

		do {
			child = B_GET_EXCL(inode, fs_makeBlock(*bp));
			child_level = level - 1;
			rc = bt_funcs.emap.eek_blocks.func[child_level](
					       context,
					       inode,
					       child,
					       child_level,
					       offset,
					       &eek,
					       &bt_funcs.emap.eek_blocks);
			offset = rc.key;
		} while ((offset != 0) && (rc.status == FS_BT_EEK_OK));

        if (rc.status != FS_BT_EEK_OK) {
            ok = FALSE;
            if (rc.status == FS_BT_EEK_DISPOSE) {
			    fs_emaproot_remove_child(context, inode, i);
			    --i;
			    --nkeys;
            }
            /*
             * COW happened lower down which means we'll need
             * to update the new child for this key
             */
            if (rc.status == FS_BT_EEK_COW) {
                *bp = rc.cowBlock;
                fs_inodeWrite(context, inode, INO_WRITE_LOG|INO_WRITE_BTREE);
            }

            /* continue to eek out the rest of the bad ones */
        }
		if (i < nkeys)
		    offset = r->offsets[i] + 1;
	}

fs_emap_mark_used_blocks_dispose:

	if (!fs_inodeBtreeWasModified(context, inode))
		fs_releaseBtreeExclusive(context, inode);
    
    return ok;
}

void
fs_emap_flat_walk(
	fs_Context *context,
	fs_Inode *inode,
	boolean (*worker)(fs_Context *context, fs_Inode *inode, fs_WalkResult *result, void *arg0),
	void *arg0)
{
	fs_Emapflatroot *r = &inode->dinode.u.emapflatroot;
	int i;

    if (!fs_emapflat_check_extents(context, inode, TRUE)) {
        return;
    }

	for (i = 0; i < r->hdr.nkeys; i++) {
		fs_Extent e = r->extents[i];
		fs_emap_walk_extent(context, inode, e, r->offsets[i], worker, arg0);
	}
}

/*
 * Walk an extent map, calling the worker function with the
 * block number of every extent in the extent map, plus the
 * blocks of the btree index itself. Used during snap creation
 * time to clear the snapbits file from the live bitplane.
 */
void
fs_emap_walk(
	fs_Context *context,
	fs_Inode *inode,
	boolean (*worker)(fs_Context *context, fs_Inode *inode, fs_WalkResult *result, void *arg0),
	void *arg0)
{

	fs_Emaproot *r = &inode->dinode.u.emaproot;
	fs_Bt_walk walk;
	int nkeys, level;
	uint64 offset;
	int i;

	/* 
	 * eek uses walk to mark blocks
	 * since eek does not assert for inconsistencies, do some basic
	 * sanity checks of emaproot and blocks before walking
	 */

	if (!fs_emap_eek_hdr(context, inode, FALSE))
		return;

	if (r->hdr.level == 0) {
		fs_emap_flat_walk(context, inode, worker, arg0);
		return;
	}

	offset = 0;

	nkeys = r->hdr.nkeys;
	level = r->hdr.level;

	walk.worker = worker;
	walk.arg = arg0;
	memset(&walk.last_blk[0], 0xfe, sizeof walk.last_blk);	// init to garbage

	for (i = 0; i <= nkeys; i++) {
		fs_Buf *child;
		fs_Block child_block;
		int child_level;
		uint64 *bp;
		boolean first_time = TRUE;

		if (i < nkeys)
			bp = &r->left[i];
		else
			bp = &r->right;

		if (!fs_blockIsValid(fs_makeBlock(*bp)) ||
		    (*bp >= context->fileSystem->params.totalBlocks)) {
			fs_eek_print_status(context, FALSE,
					    "%s: (I#%lld) invalid emap block %lld",
					    context->fileSystem->super->
					    mountName, inode->inum, *bp);
			continue;
		}

		do {
			child_block = fs_makeBlock(*bp);
			child_level = level - 1;
			if (first_time) {	// walk this child only once.
				fs_WalkResult result;
				result.type = (child_level == 0) ?
				    FS_WALK_LEAF_NODE_TYPE :
				    FS_WALK_INNER_NODE_TYPE;
				result.block = child_block;
				if (!worker(context, inode, &result, arg0))
					break;
				first_time = FALSE;
			}
			child = B_GET_EXCL(inode, child_block);
			offset =
			    bt_funcs.emap.walk.func[child_level] (context,
								  inode,
								  child,
								  child_level,
								  offset,
								  &walk,
								  &bt_funcs.emap.walk,
								  TRUE);
		} while (offset != 0);

		if (i < nkeys)
			offset = r->offsets[i] + 1;
	}
}

/*
 * Emap remove just offset
 */
void
fs_emap_remove_block(fs_Context *context, fs_Inode *inode, uint64 page)
{
    fs_Emappair pp;
    fs_Extent left, right;
    uint64 start, p;
    uint32 offset, length;


    FS_TRACE(context, FS_TRACE_SNAP,
             ("EMAP before for ino %lld page %lld",
             inode->inum, page));

    pp = fs_emap_lookup(context, inode, page);
    length = fs_extentLength(pp.extent);
    if (length == 0) {
        FS_TRACE(context, FS_TRACE_SNAP,
                ("  COW ino %lld page %lld not found in extent",
                inode->inum, page));
        return;
    }

    // remove entire extent this page belongs to
    // ick, only can remove if this page is offset 0 in the extent ....
    p = page;
    if (page != pp.offset)
        p = pp.offset;
    fs_emap_remove(context, inode, p);

    // there's more to the extent than this page so
    // we'll add back the left and/or right portion
    if (length != 1) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("  COW ino %lld page %lld part of greater extent %d",
                 inode->inum, page, fs_extentLength(pp.extent)));
        start = fs_blockNumber(fs_extentBlock(pp.extent, 0));
    	if (start >= context->fileSystem->super->numBlks)
    	{
        	fs_volumeException(context, context->fileSystem,
                               FS_VOLUME_EXCEPTION);
    	}

        offset = page - pp.offset;
        if (offset > 0) {
            left = fs_makeExtent(offset, fs_makeBlock(start));
            fs_emap_add(context, inode, pp.offset, left);
        }
        if ((offset+1) < length) {
            right = fs_makeExtent(length-offset-1,
                                  fs_makeBlock(start+offset+1));
            fs_emap_add(context, inode, page+1, right);
        }
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("EMAP after for ino %lld", inode->inum));
}

/*
 * "atomic" emap remove and add for COW
 * needed to avoid lock order problems (PR #1582)
 */
void
fs_emap_replace_block(fs_Context *context, fs_Inode *inode, uint64 page, fs_Block block)
{
 	fs_lockExclusive(context, &inode->btreeLock, NULL);
    fs_emap_remove_block(context, inode, page);
    fs_emap_add_block(context, inode, page, block);
    if (!fs_inodeBtreeWasModified(context, inode)) {
        fs_unlockExclusive(context, &inode->btreeLock);
    }
}

#endif // SSC
