/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * $Id: main.c,v 1.20 2002/04/17 19:12:35 byron Exp $
 */

#include "btree.h"
#include <stdlib.h>

/*
 * Reads stdin for the following commands, to exercise btree code:
 *
 * Afilename		# create a filename in a single directory
 * Lfilename		# look up the filename
 * Rfilename		# remove the filename
 * Dcookie		# readdir with cookie N
 *
 * aOFFSET LEN BNO	# create an extent in a single file's inode
 * lOFFSET		# look for an extent at the given OFFSET
 * rOFFSET		# remove an extent at the given OFFSET
 * cOFFSET		# cleave an inode at the given OFFSET
 * tOFFSET		# look for an extent in the trash file at the given OFFSET
 *
 * IOFFSET LEN		# add the specified blocknumber, length to free tree
 * OOFFSET		# do a freetree lookup at the given blocknumber
 * TOFFSET		# do a freetree take from the given blocknumber
 *
 * d			# dump an extent map
 * z			# zero all extent maps
 */

static void emap_dump(fs_Context *context, fs_Inode *inode);

int
main(int argc, char *argv[])
{
	extern int bt_splits, dt_splits, dr_splits;
	fs_Context c;
	fs_Context *context = &c;
	char buf[8192];
	fs_Inode dir_inode, file_inode, trashcan_inode, freetree_inode, *trash = NULL;
	int inums = 10;
	int line = 0;
	uint64 offset;
	fs_Emappair p;
	uint64 bno;
	int length;
	fs_DirEntry *entries, *e;
	int entries_size = FS_READDIR_SPACE_MAX;
	fs_Extent extent;
	int32 num_entries;
	int32 eof;
	int cookie;
	int i;

	dir_inode.dinode = malloc(sizeof *dir_inode.dinode);
	file_inode.dinode = malloc(sizeof *file_inode.dinode);
	trashcan_inode.dinode = malloc(sizeof *file_inode.dinode);
	freetree_inode.dinode = malloc(sizeof *file_inode.dinode);
	entries = malloc(entries_size);

	fs_bt_dispatch_init();

	fs_dirroot_init(context, &dir_inode, 0x00bacafefeedfaceULL);
	fs_dirroot_init(context, &trashcan_inode, 0x00bacafefeedfaceULL);
	fs_emaproot_init(context, &file_inode);
	fs_freetree_init(context, &freetree_inode);
	
	while (gets(buf)) {
		char *b = buf;
		uint32 inum;
		static int x;
		int r;

		line++;

		if (0) {
			char *a = "aah";
			assert(fs_dir_lookupent(context, &dir_inode, a, strlen(a)) > 0);
		}

		if (0 && strcmp(buf, "Ranesthesiologists") == 0) {
			fprintf(stderr, "bt_splits: %d, dt_splits: %d, dr_splits: %d\n", bt_splits, dt_splits, dr_splits);
			exit(0);
			assert(b != NULL);
		}

		switch (*b++) {
		case '\0':
		case '#':
			break;
		case 'A':
			if (strcmp(b, "aah") == 0)
				x++;
			r = fs_dir_addent(context, &dir_inode, b, strlen(b), inums++);
			if (r != FS_BT_INSERT_OK) {
				fprintf(stderr, "A r=%d\n", r);
			}
			assert(r == FS_BT_INSERT_OK);
			break;
		case 'L':
			inum = (uint32)fs_dir_lookupent(context, &dir_inode, b, strlen(b));
			assert(inum > 0);
			printf("%s: %d\n", b, inum);
			break;
		case 'R': 
			r = fs_dir_removeent(context, &dir_inode, b, strlen(b));
			assert(r == FS_BT_REMOVE_OK || r == FS_BT_REMOVE_OK_EMPTY);
			break;
		case 'D':
			cookie = strtoul(b, 0, 0);
			fs_dir_readdir(context, &dir_inode, entries, entries_size, cookie, &num_entries, &eof);
			for (i = 0, e = entries; i < num_entries; i++) {
				printf("%s: 0x%016llx, %lld\n", e->name, e->inode, e->cookie);
				FS_READDIR_ADV_ENTRY_PTR(e);
			}
			break;

		case 'a':
			sscanf(b, "%lld %d %lld", &offset, &length, &bno);
			fs_emap_add(context, &file_inode, offset, fs_makeExtent(length,
					fs_makeBlock(bno)));
			break;
		case 'l':
			offset = strtoull(b, 0, 0);
			p = fs_emap_lookup(context, &file_inode, offset);
			printf("offset: %lld, length: %d, bno: %lld\n",
					p.offset,
					fs_extentLength(p.extent),
					fs_blockNumber(fs_extentStartBlock(p.extent))
			);
			break;
		case 't':
			offset = strtoull(b, 0, 0);
			p = fs_emap_lookup(context, trash, offset);
			printf("offset: %lld, length: %d, bno: %lld\n",
					p.offset,
					fs_extentLength(p.extent),
					fs_blockNumber(fs_extentStartBlock(p.extent))
			);
			break;
		case 'r':
			offset = strtoull(b, 0, 0);
			fs_emap_remove(context, &file_inode, offset);
			break;
		case 'c':
			offset = strtoull(b, 0, 0);
			trash = fs_emap_cleave(context, &trashcan_inode, &file_inode, offset);
			break;
		case 'I':
			sscanf(b, "%lld %d", &offset, &length);
			printf("FT add: offset: %lld, length: %d\n", offset, length);
			fs_freetree_add(context, &freetree_inode, fs_makeExtent(length, fs_makeBlock(offset)));
			break;
		case 'O':
			offset = strtoull(b, 0, 0);
			extent = fs_freetree_lookup(context, &freetree_inode, fs_makeBlock(offset));
			printf("FT lookup: offset: %lld, length: %d\n",
					fs_blockNumber(fs_extentStartBlock(extent)),
					fs_extentLength(extent)
			);
			break;
		case 'T':
			offset = strtoull(b, 0, 0);
			extent = fs_freetree_take(context, &freetree_inode, fs_makeBlock(offset));
			printf("FT take %lld: start: %lld, length: %d\n", offset,
					fs_blockNumber(fs_extentStartBlock(extent)),
					fs_extentLength(extent)
			);
			break;
		case 'd':
			printf("file inode:\n");
			emap_dump(context, &file_inode);
			if (trash != NULL) {
				printf("trash inode:\n");
				emap_dump(context, trash);
			}
			break;
        case 'z':
            while (fs_emap_shrink(context, &file_inode) != 0)
                ;
            if (trash != NULL)
                while (fs_emap_shrink(context, trash) != 0)
                    ;
            break;
		default:
			printf("unknown command: %c\n", b[-1]);
		}

	}

	return 0;
}

boolean match(fs_Extent extent, void *match_arg)
{
	return FALSE;
}

static void emap_dump(fs_Context *context, fs_Inode *inode)
{
	fs_Emaproot *r = &inode->dinode->u.emaproot;
	uint64 key = 0;
	int nkeys = r->hdr.nkeys;
	int level = r->hdr.level;
	fs_Freetree_CB cb;
	fs_Buf *child;
	uint64 blknum;
	int i;

	inode->inum = FS_FREE_TREE_INODE; // HACK: the freetree search code
					  // verifies this

	cb.match = match;
	cb.match_arg = NULL;
	cb.end = fs_makeBlock(0x7fffffff);

	if (r->hdr.level == 0) {
		fs_emap_flat_search_forward(context, inode, cb.start, cb.end, match, NULL);
		return;
	}

	for (i = 0; i <= nkeys; i++) {
		if (i < nkeys)
			blknum = r->left[i];
		else
			blknum = r->right;

		do {
			child = B_GET_EXCL(inode, fs_makeBlock(blknum));
			key = bt_funcs.emap.read.func[level-1](context, inode, child, level-1, key, &cb, &bt_funcs.emap.read, TRUE);
		} while (key != 0);

		if (i < nkeys)
			key = r->offsets[i] + 1;
	}
}
