// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-alloc.c
 *
 * RCS:  $Id: fs-alloc.c,v 1.222 2002/12/10 18:42:46 bob Exp $
 *
 * Description:  Block and inode allocation.
 *
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-06-22
 */


#if !defined(SSC)

#include "fs-includes.h"

#define strbool(e)      ((e) ? "TRUE" : "FALSE")


#define SPACE_TRACE(args)   FS_TRACE(context, FS_TRACE_SPACE, args)
#define LOG_TRACE(args)     FS_TRACE(context, FS_TRACE_LOG, args)

uint32 fs_inodeAllocs;
uint32 fs_inodeDeallocs;
uint32 fs_inodeAllocLoopMax;


/**
 * Return a sub-extent of the specified extent, starting at the specified
 * start block (counting from 0 at the start of the extent) and continuing
 * through the block before the specified end block.
 */
static inline fs_Extent fs_makeSubExtent(
    fs_Extent extent,
    int start,
    int end)
{
    assert(0 <= start);
    assert(start < end);
    assert(end <= fs_extentLength(extent));

    return fs_makeExtent(end - start, fs_extentBlock(extent, start));
}


/**
 * In-core shadow bitmaps of block/inode allocation.
 */
#if defined(FS_DEBUG_ALLOC)

/**
 * Yield the CPU after doing this many of some things.
 */
#define FS_SHAWDOW_ALLOC_YIELD_CNT 500

void
fs_initShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    fs_Parameters *params = &fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint64 refBytePages =  params->refbytesBlockCount;
    uint64 inodeFilePages
        = ceilingDiv64(params->maxInodeNumber, FS_INODES_PER_PAGE);
    uint64 block = 0, page;
    uint64 maxBlock = params->totalBlocks;
    fs_Inumber base;
#if !defined(FS_STANDALONE)
    eeeGMT_t date;
#endif

    fs_Inode *inodeRef = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Inode *inodeFile = I_GET_SHARED(0, FS_INODE_INODE);

    assert(inodeRef != NULL);
    assert(inodeFile != NULL);

    // realloc shadow block and inode bufs with real fs parameters
    // since fs_mountInitFs alloc'd a smaller one to jump start

    fs_reallocShadow(context, fileSystem);

    eee_TimeGMT(&date);
    printf("Init shadow allocator start: %s max block %lld max inode %lld\n",
            eee_humandate(date.sec), maxBlock, params->maxInodeNumber);

    // read refbytes, mark used blocks in the shadow block bitmap

    for (page = 0; page < refBytePages; page++, block += blockSize)
    {
        int i;
        fs_Buf *buf = O_GET_SHARED(inodeRef, page);
        fs_RefByte *ref;
        
        assert(buf != NULL);
        ref = buf->u.refBytes;

        for (i = 0; i < blockSize; i++)
        {
            fs_RefByte byte = ref[i];

#if !defined(FS_STANDALONE)
            // yield the CPU to prevent watchdog timeouts
            if (i % FS_SHAWDOW_ALLOC_YIELD_CNT == FS_SHAWDOW_ALLOC_YIELD_CNT - 1)
                fs_sleep(context, 10);
#endif

            if ((block + i) >= maxBlock)
                break;

            if (byte != 0)
            {
                FS_DBG_SHADOW_ALLOC_BLOCK(block+i);
                // printf("init shadow: block %lld set\n", block+i);
            }
        }
        fs_bufRelease(context, buf);
    }
    fs_inodeRelease(context, inodeRef);

    // read inode file, mark used in the shadow inode bitmap

    for (page = 1; page < inodeFilePages; page++) 
    {
        fs_Buf *buf = O_GET_SHARED(inodeFile, page);
        int i;

        if (buf == NULL)
            continue; // hole in the file?
        
        base = buf->page * params->inodesPerBlock;

        for (i = 0; i < FS_INODES_PER_PAGE; i++)
        {
            fs_Dinode *dinode = &buf->u.dinode->dinode[i];

            if (dinode->attrs != 0)
            {
                FS_DBG_SHADOW_ALLOC_INODE(base+i);
                // printf("init shadow: inum %lld set\n", base+i);
            }
        }
        fs_bufRelease(context, buf);
    }
    fs_inodeRelease(context, inodeFile);

    eee_TimeGMT(&date);
    printf("Init shadow allocator done: %s\n", eee_humandate(date.sec));

}

void
fs_initAllocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    fileSystem->blockAllocShadowSize = (fileSystem->params.totalBlocks / 8) + 1;
    fileSystem->blockAllocShadow = fs_ramAlloc(FS_RAM_CORE_MISC, fileSystem->blockAllocShadowSize);
    bzero(fileSystem->blockAllocShadow, fileSystem->blockAllocShadowSize);

    // TODO - read refbytes, mark used blocks in our bitmap

    fileSystem->inodeAllocShadowSize = (fileSystem->params.maxInodeNumber / 8) + 1;
    fileSystem->inodeAllocShadow = fs_ramAlloc(FS_RAM_CORE_MISC, fileSystem->inodeAllocShadowSize);
    bzero(fileSystem->inodeAllocShadow,  fileSystem->inodeAllocShadowSize);

    // TODO - read inode file, mark used inodes in our bitmap

} /* fs_initAllocShadow */

void
fs_reallocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    void    *oldInodeShadow = fileSystem->inodeAllocShadow;
    void    *oldBlockShadow = fileSystem->blockAllocShadow;

    // Resize shadow info for inodes.
    if (fileSystem->inodeAllocShadowSize < ((fileSystem->params.maxInodeNumber / 8) + 1))
    {
        fileSystem->inodeAllocShadow = fs_ramAlloc(FS_RAM_CORE_MISC,
                                        (fileSystem->params.maxInodeNumber / 8) + 1);
        bzero(fileSystem->inodeAllocShadow, (fileSystem->params.maxInodeNumber / 8) + 1);    
        memcpy(fileSystem->inodeAllocShadow, oldInodeShadow,
                fileSystem->inodeAllocShadowSize);
        fileSystem->inodeAllocShadowSize = (fileSystem->params.maxInodeNumber / 8) + 1;

        fs_ramDealloc(FS_RAM_CORE_MISC, oldInodeShadow);
    }

    // Resize shadow info for blocks
    if (fileSystem->blockAllocShadowSize < ((fileSystem->params.totalBlocks / 8) + 1))
    {
        fileSystem->blockAllocShadow = fs_ramAlloc(FS_RAM_CORE_MISC,
                                        (fileSystem->params.totalBlocks / 8) + 1);
        bzero(fileSystem->blockAllocShadow, (fileSystem->params.totalBlocks / 8) + 1);
        memcpy(fileSystem->blockAllocShadow, oldBlockShadow,
                                        fileSystem->blockAllocShadowSize);

#if 0
        printf("realloc shadow: block %p size %lld (old %d)\n",
                fileSystem->blockAllocShadow,
                fileSystem->params.totalBlocks / 8 + 1,
                fileSystem->blockAllocShadowSize);
#endif
        fileSystem->blockAllocShadowSize = (fileSystem->params.totalBlocks / 8) + 1;

        fs_ramDealloc(FS_RAM_CORE_MISC, oldBlockShadow);
    }
}

void
fs_freeAllocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    fs_ramDealloc(FS_RAM_CORE_MISC, fileSystem->blockAllocShadow);
    fs_ramDealloc(FS_RAM_CORE_MISC, fileSystem->inodeAllocShadow);

} /* fs_freeAllocShadow */

#else // FS_DEBUG

inline void
fs_initShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
}
inline void
fs_initAllocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
}
inline void
fs_reallocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
}
inline void
fs_freeAllocShadow(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
}

#endif // FS_DEBUG



//   B L O C K S   A N D   E X T E N T S


/**
 * A collection of extents.
 */
struct fs_ExtentVector {
    /** Current number of extents. */
    uint32 count;
    /** Maximum size of this array */
    uint32 size;
    /** The actual extents -- variable sized. */
    fs_Extent extents[1];
};


/**
 * Constants for the extent vector.
 */
enum {
    /** Initial size of an extent vector. */
    EXTENT_VECTOR_INITIAL_SIZE = 5,
    /** Multiplier for increasing size of an extent vector. */
    EXTENT_VECTOR_SIZE_MULTIPLIER = 3,
};


/**
 * Create an extent vector with the specified count and size.
 */
//static fs_ExtentVector *fs_makeExtentVector(
inline fs_ExtentVector *fs_makeExtentVector(
    uint32 count,
    uint32 size)
{
    fs_ExtentVector *vector
        = fs_ramAlloc(FS_RAM_EXTENT_VECTOR,
                      offsetof(fs_ExtentVector, extents[size]));
    assert(count <= size);
    vector->count = count;
    vector->size = size;
    return vector;
}


/**
 * Add an element to an extent vector, possibly copying the vector
 * and returning a new one.
 */
static fs_ExtentVector *fs_extentVectorAdd(
    fs_Context *context,
    fs_ExtentVector *vector,
    fs_Extent extent)
{
    uint32 count;
    
    if (vector == NULL)
        vector = fs_makeExtentVector(0, EXTENT_VECTOR_INITIAL_SIZE);

    count = vector->count;
    if (count == vector->size) {
        fs_ExtentVector *newVector
            = fs_makeExtentVector(
                count, vector->size * EXTENT_VECTOR_SIZE_MULTIPLIER);
        memcpy(newVector->extents,
               vector->extents,
               count * sizeof (fs_Extent));
        fs_ramDealloc(FS_RAM_EXTENT_VECTOR, vector);
        vector = newVector;
    }

    assert(count < vector->size);
    vector->extents[vector->count] = extent;
    vector->count = count + 1;

    return vector;
}


/**
 * Allocate an extent from an extent that has been removed from the
 * binary buddy tree, returning the excess back to the tree.
 *
 * This function should only be called if the returned extent was too
 * large.
 */
static fs_Extent fs_allocateFromBuddyExtent(
    fs_Context *context,
    fs_Extent  takenExtent,
    uint16     len,
    fs_Block   targetBlock)
{
    uint16 takenLen   = fs_extentLength(takenExtent);
    uint64 takenFirst = fs_blockNumber(fs_extentStartBlock(takenExtent));
    uint64 takenLast  = takenFirst + takenLen;

#ifdef NOTDEF 
    uint64 target = fs_blockNumber(targetBlock);
#endif
    uint64 first;
    uint64 last;

    fs_Extent extent;

    assert(takenLen > 0);
    assert(len > 0);
    assert(len < takenLen);

    //printf("allocFromBuddyExtent: %lld,%d  len %d\n",
    //       takenFirst, takenLen, len);

    // This is all very cute, but it tends to work better
    // if we simply take from the beginning, since typically
    // the next request is the next page of the file and it
    // will all work out contiguously.
#ifdef NOTDEF 
    if (target < takenFirst) {
        // start at the earliest point in the allocated extent
        first = takenFirst;
    }
    else if (target < takenLast) {
        // prefer getting the position right to getting the size right
        first = target;
        len = MIN(len, takenLast - first);
    }
    else {
        // start at the latest point in the allocated extent
        first = takenLast - len;
    }
#else
    first = takenFirst;
#endif

    last = first + len;

    assert(takenFirst <= first);
    assert(last <= takenLast);

    if (takenFirst < first) {
        fs_Extent before
            = fs_makeExtent(first - takenFirst, fs_makeBlock(takenFirst));
        context->pendingFreeExtents = fs_extentVectorAdd(context,
                                         context->pendingFreeExtents, before);
    }

    if (last < takenLast) {
        fs_Extent after = fs_makeExtent(takenLast - last, fs_makeBlock(last));
        context->pendingFreeExtents = fs_extentVectorAdd(context,
                                         context->pendingFreeExtents, after);
    }

    extent = fs_makeExtent(len, fs_makeBlock(first));
    return extent;
}


/**
 * Allocate an extent from the binary buddy tree.  An attempt is made to
 * locate the block near the target and return an extent of the specified
 * size.  If the anySize parameter is true, an extent of a convenient size
 * for the allocator will be returned;  if it is false, an extent no larger
 * than the requested size will be returned.
 *
 * If the noSnap parameter is true, the snapshots will be disabled on
 * all blocks in the extent, using the refbytes file.
 *
 * TODO - exactLocation parameter
 */
static fs_Extent fs_allocateBuddyExtent(
    fs_Context *context,
    uint16 requested,
    fs_Block target,
    boolean anySize,
    boolean noSnap,
    boolean isFirstFileExtent)
{
    fs_ExtentVector *vector = context->pendingFreeExtents;
    fs_Extent       extent;
    int             length;

    // Try to allocate from this context's
    // pending free extents before removing
    // one from the buddy tree.

    if (vector != NULL)
    {
        vector->count -= 1;
        extent = vector->extents[vector->count];
        length = fs_extentLength(extent);

        //printf("sm 0x%x req %d alloc from pending target %lld blk %lld,%d\n",
        //    context->smHandle, requested,
        //    fs_blockNumber(target),
        //    fs_blockNumber(fs_extentStartBlock(extent)),
        //    length);

        if (vector->count == 0)
        {
            fs_ramDealloc(FS_RAM_EXTENT_VECTOR, vector);
            context->pendingFreeExtents = NULL;
        }
    }
    else
    {
        extent = fs_buddyFindExtent(context, requested, target,
                                    isFirstFileExtent);
        length = fs_extentLength(extent);

        //printf("sm 0x%x req %d alloc from buddy target %lld blk %lld,%d\n",
        //    context->smHandle, requested,
        //    fs_blockNumber(target),
        //    fs_blockNumber(fs_extentStartBlock(extent)),
        //    length);

        // PR #1700 consistency between buddy and refbytes essential
        fs_refByteModifyExtent(
                context,
                noSnap ? fs_refbyteSetMeta : fs_refbyteSetLive,
                extent);
    }

    if (length > 0) {
        if (length > requested && !anySize) {
            extent = fs_allocateFromBuddyExtent(
                context, extent, requested, target);
            assert(fs_extentLength(extent) == requested);
            length = requested;
        }
        
        FS_DBG_SHADOW_ALLOC_EXTENT(extent);
    }

    return extent;
}

/**
 * Allocate a single block near the given target block number.
 */
static fs_Block fs_allocateBlock(
    fs_Context  *context,
    fs_Block    target,
    fs_Inode    *inode)
{
    boolean noSnap = fs_inodeIsNoSnap(context, inode);
#ifdef NEW_ALLOC_POLICY
    fs_Extent extent = fs_allocateBuddyExtent(context, 1, fs_makeBlock(1ULL),
                                              FALSE, noSnap, FALSE);
#else // NEW_ALLOC_POLICY
    fs_Extent extent = fs_allocateBuddyExtent(context, 1, target,
                                              FALSE, noSnap, FALSE);
#endif // NEW_ALLOC_POLICY
    
    if (fs_extentLength(extent) == 0) {
        SPACE_TRACE(("  out of space"));
        return FS_INVALID_BLOCK;
    }

    assert(fs_extentLength(extent) == 1);
    SPACE_TRACE(("allocateBlock inum %lld => %lld",
                inode->inum, fs_blockNumber(fs_extentStartBlock(extent))));
    return fs_extentStartBlock(extent);
}


/**
 * Identify a good target location for an extent, of the given length,
 * starting at the specified page of the specified file.
 *
 * Since the write requests could be out of order: e.g. write page 7 could
 * come before write page 6, we need to look for preceding or/and following
 * pages to place place correctly the ne page.
 * Currently we look up to 8 pages before or/and after the new page. Why 8?
 * 8 = 2*4pages, 4pages = max 1write (NFSv3)
 * If the page is the first one to be allocated and its not page 0, then
 * make sure to keep spaces for the preceding pages.
 *
 * TODO - exact vs. inexact matches
 */
static fs_Block fs_pickTargetBlock(
    fs_Context *context,
    fs_Inode *inode,
    uint32 len,
    uint64 page)
{
    fs_Parameters * params = &context->fileSystem->params;
    fs_Buf *        inodeBuf;
    uint64          baseBlock;
    int             i;
    uint64          bnum = 0;

#ifdef NEW_ALLOC_POLICY
    bnum = 1;
    goto fs_pickTargetBlock_exit;    
#endif // NEW_ALLOC_POLICY
    
    assert(inode != NULL);

    //Look up the preceding pages, if any.
    for (i = 1; i <= 8; i++)
    {
        if (page > (i-1))
        {
            fs_Block block;
            
            block = fs_emap_lookup_block(context, inode, page - i);
            if (fs_blockIsValid(block))
            {
                bnum = fs_blockNumber(block) + i;
                goto fs_pickTargetBlock_exit;
            }
        }
    } // for (i)

    // Look up the following pages
    for (i = 1; i <= 8; i++)
    {
        fs_Block block;

        block = fs_emap_lookup_block(context, inode, page + i);
        if (fs_blockIsValid(block))
        {
            bnum = fs_blockNumber(block) - i;
            goto fs_pickTargetBlock_exit;
        }
    } // for (i)

    // Just try to place the file near its inode.  Small files we attempt
    // to place before the inode (by 1/3 of the inode stride);  large files
    // (for now, those more than one page long, or which look as though they
    // might grow to more than one page), we place just after the inode.
    // (We also place files from the first inode page after the inode,
    // because it's at the start of the disk.)

    inodeBuf = fs_inodeGetBufferShared(context, inode->inum, inode->snapnum,
                                       FALSE, FALSE);
    assert(inodeBuf != NULL);
    if (!fs_blockIsValid(inodeBuf->block))
    {
        // TODO - handle if not valid (guess where it will land)
        fs_bufRelease(context, inodeBuf);
        goto fs_pickTargetBlock_exit;
    }
    baseBlock = fs_blockNumber(inodeBuf->block);
    fs_bufRelease(context, inodeBuf);

    if (page > 0
        || len > 0
        || inode->dinode.size > params->blockSize
        || inode->inum < FS_MAX_RESERVED_INODE)
    {
        bnum = baseBlock + page + 1;
        goto fs_pickTargetBlock_exit;
    }
    else
    {
        bnum = baseBlock + params->inodeFileStride / 3;
        goto fs_pickTargetBlock_exit;
    }

  fs_pickTargetBlock_exit:
    
    /* Make sure we don't go outside the filesystem!
     */
    return(fs_makeBlock(bnum % context->fileSystem->super->numBlks));
}

/**
 * Allocate an extent for the file with the specified inode, of
 * approximately to the specified length in blocks, starting at the given
 * page within the file.  If the anySize parameter is true, the returned
 * extent may be larger than requested; otherwise, it will be at most the
 * requested size.
 *
 * No distinction is made between allocating extents for user files
 * vs. metadata files, because this call just reserves address space.
 * Assignment of blocks to buffers and allocation of buffers are
 * separate concerns.
 *
 * TODO - an exact parameter indicates that the allocator strongly wants a
 * block located at the target, if possible.  (In conjunction w/pickTarget?)
 */
fs_Extent fs_allocateExtent(
    fs_Context *context,
    fs_Inode *inode,
    uint64 page,
    uint32 len32,
    boolean anySize)
{
    uint16 len = MIN(len32, EXTENT_LENGTH_MAX);
    fs_Block target;
    fs_Extent extent;

    SPACE_TRACE(("allocateExtent inum %lld page %lld len %d",
                 (inode == NULL) ? 0 : inode->inum, page, len));

    if (inode == NULL) {
        target = fs_makeBlock(0);
    }
    else {
        target = fs_pickTargetBlock(context, inode, len, page);
        SPACE_TRACE(("  target(%lld, %lld, %d) = %lld",
                     inode->inum, page, len, fs_blockNumber(target)));
    }

    if (context->fileSystem->bootstrap != NULL)
        return fs_bootAllocateExtent(context, inode, len);

    /* If this is the first extent and the target block is not free,
     * we will try to put the extent at the beginning of a new level
     * in order to give some spece for the files to grow.
     */
    if ((inode->dinode.size == 0) && (!fs_isBlockFree(context, target)))
    {
        extent = fs_allocateBuddyExtent(context, len, target, anySize,
                                        fs_inodeIsNoSnap(context, inode),
                                        TRUE);
    }
    else
    {
        extent = fs_allocateBuddyExtent(context, len, target, anySize,
                                        fs_inodeIsNoSnap(context, inode),
                                        FALSE);
    }

    SPACE_TRACE(("  allocateExtent(%lld, %lld, %d) => %lld ... %+d",
                 (inode == NULL) ? 0 : inode->inum, page, len,
                 fs_blockNumber(fs_extentStartBlock(extent)),
                 fs_extentLength(extent)));

    return extent;
}

/**
 * Allocate a block for btree metadata, associated with the specified
 * inode number.  The allocator will try to place the block near the rest
 * of the metadata for the inode.
 *
 * This is a special interface (hopefully temporary) to allow the btree
 * copy-on-write code to avoid needing to know the inode that it's using.
 * This will probably go away when we have deferred metadata allocation.
 *
 * TODO - remove soon
 */
fs_Block fs_allocateMetadataBlockSpecial(
    fs_Context *context,
    fs_Inode   *inode,
    fs_Block   target)
{
    SPACE_TRACE(("allocateMetadataBlock inum %lld target %lld",
                 inode->inum, fs_blockNumber(target)));

    if (context->fileSystem->bootstrap != NULL)
        return fs_extentBlock(fs_bootAllocateExtent(context, inode, 1), 0);

    return fs_allocateBlock(context, target, inode);
}


/**
 * Allocate a block for btree metadata, associated with the specified
 * inode.  The allocator will try to place the block near the rest of
 * the metadata for the inode.
 */
fs_Block fs_allocateMetadataBlock(
    fs_Context *context,
    fs_Inode *inode)
{
    fs_Buf *inodeBuf
        = fs_inodeGetBufferShared(context, inode->inum, inode->snapnum,
                                  FALSE, FALSE);
    fs_Block targetBlock = inodeBuf->block;

    fs_bufRelease(context, inodeBuf);
    return fs_allocateMetadataBlockSpecial(context, inode, targetBlock);
}


/**
 * Allocate a block for storing the block of the inode file containing
 * the specified (byte) offset into the inode file.  These blocks are
 * distributed throughout the filesystem
 */
fs_Block fs_allocateInodeFileBlock(
    fs_Context *context,
    fs_Inode *inode,
    uint64 offset)
{
    uint64 target;
    uint32 stride;
    fs_Parameters *params = &context->fileSystem->params;

    SPACE_TRACE(("allocateInodeFileBlock offset %lld", offset));

    if (context->fileSystem->bootstrap != NULL)
        return fs_extentBlock(fs_bootAllocateExtent(context, inode, 1), 0);

    stride = params->inodeFileStride;
    target = ((offset / params->blockSize) * stride) + (stride / 2);

    return fs_allocateBlock(context, fs_makeBlock(target), inode);
}


/**
 * Comparison function for ordering extents by their starting block numbers.
 *
 * TODO - check that extents don't overlap?
 */
static int fs_extentCompare(
    const void *e0,
    const void *e1)
{
    uint64 block0 = fs_blockNumber(fs_extentStartBlock(*(fs_Extent *) e0));
    uint64 block1 = fs_blockNumber(fs_extentStartBlock(*(fs_Extent *) e1));

    if (block0 < block1)
        return -1;
    if (block0 > block1)
        return 1;
    
    assert(FALSE);              // array should not have duplicates
    UNREACHED();
}


/**
 * Deallocate all extents on the pending free list for this context.
 * Done at commit time.
 */
void fs_commitDeallocate(
    fs_Context *context)
{
    uint32 i;
    
    fs_ExtentVector *vector = context->pendingFreeExtents;
    if (vector == NULL)
        return;
    context->pendingFreeExtents = NULL;

    SPACE_TRACE(("commitDeallocate %d extent%s",
                 vector->count,
                 (vector->count == 1) ? "" : "s"));
    
    qsort(vector->extents,
          vector->count,
          sizeof (fs_Extent),
          fs_extentCompare);

    for (i = 0; i < vector->count; i++) {
        fs_Extent extent = vector->extents[i];

        SPACE_TRACE(("commitDeallocate extent %lld ... +%d",
                     fs_blockNumber(fs_extentStartBlock(extent)),
                     fs_extentLength(extent)));
        
        fs_buddyAddExtent(context, vector->extents[i], FALSE);
        // PR #1700: modify ref bytes at commit time after they
        //           are added back to the buddy file
        fs_refByteModifyExtent(context, fs_refbyteDeallocate, extent);
    }
    
    fs_ramDealloc(FS_RAM_EXTENT_VECTOR, vector);
}


/**
 * Return the specified extent to the allocator for reuse.  All actual
 * deallocations are deferred until commit time, to avoid two-phase
 * locking problems.
 */
static void fs_freeExtent(
    fs_Context *context,
    fs_Extent extent)
{
#if defined(FS_DEBUG_BUF_CKSUM)
    int         i;
    fs_Block    block;
#endif  //FS_DEBUG

#if defined(FS_DEBUG_BUF_CKSUM)
    /*
     *  Free the checksums for this blocks
     */

    for (i = 0; i < fs_extentLength(extent); i++) {
        block = fs_extentBlock(extent, i);
        fs_bufReleaseCksumForBlock(context, block);
    }
#endif  //FS_DEBUG

    context->pendingFreeExtents
        = fs_extentVectorAdd(context, context->pendingFreeExtents, extent);

    FS_DBG_SHADOW_FREE_EXTENT(extent);
}


/**
 * Deallocate the specified block from the live file system.  If the
 * block is not in use by a snapshot, it is returned to the allocator;
 * otherwise, the live reference bit is merely cleared for the block.
 */
void fs_deallocateBlock(
    fs_Context *context,
    fs_Block block)
{
    fs_RefByte refByte = fs_refGetByte(context, block);

    assert(context->fileSystem->bootstrap == NULL);
    SPACE_TRACE(("deallocateBlock %lld", fs_blockNumber(block)));

    if (fs_refbyteSnapCount(refByte) > 0)
        fs_refByteModifyBlock(context, fs_refbyteClearLive, block);
    else
        fs_freeExtent(context, fs_makeExtent(1, block));
}


/**
 * Deallocate all the blocks in the specified extent.
 */
void fs_deallocateExtent(
    fs_Context *context,
    fs_Extent extent)
{
    int i, j;
    int len = fs_extentLength(extent);
    
    assert(len > 0);
    assert(context->fileSystem->bootstrap == NULL);

    SPACE_TRACE(("deallocateExtent %lld ... +%d",
                 fs_blockNumber(fs_extentBlock(extent, 0)),
                 fs_extentLength(extent)));

    for (i = 0; i < len; i = j) {
        for (j = i; j < len; j++) {
            fs_Block block = fs_extentBlock(extent, j);
            if (fs_refbyteSnapCount(fs_refGetByte(context, block)) > 0)
                break;
        }
        if (j > i)
            fs_freeExtent(context, fs_makeSubExtent(extent, i, j));
        if (j < len) {
            int m = j;
            for (; j < len; j++) {
                fs_Block block = fs_extentBlock(extent, j);
                if (fs_refbyteSnapCount(fs_refGetByte(context, block)) == 0)
                    break;
            }
            fs_refByteModifyExtent(context, fs_refbyteClearLive,
                                   fs_makeSubExtent(extent, m, j));
        }
    }
}



//   I N O D E S


/*
 * The inode summary files are bitmaps (see fs-bitmap.h), indexed by
 * the page number within the inode file.
 */


/**
 * Return the page number of the inode file containing the given inode
 * number.
 */
static uint64 fs_inumToIpage(
    fs_Context *context,
    fs_Inumber inum)
{
    return inum / context->fileSystem->params.inodesPerBlock;
}


/**
 * Return the buffer of the inode file, with an exclusive lock, for the block
 * with the specified inumber.
 *
 * Uses a trylock to acquire the buffer lock;  returns NULL if we were
 * unable to acquire the lock.
 */
static fs_Buf *fs_ipageGetBuffer(
    fs_Context *context,
    uint64     ipage,
    int        snapnum,
    boolean    sleepLock)
{
    uint32 inodesPerBlock = context->fileSystem->params.inodesPerBlock;
    return fs_inodeGetBufferExclusive(context,
                                      ipage * inodesPerBlock,
                                      snapnum,
                                      TRUE,
                                      sleepLock ? FALSE : TRUE);
}


/**
 * Return a free inode in the specified buffer, or NULL if it wasn't
 * possible to allocate one.  The inode is held exclusive.  Sets the
 * mode to the specified mode for the inode, which must not be zero.
 */
static fs_Inode *fs_allocateInodeInBuffer(
    fs_Context *context,
    fs_Buf     *buf,
    uint16     mode,
    uint64     parentInum)
{
    int i;

    fs_Parameters *params = &context->fileSystem->params;
    fs_Inumber base = buf->page * params->inodesPerBlock;

    assert(mode != 0);

    for (i = 0; i < params->inodesPerBlock; i++)
    {
        fs_Dinode *dinode = &buf->u.dinode->dinode[i];

        if (dinode->attrs == 0)
        {
            fs_BitmapOperation debugOp;
            fs_Inode *inode;
            int bufOffset;
            uint16 attrs = (mode & (FS_IFMT_MASK | FS_IATR_MASK));

            FS_LOCK_ORDER_SET_PARENT(parentInum);
            inode = fs_inodeInstantiateAllocator(context, 0, base + i, buf);

            FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
            FS_ASSERT(inode->dinode.attrs == 0);

            // Write to the buf dinode, _and_ the incore-inode copy of dinode.
            dinode->attrs = inode->dinode.attrs = attrs;
            bufOffset = FS_INODE_PAGE_OFFSET(inode->inum);
            debugOp.u.inum = inode->inum;
            debugOp.op = BITMAP_OP_ALLOC;
            fs_bufWrite(context, buf, bufOffset + offsetof(fs_Dinode, attrs),
                        sizeof(dinode->attrs),
                        FS_BUF_WRITE_LOG | FS_BUF_WRITE_LOG_BIT_SET,
                        (char *) &attrs, &debugOp);

            // Update the genNum in the buffer, but don't log it.
            // Its non-critical and logging it would cause us to
            // hold the buffer until commit, which we don't want
            // to happen to inode file pages for performance reasons.
            dinode->genNum += 1;
            dinode->genNum &= ~FS_INODE_GENNUM_SNAP_MAGIC;
            inode->dinode.genNum = dinode->genNum;
            fs_bufWrite(context, buf, bufOffset + offsetof(fs_Dinode, genNum),
                        sizeof(dinode->genNum), 0, NULL, NULL);

            return inode;
        }
    }

    return NULL;
}


/**
 * Sentinel value for trylock failure when trying to allocate an inode.
 */
#define FS_INODE_PAGE_CONTENTION     ((fs_Inode *) -1)


/**
 * Return a free inode in the specified inode file page.  Returns NULL
 * if there was no free inode on the page, despite what the summary
 * indicated.  If an inode is returned, it is locked exclusively by
 * the current context.
 *
 * Updates, if necessary, the specified inode summary files.  The dirsum
 * parameter should be non-NULL if and only if the inode to be created is
 * a directory.
 *
 * Returns FS_INODE_PAGE_CONTENTION if we couldn't acquire the lock on the
 * relevant inode page.
 *
 * TODO - cache recently probed inode offsets in the buf
 */
static fs_Inode *fs_allocateInodeInPage(
    fs_Context *context,
    uint64     ipage,
    fs_Inode   *inosum,
    fs_Inode   *dirsum,
    uint16     mode,
    uint64     parentInum,
    boolean    sleepLock)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_Inode *inode;
    fs_Buf *buf;
    int i;
    boolean full;

    SPACE_TRACE(("   allocateInodeInPage ipage %lld isDir %s",
                 ipage, strbool(dirsum != NULL)));

    buf = fs_ipageGetBuffer(context, ipage, 0, sleepLock);
    if (buf == NULL)
        return FS_INODE_PAGE_CONTENTION;

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));

    inode = fs_allocateInodeInBuffer(context, buf, mode, parentInum);

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));

    fs_bufReleaseExclusive(context, buf);

    if (inode == NULL) {
        //printf("no free inode in non-full page\n");
        //fs_warn(context, FS_INTERNAL_ERR, "no free inode in non-full page");
        return NULL;
    }

    if (inode->inum < FS_MAX_RESERVED_INODE)
    {
        printf("fs_allocateInodeInPage:  allocated reserved inode %lld\n",
            inode->inum);
        FS_ASSERT(FALSE);
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }

    full = TRUE;
    for (i = (inode->inum % params->inodesPerBlock) + 1;
         i < params->inodesPerBlock;
         i++)
    {
        fs_Dinode *dinode = &buf->u.dinode->dinode[i];
        if (dinode->attrs == 0) {
            full = FALSE;
            break;
        }
    }

    if (full) {
        SPACE_TRACE(("inode page %lld full", ipage));
        fs_bitmapSet(context, inosum, ipage, NULL);
    }
    if (dirsum != NULL) {
        SPACE_TRACE(("inode page %lld hasDir", ipage));
        fs_bitmapSet(context, dirsum, ipage, NULL);
    }

    return inode;
}


/**
 * Allocate an inode for a file or directory to be created in the specified
 * directory.
 *
 * TODO - turn the locs into indices
 */
static fs_Inode *fs_allocateInode(
    fs_Context *context,
    fs_Inumber parentInum,
    fs_Inumber targetInum,
    boolean    isDir,
    uint16     mode,
    uint32     fsetNum)
{
    fs_Inode *inosum = I_GET_SHARED(0, FS_FULL_INODE_BLOCK_INODE);
    fs_Inode *inodeFile = NULL;
    fs_FileSystem *fs = context->fileSystem;
    fs_Inode *dirsum;
    fs_Inode *inode;
    fs_Dinode *dinode;
    uint32 genNum, loops = 0;
    fs_BitmapLocList *excluded = NULL;
    boolean haveExcl = FALSE;

    assert((mode & FS_IFMT_MASK) != 0);
    assert(isDir == FS_I_IS_DIR(mode));
    assert(fsetNum < FS_FILESET_INVAL_ID);

    if (!fs_haveMininodesAvail(fs) && !fs->igrowThreadStarted &&
        !context->eek.inUse)
    {
        boolean setIt = FALSE;

        NFX_SPIN_LOCK(&fs->spin);
	if (!fs->igrowThreadStarted)
        {
            fs->igrowThreadStarted = TRUE;
            setIt = TRUE;
        }
        NFX_SPIN_UNLOCK(&fs->spin);

        if (setIt)
        {
              fs_threadFork(context, fs_igrowThread,
                            TRUE, FALSE,
                            FS_DAEMON_TYPE_INO_GROW, NULL);
        }
    }

    if (!isDir)
        dirsum = NULL;
    else
        dirsum = I_GET_SHARED(0, FS_DIRECTORY_IN_INODE_BLOCK_INODE);

    while (TRUE)
    {
        uint64 ipage;
        uint64 target = fs_inumToIpage(context, targetInum);

        SPACE_TRACE(("allocateInode target %lld isDir %s",
                     targetInum, strbool(isDir)));

        loops += 1;

        if (dirsum == NULL)
            ipage = fs_bitmapFindClear(context, inosum, target, excluded);
        else
            ipage = fs_bitmapFindClear2(context, inosum, dirsum, target,
                                        excluded);

        if (ipage == FS_INVALID_BITMAP_INDEX)
        {
            if (dirsum != NULL)
            {
                // TODO - once this happens, we may want to turn off looking
                // at the dirsum for the next few minutes for this volume,
                // or kick off an igrow thread
                SPACE_TRACE(("  restarting inode search with isDir = FALSE"));
                fs_inodeRelease(context, dirsum);
                dirsum = NULL;
                continue;
            }
            if (!haveExcl)
            {
                SPACE_TRACE(("  restarting inode search with exclusive access"));
                excluded = NULL;
                fs_inodeRelease(context, inosum);
                inosum = I_GET_EXCL(0, FS_FULL_INODE_BLOCK_INODE);
                inodeFile = I_GET_EXCL(0, FS_INODE_INODE);
                haveExcl = TRUE;
                continue;
            }
            printf("out of inodes %d allocs %d deallocs %lld avail (%d)\n",
                   fs_inodeAllocs,
                   fs_inodeDeallocs,
                   fs->params.maxInodeNumber -
                   fs->super->inodesUsed,
                   fs_haveMininodesAvail(fs));
            SPACE_TRACE(("out of inodes"));
            if (haveExcl)
            {
                fs_inodeReleaseExclusive(context, inodeFile);
                fs_inodeReleaseExclusive(context, inosum);
            }
            else
                fs_inodeRelease(context, inosum);

            fs_volumeException(context, context->fileSystem,
                               FS_VOLUME_EXCEPTION);
            return NULL;
        }

        inode = fs_allocateInodeInPage(context, ipage, inosum, dirsum,
                                       mode, parentInum, haveExcl);
        if (inode == NULL)
        {
            SPACE_TRACE(("inode page %lld full (corrected)", ipage));
            fs_bitmapSet(context, inosum, ipage, NULL);
        }
        else if (inode == FS_INODE_PAGE_CONTENTION)
        {
            SPACE_TRACE(("inode page %lld already locked, excluding", ipage));
            excluded = fs_makeBitmapLocList(context, excluded, ipage);
        }
        else
        {
            break;
        }
    }

    fs_inodeAllocs += 1;
    if (loops > fs_inodeAllocLoopMax)
    {
        //printf("new fs_inodeAllocLoopMax: %d (used: %lld)\n", loops,
        //       fs->params.maxInodeNumber - fs->super->inodesUsed);
        fs_inodeAllocLoopMax = loops;    
    }

    if (haveExcl)
    {
        fs_inodeReleaseExclusive(context, inodeFile);
        fs_inodeReleaseExclusive(context, inosum);
    }
    else
        fs_inodeRelease(context, inosum);

    if (dirsum != NULL)
        fs_inodeRelease(context, dirsum);

    SPACE_TRACE(("   dir %lld -> allocated inum %lld",
                 targetInum, inode->inum));

    FS_DBG_SHADOW_ALLOC_INODE(inode->inum);

    dinode = &inode->dinode;
    genNum = dinode->genNum;
    bzero(dinode, sizeof (fs_Dinode));
    dinode->attrs = (mode & (FS_IFMT_MASK | FS_IATR_MASK));
    dinode->genNum = genNum;
    dinode->linkCnt = isDir ? 2 : 1;
    dinode->fsetNum = fsetNum;
    dinode->userAttr = isDir ? FS_INODE_UATTR_DIRECTORY :
                                 FS_INODE_UATTR_NORMAL;
    dinode->ctime = fs_getTimeOfDay();
    dinode->mtime = dinode->ctime;
    dinode->atime = dinode->ctime;
    dinode->version = 0;    // This is a timing version and not inode version

    switch (mode & FS_IFMT_MASK) {
        case FS_IFMT_REG:
        case FS_IFMT_QUOT:
            fs_emaproot_init(context, inode);
            break;
        case FS_IFMT_DIR:
            fs_dirroot_init(context, inode, parentInum);
            break;
        case FS_IFMT_BLK:
        case FS_IFMT_CHR:
        case FS_IFMT_FIFO:
        case FS_IFMT_SOCK:
        case FS_IFMT_SYML:
            break;
        default:
            fs_error(context, FS_INTERNAL_ERR,
                     "bad mode: %#o", mode & FS_IFMT_MASK);
    }

    fs_inodeWrite(context, inode, INO_WRITE_LOG);
    fs_incrementInodeUse(context, inode->inum);

    return inode;
}


/**
 * Pick a random (ish) inode as a starting point for searching for a new
 * inode location.
 * 
 * TODO - use libc rand vs. our own private rand
 */
//static fs_Inumber fs_inodeRandom(
inline fs_Inumber fs_inodeRandom(
    fs_Context *context)
{
    return fs_random64(context, context->fileSystem->params.maxInodeNumber);
}


/**
 * Allocate a disk inode (and inode number) for a file to be created in
 * the specified directory.  The inode's block will be near the directory
 * inode's block, if possible.  If the specified directory is NULL, an
 * arbitrary inode number will be chosen.
 */
fs_Inode *fs_allocateFileInode(
    fs_Context *context,
    fs_Inode   *directory,
    uint16     perm,
    uint32     uid,
    uint32     gid,
    uint32     fileset)
{
    fs_SecurityDescNfs nfs;
    fs_Inode *inode;
    uint64 target;
    uint16 mode = ((perm & FS_IFMT_MASK) == 0) ? (perm | FS_IFMT_REG) : perm;

    assert(!FS_I_IS_DIR(mode));

#ifdef NEW_ALLOC_POLICY
    target = FS_MAX_RESERVED_INODE;
#else
    if (directory != NULL)
        target = directory->inum;
    else
        target = fs_inodeRandom(context);
#endif

    SPACE_TRACE(("allocateFileInode dir %lld", target));

    inode = fs_allocateInode(context, directory ? directory->inum : 0,
                             target, FALSE, mode, fileset);

    nfs.hdr.attr = FS_SDESC_TYPE_NFS;
    nfs.hdr.len = sizeof nfs;
    nfs.hdr.cont = 0;
    nfs.hdr.uQtId = nfs.hdr.gQtId = 0;
    nfs.uid = uid;
    nfs.gid = gid;
    nfs.mode = (perm & FS_IMODE_MASK);
    nfs.pad[0] = 0;

    fs_securityDescSet(context, inode, &nfs.hdr);

    return inode;
}


/**
 * Allocate a disk inode (and inode number) for a directory to be created
 * in the given parent directory.  For directories near the root of the
 * file system, subdirectories are created in arbitrary disk locations, so
 * as to spread them across the disk;  directories further from the root
 * are created near their parents, if possible, in inode number.
 */
fs_Inode *fs_allocateDirectoryInode(
    fs_Context *context,
    fs_Inode   *parent,
    uint16     perm,
    uint32     uid,
    uint32     gid,
    uint32     fileset)
{
    fs_Inode *inode;
    fs_SecurityDescNfs nfs;
    uint64 target;
    uint16 mode = ((perm & FS_IFMT_MASK) == 0) ? (perm | FS_IFMT_DIR) : perm;

    assert(FS_I_IS_DIR(mode));
    
#ifdef NEW_ALLOC_POLICY
    target = FS_MAX_RESERVED_INODE;
#else
    if (parent->inum == FS_ROOT_INODE || 
        fs_dirroot_get_parent(context, parent) == FS_ROOT_INODE)
    {
        target = fs_inodeRandom(context);
    }
    else {
        target = parent->inum;
    }
#endif

    SPACE_TRACE(("allocateDirectoryInode parent %lld target %lld",
                 parent->inum, target));

    inode = fs_allocateInode(context, parent->inum, target, TRUE,
                             mode, fileset);

    nfs.hdr.attr = FS_SDESC_TYPE_NFS;
    nfs.hdr.len = sizeof nfs;
    nfs.hdr.cont = 0;
    nfs.hdr.uQtId = nfs.hdr.gQtId = 0;
    nfs.uid = uid;
    nfs.gid = gid;
    nfs.mode = (perm & FS_IMODE_MASK);
    nfs.pad[0] = 0;

    fs_securityDescSet(context, inode, &nfs.hdr);

    return inode;
}


/**
 * Mark the disk inode as unallocated.
 */
static void fs_markDinodeUnused(
    fs_Context  *context,
    fs_Inode    *inode)
{
    fs_Buf      *buf;
    fs_Dinode   *dinode;
    fs_BitmapOperation debugOp;
    int         bufOffset;
    uint16      modeBits = 0xffff;

    buf = fs_inodeGetBufferExclusive(context, inode->inum, 0, FALSE, FALSE);
    dinode = &buf->u.dinode->dinode[inode->inum % FS_INODES_PER_PAGE];

    dinode->attrs = 0;
    bufOffset = FS_INODE_PAGE_OFFSET(inode->inum);

    debugOp.u.inum = inode->inum;
    debugOp.op = BITMAP_OP_DEALLOC;

    // Hold this buffer locked until after this transaction has
    // committed, otherwise other transactions could obtain this
    // inoode and commit before we do!

    fs_bufWrite(context, buf, bufOffset + offsetof(fs_Dinode, attrs),
                sizeof(dinode->attrs),
                FS_BUF_WRITE_LOG | FS_BUF_LOG_ADD_BUF |
                       FS_BUF_WRITE_LOG_BIT_RESET,
                (char *) &modeBits, &debugOp);
}


/**
 * Deallocate an inode number.  Caller must have exclusive
 * access to the inode.
 */
void fs_deallocateInode(
    fs_Context *context,
    fs_Inode   *inode)
{
    SPACE_TRACE(("deallocateInode inum %lld", inode->inum));

    assert(fs_lockOwnedExclusive(context, &inode->lock));

    fs_inodeDeallocStart(context, inode);
}


/**
 * Finish deallocation of the specified inode.
 * Done at commit time.
 */
void fs_commitInodeDeallocate(
    fs_Context  *context,
    fs_Inode    *inode)
{
    uint64 ipage = fs_inumToIpage(context, inode->inum);
    fs_Inode *inosum;

    SPACE_TRACE(("commit deallocateInode inum %lld", inode->inum));

    fs_decrementInodeUse(context, inode->inum);

    fs_inodeDeallocs += 1;

    FS_DBG_SHADOW_FREE_INODE(inode->inum);

    inosum = I_GET_SHARED(0, FS_FULL_INODE_BLOCK_INODE);
    fs_bitmapClearIfSet(context, inosum, ipage, NULL);
    fs_inodeRelease(context, inosum);

    if (FS_I_IS_DIR(inode->dinode.attrs))
    {
        fs_Inode *dirsum = I_GET_SHARED(0, FS_DIRECTORY_IN_INODE_BLOCK_INODE);
        if (!fs_bitmapClearIfSet(context, dirsum, ipage, NULL)) {
            //fs_warn(context, FS_INTERNAL_ERR,
            //      "inode %lld was a directory, but the dirsum bit was clear",
            //      inode->inum);
        }
        fs_inodeRelease(context, dirsum);
    }

    fs_markDinodeUnused(context, inode);  

    fs_inodeCommit(context, inode);
}


/**
 * Recalculate the mininum number of free blocks
 * we reserve for the specified filesystem.
 */
void
fs_recalcMinFreeBlks(fs_Context *context)
{
    fs_FileSystem *fs = context->fileSystem;
    uint64 maxInodesUsed = fs->super->inodeFileBlks * FS_INODES_PER_PAGE;
    int i, snapSlotsRemain = 0;
    union {
        uint64  dword;
        byte    b[8];
    } snapBits;

    // Calculate how many snapshot id's
    // are unused.
    snapBits.dword = fs->super->snapshots;
    for (i = 0; i < 8; i++)
    {
        snapSlotsRemain += 8 - fs_onesPerByte[snapBits.b[i]];
    }
    // Always reserve space for at least one, it
    // just feels better that way.
    snapSlotsRemain = MAX(1, snapSlotsRemain);

    //printf("numBlks: %lld\n", fs->super->numBlks);
    //printf("snapSlotsRemain: %d\n", snapSlotsRemain);
    //printf("maxInodesUsed: %lld\n", maxInodesUsed);

    // Our minFreeBlks needs to ensure enough
    // free blocks are available for critical operations
    // without holding too much needlessly. These include dump,
    // snap_create (which dump depends on), and remove/rmdir/setAttr
    // operations (which can require space if the inode file or btree
    // nodes are COW).

    NFX_SPIN_LOCK(&fs->spin);

    fs->minFreeBlks = 
  
        // support chown -R and other ops on
        // COW inodes
        fs->super->inodeFileBlks +

        // space for snapbits files (support full complement
        // of snapshots)
        ((fs->super->numBlks / FS_BLOCK_SIZE / 8) * snapSlotsRemain) +

        // space for dump temp file for incremental dump assuming
        // 50% of all possible inodes will need to be dumped and
        // namelengths are 1/8'th of maxlen on average.
        //
        // (might not be enough, but dump handles this possibility).
        ((maxInodesUsed * (FS_NAME_MAX / 8)) / FS_BLOCK_SIZE / 2) +
        ((maxInodesUsed * (uint64) sizeof(fs_Dinode)) / FS_BLOCK_SIZE / 2) +

        // fudge for snapdir expansion, trashcan dir,
        // btree inner node COW during truncate, etc.
        MAX(1000, maxInodesUsed / 1024);

    //printf("recalcMinFree new %lld (%lld,%lld,%lld,%lld,%lld)\n",
    // fs->minFreeBlks, 
    // fs->super->inodeFileBlks,
    // ((fs->super->numBlks / FS_BLOCK_SIZE / 8) * snapSlotsRemain),
    // ((maxInodesUsed * (FS_NAME_MAX / 8)) / FS_BLOCK_SIZE / 2),
    // ((maxInodesUsed * (uint64) sizeof(fs_Dinode)) / FS_BLOCK_SIZE / 2),
    // MAX(1000, maxInodesUsed / 1024));

    NFX_SPIN_UNLOCK(&fs->spin);
}


/**
 * Return TRUE when the filesystem has at least its minimum
 * reserved blocks free.
 */
inline boolean
fs_haveMinfreeAvail(fs_FileSystem *fs)
{
    return (fs->minFreeBlks < fs->super->numBlks - fs->super->blksUsed);
}


/**
 * Return the number of blocks currently available (ie.
 * unallocated) to users of the filesystem (ie. nfs clients).
 */
inline uint64 fs_numBlocksAvail(fs_FileSystem *fs)
{
    uint64 max = fs_numBlocksUserMax(fs);
    uint64 used = fs->super->blksUsed;

    // Careful here, since blksUsed could exceed usermax.
    if (max > used)
        return max - used;
    else
        return 0;
}


/**
 * Return the maximum number of blocks available to users of
 * the filesystem (ie. nfs clients). This hides free blocks
 * that we reserve in the system.
 */
inline uint64 fs_numBlocksUserMax(fs_FileSystem *fs)
{
    return fs->super->numBlks - fs->minFreeBlks;
}


/** 
 * Return TRUE when the filesystem has at least MININODES_PCT
 * and MININODES_HARD inodes available.
 */
#define MININODES_PCT     1ULL // Percent.
#define MININODES_HARD  512ULL // Keep this many free regardless of % avail
                               // so things like the trashcan will have a
                               // ready supply.

inline boolean fs_haveMininodesAvail(fs_FileSystem *fs)
{
    return ((100ULL * fs->super->inodesUsed) / fs->params.maxInodeNumber)
                       < (100ULL - MININODES_PCT) &&
           fs->params.maxInodeNumber -
            (FS_MAX_RESERVED_INODE + fs->super->inodesUsed) > MININODES_HARD;
}


/**
 * Return the number of inodes currently available (ie.
 * unallocated) to users of the filesystem (ie. nfs clients),
 * assuming the automated growth of the inode file to its max.
 */
inline uint64 fs_numInodesAvail(fs_FileSystem *fs)
{
    uint64 max = fs_numInodesUserMax(fs);
    uint64 used = fs->super->inodesUsed;

    // Careful here, since inodesUsed could exceed usermax.
    if (max > used)
        return max - used;
    else
        return 0;
}


/**
 * Return the maximum number of inodes available to users of
 * the filesystem (ie. nfs clients), assuming growth of
 * the inode file to its max (blkInodeRation == 1).  This
 * hides free inodes that we reserve in the system.
 */
inline uint64 fs_numInodesUserMax(fs_FileSystem *fs)
{
    // Round down to nearest inode page.
    uint64  maxInoNum = (fs->super->numBlks / fs->params.inodesPerBlock)
                             * fs->params.inodesPerBlock;

    return maxInoNum - (FS_MAX_RESERVED_INODE +
                      MAX(MININODES_HARD, (maxInoNum / 100ULL) * MININODES_PCT));
}


inline void fs_changeInodeUse(
	fs_Context *context,
	int change)
{
	NFX_SPIN_LOCK(&context->fileSystem->superLock);
	context->fileSystem->super->inodesUsed += change;
	assert((int)context->fileSystem->super->inodesUsed >= 0);
	NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        context->logSummary->inodesUsed += change;
}

/*
 * Add one to the count of total inodes in use in the superblock. 
 */
inline void fs_incrementInodeUse(
	fs_Context  *context,
	uint64 inum)
{
    fs_changeInodeUse(context, 1);
}

/*
 * Subtract one from the count of total inodes in use in the superblock. 
 */
inline void fs_decrementInodeUse(
	fs_Context  *context,
	uint64 inum)
{
    fs_changeInodeUse(context, -1);
}

#endif // SSC

/*------------------------------------------------------------------------------
 * Allocator statistics/utilities.
 *
 * To measure how our allocator performs, we need to see how fragmented are
 * the filesystem and the free space.
 *
 * To get the statistics of a particular file, we walk through its emap
 * knowing its inum.
 * To get the statistics of the entire file system, we go through all the
 * inodes and compile files' statistics.
 * To get the statistics of the free space, we go through the buddy pages.
 *
 * Note: all functions used to walk through the emap of a file have been copied
 * from btree/emap*.c
 *
 *------------------------------------------------------------------------------
 */

#ifdef ALLOC_FILE_SIZE_MEDIANE
/* List of all files' size.
 * Used to calculate the file size median.
 */
typedef struct fs_alloc_sizeElt_T fs_alloc_sizeElt_t;
struct fs_alloc_sizeElt_T
{
    uint64                          size;
    DLL_ENTRY(fs_alloc_sizeElt_t)   pEntry;
};
#endif // ALLOC_FILE_SIZE_MEDIANE

/* Internal structure for statistics.
 */
typedef struct
{
    boolean                     file; // vs. filesystem
    boolean                     print;
    eeeRTC_t                    beginTime;
    fs_fileStat_t               fileStat;
    fs_fsysStat_t               fsysStat;
#ifdef ALLOC_FILE_SIZE_MEDIANE
    DLL_HDR(fs_alloc_sizeElt_t) fileSizeList;
#endif // ALLOC_FILE_SIZE_MEDIANE
} fs_allocStat_t;

#define FS_ALLOC_PRINT(_stat_, _param_) \
    if (_stat_->print == TRUE)  \
        printf _param_;

/*------------------------------------------------------------------------------
 * Function prototypes.
 *------------------------------------------------------------------------------
 */

static void
fs_getFileExtents(fs_Context *      context,
                  fs_Inode *        inode,
                  fs_allocStat_t *  stat);

int32
fs_allocStatsThread(fs_Context * context);

static void
fs_getAllFileExtents(fs_Context *       context,
                     fs_FileSystem *    fs,
                     int32              snapnum,
                     fs_allocStat_t *   stat);

typedef struct
{
    uint64      offset;
    fs_Extent   extent;
} fs_alloc_extent_t;

static void
fs_alloc_freeSpace(fs_Context *     context,
                   fs_allocStat_t * stat);

static int
fs_alloc_sizeCompare(const void * a,
                     const void * b);

/*------------------------------------------------------------------------------
 * Upon operator's request, we calculate the filesystem fragmentation.
 *----------------------------------------------------------------------------*/
extern boolean
fs_verifyBuddyCount(fs_Context * context, int recalc);

int32
fs_allocStatsThread(fs_Context * context)
{
    fs_MsgAllocStatsReq *   req = (fs_MsgAllocStatsReq *) context->reqHdr;
    fs_MsgAllocStatsResp *  resp = (fs_MsgAllocStatsResp *) context->respHdr;
    fs_allocStat_t          stat;
    fs_FileSystem *         fs = fs_fileSystemForVolume(req->volId);
    FS_ASSERT(fs != NULL);

    // Kick off the brute-force recount of blocks and inodes.
    fs_threadFork(context, fs_refByteCountThread, TRUE, FALSE,
                  FS_DAEMON_TYPE_REF_CNT, NULL);
    
    fs_threadFork(context, fs_inodeCountThread, TRUE, FALSE,
                  FS_DAEMON_TYPE_INO_CNT, NULL);

    /* Reset the statistics.
     */
    bzero(&stat, sizeof(fs_allocStat_t));
    stat.print = req->print;
    eee_rtc(&(stat.beginTime));
    stat.fsysStat.init = TRUE;
#ifdef ALLOC_FILE_SIZE_MEDIANE
    DLL_INIT(&(stat.fileSizeList));
#endif // ALLOC_FILE_SIZE_MEDIANE

    /* Prepare the response.
     */
    context->respSize = sizeof(fs_MsgAllocStatsResp);
    resp->volId = req->volId;
    resp->inum = req->inum;
    resp->snapnum = req->snapnum;

    if (req->inum == 0)
    {
        stat.file = FALSE;

        /* Calculate the filesystem fragmentation.
         */
        fs_getAllFileExtents(context, fs, req->snapnum, &stat);

#ifdef ALLOC_FILE_SIZE_MEDIANE
        /* Calculate the file size mediane and free the list.
         */
        if (stat.fsysStat.nbFile > 0)
        {
            fs_alloc_sizeElt_t *    pDel = NULL;
            uint64 * sizeTab = eee_ramAllocZero(stat.fsysStat.nbFile *
                                                sizeof(uint64));
            int i = 0;
            
            while (!DLL_IS_EMPTY(&(stat.fileSizeList)))
            {
                DLL_REMOVE_HEAD(&(stat.fileSizeList), pDel, pEntry);
                sizeTab[i] = pDel->size;
                i++;
                eee_ramDealloc(pDel);
            } // while
            
            qsort(sizeTab, stat.fsysStat.nbFile, sizeof(uint64),
                  fs_alloc_sizeCompare);
            stat.fsysStat.medianeFileSize = sizeTab[stat.fsysStat.nbFile/2];
            eee_ramDealloc(sizeTab);
        }
#endif // ALLOC_FILE_SIZE_MEDIANE

        /* Calculate the free space fragmentation.
         */
        fs_alloc_freeSpace(context, &stat);

        /* Copy the result in the response.
         */
        memcpy(&(resp->fsysStat), &(stat.fsysStat), sizeof(fs_fsysStat_t));
        memcpy(&(resp->fileStat), &(stat.fileStat), sizeof(fs_fileStat_t));
    }
    else
    {
        fs_Inode * inode = I_GET_SHARED(req->snapnum, req->inum);
        
        /* Calculate the file fragmentation.
         */
        stat.file = TRUE;
        FS_ALLOC_PRINT((&stat), ("\nFile extents:\n"));

        if (!inode)
        {
            printf("inode %lld not found\n", req->inum);
            return NFX_ERR;
        }

        if (!FS_I_IS_REG(inode->dinode.attrs))
        {
            printf("inode %lld is not a regular file\n", inode->inum);
            fs_inodeRelease(context, inode);
            return NFX_ERR;
        }

        fs_getFileExtents(context, inode, &stat);

        /* Copy the result in the response.
         */
        memcpy(&(resp->fileStat), &(stat.fileStat), sizeof(fs_fileStat_t));
        
        fs_inodeRelease(context, inode);
    }

    fs_verifyBuddyCount(context, TRUE);

    return(NFX_OK);
}

/* Compare function used in qsort of files' size list.
 */
static int
fs_alloc_sizeCompare(const void * a, const void * b)
    
{
    const uint64 A = *(uint64 *) a;
    const uint64 B = *(uint64 *) b;

    if (A < B) return(-1);
    else if (A > B) return(1);
    else return(0);
}

/* Go through all inodes of the file sytem (except the reserved inodes),
 * and get the filesystem fragmentation.
 */
static void
fs_getAllFileExtents(fs_Context *       context,
                     fs_FileSystem *    fs,
                     int32              snapnum,
                     fs_allocStat_t *   stat)
{
    fs_Parameters   p = fs_computeSizeParameters(fs->super);
    uint64          num_inodes = p.maxInodeNumber;
    uint64          inum = 0;

    /* Go through all inodes of the filesystem.
     */
    for (inum = FS_MAX_RESERVED_INODE; inum < num_inodes; inum++)
    {
		eeeRTC_t delta;
        fs_Inode * inode = I_GET_SHARED(snapnum, inum);

        if (inode == NULL)
            continue;
        
        if (FS_I_IS_REG(inode->dinode.attrs))
        {
            fs_getFileExtents(context, inode, stat);
        }
        
        fs_inodeRelease(context, inode);

        eee_rtcDuration(&(stat->beginTime), &delta);
        if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&(stat->beginTime));
        }
    } // for
}

/* emap walk worker routine
 */
static boolean
fs_alloc_emap_worker(
    fs_Context *context,
    fs_Inode *inode,
    fs_WalkResult *result,
    void *arg0)
{
    fs_allocStat_t *    stat = (fs_allocStat_t *) arg0;
    fs_fileStat_t *     fileStat = &(stat->fileStat);
    uint64              len;

    if (result->type != FS_WALK_EXTENT_TYPE)
    {
        // don't care about leaf or inner nodes
        return TRUE;
    }

    len = fs_extentLength(result->extent);
    fileStat->fileSize = inode->dinode.size;
    fileStat->nbExtent++;
    fileStat->totalExtentSize += len;
    if (fileStat->init == TRUE)
    {
        fileStat->minExtentSize = len;
        fileStat->maxExtentSize = len;
        fileStat->init = FALSE;
    }
    else if (len < fileStat->minExtentSize)
    {
        fileStat->minExtentSize = len;
    }
    else if (len > fileStat->maxExtentSize)
    {
        fileStat->maxExtentSize = len;
    }
    if (stat->file == TRUE)
    {
        FS_ALLOC_PRINT(stat, ("\toffset %-10lld, start %-10lld, len %-10lld\n",
                              result->offset,
                              (result->extent.lengthAndStartBlock & EXTENT_BLOCK_MASK),
                              len));
    }
    return TRUE;
}

/* Walk through the emap of a particular file to compute its fragmentation.
 */
static void
fs_getFileExtents(fs_Context *      context,
                  fs_Inode *        inode,
                  fs_allocStat_t *  stat)
    
{
    FS_ASSERT(inode != NULL);
    FS_ASSERT(FS_I_IS_REG(inode->dinode.attrs));

    bzero(&(stat->fileStat), sizeof(fs_fileStat_t));
    stat->fileStat.init = TRUE;

    if (inode->dinode.size == 0)
    {
        goto fs_getFileExtents_exit;
    }
    
#ifdef ALLOC_FILE_SIZE_MEDIANE
    /* Add file size in the list.
     */
    if (stat->file == FALSE)
    {
        fs_alloc_sizeElt_t * elt = eee_ramAllocZero(sizeof(fs_alloc_sizeElt_t));

        FS_ASSERT(elt != NULL);
        elt->size = inode->dinode.size;
        DLL_INSERT_TAIL(&(stat->fileSizeList), elt, pEntry);
    }
#endif // ALLOC_FILE_SIZE_MEDIANE
        
    fs_emap_walk(context, inode, fs_alloc_emap_worker, stat);

    stat->fsysStat.nbFragmentedFile += (stat->fileStat.nbExtent > 1)?1:0;
    stat->fsysStat.nbFile++;
    stat->fsysStat.totalFileSize += stat->fileStat.fileSize;
    stat->fsysStat.nbExtent += stat->fileStat.nbExtent;
    stat->fsysStat.totalExtentSize += stat->fileStat.totalExtentSize;
    if (stat->fsysStat.init == TRUE)
    {
        stat->fsysStat.minExtentPerFile = stat->fileStat.nbExtent;
        stat->fsysStat.maxExtentPerFile = stat->fileStat.nbExtent;
        stat->fsysStat.minExtentSize = stat->fileStat.minExtentSize;
        stat->fsysStat.maxExtentSize = stat->fileStat.maxExtentSize;
        stat->fsysStat.inumWithMaxExtent = inode->inum;
        stat->fsysStat.init = FALSE;
    }
    else
    {
        if (stat->fileStat.nbExtent < stat->fsysStat.minExtentPerFile)
        {
            stat->fsysStat.minExtentPerFile = stat->fileStat.nbExtent;
        }
        else if (stat->fileStat.nbExtent > stat->fsysStat.maxExtentPerFile)
        {
            stat->fsysStat.maxExtentPerFile = stat->fileStat.nbExtent;
            stat->fsysStat.inumWithMaxExtent = inode->inum;
        }
        stat->fsysStat.minExtentSize = MIN(stat->fileStat.minExtentSize,
                                           stat->fsysStat.minExtentSize);
        stat->fsysStat.maxExtentSize = MAX(stat->fileStat.maxExtentSize,
                                           stat->fsysStat.maxExtentSize);
    }

  fs_getFileExtents_exit:
    return;
}

/* Go through the buddy pages, to get the statistics of the free space.
 */
#define MAX_BLOCK_IN_BUDDY_REGION   (32*1024)

static void
fs_alloc_freeSpace(fs_Context *     context,
                   fs_allocStat_t * stat)
{
#ifdef BUDDY_INCORE
    uint64          buddyFilePages = context->fileSystem->buddyRegions;
#else // BUDDY_INCORE
    fs_Inode *      buddyFile = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    uint64          buddyFilePages = ceilingDiv64(buddyFile->dinode.size,
                                                  FS_BLOCK_SIZE);
#endif // BUDDY_INCORE
    uint64          page;
    uchar8 *        freeBlockTab = eee_ramAlloc(MAX_BLOCK_IN_BUDDY_REGION *
                                                sizeof(uchar8));
    int             CurrentState = 0;
    uint64          i, lasti = 0;
    uint64          extentSize = 0;
    fs_fileStat_t * freeStat = &(stat->fileStat);
    
    FS_ASSERT(freeBlockTab != NULL);
    bzero(freeStat, sizeof(fs_fileStat_t));
    freeStat->init = TRUE;
    FS_ALLOC_PRINT(stat, ("\nFree space extents:\n"));
    
    for (page = 0; page < buddyFilePages; page++) 
    {
#ifdef BUDDY_INCORE
        fs_Buf *    buf = fs_buddyGetRegionBuf(context, page, FALSE);
#else // BUDDY_INCORE
        fs_Buf *    buf = O_GET_SHARED(buddyFile, page);
#endif // BUDDY_INCORE
        uint32 *    buddy = buf->u.word;
        int         level;
        int         index;
        int         word;
        int         mask;
        uint16      base;
        uint16      loop;
        eeeRTC_t    delta;
        int start;
        
        bzero(freeBlockTab, MAX_BLOCK_IN_BUDDY_REGION * sizeof(uchar8));
        for (level = FS_MAX_BUDDY_LEVELS ; level >= 0; level--) 
        {
            base = 1 << (FS_MAX_BUDDY_LEVELS - level);
            for (loop = 0; loop < base; loop++)
            {
                index = base + loop;
                word = index / BITS_PER_WORD;
                mask = 1 << (index & (BITS_PER_WORD - 1));

                if ((buddy[word] & mask) != 0)
                {
                    int j;
                    
                    for (j = 0; j < (1<<level); j++)
                    {
                        if (freeBlockTab[(j+loop*(1<<level))] != 0)
                        {
                            printf("Aaaaaaaaaarrrggghhhhhhhhh\n");
                            goto fs_alloc_freeSpace_exit;
                        }
                        freeBlockTab[(j+loop*(1<<level))] = 1;
                    }
                }                
            } // for (loop)
        } // for (level)
        fs_bufRelease(context, buf);

        start = 0;
        //Calculate stats
        if (page == 0)
        {
            // init
            CurrentState = freeBlockTab[0];
            FS_ALLOC_PRINT(stat, ("%10lld - ", 0LL));
            start = 1; // skip first block of first page
            freeStat->totalExtentSize = freeBlockTab[0]; // init total size
        }
        for (i = start; i < MAX_BLOCK_IN_BUDDY_REGION; i++)
        {
            if (freeBlockTab[i] != CurrentState)
            {
                FS_ALLOC_PRINT(stat, ("%10lld : %s\n",
                                      (page*MAX_BLOCK_IN_BUDDY_REGION)+i-1,
                                      freeBlockTab[i-1]? "free" : "full"));
                if (freeBlockTab[i-1] == 1)
                {
                    freeStat->nbExtent++;
                    extentSize = i - lasti;
                    if (freeStat->init == TRUE)
                    {
                        freeStat->init = FALSE;
                        freeStat->minExtentSize = extentSize;
                    }
                    else
                    {
                        freeStat->minExtentSize = MIN(extentSize,
                                                      freeStat->minExtentSize);
                    }
                    freeStat->maxExtentSize = MAX(extentSize,
                                                  freeStat->maxExtentSize);
                }
                CurrentState = freeBlockTab[i];
                FS_ALLOC_PRINT(stat, ("%10lld - ",
                                      (page*MAX_BLOCK_IN_BUDDY_REGION)+i));
                lasti = i;
            }
            if (freeBlockTab[i] == 1)
            {
                freeStat->totalExtentSize++;
            }
        } // for (i)
        lasti -= MAX_BLOCK_IN_BUDDY_REGION;
        
        eee_rtcDuration(&(stat->beginTime), &delta);
        if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&(stat->beginTime));
        }
    } // for (page)

    FS_ALLOC_PRINT(stat, ("%10lld : %s\n",
                          (buddyFilePages * MAX_BLOCK_IN_BUDDY_REGION),
                          CurrentState? "free" : "full"));

  fs_alloc_freeSpace_exit:

#ifndef BUDDY_INCORE
    fs_inodeRelease(context, buddyFile);
#endif // BUDDY_INCORE
    eee_ramDealloc(freeBlockTab);
    return;
}

/*----------------------------------------------------------------------------*/
