// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-alloc.h
 *
 * RCS:  $Id: fs-alloc.h,v 1.47 2002/12/05 21:08:57 bob Exp $
 *
 * Description:  Filesystem allocation interface.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-06-22
 */

#if !defined(FS_ALLOC_H)
#define FS_ALLOC_H


/**
 * In-core shadow bitmaps of block/inode allocation.
 */
#if defined(FS_DEBUG_ALLOC)

#define FS_DBG_SHADOW_IS_SET(_ptr,_num)                             \
    (*((_ptr) + ((_num) / 8)) & (1 << ((_num) % 8)))
#define FS_DBG_SHADOW_SET(_ptr,_num)                                \
    *((_ptr) + ((_num) / 8)) |= 1 << ((_num) % 8)
#define FS_DBG_SHADOW_CLR(_ptr,_num)                                \
    *((_ptr) + ((_num) / 8)) &= ~(1 << ((_num) % 8))
#define FS_DBG_SHADOW_ALLOC(_ptr,_num)                              \
    FS_ASSERT(!FS_DBG_SHADOW_IS_SET((_ptr),(_num)));                \
    FS_DBG_SHADOW_SET((_ptr),(_num));
#define FS_DBG_SHADOW_FREE(_ptr,_num)                               \
    FS_ASSERT(FS_DBG_SHADOW_IS_SET((_ptr),(_num)));                 \
    FS_DBG_SHADOW_CLR((_ptr),(_num));
#define FS_DBG_SHADOW_ALLOC_CHECK_BLOCK(_blknum)                    \
    FS_DBG_SHADOW_IS_SET(context->fileSystem->blockAllocShadow,     \
                        (_blknum))
#define FS_DBG_SHADOW_ALLOC_BLOCK(_blknum)                          \
    NFX_SPIN_LOCK(&context->fileSystem->spin);                      \
    FS_DBG_SHADOW_ALLOC(context->fileSystem->blockAllocShadow,      \
                        (_blknum));                                 \
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);
#define FS_DBG_SHADOW_FREE_BLOCK(_blknum)                           \
    NFX_SPIN_LOCK(&context->fileSystem->spin);                      \
    FS_DBG_SHADOW_FREE(context->fileSystem->blockAllocShadow,       \
                        (_blknum));                                 \
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);
#define FS_DBG_SHADOW_ALLOC_EXTENT(_extent)                         \
    { int i;                                                        \
        for (i = 0; i < fs_extentLength(_extent); i++) {            \
            FS_DBG_SHADOW_ALLOC_BLOCK(fs_blockNumber(               \
                                    fs_extentBlock(_extent, i)));   \
        }                                                           \
    }
#define FS_DBG_SHADOW_FREE_EXTENT(_extent)                          \
    { int i;                                                        \
        for (i = 0; i < fs_extentLength(_extent); i++) {            \
            FS_DBG_SHADOW_FREE_BLOCK(fs_blockNumber(                \
                                    fs_extentBlock(_extent, i)));   \
        }                                                           \
    }
#define FS_DBG_SHADOW_ALLOC_CHECK_INODE(_inum)                      \
    FS_DBG_SHADOW_IS_SET(context->fileSystem->inodeAllocShadow,     \
                        (_inum))
#define FS_DBG_SHADOW_ALLOC_INODE(_inum)                            \
    NFX_SPIN_LOCK(&context->fileSystem->spin);                      \
    FS_DBG_SHADOW_ALLOC(context->fileSystem->inodeAllocShadow,      \
                        (_inum));                                   \
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);
#define FS_DBG_SHADOW_FREE_INODE(_inum)                             \
    NFX_SPIN_LOCK(&context->fileSystem->spin);                      \
    FS_DBG_SHADOW_FREE(context->fileSystem->inodeAllocShadow,       \
                        (_inum));                                   \
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);

#else
#define FS_DBG_SHADOW_IS_SET(_ptr,_num)
#define FS_DBG_SHADOW_SET(_ptr,_num)
#define FS_DBG_SHADOW_CLR(_ptr,_num)
#define FS_DBG_SHADOW_ALLOC(_ptr,_num)
#define FS_DBG_SHADOW_FREE(_ptr,_num)
#define FS_DBG_SHADOW_ALLOC_BLOCK(_blknum)
#define FS_DBG_SHADOW_FREE_BLOCK(_blknum)
#define FS_DBG_SHADOW_ALLOC_EXTENT(_blknum)
#define FS_DBG_SHADOW_FREE_EXTENT(_blknum)
#define FS_DBG_SHADOW_ALLOC_INODE(_inum)
#define FS_DBG_SHADOW_FREE_INODE(_inum)
#endif


//   B L O C K S   &   E X T E N T S


/**
 * Allocate an extent for the file with the specified inode, of
 * approximately to the specified length in blocks, starting at the given
 * page within the file.  If the anySize parameter is true, the returned
 * extent may be larger than requested; otherwise, it will be at most the
 * requested size.
 *
 * No distinction is made between allocating extents for user files
 * vs. metadata files, because this call just reserves address space.
 * Assignment of blocks to buffers and allocation of buffers are
 * separate concerns.
 */
extern fs_Extent fs_allocateExtent(
    fs_Context *context,
    fs_Inode *inode,
    uint64 page,
    uint32 len32,
    boolean anySize);


/**
 * Allocate a block for btree metadata, associated with the specified
 * inode.  The allocator will try to place the block near the rest of
 * the metadata for the inode.
 */
extern fs_Block fs_allocateMetadataBlock(
    fs_Context *context,
    fs_Inode *inode);


/**
 * Allocate a block for btree metadata, associated with the specified
 * inode number.  The allocator will try to place the block near the rest
 * of the metadata for the inode.
 *
 * This is a special interface (hopefully temporary) to allow the btree
 * copy-on-write code to avoid needing to know the inode that it's using.
 * This will probably go away when we have deferred metadata allocation.
 */
extern fs_Block fs_allocateMetadataBlockSpecial(
    fs_Context *context,
    fs_Inode   *inode,
    fs_Block   target);


/**
 * Allocate a block for storing the block of the inode file containing
 * the specified (byte) offset into the inode file.  These blocks are
 * distributed throughout the filesystem
 */
extern fs_Block fs_allocateInodeFileBlock(
    fs_Context *context,
    fs_Inode *inode,
    uint64 offset);


/**
 * Deallocate the specified block.  It is returned to the allocator
 * for reuse.
 */
extern void fs_deallocateBlock(
    fs_Context *context,
    fs_Block block);


/**
 * Deallocate all the blocks in the specified extent.
 */
extern void fs_deallocateExtent(
    fs_Context *context,
    fs_Extent extent);



//   I N O D E S


/**
 * Allocate a disk inode (and inode number) for a file to be created in
 * the specified directory.  The inode's block will be near the directory
 * inode's block, if possible.  If the specified directory is NULL, an
 * arbitrary inode number will be chosen.
 *
 * The mode, uid, gid, and filesetid of the new inode will be as
 * specified.  The ctime, mtime, and atime for the file will be the
 * current time.
 *
 * The returned inode is returned with an exclusive lock.
 *
 * Returns NULL if the file system has no available inodes.
 */
extern fs_Inode *fs_allocateFileInode(
    fs_Context *context,
    fs_Inode *directory,
    uint16 mode,
    uint32 uid,
    uint32 gid,
    uint32 filesetid);


/**
 * Allocate a disk inode (and inode number) for a directory to be created
 * in the given parent directory.  For directories near the root of the
 * file system, subdirectories are created in arbitrary disk locations, so
 * as to spread them across the disk;  directories further from the root
 * are created near their parents, if possible, in inode number.
 *
 * The mode, uid, gid, and filesetid of the new inode will be as
 * specified.  The ctime, mtime, and atime for the file will be the
 * current time.
 *
 * The returned inode is returned with an exclusive lock.
 *
 * Returns NULL if the file system has no available inodes.
 */
extern fs_Inode *fs_allocateDirectoryInode(
    fs_Context *context,
    fs_Inode *parent,
    uint16 perm,
    uint32 uid,
    uint32 gid,
    uint32 filesetid);


/**
 * Deallocate an inode number.
 */
extern void fs_deallocateInode(
    fs_Context *context,
    fs_Inode   *inode);



//   F I L E   S Y S T E M   F O R M A T T I N G


/**
 * Create a file system on the volume specified by the context argument,
 * with the given number of blocks and block size.
 * Return TRUE on success FALSE on error.
 */
extern boolean fs_formatFileSystem(
    fs_Context  *context,
    uint64      numBlocks,
    uint32      blockSize,
    boolean     extLog);

/**
 * Allocate an extent at boot time.
 */
extern fs_Extent fs_bootAllocateExtent(
    fs_Context  *context,
    fs_Inode    *inode,
    uint16      len);


/**
 * Deallocate all extents on the pending free list for this context.
 * Done at commit time.
 */
extern void fs_commitDeallocate(
    fs_Context *context);

/**
 * Finish deallocation of the specified inode.
 * Done at commit time.
 */
extern inline void fs_commitInodeDeallocate(
    fs_Context  *context,
    fs_Inode    *inode);


extern inline void fs_incrementBlockUse(
    fs_Context  *context,
    int numBlocks);

extern inline void fs_decrementBlockUse(
    fs_Context  *context,
    int numBlocks);

extern inline void fs_incrementInodeUse(
    fs_Context  *context,
    uint64 inum);

extern inline void fs_decrementInodeUse(
    fs_Context  *context,
    uint64 inum);

/**
 * Recalculate the mininum number of free blocks
 * we reserve for the specified filesystem.
 */
extern void
fs_recalcMinFreeBlks(fs_Context *context);

extern inline boolean fs_haveMinfreeAvail(fs_FileSystem *fs);
extern inline boolean fs_haveMininodesAvail(fs_FileSystem *fs);
extern inline uint64 fs_numInodesAvail(fs_FileSystem *fs);
extern inline uint64 fs_numInodesUserMax(fs_FileSystem *fs);
extern inline uint64 fs_numBlocksAvail(fs_FileSystem *fs);
extern inline uint64 fs_numBlocksUserMax(fs_FileSystem *fs);

#endif /* FS_ALLOC_H */
