/*
 * Name		: fs-am.c
 * RCS		: $ID$
 * Description	: FS-Agile Mirror interface
 * Created by	: Rahul Khona
 * Date Created	: 18/01/2002
 */

#include "fs-includes.h"

int32 fs_snapClean(fs_Context *context);
int32 fs_createSnapList(fs_Context *context);

/**
 * Setup the Agile Mirror Delta process. On successful completion
 */

int32
fs_amDeltaSetupThread(
    fs_Context      *context)
{
    fs_AmSourceSession_t    *session;
    int                     oldSnapNum = -1;
    int                     newSnapNum;
    char                    *newName, *oldName;
    int32   ret = NFX_OK;
    fs_MsgAmDeltaSetupReq_t *req;
    fs_MsgAmDeltaSetupResp_t *resp;

    req = (fs_MsgAmDeltaSetupReq_t *)context->reqHdr;
    resp = (fs_MsgAmDeltaSetupResp_t *)context->respHdr;
    resp->cookie = req->cookie;

	E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("FS DeltaSetupThread newName %s, oldName %s\n", req->newName, 
			(strlen(req->oldName) > 0) ? req->oldName : "None" ));

    newName = req->newName;
    oldName = req->oldName;
    FS_ASSERT(strlen(newName) > 0);

    newSnapNum = fs_getSnapNum(context, newName);
    if (newSnapNum != -1) {

        /*
         * Snapshot already exists. Remove this one
		 * it may have been left from an aborted operation
		 * We need to recreate it
         */
		if (fs_snapRemove(context, newName) == NFX_ERR)
		{
        	resp->hdr.fsStatus = FS_ERR_EXIST;
        	ret = NFX_ERR;
        	return ret;
		}
    }

    if (strlen(oldName) > 0) {
        oldSnapNum = fs_getSnapNum(context, oldName);
        if (oldSnapNum == -1) {
            resp->hdr.fsStatus = FS_ERR_NOENT;
            ret = NFX_ERR;
            return ret;
        }
    }

    session = fs_ramAlloc(FS_RAM_AM_SOURCE_SESSION, 
                          sizeof (fs_AmSourceSession_t));
    FS_ASSERT(session != NULL);
    session->cookie = req->cookie;
    session->oldSnapNum = oldSnapNum;
    session->newSnapNum = newSnapNum;
    session->copiedSnode = 0;   
    session->newName = fs_ramAlloc(FS_RAM_SNAP_NAME, sizeof (newName) + 1);
    session->oldName = fs_ramAlloc(FS_RAM_SNAP_NAME, sizeof (oldName) + 1);
    strcpy(session->newName, newName);
    strcpy(session->oldName, oldName);
    session->done = FALSE;

	printf("context %p fileSystem %p amSourceSession %p\n",
           context, context->fileSystem, 
           context->fileSystem->amSourceSession);

	/*
	 * We don't need any AM snap other than the one
	 * target has. Clean up any unneeded AM snaps
	 */
    context->fileSystem->amSourceSession = session;
	fs_snapClean(context);

    /*
     * Now create the snapshot with the new name
     */
    context->fileSystem->amSourceSession = session;
    ret = fs_snapCreate(context, newName);
    if (ret != NFX_OK) {
        fs_ramDealloc(FS_RAM_SNAP_NAME, session->newName);
        fs_ramDealloc(FS_RAM_SNAP_NAME, session->oldName);
        fs_ramDealloc(FS_RAM_AM_SOURCE_SESSION, session);
        resp->hdr.fsStatus = FS_ERR_GENERIC;
		context->fileSystem->amSourceSession = NULL;
        return ret;
    }

    session->newSnapNum = fs_getSnapNum(context, newName);

	/*
	 * Create a list of all snaps that were generated 
	 * after last snap (session->oldName) + newSnapNum
	 * This is all the stuff we will need to transfer
	 */
    context->fileSystem->amSourceSession = session;
	fs_createSnapList(context);

    resp->snapNum = session->newSnapNum;

	E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("FS DeltaSetupThread Will revert to %s (%d)\n", 
					newName, resp->snapNum));

	printf("FS DeltaSetupThread Will revert to %s (%d)\n", 
					newName, resp->snapNum);

    context->fileSystem->amSourceSession = session;
    session->fs = context->fileSystem;
    fs_syncThread(context);
    resp->hdr.fsStatus = FS_ERR_NONE;
    return ret;
}

/**
 * Get the next list of extents that need to be mirrored to the
 * target system. We will return 1 or more extents, but never more
 * than maximum number of extents asked for. If we there aren't
 * any extents available, then we will return FS_ERR_NOENT. If
 * the extent map is successfully read, we will return FS_ERR_NONE.
 * We will return FS_ERR_IO for i/o.
 */

PRIVATE fs_Status
fs_findExtent(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          blkNum,
    uint64          *start,
    uint32          *len,
    int             value
)
{
    uint64                          page;
    uint32                          byte, boff;
    uint8                           bit, tbit;
    char                            *new = NULL;
    uint64                          startBlk;
    uint32                          runLen;
    fs_Buf                          *buf;
    /*
     * Page in snapbit file corresponding to the blknum
     */

    page = blkNum >> (BITS_PER_BYTE_SHIFT + FS_BLOCK_SHIFT);

    /*
     * Byte in the snapbit file
     */

    byte = (blkNum >> BITS_PER_BYTE_SHIFT);

    /*
     * Bit within the byte corresponding to the blkNum
     */

    bit = blkNum & (BITS_PER_BYTE - 1);

    if (byte > inode->dinode.size) {
        return FS_ERR_NOENT;
    }

    boff = byte & (FS_BLOCK_SIZE - 1);

    /*
     * Now get the buffer corresponding to the page
     */

read_next_block:
    buf = O_GET_SHARED(inode, page);
    FS_ASSERT(buf != NULL);
    
    new = &buf->u.data[boff];

    /*
     * Now find sequence of blocks starting at bit in new that
     * are set. Don't cross block boundary
     */

    for (tbit = bit; tbit < BITS_PER_BYTE; tbit++) {
        if (value) {
            if ((*new >> tbit) & 0x01) {
                    startBlk =
                        (page << FS_BLOCK_SHIFT) +
                        (byte << BITS_PER_BYTE_SHIFT) + tbit;
                    tbit++;
                    goto find_run;
            }
        } else {
            if (((*new >> tbit) & 0x01) == 0) {
                    startBlk =
                        (page << FS_BLOCK_SHIFT) +
                        (byte << BITS_PER_BYTE_SHIFT) + tbit;
                    tbit++;
                    goto find_run;
            }
        }
    }
	byte++;
	new++;

    /*
     * Look at next byte
     */

    while (byte < FS_BLOCK_SIZE) {
        bit = 0;
        for (tbit = bit; tbit < BITS_PER_BYTE; tbit++) {
            if (value) {
                if ((*new >> tbit) & 0x01) {
                        startBlk =
                            (page << FS_BLOCK_SHIFT) +
                            (byte << BITS_PER_BYTE_SHIFT) + tbit;
                        tbit++;
                        goto find_run;
                }
            } else {
                if (((*new >> tbit) & 0x01) == 0) {
                        startBlk =
                            (page << FS_BLOCK_SHIFT) +
                            (byte << BITS_PER_BYTE_SHIFT) + tbit;
                        tbit++;
                        goto find_run;
                }
            }
        }
        byte++;
        new++;
    }

    /*
     * No blocks in this page are set. Read next block
     */

    page++;
    byte = page << FS_BLOCK_SHIFT;
    bit = 0;
    fs_bufRelease(context, buf);
    if (inode->dinode.size > byte) {
        goto read_next_block;
    }

    return FS_ERR_NOENT;

find_run:

    runLen = 1;
    for (; tbit < BITS_PER_BYTE; tbit++) {
        if (value) {
            if ((*new >> tbit) & 0x01) {
                    runLen++;
                 //   tbit++;
					continue;
            }
			break;
        } else {
            if (((*new >> tbit) & 0x01) == 0) {
                    runLen++;
                 //   tbit++;
					continue;
            }
			break;
        }
    }

    if (tbit == BITS_PER_BYTE) {
        byte++;
        new++;
        goto find_len;
    }

    /*
     * We found an extent of exactly runLen block
     */

    fs_bufRelease(context, buf);
    *start = startBlk;
    *len = runLen;
    return FS_ERR_NONE;

find_len:
    while (byte < FS_BLOCK_SIZE) {
        bit = 0;
        for (tbit = bit; tbit < BITS_PER_BYTE; tbit++) {
            if (value) {
                if ((*new >> tbit) & 0x01) {
                        runLen++;
						continue;
                     //   tbit++;
                }
				break;
            } else {
                if (((*new >> tbit) & 0x01) == 0) {
                     //   tbit++;
                        runLen++;
						continue;
                }
				break;
            }
        }

        if (tbit != BITS_PER_BYTE) {
            break;
        }
        byte++;
        new++;
    }
    
    fs_bufRelease(context, buf);
    *start = startBlk;
    *len = runLen;
    return FS_ERR_NONE;
}

int32
fs_amDeltaGetExtentThread(
    fs_Context      *context)
{
    fs_MsgAmDeltaGetExtentReq_t     *req;
    fs_MsgAmDeltaGetExtentResp_t    *resp;
    uint64                          blkNum;
    fs_Inode                        *oldSnapBitInode = NULL;
    fs_Inode                        *newSnapBitInode;
    fs_AmSourceSession_t            *session;
    fs_Status                       status;
    uint64                          newSnapStartBlk, oldSnapStartBlk;
    uint32                          newSnapRunLen, oldSnapRunLen;
	fs_Block						b;	
	uint32							setIdx;

    req = (fs_MsgAmDeltaGetExtentReq_t *)context->reqHdr;
    resp = (fs_MsgAmDeltaGetExtentResp_t *)context->respHdr;
    session = context->fileSystem->amSourceSession;

    FS_ASSERT(session != NULL && session->cookie == req->cookie);

    resp->cookie = req->cookie;
    blkNum = req->fsCookie;
    resp->numExtents = 0;

    /*
     * Return the pointer to the snodes for the snapshot
     */

    if (session->copiedSnode < FS_MAX_SNAPSHOTS) {
        b = fs_firstInodeBlock(session->copiedSnode);
        session->copiedSnode++;
        resp->extents[resp->numExtents].blknum = b.block;
        resp->extents[resp->numExtents++].len = 1;
        resp->fsCookie = 1;
        resp->hdr.fsStatus = FS_ERR_NONE;
        return NFX_OK;
    }

    if (session->done && session->numCurSet == session->numMaxSets) {
        resp->hdr.fsStatus = FS_ERR_NOENT;
        resp->fsCookie = req->fsCookie;
        return resp->hdr.fsStatus;
    }

    newSnapBitInode = I_GET_SHARED(session->newSnapNum, FS_SNAPSHOT_BITS_INODE);
    if (session->oldSnapNum != -1) {
        oldSnapBitInode = I_GET_SHARED(session->oldSnapNum,
                                        FS_SNAPSHOT_BITS_INODE);
    }

    do {

        status = fs_findExtent(context, newSnapBitInode, blkNum,
                               &newSnapStartBlk, &newSnapRunLen, 1);
        if (status == FS_ERR_NOENT) {
			if (session->numCurSet < session->numMaxSets) {
				session->numCurSet++;
				session->setIdx++;
				setIdx = session->setIdx;

				session->oldSnapNum = session->snapInfo[setIdx - 1].id;
				strcpy(session->oldName, session->snapInfo[setIdx - 1].name);

				session->newSnapNum = session->snapInfo[setIdx].id;
				strcpy(session->newName, session->snapInfo[setIdx].name);

            	if (oldSnapBitInode) 
                	fs_inodeRelease(context, oldSnapBitInode);
                fs_inodeRelease(context, newSnapBitInode);

				//session->copiedSnode = 0;
				blkNum = 1;

				newSnapBitInode = I_GET_SHARED(session->newSnapNum, 
											FS_SNAPSHOT_BITS_INODE);
				oldSnapBitInode = I_GET_SHARED(session->oldSnapNum,
											FS_SNAPSHOT_BITS_INODE);

				E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
					("Next set Old: %s New: %s CurSet %d MaxSet = %d\n", 
							session->oldName, session->newName, 
								session->numCurSet, session->numMaxSets));

				printf("Next set Old: %s New: %s CurSet %d MaxSet = %d\n", 
							session->oldName, session->newName, 
								session->numCurSet, session->numMaxSets);
				continue;
			}
            session->done = TRUE;

            if (resp->numExtents == 0) {
                resp->hdr.fsStatus = FS_ERR_NOENT;
                resp->fsCookie = req->fsCookie;
            } else {
                resp->hdr.fsStatus = FS_ERR_NONE;
            }
            fs_inodeRelease(context, newSnapBitInode);
            if (oldSnapBitInode) {
                fs_inodeRelease(context, oldSnapBitInode);
            }
            return resp->hdr.fsStatus; 
        }

        if (oldSnapBitInode) {
            status = fs_findExtent(context, oldSnapBitInode, newSnapStartBlk,
                                   &oldSnapStartBlk, &oldSnapRunLen, 0);
            if (status == FS_ERR_NOENT ||
                (oldSnapStartBlk >= (newSnapStartBlk + newSnapRunLen))) {

                /*
				 * All the extents belong to both the snapshots
                 */

                goto ignore_range;
            }

            resp->extents[resp->numExtents].blknum = oldSnapStartBlk;
            if ((oldSnapStartBlk + oldSnapRunLen) < 
                (newSnapStartBlk + newSnapRunLen)) {
                resp->extents[resp->numExtents].len = oldSnapRunLen;
            } else {
                resp->extents[resp->numExtents].len =
                    newSnapRunLen - (oldSnapStartBlk - newSnapStartBlk);
            }
            blkNum = resp->fsCookie = 
            resp->extents[resp->numExtents].blknum + 
                resp->extents[resp->numExtents].len;
            resp->numExtents++;
            continue;
        } else {
				resp->hdr.fsStatus = FS_ERR_NONE;
				resp->extents[resp->numExtents].blknum = newSnapStartBlk;
				resp->extents[resp->numExtents++].len = newSnapRunLen;
				resp->fsCookie = newSnapStartBlk + newSnapRunLen;
				blkNum = resp->fsCookie;
				continue;
		}
ignore_range:
#if 0
        resp->hdr.fsStatus = FS_ERR_NONE;
        resp->extents[resp->numExtents].blknum = newSnapStartBlk;
        resp->extents[resp->numExtents++].len = newSnapRunLen;
        resp->fsCookie = newSnapStartBlk + newSnapRunLen;
        blkNum = resp->fsCookie;
#endif
		blkNum = newSnapStartBlk + newSnapRunLen;
        continue;

    } while (resp->numExtents < req->maxExtents);

	resp->hdr.fsStatus = FS_ERR_NONE;
    fs_inodeRelease(context, newSnapBitInode);
    if (oldSnapBitInode) {
        fs_inodeRelease(context, oldSnapBitInode);
    }
    return resp->hdr.fsStatus; 
}
                           
/**
 * Mirror is done, tear down all the structures
 */

int32
fs_amDeltaCompleteThread(
    fs_Context      *context)
{
    int32   ret = NFX_OK;
    fs_MsgAmDeltaCompleteReq_t  *req;
    fs_MsgAmDeltaCompleteResp_t  *resp;
    fs_AmSourceSession_t    *session = context->fileSystem->amSourceSession;

    req = (fs_MsgAmDeltaCompleteReq_t *)context->reqHdr;
    resp = (fs_MsgAmDeltaCompleteResp_t *)context->respHdr;

    FS_ASSERT(session != NULL && session->cookie == req->cookie);
    context->fileSystem->amSourceSession = NULL;

    fs_ramDealloc(FS_RAM_SNAP_NAME, session->newName);
    fs_ramDealloc(FS_RAM_SNAP_NAME, session->oldName);
    fs_ramDealloc(FS_RAM_AM_SOURCE_SESSION, session);

    resp->hdr.fsStatus = FS_ERR_NONE;
    
    return ret;
}

/**
 * Setup replication on the target filesystem. Filesystem will
 * be frozen on return to ensure that on nfs operations can proceed.
 *
 * We will not protect against multiple agile mirror sessions running
 * against same fs. AM will take care of it.
 */

int32
fs_amDeltaInstallSetupThread(
    fs_Context      *context)
{
    int32   ret = NFX_OK;
    fs_AmTargetSession_t    *session;
    int                     newSnapNum;
    char                    *newName;
    fs_MsgAmDeltaInstallSetupReq_t *req;
    fs_MsgAmDeltaInstallSetupResp_t *resp;

    req = (fs_MsgAmDeltaInstallSetupReq_t *)context->reqHdr;
    resp = (fs_MsgAmDeltaInstallSetupResp_t *)context->respHdr;
    resp->cookie = req->cookie;

    newName = req->newName;
    FS_ASSERT(strlen(newName) > 0);

    newSnapNum = fs_getSnapNum(context, newName);
    if (newSnapNum != -1) {

        /*
         * Snapshot already exists. Return an error
         */

        resp->hdr.fsStatus = FS_ERR_EXIST;
        ret = NFX_ERR;
        return ret;
    }

    session = fs_ramAlloc(FS_RAM_AM_TARGET_SESSION, 
                          sizeof (fs_AmTargetSession_t));
    session->cookie = req->cookie;

    session->snapNum = req->snapNum;
    
    session->newName = fs_ramAlloc(FS_RAM_SNAP_NAME, sizeof (newName) + 1);
    session->fs = context->fileSystem;
    context->fileSystem->amTargetSession = session;
    strcpy(session->newName, newName);

    return ret;
}

/**
 * Install one ore more extents on the target. We will mark
 * the bits live. We will trust Agile Mirror agent to send
 * us valid list only. Authentication is done by the Agile Mirror
 *
 * On a crash during this process, the whole process will have
 * to be restarted. The filesystem will be in an incosistent
 * state till InstallCommit is called as the live bits get updated
 * with each call to this function.
 */

int32
fs_amDeltaInstallExtentThread(
    fs_Context      *context)
{
    int32   ret = NFX_OK;
    fs_MsgAmDeltaInstallExtentResp_t *resp;

    resp = (fs_MsgAmDeltaInstallExtentResp_t *)context->respHdr;
#if 0
    fs_MsgAmDeltaInstallExtentReq_t *req;
    uint32  numExtents;
    uint32  i;
    uint64  block;
    fs_AMExtent_t   *extent;

    req = (fs_MsgAmDeltaInstallExtentReq_t *)context->reqHdr;
    resp = (fs_MsgAmDeltaInstallExtentResp_t *)context->respHdr;
    resp->cookie = req->cookie;
    
//    resp->extentCookie = req->extentCookie;

    numExtents = req->numExtents;
    for (i = 0; i < numExtents; i++) {
        extent = &req->extents[i];
        block = extent->blknum;
        if (!fs_refByteModifyExtent(context, fs_refbyteSetLive,
                                fs_makeExtent(extent->len,
                                              fs_makeBlock(block)))) {
            resp->hdr.fsStatus = FS_ERR_GENERIC;
            ret = NFX_ERR;
            return ret;
        };
    }
    
#endif
    resp->hdr.fsStatus = FS_ERR_NONE;
    return ret;
}

/*
 * The live filesystem bits are all marked. Filesystem is still
 * frozen at this point. Create the snapshot and thaw the filesystem
 */

int32
fs_amDeltaInstallCommitThread(
    fs_Context      *context)
{
    int32   ret = NFX_OK;
    fs_MsgAmDeltaInstallCommitReq_t *req;
    fs_MsgAmDeltaInstallCommitResp_t *resp;

    req = (fs_MsgAmDeltaInstallCommitReq_t *)context->reqHdr;
    resp = (fs_MsgAmDeltaInstallCommitResp_t *)context->respHdr;

    resp->cookie = req->cookie;
    resp->hdr.fsStatus = FS_ERR_NONE;
    
    /*
     * Do it the second time around, so that any buffers that were
     * instantiated and during block transfer from AM and are incosistent
     * with disk would get purged
     */

    fs_syncThread(context);
    fs_inodeCacheFlush(context);
    fs_bufCacheFlush(context, NULL);


    /*
     * Now create the new snapshot
     */

    /*
     * XXX - Place holder for snapRevert
     */

    ret = fs_snapRevert(context, 
				context->fileSystem->amTargetSession->snapNum);

	E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("Target FS mirror revert for %s (%d) completed\n", 
			context->fileSystem->amTargetSession->newName,
			context->fileSystem->amTargetSession->snapNum));

	printf("Target FS mirror completed. Reverted to %s (%d)\n", 
			context->fileSystem->amTargetSession->newName,
			context->fileSystem->amTargetSession->snapNum);

    fs_ramDealloc(FS_RAM_SNAP_NAME,
                  context->fileSystem->amTargetSession->newName);
    fs_ramDealloc(FS_RAM_AM_TARGET_SESSION,
                  context->fileSystem->amTargetSession);
    context->fileSystem->amTargetSession = NULL;

    return ret;
}

/**
 * Returns the name of the latest snapshot
 */

int32
fs_amLastSnapshotThread(
    fs_Context      *context)
{
    char    *names[FS_MAX_SNAPSHOTS];
    char    *nameBuf;
    int32   ret = NFX_OK;
    int     nSnapshots;
    uint32  biggestTime = 0, timeStamp;
    int     latestSnapIndex = 0;
    int     i;
    fs_MsgAmLastSnapshotResp_t  *resp;
    fs_MsgAmLastSnapshotReq_t   *req;

    resp = (fs_MsgAmLastSnapshotResp_t *)(context->respHdr);
    req = (fs_MsgAmLastSnapshotReq_t *)(context->reqHdr);
    resp->cookie = req->cookie;
    resp->name[0] = '\0';
    nameBuf = fs_ramAlloc(FS_RAM_AM_LAST_SNAP,
                          (FS_MAX_SNAPSHOTS*FS_SNAPNAME_MAX));
    FS_ASSERT(nameBuf != NULL);
    nSnapshots = fs_snapGetNames(context, names, nameBuf,
                                 (FS_MAX_SNAPSHOTS*FS_SNAPNAME_MAX));

    if (nSnapshots == 0) {
       goto ret_name;
    }
    for (i = 0; i < nSnapshots; i++) {
        timeStamp = fs_snapGetTime(context, names[i]);
        if (timeStamp > biggestTime) {
            biggestTime = timeStamp;
            latestSnapIndex = i;
        }
    }

    strcpy(resp->name, names[latestSnapIndex]);
ret_name:
    fs_ramDealloc(FS_RAM_AM_LAST_SNAP, nameBuf);
    return ret;
}

/**
 * Comparison function for snap times
 */
static int
fs_CompareSnapTime(
    const void *p0,
	const void *p1)
{
	const fs_AmSnapInfo *f0 = (fs_AmSnapInfo *) p0;
	const fs_AmSnapInfo *f1 = (fs_AmSnapInfo *) p1;

	if (f0->snapTime < f1->snapTime)
		return -1;

	return 1;
}

/**
 * Creates a list of all the snap names that will be needed
 * to be included in the transfer if the existing base is 
 * based on snap snapName
 */
int32
fs_createSnapList(fs_Context *context)
{
    char    *names[FS_MAX_SNAPSHOTS];
    char    *nameBuf;
    int     i;
    int     nSnapshots;
    uint32  timeStamp, myTimeStamp;
    fs_AmSourceSession_t    *session = context->fileSystem->amSourceSession;
	char 	*snapName = session->oldName;
	int		prefixLen = strlen(snapName) - 3;
	uint32	snapIdx;
	uint32	setIdx;

	if (strncmp("SANM_SS", snapName, 7) != 0)
	{
		E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
				("Entered fs_createSnapList with no snapName\n"));
		snapName = NULL;
		prefixLen = 0;
        FS_ASSERT(session->oldSnapNum == -1);
		session->numCurSet = 1;
		session->numMaxSets = 1;
		session->setIdx = 0;
		setIdx = session->setIdx;
        session->snapInfo[setIdx].id = session->newSnapNum;
        strcpy(session->snapInfo[setIdx].name, session->newName);
        return NFX_OK;
	}
	else
	{
		E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
				("Entered fs_createSnapList snapName %s\n", snapName));
	}

    nameBuf = fs_ramAlloc(FS_RAM_AM_LAST_SNAP,
                          (FS_MAX_SNAPSHOTS*FS_SNAPNAME_MAX));
    FS_ASSERT(nameBuf != NULL);

    nSnapshots = fs_snapGetNames(context, names, nameBuf,
                                 (FS_MAX_SNAPSHOTS*FS_SNAPNAME_MAX));

	E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("fs_createSnapList got %d snapnames\n", nSnapshots));

    if (nSnapshots == 0) {
		fs_ramDealloc(FS_RAM_AM_LAST_SNAP, nameBuf);
		return NFX_OK;
    }

	/*
	 * Put in the list all snap shots created after
	 * the last snap shot (snapName/session->oldName)
	 */

	snapIdx = 0;
	if (snapName)
		myTimeStamp = fs_snapGetTime(context, snapName);
	else
		myTimeStamp = 0;

    for (i = 0; i < nSnapshots; i++) 
	{
        timeStamp = fs_snapGetTime(context, names[i]);

        if (timeStamp >= myTimeStamp)
		{
		    strcpy(session->snapInfo[snapIdx].name, names[i]);
			session->snapInfo[snapIdx].id = fs_getSnapNum(context, names[i]);
			session->snapInfo[snapIdx].snapTime = timeStamp;
			snapIdx++;
		}
    }

	/*
	 * Sort snap name entries based on time stamp
	 */
	qsort(session->snapInfo, snapIdx, sizeof(fs_AmSnapInfo), 
						fs_CompareSnapTime);

	
	for (i = 0; i < snapIdx; i++)
	{
		E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("Snap %s id %d time %d\n", session->snapInfo[i].name,
				session->snapInfo[i].id, session->snapInfo[i].snapTime));

		printf("Snap %s id %d time %d\n", session->snapInfo[i].name,
				session->snapInfo[i].id, session->snapInfo[i].snapTime);
	}

	if (session->oldSnapNum == -1)
	{
		session->numCurSet = 1;
		session->numMaxSets = snapIdx;
		session->setIdx = 0;
		setIdx = session->setIdx;

		strcpy(session->newName, session->snapInfo[setIdx].name);
		session->newSnapNum = session->snapInfo[setIdx].id;

		E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("Starting with Old: None New: %s CurSet %d MaxSet = %d\n", 
			session->newName, session->numCurSet, session->numMaxSets));

		printf("Starting with Old: None New: %s CurSet %d MaxSet = %d\n", 
			session->newName, session->numCurSet, session->numMaxSets);
	}
	else
	{
		session->numCurSet = 1;
		session->numMaxSets = snapIdx - 1;
		session->setIdx = 1;
		setIdx = session->setIdx;

		strcpy(session->oldName, session->snapInfo[setIdx - 1].name);
		session->oldSnapNum = session->snapInfo[setIdx - 1].id;

		strcpy(session->newName, session->snapInfo[setIdx].name);
		session->newSnapNum = session->snapInfo[setIdx].id;

		E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
			("Starting with Old: %s New: %s CurSet %d MaxSet = %d\n", 
			session->oldName, session->newName, 
				session->numCurSet, session->numMaxSets));

		printf("Starting with Old: %s New: %s CurSet %d MaxSet = %d\n", 
			session->oldName, session->newName, 
				session->numCurSet, session->numMaxSets);
	}

    fs_ramDealloc(FS_RAM_AM_LAST_SNAP, nameBuf);

    return NFX_OK;
}

/**
 * Removes all AM snap shots related to this mirror
 * other then the one that target has
 */

int32
fs_snapClean(fs_Context *context)
{
    char    *names[FS_MAX_SNAPSHOTS];
    char    *nameBuf;
    int     i;
    int     nSnapshots;
    fs_AmSourceSession_t *session = context->fileSystem->amSourceSession;
	char 	*oldSnapName = session->oldName;
	char 	*snapName = session->newName;
	int		prefixLen = strlen(snapName) - 3;

	if (strncmp(oldSnapName, "SANM_SS", 7) != 0)
		oldSnapName = NULL;

	E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, 
		("FS snapClean New snap name: %s Old snap name: %s\n", 
			snapName, oldSnapName ?  oldSnapName : "None"));

	printf("FS snapClean New snap name: %s Old snap name: %s\n", 
			snapName, oldSnapName ?  oldSnapName : "None");

    nameBuf = fs_ramAlloc(FS_RAM_AM_LAST_SNAP,
                          (FS_MAX_SNAPSHOTS*FS_SNAPNAME_MAX));
    FS_ASSERT(nameBuf != NULL);

    nSnapshots = fs_snapGetNames(context, names, nameBuf,
                                 (FS_MAX_SNAPSHOTS*FS_SNAPNAME_MAX));

    if (nSnapshots == 0) {
		fs_ramDealloc(FS_RAM_AM_LAST_SNAP, nameBuf);
		return NFX_OK;
    }

	/*
	 * Delete all snap shot for this mirror except 
	 * for the last snap shot that the target has
	 */
    for (i = 0; i < nSnapshots; i++) {

		/*
		 * Dont mess with snaps not related to this mirror
		 */
		if (strncmp(names[i], snapName, prefixLen) != 0)
			continue;

		/*
		 * Dont remove the snap target has
		 */
		if (oldSnapName && strcmp(names[i], oldSnapName) == 0)
			continue;

		printf("Removing snap %s\n", names[i]);

		fs_snapRemove(context, names[i]);
	}

    fs_ramDealloc(FS_RAM_AM_LAST_SNAP, nameBuf);

    return NFX_OK;
}
