/*
 * Name:  fs-api.c
 *
 * RCS:  $Id: fs-api.c,v 1.84 2002/12/10 17:27:07 bob Exp $
 *
 * Description:  Eagle filesystem API.  Any function exported outside sm-fs
 *               should be in this file.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"        

int32 fs_ioDoneNeedYieldTimeout;
int32 fs_ioDoneCompletionDropped;
extern nfx_spinlock_t   fs_rdwrSpin;

#define REQUEST_IN_SHARED_POOL

#if !defined(FS_STANDALONE)
/*-----------------------------------------------------------------------------
 * fs_receiveEEEMessage
 *
 * Description:  Handle a received EEE message for the filesystem.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
EXPORTED void
fs_receiveEEEMessage(
    eee_desc_t   *edesc)
{
    uint32  srcAddr;
    int32   rc, msgLen, numBufs, bufType;
    char    *msgPtr;
#ifdef FS_HIST_WAIT
    struct eee_window_update    *windowMessage;
#endif

	//printf("fs_receiveEEEMessage\n");

    srcAddr = edesc->hdr.src_port;
    numBufs = EEE_GET_NUM_BUFS(edesc);
    msgLen = EEE_GET_PKT_LEN(edesc);
    if (numBufs == 1)
    {
#ifdef REQUEST_IN_SHARED_POOL
        // Serve the request from the system buffer.
        bufType = FS_CONTEXT_SHARED_REQUEST;
        msgPtr = edesc->bd[1].buf;  // TODO - first buffer not used???
		eee_deallocateDesc(edesc);
#else
		uchar8	* src;
		uint32	clen, clines;

        bufType = 0;
        clen   = msgLen + CACHELINE_SIZE - 1;
		clines = clen >> CACHELINE_SHIFT;
        msgPtr = fs_ramAlloc(FS_RAM_REQ_MSG, clen);

        if (((address_t)(edesc->bd[1].buf) & 0xF000000000) == 0x2000000000)
            src = (uchar8 *)uncached_to_cached((uintptr_t)edesc->bd[1].buf);
        else
            src = edesc->bd[1].buf;
		
		ht_memcpy_ass(msgPtr, src, clines);

        eee_deallocateBuffer(edesc->bd[1].buf);
		eee_deallocateDesc(edesc);
#endif
    }
    else
    {
        // Copy the multi-buffer message to a
        // contiguous malloc buffer and free the
        // received packet.
        bufType = 0;
        msgPtr = fs_ramAlloc(FS_RAM_REQ_MSG, msgLen);
        rc = eee_copyPacketToLocalBuffer(   msgPtr,
                                            edesc,
                                            msgLen,
                                            0);
        if (rc != msgLen)
        {
            fs_error(NULL, FS_EEE_ERR, "copypackettolocal");
        }

        // Done with the packet.
        eee_freePacket(edesc);
    }

    //printf("fs_recvMsg: type %d len %d\n", *((uint32 *)msgPtr), msgLen);
#ifdef FS_HIST_WAIT
    windowMessage = (struct eee_window_update *)msgPtr;
    if (windowMessage->wu_msgtype == EEE_MSG_WINDOW_UPDATE) {
        fs_receiveFileHistWindowUpdate(windowMessage);
        if (bufType == FS_CONTEXT_SHARED_REQUEST) {
            eee_deallocateBuffer(windowMessage);
        } else {
            fs_ramDealloc(FS_RAM_REQ_MSG, windowMessage);
        }
        return;
    }
#endif

    fs_receiveMsg(msgPtr, srcAddr, msgLen, bufType);

} /* fs_receiveEEEMessage */

/*-----------------------------------------------------------------------------
 * fs_receiveEvent
 *
 * Description  : Handle a received events for the filesystem.
 * 
 * Created By   : Isabelle Cnudde
 *
 * Date Created : 5/30/2002
 */

void
fs_receiveEvent(nfx_eventMsg_t *    evt)

{
    int32   errorType = 0;
    
    switch(evt->param.volEvent.event)
    {
    case EVENT_VOL_BROKEN:
    {
        errorType = FS_EVENT_VOL_BROKEN;
        break;
    }
    case EVENT_VOL_IOERROR:
    {
        errorType = FS_EVENT_VOL_IOERROR;
        break;
    }
    case EVENT_VOL_CONFLICT:
    {
        errorType = FS_EVENT_VOL_CONFLICT;
        break;
    }
    case EVENT_VOL_TAKEOVER:
    {
        errorType = FS_EVENT_VOL_TAKEOVER;
        break;
    }
    case EVENT_VOL_CORRUPT:
    {
        errorType = FS_EVENT_VOL_CORRUPT;
        break;
    }
    case EVENT_VOL_CREATE:
    case EVENT_VOL_DELETE:
    case EVENT_VOL_GROW:
    case EVENT_VOL_SOFTQUOTA:
    case EVENT_VOL_FULL:
    case EVENT_VOL_MODIFY:
    case EVENT_VOL_ONLINE:
    case EVENT_VOL_OFFLINE:
    case EVENT_VOL_FSYSABORT:
    {
        break; // do nothing
    }
    default:
    {
        printf("FS received an event (%d) for which it didn't subscribed!\n",
               evt->param.volEvent.event);
        FS_ASSERT(FALSE);
    }
    } // switch

    if (errorType != 0)
    {
        fs_FileSystem * fs = fs_fileSystemForVolume(evt->param.volEvent.volId);
        
        if ((fs != NULL) && (fs->state != FS_VOL_STATE_FAILING))
        {
            fs_volumeException(NULL, fs, errorType);
        }
    }

    return;
}

#endif


/*-----------------------------------------------------------------------------
 * fs_ioDone
 *
 * Description:  Volume management calls this when an I/O request
 *               has completed.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
EXPORTED void
fs_ioDone(
    evm_IOStat_t    *stat)
{
    fs_FileSystem           *fileSystem;
    fs_RdwrReq              *rdwr;
    esm_event_t             event;
    int32                   fsState, rc;

    rdwr = (fs_RdwrReq *) (address_t) stat->clientPrivate;
    fileSystem = rdwr->fileSystem;

    FS_ASSERT(rdwr >= fs_rdwrStart &&
              rdwr < fs_rdwrEnd);

#if defined(AGILE_HARDWARE)
#ifdef FS_RDWR_TIMESTAMP
    eee_rtc(&rdwr->timestamp_return); 
#endif
#endif

    FS_ASSERT(rdwr->evmReq.clientPrivate == stat->clientPrivate);

    rdwr->evmStat = *stat;

    FS_TRACE(NULL, FS_TRACE_IO,
                ("vol %llx done on %s blk %lld",
                 rdwr->evmReq.id.whole_thing,
                rdwr->evmReq.reqType == EVM_READ_REQ ? "read" : "write",
                (rdwr->evmReq.offset * EVM_LBLK_SIZE) / FS_BLOCK_SIZE));

    if (rdwr->evmReq.reqType == EVM_WRITE_REQ)
    {
        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioWritesPending -= 1;
        fs_ioPending -= 1;
        NFX_SPIN_UNLOCK(&fs_ioSpin);
    }
    else
    {
        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioReadsPending -= 1;
        fs_ioPending -= 1;
        NFX_SPIN_UNLOCK(&fs_ioSpin);
    }

    if (rdwr->fileSystem != NULL)
    {
        fsState = rdwr->fileSystem->state;
    
        if (fsState == FS_VOL_STATE_FAILING)
        {
            //printf("fs_ioDone() fsState == FS_VOL_STATE_FAILING\n");
            // Volume is in exception condition,
            // ignore IO completion.
            NFX_SPIN_LOCK(&rdwr->fileSystem->spin);
            rdwr->fileSystem->ioPendingCount -= 1;
            NFX_SPIN_UNLOCK(&rdwr->fileSystem->spin);
            fs_ioDoneError(rdwr);
            return;
        }

        if (stat->status != NFX_OK)
        {
            fs_FileSystem   *fileSystem = rdwr->fileSystem;
            int32           errorType;

            errorType = (rdwr->evmReq.reqType == EVM_READ_REQ) ?
                                           FS_READ_ERR : FS_WRITE_ERR;

            // IO error occured, ignore IO completion
            // and enter exception handler
            NFX_SPIN_LOCK(&rdwr->fileSystem->spin);
            rdwr->fileSystem->ioPendingCount -= 1;
            NFX_SPIN_UNLOCK(&rdwr->fileSystem->spin);
            fs_ioDoneError(rdwr);
            fs_volumeException(NULL, fileSystem, errorType);
            return;
        }

        NFX_SPIN_LOCK(&rdwr->fileSystem->spin);
        if (rdwr->evmReq.reqType == EVM_WRITE_REQ)
        {
            rdwr->fileSystem->ioWritesPending -= 1;
        }
        else
        {
            rdwr->fileSystem->ioReadsPending -= 1;
        }
        NFX_SPIN_UNLOCK(&rdwr->fileSystem->spin);
    }
    else
    {
        //printf("fs_ioDone() rdwr->fileSystem == NULL\n");
        fs_ioDoneNeedYieldTimeout += 1;
        fs_ioDoneError(rdwr);
        return;

        // Here we're relying on the yield timeout
        // to recover what's waiting on this.
    }

    if (rdwr->flags & FS_RDWR_FLAGS_SYNC)
    {
        NFX_SPIN_LOCK(&rdwr->fileSystem->spin);
        rdwr->fileSystem->ioPendingCount -= 1;
        NFX_SPIN_UNLOCK(&rdwr->fileSystem->spin);
    }

#if !defined(FS_STANDALONE)
    // hack to make the mem stats look right
    NFX_SPIN_LOCK(&fs_ramStatsSpinId);
    fs_ramStats[FS_RAM_STATS_ALLOC][FS_RAM_READ_DATA] += 1;
    NFX_SPIN_UNLOCK(&fs_ramStatsSpinId);
#endif

    /*
     * Send the event and the updated original
     * request to the waiting state machine.
     */
    event.event = rdwr->doneEvent;
    event.p1.u.p = rdwr;
    event.p2.u.p = NULL;
    rc = thread_highPriorityEvent(rdwr->smHandle, &event);
    if (rc != NFX_OK)
    {
        // Its possible that the read waiting for this completion
        // has already terminated (eg. during exception handling).
        //printf("fs_ioDone:  warning, IO completion event dropped (%d writes pending)\n",
        //    fs_ioWritesPending);
        fs_ioDoneCompletionDropped += 1;
        if (!(rdwr->flags & FS_RDWR_FLAGS_SYNC))
        {
            NFX_SPIN_LOCK(&rdwr->fileSystem->spin);
            rdwr->fileSystem->ioPendingCount -= 1;
            NFX_SPIN_UNLOCK(&rdwr->fileSystem->spin);
        }
        fs_ioDoneError(rdwr);
    }

} /* fs_ioDone */



int
fs_getHashVolBlock(
	int	argc,
	char **argv)
{
	evm_volId_t volId;
	uint64		blknum;
	int32	    hash;

	if (argc < 3)
	{
		printf("usage: fs_vol_blk_hash: <0xVolId> <blknum>\n");
		return -1;
	}

    volId.whole_thing = strtoll(argv[1], NULL, 16);
    blknum = strtoll(argv[2], NULL, 10);

	hash = fs_bufHashVolBlock(volId, fs_makeBlock(blknum));

	printf("hash for vol 0x%llx blknum %lld: %d\n",
			volId.whole_thing, blknum, hash);

	return 0;
}


int
fs_getHashVolInum(
    int argc,
    char **argv)
{
    evm_volId_t volId;
    uint64      inum;
    int32       hash;

    if (argc < 3)
    {
        printf("usage: fs_vol_inum_hash: <0xVolId> <inum>\n");
        return -1;
    }

    volId.whole_thing = strtoll(argv[1], NULL, 16);
    inum = strtoll(argv[2], NULL, 10);

    hash = fs_bufHashVolInum(volId, inum);

    printf("hash for vol 0x%llx inum %lld: %d\n",
            volId.whole_thing, inum, hash);

	return 0;
}


int
fs_getHashVolInumPage(
    int argc,
    char **argv)
{
    evm_volId_t volId;
    uint64      inum;
    uint64      page;
    int32       hash;

    if (argc < 4)
    {
        printf("usage: fs_vol_inum_page_hash: <0xVolId> <inum> <page>\n");
        return -1;
    }
    
    volId.whole_thing = strtoll(argv[1], NULL, 16);
    inum = strtoll(argv[2], NULL, 10);
    page = strtoll(argv[3], NULL, 10);

    hash = fs_bufHashVolInumPage(volId, inum, page);

    printf("hash for vol 0x%llx inum %lld page %lld: %d\n",
            volId.whole_thing, inum, page, hash);

    return 0;
}


#if !defined(FS_STANDALONE)
/*-----------------------------------------------------------------------------
 * fs_initModule
 *
 * Description:  Called to initialize the filesystem module.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
EXPORTED void
fs_initModule()
{
    int32   rc;

    printf("fs_initModule\n");

    fs_initAlloc();

    // One-time initialization of the fs module.  This should
    // occur on a single CPU only.

#ifdef NOTDEF // TODO - add these to LINUX_TEST target
    // install the debugging signal handlers
    fs_simInitSigHandler();

    // initialize stack backtracing - only needed if malloc gets corrupted
    btrace_init();
#endif

    // Init FP globals.
    bzero(fs_errorCounts, sizeof(fs_errorCounts));
    bzero(fs_ramStats, sizeof(fs_ramStats));
    NFX_SPIN_INIT(&fs_ramStatsSpinId,SPINLOCK_LEVEL_UNKNOWN);

    bzero(&fs_nfsStats, sizeof(fs_nfsStats));
    bzero(&fs_yieldStats, sizeof(fs_yieldStats));
    bzero(&fs_recvReqStats, sizeof(fs_recvReqStats));
    bzero(&fs_replyStats, sizeof(fs_replyStats));
    NFX_SPIN_INIT(&fs_statsReqRcvdSpinId,SPINLOCK_LEVEL_UNKNOWN);

    bzero(&fs_errStats, sizeof(fs_errStats));
    NFX_SPIN_INIT(&fs_errStatsSpinId,SPINLOCK_LEVEL_UNKNOWN);

    NFX_SPIN_INIT(&fs_ioSpin,SPINLOCK_LEVEL_UNKNOWN);

    fs_initIncoreTune();
    
    fs_rdwrInit();

    fs_lockOrderInit();

    fs_initFileSystemTable();

    fs_bufCacheInit();

    fs_inodeCacheInit();

    fs_quotaCacheInit();

    fs_dnlcInit();

    fs_dirhash_cache_init();

    fs_bt_dispatch_init();

    fs_cod_init();

#ifdef NEW_ZF
    fs_zfInit();
#endif // NEW_ZF
#ifdef FS_HIST_WAIT
    fs_backupInit();
#endif

    fs_vlcInit();
    
    // Obtain our application id.
    fs_applicationId = eee_allocateApplicationId("efs");
    if (fs_applicationId == NFX_ERR)
    {
        fs_error(NULL, FS_EEE_ERR, "allocateApplicationId");
    }
	fs_srcPortId = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
											 nfxMyCPUId,
											 fs_applicationId,
											 0,
											 NFX_COS_HIGH,
											 0);

    // Register receive message function with eee.
    rc = eee_registerReceiveFunc(   fs_applicationId,
                                    fs_receiveEEEMessage,
                                    fs_receiveEEEMessage);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_EEE_ERR, "registerReceiveFunc");
    }

    // Obtain application Id's for applications we'll be concerned
    // with.
    fs_efsSscAppId = eee_getApplicationId("ssc_efs");
    if (fs_efsSscAppId == NFX_ERR)
    {
        fs_error(NULL, FS_EEE_ERR, "getApplicationId ssc_efs");
    }
    fs_nfsAppId = eee_getApplicationId("nfs_srvr");
    if (fs_nfsAppId == NFX_ERR)
    {
        fs_error(NULL, FS_EEE_ERR, "getApplicationId nfs_srvr");
    }
    fs_mcpuAppId = eee_getApplicationId("sbm");
    if (fs_mcpuAppId == NFX_ERR)
    {
        fs_error(NULL, FS_EEE_ERR, "getApplicationId sbm");
    }
	fs_mcpuDestPortId = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
												  NFX_NFP_TXRX_ANY,
                                                   fs_mcpuAppId, EEE_CONTROL,
                                                   NFX_COS_HIGH, 0);

    DLL_INIT(&fs_volExerDllHdr);

    /* Register to volume events.
     */
    event_registerClient(NFX_EVENT_VOLUME, fs_receiveEvent);
    
    printf("fs_initModule done on slot %d cpu %d\n",
        nfxMySlotId, nfxMyCPUId);
    printf("efs_sscAppId: %d, nfsAppId: %d, mcpuAppId: %d\n",
           fs_efsSscAppId, fs_nfsAppId, fs_mcpuAppId);

} /* fs_initModule */
#endif // FS_STANDALONE
#endif // SSC
