/*
 * Name:  fs-api.h
 *
 * RCS:  $Id: fs-api.h,v 1.158 2002/12/04 19:25:55 lanc Exp $
 *
 * Description:  Filesystem client interface.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_API_H)
#define FS_API_H

#include <exports/esm-api.h>

// Internal readability aid.
#define PRIVATE     static
#define EXPORTED

typedef uint64 fs_Offset;
typedef uint64 fs_Size;

/*-----------------------------------------------------------------------------
 * Inode numbers in an fs_Handle contain the inode number,
 * the snapshot number, and a version field.
 */
#define FS_HANDLE_SNAP_MASK         0xff00000000000000ULL
#define FS_HANDLE_SNAP_SHIFT        56
#define FS_HANDLE_VERS_MASK         0x00f0000000000000ULL
#define FS_HANDLE_VERS_SHIFT        52
#define FS_HANDLE_INUM_MASK         0x000fffffffffffffULL

#define FS_HANDLE_GET_SNAP(_inum_)                \
    (((_inum_) & FS_HANDLE_SNAP_MASK) >> FS_HANDLE_SNAP_SHIFT)

#define FS_HANDLE_SET_SNAP(_inum_, _s_)           \
    (_inum_) &= ~FS_HANDLE_SNAP_MASK;             \
    (_inum_) |= ((uint64) (_s_) << FS_HANDLE_SNAP_SHIFT);

#define FS_HANDLE_GET_VERS(_inum_)                \
    (((_inum_) & FS_HANDLE_VERS_MASK) >> FS_HANDLE_VERS_SHIFT)

#define FS_HANDLE_SET_VERS(_inum_, _v_)           \
    (_inum_) &= ~FS_HANDLE_VERS_MASK;             \
    (_inum_) |= ((uint64) (_v_) << FS_HANDLE_VERS_SHIFT);

#define FS_HANDLE_GET_INUM(_inum_)                \
    ((_inum_) & FS_HANDLE_INUM_MASK)

#define FS_HANDLE_SET_INUM(_inum_, _n_)           \
    (_inum_) &= ~FS_HANDLE_INUM_MASK;             \
    (_inum_) |= (_n_);


// Maximum filename length in bytes
#define FS_NAME_MAX 255

// Maximum pathname length in a soft link
#define FS_PATH_MAX 1024

// This is provided for some checksumming support in
// fiber channel - TODO - this needs to be moved to a
// message API instead of a hard-coded value!f
#define EFS_BLOCK_SIZE  8192



/**
 * Mounting state for a volume.
 */
typedef enum {
    FS_VOL_STATE_ONLINE = 1,
    FS_VOL_STATE_ONLINE_RDONLY,
    FS_VOL_STATE_OFFLINE,
    FS_VOL_STATE_PAUSED,
    FS_VOL_STATE_MOUNTING,
    FS_VOL_STATE_UNMOUNTING,
    FS_VOL_STATE_FAILING,
    FS_VOL_STATE_BACKUP_RESTORE
} fs_FileSystemState;


// Return status codes.
// Those less than FS_ERR_DEFERRED map directly to NFSv2/3.
// TODO - move these to global errno.h
typedef enum fs_Status
{
    FS_ERR_BADPATH          = -2,
    FS_ERR_GENERIC          = -1,   // NFX_ERR
    FS_ERR_NONE             = 0,    // NFX_OK, NFS_OK, NO_ERROR
    FS_ERR_PERM             = 1,
    FS_ERR_NOENT            = 2,
    FS_ERR_IO               = 5,
    FS_ERR_NXIO             = 6,
    FS_ERR_ACCESS           = 13,
    FS_ERR_EXIST            = 17,
    FS_ERR_XDEV             = 18,
    FS_ERR_NODEV            = 19,
    FS_ERR_NOTDIR           = 20,
    FS_ERR_ISDIR            = 21,
    FS_ERR_INVAL            = 22,
    FS_ERR_FBIG             = 27,
    FS_ERR_NOSPC            = 28,
    FS_ERR_ROFS             = 30,
    FS_ERR_MLINK            = 31,
    FS_ERR_NAMETOOLONG      = 63,
    FS_ERR_NOTEMPTY         = 66,
    FS_ERR_DQUOT            = 69,
    FS_ERR_STALE            = 70,
    FS_ERR_REMOTE           = 71,
    FS_ERR_WFLUSH           = 99,

    FS_ERR_BADHANDLE        = 10001,
    FS_ERR_NOT_SYNC         = 10002,
    FS_ERR_BAD_COOKIE       = 10003,
    FS_ERR_NOTSUPP          = 10004,
    FS_ERR_TOOSMALL         = 10005,
    FS_ERR_SERVERFAULT      = 10006,
    FS_ERR_BADTYPE          = 10007,
    FS_ERR_JUKEBOX          = 10008,
	FS_ERR_BUSY				= 10009,

    FS_ERR_DEFERRED         = 20000,
    FS_ERR_WRONG_EVENT      = 20001,
    FS_ERR_LOCK_DESTROYED   = 20002,
    FS_ERR_ALREADY_MOUNTED  = 20003,

    // the next errors are used to indicate that only the resp header
    // and fsStatus field  is used. None of the other fields 
    // (e.g. pre-attributes) are initialised (even not bzero)
    // NFS should drop the request
    FS_ERR_NOT_MOUNTED      = 30001,    // not mounted, or not online
    FS_ERR_NO_RESOURCE      = 30002,    // not enough resources to handle this req

    FS_IN_PROGRESS          = 40000,    // in progress status

    // the following are used to indicate snapshot errors
    FS_ERR_SNAP_EXISTS      = 50001,    // already exists (ie create error)
    FS_ERR_SNAP_NO_EXIST    = 50002,    // does not exist (ie remove error)
    FS_ERR_SNAP_MAX         = 50003,    // maxinum reached
    FS_ERR_SNAP_INVAL_NAME  = 50004,    // invalid name (ie hourly, etc)
    FS_ERR_SNAP_NOT_ENABLED = 50005,    // not enabled
    FS_ERR_SNAP_PIN         = 50006,    // cannot remove: pinned
} fs_Status;


/*-----------------------------------------------------------------------------
 * fs_ReqType - The list of supported request types in this API.
 *
 * NB: this has to match the order of requests in requests.tab!!!!
 *
 * TODO - generate these constants (but first put comments and allow
 * explicit mapping of request numbers in requests.tab)
 */
typedef enum fs_ReqType
{
    //
    // These map to NFSv3.  don't goof with them without updating
    // requests.tab.
    // AND fs_reqTypeName (fs-shell.c)
    //
    FS_REQ_NULL         = 0,        // do nothing
    FS_REQ_GET_ATTR     = 1,        // get attributes of a file or dir
    FS_REQ_SET_ATTR     = 2,        // set attributes of a file or dir
    FS_REQ_LOOKUP       = 3,        // lookup attributes of a file
    FS_REQ_ACCESS       = 4,        // check access permission on object
    FS_REQ_READLINK     = 5,        // read a soft link
    FS_REQ_READ         = 6,        // read data from a file
    FS_REQ_WRITE        = 7,        // write data to a file
    FS_REQ_CREATE       = 8,        // create a new file
    FS_REQ_MKDIR        = 9,        // create a new directory
    FS_REQ_SYMLINK      = 10,       // create a soft link
    FS_REQ_MKNOD        = 11,       // create device file
    FS_REQ_REMOVE       = 12,       // remove a file
    FS_REQ_RMDIR        = 13,       // remove a directory
    FS_REQ_RENAME       = 14,       // rename a file or directory
    FS_REQ_HARDLINK     = 15,       // create a hard link to a file
    FS_REQ_READDIR      = 16,       // read a directory's contents
    FS_REQ_READDIRPLUS  = 17,       // read a directory's contents & attributes
    FS_REQ_STAT         = 18,       // get filesystem stats
    FS_REQ_FSINFO       = 19,       // get statis filesystem information
    FS_REQ_PATHCONF     = 20,       // get POSIX info
    FS_REQ_COMMIT       = 21,       // commit mods to stable storage
    FS_REQ_NFS_MAX      = 22,       // num of nfs client request types
    //
    // The rest of these are internal management
    //
    FS_REQ_TX_COMPL     = 22,       // signal tx complete on a buffer
    FS_REQ_SYNC         = 23,       // sync dirty buffers to storage
    FS_REQ_MOUNT        = 24,       // mount a filesystem volume
    FS_REQ_UNMOUNT      = 25,       // unmount a filesystem volume
    FS_REQ_SNAP_CREATE  = 26,       // create a snapshot
    FS_REQ_SNAP_REMOVE  = 27,       // remove a snapshot
    FS_REQ_MAKEFS       = 28,       // create a filesystem
    FS_REQ_CHECKFS      = 29,       // check a filesystem
    FS_REQ_GET_ROOT     = 30,       // obtain root directory handle
    FS_REQ_TRACE        = 31,       // set trace level
    FS_REQ_SHOW_LOCKED  = 32,       // show locked bufs/inodes
    FS_REQ_EMAPDEBUG    = 33,       // run emap debug code
    FS_REQ_FLUSH        = 34,       // flush the buffer cache
	FS_REQ_CREATE_FILESET=35,		// create file set
	FS_REQ_REMOVE_FILESET=36,		// remove file set
	FS_REQ_REPORT_QUOTA	= 37,		// report quota
	FS_REQ_SET_QUOTA	= 38,		// set quota
	FS_REQ_GET_QUOTA	= 39,		// get quota
	FS_REQ_QUOTA_ON		= 40,		// quota on
	FS_REQ_QUOTA_OFF	= 41,		// quota off
	FS_REQ_LIST_FILESETS= 42,	    // list filesets
    FS_REQ_FREEZE       = 43,       // quiesce the filesystem
    FS_REQ_THAW         = 44,       // unquiesce the filesystem
    FS_REQ_CACHESTATS   = 45,       // debug: display cachestats
    FS_REQ_MEMSTATS     = 46,       // debug: display memstats
    FS_REQ_REQSTATS     = 47,       // debug: display reqstats
    FS_REQ_CONTEXT      = 48,       // debug: display a context
    FS_REQ_LOGDATA      = 49,       // debug: display log data
    FS_REQ_CRASH        = 50,       // debug: crash on command
    FS_REQ_DUMP_SETUP   = 51,       // Setup dump
    FS_REQ_DUMP_NAMES   = 52,       // setup dump names
    FS_REQ_DUMP_START   = 53,       // start dump
    FS_REQ_DUMP_ABORT   = 54,       // abort dump
    FS_REQ_DUMP_IODONE  = 55,       // i/o complete for dump
    FS_REQ_REST_SETUP   = 56,       // Setup restore
    FS_REQ_REST_NAMES   = 57,       // setup restore names
    FS_REQ_REST_START   = 58,       // start restore
    FS_REQ_REST_ABORT   = 59,       // abort restore
    FS_REQ_REST_IODONE  = 60,       // i/o complete for restore
    FS_REQ_FSSTAT_VOL   = 61,       // given a volid return fsinfo for that
    FS_REQ_EEK          = 62,	    // eek an fs from the embedded
    FS_REQ_GROWFS	    = 63,	    // grow the filesystem
    FS_REQ_SNAP_DIAG    = 64,       // snapshot diag
    FS_REQ_AM_DELTA_SETUP = 65,     // setup Agile Mirror Delta
    FS_REQ_AM_DELTA_GET_EXTENT = 66,    // get extents for Agile Mirror
    FS_REQ_AM_DELTA_COMPLETE = 67,  // completion of delta operation
    FS_REQ_AM_DELTA_INSTALL_SETUP = 68, // setup detla install on AM target
    FS_REQ_AM_DELTA_INSTALL_EXTENT = 69,    // install extents on AM target
    FS_REQ_AM_DELTA_INSTALL_COMMIT = 70,    // complete AM target installation
    FS_REQ_AM_LAST_SNAPSHOT = 71,   // Get the name of the latest snapshot
    FS_REQ_SNAP_SHOW    = 72,       // snap show
    FS_REQ_SNAP_SCHED   = 73,       // snap schedule
    FS_REQ_SNAP_ADMIN   = 74,       // snap admin
    FS_REQ_SNAP_REVERT  = 75,       // snap revert
    FS_REQ_EEKCONT      = 76,	    // eek continuation
    FS_REQ_NFSSTATS     = 77,	    // nfs stats request
    FS_REQ_YIELDSTATS   = 78,	    // yield stats request
    FS_REQ_LOCKSTATS    = 79,	    // lock stats request
    FS_REQ_LOG_TEST     = 80,       // log unit test
    FS_REQ_ZAP          = 81,	    // debug command: introduce inconsistencies
    FS_REQ_GET_FILESET_PATH = 82,   // given an dir inode get the path to it
    FS_REQ_LOGSTATS     = 83,       // debug: display logstats
    FS_REQ_VIEWFS       = 84,       // view all known file systems
    FS_REQ_ABORT        = 85,       // abort the specified volume (debug)
    FS_REQ_VOLEXER      = 86,       // volume exer commands
    FS_REQ_ALLOCSTATS   = 87,       // utilities: allocator statistics
    FS_REQ_SET_MIN_FREE = 88,       // Set the filesystem tunables - min free
    FS_REQ_GET_MIN_FREE = 89,       // Get the fileysstem tunables - min free
    FS_REQ_MCL          = 90,       // Multi Component Lookup
    FS_REQ_ERRSTATS     = 91,       // debug: display err (eg resource) stats
    FS_REQ_BENCH_MARK   = 92,       // benchmark driver requests
    FS_REQ_INCORE_TUNE  = 93,       // set/read incore tune parameters requests
    FS_REQ_DNLCSTATS    = 94,       // get/clear dnlc statistics
    FS_REQ_SNAP_STATE   = 95,       // enable/disable snapshot scheduling
    FS_REQ_SNAP_PIN     = 96,       // pin/unpin snapshots
    FS_REQ_SNAP_RENAME  = 97,       // rename snapshots
    FS_REQ_CACHE_INV    = 98,       // invalidate cache on TXRX
    FS_REQ_MAX          = 98
/*
 * NB: these requests have to match the order of requests in requests.tab!!!!
 *     AND fs_reqTypeName (fs-shell.c)
 */

} fs_ReqType; 

/*-----------------------------------------------------------------------------
 * Major/minor number pair.
 */
typedef struct fs_DevNum
{
    uint32      major;
    uint32      minor;

} fs_DevNum;

/*-----------------------------------------------------------------------------
 * fs_AttrType - Used in request/response attrType field as well as
 * in fs_Attr struct to identify the object type and relation
 * to the transaction (pre/post) of the attribute.
 */
typedef enum fs_AttrType
{
    FS_PRE_PAR_ATTRS    = 0x01,   // Pre-op parent attributes
                                  // requested/responded.
    FS_POST_PAR_ATTRS   = 0x02,   // Post-op parent attributes
                                  // requested/responded. 
    FS_PRE_ATTRS        = 0x04,   // Pre-op file attributes 
                                  // requested/responded.
    FS_POST_ATTRS       = 0x08,   // Post-op file attributes
                                  // requested/responded.
    FS_PRE_FPAR_ATTRS   = 0x10,   // Pre-op FROM parent dir attributes
                                  // requested/responded.
    FS_POST_FPAR_ATTRS  = 0x20,   // Post-op FROM parent dir attributes
                                  // requested/responded.
    FS_PRE_TPAR_ATTRS   = 0x40,   // Pre-op TO parent dir attributes
                                  // requested/responded.
    FS_POST_TPAR_ATTRS  = 0x80    // Post-op TO parent dir attributes
                                  // requested/responded.
} fs_AttrType;

/*-----------------------------------------------------------------------------
 * fs_Proto - data structure used to provide details
 * common to all filesystem transaction request messages.
 */
typedef enum fs_Proto
{
    FS_PROTO_NFS_V2     = 1,
    FS_PROTO_NFS_V3     = 2,
    FS_PROTO_NFS_V4     = 3,
    FS_PROTO_CIFS       = 4,
    FS_PROTO_MAX        = 4

} fs_Proto;

/*-----------------------------------------------------------------------------
 * fs_Time
 */
typedef struct fs_Time
{
    uint32      secs;
    uint32      nsecs;

} fs_Time;


#define FS_TIME_LE(_fs_time_1_, _fs_time_2_)                                   \
   (((((uint64)_fs_time_1_.secs) << 32) + ((uint64)_fs_time_1_.nsecs)) <=      \
    ((((uint64)_fs_time_2_.secs) << 32) + ((uint64)_fs_time_2_.nsecs)))

#define FS_TIME_GT(_fs_time_1_, _fs_time_2_)                                   \
   (((((uint64)_fs_time_1_.secs) << 32) + ((uint64)_fs_time_1_.nsecs)) >       \
    ((((uint64)_fs_time_2_.secs) << 32) + ((uint64)_fs_time_2_.nsecs)))


/*-----------------------------------------------------------------------------
 */

typedef uint32      fs_Generation;      // Generation numbers where required.

/*-----------------------------------------------------------------------------
 * fs permission 
 *-----------------------------------------------------------------------------
 */
enum {
    FS_PERM_LOOKUP  = 1,    // lookup in a directory
    FS_PERM_READ    = 2,    // read a file
    FS_PERM_WRITE   = 3,    // write a file
    FS_PERM_DIRENT  = 4,    // create/remove directory entries
    FS_PERM_READDIR = 5,    // list directory contents
    FS_PERM_EXEC    = 6,    // has execute perms
    FS_PERM_SETATTR  = 7    // Can set attributes on an object
};


/*-----------------------------------------------------------------------------
 * fs mode
 *-----------------------------------------------------------------------------
 */
#define FS_IMODE_MASK 0000777
#define FS_IMODE_RWXUGO   (FS_IMODE_RWXU|FS_IMODE_RWXG|FS_IMODE_RWXO)
#define FS_IMODE_RUGO     (FS_IMODE_RUSR|FS_IMODE_RGRP|FS_IMODE_ROTH)
#define FS_IMODE_WUGO     (FS_IMODE_WUSR|FS_IMODE_WGRP|FS_IMODE_WOTH)
#define FS_IMODE_XUGO     (FS_IMODE_XUSR|FS_IMODE_XGRP|FS_IMODE_XOTH)

#define FS_IMODE_RWXU 0000700
#define FS_IMODE_RUSR 0000400
#define FS_IMODE_WUSR 0000200
#define FS_IMODE_XUSR 0000100

#define FS_IMODE_RWXG 0000070
#define FS_IMODE_RGRP 0000040
#define FS_IMODE_WGRP 0000020
#define FS_IMODE_XGRP 0000010

#define FS_IMODE_RWXO 0000007
#define FS_IMODE_ROTH 0000004
#define FS_IMODE_WOTH 0000002
#define FS_IMODE_XOTH 0000001



/*-----------------------------------------------------------------------------
 * fs_MsgReqHdr - data structure used to provide details
 * common to all filesystem transaction request messages.
 */
typedef struct fs_MsgReqHdr
{
    /********************
     ** BEGIN WARNING!!!!  - this must be kept in sync with fs-api.h
     ** and fssf-api.h and mcpu-sim.h message headers
     ********************/
    int32      requestType;            // Type of action being requested
    int32       len;
#if defined(FS_STANDALONE)
    int32       src;    // match fs_simMsgHdr
    int32       dest;   // match fs_simMsgHdr
    int32       nfsReqNum;
    int32       sockIndex;
#endif
    /********************
     ** END WARNING
     ********************/
    fs_AttrType attrType;               // Bitmap of requested attribute types.
    uint32      uid;                    // User ID for the request
    uint32      gid;                    // Group ID for the request
    fs_Proto    proto;                  // Client protocol.
    int32       numGids;                // Number of valid entries in gids[].
    uint32	clientId;
    uint64      clientPrivate;          // Opaque to fs, returned in
                                        // response message
#define FS_REQ_GIDS_MAX     16
    uint32      gids[FS_REQ_GIDS_MAX];  // Additional groups client belongs to.
                                        // TODO - make this variable sized???
} fs_MsgReqHdr;

NFX_ASSERT_STATIC(sizeof (fs_MsgReqHdr) % 8 == 0);


/*-----------------------------------------------------------------------------
 * fs_MsgRespHdr - data structure used to provide details
 * common to all filesystem transaction response messages.
 */
typedef struct fs_MsgRespHdr
{
    /********************
     ** BEGIN WARNING!!!!  - this must be kept in sync with fs-api.h
     ** and fssf-api.h and mcpu-sim.h message headers
     ********************/
    int32      	requestType;            // Type of action being requested
    int32       len;
#if defined(FS_STANDALONE)
    int32       src;    // match fs_simMsgHdr
    int32       dest;   // match fs_simMsgHdr
    int32       nfsReqNum;
    int32       sockIndex;
#endif
    /********************
     ** END WARNING
     ********************/

    fs_AttrType attrType;               // Bitmap of attribute types
                                        // contained in the response.
    fs_Status   fsStatus;               // Completion status of request
    uint32      pad;
    uint32      clientId;
    uint64      clientPrivate;          // Opaque to fs, returned in
                                        // response message
} fs_MsgRespHdr;

NFX_ASSERT_STATIC(sizeof (fs_MsgRespHdr) % 8 == 0);


/*-----------------------------------------------------------------------------
 * fs_Handle - Handle used externally to identify a file or
 * directory.  A generation number is used to identify stale
 * fs_Handle's in the event the inode has been re-allocated
 * to a different file.
 */
typedef struct fs_Handle
{
    evm_volId_t     volId;              // 64-bit volume identifier for the
                                        // filesystem
    uint64          inum;               // 64-bit inode number, upper 8 bits
                                        // are snapshot number
    uint32          gen;                // 32-bit generation number, protect
                                        // from re-used inodes
    uint32          mkfsTime;           // creation time for file system
    uint64          mntPnt;             // 64-bit inode number of the mount point
                                        // directory
} fs_Handle;

NFX_ASSERT_STATIC(sizeof (fs_Handle) % 8 == 0);

/*-----------------------------------------------------------------------------
 * fs_FileType - Used for type field in fs_Attr.  They map directly
 * to NFSv2 and v3.
 */
typedef enum fs_FileType
{
    FS_FTYPE_NON    = 0,        // Invalid
    FS_FTYPE_REG    = 1,        // Regular file
    FS_FTYPE_DIR    = 2,        // Directory
    FS_FTYPE_BLK    = 3,        // Block device file
    FS_FTYPE_CHR    = 4,        // Character device file
    FS_FTYPE_LNK    = 5,        // Symbolic link
    FS_FTYPE_SOCK   = 6,        // Socket
    FS_FTYPE_FIFO   = 7         // Named pipe

} fs_FileType;

/*-----------------------------------------------------------------------------
 * fs_Attr - Used to supply the attributes during a file
 * GETATTR or SETATTR request.  This covers NFSv2 and v3.
 *
 * fs_Time counts usecs vs. nsecs.
 *
 * During SETATTR, fields with value of -1 are ignored.
 */
typedef struct fs_Attr
{
    fs_AttrType     attrType;       // Type of attribute - GET only
    fs_FileType     type;           // Type of file - GET only
    uint32          mode;           // UNIX mode - GET/SET
    uint32          numLinks;       // hard link count - GET only
    uint32          uid;            // file owner - GET/SET
    uint32          gid;            // file group - GET/SET
    uint32          blockSize;      // bytes per block on this volume - GET only
    uint32          blocks;         // blocks used by this object - GET only
    uint64          size;           // file size in bytes - GET/SET
    fs_DevNum       devNum;         // major & minor numbers for special files
    uint64          fsId;           // volume file exists on (fsid) - GET only
    uint64          inode;          // Inode number (file id) - GET only
    fs_Time         atime;          // time of last read - GET/SET.
    fs_Time         mtime;          // time of last file data
                                    // modification - GET/SET.
    fs_Time         ctime;          // time of last meta modification GET only
    uint32          rdev;           // Needed for compatibility with V2.
                                    // (eg. -1 indicates a PIPE if mode is a
                                    //  char special)
    ushort16        atimeType;
    ushort16        mtimeType;
#define FS_ATTR_TIME_SET_DONT   0   // Don't alter time during setattr
#define FS_ATTR_TIME_SET_SRVR   1   // Set to server's current time
#define FS_ATTR_TIME_SET_CLIENT 2   // Set to client specified time
    

} fs_Attr;

NFX_ASSERT_STATIC(sizeof (fs_Attr) % 8 == 0);


/*-----------------------------------------------------------------------------
 * fs_DirEntry - Used in response to READDIR requests.
 */
typedef struct fs_DirEntry
{
    uint64              inode;      // Inode number (fileid)
    uint64              cookie;     // "index" of the entry in the directory.
                                    // NOTE: For NFSv2, this is only 32-bit!!!
    uint32              nameLen;    // Length of name in bytes COUNTING NULL
                                    // terminator.
    uchar8              snapNum;    // Snapshot number for this dirent.
	uchar8				pad[3];
    char                name[0];    // NULL terminated name of entry up
                                    // to FS_NAME_MAX.
} fs_DirEntry;

/*-----------------------------------------------------------------------------
 * fs_DirEntryPlus - Used in response to READDIRPLUS requests.
 */
typedef struct fs_DirEntryPlus
{
    fs_Handle           handle;     // Handle to dirent
    fs_Attr             attr;       // File attributes
    fs_DirEntry         dirEntry;

} fs_DirEntryPlus;

/*-----------------------------------------------------------------------------
 * fs_MsgNullReq - data structure used for NULL
 *                 transaction request messages.
 *
 * Client request nothing.
 */
typedef struct fs_MsgNullReq
{
    fs_MsgReqHdr        hdr;

} fs_MsgNullReq;

/*-----------------------------------------------------------------------------
 * fs_MsgNullResp - data structure used for NULL
 *                  transaction response messages.
 *
 * Response includes header only.
 */
typedef struct fs_MsgNullResp
{
    fs_MsgRespHdr        hdr;

} fs_MsgNullResp;

/*-----------------------------------------------------------------------------
 * fs_MsgReadReq - data structure used for system READ
 *                 transaction request messages.
 *
 * Client requests read of a file.
 */
typedef struct fs_MsgReadReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           fileHandle;     // File handle of file to read
    uint64              fileOffset;     // Byte offset to start read action
    uint32              readLength;     // Number of bytes to read

}  fs_MsgReadReq;

/*-----------------------------------------------------------------------------
 * fs_MsgReadResp - data structure used for system READ
 *                  transaction response messages.
 *
 * Response includes completion status, cod which points to read
 * data and bytesRead which indicates the actual number of bytes
 * read (could be less than readLength).
 */
typedef struct fs_MsgReadResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // Filehandle from original request.
    uint64              fileOffset;     // Byte offset to start read action
    nfx_cod_entry_t     cod[NFX_MAX_COD_MSG];  // Returned buffers
	uint64				fsPrivate;		// Must be returned to FS in TX_COMPL
										// message.
    uint32              nbrUsedCod;     // Number of used COD's 
#if defined(FS_STANDALONE)
    uint32              data_index;     // where the data start in this msg
    uint32              pad;
#endif
    uint32              bytesRead;      // Number of bytes actually read
    uint32              eof;            // Indicates if end of file reached.
    uint32              numAttrs;       // Number of attributes in attr[].   
    fs_Attr             attr[0];        // Optional updated file attributes.
#if defined(FS_STANDALONE)
    char                data[0];        // field for the data, after 
                                        // looking up the COD
#endif
}  fs_MsgReadResp;

/*-----------------------------------------------------------------------------
 * fs_MsgWriteReq - data structure used for filesystem WRITE
 *                  transaction request messages.
 *
 * Client requests write to a file.
 */
typedef enum fs_WriteStableHow
{
    FS_WRITE_STABLE_UNSTABLE    = 0,
    FS_WRITE_STABLE_DATA_SYNC   = 1,
    FS_WRITE_STABLE_FILE_SYNC   = 2
} fs_WriteStableHow;

typedef struct fs_MsgWriteReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           fileHandle;     // File handle of file to write
    uint64              fileOffset;     // Byte offset to start write action
    uint64              writeLength;    // Number of bytes to write
    nfx_cod_entry_t     cod[NFX_MAX_COD_MSG];  // Returned buffers
    uint32              nbrUsedCod;     // Number of used COD's 
    fs_WriteStableHow   stableHow;      // Data stability required.

}  fs_MsgWriteReq;

/*-----------------------------------------------------------------------------
 * fs_MsgWriteResp - data structure used for filesystem WRITE
 *                   transaction response messages.
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgWriteResp
{
    fs_MsgRespHdr        hdr;           // Message header
    int32                numAttrs;      // Number of attributes in attr[].
    fs_WriteStableHow    stableHow;     // Data stability provided.
    uint32               verifier;      // Write verifier cookie.
    uint32               count;         // Number of bytes written to the file
    fs_Attr              attr[0];       // Optional pre and post file attributes.

}  fs_MsgWriteResp;

/*-----------------------------------------------------------------------------
 * fs_MsgLookupReq - data structure used for filesystem LOOKUP
 *                   transaction request messages.
 *
 * Client requests lookup to determine if file or directory exists.
 */
typedef struct fs_MsgLookupReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           parentHandle;   // File handle of parent directory
    uint32              nameLen;        // Length of name in bytes NOT counting NULL
                                        // terminator
    char                fileName[0];    // NULL terminated name to lookup up to
                                        // FS_NAME_MAX

}  fs_MsgLookupReq;

/*-----------------------------------------------------------------------------
 * fs_MsgLookupResp - data structure used for filesystem LOOKUP
 *                    transaction response messages.
 *
 * Response includes completion status and fileHandle if file/directory
 * was located.
 */
typedef struct fs_MsgLookupResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // Returned file handle
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional parent dir and/or looked up
                                        // object attributes.
}  fs_MsgLookupResp;

/*-----------------------------------------------------------------------------
 * fs_MsgAccessReq - data structure used for filesystem ACCESS
 *                   transaction request messages.
 *
 * Client requests access permission check on file or directory.
 */
typedef struct fs_MsgAccessReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           fileHandle;     // Handle of object to check
    uint32              accessBits;     // Bitmask of permissions to test
#define FS_ACCESS_READ      0x01    // Read file data or dir contents
#define FS_ACCESS_LOOKUP    0x02    // Name lookup in a directory
#define FS_ACCESS_MODIFY    0x04    // Rewrite data or modify existing dirents
#define FS_ACCESS_EXTEND    0x08    // Write new data or add dirents
#define FS_ACCESS_DELETE    0x10    // Remove existing dirents
#define FS_ACCESS_EXECUTE   0x20    // Has execute permission on a file

}  fs_MsgAccessReq;

/*-----------------------------------------------------------------------------
 * fs_MsgAccessResp - data structure used for filesystem ACCESS
 *                    transaction response messages.
 *
 * Response includes accessBits to indicate which operations are
 * permitted.
 */
typedef struct fs_MsgAccessResp
{
    fs_MsgRespHdr       hdr;            // Message header
    uint32              accessBits;     // Bitmask of operations permitted
    int32               numAttrs;       // Number of attributes in attr[].
    fs_Attr             attr[0];        // Optional object attributes.

}  fs_MsgAccessResp;

/*-----------------------------------------------------------------------------
 * fs_MsgCreateReq - data structure used for filesystem CREATE
 *                   transaction request messages.
 *
 * Client requests a file be created.
 */
typedef struct fs_MsgCreateReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           parentHandle;   // File handle of parent directory
    union
    {
        fs_Attr     attr;       // Attributes of file. 
        uint64      verifier;   // Exclusive verification cookie
    } u;
    ushort16            flags;
#define FS_CREATE_UNCHECKED     0   // No error if file already exists,
                                    // behavior compatible with NFSv2.
#define FS_CREATE_GUARDED       1   // Error returned if file already exists.
#define FS_CREATE_EXCLUSIVE     2   // No error if file already exists _and_
                                    // verifier equals existing cookie for the
                                    // file.  No attr supplied in this mode.
    ushort16            nameLen;    // Length of name in bytes NOT counting NULL
                                    // terminator
    char                fileName[0];// NULL terminated name of file up to
                                    // FS_NAME_MAX

}  fs_MsgCreateReq;

/*-----------------------------------------------------------------------------
 * fs_MsgCreateResp - data structure used for filesystem CREATE
 *                    transaction response messages.
 *
 * Response includes completion status and fileHandle if create was
 * successful.
 */
typedef struct fs_MsgCreateResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // Returned file handle
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional pre/post file/dir attributes.

}  fs_MsgCreateResp;

/*-----------------------------------------------------------------------------
 * fs_MsgMknodReq - data structure used for filesystem MKNOD
 *                  transaction request messages.
 *
 * Client requests a special file be created.
 */
typedef struct fs_MsgMknodReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           parentHandle;   // File handle of parent directory
    fs_Attr             attr;           // Attributes of file. 
    fs_DevNum           devNum;         // Major/minor numbers if attr.type is
                                        // FS_FTYPE_BLK or FS_FTYPE_CHR.
    uint32              nameLen;        // Length of name in bytes NOT counting NULL
                                        // terminator.
    char                fileName[0];    // NULL terminated name of file up to
                                        // FS_NAME_MAX

}  fs_MsgMknodReq;

/*-----------------------------------------------------------------------------
 * fs_MsgMknodResp - data structure used for filesystem MKNOD
 *                   transaction response messages.
 *
 * Response includes completion status and fileHandle if mknod was
 * successful.
 */
typedef struct fs_MsgMknodResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // Returned file handle
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional pre/post file/dir attributes.

}  fs_MsgMknodResp;

/*-----------------------------------------------------------------------------
 * fs_MsgRemoveReq - data structure used for filesystem REMOVE
 *                   transaction request messages.
 *
 * Client requests file/directory be removed.  Works for files or directories.
 */
typedef struct fs_MsgRemoveReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           parentHandle;   // Handle of parent directory
    uint32              nameLen;        // Length of name in bytes NOT counting NULL
                                        // terminator.
    char                name[0];        // NULL terminated name of file/dir to
                                        // remove

}  fs_MsgRemoveReq;

/*-----------------------------------------------------------------------------
 * fs_MsgRemoveResp - data structure used for filesystem REMOVE
 *                    transaction response messages.
 *
 * Response includes completion status only. The directory entry
 * associated with fileHandle is removed (so any subsequent
 * lookups will fail).
 */
typedef struct fs_MsgRemoveResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // handle of removed inode
    int32               numAttrs;       // Number of attributes in attr[].
    boolean             wasDestroyed;	// File destroyed if last link.
    fs_Attr             attr[0];        // Optional pre/post parent dir attributes,
                                        // and post attributes for the inode if
                                        // the link count is non-zero.
}  fs_MsgRemoveResp;

/*-----------------------------------------------------------------------------
 * fs_MsgRmdirReq - data structure used for filesystem RMDIR
 *                  transaction request messages.
 *
 * Client requests directory be removed.  Works for directories only.
 */
typedef struct fs_MsgRmdirReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           parentHandle;   // Handle of parent directory
    uint32              nameLen;        // Length of name in bytes NOT counting NULL
                                        // terminator
    char                name[0];        // NULL terminated name of dir to remove

}  fs_MsgRmdirReq;

/*-----------------------------------------------------------------------------
 * fs_MsgRmdirResp - data structure used for filesystem RMDIR
 *                   transaction response messages.
 *
 * Response includes completion status only. The directory entry
 * associated with fileHandle is removed (so any subsequent
 * lookups will fail).
 */
typedef struct fs_MsgRmdirResp
{
    fs_MsgRespHdr       hdr;            // Message header
	fs_Handle			fileHandle;		// Handle of dir removed
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional pre/post parent dir attributes.

}  fs_MsgRmdirResp;

/*-----------------------------------------------------------------------------
 * fs_MsgHardLinkReq - data structure used for filesystem HARDLINK
 *                     transaction request messages.
 *
 * Client requests a hard link be created to an existing file.
 * (not supported for directories)
 */
typedef struct fs_MsgHardLinkReq
{
    fs_MsgReqHdr        hdr;                // Message header
    fs_Handle           fileHandle;         // Handle for file to link to
    fs_Handle           parentDirHandle;    // Handle of parent dir for link
    uint32              nameLen;            // Length of name in bytes NOT counting
                                            // NULL terminator
    char                linkName[0];        // NULL terminated name of link up to
                                            // FS_NAME_MAX

}  fs_MsgHardLinkReq;

/*-----------------------------------------------------------------------------
 * fs_MsgHardLinkResp - data structure used for filesystem HARDLINK
 *                      transaction response messages.
 *
 * Response includes completion status.
 */
typedef struct fs_MsgHardLinkResp
{
    fs_MsgRespHdr       hdr;                // Message header
    int32               numAttrs;           // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];            // Optional pre/post from/to directory
                                            // attributes.
}  fs_MsgHardLinkResp;

/*-----------------------------------------------------------------------------
 * fs_MsgSymLinkReq - data structure used for filesystem SYMLINK
 *                    transaction request messages.
 *
 * Client requests a symbolic link be created to a file or directory.
 */
typedef struct fs_MsgSymLinkReq
{
    fs_MsgReqHdr        hdr;                // Message header
    fs_Handle           parentHandle;       // Parent directory for linkName
    ushort16            linkNameLen;        // Length of name in bytes NOT counting
                                            // NULL terminator
    ushort16            linkToNameLen;      // Length of linkToName in bytes NOT
                                            // counting NULL terminator
    char                linkName[0];        // NULL terminated name of link to
                                            // create up to FS_NAME_MAX
    char                linkToName[0];      // NULL terminated name of file/dir to
                                            // link to up to FS_PATH_MAX
}  fs_MsgSymLinkReq;

/*-----------------------------------------------------------------------------
 * fs_msgSymLinkResp - data structure used for filesystem SYMLINK
 *                     transaction response messages.
 *
 * Response includes completion status only.   No attempt is made to
 * validate the existence of the linked-to file or directory.
 */
typedef struct fs_MsgSymLinkResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // Handle for symlink
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional post dir and pre/post parent
                                        // attributes.
}  fs_MsgSymLinkResp;

/*-----------------------------------------------------------------------------
 * fs_MsgRenameReq - data structure used for filesystem RENAME
 *                   transaction request messages.
 *
 * Client requests a file or directory be renamed.
 */
typedef struct fs_MsgRenameReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           fromDirHandle;  // Handle of current parent dir
    fs_Handle           toDirHandle;    // Handle of new parent dir
    ushort16            fromNameLen;    // Length of fromName in bytes NOT counting
                                        // NULL terminator
    ushort16            toNameLen;      // Length of toName in bytes NOT counting
                                        // NULL terminator
    char                fromName[0];    // NULL terminated From name up to
                                        // FS_NAME_MAX
    char                toName[0];      // NULL terminated To name up to FS_NAME_MAX

}  fs_MsgRenameReq;

/*-----------------------------------------------------------------------------
 * fs_MsgRenameResp - data structure used for filesystem RENAME
 *                    transaction response messages.
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgRenameResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           fileHandle;     // Handle of removed original "toName"
                                        // file, if its link count is non-zero.
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional pre/post from/to dir
                                        // attributes, and post original "toName"
                                        // attributes, if its link count is non-zero.
}  fs_MsgRenameResp;

/*-----------------------------------------------------------------------------
 * fs_MsgMkdirReq - data structure used for filesystem MKDIR
 *                  transaction request messages.
 *
 * Client requests a directory be created.
 */
typedef struct fs_MsgMkdirReq
{
    fs_MsgReqHdr        hdr;            // Message header
    fs_Handle           parentHandle;   // File handle of parent directory
    fs_Attr             dirAttr;        // Attributes of directory
    uint32              nameLen;        // Length of dirName in bytes NOT counting
                                        // NULL terminator
    char                dirName[0];     // NULL terminated name of directory up to
                                        // FS_NAME_MAX

}  fs_MsgMkdirReq;

/*-----------------------------------------------------------------------------
 * fs_MsgMkdirResp - data structure used for filesystem MKDIR
 *                   transaction response messages.
 *
 * Response includes completion status and dirHandle upon success.
 */
typedef struct fs_MsgMkdirResp
{
    fs_MsgRespHdr       hdr;            // Message header
    fs_Handle           dirHandle;      // Returned directory handle 
    int32               numAttrs;       // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];        // Optional post dir and pre/post parent
                                        // attributes.
}  fs_MsgMkdirResp;

/*-----------------------------------------------------------------------------
 * fs_MsgGetAttrReq - data structure used for filesystem GETATTR
 *                    transaction request messages. 
 *
 * Client requests GET of file or directory attributes.  
 */
typedef struct fs_MsgGetAttrReq
{
    fs_MsgReqHdr        hdr;        // Message header 
    fs_Handle           fileHandle; // Handle of file to get attributes for 

}  fs_MsgGetAttrReq;

/*-----------------------------------------------------------------------------
 * fs_MsgGetAttrResp - data structure used for filesystem GETATTR
 *                     transaction response messages. 
 *
 * Response includes completion status and attr upon success.
 */
typedef struct fs_MsgGetAttrResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    fs_Attr             attr;           // Returned file attributes. 

}  fs_MsgGetAttrResp;

/*-----------------------------------------------------------------------------
 * fs_MsgSetAttrReq - data structure used for filesystem SETATTR
 *                    transaction request messages. 
 *
 * Client requests SET of file or directory attributes.   Any settable
 * fields in attr that are -1 are ignored and will not be updated.
 */
typedef struct fs_MsgSetAttrReq
{
    fs_MsgReqHdr        hdr;        // Message header 
    fs_Handle           fileHandle; // Handle of file/dir to set attributes of 
    fs_Attr             attr;       // Caller supplied file/dir attributes. 
    fs_Time             checkCtime; // If non-NULL, we check it against object's
                                    // ctime and perform setAttr only if they
                                    // match.
}  fs_MsgSetAttrReq;

/*-----------------------------------------------------------------------------
 * fs_MsgSetAttrResp - data structure used for filesystem SETATTR
 *                     transaction response messages. 
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgSetAttrResp
{
    fs_MsgRespHdr       hdr;        // Message header 
    int32               numAttrs;   // Number of attributes in attr[].
    int32               pad;
    fs_Attr             attr[0];    // Optional updated file attributes.

}  fs_MsgSetAttrResp;

/*-----------------------------------------------------------------------------
 * fs_MsgReadDirReq - data structure used for filesystem READDIR
 *                    transaction request messages. 
 *
 * Client requests directory entries.  Client makes
 * READDIR requests iteratively.  Client uses cookie value
 * of 0 to start at first directory entry, otherwise uses
 * the cookie from a previous READDIR response.  Likewise
 * for cookieVerif (ignored for NFSv2).
 */
typedef struct fs_MsgReadDirReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    fs_Handle           fileHandle;     // Handle of directory
    uint64              cookieVerif;    // Verification cookie.
    uint64              cookie;         // Indicates place in directory
                                        // entry list to begin response
    int32               maxBytes;       // Maximum byte count of response
                                        // counting everything (NFS code
                                        // responsible for accounting for
                                        // XDR overhead).
}  fs_MsgReadDirReq;


/*-----------------------------------------------------------------------------
 * fs_MsgReadDirResp - data structure used for filesystem READDIR
 *                     transaction response messages. 
 *
 * Response includes completion status, dirEntry(ies).
 */
typedef struct fs_MsgReadDirResp
{
    fs_MsgRespHdr       hdr;                // Message header 
    fs_Attr             dirAttr;            // Attributes of directory
    uint64              cookieVerif;        // Used to verify validity of
                                            // cookie passed in subsequent
                                            // requests.
    uint32              eof;                // TRUE if end of dirents reached.
    uint32              numEntriesThis;     // Number of entries in dirEntry[]
    fs_DirEntry         dirEntry[0];        // Returned directory entry(ies) 
#define FS_READDIR_SPACE_MAX 2048   // How much space we pre-allocate to build
                                    // the dirEntry[] in the response msg.

// Advance a pointer to the next dirEntry[]
// position.  Accounts for variable name length
// and 8-byte alignment requirements.
#define FS_READDIR_ADV_ENTRY_PTR(_ptr_)                         \
    {                                                           \
        fs_DirEntry     *dent = (fs_DirEntry *) (_ptr_);        \
                                                                \
        _ptr_ = (void *)(((char *) _ptr_) +                     \
		 sizeof(fs_DirEntry) + dent->nameLen + 		\
		(sizeof(uint64) - 1));				\
	_ptr_ = (void *)((address_t)(_ptr_) & 			\
				~(sizeof (uint64) - 1)); 	\
    }

// Increment counter based on the next dirEntry[]
// position.  Accounts for variable name length
// and 8-byte alignment requirements.
#define FS_READDIR_COUNT_NEXT_ENTRY_PTR(_ptr_,_cnt_)            \
    {                                                           \
        fs_DirEntry     *dent = (fs_DirEntry *) (_ptr_);        \
                                                                \
        ((_cnt_)) += sizeof(fs_DirEntry);                       \
        ((_cnt_)) += dent->nameLen;                             \
        ((_cnt_)) += sizeof(uint64) - 1;                        \
        ((_cnt_)) &= ~(sizeof(uint64) - 1);                     \
    }

} fs_MsgReadDirResp;

/*-----------------------------------------------------------------------------
 * fs_MsgReadDirPlusReq - data structure used for filesystem READDIRPLUS
 *                        transaction request messages. 
 *
 * Client requests directory entries.  Client makes
 * READDIRPLUS requests iteratively.  Client uses cookie value
 * of 0 to start at first directory entry, otherwise uses
 * the cookie from a previous READDIRPLUS response.  Likewise
 * for cookieVerif (ignored for NFSv2).
 */
typedef struct fs_MsgReadDirPlusReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    fs_Handle           fileHandle;     // Handle of directory
    uint64              cookieVerif;    // Verification cookie.
    uint64              cookie;         // Indicates place in directory
                                        // entry list to begin response
    int32               maxDirBytes;    // Maximum byte count of response
                                        // counting only fileids and names.
    int32               maxBytes;       // Maximum byte count of response
                                        // counting everything (NFS code
                                        // responsible for accounting for
                                        // XDR overhead).
}  fs_MsgReadDirPlusReq;

/*-----------------------------------------------------------------------------
 * fs_MsgReadDirPlusResp - data structure used for filesystem READDIRPLUS
 *                         transaction response messages. 
 *
 * Response includes completion status, dirEntry(ies).
 */
typedef struct fs_MsgReadDirPlusResp
{
    fs_MsgRespHdr       hdr;                // Message header 
    uint64              cookieVerif;        // Used to verify validity of
                                            // cookie passed in subsequent
                                            // requests.
    fs_Attr             attr;               // Attributes of the directory.
    uint32              eof;                // TRUE if end of dirents reached.
    uint32              numEntriesThis;     // Number of entries in dirEntry[]
    fs_DirEntryPlus     dirEntry[0];        // Returned directory entry(ies) 
// Advance a pointer to the next dirEntry[]
// position.  Accounts for variable name length
// and 8-byte alignment requirements.
#define FS_READDIRPLUS_ADV_ENTRY_PTR(_ptr_)                     \
    {                                                           \
        fs_DirEntryPlus     *dent = (fs_DirEntryPlus *) (_ptr_);\
                                                                \
        ((char *)(_ptr_)) += sizeof(fs_DirEntryPlus);           \
        ((char *)(_ptr_)) += dent->dirEntry.nameLen;            \
        ((char *)(_ptr_)) += sizeof(uint64) - 1;                \
	_ptr_ = (void *)((address_t)(_ptr_) & ~(sizeof (uint64) - 1)); \
    }
// Increment counter based on the next dirEntry[]
// position.  Accounts for variable name length
// and 8-byte alignment requirements.
#define FS_READDIRPLUS_COUNT_NEXT_ENTRY_PTR(_ptr_,_cnt_)        \
    {                                                           \
        fs_DirEntryPlus     *dent = (fs_DirEntryPlus *) (_ptr_);\
                                                                \
        ((_cnt_)) += sizeof(fs_DirEntryPlus);                   \
        ((_cnt_)) += dent->dirEntry.nameLen;                    \
        ((_cnt_)) += sizeof(uint64) - 1;                        \
        ((_cnt_)) &= ~(sizeof(uint64) - 1);                     \
    }

}  fs_MsgReadDirPlusResp;

/*-----------------------------------------------------------------------------
 * fs_MsgReadLinkReq - data structure used for filesystem READLINK
 *                     transaction request messages. 
 *
 * Client requests a read of a symbolic link.
 */
typedef struct fs_MsgReadLinkReq
{
    fs_MsgReqHdr        hdr;                // Message header 
    fs_Handle           fileHandle;         // Handle of symbolic link to read

}  fs_MsgReadLinkReq;

/*-----------------------------------------------------------------------------
 * fs_MsgReadLinkResp - data structure used for filesystem READLINK
 *                      transaction response messages. 
 *
 * Response includes completion status and the contents
 * of the symbolic link.
 */
typedef struct fs_MsgReadLinkResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    fs_Attr             attr;           // Updated file attributes.
    uint32              nameLen;        // Length of linkName in bytes NOT counting
                                        // NULL terminator
    char                linkName[0];    // NULL terminated link name up to
                                        // FS_PATH_MAX

}  fs_MsgReadLinkResp;

/*-----------------------------------------------------------------------------
 * fs_MsgTransmitComplete - data structure used for TX_COMPLETE
 *                          transaction messages. 
 *
 * Client reports that a packet transmission that used a
 * filesystem buffer (COD) is complete.  The filesystem is now
 * permitted to free the buffer back to the system. There
 * is no response message for this transaction.
 *
 * TODO - should one of these be dropped on its way to us,
 * we'll keep the buffer(s) locked forever - perhaps an ack
 * is a good idea?
 */
typedef struct fs_MsgTransmitComplete
{
    fs_MsgReqHdr       	hdr;            // Message header 
    fs_Handle           fileHandle;     // Filehandle from original request.
    uint64              fileOffset;     // Byte offset to start read action
    uint32              bytesRead;      // Number of bytes actually read
	uint64				fsPrivate;		// Returned to FS from read response.
										// This is the state machine handle
										// that is signalled.
}  fs_MsgTransmitComplete;

/*-----------------------------------------------------------------------------
 * fs_Stat - Used to supply the stats during a filesystem
 *           STAT request.
 */
typedef struct fs_Stat
{
    uint32          fsBlockSize;    // Size in bytes of filesystem block
    uint32          pad;
    uint64          numBlks;        // Number of fsBlockSize blocks in fs
    uint64          numBlksAvail;   // Number of fsBlockSize blocks available
    uint64          numBlksAvailUser;// Number of fsBlockSize blocks available
                                    // to non-priviliged users.
    uint64          numInodes;      // Total number of file slots (inodes)
    uint64          numInodesAvail; // Total number of file slots free.
    uint64          numInodesAvailUser; // Total number of file slots free to
                                    // non-priviliged users.
    uint32          readsPerSec;    // Average reads/sec over TODO interval
    uint32          writesPerSec;   // Average writes/sec over TODO interval
    uint32          avgReadSize;    // Average read size over TODO interval
    uint32          avgWriteSize;   // Averag write size over TODO interval 

} fs_Stat;

/*-----------------------------------------------------------------------------
 * fs_Fsinfo - Used to supply the fsinfo during a filesystem
 *             FSINFO request.
 */
typedef struct fs_Fsinfo
{
    uint64      maxFileSize;        // Maximum filesize supported in bytes.
    uint32      readMax;            // Max bytes in read request.
    uint32      readPref;           // Preferred bytes in read request.
    uint32      readMult;           // Suggested multiple for size of read.
    uint32      writeMax;           // Max bytes in write request.
    uint32      writePref;          // Preferred bytes in write request.
    uint32      writeMult;          // Suggested multiple for size of write.
    uint32      readDirPref;        // Preferred bytes in readdir request.
    fs_Time     timeDelta;          // Time granularity supported.
    uint32      fsProperties;       // Properties supported.
#define FS_PROP_LINK        0x01    // Supports hard links.
#define FS_PROP_SYMLINK     0x02    // Support symbolic links.
#define FS_PROP_HOMO        0x08    // PATHCONF identical for all files/dirs.
#define FS_PROP_CANSETTIME  0x10    // Time can be set with SETATTR

} fs_Fsinfo;

/*-----------------------------------------------------------------------------
 * fs_Pathconf - Used to supply the fsinfo during a filesystem
 *               PATHCONF request.
 */
typedef struct fs_Pathconf
{
    uint32      linkMax;            // Maximum number of hard links to object.
    uint32      nameMax;            // Maximum length of filename component.
    uchar8      noTrunc;            // Will/won't truncate names > max.
    uchar8      chownRestricted;    // Only uid 0 can chown/chgrp.
    uchar8      caseInsensitive;    // Case doesn't matter to server.
    uchar8      casePreserving;     // Case preserved by server.
    
} fs_Pathconf;

/*-----------------------------------------------------------------------------
 * fs_MsgGetRootReq - data structure used for filesystem GETROOT
 *                    transaction request messages.
 *
 * Client requests root handle for a mount point in a filesystem.
 */
typedef struct fs_MsgGetRootReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume filesystem lives on
    uint32              pathLen;        // Length of pathName in bytes NOT
                                        // counting NULL terminator
    char                pathName[0];    // NULL terminated name of mount point.
                                        // If not supplied, the root directory of
                                        // the volume is assumed.
}  fs_MsgGetRootReq;

/*-----------------------------------------------------------------------------
 * fs_MsgGetRootResp - data structure used for filesystem GETROOT
 *                     transaction response messages.
 *
 * Response includes completion status and dirHandle upon success.
 */
typedef struct fs_MsgGetRootResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    fs_Handle           dirHandle;      // Returned directory handle 

}  fs_MsgGetRootResp;

/*-----------------------------------------------------------------------------
 * fs_MsgSyncReq - data structure used for filesystem SYNC
 *                 transaction request messages. 
 *
 * Client requests filesystem state be synchronized to disk.
 * Client provides fsHandle.
 */
typedef struct fs_MsgSyncReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume filesystem lives on

}  fs_MsgSyncReq;

/*-----------------------------------------------------------------------------
 * fs_MsgSyncResp - data structure used for filesystem SYNC
 *                  transaction response messages. 
 *
 * Response includes completion status only.   The response can
 * be delayed significantly due to synchronization of filesystem
 * state to disk.
 */
typedef struct fs_MsgSyncResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgSyncResp;

/*-----------------------------------------------------------------------------
 * fs_MsgMountReq - data structure used for filesystem MOUNT
 *                  transaction request messages. 
 *
 * Client requests filesystem be mounted.  Client provides
 * fsName and volumeID. This request must be successfully
 * completed before any other requests can be carried out.
 */

typedef enum fs_LogDevType
{
    FS_LOG_DEVTYPE_NULL     = 0,   // NO Logging, skip all write log actions.
    FS_LOG_DEVTYPE_NVRAM    = 1,   // Logging to the FP NVRAM.
    FS_LOG_DEVTYPE_SELF     = 2,   // Logging to filesystem log area.
    FS_LOG_DEVTYPE_VOLID    = 3,   // Logging to separate raw volume.
    FS_LOG_DEVTYPE_DEFAULT  = 4,   // Only used in mount req: indicates same as before
    FS_LOG_DEVTYPE_MAX      = 4
} fs_LogDevType; 

typedef struct fs_MsgMountReq
{
    fs_MsgReqHdr        hdr;        // Message header 
    evm_volId_t         volId;      // Volume filesystem lives on

    uint32              logDevType;
    uint32              logSize;    // Size of external device
                                    //   only valid for FS_LOG_DEVTYPE_VOLID
    evm_volId_t         logVolId;   // Volume id for ext vol logging
                                    //   only valid for FS_LOG_DEVTYPE_VOLID

    int32               rdonly;     // is filesystem read only mount
    int32               eek;		// is filesystem being mounted by eek?
    uint32              fsNameLen;  // Length of fsName in bytes NOT counting NULL
                                    // terminator
    uint32              skipReplay; // Do want to replay or skip this
                                    //   if skipped: just try to scan 
    uint32              skipEek;    // Do want to eek or skip this
    
    char                fsName[0];  // NULL terminated name of filesystem up to
                                    // FS_NAME_MAX

}  fs_MsgMountReq;

/*-----------------------------------------------------------------------------
 * fs_MsgMountResp - data structure used for filesystem MOUNT
 *                   transaction response messages. 
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgMountResp
{
    fs_MsgRespHdr       hdr;        // Message header 

}  fs_MsgMountResp;

/*-----------------------------------------------------------------------------
 * fs_MsgUnmountReq - data structure used for filesystem UNMOUNT
 *                    transaction request messages. 
 *
 * Client requests filesystem be unmounted.  Client provides
 * fsName.  Filesystem will be frozen, outstanding transactions
 * will be completed, and filesystem state will be synchronized to
 * disk.
 */
typedef struct fs_MsgUnmountReq
{
    fs_MsgReqHdr        hdr;        // Message header 
    evm_volId_t         volId;      // Block device to act on

}  fs_MsgUnmountReq;

/*-----------------------------------------------------------------------------
 * fs_MsgUnmountResp - data structure used for filesystem UNMOUNT
 *                     transaction response messages. 
 *
 * Response includes completion status only.  The response can be
 * delayed significantly due to synchronization of filesystem
 * state to disk.
 */
typedef struct fs_MsgUnmountResp
{
    fs_MsgRespHdr       hdr;        // Message header 

}  fs_MsgUnmountResp;

/*-----------------------------------------------------------------------------
 * fs_MsgStatReq - data structure used for filesystem STAT
 *                 transaction request messages. 
 *
 * Client requests stat entry for a filesystem.
 */
typedef struct fs_MsgStatReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    fs_Handle           fsHandle;       // Handle of any object in
                                        // filesystem to stat 
}  fs_MsgStatReq;

/*-----------------------------------------------------------------------------
 * fs_MsgStatResp - data structure used for filesystem STAT
 *                  transaction response messages. 
 *
 * Response includes completion status and stat.
 */
typedef struct fs_MsgStatResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    fs_Attr             attr;           // Attributes of object used in
                                        // request.
    fs_Stat             stat;           // Returned filesystem stats 

}  fs_MsgStatResp;

/*-----------------------------------------------------------------------------
 * fs_MsgFsinfoReq - data structure used for filesystem FSINFO
 *                   transaction request messages. 
 *
 * Client requests fsinfo entry for a filesystem.
 */
typedef struct fs_MsgFsinfoReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    fs_Handle           fsHandle;       // Handle of any object in
                                        // filesystem to stat 
}  fs_MsgFsinfoReq;

/*-----------------------------------------------------------------------------
 * fs_MsgFsinfoResp - data structure used for filesystem FSINFO
 *                    transaction response messages. 
 *
 * Response includes completion status and fsinfo.
 */
typedef struct fs_MsgFsinfoResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    fs_Attr             attr;           // Attributes of object used in
                                        // request.
    fs_Fsinfo           fsinfo;         // Returned filesystem fsinfo

}  fs_MsgFsinfoResp;

/*-----------------------------------------------------------------------------
 * fs_MsgPathconfReq - data structure used for filesystem PATHCONF
 *                     transaction request messages. 
 *
 * Client requests pathconf entry for an object.
 */
typedef struct fs_MsgPatconfReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    fs_Handle           fsHandle;       // Handle of object to get pathconf
                                        // for. 
}  fs_MsgPathconfReq;

/*-----------------------------------------------------------------------------
 * fs_MsgPathconfResp - data structure used for filesystem PATHCONF
 *                      transaction response messages. 
 *
 * Response includes completion status and pathconf.
 */
typedef struct fs_MsgPathconfResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    fs_Attr             attr;           // Attributes of object used in
                                        // request.
    fs_Pathconf         pathconf;       // Returned pathconf info.

}  fs_MsgPathconfResp;

/*-----------------------------------------------------------------------------
 * fs_MsgCommitReq - data structure used for filesystem COMMIT
 *                   transaction request messages. 
 *
 * Client requests modifications be written to stable storage.
 * 
 * NOTE:  This is a no-op.  We always write synchronously, so this
 *        message should not currently be used.  It can be used to
 *        obtain updated file attributes, however.
 */
typedef struct fs_MsgCommitReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    fs_Handle           fsHandle;       // Handle of object to commit.
    uint64              offset;         // File offset to start flush from.
    uint32              count;          // Num bytes to flush, 0 indicates to
                                        // end of file.
}  fs_MsgCommitReq;

/*-----------------------------------------------------------------------------
 * fs_MsgCommitResp - data structure used for filesystem COMMIT
 *                    transaction response messages. 
 *
 * Response includes completion status.
 */
typedef struct fs_MsgCommitResp
{
    fs_MsgRespHdr       hdr;            // Message header 
    int32               numAttrs;       // Number of attributes in attr[].
    uint32              verifier;       // Commit verifier cookie.
    fs_Attr             attr[0];        // Optional updated file attributes.

}  fs_MsgCommitResp;


/**
 * Create snapshot request.
 */
typedef struct {
    fs_MsgReqHdr    hdr;                // Message header 
    evm_volId_t     volId;              // Block device to act on
    uint16          nameLen;            // Length of snapDirName in bytes NOT
                                        // counting NULL terminator
    char            snapDirName[0];     // NULL terminated snapshot dir name up
                                        // to FS_NAME_MAX
} fs_MsgSnapCreateReq;


/**
 * Create snapshot reply.
 */
typedef struct {
    fs_MsgRespHdr       hdr;            // Message header 
} fs_MsgSnapCreateResp;


/**
 * Remove snapshot request.
 */
typedef struct {
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Block device to act on
    uint16              nameLen;        // Length of snapDirName in bytes NOT
                                        // counting NULL terminator
    char                snapDirName[0]; // NULL terminated name of snapshot up
                                        // to FS_NAME_MAX
} fs_MsgSnapRemoveReq;


/**
 * Remove snapshot reply.
 */
typedef struct {
    fs_MsgRespHdr       hdr;            // Message header 
} fs_MsgSnapRemoveResp;


/**
 * Revert snapshot request.
 */
typedef enum fs_SnapRevertType
{
    FS_SNAP_REVERT_NAME,                // Revert using this snapname
    FS_SNAP_REVERT_ID,                  // Revert using an internal snapshot
    FS_SNAP_REVERT_NEWEST               // Revert using newest snapshot
} fs_SnapRevertType;

typedef struct {
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Block device to act on
    fs_SnapRevertType   type;
    int32               snapnum;        // Internal snapnum to revert to
    uint16              nameLen;        // Length of snapname in bytes NOT
                                        // counting NULL terminator
    char                snapname[0];    // NULL terminated name of snapshot up
                                        // to FS_NAME_MAX
} fs_MsgSnapRevertReq;


/**
 * Revert snapshot reply.
 */
typedef struct {
    fs_MsgRespHdr       hdr;            // Message header 
} fs_MsgSnapRevertResp;


/**
 * Snapshot enable/disable request.
 */
typedef struct {
    fs_MsgReqHdr    hdr;                // Message header
    evm_volId_t     volId;              // Block device to act on
    boolean         enable;             // FALSE to disable snapshots
} fs_MsgSnapStateReq;


/**
 * Snapshot enable/disable reply.
 */
typedef struct {
      fs_MsgRespHdr       hdr;            // Message header
} fs_MsgSnapStateResp;


/**
 * Snapshot pin/unpin request.
 */
typedef struct {
    fs_MsgReqHdr    hdr;                // Message header
    evm_volId_t     volId;              // Block device to act on
    boolean         pin;                // TRUE to pin snapshots
                                        // FALSE to unpin
    uint32          nameLen;            // Length of snapname in bytes NOT
                                        // counting NULL terminator
    char            snapname[0];        // NULL terminated name of snapshot up
                                        // to FS_NAME_MAX
} fs_MsgSnapPinReq;


/**
 * Snapshot pin/unpin reply.
 */
typedef struct {
      fs_MsgRespHdr       hdr;            // Message header
} fs_MsgSnapPinResp;


/**
 * Snapshot rename request.
 */
typedef struct {
    fs_MsgReqHdr    hdr;                // Message header
    evm_volId_t     volId;              // Block device to act on
    char            newname[FS_NAME_MAX];
    char            oldname[FS_NAME_MAX];
} fs_MsgSnapRenameReq;


/**
 * Snapshot rename reply.
 */
typedef struct {
      fs_MsgRespHdr       hdr;            // Message header
} fs_MsgSnapRenameResp;


/**
 * Show snapshot request.
 */

#define FS_MAX_USER_SNAPSHOTS       48  // maximum number of user snapshots
#define FS_MAX_INTERNAL_SNAPSHOTS   16  // maximum number of internal snapshots
#define FS_SNAPSHOT_MAX             64  // total

#define FS_SNAP_SCHED_MAX_HOURS 24      // maximum number of user specified hours
#define FS_SNAPNAME_MAX         24      // truncated if longer

typedef enum fs_SnapShowType
{
     FS_SNAP_SHOW_SCHED,    // Show current create/remove schedule
     FS_SNAP_SHOW_USAGE     // Show snapshot usage for a snapshot
} fs_SnapShowType;

typedef struct {
    fs_MsgReqHdr        hdr;            // Message header 
    fs_SnapShowType     type;
    evm_volId_t         volId;
    char                snapname[FS_SNAPNAME_MAX];  // for FS_SNAP_SHOW_USAGE
} fs_MsgSnapShowReq;


/**
 * Show snapshot reply.
 */

typedef struct {
    boolean     enable;
    boolean     maxWeeklySet;
    boolean     maxDailySet;
    boolean     maxHourlySet;
    boolean     hoursSet;
    uint32      maxWeekly;
    uint32      maxDaily;
    uint32      maxHourly;
    uint32      hours[FS_SNAP_SCHED_MAX_HOURS]; // 1-24 midnight is 24
} fs_SnapSched;

typedef struct {
    char        name[FS_SNAPNAME_MAX];  // name (not applicable if internal)
    int32       snapnum;                // snapnum
    boolean     pin;                    // pin status
    uint64      size;                   // number of blocks used by this snapshot
                                        //   that are different from the live filesystem
    fs_Time     ctime;                  // create time
} fs_SnapUsage;

typedef struct {
    fs_MsgRespHdr       hdr;            // Message header
    evm_volId_t         volId;
    fs_SnapShowType     type;
    fs_SnapSched        sched;          // if FS_SNAP_SHOW_SCHED
                                        // if FS_SNAP_SHOW_USAGE,
    uint64              totalBlks;      // total number of blocks in fs
    uint64              blksUsed;       // number of blocks used in fs
                                        // if FS_SNAP_GET_NAMES
    uint32              nUsers;         // number of user snapshots
    fs_SnapUsage        user[FS_MAX_USER_SNAPSHOTS];
    uint32              nInternals;     // number of internal snapshots
    fs_SnapUsage        internal[FS_MAX_INTERNAL_SNAPSHOTS];
} fs_MsgSnapShowResp;


/**
 * Schedule snap request.
 */

typedef struct {
    fs_MsgReqHdr    hdr;                // Message header
    evm_volId_t     volId;
    fs_SnapSched    sched;
} fs_MsgSnapSchedReq;

/**
 * Schedule snap reply.
 */

typedef struct {
    fs_MsgRespHdr   hdr;
} fs_MsgSnapSchedResp;

/**
 * Admin snapshot request.
 */

#define FS_SNAP_MIDNIGHT    24
#define FS_SNAP_ON_THE_HOUR 60

typedef enum fs_SnapAdminType {
    FS_SNAP_ADMIN_SCHED,                // Schedule snapshots
    FS_SNAP_ADMIN_RENAME,               // Rename snapshot
    FS_SNAP_ADMIN_SET_INTERNAL          // Set internal snapshot params
} fs_SnapAdminType;

typedef enum fs_SnapWeekDayType {
    FS_SNAP_MONDAY =    1,
    FS_SNAP_TUESDAY =   2,
    FS_SNAP_WEDNESDAY = 3,
    FS_SNAP_THURSDAY =  4,
    FS_SNAP_FRIDAY =    5,
    FS_SNAP_SATURDAY =  6,
    FS_SNAP_SUNDAY =    7
} fs_SnapWeekDayType;

typedef struct {                        // current time
    fs_SnapWeekDayType  wday;
    uint32              hour;           // 1-24, midnight is 24
    uint32              min;            // 1-60, 60 is on the hour
} fs_SnapTime;

typedef struct {                        // internal snapshot management
    evm_volId_t volId;
    int32       internalTimeout;        // interval in minutes
    int32       maxInternalSnaps;       // max number to keep
} fs_SnapInternalParams;

typedef struct {
    fs_MsgReqHdr        hdr;            // Message header
    fs_SnapAdminType    type;
    union {
        fs_SnapTime     time;               // for FS_SNAP_ADMIN_SCHED
        char            name[FS_NAME_MAX];  // for FS_SNAP_ADMIN_RENAME
        char            new_name[FS_NAME_MAX];
        fs_SnapInternalParams snapInternals;// for FS_SNAP_ADMIN_SET_INTERNAL
    } data;
} fs_MsgSnapAdminReq;

/**
 * Admin snapshot reply.
 */

typedef struct {
    fs_MsgRespHdr   hdr;
} fs_MsgSnapAdminResp;


/*-----------------------------------------------------------------------------
 * fs_MsgMakeFsReq - data structure used for filesystem MAKEFS
 *                   transaction request messages. 
 *
 * Client requests a filesystem be created.
 */
typedef struct fs_MsgMakeFsReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Block device to mkfs on
    uint64              numBlks;        // Number of blkSize blocks
                                        // in the volume
    int32               blkSize;        // Allocation block size
    int32               ext_log;        // Use an ext dev for logging
                                        // Default = FALSE = reserve space
                                        // at end of volume to log

}  fs_MsgMakeFsReq;

/*-----------------------------------------------------------------------------
 * fs_MsgMakeFsResp - data structure used for filesystem MAKEFS
 *                    transaction response messages. 
 *
 * Response includes completion status.
 */
typedef struct fs_MsgMakeFsResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgMakeFsResp;

/*-----------------------------------------------------------------------------
 * fs_MsgViewFsReq - data structure used to req info about all filesystems
 *
 */
typedef struct fs_MsgViewFsReq
{
    fs_MsgReqHdr        hdr;            // Message header
    evm_volId_t         volId;          // Volume id
#define  FS_VIEW_LIST_OF_VOLUMES    1 
#define  FS_VIEW_ONE_VOLUME         2
    uint32              type;           // what to view 
    uint32              clientprivate;  // used to return info to requestor
}  fs_MsgViewFsReq;

/*-----------------------------------------------------------------------------
 * fs_MsgViewFsResp - data structure used give info about all filesystems
 */

#define FS_SUPER_NBR_MOUNTTIME    16

typedef struct fs_FSInfo
{
    uint32              length;       // length of this entry, aligned to 64bit
    uint32              state;        // see: fs_FileSystemState
    evm_volId_t         volId;        // Volume this fs mounted on.
    fs_Stat             stat;         // Returned filesystem stats 
    uint32              mountTime[FS_SUPER_NBR_MOUNTTIME];   // time of last N mount operations
    uint32              mountState[FS_SUPER_NBR_MOUNTTIME];  // status of last N mount operations
    uint32              mountFail[FS_SUPER_NBR_MOUNTTIME];   // reason why it failed
    uint32              umountTime[FS_SUPER_NBR_MOUNTTIME];  // time of last N controlled umount operations
    uint32              umountState[FS_SUPER_NBR_MOUNTTIME]; // status of last N controlled umount operations
    uint32              mkfsTime;     // See fs_FileSystem.super
    uint32              nbrMount;
    uint32              nbrUmount;
    uint32              nbrDirtyMount;
    uint32              nbrMountFailedLogDev;
    uint32              nbrMountFailedReplay;
    uint32              nbrMountFailedEek;
    uint32              nbrMountFailedSnap;
    uint32              nameLength;
    uint8               percentMinFree; // autogrow high water mark
    char                mountName[0]; // symbolic name fs mounted as
} fs_FSInfo;


typedef struct fs_MsgViewFsResp
{
    fs_MsgRespHdr       hdr;            // Message header 

    uint32              totalNbrFs;     // total nbr of known file systems
    uint32              nbrFs;          // nbr of FS in this msg (msgsize < 1492)
    uint32              type;           // what to view 
    uint32              clientprivate;  // used to return info to requestor
    union
    {
        fs_FSInfo           fsInfo[0];
        evm_volId_t         volId[0];
    } u;

}  fs_MsgViewFsResp;

/*-----------------------------------------------------------------------------
 * fs_MsgCheckFsReq - data structure used for filesystem CHECKFS
 *                    transaction request messages. 
 *
 * Client requests a filesystem be checked.  Volume must not be
 * mounted for reliable results.
 */
typedef struct fs_MsgCheckFsReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume to check

}  fs_MsgCheckFsReq;

/*-----------------------------------------------------------------------------
 * fs_MsgCheckFsResp - data structure used for filesystem CHECKFS
 *                     transaction response messages. 
 *
 * Response includes completion status.
 */
typedef struct fs_MsgCheckFsResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgCheckFsResp;

/*-----------------------------------------------------------------------------
 * fs_MsgTraceReq - data structure used for TRACE
 *                  transaction request messages. 
 */
typedef struct fs_MsgTraceReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    uint32              level;          // The level to set/clear.
    int32               set;            // Set or clear the level.

}  fs_MsgTraceReq;

/*-----------------------------------------------------------------------------
 * fs_MsgTraceResp - data structure used for TRACE
 *                   transaction response messages. 
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgTraceResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgTraceResp;

/*-----------------------------------------------------------------------------
 * fs_MsgShowLockedReq - data structure used for SHOW_LOCKED
 *                       transaction request messages. 
 */
typedef struct fs_MsgShowLockedReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume to show locked structs on.

}  fs_MsgShowLockedReq;

/*-----------------------------------------------------------------------------
 * fs_MsgShowLockedResp - data structure used for SHOW_LOCKED
 *                        transaction response messages. 
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgShowLockedResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgShowLockedResp;

/*-----------------------------------------------------------------------------
 * fs_MsgEmapdebugReq - data structure used for extent map debugging
 */
typedef struct fs_MsgEmapdebugReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume to debug emaps on

}  fs_MsgEmapdebugReq;

/*-----------------------------------------------------------------------------
 * fs_MsgEmapdebugResp - data structure used for extent map debugging
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgEmapdebugResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgEmapdebugResp;

/*-----------------------------------------------------------------------------
 * fs_MsgZapReq - data structure used for introducing inconsistencies into
 * a filesystem; used for debugging eek.
 */
typedef struct fs_MsgZapReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume to zap
    char action;
    uint64 inum;
    uint64 data;
}  fs_MsgZapReq;

/*-----------------------------------------------------------------------------
 * fs_MsgZapResp - 
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgZapResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgZapResp;

enum {FS_EEK_QUERY_NO, FS_EEK_QUERY_YES, FS_EEK_QUERY_PROMPT, FS_EEK_QUERY_QUIT};

/*-----------------------------------------------------------------------------
 * fs_MsgEekReq - data structure used for eek debugging
 */
typedef struct fs_MsgEekReq
{
    fs_MsgReqHdr    hdr;            // Message header 
    evm_volId_t     volId;          // Volume to debug emaps on
    int			    query_reply;	// One from enum above.
    int             default_reply;  // Default reply to all eek queries
    // int	            phase;          // which phase of eek to perform (0==all)
    // int             nthreads;       // how many threads to spawn in parallel
    uint64          private;        // SM to wake up on continuation

    fs_MsgMountReq	mountreq;	// mountreq for eek on unmounted fs

}  fs_MsgEekReq;

/*-----------------------------------------------------------------------------
 * fs_MsgEekResp - data structure used for eek debugging
 *
 * Response includes possible "ynq" query.
 */
typedef struct fs_MsgEekResp
{
    fs_MsgRespHdr       hdr;            // Message header
    boolean             has_query;
    boolean             has_status;
    boolean             quit;
    boolean             pad;
    uint64              private;
    char msg[256];

}  fs_MsgEekResp;

/*-----------------------------------------------------------------------------
 * fs_MsgFlushReq - data structure used for flushing the buffer cache
 */
typedef struct fs_MsgFlushReq
{
    fs_MsgReqHdr        hdr;            // Message header 
    evm_volId_t         volId;          // Volume to flush

}  fs_MsgFlushReq;

/*-----------------------------------------------------------------------------
 * fs_MsgFlushResp - data structure used for flushing the buffer cache
 *
 * Response includes completion status only.
 */
typedef struct fs_MsgFlushResp
{
    fs_MsgRespHdr       hdr;            // Message header 

}  fs_MsgFlushResp;


/**
 * Freeze filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to quiesce
} fs_MsgFreezeReq;


/**
 * Freeze filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgFreezeResp;


/**
 * Igrow filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to quiesce
} fs_MsgIgrowReq;


/**
 * Igrow filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgIgrowResp;

/**
 * Refbyte recount filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to quiesce
} fs_MsgRefByteCountReq;   


/**
 * Refbyte recount filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgRefByteCountResp;

/**
 * Inode recount filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to quiesce
} fs_MsgInodeCountReq;   


/**
 * Inode recount filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgInodeCountResp;

/**
 * Growfs filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to quiesce
    uint64 newNumBlks;		// the new size of the filesystem in blocks
} fs_MsgGrowfsReq;


/**
 * Growfs filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgGrowfsResp;



/**
 * Thaw filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to unquiesce
} fs_MsgThawReq;


/**
 * Thaw filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgThawResp;


/**
 * Volume management calls this when an I/O request has completed.
 */
extern void fs_ioDone(
    evm_IOStat_t    *stat);


/*
 * Quota related API
 */

/*
 * valid values of the flag that quota block can have.
 * quota is either AGGREGATE, USER or GROUP
 * quota can be either DEFAULT or USE_DEFAULTS or none but not both 
 */

#define	FS_QUOTA_AGGREGATE		0x001
#define	FS_QUOTA_USER			0x002
#define	FS_QUOTA_GROUP			0x004
#define	FS_QUOTA_DEFAULT		0x080
#define FS_QUOTA_USE_DEFAULTS   0x100
#define	FS_QUOTA_ON				0x8000	// Used only for reporting back
#define	FS_QUOTA_TURN_ON		0x4000	// Used only for reporting back

/*
 * To send quota exception to event manager
 */

#define	FS_EXCEED_SOFTLIMIT	0x80000000
#define	FS_EXCEED_HARDLIMIT	0x40000000
#define	FS_EXCEED_TIMELIMIT	0x20000000
#define	FS_EXCEED_INODE		0x00000001
#define	FS_EXCEED_BLOCK		0x00000002

#define	FS_EXCEED_INODE_SOFTLIMIT (FS_EXCEED_SOFTLIMIT | FS_EXCEED_INODE)
#define	FS_EXCEED_INODE_HARDLIMIT (FS_EXCEED_HARDLIMIT | FS_EXCEED_INODE)
#define	FS_EXCEED_INODE_TIMELIMIT (FS_EXCEED_TIMELIMIT | FS_EXCEED_INODE)
#define	FS_EXCEED_BLOCK_SOFTLIMIT (FS_EXCEED_SOFTLIMIT | FS_EXCEED_BLOCK)
#define	FS_EXCEED_BLOCK_HARDLIMIT (FS_EXCEED_HARDLIMIT | FS_EXCEED_BLOCK)
#define	FS_EXCEED_BLOCK_TIMELIMIT (FS_EXCEED_TIMELIMIT | FS_EXCEED_BLOCK)

/*
 * Infinite quota
 */

#define	FS_QUOTA_UNLIMITED		0

/*
 * Quota block structure used to communicate quota and consumption information
 * between SSC and the file system
 */

typedef struct fs_QuotaBlk {
	uint16		flag;				// Qualify the quota block
	uint64		id;					// fileset id or uid or gid
	uint64      rootInode;			// inode number of root of the fileset (for
                                    // fileset only)
	uint64		inodeSoftLimit;		// soft limit for num inodes
	uint64		inodeHardLimit;		// hard limit for num inodes
	uint32		inodeTimeLimit;		// time limit in sec for inode
									// over use over soflimit
	uint64		blockSoftLimit;		// soft limit for num blocks
	uint64		blockHardLimit;		// hard limit for num blocks
	uint32		blockTimeLimit;		// time limit in sec for block
									// over use over soflimit
	uint64		inodesConsumed;		// number of inodes consumed
	uint64		blocksConsumed;		// number of blocks consumed
									// was crossed
	uint32		inodeSoftLimitTime;	// time when inode softlimit
									// was crossed
	uint32		blockSoftLimitTime;	// time when block softlimit
									// was crossed
} fs_QuotaBlk_t;

/*
 * Set Quota message from SSC to the filesystem
 */

#define FS_SETQUOTA_MAX_QUOTABLKS	1

typedef struct fs_MsgSetQuotaReq {
	fs_MsgReqHdr		msgHdr;				// Request Header
	evm_volId_t			fsid;				// Volume Id of the file system to
											// which this applies
	char8				path[FS_PATH_MAX];	// pathname of fileset
	uint32				numberLimits;		// number of quotablks
											// restrict to 1 for now
	fs_QuotaBlk_t		limits[1];			// array of quotablks
} fs_MsgSetQuotaReq_t;

/*
 * Set Quota response from file system to SSC
 */

/*
 * Valid values for status are
 * FS_ERR_NONE	- success
 * FS_ERR_INVAL	- request was invalid
 * FS_ERR_PERM	- permission to perform operation denied
 *		- Never returned for now as we trust ssc to do authentication
 * FS_ERR_NOENT	- Fileset was non existent
 * FS_ERR_NODEV - File sysem was non existent
 */


typedef	struct	fs_SetQuotaStatus {
	uint32			index;	// position of request in limits array
	fs_Status		status;	// status
} fs_SetQuotaStatus_t;

typedef struct	fs_MsgSetQuotaResp {
	fs_MsgRespHdr		msgHdr;				// Response header
	uint32				numberResponses;	// Number of responses
											// limit to 1 for now
	fs_SetQuotaStatus_t	responses[1];		// array of responses
} fs_MsgSetQuotaResp_t;

/*
 * Get Quota requests from SSC to file system
 */

/*
 * Request defination as part of get quota message
 */

/*
 * Valid values for flag are
 * FS_QUOTA_USER	- return quota information for the user
 * FS_QUOTA_GROUP	- return quota information for the group
 * FS_QUOTA_AGGREGATE	- return quota information for the aggregate
 * FS_QUOTA_DEFAULT	- return default user/group/aggregate quota limits
 */

typedef struct fs_GetQuota {
	uint8		flag;	// Qualify entity for which query is made
	uint32		id;	// user or group id
} fs_GetQuota_t;

typedef struct fs_MsgGetQuotaReq {
	fs_MsgReqHdr	msgHdr;				// request header
	evm_volId_t		fsid;				// file system id for which the
										// request is made
	char8			path[FS_PATH_MAX];	// fileset name
	uint32			numberEntities;		// number of requests
	fs_GetQuota_t	entities[4];		// array of requests
} fs_MsgGetQuotaReq_t;

/*
 * Response for get quota request
 */

/*
 * Valid values for status field in response header are
 *
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid filesystem
 * FS_ERR_NOENT		- Invalid file set
 * FS_ERR_PERM		- Permission denied for the operation
 * FS_ERR_INVAL		- Invalid message
 */

typedef struct fs_MsgGetQuotaResp {
	fs_MsgRespHdr	msgHdr;			// response header
	uint32			numberResponse;	// number of response entries
	fs_QuotaBlk_t	responses[4];	// array of quota blks
} fs_MsgGetQuotaResp_t;

/*
 * Return quotas for all the users that have entries in the system
 */

/*
 * Valid values for flag are
 *
 * FS_QUOTA_USER	- return quota information for the user
 * FS_QUOTA_GROUP	- return quota information for the group
 * FS_QUOTA_AGGREGATE	- return quota information all for the filesets
 */

typedef struct fs_MsgReportQuotaReq {
	fs_MsgReqHdr	msgHdr;				// request header
	evm_volId_t		fsid;				// Request for this filesystem
	char8			path[FS_PATH_MAX];	// request for this fileset
	uint8			flag;				// qualify request
	uint64			startIndex;			// start index
	uint32			numberEntries;		// number of entries to return
} fs_MsgReportQuotaReq_t;

/*
 * Response for Report Quota request
 */

/*
 * Valid values for status field in response header
 *
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid File system
 * FS_ERR_NOENT		- Invalid File set
 * FS_ERR_PERM		- Permission denied
 * FS_ERR_INVAL		- Invalid message
 */

typedef struct fs_MsgReportQuotaResp {
	fs_MsgRespHdr	msgHdr;				// response header
	uint32			numberResponses;	// number of responses
	fs_QuotaBlk_t	responses[1];		// array of responses
} fs_MsgReportQuotaResp_t;

/*
 * Messages to turn quota on 
 */

typedef struct fs_MsgQuotaOnReq {
	fs_MsgReqHdr	msgHdr;					// rquest header
	evm_volId_t		fsid;					// file system on which to work
	char8			path[FS_PATH_MAX];		// fileset
} fs_MsgQuotaOnReq_t;

/*
 * Response to quota on
 */

/*
 * Valid return status
 *
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid File system
 * FS_ERR_NOENT		- Invalid File set
 * FS_ERR_PERM		- Permission denied
 * FS_ERR_INVAL		- Invalid message
 * FS_ERR_DEFFERED	- Operation will be completed asynchronously
 */

typedef struct fs_MsgQuotaOnResp {
	fs_MsgRespHdr	msgHdr;		// response header
    uint32          status;     // actual status of the quota on process:
                                // FS_QUOTA_TURN_ON, FS_QUOTA_ON
} fs_MsgQuotaOnResp_t;

/*
 * Messages to turn quota off 
 */

typedef struct fs_MsgQuotaOffReq {
	fs_MsgReqHdr	msgHdr;					// rquest header
	evm_volId_t		fsid;					// file system on which to work
	char8			path[FS_PATH_MAX];		// fileset
} fs_MsgQuotaOffReq_t;

/*
 * Response to quota off
 */

/*
 * Valid return status
 *
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid File system
 * FS_ERR_NOENT		- Invalid File set
 * FS_ERR_PERM		- Permission denied
 * FS_ERR_INVAL		- Invalid message
 */

typedef struct fs_MsgQuotaOffResp {
	fs_MsgRespHdr	msgHdr;		// response header
    uint32          status;     // actual status of the quota off process:
    // - 0 (quota is off)
    // - FS_QUOTA_TURN_ON (Quota being turned on, please try later)
} fs_MsgQuotaOffResp_t;

/*
 * Messages to create file set
 */

typedef struct fs_MsgCreateFileSetReq {
	fs_MsgReqHdr	msgHdr;					// message header
	evm_volId_t		fsid;					// file system to operate on
	char8			path[FS_PATH_MAX];		// fileset to create
} fs_MsgCreateFileSetReq_t;

/*
 * Response to create file set request
 *
 * Valid values for status are
 *
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid File system
 * FS_ERR_NOENT		- Invalid File set
 * FS_ERR_PERM		- Permission denied
 * FS_ERR_INVAL		- Invalid message
 * FS_ERR_EXIST		- fileset already exists
 * FS_ERR_NOSPC		- Could not create fileset for lack of space
 * FS_ERR_ROFS		- Read only filesystem. Not applicable for now
 * FS_ERR_ACCESS	- Insufficient access permissions
 * FS_ERR_DQUOT		- parent fileset is out of quota
 */

typedef struct fs_MsgCreateFileSetResp {
	fs_MsgRespHdr	msgHdr;		// response header
								// Nothing for now
} fs_MsgCreateFileSetResp_t;

/*
 * Messages to remove file set
 */

typedef struct fs_MsgRemoveFileSetReq {
	fs_MsgReqHdr	msgHdr;					// message header
	evm_volId_t		fsid;					// file system to operate on
	char8			path[FS_PATH_MAX];		// fileset to remove
} fs_MsgRemoveFileSetReq_t;

/*
 * Response to remove file set request
 *
 * Valid values for status are
 *
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid File system
 * FS_ERR_NOENT		- Invalid File set
 * FS_ERR_PERM		- Permission denied
 * FS_ERR_INVAL		- Invalid message
 * FS_ERR_ROFS		- Read only filesystem. Not applicable for now
 * FS_ERR_ACCESS	- Insufficient access permissions
 */

typedef struct fs_MsgRemoveFileSetResp {
	fs_MsgRespHdr	msgHdr;		// response header
								// Nothing for now
} fs_MsgRemoveFileSetResp_t;

/*
 * Messages to list file set in a file system.
 */

typedef struct fs_MsgListFileSetsReq {
	fs_MsgReqHdr	msgHdr;			// message header
	evm_volId_t		fsid;			// filesystem under consideration
	uint64			startIndex;		// record position in fileset info
	uint32			numberEntries;	// return so many entries
} fs_MsgListFileSetsReq_t;

/*
 * Response message for list file set
 *
 * Response will return inode numbers of root of the file sets. It is
 * up to the administrative software to figure out pathname for the inode
 * number returned.
 *
 * Status value in msgHdr can have following values
 * FS_ERR_NONE		- success
 * FS_ERR_NODEV		- Invalid filesystem
 * FS_ERR_PERM		- Permission denied
 */

typedef struct fs_FileSetList {
    uint32              id;         // fileset id
    uint32              parentId;   // parent id
    uint64              rootInode;  // inode number of the fileset
} fs_FileSetList_t;

typedef struct fs_MsgListFileSetsResp {
	fs_MsgRespHdr		msgHdr;			// Message header
	uint32				numberEntries;	// number of entries returned
    fs_FileSetList_t    filesets[1];
} fs_MsgListFileSetsResp_t;

/*
 * Given an inode and snapshot number, find the path to the inode
 * works only for directories
 */

typedef struct fs_MsgGetFilesetPathReq {
    fs_MsgReqHdr        msgHdr;
    evm_volId_t         volId;          // Volume filesystem lives on
    uint16              fsetId;
    uint64              inum;
    uint64              snapnum;
} fs_MsgGetFilesetPathReq_t;

typedef fs_MsgGetFilesetPathReq_t fs_MsgGetFilesetPathReq;

typedef struct fs_MsgGetFilesetPathResp {
    fs_MsgRespHdr       msgHdr;
    char                path[FS_PATH_MAX + 1];
} fs_MsgGetFilesetPathResp_t;
typedef fs_MsgGetFilesetPathResp_t fs_MsgGetFilesetPathResp;



/*
 * Filesystem statistics
 */

typedef struct fs_MsgStatVolReq
{
    fs_MsgReqHdr        hdr;
    evm_volId_t         volId;      // volid
} fs_MsgStatVolReq;

typedef fs_MsgStatVolReq    fs_MsgFsstatVolReq_t;

typedef struct fs_MsgStatVolResp
{
    fs_MsgRespHdr       hdr;            // Message header 
                                        // request.
	evm_volId_t			volId;			// return volid
    fs_Stat             stat;           // Returned filesystem stats 

}  fs_MsgStatVolResp;
typedef fs_MsgStatVolResp    fs_MsgFsstatVolResp_t;
	

/**
 * Abort filesystem request.
 */
typedef struct {
    fs_MsgReqHdr hdr;           // message header
    evm_volId_t  volId;         // volume to abort
} fs_MsgAbortReq;   


/**
 * Abort filesystem response.
 */
typedef struct {
    fs_MsgRespHdr hdr;          // message header
} fs_MsgAbortResp;

/**
 * Allocator statistics.
 */

typedef struct
{
    uint64      fileSize;   // in bytes
    uint64      nbExtent;
    uint64      totalExtentSize;
    uint64      minExtentSize;
    uint64      maxExtentSize;
    boolean     init;
} fs_fileStat_t;

typedef struct
{
    uint64      nbFragmentedFile;
    uint64      nbFile;
#ifdef ALLOC_FILE_SIZE_MEDIANE
    uint64      medianeFileSize;
#endif // ALLOC_FILE_SIZE_MEDIANE
    uint64      totalFileSize;
    uint64      nbExtent;
    uint64      totalExtentSize;
    uint64      minExtentPerFile;
    uint64      maxExtentPerFile;
    uint64      inumWithMaxExtent;
    uint64      minExtentSize;
    uint64      maxExtentSize;
    boolean     init;
} fs_fsysStat_t;

typedef struct {
    fs_MsgReqHdr        hdr;        // message header
    evm_volId_t         volId;      // filesystem id
    uint64              inum;       // file id, 0 if all filesystem
    int32               snapnum;    // snapshot
    boolean             print;      // print the extents on FP console
    uchar8              pad;
} fs_MsgAllocStatsReq;

typedef struct {
    fs_MsgRespHdr       hdr;        // message header
    evm_volId_t         volId;      // filesystem id
    uint64              inum;       // file id, 0 if all filesystem
    int32               snapnum;    // snapshot
    int32               pad;
    fs_fileStat_t       fileStat;   // file statistics if file id != 0
                                    // free space statistics if file id == 0
    fs_fsysStat_t       fsysStat;   // filesystem statistics 
} fs_MsgAllocStatsResp;

/*------------------------------------------------------------------------------
 * fs_MsgMclReq - data structure used for filesystem MULTI COMPONENT LOOKUP
 *                transaction request messages.
 *
 * This operation lookups or finds a file system object starting from
 * the directory specified by the parentHandle. MCL evaluates the pathname
 * contained in the objName where each component are separated using the
 * delimitor '/'.
 * All but the final name in the list must be the names of directories.
 */
typedef struct fs_MsgMclReq
{
    fs_MsgReqHdr    hdr;            // message header
    fs_Handle       parentHandle;   // parent handle
    uint32          objLen;         // length of object to lookup
    uint32          pad;
    char            objName[0];     // object shall be NULL terminated
                                    // (max len is FS_PATH_MAX)
} fs_MsgMclReq;                     // FS_REQ_MCL

/*------------------------------------------------------------------------------
 * fs_MsgMclResp - data structure used for filesystem MULTI COMPONENT LOOKUP
 *                 transaction response messages.
 *
 * Upon succesfull completion, the filehandle of the last component is returned.
 * Otherwise, the filehandle of the last good component is returned.
 */
typedef struct fs_MsgMclResp
{
    fs_MsgRespHdr   hdr;            // message header
    fs_Handle       parentHandle;   // returned parent file handle
    fs_Handle       handle;         // returned file handle
    int32           numAttrs;       // Number of attributes in attr[].
    int32           pad;
    fs_Attr         attr[0];        // Optional parent dir and/or looked up
                                    // object attributes.
} fs_MsgMclResp;                    // FS_REQ_MCL


/*-----------------------------------------------------------------------------
 * fs_MsgCacheInvReq - data structure used for invalidate cache
 *                     request messages.
 *                     Sent from FS to TXRX.
 *                     There is no response message from TXRX.
 *
 */

typedef struct fs_MsgCacheInvReq
{
    fs_MsgRespHdr       hdr;            // Message header

#define FS_CACHE_INV_TYPE_ALL       0   // Invalidate all caches
#define FS_CACHE_INV_TYPE_VOL       1   // Invalidate by volume
#define FS_CACHE_INV_TYPE_SNAP      2   // Invalidate by vol & snap #
#define FS_CACHE_INV_TYPE_INUM      3   // Invalidate by vol, snap & inum

    uint32              invType;        // One of the above.

    fs_Handle           handle;         // ignored     for invType=ALL
                                        // specifies vol, snap & inum
} fs_MsgCacheInvReq;


#if 0				/* XXX radek */
//   M I R R O R I N G

/**
 * Report the differences in extents between the two named snapshots
 * on the specified volume.  If oldname is NULL, the entire contents
 * of the snapshot named by newname will be reported.  The transfer
 * function is called as many times as necessary to report all
 * extents, passing in an arbitrary number of extents.  Only one call
 * to transfer will be made at a time.  After all the transfer calls
 * finish, the done callback will be made.
 *
 * The snap instaniate function returns immediately after being called.
 * It will make a series of calls to transfer.  If it fails internally
 * or if the transfer function returns a result code other than NFX_OK,
 * the abort function is called.  The void arg is provided for use by the
 * callback functions.
 */
extern int fs_snapInitiateDelta(
    evm_volId_t volume,
    char *oldname,
    char *newname,
    void *arg,
    int (*transfer)(void *arg, uint32 extentCount, fs_Extent *extents),
    int (*done)(void *arg),
    void (*abort)(void *arg));

/**
 * An opaque token representing the installation of a snapshot delta.
 */
typedef struct fs_SnapInstallation fs_SnapInstallation;

/**
 * Begin installing the differences between the snapshot named oldname
 * and the one named newname on the specified volume.  If oldname is
 * NULL, the entire contents of the snapshot named by newname are being
 * installed.  Returns a token to be used for the rest of the
 * installation process.
 *
 * Returns NULL for an invalid volume id, if the old snapshot name is not
 * NULL and does not exist, or if the new snapshot does exist.
 */
extern fs_SnapInstallation *fs_snapInstallBegin(
    evm_volId_t volume,
    char *oldname,
    char *newname);

/**
 * Install the reported extents as part of the specified snapshot
 * installation.
 *
 * Actual data blocks are assumed to have been written before they
 * are reported.  If any specified block is in use the file system,
 * the call fails and the snapshot installation is aborted, and
 * all associated resources are released.
 */
extern int fs_snapInstallTransfer(
    fs_SnapInstallation *arg,
    uint32 extentCount,
    fs_Extent *extents);

/**
 * Finish the specified snapshot installation.  Returns NFX_OK if the
 * installation was successful.  All transient resources are released.
 */
extern int fs_snapInstallCommit(
    fs_SnapInstallation *arg);

/**
 * Abort the specified snapshot installation.  All resources allocated
 * by the installation are released.
 */
extern void fs_snapInstallAbort(
    fs_SnapInstallation *arg);
#endif /* XXX radek */


/**
 * Some tidbits for debug commands.
 */
extern int
fs_getHashVolBlock(
    int argc,
    char **argv);

extern int
fs_getHashVolInum(
    int argc,
    char **argv);

extern int
fs_getHashVolInumPage(
    int argc,
    char **argv);


/*
 * Auto grow api
 */


typedef struct fs_MsgSetMinFreeReq
{
    fs_MsgReqHdr        hdr;
    evm_volId_t         volId;
    uint8               minFree;

} fs_MsgSetMinFreeReq;

typedef struct fs_MsgSetMinFreeResp
{
    fs_MsgRespHdr        hdr;

} fs_MsgSetMinFreeResp;
typedef struct fs_MsgGetMinFreeReq
{
    fs_MsgReqHdr        hdr;
    evm_volId_t         volId;

} fs_MsgGetMinFreeReq;

typedef struct fs_MsgGetMinFreeResp
{
    fs_MsgRespHdr        hdr;
    uint8               minFree;

} fs_MsgGetMinFreeResp;

#endif /* FS_API_H */

