// -*- tab-width: 8; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-arena.c
 *
 * RCS:  $Id: fs-arena.c,v 1.5 2001/09/04 21:32:32 haahr Exp $
 *
 * Description:  Temporary allocation arenas for context operations.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */

#if !defined(SSC)

#include "fs-includes.h"

enum {
    /** Maximum size of an object to allocate in the arenas. */
    ARENA_MAX_SIZE = FS_MAX_BLOCK_SIZE,
    /** Minimum size of an arena. */
    ARENA_MIN_SIZE = 1024,
    /** Alignment, in bytes, for space returned from the arena. */
    ARENA_ALIGNMENT = sizeof (double)
};


/**
 * An allocation arena for a given context.  Arenas are chained, with all
 * allocations coming from the one closest to the front.
 */
struct fs_Arena
{
    /** Total size of this arena. */
    uint32 size;

    /** Bytes allocated so far in this arena. */
    uint32 used;
    
    /** The previous arena. */
    fs_Arena *prev;

    /** Space used for actual allocation. */
    union {
        double align;
        char space[1];
    } u;
};


/**
 * Allocate memory which will be freed when this context goes away.
 */
extern void *fs_contextMemAlloc(
    fs_Context *context,
    uint32     bytes)
{
    uint32 aligned = alignto(bytes, ARENA_ALIGNMENT);
    fs_Arena *arena = context->arena;
    void *result;
    
    FS_ASSERT(bytes <= ARENA_MAX_SIZE);

    if (arena == NULL || arena->size < arena->used + aligned) {
        uint32 size = alignto(aligned, ARENA_MIN_SIZE);
        arena = fs_ramAlloc(FS_RAM_ARENA, offsetof(fs_Arena, u.space[size]));
        arena->size = size;
        arena->used = 0;
        arena->prev = context->arena;
        context->arena = arena;
    }

    result = &arena->u.space[arena->used];
    arena->used += aligned;
    FS_ASSERT(arena->used <= arena->size);
    return result;
}


/**
 * Allocate zero-initialized memory will be freed when this context goes away.
 */
extern void *fs_contextMemAllocZeroed(
    fs_Context *context,
    uint32     bytes)
{
    void *p = fs_contextMemAlloc(context, bytes);
    bzero(p, bytes);
    return p;
}


/**
 * Free the memory associated with the given context.
 */
extern void fs_contextMemFree(
    fs_Context *context)
{
    fs_Arena *arena = context->arena;
    while (arena != NULL) {
        fs_Arena *prev = arena->prev;
        fs_ramDealloc(FS_RAM_ARENA, arena);
        arena = prev;
    }
    context->arena = NULL;
}
#endif // SSC
