// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-block-set.c
 *
 * RCS:  $Id: fs-block-set.c,v 1.7 2002/06/06 21:47:44 bob Exp $
 *
 * Description:  Sets of blocks.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */

#if !defined(SSC)

#include "fs-includes.h"



//   F O R W A R D   D E C L A R A T I O N S


typedef struct fs_BlockSetLink fs_BlockSetLink;

#define BLOCK_SET_MAX_LEVEL     20

#define BLOCK_SET_BLOCKS_SHIFT 6
#define BLOCK_SET_BLOCKS_BITS  (1 << BLOCK_SET_BLOCKS_SHIFT)
#define BLOCK_SET_BLOCKS_MASK  (BLOCK_SET_BLOCKS_BITS - 1)


/**
 * A link cell in the block set.  Each link covers a 64-bit range.  The
 * array is variable-sized, based on the number of levels.
 *
 * The structure is declared at its maximum size, though shorter links
 * arrays may be found.
 */
struct fs_BlockSetLink {
    /** The group of blocks, indicated by blockNumber/64. */
    uint64 key;
    /** The bitset of blocks within the group. */
    uint64 blocks;
    /** The level of this link cell. */
    uint32 level;
    /** The links from this cell forward. (Variable-sized array.) */
    fs_BlockSetLink *forward[BLOCK_SET_MAX_LEVEL];
};


/**
 * A block set is a mapping from 64-bit block numbers to booleans.  It is
 * implemented as a skip-list mapping (block-number/64) to a 64-bit value
 * used to indicate which blocks within that 64-block range are in use.
 *
 * TODO: generalize to 64-bit to arbitrary 64-bit mappings
 */
struct fs_BlockSet {
    /** Context to allocate memory in. */
    fs_Context *context;
    /** Maximum level in the list. -- TODO: merge with header.level */
    uint32 level;
    /** The first set of links -- vestigial. */
    fs_BlockSetLink header;
};


/**
 * Create a block set, using the specified context for allocation.
 */
fs_BlockSet *fs_makeBlockSet(
    fs_Context *context)
{
    int i;
    fs_BlockSet *set = fs_contextMemAlloc(context, sizeof (fs_BlockSet));
    set->context = context;
    set->level   = 0;

    set->header.key    = ~0;
    set->header.blocks = 0;
    set->header.level  = BLOCK_SET_MAX_LEVEL;
    for (i = 0; i < BLOCK_SET_MAX_LEVEL; i++)
        set->header.forward[i] = NULL;

    return set;
}


/**
 * Return true if the given block is in the block set, false otherwise.
 */
boolean fs_blockSetContains(
    fs_BlockSet *set,
    fs_Block block)
{
    uint64 key = fs_blockNumber(block) >> BLOCK_SET_BLOCKS_SHIFT;
    uint64 bit = 1ULL << (fs_blockNumber(block) & BLOCK_SET_BLOCKS_MASK);

    int i;
    fs_BlockSetLink *cell;
    fs_BlockSetLink *prev = &set->header;
    
    assert(fs_blockIsValid(block));
    
    for (i = set->level - 1; i >= 0; i--) {
        fs_BlockSetLink *forward;
        while ((forward = prev->forward[i]) != NULL && forward->key < key) {
            prev = forward;
        }
    }

    cell = prev->forward[0];
    return (cell != NULL) && (cell->key == key) && ((cell->blocks & bit) != 0);
}


/**
 * Return true if any block in the specified extent is in the block set,
 * false otherwise.
 */
boolean fs_blockSetContainsAny(
    fs_BlockSet *set,
    fs_Extent extent)
{
    int i, n;

    for (i = 0, n = fs_extentLength(extent); i < n; i++) {
        if (fs_blockSetContains(set, fs_extentBlock(extent, i)))
            return TRUE;
    }

    return FALSE;
}


/**
 * Return true if all block in the specified extent are in the block set,
 * false otherwise.
 */
boolean fs_blockSetContainsAll(
    fs_BlockSet *set,
    fs_Extent extent)
{
    int i, n;

    for (i = 0, n = fs_extentLength(extent); i < n; i++) {
        if (!fs_blockSetContains(set, fs_extentBlock(extent, i)))
            return FALSE;
    }

    return TRUE;
}


/**
 * Return a random level for the specified block set, updating the list's
 * level if necessary.
 */
static uint32 fs_blockSetRandomLevel(
    fs_BlockSet *set)
{
    int level = 1;
    while (fs_cheapRandomBoolean() && level < BLOCK_SET_MAX_LEVEL) {
        level += 1;
    }
    if (level <= set->level) {
        return level;
    }
    else {
        set->level += 1;
        return set->level;
    }
}


/**
 * Create a new link cell for the specified set with the specified key.
 * The level will be randomly chosen, the blocks array will be initially
 * zero, and none of the links will be set.
 */
static fs_BlockSetLink *fs_blockSetMakeLink(
    fs_BlockSet *set,
    uint64 key)
{
    int level = fs_blockSetRandomLevel(set);
    fs_BlockSetLink *link
        = fs_contextMemAlloc(set->context,
                             offsetof(fs_BlockSetLink, forward[level]));
    link->level = level;
    link->key = key;
    return link;
}



/**
 * Add a block (if it's not already present) to the block set.  Return
 * true if the block was already in the set.
 */
boolean fs_blockSetAdd(
    fs_BlockSet *set,
    fs_Block block)
{
    int i;

    uint64 key = fs_blockNumber(block) >> BLOCK_SET_BLOCKS_SHIFT;
    uint64 bit = 1ULL << (fs_blockNumber(block) & BLOCK_SET_BLOCKS_MASK);

    fs_BlockSetLink **update[BLOCK_SET_MAX_LEVEL];
    fs_BlockSetLink *cell;
    fs_BlockSetLink *prev = &set->header;

    assert(fs_blockIsValid(block));
    
    for (i = set->level - 1; i >= 0; i--) {
        fs_BlockSetLink *forward;
        while ((forward = prev->forward[i]) != NULL && forward->key < key) {
            prev = forward;
        }
        update[i] = &prev->forward[i];
    }

    cell = prev->forward[0];
    if (cell != NULL && cell->key == key) {
        if ((cell->blocks & bit) != 0)
            return TRUE;
        else {
            cell->blocks |= bit;
            return FALSE;
        }
    }
    else {
        int oldLevel = set->level;
        fs_BlockSetLink *add = fs_blockSetMakeLink(set, key);
        for (i = oldLevel; i < add->level; i++) {
            update[i] = &set->header.forward[i];
        }
        for (i = 0; i < add->level; i++) {
            add->forward[i] = *update[i];
            *update[i] = add;
        }
        add->blocks = bit;
        return FALSE;
    }
}


/**
 * Add all the blocks from the specified extent (that are not already
 * present) to the specified block set.  Returns true if any of blocks
 * were already in the set.
 */
boolean fs_blockSetAddAll(
    fs_BlockSet *set,
    fs_Extent extent)
{
    int i, n;
    boolean already = FALSE;

    for (i = 0, n = fs_extentLength(extent); i < n; i++) {
        if (fs_blockSetAdd(set, fs_extentBlock(extent, i)))
            already = TRUE;
    }

    return already;
}


/**
 * Delete a block from the block set.  Return true if the block was
 * present in the set, false otherwise.
 */
boolean fs_blockSetRemove(
    fs_BlockSet *set,
    fs_Block block)
{
    uint64 key = fs_blockNumber(block) >> BLOCK_SET_BLOCKS_SHIFT;
    uint64 bit = 1ULL << (fs_blockNumber(block) & BLOCK_SET_BLOCKS_MASK);

    int i;
    fs_BlockSetLink *cell;
    fs_BlockSetLink *prev = &set->header;
    
    assert(fs_blockIsValid(block));
    
    for (i = set->level - 1; i >= 0; i--) {
        fs_BlockSetLink *forward;
        while ((forward = prev->forward[i]) != NULL && forward->key < key) {
            prev = forward;
        }
    }

    cell = prev->forward[0];
    if (cell == NULL || cell->key != key || (cell->blocks & bit) == 0)
        return FALSE;

    cell->blocks &= ~bit;
    // TODO: if (cell->blocks == 0), remove it
    return TRUE;
}


/**
 * Reniove all the blocks from the specified extent (that are present) to
 * the specified block set.  Returns false if any of blocks were not
 * in the set, true otherwise.
 */
boolean fs_blockSetRemoveAll(
    fs_BlockSet *set,
    fs_Extent extent)
{
    int i, n;
    boolean already = FALSE;

    for (i = 0, n = fs_extentLength(extent); i < n; i++) {
        if (fs_blockSetRemove(set, fs_extentBlock(extent, i)))
            already = TRUE;
    }

    return already;
}


/**
 * Return the next (lowest) bit, at or above the specified bit position,
 * in the specified word, or -1 if no bit is set.
 */
static uint64 fs_nextBit(
    uint64 word,
    int start)
{
    int i, n;
    
    word >>= start;

    if (word == 0)
        return -1;

    // TODO: unroll
    for (i = 32, n = 0; i > 0; i /= 2) {
        if ((word & ((1ULL << i) - 1)) == 0) {
            word >>= i;
            n += i;
        }
    }

    return start + n;
}


/**
 * Return the first block in a block set.  Returns FS_INVALID_BLOCK if
 * there are no blocks in the set.
 */
fs_Block fs_blockSetFirst(
    fs_BlockSet *set)
{
    fs_BlockSetLink *link;
    if (set->level == 0)
        return FS_INVALID_BLOCK;
    for (link = set->header.forward[0];
         link != NULL;
         link = link->forward[0])
    {
        int first = fs_nextBit(link->blocks, 0);
        if (first >= 0)
            return fs_makeBlock((link->key * BLOCK_SET_BLOCKS_BITS) + first);
    }
    return FS_INVALID_BLOCK;
}


/**
 * Return the next block in a block set following the specified block.
 * Returns FS_INVALID_BLOCK if the specified block is the last one in
 * the set.
 */
fs_Block fs_blockSetNext(
    fs_BlockSet *set,
    fs_Block block)
{
    uint64 blockno = fs_blockNumber(block) + 1;
    uint64 key = blockno >> BLOCK_SET_BLOCKS_SHIFT;
    uint64 bit = blockno & BLOCK_SET_BLOCKS_MASK;

    int i;
    fs_BlockSetLink *cell;
    fs_BlockSetLink *link;
    fs_BlockSetLink *prev = &set->header;

    assert(fs_blockIsValid(block));
    
    for (i = set->level - 1; i >= 0; i--) {
        fs_BlockSetLink *forward;
        while ((forward = prev->forward[i]) != NULL && forward->key < key) {
            prev = forward;
        }
    }

    cell = prev->forward[0];
    if (cell == NULL)
        return FS_INVALID_BLOCK;

    if (cell->key == key) {
        int next = fs_nextBit(cell->blocks, bit);
        if (next >= 0) {
            return fs_makeBlock((key * BLOCK_SET_BLOCKS_BITS) + next);
        }
        cell = cell->forward[0];
        if (cell == NULL)
            return FS_INVALID_BLOCK;
    }
    
    for (link = cell;
         link != NULL;
         link = link->forward[0])
    {
        int first = fs_nextBit(link->blocks, 0);
        if (first >= 0)
            return fs_makeBlock((link->key * BLOCK_SET_BLOCKS_BITS) + first);
    }
    return FS_INVALID_BLOCK;
}


/**
 * Return the first block not in the specified block set.  Returns
 * FS_INVALID_BLOCK if the set contains all blocks.
 */
fs_Block fs_blockSetFirstMissing(
    fs_BlockSet *set)
{
    fs_Block zero = fs_makeBlock(0);
    if (fs_blockSetContains(set, zero))
        return fs_blockSetNextMissing(set, zero);
    else
        return zero;
}


/**
 * Return the first block after the specified block not in the specified
 * block set.  Returns FS_INVALID_BLOCK if there are no free blocks after
 * the specified block.
 */
fs_Block fs_blockSetNextMissing(
    fs_BlockSet *set,
    fs_Block startBlock)
{
    // TODO: implement this properly, skipping over clumps of blocks.
    // in the meanwhile, this is fast enough

    fs_Block block;

    for (block = fs_makeBlock(fs_blockNumber(startBlock) + 1);
         fs_blockIsValid(block);
         block = fs_makeBlock(fs_blockNumber(block) + 1))
    {
        if (!fs_blockSetContains(set, block))
            return block;
    }

    return FS_INVALID_BLOCK;
}


/**
 * Visit all extents in the specified block set.
 */
void fs_blockSetVisitExtents(
    fs_BlockSet *set,
    void *arg,
    void (*visit)(void *arg, fs_Extent extent))
{
    fs_Block block, last;

    for (block = fs_blockSetFirst(set);
         fs_blockIsValid(block);
         block = fs_blockSetNext(set, last))
    {
        uint64 delta;
        last = fs_blockSetNextMissing(set, block);

        delta = fs_blockNumber(last) - fs_blockNumber(block);

        while (delta > EXTENT_LENGTH_MAX) {
            visit(arg, fs_makeExtent(EXTENT_LENGTH_MAX, block));
            block = fs_makeBlock(fs_blockNumber(block) + EXTENT_LENGTH_MAX);
            delta -= EXTENT_LENGTH_MAX;
        }

        assert(delta > 0);
        visit(arg, fs_makeExtent(delta, block));
    }
}


/**
 * Print a block set link, for debugging purposes.
 */
static void fs_printBlockSetLink(fs_BlockSet *set, fs_BlockSetLink *link)
{
    int i;
    int level = MIN(set->level, link->level);
    printf("# %p [%p", link, link->forward[0]);
    for (i = 1; i < level; i++)
        printf(", %p", link->forward[i]);
    printf("] ");
    if (link == &set->header) {
        printf("<header>\n");
    }
    else {
        printf("%lld ... %lld",
               link->key * BLOCK_SET_BLOCKS_BITS,
               (link->key + 1) * BLOCK_SET_BLOCKS_BITS - 1);
        for (i = 0; i < BLOCK_SET_BLOCKS_BITS; i++)
            if ((link->blocks & (1ULL << i)) != 0)
                printf(" +%d", i);
        printf("\n");
    }
}


/**
 * Print a block set, for debugging purposes.
 */
void fs_printBlockSet(
    fs_BlockSet *set)
{
    fs_BlockSetLink *link;

    printf("\n");
    printf("# fs_BlockSet %p level %d\n", set, set->level);

    for (link = &set->header; link != NULL; link = link->forward[0])
        fs_printBlockSetLink(set, link);
    
    printf("\n");
}


/**
 * Test the block set code.  Not comprehensive, but it should be
 * a little useful.
 */
void fs_testBlockSets(
    fs_Context *context)
{
    int i;
    fs_Block block;
    
    fs_BlockSet *set = fs_makeBlockSet(context);

    assert(!fs_blockSetContains(set, fs_makeBlock(0)));
    assert(!fs_blockSetContains(set, fs_makeBlock(8848973)));
    assert(!fs_blockSetAdd(set, fs_makeBlock(8848973)));
    assert(fs_blockSetContains(set, fs_makeBlock(8848973)));
    assert(fs_blockSetAdd(set, fs_makeBlock(8848973)));
    assert(!fs_blockSetContains(set, fs_makeBlock(8848972)));
    assert(!fs_blockSetAdd(set, fs_makeBlock(8848972)));
    assert(fs_blockSetContains(set, fs_makeBlock(8848972)));
    assert(fs_blockSetContains(set, fs_makeBlock(8848973)));
    assert(fs_blockNumber(fs_blockSetNext(set, fs_makeBlock(2))) == 8848972);
    assert(fs_blockNumber(
        fs_blockSetNextMissing(set, fs_makeBlock(8848971))) == 8848974);

    for (i = 0; i < 1000; i++) {
        assert(!fs_blockSetContains(set, fs_makeBlock(i)));
        assert(!fs_blockSetAdd(set, fs_makeBlock(i)));
        assert(fs_blockSetContains(set, fs_makeBlock(i)));
    }

    for (i = 10000; i < 10100; i += 3) {
        assert(!fs_blockSetContains(set, fs_makeBlock(i)));
        assert(!fs_blockSetAdd(set, fs_makeBlock(i)));
        assert(fs_blockSetContains(set, fs_makeBlock(i)));
    }

    for (i = 0; i < 1000; i++) {
        assert(fs_blockSetContains(set, fs_makeBlock(i)));
    }

    for (i = 10000; i < 10100; i += 3) {
        assert(fs_blockSetContains(set, fs_makeBlock(i)));
    }

    assert(fs_blockSetContains(set, fs_makeBlock(8848972)));
    assert(fs_blockSetContains(set, fs_makeBlock(8848973)));

    fs_printBlockSet(set);

    printf("# present:\n");
    for (i = 0, block = fs_blockSetFirst(set);
         fs_blockIsValid(block);
         i++, block = fs_blockSetNext(set, block))
    {
        printf("# %d -> %lld\n", i, fs_blockNumber(block));
    }

    printf("# missing:\n");
    for (i = 0, block = fs_blockSetFirstMissing(set);
         i < 20 && fs_blockIsValid(block);
         i++, block = fs_blockSetNextMissing(set, block))
    {
        printf("# %d -> %lld\n", i, fs_blockNumber(block));
    }
}
#endif // SSC
