// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-block-set.h
 *
 * RCS:  $Id: fs-block-set.h,v 1.4 2002/01/07 22:47:35 haahr Exp $
 *
 * Description:  Sets of blocks.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */


#ifndef FS_BLOCK_SET_H
#define FS_BLOCK_SET_H

/**
 * A block set is a set containing fs_Block values.  It can also be viewed
 * as a mapping from 64-bit block numbers to booleans.
 */
typedef struct fs_BlockSet fs_BlockSet;


/**
 * Create a block set, using the specified context for allocation.
 */
extern fs_BlockSet *fs_makeBlockSet(
    fs_Context *context);


/**
 * Return true if the given block is in the block set, false otherwise.
 */
extern boolean fs_blockSetContains(
    fs_BlockSet *set,
    fs_Block block);


/**
 * Return true if any block in the specified extent is in the block set,
 * false otherwise.
 */
extern boolean fs_blockSetContainsAny(
    fs_BlockSet *set,
    fs_Extent extent);


/**
 * Return true if all block in the specified extent are in the block set,
 * false otherwise.
 */
extern boolean fs_blockSetContainsAll(
    fs_BlockSet *set,
    fs_Extent extent);


/**
 * Add a block (if it's not already present) to the block set.  Return
 * true if the block was already in the set.
 */
extern boolean fs_blockSetAdd(
    fs_BlockSet *set,
    fs_Block block);


/**
 * Add all the blocks from the specified extent (that are not already
 * present) to the specified block set.  Returns true if any of blocks
 * were already in the set.
 */
extern boolean fs_blockSetAddAll(
    fs_BlockSet *set,
    fs_Extent extent);


/**
 * Delete a block from the block set.  Return true if the block was
 * present in the set, false otherwise.
 */
extern boolean fs_blockSetRemove(
    fs_BlockSet *set,
    fs_Block block);


/**
 * Reniove all the blocks from the specified extent (that are present) to
 * the specified block set.  Returns false if any of blocks were not
 * in the set, true otherwise.
 */
extern boolean fs_blockSetRemoveAll(
    fs_BlockSet *set,
    fs_Extent extent);


/**
 * Return the first block in a block set.  Returns FS_INVALID_BLOCK if
 * there are no blocks in the set.
 */
extern fs_Block fs_blockSetFirst(
    fs_BlockSet *set);


/**
 * Return the next block in a block set following the specified block.
 * Returns FS_INVALID_BLOCK if the specified block is the last one in
 * the set.
 */
extern fs_Block fs_blockSetNext(
    fs_BlockSet *set,
    fs_Block block);


/**
 * Return the first block not in the specified block set.  Returns
 * FS_INVALID_BLOCK if the set contains all blocks.
 */
extern fs_Block fs_blockSetFirstMissing(
    fs_BlockSet *set);


/**
 * Return the first block after the specified block not in the specified
 * block set.  Returns FS_INVALID_BLOCK if there are no free blocks after
 * the specified block.
 */
extern fs_Block fs_blockSetNextMissing(
    fs_BlockSet *set,
    fs_Block block);


/**
 * Print a block set, for debugging purposes.
 */
extern void fs_printBlockSet(
    fs_BlockSet *set);


/**
 * Test the block set code.  Not comprehensive, but it should be
 * a little useful.
 */
extern void fs_testBlockSets(
    fs_Context *context);


/**
 * Visit all extents in the specified block set.
 */
extern void fs_blockSetVisitExtents(
    fs_BlockSet *set,
    void *arg,
    void (*visit)(void *arg, fs_Extent extent));


#endif /* FS_BLOCK_SET_H */
