// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-buddy.h
 *
 * RCS:  $Id: fs-buddy.h,v 1.8 2002/09/11 18:27:21 isabellec Exp $
 *
 * Description:  Interface to binary buddy system.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2002-01-17
 */

#if !defined(FS_BUDDY_H)
#define FS_BUDDY_H


/**
 * Constants defining binary buddy regions.
 */
enum {
    /** Log2(maximum size of an extent in a region). */
    FS_MAX_BUDDY_LEVELS = 15,
    /** Number of blocks in a region page. */
    FS_BLOCKS_PER_BUDDY_REGION = 1 << FS_MAX_BUDDY_LEVELS,
};


/**
 * Find an extent in the binary buddy tree near the specified block,
 * and remove the extent from the tree.  The system will try to find
 * an extent at least as long as the requested amount, but it may return
 * something shorter.
 *
 * Returns a zero length extent if no such extent was found.
 */
extern fs_Extent fs_buddyFindExtent(
    fs_Context *context,
    uint16 len,
    fs_Block target,
    boolean isFirstFileExtent);

/**
 * Add an extent to the binary buddy tree.  The binary buddy region buffers
 * are released after the operation if the release flag is true;  otherwise,
 * they will be held until the transaction commits.
 */
extern void fs_buddyAddExtent(
    fs_Context *context,
    fs_Extent extent,
    boolean release);


/**
 * Add the blocks between the first specified block (inclusive) and the
 * final specified block (exclusive) to the binary buddy tree.
 */
void fs_buddyAddBlocks(
    fs_Context *context,
    fs_Block first,
    fs_Block last,
    boolean release);


/**
 * Add an extent to the binary buddy tree, if it's not already in there.
 *
 * This is similar to fs_buddyAddExtent, except that it's robust in the
 * presence of redundant adds;  that robustness costs performance, so
 * this routine takes time proportional to the size of the extent.
 */
void fs_buddyMaybeAddExtent(
   fs_Context *context,
   fs_Extent extent,
   boolean release);


/**
 * Mark the blocks on a region page of the binary buddy tree for eek.
 */
extern boolean fs_buddyEekMarkRegion(
    fs_Context *context,
    fs_Buf *buf,
    uint8 *bitmap,
    uint64 baseBlock);

/* Check if a block is not yet allocated.
 */
boolean
fs_isBlockFree(fs_Context * context,
               fs_Block     block);

/**
 * Initialize the buddy region bitmap with nRegions
 */
extern void fs_buddyInitBitmap(
    fs_Context *context,
    fs_FileSystem *fs,
    uint32 nRegions,
    boolean setValid);


/**
 * Thread to recalculate buddy regions
 * Fired off after log replay to start rebuilding buddy
 */
void fs_buddyRecalThread(
    esm_handle_t handle,
    void *arg);

#ifdef BUDDY_INCORE
/*
 */
fs_Buf *
fs_buddyGetRegionBuf(fs_Context *   context,
                     uint32         region,
                     boolean        exclusive);
#endif // BUDDY_INCORE

#endif /* FS_BUDDY_H */
