/*
 * Name:  fs-context.c
 *
 * RCS:  $Id: fs-context.c,v 1.147 2002/12/05 21:08:57 bob Exp $
 *
 * Description:  Create/signal/destroy filesystem threads/contexts.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"
#if defined(FS_STANDALONE)
#include "simthread.h"
#endif

fs_NfsStats         fs_nfsStats;
uint32              fs_currentActiveReqs = 0;
uint32              fs_maxActiveReqs = 0; 
fs_YieldStats       fs_yieldStats;
fs_RecvReqStats     fs_recvReqStats;
fs_ReplyStats       fs_replyStats;
nfx_spinlock_t      fs_statsReqRcvdSpinId;

uint64               fs_requestsIgnored = 0;

fs_ErrStats         fs_errStats;
nfx_spinlock_t      fs_errStatsSpinId;


#ifdef FS_DEBUG_NFS_DELTA_BUCKET
// TODO - this is duplicated in fs-shell.c - FIX THIS!
//  Previous msec values:
//  { 1,5,10,15,20,25,30,40,50,100,500,1000,5000,10000};
//  Now in usecs:
static int32 fs_nfsStatsUsecs[FS_NFS_TIME_MAX] =
    { 10,50,100,200,500,1000,1500,2000,3000,5000,10000,20000,50000,100000};
eeeRTC_t fs_nfsStatsRTC[FS_NFS_TIME_MAX];
#endif

#ifdef FS_DEBUG_PRINT_DELTA_THRESHOLD
// debug aid to print is a thread takes more than N msec's
static int64 fs_elapsedUsecThreshold = 2000000; 
#endif   // FS_DEBUG_PRINT_DELTA_THRESHOLD

int32 fs_nbrOfContexts = 0 ;
int32 fs_maxNbrOfContexts = 0 ;


/**
 * Handle a received message for the filesystem.  It
 * expects a flat contiguous buffer holding one of the
 * messages defined in our API.
 */
void
fs_receiveMsg(
    void            *msgBuf,
    uint32			src,
    int32           msgLen,
    int32           contextFlags)
{
    fs_MsgReqHdr            *reqHdr;
    int32                   rc, msgType;
    const fs_ReqTypeTbl	   	*reqType;

    reqHdr = msgBuf;
    msgType = reqHdr->requestType;

#if defined(FS_STANDALONE)
    switch(reqHdr->src)
    {
        case FS_SIM_MSG_ADDR_SSC:
            break;
        case FS_SIM_MSG_ADDR_MCPU:
#ifndef SBM_REV_2
            fs_cod_write_confirm(((mcpu_confirm_cod_t *)msgBuf)->cod);
#endif
            goto freeBuf;
        default:
            if ((reqHdr->src >= FS_SIM_MSG_ADDR_FSSF) &&
                (reqHdr->src < FS_SIM_MSG_ADDR_NFS + FS_SIM_NFS_THREAD_MAX))
            {
                break;
            }
            fs_warn(NULL, FS_UNK_MSG_ERR, "fs_receiveMsg");
            goto freeBuf;
    }
#endif
    
    //printf("rx msg type %d\n", msgType);

    if (msgType >= FS_REQ_NULL &&
        msgType < FS_REQ_MAX)
    {
        // stats, independent of filesystem (this counts also mkfs, etc...)
        NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
        fs_recvReqStats.num[msgType] += 1;
        NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);

		reqType = &fs_reqTypeTbl[msgType];

        // Special case lookup req fast path.
        if (msgType == FS_REQ_LOOKUP)
        {
            rc = fs_lookupInstantiate(reqType, reqHdr, contextFlags, src);
        }
        else
        {
            rc = (reqType->instantiateFunc)(reqType, reqHdr, contextFlags, src);
        }

        if (rc != NFX_OK)
        {
            // Instantiate func responsible for error response.
        }
    }
#ifndef SBM_REV_2
    else if (msgType == SBM_MSG_COD_RCV)
    {
        fs_sacStoreInstantiate(msgBuf, bufType, src);
    }
#endif
    else
    {
#if 0
        printf("msgType: %d\n", msgType);
        fs_warn(NULL, FS_UNK_MSG_ERR, "fs_receiveMsg");
#endif		
        goto freeBuf;
    }
        
    return;

freeBuf:

    if (contextFlags & FS_CONTEXT_SHARED_REQUEST)
    {
        eee_deallocateBuffer(msgBuf);
    }
    else
    {
#if defined(FS_STANDALONE)
        fs_ramDealloc(FS_RAM_MCPU, msgBuf);
#else
        fs_ramDealloc(FS_RAM_REQ_MSG, msgBuf);
#endif
    }

} /* fs_receiveMsg */


/**
 * Create a new thread for the same file system
 * as the specified context, and run the specified
 * function.
 */
esm_handle_t
fs_threadFork(
    fs_Context  *context,
    void        (*func)(esm_handle_t, void *),
    boolean     addToActiveList,
    boolean     allowStarvation,
    int32       daemonType,
    void        *extraArg)
{
    int32   rc = NFX_OK;
    fs_Context  *newContext = fs_ramAlloc(FS_RAM_CONTEXT, sizeof (fs_Context));

    FS_ASSERT(daemonType == FS_DAEMON_TYPE_DUMP 
           || daemonType == FS_DAEMON_TYPE_QUOTA_ON
	   || daemonType == FS_DAEMON_TYPE_EEK
           || extraArg == NULL);

#ifdef FS_USE_MALLOC_SLAB
	fs_threadContextReinit(newContext);
#else
    bzero(newContext, sizeof(fs_Context));
#endif
   
    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs_nbrOfContexts++;
    if (fs_nbrOfContexts > fs_maxNbrOfContexts)
    {
        fs_maxNbrOfContexts = fs_nbrOfContexts;
    }
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
    
    if (context != NULL)
    {
        newContext->volId = context->volId;
        newContext->fileSystem = context->fileSystem;
    }

    if (allowStarvation)    {
        newContext->flags |= FS_CONTEXT_ALLOW_LOCK_STARVATION;
    }

    //Lock lists init
    fs_lockContextInit(newContext, SHARED_LOCK_LIST_INIT);

    // For correct exception handling, each thread must
    // be associated with a filesystem.
    FS_ASSERT(newContext->fileSystem != NULL);

    if (newContext->fileSystem != NULL &&
        addToActiveList)
    {
        NFX_SPIN_LOCK(&newContext->fileSystem->spin);
        DLL_INSERT_TAIL(&newContext->fileSystem->activeThreadHdr,
                        newContext, fsEntry);
        NFX_SPIN_UNLOCK(&newContext->fileSystem->spin);
    }

    newContext->daemonType = daemonType;
    newContext->isThread = TRUE;

    if (extraArg) {
            newContext->extraArg = extraArg;
    }
    newContext->smHandle = thread_create(newContext);
    if (newContext->smHandle)
        rc = thread_start(&(newContext->smHandle), func);
    else
        rc = NFX_ERR;
    if (rc != NFX_OK)
    {
        fs_error(newContext, FS_ESM_CREATE_ERR, "fs_threadCreate");
    }

    //printf("threadFork new sm handl 0x%x for func %p\n",
    //       newContext->smHandle, func);

    /*
     * Thread comes to life here
     */
	eee_rtcStart(&newContext->startTime);
	eee_rtc(&newContext->tmpTime);

    return newContext->smHandle;
}


/**
 * Terminate a thread created with fs_threadFork.
 */
EXPORTED void
fs_threadTerminate(
    fs_Context  *context)
{
    if (context->fileSystem != NULL &&
		!DLL_IS_ISOLATED(context, fsEntry))
    {
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        DLL_REMOVE(&context->fileSystem->activeThreadHdr, context, fsEntry);
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    }
	else {
		DLL_ISOLATE_ELEM(context, fsEntry);
	}

	if (context->logSummary) {
		fs_logDestroySummary(context, context->logSummary);
		context->logSummary = NULL;
	}

	if (context->arena) {
		fs_contextMemFree(context);
	}

	context->daemonType = FS_DAEMON_TYPE_INVALID;

    fs_lockContextFree(context);

	ASSERT(context->arena == NULL);
	ASSERT(context->logSummary == NULL);
	ASSERT(DLL_IS_ISOLATED(context, fsEntry));
	ASSERT(context->eek.inUse == FALSE);
	ASSERT(context->pendingFreeExtents == NULL);

    fs_ramDealloc(FS_RAM_CONTEXT, context);

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs_nbrOfContexts--;
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
}


/**
 * Wake up a daemon (the given "handle") with event "event",
 * and set "*flag" to 1. Returns TRUE if the daemon was signalled,
 * FALSE otherwise.
 */
boolean
fs_daemonWakeup(
    fs_Context *context,
    nfx_spinlock_t *spin,
    esm_handle_t handle,
    int32 *flag,
    esm_event_t event)
{
    int32 rc;
    
    NFX_SPIN_LOCK(spin);
    if (*flag > 0)
    {
        NFX_SPIN_UNLOCK(spin);
        return FALSE;
    }

    *flag = 1;
    NFX_SPIN_UNLOCK(spin);

    rc = thread_event(handle,  &event);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_EVENT_DELIVERY_FAIL, "daemon wakeup");
    }
    
    return TRUE;
}


/**
 * Notify the specified waiter that its wait is over.
 */
void
fs_notifyWaiter(
    fs_Context  *context,
    void        *waitList,
    fs_Waiter   *waiter)
{
    int32 rc;

    assert(waiter->event.p1.u.p == waitList);
    rc = thread_event(waiter->handle, &waiter->event);
    if (rc != NFX_OK)
    {
        // Don't treat this as an error, it could occur
        // during unmount or other cases where the waiting
        // thread has already terminated.
        //fs_error(context, FS_EVENT_DELIVERY_FAIL, "waking waiter");
        printf("fs_notifyWaiter could not deliver\n");
    }
}


/**
 * Print the contents of an fs_Context.
 */
void
fs_contextDump(
    fs_Context  *context)
{
    printf("Context at %p:\n", context);
    
    printf("volId 0x%llx\n", context->volId.whole_thing);
    printf("smHandle %llx\n", (uint64)((uintptr_t)context->smHandle));
    if (context->reqHdr != NULL)
    {
        printf("reqHdr @ %p, reqType %s\n", context->reqHdr,
               context->reqType->name);
    }
    printf("srcAddr %x\n", context->srcAddr);
    printf("respHdr @ %p, size %d\n", context->respHdr, context->respSize);
    printf("initRespSize %d\n", context->initRespSize);
    printf("reqInfo @ %p\n", context->reqInfo);
    printf("logSummary @ %p\n", context->logSummary);
    printf("arena @ %p\n", context->arena);
    printf("randomSeed %lld\n", context->randomSeed);
    printf("flags %x\n", context->flags);
#ifdef FS_DEBUG_DUMP_LOCK_OWNERS
    fs_lockPrintOwned(context);
#endif

} /* fs_contextDump */


/* 
 * Cleanup all resources related to this request.
 * If this is a write req, remove all cod's
 */
static void
fs_cleanup_req(
    const fs_MsgReqHdr    *reqHdr)
{
    if (reqHdr->requestType == FS_REQ_WRITE)
    {
        fs_MsgWriteReq  *req = (fs_MsgWriteReq*)reqHdr;
        int  loop;
        uint32 codLength = 0 ;
       
        if (req->writeLength > 0)
        {
            for (loop = 0; loop < req->nbrUsedCod ; loop++)
            {
                codLength += req->cod[loop].length;

#ifdef SBM_REV_2
                fs_cod_free_cod (req->cod[loop].cod);
#else
                fs_cod_write_remove (req->cod[loop].cod);
#endif
            }
            if(codLength != req->writeLength)
            {
                printf("fs_writeThread codLength %d != req->writeLength %lld, Used Cod %d\n",
                       codLength, req->writeLength, req->nbrUsedCod);
                for (loop = 0; loop < req->nbrUsedCod ; loop++)
                {
                    printf("Nbr %d  cod 0x%llx length %d offset %d\n",
                           loop, req->cod[loop].cod, req->cod[loop].length, req->cod[loop].offset);
                
                    codLength += req->cod[loop].length;
                }
                FS_ASSERT(FALSE);
            }
        }
        else
        {
            FS_ASSERT (req->nbrUsedCod == 0);
        }
    }
}


#ifdef FS_INCORE_STATS
static inline void
fs_updateResourceStats(uint32 used, int type)
{
    if (used > fs_incoreStats->resource[type].maxUsed)
    {
        fs_incoreStats->resource[type].maxUsed = used ;
    }
}
#else
void fs_updateResourceStats(uint32 used, int type);
#define fs_updateResourceStats(u,t)
#endif

static inline int
fs_incoreNotEnoughResource(uint32 total, uint32 used,
						   int type)
{
	uint32 totalMinusRelAbs;
	if (type == FS_RESOURCE_USERBUFCACHE 
		|| type == FS_RESOURCE_METABUFCACHE_NONDIR
		|| type == FS_RESOURCE_METABUFCACHE_COMP_NONDIR
		|| type == FS_RESOURCE_METABUFCACHE_DIR
		|| type == FS_RESOURCE_INODECACHE
		|| type == FS_RESOURCE_VOL_LOG) {
		/* Relative is always 0 */
		totalMinusRelAbs = total - fs_incoreTune->resource[type].absMin;
	}
	else if (type == FS_RESOURCE_ULTRABUF ||
			 type == FS_RESOURCE_MEM)
	{
		/* total is contstant and value is precalculated */
		totalMinusRelAbs = fs_incoreTune->resource[type].totalMinusRelAbs;
	}
	else {
		totalMinusRelAbs =  total - fs_incoreTune->resource[type].absMin -
			total / 128 * fs_incoreTune->resource[type].relMin;
	}
	return
		(totalMinusRelAbs -
		 fs_nbrOfContexts * fs_incoreTune->resource[type].loadFactor) < used;
}


static int
fs_resourceError(
    fs_FileSystem           *fileSystem,
    const fs_MsgReqHdr            *reqHdr,
	const fs_ReqTypeTbl   		*reqType)
{
	int resourceType;
	int i;
	
	/* Find out which resource failed, update stats and return error */
	resourceType = FS_RESOURCE_LOG;
	if (reqType->flags & FS_REQ_FLAGS_LOG)
	{
		if (fs_incoreNotEnoughResource (fileSystem->logSize,
										fileSystem->logRoot->currLen,
										FS_RESOURCE_LOG))
		{
			goto found;
		}
	}

	resourceType = FS_BUF_CACHE_ITYPE_NONDIR;
	i = FS_BUF_CACHE_ITYPE_NONDIR;
	if (fs_incoreNotEnoughResource (
			fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimit[i],
			fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufs[i],
			FS_RESOURCE_USERBUFCACHE))
	{
		goto found;
	}

	i = FS_BUF_CACHE_ITYPE_NONDIR;
	resourceType = FS_RESOURCE_METABUFCACHE_NONDIR;
	if (fs_incoreNotEnoughResource (
			fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[i],
			fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[i],
			FS_RESOURCE_METABUFCACHE_NONDIR))
	{
		goto found;
	}
	
	resourceType = FS_RESOURCE_METABUFCACHE_COMP_NONDIR;
	i = FS_BUF_CACHE_ITYPE_COMP_NONDIR;
	if (fs_incoreNotEnoughResource (
			fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[i],
			fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[i],
			FS_RESOURCE_METABUFCACHE_COMP_NONDIR))
	{
		goto found;
	}
	
	resourceType = FS_RESOURCE_METABUFCACHE_DIR;
	i = FS_BUF_CACHE_ITYPE_DIR;
	if (fs_incoreNotEnoughResource (
			fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[i],
			fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[i],
			FS_RESOURCE_METABUFCACHE_DIR))
	{
		goto found;
	}

	resourceType = FS_RESOURCE_METABUFCACHE_COMP_DIR;
	i = FS_BUF_CACHE_ITYPE_COMP_DIR;
	if (fs_incoreNotEnoughResource (
			fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[i],
			fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[i],
			FS_RESOURCE_METABUFCACHE_COMP_DIR))
	{
		goto found;
	}

	resourceType = FS_RESOURCE_INODECACHE;
	if (fs_incoreNotEnoughResource (
			fs_inodeCache->count.numCachedLimit,
			fs_inodeCache->count.numDirty,
			FS_RESOURCE_INODECACHE))
	{
		goto found;
	}

#if !defined(FS_STANDALONE)

	resourceType = FS_RESOURCE_MEM;
#ifdef FS_USE_MALLOC_SLAB
    if (fs_incoreNotEnoughResource(
        (uint32)((mm_pool_end - mm_pool_start) / 1024),
        (uint32)(mm_pool_inuse / 1024),
        FS_RESOURCE_MEM)) 
    {
		goto found;
    }
#else  // FS_USE_MALLOC_SLAB
	if (fs_incoreNotEnoughResource (
        (uint32)(eee_ramTotalFlexSize() / 1024),
        (uint32)(eee_ramTotalFlexUsed() / 1024),
        FS_RESOURCE_MEM))
	{
		goto found;
	}
#endif // FS_USE_MALLOC_SLAB

	resourceType = FS_RESOURCE_ULTRABUF;
	if (fs_incoreNotEnoughResource (eee_getTotalNbrOfBufs(EEE_BUF_ULTRA),
									eee_getTotalNbrOfBufs(EEE_BUF_ULTRA) - 
									eee_getAvailableNbrOfBufs(EEE_BUF_ULTRA),
									FS_RESOURCE_ULTRABUF))
	{
		goto found;
	}
#endif

	if (fileSystem->logRoot && fileSystem->logRoot->volBufRoot)
	{
		vl_VolBufRoot   *volBufRoot;
		uint32          currNbrBuf;

		volBufRoot = fileSystem->logRoot->volBufRoot;

		NFX_SPIN_LOCK(&volBufRoot->idSpin);
		currNbrBuf = volBufRoot->nbr.currNbrBuf;
		// currNbrBuf = volBufRoot->nbr.maxNbrIO;
		NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

		resourceType = FS_RESOURCE_VOL_LOG;
		if (fs_incoreNotEnoughResource (VL_BUFFER_LOG_MAX,
										MIN(VL_BUFFER_LOG_MAX, currNbrBuf),
										FS_RESOURCE_VOL_LOG))
		{
			goto found;
		}
	}

 found:

//	printf("fs_incoreEnoughRes: drop %s REQ due to no %s resources\n"
//	    "  (limit abs %d rel %d load %d  threads %d)\n",
//	       fs_reqTypeTbl[reqHdr->requestType].name, 
//	       fs_incoreTune->resource[resourceType].name,
//	       fs_incoreTune->resource[resourceType].absMin,
//	       fs_incoreTune->resource[resourceType].relMin,
//	       fs_incoreTune->resource[resourceType].loadFactor,
//	       fs_nbrOfContexts);

	fs_cleanup_req(reqHdr);

#ifdef FS_INCORE_STATS
	fs_incoreStats->resource[resourceType].nbrDrop++;
#endif
	NFX_SPIN_LOCK(&fs_errStatsSpinId);
	fs_errStats.resource[resourceType] += 1;
	if (fileSystem)
	{
		fileSystem->nfsStats.errs[reqHdr->requestType] += 1;
		fileSystem->errStats.resource[resourceType] += 1;
	}
	NFX_SPIN_UNLOCK(&fs_errStatsSpinId);

	return FS_ERR_NO_RESOURCE;
	
}

/**
 * Setup the context block for a filesystem request thread. It will
 * allocate reqInfoSize bytes and point reqInfo field to it and
 * initialize other fields in the context block as
 * required.
 */
static int32
fs_threadContextInit(
	const fs_ReqTypeTbl   * reqType,
    const fs_MsgReqHdr    * reqHdr,
    int32           contextFlags,
    uint32   		srcAddr,
    fs_Context      **contextp)
{
    fs_Context      *context;
    evm_volId_t     volId;
    fs_FileSystem   *fileSystem = NULL;
    int32           respAreaSize;
	uint32			reqTypeFlags = reqType->flags;
	int rc;

    volId.whole_thing = 0;

    if (reqType->extractVolIdFunc != NULL)
    {
        volId = reqType->extractVolIdFunc(reqHdr);
        fileSystem = fs_fileSystemForVolume(volId);
		if (fileSystem == NULL && ((reqTypeFlags & FS_REQ_FLAGS_CLIENT) || 
								   !(reqTypeFlags & FS_REQ_FLAGS_NO_FS_OK))) {
			rc = FS_ERR_NOT_MOUNTED;
			goto drop_no_resource;
		}
		
        if ((reqTypeFlags & FS_REQ_FLAGS_CLIENT))
        {
            int32 itype;
            uint32 metaBufLimit;
            uint32 metaBufsUsed;
			int notEnoughResource;

            // Ignore all client originated requests if the
            // volume is not available.
            if ((fileSystem->state != FS_VOL_STATE_ONLINE) &&
				(fileSystem->state != FS_VOL_STATE_ONLINE_RDONLY)) 
            {
				rc = FS_ERR_NOT_MOUNTED;
				goto drop_no_resource;
            }

            // reject for client modify requests if snapshot is in progress
            if ((fileSystem->super->snapState.phase != FS_SNAP_PHASE_READY) &&
                (reqTypeFlags & FS_REQ_FLAGS_MODIFY))
            {
                rc = FS_ERR_NO_RESOURCE;
				goto drop_no_resource;
            }

            /*
             * Check in advance for log space, cache space,
			 * available memory.  Don't proceed with the request
			 * if either is at threshold level.
			 *
			 *  (disk space and free inodes checked later, ENOSPC
			 *   returned for those)
             */

			notEnoughResource = 0;

            if (reqTypeFlags & FS_REQ_FLAGS_LOG)
            {
				fs_updateResourceStats(fileSystem->logRoot->currLen,
									   FS_RESOURCE_LOG);
				notEnoughResource |=
					fs_incoreNotEnoughResource(fileSystem->logSize,
											   fileSystem->logRoot->currLen,
											   FS_RESOURCE_LOG);
            }

            itype = FS_BUF_CACHE_ITYPE_NONDIR;
			metaBufLimit = fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimit[itype];
			metaBufsUsed = fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufs[itype];

			fs_updateResourceStats(
				fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufs[itype],
				FS_RESOURCE_USERBUFCACHE);
			notEnoughResource |=
				fs_incoreNotEnoughResource (
                    metaBufLimit, metaBufsUsed, 
					FS_RESOURCE_USERBUFCACHE);

            itype = FS_BUF_CACHE_ITYPE_NONDIR;
			metaBufLimit = fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype];
			metaBufsUsed = fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype];
			
			fs_updateResourceStats(metaBufsUsed,
								   FS_RESOURCE_METABUFCACHE_NONDIR);
			notEnoughResource |= fs_incoreNotEnoughResource (
				metaBufLimit, metaBufsUsed,
				FS_RESOURCE_METABUFCACHE_NONDIR);

            itype = FS_BUF_CACHE_ITYPE_COMP_NONDIR;
            metaBufLimit = fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype];
            metaBufsUsed = fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype];
			fs_updateResourceStats(
				metaBufsUsed, FS_RESOURCE_METABUFCACHE_COMP_NONDIR);
			notEnoughResource |= fs_incoreNotEnoughResource (
				metaBufLimit, metaBufsUsed,
				FS_RESOURCE_METABUFCACHE_COMP_NONDIR);

            itype = FS_BUF_CACHE_ITYPE_DIR;
            metaBufLimit = fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype];
            metaBufsUsed = fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype];
			fs_updateResourceStats(
				metaBufsUsed, FS_RESOURCE_METABUFCACHE_DIR);
			notEnoughResource |= fs_incoreNotEnoughResource (
				metaBufLimit, metaBufsUsed,
				FS_RESOURCE_METABUFCACHE_DIR);

            itype = FS_BUF_CACHE_ITYPE_COMP_DIR;
            metaBufLimit = fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype];
            metaBufsUsed = fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype];
			fs_updateResourceStats(
				metaBufsUsed, FS_RESOURCE_METABUFCACHE_COMP_DIR);
			notEnoughResource |= fs_incoreNotEnoughResource (
				metaBufLimit, metaBufsUsed,
				FS_RESOURCE_METABUFCACHE_COMP_DIR);

			fs_updateResourceStats(
				fs_inodeCache->count.numDirty,
				FS_RESOURCE_INODECACHE);
			notEnoughResource |= fs_incoreNotEnoughResource (
				fs_inodeCache->count.numCachedLimit,
				fs_inodeCache->count.numDirty,
				FS_RESOURCE_INODECACHE);

#if !defined(FS_STANDALONE)

            fs_updateResourceStats(
                (uint32)(eee_ramTotalFlexUsed() / 1024),
                FS_RESOURCE_MEM);
			if (fs_incoreNotEnoughResource(
					(uint32)(eee_ramTotalFlexSize() / 1024),
					(uint32)(eee_ramTotalFlexUsed() / 1024),
					FS_RESOURCE_MEM)) {
#ifdef FS_USE_MALLOC_SLAB
				cleanup_caches();
                // NO context AVAILABLE
                // fs_bufFindAndFreeResource (context, FALSE, FALSE, FALSE, fs_incoreTune->bufFreeCount);
#endif // FS_USE_MALLOC_SLAB
				notEnoughResource |= fs_incoreNotEnoughResource(
					(uint32)(eee_ramTotalFlexSize() / 1024),
					(uint32)(eee_ramTotalFlexUsed() / 1024),
					FS_RESOURCE_MEM);
			}

			fs_updateResourceStats(
				eee_getTotalNbrOfBufs(EEE_BUF_ULTRA) - 
				eee_getAvailableNbrOfBufs(EEE_BUF_ULTRA),
				FS_RESOURCE_ULTRABUF);
			notEnoughResource |= fs_incoreNotEnoughResource (
				eee_getTotalNbrOfBufs(EEE_BUF_ULTRA),
				eee_getTotalNbrOfBufs(EEE_BUF_ULTRA) - 
				eee_getAvailableNbrOfBufs(EEE_BUF_ULTRA),
				FS_RESOURCE_ULTRABUF);
#endif // FS_STANDALONE

            if (fileSystem->logRoot && fileSystem->logRoot->volBufRoot)
            {
                vl_VolBufRoot   *volBufRoot;
                uint32          currNbrBuf;

                volBufRoot = fileSystem->logRoot->volBufRoot;

                NFX_SPIN_LOCK(&volBufRoot->idSpin);
                currNbrBuf = volBufRoot->nbr.currNbrBuf;
                // currNbrBuf = volBufRoot->nbr.maxNbrIO;
                NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

				fs_updateResourceStats(
					MIN(VL_BUFFER_LOG_MAX, currNbrBuf),
					FS_RESOURCE_VOL_LOG);
				notEnoughResource |= fs_incoreNotEnoughResource (
					VL_BUFFER_LOG_MAX,
					MIN(VL_BUFFER_LOG_MAX, currNbrBuf),
					FS_RESOURCE_VOL_LOG);
            }

			if (notEnoughResource)
				return fs_resourceError(fileSystem, reqHdr, reqType);
        }
    }

    FS_ASSERT (reqType->reqInfoSize == 0);

    context = fs_ramAlloc(FS_RAM_CONTEXT, sizeof(fs_Context));

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs_nbrOfContexts++;
    if (fs_nbrOfContexts > fs_maxNbrOfContexts)
    {
        fs_maxNbrOfContexts = fs_nbrOfContexts;
    }
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

    // Don't do this for client reqs, it hurts too much.
    if (reqHdr->requestType > FS_REQ_NFS_MAX)
    {
#ifdef FS_USE_MALLOC_SLAB
		fs_threadContextReinit(context);
#else
        bzero(context, sizeof(fs_Context));
#endif
    }
    else
    {
#ifndef FS_USE_MALLOC_SLAB
        fs_threadContextCtor(context, sizeof(fs_Context));
#endif
    }

	ASSERT(context->daemonType == FS_DAEMON_TYPE_INVALID);

	*(const fs_ReqTypeTbl **)(&context->reqType) = reqType;
    context->reqHdr = reqHdr;
    context->srcAddr = srcAddr;

    context->sharedLockCount = 0;
    context->use83Names = FALSE;
    context->caseInsensitive = FALSE;
    context->hashCollides = FALSE;

    fs_lockContextInit(context, SHARED_LOCK_LIST_INIT);

    // Initialize timing instrumentation.
	eee_rtcStart(&context->startTime);

	contextFlags |= (reqTypeFlags & (FS_REQ_FLAGS_STARV | FS_REQ_FLAGS_MODIFY));

	if (((reqTypeFlags & FS_REQ_FLAGS_SPACE) &&
		!fs_haveMinfreeAvail(fileSystem)) ||
		((reqTypeFlags & FS_REQ_FLAGS_INODE) &&
		!fs_haveMininodesAvail(fileSystem) &&
        fileSystem->super->blockInodeRatio <= 1))
    {
        contextFlags |= FS_CONTEXT_ENOSPC_CONDITION;
	}

    respAreaSize = reqType->respAreaSize;
    if (respAreaSize != 0)
    {
        // Additional space needed in response?
        if ((reqTypeFlags & FS_REQ_FLAGS_ATTRS) != 0)
        {
            *(fs_AttrType *)&(reqHdr->attrType) = reqHdr->attrType & 0xff;
            respAreaSize += sizeof(fs_Attr) * fs_onesPerByte[reqHdr->attrType];
        }

        context->respSize = respAreaSize;
        context->initRespSize = respAreaSize;

#if defined(FS_STANDALONE)
        context->respHdr = fs_ramAlloc(FS_RAM_RESP_MSG, respAreaSize);
#else
        /*
         * Management plane fragmentation does not work
         * from the fp yet, use eee_sendmsg iso eee_fwd,
         * so allocate from the heap.
         */
//        if (EEE_GET_DTYPE(srcAddr) == EEE_MGMT)
//        {
//           context->respHdr = fs_ramAlloc(FS_RAM_RESP_MSG, respAreaSize);
//        }
//        else
        {
            FS_ASSERT(respAreaSize <= EEE_BUFSIZE_ULTRA);
            context->respHdr = eee_allocateBuffer(EEE_SHARED_POOL,
                                                  EEE_BUF_ULTRA,
                                                  NFX_COS_HIGH);
            contextFlags |= FS_CONTEXT_SHARED_RESPONSE;
        }
#endif

        // Don't do this for client reqs, it hurts too much.
        if (reqHdr->requestType > FS_REQ_NFS_MAX)
        {
            bzero(context->respHdr, respAreaSize);
        }

        context->respHdr->requestType = reqHdr->requestType;
        context->respHdr->clientPrivate = reqHdr->clientPrivate;
        context->respHdr->clientId = reqHdr->clientId;
        context->respHdr->attrType = 0;
        context->respHdr->len = 0;

#if defined(FS_STANDALONE)
        context->respHdr->sockIndex = reqHdr->sockIndex;
        context->respHdr->nfsReqNum = reqHdr->nfsReqNum;
#endif
    }
    else
    {
        context->respHdr = NULL;
        context->respSize = 0;
        context->initRespSize = 0;
    }

    // We have context and response allocation, set return ptr.
    context->volId = volId;
	context->fileSystem = fileSystem;

	if (fileSystem)
    {
        NFX_SPIN_LOCK(&fileSystem->fs_statsReqRcvdSpinId);
        fileSystem->recvReqStats.num[reqHdr->requestType] += 1;
#if defined(FS_DEBUG_CHECK_CRASH)
        if (fs_debugCrashEnabled)
            fs_debugCheckCrash(context);
#endif
        NFX_SPIN_UNLOCK(&fileSystem->fs_statsReqRcvdSpinId);

        NFX_SPIN_LOCK(&fileSystem->spin);
        DLL_INSERT_TAIL(&fileSystem->activeThreadHdr,
                        context, fsEntry);

        NFX_SPIN_UNLOCK(&fileSystem->spin);

        // We will access the file system so we need to acquire
        // the access lock in shared mode.
        // This is used for interaction with quiescing operations.
        // PR #1789: unmount will acquire excl global accessLock
        //           so don't acquire it share now

		contextFlags |= ((reqTypeFlags & FS_REQ_FLAGS_NO_ACCESS_LOCK)
			^ FS_REQ_FLAGS_NO_ACCESS_LOCK);
    }

    if (reqTypeFlags & FS_REQ_FLAGS_LOG)
    {
        // Allocate log summary separately, it will
        // hang around longer.
        fs_logBeginTransaction(context);

#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
        if ((reqHdr->requestType < FS_REQ_NFS_MAX) &&
            (reqHdr->proto == FS_PROTO_NFS_V3))
        {
            NFX_SPIN_LOCK(&fileSystem->spin);
            contextFlags |= FS_CONTEXT_TRANSACTION_COUNTED;
            fileSystem->nbrOngoingTrans++;
            if (fileSystem->maxOngoingTrans < fileSystem->nbrOngoingTrans)
                fileSystem->maxOngoingTrans = fileSystem->nbrOngoingTrans;
            NFX_SPIN_UNLOCK(&fileSystem->spin);
        }
#endif           

    }

#ifdef FS_DEBUG_LOCK_ORDER
    context->ignoreLockOrder = FALSE;
#endif
	context->flags = contextFlags;
    *contextp = context;

    return NFX_OK;

 drop_no_resource:
#ifdef FS_INCORE_STATS
	fs_incoreStats->nbrDropNoResource++;
#endif
	fs_cleanup_req(reqHdr);
	return rc;
	
} /* fs_threadContextInit */


/**
 * Free the request buffer for the specified context.  If the request
 * has already been freed, this is a no-op.
 */
static void
fs_contextFreeRequest(
    fs_Context      *context)
{
	void *reqHdr = (void *)context->reqHdr;
    if (reqHdr != NULL) {
		context->reqHdr = NULL;
#if defined(FS_STANDALONE)
		assert((context->flags & FS_CONTEXT_SHARED_REQUEST) == 0);
		fs_ramDealloc(FS_RAM_REQ_MSG, reqHdr);
#else
		eee_deallocateBuffer(reqHdr);
#endif		
     }
}


/**
 * Free the response buffer for the specified context.  If the response
 * has already been freed, this is a no-op.
 */
void
fs_contextFreeResponse(
    fs_Context      *context)
{
	void *respHdr = context->respHdr;
    if (respHdr != NULL) {
        context->respHdr = NULL;
#if defined(FS_STANDALONE)
		assert((context->flags & FS_CONTEXT_SHARED_RESPONSE) == 0);
		fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);
#else		
		eee_deallocateBuffer(respHdr);
#endif
    }
}


/**
 * Default cleanup function for thread contexts.
 */
void
fs_threadContextDestroy(
    fs_Context      *context,
    int		         doLockTerm)
{
	int contextFlags = context->flags;
	fs_FileSystem *fileSystem = context->fileSystem;

	// Careful here, for the unmount thread the filesystem
	// is gone!
	if (fileSystem != NULL) {
		if (contextFlags & FS_CONTEXT_MODIFY_LOCK_ACQUIRED)
		{
            fs_unlockShared(context, &fileSystem->modifyLock);
        }
		if (contextFlags & FS_CONTEXT_ACCESS_LOCK_ACQUIRED)
        {
            fs_unlockShared(context, &fileSystem->accessLock);
        }
		if (fileSystem->state != FS_VOL_STATE_FAILING && doLockTerm)
		{
			fs_lockTerminateContext(context);
		}
		if (!DLL_IS_ISOLATED(context, fsEntry))
		{
			NFX_SPIN_LOCK(&fileSystem->spin);
			DLL_REMOVE(&fileSystem->activeThreadHdr, context, fsEntry);
			NFX_SPIN_UNLOCK(&fileSystem->spin);
		}
	}
	else {
		DLL_ISOLATE_ELEM(context, fsEntry);
	}

	fs_contextFreeResponse(context);
	fs_contextFreeRequest(context);

    if (context->logSummary) {
        fs_logDestroySummary(context, context->logSummary);
		context->logSummary = NULL;
	}
	if (context->arena) 
    {
        fs_contextMemFree(context);
    }
	if(context->pendingFreeExtents)
    {
        fs_ramDealloc(FS_RAM_EXTENT_VECTOR, context->pendingFreeExtents);
        context->pendingFreeExtents = NULL;
    }    

    fs_lockContextFree(context);

	ASSERT(context->arena == NULL);
	ASSERT(context->logSummary == NULL);
	ASSERT(DLL_IS_ISOLATED(context, fsEntry));
	ASSERT(context->eek.inUse == FALSE);
	ASSERT(context->pendingFreeExtents == NULL);
	ASSERT(context->daemonType == FS_DAEMON_TYPE_INVALID);

    fs_ramDealloc(FS_RAM_CONTEXT, context);

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs_nbrOfContexts--;
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
}

/**
 * Cleanup the dump context.
 */
static void
fs_ndmpContextDumpDestroy(
    fs_Context  *context,
    uint32       cookie)
{
    fs_FileSystem       *fs = context->fileSystem;
    fs_DumpContext_t    *dumpContext = DLL_NIL;

    fs_lockShared(context, &fs->dumpContextsLock, NULL);
    DLL_ITERATE(dumpContext, &fs->dumpContexts, dumpList)
    {
        if (dumpContext->cookie == cookie)
            break;
    }
    fs_unlockShared(context, &fs->dumpContextsLock);
    if (dumpContext == DLL_NIL)
    {
        // Dump context already cleaned up.
        return;
    }

    fs_cleanDumpContext(context, dumpContext, FALSE);

    // fs_cleanDumpContext removed dumpContext from
    // the fs list, so we can simply dealloc it now.
    fs_ramDealloc(FS_RAM_DUMP_CONTEXT, dumpContext);
}


/**
 * Cleanup the restore context.
 */
static void
fs_ndmpContextRestoreDestroy(
    fs_Context  *context,
    uint32       cookie)
{
    fs_FileSystem           *fs = context->fileSystem;
    fs_RestoreContext_t     *restoreContext = DLL_NIL;

    fs_lockShared(context, &fs->restoreContextsLock, NULL);
    DLL_ITERATE(restoreContext,&fs->restoreContexts, restoreList)
    {
        if (restoreContext->cookie == cookie)
                break;
    }
    fs_unlockShared(context, &fs->restoreContextsLock);
    if (restoreContext == DLL_NIL)
    {
        // Restore context already cleaned up.
        return;
    }

    fs_cleanRestoreContext(context, restoreContext, FALSE);

    // fs_cleanRestoreContext didn't remove it from the
    // fs list, so we must do that here before deallocing it.
    fs_lockExclusive(context, &fs->restoreContextsLock, NULL);
    DLL_REMOVE(&fs->restoreContexts, restoreContext, restoreList);
    fs_unlockExclusive(context, &fs->restoreContextsLock);

    fs_ramDealloc(FS_RAM_DUMP_CONTEXT, restoreContext);
}


/**
 * Cleanup function for NDMP related thread contexts.
 */
void
fs_ndmpContextDestroy(
    fs_Context  *context)
{
    const fs_MsgReqHdr    *reqHdr = context->reqHdr;
    uint32          cookie;

    switch (reqHdr->requestType)
    {
    case FS_REQ_DUMP_START:
        cookie = ((fs_MsgDumpStartReq_t *)reqHdr)->start.ndmpSessId;
        fs_ndmpContextDumpDestroy(context, cookie);
        break;
    case FS_REQ_DUMP_SETUP:
        cookie = ((fs_MsgDumpSetupReq_t *)reqHdr)->dump.ndmpSessId;
        fs_ndmpContextDumpDestroy(context, cookie);
        break;
    case FS_REQ_DUMP_NAMES:
        cookie = ((fs_MsgDumpNamesReq_t *)reqHdr)->names.ndmpSessId;
        fs_ndmpContextDumpDestroy(context, cookie);
        break;
    case FS_REQ_REST_START:
        cookie = ((fs_MsgRestoreStartReq_t *)reqHdr)->start.ndmpSessId;
        fs_ndmpContextRestoreDestroy(context, cookie);
        break;
    case FS_REQ_REST_SETUP:
        cookie = ((fs_MsgRestoreSetupReq_t *)reqHdr)->restore.ndmpSessId;
        fs_ndmpContextRestoreDestroy(context, cookie);
        break;
    case FS_REQ_REST_NAMES:
        cookie = ((fs_MsgRestoreNamesReq_t *)reqHdr)->names.ndmpSessId;
        fs_ndmpContextRestoreDestroy(context, cookie);
        break;
    default:
        fs_error(context, FS_INTERNAL_ERR, "ndmp context dest");
        break;
    }

    fs_lockContextFree(context);

    fs_ramDealloc(FS_RAM_CONTEXT, context);

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs_nbrOfContexts--;
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
}


/**
 * Send the response for the thread if applicable.
 */

void
fs_threadRespond(
    fs_Context          *context,
    int32               retStatus)
{
    int32       rc;
	eeeRTC_t	deltaCpu, deltaTotal;
#if defined(FS_STANDALONE)
    int32       cpuId;
    char        *resp;
#endif

    if (context->respHdr != NULL)
    {
        //printf("threadRespond: to reqType %d\n",
        //  context->respHdr->requestType);
        context->respHdr->fsStatus = retStatus;

        if ((context->respHdr->requestType >= 0) &&
            (context->respHdr->requestType < FS_REQ_NFS_MAX))
        {
            NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
            if (retStatus == NFX_OK)
                fs_replyStats.ok[context->respHdr->requestType] += 1;
            else
                fs_replyStats.err[context->respHdr->requestType] += 1;
            NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);

            NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
            if (retStatus == NFX_OK)
                context->fileSystem->replyStats.ok[context->respHdr->requestType] += 1;
            else
                context->fileSystem->replyStats.err[context->respHdr->requestType] += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
        }

#if defined(FS_STANDALONE)
        cpuId = (int32) pthread_getspecific(fs_simKeyCpuId);
        resp = (char *) context->respHdr;
        context->respHdr = NULL;
        rc = fs_simSendMsg( cpuId,
                            context->reqHdr->src,
                            resp,
                            context->respSize);
#else
        if ((context->flags & FS_CONTEXT_SHARED_RESPONSE) == 0)
        {
            rc = eee_sendMessage(   context->respHdr,
                                    context->respSize,
                                    context->srcAddr,
                                    fs_packedPortAddr(nfxMyCPUId, EEE_MGMT));
        }
        else
        {
            eee_descPtr_t       edesc = NULL;

            edesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
            if (edesc == NULL)
            {
                if (context->fileSystem)
                {
                    fs_volumeException(context, context->fileSystem,
                                       FS_VOLUME_EXCEPTION);
                }
                return;
            }

            edesc->bd[1].buf = (void *) context->respHdr;
            edesc->bd[1].len = context->respSize;
            edesc->bd[1].control = EDESC_BD_ULTRA | EDESC_BD_CMD_IN_USE;

            // let eee_forwardPacket free the ptr
            context->respHdr = NULL;

            edesc->hdr.control = (1 << EDESC_BUFCNT_SHIFT) | context->respSize;
            edesc->hdr.src_port = fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL);
            edesc->hdr.dest_port = context->srcAddr;
            rc = eee_forwardPacket(edesc);

        }
#endif
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EEE_ERR, "thread respond");
        }

        // update timing info
		eee_rtcDuration(&context->tmpTime, &deltaCpu);
		eee_rtcDuration(&context->startTime, &deltaTotal);
        if (context->fileSystem)
		{
			uint32      requestType;
#ifdef FS_DEBUG_NFS_DELTA_BUCKET
            int32   i, bucket = 0;
			static boolean first = TRUE;


			if (first) {
				first = FALSE;
				// Get the thresholds in rtc form.
				for (i = 0; i < FS_NFS_TIME_MAX; i++)
				{
					eee_usecToRTC((uint64)fs_nfsStatsUsecs[i],
								  &(fs_nfsStatsRTC[i]));
				}
			}
#endif

			if (context->reqHdr != NULL)
            {
				if (context->reqHdr->requestType < FS_REQ_NFS_MAX)
                {
                    NFX_SPIN_LOCK(&context->fileSystem->spin);
					context->fileSystem->currentActiveReqs -= 1;
                    NFX_SPIN_UNLOCK(&context->fileSystem->spin);
					requestType = context->reqHdr->requestType;
				}
				else if (context->reqHdr->requestType == FS_REQ_DUMP_START)
					requestType = FS_REQ_NFS_MAX;
				else if (context->reqHdr->requestType == FS_REQ_REST_START)
					requestType = FS_REQ_NFS_MAX + 1;
				else if (context->daemonType != FS_DAEMON_TYPE_INVALID)
					requestType = context->daemonType + 2 + FS_REQ_NFS_MAX;
				else
                {
					// An unused value.
					requestType = 0xFFFFFFFF;
				}
				if (requestType != 0xFFFFFFFF)
                {
                    NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
					(void)eee_rtcIncr(&(context->fileSystem->yieldStats.num
										[requestType][FS_YIELD_TYPE_ON_CPU]),
									  &deltaCpu);
					(void)eee_rtcIncr(&(context->fileSystem->yieldStats.num
										[requestType][FS_YIELD_TYPE_TOTAL]),
									  &deltaTotal);
					NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
					NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
					if (context->reqHdr->requestType < FS_REQ_NFS_MAX)
						fs_currentActiveReqs -= 1;
					(void)eee_rtcIncr(&(fs_yieldStats.num
										[requestType][FS_YIELD_TYPE_ON_CPU]),
									  &deltaCpu);
					(void)eee_rtcIncr(&(fs_yieldStats.num
										[requestType][FS_YIELD_TYPE_TOTAL]),
									  &deltaTotal);
					NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
				}
#ifdef FS_DEBUG_NFS_DELTA_BUCKET
				if (requestType < FS_REQ_NFS_MAX)
                {
					for (i = 0; i < FS_NFS_TIME_MAX; i++)
					{
						if (eee_rtcCompare(&deltaCpu, &(fs_nfsStatsRTC[i])) < 0)
						{
							bucket = i;
							break;
						}
					}
					if (i >= FS_NFS_TIME_MAX)
						bucket = FS_NFS_TIME_MAX - 1;

					NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
					context->fileSystem->nfsStats.num[requestType][bucket] += 1;
					NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);

					NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
					fs_nfsStats.num[requestType][bucket] += 1;
					NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
#ifdef FS_DEBUG_PRINT_DELTA_THRESHOLD
					if (delta > fs_elapsedUsecThreshold)
					{
						printf("Thread %s %lld usecs\n",
							context->reqType->name, delta);
					}
#endif   // FS_DEBUG_PRINT_DELTA_THRESHOLD
				}
#endif   // FS_DEBUG_NFS_DELTA_BUCKET
			}
        }
    }

} /* fs_threadRespond */


/**
 * Send ndmp response from context.
 */
void
fs_ndmpResponse(
    fs_Context  *context,
    int32       status)
{
    const fs_MsgReqHdr    *reqHdr = context->reqHdr;
    int32           rc;

    rc = status == NFX_OK ? FS_ERR_NONE : FS_ERR_INVAL;

    if (reqHdr->requestType == FS_REQ_DUMP_START)
    {
        fs_sendNdmpResponse(context, reqHdr->requestType,
                        ((fs_MsgDumpStartReq_t *)reqHdr)->start.ndmpSessId,
                            reqHdr->clientPrivate,
                            rc);
    }
    else if (reqHdr->requestType == FS_REQ_REST_START)
    {
        fs_sendNdmpResponse(context, reqHdr->requestType,
                        ((fs_MsgRestoreStartReq_t *)reqHdr)->start.ndmpSessId,
                            reqHdr->clientPrivate,
                            rc);
    } else if (reqHdr->requestType == FS_REQ_DUMP_SETUP) {
        fs_sendNdmpResponse(context, reqHdr->requestType,
                        ((fs_MsgDumpSetupReq_t *)reqHdr)->dump.ndmpSessId,
                            reqHdr->clientPrivate,
                            rc);
    } else if (reqHdr->requestType == FS_REQ_DUMP_NAMES) {
        fs_sendNdmpResponse(context, reqHdr->requestType,
                        ((fs_MsgDumpNamesReq_t *)reqHdr)->names.ndmpSessId,
                            reqHdr->clientPrivate,
                            rc);
    } else if (reqHdr->requestType == FS_REQ_REST_SETUP) {
        fs_sendNdmpResponse(context, reqHdr->requestType,
                        ((fs_MsgRestoreSetupReq_t *)reqHdr)->restore.ndmpSessId,
                            reqHdr->clientPrivate,
                            rc);
    } else if (reqHdr->requestType == FS_REQ_REST_NAMES) {
        fs_sendNdmpResponse(context, reqHdr->requestType,
                        ((fs_MsgRestoreNamesReq_t *)reqHdr)->names.ndmpSessId,
                            reqHdr->clientPrivate,
                            rc);
    }
}


/**
 * Send an error back to the requestor.  This is used when we fail
 * to create a context for handling the request (typically due to
 * the volume not being available).
 */

/* Do this since we reuse the request msg for the response here. */
NFX_ASSERT_STATIC(sizeof(fs_MsgRespHdr) < sizeof(fs_MsgReqHdr));

void
fs_nullContextResponse(
    const fs_MsgReqHdr    *reqHdr,
    int32           contextFlags,
    uint32   		srcAddr,
    fs_Status       status)
{
    fs_MsgRespHdr   *resp = (fs_MsgRespHdr *) reqHdr;
    fs_MsgReqHdr    tmpReq = *reqHdr;
#if defined(FS_STANDALONE)
    int32           cpuId;
#endif
    int32           rc;

    bzero(resp, sizeof *resp);
    resp->requestType = tmpReq.requestType;
    resp->len = sizeof *resp;
    resp->fsStatus = status;
    resp->clientPrivate = tmpReq.clientPrivate;
    resp->clientId = tmpReq.clientId;


#if defined(FS_STANDALONE)
    resp->src = tmpReq.dest;
    resp->dest = tmpReq.src;
    resp->nfsReqNum = tmpReq.nfsReqNum;
    resp->sockIndex = tmpReq.sockIndex;

    cpuId = (int32) pthread_getspecific(fs_simKeyCpuId);
    rc = fs_simSendMsg(cpuId,
                       tmpReq.src,
                       (void *) resp,
                       sizeof *resp);
#else
    // TODO - check if system buffer is used for response (vs. malloc
    // buffer), use the appropriate send function (this currently
    // assumes a malloc'd buffer is pointed to by respHdr).
    rc = eee_sendMessage(resp,
                         sizeof *resp,
                         srcAddr,
                         fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));

    if (contextFlags & FS_CONTEXT_SHARED_REQUEST)
        eee_deallocateBuffer((void *)reqHdr);
    else 
#if defined(FS_STANDALONE)
        fs_ramDealloc(FS_RAM_MCPU, (void *) reqHdr);
#else
        fs_ramDealloc(FS_RAM_REQ_MSG, (void *) reqHdr);
#endif

#endif

    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_EEE_ERR, "nullContext respond");
    }

} /* fs_nullContextResponse */



/**
 * This is where request threads first come to life in
 * filesystem code.  When this returns, the thread dies.
 */
void
fs_threadBegin(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;
    const fs_MsgReqHdr    *reqHdr = context->reqHdr;
    int32           reqType = reqHdr->requestType;
    int32           rc;

    if (context->reqHdr &&
		context->fileSystem)
    {
        eeeRTC_t	delta;
		uint32      requestType;

		if (context->reqHdr->requestType < FS_REQ_NFS_MAX)
			requestType = context->reqHdr->requestType;
		else if (context->reqHdr->requestType == FS_REQ_DUMP_START)
			requestType = FS_REQ_NFS_MAX;
		else if (context->reqHdr->requestType == FS_REQ_REST_START)
			requestType = FS_REQ_NFS_MAX + 1;
		else if (context->daemonType != FS_DAEMON_TYPE_INVALID)
			requestType = context->daemonType + 2 + FS_REQ_NFS_MAX;
		else {
			// An unused value.
			requestType = 0xFFFFFFFF;
		}

		// get delta, and set new context->tmpTime
        eee_rtcDuration(&context->startTime, &delta);

        /*
         * Always account for tmpTime for all requests. We use this
         * to determine how long a request has been running
         */

		eee_rtc(&context->tmpTime);
		if (requestType != 0xFFFFFFFF) {
			NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
			eee_rtcIncr(&(context->fileSystem->yieldStats.num
						  [requestType][FS_YIELD_TYPE_ADMIN]), &delta);
			NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);

			NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
			eee_rtcIncr(&(fs_yieldStats.num[requestType][FS_YIELD_TYPE_ADMIN]),
						&delta);
			NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
		}
    } 

    // There's a potential race here where the thread could
    // start before the smHandle is set in the context (very
    // unlikely on the embedded system, but has been witnessed
    // in the standalone where a switch could happen at any time),
    // so set the smHandle here to be sure.
    context->smHandle = handle;

    // Lock the global access lock, if necessary.
    if (context->flags & FS_CONTEXT_ACCESS_LOCK_NEEDED)
    {
        fs_lockShared(context, &context->fileSystem->accessLock, NULL);
        context->flags &= ~FS_CONTEXT_ACCESS_LOCK_NEEDED;
        context->flags |= FS_CONTEXT_ACCESS_LOCK_ACQUIRED;
    }
    // Lock the global modify lock, if necessary.
    if (context->flags & FS_CONTEXT_MODIFY_LOCK_NEEDED)
    {
        fs_lockShared(context, &context->fileSystem->modifyLock, NULL);
        context->flags &= ~FS_CONTEXT_MODIFY_LOCK_NEEDED;
        context->flags |= FS_CONTEXT_MODIFY_LOCK_ACQUIRED;
    }

    FS_ASSERT(context->reqType->threadFunc != NULL);

    FS_TRACE(context, FS_TRACE_REQ, ("smHandle %llx ++++ %s ++++",
             (uint64)((uintptr_t)handle), context->reqType->name));


#if !defined(FS_STANDALONE)
    /*
     * Grow the fs for allocating vnops
     */

    switch (reqType) {

    case FS_REQ_WRITE:
    case FS_REQ_CREATE:
    case FS_REQ_MKDIR:
    case FS_REQ_SYMLINK:
    case FS_REQ_MKNOD:
    case FS_REQ_RENAME:
    case FS_REQ_HARDLINK:
        fs_growIfNeeded(context);
        break;
#ifdef FS_DEBUG_1
    case FS_REQ_DUMP_START:
        printf("Dump Started\n");
#endif
    }
#endif

    if (reqType < FS_REQ_NFS_MAX)
    {
        if (context->fileSystem) {
            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->currentActiveReqs += 1;
            if (context->fileSystem->currentActiveReqs >
                context->fileSystem->maxActiveReqs)
            {
                context->fileSystem->maxActiveReqs =
                    context->fileSystem->currentActiveReqs;
            }
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);
        }
        NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
        fs_currentActiveReqs += 1;
        if (fs_currentActiveReqs > fs_maxActiveReqs)
            fs_maxActiveReqs = fs_currentActiveReqs;
            
        NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
    }

    rc = (context->reqType->threadFunc)(context);

    FS_TRACE(context, FS_TRACE_REQ,
             ("smHandle %llx ---- %s ---- returns %d",
             (uint64)((uintptr_t)handle),
             context->reqType->name,
             rc));

#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
    if (context->flags & FS_CONTEXT_TRANSACTION_COUNTED)
    {
		boolean  doSignal = FALSE;

        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->nbrOngoingTrans--;
        context->flags &= ~FS_CONTEXT_TRANSACTION_COUNTED;
        if (context->fileSystem->nbrOngoingTrans == 0)
			doSignal = TRUE;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);

        if (doSignal)
        {
            // only wakeup the log IO, if no one else will do it
            fs_signalVolLogWrite (context);
        }
            
        
//        printf ("nbrOngoingTransactions (resp) still set, %s (cnt %d) rc %d\n", 
//                fs_reqTypeTbl[reqType].name, context->fileSystem->nbrOngoingTrans, rc);
    }
#endif           

#ifdef FS_DEBUG_1
    if (reqType == FS_REQ_DUMP_START) {
        printf("threadFunc returned %d\n", rc);
    }
#endif
    if (context->respSize > context->initRespSize)
    {
        printf("respSize %d is > initRespSize %d\n",
               context->respSize, context->initRespSize);
        FS_ASSERT(FALSE);
    }

    (context->reqType->respFunc)(context, rc);
#ifdef FS_DEBUG_1
    if (reqType == FS_REQ_DUMP_START) {
        printf("response up for dump done\n");
    }
#endif
    (context->reqType->clnupFunc)(context, TRUE);
#ifdef FS_DEBUG_1
    if (reqType == FS_REQ_DUMP_START) {
        printf("Cleanup up for dump done\n");
    }
#endif

    // Action complete, response sent, we're done.

} /* fs_threadBegin */


/**
 * Create a thread, initialize its context.
 */
int32
fs_threadInstantiate(
	const fs_ReqTypeTbl *reqType,
    const fs_MsgReqHdr  *reqHdr,
    int32               contextFlags,
    uint32				srcAddr)
{
    fs_Context          *context = NULL;
    int32               rc = NFX_OK;

    rc = fs_threadContextInit(reqType, reqHdr, contextFlags,
							  srcAddr, &context);
    if (rc != NFX_OK)
    {
        fs_requestsIgnored += 1;
        if (context != NULL)
        {
            context->reqHdr = reqHdr;
            (reqType->respFunc)(context, rc);
            (reqType->clnupFunc)(context, TRUE);
        }
        else
        {
            (reqType->ncRespFunc)(reqHdr, contextFlags, srcAddr, rc);
        }
        return rc;
    }

    context->isThread = TRUE;
    context->smHandle = thread_create(context);
    if (context->smHandle != 0)
    {
        rc = thread_start(&(context->smHandle), fs_threadBegin);
    }
    else
    {
        rc = NFX_ERR;
    }
    
    if (rc != NFX_OK)
    {
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
        if (context->flags & FS_CONTEXT_TRANSACTION_COUNTED)
        {
			boolean doSignal = FALSE;

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->nbrOngoingTrans--;
            context->flags &= ~FS_CONTEXT_TRANSACTION_COUNTED;
            if (context->fileSystem->nbrOngoingTrans == 0)
				doSignal = TRUE;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            if (doSignal)
            {
                // only wakeup the log IO, if no one else will do it
                fs_signalVolLogWrite (context);
            }
            
            //printf ("nbrOngoingTransactions (smHandle) still set, %s (cnt %d)\n", 
            //        reqType->name, context->fileSystem->nbrOngoingTrans);
        }
#endif           

        (reqType->respFunc)(context, FS_ERR_NO_RESOURCE);
        (reqType->clnupFunc)(context, TRUE);

        NFX_SPIN_LOCK(&fs_errStatsSpinId);
        context->fileSystem->errStats.resource[FS_RESOURCE_NO_ESM] += 1;
        fs_errStats.resource[FS_RESOURCE_NO_ESM] += 1;
        NFX_SPIN_UNLOCK(&fs_errStatsSpinId);

        // fs_warn(NULL, FS_ESM_CREATE_ERR, "thread esm");
        return NFX_ERR;
    }

    //printf("created thread sm handle 0x%x for context %p reqType %s\n",
    //  context->smHandle, context, fs_reqTypeTbl[reqHdr->requestType].name);

    return NFX_OK;

} /* fs_threadInstantiate */


#if !defined(FS_STANDALONE)

/**
 * eee timeout handler for fs_yield/FS_YIELD_TYPE_SLEEP 
 */
static int32
fs_yieldTimeoutHandler(
    uint32  timer,
    void    *param1,
    void    *param2,
    void    *param3,
    void    *param4)
{
    fs_Context  *context = (fs_Context *) param2;
    esm_event_t event;
    int32       rc;

    event.event = FS_THREAD_EVENT_YIELD_SLEEP_WAKEUP;
    event.p1.u.dw = 0x123;
    event.p2.u.dw = 0x456;
    rc = esm_deliverEvent((esm_handle_t) (address_t)param1, &event);
    if (rc != NFX_OK)
    {
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }

    return NFX_OK;
}

#endif


/**
 * Yield funnel.  All filesystem yields go through here
 * in order to provide a control point for exception
 * handling (in addition to providing a nice instrumentation  point).
 * time is in seconds
 *
 * FS sleeps also funnel through here with type FS_YIELD_TYPE_SLEEP,
 * and time in msecs, 0 to specify not to sleep but return asap.
 *
 * time is in secs, in msecs for type == FS_YIELD_TYPE_SLEEP
 */

char *fs_yield_type_name [FS_YIELD_TYPE_MAX] = {
    "cpu",
    "user read",
    "user write",
    "meta read",
    "meta write",
    "log read/write",
    "lock",
    "tx_complete",
    "cod Rx",
    "Admin",
    "Sleep",
    "IO retry",
    "total",
    "tape read",
    "tape write",
    "user IO commit"
};


void
fs_yield(
    fs_Context  *context,
    esm_event_t *event,
    int32       time,          // secs, in msecs for type == FS_YIELD_TYPE_SLEEP
    int32       type)
{
    eeeRTC_t	tvYieldStart, elapsed;
    esm_event_t localEvent, wakeupEvent;
    int32       yieldTime = time;
#if !defined(FS_STANDALONE)
    int32       rc, timer = TIMER_ERR_NO_TIMER_AVAILABLE;
#endif
    boolean     volFailed = FALSE;

    FS_ASSERT(type < FS_YIELD_TYPE_MAX);

    eee_rtc(&tvYieldStart);
    if (context->reqHdr &&
        context->fileSystem)
    {
        // Accumulate CPU usage time.
		uint32      requestType;

		if (context->reqHdr->requestType < FS_REQ_NFS_MAX)
			requestType = context->reqHdr->requestType;
		else if (context->reqHdr->requestType == FS_REQ_DUMP_START)
			requestType = FS_REQ_NFS_MAX;
		else if (context->reqHdr->requestType == FS_REQ_REST_START)
			requestType = FS_REQ_NFS_MAX + 1;
		else if (context->daemonType != FS_DAEMON_TYPE_INVALID)
			requestType = context->daemonType + 2 + FS_REQ_NFS_MAX;
		else {
			// An unused value.
			requestType = 0xFFFFFFFF;
		}

		if (requestType != 0xFFFFFFFF) {
			eee_rtcDuration(&context->tmpTime, &elapsed);
			NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
			eee_rtcIncr(&(context->fileSystem->yieldStats.num
						  [requestType][FS_YIELD_TYPE_ON_CPU]),
						&elapsed);
			NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);

			NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
			eee_rtcIncr(&(fs_yieldStats.num[requestType][
						  FS_YIELD_TYPE_ON_CPU]),
						&elapsed);
			NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
		}
    }

    if (context->fileSystem &&
        context->fileSystem->state == FS_VOL_STATE_FAILING)
    {
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
        // This will not return here.
        FS_ASSERT(FALSE);
    }

#if !defined(FS_STANDALONE)

    if (type == FS_YIELD_TYPE_SLEEP)
    {
        if (yieldTime)
        {
            timer = eee_timerCreate(yieldTime, EEE_TIMER_ONCE);
            if (timer == TIMER_ERR_NO_TIMER_AVAILABLE)
            {
                fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
                // This will not return here.
                FS_ASSERT(FALSE);
            }

            rc = eee_timerStart(timer, fs_yieldTimeoutHandler,
                                (void *) (address_t)esm_getCurrHandle(),
								context, NULL, NULL);
            if (rc != NFX_OK)
            {
                eee_timerDestroy(timer);
                fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
                // This will not return here.
                FS_ASSERT(FALSE);
            }
        }
        else
        {
            // no wait time, just send event to myself

            wakeupEvent.event = FS_THREAD_EVENT_YIELD_SLEEP_WAKEUP;
            wakeupEvent.p1.u.dw = 0x123;
            wakeupEvent.p2.u.dw = 0x456;

            rc = esm_deliverEvent(context->smHandle, &wakeupEvent);
            if (rc != NFX_OK)
            {
                fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
                // This will not return here.
                FS_ASSERT(FALSE);
            }
        }

        // no yield timeout here, just depend on the EEE timer
        // or a posted event to do its thing

        yieldTime = 0;
    }
#else
    // TODO: have timers for FS standalone
    if (type == FS_YIELD_TYPE_SLEEP)
    {
        if (time != 0)
        {
            yieldTime = time; // yieldTime in msecs
        }
        else
        {
            // no wait time, just send event to myself            
            wakeupEvent.event = FS_THREAD_EVENT_YIELD_SLEEP_WAKEUP;
            wakeupEvent.p1.u.dw = 0x123;
            wakeupEvent.p2.u.dw = 0x456;
            
            if (esm_deliverEvent(context->smHandle, &wakeupEvent) != NFX_OK)
            {
                fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
                // This will not return here.
                FS_ASSERT(FALSE);
            }
        }
    }
    else
    {
        yieldTime = time * 1000; // yieldTime in msecs
    }
    
#endif

    // remark: yieldtime in seconds, in standalone this is in msecs !!
    thread_yield(&localEvent, yieldTime);

#if !defined(FS_STANDALONE)
    if (time && (type == FS_YIELD_TYPE_SLEEP))
    {
        if (localEvent.event != FS_THREAD_EVENT_YIELD_SLEEP_WAKEUP)
        {
            // stop timer if an event occured before the timer expired
            eee_timerStop(timer);
        }
        eee_timerDestroy(timer);
    }
#endif

    // handle volume failure before events (ref PR #1457)
    // since the  FS_THREAD_EVENT_EXCEPTION can be queued after this local event
    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    if (context->fileSystem &&
        context->fileSystem->state == FS_VOL_STATE_FAILING)
    {
        volFailed = TRUE;
    }
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

    if (volFailed || (localEvent.event == FS_THREAD_EVENT_EXCEPTION))
    {
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
        FS_ASSERT(FALSE);
    }
    else if (localEvent.event == FS_THREAD_EVENT_TERMINATE)
    {
        // PR #1789: remove this context from a lock's waiting list if necc
        if (context->lockWait.smHandle && context->lockWait.lock)
        {
            fs_lockDeleteWaiter(context);
        }

#ifdef FS_DEBUG_LOCK
        // PR #1789: verify context does not hold any locks
        if ((!DLL_IS_EMPTY(&context->ownerList)) || context->sharedLockCount != 0)
        {
            printf("esm %llx terminated holding locks\n",
				   (uint64)((uintptr_t)context->smHandle));
            FS_ASSERT(FALSE);
        }
#endif
        if (context->reqHdr != NULL)
        {
            // Send an error response to this context's requestor.
            (context->reqType->respFunc) (context, NFX_ERR);
            (context->reqType->clnupFunc) (context, FALSE);
        }
        else
        {
            // Its a system related thread, terminate in the
            // usual fashion.
            fs_threadTerminate(context);
        }

        // Return this thread to the executive.
        thread_exit();
    }
    else if (localEvent.event == THREAD_EVENT_TIMEOUT)
    {
        if (type == FS_YIELD_TYPE_IO_LOG ||
#ifdef FS_STANDALONE
            type == FS_YIELD_TYPE_SLEEP ||
#endif // FS_STANDALONE
            type == FS_YIELD_TYPE_IO_COMMIT)
        {
            // Let these ones get quietly back to our caller.
        }
        else
        {
            eeeGMT_t    currTime;
            eeeGMT_t    tvYieldStartGMT;
        
            eee_TimeGMT(&currTime);
			eee_rtcToGMT(&tvYieldStart, &tvYieldStartGMT);
            printf("fs_yield timeout on %s, startTime %s, currTime %s\n",
                   fs_yield_type_name[type],
                   eee_humandate(tvYieldStartGMT.sec),
                   eee_humandate(currTime.sec));

#ifdef FS_DEBUG_LOCK
            if (type == FS_YIELD_TYPE_LOCK &&
                context->lockWait.lock != NULL)
            {
                char lockDesc[FS_LOCK_DESCRIPTION];

                printf("Waiting for: %s\n",
                       fs_lockDescribe(context, context->lockWait.lock, lockDesc));
            }
#endif

            if (type == FS_YIELD_TYPE_TXCOMPL || type == FS_YIELD_TYPE_LOCK)
            {
                // Let these ones get back to our caller.
            }
            else
            {
                fs_volumeException(context, context->fileSystem,
                                   FS_EVENT_TIMEOUT_ERR);
                // This will not return here.
                FS_ASSERT(FALSE);
            }
        }
    }
    else if (localEvent.event == FS_THREAD_EVENT_YIELD_SLEEP_WAKEUP)
    {
        // Let this one get back to our caller.
    }

    if (event != NULL)
    {
        *event = localEvent;
    }

    // Accumulate yield time.
    if (context->reqHdr)
    {
		uint32      requestType;

		eee_rtcDuration(&tvYieldStart, &elapsed);
		eee_rtc(&context->tmpTime);
        if (context->reqHdr->requestType < FS_REQ_NFS_MAX)
			requestType = context->reqHdr->requestType;
		else if (context->reqHdr->requestType == FS_REQ_DUMP_START)
			requestType = FS_REQ_NFS_MAX;
        else if (context->reqHdr->requestType == FS_REQ_REST_START)
			requestType = FS_REQ_NFS_MAX + 1;
		else if (context->daemonType != FS_DAEMON_TYPE_INVALID)
			requestType = context->daemonType + 2 + FS_REQ_NFS_MAX;
		else {
			// An unused value.
			requestType = 0xFFFFFFFF;
		}

		if (requestType != 0xFFFFFFFF) {
			NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
			eee_rtcIncr(&(context->fileSystem->yieldStats.num[requestType][type]),
						&elapsed);
			NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);

			NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
			eee_rtcIncr(&(fs_yieldStats.num[requestType][type]), &elapsed);
			NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
		}
	}
}


/**
 * Yield the processor for the specified period.
 */
extern void
fs_sleep(fs_Context *   context,
         int32          msecs)
{
    fs_yield(context, NULL, msecs, FS_YIELD_TYPE_SLEEP);
}


/**
 * Send the error response for an NDMP request when
 * we failed to create a context for the request.
 */
void
fs_nullContextSendNdmpResponse(
    const fs_MsgReqHdr  *reqHdr,
    int32               contextFlags,
    uint32              srcAddr,
    fs_Status           status)
{

    fs_MsgNdmpResp_t    msgBuf;
    fs_ReqType          msgType = reqHdr->requestType;
    int                 rc;
    uint32               cookie;
#if defined(FS_STANDALONE)
    uint64              clientPrivate = reqHdr->clientPrivate;
    int32               cpuId;
    char                *resp;
#endif

    switch (msgType) {

    case FS_REQ_DUMP_SETUP:
        cookie = ((fs_MsgDumpSetupReq_t *)reqHdr)->dump.ndmpSessId;
        break;
    case FS_REQ_DUMP_NAMES:
        cookie = ((fs_MsgDumpNamesReq_t *)reqHdr)->names.ndmpSessId;
        break;
    case FS_REQ_DUMP_START:
        cookie = ((fs_MsgDumpNamesReq_t *)reqHdr)->names.ndmpSessId;
        break;
    case FS_REQ_REST_SETUP:
        cookie = ((fs_MsgRestoreSetupReq_t *)reqHdr)->restore.ndmpSessId;
        break;
    case FS_REQ_REST_NAMES:
        cookie = ((fs_MsgRestoreNamesReq_t *)reqHdr)->names.ndmpSessId;
        break;
    case FS_REQ_REST_START:
        cookie = ((fs_MsgRestoreStartReq_t *)reqHdr)->start.ndmpSessId;
        break;
    default:
        FS_ASSERT(FALSE);
        return;
    }

    bzero(&msgBuf, sizeof (msgBuf));
    msgBuf.clientCookie = cookie;
    msgBuf.requestType = msgType;
    msgBuf.fsStatus = status;
    if (msgType == FS_REQ_DUMP_SETUP) {
        printf("nullContextNdmpResponse: for dump setup\n");
    }
#if defined(FS_STANDALONE)
        cpuId = (int32) pthread_getspecific(fs_simKeyCpuId);
        resp = (char *)&msgBuf;
/**
        rc = fs_simSendMsg( cpuId,
                            context->reqHdr->src,
                            resp,
                            context->respSize);
**/
#else
        // TODO - check if system buffer is used for response (vs. malloc
        // buffer), use the appropriate send function (this currently
        // assumes a malloc'd buffer is pointed to by respHdr).
        rc = eee_sendMessage(   &msgBuf,
                                sizeof (msgBuf),
                                srcAddr,
                                fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));

		if (contextFlags & FS_CONTEXT_SHARED_REQUEST)
            eee_deallocateBuffer((void *)reqHdr);
        else 
            fs_ramDealloc(FS_RAM_REQ_MSG, (void *)reqHdr);
#endif

}

EXPORTED int
fs_sendNdmpResponse(
    fs_Context          *context,
    fs_ReqType          msgType,
    uint32               cookie,
    uint64              clientPrivate,
    fs_Status           status)
{

    fs_MsgNdmpResp_t    msgBuf;
    int                 rc;
#if defined(FS_STANDALONE)
    int32               cpuId;
    char                *resp;
#endif

    bzero(&msgBuf, sizeof (msgBuf));
    msgBuf.clientCookie = cookie;

    if (msgType == FS_REQ_DUMP_SETUP) {
        printf("sendNdmpResponse: for dump setup\n");
    }
#if 0
    msgBuf.hdr.requestType = msgType;
    msgBuf.hdr.fsStatus = status;
    msgBuf.hdr.clientPrivate = clientPrivate;
#endif
    msgBuf.clientCookie = cookie;
    msgBuf.requestType = msgType;
    msgBuf.fsStatus = status;
#if defined(FS_STANDALONE)
        cpuId = (int32) pthread_getspecific(fs_simKeyCpuId);
        resp = (char *)&msgBuf;
        rc = NFX_OK;
/**
        rc = fs_simSendMsg( cpuId,
                            context->reqHdr->src,
                            resp,
                            context->respSize);
**/
#else
        // TODO - check if system buffer is used for response (vs. malloc
        // buffer), use the appropriate send function (this currently
        // assumes a malloc'd buffer is pointed to by respHdr).
        rc = eee_sendMessage(   &msgBuf,
                                sizeof (msgBuf),
                                context->srcAddr,
                                fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));
#endif
    return rc;
}

#endif // SSC
