/*
 * $Id: fs-dir.c,v 1.7 2002/12/07 01:30:49 lanc Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 */

/*
 * Flat directory code.
 *
 * Directories as specified in this file take two forms: a compact form
 * to fit on a small (1k) page, and a long form to fit on regular 8k
 * filesystem buffers.
 *
 * Both forms are linear search, insert-anywhere flat directories.
 * 
 * The basic directory entry data structures are:
 * 
 * // 40 byte long-form directory entry.
 * 
 * typedef struct fs_Dirent {
 * 	   uint16 len;
 *     uint16 aliasSeq;        // 8.3 ~sequence number
 *     uint32 aliasCksumRand;  // 8.3 cksum or random num
 * 	   uint64 inum;
 * 	   char name[16];
 *     uint32 hash;
 *     uint16 spare;
 * 	   uint16 cont;
 * } fs_Dirent;
 * 
 * typedef struct fs_Dircont {
 * 	   uint16 len;
 * 	   char name[36];
 * 	   uint16 cont;
 * } fs_Dircont;
 * 
 * 
 * // 32 byte short-form directory entry.
 * 
 * typedef struct fs_DirentCompact {
 *     uint16 len;
 *     uint16 aliasSeq;        // 8.3 ~sequence number
 *     uint32 aliasCksumRand;  // 8.3 cksum or random num
 *     uint64 inum;
 *     char name[16];
 * } fs_DirentCompact;
 *
 * The short form is very straightforward: no name may exceed the
 * 16-byte name length, and an entry is valid if its "len" field
 * is nonzero.
 *
 * The long form is somewhat more complicated: An entry contains
 * a valid inum, etc., if its length field is nonzero. However,
 * a zero length field may denote a free entry or a continuation
 * entry. Both free entries and continuation entries have a linked
 * list pointer (the "cont" field) which strings them up either
 * into the freelist (d->freelist) or the list formed by a long
 * name.
 *
 * !!!!!!!! NOTE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * !!                                                                !!
 * !!  The fields within the 3 structures are positioned             !!
 * !!  such that we can do some dirty casting between them.          !!
 * !!  The compact form overlays the long form completely, and       !!
 * !!  the len and cont fields overlay eachother between the         !!
 * !!  long and continuation.  Don't goof with this!                 !!
 * !!                                                                !!
 * !!!!!!!! NOTE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 *  
 */

#include "fs-includes.h"


//#define CHECK_PAGE(c,i,d)  check_page(c,i,d)
#define CHECK_PAGE(c,i,d)

#define READDIR_TRACE(args) FS_TRACE(context, FS_TRACE_READDIR, args)

static void fs_dirpage_init(fs_Dirleaf *d);
static void fs_dir_removename(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i);

#define DIR_IS_COMPACT(_inode)                                  \
    ((_inode)->dinode.attrs & FS_IFLG_COMPACT_DIR)

#define TOUCH_ELEM(buf, p) dirbuf_touch(context, buf, p, sizeof *(p))
#define TOUCH(buf, p, len) dirbuf_touch(context, buf, p, len)

// Don't make a dnlc entry if the flags would prevent
// the entry from being visible to a cifs client.
#define NO_DNLC_ENTRY \
    (FS_DIR_ALIAS_SEQ_FLG_CASE | FS_DIR_ALIAS_SEQ_FLG_ILCH | FS_DIR_ALIAS_SEQ_FLG_DUPE)

static
char convertToUpper [256] =
{
      0,   1,   2,   3,   4,   5,   6,   7,   8,   9, 
     10,  11,  12,  13,  14,  15,  16,  17,  18,  19, 
     20,  21,  22,  23,  24,  25,  26,  27,  28,  29, 
     30,  31,  32,  33,  34,  35,  36,  37,  38,  39, 
     40,  41,  42,  43,  44,  45,  46,  47,  48,  49, 
     50,  51,  52,  53,  54,  55,  56,  57,  58,  59, 
     60,  61,  62,  63,  64,  65,  66,  67,  68,  69, 
     70,  71,  72,  73,  74,  75,  76,  77,  78,  79, 
     80,  81,  82,  83,  84,  85,  86,  87,  88,  89, 
     90,  91,  92,  93,  94,  95,  96, 'A', 'B', 'C', 
    'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
    'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
    'X', 'Y', 'Z', 123, 124, 125, 126, 127, 128, 129, 
    130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 
    140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 
    150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 
    160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 
    170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 
    180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 
    190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 
    200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 
    210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 
    220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 
    230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 
    240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 
    250, 251, 252, 253, 254, 255 
};

static int fsTest83 = 0;
static int fsTestCase = 0;


/**
 * Table of ascii chars that are legal in a cifs filename.
 * This table is generated with the following code:
 *

    for (i = 0; i < 256; i++)
    {
        if (i <= 0x20) // control chars and <space>
        {
            printf("0, ");
        }
        else if (i >= 0x7f) // DEL and extended ascii chars
        {
            printf("0, ");
        }
        else switch(i)
        {
            case '"':
            case '*':
            case '/':
            case '<':
            case '>':
            case '?':
            case '\\':
            case '|':
            case '+':
            case ',':
            case ':':
            case ';':
            case '=':
            case '[':
            case ']':
                printf("0, ");
                break;
            default:
                if (i == '\'')
                    printf("'\\%c',", i);
                else
                    printf("'%c',", i);
        }

        if ((i+1) % 16 == 0)
        {
            printf("\n    ");
        }
    }
 *
 */
static
char cifsLegalChar[256] =
{
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, '!',0, '#','$','%','&','\'','(',')',0, 0, 0, '-','.',0, 
    '0','1','2','3','4','5','6','7','8','9',0, 0, 0, 0, 0, 0, 
    '@','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O',
    'P','Q','R','S','T','U','V','W','X','Y','Z',0, 0, 0, '^','_',
    '`','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
    'p','q','r','s','t','u','v','w','x','y','z','{',0, '}','~',0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
};


static inline void
dirbuf_touch(fs_Context *context, fs_Buf *buf, void *p0, int len)
{ 
        char *p = p0;
        /* range check with unsigned compare! */
        FS_ASSERT((unsigned) (p - buf->u.data) < PAGE_SIZE);
        FS_ASSERT((unsigned) (p + len - buf->u.data) <= PAGE_SIZE);
        FS_BUF_WRITE(buf, p - buf->u.data, len);
}


/*
 * Pop a free entry from the freelist in a directory page.
 */
static int
fs_dir_popent(fs_Context *context, fs_Buf *buf)
{
	fs_Dirleaf *d = buf->u.dirleaf;
	int free;

	FS_ASSERT(d->freelist < FS_DIRLEAF_ENTRIES_MAX);
	free = d->freelist;
	d->freelist = d->u[free].ent.cont;
	d->u[free].ent.cont = FS_DIRCONT_XXX;

	d->count++;

	TOUCH_ELEM(buf, &d->freelist);
	TOUCH_ELEM(buf, &d->count);

	return free;
}

/*
 * Free the i'th directory entry and push it on the free list.
 */
static void
fs_dir_pushent(fs_Context *context, fs_Buf *buf, int i)
{
	fs_Dirleaf *d = buf->u.dirleaf;

	FS_ASSERT(i < FS_DIRLEAF_ENTRIES_MAX);

	d->u[i].ent.len = 0;
	d->u[i].ent.cont = d->freelist;
	d->freelist = i;
	--d->count;

	TOUCH_ELEM(buf, &d->u[i].ent);
	TOUCH_ELEM(buf, &d->freelist);
	TOUCH_ELEM(buf, &d->count);
}


/**
 * Set the alias sequence number for the specified
 * dirent.
 */
static inline void
fs_dirSetAliasSeq(fs_Dirent *dirent, int num)
{
    FS_ASSERT(num <= FS_DIR_ALIAS_SEQ_RAND);

    dirent->aliasSeq &= ~FS_DIR_ALIAS_SEQ_MASK;
    dirent->aliasSeq |= num;
}


/**
 * Get the alias sequence number for the specified
 * dirent.
 */
static inline int
fs_dirGetAliasSeq(fs_Dirent *dirent)
{
    return (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_MASK);
}


/**
 * Returns the validity of a char for 8.3 filename
 * compatibility.
 */
enum {
    CHAR_VALID,
    CHAR_NOT_VALID,
    CHAR_IS_DOT
};
static inline int
fs_dirChar83Checker(char c)
{
    if (cifsLegalChar[(int)c] == 0)
        return CHAR_NOT_VALID;

    if (c == '.')
        return CHAR_IS_DOT;

    return CHAR_VALID;
}


/**
 * Returns TRUE if the name is an 8.3 compatible
 * filename.  Returns TRUE or FALSE in hasIllChar
 * if name contains illegal characters.
 */
static boolean
fs_dirNameIs83Compat(char *name, int namelen, boolean *hasIllChar)
{
    int valid, i, numDots = 0, extLen = 0;
    boolean rc = TRUE;

    // TODO - this will need to check longname for translatability
    // within the code set defined for this filesystem.


    // Scan the entire name for illegal characters
    // so we can set hasIllChar correctly.
    if (hasIllChar != NULL)
        *hasIllChar = FALSE;
    for (i = 0; i < namelen; i++)
    {
        valid = fs_dirChar83Checker(name[i]);
        if (valid == CHAR_NOT_VALID)
        {
            //printf("name %s NOT 8.3 compat (bad char %c)\n", name, name[i]);
            if (hasIllChar != NULL)
                *hasIllChar = TRUE;
            return FALSE;
        }

        if (valid == CHAR_IS_DOT)
        {
            numDots += 1;
            if (numDots > 1)
            {
                //printf("name %s NOT 8.3 compat (too many dots)\n", name);
                rc = FALSE;
            }
        }
        else if (numDots)
        {
            extLen += 1;
            if (extLen > 3)
            {
                //printf("name %s NOT 8.3 compat (extension too long)\n", name);
                rc = FALSE;
            }
        }
        if (i == 8 &&
            numDots == 0)
        {
                //printf("name %s NOT 8.3 compat (base too long)\n", name);
            rc = FALSE;
        }

        if (rc == FALSE &&
            hasIllChar == NULL)
        {
            break;
        }
    }

    if (namelen > 12)
    {
        //printf("name %s NOT 8.3 compat (too long)\n", name);
        return FALSE;
    }

    //if (rc == TRUE)
    //      printf("name %s IS 8.3 compat\n", name);

    return rc;
}


/**
 * Returns TRUE if the name could be an 8.3 compatible
 * name generated by this system in random number format.
 */
static boolean
fs_dirNameIsRandom(char *name, int namelen, uint32 *randomNum)
{
    FS_ASSERT(fs_dirNameIs83Compat(name, namelen, NULL));

    if (namelen >= 8 &&
        name[0] == '~')
    {
        int i;
        char tmpName[8];

        for (i = 1; i < 8; i++)
        {
            if (!isxdigit(name[i]))
                return FALSE;
            tmpName[i-1] = name[i];
        }

        if (randomNum != NULL)
        {
            tmpName[7] = '\0';
            *randomNum = atolx(tmpName);
        }
        return TRUE;
    }
       
    return FALSE;
}


/**
 * Returns TRUE if the name is formatted as a 
 * generated 8.3 name and returns sequence
 * number in seq.
 */
static boolean
fs_dirNameIs83Generated(char *name, int namelen, int *seq)
{
    char    *p;
    char    digits[FS_DIR_ALIAS_SEQ_CHARS_MAX + 1];
    int     i;
    boolean rc = FALSE;

    FS_ASSERT(fs_dirNameIs83Compat(name, namelen, NULL));

    if ((p = strnchr(name, '.', namelen)) != NULL)
    {
        p -= 1;
    }
    else
    {
        p = name + namelen - 1;
    }

    // p should point to the last digit of
    // the sequence number now.

    for (i = 0;
         i < FS_DIR_ALIAS_SEQ_CHARS_MAX && p >= name;
         i++, p--)
    {
        if (*p == '~')
            break;

        if (!isdigit(*p))
            break;

        digits[i] = *p;
    }

    if (p >= name && i > 0 && *p == '~')
    {
        char    digitsOrdered[FS_DIR_ALIAS_SEQ_CHARS_MAX + 1];
        int     x, y = i - 1;

        // It conforms, get the sequence number.
        // First we have to re-order the digits, they're
        // reversed in digits[].
        for (x = 0; x <= y; x++, i--)
            digitsOrdered[x] = digits[i - 1];
        digitsOrdered[x] = '\0';
        *seq = atoi(digitsOrdered);

        if (*seq > 0)
        {
            rc = TRUE;
            //printf("%s name IS generated format, seq %d\n", name, *seq);
        }
    }

    //if (rc != TRUE)
    //{
    //  printf("%s name NOT generated format seq %d\n", name, *seq);
    //}

    return rc;
}


/**
 * Construct the 8.3 base component (ie. up to the
 * tilde) of the specified long filename up the max
 * char in the result.
 *
 * Returns the number of characters in the result.
 */
static int
fs_dirConstruct83Base(fs_Context *context, uchar8 *longname, int namelen, char *base, int max)
{
    int i, rc = 0;
    uchar8 *p;

    p = strnrchr(longname, '.', namelen);
    if (p == NULL)
    {
        p = longname + namelen;
    }

    for (i = 0;
         i < namelen && rc < max && &longname[i] < p;
         i++)
    {
        if (fs_dirChar83Checker(longname[i]) == CHAR_VALID)
        {
            if (fsTest83 || context->use83Names)
                base[rc++] = convertToUpper[longname[i]];
            else
                base[rc++] = longname[i];
        }
    }

    return rc;
}


/**
 * Construct the .3 component (including the dot)
 * of the specified long filename.  Returns the number
 * of characters in the result.
 */
static int
fs_dirConstruct83Extension(fs_Context *context, char *longname, int namelen, char *ext)
{
    char    *p;
    int     i, rc = 0;
    int     numTrunc;

    //printf("reconext: longname %p, %s len %d\n", longname, longname, namelen);

    // Firstly, truncate all dots at the end
    // of the name.
    for (i = namelen - 1, numTrunc = 0;
         i >= 0 && longname[i] == '.';
         i--)
    {
        numTrunc += 1;
        longname[i] = '\0';
        //printf("reconext: kill dot at index %d\n", i);
    }

    // Search for the last dot.
    p = strnrchr(longname, '.', namelen);
    if (p != NULL)
    {
        char  backwards[3];
        char  *endp = longname + namelen - numTrunc - 1;
        int   num;

        //printf("reconext: dot at %p\n", p);
        // There's still a dot somewhere in it,
        // return it and up to 3 valid chars from
        // the end of the longname.  We use the end
        // characters to produce better hashes for
        // names that have counters as their extension
        // (No, this isn't the Microsoft way, but the
        // rules here are kinda loose).

        num = 0;
        while (endp > p &&
               num < 3)
        {
            if (fs_dirChar83Checker(*endp) == CHAR_VALID)
            {
                if (fsTest83 || context->use83Names)
                    backwards[num++] = convertToUpper[(int) *endp];
                else
                    backwards[num++] = *endp;
                //printf("reconext: char %c at %p\n", *p, p);
            }
            endp -= 1;
        }

        if (num > 0)
        {
            ext[0] = '.';
            rc = 1;  // count the dot
        }

        while (num > 0)
        {
            ext[rc++] = backwards[--num];
        }
    }

    //printf("reconext: numTrunc %d\n", numTrunc);

    // Replace any dots truncated.
    for (i = 0; i < numTrunc; i++)
    {
        //printf("reconext: char %c at %p\n", *p, p);
        longname[namelen - 1 - i] = '.';
    }

    //printf("reconext: rc %d\n", rc);

    return rc;
}


/*
 * Copy a name from a dirleaf (the ith entry) into the memory pointed at dst.
 */
static void
unpack_name(char *dst, fs_Dirleaf *d, fs_Dirent *dirent)
{
	int len = dirent->len;
	fs_Dircont *cont;
	
	dst[len] = '\0';

	memcpy(dst, dirent->name, MIN(len, sizeof dirent->name));

	if (len <= sizeof dirent->name)
		return;

	dst += sizeof dirent->name;
	len -= sizeof dirent->name;
	cont = &d->u[dirent->cont].cont;

	for (;;) {
		int block = MIN(len, sizeof cont->name);
		memcpy(dst, cont->name, block);
		dst += block;
		len -= block;
		if (len == 0)
			break;
		FS_ASSERT(cont->cont < FS_DIRLEAF_ENTRIES_MAX);
		cont = &d->u[cont->cont].cont;
	}
}


/**
 * Generate the hash for the specified name and type.
 */
static uint32
fs_dir_hashname(char *name, int len, int nameType)
{
    char    *p;
    int     i, skipIndex;
    int     skip[16], remain;
    boolean foundTilde = FALSE;
	uint32  hash;

    switch(nameType)
    {
    case DIR_NAME_TYPE_RAND: {
        char tmpName[8];
        FS_ASSERT(name[0] == '~');
        FS_ASSERT(len >= 8);
        memcpy(tmpName, &name[1], 7);
        tmpName[7] = '\0';
        hash = atolx(tmpName);
        // printf("hashname: %s rand 0x%x\n", name, hash);
        return hash;
    }

    case DIR_NAME_TYPE_SEQ:
        bzero(skip, sizeof skip);
        // Set p to point to the last digit of
        // the sequence number.
        if ((p = strnchr(name, '.', len)) != NULL)
        {
            p -= 1;
        }
        else
        {
            p = name + len - 1;
        }
    
        // Don't compute cksum over the ~ and the
        // maximal sequence number.
        skipIndex = p - name;
        remain = (p - name) + 1;
        for (i = 0;
             i < FS_DIR_ALIAS_SEQ_CHARS_MAX + 1 && remain > 0;
             i++, p--)
        {
            skip[skipIndex - i] = 1;
            remain--;
    
            if (*p == '~')
                foundTilde = TRUE;
    
            if (foundTilde && remain <= 8 - 1 - FS_DIR_ALIAS_SEQ_CHARS_MAX)
                break;
        }
        /* FALLTHROUGH */
 
    case DIR_NAME_TYPE_LONG:
    default:
       /*
        * Compute a 32-bit hash on a name in a case-insensitive fashion.
        * TODO: case folding for UTF-8.
        */
	    hash = ~0;
	    for (i = 0; i < len; i++)
        {
            if (nameType == DIR_NAME_TYPE_LONG || !skip[i])
                hash = crc32[(hash ^ convertToUpper[(int)name[i]]) & 0xff] ^ (hash >> 8);
        }
        break;
    }


    //printf("hashname: type %d name %s len %d hash 0x%x\n", nameType, name, len, hash);

    return hash;
}


/**
 * Return the reconstructed 8.3 alias for the specified
 * dirent.
 */
static void
fs_dirReconstruct83Name(fs_Context *context, fs_Dirleaf *d, fs_Dirent *dirent, char *dest, int *namelen)
{
    int     seq, len;
    char    *namep = dirent->name;
    char    unpacked[FS_NAME_MAX + 1];

    if (dirent->len > sizeof dirent->name) {
        unpack_name(unpacked, d, dirent);
        namep = unpacked;
    }

    seq = fs_dirGetAliasSeq(dirent);
    if (seq == FS_DIR_ALIAS_SEQ_RAND)
    {
        len = sprintf(dest, "~%07X", dirent->aliasCksumRand);
        len += fs_dirConstruct83Extension(context, namep, dirent->len,
                                               &dest[len]);
    }
    else
    {
        // Get length of sequence number first.
        len = sprintf(&dest[0], "%d", seq);

        len = fs_dirConstruct83Base(context, namep, dirent->len,
                                    dest, 8-1-len);
        dest[len] = '~';
        len++;
        len += sprintf(&dest[len], "%d", seq);

        len += fs_dirConstruct83Extension(context, namep, dirent->len,
                                          &dest[len]);
    }

    FS_ASSERT(len <= 12);
    dest[len] = '\0';

    *namelen = len;
}


/*
 * Compare a name against a directory entry which might be split amongst
 * any number of continuation structures. The comparisons in the first
 * dirent structure are special-cased because its array is slightly
 * shorter than the ones in the continuations.
 */
int
fs_dir_namecmp(fs_Context *context, fs_Inode *inode, fs_Dirleaf *d,
               fs_Dirent *dirent, fs_DirSearchCb *cb)
{
	fs_Dircont *dircont;
	int cont, i, r;
    int (*cmp_fun)(char *, char*, int);

    //printf("namecmp %s,%d/%s,%d type %d hash 0x%x\n",
    //    cb->name, cb->len, dirent->name, dirent->len, cb->nameType, cb->hash);

    switch(cb->nameType)
    {
    case DIR_NAME_TYPE_SEQ:
        if (cb->hash == dirent->aliasCksumRand)
        {
            cb->hashMatch = TRUE;
            if (fs_dirGetAliasSeq(dirent) == cb->seq)
            {
                char    dirAlias[16];
                int     dirAliasLen;
    
                fs_dirReconstruct83Name(context, d, dirent, dirAlias, &dirAliasLen);
    
                if (cb->len == dirAliasLen &&
                    strncasecmp(dirAlias, cb->name, cb->len) == 0)
                {
                    //printf("namecmp seq match\n");
                    cb->aliasFlags = (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_MASK);
                    return 0;
                } 
                //printf("namecmp len or name not match cblen %d aliaslen %d alias %s\n",
                //  cb->len, dirAliasLen, dirAlias);
            }
        }
        //printf("namecmp seq not match (hash 0x%x seq %d cb.seq %d)\n",
        //    dirent->aliasCksumRand, fs_dirGetAliasSeq(dirent), cb->seq);
        return -1;
    case DIR_NAME_TYPE_RAND:
        if (cb->hash == dirent->aliasCksumRand)
        {
            cb->hashMatch = TRUE;
            if (fs_dirGetAliasSeq(dirent) == FS_DIR_ALIAS_SEQ_RAND) {
                //printf("namecmp rand match\n");
                cb->aliasFlags = (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_MASK);
                return 0;
            }
        }
        //printf("namecmp rand no match\n");
        return -1;
    case DIR_NAME_TYPE_LONG:
        // Just doing the length check would be quicker for
        // negative responses, and for a compact it seems silly
        // to compute and compare the hash instead of just
        // comparing the names, but we MUST return hashMatch
        // status in order for alias creation logic to make the
        // correct decision.
        if (DIR_IS_COMPACT(inode)) {
            if (cb->hash == fs_dir_hashname(dirent->name, dirent->len, DIR_NAME_TYPE_LONG))
                cb->hashMatch = TRUE;
            else
                return -1;
        }
        else {
            if (cb->hash == dirent->hash)
                cb->hashMatch = TRUE;
            else
                return -1;
        }

        if (cb->len != dirent->len) {
            //printf("namecmp long wrong len\n");
            return -1;
        }

        /* FALLTHROUGH */
    }

    if (fsTest83 || fsTestCase || context->caseInsensitive) {
        cmp_fun = (void *) strncasecmp;
        //printf("namecmp doing strncasecmp\n");
    }
    else {
        cmp_fun = (void *) memcmp;
        //printf("namecmp doing memcmp\n");
    }

	/*
	 * Does the name fit in the first dirent?
	 */
	if (cb->len <= sizeof dirent->name) {
        r = cmp_fun(dirent->name, cb->name, cb->len);
        if (r == 0)
            cb->aliasFlags = (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_MASK);
        return r;
    }

	/*
	 * Does the name match up to the first dirent bytes?
	 */
	if ((r = cmp_fun(dirent->name, cb->name, sizeof dirent->name)) != 0) {
        if (r == 0)
            cb->aliasFlags = (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_MASK);
		return r;
    }

	/*
	 * Now proceed with continuations. "block" is either a full continuation
	 * worth of bytes, or the remainder of the bytes at the end of the filename.
	 */
	i = sizeof dirent->name;
	cont = dirent->cont;
	for (;;) {
		int block = MIN(sizeof dircont->name, cb->len - i);
		dircont = &d->u[cont].cont;
		cont = dircont->cont;
		if ((r = cmp_fun(dircont->name, &cb->name[i], block)) != 0) {
            if (r == 0)
                cb->aliasFlags = (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_MASK);
			return r;
        }
		i += block;
		if (i == cb->len) {
            if (r == 0)
                cb->aliasFlags = (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_MASK);
			return 0;
        }
	}

	/*NOTREACHED*/
	return 0;
}


/**
 * Return 0 if the hash matches the specified nameType
 * hash in the dirent.
 */
int
fs_dir_hashcmp(fs_Dirent *dirent, uint32 hash, int nameType)
{
    //printf("hashcmp hash 0x%x type %d\n", hash, nameType);

    switch(nameType)
    {
    case DIR_NAME_TYPE_LONG:
        //printf("dirent->hash 0x%x\n", dirent->hash);
        if (dirent->hash == hash)
            return 0;
        break;
    case DIR_NAME_TYPE_SEQ:
    case DIR_NAME_TYPE_RAND:
        //printf("dirent->aliascksumrand 0x%x\n", dirent->aliasCksumRand);
        if (dirent->aliasCksumRand == hash)
            return 0;
        break;
    }

    return -1;
}


/*
 * Iterate over the directory visiting each
 * entry with a matching hash with iter_fun().
 */
static void
dir_iterate(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_DirSearchCb  *cb,
    void            (*iter_fun)(
            fs_Inode    *inode,
            fs_Dirent   *dirent,
            fs_DirSearchCb  *cb))
{
	fs_Buf *buf;
    int i;

    if (DIR_IS_COMPACT(inode))
    {
        fs_DirleafCompact *dc;

        // Compact form has no hash for the long name,
        // so this function cannot be used for such
        // comparisons.
        FS_ASSERT(cb->nameType != DIR_NAME_TYPE_LONG);

        buf = O_GET_SHARED(inode, 0);
	    dc = buf->u.dirleafCompact;
        cb->page = 0;

	    for (i = 0; i < FS_DIRLEAF_COMPACT_ENTRIES_MAX; i++)
        {
		    if (dc->entries[i].len == 0)
			    continue;
            if (fs_dir_hashcmp((fs_Dirent *) &dc->entries[i], cb->hash, cb->nameType) == 0)
            {
                cb->index = i;
                iter_fun(inode, (fs_Dirent *) &dc->entries[i], cb);
            }
	    }
	    fs_bufRelease(context, buf);
    }
    else
    {
        fs_Dirleaf *d;
        uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
        uint64 page;
        boolean dirhashDone;

        // Try to iterate using the dirhash.
        //
        // WARNING!!! - the nature of the dirhash makes it
        // difficult to guarantee that we visit each entry
        // only once (thus, we don't even try) - so the caller
        // must be prepared to handle this!

        dirhashDone = fs_dirhash_iterate(context, inode, cb, iter_fun);
        if (dirhashDone)
            return;

        // Must do a slow linear iterate.
	    for (page = 0; page < pagecount; page++)
        {
            int i;
		    buf = O_GET_SHARED(inode, page);
		    d = buf->u.dirleaf;
    
		    if (page == 0)
			    FS_ASSERT(d->magic == DIRLEAF_LONG_MAGIC);
    
            cb->page = page;
		    for (i = 0; i < FS_DIRLEAF_ENTRIES_MAX; i++)
            {
			    if (d->u[i].ent.len == 0)
				    continue;
                if (fs_dir_hashcmp(&d->u[i].ent, cb->hash, cb->nameType) == 0)
                {
                    cb->index = i;
                    iter_fun(inode, &d->u[i].ent, cb);
                }
		    }
		    fs_bufRelease(context, buf);
	    }
    }
}


/*
 * How many name chunks does it take to fit this name in a directory page?
 */
static int
name_chunks(int len)
{
	fs_Dirent *dirent;
	fs_Dircont *dircont;

	len -= sizeof dirent->name;

	if (len <= 0)
		return 1;

	return 1 + ceilingDiv64(len, sizeof dircont->name);
}

/*
 * Pack a name into a dirleaf structure splitting it across continuations.
 * Assumes there is room in the leaf.
 */
static void
pack_name(fs_Context *context, fs_Buf *buf, fs_Dirent *dirent, int chunks, char *name, int len)
{
	fs_Dirleaf *d = buf->u.dirleaf;
	fs_Dircont *dircont;
	int i, p;

	// printf("pack name: %s %d chunks\n", name, chunks);

	/*
	 * Copy the first part of the name into the dirent.
	 */
	memcpy(dirent->name, name, MIN(len, sizeof dirent->name));

	/*
	 * Does the name fit in the entry itself?
	 */
	if (chunks == 1)
		return;

	/*
	 * Otherwise allocate continuations for the rest of the name.
	 */
	--chunks;
	p = sizeof dirent->name;
	for (i = 0; i < chunks; i++) {
		int block = MIN(sizeof dircont->name, len - p);

		dirent->cont = fs_dir_popent(context, buf);
		TOUCH(buf, &dirent->cont, sizeof dirent->cont);
		dirent = &d->u[dirent->cont].ent;
		dirent->len = 0; // sentinel value, is skipped by readdir
		dirent->cont = FS_DIRCONT_EOL;
		memcpy(((fs_Dircont *)dirent)->name, &name[p], block);
		p += block;
		TOUCH_ELEM(buf, dirent);
	}
	FS_ASSERT(len - p <= sizeof dircont->name);
}


/*
 * Debug asserts for page consistency
 */
void
check_page(fs_Context *context, fs_Inode *inode, fs_Dirleaf *d)
{
	enum {BOGUS, ALLOC, FREE, CONT};
	char index[FS_DIRLEAF_ENTRIES_MAX];
	int alloc, free, cont;
	int i;

	bzero(index, sizeof index);

	/*
	 * Find allocated slots.
	 */
	alloc = 0;
	for (i = 0; i < sizeof index; i++) {
		if (d->u[i].ent.len != 0) {

            char unpacked[FS_NAME_MAX];
            fs_DirSearchCb cb;
            int32 rc;

			index[i] = ALLOC;
			alloc++;

            unpack_name(unpacked, d, &d->u[i].ent);
            cb.inum = inode->inum;
            cb.name = unpacked;
            cb.len = d->u[i].ent.len;
            cb.nameType = DIR_NAME_TYPE_LONG;
            cb.hash = fs_dir_hashname(cb.name, cb.len, cb.nameType);
            FS_ASSERT(cb.hash == d->u[i].ent.hash);
            rc = fs_dirhash_lookup(context, inode, &cb);
            FS_ASSERT(rc == FS_DIRHASH_FOUND || rc == FS_DIRHASH_UNKNOWN);

            if (d->u[i].ent.aliasCksumRand != 0) {
                char alias[16];
                int aliasLen;

                fs_dirReconstruct83Name(context, d,
                                        &d->u[i].ent,
                                        alias, &aliasLen);

                cb.inum = inode->inum;
                cb.name = alias;
                cb.len = aliasLen;
                cb.seq = fs_dirGetAliasSeq(&d->u[i].ent);
            
                if (cb.seq == FS_DIR_ALIAS_SEQ_RAND)
                    cb.nameType = DIR_NAME_TYPE_RAND;
                else
                    cb.nameType = DIR_NAME_TYPE_SEQ;
                cb.hash = fs_dir_hashname(cb.name, cb.len, cb.nameType);
                FS_ASSERT(cb.hash == d->u[i].ent.aliasCksumRand);
                rc = fs_dirhash_lookup(context, inode, &cb);
                FS_ASSERT(rc == FS_DIRHASH_FOUND || rc == FS_DIRHASH_UNKNOWN);
            }
		}
	}

	/*
	 * Find free slots.
	 */
	free = 0;
	for (i = d->freelist; i < FS_DIRLEAF_ENTRIES_MAX; i = d->u[i].cont.cont) {
		FS_ASSERT(index[i] == 0);
		index[i] = FREE;
		free++;
	}

	FS_ASSERT(d->count + free == FS_DIRLEAF_ENTRIES_MAX);

	/*
	 * If there are any free slots, the freelist pointer should point
	 * to them.
	 */
    FS_ASSERT(d->freelist == FS_DIRFREE_EOL || index[d->freelist] == FREE);
	FS_ASSERT(d->count == FS_DIRLEAF_ENTRIES_MAX || d->freelist < FS_DIRLEAF_ENTRIES_MAX);


	/*
	 * Find continuation slots.
	 */
	cont = 0;
	for (i = 0; i < sizeof index; i++) {
		int len = (int)d->u[i].ent.len - sizeof d->u[i].ent.name;
		if (len > 0) {
			int j = i;
			while (1) {
				j = d->u[j].cont.cont;
				FS_ASSERT(len > 0);
				FS_ASSERT(index[j] == 0);
				index[j] = CONT;
				cont++;
				if (d->u[j].cont.cont == FS_DIRCONT_EOL) {
					FS_ASSERT(len <= sizeof d->u[i].cont.name);
					break;
				}
				len -= sizeof d->u[i].cont.name;
			}
		}
	}

	FS_ASSERT(alloc + cont == d->count);

	for (i = 0; i < sizeof index; i++)
		FS_ASSERT(index[i] != BOGUS);


}

/*
 * Return TRUE if a broken dirleaf is disposed of by the user.
 */
static boolean
query_dispose(fs_Context *context, fs_Inode *inode, fs_Buf *buf, boolean fixIt, char *fmt, ...)
{
	va_list ap;
	char msg[256];

	va_start(ap, fmt);
	vsprintf(msg, fmt, ap);
	va_end(ap);

	context->eek.errcount++;

	if (fixIt) {
		if (!fs_eek_query(context, "inode %lld dirleaf %lld: %s.\nDispose?",
		                  buf->inum, fs_blockNumber(buf->block), msg))
			return FALSE;
		if (DIR_IS_COMPACT(inode)) {
		} else {
			fs_dirpage_init(buf->u.dirleaf);
			dirbuf_touch(context, buf, buf->u.dirleaf, sizeof *buf->u.dirleaf);
		}
		return TRUE;
	}

	fs_eek_print_status(context, FALSE,
			    "%s: (I#%lld) dirleaf %lld: %s.",
			    context->fileSystem->super->mountName,
			    buf->inum, fs_blockNumber(buf->block), msg);
	return FALSE;
}

/*
 * Return TRUE if a broken dirleaf is fixed by the user.
 */
static boolean
query_fix(fs_Context *context, boolean fixIt, char *fmt, ...)
{
	va_list ap;
	char msg[256];

	va_start(ap, fmt);
	vsprintf(msg, fmt, ap);
	va_end(ap);

	context->eek.errcount++;

	if (fixIt) {
		if (fs_eek_query(context, "%s: %s Fix?",
				 context->fileSystem->super->mountName,
				 msg)) return TRUE;
	} else {
		fs_eek_print_status(context, FALSE, "%s: %s",
				    context->fileSystem->super->mountName,
				    msg);
	}

	return FALSE;
}

/*
 * Check an inode number for eek for both compact and long
 * form directory entries.
 */
static boolean
check_inum(fs_Context *context, uint8 *bitmap, fs_Inode *dir, boolean fixIt, uint64 inum, int *linkcount)
{
	fs_Inode *inode;

	FS_LOCK_ORDER_SET_PARENT(dir->inum);
	inode = I_GET_EXCL(dir->snapnum, inum);

	if (inode == NULL) {
       		if (query_fix(context, fixIt,
			"directory %lld: references free inode %lld.",
			dir->inum, inum)) {
			return FALSE;
		}
	} else if (FS_I_IS_DIR(inode->dinode.attrs)) {
		if (!fs_eek_mark(context, "dirchild inode number", bitmap, inum)) {
			if (query_fix(context, fixIt,
			    "directory %lld: invalid directory inode %lld.",
			     dir->inum, inum)) {
				return FALSE;
			}
		} else {
			++*linkcount;
		}
	} else {
		/* lc: huh ?? // byron: this code looks mismatched to the comment!
		 * On the first clash, adjust the
		 * inode link count to 2, and set
		 * this value in the eek link
		 * counts data structure.
		 */
		if (inode->dinode.linkCnt > 1) {
			fs_eek_increment_linkcount(context, dir->inum, inum);
		} else {
			if (!fs_eek_mark(context, "directory inum entry", bitmap, inum)) {
				context->eek.errcount++;
				if (query_fix(context, fixIt,
				    "Invalid directory inum entry.")) {
					++inode->dinode.linkCnt;
					fs_inodeWrite(context, inode, INO_WRITE_LOG);
					fs_eek_increment_linkcount(context, dir->inum, inum);
					fs_eek_increment_linkcount(context, dir->inum, inum);
				}
			}
		}
	}

	if (inode != NULL && !fs_inodeWasModified(context, inode))
		fs_inodeReleaseExclusive(context, inode);

	return TRUE;
}

static fs_Buf *
copy_block_maybe(fs_Context *context, fs_Inode *inode, fs_Buf *buf)
{
	if (fs_blockIsCopyOnWrite(context, buf->block)) {
		uint64 block, oldBlock = fs_blockNumber(buf->block);
		block = fs_make_block_copy(context, inode, oldBlock);
		buf = B_GET_EXCL(inode, fs_makeBlock(block));
	}
	return buf;
}

static boolean
fs_dirleaf_compact_eek_check(
    fs_Context *context,
    fs_Inode *inode,
    fs_Buf *buf,
    boolean fixIt)
{
	fs_DirleafCompact *d = buf->u.dirleafCompact;
	uint64 inum = d->parent_inum;

	if (d->count > FS_DIRLEAF_COMPACT_ENTRIES_MAX) {
		return query_dispose(context, inode, buf, fixIt,
				     "count %d > %d",
				     d->count,
				     FS_DIRLEAF_COMPACT_ENTRIES_MAX);
	}

	if (inum >= context->fileSystem->params.maxInodeNumber) {
		return query_dispose(context, inode, buf, fixIt,
				    "%s: (I#%lld) invalid parent inum %lld",
				    context->fileSystem->super->mountName,
				    inode->inum, inum);
	}

	return TRUE;
}

static boolean
fs_dir_compact_mark_used_inodes(fs_Context *context, uint8 *bitmap, fs_Inode *dir, boolean fixIt)
{
	fs_Buf *buf = O_GET_EXCL(dir, 0);
	fs_DirleafCompact *d = buf->u.dirleafCompact;
	int removes[FS_DIRLEAF_COMPACT_ENTRIES_MAX];
	int linkcount = 2;
	int nremoves;
	int i;

	if (!fs_dirleaf_compact_eek_check(context, dir, buf, fixIt)) {
		if (!fs_bufWasModified(context, buf))
			fs_bufReleaseExclusive(context, buf);
		return FALSE;
	}

	nremoves = 0;
	bzero(removes, sizeof removes);
	for (i = 0; i < FS_DIRLEAF_COMPACT_ENTRIES_MAX; i++) {
		if (d->entries[i].len == 0)
			continue;

		if (!check_inum(context, bitmap, dir, fixIt, d->entries[i].inum, &linkcount))
			removes[nremoves++] = i;
	}

    // fixIt false for walking phase and eek'ing snapshots
    if (!fixIt) {
        fs_bufReleaseExclusive(context, buf);
        return TRUE;
    }

	/*
	 * remove filenames with bad inode numbers
	 */
	if (nremoves > 0) {
		buf = copy_block_maybe(context, dir, buf);
		d = buf->u.dirleafCompact;
	}
	for (i = 0; i < nremoves; i++) {
		d->entries[removes[i]].len = 0;
		--d->count;
		TOUCH_ELEM(buf, &d->entries[removes[i]].len);
	}
	if (nremoves > 0)
		TOUCH_ELEM(buf, &d->count);

	if (linkcount != dir->dinode.linkCnt) {
		char msg[256];
		context->eek.errcount++;
		sprintf(msg, "MISMATCH ON link count for inode %lld, EXPECTED %d, GOT %d.",
		    dir->inum, linkcount, dir->dinode.linkCnt);
        fs_eek_print_status(context, FALSE, "%s", msg);
		if (fixIt && fs_eek_query(context, "Adjust?")) {
			dir->dinode.linkCnt = linkcount;
			fs_inodeWrite(context, dir, INO_WRITE_LOG);
		}
	}

	if (!fs_bufWasModified(context, buf))
		fs_bufReleaseExclusive(context, buf);

	return TRUE;
}


static boolean
fs_dirleaf_eek_check(
    fs_Context *context,
    fs_Inode *inode,
    fs_Buf *buf,
    boolean fixIt)
{
	fs_Dirleaf *d = buf->u.dirleaf;
	enum {BOGUS, ALLOC, FREE, CONT};
	char index[FS_DIRLEAF_ENTRIES_MAX];
	int alloc, free, cont;
	int i, j;

	bzero(index, sizeof index);

	/*
	 * Find allocated slots.
	 */
	alloc = 0;
	for (i = 0; i < sizeof index; i++) {
		if (d->u[i].ent.len != 0) {
			index[i] = ALLOC;
			alloc++;
		}
	}

	/*
	 * If there are any free slots, the freelist pointer should point
	 * to them.
	 */
	if (d->count > FS_DIRLEAF_ENTRIES_MAX)
		return query_dispose(context, inode, buf, fixIt,
				     "count %d > %d",
				     d->count,
				     FS_DIRLEAF_ENTRIES_MAX);

	if (d->count < FS_DIRLEAF_ENTRIES_MAX && d->freelist >= FS_DIRLEAF_ENTRIES_MAX)
		return query_dispose(context, inode, buf, fixIt,
				     "freelist %d >= %d",
				     d->freelist,
				     FS_DIRLEAF_ENTRIES_MAX);

	/*
	 * Find free slots.
	 */
	free = 0;
	for (i = d->freelist; i < FS_DIRLEAF_ENTRIES_MAX; i = d->u[i].cont.cont) {
		if (index[i] != 0)
			break;
		index[i] = FREE;
		free++;
	}

	if (i != FS_DIRLEAF_ENTRIES_MAX && i != FS_DIRFREE_EOL)
		return query_dispose(context, inode, buf, fixIt,
				     "allocated entry %d mapped by free list", i);

	if (d->count + free != FS_DIRLEAF_ENTRIES_MAX)
		return query_dispose(context, inode, buf, fixIt,
				     "count %d + computed free %d != %d",
				     d->count, free,
				     FS_DIRLEAF_ENTRIES_MAX);

	/*
	 * Find continuation slots.
	 */
	cont = 0;
	for (i = 0, j = 0; i < sizeof index; i++) {
		if (d->u[i].ent.len > sizeof d->u[i].ent.name) {
			j = i;
			while (1) {
				j = d->u[j].cont.cont;
				if (j == FS_DIRCONT_EOL)
					break;
				if (index[j] != 0)
					goto out;
				index[j] = CONT;
				cont++;
			}
		}
	}

out:
	if (i != sizeof index)
		return query_dispose(context, inode, buf, fixIt,
				     "entry %d maps continuations to allocated entry %d",
				     i, j);

	if (d->count != alloc + cont)
		return query_dispose(context, inode, buf, fixIt,
				     "computed alloc %d + computed cont %d != %d",
				     alloc, cont, d->count);

	for (i = 0; i < sizeof index; i++)
		if (index[i] == BOGUS)
			break;

	if (i != sizeof index)
		return query_dispose(context, inode, buf, fixIt,
				     "entry %d is unreachable", i);

	return TRUE;
}

boolean
fs_dir_mark_used_inodes(fs_Context *context, uint8 *bitmap, fs_Inode *dir, boolean fixIt)
{
	uint64 pagecount = ceilingDiv64(dir->dinode.size, PAGE_SIZE);
	uint64 page;
	int linkcount = 2;
    eeeRTC_t beginTime, delta;

    if (dir->inum == FS_SNAPSHOT_DIR_INODE)
        return TRUE;

	if (DIR_IS_COMPACT(dir))
		return fs_dir_compact_mark_used_inodes(context, bitmap, dir, fixIt);

	for (page = 0; page < pagecount; page++) {
		fs_Buf *buf = O_GET_EXCL(dir, page);
		fs_Dirleaf *d = buf->u.dirleaf;
		int removes[FS_DIRLEAF_ENTRIES_MAX];
		int nremoves;
		int i;

		if (!fs_dirleaf_eek_check(context, dir, buf, fixIt)) {
			if (!fs_bufWasModified(context, buf))
				fs_bufReleaseExclusive(context, buf);
			continue;
		}
	
		nremoves = 0;
		bzero(removes, sizeof removes);
        eee_rtc(&beginTime);

		for (i = 0; i < FS_DIRLEAF_ENTRIES_MAX; i++) {
            eee_rtcDuration(&beginTime, &delta);
            if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
            {
                fs_sleep(context, 0);
                eee_rtc(&beginTime);
            }

			if (d->u[i].ent.len == 0)
				continue;

			if (!check_inum(context, bitmap, dir, fixIt, d->u[i].ent.inum, &linkcount))
				removes[nremoves++] = i;
		}

        // fixIt false for walking phase and eek'ing snapshots

        if (fixIt) {
		    if (nremoves > 0)
			    buf = copy_block_maybe(context, dir, buf);

		    for (i = 0; i < nremoves; i++)
			    fs_dir_removename(context, dir, buf, removes[i]);
        }

		if (!fs_bufWasModified(context, buf))
		    fs_bufReleaseExclusive(context, buf);
	}

	if (linkcount != dir->dinode.linkCnt) {
		char msg[256];
		context->eek.errcount++;
		sprintf(msg, "MISMATCH ON link count for inode %lld, EXPECTED %d, GOT %d.",
			    dir->inum, linkcount, dir->dinode.linkCnt);
        fs_eek_print_status(context, FALSE, "%s", msg);
		if (fixIt && fs_eek_query(context, "Adjust?")) {
			dir->dinode.linkCnt = linkcount;
			fs_inodeWrite(context, dir, INO_WRITE_LOG);
		}
	}

	return TRUE;
}

boolean
fs_dir_is_empty(fs_Context *context, fs_Inode *inode)
{
	uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
	uint64 page;

	FS_ASSERT(pagecount > 0);

	for (page = 0; page < pagecount; page++) {
		fs_Buf *buf = O_GET_SHARED(inode, page);
		fs_Dirleaf *d = buf->u.dirleaf;
		int count = d->count;

		fs_bufRelease(context, buf);
		if (count != 0)
			return FALSE;
	}

	return TRUE;
}

uint64
fs_dirroot_get_parent(fs_Context *context, fs_Inode *inode)
{
	fs_Buf *buf = O_GET_SHARED(inode, 0);
	fs_DirleafCompact *d = buf->u.dirleafCompact;
	uint64 inum = d->parent_inum;

	fs_bufRelease(context, buf);
	return inum;
}

void
fs_dirroot_set_parent(fs_Context *context, fs_Inode *inode, uint64 parent_inum)
{
	fs_Buf *buf = O_GET_EXCL(inode, 0);
	fs_DirleafCompact *d = buf->u.dirleafCompact;

	d->parent_inum = parent_inum;
	TOUCH_ELEM(buf, &d->parent_inum);
}

void 
fs_dirroot_init(fs_Context *context, fs_Inode *inode, uint64 parent_inum)
{
	fs_DirleafCompact *d;
	fs_Buf *buf;
	uint32 mode;

	mode = (inode->dinode.attrs & ~FS_IFMT_MASK) | FS_IFMT_DIR;
	inode->dinode.attrs = mode;
	inode->dinode.attrs |= FS_IFLG_COMPACT_DIR;
	inode->dinode.size = FS_COMPACT_PAGE_SIZE;

	fs_inodeWrite(context, inode, INO_WRITE_LOG);

	buf = O_CREATE(inode, 0);
	if (!fs_bufIsCompact(buf)) {
		fs_ramDealloc(FS_RAM_READ_DATA, buf->u.data);
		buf->u.data = fs_ramAlloc(FS_RAM_COMPACT_LEAF,
					  FS_COMPACT_PAGE_SIZE);
		fs_bufLeafShrunk(context, buf, TRUE);
	}
	bzero(buf->u.data, FS_COMPACT_PAGE_SIZE);
	FS_BUF_ZERO_PART(buf, 0, FS_COMPACT_PAGE_SIZE);

	d = buf->u.dirleafCompact;
	d->parent_inum = parent_inum;
	d->magic = DIRLEAF_COMPACT_MAGIC;
	TOUCH_ELEM(buf, &d->parent_inum);
	TOUCH_ELEM(buf, &d->magic);
}

/*
 * Initialize a long-form directory page.
 */
static void
fs_dirpage_init(fs_Dirleaf *d)
{
	int i;

	d->count = 0;
	d->freelist = 0;
	d->magic = 0;

	for (i = 0; i < FS_DIRLEAF_ENTRIES_MAX-1; i++) {
		d->u[i].ent.len = 0;
		d->u[i].ent.cont = i + 1;
	}
	d->u[FS_DIRLEAF_ENTRIES_MAX-1].ent.len = 0;
	d->u[FS_DIRLEAF_ENTRIES_MAX-1].ent.cont = FS_DIRFREE_EOL;
}



/*
 * Expand from short-form to long-form when a short page is full or has a
 * long name to be added to it.
 */
static void
fs_dirpage_expand(fs_Context *context, fs_Inode *inode, fs_Buf *buf)
{
	fs_Dirleaf *new = fs_ramAlloc(FS_RAM_READ_DATA, PAGE_SIZE);
	fs_DirleafCompact *d = buf->u.dirleafCompact;
	int i;

	FS_ASSERT(d->magic == DIRLEAF_COMPACT_MAGIC);

	fs_dirpage_init(new);
	new->parent_inum = d->parent_inum;
	new->magic = DIRLEAF_LONG_MAGIC;

	buf->u.dirleaf = new;
	fs_bufLeafExpanded(context, buf);

	for (i = 0; i < FS_DIRLEAF_COMPACT_ENTRIES_MAX; i++)
    {
		FS_ASSERT((unsigned)d->entries[i].len <= sizeof d->entries[i].name);
		if (d->entries[i].len != 0) {
			fs_Dirent *newent = &new->u[fs_dir_popent(context, buf)].ent;
			newent->len = d->entries[i].len;
			newent->inum = d->entries[i].inum;
			memcpy(newent->name, d->entries[i].name, newent->len);
			newent->hash = fs_dir_hashname(newent->name, newent->len, DIR_NAME_TYPE_LONG);
            newent->aliasCksumRand = d->entries[i].aliasCksumRand;
            newent->aliasSeq = d->entries[i].aliasSeq;
		}
	}

	// We know its a compact leaf, but its possible
	// that the underlying area is actually an ultra
	// buf, so check it and do the right dealloc.
	if (fs_bufIsCompact(buf))
		fs_ramDealloc(FS_RAM_COMPACT_LEAF, d);
	else
		fs_ramDealloc(FS_RAM_READ_DATA, d);

	dirbuf_touch(context, buf, new, sizeof *new);
	inode->dinode.attrs &= ~FS_IFLG_COMPACT_DIR;
	inode->dinode.size = PAGE_SIZE;
	fs_inodeWrite(context, inode, INO_WRITE_LOG);

	CHECK_PAGE(context, inode, buf->u.dirleaf);
}


/**
 * Iterate function used to mark the used sequence
 * numbers for an 8.3 alias.
 */
static void
fs_dir83SeqIter(fs_Inode *inode, fs_Dirent *dirent, fs_DirSearchCb *cb)
{
    int seq;

    seq = fs_dirGetAliasSeq(dirent);
    //printf("seqIter seq %d for page %lld index %d cksum 0x%x len %d\n",
    //        seq, cb->page, cb->index, cb->hash, dirent->len);
    if (seq <= FS_DIR_ALIAS_SEQ_MAX)
    {
        int seqByte = (seq - 1) / 8;
        int seqBit = 1 << ((seq - 1) % 8);

        //printf("marking bytes[%d] bit 0x%x\n", seqByte, seqBit);
        cb->seqBits.bytes[seqByte] |= seqBit;
    }
}


static boolean
fs_dirSearch(fs_Context *context, fs_Inode *inode, fs_DirSearchCb *cb)
{
	fs_Buf *buf;
    int i;

    cb->hash = fs_dir_hashname(cb->name, cb->len, cb->nameType);

    //printf("dirSearch name %s len %d type %d hash 0x%x\n",
    //  cb->name, cb->len, cb->nameType, cb->hash);

    if (DIR_IS_COMPACT(inode))
    {
        fs_DirleafCompact *dc;

        buf = O_GET_SHARED(inode, 0);
	    dc = buf->u.dirleafCompact;

        FS_ASSERT(dc->magic == DIRLEAF_COMPACT_MAGIC);

	    for (i = 0; i < FS_DIRLEAF_COMPACT_ENTRIES_MAX; i++) {
		    if (dc->entries[i].len == 0)
			    continue;

		    if (fs_dir_namecmp(context, inode, (fs_Dirleaf *) dc,
                               (fs_Dirent *) &dc->entries[i], cb) == 0) {
                //printf("dirSearch compact found inum %lld\n", dc->entries[i].inum);
			    cb->inum = dc->entries[i].inum;
                cb->page = 0;
                cb->index = i;
			    fs_bufRelease(context, buf);
			    return TRUE;
		    }
	    }
	    fs_bufRelease(context, buf);
        //printf("dirSearch compact not found\n");
        return FALSE;
    }
    else
    {
	    uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
	    fs_Dirleaf *d;
	    uint64 page;
        int32 dirhashFound;

        FS_ASSERT(pagecount > 0);

	    /*
	     * A shortcut exists if the directory has an associated hash
	     * table.
	     */
        dirhashFound = fs_dirhash_lookup(context, inode, cb);
    
        if (dirhashFound == FS_DIRHASH_NOT_FOUND) {
            //printf("dirSearch dirhash not found\n");
            return FALSE;
        }
        else if (dirhashFound == FS_DIRHASH_FOUND) {
            //printf("dirSearch dirhash found\n");
		    return TRUE;
        }
    
	    /*
	     * Otherwise do a slow linear search.
	     */
	    for (page = 0; page < pagecount; page++) {
            int i;
		    buf = O_GET_SHARED(inode, page);
		    d = buf->u.dirleaf;
    
		    if (page == 0)
			    FS_ASSERT(d->magic == DIRLEAF_LONG_MAGIC);
    
		    for (i = 0; i < FS_DIRLEAF_ENTRIES_MAX; i++) {
			    if (d->u[i].ent.len == 0)
				    continue;
			    if (fs_dir_namecmp(context, inode, d, &d->u[i].ent, cb) == 0) {
                    //printf("dirSearch long found inum %lld\n", d->u[i].ent.inum);
				    cb->inum = d->u[i].ent.inum;
                    cb->page = page;
                    cb->index = i;
				    fs_bufRelease(context, buf);
				    return TRUE;
			    }
		    }
		    fs_bufRelease(context, buf);
	    }

        //printf("dirSearch long not found\n");
	    return FALSE;
    }
}


/**
 * Find an available random number to use as an alias.
 */
static uint32
fs_dirGetNewRandom(fs_Context *context, fs_Inode *inode)
{
    fs_DirSearchCb  cb;
    uint32  randomNum;
    char    name[16];
    int     numTries = 0;

    cb.nameType = DIR_NAME_TYPE_RAND;
    cb.name = name;

    do {
        randomNum = fs_random64(context, 0x0fffffffULL);
        cb.len = sprintf(name, "~%07X", randomNum);

        if (++numTries == 1000)
        {
            printf("dirGetNewRandom, too many retries\n");
            fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
            return 0;
        }

    } while (fs_dirSearch(context, inode, &cb));

    FS_ASSERT(randomNum == cb.hash);

    return cb.hash;
}


static void
fs_dirAdd83Name(fs_Context *context, fs_Inode *inode, fs_Dirent *dirent,
                char *filename, int namelen, uint32 cksum)
{
    fs_DirSearchCb  cb;
    int dw, b;
    int seq = 0;

    if (cksum == 0)
    {
        union {
            uint64  dwords[2];
            char    name[16];
        } alias;
        int baselen, len1, len2;

        alias.dwords[0] = 0;
        alias.dwords[1] = 0;
        baselen = len1 = fs_dirConstruct83Base(context, filename, namelen, &alias.name[0], 6);
        // Add the ~1 so the cksummer will function properly.
        alias.name[len1++] = '~';
        alias.name[len1++] = '1';
    
        len2 = fs_dirConstruct83Extension(context, filename, namelen, &alias.name[len1]);

        if (baselen + len2 == 0)
        {
            // The name cannot be converted to a valid 8.3
            // name.  Use random number format.
            dirent->aliasCksumRand = fs_dirGetNewRandom(context, inode);
            fs_dirSetAliasSeq(dirent, FS_DIR_ALIAS_SEQ_RAND);
            return;
        }

        cksum = fs_dir_hashname(alias.name, len1 + len2, DIR_NAME_TYPE_SEQ);
    }

    bzero(&cb.seqBits, sizeof cb.seqBits);
    cb.hash = cksum;
    cb.nameType = DIR_NAME_TYPE_SEQ;
    cb.name = filename;
    cb.len = namelen;

    //printf("dir_iter start for %s,%d\n", filename, namelen);

    // Iterate the directory marking the used sequence
    // numbers for this hash.
    dir_iterate(context, inode, &cb, fs_dir83SeqIter);

    // Now find the lowest available sequence number
    // and use it.
    for (dw = 0; dw < DIR_SEQ_NUM_DWORDS; dw++)
    {
        if (cb.seqBits.dwords[dw] != 0xffffffffffffffff)
        {
            for (b = dw * 8; b < DIR_SEQ_NUM_BYTES; b++)
            {
                if (cb.seqBits.bytes[b] != 0xff)
                {
                    int bit;

                    for (bit = 0; bit < 8; bit++)
                    {
                        if ((cb.seqBits.bytes[b] & (1 << bit)) == 0)
                        {
                            seq = (b * 8) + bit + 1;
                            if (seq <= FS_DIR_ALIAS_SEQ_MAX)
                            {
                                //printf("using new seq %d (b:%d bit:0x%x)\n", seq, b, bit);
                                fs_dirSetAliasSeq(dirent, seq);
                                dirent->aliasCksumRand = cksum;
                            }
                            else
                            {
                                // Reached end of sequence numbers for this
                                // cksum, use random number format.
                                fs_dirSetAliasSeq(dirent, FS_DIR_ALIAS_SEQ_RAND);
                                dirent->aliasCksumRand = fs_dirGetNewRandom(context, inode);
                                //printf("using random format for %s (maxSeq %d) 0x%x\n",
                                //  filename, seq, dirent->aliasCksumRand);
                            }
                            return;
                        }
                    }
                }
            }
        }
    }
}


/**
 * Return TRUE if the specified alias exists.
 */
static boolean
fs_dirAliasSeqExists(fs_Context *context, fs_Inode *inode, char *name,
                     int namelen, uint32 cksum, int seq)
{
    fs_DirSearchCb  cb;

    cb.name = name;
    cb.len = namelen;
    cb.hash = cksum;
    cb.seq = seq;
    cb.nameType = DIR_NAME_TYPE_SEQ;

    //printf("fs_dirAliasSeqExists %s len %d cksum 0x%x seq %d\n",
    //    name, namelen, cksum, seq);

    return (fs_dirSearch(context, inode, &cb));
}


/**
 * Return TRUE if the specified random number alias exists.
 */
static boolean
fs_dirAliasRandExists(fs_Context *context, fs_Inode *inode, char *name,
                      int namelen, uint32 randomNum)
{
    fs_DirSearchCb  cb;

    cb.name = name;
    cb.len = namelen;
    cb.hash = randomNum;
    cb.nameType = DIR_NAME_TYPE_RAND;

    //printf("fs_dirAliasRandExists %s len %d rand 0x%x\n",
    //    name, namelen, randomNum);

    return (fs_dirSearch(context, inode, &cb));
}


/**
 * Add an appropriate 8.3 alias if necessary.
 */
static void
fs_dir_addent_83(fs_Context *context, fs_Inode *inode, uint64 page, char *name,
                 int len, fs_Dirent *dirent, int i, boolean isCompact)
{
    int32 seq;
    uint32 hash, randomNum;
    boolean hasIllChar;

    if (!fs_dirNameIs83Compat(name, len, &hasIllChar))
    {
        fs_dirAdd83Name(context, inode, dirent, name, len, 0);
        if (hasIllChar)
            dirent->aliasSeq |= FS_DIR_ALIAS_SEQ_FLG_ILCH;

        if (!isCompact) {
            //printf("addent_83 seq dirhash ins %lld,%d,0x%x\n",
            //  page, i, dirent->aliasCksumRand);
            fs_dirhash_insert(context, inode, page, i, dirent->aliasCksumRand);
        }
    }
    else if (fs_dirNameIs83Generated(name, len, &seq))
    {
        hash = fs_dir_hashname(name, len, DIR_NAME_TYPE_SEQ);

        if (fs_dirAliasSeqExists(context, inode, name, len, hash, seq))
        {
            // An alias that matches this entry already exists.
            // Create a unique alias for this entry to use for
            // cifs clients.
            fs_dirAdd83Name(context, inode, dirent, name, len, hash);
            dirent->aliasSeq |= FS_DIR_ALIAS_SEQ_FLG_DUPE;
        }
        else
        {
            fs_dirSetAliasSeq(dirent, seq);
            dirent->aliasCksumRand = hash;
        }

        if (!isCompact) {
            //printf("addent_83 gen dirhash ins %lld,%d,0x%x\n", page, i, hash);
            fs_dirhash_insert(context, inode, page, i, hash);
        }
    }
    else if (fs_dirNameIsRandom(name, len, &randomNum))
    {
        if (fs_dirAliasRandExists(context, inode, name, len, randomNum))
        {
            // An alias that matches this entry already exists.
            // Create a unique alias for this entry to use for
            // cifs clients.
            fs_dirAdd83Name(context, inode, dirent, name, len, 0);
            dirent->aliasSeq |= FS_DIR_ALIAS_SEQ_FLG_DUPE;
            hash = dirent->aliasCksumRand;
        }
        else
        {
            fs_dirSetAliasSeq(dirent, FS_DIR_ALIAS_SEQ_RAND);
            dirent->aliasCksumRand = hash = randomNum;
        }

        if (!isCompact) {
            //printf("addent_83 rand dirhash ins %lld,%d,0x%x\n", page, i, hash);
            fs_dirhash_insert(context, inode, page, i, hash);
        }
    }
    else if (context->hashCollides)
    {
        //printf("add83 for hash collides\n");
        fs_dirAdd83Name(context, inode, dirent, name, len, 0);
        dirent->aliasSeq |= FS_DIR_ALIAS_SEQ_FLG_CASE;
        if (hasIllChar)
            dirent->aliasSeq |= FS_DIR_ALIAS_SEQ_FLG_ILCH;
        if (!isCompact) {
            //printf("addent_83 hashCollides dirhash ins %lld,%d,0x%x\n",
            //  page, i, dirent->aliasCksumRand);
            fs_dirhash_insert(context, inode, page, i, dirent->aliasCksumRand);
        }
    }
    context->hashCollides = FALSE;
}


static int
fs_dir_compact_addent(fs_Context *context, fs_Inode *inode, fs_Buf *buf, char *name, int len, uint64 inum)
{
	fs_DirleafCompact *d = buf->u.dirleafCompact;
	int i;

	FS_ASSERT(d->magic == DIRLEAF_COMPACT_MAGIC);
	
	for (i = 0; i < FS_DIRLEAF_COMPACT_ENTRIES_MAX; i++) {
		if (d->entries[i].len == 0)
			break;
	}

	FS_ASSERT(i < FS_DIRLEAF_COMPACT_ENTRIES_MAX);
	FS_ASSERT((unsigned)len <= sizeof d->entries[0].name);

	d->entries[i].len = len;
	d->entries[i].inum = inum;
    d->entries[i].aliasCksumRand = 0;
    d->entries[i].aliasSeq = 0;
	memcpy(d->entries[i].name, name, len);
	d->count++;

    fs_dir_addent_83(context, inode, 0, name, len,
                     (fs_Dirent *) &d->entries[i], i, TRUE);

    if ((d->entries[i].aliasSeq & NO_DNLC_ENTRY) == 0)
    {
        fs_dnlcInsert(context->fileSystem,
                      inode->inum,
                      name,
                      len,
                      inum);
    }
    else
    {
        fs_dnlcRemove(context->fileSystem,
                      inode->inum,
                      name,
                      len);
    }

	TOUCH_ELEM(buf, &d->entries[i]);
	TOUCH_ELEM(buf, &d->count);

	return i;
}

int
fs_dir_addent(fs_Context *context, fs_Inode *inode, char *name, int len, uint64 inum)
{
	uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
	uint64 page;
	int chunks;

	FS_ASSERT(pagecount > 0);

	if (DIR_IS_COMPACT(inode)) {
		fs_Buf *buf = O_GET_EXCL(inode, 0);
		fs_DirleafCompact *dc = buf->u.dirleafCompact;

		if (dc->count < FS_DIRLEAF_COMPACT_ENTRIES_MAX &&
            len <= sizeof dc->entries[0].name)
        {
			fs_dir_compact_addent(context, inode, buf, name, len, inum);
			return NFX_OK;
		}

		fs_dirpage_expand(context, inode, buf);
	}

	chunks = name_chunks(len);

	page = 0;
	while (1) {
		fs_Buf *buf;
		fs_Dirleaf *d;
		int i;

		if (page < pagecount) {
			buf = O_GET_EXCL(inode, page);
			d = buf->u.dirleaf;
		} else {
			buf = O_CREATE(inode, page);
			d = buf->u.dirleaf;

			inode->dinode.size += PAGE_SIZE;
			fs_inodeWrite(context, inode, INO_WRITE_LOG);

			fs_dirpage_init(d);
			dirbuf_touch(context, buf, d, sizeof *d);
		}

		if (page == 0)
			FS_ASSERT(d->magic == DIRLEAF_LONG_MAGIC);

		if (chunks > (FS_DIRLEAF_ENTRIES_MAX - d->count)) {
			FS_ASSERT(page < pagecount);
			fs_bufReleaseExclusive(context, buf);
			page++;
			continue;
		}

		CHECK_PAGE(context, inode, d);

		i = fs_dir_popent(context, buf);
        d->u[i].ent.spare = 0;
		d->u[i].ent.len = len;
		d->u[i].ent.inum = inum;
		d->u[i].ent.hash = fs_dir_hashname(name, len, DIR_NAME_TYPE_LONG);
		pack_name(context, buf, &d->u[i].ent, chunks, name, len);
        //printf("addent dirhash ins %lld,%d,0x%x\n", page, i, d->u[i].ent.hash);
		fs_dirhash_insert(context, inode, page, i, d->u[i].ent.hash);
        d->u[i].ent.aliasCksumRand = 0;
        d->u[i].ent.aliasSeq = 0;

		CHECK_PAGE(context, inode, d);

        fs_dir_addent_83(context, inode, page, name, len, &d->u[i].ent, i,
                         FALSE);

        if ((d->u[i].ent.aliasSeq & NO_DNLC_ENTRY) == 0)
        {
            fs_dnlcInsert(context->fileSystem,
                          inode->inum,
                          name,
                          len,
                          inum);
        }
        else
        {
            fs_dnlcRemove(context->fileSystem,
                          inode->inum,
                          name,
                          len);
        }

		TOUCH_ELEM(buf, &d->u[i].ent);
		CHECK_PAGE(context, inode, d);
		return NFX_OK;
	}

	/*NOTREACHED*/
	return NFX_OK;
}


uint64
fs_dir_lookupent(fs_Context *context, fs_Inode *inode, char *name, int len)
{
    fs_DirSearchCb  cb;
    boolean hasIllChar;
    boolean is83Compat;

	/*
	 * Special case "." and ".."
	 */
	if (name[0] == '.') {
		if (len == 1)
			return inode->inum;
		if (len == 2 && name[1] == '.')
			return fs_dirroot_get_parent(context, inode);
	}

    cb.name = name;
    cb.len = len;
    cb.inum = 0;
    cb.hashMatch = FALSE;
    cb.aliasFlags = 0;

    is83Compat = fs_dirNameIs83Compat(name, len, &hasIllChar);

    cb.nameType = DIR_NAME_TYPE_LONG;
    //printf("lookupent: fs_dirSearch %s long\n", name);
    if (!fs_dirSearch(context, inode, &cb) ||
        (cb.aliasFlags & FS_DIR_ALIAS_SEQ_FLG_DUPE))
    {
        if ((fsTest83 || fsTestCase ||
             (context->reqHdr && context->reqHdr->proto == FS_PROTO_CIFS) ||
             context->caseInsensitive) &&
            is83Compat)
        {
            uint64 savedInum = cb.inum;

            cb.inum = 0;
            if (fs_dirNameIsRandom(name, len, NULL))
            {
                cb.nameType = DIR_NAME_TYPE_RAND;
                fs_dirSearch(context, inode, &cb);
            }

            if (cb.inum == 0 && fs_dirNameIs83Generated(name, len, &cb.seq))
            {
                cb.nameType = DIR_NAME_TYPE_SEQ;
                fs_dirSearch(context, inode, &cb);
            }

            // If we didn't find the alias that 
            // is a dupe of the longname, then
            // return the longname.
            if (cb.aliasFlags & FS_DIR_ALIAS_SEQ_FLG_DUPE) {
                if (cb.inum == 0) {
                    cb.inum = savedInum;
                }
            }
        }
    }

    //printf("lookupent: cb.inum %lld\n", cb.inum);

    context->hashCollides = cb.hashMatch;

    if (!context->caseInsensitive && !fsTestCase)
    {
        uint64 snapnumInum;

        FS_HANDLE_SET_INUM(snapnumInum, inode->inum);
        FS_HANDLE_SET_SNAP(snapnumInum, inode->snapnum);

        if (!(cb.aliasFlags & NO_DNLC_ENTRY) && !hasIllChar)
        {
            if (cb.inum)
            {                
                fs_dnlcInsert(context->fileSystem, snapnumInum,
                              name, len, cb.inum);
            }
            else
            {            
                fs_dnlcNegate(context->fileSystem, snapnumInum,
                              name, len);
            }
        }
    }

    return cb.inum;
}

static int
fs_dir_compact_removeent(fs_Context *context, fs_Inode *inode, fs_Buf *buf,
                         char *name, int len, int i)
{
	fs_DirleafCompact *d = buf->u.dirleafCompact;

	FS_ASSERT(d->magic == DIRLEAF_COMPACT_MAGIC);

    if (d->entries[i].aliasCksumRand)
    {
        char alias[16];
        int aliasLen;

        fs_dirReconstruct83Name(context, (fs_Dirleaf *) d,
                                (fs_Dirent *) &d->entries[i],
                                alias, &aliasLen);

        if ((d->entries[i].aliasSeq & NO_DNLC_ENTRY) == 0)
        {
            fs_dnlcNegate(context->fileSystem, inode->inum, alias, aliasLen);
        }
    }

    // Remove the dnlc entry for the long name.
    FS_ASSERT(FS_DNLC_NAME_MAX <= sizeof d->entries[i].name);
    if (d->entries[i].len <= FS_DNLC_NAME_MAX)
    {
        if ((d->entries[i].aliasSeq & NO_DNLC_ENTRY) == 0)
        {
            fs_dnlcNegate(context->fileSystem, inode->inum, d->entries[i].name,
                          d->entries[i].len);
        }
    }

	d->entries[i].len = 0;
	--d->count;

	TOUCH_ELEM(buf, &d->entries[i].len);
	TOUCH_ELEM(buf, &d->count);

	return NFX_OK;
}

static void
fs_dir_removename(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i)
{
	fs_Dirleaf *d = buf->u.dirleaf;
	int nexti = d->u[i].ent.cont;
	int len = d->u[i].ent.len;

    //printf("rement alias dirhash rem %lld,%d,0x%x\n", buf->page, i, d->u[i].ent.hash);
    fs_dirhash_remove(context, inode, buf->page, i,
                      d->u[i].ent.hash);

    if (d->u[i].ent.aliasCksumRand)
    {
        char alias[16];
        int aliasLen;

        fs_dirReconstruct83Name(context, d, &d->u[i].ent,
                                alias, &aliasLen);

        if ((d->u[i].ent.aliasSeq & NO_DNLC_ENTRY) == 0)
        {
            fs_dnlcNegate(context->fileSystem, inode->inum, alias, aliasLen);
        }

        //printf("rement alias dirhash rem %lld,%d,0x%x\n",
        //  buf->page, i, d->u[i].ent.aliasCksumRand);
        fs_dirhash_remove(context, inode, buf->page, i,
                          d->u[i].ent.aliasCksumRand);
    }

    // Remove the dnlc entry for the long name.
    FS_ASSERT(FS_DNLC_NAME_MAX <= sizeof d->u[i].ent.name);
    if (d->u[i].ent.len <= FS_DNLC_NAME_MAX)
    {
        if ((d->u[i].ent.aliasSeq & NO_DNLC_ENTRY) == 0)
        {
            fs_dnlcNegate(context->fileSystem, inode->inum, d->u[i].ent.name,
                          d->u[i].ent.len);
        }
    }

	fs_dir_pushent(context, buf, i);

	/*
	 * Zap continuations.
	 */
	len -= sizeof d->u[i].ent.name;
	while (len > 0) {
		i = nexti;
		nexti = d->u[nexti].ent.cont;
		fs_dir_pushent(context, buf, i);
		len -= sizeof ((fs_Dircont *) &d->u[i].ent)->name;
	}
}

int
fs_dir_removeent(fs_Context *context, fs_Inode *inode, char *name, int len)
{
    fs_DirSearchCb  cb;
    boolean         found;

    cb.name = name;
    cb.len = len;
    cb.inum = 0;
    cb.aliasFlags = 0;

    cb.nameType = DIR_NAME_TYPE_LONG;
    //printf("removeent: fs_dirSearch long\n");
    found = fs_dirSearch(context, inode, &cb);

    if ((!found || (cb.aliasFlags & FS_DIR_ALIAS_SEQ_FLG_DUPE)) && 
        (fsTest83 || fsTestCase || context->use83Names ||
         context->caseInsensitive) &&
        fs_dirNameIs83Compat(name, len, NULL))
    {
        int32 savedIndex = cb.index;

        if (fs_dirNameIsRandom(name, len, NULL))
        {
            cb.nameType = DIR_NAME_TYPE_RAND;
            //printf("removeent: fs_dirSearch rand\n");
            found = fs_dirSearch(context, inode, &cb);
        }

        if (!found && fs_dirNameIs83Generated(name, len, &cb.seq))
        {
            cb.nameType = DIR_NAME_TYPE_SEQ;
            //printf("removeent: fs_dirSearch seq\n");
            found = fs_dirSearch(context, inode, &cb);
        }

        if (cb.aliasFlags & FS_DIR_ALIAS_SEQ_FLG_DUPE) {
            if (cb.inum == 0) {
                cb.index = savedIndex;
            }
        }
    }

    FS_ASSERT(found);

    if (found)
    {
        fs_Buf      *buf = O_GET_EXCL(inode, cb.page);

        if (DIR_IS_COMPACT(inode))
        {
            fs_dir_compact_removeent(context, inode, buf, name, len, cb.index);
        }
        else
        {
            CHECK_PAGE(context, inode, buf->u.dirleaf);
		    fs_dir_removename(context, inode, buf, cb.index);
            CHECK_PAGE(context, inode, buf->u.dirleaf);
        }

		return NFX_OK;
    }

    return NFX_ERR;
}


static void
fetchName(fs_Context *context, fs_Dirleaf *d, fs_Dirent *dirent,
          char *alias, int *lenp, boolean *useAlias)
{
    if ((fsTest83 || context->use83Names) &&
        dirent->aliasCksumRand)
    {
        fs_dirReconstruct83Name(context, d, dirent, alias, lenp);
        *useAlias = TRUE;
    }
    else if ((fsTestCase || context->caseInsensitive) && 
             (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_CASE))
    {
        fs_dirReconstruct83Name(context, d, dirent, alias, lenp);
        *useAlias = TRUE;
    }
    else if ((context->reqHdr &&
              context->reqHdr->proto == FS_PROTO_CIFS &&
              (dirent->aliasSeq & FS_DIR_ALIAS_SEQ_FLG_DUPE)))
    {
        fs_dirReconstruct83Name(context, d, dirent, alias, lenp);
        *useAlias = TRUE;
    }
    else
    {
        *lenp = dirent->len;
        if (fsTest83 || context->use83Names)
        {
            int x;
            for (x = 0; x < dirent->len; x++)
            {
                alias[x] = convertToUpper[(int)dirent->name[x]];
            }
            *useAlias = TRUE;
        }
        else
            *useAlias = FALSE;
    }
}


#define FIRST_DIRENT_COOKIE 1000

static void
fs_dir_readdir_compact(fs_Context *context, fs_Inode *inode,
                    fs_DirEntry *entries, char *endp,
                    uint64 cookie,
                    int32 *num_entries, int32 *eof)
{
	fs_Buf *buf = O_GET_SHARED(inode, 0);
	fs_DirleafCompact *d = buf->u.dirleafCompact;
	int ndone;
	int i;

	FS_ASSERT(d->magic == DIRLEAF_COMPACT_MAGIC);

	if (cookie < FIRST_DIRENT_COOKIE)
		i = 0;
	else
		i = cookie - FIRST_DIRENT_COOKIE + 1;

	for (ndone = 0; i < FS_DIRLEAF_COMPACT_ENTRIES_MAX; i++) {
		int len;
        char *namep;
        char alias[16];
        boolean useAlias;

		if (d->entries[i].len == 0)
			continue;

        fetchName(context, (fs_Dirleaf *) d, (fs_Dirent *) &d->entries[i],
                  alias, &len, &useAlias);
        if (useAlias)
            namep = alias;
        else
            namep = d->entries[i].name;

		if ((char *)entries + alignto(sizeof *entries + len + 1, sizeof (uint64)) > endp)
			break;

		entries->inode = d->entries[i].inum;
		memcpy(entries->name, namep, len);
		entries->name[len] = '\0';
		entries->nameLen = len + 1;			/* counts NUL */
		entries->cookie = i + FIRST_DIRENT_COOKIE;

		FS_READDIR_ADV_ENTRY_PTR(entries);
		ndone++;
	}

	*num_entries += ndone;
	*eof = (i == FS_DIRLEAF_COMPACT_ENTRIES_MAX);

	fs_bufRelease(context, buf);

    READDIR_TRACE(("fs_readdir compact ret %d entries eof %d", *num_entries, *eof));
}

void fs_dir_readdir(fs_Context *context, fs_Inode *inode,
                    fs_DirEntry *entries,
                    int32 max_size,
                    uint64 cookie,
                    boolean show_dot_snapshot,
                    int32 *num_entries, int32 *eof)
{
	char *endp = (char *)entries + max_size;
	uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
	uint64 page;
	int i;

    READDIR_TRACE(
		("fs_readdir cookie %lld inode %p snapnum %d inum %lld entries %p max %d",
			cookie, inode, inode->snapnum, inode->inum, entries, max_size));

	*num_entries = 0;

	/*
	 * A zero cookie implies that this is the first readdir
	 * request. So, fill in the reply structure with "." and ".."
	 * first.
	 */
	if (cookie == 0) {
		char *startp = (char *)entries;

		entries->inode = inode->inum;
		strcpy(entries->name, ".");
		entries->nameLen = 2;	/* counts NUL */
		entries->cookie = 1;
		++*num_entries;

		FS_READDIR_ADV_ENTRY_PTR(entries);

		entries->inode = fs_dirroot_get_parent(context, inode);
		strcpy(entries->name, "..");
		entries->nameLen = 3;	/* counts NUL */
		entries->cookie = 2;
		++*num_entries;

		FS_READDIR_ADV_ENTRY_PTR(entries);

		if (show_dot_snapshot) {
			entries->inode = FS_SNAPSHOT_DIR_INODE;
			strcpy(entries->name, ".snapshot");
			entries->nameLen = 10;	/* counts NUL */
			entries->cookie = 3;
			++*num_entries;

			FS_READDIR_ADV_ENTRY_PTR(entries);
		}

		max_size -= (char *)entries - startp;
	}

	if (DIR_IS_COMPACT(inode)) {
		fs_dir_readdir_compact(context, inode, entries, endp,
				               cookie, num_entries, eof);
		return;
	}

	if (cookie >= FIRST_DIRENT_COOKIE) {
		// printf("cookie: %lld\n", cookie);
		i = cookie - FIRST_DIRENT_COOKIE + 1;
		page = i / FS_DIRLEAF_ENTRIES_MAX;
		i = i % FS_DIRLEAF_ENTRIES_MAX;
	} else {
		i = 0;
		page = 0;
	}

	for (; page < pagecount; page++) {
		fs_Buf *buf = O_GET_SHARED(inode, page);
		fs_Dirleaf *d = buf->u.dirleaf;
		int j = page*FS_DIRLEAF_ENTRIES_MAX + FIRST_DIRENT_COOKIE;
		int ndone = 0;

		if (page == 0)
			FS_ASSERT(d->magic == DIRLEAF_LONG_MAGIC);

		for (; i < FS_DIRLEAF_ENTRIES_MAX; i++) {

			int len;
            char alias[16];
            boolean useAlias;

			if (d->u[i].ent.len == 0)
				continue;

            fetchName(context, d, &d->u[i].ent, alias, &len, &useAlias);
	
			if ((char *)entries + alignto(sizeof *entries + len + 1, sizeof (uint64)) > endp) {
				fs_bufRelease(context, buf);
				*num_entries += ndone;
				*eof = 0;
                READDIR_TRACE(("fs_readdir ret %d entries eof %d", *num_entries, *eof));
				return;
			}
	
			entries->inode = d->u[i].ent.inum;

            if (useAlias) {
                memcpy(entries->name, alias, len);
            }
            else {
			    unpack_name(entries->name, d, &d->u[i].ent);
            }
			entries->name[len] = '\0';
			entries->nameLen = len + 1;			/* counts NUL */
			entries->cookie = i + j;
			// printf("readdir: %s, %lld\n", entries->name, entries->cookie);
	
			FS_READDIR_ADV_ENTRY_PTR(entries);
			ndone++;
		}

		i = 0;
		fs_bufRelease(context, buf);
		*num_entries += ndone;
	}

	*eof = (page == pagecount);

    READDIR_TRACE(("fs_readdir ret %d entries eof %d", *num_entries, *eof));
}
