// -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-dnlc.c
 *
 * RCS:  $Id: fs-dnlc.c,v 1.5 2002/11/26 22:53:58 bob Exp $
 *
 * Description:  Directory name lookup cache
 * 
 */

#if !defined(SSC)

#include "fs-includes.h"

/**
 * The global dnlc.
 */
fs_Dnlc fs_dnlc;


/**
 * Return the hash code for the specified dnlc entry.
 */
static int32
fs_dnlcHash(
    evm_volId_t volId,
    uint64      parentInumSnap,
    char        *name,
    int32       len)
{
    uint64 hash;
    uint64 inum = FS_HANDLE_GET_INUM(parentInumSnap);
    int32  snapnum = FS_HANDLE_GET_SNAP(parentInumSnap);

    hash = volId.details.u.parts.clusterUnique + (inum << 5);
    hash += (hash >> FS_DNLC_HASH_SHIFT);
    hash += snapnum;

    return (fs_checksum(name, len, hash) % FS_DNLC_HASH_MAX);
}


/**
 * Return a free dnlc entry by recycling the LRU or
 * allocating a new one.
 *
 * Must be called with spinlock held.
 */
static fs_DnlcEntry *
fs_dnlcFindFree()
{
    fs_DnlcEntry    *entry;

    if (fs_dnlc.stats.used < fs_dnlc.stats.usedLimit)
    {
        entry = fs_ramAlloc(FS_RAM_DNLC_ENTRY, sizeof *entry);
        DLL_ISOLATE_ELEM(entry, nameEntry);
        DLL_ISOLATE_ELEM(entry, mruEntry);
        
        fs_dnlc.stats.used += 1;
    }
    else
    {
        DLL_REMOVE_TAIL(&fs_dnlc.mruHdr, entry, mruEntry);

        if (!DLL_IS_ISOLATED(entry, nameEntry))
        {
            int32 hash;

            hash = fs_dnlcHash(entry->volId, entry->parentInumSnap,
                               entry->name, entry->len);
            DLL_REMOVE(&fs_dnlc.nameHdr[hash], entry, nameEntry);
            FS_ASSERT(fs_dnlc.depth[hash] > 0);
            fs_dnlc.depth[hash] -= 1;
        }
    }

    return entry;
}


/**
 * Initialize the dnlc.
 */
void
fs_dnlcInit()
{
    int32   i;

    for (i = 0; i < FS_DNLC_HASH_MAX; i++)
    {
        DLL_INIT(&fs_dnlc.nameHdr[i]);
        fs_dnlc.depth[i] = 0;
    }

    NFX_SPIN_INIT(&fs_dnlc.spin,SPINLOCK_LEVEL_UNKNOWN);
    DLL_INIT(&fs_dnlc.mruHdr);
    fs_dnlc.stats.usedLimit = FS_DNLC_USED_MAX;
}


/**
 * Search the dnlc for the specified entry in the
 * specified hash bucket.  Return pointer to it if
 * found.
 *
 * Must be called with spinlock held.
 */
static fs_DnlcEntry *
fs_dnlcEntryExists(
    evm_volId_t volId,
    uint64      parentInumSnap,
    char        *name,
    int32       len,
    int32       hash)
{
    fs_DnlcEntry    *entry;

    DLL_ITERATE(entry, &fs_dnlc.nameHdr[hash], nameEntry)
    {
        if (entry->volId.whole_thing == volId.whole_thing &&
            entry->parentInumSnap == parentInumSnap &&
            entry->len == len &&
            memcmp(entry->name, name, len) == 0)
        {
            return entry;
        }
    }

    return NULL;
}


/**
 * Lookup the specified entry in the dnlc by name filling
 * in parent and child attributes and child inode genNum
 * if found.   Returns whether the entry exists in the
 * directory, doesn't exist in the directory, or the dnlc
 * doesn't know and a full probe of the directory is
 * required.
 */
int32
fs_dnlcLookup(
    fs_FileSystem   *fileSystem,
    fs_Proto        protocol,
    uint64          parentInumSnap,
    char            *name,
    int32           len,
    fs_Attr         *attrp,
    uint32          *genNum)
{
    fs_DnlcEntry    *entry;
    int32           hash;

    //printf("fs_dnlcLookup entered parent 0x%llx name %s len %d\n",
    //  parentInumSnap, name, len);

    if (len > FS_DNLC_NAME_MAX)
    {
        NFX_SPIN_LOCK(&fs_dnlc.spin);
        fs_dnlc.stats.missTooBig += 1;
        fileSystem->dnlcStats.missTooBig += 1;
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("fs_dnlcLookup miss too long\n");
        return FS_DNLC_UNKNOWN;
    }

    hash = fs_dnlcHash(fileSystem->volId, parentInumSnap, name, len);

    NFX_SPIN_LOCK(&fs_dnlc.spin);
    entry = fs_dnlcEntryExists(fileSystem->volId, parentInumSnap, name, len, hash);
    if (entry == NULL)
    {
        fs_dnlc.stats.missNotFound += 1;
        fileSystem->dnlcStats.missNotFound += 1;
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("fs_dnlcLookup miss not found\n");
        return FS_DNLC_UNKNOWN;
    }

    DLL_MOVE_TO_HEAD(&fs_dnlc.mruHdr, entry, mruEntry);

    if (entry->negative)
    {
        if (!fs_inodeGetAttrs(fileSystem, attrp, 
                              FS_HANDLE_GET_INUM(parentInumSnap),
                              FS_HANDLE_GET_SNAP(parentInumSnap), protocol, 
                              FS_POST_PAR_ATTRS, NULL))
        {
            fs_dnlc.stats.missInode += 1;
            fileSystem->dnlcStats.missInode += 1;
            NFX_SPIN_UNLOCK(&fs_dnlc.spin);
            //printf("fs_dnlcLookup hit negative, no parent inode\n");
            return FS_DNLC_UNKNOWN;
        }

        fs_dnlc.stats.hitsNegative += 1;
        fileSystem->dnlcStats.hitsNegative += 1;
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("fs_dnlcLookup hit negative\n");
        return FS_DNLC_NOT_EXISTS;
    }

    fs_dnlc.stats.hits += 1;
    fileSystem->dnlcStats.hits += 1;

    // Now call into the inode cache to get parent and child
    // attributes while holding the dnlc spinlock to ensure
    // a consistent response.
    if (!fs_inodeGetAttrs(fileSystem, attrp, entry->inum,
                          FS_HANDLE_GET_SNAP(parentInumSnap), protocol, 
                          FS_POST_ATTRS, genNum))
    {
        fs_dnlc.stats.missInode += 1;
        fileSystem->dnlcStats.missInode += 1;
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("fs_dnlcLookup hit, no child inode\n");
        return FS_DNLC_UNKNOWN;
    }

    attrp += 1;
    if (!fs_inodeGetAttrs(fileSystem, attrp, FS_HANDLE_GET_INUM(parentInumSnap),
                          FS_HANDLE_GET_SNAP(parentInumSnap), protocol, 
                          FS_POST_PAR_ATTRS, NULL))
    {
        fs_dnlc.stats.missInode += 1;
        fileSystem->dnlcStats.missInode += 1;
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("fs_dnlcLookup hit, no parent inode\n");
        return FS_DNLC_UNKNOWN;
    }

    NFX_SPIN_UNLOCK(&fs_dnlc.spin);

    //printf("fs_dnlcLookup hit! inum %lld\n", entry->inum);
    return FS_DNLC_EXISTS;
}


/**
 * Insert the specified entry in the dnlc.  Its the
 * caller's responsibility to ensure that the name
 * isn't already in the cache.
 *
 * Must be called with spinlock held.
 */
static void
fs_doDnlcInsert(
    evm_volId_t volId,
    uint64      parentInumSnap,
    char        *name,
    int32       len,
    uint64      inum,
    int32       hash,
    char        negative)
{
    fs_DnlcEntry    *entry;

    FS_ASSERT(len <= FS_DNLC_NAME_MAX);

    entry = fs_dnlcFindFree();
    entry->volId.whole_thing = volId.whole_thing;
    entry->parentInumSnap = parentInumSnap;
    memcpy(entry->name, name, len);
    entry->len = len;
    entry->inum = inum;
    entry->negative = negative;

    DLL_INSERT_HEAD(&fs_dnlc.nameHdr[hash], entry, nameEntry);
    fs_dnlc.depth[hash] += 1;
    DLL_INSERT_HEAD(&fs_dnlc.mruHdr, entry, mruEntry);
}


/**
 * Insert the specified entry into the dnlc if it
 * doesn't already exist in the cache.  Returns TRUE
 * if the insertion takes place or its already in the
 * cache, FALSE if it cannot be inserted.
 */
boolean
fs_dnlcInsert(
    fs_FileSystem   *fileSystem,
    uint64          parentInumSnap,
    char            *name,
    int32           len,
    uint64          inum)
{
    fs_DnlcEntry    *entry;
    int32           hash;

    //printf("dnlcInsert entered\n");

    if (len > FS_DNLC_NAME_MAX)
    {
        //printf("dnlcInsert too long\n");
        return FALSE;
    }

    hash = fs_dnlcHash(fileSystem->volId, parentInumSnap, name, len);
    NFX_SPIN_LOCK(&fs_dnlc.spin);
    entry = fs_dnlcEntryExists(fileSystem->volId, parentInumSnap, name, len, hash);
    if (entry != NULL)
    {
        if (entry->negative)
        {
            entry->negative = 0;
            entry->inum = inum;
            fs_dnlc.stats.setPositives += 1;
            fileSystem->dnlcStats.setPositives += 1;
            //printf("dnlcInsert exists was negative\n");
        }
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("dnlcInsert exists\n");
        return TRUE;
    }
    fs_doDnlcInsert(fileSystem->volId, parentInumSnap, name, len, inum, hash, 0);
    fs_dnlc.stats.inserts += 1;
    fileSystem->dnlcStats.inserts += 1;
    NFX_SPIN_UNLOCK(&fs_dnlc.spin);
    //printf("dnlcInsert inserted\n");
    return TRUE;
}


/**
 * Negate the specified entry in the dnlc.
 * Returns TRUE if the entry was negated (or a
 * negative entry inserted), FALSE if it couldn't
 * be.
 */
boolean
fs_dnlcNegate(
    fs_FileSystem   *fileSystem,
    uint64          parentInumSnap,
    char            *name,
    int32           len)
{
    fs_DnlcEntry    *entry;
    int32           hash;

    //printf("dnlcNegate entered\n");

    if (len > FS_DNLC_NAME_MAX)
    {
        //printf("dnlcNegate too long\n");
        return FALSE;
    }

    hash = fs_dnlcHash(fileSystem->volId, parentInumSnap, name, len);

    NFX_SPIN_LOCK(&fs_dnlc.spin);
    entry = fs_dnlcEntryExists(fileSystem->volId, parentInumSnap, name, len, hash);

#if 0 // set to 1 to turn off negative caching
    if (entry != NULL) {
            DLL_REMOVE(&fs_dnlc.nameHdr[hash], entry, nameEntry);
            FS_ASSERT(fs_dnlc.depth[hash] > 0);
            fs_dnlc.depth[hash] -= 1;
            DLL_MOVE_TO_TAIL(&fs_dnlc.mruHdr, entry, mruEntry);
    }
    NFX_SPIN_UNLOCK(&fs_dnlc.spin);
    return TRUE;
#endif

    if (entry != NULL)
    {
        if (!entry->negative)
        {
            entry->negative = 1;
            fs_dnlc.stats.setNegatives += 1;
            fileSystem->dnlcStats.setNegatives += 1;
            //printf("dnlcNegate exists was positive\n");
        }
        NFX_SPIN_UNLOCK(&fs_dnlc.spin);
        //printf("dnlcNegate exists\n");
        return TRUE;
    }
    fs_doDnlcInsert(fileSystem->volId, parentInumSnap, name, len, 0, hash, 1)
;
    fs_dnlc.stats.inserts += 1;
    fileSystem->dnlcStats.inserts += 1;
    NFX_SPIN_UNLOCK(&fs_dnlc.spin);
    //printf("dnlcNegate inserted\n");
    return TRUE;
}


/**
 * Remove the specified entry from the dnlc.
 * Returns TRUE if the entry was removed or
 * FALSE if it wasn't found.
 */
boolean
fs_dnlcRemove(
    fs_FileSystem   *fileSystem,
    uint64          parentInumSnap,
    char            *name,
    int32           len)
{
    fs_DnlcEntry    *entry;
    int32           hash;
    
    //printf("dnlcRemove entered\n");

    if (len > FS_DNLC_NAME_MAX)
    {
        //printf("dnlcRemove too long\n");
        return FALSE;
    }
    
    hash = fs_dnlcHash(fileSystem->volId, parentInumSnap, name, len);

    NFX_SPIN_LOCK(&fs_dnlc.spin);
    entry = fs_dnlcEntryExists(fileSystem->volId, parentInumSnap, name, len, hash);
    
    if (entry != NULL)
    {
        DLL_REMOVE(&fs_dnlc.nameHdr[hash], entry, nameEntry);
        FS_ASSERT(fs_dnlc.depth[hash] > 0);
        fs_dnlc.depth[hash] -= 1;
        fs_dnlc.stats.removes += 1;
        fileSystem->dnlcStats.removes += 1;
        DLL_MOVE_TO_TAIL(&fs_dnlc.mruHdr, entry, mruEntry);
    }
    NFX_SPIN_UNLOCK(&fs_dnlc.spin);

    return (entry != NULL);
}


/**
 * Flush the dnlc for the specified filesystem.
 */
void
fs_dnlcDestroyVolume(
    fs_FileSystem   *fileSystem)
{
    fs_DnlcEntry    *entry, *next;
    int32           hash;

    //printf("dnlcDestroy entered\n");
    NFX_SPIN_LOCK(&fs_dnlc.spin);
    for (hash = 0; hash < FS_DNLC_HASH_MAX; hash++)
    {
        DLL_ITERATE_NEXT(entry, next, &fs_dnlc.nameHdr[hash], nameEntry)
        {
            if (entry->volId.whole_thing == fileSystem->volId.whole_thing)
            {
                //printf("dnlcDestroy removed inum %lld\n", entry->inum);
                DLL_REMOVE(&fs_dnlc.nameHdr[hash], entry, nameEntry);
                FS_ASSERT(fs_dnlc.depth[hash] > 0);
                fs_dnlc.depth[hash] -= 1;
                DLL_MOVE_TO_TAIL(&fs_dnlc.mruHdr, entry, mruEntry);
            }
        }
    }
    NFX_SPIN_UNLOCK(&fs_dnlc.spin);
    //printf("dnlcDestroy done\n");
}


/**
 * Flush the dnlc for the specified filesystem,snapnum.
 */
void
fs_dnlcDestroySnapnum(
    fs_FileSystem   *fileSystem,
    int32           snapnum)
{
    fs_DnlcEntry    *entry, *next;
    int32           hash;

    //printf("dnlcDestroySnap entered\n");

    NFX_SPIN_LOCK(&fs_dnlc.spin);
    for (hash = 0; hash < FS_DNLC_HASH_MAX; hash++)
    {
        DLL_ITERATE_NEXT(entry, next, &fs_dnlc.nameHdr[hash], nameEntry)
        {
            if (entry->volId.whole_thing == fileSystem->volId.whole_thing &&
                FS_HANDLE_GET_SNAP(entry->parentInumSnap) == snapnum)
            {
                //printf("dnlcDestroySnap removed inum %lld\n", entry->inum);
                DLL_REMOVE(&fs_dnlc.nameHdr[hash], entry, nameEntry);
                FS_ASSERT(fs_dnlc.depth[hash] > 0);
                fs_dnlc.depth[hash] -= 1;
                DLL_MOVE_TO_TAIL(&fs_dnlc.mruHdr, entry, mruEntry);
            }
        }
    }
    NFX_SPIN_UNLOCK(&fs_dnlc.spin);
    //printf("dnlcDestroySnap done\n");
}

#endif // SSC
