/*
 * Name		: fs-dump.c
 * RCS		: $Id: fs-dump.c,v 1.96 2002/12/04 19:18:53 rahulk Exp $
 * Description	: Dump implementation
 * Created by	: Rahul Khona
 * Date Created	: 8/10/2001
 */
#if !defined(SSC)
#include "fs-includes.h"
#include "fs-dump.h"        // private declarations
#include "fs-ndmp-api.h"
#include <exports/ndmp-fs-api.h>

#ifdef FS_DEBUG
#define D   printf
#else
#define D
#endif

#ifdef FS_HIST_WAIT
fs_DumpSessionsHdr_t fs_dumpSessionsHdr;
nfx_spinlock_t	            fs_dumpSessionsSpinLock;
#endif

#define testSpecialDirectories(dirp)    \
    {                               \
        if (strcmp(dirp->name, ".") == 0                                \
         || strcmp(dirp->name, "..") == 0                               \
         || strcmp(dirp->name, ".snapshot") == 0)                       \
        {                                                               \
            FS_READDIR_ADV_ENTRY_PTR(dirp);                             \
            continue;                                                   \
        }                                                               \
    }

/*
 * Dump start of the directory
 */

PRIVATE fs_Status   fs_dumpDirNodeStart(fs_Context *, fs_DumpContext_t *,
                                        fs_Inumber, uint64, char *, int, int);
PRIVATE fs_Status fs_dumpDirNodeEnd(fs_Context *, fs_DumpContext_t *, uint64);
PRIVATE fs_Status fs_dumpReg(fs_Context *, fs_DumpContext_t *, fs_Inode *,
                             char *, int, uint8);
PRIVATE fs_Status fs_dumpSymLink(fs_Context *, fs_DumpContext_t *, fs_Inode *,
                                 char *, int);
PRIVATE fs_Status fs_dumpSpecialNode(fs_Context *context, fs_DumpContext_t *,
                                      fs_Inode *, char *, int, uint8);
PRIVATE fs_Status fs_dumpTree(fs_Context *, fs_DumpContext_t *, fs_Inode **,
                              char *, int, int);
PRIVATE void
fs_doFullFileHistories(fs_Context *context, fs_DumpContext_t *dumpContext);

PRIVATE fs_Status fs_dumpDirs(fs_Context *ctx, fs_DumpContext_t *dumpContext,
                         fs_Inode *root, char *pathEnd,
                         fs_DirEntry *readirBuf, uint32 readirBufLen);
PRIVATE fs_Status
fs_partialDumpDirNodeFileHistory(fs_Context       *context, 
                                 fs_DumpContext_t *dumpContext,
                                 fs_Inode         *root);


PRIVATE fs_Status
fs_dumpDirNodeStart(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inumber          inum,
    uint64              id,
    char                *name,
    int                 nameLen,
    int                 isFset
)
{
    fs_DumpOperation_t          *obj;
    fs_Inode                    *inode;
    uint32                      objsize;
    fs_Status                   status;
    uint32                      nheaders;
    uint32                      left;
    uint32                      totLen, availLen;
    char                        *cp, *tp;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);

    FS_ASSERT(name != NULL && dumpContext != NULL && context != NULL);

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
        ("Dumping node start - current Id %lld", id));
    obj = &dumpContext->record.u.operation;
    obj->header.headerId = id;
    obj->header.headerSubId = 0;
    obj->len = nameLen;
    obj->sourceNameLen = nameLen;
    left = FS_DR_TBSIZE - sizeof (fs_DumpOperation_t);

    nheaders = 1;
    if (nameLen > left) {
        nheaders++;
    }
    obj->header.headerBlocks = nheaders;

    if (!isFset) {
        obj->header.op = FS_DR_CREATE_DIR;
    } else {
        obj->header.op = FS_DR_CREATE_FSET;
    }
    obj->header.dataExtents = 0;

    inode = I_GET_SHARED(dumpContext->snapnum, inum);

    obj->header.attr.uid = inode->sd->u.nfs.uid;
    obj->header.attr.gid = inode->sd->u.nfs.gid;
    obj->header.attr.sizeDev.size = inode->dinode.size;
    obj->header.attr.atime = inode->dinode.atime;
    obj->header.attr.mtime = inode->dinode.mtime;
    obj->header.attr.ctime = inode->dinode.ctime;
    obj->header.attr.mode = (inode->sd->u.nfs.mode & FS_IMODE_MASK) |
                             (inode->dinode.attrs & FS_IATR_MASK);
    obj->header.attr.simpleAttrs = inode->dinode.userAttr;
    fs_inodeRelease(context, inode);

    objsize = sizeof (fs_DumpOperation_t) + nameLen;
    totLen = nameLen + 1;
    cp = name;
    tp = &dumpContext->record.u.dummy[sizeof (fs_DumpOperation_t)];
    availLen = &dumpContext->record.u.dummy[FS_DR_TBSIZE] - tp;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
             ("nodeStart:dumped %s in header", cp));
    if (availLen >= totLen) {
        bzero(tp + nameLen, availLen - nameLen);
        memcpy(tp, cp, nameLen);
    } else {
        memcpy(tp, cp, availLen);
        status = fs_writeTape(context, dumpContext);
        if (status != FS_ERR_NONE) {
            obj->header.flags = 0;
            return status;
        }
        totLen -= availLen;
        tp = dumpContext->record.u.dummy;
        bzero(tp + totLen, FS_DR_TBSIZE - totLen);
        cp += availLen;
        memcpy(tp, cp, totLen);
        tp += totLen;
    }
    
    status = fs_writeTape(context, dumpContext);
    if (status != FS_ERR_NONE) {
        return status;
    }

#ifdef BACKUP_FSET
    /*
     * If the inode is that of fileset root, dump quota limits as well
     */

    if (isFset) {
        status = fs_dumpFsetLimits(context, dumpContext, fset);
        if (status != FS_ERR_NONE) {
            return status;
        }
        status = fs_dumpUQuotaLimits(context, dumpContext, fset);
        if (status != FS_ERR_NONE) {
            return status;
        }
        status = fs_dumpGQuotaLimits(context, dumpContext, fset);
        if (status != FS_ERR_NONE) {
            return status;
        }
    }
#endif  //BACKUP_FSET

    return status;
    
}

/*
 * Dump the end of the directory
 */

PRIVATE fs_Status
fs_dumpDirNodeEnd(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    uint64              id
)
{
    fs_DumpOperation_t          *obj;
    fs_Status                   status;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Dumping node end - current Id %lld", id));

    obj = &dumpContext->record.u.operation;
    obj->header.headerId = id;
    obj->header.headerSubId = 0;
    obj->header.headerBlocks = 1;
    obj->header.dataExtents = 0;
    obj->len = 0;
    obj->sourceNameLen = 0;
    obj->header.op = FS_DR_DIR_END;
    status = fs_writeTape(context, dumpContext);
    return status;
}

/*
 * Dump pathname to link
 */

PRIVATE fs_Status
fs_dumpLink(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inode            *inode
)
{
    char        *linkPath;
    char         buf[FS_DR_TBSIZE];
    int32       copyBytes, len;
    char        *cp;
    fs_Status   status = FS_ERR_NONE;

    linkPath = fs_getLinkPath(context, dumpContext, inode, &len);
    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    cp = linkPath;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("dumpLink - linkPath %s len %d\n", linkPath, len));
    while (len > 0) {
        if (len < FS_DR_TBSIZE) {
            bzero(buf + len, FS_DR_TBSIZE - len);
            memcpy(buf, cp, len);
            copyBytes = len;
        } else {
            memcpy(buf, cp, FS_DR_TBSIZE);
            copyBytes = FS_DR_TBSIZE;
        }
        status = fs_writeDataTape(context, dumpContext, buf);
        if (status != FS_ERR_NONE) {
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("dumpLink - write data tape returned %d\n",
                      status));
            return status;
        }
        cp += copyBytes;
        len -= copyBytes;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("dumpLink - returning %d\n", status));
    return status;
}

/*
 * Dump the blocks for a given file and starting extent
 */

EXPORTED fs_Status
fs_dumpBlocks(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inode            *inode,
    uint64              start,
    fs_DumpExtent_t     *drExtent,
    uint32              numExtents
)
{

    fs_Buf              *buf;
    uint32              i;
    uint64              offset;
    uint64              numBlocks;
    int                 j;
    fs_Status           status;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("current name %s\n", dumpContext->currentName));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Inode size %lld", inode->dinode.size));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("sdrextent is %p", drExtent));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("dumpBlocks: inode size %lld", inode->dinode.size));
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Number of extents %d", numExtents));
    for (i = 0; i < numExtents; i++) {
        offset = drExtent->offset >> FS_BLOCK_SHIFT;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("dr len %lld", drExtent->len));
        numBlocks = drExtent->len;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("numblocks %lld", numBlocks));
        while (numBlocks > 0) {
            //FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             //           ("dumpBlocks1: offset %lld", offset));
            buf = O_GET_SHARED(inode, offset);
            FS_ASSERT(buf != NULL);
            //FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             //        ("Inode num %lld inode snap num %d buf snap num 0x%llx",
              //          inode->inum, inode->snapnum, buf->snapshots));
            for (j = 0; j < (FS_BLOCK_SIZE / FS_DR_TBSIZE) && numBlocks > 0;
                 j++) {
                status = fs_writeDataTape(context, dumpContext,
                                          &buf->u.data[j * FS_DR_TBSIZE]);
                if (status != FS_ERR_NONE) {
                    fs_bufRelease(context, buf);
                    return status;
                }
                numBlocks--;
            }

            fs_bufRelease(context, buf);
            fs_readAhead(context, inode, offset << FS_BLOCK_SHIFT,
						 FS_BLOCK_SIZE, 1);
            offset++;
        }
        
        drExtent++;
    }
    status = FS_ERR_NONE;
    return status;
    
}


/*
 * Dump the regular file to tape
 */

PRIVATE fs_Status
fs_dumpReg(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inode            *inode,
    char                *name,
    int                 nameLen,
    uint8               type
)
{
    fs_DumpOperation_t          *obj;
    fs_DumpDataExtension_t      *dataObj;
    uint32                      objsize;
    fs_Status                   status;
    uint32                      maxNumExtents = 0;
    uint32                      numExtents = 0;
    uint64                      curOffset = 0, start = 0;
    fs_Emappair                 p;
    fs_Extent                   extent;
    uint64                      len;
    fs_Block                    block;
    fs_DumpExtent_t             *drExtent, *sdrExtent = NULL;
    uint64                      curId;
    uint32                      totLen, availLen;
    char                        *cp, *tp;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Dumping reg - current Id %lld",
                                     dumpContext->currentId));
    obj = &dumpContext->record.u.operation;
    curId = obj->header.headerId = dumpContext->currentId;
    obj->header.headerSubId = 0;
    obj->len = nameLen;
    obj->sourceNameLen = nameLen;
    obj->header.headerBlocks = 1;
    if (nameLen > (FS_DR_TBSIZE - sizeof (obj))) {
        obj->header.headerBlocks++;
    }
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Link Count is %d",
                                     inode->dinode.linkCnt));
    if (inode->dinode.linkCnt == 1) {
        obj->header.op = type;
    } else if (!fs_isInodeCreated(context, dumpContext, inode)) {
        obj->header.op = type;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Creating  %s",
                                         dumpContext->currentName));
        fs_markInodeCreated(context, dumpContext, inode,
                            dumpContext->currentName, 
                            strlen(dumpContext->currentName), 0);
    } else {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Calling hardlink"));
        obj->header.op = FS_DR_CREATE_LINK;
    }
    objsize = sizeof (fs_DumpOperation_t) + nameLen;
    objsize += 7;
    objsize &= ~7;   // align to 8 byte boundary

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("objsize is %d", objsize));
    maxNumExtents =  (FS_DR_TBSIZE - objsize) / sizeof (fs_DumpOperation_t);
    obj->header.dataExtents = 0;
    obj->header.attr.uid = inode->sd->u.nfs.uid;
    obj->header.attr.gid = inode->sd->u.nfs.gid;
    obj->header.attr.sizeDev.size = inode->dinode.size;
    obj->header.attr.atime = inode->dinode.atime;
    obj->header.attr.mtime = inode->dinode.mtime;
    obj->header.attr.ctime = inode->dinode.ctime;
    obj->header.attr.mode = (inode->sd->u.nfs.mode & FS_IMODE_MASK) |
                             (inode->dinode.attrs & FS_IATR_MASK);
    obj->header.attr.simpleAttrs = inode->dinode.userAttr;
    totLen = nameLen + 1;
    cp = name;
    tp = &dumpContext->record.u.dummy[sizeof (fs_DumpOperation_t)];
    availLen = &dumpContext->record.u.dummy[FS_DR_TBSIZE] - tp;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("dumpReg:dumped %s in header", cp));
    if (availLen >= totLen) {
        bzero(tp + nameLen, availLen - nameLen);
        memcpy(tp, cp, nameLen);
    } else {
        memcpy(tp, cp, availLen);
        status = fs_writeTape(context, dumpContext);
        if (status != FS_ERR_NONE) {
            obj->header.flags = 0;
            return status;
        }
        totLen -= availLen;
        tp = dumpContext->record.u.dummy;
        bzero(tp + totLen, FS_DR_TBSIZE - totLen);
        cp += availLen;
        memcpy(tp, cp, totLen);
        tp += totLen;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Write record"));

        /*
         * Fake that we have at least 1 record
         */
        numExtents = 1;
        goto write_record;
    }

    drExtent = (fs_DumpExtent_t *)(dumpContext->record.u.dummy + objsize);
    sdrExtent = drExtent;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("getting extents"));

    if (obj->header.op == FS_DR_CREATE_LINK) {
        numExtents = 1;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("Call dumpLink"));
        goto write_header;
    }
    start = curOffset;
    while (curOffset < inode->dinode.size && numExtents < maxNumExtents) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("dumpRegs1:curOffset is %lld", curOffset));
		
	     //If have been using the cpu since a long time, give others a chance
	    fs_yieldPeriodic(context);
		
        p = fs_emap_lookup(context, inode, curOffset >> FS_BLOCK_SHIFT);
		if (fs_extentLength(p.extent) == 0) {
            curOffset += FS_BLOCK_SIZE;
            continue;
        }
        extent = p.extent;
        len = fs_extentLength(extent);
        block = fs_extentStartBlock(extent);
        if (numExtents < maxNumExtents) {
            drExtent->offset = curOffset;

            /*
             * Store extent length in tape record multiples
             */

            len = MIN(len << FS_BLOCK_SHIFT, inode->dinode.size - curOffset);
            len = (len + FS_DR_TBSIZE - 1) & (~(FS_DR_TBSIZE - 1));
            drExtent->len = (len >> FS_DR_TBSHIFT);
            numExtents++;
//            D("dumpRegs1:drExtent %p offset %lld len %lld", drExtent,
//               drExtent->offset, drExtent->len);
            drExtent++;
            curOffset += len;
        }
    }

write_header:
    obj->header.dataExtents = numExtents;
    /*
     * First write the object record
     */

write_record:
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("writing header"));
    status = fs_writeTape(context, dumpContext);
    if (status != FS_ERR_NONE) {
        return status;
    }


    /*
     * Now copy the extents that we dumped
     */

    if (numExtents == 0) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("num extents 0"));
        return FS_ERR_NONE;
    }

    if (obj->header.op == FS_DR_CREATE_LINK) {
        return fs_dumpLink(context, dumpContext, inode);
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("sdr len = %lld", sdrExtent->len));

    if (sdrExtent != NULL) {
        status = fs_dumpBlocks(context, dumpContext, inode, start, sdrExtent,
                               numExtents);
    }
    if (status != FS_ERR_NONE) {
        return status;
    }

more:
    if (curOffset >= inode->dinode.size) {
        return FS_ERR_NONE;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("more records"));
    start = curOffset;

    /*
     * If there are more blocks to dump, then create extention objects
     */

    numExtents = 0;
    dataObj = &dumpContext->record.u.extension;
    dataObj->header.headerId = curId;
    dataObj->header.headerSubId++;
    dataObj->header.op = FS_DR_DATA;
    objsize = sizeof (fs_DumpDataExtension_t);
	maxNumExtents = FS_DR_MAXEXTENTS;
    drExtent = dataObj->extents;
    sdrExtent = drExtent;
    while (curOffset < inode->dinode.size && numExtents < maxNumExtents) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("dumpRegs2:curOffset is %lld", curOffset));
		
	     //If have been using the cpu since a long time, give others a chance
	    fs_yieldPeriodic(context);
		
        p = fs_emap_lookup(context, inode, curOffset >> FS_BLOCK_SHIFT);
		if (fs_extentLength(p.extent) == 0) {
            curOffset += FS_BLOCK_SIZE;
            continue;
        }
        extent = p.extent;
        len = fs_extentLength(extent);
        block = fs_extentStartBlock(extent);
        if (numExtents < maxNumExtents) {
            drExtent->offset = curOffset;

            /*
             * Store extent length in tape record multiples
             */

            len = MIN(len << FS_BLOCK_SHIFT, inode->dinode.size - curOffset);
            len = (len + FS_DR_TBSIZE - 1) & (~(FS_DR_TBSIZE - 1));
            drExtent->len = (len >> FS_DR_TBSHIFT);
            numExtents++;
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("dumpRegs2:drExtent %p offset %lld len %lld", drExtent,
               drExtent->offset, drExtent->len));
            drExtent++;
            curOffset += len;

        }
    }
    dataObj->header.dataExtents = numExtents;
    /*
     * First write the object record
     */

    status = fs_writeTape(context, dumpContext);
    if (status != FS_ERR_NONE) {
        return status;
    }

    status = fs_dumpBlocks(context, dumpContext, inode, start, sdrExtent,
                           numExtents);
    if (status != FS_ERR_NONE) {
        return status;
    }

    goto more;
}

/*
 * Dump a symlink to tape
 */

PRIVATE fs_Status
fs_dumpSymLink(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inode            *inode,
    char                *name,
    int                 nameLen
)
{
    char                        buf[FS_DR_TBSIZE];
    fs_DumpOperation_t          *obj;
    uint32                      objsize;
    fs_Status                   status;
    fs_DumpExtent_t             *drExtent;
    uint64                      curId;
    uint32                      totLen, availLen, copyLen;
    char                        *cp, *tp;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Dumping Symlink - current Id %lld", dumpContext->currentId));

    if (inode->dinode.size > FS_INODE_INLINE_LEN) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Dumping Symlink as reg - current Id %lld",
                   dumpContext->currentId));
        return fs_dumpReg(context, dumpContext, inode, name, nameLen,
                          FS_DR_CREATE_SYML);
    }
    obj = &dumpContext->record.u.operation;
    curId = obj->header.headerId = dumpContext->currentId;
    obj->header.headerSubId = 0;
    obj->len = nameLen;
    obj->sourceNameLen = nameLen;
    obj->header.headerBlocks = 1;
    if (nameLen > (FS_DR_TBSIZE - sizeof (obj))) {
        obj->header.headerBlocks++;
    }
    if (inode->dinode.linkCnt == 1) {
        obj->header.op = FS_DR_CREATE_SYML;
    } else if (!fs_isInodeCreated(context, dumpContext, inode)) {
        obj->header.op = FS_DR_CREATE_SYML;
        fs_markInodeCreated(context, dumpContext, inode,
                            dumpContext->currentName, 
                            strlen(dumpContext->currentName), 0);
    } else {
        obj->header.op = FS_DR_CREATE_LINK;
    }
    objsize = sizeof (fs_DumpOperation_t) + nameLen;
    objsize += 7;
    objsize &= ~7;   // align to 8 byte boundary
    obj->header.dataExtents = 1;
    obj->header.attr.uid = inode->sd->u.nfs.uid;
    obj->header.attr.gid = inode->sd->u.nfs.gid;
    obj->header.attr.sizeDev.size = inode->dinode.size;
    obj->header.attr.atime = inode->dinode.atime;
    obj->header.attr.mtime = inode->dinode.mtime;
    obj->header.attr.ctime = inode->dinode.ctime;
    obj->header.attr.mode = (inode->sd->u.nfs.mode & FS_IMODE_MASK) |
                             (inode->dinode.attrs & FS_IATR_MASK);
    obj->header.attr.simpleAttrs = inode->dinode.userAttr;
    totLen = nameLen + 1;
    cp = name;
    tp = &dumpContext->record.u.dummy[sizeof (fs_DumpOperation_t)];
    availLen = &dumpContext->record.u.dummy[FS_DR_TBSIZE] - tp;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("dumpSymlink:dumped %s in header", cp));
    if (availLen >= totLen) {
        bzero(tp + nameLen, availLen - nameLen);
        memcpy(tp, cp, nameLen);
    } else {
        memcpy(tp, cp, availLen);
        status = fs_writeTape(context, dumpContext);
        if (status != FS_ERR_NONE) {
            obj->header.flags = 0;
            return status;
        }
        totLen -= availLen;
        tp = dumpContext->record.u.dummy;
        bzero(tp + totLen, FS_DR_TBSIZE - totLen);
        cp += availLen;
        memcpy(tp, cp, totLen);
        tp += totLen;
    }
    
    drExtent = (fs_DumpExtent_t *)(dumpContext->record.u.dummy + objsize);
    drExtent->offset = 0;
    drExtent->len = (inode->dinode.size + FS_DR_TBSIZE - 1) >> FS_DR_TBSHIFT;
    status = fs_writeTape(context, dumpContext);
    if (status != FS_ERR_NONE) {
        return status;
    }

    if (obj->header.op == FS_DR_CREATE_LINK) {
        return fs_dumpLink(context, dumpContext, inode);
    }

    totLen = inode->dinode.size;
    cp = inode->dinode.u.data;
    while (totLen > 0) {
        if (totLen < FS_DR_TBSIZE) {
            bzero(buf + totLen, FS_DR_TBSIZE - totLen);
            memcpy(buf, cp, totLen);
            copyLen = totLen;
        } else {
            memcpy(buf, cp, FS_DR_TBSIZE);
            copyLen = FS_DR_TBSIZE;
        }
        cp     += copyLen;
        totLen -= copyLen;
        status = fs_writeDataTape(context, dumpContext, buf);
        if (status != FS_ERR_NONE) {
            break;
        }
    }
    return status;   
}

/*
 * Dump special nodes
 */

PRIVATE fs_Status
fs_dumpSpecialNode(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inode            *inode,
    char                *name,
    int                 nameLen,
    uint8               type
)
{
    fs_DumpOperation_t          *obj;
    uint32                      objsize;
    fs_Status                   status;
    uint64                      curId;
    uint32                      totLen, availLen;
    char                        *cp, *tp;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Dumping spec node - current Id %lld", dumpContext->currentId));
    obj = &dumpContext->record.u.operation;
    obj->header.headerBlocks = 1;
    if (nameLen > (FS_DR_TBSIZE - sizeof (obj))) {
        obj->header.headerBlocks++;
    }
    curId = obj->header.headerId = dumpContext->currentId;
    obj->header.headerSubId = 0;
    obj->len = nameLen;
    obj->sourceNameLen = nameLen;
    if (inode->dinode.linkCnt == 1) {
        obj->header.op = type;
    } else if (!fs_isInodeCreated(context, dumpContext, inode)) {
        obj->header.op = type;
        fs_markInodeCreated(context, dumpContext, inode,
                            dumpContext->currentName, 
                            strlen (dumpContext->currentName), 0);
    } else {
        obj->header.op = FS_DR_CREATE_LINK;
    }
    objsize = sizeof (fs_DumpOperation_t) + nameLen;
    objsize += 7;
    objsize &= ~7;   // align to 8 byte boundary
    obj->header.dataExtents = 0;
    obj->header.attr.uid = inode->sd->u.nfs.uid;
    obj->header.attr.gid = inode->sd->u.nfs.gid;
    obj->header.attr.sizeDev.size = 0;
    if (FS_I_IS_BLK(inode->dinode.attrs) || FS_I_IS_CHR(inode->dinode.attrs)) {
        obj->header.attr.sizeDev.devNum.major =
            inode->dinode.u.devNum.major;
        obj->header.attr.sizeDev.devNum.minor =
            inode->dinode.u.devNum.minor;
    }
    obj->header.attr.atime = inode->dinode.atime;
    obj->header.attr.mtime = inode->dinode.mtime;
    obj->header.attr.ctime = inode->dinode.ctime;
    obj->header.attr.mode = (inode->sd->u.nfs.mode & FS_IMODE_MASK) |
                             (inode->dinode.attrs & FS_IATR_MASK);
    obj->header.attr.simpleAttrs = inode->dinode.userAttr;
    totLen = nameLen + 1;
    cp = name;
    tp = &dumpContext->record.u.dummy[sizeof (fs_DumpOperation_t)];
    availLen = &dumpContext->record.u.dummy[FS_DR_TBSIZE] - tp;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("dumpSpec:dumped %s in header", cp));
    if (availLen >= totLen) {
        bzero(tp + nameLen, availLen - nameLen);
        memcpy(tp, cp, nameLen);
    } else {
        memcpy(tp, cp, availLen);
        status = fs_writeTape(context, dumpContext);
        if (status != FS_ERR_NONE) {
            obj->header.flags = 0;
            return status;
        }
        totLen -= availLen;
        tp = dumpContext->record.u.dummy;
        bzero(tp + totLen, FS_DR_TBSIZE - totLen);
        cp += availLen;
        memcpy(tp, cp, totLen);
        tp += totLen;
    }
    
    status = fs_writeTape(context, dumpContext);
    return status;   
}


/*
 * Dump a tree
 */

PRIVATE fs_Status
fs_dumpTree(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    fs_Inode            **arg_inode,
    char                *name,
    int                 descent,
    int                 isFset
)
{
	DLL_HDR(fs_DumpTreeContext_t)	dumpTreeContexts;
	fs_DumpTreeContext_t			*dtCtx, *dtCtx2;
    fs_Inumber          tinum;
    uint16              inodeFsetNum;
    fs_Inode            *child = NULL;
	boolean				goingDown, goingUp = FALSE;
    fs_Status           status = FS_ERR_NONE;
	
	//Initialize the dump tree contexts list
	DLL_INIT(&dumpTreeContexts);
	
	//Create initial dump tree context
	dtCtx = fs_ramAlloc(FS_RAM_DUMP_DIRENT, sizeof (fs_DumpTreeContext_t));
	bzero(dtCtx, sizeof (fs_DumpTreeContext_t));
	dtCtx->name		= name;
    dtCtx->nameLen  = strlen(name);
    dtCtx->cp       = dumpContext->currentName;
    dtCtx->cp       += strlen(dumpContext->currentName);
	dtCtx->isFset	= isFset;
	dtCtx->inode	= *arg_inode;
	dtCtx->inum		= dtCtx->inode->inum;
	dtCtx->nodeId   = ++dumpContext->currentId;
	
	do {	
/*printf("***** v %llx s %u start dir inum=%llu (down %u up %d name %s)", 
		context->volId.whole_thing, dumpContext->snapnum, 
		dtCtx->inum, goingDown, goingUp, dtCtx->name);
*/
		goingDown = FALSE;
		inodeFsetNum = dtCtx->inode->dinode.fsetNum;
		
		if (goingUp) {
			goingUp = FALSE;
			goto goUp;
		}	
		
	     //If have been using the cpu since a long time, give others a chance
	    fs_yieldPeriodic(context);
	    FS_ASSERT(fs_lockHeldShare(context, &dtCtx->inode->lock));
	    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Dumping tree  - current Id %lld", dumpContext->currentId));

	    //Dump the directory root object
	    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                 ("Node start - %s", dtCtx->name));

        FS_ASSERT(dtCtx->cp == 
                  &dumpContext->currentName[strlen(dumpContext->currentName)]);
        *(dtCtx->cp++) = '/';
        *(dtCtx->cp) = '\0';           //Keep the string null terminated
        
	    status = fs_dumpDirNodeStart(context, dumpContext, dtCtx->inum,
									 dtCtx->nodeId, dtCtx->name, dtCtx->nameLen,
                                     dtCtx->isFset);
	    if (status != FS_ERR_NONE) {
	        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("going up with status %d", status));
		    child = dtCtx->inode;
			//Free current context
		    fs_ramDealloc(FS_RAM_DUMP_DIRENT, dtCtx);
			//Retrieve upper context			
			DLL_REMOVE_TAIL(&dumpTreeContexts, dtCtx, dumpTreeList);
			if (dtCtx == NULL || dtCtx == DLL_NIL)
				break;
			goingUp = TRUE;
			continue;
	    }
	    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("returned from start"));
        
        status = fs_dumpDirs(context, dumpContext, dtCtx->inode, dtCtx->cp,
                            (fs_DirEntry *)dtCtx->buf, sizeof(dtCtx->buf));
	    if (status != FS_ERR_NONE) {
	        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("going up with status %d", status));
		    child = dtCtx->inode;
			//Free current context
		    fs_ramDealloc(FS_RAM_DUMP_DIRENT, dtCtx);
			//Retrieve upper context			
			DLL_REMOVE_TAIL(&dumpTreeContexts, dtCtx, dumpTreeList);
			if (dtCtx == NULL || dtCtx == DLL_NIL)
				break;
			goingUp = TRUE;
			continue;
	    }

	    /*
	     * Read the directory
	     */
	    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("eof is %d, descent is %d", dtCtx->eof, descent));

#if 0
		dtCtx->cp = &dumpContext->currentName[strlen(dumpContext->currentName)];
	    //*(dtCtx->cp)++ = '/';
        *(dtCtx->cp) = '/';
        dtCtx->cp++;

        /*
         * Keep the string null terminated
         */

        *(dtCtx->cp) = '\0';    
#endif
	    while (!dtCtx->eof && descent) {
            dtCtx->dirp = (fs_DirEntry *)dtCtx->buf;
	        fs_dir_readdir(context, dtCtx->inode, dtCtx->dirp, 
							sizeof(dtCtx->buf), dtCtx->cookie, FALSE,
						   	&dtCtx->numEntries, &dtCtx->eof);
	        fs_inodeRelease(context, dtCtx->inode);

	        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("numEntries is %d", dtCtx->numEntries));
	        for (dtCtx->i = 0; dtCtx->i < dtCtx->numEntries; dtCtx->i++) {
	            tinum = dtCtx->dirp->inode;
	            dtCtx->cookie = dtCtx->dirp->cookie;
				FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("cookie %lld", 
													dtCtx->cookie));
                testSpecialDirectories(dtCtx->dirp);
	            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("found name ++%s++ i%lld", dtCtx->dirp->name, tinum));

                //Make sure current name would not go beyond FS_MAXPATHLEN 
                if ((dtCtx->cp - dumpContext->currentName)
                    + dtCtx->dirp->nameLen >= FS_MAXPATHLEN)   {
                    status = FS_ERR_NAMETOOLONG;
                    break;
                }

	            child = I_GET_SHARED(dumpContext->snapnum, tinum);
	            FS_ASSERT(child != NULL);
               
#if 0
	            strcpy(dtCtx->cp, dtCtx->dirp->name);
                
	            switch (child->dinode.attrs & FS_IFMT_MASK) {

	            case FS_IFMT_DIR:
	                dumpContext->currentId++;
					dtCtx2 = fs_ramAlloc(FS_RAM_DUMP_DIRENT, 
										sizeof (fs_DumpTreeContext_t));
					bzero(dtCtx2, sizeof (fs_DumpTreeContext_t));
					dtCtx2->name	= dtCtx->dirp->name;
					dtCtx2->isFset	= child->dinode.fsetNum != inodeFsetNum;
					dtCtx2->inode	= child;
					dtCtx2->inum	= child->inum;
					dtCtx2->nodeId	= dumpContext->currentId;
					DLL_INSERT_TAIL(&dumpTreeContexts, dtCtx, dumpTreeList);
	                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                             ("going down from %s to %s",
                             dtCtx->name, dtCtx2->name));
					dtCtx = dtCtx2;										  
					//We are now in the context of the new dir
					goingDown = TRUE;
    	            break;

                default:
                    break;
	            case FS_IFMT_SYML:
	                dumpContext->currentId++;
	                status = fs_dumpSymLink(context, dumpContext, child,
	                                     dtCtx->dirp->name);
	                break;

	            case FS_IFMT_REG:
	                dumpContext->currentId++;
	                status = fs_dumpReg(context, dumpContext, child,
	                                    dtCtx->dirp->name, FS_DR_CREATE_FILE);
	                break;

	            case FS_IFMT_SOCK:
	                dumpContext->currentId++;
	                status = fs_dumpSpecialNode(context, dumpContext, child,
	                                    dtCtx->dirp->name, FS_DR_CREATE_SOCK);
	                break;

	            case FS_IFMT_FIFO:
	                dumpContext->currentId++;
	                status = fs_dumpSpecialNode(context, dumpContext, child,
										dtCtx->dirp->name, FS_DR_CREATE_FIFO);
	                break;

	            case FS_IFMT_BLK:
	                dumpContext->currentId++;
	                status = fs_dumpSpecialNode(context, dumpContext, child,
	                                    dtCtx->dirp->name, FS_DR_CREATE_BLOCK);
	                break;

	            case FS_IFMT_CHR:
	                dumpContext->currentId++;
	                status = fs_dumpSpecialNode(context, dumpContext, child,
	                                    dtCtx->dirp->name, FS_DR_CREATE_CHAR);
	                break;

	            default:
	                /*
	                 * Note - We would not be in this path, if there were any
	                 * symbolic links in the filesystem.
	                 */
	                fs_ndmpSendDumpLog(context, dumpContext, status,
	                                   NDMP_LOG_MSG_TYP_INFO, FS_ERR_NONE, NULL,
										"Invalid Inode");
	                continue;
                }
				if (goingDown) break; 
#endif
                if ((child->dinode.attrs & FS_IFMT_MASK) == FS_IFMT_DIR) {
    	            strcpy(dtCtx->cp, dtCtx->dirp->name);
	                dumpContext->currentId++;
					dtCtx2 = fs_ramAlloc(FS_RAM_DUMP_DIRENT, 
										sizeof (fs_DumpTreeContext_t));
					bzero(dtCtx2, sizeof (fs_DumpTreeContext_t));
					dtCtx2->name	= dtCtx->dirp->name;
                    dtCtx2->nameLen = dtCtx->dirp->nameLen - 1;
                    dtCtx2->cp      = dtCtx->cp;
                    dtCtx2->cp      += dtCtx2->nameLen;
					dtCtx2->isFset	= child->dinode.fsetNum != inodeFsetNum;
					dtCtx2->inode	= child;
					dtCtx2->inum	= child->inum;
					dtCtx2->nodeId	= dumpContext->currentId;
					DLL_INSERT_TAIL(&dumpTreeContexts, dtCtx, dumpTreeList);
	                FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
                             ("going down from %s to %s",
                             dtCtx->name, dtCtx2->name));
					dtCtx = dtCtx2;										  
					//We are now in the context of the new dir
					goingDown = TRUE;
    	            break;
	            }
goUp:			
                //Get rid of the child name to handle the next one
	            *dtCtx->cp = '\0';
	            fs_inodeRelease(context, child);
	            if (status != FS_ERR_NONE) {
	                break;
	            }
	            FS_READDIR_ADV_ENTRY_PTR(dtCtx->dirp);
	        }
			if (goingDown) break; 
	        dtCtx->inode = I_GET_SHARED(dumpContext->snapnum, dtCtx->inum);
			if (dtCtx->inode == NULL)	{
				FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("***** inode==NULL & inum=%llu", dtCtx->inum));
				FS_ASSERT(FALSE);
			}

	        if (dtCtx->eof) {
	            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("Eof is set, numEntries is %d", dtCtx->numEntries));
	            break;
	        }
	    }
		if (goingDown) continue; 
	    fs_inodeRelease(context, dtCtx->inode);
	    *(--dtCtx->cp) = '\0';          //Get rid of trailing /

        if (status == FS_ERR_NONE)  {
    	    status = fs_dumpDirNodeEnd(context, dumpContext, dtCtx->nodeId);
        }
	    child = I_GET_SHARED(dumpContext->snapnum, dtCtx->inum);
		
		//Retrieve upper context			
		dtCtx2 = dtCtx;
		DLL_REMOVE_TAIL(&dumpTreeContexts, dtCtx, dumpTreeList);
		FS_TRACE(context,FS_TRACE_DUMP_RESTORE,
                 ("going up from %s back to %s", dtCtx2->name, 
				(dtCtx != NULL && dtCtx != DLL_NIL) ? dtCtx->name : "[END]"));
        
		//Free previous context
	    fs_ramDealloc(FS_RAM_DUMP_DIRENT, dtCtx2);
       
		goingUp = TRUE;
	} while (dtCtx != NULL && dtCtx != DLL_NIL);
	
	*arg_inode = child;
	
    return status;
}



/*
 * Dump a partial tree
 */

PRIVATE fs_Status
fs_dumpPartialTree(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    char                *name
)
{
    char                *p, *c;    
    fs_Inode            *curInode;
    fs_Inode            *parentDir;
    fs_Inumber          parentInum;
    fs_path_t           *path;
    int32               fsetNum = -1;
    fs_Status           status = FS_ERR_NONE;
/*	
elog_enabled = TRUE;
elog_severity = LOG_DEBUG;
elog_appl_data[fs_applicationId].severity = LOG_DEBUG;
elog_appl_data[fs_applicationId].bitmap |= (1 << FS_TRACE_DUMP_RESTORE);
*/    
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
             ("Partial Dump of %s - start", name));
              
    path = fs_makePath(name);
    parentDir = I_GET_SHARED(dumpContext->snapnum, FS_ROOT_INODE);
    parentInum = parentDir->inum;
    p = fs_getNextComponent(path);
    while (p != NULL) {
        fsetNum = parentDir->dinode.fsetNum;
        parentInum  = parentDir->inum;
        curInode = fs_lookupShared(context, context->fileSystem, 
                                    dumpContext->snapnum, parentDir, p);
        
        if (curInode == NULL)   {
            fs_inodeRelease(context, parentDir);
            status = FS_ERR_NOENT;
            break;
        }
        c = p;
        
        // Proceed with the next component in the path                 
        FS_ASSERT(fs_lockHeldShare(context, &curInode->lock));
        p = fs_getNextComponent(path);
        if (p != NULL) {
            // If the current parent is not a directory then path is invalid
            if (!FS_I_IS_DIR(curInode->dinode.attrs)) {
                fs_inodeRelease(context, curInode);
                status = FS_ERR_INVAL;
                break;
            }
            parentDir   = curInode;
            curInode    = NULL;
        }
    }
   
    if (status == FS_ERR_NONE) {        
        //Only allow partial dump of a directory
        if (curInode->dinode.attrs & FS_IFMT_DIR)   {
            //Store inum of the root of the partial tree
            dumpContext->partialTreeRootInum = curInode->inum;
            
            //Links should be referenced from here to be restored properly
            dumpContext->currentName[0] = '\0';
            //strcpy(dumpContext->currentName, name);

            status = fs_dumpTree(context, dumpContext, &curInode, "/"/*c*/, 1,
                                 fsetNum != curInode->dinode.fsetNum);
       
            fs_ndmpFlushDirHistory(context, dumpContext);
            if (status == FS_ERR_NONE && dumpContext->abort == 0) {
                status = fs_partialDumpDirNodeFileHistory(context, dumpContext,
                                                          curInode);
                fs_ndmpFlushNodeHistory(context, dumpContext);
            }
        } else {
            status = FS_ERR_NOTDIR;
        }
        fs_inodeRelease(context, curInode);
    }        
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
             ("Partial Dump of %s - end (status %u)", name, status));
    fs_destroyPath(path);
    return status;
}



/*
 * Write the tape start record
 */

EXPORTED fs_Status
fs_writeStartRecord(
    fs_Context              *context,
    fs_DumpContext_t        *dumpContext,
    uint8                   backupType,
    uint8                   type,
    fs_Time                 dumpDate,
    fs_Time                 prevDumpDate
)
{
    fs_DumpStartHeader_t        *start;
    fs_Status                   status;
    uint32                      sum = 0;
    uint64                      curId;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_ASSERT(type == FS_DR_START || type == FS_DR_END);
    FS_ASSERT(backupType == FS_DR_FULL || backupType == FS_DR_INCR 
           || backupType == FS_DR_PARTIAL_FULL);
    curId = dumpContext->currentId++;
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Dumping write start  - current Id %lld",
                                     dumpContext->currentId));
    start = &dumpContext->record.u.start;
    bzero(dumpContext->record.u.dummy, FS_DR_TBSIZE);
    if (type == FS_DR_START) {
        dumpContext->currentId = 0;
        start->header.op = FS_DR_TAPE_START;
    } else {
        start->header.op = FS_DR_TAPE_END;
    }

    start->header.headerId = curId;
    start->header.headerSubId = 0;
    start->magic = FS_DR_MAGIC;
    start->type = backupType;
    start->flags = type;
    start->level = dumpContext->level;
    start->pathSeparator = '/';
    start->dumpTime = dumpDate;
    start->prevDumpTime = prevDumpDate;
    start->snapshot = dumpContext->snapnum;
    start->prevSnapshot = dumpContext->prevSnapNum;
    sum += (uint32)(FS_DR_MAGIC + type + start->pathSeparator +
                    dumpDate.secs + dumpDate.nsecs + prevDumpDate.secs +
                    prevDumpDate.nsecs);
    start->checkSum = FS_DR_CHECKSUM - sum;
    start->header.version = FS_DR_VERSION;
    status = fs_writeTape(context, dumpContext);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Dumped write start  - current Id %lld",
                                     dumpContext->currentId));
    return status;   
}

/*
 * Perform a tree style full dump for a subdirectory passed as arguments.
 * If none are specified, then start with root
 */

EXPORTED fs_Status
fs_treeFullDump(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext
)
{
    fs_Inode            *root;
    fs_Status           status;
    int                 len;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Dumping full/partial tree  - current Id %lld",
                                     dumpContext->currentId));

    if (dumpContext->cwd == NULL
     || ((len = strlen(dumpContext->cwd)) == 0
     || (len == 1 && (dumpContext->cwd[0]=='/' || dumpContext->cwd[0]=='\\'))))
    {
        /*
         * We are dumping the root of the fileystem and possibly its
         * descendents
         */

        status = fs_writeStartRecord(context, dumpContext, FS_DR_FULL, 
                                     FS_DR_START, dumpContext->dumpTime,
                                     dumpContext->prevDumpTime);
        if (status != FS_ERR_NONE) {
            return status;
        }
        root = I_GET_SHARED(dumpContext->snapnum, FS_ROOT_INODE);
        dumpContext->currentName[0] = '\0';
        status = fs_dumpTree(context, dumpContext, &root, "/", 1, 1); 
        fs_ndmpFlushDirHistory(context, dumpContext);
        if (status == FS_ERR_NONE && dumpContext->abort == 0) {
                fs_doFullFileHistories(context, dumpContext);
                fs_ndmpFlushNodeHistory(context, dumpContext);
        }
        fs_inodeRelease(context, root);
        if (status != FS_ERR_NONE) {
            return status;
        }
    } else {
            
        /*
         * We are being asked to dump a sub tree of the filesystem
         */

        if (len >= FS_MAXPATHLEN)
            return FS_ERR_NAMETOOLONG;  
        
        status = fs_writeStartRecord(context, dumpContext, FS_DR_PARTIAL_FULL, 
                                     FS_DR_START, dumpContext->dumpTime,
                                     dumpContext->prevDumpTime);
        if (status != FS_ERR_NONE) {
            return status;
        }

        status = fs_dumpPartialTree(context, dumpContext, dumpContext->cwd);
        if (status != FS_ERR_NONE) {
            return status;
        }
        
    }
    dumpContext->currentId++;
    status = fs_writeStartRecord(context, dumpContext, FS_DR_FULL, FS_DR_END,
                                 dumpContext->dumpTime,
                                 dumpContext->prevDumpTime);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, 
              ("completed tree full/partial dump "));
    return status;
}

/*
 * Cleanup the dump context
 */

EXPORTED void
fs_cleanDumpContext(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    boolean             isException
)
{
    int                 i;
    fs_dumpRenameLink_t *entry;
    fs_DumpNameList_t   *nameEntry = DLL_NIL;

    /*
     * If have been using the cpu since a long time, give others a chance
     */

	if (!isException) {
			fs_yieldPeriodic(context);
	}

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("cleaning context   - current Id %lld",
                                     dumpContext->currentId));

    if (dumpContext->cwd) {
        fs_ramDealloc(FS_RAM_DUMP_NAME, dumpContext->cwd);
    }
    if (dumpContext->label) {
        fs_ramDealloc(FS_RAM_DUMP_NAME, dumpContext->label);
    }
    if (dumpContext->prevSnapName) {
        fs_ramDealloc(FS_RAM_DUMP_NAME, dumpContext->prevSnapName);
    }
    for (i = 0; i < 1023; i++) {
        DLL_REMOVE_HEAD(&dumpContext->renames[i], entry, link);
        if (entry != DLL_NIL && entry) {
            fs_ramDealloc(FS_RAM_DUMP_RENAMELINK, entry);
        }
    }
    for (i = 0; i < 127; i++) {
        DLL_REMOVE_HEAD(&dumpContext->links[i], entry, link);
        if (entry != DLL_NIL && entry) {
            fs_ramDealloc(FS_RAM_DUMP_RENAMELINK, entry);
        }
    }

    do {
        DLL_REMOVE_HEAD(&dumpContext->nameList, nameEntry, link);
        if (nameEntry != DLL_NIL && nameEntry) {
            fs_ramDealloc(FS_RAM_DUMP_NAME, nameEntry->name);
            fs_ramDealloc(FS_RAM_DUMP_NAMEENTRY, nameEntry);
        }
    } while (nameEntry != DLL_NIL && nameEntry);

    for (i = 0; i < FS_DUMP_TAPE_IO_BUFFERS; i++) {
            if (dumpContext->queueHdrs[i].ioBuf != NULL) {
                    fs_ramDealloc(FS_RAM_DUMP_IOBUF,
                                    dumpContext->queueHdrs[i].ioBuf);
                    dumpContext->queueHdrs[i].ioBuf = NULL;
            }
    }

    /*
     * Now remove the dumpContext from the filesystem
     */

    if (!isException)
        fs_lockExclusive(context, &context->fileSystem->dumpContextsLock, NULL);
    DLL_REMOVE(&context->fileSystem->dumpContexts, dumpContext, dumpList);
    if (!isException)
        fs_unlockExclusive(context, &context->fileSystem->dumpContextsLock);
}

EXPORTED fs_Status
fs_dump(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext
)
{

    fs_Status           status;
    int32               rc;
    uint32              now;
    char                snapName[FS_DUMP_MAX_SNAP_LEN];
    esm_event_t         event;

    /*
     * If have been using the cpu since a long time, give others a chance
     */
    fs_yieldPeriodic(context);
    now = fs_getTimeOfDaySecs();
    sprintf(snapName, "dump_snap_%ul", now);
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("New snapshot is %s", snapName));
    rc = fs_snapCreate(context, snapName);
    if (rc != NFX_OK) { 

        /*
         * log the error with ndmp and return failure
         */
        fs_ndmpSendDumpLog(context, dumpContext, FS_ERR_NXIO,
                           NDMP_LOG_MSG_TYP_ERR, FS_ERR_NONE, NULL,
                            "Cannot create snapshot");
        return FS_ERR_NXIO;
    }
    dumpContext->snapnum = fs_getSnapNum(context, snapName);
    if (dumpContext->snapnum == 0) {
        /*
         * log the error with ndmp and return failure
         */
        fs_ndmpSendDumpLog(context, dumpContext, FS_ERR_IO,
                           NDMP_LOG_MSG_TYP_ERR, FS_ERR_NONE, NULL,
                            "Cannot create snapshot");
        return FS_ERR_IO;
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("snapnum = %d", dumpContext->snapnum));

retry:
    if (dumpContext->level > 0) {

#ifdef FS_DEBUG
		eee_rtc(&dumpContext->startTime);
#endif
        /*
         * Incremental dump
         */
        dumpContext->prevSnapNum = fs_getSnapNum(context,
                                                 dumpContext->prevSnapName);
        if (dumpContext->prevSnapNum == -1) {
            dumpContext->level = 0;
            /*
             * Log the fact that we are converting incremental to full
             * dump
             */
            fs_ndmpSendDumpLog(context, dumpContext, FS_ERR_NONE,
                               NDMP_LOG_MSG_TYP_INFO, FS_ERR_NONE, NULL,
                                "Cannot find previous snapshot - converting"                                    " to full dump");
            goto retry;
        }
        rc = fs_ndmpSetDumpDate(context, dumpContext, snapName, now);
        if (rc != NFX_OK) {

            /*
             * log the error with ndmp and return failure
             */
            fs_ndmpSendDumpLog(context, dumpContext, FS_ERR_IO,
                               NDMP_LOG_MSG_TYP_ERR, FS_ERR_NONE, NULL,
                                "Cannot set dump dates");
            return FS_ERR_IO;
        }
        fs_snapLockSnap(context, snapName);
        fs_createFileHistories(context, dumpContext);
        status = fs_incrementalDump(context, dumpContext);
        fs_snapUnLockSnap(context, snapName);
    } else {
        /*
         * Full dump
         */

#ifdef FS_DEBUG
		eee_rtc(&dumpContext->startTime);
#endif
        rc = fs_ndmpSetDumpDate(context, dumpContext, snapName, now);
        if (rc != NFX_OK) {

            /*
             * log the error with ndmp and return failure
             */
            fs_ndmpSendDumpLog(context, dumpContext, FS_ERR_IO,
                               NDMP_LOG_MSG_TYP_ERR, FS_ERR_NONE, NULL,
                                "Cannot set dump dates");
            return FS_ERR_IO;
        }

        /*
         * Fork dump threads
         */

        dumpContext->threadStatus = FS_ERR_NONE;
        dumpContext->workerThread = fs_threadFork(context,
                            fs_objectFullDumpThread,
                              TRUE, TRUE, FS_DAEMON_TYPE_DUMP,
                              (void *)dumpContext);

        fs_snapLockSnap(context, snapName);
        status = fs_treeFullDump(context, dumpContext);

        /*
         * Ask all the threads to die
         */

        NFX_SPIN_LOCK(&dumpContext->spinLock);
        //FS_ASSERT(dumpContext->head == dumpContext->tail);
        while (dumpContext->head != dumpContext->tail) {
                NFX_SPIN_UNLOCK(&dumpContext->spinLock);
                fs_sleep(context, 0);
                NFX_SPIN_LOCK(&dumpContext->spinLock);
        }
        FS_ASSERT(dumpContext->head == dumpContext->tail);
        dumpContext->head = dumpContext->tail = 0;
        dumpContext->inumArray[dumpContext->tail++] = FS_DUMP_KILL_INUM;
        dumpContext->killedThreads = 1;
        if (dumpContext->workWaiter) {
                event.event = FS_THREAD_EVENT_SCAN;
                thread_event(dumpContext->workWaiter, &event);
                dumpContext->workWaiter = 0;
        }
        NFX_SPIN_UNLOCK(&dumpContext->spinLock);  //Anh's fix
        fs_yield(context, &event, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_IO_MW);
        FS_ASSERT(event.event == FS_THREAD_EVENT_WORK_DONE);
        //NFX_SPIN_UNLOCK(&dumpContext->spinLock);
        fs_snapUnLockSnap(context, snapName);
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("completed full dump "));
    }
    if (strlen(snapName) && (dumpContext->abort || status != FS_ERR_NONE))
    {
        /*
         * Remove the snap shot after we are done. Only for
         * aborted cases.
         */
        fs_snapRemove(context, snapName);
    }

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("Returning from fs_dump %d", status));
    return status;
    
}

/*
 * I/O Routines
 */

EXPORTED fs_Status
fs_writeTape(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext
)
{
#ifndef FS_STANDALONE
    fs_Status           status = FS_ERR_NONE;
#endif
    if (dumpContext->abort) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("Failing write Tape because of abort"));
        return FS_ERR_IO;
    }
#ifndef FS_STANDALONE

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    status = fs_ndmpWriteTape(context, dumpContext, dumpContext->record.u.dummy);

#if 0
    /*
     * Wait for the i/o done notification
     */

    dumpContext->tapeWaiter = context->smHandle;
    fs_yield(context, &event, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_IO_MW);
    if (event.event != FS_THREAD_EVENT_DUMP_REST_IOCOMPLETE) {
        FS_ASSERT(0);
    }
    
    status = (fs_Status)(event.p1.u.w);
#endif
#ifdef FS_DEBUG
    if (status != FS_ERR_NONE) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                ("ndmp (writeTape) returned following status %d", status));
    }
#endif
    return status;
#else
    (void)write(dumpContext->fd, dumpContext->record.u.dummy, FS_DR_TBSIZE);
    return FS_ERR_NONE;
#endif
}

EXPORTED fs_Status
fs_writeDataTape(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext,
    char                *buf
)
{
#ifndef FS_STANDALONE
    fs_Status           status = FS_ERR_NONE;
#endif
    if (dumpContext->abort) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("Failing write data Tape because of abort"));
        return FS_ERR_IO;
    }
#ifndef FS_STANDALONE

    /*
     * If have been using the cpu since a long time, give others a chance
     */

    fs_yieldPeriodic(context);
    status = fs_ndmpWriteTape(context, dumpContext, buf);

#if 0
    /*
     * Wait for the i/o done notification
     */

    dumpContext->tapeWaiter = context->smHandle;
    fs_yield(context, &event, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_IO_MW);
    if (event.event != FS_THREAD_EVENT_DUMP_REST_IOCOMPLETE) {
        FS_ASSERT(0);
    }
    
    status = (fs_Status)(event.p1.u.w);
#endif
#ifdef FS_DEBUG
    if (status != FS_ERR_NONE) {
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("ndmp (writeDataTape)returned following status %d", status));
    }
#endif
    return status;
#else
    (void)write(dumpContext->fd, buf, FS_DR_TBSIZE);
    return FS_ERR_NONE;
#endif
}

int32
fs_fsstatThread(
    fs_Context      *context
)
{
    fs_FileSystem   *fileSystem;
    fs_MsgFsstatVolResp_t  *resp;
    fs_MsgFsstatVolReq_t   *req;
    int32           rc;

    resp = (fs_MsgFsstatVolResp_t *) context->respHdr;
    req = (fs_MsgFsstatVolReq_t *) context->reqHdr;

    fileSystem = fs_fileSystemForVolume(context->volId);
    if (fileSystem != NULL)
    {
        rc = NFX_OK;

		resp->volId = context->volId;
        resp->stat.fsBlockSize = fileSystem->super->blkSize;
        resp->stat.numBlks = fileSystem->super->numBlks;
        resp->stat.numBlksAvail = resp->stat.numBlksAvailUser  =
            fileSystem->super->numBlks - fileSystem->super->blksUsed;

        resp->stat.numInodes = fileSystem->params.maxInodeNumber;
        resp->stat.numInodesAvail =
        resp->stat.numInodesAvailUser =
								fileSystem->params.maxInodeNumber -
								fileSystem->super->inodesUsed;
    }
    else
    {
        rc = NFX_ERR;
    }


    return rc;

} /* fs_fsstatThread */

/*
 *  Is filesystem doing dump or restore
 */

EXPORTED
boolean fs_isBackupRestoreInProgress(
				fs_Context	*context)
{
        fs_lockShared(context, &context->fileSystem->dumpContextsLock, NULL);
		if (!DLL_IS_EMPTY(&context->fileSystem->dumpContexts)) {
				fs_unlockShared(context,
                                &context->fileSystem->dumpContextsLock);
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                                ("Backup in Progress"));
				return TRUE;
		}
        fs_unlockShared(context, &context->fileSystem->dumpContextsLock);
        fs_lockShared(context, &context->fileSystem->restoreContextsLock, NULL);
		if (!DLL_IS_EMPTY(&context->fileSystem->restoreContexts)) {
				fs_unlockShared(context,
                                &context->fileSystem->restoreContextsLock);
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                                ("Restore in Progress"));
				return TRUE;
		}
        fs_unlockShared(context, &context->fileSystem->restoreContextsLock);
		return FALSE;
}


/*
 * Walk through all the inodes and dump their attributes
 */

EXPORTED void
fs_fullDumpNodeFileHistoryThread(
    esm_handle_t    smHandle,
    void            *arg
)
{
        fs_Context          *context = (fs_Context *)arg;
        fs_fullDumpFileHistoryContext_t *dumpDirContext =
                   (fs_fullDumpFileHistoryContext_t *)context->extraArg;
        fs_DumpContext_t    *dumpContext = dumpDirContext->dumpContext;
        uint64              startInum = dumpDirContext->startInum;
        uint64              endInum = dumpDirContext->endInum;
        uint64              inum;
        fs_Inode            *inode;


#if 0
        fs_ramDealloc(FS_RAM_FORK_CONTEXT, forkContext);
#endif
        fs_ramDealloc(FS_RAM_FULL_FH_CONTEXT, dumpDirContext);

        for (inum = startInum; inum < endInum; inum++) {
            /*
             * TBD Sleep here irrespective for every inode. Calling
             * Yield Periodic does not work as inode get code does
             * not call yield periodic and if we don't sleep but we
             * are close to the boundary of time we can use before
             * we watchdog reset, then we panic.
             *
             * Eventually experiment with number of inodes we can
             * get before sleeping and sleep after processing so
             * many inodes
             */
            fs_sleep(context, 0);
            if (dumpContext->abort) {
                    break;
            }
            if (inum < FS_MAX_RESERVED_INODE && inum != FS_ROOT_INODE) {
                    continue;
            }
            inode = I_GET_SHARED(dumpContext->snapnum, inum);
            if (!inode) {
                    continue;
            }
            if (FS_I_IS_QUOT(inode->dinode.attrs)) {
                    fs_inodeRelease(context, inode);
                    continue;
            }
            
            if (inode->dinode.piu.parentInum == FS_TEMP_DUMP_SWAP) {
                    fs_inodeRelease(context, inode);
                    continue;
            }
#ifndef FS_STANDALONE
            fs_ndmpDumpNodeHistory(context, dumpContext, inode);
#endif
            fs_inodeRelease(context, inode);
        }
        fs_threadTerminate(context);
        NFX_SPIN_LOCK(&dumpContext->spinLock);
        dumpContext->activeThreads--;
        NFX_SPIN_UNLOCK(&dumpContext->spinLock);
        if (dumpContext->activeThreads == 0) {
                esm_event_t event;
                event.event = FS_THREAD_EVENT_DUMP_FH_COMPLETE;
                thread_event(dumpContext->smHandle, &event);
        }
}

/*
 *  Send Dir file histories followed by nodes. Veritas does not
 *  like mixing of dirs and nodes
 */

#define FS_DUMP_FH_THREADS 5
PRIVATE void
fs_doFullFileHistories(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext
)
{
        uint64  maxInum = context->fileSystem->params.maxInodeNumber;
        uint32  numThreads = FS_DUMP_FH_THREADS;
        uint64 inumPerThread = maxInum / numThreads;
        int     i = 0;
        uint64  startInum = FS_ROOT_INODE; //FS_MAX_RESERVED_INODE;
        fs_fullDumpFileHistoryContext_t *ctx;
        esm_event_t event;

        NFX_SPIN_LOCK(&dumpContext->spinLock);
        dumpContext->activeThreads = numThreads;
        NFX_SPIN_UNLOCK(&dumpContext->spinLock);
        i = 0;
        for (; i < numThreads; i++) {
                ctx = fs_ramAlloc(FS_RAM_FULL_FH_CONTEXT,
                                  sizeof (fs_fullDumpFileHistoryContext_t));
                FS_ASSERT(ctx != NULL);
                ctx->dumpContext = dumpContext;
                ctx->startInum = startInum;
                ctx->endInum = startInum + inumPerThread;
                fs_threadFork(context, fs_fullDumpNodeFileHistoryThread,
                              TRUE, TRUE, FS_DAEMON_TYPE_DUMP,
                              (void *)ctx);
                startInum += inumPerThread;
        }
        fs_yield(context, &event, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_ADMIN);
        FS_ASSERT(event.event = FS_THREAD_EVENT_DUMP_FH_COMPLETE);
}
#define INODES_PER_BLOCK    32
#define INODES_PER_BLOCK_SHIFT    5
PRIVATE fs_Status
fs_dumpDirs(
        fs_Context          *context,
        fs_DumpContext_t    *dumpContext,
        fs_Inode            *root,
        char                *pathEnd,
        fs_DirEntry         *readirBuf,
        uint32              readirBufLen
)
{
        fs_DirEntry *dirp;
        int     eof = 0, i/*, j*/;
        uint32  numEntries;
        uint64  cookie = 0;
        fs_Inumber  inum = root->inum;
        fs_Status   status = FS_ERR_NONE;
        esm_event_t event;
        fs_Inode    *inode;
        uint64      curBlock;
//        fs_DirEntry *curDirp;
        char        *cp = pathEnd;

        curBlock = 0;
        while (!eof) {
            dirp = readirBuf;
            fs_yieldPeriodic(context);
	        fs_dir_readdir(context, root, dirp, readirBufLen,
                            cookie, FALSE, &numEntries, &eof);
                //fs_inodeRelease(context, root);
                fs_yieldPeriodic(context);
#if 1
                NFX_SPIN_LOCK(&dumpContext->spinLock);
                dumpContext->head = 0;
                dumpContext->tail = 0;
                for (i = 0; i < numEntries; i++) {
                        cookie = dirp->cookie;
                        if (strcmp(dirp->name, ".snapshot") == 0) {
                            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                                     ("found snapshot"));
                            FS_READDIR_ADV_ENTRY_PTR(dirp);
                            continue;
                        }
                        dumpContext->inumArray[dumpContext->tail] =
                                dirp->inode;
                        dumpContext->tail++;
                        if (dumpContext->tail >= FS_DUMP_MAX_INUM_ARRAY) {
                                            break;
                        }
                        FS_READDIR_ADV_ENTRY_PTR(dirp);
                }
                if (dumpContext->workWaiter) {
                        event.event = FS_THREAD_EVENT_SCAN;
                        thread_event(dumpContext->workerThread, &event);
                        dumpContext->workWaiter = 0;
                }
                NFX_SPIN_UNLOCK(&dumpContext->spinLock);
                dirp = readirBuf;
                cookie = 0;
#endif
                for (i = 0; i < numEntries; i++) {
                        cookie = dirp->cookie;
                        if (strcmp(dirp->name, ".snapshot") == 0) {
                            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                                     ("found snapshot"));
                            FS_READDIR_ADV_ENTRY_PTR(dirp);
                            continue;
                        }

#ifndef FS_STANDALONE
                        fs_ndmpDumpDirHistory(context, dumpContext, inum, 
                                                dirp->inode, dirp->name,
                                                dirp->nameLen - 1);
#endif
                        if (strcmp(dirp->name, ".") == 0 ||
                            strcmp(dirp->name, "..") == 0) {
                            FS_READDIR_ADV_ENTRY_PTR(dirp);
                            continue;
                        }

#if 0
                        if ((dirp->inode >> INODES_PER_BLOCK_SHIFT) != 
                                        curBlock) {
                                curBlock = (dirp->inode >>
                                                INODES_PER_BLOCK_SHIFT);
                                NFX_SPIN_LOCK(&dumpContext->spinLock);
                                dumpContext->head = 0;
                                dumpContext->tail = 0;
                                curDirp = dirp;
                                for (j = i; j < numEntries; j++) {
                                        dumpContext->inumArray[
                                            dumpContext->tail] =
                                            curDirp->inode;
                                    dumpContext->tail++;
                                    if ((curDirp->inode >>
                                        INODES_PER_BLOCK_SHIFT) != curBlock) {
                                            break;
                                    }
                                    if (dumpContext->tail >=
                                        dumpContext->FS_DUMP_MAX_INUM_ARRAY) {
                                            break;
                                    }
                                    FS_READDIR_ADV_ENTRY_PTR(curDirp);

                                }
                                if (dumpContext->workWaiter) {
                                        event.event = FS_THREAD_EVENT_SCAN;
                                        thread_event(dumpContext->workerThread,
                                                     &event);
                                        dumpContext->workWaiter = 0;
                                }
                                NFX_SPIN_UNLOCK(&dumpContext->spinLock);
                        }
#endif

                        //Check current name would not go beyond FS_MAXPATHLEN 
                        if ((cp - dumpContext->currentName) + dirp->nameLen 
                              >= FS_MAXPATHLEN) {
                            status = FS_ERR_NAMETOOLONG;
                            break;
                        }
                        inode = I_GET_SHARED(dumpContext->snapnum,
                                             dirp->inode);

                        strcpy(cp, dirp->name);
                        switch (inode->dinode.attrs & FS_IFMT_MASK) {
                        case FS_IFMT_DIR:
                                break;
                        case FS_IFMT_SYML:
                            dumpContext->currentId++;
                            status = fs_dumpSymLink(context, dumpContext,
                                            inode, dirp->name, dirp->nameLen-1);
                            break;

                        case FS_IFMT_REG:
                            dumpContext->currentId++;
                            status = fs_dumpReg(context, dumpContext, inode,
                                                dirp->name, dirp->nameLen - 1,
                                                FS_DR_CREATE_FILE);
                            break;

                        case FS_IFMT_SOCK:
                            dumpContext->currentId++;
                            status = fs_dumpSpecialNode(context, dumpContext,
                                    inode, dirp->name, dirp->nameLen - 1,
                                    FS_DR_CREATE_SOCK);
                            break;

                        case FS_IFMT_FIFO:
                            dumpContext->currentId++;
                            status = fs_dumpSpecialNode(context, dumpContext,
                                    inode, dirp->name, dirp->nameLen - 1,
                                    FS_DR_CREATE_FIFO);
                            break;

                        case FS_IFMT_BLK:
                            dumpContext->currentId++;
                            status = fs_dumpSpecialNode(context, dumpContext,
                                    inode, dirp->name, dirp->nameLen - 1,
                                    FS_DR_CREATE_BLOCK);
                            break;

                        case FS_IFMT_CHR:
                            dumpContext->currentId++;
                            status = fs_dumpSpecialNode(context, dumpContext,
                                    inode, dirp->name, dirp->nameLen - 1,
                                    FS_DR_CREATE_CHAR);
                            break;

                        default:
                            fs_ndmpSendDumpLog(context, dumpContext,
                                            FS_ERR_INVAL, NDMP_LOG_MSG_TYP_INFO,
                                            FS_ERR_NONE, NULL, "Invalid Inode");
                            break;
                        }

                        /*
                         * Release the inode before bailing
                         */

                        fs_inodeRelease(context, inode);
                        if (status != FS_ERR_NONE) {
                                break;
                        }

                        FS_READDIR_ADV_ENTRY_PTR(dirp);
                }

                if (status != FS_ERR_NONE) {
                        break;
                }
        }

        return status;
}

EXPORTED void
fs_objectFullDumpThread(
    esm_handle_t    handle,
    void            *arg
)
{
        fs_Context          *context = (fs_Context *)arg;
        fs_DumpContext_t    *dumpContext =
                (fs_DumpContext_t *)context->extraArg;
        fs_Inode    *inode;
        fs_Inumber  inum;
        esm_event_t event;

more_work:
        NFX_SPIN_LOCK(&dumpContext->spinLock);
        while (dumpContext->head == dumpContext->tail) {
#ifdef FS_DEBUG
                dumpContext->workerThreadsPaused++;
#endif
                dumpContext->workWaiter = handle;
                NFX_SPIN_UNLOCK(&dumpContext->spinLock);
                fs_yield(context, NULL, FS_YIELD_TIME_FOREVER,
                         FS_YIELD_TYPE_ADMIN);
                //fs_sleep(context, 0);
                NFX_SPIN_LOCK(&dumpContext->spinLock);
        }
        inum = dumpContext->inumArray[dumpContext->head];
        dumpContext->head++;
#ifdef FS_DEBUG
        dumpContext->objectsProcessedByThreads++;
#endif
        if (dumpContext->head >= FS_DUMP_MAX_INUM_ARRAY) {
                dumpContext->head = 0;
        }

        if (inum == FS_DUMP_KILL_INUM) {
                dumpContext->killedThreads--;
                if (dumpContext->killedThreads == 0) {
                    event.event =  FS_THREAD_EVENT_WORK_DONE;
                    thread_event(dumpContext->smHandle, &event);
                }
                NFX_SPIN_UNLOCK(&dumpContext->spinLock);
                printf("Thread dieing\n");
                fs_threadTerminate(context);
                return;
        }
        NFX_SPIN_UNLOCK(&dumpContext->spinLock);

        inode = I_GET_SHARED(dumpContext->snapnum, inum);
        if (inode == NULL) {
                goto more_work;
        }
        if (FS_I_IS_DIR(inode->dinode.attrs)) {
            //fs_dir_readdir_firstblock_async(context,inode);
        } else if (FS_I_IS_REG(inode->dinode.attrs)) {
                uint32  maxIo = MIN(inode->dinode.size/FS_BLOCK_SIZE,
                                    8);
            (void)fs_bufGetPageAsync(context, inode, 0, maxIo,
                                     FS_BUFLIST_META);
        }
        fs_inodeRelease(context, inode);
        while (context->fileSystem->ioPendingCount >= 
                        context->fileSystem->ioMaxReadsPending) {
                fs_sleep(context, 0);
        }

        goto more_work;
}




/*
 * Send partial tree file histories
 */

typedef struct fs_partialDumpDirNodeFileHistoryContext_t    {
    DLL_ENTRY(struct fs_partialDumpDirNodeFileHistoryContext_t) list;
    fs_Inumber  rootInum;
    fs_Inode    *root;
    fs_DirEntry *dirents;
    fs_DirEntry *dirp;
    fs_Inode    *inode;
    uint32      numEntries;
    int         eof;
    int         i;
    uint64      cookie;
    char        buf[FS_BLOCK_SIZE];
} fs_partialDumpDirNodeFileHistoryContext;


PRIVATE fs_Status
fs_partialDumpDirNodeFileHistory(
        fs_Context          *context,
        fs_DumpContext_t    *dumpContext,
        fs_Inode            *root
)
{
    DLL_HDR(fs_partialDumpDirNodeFileHistoryContext) contexts;
    fs_partialDumpDirNodeFileHistoryContext *ctx, *ctx2;    
    uint64      snapnum = root->snapnum;
    fs_Status   status = FS_ERR_NONE;
    boolean     goingDown;

    //Initialize the contexts list
    DLL_INIT(&contexts);

    //Create initial context
    ctx = fs_ramAlloc(FS_RAM_DUMP_DIRENT, 
                      sizeof(fs_partialDumpDirNodeFileHistoryContext));
    ctx->rootInum   = root->inum;
    ctx->root       = root;
    ctx->eof        = 0;
    ctx->cookie     = 0;
    ctx->dirents    = (fs_DirEntry*)ctx->buf;
    
    //Send node history for the root of the partial tree
#ifndef FS_STANDALONE
    fs_ndmpDumpNodeHistory(context, dumpContext, root);
#endif
                
    do {
        goingDown = FALSE;

        while (!ctx->eof) {
            ctx->dirp = ctx->dirents;
            fs_yieldPeriodic(context);
	        fs_dir_readdir(context, ctx->root, ctx->dirp, FS_BLOCK_SIZE,
                            ctx->cookie, FALSE, &ctx->numEntries, &ctx->eof);
            fs_inodeRelease(context, ctx->root);
            fs_yieldPeriodic(context);
            for (ctx->i = 0; ctx->i < ctx->numEntries; ctx->i++) {
                ctx->cookie = ctx->dirp->cookie;
                         
                testSpecialDirectories(ctx->dirp);

                ctx->inode = I_GET_SHARED(snapnum, ctx->dirp->inode);
                FS_ASSERT(ctx->inode != NULL);
#ifndef FS_STANDALONE
                fs_ndmpDumpNodeHistory(context, dumpContext, ctx->inode);
#endif
                if ((ctx->inode->dinode.attrs & FS_IFMT_MASK) == FS_IFMT_DIR)  
                {
                    ctx2 = fs_ramAlloc(FS_RAM_DUMP_DIRENT,
                              sizeof(fs_partialDumpDirNodeFileHistoryContext));
                    ctx2->rootInum  = ctx->dirp->inode;
                    ctx2->root      = ctx->inode;
                    ctx2->eof       = 0;
                    ctx2->cookie    = 0;
                    ctx2->dirents   = (fs_DirEntry*)ctx2->buf;
                    DLL_ISOLATE_ELEM(ctx, list);
                    DLL_INSERT_TAIL(&contexts, ctx, list);
                    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                             ("going down to %s", ctx->dirp->name));
                    ctx = ctx2;
                    //We are now in the context of the new dir
                    goingDown = TRUE;
                    break;
                    
goUp:
                    //Retrieve upper context
                    ctx2 = ctx;
                    DLL_REMOVE_TAIL(&contexts, ctx, list);
                    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                             ("going up from %s", 
                             (ctx!=NULL && ctx!=DLL_NIL) ?
                             ctx->dirp->name : "[END]"));
                    //Free previous context
                    fs_ramDealloc(FS_RAM_DUMP_DIRENT, ctx2);
                    
                    if (ctx == NULL || ctx == DLL_NIL)  {
                         //We are done, return here
                         return status;
                    } 
                }
               
                fs_inodeRelease(context, ctx->inode);

                FS_READDIR_ADV_ENTRY_PTR(ctx->dirp);
            }
            if (goingDown) break;
            
            ctx->root = I_GET_SHARED(snapnum, ctx->rootInum);
            FS_ASSERT(ctx->root != NULL);
        }
        if (goingDown) continue;

        goto goUp;
        
    } while (1);

    return status;
}

#ifdef FS_HIST_WAIT
void
fs_backupInit()
{
    NFX_SPIN_INIT(&fs_dumpSessionsSpinLock,SPINLOCK_LEVEL_UNKNOWN);
    DLL_INIT(&fs_dumpSessionsHdr);
}

void
fs_receiveFileHistWindowUpdate(struct eee_window_update *windowUpdate)
{
        fs_DumpSession_t    *session;
        boolean             found = FALSE;
        esm_event_t         event;
        NFX_SPIN_LOCK(&fs_dumpSessionsSpinLock);
        DLL_ITERATE(session, &fs_dumpSessionsHdr, link) {
                if (session->dumpContext->ndmp_ssc_src_port ==
                                windowUpdate->wu_srcaddr) {
                        found = TRUE;
                        break;
                }
        }
        if (found) {
                session->charSpace = windowUpdate->wu_char_space;
                session->mbufSpace = windowUpdate->wu_mbuf_space;
                if (session->dumpContext->histWait) {
                    session->dumpContext->histWait = FALSE;
                    event.event = FS_THREAD_EVENT_SEND_HIST;
                    thread_event(session->dumpContext->histWaiter, &event);
                }
        }
        NFX_SPIN_UNLOCK(&fs_dumpSessionsSpinLock);
}
#endif


#endif  //SSC
