/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * $Id: fs-eek.c,v 1.121 2002/12/03 01:00:59 lanc Exp $
 */

#if !defined(SSC)

#include "fs-includes.h"

typedef struct fs_Eek_bits fs_Eek_bits;
struct fs_Eek_bits {
	boolean(*extract)(fs_RefByte refbyte);
	void (*insert)(fs_RefByte *refbytep, int bit);
	void (*query)(fs_Context *context, fs_Eek_bits *bits,
		       fs_Buf *buf, uint8 *bitmap, uint64 blocknum,
		       int computed, int actual);
	jmp_buf jmp;
	uint64 roundup;
	uint8 *buddyBitmap;	// bitmap of buddy regions already rebuilt
	int maxBuddyRegions;
};

/*
 * Print eek status on FP console and send to SSC as well
 */
void
fs_eek_print_status(fs_Context *context, boolean printDate, char *fmt, ...)
{
    eeeGMT_t date;
    extern int vsprintf(char *str, const char *format, va_list ap);
#if !defined(FS_STANDALONE)
    fs_MsgEekResp *resp;
    int32 rc;
    fs_MsgEekReq *req;
    esm_event_t event;
#endif
    char string[256], vstring[256], dates[32];
	va_list ap;

    va_start(ap, fmt);
    eee_TimeGMT (&date);

    vsnprintf(vstring, 256, fmt, ap);
    if (printDate) {
        sprintf(dates, "%s", eee_humandate(date.sec));
        sprintf(string, "%s %s", vstring, dates);
    } else {
        sprintf(string, "%s", vstring);
    }
    va_end(ap);

#if defined(FS_STANDALONE)
    printf("%s\n", string);
#else

    /*
     * send to ssc as well when eek is invoked by admin via fscmd
     * watchout:
     * too many eee_sendMessage()s can deplete eee small shared buffers
     * the same buffers used by EVM for io ....
     */

    fs_lockExclusive(context, context->eek.lock, NULL);
    if (context->eek.sendToSsc) {
        resp = (fs_MsgEekResp *) context->respHdr;
        resp->has_query = FALSE;
        resp->has_status = TRUE;
        snprintf(resp->msg, 256, string);
        resp->private = context->smHandle;
        rc = eee_sendMessage(context->respHdr,
                            context->respSize,
                            context->srcAddr,
                            fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));
        if (rc != NFX_OK) {
            printf("eek: cannot send status to SSC\n");
            printf("status: %s\n", string);
            fs_unlockExclusive(context, context->eek.lock);
            return;
        }

        /*
         * even though we don't care about the reply
         * fscmd needs it to continue waiting and receiving the
         * next status/query message
         * note: use FS_YIELD_TYPE_TXCOMPL so we will not assert
         */

        fs_yield(context, &event, FS_YIELD_TIME_ADMIN, FS_YIELD_TYPE_TXCOMPL);
        if (event.event != FS_THREAD_EVENT_EEKCONT) {
            context->eek.sendToSsc = FALSE;
            printf("eek: no response from SSC, status messages to SSC disabled for %s\n",
                    context->fileSystem->super->mountName);
            fs_unlockExclusive(context, context->eek.lock);
            return;
        }
        context->respHdr->requestType = FS_REQ_EEKCONT;
        req = event.p1.u.p;
        context->respHdr->requestType = req->hdr.requestType;
        context->respHdr->clientPrivate = req->hdr.clientPrivate;
        context->respHdr->clientId  = req->hdr.clientId;

        if (event.p2.u.w & FS_CONTEXT_SHARED_REQUEST)
        {
            eee_deallocateBuffer(req);
        }
        else
        {
            fs_ramDealloc(FS_RAM_REQ_MSG, req);
        }
    }
    fs_unlockExclusive(context, context->eek.lock);
#endif
}

static int get_bit(uint8 *bitmap, uint64 bitnum, uint64 max_bitnum)
{
	assert(bitnum < max_bitnum);
	return (bitmap[bitnum >> 3] >> (bitnum & 0x7)) & 0x1;
}

/*
 * Set a bit in the bitmap, return old value of bit. Return value is used
 * in eek_mark to detect duplicates.
 */
static int set_bit(uint8 *bitmap, uint64 bitnum, uint64 max_bitnum)
{
	uint32 bit = 1 << (bitnum & 0x1f);
	uint32 *map = &((uint32 *)bitmap)[bitnum >> 5];

	//assert(bitnum < max_bitnum);
	return (ATOMIC_OR_INT(map, bit) & bit) != 0;
}

static int clear_bit(fs_Context *context, uint8 *bitmap, uint64 bitnum, uint64 max_bitnum)
{
	uint32 bit = 1 << (bitnum & 0x1f);
	uint32 *map = &((uint32 *)bitmap)[bitnum >> 5];

	if (bitnum >= max_bitnum) {
		fs_eek_print_status(context, FALSE,
				    "%s: INVALID btree block %lld to clear (max %lld)",
				    context->fileSystem->super->mountName,
				    bitnum, max_bitnum);
		return 0;
	}
	return (ATOMIC_AND_INT(map, ~bit) & bit) != 0;
}

/*
 * Compare two bits, print an error msg if they are different.
 *
 * One of the bits comes from a bitmap, one which eek constructs on the fly.
 * The other is the real bit from the FS image.
 *
 * XXX: GROSS HACK: if "bit" is negative, then the senses of the two bits are
 * reversed: the bitmap is a real FS summary bitmap, and the single bit is the
 * "expected bit".
 */
static boolean
compare_bits(fs_Context *context, char *msg, uint8 *bitmap, uint64 bitnum, int bit, uint64 max_bitnum)
{
	int expected_bit = get_bit(bitmap, bitnum, max_bitnum);
	if (bit < 0) {
		int swap = -bit;
		bit = expected_bit;
		expected_bit = swap;
	}
	if (expected_bit != bit) {
		fs_eek_print_status(context, FALSE,
                            "%s: MISMATCH ON %s %lld, EXPECTED %d, GOT %d",
                            context->fileSystem->super->mountName,
                            msg, bitnum, expected_bit, bit);
        context->eek.errcount++;
    }
	return (expected_bit == bit);
}

static boolean
fs_eek_is_interactive(fs_Context *context)
{
    return context->eek.interactive ? TRUE : FALSE;
}

boolean
fs_eek_query(fs_Context *context, char *fmt, ...)
{
	boolean reply = FALSE;
    boolean send_status = FALSE;
    char string[128];
	va_list ap;
    int rsp;

    va_start(ap, fmt);
    vsnprintf(string, 128, fmt, ap);
    va_end(ap);
    
    if (fs_eek_is_interactive(context))
    {
        fs_lockExclusive(context, context->eek.lock, NULL);
        switch (context->eek.defaultReply)
        {
            case FS_EEK_QUERY_NO:
                send_status = TRUE;
                break;

            case FS_EEK_QUERY_YES:
                context->eek.fixcount++;
                reply = TRUE;
                send_status = TRUE;
                break;

            case FS_EEK_QUERY_PROMPT:

                // We're running in interactive mode,
                // query user for action to take.

                va_start(ap, fmt);
                rsp = fs_eek_vquery(context, fmt, ap);
                va_end(ap);

                switch (rsp)
                {
                    case FS_EEK_QUERY_QUIT:
                        break;
                    case FS_EEK_QUERY_NO:
                        break;
                    case FS_EEK_QUERY_YES:
                        context->eek.fixcount++;
                        reply = TRUE;
                        break;
                    default:
                        assert(TRUE);
                }
#if !defined(FS_STANDALONE)
                // also print out on FP console
                printf("%s %s\n", string, (reply ? "Yes" : "No"));
#endif
                fs_unlockExclusive(context, context->eek.lock);
                return reply;
            default:
                break;
        }
        fs_unlockExclusive(context, context->eek.lock);
    }

    if (send_status)
        fs_eek_print_status(context, FALSE, "%s %s",
                    string, (reply ? "Yes" : "No"));
    else
        printf("%s %s\n", string, (reply ? "Yes" : "No"));

	return reply;
}

/*
 * Hide this behind a conditional so time is not spent on sprintf
 * in the common case (no errors).
 */
#define MAKE_MSG(last_arg, buf)\
    va_start(ap, last_arg);\
    vsnprintf(buf, sizeof (buf), fmt, ap);\
    va_end(ap);
    
boolean
fs_eek_mark(fs_Context *context, char *fmt, uint8 *bitmap, uint64 bitnum, ...)
{
	uint64 max_bitnum = context->fileSystem->super->numBlks;
	int old_bit;
	char msg[256];
	va_list ap;

	if (!fs_blockIsValid(fs_makeBlock(bitnum))) {
		fs_lockExclusive(context, context->eek.lock, NULL);
		MAKE_MSG(bitnum, msg);
		fs_eek_print_status(context, FALSE, "%s: INVALID %s",
                            context->fileSystem->super->mountName, msg); 
		context->eek.errcount++;
		fs_unlockExclusive(context, context->eek.lock);
		return FALSE;
	}

	if (bitnum >= max_bitnum) {
		MAKE_MSG(bitnum, msg);
		fs_eek_print_status(context, FALSE,
          	      "%s: OVERFLOW %s at %lld exceeds max: %lld",
           	     context->fileSystem->super->mountName,
           	     msg, bitnum, max_bitnum);
		fs_lockExclusive(context, context->eek.lock, NULL);
		context->eek.errcount++;
		fs_unlockExclusive(context, context->eek.lock);
		return FALSE;
	}

	old_bit = set_bit(bitmap, bitnum, max_bitnum);

	// printf(" MAP %s #%lld\n", msg, bitnum);

	if (old_bit != 0) {
		MAKE_MSG(bitnum, msg);
		fs_eek_print_status(context, FALSE,
                            "%s: DUPLICATE %s allocated at %lld",
                            context->fileSystem->super->mountName,
                            msg, bitnum);
		fs_lockExclusive(context, context->eek.lock, NULL);
		context->eek.errcount++;
		fs_unlockExclusive(context, context->eek.lock);
		return FALSE;
	}

	return TRUE;
}

/*
 * Clear a marked bit when disposing of a marked block.
 */
void
fs_eek_unmark(fs_Context *context, uint8 *bitmap, uint64 bitnum)
{
	uint64 max_bitnum = context->fileSystem->super->numBlks;
	int old_bit;

	if (!fs_blockIsValid(fs_makeBlock(bitnum))) {
		fs_eek_print_status(context, FALSE,
                            "%s: INVALID btree block to unmark",
                            context->fileSystem->super->mountName);
		context->eek.errcount++;
		return;
	}

	/*
	 * called by btree to unmark nodes for an eek dispose status
	 * not an internal eek error so don't assert
	 */
	if (bitnum >= max_bitnum) {
		fs_eek_print_status(context, FALSE,
                            "%s: OVERFLOW btree block %lld to unmark (max %lld)",
                            context->fileSystem->super->mountName,
                            bitnum, max_bitnum);
		context->eek.errcount++;
		return;
	}

	old_bit = clear_bit(context, bitmap, bitnum, max_bitnum);
	if (old_bit == 0) {
        /*
         * we may be in the process of marking an inode so it may
         * not have been set in the bitmap yet
         * just printout the warning
         */
        fs_eek_print_status(context, FALSE,
                "WARNING: potential DUPLICATE clear of btree block %lld",
                context->fileSystem->super->mountName, bitnum);
        // context->eek.errcount++;
        return;
    }
}


#define MAX_RESERVED_BLOCKNUMBERS 66

/*
 * Compare a block of refcount bytes against eek's own bitmap.
 * Each refcount byte has its active FS bit extracted when
 * checking the active FS, or the snapcount+active extracted
 * when checking the free tree.
 */
boolean
fs_eek_compare_block(
	fs_Context *context,
	fs_Eek_bits *bits,
	uint8 *bitmap,
	fs_Buf *refbits_buf,
	uint64 block_offset,
	uint64 max_blocks)
{
	uint64 bitnum = block_offset;
	uint8 *refbits = refbits_buf->u.data;
	uint64 max_bitnum;
	uint64 i;
	boolean ok = TRUE;

	/*
	 * When comparing the first block, skip over reserved block numbers.
	 */
	if (block_offset == 0)
		i = MAX_RESERVED_BLOCKNUMBERS;
	else
		i = 0;

	/*
	 * At the end of the filesystem, don't check any uninitialized
	 * refbytes which might lie in the remainder of the final page.
	 */
	if (block_offset + PAGE_SIZE > max_blocks)
		max_bitnum = max_blocks % PAGE_SIZE;
	else
		max_bitnum = PAGE_SIZE;

	/*
	 * Loop through the bitmap comparing actual to computed. If they don't
	 * match, ask the user to clear the bit or to relink the block into
	 * lost+found. Since this is done as a post-pass, we clear any bit
	 * in the computed bitmap which doesn't need to be relinked into
	 * lost+found.
	 */
	for (; i < max_bitnum; i++) {
		int computed_bit = get_bit(bitmap, bitnum + i, context->fileSystem->super->numBlks);
		int actual_bit = bits->extract(refbits[i]);
		if (computed_bit != actual_bit) {
			ok = FALSE;
			bits->query(context, bits, refbits_buf, bitmap, bitnum + i, computed_bit, actual_bit);
		} else {
			clear_bit(context, bitmap, bitnum+i, context->fileSystem->super->numBlks);
		}
	}
	
	return ok;
}

#define BITS_PER_PAGE (PAGE_SIZE*8)

/*
 * Compare a block of snap bits against eek's own bitmap.
 */
boolean
fs_eek_compare_block_bitwise(fs_Context *context, uint8 *bitmap, uint8 *snapbits, uint64 block_offset, uint64 max_blocks)
{
	uint64 bitnum = block_offset;
	uint64 max_bitnum;
	uint64 i;
	boolean ok = TRUE;

	/*
	 * When comparing the first block, skip over reserved block numbers.
	 */
	if (block_offset == 0)
		i = MAX_RESERVED_BLOCKNUMBERS;
	else
		i = 0;

	/*
	 * At the end of the filesystem, don't check any uninitialized
	 * snapbits which might lie in the remainder of the final page.
	 */
	if (block_offset + BITS_PER_PAGE > max_blocks)
		max_bitnum = max_blocks % BITS_PER_PAGE;
	else
		max_bitnum = BITS_PER_PAGE;

	/*
	 * XXX: replace this with bitwise query.
	 */
	for (; i < max_bitnum; i++) {
		int bit = get_bit(snapbits, (bitnum+i) % BITS_PER_PAGE, context->fileSystem->super->numBlks);
		// if (bit)
		// 	printf(" snapbits map %lld\n", bitnum+i);

		ok &= compare_bits(context,
                "block number", bitmap, bitnum + i,
                bit, context->fileSystem->super->numBlks);
	}
	
	return ok;
}

/** 
 * Return the /lost+found directory
 * NULL if not found or corruption
 */ 
static fs_Inode *
fs_eek_get_lost_found(fs_Context *context, fs_Inode *root, uint8 *bitmap)
{
	char *const name = "lost+found";
	fs_Inode *lf;
	uint64 inum;
	
    if (!fs_eek_is_interactive(context))
    {
        return NULL;
    }

    /*
     * verify integrity of emap btree for ROOT before using it
     */

    if (!fs_eek_meta_inode(context, FS_ROOT_INODE, 0)) {
        fs_eek_print_status(context, FALSE,
                            "INVALID emap for root");
        context->eek.errcount++;
        return NULL;
    }

    inum = fs_dir_lookupent(context, root, name, strlen(name));
    if (inum == 0) {
        fs_eek_print_status(context, FALSE, "%s: cannot find lost+found",
                            context->fileSystem->super->mountName);
        return NULL;
    }
    lf = I_GET_EXCL(0, inum);
    if (lf == NULL) {
        fs_eek_print_status(context, FALSE, "%s: cannot find lost+found",
                            context->fileSystem->super->mountName);
    }
    if (!fs_eek_meta_inode(context, lf->inum, 0)) {
        /*
         * TBD: try to fix it ??
         */
        fs_eek_print_status(context, FALSE, "INVALID emap for lost+found");
        context->eek.errcount++;
        return NULL;
    }
	return lf;
}

#if defined(FS_EEK_RELINK_BLOCKS)
static void
fs_eek_relink_lost_blocks(fs_Context *context, uint8 *bitmap, uint64 num_blocks)
{
    eeeRTC_t beginTime, delta;
    fs_Inode *root = I_GET_EXCL(0, FS_ROOT_INODE);
	fs_Inode *lf = fs_eek_get_lost_found(context, root, bitmap);
	fs_Inode *inode;
	char name[32];
	uint64 i;

    if (lf == NULL)
        goto fs_eek_relink_lost_blocks_exit;

    eee_rtc(&beginTime);

	for (i = MAX_RESERVED_BLOCKNUMBERS; i < num_blocks; i++) {
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
			eee_rtc(&beginTime);
        }
		if (get_bit(bitmap, i, context->fileSystem->super->numBlks)) {
			inode = fs_allocateFileInode(context, lf, 0400, 0, 0, 0);
			sprintf(name, "B#%lld", i);
			/*
			 * XXX: this can fail if "name" is not unique; but this should
			 * never happen unless someone maliciously adds files to lost+found;
			 * ideally eek should check for this and generate unique filenames.
			 */
			fs_dir_addent(context, lf, name, strlen(name), inode->inum);
            inode->dinode.size = FS_BLOCK_SIZE;
            inode->dinode.attrs = FS_IFMT_REG;
            inode->dinode.linkCnt = 1;
			fs_emap_add_block(context, inode, 0, fs_makeBlock(i));
            inode->dinode.numblocks += 1;
            fs_inodeWrite(context, inode, 0);
		}
	}

	if (!fs_inodeWasModified(context, lf))
		fs_inodeReleaseExclusive(context, lf);
    else
        assert(context->eek.interactive == TRUE);

fs_eek_relink_lost_blocks_exit:

	if (!fs_inodeWasModified(context, root))
		fs_inodeReleaseExclusive(context, root);
    else
        assert(context->eek.interactive == TRUE);

}
#endif

/*
 * Move unreferenced inodes to the lost+found
 * Or free if lost+found is missing or found inconsistent
 */
void
fs_eek_relink_lost_inodes(fs_Context *context, uint8 *bitmap, uint64 num_inodes)
{
    eeeRTC_t beginTime, delta;
    fs_Inode *root = I_GET_EXCL(0, FS_ROOT_INODE);
	fs_Inode *lf = fs_eek_get_lost_found(context, root, bitmap);
	char name[32];
	uint64 i;
    int32 deallocs = 0;

    if (context->eek.interactive == FALSE)
        goto fs_eek_relink_lost_inodes_return;

    eee_rtc(&beginTime);

	for (i = 0; i < num_inodes; i++) {
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
			eee_rtc(&beginTime);
        }
		if (get_bit(bitmap, i, context->fileSystem->super->numBlks)) {
			sprintf(name, "I#%lld", i);
            if (lf == NULL) {
                /*
                 * no lost+found, can't relink so just get rid of them
                 */
                fs_Inode *inode = I_GET_EXCL(0, i);
                assert(inode);
                fs_truncateFile(context, inode, 0, TRUE, NULL);
                fs_deallocateInode(context, inode);
                /* 
                 * TBD: until eek is made transactionless
                 * commit if necc since only a limited number of inodes
                 * can be deallocated within a single transaction
                 */

                if (++deallocs == FS_LOG_MAX_INODE_DEALLOC) {
                    deallocs = 0;
	                fs_logCommitEntry(context);
	                fs_logBeginTransaction(context);
                }
            } else {
			    /*
			     * XXX: this can fail if "name" is not unique; but this should
			     * never happen unless someone maliciously adds files to lost+found;
			     * ideally eek should check for this and generate unique filenames.
			     */
			    fs_dir_addent(context, lf, name, strlen(name), i);
            }
		}
	}

fs_eek_relink_lost_inodes_return:

    if (lf) {
	    if (!fs_inodeWasModified(context, lf))
		    fs_inodeReleaseExclusive(context, lf);
        else
            assert(context->eek.interactive == TRUE);
    }

	if (!fs_inodeWasModified(context, root))
		fs_inodeReleaseExclusive(context, root);
    else
        assert(context->eek.interactive == TRUE);
}

static void
fs_eek_mark_symlink(fs_Context *context, uint8 *bitmap, fs_Inode *inode)
{
	fs_Eekmark marktype;
    if (inode->dinode.size > FS_INODE_INLINE_LEN)
    {

#if 0
        /*
         * Not excersized any more
         */

        fs_Block    block = fs_extentStartBlock(inode->dinode.u.extent[0]);

        if (fs_blockIsValid(block))
        {
            if (!fs_eek_mark(context, "(I#%lld) symlink block number",
		bitmap, fs_blockNumber(block), inode->inum) &&
                fs_eek_query(context, "Clear?"))
            {
                inode->dinode.size = 0;
                // XXX: TODO:  write synchronously???
            }
        }
#else
        marktype = FS_EEK_MARK_NODES | FS_EEK_MARK_EXTENTS;
        fs_emap_fix_used_blocks(context, bitmap, inode, marktype);
#endif
        // XXX: TODO: handle this else case
    }
}

/*
 * eek's walk worker routine to mark used blocks
 */
static boolean
fs_eek_walk_worker(fs_Context *context,
        fs_Inode *inode,
        fs_WalkResult *result,
        void *arg0)
{
    uint8 *bitmap = arg0;
    uint16 b, n;

    if (result->type == FS_WALK_EXTENT_TYPE) {
        n = fs_extentLength(result->extent);
        for (b = 0; b < n; b++) {
            if (!fs_eek_mark(context, "used block (inum %lld extent offset %lld start %lld len %d)", bitmap,
                             fs_blockNumber(fs_extentBlock(result->extent, b)),
			     inode->inum, result->offset,
			     fs_blockNumber(fs_extentStartBlock(result->extent)),
			     fs_extentLength(result->extent))) {
                context->eek.errcount++;
		return FALSE;
            }
        }
    } else {
        assert((result->type == FS_WALK_INNER_NODE_TYPE) ||
               (result->type == FS_WALK_LEAF_NODE_TYPE));

        if (!fs_eek_mark(context, "used block (inum %lld %s btree node)",
		bitmap, fs_blockNumber(result->block),
		inode->inum, (result->type == FS_WALK_INNER_NODE_TYPE) ? "inner" : "leaf")) {
            context->eek.errcount++;
	    return FALSE;
        }
    }
    return TRUE;
}

/*
 * eek a meta inode's emap btree
 * Return TRUE if ok
 * FALSE if corruption
 */
boolean
fs_eek_meta_inode(fs_Context *context,
                  uint64 inum,
                  int snapnum)
{
    int errors;
    boolean ok = TRUE;
    uint64 num_blocks = context->fileSystem->super->numBlks;
    fs_Inode *inode = I_GET_SHARED(snapnum, inum);
    uint64 maxBitnum;
    uint8 *bitmap = NULL;

    if (inode == NULL) {
        return ok;
    }
    maxBitnum = ceilingDiv64(num_blocks, 8);
    bitmap = fs_ramAlloc(FS_RAM_CORE_MISC, maxBitnum);
    if (bitmap == NULL) {
        fs_eek_print_status(context, FALSE,
                            "%s: FATAL: not enough memory to eek meta inode %lld",
					        context->fileSystem->super->mountName, inum);
        fs_inodeRelease(context, inode);
        return FALSE;
    }
	bzero(bitmap, maxBitnum);

    errors = context->eek.errcount;
    context->eek.errcount = 0;

    fs_emap_walk(context, inode, fs_eek_walk_worker, bitmap);
    if (context->eek.errcount != 0) {
        ok = FALSE;
    }

    fs_inodeRelease(context, inode);
    context->eek.errcount = errors;

    fs_ramDealloc(FS_RAM_CORE_MISC, bitmap);
    return ok;
}

/*
 * Walk inodes and mark all blocks used
 * fixIt is false for:
 * 1. The first phase of marking used blocks when no corrections should be
 *    made until the buddy has been reconciled with reference
 * 2. When eek'ing snaphots
 *
 * This phase will not count errors but if nerrors is not NULL,
 * number of errors encountered will be set.  This avoids double counting
 * the same inconsistent block for walking and fixing phases.
 *
 * Return FALSE for unrecoverable error (eg corrupted inode file)
 */

/*
 * XXX: parametrize NTHREADS
 */
#define NTHREADS 64

/*
 * XXX: make this a generic struct for all eekThreads
 */
struct blocksargs {
	esm_handle_t parent;

	/*
	 * Args to max_used_blocks, passed out to worker threads.
	 */
	uint8 *bitmap;
	int snapnum;
	uint64 num_inodes;
	boolean fixIt;
	fs_Lock *lockp;

	int nworkers;

	/*
	 * Spinlock protected variables.
	 */
	nfx_spinlock_t spin;
	uint64 inum;		// base inum of block of inode file being worked on
	int nfinished;		// number of threads which have finished working
	int errors;		// total number of errors encountered in phase
};

static void
fs_eekThreadWakeup(fs_Context *context, struct blocksargs *args)
{
	boolean wakeup;

	NFX_SPIN_LOCK(&args->spin);
	args->nfinished++;
	wakeup = (args->nfinished == NTHREADS);
	NFX_SPIN_UNLOCK(&args->spin);

	if (wakeup) {
		fs_FileSystem *fs = context->fileSystem;
		esm_event_t event;
		int32 dummy = 0;
	
		event.event = FS_THREAD_EVENT_START;
		event.p1.u.p = NULL;
		event.p2.u.w = 33;	// debug
	
		fs_daemonWakeup(context, &fs->spin, args->parent, &dummy, event);
	}
}

static void
fs_eekThreadSpawn(fs_Context *context, void (*threadfunc)(esm_handle_t handle, void *arg), struct blocksargs *args)
{
	int i;

	NFX_SPIN_INIT(&args->spin, SPINLOCK_LEVEL_UNKNOWN);

	/*
	 * XXX: check to see if the threadFork fails.
	 */
	for (i = 0; i < NTHREADS; i++) {
		args->nworkers++;
		fs_threadFork(context, threadfunc, TRUE, FALSE, FS_DAEMON_TYPE_EEK, args);
	}

	fs_yield(context, NULL, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_ADMIN);
}

static void
fs_eek_mark_one_inode_block(fs_Context *context, int snapnum, uint8 *bitmap, fs_Buf *buf, uint64 inum)
{
	int i;

	for (i = 0; i < FS_INODES_PER_PAGE; i++) {
		fs_Inode *inode, dummy;

		dummy.inum = inum + i;
		dummy.snapnum = snapnum;
		dummy.dinode = buf->u.dinode->dinode[i];

		if (dummy.dinode.attrs == 0)
			continue;

		inode = &dummy;

		switch (inode->dinode.attrs & FS_IFMT_MASK) {
		case FS_IFMT_SYML:
			if (inode->dinode.size > FS_INODE_INLINE_LEN)
				fs_emap_walk(context, inode, fs_eek_walk_worker, bitmap);
			break;
		case FS_IFMT_DIR:
		case FS_IFMT_REG:
		case FS_IFMT_QUOT:
			fs_emap_walk(context, inode, fs_eek_walk_worker, bitmap);
			break;
		case FS_IFMT_CHR:
		case FS_IFMT_BLK:
		case FS_IFMT_FIFO:
		case FS_IFMT_SOCK:
			break;
		default:
			fs_eek_print_status(context, FALSE,
					    "%s: WARNING: ino %lld unknown type %o skipped",
					    context->fileSystem->super->mountName, inode->inum,
					    inode->dinode.attrs & FS_IFMT_MASK);
			break;
		}
	}
}

void
fs_eekBlocksThread(esm_handle_t handle, void *arg)
{
	fs_Inode *inodes;
	fs_Context *context = arg;
	struct blocksargs *args = context->extraArg;
	eeeRTC_t beginTime, delta;
	uint64 inum;
	uint8 *bitmap;
	int snapnum;

	bitmap = args->bitmap;
	snapnum = args->snapnum;
	context->eek.lock = args->lockp;
	
	context->eek.errcount = 0;
	
	eee_rtc(&beginTime);

	inodes = I_GET_SHARED(snapnum, FS_INODE_INODE);

	while (1) {
		fs_Buf *buf;

		/*
		 * Get a new page of the inode file to work on.
		 */
		NFX_SPIN_LOCK(&args->spin);
		inum = args->inum;
		args->inum += FS_INODES_PER_PAGE;
		NFX_SPIN_UNLOCK(&args->spin);

		if (inum >= args->num_inodes)
			break;

		if (inum == 0)
			buf = B_GET_SHARED(NULL, fs_firstInodeBlock(snapnum));
		else
			buf = O_GET_SHARED(inodes, inum / FS_INODES_PER_PAGE);

		if (buf != NULL) {
			fs_eek_mark_one_inode_block(context, snapnum, bitmap, buf, inum);
			fs_bufRelease(context, buf);
		}

		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0) {
			fs_sleep(context, 0);
			eee_rtc(&beginTime);
		}
	}

	fs_inodeRelease(context, inodes);

	NFX_SPIN_LOCK(&args->spin);
	args->errors += context->eek.errcount;
	NFX_SPIN_UNLOCK(&args->spin);

	fs_eekThreadWakeup(context, args);
}


boolean
fs_eek_mark_used_blocks(fs_Context *context,
            uint8 *bitmap,
            int snapnum,
            uint64 num_inodes,
            boolean fixIt,
            int *nerrors)
{
	uint64 num_blocks = context->fileSystem->super->numBlks;
	esm_handle_t me = context->smHandle;
	struct blocksargs args;

	bzero(bitmap, ceilingDiv64(num_blocks, 8));

	args.bitmap = bitmap;
	args.snapnum = snapnum;
	args.num_inodes = num_inodes;
	args.fixIt = fixIt;
	args.errors = 0;
	args.lockp = context->eek.lock;

	args.inum = 0;
	args.nworkers = 0;
	args.nfinished = 0;

	args.parent = me;

	fs_eekThreadSpawn(context, fs_eekBlocksThread, &args);

	if (nerrors != NULL)
		*nerrors = args.errors;

	return TRUE;
}

/*
 * Fix all blocks allocated to inodes
 * Return TRUE if a second pass of checking used blocks is neccessary
 */
boolean
fs_eek_fix_used_blocks(fs_Context *context, uint8 *bitmap, int snapnum, uint64 num_inodes)
{
    eeeRTC_t beginTime, delta;
	uint64 i;
	fs_Eekmark marktype;
    boolean checkUsedBlocks = FALSE;
    uint64 num_blocks = context->fileSystem->super->numBlks;
    int32 deallocs = 0;

    bzero(bitmap, ceilingDiv64(num_blocks, 8));
    eee_rtc(&beginTime);

	for (i = 0; i < num_inodes; i++) {
		fs_Inode *inode = I_GET_EXCL(snapnum, i);
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
			eee_rtc(&beginTime);
        }
		if (inode == NULL)
			continue;
		switch (inode->dinode.attrs & FS_IFMT_MASK) {
		case FS_IFMT_DIR:
		case FS_IFMT_REG:
			marktype = FS_EEK_MARK_NODES | FS_EEK_MARK_EXTENTS;
            if (!fs_emap_fix_used_blocks(context, bitmap, inode, marktype)) {
                /* 
                 * TBD: until eek is made transactionless
                 * commit if necc since only a limited number of inodes
                 * can be deallocated within a single transaction
                 */

                if (deallocs++ >= FS_LOG_MAX_INODE_DEALLOC) {
	                fs_logCommitEntry(context);
	                fs_logBeginTransaction(context);
                    // the commit will unlock the modified inode so reaquire it
	                if (!fs_lockOwnedExclusive(context, &inode->lock)) {
		                inode = I_GET_EXCL(snapnum, i);
                    }
                }

                /*
                 * inode corruption, remove entire file
                 */
                fs_truncateFile(context, inode, 0, TRUE, NULL);
                fs_deallocateInode(context, inode);
                fs_eek_print_status(context, FALSE,
                        "%s: (I#%lld) CORRUPTED, deallocated inode\n",
                        context->fileSystem->super->mountName,
                        inode->inum);
			    checkUsedBlocks = TRUE;
            }
			break;
		case FS_IFMT_SYML:
#if 0
            if (inode->dinode.size <= FS_INODE_INLINE_LEN) {
                fs_eek_mark_symlink(context, bitmap, inode);
            } else {
                marktype = FS_EEK_MARK_NODES | FS_EEK_MARK_EXTENTS;
                fs_emap_fix_used_blocks(context, bitmap, inode, marktype);
            }
#endif
            fs_eek_mark_symlink(context, bitmap, inode);
            break;
		case FS_IFMT_CHR:
		case FS_IFMT_BLK:
		case FS_IFMT_FIFO:
		case FS_IFMT_SOCK:
		case FS_IFMT_QUOT:
			/* XXX: TODO: check special inodes */
			break;
		default:
			/* XXX: recover gracefully here */
            fs_eek_print_status(context, FALSE,
                                "%s: WARNING: ino %lld unknown type %o skipped",
                                context->fileSystem->super->mountName,
                                inode->inum, inode->dinode.attrs & FS_IFMT_MASK);
            break;
		}
        if (!fs_inodeWasModified(context, inode))
		    fs_inodeReleaseExclusive(context, inode);
        else
            assert(context->eek.interactive == TRUE);
	}
    return checkUsedBlocks;
}

/*
 * Compare a page of reference bytes against the computed
 * used blocks in the bitmap
 */
void
fs_eek_compare_used_blocks_by_refcounts(fs_Context *context,
			   fs_Eek_bits *bits,
			   uint8 *bitmap,
			   fs_Inode *refcounts,
			   uint64 num_blocks)
{
    eeeRTC_t beginTime, delta;
	uint64 i;

    eee_rtc(&beginTime);

	for (i = 0; i < num_blocks; i += PAGE_SIZE) {
		fs_Buf *buf = O_GET_EXCL(refcounts, i / PAGE_SIZE);
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }
		if (setjmp(bits->jmp) == 0) {
			fs_eek_compare_block(context, bits, bitmap, buf, i, num_blocks);
		} else {
			/* skip to next buddy region */
			i = (i + bits->roundup) &~ bits->roundup;
		}
        if (!fs_bufWasModified(context, buf))
		    fs_bufReleaseExclusive(context, buf);
        else
            assert(context->eek.interactive == TRUE);
	}
}

/*
 * Compare a page of snapbits against the computed
 * used blocks in the bitmap
 */
void
fs_eek_compare_used_blocks_by_snapbits(fs_Context *context,
			   uint8 *bitmap,
			   fs_Inode *snapbits,
			   uint64 num_blocks)
{
    eeeRTC_t beginTime, delta;
	uint64 i;

    eee_rtc(&beginTime);

	for (i = 0; i < num_blocks; i += PAGE_SIZE*8) {
		fs_Buf *buf = O_GET_EXCL(snapbits, i / (PAGE_SIZE*8));

        if (buf == NULL) {
            break;  // growfs extras
        }
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }
		fs_eek_compare_block_bitwise(context, bitmap, buf->u.data, i, num_blocks);
        if (!fs_bufWasModified(context, buf))
		    fs_bufReleaseExclusive(context, buf);
        else
            assert(context->eek.interactive == TRUE);
	}
}

boolean
fs_eek_refbyte_is_free(fs_RefByte refbyte)
{
	return !fs_refbyteIsLive(refbyte) && fs_refbyteSnapCount(refbyte) == 0;
}

/*
 * Rebuild (from the refcounts) the ith region of the buddy tree.
 */
void
fs_eek_rebuild_buddy_region(fs_Context *context, uint64 i)
{
    eeeRTC_t beginTime, delta;
	fs_Inode *buddies = I_GET_EXCL(0, FS_FREE_BUDDY_TREE_INODE);
	fs_Buf *buf = O_GET_EXCL(buddies, i);
    uint64 j;

	bzero(buf->u.data, PAGE_SIZE);
	FS_BUF_ZERO_ALL(buf);

    eee_rtc(&beginTime);

	for (j = 0; j < FS_BLOCKS_PER_BUDDY_REGION; j++) {
		uint64 blocknum = (i * FS_BLOCKS_PER_BUDDY_REGION) + j;
		fs_Block b = fs_makeBlock(blocknum);

		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }
		if (fs_eek_refbyte_is_free(fs_refGetByte(context, b))) {
            fs_buddyMaybeAddExtent(context, fs_makeExtent(1, b), FALSE);
            context->eek.freecount++;
		}
	}
        
	if (!fs_inodeWasModified(context, buddies))
	    fs_inodeReleaseExclusive(context, buddies);
    else
        assert(context->eek.interactive == TRUE);
}

/*
 * Mark the free blocks listed in the binary buddy tree and check it for
 * internal consistency.
 */
void
fs_eek_mark_free_blocks(fs_Context *context, uint8 *bitmap)
{
    eeeRTC_t beginTime, delta;
	fs_Inode *buddies = I_GET_EXCL(0, FS_FREE_BUDDY_TREE_INODE);
    uint64 i;

    if (buddies == NULL) {
        fs_eek_print_status(context, FALSE,
                            "%s: FS_FREE_BUDDY_TREE_INODE not found",
                            context->fileSystem->super->mountName);
        context->eek.errcount++;
        return;
    }

    eee_rtc(&beginTime);

	for (i = 0; i < buddies->dinode.size / PAGE_SIZE; i++) {
		fs_Buf *buf = fs_buddyGetRegionBuf(context, i, TRUE);
        if (buf == NULL) {
            fs_eek_print_status(context, FALSE,
                "%s: page %lld for FREETREE inode not found",
                context->fileSystem->super->mountName, i);
            context->eek.errcount++;
            return;
        }
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }
		if (!fs_buddyEekMarkRegion(context, buf, bitmap, i * FS_BLOCKS_PER_BUDDY_REGION))
        {
            // PR#1520 increment errcount
            context->eek.errcount++;
		    if (fs_eek_query(context, "%s: rebuild free map region #%lld?",
                            context->fileSystem->super->mountName, i))
		    {
		        fs_eek_rebuild_buddy_region(context, i);
		    }
        }
        if (!fs_bufWasModified(context, buf))
		    fs_bufReleaseExclusive(context, buf);
        else
            assert(context->eek.interactive == TRUE);
	}
	if (!fs_inodeWasModified(context, buddies))
	    fs_inodeReleaseExclusive(context, buddies);
    else
        assert(context->eek.interactive == TRUE);
}

/*
 * Compare the computed used blocknum marked in bitmap against the
 * reference byte (actual).  Adjust the reference byte if not the same.
 */
void
fs_eek_used_query(
	fs_Context *context,
	fs_Eek_bits *bits,
	fs_Buf *refbits_buf,
	uint8 *bitmap,
	uint64 blocknum,
	int computed,
	int actual)
{
	uint64 max_blocks = context->fileSystem->super->numBlks;
	fs_Block b = fs_makeBlock(blocknum);
    char err[256];

	assert(computed != actual);
    context->eek.errcount++;

    if (computed == 1) {
        sprintf(err, "block %lld incorrectly marked free.", blocknum);
        fs_eek_print_status(context, FALSE, "%s", err);
        if (context->eek.interactive) {
            /*
             * block used by some inode but reference byte cleared
             */
            if (fs_eek_query(context, "Adjust?")) {
                fs_refByteModifyBlock(context, fs_refbyteForceLive, b);
                /* buddy is adjusted during the check free block phase */
            }
            clear_bit(context, bitmap, blocknum, max_blocks);
        }
    } else {
        /*
         * block not used (in live) but live bit set in reference byte
         */
        sprintf(err, "block %lld incorrectly marked allocated.", blocknum);
        fs_eek_print_status(context, FALSE, "%s", err);
        if (context->eek.interactive) {
            if (fs_eek_query(context, "Adjust?")) {
	            if (blocknum < max_blocks) {
                    fs_RefByte byte = fs_refGetByte(context, b);
                    if (fs_refbyteSnapCount(byte) == 0) {
                        fs_refByteModifyBlock(context, fs_refbyteZero, b);
                        fs_buddyMaybeAddExtent(context, fs_makeExtent(1, b), FALSE);
                        context->eek.freecount++;
                    } else {
                        /*
                         * Just clear the live bit since a snapshot is still using it
                         */
                        fs_refByteModifyBlock(context, fs_refbyteClearLive, b);
                    }
                }
            }
        }
    }
}

/*
 * Mismatched found between the computed buddy free block
 * and the (actual) reference byte.
 * Since the reference byte is always the true version, rebuild the buddy.
 */
void
fs_eek_free_query(
	fs_Context *context,
	fs_Eek_bits *bits,
	fs_Buf *refbits_buf,
	uint8 *bitmap,
	uint64 blocknum,
	int computed,
	int actual)
{
	uint64 i = blocknum / FS_BLOCKS_PER_BUDDY_REGION; // ith buddy region
	char err[80];

	assert(computed != actual);
    assert(bits->buddyBitmap);

    /*
     * check if this region has already been checked
     */
    if (get_bit(bits->buddyBitmap, i, bits->maxBuddyRegions)) {
        return;
    }
    set_bit(bits->buddyBitmap, i, bits->maxBuddyRegions);

    context->eek.errcount++;
	/*
	 * eek buddy puts "1"s in the bitmap where it thinks there are free
	 * blocks.
	 */
	sprintf(err, "Free map incorrectly marks block %lld as %s.", blocknum,
		    computed == 1 ? "free" : "allocated");

    if (context->eek.interactive) {
	    if (fs_eek_query(context, "%s: %s Rebuild free map region #%lld?",
                        context->fileSystem->super->mountName, err, i))
	    {
		    fs_eek_rebuild_buddy_region(context, i);
		    longjmp(bits->jmp, 1);
	    }
    } else {
        fs_eek_print_status(context, FALSE,
                            "%s: %s Inconsistent free map region #%lld",
                            context->fileSystem->super->mountName, err, i);
    }
}

/*
 * Mark the first few bits in the bitmap for reserved inode numbers;
 * these are files which are not reachable from any directory.
 */
static void
hard_mark(fs_Context *context, uint8 *bitmap, uint64 inum)
{
	fs_eek_mark(context, "reserved inum", bitmap, inum);
}

/*
 * Mark all inodes reachable from some directory
 */
void
fs_eek_mark_used_inodes(fs_Context *context, uint8 *bitmap, int snapnum, uint64 num_inodes, boolean fixIt)
{
    eeeRTC_t beginTime, delta;
	fs_Eek *eek = &context->eek;
	uint64 i;
    uint64 inodes = 0, inodesUsed;

	eek->linkcounts = NULL;
	bzero(bitmap, ceilingDiv64(num_inodes, 8));

	/*
	 * Mark inodes allocated which cannot be reached from the root
	 * directory (including the root itself.)
	 */
	hard_mark(context, bitmap, FS_INODE_INODE /*1*/);
	hard_mark(context, bitmap, FS_ROOT_INODE /*2*/);
	// inodes 3,4,6 and 7 not snapped so they should not be hard_mark'd
	if (snapnum == 0) {
		hard_mark(context, bitmap, FS_REF_COUNT_INODE /*3*/);
		hard_mark(context, bitmap, FS_FREE_BUDDY_TREE_INODE /*4*/);
		hard_mark(context, bitmap, FS_FULL_INODE_BLOCK_INODE /*6*/);
		hard_mark(context, bitmap, FS_DIRECTORY_IN_INODE_BLOCK_INODE /*7*/);
	}
	hard_mark(context, bitmap, FS_SECURITY_INODE /*5*/);
	hard_mark(context, bitmap, FS_FILESET_INFO_INODE /*8*/);
	hard_mark(context, bitmap, FS_FILESET_USER_QUOTA_INODE /*9*/);
	hard_mark(context, bitmap, FS_FILESET_GROUP_QUOTA_INODE /*10*/);
	hard_mark(context, bitmap, FS_NULL_INODE /*11*/);
	hard_mark(context, bitmap, FS_TRASHCAN_INODE /*12*/);
	hard_mark(context, bitmap, FS_SNAPSHOT_DIR_INODE /*13*/);
	// inode 14 should be hard_mark'd unless it's live
	if (snapnum != 0) hard_mark(context, bitmap, FS_SNAPSHOT_BITS_INODE /*14*/);
	hard_mark(context, bitmap, FS_LINK_INODE /*15*/);
	hard_mark(context, bitmap, FS_TEMPDIR_INODE /*16*/);
	hard_mark(context, bitmap, FS_TEMP_DUMP_SWAP /*17*/);
	hard_mark(context, bitmap, FS_AUDIT_INODE /*18*/);
	hard_mark(context, bitmap, FS_SYSTEM_HIDDEN_INODE /*19*/);
	hard_mark(context, bitmap, FS_EXTENDED_ATTR_INODE /*20*/);
	hard_mark(context, bitmap, FS_REF_OVERFLOW_INODE /*21*/);
	hard_mark(context, bitmap, FS_CIFS_SHARES_INODE /*22*/);

    eee_rtc(&beginTime);

	for (i = 0; i < num_inodes; i++) {
		fs_Inode *inode = I_GET_EXCL(snapnum, i);
		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }
		if (inode == NULL)
			continue;
		if (FS_I_IS_DIR(inode->dinode.attrs))
			fs_dir_mark_used_inodes(context, bitmap, inode, fixIt);
		if (!fs_inodeWasModified(context, inode))
			fs_inodeReleaseExclusive(context, inode);
        else
            assert(context->eek.interactive == TRUE);

        // keep track of inodes used to eek against superblk's version
        if (snapnum) continue;
        if (i >= FS_MAX_RESERVED_INODE) {
            inodes++;
        }
	}

    if (snapnum)
        return;

    // compare inodes found with the value in the superblock

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    inodesUsed = context->fileSystem->super->inodesUsed;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    if (inodes != inodesUsed) {
        fs_eek_print_status(context, FALSE,
                            "MISMATCHED inodes used in superblock, EXPECTED %d, GOT %d",
                            inodes, inodesUsed);
        context->eek.errcount++;
        if (fixIt && fs_eek_query(context, "Fix?")) {
            NFX_SPIN_LOCK(&context->fileSystem->superLock);
            context->fileSystem->super->inodesUsed = inodes;
            NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
            fs_writeSuperblocksSync(context, context->fileSystem->super);
        }
    }
}

/*
 * Keep a side-tally of the link count of an inode whose dinode.linkCnt is greater than
 * one. If the dinode.linkCnt is one, this routine will never get called but a
 * mismatch will get detected in fs_eek_mark() when more than one directory entry
 * refers to that inode.
 *
 * XXX: TODO: convert linked-list implementation to one using skip lists.
 */
void
fs_eek_increment_linkcount(fs_Context *context, uint64 parent, uint64 inum)
{
	fs_Eek *eek = &context->eek;
	fs_Eek_linkcounts *link = eek->linkcounts;

	while (link != NULL) {
		if ((link->inum == inum) && (link->parent == parent)) {
			link->count++;
			return;
		}
		link = link->next;
	}
	link = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof *link);
	link->inum = inum;
	link->parent = parent;
	link->count = 1;
	link->visited = FALSE;
	link->next = eek->linkcounts;
	eek->linkcounts = link;
}

/*
 * Return number of links found for parent/inum pair
 */
uint32
fs_eek_find_linkcount(fs_Context *context, uint64 inum)
{
    fs_Eek *eek = &context->eek;
    fs_Eek_linkcounts *link = eek->linkcounts;
    uint32 count = 0;

    while (link != NULL) {
        if (link->inum == inum) {
	    count += link->count;
	    link->visited = TRUE;
        }
        link = link->next;
    }
    return count;
}

/*
 * Return number of links found for parent/inum pair
 */
uint32
fs_eek_find_parent_linkcount(fs_Context *context, uint64 parent, uint64 inum)
{
    fs_Eek *eek = &context->eek;
    fs_Eek_linkcounts *link, **lp = &eek->linkcounts;
    uint32 count;

    while ((link = *lp) != NULL) {
        if ((link->parent == parent) && (link->inum == inum)) {
            *lp = link->next;
            count = link->count;
            fs_ramDealloc(FS_RAM_CORE_MISC, link);
            //printf("eek parent linkcount: found %d for parent %lld ino %lld\n",
            //  count, parent, inum);
            return count;
        }
        lp = &link->next;
    }
    //printf("eek parent linkcount: not found for parent %lld ino %lld\n",
    //      parent, inum);
    return 0;
}

/*
 * Search the linkcounts list for the given inum expected to match the given
 * (greater than one) linkcount. If the inode is not found, flag a mismatch.
 */
void
fs_eek_check_linkcount(fs_Context *context, uint8 *bitmap, fs_Inode *inode, boolean fixIt)
{
	uint64 max_blocks = context->fileSystem->super->numBlks;
	uint32 count;

	count = fs_eek_find_linkcount(context, inode->inum);
	if (count != 0) {
		if (count != inode->dinode.linkCnt) {
			context->eek.errcount++;
			if (fixIt && fs_eek_query(context,
					 "%s: inode %lld: link count: EXPECTED %d, GOT %d. Adjust?",
					 context->fileSystem->super->mountName,
					 inode->inum, count,
					 inode->dinode.linkCnt)) {
				inode->dinode.linkCnt = count;
				fs_inodeWrite(context, inode, INO_WRITE_LOG);
			} else {
				fs_eek_print_status(context, FALSE,
						    "%s: inode %lld: link count: EXPECTED %d, GOT %d.",
						    context->fileSystem->super->mountName,
						    inode->inum, count, inode->dinode.linkCnt);
			}
		}
		clear_bit(context, bitmap, inode->inum, max_blocks);
		return;
	}

	/*
	 * Adjust the link-count to 1 and set the bit in the bitmap so it
	 * gets relinked into lost+found later.
	 */
	set_bit(bitmap, inode->inum, max_blocks);
	context->eek.errcount++;
	if (fixIt) {
        if (fs_eek_query(context, "%s: inode %lld unreferenced. Relink?",
                        context->fileSystem->super->mountName,
                        inode->inum)) {
		    inode->dinode.linkCnt = 1;
            fs_inodeWrite(context, inode, INO_WRITE_LOG);
        }
	} else {
        fs_eek_print_status(context, FALSE, "%s: inode %lld unreferenced.",
                            context->fileSystem->super->mountName,
                            inode->inum);
    }
}

/*
 * Run this at the end of the scan-inodes phase: any inodes left in the linkcounts list
 * have dinode.linkCnt > 1, but no directory entries pointing to them!
 *
 * XXX: this should merge the entries with like inums, and relink them into
 * lost+found.
 *
 * NOTE: this should get run before fs_eek_scavenge_parent_linkcounts(); the latter
 * function frees the data structure.
 */
void
fs_eek_scavenge_linkcounts(fs_Context *context)
{
	fs_Eek *eek = &context->eek;
	fs_Eek_linkcounts *link = eek->linkcounts;
	
	while (link != NULL) {
		if (!link->visited) {
			uint32 count = fs_eek_find_linkcount(context, link->inum);
			fs_eek_print_status(context, FALSE,
				"%s: MISMATCH ON link count for inode %lld, EXPECTED %d, GOT 0",
				context->fileSystem->super->mountName,
				link->inum, count);
			eek->errcount++;
		}
		link = link->next;
	}
}

/*
 * Run this after eeking FS_LINK_INODE: any parent/inum pair left on this list
 * had no corresponding entry in the FS_LINK_INODE file.
 *
 * The exception is the first one stored via the inode's piu.parentInum.
 * Number of links should be 1 in this case.
 */
void
fs_eek_scavenge_parent_linkcounts(fs_Context *context, uint32 snapnum, boolean fixIt)
{
    eeeRTC_t beginTime, delta;
    fs_Inode *inode;
	fs_Eek *eek = &context->eek;
	fs_Eek_linkcounts *link, **lp = &eek->linkcounts;
	
    eee_rtc(&beginTime);

	while ((link = *lp) != NULL) {

        eee_rtcDuration(&beginTime, &delta);
        if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }

        if (link->count > 1) {
		    fs_eek_print_status(context, FALSE,
                    "MISMATCHED num links entry for parent I#%lld ino I#%lld, EXPECTED %d, GOT 0",
                    link->parent, link->inum, link->count);
		    eek->errcount++;
            if (fixIt && fs_eek_query(context, "Fix?")) {

                // commit what we have so far since addLink may need to
                // acquire FS_LINK_INODE bufs as well
                fs_logCommitEntry(context);
                fs_logBeginTransaction(context);

                inode = I_GET_EXCL(0, link->inum);
                fs_addLink(context, 0, inode, link->parent);
                if (!fs_inodeWasModified(context, inode))
                    fs_inodeReleaseExclusive(context, inode);

                fs_logCommitEntry(context);
                fs_logBeginTransaction(context);
            }
        } else {
            inode = I_GET_EXCL(snapnum, link->inum);
            if (inode == NULL) {
                // dir scan should've caught this case
		        fs_eek_print_status(context, FALSE,
                                    "I#%lld not found for links entry with parent I#%lld",
                                    link->inum, link->parent);
		        eek->errcount++;
            }
            if (inode && !fs_inodeWasModified(context, inode))
                fs_inodeReleaseExclusive(context, inode);
        }
		*lp = link->next;
		fs_ramDealloc(FS_RAM_CORE_MISC, link);
	}
}

/*
 * Compare the inodes found via directory entries (bitmap) against the inode file
 */
void
fs_eek_compare_used_inodes(fs_Context *context, int snapnum, uint8 *bitmap, uint64 num_inodes, boolean fixIt)
{
    eeeRTC_t beginTime, delta;
	uint64 i;

    eee_rtc(&beginTime);

	for (i = 0; i < num_inodes; i += 1) {
		fs_Inode *inode = I_GET_EXCL(snapnum, i);
		fs_Dinode *dinode;

		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }

		if (inode == NULL) {
			static unsigned char zeroes[FS_DINODE_SIZE];
			dinode = (fs_Dinode *)zeroes;
		} else {
			dinode = &inode->dinode;
		}

        // skip QUOTA files, they are not referenced by any directory 
        // clear from bitmap so it doesn't get relinked
        if (inode && (inode->dinode.attrs & FS_IFMT_MASK) == FS_IFMT_QUOT) {
			clear_bit(context, bitmap, i, num_inodes);
			fs_inodeReleaseExclusive(context, inode);
            continue;
        }

		if (dinode->attrs != 0 && !FS_I_IS_DIR(dinode->attrs) &&
		    dinode->linkCnt > 1) {
		  	fs_eek_check_linkcount(context, bitmap, inode, fixIt);
		} else if (dinode->attrs != 0 && get_bit(bitmap, i, num_inodes) == 0) {
			context->eek.errcount++;
			if (fixIt) {
                if (fs_eek_query(context, "%s: (I#%lld) unreferenced. Relink?",
                    context->fileSystem->super->mountName, i))
				    set_bit(bitmap, i, num_inodes);
			}
            else {
                fs_eek_print_status(context, FALSE, "%s: (I#%lld) unreferenced.",
                                    context->fileSystem->super->mountName, i);
            }
		} else if (dinode->attrs == 0 && get_bit(bitmap, i, num_inodes) != 0) {
			/* should catch this case in dirleaf.c */
            fs_eek_print_status(context, FALSE,
                    "%s: (I#%lld) UNALLOCATED but reference from a directory entry",
                    context->fileSystem->super->mountName, i);
		} else {
			clear_bit(context, bitmap, i, num_inodes);
		}
		if (inode != NULL && !fs_inodeWasModified(context, inode)) {
			fs_inodeReleaseExclusive(context, inode);
		}
        else {
            if (inode) assert(context->eek.interactive == TRUE);
        }
	}
	
	fs_eek_scavenge_linkcounts(context);
}

/*
 * XXX: TODO: merge this with "compare used inodes".
 */
void
fs_eek_check_summary(fs_Context *context, uint64 num_inodes)
{
    eeeRTC_t beginTime, delta;
    fs_Inode *inodes = I_GET_EXCL(0, FS_INODE_INODE);
	fs_Inode *inosum = I_GET_EXCL(0, FS_FULL_INODE_BLOCK_INODE);
    fs_Inode *dirsum = I_GET_EXCL(0, FS_DIRECTORY_IN_INODE_BLOCK_INODE);
    fs_Eek *eek = &context->eek;
	uint64 i;
    uint64 iblks, isize, inodeFileBlks, inodeFileSize;
    int32 b, n, firstBits = FS_MAX_RESERVED_INODE / FS_INODES_PER_PAGE;

    /*
     * Verify the first bits (for reserved inodes) are always set
     */
    for (b = 0, n = 0; n < FS_MAX_RESERVED_INODE; b++, n += FS_INODES_PER_PAGE) {
        uint64 target = (uint64) b / context->fileSystem->params.inodesPerBlock;
        uint64 ipage = fs_bitmapFindClear(context, inosum, target, NULL);
        uint64 dpage = fs_bitmapFindClear(context, dirsum, target, NULL);
        if (ipage < firstBits) {
            eek->errcount++;
            fs_eek_print_status(context, FALSE,
                    "Reserved bit %lld cleared for inode summary", ipage);
            if (eek->interactive) {
                fs_bitmapSet(context, inosum, b, NULL);
                eek->fixcount++;
            }
        }
        if (dpage < firstBits) {
            eek->errcount++;
            fs_eek_print_status(context, FALSE,
                    "Reserved bit %lld cleared for directory summary", ipage);
            if (eek->interactive) {
                fs_bitmapSet(context, dirsum, b, NULL);
                eek->fixcount++;
            }
        }
    }

    // init inode parameters for the superblock check
    // count the super nodes (page 0)

    iblks = 1;
    isize = FS_BLOCK_SIZE;

    eee_rtc(&beginTime);

	for (i = FS_MAX_RESERVED_INODE; i < num_inodes; i += FS_INODES_PER_PAGE) {
		fs_Buf *buf = O_GET_EXCL(inodes, i / FS_INODES_PER_PAGE);
		fs_Buf *ibuf = O_GET_EXCL(inosum, i / (FS_INODES_PER_PAGE * PAGE_SIZE * 8));
		fs_Buf *dbuf = O_GET_EXCL(dirsum, i / (FS_INODES_PER_PAGE * PAGE_SIZE * 8));
		int bitnum = (i / FS_INODES_PER_PAGE) % (PAGE_SIZE * 8);
		fs_Dinode *dinodes;
		boolean full, hasdir;
		int j;

		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }

		if (buf == NULL) {
			static unsigned char zeroes[PAGE_SIZE];
			dinodes = (fs_Dinode *)zeroes;
		} else {
			dinodes = buf->u.dinode->dinode;
            // keep track of inode parameters to check against superblocks'
            iblks += 1;
            isize = ((i / FS_INODES_PER_PAGE) + 1) * FS_BLOCK_SIZE;
		}

		full = TRUE;
		hasdir = FALSE;
		for (j = 0; j < FS_INODES_PER_PAGE; j++) {
			full &= (dinodes[j].attrs != 0);
			hasdir |= FS_I_IS_DIR(dinodes[j].attrs);
		}

		if (!compare_bits(context, "inode summary", ibuf->u.data, bitnum,
                        -full, context->fileSystem->super->numBlks)) {
			/*
			 * Find and print the inode numbers of the free inodes
			 * in this page.
			 */
			if (!full) {
				for (j = 0; j < FS_INODES_PER_PAGE; j++) {
					if (dinodes[j].attrs == 0)
						printf("\tinode %lld has mode zero\n", j + i);
				}
			}
            /*
             * Adjust the bit accordingly
             */
			eek->errcount++;
            if (fs_eek_query(context, "Fix?")) {
                if (full) fs_bitmapSet(context, inosum, bitnum, NULL);
                else fs_bitmapClearIfSet(context, inosum, bitnum, NULL);
            }
        }

		if (hasdir) {
			/*
			 * If there is a directory in this page, we
			 * don't care about the value of the dirsum bit --
			 * it could be 1 or 0. This is because removing
			 * a directory clears the dirsum bit even if there
			 * are other directories in the same inode page.
			 */
		} else if (!compare_bits(context, "directory summary", dbuf->u.data, bitnum,
                                 0, context->fileSystem->super->numBlks)) {
            /*
             * Clear the bit since there are no directories in this page
             */
			eek->errcount++;
            if (fs_eek_query(context, "Fix?")) {
                fs_bitmapClearIfSet(context, dirsum, bitnum, NULL);
            }
		}
		
		if (buf != NULL) {
            if (!fs_bufWasModified(context, buf))
		        fs_bufReleaseExclusive(context, buf);
            else
                assert(context->eek.interactive == TRUE);
		}

        if (!fs_bufWasModified(context, ibuf))
		    fs_bufReleaseExclusive(context, ibuf);
        else
            assert(context->eek.interactive == TRUE);

        if (!fs_bufWasModified(context, dbuf))
		    fs_bufReleaseExclusive(context, dbuf);
        else
            assert(context->eek.interactive == TRUE);
	}

	if (!fs_inodeWasModified(context, inosum))
	    fs_inodeReleaseExclusive(context, inosum);
    else
        assert(context->eek.interactive == TRUE);

	if (!fs_inodeWasModified(context, dirsum))
	    fs_inodeReleaseExclusive(context, dirsum);
    else
        assert(context->eek.interactive == TRUE);

	if (!fs_inodeWasModified(context, inodes))
	    fs_inodeReleaseExclusive(context, inodes);
    else
        assert(context->eek.interactive == TRUE);

    // compare iblks,isize with superblocks' inodeFileBlks,inodeFileSize

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    inodeFileBlks = context->fileSystem->super->inodeFileBlks;
    inodeFileSize = context->fileSystem->super->inodeFileSize;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);

    if (iblks != inodeFileBlks) {
        fs_eek_print_status(context, FALSE,
                            "MISMATCHED inodeFileBlks in superblock, EXPECTED %d, GOT %d",
                            iblks, inodeFileBlks);
        context->eek.errcount++;
        if (eek->interactive && fs_eek_query(context, "Fix?")) {
            NFX_SPIN_LOCK(&context->fileSystem->superLock);
            context->fileSystem->super->inodeFileBlks = iblks;
            NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
            fs_writeSuperblocksSync(context, context->fileSystem->super);
        }
    }

    if (isize != inodeFileSize) {
        fs_eek_print_status(context, FALSE,
                            "MISMATCHED inodeFileSize in superblock, EXPECTED %d, GOT %d",
                            isize, inodeFileSize);
        context->eek.errcount++;
        if (eek->interactive && fs_eek_query(context, "Fix?")) {
            NFX_SPIN_LOCK(&context->fileSystem->superLock);
            context->fileSystem->super->inodeFileSize = isize;
            NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
            fs_writeSuperblocksSync(context, context->fileSystem->super);
        }
    }
}

/*
 * Steps to eek a snapshot:
 *
 * 1. The emap btree of the snapnum's inode file is
 *    verified before accessing it's contents (via O_GET).
 *    The snapbits file was already eek'd in fs_eek_check_refcounts().
 * 2. Perform some basic sanity checks of the snapbits file since
 *    everything is reconciled to this (snapbits file == truth)
 * 3. Compare this snapshot's used blocks against the snapbits file
 * 4. Compare this snapshot's inodes found via directory entires
 *    against the snapshot's inode file
 *
 * No corrections are made to a snapshot.
 * If any inconsistency is found, the snapshot is removed.
 */
void
fs_eek_snapshot(
	fs_Context *context,
	uint8 *bitmap,
	int snapnum,
	uint64 num_blocks,
	uint64 num_inodes)
{
	fs_Inode *snapbits = I_GET_EXCL(snapnum, FS_SNAPSHOT_BITS_INODE);
    int save_errors, errors;

	if (snapbits == NULL)
		return;

    /*
     * Just count errors from eek'ing this snapshot
     * Save current and restore when done
     */

    save_errors = context->eek.errcount;
    context->eek.errcount = 0;
	
    /*
     * verify integrity of emap btree for the snapnum's inode file
     */

    if (!fs_eek_meta_inode(context, FS_INODE_INODE, snapnum)) {
        fs_eek_print_status(context, FALSE,
                            "%s: INVALID emap for inode file (snapnum %d)",
                            context->fileSystem->super->mountName,
                            snapnum);
        context->eek.errcount++;
        goto fs_eek_snapshot_exit;
    }

	fs_eek_print_status(context, TRUE,
            "== %s == snapshot %d ==",
            context->fileSystem->super->mountName, snapnum);

	fs_eek_print_status(context, TRUE,
            "== %s == mark used blocks ==", 
           context->fileSystem->super->mountName);

    /*
     * Compare this snapshot's used blocks against the snapbits file
     */

	if (!fs_eek_mark_used_blocks(context, bitmap, snapnum, num_inodes, FALSE, &errors)) {
        return;
    }

    context->eek.errcount += errors;

	fs_eek_print_status(context, TRUE,
            "== %s == compare used blocks ==",
            context->fileSystem->super->mountName);
	fs_eek_compare_used_blocks_by_snapbits(context, bitmap, snapbits, num_blocks);

    // abort if any errors encountered after marking and comparing blocks
    // since the mark used inodes phase assumes no duplicate block errors
    // or they've been fixed

    if (context->eek.errcount) {
        goto fs_eek_snapshot_exit;
    }

    /*
     * Compare this snapshot's inodes found via directory entires
     * against the snapshot's inode file
     */

	fs_eek_print_status(context, TRUE,
            "== %s == mark used inodes ==",
            context->fileSystem->super->mountName);
	fs_eek_mark_used_inodes(context, bitmap, snapnum, num_inodes, FALSE);

	fs_eek_print_status(context, TRUE,
            "== %s == compare used inodes ==", 
           context->fileSystem->super->mountName);
	fs_eek_compare_used_inodes(context, snapnum, bitmap, num_inodes, FALSE);

	fs_eek_print_status(context, TRUE, "== %s == verify link inode ==", 
                        context->fileSystem->super->mountName);
    if (!fs_eek_meta_inode(context, FS_LINK_INODE, snapnum)) {
        fs_eek_print_status(context, FALSE, "UNRECOVERAGE error: invalid emap for link inode");
        context->eek.errcount++;
        goto fs_eek_snapshot_exit;
    }
    fs_eek_link_inode(context, snapnum, FALSE);

fs_eek_snapshot_exit:

    fs_inodeReleaseExclusive(context, snapbits);

	fs_eek_print_status(context, FALSE,
            "== %s == snapshot %d done == Errors: %d",
            context->fileSystem->super->mountName, snapnum,
            context->eek.errcount);
    
    /*
     * snapshots cannot be corrected, query for removal
     */

    if (context->eek.errcount) {
        if (context->eek.interactive) {
            if (fs_eek_query(context,
                            "%s: CORRUPTED snapshot %d. Remove?",
                            context->fileSystem->super->mountName, snapnum)) {
	            fs_logCommitEntry(context);
                fs_snapRemoveSnapnum(context, snapnum);
	            fs_logBeginTransaction(context);
            }
        } else {
            fs_eek_print_status(context, FALSE,
                                "%s: CORRUPTED snapshot %d.",
                                context->fileSystem->super->mountName, snapnum);
        }
    }

    /* restore error count */
    context->eek.errcount = save_errors;
}

/*
 * Load the refcounts file and all the snap bits files, checking that the snap
 * bits sum up to the correct refcount.
 */
boolean
fs_eek_check_refcounts(fs_Context *context,
               uint8 *bitmap,
		       uint64 num_blocks)
{
    eeeRTC_t beginTime, delta;
    fs_Inode *refcounts = I_GET_SHARED(0, FS_REF_COUNT_INODE);
	fs_Inode *snapbits[FS_MAX_SNAPSHOTS];
	uint8 count[PAGE_SIZE];
    int32 max_bit_this_page;
	int j, k;
    uint64 i, max_block, blks = 0, blksUsed;
    boolean ok = TRUE;
    boolean fixIt = fs_eek_is_interactive(context);

    bzero(snapbits, sizeof(snapbits));

	for (j = 1; j < FS_MAX_SNAPSHOTS; j++) {
        if (!fs_eek_meta_inode(context, FS_SNAPSHOT_BITS_INODE, j)) {
            fs_eek_print_status(context, FALSE,
                                "%s: INVALID emap for snapbits file (snapnum %d)",
                                context->fileSystem->super->mountName, j);
            context->eek.errcount++;
            ok = FALSE;
            // TBD: remove snapshot ??
            goto fs_eek_check_refcounts_exit;
        } else {
		    snapbits[j] = I_GET_EXCL(j, FS_SNAPSHOT_BITS_INODE);
            // TBD: basic sanity check of snapbits file
        }
    }
	
    eee_rtc(&beginTime);

	for (i = 0; i < num_blocks; i += PAGE_SIZE) {
		fs_Buf *ref = O_GET_EXCL(refcounts, i / PAGE_SIZE);

		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
        }

        if (i + PAGE_SIZE > num_blocks)
            max_block = num_blocks % PAGE_SIZE;
        else
            max_block = PAGE_SIZE;
        /*
         * verify valid ref bytes
         */
        for (j = 0; j < max_block; j++) {
            fs_RefByte byte = ref->u.refBytes[j];
            uint32 snapcount = fs_refbyteSnapCount(byte);

            if (snapcount > FS_REFBYTE_SNAP_MAX) {
                fs_eek_print_status(context, FALSE,
                        "%s: INVALID ref byte %x at page %lld offset %d",
                        context->fileSystem->super->mountName,
                        byte, i / PAGE_SIZE, j);
                context->eek.errcount++;
                ok = FALSE;
                goto fs_eek_check_refcounts_exit;
            }
            // keep track of blks used to eek against the superblk's version
            if (byte != 0) {
                blks++;
            }
        }
		bzero(count, sizeof count);
		for (j = 1; j < FS_MAX_SNAPSHOTS; j++) {
			uint8 *bitmap;
			fs_Buf *buf;
			uint64 rbyte = i % (PAGE_SIZE*8);
			
			if (snapbits[j] == NULL)
				continue;

			buf = O_GET_EXCL(snapbits[j], i / (PAGE_SIZE*8));
            if (buf == NULL) {
                continue;  // growfs extras
            }
			bitmap = buf->u.data;
				
            /*
             * At the end of the filesystem, don't check any uninitialized
             * refbytes which might lie in the remainder of the final page.
             */
            if (i + PAGE_SIZE > num_blocks)
                max_bit_this_page = num_blocks % PAGE_SIZE;
            else
                max_bit_this_page = PAGE_SIZE;

			for (k = 0; k < max_bit_this_page; k++)
				count[k] += get_bit(bitmap, rbyte+k, context->fileSystem->super->numBlks);

            if (!fs_bufWasModified(context, buf))
		        fs_bufReleaseExclusive(context, buf);
            else
                assert(context->eek.interactive == TRUE);
		}
		for (j = 0; j < PAGE_SIZE; j++) {
			if (count[j] != fs_refbyteSnapCount(ref->u.data[j])) {
				fs_eek_print_status(context, FALSE,
                        "%s: MISMATCH ON refcount %lld, GOT %d, EXPECTED %d",
                        context->fileSystem->super->mountName,
						i + j, fs_refbyteSnapCount(ref->u.data[j]), count[j]);
				context->eek.errcount++;
                ok = FALSE;
                break;
			}
		}
        if (!fs_bufWasModified(context, ref))
		    fs_bufReleaseExclusive(context, ref);
        else
            assert(context->eek.interactive == TRUE);
        if (!ok) break;
	}


    // compare blks found with the value in the superblock

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    blksUsed = context->fileSystem->super->blksUsed;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    if (blks != blksUsed) {
        fs_eek_print_status(context, FALSE,
                            "MISMATCHED blocks used count in superblock, EXPECTED %lld, GOT %lld",
                            blks, blksUsed);
        context->eek.errcount++;
        if (fixIt && fs_eek_query(context, "Fix?")) {
            NFX_SPIN_LOCK(&context->fileSystem->superLock);
            context->fileSystem->super->blksUsed = blks;
            NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
            fs_writeSuperblocksSync(context, context->fileSystem->super);
        }
    }

fs_eek_check_refcounts_exit:

	for (j = 1; j < FS_MAX_SNAPSHOTS; j++) {
		if (snapbits[j] != NULL) {
	        if (!fs_inodeWasModified(context, snapbits[j]))
		        fs_inodeReleaseExclusive(context, snapbits[j]);
            else
                assert(context->eek.interactive == TRUE);
        }
	}

    fs_inodeRelease(context, refcounts);
    return ok;
}

boolean
fs_eek_refbyte_is_used(fs_RefByte refbyte)
{
	return fs_refbyteIsLive(refbyte);
}

/*
 * Mark and compare used blocks
 * Return TRUE if success
 */
boolean
fs_eek_used_blocks(fs_Context *context, uint8 *bitmap, boolean fixIt)
{
    fs_Inode *refcounts = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Parameters p;
    fs_Eek_bits bits;
    uint64 num_blocks = context->fileSystem->super->numBlks;
    uint64 num_inodes;

    p = fs_computeSizeParameters(context->fileSystem->super);
    num_inodes = p.maxInodeNumber;
    context->eek.freecount = 0;

    /*
     * Mark and compare all blocks used by inodes
     * This phase does NOT make any corrections since the buddy has
     * not been reconciled with the reference file at this point
     */

	fs_eek_print_status(context, TRUE, "== %s == mark used blocks ==", 
                        context->fileSystem->super->mountName);
	if (!fs_eek_mark_used_blocks(context, bitmap, 0, num_inodes, fixIt,
            (fixIt ? NULL : &context->eek.passOneErrors))) {
        return FALSE;
    }

	fs_eek_print_status(context, TRUE, "== %s == compare used blocks ==", 
                        context->fileSystem->super->mountName);
	//bits.extract = fs_refbyteIsLive;
	bits.extract = fs_eek_refbyte_is_used;
	bits.query = fs_eek_used_query;
	fs_eek_compare_used_blocks_by_refcounts(context, &bits, bitmap, refcounts, num_blocks);

    fs_inodeRelease(context, refcounts);

    /* 
     * Make sure all deallocated blocks are no longer in the buffer cache
     * (via fs_eek_used_query)
     */
    if (context->eek.freecount) {
        fs_bufCacheFlush(context, NULL);
    }

    return TRUE;
}

/*
 * Mark and compare free blocks
 * Return TRUE if success
 */
boolean
fs_eek_free_blocks(fs_Context *context, uint8 *bitmap, boolean fixIt)
{
    fs_Inode *refcounts = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Inode *buddy = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    uint64 maxBuddyRegions = ceilingDiv64(buddy->dinode.size, FS_BLOCK_SIZE);
    uint64 num_blocks = context->fileSystem->super->numBlks;
    fs_Eek_bits bits;

	fs_inodeRelease(context, buddy);

	/*
	 * Make sure binary buddy tree has an accurate count of free
	 * blocks by marking which blocks are listed as free, and
	 * then testing against the COMPLEMENT of the refcounts bits.
	 */

    context->eek.freecount = 0;
	bzero(bitmap, ceilingDiv64(num_blocks, 8));

	fs_eek_print_status(context, TRUE, "== %s == mark free blocks ==", 
                        context->fileSystem->super->mountName);
	fs_eek_mark_free_blocks(context, bitmap);

	fs_eek_print_status(context, TRUE, "== %s == compare free blocks ==", 
                        context->fileSystem->super->mountName);
    bzero(&bits, sizeof(bits));
	bits.extract = fs_eek_refbyte_is_free;
	bits.query = fs_eek_free_query;
    bits.maxBuddyRegions = maxBuddyRegions;
    // allocate bitmap for max buddy regions -> div 8
    bits.buddyBitmap = fs_ramAlloc(FS_RAM_CORE_MISC, ceilingDiv64(maxBuddyRegions, 8));
    if (bits.buddyBitmap == NULL) {
        fs_eek_print_status(context, FALSE,
                            "%s: FATAL: not enough memory to eek free blocks",
					        context->fileSystem->super->mountName);
        fs_inodeRelease(context, refcounts);
        return FALSE;
    }
    bzero(bits.buddyBitmap, ceilingDiv64(maxBuddyRegions, 8));

	/*
	 * roundup: how many pages in the refbytes to skip when we rebuild
	 * a buddy region. i.e., rebyte pages per buddy regions.
	 *
	 * XXX: FS_BLOCK_SIZE denotes the number of blocks mapped by
	 * a page of refbytes. Should be a different manifest constant
	 * for this? e.g., BLOCKS_PER_REFBYTE_PAGE.
	 */
	bits.roundup = FS_BLOCKS_PER_BUDDY_REGION / FS_BLOCK_SIZE;
	fs_eek_compare_used_blocks_by_refcounts(context, &bits, bitmap, refcounts, num_blocks);

    fs_inodeRelease(context, refcounts);

    /* 
     * Make sure all deallocated blocks are no longer in the buffer cache
     * (via fs_eek_mark_free_blocks and fs_eek_free_query)
     */
    if (context->eek.freecount) {
        fs_bufCacheFlush(context, NULL);
    }
    fs_ramDealloc(FS_RAM_CORE_MISC, bits.buddyBitmap);

    return TRUE;
}

/*
 * Steps to eek a filesystem:
 *
 * 1. The emap btree of meta inodes (eg FS_INODE_INODE, FS_REF_COUNT_INODE)
 *    are verfified before accessing it's contents (via O_GET).
 * 2. Perform some basic sanity checks of the reference file since
 *    everything is reconciled to this (reference file == truth):
 *      + verify valid reference bytes
 *      + verify emap btrees for all snapbits files
 *      + read snapbits files and verify correct reference snapcounts
 * 3. For all inodes in FS_INODE_INODE file, mark all blocks used by the
 *    inode including the btree nodes.
 *    After step 3. all used blocks are marked in the bitmap.
 * 4. Reconcile all used blocks with the reference file.
 *    Adjustments are made to reconcile used blocks with the
 *    reference bytes and the free buddy tree:
 *      + if blocknum is set in the bitmap but the reference byte
 *        is clear, set the reference byte
 *      + if blocknum is not set but the reference byte is, 
 *        clear reference byte and add block back to buddy
 *        (just clear live if snapcount is non zero)
 * 5. Mark all free blocks according to the buddy free tree.
 *    After step 5. all free blocks are marked in the bitmap.
 * 6. Reconcile the free buddy blocks with the reference file.
 *    Adjustments are made to reconcile the buddy with the
 *    reference bytes:
 *      + if blocknum is not consistent with the reference byte,
 *        rebuild the buddy region for blocknum
 * 
 * At this point, the reference file has been reconciled with the
 * free blocks and the used blocks in the system.
 *
 * If eek is in interactive mode, corrections can now be made to
 * any inconsistencies found.
 *    
 * 7. Fix all btree inconsistencies:
 *      + if emap, corrupted blocks are free'd and inode is deallocated
 *      + if directory, corrupted blocks are free'd and the directory indexer
 *        is recalculated if neccessary
 *
 * How the directory indexer is recalculated:
 *      + for all indices, get the hash codes
 *      + for each hash, verify a directory entrie exsists with the same hash
 *        (using the btree eek scan mechanism)
 *      + remove directory index for hash if not found
 *
 * A note here about btree corrections and cow buffers:
 *  The current buffer cache API (fs_bufWrite()) does not allow writes
 *  to meta blocks (ie btree blocks).  The btree will cow on demand
 *  when corrections need to be made to a cow buffer.  The new cow'd block
 *  will be passed back to the parent so it's key/block pair can be
 *  updated with the new blocknum.  This can trickle cow all the way back
 *  up to the root and an inode write can trigger more cow for the
 *  buffer containing the inode.  This cow mechanism requires a valid
 *  INODE file which is assumed here in step 7.
 *  After this correction, the live's version will be fixed but the
 *  corruption is still with the snapshot's version.  For now, the
 *  snapshot is removed if any inconsistencies are found.
 *
 * 8. Mark and verify inodes (inodes referenced by some directory).
 *    Also include the following directory meta inodes:
 *      + the "hash chain head" inode allocated for directory hash collisions
 *      + the directory indexer
 *        (for hash collisions or directory leaf splits)
 *    Make the following verifications:
 *      + verify valid inum in a directory entry
 *      + verify inum in a directory entry is allocated (via FS_INODE_INODE)
 *      + verify a directory index is allocated if necc for a directory entry
 *      If any inconsistencies found, the directory entries are removed
 *      and the directory indexer if recalculated if neccessary.
 *    After step 8, all inodes are marked in the bitmap.
 * 9. Reconcile the used inodes against the inode file:
 *      + if inode is allocated via FS_INODE_INODE but the inum is
 *        not set in the bitmap, relink this "dangling" inode to lost+found
 * 10. Reconcile the inode and directory summary files:
 *      + verify the first bits reserved for the meta inodes are set
 *      + verify bit is cleared in the inode summary if a page is not full
 *      + verify bit is set in the inode summary if a page is full
 *      + verify bit is cleared in the directory summary if page does
 *        not contain any directories
 */

int
fs_eek(fs_Context *context)
{
    int rc;
	fs_Eek *eek = &context->eek;
	fs_Lock lock;
	fs_SuperBlock *super;
	uint64 num_blocks, num_inodes;
	fs_Parameters p;
	uint8 *bitmap = NULL;
	int i;
    fs_FileSystemState oldState;
    eeeGMT_t date;
    eeeGMT_t startDate;

	if (context->fileSystem == NULL) {
		fs_eek_print_status(context, FALSE, "No Filesystem available for eek");
		return NFX_ERR;
	}
	if (context->fileSystem->super == NULL) {
		fs_eek_print_status(context, FALSE, "No in core super block available for eek");
		return NFX_ERR;
	}
    if (context->fileSystem->state == FS_VOL_STATE_PAUSED) {
		fs_eek_print_status(context, FALSE, "Eek already running for %s",
                            context->fileSystem->super->mountName);
        return NFX_OK;
    }

	context->eek.lock = &lock;
	fs_lockInit(context->eek.lock, FS_LOCK_TYPE_EEK);

    /*
     * Set inUse so inode and buffer cache modules will not log eek mods
     */
	eek->inUse = 1;

    eee_TimeGMT (&startDate);

    oldState = context->fileSystem->state;
    context->fileSystem->state = FS_VOL_STATE_PAUSED;
	super = context->fileSystem->super;
	num_blocks = super->numBlks;
	eek->errcount = 0;

	p = fs_computeSizeParameters(super);
	num_inodes = p.maxInodeNumber;

	/*
	 * Allocate a bitmap for the size of the filesystem.
	 */

	bitmap = fs_ramAllocFail(FS_RAM_CORE_MISC, ceilingDiv64(num_blocks, 8));
    if (bitmap == NULL) {
        fs_eek_print_status(context, FALSE,
                            "%s: FATAL: not enough memory for eek",
					        context->fileSystem->super->mountName);
        eek->errcount++;
        goto fs_eek_exit;
    }
	bzero(bitmap, ceilingDiv64(num_blocks, 8));

	fs_logBeginTransaction(context);
	FS_LOCK_ORDER_DISABLE(context);

    /*
     * Verify integrity of the meta inodes' emap btrees
     * some things should not be corrected
     */

	fs_eek_print_status(context, TRUE, "== %s == verifying meta inodes ==", 
                        context->fileSystem->super->mountName);
    if (!fs_eek_meta_inode(context, FS_REF_COUNT_INODE, 0)) {
        fs_eek_print_status(context, FALSE,
                            "%s: UNRECOVERABLE error: invalid emap for reference file",
                            context->fileSystem->super->mountName);
        eek->errcount++;
        goto fs_eek_exit;
    }
    if (!fs_eek_meta_inode(context, FS_FREE_BUDDY_TREE_INODE, 0)) {
        fs_eek_print_status(context, FALSE,
                            "%s: UNRECOVERABLE error: invalid emap for buddy free tree file",
                            context->fileSystem->super->mountName);
        eek->errcount++;
        goto fs_eek_exit;
    }
    if (!fs_eek_meta_inode(context, FS_INODE_INODE, 0)) {
        fs_eek_print_status(context, FALSE,
                            "%s: UNRECOVERABLE error: invalid emap for inode file",
                            context->fileSystem->super->mountName);
        eek->errcount++;
        goto fs_eek_exit;
    }

    /*
     * Perform some basic sanity checks of the reference file
     * since everything is reconciled to it
     */

	fs_eek_print_status(context, TRUE, "== %s == mark and compare reference counts ==", 
                        context->fileSystem->super->mountName);
	if (!fs_eek_check_refcounts(context, bitmap, num_blocks)) {
        fs_eek_print_status(context, FALSE, "%s: UNRECOVERABLE errors found in reference file",
                            context->fileSystem->super->mountName);
        goto fs_eek_exit;
    }

    /*
     * Eek used and free blocks
     * Reconcile used and free with the reference file
     */

    if (!fs_eek_used_blocks(context, bitmap, FALSE) ||
        !fs_eek_free_blocks(context, bitmap, FALSE)) {
        goto fs_eek_exit;
    }

    /*
     * Continue with the "fix it" phase now that the buddy is made
     * consistent with reference file
     */

    if (fs_eek_is_interactive(context)) {
	    fs_eek_print_status(context, TRUE, "== %s == correcting used blocks ==", 
                            context->fileSystem->super->mountName);
	    fs_eek_fix_used_blocks(context, bitmap, 0, num_inodes);
	    fs_logCommitEntry(context);
	    fs_logBeginTransaction(context);
    } else {
        /*
         * Include the errors during the "mark it" phase
         * total errcount was cleared after this phase to avoid
         * double counting the same errors
         */
        context->eek.errcount += context->eek.passOneErrors;
    }

    /*
     * Mark and compare inodes
     */

	fs_eek_print_status(context, TRUE, "== %s == mark used inodes ==", 
                        context->fileSystem->super->mountName);
	fs_eek_mark_used_inodes(context, bitmap, 0, num_inodes, context->eek.interactive);

	fs_eek_print_status(context, TRUE, "== %s == compare used inodes ==", 
                        context->fileSystem->super->mountName);
	fs_eek_compare_used_inodes(context, 0, bitmap, num_inodes, context->eek.interactive);
	fs_eek_relink_lost_inodes(context, bitmap, num_inodes);

	fs_eek_print_status(context, TRUE, "== %s == verify link inode ==", 
                        context->fileSystem->super->mountName);
    if (!fs_eek_meta_inode(context, FS_LINK_INODE, 0)) {
        // since this cannot be fixed, it's considered an unrecoverable error
        // but continue eek'ing rest of system
        fs_eek_print_status(context, FALSE, "UNRECOVERAGE error: invalid emap for link inode");
        context->eek.errcount++;
    } else {
        fs_eek_link_inode(context, 0, fs_eek_is_interactive(context));
    }

	fs_logCommitEntry(context);
	fs_logBeginTransaction(context);

    /*
     * Eek inode and directory summaries
     */

	fs_eek_print_status(context, TRUE, "== %s == mark and compare summary bits ==", 
                        context->fileSystem->super->mountName);
	fs_eek_check_summary(context, num_inodes);

	fs_eek_print_status(context, TRUE, "== %s == eek snapshots ==", 
                        context->fileSystem->super->mountName);
	
    /*
     * Eek snapshots
     */

	for (i = 1; i < FS_MAX_SNAPSHOTS; i++)
		fs_eek_snapshot(context, bitmap, i, num_blocks, num_inodes);
		
fs_eek_exit:

    if (bitmap) {
	    fs_ramDealloc(FS_RAM_CORE_MISC, bitmap);
	    fs_logCommitEntry(context);
    }

	FS_LOCK_ORDER_ENABLE(context);

    eee_TimeGMT (&date);

    fs_eek_print_status(context, TRUE, "== %s == done == Errors: %d Fixed: %d ==", 
                        context->fileSystem->super->mountName,
                        eek->errcount, eek->fixcount);
    fs_eek_print_status(context, FALSE, "  (%d hour, %d min, %d sec)", 
            (((date.sec - startDate.sec)) / SEC_PER_HOUR),
            (((date.sec - startDate.sec) % SEC_PER_HOUR) / SEC_PER_MIN),
            ((date.sec - startDate.sec) % SEC_PER_MIN));

	eek->inUse = 0;
    context->fileSystem->state = oldState;

    /*
     * TBD: truncate the log after a successful eek
     */

    if (eek->errcount <= eek->fixcount) {
        rc = NFX_OK;
    } else {
        rc = NFX_ERR;
    }

	return rc;
}


/**
 * Perform a mini-eek, just verifying the buddy file
 * agrees with the refbytes file.
 */
int
fs_eek_ref_vs_buddy(fs_Context *context)
{
	fs_Eek *eek = &context->eek;
	fs_SuperBlock *super;
	fs_Inode *refcounts;
	uint64 num_blocks;
	fs_Parameters p;
	fs_Eek_bits bits;
	uint8 *bitmap;
    fs_FileSystemState oldState;
    fs_Inode *buddy = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    uint64 maxBuddyRegions = ceilingDiv64(buddy->dinode.size, FS_BLOCK_SIZE);
    eeeGMT_t date;
    eeeGMT_t startDate;

    fs_inodeRelease(context, buddy);

	if (context->fileSystem == NULL) {
		printf("No Filesystem available for eek\n");
		eek->inUse = 0;
		return NFX_ERR;
	}
	if (context->fileSystem->super == NULL) {
		printf("No in core super block available for eek\n");
		eek->inUse = 0;
		return NFX_ERR;
	}

    /*
     * inUse set so inode and buffer cache modules will not log eek mods
     */
	eek->inUse = 1;

    eee_TimeGMT (&startDate);

    oldState = context->fileSystem->state;
    context->fileSystem->state = FS_VOL_STATE_PAUSED;

	super = context->fileSystem->super;

	num_blocks = super->numBlks;
	eek->errcount = 0;
	eek->passOneErrors = 0;

	p = fs_computeSizeParameters(super);

	/*
	 * Allocate a bitmap for the size of the filesystem.
	 */
	bitmap = fs_ramAlloc(FS_RAM_CORE_MISC, ceilingDiv64(num_blocks, 8));
	assert(bitmap != NULL);
	bzero(bitmap, ceilingDiv64(num_blocks, 8));

	fs_logBeginTransaction(context);

	FS_LOCK_ORDER_DISABLE(context);

	/*
	 * Make sure binary buddy tree has an accurate count of free
	 * blocks by marking which blocks are listed as free, and
	 * then testing against the COMPLEMENT of the refcounts bits.
	 */
	fs_eek_print_status(context, TRUE, "== %s == min-eek mark free blocks ==", 
                        context->fileSystem->super->mountName);
	fs_eek_mark_free_blocks(context, bitmap);

	fs_eek_print_status(context, TRUE, "== %s == min-eek compare free blocks ==", 
                        context->fileSystem->super->mountName);

	refcounts = I_GET_EXCL(0, FS_REF_COUNT_INODE);

    bzero(&bits, sizeof(bits));
	bits.extract = fs_eek_refbyte_is_free;
	bits.query = fs_eek_free_query;
    bits.maxBuddyRegions = maxBuddyRegions;
    // allocate bitmap for max buddy regions -> div 8
    bits.buddyBitmap = fs_ramAlloc(FS_RAM_CORE_MISC, ceilingDiv64(maxBuddyRegions, 8));
    assert(bits.buddyBitmap);
    bzero(bits.buddyBitmap, ceilingDiv64(maxBuddyRegions, 8));
	/*
	 * roundup: how many pages in the refbytes to skip when we rebuild
	 * a buddy region. i.e., rebyte pages per buddy regions.
	 *
	 * XXX: FS_BLOCK_SIZE denotes the number of blocks mapped by
	 * a page of refbytes. Should be a different manifest constant
	 * for this? e.g., BLOCKS_PER_REFBYTE_PAGE.
	 */
	bits.roundup = FS_BLOCKS_PER_BUDDY_REGION / FS_BLOCK_SIZE;
	fs_eek_compare_used_blocks_by_refcounts(context, &bits, bitmap, refcounts, num_blocks);

	if (!fs_inodeWasModified(context, refcounts))
		fs_inodeReleaseExclusive(context, refcounts);
    else
        assert(context->eek.interactive == TRUE);

    fs_ramDealloc(FS_RAM_CORE_MISC, bits.buddyBitmap);
	fs_ramDealloc(FS_RAM_CORE_MISC, bitmap);

	fs_logCommitEntry(context);

	FS_LOCK_ORDER_ENABLE(context);

    eee_TimeGMT (&date);

    fs_eek_print_status(context, TRUE, "== %s == mini-eek done == Errors: %d Fixed: %d ==",
                        context->fileSystem->super->mountName,
                        eek->errcount, eek->fixcount);
    fs_eek_print_status(context, FALSE, "  (%d hour, %d min, %d sec)", 
            (((date.sec - startDate.sec)) / SEC_PER_HOUR),
            (((date.sec - startDate.sec) % SEC_PER_HOUR) / SEC_PER_MIN),
            ((date.sec - startDate.sec) % SEC_PER_MIN));

	eek->inUse = 0;
    context->fileSystem->state = oldState;

	return (eek->errcount == 0) ? NFX_OK : NFX_ERR;
}

#endif // SSC
