/*
 * Name		: fs-evm.c
 * RCS		: $ID$
 * Description	: FS-EVM grow interface
 * Created by	: Rahul Khona
 * Date Created	: 28/01/2002
 */
#include "fs-includes.h"
#if defined(SSC)
#include <time.h>
#endif
#include "fs-evm.h"
#ifndef FS_STANDALONE
#include "exports/evm-api.h"

static int
send_evm_message(void *msg, int msgLen)
{
    int                 rc;
        rc = eee_sendMessage(msg, msgLen,
			     PACK_UNICAST_PORT_ADDRESS(nfxSscSlotId,
				  NFX_SSC_UNIX,
				  eee_getApplicationId("evm"), EEE_MGMT,
				  NFX_COS_HIGH,PORT_BEST_FIT),
                        	  fs_packedPortAddr(nfxMyCPUId, EEE_MGMT));


    if (rc < 0) {
        printf("Failed to communicate with EVM\n");
        return NFX_ERR;
    }
    return NFX_OK;
}
#endif

/*
 * Set the minimum free high water mark
 */

int32
fs_setMinFreeThread(
    fs_Context      *context
)
{
    fs_MsgSetMinFreeReq  *tune;
    fs_MsgSetMinFreeResp *resp;
    fs_FileSystem   *fs = context->fileSystem;

	resp = (fs_MsgSetMinFreeResp *)context->respHdr;

    tune = (fs_MsgSetMinFreeReq *)context->reqHdr;
    if (tune->minFree > 100) {
        resp->hdr.fsStatus = FS_ERR_INVAL;
        return NFX_ERR;
    }
    resp->hdr.fsStatus = FS_ERR_NONE;
    NFX_SPIN_LOCK(&fs->spin);
    fs->percentMinFree = 100 - tune->minFree;
    NFX_SPIN_UNLOCK(&fs->spin);
    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    fs->super->percentMinFree = fs->percentMinFree;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    fs_writeSuperblocksSync(context, fs->super);
    return NFX_OK;
}

int32
fs_getMinFreeThread(
    fs_Context      *context
)
{
    fs_MsgGetMinFreeResp *resp;
    fs_FileSystem   *fs = context->fileSystem;
	resp = (fs_MsgGetMinFreeResp *)context->respHdr;
    resp->minFree = 100 - fs->percentMinFree;
    resp->hdr.fsStatus = FS_ERR_NONE;
    return NFX_OK;
}

/*
 * Check Global Grow need
 * Returns TRUE when needed
 */

// minimum ms between successive growing of fs.
#define FS_EVM_GROW_DELAY 100000

#ifndef FS_STANDALONE
boolean fs_growIfNeeded(
    fs_Context      *context)
{
    fs_FileSystem   *fs = context->fileSystem;
    uint8           minFree;
    uint64          nBlocks;
    uint64          nAvailBlocks;
    fs_EvmMsgGrowFileSystem_t msg;
    boolean         rc = FALSE;

    /*
     * Look at the super block fields without any locks as
     */

    NFX_SPIN_LOCK(&fs->spin);
    nAvailBlocks = fs_numBlocksAvail(fs);
	nBlocks = context->fileSystem->params.totalBlocks;
    minFree = context->fileSystem->percentMinFree;
    if ((nAvailBlocks * 100 / nBlocks) < minFree)   {
		eeeRTC_t		duration;
		static boolean	first = TRUE;
		static eeeRTC_t	growDelay;

		rc = TRUE;
		eee_rtcDuration(&context->fileSystem->whenGrowSent, &duration);
		if (first) {
			first = FALSE;
			eee_msToRTC((uint32)FS_EVM_GROW_DELAY, &growDelay);
		}
		if (eee_rtcCompare(&duration, &growDelay) > 0) {
			eeeRTC_t now;

            (void)eee_rtcAdd(&context->fileSystem->whenGrowSent, &duration,
							 &now);
			FS_TRACE(context,  FS_TRACE_REQ,
    				("Sending a Grow message as AvailBlocks %lld"
                     " Blocks %lld MinFree %d now %lld prev sent %lld\n",
					 nAvailBlocks, nBlocks, minFree, eee_rtcToSec(&now),
					 eee_rtcToSec(&context->fileSystem->whenGrowSent)));
            msg.msgId = EVM_FSYS_FULL;//FS_REQ_EVM_GROW;
            msg.volId = context->volId;
            msg.blocksAvail = nAvailBlocks;
            msg.fsSize = context->fileSystem->params.totalBlocks;

            // send message to evm to grow
            (void)send_evm_message(&msg, sizeof msg);
    
            eee_rtcIncr(&context->fileSystem->whenGrowSent, &duration);
        }
    }
    NFX_SPIN_UNLOCK(&fs->spin);

    return rc;
}
#else
/*
 * Don't grow till growfs is fixed
 */
boolean fs_growIfNeeded(
    fs_Context      *context)
{
    return FALSE;
}
#endif
