/*
 * Name         : fs-fileset.c
 * RCS          : $Id: fs-fileset.c,v 1.62 2002/12/05 19:46:04 laurentb Exp $
 * Description	: filset implementation data structures
 * Created by	: Rahul Khona
 * Date Created	: 8/13/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

/*
 * Create a fileset. A fileset is successfully created if the
 * root of the fileset does not already exist or if it exists
 * and does not contain any directory entries.
 */

int32
fs_createFileSetThread( fs_Context	*context)
{
	fs_FileSystem	*core = context->fileSystem;
	fs_MsgCreateFileSetReq_t		*req;
	fs_MsgCreateFileSetResp_t		*resp;
	fs_Inode						*fsetInfoInode = NULL;
	fs_Inode						*root = NULL;
	int32							status;
	int								rootMode = FS_IFMT_DIR | FS_IMODE_MASK;
	fs_FileSetIncoreRecord_t		*parent = NULL, *fset = NULL;
	fs_FileSetIncoreRecord_t		*travFset = NULL;
	fs_Inode						*parentDir = NULL, *component = NULL;
	fs_path_t						*path = NULL;
	char							*p = NULL, *c = NULL;
	int								createdRoot = 0;
	int32							rc;

	context->respSize = sizeof(fs_MsgCreateFileSetResp_t);
	resp = (fs_MsgCreateFileSetResp_t *) context->respHdr;
	req = (fs_MsgCreateFileSetReq_t *) context->reqHdr;

	FS_ASSERT(context->logSummary != NULL);

	/*
	 * XXX - Reserve space in the log for this transaction,
	 * so that we don't deadlock with ourselves when trying
	 * to allocate space in the log during this transaction
	 * and only way to free log space is to flush and free
	 * buffers owned by us
 	 *
	 * Maximum Space required by us would be
	 * 8KB for new fileset record
	 * 128 bytes  for inode map file inode update
	 * 256 bytes for quota inodes
	 * 128 bytes for root inode
	 * 1024 bytes for parent directory incase root needs to be created
	 * 128 bytes for the parent directory
	 * 7KB for inode map split
 	 * 7KB for directory split
	 *
	 * Total around 24kb of log space
	 */


try_again:
	/*
	 * Validate that path is an absolute path
	 */

	if (req->path[0] != '/' && req->path[0] != '\\') {
		return FS_ERR_INVAL;
	}

	/*
	 * Now extract the path and lookup the root
	 */

	path = fs_makePath(req->path);

	/*
	 * Start with root inode. Since we don't have a notion of holding
	 * a reference to the inode and inode lock is used to protect
	 * its identity as well as contents, we will always look for
	 * exclusive lock in this path. This is rare enough operation so
	 * it is OK. Ideally we would like to go through the whole path
	 * holding shared locks on a parent dir at time and searching	
	 * for component in it. If at the end we don't find it, we want
	 * to scan the last parent again holding it in exclusive mode and
	 * creating the entry if not found. This is not possible because
	 * we cannot upgrade from shared to exclusive (and rightly so) and
 	 * we cannot drop the lock on the parentdir without rescanning the
	 * whole path.
	 *
	 * We can perhaps re get the parent inode in exclusive mode and
	 * verify its generation count and if they are same, we can
	 * continue with the process, but for now exclusive locks seem
	 * simpler without any significant impact on the system because
	 * of rarity of this operation.
	 */

	parentDir = I_GET_EXCL(0, FS_ROOT_INODE);
	parent = parentDir->fset;
#if 0
	if (parentDir->fset == NULL) {
		parent = parentDir->fset = fs_getFset(context, core, parentDir);
		FS_ASSERT(parentDir->fset != NULL);
	}
#else
	parent = parentDir->fset = fs_getFset(context, core, parentDir);
	FS_ASSERT(parentDir->fset != NULL);
#endif

	p = fs_getNextComponent(path);
	if (p == NULL)	{	//Try to create a fileset on root directory
		fs_inodeReleaseExclusive(context, parentDir);
		return FS_ERR_EXIST;
	}
			
	while (p != NULL) {

		FS_TRACE(context, FS_TRACE_REQ, ("1 path component is %s\n", p));

		/*
		 * Always look in live filesystem
		 */

		component = fs_lookup(context, core, 0, parentDir, p);
		if (component == NULL) {
			FS_ASSERT(fs_lockOwnedExclusive(context,
						     &parentDir->lock));
			c = fs_getNextComponent(path);

			/*
			 * If it is not the end of the path, path is invalid
			 * otherwise we need to create component
			 */

			if (c != NULL) {
				fs_inodeReleaseExclusive(context, parentDir);
				fs_destroyPath(path);
				return FS_ERR_INVAL;
			}

			/*
			 * We need to create p in parent dir
			 */


			FS_ASSERT(parentDir != NULL &&
					 fs_lockOwnedExclusive(context, &parentDir->lock));
			break;
		}

		/*
		 * Proceed with the next component in the path
		 */

		FS_ASSERT(fs_lockOwnedExclusive(context, &component->lock));
		p = fs_getNextComponent(path);
		if (p != NULL) {
#if 0
			/*
			 * If the current parent is not a directory then
		 	 * the path is invalid
			 */

			if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
				fs_inodeReleaseExclusive(context, parentDir);
				fs_destroyPath(path);
				return FS_ERR_INVAL;
			}

			if (component->fset == NULL) {
				component->fset = fs_getFset(context, core, component);
				if (component->fset == NULL) {
					component->fset = parentDir->fset;
				}
			}
#else
			component->fset = fs_getFset(context, core, component);
			FS_ASSERT(component->fset != NULL);
#endif
			FS_ASSERT(component->fset != NULL);
			parentDir = component;
			component = NULL;
			parent = parentDir->fset;
		}
	}

	/*
	 * At this point, either the component does not exist or
	 * we found a component
	 */

	if (component != NULL) {
		FS_ASSERT(p == NULL);
		root = component;
		if (!FS_I_IS_DIR(root->dinode.attrs)) {
			fs_inodeReleaseExclusive(context, root);
			fs_destroyPath(path);
			return FS_ERR_INVAL;
		}
		if (!fs_dir_is_empty(context, root)) {
			fs_inodeReleaseExclusive(context, root);
			fs_destroyPath(path);
			return FS_ERR_EXIST;
		}

		/*
		 * If the component is root of the fileset, then we already
		 * have a fileset created for this root, return success
		 */

		if (component->fset->cachedRecord.rootInode == component->inum) {
			fs_inodeReleaseExclusive(context, component);
			return FS_ERR_NONE;
		}

		/*
		 * Protect against quota on going on the fileset containing
		 * this inode or any of its ancestor filesets. If quota on is
		 * in progress, try after sometime
		 */

		travFset = component->fset;
		if (!fs_trylockShared(context, &travFset->traverseLock)) {
				fs_inodeReleaseExclusive(context, component);
				fs_lockShared(context, &travFset->lock, NULL);
				fs_unlockShared(context, &travFset->lock);
				fs_destroyPath(path);
				travFset = NULL;
				goto try_again;
		}

	} else {

        // Check the fileset name is valid
        status = fs_nameIsValid(p, strlen(p));
        if (status != FS_ERR_NONE) {
            fs_inodeReleaseExclusive(context, parentDir);
            fs_destroyPath(path);
        	if (travFset != NULL) {
        		fs_unlockShared(context, &travFset->traverseLock);
        	}
            return status;
        }
           
    	// See if we can allocate a new fileset
		NFX_SPIN_LOCK(&core->filesetListLock);
        if (core->filesetsCount > FS_FILESET_MAX_ID)  {
    		NFX_SPIN_UNLOCK(&core->filesetListLock);
            fs_inodeReleaseExclusive(context, parentDir);
            fs_destroyPath(path);
        	if (travFset != NULL) {
        		fs_unlockShared(context, &travFset->traverseLock);
        	}
	    	return FS_ERR_FBIG;
		} else {
            //reserve it so that no other can be created if we reach the limit
            core->filesetsCount++;
        }
		NFX_SPIN_UNLOCK(&core->filesetListLock);
        
		/*
		 * We need to create an entry in the parent directory
		 * for p. Before we do that, protect against quota on.
		 */

		travFset = parentDir->fset;
		if (!fs_trylockShared(context, &travFset->traverseLock)) {
    		NFX_SPIN_LOCK(&core->filesetListLock);
            core->filesetsCount--;
    		NFX_SPIN_UNLOCK(&core->filesetListLock);
			fs_inodeReleaseExclusive(context, parentDir);
			fs_lockShared(context, &travFset->lock, NULL);
			fs_unlockShared(context, &travFset->lock);
			fs_destroyPath(path);
			travFset = NULL;
			goto try_again;
		}

		FS_ASSERT(p != NULL);
		FS_ASSERT(fs_lockOwnedExclusive(context, &parentDir->lock));
		component = fs_create(context, core, 0, parentDir, p, rootMode);
        status = fs_addLink(context, 0, component, parentDir->inum);
        if (status != FS_ERR_NONE) {
    		NFX_SPIN_LOCK(&core->filesetListLock);
            core->filesetsCount--;
    		NFX_SPIN_UNLOCK(&core->filesetListLock);
            fs_inodeReleaseExclusive(context, parentDir);
            fs_unlockShared(context, &travFset->lock);
            fs_destroyPath(path);
            return status;
        }
		FS_ASSERT(component != NULL &&
			  fs_lockOwnedExclusive(context, &component->lock));

		root = component;
		root->fset = parentDir->fset;
		createdRoot = 1;

		/*
	 	 * If we dirtied the parentdir in the transaction, don't release it
		 */

        if (!fs_inodeWasModified(context, parentDir)) {
				fs_inodeReleaseExclusive(context, parentDir);
		}
	}

	fs_destroyPath(path);

	FS_ASSERT(parent != NULL);

	/*
	 * Lock down the fsetInfo file in exclusive mode so that
	 * we don't violate any locking hierarchies later on
	 */

	fsetInfoInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
	fs_lockExclusive(context, &parent->lock, NULL);

	FS_ASSERT(root != NULL && parent != NULL &&
		  fs_lockOwnedExclusive(context, &root->lock) &&
		  fs_lockOwnedExclusive(context, &parent->lock));

	fset = fs_getFset(context, core, root);

	if (fset == NULL || fset->cachedRecord.rootInode != root->inum) {

		//We have checked upfront that there was enough space

		if (fsetInfoInode->dinode.size >= FS_FILESET_INFO_MAX_SIZE) {
			FS_ASSERT(fs_fsetInfoHasFreeFsets(context, core, fsetInfoInode)) 
		}
		fset = fs_makeFileSet(context, core, parent, root, fsetInfoInode);
		root->fset = fset;
	} else {

		/*
		 * Fileset already existed
		 */

		FS_ASSERT(createdRoot == 0);
	}
	if (fset == NULL) {

		/*
		 * We can never fail for resources and i/o failures
		 * panic the system
		 */

		FS_ASSERT(0);
	}

	fs_unlockExclusive(context, &parent->lock);

	/*
	 * If we have dirtied the inode in this transaction don't do a put
	 * on it.
	 */

    if (!fs_inodeWasModified(context, fsetInfoInode)) {
			fs_inodeReleaseExclusive(context, fsetInfoInode);
	}

	/*
	 * If we have dirtied the inode in this transaction don't do a put
	 * on it.
	 */

    if (!fs_inodeWasModified(context, root)) {
			fs_inodeReleaseExclusive(context, root);
	}

	/*
	 * Transaction commit would unlock the inodes
	 */

    rc = fs_logCommitEntry(context);

	/*
	 * If we blocked out quota on, release it
	 */

	if (travFset != NULL) {
		fs_unlockShared(context, &travFset->traverseLock);
	}

    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
        return FS_ERR_IO;
    }

	return FS_ERR_NONE;
}

/*
 * Remove a fileset.
 */

int32
fs_removeFileSetThread( fs_Context	*context)
{
	fs_FileSystem	*core = context->fileSystem;
	fs_MsgRemoveFileSetReq_t		*req;
	fs_MsgRemoveFileSetResp_t		*resp;
	fs_Inode			*fsetInfoInode;
	fs_Inode			*root = NULL, *parentDir = NULL;
	fs_Inode			*component = NULL;
	fs_FileSetIncoreRecord_t	*fset;
	fs_path_t			*path;
	char				*p, *t;
	int32				rc;
    fs_Inumber          inum;

	context->respSize = sizeof(fs_MsgRemoveFileSetResp_t);
	resp = (fs_MsgRemoveFileSetResp_t *) context->respHdr;
	req = (fs_MsgRemoveFileSetReq_t *) context->reqHdr;

	FS_ASSERT(context->logSummary != NULL);

	/*
	 * XXX - Reserve space in the log for this transaction,
	 * so that we don't deadlock with ourselves when trying
	 * to allocate space in the log during this transaction
	 * and only way to free log space is to flush and free
	 * buffers owned by us
 	 *
	 * Maximum Space required by us would be
	 * 8KB for new fileset record
	 * 128 bytes  for inode map file inode update
	 * 256 bytes for quota inodes
	 * 128 bytes for root inode
	 * 1024 bytes for parent directory incase root needs to be created
	 * 128 bytes for the parent directory
	 * 7KB for inode map split
 	 * 7KB for directory split
	 *
	 * Total around 24kb of log space
	 */

	/*
	 * Now extract the path and lookup the root
	 */

	/*
	 * Validate that path is an absolute path
	 */

	if (req->path[0] != '/' && req->path[0] != '\\') {
		return FS_ERR_INVAL;
	}
    
previousSlash:
    t = strrchr(req->path, '/');
    if (t == NULL)  {
        t = strrchr(req->path, '\\');
    }
    ASSERT(t != NULL);

    if (t == &req->path[0]) {
		if (t[1] == '\0') {	//removing root directory fileset not allowed
    		return FS_ERR_INVAL;
		}
		*t++ = '\0';
		parentDir = I_GET_EXCL(0, FS_ROOT_INODE);
		if (parentDir->fset == NULL) {
			parentDir->fset = fs_getFset(context, core, parentDir);
		}
		goto find_child;
    } else {
        if (t[1] == '\0') { //path has a trailing /, remove it
            *t = '\0';
            goto previousSlash;        
        }
    }
    *t++ = '\0';
	path = fs_makePath(req->path);

	/*
	 * Start with root inode
	 */

	parentDir = I_GET_EXCL(0, FS_ROOT_INODE);
	if (parentDir->fset == NULL) {
		parentDir->fset = fs_getFset(context, core, parentDir);
	}

	p = fs_getNextComponent(path);
	while (p != NULL) {

		/*
		 * Always look in live filesystem
		 */

		component = fs_lookup(context, core, 0, parentDir, p);
		if (component == NULL) {
			FS_ASSERT(fs_lockOwnedExclusive(context, &parentDir->lock));
			break;
		}
		
		FS_ASSERT(fs_lockOwnedExclusive(context, &component->lock));
		p = fs_getNextComponent(path);
		if (p != NULL) {
			parentDir = component;
			component = NULL;
			if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
				fs_inodeReleaseExclusive(context, parentDir);
				parentDir = NULL;
				break;
			}
		}
	}
	fs_destroyPath(path);

	if (component == NULL) {
		fs_inodeReleaseExclusive(context, parentDir);
		return FS_ERR_NOENT;
	}

	if (!FS_I_IS_DIR(component->dinode.attrs)) {
		fs_inodeReleaseExclusive(context, component);
		return FS_ERR_NOENT;
	}
	
    parentDir = component;

find_child:
	
    inum = fs_dir_lookupent(context, parentDir, t, strlen(t));
    if (inum == 0) {
		fs_inodeReleaseExclusive(context, parentDir);
        *--t = '/';
		return FS_ERR_NOENT;
    }

    component = I_GET_EXCL(0, inum);
	if (!FS_I_IS_DIR(component->dinode.attrs)) {
		fs_inodeReleaseExclusive(context, component);
		fs_inodeReleaseExclusive(context, parentDir);
        *--t = '/';
		return FS_ERR_NOENT;
	}

	root = component;
    fset = root->fset = fs_getFset(context, core, root);

	/*
	 * The root of fset is not the same as the root directory that
	 * user has specified return a failure
	 */

	if (root->inum != fset->cachedRecord.rootInode) {
		fs_inodeReleaseExclusive(context, root);
		fs_inodeReleaseExclusive(context, parentDir);
        *--t = '/';
		return FS_ERR_NOENT;
	}
	
	/*
	 * If the directory is not empty then don't allow fileset to be removed.
	 */
	if (!fs_dir_is_empty(context, root)) {
		fs_inodeReleaseExclusive(context, root);
		fs_inodeReleaseExclusive(context, parentDir);
        *--t = '/';
		return FS_ERR_NOTEMPTY;
	}

    // Check that quota for this fileset is not being turned on 
    NFX_SPIN_LOCK(&fset->spinlock);
    if (fset->incoreFlag & FS_FILESET_QUOTA_ON_THREAD)  {
        NFX_SPIN_UNLOCK(&fset->spinlock);
        fs_inodeReleaseExclusive(context, root);
        fs_inodeReleaseExclusive(context, parentDir);
        *--t = '/';
        return FS_ERR_BUSY;
    }
    NFX_SPIN_UNLOCK(&fset->spinlock);
    
	/*
	 * Block out quota on from looking at this fileset while we are
	 * removing it. If quota on is looking at this fileset, then
	 * wait for it then return a retry error
	 */

	if (!fs_trylockExclusive(context, &fset->traverseLock)) {
		fs_inodeReleaseExclusive(context, root);
		fs_inodeReleaseExclusive(context, parentDir);
        *--t = '/';
		return FS_ERR_BUSY;
	}

	/*
	 * Lock down the fsetInfo file in exclusive mode so that
	 * we don't violate any locking hierarchies later on
	 */

	fsetInfoInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);

	fs_removeFileSet(context, core, fset, root, fsetInfoInode);
    if (!fs_inodeWasModified(context, fsetInfoInode)) {
        fs_inodeReleaseExclusive(context, fsetInfoInode);
    }

    /*
     * Remove the entry from its parent as well
     */

    fs_removeDirNode(context, parentDir, root, t, strlen(t), NULL, NULL, NULL);
    *--t = '/';

	/*
	 * If we have not dirtied the inode in this transaction release it.
	 */
    if (!fs_inodeWasModified(context, parentDir)) {
			fs_inodeReleaseExclusive(context, parentDir);
	}

	/*
	 * Transaction commit would unlock the inodes
	 */
    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
        return FS_ERR_IO;
    }

    //Decrement the number of filesets
	NFX_SPIN_LOCK(&core->filesetListLock);
    FS_ASSERT(core->filesetsCount != 1)
    core->filesetsCount--;
    NFX_SPIN_UNLOCK(&core->filesetListLock);

	return FS_ERR_NONE;
}

/*
 * List the filesets for a given filesystem
 */

int32
fs_listFsetThread( fs_Context *context)
{
	uint32	requests;
	uint64	start;
	fs_MsgListFileSetsReq_t	    *req;
	fs_MsgListFileSetsResp_t	*resp;

	resp = (fs_MsgListFileSetsResp_t *) context->respHdr;
	req = (fs_MsgListFileSetsReq_t *) context->reqHdr;

	requests = req->numberEntries;
	start = req->startIndex;

    FS_ASSERT(context->respSize >= sizeof (fs_MsgListFileSetsResp_t) +
                                    (requests * sizeof (fs_FileSetList_t)));

#if 0
    WE CAN NO LONGER MALLOC RESPONSE. SET TO MAXIMUM IN REQUEST.TAB, THEN
    JUST ADJUST THE SIZE HERE
	context->respSize = sizeof (fs_MsgListFileSetsResp_t) +
				(requests * sizeof (fs_FileSetList_t));
	/*
	 * Create our response space as we would not have allocated
	 * space for this request at context creation time because
	 * of variable nature of the request
	 */

	context->respHdr = fs_ramAlloc(FS_RAM_RESP_MSG, context->respSize);
	resp = (fs_MsgListFileSetsResp_t *)context->respHdr;
	bzero(context->respHdr, context->respSize);
	context->respHdr->requestType = req->msgHdr.requestType;
	context->respHdr->clientPrivate = req->msgHdr.clientPrivate;
	context->respHdr->clientId = req->msgHdr.clientId;
	context->initRespSize = context->respSize;
#endif
    
	return fs_listFileSet(context, context->fileSystem, req, resp);
}

/*
 * Free the fileset. Give back its memory
 */

EXPORTED void
fs_holdFileSet(
	fs_Context			*context,	// context
	fs_FileSystem			*core,		// core
	fs_FileSetIncoreRecord_t	*fset	// fileset to be freed
)
{
	NFX_SPIN_LOCK(&fset->spinlock);
	fset->count++;
	NFX_SPIN_UNLOCK(&fset->spinlock);
}

EXPORTED void
fs_releFileSet(
	fs_Context			*context,	// context
	fs_FileSystem			*core,		// core
	fs_FileSetIncoreRecord_t	*fset	// fileset to be freed
)
{
	NFX_SPIN_LOCK(&fset->spinlock);
	fset->count++;
	NFX_SPIN_UNLOCK(&fset->spinlock);
}

EXPORTED int
fs_isFileSetHeld(
	fs_Context			*context,	// context
	fs_FileSystem			*core,		// core
	fs_FileSetIncoreRecord_t	*fset	// fileset to be freed
)
{
	int	rc;
	NFX_SPIN_LOCK(&fset->spinlock);
	rc = fset->count;
	NFX_SPIN_UNLOCK(&fset->spinlock);
	return rc;
}
	
PRIVATE void
fs_freeFileSet(
	fs_Context			        *context,	// context
	fs_FileSystem			    *core,		// core
	fs_FileSetIncoreRecord_t	*fset   	// fileset to be freed
)
{
	FS_ASSERT(fs_lockOwnedExclusive(context, &fset->lock));
	FS_ASSERT(fs_lockOwnedExclusive(context, &fset->traverseLock));
	NFX_SPIN_LOCK(&fset->spinlock);
    DLL_REMOVE(&core->filesetList, fset, fsetList);
	core->filesetsCached--;
	NFX_SPIN_UNLOCK(&fset->spinlock);
    fs_unlockExclusive(context, &fset->lock);
    fs_unlockExclusive(context, &fset->traverseLock);
	fs_ramDealloc(FS_RAM_FILESET, fset);
}

/*
 * Initialize the fileset structure
 */

PRIVATE	void
fs_initializeFileSet(
	fs_FileSetIncoreRecord_t	*fset,
	uint64				root,
	fs_FileSetIncoreRecord_t	*parent)
{
	fs_FileSetDiskRecord_t	*incore = &fset->cachedRecord;

	/*
	 * XXX - Assert that the appropriate lock is held
	 */


	fset->fsetList.prev = fset->fsetList.next = NULL;
	fset->hashList.prev = fset->hashList.next = NULL;
	fset->fileSetId = FS_FILESET_INVAL_ID;
	fset->incoreFlag = FS_FILESET_NEEDS_INIT;
	fset->count = 0;
	fset->parent = parent;
	bzero(incore, sizeof (fs_FileSetDiskRecord_t));
	incore->rootInode = root;
}

/*
 * Put the fileset structure on the hash queues
 */

PRIVATE void
fs_hashInFileSet(
	fs_FileSystem			*core,
	fs_FileSetIncoreRecord_t	*fset,
	uint32				id,
	int				hash
)
{
	FS_ASSERT(hash == FS_FILESET_HASH(id));

	/*
	 * XXX - Assert that the appropriate list lock is held
	 */
	DLL_INSERT_HEAD(&core->fileSetHash[hash], fset, hashList);
}

PRIVATE void
fs_unHashFileSet(
	fs_FileSystem			*core,
	fs_FileSetIncoreRecord_t	*fset,
	uint32				id,
	int				hash
)
{
	FS_ASSERT(hash == FS_FILESET_HASH(id));

	/*
	 * XXX - Assert that the appropriate list lock is held
	 */
	DLL_REMOVE(&core->fileSetHash[hash], fset, hashList);
}

/*
 * Allocate an incore fileset structure if not already available in core
 */
PRIVATE fs_FileSetIncoreRecord_t *
fs_allocateIncoreFileSetRecord(
	fs_Context			*context,
	fs_FileSystem			*core,
	fs_FileSetIncoreRecord_t	*parent,
	uint32				fsetId,	// FS_FILESET_INVAL_ID overloads to
								// mean don't hash in or put on list
	int					lookupOnly)
{
	int	steal = 0;
	int	hash = FS_FILESET_HASH(fsetId);
	fs_FileSetIncoreRecord_t *fset, *nfset = NULL;
	nfx_spinlock_t		*listlock;
	listlock = &core->filesetListLock;

	/*
	 * XXX - Assert that the list lock is not held
	 */

retry:

	NFX_SPIN_LOCK(listlock);
	DLL_ITERATE(fset, &core->fileSetHash[hash], hashList) {
		if (fset == NULL || fset == DLL_NIL ||
			fset->fileSetId == fsetId) {
			break;
		}
	}

	/*
	 * fset structure was found. Return it
	 */

	if (fset != NULL && fset != DLL_NIL) {

		/*
		 * If we allocated memory for nfset, put it on the
	 	 * freelist.
		 */
		if (nfset != NULL) {
			DLL_INSERT_HEAD(&core->filesetList,
					   nfset, fsetList);
		}
		NFX_SPIN_UNLOCK(listlock);
		return fset;
	}

	if (lookupOnly) {
		NFX_SPIN_UNLOCK(listlock);
		return NULL;
	}

	if (nfset == NULL &&
	    (core->filesetsCached >= core->filesetCacheLimit || steal)) {
		DLL_REMOVE_HEAD(&core->filesetList, fset, fsetList);

		FS_ASSERT(FALSE);	//Never happens in the current implementation
#if 0		
		if (fset == NULL || fset == DLL_NIL) {
			FS_ASSERT(steal > 0);
			NFX_SPIN_UNLOCK(listlock);

			if (steal > 10) {

				/*
				 * We retried many a times.
				 * must be a bug panic
				 */

				fs_panic("%s:%s:%d",
					 "Cannot allocate fset structures",
					 __FILE__, __LINE__);
			}

			/*
			 * There isn't memory available and there aren't
			 * any free structures available. Sleep for a second
			 */

			FS_ASSERT(FALSE); //sleep(1);
			steal++;
			goto retry;
		}

		/*
		 * We can assert this because we have just removed the fileset
		 * from the freelist and and not yet put it on the hash list
		 * and are holding the list lock
		 */

		FS_ASSERT(fset->count == 1);
		fs_initializeFileSet(fset, fsetId, parent);

		/*
		 * Requested id is overloaded. If we are called with
		 * FS_FILESET_INVAL_ID, then caller will intiialize it
		 * and hash it in, when appropriate
		 */

		if (fsetId != FS_FILESET_INVAL_ID) {
			DLL_INSERT_TAIL(&core->filesetList, fset, fsetList);

			/*
		 	 * We did not drop the list lock at any time, so no new
		 	 * entry would be created. We can safely assume that
		 	 * the fileset structure of identity we need is not created
		 	 */

			fs_hashInFileSet(core, nfset, fsetId, hash);
		}
#endif
		NFX_SPIN_UNLOCK(listlock);
		return fset;
	}


	/*
	 * We need to allocate more memory. Memory alloc never sleeps
	 * nor can it fail. It panics if we cannot get the memory
	 * so be very careful in allocation. Don't allocate any extra
	 */

	if (core->filesetsCached < core->filesetCacheLimit) {
		nfset = fs_ramAlloc(FS_RAM_FILESET,
				    sizeof (fs_FileSetIncoreRecord_t));
		bzero(nfset, sizeof (fs_FileSetIncoreRecord_t));
		/*
		 * Initialize the spinlock as soon as we allocated memory, so
		 * that we don't have to bother about reused or new structure
		 * in init routines
		 */

		NFX_SPIN_INIT(&nfset->spinlock,SPINLOCK_LEVEL_UNKNOWN);
        fs_lockInit(&nfset->lock, FS_LOCK_TYPE_FILESET);
        fs_lockInit(&nfset->traverseLock, FS_LOCK_TYPE_FILESET_TRAVERSAL);

		/*
		 * Increment the count of cached filesets
		 */

		core->filesetsCached++;

		/*
		 * We had previously allocated a structure and a new one
		 * with identify of our desire is not created, assign
		 * our identity to it, hash it in and return it
		 */

		fset = nfset;
		fs_initializeFileSet(fset, fsetId, parent);

		/*
		 * Requested id is overloaded. If we are called with
		 * FS_FILESET_INVAL_ID, then caller will initialize it
		 * and hash it in, when appropriate
		 */

		if (fsetId != FS_FILESET_INVAL_ID) {
			DLL_INSERT_TAIL(&core->filesetList, fset, fsetList);

			/*
		 	 * We did not drop the list lock at any time, so no new
		 	 * entry would be created. We can safely assume that
		 	 * the fileset structure of identity we need is not created
		 	 */

			fs_hashInFileSet(core, nfset, fsetId, hash);
		}

		NFX_SPIN_UNLOCK(listlock);
		return fset;
#if 0		
	} else {

		NFX_SPIN_UNLOCK(listlock);

		/*
		 * We are out of limit, try stealing some from the freelist
		 */

		steal = 1;
		FS_ASSERT(FALSE); //sleep(1);
#endif
	}

	goto retry;

	/* NOTREACHED */
}

EXPORTED int
fs_fsetInfoHasFreeFsets(
	fs_Context		*context,
	fs_FileSystem	*core,
	fs_Inode		*fsetInfoInode)
{

	uint64					offset = 0;
	int						found = 0;
	fs_FileSetDiskRecordPadded	*rec;
	fs_Buf					*buf;
	int						i;

	while (offset < fsetInfoInode->dinode.size && !found) {
		
		buf = O_GET_EXCL(fsetInfoInode, offset >> FS_BLOCK_SHIFT);

		FS_ASSERT(buf != NULL);

		rec = buf->u.fset->recs;

		for (i = 0; i < FS_FILESET_RECS_PER_PAGE; i++) {
			if (rec->drec.rootInode == 0) {
				found++;
				break;
			}
			rec++;
		}
		offset += FS_BLOCK_SIZE;
		fs_bufReleaseExclusive(context, buf);
	}
	return found;
}

/*
 * Do a transaction to write the new fileset information. It can
 * perform allocation to the filesetInfo file
 */

PRIVATE int
fs_initializeFilesetRecord(
	fs_Context					*context,
	fs_FileSystem				*core,
	fs_FileSetIncoreRecord_t	*parent,
	fs_Inode					*root,
	fs_Inode					*fsetInfoInode,
	fs_FileSetIncoreRecord_t	*fset)
{
	uint64			offset = 0;
	fs_Buf			*buf = NULL;
	int			found = 0;
	fs_FileSetDiskRecordPadded	*rec = NULL;
	fs_FileSetDiskRecord_t	*parentCache = NULL;
	int			i;
	int			id = 0;

	/*
	 * Assert that the fsetInfoInode is held exclusively, root
 	 * is held exclusively and uInode and gInode are held exclusively
	 * Assert that parent is exclusively held
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock));
	FS_ASSERT(fs_lockOwnedExclusive(context, &root->lock));
	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &root->lock) &&
		  ((parent == NULL && root->inum == FS_ROOT_INODE) ||
		   fs_lockOwnedExclusive(context, &parent->lock)));
	FS_ASSERT(fs_lockOwnedExclusive(context, &fset->lock));
	FS_ASSERT(fset->fileSetId == FS_FILESET_INVAL_ID);

	if (parent != NULL) {
		parentCache = &parent->cachedRecord;
	}

	/*
	 * FileSetInfo files are typically not very big and make fileset
	 * operations even more rare. Our disk layout does not optimize
	 * quick searching for available fileset record.
	 *
	 * Sequentially scan through the fileset info file and see if
	 * any slot is free. A record is deemed free, if the root inode
	 * is 0.
	 */

	while (offset < fsetInfoInode->dinode.size && !found) {
		
		buf = O_GET_EXCL(fsetInfoInode, offset >> FS_BLOCK_SHIFT);

		FS_ASSERT(buf != NULL);

		rec = buf->u.fset->recs;

		for (i = 0; i < FS_FILESET_RECS_PER_PAGE; i++) {
			if (rec->drec.rootInode == 0) {
				found++;
				break;
			}
			rec++;
			id++;
		}
		offset += FS_BLOCK_SIZE;
		if (!found) {
			fs_bufReleaseExclusive(context, buf);
		}
	}

	/*
	 * We found a record. Initialize it appropriately and do a transaction
	 * to write the buffer
	 */
	if (found) {
		rec->drec.parentId = parent->fileSetId;
		rec->drec.rootInode = root->inum;
		rec->drec.fileSetFlags = 0;
		bzero(&rec->drec.fileSetConsumption,
		      sizeof (fs_ConsumptionRecord_t));
        rec->drec.lastChanged = fs_getTimeOfDaySecs();

		/*
		 * Copy defaults from parent
		 */

		if (parentCache != NULL) {
        	rec->drec.fileSetFlags |= FS_FILESET_USE_QDEFAULTS;
			bcopy(&rec->drec.defaultAggregateQuota,
				  &parentCache->defaultAggregateQuota,
				  sizeof (fs_QuotaRecord_t));
			bcopy(&rec->drec.fileSetQuota, &parentCache->defaultAggregateQuota,
				  sizeof (fs_QuotaRecord_t));
		} else {

			/*
			 * We are the root inode. Set defaults as unlimited
			 */
			bzero(&rec->drec.defaultAggregateQuota, sizeof (fs_QuotaRecord_t));
            bzero(&rec->drec.fileSetQuota, sizeof (fs_QuotaRecord_t));
		}

		/*
		 * Copy the new records in cache as well
		 */

		bcopy(&rec->drec, &fset->cachedRecord,
              sizeof (fs_FileSetDiskRecord_t));

		/*
		 * Update the fset with the new id. assert we are holding the
		 * fset in exclusive mode
		 */

		FS_ASSERT(fs_lockOwnedExclusive(context, &fset->lock));

		fset->fileSetId = id;

		/*
	 	 * Re hash in the fset and enter it into the list as well
	 	 */

		NFX_SPIN_LOCK(&core->filesetListLock);

		fs_hashInFileSet(core, fset, fset->fileSetId,
					     FS_FILESET_HASH(fset->fileSetId));

		DLL_INSERT_TAIL(&core->filesetList, fset, fsetList);
		NFX_SPIN_UNLOCK(&core->filesetListLock);

		FS_BUF_WRITE(buf, 
             (id * sizeof (fs_FileSetDiskRecordPadded)) & (FS_BLOCK_SIZE-1),
             sizeof (fs_FileSetDiskRecordPadded));

		/*
		 * Transaction commit would release the buffer lock
		 */

		return 1;
	}

	if (fsetInfoInode->dinode.size == FS_FILESET_INFO_MAX_SIZE) {
			return 0;
	}

	/*
	 * We need to allocate
	 */

	FS_ASSERT(offset == fsetInfoInode->dinode.size);
	
	buf = O_CREATE(fsetInfoInode, offset >> FS_BLOCK_SHIFT);

	/*
	 * If we could not do the allocation return failure
	 */

	if (buf == NULL) {
		fs_error(context, FS_INTERNAL_ERR,
			 "Could not allocate blocks to fileset info file");
		return 0;
	}

	/*
	 * Initialize the buffer.data
	 */

	bzero(buf->u.data, FS_BLOCK_SIZE);
    FS_BUF_ZERO_ALL(buf);

	rec = buf->u.fset->recs;
	if (parent != NULL) {
		rec->drec.parentId = parent->fileSetId;
	} else {
		rec->drec.parentId = FS_FILESET_INVAL_ID;
	}
	rec->drec.rootInode = root->inum;
	rec->drec.fileSetFlags = 0;
	bzero(&rec->drec.fileSetConsumption,
	      sizeof (fs_ConsumptionRecord_t));
    rec->drec.lastChanged = fs_getTimeOfDaySecs();

	/*
	 * Copy defaults from parent
	 */
	if (parentCache != NULL) {
    	rec->drec.fileSetFlags |= FS_FILESET_USE_QDEFAULTS;
		bcopy(&rec->drec.defaultAggregateQuota,
		  	  &parentCache->defaultAggregateQuota,
		  	  sizeof (fs_QuotaRecord_t));
		bcopy(&rec->drec.fileSetQuota, &parentCache->defaultAggregateQuota,
			  sizeof (fs_QuotaRecord_t));
	} else {

		/*
	 	 * We are the root inode. Set defaults as unlimited
	 	 */

		bzero(&rec->drec.defaultAggregateQuota, sizeof (fs_QuotaRecord_t));
        bzero(&rec->drec.fileSetQuota, sizeof (fs_QuotaRecord_t));

	}

	/*
	 * Copy the new records in cache as well
	 */

	bcopy(&rec->drec, &fset->cachedRecord, sizeof (fs_FileSetDiskRecord_t));

	/*
	 * Assert that we hold fset exclusively
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fset->lock));

	/*
	 * Assign the new id to incore fset structure.
	 */

	fset->fileSetId = offset / sizeof (fs_FileSetDiskRecordPadded);

	/*
	 * Re hash in the fset and enter it into the list as well
	 */
	NFX_SPIN_LOCK(&core->filesetListLock);
	fs_hashInFileSet(core, fset, fset->fileSetId,
					 FS_FILESET_HASH(fset->fileSetId));
	DLL_INSERT_TAIL(&core->filesetList, fset, fsetList);
	NFX_SPIN_UNLOCK(&core->filesetListLock);

	FS_BUF_WRITE(buf, 0, FS_BLOCK_SIZE);

	/*
	 * Also update the fsetInfo file size now
	 */
	fsetInfoInode->dinode.size += FS_BLOCK_SIZE;
    fs_inodeWrite(context, fsetInfoInode, INO_WRITE_LOG);

	return 1;
}

/*
 * Do a transaction to free the fileset information.
 */

PRIVATE void
fs_deinitializeFileSetRecord(
	fs_Context			*context,
	fs_FileSystem			*core,
	fs_Inode			*fsetInfoInode,
	fs_FileSetIncoreRecord_t	*fset
)
{
	
	uint64			id = fset->fileSetId;
	uint64			offset = id * sizeof (fs_FileSetDiskRecordPadded);
	fs_Buf			*buf;
	fs_FileSetDiskRecord_t	*rec = NULL;
	int32			boff;

	/*
	 * Assert that the fsetInfoInode is held exclusively, root
 	 * is held exclusively.
	 * Assert that parent is exclusively held
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fset->lock));

	/*
	 * Free the corresponding record by setting the root inode to 0.
	 */

	boff = (offset & (FS_BLOCK_SIZE - 1));
	buf = O_GET_EXCL(fsetInfoInode, offset >> FS_BLOCK_SHIFT);

	FS_ASSERT(buf != NULL);

	rec = &buf->u.fset->recs[boff >> FS_FILESET_REC_SHIFT].drec;

	rec->rootInode = 0;
	rec->fileSetFlags = 0;

	FS_BUF_WRITE(buf, boff, sizeof (fs_FileSetDiskRecord_t));

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fset->lock));

	return;
}

EXPORTED fs_FileSetIncoreRecord_t *
fs_getFset(
	fs_Context	*context,
	fs_FileSystem	*core,
	fs_Inode	*inode	// Get fileset belonging to this inode
)
{
	fs_FileSetIncoreRecord_t	*fset;

    FS_ASSERT(context != NULL && core != NULL && inode != NULL);

	fset =  fs_allocateIncoreFileSetRecord(context, core, NULL,
					      inode->dinode.fsetNum, 1);
	return fset;
}

/*
 * Readin all the fsets and create the incore list. Should be
 * called in at mount time only
 */

EXPORTED void
fs_buildFileSetList(
	fs_Context	*context,
	fs_FileSystem	*core)
{
	fs_Inode			*fsetInode;
	fs_Buf				*buf;
	uint64				offset = 0;
	int				id = 0;
    int             fsetCount = 0;
	fs_FileSetIncoreRecord_t	*fset, *tfset;
	int				i;
	fs_FileSetDiskRecordPadded		*rec;

	fsetInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
	while (offset < fsetInode->dinode.size) {
		buf = O_GET_SHARED(fsetInode, offset >> FS_BLOCK_SHIFT);
		FS_ASSERT(buf != NULL);
		rec = buf->u.fset->recs;
		for (i = 0; i < FS_FILESET_RECS_PER_PAGE; i++) {
			if (rec->drec.rootInode == 0) {
                id++;
				rec++;
				continue;
			}
			fset = fs_allocateIncoreFileSetRecord(context, core, NULL, id, 0);
			fs_lockShared(context, &fset->lock, NULL);
			fset->fileSetId = id;
			fset->incoreFlag = 0;
			bcopy(rec, &fset->cachedRecord, sizeof (fs_FileSetDiskRecord_t));
			//If quota on flag is set on disk, then set it in the incore flags
			if (fset->cachedRecord.fileSetFlags & FS_FILESET_QUOTA_ON)
				fset->incoreFlag |= FS_FILESET_QUOTA_ON;
			fs_unlockShared(context, &fset->lock);
            fsetCount++;
			id++;
			rec++;
		}
		fs_bufRelease(context, buf);
		offset += FS_BLOCK_SIZE;
	}

	FS_ASSERT(id != 0);
	FS_ASSERT(core->filesetsCached == fsetCount);

	/*
	 * At this point all the filesets are hashed in and are incore
	 * only thing that is missing is the linkage between children
	 * and parent. Walk the list and assign appropriate ids. 
	 */

	
	NFX_SPIN_LOCK(&core->filesetListLock);
    //Initialize the number of filesets in the file system
    core->filesetsCount = fsetCount;
	DLL_ITERATE(fset, &core->filesetList, fsetList) {

		if (fset == NULL || fset == DLL_NIL) {
			break;
		}

		/*
		 * Root of the filesystem always has 0th fileset
	 	 */
		if (fset->fileSetId == 0) {
			fset->parent = NULL;
			continue;
		}

		DLL_ITERATE(tfset, &core->filesetList, fsetList) {
			if (tfset == NULL || tfset == DLL_NIL) {
				break;
			}
			if (tfset->fileSetId == fset->cachedRecord.parentId) {
				fset->parent = tfset;
                //If fileset uses quota defaults, copy them from its parent
                if (fset->cachedRecord.fileSetFlags & FS_FILESET_USE_QDEFAULTS)
                {
                    bcopy(&tfset->cachedRecord.defaultAggregateQuota, 
                          &fset->cachedRecord.fileSetQuota,
                          sizeof(fs_QuotaRecord_t));
                }
				break;
			}
		}
		
	}
	NFX_SPIN_UNLOCK(&core->filesetListLock);
	fs_inodeReleaseExclusive(context, fsetInode);
}

/*
 * Clean filesets and quota incore structures
 * Has to be called with fs access lock held exclusive
 */
EXPORTED void fs_cleanFileSetNQuotaLists(
	fs_Context      *context,
	fs_FileSystem   *core)
{
    fs_QuotaIncoreRecord_t      *quota;
    fs_FileSetIncoreRecord_t    *fileset;
    int                         i;
        
    NFX_SPIN_LOCK(&fs_quotaCache.spin);
    for (i = 0; i < FS_QCACHE_HASH_MAX; i++) {
    	DLL_ITERATE(quota, &fs_quotaCache.userQuotaHashList[i], hashList)
        {
			if (quota == NULL || quota == DLL_NIL) {
				break;
			}
            if (quota->volId.whole_thing == core->volId.whole_thing)    {
		        DLL_REMOVE(&fs_quotaCache.userQuotaHashList[i], quota, 
                           hashList);
	            if (quota->count == 0) {
		            DLL_REMOVE(&fs_quotaCache.quotaFreeList, quota, freeList);
                }
                fs_ramDealloc(FS_RAM_QUOTA, quota);
                fs_quotaCache.quotasCached--;
			}
		}

    	DLL_ITERATE(quota, &fs_quotaCache.groupQuotaHashList[i], hashList)
        {
			if (quota == NULL || quota == DLL_NIL) {
				break;
			}
            if (quota->volId.whole_thing == core->volId.whole_thing)    {
		        DLL_REMOVE(&fs_quotaCache.groupQuotaHashList[i], quota, 
                           hashList);
	            if (quota->count == 0) {
		            DLL_REMOVE(&fs_quotaCache.quotaFreeList, quota, freeList);
                }
                fs_ramDealloc(FS_RAM_QUOTA, quota);
                fs_quotaCache.quotasCached--;
			}
		}
    }
	NFX_SPIN_UNLOCK(&fs_quotaCache.spin);
    
	NFX_SPIN_LOCK(&core->filesetListLock);
    DLL_REMOVE_HEAD(&core->filesetList, fileset, fsetList);
    while (fileset != NULL) {
        fs_ramDealloc(FS_RAM_FILESET, fileset);
        DLL_REMOVE_HEAD(&core->filesetList, fileset, fsetList);
        core->filesetsCached--;
        core->filesetsCount--;
    }
    FS_ASSERT(core->filesetsCached == 0);
    FS_ASSERT(core->filesetsCount == 0);
    
    //Now reinitialize the fileset hash lists
    for (i = 0; i < FS_FILESET_CACHE_HASH_MAX; i++) {
        DLL_INIT(&(core->fileSetHash[i]));
    }
	NFX_SPIN_UNLOCK(&core->filesetListLock);
}

/*
 * Create a new fileset. On success it returns a fileset structure
 * appropriately inserted in to fileset lists and all the necessary
 * allocations done. The transaction is logged on return. 
 *
 * Caller is expected to pass in exclusively held fsetInfoInode
 * and fsetRoot
 *
 * Locking hierarchy also implies that imap file is likely to be
 * below any other inode as we allocate quota inode holding root
 * inode of the fileset as well as filesetinfo file.
 */

EXPORTED fs_FileSetIncoreRecord_t *
fs_makeFileSet(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*parent,	// parent of the fileset
	fs_Inode					*fsetRoot,	// Fileset for this inode
	fs_Inode					*fsetInfoInode	// fileset info file inode
)
{
#if 0
TOBERM
	fs_Inode					*uInode, *gInode;
#endif
	fs_FileSetIncoreRecord_t	*fset;
	int							rc;

	/*
	 * Assert that the arugments are valid
	 */

    FS_ASSERT(context != NULL);
    FS_ASSERT(core != NULL);
    FS_ASSERT(fsetRoot != NULL);
    FS_ASSERT(FS_I_IS_DIR(fsetRoot->dinode.attrs));
    FS_ASSERT(parent != NULL || fsetRoot->inum == FS_ROOT_INODE);

	/*
	 * Assert that fsetInfoInode is held exclusively. This
	 * protects new fsets from being created and fsets from
	 * being removed. Also assert that fsetRoot is exclusively
	 * held at this point. Also assert that parent is exclusively
	 * held
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fsetRoot->lock) &&
		  (parent == NULL || fs_lockOwnedExclusive(context, &parent->lock)));

	/*
	 * Now search for the fileset in the incore structures. If
	 * one exists, we return immediately, otherwise create it
	 * and link it in the appropriate chains.  We don't allow creating
	 * a fset in the middle of an existing hierarchy so parent chains
	 * are not likely to be disturbed. 
	 *
	 * Fset will be associated with the inode, so while
	 * inode is incore, the fset cannot be removed.
	 */

	fset = fs_allocateIncoreFileSetRecord(context, core,
					      parent, fsetRoot->dinode.fsetNum, 1);

	if (fset == NULL || fset->cachedRecord.rootInode != fsetRoot->inum) {

		/*
		 * We could not find the fileset incore. See if we have
		 * enough room to allocate space in fileset file
		 */

		if (fsetInfoInode->dinode.size >= FS_FILESET_INFO_MAX_SIZE) {

			/*
			 * Find spare room in the file. If we don't have return
			 * failure
			 */

			if (!fs_fsetInfoHasFreeFsets(context, core, fsetInfoInode)) {
				return NULL;
			}
		}
		fset = fs_allocateIncoreFileSetRecord(context, core,
								  			  parent, FS_FILESET_INVAL_ID, 0);
		
	}

	NFX_SPIN_LOCK(&fset->spinlock);
	if (!(fset->incoreFlag & FS_FILESET_NEEDS_INIT)) {
		FS_ASSERT(fset->fileSetId < FS_FILESET_INVAL_ID);
		NFX_SPIN_UNLOCK(&fset->spinlock);
		return fset;
	}
	NFX_SPIN_UNLOCK(&fset->spinlock);

	/*
	 * This is a newly allocated fileset, assert that its id is not
	 * initialized yet
	 */

	FS_ASSERT(fset->fileSetId == FS_FILESET_INVAL_ID);


	/*
	 * Lock it down
	 */

	fs_lockExclusive(context, &fset->lock, NULL);

#if 0
TOBEUPDATED
	/*
	 * Allocate and initialize user and quota inodes
	 */

	uInode = fs_allocateFileInode(context, NULL, FS_IFMT_QUOT, 0, 0, 0);

	if (uInode == NULL) {

		/*
		 * We don't have a way of aborting the transaction.
		 * We should not commit the transaction and return
		 * failure
		 */

		FS_ASSERT(0);
//		fs_error(context, FS_INTERNAL_ERR, "Could not allocate inode");
	}

	FS_ASSERT(uInode != NULL);

	gInode = fs_allocateFileInode(context, NULL, FS_IFMT_QUOT, 0, 0, 0);

	if (gInode == NULL) {

		/*
		 * We don't have a way of aborting the transaction.
		 * We should not commit the transaction and return
		 * failure
		 */

		/*
		 * We need a way to invalidate incore inode
		 */
		fs_error(context, FS_INTERNAL_ERR, "Could not allocate inode");
		fs_inodeReleaseExclusive(context, uInode);
	}

	FS_ASSERT(gInode != NULL);

	FS_ASSERT(uInode != NULL && gInode != NULL);


	/*
	 * allocate and Initialize fileset record
	 */

	rc = fs_initializeFilesetRecord(context, core, parent, fsetRoot,
				   	fsetInfoInode, fset, uInode, gInode);
#endif
	/*
	 * allocate and Initialize fileset record
	 */

	rc = fs_initializeFilesetRecord(context, core, parent, fsetRoot,
				   	fsetInfoInode, fset);

	/*
	 * We failed in allocating the fileset record. Return failure
	 */

	if (!rc) {

		fs_error(context, FS_INTERNAL_ERR,
			"Could not allocate space for fileset record");
		fs_unlockExclusive(context, &fset->lock);
#if 0
TOBERM
		fs_inodeReleaseExclusive(context, uInode);
		fs_inodeReleaseExclusive(context, gInode);
#endif
	}

	FS_ASSERT(fset->fileSetId != FS_FILESET_INVAL_ID);

	fs_unlockExclusive(context, &fset->lock);

	/*
	 * Do a transaction to mark the inode as belonging to a particular
	 * fileset
	 */

	fsetRoot->dinode.fsetNum = fset->fileSetId;
	fsetRoot->fset = fset;
	fset->parent = parent;

	/*
	 * Now release all the data stuctures for updates and commit
	 * of the transaction
	 */

    fs_inodeWrite(context, fsetRoot, INO_WRITE_LOG);

	NFX_SPIN_LOCK(&fset->spinlock);
	fset->incoreFlag &= ~FS_FILESET_NEEDS_INIT;
	NFX_SPIN_UNLOCK(&fset->spinlock);

	/*
	 * Assert that we are still holding fsetInfoInode exclusively.
	 * We cannot assert that we hold fsetRoot as we did an
	 * fs_inodeWrite and transaction might get logged and lock
	 * released befor this point, if we ever introduce preemption,
	 * till then we assert
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fsetRoot->lock) &&
		  ((parent == NULL && fsetRoot->inum == FS_ROOT_INODE) ||
		   fs_lockOwnedExclusive(context, &parent->lock)));

	return fset;
}

/*
 * This routine is called at mkfs time with user quota inode
 * and group quota inode held. This is a clear locking hierarchy
 * violation, though we don't expect a deadlock as this is mkfs time
 * and filesystem is not in use
 */

EXPORTED fs_FileSetIncoreRecord_t *
fs_makeDefaultFileSet(
	fs_Context					*context,
	fs_FileSystem					*core,
	fs_FileSetIncoreRecord_t	*parent,	// parent of the fileset
	fs_Inode					*fsetRoot,	// Fileset for this inode
	fs_Inode					*fsetInfoInode	// fileset info file inode
)
{
	fs_FileSetIncoreRecord_t	*fset;
	int							rc;

	/*
	 * Assert that the arugments are valid
	 */

	FS_ASSERT(context != NULL && core != NULL &&
			  fsetRoot != NULL && FS_I_IS_DIR(fsetRoot->dinode.attrs) &&
			  (parent != NULL || fsetRoot->inum == FS_ROOT_INODE));

	/*
	 * Assert that fsetInfoInode is held exclusively. This
	 * protects new fsets from being created and fsets from
	 * being removed. Also assert that fsetRoot is exclusively
	 * held at this point. Also assert that parent is exclusively
	 * held
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fsetRoot->lock) &&
		  (parent == NULL || fs_lockOwnedExclusive(context, &parent->lock)));

#if 0	//Code from makeFileSet, not needed here. 
		//Remove it for better understanding. LB
	/*
	 * Now search for the fileset in the incore structures. If
	 * one exists, we return immediately, otherwise create it
	 * and link it in the appropriate chains.  We don't allow creating
	 * a fset in the middle of an existing hierarchy so parent chains
	 * are not likely to be disturbed. 
	 *
	 * Fset will be associated with with the inode, so while
	 * inode is incore, the fset cannot be removed.
	 */

	fset = fs_allocateIncoreFileSetRecord(context, core,
					      parent, fsetRoot->dinode.fsetNum, 1);

	if (fset == NULL || fset->cachedRecord.rootInode != fsetRoot->inum) {

		/*
		 * We could not find the fileset incore. See if we have
		 * enough room to allocate space in fileset file
		 */

		if (fsetInfoInode->dinode.size >= FS_FILESET_INFO_MAX_SIZE) {

			/*
			 * Find spare room in the file. If we don't have return
			 * failure
			 */

			if (!fs_fsetInfoHasFreeFsets(context, core, fsetInfoInode)) {
				return NULL;
			}
		}
		fset = fs_allocateIncoreFileSetRecord(context, core,
								  			  parent, FS_FILESET_INVAL_ID, 0);
		
	}
#else
	fset = fs_allocateIncoreFileSetRecord(context, core, parent,
											FS_FILESET_INVAL_ID, 0);
#endif

	NFX_SPIN_LOCK(&fset->spinlock);
	if (!(fset->incoreFlag & FS_FILESET_NEEDS_INIT)) {
		FS_ASSERT(fset->fileSetId < FS_FILESET_INVAL_ID);
		NFX_SPIN_UNLOCK(&fset->spinlock);
		return fset;
	}
	NFX_SPIN_UNLOCK(&fset->spinlock);

	/*
	 * This is a newly allocated fileset, assert that its id is not
	 * initialized yet
	 */

	FS_ASSERT(fset->fileSetId == FS_FILESET_INVAL_ID);


	/*
	 * Lock it down
	 */

	fs_lockExclusive(context, &fset->lock, NULL);


	/*
	 * allocate and Initialize fileset record
	 */

	rc = fs_initializeFilesetRecord(context, core, parent, fsetRoot,
				   	fsetInfoInode, fset);

	/*
	 * We failed in allocating the fileset record. Return failure
	 */

	if (!rc) {

		fs_error(context, FS_INTERNAL_ERR,
			"Could not allocate space for fileset record");
		fs_unlockExclusive(context, &fset->lock);
	}

	FS_ASSERT(fset->fileSetId != FS_FILESET_INVAL_ID);

	fs_unlockExclusive(context, &fset->lock);

	/*
	 * Do a transaction to mark the inode as belonging to a particular
	 * fileset
	 */

	fsetRoot->dinode.fsetNum = fset->fileSetId;
	fsetRoot->fset = fset;
	fset->parent = parent;

	/*
	 * Now release all the data stuctures for updates and commit
	 * of the transaction
	 */

    fs_inodeWrite(context, fsetRoot, INO_WRITE_LOG);

	NFX_SPIN_LOCK(&fset->spinlock);
	fset->incoreFlag &= ~FS_FILESET_NEEDS_INIT;
	NFX_SPIN_UNLOCK(&fset->spinlock);

    //This file system has now one fileset
	NFX_SPIN_LOCK(&core->filesetListLock);
    core->filesetsCount = 1;
    NFX_SPIN_UNLOCK(&core->filesetListLock);
    
	/*
	 * Assert that we are still holding fsetInfoInode exclusively.
	 * We cannot assert that we hold fsetRoot as we did an
	 * fs_inodeWrite and transaction might get logged and lock
	 * released befor this point, if we ever introduce preemption,
	 * till then we assert
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fsetRoot->lock) &&
		  ((parent == NULL && fsetRoot->inum == FS_ROOT_INODE) ||
		   fs_lockOwnedExclusive(context, &parent->lock)));

	return fset;
}

/*
 * Remove a user / group quota file associated with a fileset
 */
void
fs_removeUGquotaFile(fs_Context		*context,
					fs_FileSystem	*core,
					uint64			quotaInum)
{
	fs_Inode	*quotaInode;
	
	//Assert that we have correct arguments.
	FS_ASSERT(context != NULL && core != NULL);
	
	//Quotas are applicable only to live filesystem. Snapnum is always 0
	//get exclusive lock since we want to remove the file and the inode

	quotaInode = I_GET_EXCL(0, quotaInum);
	
	FS_ASSERT(quotaInode != NULL);
	
	//assert that quotaInode is held exclusively
	FS_ASSERT(fs_lockHeldExclusive(context, &quotaInode->lock) != 0);
	
	//assert we have the right inode
	FS_ASSERT(quotaInode->inum == quotaInum);

	//deallocate quota file blocks
	fs_truncateFile(context, quotaInode, 0, FALSE, NULL);

	//deallocate quota file inode
	fs_deallocateInode(context, quotaInode);
}


/*
 * Remove a fileset. It is assumed that fileset's root is already 
 * removed from the namespace. Caller will remove the inode in the
 * same transaction as well. Caller should be holding the fset and
 * when done release the fset
 */

EXPORTED void
fs_removeFileSet(
	fs_Context					*context,
	fs_FileSystem				*core,
	fs_FileSetIncoreRecord_t	*fset,			// fileset being removed
	fs_Inode					*fsetRoot,		// root of the fileset
	fs_Inode					*fsetInfoInode	// fileset info file
)
{

	FS_ASSERT(context != NULL && core != NULL && fset != NULL);

	/*
	 * Assert that the root inode is exclusively locked as well
	 * as fsetInfoInode is held exclusively
	 */

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fsetRoot->lock));

    fsetRoot->dinode.fsetNum = fset->parent->fileSetId;
    fs_inodeWrite(context, fsetRoot, INO_WRITE_LOG);
    fsetRoot->fset = NULL;
	fs_lockExclusive(context, &fset->lock, NULL);
	
	/*
	 * Remove the fileset from the hash, so nobody else can look at it.
	 * We are holding the filesetinfo inode in exclusive mode, so nobody
	 * else would be able to create an fset of the same name while we
	 * are removing it.
	 */

	NFX_SPIN_LOCK(&core->filesetListLock);
	fs_unHashFileSet(core, fset, fset->fileSetId,
					 FS_FILESET_HASH(fset->fileSetId));
	NFX_SPIN_UNLOCK(&core->filesetListLock);

	/*
	 * Within the transaction write zeroed fileset disk record
	 */

	fs_deinitializeFileSetRecord(context, core, fsetInfoInode, fset);

	FS_ASSERT(fs_lockOwnedExclusive(context, &fsetInfoInode->lock) &&
		  fs_lockOwnedExclusive(context, &fsetRoot->lock));


	/*
	 * Remove fileset from fileset list. It will unlock the fset as well.
	 */

	fs_freeFileSet(context, core, fset);
	
	return;
}

/*
 * For a given path find the fileset it belongs to
 */

EXPORTED fs_Status
fs_findFset(
	fs_Context					*context,
	fs_FileSystem				*core,
	char						*pt,
	fs_Inode					**root,
	fs_FileSetIncoreRecord_t	**rfset)
{

	fs_path_t	*path;
	fs_Inode	*parentDir = NULL, *component = NULL;
	char		*p;

	*rfset = NULL;
	if (pt == NULL || (pt[0] != '/' && pt[0] != '\\')) {
		return FS_ERR_INVAL;
	}
	path = fs_makePath(pt);

	/*
	 * Start with root inode
	 */

	component = parentDir = I_GET_SHARED(0, FS_ROOT_INODE);
	parentDir->fset = fs_getFset(context, core, parentDir);
	FS_ASSERT(parentDir->fset != NULL &&
			  parentDir->fset->cachedRecord.rootInode == parentDir->inum);

	p = fs_getNextComponent(path);
	while (p != NULL) {

		/*
		 * Always look in live filesystem
		 */

		component = fs_lookupShared(context, core, 0, parentDir, p);
		if (component == NULL) {
			FS_ASSERT(fs_lockOwnedShared(context, &parentDir->lock, TRUE));
			break;
		}

		
		FS_ASSERT(fs_lockOwnedShared(context, &component->lock, TRUE));
		p = fs_getNextComponent(path);
		if (p != NULL) {
			parentDir = component;
			component = NULL;
			if (!FS_I_IS_DIR(parentDir->dinode.attrs)) {
				fs_inodeRelease(context, parentDir);
				parentDir = NULL;
				break;
			}
		}
	}
	fs_destroyPath(path);

	if (component == NULL) {
		fs_inodeRelease(context, parentDir);
		return FS_ERR_NOENT;
	}

	if (!FS_I_IS_DIR(component->dinode.attrs) ||
	    component->fset->cachedRecord.rootInode != component->inum) {
		fs_inodeRelease(context, component);
		return FS_ERR_NOENT;
		
	}
	*root = component;
	*rfset = component->fset;
	return FS_ERR_NONE;
}

/*
 * List the filesets in the file system 
 */

EXPORTED int32
fs_listFileSet(
	fs_Context				    *context,
	fs_FileSystem				*core,
	fs_MsgListFileSetsReq_t	    *req,
	fs_MsgListFileSetsResp_t	*resp
)
{

	/*
	 * It is assumed that the caller has already allocated memory for
	 * the required amount of entries in response structure
	 */

	fs_Inode					*fsetInfoInode;
	int							id = req->startIndex;
	fs_FileSetIncoreRecord_t	*fset;
	fs_FileSetList_t			*list;
	fs_FileSetDiskRecord_t		*dreq;

	/*
	 * Assert that we are not holding any locks
	 */
	FS_ASSERT(context != NULL && core != NULL && req != NULL && resp != NULL);

	/*
	 * Lock the fileset info inode in shared mode, so that we cannot
	 * create or remove filesets while this operation is in progress,
	 */

	fsetInfoInode = I_GET_SHARED(0, FS_FILESET_INFO_INODE);

	/*
	 * Also hold the file system's filesetlistLock.
	 */

	NFX_SPIN_LOCK(&core->filesetListLock);
	
	/*
	 * Iterate through all the filesets and copy quota and consumption
	 * data into the response
	 */

	list = resp->filesets;
	resp->numberEntries = 0;
	DLL_ITERATE(fset, &core->filesetList, fsetList) {
		if (fset == NULL || fset == DLL_NIL) {
				break;
		}
		if (fset->fileSetId >= id) {

			if (fset->cachedRecord.rootInode < FS_ROOT_INODE) {
					continue;
			}

			fs_holdFileSet(context, core, fset);

			/*
			 * Protect against quota reservations
		 	 */

			dreq = &fset->cachedRecord;
			NFX_SPIN_LOCK(&fset->spinlock);
            list->id = fset->fileSetId;
            list->parentId = dreq->parentId;
            list->rootInode = dreq->rootInode;
			NFX_SPIN_UNLOCK(&fset->spinlock);
			fs_releFileSet(context, core, fset);

			resp->numberEntries++;
			list++;
		}
		if (resp->numberEntries >= req->numberEntries) {
				break;
		}
	}

    context->respSize = sizeof(fs_MsgListFileSetsResp_t)
                      + (sizeof(fs_FileSetList_t) * (resp->numberEntries - 1));
    resp->msgHdr.len = context->respSize;
            
	NFX_SPIN_UNLOCK(&core->filesetListLock);
	fs_inodeRelease(context, fsetInfoInode);
	return FS_ERR_NONE;
}

/*
 * Return the disk fset record for snapshot
 */

EXPORTED fs_FileSetDiskRecord_t *
fs_getSnappedFset(
	fs_Context	*context,
	fs_FileSystem	*core,
	fs_Inode	*inode	// Get fileset belonging to this inode
)
{
	fs_FileSetDiskRecord_t	    *dfset;
    fs_Inode                    *fsetInfoInode;
    fs_Buf                      *buf;
    uint64                      offset;
    uint32                      boff;

    FS_ASSERT(inode->snapnum != 0);
    FS_ASSERT(context != NULL && core != NULL && inode != NULL);

	dfset =  (fs_FileSetDiskRecord_t *)fs_ramAlloc(FS_RAM_FILESET,
                                            sizeof (fs_FileSetDiskRecord_t));

    FS_ASSERT(dfset != NULL);

    /*
     * Read in the fset from disk
     */

    fsetInfoInode = I_GET_SHARED(inode->snapnum, FS_FILESET_INFO_INODE);
    offset =((inode->dinode.fsetNum * sizeof (fs_FileSetDiskRecordPadded)) >>
                FS_BLOCK_SHIFT);
    buf = O_GET_EXCL(fsetInfoInode, offset);
    FS_ASSERT(buf != NULL);
    boff = (uint32)(offset & (FS_BLOCK_SIZE - 1));
    bcopy(&buf->u.data[boff], (char *)dfset,
          sizeof (fs_FileSetDiskRecordPadded));
    fs_bufRelease(context, buf);
	return dfset;
}

/*
 * Given a fileset Id, return the complete path to it. Ensure that
 * no locks are held while this is called
 */

#define FS_PARENT_ARRAY_LEN 10
typedef struct fs_parentArray {
    struct fs_parentArray   *next, *prev;
    uint16      nEntries;
    fs_Inumber  array[FS_PARENT_ARRAY_LEN];
} fs_ParentArray_t;

PRIVATE fs_Status
fs_getDirEntry(fs_Context *context,
               fs_Inode   *dir,
               fs_Inumber entryInum,
               char *buf)
{
    fs_DirEntry *dirents, *dirp;
    int         eof;
    uint64      cookie;
    int         numEntries;
    int         i;
    
    FS_ASSERT(fs_lockHeldShare(context, &dir->lock));
    dirents = fs_ramAlloc(FS_RAM_GET_FILESET_PATH, FS_MAXPATHLEN + 1);
    FS_ASSERT(dirents != NULL);
    dirp = dirents;
    eof = 0;
    cookie = 0;
    while (!eof) {
        fs_dir_readdir(context, dir, dirents, FS_MAXPATHLEN + 1,
						cookie, FALSE, &numEntries, &eof);
        dirp = dirents;

        FS_TRACE(context, FS_TRACE_REQ, ("numEntries is %d\n", numEntries));
        for (i = 0; i < numEntries; i++) {
            cookie = dirp->cookie;
            if (dirp->inode == entryInum) {
                strncpy(buf, dirp->name, dirp->nameLen);
                buf[dirp->nameLen] = '\0';
                fs_ramDealloc(FS_RAM_GET_FILESET_PATH, dirents);
//                fs_inodeRelease(context, dir);
                return FS_ERR_NONE;
            }
            FS_READDIR_ADV_ENTRY_PTR(dirp);
        }
    }
    fs_ramDealloc(FS_RAM_GET_FILESET_PATH, dirents);
//    fs_inodeRelease(context, dir);
    return FS_ERR_NOENT;
}

PRIVATE char*
fs_getFileSetPath(
    fs_Context      *context,
    uint32          fsetId,
    fs_Inumber      filesetRootInum,
    uint64          snapNum
)
{
    char        *buf, *cp;
    fs_ParentArray_t    *parray = NULL, *narray = NULL;
    int                 nEntries = 0;
    fs_Inumber          parentInum;
    fs_Inode            *inode;
    fs_Inumber          inum = filesetRootInum;
    uint64              cookie = 0;
    fs_Status           status;

    if (filesetRootInum == FS_ROOT_INODE) {
        return "/";
    }

    inode = I_GET_SHARED(snapNum, inum);
    if (inode == NULL) {
        while (parray) {
            narray = parray->next;
//            fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
            parray = narray;
        }
        return NULL;
    }

    if (inode->dinode.fsetNum != fsetId) {
        while (parray) {
            narray = parray->next;
//           fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
            parray = narray;
        }
        fs_inodeRelease(context, inode);
        return NULL;
        
    }

    if (!FS_I_IS_DIR(inode->dinode.attrs)) {
        while (parray) {
            narray = parray->next;
//            fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
            parray = narray;
        }
        fs_inodeRelease(context, inode);
        return NULL;
    }

    fs_inodeRelease(context, inode);
    status = fs_findParent(context, snapNum, inum, &parentInum, &cookie);
    if (status != FS_ERR_NONE) {
        while (parray) {
            narray = parray->next;
//            fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
            parray = narray;
        }
        fs_inodeRelease(context, inode);
        return NULL;
    }

    narray = parray = fs_ramAlloc(FS_RAM_GET_FILESET_PATH,
                                  sizeof (fs_ParentArray_t));
    bzero(narray, sizeof (fs_ParentArray_t));
    FS_ASSERT(parray != NULL);
    narray->next = NULL;
    narray->array[narray->nEntries++] = filesetRootInum;
    narray->array[narray->nEntries++] = parentInum;
    inum = parentInum;
    while (inum != FS_ROOT_INODE) {
        cookie = 0;
        status = fs_findParent(context, snapNum, inum, &parentInum, &cookie);
        if (status != FS_ERR_NONE) {
            while (parray) {
                narray = parray->next;
                fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
                parray = narray;
            }
            fs_inodeRelease(context, inode);
            return NULL;
        }
        narray->array[narray->nEntries++] = parentInum;
        if (narray->nEntries >= FS_PARENT_ARRAY_LEN
         && parentInum != FS_ROOT_INODE)
        {
            narray->next =  fs_ramAlloc(FS_RAM_GET_FILESET_PATH,
                                        sizeof (fs_ParentArray_t));
            narray->next->prev = narray;
            narray->next->next = NULL;
            narray = narray->next;
            narray->nEntries = 0;
            FS_ASSERT(narray);
        }
        inum = parentInum;
    }

    /*
     * Now that we have collected all the ancestors, construct the path
     */

    buf = fs_ramAlloc(FS_RAM_GET_FILESET_PATH, FS_MAXPATHLEN + 1);
    buf[0] = '\0';
    cp = buf;
    while (narray->next) {
        narray = narray->next;
    }

    nEntries = narray->nEntries - 1;
    while (nEntries >= 0) {
        parentInum = narray->array[nEntries];
        nEntries--;
        if (nEntries >= 0) {
            inum = narray->array[nEntries];
        } else if (narray->prev == NULL) {
            break;  // This is the fileset root. We don't need to scan further
        } else {
            narray = narray->prev;
            nEntries = narray->nEntries - 1;
            inum = narray->array[nEntries];
        }
        inode = I_GET_SHARED(0, parentInum);
        if (!FS_I_IS_DIR(inode->dinode.attrs)) {
            fs_inodeRelease(context, inode);
            while (parray) {
                narray = parray->next;
                fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
                parray = narray;
            }
            fs_ramDealloc(FS_RAM_GET_FILESET_PATH, buf);
            return NULL;
        }
        *cp++ = '/';
        status = fs_getDirEntry(context, inode, inum, cp);
        fs_inodeRelease(context, inode);
        if (status != FS_ERR_NONE) {
            while (parray) {
                narray = parray->next;
                fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
                parray = narray;
            }
            fs_ramDealloc(FS_RAM_GET_FILESET_PATH, buf);
            return NULL;
        }
        cp += strlen(cp);
    }
    while (parray) {
        narray = parray->next;
        fs_ramDealloc(FS_RAM_GET_FILESET_PATH, parray);
        parray = narray;
    }
    return buf;
}

int32
fs_getFilesetPathThread(fs_Context *context)
{
    fs_MsgGetFilesetPathReq_t *req;
    fs_MsgGetFilesetPathResp_t *resp;
    uint16          fsetId;
    fs_Inumber      inum;
    uint64          snapNum;
    char            *buf;

    req = (fs_MsgGetFilesetPathReq_t *)context->reqHdr;
	resp = (fs_MsgGetFilesetPathResp_t *) context->respHdr;
    inum = req->inum;
    snapNum = req->snapnum;
    fsetId = req->fsetId;
    buf = fs_getFileSetPath(context, fsetId, inum, snapNum);

    /*
     * Don't free the buffer if the fileset path is "/" as 
     * we did not malloc for that case
     */

    if (buf) {
        strcpy(resp->path, buf);
        if (strcmp(buf, "/") != 0) {
            fs_ramDealloc(FS_RAM_GET_FILESET_PATH, buf);
        }
        return FS_ERR_NONE;
    }
    return FS_ERR_NOENT;
}

#endif // SSC
