/*
 * Name			: fs-fileset.h
 * RCS			: $ID$
 * Description	: Quota implementation data structures
 * Created by	: Rahul Khona
 * Date Created	: 8/10/2001
 */

#ifndef _FS_FILESET_H
#define _FS_FILESET_H

/*
 * This file describes the incore file set info record. It hangs off a
 * freelist and a hashlist which are anchored in fs_FileSystem structure
 * for now. Later they should be moved to a global queue.
 *
 * lock is an iolock structure which enables only one context to log
 * and update ondisk values
 *
 * Spin lock would protect incoreRecords which are used for tracking
 * incore reservations. Once we are ready to commit the transaction
 * we would lock the structure exclusively, apply the reservation for the
 * particular request which holds it exclusively to the disk records and
 * keep the structure locked till it is logged. The reason is that quota
 * structures can be point of contention, if several allocation requests
 * come in for the same fileset. By separating incore reservation of quotas
 * from actual update of quotas to disk, we don't have to keep the
 * quota structure for the user locked for the whole duration and
 * thus reduce contention.
 *
 * In addition disk quota have to be written in log order, so that
 * replay of transaction log will not reset the quota and consumption
 * information incorrectly. The problem would be best explained by
 * an example. If 2 requests came in for fileset A simultaneously to allocate
 * 5 blocks to file 1 and 10 blocks to file 2. If the consumption at the
 * before these requests was say Y. If the ondisk updates were in the
 * order of Y + 5, followed by (Y + 5) + 10 and if the log was written
 * in (Y + 5) + 10 and Y + 5, then replay could show the consumption
 * as Y + 5 where as it should have been Y + 15. Remember that on disk
 * we record absolute consumption value and not the deltas that need
 * to be applied as we log the disk blocks and not the operations.
 *
 * //NOT IMPLEMENTED TODAY
 * count is protected by the list spin lock. It indicates how many
 * contexts have this structure in use. Routines that make the structure
 * in use should increment the count holding the freelist lock and which
 * release the structure should decrement the count. If the count drops to
 * 0, the structure should be put back on freelist and if the count changes
 * from 0 to 1, it should be taken off the freelist.
 *
 * //TODO CHANGE THE FOLLOWING COMMENT TO:
 * We cache the fileset on disk structure in the incore structure
 * //TODO SO REMOVE THIS:
 * We cache the rootinode of the fileset in the incore structure also
 * as on disk structure can be freed while the incore structure is still
 * in use.
 *
 * //TODO REMOVE THE FOLLOWING COMMENT ?
 * allocation version refers to the allocation version number of the
 * filesystem when quota on was issued. It helps us in determining which
 * files don't have their quota accounting performed while quota on 
 * process is in progress. The basic theory is that every time a block
 * or an inode is to be udpated a global counter would be incremented
 * and the value recorded in the associated inode. Whenever quota on is
 * issued that value would be cached in the fileset for which it is issued.
 * Quota on process would scan through the directory substree and for each
 * inode, if the inode's allocation version was greater than the
 * cached allocation version, then the consumption accounting would have
 * already been done and it can be skipped.
 */

/*
 * Quota record on disk. This will be used in variety of quota files
 * such as user quota file, group quota file and fileset info file
 */

typedef struct fs_QuotaRecord {
	uint64	inodeSoftLimit;	//Soft limit for number of inodes
	uint64	inodeHardLimit;	//Hard limit for number of inodes
	uint64	inodeTimeLimit;	//Time limit for number of inodes
	uint64	blockSoftLimit;	//Soft limit for number of blocks
	uint32	blockHardLimit;	//Hard limit for number of blocks
	uint32	blockTimeLimit;	//Time limit for number of blocks
} fs_QuotaRecord_t;	/* 40 bytes */

/*
 * Resource consumption record on disk. This will be used in variety of
 * quota files such as user quota, group quota nad file set info file
 *
 * Time would always be stored on disk starting at Jan 1, 1970. 
 */

typedef struct fs_ConsumptionRecord {
	uint64	inodes;				// Inodes consumed
	uint64	blocks;				// blocks consumed
	uint64	timeExtraInodes;	// time when inode softlimit was exceeded
	uint64	timeExtraBlocks;	// time when block softlimit was exceeded
} fs_ConsumptionRecord_t;	/* 32 bytes */

/*
 * Fileset record structure on disk. Each record would be 2048 bytes on disk.
 * Enough room is left for future growth of fileset information. It might
 * change as part of this release or future disk layout revisions as well
 */

typedef enum {
    FS_QUOTA_OFF,           // feature disabled
    FS_QUOTA_BUILDING,      // quota is being turned on
    FS_QUOTA_TRACKED,       // only usage calculations is performed
    FS_QUOTA_ENFORCED,      // quota limits are enforced as well
} fs_quota_state;

/*
 * Valid values for fileSetFlags
 */

typedef struct fs_FileSetDiskRecord {
	uint64					rootInode;			// root inode for this fileset
	uint32					parentId;			// position of parent file sets
												// record in fileset info file
	uint32					fileSetFlags;		// flags for file set
    fs_quota_state          quotaState;         // quota feature state
	fs_QuotaRecord_t		defaultAggregateQuota;	// default quotas
	fs_QuotaRecord_t		fileSetQuota;		// Aggr. quota for this fileset
	fs_ConsumptionRecord_t	fileSetConsumption;	// consumption info
	uint32					lastChanged;	    // time when last modified
                                                // to incorporate new limits
    uint32                  bandWidth;          // Bandwidth quota not
                                                // enforced yet
} fs_FileSetDiskRecord_t;	/* 512 bytes */

typedef union fs_FileSetDiskRecordPadded {
    fs_FileSetDiskRecord_t  drec;
    char8                   bytes[512];
} fs_FileSetDiskRecordPadded;

NFX_ASSERT_STATIC(FS_BLOCK_SIZE % sizeof (fs_FileSetDiskRecordPadded) == 0);
#define FS_FILESET_RECS_PER_PAGE    \
    (FS_BLOCK_SIZE / sizeof (fs_FileSetDiskRecordPadded))

#define FS_FILESET_REC_SHIFT    9

NFX_ASSERT_STATIC((1 << FS_FILESET_REC_SHIFT) == 
                sizeof (fs_FileSetDiskRecordPadded));

#define	FS_FILESET_QUOTA_ON			0x00000001	// Quota is on
#define	FS_FILESET_QUOTA_TURN_ON	0x00000002	// quota is turning on
#define	FS_FILESET_QUOTA_ON_THREAD 	0x00000004	// quota thread present
#define	FS_FILESET_NEEDS_INIT		0x00000008	// fset needs to be read
#define	FS_FILESET_DIRTY			0x00000010	// fset needs to be flushed
#define	FS_FILESET_USE_QDEFAULTS	0x00000020	// fset uses quota defaults
#define FS_FSET_QNOT_BLK_TIMELIMIT	0x80000000	// fset quota notification flags
#define FS_FSET_QNOT_BLK_HARDLIMIT	0x40000000
#define FS_FSET_QNOT_INO_TIMELIMIT	0x20000000
#define FS_FSET_QNOT_INO_HARDLIMIT	0x10000000


typedef struct fs_FileSetIncoreRecord fs_FileSetIncoreRecord_t;
struct fs_FileSetIncoreRecord {
	DLL_ENTRY(fs_FileSetIncoreRecord_t)	fsetList;
	DLL_ENTRY(fs_FileSetIncoreRecord_t)	hashList;
	fs_Lock									lock;		// lock the structure
	nfx_spinlock_t							spinlock;	// protect incore
														//reservations
	int32									count;		// unused
	uint32									fileSetId;	// fileset id
	struct	fs_FileSetIncoreRecord 			*parent;	// parent of this
														// fileset
	uint32									incoreFlag;	// incore flags
	uint64									inodeSeen;	// inode seen
														// by quota on
	fs_FileSetDiskRecord_t					cachedRecord;	// incore copy

	/*
	 * Traverse lock is used to protect traversal through filesets. In
	 * locking hierarchy it is above the inode lock and can be held while
	 * initializing a transaction as it does not prevent flushing of
	 * fileset buffers and freeing them.
	 *
 	 * Traverse locks can be nested in the tree order i.e parent first
	 * followed by its sibling. Only 1 sibling can be locked in a parent
	 */

	fs_Lock									traverseLock;	
};

/*
 * Maximum size of the file set info file.
 */

#define	FS_FILESET_MAX_ID			63
#define	FS_FILESET_INVAL_ID			(FS_FILESET_MAX_ID + 1)
#define	FS_FILESET_INFO_MAX_SIZE	\
	(sizeof (fs_FileSetDiskRecord_t) * (FS_FILESET_MAX_ID + 1))

#endif	/* _FS_FILESET_H */

