// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-inlines.c
 *
 * RCS:  $Id: fs-inlines.c,v 1.27 2002/10/15 22:06:33 maximk Exp $
 *
 * Description:  Inline functions.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-06-25
 */

#if !defined(SSC)

#if !defined(FS_INLINES_C)
#define FS_INLINES_C

/*
 * The INLINE_FUNCTION macro has been declared externally if we're being
 * included in the compilation of another file, so this conditional is
 * used to determine whether we're being compiled standalone.
 */
#if !defined(INLINE_FUNCTION)

#include "fs-includes.h"

#if defined(FS_USE_INLINES)

/*
 * If this is a production system, we don't need the standalone versions
 * of the inline functions, and they'll probably confuse matters.
 */
#define INLINE_FUNCTION(proto, body) 
#else
/*
 * When we're compiled for development purposes, we don't want to inline
 * functions, because it gets in the way of debugging.  In that case, the
 * functions are defined in this file.
 */
#define INLINE_FUNCTION(proto, body) proto body
#endif

#endif  /* INLINE_FUNCTION */



/*
 * F U N C T I O N    D E F I N I T I O N S
 */


//   U T I L I T I E S


/**
 * Return the ceiling of the quotient of two 64-bit values.
 */
INLINE_FUNCTION(
    uint64 ceilingDiv64(uint64 num, uint64 den),
    {
        return (num + den - 1) / den;
    })



//   B L O C K S   &   E X T E N T S


/**
 * Create a block struct for the specified logical block number.
 * The block number is an unsigned 48-bit quantity.
 *
 * TODO:  Should we pass in the context here, so we can
 * check that the block number is within the size of the FS?
 */
INLINE_FUNCTION(
    fs_Block fs_makeBlock(uint64 logicalBlockNumber),
    {
        fs_Block block;
        block.block = logicalBlockNumber;
        return block;
    })


/**
 * Create the canonical invalid block.
 */
INLINE_FUNCTION(
    fs_Block fs_makeInvalidBlock(),
    {
        fs_Block block;
        block.block = FS_INVALID_BLKNUM;
        return block;
    })


/**
 * Return the logical block number for the given block.
 */
INLINE_FUNCTION(
    uint64 fs_blockNumber(fs_Block block),
    {
        return block.block;
    })


/**
 * Return true if two blocks refer to the same storage blocks.
 */
INLINE_FUNCTION(
    uint64 fs_blockEq(fs_Block block1, fs_Block block2),
    {
        return block1.block == block2.block;
    })


/**
 * Return true if the block has a valid block number.
 */
INLINE_FUNCTION(
    boolean fs_blockIsValid(fs_Block block),
    {
        return block.block < FS_VALID_BLOCK_LIMIT;
    })


/**
 * Create an extent of the specified length starting at the specified
 * block.
 */
INLINE_FUNCTION(
    fs_Extent fs_makeExtent(uint16 length, fs_Block startBlock),
    {
        fs_Extent extent;
        extent.lengthAndStartBlock
            = ((((uint64) length) << EXTENT_BLOCK_BITS)
               | fs_blockNumber(startBlock));
        return extent;
    })


/**
 * Return the length in blocks of the specified extent.
 */
INLINE_FUNCTION(
    uint16 fs_extentLength(fs_Extent extent),
    {
        return (uint16) (extent.lengthAndStartBlock >> EXTENT_BLOCK_BITS);
    })


/**
 * Return the block at the specified block offset within the extent.
 * The offsets are numbered from 0, inclusive, to the size of the
 * extent, exclusive.
 */
INLINE_FUNCTION(
    fs_Block fs_extentBlock(fs_Extent extent, uint16 blockOffset),
    {
        FS_ASSERT(blockOffset < fs_extentLength(extent));
        return fs_makeBlock(
            (extent.lengthAndStartBlock & EXTENT_BLOCK_MASK) + blockOffset);
    })


/**
 * Return the start block of the specified extent, even if the extent
 * has zero length.  This function should only be used for debugging.
 */
INLINE_FUNCTION(
    fs_Block fs_extentStartBlock(fs_Extent extent),
    {
        return fs_makeBlock(extent.lengthAndStartBlock & EXTENT_BLOCK_MASK);
    })


/**
 * Split the given extent at the given blockOffset,
 * 
 */
INLINE_FUNCTION(
    fs_Extentpair fs_splitExtent(fs_Extent extent, uint16 blockOffset),
    {
       	uint32 length = fs_extentLength(extent);
	uint64 start = fs_blockNumber(fs_extentBlock(extent, 0));
	fs_Extentpair r;

	r.left = fs_makeExtent(blockOffset, fs_makeBlock(start));
	r.right = fs_makeExtent(length - blockOffset, fs_makeBlock(start + blockOffset));

	return r;
    })

/**     
 * Merge two contiguous extents
 *      
 */     
INLINE_FUNCTION(
    fs_Extent fs_mergeExtents(fs_Extent left, fs_Extent right),
    {
        uint32 length2 = fs_extentLength(left) + fs_extentLength(right);
        fs_Block start = fs_extentBlock(left, 0);

        assert(length2 <= EXTENT_LENGTH_MAX);
        assert(fs_extentLength(left) + fs_blockNumber(start) ==
                fs_blockNumber(fs_extentBlock(right, 0)));
        
        return fs_makeExtent(length2, start);
    })

/**     
 * Return TRUE if two extents at the given offsets in the same
 * extent map are adjacent and so would become candidates for
 * merging.
 */     
INLINE_FUNCTION(
    boolean fs_areAdjacent(fs_Extent left, fs_Extent right, uint64 oleft, uint64 oright),
    {
	return (fs_extentLength(left) + fs_blockNumber(fs_extentBlock(left, 0)) == 
		fs_blockNumber(fs_extentBlock(right, 0))) &&
	       (fs_extentLength(left) + oleft == oright) &&
               (fs_extentLength(left) + fs_extentLength(right) <= EXTENT_LENGTH_MAX);
    })



/**
 * Return the packed eee port address for this file processor and the
 * supplied data type.
 */
INLINE_FUNCTION(
    uint32 fs_packedPortAddr(int32 cpuId, int32 dataType),
    {
        return fs_srcPortId | (dataType << EDESC2_DATA_TYPE_SHIFT);
    })



//   B I T   O P E R A T I O N S


/**
 * Align the specified unsigned 32-bit value to a multiple of the
 * specified alignment, which must be a power of two no smaller than the
 * original value.
 */
INLINE_FUNCTION(
    uint32 alignto(uint32 n, uint32 align),
    {
        return (n + align - 1) &~ (align - 1);
    })

    
/**
 * Align the specified unsigned 32-bit value to a multiple of the
 * specified alignment, which must be a power of two no smaller than the
 * original value.
 */
INLINE_FUNCTION(
    uint64 alignto64(uint64 n, uint64 align),
    {
        return (n + align - 1) &~ (align - 1);
    })


/**
 * Return true for even numbers.
 */
INLINE_FUNCTION(
    boolean isEven(uint32 n),
    {
        return ((n & 1) == 0);
    })


/**
 * Return the smallest bit set in a 32-bit word.
 */
INLINE_FUNCTION(
    uint32 leastBit(uint32 n),
    {
        return n &~ (n - 1);
    })

/**
 * Return the smallest bit set in a 64-bit word.
 */
INLINE_FUNCTION(
    uint64 leastBit64(uint64 n),
    {
        return n &~ (n - 1);
    })


/**
 * Return true if the specified quantity is a power of two.
 */
INLINE_FUNCTION(
    boolean isPowerOfTwo(uint32 n),
    {
        return n == leastBit(n);
    })


/**
 * Return true if the specified quantity is a power of two.
 */
INLINE_FUNCTION(
    boolean isPowerOfTwo64(uint64 n),
    {
        return n == leastBit64(n);
    })


/**
 * Round the specified quantity up to the next power of two.
 */
INLINE_FUNCTION(
    uint32 powerOfTwoCeiling(uint32 n),
    {
        while (TRUE) {
            uint32 lowBit = leastBit(n);
            if (n == lowBit)
                return n;
            n += lowBit;
        }
    })


/**
 * Round the specified quantity down to the next power of two.
 */
INLINE_FUNCTION(
    uint32 powerOfTwoFloor(uint32 n),
    {
        while (TRUE) {
            uint32 lowBit = leastBit(n);
            if (n == lowBit)
                return n;
            n -= lowBit;
        }
    })


/**
 * Single iteration step for base 2 logarithm functions.
 */
#define LOG_STEP(log, type, value, bits)        \
    do {                                        \
        type shift = 1 << (bits);               \
        type x = value >> shift;                \
        if (x != 0) {                           \
            value = x;                          \
            log += shift;                       \
        }                                       \
    } while (0)


/**
 * Return the base two logarithm for a power of two.
 */
INLINE_FUNCTION(
    int log2_32(uint32 n),
    {
        uint32 log2 = 0;
        uint32 m = n;

        assert(n > 0);
        assert(isPowerOfTwo(n));

        LOG_STEP(log2, uint32, m, 4);
        LOG_STEP(log2, uint32, m, 3);
        LOG_STEP(log2, uint32, m, 2);
        LOG_STEP(log2, uint32, m, 1);
        LOG_STEP(log2, uint32, m, 0);

        //printf("log2(%d) = %d\n", n, log2);
        return log2;
    })


/**
 * Return the base two logarithm for a power of two.
 */
INLINE_FUNCTION(
    int log2_64(uint64 n),
    {
        uint64 log2 = 0;
        uint64 m = n;

        assert(n > 0);
        assert(isPowerOfTwo64(n));

        LOG_STEP(log2, uint64, m, 5);
        LOG_STEP(log2, uint64, m, 4);
        LOG_STEP(log2, uint64, m, 3);
        LOG_STEP(log2, uint64, m, 2);
        LOG_STEP(log2, uint64, m, 1);
        LOG_STEP(log2, uint64, m, 0);

        //printf("log2(%lld) = %lld\n", n, log2);
        return log2;
    })


/**
 * Count the bits in a 64-bit word.
 *
 * TODO - take advantage of hardware?
 * TODO - optimize using a static array of counts per byte
 */
INLINE_FUNCTION(
    uint32 fs_popCount64(uint64 word),
    {
        int popc = 0;
        uint64 w = word;
        while (w != 0) {
            popc += 1;
            w &= ~leastBit64(w);
        }
        return popc;
    })



#undef LOG_STEP

#endif /* !defined(FS_INLINES_C) */

#endif // SSC
