// -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-inode.c
 *
 * RCS:  $Id: fs-inode.c,v 1.243 2002/12/06 18:05:47 bob Exp $
 *
 * Description:  Code related to inodes.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

#define INO_TRACE(args)   FS_TRACE(context, FS_TRACE_INO, args)

// There can be a few more inodes in the cache than is
// indicated in usedCount, so don't complain unnecessarily.
#define LOOP_DBG_FUDGE  500


/**
 * Print the contents of an in-core inode.
 */
void
fs_inodeDump(
    char        *str,
    fs_Inode    *inode)
{
    printf("%s i%lld/%d %p lk.own %lld lk.mode %d flags %x\n",
           str,
           inode->inum, inode->snapnum, inode,
           (uint64)((uintptr_t)inode->lock.owner), inode->lock.mode,
		   inode->flags);
}


/**
 * Print the full contents inode cache, for debugging purposes.
 */
void fs_inodeCacheDump(
    fs_Context *context)
{
    fs_Inode *inode;
    int32    hash;

    printf("Inode cache:\n");
    printf("cached %d limit %d dirty %d\n",
            fs_inodeCache->count.numCached,
            fs_inodeCache->count.numCachedLimit,
            fs_inodeCache->count.numDirty);

    printf("\nMRU list:\n");
    DLL_ITERATE_BACK(inode, &fs_inodeCache->mruHdr, mruEntry)
    {
        fs_inodeDump("mru", inode);
    }

    printf("\nHash buckets:\n");
    for (hash = 0; hash < FS_ICACHE_HASH_MAX; hash += 1)
    {
        char    bucketStr[32];

        sprintf(bucketStr, "bucket %d ", hash);
        DLL_ITERATE(inode, &fs_inodeCache->cache[hash], cacheEntry)
        {
            fs_inodeDump(bucketStr, inode);
        }
    }
    printf("\n");
}


/**
 * Return true if the specified inode is marked "no snapshot".  (Also
 * returns true for a null inode.)
 */
boolean fs_inodeIsNoSnap(
    fs_Context *context,
    fs_Inode *inode)
{
    if (inode == NULL)
        return FALSE;
    return (inode->dinode.attrs & FS_IFLG_NO_SNAP) != 0;
}


/**
 * Return the percentage of the inode cache
 * that is currently available (not dirty).
 */
int32
fs_inodePercentAvail()
{
#ifdef FS_DYN_CACHE_SIZE
    int32 numAvail = fs_inodeCache->count.numCachedLimitLWM -
            fs_inodeCache->count.numDirty;

    return (numAvail * 100) / fs_inodeCache->count.numCachedLimitLWM;
#else
    int32 numAvail = fs_inodeCache->count.numCachedLimit -
            fs_inodeCache->count.numDirty;

    return (numAvail * 100) / fs_inodeCache->count.numCachedLimit;
#endif
}


/**
 * Function to indicate whether the specified inode is
 * available for reclamation.
 */
static inline boolean
fs_inodeIsAvail(
    fs_Inode    *inode)
{
    return ((inode->flags & ~INO_FLAGS_RECLAIM_OK) == 0);
}


/**
 * Compute security descriptor hash based on secId
 * in security file.
 */
static inline uint32
fs_securityDescSecIdHash(
    uint32  secId)
{
    return secId % FS_SDESC_HASH_MAX;
}


/**
 * Compute security descriptor hash based on uid.
 */
static inline uint32
fs_securityDescUidHash(
    uint32  uid)
{
    return uid % FS_SDESC_HASH_MAX;
}


/**
 * Caller releases interest in the specified cached 
 * security descriptor.
 */
static void
fs_securityDescRelease(
    fs_Context          *context,
    fs_SecurityDescCore *sdc)
{
    int32   type = (sdc->attr & FS_SDESC_TYPE_MASK);

    //printf("sd rel\n");

    NFX_SPIN_LOCK(&fs_securityDesc.spin);
    sdc->refCount -= 1;
    if (sdc->refCount == 0)
    {
        //printf("sd rel onto free list\n");
        DLL_INSERT_HEAD(&fs_securityDesc.freeHdr[type], sdc, freeEntry);
        fs_securityDesc.numFree[type] += 1;
    }
    NFX_SPIN_UNLOCK(&fs_securityDesc.spin);
}


/**
 * Locate a free in-core security descriptor or malloc
 * a new one if limits allow.
 *
 * !!! Must be called with the cache spinlock held !!!
 */
static fs_SecurityDescCore *
fs_securityDescFindFree(
    fs_Context  *context,
    int32       type)
{
    fs_SecurityDescCore *sdc;

    ASSERT_LOCK_OWNED(&fs_securityDesc.spin);

    if (fs_securityDesc.numUsed[type] < fs_securityDesc.numUsedLimit[type])
    {
        //printf("sd alloc new\n");
        sdc = fs_ramAlloc(FS_RAM_SDESC, sizeof *sdc);
        bzero(sdc, sizeof *sdc);
        sdc->attr = type;
        fs_securityDesc.numUsed[type] += 1;
    }
    else
    {
        int32   hash;

        //printf("sd reuse\n");
        DLL_REMOVE_TAIL(&fs_securityDesc.freeHdr[type], sdc, freeEntry);
        if (sdc == NULL)
        {
            printf("No free in-core security descriptors\n");
            NFX_SPIN_UNLOCK(&fs_securityDesc.spin);
            fs_volumeException(context, context->fileSystem, 
                                FS_VOLUME_EXCEPTION);
            /* NOTREACHED */
        }
        FS_ASSERT(sdc->refCount == 0);

        fs_securityDesc.numFree[type] -= 1;

        if (!DLL_IS_ISOLATED(sdc, secIdEntry))
        {
            hash = fs_securityDescSecIdHash(sdc->secId);
            DLL_REMOVE(&fs_securityDesc.secIdHdr[hash], sdc, secIdEntry);
        }

        if (!DLL_IS_ISOLATED(sdc, uidEntry))
        {
            if (type == FS_SDESC_TYPE_NFS)
            {
                hash = fs_securityDescUidHash(sdc->u.nfs.uid);
            }
            else
            {
                hash = -1;  // compiler warning
                // Not supported yet.
                FS_ASSERT(FALSE);
            }
            DLL_REMOVE(&fs_securityDesc.uidHdr[hash], sdc, uidEntry);
        }
    }

    return sdc;
}


/**
 * Insert the specified security descriptor into the cache.
 * Assumed duplicate check already performed.  Returns a
 * pointer to the cached copy (pinned).
 *
 * !!! Must be called with the cache spinlock held !!!
 */
static fs_SecurityDescCore *
fs_securityDescCacheInsertNfs(
    fs_Context          *context,
    uint8               snapNum,
    uint32              secId,
    uint32              hash,
    fs_SecurityDescNfs  *insert)
{
    fs_SecurityDescCore *sdc;
    fs_SecurityDescNfs  *nfs;
    int32               uidHash = fs_securityDescUidHash(insert->uid);

    ASSERT_LOCK_OWNED(&fs_securityDesc.spin);

    sdc = fs_securityDescFindFree(context, FS_SDESC_TYPE_NFS);
    nfs = &sdc->u.nfs;

    sdc->volId.whole_thing = context->volId.whole_thing;
    sdc->volId = context->volId;
    sdc->snapNum = snapNum;
    sdc->attr = FS_SDESC_TYPE_NFS;
    sdc->secId = secId;
    sdc->refCount = 1;
    *nfs = *insert;

    //printf("sd cache insert nfs secId %d uid %d gid %d mode %o\n",
    //    secId, insert->uid, insert->gid, insert->mode);

    DLL_INSERT_HEAD(&fs_securityDesc.secIdHdr[hash], sdc, secIdEntry);
    DLL_INSERT_HEAD(&fs_securityDesc.uidHdr[uidHash], sdc, uidEntry);

    return sdc;
}


/**
 * Flush the security descriptor cache for the
 * specified filesystem.
 */
static void
fs_securityDescCacheFlush(
    fs_Context  *context,
    int32       snapNum)
{
    fs_SecurityDescCore *sdc, *sdcNext;
    uint32              hash;
    int32               type;

    //printf("sd cache flush\n");

    NFX_SPIN_LOCK(&fs_securityDesc.spin);
    for (type = 0; type < FS_SDESC_TYPE_MAX; type++)
    {
        DLL_ITERATE_NEXT(sdc, sdcNext, &fs_securityDesc.freeHdr[type],
                         freeEntry)
        {
            if (sdc->volId.whole_thing == context->volId.whole_thing)
            {
                if (snapNum && (sdc->snapNum != snapNum))
                    continue;

                if (sdc->refCount != 0)
                    continue;

                DLL_REMOVE(&fs_securityDesc.freeHdr[type], sdc, freeEntry);
                if (!DLL_IS_ISOLATED(sdc, uidEntry))
                {
                    FS_ASSERT(sdc->attr & FS_SDESC_TYPE_NFS);
                    hash = fs_securityDescUidHash(sdc->u.nfs.uid);
                    DLL_REMOVE(&fs_securityDesc.uidHdr[hash],
                               sdc, uidEntry);
                }
                if (!DLL_IS_ISOLATED(sdc, secIdEntry))
                {
                    hash = fs_securityDescSecIdHash(sdc->secId);
                    DLL_REMOVE(&fs_securityDesc.secIdHdr[hash],
                               sdc, secIdEntry);
                }
                fs_ramDealloc(FS_RAM_SDESC, sdc);
                fs_securityDesc.numUsed[type] -= 1;
                fs_securityDesc.numFree[type] -= 1;
            }
        }
    }
        
    NFX_SPIN_UNLOCK(&fs_securityDesc.spin);
}


/**
 * Scan the cache for an entry matching the specified secId.
 * Return a pointer to it (pinned) if found.
 *
 * !!! Must be called with the cache spinlock held !!!
 */
static fs_SecurityDescCore *
fs_securityDescCacheScanSecId(
    fs_Context  *context,
    uint8       snapNum,
    uint32      hash,
    uint32      secId)
{
    fs_SecurityDescCore *sdc;

    ASSERT_LOCK_OWNED(&fs_securityDesc.spin);

    DLL_ITERATE(sdc, &fs_securityDesc.secIdHdr[hash], secIdEntry)
    {
        if (sdc->volId.whole_thing == context->volId.whole_thing &&
            sdc->snapNum == snapNum &&
            sdc->secId == secId)
        {
            //printf("sd cache scan found secId %d uid %d mode %o\n",
            //        secId, sdc->u.nfs.uid, sdc->u.nfs.mode);
            sdc->refCount += 1;
            if (sdc->refCount == 1)
            {
                //printf("sd rem free\n");
                int32 type = (sdc->attr & FS_SDESC_TYPE_MASK);
                FS_ASSERT(!DLL_IS_ISOLATED(sdc, freeEntry));
                DLL_REMOVE(&fs_securityDesc.freeHdr[type], sdc, freeEntry);
                fs_securityDesc.numFree[type] -= 1;
            }
            return sdc;
        }
    }

    return NULL;
}


/**
 * Scan the cache for an entry matching the specified nfs
 * descriptor. Return a pointer to it (pinned) if found.
 *
 * !!! Must be called with the cache spinlock held !!!
 */
static fs_SecurityDescCore *
fs_securityDescCacheScanNfs(
    fs_Context          *context,
    uint8               snapNum,
    fs_SecurityDescNfs  *nfs)
{
    fs_SecurityDescCore *sdc;
    int32               hash = fs_securityDescUidHash(nfs->uid);

    ASSERT_LOCK_OWNED(&fs_securityDesc.spin);

    DLL_ITERATE(sdc, &fs_securityDesc.uidHdr[hash], uidEntry)
    {
        if (sdc->u.hdr.cksum == nfs->hdr.cksum)
        {
            if (sdc->volId.whole_thing == context->volId.whole_thing &&
                sdc->snapNum == snapNum &&
                (sdc->u.hdr.attr & FS_SDESC_TYPE_MASK) == FS_SDESC_TYPE_NFS &&
                sdc->u.nfs.uid == nfs->uid &&
                sdc->u.nfs.gid == nfs->gid &&
                sdc->u.nfs.mode == nfs->mode)
            {
                //printf("sd cache scan found nfs uid %d mode %o secId %d\n",
                //  nfs->uid, nfs->mode, sdc->secId);
                sdc->refCount += 1;
                if (sdc->refCount == 1)
                {
                    //printf("sd rem free\n");
                    FS_ASSERT(!DLL_IS_ISOLATED(sdc, freeEntry));
                    DLL_REMOVE(&fs_securityDesc.freeHdr[FS_SDESC_TYPE_NFS],
                               sdc, freeEntry);
                    fs_securityDesc.numFree[FS_SDESC_TYPE_NFS] -= 1;
                }
                return sdc;
            }
        }
    }

    return NULL;
}


/**
 * Reclaim security descriptors from inodes on the
 * free list until the number of free descriptors
 * hits the free hwm.
 */
static void
fs_securityDescReclaim(
    fs_Context  *context,
    int32       type)
{
    fs_Inode        *inode;

    //NFX_SPIN_LOCK(&fs_securityDesc->spin);
    NFX_SPIN_LOCK(&fs_securityDesc.spin);
    fs_securityDesc.numReclaims[type] += 1;
    //NFX_SPIN_UNLOCK(&fs_securityDesc->spin);
    NFX_SPIN_UNLOCK(&fs_securityDesc.spin);

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    DLL_ITERATE(inode, &fs_inodeCache->mruHdr, mruEntry)
    {
        if (fs_securityDesc.numFree[type] >= FS_SDESC_NUM_FREE_HWM)
        {
            break;
        }

        if (inode->sd != NULL)
        {
            //printf("sd reclaim inum %lld\n", inode->inum);
            fs_securityDescRelease(context, inode->sd);
            inode->sd = NULL;
        }
    }
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}


/**
 * Insert the specified security descriptor into the cache
 * holding it pinned and return a pointer to the cached entry.
 * If a matching security descriptor is already in the cache
 * it will be pinned and its pointer will be returned.
 */
fs_SecurityDescCore *
fs_securityDescCacheInsert(
    fs_Context          *context,
    int32               snapNum,
    uint32              secId,
    fs_SecurityDescHdr  *insert)
{
    fs_SecurityDescCore *sdc;
    uint32              hash = fs_securityDescSecIdHash(secId);
    int32               type;

    // Checksum already generated by caller.
    FS_ASSERT(insert->cksum != 0);

    NFX_SPIN_LOCK(&fs_securityDesc.spin);

    // First scan for duplicate entry.  If found, return it.
    sdc = fs_securityDescCacheScanSecId(context, snapNum, hash, secId);
    if (sdc != NULL)
    {
        NFX_SPIN_UNLOCK(&fs_securityDesc.spin);
        return sdc;
    }

    // Entry not found, insert it.
    type = insert->attr & FS_SDESC_TYPE_MASK;
    if (type == FS_SDESC_TYPE_NFS)
    {
        sdc = fs_securityDescCacheInsertNfs(context, snapNum, secId,
                                            hash, (void *) insert);
    }
    else
    {
        // Not supported yet.
        FS_ASSERT(FALSE);
        //sdc = fs_securityDescInsertCifs(context, snapNum, secId, 
        //                                  hash, insert);
    }

    NFX_SPIN_UNLOCK(&fs_securityDesc.spin);

    // Take this opportunity to reclaim some cached
    // descriptors if the free list is low.
    if (fs_securityDesc.numUsed[type] >= fs_securityDesc.numUsedLimit[type] &&
        fs_securityDesc.numFree[type] < FS_SDESC_NUM_FREE_LWM)
    {
        fs_securityDescReclaim(context, type);
    }
    return sdc;
}


/**
 * Search the security file for the descriptor at the specified
 * secId.  If found, insert it in the cache and return a pointer
 * to it (pinned).
 */
static fs_SecurityDescCore *
fs_securityDescGetDisk(
    fs_Context  *context,
    uint8       snapNum,
    uint32      secId)
{
    fs_SecurityDescCore     *sdc = NULL;
    fs_Inode                *sdInode = I_GET_SHARED(snapNum, FS_SECURITY_INODE);
    fs_Buf                  *buf = O_GET_SHARED(sdInode, secId / FS_SDESC_ID_PER_PAGE);

    //printf("sd get disk secId %d\n", secId);
    if (buf == NULL)
    {
        //printf("no buf at page %d\n", secId / FS_BLOCK_SIZE);
        fs_inodeRelease(context, sdInode);
        return NULL;
    }

    if (secId >= fs_securityDesc.startId[FS_SDESC_TYPE_CIFS])
    {
        // Not supported yet.
        printf("CIFS security descriptor not supported yet\n");
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }
    else
    {
        fs_SecurityDescNfs  *nfs;
        int32 index;

        index = secId % FS_SDESC_ID_PER_PAGE;
        nfs = &buf->u.sdnfs->nfs[index];
        if ((nfs->hdr.attr & FS_SDESC_TYPE_MASK) == FS_SDESC_TYPE_NFS)
        {
            //printf("sd found secId %d in disk buf\n", secId);
            sdc = fs_securityDescCacheInsert(context, snapNum, secId,
                                                   &nfs->hdr);
        }
        else
        {
            //printf("sd failed to find secId %d in disk buf\n",
            //  secId);
        }
    }

    fs_bufRelease(context, buf);
    fs_inodeRelease(context, sdInode);

    return sdc;
}


/**
 * Fetch security descriptor for the specified secId.  A
 * reference count is maintained in the cache to ensure that
 * the pointer will remain valid.
 */
static fs_SecurityDescCore *
fs_securityDescGet(
    fs_Context  *context,
    uint8       snapNum,
    uint32      secId)
{
    fs_SecurityDescCore     *sdc;
    uint32                  hash = fs_securityDescSecIdHash(secId);

    //printf("sd desc get called snap %d secId %d\n", snapNum, secId);

    NFX_SPIN_LOCK(&fs_securityDesc.spin);
    sdc = fs_securityDescCacheScanSecId(context, snapNum, hash, secId);
    NFX_SPIN_UNLOCK(&fs_securityDesc.spin);

    if (sdc != NULL)
    {
        return sdc;
    }
    else
    {
        // Didn't find it in the cache.  Try the security file.
        return fs_securityDescGetDisk(context, snapNum, secId);
    }
}


/**
 * Search the security file for an NFS descriptor matching
 * the specified descriptor, instantiating one if its
 * not found.  Pin it in the cache and return the ptr.
 */
fs_SecurityDescCore *
fs_securityDescInsertFileNfs(
    fs_Context          *context,
    fs_Inode            *sdInode,
    fs_SecurityDescNfs  *nfsInsert)
{
    fs_Buf  *buf;
    uint32  startSecId = fs_securityDesc.startId[FS_SDESC_TYPE_NFS];
    uint32  page, insertPage = -1, insertIndex = -1, secId = -1;
    boolean found = FALSE, foundFreeSlot = FALSE;

    //printf("sd insert file nfs uid %d mode %o\n",
    //  nfsInsert->uid, nfsInsert->mode);

    FS_ASSERT(fs_lockOwnedExclusive(context, &sdInode->lock));

    nfsInsert->hdr.attr = FS_SDESC_TYPE_NFS;
    nfsInsert->hdr.len = sizeof(fs_SecurityDescNfs);
    nfsInsert->hdr.cont = 0;
    nfsInsert->hdr.cksum = 0;
    nfsInsert->hdr.uQtId = nfsInsert->hdr.gQtId = 0;
    nfsInsert->pad[0] = 0;
    nfsInsert->hdr.cksum = uint32_checksum(nfsInsert,
                                       sizeof(fs_SecurityDescNfs), 0);

    for (page = startSecId / FS_SDESC_ID_PER_PAGE; ; page++)
    {
        fs_SecurityDescNfs  *nfs;
        int32               i;

        // If we hit a hole in the file, the search is
        // over.
        buf = O_GET_EXCL(sdInode, page);
        if (buf == NULL)
        {
            //printf("sd insert file nfs hole in file\n");
            break;
        }

        for (i = 0; i < FS_SDESC_ID_PER_PAGE; i++)
        {
            nfs = &buf->u.sdnfs->nfs[i];
            if (nfs->hdr.attr != 0)
            {
                if (nfs->hdr.cksum == nfsInsert->hdr.cksum &&
                    nfs->uid == nfsInsert->uid &&
                    nfs->gid == nfsInsert->gid &&
                    nfs->mode == nfsInsert->mode)
                {
                    //printf("sd insert file found in file\n");
                    secId = startSecId + (page * FS_SDESC_ID_PER_PAGE) + i;
                    found = TRUE;
                    break;
                }
            }
            else if (!foundFreeSlot)
            {
                insertPage = page;
                insertIndex = i;
                foundFreeSlot = TRUE;
            }
        }
        if (found)
        {
            break;
        }

        fs_bufReleaseExclusive(context, buf);
    }

    if (!found)
    {
        if (foundFreeSlot)
        {
            FS_ASSERT(insertPage != -1);
            FS_ASSERT(insertIndex != -1);
            //printf("sd insert file free slot\n");
            buf = O_GET_EXCL(sdInode, insertPage);
        }
        else
        {
            //printf("sd insert file create new page\n");
            buf = O_CREATE(sdInode, page);
            bzero(buf->u.data, FS_BLOCK_SIZE);
            FS_BUF_ZERO_ALL(buf);
            insertPage = page;
            insertIndex = 0;

            if ((page + 1) * FS_BLOCK_SIZE > sdInode->dinode.size)
            {
                sdInode->dinode.size = (page + 1) * FS_BLOCK_SIZE;
            }
        }

        memcpy(&buf->u.sdnfs->nfs[insertIndex],
               nfsInsert, sizeof *nfsInsert);

        FS_BUF_WRITE(buf, insertIndex * sizeof *nfsInsert, sizeof *nfsInsert);

        // Write the inode too, to ensure lock order will be
        // maintained (otherwise we might release the inode but
        // not the buf).
        sdInode->dinode.mtime = fs_getTimeOfDay();
        fs_inodeWrite(context, sdInode, INO_WRITE_LOG);

        secId = startSecId + (insertPage * FS_SDESC_ID_PER_PAGE) + insertIndex;
        // printf("insert new sd buf write secId %d\n", secId);
    }
    else
    {
        fs_bufReleaseExclusive(context, buf);
    }

    FS_ASSERT(secId != -1);
    return fs_securityDescCacheInsert(context, sdInode->snapnum, secId,
                                      &nfsInsert->hdr);
}


/**
 * Set the security descriptor for the specified inode to
 * the specified descriptor.  Will create a new descriptor
 * on disk if necessary, update the in-core cache, and update
 * the inode.
 */
void
fs_securityDescSet(
    fs_Context          *context,
    fs_Inode            *inode,
    fs_SecurityDescHdr  *hdr)
{
    fs_Inode                *sdInode;
    fs_SecurityDescCore     *sdc = NULL;
    int32                   type = (hdr->attr & FS_SDESC_TYPE_MASK);

    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    //printf("sd set called\n");

    // Make sure the checksum is up to date.
    hdr->cksum = 0;
    hdr->cksum = uint32_checksum(hdr, hdr->len, 0);

    if (inode->sd)
    {
        //printf("sd set release existing sd\n");
        // Release the existing descriptor from the inode.
        fs_securityDescRelease(context, inode->sd);
        inode->sd = NULL;
    }

    sdInode = I_GET_EXCL(inode->snapnum, FS_SECURITY_INODE);
    NFX_SPIN_LOCK(&fs_securityDesc.spin);
    if (type == FS_SDESC_TYPE_NFS)
    {
        fs_SecurityDescNfs *nfs = (void *) hdr;

        sdc = fs_securityDescCacheScanNfs(context, inode->snapnum, nfs);
    }
    else
    {
        // Not supported yet.
        FS_ASSERT(FALSE);
    }
    NFX_SPIN_UNLOCK(&fs_securityDesc.spin);

    if (sdc == NULL)
    {
        //printf("sd set cache missed, checking file\n");
        // Didn't find it in the cache,
        // search the security file, instantiating
        // if necessary.
        if (type == FS_SDESC_TYPE_NFS)
        {
            sdc = fs_securityDescInsertFileNfs(context, sdInode, (void *) hdr);
        }
        else
        {
            // Not supported yet.
            FS_ASSERT(FALSE);
        }
    }

    if (!fs_inodeWasModified(context, sdInode))
    {
        fs_inodeReleaseExclusive(context, sdInode);
    }

    inode->sd = sdc;
    if (inode->dinode.securityId != sdc->secId)
    {
        inode->dinode.securityId = sdc->secId;
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
    }
}


/**
 * Find a free inode pointer, by walking backwards through the
 * most-recently-used list of inodes.  Returns NULL if all inodes
 * are still in use, else a free inode pointer NOT LOCKED.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
PRIVATE fs_Inode *
fs_inodeFindFree(
    fs_Context *context)
{
    fs_Inode *inode;

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    // Just pop the next one off the free list.
    if (!DLL_IS_EMPTY(&fs_inodeCache->mruHdr))
    {
        DLL_REMOVE_TAIL(&fs_inodeCache->mruHdr, inode, mruEntry);

        FS_ASSERT(!fs_lockHeldShare(context, &inode->lock));
        FS_ASSERT(!fs_lockHeldExclusive(context, &inode->lock));
        FS_ASSERT(fs_inodeIsAvail(inode));
        FS_ASSERT(inode->logSegDep.count == 0);
        FS_ASSERT((inode->flags & INO_FLAGS_RESTORE_IN_PROGRESS) == 0)

        inode->flags = 0 ;
        inode->zfflags = 0 ;
        inode->cacheflags = 0 ;
        inode->headLogSeg = -1;

        FS_ASSERT(inode->logSegDep.count == 0);

        if (inode->readAheadState != NULL)
        {
            FS_ASSERT(DLL_IS_EMPTY(
                           &inode->readAheadState->pendingReadList));
            fs_ramDealloc(FS_RAM_READ_AHEAD, inode->readAheadState);
            inode->readAheadState = NULL;
        }

        return inode;
    }
    
    return NULL;
}


/**
 * Hash for volId, inum, snapnum.
 */
static int32
fs_inodeCacheHash(
    evm_volId_t volId,
    uint64      inum,
    int32       snapnum)
{
#ifdef NOTDEF
    struct {
        evm_volId_t volId;
        uint64      inum;
        uint64      snapnum;
    } info;

    info.volId = volId;
    info.inum = inum;
    info.snapnum = snapnum;

    // this hash function is too damn expensive!
    //return fs_coreHash(&info, sizeof(info) / sizeof(uint64), 0) %
    //                            FS_ICACHE_HASH_MAX;
    return uint32_checksum(&info, sizeof(info), 0) % FS_ICACHE_HASH_MAX;
#else
    uint64 hash;

    hash = volId.details.u.parts.clusterUnique + (inum << 8) + snapnum;
    hash += (hash >> FS_ICACHE_HASH_SHIFT);
    return (hash % FS_ICACHE_HASH_MAX);

#endif
}


/**
 * Return TRUE if inode is COW and return buffer exclusive.
 */
PRIVATE fs_Buf*
fs_inodeIsCow(
    fs_Context      *context,
    fs_Inode        *inode)
{
    fs_Buf  *buf;
    boolean isCOW;

    // optimized: check if inode is still on COW list first
    // before going to buffer cache/disk to check for COW status

    if (DLL_IS_ISOLATED(inode, cowEntry))
    {
        //printf("inode is cow: ino %lld off cow list\n", inode->inum);
        return NULL;
    }

    if (context->fileSystem->state == FS_VOL_STATE_MOUNTING)
    {
        // Must be doing replay, implicitly non-cow.
        return NULL;
    }
        
    buf = fs_inodeGetBufferExclusive(context, inode->inum,
                                     inode->snapnum, FALSE, FALSE);
    FS_ASSERT(buf != NULL);
    isCOW = fs_bufIsCopyOnWrite(context, buf);
    if (isCOW)
        return buf;
    fs_bufReleaseExclusive(context, buf);
    return NULL;
}


/**
 * Add an inode to the list of inodes that might need COW.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
PRIVATE void
fs_inodeAddCow(
    fs_Context      *context,
    fs_Inode        *inode)
{
    int32           hash;
    uint64          offset;

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    // meta or no-snap inodes do not go on the cow list
    if ((inode->inum < FS_MAX_META_INODE) || fs_inodeIsNoSnap(context, inode))
        return;

    offset = ((inode->inum * sizeof (fs_Dinode)) / FS_BLOCK_SIZE) + 1;
    hash = offset % FS_ICACHE_COW_HASH_MAX;
    FS_ASSERT(DLL_IS_ISOLATED(inode, cowEntry));
    DLL_INSERT_HEAD(&fs_inodeCache->cowHdr[hash], inode, cowEntry);
    //printf("inode add cow: ino %lld %p offset %lld hash %d\n",
    //      inode->inum, inode, offset, hash);
}


/**
 * Remove all or just one inode from the list that might need COW.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
PRIVATE void
fs_inodeDeleteCow(
    fs_Context      *context,
    fs_Inode        *inode,
    boolean         all)
{
    fs_Inode        *i, *iNext;
    int32           hash;
    uint64          offset, inoOffset;

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    // meta or no-snap inodes are not inserted to the cow list
    if ((inode->inum < FS_MAX_META_INODE) || fs_inodeIsNoSnap(context, inode))
    {
        FS_ASSERT(DLL_IS_ISOLATED(inode, cowEntry));
        return;
    }

    offset = ((inode->inum * sizeof (fs_Dinode)) / FS_BLOCK_SIZE) + 1;
    hash = offset % FS_ICACHE_COW_HASH_MAX;

    if (all)
    {
        //printf("inode delete cow: for all offset %lld hash %d\n",
        //      offset, hash);
        DLL_ITERATE_NEXT(i, iNext, &fs_inodeCache->cowHdr[hash], cowEntry)
        {
            FS_ASSERT(!DLL_IS_ISOLATED(i, cowEntry));
            // account for collisions
            inoOffset = ((i->inum * sizeof (fs_Dinode)) / FS_BLOCK_SIZE) + 1;
            if (inoOffset == offset)
            {
                //printf("  ino %lld %p\n", i->inum, i);
                FS_ASSERT(!DLL_IS_ISOLATED(i, cowEntry));
                DLL_REMOVE(&fs_inodeCache->cowHdr[hash], i, cowEntry);
            }
        }
    }
    else
    {
        if (!DLL_IS_ISOLATED(inode, cowEntry))
        {
            DLL_REMOVE(&fs_inodeCache->cowHdr[hash], inode, cowEntry);
            //printf("inode delete cow: for ino %lld %p offset %lld hash %d\n",
            //      inode->inum, inode, offset, hash);
        }
    }
}


/**
 * Add all inodes back to the lists that might need COW.
 * Called after a snapshot operation (ie create).
 */
void
fs_inodeCacheEnableCow(
    fs_Context *context)
{
    fs_Inode *inode, *inodeNext;

    //printf("ino cache enable cow:\n");

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    DLL_ITERATE_NEXT(inode, inodeNext, &fs_inodeCache->mruHdr, mruEntry)
    {
        if ((inode->fileSystem == context->fileSystem) &&
             DLL_IS_ISOLATED(inode, cowEntry) &&
             ((inode->dinode.attrs & FS_IFLG_NO_SNAP) == 0) &&
             (inode->snapnum == 0))
        {
            //printf("  ino %p %lld snap %d (hash %d)\n",
            //      inode, inode->inum, inode->snapnum, hash);
            fs_inodeAddCow(context, inode);
        }
    }
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}


/**
 * Insert the inode into the cache.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
static void
fs_inodeCacheInsert(
    fs_Context  *context,
    fs_Inode    *inode)
{
    int32   hash;

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    hash = fs_inodeCacheHash(inode->fileSystem->volId, inode->inum, inode->snapnum);

    FS_ASSERT(DLL_IS_ISOLATED(inode, cacheEntry));

#if defined(FS_DEBUG_DUP_INODE_CACHE_INSERT)

#ifdef FS_INODE_CHECK_HASH_DEPTH
    fs_inodeCache->nbrLookups += 1;
#endif

    {
        fs_Inode *dbg, *dbgNext;
        DLL_ITERATE_NEXT(dbg, dbgNext,
                        &fs_inodeCache->cache[hash],
                        cacheEntry)
        {
#ifdef FS_INODE_CHECK_HASH_DEPTH
            fs_inodeCache->nbrScanned += 1;
#endif
            if (dbg->inum == inode->inum &&
                dbg->snapnum == inode->snapnum &&
                dbg->fileSystem == inode->fileSystem)
            {
                //fs_inodeCacheDump(context);
                fs_error(
                    context, FS_INTERNAL_ERR,
                    "duplicate inode cache insert ino %lld snap %d)\n",
                    inode->inum, inode->snapnum);
            }
        }
    }
#endif

    DLL_INSERT_HEAD(&fs_inodeCache->cache[hash], inode, cacheEntry);
#ifdef FS_INODE_CHECK_HASH_DEPTH
    fs_inodeCache->depth[hash]++;
#endif

    FS_ASSERT(DLL_IS_ISOLATED(inode, mruEntry));

    fs_inodeCache->count.numCached += 1;
    inode->fileSystem->inodeCacheCount.numCached += 1;

} /* fs_inodeCacheInsert */


/**
 * Insert the inode onto the mru list of inodes.
 *
 * returns TRUE is the inode is deallocated
 *
 * Must be called with the fs_inodeCache Spinlock held.
 *
 */
static void
fs_inodeCacheInsertMruOnly(
    fs_Context  *context,
    fs_Inode    *inode)
{
    FS_ASSERT(inode->flags == 0);
    FS_ASSERT(inode->zfflags == 0);
    FS_ASSERT(inode->cacheflags == 0);
    FS_ASSERT(DLL_IS_ISOLATED(inode, mrdEntry));
    FS_ASSERT(DLL_IS_ISOLATED(inode, cacheEntry));
    FS_ASSERT(DLL_IS_ISOLATED(inode, mruEntry));
    FS_ASSERT(DLL_IS_ISOLATED(inode, cowEntry));
    FS_ASSERT(inode->logRefCount == 0);
    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    if (fs_inodeCache->count.numCached >= fs_inodeCache->count.numCachedLimit)
    {
        // Cache is already full, just free this inode.
        if (inode->readAheadState != NULL)
        {
            FS_ASSERT(DLL_IS_EMPTY(
                           &inode->readAheadState->pendingReadList));
            fs_ramDealloc(FS_RAM_READ_AHEAD, inode->readAheadState);
            inode->readAheadState = NULL;
        }
        fs_ramDealloc(FS_RAM_INODE, inode);
        return;
    }

    DLL_INSERT_TAIL(&fs_inodeCache->mruHdr, inode, mruEntry);
    fs_inodeCache->count.numCached += 1;
    inode->fileSystem->inodeCacheCount.numCached += 1;
    
    return;

} /* fs_inodeCacheInsertMruOnly */


/**
 * Remove the inode from the cache.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
static void
fs_inodeCacheRemove(
    fs_Context  *context,
    fs_Inode    *inode)
{
    FS_ASSERT((inode->flags == 0 && inode->cacheflags == 0) ||
			  inode->flags == INO_FLAGS_RESTORE_IN_PROGRESS);
    FS_ASSERT(DLL_IS_ISOLATED(inode, mrdEntry));
    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    if (!DLL_IS_ISOLATED(inode, cacheEntry))
    {
        int32 hash = fs_inodeCacheHash(inode->fileSystem->volId, inode->inum,
                                       inode->snapnum);
        DLL_REMOVE(&fs_inodeCache->cache[hash], inode, cacheEntry);
#ifdef FS_INODE_CHECK_HASH_DEPTH
        fs_inodeCache->depth[hash]--;
#endif
    }

    if (!DLL_IS_ISOLATED(inode, mruEntry))
    {
        DLL_REMOVE(&fs_inodeCache->mruHdr, inode, mruEntry);
    }

    // remove from cow list
    fs_inodeDeleteCow(context, inode, FALSE);

    FS_ASSERT(fs_inodeCache->count.numCached > 0);
    FS_ASSERT(inode->fileSystem->inodeCacheCount.numCached > 0);
    fs_inodeCache->count.numCached -= 1;
    inode->fileSystem->inodeCacheCount.numCached -= 1;

    if (inode->sd != NULL)
    {
        fs_securityDescRelease(context, inode->sd);
        inode->sd = NULL;
    }

} /* fs_inodeCacheRemove */

/**
 * free resource for inode cache
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
void
fs_inodeFreeResource(
    fs_Context          *context,
    int32               reduceCount)
{
#ifdef FS_DYN_CACHE_SIZE
    fs_Inode    *freeInode;
    int32        count;

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    for (count = 0 ; count < reduceCount; count++)
    {
        if (fs_inodeCache->count.numCachedLimit <= fs_inodeCache->count.numCachedLimitLWM)
        {
            // don't drop below this level 
            if (count == 0)
                printf ("fs_inodeFreeResource don't drop below numCachedLimitLWM\n") ;
            return;
        }

        freeInode = fs_inodeFindFree(context);
        if (freeInode == NULL)
        {
            printf ("fs_inodeFreeResource could not fs_inodeFindFree\n") ;
            return;
        }
   
        fs_inodeCache->count.numCachedLimit--;

        // todo decr stats for this type
   
        fs_inodeCacheRemove(context, freeInode);

        FS_ASSERT(freeInode->flags == 0);
        FS_ASSERT(freeInode->cacheflags == 0);
        FS_ASSERT(DLL_IS_ISOLATED(freeInode, mrdEntry));
        FS_ASSERT(DLL_IS_ISOLATED(freeInode, cacheEntry));
        FS_ASSERT(DLL_IS_ISOLATED(freeInode, mruEntry));
        FS_ASSERT(DLL_IS_ISOLATED(freeInode, cowEntry));
        FS_ASSERT(freeInode->logRefCount == 0);

        fs_ramDealloc(FS_RAM_INODE, freeInode);
    }
#endif // FS_DYN_CACHE_SIZE
}

/**
 * Allocate a new in-core inode or reclaim the LRU one.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 *
 * Returns inode with EXCL lock held.
 */
static fs_Inode *fs_inodeAllocIncore(
    fs_Context 	*context,
	uint64		inum)
{
    fs_Inode    *inode;
#ifdef FS_DYN_CACHE_SIZE
	boolean      resourceAvail = TRUE;
#endif

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

#ifdef FS_DYN_CACHE_SIZE
    if (((uint32)(eee_ramTotalFlexAvail() /1024)) < fs_incoreTune->memMinAvailable)
    {
		// the resource usage is over the limit, , 
		// knock some out
		// reduce the used limit,
        fs_bufFindAndFreeResource (context, FALSE, FALSE, TRUE, fs_incoreTune->bufFreeCount);

        resourceAvail = FALSE;
    }

	if (resourceAvail)
    {
        if (fs_inodeCache->count.numCached >= fs_inodeCache->count.numCachedLimit)
        {
        	// OK, increase the used buf limit.
            fs_inodeCache->count.numCachedLimit = 
            	MIN((fs_inodeCache->count.numCached + 1),
					 fs_inodeCache->count.numCachedLimitHWM);

			// todo incr stats for this type
        }
    }
    else
    {
    	if (fs_inodeCache->count.numCached < fs_inodeCache->count.numCachedLimit)
        {
        	// OK, decrease the used inode limit.
            fs_inodeCache->count.numCachedLimit =
            	MAX(fs_inodeCache->count.numCached,
                    fs_inodeCache->count.numCachedLimitLWM);

			// todo incr stats for this type
        }
    }    
#endif // FS_DYN_CACHE_SIZE

    if (fs_inodeCache->count.numCached < fs_inodeCache->count.numCachedLimit)
    {
        inode = fs_ramAlloc(FS_RAM_INODE, sizeof(fs_Inode));
        bzero(inode, sizeof(fs_Inode));
		inode->inum = inum; // for lockorder checking
        inode->headLogSeg = -1;
        fs_lockInit(&inode->lock, FS_LOCK_TYPE_INODE);
        fs_lockInit(&inode->btreeLock, FS_LOCK_TYPE_BTREE);
        fs_lockExclusive(context, &inode->lock, NULL);
	    NFX_SPIN_INIT(&inode->spin,SPINLOCK_LEVEL_UNKNOWN);
        // insert to cow list
        fs_inodeAddCow(context, inode);
        return inode;
    }
    else
    {
        boolean acquired;

        // This cache has reached its limit, knock something
        // out to make room for the new inode.

        inode = fs_inodeFindFree(context);
        if (inode == NULL)
        {
            // All cached inodes are in use.  The caller
            // must handle this case.
            return NULL;
        }

        fs_inodeCacheRemove(context, inode);

        inode->inum = inum; // for lockorder checking

        // insert to cow list
        fs_inodeAddCow(context, inode);

        // We should not be blocked getting the lock since
        // we hold the cache spinlock and the inode is supposed
        // to be free.
        acquired = fs_trylockExclusive(context, &inode->lock);
        FS_ASSERT(acquired);

        return inode;
    }
}


/**
 * Print all the locked, in-cache inodes.  Best to call with the inode cache
 * spin lock held.
 */
void fs_inodeShowLocked()
{
    int32           hash;

    for (hash = 0; hash < FS_ICACHE_HASH_MAX; hash++)
    {
        fs_Inode *inode;
        FS_LOOP_DBG_DECL(dbgNumLoops);

        DLL_ITERATE(inode,
                       &fs_inodeCache->cache[hash],
                       cacheEntry)
        {
            FS_LOOP_DBG_CHECK(dbgNumLoops, fs_inodeCache->count.numCached + LOOP_DBG_FUDGE);
            if (inode->lock.mode != FS_LOCK_MODE_AVAIL) {
                printf("inode at %p num %lld lock mode %d owner %llx\n",
                       inode, inode->inum, inode->lock.mode,
                       (uint64)((uintptr_t)inode->lock.owner));
            }
        }
    }
}


/**
 * Return a buffer from the inode file, with the specified lock mode,
 * given the inode number.
 */
static fs_Buf *fs_inodeGetBuffer(
    fs_Context *context,
    fs_Inumber inum,
    int snapnum,
    boolean isCreate,
    boolean isExclusive,
    boolean isTryLock)
{
    if (inum < FS_MAX_META_INODE)
	{
        fs_Block block = fs_firstInodeBlock(snapnum);
        fs_Buf *buf;
        assert(!isTryLock);
        if (isExclusive)
            buf = B_GET_EXCL(NULL, block);
        else
            buf = B_GET_SHARED(NULL, block);

        assert(buf != NULL);
        buf->page = 0;
        buf->inum = FS_INODE_INODE;
        return buf;
    }
    
    else {
        fs_Parameters *params = &context->fileSystem->params;
        uint64 page = (inum * sizeof (fs_Dinode)) / params->blockSize;
        fs_Inode *inodeFile = I_GET_SHARED(snapnum, FS_INODE_INODE);
        fs_Buf *buf;

        if (inodeFile == NULL)
        {
            FS_ASSERT(snapnum != 0);
            // Snapshot doesn't exist.
            return NULL;
        }


getBufAgain:

        // to avoid lock order problems use a different lock
        // instead of the btree lock (PR #1582)
        if (isCreate)
            fs_lockExclusive(context, &context->fileSystem->inodeLock, NULL);
        else
            fs_lockShared(context, &context->fileSystem->inodeLock, NULL);

        buf = fs_bufGetPage(context,
                            inodeFile,
                            page,
                            (isExclusive
                             ? BUF_GET_MODE_EXCL
                             : BUF_GET_MODE_SHARED),
                            BUF_GET_OPT_SLEEP
                            | (isTryLock ? BUF_GET_OPT_TRY_LOCK : 0),
                            FS_BLOCK_SIZE);

        if (buf == FS_BUF_TRY_LOCK_FAILED) {
            if (isCreate)
                fs_unlockExclusive(context, &context->fileSystem->inodeLock);
            else
                fs_unlockShared(context, &context->fileSystem->inodeLock);
            fs_inodeRelease(context, inodeFile);
            return NULL;
        }
        
        if (buf == NULL && isCreate)
		{
			FS_ASSERT(isExclusive); // required for creating inodes

            buf = O_GET_EXCL(inodeFile, page);
            if (buf != NULL) {
                if (isCreate)
                    fs_unlockExclusive(context, &context->fileSystem->inodeLock);
                else
                    fs_unlockShared(context, &context->fileSystem->inodeLock);
            }
            else {
                // Do this as a nested transaction to avoid
                // lock order problems (bugid 1112).

                fs_logNestTransaction(context);

                // Size and numblocks for the inode file are
                // maintained in the superblock to avoid lock
                // order problems on the inode file inode.  This
                // happens rare enough to not pose a concurrency
                // problem.  Note that this means that we only have
                // this information for the live inode file, not for
                // snapshots.
                context->fileSystem->super->inodeFileBlks += 1;
                if ((page + 1) * FS_BLOCK_SIZE >
                        context->fileSystem->super->inodeFileSize)
                {
                    context->fileSystem->super->inodeFileSize =
                                    (page + 1) * FS_BLOCK_SIZE;
                }
                fs_writeSuperblocksTrans(context, context->fileSystem->super);

                buf = O_CREATE(inodeFile, page);
                assert(buf != NULL);

                if (isCreate)
                    fs_unlockExclusive(context, &context->fileSystem->inodeLock);
                else
                    fs_unlockShared(context, &context->fileSystem->inodeLock);

                // zero the inode page
                bzero(buf->u.rawdata, context->fileSystem->params.blockSize);
                FS_BUF_ZERO_ALL(buf);

                // Commit the nested transaction.
                fs_logCommitEntry(context);

                // Calculate the min free blocks for current
                // conditions.
                fs_recalcMinFreeBlks(context);
                
                // Since commiting unlocked buf, we need to re-acquire it.
                goto getBufAgain;
            }
        }
        else
        {
            if (isCreate)
                fs_unlockExclusive(context, &context->fileSystem->inodeLock);
            else
                fs_unlockShared(context, &context->fileSystem->inodeLock);
        }
        

        fs_inodeRelease(context, inodeFile);
        return buf;
    }
}


/**
 * Return a buffer from the inode file, with an exclusive lock, given the
 * inode number.
 */
fs_Buf *fs_inodeGetBufferExclusive(
    fs_Context *context,
    fs_Inumber inum,
    int snapnum,
    boolean create,
    boolean tryLock)
{
    return fs_inodeGetBuffer(context, inum, snapnum, create, TRUE, tryLock);
}


/**
 * Return a buffer from the inode file, with a shared lock, given the
 * inode number.
 */
fs_Buf *fs_inodeGetBufferShared(
    fs_Context *context,
    fs_Inumber inum,
    int snapnum,
    boolean create,
    boolean tryLock)
{
    return fs_inodeGetBuffer(context, inum, snapnum, create, FALSE, tryLock);
}
    

/**
 * Brute-force recount of all inodes -- used after a system crash.
 */
void fs_inodeCountAllocated(
    fs_Context *context
)
{
    fs_Inode *inodeFile = I_GET_SHARED(0, FS_INODE_INODE);
    fs_Parameters *params = &context->fileSystem->params;
    uint64 pageno, max_pageno;
    uint64  totInodesUsed = 0;

    max_pageno = ceilingDiv64(params->maxInodeNumber, FS_INODES_PER_PAGE);
    
    for (pageno = 1; pageno < max_pageno; pageno++) {
        fs_Buf *buf = O_GET_EXCL(inodeFile, pageno);
        int inodesUsed = 0;
        int i;

        if (buf == NULL)
            continue; // hole in the file?
        
        for (i = 0; i < FS_INODES_PER_PAGE; i++)
            inodesUsed += (buf->u.dinode->dinode[i].attrs != 0);
        
        totInodesUsed += inodesUsed;

        fs_bufReleaseExclusive(context, buf);
    }

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    //if (totInodesUsed != context->fileSystem->super->inodesUsed)
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("fs_inodeCountThread recalc %lld (calc %lld) max %lld - %lld%%.",
                 context->fileSystem->super->inodesUsed,
                 totInodesUsed, 
                 context->fileSystem->params.maxInodeNumber,
                 100LL * context->fileSystem->super->inodesUsed / 
                   context->fileSystem->params.maxInodeNumber));
    }
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    fs_inodeRelease(context, inodeFile);
}


/**
 * Thread for re-counting inode usage after a crash.
 */
void
fs_inodeCountThread(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;

    //FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
    //        ("fs_inodeCountThread: inode count recalc thread started."));

    // PR #1789: acquire global access lock
    fs_lockShared(context, &context->fileSystem->accessLock, NULL);

    fs_inodeCountAllocated(context);

    // release global access lock
    fs_unlockShared(context, &context->fileSystem->accessLock);

    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("fs_inodeCountThread exiting (Used %lld total %lld - %lld%%).",
             context->fileSystem->super->inodesUsed, 
             context->fileSystem->params.maxInodeNumber,
             100LL * context->fileSystem->super->inodesUsed / context->fileSystem->params.maxInodeNumber));

    fs_threadTerminate(context);
}


/**
 * Lookup an inode in cache by inode number. Return NULL if the inode
 * was not in the cache.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
static fs_Inode *fs_inodeCacheLookup(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum
    )
{
    fs_Inode *inode;
    int32       hash = fs_inodeCacheHash(context->volId, inum, snapnum);
    fs_FileSystem *fileSystem = context->fileSystem;

    FS_LOOP_DBG_DECL(dbgNumLoops);

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    DLL_ITERATE(inode, &fs_inodeCache->cache[hash], cacheEntry)
    {
        FS_LOOP_DBG_CHECK(dbgNumLoops,
                          fs_inodeCache->count.numCached + LOOP_DBG_FUDGE);
        if (inode->inum == inum &&
            inode->snapnum == snapnum &&
            inode->fileSystem == fileSystem)
        {
            return inode;
        }
    }

    return NULL;
}
    

/**
 * Lookup an inode in the inode cache by inode number.  Returns NULL if
 * the inode was not in the cache, INO_NO_LONGER_VALID if the inode
 * is pending deallocation or was destroyed while we were waiting for
 * the lock, or a valid, locked inode otherwise.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 *
 * If buf is non-NULL, it is assumed to be held shared and
 * it will be released by this function if it returns a non-NULL
 * status.
 */

#define INO_NO_LONGER_VALID     1

static fs_Inode *fs_inodeCacheGet(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum,
    int32           lockFlags)
{
    fs_Inode    *inode;
    int32       hash = fs_inodeCacheHash(context->volId, inum, snapnum);
    fs_FileSystem *fileSystem = context->fileSystem;

    FS_LOOP_DBG_DECL(dbgNumLoops);

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

#ifdef FS_INODE_CHECK_HASH_DEPTH
    fs_inodeCache->nbrLookups += 1;
#endif

    DLL_ITERATE(inode, &fs_inodeCache->cache[hash], cacheEntry)
    {
#ifdef FS_INODE_CHECK_HASH_DEPTH
        fs_inodeCache->nbrScanned += 1;
#endif
        FS_LOOP_DBG_CHECK(dbgNumLoops, fs_inodeCache->count.numCached + LOOP_DBG_FUDGE);
        if (inode->inum == inum &&
            inode->snapnum == snapnum &&
            inode->fileSystem == fileSystem)
        {
            if (!(inode->flags & INO_FLAGS_PENDING_DEALLOC)
                && fs_lockmgr(context, &inode->lock, &fs_inodeCache->spin,
                              lockFlags))
            {
                if (!DLL_IS_ISOLATED(inode, mruEntry))
                {
                    // Remove the inode from the free list.
                    DLL_REMOVE(&fs_inodeCache->mruHdr, inode, mruEntry);
                }
                return inode;
            }
            else
            {
                // Check for destruction in lock return.  Higher layer
                // should treat it as a lookup failure.  This could happen
                // if the inode was deleted while we were waiting our turn.
                return (fs_Inode *) INO_NO_LONGER_VALID;
            }
        }
    }

    return NULL;

} /* fs_inodeCacheGet */


/**
 * Signal the inode cleaner threads to wakeup and clean
 * inodes.
 */
static void
fs_inodeSignalCleaners(
    fs_Context      *context)
{
    esm_event_t     event;
    int32           i, rc;

    event.event = FS_THREAD_EVENT_START;
    event.p1.u.w = 0;
    event.p2.u.w = FS_INODE_CLEANER_ALL;  // debug

    for (i = 0; i < FS_FILE_SYSTEM_HASH_MAX; i++)
    {
        fs_FileSystem   *fs;

        NFX_SPIN_LOCK(&fs_fileSystemTableLock);
        for (fs = fs_fileSystems[i]; fs != NULL; fs = fs->next)
        {
            if (fs->state == FS_VOL_STATE_ONLINE)
            {
                rc = fs_daemonWakeup(context, &fs->spin,
                                     fs->inodeCleanerHandle,
                                     &fs->inodeCleanerSignal, event);

                INO_TRACE(("vol 0x%llx cleaner signalled, daemon %s awakened",
                    fs->volId.whole_thing, rc == TRUE ? "was" : "was not"));
            }
        }
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
    }

} /* fs_inodeSignalCleaners */


uint32 fs_inodeAllocRetries = 0;

/**
 * Put the caller to sleep for a bit to wait for
 * in-core inodes to come available.
 */
static void
fs_inodeWaitAvail(
    fs_Context  *context,
    int32       *numRetries)
{
    if (!context->isThread)
    {
        printf("inode cache 100%% busy for non-thread\n");
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }
    else if (++*numRetries > 1000)
    {
        printf("inode cache 100%% busy, retries exhausted\n"); 
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }
    else
    {
        fs_inodeAllocRetries += 1;
        fs_sleep(context, 10);
    }
}


/**
 * Instantiate the newly allocated inode into the
 * cache.
 */
fs_Inode *
fs_inodeInstantiateAllocator(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum,
	fs_Buf			*buf)
{
    fs_FileSystem   *fileSystem;
    fs_Inode        *inode;
    uint32          numRetries = 0;

    FS_TRACE(context, FS_TRACE_INO,
             ("alloc inst %lld snap %d buf %p",
              inum, snapnum, buf));

	FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));

    fileSystem = context->fileSystem;

retry:

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
#ifdef FS_DEBUG_NEW_INODE_INSERT
    inode = fs_inodeCacheLookup(context, snapnum, inum);
    FS_ASSERT(inode == NULL);
#endif

    // Allocate/recycle an in-core inode.
    inode = fs_inodeAllocIncore(context, inum);
    if (inode == NULL)
    {
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        fs_inodeWaitAvail(context, &numRetries);
        goto retry;
    }
    inode->snapnum = snapnum;
    inode->fileSystem = fileSystem;
    inode->flags = 0;
    inode->zfflags = 0;
    inode->cacheflags = 0;

    // We have EXCLUSIVE lock on the inode.
    // Add it to the cache.
    fs_inodeCacheInsert(context, inode);

    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

    // Take this opportunity to wakeup the cleaner daemons
    // if necessary.
    if (fs_inodePercentAvail() < fs_incoreTune->inodePctAvailLoWtrMk)
    {
        fs_inodeSignalCleaners(context);
    }

    // Copy the dinode.
    memcpy(&inode->dinode,
           &buf->u.dinode->dinode[inum % FS_DINODE_PAGE_MAX],
           sizeof(fs_Dinode));

    FS_TRACE(context, FS_TRACE_INO, ("  got inum %lld snap %d",
             inum, snapnum));

    return inode;

} /* fs_inodeInstantiateAllocator */


static int32 fs_lostSecurityDesc = 0;

/**
 * Get an existing inode from the cache or disk, instantiate
 * in the cache as needed.
 */
fs_Inode *
fs_inodeGet(
	fs_Context      *context,
    int32           snapnum,
    uint64          inum,
	int32			lockMode)
{
    fs_FileSystem   *fileSystem;
    fs_Buf          *buf;
    fs_Inode        *inode;
    uint32          numRetries = 0;

    FS_TRACE(context, FS_TRACE_INO,
             ("iget %lld snap %d lockMode %d",
              inum, snapnum,
              lockMode));

    fileSystem = context->fileSystem;

retry:

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    inode = fs_inodeCacheGet(context, snapnum, inum, lockMode);
    if (inode == NULL)
    {
        FS_TRACE(context, FS_TRACE_INO, ("  i_miss %lld snap %d",
                 inum, snapnum));

        // We missed in the cache, try to read the inode from disk.

        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

        buf = fs_inodeGetBufferShared(context, inum, snapnum, FALSE, FALSE);
        //printf("~~~ fs_inodeGetBuffer(%lld, %d) ->"
        //       " buf %p block %lld inode %lld page %lld\n", inum, snapnum,
        //       buf, fs_blockNumber(buf->block), buf->inum, buf->page);

        if (buf == NULL ||
            buf->u.dinode->dinode[inum % FS_DINODE_PAGE_MAX].attrs == 0)
        {
            // Lookup failure (inode is not allocated).

			FS_TRACE(context, FS_TRACE_INO,
                    ("iget %lld snap %d lookup returned %s",
                     inum, snapnum, buf == NULL ? "NULL buf" :
                    "unallocated inode"));
 
            if (buf != NULL)
                fs_bufRelease(context, buf);

            return NULL; 
		}

        // We have the dinode buffer and its okay for us to
        // proceed to instantiate.  However, in case we raced
        // with someone, we need to check again if its already
        // in the cache.  If its already in the cache, unlock the buf
        // (since we don't want to block waiting for the inode lock
        // while holding the buf, otherwise we could deadlock with
        // inode deallocation).

        NFX_SPIN_LOCK(&fs_inodeCache->spin);
        inode = fs_inodeCacheLookup(context, snapnum, inum);
        if (inode != NULL)
        {
            fs_bufRelease(context, buf);
            if (!(inode->flags & INO_FLAGS_PENDING_DEALLOC)
                && (fs_lockmgr(context, &inode->lock, &fs_inodeCache->spin,
                               lockMode)))
            {
                if (!DLL_IS_ISOLATED(inode, mruEntry))
                {
                    DLL_REMOVE(&fs_inodeCache->mruHdr, inode, mruEntry);
                }
                FS_TRACE(context, FS_TRACE_INO,
                    ("  got inum %lld snap %d on refetch",
                     inum, snapnum));
            }
            else
            {
                FS_TRACE(context, FS_TRACE_INO,
                    ("  got invalid inum %lld snap %d on refetch",
                     inum, snapnum));
                inode = NULL;
            }
            
            // buf already released

            NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

            goto cacheGetOut;
        }

        // Allocate/recycle an in-core inode.
        inode = fs_inodeAllocIncore(context, inum);
        if (inode == NULL)
        {
            NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
            fs_bufRelease(context, buf);
            fs_inodeWaitAvail(context, &numRetries);
            goto retry;
        }
        inode->snapnum = snapnum;
        inode->fileSystem = fileSystem;
        inode->flags = 0;
        inode->zfflags = 0;
        inode->cacheflags = 0;

        // We have EXCLUSIVE lock on the inode.
        // Add it to the cache.
        fs_inodeCacheInsert(context, inode);

        fs_inodeCache->count.missCount += 1;
        inode->fileSystem->inodeCacheCount.missCount += 1;
#ifdef FS_DEBUG_INODE_CNT_MISS_TYPE
        if (inum < FS_MAX_META_INODE)
        {
            fs_inodeCache->count.missCntType[inum] += 1;
            inode->fileSystem->inodeCacheCount.missCntType[inum] += 1;
        }
#endif
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

        // Take this opportunity to wakeup the cleaner daemons
        // if necessary.
        if (fs_inodePercentAvail() < fs_incoreTune->inodePctAvailLoWtrMk)
        {
            fs_inodeSignalCleaners(context);
        }

        // At this point, the inode is locked and in the cache.  The cache
        // spinlock is released.  Update the inode->dinode from the buf.

        memcpy(&inode->dinode,
               &buf->u.dinode->dinode[inum % FS_DINODE_PAGE_MAX],
               sizeof(fs_Dinode));

		// Done with buf, release it.
        fs_bufRelease(context, buf);

        if (lockMode & FS_LOCK_MODE_SHARE)
        {
            // Demote the lock.
            fs_lockDemote(context, &inode->lock, FS_LOCK_MODE_SHARE);
        }
    }
    else if (inode == (fs_Inode *) INO_NO_LONGER_VALID)
    {
        // Destroyed while we were waiting for it.
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        return NULL;
    }
    else
    {
        // Hit in the cache.

        fs_inodeCache->count.hitCount += 1;
        inode->fileSystem->inodeCacheCount.hitCount += 1;
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
    }

    FS_TRACE(context, FS_TRACE_INO, ("  got inum %lld snap %d",
             inum, snapnum));

cacheGetOut:

    if (inode->sd == NULL &&
        inode->inum != FS_SECURITY_INODE &&
        inode->inum != FS_SNAPSHOT_BITS_INODE)
    {
        inode->sd = fs_securityDescGet(context,
                                       snapnum,
                                       inode->dinode.securityId);
        if (inode->sd == NULL)
        {
            // Failed to find the security descriptor!  This
            // is a bug, but as a workaround, default the inode
            // to root's security descriptor.


            printf("failed to find securityId for volId %lld, inum %lld, offset %d\n",
                   context->volId.whole_thing, inode->inum,
                   inode->dinode.securityId);

            fs_lostSecurityDesc += 1;
            inode->sd = fs_securityDescGet(context,
                                           snapnum,
                                           0);
        }
        FS_ASSERT(inode->sd != NULL);
    }

#ifdef NEW_ZF
    if (!DLL_IS_EMPTY(&(fileSystem->zfInodeHdr)))
    {
        NFX_SPIN_LOCK(&(fileSystem->zfSpin));
        if (fs_zfGetInode(context, inode->inum) != NULL)
        {
            //printf("inum %10lld is part of zero fill list\n", inode->inum);
            inode->zfflags |= INO_FLAGS_UZERO;
        }
        else
        {
            //printf("inum %10lld is not part of zero fill list\n", inode->inum);
            inode->zfflags &= ~INO_FLAGS_UZERO;
        }
        NFX_SPIN_UNLOCK(&(fileSystem->zfSpin));
    }
#endif // NEW_ZF
    
    return inode;

} /* fs_inodeGet */


/**
 * Fetch the attributes for the specified inode if they're
 * in the cache.  Return TRUE if they were fetched, FALSE
 * otherwise.
 *
 * !!! This does not block. !!!
 */
boolean
fs_inodeGetAttrs(
    fs_FileSystem   *fileSystem,
    fs_Attr         *attrp,
    uint64          inum,
    int32           snapnum,
    fs_Proto        protocol,
    fs_AttrType     attrType,
    uint32          *genNum)
{
    fs_Inode        *inode;
    int32           hash = fs_inodeCacheHash(fileSystem->volId, inum, snapnum);

    FS_TRACE(NULL, FS_TRACE_INO,
             ("igetattr %lld snap %d",
              inum, snapnum));

    NFX_SPIN_LOCK(&fs_inodeCache->spin);

    DLL_ITERATE(inode, &fs_inodeCache->cache[hash], cacheEntry)
    {
        if (inode->inum == inum &&
            inode->snapnum == snapnum &&
            inode->fileSystem == fileSystem)
        {
            if (inode->sd == NULL)
            {
                // sd is not yet initiliased, return as not found
                NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
                return FALSE;
            }
            
            fs_doGetAttr(fileSystem->volId, attrp, inode, 0, 0, protocol, attrType);
            if (genNum != NULL)
            {
                *genNum = inode->dinode.genNum;
            }
            NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
            return TRUE;
        }
    }

    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
    return FALSE;

} /* fs_inodeGetAttrs */


/**
 * Instantiate a meta inode in the cache.
 */ 
fs_Inode *
fs_inodeInstantiateMetanode(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum)
{
    fs_FileSystem   *fileSystem;
    fs_Inode        *inode;
    uint32          numRetries = 0;

    FS_TRACE(context, FS_TRACE_INO,
             ("meta inst %lld snap %d",
              inum, snapnum));

	FS_ASSERT(inum < FS_MAX_META_INODE);

    fileSystem = context->fileSystem;

retry:

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
#ifdef FS_DEBUG_NEW_INODE_INSERT
    inode = fs_inodeCacheLookup(context, snapnum, inum);
    FS_ASSERT(inode == NULL);
#endif

    inode = fs_inodeAllocIncore(context, inum);
    if (inode == NULL)
    {
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        fs_inodeWaitAvail(context, &numRetries);
        goto retry;
    }
    inode->snapnum = snapnum;
    inode->fileSystem = fileSystem;
    inode->flags = 0;
    inode->zfflags = 0;
    inode->cacheflags = 0;

    // We have EXCLUSIVE lock on the inode.
    // Add it to the cache.
    fs_inodeCacheInsert(context, inode);
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

    // Clean up the newly allocated inode.
    bzero(&inode->dinode, sizeof(fs_Dinode));

	return inode;

} /* fs_inodeInstantiateMetanode */


/**
 * Add the specified inode to the free list
 * if its available for reclamation.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
static void
fs_inodeMaybeAddToFreeList(
    fs_Context  *context,
    fs_Inode    *inode)
{
    FS_ASSERT(DLL_IS_ISOLATED(inode, mruEntry));
    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    if (!fs_lockHeldShare(context, &inode->lock) &&
        !fs_lockHeldExclusive(context, &inode->lock) &&
        fs_inodeIsAvail(inode))
    {
        // Inode is free for reclamation, add it to
        // the MRU list.
        DLL_INSERT_HEAD(&fs_inodeCache->mruHdr, inode, mruEntry);
    }
}


/**
 * Called to unlock and release an inode, held with a shared lock, back
 * to the cache.
 */
extern void fs_inodeRelease(
    fs_Context      *context,
    fs_Inode        *inode)
{
    FS_TRACE(context, FS_TRACE_INO,
             ("irelease %lld snap %d",
             inode->inum, inode->snapnum));

    FS_ASSERT(fs_lockHeldShare(context, &inode->lock) ||
              fs_lockOwnedExclusive(context, &inode->lock));

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    fs_unlockShared(context, &inode->lock);
    fs_inodeMaybeAddToFreeList(context, inode);
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}


/**
 * Called to unlock and release an inode, held with a exclusive lock,
 * back to the cache.
 *
 * Should normally only be used if inode was NOT modified.  If inode was
 * modified, fs_inodeWrite() should be used (eventually followed by
 * fs_logCommitEntry()).
 */
extern void fs_inodeReleaseExclusive(
    fs_Context      *context,
    fs_Inode        *inode)
{
    FS_TRACE(context, FS_TRACE_INO,
             ("irelease exclusive %lld snap %d",
             inode->inum, inode->snapnum));

    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    fs_unlockExclusive(context, &inode->lock);
    fs_inodeMaybeAddToFreeList(context, inode);
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}

/**
 * Called to unlock and release an inode, held with a exclusive lock,
 * back to the cache. Release all the exclusive holds as the caller
 * knows that it does not need the object again and there aren't
 * any pointers to this inode
 *
 * Should normally only be used if inode was NOT modified.  If inode was
 * modified, fs_inodeWrite() should be used (eventually followed by
 * fs_logCommitEntry()).
 */
extern void fs_inodeReleaseExclusiveAll(
    fs_Context      *context,
    fs_Inode        *inode)
{
    FS_TRACE(context, FS_TRACE_INO,
             ("irelease exclusive %lld snap %d",
             inode->inum, inode->snapnum));

    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    fs_unlockExclusiveAll(context, &inode->lock);
    fs_inodeMaybeAddToFreeList(context, inode);
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}


/**
 * Mark the specified as dirty and add it to the dirty
 * list if its not already.
 * 
 * Must be called with the inode->spin lock held.
 */
static void
fs_inodeMarkDirty(
	fs_Context	*context,
	fs_Inode	*inode)
{

    ASSERT_LOCK_OWNED(&inode->spin);

    // Add the inode to the dirty list if it isn't
    // already.
    if (!(inode->flags & INO_FLAGS_DIRTY))
    {
        FS_ASSERT(DLL_IS_ISOLATED(inode, mrdEntry));

        NFX_SPIN_LOCK(&fs_inodeCache->spin);
        DLL_INSERT_HEAD(&inode->fileSystem->inodeMrdHdr, inode, mrdEntry);
        fs_inodeCache->count.numDirty += 1;
        inode->fileSystem->inodeCacheCount.numDirty += 1;
        inode->cacheflags |= INO_FLAGS_MRD;
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        inode->flags |= INO_FLAGS_DIRTY;
    }
}


/**
 * Set the inode onto the dirty list for the specified
 * log segment.
 */
void
fs_inodeSetDirtySegment(
    fs_Context  *context,
    void        *obj,
    int32       segNum,
    int32       offset)
{
    fs_Inode    *inode = obj;

    FS_ASSERT(inode->flags & INO_FLAGS_DIRTY);
    FS_ASSERT(segNum < FS_LOG_MAX_SEGMENTS);
    FS_ASSERT(inode->headLogSeg == -1 ||
              fs_logSegIsOlder(context->fileSystem->logRoot, segNum, inode->headLogSeg));

    //printf("4: offset %d segNum %d\n", offset, segNum);

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    if (DLL_IS_ISOLATED(inode, mrdEntry))
    {
        //printf("2: offset %d segNum %d\n", offset, segNum);
        // Writeback already in progress on this
        // inode, don't put it back on the list.
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        return;
    }

    if (inode->cacheflags & INO_FLAGS_MRD)
    {
        DLL_REMOVE(&inode->fileSystem->inodeMrdHdr, inode, mrdEntry);
        inode->cacheflags &= ~INO_FLAGS_MRD;
    }
    else if (!DLL_IS_ISOLATED(inode, mrdEntry))
    {
        DLL_REMOVE(&inode->fileSystem->inodeLogSegmentHdr[inode->headLogSeg],
                   inode, mrdEntry);
    }
    DLL_INSERT_HEAD(&inode->fileSystem->inodeLogSegmentHdr[segNum],
                    inode, mrdEntry);
    inode->headLogSeg = segNum;
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}


/*-----------------------------------------------------------------------------
 * fs_inodeWrite
 *
 * Description: Called to inidicate that an inode has been written.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
EXPORTED void
fs_inodeWrite(
    fs_Context      *context,
    fs_Inode        *inode,
    uint32          flags)
{
    boolean     addLock = FALSE;
    boolean     addBtreeLock = FALSE;
    fs_Buf      *cowBuf;

    FS_TRACE(context, FS_TRACE_INO,
             ("iwrite %lld snap %d flags %#x", 
              inode->inum, inode->snapnum, flags));

    FS_ASSERT(((flags & INO_WRITE_BTREE) &&
    	       fs_lockOwnedExclusive(context, &inode->btreeLock)) ||
    	      fs_lockOwnedExclusive(context, &inode->lock));

    FS_ASSERT(inode->snapnum == 0 ||
              inode->inum == FS_SNAPSHOT_BITS_INODE);

    NFX_SPIN_LOCK(&inode->spin);
    if (flags & INO_WRITE_LOG)
    {
        // Add the lock to the in-core transaction summary
        // if it isn't already.
	    if (flags & INO_WRITE_BTREE)
	    {
            if (!(inode->flags & INO_FLAGS_BTREE_TRANS_LOCK))
	        {
                inode->flags |= INO_FLAGS_BTREE_TRANS_LOCK;
                addBtreeLock = TRUE;
            }
	    }
        else if (!(inode->flags & INO_FLAGS_TRANS_LOCK))
        {
            inode->flags |= INO_FLAGS_TRANS_LOCK;
            addLock = TRUE;
        }

        inode->flags |= INO_FLAGS_TRANS_WRITE;
    }

    fs_inodeMarkDirty(context, inode);
    NFX_SPIN_UNLOCK(&inode->spin);

    // PR #1857: COW should be handled within the txn that triggers it
    cowBuf = fs_inodeIsCow(context, inode);
    if (cowBuf)
    {
        int32 bufOffset = FS_INODE_PAGE_OFFSET(inode->inum);

        // COW the buffer which logs the COW txn records
        // but will not include the new inode data
        fs_bufWrite(context, cowBuf, bufOffset, sizeof(fs_Dinode),
                    0, NULL, NULL);

        //printf("ino write: ino %p %lld COW %lld\n",
        //      inode, inode->inum, cowBuf->page);

        // remove all other inodes from this COW buf list
        // watchout: 
        // this works since the entire inode cache is flushed
        // after each snapshot, so another COW for the same
        // INODE page/offset will not occur

        NFX_SPIN_LOCK(&fs_inodeCache->spin);
        fs_inodeDeleteCow(context, inode, TRUE);
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        fs_bufReleaseExclusive(context, cowBuf);
    }

    if (addLock)
    {
        FS_TRACE(context, FS_TRACE_INO,
                ("  log ino %lld handle %llx", 
                 inode->inum, (uint64)((uintptr_t)context->smHandle)));
        fs_logAddInodeLock(context, context->logSummary, inode);
    }
    else if (addBtreeLock)
    {
        FS_TRACE(context, FS_TRACE_INO,
                 ("log btree ino %lld handle %llx", 
                 inode->inum, (uint64)((uintptr_t)context->smHandle)));
    	fs_logAddBtreeLock(context, context->logSummary, inode);
    }


} /* fs_inodeWrite */


/**
 * Adds an inode lock to the list of locked objects for
 * the current transaction.  It will be unlocked when the
 * transaction commits.
 */
EXPORTED void
fs_inodeLockTransaction(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_LogSummary   *logSummary)
{
    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    // Add the inode lock to the transaction if its
    // not already.
    NFX_SPIN_LOCK(&inode->spin);
    if (!(inode->flags & INO_FLAGS_TRANS_LOCK))
    {
        inode->flags |= INO_FLAGS_TRANS_LOCK;
        NFX_SPIN_UNLOCK(&inode->spin);
        FS_TRACE(context, FS_TRACE_INO, ("  trans lock ino %lld handle %llx", 
                 inode->inum, (uint64)((uintptr_t)context->smHandle)));
        fs_logAddInodeLock(context, logSummary, inode);
    }
    else
    {
        NFX_SPIN_UNLOCK(&inode->spin);
    }
}


/**
 * Called during transaction commit to copy the
 * specified inode to the log.
 */
EXPORTED void
fs_inodeUpdateLog(
    fs_Context  *context,
    fs_Inode    *inode)
{
    NFX_SPIN_LOCK(&inode->spin);
    if ((inode->flags & INO_FLAGS_TRANS_WRITE) &&
        !(inode->flags & INO_FLAGS_PENDING_DEALLOC))
    {
        inode->flags &= ~INO_FLAGS_TRANS_WRITE;
        NFX_SPIN_UNLOCK(&inode->spin);

        if (context->eek.inUse == 0)
        {
            // Copy the inode to the log.
            fs_logAddInodeRecord(context, inode);
        }
    }
    else
    {
        NFX_SPIN_UNLOCK(&inode->spin);
    }

} /* fs_inodeUpdateLog */


/**
 * Called during tranaction commit to finish the act of destroying
 * an inode.
 */
static void
fs_inodeDemolish(
    fs_Context  *context,
    fs_Inode    *inode)
{
    INO_TRACE(("esm %llx inode demolish %lld",
			    (uint64)((uintptr_t)context->smHandle), inode->inum));

    FS_ASSERT(inode->flags & INO_FLAGS_PENDING_DEALLOC);
    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    if (inode->logSegDep.count > 0)
    {
        fs_logRemoveDeps(context, &inode->logSegDep, NULL);
    }

    if (inode->logRefCount != 0)
    {
        // Don't add it to the free list, we'll
        // do it once the logRefCount reaches 0.

        fs_inodeRemoveFromCache(context, inode, FALSE);

		//printf("SET FREE %p 0x%x\n", inode, inode->flags);
        inode->flags |= INO_FLAGS_FREE_LOGBUF;
    }
    else
    {
        fs_inodeRemoveFromCache(context, inode, TRUE);
    }

} /* fs_inodeDemolish */


/**
 * Called during transaction commit to unlock the
 * specified inode.
 */
EXPORTED void
fs_inodeCommit(
    fs_Context  *context,
    fs_Inode    *inode)
{
    INO_TRACE(("esm %llx commit inum %lld flags 0x%x",
			   (uint64)((uintptr_t)context->smHandle), inode->inum, inode->flags));

    if (inode->flags & INO_FLAGS_PENDING_DEALLOC)
    {
        // Finish destruction of this inode.
        fs_inodeDemolish(context, inode);
        return;
    }

    // ?? NFX_SPIN_LOCK(&inode->spin);
    // fs_lockOwnedExclusive needed to update flags without spinlock
    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    NFX_SPIN_LOCK(&inode->spin);
    inode->flags &= ~INO_FLAGS_TRANS_LOCK;
    NFX_SPIN_UNLOCK(&inode->spin);

    // Release the inode.
    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    fs_lockCommit(context, &inode->lock);
    fs_inodeMaybeAddToFreeList(context, inode);
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

} /* fs_inodeCommit */


/**
 * Return true if the inode has been dirtied by the current transaction.
 */
EXPORTED
boolean fs_inodeWasModified(
    fs_Context *context,
    fs_Inode *inode)
{
    assert(fs_lockOwnedExclusive(context, &inode->lock));
    return (inode->flags & INO_FLAGS_TRANS_LOCK) != 0;
}


/**
 * Return true if the inode btree has been dirtied by the current transaction.
 */
EXPORTED
boolean fs_inodeBtreeWasModified(
    fs_Context *context,
    fs_Inode *inode)
{
    assert(fs_lockOwnedExclusive(context, &inode->btreeLock));
    return (inode->flags & INO_FLAGS_BTREE_TRANS_LOCK) != 0;
}


/**
 * Called during transaction commit to unlock the specified
 * inode btree.
 */
EXPORTED void
fs_inodeBtreeCommit(
    fs_Context  *context,
    fs_Inode    *inode)
{
    NFX_SPIN_LOCK(&inode->spin);
    inode->flags &= ~INO_FLAGS_BTREE_TRANS_LOCK;
    NFX_SPIN_UNLOCK(&inode->spin);

    // Release the inode btree lock.
    fs_lockCommit(context, &inode->btreeLock);

} /* fs_inodeBtreeCommit */


/**
 * Signal the inode cleaner for the specified filesystem.
 */
EXPORTED void
fs_inodeSignalCleaner(
	fs_Context	*context,
    uint32      type)
{
	fs_FileSystem	*fs = context->fileSystem;
    esm_event_t     event;
    int32           rc;

	if (fs->state == FS_VOL_STATE_ONLINE)
    {
        event.event = FS_THREAD_EVENT_START;
        event.p1.u.w = 0;
        event.p2.u.w = type;  // debug

        rc = fs_daemonWakeup(context, &fs->spin,
                             fs->inodeCleanerHandle,
                             &fs->inodeCleanerSignal, event);
    
        INO_TRACE(("vol 0x%llx cleaner signalled, daemon %s awakened",
           		    fs->volId.whole_thing, rc == TRUE ? "was" : "was not"));
    }
}


/**
 * Clean up the specified inode by writing its contents to
 * its dinode buffer.  Returns TRUE if it commited a transaction
 * in the process.
 */
PRIVATE boolean
fs_inodeClean(
    fs_Context  *context,
    fs_Inode    *inode)
{
    fs_Buf  *buf;
    int32   bufOffset;
    boolean isCOW;

    FS_ASSERT(fs_lockHeldShare(context, &inode->lock) ||
              fs_lockOwnedExclusive(context, &inode->lock));
	FS_ASSERT(inode->snapnum == 0 ||
              inode->inum == FS_SNAPSHOT_BITS_INODE);

    if (inode->flags & INO_FLAGS_RESTORE_IN_PROGRESS)
    {
        fs_bufSyncInodeBufs(context, inode);
#ifdef NEW_ZF
        if (inode->zfflags & INO_FLAGS_UZERO) {
                /*
                 * Clean zero fill log records if any
                 */
                fs_logAddUserZeroRecord(context, inode, 0, 0, 0,
                                        FALSE, TRUE, NULL);
        }
#endif
        
        NFX_SPIN_LOCK(&inode->spin);
        inode->flags &= ~INO_FLAGS_RESTORE_IN_PROGRESS;
        NFX_SPIN_UNLOCK(&inode->spin);
    }

    NFX_SPIN_LOCK(&inode->spin);
    if (inode->logRefCount != 0)
    {
		//printf("SET WAIT %p 0x%x\n", inode, inode->flags);
        inode->flags |= INO_FLAGS_WAIT_LOGBUF;
        inode->cleanerHandle = context->smHandle;
    }
    NFX_SPIN_UNLOCK(&inode->spin);

    if (inode->flags & INO_FLAGS_WAIT_LOGBUF)
    {
        fs_signalVolLogWrite (context);

        fs_yield(context, NULL, FS_YIELD_TIME_ADMIN, FS_YIELD_TYPE_ADMIN);
        FS_ASSERT(inode->logRefCount == 0);
		//printf("CLEAR WAIT %p 0x%x\n", inode, inode->flags);
        // ?? NFX_SPIN_LOCK(&inode->spin);
        NFX_SPIN_LOCK(&inode->spin);
        inode->flags &= ~INO_FLAGS_WAIT_LOGBUF;
        inode->cleanerHandle = 0;
        NFX_SPIN_UNLOCK(&inode->spin);
    }

    buf = fs_inodeGetBufferExclusive(context, inode->inum,
                                     inode->snapnum, FALSE, FALSE);
    FS_ASSERT(buf != NULL);

    isCOW = fs_bufIsCopyOnWrite(context, buf);

    // COW should never be defered
    FS_ASSERT(!isCOW || inode->inum != FS_SNAPSHOT_BITS_INODE);

    FS_ASSERT(inode->dinode.attrs != 0 ||
              inode->inum == FS_SNAPSHOT_BITS_INODE);

    // Copy the new dinode to the buf.
    bufOffset = FS_INODE_PAGE_OFFSET(inode->inum);
    memcpy(buf->u.data + bufOffset, &inode->dinode, sizeof(fs_Dinode));

    // Indicate that the buf was written, which dirties the buf
    // and manages any COW concerns for it.
    fs_bufWrite(context, buf, bufOffset, sizeof(fs_Dinode),
                0, NULL, NULL);
    
    // Transfer the log dependencies from the inode to the dirtied buf.
    fs_logTransferDeps(context, &inode->logSegDep, inode->headLogSeg,
                       buf, FS_LOG_DEP_BUF);
    inode->headLogSeg = -1;

    fs_bufReleaseExclusive(context, buf);

    NFX_SPIN_LOCK(&inode->spin);
    inode->flags &= ~INO_FLAGS_DIRTY;
    NFX_SPIN_UNLOCK(&inode->spin);

    INO_TRACE(("cleaned inum %lld", inode->inum));

    return isCOW;

} /* fs_inodeClean */


/**
 * Called to decrement inode cache dirty count and notify
 * any waiting threads.  Returns TRUE if the specified smHandle
 * was notified.
 * 
 * Must be called with the fs_inodeCache Spinlock held.
 */
PRIVATE boolean
fs_inodeDecrDirtyCount(
    fs_Context      *context,
    fs_Inode        *inode)
{
    fs_FileSystem   *fileSystem = inode->fileSystem;
    fs_Waiter       *waiter, *waiterNext;
    boolean         rc = FALSE;

    FS_ASSERT(fs_inodeCache->count.numDirty > 0);
    FS_ASSERT(fileSystem->inodeCacheCount.numDirty > 0);
    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    fs_inodeCache->count.numDirty -= 1;
    fileSystem->inodeCacheCount.numDirty -= 1;
    fs_inodeCache->count.numCleaned += 1;
    fileSystem->inodeCacheCount.numCleaned += 1;

    //printf("decrDirty fs %p %d/%d\n", fileSystem,
    //      fs_inodeCache->count.numDirty,
    //      fileSystem->inodeCacheCount.numDirty);

    DLL_ITERATE_NEXT(waiter, waiterNext, &fileSystem->inodeWaiterHdr, dllEntry)
    {
        waiter->numWait -= 1;
        if (waiter->numWait == 0)
        {
            DLL_REMOVE(&fileSystem->inodeWaiterHdr, waiter, dllEntry);
            fs_notifyWaiter(context, fs_inodeCache, waiter);
            fs_ramDealloc(FS_RAM_WAITER, waiter);
            rc = TRUE;
        }
    }

    return rc;

} /* fs_inodeDecrDirtyCount */


/**
 * Inode cleaner thread.  There is one of these running per mounted
 * filesystem, as well as possibly additional ones spawned whenever
 * a thread desires to block on sync'ing a filesystem's inodes out to
 * disk.
 */


#ifdef FS_DEBUG_INODE_CLEANER_TIMING
eeeRTC_t fs_inodeCleanerPrintRuntimeThreshold;
int32 fs_inodeCleanerOverThreshold = 0;
int32 fs_inodeCleanerMaxRuntime = 0;
#endif

eeeRTC_t fs_inodeCleanerMaxRuntime;
int32 fs_inodeCleanerMaxPerLoop = 0;
int32 fs_inodeCleanerCount[FS_INODE_CLEANER_MAX];


/**
 * Clean a batch of inodes and optionally notify
 * waiterHandle.  Returns TRUE if waiterHandle was
 * notified, FALSE otherwise.
 */
static boolean
fs_inodeCleanBatch(
    fs_Context      *context,
    fs_InodeDllHdr  *dllHdr,
    int32           what)
{
    fs_Inode        *inode;
    int32           numDone = 0;
    boolean         locked, waiterNotified = FALSE;
#ifdef FS_DEBUG_INODE_CLEANER_TIMING
    eeeRTC_t	startTime;
    eeeRTC_t	delta;
#endif
	static boolean first = TRUE;
    
	if (first) {
		first = FALSE;
		// Do some initialization.
#ifdef FS_DEBUG_INODE_CLEANER_TIMING
		eee_msToRTC(2000, &fs_inodeCleanerPrintRuntimeThreshold);
#endif
		fs_inodeCleanerMaxRuntime = *(eee_rtcZero());
	}
#ifdef FS_DEBUG_INODE_CLEANER_TIMING
	eee_rtcStart(&startTime);
#endif

    NFX_SPIN_LOCK(&fs_inodeCache->spin);

//#ifdef FS_DEBUG_INODE_CLEANER_TIMING
#if 1
    FS_ASSERT(what < FS_INODE_CLEANER_MAX);
    fs_inodeCleanerCount[what] += 1;
#endif

    while (!DLL_IS_EMPTY(dllHdr) &&
           !waiterNotified)
    {
        if (numDone++ >= fs_incoreTune->inodeCleanPerPass)
        {
            NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
            //printf("cleaned %d inodes\n", numDone);
            return waiterNotified;
        }

        DLL_REMOVE_TAIL(dllHdr, inode, mrdEntry);
        inode->cacheflags &= ~INO_FLAGS_MRD;

        locked = fs_trylockSharedSleep(context, &inode->lock,
                                       &fs_inodeCache->spin);
        if (!locked)
            continue;
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

        locked = fs_trylockSharedSleep(context, &inode->btreeLock, NULL);
        FS_ASSERT(locked);

        fs_inodeClean(context, inode);

        NFX_SPIN_LOCK(&fs_inodeCache->spin);
        waiterNotified = fs_inodeDecrDirtyCount(context, inode);
        fs_unlockShared(context, &inode->btreeLock);
        fs_unlockShared(context, &inode->lock);
        fs_inodeMaybeAddToFreeList(context, inode);
    }

    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

#ifdef FS_DEBUG_INODE_CLEANER_TIMING
	eee_rtcDuration(&startTime, &delta);
	if (eee_rtcCompare(&delta, &fs_inodeCleanerMaxRuntime) > 0)
    {
        fs_inodeCleanerMaxRuntime = delta;
		if (eee_rtcCompare(&delta, &fs_inodeCleanerPrintRuntimeThreshold) > 0) // > 2 sec
        {
			//eeeRTC_t endTime;
			//uint32	 deltaMS;

			//deltaMS = (uint32)eee_rtcToMS(&delta)
            fs_inodeCleanerOverThreshold++;
            if (numDone > 0)
            {
                //printf(
                //"fs_inodeCleanerThread:  numdone %d runtime %d avg %d\n",
                //    numDone, deltaMS, deltaMS/numDone);
            }
            else
            {
                //printf(
                //    "fs_inodeCleanerThread:  numdone %d runtime %d\n",
                //        numDone, deltaMS);
            }
			//eee_rtcAdd(&startTime, &delta, &endTime);
			//printf("%s\n", eee_humandate((uint32)eee_rtcToSec(&endTime)));
        }
    }
#endif
    if (numDone > fs_inodeCleanerMaxPerLoop)
    {
        fs_inodeCleanerMaxPerLoop = numDone;
    }

    //printf("cleaned %d inodes\n", numDone);
    return waiterNotified;
}


/**
 * Clean inodes on the dirty log segment headers
 * until there are no more dirty or waiter
 * notification is made.
 */
void
fs_inodeCleanerTmp(
    fs_Context      *context)
{
    fs_FileSystem   *fileSystem = context->fileSystem;
    esm_event_t     event;
    int32           i, segNum;
    boolean         cleanDone = FALSE, waiterNotified = FALSE;
    uint64          cleanedCount;

    // Wait for start event.
    fs_yield(context, &event, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_ADMIN);
    cleanedCount = event.p1.u.dw;

    // PR #1789: acquire global access lock
    fs_lockShared(context, &fileSystem->accessLock, NULL);

    segNum = fs_logGetHeadSegnum(context, fileSystem);

    for (i = 0; i < FS_LOG_MAX_SEGMENTS; i++)
    {
        while (!DLL_IS_EMPTY(&fileSystem->inodeLogSegmentHdr[segNum]))
        {
            waiterNotified = fs_inodeCleanBatch(context,
                                  &fileSystem->inodeLogSegmentHdr[segNum],
                                  event.p2.u.w);

            if (waiterNotified ||
                (cleanedCount &&
                 fileSystem->inodeCacheCount.numCleaned >= cleanedCount))
            {
                cleanDone = TRUE;
                break;
            }

            // release global access lock before sleep
            fs_unlockShared(context, &fileSystem->accessLock);
    
            // Release CPU for a bit.
    
            INO_TRACE(("vol 0x%llx cleaner rescheduling self, numDirty %d",
                       context->fileSystem->volId.whole_thing,
                       fs_inodeCache->count.numDirty));
            fs_sleep(context, 0);
    
            // PR #1789: acquire global access lock
            fs_lockShared(context, &fileSystem->accessLock, NULL);
        }

        if (cleanDone)
        {
            break;
        }

        if (++segNum >= FS_LOG_MAX_SEGMENTS)
        {
            segNum = 0;
        }
    }

    if (!cleanDone)
    {
        // There may be dirty inodes that aren't logged,
        // try cleaning them.
        while (!DLL_IS_EMPTY(&fileSystem->inodeMrdHdr))
        {
            waiterNotified = fs_inodeCleanBatch(context,
                                  &fileSystem->inodeMrdHdr,
                                  event.p2.u.w);

            if (waiterNotified ||
                (cleanedCount &&
                 fileSystem->inodeCacheCount.numCleaned >= cleanedCount))
            {
                break;
            }
        }
    }

    fs_bufSignalWriteback(context, FS_BUFLIST_META, FS_WB_WAKEUP_INO_SYNC);

    // release global access lock
    fs_unlockShared(context, &fileSystem->accessLock);
}


/**
 * Return TRUE if a dirty inode is potentially
 * holding up log truncation of the current head
 * log segment.
 */
boolean
fs_inodeBlockingLogTruncation(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    int32 headSeg = fs_logGetHeadSegnum(context, fileSystem);

    if (!DLL_IS_EMPTY(&fileSystem->inodeLogSegmentHdr[headSeg]))
    {
        return TRUE;
    }

    return FALSE;
}


/**
 * Main inode cleaner thread for a volume.  This
 * thread lives for the entire duration of a mount.
 */
void
fs_inodeCleanerMain(
    fs_Context  *context)
{
    fs_FileSystem   *fileSystem = context->fileSystem;
    esm_event_t     event;
    int32           i, segNum;
    boolean         didClean;

	// If we're the main cleaner thread for the filesystem, we
	// must obtain the access lock, otherwise we can deadlock
	// with eek.  We DO NOT get the lock if we're an ancilliary
	// cleaner thread, since things like fs_freeze grab it EXCL
	// and then spawn us to do a sync.

    event.p2.u.w = FS_INODE_CLEANER_START;

    fs_lockShared(context, &fileSystem->accessLock, NULL);

    while (TRUE)
    {
        didClean = FALSE;

        while (fs_inodeBlockingLogTruncation(context, fileSystem) ||
               fs_logAboveWritebackThreshold(context, fileSystem) ||
               fs_inodePercentAvail() < fs_incoreTune->inodePctAvailHiWtrMk)
        {
            if (fileSystem->inodeCacheCount.numDirty == 0)
            {
                // Nothing I can do to help, go back to sleep.
                break;
            }

            segNum = fs_logGetHeadSegnum(context, fileSystem);
            for (i = 0; i < FS_LOG_MAX_SEGMENTS; i++)
            {
                if (!fs_inodeBlockingLogTruncation(context, fileSystem) &&
                    !fs_logAboveWritebackThreshold(context, fileSystem) &&
                    (fs_inodePercentAvail() > fs_incoreTune->inodePctAvailHiWtrMk))
                {
                    // No urgency to do any more, wait for next
                    // wakeup event.
                    break;
                }
                while (!DLL_IS_EMPTY(&fileSystem->inodeLogSegmentHdr[segNum]))
                {
                    fs_inodeCleanBatch(context,
                                   &fileSystem->inodeLogSegmentHdr[segNum],
                                   event.p2.u.w);
    
                    didClean = TRUE;
    
                    fs_unlockShared(context, &fileSystem->accessLock);
                    fs_sleep(context, 10);
                    fs_lockShared(context, &fileSystem->accessLock, NULL);
                }
                if (++segNum >= FS_LOG_MAX_SEGMENTS)
                {
                    segNum = 0;
                }
            }

            // There may be dirty inodes that aren't logged,
            // try cleaning them. 
            if (!DLL_IS_EMPTY(&fileSystem->inodeMrdHdr))
            {
                fs_inodeCleanBatch(context,
                                   &fileSystem->inodeMrdHdr,
                                   event.p2.u.w);
                didClean = TRUE;
            }

            fs_unlockShared(context, &fileSystem->accessLock);
            fs_sleep(context, 10);
            fs_lockShared(context, &fileSystem->accessLock, NULL);
        }

        if (didClean)
        {
            fs_bufSignalWriteback(context, FS_BUFLIST_META, FS_WB_WAKEUP_INO_SYNC);
        }

        NFX_SPIN_LOCK(&fs_inodeCache->spin);
        fileSystem->inodeCleanerSignal = 0;
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

        fs_unlockShared(context, &fileSystem->accessLock);
        fs_yield(context, &event, FS_YIELD_TIME_FOREVER,
                 FS_YIELD_TYPE_ADMIN);
        fs_lockShared(context, &fileSystem->accessLock, NULL);
        FS_ASSERT(fileSystem->inodeCleanerSignal == 1);
    }
}


/**
 * Thread wrapper for inode cleaner.
 */
void
fs_inodeCleanerThread(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;
    fs_FileSystem   *fileSystem = context->fileSystem;

    boolean         mainCleaner = (context->smHandle ==
                                   fileSystem->inodeCleanerHandle);
    if (mainCleaner)
    {
        fs_inodeCleanerMain(context);
        /* Doesn't terminate */
    }
    else
    {
        fs_inodeCleanerTmp(context);
        fs_threadTerminate(context);
    }
}


/**
 * Start the process of syncing the specified context's
 * inode cache.
 */
PRIVATE void
fs_doInodeSyncCache(
    fs_Context  *context,
    boolean     spawnNew)
{
	esm_event_t     event;
    uint64          cleanedCount;
    int32           rc;

    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    
    if (context->fileSystem->inodeCacheCount.numDirty == 0) {
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        INO_TRACE(("inode cache sync returned immediately (nothing to do)"));
        return;
    }

    if (spawnNew)
    {
        fs_Waiter       *waiter;

        waiter = fs_ramAlloc(FS_RAM_WAITER, sizeof *waiter);
        waiter->handle = context->smHandle;
        waiter->event.event  = FS_THREAD_EVENT_SYNC;
        waiter->event.p1.u.p = fs_inodeCache;
        waiter->event.p2.u.p = NULL;

	    DLL_ISOLATE_ELEM(waiter, dllEntry);
        DLL_INSERT_TAIL(&context->fileSystem->inodeWaiterHdr, waiter, dllEntry);
        waiter->numWait = context->fileSystem->inodeCacheCount.numDirty;
        waiter->numWaitStart = waiter->numWait;

        // This case relies on waiter notification, so set
        // cleanedCount to 0.
        cleanedCount = 0;
    }
    else
    {
        // This case relies on the cleanedCount being reached.
        cleanedCount = context->fileSystem->inodeCacheCount.numCleaned +
                          context->fileSystem->inodeCacheCount.numDirty;
    }

    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

    INO_TRACE(("inode cache sync for volId %lld, numDirty %d",
            context->volId.whole_thing,
            context->fileSystem->inodeCacheCount.numDirty));

    event.event = FS_THREAD_EVENT_START;
    event.p1.u.dw = cleanedCount;
    event.p2.u.w = FS_INODE_CLEANER_START;  // debug

    if (spawnNew)
    {
        esm_handle_t    newThread;


        newThread = fs_threadFork(context, fs_inodeCleanerThread, TRUE, FALSE,
                                  FS_DAEMON_TYPE_INO_CLNR, NULL);

        // Post an event to the new thread to start the
        // cleaner routine.
        rc = thread_event(newThread, &event);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EEE_ERR, "fs_inodeSyncCache");
        }

        // Now wait for the notification.
        fs_yield(context, &event, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_MW);
        if (event.event != FS_THREAD_EVENT_SYNC)
        {
            fs_error(context, FS_INTERNAL_ERR,
                     "fs_doInodeSyncCache: event %d", event.event);
        }
    }
    else
    {
        // post an event to myself to jump start the cleaner routine
        rc = thread_event(context->smHandle, &event);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EEE_ERR, "fs_inodeSyncCache");
        }

        // Call tmp cleaner function directly.  When it returns,
        // we're done!
        fs_inodeCleanerTmp(context);
    }

} /* fs_doInodeSyncCache */


/**
 * Clean all of the specified filesystem's inodes and
 * wait for it to complete.  The sync is done in the
 * callers thread.
 */
EXPORTED void
fs_inodeSyncCache(
    fs_Context *context)
{
    fs_doInodeSyncCache(context, FALSE);
}


/**
 * Clean all of the specified filesystem's inodes and
 * wait for it to complete.  A new thread is created
 * to accomplish the sync.
 */
EXPORTED void
fs_inodeSyncCacheSpawnNew(
    fs_Context *context)
{
    fs_doInodeSyncCache(context, TRUE);
}


#if !defined(FS_STANDALONE)
/**
 * Timer function to signal inode cleaner to wake up.
 */
PRIVATE int32
fs_inodeSyncTimeout(
    uint32  timer,
    void    *param1,
    void    *param2,
    void    *param3,
    void    *param4)
{
    fs_inodeSignalCleaners(NULL);
    return NFX_OK;
}
#endif


/**
 * Set the upper limit on the inode cache size.
 */
void
fs_inodeSetCacheLimit()
{
    uint64  memAvail, memUse;
    int32   sizeEach;

    NFX_SPIN_LOCK(&fs_inodeCache->spin);

    memAvail = eee_ramTotalFlexSize();
    memUse = (memAvail / 100) * fs_incoreTune->inodePctMem;
#ifdef FS_USE_MALLOC_SLAB
    sizeEach = sizeof(fs_Inode);
    // (memUse / fs_inode_cache->c_slabsize) * fs_inode_cache->c_num;
#else
    sizeEach = (((sizeof(fs_Inode) - 1) / MM_FLEX_MODULO + 1) * MM_FLEX_MODULO) + MM_FLEX_OVERHEAD ;
#endif
    fs_inodeCache->count.numCachedLimit = memUse / sizeEach;
    // printf("inodes cache limit: %d old: %d\n", fs_inodeCache->count.numCachedLimit);
    
#ifdef FS_DYN_CACHE_SIZE
    fs_inodeCache->count.numCachedLimitLWM = fs_inodeCache->count.numCachedLimit;
    fs_inodeCache->count.numCachedLimitHWM = (1 << 31);  // no limit yet
#endif

    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
}

/**
 * Initialize the inode cache.
 */
EXPORTED void
fs_inodeCacheInit()
{
    int32   i;

    fs_inodeCache = (void *) eee_ramAlloc(sizeof *fs_inodeCache);
    FS_ASSERT(fs_inodeCache != NULL);
    bzero(fs_inodeCache, sizeof *fs_inodeCache);

    NFX_SPIN_INIT(&fs_inodeCache->spin,SPINLOCK_LEVEL_UNKNOWN);
    for (i = 0; i < FS_ICACHE_HASH_MAX; i++)
    {
        DLL_INIT(&fs_inodeCache->cache[i]);
    }
    DLL_INIT(&fs_inodeCache->mruHdr);
    for (i = 0; i < FS_ICACHE_COW_HASH_MAX; i++)
    {
        DLL_INIT(&fs_inodeCache->cowHdr[i]);
    }

    fs_inodeSetCacheLimit();

#if !defined(FS_STANDALONE)
    // Create a timer to periodically sync the inode cache.
    fs_inodeCache->syncTimer = eee_timerCreate(FS_INODE_SYNC_TIMEOUT,
                                               EEE_TIMER_REPEAT);
    if (fs_inodeCache->syncTimer == TIMER_ERR_NO_TIMER_AVAILABLE)
    {
        fs_error(NULL, FS_RAM_ALLOC_ERR, "create icache sync timer");
    }
    else
    {
        if (eee_timerStart(fs_inodeCache->syncTimer, fs_inodeSyncTimeout,
                           NULL, NULL, NULL, NULL) != NFX_OK)
        {
            fs_error(NULL, FS_INTERNAL_ERR, "start icache sync timer");
        }
    }
#endif

    // Initialize security descriptor cache.
    bzero(&fs_securityDesc, sizeof fs_securityDesc);
    NFX_SPIN_INIT(&fs_securityDesc.spin,SPINLOCK_LEVEL_UNKNOWN);
    for (i = 0; i < FS_SDESC_HASH_MAX; i++)
    {
        DLL_INIT(&fs_securityDesc.secIdHdr[i]);
        DLL_INIT(&fs_securityDesc.uidHdr[i]);
    }
    for (i = 0; i < FS_SDESC_TYPE_MAX; i++)
    {
        DLL_INIT(&fs_securityDesc.freeHdr[i]);
        fs_securityDesc.numUsedLimit[i] = FS_SDESC_NUM_USED_LIMIT;
    }

    fs_securityDesc.startId[FS_SDESC_TYPE_NFS]   = FS_SDESC_ID_START_NFS;
    fs_securityDesc.startId[FS_SDESC_TYPE_CIFS]  = FS_SDESC_ID_START_CIFS;

} /* fs_inodeCacheInit */


/**
 * Called to invalidate an inode from the cache.
 */
EXPORTED void
fs_inodeRemoveFromCache(
    fs_Context      *context,
    fs_Inode        *inode,
    boolean         leaveOnMru)
{
    INO_TRACE(("esm %llx inode destroy %lld",
			   (uint64)((uintptr_t)context->smHandle), inode->inum));

    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    // no need, we fs_lockOwnedExclusive --> NFX_SPIN_LOCK(&inode->spin);
    NFX_SPIN_LOCK(&fs_inodeCache->spin);

    // Remove the inode from the dirty list.
    if (!DLL_IS_ISOLATED(inode, mrdEntry))
    {
        if (inode->cacheflags & INO_FLAGS_MRD)
        {
            inode->cacheflags &= ~INO_FLAGS_MRD;
            DLL_REMOVE(&inode->fileSystem->inodeMrdHdr, inode, mrdEntry);
        }
        else
        {
            DLL_REMOVE(&inode->fileSystem->inodeLogSegmentHdr[inode->headLogSeg],
                       inode, mrdEntry);
            inode->headLogSeg = -1;
        }
    }
    // Clear dirty flag and decrement dirty count.
    if (inode->flags & INO_FLAGS_DIRTY)
    {
        inode->flags &= ~INO_FLAGS_DIRTY;
        fs_inodeDecrDirtyCount(context, inode);
    }

    // Remove the inode from the cache.
    inode->flags = 0;
    inode->zfflags = 0;
    inode->cacheflags = 0;
    fs_inodeCacheRemove(context, inode);

    // Call destroy on the lock so any waiters will be
    // notified of its destruction.
    fs_lockDestroy(context, &inode->lock);

    if (leaveOnMru)
    {
        // Reinit the inode lock and put the inode on the free
        // list.
        // unlock before calling fs_inodeCacheInsertMruOnly who
        // can dealloc the inode
        inode->inum = 0;
        fs_lockInit(&inode->lock, FS_LOCK_TYPE_INODE);
        fs_inodeCacheInsertMruOnly(context, inode);
    }
    
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

} /* fs_inodeRemoveFromCache */


/**
 * Called from the allocator during the first phase
 * of deallocating an inode.
 */
EXPORTED void
fs_inodeDeallocStart(
    fs_Context  *context,
    fs_Inode    *inode)
{
    // fs_lockOwnedExclusive needed to update flags without spinlock
    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

    inode->flags |= INO_FLAGS_PENDING_DEALLOC;

    fs_inodeLockTransaction(context, inode, context->logSummary);

    // Add a log entry indicating that this inode was flushed.
    fs_logAddInodeFlushRecord(context, inode);

} /* fs_inodeDeallocStart */


/*-----------------------------------------------------------------------------
 * fs_inodeCacheFlushSnap
 *
 * Description: Called to flush the contents of the inode cache for the
 *              specified volume and snapNum (if non zero).
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
EXPORTED void
fs_inodeCacheFlushSnap(
    fs_Context *context,
    int32       snapNum)
{
    fs_FileSystem   *fileSystem;
    fs_Inode        *inode, *inodeNext;
    int32           rc, numDone, numCached;
    FS_LOOP_DBG_DECL(dbgNumLoops);

    numDone = 0;
    fileSystem = context->fileSystem;
    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    numCached = fs_inodeCache->count.numCached;
    DLL_ITERATE_NEXT(inode, inodeNext, &fs_inodeCache->mruHdr, mruEntry)
    {
        FS_LOOP_DBG_CHECK(dbgNumLoops, numCached + LOOP_DBG_FUDGE);

        //We can not yield here because we are holding the cache spin lock
        //The loop count is checked above, so we can refresh the watchdog 
        //without any risk of infinite loop.
        //watchdog_kick is such a short function that it is ok to call it at
        //each loop.
#ifndef FS_STANDALONE
#ifdef USE_WATCHDOG
        watchdog_kick();
#endif // USE_WATCHDOG
#endif // FS_STANDALONE

        if (inode->fileSystem != fileSystem)
            continue;

        if (snapNum && inode->snapnum != snapNum)
            continue;

        FS_ASSERT(fs_inodeIsAvail(inode));

        rc = fs_trylock(context, &inode->lock, FS_LOCK_MODE_EXCL);
        if (rc == TRUE)
        {
            if (!fs_inodeIsAvail(inode))
            {
                fs_unlockExclusive(context, &inode->lock);
                continue;
            }

            inode->flags = 0;
            inode->zfflags = 0;
            inode->cacheflags = 0;

            INO_TRACE(("flush ino %p inum %lld snap %d",
                inode, inode->inum, inode->snapnum));

            fs_inodeCacheRemove(context, inode);

            fs_lockDestroy(context, &inode->lock);

            if (inode->readAheadState != NULL)
            {
                FS_ASSERT(DLL_IS_EMPTY(
                           &inode->readAheadState->pendingReadList));
                fs_ramDealloc(FS_RAM_READ_AHEAD, inode->readAheadState);
                inode->readAheadState = NULL;
            }

            fs_ramDealloc(FS_RAM_INODE, inode);

            numDone += 1;
        }
        else
        {
            FS_ASSERT(FALSE);
            INO_TRACE(("can't flush ino %p inum %lld", inode, inode->inum));
        }
    }
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

    INO_TRACE(("flushed %d inodes for vol %llx",
                numDone, context->volId.whole_thing));

    // Now flush any cached security descriptors.
    fs_securityDescCacheFlush(context, 0);

    if (snapNum == 0)
    {
        fs_dnlcDestroyVolume(context->fileSystem);
        fs_dirhash_destroy_volume(context->fileSystem);
    }
    else
    {
        fs_dnlcDestroySnapnum(context->fileSystem, snapNum);
        fs_dirhash_destroy_snapnum(context->fileSystem, snapNum);
    }

} /* fs_inodeCacheFlush */


/**
 * Destroy the specified inode.
 *
 * Must be called with the fs_inodeCache Spinlock held.
 */
static void
fs_inodeCacheDestroyInode(
    fs_Context  *context,
    fs_Inode    *inode)
{
    INO_TRACE(("destroy ino %p inum %lld snap %d",
        inode, inode->inum, inode->snapnum));

    ASSERT_LOCK_OWNED(&fs_inodeCache->spin);

    if (!DLL_IS_ISOLATED(inode, mrdEntry))
    {
        if (inode->cacheflags & INO_FLAGS_MRD)
        {
            inode->cacheflags &= ~INO_FLAGS_MRD;
            DLL_REMOVE(&inode->fileSystem->inodeMrdHdr, inode, mrdEntry);
        }
        else
        {
            DLL_REMOVE(&inode->fileSystem->inodeLogSegmentHdr[inode->headLogSeg],
                       inode, mrdEntry);
        }
    }

    if (inode->flags & INO_FLAGS_DIRTY)
    {
        FS_ASSERT(fs_inodeCache->count.numDirty > 0);
        FS_ASSERT(inode->fileSystem->inodeCacheCount.numDirty > 0);
        fs_inodeCache->count.numDirty -= 1;
        inode->fileSystem->inodeCacheCount.numDirty -= 1;
    }
    inode->flags = 0; // avoid FS_ASSERT in inodeCacheRemove
    inode->zfflags = 0; // avoid FS_ASSERT in inodeCacheRemove
    inode->cacheflags = 0; // avoid FS_ASSERT in inodeCacheRemove
    fs_inodeCacheRemove(context, inode);
}

/**
 * Destroy the contents of the inode cache for the
 * specified volume without trying to sync to disk,
 * and without regard to inode locks.
 *
 * This is called when a volume is being forcibly
 * unmounted.
 */
void
fs_inodeCacheDestroyVolume(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    fs_Inode        *inode, *inodeNext;
    int32           numDone, numCached, hash;
    FS_LOOP_DBG_DECL(dbgNumLoops);

    numDone = 0;
    NFX_SPIN_LOCK(&fs_inodeCache->spin);
    numCached = fs_inodeCache->count.numCached;
    DLL_ITERATE_NEXT(inode, inodeNext, &fs_inodeCache->mruHdr, mruEntry)
    {
        FS_LOOP_DBG_CHECK(dbgNumLoops, numCached + LOOP_DBG_FUDGE);

        if (inode->fileSystem != fileSystem)
            continue;

        fs_inodeCacheDestroyInode(context, inode);

        if (inode->readAheadState != NULL)
        {
            FS_ASSERT(DLL_IS_EMPTY(
                           &inode->readAheadState->pendingReadList));
            fs_ramDealloc(FS_RAM_READ_AHEAD, inode->readAheadState);
            inode->readAheadState = NULL;
        }
        fs_ramDealloc(FS_RAM_INODE, inode);
        numDone += 1;
    }
    FS_LOOP_DBG_INIT(dbgNumLoops);
    for (hash = 0; hash < FS_ICACHE_HASH_MAX; hash += 1)
    {
        DLL_ITERATE_NEXT(inode, inodeNext, &fs_inodeCache->cache[hash],
                         cacheEntry)
        {
            FS_LOOP_DBG_CHECK(dbgNumLoops, numCached + LOOP_DBG_FUDGE);

            if (inode->fileSystem != fileSystem)
                continue;

            fs_inodeCacheDestroyInode(context, inode);

            if (inode->readAheadState != NULL)
            {
                FS_ASSERT(DLL_IS_EMPTY(
                           &inode->readAheadState->pendingReadList));
                fs_ramDealloc(FS_RAM_READ_AHEAD, inode->readAheadState);
                inode->readAheadState = NULL;
            }
            fs_ramDealloc(FS_RAM_INODE, inode);
            numDone += 1;
        }
    }
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

    INO_TRACE(("destroyed %d inodes for vol %llx",
                numDone, context->volId.whole_thing));

    fs_securityDescCacheFlush(context, 0);

    fs_dnlcDestroyVolume(context->fileSystem);
    fs_dirhash_destroy_volume(context->fileSystem);

} /* fs_inodeCacheDestroyVolume */


/**
 * Increment the log buffer reference
 * count on the specified inode.
 */
void
fs_inodeLogRefIncr(
    fs_Context  *context,
    fs_Inode    *inode)
{
    FS_ASSERT(fs_lockHeldExclusive(context, &inode->lock));

    NFX_SPIN_LOCK(&inode->spin);
    inode->logRefCount += 1;
    NFX_SPIN_UNLOCK(&inode->spin);

} /* fs_inodeLogRefIncr */


/**
 * Decrement the log buffer reference
 * count on the specified inode.
 */
void
fs_inodeLogRefDecr(
    fs_Context  *context,
    fs_Inode    *inode)
{
    boolean     doCleanerEvent = FALSE;
    boolean     doFree = FALSE;

    // We guarantee that the inode pointer is valid
    // since it is dirty and we won't free a dirty
    // inode until the count goes to 0.
    // 
    // TODO - some way to assert this?

    NFX_SPIN_LOCK(&inode->spin);
    FS_ASSERT(inode->logRefCount > 0);
    inode->logRefCount -= 1;
    if (inode->logRefCount == 0)
    {
        if (inode->flags & INO_FLAGS_WAIT_LOGBUF)
        {
            doCleanerEvent = TRUE;
        }
        else if (inode->flags & INO_FLAGS_FREE_LOGBUF)
        {
            doFree = TRUE;
        }
    }
    NFX_SPIN_UNLOCK(&inode->spin);

    if (doCleanerEvent)
    {
        esm_event_t     event;
        int32           rc;

        event.event = FS_THREAD_EVENT_YIELD_WAKE;
        // event.p1.u.p = inode;
        event.p1.u.p = NULL;
        event.p2.u.w = FS_INODE_CLEANER_REFDECR;
		//printf("EVENT WAIT %p 0x%x\n", inode, inode->flags);
        rc = esm_deliverEvent(inode->cleanerHandle, &event);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EVENT_DELIVERY_FAIL,
                     "logbuf cleaner signal");
        }
    }
    else if (doFree)
    {
        // Reinit the inode lock and put the inode on the free
        // list.
		//printf("DO FREE %p 0x%x\n", inode, inode->flags);
        inode->flags = 0;
        inode->zfflags = 0;
        inode->cacheflags = 0;
        inode->inum = 0;
        fs_lockInit(&inode->lock, FS_LOCK_TYPE_INODE);
        NFX_SPIN_LOCK(&fs_inodeCache->spin);
        fs_inodeCacheInsertMruOnly(context, inode);
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
    }

} /* fs_inodeLogRefDecr */


/*
 * Lookup an entry in the given directory and return exclusive locked inode
 */

EXPORTED fs_Inode *
fs_lookup(
	fs_Context	*context,
	fs_FileSystem	*core,
	uint64		snapnum,	// snapshot number
	fs_Inode	*parent,	// directory in which search is made
	char		*name		// name of the file/dir being lookedup
)
{
	uint64				inum;
	fs_Inode			*inode;
	fs_FileSetIncoreRecord_t	*fset, *pFset;

	FS_ASSERT(context != NULL && core != NULL && parent != NULL &&
		  name != NULL);
	FS_ASSERT(fs_lockOwnedExclusive(context, &parent->lock));

	inum = fs_dir_lookupent(context, parent, name, strlen(name));

	/*
	 * We did not find the entry in the directory
	 */

	if (inum == 0) {
		return NULL;
	}

	/*
	 * XXX - Fileset initialization belongs in other lookup paths as
	 * well
	 */

	pFset =  parent->fset;
	inode = I_GET_EXCL(snapnum, inum);
	fs_inodeReleaseExclusive(context, parent);
	fset = fs_getFset(context, core, inode);
	FS_ASSERT(fset != NULL);

	inode->fset = fset;
	return inode;
}

EXPORTED fs_Inode *
fs_lookupKeepParentLocked(
	fs_Context	*context,
	fs_FileSystem	*core,
	uint64		snapnum,	// snapshot number
	fs_Inode	*parent,	// directory in which search is made
	char		*name		// name of the file/dir being lookedup
)
{
	uint64				inum;
	fs_Inode			*inode;
	fs_FileSetIncoreRecord_t	*fset;

	FS_ASSERT(context != NULL && core != NULL && parent != NULL &&
		  name != NULL);
	FS_ASSERT(fs_lockOwnedExclusive(context, &parent->lock));

	inum = fs_dir_lookupent(context, parent, name, strlen(name));

	/*
	 * We did not find the entry in the directory
	 */

	if (inum == 0) {
		return NULL;
	}

	/*
	 * XXX - Fileset initialization belongs in other lookup paths as
	 * well
	 */

	inode = I_GET_EXCL(snapnum, inum);
	fset = fs_getFset(context, core, inode);
	FS_ASSERT(fset != NULL);
	inode->fset = fset;
	return inode;
}

EXPORTED fs_Inode *
fs_lookupShared(
	fs_Context	*context,
	fs_FileSystem	*core,
	uint64		snapnum,	// snapshot number
	fs_Inode	*parent,	// directory in which search is made
	char		*name		// name of the file/dir being lookedup
)
{
	uint64				inum;
	fs_Inode			*inode;
	fs_FileSetIncoreRecord_t	*fset;

	FS_ASSERT(context != NULL && core != NULL && parent != NULL &&
		  name != NULL);
	FS_ASSERT(fs_lockHeldShare(context, &parent->lock));

	inum = fs_dir_lookupent(context, parent, name, strlen(name));

	/*
	 * We did not find the entry in the directory
	 */

	if (inum == 0) {
		return NULL;
	}

	/*
	 * XXX - Fileset initialization belongs in other lookup paths as
	 * well
	 */

	inode = I_GET_SHARED(snapnum, inum);
	fs_inodeRelease(context, parent);
	fset = fs_getFset(context, core, inode);
	FS_ASSERT(fset != NULL);
	inode->fset = fset;
	return inode;
}


/*
 * Create an empty file in the directory
 */

/*
 * XXX - Fileset initialization belongs in other lookup paths as
 * well
 *
 * TODO - uid, gid
 */

EXPORTED fs_Inode *
fs_create(
    fs_Context	*context,
    fs_FileSystem	*core,
    uint64		snapnum,	// snapshot number
    fs_Inode	*parent,	// parent inode
    char		*name,		// name of the file/dir to be created
    int			mode		// mode bits for the new file
)
{
	uint64				inum;
	fs_Inode			*inode;
	fs_FileSetIncoreRecord_t	*fset;
	int					namelen = strlen(name);
    int64               parentNumBlocks;
    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL, *list3 = NULL;

	FS_ASSERT(context != NULL && core != NULL && parent != NULL &&
		  name != NULL);
	FS_ASSERT(fs_lockOwnedExclusive(context, &parent->lock));

	inum = fs_dir_lookupent(context, parent, name, strlen(name));
    
	/*
	 * We did not find the entry in the directory
	 */
    
    if (inum == 0) 
    {
        fs_Time now;

        /*
         * Entry not found. Create it
         */
        
        FS_TRACE(context, FS_TRACE_REQ, ("parent link count is %d\n",
                                         parent->dinode.linkCnt));
        if (FS_I_IS_DIR(mode)) 
        {
            inode = fs_allocateDirectoryInode(context, parent, mode, 0, 0,
                                              parent->dinode.fsetNum);
        }
        else 
        {
            inode = fs_allocateFileInode(context, parent, mode, 0, 0,
                                         parent->dinode.fsetNum);
        }
        
        FS_TRACE(context, FS_TRACE_REQ, ("after parent link count is %d\n",
                                         parent->dinode.linkCnt));
        if (inode == NULL) 
        {
            fs_error(context, FS_INTERNAL_ERR, "ialloc");
            FS_ASSERT(0);
            return NULL;
        }
        FS_TRACE(context, FS_TRACE_REQ, ("Created inode %lld\n", inode->inum));
        FS_ASSERT(parent->fset != NULL);
        inode->fset = parent->fset;
        
         //Add this inode in parent filesets usage
        fs_QuotaAdjustmentNoCheck(  context,
                                    context->fileSystem,
                                    parent,
                                    inode,
                                    1,
                                    FS_COUNT_INODES,
                                    0,
                                    0,
                                    &list1);
       
        if (FS_I_IS_DIR(mode))  {
            //Add the first block of the directory in parent filesets usage
            fs_QuotaAdjustmentNoCheck(  context,
                                        context->fileSystem,
                                        parent, 
                                        inode,
                                        1,
                                        FS_COUNT_BLOCKS,
                                        0,
                                        0,
                                        &list2);
        }
            
        // Initialize the inode extentmap root
        
        FS_TRACE(context, FS_TRACE_REQ,
                 ("after dirrot init - parent link count is %d\n",
                  parent->dinode.linkCnt));
        
        // We can never fail for resources.
        parentNumBlocks = parent->dinode.numblocks;
        fs_dir_addent(context, parent, name, namelen, inode->inum);
        FS_TRACE(context, FS_TRACE_REQ,
                 ("after dirent add - parent link count is %d\n",
                  parent->dinode.linkCnt));
        FS_ASSERT(FS_I_IS_DIR(inode->dinode.attrs));
                
        //Check if directory size has changed
        parentNumBlocks = parent->dinode.numblocks - parentNumBlocks;
        if (parentNumBlocks != 0)   
        {
            fs_QuotaAdjustmentNoCheck(  context,
                                        context->fileSystem,
                                        parent,
                                        NULL,
                                        parentNumBlocks,
                                        FS_COUNT_BLOCKS,
                                        parent->sd->u.nfs.uid,
                                        parent->sd->u.nfs.gid,
                                        &list3);
        }
        
        if (FS_I_IS_DIR(mode)) {
            parent->dinode.linkCnt += 1;
        }

        now = fs_getTimeOfDay();
        parent->dinode.ctime = now;
        parent->dinode.mtime = now;
        parent->dinode.atime = now;
        fs_inodeWrite(context, parent, INO_WRITE_LOG);

        /*
         * If we did any quota reservations, commit them.
    	 * We commit the directory size change list first since it is less 
         * likely to exist. Therefore, we prevent the unnecessary hold of quota
         * inodes.
         */
        if (list3) {
            fs_commitContextList(context, context->fileSystem, list3, FALSE);
            fs_freeQuotaContextList(context, context->fileSystem, list3);
        }       
        if (list1) {
            list1->newInode = inode;
            fs_commitContextList(context, context->fileSystem, list1, TRUE);
            fs_freeQuotaContextList(context, context->fileSystem, list1);
        }       
        if (list2) {
            fs_commitContextList(context, context->fileSystem, list2, FALSE);
            fs_freeQuotaContextList(context, context->fileSystem, list2);
        }       
	}
    else {
		inode = I_GET_EXCL(snapnum, inum);

		/*
		 * We can never fail for resources. On I/O error the
		 * filesystem would have been disabled or machine paniced
		 */

		FS_ASSERT(inode != NULL);
		fset = fs_getFset(context, core, inode);
        FS_ASSERT(fset != NULL);
        inode->fset = fset;
        FS_ASSERT(inode->dinode.attrs != 0);
    }
    return inode;
    
}


/**
 * Update inode access time.
 */
void fs_inodeAtimeUpdate(
    fs_Context  *context,
    fs_Inode    *inode)
{
    fs_Time now;
#define FS_ATIME_MIN_CHANGE_SECS    (15 * 60)

    if (! fs_incoreTune->enableAtimeUpdate)
    {
        return;
    }

    if ((context->fileSystem->flags & FS_IGNORE_ATIME_UPDATE) ||
        inode->snapnum != 0)
    {
        return;
    }

    now = fs_getTimeOfDay();
    if (now.secs >= inode->dinode.atime.secs &&
        now.secs - inode->dinode.atime.secs < FS_ATIME_MIN_CHANGE_SECS)
    {
        // Don't bother to do this unless the time is
        // significantly different.
        return;
    }

    FS_ASSERT(fs_lockHeldShare(context, &inode->lock) ||
              fs_lockHeldExclusive(context, &inode->lock));
    NFX_SPIN_LOCK(&inode->spin);
    inode->dinode.atime = now;
    fs_inodeMarkDirty(context, inode);
    NFX_SPIN_UNLOCK(&inode->spin);
}


/**
 * Insert inum,snapnum into the cache and return the inode pointer
 *
 * Called during log replay to get an inode log record into the cache
 * without using the inode file from disk
 */
fs_Inode *fs_inodeInsertForReplay(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum,
    int32           lockMode)
{
    fs_Inode        *inode;
    fs_FileSystem   *fileSystem = context->fileSystem;
    uint32          numRetries = 0;

retry:

    NFX_SPIN_LOCK(&fs_inodeCache->spin);

    // check if already in cache

    inode = fs_inodeCacheGet(context, snapnum, inum, lockMode);
    FS_ASSERT(inode != (fs_Inode *) INO_NO_LONGER_VALID);
    if (inode)
    {
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        return inode;
    }

    // if not, allocate and insert into cache

    inode = fs_inodeAllocIncore(context, inum);
    if (inode == NULL)
    {
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
        fs_inodeWaitAvail(context, &numRetries);
        goto retry;
    }
    inode->snapnum = snapnum;
    inode->fileSystem = fileSystem;
    inode->flags = INO_FLAGS_REPLAY_INSERT;
    inode->zfflags = 0;
    inode->cacheflags = 0;

    fs_inodeCacheInsert(context, inode);
    NFX_SPIN_UNLOCK(&fs_inodeCache->spin);

    // fs_inodeAllocIncore() aquires exclusive lock demote if necc

    if (lockMode & FS_LOCK_MODE_SHARE)
    {
        fs_lockDemote(context, &inode->lock, FS_LOCK_MODE_SHARE);
    }

    return inode;
}




#ifdef FS_DEBUG_INODE_COUNTS

/**
 * Check that the inode counts are correct.
 */
void fs_inodeCheckCounts(
    fs_Context *context)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint64 inum;
    uint64 count = 0;

    for (inum = 0; inum < params->maxInodeNumber; inum++) {
        fs_Inode *inode = I_GET_SHARED(0, inum);
        if (inode != NULL) {
            fs_Dinode *dinode = inode->dinode;
            char data[(FS_INODE_INLINE_LEN * 2) + 1];
            int i;

            for (i = 0; i < FS_INODE_INLINE_LEN; i++)
                sprintf(&data[i * 2], "%02x", dinode->u.data[i] & 0xff);
            data[FS_INODE_INLINE_LEN * 2] = '\0';
            printf("%6lld: size %6lld  attrs %o   %s\n", inum,
                   dinode->size, dinode->attrs, data);
            fs_inodeRelease(context, inode);
            count++;
        }
    }
    printf("%lld inodes allocated (of %lld)\n", count, params->maxInodeNumber);
}

#endif // FS_DEBUG_INODE_COUNTS

#endif // SSC

