// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-inode.h
 *
 * RCS:  $Id: fs-inode.h,v 1.86 2002/11/26 22:53:59 bob Exp $
 *
 * Description:  Data structures for inode and related items.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_INODE_H)
#define FS_INODE_H


/**
 * Parent_inum_level_nkeys is a bitfield: the low 56 bits are the inode number,
 * and the high 8 bits are 4 bits each of level and nkeys:
 *
 * <63 level 60> <59 nkeys 56> <55 inum 0>
 */
#define FS_DIRROOT_MAXKEYS 3
typedef struct fs_Dirroot
{
    uint64 parent_inum_level_nkeys;
    uint64 right;
    uint64 left[FS_DIRROOT_MAXKEYS];
    uint32 hash[FS_DIRROOT_MAXKEYS];
    uint32 spare;

} fs_Dirroot;


/**
 * Root node of inode B+ tree, housed in the inode itself
 */
typedef struct fs_Emaproot_hdr { /* header common to btree and flat roots. */
    int32 nkeys;
    int32 level;
} fs_Emaproot_hdr;

#define FS_INOROOT_MAXKEYS 2
typedef struct fs_Emaproot
{
    fs_Emaproot_hdr hdr;
    uint64 offsets[FS_INOROOT_MAXKEYS];
    uint64 left[FS_INOROOT_MAXKEYS];
    uint64 right;
} fs_Emaproot;


/**
 * Root node of inode B+ tree when it's both the root and the leaf.
 */

#define FS_INOFLATROOT_MAXKEYS 2
typedef struct fs_Emapflatroot
{
    fs_Emaproot_hdr hdr;
    uint64 offsets[FS_INOFLATROOT_MAXKEYS];
    fs_Extent extents[FS_INOFLATROOT_MAXKEYS];
    uint32 spare[2];
} fs_Emapflatroot;

#define FS_INOROOT_MAXLEVEL 8 // maximum height of a btree


#define FS_DINODE_SIZE 128

/**
 * The on-disk inode.
 */
struct fs_Dinode
{
    uint64      size;     // Size of file in bytes
    ushort16    attrs;    // Format and simple attribute fields.  Remainder
                          // of 'mode' bits are in the security file.
#define FS_IFMT_MASK              0170000
#define FS_IFMT_SOCK              0140000
#define FS_IFMT_SYML              0120000
#define FS_IFMT_REG               0100000
#define	FS_IFMT_QUOT              0110000
#define FS_IFMT_BLK               0060000
#define FS_IFMT_DIR               0040000
#define FS_IFMT_CHR               0020000
#define FS_IFMT_FIFO              0010000

#define FS_IATR_MASK              0007000
#define FS_IATR_SUID              0004000
#define FS_IATR_SGID              0002000
#define FS_IATR_DNC               0001000

#define FS_IFLG_MASK              0000777
#define FS_IFLG_NO_SNAP           0000001 // don't snapshot
#define FS_IFLG_DIR_SPLIT         0000002 // dir has split
//#define FS_IFLG_UNUSED            0000004 // not currently used
#define FS_IFLG_COMPACT_EMAP      0000010 // emap is in compact leaf form
#define FS_IFLG_COMPACT_DIR       0000020 // dir is in compact form
#define FS_IFLG_VIRUS_MASK        0000140 
#define FS_IFLG_VIRUS_UNK         0000000 // virus status unknown
#define FS_IFLG_VIRUS_CLEAN       0000040 // file is virus free
#define FS_IFLG_VIRUS_INFECTED    0000100 // file is infected
#define FS_IFLG_VIRUS_UNDEF       0000140 // unused
#define FS_IFLG_EATTR             0000200 // inode has extended attributes
//#define FS_IFLG_UNUSED            0004000 // not currently used
#define FS_IFLG_SNAP_PIN          0000120 // snapbits inode pinned

#define FS_I_IS_SYML(m) (((m) & FS_IFMT_MASK) == FS_IFMT_SYML)
#define FS_I_IS_REG(m)  (((m) & FS_IFMT_MASK) == FS_IFMT_REG)
#define FS_I_IS_DIR(m)  (((m) & FS_IFMT_MASK) == FS_IFMT_DIR)
#define FS_I_IS_CHR(m)  (((m) & FS_IFMT_MASK) == FS_IFMT_CHR)
#define FS_I_IS_BLK(m)  (((m) & FS_IFMT_MASK) == FS_IFMT_BLK)
#define FS_I_IS_FIFO(m) (((m) & FS_IFMT_MASK) == FS_IFMT_FIFO)
#define FS_I_IS_SOCK(m) (((m) & FS_IFMT_MASK) == FS_IFMT_SOCK)
#define	FS_I_IS_QUOT(m) (((m) & FS_IFMT_MASK) == FS_IFMT_QUOT)

#define	FS_I_IS_COMPACT_EMAP(m) ((m) & FS_IFLG_COMPACT_EMAP)
#define	FS_I_IS_COMPACT_DIR(m)  ((m) & FS_IFLG_COMPACT_DIR)

    ushort16    userAttr;
#define FS_INODE_UATTR_READONLY   0x0001 // User attr - read only
#define FS_INODE_UATTR_HIDDEN     0x0002 // User attr - hidden file
#define FS_INODE_UATTR_SYSTEM     0x0004 // User attr - system file
#define FS_INODE_UATTR_VOLUME     0x0008 // User attr - obselete
#define FS_INODE_UATTR_DIRECTORY  0x0010 // User attr - directory
#define FS_INODE_UATTR_ARCHIVE    0x0020 // User attr - archive
#define FS_INODE_UATTR_DEVICE     0x0040 // User attr - device
#define FS_INODE_UATTR_NORMAL     0x0080 // User attr - normal
#define FS_INODE_UATTR_TEMPORARY  0x0100 // User attr - temporary
#define FS_INODE_UATTR_SPARSE     0x0200 // User attr - sparse
#define FS_INODE_UATTR_REPARSE_PT 0x0400 // User attr - special symlink
#define FS_INODE_UATTR_COMPRESSED 0x0800 // User attr - data compressed
#define FS_INODE_UATTR_OFFLINE    0x1000 // User attr - offline storage
#define FS_INODE_UATTR_NCI        0x2000 // User attr - non content indexed
#define FS_INODE_UATTR_ENCRYPTED  0x4000 // User attr - offline storage

    uint32      linkCnt;    // Number of links to this file/directory
#define FS_INODE_LINK_MAX   UINT32_MAX

    fs_Time     ctime;
    fs_Time     mtime;
    fs_Time     atime;

    uint64      numblocks;  // 	Number of blocks allocated.
    union {
        uint64      parentInum; // Inode number of parent for single link files
        uint64      indexInum;  // Inode number of directory index file
    } piu;
    uint32      genNum;     // Inode generation number
#define FS_INODE_GENNUM_SNAP_MAGIC  0x80000000 // upper bit of gennum reserved 
#define FS_INODE_GENNUM_SNAP_MAGIC2 0x40000000 // another bit of gennum reserved
    uint32      securityId; // 32-byte offset in security file for security desc.
    uint32      fsetNum;    // fset number
    uint32      version;    // monotonically incr for inode/data changes for
                            // dump
    uint32      spare[2];

#define FS_INODE_INLINE_LEN         48
    union {
        char8                   data[FS_INODE_INLINE_LEN];
        fs_DevNum               devNum;
        fs_Emaproot             emaproot;
        fs_Emapflatroot         emapflatroot;
    } u;
};

NFX_ASSERT_STATIC(sizeof (fs_Dinode) == FS_DINODE_SIZE);

#define FS_INODES_PER_PAGE  (FS_BLOCK_SIZE / sizeof(fs_Dinode))

#define FS_INODE_PAGE_OFFSET(_inum_)                            \
    (((_inum_) % FS_INODES_PER_PAGE) * sizeof(fs_Dinode))


/**
 * Inode flags.
 */
enum {
    INO_FLAGS_TRANS_WRITE       = 0x001, // Current transaction modified
                                         // inode
    INO_FLAGS_DIRTY             = 0x002, // Inode is dirty in-core

    INO_FLAGS_TRANS_LOCK        = 0x004, // Transaction locked inode until
                                         // commit.
    INO_FLAGS_BTREE_TRANS_LOCK  = 0x008, // Transaction locked btree until
                                         // commit.
    INO_FLAGS_PENDING_DEALLOC   = 0x010, // Inode pending deallocation.
                                         // Will occur at commit.
    INO_FLAGS_WAIT_LOGBUF       = 0x020, // Inode cleaner is blocked awaiting
                                         // pending log bufs to be flushed to
                                         // disk.
    INO_FLAGS_FREE_LOGBUF       = 0x040, // Inode is pending free operation,
                                         // waiting for all log bufs that hold
                                         // this inode to flush to disk.
    INO_FLAGS_REPLAY_INSERT     = 0x080, // Inode was inserted in cache during
                                         // replay (debug helper).
    INO_FLAGS_RESTORE_IN_PROGRESS = 0x100,// restore is using this inode,
                                          // don't reclaim
};

enum {
    INO_FLAGS_MRD               = 0x1, // inode is on mrd list
};

enum {
    INO_FLAGS_UZERO             = 0x1, // inode needs user zero fill
};

/**
 * Flags that can be set in an inode yet the inode
 * remains reclaimable.
 */
#define INO_FLAGS_RECLAIM_OK    (INO_FLAGS_REPLAY_INSERT |      \
                                 INO_FLAGS_RESTORE_IN_PROGRESS| \
                                 INO_FLAGS_UZERO)



/**
 * Stuff for tracking read-ahead of an inode.
 */
typedef struct fs_PendingRead fs_PendingRead;
struct fs_PendingRead
{
    DLL_ENTRY(fs_PendingRead)       list;       // list of reads
    uint64                          offset;     // read offset
    uint32                          len;        // read len
    esm_handle_t                    smHandle;
    uint8                           sequential; // whether we have to perform
                                                // sequential i/o or not
};
typedef struct fs_ReadAheadState fs_ReadAheadState;
struct fs_ReadAheadState
{
    uint64                  erOffset;   // expected next read offset
    uint64                  raOffset;   // read ahead offset
    uint64                  raTrigger;  // Read ahead when this point reached
    uint32                  raSize;     // Number of blocks to read ahead
    uint8                   uxRead;     // Inspite of read ahead we had
                                        // to block for i/o.
    uint8                   nPendingReads;  // number of pending reads
    DLL_HDR(fs_PendingRead) pendingReadList;    // pending read list hdr
};


/**
 * An in-core inode.
 */
struct fs_Inode
{
    fs_Lock                 lock;
    fs_FileSystem           *fileSystem;
    uint64                  inum;
    DLL_ENTRY(fs_Inode)     cacheEntry; // Entry for cache hash chain.
    DLL_ENTRY(fs_Inode)     mruEntry;   // Entry for inode MRU list (now freelist).
    DLL_ENTRY(fs_Inode)     mrdEntry;   // Entry for inode dirty list.
    DLL_ENTRY(fs_Inode)     cowEntry;   // Entry for inode cow list.
    fs_Lock                 btreeLock;	// Lock for extent maps in the Dinode.
    nfx_spinlock_t          spin;
    esm_handle_t            cleanerHandle; // Handle of cleaner thread that
                                        // is currently blocked on this inode.
    int32                   logRefCount;// Count of committed transactions in
                                        // log buffers not yet on disk that
                                        // have mods for this inode.
    fs_SecurityDescCore     *sd;        // Security descriptor.
    fs_FileSetIncoreRecord_t	*fset;	// fileset
    fs_LogSegmentDeps       logSegDep;  // Log dependencies
    fs_ReadAheadState       *readAheadState; // Valid if read-ahead started.
    int16                   headLogSeg; // Log segment number this inode first
                                        // resides at.
    ushort16                snapnum;    // Snapshot this inode is part of.
    ushort16                flags;      // all protected by inode->spin
    uint8                   zfflags;    // protected by fileSystem->zfSpin->spinlock
    uint8                   cacheflags; // protected by fs_inodeCache->spinlock
    fs_Dinode               dinode;     // The disk inode.
};


/**
 * Initialize the inode cache.
 */
extern void
fs_inodeCacheInit();


/**
 * Called to get and lock an inode in the cache.
 */
extern fs_Inode *fs_inodeGet(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum,
    int32           lockMode);


/**
 * Return a buffer from the inode file, with a shared lock, given the
 * inode number.
 */
extern fs_Buf *fs_inodeGetBufferShared(
    fs_Context  *context,
    fs_Inumber  inumber,
    int         snapnum,
    boolean     create,
    boolean     isTryLock);


/**
 * Return a buffer from the inode file, with an exclusive lock, given the
 * inode number.
 */
extern fs_Buf *fs_inodeGetBufferExclusive(
    fs_Context  *context,
    fs_Inumber  inumber,
    int         snapnum,
    boolean     create,
    boolean     isTryLock);


/**
 * Called to unlock and release an inode, held with a shared lock, back
 * to the cache.
 */
extern void fs_inodeRelease(
    fs_Context      *context,
    fs_Inode        *inode);


/**
 * Called to unlock and release an inode, held with a exclusive lock,
 * back to the cache. Give up all the exclusive holds on this lock
 *
 * Should normally only be used if inode was NOT modified.  If inode was
 * modified, fs_inodeWrite() should be used (eventually followed by
 * fs_logCommitEntry()).
 */
extern void fs_inodeReleaseExclusive(
    fs_Context      *context,
    fs_Inode        *inode);

extern void fs_inodeReleaseExclusiveAll(
    fs_Context      *context,
    fs_Inode        *inode);

/**
 * Inode write options.  (Not declared as an enum type, because they can be
 * combined.)
 */
enum {
    INO_WRITE_LOG	= 0x0001,
    INO_WRITE_BTREE	= 0x0002,
};


/**
 * Called to inidicate that an inode has been written. Takes care of all
 * COW and logging concerns.
 */
extern void fs_inodeWrite(
    fs_Context      *context,
    fs_Inode        *inode,
    uint32          flags);


/**
 * Adds an inode lock to the list of locked objects for
 * the current transaction.  It will be unlocked when the
 * transaction commits.
 */
extern void
fs_inodeLockTransaction(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_LogSummary   *logSummary);


/**
 * Return true if the inode has been dirtied by the current transaction.
 * The current context must hold the inode exclusively.
 */
extern boolean fs_inodeWasModified(
    fs_Context *context,
    fs_Inode *inode);


/**
 * Return true if the inode btree has been dirtied by the current transaction.
 * The current context must hold the inode btree lock exclusively.
 */
extern boolean fs_inodeBtreeWasModified(
    fs_Context *context,
    fs_Inode *inode);


/**
 * Called during transaction commit to copy the
 * specified inode to the log.
 */
extern void
fs_inodeUpdateLog(
    fs_Context  *context,
    fs_Inode    *inode);


/**
 * Called at transaction commit to unlock the specified inode.
 */
extern void
fs_inodeCommit(
    fs_Context      *context,
    fs_Inode        *inode);

/**
 * Called at transaction commit to unlock the specified inode btree.
 */
extern void
fs_inodeBtreeCommit(
    fs_Context      *context,
    fs_Inode        *inode);


/**
 * Called to invalidate an inode from the cache.
 */
extern void
fs_inodeRemoveFromCache(
    fs_Context      *context,
    fs_Inode        *inode,
    boolean         leaveOnMru);


/**
 * Flush the contents of the inode cache for the
 * specified volume and snapNum.
 */
extern void
fs_inodeCacheFlushSnap(
    fs_Context *context,
    int32       snapNum);


/**
 * Flush the contents of the inode cache for the
 * specified volume.
 */
static inline void
fs_inodeCacheFlush(
    fs_Context *context)
{
    fs_inodeCacheFlushSnap(context, 0);
}


/**
 * Clean all of the specified filesystem's inodes and
 * wait for it to complete.  The sync is done in the
 * callers thread.
 */
extern void
fs_inodeSyncCache( 
    fs_Context *context);


/**
 * Clean all of the specified filesystem's inodes and
 * wait for it to complete.  A new thread is created
 * to accomplish the sync.
 */
extern void
fs_inodeSyncCacheSpawnNew(
    fs_Context *context);


/**
 * Signal the inode cleaner thread to wakeup and clean
 * inodes for the specified filesystem.
 */
enum {
    FS_INODE_CLEANER_INVALID,
    FS_INODE_CLEANER_LOG,
    FS_INODE_CLEANER_ALL,
    FS_INODE_CLEANER_START,
    FS_INODE_CLEANER_REFDECR,
    FS_INODE_CLEANER_PAD,
    FS_INODE_CLEANER_MAX,
};

// Must be even in order to avoid padding problem in
// cachestats response.
NFX_ASSERT_STATIC(FS_INODE_CLEANER_MAX % 2 == 0);

extern int32 fs_inodeCleanerCount[FS_INODE_CLEANER_MAX];
extern int32 fs_inodeCleanerMaxPerLoop;

extern void
fs_inodeSignalCleaner(
    fs_Context      *context,
    uint32          type);


/**
 * Truncate a file to a newly specified size.
 */
extern void fs_truncateFile(
    fs_Context              *context,
    fs_Inode                *inode,
    uint64	                newSize,
	boolean		            dontTrashCan,
    fs_QuotaContextList_t   **rlist);
   

/**
 * Update the inode's atime. If ignore atime is set on fs
 * then ignore it
 */
extern void fs_inodeAtimeUpdate(
    fs_Context  *context,
    fs_Inode    *inode);


/**
 * Return true if the specified inode is marked "no snapshot".  (Also
 * returns true for a null inode.)
 */
extern boolean fs_inodeIsNoSnap(
    fs_Context *context,
    fs_Inode *inode);

/**
 * Print some of the contents of an in-core inode.
 */
extern void
fs_inodeDump(
    char     *str,
    fs_Inode *inode);


/**
 * Show all locked inodes in the cache.
 */
extern void
fs_inodeShowLocked();


#ifdef FS_DEBUG_INODE_COUNTS
/**
 * Check that the inode counts are correct.
 */
extern void fs_inodeCheckCounts(
    fs_Context *context);
#endif


// Get inode with shared lock, sleep if necessary.
#define I_GET_SHARED(_snap_,_inum_)                 \
    fs_inodeGet(    context,                        \
                    (_snap_),                       \
                    (_inum_),                       \
                    FS_LOCK_MODE_SHARE|FS_LOCK_MODE_WAIT)

// Get inode with shared trylock, sleep if I/O necessary.
#define I_GET_TRY_SHARED(_snap_,_inum_)             \
    fs_inodeGet(    context,                        \
                    (_snap_),                       \
                    (_inum_),                       \
                    FS_LOCK_MODE_SHARE)

// Get inode with exclusive lock, sleep if necessary.
#define I_GET_EXCL(_snap_,_inum_)                   \
    fs_inodeGet(    context,                        \
                    (_snap_),                       \
                    (_inum_),                       \
                    FS_LOCK_MODE_EXCL|FS_LOCK_MODE_WAIT)

// Get inode with exclusive trylock, sleep if I/O necessary.
#define I_GET_TRY_EXCL(_snap_,_inum_)               \
    fs_inodeGet(    context,                        \
                    (_snap_),                       \
                    (_inum_),                       \
                    FS_LOCK_MODE_EXCL)

/**
 * Destroy the contents of the inode cache for the
 * specified volume without trying to sync to disk,
 * and without regard to inode locks.
 *
 * This is called when a volume is being forcibly
 * unmounted.
 */
void
fs_inodeCacheDestroyVolume(
    fs_Context      *context,
    fs_FileSystem   *fileSystem);


/**
 * Inode cleaner thread.  There is one of these running per mounted
 * filesystem, as well as possibly additional ones spawned whenever
 * a thread desires to block on sync'ing a filesystem's inodes out to
 * disk.
 */
extern void 
fs_inodeCleanerThread(
    esm_handle_t    handle,
    void            *arg);


/**
 * A thread to grow the inode file.
 */
extern void
fs_igrowThread(
    esm_handle_t    handle,
    void            *arg);

/**
 * Thread for re-counting inode usage after a crash.
 */
void
fs_inodeCountThread(
    esm_handle_t    handle,
    void            *arg);


/**
 * Increment the log buffer reference
 * count on the specified inode.
 */
extern void
fs_inodeLogRefIncr(
    fs_Context  *context,
    fs_Inode    *inode);


/**
 * Decrement the log buffer reference
 * count on the specified inode.
 */
extern void
fs_inodeLogRefDecr(
    fs_Context  *context,
    fs_Inode    *inode);


/**
 * Insert inum,snapnum into the cache and return the inode pointer
 *
 * Called during log replay to get an inode log record into the cache
 * without using the inode file from disk
 */
extern fs_Inode *fs_inodeInsertForReplay(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum,
    int32           mode);


/**
 * Instantiate the newly allocated inode into the
 * cache.
 */
extern fs_Inode *
fs_inodeInstantiateAllocator(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum,
    fs_Buf          *buf);


/**
 * Instantiate a meta inode in the cache.
 */
extern fs_Inode *
fs_inodeInstantiateMetanode(
    fs_Context      *context,
    int32           snapnum,
    uint64          inum);


/**
 * Called from the allocator during the first phase
 * of deallocating an inode.
 */
extern void
fs_inodeDeallocStart(
    fs_Context  *context,
    fs_Inode    *inode);

/**
 * Return the percentage of the inode cache
 * that is currently available (not dirty).
 */
extern int32
fs_inodePercentAvail();

/**
 * Add all inodes back to the lists that might need COW.
 * Called after a snapshot operation (ie create).
 */
extern void
fs_inodeCacheEnableCow(
       fs_Context *context);


/**
 * Set the inode onto the dirty list for the specified
 * log segment.
 */
extern void
fs_inodeSetDirtySegment(
    fs_Context  *context,
    void        *obj,
    int32       segNum,
    int32       offset);

/**
 * Return TRUE if a dirty inode is potentially
 * holding up log truncation of the current head
 * log segment.
 */
extern boolean
fs_inodeBlockingLogTruncation(
    fs_Context      *context,
    fs_FileSystem   *fileSystem);

/**
 * Set the security descriptor for the specified inode
 * as indicated.  The caller must hold the inode lock in
 * exclusive mode.
 */
extern void
fs_securityDescSet(
    fs_Context         *context,
    fs_Inode           *inode,
    fs_SecurityDescHdr *hdr);


/**
 * Fetch the attributes for the specified inode if they're
 * in the cache.  Return TRUE if they were fetched, FALSE
 * otherwise.
 *
 * !!! This does not block. !!!
 */
extern boolean
fs_inodeGetAttrs(
    fs_FileSystem   *fileSystem,
    fs_Attr         *attrp,
    uint64          inum,
    int32           snapnum,
    fs_Proto        protocol,
    fs_AttrType     attrType,
    uint32          *genNum);

/**
 * free resource for inode cache
 *
 * Caller must hold spinlock.
 */
extern void
fs_inodeFreeResource(
    fs_Context          *context,
    int32               reduceCount);

/**
 * Set the upper limit on the inode cache size.
 */
extern void
fs_inodeSetCacheLimit();


#endif /* FS_INODE_H */
