// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-layout.h
 *
 * RCS:  $Id: fs-layout.c,v 1.10 2002/08/01 17:56:15 sandrineb Exp $
 *
 * Description:  File system layout.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-29
 */

#if !defined(SSC)

#include "fs-includes.h"



//   P A R A M E T E R S


/*
 * All of these parameters should probably be in the fs_info block,
 * and variable rather than fixed.
 */



//   P A R A M E T E R S


/**
 * Turn a number into a shorter version, using K, M, G, T, or P suffixes,
 * if it's a multiple of 1024^n.
 *
 * WARNING: NOT MP SAFE, NOT REENTRANT
 */
static char *clean(uint64 n)
{
    if (n == 0)
        return "";
    else {
        int p = 0;
        char suffixes[] = "KMGTP";
        static char buf[40];
        while (p < (sizeof suffixes) && (n % 1024) == 0) {
            p++;
            n /= 1024;
        }
        if (p == 0)
            return "";
        sprintf(buf, " (%lld%c)", n, suffixes[p - 1]);
        return buf;
    }
}


/**
 * For debugging purposes, print the FS parameters.
 */
void fs_printParameters(fs_Parameters *params)
{
#define P(p) \
    printf("# %-22s %16lld%s\n", #p ":", (uint64) params->p, clean(params->p))

    P(totalBlocks);
    P(totalBytes);
    P(blockSize);
    P(wordsPerBlock);
    P(blockBytesShift);
    P(blockBytesMask);
    P(blockBitsShift);
    P(blockInodeRatio);
    P(inodesPerBlock);
    P(inodeFileStride);
    P(maxInodeBlocks);
    P(maxInodeNumber);
    P(inodeSummaryStride);
    P(inodeSummaryCount);
    P(buddyRegionCount);
    P(refbytesBlockCount);

#undef P
}


/**
 * Compute the sizes of allocation structures used by the file system,
 * based on the number of blocks in the file system and the block size.
 */
fs_Parameters fs_computeSizeParameters(
    fs_SuperBlock *super)
{
    fs_Parameters p;

    // set the initial values

    p.totalBlocks          = super->numBlks;
    p.blockSize            = super->blkSize;
    p.blockInodeRatio      = super->blockInodeRatio;

    assert(isPowerOfTwo(p.blockSize));

    // compute the derived values

    p.wordsPerBlock
        = p.blockSize / sizeof (uint32);

    p.blockBytesShift
        = log2_32(p.blockSize);
    p.blockBytesMask
        = p.blockSize - 1;
    p.blockBitsShift
        = p.blockBytesShift + log2_32(BITS_PER_BYTE);

    p.totalBytes
        = p.totalBlocks * p.blockSize;
    p.inodesPerBlock
        = p.blockSize / sizeof (fs_Dinode);
    p.inodeFileStride
        = p.blockInodeRatio * p.inodesPerBlock;
    p.maxInodeBlocks
        = MAX(p.totalBlocks / p.inodeFileStride, 2);
    p.maxInodeNumber
        = p.maxInodeBlocks * p.inodesPerBlock;

    p.inodeSummaryStride
        = p.inodeFileStride * p.blockSize * BITS_PER_BYTE;
    p.inodeSummaryCount
        = ceilingDiv64(p.totalBlocks, p.inodeSummaryStride);

    p.buddyRegionCount
        = ceilingDiv64(p.totalBlocks, FS_BLOCKS_PER_BUDDY_REGION);

    // Compute enough refByte blocks to cover the sum of all buddy
    // regions, this makes growfs easier.
    p.refbytesBlockCount 
        = p.buddyRegionCount * (FS_BLOCKS_PER_BUDDY_REGION / p.blockSize);

    // fs_printParameters(&p);
#ifdef ISABELLE_STANDALONE
    printf("\nfs_computeSizeParameters()\n");
    fs_printParameters(&p);
#endif // ISABELLE_STANDALONE

    return p;
}


/**
 * Return the number of superblock copies found in a file system.
 */
int32 fs_superBlockCount()
{
    return 2;
}


/**
 * Return the block for the nth copy of the superblock, where blocks are
 * of the specific block size.  The zeroth copy is the primary superblock.
 */
fs_Block fs_superBlockBlock(uint32 copy)
{
    switch (copy) {
        case 0:
            return fs_makeBlock(0);
        case 1:
            return fs_makeBlock(1 + FS_MAX_SNAPSHOTS);
        default:
            assert(copy < 2);
    }
    UNREACHED();
}


/**
 * Return the block for the first block of the inode file in the
 * specified snapshot (where the live file system is snapshot zero).
 */
fs_Block fs_firstInodeBlock(uint32 snapshot)
{
    return fs_makeBlock(1 + snapshot);
}
#endif // SSC
