// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-layout.h
 *
 * RCS:  $Id: fs-layout.h,v 1.33 2002/11/26 22:53:59 bob Exp $
 *
 * Description:  File system layout-related definitions.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-07-19
 */

#if !defined(FS_LAYOUT_H)
#define FS_LAYOUT_H



//   P A R A M E T E R S


/**
 * Various layout parameters for the file system, computed from the
 * size of the file system and the block size.
 *
 * Perhaps these should all move into the superblock?
 */
typedef struct fs_Parameters {

    // basic inputs
    
    /** The size of the file system, in blocks. */
    uint64 totalBlocks;

    /** The size of a block, in bytes. */
    uint32 blockSize;

    /** The target block:inode ratio. */
    uint32 blockInodeRatio;


    // derived values

    /** The number of uint32 words that can be stored in a block. */
    uint32 wordsPerBlock;

    /** Number of bits to shift a block number to make it a byte offset. */
    uint32 blockBytesShift;

    /** Mask to get a byte offset within a block. */
    uint32 blockBytesMask;

    /** Number of bits to shift a block number to make it a bit offset. */
    uint32 blockBitsShift;

    /** The size of the file system, in bytes. */
    uint64 totalBytes;

    /** The number of inodes we can fit in a block. */
    uint32 inodesPerBlock;

    /** The number of blocks covered by one inode file block. */
    uint32 inodeFileStride;

    /** The number of blocks dedicated to inodes, if all inode blocks
     *  are allocated, using the target block:inode ratio. */
    uint32 maxInodeBlocks;

    /** The maximum inode number for the target block:inode ratio. */
    uint64 maxInodeNumber;

    /** The number of blocks covered by one full- or directory-in-inode
     *  summary block, using the target block:inode ratio. */
    uint32 inodeSummaryStride;

    /** The number of each type of inode summary block. */
    uint32 inodeSummaryCount;

    /** The number of blocks of reference bytes in the system. */
    uint32 refbytesBlockCount;

    /** The number of binary buddy tree regions. */
    uint32 buddyRegionCount;

} fs_Parameters;


/**
 * Compute the sizes of allocation structures used by the file system,
 * based on the number of blocks in the file system and the block size.
 *
 * TODO: make blockInodeRatio and blocksPerRefCluster parameters, too.
 */
extern fs_Parameters fs_computeSizeParameters(
    fs_SuperBlock *super);


/**
 * For debugging purposes, print the FS parameters.
 */
extern void fs_printParameters(
    fs_Parameters *params);


/**
 * Reserved inode numbers.  All of these must fit in a single block, the
 * first block of the inode file, which (for 8K block and 256-byte inodes)
 * limits us to 32 inodes.
 *
 * Each snapshot contains a copy of the first block of the inode file,
 * with the data for that snapshot.  The file system is laid out with
 * the superblock first, then the first 
 */
enum {
    /** A bad inode number.  We could use this, if needed. */
    FS_BAD_INODE = 0,

    /** The inode of the inode file. */
    FS_INODE_INODE = 1,

    /** The inode of the root directory for the file system. */
    FS_ROOT_INODE = 2,

    /** The inode of the reference counts. */
    FS_REF_COUNT_INODE = 3,

    /** Binary buddy tree of free blocks. */
    FS_FREE_BUDDY_TREE_INODE = 4,

    /** The inode of the security descriptor file. */
    FS_SECURITY_INODE = 5,

    /** The inode of the full inode block summary file. */
    FS_FULL_INODE_BLOCK_INODE = 6,

    /** The inode of the directory in inode block summary file. */
    FS_DIRECTORY_IN_INODE_BLOCK_INODE = 7,

    /** The inode of the fileset info file */
    FS_FILESET_INFO_INODE = 8,

    /** Default fileset user quota inode */
    FS_FILESET_USER_QUOTA_INODE = 9,

    /** Default fileset group quota inode */
    FS_FILESET_GROUP_QUOTA_INODE = 10,

    /** Null inode, returned for snapshot dirents for
     ** files that dont exist in the snapshot.  */
    FS_NULL_INODE = 11,

    /** Trashcan inode. */
    FS_TRASHCAN_INODE = 12,

    /** Directory of snapshot names. */
    FS_SNAPSHOT_DIR_INODE = 13,

    /** Bitmap of blocks contained in a snapshot. */
    FS_SNAPSHOT_BITS_INODE = 14,

    /** Links created (used to accelerate dump). */
    FS_LINK_INODE = 15,

    /** Temp Dir to create temporary root while restore */
    FS_TEMPDIR_INODE = 16,

    /** Temp swap Dir **/
    FS_TEMP_DUMP_SWAP = 17,

    /** Audit file **/
    FS_AUDIT_INODE = 18,

    /** System hidden directory **/
    FS_SYSTEM_HIDDEN_INODE = 19,

    /** Extended attributes/mapping file **/
    FS_EXTENDED_ATTR_INODE = 20,

    /** Reserved for future support of > 64 snapshots */
    FS_REF_OVERFLOW_INODE = 21,

    /** Directory for CIFS shares */
    FS_CIFS_SHARES_INODE = 22,

    /** Maximum used meta inode number. */
    FS_MAX_USED_META_INODE,

    /** Maximum meta inode number must be below this. */
    FS_MAX_META_INODE = 64,

    /** Inode numbers 64 through 127 are reserved for snapshot id's. */
    FS_SNAPSHOT_ID_INODE_FIRST = 64,
    FS_SNAPSHOT_ID_INODE_LAST = 127,

    /** This and above can be dynamically allocated. */
    FS_MAX_RESERVED_INODE = 128,
};


/**
 * Reserved snapshot numbers.
 */
enum {
    /** The snapshot id of the live filesystem.  (Much code assumes
     *  that this is zero.) */
    FS_LIVE_SNAP = 0,

    /** The snapshot id of the snapshot directories. */
    FS_SNAPDIR_SNAP = 1,

    /** The maximum number of snapshots for a file system, including the
     *  reserved ids. */
    FS_MAX_SNAPSHOTS = 64
};

NFX_ASSERT_STATIC(FS_MAX_SNAPSHOTS <=
              (FS_SNAPSHOT_ID_INODE_LAST - FS_SNAPSHOT_ID_INODE_FIRST) + 1);



//   S U P E R B L O C K S


/**
 * Return the number of superblock copies found in a file system.
 */
extern int32 fs_superBlockCount();


/**
 * Return the block for the nth copy of the superblock, where blocks are
 * of the specific block size.  The zeroth copy is the primary superblock.
 */
extern fs_Block fs_superBlockBlock(uint32 copy);


/**
 * Return the block for the first block of the inode file in the
 * specified snapshot (where the live file system is snapshot zero).
 */
extern fs_Block fs_firstInodeBlock(uint32 snapshot);


#endif /* FS_LAYOUT_H */
