/*
 * Name:  fs-lock.c
 *
 * RCS:  $Id: fs-lock.c,v 1.120 2002/11/27 17:39:25 janv Exp $
 *
 * Description:  File system locks
 * 
 * Created By:  Bob Fozard and Paul Haahr
 *
 * Date Created:  2001-07-24
 */

#if !defined(SSC)

#include "fs-includes.h"

#define LOCK_TRACE(args)   FS_TRACE(context, FS_TRACE_LCK, args)

#define REVERSE_OFFSET(_type_, _field_, _lock_) \
    ((_type_ *) (((char *) _lock_) - offsetof(_type_, _field_)))

#if defined FS_DEBUG_LOCK
// #define FS_KEEP_TRACK_OF_LOCKS(context)   (TRUE)
#define FS_KEEP_TRACK_OF_LOCKS(context)   (context->sharedLockSize != 0)
#else   // ! FS_DEBUG_LOCK
#define FS_KEEP_TRACK_OF_LOCKS(context)   (context->sharedLockSize != 0)
// #define FS_KEEP_TRACK_OF_LOCKS(context)   (FALSE)
#endif  // ! FS_DEBUG_LOCK        
        
boolean fs_doLockOrderChecking = FALSE;

// #define FS_DEBUG_CHECK_SHAREDLOCK

int fs_lockMaxSharedLockListSize = 0;
#ifdef FS_DEBUG_CHECK_SHAREDLOCK
#define FS_DEBUG_CHECK_SHAREDLOCK_MAX  1024
int fs_lockAddSharedLock = 0;
int fs_lockAddSharedLockGrow = 0;
int fs_lockSharedLockCount[FS_DEBUG_CHECK_SHAREDLOCK_MAX];
int fs_lockSharedLockGet[FS_DEBUG_CHECK_SHAREDLOCK_MAX];
int fs_lockSharedLockRm[FS_DEBUG_CHECK_SHAREDLOCK_MAX];
#endif

uint32 fs_lockWaitStats[FS_LOCK_OWNER_MAX]  // requestor req type
                        [FS_LOCK_TYPE_MAX]   // lock type
                        [FS_LOCK_OWNER_MAX]; // owner req type
fs_LockStatBuf fs_lockStatBuf;


//   S T A T I C   D A T A

//
// TODO - we should probably use the logId (transaction id) vs. the
// smHandle for the owner field of the lock, since the logId's are
// short-lived and some smHandle's live on for the life of a mount.
// (Need to make sure that non-intrusive ops obtain a transaction ID!)
//

/**
 * fs_lockCompat  -         AVAIL SHARE  EXCL
 *                   AVAIL   +      +     +
 *                   SHARE   +      +     -
 *                   EXCL    +      -     -
 */
static boolean fs_lockCompat[FS_LOCK_MODE_MAX][FS_LOCK_MODE_MAX] = {
    { TRUE,  TRUE,  TRUE  },
    { TRUE,  TRUE,  FALSE },
    { TRUE,  FALSE, FALSE }
};


/**
 * Names for lock modes.
 */
static char *fs_lockModes[FS_LOCK_MODE_MAX] = { "avail", "shared", "excl" };



//   L O C K   D E B U G G I N G


#ifdef FS_DEBUG_LOCK_ORDER

/**
 * Map from fs_LockType index to correct slot in
 * fs_LockOrderObjects[].
 */
int32 fs_lockOrderObjectMap[FS_LOCK_TYPE_MAX];


/**
 * Lock order for high level objects.  Note that TYPE_INODE
 * and TYPE_BUF don't really fit in the levels indicated,
 * they are handled with finer granularity in the lock order
 * code.
 */
fs_LockOrder fs_lockOrderObject[FS_LOCK_TYPE_MAX] =
{
    /* Type                             isLeaf      ignore*/
    { FS_LOCK_TYPE_INVALID,             FALSE,      FALSE },
    { FS_LOCK_TYPE_MODIFY,              FALSE,      FALSE },
    { FS_LOCK_TYPE_ACCESS,              FALSE,      FALSE },
    { FS_LOCK_TYPE_FREEZE,              FALSE,      FALSE },
    { FS_LOCK_TYPE_INODE,               FALSE,      FALSE },
    { FS_LOCK_TYPE_BUF,                 FALSE,      FALSE },
    { FS_LOCK_TYPE_FILESET,             FALSE,      FALSE },
    { FS_LOCK_TYPE_QUOTA,               FALSE,      FALSE },
    { FS_LOCK_TYPE_FILESET_TRAVERSAL,   FALSE,      FALSE },
    { FS_LOCK_TYPE_BTREE,               FALSE,      TRUE  },
    { FS_LOCK_TYPE_BOOT,                FALSE,      TRUE  },
};


/**
 * Map from inum to correct slot in
 * fs_LockOrderInode[].
 */
int32 fs_lockOrderInodeMap[FS_MAX_USED_META_INODE];


/**
 * Lock order for reserved inodes.
 */
fs_LockOrder fs_lockOrderInode[FS_MAX_USED_META_INODE] =
{
    /* Type                             isLeaf      ignore*/
    { FS_BAD_INODE,                     FALSE,      FALSE },
    { FS_ROOT_INODE,                    FALSE,      FALSE },
    { FS_TRASHCAN_INODE,                FALSE,      FALSE },
    { FS_FILESET_INFO_INODE,            FALSE,      FALSE },
    { FS_FILESET_USER_QUOTA_INODE,      FALSE,      FALSE },
    { FS_FILESET_GROUP_QUOTA_INODE,     FALSE,      FALSE },
    { FS_LINK_INODE,                    FALSE,      FALSE },
    { FS_TEMPDIR_INODE,                 FALSE,      FALSE },
    { FS_FULL_INODE_BLOCK_INODE,        FALSE,      FALSE },
    { FS_DIRECTORY_IN_INODE_BLOCK_INODE,FALSE,      FALSE },
    { FS_INODE_INODE,                   FALSE,      FALSE },
    { FS_FREE_BUDDY_TREE_INODE,         TRUE,       FALSE },
    { FS_REF_COUNT_INODE,               TRUE,       FALSE },
    { FS_SNAPSHOT_DIR_INODE,            FALSE,      TRUE },
    { FS_SNAPSHOT_BITS_INODE,           FALSE,      TRUE },
    { FS_INODE_FIVE_UNUSED,             FALSE,      TRUE },
    { FS_NULL_INODE,                    TRUE,       TRUE },
};


/**
 * Initialize the lock order structures.
 */
void
fs_lockOrderInit()
{
    int32   i, j;

    // Initialize the maps that allow us to directly
    // index into the more human-readable lock order
    // structures with the lock type enums.

    for (i = 0; i < FS_LOCK_TYPE_MAX; i++)
    {
        for (j = 0; j < FS_LOCK_TYPE_MAX; j++)
        {
            if (fs_lockOrderObject[j].type == i)
                fs_lockOrderObjectMap[i] = j;
        }
    }

    for (i = 0; i < FS_MAX_USED_META_INODE; i++)
    {
        for (j = 0; j < FS_MAX_USED_META_INODE; j++)
        {
            if (fs_lockOrderInode[j].type == i)
                fs_lockOrderInodeMap[i] = j;
        }
    }

} /* fs_lockOrderInit */


static void
fs_lockOrderUnlockInode(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_LockOrder    *objectOrder)
{
	fs_LockOrder	*inodeOrder, *ownedOrder;
	fs_Lock			*owned;
	fs_Inode		*ownedInode;
	int32			inodeHeight, ownedHeight;

	if (inode->inum >= FS_MAX_META_INODE)
		return;

    // Some inodes to ignore.  Something trying to do inode lookups
    // can cause these to be instantiated and locked for short
    // periods, so we need to trap them here. 
    if (inode->inum == 0 ||
        (inode->inum >= FS_MAX_USED_META_INODE &&
         inode->inum < FS_MAX_META_INODE))
    {
        return;
    }

    inodeHeight = fs_lockOrderInodeMap[inode->inum];
    FS_ASSERT(inodeHeight > 0 &&
              inodeHeight < FS_MAX_USED_META_INODE);

    inodeOrder = &fs_lockOrderInode[inodeHeight];

	if (inodeOrder->ignore)
		return;

    if (inodeOrder->isLeaf)
    {
        FS_ASSERT(context->leafHeld == inode);

        LOCK_TRACE(("order inode unlock inum %lld was leaf", 
                inode->inum));
        context->leafHeld = NULL;
        return;
    }

	context->maxReservedInodeHeight = 0;

    DLL_ITERATE(owned, &context->ownerList, ownerEntry)
    {
        if (owned->type != FS_LOCK_TYPE_INODE)
   			continue;

    	ownedInode = REVERSE_OFFSET(fs_Inode, lock, owned);
		if (ownedInode->inum == FS_ROOT_INODE ||
			ownedInode->inum >= FS_MAX_META_INODE)
		{
			continue;
		}

    	ownedHeight = fs_lockOrderInodeMap[ownedInode->inum];
    	FS_ASSERT(ownedHeight > 0 &&
              	  ownedHeight < FS_MAX_USED_META_INODE);

    	ownedOrder = &fs_lockOrderInode[ownedHeight];
    	if (ownedOrder->ignore)
        	continue;

		if (ownedHeight > context->maxReservedInodeHeight)
        {
            LOCK_TRACE(("unlock, new ino height is %d, inum %lld",
                   ownedHeight, ownedInode->inum));
			context->maxReservedInodeHeight = ownedHeight;
        }
	}

} /* fs_lockOrderUnlockInode */


/**
 * Clean up lock order checking related info for
 * releasing the specified lock.
 */
void
fs_lockOrderUnlock(
    fs_Context  *context,
    fs_Lock     *lock)
{
    fs_LockOrder    *objectOrder;
    int32           i, objectHeight;

    if (!fs_doLockOrderChecking)
        return;

	// No lock order checking during mkfs, etc.
	if (context->fileSystem->state != FS_VOL_STATE_ONLINE)
		return;

    if (context->ignoreLockOrder)
        return;

    objectHeight = fs_lockOrderObjectMap[lock->type];
    FS_ASSERT(objectHeight > 0 &&
              objectHeight < FS_LOCK_TYPE_MAX);

    objectOrder = &fs_lockOrderObject[objectHeight];
    if (objectOrder->ignore)
        return;

    NFX_SPIN_LOCK(&context->lockListSpin);

    if (objectOrder->isLeaf)
    {
        FS_ASSERT(context->leafHeld == lock);

        LOCK_TRACE(("order unlock object height %d was leaf",
                    objectHeight));

        context->leafHeld = NULL;
        NFX_SPIN_UNLOCK(&context->lockListSpin);
        return;
    }

    if (objectHeight < context->lockHeight)
    {
        // Release of this lock doesn't change
        // current position in lock heirarchy.
        NFX_SPIN_UNLOCK(&context->lockListSpin);
        return;
    }

    switch(lock->type)
    {
        case FS_LOCK_TYPE_BUF:
            // Nothing to do here.
            break;

        case FS_LOCK_TYPE_INODE: {
            fs_Inode *inode = REVERSE_OFFSET(fs_Inode, lock, lock);
            fs_lockOrderUnlockInode(context, inode, objectOrder);
            break;
        }
        case FS_LOCK_TYPE_QUOTA:
        case FS_LOCK_TYPE_FILESET:
        case FS_LOCK_TYPE_FILESET_TRAVERSAL:
        case FS_LOCK_TYPE_BOOT:
        case FS_LOCK_TYPE_MODIFY:
        case FS_LOCK_TYPE_ACCESS:
        case FS_LOCK_TYPE_FREEZE:
        case FS_LOCK_TYPE_BTREE: {
            fs_Lock         *owned, *shared;
            fs_LockOrder    *ownedOrder, *sharedOrder;
            int32           ownedHeight, sharedHeight;

            // Set the new height to the highest lock
            // still held.
            context->lockHeight = 0;
            DLL_ITERATE(owned, &context->ownerList, ownerEntry)
            {
                ownedHeight = fs_lockOrderObjectMap[owned->type];
                FS_ASSERT(ownedHeight > 0 &&
                          ownedHeight < FS_LOCK_TYPE_MAX);

                ownedOrder = &fs_lockOrderObject[ownedHeight];
                if (ownedOrder->ignore)
                    continue;

                if (ownedHeight > context->lockHeight)
                {
                    LOCK_TRACE(("order unlock object new height %d",
                            ownedHeight));
                    context->lockHeight = ownedHeight;
                }
            }
	        for (i = 0; i < context->sharedLockCount; i++)
            {
                shared = context->sharedLocksPtr[i].lock;
                sharedHeight = fs_lockOrderObjectMap[shared->type];
                FS_ASSERT(sharedHeight > 0 &&
                          sharedHeight < FS_LOCK_TYPE_MAX);

                sharedOrder = &fs_lockOrderObject[sharedHeight];
                if (sharedOrder->ignore)
                    continue;

                if (sharedHeight > context->lockHeight)
                {
                    LOCK_TRACE(("order unlock object new height %d",
                            sharedHeight));
                    context->lockHeight = sharedHeight;
                }
            }

            break;
        }

        default:
            fs_error(context, FS_INTERNAL_ERR, "lockorder: unknown type");
            break;
    }

    NFX_SPIN_UNLOCK(&context->lockListSpin);

} /* fs_lockOrderUnlock */


/**
 * Perform lock order checking on the specified
 * buf acquisition attempt.  Returns TRUE if a
 * violation occured.
 */
static boolean
fs_lockOrderLockBuf(
    fs_Context      *context,
    fs_Buf          *buf,
    fs_LockOrder    *objectOrder,
    boolean         isTrylock)
{
    fs_LockOrder    *inodeOrder;
    fs_Lock         *owned, *shared;
    fs_Buf          *ownedBuf, *sharedBuf;
    int32           i, inodeHeight;

    if (buf->inum == 0 || buf->page == FS_BUF_NO_PAGE)
    {
        // If the buf is not for a file, assume its
        // for a btree and lock order isn't checked
        // here for those, so ignore it.
        return FALSE;
    }

	// Ignore order involving reserved disk inode page.
	//
	// TODO - remove this by convering inode allocation
	// to non-accumulating lock (ie. use bitmap style
	// logging for them)
	if (buf->inum == FS_INODE_INODE &&
		buf->page == 0)
	{
		return FALSE;
	}

    // Ignore trylock attempts.
    if (isTrylock)
    {
        return FALSE;
    }

	if (buf->inum < FS_MAX_USED_META_INODE)
	{
    	inodeHeight = fs_lockOrderInodeMap[buf->inum];
    	FS_ASSERT(inodeHeight > 0 &&
              	  inodeHeight < FS_MAX_USED_META_INODE);

    	inodeOrder = &fs_lockOrderInode[inodeHeight];
    	if (inodeOrder->ignore)
        	return FALSE;
	}

    // Scan the exclusively held buf locks for this
	// inode, verify we don't already hold a higher
	// numbered page.
    DLL_ITERATE(owned, &context->ownerList, ownerEntry)
    {
        if (owned->type != FS_LOCK_TYPE_BUF)
            continue;

        ownedBuf = REVERSE_OFFSET(fs_Buf, lock, owned);
        if (ownedBuf->inum != buf->inum)
            continue;

        if (ownedBuf->inum == 0 || ownedBuf->page == FS_BUF_NO_PAGE)
        {
            // Assume its a btree page, ignore it.
            continue;
        }

        if (ownedBuf->page > buf->page)
		{
			printf("inum %lld ownedBuf->page %lld > buf-> page %lld\n",
				buf->inum, ownedBuf->page, buf->page);
            return TRUE;
		}
    }

    for (i = 0; i < context->sharedLockCount; i++)
	{
        shared = context->sharedLocksPtr[i].lock;
        if (shared->type != FS_LOCK_TYPE_BUF)
            continue;

        sharedBuf = REVERSE_OFFSET(fs_Buf, lock, shared);
        if (sharedBuf->inum != buf->inum)
            continue;

        if (sharedBuf->page > buf->page)
		{
			printf("inum %lld sharedBuf->page %lld > buf-> page %lld\n",
				buf->inum, sharedBuf->page, buf->page);
            return TRUE;
		}
    }

    return FALSE;

} /* fs_lockOrderLockBuf */


/**
 * Perform lock order checking on the specified
 * inode acquisition attempt.  Returns TRUE if a
 * violation occured.
 */
static boolean
fs_lockOrderLockInode(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_LockOrder    *objectOrder,
    boolean         isTrylock)
{
    fs_LockOrder    *inodeOrder;
    fs_Lock         *owned, *shared;
    fs_Inode        *ownedInode, *sharedInode;
    uint64          conflictInum = 0;
    int32           i, inodeHeight;
    boolean         hasParent = FALSE, orderWrong = FALSE;

    // Some inodes to ignore.  Something trying to do inode lookups
    // can cause these to be instantiated and locked for short
    // periods, so we need to trap them here. 
    if (inode->inum == 0 ||
        (inode->inum >= FS_MAX_USED_META_INODE &&
         inode->inum < FS_MAX_META_INODE))
    {
        return FALSE;
    }

	if (inode->inum != FS_ROOT_INODE &&
		inode->inum <= FS_MAX_USED_META_INODE)
	{
        inodeHeight = fs_lockOrderInodeMap[inode->inum];
        FS_ASSERT(inodeHeight > 0 &&
                  inodeHeight < FS_MAX_USED_META_INODE);

        inodeOrder = &fs_lockOrderInode[inodeHeight];
        if (inodeOrder->ignore)
            return FALSE;

        if (inodeOrder->isLeaf)
        {
            if (context->leafHeld != NULL &&
				context->leafHeld != inode)
			{
				// Lock order violation.
				printf("Leaf inode %p already held\n",
					context->leafHeld);
                return TRUE;
			}
            else
            {
                LOCK_TRACE(("order inode lock inum %lld is leaf", 
                            inode->inum));
                context->leafHeld = inode;
                return FALSE;
            }
        }

		if (inodeHeight < context->maxReservedInodeHeight)
		{
            if (!isTrylock)
            {
                // Lock order violation.
                printf(
			"inode->inum %lld height %d < held maxReservedInode height %d\n",
                    inode->inum, inodeHeight, context->maxReservedInodeHeight);
                return TRUE;
            }
            else
            {
                return FALSE;
            }
		}
		else
		{
            LOCK_TRACE(("order lock new ino height is %d inum %lld",
                   inodeHeight, inode->inum));
			context->maxReservedInodeHeight = inodeHeight;
			return FALSE;
		}
	}
	else if (!isTrylock)
	{
	    // If the parent inum is indicated, validate that
	    // this context holds the parent, otherwise validate
	    // that no held inode is greater then the requested
	    // one.
        DLL_ITERATE(owned, &context->ownerList, ownerEntry)
        {
            if (owned->type != FS_LOCK_TYPE_INODE)
			    continue;
    
		    ownedInode = REVERSE_OFFSET(fs_Inode, lock, owned);
		    if (context->lockOrderParentInum == ownedInode->inum)
			{
				// No problem, caller holds parent.
			    hasParent = TRUE;
			    break;
			}
    
		    if (ownedInode->inum > inode->inum)
		    {
                conflictInum = ownedInode->inum;
                orderWrong = TRUE;
		    }
        }
    	for (i = 0; i < context->sharedLockCount; i++)
	    {
            shared = context->sharedLocksPtr[i].lock;
            if (shared->type != FS_LOCK_TYPE_INODE)
			    continue;
    
		    sharedInode = REVERSE_OFFSET(fs_Inode, lock, shared);
		    if (context->lockOrderParentInum == sharedInode->inum)
			{
				// No problem, caller holds parent.
			    hasParent = TRUE;
			    break;
			}
    
		    if (sharedInode->inum > inode->inum)
		    {
                conflictInum = sharedInode->inum;
                orderWrong = TRUE;
		    }
		}

        if (orderWrong && !hasParent)
        {
		    // Lock order violation.
			printf("heldInode->inum %lld > inode->inum %lld\n",
				conflictInum, inode->inum);
            return TRUE;
        }

    	return FALSE;
	}
    else
    {
        // A trylock attempt at a non-reserved inode,
        // no checking, nothing to store here.
        return FALSE;
    }

} /* fs_lockOrderLockInode */


/**
 * Perform lock order checking on the specified lock
 * acquisition attempt.
 */
void
fs_lockOrderLock(
    fs_Context  *context,
    fs_Lock     *lock,
    int32       mode,
    boolean     isTrylock)
{
    fs_LockOrder    *objectOrder;
    int32           objectHeight;
    char            desc[FS_LOCK_DESCRIPTION];

    if (!fs_doLockOrderChecking)
        return;

	// No lock order checking during mkfs, etc.
	if (context->fileSystem->state != FS_VOL_STATE_ONLINE)
		return;

    if (context->ignoreLockOrder)
        return;

    objectHeight = fs_lockOrderObjectMap[lock->type];
    FS_ASSERT(objectHeight > 0 &&
              objectHeight < FS_LOCK_TYPE_MAX);

    objectOrder = &fs_lockOrderObject[objectHeight];
    if (objectOrder->ignore)
        return;

    NFX_SPIN_LOCK(&context->lockListSpin);

    if (objectOrder->isLeaf)
    {
        if (context->leafHeld != NULL &&
			context->leafHeld != lock)
		{
			printf("Leaf already held: %s\n",
                   fs_lockDescribe(context, context->leafHeld, desc));
            goto lockOrderViolated;
		}
        else
        {
            LOCK_TRACE(("order lock object height %d is leaf",
                        objectHeight));
            context->leafHeld = lock;
        }
        NFX_SPIN_UNLOCK(&context->lockListSpin);
        return;
    }

    switch(lock->type)
    {
        case FS_LOCK_TYPE_BUF: {
            fs_Buf *buf = REVERSE_OFFSET(fs_Buf, lock, lock);
            if (fs_lockOrderLockBuf(context, buf, objectOrder, isTrylock))
			{
				printf("Buf lockorder violated\n");
                goto lockOrderViolated;
			}
            break;
        }
        case FS_LOCK_TYPE_INODE: {
            fs_Inode *inode = REVERSE_OFFSET(fs_Inode, lock, lock);
            if (fs_lockOrderLockInode(context, inode, objectOrder, isTrylock))
			{
				printf("Inode lockorder violated\n");
                goto lockOrderViolated;
			}
			context->lockOrderParentInum = 0;
            break;
        }

        default:
            if (objectHeight < context->lockHeight)
            {
                if (!isTrylock)
                {
                    printf("Object lockorder violated\n");
                    goto lockOrderViolated;       
                }
            }
            else
            {
                context->lockHeight = objectHeight;
                LOCK_TRACE(("order lock object new height %d",
                            objectHeight));
            }
            break;
    }

    NFX_SPIN_UNLOCK(&context->lockListSpin);
    return;

lockOrderViolated:

    // Violation, scream real loud.
    NFX_SPIN_UNLOCK(&context->lockListSpin);
    fs_contextDump(context);
    fs_error(context, FS_INTERNAL_ERR,
        "lockorder violation, trying %s",
        fs_lockDescribe(context, lock, desc));

} /* fs_lockOrderLock */

#else  // FS_DEBUG_LOCK_ORDER

/**
 * Perform lock order checking on the specified lock
 * acquisition attempt.
 */
static inline void
fs_lockOrderLock(
    fs_Context  *context,
    fs_Lock     *lock,
    int32       mode,
    boolean     isTrylock)
{


} /* fs_lockOrderLock */


/**
 * Clean up lock order checking related info for
 * releasing the specified lock.
 */
static inline void
fs_lockOrderUnlock(
    fs_Context  *context,
    fs_Lock     *lock)
{

} /* fs_lockOrderUnlock */


/**
 * Initialize the lock order structures.
 */
void
fs_lockOrderInit()
{

}


#endif // FS_DEBUG_LOCK_ORDER



#if defined FS_DEBUG_LOCK

/**
 * Reset the exclusive lock count for the specified lock to zero for
 * a release.
 */
static inline void fs_lockResetExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
    assert(lock->exclusiveCount > 0);
    lock->exclusiveCount = 0;
}


/**
 * Initialize the exclusive lock count for the specified lock to one
 * for the first time it's claimed.
 */
static inline void fs_lockInitializeExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
    assert(lock->exclusiveCount == 0);
    lock->exclusiveCount = 1;
}


/**
 * Increment the exclusive lock count for the specified lock.
 */
static inline void fs_lockIncrementExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
    assert(lock->owner == context->smHandle);
    lock->exclusiveCount += 1;
}


/**
 * Decrement the exclusive lock count for the specified lock.
 */
static inline void fs_lockDecrementExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
    assert(lock->owner == context->smHandle);
    assert(lock->exclusiveCount > 0);
    lock->exclusiveCount -= 1;
}

/**
 * Decrement the exclusive lock count for the specified lock, give
 * up all the counts.
 */
static inline void fs_lockDecrementExclusiveCountAll(
    fs_Context *context,
    fs_Lock    *lock)
{
    assert(lock->owner == context->smHandle);
    assert(lock->exclusiveCount > 0);
    lock->exclusiveCount = 0;
}

/**
 * Return the count of exclusive acquisitions for the specified lock,
 * if they are being tracked, or zero if not known.
 */
static inline int fs_lockGetExclusiveCount(
    fs_Context *context,
    fs_Lock *lock)
{
    if (!FS_KEEP_TRACK_OF_LOCKS(context)) {
        if (lock->mode == FS_LOCK_MODE_EXCL)
            return lock->count;
        else
            return 0;
    }
    return lock->exclusiveCount;
}



#else   // ! FS_DEBUG_LOCK

/**
 * Reset the exclusive lock count for the specified lock to zero for
 * a release.
 */
static inline void fs_lockResetExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
}


/**
 * Initialize the exclusive lock count for the specified lock to 1
 * for the first time it's claimed.
 */
static inline void fs_lockInitializeExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
}


/**
 * Increment the exclusive lock count for the specified lock.
 */
static inline void fs_lockIncrementExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
}


/**
 * Decrement the exclusive lock count for the specified lock.
 */
static inline void fs_lockDecrementExclusiveCount(
    fs_Context *context,
    fs_Lock    *lock)
{
}


/**
 * Return the count of exclusive acquisitions for the specified lock,
 * if they are being tracked, or zero if not known.
 */
static inline int fs_lockGetExclusiveCount(
    fs_Context *context,
    fs_Lock *lock)
{
    if (lock->mode == FS_LOCK_MODE_EXCL)
        return lock->count;
    else
        return 0;
}

#endif

#ifdef FS_DEBUG_CHECK_SHAREDLOCK
void fs_lockPrintSharedLockStats(
    fs_Context *context)
{
    int loop;
    
    printf ("fs_lockMaxSharedLockListSize    %d\n", fs_lockMaxSharedLockListSize);
    printf ("fs_lockAddSharedLock            %d\n", fs_lockAddSharedLock);
    printf ("fs_lockAddSharedLockGrow        %d\n", fs_lockAddSharedLockGrow);
    printf ("\n");
    for (loop = 0 ; loop < FS_DEBUG_CHECK_SHAREDLOCK_MAX; loop++)
    {
        if (fs_lockSharedLockCount[loop])
            printf ("fs_lockSharedLockCount     %d  %d\n", loop, fs_lockSharedLockCount[loop]);
    }
    printf ("\n");
    for (loop = 0 ; loop < FS_DEBUG_CHECK_SHAREDLOCK_MAX; loop++)
    {
        if (fs_lockSharedLockGet[loop])
            printf ("fs_lockSharedLockGet       %d  %d\n", loop, fs_lockSharedLockGet[loop]);
    }
    printf ("\n");
    for (loop = 0 ; loop < FS_DEBUG_CHECK_SHAREDLOCK_MAX; loop++)
    {
        if (fs_lockSharedLockRm[loop])
            printf ("fs_lockSharedLockRm        %d  %d\n", loop, fs_lockSharedLockRm[loop]);
    }
}
#endif


/**
 * Remove the specified lock from the chain of owned locks for the
 * specified context.
 */
static inline void fs_lockRemoveOwned(
    fs_Context *context,
    fs_Lock    *lock)
{
    //char desc[FS_LOCK_DESCRIPTION];
    //printf("%6x -= %s\n", context->smHandle,
    //       fs_lockDescribe(context, lock, desc));

    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return;

    assert(lock->owner != 0);

    NFX_SPIN_LOCK(&context->lockListSpin);

    FS_ASSERT( ! DLL_IS_ISOLATED(lock, ownerEntry));

    DLL_REMOVE (&context->ownerList, lock, ownerEntry);

    NFX_SPIN_UNLOCK(&context->lockListSpin);
}


/**
 * Add the specified lock to the chain of owned locks for the
 * specified context.
 */
static void fs_lockAddOwned(
    fs_Context *context,
    fs_Lock    *lock)
{
    //char desc[FS_LOCK_DESCRIPTION];
    //printf("%6x += %s\n", context->smHandle,
    //       fs_lockDescribe(context, lock, desc));
       
    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return;

    assert(lock->mode == FS_LOCK_MODE_EXCL);
    assert(lock->owner != 0);
    FS_ASSERT(DLL_IS_ISOLATED(lock, ownerEntry));
    NFX_SPIN_LOCK(&context->lockListSpin);
    DLL_INSERT_HEAD(&context->ownerList, lock, ownerEntry);
    NFX_SPIN_UNLOCK(&context->lockListSpin);
}

/**
 * Inform the lock manager that the specified context is disappearing.
 * All locks held at this point should cause an error.
 */
void fs_lockTerminateContext(
    fs_Context *context)
{
    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return;

    if ((! DLL_IS_EMPTY(&context->ownerList)) || context->sharedLockCount != 0) 
    {
        int i;
        fs_Lock *lock;
        char desc[FS_LOCK_DESCRIPTION];

        NFX_SPIN_LOCK(&context->lockListSpin);

        DLL_ITERATE(lock, &context->ownerList, ownerEntry)
            {
                printf("esm %llx died holding %s\n", 
					   (uint64)((uintptr_t)context->smHandle),
                       fs_lockDescribe(context, lock, desc));
            }
        
        for (i = 0; i < context->sharedLockCount; i++) {
            fs_SharedLockRef *shared = &context->sharedLocksPtr[i];
            assert(shared->lock != NULL);
            printf("esm %llx died sharing (%d time%s) %s\n",
				   (uint64)((uintptr_t)context->smHandle),
                   shared->count, shared->count == 1 ? "" : "s",
                   fs_lockDescribe(context, shared->lock, desc));
        }

        NFX_SPIN_UNLOCK(&context->lockListSpin);

		fs_contextDump(context);
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
        // fs_panic("esm %7x failed to release locks before terminating", context->smHandle);
    }
}



/**
 * Add a lock to the list of shared locks held by this context.
 */
static void fs_lockAddShared(
    fs_Context *context,
    fs_Lock *lock)
{
    fs_SharedLockRef *shared, *newShared;
    int i;

    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return;

    assert(lock->owner == 0 || lock->owner == context->smHandle);

    NFX_SPIN_LOCK(&context->lockListSpin);
    shared = &context->sharedLocksPtr[context->sharedLockCount - 1];
    for (i = context->sharedLockCount; i > 0; i--) {
        if (shared->lock == lock) {
            assert(shared->count < lock->count);
            shared->count += 1;
            NFX_SPIN_UNLOCK(&context->lockListSpin);
            return;
        }
        shared--;
    }
    
    if (context->sharedLockCount >= context->sharedLockSize) {
        //List needs to grow
        fs_SharedLockRef *newList;
        newList = fs_ramAlloc(FS_RAM_CTX_SHARED_LOCKS,
                        sizeof(fs_SharedLockRef) * 2 * context->sharedLockSize);
        memcpy(newList, context->sharedLocksPtr, 
               sizeof(fs_SharedLockRef) * context->sharedLockSize);
#ifdef FS_DEBUG_LOCK
        if ((address_t)context->sharedLocksPtr != (address_t)&context->sharedLocks)
#endif
        {
            fs_ramDealloc(FS_RAM_CTX_SHARED_LOCKS, context->sharedLocksPtr);
        }
        context->sharedLocksPtr = newList;
        context->sharedLockSize *= 2;
#ifdef FS_DEBUG_CHECK_SHAREDLOCK
        fs_lockAddSharedLockGrow++;
        if (context->sharedLockSize > fs_lockMaxSharedLockListSize)   {
            fs_lockMaxSharedLockListSize = context->sharedLockSize;
        }
#endif
    }
    
    newShared = &context->sharedLocksPtr[context->sharedLockCount];
    context->sharedLockCount += 1;
    newShared->lock = lock;
    newShared->count = 1;

#ifdef FS_DEBUG_CHECK_SHAREDLOCK
    fs_lockAddSharedLock++;
    if (context->sharedLockCount >= FS_DEBUG_CHECK_SHAREDLOCK_MAX)        
        fs_lockSharedLockCount[FS_DEBUG_CHECK_SHAREDLOCK_MAX - 1]++;
    else
        fs_lockSharedLockCount[context->sharedLockCount]++;
#endif

    NFX_SPIN_UNLOCK(&context->lockListSpin);
}


/**
 * Return the number of shared acquisitions of a lock by the specified
 * context.  Returns zero if the lock is not currently held.
 */
static int fs_lockGetSharedCount(
    fs_Context *context,
    fs_Lock *lock)
{
    int i;
    fs_SharedLockRef *shared;

    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return 0;

    NFX_SPIN_LOCK(&context->lockListSpin);

#ifdef FS_DEBUG_CHECK_SHAREDLOCK
    if (context->sharedLockCount >= FS_DEBUG_CHECK_SHAREDLOCK_MAX)        
        fs_lockSharedLockGet[FS_DEBUG_CHECK_SHAREDLOCK_MAX - 1]++;
    else
        fs_lockSharedLockGet[context->sharedLockCount]++;
#endif

    shared = context->sharedLocksPtr;
    for (i = context->sharedLockCount; i > 0; i--) {
        if (shared->lock == lock)   {
            NFX_SPIN_UNLOCK(&context->lockListSpin);
            return shared->count;
        }
        shared++;
    }
    NFX_SPIN_UNLOCK(&context->lockListSpin);

    return 0;
}


/**
 * Remove a single reference to a lock from the list of shared locks held
 * by this content.  Returns TRUE if the last reference to the specified
 * lock is removed.
 */
static boolean fs_lockRemoveShared(
    fs_Context *context,
    fs_Lock *lock)
{
    int i;
	boolean rc = FALSE;
    fs_SharedLockRef *shared;

    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return FALSE;

    NFX_SPIN_LOCK(&context->lockListSpin);

#ifdef FS_DEBUG_CHECK_SHAREDLOCK
    if (context->sharedLockCount >= FS_DEBUG_CHECK_SHAREDLOCK_MAX)        
        fs_lockSharedLockRm[FS_DEBUG_CHECK_SHAREDLOCK_MAX - 1]++;
    else
        fs_lockSharedLockRm[context->sharedLockCount]++;
#endif

    shared = &context->sharedLocksPtr[context->sharedLockCount - 1];
    for (i = context->sharedLockCount; i > 0; i--) {
        if (shared->lock == lock) {
            if (shared->count == 1) {
                fs_SharedLockRef *last
                    = &context->sharedLocksPtr[--context->sharedLockCount];
                if (shared != last)
                    *shared = *last;
                last->lock  = NULL;
                last->count = 0;
                shared = last;
				rc = TRUE;
            }
            else {
                shared->count -= 1;
            }
            if (shared->count > lock->count) {
                char desc[FS_LOCK_DESCRIPTION];
                printf("%s: shared->count %d lock->count %d\n",
                       fs_lockDescribe(context, lock, desc),
                       shared->count, lock->count);
            }
            assert(shared->count <= lock->count);
            NFX_SPIN_UNLOCK(&context->lockListSpin);
            return rc;
        }
        shared--;
    }
    
    NFX_SPIN_UNLOCK(&context->lockListSpin);

    {
        char desc[FS_LOCK_DESCRIPTION];
        fs_error(context, FS_INTERNAL_ERR,
                 "unlock shared: not sharing lock %s",
                 fs_lockDescribe(context, lock, desc));
    }

	return rc;
}


/**
 * Remove all references to a lock from the list of shared locks held by
 * this content, if any.
 */
static void fs_lockRemoveSharedAll(
    fs_Context *context,
    fs_Lock *lock)
{
    int i;
    fs_SharedLockRef *shared = context->sharedLocksPtr;

    if (!FS_KEEP_TRACK_OF_LOCKS(context))
        return;

    NFX_SPIN_LOCK(&context->lockListSpin);

#ifdef FS_DEBUG_CHECK_SHAREDLOCK
    if (context->sharedLockCount >= FS_DEBUG_CHECK_SHAREDLOCK_MAX)        
        fs_lockSharedLockRm[FS_DEBUG_CHECK_SHAREDLOCK_MAX - 1]++;
    else
        fs_lockSharedLockRm[context->sharedLockCount]++;
#endif

    for (i = context->sharedLockCount; i > 0; i--) {
        if (shared->lock == lock) {
            fs_SharedLockRef *last
                = &context->sharedLocksPtr[--context->sharedLockCount];
            if (shared != last)
                *shared = *last;
            last->lock  = NULL;
            last->count = 0;
            NFX_SPIN_UNLOCK(&context->lockListSpin);
            return;
        }
        shared++;
    }
    NFX_SPIN_UNLOCK(&context->lockListSpin);
}


/**
 * Print the locks owned by the specified context.
 */
void fs_lockPrintOwned(
    fs_Context *context)
{
    int i, lineNbr = 0;
    fs_Lock *lock;

    printf ("context %p esm 0x%llx reqType %s\n", context,
			 (uint64)((uintptr_t)context->smHandle),
			 (context->reqHdr != NULL) ? 
			fs_reqTypeTbl[context->reqHdr->requestType].name : "?");

    if (!FS_KEEP_TRACK_OF_LOCKS(context))   {
        printf("    Locks not available for this thread\n");
        return;
    }
    
    NFX_SPIN_LOCK(&context->lockListSpin);

    DLL_ITERATE(lock, &context->ownerList, ownerEntry)
    {
        char desc[FS_LOCK_DESCRIPTION];
        printf("%2d  Owns %s\n", lineNbr++, fs_lockDescribe(context, lock, desc));
    }

    for (i = 0; i < context->sharedLockCount; i++) {
        char desc[FS_LOCK_DESCRIPTION];
        fs_SharedLockRef *shared = &context->sharedLocksPtr[i];
        printf("%2d  Shares (%d time%s) %s\n",
               lineNbr++,
               shared->count, (shared->count == 1) ? "" : "s",
               fs_lockDescribe(context, shared->lock, desc));
    }
	if (context->lockWait.lock != NULL)
	    printf("%2d  Waiting for lock %p\n", lineNbr, context->lockWait.lock);
    else
	    printf("%2d  NOT waiting for a lock\n", lineNbr);

    NFX_SPIN_UNLOCK(&context->lockListSpin);
}


/**
 * Print the locks owned by all the running contexts of all the online fs.
 */
void fs_lockPrintAllOwned(void)
{
    fs_Context      *threadContext;
    uint32          loop;
    fs_FileSystem   *fs;

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);

    for (loop = 0; loop < FS_FILE_SYSTEM_HASH_MAX; loop++)
    {
        for (fs = fs_fileSystems[loop]; fs != NULL; fs = fs->next)
        {
            if ((fs->state == FS_VOL_STATE_ONLINE) ||
                (fs->state != FS_VOL_STATE_ONLINE_RDONLY))
            {
                printf ("\n\n\nLocks for filesystem %s \n\n", 
                        fs->super->mountName);

                NFX_SPIN_LOCK(&fs->spin);
                DLL_ITERATE(threadContext, &fs->activeThreadHdr, fsEntry)
                {
                    fs_lockPrintOwned (threadContext);
                }
                NFX_SPIN_UNLOCK(&fs->spin);
                printf("MetaWriteback esm 0x%llx\n", 
                       (uint64)((uintptr_t)fs->bufList[FS_BUFLIST_META].wbHandle));
                printf("UserWriteback esm 0x%llx\n", 
                       (uint64)((uintptr_t)fs->bufList[FS_BUFLIST_USER].wbHandle));
            }
        }
    }

#ifdef FS_DEBUG_CHECK_SHAREDLOCK
    fs_lockPrintSharedLockStats(NULL);
#endif

    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
}

int fs_uiLockPrintAllOwned(	
    int	argc,
	char **argv)
{
    fs_lockPrintAllOwned();

    return NFX_OK;
}


/**
 * Write a description of the object covered by the specified lock to the
 * specifed string buffer.  Returns the number of characters appended.
 */
static int fs_lockDescribeObject(
    fs_Context *context,
    fs_Lock    *lock,
    char       *string,
    int32      remain)
{
    char       *p = string;
    int32      this;

    this = snprintf(p, remain, " - ");
    p += this;
    remain -= this;

    switch (lock->type) {

        case FS_LOCK_TYPE_BUF: {
            fs_Buf *buf = REVERSE_OFFSET(fs_Buf, lock, lock);
            this = snprintf(p, remain, "buffer");
            p += this;
            remain -= this;

            if (fs_blockIsValid(buf->block))
            {
                this = snprintf(p, remain, " block %lld", fs_blockNumber(buf->block));
                p += this;
                remain -= this;
            }
            
            if (buf->inum != 0) 
            {
                this = snprintf(p, remain, " inode %lld ", buf->inum);
                p += this;
                remain -= this;

                if (buf->page == FS_BUF_NO_PAGE)
                    this = snprintf(p, remain, "btree page");
                else
                    this = snprintf(p, remain, "page %lld", buf->page);
                p += this;
                remain -= this;
            }
            break;
        }

        case FS_LOCK_TYPE_INODE: {
            fs_Inode *inode = REVERSE_OFFSET(fs_Inode, lock, lock);
            if (FS_I_IS_QUOT(inode->dinode.attrs))
                this = snprintf(p, remain, "quota inode %lld", inode->inum);
            else
                this = snprintf(p, remain, "inode %lld", inode->inum);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_BTREE: {
            fs_Inode *inode = REVERSE_OFFSET(fs_Inode, btreeLock, lock);
            this = snprintf(p, remain, "btree inode %lld", inode->inum);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_QUOTA: {
            fs_QuotaIncoreRecord_t *quota;
            quota = REVERSE_OFFSET(fs_QuotaIncoreRecord_t, lock, lock);
            this = snprintf(p, remain, "quota fset %u %cid %u", quota->fileSetId,
                         ((quota->flag & FS_QUOTA_INCORE_USER) ? 'u' : 'g'),
                         quota->id);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_FILESET: {
            fs_FileSetIncoreRecord_t *fset;
            fset = REVERSE_OFFSET(fs_FileSetIncoreRecord_t, lock, lock);
            this = snprintf(p, remain, "fset %u", fset->fileSetId);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_FILESET_TRAVERSAL: {
            fs_FileSetIncoreRecord_t *fset;
            fset = REVERSE_OFFSET(fs_FileSetIncoreRecord_t, traverseLock, lock);
            this = snprintf(p, remain, "fset %u traversal", fset->fileSetId);
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_BOOT: {
            this = snprintf(p, remain, "boot allocator");
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_MODIFY: {
            this = snprintf(p, remain, "global modify");
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_ACCESS: {
            this = snprintf(p, remain, "global access");
            p += this;
            remain -= this;
            break;
        }

        case FS_LOCK_TYPE_FREEZE: {
            this = snprintf(p, remain, "freezer");
            p += this;
            remain -= this;
            break;
        }

        default: {
            this = snprintf(p, remain, "unknown (%d)", lock->type);
            p += this;
            remain -= this;
            break;
        }

    }

    return p - string;
}


/**
 * Print a description of the specified lock into the specified string,
 * which must have at least FS_LOCK_DESCRIPTION bytes.  Returns the
 * description.
 *
 * Given the way fs_trace works, this is far too expensive to use with
 * it in a production system.  Ideally, we could add a %q or somesuch
 * for printing locks, but that's not possible with standard printf.
 */
const char *fs_lockDescribe(
    fs_Context *context,
    fs_Lock    *lock,
    char       string[FS_LOCK_DESCRIPTION])
{
    char *p = string;
    int exclusiveCount = fs_lockGetExclusiveCount(context, lock);
    fs_LockWait *next;
    int32       this, remain = FS_LOCK_DESCRIPTION - 1;

    assert(string != NULL);

    this = snprintf(string, remain, "{ lock %p : ", lock);
    p += this;
    remain -= this;

    if (lock->mode < (sizeof fs_lockModes) / (sizeof *fs_lockModes))
        this = snprintf(p, remain, "%s", fs_lockModes[lock->mode]);
    else
        this = snprintf(p, remain, "mode %d", lock->mode);
    p += this;
    remain -= this;

    if (lock->owner != 0)
    {
        this = snprintf(p, remain, " owner esm 0x%llx", (uint64)((uintptr_t)lock->owner));
        p += this;
        remain -= this;
    }
    
    if (lock->count > 0)
    {
        this = snprintf(p, remain, " count %d", lock->count);
        p += this;
        remain -= this;
    }

    if (lock->mode == FS_LOCK_MODE_EXCL && exclusiveCount != lock->count)
    {
        this = snprintf(p, remain, " exclusive %d", exclusiveCount);
        p += this;
        remain -= this;
    }

    if (lock->mode == FS_LOCK_MODE_EXCL) 
    {
        this = snprintf(p, remain, " ra ");
        p += this;
        remain -= this;
#if defined(FS_STANDALONE)
        this = btrace_sprint_function(p, lock->ra);
#else
        this = snprintf(p, remain, " %#lx ", (address_t)lock->ra);
#endif
        p += this;
        remain -= this;
    }

    if (lock->waitCount > 0)
    {
        this = snprintf(p, remain, " waitcount %d", lock->waitCount);
        p += this;
        remain -= this;
    }

    if ((lock->flags & FS_LOCK_FLAG_DESTROY) != 0)
    {
        this = snprintf(p, remain, " destroyed");
        p += this;
        remain -= this;
    }

    if ((lock->flags & FS_LOCK_FLAG_DIRTY) != 0)
    {
        this = snprintf(p, remain, " dirty");
        p += this;
        remain -= this;
    }

    this = fs_lockDescribeObject(context, lock, p, remain);   // DEBUG
    p += this;
    remain -= this;

    this = snprintf(p, remain, " }\n      {");
    p += this;
    remain -= this;

    this = snprintf(p, remain, "Wait esm list ");
    p += this;
    remain -= this;

    next = lock->waiting; 
    while (next != NULL)
    {
        this = snprintf(p, remain, " 0x%llx %s ", (uint64)((uintptr_t)next->smHandle),
					  (next->mode == FS_LOCK_MODE_EXCL) ? "Ex" : "Sh");
        p += this;
        remain -= this;

        if (this <= 0)
            break;

        next = next->next;
    }

    this = snprintf(p, remain, "}");
    p += this;
    remain -= this;

    return string;
}


//   P R I M I T I V E S


/**
 * Attempt to acquire the specified lock in the specified mode.  Returns
 * true if the lock was acquired and false otherwise.  Never sleeps.
 *
 * Must be called with lock->spin held.
 */
static boolean fs_trylockSpin(
    fs_Context  *context,
    fs_Lock     *lock,
    fs_LockMode lockMode,
    void        *ra)
{
    boolean isExclusive;
    char desc[FS_LOCK_DESCRIPTION];
    
    LOCK_TRACE(("lock %s %s", fs_lockModes[lockMode],
                fs_lockDescribe(context, lock, desc)));

    assert(context->smHandle != 0);

    switch (lockMode) {
        case FS_LOCK_MODE_SHARE:
            isExclusive = FALSE;
            break;
        case FS_LOCK_MODE_EXCL:
            isExclusive = TRUE;
            break;
        default:
            isExclusive = FALSE;
            assert(FALSE);
    }

    if (!fs_lockIsLive(context, lock)) {
        // Lock is marked for destruction, disallow new callers.
        return FALSE;
    }

    if (fs_lockOwnedExclusive(context, lock)) {
        // The owning transaction has exclusive access to the
        // object.  Let any of its lock requests proceed.  We
        // maintain the counts here, but in the event the
        // owner dirties the object the count is simply zeroed
        // during commit so there doesn't need to be an equal
        // number of lock/unlock calls - if the object isn't
        // dirtied (thus no subsequent commit), then the transaction
        // must ensure symmetry between lock/unlock calls, otherwise
        // the lock will never become free.
        lock->count += 1;
        if (isExclusive)
            fs_lockIncrementExclusiveCount(context, lock);
        else
            fs_lockAddShared(context, lock);
        return TRUE;
    }

    if (!fs_lockCompat[lockMode][lock->mode])
        return FALSE;
    
#if defined FS_DEBUG_LOCK
  	if (lock->waiting)
	{
		// To provide a measure of fairness and more importantly
		// to avoid live-lock, disallow the request if others
		// are already waiting in line for it (This could occur
		// if a lock is held in shared mode and others continuously
		// pile themselves onto it while an exclusive requestor
		// is forced to wait).

        if (FS_KEEP_TRACK_OF_LOCKS(context))    {
            // If this context already has it shared, we still have to grant it
            // to avoid a deadlock with the exclusive requestor waiting
            // 
            // For quota/fileset incore quotas, we can not do that (deadlock),
            // but they are never held for long so we should never reach the
            // lock timeout
            if (lock->type != FS_LOCK_TYPE_QUOTA
             && lock->type != FS_LOCK_TYPE_FILESET
             && !fs_lockOwnedShared(context, lock, TRUE)) {
        		return FALSE;
            }
        } else {
            // If we don't keep track of locks for that context, we are not sure
            // whether it holds it already. Allow lock access starvation.
            lock->flags |= FS_LOCK_FLAG_ALLOW_STARVATION;    
        }
	}
#else    
    if (lock->waiting) 
    {
        if (lock->type != FS_LOCK_TYPE_QUOTA
            && lock->type != FS_LOCK_TYPE_FILESET)
        {
            return FALSE;
        }
    }
#endif

    // Grant it.

    lock->count += 1;
    if (isExclusive)
        fs_lockInitializeExclusiveCount(context, lock);
    else
        fs_lockAddShared(context, lock);

    if (lockMode > lock->mode) {
        assert(lock->owner == 0);
        lock->mode = lockMode;
        if (isExclusive) {
            lock->owner = context->smHandle;
            lock->ra = ra;
            fs_lockAddOwned(context, lock);
#ifdef FS_NO_LOCK_TIMEOUT
            if (context->flags & FS_CONTEXT_DISALLOW_EXCL_LOCK_TIMEOUT) {
                    lock->flags |= FS_LOCK_DISALLOW_EXCL_TIMEOUT;
            }
#endif
        }
    }
    //fs_lockPrintOwned(context); // DEBUG
    return TRUE;
}


/**
 * Attempt to acquire the specified lock in the specified mode.  Returns
 * true if the lock was acquired and false otherwise.  Never sleeps.
 */
boolean fs_trylock(
    fs_Context  *context,
    fs_Lock     *lock,
    fs_LockMode mode)
{
    boolean acquired;
    void *ra = (void *) (address_t) get_ra();

    if (!fs_lockIsLive(context, lock)) {
        // Lock is marked for destruction, disallow new callers.
        return FALSE;
    }

    NFX_SPIN_LOCK(&lock->spin);
    acquired = fs_trylockSpin(context, lock, mode, ra);
    NFX_SPIN_UNLOCK(&lock->spin);

    if (acquired)
        fs_lockOrderLock(context, lock, mode, TRUE);

    return acquired;
}


/**
 * Sleep to wait for a lock becoming available.  By the time this function
 * is called, we're already set up to wait on the lock.
 */
static int32 fs_lockSleep(
    fs_Context      *context,
    fs_Lock         *lock,
    nfx_spinlock_t  *spinp,
    int32           event)
{
    esm_event_t wakeEvent;
    char  desc[FS_LOCK_DESCRIPTION];

    // Release caller's lock, if supplied.
    if (spinp != NULL)
        NFX_SPIN_UNLOCK(spinp);

    LOCK_TRACE(("yielding for wait on lock %s",
                fs_lockDescribe(context, lock, desc)));

    do {
        // Release the CPU.
#ifdef FS_NO_LOCK_TIMEOUT
        if (lock->flags & FS_LOCK_DISALLOW_EXCL_TIMEOUT) {
retry:
                fs_yield(context, &wakeEvent, FS_YIELD_TIME_LOCK,
                         FS_YIELD_TYPE_LOCK);
                if (wakeEvent.event == THREAD_EVENT_TIMEOUT) {
                    FS_ELOG(context, LOG_NOTICE, FS_TRACE_LCK, 
                        (   "lock wait timeout, retrying"));
                    goto retry;
                }
        } else {
                fs_yield(context, &wakeEvent, FS_YIELD_TIME_LOCK,
                         FS_YIELD_TYPE_LOCK);
        }
#else
        fs_yield(context, &wakeEvent, FS_YIELD_TIME_LOCK, FS_YIELD_TYPE_LOCK);
#endif

        if (wakeEvent.event == THREAD_EVENT_TIMEOUT)    {
#ifdef FS_NO_LOCK_TIMEOUT
            FS_ASSERT((lock->flags & FS_LOCK_DISALLOW_EXCL_TIMEOUT) == 0);
#endif
            NFX_SPIN_LOCK(&lock->spin);
            if (lock->flags & FS_LOCK_FLAG_ALLOW_STARVATION)    {
                NFX_SPIN_UNLOCK(&lock->spin);
                FS_ELOG(context, LOG_NOTICE, FS_TRACE_LCK, 
                        ("lock wait timeout, starvation allowed"));
                continue;
            }
            NFX_SPIN_UNLOCK(&lock->spin);
            fs_volumeException(context, context->fileSystem,
                               FS_EVENT_TIMEOUT_ERR);
            // This will not return here.
            FS_ASSERT(FALSE);
        }
    }while (0);
    
    LOCK_TRACE(("resuming from wait on lock %s",
                fs_lockDescribe(context, lock, desc)));

    // Re-acquire caller's lock, if supplied.
    if (spinp != NULL)
        NFX_SPIN_LOCK(spinp);

    // Make sure the event we got is the event
    // we were waiting for!  If not, return error.
    //
    // TODO - may want to enhance this so that
    // our caller can get the event,p1.u.p2 info
    // in the event it isn't what we're waiting
    // for.  (Do this by plugging the values into
    // the context structure)
    if (wakeEvent.event != event)
    {
        printf("fs_lock: wrong event: got %d/%p/%p, exp %d/%p\n",
               wakeEvent.event, wakeEvent.p1.u.p, wakeEvent.p2.u.p, event, lock);
        fs_error(context, FS_INTERNAL_ERR, "lock wrong event");
        return FS_ERR_WRONG_EVENT;
    }

    if (wakeEvent.p1.u.p == NULL)
    {
        // Lock destroyed
        return FS_ERR_LOCK_DESTROYED;
    }

    if (wakeEvent.p1.u.p != lock)
    {
        printf("fs_lock: wrong param1: got %d/%p/%p, exp %d/%p\n",
               wakeEvent.event, wakeEvent.p1.u.p, wakeEvent.p2.u.p, event, lock);
        fs_error(context, FS_INTERNAL_ERR, "lock wrong event");
        return FS_ERR_WRONG_EVENT;
    }
#ifdef FS_DEBUG_LOCK
	context->lockWait.lock = NULL;
#endif	//FS_DEBUG_LOC

    // Lock is granted.
    return NFX_OK;
}


/**
 * Enqueue a state machine to the waiting list for a lock.
 * Must be called with lock->spin held.
 */
static void fs_lockEnqueueWaiter(
    fs_Context  *context,
    fs_Lock     *lock,
    fs_LockWait *wait)
{
    fs_Context  *ownerContext;
    int32       whatOwns, whatThis;

    assert(wait->next == NULL);

    if (lock->waiting == NULL)
    {
        if (lock->waitCount != 0)
        {
            printf("waitcount is %d\n", lock->waitCount);
        }
        assert(lock->waitCount == 0);
        lock->waiting = wait;
    }
    else
    {
        fs_LockWait *next;
        fs_LockWait **prevp;
        FS_LOOP_DBG_DECL(dbgNumLoops);

        for (prevp = &lock->waiting;
             (next = *prevp) != NULL;
             prevp = &next->next)
        {
            FS_LOOP_DBG_CHECK(dbgNumLoops, 5000);
        }
        *prevp = wait;
    }

    lock->waitCount += 1;
    lock->totalWaits += 1;
#ifdef FS_LOCK_WAITER_TIME
	eee_rtc(&wait->time);
#endif
    // save lock pointer to dequeue waiter at thread terminate time
    wait->lock = lock;

    // Keep track by request type and lock type
    // of what other things get in our way.
    if (lock->owner != 0)
    {
        ownerContext = esm_getUserInfo(lock->owner);
        if (ownerContext->daemonType)
        {
            whatOwns = ownerContext->daemonType + FS_LOCK_DAEMON_OFFSET;
        }
        else if (ownerContext->reqHdr &&
                 ownerContext->reqHdr->requestType < FS_REQ_NFS_MAX)
        {
            whatOwns = ownerContext->reqHdr->requestType;
        }
        else
        {
            whatOwns = FS_LOCK_OWNER_UNK;
        }
    }
    else
    {
        whatOwns = FS_LOCK_OWNER_SHARED;
    }
    
    if (context->daemonType)
    {
        whatThis = context->daemonType + FS_LOCK_DAEMON_OFFSET;
    }
    else if (context->reqHdr &&
              context->reqHdr->requestType < FS_REQ_NFS_MAX)
    {
        whatThis = context->reqHdr->requestType;
    }
    else
    {
        whatThis = FS_LOCK_OWNER_UNK;
    }

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs_lockWaitStats[whatThis]
                    [lock->type]
                    [whatOwns] += 1;
    if (lock->type == FS_LOCK_TYPE_BUF)
    {
        fs_Buf *buf = REVERSE_OFFSET(fs_Buf, lock, lock);

        if (buf->inum < FS_MAX_META_INODE)
        {
            if (buf->page != FS_BUF_NO_PAGE)
                fs_lockStatBuf.metaBufInumPage[buf->inum] += 1;
            else
                fs_lockStatBuf.metaBufInumOther[buf->inum] += 1;
        }
        else
        {
            if (buf->page != FS_BUF_NO_PAGE)
                fs_lockStatBuf.userBufInumPage += 1;
            else
                fs_lockStatBuf.userBufInumOther += 1;
        }
    }
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
}


/**
 * Delete a waiter from a lock's waiting list
 * Called at thread terminate time
 */
void fs_lockDeleteWaiter(
    fs_Context  *context)
{
    fs_Lock *lock;
    fs_LockWait *wait = &context->lockWait;
    boolean found = FALSE;

    lock = wait->lock;
    assert(lock);

    NFX_SPIN_LOCK(&lock->spin);

    assert(lock->waiting != NULL);
    lock->waitCount -= 1;
    lock->totalWaits -= 1;

    if (lock->waiting == wait)
    {
        lock->waiting = wait->next;
    }
    else
    {
        fs_LockWait *currp;
        fs_LockWait *prevp;
        FS_LOOP_DBG_DECL(dbgNumLoops);

        for (prevp = currp = lock->waiting;
             currp;
             prevp = currp, currp = currp->next)
        {
            FS_LOOP_DBG_CHECK(dbgNumLoops, 5000);
            if (currp == wait)
            {
                found = TRUE;
                prevp->next = wait->next;
                break;
            }
        }
        assert(found);
    }

    NFX_SPIN_UNLOCK(&lock->spin);
}

/**
 * Wait for a filesystem lock.
 *
 * Must be called with lock->spin held;  the spin lock is dropped before
 * the call returns.
 */
int32 fs_lockWait(
    fs_Context     *context,
    fs_Lock        *lock,
    fs_LockMode    mode,
    boolean        sleep,
    nfx_spinlock_t *spinp,
    int32          event,
    void           *p2,
    boolean        allocateWaitStruct,
    void           *ra)
{
    fs_LockWait *wait;
    char desc[FS_LOCK_DESCRIPTION];

    LOCK_TRACE(("must wait on lock %s",
                fs_lockDescribe(context, lock, desc)));

	// Let this through.  Yeah, its deadlock bait (until
	// we completely track shared locks) but we should 
	// allow it for now, certain cases can make good use
	// of it (eg. wb daemon holds shared lock on buf when
	// doing write of cloned buf, and it might also try to
	// get excl lock on shared buf to kick it off again)
	//
    //assert(fs_lockGetSharedCount(context, lock) == 0);
    assert(!fs_lockOwnedExclusive(context, lock));

    // Add this requestor to the wait queue.
    if (allocateWaitStruct) {
        // TODO - would be good to move the alloc outside of where
        // we hold the spinlock, but not strictly necessary.
        wait = fs_ramAlloc(FS_RAM_LOCK_WAIT, sizeof (fs_LockWait));
        wait->flags = FS_LOCK_WAIT_FLAG_ALLOC;
    }
    else {
        wait = &context->lockWait;
        wait->flags = 0;
        assert(wait->smHandle == 0);
    }
    wait->smHandle = context->smHandle;
    wait->event    = event;
    wait->p2       = p2;
    wait->mode     = mode;
    wait->next     = NULL;
    wait->ra       = ra;
    wait->sleep    = sleep;

    fs_lockEnqueueWaiter(context, lock, wait);

    NFX_SPIN_UNLOCK(&lock->spin);

    if (sleep) {
        return fs_lockSleep(context, lock, spinp, event);
    }
    else {
        LOCK_TRACE(("return for wait on lock %s",
                 fs_lockDescribe(context, lock, desc)));
        return FS_ERR_DEFERRED;
    }
}



/**
 * Attempt to acquire the specified lock in the specified mode, sleeping
 * until the lock becomes available.  The specified spin lock is released
 * while sleeping.
 *
 * Returns true if the lock is successfully acquired, false if the lock
 * was destroyed before it could be acquired, and never returns if the
 * lock never becomes available.
 */
boolean
fs_trylockSleep(
    fs_Context     *context,
    fs_Lock        *lock,
    fs_LockMode    mode,
    nfx_spinlock_t *spin,
    void           *ra)
{
    fs_lockOrderLock(context, lock, mode, FALSE);

    NFX_SPIN_LOCK(&lock->spin);

    if (fs_trylockSpin(context, lock, mode, ra)) {
        NFX_SPIN_UNLOCK(&lock->spin);
        return TRUE;
    }
    else {
        int rc = fs_lockWait(context, lock, mode, TRUE,
                             spin, FS_THREAD_EVENT_LOCK, NULL, FALSE, ra);
        return (rc == NFX_OK);
    }
}

/**
 * Return true if the specified context holds the specified lock
 * in exclusive mode.
 */
boolean
fs_lockOwnedExclusive(
	fs_Context *context,
	fs_Lock    *lock)
{
	return fs_lockOwned(context->smHandle, lock, FS_LOCK_MODE_EXCL);
}

/**
 * Return true if the specified context holds the specified lock
 * in shared mode.  Pref is the expected mode (simply returns this
 * value if shared locks are not tracked for this context.
 */
boolean
fs_lockOwnedShared(
	fs_Context *context,
	fs_Lock    *lock,
	boolean    pref)
{
    if (!FS_KEEP_TRACK_OF_LOCKS(context))
		return pref;
    else
        return (fs_lockGetSharedCount(context, lock) != 0);
}

/**
 * Attempt to acquire the specified lock in the specified mode, or queue
 * the specified context in the lock's acquisition queue.  Returns true if
 * the lock was acquired and false if not.
 */
boolean
fs_trylockEnqueue(
    fs_Context  *context,
    fs_Context  *notifyContext,
    fs_Lock     *lock,
    fs_LockMode mode,
    int32       event,
    void        *p2,
    boolean     allocateWaitStruct)
{
    int rc;
    void *ra = (void *) (address_t) get_ra();
    
    NFX_SPIN_LOCK(&lock->spin);

    if (!fs_lockIsLive(context, lock)) {
        NFX_SPIN_UNLOCK(&lock->spin);
        return FALSE;
    }

    if (fs_trylockSpin(context, lock, mode, ra)) {
        NFX_SPIN_UNLOCK(&lock->spin);
        fs_lockOrderLock(context, lock, mode, TRUE);
        return TRUE;
    }

    // Will do lockorder check when lock is acquired.
    rc = fs_lockWait(notifyContext, lock, mode, FALSE,
                     NULL, event, p2, allocateWaitStruct, ra);
    assert(rc == FS_ERR_DEFERRED);
    return FALSE;
}


/**
 * Dequeue and notify a SM waiting for a lock.
 *
 * Must be called with lock->spin held.
 */
static fs_LockWait *fs_lockDequeueWaiter(
    fs_Context  *context,
    fs_Lock     *lock,
    boolean     isDestroy)
{
    esm_handle_t    handle;
    esm_event_t     event;
    int32           rc;

    fs_LockWait     *waiting = lock->waiting;
    fs_LockWait     *next = waiting->next;

    assert(waiting != NULL);
    assert(lock->waitCount > 0);

    lock->waitCount -= 1;
    assert(waiting->lock == lock);
    waiting->lock = NULL;

    if (!isDestroy) {
        fs_Context *waiter = esm_getUserInfo(waiting->smHandle);
        lock->count += 1;

        assert(waiter->smHandle == waiting->smHandle);

        if (waiting->mode > lock->mode)
            lock->mode = waiting->mode;

        switch (waiting->mode) {
            
            case FS_LOCK_MODE_SHARE:
                // is this safe to do to another context?
                fs_lockAddShared(waiter, lock);
                break;
            
            case FS_LOCK_MODE_EXCL:
                fs_lockInitializeExclusiveCount(context, lock);
                lock->owner = waiter->smHandle;
                lock->ra = waiting->ra;
                fs_lockAddOwned(waiter, lock);
                break;

            default:
                assert(FALSE);
        }

        // If the waiting context didn't sleep, then we didn't
        // do the lock order check on it yet, so do it now.
        if (!waiting->sleep)
        {
            fs_lockOrderLock(waiter, lock, waiting->mode, TRUE);
        }
    }

    // Grab these, we can't touch waiting after
    // we've notified, since another CPU could start
    // running it and re-use waiting.
    handle = waiting->smHandle;
    event.event = waiting->event;
    event.p2.u.p = waiting->p2;
    event.p1.u.p = isDestroy ? NULL : lock;

    // Clear the handle, for debugging.
    waiting->smHandle = 0;

#ifdef FS_LOCK_WAITER_TIME
	eee_rtcDuration(&waiting->time, &lock->waitTime);
#endif

    if ((waiting->flags & FS_LOCK_WAIT_FLAG_ALLOC) != 0)
    {
        // TODO - return these too, so caller can chain them 
        // and dealloc outside of spinlocked loop.
        fs_ramDealloc(FS_RAM_LOCK_WAIT, waiting);
    }

    // Notify.
    rc = thread_event(handle, &event);
    if (rc != NFX_OK) {
        fs_error(context, FS_EVENT_DELIVERY_FAIL, "lockavail");
    }

    lock->waiting = next;
    return next;
}


/**
 * Dequeue waiters on the specified lock until the lock is held by one
 * which would block, or all waiters if isDestroy is TRUE.
 *
 * lock->spin must be held when calling this function.
 */
static void fs_lockDequeueWaiters(
    fs_Context  *context,
    fs_Lock     *lock,
    boolean     isDestroy)
{
    fs_LockWait *waiter = lock->waiting;
    FS_LOOP_DBG_DECL(dbgNumLoops);

    while (waiter != NULL)
    {
        char desc[FS_LOCK_DESCRIPTION];

        FS_LOOP_DBG_CHECK(dbgNumLoops, 5000);

        if (!isDestroy && !fs_lockCompat[lock->mode][waiter->mode])
            break;

        LOCK_TRACE(("esm %llx wait over on lock %s",
					(uint64)((uintptr_t)waiter->smHandle),
					fs_lockDescribe(context, lock, desc)));

        waiter = fs_lockDequeueWaiter(context, lock, isDestroy);
    }
}


/**
 * Steal the specified lock from another context.  The lock must be
 * owned exclusively.
 */
void fs_lockSteal(
    fs_Context      *context,
    fs_Lock         *lock)
{
    char desc[FS_LOCK_DESCRIPTION];
    fs_Context *old = esm_getUserInfo(lock->owner);

    LOCK_TRACE(("stealing %s (old context %p)",
             fs_lockDescribe(context, lock, desc), old));

    assert(old != NULL);
    assert(lock->owner != context->smHandle);
    fs_lockRemoveOwned(old, lock);
    fs_lockOrderUnlock(old, lock); // DEBUG
    fs_lockOrderLock(context, lock, FS_LOCK_MODE_EXCL, FALSE); // DEBUG
    lock->owner = context->smHandle;
    fs_lockAddOwned(context, lock);
}


/**
 * Give the specified lock to another context.  The lock must be
 * owned exclusively.
 */
void fs_lockGiveExclusive(
    fs_Context *context,
    fs_Context *newOwner,
    fs_Lock *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    LOCK_TRACE(("giving excl %s to %p",
                fs_lockDescribe(context, lock, desc), newOwner));

    assert(fs_lockOwnedExclusive(context, lock));
    fs_lockRemoveOwned(context, lock);
    fs_lockOrderUnlock(context, lock); // DEBUG
    fs_lockOrderLock(newOwner, lock, FS_LOCK_MODE_EXCL, FALSE); // DEBUG
    lock->owner = newOwner->smHandle;
    fs_lockAddOwned(newOwner, lock);
}


/**
 * Give the specified lock to another context.  The lock must be
 * owned shared.  Only a single instance of a shared hold on the
 * lock is transferred.  If the caller currently holds additional
 * shared references to the lock, they remain and are the
 * responsibility of the caller.
 */
void fs_lockGiveShared(
    fs_Context *context,
    fs_Context *newOwner,
    fs_Lock *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    LOCK_TRACE(("giving shared %s to %p",
                fs_lockDescribe(context, lock, desc), newOwner));

    assert(fs_lockOwnedShared(context, lock, TRUE));
    fs_lockRemoveShared(context, lock);
    fs_lockOrderUnlock(context, lock); // DEBUG
    fs_lockOrderLock(newOwner, lock, FS_LOCK_MODE_SHARE, FALSE); // DEBUG
    fs_lockAddShared(newOwner, lock);
}


/**
 * Perform lockCommit, optionally dequeueing
 * waiters.
 */
static void fs_doLockCommit(
    fs_Context  *context,
    fs_Lock     *lock,
    boolean     dequeueWaiters)
{
    assert(fs_lockOwnedExclusive(context, lock));

    lock->flags &= ~(FS_LOCK_FLAG_DIRTY | FS_LOCK_FLAG_ALLOW_STARVATION);
    fs_lockRemoveOwned(context, lock);
    lock->count = 0;
    lock->owner = 0;
    lock->mode = FS_LOCK_MODE_AVAIL;
    lock->ra = NULL;
    fs_lockRemoveSharedAll(context, lock);
    fs_lockResetExclusiveCount(context, lock);

    if (dequeueWaiters)
        fs_lockDequeueWaiters(context, lock, FALSE);
}


/**
 * Give the specified lock to another context, demoting
 * it to shared mode.  All references to the lock by the
 * giving context are removed, similar to fs_lockCommit.
 */
void fs_lockGiveDemote(
    fs_Context  *context,
    fs_Context  *newOwner,
    fs_Lock     *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    assert(lock->owner == context->smHandle);
    assert(lock->mode == FS_LOCK_MODE_EXCL);

    LOCK_TRACE(("giving w/demote %s to %p",
                fs_lockDescribe(context, lock, desc), newOwner));

    NFX_SPIN_LOCK(&lock->spin);
    fs_doLockCommit(context, lock, FALSE);
    lock->count = 1;
    lock->mode = FS_LOCK_MODE_SHARE;
    fs_lockAddShared(newOwner, lock);
    fs_lockOrderLock(newOwner, lock, FS_LOCK_MODE_SHARE, FALSE); // DEBUG

    fs_lockDequeueWaiters(context, lock, FALSE);
    NFX_SPIN_UNLOCK(&lock->spin);
}


/**
 * Release a filesystem lock that was acquired in exclusive mode.
 */
void fs_unlockExclusive(
    fs_Context *context,
    fs_Lock    *lock)
{
	boolean removed = FALSE;
    char desc[FS_LOCK_DESCRIPTION];
    LOCK_TRACE(("unlocking exclusive %s",
                fs_lockDescribe(context, lock, desc)));


    NFX_SPIN_LOCK(&lock->spin);

    assert(fs_lockOwnedExclusive(context, lock));

    if ((lock->flags & FS_LOCK_FLAG_DIRTY) != 0 && lock->count == 1) {
        // If the object is dirtied, we shouldn't be releasing our last
        // exclusive lock on it.
        fs_error(context, FS_INTERNAL_ERR,
                 "rel final excl lock on dirty buf %s",
                 fs_lockDescribe(context, lock, desc));
    }

    lock->count -= 1;
    fs_lockDecrementExclusiveCount(context, lock);
    if (lock->count == 0) {
        fs_lockRemoveOwned(context, lock);
        lock->mode = FS_LOCK_MODE_AVAIL;
        lock->owner = 0;
        lock->ra = NULL;
#ifndef FS_NO_LOCK_TIMEOUT
        lock->flags &= ~FS_LOCK_FLAG_ALLOW_STARVATION;
#else
        lock->flags &= ~(FS_LOCK_FLAG_ALLOW_STARVATION |
                         FS_LOCK_DISALLOW_EXCL_TIMEOUT);
#endif
		removed = TRUE;
    }

    fs_lockDequeueWaiters(context, lock, FALSE);

    NFX_SPIN_UNLOCK(&lock->spin);

	if (removed)
        fs_lockOrderUnlock(context, lock); // DEBUG
}

/**
 * Release a filesystem lock that was acquired in exclusive mode.
 * If the caller had acquired the lock multiple times, release all
 */
void fs_unlockExclusiveAll(
    fs_Context *context,
    fs_Lock    *lock)
{
	boolean removed = FALSE;
    char desc[FS_LOCK_DESCRIPTION];
    LOCK_TRACE(("unlocking exclusive %s",
                fs_lockDescribe(context, lock, desc)));


    NFX_SPIN_LOCK(&lock->spin);

    assert(fs_lockOwnedExclusive(context, lock));

    if ((lock->flags & FS_LOCK_FLAG_DIRTY) != 0 && lock->count == 1) {
        // If the object is dirtied, we shouldn't be releasing our last
        // exclusive lock on it.
        fs_error(context, FS_INTERNAL_ERR,
                 "rel final excl lock on dirty buf %s",
                 fs_lockDescribe(context, lock, desc));
    }

    lock->count = 0;
#if defined FS_DEBUG_LOCK
    fs_lockDecrementExclusiveCountAll(context, lock);
#endif
    if (lock->count == 0) {
        fs_lockRemoveOwned(context, lock);
        lock->mode = FS_LOCK_MODE_AVAIL;
        lock->owner = 0;
        lock->ra = NULL;
#ifndef FS_NO_LOCK_TIMEOUT
        lock->flags &= ~FS_LOCK_FLAG_ALLOW_STARVATION;
#else
        lock->flags &= ~(FS_LOCK_FLAG_ALLOW_STARVATION |
                         FS_LOCK_DISALLOW_EXCL_TIMEOUT);
#endif
		removed = TRUE;
    }

    fs_lockDequeueWaiters(context, lock, FALSE);

    NFX_SPIN_UNLOCK(&lock->spin);

	if (removed)
        fs_lockOrderUnlock(context, lock); // DEBUG
}


/**
 * Release a filesystem lock that was acquired in shared mode, with
 * the lock's spin lock already held.  Returns TRUE if the last
 * reference to the lock by this context is removed (DEBUG only).
 */
static boolean fs_unlockSharedNoSpin(
    fs_Context *context,
    fs_Lock    *lock)
{
	boolean	last;

    assert(lock->count > 0);

    lock->count -= 1;
    if (lock->count == 0) {
        lock->mode = FS_LOCK_MODE_AVAIL;
        lock->owner = 0;
        lock->flags &= ~FS_LOCK_FLAG_ALLOW_STARVATION;
    }
    last = fs_lockRemoveShared(context, lock);

    fs_lockDequeueWaiters(context, lock, FALSE);

	return last;
}


/**
 * Release a filesystem lock that was acquired in shared mode.
 */
extern void fs_unlockShared(
    fs_Context *context,
    fs_Lock    *lock)
{
	boolean last;
    char desc[FS_LOCK_DESCRIPTION];
    LOCK_TRACE(("unlocking shared %s",
             fs_lockDescribe(context, lock, desc)));

    NFX_SPIN_LOCK(&lock->spin);
    last = fs_unlockSharedNoSpin(context, lock);
    NFX_SPIN_UNLOCK(&lock->spin);

	if (last)
   		fs_lockOrderUnlock(context, lock);
}


/**
 * Mark a lock as having its object dirtied.
 */
void fs_lockSetDirty(
    fs_Context *context,
    fs_Lock    *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    assert(fs_lockOwnedExclusive(context, lock));

    LOCK_TRACE(("dirty lock %s",
                fs_lockDescribe(context, lock, desc)));

    NFX_SPIN_LOCK(&lock->spin);
    lock->flags |= FS_LOCK_FLAG_DIRTY;
    NFX_SPIN_UNLOCK(&lock->spin);
}


/**
 * Clear the object dirty flag in the lock.
 */
void fs_lockUnsetDirty(
    fs_Context *context,
    fs_Lock    *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    assert(fs_lockOwnedExclusive(context, lock));

    LOCK_TRACE(("undirty lock %s",
                fs_lockDescribe(context, lock, desc)));

    NFX_SPIN_LOCK(&lock->spin);
    lock->flags &= ~FS_LOCK_FLAG_DIRTY;
    NFX_SPIN_UNLOCK(&lock->spin);
}


/**
 * Release a filesystem lock at transaction commit.
 */
void fs_lockCommit(
    fs_Context *context,
    fs_Lock    *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    LOCK_TRACE(("commit unlock %s",
                fs_lockDescribe(context, lock, desc)));

    assert((lock->flags & FS_LOCK_FLAG_DIRTY) != 0);
    assert(fs_lockOwnedExclusive(context, lock));

    NFX_SPIN_LOCK(&lock->spin);
    fs_doLockCommit(context, lock, TRUE);
    NFX_SPIN_UNLOCK(&lock->spin);

  	fs_lockOrderUnlock(context, lock);
}


/**
 * Change the mode of an already held lock.
 * The specified newMode must be a "lesser" mode than the currently
 * held mode (eg. move from EXCL to SHARE).
 *
 * In practice, this can only move a mode from exclusive to shared.
 */
void fs_lockDemote(
    fs_Context      *context,
    fs_Lock         *lock,
    fs_LockMode     newMode)
{
    char desc[FS_LOCK_DESCRIPTION];

    assert(fs_lockOwnedExclusive(context, lock));
    assert(newMode == FS_LOCK_MODE_SHARE);
    assert(newMode < lock->mode);

    LOCK_TRACE(("demote to %s %s",
             fs_lockModes[newMode], fs_lockDescribe(context, lock, desc)));

    NFX_SPIN_LOCK(&lock->spin);

    fs_lockDecrementExclusiveCount(context, lock);
    if (lock->count == 1)
	{
    	fs_lockRemoveOwned(context, lock);
    	lock->owner = 0;
    	lock->mode = newMode;
        lock->ra = NULL;
	}
    fs_lockAddShared(context, lock);

    fs_lockDequeueWaiters(context, lock, FALSE);

    NFX_SPIN_UNLOCK(&lock->spin);
}


/**
 * Mark a lock as destroyed.  Caller must hold EXCLUSIVE access.
 * All waiters are delivered their event with param1 == NULL
 * (normally is lock).
 */
void fs_lockDestroy(
    fs_Context      *context,
    fs_Lock         *lock)
{
    char desc[FS_LOCK_DESCRIPTION];

    assert(fs_lockOwnedExclusive(context, lock));

    LOCK_TRACE(("destroy %s",
             fs_lockDescribe(context, lock, desc)));

    NFX_SPIN_LOCK(&lock->spin);

    lock->flags |= FS_LOCK_FLAG_DESTROY;

    fs_lockDequeueWaiters(context, lock, TRUE);

    assert(lock->waitCount == 0);

    fs_lockResetExclusiveCount(context, lock);
    
    fs_lockRemoveOwned(context, lock);

    lock->owner = 0;
    lock->count = 0;
    lock->flags = FS_LOCK_FLAG_DESTROY;
    lock->mode  = FS_LOCK_MODE_AVAIL;
    lock->ra    = NULL;

    NFX_SPIN_UNLOCK(&lock->spin);

    fs_lockOrderUnlock(context, lock); // DEBUG
}

/**
 * Free context locks lists.
 */
void
fs_lockContextFree(
	fs_Context  *context)
{

    if (context->sharedLocksPtr)   
    {
        FS_ASSERT (context->sharedLockSize > 0);
        
#ifdef FS_DEBUG_LOCK
        if ((address_t)context->sharedLocksPtr != (address_t)&context->sharedLocks)
#endif
        {
            fs_ramDealloc(FS_RAM_CTX_SHARED_LOCKS, context->sharedLocksPtr); 
        }
    }
    else
    {
        FS_ASSERT (context->sharedLockSize == 0);
    }
    context->sharedLockSize = 0;
    context->sharedLockCount = 0;
    context->sharedLocksPtr = NULL;

    context->lockWait.smHandle = 0;
    context->lockWait.next = NULL;
    DLL_INIT(&context->ownerList);
}

/**
 * Initialize context locks lists.
 */
void fs_lockContextInit(
    fs_Context  *context,
    uint32       sharedLocksListSize)
{
    DLL_INIT(&context->ownerList);

#if defined FS_DEBUG_LOCK
    if (sharedLocksListSize == 0)
        sharedLocksListSize = SHARED_LOCK_LIST_INIT;
#else
    if (context->flags & FS_CONTEXT_ALLOW_LOCK_STARVATION)  {
        sharedLocksListSize = 0;
    }
#endif

// quick hack to disable sharedlocks for the writeback deamon
if (sharedLocksListSize == 512)
 sharedLocksListSize = 0;
    
            
    if (sharedLocksListSize == 0)
    {
        context->sharedLocksPtr = NULL;
    }
#ifdef FS_DEBUG_LOCK
    else
    if (sharedLocksListSize == SHARED_LOCK_LIST_INIT)
    {
        context->sharedLocksPtr = (fs_SharedLockRef *)&context->sharedLocks;
    }
#endif
    else
    {
        context->sharedLocksPtr =
            fs_ramAlloc(FS_RAM_CTX_SHARED_LOCKS,
                        sizeof(fs_SharedLockRef) * sharedLocksListSize);

    }
    context->sharedLockSize = sharedLocksListSize;
    context->sharedLockCount = 0;

    NFX_SPIN_INIT(&context->lockListSpin,SPINLOCK_LEVEL_UNKNOWN);

    context->lockWait.smHandle = 0;
    context->lockWait.next = NULL;

#ifdef FS_DEBUG_LOCK_ORDER
    context->ignoreLockOrder = FALSE;
    context->lockHeight = 0;
    context->maxReservedInodeHeight = 0;
    context->lockOrderParnetInum = 0;
#endif
}

boolean
fs_lockmgr(fs_Context *context, fs_Lock *lock, nfx_spinlock_t *spin,
		   uint32 lockFlags)
{
    void *ra = (void *)get_ra();

    uint32 lockMode = (lockFlags & (FS_LOCK_MODE_SHARE|FS_LOCK_MODE_EXCL));
    fs_lockOrderLock(context, lock, lockMode, FALSE);
    NFX_SPIN_LOCK(&lock->spin);
    if (fs_trylockSpin(context, lock, lockMode, ra)) 
    {
        NFX_SPIN_UNLOCK(&lock->spin);
        return TRUE;
    }
    if (lockFlags & FS_LOCK_MODE_WAIT) 
    {
        // this will unlock the spinlock
        return (fs_lockWait(context, lock, lockMode, TRUE, spin,
                            FS_THREAD_EVENT_LOCK, NULL, FALSE, ra) == NFX_OK);
    }
    else
    {
        NFX_SPIN_UNLOCK(&lock->spin);
        return FALSE;
    }
}

#endif // SSC
