// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-lock.h
 *
 * RCS:  $Id: fs-lock.h,v 1.50 2002/11/20 01:06:22 rick Exp $
 *
 * Description:  File system lock interface
 * 
 * Created By:  Bob Fozard and Paul Haahr
 *
 * Date Created:  2001-07-24
 */

#if !defined(FS_LOCK_H)
#define FS_LOCK_H


/**
 * Types of objects to lock.  Used only for debugging.
 */
typedef enum {
    /** An invalid lock type, to catch uninitialized locks. */
    FS_LOCK_TYPE_INVALID = 0,
    /** Buffer */
    FS_LOCK_TYPE_BUF,
    /** Inode */
    FS_LOCK_TYPE_INODE,
    /** An in-core quote record. */
    FS_LOCK_TYPE_QUOTA,
    /** An in-core file set record */
    FS_LOCK_TYPE_FILESET,
    /** Traversing an in-core file set record */
    FS_LOCK_TYPE_FILESET_TRAVERSAL,
    /** The boot allocator. */
    FS_LOCK_TYPE_BOOT,
    /** The system-wide modification lock. */
    FS_LOCK_TYPE_ACCESS,
    /** The freezer thread lock. */
    FS_LOCK_TYPE_MODIFY,
    /** The freezer thread lock. */
    FS_LOCK_TYPE_FREEZE,
    /** A lock on the btree part of a Dinode */
    FS_LOCK_TYPE_BTREE,
    /** Lock to allow access to tape for dump threads */
    FS_LOCK_TYPE_TAPE,
    /** Lock to allow eek multithreading */
    FS_LOCK_TYPE_EEK,
    /** snapIndo accessLocks */
    FS_LOCK_TYPE_SNAP,
    /** End */
    FS_LOCK_TYPE_MAX
} fs_LockType;

// Must be even in order to avoid padding problem in
// logstats response.
NFX_ASSERT_STATIC(FS_LOCK_TYPE_MAX % 2 == 0);

/**
 * Used for tracking lock waits.
 */
#define FS_LOCK_DAEMON_OFFSET (FS_REQ_NFS_MAX)
#define FS_LOCK_OWNER_SHARED  (FS_DAEMON_TYPE_MAX + FS_LOCK_DAEMON_OFFSET)
#define FS_LOCK_OWNER_UNK     (FS_LOCK_OWNER_SHARED + 1)
#define FS_LOCK_OWNER_MAX     (FS_LOCK_OWNER_UNK + 1)

// Must be even in order to avoid padding problem in
// logstats response.
NFX_ASSERT_STATIC(FS_LOCK_OWNER_MAX % 2 == 0);

extern int fs_lockMaxSharedLockListSize;
extern uint32 fs_lockWaitStats[FS_LOCK_OWNER_MAX]  // requestor req type
                               [FS_LOCK_TYPE_MAX]   // lock type
                               [FS_LOCK_OWNER_MAX]; // owner req type




#if defined FS_DEBUG_LOCK_ORDER
/**
 * Lock order debugging.
 */
typedef struct fs_LockOrder
{
    fs_LockType     type;       // from above enum list
    boolean         isLeaf;     // no locks acquired after this held
    boolean         ignore;     // ignore this lock type 

} fs_LockOrder;

/**
 * To validate regular inode lock order when the caller knows
 * the parent directory, use this when the caller holds the
 * parent, otherwise strict inode number lock order will be
 * enforced.
 */
#define FS_LOCK_ORDER_SET_PARENT(_inum_)        \
        context->lockOrderParentInum = (_inum_);
#define FS_LOCK_ORDER_DISABLE(_context_)        \
        (_context_)->ignoreLockOrder = TRUE;
#define FS_LOCK_ORDER_ENABLE(_context_)         \
        (_context_)->ignoreLockOrder = FALSE;

#else // FS_DEBUG_LOCK_ORDER

#define FS_LOCK_ORDER_SET_PARENT(_inum_)
#define FS_LOCK_ORDER_DISABLE(_context_)
#define FS_LOCK_ORDER_ENABLE(_context_)

#endif // FS_DEBUG_LOCK_ORDER


/**
 * File system lock structure.
 */
struct fs_Lock {

    /** Lock used to protect wait list and other internals. */
    nfx_spinlock_t  spin;

    /** Number of owners.  (Really, count of acquire operations.) */
    int32           count;

    /** Locking mode:  available, shared, (write), exclusive */
    uint16          mode;

    /** Bitset of flags. */
    uint16          flags;

    /** State machine handle of "major" owner (exclusor or writer). */
    esm_handle_t    owner;

    /** Queue of waiting requestors. */
    fs_LockWait     *waiting;

    /** Record the address of the first caller for 
     *  the exclusive holder. */
    void            *ra;

    /** Accumulated wait time on this lock in rtc form */
    eeeRTC_t        waitTime;

    /** Total number of times a context had to wait for
     *  this lock. */
    uint64          totalWaits;

    /** The next lock owned by the context holding this lock. */
    DLL_ENTRY(fs_Lock)  ownerEntry;

    /** The count of exclusive acquisitions of this lock. */
    int32           exclusiveCount;

    /** Number of waiting requestors on waiting list.
     *  TODO - take this out, it's just for debug. */
    int32           waitCount;

    /** Type of object controlled by lock. */
    fs_LockType     type;
};


/**
 * Description of a shared lock and the number of times it is held.
 */
typedef struct {
    /** The lock. */
    fs_Lock *lock;
    /** The number of times it is held. */
    int count;
} fs_SharedLockRef;


/**
 * Lock structure flags.
 */
enum {
    /** Lock is marked for destruction. */
    FS_LOCK_FLAG_DESTROY = 0x01,
    /** Object lock pertains to was dirtied by current owner. */
    FS_LOCK_FLAG_DIRTY = 0x02,
    /** Lock access starvation is allowed. */
    FS_LOCK_FLAG_ALLOW_STARVATION = 0x04,
#ifdef FS_NO_LOCK_TIMEOUT
     FS_LOCK_DISALLOW_EXCL_TIMEOUT = 0x08
#endif
};


/**
 * Lock wait list element.
 */
struct fs_LockWait {
    esm_handle_t   smHandle;
    int32          event;
    void           *p2;
    uint16         mode;
    uint16         flags;
    fs_LockWait    *next;
    void           *ra;
    boolean        sleep;
    eeeRTC_t       time;
    fs_Lock        *lock;
};


/**
 * Lock wait element flags.
 */
enum {
    /** Wait structure was alloc'd by fs_lock() and needs to be freed. */
    FS_LOCK_WAIT_FLAG_ALLOC = 0x1
};


/**
 * Lock modes.
 */
typedef enum fs_LockMode {
    /** Internal use only. */
    FS_LOCK_MODE_AVAIL = 0,
    /** I can share access with others. */
    FS_LOCK_MODE_SHARE = 1,
    /** I need exclusive access. */
    FS_LOCK_MODE_EXCL = 2,
    /** Wait for lock if not available */
    FS_LOCK_MODE_WAIT = 4,
    /** Internal use only. */
    FS_LOCK_MODE_MAX,
} fs_LockMode;


/**
 * Set the type of a lock, for debugging purposes.
 */
static inline void fs_lockSetType(
    fs_Lock     *lock,
    fs_LockType type)
{
    lock->type = type;
}


/**
 * Initialize the specified lock guarding the specified type of object.
 */
static inline void
fs_lockInit(
    fs_Lock     *lock,
    fs_LockType type)
{
	extern void bzero(void * dst, __SIZE_TYPE__ length);

	assert(type != 0);
    bzero(lock, sizeof *lock);
    DLL_ISOLATE_ELEM(lock, ownerEntry);

    NFX_SPIN_INIT(&lock->spin,SPINLOCK_LEVEL_UNKNOWN);
    fs_lockSetType(lock, type); // DEBUG
}


/**
 * Return true if the specified lock is live;  that is, if it has not been
 * marked for destruction.
 *
 * Note that this function does not acquire the lock's spin lock, and is
 * meant to be called (potentially) without the spin lock held.  That
 * means that we need to be very careful if running in an incoherent
 * memory environment and must ensure that we don't reuse destroyed locks
 * while anyone is potentially looking at one.
 */
static inline boolean
fs_lockIsLive(
    fs_Context  *context,
    fs_Lock     *lock)
{
    return ((lock->flags & FS_LOCK_FLAG_DESTROY) == 0);
}




/**
 * Attempt to acquire the specified lock in the specified mode.  Returns
 * true if the lock was acquired and false otherwise.  Never sleeps.
 */
extern boolean fs_trylock(
    fs_Context  *context,
    fs_Lock     *lock,
    fs_LockMode mode);


/**
 * Attempt to acquire the specified lock in shared mode.  Returns true if
 * the lock was acquired and false otherwise.  Never sleeps.
 */
static inline boolean
fs_trylockShared(
	fs_Context  *context,
	fs_Lock     *lock)
{
    return fs_trylock(context, lock, FS_LOCK_MODE_SHARE);
}


/**
 * Attempt to acquire the specified lock exclusively.  Returns true if the
 * lock was acquired and false otherwise.  Never sleeps.
 */
static inline boolean
fs_trylockExclusive(
	fs_Context  *context,
	fs_Lock     *lock)
{
    return fs_trylock(context, lock, FS_LOCK_MODE_EXCL);
}


/**
 * Attempt to acquire the specified lock in the specified mode, or queue
 * the specified context in the lock's acquisition queue.  Returns true if
 * the lock was acquired and false if not.
 *
 * If the lock is not acquired, the lock may have been destroyed;  after
 * a false return, the caller should ensure that the lock is still live,
 * using fs_lockIsLive().
 *
 * If the lock is still live, the specified context is put on the
 * wait queue for the lock, using the specified event and (second) yield
 * parameter.  When the lock is acquired, the event will be delivered to
 * the context's state machine, with the lock's address as the first
 * parameter;  if the lock is destroyed, the event will be delivered,
 * but with a null first parameter.
 *
 * The specified context's wait structure is used if allocateWaitStruct is
 * false;  otherwise, a separate wait structure is allocated for this
 * request.
 */
extern boolean fs_trylockEnqueue(
    fs_Context  *context,
    fs_Context  *notifyontext,
    fs_Lock     *lock,
    fs_LockMode mode,
    int32       event,
    void        *p2,
    boolean     allocateWaitStruct);

/**
 * Wait for a filesystem lock.
 *
 * Must be called with lock->spin held;  the spin lock is dropped before
 * the call returns.
 */
extern int32 fs_lockWait(
    fs_Context     *context,
    fs_Lock        *lock,
    fs_LockMode    mode,
    boolean        sleep,
    nfx_spinlock_t *spinp,
    int32          event,
    void           *p2,
    boolean        allocateWaitStruct,
    void           *ra);

/**
 * Attempt to acquire the specified lock in the specified mode, sleeping
 * until the lock becomes available.  The specified spin lock is released
 * while sleeping.
 *
 * Returns true if the lock is successfully acquired, false if the lock
 * was destroyed before it could be acquired, and never returns if the
 * lock never becomes available.
 */
extern boolean
fs_trylockSleep(
    fs_Context     *context,
    fs_Lock        *lock,
    fs_LockMode    mode,
    nfx_spinlock_t *spin,
    void           *ra);


/**
 * Attempt to acquire the specified lock in shared mode, sleeping
 * until the lock becomes available.  The specified spin lock is released
 * before yielding and reacquired after lock acquisition.
 *
 * Returns true if the lock is successfully acquired, false if the lock
 * was destroyed before it could be acquired, and never returns if the
 * lock never becomes available.
 */
static inline boolean
fs_trylockSharedSleep(
	fs_Context     *context,
	fs_Lock        *lock,
	nfx_spinlock_t *spin)
{
    return fs_trylockSleep(context, lock, FS_LOCK_MODE_SHARE,
                           spin, (void *) (address_t) get_ra());
}


/**
 * Attempt to acquire the specified lock exclusively.  Returns true if the
 * lock was acquired and false otherwise.  The specified spin lock is
 * released before yielding and reacquired after lock acquisition.
 *
 * Returns true if the lock is successfully acquired, false if the lock
 * was destroyed before it could be acquired, and never returns if the
 * lock never becomes available.
 */
static inline boolean
fs_trylockExclusiveSleep(
	fs_Context     *context,
	fs_Lock        *lock,
	nfx_spinlock_t *spin)
{
    return fs_trylockSleep(context, lock, FS_LOCK_MODE_EXCL,
                           spin, (void *) (address_t) get_ra());
}


/**
 * Change the mode of an already held lock to a greater (more exclusive)
 * locking mode.  This is ONLY used by the inodeWrite() function to
 * provide exclusive access to the buf while modifying buf state for
 * logging, COW, etc.
 *
 * If the lock cannot be reacquired in the new mode immediately, this
 * function will sleep until it can.  If the lock is destroyed, this
 * function will abort the system.
 */
extern void fs_lockPromote(
    fs_Context      *context,
    fs_Lock         *lock,
    fs_LockMode     newMode);


/**
 * Steal the specified lock from another context.  The lock must be
 * owned exclusively.
 */
extern void fs_lockSteal(
    fs_Context      *context,
    fs_Lock         *lock);


/**
 * Give the specified lock to another context.  The lock must be
 * owned exclusively.
 */
extern void fs_lockGiveExclusive(
    fs_Context *context,
    fs_Context *newOwner,
    fs_Lock *lock);


/**
 * Give the specified lock to another context.  The lock must be
 * owned shared.  Only a single instance of a shared hold on the
 * lock is transferred.  If the caller currently holds additional
 * shared references to the lock, they remain and are the
 * responsibility of the caller.
 */
void fs_lockGiveShared(
    fs_Context *context,
    fs_Context *newOwner,
    fs_Lock *lock);


/**
 * Give the specified lock to another context, demoting
 * it to shared mode.  All references to the lock by the
 * giving context are removed, similar to fs_lockCommit.
 */
extern void fs_lockGiveDemote(
    fs_Context  *context,
    fs_Context  *newOwner,
    fs_Lock     *lock);

/**
 * Release a filesystem lock that was acquired in exclusive mode.
 */
extern void fs_unlockExclusive(
    fs_Context      *context,
    fs_Lock         *lock);
/**
 * Release a filesystem lock that was acquired in exclusive mode. Give
 * up all the exclusive locks for the caller
 */
extern void fs_unlockExclusiveAll(
    fs_Context      *context,
    fs_Lock         *lock);


/**
 * Release a filesystem lock that was acquired in exclusive mode.
 */
extern void fs_unlockShared(
    fs_Context      *context,
    fs_Lock         *lock);


/**
 * Called to mark a lock as having its object dirtied.
 */
extern void fs_lockSetDirty(
    fs_Context      *context,
    fs_Lock         *lock);


/**
 * Clear the object dirty flag in the lock.
 */
extern void fs_lockUnsetDirty(
    fs_Context *context,
    fs_Lock    *lock);


/**
 * Called to release a filesystem lock at transaction commit.
 */
extern void fs_lockCommit(
    fs_Context      *context,
    fs_Lock         *lock);


/**
 * Called to change the mode of an already held lock.  newMode must be a
 * "lesser" mode than the currently held mode.
 */
extern void fs_lockDemote(
    fs_Context      *context,
    fs_Lock         *lock,
    fs_LockMode     newMode);


/**
 * Mark a lock as destroyed.  Caller must hold EXCLUSIVE access.  All
 * waiters are delivered an error event.
 */
extern void fs_lockDestroy(
    fs_Context      *context,
    fs_Lock         *lock);

/**
 * Initialize context locks lists.
 */ 
extern void fs_lockContextInit(
    fs_Context  *context,
    uint32       sharedLocksListSize);

/**
 * Free context locks lists.
 */
extern void
fs_lockContextFree(
	fs_Context  *context);


/**
 * Number of characters needed for the description of a lock.
 */
enum { FS_LOCK_DESCRIPTION = 200 };


/**
 * Print a description of the specified lock into the specified string,
 * which must have at least FS_LOCK_DESCRIPTION bytes.  Returns the
 * description.
 */
extern const char *fs_lockDescribe(
    fs_Context *context,
    fs_Lock    *lock,
    char       string[FS_LOCK_DESCRIPTION]);



//   P R E D I C A T E S


/**
 * Acquire the specified lock in the specified mode.
 */
static inline void
fs_lock(
	fs_Context     *context,
	fs_Lock        *lock,
	fs_LockMode    mode,
	nfx_spinlock_t *spin,
	void           *ra)
{
    if (!fs_trylockSleep(context, lock, mode, spin, ra)) {
        fs_error(context, FS_INTERNAL_ERR,
                 "fs_lock: lock %p was destroyed", lock);
    }
}

/**
 * Acquire the specified lock exclusively.  Will sleep to obtain the
 * lock if it is not immediately available.  If the specified spinlock
 * is not NULL, it is released before yielding and reacquired after
 * the lock is acquired.
 *
 * If the lock is destroyed, this call will abort the server.
 */
static inline void
fs_lockExclusive(
    fs_Context     *context,
    fs_Lock        *lock,
    nfx_spinlock_t *spin)
{
    fs_lock(context, lock, FS_LOCK_MODE_EXCL, spin,
            (void *) (address_t) get_ra());
    assert(lock->count > 0);
}


/**
 * Acquire the specified lock in shared mode.  Will sleep to obtain the
 * lock if it is not immediately available.  If the specified spinlock
 * is not NULL, it is released before yielding and reacquired after
 * the lock is acquired.
 *
 * If the lock is destroyed, this call will abort the server.
 */
static inline void
fs_lockShared(
	fs_Context *context,
	fs_Lock *lock,
	nfx_spinlock_t *spin)
{
    fs_lock(context, lock, FS_LOCK_MODE_SHARE, spin,
            (void *) (address_t) get_ra());
    assert(lock->count > 0);
}


/*
 * Used only for asserts
 *
 * Indicates that lock is owned currently in exclusive mode
 * by the specified handle. We don't acquire the lock spin lock to
 * look at the values as assumption is that each of the field
 * can be accessed atomically in the main memory and a context
 * cannot be running on more than 1 processor at any given point
 * of time.
 */
static inline boolean
fs_lockOwned(
	esm_handle_t    handle,
	fs_Lock		    *lock,
	int		        mode)
{
    //char desc[FS_LOCK_DESCRIPTION];
	//LOCK_TRACE(("owned? %s %s",
    //         fs_lockModes[mode],
	//		 fs_lockDescribe(context, lock, desc));
	assert(mode == FS_LOCK_MODE_EXCL);

	return ((lock->count > 0)
            && (lock->mode == mode)
            && (lock->owner == handle));
}


/**
 * Return true if the specified context holds the specified lock
 * in exclusive mode.
 */
extern boolean
fs_lockOwnedExclusive(
	fs_Context *context,
	fs_Lock    *lock);


/**
 * Return true if the specified handle holds the specified lock
 * in exclusive mode.
 */
static inline boolean
fs_lockOwnedExclusiveHandle(
	esm_handle_t    handle,
	fs_Lock         *lock)
{
	return fs_lockOwned(handle, lock, FS_LOCK_MODE_EXCL);
}


/**
 * Return true if the specified context holds the specified lock
 * in shared mode.  Pref is the expected mode (simply returns this
 * value if shared locks are not tracked for this context.
 */
extern boolean
fs_lockOwnedShared(
	fs_Context *context,
	fs_Lock    *lock,
	boolean    pref);

/*
 * Held routines are only to be used under asserts for debug purposes
 * as they are not very solid implementations and cannot accurately
 * state whether lock is held or not
 */


/**
 * Return true if the lock is held in the given mode by any body.
 * This is a very loose test. It cannot be garuanteed that holder
 * of the lock will not release it while we think it is held
 */
static inline boolean
fs_lockHeld(
	fs_Context *context,
	fs_Lock    *lock,
	int        mode)
{
    return (lock->count > 0) && (lock->mode == mode);
}

/**
 * Return true if the specified lock is held in exclusive mode by any
 * context.
 */
static inline boolean
fs_lockHeldExclusive(
	fs_Context *context,
	fs_Lock    *lock)
{
	return fs_lockHeld(context, lock, FS_LOCK_MODE_EXCL);
}


/**
 * Return true if the specified lock is held in shared mode by any
 * context.
 */
static inline boolean
fs_lockHeldShare(
	fs_Context *context,
	fs_Lock    *lock)
{
	return fs_lockHeld(context, lock, FS_LOCK_MODE_SHARE);
}

/**
 * Inform the lock manager that the specified context is disappearing.
 * All locks held at this point should cause an error.
 */
extern void fs_lockTerminateContext(
    fs_Context *context);



/**
 * Print the locks owned by the specified context.
 */
extern void fs_lockPrintOwned(
    fs_Context *context);


/**
 * Initialize the lock order structures.
 */
extern void
fs_lockOrderInit();


/**
 * Delete a waiter from a lock's waiting list
 * Called at thread terminate time
 */
extern void fs_lockDeleteWaiter(
    fs_Context  *context);

extern boolean
fs_lockmgr(fs_Context *context, fs_Lock *lock, nfx_spinlock_t *spin,
           uint32 flags);

#endif /* FS_LOCK_H */
