/*
 * Name:  fs-mount.c
 *
 * RCS:  $Id: fs-mount.c,v 1.156 2002/12/10 18:42:46 bob Exp $
 *
 * Description:  Handling for mount/unmount requests.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"


//   C O R E   I N F O   T A B L E



/**
 * An array of the in-core root structures for filesystems.
 */
fs_FileSystem *fs_fileSystems[FS_FILE_SYSTEM_HASH_MAX];


/**
 * A spinlock for the file system table.
 */
nfx_spinlock_t fs_fileSystemTableLock;


/**
 * Initialize the file system table.
 */
void fs_initFileSystemTable()
{
    bzero(fs_fileSystems, sizeof(fs_fileSystems));
    NFX_SPIN_INIT(&fs_fileSystemTableLock,SPINLOCK_LEVEL_UNKNOWN);
}


/**
 * Return the hash code for a specified volume id.
 */
static int fs_fileSystemHash(
    evm_volId_t volId)
{
	// This hash is way expensive
    //return fs_coreHash(&volId, sizeof(evm_volId_t) / sizeof(uint64), 0) %
    //                        FS_FILE_SYSTEM_HASH_MAX;
	return volId.whole_thing % FS_FILE_SYSTEM_HASH_MAX;
}


/**
 * Return the FileSystem structure for a specified volume id, or NULL if the
 * volume is not known.  Assumes that the fileSystem spinlock is held in the
 * current context.
 */
fs_FileSystem *fs_fileSystemForVolumeLocked(
    evm_volId_t volId)
{
    fs_FileSystem *p;

    for (p = fs_fileSystems[fs_fileSystemHash(volId)];
         p != NULL;
         p = p->next)
	{
        if (p->volId.whole_thing == volId.whole_thing)
            return p;
    }

    return NULL;
}


/**
 * Return the FileSystem structure for a specified volume id, or NULL if the
 * volume is not known.
 */
fs_FileSystem *fs_fileSystemForVolume(
    evm_volId_t volId)
{
    fs_FileSystem *info;
    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    info = fs_fileSystemForVolumeLocked(volId);
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
    return info;
}


/**
 * Get list of volume ids up to maxVols
 * Return number of volume ids in list
 */
int32 fs_fileSystemGetVols(
    fs_Context *context,
    evm_volId_t *vols,
    int32 maxVols)
{
    fs_FileSystem *p;
    int32   i, n = 0;

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    for (i = 0; i < FS_FILE_SYSTEM_HASH_MAX; i++) {
        for (p = fs_fileSystems[i]; p != NULL; p = p->next, vols++, n++) {
            *vols = p->volId;
            if (n >= maxVols) {
                return n;
            }
        }
    }
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
    return n;
}


/**
 * Read a superblock for the volume specifed by this block.  Any superblock
 * that's in good enough shape is used.  Returns NULL if no superblock could
 * be read.
 */
static fs_SuperBlock *
fs_readAnySuperBlock(
    fs_Context *context)
{
    int i, n;

    int errorType = FS_INTERNAL_ERR;
    fs_SuperBlock *super = NULL;

    for (i = 0, n = fs_superBlockCount(); i < n; i++)
    {
        fs_RdwrReq *rdwr;
        uint32 savedCksum;

        if (super != NULL) {
            fs_ramDealloc(FS_RAM_READ_DATA, super);
            super = NULL;
        }

        rdwr = fs_rdwrAlloc();
        fs_volReadBlock(context, rdwr, fs_superBlockBlock(i));
        super = rdwr->evmStat.rDataHandle.dataPtr;
        fs_rdwrDealloc(rdwr);

        savedCksum = super->crc32;
        super->crc32 = 0 ;
        if (savedCksum != fs_checksum(super, sizeof(fs_SuperBlock), ~0))
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                    ("bad CRC in superblock %d", i));
            errorType = FS_BAD_SUPER_MAGIC;
            continue;
        } 

        if (super->magic != FS_SUPER_MAGIC)
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                    ("bad magic number in superblock %d", i));
            errorType = FS_BAD_SUPER_MAGIC;
            continue;
        }

        if (super->version != FS_SUPER_CURRENT_VERSION)
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                    ("bad version %d in superblock %d, expected %d",
                     super->version, i, FS_SUPER_CURRENT_VERSION));
            errorType = FS_BAD_SUPER_VERSION;
            continue;
        }

        if (super->flags & FS_SUPER_FLAG_MKFS)
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                    ("mkfs never finished superblock %d", i));
            errorType = FS_MKFS_INCOMPLETE;
            continue;
        }

        // hack to make the mem stats look right
        NFX_SPIN_LOCK(&fs_ramStatsSpinId);
        fs_ramStats[FS_RAM_STATS_ALLOC][FS_RAM_READ_DATA] -= 1;
        fs_ramStats[FS_RAM_STATS_ALLOC][FS_RAM_SUPER] += 1;
        NFX_SPIN_UNLOCK(&fs_ramStatsSpinId);

        return super;
    }

    if (super != NULL)
    {
        fs_ramDealloc(FS_RAM_READ_DATA, super);
        super = NULL;
    }
    fs_warn(context, errorType, "fs_readAnySuperBlock");
    return NULL;
}


/**
 * Ensure that a file system is clean, based on its superblock.  Tries
 * to clean the file system by replaying the log, if it isn't already
 * clean.  Returns TRUE if the file system is usable, and TRUE in
 * didReplay if a log replay was conducted.
 */
static boolean
fs_ensureClean(
    fs_Context    *context,
    fs_SuperBlock *super,
    boolean       skipReplay,
	boolean		  *didReplay)
{
    int32 rc;
    uint32 logSize, bufSize;
    fs_FileSystem *fs = context->fileSystem;
    fs_Time  now = fs_getTimeOfDay();

    fs->logRoot->logDevType = super->logDevType;
	*didReplay = FALSE;

    if (super->flags & FS_SUPER_FLAG_CLEAN)
    {
        FS_ASSERT((super->flags & FS_SUPER_FLAG_MUST_EEK) == 0);

        switch (fs->logRoot->logDevType) 
        {
        case FS_LOG_DEVTYPE_NVRAM:
            rc = nvram_find(NVRAM_CLIENT_EFS, fs->volId.whole_thing,
							&fs->logRoot->logDev.nvHandle);
            if (rc != NFX_OK) {
                rc = nvram_allocate (NVRAM_CLIENT_EFS, context->volId.whole_thing,
                                     fs->logSize, &fs->logRoot->logDev.nvHandle);
                if (rc != NFX_OK)
                {
                    // TODO - this ultimately should be a warning and not
                    // take the system down!
                    fs_warn(context, FS_INTERNAL_ERR, "nvram alloc");
                    return FALSE;
                }
            }
            else
            {
                // found
                // TODO - if !empty (by reading head/tail), print a
                // warning message
            }
            rc = fs_logInitDevice(context);
            if (rc != NFX_OK) 
            {
                fs_warn(context, FS_INTERNAL_ERR,
                        "failed to init log device: %d", rc);
                return FALSE;
            }
            break;
        case FS_LOG_DEVTYPE_SELF:
        case FS_LOG_DEVTYPE_VOLID:
            if ((rc = fs_logReadVolLogHdr(context, fs->logRoot,
                                          &logSize, &bufSize)) != NFX_OK)
            {
                rc = fs_logInitDevice(context);
                if (rc != NFX_OK) 
                {
                    fs_warn(context, FS_INTERNAL_ERR,
                            "failed to init log device: %d", rc);
                    super->flags |= FS_SUPER_FLAG_MUST_EEK;
                    return FALSE;
                }
            }
            else
            {
                if ((rc = fs_logReplay(context, TRUE, fs->logRoot)) != NFX_OK)
                {
                    fs_warn(context, FS_LOG_BAD_DEV_ERR,
                            "ensure clean, replay log error (%d)", rc);
                    super->flags |= FS_SUPER_FLAG_MUST_EEK;
                    return FALSE;
                }

                if ((fs->logRoot->head != fs->logRoot->tail) ||
                    (fs->logRoot->currLen != 0))
                {
                    fs_warn(context, FS_LOG_BAD_DEV_ERR,
                            "ensure clean, super block marked clean "
                            "but log head %d tail %d size%d",
                            fs->logRoot->head, fs->logRoot->tail,
                            fs->logRoot->currLen);
                    super->flags |= FS_SUPER_FLAG_MUST_EEK;
                    return FALSE;
                }
            }
            break;
        case FS_LOG_DEVTYPE_NULL:
            rc = fs_logInitDevice(context);
            break;
        default:
            fs_warn(context, FS_LOG_BAD_DEV_ERR, "ensure clean find clean");
            return FALSE;
        }

        return TRUE;
    }

    if (super->flags & FS_SUPER_FLAG_MUST_EEK)
    {
        fs_logStopDaemon(context, FALSE);

        super->logInitTimeSec = now.secs;
        super->logInitTimeNsec = now.nsecs;

        rc = fs_logInitDevice(context);
        if (rc != NFX_OK) 
        {
            fs_warn(context, FS_INTERNAL_ERR,
                    "failed to init log device: %d", rc);
            return FALSE;
        }
        return TRUE;
    }

	// Filesystem is dirty
    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("Volume is dirty, will check for log (devtype %d logsize %d)",
             fs->logRoot->logDevType, fs->logSize));

//printf("MOUNT FAILS UNTIL REPLAY IS STABLE\n");
//return FALSE;

    // Go and try to find the log
    switch (fs->logRoot->logDevType) 
    {
    case FS_LOG_DEVTYPE_NVRAM:
        rc = nvram_find(NVRAM_CLIENT_EFS, fs->volId.whole_thing,
						&fs->logRoot->logDev.nvHandle);
        break;
    case FS_LOG_DEVTYPE_SELF:
    case FS_LOG_DEVTYPE_VOLID:
        rc = fs_logReadVolLogHdr (context, fs->logRoot, &logSize, &bufSize);
        break;
    case FS_LOG_DEVTYPE_NULL:
        // No log, so can't find it
        rc = NFX_ERR;
        break;
    default:
        fs_warn(context, FS_LOG_BAD_DEV_ERR, "ensure clean find dirty");
        return FALSE;
    }

    if (rc != NFX_OK) 
    {
        // Log device was dirty, but we could not find it !!!
        // Allocate and init a new log device

        super->flags |= FS_SUPER_FLAG_MUST_EEK;

        switch (fs->logRoot->logDevType) 
        {
        case FS_LOG_DEVTYPE_NVRAM:
            rc = nvram_allocate (NVRAM_CLIENT_EFS, context->volId.whole_thing,
                                 fs->logSize, &fs->logRoot->logDev.nvHandle);
            if (rc != NFX_OK)
            {
                fs_warn(context, FS_INTERNAL_ERR, "nvram alloc");
                return FALSE;
            }
            break;
        case FS_LOG_DEVTYPE_SELF:
        case FS_LOG_DEVTYPE_VOLID:
            break;
        case FS_LOG_DEVTYPE_NULL:
            break;
        default:
            fs_warn(context, FS_LOG_BAD_DEV_ERR, "ensure clean reinit");
            return FALSE;
        }

        fs_logStopDaemon(context, FALSE);

        super->logInitTimeSec = now.secs;
        super->logInitTimeNsec = now.nsecs;

        rc = fs_logInitDevice(context);
        if (rc != NFX_OK) 
        {
            fs_warn(context, FS_INTERNAL_ERR,
                     "failed to init log device: %d", rc);
			return FALSE;
        }

        fs_warn(context, FS_INTERNAL_ERR, "dirty filesystem, no log");
        return FALSE;
    }

	// We've found our log, try to replay it.

    *didReplay = TRUE;
    rc = fs_logReplay(context, skipReplay, fs->logRoot);
    if (skipReplay)
    {
        rc = NFX_OK;
        *didReplay = FALSE;
    }

    if (rc != NFX_OK)
    {
        super->flags |= FS_SUPER_FLAG_MUST_EEK;
        fs_warn(context, FS_INTERNAL_ERR, "log replay failed: %d", rc);
        return FALSE;
    }

    // replay succesfull finished
    // update blksUsed and inodesUsed in superblock
    // transactionnal super block updates could have changed the values.
    context->fileSystem->super->blksUsed = context->fileSystem->commitedBlksUsed;
    context->fileSystem->super->inodesUsed = context->fileSystem->commitedInodesUsed;

    return TRUE;
}


/**
 * Create the in-core structure for a given file system, which is specified
 * by the given context's volume.  A superblock is passed in and ownership
 * of it is handed over to the fileSystem stucture.
 */
fs_FileSystem *
fs_makeFileSystem(
    fs_Context    *context,
    fs_SuperBlock *super)
{
    int32   hash;
	int32	fsetLimit;

    fs_FileSystem   *fileSystem;
    fs_FileSystem   *oldFileSystem;
	int             i;
    
    // TODO - define these limits properly
    // Set very low for stress testing
	fsetLimit = 10000;	// arrive at suitable limits

    fileSystem = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof (fs_FileSystem));
    bzero(fileSystem, sizeof(fs_FileSystem));

    NFX_SPIN_INIT(&(fileSystem->spin),SPINLOCK_LEVEL_UNKNOWN);
    NFX_SPIN_INIT(&(fileSystem->fs_statsReqRcvdSpinId),SPINLOCK_LEVEL_UNKNOWN);
    NFX_SPIN_INIT(&(fileSystem->superLock),SPINLOCK_LEVEL_UNKNOWN);
    fileSystem->volId = context->volId;
    fileSystem->state = FS_VOL_STATE_MOUNTING;
    fileSystem->super = super;
    fileSystem->logSize = super->logSize;
    fileSystem->params = fs_computeSizeParameters(super);

    fileSystem->logRoot = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof(fs_LogFsRoot));
    bzero(fileSystem->logRoot, sizeof(fs_LogFsRoot));
    NFX_SPIN_INIT(&fileSystem->logRoot->spin,SPINLOCK_LEVEL_UNKNOWN);
    fs_logInitLogSegments(context, fileSystem->logRoot, fileSystem->logSize);

    fs_lockInit(&fileSystem->accessLock, FS_LOCK_TYPE_ACCESS);
    fs_lockInit(&fileSystem->modifyLock, FS_LOCK_TYPE_MODIFY);
    fs_lockInit(&fileSystem->freezeLock, FS_LOCK_TYPE_FREEZE);
    fs_lockInit(&fileSystem->dumpContextsLock, FS_LOCK_TYPE_TAPE);
    fs_lockInit(&fileSystem->restoreContextsLock, FS_LOCK_TYPE_TAPE);

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    oldFileSystem = fs_fileSystemForVolumeLocked(context->volId);
    if (oldFileSystem != NULL)
    {
        fs_warn(context, FS_ALREADY_MOUNTED_ERR, "fs_makeFileSystem");
        fs_ramDealloc(FS_RAM_CORE_MISC, fileSystem->logRoot);
        fs_ramDealloc(FS_RAM_CORE_MISC, fileSystem);
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
        return NULL;
    }
    hash = fs_fileSystemHash(context->volId);
    fileSystem->next = fs_fileSystems[hash];
    fs_fileSystems[hash] = fileSystem;
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
	fileSystem->filesetCacheLimit = fsetLimit;

	NFX_SPIN_INIT(&fileSystem->filesetListLock,SPINLOCK_LEVEL_UNKNOWN);

	/*
	 * Now initialize the fset lists
	 */

	for (i = 0; i < FS_FILESET_CACHE_HASH_MAX; i++) {
		DLL_INIT(&(fileSystem->fileSetHash[i]));
	}
	DLL_INIT(&fileSystem->filesetList);

    fs_initAllocShadow(context, fileSystem);

    DLL_INIT(&fileSystem->activeThreadHdr);

    DLL_INIT(&fileSystem->inodeMrdHdr);
    for (i = 0; i < FS_LOG_MAX_SEGMENTS; i++)
    {
        DLL_INIT(&fileSystem->inodeLogSegmentHdr[i]);
    }
    DLL_INIT(&fileSystem->inodeWaiterHdr);

    /*
     * Initialize dump/restore contexts lists
     */
    DLL_INIT(&fileSystem->restoreContexts);
    DLL_INIT(&fileSystem->dumpContexts);

#if defined(FS_DEBUG_BUF_CKSUM)
    for (i = 0; i < FS_BUFLIST_HASH_MAX; i++)
    {
        DLL_INIT(&fileSystem->bufCksumHdr[i]);
        // bzero: fileSystem->bufCksumNbrEntries[i] = 0;
    }
#endif

    fs_lockInit(&fileSystem->inodeLock, FS_LOCK_TYPE_INODE);

    /*
     * Initialize the valid buddy region bitmap
     */
    NFX_SPIN_INIT(&fileSystem->buddyBitmapSpin,SPINLOCK_LEVEL_UNKNOWN);
    fileSystem->buddyRegions = 0;
    fs_buddyInitBitmap(context, fileSystem,
                       fileSystem->params.buddyRegionCount, TRUE);

#ifdef NEW_ZF
    /* Initialize the list of user data to zero.
     */
    //printf("fs_makeFileSystem() - init zfInodeHdr\n");
    DLL_INIT(&(fileSystem->zfInodeHdr));
    fileSystem->zfThreadHandle = 0;
    NFX_SPIN_INIT(&(fileSystem->zfSpin),SPINLOCK_LEVEL_UNKNOWN);
#endif // NEW_ZF
    
    return fileSystem;
}


/**
 * Set the log type to self.
 */
static int32
fs_setLogDevTypeToSelf(
    fs_Context      *context,
    fs_SuperBlock   *super,
    fs_MsgMountReq  *req)
{
    switch (super->logDevType)
    {
    case FS_LOG_DEVTYPE_NVRAM:
    case FS_LOG_DEVTYPE_VOLID:
    case FS_LOG_DEVTYPE_NULL:
    case FS_LOG_DEVTYPE_SELF:
        if (super->logSpaceSize)
        {
            super->logDevType = FS_LOG_DEVTYPE_SELF;
            super->logSize = super->logSpaceSize;
            super->logVolId.whole_thing = 0;
        }
        else
        {
            fs_warn(context, FS_LOG_BAD_DEV_ERR,
                "fs_mount: Can't switch to FS_LOG_DEVTYPE_SELF, no"
                "space reserved");
            return NFX_ERR;
        }
        break;
    default:
        fs_warn(context, FS_INTERNAL_ERR,
                "fs_mount: unknown super->logDevType %d", super->logDevType);
        return NFX_ERR;
    }

    //printf("setting new log type to self, size in super to %d\n", super->logSize);

    return NFX_OK;
}


/**
 * Set the log type to external volume.
 */
static int32
fs_setLogDevTypeToVol(
    fs_Context      *context,
    fs_SuperBlock   *super,
    fs_MsgMountReq  *req)
{    
    int32   adjustedReqLogSize;

    req->logSize = MIN(req->logSize, (FS_LOG_MAX_NBR_BLOCKS * FS_BLOCK_SIZE));

    if (req->logDevType == FS_LOG_DEVTYPE_DEFAULT)
    {
        adjustedReqLogSize = req->logSize;
    }
    else
    {
        adjustedReqLogSize = req->logSize -
                                    sizeof(vl_LogHdr) -
                   sizeof(vl_BufDiskHdr) * req->logSize / VL_BUFFER_SIZE;
    }

    // Make sure its an even multiple of segments.
    adjustedReqLogSize = (adjustedReqLogSize / FS_LOG_MAX_SEGMENTS) * FS_LOG_MAX_SEGMENTS;

    switch (super->logDevType)
    {
    case FS_LOG_DEVTYPE_NVRAM:
    case FS_LOG_DEVTYPE_SELF:
        if (!(super->flags & FS_SUPER_FLAG_CLEAN))
        {
            fs_warn(context, FS_LOG_BAD_DEV_ERR, 
                    "fs_mount: Can't change logDevType from NVRAM or SELF to "
                    "VOLID while not clean");
            return NFX_ERR;
        }
        break;
    case FS_LOG_DEVTYPE_VOLID:
        if ((super->logSize != adjustedReqLogSize ||
             super->logVolId.whole_thing != req->logVolId.whole_thing) &&
            !(super->flags & FS_SUPER_FLAG_CLEAN))
        {
            //printf("ssz %d adjsz %d svid 0x%llx rvid 0x%llx\n",
            //       super->logSize, adjustedReqLogSize,
            //       super->logVolId.whole_thing, req->logVolId.whole_thing);
 
            fs_warn(context, FS_LOG_BAD_DEV_ERR, 
                     "fs_mount: Can't change log device/size for VOLID "
                     "while not clean");
            return NFX_ERR;
        }
        break;    
    case FS_LOG_DEVTYPE_NULL:
        break;
    default:
        fs_warn(context, FS_INTERNAL_ERR,
                "fs_mount: unknown req->logDevType %d", super->logDevType);
        return NFX_ERR;
    }

    super->logDevType = FS_LOG_DEVTYPE_VOLID;
    super->logSize = adjustedReqLogSize;
    super->logVolId = req->logVolId;

    //printf("setting new log type to vol, size in super to %d\n", super->logSize);

    return NFX_OK;
}


/**
 * Set the log type to nvram.
 */
static int32
fs_setLogDevTypeToNvram(
    fs_Context      *context,
    fs_SuperBlock   *super,
    fs_MsgMountReq  *req)
{
    switch (super->logDevType)
    {
    case FS_LOG_DEVTYPE_VOLID:
    case FS_LOG_DEVTYPE_SELF:
        if (!(super->flags & FS_SUPER_FLAG_CLEAN))
        {
            fs_warn(context, FS_LOG_BAD_DEV_ERR, 
                    "fs_mount: Can't change logDevType from VOLID or SELF to "
                    "NVRAM while not clean");
            return NFX_ERR;
        }
        break;
    case FS_LOG_DEVTYPE_NULL:
    case FS_LOG_DEVTYPE_NVRAM:        
        break;
    default:
        fs_warn(context, FS_INTERNAL_ERR,
                "fs_mount: unknown req->logDevType %d", super->logDevType);
        return NFX_ERR;
    }

    super->logDevType = FS_LOG_DEVTYPE_NVRAM;
    super->logSize = req->logSize;

    //printf("setting new log type to nvram, size in super to %d\n", super->logSize);

    return NFX_OK;
}


/**
 * Set the log type to null.
 */
static int32
fs_setLogDevTypeToNull(
    fs_Context      *context,
    fs_SuperBlock   *super,
    fs_MsgMountReq  *req)
{        

    switch (super->logDevType)
    {
    case FS_LOG_DEVTYPE_NVRAM:
    case FS_LOG_DEVTYPE_VOLID:
    case FS_LOG_DEVTYPE_SELF:
        if (!(super->flags & FS_SUPER_FLAG_CLEAN))
        {
            fs_warn(context, FS_LOG_BAD_DEV_ERR, 
                    "fs_mount: Can't change logDevType from VOLID or SELF or NVRAM to "
                    "NULL while not clean");
            return NFX_ERR;
        }
        break;
    case FS_LOG_DEVTYPE_NULL:
        break;
    default:
        fs_warn(context, FS_INTERNAL_ERR,
                "fs_mount: unknown req->logDevType %d", super->logDevType);
        return NFX_ERR;
    }

    super->logDevType = FS_LOG_DEVTYPE_NULL;
    super->logSize = 16 * 1024 * 1024;

    //printf("setting new log type to null, size in super to %d\n", super->logSize);

    return NFX_OK;
}


/*-----------------------------------------------------------------------------
 * 
 * Set the LogDevType depending of previous Type (to be found in the 
 * super block, and the CLEAN state
 */
static int32
fs_setLogDevType (
    fs_Context      *context,
    fs_SuperBlock   *super,
    fs_MsgMountReq  *req)
{
    int32   reqDevType, rc;

    FS_TRACE(context, FS_TRACE_REQ,
            ("setLogDevType super %d req %d svid %llx rvid %llx ssz %d rsz %d",
            super->logDevType, req->logDevType,
            super->logVolId.whole_thing, req->logVolId.whole_thing,
             super->logSize, req->logSize));

    if (req->logDevType == FS_LOG_DEVTYPE_DEFAULT)
    {
        reqDevType = super->logDevType;
        req->logVolId = super->logVolId;
        req->logSize = super->logSize;
    }
    else
    {
        reqDevType = req->logDevType;
    }

    switch (reqDevType)
    {
    case FS_LOG_DEVTYPE_NVRAM:
        rc = fs_setLogDevTypeToNvram(context, super, req);
        break;
    case FS_LOG_DEVTYPE_VOLID:
        rc = fs_setLogDevTypeToVol(context, super, req);
        break;
    case FS_LOG_DEVTYPE_SELF:
        rc = fs_setLogDevTypeToSelf(context, super, req);
        break;
    case FS_LOG_DEVTYPE_NULL:
        rc = fs_setLogDevTypeToNull(context, super, req);
        break;
    default:
        fs_warn(context, FS_INTERNAL_ERR,
            "fs_mount: unknown req->logDevType %d", reqDevType);
        return NFX_ERR;
    }

    context->fileSystem->logSize = super->logSize;

    return rc;
}


/**
 * Create initial fileSystem structure prior to reading
 * superblock and updating it with correct parameters.
 * This is needed to facilitate exception handling in
 * reading the initial superblock.
 */
static fs_FileSystem *
fs_mountInitFs(
    fs_Context      *context,
    fs_SuperBlock   *super)
{
    fs_FileSystem   *fs;

    fs = context->fileSystem = fs_makeFileSystem(context, super);
    if (fs == NULL)
    {
        return NULL;
    }

    // Add ourselves to active thread list for the fs to
    // ensure exception during mkfs is properly handled.
    NFX_SPIN_LOCK(&fs->spin);
    DLL_INSERT_TAIL(&fs->activeThreadHdr, context, fsEntry);
    NFX_SPIN_UNLOCK(&fs->spin);
    
    return fs;
}


/**
 * Remove the specified filesystem from the table.
 */
static void
fs_removeFileSystem(
    fs_Context      *context,
    fs_FileSystem   *fs)
{
    fs_FileSystem               **prevfs;

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    for (prevfs = &fs_fileSystems[fs_fileSystemHash(context->volId)];;
         prevfs = &fs->next)
    {
        fs = *prevfs;
        if (fs == NULL)
        {
            fs_error(context, FS_VOLUME_NOT_MOUNTED, "fs_unmountThread");
        }
        if (fs->volId.whole_thing == context->volId.whole_thing)
            break;
    }
    assert(fs != NULL);
    *prevfs = fs->next;
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
}


PRIVATE void
fs_setTunables(fs_Context *context, fs_FileSystem *fs)
{
#if 0
    fs->percentMinFree = FS_DEFAULT_FS_MIN_FREE;
#else
    fs->percentMinFree = fs->super->percentMinFree;
    if (fs->percentMinFree > 100) {
        fs->percentMinFree = FS_DEFAULT_FS_MIN_FREE;
    }
#endif
}


/**
 * Do the work of mounting the filesystem.
 */
int32
fs_doMount(
	fs_Context		*context,
	fs_MsgMountReq	*req)
{
    fs_FileSystem   *fs = NULL;
    fs_SuperBlock   *super = NULL, *tmpSuper = NULL;
    boolean         mountProceed, replayDone, setBuddyBitsValid;
    boolean         disMount = FALSE, mustEek = FALSE, logInit = FALSE;
    int32           rc = NFX_ERR, loop;

	FS_TRACE(context, FS_TRACE_REQ, ("In mount thread"));

    if (fs_fileSystemForVolume(context->volId) != NULL)
    {
        //fs_warn(context, FS_ALREADY_MOUNTED_ERR, "fs_mountThread");
        FS_ELOG(context, LOG_WARNING, FS_TRACE_REQ,
                ("FILESYSTEM INFO: already mounted"));
        return FS_ERR_ALREADY_MOUNTED;
    }
    
    // Create a temporary in-core superblock and filesystem
    // to support exception handling in case one occurs prior
    // to instantiating correct in-core structures.
    tmpSuper = fs_ramAlloc(FS_RAM_SUPER, FS_MAX_BLOCK_SIZE);
    fs_mkfsInitializeSuperblock(FS_MIN_BLOCKS, tmpSuper, FS_MAX_BLOCK_SIZE);
    fs = fs_mountInitFs(context, tmpSuper);
    if (fs == NULL)
    {
        goto fs_doMountErr;
    }

    super = fs_readAnySuperBlock(context);
    if (super == NULL)
    {
        goto fs_doMountErr;
    }

    // Update in-core superblock with read superblock
    // and release the temporary one.
    fs->super = super;
    fs->params = fs_computeSizeParameters(super);
    fs->logSize = super->logSize;
    fs_ramDealloc(FS_RAM_SUPER, tmpSuper);
    tmpSuper = NULL;

    // shift the previous mount info
    for (loop = (FS_SUPER_NBR_MOUNTTIME - 1); loop > 0; loop--)
    {
        super->mountTime[loop] = super->mountTime[loop - 1];
        super->mountState[loop] = super->mountState[loop - 1];
        super->mountFail[loop] = super->mountFail[loop - 1];
    }

    super->nbrMount++;

    // put the volume name in the super block
    if (req->fsNameLen != 0)
    {
        strncpy(super->mountName, req->fsName,
                MIN(req->fsNameLen, FS_NAME_MAX));
        super->mountName[MIN(req->fsNameLen, FS_NAME_MAX)] = '\0';
    }

    // TODO - convert timestamp
    //printf("Volume %#llx last mounted as %s at %s\n",
    //       context->volId.whole_thing, super->mountName,
    //       fs_humandate(super->mountTime));

    if (fs_setLogDevType (context, super, req) != NFX_OK)
    {
        super->nbrMountFailedLogDev++;
        super->mountTime[0] = fs_getTimeOfDaySecs();
        super->mountState[0] = FS_VOL_STATE_OFFLINE;
        super->mountFail[0] = FS_MOUNT_FAIL_LOGDEV;
        fs_writeSuperblocksSync(context, super);

        goto fs_doMountErr;
    }

    // now that the logsize is set properly, we can init the segments
    fs_logInitLogSegments(context, fs->logRoot, fs->logSize);

    fs->logDevTypeTable = &fs_logDevTypeTable[super->logDevType];
    if (super->logDevType == FS_LOG_DEVTYPE_NVRAM)
        fs->persistentUserBufCache = TRUE;
    else
        fs->persistentUserBufCache = FALSE;

#ifdef NEW_ZF
    /* Initialize the list of user data to zero.
     */
    //printf("fs_doMount() - init zfInodeHdr\n");
    DLL_INIT(&(fs->zfInodeHdr));
    fs->zfThreadHandle = 0;
    NFX_SPIN_INIT(&(fs->zfSpin),SPINLOCK_LEVEL_UNKNOWN);
#endif // NEW_ZF
    
    /*
     * Start the writeback daemons now, log replay might need it
     * Log replay will call directly inode sync if necc
     */
    fs_bufStartWritebackDaemon(context);

    mountProceed = fs_ensureClean(context, super, req->skipReplay, &replayDone);
    logInit = TRUE;

    /* Initialize the in-core buddies.
     */
    fs->buddies = fs_ramAlloc(FS_RAM_CORE_MISC,
                              fs->params.buddyRegionCount * sizeof(fs_buddy_t));
    for (loop = 0; loop < fs->params.buddyRegionCount; loop++)
    {
        NFX_SPIN_INIT(&(fs->buddies[loop].spin),SPINLOCK_LEVEL_UNKNOWN);
        fs->buddies[loop].flags = 0;
        fs->buddies[loop].summary = NULL;
    } // for
    
    // reinitialize the buddy bitmap with new region count
    // set buddy bits valid only if log was clean and eek not necc
    setBuddyBitsValid = (replayDone == FALSE) &&
                        ((super->flags & FS_SUPER_FLAG_MUST_EEK) == 0);
    fs_buddyInitBitmap(context, fs,
                       fs->params.buddyRegionCount, setBuddyBitsValid);

    // Continue with a snapshot operation, if one was in progress
    // PR #1756: make sure snapshot resumes before eek and buddy
    // recalculation since snapshot might modify the reference bytes

    if (super->snapState.phase != FS_SNAP_PHASE_READY)
    {
        rc = fs_snapResume(context, super->snapState);
        if (rc != NFX_OK)
        {
            super->nbrMountFailedSnap++;
            fs_warn(context, FS_INTERNAL_ERR,
                    "fs_mountThread: snapshot continuation failed");
            fs_writeSuperblocksSync(context, super);
            goto fs_doMountErr;
        }
    }

    // check if we must eek before continuing the mount

    if ((super->flags & FS_SUPER_FLAG_MUST_EEK) && !req->eek)
    {
        // set up so eek is in the fix-it mode
        context->eek.defaultReply = FS_EEK_QUERY_YES;
        context->eek.interactive = TRUE;
        mustEek = TRUE;

        // since eagle agent is waiting for this thread which
        // may timeout because of eek, respond now with in-progress status
        // an on-line event after a successful mount will trigger the nfs share

        fs_threadRespond(context, FS_IN_PROGRESS);
#if defined(FS_STANDALONE)
        context->respHdr = NULL;
#else
        fs_contextFreeResponse(context);
#endif
    }

	if (!mountProceed || mustEek || req->eek)
	{
        // if we must eek, eek it and continue the mount if success
        if (mustEek)
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                ("Filesystem inconsistent, check forced"));
            disMount = FALSE;
        }
        else
        {
            disMount = TRUE;
        }

        rc = NFX_ERR;

		if (mustEek || req->eek)
		{
            if (super->snapState.phase != FS_SNAP_PHASE_READY)
            {
                FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                    ("Can't recover filesystem (eek), snapshot pending"));
                disMount = TRUE;
                super->flags |= FS_SUPER_FLAG_MUST_EEK;
                super->flags &= ~FS_SUPER_FLAG_CLEAN;
            }
            else
            {
                fs_lockExclusive(context, &context->fileSystem->accessLock,
                                 NULL);
                fs_lockExclusive(context, &context->fileSystem->modifyLock,
                                 NULL);

                // replay might've read in corrupted blocks
                // write them out before eek tries to fix them
                // (to avoid the duplicate cache problem)

                fs_syncThread(context);
                fs_inodeCacheFlush(context);
                fs_bufCacheFlush(context, NULL);

                rc = fs_eek(context);

                fs_unlockExclusive(context, &context->fileSystem->modifyLock);
                fs_unlockExclusive(context, &context->fileSystem->accessLock);
                if (rc != NFX_OK)
                {
                    FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                        ("Failed to recover filesystem (eek failure)"));
                    disMount = TRUE;
                    super->flags |= FS_SUPER_FLAG_MUST_EEK;
                    super->flags &= ~FS_SUPER_FLAG_CLEAN;
                }
                else
                {
                    // clear the must eek flag if eek was successful
                    // but don't clear the clean flag so we will kick off
                    // the brute-force recount of blocks and inodes below

                    super->flags &= ~FS_SUPER_FLAG_MUST_EEK;
                }
            }
		}
	}

    if (disMount)
    {
        if (!mountProceed)
        {
            super->nbrMountFailedReplay++;
       	    super->mountTime[0] = fs_getTimeOfDaySecs();
       	    super->mountState[0] = FS_VOL_STATE_OFFLINE;
       	    super->mountFail[0] = FS_MOUNT_FAIL_REPLAY;
        }
	
       	fs_writeSuperblocksSync(context, super);
        // eek might've been modifications, sync and flush 
        if (req->eek)
        {
            fs_syncThread(context);
            fs_inodeCacheFlush(context);
            fs_bufCacheFlush(context, NULL);
        }
        goto fs_doMountErr;
    }

#ifndef BUDDY_INCORE
    // start the buddy background thread only if replay was done
    // or eek was done (eek might not have fixed all buddy regions)

    if (replayDone || mustEek)
    {
        fs->buddyThreadHandle = fs_threadFork(context, fs_buddyRecalThread, 
                                              TRUE, FALSE,
                                              FS_DAEMON_TYPE_BUDDY, NULL);
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("mount: started buddy recal thread with max regions %d",
                 fs->buddyRegions));
    }
#endif // BUDDY_INCORE

#ifdef NEW_ZF
    /* Start the thread to zero fill the user data that haven't been commited.
     */
    if ((replayDone == TRUE) && (!DLL_IS_EMPTY(&(fs->zfInodeHdr))))
    {
        fs->zfThreadHandle = fs_threadFork(context, fs_userZeroThread, 
                                           TRUE, FALSE, // ???
                                           FS_DAEMON_TYPE_UZERO, NULL);
    }
#endif // NEW_ZF

#ifdef FS_RECALC_REF_AND_INODE
    //
    // Kick off the brute-force recount of blocks and inodes.
    //
    // NO longer needed, just done for verification.
    if (!(super->flags & FS_SUPER_FLAG_CLEAN))
    {
        fs_threadFork(context, fs_refByteCountThread, TRUE, FALSE,
                      FS_DAEMON_TYPE_REF_CNT, NULL);

        fs_threadFork(context, fs_inodeCountThread, TRUE, FALSE,
                      FS_DAEMON_TYPE_INO_CNT, NULL);
    }
#endif

    // TODO: if it was not clean, but ensureclean replayed completely
    // change to the new logdevice as/if indicated

    // Update the superblock
    super->flags &= ~FS_SUPER_FLAG_CLEAN;

    // printf("Mount as %s at time %s\n", super->mountName,
    //    fs_humandate(super->mountTime));

    super->mountTime[0] = fs_getTimeOfDaySecs();
    super->mountState[0] = FS_VOL_STATE_ONLINE;
    if (req->rdonly) 
    {
        super->mountState[0] = FS_VOL_STATE_ONLINE_RDONLY;
    }
    super->mountFail[0] = FS_MOUNT_NO_FAIL;
    
    fs_writeSuperblocksSync(context, super);

    // Calculate the min free blocks for current
    // conditions.
    fs_recalcMinFreeBlks(context);

    // Initialize the shadow allocator
    fs_initShadow(context, fs);

    // Create the trashcan daemon thread.
    fs->trashcanSignal = 1;
    fs->trashcanHandle = fs_threadFork(context, fs_trashcanThread, TRUE,
                                       FALSE, FS_DAEMON_TYPE_TRASH, NULL);
    //printf("trashcanThread handle 0x%x\n", fs->trashcanHandle);

    // Create the inode cleaner thread.
    fs->inodeCleanerSignal = 1;
    fs->inodeCleanerHandle = fs_threadFork(context, fs_inodeCleanerThread,
                                           TRUE, FALSE,
                                           FS_DAEMON_TYPE_INO_CLNR, NULL);
    //printf("inodeCleaner handle 0x%x\n", fs->inodeCleanerHandle);

	// Load the filesets before we mark the filesystem online
	FS_TRACE(context, FS_TRACE_REQ, ("Building fset"));
	fs_buildFileSetList(context, fs);
	FS_TRACE(context, FS_TRACE_REQ, ("Built fset"));

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs->state = FS_VOL_STATE_ONLINE;
    if (req->rdonly) {
        fs->state = FS_VOL_STATE_ONLINE_RDONLY;
    }

    /*
     * Set tunables if any required
     */
    fs_setTunables(context, fs);
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

    // initialize snapshot if ready

    if (super->snapState.phase == FS_SNAP_PHASE_READY)
    {
        fs_snapInit(context);
    }
    
    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("Volume online"));

    return NFX_OK;

// PR #1811 common error handling area

fs_doMountErr:

    // Remove ourselves from the active list
    // Added during fs_mountInitFs()
    if (!DLL_IS_ISOLATED(context, fsEntry))
    {
        NFX_SPIN_LOCK(&fs->spin);
        DLL_REMOVE(&fs->activeThreadHdr, context, fsEntry);
        NFX_SPIN_UNLOCK(&fs->spin);
    }

    // Dealloc any supers
    if (tmpSuper) fs_ramDealloc(FS_RAM_SUPER, tmpSuper);
    if (super) fs_ramDealloc(FS_RAM_SUPER, super);

    if (!fs)
    {
        // didn't get very far, return
        context->fileSystem = NULL;
        return NFX_ERR;
    }

    // Stop the log daemon
    if (logInit) 
        fs_logStopDaemon(context, FALSE);

    // Stop the writeback daemons
    fs_bufStopWritebackDaemonWait(context, FALSE);

    // Remove the fs from the global table
    fs_removeFileSystem(context, fs);

    if (fs->logRoot) fs_ramDealloc(FS_RAM_CORE_MISC, fs->logRoot);
    if (fs->buddyBitmap) fs_ramDealloc(FS_RAM_CORE_MISC, fs->buddyBitmap);
    if (fs->buddies) fs_ramDealloc(FS_RAM_CORE_MISC, fs->buddies);
    fs_ramDealloc(FS_RAM_CORE_MISC, fs);

    // WARNING:
    // currently no errors after filesystem threads have started
    // so no need to terminate them

    context->fileSystem = NULL;
    return NFX_ERR;

} /* fs_doMount */


/**
 * Handle the mount request.
 */
int32
fs_mountThread(
    fs_Context      *context)
{
    int32   rc;

    rc = fs_doMount(context, (fs_MsgMountReq *) context->reqHdr);
    if (rc == NFX_OK)
    {
#if !defined(FS_STANDALONE)
        /* Send an event that the volume is online.
         */
        fs_postEvt(context, EVENT_VOL_ONLINE, EVENT_STATE_UP);

        /*
         * Cleanup dump/restore residue. Remove files only if
         * we are not mounted as part of eek
         */

        if (!context->eek.inUse) {
                fs_cleanFileHistories(context, NULL);
                fs_cleanRestoreResidue(context);
        }
#endif
    }

    return rc;

} /* fs_mountThread */


/**
 * Signal active threads for the specified filesystem for
 * termination, block here until they've all terminated.
 */
static void
fs_unmountTerminateThreads(
    fs_Context      *context,
    fs_FileSystem   *fs)
{
    fs_Context      *threadContext;
    esm_event_t     event;
    int32           rc, numWaits = 0;

    FS_ASSERT(context->isThread);

    //printf("unmountTerminateThreads ioPending %d\n", fs->ioPendingCount);

    // Remove this context from the active thread list and
    // signal remaining contexts for termination.
    NFX_SPIN_LOCK(&fs->spin);
    if (!DLL_IS_ISOLATED(context, fsEntry))
    {
        DLL_REMOVE(&context->fileSystem->activeThreadHdr, context, fsEntry);
    }
    DLL_ITERATE(threadContext, &fs->activeThreadHdr, fsEntry)
    {
        //printf("unmount signalling smHandle %#x\n",
        //        threadContext->smHandle);

        event.event = FS_THREAD_EVENT_TERMINATE;
        rc = thread_event(threadContext->smHandle, &event);
        if (rc != NFX_OK)
        {
            fs_error(NULL, FS_EVENT_DELIVERY_FAIL,
                "unmount thread term");
        }
    }
    NFX_SPIN_UNLOCK(&fs->spin);

// about 60 seconds
#define MAX_WAITS   50
#define WAIT_MSECS  50

    // Now wait until all threads are terminated.
    NFX_SPIN_LOCK(&fs->spin);
    while (numWaits++ < MAX_WAITS)
    {
        if (!DLL_IS_EMPTY(&fs->activeThreadHdr))
        {
            NFX_SPIN_UNLOCK(&fs->spin);
            fs_sleep(context, numWaits * WAIT_MSECS);
            NFX_SPIN_LOCK(&fs->spin);
    
            if (numWaits > MAX_WAITS / 2)
            {
                FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                        ("unmount, still waiting for threads to terminate."));

                DLL_ITERATE(threadContext, &fs->activeThreadHdr, fsEntry)
                {
                    // try again to terminate
                    event.event = FS_THREAD_EVENT_TERMINATE;
                    rc = thread_event(threadContext->smHandle, &event);
    
                    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                            ("  retry terminate thread handle 0x%llx",
                             (uint64)((uintptr_t)threadContext->smHandle)));
                }
            }
        }
    }
    NFX_SPIN_UNLOCK(&fs->spin);
#undef MAX_WAITS
#undef WAIT_MSECS

}


/**
 * Wait for the log to indicate empty.
 */
boolean
fs_waitLogEmpty(
    fs_Context  *context)
{
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;
    int32               numWaits = 0;
    boolean             isEmpty = FALSE;

// about 60 seconds
#define MAX_WAITS   50
#define WAIT_MSECS  50

    while (numWaits++ < MAX_WAITS)
    {
        isEmpty = devTypeTable->logIsEmpty(context);
        if (isEmpty == TRUE)
            break;

        fs_sleep(context, numWaits * WAIT_MSECS);
    }

    if (!isEmpty)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                ("Unmount error log not empty after %d waits", numWaits));
        fs_warn(context, FS_INTERNAL_ERR,
                "unmount, sync, flush but still something in Log");
    }
#undef MAX_WAITS
#undef WAIT_MSECS

    if (numWaits > 1)
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("Unmount flushed the log after %d waits", numWaits));
    }

    return isEmpty;
}


/**
 * Do the real work of unmounting a volume.
 */
void
fs_doUnmount(
    fs_Context      *context,
    fs_FileSystem   *fs,
    boolean         isForce)
{
    int32                       i, j, loop;
    boolean                     logEmpty = FALSE;

    fs = context->fileSystem;

    if (!isForce)
    {
        // PR #1834: terminate any snapshot operation in progress
        if (fs_snapTerminateActive(context))
        {
            FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                    ("umount: terminating in progress snapshot operation"));
        }
        
        // PR #1789: let all current active threads complete
        // so all their acquired locks are commited
        fs_lockExclusive(context, &fs->accessLock, NULL);

        // Destroy all remaining threads.
        // If it is a forced unmout, there is no more active context.
        // This will terminate system threads that have released
        // the global accessLock before yielding/sleeping.
        fs_unmountTerminateThreads(context, fs);

        // Sync and flush the caches.
        fs_syncThread(context);
        fs_inodeCacheFlush(context);
        fs_bufCacheFlush(context, NULL);

        // Ensure all pending IO is completed.
        fs_volWaitPendingIO(context, FALSE);

        // Stop the buf writeback daemons.  For
        // a forced unmount, its already been done.
        fs_bufStopWritebackDaemonWait(context, FALSE);

#if defined(FS_DEBUG)
        // detect log not empty condition
        for (i = 0; i < FS_BUFLIST_MAX; i++)
        {
            for (j = 0; j < FS_BUF_CACHE_ITYPE_MAX; j++)
            {
                FS_ASSERT(fs->bufList[i].count.dirtyBufs[j] == 0);
            }
        }
        FS_ASSERT(DLL_IS_EMPTY(&fs->inodeMrdHdr));
        FS_ASSERT(fs->inodeCacheCount.numDirty == 0);
#endif

        // Stop the log daemon.  For
        // a forced unmount, its already been done.
        logEmpty = fs_waitLogEmpty(context);
        fs_logStopDaemon(context, FALSE);

        // Ensure all pending IO is completed.
        fs_volWaitPendingIO(context, FALSE);

        // Final cache flush.
        fs_inodeCacheFlush(context);
        fs_bufCacheFlush(context, NULL);

        // release the global access locks
        fs_unlockExclusive(context, &fs->accessLock);
    }

#if defined(FS_DEBUG)
    for (i = 0; i < FS_BUFLIST_MAX; i++)
    {
        for (j = 0; j < FS_BUF_CACHE_ITYPE_MAX; j++)
        {
            FS_ASSERT(fs->bufList[i].count.dirtyBufs[j] == 0);
            FS_ASSERT(fs->bufList[i].count.usedBufs[j] == 0);
        }
    }
    FS_ASSERT(DLL_IS_EMPTY(&fs->inodeMrdHdr));
    FS_ASSERT(fs->inodeCacheCount.numDirty == 0);
    FS_ASSERT(fs->inodeCacheCount.numCached == 0);
#endif

    if (isForce)
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("Forcibly unmounting volume"));
    }
    else
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("Unmounting volume"));

        fs->super->nbrUmount++;

        // shift the previous mount info
        for (i = (FS_SUPER_NBR_MOUNTTIME - 1); i > 0; i--)
        {
            fs->super->umountTime[i] = fs->super->umountTime[i - 1];
            fs->super->umountState[i] = fs->super->umountState[i - 1];
        }
        fs->super->umountTime[0] = fs_getTimeOfDaySecs();

        if (logEmpty)
        {
            boolean buddyDone = TRUE;

            // don't mark clean if the buddy recal thread has not completed
            NFX_SPIN_LOCK(&fs->spin);
            if (fs->buddyThreadHandle != 0)
                buddyDone = FALSE;
            NFX_SPIN_UNLOCK(&fs->spin);
            if (buddyDone)
                fs->super->flags |= FS_SUPER_FLAG_CLEAN;
            fs->super->umountState[0] = FS_UMOUNT_NORMAL;
        }
        else
        {
            fs->super->flags &= ~FS_SUPER_FLAG_CLEAN;
            fs->super->flags |= FS_SUPER_FLAG_MUST_EEK;
            fs->super->umountState[0] = FS_UMOUNT_FAIL_LOG_NOT_EMPTY;
        }

        // For correct exception handling in case of error during
        // write of superblocks, put this context back on the active
        // list for this filesystem.
        NFX_SPIN_LOCK(&fs->spin);
        DLL_INSERT_HEAD(&context->fileSystem->activeThreadHdr,
                        context, fsEntry);
        NFX_SPIN_UNLOCK(&fs->spin);

        fs_writeSuperblocksSync(context, fs->super);
    }

    /* Free the in-core buddies.
     */
    if (fs->buddies != NULL)
    {
        for (loop = 0; loop < fs->params.buddyRegionCount; loop++)
        {
            if (fs->buddies[loop].summary != NULL)
            {
                fs_ramDealloc(FS_RAM_REGION_SUMMARY,
                              fs->buddies[loop].summary);
                fs->buddies[loop].summary = NULL;
            }
        } // for
        fs_ramDealloc(FS_RAM_CORE_MISC, fs->buddies);
    }

#ifdef NEW_ZF
    /* Destroy the list of user data to zero if not empty.
     */
    NFX_SPIN_LOCK(&(fs->zfSpin));
    if (!DLL_IS_EMPTY(&(fs->zfInodeHdr)))
    {
        fs_zfInode * currInode = DLL_NIL;
        fs_zfInode * nextInode = DLL_NIL;
        
        DLL_ITERATE_NEXT(currInode, nextInode, &(fs->zfInodeHdr), entry)
        {
            fs_zfDelPage(context, currInode->inum, 0, TRUE);
        } // DLL_ITERATE_NEXT
        
        FS_ASSERT(DLL_IS_EMPTY(&(fs->zfInodeHdr)));
        fs->zfThreadHandle = 0;
    }
    NFX_SPIN_UNLOCK(&(fs->zfSpin));
#endif // NEW_ZF
        
    // Remove fileSystem from the table.
    fs_removeFileSystem(context, fs);

	/*
	 * Free all the quotas and fsets that we have allocated
	 */

    fs_cleanFileSetNQuotaLists(context, fs);

    /*
     * Free any dump/restore residue
     */

	fs_destroyDumpContextOnUmount(context);
	fs_destroyRestoreContextOnUmount(context);

    fs_freeAllocShadow(context, fs);

#if defined(FS_DEBUG_BUF_CKSUM)
    fs_bufReleaseAllCksum (context);
#endif

    if (fs->buddyBitmap)
		fs_ramDealloc(FS_RAM_CORE_MISC, fs->buddyBitmap);

    fs_ramDealloc(FS_RAM_SUPER, fs->super);
    fs_ramDealloc(FS_RAM_CORE_MISC, fs->logRoot);
    fs_ramDealloc(FS_RAM_CORE_MISC, fs);
    context->fileSystem = NULL;
}


/**
 * Handle the unmount request.
 */
int32
fs_unmountThread(
    fs_Context      *context)
{
    fs_FileSystem   *fs;

	FS_TRACE(context, FS_TRACE_REQ, ("+++ unmountThread +++\n"));

    NFX_SPIN_LOCK(&fs_fileSystemTableLock);
    fs = fs_fileSystemForVolumeLocked(context->volId);
    if (fs == NULL ||
        fs->state == FS_VOL_STATE_UNMOUNTING ||
        fs->state == FS_VOL_STATE_FAILING)
    {
        //fs_warn(context, FS_VOLUME_NOT_MOUNTED, "fs_unmountThread");
        FS_ELOG(NULL, LOG_WARNING, FS_TRACE_REQ,
                ("FILESYSTEM INFO: volId (0x%llx) not mounted",
                 context->volId.whole_thing));
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
        return FS_ERR_NOT_MOUNTED;
    }
    fs->state = FS_VOL_STATE_UNMOUNTING;
    NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

    context->fileSystem = fs;

    fs_doUnmount(context, fs, FALSE);

    context->respSize = sizeof(fs_MsgRespHdr);
    context->fileSystem = NULL;

	FS_TRACE(context, FS_TRACE_REQ, ("--- unmountThread ---\n"));


#if !defined(FS_STANDALONE)
    /* Send an event that the volume is offline.
     */
     fs_postEvt(context, EVENT_VOL_OFFLINE, EVENT_STATE_DOWN);
#endif

    return NFX_OK;

} /* fs_unmountThread */

#endif // SSC


