/*
 * Name:  fs-nvram.c
 *
 * RCS:  $Id: fs-nvram.c,v 1.13 2002/09/23 20:01:26 janv Exp $
 *
 * Description:  The file system specific functions for nvram
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  11/10/2001
 */

#if !defined(SSC)

#include "fs-includes.h"


/* 
 * Name: ck_sum
 *
 * Description: calculate/verify the checksum
 *
 *   Input: 
 *   - start ptr (must be int aligned !!)
 *   - end ptr (not used) (must be int aligned !!)
 *   - ptr to cksum
 *   Output 
 *   - if cksum == NULL : just return the result
 *     else  update the check sum
 *   - result == 0 if OK
 *
 *   To verify: call with arg cksum set to NULL
 *   To set: init (*cksum) to 0, call the function
 *           with arg *cksum set
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 */
static int 
ck_sum (void *start, void *end, uint32 *cksum)
{
    int *st = start;
    int *en = end;
    int sum = 0;
    int result;
  
    while (st != en) 
    {
        sum += *st;
        st++;
    }

    if (cksum == NULL)
    {
        result = - sum ;
    }
    else
    {
        result = sum + *cksum;
        *cksum = - sum ;
    }

    return (result) ;  
}

/*-----------------------------------------------------------------------------
 * fs_logReadNvramDevHdr
 *
 * Description: Called to read the device header
 *              and also to verify the cksum, version, magic etc
 *              return NULL if not found.
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  9/20/2001
 */

extern int32
fs_logReadNvramDevHdr(
    fs_LogFsRoot    *logRoot,
    fs_LogDevHdr    *devHdr)
{
    nvram_rw_t          rw;
    int32               lenp[1], rc;

    FS_ASSERT (logRoot->logDevType == FS_LOG_DEVTYPE_NVRAM);

    lenp[0] = sizeof(fs_LogDevHdr);

    rw.handle = logRoot->logDev.nvHandle;
    rw.numVecs = 1;
    rw.datap = (uchar8 **) &devHdr;
    rw.lenp = lenp;
    rw.offset = 0;
    rw.doWrap = 0;

    rc = nvram_read(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_READ_ERR, "nvram");
        return (NFX_ERR);
    }

    if (devHdr->magic != FS_LOG_MAGIC)
    {
        FS_TRACE(NULL, FS_TRACE_LOG, ("fs_logReadDevHdr: bad magic"));
        goto Read2ndDevHdr;
    }
    if (devHdr->version != FS_LOG_VERSION_ONE)
    {
        FS_TRACE(NULL, FS_TRACE_LOG, ("fs_logReadDevHdr: bad version"));
        goto Read2ndDevHdr;
    }
    if (ck_sum (devHdr, &(devHdr->spare), NULL) != 0)
    {
        FS_TRACE(NULL, FS_TRACE_LOG, ("fs_logReadDevHdr: bad cksum"));
        goto Read2ndDevHdr;
    }

    return (NFX_OK);

Read2ndDevHdr:

    rw.offset = sizeof(fs_LogDevHdr);

    rc = nvram_read(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_READ_ERR, "nvram");
        return (NFX_ERR);
    }

    if (devHdr->magic != FS_LOG_MAGIC)
    {
        fs_error(NULL, FS_LOG_BAD_MAGIC, "nvram");
        return (NFX_ERR);
    }
    if (devHdr->version != FS_LOG_VERSION_ONE)
    {
        fs_error(NULL, FS_LOG_BAD_VERSION, "nvram");
        return (NFX_ERR);
    }
    if (ck_sum (devHdr, &(devHdr->spare), NULL) != 0)
    {
        fs_error(NULL, FS_LOG_BAD_VERSION, "nvram");
        return (NFX_ERR);
    }

    // the  2nd DevHdr is OK, copy this to the first !!
    rw.offset = 0;
    rc = nvram_read(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_WRITE_ERR, "nvram");
        return (NFX_ERR);
    }

    return (NFX_OK);
};

/*-----------------------------------------------------------------------------
 * fs_logNvramWriteDevHdr
 *
 * Description: Called to write the device header
 *              and also update the cksum, version, magic etc
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  9/20/2001
 */

extern int32
fs_logNvramWriteDevHdr(
    fs_Context      *context,
    int32           operation,
    uint32          count)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    nvram_rw_t          rw;
    int32               lenp[1], rc;
    fs_LogDevHdr        devHdr;
    fs_LogDevHdr        *pdevHdr = &devHdr;


    rc = fs_logReadNvramDevHdr(logRoot, &devHdr);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_READ_ERR, "fs_logWriteDevHdr");
        return NFX_ERR;
    }

    lenp[0] = sizeof(fs_LogDevHdr);

    rw.handle = logRoot->logDev.nvHandle;
    rw.numVecs = 1;
    rw.datap = (uchar8 **) &pdevHdr;
    rw.lenp = lenp;
    rw.offset = 0;
    rw.doWrap = 0;

    switch (operation)
    {
    case FS_LOG_WRITE_HEAD:
        FS_TRACE(NULL, FS_TRACE_LOG, ("fs_logWriteNvramDevHdr: head %d (%d old)", 
                                      count, devHdr.head));
        devHdr.head = count;
        break;
    case FS_LOG_WRITE_INCR_COMMITS:
        FS_TRACE(NULL, FS_TRACE_LOG, ("fs_logWriteNvramDevHdr: nbrCommits %d + %d", 
                                      devHdr.nbrCommits, count));
        devHdr.nbrCommits += count;
        break;
    case FS_LOG_WRITE_DECR_COMMITS:
        FS_TRACE(NULL, FS_TRACE_LOG, ("fs_logWriteNvramDevHdr: nbrCommits %d - %d", 
                                      devHdr.nbrCommits, count));
        devHdr.nbrCommits -= count;
        break;        
    }

    devHdr.cksum = 0;
    ck_sum (&devHdr, &(devHdr.spare), &(devHdr.cksum));

    rc = nvram_write(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_WRITE_ERR, "nvram");
        return (NFX_ERR);
    }

    // write 2 times, if crash during first write, we still have the 2nd
    rw.offset = sizeof(fs_LogDevHdr);
    rc = nvram_write(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_WRITE_ERR, "nvram");
        return (NFX_ERR);
    }

    return (NFX_OK);
} // fs_logWriteNvramDevHdr


/*-----------------------------------------------------------------------------
 * fs_logReadNvram
 *
 * Description: Called to read something in the NVRAM
 *              No path of return, if it fails
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  9/20/2001
 */

extern void
fs_logReadNvram(
    fs_LogFsRoot    *logRoot,
    uchar8          *data,
    uint32          offset,
    uint32          size)
{
    nvram_rw_t          rw;
    int32               lenp[1], rc;

    FS_ASSERT (logRoot->logDevType == FS_LOG_DEVTYPE_NVRAM);

    rw.handle = logRoot->logDev.nvHandle;
    rw.numVecs = 1;
    rw.datap = &data ;
    lenp[0] = size;
    rw.lenp = lenp;
    rw.offset = offset;
    rw.doWrap = 1;
    rw.wrapReserve = FS_LOG_RESERVED_BYTES;

    rc = nvram_read(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_READ_ERR, "nvram");
    }

} //fs_logReadNvram


/*-----------------------------------------------------------------------------
 * fs_logNvramWrite
 *
 * Description: Called to write something in the NVRAM
 *              No path of return, if it fails
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  9/20/2001
 */

extern void
fs_logNvramWrite(
    fs_Context          *context,
    uint32              *offset,
    uchar8              *datap[LOG_IOVEC_MAX],
    uint32              lenp[LOG_IOVEC_MAX],
    fs_LogStableWrite   stable,
    uint64              inum)
{
    fs_LogUndoSummary   *undo, *undoNext;
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    nvram_rw_t          rw;
    int32               rc, bytesDone, vec;
    fs_LogId            eol = FS_LOG_ID_EOL;

    FS_ASSERT (logRoot->logDevType == FS_LOG_DEVTYPE_NVRAM);

    datap[LOG_IOVEC_EOL] = (uchar8 *) &eol;
    lenp[LOG_IOVEC_EOL] = FS_LOG_ID_BYTES;

    // Back up the tail pointer so that we'll overwrite the
    // previous EOL.
    if (logRoot->head != logRoot->tail)
    {
        FS_LOG_DECR_PTR(((fs_FileSystem *)context->fileSystem)->logSize,
                        logRoot->tail,
                        FS_LOG_ID_BYTES);
    }

    *offset = logRoot->tail;

    rw.handle = logRoot->logDev.nvHandle;
    rw.numVecs = LOG_IOVEC_MAX;
    rw.datap = datap;
    rw.lenp = lenp;
    rw.offset = *offset;
    rw.doWrap = 1;
    rw.wrapReserve = FS_LOG_RESERVED_BYTES;

    rc = nvram_write(&rw);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_LOG_READ_ERR, "nvram");
    }

    for (bytesDone = 0, vec = 0; vec < LOG_IOVEC_MAX; vec += 1)
    {
        bytesDone += lenp[vec];
    }

    FS_LOG_INCR_PTR(context->fileSystem->logSize,
                    logRoot->tail,
                    bytesDone);

    if ((stable == FS_LOG_WRITE_COMMIT_STABLE) ||
        (stable == FS_LOG_WRITE_COMMIT_UNSTABLE))
    {
        DLL_ITERATE_NEXT(undo, undoNext, &context->logSummary->undoHdr,
                         logEntry)
        {
            fs_Buf  *buf = undo->objSumm.objectPtr;

            FS_ASSERT(undo->objSumm.objectType == FS_LOG_DEP_BUF);
            FS_ASSERT(buf->flags & FS_BUF_TYPE_META);

            DLL_REMOVE(&context->logSummary->undoHdr, undo, logEntry);

            NFX_SPIN_LOCK(&buf->spin);
            DLL_REMOVE(&buf->undoHdr, undo, bufEntry);
            DLL_INSERT_TAIL(&buf->candoHdr, undo, bufEntry);
            NFX_SPIN_UNLOCK(&buf->spin);
        }
    }

} // fs_logNvramWrite


/*-----------------------------------------------------------------------------
 * fs_logNvramInitDevice
 *
 * Description: Called to initialize an nvram log device.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

extern int32
fs_logNvramInitDevice(
    fs_Context      *context)
{
    fs_LogFsRoot    *logRoot;
    fs_LogDevHdr    devHdr;
    fs_LogId        eol;
    uchar8          *datap[3];
    int32           lenp[3];
    int32           rc;
    nvram_rw_t      rw;

    FS_TRACE(context, FS_TRACE_LOG, ("fs_logInitNvram"));

    logRoot = (context->fileSystem)->logRoot;

    FS_ASSERT (logRoot->logDevType == FS_LOG_DEVTYPE_NVRAM);

    logRoot->headPtr = 0;           // offset on dev where headPtr is
    logRoot->head = FS_LOG_RESERVED_BYTES;
    logRoot->tail = FS_LOG_RESERVED_BYTES;
    logRoot->currLen = 0;
    logRoot->nextId = FS_LOG_ID_MIN;

    devHdr.head =  FS_LOG_RESERVED_BYTES;
    devHdr.magic = FS_LOG_MAGIC;
    devHdr.version = FS_LOG_VERSION_ONE;
    devHdr.timeStamp = 0x76543210;  // TODO -get timestamp
    devHdr.logSize = context->fileSystem->logSize;
    devHdr.nbrCommits = 0;
    devHdr.cksum = 0;
    ck_sum (&devHdr, &(devHdr.spare), &(devHdr.cksum));
    
    eol = FS_LOG_ID_EOL;

    lenp[0] = sizeof(fs_LogDevHdr);
    datap[0] = (uchar8 *) &devHdr;
    lenp[1] = sizeof(fs_LogDevHdr);
    datap[1] = (uchar8 *) &devHdr;
    lenp[2] = FS_LOG_ID_BYTES;
    datap[2] = (uchar8 *) &eol;
    
    rw.handle = logRoot->logDev.nvHandle;
    rw.numVecs = 3;
    rw.datap = datap;
    rw.lenp = lenp;
    rw.offset = (uint32) 0;
    rw.doWrap = 0;
    rc = nvram_write(&rw);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_LOG_WRITE_ERR, "nvram");
        return NFX_ERR;
    }
    
    return NFX_OK;
} /* fs_logInitNvram */


extern int32
fs_logNvramCommitWrite(
    fs_Context          *context,
    fs_Inode            *inode,
    fs_WriteStableHow   *stableHow)
{
    int32   rc;

    // Write the commit record to the log.
    rc = fs_logCommit(context, FS_LOG_WRITE_COMMIT_STABLE, 0);
    *stableHow = FS_WRITE_STABLE_FILE_SYNC;

    return rc;
}


extern int32
fs_logNvramRecordCksum(
    fs_Context      *context,
    fs_LogRecHdr    *recordHdr,
    int32           recordHdrLen,
    void            *data,
    int32           origLen)
{
    int32               sumLen, i, n, dataLen;
    uint32              *sumdatap, sum;
    FS_LOOP_DBG_DECL(dbgNumLoops);

    // Align the tail of the data to 32-bit boundary.
    dataLen = alignto(origLen, sizeof (uint32));

    sum = 0;

    // checksum header
    for (sumLen = 0, sumdatap = (uint32 *) recordHdr;
         sumLen < recordHdrLen;
         sumLen += sizeof (uint32), sumdatap++)
    {
        FS_LOOP_DBG_CHECK(dbgNumLoops, 5000);
        sum += *sumdatap;
    }

    // checksum aligned portion of payload
    for (i = 0, n = origLen / sizeof (uint32), sumdatap = (uint32 *) data;
         i < n;
         i++, sumLen += sizeof (uint32), sumdatap++)
    {
        FS_LOOP_DBG_CHECK(dbgNumLoops, 5000);
        sum += *sumdatap;
    }

    // checksum unaligned tail of payload
    if (dataLen != origLen) {
        int32 extra = dataLen - origLen;
        uint32 word = 0;

        memcpy(&word, sumdatap, (sizeof (uint32)) - extra);
        sum += word;
        sumLen += sizeof (uint32);
    }

    //recordHdr->cksum = ~sum;
    // removed the cksum field for volume logging / perf reasons
    FS_ASSERT(FALSE);

    return NFX_OK;
}


/*-----------------------------------------------------------------------------
 * fs_logNvramReadInit, fs_logNvramRead, fs_logNvramReadDone, fs_logNvramVerifyCksum
 *
 * Description: NVRAM read routines
 *
 * Created By:  Lan Chin
 *
 * Date Created:  2/22/2002
 */

extern int32
fs_logNvramReadInit(
    fs_Context  *context,
    fs_LogInfo  *info)
{
    int32           rc;
    fs_LogDevHdr    devHdr;

    rc = fs_logReadNvramDevHdr(context->fileSystem->logRoot, &devHdr);
    if (rc == NFX_ERR)
        return rc;

    info->nbrCommits = devHdr.nbrCommits;
    info->logSize = devHdr.logSize;
    info->head = devHdr.head;
    info->tail = devHdr.head;
    info->readChunkSize = NVRAM_RW_MAX;
    info->maxSize = NVRAM_RW_MAX;

    return NFX_OK;
}

extern int32
fs_logNvramRead(
    fs_Context  *context,
    uchar8      **datap,
    uint32      offset,
    uint32      *bytesRead,
    uint32      tail)
{
    if (*bytesRead != NVRAM_RW_MAX)
    {
        return NFX_ERR;
    }
    if (*datap == NULL)
    {
        *datap = fs_ramAlloc(FS_RAM_LOG_READ_BUF, NVRAM_RW_MAX);
        assert(datap);
    }

    // never returns if err
    fs_logReadNvram(context->fileSystem->logRoot, *datap, offset, *bytesRead);
    *bytesRead = NVRAM_RW_MAX;

    return NFX_OK;
}

extern void
fs_logNvramReadDone(
    fs_Context  *context,
    uchar8      *datap)
{
    fs_ramDealloc(FS_RAM_LOG_READ_BUF, datap);
    return;
}

extern int32
fs_logNvramVerifyCksum(
    fs_Context  *context,
    uint32      *data,
    uint32      length)
{
    int32   sumLen;
    uint32  *sumdatap, sum;

    FS_ASSERT((length & 0x03) == 0);   // must align to int32
    sum = 0;
    for (sumLen = 0, sumdatap = data;
         sumLen < length;
         sumLen += sizeof (uint32), sumdatap++)
    {
         sum += *sumdatap;
    }
    if (sum != -1)
    {
        return NFX_ERR;
    }
    
    return NFX_OK;
}

extern int32
fs_logNvramIsEmpty(
    fs_Context  *context)
{
    return (FS_LOG_IS_EMPTY(context->fileSystem->logRoot));
}


#endif // SSC
