/*
 * Name:  fs-primitive.h
 *
 * RCS:  $Id: fs-primitive.h,v 1.125 2002/11/26 22:53:59 bob Exp $
 *
 * Description:  Primitive structures and definitions for the filesystem.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_PRIMITIVE_H)
#define FS_PRIMITIVE_H

#if !defined(MIN)
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) < (b)) ? (b) : (a))
#endif

#if !defined(offsetof)
# define offsetof(T,F) ((unsigned int)((char *)&((T *)0L)->F - (char *)0L))
#endif

#if defined(__GNUC__)
typedef volatile void noreturn;
#else
typedef void noreturn;
#endif

enum {
    BITS_PER_BYTE = 8,
    BITS_PER_WORD = BITS_PER_BYTE * sizeof (uint32), // TODO - generalize
    BITS_PER_BYTE_SHIFT=3
};


/**
 * Forward struct declarations.
 */

typedef struct fs_Arena              fs_Arena;
typedef struct fs_BitmapLocList      fs_BitmapLocList;
typedef struct fs_BootAllocator      fs_BootAllocator;
typedef struct fs_Buf                fs_Buf;
typedef struct fs_BufCksum           fs_BufCksum;
typedef struct fs_BufList            fs_BufList;
typedef struct fs_BufListGlobal      fs_BufListGlobal;
typedef struct fs_Dependencies       fs_Dependencies;
typedef struct fs_DepLog             fs_DepLog;
typedef struct fs_Context            fs_Context;
typedef struct fs_SecurityDescCore   fs_SecurityDescCore;
typedef struct fs_SecurityDescCache  fs_SecurityDescCache;
typedef struct fs_SecurityDescHdr    fs_SecurityDescHdr;
typedef struct fs_SecurityDescNfs    fs_SecurityDescNfs;
typedef struct fs_SecurityDescCifs   fs_SecurityDescCifs;
typedef struct fs_Dinode             fs_Dinode;
typedef struct fs_ExtentVector       fs_ExtentVector;
typedef struct fs_FileSystem         fs_FileSystem;
typedef struct fs_Inode              fs_Inode;
typedef struct fs_Lock               fs_Lock;
typedef struct fs_LockWait           fs_LockWait;
typedef struct fs_RegionSummary      fs_RegionSummary;
typedef struct fs_SuperBlock         fs_SuperBlock;
typedef struct fs_Waiter             fs_Waiter;
typedef struct fs_RdwrReq            fs_RdwrReq;
typedef struct fs_LogBlkId           fs_LogBlkId;
typedef struct fs_LogDevHdr          fs_LogDevHdr;
typedef struct fs_LogFsRoot          fs_LogFsRoot;
typedef struct fs_LogInodeArr        fs_LogInodeArr;
typedef struct fs_LogLockedObjects   fs_LogLockedObjects;
typedef struct fs_LogRecAlloc        fs_LogRecAlloc;
typedef struct fs_LogRecBitmap       fs_LogRecBitmap;
typedef struct fs_LogRecCounter      fs_LogRecCounter;
typedef struct fs_LogRecInode        fs_LogRecInode;
typedef struct fs_LogRecInodeFlush   fs_LogRecInodeFlush;
typedef struct fs_LogRecBufFlush     fs_LogRecBufFlush;
typedef struct fs_LogRecMetaOne      fs_LogRecMetaOne;
typedef struct fs_LogRecMetaTwo      fs_LogRecMetaTwo;
typedef struct fs_LogRecSumm         fs_LogRecSumm;
typedef struct fs_LogObjectSumm      fs_LogObjectSumm;
typedef struct fs_LogRecSummArr      fs_LogRecSummArr;
typedef struct fs_LogRecUser         fs_LogRecUser;
typedef struct fs_LogRecUserRm       fs_LogRecUserRm;
#ifdef NEW_ZF
typedef struct fs_LogRecMetaZero     fs_LogRecMetaZero;
typedef struct fs_LogRecUserZero     fs_LogRecUserZero;
typedef struct fs_zfPage_t           fs_zfPage;
typedef struct fs_zfInode_t          fs_zfInode;
#else // NEW_ZF
typedef struct fs_LogRecZero         fs_LogRecZero;
#endif // NEW_ZF
typedef struct fs_LogRecord          fs_LogRecord;
typedef struct fs_LogReplayHdr       fs_LogReplayHdr;
typedef struct fs_LogReplayRecord    fs_LogReplayRecord;
typedef struct fs_LogSummary         fs_LogSummary;
typedef struct fs_LogUndoSummary     fs_LogUndoSummary;
typedef struct fs_BitmapOperation    fs_BitmapOperation;
typedef struct fs_LogRecMarkDone	 fs_LogRecMarkDone;
typedef struct fs_LogReplayBitmap    fs_LogReplayBitmap;
typedef struct fs_LogReplay          fs_LogReplay;
typedef struct fs_DirSearchCb        fs_DirSearchCb;



/**
 * Type renamings -- TODO: wrap these in structs.
 */

typedef uint8  fs_RefByte;
typedef uint64 fs_Inumber;
typedef uint32 fs_LogId;


/*-----------------------------------------------------------------------------
 * Miscellaneous filesystem primitives, macros.
 */
#define FS_BLOCK_SIZE           8192 // TODO - remove
#define FS_BLOCK_SHIFT          13
#define FS_COMPACT_PAGE_SIZE    1024
#define PAGE_SIZE               FS_BLOCK_SIZE



/**
 * Maximum size of a file system block.
 */
enum { FS_MAX_BLOCK_SIZE = 8 * 1024 };


/**
 * Wrapper for a 48-bit block number.
 */
typedef struct fs_Block {
    uint64 block;
} fs_Block;


/**
 * A canonical invalid block number.
 */
#define FS_INVALID_BLKNUM  0xffffdeadbeefULL
#define FS_INVALID_BLOCK   ((fs_Block) { FS_INVALID_BLKNUM })


/**
 * Magic numbers.
 */
#define FS_VALID_BLOCK_LIMIT       (0xffff00000000ULL)
#define FS_VALID_INODE_LIMIT       (0xffff00000000ULL)


/**
 * Extent bit allocation.
 */
enum {
    EXTENT_BLOCK_BITS  = 48,
    EXTENT_LENGTH_BITS = 16
};

#define EXTENT_BLOCK_MASK       ((1ULL << EXTENT_BLOCK_BITS) - 1)
#define EXTENT_LENGTH_MAX       ((1 << EXTENT_LENGTH_BITS) - 1)


/**
 * MIN, MAX blocks in a filesystem.
 */
#define FS_MIN_BLOCKS             5000ULL
#define FS_MAX_BLOCKS             FS_VALID_BLOCK_LIMIT
#define FS_MAX_BLOCKS_SUPPORTED   12207000000ULL // about 100 TB - any larger
                                                 // and we could have trouble 
                                                 // getting enough memory to
                                                 // eek it.



/**
 * An extent descriptor:  a 16-bit length plus a 48-bit starting block.
 * In the internal representation,
 */
typedef struct fs_Extent {
    uint64 lengthAndStartBlock;
} fs_Extent;


/*-----------------------------------------------------------------------------
 * An extent map pair.
 */
typedef struct fs_Emappair {
    uint64      offset;
    fs_Extent   extent;
} fs_Emappair;

/*
 * A pair of extents, used by fs_splitExtent
 */
typedef struct fs_Extentpair {
    fs_Extent left, right;
} fs_Extentpair;


/**
 * Buffer list indices.
 */
typedef enum {
    FS_BUFLIST_META = 0,       // Buffer cache list for meta bufs
    FS_BUFLIST_USER = 1,       // Buffer cache list for user bufs
    FS_BUFLIST_MAX = 2
} fs_BufListIndex;


/**
 * Structure for tracking which log segments have
 * dependencies on an object.
 */
typedef struct fs_LogSegmentDeps
{
#define FS_LOG_MAX_SEGMENTS     64
    union {
        uchar8          bytes[FS_LOG_MAX_SEGMENTS / BITS_PER_BYTE];
        uint64          dword;
    } u;
    int32           count;  // How many segments are affected.
} fs_LogSegmentDeps;


/**
 * Describe the function of a logged bitmap
 * operation to aid debugging.
 */
struct fs_BitmapOperation
{
    union {
        uint64      inum;
        fs_Extent   extent;
    } u;
#define BITMAP_OP_ALLOC     1
#define BITMAP_OP_DEALLOC   2
#define BITMAP_OP_MERGE     3
    int32           op;
	uint32			pad;         // to align to 64 bits 
};


/**
 * A (context, event, parameters) tuple describing some state machine
 * that's waiting for something.
 */
struct fs_Waiter {
    /** State machine handle for the waiting context. */
    esm_handle_t handle;
    /** Event to wait for. */
    esm_event_t event;
    /** Count of something to wait for. */
    int32 numWait;
    /** Start count of what its waiting for,  a debug aid */
    int32 numWaitStart;
    /** For chaining waiters together. */
    DLL_ENTRY(fs_Waiter)    dllEntry;
};


/**
 * Daemon types for statistics, etc.
 */
enum {
    FS_DAEMON_TYPE_INVALID  = 0,
    FS_DAEMON_TYPE_VOLLOG,      // volume logging daemon
    FS_DAEMON_TYPE_WB_META,     // meta buf writeback daemon
    FS_DAEMON_TYPE_WB_USER,     // user buf writeback daemon
    FS_DAEMON_TYPE_INO_CLNR,    // inode cleaner thread
    FS_DAEMON_TYPE_INO_GROW,    // inode file grow thread
    FS_DAEMON_TYPE_INO_CNT,     // used inode recount thread (once per mount)
    FS_DAEMON_TYPE_REF_CNT,     // use block recount thread (once per mount)
    FS_DAEMON_TYPE_SNAP,        // snapshot thread
    FS_DAEMON_TYPE_FREEZER,     // debug thread for fs_freeze/thaw
    FS_DAEMON_TYPE_ALLOC,       // deferred allocation thread (defunct)
    FS_DAEMON_TYPE_BUDDY,       // buddy bits recalc thread (once per mount)
    FS_DAEMON_TYPE_TRASH,       // trashcan thread
    FS_DAEMON_TYPE_DUMP,        // Dump daemon takes additional args
    FS_DAEMON_TYPE_QUOTA_ON,    // Quota on daemon takes additional args
#ifdef NEW_ZF
    FS_DAEMON_TYPE_UZERO,       // zero fill non-commited user data
#else // NEW_ZF
    FS_DAEMON_TYPE_UNUSED,      // pad
#endif // NEW_ZF
    FS_DAEMON_TYPE_EEK,
    FS_DAEMON_TYPE_UNUSED2,      // pad
    FS_DAEMON_TYPE_MAX,        
};
    
// Must be even in order to avoid padding problem in
// lockstats response.
NFX_ASSERT_STATIC(FS_DAEMON_TYPE_MAX % 2 == 0);

/*-----------------------------------------------------------------------------
 * Error handling.  Each of these has an associated severity level,
 * and depending on the level, appropriate actions are taken in fs_error().
 */
typedef enum fs_ErrorType
{
    FS_DELIVER_EVT_ERR,
    FS_UNK_MSG_ERR,
    FS_SIM_ARGS_ERR,
    FS_SIM_OPEN_ERR,
    FS_SIM_LSEEK_ERR,
    FS_SIM_IO_ERR,
    FS_EEE_ERR,
    FS_INTERNAL_ERR,
    FS_RAM_ALLOC_ERR,
    FS_EVENT_DELIVERY_FAIL,
    FS_ESM_CREATE_ERR,
    FS_ESM_DESTROY_ERR,
    FS_READ_ERR,
    FS_WRITE_ERR,
    FS_LOG_ENTRY_ERR,
    FS_LOG_BAD_DEV_ERR,
    FS_LOG_WRITE_ERR,
    FS_LOG_READ_ERR,
    FS_LOG_ID_NOT_FOUND,
    FS_LOG_DEP_NOT_FOUND,
    FS_LOG_FULL_ERR,
    FS_VOLUME_NOT_MOUNTED,
    FS_ALREADY_MOUNTED_ERR,
    FS_BAD_SUPER_MAGIC,
    FS_BAD_SUPER_VERSION,
    FS_MKFS_INCOMPLETE,
    FS_LOG_BAD_VERSION,
    FS_LOG_BAD_MAGIC,
    FS_VOLUME_EXCEPTION,
    FS_EVENT_TIMEOUT_ERR,
    FS_EVENT_VOL_BROKEN,
    FS_EVENT_VOL_IOERROR,
    FS_EVENT_VOL_CONFLICT,
    FS_EVENT_VOL_TAKEOVER,
    FS_EVENT_VOL_CORRUPT,
    FS_ERROR_TYPE_MAX
} fs_ErrorType;
extern int32 fs_errorCounts[FS_ERROR_TYPE_MAX];


#if defined(__GNUC__)
#define PRINTF_LIKE(n) __attribute__ ((format (printf, (n), (n) + 1)))
#else
#define PRINTF_LIKE(n)
#endif


/**
 * Signal an unrecoverable error.
 */
extern noreturn fs_error(
    fs_Context      *context,
    int32           errorType,
    char            *format,
    ...
)
    PRINTF_LIKE(3);


/**
 * Signal an error or warning which we can recover from.
 */
extern void fs_warn(
    fs_Context      *context,
    int32           errorType,
    char            *format,
    ...
)
    PRINTF_LIKE(3);


/*-----------------------------------------------------------------------------
 * Debugging.
 */

#ifdef FS_DEBUG
#define FS_ASSERT(_what_)                                       \
    if (!(_what_))                                              \
    {                                                           \
        printf("%s:%d: FS_ASSERT FAILED: %s in %s\n",           \
              __FILE__, __LINE__, #_what_, __FUNCTION__);       \
        fs_abort();                                             \
    }
#else
#define FS_ASSERT(_what_)
#endif

#ifdef assert
#undef assert
#endif
#define assert(what) FS_ASSERT(what)

extern noreturn fs_abort();
extern noreturn fs_panic(char *fmt, ...) PRINTF_LIKE(1);

#define UNREACHED()                                 \
    do {                                            \
        int UNREACHED = 0; FS_ASSERT(UNREACHED);    \
    } while (1)

typedef enum fs_traceLvlNum {
    FS_TRACE_REQ,
    FS_TRACE_BUF,
    FS_TRACE_INO,
    FS_TRACE_LCK,
    FS_TRACE_LOG,
    FS_TRACE_IO,
    FS_TRACE_SPACE,
    FS_TRACE_SNAP,
    FS_TRACE_COD,
    FS_TRACE_VERBOSE,
    FS_TRACE_VOLLOG,
    FS_TRACE_READDIR,
    FS_TRACE_DUMP_RESTORE,
    FS_TRACE_LVL_MAX
} fs_TraceLvlNum;

#define FS_TRACE_FP_MAX     2
extern char *fs_traceLvlName[FS_TRACE_LVL_MAX];


#if defined(FS_DEBUG)
#define FS_LOOP_DBG_DECL(_varName_) int32 _varName_ = 0;
#define FS_LOOP_DBG_INIT(_varName_) _varName_ = 0;
#define FS_LOOP_DBG_CHECK(_varName_,_max_)              \
    if (++(_varName_) > (_max_))  FS_ASSERT(FALSE);
#else
#define FS_LOOP_DBG_DECL(_varName_)
#define FS_LOOP_DBG_INIT(_varName_)
#define FS_LOOP_DBG_CHECK(_varName_,_max_)
#endif

#if defined(FS_DEBUG_SHADOW_BUF)
#define FS_DBG_INIT_SHADOW(_buf_)                                       \
    if ((_buf_)->shadow == NULL)                                        \
    {                                                                   \
        (_buf_)->shadow = fs_ramAlloc(FS_RAM_BUF_SHADOW_DATA,           \
                                                       FS_BLOCK_SIZE);  \
    }                                                                   \
    if ((_buf_)->flags & FS_BUF_COMPACT)                                \
        eee_ramCopy((_buf_)->shadow, (_buf_)->u.data, FS_COMPACT_PAGE_SIZE);   \
    else                                                                \
        eee_ramCopy((_buf_)->shadow, (_buf_)->u.data, FS_BLOCK_SIZE);

#define FS_DBG_UPDATE_SHADOW(_buf_,_offset_,_len_)                      \
    eee_ramCopy((_buf_)->shadow + (_offset_),                           \
                (_buf_)->u.data + (_offset_),                           \
                (_len_));
#define FS_DBG_FREE_SHADOW(_buf_)                                       \
    if((_buf_)->shadow != NULL)                                         \
    {                                                                   \
        fs_ramDealloc(FS_RAM_BUF_SHADOW_DATA, (_buf_)->shadow);         \
        (_buf_)->shadow = NULL;                                         \
    }
#else
#define FS_DBG_INIT_SHADOW(_buf_)
#define FS_DBG_UPDATE_SHADOW(_buf_,_offset_,_len_)
#define FS_DBG_FREE_SHADOW(_buf_)
#endif

#if defined(FS_DEBUG_SHADOW_BUF)
#define FS_DBG_MEMCMP(_p1_,_p2_,_len_)          \
    if (memcmp((_p1_),(_p2_),(_len_)) != 0)     \
    {                                           \
        printf("fs_memdiff %p,%p\n",            \
            (_p1_),(_p2_));                     \
        fs_memdiff(_p1_, _p2_, _len_);          \
    }

static inline void fs_memdiff(void *p0, void *q0, int len)
{
    char *p = p0;
    char *q = q0;
    int i;

    for (i = 0; i < len; i++) {
        if (p[i] != q[i]) {
            printf("fs_memdiff at byte offset %d (of %d)\n", i, len);
            printf("bytes at %d are %hhu, %hhu\n", i, p[i], q[i]);
            FS_ASSERT(p[i] == q[i]);
        }
    }
}
#else
#define FS_DBG_MEMCMP(_p1_,_p2_,_len_)  
#endif

#endif /* FS_PRIMITIVE_H */
