// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-ref.h
 *
 * RCS:  $Id: fs-ref.h,v 1.9 2002/07/23 04:55:37 janv Exp $
 *
 * Description:  Reference byte definitions and prototypes.
 *
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-26
 */

#ifndef FS_REF_H
#define FS_REF_H

/**
 * One reference byte is allocated for every block in the file system.
 * The bytes contain a flag indicating whether the block is allocated
 * in the live file system, whether the block should never appear in a
 * snapshot, and a count of the number of snapshots the block is in.
 *
 * The bits are laid out as:
 *
 *       7       6       5       4       3       2       1       0
 *   +-----------------------------------------------+-------+-------+
 *   |           snapshot reference count            |nosnap?|islive?|
 +   +-----------------------------------------------+-------+-------+
 *
 * The refbyte for a free block is all zero bits.
 *
 * It is invalid to have the never-snapshot flag set to true in a
 * byte where the snapshot reference count is non-zero or the live
 * flag is false, and the accessor functions maintain those conditions.
 *
 * TODO - wrap in a struct?
 */
//typedef uint8 fs_RefByte;

/** Reference byte flag for blocks in the live filesystem. */
#define FS_REFBYTE_IS_LIVE              0x1

/** Reference byte flag for blocks that never get snapshotted. */
#define FS_REFBYTE_NEVER_SNAPSHOT       0x2

/** Reference byte mask and shift for snapshot reference count. */
#define FS_REFBYTE_SNAP_COUNT_SHIFT     2
#define FS_REFBYTE_SNAP_COUNT_MASK      0xfc

/** Maximum number of snapshots.  (Exclusive upper limit.) */
#define FS_REFBYTE_SNAP_MAX             64


// TODO - inlining some of these accessors?
// TODO - add context parameters?


/**
 * Return a reference byte with the given is-live flag, never-snapshot flag,
 */
extern fs_RefByte fs_makeRefByte(
    boolean isLive,
    boolean neverSnapshot,
    unsigned snapCount);


/**
 * Return true if the specified reference byte indicates a live block.
 */
extern boolean fs_refbyteIsLive(
    fs_RefByte refbyte);


/**
 * Return true if the specified reference byte indicates a block that
 * should never be in a snapshot.
 */
extern boolean fs_refbyteNeverSnapshot(
    fs_RefByte refbyte);


/**
 * Return the snapshot count from the specified reference byte.
 */
extern unsigned fs_refbyteSnapCount(
    fs_RefByte refbyte);


/**
 * Return a zero refbyte. (used as a transform function to the Modify routines)
 */
extern fs_RefByte fs_refbyteZero(fs_RefByte refbyte);

/**
 * Return the specified reference byte with the live bit set.
 */
extern fs_RefByte fs_refbyteSetLive(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte with the live bit set.  Does not
 * check whether the live bit was previously set.
 */
extern fs_RefByte fs_refbyteForceLive(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte with the live bit clear.
 */
extern fs_RefByte fs_refbyteClearLive(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte modified so that it can be
 * added back into the allocator's structures.
 *
 * In practice, this is a partial function that always returns 0.
 * However, it asserts that the block had been live and no snapshots
 * are using it.
 */
extern fs_RefByte fs_refbyteDeallocate(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte with the never-snapshot bit set.
 */
extern fs_RefByte fs_refbyteDisableSnapshots(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte with the never-snapshot bit clear.
 */
extern fs_RefByte fs_refbyteEnableSnapshots(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte with the snapshot reference count
 * set to the specified value.
 */
extern fs_RefByte fs_refbyteSetSnapCount(
    fs_RefByte refbyte,
    unsigned snapCount);


/**
 * Return the specified reference byte with the snapshot reference count
 * incremented by one.
 */
extern fs_RefByte fs_refbyteIncrSnapCount(
    fs_RefByte refbyte);


/**
 * Return the specified reference byte with the snapshot reference count
 * decremented by one.
 */
extern fs_RefByte fs_refbyteDecrSnapCount(
    fs_RefByte refbyte);


/**
 * Return a reference byte with the live and never-snapshot flags true
 * and the reference count zero.  The previous value for the byte should
 * be zero.
 */
extern fs_RefByte fs_refbyteSetMeta(
    fs_RefByte refbyte);


/**
 * Return the reference byte for the specified block.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
extern fs_RefByte fs_refGetByte(
    fs_Context *context,
    fs_Block block);


/**
 * Modify the reference bytes for all blocks in the range specified by the
 * given block count and start position, according to the specified
 * transformation function.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
extern boolean fs_refByteModify(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    uint32     count,
    fs_Block   block);


/**
 * Modify the reference byte for the given block number, according
 * to the given transformation function.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
extern boolean fs_refByteModifyBlock(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Block block);


/**
 * Modify the reference byte for all teh blocks in the specified extent,
 * according to the given transformation function.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
extern boolean fs_refByteModifyExtent(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Extent  extent);

/**
 * Apply the given transform to the refbytes past the end of the filesystem.
 * (Used by growfs and mkfs).
 */
extern void fs_refByteModifyPastEnd(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Parameters *p);

/**
 * Count all allocated blocks, according to the reference bytes.
 *
 * The result is stored in the superblock as a total count of used
 * blocks.
 */
extern void
fs_refByteCountThread(
    esm_handle_t    handle,
    void            *arg);
    

/**
 * Print all allocated blocks, according to the reference bytes.
 */
void fs_refBytePrintAllocated(
    fs_Context *context);

#endif // FS_REF_H
