// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-skip-heap.h
 *
 * RCS:  $Id: fs-skip-heap.h,v 1.14 2002/01/23 23:48:20 haahr Exp $
 *
 * Description:  Skip list implementation of intra-block sorted sets.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */

#ifndef FS_SKIP_HEAP_H
#define FS_SKIP_HEAP_H


//   T E M P L A T E S


/**
 * A template for a skip heap
 */
typedef struct fs_SkipHeapTemplate fs_SkipHeapTemplate;


/**
 * Create a skip-heap template for a specified block size, payload size
 * and comparison function.  A template is allocated using fs_ramAlloc and
 * should not be freed until all heaps using the given template disappear.
 *
 * A given payload size should always produce exactly the same template;
 * this is essential for running out of an disk block.
 *
 * Objects stored in a skip heap can be treated as either key/value pairs
 * or as just elements in the heap -- the distinction is made entirely
 * within the comparison function.
 *
 * The payload size is the size of each slot.  If variable-sized objects
 * are stored in the heap, pick the largest size, or implement continuations
 * using fs_skipHeapAcquireSlot() and fs_skipHeapReleaseSlot().
 *
 * The reserve size is the amount of space to reserve (always at the end
 * of the buffer) for client data which needs to be recorded in the skip
 * heap buffer but isn't a skip heap entry proper.
 *
 * The comparison function should return a negative integer if the first
 * value is less than the second, a positive integer if the first is
 * greater than the second, and zero if the two are equals.  To support
 * uses where the payloads are not stored in the same representation as
 * search keys, the comparison function is always called with a value from
 * the heap as its first argument and the search key second.
 *
 * To use the fs_skipHeapFindNearest operation, the comparison function
 * should return a result proportional in magnitude to the distance between
 * its arguments.  If it does not, the nearest will be arbitrarily chosen
 * from between the neighbors on either side of the search value.
 *
 * TODO: parameterize by file system block size?
 * TODO: parameterize with skip list level distribution probabilities?
 */
extern fs_SkipHeapTemplate *fs_makeSkipHeapTemplate(
    int blockSize,
    int payloadSize,
    int reserveSize,
    int (*compare)(void *payload1, void *payload2));


/**
 * Return the capacity of heaps with the specified template.
 */
extern int fs_skipHeapTemplateCapacity(
    fs_SkipHeapTemplate *template);


/**
 * Return the block size of heaps with the specified template.
 */
extern int fs_skipHeapTemplateBlockSize(
    fs_SkipHeapTemplate *template);


/**
 * Return the payload size of heaps with the specified template.
 */
extern int fs_skipHeapTemplatePayloadSize(
    fs_SkipHeapTemplate *template);



//   S K I P   H E A P S


/**
 * A skip heap is a skip-list implementation of a sorted collection of
 * fixed-size objects within a single file-system block.
 *
 * These objects are lightweight and can be recreated very cheaply, given
 * the template and underlying buffer.
 *
 * Skip heaps may be passed by value or by reference.  I used a
 * value-passing convention for the entire API, because it reinforces
 * the notion that we don't want to store anything mutable in the
 * skip heap header, as opposed to the buffer.
 *
 * All of the operations on skip heaps are supported only when the
 * underlying buffer is locked by the caller.
 */
typedef struct fs_SkipHeap {
    /** The context for operations on this skip heap. */
    fs_Context *context;
    /** The layout for this skip heap. */
    fs_SkipHeapTemplate *template;
    /** The underlying buffer containing the skip heap's data. */
    fs_Buf *buffer;
} fs_SkipHeap;


/**
 * Create a skip heap in the specified context, using the specified buffer
 * to contain its data and the specified template for its layout.
 */
extern fs_SkipHeap fs_makeSkipHeap(
    fs_Context *context,
    fs_SkipHeapTemplate *template,
    fs_Buf *buffer);


/**
 * Initialize a block containing a skip heap.
 */
extern void fs_skipHeapInit(
    fs_SkipHeap *heap);

/**
 * Initialize a block containing a skip heap without making
 * log entries.
 */
extern void fs_skipHeapInitNoLog(
    fs_SkipHeap *heap);


/**
 * Return a pointer to the client-reserved space in the skip heap.
 */
extern void *fs_skipHeapGetReserve(
    fs_SkipHeap *heap);



//   S I Z I N G


/**
 * Return true if the given skip heap is empty;  that is, it contains no
 * entries yet.
 */
extern boolean fs_skipHeapIsEmpty(
    fs_SkipHeap *heap);


/**
 * Return true if the given skip heap is full;  that is, an attempt to
 * add a new value or acquire a free slot.
 */
extern boolean fs_skipHeapIsFull(
    fs_SkipHeap *heap);


/**
 * Return the number of values currently stored in the given skip heap.
 */
extern int fs_skipHeapCount(
    fs_SkipHeap *heap);


/**
 * Return the total number of slots in the given skip heap.
 */
extern int fs_skipHeapCapacity(
    fs_SkipHeap *heap);


/**
 * Return the number of free slots in the given skip heap.
 */
extern int fs_skipHeapRemaining(
    fs_SkipHeap *heap);



//   B A S I C   O P E R A T I O N S


/**
 * Search for the specified value in the specified skip heap.  If the
 * value is found, returns a pointer to the slot containing it.
 */
extern void *fs_skipHeapGet(
    fs_SkipHeap *heap,
    void *value);


/**
 * Search for a value near the specified value in the specified skip heap.
 * If the skip heap is non-empty, returns a pointer to the slot containing
 * the nearest value; otherwise, returns NULL.
 *
 * This function depends on a comparison function which returns relative
 * distance values.  If the comparison function does not return meaningful
 * distances, one of the neighbors will be chosen arbitrarily.
 */
extern void *fs_skipHeapFindNearest(
    fs_SkipHeap *heap,
    void *value);


/**
 * Add the specified value to the specified skip heap.  Returns true if
 * the value was successfully inserted, or false if there was no space
 * for the new value in the skip heap.  If an entry equal to this value
 * already exists in the heap it is replaced.
 *
 * TODO: is silent replacement the right thing?
 */
extern boolean fs_skipHeapAdd(
    fs_SkipHeap *heap,
    void *value);

 
/**
 * Remove the entry with the specified value from the skip heap.  Returns
 * true if the value was present in the heap, and false otherwise.
 */
extern boolean fs_skipHeapRemove(
    fs_SkipHeap *heap,
    void *value);



//   I T E R A T I O N


/**
 * Return the first element in the specified skip heap, in the ordering
 * defined by the comparison function.  Returns NULL for an empty heap.
 */
extern void *fs_skipHeapFirst(
    fs_SkipHeap *heap);


/**
 * Return the element following the specified element in the skip heap,
 * in the ordering defined by the comparison function.  Returns NULL
 * at the end of the heap.
 */
extern void *fs_skipHeapNext(
    fs_SkipHeap *heap,
    void *prev);



//   C O N T I N U A T I O N   S U P P O R T


/**
 * Create a slot for the specified value in the skip heap, but do nothing
 * to actually insert the value.  This mechanism is intended for skip
 * heaps using continuation structures.  This function returns NULL if
 * the value was already present in the heap or if there is no space for
 * the new entry.
 *
 * The caller must properly fill in the returned slot before making
 * any further calls using the skip heap, except that slots for the
 * continuation structures may be acquired before the value is filled
 * in.  It is the caller's responsibility to touch the buffer with
 * fs_bufWrite when making changes to the returned slot.
 */
extern void *fs_skipHeapInsert(
    fs_SkipHeap *heap,
    void *value);


/**
 * Acquire a slot of the specified skip heap, for use in continuations.
 * The slot will not be used internally by the heap until it is released.
 * Returns NULL if there are no free slots.  It is the caller's
 * responsibility to touch the buffer with fs_bufWrite when making changes
 * to the returned slot.
 */
extern void *fs_skipHeapAcquireSlot(
    fs_SkipHeap *heap);


/**
 * Release a slot that has been acquired for use in continuations for
 * the specified heap.  Releasing a slot that has not been acquired
 * is an error which may corrupt the heap structures, but which may
 * not be detected.
 */
extern void fs_skipHeapReleaseSlot(
    fs_SkipHeap *heap,
    void *slot);


/**
 * Translate a pointer to a slot in the given skip heap into an index
 * within the heap.  It is an error if the pointer is not a valid pointer
 * into the skip heap.
 */
extern int fs_skipHeapSlotToIndex(
    fs_SkipHeap *heap,
    void *slot);


/**
 * Translate an index within the given skip heap into an pointer to the
 * corresponding slot in the heap.  Returns NULL if the index is not
 * valid index for the heap.
 */
extern void *fs_skipHeapIndexToSlot(
    fs_SkipHeap *heap,
    int index);


/**
 * Set the next link for the specified slot to point at the specified next
 * slot.  Both slots must have been allocated for continuations, though
 * the second slot may be null.  It is an unchecked error to use this
 * function if either slot is part of the base skip-list was not allocated
 * as a continuation.
 *
 * Use fs_skipHeapNext to find the next entry.
 */
extern void fs_skipHeapSetNext(
    fs_SkipHeap *heap,
    void *slot,
    void *next);




//   S A N I T Y   C H E C K S


/**
 * Check that a skip heap is well-formed and obeys all the representation
 * invariants.
 */
extern void fs_skipHeapCheck(
    fs_SkipHeap *heap);

/**
 * Compare two unsigned 32-bit values, returning zero if they are equal, a
 * negative value if the first is less than the second, and a positive
 * value if the first is greater.  The magnitude of the returned result is
 * proportional to the distance between the two, clamped at 2^31-1.
 */
extern int fs_uint32compare(void *payload1, void *payload2);

#endif /* FS_SKIP_HEAP_H */
