// -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-snap.h
 *
 * RCS:  $Id: fs-snap.h,v 1.26 2002/12/04 19:25:55 lanc Exp $
 *
 * Description:  Snapshots.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-09-21
 */

#ifndef FS_SNAP_H
#define FS_SNAP_H

//#define FS_SNAP_EEK

/**
 * Snapshot phase.
 */
typedef enum {
    /** No snapshot operation in progress. */
    FS_SNAP_PHASE_READY = 0,

    /** Snapshot creation:  allocating snap bits and sync'ing. */
    FS_SNAP_PHASE_CREATE_ALLOC,
    /** Snapshot creation:  reading reference counts, writing snap bits. */
    FS_SNAP_PHASE_CREATE_SNAP_BITS,
    /** Snapshot creation:  incrementing reference counts. */
    FS_SNAP_PHASE_CREATE_REF_COUNTS,
    /** Snapshot creation:  writing the global metadata. */
    FS_SNAP_PHASE_CREATE_META,

    /** Snapshot removal:  decrementing the reference counts. */
    FS_SNAP_PHASE_REMOVE_REF_COUNTS,
    /** Snapshot removal:  update snap bits. */
    FS_SNAP_PHASE_REMOVE_SNAP_BITS,
    /** Snapshot removal:  update snap bits and free blocks. */
    FS_SNAP_PHASE_REMOVE_SNAP_BITS_AND_FREE,
    /** Snapshot removal:  freeing blocks based on snap bits. */
    FS_SNAP_PHASE_REMOVE_FREE_BLOCKS,
    /** Snapshot removal:  freeing the snap bits. */
    FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS,

    /** Snapshot revert:   remove newer snapshots from the live system */
    FS_SNAP_PHASE_REVERT_REMOVE_SNAPS,
    /** Snapshot revert:   regenerate reference counts from snapbits file */
    FS_SNAP_PHASE_REVERT_REF_COUNTS,
    /** Snapshot revert:   recalculate free buddy tree from reference counts */
    FS_SNAP_PHASE_REVERT_BUDDY,
    /** Snapshot revert:   regenerate inode and directory block files */
    FS_SNAP_PHASE_REVERT_INODE_BLOCKS,
    /** Snapshot revert:   copy snapshot snodes to live system */
    FS_SNAP_PHASE_REVERT_META,
    /** Snapshot revert:   clear the snap bits file from live system */
    FS_SNAP_PHASE_REVERT_CLEAR_SNAP_BITS,

} fs_SnapPhase;


/**
 * A bad count of blocks.
 */
#define FS_BAD_COUNT (0xffffffffffffffffULL)


/**
 * mask for snapshot bits in the superblock
 *  user snapshot id bits are from 1 to 48
 *  internal snapshot id bits from 49 to 63
 * 0 is the live file system
 */
#define FS_SNAP_USER_ID_MASK        0x0001fffffffffffe
#define FS_SNAP_INTERNAL_ID_MASK    0xfffe000000000000

/**
 * Internal snapshot create interval
 */

#define FS_SNAP_INTERNAL_TIMEOUT    10  // every two minutes
#define FS_SNAP_INTERNAL_MIN        60  // in seconds

/**
 * maximum internal snapshots is 15 since 48 is reserved for user
 * and 1 is for the live
 */

#define FS_SNAP_MAX_INTERNAL            (FS_MAX_SNAPSHOTS - FS_MAX_USER_SNAPSHOTS - 1)
#define FS_SNAP_DEFAULT_INTERNAL_SNAPS  0   // default number of internal snaps to keep
                                            // TBD: disable for now

typedef struct fs_SnapInternalInfo
{
    int32   snapnum;
    fs_Time ctime;      // create time
} fs_SnapInternalInfo;


/**
 * Snapshot schedule time
 * For now, always on sunday for weekly and at midnight for weekly and daily
 */

#define FS_SNAP_IS_DAILY(time) (time->hour == FS_SNAP_MIDNIGHT)

#define FS_SNAP_IS_WEEKLY(time) \
    ((time->wday == FS_SNAP_SUNDAY) && FS_SNAP_IS_DAILY(time))

/**
 * Snapshot hour check
 * also include check for minutes in case hourly cron isn't happening
 *
 * hour_bits is a bitmap where bit 24 is midnight
 * hour is a uint32 1-24 where int 24 is midnight
 *
 * define FS_SNAP_TEST_SCHED for unit testing so the minute check is ignored
 */

#define FS_SNAP_TEST_SCHED

#ifdef FS_SNAP_TEST_SCHED
#define FS_SNAP_IS_HOURLY(hour, hour_bits, min) ((hour_bits >> hour) & 0x1)
#else
#define FS_SNAP_IS_HOURLY(hour, hour_bits, min) \
    (((hour_bits >> hour) & 0x1) && FS_SNAP_CHECK_MINUTES(min))
#endif

/**
 * The minute check is for a sanity check in case the hourly cron isn't
 *  happening as advertised
 */

#define FS_SNAP_CHECK_MINUTES(min) \
    ((min == 60) || (min == 59) || (min == 1) || (min == 2))

/**
 * Snapshot times, for recording in superblock.
 */

#define FS_SNAP_DEFAULT_MAX_WEEKLY  0
#define FS_SNAP_DEFAULT_MAX_DAILY   2
#define FS_SNAP_DEFAULT_MAX_HOURLY  8
#define FS_SNAP_DEFAULT_HOURS       ((1 << 8) | ( 1 << 12) | (1 << 16) | (1 << 20))

/**
 * Default snap hours if maxHourly and no hours were specified (per NetApp)
 */
#define FS_SNAP_ALL_HOURS           0x01ffffff

typedef struct {
    boolean enable;                     // FALSE to disable snapshot scheduling
    uint8   maxWeekly;                  // max weekly snapshots
    uint8   maxDaily;                   // max daily snapshots
    uint8   maxHourly;                  // max hourly snapshots
    uint32  hours;                      // bitmap of hours
                                        // bit 1-24, 24 midnight (bit 0 not used)
} fs_SnapTimes;

/**
 * Snapshot state, for recording in superblock.
 */

#define FS_SNAP_VERSION_ONE         0x1
#define FS_SNAP_CHECK_VERSION(context) \
    ((context->fileSystem->super->snapState.version == FS_SNAP_VERSION_ONE) ? TRUE : FALSE)

typedef struct {
    /** Snapshot version. */
    uint8 version;
    /** Snapshot being modified. */
    uint8 id;
    /** Phase-in-progress. */
    fs_SnapPhase phase;
    fs_SnapTimes times;
} fs_SnapState;

/**
 * Default snapshot names
 */
#define FS_SNAP_WEEKLY_NAME "weekly."
#define FS_SNAP_DAILY_NAME "daily."
#define FS_SNAP_HOURLY_NAME "hourly."

#define FS_SNAP_RESERVE_NAME(name) \
    ((strncmp(name, FS_SNAP_WEEKLY_NAME, strlen(FS_SNAP_WEEKLY_NAME)) == 0) || \
     (strncmp(name, FS_SNAP_DAILY_NAME, strlen(FS_SNAP_DAILY_NAME)) == 0) || \
     (strncmp(name, FS_SNAP_HOURLY_NAME, strlen(FS_SNAP_HOURLY_NAME)) == 0))

/**
 * Snapshot stats
 */
typedef struct fs_SnapStats
{
    uint32  snapCreates;        // number of snap creates
    uint32  snapRemoves;        // number of snap removes
    uint32  snapReverts;        // number of snap reverts

    uint32  refCntIncs;         // number reference snap counter increments
    uint32  refCntDecs;         // number reference snap counter decrements
    uint32  liveClears;         // number reference live bit clears
    uint32  snapBitBlocks;      // number of blocks included in snapbits file
    uint32  notFrees;           // number of blocks that are not deallocated
    uint32  liveNotFrees;       // because live bit set
    uint32  snapNotFrees;       // because snap counter > 0
    uint32  blockFrees;         // number of blocks deallocated after snap remove

                                // reference bytes set after snap revert
    uint32  revertSnapBitBlks;  // number of blocks included in snapbits file
    uint32  revertLiveClears;   // number reference live bit clears
    uint32  revertLiveSets;     // number reference live bit sets
    uint32  revertAlreadyLives; // already lives
    uint32  revertNeverSnaps;   // number of never snap blocks found
    uint32  revertBuddyAdds;    // blocks returned to buddy after snap revert
                                // number of inodes set in use after snap revert
    uint32  revertInodes;       //   regular inodes
    uint32  revertDInodes;      //   directory inodes
                                // last known superblock values
    uint64  totalBlocks;
    uint64  usedBlocks;
    uint64  usedInodes;
} fs_SnapStats; 


/**
 * snapshot data structures per fileSystem
 *
 * lock: used to allow only one snapshot operation per filesystem
 *
 *  Only one snapshot operation is allowed because of the snap state saved in
 *  the superblock.  Since we do the FS_SNAP_PHASE_CREATE_ALLOC phase before the
 *  freeze, guarantee only one snapshot operation by acquiring this snap lock first.
 *
 * accessLock: used to lock down a snapshot so it will not be removed
 *             while an application (eg NDMP dump restore) is accessing it
 *
 * For internal snapshot management
 *
 * spin:                spin lock for active
 * active:              set by the snapshot thread and clear when
 *                      internal snapshot done
 *                      (used to prevent posting event when snapshot is
 *                       acquiring freeze lock ie when eek has it)
 * timer:               from eee_timerStart()
 * timeout:             value for eee_timerStart()
 *                      (default is FS_SNAP_INTERNAL_TIMEOUT)
 * maxInternalSnaps:    maxinum number of internal snapshots
 *                      (default is FS_SNAP_DEFAULT_INTERNAL_SNAPS)
 * snapThreadHandle:    from fs_threadFork()
 * lastSnapCreate:      last snapshot create time
 * terminate:           TRUE if in progress snapshots should be terminated (PR #1834)
 */

typedef struct {
    fs_Lock         accessLocks[FS_MAX_SNAPSHOTS];
    fs_Lock         lock;
    int32           timer;
    int32           timeout;
    int32           maxInternalSnaps;
    esm_handle_t    snapThreadHandle;
    eeeRTC_t        lastSnapCreate;
    nfx_spinlock_t  spin;
    boolean         active;
    boolean         terminate;
} fs_SnapInfo;


/**
 * Called at mount time to initialize snapshots
 */
extern void fs_snapInit(
    fs_Context  *context);


/**
 * Called at mkfs time to set default snap state
 */
extern void fs_snapSetDefaultState(
    fs_SnapState    *snapState);


/**
 * Create a snapshot in the specified context's file system, with the
 * specified name.
 */
extern int32 fs_snapCreate(
    fs_Context *context,
    char *name);


/**
 * Remove a snapshot in the specified context's file system, with the
 * specified name.
 */
extern int32 fs_snapRemove(
    fs_Context *context,
    char *name);


/**
 * Remove a snapshot in the specified context's file system, with the
 * specified snapnum.
 */
int32 fs_snapRemoveSnapnum(
    fs_Context *context,
    int32 snapnum);


/**
 * Revert back to a snapshot with snapnum
 *
 * Assumptions (relevant only for agile mirror):
 *
 *  1. the snapshot supernodes are on disk
 *  2. all blocks used by the snapshot are on disk  
 *  3. same superblock version for both live and snapshot 
 *  4. volume big enough to revert
 *     (agile mirror will grow vol and re adjust fs using fs_growfs()? if necc)
 *
 * if NFX_OK, the snapshot's view becomes the live file system
 * if NFX_ERR, the snapnum is not valid
 */
extern int32 fs_snapRevert(
    fs_Context *context,
    int32 snapnum);


/**
 * Create an internal snapshot in the specified context's file system.
 * The oldest internal snapshot will be removed before creating this new snapshot.
 *
 * This internal snapshot is just like a user created one except
 * there is no entry for it in the .snapshot directory.
 *
 * Returns NFX_OK or NFX_ERR
 * If NFX_OK, created snapnum returned (if not NULL)
 */

extern int32 fs_snapCreateInternal(
    fs_Context *context,
    int32 *snapnum,
    boolean quiesce);


/**
 * Remove an internal snapshot in the specified context's file system.
 *
 * This API is here for completeness since normally a snap create internal will
 * remove the oldest internal snapshot.
 * Also useful during unit testing.
 *
 * Returns NFX_OK or NFX_ERR
 */

extern int32 fs_snapRemoveInternal(
    fs_Context *context,
    int32 snapnum);


/**
 * Revert the specified context's file system to an internal snapshot.
 * If snapnum is 0, revert to newest internal snapshot.
 *
 * Returns NFX_OK or NFX_ERR
 */

extern int32 fs_snapRevertInternal(
    fs_Context *context,
    int32 snapnum);


/**
 * Set some parameters for taking internal snapshots for
 * the specified context's file system:
 *  1. nsnaps - keep this many internal snapshots (default is 2)
 *  2. minutes - interval to take the internal snapshot (default is 10 minutes)
 *  3. mkfsTime - TRUE if called during mkfs time
 *
 * Returns NFX_OK or NFX_ERR
 */

extern int32 fs_snapSetInternal(
    fs_Context *context,
    int32 max_snapshots,
    int32 minutes,
    boolean mkfsTime);


/**
 * Resume a snapshot creation, removal or revert operation, based on the snapshot
 * state from the superblock.
 */
extern int32 fs_snapResume(
    fs_Context *context,
    fs_SnapState state);


/**
 * Set a snapshot schedule
 * if write is TRUE, the superblock is written
 */
extern int32 fs_snapSetSched(
    fs_Context *context,
    fs_SnapSched *sched,
    boolean write);


/**
 * Get schedule for snapshot
 */
extern int32 fs_snapGetSched(
    fs_Context *context,
    fs_SnapSched *sched);


/**
 * Schedule snapshot
 */
extern void fs_snapSched(
    fs_Context *context,
    fs_SnapTime *time);


/**
 * Enable or disable snapshots
 */
extern void fs_snapState(
    fs_Context *context,
    boolean enable);


/**
 * Get snapshot info: name, snapnum, ctime
 * usage must be big enough to hold FS_MAX_SNAPSHOTS
 * n modified to return number of snapshots found
 *
 * If internal is TRUE, get info for internal snapshots
 * Else return info for user created snapshots
 */
extern int32 fs_snapGetInfo(
    fs_Context *context,
    fs_SnapUsage *usage,
    int32 *n,
    boolean internal);


/**
 * Called during mount unclean to check if snap revert
 * will recalculate blksUsed in superblock
 */
extern boolean fs_snapRecalculateBlksUsed(
    fs_Context *context);


/**
 * Called during mount unclean to check if snap revert
 * will recalculate inodesUsed in superblock
 */
extern boolean fs_snapRecalculateInodesUsed(
    fs_Context *context);


/**
 * Given a snapshot name find the snapshot number
 */
extern int fs_getSnapNum(
        fs_Context *context,
        char       *name);


/**
 * Get the names of all snapshots for this volume.  Returns the count of
 * names returned.  The names of the snapshots are filled into the names
 * array.  The actual bytes in the names are written to the namebuf;  any
 * names which there isn't space for, according to namebufBytes, are
 * omitted.
 */
extern int32 fs_snapGetNames(
    fs_Context *context,
    char *names[FS_MAX_SNAPSHOTS],
    char *namebuf,
    uint32 namebufBytes);


/**
 * Return the time, in seconds since the start of 1970, the specified
 * snapshot was taken.
 *
 * Returns 0 if the named snapshot does not exists.
 */
extern uint32 fs_snapGetTime(
    fs_Context *context,
    char *name);


/**
 * Find the newest snapshot (user or internal)
 * -1 if no snapshots found
 */
extern int32 fs_snapGetNewest(fs_Context *context);


/**
 * Dump out the blocks in the snapbits file that are set (allocated)
 */
extern void fs_snapDiagBitsFile(
        fs_Context *context,
        char *name,
        int snapnum);


/**
 * Get global snapshot stats
 */
extern void fs_snapDiagGetStats(
        fs_Context *context,
        fs_SnapStats *stats);


/**
 * Lock down a snapshot to prevent snapshot removal
 * Returns NFX_OK or NFX_ERR
 */
extern int32 fs_snapLockSnap(
        fs_Context *context,
        char *name);


/**
 * Unlock a snapshot that was previously locked
 * Returns NFX_OK or NFX_ERR
 */
extern int32 fs_snapUnLockSnap(
        fs_Context *context,
        char *name);


/**
 * Rename snapshot with oldname to newname
 */
extern int32 fs_snapRename(
        fs_Context *context,
        char *oldname,
        char *newname);


/**
 * Pin a snapshot so it can not be removed until it is un pinned
 */
extern int32 fs_snapPin(
        fs_Context *context,
        int snapnum);


/**
 * Un pin a snapshot that was previously pinned
 */
extern int32 fs_snapUnpin(
        fs_Context *context,
        int snapnum);


/**
 * Terminate any in progress snapshot operation
 */
extern boolean fs_snapTerminateActive(
        fs_Context  *context);

#endif // FS_SNAP_H
