/*
 * Name:  fs-trashcan.c
 *
 * RCS:  $Id: fs-trashcan.c,v 1.43 2002/11/26 22:54:00 bob Exp $
 *
 * Description:  Trashcan functions.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  9/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

/**
 * Insert "inode" into the trashcan.
 */
void
fs_trashcanAdd(fs_Context *context, fs_Inode *trashcanInode, fs_Inode *inode)
{
	char name[32];	// XXX: magic number -- represent a uint64 in ASCII

	sprintf(name, "%016llX", inode->inum);
	fs_dir_addent(context, trashcanInode, name, strlen(name), inode->inum);
	fs_trashcanWakeup(context);
}

/**
 * Remove the extents from a file one page of extents per transaction
 * and destroy the dirent and inode.
 */
// define FS_DEBUG_PRINT_TRASHCAN_PERF 1

#ifdef FS_DEBUG_PRINT_TRASHCAN_PERF
static int sleepLoop = 0, sleepLoopMax = 0, nblocksMax = 0, sleepCnt = 0;
static eeeRTC_t sleepTimeMax;
#endif

void
fs_trashDirent(fs_Context *context, fs_DirEntry *dirent)
{
    fs_Inode *inode, *trashcanInode;
    uint64 nblocks, blksRemoved = 0, orig_inum = 0;
	uint32 orig_genNum;
    int loopCnt = 0, umountCnt = 0;
    int doSleep = FALSE;
    eeeRTC_t  beginTime, delta;
    fs_LogFsRoot *logRoot = context->fileSystem->logRoot;
#ifdef FS_DEBUG_PRINT_TRASHCAN_PERF
    eeeRTC_t  startTime;
    static boolean first = TRUE;

    if (first) {
        first = FALSE;
        sleepTimeMax = *(eee_rtcZero());
    }
#endif

#define FS_CHECK_CPU_USAGE  10

    FS_TRACE(context, FS_TRACE_REQ,
            ("smHandle %llx ++++ trash emptying ino %lld ++++", 
            (uint64)((uintptr_t)context->smHandle), dirent->inode));

    // PR #1789: acquire global access lock
    fs_lockShared(context, &context->fileSystem->accessLock, NULL);
    fs_lockShared(context, &context->fileSystem->modifyLock, NULL);

    eee_rtc(&beginTime);

#ifdef FS_DEBUG_PRINT_TRASHCAN_PERF
    startTime = beginTime;
    sleepLoop = 0;
    sleepCnt = 0;
#endif

    if (logRoot->currLen > 
        (FS_LOG_LOOP_BYTES(context->fileSystem->logSize) * 
                                   fs_incoreTune->logCommitHiWtrMk /100)) {
        
        // Wakeup the inode cleaner threads.
        fs_inodeSignalCleaner(context, FS_INODE_CLEANER_LOG);

        // Wakeup the meta buffer writeback daemon.
        fs_bufSignalWriteback(context, FS_BUFLIST_META, FS_WB_WAKEUP_LOG);

        do {

            fs_sleep (context, 0);

        } while (logRoot->currLen > 
             (FS_LOG_LOOP_BYTES(context->fileSystem->logSize) * 
                                    fs_incoreTune->logCommitHiWtrMk /100));

    }

    inode = I_GET_EXCL(0, dirent->inode);
    
    while ((nblocks = fs_emap_shrink(context, inode)) != 0)
	{
        blksRemoved += nblocks;
        loopCnt += 1;
        umountCnt += 1;

		/*
 		 * Be sure commit will unlock for us, since shrink
		 * might not have modified the inode.
		 */
		fs_inodeLockTransaction(context, inode, context->logSummary);

        vl_addUnstableId(context, &inode->inum);
        fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, inode->inum);

        if (logRoot->currLen > 
            (FS_LOG_LOOP_BYTES(context->fileSystem->logSize) * 
                                   fs_incoreTune->logCommitHiWtrMk /100)) {
        
            // Wakeup the inode cleaner threads.
            fs_inodeSignalCleaner(context, FS_INODE_CLEANER_LOG);

            // Wakeup the meta buffer writeback daemon.
            fs_bufSignalWriteback(context, FS_BUFLIST_META, FS_WB_WAKEUP_LOG);

            do {

                fs_sleep (context, 0);

            } while (logRoot->currLen > 
                            (FS_LOG_LOOP_BYTES(context->fileSystem->logSize) * 
                                    fs_incoreTune->logCommitHiWtrMk /100));
        } 

        if (logRoot->volBufRoot)    {
            vl_VolBufRoot   *volBufRoot = logRoot->volBufRoot;
            uint32          currNbrBuf;
            
            NFX_SPIN_LOCK(&volBufRoot->idSpin);
            currNbrBuf = volBufRoot->nbr.currNbrBuf;
            NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

            if (currNbrBuf > VL_BUFFER_LOG_MAX - 
                         fs_incoreTune->resource[FS_RESOURCE_VOL_LOG].absMin)
            {
                fs_sleep (context, 10);
            }
                        
        }        

		if (context->logSummary == NULL) {
            fs_logBeginTransaction(context);
		}

#ifdef FS_DEBUG_PRINT_TRASHCAN_PERF
        sleepLoop++;
#endif
        if ((loopCnt % FS_CHECK_CPU_USAGE) == 0)
        {
            // sleep to prevent resource (log, cache, ...) overrun
            doSleep = TRUE;
        }
        else
		{
			eee_rtcDuration(&beginTime, &delta);
			if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
			{
				// sleep to prevent CPU overrun
				doSleep = TRUE;
			}
		}
        
        if (doSleep)
        {
            doSleep = FALSE;
            
#ifdef FS_DEBUG_PRINT_TRASHCAN_PERF
            sleepCnt++;
            if (sleepLoopMax < sleepLoop)
                sleepLoopMax = sleepLoop;
            sleepLoop=0;
			if (eee_rtcCompare(&delta, &sleepTimeMax) > 0)
                sleepTimeMax = delta;
            if (nblocksMax < nblocks)
                nblocksMax = nblocks;
#endif
            
            // release global locks before sleep
            fs_unlockShared(context, &context->fileSystem->modifyLock);
            fs_unlockShared(context, &context->fileSystem->accessLock);

            fs_sleep(context, 0);
            
            // reacquire global locks
            fs_lockShared(context, &context->fileSystem->accessLock, NULL);
            fs_lockShared(context, &context->fileSystem->modifyLock, NULL);

            eee_rtc(&beginTime);
        }

		inode = I_GET_EXCL(0, dirent->inode);
    }

#ifdef FS_DEBUG_PRINT_TRASHCAN_PERF
    printf("Trashcan inode %lld  Time: %lld msecs\n", 
           dirent->inode, eee_rtcToMS(&startTime));
    printf("  sleepCnt %d usecmax %lld maxBlocks %d total blocks %lld\n", 
           sleepCnt, eee_rtcToUsec(&sleepTimeMax), nblocksMax, blksRemoved);
#endif

#if defined FS_DEBUG_TRASH_EMAP_LOOKUP
{
    uint64 page;
    for (page = 0; page < inode->dinode.size / FS_BLOCK_SIZE; page++)
    {
        fs_Block    block;

        block = fs_emap_lookup_block(context, inode, page);
        if (fs_blockIsValid(block))
        {
            printf("found extent after trash truncation, page %lld\n", page);
            FS_ASSERT(FALSE);
        }

		eee_rtcDuration(&beginTime, &delta);
		if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            // release global locks before sleep
            fs_unlockShared(context, &context->fileSystem->modifyLock);
            fs_unlockShared(context, &context->fileSystem->accessLock);

            fs_sleep(context, 0);
            
            // reacquire global locks
            fs_lockShared(context, &context->fileSystem->accessLock, NULL);
            fs_lockShared(context, &context->fileSystem->modifyLock, NULL);

            eee_rtc(&beginTime);
        }
    }
}
#endif

	/*
 	 * Be sure commit will unlock for us, since shrink
	 * might not have modified the inode.
	 */
	fs_inodeLockTransaction(context, inode, context->logSummary);

	fs_logCommitEntry(context);

	if (context->logSummary == NULL) {
    	fs_logBeginTransaction(context);
	}

	trashcanInode = I_GET_EXCL(0, FS_TRASHCAN_INODE);
	inode = I_GET_EXCL(0, dirent->inode);
    fs_dir_removeent(context, trashcanInode, dirent->name,
                     strlen(dirent->name));

    // Dirty trick:  we're overloading the numblocks field in the
    // trashed inode to point to the original inode.
    orig_inum = inode->dinode.numblocks;
    orig_genNum = inode->dinode.genNum;
    fs_deallocateInode(context, inode);

    /*
     * Be sure commit will unlock for us, since removeent 
     * might not have modified the inode.
     */
    fs_inodeLockTransaction(context, trashcanInode, context->logSummary);
    fs_logCommitEntry(context);

    if (orig_inum && blksRemoved)
    {
        fs_Inode *orig_inode;

        /*
         * Decrement the block count from the original trashed
         * file.  We do this as a separate transaction to avoid
         * inode deadlock.  This means we don't have a consistency
         * guarantee over the block count of the truncated file in
         * case we crash before committing this.
         */

        if (context->logSummary == NULL) {
            fs_logBeginTransaction(context);
        }

        orig_inode = I_GET_EXCL(0, orig_inum);

        /*
         * The original file may have disappeared since the trash process
         * began -- check to see if the inode is nonnull.
         */
        if (orig_inode != NULL)
        {
            if (orig_inode->dinode.genNum == orig_genNum)
            {
                orig_inode->dinode.numblocks -= blksRemoved;
                fs_inodeWrite(context, orig_inode, INO_WRITE_LOG);
                fs_logCommitEntry(context);
            }
            else
            {
                fs_inodeReleaseExclusive(context, orig_inode);
            }
        }
    }

    fs_unlockShared(context, &context->fileSystem->modifyLock);
    // release global access lock before return
    fs_unlockShared(context, &context->fileSystem->accessLock);

    FS_TRACE(context, FS_TRACE_REQ,
            ("smHandle %llx ---- trash emptied ino %lld ----", 
            (uint64)((uintptr_t)context->smHandle), dirent->inode));
}


/**
 * Thread to empty the trash directory.
 */
void
fs_trashcanThread(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;
    fs_FileSystem   *fs = context->fileSystem;
    fs_DirEntry     *dirent, entries[6];	/* XXX: magic number: should fit ".", ".." and one name */
    int32 num_entries, eof;
    fs_Inode *trashcanInode;

    while (1) {

        // PR #1789: acquire global access lock
        fs_lockShared(context, &fs->accessLock, NULL);

		if (context->logSummary == NULL) {
            fs_logBeginTransaction(context);
		}

		trashcanInode = I_GET_SHARED(0, FS_TRASHCAN_INODE);

		if (fs_dir_is_empty(context, trashcanInode)) {
	    	fs_inodeRelease(context, trashcanInode);

            // release global access lock before yielding
            fs_unlockShared(context, &fs->accessLock);

	    	NFX_SPIN_LOCK(&fs->spin);
	    	fs->trashcanSignal = 0;
	    	NFX_SPIN_UNLOCK(&fs->spin);

            fs_lockTerminateContext(context);

            //printf("trashcan demon going to sleep\n");
            fs_yield(context, NULL, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_ADMIN);
            //printf("trashcan demon waking up\n");
	    	continue;
		}

		fs_dir_readdir(context, trashcanInode, entries, sizeof entries, 0,
		       			FALSE, &num_entries, &eof);
		fs_inodeRelease(context, trashcanInode);

        // release global access lock
        fs_unlockShared(context, &fs->accessLock);

		FS_ASSERT(num_entries >= 3);	/* ".", "..", and one filename */

		dirent = entries;
		FS_READDIR_ADV_ENTRY_PTR(dirent);	/* skip "." */
		FS_READDIR_ADV_ENTRY_PTR(dirent);	/* skip ".." */

		fs_trashDirent(context, dirent);
   	}

	/* NOTREACHED */
}


/**
 * Wake up the trashcan thread if its sleeping.
 */
void
fs_trashcanWakeup(fs_Context *context)
{
    fs_FileSystem *fs = context->fileSystem;
    esm_event_t event;

    event.event = FS_THREAD_EVENT_START;
    event.p1.u.p = NULL;
    event.p2.u.w = 33; // debug

    fs_daemonWakeup(context, &fs->spin, fs->trashcanHandle,
					&fs->trashcanSignal, event);
}
#endif // SSC
