/*
 * Name:  fs-vol-exer.c
 *
 * RCS:  $Id: fs-vol-exer.c,v 1.13 2002/12/10 17:27:07 bob Exp $
 *
 * Description:  Volume exerciser
 * 
 */

#if !defined(SSC)

#include "fs-includes.h"
#include "fs-volexer.h"

#define EXER_TRACE  if(fs_volexerTrace) printf

#define EVMBLKS_PER_FS_BLOCK    (FS_BLOCK_SIZE / EVM_LBLK_SIZE)

// Kick off up to this many IO's on a single
// event.
#define FS_VOL_EXER_MAX_IO_KICK 32

enum {
    FS_VOL_EXER_INITING     = 0,
    FS_VOL_EXER_RUNNING     = 1,
    FS_VOL_EXER_STOPPING    = 2,
    FS_VOL_EXER_STOPPED     = 3,
};


#define FS_VOL_EXER_STATE_BUSY       0x80
#define FS_VOL_EXER_STATE_GEN_MASK   0x7f


// List of active volume exerciser instances.
fs_VolExerDll fs_volExerDllHdr;
int32 fs_volexerTrace = 0;

uint64 fs_volexer_stats_usec[FS_TIME_STATS_BUCKETS_MAX] = { 10000, 50000, 100000, 200000, 300000, 400000, 500000, 1000000 };

/**
 * Return the volExer struct for the specified
 * volume if it exists.
 */
fs_VolExer *
fs_volExerForVolumeLocked(
    evm_volId_t     volId)
{
    fs_VolExer  *exer;

    DLL_ITERATE(exer, &fs_volExerDllHdr, dllEntry)
    {
        if (exer->volId.whole_thing == volId.whole_thing)
        {
            return exer;
        }
    }
    return NULL;
}


/**
 * Start, stop, or stat the volume exerciser for
 * the specified volume.
 */
int32
fs_volExerThread(
    fs_Context  *context)
{
    fs_MsgVolExerReq    *req = (fs_MsgVolExerReq *) context->reqHdr;
    fs_MsgVolExerResp   *resp = (fs_MsgVolExerResp *) context->respHdr;
    fs_VolExer          *exer;
    esm_event_t         event;
    uint64              i;
    int32               rc;

    FS_TRACE(context, FS_TRACE_REQ, ("volexer"));

    context->respSize = sizeof (fs_MsgVolExerResp);
	resp->count_per_usec = count_per_usec;
    resp->volId = req->volId;
    resp->what = req->what;

    switch(req->what)
    {
    case FS_VOL_EXER_START:
        NFX_SPIN_LOCK(&fs_fileSystemTableLock);
        if (fs_fileSystemForVolumeLocked(context->volId) != NULL)
        {
            NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
            printf("volume 0x%llx is mounted, can't start exerciser\n",
                   context->volId.whole_thing);
            return NFX_ERR;
        }
        if (fs_volExerForVolumeLocked(context->volId) != NULL)
        {
            NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
            printf("volume 0x%llx is already running exerciser\n",
                   context->volId.whole_thing);
            return NFX_ERR;
        }
        exer = eee_ramAlloc(sizeof *exer);
        if (exer == NULL)
        {
            printf("failed to alloc memory, can't start exerciser\n");
            return NFX_ERR;
        }
        bzero(exer, sizeof *exer);
        exer->volId = context->volId;
        DLL_INSERT_HEAD(&fs_volExerDllHdr, exer, dllEntry);
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

        // Init all random generators to the same
        // seed.
        if ((exer->seed = req->seed) == 0)
        {
            exer->seed = fs_getTimeOfDaySecs();
        }
        for (i = 0; i < FS_VOL_EXER_ACTION_MAX; i++)
        {
            initstate_r(exer->seed, exer->rw[i].randomState, RANDOM_STATE_SIZE,
                        &exer->rw[i].randData);
            exer->rw[i].nextOffset = -1;
            exer->rw[i].nextNumReqs = -1;
        }
        initstate_r(exer->seed, exer->actionRandomState, RANDOM_STATE_SIZE,
                    &exer->actionRandData);

        // Set the numBlks values appropriate to simulate a log
        // device at the end of the volume (where sequential access
        // will take place in a loop).
        exer->numBlksTotal = req->numBlks;
        exer->numBlksSequential = req->numBlksSequential;
        exer->numBlksRandom = exer->numBlksTotal - exer->numBlksSequential;
        exer->currSequentialROffset = exer->numBlksRandom; 
        exer->currSequentialWOffset = exer->numBlksRandom; 
        exer->sequentialIoSize = req->sequentialIoSize; 
        exer->load = req->load;
        exer->loadType = req->loadType;
        exer->ioType = req->ioType;
        exer->attrib = req->attrib;
        exer->isUser = req->isUser;
        if (exer->isUser)
        {
            exer->writeCod = fs_cod_get_new_cod();
            rc = fs_cod_create_null(exer->writeCod,
                                    FS_BLOCK_SIZE, 0);
            if (rc != NFX_OK)
            {
                printf("failed to create exer write cod\n");
                return rc;
            }
        }

        // Find the nearest power of 2 greater than numBlksRandom
        // to use as a mask.
        for (i = 1; i < UINT64_MAX; i *= 2)
        {
            if (i > exer->numBlksRandom)
            {
                exer->blksRandMask = (i - 1);
                break;
            }
        }

        exer->reqState = eee_ramAlloc(exer->numBlksTotal);
        if (exer->reqState == NULL)
        {
            NFX_SPIN_LOCK(&fs_fileSystemTableLock);
            DLL_REMOVE(&fs_volExerDllHdr, exer, dllEntry);
            NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
            eee_ramDealloc(exer);
            printf("failed to alloc memory, can't start exerciser, try lower volume size\n");
            return NFX_ERR;
        }
        bzero(exer->reqState, exer->numBlksTotal);

        rc = fs_volexer_sm_create(&exer->handle, exer);
        if (rc != NFX_OK)
        {
            fs_error(NULL, FS_ESM_CREATE_ERR, "volexer esm");
        }

        event.event = FS_VOLEXER_EVENT_GO;
        rc = esm_deliverEvent(exer->handle, &event);
        if (rc != NFX_OK)
        {
            fs_warn(NULL, FS_EVENT_DELIVERY_FAIL, "volexer go");
        }
        break;

    case FS_VOL_EXER_STOP:
        NFX_SPIN_LOCK(&fs_fileSystemTableLock);
        exer = fs_volExerForVolumeLocked(context->volId);
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

        if (exer == NULL)
        {
            printf("exerciser not running on vol 0x%llx\n",
                   context->volId.whole_thing);
            return NFX_ERR;
        }
        event.event = FS_VOLEXER_EVENT_STOP;
        rc = esm_deliverEvent(exer->handle, &event);
        if (rc != NFX_OK)
        {
            fs_warn(NULL, FS_EVENT_DELIVERY_FAIL, "volexer stop");
        }
        // FALLTHROUGH

    case FS_VOL_EXER_STAT:
        NFX_SPIN_LOCK(&fs_fileSystemTableLock);
        exer = fs_volExerForVolumeLocked(context->volId);
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);

        if (exer == NULL)
        {
            printf("exerciser not running on vol 0x%llx\n",
                   context->volId.whole_thing);
            return NFX_ERR;
        }

        resp->volId = exer->volId;
        resp->load = exer->load;
        resp->seed = exer->seed;
        resp->numBlks = exer->numBlksTotal;
        resp->numBlksSequential = exer->numBlksSequential;
        resp->sequentialIoSize = exer->sequentialIoSize;
        resp->ioType = exer->ioType;
        resp->loadType = exer->loadType;
        resp->readPending = exer->rw[EVM_READ_REQ].totalPendingCount;
        resp->readDone = exer->rw[EVM_READ_REQ].totalDoneCount;
        resp->writePending = exer->rw[EVM_WRITE_REQ].totalPendingCount;
        resp->writeDone = exer->rw[EVM_WRITE_REQ].totalDoneCount;
        resp->genNumErrors = exer->genNumErrors;
        resp->dataErrors = exer->dataErrors;
        resp->isUser = exer->isUser;

        for (i = 0; i < FS_TIME_STATS_BUCKETS_MAX; i++) 
        {
            resp->respTimeRead[i] = exer->respTimeRead[i];
        }
        resp->respTimeMaxRead = exer->respTimeMaxRead;
        resp->respTimeMinRead = exer->respTimeMinRead;
        resp->ioQDepthMaxRead = exer->ioQDepthMaxRead;
        resp->ioQDepthMinRead = exer->ioQDepthMinRead;
        resp->esmTimeMaxRead = exer->esmTimeMaxRead;
        resp->esmTimeMinRead = exer->esmTimeMinRead;
        resp->totalRespTimeRead = exer->totalRespTimeRead;

        for (i = 0; i < FS_TIME_STATS_BUCKETS_MAX; i++) 
        {
            resp->respTimeWrite[i] = exer->respTimeWrite[i];
        }
        resp->respTimeMaxWrite = exer->respTimeMaxWrite;
        resp->respTimeMinWrite = exer->respTimeMinWrite;
        resp->ioQDepthMaxWrite = exer->ioQDepthMaxWrite;
        resp->ioQDepthMinWrite = exer->ioQDepthMinWrite;
        resp->esmTimeMaxWrite = exer->esmTimeMaxWrite;
        resp->esmTimeMinWrite = exer->esmTimeMinWrite;
        resp->totalRespTimeWrite = exer->totalRespTimeWrite;

#if defined(AGILE_HARDWARE)
		eee_rtcDuration(&exer->startTime,
                        &resp->runTime);
#endif
        break;

    default:
        printf("unknown exerciser req %d\n", req->what);
        break;
    }

    return NFX_OK;
}


/**
 * Callback function for EVM IO completion.
 */
void
fs_volexerIoDone(
    evm_IOStat_t    *stat)
{
    fs_RdwrReq              *rdwr;
    esm_event_t             event;
    int32                   rc;

    rdwr = (fs_RdwrReq *) (address_t) stat->clientPrivate;

    FS_ASSERT(rdwr->evmReq.clientPrivate == stat->clientPrivate);

    rdwr->evmStat = *stat;
#if defined(AGILE_HARDWARE)
    eee_rtc (&rdwr->timestamp_return); 
#endif

    event.event = rdwr->doneEvent;
    event.p1.u.p = rdwr;
    event.p2.u.p = NULL;
    rc = thread_event(rdwr->smHandle, &event);
    if (rc != NFX_OK)
    {
        fs_error(NULL, FS_EVENT_DELIVERY_FAIL, "volexer IO done handle 0x%llx",
				 (uint64)((uintptr_t)rdwr->smHandle));
    }
}


/**
 * Free the data pointer and the rdwr req structure.
 */
static void
fs_volexerFreeRdwr(
	fs_VolExer	*exer,
    fs_RdwrReq  *rdwr)
{
    if (rdwr->evmReq.reqType == EVM_READ_REQ)
    {
        if (rdwr->evmReq.dataType == EVM_USER_DATA_REQ)
        {
            fs_cod_free_cod(rdwr->evmStat.rDataHandle.codHandle);
        }
        else
        {
            // Free the data pointer.
            fs_ramDealloc(FS_RAM_BUF_DATA, rdwr->evmStat.rDataHandle.dataPtr);
        }
    }
    else
    {
        if (rdwr->evmReq.dataType == EVM_META_DATA_REQ)
        {
            // Free the data pointer.
            fs_ramDealloc(FS_RAM_BUF_DATA, rdwr->evmReq.wDataHandle.dataPtr);
        }
    }

    eee_ramDealloc(rdwr);
}


/**
 * Initialize the specified data buffer with the
 * pattern for the specified offset and genNum.
 */
static void
fs_volexerInitWriteData(
    fs_VolExer  *exer,
    void        *ptr, 
    uint64      offset)
{
    uint32  i;
    uint64  *llPtr = (uint64 *) ptr;
    uint64  byteOffset;

    *llPtr = exer->reqState[offset] & FS_VOL_EXER_STATE_GEN_MASK;

    for (i = 1, byteOffset = (offset * FS_BLOCK_SIZE) + sizeof(uint64);
         i < FS_BLOCK_SIZE / sizeof(uint64);
         i += 1, byteOffset += sizeof(uint64))
    {
        *(llPtr + i) = byteOffset;
    }
}


/**
 * Stop the test.
 */
static void
fs_volexerStop(
    fs_VolExer  *exer)
{
    if (exer->ioPending == 0 && exer->state == FS_VOL_EXER_STOPPING)
    {
        printf("volume exerciser 0 IO's pending, terminated\n");
        if (fs_volexer_sm_destroy(exer->handle) == NFX_OK) {

            NFX_SPIN_LOCK(&fs_fileSystemTableLock);
            DLL_REMOVE(&fs_volExerDllHdr, exer, dllEntry);
            NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
            exer->state = FS_VOL_EXER_STOPPED;
            if (exer->writeCod)
            {
                fs_cod_free_cod(exer->writeCod);
            }
            eee_ramDealloc(exer->reqState);
            eee_ramDealloc(exer);
        }
    }
    else
    {
        printf("volume exerciser terminate pending %d IO's\n",
            exer->ioPending);
        exer->state = FS_VOL_EXER_STOPPING;
    }
}


/**
 * Check the specified data buffer for the
 * correct pattern for the specified offset;
 */
static void
fs_volexerCheckReadData(
    fs_VolExer  *exer,
    void        *ptr,
    uint64      offset)
{
    uint32  i;
    uint64  genNum, *llPtr = (uint64 *) ptr;
    uint64  byteOffset;

    if (offset >= exer->numBlksTotal)
    {
        printf(
   "exerciser vol 0x%llx internal error, block %lld >= numBlksTotal %lld\n",
               exer->volId.whole_thing, offset, exer->numBlksTotal);
        fs_volexerStop(exer);
        return;
    }

    if (exer->reqState[offset] == 0)
    {
        printf(
            "exerciser vol 0x%llx internal error, block %lld never written\n",
               exer->volId.whole_thing, offset);
        fs_volexerStop(exer);
        return;
    }

    genNum = *llPtr;
    if ((exer->reqState[offset] & FS_VOL_EXER_STATE_GEN_MASK) != genNum)
    {
        printf(
     "exerciser vol 0x%llx compare error, block %lld wrong genNum %lld/%d\n",
               exer->volId.whole_thing, offset, genNum,
               (exer->reqState[offset] & FS_VOL_EXER_STATE_GEN_MASK));
        exer->genNumErrors += 1;
        // continue to check data consistency
    }

    for (i = 1, byteOffset = (offset * FS_BLOCK_SIZE) + sizeof(uint64);
         i < FS_BLOCK_SIZE / sizeof(uint64);
         i += 1, byteOffset += sizeof(uint64))
    {
        if (*(llPtr + i) != byteOffset)
        {
            printf(
  "exerciser vol 0x%llx compare error, block %lld wrong data 0x%llx/0x%llx\n",
                    exer->volId.whole_thing, offset, *(llPtr + i),
                    byteOffset);
            exer->dataErrors += 1;
            // only count one per block
            return;
        }
    }
}


/**
 * Update the incore state and generation number
 * for the specified action, offset, numReqs.
 */
static void
fs_volexerUpdateStateGenNum(
    fs_VolExer  *exer,
    uint32      action,
    uint64      offset,
    int32       numReqs)
{
    int32   i;

    // Update state and genNum as required.
    if (action != FS_VOL_EXER_ACTION_NONE)
    {
        for (i = 0; i < numReqs; i++)
        {
            uchar8  gen = exer->reqState[(offset) + i] &
                                  FS_VOL_EXER_STATE_GEN_MASK;

            if (action == EVM_WRITE_REQ)
            {
                if (gen == FS_VOL_EXER_STATE_GEN_MASK)
                    gen = 1;
                else
                    gen += 1;
                exer->reqState[(offset) + i] = gen;
            }

            exer->reqState[(offset) + i] |= FS_VOL_EXER_STATE_BUSY;

            EXER_TRACE("exer 0x%llx action %d block %lld num %d/%d gen %d\n",
               exer->volId.whole_thing, action, (offset) + i,
               i, numReqs, gen);
        }
    }
}


/**
 * Check for a busy condition on the specified offset,numReqs,
 * return ACTION_NONE if we need to wait before proceeding with
 * the specified request.
 */
static void
fs_volexerCheckBusy(
    fs_VolExer  *exer,
    uint32      *action,
    uint64      *offset,
    uint32      *numReqs)
{
    int32   i;

    // Check if any of the offsets are currently
    // busy.
    for (i = 0; i < *numReqs; i++)
    {
        if (exer->reqState[(*offset) + i] & FS_VOL_EXER_STATE_BUSY)
        {
            // Busy condition, stall for now, we'll retry
            // this offset,reqLen next time here for this
            // action.
            exer->rw[*action].nextOffset = *offset;
            exer->rw[*action].nextNumReqs = *numReqs;

            EXER_TRACE("exer 0x%llx action %d->none block %lld busy\n",
                       exer->volId.whole_thing, *action, (*offset) + i);

            *action = FS_VOL_EXER_ACTION_NONE;
        }
    }
}


/**
 * Return the next sequential offset and numReqs for the
 * specified action.   Updates exer->reqState[], possibly
 * returning ACTION_NONE if the next offset(s) is(are)
 * busy.
 */
static void
fs_volexerGetNextSeqOffsetLen(
    fs_VolExer  *exer,
    uint32      *action,
    uint64      *offset,
    int32       *numReqs)
{
    if ((exer->currSequentialROffset + exer->sequentialIoSize) >= exer->numBlksTotal) 
            exer->currSequentialROffset = exer->numBlksRandom;

    if ((exer->currSequentialWOffset + exer->sequentialIoSize) >= exer->numBlksTotal) 
            exer->currSequentialWOffset = exer->numBlksRandom;

    *offset = (*action == EVM_READ_REQ) ? exer->currSequentialROffset :
                                          exer->currSequentialWOffset;
    *numReqs = exer->sequentialIoSize;

    // We might need to wait before starting this request.
    fs_volexerCheckBusy(exer, action, offset, numReqs);

    // Update state and genNum as required.
    if (*action != FS_VOL_EXER_ACTION_NONE)
    {
        fs_volexerUpdateStateGenNum(exer, *action, *offset, *numReqs);
        if (*action == EVM_READ_REQ) {

            exer->currSequentialROffset += exer->sequentialIoSize;
            exer->seqReadDone += exer->sequentialIoSize;

        } else {

            exer->currSequentialWOffset += exer->sequentialIoSize; 
            exer->seqWriteDone += exer->sequentialIoSize;

        }
    }
}


/**
 * Return the next random offset and numReqs for the
 * specified action.   Updates exer->reqState[], possibly
 * returning ACTION_NONE if the next offset(s) is(are)
 * busy.
 */
static void
fs_volexerGetNextRandOffsetLen(
    fs_VolExer  *exer,
    uint32      *action,
    uint64      *offset,
    int32       *numReqs)
{
    uint32  rand;

    if (exer->rw[*action].nextOffset == -1)
    {
        // Get the next offset and numReqs from the random
        // generator.
        random_r(&exer->rw[*action].randData, &rand);
        *offset = rand & exer->blksRandMask;
        *offset = MIN(*offset, exer->numBlksRandom - 1);
        *numReqs = MAX(1, MIN((rand & 0xf00) >> 8, 8));
        while (*offset + *numReqs > exer->numBlksRandom)
        {
            *numReqs -= 1;
        }
    }
    else
    {
        // We were stalled due to a busy condition
        // last time, try the stalled offset again.
        *offset = exer->rw[*action].nextOffset;
        *numReqs = exer->rw[*action].nextNumReqs;

        exer->rw[*action].nextOffset = -1;
    }

    // We might need to wait before starting this request.
    fs_volexerCheckBusy(exer, action, offset, numReqs);

    // Update state and genNum as required.
    if (*action != FS_VOL_EXER_ACTION_NONE)
    {
        fs_volexerUpdateStateGenNum(exer, *action, *offset, *numReqs);
    }
}


/**
 * Kick off the next IO request.  Returns TRUE if one
 * could be started, otherwise FALSE in the case of the
 * next one must wait due to the offset being busy.
 */
static boolean
fs_volexerStartNextIO(
    fs_VolExer  *exer)
{
    fs_RdwrReq	*rdwr;
    uint64  	offset;
    uint32  	loadType, numReqs, action, i;
    uchar8  	*ptr;
    boolean 	rc;

    action = -1;

    // Select next action (read, write).
    if (exer->ioType == FS_VOL_EXER_IOTYPE_RO)
    {
        exer->state = FS_VOL_EXER_RUNNING;
        action = EVM_READ_REQ;
    }
    else if (exer->ioType == FS_VOL_EXER_IOTYPE_WO)
    {
        exer->state = FS_VOL_EXER_RUNNING;
        action = EVM_WRITE_REQ;
    }
    else if (exer->state == FS_VOL_EXER_INITING)
    {
        if (exer->rw[EVM_WRITE_REQ].totalDoneCount >= exer->load * 8)
        {
            exer->state = FS_VOL_EXER_RUNNING;
        }
        else
        {
            action = EVM_WRITE_REQ;
        }
    }

    // Select next offset, length.
    if (exer->loadType == FS_VOL_EXER_TYPE_BOTH)
    {
        uint32  rand;

        // Randomize whether to do a sequential or random.
        random_r(&exer->actionRandData, &rand);
        loadType = rand % FS_VOL_EXER_TYPE_BOTH;
    }
    else
    {
        loadType = exer->loadType;
    }

    if (action == -1)
    {
        if (loadType == FS_VOL_EXER_TYPE_SEQUENTIAL && 
            exer->seqReadDone >= exer->seqWriteDone)
        {
            // Must write, reads have caught up.
            action = EVM_WRITE_REQ;
        }
        else
        {
            uint32  rand;

            // Random next action.
            random_r(&exer->actionRandData, &rand);
            action = rand % FS_VOL_EXER_ACTION_MAX;
            FS_ASSERT(action == EVM_READ_REQ || action == EVM_WRITE_REQ);

            // do more reads if the writes get too far ahead
            if (loadType == FS_VOL_EXER_TYPE_SEQUENTIAL && 
                exer->seqWriteDone > exer->seqReadDone &&
                exer->seqWriteDone - exer->seqReadDone >
                    exer->numBlksSequential / 2)
            {
                action = EVM_READ_REQ;
            }
        }
    }

    if (loadType == FS_VOL_EXER_TYPE_SEQUENTIAL)
    {
        fs_volexerGetNextSeqOffsetLen(exer, &action, &offset, &numReqs);
    }
    else 
    {
        fs_volexerGetNextRandOffsetLen(exer, &action, &offset, &numReqs);
    }

    EXER_TRACE("volExerStartNext: action %d offset %lld numReqs %d\n",
               action, offset, numReqs);

    switch(action)
    {
    case EVM_READ_REQ:
        for (i = 0; i < numReqs; i++)
        {
            rdwr = eee_ramAlloc(sizeof *rdwr);
    
            rdwr->flags = 0;
            rdwr->smHandle = exer->handle;
            rdwr->doneEvent = FS_VOLEXER_EVENT_IO_DONE;
            rdwr->evmReq.id = exer->volId;
            rdwr->evmReq.reqSize = EVMBLKS_PER_FS_BLOCK;
            rdwr->evmReq.offset = ((offset + i) * FS_BLOCK_SIZE) /
                                                   EVM_LBLK_SIZE;
            rdwr->evmReq.clientPrivate = (address_t) rdwr;
            rdwr->evmReq.reqType = EVM_READ_REQ;
            if (exer->isUser)
            {
                rdwr->evmReq.dataType = EVM_USER_DATA_REQ;
            }
            else
            {
                rdwr->evmReq.dataType = EVM_META_DATA_REQ;
            }
            rdwr->evmReq.codOffset = 0;
            rdwr->evmReq.attributes = exer->attrib;
#if defined(AGILE_HARDWARE)
            eee_rtc(&rdwr->timestamp_start); 
#endif
            rc = evm_IOReq(&rdwr->evmReq,
                           fs_volexerIoDone);
            if (rc != NFX_OK)
            {
                eee_ramDealloc(rdwr);
                printf(
     "volexer vol 0x%llx failed to start READ req rc 0x%x, terminating test\n",
                    exer->volId.whole_thing, rc);
                fs_volexerStop(exer);
            }
            else
            {
                ++exer->ioPending;
                ++(exer->rw[action].totalPendingCount);
                (loadType == FS_VOL_EXER_TYPE_SEQUENTIAL) ? 
                                    ++(exer->rw[action].seqPendingCount) : 0;
            }
        }
        rc = TRUE;
        break;
    case EVM_WRITE_REQ:
        for (i = 0; i < numReqs; i++)
        {
            rdwr = eee_ramAlloc(sizeof *rdwr);

            if (exer->isUser)
            {
                rdwr->evmReq.dataType = EVM_USER_DATA_REQ;
                rdwr->evmReq.wDataHandle.codHandle = exer->writeCod;
            }
            else
            {
                ptr = fs_ramAlloc(FS_RAM_BUF_DATA, FS_BLOCK_SIZE);
                fs_volexerInitWriteData(exer, ptr, offset + i);
                rdwr->evmReq.dataType = EVM_META_DATA_REQ;
                rdwr->evmReq.wDataHandle.dataPtr = ptr;
            }
    
            rdwr->flags = 0;
            rdwr->smHandle = exer->handle;
            rdwr->doneEvent = FS_VOLEXER_EVENT_IO_DONE;
            rdwr->evmReq.id = exer->volId;
            rdwr->evmReq.reqSize = EVMBLKS_PER_FS_BLOCK;
            rdwr->evmReq.offset = ((offset + i) * FS_BLOCK_SIZE) /
                                                   EVM_LBLK_SIZE;
            rdwr->evmReq.clientPrivate = (address_t) rdwr;
            rdwr->evmReq.reqType = EVM_WRITE_REQ;
            rdwr->evmReq.codOffset = 0;
            rdwr->evmReq.attributes = exer->attrib;

#if defined(AGILE_HARDWARE)
            eee_rtc(&rdwr->timestamp_start); 
#endif
            rc = evm_IOReq(&rdwr->evmReq,
                           fs_volexerIoDone);
            if (rc != NFX_OK)
            {
                fs_volexerFreeRdwr (exer, rdwr);
                printf(
    "volexer vol 0x%llx failed to start WRITE req rc 0x%x, terminating test\n",
                    exer->volId.whole_thing, rc);
                fs_volexerStop(exer);
            }
            else
            {
                exer->ioPending += 1;
                ++(exer->rw[action].totalPendingCount);
                (loadType == FS_VOL_EXER_TYPE_SEQUENTIAL) ? 
                                    ++(exer->rw[action].seqPendingCount) : 0;
            }
        }
        rc = TRUE;
        break;
    default:
        // Taking no action this pass, desired offset
        // must be busy.
        rc = FALSE;
        break;
    }

    return rc;
}


/**
 * Action function to start the exerciser.
 */
int32
fs_volexer_go(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_VolExer  *exer = smCookie;
    int32       i;
    uint32      rand;

    EXER_TRACE("volexer_go on volId 0x%llx\n", exer->volId.whole_thing);

    // Start a bunch of writes to get things rolling.
    exer->state = FS_VOL_EXER_INITING;

    random_r(&exer->actionRandData, &rand);

    rand = rand % MIN (FS_VOL_EXER_MAX_IO_KICK, exer->load);

    // ensure at least one io.
    rand = (rand == 0) ? 1 : rand;

    for (i = 0; i < rand; i++)
    {
        if (fs_volexerStartNextIO(exer) == FALSE)
            break;
    }

#if defined(AGILE_HARDWARE)
    eee_rtcStart(&exer->startTime);
#endif

    return nextState;

} /* fs_volexer_go */

/**
 * Action function for an IO completion event.
 */
int32
fs_volexer_ioDone(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_VolExer  *exer = smCookie;
    fs_RdwrReq  *rdwr = event.p1.u.p;
    eeeRTC_t	end_time;
	eeeRTC_t	ioTime;
	eeeRTC_t	esmTime;
    uint32      i;
	uint32      rand;
	static eeeRTC_t fs_volexer_stats_rtc[FS_TIME_STATS_BUCKETS_MAX];
	static boolean first = TRUE;


	if (first) {
		first = FALSE;
		for (i=0; i<FS_TIME_STATS_BUCKETS_MAX; i++) {
			eee_msToRTC((uint32)fs_volexer_stats_usec[i],
						&(fs_volexer_stats_rtc[i]));
		}
	}
    exer->ioPending -= 1;
    --(exer->rw[rdwr->evmReq.reqType].totalPendingCount);
    ++(exer->rw[rdwr->evmReq.reqType].totalDoneCount);

    if (rdwr->evmReq.offset >= exer->numBlksRandom || 
                        exer->loadType == FS_VOL_EXER_TYPE_SEQUENTIAL) {

        --(exer->rw[rdwr->evmReq.reqType].seqPendingCount);
        ++(exer->rw[rdwr->evmReq.reqType].seqDoneCount);

    }

    exer->reqState[rdwr->evmReq.offset / EVMBLKS_PER_FS_BLOCK] &=
                         ~FS_VOL_EXER_STATE_BUSY;

    if (rdwr->evmStat.status != NFX_OK)
    {
        fs_volexerStop(exer);
        fs_volexerFreeRdwr(exer, rdwr);
        return nextState;
    }

    //
    //  Gather timing statistics only for successful io completions.
    //
	(void)eee_rtcSub(&rdwr->timestamp_return,
						  &rdwr->timestamp_start,
						  &ioTime);
    eee_rtc(&end_time);
	(void)eee_rtcSub(&end_time,
					 &rdwr->timestamp_return,
					 &esmTime);

    if (rdwr->evmReq.reqType == EVM_READ_REQ)
    {
		(void)eee_rtcIncr(&exer->totalRespTimeRead, &ioTime);

        for (i = 0; i < FS_TIME_STATS_BUCKETS_MAX; i++) 
        {

            if (eee_rtcCompare(&ioTime, &(fs_volexer_stats_rtc[i])) < 0) { 

                exer->respTimeRead[i]++;
                break;
			}
        }

        if (i == FS_TIME_STATS_BUCKETS_MAX) 
        {
            exer->respTimeRead[i-1]++;
        }

		if (eee_rtcCompare(&exer->respTimeMaxRead, &ioTime) < 0) {
			exer->respTimeMaxRead = ioTime;
		}
        exer->ioQDepthMaxRead = MAX(exer->ioQDepthMaxRead, exer->ioPending);
		if ((eee_rtcCompare(&exer->respTimeMinRead, eee_rtcZero()) == 0) ||
			(eee_rtcCompare(&exer->respTimeMinRead, &ioTime) > 0)) {
			exer->respTimeMinRead = ioTime;
		}
        exer->ioQDepthMinRead = (exer->ioQDepthMinRead == 0) ? exer->ioPending : MIN(exer->ioQDepthMinRead, exer->ioPending);
		if (eee_rtcCompare(&exer->esmTimeMaxRead, &esmTime) < 0) {
			exer->esmTimeMaxRead = esmTime;
		}
		if ((eee_rtcCompare(&exer->esmTimeMinRead, eee_rtcZero()) == 0) ||
			(eee_rtcCompare(&exer->esmTimeMinRead, &esmTime) > 0)) {
			exer->esmTimeMinRead = esmTime;
		}
	}
    else 
    {
		(void)eee_rtcIncr(&exer->totalRespTimeWrite, &ioTime);

        for (i = 0; i < FS_TIME_STATS_BUCKETS_MAX; i++) 
        {

            if (eee_rtcCompare(&ioTime, &(fs_volexer_stats_rtc[i])) < 0) { 

                exer->respTimeWrite[i]++;
                break;
			}
        }

        if (i == FS_TIME_STATS_BUCKETS_MAX) 
        {
            exer->respTimeWrite[i-1]++;
        }

		if (eee_rtcCompare(&exer->respTimeMaxWrite, &ioTime) < 0) {
			exer->respTimeMaxWrite = ioTime;
		}
        exer->ioQDepthMaxWrite = MAX(exer->ioQDepthMaxWrite, exer->ioPending);
		if ((eee_rtcCompare(&exer->respTimeMinWrite, eee_rtcZero()) == 0) ||
			(eee_rtcCompare(&exer->respTimeMinWrite, &ioTime) > 0)) {
			exer->respTimeMinWrite = ioTime;
		}
        exer->ioQDepthMinWrite = (exer->ioQDepthMinWrite == 0) ? exer->ioPending : MIN(exer->ioQDepthMinWrite, exer->ioPending);
		if (eee_rtcCompare(&exer->esmTimeMaxWrite, &esmTime) < 0) {
			exer->esmTimeMaxWrite = esmTime;
		}
		if ((eee_rtcCompare(&exer->esmTimeMinWrite, eee_rtcZero()) == 0) ||
			(eee_rtcCompare(&exer->esmTimeMinWrite, &esmTime) > 0)) {
			exer->esmTimeMinWrite = esmTime;
		}
    }
    
    if (!exer->isUser &&
        rdwr->evmReq.reqType == EVM_READ_REQ &&
        exer->ioType != FS_VOL_EXER_IOTYPE_RO &&
        (rdwr->evmReq.offset / EVMBLKS_PER_FS_BLOCK) >= exer->numBlksRandom)
    {
        // Compare cksum on read data.
        fs_volexerCheckReadData(exer, rdwr->evmStat.rDataHandle.dataPtr,
                            rdwr->evmReq.offset / EVMBLKS_PER_FS_BLOCK);
    }

    fs_volexerFreeRdwr(exer, rdwr);

    if (exer->state == FS_VOL_EXER_RUNNING ||
        exer->state == FS_VOL_EXER_INITING)
    {
        if (exer->ioPending < exer->load)
        {
            random_r(&exer->actionRandData, &rand);

            rand = rand % MIN(FS_VOL_EXER_MAX_IO_KICK,
                              exer->load - exer->ioPending);

            // ensure at least one io.
            rand = (rand == 0) ? 1 : rand;

            for (i = 0; i < rand; i++)
            {
                if (fs_volexerStartNextIO(exer) == FALSE)
                    break;
            }
        }
    }
    else if (exer->state == FS_VOL_EXER_STOPPING)
    {
        if (exer->ioPending == 0)
        {
            fs_volexerStop(exer);
        }
    }

    return nextState;

} /* fs_volexer_ioDone */


/**
 * Action function to stop the exerciser.
 */
int32
fs_volexer_stop(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_VolExer  *exer = smCookie;

    fs_volexerStop(exer);

    return nextState;

} /* fs_volexer_stop */


/**
 * Create action.
 */
int32
fs_volexer_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    // do nothing here.
    return nextState;

} /* fs_volexer_create */


/* 
 * bad_state action.
 */
int32
fs_volexer_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    FS_ASSERT(FALSE);

    return nextState;

} /* fs_volexer_bad_state */

#endif // SSC
