/*------------------------------------------------------------------------------
 * RCS: $Id: fs-vol-log.c,v 1.124 2002/12/10 19:34:28 bob Exp $
 * Name         : fs-vol-log.c
 *
 * Description  : The file system specific functions for 
 *                handling the log from a volume
 *
 * Created By   : Jan Vanhoutte
 *
 * Date Created :  11/10/2001
 *
   Volume logging architecture

            filesystem
                ^
                |  API in fs_log.h [ fs_logAddRecord and fs_logDeleteRecord ]
                v
          fs log manager  (log size Mega bytes, truncate if certain % is in use)
                ^
                |
                v
        [ optional a layer to compress the logging for one transaction ]
                ^
                |
                v
           the volume logger 
           Use of volume log buffers to keep the disk spinning


   The files

   The volume log buffers
   - 0 or 1 I/O log buffer: 
        - write async to disk, 
        - when "write finished" 
          - inform all dependent tasks 
            if sync write: esm_yield
            if async write: inform that log is saved
          - start to write next
   - 0, 1 or more buffers ready to flush to log volume
   - current vol log buffer: write the log summaries
 
 *
 *------------------------------------------------------------------------------
 */

#if !defined(SSC)

//#define OFFSET_47573224

/*------------------------------------------------------------------------------
 * Header files.
 *------------------------------------------------------------------------------
 */
#include "fs-includes.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */
#define VL_RIGHT   1
#define VL_LEFT    -1

// Enough space for a maximal log record.
#define VL_TMPBUF_SIZE  (FS_LOG_RECORD_LEN_MAX + VL_BUFFER_SIZE)

#ifdef FS_DEBUG_VOL_LOG_BUCKETS
int32 fs_vl_dep_bucket[FS_VL_STATS_NBR_DEP_BUCKETS] = {1,2,4,8,16,100000};
#endif // FS_DEBUG_VOL_LOG_BUCKETS

// Volume Log Cache
fs_vlcCache vlcCache;

/*------------------------------------------------------------------------------
 * Function prototypes.
 *------------------------------------------------------------------------------
 */
static void vl_IONextBuf (fs_Context *context, vl_VolBufRoot *volBufRoot);

#define vl_validRecord(_value_) \
    (((_value_) < (VL_MAX_BUF_ID)) ? (TRUE) : (FALSE))

/*------------------------------------------------------------------------------
 * Name: ck_sum
 *
 * Description: calculate/verify the checksum
 *
 *   Input: 
 *   - start ptr (must be int aligned !!)
 *   - end ptr (not used) (must be int aligned !!)
 *   - ptr to cksum
 *   Output 
 *   - if cksum == NULL : just return the result
 *     else  update the check sum
 *   - result == 0 if OK
 *
 *   To verify: call with arg cksum set to NULL
 *   To set: init (*cksum) to 0, call the function
 *           with arg *cksum set
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 */
static int 
ck_sum (void *start, void *end, uint32 *cksum)
{
    int *st = start;
    int *en = end;
    int sum = 0;
    int result;
  
    while (st != en) 
    {
        sum += *st;
        st++;
    }

    if (cksum == NULL)
    {
        result = - sum ;
    }
    else
    {
        result = sum + *cksum;
        *cksum = - sum ;
    }

    return (result) ;  
}  // ck_sum

/*------------------------------------------------------------------------------
 *  Caller must hold volBufRoot->idSpin
 */
static void
vl_wakeupIO (
    fs_Context  *context)
{
    esm_event_t    event;
    vl_VolBufRoot  *volBufRoot;
    uint32         rc;

    FS_TRACE(context, FS_TRACE_VOLLOG, ("vl_wakeupIO"));

    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    if ((volBufRoot->nbr.currNbrIO >= VL_BUFFER_IO_MAX) ||
        volBufRoot->wakeupSignalled)
    {
        return;
    }

    volBufRoot->wakeupSignalled = 1;
    
    event.event = FS_VL_EVENT_WAKEUP;
    event.p1.u.p = NULL;
    event.p2.u.p = NULL;
    rc = esm_deliverHighPriorityEvent(volBufRoot->ioWrHandle, &event);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_EVENT_DELIVERY_FAIL, "vl_wakeupIO");
    }
}

/*------------------------------------------------------------------------------
 *
 */
uint32
vl_hashFileHandle (const void *key)
{
    uint64 v ;
    uint32 i ;
  
    v = *((uint64*)key) ;
    i = (uint32)v;
  
    return (i); 
}

int32
vl_cmpFileHandle (const void *key1, const void *key2)
{
    uint64 v1, v2 ;
    v1 = *((uint64*)key1) ;
    v2 = *((uint64*)key2) ;
    if (v1 == v2)
        return 0;
    else
        return 1;
}

void
vl_destructFileHandle (void *key, void *obj)
{
    FS_TRACE(NULL, FS_TRACE_VOLLOG, ("vl_destructFileHandle %p", obj));

    // could happen during abort
    // FS_ASSERT (fhEntry->nbrEntries == 0);
    
    fs_ramDealloc (FS_RAM_LOG_WRITE_BUF, obj);
}


/*------------------------------------------------------------------------------
 * Increment the count of transactions pending write to the log.
 */
extern void
vl_addUnstableId(fs_Context *   context,
                 uint64 *       inum)
{
    vl_unstableFileHandle  *fhEntry;
    vl_VolBufRoot *volBufRoot = context->fileSystem->logRoot->volBufRoot;

    NFX_SPIN_LOCK(&volBufRoot->hidIdSpin);
    if ((fhEntry = (vl_unstableFileHandle *)
         chash_find(volBufRoot->fileHandleHid, inum)) == NULL)
    {
        fhEntry = fs_ramAlloc(FS_RAM_LOG_WRITE_BUF,
                              sizeof(vl_unstableFileHandle));
        DLL_INIT(&(fhEntry->waiterHdr));
        fhEntry->nbrEntries = 1;
        fhEntry->inum = *inum;
        if ((chash_insert(volBufRoot->fileHandleHid,
                          &fhEntry->inum, fhEntry)) < 0)
        {
            fs_error(NULL, FS_INTERNAL_ERR, "Add fileHandle to hash table");
        }
    }
    else
    {
        fhEntry->nbrEntries++;
    }
    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("Add entry for unstable commit %p inum %lld nbr %d", 
              context, *inum, fhEntry->nbrEntries));
    NFX_SPIN_UNLOCK(&volBufRoot->hidIdSpin);
    return ;
}

extern void
vl_removeUnstableId(
    fs_Context         *context,
    uint64             *inum)
{
    vl_unstableFileHandle *fhEntry;
    fs_Waiter *waiter, *waiterNext;
    vl_VolBufRoot *volBufRoot = context->fileSystem->logRoot->volBufRoot;
    uint32                 rc;

    NFX_SPIN_LOCK(&volBufRoot->hidIdSpin);
    if ((fhEntry = (vl_unstableFileHandle *)
         chash_find(volBufRoot->fileHandleHid, inum)) == NULL)
    {
        fs_error(NULL, FS_INTERNAL_ERR, "NO fileHandle found in hash table");
    }
    else
    {
        FS_ASSERT (fhEntry->nbrEntries > 0);
        fhEntry->nbrEntries--;
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("Remove entry for unstable commit %p inum %lld nbr %d", 
                  context, *inum, fhEntry->nbrEntries));

        DLL_ITERATE_NEXT(waiter, waiterNext, &fhEntry->waiterHdr, dllEntry)
        {
            FS_ASSERT (waiter->numWait > 0);
            waiter->numWait-- ;
            if (waiter->numWait == 0)
            {
                rc = esm_deliverHighPriorityEvent(waiter->handle,
                                                  &waiter->event);
                if (rc != NFX_OK)
                {
                    fs_error(NULL, FS_EVENT_DELIVERY_FAIL, "logCompl event");
                }
                DLL_REMOVE(&fhEntry->waiterHdr, waiter, dllEntry);
                fs_ramDealloc(FS_RAM_WAITER, waiter);
            }
        } // DLL_ITERATE_NEXT
        
        if (fhEntry->nbrEntries == 0)
        {
            FS_ASSERT (DLL_IS_EMPTY(&fhEntry->waiterHdr));
            // was last entry on list
            if (chash_remove(volBufRoot->fileHandleHid, inum) != 0)
            {
                fs_error(NULL, FS_INTERNAL_ERR,
                         "fileHandle could not be removed from hash table");
            }
        }
    }
    NFX_SPIN_UNLOCK(&volBufRoot->hidIdSpin);
    return;
}

/*------------------------------------------------------------------------------
 * For buf needing allocation, perform allocation and
 * if its an UNSTABLE write record a zero-fill transaction
 * for it in the log.
 */
extern void
vl_addUserBuf (
    fs_Context        *context,
    fs_Inode          *inode,
    fs_Buf            *buf,
    fs_WriteStableHow stableHow)
{
#ifdef FS_DEBUG_CHECK_CRASH
    extern boolean fs_debugZeroFillYield;
#endif

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
    FS_TRACE(context, FS_TRACE_VOLLOG,
            ("Add buf for DirectDisk %p %p inum %lld", 
             context, buf, buf->inum));

    if (!fs_blockIsValid(buf->block))
    {
        FS_ASSERT(buf->zeroFillLogId == 0);

        fs_bufAllocBuf (context, inode, buf, TRUE);
        FS_ASSERT(fs_blockIsValid(buf->block));

        if (stableHow == FS_WRITE_STABLE_UNSTABLE)
        {
            // Add a new transaction to the log, to cause
            // zero-fill of the buf during replay.
            fs_logZeroFillAndCommit(context, buf);

#ifdef FS_DEBUG_CHECK_CRASH
            if (fs_debugZeroFillYield)
            {
                // This simply causes the thread to yield for
                // a period to allow some log debugging (we cause
                // a crash after this and test replay unreferenced
                // zero-fill records).
                fs_debugZeroFillYield = FALSE;
                printf("Logid %d will yield for 30 seconds (tail %d)...\n",
                    context->logSummary->id,
                    context->fileSystem->logRoot->tail);
                fs_logDumpIncoreData(context, context->logSummary);
                fs_sleep(context, 30000);
                printf("Logid %d resuming from yield (tail %d)...\n",
                    context->logSummary->id,
                    context->fileSystem->logRoot->tail);
            }
#endif
        }
    }
}

/*------------------------------------------------------------------------------
 * Block until log is written out to disk
 */
static int32
vl_waitLogIsFlushed(
    fs_Context         *context,
    vl_VolBufRoot      *volBufRoot)
{
    esm_event_t            wakeEvent;
    int32                  count = 0; 
    
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
    while (1)
    {
        // just sleep for 1 sec (the min time) 
        fs_yield(context, &wakeEvent, 2, FS_YIELD_TYPE_IO_LOG);

        if (wakeEvent.event == THREAD_EVENT_TIMEOUT)
        {
            if (count++ > 1)
            {
                FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                        ("vl_waitLogIsFlushed: force a postponed log write"));
            }
            
            //   force a vol log write
            NFX_SPIN_LOCK(&volBufRoot->idSpin);
            volBufRoot->forceWrite += 1;
            context->fileSystem->logRoot->logStats.forceWrite++;
            context->fileSystem->logRoot->logStats.forceWriteDone++;
            vl_IONextBuf (context, volBufRoot);
            NFX_SPIN_UNLOCK(&volBufRoot->idSpin);
            // continue with the normal yield
        }
        else
        if ((wakeEvent.event == FS_THREAD_EVENT_LOG_COMPL) &&
            (wakeEvent.p1.u.p == NULL) &&
            (wakeEvent.p2.u.p == NULL))
        {
            return (NFX_OK);
        }
        else
        {
            fs_error(context, FS_INTERNAL_ERR,
                     "expected TO or log compl event, got %d %p %p",
                     wakeEvent.event, wakeEvent.p1.u.p, wakeEvent.p2.u.p);
            return (NFX_ERR);
        }
    }
#else
    fs_yield(context, &wakeEvent, FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_LOG);

    // Make sure the event we got is the event
    // we were waiting for!  If not, return error.
    if (!((wakeEvent.event == FS_THREAD_EVENT_LOG_COMPL) &&
          (wakeEvent.p1.u.p == NULL) && (wakeEvent.p2.u.p == NULL)))
    {
        fs_error(NULL, FS_INTERNAL_ERR, "Received bad event %d %p %p",
                 wakeEvent.event, wakeEvent.p1.u.p, wakeEvent.p2.u.p);
        return (NFX_ERR);
    }
    return (NFX_OK);
#endif
}

/*------------------------------------------------------------------------------
 * Block here until all transactions affecting the 
 * specified inode are stable in the log.
 */
extern void
vl_waitInodeStable(
    fs_Context         *context,
    uint64             inum)
{
    vl_unstableFileHandle  *fhEntry;
    vl_VolBufRoot *volBufRoot = context->fileSystem->logRoot->volBufRoot;

    NFX_SPIN_LOCK(&volBufRoot->hidIdSpin);

    if ((fhEntry = (vl_unstableFileHandle *)
         chash_find(volBufRoot->fileHandleHid, &inum)) == NULL)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_checkUnstableDone no hash %p inum %lld", 
                  context, inum));
        // no entry found: all (unstable) log entries are written to disk 
        NFX_SPIN_UNLOCK(&volBufRoot->hidIdSpin);
    }
    else
    {
        fs_Waiter       *waiter;

        FS_ASSERT (fhEntry->nbrEntries > 0);
        waiter = fs_ramAlloc(FS_RAM_WAITER, sizeof *waiter);
        waiter->handle = context->smHandle;
        waiter->event.event  = FS_THREAD_EVENT_LOG_COMPL;
        waiter->event.p1.u.p = NULL; 
        waiter->event.p2.u.p = NULL; 
        waiter->numWait = fhEntry->nbrEntries;
        DLL_ISOLATE_ELEM(waiter, dllEntry);
        DLL_INSERT_TAIL(&fhEntry->waiterHdr, waiter, dllEntry);
        NFX_SPIN_UNLOCK(&volBufRoot->hidIdSpin);
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_checkUnstableDone yield   %p inum %lld", 
                  context, inum));
        vl_waitLogIsFlushed(context, volBufRoot);
    }

    return;
}

/*-----------------------------------------------------------------------------
 * fs_logReadVol
 *
 * Description: Called to read something in the volume logger
 *              No path of return, if it fails
 *
 * vl_readBlock   read (and caches) a data block from disk
 * vl_readBuf     uses vl_readBlock to get DiskBuf
 *                and executes the cksum
 * fs_logReadVol  uses vl_readBuf
 *                if requested size fits: OK return ptr
 *                if offset+size maps in 2 or more log buffers
 *                   one local buffer is allocated and memcpy.
 *
 * Clean cache, if we write to disk
 */

extern void
vl_clearBlockCache(fs_Context * context)
    
{
    fs_FileSystem * fs = context->fileSystem;
    vl_VolBufRoot * volBufRoot = fs->logRoot->volBufRoot;
    vl_BlockCache * blockCache = &(volBufRoot->blockCache);
    fs_vlcEntry *   currEntry = NULL;
    fs_vlcEntry *   nextEntry = NULL;
    
    FS_TRACE(context, FS_TRACE_VOLLOG, ("vl_clearBlockCache"));
    //printf("vl_clearBlockCache()\n");

    if (blockCache->validCache == FALSE)
    {
        // nothing to clear
        return;
    }

    // remove all entries in the cache related to this volume
    DLL_ITERATE_NEXT(currEntry, nextEntry, &(blockCache->hdr), entry)
    {
        int32 rc;

        FS_ASSERT(currEntry->volId.whole_thing == context->volId.whole_thing);

        if (currEntry->ultrabuf != NULL)
            fs_ramDealloc(FS_RAM_READ_DATA, currEntry->ultrabuf);
        rc = chash_remove(vlcCache.hashTab, (void *) currEntry);
        FS_ASSERT(rc == NFX_OK);
        DLL_REMOVE(&(blockCache->hdr), currEntry, entry);
        fs_ramDealloc(FS_RAM_VLC, currEntry);
    } // DLL_ITERATE_NEXT
    FS_ASSERT(DLL_IS_EMPTY(&(blockCache->hdr)));
    
    if (volBufRoot->tmpBuffer)
    {
        fs_ramDealloc(FS_RAM_LOG_REC, volBufRoot->tmpBuffer);
        volBufRoot->tmpBuffer = NULL;
    }
    
    blockCache->validCache = FALSE;
}

/* Read ahead a log block.
 * The response will be received asynchronously by the IO daemon
 * (not by this thread!).
 * Note: caller is responsible to lock the cache.
 */
extern void
fs_logVolReadAhead(fs_Context *     context,
                   vl_VolBufRoot *  volBufRoot,
                   evm_volOffset_t  readOffset,
                   uint32           nbBuf)
{
    fs_RdwrReq *    rdwr = fs_rdwrAlloc();
    esm_handle_t    handle = context->smHandle;
#ifdef ISABELLE // debug
    vl_BlockCache * blockCache = &volBufRoot->blockCache;
#endif // ISABELLE
    fs_vlcEntry *   newEntry = fs_vlcAddEntry(context, readOffset, nbBuf);
    
    //printf("Send IO for readOffset %lld nbBuf %d\n", readOffset, nbBuf);
    FS_ASSERT((readOffset % volBufRoot->bufSize) == 0);
    FS_ASSERT((0 < nbBuf) && (nbBuf <= VL_NB_READ_BUF));
    FS_ASSERT(newEntry != NULL);

    context->smHandle = volBufRoot->ioWrHandle;
    fs_volReadAsync(context, rdwr, readOffset, (nbBuf * volBufRoot->bufSize),
                    FS_VL_EVENT_IO_DONE, SCSI_MSG_ATTR_NO_REORDER,
                    (void *) (address_t) newEntry);
    context->smHandle = handle;
    
#ifdef ISABELLE // debug
    blockCache->stats.nbReadAhead++;    
#endif // ISABELLE
    return;
}
                 
static char *
vl_readBlock(fs_Context *       context,
             vl_VolBufRoot *    volBufRoot,
             evm_volOffset_t    offset,    // block offset in bytes
             evm_volSize_t      size,
             boolean *          fromCache) // indicates if returned from cache
    
{
    fs_RdwrReq *    rdwr;
    fs_FileSystem * fs = context->fileSystem;
    fs_LogFsRoot *  logRoot = fs->logRoot;
    vl_BlockCache * blockCache = &volBufRoot->blockCache;
    fs_vlcEntry *   vlcEntry = NULL;

    //printf("Read block offset %lld\n", offset);
    FS_ASSERT(size <= FS_BLOCK_SIZE);
    FS_ASSERT(((offset / volBufRoot->bufSize) * volBufRoot->bufSize) == offset);

    *fromCache = FALSE;
    if (blockCache->validCache == FALSE)
    {
        fs_logVolInitCache(blockCache);
    }
    else
    {
        NFX_SPIN_LOCK(&(blockCache->spin));
        vlcEntry = fs_vlcGetEntry(context, offset);
        if (vlcEntry != NULL)
        {
#ifdef DEBUG
            boolean found = FALSE;
#endif
            uint32  i;

            FS_ASSERT(size == volBufRoot->bufSize);

#ifdef DEBUG
            for (i = 0; i < vlcEntry->nbBuf; i++)
            {
                if ((vlcEntry->offset + (i * volBufRoot->bufSize)) == offset)
                {
                    found = TRUE;
                    break;
                }
            } // for
            FS_ASSERT(found == TRUE);
#else // DEBUG
            i = (offset - vlcEntry->offset) / volBufRoot->bufSize;
            FS_ASSERT((0<=i) && (i<VL_NB_READ_BUF));
#endif // DEBUG
            
            if (vlcEntry->valid == TRUE)
            {
                FS_TRACE(context, FS_TRACE_VOLLOG,
                         ("vl_readBlock from cache offset %lld size %lld", 
                          offset, size));
                NFX_SPIN_UNLOCK(&(blockCache->spin));
                *fromCache = TRUE;
                return(vlcEntry->ultrabuf + (i * volBufRoot->bufSize));
            }
            else if (vlcEntry->reading == TRUE) 
            {
                esm_event_t wakeEvent;
#ifdef DEBUG
                eeeRTC_t t1, t2, diff;
                eee_rtc(&t1);
#endif // DEBUG
                //FS_ASSERT(i == 0);
                //if (i != 0) printf("vl_readBlock() i(%d) != 0\n", i);
                
                /* The data we need are being read,
                 * let's wait for the response...
                 */
                vlcEntry->guyWaiting = context->smHandle;
                //printf("Wait IO for offset %lld I am %d\n",
                //vlcEntry->offset, vlcEntry->guyWaiting);
                NFX_SPIN_UNLOCK(&(blockCache->spin));
                
                fs_yield(context, &wakeEvent,
                         FS_YIELD_TIME_IO, FS_YIELD_TYPE_IO_LOG);
                if (!((wakeEvent.event == FS_THREAD_EVENT_LOG_COMPL) &&
                      (wakeEvent.p1.u.p == NULL) && (wakeEvent.p2.u.p == NULL)))
                {
                    fs_error(NULL, FS_INTERNAL_ERR,
                             "Received bad event %d %p %p "
                             "expected %d (nil) (nil)",
                             wakeEvent.event, wakeEvent.p1.u.p,
                             wakeEvent.p2.u.p, FS_THREAD_EVENT_LOG_COMPL);
                }
                
                NFX_SPIN_LOCK(&(blockCache->spin));
                FS_ASSERT(vlcEntry->valid == TRUE);
                FS_ASSERT(vlcEntry->reading == FALSE);
                FS_ASSERT(vlcEntry->ultrabuf != NULL);
                FS_ASSERT(vlcEntry->offset == offset - (i * volBufRoot->bufSize));
                vlcEntry->guyWaiting = 0;
                NFX_SPIN_UNLOCK(&(blockCache->spin));
                *fromCache = TRUE;
                
#ifdef ISABELLE // debug
                eee_rtc(&t2);
                eee_rtcSub(&t2, &t1, &diff);
                
                blockCache->stats.nbYield++;
#ifdef FS_STANDALONE
                blockCache->stats.totalYieldTime += diff;
#else // FS_STANDALONE
                blockCache->stats.totalYieldTime += eee_rtcToUsec(&diff);
#endif // FS_STANDALONE
#endif // ISABELLE
                return(vlcEntry->ultrabuf + (i * volBufRoot->bufSize));
            }
            else
            {
                // should never come there
                FS_ASSERT(FALSE);
            }
        }
        NFX_SPIN_UNLOCK(&(blockCache->spin));
    }

    FS_ASSERT(blockCache->replay == FALSE); // should not be here during replay!

    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readBlock offset %lld size %lld", offset, size));
        rdwr = fs_rdwrAlloc();
        if (logRoot->logDevType == FS_LOG_DEVTYPE_SELF)
        {
            fs_volRead (context, rdwr, offset, size);
        }
        else
        {
            volBufRoot->ioRdContext->smHandle = context->smHandle;
            fs_volRead (volBufRoot->ioRdContext, rdwr, offset, size);
        }
        
        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioTotalLogReads += 1;
        NFX_SPIN_UNLOCK(&fs_ioSpin);
        
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->ioTotalLogReads += 1;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
        
        FS_ASSERT((size / volBufRoot->bufSize) == 1);
    
        NFX_SPIN_LOCK(&(blockCache->spin));
        
        vlcEntry = fs_vlcAddEntry(context, offset, size/volBufRoot->bufSize);
        FS_ASSERT(vlcEntry != NULL);
        vlcEntry->valid = TRUE;
        vlcEntry->reading = FALSE;
        vlcEntry->ultrabuf = rdwr->evmStat.rDataHandle.dataPtr;
        
        fs_rdwrDealloc(rdwr);
        
        NFX_SPIN_UNLOCK(&(blockCache->spin));
    }

    return(vlcEntry->ultrabuf);
}

static vl_BufDiskHdr *
vl_readBuf (
    fs_Context      *context,
    vl_VolBufRoot   *volBufRoot,
    evm_volOffset_t offset,       // log file offset (including diskHdr overhead
    evm_volSize_t   size)
{
    vl_BufDiskHdr   *buf;
    char            *block;
    evm_volOffset_t blockOffset;
    boolean         fromCache;

    //printf("\tvl_readBuf() offset %lld size %lld\n", offset, size);
    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("vl_readBuf offset %lld size %lld", offset, size));

    FS_ASSERT (size <= volBufRoot->bufSize);
    FS_ASSERT (((offset / volBufRoot->bufSize) *
                volBufRoot->bufSize) == offset);

    blockOffset = ((evm_volOffset_t)offset /
                   volBufRoot->bufSize) * volBufRoot->bufSize ;
    block = vl_readBlock (context, volBufRoot, blockOffset,
                          volBufRoot->bufSize, &fromCache);
    buf = (vl_BufDiskHdr*)&block[offset - blockOffset] ;
    
    //printf ("Read %d %p offset %lld size %d cksum %d %d head %d time %d ioWriteTime %d\n", 
    //        buf->bufId, buf, offset, buf->size, buf->cksum, 
    //        ck_sum (buf, &(((char*)(buf))[buf->size]), NULL),
    //        buf->head, buf->timeStamp, buf->ioWriteTime);

    if (buf->magic != VL_LOG_MAGIC)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readBuf wrong Magic %d", buf->magic));
        return NULL;
    }

    if (buf->timeStamp != volBufRoot->timeStamp)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readBuf wrong timestamp 0x%x 0x%x", 
                  buf->timeStamp, volBufRoot->timeStamp));
        return NULL;
    }

    if ((buf->logInitTimeSec != volBufRoot->logInitTimeSec) ||
        (buf->logInitTimeNsec != volBufRoot->logInitTimeNsec))
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readBuf wrong init log timestamp 0x%x 0x%x 0x%x 0x%x", 
                  buf->logInitTimeSec, volBufRoot->logInitTimeSec,
                  buf->logInitTimeNsec, volBufRoot->logInitTimeNsec));
        return NULL;
    }

    if (!fromCache)
    {
        // read from disk, not from cache, need to check CRC
        if (ck_sum (buf, &(((char*)(buf))[buf->size]), NULL) != 0)
        {
            fs_warn(context, FS_LOG_READ_ERR, "vl_readBuf, wrong ck_sum at offset %lld ", offset);
            FS_TRACE(context, FS_TRACE_VOLLOG, ("vl_readBuf wrong ck_sum"));
            return NULL;
        }
    }
    // else, from cache
    // save on CRC calculation, each time we read a record from the same buf

    return (buf);
}

#define VL_LOG_DUMP_INPUT_TYPE_VOL_OFFSET   1
#define VL_LOG_DUMP_INPUT_TYPE_BUF_OFFSET   2
#define VL_LOG_DUMP_INPUT_TYPE_LOG_OFFSET   3
#define VL_LOG_DUMP_INPUT_TYPE_BUF_INDEX    4
extern void
vl_logDumpVollogBuf(
    fs_Context      *context,
    uint32          inputType,  // what kind of input is the offset
    uint64          offset,     // 
    uint32          size)       // hex dump of data in logbuffer
{
    vl_VolBufRoot   *volBufRoot;
    uint32          bufIndex;
    uint32          logOffset;
    evm_volOffset_t volOffset, bufOffset;
    vl_BufDiskHdr   *buf;

    volBufRoot = context->fileSystem->logRoot->volBufRoot ;
 
    switch (inputType)
    {
    case VL_LOG_DUMP_INPUT_TYPE_VOL_OFFSET:
        volOffset = offset;
        bufOffset = (offset / volBufRoot->bufSize) * volBufRoot->bufSize ;
        bufIndex = (bufOffset - volBufRoot->startOffset) / volBufRoot->bufSize;
        logOffset = (uint32)((volOffset - volBufRoot->startOffset) -
                             sizeof (vl_LogHdr) - 
            (sizeof (vl_BufDiskHdr) * (bufIndex + 1)));
        break;
    case VL_LOG_DUMP_INPUT_TYPE_BUF_OFFSET:
        bufOffset = (offset / volBufRoot->bufSize) * volBufRoot->bufSize ;
        if (offset == bufOffset)
        {
            printf("bufOffset (%lld) was not at buf boundary, set to %lld\n",
                   offset, bufOffset);
        }
        bufIndex = (bufOffset - volBufRoot->startOffset) / volBufRoot->bufSize;
        volOffset = bufOffset + sizeof (vl_BufDiskHdr);
        if (bufIndex == 0)
        {
            volOffset += sizeof (vl_LogHdr);
        }
        logOffset = (uint32)((volOffset - volBufRoot->startOffset) -
                             sizeof (vl_LogHdr) - 
                             (sizeof (vl_BufDiskHdr) * (bufIndex + 1)));
        break;
    case VL_LOG_DUMP_INPUT_TYPE_LOG_OFFSET:
        logOffset = (uint32)offset;
        bufIndex = (offset + sizeof (vl_LogHdr)) / (volBufRoot->bufSize -
                                                    sizeof (vl_BufDiskHdr));
        volOffset = volBufRoot->startOffset + (evm_volOffset_t)offset + 
            sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * (bufIndex + 1);
        bufOffset = (volOffset / volBufRoot->bufSize) * volBufRoot->bufSize;
        break;
    case VL_LOG_DUMP_INPUT_TYPE_BUF_INDEX:
        bufIndex = (uint32)offset ;
        bufOffset = volBufRoot->startOffset + bufIndex * volBufRoot->bufSize ;
        // go to the beginning of the payload in this buf
        volOffset = bufOffset + sizeof (vl_BufDiskHdr);
        if (bufIndex == 0)
        {
            volOffset += sizeof (vl_LogHdr);
        }
        logOffset = (uint32)((volOffset - volBufRoot->startOffset) -
                             sizeof (vl_LogHdr) - 
                             (sizeof (vl_BufDiskHdr) * (bufIndex + 1)));
        break;
    default:
        printf("vl_logDumpVollogBuf: Wrong input type\n");
        return;
    }

    printf("Read a vollog buffer at Bufoffset %lld, VolOffset %lld, "
           "start %lld\n", bufOffset, volOffset, volBufRoot->startOffset);
    printf("   BufIndex %d, log offset %d, logsize %d\n",
           bufIndex, logOffset, context->fileSystem->logSize);
    
    if (logOffset > context->fileSystem->logSize)
    {
        printf ("   Error logOffset > context->fileSystem->logSize\n");
        return;
    }
    
    buf = vl_readBuf (context, volBufRoot, bufOffset, volBufRoot->bufSize);
    if (buf)
    {
        printf ("   tail %d head %d size %d bufId %d\n", 
                buf->tail, buf->head, buf->size, buf->bufId);
    }
    else
    {
        printf ("   Error to get this buf\n");
    }
}

extern uchar8 *
fs_logReadVol(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          offset,    // log offset
    uint32          *size)
{
    vl_VolBufRoot * volBufRoot = context->fileSystem->logRoot->volBufRoot;
    fs_LogInfo *    logInfo = &(context->fileSystem->logRoot->logInfo);
    evm_volOffset_t volOffset, bufOffset, bufOffsetEnd;
    vl_BufDiskHdr * buf = NULL;
    uint32          bytesThis = 0, bytesDone;
    int32           bytesTodo;
    uint32          nbrOffBuf;
    uint32          prevBufId;
#ifdef OFFSET_47573224
    boolean offset47573224 = FALSE;
#endif // OFFSET_47573224
    FS_LOOP_DBG_DECL(dbgNumLoops);

    //printf("fs_logReadVol(%d, %d)\n", offset, *size);
    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));
    FS_ASSERT (offset < context->fileSystem->logSize);

    nbrOffBuf = ((offset + sizeof (vl_LogHdr)) /
                 (volBufRoot->bufSize - sizeof (vl_BufDiskHdr)) + 1);
    volOffset = (volBufRoot->startOffset + (evm_volOffset_t)offset +
                 sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * nbrOffBuf);
    bufOffset = (volOffset / volBufRoot->bufSize) * volBufRoot->bufSize;
    bufOffsetEnd = (((volOffset + *size) / volBufRoot->bufSize) *
                    volBufRoot->bufSize);

#ifdef OFFSET_47573224
    if (offset >= 47573224)
    {
        offset47573224 = TRUE;
        printf("*** fs_logReadVol(%d)\n", offset);
        printf("*** nbrOffBuf %d volOffset %lld bufOffset %lld bufOffsetEnd %lld\n",
               nbrOffBuf, volOffset, bufOffset, bufOffsetEnd);
    }
#endif // OFFSET_47573224

    // warning if next buffer add some more sizeof (vl_BufDiskHdr)
    if ((bufOffset == bufOffsetEnd) &&
        ((offset + *size) <= context->fileSystem->logSize))
    {
        // requested data all fits in one vol buf buffer

        buf = vl_readBuf (context, volBufRoot, bufOffset, volBufRoot->bufSize);
    
        if (buf == NULL)
        {
            // no valid data
            *size = 0;
            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_readVol (1) offset %d size %d (%d %lld %lld %lld)", 
                      offset, *size, nbrOffBuf, volOffset, bufOffset,
                      volBufRoot->startOffset));
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("LOG READ ERROR  invalid buf (1) offset %d !!!", offset));
            return NULL ;
        }
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readVol (2) offset %d size %d (%d %lld %lld %lld "
                  "%d %d %d)",  offset, *size, nbrOffBuf, volOffset,
                  bufOffset, volBufRoot->startOffset, volBufRoot->bufSize,
                  buf->size, *size));
        // READ_MORE
        if (logInfo->tail >= offset)
            *size = MIN((buf->size - (volOffset - bufOffset)),
                        (logInfo->tail - offset));
        else
            *size = buf->size - (volOffset - bufOffset);
#ifdef OFFSET_47573224
        if (offset47573224)
        {
            printf("*** *size %d\n", *size);
        }
#endif // OFFSET_47573224
        return ((uchar8*)&(((char*)buf)[volOffset - bufOffset])) ;
    }
    
    if (volBufRoot->tmpBuffer)
        fs_ramDealloc(FS_RAM_LOG_REC, volBufRoot->tmpBuffer);
    volBufRoot->tmpBuffer = fs_ramAlloc(FS_RAM_LOG_REC, VL_TMPBUF_SIZE);
    
    bytesDone = 0;
    bytesTodo = MIN(*size, VL_TMPBUF_SIZE);
    prevBufId = VL_MAX_BUF_ID ;

    while (bytesTodo > 0)
    {
        FS_LOOP_DBG_CHECK(dbgNumLoops, 10);

#ifdef OFFSET_47573224
        if (offset47573224)
        {
            printf("*** while (%d > 0)\n", bytesTodo);
        }
#endif // OFFSET_47573224
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readVol read buf %d %d", bytesDone, bytesTodo));
        buf = vl_readBuf (context, volBufRoot, bufOffset, volBufRoot->bufSize);
        
        if (buf == NULL) // No valid buffer (cksum or other problem)
        {
            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_readVol (3) offset %d size %d (%d %d %d %d %d)", 
                      offset, *size, prevBufId, buf->bufId, 
                      bytesTodo, bytesThis, bytesDone));
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("LOG READ ERROR invalid buf (2)  offset %d !!!", offset));
            return NULL ;
        }

        // READ_MORE
        if (logInfo->tail >= offset)
            bytesThis = MIN((buf->size - (volOffset - bufOffset)),
                            (logInfo->tail - offset));
        else
            bytesThis = buf->size - (volOffset - bufOffset);
        
#ifdef OFFSET_47573224
        if (offset47573224)
        {
            printf("*** bytesThis %d\n", bytesThis);
        }
#endif // OFFSET_47573224

        // For debug, should not read past tail
        if ((bytesThis == 0) &&
            (buf->size < volBufRoot->bufSize) &&    // buffer not completely used
            (offset < context->fileSystem->logSize))// and not and end of log : this is end of log
        {
            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_readVol (4) offset %d size %d "
                      "(%d %d %d %d %d %d %lld %lld)", 
                      offset, *size, prevBufId, buf->bufId, 
                      bytesTodo, bytesThis, bytesDone,
                      buf->size, volOffset, bufOffset));
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("LOG READ ERROR offset %d id %d size %d !!!",
                     offset, buf->bufId, buf->size));
            return NULL ;
        }
        
        // For debug, should not read past tail
        if ((prevBufId < VL_MAX_BUF_ID) &&
            (prevBufId > buf->bufId) && // this buffer is an older buffer
            (buf->bufId != 0))          // If not wrap around
        {
            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_readVol (5) offset %d size %d (%d %d %d %d %d)", 
                      offset, *size, prevBufId, buf->bufId, 
                      bytesTodo, bytesThis, bytesDone));
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("LOG READ ERROR offset %d prevId %d id %d !!!",
                     offset, prevBufId, buf->bufId));
            return NULL ;
        }
        prevBufId = buf->bufId;

        memcpy (&volBufRoot->tmpBuffer[bytesDone], 
                (uchar8*)&(((char*)buf)[volOffset - bufOffset]), 
                bytesThis);

        FS_ASSERT (buf->size <= volBufRoot->bufSize);

        if (buf->size == volBufRoot->bufSize)
        {
            volOffset += sizeof (vl_BufDiskHdr) + bytesThis;
            bufOffset += volBufRoot->bufSize;
        }
        
        // should not go to next buf if this one is not fully used
        bytesDone += bytesThis;
        bytesTodo -= bytesThis;
        offset += bytesThis;
#ifdef OFFSET_47573224
        if (offset47573224)
        {
            printf("*** bytesDone %d bytesTodo %d offset %d\n",
                   bytesDone, bytesTodo, offset);
        }
#endif // OFFSET_47573224

        if (offset > context->fileSystem->logSize)
        {
            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_readVol (6) offset %d %d size %d "
                      "(%d %lld %lld %d %d %d)", 
                      context->fileSystem->logSize, offset, 
                      *size, nbrOffBuf, volOffset, bufOffset,
                      bytesTodo, bytesThis, bytesDone));
            FS_ASSERT (FALSE);
        }
        
        if (offset >= context->fileSystem->logSize)
        {
            volOffset = volBufRoot->startOffset + sizeof (vl_LogHdr) +
                sizeof (vl_BufDiskHdr);
            bufOffset = (volOffset / volBufRoot->bufSize) *
                volBufRoot->bufSize ;
            offset = 0;
        }
#ifdef OFFSET_47573224
        if (offset47573224)
        {
            printf("*** volOffset %lld bufOffset %lld offset %d\n",
                   volOffset, bufOffset, offset);
        }
#endif // OFFSET_47573224
    } // while (bytesTodo > 0)

    *size = bytesDone;
#ifdef OFFSET_47573224
    if (offset47573224)
    {
        printf("*** *size %d\n", *size);
    }
#endif // OFFSET_47573224

    if (bytesDone == 0)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_readVol (7) offset %d size %d (%d %lld %lld)", 
                  offset, *size, nbrOffBuf, volOffset, bufOffset));
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("LOG READ ERROR bytesDone == 0 offset %d  !!!", offset));
        return (NULL);
    }

    FS_TRACE(context, FS_TRACE_VOLLOG,
            ("vl_readVol (8) offset %d size %d (%d %lld %lld)", 
            offset, *size, nbrOffBuf, volOffset, bufOffset));

    return (volBufRoot->tmpBuffer);
} //fs_logReadVol

/*-----------------------------------------------------------------------
 * find the latest volume log buffer on disk
 *
 * Assume the buffer id is always incrementing, but can wrap around with 
 * a value which is much greater than the number of entries in the log.
 *
 * The buffers which are not in use have an invalid entry (format vol log).
 * The not used entries are always the first entries of the log. After a wrap
 * around, all the buffer are at least used once.
 *
 * BufId  is a number from 0..VL_MAX_BUF_ID (large number)
 * BufNbr is a number from 0..nbrOfBufsInVolLog  (sizeof(vollog) / volBufRoot->bufSize)
 *
 */

static vl_BufDiskHdr *
vl_findMostRecentBuf (
    fs_Context *context)
{
    uint32          step, dir, cnt, found, vl_max_nbr_probes ; 
    uint32          bufNbr_1, bufNbr_2;
    uint32          bufId_1, bufId_2;
    vl_BufDiskHdr   *buf = NULL;
    uint32          nbrOfBufsInVolLog;
    vl_VolBufRoot   *volBufRoot = context->fileSystem->logRoot->volBufRoot ;
    vl_VolBuf       *currBuf;
    uint32          bufNbr, bufNbrLast;
    evm_volOffset_t volOffset, bufOffset;

    // The only instance running is during replay
    // Todo: locking if needed during runtime 

    FS_ASSERT((context->fileSystem->logRoot->logDevType ==
               FS_LOG_DEVTYPE_SELF) ||
              (context->fileSystem->logRoot->logDevType ==
               FS_LOG_DEVTYPE_VOLID));

    nbrOfBufsInVolLog = VL_GET_NBR_OF_BUF(context->fileSystem->logSize, 
                                          volBufRoot->bufSize);
    vl_max_nbr_probes = log2_32 (powerOfTwoCeiling(context->fileSystem->logSize 
                                                   / volBufRoot->bufSize)) + 5 ;
    // 5 is a savety margin

    // should never see more than 1 CRC error contiguously on one log device 
#define VL_MAX_NBR_RETRIES   2

    cnt = 0 ;
    bufNbr_1 = 0; 

    do
    {
        buf = vl_readBuf(context, 
                         volBufRoot, 
                         (volBufRoot->startOffset + bufNbr_1 *
                          volBufRoot->bufSize), 
                         (evm_volSize_t)sizeof (vl_BufDiskHdr));
        cnt++;
    } while ((buf == NULL) && (bufNbr_1++ < VL_MAX_NBR_RETRIES));

    if (buf == NULL)
    {
        // totally empty log
        // replay finished
        fs_warn(context, FS_LOG_READ_ERR,
                "vl_findMostRecentBuf, failed finding good log entries");
        return (buf);
    }
    
    dir = VL_RIGHT ;
    step = (nbrOfBufsInVolLog - cnt) / 2;
    bufNbr_2 = step;
    cnt = 0 ;

    bufId_1 = buf->bufId ;
    bufId_2 = bufId_1;  // not needed, just to fix a compiler warning
    
    found = FALSE;
    while (!found)
    {
        buf = vl_readBuf(context, 
                         volBufRoot, 
                         (volBufRoot->startOffset + bufNbr_2 *
                          volBufRoot->bufSize), 
                         (evm_volSize_t)sizeof (vl_BufDiskHdr));
        cnt++;
        if (cnt > vl_max_nbr_probes)
            return NULL;
        
        if (buf != NULL)
        {
            bufId_2 = buf->bufId;

            if (bufId_2 > bufId_1)
            {
                // normally, 2 more recent
        
                if (bufId_2 > (bufId_1 + VL_BUF_ID_WRAP))
                {
                    // bufId wrap around, 1 is more recent
                    dir = VL_LEFT;
                }
                else
                {
                    // 2 more recent
                    dir = VL_RIGHT;
                }
            } 
            else 
            {
                // normally, 1 more recent
        
                if (bufId_1 > (bufId_2 + VL_BUF_ID_WRAP))
                {
                    // recId wrap around, 2 is more recent
                    dir = VL_RIGHT;
                }
                else
                {
                    // 1 more recent
                    dir = VL_LEFT ;
                }
            }
        }
        else
        {
            dir = VL_LEFT ;
        }

        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_findMostRecentBuf read BufNbr %d %p %d %d %d %d", 
                  bufNbr_2, buf, cnt, bufId_2, bufId_1, dir));

        //printf ("Nbr %d %d Id %d %d step %d dir %d in nbr %d iter %d \n",
        //        bufNbr_1, bufNbr_2, bufId_1, bufId_2,
        //        step, dir, nbrOfBufsInVolLog, cnt);

        step = (bufNbr_2 - bufNbr_1);

        if (dir == VL_RIGHT)
        {
            bufNbr_1 = bufNbr_2;
            bufId_1 = bufId_2;
        }
        
        step = (step + 1) / 2;

        if (step == 1 && dir == VL_RIGHT && bufNbr_2 == (nbrOfBufsInVolLog - 1))
        {
            // printf ("end of list\n");
            step = 0;
        }
        
        bufNbr_2 = bufNbr_2 + step * dir;

        if (bufNbr_2 >= nbrOfBufsInVolLog)
        {
            // printf ("Pass end of list\n");
            bufNbr_2 = nbrOfBufsInVolLog - 1;
        }

        if (bufNbr_1 == bufNbr_2)
            found = TRUE;
         
    }

    buf = vl_readBuf (context, 
                      volBufRoot, 
                      volBufRoot->startOffset + bufNbr_1 * volBufRoot->bufSize, 
                      (evm_volSize_t)sizeof (vl_BufDiskHdr));

    FS_ASSERT (buf->timeStamp == volBufRoot->timeStamp);

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("vl_findMostRecentBuf Found BufNbr %d (%p %d %d)", 
              bufNbr_1, buf, nbrOfBufsInVolLog, cnt));

    bufNbr = (buf->tail + sizeof (vl_LogHdr)) /
        (volBufRoot->bufSize - sizeof (vl_BufDiskHdr));

    if (bufNbr != bufNbr_1)
    {
        bufNbrLast = ((context->fileSystem->logSize - 1) + sizeof (vl_LogHdr)) / 
            (volBufRoot->bufSize - sizeof (vl_BufDiskHdr));

        if ((bufNbr == (bufNbr_1 - 1)) ||
            ((bufNbr_1 == 0) && (bufNbr == bufNbrLast)))
        {
            // this means that the log tail is in the prev buffer

            FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                    ("vl_findMostRecentBuf: get previous most recent buf %d (%d %d)", 
                     bufNbr, bufNbr_1, buf->tail));

            bufNbr_1 = bufNbr;
            
            buf = vl_readBuf (context, 
                              volBufRoot, 
                              volBufRoot->startOffset + bufNbr_1 * volBufRoot->bufSize, 
                              (evm_volSize_t)sizeof (vl_BufDiskHdr));

            if (buf == NULL)
            {
                fs_warn(context, FS_LOG_READ_ERR,
                        "vl_findMostRecentBuf: no valid buf found %d %d", 
                        bufNbr_1, buf->tail);
                return NULL;
            }

            FS_ASSERT (buf->timeStamp == volBufRoot->timeStamp);

            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_findMostRecentBuf Found BufNbr %d (%p %d %d)", 
                      bufNbr_1, buf, nbrOfBufsInVolLog, cnt));
        }
        else
        {
            fs_warn(context, FS_LOG_READ_ERR,
                    "vl_findMostRecentBuf bufNbr does not match %d %d %d", 
                    bufNbr, bufNbr_1, buf->tail);
            return NULL;
        }
    }
    
    volOffset = volBufRoot->startOffset + (evm_volOffset_t)buf->tail +
        sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * (bufNbr_1 + 1);
    bufOffset = (volOffset / volBufRoot->bufSize) * volBufRoot->bufSize;

    // This is the most recent
    volBufRoot->currBufId = buf->bufId + 1;
    if (volBufRoot->currBufId >= VL_MAX_BUF_ID)
    {
        volBufRoot->currBufId = 0;
    }

    volBufRoot->currBufNbr = bufNbr_1 + 1;
    if (volBufRoot->currBufNbr >=
        VL_GET_NBR_OF_BUF(context->fileSystem->logSize, volBufRoot->bufSize))
    {
        volBufRoot->currBufNbr = 0;
    }
    
    currBuf = volBufRoot->hdr.tail;
    currBuf->coreHdr.head = buf->head;
    currBuf->coreHdr.tail = buf->tail;
    currBuf->coreHdr.currCommits = 0;
    currBuf->coreHdr.currNbr = bufNbr_1;

    currBuf->coreHdr.size = (volOffset - bufOffset);
    currBuf->diskLogBuf->diskHdr.size = currBuf->coreHdr.size;

    //printf("Update size after replay %d %d %d %lld %lld %d %lld\n",
    // 	bufNbr, currBuf->coreHdr.currNbr, currBuf->coreHdr.size,
	//	volOffset, bufOffset, buf->nbrCommits, volBufRoot->startOffset);
    //printf ("  head %d tail %d id %d size %d\n",
    //    buf->head, buf->tail, buf->bufId, buf->size);

    memcpy (&(currBuf->diskLogBuf->diskHdr), buf, buf->size);

    currBuf->diskLogBuf->diskHdr.bufId = volBufRoot->currBufId - 1;


    volBufRoot->head = buf->head;
    volBufRoot->currCommits = 0;
    
    //printf ("Replay found buf index %d in %d total in %d iterations\n",
    //bufNbr_1, nbrOfBufsInVolLog, cnt);
    //printf ("  head %d tail %d id %d size %d\n",
    //buf->head, buf->tail, buf->bufId, buf->size);

    return (buf);

}  // vl_findMostRecentBuf

/*------------------------------------------------------------------------------
 *
 */
static inline vl_VolBuf * 
vl_allocLogBuf(int32 what, int32 nbytes)
{
    vl_VolBuf *buf;

    buf = (vl_VolBuf*)fs_ramAlloc(what, nbytes);
    buf->diskLogBuf = (vl_Disk_Buffer*)fs_ramAlloc(FS_RAM_WRITE_DATA, sizeof(vl_Disk_Buffer));
    return (buf);
}

static inline void
vl_deallocLogBuf(int32 what, vl_VolBuf *buf)
{
    fs_ramDealloc(FS_RAM_WRITE_DATA, buf->diskLogBuf);
    if (buf->coreHdr.ioDoneExp != NULL) {

        fs_ramDealloc(FS_RAM_LOG_BUF_DEP, buf->coreHdr.ioDoneExp);
        buf->coreHdr.ioDoneExp = NULL;

    } 
    fs_ramDealloc(what, buf);
}

static inline void
vl_freeLogBuf(int32 what, vl_VolBuf *buf)
{
    fs_ramDealloc(FS_RAM_WRITE_DATA, buf->diskLogBuf);
    if (buf->coreHdr.ioDoneExp != NULL) {

        fs_ramDealloc(FS_RAM_LOG_BUF_DEP, buf->coreHdr.ioDoneExp);
        buf->coreHdr.ioDoneExp = NULL;

    } 
    fs_ramDealloc(what, buf);
}

/*------------------------------------------------------------------------------
 *
 */
void
vl_getNewBuf (
    fs_Context      *context,
    vl_VolBufRoot   *volBufRoot,
    uint32          currCommits)
{
    // volBufRoot->idSpin locked by caller

    vl_VolBuf       *currBuf;

    currBuf = vl_allocLogBuf(FS_RAM_LOG_BUF, sizeof(vl_VolBuf));

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("vl_getNewBuf %d %d %p %p", 
              volBufRoot->currBufId, volBufRoot->currBufNbr, 
              currBuf, &currBuf->diskLogBuf));

    NFX_SPIN_INIT(&currBuf->coreHdr.idSpin,SPINLOCK_LEVEL_UNKNOWN);
    currBuf->coreHdr.state = VL_BUF_STATE_CURR;
    currBuf->coreHdr.size = sizeof(vl_BufDiskHdr);
    currBuf->coreHdr.currCommits = currCommits;
    currBuf->coreHdr.currNbr = volBufRoot->currBufNbr;
    currBuf->coreHdr.ioFinished = VL_BUF_IOFINISHED_NOT_SET;
    NFX_SPIN_LOCK(&context->fileSystem->spin);
    currBuf->coreHdr.blksUsed = context->fileSystem->commitedBlksUsed;
    currBuf->coreHdr.inodesUsed = context->fileSystem->commitedInodesUsed;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);

    DLL_ISOLATE_ELEM(currBuf, entry);

    currBuf->coreHdr.ioDoneUsed = 0 ;
#ifdef FS_DEBUG_VOL_LOG_BUCKETS
    bzero (&currBuf->coreHdr.ioDone, sizeof (vl_BufIODone) * VL_IO_DONE_ARR_MAX);
#endif
    currBuf->coreHdr.ioDoneExp = NULL;
    
    currBuf->diskLogBuf->diskHdr.bufId = volBufRoot->currBufId;
    currBuf->diskLogBuf->diskHdr.head = volBufRoot->head;
    // cksum, size, head, etc will be initilazed at I/O time (vl_setReadyForIO)

    if (volBufRoot->currBufNbr == 0)
    {
        vl_LogHdr       *logHdr;

        logHdr = (vl_LogHdr*)(&currBuf->diskLogBuf->data[0]);
        logHdr->magic = volBufRoot->magic;
        logHdr->version = VL_LOG_VERSION_ONE;
        logHdr->timeStamp = volBufRoot->timeStamp;
        logHdr->logInitTimeSec = volBufRoot->logInitTimeSec;
        logHdr->logInitTimeNsec = volBufRoot->logInitTimeNsec;
        logHdr->logSize = context->fileSystem->logSize;
        logHdr->bufSize = volBufRoot->bufSize;
        logHdr->cksum = 0;
        ck_sum (logHdr, logHdr + 1, &(logHdr->cksum));

        currBuf->coreHdr.size += sizeof(vl_LogHdr);
    }
    
    // prepare Id and Nbr for next buffer
    volBufRoot->currBufId++;
    if (volBufRoot->currBufId >= VL_MAX_BUF_ID)
        volBufRoot->currBufId = 0;
    
    volBufRoot->currBufNbr++;
    if (volBufRoot->currBufNbr >=
        VL_GET_NBR_OF_BUF(context->fileSystem->logSize, volBufRoot->bufSize))
        volBufRoot->currBufNbr = 0;

    DLL_INSERT_TAIL(&(volBufRoot->hdr), currBuf, entry);
    volBufRoot->nbr.currNbrBuf++;
    if (volBufRoot->nbr.maxNbrBuf < volBufRoot->nbr.currNbrBuf)
        volBufRoot->nbr.maxNbrBuf = volBufRoot->nbr.currNbrBuf;
}

static void
vl_initVolBufRoot (
    fs_Context      *context,
    fs_LogFsRoot    *logRoot)
{
    vl_VolBufRoot   *volBufRoot;
    vl_BlockCache   *blockCache;
 
    volBufRoot = (vl_VolBufRoot *) fs_ramAlloc(FS_RAM_CORE_MISC,
                                               sizeof(vl_VolBufRoot));
    bzero (volBufRoot, sizeof(vl_VolBufRoot));
    NFX_SPIN_INIT(&volBufRoot->idSpin,SPINLOCK_LEVEL_UNKNOWN);

    NFX_SPIN_LOCK(&volBufRoot->idSpin);

    logRoot->volBufRoot = volBufRoot;

    DLL_INIT(&(volBufRoot->hdr));

    volBufRoot->bufSize = VL_BUFFER_SIZE;
    volBufRoot->magic = VL_LOG_MAGIC;
    volBufRoot->timeStamp = context->fileSystem->super->mkfsTime;
    volBufRoot->logInitTimeSec = context->fileSystem->super->logInitTimeSec;
    volBufRoot->logInitTimeNsec = context->fileSystem->super->logInitTimeNsec;

    switch (logRoot->logDevType)
    {
        case FS_LOG_DEVTYPE_SELF:
            volBufRoot->startOffset =
                context->fileSystem->super->logStartOffset;
            break;
        case FS_LOG_DEVTYPE_VOLID:
            volBufRoot->startOffset = 0;
            break;
    }

    vl_getNewBuf (context, volBufRoot, 0) ;

    blockCache = &(volBufRoot->blockCache);
    NFX_SPIN_INIT(&blockCache->spin, SPINLOCK_LEVEL_UNKNOWN);
    blockCache->validCache = FALSE;

    volBufRoot->copyOfCurrentBuf = NULL;
    vl_startIODaemon (context);

    NFX_SPIN_INIT(&volBufRoot->hidIdSpin, SPINLOCK_LEVEL_UNKNOWN);
    volBufRoot->fileHandleHid = chash_create(100, // Todo: dimensioning
                                             vl_hashFileHandle,
                                             vl_cmpFileHandle,
                                             vl_destructFileHandle);
    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);
}

extern int32
fs_logReadVolLogHdr (
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          *logSize,
    uint32          *bufSize)
{
    vl_LogHdr       *logHdr;
    vl_BufDiskHdr   *buf;
    vl_VolBufRoot   *volBufRoot = logRoot->volBufRoot ;

    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    if (volBufRoot == NULL)
    {
        vl_initVolBufRoot (context, logRoot);
        volBufRoot = logRoot->volBufRoot;
    }
    
    buf = vl_readBuf(context, 
                     volBufRoot, 
                     volBufRoot->startOffset, // + 0, the first bufNbr
                     (evm_volSize_t)sizeof (vl_BufDiskHdr) + sizeof (vl_LogHdr));
    if (buf == NULL)
    {
        // invalid log
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("fs_logReadVolLogHdr() - Invalid log %lld.\n",
                 volBufRoot->startOffset));
        return (NFX_ERR);
    }

    logHdr = (vl_LogHdr *)(buf + 1);

    if (logHdr->magic != VL_LOG_MAGIC)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_logReadVolLogHdr: bad magic %d", logHdr->magic));
        return (NFX_ERR);
    }
    if (logHdr->version != VL_LOG_VERSION_ONE)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_logReadVolLogHdr: bad version %d", logHdr->version));
        return (NFX_ERR);
    }
    if (logHdr->timeStamp != volBufRoot->timeStamp)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_logReadVolLogHdr wrong timestamp 0x%x 0x%x", 
                  logHdr->timeStamp, volBufRoot->timeStamp));
        return (NFX_ERR);
    }
    if ((logHdr->logInitTimeSec != volBufRoot->logInitTimeSec) ||
        (logHdr->logInitTimeNsec != volBufRoot->logInitTimeNsec))
    {
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_logReadVolLogHdr wrong init log timestamp 0x%x 0x%x 0x%x 0x%x", 
                  logHdr->logInitTimeSec, volBufRoot->logInitTimeSec,
                  logHdr->logInitTimeNsec, volBufRoot->logInitTimeNsec));
        return (NFX_ERR);
    }
    if (ck_sum (logHdr, logHdr + 1, NULL) != 0)
    {
        FS_TRACE(context, FS_TRACE_VOLLOG, ("vl_logReadVolLogHdr: bad cksum"));
        return (NFX_ERR);
    }

    *logSize = logHdr->logSize;
    *bufSize = logHdr->bufSize;
    FS_ASSERT(logHdr->logSize == context->fileSystem->logSize);

    //volBufRoot->bufSize = logHdr->bufSize;
    //volBufRoot->magic = logHdr->magic;

    volBufRoot->logInitTimeSec = logHdr->logInitTimeSec;
    volBufRoot->logInitTimeNsec = logHdr->logInitTimeNsec;

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("fs_logReadVolLogHdr size %d, bufsize %d, M 0x%x, T 0x%x 0x%x 0x%x", 
              logHdr->logSize, logHdr->bufSize, 
              logHdr->magic, logHdr->timeStamp,
              logHdr->logInitTimeSec, logHdr->logInitTimeNsec));

    return (NFX_OK);
} // fs_logReadVolLogHdr

/*-----------------------------------------------------------------------------
 * fs_logVolWriteDevHdr
 *
 * Description: Called to find/allocate and initialize the volume log device.
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  3/20/2001
 */

extern int32
fs_logVolWriteDevHdr (
    fs_Context      *context,
    int32           operation,
    uint32          count)
{
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;
    vl_VolBufRoot   *volBufRoot = logRoot->volBufRoot ;

    FS_ASSERT (volBufRoot != NULL);    
    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    if (logRoot->inReplay)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("writedevhdr in state %d: operation %d, value %d, head %d",
                 context->fileSystem->state, operation,
                 count, volBufRoot->head));
        FS_ASSERT(FALSE);
    }

    NFX_SPIN_LOCK(&volBufRoot->idSpin);
    switch (operation)
    {
    case FS_LOG_WRITE_HEAD:
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("fs_logVolWriteDevHdr: head %d (%d old)", 
                  count, volBufRoot->head));
        volBufRoot->head = count;
        vl_wakeupIO (context);
        break;
    }
    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

    return NFX_OK;
}

/*-----------------------------------------------------------------------------
 *
 */
static void
vl_bufDoIO (
    fs_Context      *context,
    vl_VolBufRoot   *volBufRoot,
    vl_VolBuf       *buf,
    boolean         current)
{
    evm_volOffset_t offset;
    fs_RdwrReq      *rdwr;
    int32           loop;
    vl_Disk_Buffer  *diskBuf;
    uint64          attributes;
    vl_BufIODone    *ioDone;
#ifdef LOG_TUNE_STATISTICS
    fs_LogStats     *logStats;

    logStats = &(context->fileSystem->logRoot->logStats);
#endif
    
    FS_ASSERT(buf->coreHdr.ioFinished == VL_BUF_IOFINISHED_NOT_SET);

    // write the first buf with the LogHdr
    buf->diskLogBuf->diskHdr.magic = volBufRoot->magic;
    buf->diskLogBuf->diskHdr.timeStamp = volBufRoot->timeStamp;
    buf->diskLogBuf->diskHdr.logInitTimeSec = volBufRoot->logInitTimeSec;
    buf->diskLogBuf->diskHdr.logInitTimeNsec = volBufRoot->logInitTimeNsec;
//    buf->diskLogBuf->diskHdr.ioWriteTime = date.sec;
    buf->diskLogBuf->diskHdr.size = buf->coreHdr.size;

    if (buf->entry.next == DLL_NIL)
    {
        // No other buffer in the volume logging queue
        // This buf contains the latest up to date info
        buf->coreHdr.head = volBufRoot->head;
        buf->coreHdr.tail = (uint32)context->fileSystem->logRoot->tail;
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        buf->coreHdr.blksUsed = context->fileSystem->commitedBlksUsed;
        buf->coreHdr.inodesUsed = context->fileSystem->commitedInodesUsed;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    }
    
    buf->diskLogBuf->diskHdr.head = buf->coreHdr.head;
    buf->diskLogBuf->diskHdr.tail = buf->coreHdr.tail;
    buf->diskLogBuf->diskHdr.blksUsed = buf->coreHdr.blksUsed;
    buf->diskLogBuf->diskHdr.inodesUsed = buf->coreHdr.inodesUsed;   

    buf->diskLogBuf->diskHdr.cksum = 0;
    ck_sum (&buf->diskLogBuf->diskHdr, 
            &(((char*)(&buf->diskLogBuf->diskHdr))[buf->diskLogBuf->diskHdr.size]), 
            &(buf->diskLogBuf->diskHdr.cksum));

#ifdef LOG_TUNE_STATISTICS
    logStats->ioDoneUsed += buf->coreHdr.ioDoneUsed;
    logStats->bufUsed += 1;
    logStats->ioDoneMax = MAX(logStats->ioDoneMax, buf->coreHdr.ioDoneUsed);
    logStats->ioDoneMin = logStats->ioDoneMin ?
                          MIN(logStats->ioDoneMin, buf->coreHdr.ioDoneUsed) :
                          buf->coreHdr.ioDoneUsed;
    buf->coreHdr.ioDoneUsed ? ((buf->coreHdr.ioDoneUsed <= VL_IO_DONE_ARR_MAX) ?
                                           logStats->nbrFixedioDoneUsed++ :
                                           logStats->nbrVariableioDoneUsed++) :
                                           0;
#endif
                                            

    for (loop = 0; loop < buf->coreHdr.ioDoneUsed; loop++) {

        ioDone = (loop < VL_IO_DONE_ARR_MAX) ?
                              &buf->coreHdr.ioDone[loop] :
                              &buf->coreHdr.ioDoneExp[loop - VL_IO_DONE_ARR_MAX];

            if (ioDone->state == VL_IO_DONE_STATE_READY) {

                ioDone->state = VL_IO_DONE_STATE_IO;

            } else if (ioDone->state == VL_IO_DONE_STATE_UNSTABLE_READY) {

                ioDone->state = VL_IO_DONE_STATE_UNSTABLE_IO;

            }

    }

    if (volBufRoot->blockCache.validCache)
    {
        vl_clearBlockCache(context);
    }
    
    offset = volBufRoot->startOffset + 
        (evm_volOffset_t)(buf->coreHdr.currNbr * volBufRoot->bufSize);

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("vl_bufDoIO %d %p %lld", 
              buf->diskLogBuf->diskHdr.bufId, buf, offset));

    //printf ("Write %d %p offset %lld size %d cksum %d %d head %d Time %d ioWriteTime %d\n", 
    //        buf->diskLogBuf->diskHdr.bufId, buf, offset, 
    //        buf->diskLogBuf->diskHdr.size, buf->diskLogBuf->diskHdr.cksum,
    //        ck_sum (&buf->diskLogBuf->diskHdr, 
    //        &(((char*)(&buf->diskLogBuf->diskHdr))[buf->diskLogBuf->diskHdr.size]), NULL),
    //        buf->diskLogBuf->diskHdr.head, buf->diskLogBuf->diskHdr.timeStamp, 
    //        buf->diskLogBuf->diskHdr.ioWriteTime);
    if (current) {

        // if this is the current buffer, make a copy to avoid overwriting the same buffer.
        // CRC calculations in EVM.
        if (volBufRoot->copyOfCurrentBuf == NULL)
        {
            volBufRoot->copyOfCurrentBuf = fs_ramAlloc(FS_RAM_WRITE_DATA, 
                                                       sizeof(vl_Disk_Buffer));
        }
        diskBuf = volBufRoot->copyOfCurrentBuf;
        memcpy (diskBuf, &(buf->diskLogBuf->diskHdr),
                buf->diskLogBuf->diskHdr.size);
    }

    volBufRoot->nbr.currNbrIO++;
    volBufRoot->nbr.totalNbrIO++;
    if (volBufRoot->nbr.maxNbrIO < volBufRoot->nbr.currNbrIO)
        volBufRoot->nbr.maxNbrIO = volBufRoot->nbr.currNbrIO;
    
    rdwr = fs_rdwrAlloc();

    diskBuf = (vl_Disk_Buffer*) &(buf->diskLogBuf->diskHdr);

    if (context->fileSystem->logRoot->logDevType == FS_LOG_DEVTYPE_SELF)
    {
        FS_ASSERT(offset / FS_BLOCK_SIZE >=
                  context->fileSystem->super->numBlks + 1);
    }
    
    NFX_SPIN_LOCK(&fs_ioSpin);
    fs_ioTotalLogWrites += 1;
    NFX_SPIN_UNLOCK(&fs_ioSpin);

    NFX_SPIN_LOCK(&context->fileSystem->spin);
    context->fileSystem->ioTotalLogWrites += 1;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);

#ifdef FS_STANDALONE
    attributes = 0;
#else
    attributes = SCSI_MSG_ATTR_NO_REORDER;

    if (fs_incoreTune->vollogLoopBack == 1)
    {
        attributes |= SCSI_MSG_ATTR_LOOPBACK;
    }

    if (fs_incoreTune->vollogLoopBack == 2)
    {
        attributes = 0;
    }

#endif
    
    fs_volWriteAsync(context,
                     rdwr,
                     volBufRoot->ioWrContext->smHandle,
                     offset,
                     VL_BUFFER_SIZE,
                     diskBuf,
                     FS_VL_EVENT_IO_DONE,
                     attributes,
                     buf);    // private, esm 
}

static void
vl_setReadyForIO (
    fs_Context      *context,
    vl_VolBufRoot   *volBufRoot,
    vl_VolBuf       *buf)
{
    // volBufRoot->idSpin locked by caller
    // called, when current buffer is full, and can be written out to disk

    // The buf->coreHdr.idSpin is still locked by caller !!!!
    // At the end, this is unlocked, and the new buf is locked

    uint32 currCommits;

    vl_BufDiskHdr *diskHdr = &buf->diskLogBuf->diskHdr;

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("vl_setReadyForIO %d %p state %d", 
              diskHdr->bufId, buf, buf->coreHdr.state));
    FS_ASSERT((diskHdr->bufId == (volBufRoot->currBufId - 1)) ||
              (volBufRoot->currBufId == 0));

    currCommits = buf->coreHdr.currCommits;

    switch (buf->coreHdr.state)
    {
    case VL_BUF_STATE_IO_CURR:
        buf->coreHdr.state = VL_BUF_STATE_IO_READY;
        NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);
        break;
    case VL_BUF_STATE_IO_READY:
        // was already ready for IO, cant come here twice.
        fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
        break;
    case VL_BUF_STATE_IO:
        // in IO, can't come here again
        fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
        break;
    case VL_BUF_STATE_CURR:
        buf->coreHdr.state = VL_BUF_STATE_READY;
        // vl_wakeupIO (context);
        NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);
        vl_IONextBuf (context, volBufRoot);
        break;
    case VL_BUF_STATE_READY:
        // was already ready for IO, can't come here twice
        fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
        break;
    default:
        fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
        break;
    }

    vl_getNewBuf(context, volBufRoot, currCommits);
    buf = volBufRoot->hdr.tail;

    NFX_SPIN_LOCK(&buf->coreHdr.idSpin);
}

static uint32
vl_appendToCurrBuf (
    fs_Context          *context,
    fs_LogFsRoot        *logRoot,
    uint32              offset,
    uchar8              *data, 
    uint32              len,
    fs_LogStableWrite   stable,
    uint64              inum)
{
    // locked by caller
    // You can only append to buf, but don't change bufDiskHdr, use bufCoreHdr
    vl_VolBuf       *buf;
    vl_VolBufRoot   *volBufRoot = logRoot->volBufRoot ;
    uint32          bytesTodo = len;
    uint32          bytesThis ;
#ifdef FS_DEBUG_OFFSETS_VOL_LOG
    uint32 bufNbr;
    evm_volOffset_t volOffset, bufOffset;
#endif
    
    // Get current buf == get last buffer, must exist
    FS_ASSERT (volBufRoot != NULL);    
    buf = volBufRoot->hdr.tail ;

    NFX_SPIN_LOCK(&buf->coreHdr.idSpin);

    FS_ASSERT (buf != NULL);
    FS_ASSERT (buf != DLL_NIL);
    FS_ASSERT ((buf->diskLogBuf->diskHdr.bufId ==
                (volBufRoot->currBufId - 1)) ||
               (volBufRoot->currBufId == 0));
    FS_ASSERT ((buf->coreHdr.state == VL_BUF_STATE_CURR) ||
               (buf->coreHdr.state = VL_BUF_STATE_IO_CURR));
    
#ifdef FS_DEBUG_OFFSETS_VOL_LOG
    // only for debugging: check if log offset matches the current buffer size
    {
        
        bufNbr = (offset + sizeof (vl_LogHdr)) /
            (volBufRoot->bufSize - sizeof (vl_BufDiskHdr));
        volOffset = volBufRoot->startOffset + (evm_volOffset_t)offset + 
            sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * (bufNbr + 1);
        bufOffset = (volOffset / volBufRoot->bufSize) * volBufRoot->bufSize ;
        // printf("Append %d %d %d %d %d %lld %lld\n", 
        // offset, len, bufNbr, buf->coreHdr.currNbr, buf->coreHdr.size, volOffset, bufOffset);

        FS_ASSERT (bufNbr == buf->coreHdr.currNbr);
        FS_ASSERT ((volOffset - bufOffset) == buf->coreHdr.size);
    }
#endif
    
    bytesThis = MIN(bytesTodo, volBufRoot->bufSize - buf->coreHdr.size);
    bytesThis = MIN(bytesThis, context->fileSystem->logSize - offset);
        
    while (bytesTodo)
    {
        vl_BufDiskHdr *diskBuf = &(buf->diskLogBuf->diskHdr);
        
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("vl_appendToCurrBuf offset %d size %d", offset, bytesThis));

        //printf ("BufId A1 %d Curr %d State %d %p %d %d\n", 
        //        buf->diskHdr.bufId, volBufRoot->currBufId - 1, 
        //        buf->coreHdr.state, buf, bytesTodo, bytesThis);
        memcpy(&(((char*)diskBuf)[buf->coreHdr.size]), data, bytesThis);
        buf->coreHdr.size += bytesThis;

        buf->coreHdr.head = volBufRoot->head;
        buf->coreHdr.tail = (uint32)context->fileSystem->logRoot->tail;

        /* Remark: tail and nbrCommits does not include this record
         * While doing IO and nothing is backlogged, we will update
         * to the latest, info.
         */

        if ((volBufRoot->bufSize == buf->coreHdr.size) ||
            (bytesThis == (context->fileSystem->logSize - offset)))
        {
            // set this buffer ready for IO, 
            // and get a new buffer

            // The buf->coreHdr.idSpin is still locked, but will be
            // unlocked in vl_setReadyForIO, and the new buffer will
            // will have the buf->coreHdr.idSpin locked

            vl_setReadyForIO (context, volBufRoot, buf);
            buf = volBufRoot->hdr.tail ;
        }

        data += bytesThis;
        bytesTodo -= bytesThis ;

        if (bytesThis == (context->fileSystem->logSize - offset))
        {
            //printf("Wrap around needed %d %d %d %d %d %d %d %d\n", 
            //       context->fileSystem->logSize, bytesThis, volBufRoot->currBufNbr, 
            //       (offset + sizeof (vl_LogHdr)) / (volBufRoot->bufSize - sizeof (vl_BufDiskHdr)) + 1,
            //       VL_GET_NBR_OF_BUF(context->fileSystem->logSize, volBufRoot->bufSize),
            //       offset, volBufRoot->bufSize, buf->coreHdr.size);

            if ( ! (volBufRoot->currBufNbr == 1))
            {
                printf("currBufNbr %d len %d bytesThis %d logSize %d offset %d\n", 
                       volBufRoot->currBufNbr, len, bytesThis, context->fileSystem->logSize, offset);
                
                // FS_ASSERT(volBufRoot->currBufNbr == 1);
            }

            offset = 0 ;
        }
        else
        {
            offset += bytesThis;
        }
        
        bytesThis = MIN(bytesTodo, volBufRoot->bufSize - buf->coreHdr.size);
        bytesThis = MIN(bytesThis, context->fileSystem->logSize - offset);
    }
    // printf ("BufId A1 %d %p %d\n", buf->diskHdr.bufId, buf, buf->coreHdr.size);

    NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);

    return (offset);

}  // vl_appendToCurrBuf

extern void
fs_logVolWrite(
    fs_Context          *context,
    uint32              *offset,
    uchar8              *datap[LOG_IOVEC_MAX],
    uint32              lenp[LOG_IOVEC_MAX],
    fs_LogStableWrite   stable,
    uint64              inum)
{
    uint32          loop, bytesDone;
    uint32          nextOffset;
    fs_LogFsRoot    *logRoot = context->fileSystem->logRoot;
    vl_VolBufRoot   *volBufRoot = logRoot->volBufRoot;
    int32           UseIoDone = -1, loop1;
    vl_BufIODone    *ioDone;
    vl_VolBuf       *buf;
    
    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    NFX_SPIN_LOCK(&volBufRoot->idSpin);
    // lock the currBuf for writing the complete record

    *offset = logRoot->tail;
    nextOffset = *offset;

    for (bytesDone = 0, loop = 0 ; loop < LOG_IOVEC_MAX ; loop++)
    {
        if (lenp[loop] == 0)
            continue;

        nextOffset = vl_appendToCurrBuf (context, 
                                         logRoot, 
                                         nextOffset, 
                                         datap[loop], 
                                         lenp[loop], 
                                         stable, 
                                         inum);
        bytesDone += lenp[loop];
    }
    

    NFX_SPIN_LOCK(&logRoot->spin);
    FS_LOG_INCR_PTR(context->fileSystem->logSize,
                    logRoot->tail,
                    bytesDone);
    NFX_SPIN_UNLOCK(&logRoot->spin);

    buf = volBufRoot->hdr.tail ;

    if (stable == FS_LOG_WRITE_COMMIT_STABLE || stable == FS_LOG_WRITE_COMMIT_UNSTABLE) 
    {
        fs_LogSummary        *logSummary = context->logSummary;
        fs_LogLockedObjects  *objects;
        fs_Buf               *buffer, **bufListPtr;
        fs_Inode             *inode;
        int32                 numBufDeps = 0;

        NFX_SPIN_LOCK(&buf->coreHdr.idSpin);

        UseIoDone = buf->coreHdr.ioDoneUsed;

        FS_ASSERT ((UseIoDone < VL_IO_DONE_ARR_MAX + VL_IO_DONE_ARR_EXP)); 

        // if commit record, update the blksUsed and inodesUsed
        // in the context->fileSystem and buf->coreHdr
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->commitedBlksUsed += context->logSummary->blksUsed;
        context->fileSystem->commitedInodesUsed += context->logSummary->inodesUsed;
        buf->coreHdr.blksUsed = context->fileSystem->commitedBlksUsed;
        buf->coreHdr.inodesUsed = context->fileSystem->commitedInodesUsed;
        
        //printf("Commit %lld %lld - %lld %lld\n", 
        //       context->fileSystem->super->blksUsed, context->fileSystem->commitedBlksUsed, 
        //       context->fileSystem->super->inodesUsed, context->fileSystem->commitedInodesUsed);

        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
        
        if (UseIoDone == VL_IO_DONE_ARR_MAX && buf->coreHdr.ioDoneExp == NULL) 
        {
            buf->coreHdr.ioDoneExp = (vl_BufIODone *) 
                fs_ramAlloc(FS_RAM_LOG_BUF_DEP, VL_IO_DONE_EXP_SIZE); 
#ifdef FS_DEBUG_VOL_LOG_BUCKETS
            memset (currBuf->coreHdr.ioDoneExp, 0,  VL_IO_DONE_EXP_SIZE);
#endif
        } 

        if (UseIoDone < VL_IO_DONE_ARR_MAX) 
        {
            ioDone = &buf->coreHdr.ioDone[UseIoDone];
#ifdef FS_DEBUG_VOL_LOG_BUCKETS
            FS_ASSERT(ioDone->state == VL_IO_DONE_STATE_FREE);
#endif
        } 
        else 
        {
            ioDone = &buf->coreHdr.ioDoneExp[UseIoDone - VL_IO_DONE_ARR_MAX];
#ifdef FS_DEBUG_VOL_LOG_BUCKETS
            FS_ASSERT(ioDone->state == VL_IO_DONE_STATE_FREE);
#endif
        }

        if (stable == FS_LOG_WRITE_COMMIT_STABLE) 
        {
            ioDone->state = VL_IO_DONE_STATE_READY;
            ioDone->u.smHandle = context->smHandle;
            ++buf->coreHdr.ioDoneUsed;
        } 
        else 
            // if (stable == FS_LOG_WRITE_COMMIT_UNSTABLE) 
        {
            ioDone->state = VL_IO_DONE_STATE_UNSTABLE_READY;
            ioDone->u.inum = inum;
            ++buf->coreHdr.ioDoneUsed;
        }

        if (logSummary->bufCount > VL_DEF_NBR_DEP_BUF) 
        {
            ioDone->bufList = (fs_Buf**)fs_ramAlloc(FS_RAM_LOG_BUF_DEP, 
                                      sizeof(fs_Buf*) * logSummary->bufCount);
            bufListPtr = ioDone->bufList;
        } 
        else 
        {

            ioDone->bufList = NULL;
            bufListPtr = ioDone->buf;
        }

#ifdef LOG_TUNE_STATISTICS
        logSummary->bufCount ? ((logSummary->bufCount <= VL_DEF_NBR_DEP_BUF) ?
                                logRoot->logStats.nbrFixedBufDep++ :
                                logRoot->logStats.nbrVariableBufDep++) : 0;

        logRoot->logStats.maxBufDep = MAX (logSummary->bufCount, 
                                           logRoot->logStats.maxBufDep);
        
        logRoot->logStats.minBufDep = logRoot->logStats.minBufDep ?
            MIN (logSummary->bufCount, logRoot->logStats.minBufDep) :
            logSummary->bufCount;
#endif

        objects = &logSummary->bufArr;
        FS_ASSERT(objects != NULL);
      
        for (loop1 = 0; loop1 < logSummary->bufCount; loop1++) {

            if (loop1 > 0 && (loop1 % FS_LOG_OBJ_ARR_MAX) == 0)
            {
                objects = objects->next;
            }
            buffer = objects->obj[loop1 % FS_LOG_OBJ_ARR_MAX];
            
            if ((buffer) && (buffer->flags & FS_BUF_TYPE_META))
            {
                *(bufListPtr++) = buffer;
                fs_bufLogRefIncr (context, buffer);
                numBufDeps++;
            }
        }
       ioDone->nbrBufDep = numBufDeps;

#ifdef LOG_TUNE_STATISTICS
       logSummary->inodeCount ? ((logSummary->inodeCount <= VL_DEF_NBR_DEP_INODE) ?
                                 logRoot->logStats.nbrFixedInodeDep++ :
                                 logRoot->logStats.nbrVariableInodeDep++) :
           0;
       
       logRoot->logStats.maxInodeDep = MAX(logRoot->logStats.maxInodeDep, 
                                           logSummary->inodeCount);
       
       logRoot->logStats.minInodeDep = (logRoot->logStats.minInodeDep != 0) ?
           MIN (logRoot->logStats.minInodeDep, logSummary->inodeCount) :
           logRoot->logStats.minInodeDep;
#endif
      
       // Inherit the undoHdr from the logSummary
       ioDone->undoHdr.head = logSummary->undoHdr.head;
       ioDone->undoHdr.tail = logSummary->undoHdr.tail;
       DLL_INIT(&logSummary->undoHdr);

       ioDone->nbrInodeDep = logSummary->inodeCount;

       if (ioDone->nbrInodeDep > VL_DEF_NBR_DEP_INODE)
       {
           ioDone->u_inode.inodeList =
               (fs_Inode**)fs_ramAlloc(FS_RAM_LOG_BUF_DEP, 
                                       sizeof(fs_Inode*) * ioDone->nbrInodeDep);
       }

        objects = &logSummary->inodeArr;
        FS_ASSERT(objects != NULL);
        
        for (loop1 = 0; loop1 < ioDone->nbrInodeDep; loop1++)
        {
            if (loop1 > 0 && (loop1 % FS_LOG_OBJ_ARR_MAX) == 0)
            {
                objects = objects->next;
            }
            inode = objects->obj[loop1 % FS_LOG_OBJ_ARR_MAX];
            if (ioDone->nbrInodeDep <= VL_DEF_NBR_DEP_INODE)
            {
                ioDone->u_inode.inode[loop1] = inode;
            } else
            {
                ioDone->u_inode.inodeList[loop1] = inode;
            }
            if (inode != NULL)
            {
                fs_inodeLogRefIncr (context, inode);
            }
        }
        
        NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);

#ifdef FS_DEBUG_VOL_LOG_BUCKETS
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS ; loop1++)
        {
            if (logSummary->bufCount < fs_vl_dep_bucket[loop1])
            {
                volBufRoot->nbr.nbrDepBuf[loop1]++;
                break;
            }
        }

        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS ; loop1++)
        {
            if (logSummary->inodeCount < fs_vl_dep_bucket[loop1])
            {
                volBufRoot->nbr.nbrDepInode[loop1]++;
                break;
            }
        }
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS ; loop1++)
        {
            if (logSummary->undoCount < fs_vl_dep_bucket[loop1])
            { 
                volBufRoot->nbr.nbrDepUndo[loop1]++;
                break;
            }
        }
        if (volBufRoot->nbr.maxDepBuf < logSummary->bufCount)
            volBufRoot->nbr.maxDepBuf = logSummary->bufCount;
        if (volBufRoot->nbr.maxDepInode < logSummary->inodeCount)
            volBufRoot->nbr.maxDepInode = logSummary->inodeCount;
        if (volBufRoot->nbr.maxDepUndo < logSummary->undoCount)
            volBufRoot->nbr.maxDepUndo = logSummary->undoCount;
#endif // FS_DEBUG_VOL_LOG_BUCKETS

        volBufRoot->currCommits++;

        if (stable == FS_LOG_WRITE_COMMIT_STABLE) 
        {
            vl_IONextBuf (context, volBufRoot);
        } else 
        if (stable == FS_LOG_WRITE_COMMIT_UNSTABLE) 
        {
            vl_wakeupIO (context);
        }
    }
    
    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

} // fs_logVolWrite

/*-----------------------------------------------------------------------------
 * fs_logVolInitDevice
 *
 * Description: Called to initialize the volume log device.
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  3/20/2001
 */

extern int32
fs_logVolInitDevice(
    fs_Context      *context)
{
    vl_VolBufRoot   *volBufRoot;
    vl_VolBuf       *buf;
    fs_LogFsRoot    *logRoot;

    logRoot = context->fileSystem->logRoot;

    FS_TRACE(context, FS_TRACE_VOLLOG, ("fs_logVolInitDevice"));

    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    logRoot->headPtr = 0;           // offset on dev where headPtr is
    logRoot->head = 0;
    logRoot->tail = 0;
    logRoot->currLen = 0;
    logRoot->nextId = FS_LOG_ID_MIN;

    {
        uint64 blksUsed, inodesUsed;

        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        blksUsed = context->fileSystem->super->blksUsed;
        inodesUsed = context->fileSystem->super->inodesUsed;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->commitedBlksUsed = blksUsed;
        context->fileSystem->commitedInodesUsed = inodesUsed;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    }

    volBufRoot = logRoot->volBufRoot;

    if (volBufRoot == NULL)
    {
        vl_initVolBufRoot (context, logRoot);
        volBufRoot = logRoot->volBufRoot;
    }
    else
    {
        FS_ASSERT(volBufRoot->currBufNbr == 1); // just allocated the first one
        FS_ASSERT(volBufRoot->logInitTimeSec ==
                  context->fileSystem->super->logInitTimeSec);
        FS_ASSERT(volBufRoot->logInitTimeNsec ==
                  context->fileSystem->super->logInitTimeNsec);
    }

    NFX_SPIN_LOCK(&volBufRoot->idSpin);
        
    FS_ASSERT (volBufRoot->timeStamp == context->fileSystem->super->mkfsTime);

    // currBufNbr == 0, logHdr will be initialized
    buf = volBufRoot->hdr.head ;

    buf->coreHdr.currCommits = volBufRoot->currCommits;
    buf->coreHdr.state = VL_BUF_STATE_IO_CURR;
    vl_bufDoIO (context, volBufRoot, buf, TRUE);
    
    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

    return NFX_OK;
} /* fs_logVolInitDevice */

/*------------------------------------------------------------------------------
 * the volume log daemon and state machine functions
 */
extern int32
fs_vl_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    FS_TRACE(NULL, FS_TRACE_VOLLOG, ("fs_vl_create"));

    // do nothing here.
    return nextState;

} /* fs_vl_create */

/*------------------------------------------------------------------------------
 *
 */
static void
vl_IONextBuf (
    fs_Context      *context,
    vl_VolBufRoot   *volBufRoot)
{
    // check if new IO can be started
    // NFX_SPIN_LOCK(&volBufRoot->idSpin);   must be done by caller !!  

    vl_VolBuf       *buf;
    vl_VolBuf       *nextBuf;
    fs_LogFsRoot    *logRoot;
    int32           moreBufs;

    moreBufs =  fs_incoreTune->vollogIoMax - volBufRoot->nbr.currNbrIO;
    logRoot = context->fileSystem->logRoot;

    DLL_ITERATE_NEXT(buf, nextBuf, &volBufRoot->hdr, entry)
        {
            FS_TRACE(context, FS_TRACE_VOLLOG,
                     ("vl_IONextBuf %d %p state %d", 
                      buf->diskLogBuf->diskHdr.bufId, buf, buf->coreHdr.state));

            if (moreBufs <= 0)
            {
                return;
            }
            NFX_SPIN_LOCK(&buf->coreHdr.idSpin);

            switch (buf->coreHdr.state)
            {
            case VL_BUF_STATE_IO:
                // ignore, IO ongoing
                break;
            case VL_BUF_STATE_CURR:
                //
                // Perform IO to disk iff new data is added.
                //

                // check if new data is added.
                if ((volBufRoot->currCommits != buf->coreHdr.currCommits) ||
                    (volBufRoot->head != buf->diskLogBuf->diskHdr.head))
                {
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
                    if ((context->fileSystem->nbrOngoingTrans < fs_incoreTune->activeTransToWait) ||
                        (volBufRoot->forceWrite) || 
                        (context->reqHdr == NULL) ||
                        (context->reqHdr->requestType >= FS_REQ_NFS_MAX))
#endif
                    {
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
                        volBufRoot->postponedLogWrite = FALSE;
                        if (volBufRoot->forceWrite > 0)
                            volBufRoot->forceWrite -= 1;
#endif
                        buf->coreHdr.currCommits = volBufRoot->currCommits;
                        buf->coreHdr.state = VL_BUF_STATE_IO_CURR;
                        vl_bufDoIO (context, volBufRoot, buf, TRUE);
                        logRoot->logStats.ioPartialLogWrites++;
                        logRoot->logStats.ioPartialBytes += buf->coreHdr.size;
                        moreBufs--;
                    }
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
                    else
                    {
                        volBufRoot->postponedLogWrite = TRUE;
                        
                        logRoot->logStats.postponedPartialLogWrites++;
                        // wait to write the log, until the next transaction commit
                        // or, better, until ready.
                    }
#endif
                }
                break;

            case VL_BUF_STATE_IO_CURR:
                // ignore, io is ongoing.
                break;
                 
            case VL_BUF_STATE_READY:
                buf->coreHdr.state = VL_BUF_STATE_IO ;
                logRoot->logStats.ioTotalLogWrites++;
                logRoot->logStats.ioTotalBytes += buf->coreHdr.size;
                vl_bufDoIO (context, volBufRoot, buf, FALSE);
                moreBufs--;
                break;

            case VL_BUF_STATE_IO_READY:
                // If there is a buf waiting in line in front
                // of this one, we cannot schedule the next
                // one for IO.
                NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);
                return;

            default:
                FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                        ("buf = %p, state = %d", buf, buf->coreHdr.state));
                fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
                break;
            }
            
            NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);
        }
}

extern int32
fs_vl_wakeup(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    fs_Context      *context;
    vl_VolBufRoot   *volBufRoot;

    context = (fs_Context *) smCookie;
    volBufRoot = context->fileSystem->logRoot->volBufRoot ;

    NFX_SPIN_LOCK(&volBufRoot->idSpin);
    volBufRoot->wakeupSignalled = 0;
    vl_IONextBuf (context, volBufRoot);
    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

    return nextState;

} /* fs_vl_wakeup */


static void
vl_ioDoneRmBufDependencies (
    fs_Context *context,
    vl_VolBuf *buf)
{
    fs_LogUndoSummary  *undo, *undoNext;
    fs_Buf	           **bufListPtr;
    esm_event_t        event;
    int32              loop;
    int32              loop1;
    int32              rc;
    int32              nbrBufDep;
    int32              nbrInodeDep;
    vl_BufIODone       *ioDone;
    
    for (loop = 0; loop < buf->coreHdr.ioDoneUsed; loop++) {

        ioDone = (loop < VL_IO_DONE_ARR_MAX) ?
                 &(buf->coreHdr.ioDone[loop]) :
                 &(buf->coreHdr.ioDoneExp[loop - VL_IO_DONE_ARR_MAX]);


        if ((ioDone->state == VL_IO_DONE_STATE_IO) || 
            (ioDone->state == VL_IO_DONE_STATE_UNSTABLE_IO)) 
        {
            bufListPtr = ioDone->bufList == NULL ?  ioDone->buf :
                                                    ioDone->bufList;
            nbrBufDep = ioDone->nbrBufDep;

            for (loop1 = 0 ; loop1 < nbrBufDep ; loop1++) 
            {
                fs_bufLogRefDecr (context, *bufListPtr++);
            }

            if (ioDone->bufList != NULL) 
            {
                fs_ramDealloc(FS_RAM_LOG_BUF_DEP, ioDone->bufList);
            }
           
            DLL_ITERATE_NEXT(undo, undoNext, &ioDone->undoHdr, logEntry) 
            {
                fs_Buf  *fsbuf = undo->objSumm.objectPtr;

                FS_ASSERT(undo->objSumm.objectType == FS_LOG_DEP_BUF);
                FS_ASSERT(fsbuf->flags & FS_BUF_TYPE_META);

                DLL_REMOVE(&ioDone->undoHdr, undo, logEntry);

                // Move the undo record to the cando list.
                NFX_SPIN_LOCK(&fsbuf->spin);
                DLL_REMOVE(&fsbuf->undoHdr, undo, bufEntry);
                DLL_INSERT_TAIL(&fsbuf->candoHdr, undo, bufEntry);
                NFX_SPIN_UNLOCK(&fsbuf->spin);

                if (fs_trylockEnqueue(context, context, &fsbuf->lock,
                                      FS_LOCK_MODE_EXCL,
                                      FS_VL_EVENT_FS_BUF_LOCK,
                                      fsbuf, TRUE))
                {
                    fs_bufConvertCando(context, fsbuf);
                    fs_bufReleaseExclusive(context, fsbuf);
                }
            }

            nbrInodeDep =  ioDone->nbrInodeDep;

            for (loop1 = 0; loop1 < nbrInodeDep; loop1++) 
            {
                fs_Inode *inode;
                inode = (nbrInodeDep <= VL_DEF_NBR_DEP_INODE) ?
                    ioDone->u_inode.inode[loop1] :
                    ioDone->u_inode.inodeList[loop1];

                if (inode != NULL) 
                {
                    fs_inodeLogRefDecr (context, inode);
                }
            }

            if (nbrInodeDep > VL_DEF_NBR_DEP_INODE) 
            {
                fs_ramDealloc(FS_RAM_LOG_BUF_DEP, ioDone->u_inode.inodeList);
            }
        }

        if (ioDone->state == VL_IO_DONE_STATE_IO) 
        {
            event.event = FS_THREAD_EVENT_LOG_COMPL;
            event.p1.u.p = NULL;
            event.p2.u.p = NULL;
            rc = esm_deliverHighPriorityEvent(ioDone->u.smHandle, &event);

            if (rc != NFX_OK) 
            {
                fs_error(NULL, FS_EVENT_DELIVERY_FAIL, "logCompl event");
            }
            ioDone->state = VL_IO_DONE_STATE_FREE;
        } 
        else 
        if (ioDone->state == VL_IO_DONE_STATE_UNSTABLE_IO) 
        {
            vl_removeUnstableId (context, &ioDone->u.inum);
            ioDone->state = VL_IO_DONE_STATE_FREE;
        }
    }
}

static void
vl_ioDoneDependencies (
    fs_Context *context,
    vl_VolBuf *iobuf)
{
    vl_VolBufRoot   *volBufRoot;
    vl_VolBuf       *buf, *bufNext;
    
    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    // NFX_SPIN_LOCK(&volBufRoot->idSpin);   must be done by caller !!  

    // buffer must be head of list
    FS_ASSERT(iobuf == volBufRoot->hdr.head);
    
    DLL_ITERATE_NEXT(buf, bufNext, &volBufRoot->hdr, entry)
        {
            NFX_SPIN_LOCK(&buf->coreHdr.idSpin);

            if (buf->coreHdr.ioFinished == VL_BUF_IOFINISHED_AND_QUEUED)
            {
                switch (buf->coreHdr.state)
                {
                case VL_BUF_STATE_IO:
                    vl_ioDoneRmBufDependencies (context, buf);
    
                    DLL_REMOVE(&volBufRoot->hdr, buf, entry);
                    NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);
                    vl_deallocLogBuf(FS_RAM_LOG_BUF, buf);
                    volBufRoot->nbr.currNbrBuf--;
                    buf = NULL;
                    vl_IONextBuf (context, volBufRoot);
                    break;
                case VL_BUF_STATE_IO_CURR:
                    vl_ioDoneRmBufDependencies (context, buf);
                    buf->coreHdr.state = VL_BUF_STATE_CURR;                    
                    buf->coreHdr.ioFinished = VL_BUF_IOFINISHED_NOT_SET;
                    if ((volBufRoot->currCommits != buf->coreHdr.currCommits) ||
                        (volBufRoot->head != buf->diskLogBuf->diskHdr.head)) 
                    {
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
                        if ((context->fileSystem->nbrOngoingTrans <
                             fs_incoreTune->activeTransToWait) ||
                            (volBufRoot->forceWrite))
#endif
                        {
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
                            volBufRoot->postponedLogWrite = FALSE;
                            if (volBufRoot->forceWrite > 0)
                                volBufRoot->forceWrite -= 1;
#endif
                            buf->coreHdr.currCommits = volBufRoot->currCommits;
                            buf->coreHdr.state = VL_BUF_STATE_IO_CURR;
                            vl_bufDoIO (context, volBufRoot, buf, TRUE);
                            context->fileSystem->logRoot->logStats.ioPartialLogWrites++;
                            context->fileSystem->logRoot->logStats.ioPartialBytes += buf->coreHdr.size;
                        }
#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
                        else
                        {
                            volBufRoot->postponedLogWrite = TRUE;

                            context->fileSystem->logRoot->logStats.postponedPartialLogWrites++;
                            // wait to write the log, until the next transaction commit
                            // or, better, until ready.
                        }
#endif
                    }
                    break;

                case VL_BUF_STATE_IO_READY:
                    vl_ioDoneRmBufDependencies (context, buf);
                    buf->coreHdr.state = VL_BUF_STATE_IO;
                    buf->coreHdr.ioFinished = VL_BUF_IOFINISHED_NOT_SET;
                    context->fileSystem->logRoot->logStats.ioTotalLogWrites++;
                    context->fileSystem->logRoot->logStats.ioTotalBytes += buf->coreHdr.size;
                    vl_bufDoIO (context, volBufRoot, buf, FALSE);
                    break;

                case VL_BUF_STATE_CURR:
                    // no IO scheduled
                    fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
                    break;
                case VL_BUF_STATE_READY:
                    // no IO scheduled
                    fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
                    break;
                default:
                    fs_error(NULL, FS_LOG_WRITE_ERR, "buf state error");
                    break;
                }
            }
            else
            {
                // io not yet finished for this buf
                FS_ASSERT(buf->coreHdr.ioFinished == VL_BUF_IOFINISHED_NOT_SET);
                NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);

                // stop scanning the queue
                return;
            }

            if (buf)    {
                NFX_SPIN_UNLOCK(&buf->coreHdr.idSpin);
            }
        }
}

/*------------------------------------------------------------------------------
 *
 */
extern int32
fs_vl_ioDone(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context *    context = smCookie;
    fs_FileSystem * fs = context->fileSystem;
    fs_RdwrReq *    rdwr = event.p1.u.p;
    vl_VolBuf *     buf;
    fs_LogFsRoot *  logRoot = fs->logRoot;
    vl_VolBufRoot * volBufRoot = logRoot->volBufRoot;

    NFX_SPIN_LOCK(&context->fileSystem->spin);
    fs->ioPendingCount -= 1;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    
    if (rdwr->evmStat.status != NFX_OK)
    {
        fs_rdwrDealloc(rdwr);
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("fs_vl_ioDone io error returned, aborting"));
        fs_volumeException(context, context->fileSystem, FS_WRITE_ERR);
        return nextState;
    }
    
    if (fs->state == FS_VOL_STATE_FAILING)
    {
        fs_ioDoneError(rdwr);
        return nextState;
    }
    
    if (rdwr->evmReq.reqType == EVM_WRITE_REQ)
    {
        // if copyOfCurrentBuf: mark as free
        buf = (vl_VolBuf *) rdwr->private;
        fs_rdwrDealloc(rdwr);
        
        FS_TRACE(context, FS_TRACE_VOLLOG,
                 ("fs_vl_ioDone %d %p state %d", 
                  buf->diskLogBuf->diskHdr.bufId, buf, buf->coreHdr.state));
        
        NFX_SPIN_LOCK(&volBufRoot->idSpin);
        
        FS_ASSERT (volBufRoot->nbr.currNbrIO > 0);
        volBufRoot->nbr.currNbrIO--;
        
        buf->coreHdr.ioFinished = VL_BUF_IOFINISHED_AND_QUEUED;
        
        // if this buf is not the first buf on the list
        if (buf != volBufRoot->hdr.head)
        {
#if !defined(NFX_SMP)
			// Don't warn for SMP
            // the ioDone is out of sequence
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("fs_vl_ioDone: received bufId %d while waiting for %d", 
                     buf->diskLogBuf->diskHdr.bufId,
                     volBufRoot->hdr.head->diskLogBuf->diskHdr.bufId));
            // FS_ASSERT(FALSE);
#endif			
            
        }
        else
        {
            vl_ioDoneDependencies (context, buf);
        }
        
        NFX_SPIN_UNLOCK(&volBufRoot->idSpin);
    }
    else if (rdwr->evmReq.reqType == EVM_READ_REQ)
    {
        /* Response of a log read ahead during log replay.
         */
        vl_BlockCache * blockCache = &volBufRoot->blockCache;
        fs_vlcEntry *   vlcEntry = (fs_vlcEntry *) rdwr->private;
        FS_ASSERT(vlcEntry != NULL);

        //printf("Received IO for offset %lld nbBuf %d\n",
        //(rdwr->evmReq.offset * EVM_LBLK_SIZE),
        //(uint32) (rdwr->evmReq.reqSize * EVM_LBLK_SIZE / volBufRoot->bufSize));

        if (blockCache->validCache == TRUE)
        {       
            NFX_SPIN_LOCK(&(blockCache->spin));
            
            FS_ASSERT(vlcEntry->volId.whole_thing == context->volId.whole_thing);
            FS_ASSERT(vlcEntry->valid == FALSE);
            FS_ASSERT(vlcEntry->reading == TRUE);
            FS_ASSERT(vlcEntry->ultrabuf == NULL);
            FS_ASSERT(vlcEntry->offset == (rdwr->evmReq.offset * EVM_LBLK_SIZE));
            FS_ASSERT(vlcEntry->nbBuf == (uint32) (rdwr->evmReq.reqSize *
                                                   EVM_LBLK_SIZE /
                                                   volBufRoot->bufSize));

            vlcEntry->valid = TRUE;
            vlcEntry->reading = FALSE;
            vlcEntry->ultrabuf = rdwr->evmStat.rDataHandle.dataPtr;
            
            /* Send event to deblock the guy waiting (if any).
             */
            if (vlcEntry->guyWaiting != 0)
            {
                esm_event_t event;
                int32       rc = NFX_OK;
                
                //printf("Send IO for offset %lld to waiting guy %d (cidx %d)\n",
                //blockCache->cache[cidx].offset,
                //blockCache->cache[cidx].guyWaiting, cidx);
                event.event = FS_THREAD_EVENT_LOG_COMPL;
                event.p1.u.p = NULL;
                event.p2.u.p = NULL;
                rc = thread_highPriorityEvent(vlcEntry->guyWaiting, &event);
                FS_ASSERT(rc == NFX_OK);
            }
            
            NFX_SPIN_UNLOCK(&(blockCache->spin));
        }
        else
        {
            printf("Received IO for offset %lld nbBuf %d\n",
                   (rdwr->evmReq.offset * EVM_LBLK_SIZE),
                   (uint32) (rdwr->evmReq.reqSize * EVM_LBLK_SIZE /
                             volBufRoot->bufSize));
            printf("Cache is not valid anymore. Delete received IO response.\n");
            fs_ramDealloc(FS_RAM_READ_DATA, rdwr->evmStat.rDataHandle.dataPtr);
        }
        
        fs_rdwrDealloc(rdwr);
    }
    else
    {
        // cannot be something else than read or write!
        FS_ASSERT(FALSE);
    }

    return nextState;

} /* fs_vl_ioDone */


int32
fs_vl_fsBufLock(
    esm_handle_t handle,
    int32        nextState,
    esm_event_t  event,
    void *       smCookie)
{
    fs_Context  *context = smCookie;
    fs_Buf      *buf = event.p2.u.p;

    //printf("fs_vl_fsBufLock()\n");

    fs_bufConvertCando(context, buf);
    fs_bufReleaseExclusive(context, buf);

    return nextState;

} /* fs_vl_fsBufLock */


extern int32
fs_vl_stop(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    fs_Context      *context = smCookie;
    vl_VolBufRoot   *volBufRoot;
    fs_LogFsRoot    *logRoot;
    vl_VolBuf       *buf, *bufNext;
    int32 rc;

    context = smCookie;

    FS_TRACE(context, FS_TRACE_VOLLOG, ("fs_vl_stop"));

    logRoot = context->fileSystem->logRoot;

    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    volBufRoot = logRoot->volBufRoot;

    if (volBufRoot->copyOfCurrentBuf != NULL) {

        fs_ramDealloc (FS_RAM_WRITE_DATA, volBufRoot->copyOfCurrentBuf);
        volBufRoot->copyOfCurrentBuf = NULL;

    }

    vl_clearBlockCache(context);

    NFX_SPIN_LOCK(&volBufRoot->idSpin);

    if (volBufRoot->ioWrHandle)
    {
        FS_ASSERT(handle == volBufRoot->ioWrHandle);

        rc = fs_vl_sm_destroy(volBufRoot->ioWrHandle);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_ESM_DESTROY_ERR, "wb esm ioWrHandle");
        }
    }

    if (volBufRoot->ioWrContext)
    {
        fs_lockContextFree(volBufRoot->ioWrContext);
        fs_ramDealloc(FS_RAM_CORE_MISC, volBufRoot->ioWrContext);
    }
    if (volBufRoot->ioRdContext)
    {       
        fs_lockContextFree(volBufRoot->ioRdContext);
        fs_ramDealloc(FS_RAM_CORE_MISC, volBufRoot->ioRdContext);
    }
    
    DLL_ITERATE_NEXT(buf, bufNext, &volBufRoot->hdr, entry)
    {
        DLL_REMOVE(&volBufRoot->hdr, buf, entry);

        vl_freeLogBuf(FS_RAM_LOG_BUF, buf);
        volBufRoot->nbr.currNbrBuf--;
    }
    
    chash_destroy (volBufRoot->fileHandleHid);
    volBufRoot->fileHandleHid = -1;

    logRoot->volBufRoot = NULL;

    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

    fs_ramDealloc(FS_RAM_CORE_MISC, volBufRoot);

    return nextState;
}

extern int32
fs_vl_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *smCookie)
{
    fs_Context      *context;

    context = smCookie;

    FS_TRACE(context, FS_TRACE_VOLLOG, ("fs_vl_bad_state"));

    FS_ASSERT(FALSE);

    return nextState;

} /* fs_vl_bad_state */

extern void
vl_startIODaemon(
    fs_Context *context)
{
    vl_VolBufRoot *volBufRoot;
    fs_LogFsRoot  *logRoot;
    int32 rc;
        
    FS_TRACE(context, FS_TRACE_VOLLOG, ("vl_startIODaemon"));

    logRoot = context->fileSystem->logRoot;

    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    volBufRoot = logRoot->volBufRoot;

    volBufRoot->ioWrContext = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof (fs_Context));
    bzero(volBufRoot->ioWrContext, sizeof(fs_Context));

    //Lock lists init
    fs_lockContextInit(volBufRoot->ioWrContext, SHARED_LOCK_LIST_INIT);
    volBufRoot->ioWrContext->flags |= FS_CONTEXT_ALLOW_LOCK_STARVATION;

    if (logRoot->logDevType == FS_LOG_DEVTYPE_SELF)
    {
        volBufRoot->ioWrContext->volId = context->volId;
    }
    else
    {
        volBufRoot->ioWrContext->volId = context->fileSystem->super->logVolId;
    }
    volBufRoot->ioWrContext->fileSystem = context->fileSystem;
    volBufRoot->ioWrContext->daemonType = FS_DAEMON_TYPE_VOLLOG;
    volBufRoot->nbr.currNbrIO = 0 ;
    
    rc = fs_vl_sm_create(&volBufRoot->ioWrHandle, volBufRoot->ioWrContext);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_ESM_CREATE_ERR, "vl io esm");
    }
    
    volBufRoot->ioWrContext->smHandle = volBufRoot->ioWrHandle;

    if (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID)
    {
        // Need to create another context for reading the external volume !!!
        volBufRoot->ioRdContext = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof (fs_Context));
        bzero(volBufRoot->ioRdContext, sizeof(fs_Context));
        
        //Lock lists init
        fs_lockContextInit(volBufRoot->ioRdContext, SHARED_LOCK_LIST_INIT);
    
        volBufRoot->ioRdContext->volId = context->fileSystem->super->logVolId;
        volBufRoot->ioRdContext->fileSystem = context->fileSystem;
    
        volBufRoot->ioRdContext->smHandle = context->smHandle;
    }
}

extern void 
vl_stopIODaemon(
    fs_Context *context,
    boolean isForce)
{
    fs_LogFsRoot    *logRoot;
    vl_VolBufRoot   *volBufRoot;
    int32           numWaits = 0;

    FS_TRACE(context, FS_TRACE_VOLLOG, ("vl_stopIODaemon"));

    logRoot = context->fileSystem->logRoot;
    volBufRoot = logRoot->volBufRoot;

    FS_ASSERT ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
               (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID));

    if (volBufRoot &&
        volBufRoot->ioWrHandle)
    {
        esm_event_t event;
        int32       rc;
        
        event.event = FS_VL_EVENT_STOP;
	event.p1.u.p = NULL;
	event.p2.u.p = NULL;
        rc = esm_deliverEvent(volBufRoot->ioWrHandle, &event);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EEE_ERR, "vl_stopIODaemon deliver event failed");
        }
    }

// about 60 seconds
#define MAX_WAITS   50
#define WAIT_MSECS  50
    while (numWaits++ < MAX_WAITS)
    {
        if (logRoot->volBufRoot != NULL)
        {
            if (numWaits > MAX_WAITS / 2)
            {
                FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                        ("unmount, still waiting for logging daemon to terminate"));
            }
            if (isForce)
            {
                // We're in exception condition already,
                // don't call fs_sleep, since it will
                // call the exception handler.
                thread_sleep(NULL, numWaits * WAIT_MSECS);
            }
            else
            {
                fs_sleep(context, numWaits * WAIT_MSECS);
            }
        }
        else
        {
            break;
        }
    }

    if (numWaits >= MAX_WAITS)
    {
        if (isForce)
        {
            fs_error(context, FS_INTERNAL_ERR, "vl_stopIODaemon failed");
        }
        else
        {
            fs_volumeException(context, context->fileSystem,
                               FS_VOLUME_EXCEPTION);
        }
    }
#undef MAX_WAITS
#undef WAIT_MSECS

}

/*----------------------------------------------------------------
 *
 * Vol logging need to lock the read cache while reading the volume
 * During normal operation, the volume is only written, and the cache
 * is invalid.
 * During replay (wyith the lock on), no writing is done. 
 * Only during debug, while reading the log data, the lock will
 * temporary block writing. 
 *
 */

extern int32
fs_logVolCommitWrite(
    fs_Context          *context,
    fs_Inode            *inode,
    fs_WriteStableHow   *stableHow)
{
    int32   rc;

    if (*stableHow == FS_WRITE_STABLE_UNSTABLE)
    {
        // Record that this inode has an unstable log transaction
        // associated with it (the currently commiting transaction).
        // and write the commit record to the log without blocking.
        vl_addUnstableId(context, &inode->inum);
        rc = fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, inode->inum);
    }
    else
    {
        // Start I/O for all dirty user bufs for this inode
        // and block until they're completed.
        fs_bufSyncInodeBufs(context, inode);

        // Write the commit log entry and block until its
        // completed.
        rc = fs_logCommit(context, FS_LOG_WRITE_COMMIT_STABLE, 0);
        rc = fs_logVolCommitTrans(context);
    }

    return rc;
}


extern int32
fs_logVolCommitTrans(
    fs_Context          *context)
{
    vl_VolBufRoot   *volBufRoot;
    
    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    return (vl_waitLogIsFlushed(context, volBufRoot));
}

extern int32
fs_logVolSyncInodeBuf(
    fs_Context  *context,
    fs_Inode    *inode)
{
    // Schedule I/O for dirty user bufs and wait for
    // them to complete (up to 60 seconds).

    if (!fs_bufSyncInodeBufsAndTimeout(context, inode, 60))
    {
        return FS_ERR_NO_RESOURCE;
    }

    return NFX_OK;
}

extern int32
fs_logVolSyncInodeLog(
    fs_Context  *context,
    uint64      inum)
{
    // Wait for any transactions affecting this inode
    // to be stable in the log.
    vl_waitInodeStable(context, inum);

    return NFX_OK;
}

/**
 * When we read from volume, we read in chunks of VL_BUFFER_SIZE,
 * and one log record can span 2+ blocks.
 * To avoid copy, we read a small size, and gets the ptr to the
 * the data (no copy needed). Only in the limited cases when a log
 * record spans 2+ blocks. we copy to a temporary buffer.
 * The optimal min size is at least the record header + the avg data size.
 */
#define  VOL_MIN_TO_READ   (sizeof(fs_LogRecord) + 16)

extern int32
fs_logVolReadInit(
    fs_Context  *context,
    fs_LogInfo  *info)
{
    vl_BufDiskHdr   *vlBuf;
    int32           bufSize;

    if (fs_logReadVolLogHdr(context, context->fileSystem->logRoot,
                            &info->logSize, &bufSize) != NFX_OK)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("log vol: bad device header"));
        return NFX_ERR;
    }
    vlBuf = vl_findMostRecentBuf(context);
    if (vlBuf == NULL)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("log vol: cannot find most recent buf"));
        return NFX_ERR;
    }

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    context->fileSystem->super->blksUsed = vlBuf->blksUsed;
    context->fileSystem->super->inodesUsed = vlBuf->inodesUsed;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    NFX_SPIN_LOCK(&context->fileSystem->spin);
    context->fileSystem->commitedBlksUsed = vlBuf->blksUsed;
    context->fileSystem->commitedInodesUsed = vlBuf->inodesUsed;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);

    info->nbrCommits = 0;  // unknown
    info->head = vlBuf->head;
    info->tail = vlBuf->tail;
    info->readChunkSize = VOL_MIN_TO_READ;

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("read init: logSize %d, nbrCommits %d, head %d, tail %d",
             info->logSize, info->nbrCommits, info->head, info->tail));
    return NFX_OK;
}

extern int32
fs_logVolRead(
    fs_Context  *context,
    uchar8      **datap,
    uint32      offset,
    uint32      *bytesRead,
    uint32      tail)
{
    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("log vol read: offset %u bytesRead %u (tail 0x%x)",
             offset, *bytesRead, tail));

    if (*bytesRead == 0)
    {
        FS_ELOG(context, LOG_WARNING, FS_TRACE_LOG,
                ("requested 0 bytes in logVolRead"));
	return 0;
    }

    // adjust for circular buffer wrap around

    if (tail >= offset)
        *bytesRead = MIN(*bytesRead, tail - offset);
    else
        *bytesRead = MIN(*bytesRead, tail + context->fileSystem->logSize - offset);

    if (*bytesRead != 0)
    {
        *datap = fs_logReadVol(context, context->fileSystem->logRoot,
                               offset, bytesRead);
    }
    else
    {
        *datap = NULL;
    }
    
    if (*datap == NULL)
        *bytesRead = 0;     // indicate eol (end of log)

    FS_TRACE(context, FS_TRACE_VOLLOG,
             ("  rc bytesRead %d", *bytesRead));

    //if (*bytesRead == 0)
    //{
    //printf("datap %p tail %d offset %d logsize %d\n",
    //*datap, tail, offset, context->fileSystem->logSize);
    //}

    return NFX_OK;
}

extern void
fs_logVolReadDone(
    fs_Context  *context,
    uchar8      *datap)
{
    vl_clearBlockCache(context);
    return;
}

extern int32
fs_logVolIsEmpty(
    fs_Context  *context)
{
    vl_VolBufRoot   *volBufRoot;
    vl_VolBuf       *buf;
    int32           isEmpty = FALSE;

    if (context->fileSystem->logRoot->currLen != 0)
    {
        return FALSE;
    }

    FS_ASSERT(context->fileSystem->logRoot->tail == context->fileSystem->logRoot->head);
    
    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    NFX_SPIN_LOCK(&volBufRoot->idSpin);

    if (DLL_IS_EMPTY(&volBufRoot->hdr))
    {
        FS_ASSERT (volBufRoot->nbr.currNbrIO == 0);
        isEmpty = TRUE;
    }
    else
    {
        buf = volBufRoot->hdr.head;

        if ((buf->coreHdr.state == VL_BUF_STATE_CURR) &&
            (buf->coreHdr.ioFinished == VL_BUF_IOFINISHED_NOT_SET))
        {
            if ((volBufRoot->currCommits == buf->coreHdr.currCommits) &&
                (volBufRoot->head == buf->diskLogBuf->diskHdr.head))
            {
                FS_ASSERT (volBufRoot->nbr.currNbrIO == 0);
                isEmpty = TRUE;
            }
        }
        else
        {
            // io must be ongoing, or is signalled (vl_wakeupIO) to start
            FS_ASSERT (volBufRoot->nbr.currNbrIO != 0);
        }
    }
    
    NFX_SPIN_UNLOCK(&volBufRoot->idSpin);

    if (! isEmpty)
    {
        fs_signalVolLogWrite(context);
    }
    
    return (isEmpty);
}


extern void
fs_vl_showStats (
    fs_Context      *context)
{
    vl_VolBufRoot   *volBufRoot;
#ifdef FS_DEBUG_VOL_LOG_BUCKETS
    int32           loop1;
#endif

    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    if (volBufRoot)
    {
        NFX_SPIN_LOCK(&volBufRoot->idSpin);
        printf ("Current number of volume log buffers in use %d (max %d)\n",
                volBufRoot->nbr.currNbrBuf, volBufRoot->nbr.maxNbrBuf);
        printf ("Current number of volume log buffers in i/o %d (max %d)\n",
                volBufRoot->nbr.currNbrIO, volBufRoot->nbr.maxNbrIO);
        printf ("Total number of volume log buffers written  %d\n",
                volBufRoot->nbr.totalNbrIO);

#ifdef FS_DEBUG_VOL_LOG_BUCKETS
        printf ("Commit dependency buckets  ");
        printf ("\nBetween   ");
        printf ("       0  ");
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS - 1 ; loop1++)
        {
            printf ("%8d  ", fs_vl_dep_bucket[loop1]);
        }
        printf ("\nAnd (<)   ");
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS - 1 ; loop1++)
        {
            printf ("%8d  ", fs_vl_dep_bucket[loop1]);
        }
        printf ("     MAX  = value\n");

        printf ("Buf       ");
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS ; loop1++)
        {
            printf ("%8d  ", volBufRoot->nbr.nbrDepBuf[loop1]);
        }
        printf ("= %d \n", volBufRoot->nbr.maxDepBuf);
        printf ("Inode     ");
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS ; loop1++)
        {
            printf ("%8d  ", volBufRoot->nbr.nbrDepInode[loop1]);
        }
        printf ("= %d \n", volBufRoot->nbr.maxDepInode);
        printf ("Undo      ");
        for (loop1 = 0 ; loop1 < FS_VL_STATS_NBR_DEP_BUCKETS ; loop1++)
        {
            printf ("%8d  ", volBufRoot->nbr.nbrDepUndo[loop1]);
        }
        printf ("= %d \n", volBufRoot->nbr.maxDepUndo);
#endif // FS_DEBUG_VOL_LOG_BUCKETS

        NFX_SPIN_UNLOCK(&volBufRoot->idSpin);
    }
}

extern void
fs_vl_getStats (
    fs_Context      *context,
    boolean         isClear,
    fs_vl_stats     *msg)
{
    vl_VolBufRoot   *volBufRoot;

    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    if (volBufRoot)
    {
        NFX_SPIN_LOCK(&volBufRoot->idSpin);
        msg->currNbrBuf = volBufRoot->nbr.currNbrBuf;
        msg->currNbrIO = volBufRoot->nbr.currNbrIO;
        msg->maxNbrBuf = volBufRoot->nbr.maxNbrBuf;
        msg->maxNbrIO = volBufRoot->nbr.maxNbrIO;
        msg->totalNbrIO = volBufRoot->nbr.totalNbrIO;

        if (isClear)
        {
            volBufRoot->nbr.maxNbrBuf = 0;
            volBufRoot->nbr.maxNbrIO = 0;
            volBufRoot->nbr.totalNbrIO = 0;
        }
        
        NFX_SPIN_UNLOCK(&volBufRoot->idSpin);
    }
}


/*
 * wakeup the writeback of log buffers, if needed.
 */

extern void
fs_signalVolLogWrite (
    fs_Context      *context)
{
    vl_VolBufRoot   *volBufRoot;

    volBufRoot = context->fileSystem->logRoot->volBufRoot;

    if (volBufRoot)
    {
        NFX_SPIN_LOCK(&volBufRoot->idSpin);

#ifdef FS_VOLLOG_WAIT_FOR_ONGOING_TRANS
        volBufRoot->forceWrite += 1;
#endif
        vl_wakeupIO (context);

        NFX_SPIN_UNLOCK(&volBufRoot->idSpin);
    }
    
}

/* Initialize the cache.
 */
extern void
fs_logVolInitCache(vl_BlockCache * blockCache)
    
{
    //printf("fs_logVolInitCache()\n");

    // done in vl_initVolBufRoot()
    //NFX_SPIN_INIT(&(blockCache->spin),SPINLOCK_LEVEL_UNKNOWN);
    blockCache->validCache = TRUE;
    blockCache->replay = FALSE;
    blockCache->entrySize = VLC_LOGBUF_SIZE;
    DLL_INIT(&(blockCache->hdr));
    
    return;
}

/*------------------------------------------------------------------------------
 * Trigger the asynchronous read of the complete log of a filesystem.
 */
extern void
fs_logVolReadCompleteLog(fs_Context *   context,
                         fs_LogInfo *   logInfo)
{
    fs_FileSystem * fs = context->fileSystem;
    vl_VolBufRoot * volBufRoot = fs->logRoot->volBufRoot;
    boolean         endOfLog = FALSE;
    uint32          nbReadReq = 0;
    evm_volOffset_t logOffset, headOffset ,tailOffset, stopOffset;
    evm_volOffset_t startReadingFromHere;

    /* Check first if there is something to read!
     */
    if (logInfo->head == logInfo->tail)
        return;
          
    /* Convert log offset to volume offset.
     */
    FS_LOG2VOLOFFSET(logInfo->head, headOffset);
    FS_LOG2VOLOFFSET(logInfo->tail, tailOffset);
    if (headOffset <= tailOffset)
    {
        stopOffset = tailOffset;
    }
    else
    {
        FS_LOG2VOLOFFSET(logInfo->logSize, stopOffset);
    }
    //FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
    //("start %lld head %lld tailOffset %lld stopOffset %lld",
    //volBufRoot->startOffset, headOffset, tailOffset, stopOffset));

    /* Start reading at a 8k boundary so a page fits entirely into an ultrabuf.
     */
    startReadingFromHere = logOffset =
        (headOffset - (headOffset % VLC_BLOCK_SIZE));

    /* Read the whole log.
     */
    while (!endOfLog)
    {
        boolean         contiguous = TRUE;
        evm_volOffset_t readOffset = logOffset;
        uint32          cnt = 0;

        /* Find out how many read ahead I can do.
         * Take care of the warp around (=> contiguous).
         */
        while ((cnt < VL_NB_READ_BUF) && (contiguous == TRUE))
        {
            if (logOffset <= tailOffset)
            {
                cnt++;
                logOffset += volBufRoot->bufSize;
            }
            else if ((headOffset > tailOffset) &&
                     (logOffset <= stopOffset) &&
                     (logOffset >= startReadingFromHere))
            {
                cnt++;
                if (logOffset < stopOffset)
                {
                    logOffset += volBufRoot->bufSize;
                }
                else
                {
                    logOffset = volBufRoot->startOffset;
                    contiguous = FALSE;
                }
            }
            else
            {
                //printf("No more stuff to read ahead.\n");
                endOfLog = TRUE;
                break;
            }
        } // while
        
        if (cnt > 0)
        {
            /* Read ahead.
             * To avoid "eating" all the resources,
             * do a synchronous read if too many read ahead have been triggered,
             * so the asynchronous responses have the time to be back.
             */
            nbReadReq++;
            if ((fs_ioPending > FS_RDWR_MAX/2) || (nbReadReq % 1024 == 0))
            {
                fs_RdwrReq *    rdwr = fs_rdwrAlloc();
                fs_vlcEntry *   vlcEntry = NULL;
    
                fs_volRead(context, rdwr,
                           readOffset, (cnt * volBufRoot->bufSize));

                vlcEntry = fs_vlcAddEntry(context, readOffset, cnt);
                FS_ASSERT(vlcEntry != NULL);
                vlcEntry->valid = TRUE;
                vlcEntry->reading = FALSE;
                vlcEntry->ultrabuf = rdwr->evmStat.rDataHandle.dataPtr;
                
                fs_rdwrDealloc(rdwr);
            }
            else
            {
                fs_logVolReadAhead(context, volBufRoot, readOffset, cnt);
            }
        }
    } // while
    
    //FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
    //("replay: issued %d log read requests (fs_logVolReadCompleteLog)",
    //nbReadReq));
    return;
} // fs_logVolReadCompleteLog

/*------------------------------------------------------------------------------
 * Volume Log Cache
 *------------------------------------------------------------------------------
 */
#define VLC_TRACE(_par_) //printf _par_; printf("\n")

void
fs_vlcInit(void)
{
    VLC_TRACE(("fs_vlcInit()"));

    NFX_SPIN_INIT(&(vlcCache.spin), SPINLOCK_LEVEL_UNKNOWN);
    vlcCache.hashTab = chash_create(VLC_HASH_BUCKETS,
                                    fs_vlcHash,
                                    fs_vlcCompare,
                                    fs_vlcRelease);
    vlcCache.size = VLC_MAX_SIZE;
    vlcCache.used = 0;
    DLL_INIT(&(vlcCache.pendingMountList));

    return;
}

uint32
fs_vlcHash(const void * key)
    
{
    fs_vlcEntry * entry = (fs_vlcEntry *) key;

    return((uint32) (((entry->volId.details.u.parts.clusterUnique + entry->page) >> VLC_HASH_SHIFT) + entry->volId.details.u.parts.clusterUnique + entry->page) % VLC_HASH_BUCKETS);
}

int32
fs_vlcCompare(const void * key1,
              const void * key2)
{
    fs_vlcEntry * entry1 = (fs_vlcEntry *) key1;
    fs_vlcEntry * entry2 = (fs_vlcEntry *) key2;

    if ((entry1->volId.whole_thing == entry2->volId.whole_thing) &&
        (entry1->page == entry2->page))
        return(0);
    else
        return(1);
}

void
fs_vlcRelease(void * key,
              void * obj)
{
    return; // do nothing
}

fs_vlcEntry *
fs_vlcGetEntry(fs_Context *     context,
               evm_volOffset_t  offset)
{
    vl_VolBufRoot * volBufRoot = context->fileSystem->logRoot->volBufRoot;
    vl_BlockCache * blockCache = &(volBufRoot->blockCache);
    fs_vlcEntry     searchEntry;
    searchEntry.volId = context->volId;
    searchEntry.page = (offset - volBufRoot->startOffset) / blockCache->entrySize;
    
    return((fs_vlcEntry *) chash_find(vlcCache.hashTab, (void *) &searchEntry));
}

fs_vlcEntry *
fs_vlcAddEntry(fs_Context *     context,
               evm_volOffset_t  offset,
               uint32           nbBuf)
{
    int32           rc;
    vl_VolBufRoot * volBufRoot = context->fileSystem->logRoot->volBufRoot;
    vl_BlockCache * blockCache = &(volBufRoot->blockCache);
    fs_vlcEntry *   newEntry = fs_ramAlloc(FS_RAM_VLC, sizeof(fs_vlcEntry));
    FS_ASSERT(newEntry != NULL);

    VLC_TRACE(("fs_vlcAddEntry(%lld)", offset));
    
    newEntry->volId = context->volId;
    newEntry->page = (offset - volBufRoot->startOffset) / blockCache->entrySize;
    newEntry->valid = FALSE;
    newEntry->reading = TRUE;
    newEntry->nbBuf = nbBuf;
    newEntry->ultrabuf = NULL;
    newEntry->offset = offset;
    newEntry->guyWaiting = 0;
    rc = chash_insert(vlcCache.hashTab, (void *) newEntry, (void *) newEntry);
    FS_ASSERT(rc == NFX_OK);
    DLL_ISOLATE_ELEM(newEntry, entry);
    DLL_INSERT_TAIL(&(blockCache->hdr), newEntry, entry);
    
    return(newEntry);
}

void
fs_vlcDelEntry(fs_Context *     context,
               evm_volOffset_t  offset)
{
    int32           rc;
    vl_VolBufRoot * volBufRoot = context->fileSystem->logRoot->volBufRoot;
    vl_BlockCache * blockCache = &(volBufRoot->blockCache);
    fs_vlcEntry *   vlcEntry = fs_vlcGetEntry(context, offset);
    FS_ASSERT(vlcEntry != NULL);
    
    VLC_TRACE(("fs_vlcDelEntry(%lld)", offset));
    
    rc = chash_remove(vlcCache.hashTab, (void *) vlcEntry);
    FS_ASSERT(rc == NFX_OK);
    DLL_REMOVE(&(blockCache->hdr), vlcEntry, entry);
    fs_ramDealloc(FS_RAM_VLC, vlcEntry);
    
    return;
}

/*----------------------------------------------------------------------------*/

#endif // SSC

