/*------------------------------------------------------------------------------
 * RCS:  $Id: fs-zeroFill.c,v 1.10 2002/12/05 17:23:47 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fs-zeroFill.c
 *
 * Description  : FileSystem (FS)
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-OCT-2002
 *
 *------------------------------------------------------------------------------
 */

#include "fs-includes.h"
#include "fs-log-private.h"

#ifdef NEW_ZF

uint32 fs_zfInodeHashTab;
uint32 fs_zfPageHashTab;

#define ZF_TRACE(_par_) //printf _par_; printf("\n")

/*------------------------------------------------------------------------------
 *
 *------------------------------------------------------------------------------
 */

void
fs_zfInit(void)
{
    ZF_TRACE(("fs_zfInit()"));
    
    fs_zfInodeHashTab = chash_create(FS_ZF_INUM_BUCKETS,
                                     fs_zfInumHash,
                                     fs_zfInumCompare,
                                     fs_zfInumRelease);
    fs_zfPageHashTab = chash_create(FS_ZF_PAGE_BUCKETS,
                                    fs_zfPageHash,
                                    fs_zfPageCompare,
                                    fs_zfPageRelease);
    return;
}

uint32
fs_zfInumHash(const void * key)
    
{
    fs_zfInode * inode = (fs_zfInode *) key;
    return(fs_bufHashVolInum(inode->volId, inode->inum));
}

int32
fs_zfInumCompare(const void * key1,
                 const void * key2)
{
    fs_zfInode * inode1 = (fs_zfInode *) key1;
    fs_zfInode * inode2 = (fs_zfInode *) key2;

    if ((inode1->volId.whole_thing == inode2->volId.whole_thing) &&
        (inode1->inum == inode2->inum))
        return(0);
    else
        return(1);
}

void
fs_zfInumRelease(void * key,
                 void * obj)
{
    return; // do nothing
}

uint32
fs_zfPageHash(const void * key)
    
{
    fs_zfPage * page = (fs_zfPage *) key;
    return(fs_bufHashVolInumPage(page->volId, page->inum, page->page));
}

int32
fs_zfPageCompare(const void * key1,
                 const void * key2)
{
    fs_zfPage * page1 = (fs_zfPage *) key1;
    fs_zfPage * page2 = (fs_zfPage *) key2;

    if ((page1->volId.whole_thing == page2->volId.whole_thing) &&
        (page1->inum == page2->inum) && (page1->page == page2->page))
        return(0);
    else
        return(1);
}

void
fs_zfPageRelease(void * key,
                 void * obj)
{
    return; // do nothing
}

/*------------------------------------------------------------------------------
 *
 *------------------------------------------------------------------------------
 */

fs_zfInode *
fs_zfGetInode(fs_Context *  context,
              uint64        inum)
{
    fs_zfInode  searchInode;
    searchInode.volId = context->volId;
    searchInode.inum = inum;
    
    return((fs_zfInode *) chash_find(fs_zfInodeHashTab, (void *) &searchInode));
}

fs_zfPage *
fs_zfGetPage(fs_Context *   context,
             uint64         inum,
             uint64         page)
{
    fs_zfPage   searchPage;
    searchPage.volId = context->volId;
    searchPage.inum = inum;
    searchPage.page = page;
    
    return((fs_zfPage *) chash_find(fs_zfPageHashTab, (void *) &searchPage));
}

fs_zfInode *
fs_zfAddInode(fs_Context *  context,
              uint64        inum)
{
    int32 rc;
    fs_zfInode * newInode = fs_ramAlloc(FS_RAM_ZF, sizeof(fs_zfInode));
    FS_ASSERT(newInode != NULL);

    ZF_TRACE(("fs_zfAddInode(%lld)", inum));
    
    newInode->volId = context->volId;
    newInode->inum = inum;
    DLL_INIT(&(newInode->pageList));
    DLL_ISOLATE_ELEM(newInode, entry);
    DLL_INSERT_TAIL(&(context->fileSystem->zfInodeHdr), newInode, entry);
    rc = chash_insert(fs_zfInodeHashTab, (void *) newInode, (void *) newInode);
    FS_ASSERT(rc == NFX_OK);
    
    return(newInode);
}

void
fs_zfAddPage(fs_Context *       context,
             uint64             inum,
             uint64             startPage,
             uint32             offsetInFirstPage,
             uint32             nbPage,
             fs_LogRecord *     record,
             fs_LogObjectSumm * obj)    // offset+length should be prefilled
{
    uint32          i;
    fs_zfInode *    zfInode = fs_zfGetInode(context, inum);

    ZF_TRACE(("fs_zfAddPage(%lld, %lld, %d)", inum, startPage, nbPage));
    ZF_TRACE(("obj = {offset = %d, length = %d}", obj->offset, obj->length));
    
    if (zfInode == NULL)
        zfInode = fs_zfAddInode(context, inum);

    obj->objectType = FS_LOG_DEP_UZERO;

    for (i = 0; i < nbPage; i++)
    {
        int32 rc;
        fs_zfPage * newPage = fs_zfGetPage(context, inum, startPage + i);

        if (newPage == NULL)
        {
            fs_zfPage * newPage = fs_ramAlloc(FS_RAM_ZF, sizeof(fs_zfPage));
            FS_ASSERT(newPage != NULL);
            
            bzero((void *) newPage, sizeof(fs_zfPage)); // need for logSegDep
            obj->objectPtr = newPage;
            fs_logAddDep(context, &record->hdr, obj, FALSE);
            ZF_TRACE(("fs_logAddDep(%llx, %d)",
                      newPage->logSegDep.u.dword, newPage->logSegDep.count));
            newPage->volId = context->volId;
            newPage->inum = inum;
            newPage->page = startPage + i;
            if (i == 0)
                newPage->offset = offsetInFirstPage;
            else
                newPage->offset = 0;
            DLL_ISOLATE_ELEM(newPage, entry);
            DLL_INSERT_TAIL(&(zfInode->pageList), newPage, entry);
            rc = chash_insert(fs_zfPageHashTab,
                              (void *) newPage, (void *) newPage);
            FS_ASSERT(rc == NFX_OK);
        }
        else
        {
            FS_ASSERT(i == 0);
            newPage->offset = MIN(newPage->offset, offsetInFirstPage);
            obj->objectPtr = newPage;
            fs_logAddDep(context, &record->hdr, obj, FALSE);
            ZF_TRACE(("fs_logAddDep(%llx, %d)",
                      newPage->logSegDep.u.dword, newPage->logSegDep.count));
        }
    } // for
    
    return;
}

void
fs_zfDelInode(fs_Context *  context,
              uint64        inum)
{
    int32 rc;
    fs_zfInode * zfInode = fs_zfGetInode(context, inum);
    FS_ASSERT(zfInode != NULL);
    FS_ASSERT(DLL_IS_EMPTY(&(zfInode->pageList)));
    
    ZF_TRACE(("fs_zfDelInode(%lld)", inum));
    
    DLL_REMOVE(&(context->fileSystem->zfInodeHdr), zfInode, entry);
    rc = chash_remove(fs_zfInodeHashTab, (void *) zfInode);
    FS_ASSERT(rc == NFX_OK);
    fs_ramDealloc(FS_RAM_ZF, zfInode);
    
    return;
}

void
fs_zfDelPage(fs_Context *   context,
             uint64         inum,
             uint64         startPage,
             boolean        tillEnd)
{
    fs_zfInode * zfInode = fs_zfGetInode(context, inum);

    ZF_TRACE(("fs_zfDelPage(%lld, %lld, %s)", inum, startPage,
             tillEnd? "till end" : "only this page"));
    
    if (zfInode != NULL)
    {
        int32 rc;
        
        if (tillEnd == FALSE)
        {
            fs_zfPage * zfPage = fs_zfGetPage(context, inum, startPage);
            if (zfPage != NULL)
            {
                ZF_TRACE(("fs_logRemoveDeps(%llx, %d)",
                          zfPage->logSegDep.u.dword, zfPage->logSegDep.count));
                fs_logRemoveDeps(context, &(zfPage->logSegDep), NULL);
                DLL_REMOVE(&(zfInode->pageList), zfPage, entry);
                rc = chash_remove(fs_zfPageHashTab, (void *) zfPage);
                FS_ASSERT(rc == NFX_OK);
                fs_ramDealloc(FS_RAM_ZF, zfPage);
            }
        }
        else // tillEnd == TRUE
        {
            fs_zfPage * currPage = DLL_NIL;
            fs_zfPage * nextPage = DLL_NIL;
            
            DLL_ITERATE_NEXT(currPage, nextPage, &(zfInode->pageList), entry)
            {
                if (currPage->page >= startPage)
                {
                    ZF_TRACE(("fs_logRemoveDeps(%llx, %d)",
                              currPage->logSegDep.u.dword,
                              currPage->logSegDep.count));
                    fs_logRemoveDeps(context, &(currPage->logSegDep), NULL);
                    DLL_REMOVE(&(zfInode->pageList), currPage, entry);
                    rc = chash_remove(fs_zfPageHashTab, (void *) currPage);
                    FS_ASSERT(rc == NFX_OK);
                    fs_ramDealloc(FS_RAM_ZF, currPage);
                }
            } // DLL_ITERATE_NEXT
        }
        
        if (DLL_IS_EMPTY(&(zfInode->pageList)))
        {
            DLL_REMOVE(&(context->fileSystem->zfInodeHdr), zfInode, entry);
            rc = chash_remove(fs_zfInodeHashTab, (void *) zfInode);
            FS_ASSERT(rc == NFX_OK);
            fs_ramDealloc(FS_RAM_ZF, zfInode);
        }
    }
    
    return;
}

void
fs_zfDisplayList(fs_Context *   context)

{
    fs_FileSystem * fs = context->fileSystem;
    fs_zfInode *    currInode;
    
    if (!DLL_IS_EMPTY(&(fs->zfInodeHdr)))
    {
        printf("%s: list of user data page to zero:\n", fs->super->mountName);
        DLL_ITERATE(currInode, &(fs->zfInodeHdr), entry)
        {
            fs_zfPage * currPage;
            printf("inum %lld ", currInode->inum);
            DLL_ITERATE(currPage, &(currInode->pageList), entry)
            {
                printf("- %lld ", currPage->page);
                if (currPage->offset != 0)
                    printf("(%d) ", currPage->offset);
            }
            printf("\n");
        } // DLL_ITERATE
    }
    else
    {
        printf("%s: list of user data page to zero: EMPTY\n",
               fs->super->mountName);
    }
}

/*------------------------------------------------------------------------------
 * Zero the user data pages that haven't be commited.
 */

void
fs_zfZeroPage(fs_Context *  context,
              fs_Inode *    inode,
              fs_zfPage *   zfPage)
{
    fs_FileSystem * fs = context->fileSystem;
    fs_Buf *        buf = NULL;
    fs_Block        block = fs_emap_lookup_block(context, inode, zfPage->page);
    boolean         doRead = FALSE;

    ZF_TRACE(("fs_zfZeroPage(inum %lld, page %lld, offset %d)",
              inode->inum, zfPage->page, zfPage->offset));
    
    // Get the user data buffer.
    if (zfPage->offset != 0)
        doRead = TRUE;
    fs_logReplayFindBuf(context, NULL,
                        inode->inum,
                        zfPage->page,
                        block,
                        FALSE, //isMeta
                        doRead,
                        &buf);
    FS_ASSERT(buf != NULL);
    buf->flags &= ~FS_BUF_LOG_REPLAY; // don't need this flag

    // Fill it with zeroes.
    if (!buf->u.user.codHandle)
    {
        buf->u.user.codHandle = fs_cod_get_new_cod();
        fs_cod_create_null(buf->u.user.codHandle, FS_BLOCK_SIZE, 0);
    }
    else
    {
        FS_ASSERT(buf->u.user.codOffset == 0);
        fs_cod_fill_null(&(buf->u.user.codHandle),
                         (FS_BLOCK_SIZE - zfPage->offset - 1),
                         (zfPage->offset + 1));
    }
    buf->u.user.codLength = FS_BLOCK_SIZE;
    buf->u.user.codOffset = 0;

    // The buffer is now dirty.
    fs_logReplayMarkBufDirty(context, &fs->bufList[FS_BUFLIST_USER], buf);
    
    // Transfer the log depencies to the buffer,
    fs_logDoDepTransfer(context, &(buf->logSegDep), &(zfPage->logSegDep));

    // Release the buffer
    fs_bufReleaseExclusive(context, buf);

    return;
}

void
fs_userZeroThread(esm_handle_t  handle,
                  void *        arg)

{
    fs_Context *    context = arg;
    fs_FileSystem * fs = context->fileSystem;
    eeeRTC_t        beginTime, delta;

    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ, ("start user zero thread"));
    fs_lockShared(context, &(fs->accessLock), NULL);
    eee_rtc(&beginTime);

    // Zero the first page in the ZF list, remove it and start again
    // till list is empty.
    while (!DLL_IS_EMPTY(&(fs->zfInodeHdr)))
    {
        fs_zfInode *    zfInode = NULL;
        fs_zfPage *     zfPage = NULL;
        fs_Inode *      inode = NULL;
        uint64          inum;
        int32           rc;
        
        // Lock the ZF list.
        NFX_SPIN_LOCK(&(fs->zfSpin));

        // Make sure the list is not empty, otherwise we are done.
        if (DLL_IS_EMPTY(&(fs->zfInodeHdr)))
        {
            NFX_SPIN_UNLOCK(&(fs->zfSpin));
            break;
        }

        // Take the first inode from the list
        zfInode = fs->zfInodeHdr.head;
        FS_ASSERT(zfInode != DLL_NIL);
        inum = zfInode->inum;
        
        // Unlock the ZF list.
        NFX_SPIN_UNLOCK(&(fs->zfSpin));

        // Get the corresponding inode.
        inode = I_GET_SHARED(0, inum);
        FS_ASSERT(inode != NULL);
            
        // Lock the ZF list.
        NFX_SPIN_LOCK(&(fs->zfSpin));

        // Make sure the list is still not empty and
        // that the inode we got is still in the list.
        if (DLL_IS_EMPTY(&(fs->zfInodeHdr)))
        {
            NFX_SPIN_UNLOCK(&(fs->zfSpin));
            fs_inodeRelease(context, inode);
            inode = NULL;
            break;
        }
        
        zfInode = fs->zfInodeHdr.head;
        FS_ASSERT(zfInode != DLL_NIL);
        if (zfInode->inum != inum)
        {
            NFX_SPIN_UNLOCK(&(fs->zfSpin));
            fs_inodeRelease(context, inode);
            inode = NULL;
            continue;
        }
        
        // Take the first page in the list and remove it from the list.
        DLL_REMOVE_HEAD(&(zfInode->pageList), zfPage, entry);
        rc = chash_remove(fs_zfPageHashTab, (void *) zfPage);
        FS_ASSERT(rc == NFX_OK);
        
        // If no more pages for this inode, remove inode from ZF list.
        if (DLL_IS_EMPTY(&(zfInode->pageList)))
        {
            fs_zfDelInode(context, zfInode->inum);
        }

        // Unlock the ZF list.
        NFX_SPIN_UNLOCK(&(fs->zfSpin));
                    
        // Zero the page (inserted in cache...) and free the page.
        fs_zfZeroPage(context, inode, zfPage);
        fs_ramDealloc(FS_RAM_ZF, zfPage);
        
        // Release the inode
        fs_inodeRelease(context, inode);
        inode = NULL;

        // Don't hold the CPU for too long. Share it :)
        eee_rtcDuration(&beginTime, &delta);
        if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_unlockShared(context, &(fs->accessLock));
            fs_sleep(context, 0);
            eee_rtc(&beginTime);
            fs_lockShared(context, &(fs->accessLock), NULL);
        }
    } // while

    // The End.
    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ, ("end user zero thread"));
    NFX_SPIN_LOCK(&(fs->zfSpin));
    context->fileSystem->zfThreadHandle = 0;
    NFX_SPIN_UNLOCK(&(fs->zfSpin));
    fs_unlockShared(context, &(fs->accessLock));
    fs_threadTerminate(context);
    return;
}

/*----------------------------------------------------------------------------*/
#endif // NEW_ZF
