/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: file system debugger
 *
 * RCS:  $Id: fsdb.c,v 1.21 2002/11/27 17:39:26 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Read and dumps the on disk data for the file system
 *
 * Created by:	Jan Vanhoutte
 *
 * Date Created: May 20, 2002
 *
 *-----------------------------------------------------------------
 */
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#if defined(FS_STANDALONE)
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#endif

#include "fsdb-context.h"

#ifndef FS_STANDALONE
uint32 alignto(uint32 n, uint32 align);
#endif // FS_STANDALONE


fsdb_DumpFlags  fsdb_default_dump;
fsdb_DumpFlags  fsdb_local_dump;
fsdb_DumpFlags  fsdb_hex_byte_dump;
fsdb_DumpFlags  *fsdb_dump = &fsdb_default_dump;
fsdb_VolumeInfo fsdb_volume;

int    fsdb_fgets_uint64 (uint64 *value);
int    fsdb_fgets_uint32 (uint32 *value);
int    fsdb_fgets_int (int *value);
int    fsdb_continue (void);
int    fsdb_radix_convert(char *str, unsigned long word, int size, int radix);
void   fsdb_dump_bin(void *ptr, int len);
void   fsdb_dumpWord (void *ptr);
void   fsdb_dumpData (void *data, uint64 offset, uint32 size);
void   fsdb_dumpLine (char *format, ...);
void   fsdb_dumpError (char *format, ...);

int    fsdb_dumpSuperBlock (fs_SuperBlock *super);
int    fsdb_dumpDinode (fs_Dinode *dinode);
void   fsdb_dumpDinodeData (fs_Dinode *dinode);
void   fsdb_dumpDinodeDevNum (fs_Dinode *dinode);
void   fsdb_dumpDinodeEmaproot (fs_Dinode *dinode);
void   fsdb_dumpDinodeEmapflatroot (fs_Dinode *dinode);
int    fsdb_dumpRefByte (fs_RefByte *refByte);
void   fsdb_dumpBtree (fs_Btree *btree);
int32  fsdb_dirleaf_check (fs_Dirleaf *dirleaf);
void   fsdb_dumpDirleaf (fs_Dirleaf *dirleaf);
void   fsdb_dumpDirleafCompact (fs_DirleafCompact *dirleaf);
int    fsdb_dumpEmapleaf (fs_Emapleaf *emapleaf);
void   fsdb_dumpDinodePage (fs_DinodePage *dinodePage);
void   fsdb_dumpFileSetQuotaRecord (char *str, fs_QuotaRecord_t *quota);
void   fsdb_dumpFileSetConsumptionRecord (char *str, fs_ConsumptionRecord_t *consumption);
void   fsdb_dumpFileSet (fs_FileSetDiskRecord_t *fileset);
int    fsdb_dumpBuddy (uint32 *buddy);
void   fsdb_dumpFullInode (char8 *data);
void   fsdb_dumpDirInInode (char8 *data);


int    fsdb_dumpVollogBuf (vl_BufDiskHdr *logBuf);
void   fsdb_dumpLogRecord (void);
vl_BufDiskHdr *fsdb_readVolBuf(uint64 offset, uint32 size);
uint64 fsdb_findMostRecentLogBuf (uint64 startOffset, uint32 logSize, uint32 bufSize);
int    fsdb_dumpLog (uint64 startOffset, uint32 logSize, uint32 bufSize, uint64 *offset, int criteria, uint64 value);
void   fsdb_dumpLogRecord (void);

void   fsdb_dumpOffset (uint64 offset);
void   fsdb_set_debug (fsdb_SetDumpAction action);
void   fsdb_dump_block (void);
void   fsdb_do_read (uint64 offset, uint32 size);
void   fsdb_dumpLunLabel (void);

uint64 fsdb_emap_flat_root_lookup(fs_Emapflatroot *fr, uint64 page);
uint64 fsdb_emap_root_lookup(fs_Emaproot *r, uint64 page);
uint64 fsdb_emap_leaf_lookup(fs_Emapleaf *emapleaf, uint64 page);
void   fsdb_block_lookup (uint64 block);
uint64 fsdb_inum_page_lookup (uint64 inum, uint64 page);
int32  fsdb_superBlockCount (void);
uint64 fsdb_superBlockBlock (uint32 copy);
void   fsdb_read_predefined (void);
void   fsdb_read_menu (void);
void   fsdb_vol_menu (void);

void   fsdb_vol_info(void);
void   fsdb_data_info(void);
void   fsdb_info(void);
void   fsdb_util_menu (void);

void   fsdb_menu (void);

extern int    libc_check_crc  (int argc, char **argv);

#define STR_MAX_SIZE   128

char   dumpLine [1024];
int32  dumpCount = 0;
int32  lineCount = 0;

#ifndef FS_STANDALONE
uint32 alignto(uint32 n, uint32 align)
{
    return (n + align - 1) &~ (align - 1);
}
#endif // FS_STANDALONE

int
fsdb_fgets_uint64 (
    uint64    *value)
{
    char     str[STR_MAX_SIZE];
    int      loop;
    
    str[0] = 0;
    fgets(str, STR_MAX_SIZE, stdin);
    if ((str[0] == '0') && ((str[1] == 'x') || (str[1] == 'X')))
    {
        return (sscanf(str, "%llx", value));
    }
    else
    {
        return (sscanf(str, "%lld", value));
    }
}

int
fsdb_fgets_uint32 (
    uint32    *value)
{
    char     str[STR_MAX_SIZE];

    str[0] = 0;
    fgets(str, STR_MAX_SIZE, stdin);
    if ((str[0] == '0') && ((str[1] == 'x') || (str[1] == 'X')))
    {
        return (sscanf(str, "%x", value));
    }
    else
    {
        return (sscanf(str, "%d", value));
    }
}

int
fsdb_fgets_int (
    int     *value)
{
    char     str[STR_MAX_SIZE];

    str[0] = 0;
    fgets(str, STR_MAX_SIZE, stdin);
    return (sscanf(str, "%d", value));
}

int
fsdb_continue(void)
{
    char     str[STR_MAX_SIZE];
    
    str[0] = 0;
    printf("Continue or 'q' .... ");
    fgets(str, STR_MAX_SIZE, stdin);
    if ((str[0] == 'q') || (str[0] == 'Q'))
    {
        return FALSE;
    }
    return TRUE;
}

void
fsdb_dumpLine (
    char            *format,
    ...
)
{
    va_list ap;

    if (fsdb_dump->level < FSDB_DUMP_LEVEL_MEDIUM)
    {
        return;
    }
    
    va_start(ap, format);
    vprintf(format, ap);
    va_end(ap);
}

void
fsdb_dumpError (
    char            *format,
    ...
)
{
    va_list ap;

    if (fsdb_dump->level < FSDB_DUMP_LEVEL_MEDIUM)
    {
        return;
    }
    
    va_start(ap, format);
    vprintf(format, ap);
    va_end(ap);
}


int
fsdb_radix_convert(char *str, unsigned long word, int size, int radix)
{
	int		j;

	str[size] = 0;
	for (j = size - 1; j >= 0; --j)
	{
		int digit = word % radix;
		word /= radix;
		if (digit < 10)
			str[j] = '0' + digit;
		else
			str[j] = 'A' + digit - 10;
	}
    return (size);
}


void
fsdb_dump_bin(void *ptr, int len)
{
    int      loop;
    int64    bit;
    char    *data = ptr;
    short   *data16 = ptr;
    int32   *data32 = ptr;
    int64   *data64 = ptr;


    len--;
    for (loop = len ; loop >= 0 ; loop--)
    {
        switch (fsdb_dump->size)
        {
        default:
        case FSDB_DUMP_1_BYTE:
            bit = *data & (1 << loop);
            break;
        case FSDB_DUMP_2_BYTES:
            bit = *data16 & (1 << loop);
            break;
        case FSDB_DUMP_4_BYTES:
            bit = *data32 & (1 << loop);
            break;
        case FSDB_DUMP_8_BYTES:
            bit = *data64 & (1 << loop);
            break;
        }
        
        if (bit)
            printf("1");
        else
            printf("0");
    }
    printf(" ");
}

void 
fsdb_dumpWord (
    void *ptr)
{
    char   *data = ptr;
    short  *data16 = ptr;
    int32  *data32 = ptr;
    int64  *data64 = ptr;

    switch (fsdb_dump->size)
    {
    default:
    case FSDB_DUMP_1_BYTE:
        switch (fsdb_dump->format)
        {
        default:
        case FSDB_DUMP_HEX:
            printf("%02x ", *data & 0xff); 
            break;
        case FSDB_DUMP_DEC:
            printf("%3d ", *data & 0xff); 
            break;
        case FSDB_DUMP_OCT:
            printf("%03o ", *data & 0xff); 
            break;
        case FSDB_DUMP_BIN:
            fsdb_dump_bin(data, 8);
            break;
        }
        break;
    case FSDB_DUMP_2_BYTES:
        switch (fsdb_dump->format)
        {
        default:
        case FSDB_DUMP_HEX:
            printf("%04x ", *data16 & 0xffff); 
            break;
        case FSDB_DUMP_DEC:
            printf("%5d ", *data16 & 0xffff); 
            break;
        case FSDB_DUMP_OCT:
            printf("%06o ", *data16 & 0xffff); 
            break;
        case FSDB_DUMP_BIN:
            fsdb_dump_bin(data16, 16);
            break;
        }
        break;
    case FSDB_DUMP_4_BYTES:
        switch (fsdb_dump->format)
        {
        default:
        case FSDB_DUMP_HEX:
            printf("%08x ", *data32); 
            break;
        case FSDB_DUMP_DEC:
            printf("%10d ", *data32); 
            break;
        case FSDB_DUMP_OCT:
            printf("%012o ", *data32); 
            break;
        case FSDB_DUMP_BIN:
            fsdb_dump_bin(data32, 32);
            break;
        }
                break;
    case FSDB_DUMP_8_BYTES:
        switch (fsdb_dump->format)
        {
        default:
        case FSDB_DUMP_HEX:
            printf("%016llx ", *data64); 
            break;
        case FSDB_DUMP_DEC:
            printf("%16lld ", *data64); 
            break;
        case FSDB_DUMP_OCT:
            printf("%024llo ", *data64); 
            break;
        case FSDB_DUMP_BIN:
            fsdb_dump_bin(data64, 64);
            break;
        }
        break;
    }
}



void 
fsdb_dumpData (
    void *ptr,
    uint64 offset,
    uint32 len)
{
	int    charCount = 0 ;
    int    wordCount;
    char   *data = ptr;
    short  *data16;
    int32  *data32;
    int64  *data64;
	char   ch;
    int    charPerLine = 16;   // should be multiple of 8
    int    wordsPerLine;
    int    wordSize = 1;

    switch (fsdb_dump->size)
    {
    default:
    case FSDB_DUMP_1_BYTE:
        break;
    case FSDB_DUMP_2_BYTES:
        wordSize = 2;
        data16 = (short  *)((address_t)(data + 1) & (address_t)~1);
        if ((char*)data16 != data)
        {
            printf("Data not word (2 bytes) aligned, first byte ");
            fsdb_dump->size = FSDB_DUMP_1_BYTE;
            fsdb_dumpWord(ptr);
            fsdb_dump->size = FSDB_DUMP_2_BYTES;
            printf("\n");
            charCount = ((char*)data16 - data) ;
        }
        break;
    case FSDB_DUMP_4_BYTES:
        wordSize = 4;
        data32 = (int32  *)((address_t)(data + 3) & (address_t)~3);
        if ((char*)data32 != data)
        {
            printf("Data not word (4 bytes) aligned, first bytes ");
            fsdb_dump->size = FSDB_DUMP_1_BYTE;
            fsdb_dumpWord(ptr);
            fsdb_dump->size = FSDB_DUMP_4_BYTES;
            printf("\n");
            charCount = ((char*)data32 - data) ;
        }
        break;
    case FSDB_DUMP_8_BYTES:
        wordSize = 8;
        data64 = (int64  *)((address_t)(data + 7) & (address_t)~7);
        if ((char*)data64 != data)
        {
            printf("Data not word (8 bytes) aligned, first byte ");
            fsdb_dump->size = FSDB_DUMP_1_BYTE;
            fsdb_dumpWord(ptr);
            fsdb_dump->size = FSDB_DUMP_8_BYTES;
            printf("\n");
            charCount = ((char*)data64 - data) ;
        }
        break;
    }

    if (fsdb_dump->format == FSDB_DUMP_BIN)
    {
        charPerLine = 8 ;
    }
    
    wordsPerLine = charPerLine / wordSize;

	while (charCount < len)
    {
		printf("\n%08llx: %04x: %s: ", offset + charCount, charCount, fsdb_dump->typeStr);
		sprintf(dumpLine, "\n%04x: %s: ", charCount, fsdb_dump->typeStr);
        dumpCount = sizeof(dumpLine);

		for(wordCount = charCount; 
            wordCount < MIN(charCount+charPerLine, len); 
            wordCount += wordSize )
        {
            fsdb_dumpWord(&data[wordCount]);
/*
            dumpCount += fsdb_radix_convert(&dumpLine[dumpCount], 
                                            &data[wordCount], 
                                            fsdb_dump->wordSize, 
                                            fsdb_dump->radix);
*/
        }
        
		printf("    ");

		for(wordCount = charCount; 
            wordCount < MIN(charCount+charPerLine, len); 
            wordCount++ ) 
        {
			ch = data[wordCount] & 0xff;
			if( (ch >= 0x20) && (ch < 0x7F))
				printf("%c", ch); 
			else
			if( ch == 0 ) 
				printf("%c",'.'); 
			else
				printf("%c",'_'); 
		}

        charCount += charPerLine;
	}
	printf("\n");

}


int
fsdb_dumpSuperBlock (fs_SuperBlock *super)
{
    // return TRUE is valid
    uint32 savedCksum, newCksum;
    uint32 loop_1, loop_2;
    int validSuperBlock = TRUE;

    if (super->magic != FS_SUPER_MAGIC)
    {
        fsdb_dumpError("Super block has wrong magic: 0x%x (should be 0x%x)\n",
               super->magic, FS_SUPER_MAGIC);
        fsdb_dumpData (super, fsdb_volume.offset, sizeof(fs_SuperBlock));
        return FALSE;
    }
    else
    {
        fsdb_dumpLine("Super block magic:                       0x%x\n", super->magic);
    }

    if ( ! ((super->version == 4) ||
            (super->version == 5) ||
            (super->version == 6) ||
            (super->version == 7) ||
            (super->version == 8) ||
            (super->version == 10) ||
            (super->version == 11) ||
            (super->version == FS_SUPER_CURRENT_VERSION)))
    {
        fsdb_dumpError("Super block has wrong version: 0x%x (should be 0x%x)\n",
               super->version, FS_SUPER_CURRENT_VERSION);
        fsdb_dumpData (super, fsdb_volume.offset, sizeof(fs_SuperBlock));
        return FALSE;
    }
    else
    {
        fsdb_dumpLine("Super block version:                     0x%x\n", super->version);
    }

    savedCksum = super->crc32;
    super->crc32 = 0 ;
    newCksum = crc32_checksum(super, sizeof(fs_SuperBlock), ~0);
    super->crc32 = savedCksum;
    
    if (savedCksum != newCksum)
    {
        fsdb_dumpError("Super block has wrong CRC: 0x%x (should be 0x%x)\n",
                       savedCksum, newCksum);
        fsdb_dumpError("... continue\n");
        validSuperBlock = FALSE;
    } 
    else
    {
        fsdb_dumpLine("Super block CRC32:                       0x%x\n", savedCksum);
    }
    
    if (super->flags & ~(FS_SUPER_FLAG_CLEAN | FS_SUPER_FLAG_MKFS))
    {
        fsdb_dumpError("Super block has wrong flags: 0x%x (only bit 0 and 1 is used)\n",
               super->flags);
    }
    fsdb_dumpLine("Super block flags                        0x%x\n", super->flags);
    if (super->flags & FS_SUPER_FLAG_CLEAN)
        fsdb_dumpLine("            is CLEAN\n");
    if (super->flags & FS_SUPER_FLAG_MKFS)  
        fsdb_dumpLine("             is mkfs's\n");

    fsdb_dumpLine("\n");
    fsdb_dumpLine("Super block: number of bytes per blocks  %d (0x%x)\n", 
                  super->blkSize, super->blkSize);
    fsdb_dumpLine("Super block: spareword                   %d (0x%x)\n", 
                  super->spareword, super->spareword);
    fsdb_dumpLine("Super block: total number of blocks      %lld (0x%llx)\n", 
                  super->numBlks, super->numBlks);
    fsdb_dumpLine("Super block: number of blocks in use     %lld (0x%llx)\n", 
                  super->blksUsed, super->blksUsed);
    fsdb_dumpLine("Super block: number of inodes in use     %lld (0x%llx)\n", 
                  super->inodesUsed, super->inodesUsed);
    fsdb_dumpLine("Super block: ratio of blocks to inode    %d\n", 
                  super->blockInodeRatio);
    fsdb_dumpLine("\n");
    fsdb_dumpLine("Super block: how many times mounted      %d\n", super->nbrMount);
    fsdb_dumpLine("Super block: how many times unmounted    %d\n", super->nbrUmount);
    fsdb_dumpLine("Super block: how many times dirty mount  %d\n", super->nbrDirtyMount);
    fsdb_dumpLine("Super block: mount failed due to log     %d\n", super->nbrMountFailedLogDev);
    fsdb_dumpLine("Super block: mount failed due to replay  %d\n", super->nbrMountFailedReplay);
    fsdb_dumpLine("Super block: mount failed due to eek     %d\n", super->nbrMountFailedEek);
    fsdb_dumpLine("Super block: mount failed due to snap    %d\n", super->nbrMountFailedSnap);
    fsdb_dumpLine("\n");
    fsdb_dumpLine("Super block: date of mkfs                %s", ctime((time_t *)&super->mkfsTime));
    fsdb_dumpLine("Super block: log init time               %s", ctime((time_t *)&super->logInitTimeSec));
    fsdb_dumpLine("Super block: log init nsec               %d\n", super->logInitTimeNsec);
    fsdb_dumpLine("\n");
    fsdb_dumpLine("Super block: number of filesets          %d\n", super->numFset);
    switch (super->logDevType)
    {
    case FS_LOG_DEVTYPE_NULL:
        fsdb_dumpLine("Super block: log device type             NULL\n");
        fsdb_dumpLine("Super block: (volId not used)            %lld (0x%llx)\n", 
                      super->logVolId, super->logVolId);
        break;
    case FS_LOG_DEVTYPE_NVRAM:
        fsdb_dumpLine("Super block: log device type             NVRAM\n");
        fsdb_dumpLine("Super block: (volId not used)            %lld (0x%llx)\n", 
                      super->logVolId, super->logVolId);
        break;
    case FS_LOG_DEVTYPE_SELF:
        fsdb_dumpLine("Super block: log device type             SELF\n");
        fsdb_dumpLine("Super block: (volId not used)            %lld (0x%llx)\n", 
                      super->logVolId, super->logVolId);
        break;
    case FS_LOG_DEVTYPE_VOLID:
        fsdb_dumpLine("Super block: log device type             EXT VOL\n");
        fsdb_dumpLine("Super block: log device volume id        %lld (0x%llx)\n", 
                      super->logVolId, super->logVolId);
        break;
    default: 
        fsdb_dumpLine("Super block: log device type             UNKNOWN\n");
        fsdb_dumpLine("Super block: (volId not used)            %lld (0x%llx)\n", 
                      super->logVolId, super->logVolId);
        break;
    }
    fsdb_dumpLine("Super block: log size (bytes)            %d\n", super->logSize);
    fsdb_dumpLine("Super block: local reserved log space    %d\n", super->logSpaceSize);
    fsdb_dumpLine("Super block: log start offset            %lld\n", super->logStartOffset);
    fsdb_dumpLine("\n");
    fsdb_dumpLine("Super block: snapshot bitmap             0x%016llx\n", super->snapshots);
    fsdb_dumpLine("Super block: snapshot version            %d\n", super->snapState.version);
    fsdb_dumpLine("Super block: snapshot id                 %d\n", super->snapState.id);
    fsdb_dumpLine("Super block: snapshot phase              %d\n", super->snapState.phase);
    fsdb_dumpLine("Super block: snap times enable           %d\n", super->snapState.times.enable);
    fsdb_dumpLine("Super block: snap times max weekly       %d\n", super->snapState.times.maxWeekly);
    fsdb_dumpLine("Super block: snap times max daily        %d\n", super->snapState.times.maxDaily);
    fsdb_dumpLine("Super block: snap times max hourly       %d\n", super->snapState.times.maxHourly);
    fsdb_dumpLine("Super block: snap times hour bitmap      0x%08x\n", super->snapState.times.hours);
    fsdb_dumpLine("\n");
    fsdb_dumpLine("Super block: mount name                  %s\n", super->mountName);
    fsdb_dumpLine("Super block: Perc. of always av. blocks  %d\n", super->percentMinFree);
    fsdb_dumpLine("\n");

    for (loop_1 = 0 ; loop_1 < FS_SUPER_NBR_MOUNTTIME; loop_1++)
    {
        if (super->mountTime[loop_1] == 0)
        {
            break;
        }
    }
    fsdb_dumpLine("Super block: result of last %d mount operation(s)\n", loop_1);
    for (loop_2 = 0 ; loop_2 < loop_1; loop_2++)
    {
        if (super->mountFail[loop_2] == FS_MOUNT_NO_FAIL)
        {
            switch (super->mountState[loop_2])
            {
            case FS_VOL_STATE_ONLINE:
                fsdb_dumpLine ("  online         ");
                break;
            case FS_VOL_STATE_ONLINE_RDONLY:
                fsdb_dumpLine ("  online RO      ");
                break;
            case FS_VOL_STATE_OFFLINE: 
                fsdb_dumpLine ("  offline        ");
                break;
            default:
                fsdb_dumpLine ("  unknown state  ");
                break;
            }
        }
        else
        {
            switch (super->mountFail[loop_2])
            {
            case FS_MOUNT_FAIL_LOGDEV:
                fsdb_dumpLine ("  Failed due to wrong log device  ");
                break;
            case FS_MOUNT_FAIL_REPLAY:
                fsdb_dumpLine ("  Failed during replay  ");
                break;
            case FS_MOUNT_FAIL_EEK:
                fsdb_dumpLine ("  Failed during eek  ");
                break;
            case FS_MOUNT_FAIL_SNAP:
                fsdb_dumpLine ("  Failed during snap resume  ");
                break;
            default:
                fsdb_dumpLine ("  Failed for unknown reason %d  ", super->mountFail[loop_2]);
                break;
            }
        }
        fsdb_dumpLine ("  Mount time %s", ctime((time_t *)&super->mountTime[loop_2]));
        // fsdb_dumpLine ("\n");  // ctime adds the \n
    }
    
    for (loop_1 = 0 ; loop_1 < FS_SUPER_NBR_MOUNTTIME; loop_1++)
    {
        if (super->umountTime[loop_1] == 0)
        {
            break;
        }
    }
    fsdb_dumpLine("Result of last %d umount operation(s)\n", loop_1);
    for (loop_2 = 0 ; loop_2 < loop_1; loop_2++)
    {
        switch (super->umountState[loop_2])
        {
        case FS_UMOUNT_NORMAL:
            break;
        default:
            fsdb_dumpLine ("  unknown state");
            break;
        }
        fsdb_dumpLine ("  umount time %s", ctime((time_t *)&super->umountTime[loop_2]));
        // fsdb_dumpLine ("\n");  // ctime adds the \n
    }
    fsdb_dumpLine ("\n");

    fsdb_dumpLine("Super block: nbr of grow operation(s)    %d\n", super->nbrGrow);
    for (loop_1 = 0 ; loop_1 < FS_SUPER_NBR_MOUNTTIME; loop_1++)
    {
        if (super->growTime[loop_1] == 0)
        {
            break;
        }
    }
    fsdb_dumpLine("Date of last %d growTime operation(s)\n", loop_1);
    for (loop_2 = 0 ; loop_2 < loop_1; loop_2++)
    {
        fsdb_dumpLine ("  growfs time %s", ctime((time_t *)&super->growTime[loop_2]));
        // fsdb_dumpLine ("\n");  // ctime adds the \n
    }
    fsdb_dumpLine ("\n");

    if (validSuperBlock)
    {
        fsdb_volume.mkfsTime = super->mkfsTime;
    }

    return (validSuperBlock);
}

int
fsdb_dumpDinode (fs_Dinode *dinode)
{
    int inUse = TRUE;  // also means valid 

    fsdb_dumpLine("Dinode: file size (bytes)                %lld (0x%llx)\n", 
                  dinode->size, dinode->size); 
    fsdb_dumpLine("Dinode: attrs                            %d (0x%x 0%o)\n", 
                  dinode->attrs, dinode->attrs, dinode->attrs);

    if ((dinode->attrs & FS_IFMT_MASK) == 0)
    {
        inUse = FALSE;
    }
    
    switch (dinode->attrs & FS_IFMT_MASK)
    {
    case FS_IFMT_SOCK:
        fsdb_dumpLine("                SOCKET\n");
        break;
    case FS_IFMT_SYML:
        fsdb_dumpLine("                SYMBOLIC LINK\n");
        break;
    case FS_IFMT_REG:
        fsdb_dumpLine("                REGULAR FILE\n");
        break;
    case FS_IFMT_BLK:
        fsdb_dumpLine("                BLOCK FILE\n");
        break;
    case FS_IFMT_DIR:
        fsdb_dumpLine("                DIRECTORY\n");
        break;
    case FS_IFMT_CHR:
        fsdb_dumpLine("                CHAR FILE\n");
        break;
    case FS_IFMT_FIFO:
        fsdb_dumpLine("                FIFO FILE\n");
        break;
    case FS_IFMT_QUOT:
        fsdb_dumpLine("                QUOTA FILE\n");
        break;
    case 0:
        fsdb_dumpLine("                MODE == 0 : unused entry\n");
        break;
    default:
        fsdb_dumpLine("                UNKNOWN FILE TYPE\n");
        inUse = FALSE;
        break;
    }
    if (dinode->attrs & FS_IATR_SUID)
        fsdb_dumpLine("                SET USER ID on execution\n");
    if (dinode->attrs & FS_IATR_SGID)
        fsdb_dumpLine("                SET GROUP ID on execution\n");
    if (dinode->attrs & FS_IATR_DNC)
    {
        if (FS_I_IS_REG(dinode->attrs))
            fsdb_dumpLine("                do-not-cache flag (regular file)\n");
        else
            fsdb_dumpLine("                restriction deletion flag (dir)\n");
    }
    
    if (dinode->attrs & FS_IFLG_NO_SNAP)
        fsdb_dumpLine("                don't snapshot\n");
    if (dinode->attrs & FS_IFLG_DIR_SPLIT)
        fsdb_dumpLine("                dir has split\n");
    if (dinode->attrs & FS_IFLG_COMPACT_EMAP)
        fsdb_dumpLine("                emap is in compact leaf form\n");
    if (dinode->attrs & FS_IFLG_COMPACT_DIR)
        fsdb_dumpLine("                dir is in compact form\n");

// TODO: display userAttr field

    fsdb_dumpLine("Dinode: link count                       %d (0x%x)\n", 
                  dinode->linkCnt, dinode->linkCnt);
    fsdb_dumpLine("Dinode: ctime of last attr modification  %s",
                  ctime((time_t *)&dinode->ctime));
    fsdb_dumpLine("Dinode: mtime of last data modification  %s", 
                  ctime((time_t *)&dinode->mtime));
    fsdb_dumpLine("Dinode: atime of last read/write         %s", 
                  ctime((time_t *)&dinode->atime));
    fsdb_dumpLine("Dinode: inode generation number          %d (0x%x)\n", 
                  dinode->genNum, dinode->genNum);
    fsdb_dumpLine("Dinode: fset number                      %d (0x%x)\n", 
                  dinode->fsetNum, dinode->fsetNum);
    fsdb_dumpLine("Dinode: parent Inum                      %lld (0x%llx)\n", 
                  dinode->piu.parentInum, dinode->piu.parentInum);
    fsdb_dumpLine("Dinode: index Inum                       %lld (0x%llx)\n", 
                  dinode->piu.indexInum, dinode->piu.indexInum);
    fsdb_dumpLine("Dinode: number of allocated blocks       %lld (0x%llx)\n", 
                  dinode->numblocks, dinode->numblocks);
    fsdb_dumpLine("Dinode: security id                      %d (0x%x)\n", 
                  dinode->securityId, dinode->securityId);

    fsdb_dumpLine("\n");

    switch (dinode->attrs & FS_IFMT_MASK)
    {
    case FS_IFMT_BLK:
    case FS_IFMT_CHR:
        fsdb_dumpDinodeDevNum (dinode);
        break;
    case FS_IFMT_SOCK:
    case FS_IFMT_FIFO:
        // empty
        break;
    case FS_IFMT_SYML:
        if (dinode->size <= 128)
        {
            fsdb_dumpLine("Dinode: symlink name                     %.128s\n", dinode->u.data);
            break;
        }
        // else fall through to regular file:
    case FS_IFMT_REG:
    case FS_IFMT_QUOT:
    case FS_IFMT_DIR:
        if (dinode->u.emaproot.hdr.level == 0)
        {
            fsdb_dumpDinodeEmapflatroot (dinode);
        }
        else
        {
            fsdb_dumpDinodeEmaproot (dinode);
        }
        break;
    case 0:
        fsdb_dumpLine("                MODE == 0\n");
        fsdb_dumpDinodeData (dinode);
        break;
    default:
        fsdb_dumpLine("                UNKNOWN FILE TYPE\n");
        fsdb_dumpDinodeData (dinode);
        break;
    }

    return (inUse);
}

void
fsdb_dumpDinodeData (fs_Dinode *dinode)
{
    fsdb_dumpData(dinode->u.data, 0, FS_INODE_INLINE_LEN);
}

void
fsdb_dumpDinodeDevNum (fs_Dinode *dinode)
{
    fsdb_dumpLine("Dinode: major                            %d (0x%x)\n", 
                  dinode->u.devNum.major, dinode->u.devNum.major);
    fsdb_dumpLine("Dinode: minor                            %d (0x%x)\n", 
                  dinode->u.devNum.minor, dinode->u.devNum.minor);
    
}

void
fsdb_dumpDinodeEmaproot (fs_Dinode *dinode)
{
    int  loop;

    fsdb_dumpLine("Dinode - emaproot   : nkeys              %d (0x%x)\n",
                  dinode->u.emaproot.hdr.nkeys,
                  dinode->u.emaproot.hdr.nkeys);
    fsdb_dumpLine("Dinode - emaproot   : level              %d (0x%x)\n",
                  dinode->u.emaproot.hdr.level,
                  dinode->u.emaproot.hdr.level);

    if ((dinode->u.emaproot.hdr.nkeys < 0) ||
        (dinode->u.emaproot.hdr.nkeys >= FS_INOROOT_MAXKEYS))
    {
        fsdb_dumpError("Dinode - emaproot: nkeys (%d) >= MAX (%d)\n", 
                       dinode->u.emaproot.hdr.nkeys, FS_INOROOT_MAXKEYS);
        return;
    }

    // for (loop = 0 ; loop < FS_INOROOT_MAXKEYS; loop++)
    for (loop = 0 ; loop < dinode->u.emaproot.hdr.nkeys; loop++)
    { 
        fsdb_dumpLine("Dinode - emaproot %2d: offset len start   %lld %lld %lld\n",
                      loop,
                      dinode->u.emaproot.offsets[loop],
                      dinode->u.emaproot.left[loop] >> EXTENT_BLOCK_BITS,
                      dinode->u.emaproot.left[loop] & EXTENT_BLOCK_MASK);
        fsdb_dumpLine("                      offset len start   0x%llx 0x%llx 0x%llx\n",
                      dinode->u.emaproot.offsets[loop],
                      dinode->u.emaproot.left[loop] >> EXTENT_BLOCK_BITS,
                      dinode->u.emaproot.left[loop] & EXTENT_BLOCK_MASK);

        /*
        fsdb_dumpLine("Dinode - emaproot %2d: offsets            %lld (0x%llx)\n",
                      loop,
                      dinode->u.emaproot.offsets[loop],
                      dinode->u.emaproot.offsets[loop]);

        fsdb_dumpLine("Dinode - emaproot %2d:   left len         %lld (0x%llx)\n",
                      loop,
                      dinode->u.emaproot.left[loop] >> EXTENT_BLOCK_BITS,
                      dinode->u.emaproot.left[loop] >> EXTENT_BLOCK_BITS);
        fsdb_dumpLine("Dinode - emaproot %2d:   left start       %lld (0x%llx)\n",
                      loop,
                      dinode->u.emaproot.left[loop] & EXTENT_BLOCK_MASK,
                      dinode->u.emaproot.left[loop] & EXTENT_BLOCK_MASK);
        */
    }
    fsdb_dumpLine("Dinode - emaproot   : right              %lld (0x%llx)\n",
                  dinode->u.emaproot.right,
                  dinode->u.emaproot.right);
}

void
fsdb_dumpDinodeEmapflatroot (fs_Dinode *dinode)
{
    int  loop;
    fsdb_dumpLine("Dinode - emapflatroot   : nkeys             %d (0x%x)\n",
                  dinode->u.emapflatroot.hdr.nkeys,
                  dinode->u.emapflatroot.hdr.nkeys);
    fsdb_dumpLine("Dinode - emapflatroot   : level             %d (0x%x)\n",
                  dinode->u.emapflatroot.hdr.level,
                  dinode->u.emapflatroot.hdr.level);

    if ((dinode->u.emapflatroot.hdr.nkeys < 0) ||
        (dinode->u.emapflatroot.hdr.nkeys >= FS_INOFLATROOT_MAXKEYS))
    {
        fsdb_dumpError("Dinode - emapflatroot: nkeys (%d) >= MAX (%d)\n", 
                       dinode->u.emapflatroot.hdr.nkeys, FS_INOFLATROOT_MAXKEYS);
        return;
    }

    // for (loop = 0 ; loop < FS_INOFLATROOT_MAXKEYS; loop++)
    for (loop = 0 ; loop < dinode->u.emapflatroot.hdr.nkeys ; loop++)
    { 
        fsdb_dumpLine("Dinode - emapflatroot %2d: offset len start  %lld %lld %lld\n",
                      loop,
                      dinode->u.emapflatroot.offsets[loop],
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        fsdb_dumpLine("                          offset len start  0x%llx 0x%llx 0x%llx \n",
                      dinode->u.emapflatroot.offsets[loop],
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK);

        /*
        fsdb_dumpLine("Dinode - emapflatroot %2d: offsets           %lld (0x%llx)\n",
                      loop,
                      dinode->u.emapflatroot.offsets[loop],
                      dinode->u.emapflatroot.offsets[loop]);

        fsdb_dumpLine("Dinode - emapflatroot %2d: extents len       %lld (0x%llx)\n",
                      loop,
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS);
        fsdb_dumpLine("Dinode - emapflatroot %2d: extents start     %lld (0x%llx)\n",
                      loop,
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK,
                      dinode->u.emapflatroot.extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        */
    }
}

void
fsdb_dumpBtree (fs_Btree *btree)
{
    int  loop, index;

    fsdb_dumpLine("Btree - level                             %d (0x%x)\n",
                      btree->level,
                      btree->level);
    fsdb_dumpLine("Btree - nkeys                             %d (0x%x)\n",
                      btree->nkeys,
                      btree->nkeys);

    if ((btree->nkeys < 0) ||
        (btree->nkeys >= FS_BTREE_MAXKEYS))
    {
        fsdb_dumpError("Btree: nkeys (%d) >= MAX (%d)\n", 
                       btree->nkeys, FS_BTREE_MAXKEYS);
        return;
    }

    fsdb_dumpLine("Btree: index values [0..%d]\n", btree->nkeys);
    
    for (loop = 0 ; loop < btree->nkeys; loop++)
    {
        if (loop && ((loop % 15) == 0))
            fsdb_dumpLine ("\n");
        
        fsdb_dumpLine("%3d ", btree->index[loop]);
    }
    fsdb_dumpLine ("\n");

    for (loop = 0 ; loop < btree->nkeys; loop++)
    {
        index = btree->index[loop];

        fsdb_dumpLine("Btree: index %3d %3d : keys - left        %lld  (0x%llx) - %lld (0x%llx)\n",
                      loop, index,
                      btree->keys[index],
                      btree->keys[index],
                      btree->left[index],
                      btree->left[index]);
    }
    
    fsdb_dumpLine("Btree - right                             %lld (0x%llx)\n",
                      btree->right,
                      btree->right);
}

int
fsdb_dumpRefByte (fs_RefByte *refByte)
{
    int  loop;
    int  inUse = 0;
    int  isLiveError = 0;
    int  snapError = 0;
    int  validRefBytePage = TRUE;

    for (loop = 0 ; loop < FS_BLOCK_SIZE; loop++)
    {
        if (refByte[loop])
        {
            fsdb_dumpLine("Ref byte %4d : %s : %s : snapcount %d\n",
                          loop,
                          ((refByte[loop] & FS_REFBYTE_IS_LIVE) == 0) ? "Error !!" : "isLive  ",
                          ((refByte[loop] & FS_REFBYTE_NEVER_SNAPSHOT) == 1) ? "No snap " : "snap    ",
                          (refByte[loop] >> FS_REFBYTE_SNAP_COUNT_SHIFT));
            
            if ((refByte[loop] & FS_REFBYTE_IS_LIVE) == 0)
                isLiveError++;

            if (((refByte[loop] & FS_REFBYTE_NEVER_SNAPSHOT) == 1) && 
                ((refByte[loop] >> FS_REFBYTE_SNAP_COUNT_SHIFT) != 0))
                snapError++;

            inUse++;
        }
    }
    fsdb_dumpLine("Ref bytes: total in use %d - free %d\n", inUse, FS_BLOCK_SIZE - inUse);
    if ((isLiveError + snapError) > 0)
    {
        fsdb_dumpLine("Ref bytes: 'in live' error cnt %d  'snap' error cnt %d\n", isLiveError, snapError);
        validRefBytePage = FALSE;
    }
    
    fsdb_dumpLine("\n");

    return (validRefBytePage);    
}

/*
 * Perform consistency checks on directory leaf.
 * Print an error message 
 * return NFX_ERR if not value dir leaf page
 */ 
int32
fsdb_dirleaf_check(fs_Dirleaf *dirleaf)
{
    enum {BOGUS, ALLOC, FREE, CONT};
	uint8   index[FS_DIRLEAF_ENTRIES_MAX];
    int     alloc, free, cont;
	int     i;

	bzero(index, sizeof index);

	/*
	 * Find allocated slots.
	 */
	alloc = 0;
	for (i = 0; i < sizeof index; i++) {
		if (dirleaf->u[i].ent.len != 0) {
			index[i] = ALLOC;
			alloc++;
		}
	}

	/*
	 * If there are any free slots, the freelist pointer should point
	 * to them.
	 */
    if (dirleaf->count != FS_DIRLEAF_ENTRIES_MAX &&
        dirleaf->freelist >= FS_DIRLEAF_ENTRIES_MAX)
    {
        fsdb_dumpError("dirleaf freelist incorrect (count %d freelist %d)\n", 
                       dirleaf->count, dirleaf->freelist);
        return NFX_ERR;
    }

	/*
	 * Find free slots.
	 */
	free = 0;
	for (i = dirleaf->freelist; i < FS_DIRLEAF_ENTRIES_MAX; i = dirleaf->u[i].cont.cont) {
        if (index[i] != 0)
        {
            fsdb_dumpError("dirleaf index %d on freelist twice\n",  i);
            return NFX_ERR;
        }
		index[i] = FREE;
		free++;
	}

    if (dirleaf->count + free != FS_DIRLEAF_ENTRIES_MAX)
    {
        fsdb_dumpError("dirleaf count (%d) + free count (%d) != MAX (%d)\n",
            dirleaf->count, free, FS_DIRLEAF_ENTRIES_MAX);
        return NFX_ERR;
    }

	/*
	 * Find continuation slots.
	 */
	cont = 0;
	for (i = 0; i < sizeof index; i++) {
		int len = (int)dirleaf->u[i].ent.len - sizeof dirleaf->u[i].ent.name;
		if (len > 0) {
			int j = i;
			while (1) {
				j = dirleaf->u[j].cont.cont;

                if (len <= 0)
                {
                    fsdb_dumpError("dirleaf index %d len exausted before cont entries\n",
                                   i);
                    return NFX_ERR;
                }

                if (index[j] != 0)
                {
                    fsdb_dumpError("dirleaf index %d in cont chain already referenced\n",
                                   j);
                    return NFX_ERR;
                }

				index[j] = CONT;
				cont++;
				if (dirleaf->u[j].cont.cont == FS_DIRCONT_EOL) {

                    if (len > sizeof dirleaf->u[i].cont.name)
                    {
                        fsdb_dumpError("dirleaf index %d cont entries exausted before len\n",
                                       i);
                        return NFX_ERR;
                    }

					break;
				}
				len -= sizeof dirleaf->u[i].cont.name;
			}
		}
	}

    if (alloc + cont != dirleaf->count)
    {
        fsdb_dumpError("dirleaf count (%d) != alloc (%d) + cont (%d)\n",
            dirleaf->count, alloc, cont);
        return NFX_ERR;
    }

	for (i = 0; i < sizeof index; i++)
    {
        if (index[i] == BOGUS)
        {
            fsdb_dumpError("dirleaf index %d not accounted for\n", i);
            return NFX_ERR;
        }
    }

    return NFX_OK;
}

void 
fsdb_dumpDirleaf (fs_Dirleaf *dirleaf)
{
    int  loop;
    int  used = 0;
    int  cont;
    int  index;
    int  leaf;
    int  len;
#define FSDB_MAX_NAME_ENTRY_LEN  34
    char name[FSDB_MAX_NAME_ENTRY_LEN];
    
    if (fsdb_dirleaf_check(dirleaf) != NFX_OK)
    {
        return;
    }
    
    for (loop = 0 ; loop < FS_DIRLEAF_ENTRIES_MAX; loop++)
    {
        if (dirleaf->u[loop].ent.len > FS_NAME_MAX)
        {
            fsdb_dumpError("dirleaf key %3d : error: dirent.len (%d) > FS_NAME_MAX (%d)\n",
                           loop,
                           dirleaf->u[loop].ent.len, 
                           FS_NAME_MAX);
            return;
        }

        fsdb_dumpLine("dirleaf - dir %3d : inum                      %lld (0x%llx)\n",
                      loop,
                      dirleaf->u[loop].ent.inum,
                      dirleaf->u[loop].ent.inum);
        fsdb_dumpLine("dirleaf - dir %3d : hash                      %d (0x%x)\n",
                      loop,
                      dirleaf->u[loop].ent.hash,
                      dirleaf->u[loop].ent.hash);
        fsdb_dumpLine("dirleaf - dir %3d : cont - len                %d - %d\n",
                      loop,
                      dirleaf->u[loop].ent.cont,
                      dirleaf->u[loop].ent.len);
        fsdb_dumpLine("dirleaf - dir %3d : aliasSeq                  %d (0x%x)\n",
                      loop,
                      dirleaf->u[loop].ent.aliasSeq,
                      dirleaf->u[loop].ent.aliasSeq);
        fsdb_dumpLine("dirleaf - dir %3d : aliasCksumRand            %d (0x%x)\n",
                      loop,
                      dirleaf->u[loop].ent.aliasCksumRand,
                      dirleaf->u[loop].ent.aliasCksumRand);

        len = MIN(FSDB_MAX_NAME_ENTRY_LEN - 1, dirleaf->u[loop].ent.len);
        memcpy(name, dirleaf->u[loop].ent.name, len);
        name[len] = 0;
        fsdb_dumpLine("dirleaf - dir %3d : name                      '%s'\n",
                      loop,
                      name);

        fsdb_dumpLine("\n");
        used++;
    }
    
    fsdb_dumpLine("\n");

    fsdb_dumpLine("dirleaf - parent_inum                         %d (0x%x)\n",
                  dirleaf->parent_inum,
                  dirleaf->parent_inum);
    fsdb_dumpLine("dirleaf - magic                               %d (0x%x)\n",
                  dirleaf->magic,
                  dirleaf->magic);
    fsdb_dumpLine("dirleaf - count                               %d (0x%x)\n",
                  dirleaf->count,
                  dirleaf->count);
    fsdb_dumpLine("dirleaf - freelist                            %d (0x%x)\n",
                  dirleaf->freelist,
                  dirleaf->freelist);
    fsdb_dumpLine("dirleaf - FS_DIRLEAF_ENTRIES_MAX              %d (0x%x)\n",
                  FS_DIRLEAF_ENTRIES_MAX,
                  FS_DIRLEAF_ENTRIES_MAX);

}


void 
fsdb_dumpDirleafCompact (fs_DirleafCompact *dirleaf)
{
    int  loop;
    int  used = 0;
    int  cont;
    int  index;
    int  leaf;
    int  len;
#define FSDB_MAX_NAME_ENTRY_LEN  34
    char name[FSDB_MAX_NAME_ENTRY_LEN];
    
    
    for (loop = 0 ; loop < FS_DIRLEAF_COMPACT_ENTRIES_MAX; loop++)
    {
        if (dirleaf->entries[loop].len > sizeof(dirleaf->entries[loop].name))
        {
            fsdb_dumpError("dirleaf key %3d : error: dirent.len (%d) too big (%d)\n",
                           loop,
                           dirleaf->entries[loop].len, 
                           sizeof(dirleaf->entries[loop].name));
            return;
        }

        fsdb_dumpLine("dirleaf - dir %3d : inum                      %lld (0x%llx)\n",
                      loop,
                      dirleaf->entries[loop].inum,
                      dirleaf->entries[loop].inum);
        fsdb_dumpLine("dirleaf - dir %3d : len                       %d\n",
                      loop,
                      dirleaf->entries[loop].len);
        fsdb_dumpLine("dirleaf - dir %3d : aliasSeq                  %d (0x%x)\n",
                      loop,
                      dirleaf->entries[loop].aliasSeq,
                      dirleaf->entries[loop].aliasSeq);
        fsdb_dumpLine("dirleaf - dir %3d : aliasCksumRand            %d (0x%x)\n",
                      loop,
                      dirleaf->entries[loop].aliasCksumRand,
                      dirleaf->entries[loop].aliasCksumRand);

        len = MIN(FSDB_MAX_NAME_ENTRY_LEN - 1, dirleaf->entries[loop].len);
        memcpy(name, dirleaf->entries[loop].name, len);
        name[len] = 0;
        fsdb_dumpLine("dirleaf - dir %3d : name                      '%s'\n",
                      loop,
                      name);

        fsdb_dumpLine("\n");
        used++;
    }
    
    fsdb_dumpLine("\n");

    fsdb_dumpLine("dirleaf - parent_inum                         %d (0x%x)\n",
                  dirleaf->parent_inum,
                  dirleaf->parent_inum);
    fsdb_dumpLine("dirleaf - magic                               %d (0x%x)\n",
                  dirleaf->magic,
                  dirleaf->magic);
    fsdb_dumpLine("dirleaf - count                               %d (0x%x)\n",
                  dirleaf->count,
                  dirleaf->count);
    fsdb_dumpLine("dirleaf - FS_DIRLEAF_COMPACT_ENTRIES_MAX      %d (0x%x)\n",
                  FS_DIRLEAF_COMPACT_ENTRIES_MAX,
                  FS_DIRLEAF_COMPACT_ENTRIES_MAX);

}


int
fsdb_dumpEmapleaf (fs_Emapleaf *emapleaf)
{
    int loop, index;
    int validEmapleaf = TRUE;
    
    fsdb_dumpLine("Emap leaf page - nkeys                        %d (0x%x)\n",
                  emapleaf->nkeys,
                  emapleaf->nkeys);

    if (emapleaf->nkeys >= FS_EMAP_EXTENTS_MAX)
    {
        fsdb_dumpError("Emap leaf page - nkeys %d >= FS_EMAP_EXTENTS_MAX %d\n",
                       emapleaf->nkeys, FS_EMAP_EXTENTS_MAX);
        return FALSE;
    }

    fsdb_dumpLine("Emap leaf page: index values [0..%d]\n", emapleaf->nkeys);
    
    for (loop = 0 ; loop < emapleaf->nkeys; loop++)
    {
        if (loop && ((loop % 15) == 0))
            fsdb_dumpLine ("\n");
        
        fsdb_dumpLine("%3d ", emapleaf->index[loop]);
    }
    fsdb_dumpLine ("\n");
    
    for (loop = 0 ; loop < emapleaf->nkeys; loop++)
    {
        index = emapleaf->index[loop];

        fsdb_dumpLine("Emap leaf page: index %3d %3d : offset  len start  %lld %lld %lld\n",
                      loop, index,
                      emapleaf->offsets[index],
                      emapleaf->extents[index].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                      emapleaf->extents[index].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        fsdb_dumpLine("                                offset  len start  0x%llx 0x%llx 0x%llx\n",
                      emapleaf->offsets[index],
                      emapleaf->extents[index].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                      emapleaf->extents[index].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        /*
        fsdb_dumpLine("Emap leaf page: index %3d : index             %d (0x%x)\n",
                      loop,
                      emapleaf->index[loop],
                      emapleaf->index[loop]);

        fsdb_dumpLine("Emap leaf page: index %3d : offset            %lld (0x%llx)\n",
                      loop,
                      emapleaf->offsets[loop],
                      emapleaf->offsets[loop]);
        fsdb_dumpLine("Emap leaf page: index %3d : length and        %lld (0x%llx)\n",
                      loop,
                      emapleaf->extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                      emapleaf->extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS);
        fsdb_dumpLine("Emap leaf page: index %3d :        and start  %lld (0x%llx)\n",
                      loop,
                      emapleaf->extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK,
                      emapleaf->extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        */
    }
    
    return (validEmapleaf);
}

void 
fsdb_dumpDinodePage (fs_DinodePage *dinodePage)
{
    int loop;
    int used = 0;
    
    for (loop = 0 ; loop < FS_DINODE_PAGE_MAX; loop++)
    {
        if (dinodePage->dinode[loop].attrs & FS_IFMT_MASK)
        {
            printf("Dinode page %d\n", loop);

            if (fsdb_volume.offset == FS_BLOCK_SIZE)
            {
                switch(loop)
                {
                case FS_INODE_INODE:
                    printf("FS_INODE_INODE\n");
                    break;
                case FS_ROOT_INODE:
                    printf("FS_ROOT_INODE\n");
                    break;
                case FS_REF_COUNT_INODE:
                    printf("FS_REF_COUNT_INODE\n");
                    break;
                case FS_FREE_BUDDY_TREE_INODE:
                    printf("FS_FREE_BUDDY_TREE_INODE\n");
                    break;
                case FS_SECURITY_INODE:
                    printf("FS_SECURITY_INODE\n");
                    break;
                case FS_FULL_INODE_BLOCK_INODE:
                    printf("FS_FULL_INODE_BLOCK_INODE\n");
                    break;
                case FS_DIRECTORY_IN_INODE_BLOCK_INODE:
                    printf("FS_DIRECTORY_IN_INODE_BLOCK_INODE\n");
                    break;
                case FS_FILESET_INFO_INODE:
                    printf("FS_FILESET_INFO_INODE\n");
                    break;
                case FS_FILESET_USER_QUOTA_INODE:
                    printf("FS_FILESET_USER_QUOTA_INODE\n");
                    break;
                case FS_TRASHCAN_INODE:
                    printf("FS_TRASHCAN_INODE\n");
                    break;
                case FS_SNAPSHOT_DIR_INODE:
                    printf("FS_SNAPSHOT_DIR_INODE\n");
                    break;
                case FS_SNAPSHOT_BITS_INODE:
                    printf("FS_SNAPSHOT_BITS_INODE\n");
                    break;
                case FS_LINK_INODE:
                    printf("FS_LINK_INODE\n");
                    break;
                case FS_TEMPDIR_INODE:
                    printf("FS_TEMPDIR_INODE\n");
                    break;
                }
            }
            
            fsdb_dumpDinode (&dinodePage->dinode[loop]);
            used++;
            printf("\n");
        }
    }

    fsdb_dumpLine("Dinode page - FS_DINODE_PAGE_MAX %3d - Used entries %3d\n",
                  FS_DINODE_PAGE_MAX, used);
}

void 
fsdb_dumpFileSetQuotaRecord (char *str, fs_QuotaRecord_t *quota)
{
    fsdb_dumpLine("%s Inode soft limit        %lld (0x%llx)\n",
                  str, quota->inodeSoftLimit, quota->inodeSoftLimit);
    fsdb_dumpLine("%s Inode hard limit        %lld (0x%llx)\n",
                  str, quota->inodeHardLimit, quota->inodeHardLimit);
    fsdb_dumpLine("%s Inode time limit        %lld (0x%llx)\n",
                  str, quota->inodeTimeLimit, quota->inodeTimeLimit);
    fsdb_dumpLine("%s Nbr of block soft limit %lld (0x%llx)\n",
                  str, quota->blockSoftLimit, quota->blockSoftLimit);
    fsdb_dumpLine("%s Nbr of block hard limit %lld (0x%llx)\n",
                  str, quota->blockHardLimit, quota->blockHardLimit);
    fsdb_dumpLine("%s Nbr of block time limit %lld (0x%llx)\n",
                  str, quota->blockTimeLimit, quota->blockTimeLimit);
}

void 
fsdb_dumpFileSetConsumptionRecord (char *str, fs_ConsumptionRecord_t *consumption)
{
    fsdb_dumpLine("%s Inodes consumed         %lld (0x%llx)\n",
                  str, consumption->inodes, consumption->inodes);
    fsdb_dumpLine("%s Blocks consumed         %lld (0x%llx)\n",
                  str, consumption->blocks, consumption->blocks);
    fsdb_dumpLine("%s Time inode soft exceed  %lld (0x%llx)\n",
                  str, consumption->timeExtraInodes, consumption->timeExtraInodes);
    fsdb_dumpLine("%s time block soft exceed  %lld (0x%llx)\n",
                  str, consumption->timeExtraBlocks, consumption->timeExtraBlocks);
}

void
fsdb_dumpFileSet (fs_FileSetDiskRecord_t *fileset)
{
    fsdb_dumpLine("Fileset : parentId                            %d (0x%x)\n",
                  fileset->parentId, fileset->parentId);
    fsdb_dumpLine("Fileset : root Inode                          %lld (0x%llx)\n",
                  fileset->rootInode, fileset->rootInode);
    fsdb_dumpLine("Fileset : flags                               %d (0x%x)\n",
                  fileset->fileSetFlags, fileset->fileSetFlags);

    fsdb_dumpFileSetQuotaRecord ("Fileset default aggr ", &fileset->defaultAggregateQuota);
    fsdb_dumpFileSetQuotaRecord ("Fileset quota        ", &fileset->fileSetQuota);
    
    fsdb_dumpFileSetConsumptionRecord ("Fileset consumption  ", &fileset->fileSetConsumption);

    fsdb_dumpLine("Fileset : time last modifications             %s",
                  ctime((time_t *)&fileset->lastChanged));
    fsdb_dumpLine("Fileset : bandwitdh                           %d (0x%x)\n",
                  fileset->bandWidth, fileset->bandWidth);
}


int
fsdb_dumpBuddy (uint32 *buddy)
{
    uint64   blockNbr;
    uint64   region;
    uint32   freeBlocks;
    int      level;
    int      index;
    int      word;
    int      mask;
    int      count;
    uint     base;
    uint     dumpBase;
    uint     subBase;
    uint     blank;
    uint     loop;
    int      freeCount[FS_MAX_BUDDY_LEVELS + 1];
    int      errorCnt = 0;
    int      validBuddyTree = TRUE;
    
    blockNbr = fsdb_volume.offset / FS_BLOCK_SIZE;
    region = blockNbr >> FS_MAX_BUDDY_LEVELS;
    // offsetInRegion = blockNbr & ((1 << FS_MAX_BUDDY_LEVELS) - 1);

    fsdb_dumpLine("\nBuddy Block %lld (0x%llx) : Region %lld (0x%llx)\n",
                  blockNbr, blockNbr, region, region);

    if ((buddy[0] & 0x01) != 0)
    {
        fsdb_dumpError("Buddy page: the first bit should be 0, but is 1 !!!!!\n");
        validBuddyTree = FALSE;
    }

    freeBlocks = 0;
    
    for (level = FS_MAX_BUDDY_LEVELS ; level >= 0; level--) 
    {
        count = 0;
        
        base = 1 << (FS_MAX_BUDDY_LEVELS - level);

        for (loop = 0; loop < base; loop++)
        {
            index = base + loop;
            word = index / BITS_PER_WORD;
            mask = 1 << (index & (BITS_PER_WORD - 1));
            if ((buddy[word] & mask) != 0)
            {
                int  level_1, loop_1, base_1, index_1;

                count++;

                // for debugging: check that all bits below tree are 0
                if (level > 0)
                {
                    for (level_1 = level - 1; level_1 >= 0; level_1--)
                    {
                        base_1 = 1 << (level - level_1);
                        for (loop_1 = 0; loop_1 < base_1; loop_1++)
                        {
                            index_1 = (index << (level - level_1)) + loop_1;
                            word = index_1 / BITS_PER_WORD;
                            mask = 1 << (index_1 & (BITS_PER_WORD - 1));
                            if ((buddy[word] & mask) != 0)
                            {
                                fsdb_dumpError("Buddy page error: tree under level %d off %d index %d is not all 0\n", 
                                               level, loop, index);
                                fsdb_dumpError("                             level %d off %d index %d is set to 1 \n", 
                                               level_1, loop_1, index_1);

                                validBuddyTree = FALSE;
                                errorCnt++;
                                if (errorCnt > 10)
                                {
                                    return FALSE;
                                }
                            }
                        }
                    }
                }
            }
        }
        freeCount[level] = count;
        freeBlocks += count * (1 << level);
        //fsdb_dumpLine("Buddy free count for level %d : %d\n",
        //              level, count);
    }
   
    if (freeBlocks > (1 << FS_MAX_BUDDY_LEVELS))
    {
        fsdb_dumpError("Nbr of free blocks in this region: %d > Total: %d\n", 
                       freeBlocks, 1 << FS_MAX_BUDDY_LEVELS);
        validBuddyTree = FALSE;
        return (validBuddyTree);
    }
    if ((buddy[0] & 0x01) != 0)
    {
        validBuddyTree = FALSE;
        return (validBuddyTree);
    }
    

#define FSDB_DUMP_BUDDY_SPLIT   8 

    printf("Lev - free - in use buddy bits\n");
    for (level = FS_MAX_BUDDY_LEVELS ; level >= 0; level--) 
    {
        base = 1 << (FS_MAX_BUDDY_LEVELS - level);
        printf("%3d - %4d - ", level, freeCount[level]);
        for (loop = 0; loop < base; loop++)
        {
            if (fsdb_local_dump.showBuddyBits)
                if ((loop > 0) && ((loop % 128) == 0))
                    printf("\n           - ");

            index = base + loop;
            word = index / BITS_PER_WORD;
            mask = 1 << (index & (BITS_PER_WORD - 1));

            if (level >= FSDB_DUMP_BUDDY_SPLIT)
            {
                if ((buddy[word] & mask) != 0)
                    printf("1");
                else
                    printf("0");

                for (blank = 0 ; blank < ((1<<(level-FSDB_DUMP_BUDDY_SPLIT))-1); blank++)
                    printf(" ");
            }
            else
            {
                if (fsdb_local_dump.showBuddyBits)
                {
                    if ((buddy[word] & mask) != 0)
                        printf("1");
                    else
                        printf(".");
                }
                else
                {
                    if ((buddy[word] & mask) != 0)
                        printf("- %4d (%5d) ", loop, index);
                }
            }
        }
        printf("\n");
    }

    printf("\n");
    printf("Nbr of free blocks in this region: %d Total: %d\n", freeBlocks, 1 << FS_MAX_BUDDY_LEVELS);
    printf("\n");

    {
#define MAX_BLOCK_IN_BUDDY_REGION   (32*1024)

        int freeBlockTab[MAX_BLOCK_IN_BUDDY_REGION];
        int i;
        int nbFreeBlock = 0;
        int CurrentState = 0;
        int line = 0;

        bzero(&freeBlockTab, MAX_BLOCK_IN_BUDDY_REGION*sizeof(int));
        
        printf("Buddy allocations:\n");
        for (level = FS_MAX_BUDDY_LEVELS ; level >= 0; level--) 
        {
            base = 1 << (FS_MAX_BUDDY_LEVELS - level);
            for (loop = 0; loop < base; loop++)
            {
                index = base + loop;
                word = index / BITS_PER_WORD;
                mask = 1 << (index & (BITS_PER_WORD - 1));

                if ((buddy[word] & mask) != 0)
                {
                    int j;
                    
                    for (j = 0; j < (1<<level); j++)
                    {
                        if (freeBlockTab[j+loop*(1<<level)] != 0)
                        {
                            printf("Aaaaaaaaaarrrggghhhhhhhhh\n");
                            return(FALSE);
                        }
                        freeBlockTab[j+loop*(1<<level)] = 1;
                    }
                }
                
            } // for (loop)
        } // for (level)
        
        CurrentState = freeBlockTab[0];
        printf("%5d - ",0);
        for (i = 1; i < MAX_BLOCK_IN_BUDDY_REGION; i++)
        {
            if (freeBlockTab[i] != CurrentState)
            {
                printf("%5d : %s\n", i-1,freeBlockTab[i-1]? "free" : "full");
                line++;
                if (line >= 20)
                {
                    boolean rc = fsdb_continue();
                    line = 0;
                    if (rc == FALSE)
                        break;
                }
                CurrentState = freeBlockTab[i];
                printf("%5d - ", i);
            }
            if (freeBlockTab[i] == 1)
            {
                nbFreeBlock++;
            }
        } // for
        printf("%5d : %s\n", MAX_BLOCK_IN_BUDDY_REGION,
               CurrentState? "free" : "full");
        printf("\n\n");

        if (freeBlocks != nbFreeBlock)
        {
            return(FALSE);
        }
    }

    return (validBuddyTree);
}

void   fsdb_dumpFullInode (char8 *data)
{
    int index, word, mask;
    int free = 0;
    
    for (index = 0; index < (FS_BLOCK_SIZE * BITS_PER_BYTE); index++)
    {
        word = index / BITS_PER_BYTE;
        mask = 1 << (index & (BITS_PER_BYTE - 1));
        if ((data[word] & mask) == 0)
        {
            if (free && ((free % 10) == 0))
                printf("\n");
            
            free++;
            printf ("%5d ", index);
        }
    }
    
    printf("\n\n");
    printf("Nbr of inode pages with free entries: %d Total: %d\n", free, (FS_BLOCK_SIZE * BITS_PER_BYTE));
    printf("\n");
}

void   fsdb_dumpDirInInode (char8 *data)
{
    int index, word, mask;
    int noDirInInode = 0;
    
    for (index = 0; index < (FS_BLOCK_SIZE * BITS_PER_BYTE); index++)
    {
        word = index / BITS_PER_BYTE;
        mask = 1 << (index & (BITS_PER_BYTE - 1));
        if ((data[word] & mask) == 0)
        {
            if (noDirInInode && ((noDirInInode % 10) == 0))
                printf("\n");
            
            noDirInInode++;
            printf ("%5d ", index);
        }
    }
    
    printf("\n\n");
    printf("Nbr of inode pages with no directory entry: %d Total: %d\n", 
           noDirInInode, (FS_BLOCK_SIZE * BITS_PER_BYTE));
    printf("\n");
}


int
fsdb_dumpVollogBuf (vl_BufDiskHdr *logBuf)
{
    vl_LogHdr *logHdr = (vl_LogHdr*)(logBuf + 1);    
    uint32 savedCksum, newCksum;
    int  rc = TRUE;

    fsdb_dumpLine("\n");
    if (logHdr->version == VL_LOG_VERSION_ONE)
    {
        // this is the first buffer in the log
        fsdb_dumpLine("Volume log buf header version            0x%x\n", logHdr->version);
        fsdb_dumpLine("Volume log buf header timestamp (mkfs)   %s", ctime((time_t *)&logHdr->timeStamp));
        fsdb_dumpLine("Volume log buf header log init time      %s", ctime((time_t *)&logHdr->logInitTimeSec));
        fsdb_dumpLine("Volume log buf header log init time nsec %d\n", logHdr->logInitTimeNsec);
        fsdb_dumpLine("Volume log buf header logSize            %d (0x%x)\n", logHdr->logSize, logHdr->logSize);
        fsdb_dumpLine("Volume log buf header bufSize            %d (0x%x)\n", logHdr->bufSize, logHdr->bufSize);
        savedCksum = logHdr->cksum;
        logHdr->cksum = 0 ;
        newCksum = uint32_checksum(logHdr, sizeof(vl_LogHdr), 0);
        logHdr->cksum = savedCksum;
        if (savedCksum != newCksum)
        {
            fsdb_dumpError("Volume log buf header has wrong CRC: 0x%x (should be 0x%x)\n",
                           savedCksum, newCksum);
            fsdb_dumpError("... continue\n");
            rc = FALSE;
        }
        else
        {
            fsdb_dumpLine("Volume log buf header cksum              0x%x\n", logHdr->cksum);
        }
        fsdb_dumpLine("\n");
    }

    if (logBuf->magic != VL_LOG_MAGIC)
    {
        fsdb_dumpError("Vol log buffer has wrong magic: 0x%x (should be 0x%x)\n",
               logBuf->magic, VL_LOG_MAGIC);
        fsdb_dumpData (logBuf, fsdb_volume.offset, sizeof(vl_BufDiskHdr));
        return FALSE;
    }
    else
    {
        fsdb_dumpLine("Volume log buf Magic                     0x%x\n", logBuf->magic);
    }

    if (fsdb_volume.mkfsTime &&
        (logBuf->timeStamp != fsdb_volume.mkfsTime))
    {
        fsdb_dumpLine("Volume log buf wrong timestamp (mkfs)    %s",
                      ctime((time_t *)&logBuf->timeStamp));
        fsdb_dumpLine(" - Must be: %s", 
                      ctime((time_t *)&fsdb_volume.mkfsTime));
        // rc = FALSE;
    }
    else
    {
        fsdb_dumpLine("Volume log buf timestamp (mkfs)          %s", ctime((time_t *)&logBuf->timeStamp));
    }
    
    fsdb_dumpLine("Volume log buf log init time             %s", ctime((time_t *)&logBuf->logInitTimeSec));
    fsdb_dumpLine("Volume log buf log init time nsec        %d\n", logBuf->logInitTimeNsec);

    fsdb_dumpLine("Volume log buf ioWriteTime               %s", ctime((time_t *)&logBuf->ioWriteTime));

    fsdb_dumpLine("Volume log buf Id (incr. nbr)            %d (0x%x)\n", logBuf->bufId, logBuf->bufId);
    if ((logBuf->size >= sizeof(vl_BufDiskHdr)) && (logBuf->size <= VL_BUFFER_SIZE))
    {
        savedCksum = logBuf->cksum;
        logBuf->cksum = 0 ;
        newCksum = uint32_checksum(logBuf, logBuf->size, 0);
        logBuf->cksum = savedCksum;
        if (savedCksum != newCksum)
        {
            fsdb_dumpError("Volume log buf has wrong CRC: 0x%x (should be 0x%x)\n",
                           savedCksum, newCksum);
            fsdb_dumpError("... continue\n");
            rc = FALSE;
        }
        else
        {
            fsdb_dumpLine("Volume log buf cksum                     0x%x\n", logBuf->cksum);
        }
        fsdb_dumpLine("Volume log buf size                      %d (0x%x)\n", logBuf->size, logBuf->size);
    }
    else
    {
        fsdb_dumpError("Volume log buf has invalid lenght %d (0x%x) [%d..%d]\n",
                       logBuf->size, logBuf->size, sizeof(vl_BufDiskHdr), VL_BUFFER_SIZE);
        fsdb_dumpError("... continue\n");
        rc = FALSE;
    }
    fsdb_dumpLine("Volume log buf head                      %d (0x%x)\n", logBuf->head, logBuf->head);
    fsdb_dumpLine("Volume log buf tail                      %d (0x%x)\n", logBuf->tail, logBuf->tail);
    fsdb_dumpLine("Volume log buf blksUsed                  %lld (0x%llx)\n", logBuf->blksUsed, logBuf->blksUsed);
    fsdb_dumpLine("Volume log buf inodesUsed                %lld (0x%llx)\n", logBuf->inodesUsed, logBuf->inodesUsed);
    fsdb_dumpLine("\n");

    return (rc);
}

vl_BufDiskHdr *
fsdb_readVolBuf(uint64 offset, uint32 size)
{
    vl_BufDiskHdr *buf;
    uint32 savedCksum, newCksum;
    
    if ( ! (fsdb_volume.dataAvailable &&
            (offset == fsdb_volume.offset)))
    {
        fsdb_do_read(offset, size);
    }
    
    if ( ! fsdb_volume.dataAvailable)
        return NULL;

    buf = (vl_BufDiskHdr*)&fsdb_volume.data;
    if (buf->magic != VL_LOG_MAGIC)
    {
        printf("vl_readBuf wrong Magic 0x%x (must be 0x%x) offset %lld\n", buf->magic, VL_LOG_MAGIC, offset);
        return NULL;
    }

    if (buf->timeStamp != fsdb_volume.mkfsTime)
    {
        printf ("vl_readBuf wrong timestamp %s Must be: %s\n",
                ctime((time_t *)&buf->timeStamp),  ctime((time_t *)&fsdb_volume.mkfsTime));
        return NULL;
    }

    savedCksum = buf->cksum;
    buf->cksum = 0 ;
    newCksum = uint32_checksum(buf, buf->size, 0);
    buf->cksum = savedCksum;
    if (savedCksum != newCksum)
    {
        fsdb_dumpError("Volume log buf header has wrong CRC: 0x%x (should be 0x%x)\n",
                       savedCksum, newCksum);
        return NULL;
    }

    return (buf);
}


#define  VL_GET_NBR_OF_BUF(_logSize_, _bufSize_)                            \
       ((((_logSize_ - 1) + sizeof(vl_LogHdr)) / (_bufSize_ - sizeof (vl_BufDiskHdr))) + 1)

uint64                                            // return offset or -1 if not found
fsdb_findMostRecentLogBuf (uint64 startOffset, uint32 logSize, uint32 bufSize)
{

    uint32          step, dir, cnt, found, vl_max_nbr_probes ; 
    uint32          bufNbr_1, bufNbr_2;
    uint32          bufId_1, bufId_2;
    vl_BufDiskHdr   *buf = NULL;
    uint32          nbrOfBufsInVolLog;
    vl_VolBufRoot   *volBufRoot;
    vl_VolBuf       *currBuf;
    uint32          bufNbr, bufNbrLast;
    evm_volOffset_t volOffset, bufOffset;

    nbrOfBufsInVolLog = VL_GET_NBR_OF_BUF(logSize, bufSize);

    vl_max_nbr_probes = 30;
    // 5 is a safety margin

// should never see more than 1 CRC error contiguously on one log device 
#define VL_MAX_NBR_RETRIES   2
#define VL_RIGHT   1
#define VL_LEFT    -1

    cnt = 0 ;
    bufNbr_1 = 0; 

    do
    {
        buf = fsdb_readVolBuf(startOffset + bufNbr_1 * bufSize, bufSize);
        cnt++;
    } while ((buf == NULL) && (bufNbr_1++ < VL_MAX_NBR_RETRIES));

    if (buf == NULL)
    {
        // totally empty log
        // replay finished
        printf ("fsdb_findMostRecentBuf, failed finding good log entries\n");
        return (-1);
    }
    
    dir = VL_RIGHT ;
    step = (nbrOfBufsInVolLog - cnt) / 2;
    bufNbr_2 = step;
    cnt = 0 ;

    bufId_1 = buf->bufId ;
    bufId_2 = bufId_1;  // not needed, just to fix a compiler warning
    
    found = FALSE;
    while ( ! found)
    {
        buf = fsdb_readVolBuf(startOffset + bufNbr_2 * bufSize, bufSize);
        cnt++;
        if (cnt > vl_max_nbr_probes)
            return -1;
        
        if (buf != NULL)
        {
            bufId_2 = buf->bufId;

            if (bufId_2 > bufId_1)
            {
                // normally, 2 more recent
        
                if (bufId_2 > (bufId_1 + VL_BUF_ID_WRAP))
                {
                    // bufId wrap around, 1 is more recent
                    dir = VL_LEFT;
                }
                else
                {
                    // 2 more recent
                    dir = VL_RIGHT;
                }
            } 
            else 
            {
                // normally, 1 more recent
        
                if (bufId_1 > (bufId_2 + VL_BUF_ID_WRAP))
                {
                    // recId wrap around, 2 is more recent
                    dir = VL_RIGHT;
                }
                else
                {
                    // 1 more recent
                    dir = VL_LEFT ;
                }
            }
        }
        else
        {
            dir = VL_LEFT ;
        }

        step = (bufNbr_2 - bufNbr_1);

        if (dir == VL_RIGHT)
        {
            bufNbr_1 = bufNbr_2;
            bufId_1 = bufId_2;
        }
        
        step = (step + 1) / 2;

        if (step == 1 && dir == VL_RIGHT && bufNbr_2 == (nbrOfBufsInVolLog - 1))
        {
            step = 0;
        }
        
        bufNbr_2 = bufNbr_2 + step * dir;

        if (bufNbr_2 >= nbrOfBufsInVolLog)
        {
            bufNbr_2 = nbrOfBufsInVolLog - 1;
        }

        if (bufNbr_1 == bufNbr_2)
            found = TRUE;
         
        printf("fsdb_findMostRecentBuf: buf %d %d %d %d %d\n", 
               bufNbr_1, bufNbr_2, bufId_1, bufId_2, dir);
     }

    buf = fsdb_readVolBuf(startOffset + bufNbr_1 * bufSize, bufSize);

    bufNbr = (buf->tail + sizeof (vl_LogHdr)) / (bufSize - sizeof (vl_BufDiskHdr));

    if (bufNbr != bufNbr_1)
    {
        bufNbrLast = ((logSize - 1) + sizeof (vl_LogHdr)) / 
				                                (bufSize - sizeof (vl_BufDiskHdr));

        if ((bufNbr == (bufNbr_1 - 1)) ||
            ((bufNbr_1 == 0) && (bufNbr == bufNbrLast)))
        {
            // this means that the log tail is in the prev buffer

            printf("fsdb_findMostRecentBuf: get previous most recent buf %d (%d %d)\n", 
                   bufNbr, bufNbr_1, buf->tail);
            bufNbr_1 = bufNbr;
            
            buf = fsdb_readVolBuf(startOffset + bufNbr_1 * bufSize, bufSize);

            if (buf == NULL)
            {
                printf ("fsdb_findMostRecentBuf: no valid buf found %d %d\n", 
                        bufNbr_1, buf->tail);
                return -1;
            }


        }
        else
        {
            printf ("fsdb_findMostRecentBuf bufNbr does not match %d %d %d\n", 
                    bufNbr, bufNbr_1, buf->tail);
            return -1;
        }
    }
    
    printf ("fsdb found buf index %d in %d total in %d iterations\n",
            bufNbr_1, nbrOfBufsInVolLog, cnt);
    printf ("  head %d tail %d id %d size %d\n",
            buf->head, buf->tail, buf->bufId, buf->size);

    return (startOffset + bufNbr_1 * bufSize);

}  // fsdb_findMostRecentBuf


/**
 * Dump the contents of a log record for debugging.
 * Return the new offset or -1 if FS_LOG_ID_EOL
 * Called with offset 0, will give the first log record.
 *
 * copy from fs-log.c
 */
static uint32
fsdb_logDumpOneRecord(
    char            **str,
    fs_LogRecord    *record)
{
    uint32              recLen;
    char                *s;
    uint32              savedCksum, cksum;
    uint32              loop;

    s = *str;

    s += sprintf(s, " hdr.id %d\n", record->hdr.id);
    s += sprintf(s, " hdr.type %d\n", record->hdr.recType);
    s += sprintf(s, " hdr.flags 0x%x\n", record->hdr.flags);

    switch(record->hdr.recType)
    {
    case FS_LOG_REC_META_ONE:
        s += sprintf(s, "META_ONE type\n");
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
               record->u.metaOne.blkId.inum, record->u.metaOne.blkId.page,
               record->u.metaOne.blkId.block.block);
        s += sprintf(s, "offset %d, ", record->u.metaOne.offset);
        s += sprintf(s, "length %d\n", record->u.metaOne.length);
        //s += fs_logDumpData(record->u.metaOne.data, record->u.metaOne.length, s);
        recLen = alignto(record->u.metaOne.length, FS_LOG_ALIGN_64)
            + sizeof(fs_LogRecMetaOne);
        break;
    case FS_LOG_REC_META_TWO:
        s += sprintf(s, "META_TWO type\n");
        s += sprintf(s, "srcBlkNum %lld\n", record->u.metaTwo.srcBlkNum);
        s += sprintf(s, "dstBlkId inum %lld page %lld block %lld\n",
               record->u.metaTwo.dstBlkId.inum, record->u.metaTwo.dstBlkId.page,
               record->u.metaTwo.dstBlkId.block.block);
        s += sprintf(s, "offset %d, ", record->u.metaTwo.offset);
        s += sprintf(s, "length %d\n", record->u.metaTwo.length);
        //s += fs_logDumpData(record->u.metaTwo.data, record->u.metaTwo.length, s);
        recLen = alignto(record->u.metaTwo.length, FS_LOG_ALIGN_64)
            + sizeof(fs_LogRecMetaOne);
        break;
    case FS_LOG_REC_USER:
        s += sprintf(s, "USER type\n");
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
               record->u.user.blkId.inum, record->u.user.blkId.page,
               record->u.user.blkId.block.block);
        s += sprintf(s, "codHandle 0x%llx\n", record->u.user.codHandle);
        recLen = sizeof(fs_LogRecUser);
        break;
    case FS_LOG_REC_ALLOC:
        s += sprintf(s, "ALLOC type\n");
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
               record->u.alloc.blkId.inum, record->u.alloc.blkId.page,
               record->u.alloc.blkId.block.block);
        if (record->u.alloc.isMeta)
            s += sprintf(s, "META\n");
        else
            s += sprintf(s, "USER\n");
        if (record->u.alloc.isCow)
            s += sprintf(s, "COW\n");
        recLen = sizeof(fs_LogRecAlloc);
        break;
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        s += sprintf(s, "META_ZERO type\n");
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
                     record->u.metaZero.blkId.inum,
                     record->u.metaZero.blkId.page,
                     record->u.metaZero.blkId.block.block);
        s += sprintf(s, "offset %d\n", record->u.metaZero.offset);
        s += sprintf(s, "length %d\n", record->u.metaZero.length);
        recLen = sizeof(fs_LogRecMetaZero); 
       break;
    case FS_LOG_REC_USER_ZERO:
        s += sprintf(s, "USER_ZERO type\n");
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
                     record->u.userZero.blkId.inum,
                     record->u.userZero.blkId.page,
                     record->u.userZero.blkId.block.block);
        s += sprintf(s, "offset in 1st page %d\n", record->u.userZero.offset);
        s += sprintf(s, "nb page %d\n", record->u.userZero.nbPage);
        s += sprintf(s, "action: %s\n", (record->u.userZero.zero?
                                         "zero fill" : "don't zero fill"));
        recLen = sizeof(fs_LogRecUserZero); 
       break;
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
        s += sprintf(s, "ZERO_FILL type\n");
        if (record->u.zero.isMeta)
            s += sprintf(s, "META\n");
        else
            s += sprintf(s, "USER\n");
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
               record->u.zero.blkId.inum, record->u.zero.blkId.page,
               record->u.zero.blkId.block.block);
        s += sprintf(s, "offset %d, ", record->u.zero.offset);
        s += sprintf(s, "length %d\n", record->u.zero.length);
        recLen = sizeof(fs_LogRecZero); 
       break;
#endif // NEW_ZF
    case FS_LOG_REC_COMMIT:
        s += sprintf(s, "COMMIT type\n");
        s += sprintf(s, "numRecs %d\n", record->hdr.numRecs);
        recLen = 0;
        break;
    case FS_LOG_REC_BITMAP:
        s += sprintf(s, "BITMAP type\n");
        s += sprintf(s, "debugOp %d inum 0x%llx extent len %lld block %lld\n", 
                     record->u.bitmap.debugOp.op, 
                     record->u.bitmap.debugOp.u.inum, 
                     record->u.bitmap.debugOp.u.extent.lengthAndStartBlock >> EXTENT_BLOCK_BITS, 
                     record->u.bitmap.debugOp.u.extent.lengthAndStartBlock  & EXTENT_BLOCK_MASK);
        s += sprintf(s, "blkId inum %lld page %lld block %lld\n",
               record->u.bitmap.blkId.inum, record->u.bitmap.blkId.page,
               record->u.bitmap.blkId.block.block);
        s += sprintf(s, "offset %d, length %d, %s\n", 
                        record->u.bitmap.offset, record->u.bitmap.length,
                        ((record->u.bitmap.op == FS_LOG_REC_BITMAP_SET) ? "SET" : "RESET"));
        if (record->u.bitmap.length < FS_LOG_BITMAP_MAX_BYTES)
        {
            for (loop = 0; loop < record->u.bitmap.length; loop++)
            {
                s += sprintf(s, "0x%02x ", record->u.bitmap.bits[loop]);
            }
            s += sprintf(s, "\n");
        }
        recLen = sizeof(fs_LogRecBitmap);
        break;
    case FS_LOG_REC_INODE:
        s += sprintf(s, "INODE type\n");
        s += sprintf(s, "inum %lld\n",
                     record->u.inode.inum);
        recLen = sizeof(fs_LogRecInode);
        break;
    case FS_LOG_REC_INODE_FLUSH:
        s += sprintf(s, "INODE_FLUSH type\n");
        s += sprintf(s, "inum %lld\n",
                     record->u.inodeFlush.inum);
        recLen = sizeof(fs_LogRecInodeFlush);
        break;
    case FS_LOG_REC_BUF_FLUSH:
        s += sprintf(s, "BUF_FLUSH type\n");
        s += sprintf(s, "list %d inum %lld page %lld blknum %lld\n",
                     record->u.bufFlush.bufList,
                     record->u.bufFlush.inum,
                     record->u.bufFlush.startPage,
                     record->u.bufFlush.block.block);
        recLen = sizeof(fs_LogRecBufFlush);
        break;
    case FS_LOG_REC_MARK_DONE:
        s += sprintf(s, "MARK DONE type\n");
        s += sprintf(s, "Mark Transaction %d done\n", record->u.markDone.logId);
        recLen = sizeof(fs_LogRecMarkDone);
        break;
    default:
        s += sprintf(s, "UNKNOWN type\n");
        recLen = 0;
        break;
    }

    recLen += sizeof(fs_LogRecHdr);

    s += sprintf(s, "recLen: %d\n", recLen);
    s += sprintf(s, "\n");

    *str = s;

    return (recLen);
}

int
fsdb_dumpLog (
    uint64 startOffset, 
    uint32 logSize, 
    uint32 bufSize, 
    uint64 *offset, 
    int criteria, 
    uint64 value)
{
    int              bufIndex;
    uint64           logOffset = *offset;
    uint64           volOffset;
    uint64           bufOffset;
    vl_BufDiskHdr   *buf = NULL;
    char             string[2048];
    char            *s = string;
    int              recLen = 0;
    int              match;
    fs_LogRecord    *record;
    char             data[sizeof(fs_LogRecord)];
    uint64           off_1; 
    int              size_1, bufId_1;

    // align to 8 bytes
    logOffset = (logOffset + 7) & ~7;
    if (logOffset != *offset)
    {
        printf ("Aligned log offset to 8 bytes: %lld\n", logOffset);
    }
    *offset = logOffset;

    if (logOffset > (uint64)logSize)
    {
        printf ("fsdb_dumpLog: invalid log offset %lld (max %d)\n", logOffset, logSize);
        return FALSE;
    }

	match = FALSE;
	if (criteria == 0)
		match = TRUE;

    bufIndex = ((uint32)logOffset + sizeof (vl_LogHdr)) / (bufSize - sizeof (vl_BufDiskHdr));
    volOffset = startOffset + logOffset + 
        sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * (bufIndex + 1);
    bufOffset = (volOffset / bufSize * bufSize) ;
                
    if ((buf = fsdb_readVolBuf(bufOffset, bufSize)) == NULL)
    {
        printf ("fsdb_dumpLog: invalid data at log offset %lld buf offset %lld\n", logOffset, bufOffset);
        return FALSE;
    }

    off_1 = volOffset - bufOffset - sizeof(vl_BufDiskHdr);

    record = (fs_LogRecord*)&((vl_Disk_Buffer*)buf)->data[off_1];
 
    if (buf->size > FS_BLOCK_SIZE)
    {
        printf ("fsdb_dumpLog: invalid buf size %d\n", buf->size);
        return FALSE;
    }

    if ((buf->size - off_1) < sizeof(fs_LogRecord))
    {
        bufId_1 = buf->bufId;

        size_1 = buf->size - off_1 - sizeof(vl_BufDiskHdr);
        
        if (size_1 <= 0)
        {
            printf ("fsdb_dumpLog: end of log\n");
            return FALSE;
        }

        // copy the first part into local data buffer
        memcpy(data, &((vl_Disk_Buffer*)buf)->data[off_1], size_1);
        
        if ((logOffset + size_1) >= logSize)
        {
            printf ("Start at beginning of log\n");
            bufIndex = 0;
            off_1 = sizeof (vl_LogHdr);
            volOffset = startOffset + sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr);
        }
        else
        {
            // read next buffer
            bufIndex += 1;
            off_1 = 0 ;
            volOffset = startOffset + logOffset + size_1 + 
                sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * (bufIndex + 1);
        }
        bufOffset = (volOffset / bufSize * bufSize) ;

        if ((buf = fsdb_readVolBuf(bufOffset, bufSize)) == NULL)
        {
            printf ("fsdb_dumpLog: invalid data at log offset %lld buf offset %lld\n", 
                    logOffset, bufOffset);
            return FALSE;
        }
        
        if (buf->size > FS_BLOCK_SIZE)
        {
            printf ("fsdb_dumpLog: invalid buf size %d\n", buf->size);
            return FALSE;
        }

        if ((bufId_1 + 1) != buf->bufId)
        {
            // todo wrap around .....
            printf ("reading wrong bufId  %d %d\n", bufId_1, buf->bufId);
            // todo : do we need to stop here ?? 
            // return FALSE;
        }

        memcpy(&data[size_1], &((vl_Disk_Buffer*)buf)->data[off_1], sizeof(fs_LogRecord) - size_1);
        
        record = (fs_LogRecord*)data;
    }

    s += sprintf(s, "LogRecord at offset %lld\n", logOffset);
    recLen = fsdb_logDumpOneRecord(&s, record);

    // find valid record
    if ( ! ((record->hdr.id >= FS_LOG_ID_MIN) &&
            (record->hdr.id < FS_LOG_ID_MAX) &&
            (record->hdr.recType >= FS_LOG_REC_META_ONE) &&
            (record->hdr.recType < FS_LOG_REC_TYPE_MAX) &&
            (recLen <= (FS_BLOCK_SIZE + sizeof(fs_LogRecord)))))
    {
        *offset += 8;
        if (*offset > logSize)
        {
            *offset = *offset - logSize;
        }
        printf ("Not on record boundary: increment log offset to: %lld, id %d type %d len %d\n", 
                *offset, record->hdr.id, record->hdr.recType, recLen);
        return TRUE;
    }

    switch (criteria)
    {
	case 1: // transId
		if (record->hdr.id == (uint32)(value & 0xffffffff))
			match = TRUE;
		break;
    case 2: // blocknbr
        switch(record->hdr.recType)
        {
        case FS_LOG_REC_META_ONE:
            if (record->u.metaOne.blkId.block.block == value)
                match = TRUE;
            break;
        case FS_LOG_REC_META_TWO:
            if (record->u.metaTwo.dstBlkId.block.block == value)
                match = TRUE;
            break;
        case FS_LOG_REC_USER:
            if (record->u.user.blkId.block.block == value)
                match = TRUE;
            break;
        case FS_LOG_REC_ALLOC:
            if (record->u.alloc.blkId.block.block == value)
                match = TRUE;
            break;
#ifdef NEW_ZF
        case FS_LOG_REC_META_ZERO:
            if (record->u.metaZero.blkId.block.block == value)
                match = TRUE;
            break;
        case FS_LOG_REC_USER_ZERO:
            if (record->u.userZero.blkId.block.block == value)
                match = TRUE;
            break;
#else // NEW_ZF
        case FS_LOG_REC_ZERO_FILL:
            if (record->u.zero.blkId.block.block == value)
                match = TRUE;
            break;
#endif // NEW_ZF
        case FS_LOG_REC_BITMAP:
            if (record->u.bitmap.blkId.block.block == value)
                match = TRUE;
            if (((record->u.bitmap.debugOp.u.extent.lengthAndStartBlock & EXTENT_BLOCK_MASK) <= value) &&
                ((record->u.bitmap.debugOp.u.extent.lengthAndStartBlock  & EXTENT_BLOCK_MASK) + 
                 ((record->u.bitmap.debugOp.u.extent.lengthAndStartBlock >> EXTENT_BLOCK_BITS)) > value))
                match = TRUE;
            break;
        case FS_LOG_REC_BUF_FLUSH:
            if (record->u.bufFlush.block.block == value)
                match = TRUE;
        }
        break;
    case 3: // inum
        switch(record->hdr.recType)
        {
        case FS_LOG_REC_META_ONE:
            if (record->u.metaOne.blkId.inum == value)
                match = TRUE;
            break;
        case FS_LOG_REC_META_TWO:
            if (record->u.metaTwo.dstBlkId.inum == value)
                match = TRUE;
            break;
        case FS_LOG_REC_USER:
            if (record->u.user.blkId.inum == value)
                match = TRUE;
            break;
        case FS_LOG_REC_ALLOC:
            if (record->u.alloc.blkId.inum == value)
                match = TRUE;
            break;
#ifdef NEW_ZF
        case FS_LOG_REC_META_ZERO:
            if (record->u.metaZero.blkId.inum == value)
                match = TRUE;
            break;
        case FS_LOG_REC_USER_ZERO:
            if (record->u.userZero.blkId.inum == value)
                match = TRUE;
            break;
#else // NEW_ZF
        case FS_LOG_REC_ZERO_FILL:
            if (record->u.zero.blkId.inum == value)
                match = TRUE;
            break;
#endif // NEW_ZF
        case FS_LOG_REC_BITMAP:
            if (record->u.bitmap.blkId.inum == value)
                match = TRUE;
            if (record->u.bitmap.debugOp.u.inum == value)
                match = TRUE;
             break;
        case FS_LOG_REC_INODE:
            if (record->u.inode.inum == value)
                match = TRUE;
            break;
        case FS_LOG_REC_INODE_FLUSH:
            if (record->u.inodeFlush.inum == value)
                match = TRUE;
            break;
        case FS_LOG_REC_BUF_FLUSH:
            if (record->u.bufFlush.inum == value)
                match = TRUE;
        }
        break;
	}

    if (match)
    {
        printf ("%s", string);
    }

    *offset += recLen ;
    if (*offset > logSize)
    {
        *offset = *offset - logSize;
    }

    return TRUE;
}



void   
fsdb_dumpLogRecord ()
{
    int               input;
    int               def;
    int               done = FALSE;
    int               bufIndex;
    uint64            logOffset;
    uint64            volOffset;
    uint64            bufOffset;
    uint64            startOffset;
    uint64            tailOffset;
    fs_SuperBlock    *super;
    uint32            logSize;
    vl_BufDiskHdr    *logBuf;
    int               bufSize = VL_BUFFER_SIZE;             // todo: check VL_BUFFER_SIZE !!!!
    uint32            nbrEntries;
    int               validVollogBufHdr = FALSE;
    vl_LogHdr        *logHdr;
    
    bufOffset = 0;

    printf ("Read the super block\n");
    
    fsdb_do_read(bufOffset, bufSize);
    if ( ! fsdb_volume.dataAvailable)
    {
        printf ("Could not read the super block\n");
        return ;
    }
    super = (fs_SuperBlock *)&fsdb_volume.data;
    if ( ! fsdb_dumpSuperBlock (super))
    {
        printf ("Super block not valid\n");
        return ;
    }
    fsdb_volume.dataType = FSDB_DATA_TYPE_SUPER;
    if (super->logDevType != FS_LOG_DEVTYPE_SELF)
    {
        printf ("Volume logging not on disk, can't read\n");
        return ;
    }

    logSize = super->logSize;
    if (super->version < 7)
    {
        startOffset = (super->numBlks + 1) * FS_BLOCK_SIZE;
    }
    else
    {
        startOffset = super->logStartOffset;
    }
    
    fsdb_volume.mkfsTime = super->mkfsTime;

    printf ("Volume logging start at disk offset %lld (0x%llx) block %lld (0x%llx)\n", 
            startOffset, startOffset, super->numBlks + 1, super->numBlks + 1);
    printf ("Volume logging size %d (0x%x)\n\n", 
            logSize, logSize);
    
    if ( ! fsdb_continue())
        return;

    printf ("Read the volume log header\n");
    
    fsdb_do_read(startOffset, bufSize);
    if ( ! fsdb_volume.dataAvailable)
    {
        printf ("Could not read the log header\n");
        return ;
    }
    logBuf = (vl_BufDiskHdr *)&fsdb_volume.data;
    if ( ! fsdb_dumpVollogBuf(logBuf))
    {
        printf ("Volume log header not valid\n");
        return ;
    }
    logHdr = (vl_LogHdr*)(logBuf + 1);    
    bufSize = logHdr->bufSize;
    printf ("Volume log buffer size = %d\n", bufSize);

    fsdb_volume.dataType = FSDB_DATA_TYPE_VOL_LOG_BUF;
    
    if ( ! fsdb_continue())
        return;

    while (! done)
    {
        printf("\n");
        printf("1: Find most recent vol log buffer\n");
        printf("2: Print all kinds of offset for given log offset\n");
        printf("3: Dump next     volume log buffer header\n");
        printf("4: Dump previous volume log buffer header\n");
        printf("5: Dump log records starting from given log offset\n");
        printf("6: Dump log records for givem inum starting from given log offset\n");
        printf("7: Dump log records for given blocknumber starting from given log offset\n");
        printf("8: Dump log records for given transaction id starting from given log offset\n");
        printf("0: Return\n");
        def = 0;
        printf("---> Input [default: %d]: ", def);
        input = def;
        fsdb_fgets_int (&input);

        switch(input)
        {
        case 1:
            tailOffset = fsdb_findMostRecentLogBuf (startOffset, logSize, bufSize);

            if (tailOffset == -1)
            {
                printf("Most recent buffer not found\n");
                break;
            }
            
            if ( ! fsdb_continue())
                break;

            if (fsdb_dumpVollogBuf((vl_BufDiskHdr *)fsdb_volume.data))
                validVollogBufHdr = TRUE;
            
            break;
        case 2:
            printf("Log offset [0..%d]: ", logSize);
            logOffset = 0;
            fsdb_fgets_uint64(&logOffset);
            if ((logOffset >= 0) && (logOffset < logSize))
            {
                bufIndex = (logOffset + sizeof (vl_LogHdr)) / (bufSize - sizeof (vl_BufDiskHdr));
                volOffset = startOffset + logOffset + 
                    sizeof (vl_LogHdr) + sizeof (vl_BufDiskHdr) * (bufIndex + 1);
                bufOffset = (volOffset / FS_BLOCK_SIZE * FS_BLOCK_SIZE) ;
                
                fsdb_do_read(bufOffset, bufSize);
                if (fsdb_volume.dataAvailable) 
                {
                    if (fsdb_dumpVollogBuf ((vl_BufDiskHdr*)fsdb_volume.data))
                        validVollogBufHdr = TRUE;
                }
                
                printf ("\n");
                printf ("Log offset               %lld\n", logOffset);
                printf ("Buffer index             %d\n", bufIndex);                
                printf ("Volume offset            %lld\n", volOffset);
                printf ("Volume start offset      %lld\n", startOffset);
                printf ("Buffer offset            %lld\n", (volOffset / bufSize) * bufSize);
                printf ("Disk block number        %lld\n", (bufOffset / FS_BLOCK_SIZE));
                printf ("\n");
            }
            break;
        case 3:
            if ( ! validVollogBufHdr)
            {
                bufOffset = startOffset;
            }
            else
            {
                bufOffset = fsdb_volume.offset + bufSize;
                if (bufOffset > fsdb_volume.currSize)
                    bufOffset = startOffset;
            }
            bufIndex = (bufOffset - startOffset) / bufSize;
            volOffset = bufOffset + sizeof (vl_BufDiskHdr);
            if (bufIndex == 0)
                volOffset += sizeof (vl_LogHdr);
            logOffset = (volOffset - startOffset) - sizeof (vl_LogHdr) - 
                                 (sizeof (vl_BufDiskHdr) * (bufIndex + 1));

            fsdb_do_read(bufOffset, bufSize);
            if (fsdb_volume.dataAvailable) 
            {
                if (fsdb_dumpVollogBuf ((vl_BufDiskHdr*)fsdb_volume.data))
                    validVollogBufHdr = TRUE;

                printf ("\n");
                printf ("Log offset               %lld\n", logOffset);
                printf ("Buffer index             %d\n", bufIndex);                
                printf ("Volume offset            %lld\n", volOffset);
                printf ("Volume start offset      %lld\n", startOffset);
                printf ("Buffer offset            %lld\n", (volOffset / bufSize) * bufSize);
                printf ("Disk block number        %lld\n", (bufOffset / FS_BLOCK_SIZE));
                printf ("\n");
            }
            break;
        case 4:
            if ( ! validVollogBufHdr)
            {
                bufOffset = startOffset;
            }
            else
            {
                bufOffset = fsdb_volume.offset - bufSize;
                if (bufOffset < startOffset)
                    bufOffset = fsdb_volume.currSize / FS_BLOCK_SIZE * FS_BLOCK_SIZE;
            }
            bufIndex = (bufOffset - startOffset) / bufSize;
            volOffset = bufOffset + sizeof (vl_BufDiskHdr);
            if (bufIndex == 0)
                volOffset += sizeof (vl_LogHdr);
            logOffset = (volOffset - startOffset) - sizeof (vl_LogHdr) - 
                                 (sizeof (vl_BufDiskHdr) * (bufIndex + 1));
            
            fsdb_do_read(bufOffset, bufSize);
            if (fsdb_volume.dataAvailable) 
            {
                if (fsdb_dumpVollogBuf ((vl_BufDiskHdr*)fsdb_volume.data))
                    validVollogBufHdr = TRUE;

                printf ("\n");
                printf ("Log offset               %lld\n", logOffset);
                printf ("Buffer index             %d\n", bufIndex);                
                printf ("Volume offset            %lld\n", volOffset);
                printf ("Volume start offset      %lld\n", startOffset);
                printf ("Buffer offset            %lld\n", (volOffset / bufSize) * bufSize);
                printf ("Disk block number        %lld\n", (bufOffset / FS_BLOCK_SIZE));
                printf ("\n");
            }
            break;
        case 5:
            printf("Log offset [0..%d]: ", logSize);
            logOffset = 0;
            fsdb_fgets_uint64(&logOffset);
            printf("Nbr of entries [1..?]: ");
            nbrEntries = 1;
            fsdb_fgets_uint32(&nbrEntries);

            while ((logOffset >= 0) && 
                   (nbrEntries-- > 0)) 
            {
                if ( ! fsdb_dumpLog(startOffset, logSize, bufSize, &logOffset, 0, 0))
                    break;
            }
            break;
        case 6:
        {
            uint64 inum;
            
            printf("Log offset [0..%d]: ", logSize);
            logOffset = 0;
            fsdb_fgets_uint64(&logOffset);
            printf("Nbr of entries [1..?]: ");
            nbrEntries = 1;
            fsdb_fgets_uint32(&nbrEntries);
            printf("inum:  ");
            inum = 1;
            fsdb_fgets_uint64(&inum);

            while ((logOffset >= 0) && 
                   (nbrEntries-- > 0)) 
            {
                if ( ! fsdb_dumpLog(startOffset, logSize, bufSize, &logOffset, 3, inum))
                    break;
            }
            break;
        }
        case 7:
        {
            uint64 blockNbr;
            
            printf("Log offset [0..%d]: ", logSize);
            logOffset = 0;
            fsdb_fgets_uint64(&logOffset);
            printf("Nbr of entries [1..?]: ");
            nbrEntries = 1;
            fsdb_fgets_uint32(&nbrEntries);
            printf("blockNbr:  ");
            blockNbr = 1;
            fsdb_fgets_uint64(&blockNbr);

            while ((logOffset >= 0) && 
                   (nbrEntries-- > 0)) 
            {
                if ( ! fsdb_dumpLog(startOffset, logSize, bufSize, &logOffset, 2, blockNbr))
                    break;
            }
            break;
        }
        case 8:
        {
            uint64 transId;
            
            printf("Log offset [0..%d]: ", logSize);
            logOffset = 0;
            fsdb_fgets_uint64(&logOffset);
            printf("Nbr of entries [1..?]: ");
            nbrEntries = 1;
            fsdb_fgets_uint32(&nbrEntries);
            printf("transaction Id:  ");
            transId = 1;
            fsdb_fgets_uint64(&transId);

            while ((logOffset >= 0) && 
                   (nbrEntries-- > 0)) 
            {
                if ( ! fsdb_dumpLog(startOffset, logSize, bufSize, &logOffset, 1, transId))
                    break;
            }
            break;
        }
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", input);
            break;
        }
    }
}


/******************************************************************************
 * print the offset/blocknbr
 */

void 
fsdb_dumpOffset(uint64 offset)
{
    printf("Offset: %lld (0x%llx)\n", offset, offset);

    printf("SDM blocks (%4d) %lld", SDM_BLOCK_SIZE, offset / SDM_BLOCK_SIZE);
    if ((offset % SDM_BLOCK_SIZE) != 0)
    {
        printf ("(off in block %lld)", offset % SDM_BLOCK_SIZE);
    }
    printf("\n");

    printf("FS  blocks (%4d) %lld", FS_BLOCK_SIZE,  offset / FS_BLOCK_SIZE);
    if ((offset % FS_BLOCK_SIZE) != 0)
    {
        printf ("(off in block %lld)", offset % FS_BLOCK_SIZE);
    }
    printf("\n");
    
}







/******************************************************************************
 * setup dump output format 
 */

void
fsdb_set_debug (fsdb_SetDumpAction action)
{
    int      done = FALSE;

    switch (action)
    {
    case FSDB_SET_DUMP_DEFAULT:
        fsdb_default_dump.level = FSDB_DUMP_LEVEL_MAXIMUM;
        fsdb_default_dump.size = FSDB_DUMP_1_BYTE;
        fsdb_default_dump.format = FSDB_DUMP_HEX;
        fsdb_default_dump.showBuddyBits = FALSE;
        strcpy(fsdb_default_dump.typeStr, "hex");

        fsdb_hex_byte_dump.level = FSDB_DUMP_LEVEL_MAXIMUM;
        fsdb_hex_byte_dump.size = FSDB_DUMP_1_BYTE;
        fsdb_hex_byte_dump.format = FSDB_DUMP_HEX;
        fsdb_hex_byte_dump.showBuddyBits = FALSE;
        strcpy(fsdb_hex_byte_dump.typeStr, "hex");

        fsdb_dump = &fsdb_default_dump;
        break;
    case FSDB_SET_DUMP_LOCAL:
        done = FALSE;
        while ( ! done)
        {
            printf("DUMP level (min = 0, max = %d): ", FSDB_DUMP_LEVEL_MAX - 1);
            fsdb_local_dump.level = FSDB_DUMP_LEVEL_MAX - 1;
            fsdb_fgets_uint32(&fsdb_local_dump.level);
            if ((fsdb_local_dump.level >= 0) &&
                (fsdb_local_dump.level < FSDB_DUMP_LEVEL_MAX))
            {
                done = TRUE;
            }
        }
        
        done = FALSE;
        while ( ! done)
        {
            printf("DUMP size (char [0], short [1], 4 byte [2], 8 byte [3]): ");
            fsdb_local_dump.size = FSDB_DUMP_4_BYTES;
            fsdb_fgets_uint32(&fsdb_local_dump.size);
            if ((fsdb_local_dump.size >= 0) &&
                (fsdb_local_dump.size < FSDB_DUMP_BYTES_MAX))
            {
                done = TRUE;
            }
        }

        done = FALSE;
        while ( ! done)
        {
            printf("DUMP level (hex [0], dec [1], oct [2], bin [3]): ");
            fsdb_local_dump.format = FSDB_DUMP_HEX;
            fsdb_fgets_uint32(&fsdb_local_dump.format);
            if ((fsdb_local_dump.format >= 0) &&
                (fsdb_local_dump.format < FSDB_DUMP_FORMAT_MAX))
            {
                done = TRUE;
            }
        }

        done = FALSE;
        while ( ! done)
        {
            printf("DUMP all the buddy bits [1], disable [0]): ");
            fsdb_local_dump.showBuddyBits = FALSE;
            fsdb_fgets_uint32(&fsdb_local_dump.showBuddyBits);
            if ((fsdb_local_dump.showBuddyBits == 0) ||
                (fsdb_local_dump.showBuddyBits == 1))
            {
                done = TRUE;
            }
        }

        fsdb_dump = &fsdb_local_dump;

        switch (fsdb_dump->format)
        {
        default:
        case FSDB_DUMP_HEX:
            strcpy(fsdb_dump->typeStr, "hex");
            break;
        case FSDB_DUMP_DEC:
            strcpy(fsdb_dump->typeStr, "dec");
            break;
        case FSDB_DUMP_OCT:
            strcpy(fsdb_dump->typeStr, "oct");
            break;
        case FSDB_DUMP_BIN:
            strcpy(fsdb_dump->typeStr, "bin");
            break;
        }
        // fall trough
    case FSDB_SHOW_DUMP:
        printf("\n");
        printf("DUMP level (min = 0, max = %d):     %d\n", FSDB_DUMP_LEVEL_MAX-1, fsdb_dump->level);
        printf("DUMP format                        %s (%d)\n", fsdb_dump->typeStr, fsdb_dump->format);
        printf("DUMP size (nbr of bytes)           %d\n", 1 << fsdb_dump->size);
        printf("DUMP all the buddy bits            %s\n", fsdb_dump->showBuddyBits ? "enabled" : "disabled");
        break;
    }
}

/******************************************************************************
 * dump the data in the block
 */


char *fsdb_dataTypeString[FSDB_DATA_TYPE_MAX] = {
    "No data",
    "Raw data",
    "Super block",
    "Disk inode Page",
    "Ref Byte Page",
    "Btree Page",
    "Dir Leaf Page",
    "Dir Index Page",
    "Emap Leaf Page",
    "File Sets Page",
    "Buddy Page",
    "Full inode block",
    "Dir in inode block",
    "Volume log buffers",
    "Disk Inode (dump one entry)"
};


void
fsdb_dump_block (void)
{
    int      input;
    int      def;
    int      done = FALSE;
    uint32   size, offset;
    int      type;
    int      loop;

    if ( ! fsdb_volume.dataAvailable)
    {
        printf("No disk data available to dump\n");
        return;
    }
    
    printf("Current block\n");
    printf("FS  block  %16lld (0x%016llx)\n", 
           fsdb_volume.offset / FS_BLOCK_SIZE,  fsdb_volume.offset / FS_BLOCK_SIZE);
    printf("EVM block  %16lld (0x%016llx)\n", 
           fsdb_volume.offset / SDM_BLOCK_SIZE, fsdb_volume.offset / SDM_BLOCK_SIZE);
    printf("offset     %16lld (0x%016llx)\n", 
           fsdb_volume.offset, fsdb_volume.offset);
    printf("\n");
            
    while (! done)
    {
        for (loop = 1 ; loop < FSDB_DATA_TYPE_MAX ; loop++)
        {
            printf("%2d:  %s\n", loop, fsdb_dataTypeString[loop]);
        }

        printf(" 0:  Return\n");
        def = fsdb_volume.dataType;
        printf("---> Input  [default: %d]: ", def);

        type = def;
        fsdb_fgets_int (&type);

        switch (type)
        {
        case FSDB_DATA_TYPE_UNKNOWN:
        {
            int done;
            
            done = FALSE;
            while ( ! done)
            {
                printf("offset in block (bytes) [0..%d] [default 0]: ", FS_BLOCK_SIZE-1);
                offset = 0;
                fsdb_fgets_uint32(&offset);
                if ((offset >= 0) && (offset < FS_BLOCK_SIZE))
                {
                    done = TRUE;
                }
            }
            
            done = FALSE;
            while ( ! done)
            {
                printf("size in block (bytes) [1..%d] [default %d]: ",
                       FS_BLOCK_SIZE - offset, FS_BLOCK_SIZE - offset);
                size = FS_BLOCK_SIZE - offset;
                fsdb_fgets_uint32(&size);
                if ((size > 0) && (size <= FS_BLOCK_SIZE - offset))
                {
                    done = TRUE;
                }
            }

            fsdb_dumpData ((void*)&fsdb_volume.data[offset], 
                           fsdb_volume.offset + offset, 
                           size);
            break;
        }
        case FSDB_DATA_TYPE_SUPER:
            fsdb_dumpSuperBlock ((fs_SuperBlock *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_D_INODE:
            fsdb_dumpDinodePage ((fs_DinodePage *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_REF_BYTE:
            fsdb_dumpRefByte ((fs_RefByte *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_BTREE:
            fsdb_dumpBtree ((fs_Btree *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_DIR_LEAF:
            fsdb_dumpDirleaf ((fs_Dirleaf *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_DIR_LEAF_COMPACT:
            fsdb_dumpDirleafCompact ((fs_DirleafCompact *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_EMAP_LEAF:
            fsdb_dumpEmapleaf ((fs_Emapleaf *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_FILE_SETS:
            fsdb_dumpFileSet ((fs_FileSetDiskRecord_t *)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_BUDDY:
            fsdb_dumpBuddy ((uint32*)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_FULL_INODE_BLOCK:
            fsdb_dumpFullInode ((char8*)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_DIR_IN_INODE_BLOCK:
            fsdb_dumpDirInInode ((char8*)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_VOL_LOG_BUF:
            fsdb_dumpVollogBuf ((vl_BufDiskHdr*)fsdb_volume.data);
            break;
        case FSDB_DATA_TYPE_D_INODE_ONE_ENTRY:
        {
            int done;
            fs_Dinode *dinode;
            
            done = FALSE;
            while ( ! done)
            {
                printf("Which entry [0..%d] [default 0]: ", FS_INODES_PER_PAGE-1);
                input = 0;
                fsdb_fgets_int (&input);
                if ((input >= 0) && (input < FS_INODES_PER_PAGE))
                {
                    done = TRUE;
                }
            }
            dinode = &(((fs_DinodePage *)fsdb_volume.data)->dinode[input]);
            
            fsdb_dumpDinode (dinode);
            break;
        }
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", type);
            break;
        }
    }
}



/******************************************************************************
 * do actual read of a FS block
 */


void
fsdb_dumpLunLabel (void)
{
    int      input;
    int      done = FALSE;
    int      def, defLun = 0;
    uint64   offset;
    int      action;

#if defined(FS_STANDALONE)
    printf("No LUN's configured in standalone\n");
    return ;
#endif    

    if (fsdb_volume.nbrLuns < 1)
    {
        printf("No LUN's configured\n");
        return ;
    }

    while (! done)
    {
        action = FSDB_LUNLABEL_NO_ACTION;

        printf("\n");
        printf("1: Read LUN label\n");
        printf("2: Write LUN - management area\n");
        printf("3: Write LUN - takeover area\n");
        printf("0: Return\n");
        if (defLun >= fsdb_volume.nbrLuns)
        {
            def = 0;
            defLun = 0;
        }
        else
        {
            def = 1;
        }
        printf("---> Input [default: %d]: ", def);
        input = def;
        fsdb_fgets_int (&input);

        if (input == 0)
        {
            return;
        }
        if (input == 2)
        { 
            action = FSDB_LUNLABEL_WRITE_MANAGEMENT;
        }
        if (input == 3)
        { 
            action = FSDB_LUNLABEL_WRITE_TAKEOVER;
        }
        
        printf("Which lun [0..%d] [default %d]: ", 
               fsdb_volume.nbrLuns - 1, defLun);
        input = defLun;
        fsdb_fgets_int (&input);

        if ((input < 0) || 
            (input >= fsdb_volume.nbrLuns))
        {
            return;
        }

        fsdb_volume.dataType = FSDB_DATA_TYPE_UNKNOWN;
        fsdb_volume.dataAvailable = FALSE;

        offset = fsdb_volume.lunInfo[input].size;
     
        if (fsdb_evm_read_label(&fsdb_volume, input, action))
        {
            printf ("No valid data for LUN %d offset: %lld  block FS: %lld EVM: %lld\n\n",
                    input, offset, offset / FS_BLOCK_SIZE, offset / SDM_BLOCK_SIZE);
            return;
        }
        defLun++;
    }
}


void
fsdb_do_read(uint64 offset, uint32 bufSize)
{
#if defined(FS_STANDALONE)
    int      fd, rc;
#endif
    uint64   lun_offset;
    int      loop, lunNbr;
    uint32   cksum;
    uint64   seekset;

    if ( ! fsdb_volume.ready)
    {
        printf("Volume/lun not yet configured to read\n");
        return;
    }

    fsdb_volume.dataAvailable = FALSE;

/*
    printf("Reading FS  block  %16lld (0x%016llx)\n", offset / FS_BLOCK_SIZE,  offset / FS_BLOCK_SIZE);
    printf("Reading EVM block  %16lld (0x%016llx)\n", offset / SDM_BLOCK_SIZE, offset / SDM_BLOCK_SIZE);
    printf("Reading offset     %16lld (0x%016llx)\n", offset, offset);
    printf("\n");
*/            
#if defined(FS_STANDALONE)

    fd = open64(fsdb_volume.mountName,
                O_RDWR | O_CREAT,
                S_IREAD | S_IWRITE);
    if (fd <= 0)
    {
        printf("Read req, %s not found\n", fsdb_volume.mountName);
        return;
    }
    seekset = lseek64(fd,
                      offset,
                      SEEK_SET);
    if (seekset != offset)
    {
        printf("failed to seek to offset %lld block %lld for read\n", 
               offset, offset / FS_BLOCK_SIZE);
        return;
    }

    fsdb_volume.dataType = FSDB_DATA_TYPE_UNKNOWN;

    rc = read(fd, fsdb_volume.data, bufSize);
    
    if (rc != bufSize)
    {
        printf("failed to read offset %lld block %lld (read only %d bytes)\n", 
               offset, offset / FS_BLOCK_SIZE, rc);
        return;
    }

    close(fd);

#else

    if (fsdb_volume.nbrLuns < 1)
    {
        printf("No LUN's configured\n");
        return ;
    }

    lun_offset = offset;
    lunNbr = -1;

    for (loop = 0 ; loop < fsdb_volume.nbrLuns; loop++)
    {
        if (fsdb_volume.lunInfo[loop].startBlk != 0)
        {
            printf("Start block (%d) for lun %d must be 0\n", 
                   fsdb_volume.lunInfo[loop].startBlk, loop);
            return ;
        }

        if (lun_offset < fsdb_volume.lunInfo[loop].size)
        {
            lunNbr = loop;
            break;
        }
        else
        {
            lun_offset -= fsdb_volume.lunInfo[loop].size;
        }
    }

    if (lunNbr == -1)
    {
        printf("Offset not found on this LUN set (offset: %lld)\n", offset);
        return;
    }
    
    if (fsdb_volume.lunInfo[lunNbr].sdmHandle == SDM_INVALID_HANDLE)
    {
        printf ("No valid SdmHandle Lun %d \n\n", lunNbr);
        return;
    }

    fsdb_volume.dataType = FSDB_DATA_TYPE_UNKNOWN;

    if (fsdb_sdm_disk_read(fsdb_volume.lunInfo[lunNbr].sdmHandle, 
                           lun_offset,
                           bufSize,
                           fsdb_volume.data) != NFX_OK)
    {
        printf ("No valid data \n\n");
        return;
    }
#endif

    cksum = uint32_checksum(fsdb_volume.data, bufSize, 0);
    // printf("checksum 0x%x\n", cksum);

    fsdb_volume.dataAvailable = TRUE;
    fsdb_volume.offset = offset;
}

/******************************************************************************
 * some auxiliary functions for super blocks 
 * !!!  copy of fs-layout.c
 */

int32 fsdb_superBlockCount(void)
{
    return 2;
}
uint64 fsdb_superBlockBlock(uint32 copy)
{
    switch (copy) {
        case 0:
            return 0;
        case 1:
            return 1 + FS_MAX_SNAPSHOTS;
        default:
            printf("Invalid superblock number %d\n", copy);
    }
    return -1;
}

/******************************************************************************
 * find all info related to block
 */

uint64
fsdb_emap_leaf_lookup(fs_Emapleaf *emapleaf, uint64 page)
{
    int      loop, index;
    uint64   blockNbr = FS_INVALID_BLKNUM;

    if (emapleaf->nkeys >= FS_EMAP_EXTENTS_MAX)
    {
        fsdb_dumpError("Emap leaf page - nkeys %d >= FS_EMAP_EXTENTS_MAX %d\n",
                       emapleaf->nkeys, FS_EMAP_EXTENTS_MAX);
        return FALSE;
    }

    for (loop = 0 ; loop < emapleaf->nkeys; loop++)
    {
        index = emapleaf->index[loop];

        if ((page >= emapleaf->offsets[index]) &&
            (page < emapleaf->offsets[index] + 
             (emapleaf->extents[index].lengthAndStartBlock >> EXTENT_BLOCK_BITS)))
        {
            break;
        }
    }
    if (loop < emapleaf->nkeys)
    {
        blockNbr = (emapleaf->extents[index].lengthAndStartBlock & EXTENT_BLOCK_MASK) +
            (page - emapleaf->offsets[index]);
        
        printf ("Found block for page %lld in emap leaf page (i: %d index: %d)\n", 
                page, loop, index);
        printf ("Offset %lld len %lld Startblock %lld\n", 
                emapleaf->offsets[index], 
                emapleaf->extents[index].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                emapleaf->extents[index].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        printf ("Found block: %lld (0x%llx)\n", blockNbr, blockNbr);
            
    }
    return (blockNbr);
}

uint64
fsdb_emap_flat_root_lookup(fs_Emapflatroot *fr, uint64 page)
{
    int      loop;
    uint64   blockNbr = FS_INVALID_BLKNUM;

    if (fr->hdr.level != 0)
    {
        printf("Error using emap flat root lookup while level (%d) != 0\n", fr->hdr.level);
        return (blockNbr);
    }
    
    if ((fr->hdr.nkeys < 0) ||
        (fr->hdr.nkeys >= FS_INOFLATROOT_MAXKEYS))
    {
        printf("Dinode - emapflatroot: nkeys (%d) >= MAX (%d)\n", 
               fr->hdr.nkeys, FS_INOFLATROOT_MAXKEYS);
        return (blockNbr);
    }

    for (loop = 0; loop < fr->hdr.nkeys; loop++)
    {
        if ((page >= fr->offsets[loop]) &&
            (page < fr->offsets[loop] + 
             (fr->extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS)))
        {
            break;
        }
    }
    if (loop < fr->hdr.nkeys)
    {
        blockNbr = (fr->extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK) +
            (page - fr->offsets[loop]);
        
        // printf ("Found block for inum 0x%llx page 0x%llx in emap flat root (%d)\n", 
        //         inum, page, loop);
        printf ("Found block for page 0x%llx in emap flat root (%d)\n", 
                page, loop);
        printf ("Offset %lld len %lld Startblock %lld\n", 
                fr->offsets[loop], 
                fr->extents[loop].lengthAndStartBlock >> EXTENT_BLOCK_BITS,
                fr->extents[loop].lengthAndStartBlock & EXTENT_BLOCK_MASK);
        printf ("Found block: %lld (0x%llx)\n", blockNbr, blockNbr);
            
    }
    return (blockNbr);
}

uint64
fsdb_emap_root_lookup(fs_Emaproot *r, uint64 page)
{
    uint64   blockNbr = FS_INVALID_BLKNUM;
    uint64   offset;
    int      loop;
    int      child_level;

    if (r->hdr.level == 0)
    {
        printf("Error using emap root lookup while level == 0\n");
        return (blockNbr);
    }
    
    if ((r->hdr.nkeys < 0) ||
        (r->hdr.nkeys >= FS_INOROOT_MAXKEYS))
    {
        printf("Dinode - emaproot: nkeys (%d) >= MAX (%d)\n", 
               r->hdr.nkeys, FS_INOROOT_MAXKEYS);
        return (blockNbr);
    }

    // see fs_emap_lookup - get_child_buf
    for (loop = 0; loop < r->hdr.nkeys; loop++)
    {
        if (page <= r->offsets[loop])
        {
            break;
        }
    }
    if (loop < r->hdr.nkeys)
    {
        blockNbr = r->left[loop];
    }
    else
    {
        blockNbr = r->right;
    }

    child_level = r->hdr.level - 1;
        
    printf ("Found child block for inum: page 0x%llx in emap root\n", 
            page);
    printf (" Block: %lld (0x%llx)\n", blockNbr, blockNbr);
    printf ("Read the block of the inode file\n\n");

    if ( ! fsdb_continue())
        return (blockNbr);

    offset = blockNbr * FS_BLOCK_SIZE;

    // This will overwrite r ....

    fsdb_do_read(offset, FS_BLOCK_SIZE);
    if ( ! fsdb_volume.dataAvailable)
    {
        printf ("Could not read the Inode Block\n");
        return (FS_INVALID_BLKNUM);
    }

    fsdb_volume.dataType = FSDB_DATA_TYPE_EMAP_LEAF;
    fsdb_dumpEmapleaf ((fs_Emapleaf *)fsdb_volume.data);

    if (child_level == 0)
    {
        // found the emap leaf page 
        blockNbr = fsdb_emap_leaf_lookup((fs_Emapleaf *)fsdb_volume.data, page);
        
        return (blockNbr);
    }
    
    if ( ! fsdb_continue())
        return (blockNbr);

    // todo tbi


/*
 * Perform a recursive search through the B-tree. The search results are
 * placed in the "void *data" argument so fs_bt_search itself returns void.
 * 
 * The return value is a key used to guide a sequential traversal of
 * all the leaves. If the return value is zero, then the parent picks
 * the next child branch to visit, by setting the key to "key+1".
 * XXX: explain this better.
 */
/*
        if (child_level > 0)
        {

uint64
fs_bt_search(fs_Context *context,
	     fs_Inode *inode,
	     fs_Buf *buf,
	     int level,
	     uint64 key,
	     void *data,
	     fs_Bt_search_dispatch *search,
             boolean exclusive)
{
	int child_level = level - 1;
	uint64 next_key;
	fs_Buf *child;
	int i;

	// Traverse B-tree nodes descending to the left on <=.
	i = fs_bt_find_child(buf, key);

	// only visit leaves with exclusive flag set (for free tree).
	// if (child_level > 0)
	//	exclusive = FALSE;

    // Free tree is gone, this should now always be FALSE.
    assert(exclusive == FALSE);

	child = fs_bt_get_child(context, inode, buf, i, exclusive);
	assert(child != NULL);
	next_key = search->func[child_level](context, inode, child, child_level, key, data, search, exclusive);

	// XXX: by using the next_key hack even on single probes, we limit
	// concurrency by forcing the bufRelease after the call to the
	// child search->func(). Overall, crab-walking to increase concurrency
	// could probably be revisited in the btree code.
	if (next_key == 0) {
		if (i < buf->u.btree->nkeys)
			next_key = *get_key(buf->u.btree, i) + 1;
	}

        fs_bufRelease(context, buf);
	return next_key;

        }
        else
        {

            fs_emapleaf_search
	found = fs_emapleaf_scan(l, &i, offset);
fs_emapleaf_scan(fs_Emapleaf *l, int *index, uint64 offset)
	int bot, top;
	int i;

	bot = 0;
	top = l->nkeys;
	i = 0;

	FS_ASSERT((unsigned)l->nkeys <= FS_EMAP_EXTENTS_MAX);

	// Binary search to a likely starting point.
	while (bot < top) {
		uint64 offset_i; // cache of the ith offset in the leaf

		i = (top + bot) / 2;
		offset_i = get_offset(l, i);
		if (offset == offset_i)
			break;
		if (offset < offset_i)
			top = i-1;
		else
			bot = i+1;
	}

	// Move forward over small keys.
	while (i < l->nkeys && offset > get_offset(l, i))
		i++;

	// Back up over larger keys.
	while (i > 0
               && offset < (get_offset(l, i - 1)
                             + fs_extentLength(get_extent(l, i - 1))))
        {
		--i;
        }

	*index = i;

	return i < l->nkeys && offset >= get_offset(l, i)
			    && offset < (get_offset(l, i)
                                         + fs_extentLength(get_extent(l, i)));

		p->offset = get_offset(l, i);
		p->extent = get_extent(l, i);
static inline uint64
get_offset(fs_Emapleaf *l, int i)
{
	FS_ASSERT(l->index[i] < FS_EMAP_EXTENTS_MAX);
	return l->offsets[l->index[i]];
}

static inline fs_Extent
get_extent(fs_Emapleaf *l, int i)
{
	FS_ASSERT(l->index[i] < FS_EMAP_EXTENTS_MAX);
	return l->extents[l->index[i]];
}

        }
        
// 	bt_funcs.emap.search.func[child_level](context, inode, child, child_level, offset, &p, &bt_funcs.emap.search, FALSE);



// see fs_emap_lookup
// child buf 

// 	bt_funcs.emap.search.func[child_level](context, inode, child, child_level, offset, &p, &bt_funcs.emap.search, FALSE);

        if ( ! fsdb_continue())

        // return (blockNbr);
*/

    return (blockNbr);
}



void
fsdb_block_lookup (uint64 block)
{
    int     loop;
    uint64  offset;
    
    offset = block * FS_BLOCK_SIZE;
        
    // check if super block
    for (loop = 0 ; loop < fsdb_superBlockCount(); loop++)
    {
        if (block == fsdb_superBlockBlock(loop))
        {
            fsdb_do_read(offset, FS_BLOCK_SIZE);
            if ( ! fsdb_volume.dataAvailable)
            {
                fsdb_dumpError("Error reading superblock %d offset %lld\n", loop, offset);
                return;
            }
            fsdb_volume.dataType = FSDB_DATA_TYPE_SUPER;
            fsdb_dumpSuperBlock ((fs_SuperBlock *)fsdb_volume.data);

            if ( ! fsdb_continue())
                return ;
        }
    }

    // check if predefined
    if (block == 1)
    {
        fsdb_do_read(offset, FS_BLOCK_SIZE);
        
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading disk inode %lld offset %lld\n", block, offset);
            return;
        }
        fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;
        fsdb_dumpDinodePage ((fs_DinodePage *)fsdb_volume.data);

        if ( ! fsdb_continue())
            return ;
    }

    // snap shot dinode 
    if ((block > 1) && (block <= FS_MAX_SNAPSHOTS)) 
    {
        fsdb_do_read(offset, FS_BLOCK_SIZE);
        
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading disk inode %lld offset %lld\n", block, offset);
            return;
        }

        fsdb_dumpLine("Dump snap shot page block %lld (offset %lld) as dinode: \n\n\n", block, offset);

        fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;
        fsdb_dumpDinodePage ((fs_DinodePage *)fsdb_volume.data);

        if ( ! fsdb_continue())
            return ;
    }

    // find in buddytree files
    {
        fs_Emaproot  *r;
        uint64 blockNbr;
        int level, base, index, word, mask, region, offsetInRegion, free;
        uint32 *buddy;
        
        fsdb_dumpLine("\nRead buddy tree dinode\n\n");

        offset = 1 * FS_BLOCK_SIZE;
        fsdb_do_read(1 * FS_BLOCK_SIZE, FS_BLOCK_SIZE);
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading disk inode %lld offset %lld\n", 1, offset);
            return;
        }
        fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;
        fsdb_dumpDinode ((fs_Dinode *)&((fs_DinodePage*)fsdb_volume.data)->dinode[FS_FREE_BUDDY_TREE_INODE]);
        
        // find buddy block offset for region 
        r = &(((fs_Dinode )((fs_DinodePage*)fsdb_volume.data)->dinode[FS_FREE_BUDDY_TREE_INODE]).u.emaproot);

        region = block >> FS_MAX_BUDDY_LEVELS;

        fsdb_dumpLine("\nRead buddy tree - emap lookup\n\n");

        if (r->hdr.level == 0)
        {
            if ((blockNbr = fsdb_emap_flat_root_lookup((fs_Emapflatroot *)r, region)) == FS_INVALID_BLKNUM)
            {
                fsdb_dumpError("Emapflatroot lookup error for page %d in buddy tree\n", region);
                return;
            }
        }
        else
        {
            if ((blockNbr = fsdb_emap_root_lookup(r, region)) == FS_INVALID_BLKNUM)
            {
                fsdb_dumpError("Emapflatroot lookup error for page %d in buddy tree\n", region);
                return;
            }
        }
        
        offset = blockNbr * FS_BLOCK_SIZE;
        fsdb_do_read(offset, FS_BLOCK_SIZE);
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading buddy page offset %lld\n", offset);
            return;
        }
        buddy = (uint32*)fsdb_volume.data;
        fsdb_volume.dataType = FSDB_DATA_TYPE_BUDDY;
        if ( ! fsdb_dumpBuddy ((uint32*)fsdb_volume.data))
        {
            fsdb_dumpError("No valid buddy page offset %lld\n", offset);
//            return;
        }

        offsetInRegion = block & ((1 << FS_MAX_BUDDY_LEVELS) - 1);

        printf ("\n");
        for (level = FS_MAX_BUDDY_LEVELS ; level >= 0; level--) 
        {
            base = 1 << (FS_MAX_BUDDY_LEVELS - level);
            index = base + (offsetInRegion >> level);
            word = index / BITS_PER_WORD;
            mask = 1 << (index & (BITS_PER_WORD - 1));
            printf("level %2d base %5d index %5d (%4d) mask 0x%08x : 0x%02x\n", 
                   level, base, index, index / 8, mask, buddy[word] & mask);
        }
        printf ("\n");


        free= FALSE;
        for (level = FS_MAX_BUDDY_LEVELS ; level >= 0; level--) 
        {
            base = 1 << (FS_MAX_BUDDY_LEVELS - level);
            index = base + (offsetInRegion >> level);
            word = index / BITS_PER_WORD;
            mask = 1 << (index & (BITS_PER_WORD - 1));
            // printf("level %d base %d index %d  0x%x\n", level, base, index, buddy[word] & mask);
            if ((buddy[word] & mask) != 0)
            {
                free = TRUE;
                break;
            }
        }
        if (free)
        {
            fsdb_dumpLine("Buddy tree indicates that block %lld is free (level %d)\n", block, level);
        }
        else
        {
            fsdb_dumpLine("Buddy tree indicates that block %lld is in use\n", block);
        }

        fsdb_dumpLine("\n");
        if ( ! fsdb_continue())
            return ;
    }

    // find in refbyte files
    {
        fs_Emaproot  *r;
        uint64 blockNbr;
        int region, offsetInRegion;
        
        fsdb_dumpLine("\nRead ref byte dinode\n\n");

        offset = 1 * FS_BLOCK_SIZE;
        fsdb_do_read(1 * FS_BLOCK_SIZE, FS_BLOCK_SIZE);
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading disk inode %lld offset %lld\n", 1, offset);
            return;
        }
        fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;
        fsdb_dumpDinode ((fs_Dinode *)&(((fs_DinodePage*)fsdb_volume.data)->dinode[FS_REF_COUNT_INODE]));
        
        // find ref count block offset for region 
        r = &(((fs_Dinode)((fs_DinodePage*)fsdb_volume.data)->dinode[FS_REF_COUNT_INODE]).u.emaproot);

        region = block / FS_BLOCK_SIZE;

        fsdb_dumpLine("\nRead ref byte page - emap lookup\n\n");

        if (r->hdr.level == 0)
        {
            if ((blockNbr = fsdb_emap_flat_root_lookup((fs_Emapflatroot *)r, region)) == FS_INVALID_BLKNUM)
            {
                fsdb_dumpError("Emapflatroot lookup error for page %d in buddy tree\n", region);
                return;
            }
        }
        else
        {
            if ((blockNbr = fsdb_emap_root_lookup(r, region)) == FS_INVALID_BLKNUM)
            {
                fsdb_dumpError("Emapflatroot lookup error for page %d in buddy tree\n", region);
                return;
            }
        }
        
        offset = blockNbr * FS_BLOCK_SIZE;
        fsdb_do_read(offset, FS_BLOCK_SIZE);
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading ref byte page offset %lld\n", offset);
            return;
        }
        fsdb_volume.dataType = FSDB_DATA_TYPE_REF_BYTE;
        if ( ! fsdb_dumpRefByte ((fs_RefByte *)fsdb_volume.data))
        {
            fsdb_dumpError("No valid ref byte page offset %lld\n", offset);
            return;
        }

        offsetInRegion = block % FS_BLOCK_SIZE;
        if (((fs_RefByte *)(fsdb_volume.data))[offsetInRegion] == 0)
        {
            fsdb_dumpLine("Ref byte page indicates that block %lld is free\n", 
                          block);
        }
        else
        {
            fsdb_dumpLine("Ref byte page indicates that block %lld is in use: snap %d - refbyte (0x%02x)\n", 
                          block, 
                          ((fs_RefByte *)(fsdb_volume.data))[offsetInRegion] >> FS_REFBYTE_SNAP_COUNT_SHIFT,
                          ((fs_RefByte *)(fsdb_volume.data))[offsetInRegion]);
        }

        fsdb_dumpLine("\n");
        if ( ! fsdb_continue())
            return ;
    }


    // todo tbi
    // check if block itself is used in
    // - inode inode file
    // - ref count file
    // - free buddy file
    // - full inode file
    // - dir in inode file
    // - fileset file
    // - fileset user quota file
    // - quota file
    // - .... file
    // else start checking the root inode
    //   .... user block
}


/******************************************************************************
 * find block nbr for given inum/page
 */

uint64
fsdb_inum_page_lookup(
    uint64  inum,
    uint64  page)
{
    uint64            offset;
    uint64            blockNbr = FS_INVALID_BLKNUM;
    fs_Dinode        *dinode;
    uint64            inodeInum;

    offset = (FS_HANDLE_GET_SNAP(inum) + 1) * FS_BLOCK_SIZE;

    printf ("Read the FS_INODE_INODE block\n\n");
    
    fsdb_do_read(offset, FS_BLOCK_SIZE);
    fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;

    if ( ! fsdb_volume.dataAvailable)
    {
        printf ("Could not read the first Inode Block snap %lld\n", FS_HANDLE_GET_SNAP(inum));
        return (blockNbr);
    }

    dinode = &(((fs_DinodePage *)fsdb_volume.data)->dinode[FS_INODE_INODE]);

    fsdb_dumpDinode(dinode);

    if ( ! FS_I_IS_REG(dinode->attrs))
    {
        printf ("Dinode FS_INODE_INODE is not a REGULAR file, snap %lld\n", FS_HANDLE_GET_SNAP(inum));
        return (blockNbr);
    }

#define FS_BT_MAXLEVEL 8
    if ((dinode->u.emaproot.hdr.level < 0) || (dinode->u.emaproot.hdr.level >= FS_BT_MAXLEVEL))
    {
        printf ("Dinode FS_INODE_INODE has invalid level %d [0..%d]\n", 
                dinode->u.emaproot.hdr.level, FS_BT_MAXLEVEL);
        return (blockNbr);
    }

    inodeInum = inum / FS_DINODE_PAGE_MAX;
    
    if (dinode->u.emaproot.hdr.level == 0)
    {

        printf ("\nemap flat root lookup in page %lld\n\n", inodeInum);

        if ((blockNbr = fsdb_emap_flat_root_lookup(&dinode->u.emapflatroot, inodeInum)) == FS_INVALID_BLKNUM)
        {
            fsdb_dumpError("Emapflatroot lookup error for page %d in inode page\n", page);
            return (blockNbr);
        }

        // if flat, this is the dinode page !!!!!
    }
    else
    {
        printf ("\nemap root lookup in page %lld\n\n", inodeInum);

        if ((blockNbr = fsdb_emap_root_lookup(&dinode->u.emaproot, inodeInum)) == FS_INVALID_BLKNUM)
        {
            fsdb_dumpError("Emaproot lookup error for page %d in inode page\n", page);
            return (blockNbr);
        }

        printf ("\nemap leaf page lookup in block %lld for page %lld\n\n", blockNbr, inodeInum);

        if ( ! fsdb_continue())
            return (blockNbr);

        // this is an emap leaf page
        offset = blockNbr * FS_BLOCK_SIZE;
        fsdb_do_read(offset, FS_BLOCK_SIZE);
        if ( ! fsdb_volume.dataAvailable)
        {
            fsdb_dumpError("Error reading the emap leaf page %lld\n", offset);
            return FS_INVALID_BLKNUM;
        }

        fsdb_volume.dataType = FSDB_DATA_TYPE_EMAP_LEAF;

        /* already done in fsdb_emap_root_lookup 
        if ( ! fsdb_dumpEmapleaf ((fs_Emapleaf *)fsdb_volume.data))
        {
            fsdb_dumpError("Error in emap leaf page %lld\n", offset);
            return FS_INVALID_BLKNUM;
        }
        */

        if ((blockNbr = fsdb_emap_leaf_lookup((fs_Emapleaf *)fsdb_volume.data, 
                                              inodeInum)) == FS_INVALID_BLKNUM)
        {
            fsdb_dumpError("Emap leaf lookup error for page %d\n", page);
            return (blockNbr);
        }
    }
    
    printf ("\nDinode block %lld : look for inum %lld entry %lld\n\n", 
            blockNbr, inum, inum % FS_DINODE_PAGE_MAX);

    if ( ! fsdb_continue())
        return (blockNbr);

    offset = blockNbr * FS_BLOCK_SIZE;
    fsdb_do_read(offset, FS_BLOCK_SIZE);
    if ( ! fsdb_volume.dataAvailable)
    {
        fsdb_dumpError("Error reading inode page offset %lld\n", offset);
        return FS_INVALID_BLKNUM;
    }

    fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;
    if ( ! fsdb_dumpDinode ((fs_Dinode *)&((fs_DinodePage*)fsdb_volume.data)->dinode[inum % FS_DINODE_PAGE_MAX]))
    {
        fsdb_dumpError("Inum not used in Dinode page (offset %lld entry %lld )\n", 
                       offset, inum % FS_DINODE_PAGE_MAX);
        return FS_INVALID_BLKNUM;
    }

    printf ("TBI\n");
    printf (" Block: %lld (0x%llx)\n", blockNbr, blockNbr);
    
    // todo tbi
    // this is page 0
    // look for requested page
    // try to find type ....

    return (blockNbr);
}


/******************************************************************************
 * read some well known blocks
 */

void
fsdb_read_predefined(void)
{
    int      input;
    int      done = FALSE;
    int      size;
    int      loop;
    uint64   offset;
    uint64   blockNbr;
            
#if defined(FS_STANDALONE)
    int      fd, rc;
#endif

    if ( ! fsdb_volume.ready)
    {
        printf("Volume/lun not yet configured to read\n");
        return;
    }

    while (! done)
    {
        printf("\n");
        printf("1: Read superBlock\n");
        printf("2: Read reserved dinode page (FS block 1)\n");
        printf("3: inum/page lookup\n");
        printf("4: block number lookup\n");
        printf("5: volume log records\n");
        printf("6: LUN lablels\n");
        printf("0: Return\n");
        printf("---> Input  [default: 0]: ");

        input = 0;
        fsdb_fgets_int (&input);

        switch(input)
        {
        case 1:
            printf("Which superblock [0..%d]: ", fsdb_superBlockCount() - 1);
            input = 0;
            fsdb_fgets_int (&input);
            if ((input < 0) || 
                (input >= fsdb_superBlockCount()))
            {
                printf("Wrong super block id %d, not in range [0..%d]\n", input, fsdb_superBlockCount()-1);
                continue;
            }
            blockNbr = fsdb_superBlockBlock((uint32)input);
            //blockNbr = block.block;
            offset = blockNbr * FS_BLOCK_SIZE;
                
            fsdb_do_read(offset, FS_BLOCK_SIZE);

            if (fsdb_volume.dataAvailable)
            {
                fsdb_volume.dataType = FSDB_DATA_TYPE_SUPER;
                fsdb_dumpSuperBlock ((fs_SuperBlock *)fsdb_volume.data);
            }
            break;
        case 2:
            offset = 1 * FS_BLOCK_SIZE;
                
            fsdb_do_read(offset, FS_BLOCK_SIZE);

            if (fsdb_volume.dataAvailable)
            {
                fsdb_volume.dataType = FSDB_DATA_TYPE_D_INODE;
                fsdb_dumpDinodePage ((fs_DinodePage *)fsdb_volume.data);
            }
            break;
        case 3:
        {
            uint64  inum;
            uint64  page;
            uint64  block;
            
            printf("Inum: ");
            fsdb_fgets_uint64(&inum);
            printf("Page: ");
            fsdb_fgets_uint64(&page);

            block = fsdb_inum_page_lookup(inum, page);
            break;
        }
        case 4:
        {
            uint64  block = 0;
            
            printf("block number: ");
            fsdb_fgets_uint64(&block);

            fsdb_block_lookup(block);
            break;
        }
        case 5:
            fsdb_dumpLogRecord ();
            break;
        case 6:
            fsdb_dumpLunLabel ();
            break;
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", input);
            break;
        }
    }
}

/******************************************************************************
 * read menu  
 */

void 
fsdb_read_menu (void)
{
    int      input;
    int      def;
    int      done = FALSE;
    int      size;
    int      loop;
    uint64   offset;
    uint64   blockNbr;

    if ( ! fsdb_volume.ready)
    {
        printf("Volume/lun not yet configured to read\n");
        return;
    }

    while (! done)
    {
        printf("\n");
        printf("1: Read block for given offset (bytes)\n");
        printf("2: Read block for given FS  blocknumber\n");
        def = 2;
        printf("3: Read block for given EVM blocknumber\n");
        printf("4: Read predefined blocks\n");
        if(fsdb_volume.dataAvailable)
        {
            printf("5: Dump block\n");
            def = 5;
        }
        printf("0: Return\n");
        printf("---> Input [default: %d]: ", def);

        input = def;
        fsdb_fgets_int (&input);

        switch(input)
        {
        case 1:
        {
            int done;
            
            done = FALSE;
            while ( ! done)
            {
                printf("offset in volume (bytes) [0..%lld]: ", fsdb_volume.currSize-1);
                offset = 0;
                fsdb_fgets_uint64(&offset);
                if (offset < fsdb_volume.currSize)
                {
                    done = TRUE;
                }
            }
            blockNbr = offset / FS_BLOCK_SIZE;
            offset = blockNbr * FS_BLOCK_SIZE;

            fsdb_do_read(offset, FS_BLOCK_SIZE);
            break;
        }
        case 2:
        {
            int done;
            
            done = FALSE;
            while ( ! done)
            {
                printf("FS block number in volume (block size %d) [0..%lld]: ", 
                       FS_BLOCK_SIZE, fsdb_volume.currSize / FS_BLOCK_SIZE);
                blockNbr = 0;
                fsdb_fgets_uint64 (&blockNbr);
                if (blockNbr <= fsdb_volume.currSize / FS_BLOCK_SIZE)
                {
                    done = TRUE;
                }
            }
            offset = blockNbr * FS_BLOCK_SIZE;

            fsdb_do_read(offset, FS_BLOCK_SIZE);
            break;
        }
        case 3:
        {
            int done;
            
            done = FALSE;
            while ( ! done)
            {
                printf("EVM block number in volume (block size %d) : ", SDM_BLOCK_SIZE);
                blockNbr = 0;
                fsdb_fgets_uint64 (&blockNbr);
                if (blockNbr <= fsdb_volume.currSize / SDM_BLOCK_SIZE)
                {
                    done = TRUE;
                }
            }
            offset = blockNbr * SDM_BLOCK_SIZE;

            fsdb_do_read(offset, SDM_BLOCK_SIZE);
            break;
        }
        case 4:
            fsdb_read_predefined();
            break;
        case 5:
            fsdb_dump_block();
            break;
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", input);
            break;
        }
    }
}

/******************************************************************************
 * menu to setup volume
 */

void 
fsdb_vol_menu (void)
{
    int      input;
    int      def;
    int      done = FALSE;
    char     str[STR_MAX_SIZE];
    int      size;
    int      loop;

    while (! done)
    {
        printf("\n");
        printf("1: Define volume by Name\n");
#if ! defined(FS_STANDALONE)
        printf("2: Define volume by VolId\n");
        printf("3: Define directly the LUN (wwn,lun)\n");
#endif
        printf("4: Show volume information\n");
        printf("0: Return\n");
        if (fsdb_volume.ready)
            def = 0;
        else
            def = 1;
        printf("---> Input [default: %d]: ", def);
        input = def;
        fsdb_fgets_int (&input);
        
        switch(input)
        {
        case 1:
            bzero (&fsdb_volume, sizeof(fsdb_volume));

            printf("Volume Name: ");
#if defined(FS_STANDALONE)
            strcpy(fsdb_volume.mountName, "FS_EAGLE_SIM_VOL_");
            size = strlen(fsdb_volume.mountName);
#else
            size = 0;
#endif

            str[0] = 0;
            fgets(str, STR_MAX_SIZE, stdin);
            sscanf (str, "%s", &(fsdb_volume.mountName[size]));

#if defined(FS_STANDALONE)
            fsdb_volume.currSize = 1000000000;
#else
            if (fsdb_get_volid (fsdb_volume.mountName, &(fsdb_volume.volId)) != NFX_OK)
            {
                printf ("Could not find VolId\n\n");
                continue;
            }

            printf("Vol: %s - VolId %lld (0x%llx)\n", 
                   fsdb_volume.mountName, 
                   fsdb_volume.volId.whole_thing,
                   fsdb_volume.volId.whole_thing);
        
            if (fsdb_get_evm_vol_cfg (fsdb_volume.volId, &fsdb_volume) != NFX_OK)
            {
                printf ("Could not find Disk info\n\n");
                continue;
            }
#endif
            fsdb_volume.ready = TRUE;
            break;
        case 2:
            bzero (&fsdb_volume, sizeof(fsdb_volume));

#if defined(FS_STANDALONE)
            printf("Volume Id not Implemented\n");
#else
            printf("Volume Id : ");
            fsdb_volume.volId.whole_thing = 0;
            fsdb_fgets_uint64 (&fsdb_volume.volId.whole_thing);
            
            if (fsdb_get_evm_vol_cfg (fsdb_volume.volId, &fsdb_volume) != NFX_OK)
            {
                printf ("Could not find Disk info\n\n");
                continue;
            }

            fsdb_volume.ready = TRUE;
#endif
            break;
        case 3:
            bzero (&fsdb_volume, sizeof(fsdb_volume));

#if defined(FS_STANDALONE)
            printf("Volume Id not Implemenated\n");
#else
            
            sprintf(fsdb_volume.mountName, "LUN based lookup");
            fsdb_volume.nbrLuns = 1;
            // Todo: find correct values
            fsdb_volume.currSize = fsdb_volume.lunInfo[0].size = 1000000000;
            
            printf("WWN : ");
            fsdb_volume.lunInfo[0].wwn = 0;
            fsdb_fgets_uint64 (&fsdb_volume.lunInfo[0].wwn);
            
            printf("lun number : ");
            fsdb_volume.lunInfo[0].lun = 0;
            fsdb_fgets_uint32 (&fsdb_volume.lunInfo[0].lun);

            if ((fsdb_volume.lunInfo[0].sdmHandle 
                 = fsdb_sdm_ui_getSdmHandle (
                     fsdb_volume.lunInfo[0].wwn,
                     fsdb_volume.lunInfo[0].lun)) != SDM_INVALID_HANDLE)
            {
                printf ("Could not find SdmHandle info\n\n");
                continue;
            }
            fsdb_volume.ready = TRUE;
#endif
            break;
        case 4:
            fsdb_vol_info();
            break;
        case 0:
            return;
            break;
        default:
            printf ("Invalid input %s (%d)\n", str, input);
            break;
        }
    }
}

/******************************************************************************
 * dump some general info
 */

void
fsdb_vol_info(void)
{
    int loop;
    
    if ( ! fsdb_volume.ready)
    {
        printf("Volume/lun not yet configured to read\n");
        return;
    }

#if defined(FS_STANDALONE)
    printf("Volume will read file: %s\n", fsdb_volume.mountName);
#else
    printf("Logical volume name: %s, id: 0x%llx\n",
           fsdb_volume.mountName, fsdb_volume.volId.whole_thing);
    printf("  current size (MB): %lld\n",
           fsdb_volume.currSize / EVM_MEGA_BYTES);
    // quota not stored
    printf("  online: %s, state: %s\n",
           fsdb_volume.evmOnLine ? "Yes":"No",
           fsdb_volume.evmState == EVM_VOL_READY ? "ready":"NOT ready");
    
    for (loop = 0; loop < fsdb_volume.nbrLuns; loop++)
    {
        printf("  - lun #%d\n", loop + 1);
        printf("    scsi id: 0x%llx\n"
               "    size: %lld bytes, startBlk: %d\n",
               fsdb_volume.lunInfo[loop].scsiId, 
               fsdb_volume.lunInfo[loop].size,
               fsdb_volume.lunInfo[loop].startBlk);
        
        printf("    WWN: 0x%llx  lun: %d  sdmHandle 0x%x\n",
               fsdb_volume.lunInfo[loop].wwn,
               fsdb_volume.lunInfo[loop].lun,
               fsdb_volume.lunInfo[loop].sdmHandle);
    }
#endif
}

void
fsdb_data_info(void)
{
    if ( ! fsdb_volume.dataAvailable)
    {
        printf("No disk data available to dump\n");
        return;
    }
    
    printf("Current block\n");
    printf("FS  block  %16lld (0x%016llx)\n", 
           fsdb_volume.offset / FS_BLOCK_SIZE,  fsdb_volume.offset / FS_BLOCK_SIZE);
    printf("EVM block  %16lld (0x%016llx)\n", 
           fsdb_volume.offset / SDM_BLOCK_SIZE, fsdb_volume.offset / SDM_BLOCK_SIZE);
    printf("offset     %16lld (0x%016llx)\n", 
           fsdb_volume.offset, fsdb_volume.offset);
    printf("\n");
    printf("Data: %s\n", fsdb_dataTypeString[fsdb_volume.dataType]);
    printf("\n");
}

void
fsdb_info (void)
{
    int      input;
    int      def;
    int      done = FALSE;
    int      type;
    int      size;
    int      loop;

    def = 2;
    while (! done)
    {
        printf("\n");
        for (loop = 1 ; loop < FSDB_DATA_TYPE_MAX ; loop++)
        {
            printf("%2d:  %s\n", loop, fsdb_dataTypeString[loop]);
        }
        printf("%2d:  Current volume info\n", FSDB_DATA_TYPE_MAX + 0);
        printf("%2d:  Current data info\n", FSDB_DATA_TYPE_MAX + 1);
        printf("%2d:  Current dump format\n", FSDB_DATA_TYPE_MAX + 2);

        printf(" 0: Return\n");
        printf("---> Input [default: %d]: ", def);
        input = def;
        fsdb_fgets_int (&input);

        printf("\n");
        switch(input)
        {
        case FSDB_DATA_TYPE_UNKNOWN:
            printf("No assumptions about data type\n");
            printf("See 'setup dump format' and 'dump raw data'\n");
            break;
        case FSDB_DATA_TYPE_SUPER:
        {
            fs_SuperBlock *super;
            
            printf("Only FS_SUPER_CURRENT_VERSION supported\n");
            printf("Sizeof fs_SuperBlock : %d\n", sizeof(fs_SuperBlock));
            printf("Spare bytes: %d + %d\n", sizeof(super->spare), FS_BLOCK_SIZE - sizeof(*super));
            break;
        }
        case FSDB_DATA_TYPE_D_INODE:
        {
            fs_Dinode *dinode;
            
            printf("Total %d entries of %d Bytes\n", FS_DINODE_PAGE_MAX, sizeof(fs_Dinode));
            printf("For each entry:\n");
            printf(" MODE defines the decoding of the last %d bytes\n", sizeof(dinode->u));
            printf(" MODE == 0 : entry is not used\n");
            printf(" Spare bytes per entry: %d \n", sizeof(dinode->spare));
            break;
        }
        case FSDB_DATA_TYPE_REF_BYTE:
            printf("%d entries of %d bytes\n", FS_BLOCK_SIZE / sizeof(fs_RefByte), sizeof(fs_RefByte));
            printf("Spare bytes: 0\n");
            printf("
/**
 * One reference byte is allocated for every block in the file system.
 * The bytes contain a flag indicating whether the block is allocated
 * in the live file system, whether the block should never appear in a
 * snapshot, and a count of the number of snapshots the block is in.
 *
 * The bits are laid out as:
 *
 *       7       6       5       4       3       2       1       0
 *   +-----------------------------------------------+-------+-------+
 *   |           snapshot reference count            |nosnap?|islive?|
 +   +-----------------------------------------------+-------+-------+
 *
 * The refbyte for a free block is all zero bits.
 *
 * It is invalid to have the never-snapshot flag set to true in a
 * byte where the snapshot reference count is non-zero or the live
 * flag is false, and the accessor functions maintain those conditions.
 **/" 
) ;
            break;
        case FSDB_DATA_TYPE_BTREE:
            printf("Index heap\n");
            printf(" %d entries of %d bytes\n", FS_BTREE_MAXKEYS, sizeof(uint16));
            printf("keys used to order Btree\n");
            printf(" %d entries of %d bytes\n", FS_BTREE_MAXKEYS, sizeof(uint64));
            printf("left pointers\n");
            printf(" %d entries of %d bytes\n", FS_BTREE_MAXKEYS, sizeof(uint64));
            printf("Spare bytes: %d\n", FS_BLOCK_SIZE - sizeof(fs_Btree));
            break;
        case FSDB_DATA_TYPE_DIR_LEAF:
            printf("%d dir leaf entries of %d bytes\n", FS_DIRLEAF_ENTRIES_MAX, sizeof(fs_Dirent));
            printf("Spare bytes: %d\n", FS_BLOCK_SIZE - sizeof(fs_Dirleaf));
            break;
        case FSDB_DATA_TYPE_DIR_LEAF_COMPACT:
            printf("%d dir leaf entries of %d bytes\n", FS_DIRLEAF_COMPACT_ENTRIES_MAX, sizeof(fs_DirentCompact));
            break;
        case FSDB_DATA_TYPE_EMAP_LEAF:
            printf("Index heap size: %d words of %d bytes\n", FS_EMAP_EXTENTS_MAX, sizeof(uint16));
            printf("Keys (offset) to order btree: %d words of %d bytes\n", FS_EMAP_EXTENTS_MAX, sizeof(uint64));
            printf("Extents: %d blocks of %d bytes\n", FS_EMAP_EXTENTS_MAX, sizeof(fs_Extent));
            printf("Spare bytes: %d\n", FS_BLOCK_SIZE - sizeof(fs_Emapleaf));
            break;
        case FSDB_DATA_TYPE_FILE_SETS:
            printf("TBD\n");
            break;
        case FSDB_DATA_TYPE_BUDDY:
            printf("Max %d buddy levels\n", FS_MAX_BUDDY_LEVELS);
            printf("Blocks per buddy region: %d\n", FS_BLOCKS_PER_BUDDY_REGION);
            printf("First bit == 0, set means free, 0 means in use\n");
            printf("Spare bytes: 0\n");
            break;
        case FSDB_DATA_TYPE_FULL_INODE_BLOCK:
            printf("Each bit indicates if related page in inode file is full\n");
            printf("0: one or more entries are free\n");
            printf("1: dinode page is full\n");
            printf("The above is the disk format, the search functions return the inverse\n");
            printf("Spare bytes: 0\n");
            break;
        case FSDB_DATA_TYPE_DIR_IN_INODE_BLOCK:
            printf("Each bit indicates if related page in inode file has a DIRECTORY entry\n");
            printf("1: one or more directory entries in dinode page\n");
            printf("0: no directory entries in dinode page\n");
            printf("The above is the disk format, the search functions return the inverse\n");
            printf("Spare bytes: 0\n");
            break;
        case FSDB_DATA_TYPE_VOL_LOG_BUF:
            printf("Vol log header %d Bytes (only at beginning of log)\n", sizeof(vl_LogHdr));
            printf("Vol log buffer size: %d Bytes\n", VL_BUFFER_SIZE);
            printf("Disk header %d Bytes (in each volume log buffer)\n", sizeof(vl_BufDiskHdr));
            printf("Magic (first 4 bytes of each vol log buffer): 0x%x\n", VL_LOG_MAGIC);
            printf("Spare bytes: 0\n");
            break;
        case FSDB_DATA_TYPE_D_INODE_ONE_ENTRY:
            printf("See disk inode page\n");
            break;
        case (FSDB_DATA_TYPE_MAX + 0):
            fsdb_vol_info();
            break;
        case (FSDB_DATA_TYPE_MAX + 1):
            fsdb_data_info();
            break;
        case (FSDB_DATA_TYPE_MAX + 2):
            fsdb_set_debug(FSDB_SHOW_DUMP);
            break;
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", input);
            break;
        }
        printf("\nEnter to cont...");
        fsdb_fgets_int (&input);

        def++;
        if (def > FSDB_DATA_TYPE_MAX + 2)
            def = 0;
    }
}

/******************************************************************************
 * menu for utilities
 */

void
fsdb_util_menu (void)
{
    int      input;
    int      def;
    int      done = FALSE;
    int      type;
    int      size;

    while (! done)
    {
        printf("\n");
        printf("1: CRC calculations\n");
        printf("2: Print current date\n");
        printf("3: Print date for given secs\n");
        printf("4: General info\n");
        printf("0: Return\n");
        def = 0;
        printf("---> Input [default: %d]: ", def);
        input = def;
        fsdb_fgets_int (&input);

        switch(input)
        {
        case 1:
            printf("\n");
            printf ("Data pointer: %p\n", &(fsdb_volume.data));
            libc_check_crc(0, NULL);
            break;
        case 2:
        {
            struct timeval now;
            
            printf("\n");
            gettimeofday(&now, NULL);
            printf("Current date: %d (0x%x)\n", (int32)now.tv_sec, (int32)now.tv_sec);
            printf("Current date: %s", ctime((time_t *)&now.tv_sec));
            break;
        }
        case 3:
        {
            uint32 date;
            
            printf("\n");
            printf("Date in seconds since Jan 1, 1970 (dec or 0x input): ");
            fsdb_fgets_uint32(&date);
            printf("\n");
            printf("Given date: %d (0x%x)\n", date, date);
            printf("Given date: %s", ctime((time_t *)&date));
            break;
        }
        case 4:
            fsdb_info();
            break;
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", input);
            break;
        }
    }
}

/******************************************************************************
 * main menu
 */

void
fsdb_menu (void)
{
    int      input;
    int      def;
    int      done = FALSE;

    while (! done)
    {
        printf("\n");
        printf("1: Setup volume\n");
        def = 1;
        printf("2: Setup dump format \n");
        if(fsdb_volume.ready)
        {
            printf("3: Read block\n");
            def = 3;
        }
        if(fsdb_volume.dataAvailable)
        {
            printf("4: Dump block\n");
            def = 4;
        }
        printf("5: Utilities\n");
        printf("0: Exit\n");
        printf("---> Input [default: %d]: ", def);

        input = def;
        fsdb_fgets_int (&input);

        switch(input)
        {
        case 1:
            fsdb_vol_menu();
            break;
        case 2:
            fsdb_set_debug(FSDB_SET_DUMP_LOCAL);            
            break;
        case 3:
            fsdb_read_menu();
            break;
        case 4:
            fsdb_dump_block();
            break;
        case 5:
            fsdb_util_menu();
            break;
        case 0:
            return;
            break;
        default:
            printf ("Invalid input (%d)\n", input);
            break;
        }
    }
}

/******************************************************************************
 *  main
 */

/*
void
fsdb_special(void)
{
    int      i1, i2, i3, i4;
    char     c1;
    FILE    *fp;
    char     str[2000];
    char    *s = str;
    fs_LogRecord *record;
    int      logOffset = 0;
    int      recLen;
    uint32   cksum;
    
    //fp = fopen("/n/users/isabellec/log-061702", "r");
    // fp = fopen("./log-061702", "r");
    // fp = fopen("./log-061702-1", "r");
    fp = fopen("./log-061802", "r");
    // fp = fopen("./log-061802-1", "r");
    
    // skip first lines
    for (i1=0; i1<9; i1++)
    {
        fgets(str, 2000, fp);
    }
    i3 = 0;
    for (i1=0; i1< (8192/16); i1++)
    {
        fgets(str, 2000, fp);
        for (i2 = 0 ; i2 < 16 ; i2++)
        {
            sscanf (&str[13 + i2*3], "%x", &i4);
            fsdb_volume.data[i3++] = (char)(i4 & 0xff);
        }
    }
    printf("Read %d bytes\n", i3);

    cksum = crc32_checksum(fsdb_volume.data, 8192, 0);
    
    printf("cksum for EVM 0x%x   08919cfa\n", cksum);
    
    
    fsdb_volume.dataType = FSDB_DATA_TYPE_VOL_LOG_BUF;
    fsdb_volume.dataAvailable = TRUE;
    
    fsdb_dumpVollogBuf((vl_BufDiskHdr *)fsdb_volume.data);

    logOffset = sizeof(vl_BufDiskHdr);
    record = (fs_LogRecord*)&fsdb_volume.data[logOffset];
    
    while (logOffset < 8192)
    {
        printf("LogRecord at offset %d\n", logOffset);
        record = (fs_LogRecord*)&fsdb_volume.data[logOffset];
        s = str;
        recLen = fsdb_logDumpOneRecord(&s, record);

        // find valid record
        if ( ! ((record->hdr.pad == FS_LOG_MAGIC_PAD) &&
                (record->hdr.id >= FS_LOG_ID_MIN) &&
                (record->hdr.id < FS_LOG_ID_MAX) &&
                (record->hdr.recType >= FS_LOG_REC_META_ONE) &&
                (record->hdr.recType < FS_LOG_REC_TYPE_MAX) &&
                (recLen <= (FS_BLOCK_SIZE + sizeof(fs_LogRecord)))))
        {
            logOffset += 8;
            printf ("Not on record boundary: increment log offset to: %d\n", logOffset);
        }
        else
        {
            printf ("%s", str);
            logOffset += recLen;
        }

    }

}
*/

int main(int argc, char *argv[])
{
    bzero (&fsdb_volume, sizeof(fsdb_volume));
    fsdb_volume.dataType = FSDB_DATA_TYPE_UNKNOWN;

    fsdb_set_debug(FSDB_SET_DUMP_DEFAULT);

    // fsdb_special();
    
    fsdb_menu ();
    
    return 0;
}
