/*
 * Name:  mcpu-sim.c
 *
 * RCS:  $Id: mcpu-sim.c,v 1.31 2002/09/11 19:55:50 bobonfeline Exp $
 *
 * Description:  Simulation file for the MCPU
 * 
 * Created By:  
 *
 * Date Created:  7/06/2001
 */

/*
 * description of the tables
 *
 * COD_HT   Chunk Of Data - Handle Table
 * HMT      Handle Mapping Table
 * BDL      Buffer Descriptor List
 * CHQ      Cod Handle Queue
 * IFQ      Free Index Queue
 * FBQ      Free Buffer Queue
 * SAC      Storage Area Cache
 *
 * COD_HT : contains all the COD entries and ref to HMT table
 *          If not in use the hmt_index points to the next free COD entry
 * HMT    : for given COD, points to BDL
 *          If inuse, indicate the total length
 * BDL    : For given COD/HMT points to list of data entries
 *          (in simulation I use FS_BLOCK_SIZE, so only one will be inuse)
 * SAC    : contains the real data
 *          1-to-1 relation with BDL
 *
 * CHQ    : points to first/last free COD_HT entry (and current nbr of free entries)
 * IFQ    : points to first/last free HMT entry (and current nbr of free entries)
 * FBQ    : points to first/last free BDL entry (and current nbr of free entries)
 *
 */

#if !defined(SSC)

/*

gcc -g -Wall -DMAIN_MCPU -I../../Includes -I../../Includes/exports -o mmm chash.c mcpu-sim.c 

gcc -g -Wall -I../../Includes -I../../Includes/exports -o mmm mcpu-sim.c ../sm-sbm/sbm-init.c

 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/time.h>
#include <time.h>

#include "nfsproto.h"
#include "mount.h"
#include "fs-includes.h"

#include "chash-api.h"
#include "mcpu-sim.h"




#define Panic printf

#define NFX_ERROR_BASE      0xFFFFFF00
#define SBM_ERROR_BASE      (NFX_ERROR_BASE-0x400)


/* #include "sbm-gmc.h" */
#include "../sm-sbm/sbm.h"
#if !defined(FS_STANDALONE)
#include "../sm-sbm/sbm-gmc.h"
#endif

#define TST_MCPU_SAC_BLOCK_SIZE   FS_BLOCK_SIZE
// #define TST_MCPU_SAC_BLOCK_SIZE 0x1000   // 4k
// #define TST_MCPU_SAC_BLOCK_SIZE 0x08   // 8 bytes

// #define TST_MCPU_SAC_NBR_OF_BLOCKS 0x10000000 // 
//#define TST_MCPU_SAC_NBR_OF_BLOCKS 0x1000 // 4K
#define TST_MCPU_SAC_NBR_OF_BLOCKS 50000

uint32 TST_MCPU_CHUNKS_OF_NODES = TST_MCPU_SAC_NBR_OF_BLOCKS / 2 ;


char SAC [TST_MCPU_SAC_NBR_OF_BLOCKS] [TST_MCPU_SAC_BLOCK_SIZE] ;

sbm_hmtEntry_t HMT [TST_MCPU_SAC_NBR_OF_BLOCKS];
sbm_bdlEntry_t BDL [TST_MCPU_SAC_NBR_OF_BLOCKS];

sbm_hash_node_t COD_HT [TST_MCPU_SAC_NBR_OF_BLOCKS] ;

typedef struct 
{
  uint32 curr_size ;
  uint32 first ;
  uint32 last ;
} queue_t;

queue_t IFQ ;
queue_t FBQ ;
queue_t CHQ ;

uint32 COD_hid = -1 ;

uint32 sac_blocks_in_use = 0;

#define INVALID_ENTRY               0xFFFFFFFF
#define COD_NOT_IN_USE              ((nfx_cod_t)(-1))

static pthread_mutex_t mcpu_sim_lock ;
// static pthread_mutex_t mcpu_sim_lock  = {0, 0, 0, PTHREAD_MUTEX_ERRORCHECK_NP, __LOCK_INITIALIZER};

int32
mcpu_sim_restart_from_file();
int32
mcpu_sim_write_to_file();

#ifdef MAIN_MCPU

void *
eee_ramAlloc(int32 numBytes)
{
    void *ptr = malloc(numBytes);

    //printf("allocing %#x\n", ptr);

    return ptr;
}

void
eee_ramDealloc(void *ptr)
{
    //printf("freeing %#x\n", ptr);

    free(ptr);
}

#endif


/**
 * Return the total size of the SAC.
 */
uint64
mcpu_sim_totalSacSize()
{
    return (TST_MCPU_SAC_BLOCK_SIZE * TST_MCPU_SAC_NBR_OF_BLOCKS); 
}


uint32
hash_cod (const void *key)
{
  uint64 v ;
  uint32 i ;

  v = *((uint64*)key) ;
  i = (uint32)v;

  return (i);
}


int32
cmp_cod (const void *key1, const void *key2)
{
  uint64 v1, v2 ;
  v1 = *((uint64*)key1) ;
  v2 = *((uint64*)key2) ;
  if (v1 == v2)
    return 0;
  else
    return 1;
}


int32
mcpu_sim_sbm_init()
{
  int index ;
  
  bzero (COD_HT, sizeof(COD_HT));
  bzero (HMT, sizeof(HMT));
  bzero (BDL, sizeof(BDL));
 
  for (index = 0 ; index < (TST_MCPU_SAC_NBR_OF_BLOCKS - 1); index++)
    {
      HMT[index].bdl_index = (index+1) ;
      BDL[index].next_bdl_index = (index+1) ;
    }
  HMT[index].bdl_index = INVALID_ENTRY ;
  BDL[index].next_bdl_index = (BDL_EOC) ;

  IFQ.curr_size = TST_MCPU_SAC_NBR_OF_BLOCKS ;
  FBQ.curr_size = TST_MCPU_SAC_NBR_OF_BLOCKS ;
  IFQ.first = 0 ;
  IFQ.last = index ;
  FBQ.first = 0 ;
  FBQ.last = index ;
  
  COD_hid = chash_create (TST_MCPU_CHUNKS_OF_NODES,
                          hash_cod,
                          cmp_cod,
                          NULL);

  for (index = 0 ; index < (TST_MCPU_SAC_NBR_OF_BLOCKS - 1); index++)
  {
      COD_HT[index].hmt_index = (index+1) ;
      COD_HT[index].COD = COD_NOT_IN_USE ;
  }
  COD_HT[index].hmt_index = INVALID_ENTRY ;
  COD_HT[index].COD = COD_NOT_IN_USE ;
  CHQ.curr_size = TST_MCPU_SAC_NBR_OF_BLOCKS ;
  CHQ.first = 0 ;
  CHQ.last = index ;

  pthread_mutex_init (&mcpu_sim_lock, NULL);
  
  mcpu_sim_restart_from_file();

  return NFX_OK;
}

int
mcpu_sim_sbm_setCOD(nfx_cod_t COD)
{
  int index ;
  
  // locked by caller 

  if ((0 == CHQ.curr_size) ||
      (INVALID_ENTRY == CHQ.first))
    {
      Panic ("sbm_setCOD: no more COD's available\n");
      exit (-1) ;
    }

  if (chash_find (COD_hid, &COD) != NULL)
  {
      // Panic ("sbm_setCOD: COD exist already %#llx!\n", COD);
      return (-1) ;
  }

  index = CHQ.first ;
  CHQ.curr_size--;
  CHQ.first = COD_HT[index].hmt_index ;
  if (CHQ.curr_size == 0)
    CHQ.first = CHQ.last = INVALID_ENTRY ;

  /* HMT not in use, mark as INVALID_ENTRY */
  COD_HT[index].hmt_index = INVALID_ENTRY ;

  COD_HT[index].COD = COD ;

  if (chash_insert (COD_hid, &(COD_HT[index].COD), &(COD_HT[index])) < 0)
  {
      Panic ("sbm_setCOD: Wrong parameters\n");
      return (-1) ;
  }

  return (index) ;
}

int32
mcpu_sim_sbm_freeCOD(
    nfx_cod_t   COD
)
{
  sbm_hash_node_t       * cod_p ;
  int                   cod_index ;
  uint32                hmt_index;
  uint32                bdl_index, next_bdl_index;

  pthread_mutex_lock (&mcpu_sim_lock);

  if ((cod_p = (sbm_hash_node_t*)chash_find (COD_hid, &COD)) == NULL) 
    {
      printf ("sbm_freeCOD: COD (%#llx) not in tables \n", COD);
      pthread_mutex_unlock (&mcpu_sim_lock);
      return SBM_ERR_INVALID_COD;
    }

  cod_index = (cod_p - COD_HT);
  if (COD_HT[cod_index].COD != COD)
    {
      Panic ("sbm_freeCOD: index problem (%#llx) != (%#llx)\n", 
             COD_HT[cod_index].COD , COD);
      exit (-1) ;
    }

  if (INVALID_ENTRY == COD_HT[cod_index].hmt_index)
    {
      printf ("sbm_freeCOD: HMT (%#llx) was not in use\n", COD);
      pthread_mutex_unlock (&mcpu_sim_lock);
      return SBM_ERR_INVALID_COD;
    }

  hmt_index = COD_HT[cod_index].hmt_index ;
  if (hmt_index >= TST_MCPU_SAC_NBR_OF_BLOCKS)
    {
      printf ("sbm_freeCOD: HMT index not in use, %#x\n", hmt_index);
    }
  else
    {
      bdl_index = HMT[hmt_index].bdl_index;
      /* clean up list of BD entries */
      do {
        next_bdl_index = BDL[bdl_index].next_bdl_index;

        sac_blocks_in_use--;

        /* append to BD to free FBQ list */
        if (FBQ.curr_size == 0)
          {
            FBQ.first = bdl_index ;
          }
        else
          {
            BDL[FBQ.last].next_bdl_index = bdl_index;
          }
        FBQ.last = bdl_index ;
        BDL[bdl_index].next_bdl_index = INVALID_ENTRY;
        BDL[bdl_index].len = 0 ;
        BDL[bdl_index].offset = 0 ;
        FBQ.curr_size++ ;

        bdl_index = next_bdl_index;

      } while (bdl_index != BDL_EOC);


      /* append to HM to free IFQ list */
      if (IFQ.curr_size == 0)
        {
          IFQ.first = hmt_index ;
        }
      else
        {
          HMT[IFQ.last].bdl_index = hmt_index ;
        }
      IFQ.last = hmt_index ;
      HMT[hmt_index].bdl_index = INVALID_ENTRY;
      HMT[hmt_index].COD_len = 0 ;
      IFQ.curr_size++ ;     

    }
  
  if (chash_remove (COD_hid, &COD) != 0) 
    {
      printf ("sbm_freeCOD: COD (%#llx) not in tables \n", COD);
      pthread_mutex_unlock (&mcpu_sim_lock);
      return SBM_ERR_INVALID_COD;
    }

  /* append to free COD_HT list */
  if (CHQ.curr_size == 0)
    {
      CHQ.first = cod_index ;
    }
  else
    {
      COD_HT[CHQ.last].hmt_index = cod_index;
    }

  CHQ.last = cod_index ;
  COD_HT[cod_index].hmt_index = INVALID_ENTRY;
  COD_HT[cod_index].COD = COD_NOT_IN_USE ;
  CHQ.curr_size++ ;

  pthread_mutex_unlock (&mcpu_sim_lock);

  return NFX_OK;
}

int32
mcpu_sim_find_bd(
    nfx_cod_t   COD,
    uint32      *hmt_id,
    uint32      *bdl_id
)
{
  sbm_hash_node_t       * cod_p ;
  int                   cod_index ;
  uint32                bdl_index, hmt_index;

  *hmt_id = *bdl_id = INVALID_ENTRY;

  // locked by caller !!!!!

  if ((cod_p = (sbm_hash_node_t*)chash_find (COD_hid, &COD)) == NULL) 
    {
      // printf ("mcpu_sim_find_bd: COD (%#llx) not in tables \n", COD);
      return SBM_ERR_INVALID_COD;
    }

  cod_index = (cod_p - COD_HT);
  if (COD_HT[cod_index].COD != COD)
    {
      Panic ("mcpu_sim_find_bd: index problem (%#llx) != (%#llx)\n", 
             COD_HT[cod_index].COD , COD);
      exit (-1) ;
    }

  if (INVALID_ENTRY == COD_HT[cod_index].hmt_index)
    {
      printf ("mcpu_sim_find_bd: HMT (%#llx) was not in use\n", COD);
      return SBM_ERR_INVALID_COD;
    }

  hmt_index = COD_HT[cod_index].hmt_index ;
  if (hmt_index >= TST_MCPU_SAC_NBR_OF_BLOCKS)
    {
      printf ("mcpu_sim_find_bd: HMT index not in use, %#x\n", hmt_index);
      return SBM_ERR_INVALID_COD;
    }
  else
    {
      bdl_index = HMT[hmt_index].bdl_index;
      *hmt_id = hmt_index;
      *bdl_id = bdl_index;
    }

  return NFX_OK;
}


int
mcpu_sim_store_data (
    nfx_cod_t   COD,
    uint32     length,
    char       *data
)
{
  uint32                copied = 0;
  uint32                curr_size;
  uint32                cod_index;
  uint32                hmt_index;
  uint32                bdl_index, next_bdl_index;

  // locked by caller !!!!!

  cod_index = (uint32) mcpu_sim_sbm_setCOD(COD) ;

  if (cod_index < 0)
    {
      Panic ("mcpu_sim_rx_data_from_LUC: duplicate COD (%#llx)\n", 
             COD);
      exit (-1) ;
    }
  if (COD_HT[cod_index].COD != COD)
    {
      Panic ("mcpu_sim_rx_data_from_LUC: index problem (%#llx) != (%#llx)\n", 
             COD_HT[cod_index].COD , COD);
      exit (-1) ;
    }

  if (INVALID_ENTRY != COD_HT[cod_index].hmt_index)
    {
      Panic ("mcpu_sim_rx_data_from_LUC: HMT (%#llx) was already in use\n", COD);
      exit (-1) ;
    }

  /* allocate HMT index */
  if (IFQ.curr_size <= 0)
    {
      Panic ("mcpu_sim_rx_data_from_LUC: NO more IFQ entries\n");
      exit (-1) ;
    }
  hmt_index = IFQ.first;
  IFQ.first = HMT[hmt_index].bdl_index;
  IFQ.curr_size--;
  if (IFQ.curr_size == 0)
    IFQ.first = IFQ.last = INVALID_ENTRY ;
  COD_HT[cod_index].hmt_index = hmt_index ;

  /* allocate first BD index */
  if (FBQ.curr_size <= 0)
    {
      Panic ("mcpu_sim_rx_data_from_LUC: NO more FBQ entries\n");
      exit (-1) ;
    }
  bdl_index = FBQ.first;
  FBQ.first = BDL[bdl_index].next_bdl_index;
  FBQ.curr_size--;
  if (FBQ.curr_size == 0)
    FBQ.first = FBQ.last = INVALID_ENTRY ;
  HMT[hmt_index].bdl_index = bdl_index ;
  HMT[hmt_index].COD_len = length;
  sac_blocks_in_use++;

  curr_size = MIN (length - copied, TST_MCPU_SAC_BLOCK_SIZE);
  BDL[bdl_index].len = curr_size ;
  memcpy (&SAC[bdl_index][0], &data[copied], curr_size) ;
  copied += curr_size ;
  BDL[bdl_index].offset = 0 ;
  BDL[bdl_index].next_bdl_index =  BDL_EOC ;
  
  while ((length - copied) > 0) 
    {
      /* allocate next BD index */
      if ((FBQ.curr_size <= 0) ||
          (INVALID_ENTRY == FBQ.first))
        {
          Panic ("mcpu_sim_rx_data_from_LUC: NO more FBQ entries\n");
          exit (-1) ;
        }
      next_bdl_index = FBQ.first;
      FBQ.first = BDL[next_bdl_index].next_bdl_index;
      FBQ.curr_size--;
      if (FBQ.curr_size == 0)
        FBQ.first = FBQ.last = INVALID_ENTRY ;
      sac_blocks_in_use++;
      
      curr_size = MIN (length - copied, TST_MCPU_SAC_BLOCK_SIZE);
      BDL[next_bdl_index].len = curr_size ;
      memcpy (&SAC[next_bdl_index][0], &data[copied], curr_size) ;
      copied += curr_size ;
      BDL[next_bdl_index].offset = 0 ;
      BDL[next_bdl_index].next_bdl_index =  BDL_EOC ;
      BDL[bdl_index].next_bdl_index = next_bdl_index ;
      
      bdl_index = next_bdl_index ;
    }

  return NFX_OK;  
}


/*
 *  returns the length of copied bytes
 *  returns negative if failure
 */

int
mcpu_sim_retrieve_data (
    char            *data,
    nfx_cod_t       COD,
    uint32          offset,
    uint32          length
)
{
  sbm_hash_node_t       * cod_p ;
  uint32                copied = 0;
  uint32                curr_size;
  uint32                curr_off;
  uint32                cod_index;
  uint32                hmt_index;
  uint32                bdl_index, next_bdl_index;

  // !! locked by caller !!

  if ((cod_p = (sbm_hash_node_t*)chash_find (COD_hid, &COD)) == NULL) 
    {
      printf ("mcpu_sim_retrieve_data: COD (%#llx) not in tables \n", COD);
      return NFX_ERR;
    }
  cod_index = (cod_p - COD_HT);

  if (COD_HT[cod_index].COD != COD)
    {
      Panic ("mcpu_sim_retrieve_data: index problem (%#llx) != (%#llx)\n", 
             COD_HT[cod_index].COD , COD);
      exit (-1) ;
    }

  if (INVALID_ENTRY == COD_HT[cod_index].hmt_index)
    {
      Panic ("mcpu_sim_retrieve_data: HMT (%#llx) was not in use\n", COD);
      exit (-1) ;
    }

  hmt_index = (COD_HT[cod_index].hmt_index);
  if (hmt_index >= TST_MCPU_SAC_NBR_OF_BLOCKS)
    {
      printf ("mcpu_sim_retrieve_data: HMT index not in use, %#x\n", hmt_index);
      return NFX_ERR;
    }
  else
    {
      curr_off = offset ;
      bdl_index = HMT[hmt_index].bdl_index ;
      if (BDL[bdl_index].len > curr_off)
        {
          curr_size = MIN (length - copied, BDL[bdl_index].len - curr_off);
          memcpy (&data[copied], &SAC[bdl_index][curr_off], curr_size) ;
          copied += curr_size ;
          curr_off = 0 ;
        }
      else
        {
          curr_off -= BDL[bdl_index].len ;
        }
      
      bdl_index = BDL[bdl_index].next_bdl_index ;

      while (((length - copied) > 0) &&
             (bdl_index != BDL_EOC))  
        {
          if (copied >= (HMT[hmt_index].COD_len - offset))
            {
              break;
            }

          next_bdl_index = BDL[bdl_index].next_bdl_index ;
          
          if (BDL[bdl_index].len > curr_off)
            {
              curr_size = MIN (length - copied, BDL[next_bdl_index].len - curr_off);
              memcpy (&data[copied], &SAC[next_bdl_index][curr_off], curr_size) ;
              copied += curr_size ;
              curr_off = 0 ;
            }
          else
            {
              curr_off -= BDL[bdl_index].len ;
            }

          bdl_index = next_bdl_index ;
        }
    }

  return (copied);
}


int32
mcpu_sim_sbm_showCOD (
    nfx_cod_t   COD,
    int32 level
)
{
  sbm_hash_node_t       * cod_p ;
  uint32                cod_index;
  uint32                hmt_index;
  uint32                bdl_index, next_bdl_index;

  pthread_mutex_lock (&mcpu_sim_lock);

  printf ("COD = %#llx (%lld)\n", COD, COD);
  if (level >= MCPU_SHOW_ALL)
    chash_stat (COD_hid) ;

  if ((cod_p = (sbm_hash_node_t*)chash_find (COD_hid, &COD)) == NULL) 
    {
      printf ("sbm_freeCOD: COD (%#llx) not in tables \n", COD);
      pthread_mutex_unlock (&mcpu_sim_lock);
      return SBM_ERR_INVALID_COD;
    }

  cod_index = (cod_p - COD_HT);
  if (COD_HT[cod_index].COD != COD)
    {
      Panic ("sbm_showCOD: index problem (%#llx) != (%#llx)\n", 
             COD_HT[cod_index].COD , COD);
      exit (-1) ;
    }

  if (INVALID_ENTRY == COD_HT[cod_index].hmt_index)
    {
      printf ("sbm_showCOD: HMT (%#llx) was not in use\n", COD);
      pthread_mutex_unlock (&mcpu_sim_lock);
      return SBM_ERR_INVALID_COD;
    }

  hmt_index = COD_HT[cod_index].hmt_index ;
  if (hmt_index >= TST_MCPU_SAC_NBR_OF_BLOCKS)
    {
      printf ("sbm_showCOD: HMT index not in use, %#x\n", hmt_index);
    }
  else
    {
      if (level >= MCPU_SHOW_DETAIL)
        printf ("  HMT index 0x%8x, len %8d\n", 
                hmt_index, HMT[hmt_index].COD_len);
      bdl_index = HMT[hmt_index].bdl_index;

      do {
        if (level >= MCPU_SHOW_DETAIL)
          printf ("  BDL index 0x%8x, len %5d, off %5d\n", 
                  bdl_index, BDL[bdl_index].len, BDL[bdl_index].offset);
        next_bdl_index = BDL[bdl_index].next_bdl_index;
        
        bdl_index = next_bdl_index;

      } while (bdl_index != BDL_EOC);

    }

  printf ("  CHQ  size %d, free %d \n", 
          TST_MCPU_SAC_NBR_OF_BLOCKS, CHQ.curr_size);
  printf ("  IFQ  size %d, free %d \n", 
          TST_MCPU_SAC_NBR_OF_BLOCKS, IFQ.curr_size);
  printf ("  FBQ  size %d, free %d \n", 
          TST_MCPU_SAC_NBR_OF_BLOCKS, FBQ.curr_size);
  printf ("  SAC Blocks in use %d \n", 
          sac_blocks_in_use);
  printf ("\n");
  
  pthread_mutex_unlock (&mcpu_sim_lock);

  return NFX_OK;
}

int32
mcpu_sim_sbm_show_all (int32 level)
{
  nfx_cod_t       *COD_p;
  sbm_hash_node_t  *node_p;
  int32  nbr = 0 ;
  
  chash_reset (COD_hid);
  
  while (1 == chash_next (COD_hid, (void**)&COD_p, (void**)&node_p))
    {
      mcpu_sim_sbm_showCOD (*COD_p, level) ;
      nbr++ ;
    }
  
  printf ("  Total number of COD:  %d \n\n", nbr);
  
  return (nbr);   
}


int32
mcpu_sim_sbm_show_sum (void)
{
    int32               nbr_cod = 0 ;
    int32               nbr_hmt = 0 ;
    int32               nbr_bdl = 0 ;
    int32               hmt_bytes = 0 ;
    int32               bdl_bytes = 0 ;

    uint32              array_sz;
    uint32              elements;
    uint32              curr_max_size;
    uint32              max_size;
    uint32              loop ;

    // nfx_cod_t            *COD_p;
    sbm_hash_node_t     *node_p;
    sbm_hash_node_t     *cod_p ;
    uint32              cod_index;
    uint32              hmt_index;
    uint32              bdl_index;

    pthread_mutex_lock (&mcpu_sim_lock);

    chash_reset (COD_hid);
    
    while (1 == chash_next (COD_hid, (void**)&cod_p, (void**)&node_p))
      {
        nbr_cod++ ;

        // cod_p = chash_find (COD_hid, *COD_p) ;

        cod_index = (node_p - COD_HT);

        if (COD_HT[cod_index].COD != *((nfx_cod_t*)cod_p))
          {
            Panic ("sbm_showCOD: index problem (%#llx) != (%#llx)\n", 
                   COD_HT[cod_index].COD , *((nfx_cod_t*)cod_p));
            exit (-1) ;
          }

        if (INVALID_ENTRY == COD_HT[cod_index].hmt_index)
            continue ;

        hmt_index = COD_HT[cod_index].hmt_index ;
        if (hmt_index >= TST_MCPU_SAC_NBR_OF_BLOCKS)
          {
            continue ;
          }
        else
          {
            nbr_hmt++ ;
            hmt_bytes += HMT[hmt_index].COD_len ;

            bdl_index = HMT[hmt_index].bdl_index;

            do {
              nbr_bdl++ ;
              bdl_bytes += BDL[bdl_index].len ;

              bdl_index = BDL[bdl_index].next_bdl_index;

            } while (bdl_index != BDL_EOC);
          }
      }

    chash_stat_nbr (COD_hid, &array_sz, &elements, &curr_max_size, &max_size);

    printf ("Nbr COD: %d, hmt %d, bdl %d, bytes %d, %d \n", 
            nbr_cod, nbr_hmt, nbr_bdl, hmt_bytes, CHQ.curr_size);

    if (hmt_bytes != bdl_bytes)
      printf ("Data in tables not consistent %d\n", bdl_bytes);

    printf ("hash array %d, elements %d, curr size %d, max size %d \n", 
            array_sz, elements, curr_max_size, max_size);

    if (CHQ.curr_size == 0)
    {
        if (CHQ.first != INVALID_ENTRY)
            printf ("mcpu-sim: COD: wrong CHQ.first\n");
        if (CHQ.last != INVALID_ENTRY)
            printf ("mcpu-sim: COD: wrong CHQ.last\n");
        if (nbr_cod != TST_MCPU_SAC_NBR_OF_BLOCKS)
            printf ("mcpu-sim: COD: wrong nbr_cod\n");
    }
    else
    {
        cod_index = CHQ.first ;
        for (loop = 0 ; loop < (CHQ.curr_size - 1) ; loop ++)
        {
            if (COD_HT[cod_index].COD != COD_NOT_IN_USE)
                printf ("mcpu-sim: COD: wrong free cod %d  %#llx\n", loop, COD_HT[cod_index].COD);

            if (cod_index == INVALID_ENTRY)
                printf ("mcpu-sim: COD: wrong cod_index %d %d\n", loop, CHQ.curr_size);
            else
                cod_index = COD_HT[cod_index].hmt_index ;
        }
        if (COD_HT[cod_index].hmt_index != INVALID_ENTRY)
            printf ("mcpu-sim: COD: wrong last entry %d\n", CHQ.curr_size);
        if (CHQ.last != cod_index)
            printf ("mcpu-sim: COD: wrong CHQ.last %d\n", CHQ.curr_size);
        if (nbr_cod != (TST_MCPU_SAC_NBR_OF_BLOCKS - CHQ.curr_size))
            printf ("mcpu-sim: COD: wrong nbr_cod %d \n", CHQ.curr_size);
    }

    pthread_mutex_unlock (&mcpu_sim_lock);

    return NFX_OK;  
}



int
mcpu_sim_rx_data_from_LUC (
    mcpu_luc_msg_req_t *mcpuReq
    )
{
	int        rc = 0;
    uint32     bytesThis, loop, length;
    char       *data;

    pthread_mutex_lock (&mcpu_sim_lock);

    length = 0;
    data = &(mcpuReq->data[0]);
    
    for (loop = 0 ; loop < mcpuReq->nbrUsedCod; loop++)
    {
        bytesThis = MIN (mcpuReq->writeLength - length, FS_BLOCK_SIZE);
        
        rc |= mcpu_sim_store_data (mcpuReq->cod[loop].cod, bytesThis, data);
        FS_ASSERT (mcpuReq->cod[loop].offset == 0);
        FS_ASSERT (mcpuReq->cod[loop].length == bytesThis);

        length += bytesThis;
        data += bytesThis;
    }

    FS_ASSERT (length == mcpuReq->writeLength);
        
    pthread_mutex_unlock (&mcpu_sim_lock);

	return (rc);
}

int
mcpu_sim_tx_data_to_LUC (
    fs_MsgReadResp  **fsReadResp
    )
{
    fs_MsgReadResp  *ReadResp;    // new resp with the data
    nfx_cod_t       cod;
    int32           index;
    char            *data;
    uint32          loop, length, copied;

    length = (*fsReadResp)->bytesRead;
    if (length == 0)
    {
        (*fsReadResp)->data_index = (*fsReadResp)->hdr.len ;
        return (NFX_OK);
    }

    pthread_mutex_lock (&mcpu_sim_lock);

    index = (*fsReadResp)->hdr.len ;
    ReadResp = fs_ramAlloc(FS_RAM_RESP_MSG, index + length);
    memcpy (ReadResp, *fsReadResp, index) ;
    ReadResp->data_index = index ;
    data = &(((char*)ReadResp)[index]);
    ReadResp->hdr.len = index + length ;
    fs_ramDealloc(FS_RAM_RESP_MSG, *fsReadResp);
    *fsReadResp = ReadResp;

    copied = 0;
    FS_ASSERT (ReadResp->nbrUsedCod < NFX_MAX_COD_MSG);

    for (loop = 0 ; loop < ReadResp->nbrUsedCod ; loop++)
    {
        cod = ReadResp->cod[loop].cod;
        copied += mcpu_sim_retrieve_data (&(data[copied]), 
                                          cod, 
                                          ReadResp->cod[loop].offset, 
                                          ReadResp->cod[loop].length);
    }
    
    if (copied != ReadResp->bytesRead)
    {
        ReadResp->hdr.fsStatus = FS_ERR_GENERIC;
        ReadResp->bytesRead = 0;
        ReadResp->hdr.len = index ;
        pthread_mutex_unlock (&mcpu_sim_lock);
        FS_ASSERT (FALSE);
        return (NFX_ERR);
    }

    pthread_mutex_unlock (&mcpu_sim_lock);

    return (NFX_OK);
}


ssize_t
mcpu_sim_evm_write_data (
    nfx_cod_t cod,
    int fd, 
    evm_volSize_t length)
{
    ssize_t numWritten = 0;
    uint32  offset = 0;
    char *data ;
    int copied;

    pthread_mutex_lock (&mcpu_sim_lock);

    data = fs_ramAlloc(FS_RAM_FS_COD, (int32)length);

    copied = mcpu_sim_retrieve_data (data, cod, offset, (uint32)length);

    if ((copied >= 0) &&
        (copied < length))
      {
        memset(&data[copied], 0, length - copied);
      }

    if (copied >= 0)
      numWritten = write(fd,
                         data,
                         length);
    
    fs_ramDealloc(FS_RAM_FS_COD, data);

    pthread_mutex_unlock (&mcpu_sim_lock);

    return (numWritten);
}

int 
mcpu_sim_evm_read_data (
    char *data, 
    evm_volSize_t length,
    nfx_cod_t *COD)
{
  int rc;
  
  pthread_mutex_lock (&mcpu_sim_lock);

  *COD = fs_sim_get_new_cod(FS_SIM_MSG_ADDR_MCPU);

  rc = mcpu_sim_store_data (*COD, length, data);

  pthread_mutex_unlock (&mcpu_sim_lock);

  if (rc != NFX_OK)
    {
      printf ("mcpu_sim_evm_read_data: mcpu_sim_store_data failed %d\n", rc);
      mcpu_sim_sbm_freeCOD (*COD);
    }
  
  return (rc);
}


int32
mcpu_sim_split (
    nfx_cod_t   src_COD,
    uint32      src_offset,
    nfx_cod_t   dst_COD)
{
  uint32      src_hmt_id;
  uint32      src_bdl_id;
  uint32      dst_hmt_id;
  uint32      dst_bdl_id;
  uint32      len;
  char        buffer[FS_BLOCK_SIZE];
  
  if (TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_split TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE\n");
      exit (-1) ;
    }
  
  if (src_offset > FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_split (offset + len) > FS_BLOCK_SIZE\n");
      exit (-1) ;
    }
  
  pthread_mutex_lock (&mcpu_sim_lock);

  if (NFX_OK != mcpu_sim_find_bd (src_COD, &src_hmt_id, &src_bdl_id))
    {
      printf ("mcpu_sim_split src_COD not found\n");
      exit (-1) ;
    }
  if (NFX_OK == mcpu_sim_find_bd (dst_COD, &dst_hmt_id, &dst_bdl_id))
    {
      printf ("mcpu_sim_split dst_COD already exist\n");
      exit (-1) ;
    }

  len = HMT[src_hmt_id].COD_len - src_offset ;

  if (len < 0)
    {
      printf ("mcpu_sim_split length or src_offset wrong\n");
      exit (-1) ;
    }


  memcpy (&(buffer[0]), 
          &(SAC[src_bdl_id][src_offset]), 
          len);

  HMT[src_hmt_id].COD_len -= len ;
  BDL[src_bdl_id].len -= len ;

  mcpu_sim_store_data (dst_COD, len, &(buffer[0]));

  pthread_mutex_unlock (&mcpu_sim_lock);

  return NFX_OK;
}


int32
mcpu_sim_merge (
    nfx_cod_t   src_COD,
    uint32      src_len,
    uint32      src_offset,
    nfx_cod_t   dst_COD,
    uint32      dst_offset)
{
  uint32      src_hmt_id;
  uint32      src_bdl_id;
  uint32      dst_hmt_id;
  uint32      dst_bdl_id;
  
  if (TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_merge TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE\n");
      exit (-1) ;
    }
  
  if ((dst_offset + (src_len - src_offset)) > FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_merge (offset + len) > FS_BLOCK_SIZE\n");
      exit (-1) ;
    }

  pthread_mutex_lock (&mcpu_sim_lock);

  if (NFX_OK != mcpu_sim_find_bd (src_COD, &src_hmt_id, &src_bdl_id))
    {
      printf ("mcpu_sim_merge src_COD not found\n");
      exit (-1) ;
    }
  
  if (NFX_OK != mcpu_sim_find_bd (dst_COD, &dst_hmt_id, &dst_bdl_id))
    {
      printf ("mcpu_sim_merge dst_COD not found\n");
      exit (-1) ;
    }

  memcpy (&(SAC[dst_bdl_id][dst_offset]), 
          &(SAC[src_bdl_id][src_offset]), 
          src_len);

  HMT[dst_hmt_id].COD_len = MAX (HMT[dst_hmt_id].COD_len, dst_offset + src_len);
  BDL[dst_bdl_id].len = MAX (HMT[dst_hmt_id].COD_len, dst_offset + src_len);

  pthread_mutex_unlock (&mcpu_sim_lock);

  // what about source COD
  mcpu_sim_sbm_freeCOD (src_COD);

  return NFX_OK;
}

int32
mcpu_sim_copy (
    nfx_cod_t   src_COD,
    uint32      src_len,
    uint32      src_offset,
    nfx_cod_t   dst_COD,
    uint32      dst_offset)
{
  uint32      src_hmt_id;
  uint32      src_bdl_id;
  uint32      dst_hmt_id;
  uint32      dst_bdl_id;
  
  if (TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_copy TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE\n");
      exit (-1) ;
    }
  
  if ((dst_offset + (src_len - src_offset)) > FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_copy (offset + len) > FS_BLOCK_SIZE\n");
      exit (-1) ;
    }

  pthread_mutex_lock (&mcpu_sim_lock);

  if (NFX_OK != mcpu_sim_find_bd (src_COD, &src_hmt_id, &src_bdl_id))
    {
      printf ("mcpu_sim_copy src_COD not found\n");
      exit (-1) ;
    }
  
  if (NFX_OK != mcpu_sim_find_bd (dst_COD, &dst_hmt_id, &dst_bdl_id))
    {
      printf ("mcpu_sim_copy dst_COD not found\n");
      exit (-1) ;
    }

  memcpy (&(SAC[dst_bdl_id][dst_offset]), 
          &(SAC[src_bdl_id][src_offset]), 
          src_len);

  HMT[dst_hmt_id].COD_len = MAX (HMT[dst_hmt_id].COD_len, dst_offset + src_len);
  BDL[dst_bdl_id].len = MAX (HMT[dst_hmt_id].COD_len, dst_offset + src_len);

  pthread_mutex_unlock (&mcpu_sim_lock);

  return NFX_OK;
}

int32
mcpu_sim_cod_null (
    nfx_cod_t   COD,
    uint32      len,
    uint32      offset)
{
  uint32      hmt_id;
  uint32      bdl_id;
  
  if (TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_cod_null TST_MCPU_SAC_BLOCK_SIZE != FS_BLOCK_SIZE\n");
      exit (-1) ;
    }
  
  if ((offset + len) > FS_BLOCK_SIZE)
    {
      printf ("mcpu_sim_cod_null (offset + len) > FS_BLOCK_SIZE\n");
      exit (-1) ;
    }
  
  pthread_mutex_lock (&mcpu_sim_lock);

  if (NFX_OK != mcpu_sim_find_bd (COD, &hmt_id, &bdl_id))
    {
      // just dummy data to create entry
      mcpu_sim_store_data (COD, sizeof(len), (char*)&len);

      if (NFX_OK != mcpu_sim_find_bd (COD, &hmt_id, &bdl_id))
        {
          printf ("mcpu_sim_cod_null creation failed\n");
          exit (-1) ;
        }
      
    }
  
  memset (&(SAC[bdl_id][offset]), 0, len);

  HMT[hmt_id].COD_len = MAX (HMT[hmt_id].COD_len, offset + len);
  BDL[bdl_id].len = MAX (HMT[hmt_id].COD_len, offset + len);

  pthread_mutex_unlock (&mcpu_sim_lock);

  return NFX_OK;
}


#define MCPU_SIM_COD_SIZE    16
#define MCPU_SIM_MAGIC_SIZE  8
#define MCPU_SIM_SIZE_SIZE   4
#define MCPU_SIM_MAGIC       "MAGICMCP"
#define MCPU_SIM_MAGIC_EOF   "MAGICEOF"

int32
mcpu_sim_restart_from_file()
{
  int fd;
  char datap [FS_BLOCK_SIZE+4];
  int reading = 1;
  int numRead;
  nfx_cod_t   COD;
  int nbr = 0;
  int copied ;
  
  pthread_mutex_lock (&mcpu_sim_lock);

  fd = open("FS_EAGLE_SIM_COD", O_RDONLY);
  if (fd > 0)
	{
	  while (reading)
	  {
		numRead = read(fd, datap, MCPU_SIM_MAGIC_SIZE);
		if ((numRead == MCPU_SIM_MAGIC_SIZE) &&
			(strncmp (datap, MCPU_SIM_MAGIC_EOF, MCPU_SIM_MAGIC_SIZE) == 0))
		  break;

		if ( ! ((numRead == MCPU_SIM_MAGIC_SIZE) &&
				(strncmp (datap, MCPU_SIM_MAGIC, MCPU_SIM_MAGIC_SIZE) == 0)))
		  {
			printf ("mcpu_sim_restart_from_file: Magic failed \n");
			break;
		  }

		bzero (datap, MCPU_SIM_COD_SIZE+4);
		numRead = read(fd, datap, MCPU_SIM_COD_SIZE);
		if (numRead != MCPU_SIM_COD_SIZE)
		  {
			printf ("mcpu_sim_restart_from_file: Read COD failed\n");
			break;
		  }
		sscanf (datap, "%llx", &COD);
		// in store data --> (mcpu_sim_sbm_setCOD (COD) < 0)
		
		bzero (datap, MCPU_SIM_SIZE_SIZE+4);
		numRead = read(fd, datap, MCPU_SIM_SIZE_SIZE);
		if (numRead != MCPU_SIM_SIZE_SIZE)
		  {
			printf ("mcpu_sim_restart_from_file: Read size failed\n");
			break;
		  }
		sscanf (datap, "%d", &copied);
		if (copied > FS_BLOCK_SIZE)
		  {
			printf ("mcpu_sim_restart_from_file: read size failed\n");
			break;
		  }

		numRead = read(fd, datap, copied);
		if (numRead != copied)
		  {
			printf ("mcpu_sim_restart_from_file: Read data failed\n");
			break;
		  }

		if (mcpu_sim_store_data (COD, copied, datap) < 0)
		  {
			printf ("mcpu_sim_restart_from_file: mcpu_sim_store_data failed\n");
			break;
		  }

		// printf ("Read COD %d %d %#llx\n", nbr, copied, COD);
		nbr++;
	  }

	  close (fd);
	}

  if (nbr)
	printf ("mcpu_sim_restart_from_file: restored %d COD's\n", nbr);

  pthread_mutex_unlock (&mcpu_sim_lock);

  return (nbr);
}

int32
mcpu_sim_write_to_file()
{
  int fd;
  char data [FS_BLOCK_SIZE+4];
  char cod_str [MCPU_SIM_COD_SIZE+4];
  char size_str [8];
  int numWritten;
  nfx_cod_t        *COD_p;
  sbm_hash_node_t  *node_p;
  int nbr = 0 ;
  int copied;
  int openFlags = 0;
  static int write_once = 0 ;

  pthread_mutex_lock (&mcpu_sim_lock);

  if (write_once)
	{
	  pthread_mutex_unlock (&mcpu_sim_lock);
	  return (0);
	}
  

  fd = open("FS_EAGLE_SIM_COD",  openFlags | O_RDWR | O_CREAT, S_IREAD | S_IWRITE);

  if ((fd > 0) &&
	  (COD_hid >= 0))
	{

	  chash_reset (COD_hid);

	  while (1 == chash_next (COD_hid, (void**)&COD_p, (void**)&node_p))
	  {
		numWritten = write(fd, MCPU_SIM_MAGIC, MCPU_SIM_MAGIC_SIZE);
		if (numWritten != MCPU_SIM_MAGIC_SIZE)
		  break;

		sprintf (cod_str, "%16llx", *COD_p);
		numWritten = write(fd, cod_str, MCPU_SIM_COD_SIZE);
		if (numWritten != MCPU_SIM_COD_SIZE)
		  break;

	    copied = mcpu_sim_retrieve_data (data, *COD_p, 0, FS_BLOCK_SIZE);
		if (copied < 0)
		  break;
		if (copied > FS_BLOCK_SIZE)
		  break;
		
		sprintf (size_str, "%4d", copied);
		numWritten = write(fd, size_str, MCPU_SIM_SIZE_SIZE);
		if (numWritten != MCPU_SIM_SIZE_SIZE)
		  break;

		numWritten = write(fd, data, copied);

		if (numWritten != copied)
		  break;

		// printf ("Write COD %d %d %#llx\n", nbr, numWritten, *COD_p);
		nbr++;
	  }

	  numWritten = write(fd, MCPU_SIM_MAGIC_EOF, MCPU_SIM_MAGIC_SIZE);

	  close (fd);

	  write_once = 1 ;
	}

  pthread_mutex_unlock (&mcpu_sim_lock);

  if (nbr)
	printf ("mcpu_sim_write_to_file: stored %d COD's\n", nbr);

  return (nbr);
}




#ifdef MAIN_MCPU

char *test_str = "Hello bug world, here is it" ;

int
main (int argc, char* argv[])
{
  struct timeval tv1, tv2 ;
  double tt ;
  int   rc;
  nfx_cod_t tst_cod[10];
  nfx_cod_t curr_COD = 67436672934;   // just start with a rnd cod number

  gettimeofday (&tv1, NULL);    /*  since JAN_1970  */

  if ((rc = mcpu_sim_sbm_init()) != NFX_OK)
    printf("sbm init error %#x\n", rc);

  tst_cod[0] = curr_COD++;

  mcpu_sim_rx_data_from_LUC (tst_cod[0], strlen(test_str), test_str);
  mcpu_sim_sbm_showCOD (tst_cod[0],3);

  tst_cod[1] = curr_COD++;

  mcpu_sim_rx_data_from_LUC (tst_cod[1], 5, test_str);
  mcpu_sim_sbm_showCOD (tst_cod[1], 2);

  tst_cod[2] = curr_COD++;

  mcpu_sim_rx_data_from_LUC (tst_cod[2], strlen(test_str), test_str);
  mcpu_sim_sbm_showCOD (tst_cod[2], 2);

  printf ("***\n");
  mcpu_sim_sbm_show_all (3);
  printf ("***\n");

  mcpu_sim_sbm_freeCOD (tst_cod[1]);
  mcpu_sim_sbm_freeCOD (tst_cod[2]);
  mcpu_sim_sbm_freeCOD (tst_cod[0]);

  tst_cod[0] = curr_COD++;

  mcpu_sim_sbm_showCOD (tst_cod[0], 3);
  mcpu_sim_sbm_freeCOD (tst_cod[0]);

  mcpu_sim_sbm_show_all (1);
  mcpu_sim_sbm_show_all (2);
  mcpu_sim_sbm_show_all (3);
  
  gettimeofday (&tv2, NULL);
  tt = (tv2.tv_sec - tv1.tv_sec) + 
    (double)(tv2.tv_usec - tv1.tv_usec) / 1000000.0 ;
  printf ("Total time to run %10.6f\n", tt) ;

  return (0);
}

#else

// See in fs-sim.c  fs_simMcpuExecThreadMain

#endif /* MAIN_MCPU */
#endif // SSC


