/*
 * Name:  replaybitmaptest.c
 *
 * RCS:  $Id: replaybitmaptest.c,v 1.1 2002/04/09 15:03:21 bobonfrogger Exp $
 *
 * Test program for replay bitmap.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

typedef char         uchar8;
typedef int          int32;
typedef unsigned int uint32;
typedef int          boolean;

#define fs_LogId    uint32

#define PRIVATE

typedef struct fs_LogRecHdr
{
    uint32 s[4];
}fs_LogRecHdr;


#define FS_LOG_MAX_NBR_BLOCKS    ((16 * 1024 * 1024) / FS_BLOCK_SIZE)
#define FS_BLOCK_SIZE            8192
#define BITS_PER_BYTE            8

#define FS_RAM_CORE_MISC         1

#define FALSE                    0

#define FS_LOG_ID_MIN      1
//#define FS_LOG_ID_MAX      0xfeedb00f   // Highest valid transaction id
//#define FS_LOG_ID_EOL      0xfeedbeef   // Indicates end of log.
//#define FS_LOG_ID_INVALID  0xfeedbeff   // Indicates that record does not belong to 
                                        // to a transcation (trash records during replay)
#define FS_LOG_ID_MAX      1000
#define FS_LOG_ID_EOL      1001
#define FS_LOG_ID_INVALID  1002

#define fs_ramAlloc(type, size)    malloc(size)
#define fs_ramDealloc(type, ptr)   free(ptr)

#define FS_ASSERT   assert


typedef struct fs_LogReplayBitmap
{
    uchar8  *bits;      // bitmap
    int32   numBits;    // size of bitmap in bits
    uint32  firstId;    // transaction id at index 0
	uint32  maxForwIndex;
	uint32  maxRevIndex;
} fs_LogReplayBitmap;

fs_LogReplayBitmap bitmap = {NULL, 0, -1, 0, 0};







/**
 * Release memory allocated for replay bitmap.
 */
void
fs_logReleaseReplayBitmap(
    fs_LogReplayBitmap  *bitmap)
{
    if (bitmap->bits)
    {
        fs_ramDealloc(FS_RAM_CORE_MISC, bitmap->bits);
        bitmap->bits = NULL;
    }
}


/**
 * Initialize the bitmap for recording which transactions
 * will be replayed.
 */
void
fs_logInitReplayBitmap(
    fs_LogReplayBitmap  *bitmap)
{
    uint32  numBits;
    uint32  numBytes;
    
	bzero(bitmap, sizeof *bitmap);

    // Simply allocate enough bits to cover the maximum
    // log size with nothing but commit records in it.
    // This would not be a valid log, since there must be
    // something other than just commits of nothing, but
    // this number will serve to provide for handling
    // all the ids in our circular bitmap.
    numBits = (FS_LOG_MAX_NBR_BLOCKS * FS_BLOCK_SIZE) / sizeof(fs_LogRecHdr);

    numBits = 2 * 8;
    
    numBytes = numBits / BITS_PER_BYTE;

    bitmap->bits = fs_ramAlloc(FS_RAM_CORE_MISC, numBytes);
    bzero(bitmap->bits, numBytes);

    bitmap->numBits = numBits;
    bitmap->firstId = FS_LOG_ID_INVALID;

	bitmap->maxRevIndex = numBits;

    FS_ASSERT(FS_LOG_ID_MAX > bitmap->numBits);

    //printf("initReplayBitmap numBytes %d numBits %d nbrCommits %d\n",
    //       numBytes, bitmap->numBits, logInfo->nbrCommits);
}

#define FS_LOG_INDEX_NOT_IN_BITMAP     -2

/**
 * Return the bit index in the replayBitmap for the
 * specified id.
 */
PRIVATE uint32
fs_logReplayBitmapGetIndex(
    fs_LogReplayBitmap  *bitmap,
    fs_LogId            id)
{
    uint32  index, zeroIndex;

    FS_ASSERT(id >= FS_LOG_ID_MIN &&
              id <= FS_LOG_ID_MAX);

    if (bitmap->firstId == FS_LOG_ID_INVALID)
    {
        // Initialize our point of reference in the bitmap.
printf("setup bitmap->firstId to %d\n", id);
        bitmap->firstId = id;
        return 0;
    }
    
    if (id == bitmap->firstId)
    {
        index = 0;
    }
    else 
    if (id > bitmap->firstId)
    {
        index = id - bitmap->firstId;

 		if (index >= bitmap->maxRevIndex)
		{
            if (bitmap->firstId >= bitmap->numBits)
            {
                // collison
                index = FS_LOG_INDEX_NOT_IN_BITMAP ;
            }
            else
            {
                zeroIndex = bitmap->numBits - bitmap->firstId;
               
                if (zeroIndex <= bitmap->maxForwIndex)
                {
                    // collison
                    index = FS_LOG_INDEX_NOT_IN_BITMAP ;
                }
                else
                if ((FS_LOG_ID_MAX - id) >= (zeroIndex - bitmap->maxForwIndex))
                {
                    // collison
                    index = FS_LOG_INDEX_NOT_IN_BITMAP ;
                }
                else
                {
                    index = zeroIndex - (FS_LOG_ID_MAX - id);

                    FS_ASSERT(index > bitmap->maxForwIndex);
                
                    if (index < bitmap->maxRevIndex)
                    {
                        bitmap->maxRevIndex = index;
                    }
                }
            }
		}
        else
        {
            if (index > bitmap->maxForwIndex)
            {
                bitmap->maxForwIndex = index;
            }
        }
	}
    else
    {

        index = bitmap->firstId - id;

        if (index > bitmap->numBits)
        {
            // check the forward wrap around case
        
            index = (FS_LOG_ID_MAX - bitmap->firstId) + id ;
            if (index >= bitmap->maxRevIndex)
            {
                // collison, out of available bitmap
                index = FS_LOG_INDEX_NOT_IN_BITMAP ;
            }
            else
            {
                if (index > bitmap->maxForwIndex)
                {
                    bitmap->maxForwIndex = index;
                }
            }
        }
        else
        {
            index = bitmap->numBits - (bitmap->firstId - id) ;
        
            if ((index == 0) || (index <= bitmap->maxForwIndex))
            {
                // collison, not enough space
                index = FS_LOG_INDEX_NOT_IN_BITMAP ;
            }
            else
            {
                if (index < bitmap->maxRevIndex)
                {
                    bitmap->maxRevIndex = index;
                }
            }
        }

    }

    FS_ASSERT((index == FS_LOG_INDEX_NOT_IN_BITMAP) || (index < bitmap->numBits));

    //printf("id %d index %d\n", id, index);
    return index;
}


/** 
 * Return TRUE if the specified id's bit is set
 * in the replayBitmap, FALSE otherwise.
 */
PRIVATE boolean
fs_logReplayBitmapIsSet(
    fs_LogReplayBitmap  *bitmap,
    fs_LogId            id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(bitmap, id);
    uint32 byteIndex = bitIndex / BITS_PER_BYTE;
    uchar8 mask = 1 << (bitIndex % BITS_PER_BYTE);

    if (bitIndex == FS_LOG_INDEX_NOT_IN_BITMAP)
    {
        printf ("fs_logReplayBitmapIsSet: Out of our bitmap, id %d \n", id);
        // FS_ASSERT(FALSE);
        //bitmaphashcommit(...);
        return (FALSE);
    }
    else
    {
        return (bitmap->bits[byteIndex] & mask);
    }
}


/**
 * Update the replay bitmap boundaries to ensure a
 * bit is reserved for the specified transaction id.
 */
PRIVATE void
fs_logReplayBitmapReserve(
    fs_LogReplayBitmap  *bitmap,
    fs_LogId            id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(bitmap, id);
}


/**
 * Update the replay bitmap to reflect commit of
 * the specified transaction id.
 */
PRIVATE void
fs_logReplayBitmapCommit(
    fs_LogReplayBitmap  *bitmap,
    fs_LogId            id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(bitmap, id);
    uint32 byteIndex = bitIndex / BITS_PER_BYTE;
    uchar8 mask = 1 << (bitIndex % BITS_PER_BYTE);

    if (bitIndex == FS_LOG_INDEX_NOT_IN_BITMAP)
    {
        printf ("fs_logReplayBitmapCommit: Out of our bitmap, id %d \n", id);
        // FS_ASSERT(FALSE);
        //bitmaphashcommit(...);
    }
    else
    {
        bitmap->bits[byteIndex] |= mask;
    }

    //printf("commit id %d\n", id);
}


/**
 * Update the replay bitmap to reflect done (ie. don't replay)
 * of the specified transaction id.
 */
PRIVATE void
fs_logReplayBitmapDone(
    fs_LogReplayBitmap  *bitmap,
    fs_LogId            id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(bitmap, id);
    uint32 byteIndex = bitIndex / BITS_PER_BYTE;
    uchar8 mask = 1 << (bitIndex % BITS_PER_BYTE);

    if (bitIndex == FS_LOG_INDEX_NOT_IN_BITMAP)
    {
        printf ("fs_logReplayBitmapDone: Out of our bitmap, id %d \n", id);
        // FS_ASSERT(FALSE);
        //bitmaphashdone(...);
    }
    else
    {
        bitmap->bits[byteIndex] &= ~mask;
    }

    //printf("ignore id %d\n", id);
}






main(int argc, char** argv)
{
    fs_LogId            id;
    int                 scan, index;
    FILE                *infile;
    uint32              recid;
    char                type;
    uint32              cntCommit = 0;
    uint32              cntDone = 0;
    uint32              cntRecord = 0;
#define MAX 150
    char                line [MAX] ;
        
    fs_logInitReplayBitmap(&bitmap);

    infile = fopen("logmap", "r");
    if (infile == NULL)
    {
        perror("infile");
        exit(2);
    }

    while (fgets (line, MAX, infile) != NULL)
    {
        recid = 0 ;
        
        scan = sscanf (line, "%d %c", &recid, &type);
       
        if ((scan < 2) || (recid == 0))
        {
            printf("Invalid input: %s\n", line);
            continue;
        }

        if (type == 'c')
        {
            fs_logReplayBitmapCommit(&bitmap, recid);
            cntDone++;
        }
        else
        if (type == 'd')
        {
            fs_logReplayBitmapDone(&bitmap, recid);
            cntCommit++;
        }
        else
        {
            fs_logReplayBitmapReserve(&bitmap, recid);
            cntRecord++;
        }

       printf ("Id %d index %d\n", recid, fs_logReplayBitmapGetIndex(&bitmap, recid));

    }

    fseek(infile, 0, SEEK_SET);
    
    
    printf ("Scan done: Commits %d Done %d Other records %d\n", 
            cntCommit, cntDone, cntRecord);
    

    printf ("Bitmap: numBits %d firstId %d maxForwIndex %d maxRevIndex %d\n",
            bitmap.numBits, bitmap.firstId, bitmap.maxForwIndex, bitmap.maxRevIndex);
    

/*
    for (index = 0; index < bitmap.numBits; index ++)
    {
        if (fs_logReplayBitmapIsSet (&bitmap, index))
        {
            if (index < bitmap.maxForwIndex)
            {
                printf("LogId %d (index %d) is set\n", index + bitmap.firstId, index);
            }
            else
            if (index > bitmap.maxRevIndex)
            {
                int32 id = bitmap.firstId - index;
                if (id < 0)
                {
                    id = FS_LOG_ID_MAX - id ;
                }

                printf("LogId %d (index %d) is set\n", id, index);
            }
            else
            {
                printf("Error: index %d set in unused zone \n", index);
            }
        }
    }
*/
  
    while (fgets (line, MAX, infile) != NULL)
    {
        recid = 0 ;
        
        scan = sscanf (line, "%d %c", &recid, &type);
       
        if ((scan < 2) || (recid == 0))
        {
            printf("Invalid input: %s\n", line);
            continue;
        }
        
        printf ("Id %d index %d %s\n", 
                recid, fs_logReplayBitmapGetIndex(&bitmap, recid),
                fs_logReplayBitmapIsSet (&bitmap, recid) ? "SET" : "NOT set" );
        
    }

    fs_logReleaseReplayBitmap (&bitmap);

}





