// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/**
 * Test code for file system layout.
 *
 * Paul Haahr <haahr@agilestorage.com>
 */

#include "fs-includes.h"
#include <stdio.h>



//   S I M U L A T E D   A G I L E   E N V I R O N M E N T


/**
 * Globals we wish we didn't refer to.
 */
int32 fs_applicationId = 0;
uint32 nfxMySlotId = 0;


/**
 * Abort the system, after printing a backtrace.
 */
noreturn fs_abort()
{
    btrace_print();
    abort();
}


/**
 * Print a panic message and then abort the system.
 */
noreturn fs_panic(char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    vprintf(format, ap);
    va_end(ap);

    fs_abort();
}



//  T E S T S


#define K       1024ull
#define M       (K * K)
#define G       (K * M)
#define T       (K * G)
#define P       (K * T)

#define DEFAULT_BLOCK_SIZE              (8 * K)
#define DEFAULT_BLOCK_COUNT             (T / DEFAULT_BLOCK_SIZE)
#define DEFAULT_BLOCK_INODE_RATIO       20


/**
 * Parse a 64-bit integer, allowing K, M, G, T, and P suffixes.
 */
uint64 uint64Parse(char *s)
{
    char *t;
    uint64 n = strtoull(s, &t, 0);
    if (t == NULL)
        fs_panic("bad number: %s\n", s);
    switch (*t) {
        case 'k': case 'K':     n *= K; t++; break;
        case 'm': case 'M':     n *= M; t++; break;
        case 'g': case 'G':     n *= G; t++; break;
        case 't': case 'T':     n *= T; t++; break;
        case 'p': case 'P':     n *= P; t++; break;
    }
    if (*t != '\0')
        fs_panic("bad number: %s\n", s);
    return n;
}


/**
 * Public entry point for testing.
 */
int main(int argc, char **argv)
{
    fs_Parameters params;

    fs_SuperBlock super;
    
    if (argc > 4) {
        fprintf(stderr,
                "usage: test-layout"
                " [blockCount [blockSize [blockInodeRatio]]]\n");
        return 1;
    }

    if (argc > 3)
        super.blockInodeRatio = uint64Parse(argv[3]);
    else
        super.blockInodeRatio = DEFAULT_BLOCK_INODE_RATIO;

    if (argc > 2)
        super.blkSize = uint64Parse(argv[2]);
    else
        super.blkSize = DEFAULT_BLOCK_SIZE;

    if (argc > 1)
        super.numBlks = uint64Parse(argv[1]);
    else
        super.numBlks = DEFAULT_BLOCK_COUNT;
    
    params = fs_computeSizeParameters(&super);
    printParameters(&params);
    return 0;
}
