/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf-file.c,v 1.2 2002/12/10 22:16:22 bobonfeline Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf-file.c
 *
 * Description  : File System State-Full (FSSF)
 *                Operations on the fileTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "fssf.h"
#include "fssf-hash.h"

/*------------------------------------------------------------------------------
 * Constants.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

/**/
static fssf_status_t
fssf_delFile(fssf_share_t **    file);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_delFile
 *
 * Description  : Remove a file from fileTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fssf_status_t
fssf_delFile(fssf_share_t **    file)   // input/ouput

{
    fssf_status_t   status = NFX_OK;

    FSSF_FCT_PRINT(("FSSF: fssf_delFile(0x%llx/0x%llx/%d/%d/0x%llx)",
                    (*file)->filehandle.volId.whole_thing,
                    (*file)->filehandle.inum,
                    (*file)->filehandle.gen,
                    (*file)->filehandle.mkfsTime,
                    (*file)->filehandle.mntPnt));

    if (fssf_chash_remove(fileTab, (void *) &((*file)->filehandle)) < 0)
    {
        FSSF_ERR_PRINT(("FSSF ERROR: fssf_chash_remove() failed"));
        status = NFX_ERR;
    }
    
    return(status);
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_initFileTab
 *
 * Description  : Initialize the open files table.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_initFileTab(int initSize)

{
    NFX_SPIN_INIT(&fssf_tab_lock, SPINLOCK_LEVEL_UNKNOWN);

    NFX_SPIN_LOCK(&fssf_tab_lock);
    fileTab = fssf_chash_create(FSSF_HASH_BUCKETS_OPEN_FILE,
                                fssf_file_hashFunc,
                                fssf_file_hashCompare,
                                fssf_file_hashDestroy);
    NFX_SPIN_UNLOCK(&fssf_tab_lock);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_file_hashFunc
 *
 * Description  : Hash function for the open files table.
 *                Use the inode number which is unique in a file system.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

uint32
fssf_file_hashFunc(const void * key)

{
    return(((fs_Handle *) key)->inum % FSSF_HASH_BUCKETS_OPEN_FILE);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_file_hashCompare
 *
 * Description  : Compare function for the open files table.
 *                The 2 file handles should be identical.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32
fssf_file_hashCompare(const void *  key1,
                      const void *  key2)
    
{
    if (memcmp(key1, key2, sizeof(fs_Handle)) == 0)
    {
        return(0);  // it matches
    }
    else
    {
        return(1);
    }
}

/*------------------------------------------------------------------------------
 * Name         : fssf_file_hashDestroy
 *
 * Description  : Destroy function for the open files table.
 *                Free the "open file" element.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 12-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_file_hashDestroy(void *    key,
                      void *    obj)
    
{
    fssf_ramDealloc(key);
    fssf_ramDealloc(obj);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_getFile
 *
 * Description  : Return the open file matching the given handle.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_getFile(fs_Handle *        fileHandle, // input
             fssf_share_t **    file)       // output
    
{
    FSSF_FCT_PRINT(("FSSF: fssf_getFile(0x%llx/0x%llx/%d/%d/0x%llx)",
                    fileHandle->volId.whole_thing,
                    fileHandle->inum,
                    fileHandle->gen,
                    fileHandle->mkfsTime,
                    fileHandle->mntPnt));

    *file = (fssf_share_t *) fssf_chash_find(fileTab, (void *) fileHandle);
#ifndef ISABELLE // debug only!!
    if (*file == NULL)
    {
        printf("file not in fileTab!\n");
    }
    else
    {
        printf("file found!\n");
    }
#endif // ISABELLE
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_addNewFile
 *
 * Description  : Get a entry in the table for this new open file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_addNewFile(fs_Handle *     fileHandle, // input
                fssf_share_t ** newFile)    // output
    
{
    fssf_status_t   status = NFX_OK;
    fs_Handle *     newFilehandle = NULL;
    
    FSSF_FCT_PRINT(("FSSF: fssf_addNewFile(0x%llx/0x%llx/%d/%d/0x%llx)",
                    fileHandle->volId.whole_thing,
                    fileHandle->inum,
                    fileHandle->gen,
                    fileHandle->mkfsTime,
                    fileHandle->mntPnt));

    newFilehandle = (fs_Handle *) fssf_ramAllocZero(sizeof(fs_Handle));
    *newFilehandle = *fileHandle;
    
    *newFile = (fssf_share_t *) fssf_ramAllocZero(sizeof(fssf_share_t));
    (*newFile)->filehandle = *fileHandle;
    DLL_INIT(&((*newFile)->stateList));
    DLL_INIT(&((*newFile)->lockList));
    
    if (fssf_chash_insert(fileTab, (void *) newFilehandle, (void *) *newFile)
        < 0)
    {
        fssf_ramDealloc(*newFile);
        status = NFX_ERR;
    }

    return(status);
}

/*------------------------------------------------------------------------------
 * Name         : fssf_addStateInFile
 *
 * Description  : Add a new state to the open file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_addStateInFile(fssf_state_t *  state,  // input
                    fssf_share_t *  file)   // input / output
    
{
    NFX_ASSERT((state != NULL) && (file != NULL));
    DLL_INSERT_TAIL(&(file->stateList), state, statePerFile);
    state->fileShare = file;
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_delStateFromFile
 *
 * Description  : Remove a state for a given file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_delStateFromFile(fssf_state_t *    state)  // input / output
    
{
    fssf_status_t   status = NFX_OK;
    
    NFX_ASSERT((state != NULL) && (state->fileShare != NULL));

#ifdef ISABELLE // test only
    {
        fssf_share_t * file = NULL;
        
        fssf_getFile(&(state->fileShare->filehandle), &file);
        if (file == NULL)
        {
            printf("file(0x%llx/0x%llx) not in fileTab.\n",
                   state->fileShare->filehandle.volId.whole_thing,
                   state->fileShare->filehandle.inum);
        }
        else
        {
            printf("file(0x%llx/0x%llx) found!\n",
                   state->fileShare->filehandle.volId.whole_thing,
                   state->fileShare->filehandle.inum);
        }    
    }
#endif // ISABELLE
    
    DLL_REMOVE(&(state->fileShare->stateList), state, statePerFile);
    if (DLL_IS_EMPTY(&(state->fileShare->stateList)))
    {
        FSSF_DBG_PRINT(("FSSF: delete file(0x%llx/0x%llx/%d/%d/0x%llx)",
                        state->fileShare->filehandle.volId.whole_thing,
                        state->fileShare->filehandle.inum,
                        state->fileShare->filehandle.gen,
                        state->fileShare->filehandle.mkfsTime,
                        state->fileShare->filehandle.mntPnt));
        status = fssf_delFile(&(state->fileShare));
    }
    state->fileShare = NULL;
    
    return(status);
}

/*----------------------------------------------------------------------------*/
