/* -*- tab-width: 4 -*-
 *------------------------------------------------------------------------
 * File:        fssf_chash.c
 *
 * RCS: $Id: fssf-hash.c,v 1.5 2002/11/20 01:07:06 rick Exp $
 * 
 * Purpose:
 *      provide some generic libraries for hash tables
 *      
 * some remarks
 * - the MAX_NBR_OF_HASH_TABLES is fixed
 * - Each has a predefined struct 'tables' (about 8 words)
 *   Each table has the ref to the hash calculation of the 
 *   key, the key compare and the destroy/free key and obj 
 *   functions. Also several statistics.
 *   Whatever the hash functions returns (type uint32),
 *   this is still protected with % array_elements.
 * - at creation time, the number of 'array_elements' 
 *   of size NODE_HDR_T (4 words) is dynamically allocated
 * - for each entry a small block NODE_ELEM_T is needed.
 *   this contains the ptr to key and to the data (obj)
 * - to avoid alloc of sevral small blocks (fragmentation
 *   or waste of memory), a big chunk (CHUNKS_OF_NODES) 
 *   is allocated at one moment. Extensions are done
 *   automatically. No cleanup is foreseen.
 *   (the free chunk of nodes is common for all tables).
 * - find and deletion is done solely based on the key 
 *   value.
 * Possible extensions
 * - dynamically changes the size of array_elements
 * - cleanup the free chunk of node list.
 * - make the nbr of hash tables dynamic.
 *   Easy if a ptr is returned instead of the index.
 * - add an id, name to the table. More readable.
 * - define a max nbr of elements per table. Needed ?
 *
 * File History:
 *
 *------------------------------------------------------------------------
 */
#if !defined(SSC)

//#include "fs-includes.h"
#include "fssf-hash.h"

//#define Panic fs_panic
#define Panic   NFX_ASSERT(TRUE)

#define MAX_NBR_OF_HASH_TABLES   2
#define CHUNKS_OF_NODES          1024

typedef struct NODE_ELEM_T NODE_ELEM_T;
struct NODE_ELEM_T
{
    DLL_ENTRY(NODE_ELEM_T) dll;
    void *key;
    void *obj;
};

typedef struct NODE_HDR_T
{
    DLL_HDR(NODE_ELEM_T) dll;
    uint32 curr_size;
    uint32 max_size;
} NODE_HDR_T ;

static struct
{
    uint32 valid;
    uint32 array_elements;
    NODE_HDR_T *array;
    uint32 (*hash_func) (const void *key);
    int32 (*keycmp_func) (const void *key1, const void *key2);
    void (*destruct_func) (void *key, void *obj);
    struct 
    {
        int32 index;
        int32 reach;
    } iterate;
    struct
    {
        uint32 inserts;
        uint32 removes;
        uint32 nexts;
    } stats;
    uint32 total_elements;
} tables[MAX_NBR_OF_HASH_TABLES];

static NODE_HDR_T node_list = { { NULL, NULL }, 0, 0} ;

static nfx_spinlock_t fssf_chash_spinlock;
static int fssf_chash_spinlock_init = 0 ;


/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_node_alloc 
 * 
 * Description:
 *      support function to create additional space for nodes
 *      return curr size of free list or -1 in case of failure 
 * 
 * Function History:
 *      
 * Hash tables must be locked by caller
 *
 *------------------------------------------------------------------------
 */
static int32
fssf_chash_node_alloc ()
{
  NODE_ELEM_T *node_elements;
  uint32 index ;
  
  if (NULL == (node_elements = (NODE_ELEM_T*) eee_ramAlloc(CHUNKS_OF_NODES * sizeof (NODE_ELEM_T))))
    {
      return -1 ;
    }
  memset (node_elements, 0, CHUNKS_OF_NODES * sizeof (NODE_ELEM_T));

  if (NULL == node_list.dll.head)
    DLL_INIT(&(node_list.dll)) ;

  /* initialize the new array */
  for (index = 0; index < CHUNKS_OF_NODES; index++)
    {
      DLL_INSERT_HEAD(&(node_list.dll),&(node_elements[index]),dll);
      
      /* new_node_list->key = NULL ;  see memset */
      /* new_node_list->obj = NULL ;  see memset */
    }

  node_list.curr_size += CHUNKS_OF_NODES ;
  node_list.max_size += CHUNKS_OF_NODES ;

  return node_list.curr_size ;
}


/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_create
 * 
 * Description:
 *      Create a chained hash. passed hash, compare and destroy functions.
 *      returns hid >=0 if okay, -1 in case of failure
 * 
 * Function History:
 *      
 * Lock the table
 *
 *------------------------------------------------------------------------
 */
extern uint32 
fssf_chash_create (uint32 array_elements, 
              uint32 (*hash_func) (const void *key),
              int32 (*keycmp_func) (const void *key1, const void *key2),
              void (*destruct_func) (void *key, void *obj))
{
  uint32 hid, index;
  NODE_HDR_T *array;

  if ( ! fssf_chash_spinlock_init)
  {
    fssf_chash_spinlock_init = 1;
    NFX_SPIN_INIT(&fssf_chash_spinlock,SPINLOCK_LEVEL_UNKNOWN);
  }

  if ((array_elements <= 0) || (NULL == hash_func) || (NULL == keycmp_func))
    return -1;

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  for (hid = 0; hid < MAX_NBR_OF_HASH_TABLES; hid++)
    {
      if (! tables[hid].valid)
        break;
    }
  
  if (MAX_NBR_OF_HASH_TABLES == hid)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  if (NULL == (tables[hid].array = (NODE_HDR_T*) eee_ramAlloc(array_elements * sizeof (NODE_HDR_T))))
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  memset (tables[hid].array, 0, array_elements * sizeof (NODE_HDR_T));

  tables[hid].array_elements = array_elements;
  tables[hid].hash_func = hash_func;
  tables[hid].keycmp_func = keycmp_func;
  tables[hid].destruct_func = destruct_func;
  tables[hid].valid = 1;
  /* tables[hid].total_elements = 0 ;   see memset */

  array= tables[hid].array;
  for (index = 0; index < tables[hid].array_elements; ++index)
    {
      DLL_INIT(&(array[index].dll)) ;
    }
  
  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return hid;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_destroy
 * 
 * Description:
 *      Destroy a chained hash.
 * 
 * Function History:
 *      
 * Lock the table
 *
 *------------------------------------------------------------------------
 */
extern void
fssf_chash_destroy (uint32 hid)
{
  int32 index;
  NODE_HDR_T *array;
  NODE_ELEM_T *node;

  if (MAX_NBR_OF_HASH_TABLES <= hid)
    Panic ("fssf_chash_destroy: invalid hid (%d)\n", hid);

  if (! tables[hid].valid)
    Panic ("fssf_chash_destroy: invalid hash table (hid == %d)\n", hid);

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  if (tables[hid].total_elements)
    {
      array = tables[hid].array ;
      for (index = 0; index < tables[hid].array_elements; index++)
        {
          while ( ! DLL_IS_EMPTY(&(array[index].dll)))
            {
              DLL_REMOVE_HEAD(&(array[index].dll),node,dll);

              if (NULL != tables[hid].destruct_func)
                {
                  tables[hid].destruct_func (node->key, node->obj);
                }
              node->key = NULL ;
              node->obj = NULL ;
              tables[hid].stats.removes++;
              tables[hid].total_elements--;
              array[index].curr_size--;

              DLL_INSERT_TAIL(&(node_list.dll), node, dll);
            }
          if (0 != array[index].curr_size)
            Panic ("fssf_chash_destroy: Hash table %d.%d entry not consistent.\n", hid, index);
        }
    }

  if (0 != tables[hid].total_elements)
    Panic ("fssf_chash_destroy: Hash table %d not consistent.\n", hid);

  eee_ramDealloc (tables[hid].array) ;
  /* Panic ("fssf_chash_destroy: MemFree failed.\n");    No check possible */

  memset (& (tables[hid]), 0, sizeof (tables[hid]));

  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
}

/*
 *------------------------------------------------------------------------
 * Function:    find_core
 * 
 * Description:
 *      Core finder routine.
 * 
 * Function History:
 *      
 * Hash tables must be locked by caller
 *
 *------------------------------------------------------------------------
 */
static NODE_HDR_T *
find_core (uint32 hid, void *skey, NODE_ELEM_T **result_node)
{ 
  uint32 index;
  NODE_HDR_T *array;
  NODE_ELEM_T *node;

  if (MAX_NBR_OF_HASH_TABLES <= hid)
    return NULL;
  
  if (! tables[hid].valid)
    return NULL;

  index = (tables[hid].hash_func (skey)) % (tables[hid].array_elements);

  array = tables[hid].array ;

  if (0 == array[index].curr_size)
    return NULL;

  DLL_ITERATE(node, &array[index].dll, dll) 
  {
#ifdef ISABELLE // debug only!!!
          printf("hid = %d\n", hid);
          printf("index = %d\n", index);
          if (hid == 0)
          {
              printf("skey = 0x%llx/0x%llx\n",
                     ((fs_Handle *) skey)->volId.whole_thing,
                     ((fs_Handle *) skey)->inum);
              printf("node->key = 0x%llx/0x%llx\n",
                     ((fs_Handle *) (node->key))->volId.whole_thing,
                     ((fs_Handle *) (node->key))->inum);
          }
          else
          {
              printf("skey = 0x%llx\n", *((fssf_clientid_t *) skey));
              printf("node->key = 0x%llx\n", *((fssf_clientid_t *) (node->key)));
          }
#endif // ISABELLE
          if (! tables[hid].keycmp_func (skey, node->key))
          {
              *result_node = node;
#ifdef ISABELLE // debug only!!!
              printf("found!\n");
#endif // ISABELLE
              return &array[index];
          }
  }

  *result_node = NULL;
  return NULL;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_insert
 * 
 * Description:
 *      Insert a node into the chained hash.
 *      return 0 if OK, -1 in case of failure
 * 
 * Function History:
 *      
 * Lock the table
 *
 * To Enable: - Ckeck if key is unique !!!!
 *
 *------------------------------------------------------------------------
 */
extern int32
fssf_chash_insert (uint32 hid, void *key, void *obj)
{
  uint32 index;
  NODE_HDR_T *array;
  NODE_ELEM_T *node;

  if (MAX_NBR_OF_HASH_TABLES <= hid)
    return -1;

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  if (! tables[hid].valid)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  array = find_core (hid, key, & node);
  if (array != NULL)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  index = (tables[hid].hash_func (key)) % (tables[hid].array_elements);
  
  array = tables[hid].array ;

  if (0 == node_list.curr_size)
    if (fssf_chash_node_alloc () < 0)
      {
        NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
        return -1 ;
      }

  DLL_REMOVE_HEAD(&(node_list.dll), node, dll);
  node_list.curr_size--;
  DLL_INSERT_HEAD(&(array[index].dll), node, dll);
  array[index].curr_size++;
  if (array[index].curr_size > array[index].max_size)
    array[index].max_size = array[index].curr_size;

  node->obj = obj ;
  node->key = key ;
#ifdef ISABELLE // debug only!!!
  printf("node->obj = %x\n", (uint32) (address_t) node->obj);
  printf("node->key = %x\n", (uint32) (address_t) node->key);
#endif // ISABELLE
  
  tables[hid].stats.inserts++;
  tables[hid].total_elements++;

  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return 0;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_remove
 * 
 * Description:
 *      Remove a node from the chained hash.
 *      return 0 if OK, -1 in case of failure
 * 
 * Function History:
 *      
 * Lock the table
 *
 *------------------------------------------------------------------------
 */
extern int32
fssf_chash_remove (uint32 hid, void *skey)
{
  NODE_HDR_T *array;
  NODE_ELEM_T *node;

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  array = find_core (hid, skey, & node);

  if (NULL == array)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  DLL_REMOVE(&(array->dll), node, dll);
  array->curr_size-- ;
  DLL_INSERT_HEAD(&(node_list.dll), node, dll);
  node_list.curr_size++;

  if (NULL != tables[hid].destruct_func)
    tables[hid].destruct_func (node->key, node->obj);

  tables[hid].stats.removes++;
  tables[hid].total_elements--;

  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return 0;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_find
 * 
 * Description:
 *      Find a node in the chained hash.
 *      return the obj, or NULL if not found
 * 
 * Function History:
 *      
 * Lock the table
 * It is up to the caller the lock the object !!!!
 *
 *------------------------------------------------------------------------
 */
void*
fssf_chash_find (uint32 hid, void *key)
{
  NODE_HDR_T *array;
  NODE_ELEM_T *node;

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  array = find_core (hid, key, & node);

#ifdef ISABELLE // debug only!!!
  printf("array = %x\n", (uint32) (address_t) array);
  printf("node = %x\n", (uint32) (address_t) node);
  if (node != NULL)
  {
      printf("node->obj = %x\n", (uint32) (address_t) (node->obj));
  }
#endif // ISABELLE
  
  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  if (NULL == array)
    return NULL;

  return node->obj;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_reset
 * 
 * Description:
 *      Reset the iterator for this chained hash.
 * 
 * Function History:
 *      
 * Lock the table
 *
 *------------------------------------------------------------------------
 */
int32
fssf_chash_reset (uint32 hid)
{
  if (MAX_NBR_OF_HASH_TABLES <= hid)
    return -1;

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  if (! tables[hid].valid)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  memset (& (tables[hid].iterate), 0, sizeof (tables[hid].iterate));

  // tables[hid].stats.nexts = 0;
  
  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return 0;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_next
 * 
 * Description:
 *      Get the next value, move the iterator
 * 
 * Function History:
 *
 * Lock the table    
 *
 *------------------------------------------------------------------------
 */
int32
fssf_chash_next (uint32 hid, void **keyp, void **objp)
{
  NODE_HDR_T *array;
  NODE_ELEM_T *node;
  int32 index;

  if (MAX_NBR_OF_HASH_TABLES <= hid)
    return -1;

  NFX_SPIN_LOCK(&fssf_chash_spinlock);

  if (! tables[hid].valid)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  /* if a search is ended, wait for the fssf_chash_reset */
  if (tables[hid].iterate.index < 0)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  if (tables[hid].total_elements == 0)
    {
      tables[hid].iterate.index = -1;
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;    
    }

  array = tables[hid].array ;

next_hash_next:
  /* find first non empty array entry */
  while (DLL_IS_EMPTY(&(array[tables[hid].iterate.index].dll)))
    {
      tables[hid].iterate.index++ ;
      tables[hid].iterate.reach = 0 ;
      if (tables[hid].iterate.index >= tables[hid].array_elements)
        {
          tables[hid].iterate.index = -1;
          NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
          return -1;    
        }
    }
  /* skip tables[hid].iterate.reach */
  if (tables[hid].iterate.reach >= array[tables[hid].iterate.index].curr_size)
    {
      tables[hid].iterate.index++ ;
      tables[hid].iterate.reach = 0 ;
      if (tables[hid].iterate.index >= tables[hid].array_elements)
        {
          tables[hid].iterate.index = -1;
          NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
          return -1;    
        }
      goto next_hash_next;
    }
  node = array[tables[hid].iterate.index].dll.head ;
  for (index = 0; index < tables[hid].iterate.reach ; index++)
    {
      node = node->dll.next;
      if (node == NULL)
        {
          Panic ("fssf_chash_next: wrong nbr of nodes\n");
          NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
          return -1;
        }     
    }
  tables[hid].iterate.reach++ ;
  *keyp = node->key ;
  *objp = node->obj ;
  
  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return 1;
}


/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_stat
 * 
 * Description:
 *      View the statistics for given hash table
 * 
 * Function History:
 *      
 * Lock the table    
 *
 *------------------------------------------------------------------------
 */
int32
fssf_chash_stat (uint32 hid)
{
  int32 index;
  NODE_HDR_T *array;


  if (MAX_NBR_OF_HASH_TABLES <= hid)
    {
      printf ("fssf_chash_stat: invalid hid (%d)\n", hid);
      return -1;
    }  

  NFX_SPIN_LOCK(&fssf_chash_spinlock); 

  if (! tables[hid].valid)
    {
      printf ("fssf_chash_stat: hash table (%d) not in use !\n", hid);
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  printf ("Statistics for hash table %d \n", hid);
  printf ("  Table array size          %d\n", tables[hid].array_elements);
  printf ("  Total number of entries   %d\n", tables[hid].total_elements);
  printf ("  Total number of inserts   %d\n", tables[hid].stats.inserts);
  printf ("  Total number of removes   %d\n", tables[hid].stats.removes);
  printf ("  Total number of nexts     %d\n", tables[hid].stats.nexts);
  printf ("  Total available nodes     %d\n", node_list.curr_size);
  printf ("  Max nbr nodes available   %d\n", node_list.max_size);

  array = tables[hid].array ;
  for (index = 0; index < tables[hid].array_elements ; index++)
    {
      if (array[index].max_size)
        {
          printf ("  Entry %8d, current entries %8d, max entries %8d\n", 
                  index, 
                  array[index].curr_size,
                  array[index].max_size );
        }
    }
  printf ("\n");
  
  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return 0 ;
}

/*
 *------------------------------------------------------------------------
 * Function:    fssf_chash_stat_nbr
 * 
 * Description:
 *      returns the statistics for given hash table
 *      array_sz: the size of the array
 *      elements: the current total number of elements/nodes in the table 
 *      curr_max_size: the current max on one row
 *      max_size: the overall max once on one row
 * 
 * Function History:
 *      
 * Lock the table    
 *
 *------------------------------------------------------------------------
 */
int32
fssf_chash_stat_nbr (
    uint32 hid, 
    uint32 *array_sz, 
    uint32 *elements, 
    uint32 *curr_max_size,
    uint32 *max_size)
{
  int32 index;
  NODE_HDR_T *array;

  *array_sz = *elements = *curr_max_size = *max_size = 0 ;

  if (MAX_NBR_OF_HASH_TABLES <= hid)
      return -1;

  NFX_SPIN_LOCK(&fssf_chash_spinlock); 

  if (! tables[hid].valid)
    {
      NFX_SPIN_UNLOCK(&fssf_chash_spinlock);
      return -1;
    }

  *array_sz = tables[hid].array_elements;
  *elements = tables[hid].total_elements;

  array = tables[hid].array ;
  for (index = 0; index < tables[hid].array_elements ; index++)
    {
      *curr_max_size = MAX (array[index].curr_size, *curr_max_size);
      *max_size = MAX (array[index].max_size, *max_size);
    }
  
  NFX_SPIN_UNLOCK(&fssf_chash_spinlock);

  return 0 ;
}
#endif // SSC
