/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf-struct.h,v 1.7 2002/03/27 00:09:48 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf-struct.h
 *
 * Description  : File System State-Full (FSSF)
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#ifndef _FSSF_STRUCT_H_
#define _FSSF_STRUCT_H_

/*------------------------------------------------------------------------------
 * Header files.
 *------------------------------------------------------------------------------
 */

#include "fssf.h"
#include "exports/queue-api.h"  // DLL_xxx macros

/*------------------------------------------------------------------------------
 * Constant definitions.
 *------------------------------------------------------------------------------
 */

#define FSSF_MAX_OPEN_FILE_PER_CLIENT   20
#define FSSF_MAX_LOCK_PER_FILE          2
#define FSSF_MAX_CLIENT                 4096

#define FSSF_MAX_OPEN_FILE  (FSSF_MAX_CLIENT * FSSF_MAX_OPEN_FILE_PER_CLIENT)
#define FSSF_MAX_OWNER      (FSSF_MAX_CLIENT)
#define FSSF_MAX_OPEN_STATE (FSSF_MAX_CLIENT * FSSF_MAX_OPEN_FILE_PER_CLIENT)
#define FSSF_MAX_LOCK_STATE (FSSF_MAX_OPEN_FILE * FSSF_MAX_LOCK_PER_FILE)

#define FSSF_MIN_OPEN_FILE  (FSSF_MAX_CLIENT*FSSF_MAX_OPEN_FILE_PER_CLIENT / 10)
#define FSSF_MIN_OWNER      (FSSF_MAX_CLIENT / 10)
#define FSSF_MIN_OPEN_STATE (FSSF_MAX_CLIENT*FSSF_MAX_OPEN_FILE_PER_CLIENT / 10)
#define FSSF_MIN_LOCK_STATE (FSSF_MAX_OPEN_FILE * FSSF_MAX_LOCK_PER_FILE / 10)

/* Note that the hash buckets should be a prime number far away
 * from a power of 2.
 */
#define FSSF_HASH_BUCKETS_OPEN_FILE     3001
#define FSSF_HASH_BUCKETS_OWNER         701

/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef struct fssf_share_T         fssf_share_t;
typedef struct fssf_state_T         fssf_state_t;
typedef struct fssf_owner_T         fssf_owner_t;
typedef struct fssf_lock_T          fssf_lock_t;

/* Share reservations on a file.
 * -----------------------------
 * Note: only open files are kept track of.
 */
struct fssf_share_T
{
    fs_Handle               filehandle; // handle to the open file
    fssf_shareAccess_t      shareAccess;// access already granted on the file
    fssf_shareDeny_t        shareDeny;  // access to deny others for the file
    DLL_HDR(fssf_state_t)   stateList;  // open states associated with the file
    DLL_HDR(fssf_lock_t)    lockList;   // locks made on the file
};

/* Owner.
 * ------
 * Note: only owners with at least one open file are kept track of.
 */
struct fssf_owner_T
{
    fssf_clientid_t         owner;      // owner
    DLL_HDR(fssf_state_t)   stateList;  // open states belonging to the owner
};

/* Open state of a specific owner for a specific file.
 * ---------------------------------------------------
 */
struct fssf_state_T
{
    fssf_stateid_t          stateid;        // open state id
    byte                    seq;            // 8-bits sequence number
    fssf_share_t *          fileShare;      // pointer to the open file
    fssf_owner_t *          owner;          // pointer to the 'open' owner
    fssf_shareAccess_t      shareAccess;    // access granted for the owner
    fssf_shareDeny_t        shareDeny;  // access denied to others by the owner
    fs_Offset               currPos;    // current position in file
    DLL_HDR(fssf_lock_t)    lockList;   // locks made by the owner on the file
    DLL_ENTRY(fssf_state_t) statePerFile;   // stateList's entry in fssf_share_T
    DLL_ENTRY(fssf_state_t) statePerOwner;  // stateList's entry in fssf_owner_T
    DLL_ENTRY(fssf_state_t) freeState;      // point to the next free state
};

/* Lock.
 * -----
 */
struct fssf_lock_T
{
    fssf_stateid_t          stateid;        // lock state id
    fssf_share_t *          fileShare;      // pointer to the open file
    fssf_lockType_t         lockType;       // lock type
    fs_Offset               offset;         // locked record - offset
    uint32                  len;            // locked record - length
    fssf_owner_t *          owner;          // pointer to the 'lock' owner
    DLL_ENTRY(fssf_lock_t)  lockPerFile;    // lockList's entry in fssf_share_T
    DLL_ENTRY(fssf_lock_t)  lockPerState;   // lockList's entry in fssf_state_T
    DLL_ENTRY(fssf_lock_t)  freeLock;       // point to the next free lock
};

/* Open state(s) table.
 */
typedef struct
{
    uint32                  size;           // number of elements in the table
    fssf_state_t **         tab;            // pointer to the table
    DLL_HDR(fssf_state_t)   freeList;       // pointer to the free elements
} fssf_openTab_t;

/* Lock state(s) table.
 */
typedef struct
{
    uint32                  size;           // number of elements in the table
    fssf_lock_t **          tab;            // pointer to the table
    DLL_HDR(fssf_lock_t)    freeList;       // pointer to the free elements
} fssf_lockTab_t;

/*  Global FSFS statistics.
 */
typedef struct
{
    uint64  nNullReq;           // NULL operation
    uint64  nNullRsp;
    uint64  nCloseReq;          // CLOSE operation
    uint64  nCloseRsp;
    uint64  nCreateReq;         // CREATE operation
    uint64  nCreateRsp;
    uint64  nGetattrReq;        // GETATTR operation
    uint64  nGetattrRsp;
    uint64  nLockReq;           // LOCK operation
    uint64  nLockRsp;
    uint64  nOpenReq;           // OPEN operation
    uint64  nOpenRsp;
    uint64  nReadReq;           // READ operation
    uint64  nReadRsp;
    uint64  nRemoveReq;         // REMOVE operation
    uint64  nRemoveRsp;
    uint64  nSetattrReq;        // SETATTR operation
    uint64  nSetattrRsp;
    uint64  nUnlockReq;         // UNLOCK operation
    uint64  nUnlockRsp;
    uint64  nWriteReq;          // WRITE operation
    uint64  nWriteRsp;

    uint64  nFsCreateReq;       // FS CREATE operation
    uint64  nFsCreateRsp;
    uint64  nFsLookupReq;       // FS LOOKUP operation
    uint64  nFsLookupRsp;
    
    uint64  nAlloc;             // number of alloc
    uint64  nDealloc;           // number of dealloc
} fssf_stats_t;

/* 
 */
typedef struct
{
    esm_handle_t    threadHandle;
    uint32          sender;
    char *          msgBuf;
} fssf_cb_t;

/*----------------------------------------------------------------------------*/
#endif // _FSSF_STRUCT_H_
