/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf.c,v 1.16 2002/03/27 00:09:48 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf.c
 *
 * Description  : File System State-Full (FSSF)
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "fssf.h"

/*------------------------------------------------------------------------------
 * Constants.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_processCloseReq
 *
 * Description  : Process CLOSE request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processCloseReq(esm_handle_t   handle,
                     void *         userInfo)

{
    fssf_cb_t *         cb = (fssf_cb_t *) userInfo;
    fssf_closeReq_t *   req = (fssf_closeReq_t *) cb->msgBuf;
    fssf_closeRsp_t *   rsp = NULL;
    fssf_status_t       status = NFX_OK;
    fssf_state_t *      openState = NULL;
    
    FSSF_STAT_CLOSEREQ;

    /* Initialize the response.
     */
    rsp = (fssf_closeRsp_t *) fssf_ramAllocZero(sizeof(fssf_closeRsp_t));
    rsp->hdr.msgType = FSSF_CLOSE_RSP;
    rsp->hdr.len = sizeof(fssf_closeRsp_t);
    rsp->hdr.clientPrivate = req->hdr.clientPrivate;
    rsp->hdr.protocol = req->hdr.protocol;

    /* Get entry in open states table.
     */
    status = fssf_getState(&(req->stateid), &openState);
    if (status != NFX_OK)
    {
        goto fssf_processCloseReq_exit;
    }

    /* Check if any lock from this client are still existing.
     */
    if (DLL_IS_EMPTY(&(openState->lockList)) == FALSE)
    {
        status = NFX_ERR;
        goto fssf_processCloseReq_exit;
    }

    /* Close file.
     * Remove all entries corresponding to that file/owner from our tables.
     */
    status = fssf_delStateFromOwner(openState);
    if (status != NFX_OK)
    {
        goto fssf_processCloseReq_exit;
    }

    status = fssf_delStateFromFile(openState);
    if (status != NFX_OK)
    {
        goto fssf_processCloseReq_exit;
    }

    fssf_delOpenState(&openState);
    
  fssf_processCloseReq_exit:
    
    rsp->status = status;

    /* Respond to caller.
     */
    if (fssf_sendMsg(fssfPort, cb->sender,
                     (void *) rsp, sizeof(fssf_closeRsp_t)) == NFX_OK)
    {
        FSSF_STAT_CLOSERSP;
    }
    
    /* Free memory.
     */
    fssf_ramDealloc(cb);
#ifndef FS_STANDALONE
    fssf_ramDealloc(rsp);
#endif // FS_STANDALONE
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processCreateReq
 *
 * Description  : Process CREATE request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processCreateReq(fssf_createReq_t * req)

{
    FSSF_STAT_CREATEREQ;
    FSSF_ERR_PRINT(("Not yet implemented!"));
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processGetattrReq
 *
 * Description  : Process GETATTR request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processGetattrReq(esm_handle_t handle,
                       void *       userInfo)

{
    fssf_cb_t *         cb = (fssf_cb_t *) userInfo;
    fssf_getattrReq_t * req = (fssf_getattrReq_t *) cb->msgBuf;
    fssf_getattrRsp_t * rsp = NULL;
    fssf_status_t       status = NFX_OK;
    fssf_state_t *      openState = NULL;
    
    FSSF_STAT_GETATTRREQ;

    /* Initialize the response.
     */
    rsp = (fssf_getattrRsp_t *) fssf_ramAllocZero(sizeof(fssf_getattrRsp_t));
    rsp->hdr.msgType = FSSF_GETATTR_RSP;
    rsp->hdr.len = sizeof(fssf_getattrRsp_t);
    rsp->hdr.clientPrivate = req->hdr.clientPrivate;
    rsp->hdr.protocol = req->hdr.protocol;

    /* Get entry in open states table.
     */
    status = fssf_getState(&(req->stateid), &openState);
    if (status != NFX_OK)
    {
        goto fssf_processGetattrReq_exit;
    }

    /* Return attributes.
     */
    rsp->currPos = openState->currPos;
    rsp->shareAccess = openState->shareAccess;
    rsp->shareDeny = openState->shareDeny;
    rsp->filehandle = openState->fileShare->filehandle;
    
  fssf_processGetattrReq_exit:
    
    rsp->status = status;

    /* Respond to caller.
     */
    if (fssf_sendMsg(fssfPort, cb->sender,
                     (void *) rsp, sizeof(fssf_getattrRsp_t)) == NFX_OK)
    {
        FSSF_STAT_GETATTRRSP;
    }
    
    /* Free memory.
     */
    fssf_ramDealloc(cb);
#ifndef FS_STANDALONE
    fssf_ramDealloc(rsp);
#endif // FS_STANDALONE
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processLockReq
 *
 * Description  : Process LOCK request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processLockReq(fssf_lockReq_t * req)

{
    FSSF_STAT_LOCKREQ;
    FSSF_ERR_PRINT(("Not yet implemented!"));
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processOpenReq
 *
 * Description  : Process OPEN request.
 *
 *                Note that the OPEN operation doesn't checked the access rights
 *                of the client. It is therefore  assumed that the access rights
 *                have been checked before opening the file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processOpenReq(esm_handle_t    handle,
                    void *          userInfo)

{
    fssf_cb_t *         cb = (fssf_cb_t *) userInfo;
    fssf_openReq_t *    req = (fssf_openReq_t *) cb->msgBuf;
    fssf_openRsp_t *    rsp = NULL;
    fssf_status_t       status = NFX_OK;
    fs_Handle           fileHandle;
    //fs_Attr             fileAttr;
    fssf_owner_t *      owner = NULL;
    fssf_share_t *      openFile = NULL;
    fssf_state_t *      openState = NULL;

    FSSF_STAT_OPENREQ;

    /* Initialize the response.
     */
    rsp = (fssf_openRsp_t *) fssf_ramAllocZero(sizeof(fssf_openRsp_t));
    rsp->hdr.msgType = FSSF_OPEN_RSP;
    rsp->hdr.len = sizeof(fssf_openRsp_t);
    rsp->hdr.clientPrivate = req->hdr.clientPrivate;
    rsp->hdr.protocol = req->hdr.protocol;
    
    /* Check if owner has already a state.
     * If not add him/her in our tables.
     */
    fssf_getOwner(req->owner, &owner);
    if (owner == NULL)
    {
        status = fssf_addNewOwner(req->owner, &owner);
        if (owner == NULL)
        {
            goto fssf_processOpenReq_exit;
        }
    }

    /* If the file has to be created first, do it.
     * Otherwise, check if the file exists.
     */
    if (req->createFlag == TRUE)
    {
        fs_Handle   dirHandle;
        fs_Attr     dirAttr;    // not used
        
        /* Check if the directory exists.
         */
        status = fssf_mcl(cb, &dirHandle, &dirAttr);
        if (status != NFX_OK)
        {
            goto fssf_processOpenReq_exit;
        }

        /* Create the file.
         */
        status = fssf_createFile(cb, &dirHandle, &fileHandle, &(rsp->attr));
        if (status != NFX_OK)
        {
            goto fssf_processOpenReq_exit;
        }
    }
    else
    {
        /* Check if the file exists (LOOKUP).
         */
        status = fssf_mcl(cb, &fileHandle, &(rsp->attr));
        if (status != NFX_OK)
        {
            goto fssf_processOpenReq_exit;
        }
    }
    
    /* Open the file.
     */
    
    /* Get entry in open files table.
     */
    fssf_getFile(&fileHandle, &openFile);
    if (openFile == NULL)
    {
        /* Nobody is currently accessing that file.
         * Add new entries for that file in our tables.
         */
        status = fssf_addNewFile(&fileHandle, &openFile);
        if (status != NFX_OK)
        {
            goto fssf_processOpenReq_exit;
        }
        //openFile->filehandle = fileHandle;
        openFile->shareAccess = req->shareAccess;
        openFile->shareDeny = req->shareDeny;
        //DLL_INIT(openFile->stateList);
        //DLL_INIT(openFile->lockList);
        
        fssf_addOpenState(&openState);
        if (openState == NULL)
        {
            status = NFX_ERR;
            goto fssf_processOpenReq_exit;
        }
        //openState->fileShare = openFile;
        //openState->owner = owner;
        openState->shareAccess = req->shareAccess;
        openState->shareDeny = req->shareDeny;
        openState->currPos = 0;
        DLL_INIT(&(openState->lockList));
        
        fssf_addStateInFile(openState, openFile);
        fssf_addStateInOwner(openState, owner);
    }
    else 
    {
        /* The file is already open by some clients.
         * Check if new client share reservations do not conflict with
         * current file share reservations.
         */
        if((req->shareAccess & openFile->shareDeny) != 0)
        {
            status = FSSF_ERR_SHARE_DENIED;
            goto fssf_processOpenReq_exit;
        }
        
        /* Nobody is currently accessing that file.
         * Add/update entries for that file in our tables.
         */
        //fssf_getNewOpenState(&openState);
        fssf_addOpenState(&openState);
        if (openState == NULL)
        {
            status = NFX_ERR;
            goto fssf_processOpenReq_exit;
        }
        openState->fileShare = openFile;
        openState->owner = owner;
        openState->shareAccess = req->shareAccess;
        openState->shareDeny = req->shareDeny;
        openState->currPos = 0;
        DLL_INIT(&(openState->lockList));
        
        openFile->shareAccess = openFile->shareAccess | req->shareAccess;
        openFile->shareDeny = openFile->shareDeny | req->shareDeny;
        
        fssf_addStateInFile(openState, openFile);
        fssf_addStateInOwner(openState, owner);
    }
    
  fssf_processOpenReq_exit:
    
    rsp->status = status;
    if (rsp->status == NFX_OK)
    {
        rsp->stateid = openState->stateid;
    }
    else
    {
        // TODO!!! undo what we did :(
    }

    /* Respond to caller.
     */
    if (fssf_sendMsg(fssfPort, cb->sender,
                     (void *) rsp, sizeof(fssf_openRsp_t)) == NFX_OK)
    {
        FSSF_STAT_OPENRSP;
    }
    
    /* Free memory.
     */
    fssf_ramDealloc(cb);
#ifndef FS_STANDALONE
    fssf_ramDealloc(rsp);
#endif // FS_STANDALONE
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processReadReq
 *
 * Description  : Process READ request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processReadReq(esm_handle_t    handle,
                    void *          userInfo)

{
#ifdef ISABELLE // under construction
    fssf_cb_t *         cb = (fssf_cb_t *) userInfo;
    fssf_readReq_t *    req = (fssf_readReq_t *) cb->msgBuf;
    fssf_readRsp_t      rsp;
    fssf_status_t       status = NFX_OK;
    fssf_state_t *      openState = NULL;
    
    FSSF_STAT_READREQ;

    /* Initialize the response.
     */
    bzero((void *) &rsp, sizeof(fssf_readRsp_t));
    rsp.hdr.msgType = FSSF_READ_RSP;
    rsp.hdr.len = sizeof(fssf_readRsp_t);
    rsp.hdr.clientPrivate = req->hdr.clientPrivate;
    rsp.hdr.protocol = req->hdr.protocol;
    
    /* Get entry in open states table.
     */
    status = fssf_getState(&(req->stateid), &openState);
    if (status != NFX_OK)
    {
        goto fssf_processCloseReq_exit;
    }

    // TODO!!!
    
  fssf_processReadReq_exit:
    
    rsp.status = status;
    // TODO!!!
    
    /* Respond to caller.
     */
    if (eee_sendMessage((void *) &rsp, sizeof(fssf_readRsp_t),
                        cb->sender, fssfPort) == NFX_OK)
    {
        FSSF_STAT_READRSP;
    }
    
    /* Free control block.
     */
    fssf_ramDealloc(cb);
    
#endif // ISABELLE
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processRemoveReq
 *
 * Description  : Process REMOVE request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processRemoveReq(fssf_removeReq_t * req)

{
    FSSF_STAT_REMOVEREQ;
    FSSF_ERR_PRINT(("Not yet implemented!"));
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processSetattrReq
 *
 * Description  : Process SETATTR request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processSetattrReq(fssf_setattrReq_t * req)

{
    FSSF_STAT_SETATTRREQ;
    FSSF_ERR_PRINT(("Not yet implemented!"));
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processUnlockReq
 *
 * Description  : Process UNLOCK request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processUnlockReq(fssf_unlockReq_t * req)

{
    FSSF_STAT_UNLOCKREQ;
    FSSF_ERR_PRINT(("Not yet implemented!"));
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processWriteReq
 *
 * Description  : Process WRITE request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processWriteReq(fssf_writeReq_t * req)

{
    FSSF_STAT_WRITEREQ;
    FSSF_ERR_PRINT(("Not yet implemented!"));
}

/*----------------------------------------------------------------------------*/
