/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : sim-client.c
 *
 * Description  : Simulates FSSF client.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include <stdlib.h>         // exit
#include <stdio.h>          // fprintf
#include <netdb.h>          // gethostbyname
#include <string.h>         // rindex
#include <sys/types.h>
#include <unistd.h>

#ifndef FS_STANDALONE
#include "nfx-incl.h"
#include "exports/anpssc-api.h"
#include "exports/evm-api.h"
#else
#include "../sm-fs/fs-includes.h"
#endif

#include "exports/fssf-api.h"

/*------------------------------------------------------------------------------
 * Constants.
 *------------------------------------------------------------------------------
 */

/* Wait for FS/FSSF to respond (5s).
 */
static struct timeval waitTo = {5, 0};

/*------------------------------------------------------------------------------
 * Data structures definition
 *------------------------------------------------------------------------------
 */

/* Basic message structure. */
typedef struct
{
    char    buf[4096];
    uint32  len;
} sim_msg_t;

/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */

/* The name of this program. */
static char * myName;

/*------------------------------------------------------------------------------
 * Prototype of local functions
 *------------------------------------------------------------------------------
 */

static void
print_usage(void);
static int
bad_stat(int stat);
static int
bad_resp (int stat, char *resp);

static evm_volId_t          // output
sim_getVolId(char * name);  // input

static fs_Status                        // output
sim_getRoot(evm_volId_t     volId,      // input
            char *          path,       // intput
            fs_Handle *     dirHandle); // output

static fs_Status                    // output
sim_lookup(fs_Handle *  dirHandle,  // input
           char *       obj,        // input
           fs_Handle *  objHandle); // output

static fssf_status_t
sim_open(fs_Handle *        parentHandle,   // input
         char *             fileName,       // input
         fssf_stateid_t *   stateid);       // output

static fssf_status_t                    // output
sim_close(fssf_stateid_t    stateid);   // input

/*------------------------------------------------------------------------------
 * Name         : print_usage()
 *
 * Description  : Shows the correct usage of this program.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
print_usage(void)

{
	fprintf(stderr,"Usage:\n");

	fprintf(stderr," %s lookup <volume> <filename>\n", myName);
	fprintf(stderr," %s open <volume> <filename>\n", myName);
	fprintf(stderr," %s close <stateid>\n", myName);
	fprintf(stderr,"\n");
}

/*-----------------------------------------------------------------
 * Name: bad_stat
 *
 * Description: General check of sendAgileMsg() returned status
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 June 2001
 *
 *-----------------------------------------------------------------
 */
static int
bad_stat (int stat)
{
	if (stat < 0) {
		perror ("Send to evmcfgd failed");
		return 1;
	}
	else if (!stat) {
		fprintf (stderr, "Timed out waiting for response\n");
		return 1;
	}
	return 0;
}

/*-----------------------------------------------------------------
 * Name: bad_resp
 *
 * Description: General check of response from evm_cfgd
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 June 2001
 *
 *-----------------------------------------------------------------
 */
static int
bad_resp (int stat, char *resp)
{
	evm_MsgStat_t *stat_msg = (evm_MsgStat_t *)resp;
	char *rtnstr = &resp[sizeof(evm_MsgStat_t)];

	if (stat < sizeof(evm_MsgStat_t)) {
		fprintf (stderr, "Response message truncated\n");
		return 1;
	}
/*
	Check the returned status to make sure it's from us and the status
	is OK.
 */
	else if (stat_msg->hdr.clientPrivate != getpid()) {
		fprintf (stderr, "Response wasn't from our message\n");
		return 1;
	}
/*
	If we got an error, they might have hidden a string at the end of
	the response.
 */
	else if (stat_msg->stat != NFX_OK) {
		fprintf (stderr, "Response indicates message was rejected\n");
		if ((stat > sizeof(evm_MsgStat_t)) && *rtnstr) {
			fprintf (stderr, "%s\n", rtnstr);
		}
		return 1;
	}
	return 0;
}

/*-----------------------------------------------------------------
 * Name: get_id
 *
 * Description: Given a name, get the LV id
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 June 2001
 *
 *-----------------------------------------------------------------
 */

static evm_volId_t          // output
sim_getVolId(char * name)   // input

{
	evm_MsgGetId_t  getmsg;
    sim_msg_t       msg;
	evm_MsgStat_t * respmsg = (evm_MsgStat_t *) msg.buf;

	memset(&getmsg, 0, sizeof(getmsg));	// Squeeky clean
	getmsg.hdr.msgType = EVM_VOL_GET_ID;
	getmsg.hdr.clientPrivate = getpid();
	strcpy(getmsg.name, name);

	msg.len = sendAgileMsg(-1, &getmsg, sizeof(getmsg), 
                           "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitTo, msg.buf, sizeof(msg.buf));
	NFX_ASSERT((msg.len >= sizeof(evm_MsgStat_t)) &&
               (respmsg->stat == NFX_OK));
    
	if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
    {
		respmsg->payload.lVolId.whole_thing = -1;
	}

	return(respmsg->payload.lVolId);
}

/*------------------------------------------------------------------------------
 * Name         : sim_root()
 *
 * Description  : Get root handle for a mount point in a filesystem.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fs_Status                        // output
sim_getRoot(evm_volId_t     volId,      // input
            char *          path,       // intput
            fs_Handle *     dirHandle)  // output

{
    uint32              reqSize;
    fs_MsgGetRootReq *  req = NULL;
    sim_msg_t           msg;
    fs_MsgGetRootResp * rsp = (fs_MsgGetRootResp *) msg.buf;
    
    /* Initialize the request.
     */
    reqSize = sizeof(fs_MsgLookupReq) + strlen(path);
    req = (fs_MsgGetRootReq *) calloc(reqSize, 1);
    req->hdr.requestType = FS_REQ_GET_ROOT;
    req->hdr.len = reqSize;
	req->hdr.attrType = 0;
	req->hdr.uid = 0;   // root
	req->hdr.gid = 0;   // root
    req->hdr.proto = FS_PROTO_NFS_V3;
    req->hdr.numGids = 0;
    req->hdr.clientPrivate = 0;
    req->volId = volId;
    req->pathLen = strlen(path);
    strcpy(req->pathName, path);
    
    /* Send the request to FS.
     */
	msg.len = sendAgileMsg(-1, &req, reqSize, 
                           "fs", NFX_NFP_FP2, 1, MGMTPLANE,
                           &waitTo, msg.buf, sizeof(msg.buf));
    NFX_ASSERT((msg.len >= sizeof(fs_MsgGetRootResp)) &&
               (rsp->hdr.requestType == FS_REQ_GET_ROOT));

    /* A response has been received.
     */
    printf("FS_REQ_GET_ROOT: status = %d\n", rsp->hdr.fsStatus);
    if (rsp->hdr.fsStatus == NFX_OK)
    {
        memcpy((void *) dirHandle, (const void *) &(rsp->dirHandle),
               sizeof(fs_Handle));
    }
    
    return(rsp->hdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : sim_lookup()
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fs_Status                    // output
sim_lookup(fs_Handle *  dirHandle,  // input
           char *       obj,        // input
           fs_Handle *  objHandle)  // output

{
    uint32              reqSize;
    fs_MsgLookupReq *   req = NULL;
    sim_msg_t           msg;
    fs_MsgLookupResp *  rsp = (fs_MsgLookupResp *) msg.buf;
    
    /* Initialize the request.
     */
    reqSize = sizeof(fs_MsgLookupReq) + strlen(obj);
    req = (fs_MsgLookupReq *) calloc(reqSize, 1);
    req->hdr.requestType = FS_REQ_LOOKUP;
    req->hdr.len = reqSize;
	req->hdr.attrType = FS_POST_ATTRS;
	req->hdr.uid = 0;   // root
	req->hdr.gid = 0;   // root
    req->hdr.proto = FS_PROTO_NFS_V3;
    req->hdr.numGids = 0;
    req->hdr.clientPrivate = 0;
    
    memcpy((void *) &(req->parentHandle), (const void *) dirHandle,
           sizeof(fs_Handle));
    req->nameLen = strlen(obj);
    strcpy(req->fileName, obj);
    
    /* Send the request to FS.
     */
	msg.len = sendAgileMsg(-1, &req, reqSize, 
                           "fs", NFX_NFP_FP2, 1, MGMTPLANE,
                           &waitTo, msg.buf, sizeof(msg.buf));
    NFX_ASSERT((msg.len >= sizeof(fs_MsgLookupResp)) &&
               (rsp->hdr.requestType == FS_REQ_LOOKUP));

    /* A response has been received.
     */
    printf("rsp->hdr.fsStatus = %d\n", rsp->hdr.fsStatus);
    if (rsp->hdr.fsStatus == NFX_OK)
    {
        memcpy((void *) objHandle, (const void *) &(rsp->fileHandle),
               sizeof(fs_Handle));
        NFX_ASSERT(rsp->numAttrs == 1);
        NFX_ASSERT(rsp->attr[0].attrType == FS_POST_ATTRS);
        printf("rsp->attr[0].mode = 0x%x\n", rsp->attr[0].mode);
    }
    
    return(rsp->hdr.fsStatus);
}

/*------------------------------------------------------------------------------
 * Name         : sim_open()
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fssf_status_t
sim_open(fs_Handle *        parentHandle,   // input
         char *             fileName,       // input
         fssf_stateid_t *   stateid)        // output

{
    fssf_openReq_t      req;
    sim_msg_t           msg;
    fssf_openRsp_t *    rsp = (fssf_openRsp_t *) msg.buf;

    /* Initialize the request.
     */
    bzero((void *) &req, sizeof(fssf_openReq_t));
    req.hdr.msgType = FSSF_OPEN_REQ;
    req.hdr.len = sizeof(fssf_openReq_t);
    req.hdr.clientPrivate = getpid();
    req.hdr.protocol = FS_PROTO_CIFS;
    req.owner = 0xDAD;
    memcpy((void *) &req.parentHandle, (const void *) parentHandle,
           sizeof(fs_Handle));
    strcpy(req.pathName, "");
    strcpy(req.fileName, fileName);
    req.createFlag = FALSE;
    //req.createHow;
    //req.u;
    req.shareAccess = FSSF_SHARE_ACCESS_BOTH;
    req.shareDeny = FSSF_SHARE_DENY_NONE;
    
    /* Send the request to FSSF.
     */
	msg.len = sendAgileMsg(-1, &req, sizeof(fssf_openReq_t), 
                           "fssf", NFX_NFP_FP2, 1, MGMTPLANE,
                           &waitTo, msg.buf, sizeof(msg.buf));
    NFX_ASSERT((msg.len == sizeof(fssf_openRsp_t)) &&
               (rsp->hdr.msgType == FSSF_OPEN_RSP));

    /* A response has been received.
     */
    printf("rsp->status = %d\n", rsp->status);
    if (rsp->status == NFX_OK)
    {
        *stateid = rsp->stateid;
        printf("rsp->stateid = 0x%llx\n", rsp->stateid.id);
        printf("rsp->attr.mode = 0x%x\n", rsp->attr.mode);
    }
    
    return(rsp->status);
}

/*------------------------------------------------------------------------------
 * Name         : sim_close()
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fssf_status_t                    // output
sim_close(fssf_stateid_t    stateid)    // input

{
    fssf_closeReq_t     req;
    sim_msg_t           msg;
    fssf_closeRsp_t *   rsp = (fssf_closeRsp_t *) msg.buf;

    /* Initialize the request.
     */
    bzero((void *) &req, sizeof(fssf_closeReq_t));
    req.hdr.msgType = FSSF_CLOSE_REQ;
    req.hdr.len = sizeof(fssf_closeReq_t);
    req.hdr.clientPrivate = getpid();
    req.hdr.protocol = FS_PROTO_CIFS;
    req.stateid = stateid;
    
    /* Send the request to FSSF.
     */
	msg.len = sendAgileMsg(-1, &req, sizeof(fssf_closeReq_t), 
                           "fssf", NFX_NFP_FP2, 1, MGMTPLANE,
                           &waitTo, msg.buf, sizeof(msg.buf));
    NFX_ASSERT((msg.len == sizeof(fssf_closeRsp_t)) &&
               (rsp->hdr.msgType == FSSF_CLOSE_RSP));

    /* A response has been received.
     */
    printf("rsp->status = %d\n", rsp->status);
    
    return(rsp->status);
}

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : It all begins here.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int
main(int argc, char *argv[])

{
	char *      actionStr;
    
    /* Initialize my own name.
     * Discard path if any.
     */
	myName = argv[0];
	if (rindex(myName, '/'))
    {
		strcpy(myName, rindex(myName, '/') + 1);
	}
    
    /* Make sure there's at least the action and the object strings.
     * Get the command and the object on which to execute the command.
     */
	if (argc < 2)
    {
		print_usage();
		exit(NFX_ERR);
	}
    actionStr = argv[1];
      
	if (!strcasecmp(actionStr, "lookup"))
    {
        evm_volId_t volId;
        fs_Status   status;
        fs_Handle   dirHandle;
        fs_Handle   fileHandle;
        
        if (argc != 4)
        {
            print_usage();
            goto sim_exit;
        }
        
        volId = sim_getVolId(argv[2]);
        if (volId.whole_thing == -1)
        {
            goto sim_exit;
        }
        
        status = sim_getRoot(volId, "/", &dirHandle);
        if (status != NFX_OK)
        {
            goto sim_exit;
        }
        
        status = sim_lookup(&dirHandle, argv[3], &fileHandle);
        if (status != NFX_OK)
        {
            goto sim_exit;
        }
        
        printf("Done.\n");
        exit(NFX_OK);
    }
	else if (!strcasecmp(actionStr, "open"))
    {
        evm_volId_t     volId;
        fs_Status       status;
        fs_Handle       dirHandle;
        fssf_stateid_t  stateid;
        
        if (argc != 4)
        {
            print_usage();
            goto sim_exit;
        }
        
        volId = sim_getVolId(argv[2]);
        if (volId.whole_thing == -1)
        {
            goto sim_exit;
        }
        
        status = sim_getRoot(volId, "/", &dirHandle);
        if (status != NFX_OK)
        {
            goto sim_exit;
        }
        
        status = sim_open(&dirHandle, argv[3], &stateid);
        if (status != NFX_OK)
        {
            goto sim_exit;
        }
        
        printf("Done.\n");
        exit(NFX_OK);
    }
	else if (!strcasecmp(actionStr, "close"))
    {
        fssf_status_t   status;
        fssf_stateid_t  stateid;
        
        if (argc != 3)
        {
            print_usage();
            goto sim_exit;
        }

        stateid.id = strtoq(argv[2], NULL, 0);
        
        status = sim_close(stateid);
        if (status != NFX_OK)
        {
            goto sim_exit;
        }
        
        printf("Done.\n");
        exit(NFX_OK);
    }
	else
    {
		fprintf(stderr, "Invalid command: %s\n", actionStr);
		print_usage();
		exit(NFX_ERR);
	}

  sim_exit:
    printf("Command failure.\n");
	exit(NFX_ERR);
}

/*----------------------------------------------------------------------------*/
