
#include <nfx-incl.h>
#include "ssc-stdio.h"
#include <exports/thread-api.h>
#include <exports/esm-api.h>
#include <exports/eee-api.h>
#include <exports/malloc-api.h>

extern uint32 nfxMySlotId, nfxMyCPUId;

#undef DEBUG
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif


static int my_app_id;
static uint32 src_port;
static uint32 dest_port;


void
ssc_stdio_rcv(eee_descPtr_t desc)
{
	esm_event_t ev;
	struct ssc_stdio_msg_hdr *rsp = (struct ssc_stdio_msg_hdr *)
		edesc_firstBuf(desc)->buf;
	ev.p1.u.p = desc;
	ev.event = rsp->type;
	if (thread_event(rsp->handle, &ev) != NFX_OK) {
		printf("ssc_stdio_rcv: thread_event failed\n");
		eee_freePacket(desc);
	}
}


void
ssc_stdio_init(void)
{
	my_app_id = eee_getApplicationId("stdio");
	src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId,
										my_app_id, EEE_MGMT, NFX_COS_HIGH, 0);
	dest_port = PACK_UNICAST_PORT_ADDRESS(0, 0, my_app_id, EEE_MGMT,
										  NFX_COS_HIGH, 0);
	eee_registerReceiveFunc(my_app_id, ssc_stdio_rcv, ssc_stdio_rcv);
	
};


ssc_FILE *
ssc_fopen(char *name, char *mode)
{
	int namelen = strlen(name);
	int modelen = strlen(mode);
	int sz = sizeof(struct ssc_stdio_open_msg) + namelen + modelen + 2;
	struct ssc_stdio_open_msg *msg = eee_ramAlloc(sz);
	char *msg_data = (char *)msg + sizeof(*msg);
	esm_event_t event;
	ssc_FILE *file = NULL;

	debug(("ssc_fopen: %s %s\n", name, mode));
	
	msg->type = SSC_STDIO_OPEN;
	msg->namelen = strlen(name);
	msg->modelen = strlen(mode);
	msg->handle = esm_getCurrHandle();
	memcpy(msg_data, name, namelen + 1);
	memcpy(msg_data + namelen + 1, mode, modelen + 1);
	eee_sendMessage(msg, sz, dest_port, src_port);
	eee_ramDealloc(msg);
	thread_yield(&event, 20);
	if (event.event == SSC_STDIO_OPEN) {
		eee_descPtr_t desc = event.p1.u.p;
		struct ssc_stdio_open_rsp *rsp = (struct ssc_stdio_open_rsp *)
			edesc_firstBuf(desc)->buf;
		if (rsp->fd) {
			file = eee_ramAlloc(sizeof(*file));
			if (file) {
				file->fd = rsp->fd;
			}
		}
		eee_freePacket(desc);
	}
	else {
		debug(("no response\n"));
	}
	return file;
}


int
ssc_fread(char *buf, uint32 sz, uint32 nelm, ssc_FILE *file)
{
	struct ssc_stdio_read_msg msg;
	esm_event_t event;
	int rv = -1;
	msg.type = SSC_STDIO_READ;
	msg.handle = esm_getCurrHandle();
	msg.fd = file->fd;
	msg.sz = sz;
	msg.nelm = nelm;
	eee_sendMessage(&msg, sizeof(msg), dest_port, src_port);
	thread_yield(&event, 20);
	if (event.event == SSC_STDIO_READ) {
		eee_descPtr_t desc = event.p1.u.p;
		struct ssc_stdio_read_rsp *rsp = (struct ssc_stdio_read_rsp *)
			edesc_firstBuf(desc)->buf;
		if (rsp->nelm > 0) {
			eee_copyPacketData(buf, desc, sizeof(*rsp),
							   rsp->nelm * sz, 0);
			rv = rsp->nelm;
		}

		eee_freePacket(desc);
	}
	else {
		debug(("no response\n"));
	}
	return rv;
}


int
ssc_fwrite(char *buf, uint32 sz, uint32 nelm, ssc_FILE *file)
{
	int bufsize = sz * nelm;
	int msg_sz = sizeof(struct ssc_stdio_write_msg) + bufsize;
	struct ssc_stdio_write_msg *msg = eee_ramAlloc(msg_sz);
	esm_event_t event;
	int rv = -1;

	debug(("ssc_fwrite: fd=%x sz=%d nelm=%d\n", file->fd, sz, nelm));
	
	msg->type = SSC_STDIO_WRITE;
	msg->handle = esm_getCurrHandle();
	msg->fd = file->fd;
	msg->sz = sz;
	msg->nelm = nelm;
	memcpy((char *)msg + sizeof(*msg), buf, bufsize);
	eee_sendMessage(msg, msg_sz, dest_port, src_port);
	eee_ramDealloc(msg);
	thread_yield(&event, 20);
	if (event.event == SSC_STDIO_WRITE) {
		eee_descPtr_t desc = event.p1.u.p;
		struct ssc_stdio_write_rsp *rsp = (struct ssc_stdio_write_rsp *)
			edesc_firstBuf(desc)->buf;
		rv = rsp->nelm;
		eee_freePacket(desc);
	}
	else {
		debug(("no response\n"));
	}
	return rv;
}


int
ssc_fclose(ssc_FILE *file)
{
	struct ssc_stdio_close_msg msg;
	esm_event_t event;
	msg.type = SSC_STDIO_CLOSE;
	msg.handle = esm_getCurrHandle();
	msg.fd = file->fd;
	eee_sendMessage(&msg, sizeof(msg), dest_port, src_port);
	thread_yield(&event, 20);
	if (event.event == SSC_STDIO_CLOSE) {
		eee_descPtr_t desc = event.p1.u.p;
		eee_freePacket(desc);
	}
	else {
		debug(("no response\n"));
	}
	eee_ramDealloc(file);
	return 0;
}


int
ssc_fseek(ssc_FILE *file, uint32 off, uint32 whence)
{
	struct ssc_stdio_seek_msg msg;
	esm_event_t event;
	int rv = -1;
	msg.type = SSC_STDIO_SEEK;
	msg.handle = esm_getCurrHandle();
	msg.offset = off;
	msg.whence = whence;
	msg.fd = file->fd;
	eee_sendMessage(&msg, sizeof(msg), dest_port, src_port);
	thread_yield(&event, 20);
	if (event.event == SSC_STDIO_SEEK) {
		eee_descPtr_t desc = event.p1.u.p;
		struct ssc_stdio_seek_rsp *rsp = (struct ssc_stdio_seek_rsp *)
			edesc_firstBuf(desc)->buf;
		rv = rsp->offset;
		eee_freePacket(desc);
	}
	else {
		debug(("no response\n"));
	}
	return rv;
}




