/*  
 *=============================================================================
 * File Name  :    gt64240mii-rx.c
 *
 * Description:    
 *
 * Created    :    08-06-2001
 *
 * Author     :    Tony Wong
 *
 * Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */

#include "gt64240mii.h"
#include <exports/mem-api.h>

extern int gt_mii_diagPortEna(uint32);
extern int gt64240_printBuf(char *, int, int);

/*
 *------------------------------------------------------------------------------
 * Function Name:	gt64240mii_rxPacket
 *
 * Description: 	Transmits a unicast packet to the backplane.
 *
 * Created by:		Tony Wong
 *
 * Data Created:	08-06-2001
 * 					
 * Returns: NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
gt64240mii_rxPacket( uint32 port,
		     gt64240mii_rxCtrl_t *rxc, 
		     uint32 loopCnt )
{
	int i;
	uint32 tmp;
	uint32 fError;
	uint32 cmdsts;
	uint32 pktLen;
	uchar8 *newBuf;
	eee_desc_t *eDesc = NULL;
	gt64240mii_rxdesc_t *rxCur;

	/*
	 * The first descriptor of each packet is the last to be returned, 
	 * so processing is only done upon completion of packet reception.
	 * A more efficient method is possible by working on 2nd, 3rd...
	 * descriptors and saving the first one for last if the packet 
	 * requires more than one descriptor.  
	 *
	 * Doing the normal case of processing complete packets because
	 * performance should still be fine.
	 */

	rxCur = rxc->rxCur;
	cmdsts = rxCur->cmdsts;
	while( !(cmdsts & GT_MII_CMDSTS_OWN) && loopCnt-- ) {
		/*
		 * Allocate a new receive buffer and add to rx queue.
		 */
		newBuf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_LARGE,0);
		if (!newBuf)
		{
			rxc->noBufErr++;
			gt64240mii->port[port].stats.rxNoBuf++;
			return(NFX_ERR); /* wait until more buffers are available */
		}
		/*
		 * Allocate EEE packet descriptor.
		 */
		if(cmdsts & GT_MII_CMDSTS_LAST) {
			eDesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
			if( !eDesc ) {
				eee_deallocateBuffer(newBuf);
				rxc->noDescErr++;
				return(NFX_ERR); /* wait until more descriptors are available */
			}
			eDesc->bd[0].control = 0; /* mark first buffer as unused */
		}
		/*
		 * Store buffer pointers in buffer cache for now.
		 */
		tmp = (uint32)(rxCur->bufptr);
#ifdef FC_USE_NEW_MEM_MAP
		tmp = (uint32)phys_to_virt_cached((address_t)tmp);
#else
		tmp |= 0x80000000;
#endif
/*
		{
		uchar8 *tmp1 = (uchar8 *)tmp;
		int j;
		printf("gt64200mii_rxPacket: bufPtr=%0x\n", tmp);
		for (j=0; j<64; j++)
			printf("%0x ", *tmp1++);
		}
*/
		rxc->bc[rxc->bufCnt].ptr = (uchar8 *)tmp;
		rxc->bc[rxc->bufCnt].len = (rxCur->bufsize_bytecount) & 0xffff;
		rxc->bufCnt++;

		if (cmdsts & GT_MII_CMDSTS_1ST) /* save status from first descriptor */
			rxc->first_cmdsts = cmdsts;

		/*
		 * Replenish buffer and transfer descriptor ownership 
		 * back to the chip.
		 */
#ifdef FC_USE_NEW_MEM_MAP
		tmp = (uint32)virt_to_phys((address_t)newBuf);
#else
		tmp = (uint32)(newBuf) & ~0xA0000000;
#endif
		rxCur->bufptr = (uchar8 *)tmp;
		rxCur->bufsize_bytecount = (GT_MII_RX_BUF_SIZE << 16);
		rxCur->cmdsts = (GT_MII_CMDSTS_OWN | GT_MII_CMDSTS_INTR);
		/* 
		 * Look for the last buffer in the packet 
		 */
		if( (cmdsts & GT_MII_CMDSTS_LAST) ) {
			rxc->bc[rxc->bufCnt-1].len -= 4;
			if( rxc->first_cmdsts & GT_MII_CMDSTS_ERRS ) 
			{  
				fError = 1;
				if( rxc->first_cmdsts & GT_MII_CMDSTS_SHORT ) 
					rxc->shortErr++;
				if( rxc->first_cmdsts & GT_MII_CMDSTS_LONG )
					rxc->longErr++;
				if( rxc->first_cmdsts & GT_MII_CMDSTS_OVR )
					rxc->ovrErr++;
				if( rxc->first_cmdsts & GT_MII_CMDSTS_RXCOL )
					rxc->collErr++;
				if( rxc->first_cmdsts & GT_MII_CMDSTS_ERRCRC ) {
					rxc->crcErr++;
					gt64240mii->port[port].stats.rxCrc++;
				}
			} 
			else
			{
				fError = 0;
				pktLen = 0;
				for(i=1; i <= rxc->bufCnt; i++) {
					eDesc->bd[i].buf = rxc->bc[i-1].ptr;
					eDesc->bd[i].len = rxc->bc[i-1].len;
					eDesc->bd[i].control = EDESC_BD_CMD_IN_USE |
						EDESC_BD_LARGE;
					pktLen += rxc->bc[i-1].len;
				}

				if (rxc->loopback) {
					eDesc->hdr.dest_port = gt64240mii->loopAddr;
					eDesc->hdr.src_port =  gt64240mii->loopAddr;
				}
				else {
#ifdef PROM
					if (((enet_hdr_t *)eDesc->bd[1].buf)->type != EEE_ETH_TYPE) {
						eDesc->hdr.dest_port = gt64240mii->ipAppPort;
						if ((port == 1) && (isFC()))
							eDesc->hdr.src_port = gt64240mii->loopAddrBP;
						else
							eDesc->hdr.src_port = gt64240mii->loopAddr;
					}
					else
#endif					
					{
						/* Used as backplane driver */
						pktLen -= eee_removeHeader(eDesc);
					}
				}

				eDesc->hdr.control = (pktLen & EDESC_TOTAL_LEN) |
					((rxc->bufCnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

				rxc->pkts++;
				rxc->octets += pktLen;
				gt64240mii->port[port].stats.rxPkts++;
				gt64240mii->port[port].stats.rxOctets += pktLen;

				/*
				 * Queue the packet to the EEE layer.
				 */
/*
				{
				uint32	app_id;
				app_id = EEE_GET_APP_ID(eDesc->hdr.dest_port);
				if ((rxc->loopback) && (app_id == gt64240mii->appId))
					eee_deliverPacketToApp(eDesc, app_id); 
				}
*/
				if (gt64240mii->port[port].rx_trace) {
					printf("RX BUF:\n");
					printf("bd[1].buf = 0x%x, len = 0x%x\n",
										(uint32)eDesc->bd[1].buf, eDesc->bd[1].len);
					gt64240_printBuf((char *)(uint32)eDesc->bd[1].buf,
														eDesc->bd[1].len, 0);
				}

				eee_receivePacket(eDesc);

			}

			/*
			 * Clear buffer cache.
			 */
			for(i=0; i < rxc->bufCnt; i++) {
				if(fError && rxc->bc[i].ptr) 
					eee_deallocateBuffer(rxc->bc[i].ptr);
				rxc->bc[i].ptr = 0;
				rxc->bc[i].len = 0;
			}
			rxc->bufCnt = 0;

		} /* if last buffer */

		if( ++rxCur >= rxc->rxqEnd )
			rxCur = rxc->rxqStart;

		cmdsts = rxCur->cmdsts;
		rxc->rxCur = rxCur;
	} /* while */

	return(NFX_OK);

} /* gt64240mii_rxPacket */

/*
 *------------------------------------------------------------------------------
 * Function Name:	gt64240mii_rxPktPoll
 *
 * Description: 	Called by the eee layer and in turn calls 
 *                      gt64240mii_rxPacket for each receive priority queue 
 *                      to collect packet buffers and pass them up to the 
 *                      eee layer.
 *
 * Created by:		Tony Wong
 *
 * Data Created:	08-06-2001
 * 					
 * Returns: none 
 *
 *------------------------------------------------------------------------------
 */
void
gt64240mii_rxPktPoll( void *cb, uint32 tlimit )
{
	int port, i;
	uint32 loopCnt, val;
#ifdef FOR_OLD_COMPILER
	gt64240mii_icr_t icr;
#endif
	gt64240mii_sdcmr_t sdcmr;
	uint32 icr_word = 0;
	
	loopCnt = 128;
	for (port = 0; port < GT_MII_MAX_PORTS; port++) {
		for (i=0; i < GT_MII_RX_Q_CNT; i++) {
#ifdef FOR_OLD_COMPILER
			icr.wd = gt_mii_readPortReg(port, GT_MII_ICR); /* get status */
			if (icr.bits.rxerror) {/* if fell behind with rx queue processing */
#else
			icr_word = gt_mii_readPortReg(port, GT_MII_ICR); /* get status */
			if (icr_word & 0x100) {/* if fell behind with rx queue processing */
#endif
				sdcmr.wd = gt_mii_readPortReg(port, GT_MII_SDCMR);
				sdcmr.bits.ar = 1; /* abort receive */	
				gt_mii_writePortReg(port, GT_MII_SDCMR, (uint32)sdcmr.wd);
				/* 
				 * finish processing what we can - note that receive is 
				 * not re-enabled if out of system buffers or descriptors 
				 */
				if (gt64240mii_rxPacket(port, &gt64240mii->port[port].rxc[i], 
										GT_MII_RX_Q_LEN) == NFX_OK) 
				{
					/* re-align queue pointers and resume */
#ifdef FC_USE_NEW_MEM_MAP
   					val = (uint32)virt_to_phys((address_t)gt64240mii->port[port].rxc[i].rxCur);
#else
   					val = ((uint32)(gt64240mii->port[port].rxc[i].rxCur) & 
							~V2PHYS_ADDR_MASK);
#endif
   					gt_mii_writePortReg(port, GT_MII_FRDP3, val);
   					gt_mii_writePortReg(port, GT_MII_CRDP3, val);
					gt_mii_writePortReg(port, GT_MII_ICR, 0); /* clear status */
					gt_mii_diagPortEna(port);  /* re-enable receive */
				}
			}
			else /* normal processing */
				gt64240mii_rxPacket(port, &gt64240mii->port[port].rxc[i], 
									loopCnt);
		}
	}
	return;

} /* gt64240mii_rxPktPoll */

