/*
 *----------------------------------------------------------------------------
 *
 * Name:         gt64440mii-rx.c
 *
 * Copyright (C) 2004, OnStor, Inc.
 *
 * Description : Contains Disco3 Ethernet Driver RX routines.
 *               Ported from sm-gt64240-mii module.
 *
 * Created by  : Shaun Hui
 *
 * Date Created: 04/13/04
 *
 *-----------------------------------------------------------------
 */

#ifndef SIBYTE_BC

#include "gt64440mii.h"
#include <exports/mem-api.h>

extern int gt_mii_diagPortEna(uint32);
extern int gt64440_printBuf(char *, int, int);

/*
 *------------------------------------------------------------------------------
 * Function Name:	gt64440mii_rxPacket
 *
 * Description: 	Transmits a unicast packet to the backplane.
 *
 * Created by:		Tony Wong
 *
 * Data Created:	08-06-2001
 * 					
 * Returns: NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
gt64440mii_rxPacket( uint32 port,
		     gt64440mii_rxCtrl_t *rxc, 
		     uint32 loopCnt )
{
	int i;
	uint32 tmp;
	uint32 fError;
	uint32 cmdsts;
	uint32 pktLen;
	uchar8 *newBuf;
	eee_desc_t *eDesc = NULL;
	gt64440mii_rxdesc_t *rxCur;

	/*
	 * The first descriptor of each packet is the last to be returned, 
	 * so processing is only done upon completion of packet reception.
	 * A more efficient method is possible by working on 2nd, 3rd...
	 * descriptors and saving the first one for last if the packet 
	 * requires more than one descriptor.  
	 *
	 * Doing the normal case of processing complete packets because
	 * performance should still be fine.
	 */

	rxCur = rxc->rxCur;
	cmdsts = rxCur->cmdsts;
	while( !(cmdsts & RX_O) && loopCnt-- ) {
		/*
		 * Allocate a new receive buffer and add to rx queue.
		 */
		newBuf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_LARGE,0);
		if (!newBuf)
		{
			rxc->noBufErr++;
			gt64440mii->port[port].stats.rxNoBuf++;
			return(NFX_ERR); /* wait until more buffers are available */
		}
		/*
		 * Allocate EEE packet descriptor.
		 */
		if(cmdsts & RX_L) {
			eDesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
			if( !eDesc ) {
				eee_deallocateBuffer(newBuf);
				rxc->noDescErr++;
				return(NFX_ERR); /* wait until more descriptors are available */
			}
			eDesc->bd[0].control = 0; /* mark first buffer as unused */
		}
		/*
		 * Store buffer pointers in buffer cache for now.
		 */
		tmp = (uint32)(rxCur->bufptr);
#ifdef FC_USE_NEW_MEM_MAP
		tmp = (uint32)phys_to_virt_cached((address_t)tmp);
#else
		tmp |= 0x80000000;
#endif
/*
		{
		uchar8 *tmp1 = (uchar8 *)tmp;
		int j;
		printf("gt64200mii_rxPacket: bufPtr=%0x\n", tmp);
		for (j=0; j<64; j++)
			printf("%0x ", *tmp1++);
		}
*/
		rxc->bc[rxc->bufCnt].ptr = (uchar8 *)tmp;
		rxc->bc[rxc->bufCnt].len = (rxCur->bufsize_bytecount) >> 16;
		rxc->bufCnt++;

		if (cmdsts & RX_F) /* save status from first descriptor */
			rxc->first_cmdsts = cmdsts;

		/*
		 * Replenish buffer and transfer descriptor ownership 
		 * back to the chip.
		 */
#ifdef FC_USE_NEW_MEM_MAP
		tmp = (uint32)virt_to_phys((address_t)newBuf);
#else
		tmp = (uint32)(newBuf) & ~0xA0000000;
#endif
		rxCur->bufptr = (uchar8 *)tmp;
		rxCur->bufsize_bytecount = (GT_MII_RX_BUF_SIZE & ~7);
		rxCur->cmdsts = (RX_O | RX_EI);
		/* 
		 * Look for the last buffer in the packet 
		 */
		if( (cmdsts & RX_L) ) {
			rxc->bc[rxc->bufCnt-1].len -= 4;
			if( rxc->first_cmdsts & RX_ES ) 
			{  
				fError = 1;
				if(( rxc->first_cmdsts & RX_EC) == RX_EC_LONG )
					rxc->longErr++;
				else if(( rxc->first_cmdsts & RX_EC) == RX_EC_RSRC )
					rxc->noDescErr++;
				else if(( rxc->first_cmdsts & RX_EC) == RX_EC_OVR )
					rxc->ovrErr++;
				else if(( rxc->first_cmdsts & RX_EC) == RX_EC_CRC){
					rxc->crcErr++;
					gt64440mii->port[port].stats.rxCrc++;
				}
			} 
			else
			{
				fError = 0;
				pktLen = 0;
				for(i=1; i <= rxc->bufCnt; i++) {
					eDesc->bd[i].buf = rxc->bc[i-1].ptr;
					eDesc->bd[i].len = rxc->bc[i-1].len;
					eDesc->bd[i].control = EDESC_BD_CMD_IN_USE |
						EDESC_BD_LARGE;
					pktLen += rxc->bc[i-1].len;
				}

				if (rxc->loopback) {
					eDesc->hdr.dest_port = gt64440mii->loopAddr;
					eDesc->hdr.src_port =  gt64440mii->loopAddr;
				}
				else {
#ifdef PROM
					if (((enet_hdr_t *)eDesc->bd[1].buf)->type != EEE_ETH_TYPE) {
						eDesc->hdr.dest_port = gt64440mii->ipAppPort;
						if ((port == 1) && (isFC()))
							eDesc->hdr.src_port = gt64440mii->loopAddrBP;
						else
							eDesc->hdr.src_port = gt64440mii->loopAddr;
					}
					else
#endif					
					{
						/* Used as backplane driver */
						pktLen -= eee_removeHeader(eDesc);
					}
				}

				eDesc->hdr.control = (pktLen & EDESC_TOTAL_LEN) |
					((rxc->bufCnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

				rxc->pkts++;
				rxc->octets += pktLen;
				gt64440mii->port[port].stats.rxPkts++;
				gt64440mii->port[port].stats.rxOctets += pktLen;

				/*
				 * Queue the packet to the EEE layer.
				 */
/*
				{
				uint32	app_id;
				app_id = EEE_GET_APP_ID(eDesc->hdr.dest_port);
				if ((rxc->loopback) && (app_id == gt64440mii->appId))
					eee_deliverPacketToApp(eDesc, app_id); 
				}
*/
				if (gt64440mii->port[port].rx_trace) {
					printf("RX BUF:\n");
					printf("bd[1].buf = 0x%x, len = 0x%x\n",
										(uint32)eDesc->bd[1].buf, eDesc->bd[1].len);
					gt64440_printBuf((char *)(uint32)eDesc->bd[1].buf,
														eDesc->bd[1].len, 0);
				}

				eee_receivePacket(eDesc);

			}

			/*
			 * Clear buffer cache.
			 */
			for(i=0; i < rxc->bufCnt; i++) {
				if(fError && rxc->bc[i].ptr) 
					eee_deallocateBuffer(rxc->bc[i].ptr);
				rxc->bc[i].ptr = 0;
				rxc->bc[i].len = 0;
			}
			rxc->bufCnt = 0;

		} /* if last buffer */

		if( ++rxCur >= rxc->rxqEnd )
			rxCur = rxc->rxqStart;

		cmdsts = rxCur->cmdsts;
		rxc->rxCur = rxCur;
	} /* while */

	return(NFX_OK);

} /* gt64440mii_rxPacket */

/*
 *------------------------------------------------------------------------------
 * Function Name:	gt64440mii_rxPktPoll
 *
 * Description: 	Called by the eee layer and in turn calls 
 *                      gt64440mii_rxPacket for each receive priority queue 
 *                      to collect packet buffers and pass them up to the 
 *                      eee layer.
 *
 * Created by:		Tony Wong
 *
 * Data Created:	08-06-2001
 * 					
 * Returns: none 
 *
 *------------------------------------------------------------------------------
 */
void
gt64440mii_rxPktPoll( void *cb, uint32 tlimit )
{
	int port, i;
	uint32 loopCnt, val;
	uint32 icr_word = 0;
	
	loopCnt = 128;
	for (port = 0; port < GT_MII_MAX_PORTS; port++) {
		for (i=0; i < GT_MII_RX_Q_CNT; i++) {

			/*
			 * Get Status, check for RX resource error.
			 */
			icr_word = GT_READ( ETH_INTERRUPT_CAUSE_REG(port) );
			if (icr_word & ICR_RXERROR ) {

				/*
				 * Resource error, disable Rx Q 0
				 */
				GT_WRITE(ETH_RECEIVE_QUEUE_COMMAND_REG(port), 0x100);
                                                                                
				/* 
				 * finish processing what we can - note that receive is 
				 * not re-enabled if out of system buffers or descriptors 
				 */
				if (gt64440mii_rxPacket(port, &gt64440mii->port[port].rxc[i], 
										GT_MII_RX_Q_LEN) == NFX_OK) 
				{
					/* re-align queue pointers and resume */
#ifdef FC_USE_NEW_MEM_MAP
   					val = (uint32)virt_to_phys((address_t)gt64440mii->port[port].rxc[i].rxCur);
#else
   					val = ((uint32)(gt64440mii->port[port].rxc[i].rxCur) & 
							~V2PHYS_ADDR_MASK);
#endif
					/*
					 * Update current Rx Desc.
					 * Clear status &
					 * Enable Rx Q 0
					 */
					GT_WRITE(ETH_RX_CURRENT_QUEUE_DESC_PTR_0(port), val);
					GT_WRITE(ETH_INTERRUPT_CAUSE_REG(port),0); 
					GT_WRITE(ETH_RECEIVE_QUEUE_COMMAND_REG(port), 0x1);
				}
			}
			else /* normal processing */
				gt64440mii_rxPacket(port, &gt64440mii->port[port].rxc[i], 
									loopCnt);
		}
	}
	return;

} /* gt64440mii_rxPktPoll */

#endif
