/*
 *-----------------------------------------------------------------
 *
 * Name:	hash-api.c
 *
 * RCS:  $Id: hash-api.c,v 1.7 2002/11/25 18:20:14 sandrineb Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	General purpose hash table implementation.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/3/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/hash-api.h>
#include <exports/malloc-api.h>
#include "hash.h"

/*-----------------------------------------------------------------
 * Name:	hash_create(num_buckets, hash, compare)
 *
 * Description:	Allocate memory for a hash table
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/3/01
 *
 *-----------------------------------------------------------------
 */
address_t			/* returns hash table handle, or NULL */
hash_create(
	uint32	num_buckets,
	uint32	(* hash)(
				void	* key
			),
	int32	(* compare)(
				void	* ht_data,
				void	* search_data
			),
	void 	(* release)(
				void	* data
			)
)
{
	ht_t	* htp;
	uint32	i;

	htp = (ht_t *)eee_ramAlloc(sizeof(ht_t));
	htp->head = (ht_item_t **)eee_ramAlloc(num_buckets * sizeof(ht_item_t *));

	htp->hash = hash;
	htp->compare = compare;
	htp->release = release;
	for (i=0; i< num_buckets; i++)
		htp->head[i] = NULL;

	htp->num_buckets = num_buckets;
	NFX_SPIN_INIT(&htp->ht_spin, SPINLOCK_LEVEL_UNKNOWN);

	return ((address_t)htp);

}	/* hash_create */

/*-----------------------------------------------------------------
 * Name:	hash_destroy(htp)
 *
 * Description:	Destroy a hash table, freeing memory allocated by hash
 *				table routines.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/3/01
 *
 *-----------------------------------------------------------------
 */
void
hash_destroy(
	address_t	htp
)
{
	ht_item_t	* np;
	uint32		i;

	for (i=0; i<((ht_t *)htp)->num_buckets; i++) {
		np = ((ht_t *)htp)->head[i];
		while (np != NULL) {
			ht_item_t	* temp;

			temp = np;
			np = np->next;
			if (((ht_t *)htp)->release)
				(*((ht_t *)htp)->release)(temp->data);
			eee_ramDealloc(temp);
		}
	}

	eee_ramDealloc(((ht_t *)htp)->head);
	eee_ramDealloc((void *)htp);

}	/* hash_destroy */

static inline int
do_compare(struct ht *ht, ht_item_t *np, void *data)
{
	return ht->compare ? ht->compare(np->data, data) == 0 : np->data == data;
}


/*-----------------------------------------------------------------
 * Name:	hash_insert(htp, key, data)
 *
 * Description:	Insert item into hash table.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/3/01
 *
 *-----------------------------------------------------------------
 */
int32						/* return NFX_OK or HT_ERR_DATA_IN_USE */
hash_insert(
	address_t	htp,			/* hash table */
	void		* key,			/* key for hashing function */
	void		* data			/* data structure to insert */
)
{
	ht_item_t	* np;
	uint32		hashval;
	int rc = HT_ERR_DATA_IN_USE;
	struct ht 	*ht = (struct ht *)htp;

	hashval = (*ht->hash)(key);
	if (hashval >= ht->num_buckets)
		panic("hashval(0x%x) > num_buckets(0x%x)\n",
			  hashval, ht->num_buckets);

	NFX_SPIN_LOCK(&ht->ht_spin);
	for (np = ht->head[(int)hashval]; np != NULL; np = np->next) {
		if (do_compare(ht, np, data)) {
			break;
		}
	}
	if (np == NULL) {
		np = hash_getNode();
		np->next = ht->head[(int)hashval];
		np->data = data;
		ht->head[hashval] = np;
		rc = NFX_OK;
	}
	NFX_SPIN_UNLOCK(&ht->ht_spin);

	return rc;

}	/* hash_insert */


/*-----------------------------------------------------------------
 * Name:	hash_lookup(ht, key, data)
 *
 * Description:	Find an item in the hash table.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/3/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns data structure stored in table, or NULL */
hash_lookup(
	address_t	htp,			/* hash table */
	void		* key,			/* key for hashing function */
	void		* data			/* data structure to find */
)
{
	ht_item_t	* np;
	uint32		hashval;
	struct ht 	*ht = (struct ht *)htp;
	void *temp;

	hashval = ht->hash(key);
	if (hashval >= ht->num_buckets)
		panic("hashval(0x%x) > num_buckets(0x%x)\n",
			  hashval, ht->num_buckets);
	
	NFX_SPIN_LOCK(&ht->ht_spin);
	for (np = ht->head[(int)hashval]; np != NULL; np = np->next) {
		if (do_compare(ht, np, data))
			break;
	}
	temp = np ? np->data : NULL;
	NFX_SPIN_UNLOCK(&ht->ht_spin);

	return temp;

}	/* hash_lookup */

/*-----------------------------------------------------------------
 * Name:	hash_delete(ht, key, data)
 *
 * Description:	Remove an item from a hash table.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/3/01
 *
 *-----------------------------------------------------------------
 */
void *			/* returns pointer to data structure, or NULL */
hash_delete(
	address_t	htp,			/* hash table */
	void		* key,			/* key for hashing function */
	void		* data			/* data structure to find and delete */
)
{
	ht_item_t	* np;
	ht_item_t	**prev;
	uint32		hashval;
	void		* temp;
	struct ht 	*ht = (struct ht *)htp;
	

	hashval = (*ht->hash)(key);
	if (hashval >= ht->num_buckets)
		panic("hashval(0x%x) > num_buckets(0x%x)\n",
										hashval, ht->num_buckets);
	NFX_SPIN_LOCK(&ht->ht_spin);

	prev = &ht->head[(int)hashval];
	np = ht->head[(int)hashval];
	while (np != NULL && do_compare(ht, np, data) == 0) {
		prev = &np->next;
		np = np->next;
	}
	if (np != NULL) {
		*prev = np->next;
		temp = np->data;
		hash_freeNode(np);
	}
	else {
		temp = NULL;
	}
	
	NFX_SPIN_UNLOCK(&ht->ht_spin);
	return temp;
}	/* hash_delete */

