/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ipm-util.c
 *
 * RCS:  $Id: ipm-util.c,v 1.16 2002/09/06 19:22:53 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: api to ui functions 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/11/2001
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <nfx-ports.h>
#if defined(SSC)
#include <stdio.h>
#include <string.h>
#include <exports/ncm-api.h>
#include <exports/anpssc-api.h>
#include <exports/ipmd-lib.h>
#endif /* SSC */

#include <exports/queue-api.h>
#include <exports/ipm-api.h>
#include <exports/eee-error.h>
#include "ipm.h"

int32							/* fill in ipm header */
ipm_FillIpmMsgHdr(ipm_mhdr_t *ipm_hdrp, uint32 msg, int size)
{
  ipm_hdrp->ipmh_mtype = msg;
  ipm_hdrp->ipmh_size = size;
  ipm_hdrp->ipmh_flags = 0;
  return NFX_OK;
}

int32							/* fill in ipm ip interface  */
ipm_FillIpmMsgIp(ipm_mip_t *ipm_ipp, uint32 ip, uint32 mask)
{
  ipm_ipp->ip_addr = ip;
  ipm_ipp->ip_mask = mask;
  return NFX_OK;
}

int32							/* fill in ipm rt  */
ipm_FillIpmMsgRt(ipm_mrt_t *ipm_rtp, uint32 ip, uint32 mask, uint32 nhop)
{
  ipm_rtp->rt_addr = ip;
  ipm_rtp->rt_mask = mask;
  ipm_rtp->rt_nhop = nhop;
  return NFX_OK;
}

int32							/* fill in ipm arp  */
ipm_FillIpmMsgArp(ipm_marp_t *ipm_arpp, uint32 ip, char *mac, int size)
{
  ipm_arpp->arp_addr = ip;
  if (mac)
	  memcpy(ipm_arpp->arp_mac, mac, size);
  else
	  bzero(ipm_arpp->arp_mac, 6);
  
  return NFX_OK;
}

#if defined(SSC)
/* only fillin interface identification portion of the structure */ 
int32
ipmd_FillIpmMsgIntfId(ipm_mintf_t *ipm_intfp, char	*ifname)
{
  int	type, slot, cpu, link;
  
  /* convert string */
  if((!ifname) ||
	 (ipmd_ParseIfName(ifname, &type, &slot, &link) != NFX_OK))
	return NFX_ERR;

  /* convert link number to internal cpuid and link id */
  cpu = GET_CPU_FROM_ULINK(link);
  link = GET_LINK_FROM_ULINK(link);

  /* fill in device identification */
  ipm_intfp->if_dtype = type;
  ipm_intfp->if_slot = slot;
  ipm_intfp->if_cpu = cpu;
  ipm_intfp->if_link = link;

  return NFX_OK;
}

struct timeval	waitTo = {5, 0}; /* wait 5 seconds */
struct timeval	txrxWaitTo = {2, 0}; /* wait 2 seconds */

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_SendMsg
 *
 * Description: send contrl msg to txrx 
 *
 * Created by:	Su Wang
 *
 * Date Created: 02/08/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* -1 fail, 0 no response, >0 response */
ipmd_SendMsg(void		*msg,	/* ptr to real msg */
			 uint32		size,	/* size of msg */
			 int		slot,
			 int		cpu,
			 char		*app,
			 void		*resp,	/* does user want response */
			 int		resp_len,
			 int		bcast,
			 int		sock)	/* is this bcast msg */
{
	if (bcast) {
		/* no bcast yet, send to TXRX */
		slot = 1;
		cpu  = 0; /* NFX_NFP_TX */
		bcast = 0;
	}
	
	if (resp && bcast) /* and resp mutually exclusive */
		return -1;
	
	//printf("ipmd_SendMsg: send msg(size = %d) to slot %d, cpu %d, app %s,sock=%d \n", 
	//		size, slot, cpu, app, sock);
	return sendAgileMsg(sock, msg, size, app, cpu, slot, MGMTPLANE,
						&txrxWaitTo, resp, resp_len);
	
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_SendMsgByNcm
 *
 * Description: send msg to NCM, and let it route msg. This is used
 *				by application and ipmd to talk to one another
 *
 * Created by:	Su Wang
 *
 * Date Created: 02/08/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* -1, 0, >0 */
ipmd_SendMsgByNcm(int		sock,
				  void		*msg,
				  uint32	size,
				  char		*app,
				  char		*node,
				  void*		resp,
				  int		resp_len)
{
#if 1	/* IPM_NCM_HDR */
	ncmMsg_t	*ncm_hdrp;
	int			rc;

	/* fill in ncm hdr */
	ncm_hdrp = (ncmMsg_t*)(((char*)msg) - IPM_NCM_HDR);
	ncm_hdrp->msgType = NCM_SEND;
	strcpy(ncm_hdrp->appName, app);
	strcpy(ncm_hdrp->nodeName, node);
	
	/* update the msg size */
	size += IPM_NCM_HDR;

	/* update the response buf ptr to reflect ncm hdr */
	if (resp) {
		resp = (void*)(((char*)resp) - IPM_NCM_HDR);
		resp_len += IPM_NCM_HDR;
		ncm_hdrp->rqResp = 1;
	}
	
	rc = sendAgileMsg(sock, ncm_hdrp, size, "ncm", NFX_SSC_UNIX, SSC_SLOT,
					  MGMTPLANE, &waitTo, resp, resp_len);

	/* if rcv rely, deduct ncm hdrsize from rc */
	if (rc > 0)
		rc -= IPM_NCM_HDR;

	return rc;
#else
	/* send to application directly */
	return sendAgileMsg(sock, msg, size, app, NFX_SSC_UNIX, SSC_SLOT,
						MGMTPLANE, &waitTo, resp, resp_len);
#endif /* IPM_NCM_HDR */		
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ParseIfName
 *
 * Description: parse ifname string to type, slot, link int. 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_ParseIfName(char			*ifname,
				 int			*type,
				 int			*slot,
				 int			*link)
{
	boardType_t	intype;
	char		node[24];
  
	if (getPortInfo(ifname, node, &intype, slot, link)) {
//		printf("bad interface %s format\n", ifname);
		return NFX_ERR;
	}
	
	*type = (int)intype;
	// printf("interface %s, with type %d\n", ifname, *type);
	return NFX_OK;
}
#endif /* SSC */

/*
 *-----------------------------------------------------------------
 *
 * NAME : bsd_ShowPkt
 *
 * Description: Display a packet  
 *
 * Created by: Su Wang
 *
 * Data Created: 02/13/01
 *
 *-----------------------------------------------------------------
 */
void
bsd_ShowPkt(char* pkt, int len)
{

  int	 new_line = 1;
  uint32 count = 0;

  if (!pkt || !len)
    return;

  printf("\n");
  printf("Print packet size %d\n", len);
  while (len > count)
    {
      if (new_line)
	{
	  printf("0x%08lx:       ", (unsigned long)count);
	  new_line = 0;
	}
      
      printf("%02X ", (*pkt++) & 0xff);
      count++;
      
      if (count % 16 == 0)
	{
	  printf("\n");
	  new_line = 1;
	}
    }

  printf("\n\n");
  return;

}

void
ipm_LogErr(char *fileName, int32 lineNum, int32 errCode )
{
 	char	*errStr;
	int32	do_panic;

	do_panic = 0;
	switch (errCode)
	  {
	  case IPM_ERR_ALLOC:
		errStr = "No buffer allocated";
		break;

	  case IPM_ERR_QUEUE:
		errStr = "queueing error";
		break;

	  case IPM_ERR_MTYPE:
		errStr = "bad msg type";
		do_panic = 1;
		break;

	  case IPM_ERR_SEND:
		errStr = "bad msg send";
		break;

	  case IPM_ERR_RCV:
		errStr ="bad msg rcved";
		break;

	  default:
		errStr = "Unknown ipm error";
		do_panic = 1;
		break;
	  }
	printf("IMP Error (%d): %s! File: %s Line: %d\n", 
			errCode, errStr, fileName, lineNum);

	if (do_panic)
	  IPM_PANIC();
}


void ipm_PrintStats(ipm_stats_t		*ipm_statsp)
{
  printf("IPM stats\n");
  printf("--------------------------------------------------\n");
  printf("              Msg Forwarded			Msg Reived\n");
  printf("Admin:		%-18d%d\n", ipm_statsp->adm_fwd,  ipm_statsp->adm_rcv);
  printf("Device:		%-18d%d\n", ipm_statsp->dev_fwd,  ipm_statsp->dev_rcv);
  printf("Inform:		%-18d%d\n", ipm_statsp->info_fwd,  ipm_statsp->info_rcv);
  printf("Error:		%-18d%d\n", ipm_statsp->err_fwd,  ipm_statsp->err_rcv);
  printf("confstats_req %-18d\n", ipm_statsp->req_rcv);
#if defined(NFP_TXRX)
  printf("tcp quelen   %d,		udp quelen %d\n", ipm_statsp->connTcpNum, 
		 ipm_statsp->connUdpNum);
#endif /* NFP_TXRX */
  
  return;
}

void ipm_ClearStats(ipm_stats_t		*ipm_statsp)
{
  bzero(ipm_statsp, sizeof(*ipm_statsp));
  return;
}
