/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ipmd-lib.h
 *
 * RCS:  $Id: ipmd-lib.h,v 1.14 2002/10/24 17:21:30 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: ipm libary functions on ssc. All IP/MAC addresses 
 *				must be in network order. Interface name must follow
 *				convention "type:slot/link".
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
#ifndef _IPMD_LIB_H
#define _IPMD_LIB_H

#include <nfx-error.h>
#include <nfx-types.h>
#include <nfx-defs.h>

#if !defined(PROM)
#define LINK_AGG
#endif /* PROM */

#define IPM_VIF_STR				"vif"

#define IPM_STR_SIZE			24

#define IPM_SSC_BUFSIZE			4096

#define IPM_CMDERR_TIMEOUT		-11
#define IPM_CMDERR_NOINTF		-12
#define IPM_CMDERR_NORT			-13
#define IPM_CMDERR_DUPRT		-14
#define IPM_CMDERR_SYSNOBUF		-15
#define IPM_CMDERR_APPNOBUF		-16
#define IPM_CMDERR_IPMTIMEOUT	-17
#define IPM_CMDERR_DUPIP		-18
#define IPM_CMDERR_NOIP			-19
#define IPM_CMDERR_NOGW			-20

/* virtual interface related errors */
#define IPM_CMDERR_NOVIF		-40
#define IPM_CMDERR_DUPVIF		-41
#define IPM_CMDERR_HASIP_VIF	-42
#define IPM_CMDERR_ACTLINK_VIF	-43
#define IPM_CMDERR_BADMODE_VIF	-44
#define IPM_CMDERR_NOLINK_VIF	-45
#define IPM_CMDERR_MANYLINK_VIF	-46
#define IPM_CMDERR_MANYVIF		-47

#define IPM_GET_ERR				NFX_ERR	/* some err */
#define IPM_GET_OK				NFX_OK
#define IPM_GET_NO_BUF			IPM_CMDERR_SYSNOBUF /* buff provided too small */


/* 
 *-----------------------------------------------------------------
 * APIs for application to manipulate IP stack configuration.. 
 *
 * Expected Behavior:
 * 
 *  Add IP:
 *  TXRX		SSC					Result
 *  No			No					IP added, No err
 *  No			Yes, same fp		IP added on Txrx, No err
 *  No			Yes, diff fp		IP not added, Err
 *  Yes			Yes					Do nothing, No Err
 *  Yes			No					IP added, No err
 *
 * Delete IP:
 * TXRX			SSC					Result
 * No			Yes					IP deleted on ssc, No err
 * No			No					Err
 * No			Yes, diff fp		Err
 * Yes			Yes					IP deleted, No err
 * Yes			No					IP delete on Txrx, Err
 *
 * Add Route:
 * TXRX			SSC					Result
 * No			Yes, same gw		Rt added on Txrx, No err
 * No			Yes, diff gw		Err
 * No			No					Rt added, No err
 * Yes			No					Rt added on ssc, No Err
 * Yes			Yes					Do nothing, no error 		
 *
 * Delete Route:
 * TXRX			SSC					Result
 * No			No					Err
 * No			Yes				    Rt deleted on ssc, No Err
 * Yes			No					Rt deleted on txrx, Err
 * Yes			Yes					Rt deleted, No Err
 *-----------------------------------------------------------------
 */

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_PrntErrMsg	
 *
 * Description: print err messages
 *
 *-----------------------------------------------------------------
 */
void
ipmd_PrntErrMsg(int			errType, 
				char		*ifname,
				char		*vifname,
				uint32		ipaddr1,
				uint32		ipaddr2);

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_EnableIntfCluster	
 *
 * Description: enable a given interface across system.
 *
 *
 * parameters:
 * char	*ifname,	// device name 
 * char	*app,		// requestor's appl. name, optional 
 * int	sock,		// which socket use to send, optional
 * char	*node		// node of interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_EnableIntfCluster(char	*ifname,
					   char	*app,
					   int	sock,
					   char	*node);
#define ipmd_EnableIntf(_ifname, _app, _sock)	\
		ipmd_EnableIntfCluster(_ifname, _app, _sock, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DisableIntfCluster	
 *
 * Description: disable a given interface across system.
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DisableIntfCluster(char	*ifname,
						char	*app, 
						int		sock, 
						char	*node);
#define ipmd_DisableIntf(_ifname, _app, _sock)	\
		ipmd_DisableIntfCluster(_ifname, _app, _sock, "local")


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_EnableIntfAutoCluster	
 *
 * Description: enable gigE auto negociation.
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/08/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_EnableIntfAutoCluster(char	*ifname,
						char	*app, 
						int		sock, 
						char	*node);
#define ipmd_EnableIntfAuto(_ifname, _app, _sock)	\
		ipmd_EnableIntfAutoCluster(_ifname, _app, _sock, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DisableIntfAutoCluster	
 *
 * Description: disable gigE auto negociation.
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/08/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DisableIntfAutoCluster(char	*ifname,
							char	*app, 
							int		sock, 
							char	*node);
#define ipmd_DisableIntfAuto(_ifname, _app, _sock)	\
		ipmd_DisableIntfAutoCluster(_ifname, _app, _sock, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_SetIntfMtuCluster	
 *
 * Description: set intf Mtu size
 *
 * Created by:	Su Wang
 *
 * Date Created: 2/28/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_SetIntfMtuCluster (char	*intf,
						int		mtu,
						char	*app,	
						int		sock,
						char	*node);


/* virtual interface info */
#define IPM_MAXLINK_VIF		4 /* max. links per vif */
#define IPM_MAX_VIF			4 /* max vifs in system */
typedef struct ipm_vif_t {
	char	vif_name[IPM_STR_SIZE];
	char	vif_mac[8];
	char	vif_links[IPM_MAXLINK_VIF][IPM_STR_SIZE];
	char	vif_prefLink[IPM_STR_SIZE]; /* prefered link by user */
	char	vif_actLink[IPM_STR_SIZE]; /* active link */
	uint32	vif_linkCnt;
#define IPM_VIF_MODE_FO	0x1	/* fail-over mode */
#define IPM_VIF_MODE_LA	0x2	/* link aggregation mode */
	uint32	vif_flags;
	uint32	vif_slot;			/* slot this vif resides */
	uint32	vif_ip;
	uint32  vif_mask;
	uint32  vif_ipCnt;
} ipm_vif_t;

#if defined(LINK_AGG)

void ipmd_PrintVif(ipm_vif_t *ipm_vifp, uint32 mtype);

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_CreateVifCluster
 *
 * Description: create virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, or ERRs */
ipmd_CreateVifCluster (char		*vif, /* interface to be created */
					   char		*links, /* link contained in vif */
					   int		count,	/* num. of links */
					   uint32	flags,
					   char		*app,	
					   int		sock,
					   char		*node);

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DeleteVifCluster
 *
 * Description: delete virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, or ERRs */
ipmd_DeleteVifCluster (char		*vif, /* interface to be created */
					   char		*app,	
					   int		sock,
					   char		*node);


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_PrefVifCluster
 *
 * Description: setup prefered link within a application
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, or ERRs */
ipmd_PrefVifCluster (char		*vif, /* interface to be created */
					char		*link,
					char		*app,	
					int			sock,
					char		*node);


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_AddLinkVifCluster
 *
 * Description: add link(s) to existing virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, or ERRs */
ipmd_AddLinkVifCluster(char		*vif, /* interface to be created */
				   char		*links, /* link contained in vif */
				   int		count,	/* num. of links */
				   char		*app,	
				   int		sock,
				   char		*node);

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_RmLinkVifCluster
 *
 * Description: remove link(s) to existing virtual interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, or ERRs */
ipmd_RmLinkVifCluster(char		*vif, /* interface to be created */
					  char		*links, /* link contained in vif */
					  int		count,	/* num. of links */
					  char		*app,	
					  int		sock,
					  char		*node);


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_GetVifByNameCluster
 *
 * Description: get virtual interface info
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, or ERRs */
ipmd_GetVifByNameCluster(char		*vif, /* NULL to get all */
						 ipm_vif_t	*buf,
						 int		bufsize,
						 int		*count,
						 int		*need,
						 char		*node);
#endif /* LINK_AGG */

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_AddIpCluster
 *
 * Description: add an ip to interface across the system.
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_AddIpCluster(char*		ifname,	
				  uint32	ip,
				  uint32	mask,	
				  char		*app,
				  int		sock,
				  char		*node);
#define ipmd_AddIp(_ifname, _ip, _mask, _app, _sock)	\
		ipmd_AddIpCluster(_ifname, _ip, _mask, _app, _sock, "local")



/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DelIpCluster
 *
 * Description: delete an ip to interface across the system.
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DelIpCluster( char*		ifname,	
				   uint32		ip,
				   char			*app,
				   int			sock,
				   char			*node);
#define ipmd_DelIp(_ifname, _ip, _app, _sock)	\
		ipmd_DelIpCluster(_ifname, _ip, _app, _sock, "local")


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_AddRtCluster
 *
 * Description: add a static rt across system
 * 
 * parameters:
 * uint32		ip,		// route address
 * uint32		mask,	// route mask
 * uint32	    nexthop,// route gateway
 * char			*app	// appl. name, optional	
 * int			sock,	// socket used to send, optional
 * char			*node	// node of interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_AddRtCluster( uint32		ip,	
				   uint32		mask,	
				   uint32	    nexthop,
				   char			*app,
				   int			sock,
				   char			*node);
#define	ipmd_AddRt(_ip, _mask, _nexthop, _app, _sock)	\
		ipmd_AddRtCluster(_ip, _mask, _nexthop, _app, _sock, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DelRtCluster
 *
 * Description: delete a static rt across system

 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DelRtCluster(uint32		ip,	
				  uint32		mask,	
				  uint32	    nexthop,
				  char			*app,
				  int			sock,
				  char			*node);
#define ipmd_DelRt(_ip, _mask, _nexthop, _app, _sock)	\
		ipmd_DelRtCluster(_ip, _mask, _nexthop, _app, _sock, "local")


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_AddArpCluster
 *
 * Description: add a static arp across system
 *
 * parameters:
 * uint32		ip,		// ip address of mac
 * char			*mac,	// mac associated with ip
 * char			*app,	// appl. name, optional
 * int			sock,	// socket to send, optional
 * char			*node	// node of interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_AddArpCluster(uint32		ip,	
				   char			*mac,
				   char			*app,
				   int			sock,
				   char			*node);
#define ipmd_AddArp(_ip, _mac, _app, _sock)		\
		ipmd_AddArpCluster(_ip, _mac, _app, _sock, "local")


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DelArpCluster
 *
 * Description: delete a static arp across system

 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DelArpCluster(uint32		ip,	
				   char			*app,
				   int			sock,
				   char			*node);
#define ipmd_DelArp(_ip, _app, _sock)		\
		ipmd_DelArpCluster(_ip, _app, _sock, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ShutDownCluster()
 *
 * Description: This function can be used by application to shutdown
 *				IPMD
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32					
ipmd_ShutDownCluster(char		*app,
					 char		*node);
#define ipmd_ShutDown(_app)	ipmd_ShutDownCluster(_app, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_RestartCluster()
 *
 * Description: This function can be used by application to restart
 *				IPMD
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/13/2001
 *
 *-----------------------------------------------------------------
 */
int32					
ipmd_RestartCluster(char		*app,
					char		*node);
#define ipmd_Restart(_app)	ipmd_RestartCluster(_app, "local")

/* 
 *-----------------------------------------------------------------
 *
 * the following functions are used by application to retrive kernel 
 * IP information.
 * 
 * One unfortunate side effect is that there is no static memory inside
 * of libary functions. The application must allocate
 * large enough buffer in case there are many instances that need
 * to be retrived. On the other hand, intf, IP, RT, and ARP should be
 * limited on our system. An error should be returned if buffer is
 * not large enough, with required size hinted.
 *
 *-----------------------------------------------------------------
 */

/* device intf struct to application */
typedef struct ipm_if {
  uchar8		if_name[IPM_STR_SIZE];	/* interface name */
  uchar8		if_vifName[IPM_STR_SIZE];
  ushort16		if_slot;
  ushort16		if_cpu;
  ushort16		if_link;
  ushort16		if_flags;		/* interface flags */
  ushort16		if_mtusize;
#define IPM_AUXIF_RUNNING	0x1	/* operational state up */
#define IPM_AUXIF_AUTO		0x2	/* auto negociate on */
  ushort16		if_auxFlags;	
  char			if_mac[8];
} ipm_if_t;

/* ip intf structure to application */
typedef struct ipm_ip {
  uint32		ip_addr;		/* network order */
  uint32		ip_mask;
  uchar8		ip_ifname[IPM_STR_SIZE];	/* device associated with ip interface */
  uchar8		ip_vifName[IPM_STR_SIZE];
} ipm_ip_t;

/* ip rt structure to application */
typedef struct ipm_rt {
  uint32		rt_addr;
  uint32		rt_mask;
  uint32		rt_nhop;
  uint32		rt_flags;
  uchar8		rt_ifname[IPM_STR_SIZE];
  uchar8		rt_vifName[IPM_STR_SIZE];
} ipm_rt_t;

/* ip arp structure to application */
typedef struct ipm_arp {
  uint32		arp_addr;
  uchar8		arp_mac[8];
  uint32		arp_age;
  uchar8		arp_ifname[IPM_STR_SIZE];
  uchar8		arp_vifName[IPM_STR_SIZE];
} ipm_arp_t;


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_GetIpByNameCluster
 *
 * Description: get IP interface associated with a device. 
 *				if device is NULL, cpy all ip interface into
 *				buff in consecutive ipm_if_t.
 *
 * parameters:
 * char		*ifname,	// device name, NULL get all devices
 * ipm_ip_t *buf,		// buff to cpy to, 
 * int		bufsize,	// size of buff 
 * int		*count,		// how many record being cpy in buf 
 * int		*need,		// hint on how many buffer needed, 
 * char		*node		// node of interest
 * 
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_ERR,IPM_GET_NO_BUF, IPM_GET_OK */
ipmd_GetIpByNameCluster(char	*ifname,
						ipm_ip_t *buf, 
						int		bufsize,
						int		*count,
						int		*need,	
						char	*node);
#define ipmd_GetIpByName(_ifname, _buf, _bufsize, _count, _need)	\
ipmd_GetIpByNameCluster(_ifname, _buf, _bufsize, _count, _need, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_GetIfByNameCluster
 *
 * Description: get device if configuration, if device = NULL,
 *				all device interface are copied
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_ERR, IPM_GET_OK, IPM_GET_NO_BUF */
ipmd_GetIfByNameCluster(char	*ifname,
						ipm_if_t *buf,
						int		bufsize,
						int		*count,
						int		*need,
						char	*node);
#define ipmd_GetIfByName(_ifname, _buf, _bufsize, _count, _need)	\
ipmd_GetIfByNameCluster(_ifname, _buf, _bufsize, _count, _need, "local")


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_GetL2ConnEnt
 *
 * Description: Get entry of given IP from Lay2 connection table. 
 *				count indicates number of successful retrivals.
 *
 * Created by:	Su Wang
 *
 * Date Created: 5/28/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_ERR, IPM_GET_OK, IPM_GET_NO_BUF */
ipmd_GetL2ConnEntCluster(uint32		ip,
						 ipm_arp_t	*buf,
						 int		bufsize,
						 int		*count,
						 int		*needed,
						 char		*node);
#define ipmd_GetL2ConnEnt(_ip, _buf, _bufsize, _count, _need)	\
ipmd_GetL2ConnEntCluster(_ip, _buf, _bufsize, _count, _need, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_GetRtTblCluster
 *
 * Description: get entire route table in buf. 
 *
 * parameters:
 * ipm_rt_t	*buf,	// buff to cpy to, 
 * int		bufsize,// size of buff 
 * int		*count,	// how many being cpyed
 * int		*need,	// hint on how many buffer needed
 * char		*node	// node of interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetRtTblCluster(ipm_rt_t	*buf,	
					 int		bufsize,
					 int		*count,	
					 int		*need,
					 char		*node);
#define ipmd_GetRtTbl(_buf, _bufsize, _count, _need)	\
ipmd_GetRtTblCluster(_buf, _bufsize, _count, _need, "local")

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_GetArpTblCluster
 *
 * Description: get entire ARP table in buf. 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_NO_BUF */
ipmd_GetArpTblCluster(ipm_arp_t *buf,
					  int		bufsize,
					  int		*count,
					  int		*need,
					  char		*node);
#define ipmd_GetArpTbl(_buf, _bufsize, _count, _need)		\
ipmd_GetArpTblCluster(_buf, _bufsize, _count, _need, "local")


/* 
 *-----------------------------------------------------------------
 * The following functions retrive IP statistics from shared memory. These
 * are mainly used by SNMP and UI to display stats
 *-----------------------------------------------------------------
 */
typedef struct ipm_ifstat {
  char		ifi_ifname[IPM_STR_SIZE]; /* interface name */
  uint32	ifi_speed;			/* interface speed */
  uint32	ifi_oQlen;			/* intf outbound queue len. */

  /* if_data structure in if.h */
  /* generic interface information */
  char      ifi_type;			/* ethernet, tokenring, etc */
  char	    ifi_physical;		/* e.g., AUI, Thinnet, 10base-T, etc */
  char	    ifi_addrlen;		/* media address length */
  char	    ifi_hdrlen;			/* media header length */
  char	    ifi_recvquota;		/* polling quota for receive intrs */
  char	    ifi_xmitquota;		/* polling quota for xmit intrs */
  uint32	ifi_mtu;			/* maximum transmission unit */
  uint32	ifi_metric;			/* routing metric (external only) */
  uint32	ifi_baudrate;		/* linespeed */
  /* volatile statistics */
  uint32	ifi_ipackets;		/* packets received on interface */
  uint32	ifi_ierrors;		/* input errors on interface */
  uint32	ifi_opackets;		/* packets sent on interface */
  uint32	ifi_oerrors;		/* output errors on interface */
  uint32	ifi_collisions;		/* collisions on csma interfaces */
  uint32	ifi_ibytes;			/* total number of octets received */
  uint32	ifi_obytes;			/* total number of octets sent */
  uint32	ifi_imcasts;		/* packets received via multicast */
  uint32	ifi_omcasts;		/* packets sent via multicast */
  uint32	ifi_iqdrops;		/* dropped on input, this interface */
  uint32	ifi_noproto;		/* destined for unsupported protocol */
  uint32	ifi_recvtiming;		/* usec spent receiving when timing */
  uint32	ifi_xmittiming;		/* usec spent xmitting when timing */
  uint32	ifi_lastchange_sec;	/* time of last administrative change in sec*/
  uint32	ifi_lastchange_usec; /* time of last administrative change in micro sec*/
} ipm_ifstat_t;

typedef struct ipm_ipstat {
  uint32	ips_card;		/* card number */
  uint32    ips_defaultTTL;

  /* ifstat structure in ip_var.h */
  uint32	ips_total;		/* total packets received */
  uint32	ips_badsum;		/* checksum bad */
  uint32	ips_tooshort;		/* packet too short */
  uint32	ips_toosmall;		/* not enough data */
  uint32	ips_badhlen;		/* ip header length < data size */
  uint32	ips_badlen;		/* ip length < ip header length */
  uint32	ips_fragments;		/* fragments received */
  uint32	ips_fragdropped;	/* frags dropped (dups, out of space) */
  uint32	ips_fragtimeout;	/* fragments timed out */
  uint32	ips_forward;		/* packets forwarded */
  uint32	ips_fastforward;	/* packets fast forwarded */
  uint32	ips_cantforward;	/* packets rcvd for unreachable dest */
  uint32	ips_redirectsent;	/* packets forwarded on same net */
  uint32	ips_noproto;		/* unknown or unsupported protocol */
  uint32	ips_delivered;		/* datagrams delivered to upper level*/
  uint32	ips_localout;		/* total ip packets generated here */
  uint32	ips_odropped;		/* lost packets due to nobufs, etc. */
  uint32	ips_reassembled;	/* total packets reassembled ok */
  uint32	ips_fragmented;		/* datagrams successfully fragmented */
  uint32	ips_ofragments;		/* output fragments created */
  uint32	ips_cantfrag;		/* don't fragment flag was set, etc. */
  uint32	ips_badoptions;		/* error in option processing */
  uint32	ips_noroute;		/* packets discarded due to no route */
  uint32	ips_badvers;		/* ip version != 4 */
  uint32	ips_rawout;		    /* total raw ip packets generated */
  uint32	ips_rawin;

  uint32	rl_arpfwd;			/* ip relayer stats */
  uint32    rl_arprcv;
  uint32    rl_ipfwd;
  uint32    rl_iprcv;
  uint32	rl_arpfwderr;
  uint32    rl_arprcverr;
  uint32    rl_ipfwderr;
  uint32	rl_iprcverr;		
  uint32	rl_noproto;			/* unrecognized protocol */
  uint32    rl_fragsfwd;		/* # of fragments sent out */
  uint32    rl_fragedfwd;		/* # of ip has been fragmented */
  uint32	rl_fragsfwderr;		/* # of errs during fragmentation */
  uint32    rl_reasrcv;			/* # of ip reassembled */
  uint32	rl_fragsrcv;		/* # of fragments received */
  uint32    rl_fragsrcverr;		/* # of errs during reassemble */
  uint32	ips_toolong;		/* ip length > max ip packet size */
  uint32	ips_notmember;		/* multicasts for unregistered grps */
} ipm_ipstat_t;

#if !defined(ICMP_MAXTYPE)
#define ICMP_MAXTYPE	18
#endif /* ICMP_MAXTYPE */

typedef struct ipm_icmpstat {
  uint32	icps_card;		/* card number */
  
  /* icmpstat structure in icmp_var.h */
  /* statistics related to icmp packets generated */
  uint32	icps_error;		/* # of calls to icmp_error */
  uint32	icps_oldshort;		/* no error 'cuz old ip too short */
  uint32	icps_oldicmp;		/* no error 'cuz old was icmp */
  uint32	icps_outhist[ICMP_MAXTYPE + 1];
  /* statistics related to input messages processed */
  uint32	icps_badcode;		/* icmp_code out of range */
  uint32	icps_tooshort;		/* packet < ICMP_MINLEN */
  uint32	icps_checksum;		/* bad checksum */
  uint32	icps_badlen;		/* calculated bound mismatch */
  uint32	icps_reflect;		/* number of responses */
  uint32	icps_inhist[ICMP_MAXTYPE + 1];
  uint32	icps_bmcastecho; 	/* b/mcast echo requests dropped */
  uint32	icps_bmcasttstamp; 	/* b/mcast tstamp requests dropped */
} ipm_icpstat_t;

typedef struct ipm_tcpstat {
  uint32	tcps_card;		/* card number */

  /* tcpstat structure in tcp_var.h */
  uint32	tcps_connattempt;	/* connections initiated */
  uint32	tcps_accepts;		/* connections accepted */
  uint32	tcps_connects;		/* connections established */
  uint32	tcps_drops;			/* connections dropped */
  uint32	tcps_conndrops;		/* embryonic connections dropped */
  uint32	tcps_closed;		/* conn. closed (includes drops) */
  uint32	tcps_segstimed;		/* segs where we tried to get rtt */
  uint32	tcps_rttupdated;	/* times we succeeded */
  uint32	tcps_delack;		/* delayed acks sent */
  uint32	tcps_timeoutdrop;	/* conn. dropped in rxmt timeout */
  uint32	tcps_rexmttimeo;	/* retransmit timeouts */
  uint32	tcps_persisttimeo;	/* persist timeouts */
  uint32	tcps_keeptimeo;		/* keepalive timeouts */
  uint32	tcps_keepprobe;		/* keepalive probes sent */
  uint32	tcps_keepdrops;		/* connections dropped in keepalive */
  
  uint32	tcps_sndtotal;		/* total packets sent */
  uint32    tcps_sndfast;		/* send via fast path */
  uint32	tcps_sndpack;		/* data packets sent */
  uint32	tcps_sndbyte;		/* data bytes sent */
  uint32	tcps_sndrexmitpack;	/* data packets retransmitted */
  uint32	tcps_sndrexmitbyte;	/* data bytes retransmitted */
  uint32	tcps_sndacks;		/* ack-only packets sent */
  uint32	tcps_sndprobe;		/* window probes sent */
  uint32	tcps_sndurg;		/* packets sent with URG only */
  uint32	tcps_sndwinup;		/* window update-only packets sent */
  uint32	tcps_sndctrl;		/* control (SYN|FIN|RST) packets sent */
  
  uint32	tcps_rcvtotal;		/* total packets received */
  uint32	tcps_rcvpack;		/* packets received in sequence */
  uint32	tcps_rcvbyte;		/* bytes received in sequence */
  uint32	tcps_rcvbadsum;		/* packets received with ccksum errs */
  uint32	tcps_rcvbadoff;		/* packets received with bad offset */
  uint32	tcps_rcvshort;		/* packets received too short */
  uint32	tcps_rcvduppack;	/* duplicate-only packets received */
  uint32	tcps_rcvdupbyte;	/* duplicate-only bytes received */
  uint32	tcps_rcvpartduppack; /* packets with some duplicate data */
  uint32	tcps_rcvpartdupbyte; /* dup. bytes in part-dup. packets */
  uint32	tcps_rcvoopack;		/* out-of-order packets received */
  uint32	tcps_rcvoobyte;		/* out-of-order bytes received */
  uint32	tcps_rcvpackafterwin; /* packets with data after window */
  uint32	tcps_rcvbyteafterwin; /* bytes rcvd after window */
  uint32	tcps_rcvafterclose;	/* packets rcvd after "close" */
  uint32	tcps_rcvwinprobe;	/* rcvd window probe packets */
  uint32	tcps_rcvdupack;		/* rcvd duplicate acks */
  uint32	tcps_rcvacktoomuch;	/* rcvd acks for unsent data */
  uint32	tcps_rcvackpack;	/* rcvd ack packets */
  uint32	tcps_rcvackbyte;	/* bytes acked by rcvd acks */
  uint32	tcps_rcvwinupd;		/* rcvd window update packets */
  uint32	tcps_pawsdrop;		/* segments dropped due to PAWS */
  uint32	tcps_predack;		/* times hdr predict ok for acks */
  uint32	tcps_preddat;		/* times hdr predict ok for data pkts */
  uint32	tcps_pcbcachemiss;
  uint32	tcps_cachedrtt;		/* times cached RTT in route updated */
  uint32	tcps_cachedrttvar;	/* times cached rttvar updated */
  uint32	tcps_cachedssthresh; /* times cached ssthresh updated */
  uint32	tcps_usedrtt;		/* times RTT initialized from route */
  uint32	tcps_usedrttvar;	/* times RTTVAR initialized from rt */
  uint32	tcps_usedssthresh;	/* times ssthresh initialized from rt*/
  uint32	tcps_persistdrop;	/* timeout in persist state */
  uint32	tcps_badsyn;		/* bogus SYN, e.g. premature ACK */
  uint32	tcps_mturesent;		/* resends due to MTU discovery */
  uint32	tcps_listendrop;	/* listen queue overflows */
} ipm_tcpstat_t;

typedef struct ipm_udpstat {
  uint32	udps_card;			/* card number */

  /* udpstat structure in udp_var.h */
  uint32	udps_ipackets;		/* total input packets */
  uint32	udps_hdrops;		/* packet shorter than header */
  uint32	udps_badsum;		/* checksum error */
  uint32	udps_badlen;		/* data length larger than packet */
  uint32	udps_noport;		/* no socket on port */
  uint32	udps_noportbcast;	/* of above, arrived as broadcast */
  uint32	udps_fullsock;		/* not delivered, input socket full */
  uint32	udps_pcbcachemiss;	/* input packets missing pcb cache */
  uint32	udps_pcbhashmiss;	/* input packets not for hashed pcb */
  uint32	udps_opackets;		/* total output packets */
  uint32	udps_fastout;		/* output packets on fast path */
  uint32    udps_fastin;		/* input packets on fast path */
} ipm_udpstat_t; 

typedef struct ipm_conn {		/* tcp/udp connection table */
  uint32		conn_lAddr;
  uint32		conn_rAddr;
  ushort16		conn_lPort;
  ushort16		conn_rPort;
  uint32		conn_state;
} ipm_conn_t;

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ValidateIpmMem	
 *
 * Description: validate ipm memory.
 *
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/17/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_ValidateIpmMemCluster(char	*node);
#define ipmd_ValidateIpmMem()	ipmd_ValidateIpmMemCluster("local")	

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetIfStatsCluster
 *
 * Description: get stats associated with a device. 
 *				if device is NULL, cpy all ifstats in
 *				buff in consecutive ipm_ifstat_t.
 *
 * parameters:
 * char*		ifname,	// device name, null if stats from all devices	
 * ipm_ifstat_t	*ifstat,// buf to cpy to 
 * int			bufsize,// buf size 
 * int			*count,	// num. of record copied
 * int			*needed,// bytes needed 
 * char			*node	// node of interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetIfStatsCluster(char*			ifname,	
					   ipm_ifstat_t		*ifstat, 
					   int				bufsize, 
					   int				*count,	
					   int				*needed,
					   char				*node);
#define ipmd_GetIfStats(_ifname, _ifstat, _bufsize, _count, _needed)	\
ipmd_GetIfStatsCluster(_ifname, _ifstat, _bufsize, _count, _needed, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetIpStatsCluster
 *
 * Description: get ip stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_ipstat_t.
 *
 * parameters:
 * int			card,	// which card, -1 if stats from all cards	
 * ipm_ipstat_t	*ipstat,// buf to cpy to 
 * int			bufsize,// buf size 
 * int			*count,	// num. of record copied
 * int			*needed,// bytes needed 
 * char			*node	// node of interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetIpStatsCluster(int32		card,	
					   ipm_ipstat_t	*ipstat, 
					   int			bufsize,
					   int			*count,	
					   int			*needed,
					   char			*node);
#define ipmd_GetIpStats(_card, _ipstat, _bufsize, _count, _needed)	\
ipmd_GetIpStatsCluster(_card, _ipstat, _bufsize, _count, _needed, "local")


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetIcmpStatsCluster
 *
 * Description: get icmp stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_icpstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetIcmpStatsCluster(int32			card,	
						 ipm_icpstat_t	*icmpstat, 
						 int			bufsize,
						 int			*count,	
						 int			*needed,	
						 char			*node);
#define ipmd_GetIcmpStats(_card, _icmpstat, _bufsize, _count, _needed)	\
ipmd_GetIcmpStatsCluster(_card, _icmpstat, _bufsize, _count, _needed, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTcpStatsCluster
 *
 * Description: get Tcp stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_tcpstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTcpStatsCluster(int32			card,	
						ipm_tcpstat_t	*tcpstat, 
						int				bufsize,
						int				*count,	
						int				*needed,
						char			*node);
#define ipmd_GetTcpStats(_card, _tcpstat, _bufsize, _count, _needed)	\
ipmd_GetTcpStatsCluster(_card, _tcpstat, _bufsize, _count, _needed, "local")


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetUdpStatsCluster
 *
 * Description: get Udp stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_udpstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetUdpStatsCluster(int32			card,	
						ipm_udpstat_t	*udpstat, 
						int				bufsize,
						int				*count,
						int				*needed,
						char			*node);
#define ipmd_GetUdpStats(_card, _udpstat, _bufsize, _count, _needed)	\
ipmd_GetUdpStatsCluster(_card, _udpstat, _bufsize, _count, _needed, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTcpConnTableCluster
 *
 * Description: get system wide TCP connection table.
 *
 * 
 * parameters:
 * ipm_conn_t	*tcpconn,	// buffer to copy to
 * int			bufsize,	// buffer size
 * int			index,		// where to start, <=0 start from begining 
 * int			incount,	// how many wanted,<0 to end if list 
 * int			*count,		// number of record 
 * int			*needed,	// bytes needed
 * char			*node		// node of interest
 * 
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTcpConnTableCluster(ipm_conn_t		*tcpconn,
							int				bufsize,
							int				index, 
							int				incount,
							int				*count,
							int				*needed,
							char			*node);
#define ipmd_GetTcpConnTable(_tcp_conn, _bufsize, _index, _incount, _count, _needed)	\
ipmd_GetTcpConnTableCluster(_tcp_conn, _bufsize, _index, _incount, _count, _needed, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetUdpConnTableCluster
 *
 * Description: get system wide UDP connection table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetUdpConnTableCluster(ipm_conn_t	*udpconn, 
							int			bufsize,
							int			index,
							int			incount,
							int			*count,
							int			*needed,
							char		*node);
#define ipmd_GetUdpConnTable(_udpconn, _bufsize, _index, _incount, _count, _needed)	\
ipmd_GetUdpConnTableCluster(_udpconn, _bufsize, _index, _incount, _count, _needed, "local")


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTcpConnCluster
 *
 * Description: find a TCP connection in record.
 *
 * parameters:
 * ipm_conn_t	*connp,	 // used to search
 * ipm_conn_t	*result, // record copied here if found
 * char			*node	// node if interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_ERR, IPM_GET_OK */
ipmd_GetTcpConnCluster(ipm_conn_t	*connp,
					   ipm_conn_t	*result,
					   char			*node);
#define ipmd_GetTcpConn(_connp, _result)	\
ipmd_GetTcpConnCluster(_connp, _result, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetUdpConnCluster
 *
 * Description: find a UDP connection in record.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR*/
ipmd_GetUdpConnCluster(ipm_conn_t	*connp,
					   ipm_conn_t	*result,
					   char			*node);
#define ipmd_GetUdpConn(_connp, _result)	\
ipmd_GetUdpConnCluster(_connp, _result, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTcpNxtConnCluster
 *
 * Description: find next TCP connection in record.
 *
 * parameters:
 * ipm_conn_t	*connp,	 // used to search, NULL returns first record
 * ipm_conn_t	*nextp, // record copied here if found
 * char			*node	// node if interest
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR */
ipmd_GetTcpNxtConnCluster(ipm_conn_t	*connp, 
						  ipm_conn_t	*nextp,
						  char			*node);
#define ipmd_GetTcpNxtConn(_connp, _nextp)	\
ipmd_GetTcpNxtConnCluster(_connp, _nextp, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetUdpNxtConnCluster
 *
 * Description: find next UDP connection in record.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR */
ipmd_GetUdpNxtConnCluster(ipm_conn_t	*connp,
						  ipm_conn_t	*nextp,
						  char			*node);
#define ipmd_GetUdpNxtConn(_connp, _nextp)	\
ipmd_GetUdpNxtConnCluster(_connp, _nextp, "local")	

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTcpConnSSCCluster
 *
 * Description: get tcp conn table on ssc.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTcpConnSSCCluster(ipm_conn_t	*tcpconn,
						  int			bufsize,
						  int			*count,
						  int			*needed,
						  char			*node);
#define ipmd_GetTcpConnSSC(_tcpconn, _bufsize, _count, _needed)	\
ipmd_GetTcpConnSSCCluster(_tcpconn, _bufsize, _count, _needed, "local")

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetUdpConnSSCCluster
 *
 * Description: get udp conn table on ssc.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/11/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetUdpConnSSCCluster(ipm_conn_t	*udpconn,
						  int			bufsize,
						  int			*count,	
						  int			*needed,
						  char			*node);
#define ipmd_GetUdpConnSSC(_udpconn, _bufsize, _count, _needed)	\
ipmd_GetUdpConnSSCCluster(_udpconn, _bufsize, _count, _needed, "local")


/*-----------------------------------------------------------------
 *
 * Name: ipmd_EnableIntfOnCardCluster
 *
 * Description: enable all interface on a card
 *
 * Created by:	Su Wang
 *
 * Date Created: 03/20/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_EnableIntfOnCardCluster(int	card,
							 char	*node);

/*-----------------------------------------------------------------
 *
 * Name: ipmd_disableIntfOnCardCluster
 *
 * Description: disable all interface on a card
 *
 * Created by:	Su Wang
 *
 * Date Created: 03/20/2002
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DisableIntfOnCardCluster(int	card,
							  char	*node);


#endif /* _IPMD_LIB_H */
