/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc.c
 *
 * RCS:		$Id: ispfc.c,v 1.34 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver initialization and bulk of operational
 *              code.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/5/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_nvram.h"
#include "ispfc_mbox.h"
#include "ispfc_scsi.h"
#include "ispfc_scsi_tmo.h"
#include "ispfc_ct.h"
#include "ispfc_ioctl.h"
#include "ispfc_link.h"
#include "ispfc_device.h"
#include "ispfc_aen_handler.h"
#include "ispfc.h"
#include "ispfc_td.h"
#include "ispfc_iocb.h"
#include "isp2300_fw.h"
#include <exports/cm-api.h>
#include <exports/env-api.h>

int32    ispfc_completion_handler(ispfc_cb_t *, uint32);
#if defined(ISPFC_SIM)
void	 ispfc_process_simulated_completions(ispfc_cb_t *);
#endif /* defined(ISPFC_SIM) */

void *	  ispfc_global_memory = NULL;
uint32	  ispfc_global_memory_len = 0;

/*******************************************************************************
 * ispfc_early_init -
 *
 *   Driver early intialization entry point.
 ******************************************************************************/
void 
ispfc_early_init()
{
	VERIFY(ISPFC_REQUEST_QUEUE_SIZE < 65536);
	VERIFY(ISPFC_RESPONSE_QUEUE_SIZE < 65536);
	VERIFY(sizeof(ispfc_init_fw_cb_t) <= 2 * sizeof(ispfc_command_entry_t));

	/* Allocate request and response rings -- must be 64 byte aligned. */
	ispfc_global_memory_len = sizeof(ispfc_command_entry_t) * 
		(ISPFC_REQUEST_QUEUE_SIZE + ISPFC_RESPONSE_QUEUE_SIZE + 3);
	ispfc_global_memory = HAL_ALIGN_TO(eee_allocateBlock(ispfc_global_memory_len), sizeof(ispfc_command_entry_t));
	VERIFY(ispfc_global_memory);
	CACHE_FLUSH(ispfc_global_memory, 
		    (address_t)ispfc_global_memory + ispfc_global_memory_len);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	ispfc_global_memory = (void *)cached_to_uncached((address_t)ispfc_global_memory);
#else
	(address_t)ispfc_global_memory |= 0x20000000L; /* Turn on uncached access */
#endif
#endif
	bzero(ispfc_global_memory, ispfc_global_memory_len);
}

/*******************************************************************************
 * ispfc_pci_init -
 *
 *   PCI intialization entry point. bus_id/dev_id == -1 is a wildcard; hunt
 *   for the ISP 2300 device. There should only be one.
 ******************************************************************************/
int 
ispfc_pci_init(int bus_id, int dev_id)
{
	int32			i, j;
	int32			found = 0;
	uint32			value;
	hal_dispatch_entry_t *	hal_dte;

#define MAX_PCI_BUSSES		2
#define MAX_DEVS_PER_PCI_BUS	9

#if !defined(ISPFC_SIM)
	for (i = ((bus_id == -1) ? 0 : bus_id); i < ((bus_id == -1) ? MAX_PCI_BUSSES : bus_id+1); ++i) {
		for (j = ((dev_id == -1) ? 0 : dev_id); j < ((dev_id == -1) ? MAX_DEVS_PER_PCI_BUS : dev_id+1); ++j) {		
			/* Verify it's a 2300 */
			if (pci_configRead16(i, j, ISPFC_REG_PCI_VENDOR_ID) == 0x1077 &&
			    pci_configRead16(i, j, ISPFC_REG_PCI_DEVICE_ID) == 0x2300) {
				FC_DBG(0, ("ispfc_pci_init: A QLogic 2300 found at bus_id %d, dev_id %d\n",
					   i, j));
				bus_id = i;
				dev_id = j;
				++found;
				goto found;
			}
		}
	}
 found:
 	if (!found)
		return(NFX_ERR);
	
	/* 
	 * Clear bit 0 of ROM Base Address register to take RISC out
	 * of reset.  
	 */
	value = pci_configRead32(bus_id, dev_id, ISPFC_REG_PCI_ROM_BASE);
	value &= 0xFFFFFFFE;
	pci_configWrite32(bus_id, dev_id, ISPFC_REG_PCI_ROM_BASE, value);

	/* Set the Cmd register -- program for memory mapped register access */
	value = CMD_SYS_ERR_ENABLE | 
		CMD_PARITY_RESPONSE | 
		/* CMD_MEM_WRITE_INVALID | */
		CMD_BUS_MASTER |
		CMD_MEM_ADDR_ENABLE;
	pci_configWrite16(bus_id, dev_id, ISPFC_REG_PCI_COMMAND, value);

	/* Set the Cache Line Size register to 32 bytes (8 4-byte words) */
	pci_configWrite8(bus_id, dev_id, ISPFC_REG_PCI_CACHE_LINE, CACHELINE_SIZE/4);

	/* Set the Latency Timer register to 0x80 */
	pci_configWrite8(bus_id, dev_id, ISPFC_REG_PCI_LATENCY, 0x80);

	/* Set the memory base address registers */
	pci_configWrite32(bus_id, dev_id, ISPFC_REG_PCI_MEM_BASE, ISPFC_BASE_ADDR);

	FC_DBG(0, ("ispfc: ispfc_pci_init: Config space dump of bus_id %d, dev_id %d: \n", bus_id, dev_id));
        FC_DBG(0, ("       Vendor......0x%x\n", pci_configRead16(bus_id, dev_id, ISPFC_REG_PCI_VENDOR_ID)));
        FC_DBG(0, ("       Device......0x%x\n", pci_configRead16(bus_id, dev_id, ISPFC_REG_PCI_DEVICE_ID)));
        FC_DBG(0, ("       Command.....0x%x\n", pci_configRead16(bus_id, dev_id, ISPFC_REG_PCI_COMMAND)));
        FC_DBG(0, ("       Status......0x%x\n", pci_configRead16(bus_id, dev_id, ISPFC_REG_PCI_STATUS)));
        FC_DBG(0, ("       Revision....0x%x\n", pci_configRead8(bus_id, dev_id, ISPFC_REG_PCI_REVISION)));
        FC_DBG(0, ("       Class.......0x%02x%04x\n",
		     pci_configRead8(bus_id, dev_id, ISPFC_REG_PCI_CLASS),
		     pci_configRead16(bus_id, dev_id, ISPFC_REG_PCI_CLASS+1)));
        FC_DBG(0, ("       CacheLine...0x%x\n", pci_configRead8(bus_id, dev_id, ISPFC_REG_PCI_CACHE_LINE)));
        FC_DBG(0, ("       Latency.....0x%x\n", pci_configRead8(bus_id, dev_id, ISPFC_REG_PCI_LATENCY)));
        FC_DBG(0, ("       HeaderType..0x%x\n", pci_configRead16(bus_id, dev_id, ISPFC_REG_PCI_HEADER_TYPE)));
        FC_DBG(0, ("       I/O Base....0x%x\n", pci_configRead32(bus_id, dev_id, ISPFC_REG_PCI_IO_BASE)));
        FC_DBG(0, ("       Mem Base....0x%x\n", pci_configRead32(bus_id, dev_id, ISPFC_REG_PCI_MEM_BASE)));
        FC_DBG(0, ("       ROM Base....0x%x\n", pci_configRead32(bus_id, dev_id, ISPFC_REG_PCI_ROM_BASE)));
#endif /* !defined(ISPFC_SIM) */

	/* Create a HAL dispatch table entry and add dev/bus ID */
	hal_dte = hal_dt_entry_new();
	hal_dte->hal_bus_id = bus_id;
	hal_dte->hal_dev_id = dev_id;
	hal_dte->hal_valid++;

	return(NFX_OK);
} 

/*******************************************************************************
 * ispfc_init -
 *
 *   Driver intialization entry point.
 ******************************************************************************/
void 
ispfc_init()
{
	ispfc_cb_t *		cb;
	hal_dispatch_entry_t *  hal_dte = hal_dt_entry_get();
	int			i;
	int			rc;

	FC_DBG(0, ("ispfc: ispfc_init: \n"));
	
	if (hal_dte == NULL) {
		FC_DBG(0, ("ispfc: ispfc_init -- no QLogic 2300 found.\n"));
		return;
	}
		
	/* Verify pointer length */
	VERIFY(sizeof(void *) == sizeof(ps_uint_t));

	/* Allocate adapter control block. */
	cb = (ispfc_cb_t *)
		HAL_ALIGN_TO(eee_ramAlloc(sizeof(ispfc_cb_t) + sizeof(uint64)), sizeof(uint64));
	VERIFY(cb);
	bzero(cb, sizeof(ispfc_cb_t));

	cb->request = (ispfc_command_entry_t *)ispfc_global_memory;
	cb->response = (ispfc_status_entry_t *)(cb->request + ISPFC_REQUEST_QUEUE_SIZE);
	cb->fw_init_cb = (ispfc_init_fw_cb_t *)(cb->response + ISPFC_RESPONSE_QUEUE_SIZE);
	FC_DBG(0, ("ispfc: cb: %p\n", cb));
	FC_DBG(0, ("ispfc: request ring: %p to %p (%d/%d)\n", 
			   &cb->request[0], 
			   &cb->request[ISPFC_REQUEST_QUEUE_SIZE-1],
			   ISPFC_REQUEST_QUEUE_SIZE,
			   sizeof(ispfc_command_entry_t) * ISPFC_REQUEST_QUEUE_SIZE));
	FC_DBG(0, ("ispfc: response ring: %p to %p (%d/%d)\n", 
			   &cb->response[0], 
			   &cb->response[ISPFC_RESPONSE_QUEUE_SIZE-1],
			   ISPFC_RESPONSE_QUEUE_SIZE,
			   sizeof(ispfc_command_entry_t) * ISPFC_RESPONSE_QUEUE_SIZE));
	FC_DBG(0, ("ispfc: fw_init_cb: %p(%d)\n", 
			   cb->fw_init_cb,
			   sizeof(ispfc_init_fw_cb_t)));

	cb->adap_init_state = eee_ramAlloc(sizeof(ispfc_adap_init_state_t));
	cb->conn_init_state = eee_ramAlloc(sizeof(ispfc_conn_init_state_t));
	cb->link_init_state = eee_ramAlloc(sizeof(ispfc_link_init_state_t));
	cb->dev_disc_state = eee_ramAlloc(sizeof(ispfc_dev_discovery_state_t));
	cb->mbox_sm = eee_ramAlloc(sizeof(mbox_cmd_sm_t));
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	cb->dev_disc_state = (void *)cached_to_uncached((address_t)cb->dev_disc_state);
#else
	(ps_uint_t)cb->dev_disc_state |= 0x20000000L; /* Turn on uncached access */
#endif
#endif
	bzero(cb->adap_init_state, sizeof(ispfc_adap_init_state_t));
	bzero(cb->conn_init_state, sizeof(ispfc_conn_init_state_t));
	bzero(cb->link_init_state, sizeof(ispfc_link_init_state_t));
	bzero(cb->dev_disc_state, sizeof(ispfc_dev_discovery_state_t));
	bzero(cb->mbox_sm, sizeof(mbox_cmd_sm_t));

	for (i = 0; i < sizeof(cb->aen_handler)/sizeof(void *); ++i) {
		cb->aen_handler[i] = eee_ramAlloc(sizeof(ispfc_aen_handler_t));
		bzero(cb->aen_handler[i], sizeof(ispfc_aen_handler_t));
	}

#if defined(EVAL_BOARD) || defined (AGILE_HARDWARE)
	/* Point at register memory base address */
	cb->reg_mem_base = (void *)(ISPFC_BASE_ADDR);
#ifdef EVAL_BOARD
	cb->reg_mem_base = (void *)((address_t)cb->reg_mem_base | 0xA0000000);
#endif /* EVAL_BOARD */
#endif /* defined(EVAL_BOARD) || defined (AGILE_HARDWARE) */

	/* Point at firmware */
	cb->firmware_version = firmware_version;
	cb->risc_code_addr = risc_code_addr01;
	cb->risc_code_length = risc_code_length01;
	cb->risc_code = risc_code01;

#if defined(BIG_ENDIAN_EVAL_BOARD)
	/* Half-word swap firmware for Eval board. */
	cb->risc_code = (ushort16 *)HAL_ALIGN_TO(eee_ramAlloc(risc_code_length01 * 2 + CACHELINE_SIZE), CACHELINE_SIZE);
#ifdef FC_USE_NEW_MEM_MAP
	cb->risc_code = (ushort16 *)cached_to_uncached((address_t)cb->risc_code);
#else
	(ps_uint_t)cb->risc_code |= 0x20000000L; /* Turn on uncached access */
#endif
	{
		ushort16 *src, *dest;
		src = (ushort16 *)risc_code01;
		dest = (ushort16 *)cb->risc_code;
		for (i = 0; i < risc_code_length01; ++i) {
			*dest = *src;
			SWAP2(dest);
			++dest; ++src;
		}
	}
#endif

	/* Initialize FW init control block */
	cb->fw_init_cb->version = 1;
	cb->fw_init_cb->fw_options = IFW_FO_EXTENDED_CONTROL_BLOCK |
		IFW_FO_NAME_OPTION |
		IFW_FO_ENABLE_ADISC | 
		IFW_FO_ASSERT_LIHA | 
		IFW_FO_ASSERT_LIPA | 
		IFW_FO_FORCE_PLOGI_ON_LIP;
	cb->fw_init_cb->max_frame_len = 2048;
	cb->fw_init_cb->max_IOCB_resources = 1; /* Not currently used */
	cb->fw_init_cb->execution_throttle = ISPFC_MAX_SCSI_TAGS * ISPFC_MAX_LUNS_PER_TARGET;
	cb->fw_init_cb->retry_count = 0;
	cb->fw_init_cb->retry_delay = 0;
	cb->fw_init_cb->hard_address = 120;
	cb->fw_init_cb->login_time_out = ISPFC_RA_TOV_DEFAULT/MSEC_PER_SEC;
	cb->fw_init_cb->request_out_pointer = 0;
	cb->fw_init_cb->response_in_pointer = 0;
	cb->fw_init_cb->request_queue_len = ISPFC_REQUEST_QUEUE_SIZE;
	cb->fw_init_cb->response_queue_len = ISPFC_RESPONSE_QUEUE_SIZE;
	cb->fw_init_cb->request_queue_addr[1] = 0;
	cb->fw_init_cb->request_queue_addr[0] = VA_TO_IO32(cb->request);
	cb->fw_init_cb->response_queue_addr[1] = 0;
	cb->fw_init_cb->response_queue_addr[0] = VA_TO_IO32(cb->response);
	cb->fw_init_cb->additional_fw_options = IFW_AFO_LOOP_ONLY;
	cb->fw_init_cb->special_options = IFW_SO_AUTO_NEGOTIATE_SPEED | IFW_SO_50_OHM;

	/* Read nodename/portname from NVRAM and set adapter_wwn */
	ispfc_nvram_config(cb);
	WWN_SET(&cb->adapter_wwn, cb->fw_init_cb->portname);

#if defined(BIG_ENDIAN_EVAL_BOARD)
	/* Swap certain fields of firmware init cb for Eval board. */
	{
		SWAP2(&cb->fw_init_cb->fw_options);
		SWAP2(&cb->fw_init_cb->max_frame_len);
		SWAP2(&cb->fw_init_cb->max_IOCB_resources);
		SWAP2(&cb->fw_init_cb->hard_address);
		SWAP2(&cb->fw_init_cb->execution_throttle);
		SWAP2(&cb->fw_init_cb->request_queue_len);
		SWAP2(&cb->fw_init_cb->response_queue_len);
		SWAP4(&cb->fw_init_cb->request_queue_addr[0]);
		SWAP4(&cb->fw_init_cb->response_queue_addr[0]);
		SWAP2(&cb->fw_init_cb->additional_fw_options);
		SWAP2(&cb->fw_init_cb->special_options);
	}
#endif

	/* Set default timeout values */
	cb->lip_timeout = LOOP_UP_TIMEOUT;
	cb->mbox_timeout = ISPFC_MBOX_TIMEOUT_DEFAULT;
	cb->switch_ra_tov = ISPFC_RA_TOV_DEFAULT;

	/* Initialize handle map */
	ispfc_device_handle_init(cb);

	/* Add switch table entries to HAL switch table entry */
	hal_dte->hal_scsi_send = ispfc_scsi_send;
	hal_dte->hal_scsi_task_mgmt_send = ispfc_scsi_task_mgmt_send;
	hal_dte->hal_ct_send = ispfc_ct_send;
	hal_dte->hal_ioctl = ispfc_ioctl;
	hal_dte->hal_handle = cb;

	sprintf(cb->debug_name, "ispfc:sp%d.%d", nfxMySlotId, nfxMyCPUId);

	/* Reset chip */
	ispfc_soft_reset_adapter(cb);

	/* 
	 * Initialize link state machine and register
	 * ispfc_link_management_handler thread 
	 */
	if (ispfc_is_autolinkenable_on())
		cb->link_state = ISPFC_CLS_STATE_RESET_REQUESTED;
	else {
		cb->link_state = ISPFC_CLS_STATE_OFFLINE;
		cb->link_state |= ISPFC_CLF_FC_OFFLINE_LATCH;
		printf("************************************************************\n");
		printf("WARNING: Fibrechannel link sp%d.%d is OFFLINE/DISABLED.     \n", nfxMySlotId, nfxMyCPUId);
		printf("  To reenable, use the nfxsh command 'port enable sp%d.%d'.\n", nfxMySlotId, nfxMyCPUId);
		printf("************************************************************\n");
	}
	rc = eee_registerPollingFunc(EEE_LOW_PRIORITY, (void *)ispfc_link_management_handler, cb);
	VERIFY(rc >= 0);

	/*
         * Register ispfc_device_management_handler thread
         */
        rc = eee_registerPollingFunc(EEE_HIGH_PRIORITY, (void *)ispfc_device_management_handler, cb);
        VERIFY(rc >= 0);

	/*
	 * Register ispfc_ct_desc_start_handler thread
	 */
        rc = eee_registerPollingFunc(EEE_LOW_PRIORITY, (void *)ispfc_ct_desc_start_handler, cb);
        VERIFY(rc >= 0);

	/*
	 * Register ispfc_scsi_task_management_start_handler thread
	 */

	/*
	 * Register ispfc_completion_handler thread
	 */
	rc = eee_registerPollingFunc(EEE_HIGH_PRIORITY, (void *)ispfc_completion_handler, cb);
	VERIFY(rc >= 0);

	/*
	 * Register ispfc_timeout_handler thread -- it will be called every TIMEOUT_HANDLER_PERIOD mSecs.
	 */
	cb->ispfc_timeout_handler_timer_handle = eee_timerCreate(TIMEOUT_HANDLER_PERIOD, EEE_TIMER_REPEAT);
	VERIFY(cb->ispfc_timeout_handler_timer_handle >= 0);
	rc = eee_timerStart(cb->ispfc_timeout_handler_timer_handle, (void *)ispfc_timeout_handler,
			    cb, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);

	/*
	 * Initialize topology discovery state machine.
	 */
	fctd_init(ISPFC_TD_POLL_PERIOD);
}

/*******************************************************************************
 * ispfc_completion_handler -
 *
 *   The "backend interrupt handler".
 ******************************************************************************/
int32 
ispfc_completion_handler(ispfc_cb_t *cb, uint32 timer_ref)
{
	ushort16 r_to_h_hi, r_to_h_low;

#if defined(ISPFC_SIM)
	ispfc_process_simulated_completions(cb);
#else
	/*
	 * Check if there's an interrupt pending.
	 */
	r_to_h_low = ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_LOW);
	if (!(r_to_h_low & R_TO_H_LOW_RISC_INT)) {
		/*
		 * Are there responses left unprocessed since last
		 * poll pass? If so, do more now.
		 */
		if (cb->response_in != cb->response_out)
			ispfc_process_iocb_completions(cb, ISPFC_RESPONSE_THROTTLE);
		return(NFX_OK);
	}
	r_to_h_hi = ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_HI);

	/*
	 * If link is offline, don't bother going any further
	 */
	if (LINK_STATE(cb) == ISPFC_CLS_STATE_OFFLINE) {
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
		goto done;
	}

	switch (r_to_h_low & R_TO_H_LOW_STATUS_MASK) {
	case R_TO_H_LOW_STATUS_RESPONSE_UPDATED:	/* IOCB completion? */
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
#if 0
		cb->response_in = ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_IN);
		if (cb->response_in != r_to_h_hi)
			P("R_IO 0x%x, RTOH_HI 0x%x\n", cb->response_in, r_to_h_hi);
#else
		cb->response_in = r_to_h_hi;
#endif
		/* Only process IOCB completions if it "makes sense" */
		if (cb->response_in != cb->response_out) 
			ispfc_process_iocb_completions(cb, ISPFC_RESPONSE_THROTTLE);
		break;

	case R_TO_H_LOW_STATUS_ASYNC_EVENT:		/* AEN signalled ? */
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
		ISPFC_REG_WRITE(cb, ISPFC_REG_SEMAPHORE, 0);
		ispfc_aen_handler(cb, r_to_h_hi);
		break;

	case R_TO_H_LOW_STATUS_ROM_SUCCESSFULL:		/* Mailbox command completion */
	case R_TO_H_LOW_STATUS_ROM_FAILED:
	case R_TO_H_LOW_STATUS_MBOX_SUCCESSFULL:
	case R_TO_H_LOW_STATUS_MBOX_FAILED:
		ispfc_mbox_done(cb);
		break;

	default:
	  /* 
		VERIFY(0);
	  */
	  ;
	}
#endif /* defined(ISPFC_SIM) */

 done:

	return(NFX_OK);
}

#if defined(ISPFC_SIM)
void ispfc_process_simulated_completions(ispfc_cb_t *cb)
{
	ispfc_dev_t *	dev;
	scsi_desc_t *	sd;
	ct_desc_t *	ctd;
	
	/*
	 * Process SCSI descriptor completions 
	 */
	FOR_ALL_DEVS({
		again_scsi:
		if (sd = dev->scsi_desc_busy) {
			scsi_desc_t *	next;
			eeeRTC_t now;
			do {
				next = sd->next;
				/*     
				 * Simulated completion time is >= 10mS.
				 */
				eee_rtc(&now);
				if (eee_rtcCompare(&now, &sd->sim_response_timestamp) >= 0) {
					ispfc_dequeue_desc_off_dev_busy_queue(cb, dev, sd);
					ispfc_dispose_of_scsi_desc(cb, dev, sd, 0);
					goto again_scsi;
				}
				sd = next;
			} while (sd && sd != dev->scsi_desc_busy);
		}
	});

	/*
	 * Process CT descriptor completions 
	 */
 again_ct:
	if (ctd = cb->ct_desc_busy) {
		ct_desc_t *	next;
		eeeRTC_t delta;
		do {
			next = ctd->next;
			/*     
			 * Simulated completion time is >= 20000uS.
			 */
			eee_rtcDuration(&ctd->first_issued_timestamp, &delta);
			if (eee_rtcToUsec(&delta) >= 20000LL)) {
				ispfc_dequeue_desc_off_ct_busy_queue(cb, ctd);
				ispfc_dispose_of_ct_desc(cb, ctd, 0);
				goto again_ct;
			}
			ctd = next;
		} while (ctd && ctd != cb->ct_desc_busy);
	}
}
#endif /* defined(ISPFC_SIM) */

int32
ispfc_is_autolinkenable_on(void)
{
	char8  str[ENV_MAX_STR_LEN];

	if (seepenv_get(ENV_KEY_AUTOLINKENABLE, str) == NFX_OK &&
	    strcmp(str, "on") == 0)
		return(1);
	else
		return(0);
}

void
ispfc_set_autolinkenable_on(void)
{
	if (!ispfc_is_autolinkenable_on())
		seepenv_set(ENV_KEY_AUTOLINKENABLE, "on");
}

void
ispfc_set_autolinkenable_off(void)
{
	if (ispfc_is_autolinkenable_on())
		seepenv_set(ENV_KEY_AUTOLINKENABLE, "off");
}
