/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_channels.h
 *
 * RCS:		$Id: ispfc_channels.c,v 1.4 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL channel definitions.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/6/01
 *
 *------------------------------------------------------------------------------
 */

#include <nfx-incl.h>
#include "ispfc_private.h"
#include "ispfc_channels.h"
#include "ispfc_hash.h"
#include "ispfc_extern.h"

#define WWN_TO_HASH_KEY(_wwn_) \
	( (uint32) \
	  ((((uint32)(_wwn_).u.octets[0]) << 24) | \
	   (((uint32)(_wwn_).u.octets[1]) << 16) | \
	   (((uint32)(_wwn_).u.octets[2]) <<  8) | \
	   (((uint32)(_wwn_).u.octets[3]))) ^ \
	  (uint32) \
	  ((((uint32)(_wwn_).u.octets[4]) << 24) | \
	   (((uint32)(_wwn_).u.octets[5]) << 16) | \
	   (((uint32)(_wwn_).u.octets[6]) <<  8) | \
	   (((uint32)(_wwn_).u.octets[7]))))

typedef struct priv_channel_name {
	channel_name_t			channel;
	int				channel_ref_count;
	struct priv_channel_name *	next;
	struct priv_channel_name *	prev;
} priv_channel_name_t;

static ispfc_hash_t *		channels_hash = NULL;
static priv_channel_name_t *	channels_root = NULL;
static int			channels_count = 0;

/*
 * ispfc_channel_add - add a channel to the global channel list
 */
int
ispfc_channel_add(channel_name_t *ch)
{
	int rc;
	priv_channel_name_t	*pch;
	ispfc_hash_key_t	key;

	if (channels_hash == NULL)
		channels_hash = ispfc_hash_table_alloc(ISPFC_HASH_TYPE1);
	else {
		key = WWN_TO_HASH_KEY(ch->channel_wwn);
		pch = ispfc_hash_table_obj_find(channels_hash, key);
		if (pch != NULL) {
			ASSERT(strcmp(ch->channel_name, pch->channel.channel_name) == 0);
			++pch->channel_ref_count;
			return(NFX_OK);
		}
	}

	pch = (priv_channel_name_t *)eee_ramAlloc(sizeof(priv_channel_name_t));
	bzero(pch, sizeof(priv_channel_name_t));
	memcpy(&pch->channel, ch, sizeof(channel_name_t));
	pch->channel_ref_count = 1;

	/*
	 * Insert the channel into the hash table
	 */
	key = WWN_TO_HASH_KEY(ch->channel_wwn);	
	rc = ispfc_hash_table_obj_insert(channels_hash, key, pch);
	VERIFY(rc == NFX_OK);
	
	/*
	 * ... and into the linear linked list.
	 */
	if (channels_root == NULL) {
		pch->next = pch;
		pch->prev = pch;
		channels_root = pch;
	} 
	else {
		pch->next = channels_root;
		pch->prev = channels_root->prev;
		channels_root->prev->next = pch;
		channels_root->prev = pch;
	}
	channels_count++;

	return(NFX_OK);
}

/*
 * ispfc_channel_remove - remove a channel to the global channel list, or at least
 *                        lower the reference count. Last one out deletes the
 *                        entry.
 */
int
ispfc_channel_remove(channel_name_t *ch)
{
	priv_channel_name_t     *pch;
	ispfc_hash_key_t	key;

	VERIFY(channels_hash);
	VERIFY(channels_root);
	VERIFY(channels_count);
	
	key = WWN_TO_HASH_KEY(ch->channel_wwn);
	pch = ispfc_hash_table_obj_find(channels_hash, key);
	VERIFY(pch);
	VERIFY(pch->channel_ref_count);
	if (--pch->channel_ref_count == 0) {
		channels_count--;

		/*
		 * Remove the channel from the hash table
		 */
		ispfc_hash_table_obj_delete(channels_hash, key, pch);

		/* 
		 * ... and from the linear linked list
		 */
		pch->next->prev = pch->prev;
		pch->prev->next = pch->next;

		if (channels_root == pch)
			channels_root = pch->next;
		
		if (pch->next == pch) {
			VERIFY(pch->prev == pch);
			VERIFY(channels_root == pch);
			VERIFY(channels_count == 0);
			channels_root = NULL;
		}
		pch->next = pch->prev = NULL;
	}

	eee_ramDealloc((void *)pch);

	return(NFX_OK);
}

/*
 * ispfc_channel_lookup - Translates from channel WWN to channel
 *                        pathname name
 */
int 
ispfc_channel_lookup(channel_name_t *ch)
{
	priv_channel_name_t     *pch;
	ispfc_hash_key_t	key;

	if (channels_hash == NULL)
		return(NFX_ERR);
	key = WWN_TO_HASH_KEY(ch->channel_wwn);
	pch = ispfc_hash_table_obj_find(channels_hash, key);
	if (pch) {
		VERIFY(pch->channel_ref_count);
		memcpy(ch, &pch->channel, sizeof(channel_name_t));
		return(NFX_OK);
	}
	else
		return(NFX_ERR);
}

/*
 * ispfc_channel_set_fl_port - Sets the FL port ID of an FL-port channel.
 */
int 
ispfc_channel_set_fl_port(channel_name_t *ch)
{
	priv_channel_name_t     *pch;
	ispfc_hash_key_t	key;

	VERIFY(channels_hash);
	key = WWN_TO_HASH_KEY(ch->channel_wwn);
	pch = ispfc_hash_table_obj_find(channels_hash, key);
	if (pch) {
		VERIFY(pch->channel.channel_interface_type == HAL_IF_FC_F_PORT ||
		       pch->channel.channel_interface_type == HAL_IF_FC_FL_PORT);
		pch->channel.u.fc_f_or_fc_fl.channel_fl_port = ch->u.fc_f_or_fc_fl.channel_fl_port;
		return(NFX_OK);
	}
	else
		return(NFX_ERR);
}

void 
ispfc_channels_dump(void)
{
	priv_channel_name_t     *pch;
	char			format[32];
	
	pch = channels_root;
	do {
		/* Dump channel WWN */
		P("%s", wwn_to_str(&pch->channel.channel_wwn, NULL));

		/* Dump channel name */
		sprintf(format, "%s%ds", "%", 16 /* MAX_CHANNEL_NAME_LEN */);
		P(format, pch->channel.channel_name);
		P(" ");

		/* Dump interface type */
		switch(pch->channel.channel_interface_type) {
		case HAL_IF_FC_L_PORT:
			P("L_PORT  "); break;
		case HAL_IF_FC_NL_PORT:
			P("NL_PORT "); break;
		case HAL_IF_FC_N_PORT:
			P("N_PORT  "); break;
		case HAL_IF_FC_F_PORT:
			P("F_PORT  "); break;
		case HAL_IF_FC_FL_PORT:
			P("FL_PORT "); break;
		default:
			VERIFY(0);
		}

		switch (pch->channel.channel_interface_type) {
		case HAL_IF_FC_L_PORT:
			P("bus id %2d, dev id %2d ", 
			  pch->channel.u.fc_l.channel_ha_bus_id,
			  pch->channel.u.fc_l.channel_ha_dev_id);
			break;

		case HAL_IF_FC_NL_PORT:
		case HAL_IF_FC_N_PORT:
			P("bus id %2d, dev id %2d, ", 
			  pch->channel.u.fc_n_or_fc_nl.channel_ha_bus_id,
			  pch->channel.u.fc_n_or_fc_nl.channel_ha_dev_id);
			P("peer WWN %s ", wwn_to_str(&pch->channel.u.fc_n_or_fc_nl.channel_peer_wwn, NULL));
			break;

		case HAL_IF_FC_F_PORT:
		case HAL_IF_FC_FL_PORT:
			P("switch phys port %2d, ", 
			  pch->channel.u.fc_f_or_fc_fl.channel_phys_port);
			P("switch WWN %s, ", 
			  wwn_to_str(&pch->channel.u.fc_f_or_fc_fl.channel_switch_wwn, NULL));
			P("FL_port 0x%06x ", pch->channel.u.fc_f_or_fc_fl.channel_fl_port);
			break;
		}
		P("\n");

		pch = pch->next;
	} while (pch != channels_root);
}
	
