/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_device.c
 *
 * RCS:		$Id: ispfc_device.c,v 1.54 2002/11/26 03:18:53 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver device code.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/6/01
 *
 *------------------------------------------------------------------------------
 */

#if defined(LINUX_TEST)
#include <stdlib.h>
#endif
#include "ispfc_private.h"
#include "ispfc_iocb.h"
#include "ispfc_link.h"
#include "ispfc_mbox.h"
#include "ispfc_ct.h"
#include "ispfc_device.h"
#include "ispfc_hash.h"
#include "ispfc_channels.h"
#include "ispfc_scsi.h"
#include "ispfc_scsi_tmo.h"

#if defined(ISPFC_SIM)
extern int	ispfc_fcadmin_disk_count;
extern int	ispfc_fcadmin_tape_count;
extern wwn_t	ispfc_write_protected_device;
#endif

void	 ispfc_reestablish_connection_1(ispfc_cb_t *, void *, int32);
void	 ispfc_reestablish_connection_ct_done(struct ct_desc *);
void	 ispfc_reestablish_connection_defer(ispfc_cb_t *, ispfc_dev_discovery_state_t *);
int32	 ispfc_reestablish_connection_deferred_timeout(int32, ispfc_cb_t *, ispfc_dev_discovery_state_t *);
void	 ispfc_reestablish_connection_done(ispfc_cb_t *, void *, int32);

void	 ispfc_dev_login_done_callout(ispfc_cb_t *, void *, int32);
void	 ispfc_dev_login_cleanup_done_callout(ispfc_cb_t *, void *, int32);
void	 ispfc_dev_logout_done_callout(ispfc_cb_t *, void *, int32);
int32	 ispfc_dev_deferred_login_timer_done_callout(int32, ispfc_cb_t *, ispfc_dev_t *);
int32	 ispfc_dev_paused_timer_done_callout(int32, ispfc_cb_t *, ispfc_dev_t *);
void	 ispfc_dev_get_port_database_done_callout(ispfc_cb_t *, void *, int32);

void     ispfc_cmd_timeout(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void     ispfc_dev_abort_taskset(ispfc_cb_t *, ispfc_dev_t *, byte); 
void     ispfc_dev_abort_taskset_0(ispfc_cb_t *, void *, int32);
void     ispfc_dev_abort_taskset_1(int32, ispfc_cb_t *, ispfc_dev_t *, scsi_task_mgmt_t *);

void     ispfc_dev_reset_device_0(ispfc_cb_t *, void *, int32);
void     ispfc_dev_reset_device_1(int32, ispfc_cb_t *, ispfc_dev_t *);

#define ISPFC_DEV_DISC_RETRY 3

/*******************************************************************************
 * ispfc_reestablish_connection -
 *
 *   Primary entry point into reestablish connection state machine. State 
 *   machine is initialized and the first state is entered.  
 ******************************************************************************/
void  
ispfc_reestablish_connection(ispfc_cb_t *cb, void (*reestablish_callback)(ispfc_cb_t *, void *, int32), void *reestablish_cookie)
{
	ispfc_dev_discovery_state_t	*st = cb->dev_disc_state;

	VERIFY(st);
	
	bzero(st, sizeof(ispfc_dev_discovery_state_t));

	if (cb->interface_type == HAL_IF_FC_L_PORT ||
	    cb->interface_type == HAL_IF_FC_N_PORT)
		st->dev_disc_state = 0;
	else 
		st->dev_disc_state = 2;
	st->dev_disc_retry = ISPFC_DEV_DISC_RETRY;
	st->dev_disc_callback = reestablish_callback;
	st->dev_disc_cookie = reestablish_cookie;

	ispfc_reestablish_connection_1(cb, st, NFX_OK);
}

/*******************************************************************************
 * ispfc_reestablish_connection_done -
 *
 *   Restart the reestablish connection state machine if a nested LIP/NSCN event
 *   arrived while we were processing the current state change notification.
 ******************************************************************************/
void
ispfc_reestablish_connection_done(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_discovery_state_t *st = cookie;

	++cb->reestablish_count;

	/* Cleanup state */
	if (cb->interface_type != HAL_IF_FC_L_PORT && 
	    cb->interface_type != HAL_IF_FC_N_PORT &&
	    st->u.fabric.ctd.e_desc) {
		eee_deallocateDesc(st->u.fabric.ctd.e_desc);
		st->u.fabric.ctd.e_desc = NULL;
	}

	if (status != REESTABLISH_STATUS_SUCCESS)
		(*st->dev_disc_callback)(cb, st->dev_disc_cookie, REESTABLISH_STATUS_FAILED);	
	else {
		/* Nested LIP/NSCN ? */
		if (cb->link_state & ISPFC_CLF_NESTED_LIP ||
		    cb->link_state & ISPFC_CLF_NESTED_NSCN) {
			cb->link_state &= ~(ISPFC_CLF_NESTED_LIP | ISPFC_CLF_NESTED_NSCN);
			if (cb->interface_type == HAL_IF_FC_L_PORT ||
			    cb->interface_type == HAL_IF_FC_N_PORT)
				st->dev_disc_state = 0;
			else 
				st->dev_disc_state = 2;
			st->dev_disc_retry = ISPFC_DEV_DISC_RETRY;
			ispfc_reestablish_connection_1(cb, st, NFX_OK);
		}
		else {
			(*st->dev_disc_callback)(cb, st->dev_disc_cookie, status);
		}
	}
}

#define ISPFC_CT_SUCCESS		NFX_OK
#define ISPFC_CT_LOGICAL_FAILURE	NFX_ERR-1
#define ISPFC_CT_TRANSPORT_FAILURE	NFX_ERR-2
#define ISPFC_CT_TIMEOUT		NFX_ERR-3

#define ISPFC_CT_DEFERRED_TIMER_PERIOD	2000

/*******************************************************************************
 * ispfc_reestablish_connection_1 -
 * 
 *  Runs the device discovery state machine.
 *	   if ("LOOP (no FL-port)")
 *		 --> state 0
 *	   else
 *		 --> state ??
 *
 *	   State 0: 
 *		 o mailbox(GET_LOOP_MAP)
 *		 --> state 1
 *	   State 1:
 *		 TIMEOUT --> return(FAILURE)
 *		 FAILURE --> state 0 (retry)
 *		 SUCCESS --> return(SUCCESS)
 *
 *	   State 2:		// GIEL
 *		 o CT(GIEL)
 *		 --> state 3
 *	   State 3:		// GIELN
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> state 2 (after deferred delay)
 *		 o CT(GIELN)
 *		 --> state 4
 *	   State 4:		// GFN
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> deferred retry
 *		 o CT(GFN)
 *	   State 5:		// GPL
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> deferred retry
 *		 o CT(GPL)
 *	   State 6:		// GPPN
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> deferred retry
 *		 o if (port count exhausted)
 *		      --> state 8
 *		   else
 *		      o CT(GPPN)
 *		      --> state 7
 *         State 7:		
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> deferred retry
 *		 --> state 6
 *	   State 8:		// GID_FT
 *		 o CT(GID_FT)
 *               --> state 9
 *	   State 9:
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> deferred retry
 *		 o if (port count exhausted)
 *		      --> state 11
 *		   else	
 *		      o CT(GFPN_ID)
 *		      --> state 10
 *	   State 10:
 *		 T_FAILURE --> return(FAILURE)
 *		 L_FAILURE --> deferred retry
 *		 o if (port count exhausted)
 *		      --> state 11
 *		   else	
 *		      --> state 9
 *         State 11:
 *               --> return(SUCCESS)
 *
 ******************************************************************************/
void
ispfc_reestablish_connection_1(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_discovery_state_t *st = cookie;
	ct_desc_t *		     ctd = &st->u.fabric.ctd;
	eee_desc_t *		     e_desc;
	eee_descBuf_t *		     bd;
	int			     rc;
	int			     i;
	byte			     dev_alpa;
	ispfc_dev_t *		     dev;
	channel_name_t		     ch;
	wwn_t *			     ie_name;
	wwn_t *			     port_name;

 next_state:

	FC_DBG(1, ("%s: ispfc_reestablish_connection_1: state = %d\n", 
		   cb->debug_name, st->dev_disc_state));

	switch(st->dev_disc_state) {
	case 0:				/* Start of LOOP state machine */
		/* Out of retries? Return failure */
		if (st->dev_disc_retry == 0) {
			/* Failure */
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}

		if (cb->interface_type == HAL_IF_FC_N_PORT) {
			/* Fire off a GET LOOP MAP mailbox command */
			cb->mbox_in_valid_count = 7;
			cb->mbox_out_valid_count = 2;
			cb->mbox_in[0] = ISPFC_GET_ID_LIST;
			cb->mbox_in[1] = VA_TO_IO32(&st->peer_id) >> 16;
			cb->mbox_in[2] = VA_TO_IO32(&st->peer_id) & 0xFFFF;
			cb->mbox_in[3] = 0;
			cb->mbox_in[4] = 0;
			cb->mbox_in[5] = 0;
			cb->mbox_in[6] = 0;
			st->dev_disc_state = 1;
			ispfc_mbox_start(cb, ispfc_reestablish_connection_1, st);
		}
		else {
			/* Fire off a GET LOOP MAP mailbox command */
			cb->mbox_in_valid_count = 8;
			cb->mbox_in[0] = ISPFC_GET_FCAL_POSITION_MAP;
			cb->mbox_in[1] = 0;
			cb->mbox_in[2] = VA_TO_IO32(&st->fcal_map) >> 16;
			cb->mbox_in[3] = VA_TO_IO32(&st->fcal_map) & 0xFFFF;
			cb->mbox_in[4] = 0;
			cb->mbox_in[5] = 0;
			cb->mbox_in[6] = 0;
			cb->mbox_in[7] = 0;
			st->dev_disc_state = 1;
			ispfc_mbox_start(cb, ispfc_reestablish_connection_1, st);
		}
		return;

	case 1:
		if (status == MBOX_CMD_STATUS_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == MBOX_CMD_STATUS_FAILED) {
			--st->dev_disc_retry;
			st->dev_disc_state = 0;
			goto next_state;
		}

#if defined(ISPFC_SIM)
		/* Not simulated tapes with AL topologies */
		if (ispfc_fcadmin_disk_count > ISPFC_SIM_MAX_DISKS)
			ispfc_fcadmin_disk_count = ISPFC_SIM_MAX_DISKS;

		if (cb->interface_type == HAL_IF_FC_N_PORT) {
			cb->mbox_out[1] = 2;
			st->peer_id[0] = (cb->host_port_id == 0xef) ? 0x10000e8 : 0x00000ef;
			st->peer_id[1] = 0x7efffffe;
		}
		else {
			st->fcal_map.port_count = ispfc_fcadmin_disk_count+1;
			st->fcal_map.port_map[0] = ispfc_loopid_to_alpa(cb->host_loop_id);
			st->fcal_map.port_map[1] = ispfc_loopid_to_alpa(0x1);
			st->fcal_map.port_map[2] = ispfc_loopid_to_alpa(0x2);
			st->fcal_map.port_map[3] = ispfc_loopid_to_alpa(0x3);
			st->fcal_map.port_map[4] = ispfc_loopid_to_alpa(0x10);
			st->fcal_map.port_map[5] = ispfc_loopid_to_alpa(0x11);
			st->fcal_map.port_map[6] = ispfc_loopid_to_alpa(0x12);
			st->fcal_map.port_map[7] = ispfc_loopid_to_alpa(0x20);
			st->fcal_map.port_map[8] = ispfc_loopid_to_alpa(0x21);
			st->fcal_map.port_map[9] = ispfc_loopid_to_alpa(0x22);
			st->fcal_map.port_map[10] = ispfc_loopid_to_alpa(0x30);
			st->fcal_map.port_map[11] = ispfc_loopid_to_alpa(0x31);
			st->fcal_map.port_map[12] = ispfc_loopid_to_alpa(0x32);
			st->fcal_map.port_map[13] = ispfc_loopid_to_alpa(0x33);
			st->fcal_map.port_map[14] = ispfc_loopid_to_alpa(0x40);
			st->fcal_map.port_map[15] = ispfc_loopid_to_alpa(0x41);
			st->fcal_map.port_map[16] = ispfc_loopid_to_alpa(0x42);
			st->fcal_map.port_map[17] = ispfc_loopid_to_alpa(0x50);
			st->fcal_map.port_map[18] = ispfc_loopid_to_alpa(0x51);
			st->fcal_map.port_map[19] = ispfc_loopid_to_alpa(0x52);
			st->fcal_map.port_map[20] = ispfc_loopid_to_alpa(0x60);
		}
#endif /* defined(ISPFC_SIM) */

		++cb->dev_generation;

		if (cb->interface_type == HAL_IF_FC_N_PORT) {
			VERIFY(cb->mbox_out[1] == 2);
			for (i = 0; i < 2; ++i) {
				st->peer_id[i] &= 0x00ffffff;
				if (st->peer_id[i] == 0x00fffffe) /* F_PORT FC ID */
					continue;
				dev = ispfc_device_lookup_by_phys_dev(cb, st->peer_id[i]);
				if (dev == NULL) {
					WWN_SET(&ch.channel_wwn, &cb->adapter_wwn);
					rc = ispfc_channel_lookup(&ch);
					VERIFY(rc == NFX_OK);
					dev = ispfc_device_new(cb, st->peer_id[i], &ch);
					VERIFY(dev);
				}
				else {
					/* Verify that the WWN of the device hasn't changed. */
					dev->dev_state |= ISPFC_DF_ID_VERIFICATION_REQUESTED;
					dev->dev_generation = cb->dev_generation;
				}
			}
		}
		else {
			/* 
			 * Traverse loop map and create new device instances
			 */
			for (i = 0; i < st->fcal_map.port_count; i++) {
				dev_alpa = st->fcal_map.port_map[i];
				/* Skip host port */
				if (dev_alpa == ispfc_loopid_to_alpa(cb->host_loop_id))
					continue;
				dev = ispfc_device_lookup_by_phys_dev(cb, dev_alpa);
				if (dev == NULL) {
					WWN_SET(&ch.channel_wwn, &cb->adapter_wwn);
					rc = ispfc_channel_lookup(&ch);
					VERIFY(rc == NFX_OK);
					dev = ispfc_device_new(cb, dev_alpa, &ch);
					VERIFY(dev);
				}
				else {
					/* Verify that the WWN of the device hasn't changed. */
					dev->dev_state |= ISPFC_DF_ID_VERIFICATION_REQUESTED;
					dev->dev_generation = cb->dev_generation;
				}
			}
		}
		/* 
		 * ... and remove device instances for which physical
		 * devices have disappeared.
		 */
		FOR_ALL_DEVS({
			if (dev->dev_generation != cb->dev_generation) {
				dev->dev_state |= ISPFC_DF_INVALIDATION_REQUESTED;
			}
		});
		ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_SUCCESS);
		return;

	case 2:				/* Start of FABRIC state machine */
		/* Initialize common CT descriptor */
		if (st->u.fabric.ctd.e_desc == NULL) {
			st->u.fabric.ctd.e_desc = eee_allocateDesc(EEE_SHARED_POOL, 0);
			bzero(st->u.fabric.ctd.e_desc, sizeof(eee_desc_t));
		}
		e_desc = st->u.fabric.ctd.e_desc;
		bd = e_desc->bd;
		bd[1].buf = (void *)&st->ct_req;
		bd[1].len = sizeof(ispfc_ct_req_t);
		bd[1].control = EDESC_BD_CMD_IN_USE;
		bd[2].buf = (void *)&st->ct_GIEL_resp;
		bd[2].len = sizeof(ispfc_ct_resp_t);
		bd[2].control = EDESC_BD_CMD_IN_USE;
		e_desc->hdr.control = 
			((2 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK) | 
			((sizeof(ispfc_ct_req_t) + sizeof(ispfc_ct_resp_t)) & EDESC_TOTAL_LEN);

		/* Fire off GIEL CT request */
		ctd->callback = ispfc_reestablish_connection_ct_done;
		ctd->handle = st;
		st->dev_disc_state = 3;
		st->dev_disc_generic_counter_a = 0; /* Will be used as IE index */
		ispfc_ms_GIEL(cb, ctd);
		return;
		
	case 3:			/* GIEL pending */
		/* 
		 * QLogic SanBox2 switches with firmware rev.
		 * SWImageVersion (1) - build date   V1.0-2-0   (Mon Jun 25 16:31:49 2001)
		 * don't support MS queries. Skip the topology discovery part. Problem is, we don't know channel info.
		 */
		cb->link_state &= ~ISPFC_CLF_FC_GS3_NOT_SUPPORTED;
		if (status == ISPFC_CT_TRANSPORT_FAILURE) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE || status == ISPFC_CT_TIMEOUT) {
			if (st->dev_disc_deferred_retry) {
				cb->link_state |= ISPFC_CLF_FC_GS3_NOT_SUPPORTED;
				st->dev_disc_state = 8;
				goto next_state;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 2;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			/* IE count exhausted ? */
			if (st->dev_disc_generic_counter_a < st->ct_GIEL_resp.u.GIEL.ie_count) {
				ie_name = (wwn_t *)st->ct_GIEL_resp.u.GIEL.ie[st->dev_disc_generic_counter_a].ie_name;
				FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
					   "GIEL.ie[%d].ie_name = %s\n",
					   cb->debug_name, st->dev_disc_state,
					   st->dev_disc_generic_counter_a,
					   wwn_to_str(ie_name, NULL)));
				/* Fire off a GIELN CT request */
				e_desc = st->u.fabric.ctd.e_desc;
				bd = e_desc->bd;
				bd[2].buf = (void *)&st->ct_GIELN_resp;
				ctd->callback = ispfc_reestablish_connection_ct_done;
				ctd->handle = st;
				st->dev_disc_state = 4;
				ispfc_ms_GIELN(cb, ie_name, ctd);
				return;
			}
			else {
				st->dev_disc_state = 8;
				goto next_state;
			}
		}
		break;

	case 4:			/* GIELN pending */
		if (status == ISPFC_CT_TRANSPORT_FAILURE || status == ISPFC_CT_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE) {
			if (st->dev_disc_deferred_retry) {
				ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
				return;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 3;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
				   "GIELN.ie_logical_name(%d) = '%s'\n", 
				   cb->debug_name, st->dev_disc_state,
				   st->ct_GIELN_resp.u.GIELN.ie_logical_name.len,
				   st->ct_GIELN_resp.u.GIELN.ie_logical_name.name));

			/* Fire off a GFN CT request */
			e_desc = st->u.fabric.ctd.e_desc;
			bd = e_desc->bd;
			bd[2].buf = (void *)&st->ct_GFN_resp;
			ctd->callback = ispfc_reestablish_connection_ct_done;
			ctd->handle = st;
			st->dev_disc_state = 5;
			ie_name = (wwn_t *)st->ct_GIEL_resp.u.GIEL.ie[st->dev_disc_generic_counter_a].ie_name;
			ispfc_ms_GFN(cb, ie_name, ctd);
			return;
		}
		break;

	case 5:			/* GFN pending */
		if (status == ISPFC_CT_TRANSPORT_FAILURE || status == ISPFC_CT_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE) {
			if (st->dev_disc_deferred_retry) {
				ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
				return;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 3;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
				   "GFN.fabric_name = %s\n", 
				   cb->debug_name, st->dev_disc_state,
				   wwn_to_str((wwn_t *)st->ct_GFN_resp.u.GFN.fabric_name, NULL)));

			/* Fire off a GPL CT request */
			e_desc = st->u.fabric.ctd.e_desc;
			bd = e_desc->bd;
			bd[2].buf = (void *)&st->ct_GPL_resp;
			ctd->callback = ispfc_reestablish_connection_ct_done;
			ctd->handle = st;
			st->dev_disc_state = 6;
			ie_name = (wwn_t *)st->ct_GIEL_resp.u.GIEL.ie[st->dev_disc_generic_counter_a].ie_name;
			ispfc_ms_GPL(cb, ie_name, ctd);
			st->dev_disc_generic_counter_b = 0; /* Will be used as port index */
			return;
		}
		break;

	case 6:			/* GPL pending */
		if (status == ISPFC_CT_TRANSPORT_FAILURE || status == ISPFC_CT_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE) {
			if (st->dev_disc_deferred_retry) {
				ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
				return;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 3;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			/* Exhausted ports ? */
			if (st->dev_disc_generic_counter_b < st->ct_GPL_resp.u.GPL.port_count) {
				port_name = (wwn_t *)st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_name;
				FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
					   "GPL.port[%d].port_name/port_type = %s/0x%x\n", 
					   cb->debug_name, st->dev_disc_state,
					   st->dev_disc_generic_counter_b,
					   wwn_to_str(port_name, NULL),
					   st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_type));

				/* Fire off a GPPN CT request */
				e_desc = st->u.fabric.ctd.e_desc;
				bd = e_desc->bd;
				bd[2].buf = (void *)&st->ct_GPPN_resp;
				ctd->callback = ispfc_reestablish_connection_ct_done;
				ctd->handle = st;
				st->dev_disc_state = 7;
				ispfc_ms_GPPN(cb, port_name, ctd);
				return;
			}
			else {
				/* Go do next IE */
				status = ISPFC_CT_SUCCESS;
				++st->dev_disc_generic_counter_a;
				st->dev_disc_state = 3;
				goto next_state;
			}
		}
		break;

	case 7:			/* GPPN pending */
		if (status == ISPFC_CT_TRANSPORT_FAILURE || status == ISPFC_CT_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE) {
			if (st->dev_disc_deferred_retry) {
				ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
				return;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 3;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			channel_name_t		ch1, ch2;
			int			ret_val;
#ifdef DEBUG
			char			switch_name[32], port_name[32];
#endif

			FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
				   "GPPN.port_number = %d\n", 
				   cb->debug_name, st->dev_disc_state,
				   st->ct_GPPN_resp.u.GPPN.port_number));

#ifdef DEBUG
			wwn_to_str((wwn_t *)st->ct_GIEL_resp.u.GIEL.ie[st->dev_disc_generic_counter_a].ie_name, switch_name);
			wwn_to_str((wwn_t *)st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_name, port_name);
			FC_DBG(1, ("%s: port %d, type 0x%x (%s) found on switch '%s' (%s)\n",
				   cb->debug_name,
				   st->ct_GPPN_resp.u.GPPN.port_number, 
				   st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_type, port_name,
				   st->ct_GIELN_resp.u.GIELN.ie_logical_name.name, switch_name));
#endif
			/*
			 * Skip if port isn't an FL- or F- port
			 */
			if (st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_type == PORT_TYPE_F ||
			    st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_type == PORT_TYPE_FL) {
				bzero(&ch2, sizeof(channel_name_t));
				WWN_SET(&ch1.channel_wwn, 
					st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_name);
				WWN_SET(&ch2.channel_wwn, 
					st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_name);
				sprintf(ch2.channel_name, "%s.%d:", 
					st->ct_GIELN_resp.u.GIELN.ie_logical_name.name,
					st->ct_GPPN_resp.u.GPPN.port_number);
				ch2.channel_interface_type = 
					(st->ct_GPL_resp.u.GPL.port[st->dev_disc_generic_counter_b].port_type == PORT_TYPE_F) ?
					HAL_IF_FC_F_PORT : HAL_IF_FC_FL_PORT;
				ch2.u.fc_f_or_fc_fl.channel_phys_port = st->ct_GPPN_resp.u.GPPN.port_number;
				WWN_SET(&ch2.u.fc_f_or_fc_fl.channel_switch_wwn, 
					st->ct_GIEL_resp.u.GIEL.ie[st->dev_disc_generic_counter_a].ie_name);
				if (ispfc_channel_lookup(&ch1) == NFX_OK) {
					/* Channel already exists -- verify that name hasn't changed */
					if (strcmp(ch1.channel_name, ch2.channel_name) != 0) {
						ret_val = ispfc_channel_remove(&ch1); 
						VERIFY(ret_val == NFX_OK);
						ret_val = ispfc_channel_add(&ch2); 
						VERIFY(ret_val == NFX_OK);
					}
				}
				else {
					ret_val = ispfc_channel_add(&ch2);
					VERIFY(ret_val == NFX_OK);
				}
			}

			/* Go do next port */
			status = ISPFC_CT_SUCCESS;
			++st->dev_disc_generic_counter_b;
			st->dev_disc_state = 6;
			goto next_state;
		}
		break;

	case 8:			/* Start of Nameserver discovery */
		/* Fire off a GID_FT request */
		e_desc = st->u.fabric.ctd.e_desc;
		bd = e_desc->bd;
		bd[2].buf = (void *)&st->ct_GID_FT_resp;
		ctd->callback = ispfc_reestablish_connection_ct_done;
		ctd->handle = st;
		st->dev_disc_state = 9;
		st->dev_disc_generic_counter_a = 0; /* Will be used as port index */
		++cb->dev_generation;
		ispfc_ns_GID_FT(cb, ctd);
		return;

	case 9:
		if (status == ISPFC_CT_TRANSPORT_FAILURE || status == ISPFC_CT_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE) {
			if (st->dev_disc_deferred_retry) {
				/* There may be no devices available */
				st->dev_disc_state = 11;
				goto next_state;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 8;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			ispfc_ct_resp_GID_FT_t	*p = &st->ct_GID_FT_resp.u.GID_FT.list[st->dev_disc_generic_counter_a];

			FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
				   "GID_FT.list[%d] = 0x%08x\n",
				   cb->debug_name, st->dev_disc_state,
				   st->dev_disc_generic_counter_a,
				   p->portid));
			
			/* 
			 * Ancor returns the host entry for GID_FT,
			 * even though we haven't registered an FC-4
			 * type. Filter it out.  
			 */
			if (p->portid == cb->host_port_id) {
				FC_DBG(1, ("%s: Skipping host entry ... portID 0x%x\n", 
					   cb->debug_name, p->portid));
				if (p->control_byte & 0x80)
					st->dev_disc_state = 11;
				else {
					++st->dev_disc_generic_counter_a;
					st->dev_disc_state = 9;
					status = ISPFC_CT_SUCCESS;
				}
				goto next_state;
			}
			 
			/* Fire off a GFPN_ID request */
			e_desc = st->u.fabric.ctd.e_desc;
			bd = e_desc->bd;
			bd[2].buf = (void *)&st->ct_GFPN_ID_resp;
			ctd->callback = ispfc_reestablish_connection_ct_done;
			ctd->handle = st;
			st->dev_disc_state = 10;
			ispfc_ns_GFPN_ID(cb, p->portid, ctd);
			return;
		}
		break;

	case 10:
		if (status == ISPFC_CT_TRANSPORT_FAILURE || status == ISPFC_CT_TIMEOUT) {
			ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
			return;
		}
		else
		if (status == ISPFC_CT_LOGICAL_FAILURE) {
			if (st->dev_disc_deferred_retry) {
				ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_FAILED);
				return;
			}
			else {
				/* Start a deferred timer, to allow for retry */
				st->dev_disc_state = 9;
				ispfc_reestablish_connection_defer(cb, st);
				return;
			}
		}
		else {
			ispfc_ct_resp_GID_FT_t	*p = &st->ct_GID_FT_resp.u.GID_FT.list[st->dev_disc_generic_counter_a];

			FC_DBG(2, ("%s: ispfc_reestablish_connection_1: state = %d: "
				   "GFPN_ID.node_name = %s\n", 
				   cb->debug_name, st->dev_disc_state,
				   wwn_to_str((wwn_t *)st->ct_GFPN_ID_resp.u.GFPN_ID.node_name, NULL)));

			if (!(cb->link_state & ISPFC_CLF_FC_GS3_NOT_SUPPORTED)) {
				/* Verify that the channel for this device exists */
				WWN_SET(&ch.channel_wwn, st->ct_GFPN_ID_resp.u.GFPN_ID.node_name);
				rc = ispfc_channel_lookup(&ch);

				/*
				 * If channel doesn't exist, we're
				 * probably seeing the QLogic SANBOX2
				 * switch running 1.1.13 firmware; the
				 * GFPN_ID query returned a screwed up
				 * fabric port name -- hack/fix it
				 * here. 
				 */
				if (rc != NFX_OK) {
					char s1[32];
					char s2[32];
#if defined(LITTLE_ENDIAN_EVAL_BOARD) || defined(AGILE_HARDWARE)
					inplace_ntohll(&ch.channel_wwn);
#endif
					*(uint64 *)&ch.channel_wwn &= ~0x00FFFF0000000000LL;
					*(uint64 *)&ch.channel_wwn |= (uint64)(p->portid & 0x00FF00) << 40;
#if defined(LITTLE_ENDIAN_EVAL_BOARD) || defined(AGILE_HARDWARE)
					inplace_ntohll(&ch.channel_wwn);
#endif
					FC_DBG(2, ("%s: ispfc_reestablish_connection_1: FPN %s doesn't exist; fix it to %s\n",
						   cb->debug_name, 
						   wwn_to_str((wwn_t *)st->ct_GFPN_ID_resp.u.GFPN_ID.node_name, s1),
						   wwn_to_str(&ch.channel_wwn, s2)));
					rc = ispfc_channel_lookup(&ch); 
					if (rc != NFX_OK) {
						FC_DBG(0, ("%s: Failed to find channel %s: "
							   "retrying assuming that port types are not known yet.\n",
							   cb->debug_name, 
							   wwn_to_str((wwn_t *)st->ct_GFPN_ID_resp.u.GFPN_ID.node_name, s1)));
						--st->dev_disc_retry;
						st->dev_disc_state = 2;
						goto next_state;
					}
				}
			}
			else {
				/* Use adapter channel instead. */
				WWN_SET(&ch.channel_wwn, &cb->adapter_wwn);
				rc = ispfc_channel_lookup(&ch);
				VERIFY(rc == NFX_OK);
			} 
			
			/* Existing device or new one ? */
			dev = ispfc_device_lookup_by_phys_dev(cb, p->portid);
			if (dev == NULL) {
				dev = ispfc_device_new(cb, p->portid, &ch);
				VERIFY(dev);
			}
			else {
				/* Verify that the WWN of the device hasn't changed. */
				dev->dev_state |= ISPFC_DF_ID_VERIFICATION_REQUESTED;
				dev->dev_generation = cb->dev_generation;
			}

			if (!(cb->link_state & ISPFC_CLF_FC_GS3_NOT_SUPPORTED)) {
				/*
				 * Set the FL-port for the channel if it isn't
				 * already set. Reset it if it's different,
				 * i.e. the domain number has changed.  
				 */
				if (ch.u.fc_f_or_fc_fl.channel_fl_port == 0 ||
				    ch.u.fc_f_or_fc_fl.channel_fl_port != ISPFC_DEV_PHYS_DEV_TO_FLA(p->portid)) {
					ch.u.fc_f_or_fc_fl.channel_fl_port = ISPFC_DEV_PHYS_DEV_TO_FLA(p->portid);
					rc = ispfc_channel_set_fl_port(&ch);
					VERIFY(rc == NFX_OK);
				}
				VERIFY(ch.u.fc_f_or_fc_fl.channel_fl_port == ISPFC_DEV_PHYS_DEV_TO_FLA(p->portid));
			}

			/* Exhausted ports ? */ 
			if (p->control_byte & 0x80)
				st->dev_disc_state = 11;
			else {
				++st->dev_disc_generic_counter_a;
				st->dev_disc_state = 9;
				status = ISPFC_CT_SUCCESS;
			}
			goto next_state;
		}
		break;

	case 11:
		/* 
		 * ... and remove device instances for which physical
		 * devices have disappeared.
		 */
		FOR_ALL_DEVS({
			if (dev->dev_generation != cb->dev_generation) {
				dev->dev_state |= ISPFC_DF_INVALIDATION_REQUESTED;
			}
		});
		ispfc_reestablish_connection_done(cb, st, REESTABLISH_STATUS_SUCCESS);
		return;

	default:
		VERIFY(0);
	}
}

void 
ispfc_reestablish_connection_ct_done(struct ct_desc *ctd)
{
	ispfc_cb_t *cb = hal_dt_entry_get()->hal_handle;
	VERIFY(cb);

	if ((ctd->status & CT_STATUS_MASK) != CT_STATUS_OK) {
		if ((ctd->status & CT_STATUS_MASK) == CT_STATUS_TIMEOUT)
			ispfc_reestablish_connection_1(cb, ctd->handle, ISPFC_CT_TIMEOUT);
		else
			ispfc_reestablish_connection_1(cb, ctd->handle, ISPFC_CT_TRANSPORT_FAILURE);
		return;
	}
	VERIFY(ctd->status & CT_STATUS_RESP_VALID);
	if (((ctd->status & CT_STATUS_RESP_MASK) >> CT_STATUS_RESP_SHIFT) != CT_STATUS_RESP_FS_ACC) {
		ispfc_reestablish_connection_1(cb, ctd->handle, ISPFC_CT_LOGICAL_FAILURE);
		return;
	}
	
	ispfc_reestablish_connection_1(cb, ctd->handle, ISPFC_CT_SUCCESS);
}

void
ispfc_reestablish_connection_defer(ispfc_cb_t *cb, ispfc_dev_discovery_state_t *st)
{
	++st->dev_disc_deferred_retry;
	st->dev_disc_timer_handle = eee_timerCreate(ISPFC_CT_DEFERRED_TIMER_PERIOD, EEE_TIMER_ONCE);
	VERIFY(st->dev_disc_timer_handle >= 0);
	eee_timerStart(st->dev_disc_timer_handle, (void *)ispfc_reestablish_connection_deferred_timeout,
		       cb, st, NULL, NULL);
}

int32
ispfc_reestablish_connection_deferred_timeout(int32 timer_handle, ispfc_cb_t *cb, ispfc_dev_discovery_state_t *st)
{
	int32 rc;

	VERIFY(st->dev_disc_timer_handle == timer_handle);
	rc = eee_timerDestroy(st->dev_disc_timer_handle);
	VERIFY(rc == NFX_OK);

	ispfc_reestablish_connection_1(cb, st, ISPFC_CT_SUCCESS);

	return(NFX_OK);
}

int32 
ispfc_device_management_handler(ispfc_cb_t *cb)
{
	ispfc_dev_t *		dev;
	ispfc_dev_t *		next;
	ispfc_dev_t *		lru_dev;
	int			rc;
	scsi_desc_t *		sd;
	scsi_task_mgmt_t *	tmo;

	/* 
	 * If link not up, don't bother; but do check if device
	 * invalidation requests are pending.
	 */
	if (LINK_STATE(cb) != ISPFC_CLS_STATE_UP) {
		/* FOR_ALL_DEVS({ */
		if ((dev = cb->dev_list)) {
			do {
				next = dev->next;

				if ((dev->dev_state & ISPFC_DF_INVALIDATION_REQUESTED) && dev->scsi_desc_busy == NULL &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_LOGIN_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_LOGOUT_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_TMO_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_RESET_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_GPDB_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_ID_VERIFICATION_PENDING &&
				    DEV_STATE(dev) != ISPFC_DS_STATE_DEFERRED_LOGIN) {
					if (DEV_STATE(dev) == ISPFC_DS_STATE_UP) {
						if (cb->interface_type != HAL_IF_FC_L_PORT &&
						    cb->interface_type != HAL_IF_FC_N_PORT) {
							/* If mailbox busy, bail */
							if (ispfc_mbox_is_busy(cb))
								goto next_device;
							
							/* Fire off a ISPFC_LOGOUT_FABRIC_PORT mailbox command */
							VERIFY(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED);
							cb->mbox_in_valid_count = 2;
							cb->mbox_in[0] = ISPFC_LOGOUT_FABRIC_PORT;
							cb->mbox_in[1] = dev->mapping_id << 8;
							SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGOUT_PENDING);
							ispfc_mbox_start(cb, ispfc_dev_logout_done_callout, dev);
						}
						else {
							/* For loop devices we don't need to explicitly logout */
							SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGOUT_PENDING);
							ispfc_dev_logout_done_callout(cb, dev, MBOX_CMD_STATUS_SUCCESS);
						}
					}
					else 
						ispfc_device_invalidate(cb, dev); 
				}
			next_device:
				dev = next;
			} while (cb->dev_list && dev != cb->dev_list);
		}
		/* }); */
		return(NFX_OK);
	}

	/* FOR_ALL_DEVS({ */
	if ((dev = cb->dev_list)) {
		do {
			next = dev->next;

			switch (DEV_STATE(dev)) {
			case ISPFC_DS_STATE_GPDB_PENDING:
			case ISPFC_DS_STATE_ID_VERIFICATION_PENDING:
			case ISPFC_DS_STATE_LOGIN_PENDING:
			case ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING:
			case ISPFC_DS_STATE_LOGOUT_PENDING:
			case ISPFC_DS_STATE_TMO_PENDING:
			case ISPFC_DS_STATE_RESET_PENDING:
				/* Waiting for mailbox command to complete -- do nothing. */
				break;

			case ISPFC_DS_STATE_DEFERRED_LOGIN:
			case ISPFC_DS_STATE_PAUSED:
			case ISPFC_DS_STATE_QUEUE_FLUSH_PENDING:
				/* Waiting for timer to pop -- do nothing */
				break;

			case ISPFC_DS_STATE_INVALIDATE_PENDING:
				/* Waiting for scsi_desc_busy queue to empty -- do nothing */
				break;

			case ISPFC_DS_STATE_QUIESCED:
				/* Waiting for Busy queue to empty or the device global timer to pop -- do nothing */
				break;

			case ISPFC_DS_STATE_LOGGED_OUT:
				/* Stable state -- nothing to do, unless there's something in the waiting queue. */
				VERIFY(dev->scsi_desc_busy == NULL);
				/* If something in the waiting queue, better get this device back up */
				if (dev->scsi_desc_waiting) {
					SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);
				}
				break;

			case ISPFC_DS_STATE_LOGIN_REQUESTED:
				if (dev->scsi_desc_busy)
					break;
				if (dev->dev_state & ISPFC_DF_INVALIDATION_REQUESTED) {
					ispfc_device_invalidate(cb, dev); 
					break;
				}

				/* If mailbox busy, bail */
				if (ispfc_mbox_is_busy(cb))
					break;

				if (cb->interface_type == HAL_IF_FC_L_PORT || 
				    cb->interface_type == HAL_IF_FC_N_PORT) {
					/* Fire off a PORT LOGIN mailbox command */
					VERIFY(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED);
					cb->mbox_in_valid_count = 3;
					cb->mbox_in[0] = ISPFC_PORT_LOGIN;
					cb->mbox_in[1] = dev->mapping_id << 8;
					cb->mbox_in[2] = 1; /* Force login */
					SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_PENDING);
					ispfc_mbox_start(cb, ispfc_dev_login_done_callout, dev);
				}
				else {
					/* 
					 * Allocate a mapping ID if we don't
					 * already have one -- if we can't, kick
					 * the LRU dev out. 
					 */
					if (!(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED)) {
						rc = ispfc_fabric_map_allocate_id(cb);
						if (rc == NFX_ERR) {
							lru_dev = ispfc_fabric_map_find_lru(cb);
							VERIFY(lru_dev);
							lru_dev->dev_state |= ISPFC_DF_LOGOUT_REQUESTED;
							break;
						}
						else
							dev->mapping_id = rc;
					}							
					/* Fire off a FABRIC PORT LOGIN mailbox command */
					dev->dev_state |= ISPFC_DF_MAPPING_ID_ALLOCATED;
					cb->mbox_in_valid_count = 4;
					cb->mbox_in[0] = ISPFC_LOGIN_FABRIC_PORT;
					cb->mbox_in[1] = dev->mapping_id << 8;
					cb->mbox_in[2] = (dev->phys_dev_id & ISPFC_DEV_PHYS_DEV_ID_MASK_HI) >> 16;
					cb->mbox_in[3] = dev->phys_dev_id & ISPFC_DEV_PHYS_DEV_ID_MASK_LO;
					SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_PENDING);
					ispfc_mbox_start(cb, ispfc_dev_login_done_callout, dev);
				}
				break;

			case ISPFC_DS_STATE_UP:
				/*
				 * Check option/action flags.
				 */
				if (dev->dev_state & ISPFC_DF_STATE_ACTION_FLAGS_MASK) {
					if (dev->dev_state & ISPFC_DF_INVALIDATION_REQUESTED) { 
						if (cb->interface_type != HAL_IF_FC_L_PORT &&
						    cb->interface_type != HAL_IF_FC_N_PORT) {
							/* If mailbox busy, bail */
							if (ispfc_mbox_is_busy(cb))
								break;

							/* Fire off a ISPFC_LOGOUT_FABRIC_PORT mailbox command */
							VERIFY(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED);
							cb->mbox_in_valid_count = 2;
							cb->mbox_in[0] = ISPFC_LOGOUT_FABRIC_PORT;
							cb->mbox_in[1] = dev->mapping_id << 8;
							SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGOUT_PENDING);
							ispfc_mbox_start(cb, ispfc_dev_logout_done_callout, dev);
						}
						else {
							/* For loop devices we don't need to explicitly logout */
							SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGOUT_PENDING);
							ispfc_dev_logout_done_callout(cb, dev, MBOX_CMD_STATUS_SUCCESS);
						}
					}
					else
					if (dev->dev_state & ISPFC_DF_ID_VERIFICATION_REQUESTED) {
						/* If mailbox busy, bail */
						if (ispfc_mbox_is_busy(cb))
							break;

						/* Fire off a GET PORT DATA BASE mailbox command */
						cb->mbox_in_valid_count = 8;
						cb->mbox_in[0] = ISPFC_GET_PORT_DB;
						cb->mbox_in[1] = dev->mapping_id << 8;
						cb->mbox_in[2] = VA_TO_IO32(dev->port_database) >> 16;
						cb->mbox_in[3] = VA_TO_IO32(dev->port_database) & 0x00FFFF;
						cb->mbox_in[4] = 0;
						cb->mbox_in[5] = 0;
						cb->mbox_in[6] = 0;
						cb->mbox_in[7] = 0;
						SET_DEV_STATE(dev, ISPFC_DS_STATE_ID_VERIFICATION_PENDING);
						ispfc_mbox_start(cb, ispfc_dev_get_port_database_done_callout, dev);
						dev->dev_state &= ~ISPFC_DF_ID_VERIFICATION_REQUESTED;
						WWN_SET(&dev->dev_old_portname, &dev->dev_portname);
					}
					else
					if (dev->dev_state & ISPFC_DF_LOGOUT_REQUESTED) {
						VERIFY(cb->interface_type != HAL_IF_FC_L_PORT && cb->interface_type != HAL_IF_FC_N_PORT);
						/* If mailbox busy, bail */
						if (ispfc_mbox_is_busy(cb))
							break;

						/* Fire off a ISPFC_LOGOUT_FABRIC_PORT mailbox command */
						VERIFY(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED);
						cb->mbox_in_valid_count = 2;
						cb->mbox_in[0] = ISPFC_LOGOUT_FABRIC_PORT;
						cb->mbox_in[1] = dev->mapping_id << 8;
						SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGOUT_PENDING);
						ispfc_mbox_start(cb, ispfc_dev_logout_done_callout, dev);
						dev->dev_state &= ~ISPFC_DF_LOGOUT_REQUESTED;
					}
					else
					if (dev->dev_state & ISPFC_DF_REGISTRATION_REQUESTED) {
						ispfc_device_register(cb, dev);
						dev->dev_state &= ~ISPFC_DF_REGISTRATION_REQUESTED;
					}
				}
				else {
					/* 
					 * If anything in the waiting queues, start it.
					 */
					while (dev->dev_cmd_busy_count < ISPFC_MAX_SCSI_TAGS * ISPFC_MAX_LUNS_PER_TARGET &&
					       (sd = ispfc_dequeue_desc_off_dev_waiting_queue(cb, dev))) {
						rc = ispfc_scsi_dispatch_desc(cb, dev, sd);
						/* If dispatch failed, stick it back on the waiting list and bail */
						if (rc != NFX_OK) {
							ispfc_queue_desc_on_dev_waiting_queue(cb, dev, sd);
							break;
						}
					}
				}
				break;

			case ISPFC_DS_STATE_TMO_REQUESTED:
				/*
				 * Start one request in the waiting queue. There may only be one busy at a time.
				 */
				VERIFY(dev->scsi_task_mgmt_waiting);
				if (dev->scsi_task_mgmt_busy == NULL) {
					tmo = ispfc_dequeue_tmo_off_dev_waiting_queue(cb, dev);
					rc = ispfc_scsi_task_mgmt_dispatch(cb, dev, tmo);
					/* If dispatch failed, stick it back on the waiting list and bail */
					if (rc != NFX_OK)
						ispfc_queue_tmo_on_dev_waiting_queue(cb, dev, tmo);
				}
				break;

			case ISPFC_DS_STATE_RESET_REQUESTED:
				/* If mailbox busy, bail */
				if (ispfc_mbox_is_busy(cb))
					break;

				if (cb->interface_type == HAL_IF_FC_L_PORT || cb->interface_type == HAL_IF_FC_N_PORT) {
					/* Fire off a ISPFC_LIP_RESET mailbox command */
					cb->mbox_in_valid_count = 4;
					cb->mbox_in[0] = ISPFC_LIP_RESET;
					cb->mbox_in[1] = dev->mapping_id << 8;
					cb->mbox_in[2] = 0;
					cb->mbox_in[3] = 0;
					SET_LINK_STATE(cb, ISPFC_CLS_STATE_LIP_GENERATED);
					ispfc_mbox_start(cb, ispfc_dev_reset_device_0, dev);
				}
				else {
					ispfc_lfa_cb_t *	req;
					ispfc_lfa_resp_t *	resp;
					channel_name_t		ch;

					dev->lfa_req = eee_ramAlloc(sizeof(ispfc_lfa_cb_t));
					VERIFY(dev->lfa_req);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
					dev->lfa_req = (void *)cached_to_uncached((address_t)dev->lfa_req);
#else
					(ps_uint_t)dev->lfa_req |= 0x20000000L; /* Turn on uncached access */
#endif
#endif
					req = dev->lfa_req;
					dev->lfa_resp = eee_ramAlloc(sizeof(ispfc_lfa_resp_t));
					VERIFY(dev->lfa_resp);
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
					dev->lfa_resp = (void *)cached_to_uncached((address_t)dev->lfa_resp);
#else
					(ps_uint_t)dev->lfa_resp |= 0x20000000L; /* Turn on uncached access */
#endif
#endif
					resp = dev->lfa_resp;

					WWN_SET(&ch.channel_wwn, &dev->channel_wwn);
					rc = ispfc_channel_lookup(&ch);
					VERIFY(rc == NFX_OK);
					VERIFY(ch.channel_interface_type == HAL_IF_FC_FL_PORT);
					
					bzero(req, sizeof(ispfc_lfa_cb_t));
					req->lfa_response_buf_len = 4;
					req->lfa_response_buf_lo32 = VA_TO_IO32(resp);
					req->lfa_command_len = 4;
					req->lfa_fl_port = ch.u.fc_f_or_fc_fl.channel_fl_port;
					req->lfa_command_code = LFA_COMMAND_CODE_LINIT;
					req->u.linit.lfa_linit_init_function = LFA_LINIT_INIT_FUNCTION_NORMAL;
					req->u.linit.lfa_linit_lip_byte_3 = ISPFC_DEV_PHYS_DEV_TO_ALPA(dev->phys_dev_id);
					req->u.linit.lfa_linit_lip_byte_4 = 0x00;

#if 0					/* Seems like Qlogic byte swaps LFA autonomously */
					inplace_ntohs(&req->lfa_response_buf_len);
					inplace_ntohl(&req->lfa_response_buf_lo32);
					inplace_ntohs(&req->lfa_command_len);
					inplace_ntohl(&req->lfa_fl_port);
					inplace_ntohs(&req->lfa_command_code);
					inplace_ntohs(&req->u.linit.lfa_linit_init_function);
#endif
					/* Fire off a SEND LFA mailbox command */
					cb->mbox_in_valid_count = 8;
					cb->mbox_in[0] = ISPFC_SEND_LFA;
					cb->mbox_in[1] = sizeof(ispfc_lfa_cb_t)/sizeof(ushort16);
					cb->mbox_in[2] = VA_TO_IO32(req) >> 16;
					cb->mbox_in[3] = VA_TO_IO32(req) & 0xFFFF;
					cb->mbox_in[4] = 0;
					cb->mbox_in[5] = 0;
					cb->mbox_in[6] = 0;
					cb->mbox_in[7] = 0;
					SET_LINK_STATE(cb, ISPFC_CLS_STATE_NSCN_GENERATED);
					ispfc_mbox_start(cb, ispfc_dev_reset_device_0, dev);
				}
				SET_DEV_STATE(dev, ISPFC_DS_STATE_RESET_PENDING);
				break;

			default:
				FC_DBG(0, ("%s: Invalid state for physdev 0x%x: state 0x%x\n", 
					   cb->debug_name, dev->phys_dev_id, dev->dev_state));
				VERIFY(0);
			}
			dev = next;
		} while (dev != cb->dev_list && cb->dev_list);
	}
	/* }); */

	return(NFX_OK);
}

void
ispfc_dev_login_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_t *	dev = cookie;
	int		rc;
	int		invalidate_now = 0;
	byte		mapping_id;
	uint32		port_id;

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_LOGIN_PENDING);

#if defined(ISPFC_SIM) && 0
	/* Simulate 3 login failures and then success */
	if (dev->device_login_level < 3)
		status = MBOX_CMD_STATUS_FAILED;
#endif

	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else
	if (status == MBOX_CMD_STATUS_FAILED) {
		/* If fabric, free mapping id */
		if (cb->mbox_out[0] == 0x4007) {
			VERIFY(cb->interface_type != HAL_IF_FC_L_PORT && cb->interface_type != HAL_IF_FC_N_PORT);
			/* 
			 * Port ID used: The selected port ID is
			 * allocated for another loop ID -- See
			 * Outgoing Mailbox 1 register.  
			 */
			 mapping_id = cb->mbox_out[1];

			 FC_DBG(0, ("%s: Login failed for phys_dev_id 0x%x: FC Port ID 0x%x already mapped to ID 0x%x "
				    "-- forcing logout and retry.\n",
				    cb->debug_name, dev->phys_dev_id, dev->phys_dev_id, mapping_id));

			 /* 
			  * This had better be a residue from a failed
			  * login, else our state machine is seriously
			  * confused.
			  */
			 VERIFY(!ispfc_fabric_map_is_allocated_id(cb, mapping_id));

			 rc = ispfc_fabric_map_deallocate_id(cb, dev->mapping_id);
			 VERIFY(rc == NFX_OK);
			 dev->dev_state &= ~ISPFC_DF_MAPPING_ID_ALLOCATED;

			 /* Fire off a ISPFC_LOGOUT_FABRIC_PORT mailbox command */
			 cb->mbox_in_valid_count = 2;
			 cb->mbox_in[0] = ISPFC_LOGOUT_FABRIC_PORT;
			 cb->mbox_in[1] = mapping_id << 8;
			 SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING);
			 ispfc_mbox_start(cb, ispfc_dev_login_cleanup_done_callout, dev);

			 return;
		}
		else 
		if (cb->mbox_out[0] == 0x4008) {
			VERIFY(cb->interface_type != HAL_IF_FC_L_PORT && cb->interface_type != HAL_IF_FC_N_PORT);
			/* 
			 * Loop ID used: The selected loop ID is
			 * allocated for another port ID -- Outgoing
			 * Mailbox 1 and 2 registers.  
			 */
			port_id = (cb->mbox_out[1] << 16) | cb->mbox_out[2];
			
			FC_DBG(0, ("%s: Login failed for phys_dev_id 0x%x: Mapping ID 0x%x already consumed by FC port ID 0x%x "
				   "-- forcing logout and retry.\n",
				   cb->debug_name, dev->phys_dev_id, dev->mapping_id, port_id));

			rc = ispfc_fabric_map_deallocate_id(cb, dev->mapping_id);
			VERIFY(rc == NFX_OK);
			dev->dev_state &= ~ISPFC_DF_MAPPING_ID_ALLOCATED;

			/* Fire off a ISPFC_LOGOUT_FABRIC_PORT mailbox command */
			cb->mbox_in_valid_count = 2;
			cb->mbox_in[0] = ISPFC_LOGOUT_FABRIC_PORT;
			cb->mbox_in[1] = dev->mapping_id << 8;
			SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING);
			ispfc_mbox_start(cb, ispfc_dev_login_cleanup_done_callout, dev);

			return;
		}
		else 
		if (cb->mbox_out[0] == 0x4005) {
			/*
			 * Command error: The command did not complete
			 * for the reason specified in Outgoing
			 * Mailbox 1 register. 
			 */
			if (cb->mbox_out[1] == 6) {
				/* 
				 * Remote device doesn't support
				 * target function -- invalidate
				 * immediately.
				 */
				invalidate_now = 1;
			}
		}

		if (cb->interface_type != HAL_IF_FC_L_PORT && cb->interface_type != HAL_IF_FC_N_PORT) {
			rc = ispfc_fabric_map_deallocate_id(cb, dev->mapping_id);
			VERIFY(rc == NFX_OK);
			dev->dev_state &= ~ISPFC_DF_MAPPING_ID_ALLOCATED;
		}
		
		/* 
		 * Check deferred login pending count -- if
		 * exceeded, invalidate the device.
		 */
		if (invalidate_now || ++dev->device_login_level > DEVICE_DEFERRED_LOGIN_RETRY)
			ispfc_device_invalidate(cb, dev);
		else {
			/* Start a timer to pop in DEVICE_DEFERRED_LOGIN_TIMER msecs. */
                        dev->deferred_login_timer_handle = eee_timerCreate(DEVICE_DEFERRED_LOGIN_TIMER,
                                                                           EEE_TIMER_ONCE);
                        VERIFY(dev->deferred_login_timer_handle >= 0);
                        rc = eee_timerStart(dev->deferred_login_timer_handle,
                                            (void *)ispfc_dev_deferred_login_timer_done_callout,
                                            cb, dev, NULL, NULL);
			VERIFY(rc == NFX_OK);
                        SET_DEV_STATE(dev, ISPFC_DS_STATE_DEFERRED_LOGIN);
		}
	}
	else {
		/* Fire off a GET PORT DATA BASE mailbox command */
		cb->mbox_in_valid_count = 8;
		cb->mbox_in[0] = ISPFC_GET_PORT_DB;
		cb->mbox_in[1] = dev->mapping_id << 8;
		cb->mbox_in[2] = VA_TO_IO32(dev->port_database) >> 16;
		cb->mbox_in[3] = VA_TO_IO32(dev->port_database) & 0x00FFFF;
		cb->mbox_in[4] = 0;
		cb->mbox_in[5] = 0;
		cb->mbox_in[6] = 0;
		cb->mbox_in[7] = 0;
		SET_DEV_STATE(dev, ISPFC_DS_STATE_GPDB_PENDING);
		ispfc_mbox_start(cb, ispfc_dev_get_port_database_done_callout, dev);
	}
}

void 
ispfc_dev_login_cleanup_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_t *	dev = cookie;

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING);

	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}

        SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);
}

void 
ispfc_dev_logout_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_t *	dev = cookie;
	int		rc;

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_LOGOUT_PENDING);

#if defined(ISPFC_SIM)
	status = MBOX_CMD_STATUS_SUCCESS;
#endif

	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		/* Device is logged out; free mapping ID if necessary */
		if (cb->interface_type != HAL_IF_FC_L_PORT && cb->interface_type != HAL_IF_FC_N_PORT) {
			dev->dev_state &= ~ISPFC_DF_MAPPING_ID_ALLOCATED;
			rc = ispfc_fabric_map_deallocate_id(cb, dev->mapping_id);
			dev->mapping_id = -1;
		}

		/*
		 * Next state is LOGGED_OUT or INVALIDATION_PENDING,
		 * depending on the setting of the
		 * ISPFC_DF_INVALIDATION_REQUESTED action flag.  
		 */
		if (dev->dev_state & ISPFC_DF_INVALIDATION_REQUESTED) {
			/* 
			 * If no pending busy descriptors, we can invalidate
			 * the device immediately. Else we need to wait for
			 * pending descriptors to "trickle" out.
			 */
			if (dev->scsi_desc_busy == NULL) {
				ispfc_dev_t *		new_dev;
				uint32			new_phys_dev_id;
				channel_name_t		ch;
				int			rc;

				/*
				 * If we're asked to reinit the device,
				 * we need to create a new instance
				 * with the same physical ID.  
				 */
				if (dev->dev_state & ISPFC_DF_REINIT_REQUESTED) {
					new_phys_dev_id = dev->phys_dev_id;
					WWN_SET(&ch.channel_wwn.u, &dev->channel_wwn);
					rc = ispfc_channel_lookup(&ch);
					VERIFY(rc == NFX_OK);
					ispfc_device_invalidate(cb, dev);

					new_dev = ispfc_device_lookup_by_phys_dev(cb, new_phys_dev_id);
					VERIFY(new_dev == NULL);
					new_dev = ispfc_device_new(cb, new_phys_dev_id, &ch);
					VERIFY(dev);
				}
				else 
					ispfc_device_invalidate(cb, dev);
			}
			else {
				SET_DEV_STATE(dev, ISPFC_DS_STATE_INVALIDATE_PENDING);
			}
		}
		else {
			/* Move to LOGGED_OUT state */
			SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGGED_OUT);
		}
	}
}

int32
ispfc_dev_deferred_login_timer_done_callout(int32 timer_handle, ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	int32 rc;

        VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_DEFERRED_LOGIN);
        VERIFY(dev->deferred_login_timer_handle == timer_handle);
        rc = eee_timerDestroy(dev->deferred_login_timer_handle);
	VERIFY(rc == NFX_OK);
        dev->deferred_login_timer_handle = 0;

        SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);

	return(NFX_OK);
}

int32 
ispfc_dev_paused_timer_done_callout(int32 timer_handle, ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	int32 rc;

        VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_PAUSED);
        VERIFY(dev->paused_timer_handle == timer_handle);
        rc = eee_timerDestroy(dev->paused_timer_handle);
	VERIFY(rc == NFX_OK);
        dev->paused_timer_handle = 0;

	SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);

	return(NFX_OK);
}

void
ispfc_dev_get_port_database_done_callout(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_t *		dev = cookie;
	ispfc_port_database_t * port_db = dev->port_database;
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
	char *			simdir = NULL;
#endif

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_GPDB_PENDING ||
	       DEV_STATE(dev) == ISPFC_DS_STATE_ID_VERIFICATION_PENDING);

#if defined(ISPFC_SIM)
	port_db->master_state = PORT_LOGGED_IN;
#endif

	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else
	if (status == MBOX_CMD_STATUS_FAILED ||
	    port_db->master_state != PORT_LOGGED_IN) {
		SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);
	}
	else {
#if !defined(ISPFC_SIM)
		/* Get WWNs from data base entry and register device */
		WWN_SET(&dev->dev_nodename, &port_db->nodename);
		WWN_SET(&dev->dev_portname, &port_db->portname);

		if (!(port_db->PRLI_service_params_3 & 0x10)) {
			FC_DBG(0, ("%s: portname 0x%llx, nodename 0x%llx, PRLI len %d service parms 0x%04x 0x%04x\n", 
				   cb->debug_name,
				   dev->dev_portname.u.wwn, dev->dev_nodename.u.wwn,
				   port_db->PRLI_payload_len, port_db->PRLI_service_params_0, port_db->PRLI_service_params_3));
			FC_DBG(0, ("%s: invalidating initiator ....\n", cb->debug_name));
			SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);
			dev->dev_state |= ISPFC_DF_INVALIDATION_REQUESTED;
			return;
		}
#else
		dev->dev_portname.u.wwn = 0x2000000000000000 | dev->phys_dev_id;
		dev->dev_nodename.u.wwn = 0x2000000000000000 | dev->phys_dev_id;

		dev->sim_dev_type = ((dev->phys_dev_id & 0x00FF0000) >> 16) >= 2; 

#if 0
		/* Simulate an inplace device swap */
		if (DEV_STATE(dev) == ISPFC_DS_STATE_ID_VERIFICATION_PENDING)
			dev->dev_old_portname.u.wwn++;
#endif

#endif /* !defined(ISPFC_SIM) */

		/* Verify WWN if necessary */
		if (DEV_STATE(dev) == ISPFC_DS_STATE_ID_VERIFICATION_PENDING &&
		    !WWN_EQUAL(&dev->dev_portname, &dev->dev_old_portname)) {
			/* 
			 * Someone swapped a device in place. Mark the
			 * current instance for invalidation and
			 * indicate a new instance is to be created.
			 */
			SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);
			dev->dev_state |= ISPFC_DF_INVALIDATION_REQUESTED;
			dev->dev_state |= ISPFC_DF_REINIT_REQUESTED;
		}
		else {
#if defined(ISPFC_SIM)
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
			simdir = getenv("SIMDIR");
			if (dev->sim_dev_type == 0) {
				sprintf(dev->scsi_vendor_id,  "AGILESTOR");
				sprintf(dev->scsi_product_id, "LINUX DISK 2001 ");
				sprintf(dev->scsi_revision,   "42  ");
				sprintf(dev->scsi_serialno,   "%08x", dev->phys_dev_id);
				dev->capacity = ISPFC_SIM_DISK_CAPACITY;
				dev->sector_size = ISPFC_SIM_DISK_SECTOR_SIZE;
				sprintf(dev->filename, ISPFC_SIM_DISK_FMT, 
					simdir ? simdir : "/tmp", 
					dev->sector_size,
					dev->dev_portname.u.wwn);
				/* Open the sim disk file */
				if ((dev->fd = open(dev->filename, O_RDWR)) == -1) {
					dev->fd = open(dev->filename, O_RDWR | O_CREAT|O_TRUNC, 0644);
					if (dev->fd == -1)
						panic("open failed on %s: %s\n", dev->filename, strerror(errno));
					VERIFY(dev->fd > 0);
					lseek(dev->fd, dev->capacity * dev->sector_size-8, SEEK_SET);
					write(dev->fd, dev->scsi_vendor_id, 8);
				}
			}
			else {
				sprintf(dev->scsi_vendor_id,  "AGILESTOR");
				sprintf(dev->scsi_product_id, "LINUX TAPE 2001 ");
				sprintf(dev->scsi_revision,   "42  ");
				sprintf(dev->scsi_serialno,   "%08x", dev->phys_dev_id);
				dev->fixed_blocksize = ISPFC_SIM_TAPE_DEFAULT_BLOCKSIZE;
				dev->max_bytes_to_eod = ISPFC_SIM_TAPE_MAX_BYTES;
				dev->compression_enabled = 0;
				dev->write_protected = WWN_EQUAL(&dev->dev_portname, &ispfc_write_protected_device);
				
				/*
				 * Assume device starting from powerup
				 * and hence tape is positioned on
				 * first block of first file.  
				 */
				dev->dev_state |= ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE;
				dev->current_fileno = 0;
				dev->current_blockno = 0;
				dev->current_data_fd = -1;
				dev->current_dir_fd = -1;
				dev->current_dir = NULL;
			}
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
			sprintf(dev->scsi_vendor_id,  "AGILESTOR");
			sprintf(dev->scsi_product_id, "RAM DISK 2001   ");
			sprintf(dev->scsi_revision,   "42  ");
			sprintf(dev->scsi_serialno,   "%08x", dev->phys_dev_id);
			dev->capacity = 1024; /* 0.5MB */
			dev->sector_size = 512;
			/* Allocate space for sim RAM disk */
			dev->ram_disk_buf = eee_ramAlloc(dev->capacity * dev->sector_size);
			if (dev->ram_disk_buf == NULL) {
				panic("failed to allocate space for RAM disk\n");
			}
#endif
#endif /* defined(ISPFC_SIM) */
			if (!(dev->dev_state & ISPFC_DF_REGISTERED))
				ispfc_device_register(cb, dev);
			SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);
		}

		/* Sync device */
		ispfc_send_marker(cb, dev, 0, MARKER_SYNC_TARGET);
	}
}

int32 
ispfc_timeout_handler(int32 timer_handle, ispfc_cb_t *cb)
{
	ispfc_dev_t *	dev;
	ispfc_dev_t *	next_dev;
	scsi_desc_t *	sd;
	scsi_desc_t *	next_sd;

	/* If link not up, don't bother timing out SCSI commands */
	if (LINK_STATE(cb) != ISPFC_CLS_STATE_UP)
		return(NFX_OK);

	/* FOR_ALL_DEVS({ */
	if ((dev = cb->dev_list)) {
		do {
			next_dev = dev->next;

			if (DEV_STATE(dev) == ISPFC_DS_STATE_UP ||
			    DEV_STATE(dev) == ISPFC_DS_STATE_QUIESCED) {

				if ((sd = dev->scsi_desc_busy)) {
					/* 
					 * Bump global device timer. If it's
					 * popped, no forward progress is being
					 * made on the device. Initiate timeout
					 * handling.
					 */
					if (dev->dev_global_timeout_timer-- == 0) {
						VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_QUIESCED);
						FC_DBG(0, ("%s: phys_dev_id 0x%x, handle %d: global device timer timeout, "
							   "initiating device recovery.\n",
							   cb->debug_name, dev->phys_dev_id, dev->dev_handle));
						
						/* Move all busy descriptors to TIMED OUT state. */
						do {	   
							sd->state = SCSI_DESC_STATE_TIMED_OUT;
							sd = sd->next; 
						} while (sd != dev->scsi_desc_busy);
						ispfc_cmd_timeout(cb, dev, sd);
						goto next_dev;
					}

					do {
						next_sd = sd->next;

						VERIFY(sd->state == SCSI_DESC_STATE_BUSY);

						/* 
						 * Do nothing if no timeout, or
						 * if this command has already
						 * timed out in which case we
						 * better already be quiesced.  
						 */
						if (sd->current_timeout_count < 0 || 
						    sd->current_timeout_count-- != 0 || 
						    DEV_STATE(dev) == ISPFC_DS_STATE_QUIESCED)
							continue;
						VERIFY(sd->current_timeout_count >= 0 || 
						       DEV_STATE(dev) != ISPFC_DS_STATE_QUIESCED);

						/* 
						 * A command timer has popped
						 * ... quiesce the device to
						 * allow command to complete, in
						 * case it's being starved.
						 */
						FC_DBG(0, ("%s: phys_dev_id 0x%x, handle %d (%p) cdb %s: command timeout, "
							   "quiescing drive to allow outstanding I/O to complete.\n",
							   cb->debug_name, dev->phys_dev_id, dev->dev_handle,
							   sd, ispfc_scsi_cdb_to_string(sd->cdb)));
						SET_DEV_STATE(dev, ISPFC_DS_STATE_QUIESCED);
					} while ((sd = next_sd) != dev->scsi_desc_busy);
				}
			}
		next_dev:
			dev = next_dev;
		} while (LINK_STATE(cb) == ISPFC_CLS_STATE_UP && dev != cb->dev_list);
	}
	/* }); */

	return(NFX_OK);
}


ispfc_dev_t*
ispfc_device_new(ispfc_cb_t *cb, uint32 phys_dev_id, channel_name_t *ch)
{
        ispfc_dev_t   *dev;

#if 0	/* XXX RAM allocator not aligning to double-word addresses */
        dev = (ispfc_dev_t *)eee_ramAlloc(sizeof(ispfc_dev_t));
#else
        dev = HAL_ALIGN_TO(eee_ramAlloc(sizeof(ispfc_dev_t) + sizeof(uint64)), sizeof(uint64));
#endif

        bzero(dev, sizeof(ispfc_dev_t));
        VERIFY(dev);
	dev->port_database = (ispfc_port_database_t *)eee_ramAlloc(sizeof(ispfc_port_database_t));
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	dev->port_database = (void *)cached_to_uncached((address_t)dev->port_database);
#else
	(ps_uint_t)dev->port_database |= 0x20000000L; /* Turn on uncached access */
#endif
#endif
	VERIFY(dev->port_database);
	dev->cb = cb;
        dev->phys_dev_id = phys_dev_id;
        if (ch)
                WWN_SET(&dev->channel_wwn, &ch->channel_wwn);

	/*
	 * Allocate handle for device
	 */
	dev->dev_handle = ispfc_device_handle_allocate(cb);
	VERIFY(dev->dev_handle >= 0);

	/* Loop devices implicitly mapped */
	if (cb->interface_type == HAL_IF_FC_L_PORT ||
	    cb->interface_type == HAL_IF_FC_N_PORT) {
		dev->dev_state |= ISPFC_DF_MAPPING_ID_ALLOCATED;
		dev->mapping_id = ispfc_alpa_to_loopid(phys_dev_id);
	}

	/* 
	 * Synchronize generate number with that of the adapter control block 
	 */
	dev->dev_generation = cb->dev_generation;

	/*
	 * Set initial state machine state
	 */
	SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);

        /*
         * Allocate device hash tables on first dev alloc
         */
        if (cb->dev_count == 0) {
                if (cb->interface_type == HAL_IF_FC_L_PORT || cb->interface_type == HAL_IF_FC_N_PORT)
                        cb->dev_physid_hash = ispfc_hash_table_alloc(ISPFC_HASH_TYPE1);
                else
			cb->dev_physid_hash = ispfc_hash_table_alloc(ISPFC_HASH_TYPE2);
		cb->dev_handle_hash = ispfc_hash_table_alloc(ISPFC_HASH_TYPE2);
        }

        /*
         * Insert dev context into list
         */
        if (cb->dev_list == NULL) {
                /* Stick at head of list */
                cb->dev_list = dev;
                dev->next = dev->prev = dev;
        }
        else {
                /* Stick at end of list */
                dev->next = (ispfc_dev_t *)cb->dev_list;
                dev->prev = ((ispfc_dev_t *)cb->dev_list)->prev;
                ((ispfc_dev_t *)cb->dev_list)->prev->next = dev;
                ((ispfc_dev_t *)cb->dev_list)->prev = dev;
        }

        /*
         * Insert dev context into physID hash table
         */
        if (ispfc_hash_table_obj_find(cb->dev_physid_hash, dev->phys_dev_id))
		PANIC("ispfc: Attempting to add physid 0x%x a second time.\n", dev->phys_dev_id);
        ispfc_hash_table_obj_insert(cb->dev_physid_hash, dev->phys_dev_id, (ispfc_hash_obj_t)dev);

        /*
         * Insert dev context into handle hash table
         */
        if (ispfc_hash_table_obj_find(cb->dev_handle_hash, dev->dev_handle))
		PANIC("ispfc: Attempting to add handle 0x%x a second time.\n", dev->dev_handle);
        ispfc_hash_table_obj_insert(cb->dev_handle_hash, dev->dev_handle, (ispfc_hash_obj_t)dev);

        cb->dev_count++;

        return(dev);
}

ispfc_dev_t *
ispfc_device_lookup_by_phys_dev(ispfc_cb_t *cb, uint32 phys_dev_id)
{
        if (cb->dev_physid_hash)
                return(ispfc_hash_table_obj_find(cb->dev_physid_hash, phys_dev_id));
        else
                return(NULL);
}

ispfc_dev_t *
ispfc_device_lookup_by_handle(ispfc_cb_t *cb, uint32 dev_handle)
{
        if (cb->dev_handle_hash)
                return(ispfc_hash_table_obj_find(cb->dev_handle_hash, dev_handle));
        else
                return(NULL);
}

ispfc_dev_t *
ispfc_device_lookup_by_channel_and_loopid(ispfc_cb_t *cb, wwn_t channel_wwn, int loop_id)
{
        channel_name_t  ch;
        int             rc;
        uint32          phys_dev;

        if (channel_wwn.u.wwn == null_wwn.u.wwn ||
            channel_wwn.u.wwn == cb->adapter_wwn.u.wwn)
                return(ispfc_device_lookup_by_phys_dev(cb, loop_id));
        else {
                ch.channel_wwn.u.wwn = channel_wwn.u.wwn;
                rc = ispfc_channel_lookup(&ch);
                if (rc == NFX_ERR)
                        return(NULL);
		/* VERIFY(PHYS_DEV_TO_FLA(ch.u.fc_f_or_fc_fl.channel_fl_port) != 0); */
                phys_dev = ch.u.fc_f_or_fc_fl.channel_fl_port | ispfc_loopid_to_alpa(loop_id);
                return(ispfc_device_lookup_by_phys_dev(cb, phys_dev));
        }
}

ispfc_dev_t * 
ispfc_device_lookup_by_portname(ispfc_cb_t *cb, wwn_t portname)
{
	ispfc_dev_t *	dev;

	FOR_ALL_DEVS({
		if (WWN_EQUAL(&dev->dev_portname, &portname))
			return(dev);
	});
	return(NULL);
}

void
ispfc_device_free(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	int	rc;

        VERIFY(cb->dev_count > 0);

        /*
         * Delete dev context from list
         */
        dev->prev->next = dev->next;
        dev->next->prev = dev->prev;

        cb->dev_list = dev->next;

        dev->next = dev->prev = NULL;

        if (cb->dev_list == dev)
                cb->dev_list = NULL;

        /*
         * Free it from physid hash table
         */
        if (ispfc_hash_table_obj_find(cb->dev_physid_hash, dev->phys_dev_id) == NULL)
		PANIC("ispfc: Attempting to free a physid 0x%x which isn't in hash table.\n", dev->phys_dev_id);
        ispfc_hash_table_obj_delete(cb->dev_physid_hash, dev->phys_dev_id, dev);

        /*
         * Free it from handle hash table
         */
        if (ispfc_hash_table_obj_find(cb->dev_handle_hash, dev->dev_handle) == NULL)
		PANIC("ispfc: Attempting to free a handle 0x%x which isn't in hash table.\n", dev->dev_handle);
        ispfc_hash_table_obj_delete(cb->dev_handle_hash, dev->dev_handle, dev);

	/*
	 * Deallocate handle for device
	 */
	rc = ispfc_device_handle_deallocate(cb, dev->dev_handle);
	VERIFY(rc == NFX_OK);

	/*	   
	 * Free structures associated with device
	 */
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
	dev->port_database = (void *)uncached_to_cached((address_t)dev->port_database);
#else
	(ps_uint_t)dev->port_database &= ~0x20000000L; /* Turn off uncached access */
#endif
#endif
	eee_ramDealloc((void *)dev->port_database);
	eee_ramDealloc((void *)dev);

        /*
         * Free hash tables on last free
         */
        if (--cb->dev_count == 0) {
                ispfc_hash_table_free(cb->dev_physid_hash);
                cb->dev_physid_hash = NULL;
                ispfc_hash_table_free(cb->dev_handle_hash);
                cb->dev_handle_hash = NULL;
        }
}

int32
ispfc_device_invalidate(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	scsi_desc_t *	sd;

	SET_DEV_STATE(dev, ISPFC_DS_STATE_INVALID);

	/*
	 * Commands busy/pending? There better not be.
	 */
	VERIFY(dev->dev_cmd_busy_count == 0 && dev->scsi_desc_busy == NULL);

	/* 
	 * This fabric device better not be mapped
	 */
	VERIFY(cb->interface_type == HAL_IF_FC_L_PORT || 
	       cb->interface_type == HAL_IF_FC_N_PORT ||
	       !(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED));

	/*
	 * Flush waiting queue.
	 */
	while ((sd = ispfc_dequeue_desc_off_dev_waiting_queue(cb, dev))) {
		sd->prev = sd->next = NULL;
		SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_INVALID_DEVICE);
		SCSI_RECEIVE(sd);
	}
	
	/* 
	 * Delete device path instance from DCS/DCA.
	 */
	if (dev->dev_state & ISPFC_DF_REGISTERED)
		ispfc_device_deregister(cb, dev);

	ispfc_device_free(cb, dev);

	return(NFX_OK);
}

int
ispfc_device_register(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	char	channel_buf[WWN_STR_LEN];
	char	nodename_buf[WWN_STR_LEN];
	int	rc;
	wwn_t	aligned_dev_nodename;
	wwn_t	aligned_channel_wwn;

	FC_DBG(1, ("%s: Registering device: phys_dev_id 0x%x, handle 0x%x, nodename %s, channel_wwn %s\n",
		   cb->debug_name, dev->phys_dev_id, dev->dev_handle, 
		   wwn_to_str(&dev->dev_nodename, nodename_buf), 
		   wwn_to_str(&dev->channel_wwn, channel_buf)));
	WWN_SET(&aligned_dev_nodename, &dev->dev_nodename);
	WWN_SET(&aligned_channel_wwn, &dev->channel_wwn);
#if defined(LITTLE_ENDIAN_EVAL_BOARD) || defined(AGILE_HARDWARE)
	inplace_ntohll(&aligned_dev_nodename);
	inplace_ntohll(&aligned_channel_wwn);
#endif
	rc = SCSI_DEVICE_NOTIFY(SCSI_LLNOTIFY_DEV_DETECTED,
				SCSI_LLNOTIFY_FC,
				dev->dev_handle,
				aligned_dev_nodename.u.wwn, 
				aligned_channel_wwn.u.wwn, 
				NULL,
				0);
	VERIFY(rc == NFX_OK);

	dev->dev_state |= ISPFC_DF_REGISTERED;
	return(NFX_OK);
}

int
ispfc_device_deregister(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	char	channel_buf[WWN_STR_LEN];
	char	nodename_buf[WWN_STR_LEN];
	int	rc;
	wwn_t	aligned_dev_nodename;
	wwn_t	aligned_channel_wwn;

	FC_DBG(1, ("%s: Deregistering device: phys_dev_id 0x%x, handle 0x%x, nodename %s, channel_wwn %s\n",
		   cb->debug_name, dev->phys_dev_id, dev->dev_handle, 
		   wwn_to_str(&dev->dev_nodename, nodename_buf), 
		   wwn_to_str(&dev->channel_wwn, channel_buf)));
	WWN_SET(&aligned_dev_nodename, &dev->dev_nodename);
	WWN_SET(&aligned_channel_wwn, &dev->channel_wwn);
#if defined(LITTLE_ENDIAN_EVAL_BOARD) || defined(AGILE_HARDWARE)
	inplace_ntohll(&aligned_dev_nodename);
	inplace_ntohll(&aligned_channel_wwn);
#endif
	rc = SCSI_DEVICE_NOTIFY(SCSI_LLNOTIFY_DEV_REMOVED,
				SCSI_LLNOTIFY_FC,
				dev->dev_handle,
				aligned_dev_nodename.u.wwn, 
				aligned_channel_wwn.u.wwn, 
				NULL,
				0);
	VERIFY(rc == NFX_OK);

	dev->dev_state &= ~ISPFC_DF_REGISTERED;
	return(NFX_OK);
}

int32 
ispfc_device_pause(ispfc_cb_t *cb, ispfc_dev_t *dev, int delay_secs)
{
	int32	rc;

	if (DEV_STATE(dev) == ISPFC_DS_STATE_UP) {
		/* Start a timer to pop in delay_secs seconds. */
		dev->paused_timer_handle = eee_timerCreate(delay_secs * MSEC_PER_SEC,
							   EEE_TIMER_ONCE);
		VERIFY(dev->paused_timer_handle >= 0);
		rc = eee_timerStart(dev->paused_timer_handle,
				    (void *)ispfc_dev_paused_timer_done_callout,
				    cb, dev, NULL, NULL);
		VERIFY(rc == NFX_OK);
		SET_DEV_STATE(dev, ISPFC_DS_STATE_PAUSED);
	}
	return(NFX_OK);
}

int32 
ispfc_device_login(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	if (DEV_STATE(dev) == ISPFC_DS_STATE_UP) {
		SET_DEV_STATE(dev, ISPFC_DS_STATE_LOGIN_REQUESTED);
	}
	return(NFX_OK);
}


/*
 * Timeout escallation strategy:
 *  1st timeout (current_to_retry_count = 0)
 *	abort taskset
 *  2nd timeout (current_to_retry_count = 1)
 *      reset device
 *  3rd timeout (current_to_retry_count = 2)
 *	reset adapter
 *  nth timeout (timeout_retry_count > 2)
 *	reset device
 *	reset adapter
 *      give up all hope ... this device is failing
 */

void
ispfc_cmd_timeout(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	switch (sd->current_to_retry_count) {
	case 0:
		FC_DBG(0, ("%s: Level 0 timeout - phys_dev_id 0x%x, handle %d: Aborting taskset: "
			   "(%p,%s,%d/%d)\n",
			   cb->debug_name, dev->phys_dev_id, dev->dev_handle,
			   sd, ispfc_scsi_cdb_to_string(sd->cdb),
			   sd->current_timeout_count, sd->current_to_retry_count));
		ispfc_dev_abort_taskset(cb, dev, (sd->path & SCSI_PATH_LUN_MASK) >> SCSI_PATH_LUN_SHIFT);
		break;
	case 1:
		FC_DBG(0, ("%s: Level 1 timeout - phys_dev_id 0x%x, handle %d: Resetting device: "
			   "(%p,%s,%d/%d)\n",
			   cb->debug_name, dev->phys_dev_id, dev->dev_handle,
			   sd, ispfc_scsi_cdb_to_string(sd->cdb),
			   sd->current_timeout_count, sd->current_to_retry_count));
		SET_DEV_STATE(dev, ISPFC_DS_STATE_RESET_REQUESTED);
		break;

	case 2:
		FC_DBG(0, ("%s: Level 2 timeout - phys_dev_id 0x%x, handle %d: Resetting adapter: "
			   "(%p,%s,%d/%d)\n",
			   cb->debug_name, dev->phys_dev_id, dev->dev_handle,
			   sd, ispfc_scsi_cdb_to_string(sd->cdb),
			   sd->current_timeout_count, sd->current_to_retry_count));
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		break;

	default:
		FC_DBG(0, ("%s: Level n timeout - phys_dev_id 0x%x, handle %d: Resetting adapter and scheduling device reset: "
			   "(%p,%s,%d/%d)\n",
			   cb->debug_name, dev->phys_dev_id, dev->dev_handle,
			   sd, ispfc_scsi_cdb_to_string(sd->cdb),
			   sd->current_timeout_count, sd->current_to_retry_count));
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		SET_DEV_STATE(dev, ISPFC_DS_STATE_RESET_REQUESTED);
		/* XXX Link reset probably moves dev state to
                   login_requested, wipping out the reset requested;
                   need to make some changes ..... */
		break;
	}

	sd->current_to_retry_count++;

	return;
}

void
ispfc_dev_abort_taskset(ispfc_cb_t *cb, ispfc_dev_t *dev, byte lun)
{
	scsi_task_mgmt_t *tmo;

	VERIFY(dev->scsi_task_mgmt_busy == NULL);

	tmo = (scsi_task_mgmt_t *)eee_ramAlloc(sizeof(scsi_task_mgmt_t));
	bzero(tmo, sizeof(scsi_task_mgmt_t));

	tmo->hal_handle = dev->dev_handle;
	tmo->lun = lun;
	tmo->command = SCSI_TMO_ABORT_TASKSET;
	tmo->callback = (void (*)(scsi_task_mgmt_t *))eee_ramDealloc;
	tmo->priv_callback = ispfc_dev_abort_taskset_0;

	ispfc_queue_tmo_on_dev_waiting_queue(cb, dev, tmo);
	SET_DEV_STATE(dev, ISPFC_DS_STATE_TMO_REQUESTED);
}

/*******************************************************************************
 * ispfc_dev_abort_taskset_[0/1]
 * 
 *  Runs the device ABORT TASKSET state machine.
 *
 *	   State 0:
 *		 TIMEOUT --> reset adapter
 *               o Start Queue Flush timer
 *		 --> state 1
 *	   State 1:
 *		 o if (busy queue empty)
 *		   --> set device state to UP
 *		 else
 *		   --> reset adapter
 *
 ******************************************************************************/
void
ispfc_dev_abort_taskset_0(ispfc_cb_t *cb, void *cookie, int32 status)
{
	scsi_task_mgmt_t * tmo = cookie;
	ispfc_dev_t *	   dev = ispfc_hash_table_obj_find(cb->dev_handle_hash, tmo->hal_handle);

	FC_DBG(1, ("%s: ispfc_dev_abort_taskset_0: status = %d\n", 
		   cb->debug_name, status));

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_TMO_PENDING);
	VERIFY(dev->scsi_task_mgmt_busy == tmo);

	dev->scsi_task_mgmt_busy = NULL;
	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		FC_DBG(1, ("%s: ispfc_dev_abort_taskset_0: Starting queue flush timer\n",
			   cb->debug_name));

		dev->queue_flush_timer_handle = eee_timerCreate(DEVICE_QUEUE_FLUSH_TIMER, EEE_TIMER_ONCE);
		VERIFY(dev->queue_flush_timer_handle >= 0);
		eee_timerStart(dev->queue_flush_timer_handle, (void *)ispfc_dev_abort_taskset_1,
			       cb, dev, tmo, NULL);
		SET_DEV_STATE(dev, ISPFC_DS_STATE_QUEUE_FLUSH_PENDING);
	}
}

void
ispfc_dev_abort_taskset_1(int32 timer_handle, ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_task_mgmt_t *tmo)
{
	int rc;

	VERIFY(timer_handle == dev->queue_flush_timer_handle);
	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_QUEUE_FLUSH_PENDING);
	rc = eee_timerDestroy(dev->queue_flush_timer_handle);
	VERIFY(rc == NFX_OK);
	dev->queue_flush_timer_handle = 0;

	if (dev->scsi_desc_busy) {
		FC_DBG(0, ("%s: ispfc_dev_abort_taskset_1: Queue flush timer expired ... resetting adapter.\n",
			   cb->debug_name));
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		ispfc_send_marker(cb, dev, tmo->lun, MARKER_SYNC_LUN);
		SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);
	}

	tmo->status = NFX_OK;
	(*tmo->callback)(tmo);
}

/*******************************************************************************
 * ispfc_dev_reset_device_[0/1]
 * 
 *  Runs the device reset state machine.
 *
 *	   State 0:
 *		 TIMEOUT --> reset adapter
 *               o Start Queue Flush timer
 *		 --> state 1
 *	   State 1:
 *		 o if (busy queue empty)
 *		   --> set device state to UP
 *		 else
 *		   --> reset adapter
 *
 ******************************************************************************/
void
ispfc_dev_reset_device_0(ispfc_cb_t *cb, void *cookie, int32 status)
{
	ispfc_dev_t *	   dev = cookie;

	FC_DBG(1, ("%s: ispfc_dev_reset_device_0: status = %d\n", 
		   cb->debug_name, status));

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_RESET_PENDING);

	if (status == MBOX_CMD_STATUS_TIMEOUT) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		if (cb->interface_type != HAL_IF_FC_L_PORT && cb->interface_type != HAL_IF_FC_N_PORT) {
			ispfc_lfa_cb_t *	req = dev->lfa_req;
			ispfc_lfa_resp_t *	resp = dev->lfa_resp;
			
			inplace_ntohl(&resp->lfa_response_code);
			if (resp->lfa_response_code != LFA_RESPONSE_CODE_LS_ACC) {
				FC_DBG(0, ("%s: LINIT LIP(%x,%x) on remote loop failed: response code 0x%x\n", 
					   cb->debug_name,
					   req->u.linit.lfa_linit_lip_byte_3,
					   req->u.linit.lfa_linit_lip_byte_4,
					   resp->lfa_response_code));
			}
			if (resp->u.linit.lfa_linit_status != LFA_LINIT_STATUS_SUCCESS) {
				FC_DBG(0, ("%s: Failure of LINIT on remote loop. (status 0x%x)\n", 
					   cb->debug_name,
					   resp->u.linit.lfa_linit_status));
			}
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
			dev->lfa_req = (void *)uncached_to_cached((address_t)dev->lfa_req);
#else
			(ps_uint_t)dev->lfa_req &= ~0x20000000L; /* Turn off uncached access */
#endif
#endif
			eee_ramDealloc(dev->lfa_req);
			dev->lfa_req = NULL;
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#ifdef FC_USE_NEW_MEM_MAP
			dev->lfa_resp = (void *)uncached_to_cached((address_t)dev->lfa_resp);
#else
			(ps_uint_t)dev->lfa_resp &= ~0x20000000L; /* Turn off uncached access */
#endif
#endif
			eee_ramDealloc(dev->lfa_resp);
			dev->lfa_resp = NULL;
		}

		FC_DBG(1, ("%s: ispfc_dev_reset_device_0: Starting queue flush timer\n",
			   cb->debug_name));

		dev->queue_flush_timer_handle = eee_timerCreate(DEVICE_QUEUE_FLUSH_TIMER, EEE_TIMER_ONCE);
		VERIFY(dev->queue_flush_timer_handle >= 0);
		eee_timerStart(dev->queue_flush_timer_handle, (void *)ispfc_dev_reset_device_1,
			       cb, dev, NULL, NULL);
		SET_DEV_STATE(dev, ISPFC_DS_STATE_QUEUE_FLUSH_PENDING);
	}
}

void
ispfc_dev_reset_device_1(int32 timer_handle, ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	int rc;

	VERIFY(timer_handle == dev->queue_flush_timer_handle);
	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_QUEUE_FLUSH_PENDING);
	rc = eee_timerDestroy(dev->queue_flush_timer_handle);
	VERIFY(rc == NFX_OK);
	dev->queue_flush_timer_handle = 0;

	if (dev->scsi_desc_busy) {
		FC_DBG(0, ("%s: ispfc_dev_reset_device_1: Queue flush timer expired ... resetting adapter.\n",
			   cb->debug_name));
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
	}
	else {
		ispfc_send_marker(cb, dev, 0, MARKER_SYNC_TARGET);
		SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);
	}
}

/*******************************************************************************
 * ispfc_queue_desc_on_dev_busy_queue -
 *
 *   Insert descriptor into device's BUSY queue, sorted by timeout value. The
 *   device global timeout timer is adjusted to the maximum time out value
 *   of all descriptors currently pending/busy.
 ******************************************************************************/
void
ispfc_queue_desc_on_dev_busy_queue(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
        scsi_desc_t *	p;

	FC_DBG(20, ("%s: queueing SCSI desc %p on B q of device %p\n", 
		   cb->debug_name, sd, dev));

        VERIFY(sd->state == SCSI_DESC_STATE_WAITING);
        sd->state = SCSI_DESC_STATE_BUSY;
        cb->scsi_desc_busy_count++;
        dev->dev_cmd_busy_count++;

        /*
         * Insert into list sorted by timeout value.
         */
        if ((p = dev->scsi_desc_busy)) {
                while (p->next != dev->scsi_desc_busy &&
                       sd->timeOutCnt < p->timeOutCnt)
                        p = p->next;
                if (p->timeOutCnt > sd->timeOutCnt) {
                        /* Insert after p */
                        p->next->prev = sd;
                        sd->next = p->next;
                        sd->prev = p;
                        p->next = sd;
                }
                else {
                        /* Insert before p */
                        p->prev->next = sd;
                        sd->prev = p->prev;
                        sd->next = p;
                        p->prev = sd;
                        if (p == dev->scsi_desc_busy)
                                dev->scsi_desc_busy = sd;
                }
        }
        else {
                dev->scsi_desc_busy = sd;
                sd->next = sd->prev = sd;
        }

        /*
         * Reset the global timer value to the max -- we don't know in
         * which order commands get executed on drives.
         */
        dev->dev_global_timeout_timer = dev->scsi_desc_busy->timeOutCnt + DEVICE_QUIESCE_TIMER;
}

/*******************************************************************************
 * ispfc_queue_desc_on_dev_waiting_queue -
 *
 *   Insert descriptor into device's WAITING queue, sorted by arrival time; i.e. 
 *   new arrivals are inserted at the end of the queue.
 ******************************************************************************/
void
ispfc_queue_desc_on_dev_waiting_queue(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
        VERIFY(sd->state == SCSI_DESC_STATE_WAITING);

	FC_DBG(20, ("%s: queueing SCSI desc %p on W q of device %p\n", 
		   cb->debug_name, sd, dev));

	if (dev->scsi_desc_waiting == NULL) {
		sd->next = sd->prev = sd;
		dev->scsi_desc_waiting = sd;
	}
	else {
		sd->prev = dev->scsi_desc_waiting->prev;
		sd->next = dev->scsi_desc_waiting;
		dev->scsi_desc_waiting->prev->next = sd;
		dev->scsi_desc_waiting->prev = sd;
	}
}

/*******************************************************************************
 * ispfc_dequeue_desc_off_dev_busy_queue -
 *
 *   Remove descriptor from device's BUSY queue. The
 *   device global timeout timer is adjusted to the maximum time out value
 *   of all remaining descriptors currently pending/busy.
 ******************************************************************************/
void
ispfc_dequeue_desc_off_dev_busy_queue(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
        scsi_desc_t *	head;

	FC_DBG(20, ("%s: dequeueing SCSI desc %p off B q of device %p\n", 
		   cb->debug_name, sd, dev));

        VERIFY(sd->state != SCSI_DESC_STATE_WAITING);
        VERIFY(cb->scsi_desc_busy_count);
        VERIFY(dev->dev_cmd_busy_count);

        sd->state = SCSI_DESC_STATE_WAITING;
        cb->scsi_desc_busy_count--;
        dev->dev_cmd_busy_count--;

        head = dev->scsi_desc_busy;
        VERIFY(head && sd->next);
        if (head == sd) {
                if (head->next == head) {
                        VERIFY(head->prev == head);
                        dev->scsi_desc_busy = NULL;
                }
                else {
                        dev->scsi_desc_busy = head->next;
                }
        }
        /* dequeue it from the list */
        sd->prev->next = sd->next;
        sd->next->prev = sd->prev;

        sd->next = sd->prev = NULL;

        /*
         * Recompute next global timeout value. Just use the current
         * timeout value of the command at the head of the list. This
         * *should* be the one with the largest timer value.
         */
        if (dev->scsi_desc_busy)
                dev->dev_global_timeout_timer = dev->scsi_desc_busy->timeOutCnt + DEVICE_QUIESCE_TIMER;
        else
                dev->dev_global_timeout_timer = 0;
}

/*******************************************************************************
 * ispfc_dequeue_desc_off_dev_waiting_queue -
 *
 *   Remove descriptor from device's WAITING queue, from the head; i.e. 
 *   in FIFO order.
 ******************************************************************************/
scsi_desc_t *
ispfc_dequeue_desc_off_dev_waiting_queue(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	scsi_desc_t *	sd;

	if ((sd = dev->scsi_desc_waiting)) {
		sd->next->prev = sd->prev;
		sd->prev->next = sd->next;
		if (sd->next == sd)
			dev->scsi_desc_waiting = NULL;
		else
			dev->scsi_desc_waiting = sd->next;			
		sd->next = sd->prev = NULL;
		VERIFY(sd->state == SCSI_DESC_STATE_WAITING);
		FC_DBG(20, ("%s: dequeueing SCSI desc %p off W q of device %p\n", 
			   cb->debug_name, sd, dev));
	}

	return(sd);
}

/*******************************************************************************
 * ispfc_queue_tmo_on_dev_waiting_queue -
 *
 *   Insert TMO (task management operation) into device's WAITING queue, 
 *   sorted by arrival time; i.e. new arrivals are inserted at the end of the 
 *   queue.
 ******************************************************************************/
void
ispfc_queue_tmo_on_dev_waiting_queue(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_task_mgmt_t *tmo)
{
	FC_DBG(20, ("%s: queueing TMO %p on W q of device %p\n", 
		   cb->debug_name, tmo, dev));

	if (dev->scsi_task_mgmt_waiting == NULL) {
		tmo->next = tmo->prev = tmo;
		dev->scsi_task_mgmt_waiting = tmo;
	}
	else {
		tmo->prev = dev->scsi_task_mgmt_waiting->prev;
		tmo->next = dev->scsi_task_mgmt_waiting;
		dev->scsi_task_mgmt_waiting->prev->next = tmo;
		dev->scsi_task_mgmt_waiting->prev = tmo;
	}
}

/*******************************************************************************
 * ispfc_dequeue_tmo_off_dev_waiting_queue -
 *
 *   Remove TMO(task management operation) from device's WAITING queue, from 
 *   the head; i.e. in FIFO order.
 ******************************************************************************/
scsi_task_mgmt_t * 
ispfc_dequeue_tmo_off_dev_waiting_queue(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	scsi_task_mgmt_t *	tmo;

	if ((tmo = dev->scsi_task_mgmt_waiting)) {
		tmo->next->prev = tmo->prev;
		tmo->prev->next = tmo->next;
		if (tmo->next == tmo)
			dev->scsi_task_mgmt_waiting = NULL;
		else
			dev->scsi_task_mgmt_waiting = tmo->next; 
		tmo->next = tmo->prev = NULL;
		FC_DBG(20, ("%s: dequeueing TMO %p off W q of device %p\n", 
			   cb->debug_name, tmo, dev));
	}

	return(tmo);
}

#ifdef DEBUG
void	
ispfc_dev_check_state_machine(ispfc_dev_t *dev, int new_state)
{
	char *		state_string;
	ispfc_cb_t *	cb = dev->cb;

	VERIFY((new_state & ~ISPFC_DS_STATE_MASK) == 0);
	/* Sanity check dev state machine */
	switch (DEV_STATE(dev)) {
	case ISPFC_DS_STATE_LOGIN_PENDING:
		break;
	case ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING:
		break;
	case ISPFC_DS_STATE_LOGIN_REQUESTED:
		break;
	case ISPFC_DS_STATE_GPDB_PENDING:
		break;
	case ISPFC_DS_STATE_DEFERRED_LOGIN:
		break;
	case ISPFC_DS_STATE_LOGOUT_PENDING:
		break;
	case ISPFC_DS_STATE_LOGGED_OUT:
		break;
	case ISPFC_DS_STATE_INVALIDATE_PENDING:
		break;
	case ISPFC_DS_STATE_UP:
		break;
	case ISPFC_DS_STATE_PAUSED:
		break;
	case ISPFC_DS_STATE_QUIESCED:
		break;
	case ISPFC_DS_STATE_TMO_PENDING:
		break;
	case ISPFC_DS_STATE_TMO_REQUESTED:
		break;
	case ISPFC_DS_STATE_RESET_PENDING:
		break;
	case ISPFC_DS_STATE_RESET_REQUESTED:
		break;
	case ISPFC_DS_STATE_QUEUE_FLUSH_PENDING:
		break;
	case ISPFC_DS_STATE_ID_VERIFICATION_PENDING:
		break;
	case ISPFC_DS_STATE_INVALID:
		break;
	default:
		VERIFY(0);
	}
	switch(new_state) {
	case ISPFC_DS_STATE_LOGIN_PENDING:		state_string = "ISPFC_DS_STATE_LOGIN_PENDING"; break;
	case ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING:	state_string = "ISPFC_DS_STATE_LOGIN_CLEANUP_PENDING"; break;
	case ISPFC_DS_STATE_LOGIN_REQUESTED:		state_string = "ISPFC_DS_STATE_LOGIN_REQUESTED"; break;
	case ISPFC_DS_STATE_GPDB_PENDING:		state_string = "ISPFC_DS_STATE_GPDB_PENDING"; break;
	case ISPFC_DS_STATE_DEFERRED_LOGIN:		state_string = "ISPFC_DS_STATE_DEFERRED_LOGIN"; break;
	case ISPFC_DS_STATE_LOGOUT_PENDING:		state_string = "ISPFC_DS_STATE_LOGOUT_PENDING"; break;
	case ISPFC_DS_STATE_LOGGED_OUT:			state_string = "ISPFC_DS_STATE_LOGGED_OUT"; break;
	case ISPFC_DS_STATE_INVALIDATE_PENDING:		state_string = "ISPFC_DS_STATE_INVALIDATE_PENDING"; break;
	case ISPFC_DS_STATE_UP:				state_string = "ISPFC_DS_STATE_UP"; break;
	case ISPFC_DS_STATE_PAUSED:			state_string = "ISPFC_DS_STATE_PAUSED"; break;
	case ISPFC_DS_STATE_QUIESCED:			state_string = "ISPFC_DS_STATE_QUIESCED"; break;
	case ISPFC_DS_STATE_TMO_PENDING:		state_string = "ISPFC_DS_STATE_TMO_PENDING"; break;
	case ISPFC_DS_STATE_TMO_REQUESTED:		state_string = "ISPFC_DS_STATE_TMO_REQUESTED"; break;
	case ISPFC_DS_STATE_RESET_PENDING:		state_string = "ISPFC_DS_STATE_RESET_PENDING"; break;
	case ISPFC_DS_STATE_RESET_REQUESTED:		state_string = "ISPFC_DS_STATE_RESET_REQUESTED"; break;
	case ISPFC_DS_STATE_QUEUE_FLUSH_PENDING:	state_string = "ISPFC_DS_STATE_QUEUE_FLUSH_PENDING"; break;
	case ISPFC_DS_STATE_ID_VERIFICATION_PENDING:	state_string = "ISPFC_DS_STATE_ID_VERIFICATION_PENDING"; break;
	case ISPFC_DS_STATE_INVALID:			state_string = "ISPFC_DS_STATE_INVALID"; break;
	default:					VERIFY(0);
	}

	FC_DBG(1, ("%s: Device 0x%x: New state %s (0x%08x-->0x%08x)\n",
		   cb->debug_name, dev->phys_dev_id, state_string, dev->dev_state, new_state));
}
#endif /* DEBUG */

#ifdef DEBUG
void
ispfc_dump_devs(ispfc_cb_t *cb)
{
        ispfc_dev_t   *dev;

        /*
        "01234567890123456789012345678901234567890123456789012345678901234567890123456789
        "[0x########]:         phys_dev_id dev_handle mapping_id dev_nodename     dev_cmd_busy_count dev_state"
        "                      0x########  0x####     0x##       ################ ###                0x########
        "      scsi_desc_busy: 0x################ 0x################ ..."
        "   scsi_desc_waiting: 0x################ 0x################ ..."
        */
        if ((dev = cb->dev_list)) {
                do {
			P(
			  "[%p]:         phys_dev_id dev_handle mapping_id dev_nodename     dev_cmd_busy_count dev_state\n",
			  dev);
                        P(
			  "                      0x%08x  0x%04x     0x%02x       %s %d                  0x%08x\n",
			  dev->phys_dev_id,
			  dev->dev_handle,
			  dev->mapping_id,
			  wwn_to_str(&dev->dev_nodename, NULL),
			  dev->dev_cmd_busy_count,
			  dev->dev_state);
                        {
                                scsi_desc_t *d;

                                P(
				  "      scsi_desc_busy: ");
                                if ((d = dev->scsi_desc_busy)) {
                                        do {
                                                P("%p ", d);
                                                d = d->next;
                                        } while (d && d != dev->scsi_desc_busy);
                                }
                                P("\n");

                                P(
				  "   scsi_desc_waiting: ");
                                if ((d = dev->scsi_desc_waiting)) {
                                        do {
                                                P("%p ", d);
                                                d = d->next;
                                        } while (d && d != dev->scsi_desc_waiting);
                                }
                                P("\n");
                        }
                        dev = dev->next;
                } while (dev != cb->dev_list);
        }
}
#endif
