/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_hash.c
 *
 * RCS:		$Id: ispfc_hash.c,v 1.3 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver hash code.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/6/01
 *
 *------------------------------------------------------------------------------
 */

#include <nfx-incl.h>
#include "ispfc_private.h"
#include "ispfc_hash.h"
#include "ispfc_extern.h"

uint32 ispfc_hash_func1(ispfc_hash_key_t);
uint32 ispfc_hash_func2(ispfc_hash_key_t);
uint32 ispfc_hash_func3(ispfc_hash_key_t);

ispfc_hash_t *
ispfc_hash_table_alloc(uint32 type)
{
	ispfc_hash_t	*t;

	t = (ispfc_hash_t *)eee_ramAlloc(sizeof(ispfc_hash_t));
	bzero(t, sizeof(ispfc_hash_t));
	if (t == NULL)
		return(NULL);

	switch(type) {
	case ISPFC_HASH_TYPE1:
		t->ispfc_hash_table_size = 128;
		t->ispfc_hash_func = ispfc_hash_func1;
		break;

	case ISPFC_HASH_TYPE2:
		t->ispfc_hash_table_size = 1024;
		t->ispfc_hash_func = ispfc_hash_func2;
		break;

	case ISPFC_HASH_TYPE3:
	default:
		VERIFY(0);
	}

	t->ispfc_hash_buckets = (ispfc_hash_entry_t **)eee_ramAlloc(sizeof(ispfc_hash_t *) * t->ispfc_hash_table_size);
	bzero(t->ispfc_hash_buckets, sizeof(ispfc_hash_t *) * t->ispfc_hash_table_size);
	if (t->ispfc_hash_buckets == NULL) {
		eee_ramDealloc((void *)t);
		return(NULL);
	}

	return(t);
}

void
ispfc_hash_table_free(ispfc_hash_t *t)
{
	eee_ramDealloc((void *)t->ispfc_hash_buckets);
	eee_ramDealloc((void *)t);
}

int
ispfc_hash_table_obj_insert(ispfc_hash_t *t, ispfc_hash_key_t key, ispfc_hash_obj_t obj)
{
	uint32			hash = (*t->ispfc_hash_func)(key);
	ispfc_hash_entry_t *	e;

	ASSERT(t);
	ASSERT(hash <= t->ispfc_hash_table_size);

	e = (ispfc_hash_entry_t *)eee_ramAlloc(sizeof(ispfc_hash_entry_t));
	bzero(e, sizeof(ispfc_hash_entry_t));
	if (e == NULL)
		return(NFX_ERR);
	e->key = key;
	e->object = obj;

	if (t->ispfc_hash_buckets[hash] == NULL) {
		/* Stick at head of list */
		t->ispfc_hash_buckets[hash] = e;
		e->next = e->prev = e;
	}
	else {
		/* Stick at end of list */
		e->next = t->ispfc_hash_buckets[hash];
		e->prev = t->ispfc_hash_buckets[hash]->prev;
		t->ispfc_hash_buckets[hash]->prev->next = e;
		t->ispfc_hash_buckets[hash]->prev = e;
	}
	return(NFX_OK);
}

void
ispfc_hash_table_obj_delete(ispfc_hash_t *t, ispfc_hash_key_t key, ispfc_hash_obj_t obj)
{
	uint32			hash = (*t->ispfc_hash_func)(key);
	ispfc_hash_entry_t *	e;

	ASSERT(t);
	ASSERT(hash <= t->ispfc_hash_table_size);

	e = t->ispfc_hash_buckets[hash];
	ASSERT(e);
	do {
		if (e->key == key)
			break;
		e = e->next;
	} while (e != t->ispfc_hash_buckets[hash]);

	ASSERT(e->object == obj);
	
	e->prev->next = e->next;
	e->next->prev = e->prev;

	t->ispfc_hash_buckets[hash] = e->next;

	if (t->ispfc_hash_buckets[hash] == e)
		t->ispfc_hash_buckets[hash] = NULL;

	eee_ramDealloc((void *)e);
}

ispfc_hash_obj_t 
ispfc_hash_table_obj_find(ispfc_hash_t *t, ispfc_hash_key_t key)
{
	uint32			hash = (*t->ispfc_hash_func)(key);
	ispfc_hash_entry_t *	e;
#ifdef DEBUG
	int			count = 0;
#endif

	ASSERT(t);
	ASSERT(hash <= t->ispfc_hash_table_size);

	e = t->ispfc_hash_buckets[hash];
	if (e == NULL)
		return(NULL);

	do {
		if (e->key == key) {
#ifdef DEBUG
			if (count < ISPFC_EFFICIENCY_HIST_SIZE)
				t->ispfc_efficiency_histogram[count]++;
			else
				t->ispfc_efficiency_histogram[ISPFC_EFFICIENCY_HIST_SIZE]++;
#endif
			return(e->object);
		}
#ifdef DEBUG
		++count;
#endif
		e = e->next;
	} while (e != t->ispfc_hash_buckets[hash]);
	return(NULL);
}

void
ispfc_hash_table_dump(ispfc_hash_t *t)
{
	int			i;
	ispfc_hash_entry_t	*e;
	
	P("Hash table at %p: size %d\n", t, t->ispfc_hash_table_size);
	for (i = 0; i < t->ispfc_hash_table_size; ++i) {
		if ((e = t->ispfc_hash_buckets[i])) {
			P("[%4d]: ", i);
			do {
				P("(%p, 0x%08x) ", e->object, e->key);
				e = e->next;
			} while (e != t->ispfc_hash_buckets[i]);
			P("\n");
		}
	}
}

uint32
ispfc_hash_func1(ispfc_hash_key_t key)
{
	return(key & 0x0000007FL);
}

uint32
ispfc_hash_func2(ispfc_hash_key_t key)
{
	return((key & 0x00000F00 >> 2) | (key & 0x0000003F));
}

uint32
ispfc_hash_func3(ispfc_hash_key_t key)
{
	return(0);
}

