/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_iocb.c
 *
 * RCS:		$Id: ispfc_iocb.c,v 1.45 2002/10/30 17:48:33 nitinb Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver IOCB routines
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/12/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_iocb.h"
#include "ispfc_mbox.h"
#include "ispfc_ct.h"
#include "ispfc_device.h"
#include "ispfc_link.h"
#include "ispfc_scsi.h"

void *	 ispfc_get_iocb(ispfc_cb_t *);
void	 ispfc_send_iocb(ispfc_cb_t *);
int32	 ispfc_iocb_handle_allocate(ispfc_cb_t *, void *, uint32);
void	 ispfc_iocb_handle_deallocate(ispfc_cb_t *, int32);
void	 ispfc_iocb_handle_lookup(ispfc_cb_t *, int32, void **, uint32 *);
void	 ispfc_dump_response(ispfc_cb_t *, ispfc_iocb_status_t *);
void	 ispfc_dump_cmd2(ispfc_cb_t *, ispfc_iocb_command2_t *);
void	 ispfc_dump_cont0(ispfc_cb_t *, ispfc_iocb_cont0_t *);
void	 ispfc_dump_ct(ispfc_cb_t *, ispfc_iocb_ct_t *);

void *
ispfc_get_iocb(ispfc_cb_t *cb)
{
	ispfc_command_entry_t *e;

	if (cb->scsi_desc_busy_count + cb->ct_desc_busy_count == ISPFC_MAX_OUTSTANDING_REQUESTS)
		return(NULL);
	e = &cb->request[cb->request_in];
	INC_Q_PTR(cb->request_in, ISPFC_REQUEST_QUEUE_SIZE);
	bzero(e, sizeof(ispfc_command_entry_t));
	return(e);
}

void
ispfc_send_iocb(ispfc_cb_t *cb)
{
	ISPFC_REG_WRITE(cb, ISPFC_REG_REQUEST_IN, cb->request_in);
}

void 
ispfc_iocb_handle_init(ispfc_cb_t *cb)
{
	bzero(cb->iocb_handles, sizeof(cb->iocb_handles));
	cb->available_iocb_handles = ISPFC_MAX_OUTSTANDING_REQUESTS;
	cb->next_iocb_handle = 0;
}

int32
ispfc_iocb_handle_allocate(ispfc_cb_t *cb, void *cookie, uint32 type)
{
	int32	rc;

	VERIFY(cb->available_iocb_handles);
	while (cb->iocb_handles[cb->next_iocb_handle].alloced) {
		INC_Q_PTR(cb->next_iocb_handle, ISPFC_MAX_OUTSTANDING_REQUESTS);
	}
	cb->iocb_handles[cb->next_iocb_handle].cookie = cookie;
	cb->iocb_handles[cb->next_iocb_handle].alloced = type;
	--cb->available_iocb_handles;
	rc = cb->next_iocb_handle+1;
	INC_Q_PTR(cb->next_iocb_handle, ISPFC_MAX_OUTSTANDING_REQUESTS);
	return(rc);
}

void
ispfc_iocb_handle_deallocate(ispfc_cb_t *cb, int32 iocb_handle)
{
	VERIFY(iocb_handle > 0);
	VERIFY(cb->iocb_handles[iocb_handle-1].alloced);
	cb->iocb_handles[iocb_handle-1].alloced = 0;
	++cb->available_iocb_handles;
}

void
ispfc_iocb_handle_lookup(ispfc_cb_t *cb, int32 iocb_handle, void **cookie, uint32 *type)
{
	VERIFY(iocb_handle > 0);
	VERIFY(cb->iocb_handles[iocb_handle-1].alloced);
	*cookie = cb->iocb_handles[iocb_handle-1].cookie;
	*type = cb->iocb_handles[iocb_handle-1].alloced;
}

int
ispfc_send_marker(ispfc_cb_t *cb, ispfc_dev_t *dev, int lun, int marker_type)
{
	ispfc_iocb_marker_t *	m;

	m = ispfc_get_iocb(cb);

        m->header.entry_type = ISPFC_ENTRY_TYPE_MARKER;
        m->header.entry_cnt = 1;

	if (marker_type == MARKER_SYNC_TARGET || marker_type == MARKER_SYNC_LUN) {
		if (marker_type == MARKER_SYNC_LUN)
			m->lun = lun;
		VERIFY(dev);
		VERIFY(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED);
		m->target = dev->mapping_id;
	}
        m->modifier = marker_type;

#if defined(BIG_ENDIAN_EVAL_BOARD)
	SWAP2(&m->flags);
	SWAP2(&m->lun);
#endif

        ispfc_send_iocb(cb);

        return(NFX_OK);
}

typedef struct {
	ispfc_iocb_command2_t * ic2;
        ispfc_iocb_cont0_t *    ice;
	ispfc_dsd0_t *		next_seg;
	int			segs_remaining;
	uint32			segs_emitted;
	uint32			bytes_emitted;
} ispfc_sg_list_t;

int
ispfc_send_scsi_desc(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
#if 0
	static int		dropped_counter = 0;
#endif

	ispfc_sg_list_t 	sg_list;
	ispfc_iocb_command2_t *	ic2;
	scsi_desc_t *		sdc;
	eee_desc_t *		ed;
	eee_descBuf_t *		bd;
	eee_descBuf_t *		cs_bd;
	byte *			data_p;
	byte *			cs_p;
	int32			bufcnt;
	int32			bytecnt;
	int32			emitted_data_len;
	int32			data_in_block;
	int32			data_in_buf;
	int32			cs_blockcnt;

	void ispfc_emit_seg(uint32 addr, uint32 len)
		{ 
			sg_list.next_seg->buf = addr;
			sg_list.next_seg->len = len; 
#if defined(BIG_ENDIAN_EVAL_BOARD)
			SWAP4(&sg_list.next_seg->len);
			SWAP4(&sg_list.next_seg->buf);
#endif
			sg_list.segs_emitted++;
			sg_list.bytes_emitted += len;
			if (--sg_list.segs_remaining)
				++sg_list.next_seg;
			else {
#ifdef DEBUG
				if (ispfc_dbg_level >= 10) {
					if (sg_list.ice)
						ispfc_dump_cont0(cb, sg_list.ice);
				}
#endif
				sg_list.ice = ispfc_get_iocb(cb);
				VERIFY(sg_list.ice);
				sg_list.ice->header.entry_type = ISPFC_ENTRY_TYPE_CONTIN32;
				sg_list.ice->header.entry_cnt = 1;
				++sg_list.ic2->header.entry_cnt;
				VERIFY(sg_list.ic2->header.entry_cnt < 255);
				sg_list.segs_remaining = CONT_TYPE0_SG_CNT;
				sg_list.next_seg = &sg_list.ice->segments[0];
			}
		}

	VERIFY(DEV_STATE(dev) == ISPFC_DS_STATE_UP);
	VERIFY(dev->dev_state & ISPFC_DF_MAPPING_ID_ALLOCATED);

#if 0
	if ((++dropped_counter % 10000) == 0) {
		P("Dropping command 0x%x\n", sd);
		return(NFX_OK);
	}
#endif

	ic2 = ispfc_get_iocb(cb);
	if (ic2 == NULL)
		return(NFX_ERR);

	/* Setup IOCB header, target and LUN */
	ic2->header.entry_type = ISPFC_ENTRY_TYPE_COMMAND2;
	ic2->header.entry_cnt = 1;
	ic2->handle = ispfc_iocb_handle_allocate(cb, sd, ISPFC_ENTRY_TYPE_COMMAND2);
	VERIFY(ic2->handle != NFX_ERR);
	ic2->target = dev->mapping_id;
	ic2->lun = (sd->path & SCSI_PATH_LUN_MASK) >> SCSI_PATH_LUN_SHIFT;

	/* Setup IOCB control flags and CDB */
	if (sd->control & SD_FLAG_NOREORDER)
		ic2->control_flags = ISPFC_CONTROL_ORDER_Q;
	else
		ic2->control_flags = ISPFC_CONTROL_SIMPLE_Q;

	if (sd->status & SD_DEV_DATA_IN_EXPECT ||
	    sd->status & SD_DEV_DATA_OUT_EXPECT) {
		if (sd->status & SD_DEV_DATA_IN_EXPECT)
			ic2->control_flags |= ISPFC_CONTROL_READ;
		else
			ic2->control_flags |= ISPFC_CONTROL_WRITE;
	}
	memcpy(ic2->cdb, sd->cdb, CMD_MAX_CDB);

	/* Initialize sg_list state machine */
	sg_list.ic2 = ic2;
	sg_list.ice = NULL;
	sg_list.next_seg = &ic2->segments[0];
	sg_list.segs_remaining = CMD_TYPE2_SG_CNT;
	sg_list.segs_emitted = 0;
	sg_list.bytes_emitted = 0;

	for (sdc = sd; sdc; sdc = sdc->chain) {
		ed  = sdc->eDesc;
		while (ed) {

			if ((sdc->control & SD_OPCODE_MASK) != SD_OPCODE_READ &&
			    (sdc->control & SD_OPCODE_MASK) != SD_OPCODE_WRITE)
				CACHE_FLUSH_PACKET(ed);

			bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
			bytecnt = ed->hdr.control & EDESC_TOTAL_LEN;
			bd = &ed->bd[0];
			if (!(bd->control & EDESC_BD_CMD_IN_USE))
				++bd;
			VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
			VERIFY(bd->buf);
			data_in_buf = bd->len;
			data_p = bd->buf;
			if (sdc->control & SD_FLAG_EMBEDDED_CS) {
				cs_blockcnt = (ed->hdr.control & EDESC_TOTAL_LEN) / (ECSUM_BLOCKSIZE + ECSUM_LEN);
				cs_bd = bd + bufcnt - 1;
				VERIFY(cs_bd->control & EDESC_BD_CMD_IN_USE);
				VERIFY(cs_bd->buf);
				cs_p = cs_bd->buf + cs_bd->len - cs_blockcnt * ECSUM_LEN;
				data_in_block = ECSUM_BLOCKSIZE;
			}
			else
				data_in_block = bytecnt;

			while (bytecnt) {
				/* Emit seg for user data */
				while (data_in_block) {
					emitted_data_len = MIN(data_in_block, data_in_buf);
					ispfc_emit_seg(VA_TO_IO32(data_p), emitted_data_len);
					bytecnt -= emitted_data_len;
					data_in_block -= emitted_data_len;
					data_in_buf -= emitted_data_len;
					data_p += emitted_data_len;
					if (data_in_buf == 0 && data_in_block > 0) {
						++bd;
						VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
						VERIFY(bd->buf);
						data_in_buf = bd->len;
						data_p = bd->buf;
					}
				}

				if (!(sdc->control & SD_FLAG_EMBEDDED_CS))
					continue;
				else
					data_in_block = ECSUM_BLOCKSIZE;

				/* Emit seg for checksum */
				emitted_data_len = ECSUM_LEN;
				ispfc_emit_seg(VA_TO_IO32(cs_p), emitted_data_len);
				bytecnt -= emitted_data_len;
				cs_p += emitted_data_len;
			} /* while (bytecnt) */

			if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
				ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
				VERIFY(ed);
			}
			else
				ed = NULL;
		} /* while (ed) */
	} /* for (sdc) */

	ic2->byte_count = sg_list.bytes_emitted;
	ic2->seg_count = sg_list.segs_emitted;

#if defined(BIG_ENDIAN_EVAL_BOARD)
	SWAP2(&ic2->lun);
	SWAP2(&ic2->control_flags);
	SWAP2(&ic2->seg_count);
	SWAP4(&ic2->byte_count);
#endif
#ifdef DEBUG
	if (ispfc_dbg_level >= 10) {
		ispfc_dump_cmd2(cb, ic2);
	}
#endif

	 /* Fire it off .... */
	ispfc_send_iocb(cb);

	return(NFX_OK);
}

int 
ispfc_send_ct_desc(ispfc_cb_t *cb, ct_desc_t *ctd)
{
	ispfc_iocb_ct_t *	ic;

	ic = ispfc_get_iocb(cb);
	if (ic == NULL)
		return(NFX_ERR);

	ic->header.entry_type = ISPFC_ENTRY_TYPE_CTIOCB;
	ic->header.entry_cnt  = 1;
	ic->handle = ispfc_iocb_handle_allocate(cb, ctd, ISPFC_ENTRY_TYPE_CTIOCB);
	if (ctd->server_addr == CT_ADDRESS_NS) {
		VERIFY(NS_STATE(cb) == ISPFC_CLS_NS_STATE_MAPPED);
		ic->target = ISPFC_DIRECTORY_SERVER_PORTID;
	}
	else {
		VERIFY(MS_STATE(cb) == ISPFC_CLS_MS_STATE_MAPPED);
		ic->target = ISPFC_MANAGEMENT_SERVER_PORTID;
	}

	ic->control_flags = ISPFC_CONTROL_WRITE;
	ic->timeout = 5;
	ic->req_dsd_count = 1;
	ic->total_dsd_count = 2;

	VERIFY((ctd->e_desc->hdr.control & EDESC_TOTAL_LEN) >= 3);
	VERIFY(ctd->e_desc->bd[1].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[2].control & EDESC_BD_CMD_IN_USE);
	VERIFY(ctd->e_desc->bd[1].buf);
	VERIFY(ctd->e_desc->bd[2].buf);
	ic->req_byte_count = ctd->e_desc->bd[1].len;
	ic->resp_byte_count = ctd->e_desc->bd[2].len; 

	ic->req_dsd.buflo = VA_TO_IO32(ctd->e_desc->bd[1].buf);
	ic->req_dsd.bufhi = 0;
	ic->req_dsd.len = ctd->e_desc->bd[1].len;

	ic->resp_dsd.buflo = VA_TO_IO32(ctd->e_desc->bd[2].buf);
	ic->resp_dsd.bufhi = 0;
	ic->resp_dsd.len = ctd->e_desc->bd[2].len;

#ifdef DEBUG
	if (ispfc_dbg_level >= 10) {
		P("CT %p\n", ic);
		ispfc_dump_ct(cb, ic);
	}
#endif

#if defined(BIG_ENDIAN_EVAL_BOARD)
	SWAP2(&ic->control_flags);
	SWAP2(&ic->timeout);
	SWAP2(&ic->req_dsd_count);
	SWAP2(&ic->total_dsd_count);
	SWAP4(&ic->resp_byte_count);
	SWAP4(&ic->req_byte_count);

	SWAP4(&ic->req_dsd.buflo);
	SWAP4(&ic->req_dsd.len);
	SWAP4(&ic->resp_dsd.buflo);
	SWAP4(&ic->resp_dsd.len);
#endif

	/* Fire it off .... */
	CACHE_FLUSH(ctd->e_desc->bd[1].buf, ctd->e_desc->bd[1].buf + ctd->e_desc->bd[1].len);
	CACHE_FLUSH(ctd->e_desc->bd[2].buf, ctd->e_desc->bd[2].buf + ctd->e_desc->bd[2].len);
	ispfc_send_iocb(cb);
	
	return(NFX_OK);
}


/*******************************************************************************
 * ispfc_process_iocb_completions -
 *
 *   Handle all responses in the response ring.
 ******************************************************************************/
void
ispfc_process_iocb_completions(ispfc_cb_t *cb, int response_throttle)
{
	int			response_count = 0;
	ispfc_iocb_status_t *	is;
	byte			entry_type;
	scsi_desc_t *		sd;
	ct_desc_t *		ctd;
	ispfc_dev_t *		dev;
	int			retry;

	/* 
	 * Defensive code - check response head pointer is out of bounds.
	 */
	if (cb->response_in >= ISPFC_RESPONSE_QUEUE_SIZE) {
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		return;
	}

	/* 
	 * Defensive code - check if "jumped too far". We know how
	 * many commands are pending. If it jumps farther than the
	 * previous value plus the number of outstanding commands,
	 * then BINGO ... we have a firmware bug ... CALL QLOGIC!
	 */
#if 0
	{
		int max_response_in = cb->prev_response_in + cb->scsi_desc_busy_count + cb->ct_desc_busy_count;
				
		if (max_response_in < ISPFC_RESPONSE_QUEUE_SIZE) {
			if (cb->response_in >= cb->prev_response_in &&
			    cb->response_in <= max_response_in) 
				goto response_check_ok;
		}
		else {
			if ((cb->response_in >= cb->prev_response_in && cb->response_in < ISPFC_RESPONSE_QUEUE_SIZE) ||
			    cb->response_in <= (max_response_in % ISPFC_RESPONSE_QUEUE_SIZE))
				goto response_check_ok;
		}
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);

		P("prev_response_in 0x%x, response_in 0x%x, scsi_desc_busy_count 0x%x, ct_desc_busy_count 0x%x\n",
		  cb->prev_response_in, cb->response_in, cb->scsi_desc_busy_count, cb->ct_desc_busy_count);
		ispfc_dump_pio_trace(cb, 128);

		return;
	response_check_ok:
		;
	}
#endif
	cb->prev_response_in = cb->response_in;

	while (cb->response_out != cb->prev_response_in && response_count < response_throttle) {

		response_count++;
		is = (ispfc_iocb_status_t *) &cb->response[cb->response_out];
		INC_Q_PTR(cb->response_out, ISPFC_RESPONSE_QUEUE_SIZE);

		/* 
		 * Defensive code - mark processed status entries by
		 * overwriting the header. If we see that pattern in a
		 * status entry, we're looking at a QLogic firmware
		 * bug.
		 */
		if (*((uint32 *)&is->header) == 0xDEADBEEF) {
			FC_DBG(0, ("%s: Revisited response entry.\n", cb->debug_name));
			ispfc_dump_response(cb, is);
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
			return;
		}

		entry_type = is->header.entry_type;
		if (entry_type == ISPFC_ENTRY_TYPE_STATUS_CONT) {
			FC_DBG(1, ("%s: Ignoring status continuation response entry.\n", cb->debug_name));
#ifdef DEBUG
			if (ispfc_dbg_level >= 1) {
			    ispfc_dump_response(cb, is);
			}		
#endif
			continue;
		}
		else
		if (entry_type != ISPFC_ENTRY_TYPE_STATUS && entry_type != ISPFC_ENTRY_TYPE_CTIOCB) {
			FC_DBG(0, ("%s: Invalid response entry - 0x%x (%p)\n", cb->debug_name, entry_type, is));
			ispfc_dump_response(cb, is);
			ispfc_dump_pio_trace(cb, 128);
			ispfc_dump_registers(cb);
			continue;
		}
		if (is->header.entry_status != 0) {
			FC_DBG(0, ("%s: Bad response entry.\n", cb->debug_name));
			ispfc_dump_response(cb, is);
			continue;
		}

		retry = 0;

#ifdef DEBUG
		if (ispfc_dbg_level >= 10) {
			P("RESP %p\n", is);
			ispfc_dump_response(cb, is);
		}
#endif
		
		if (entry_type == ISPFC_ENTRY_TYPE_CTIOCB) { /* Handle CT IOCB completion */
			uint32			handle;
			uint32			expected_type;

			handle = is->handle;
			ispfc_iocb_handle_lookup(cb, handle, (void **)&ctd, &expected_type);
			if (ctd == NULL) {
				FC_DBG(0, ("%s: Bad response handle in entry.\n", cb->debug_name));
				ispfc_dump_response(cb, is);
				continue;
			}
			VERIFY(expected_type == ISPFC_ENTRY_TYPE_CTIOCB);
			ispfc_iocb_handle_deallocate(cb, handle);
			
			/* Defensive code - mark the response as "seen". */
			*((uint32 *)&is->header) = 0xDEADBEEF;

#if defined(BIG_ENDIAN_EVAL_BOARD)
			SWAP2(&is->completion_status);
			SWAP2(&is->state_flags);
			SWAP2(&is->status_flags);
#endif

			ispfc_dequeue_desc_off_ct_busy_queue(cb, ctd);

			switch (is->completion_status) {
			case ISPFC_CS_COMPLETE:
				ctd->status = CT_STATUS_OK;
				break;

			case ISPFC_CS_TIMEOUT:
				FC_DBG(0, ("%s: CT Timeout on Port 0x%x.\n",
					   cb->debug_name, ctd->server_addr));
				ctd->status = CT_STATUS_TIMEOUT;
				--ctd->retry_count;
				++retry;
				break;

			case ISPFC_CS_PORT_UNAVAIL:
			case ISPFC_CS_LOGGED_OUT:
				FC_DBG(0, ("%s: CT Port 0x%x is not logged in.\n",
					   cb->debug_name, ctd->server_addr));
				ctd->status = CT_STATUS_ERR;
				--ctd->retry_count;
				++retry;

				/* Only MS server can "logout" */
				/* But NS may be logged out if switch is down */
				//VERIFY(ctd->server_addr == CT_ADDRESS_MS);
				if (MS_STATE(cb) == ISPFC_CLS_MS_STATE_MAPPED)
					SET_MS_STATE(cb, ISPFC_CLS_MS_STATE_NOT_MAPPED);
				break;

			default:
				ctd->status = CT_STATUS_ERR;
				break;
			}

			ispfc_dispose_of_ct_desc(cb, ctd, retry);
		}
		else {					     /* Handle SCSI IOCB completion */
			uint32			handle;
			uint32			expected_type;

			handle = is->handle;
			ispfc_iocb_handle_lookup(cb, handle, (void **)&sd, &expected_type);
			if (sd == NULL) {
				FC_DBG(0, ("%s: Bad response handle in entry.\n", cb->debug_name));
				ispfc_dump_response(cb, is);
				continue;
			}
			ispfc_iocb_handle_deallocate(cb, handle);
			dev = ispfc_device_lookup_by_handle(cb, sd->halHandle);
			VERIFY(dev);

			/* Defensive code - mark the response as "seen". */
			*((uint32 *)&is->header) = 0xDEADBEEF;

#if defined(BIG_ENDIAN_EVAL_BOARD)
			SWAP2(&is->scsi_status);
			SWAP2(&is->completion_status);
			SWAP2(&is->state_flags);
			SWAP2(&is->status_flags);
			SWAP2(&is->resp_info_len);
			SWAP2(&is->sense_data_len);
			SWAP4(&is->residual);
#endif

			if (is->completion_status != ISPFC_CS_COMPLETE &&
			    is->completion_status != ISPFC_CS_DATA_UNDERRUN) {
				FC_DBG(1, ("%s: (%p) 0x%x, completion_status 0x%x, retryCnt %d\n",
					   cb->debug_name, sd, dev->phys_dev_id,
					   is->completion_status,
					   sd->retryCnt));
			}

			switch (is->completion_status) {
			case ISPFC_CS_COMPLETE:
				/* 
				 * Command completed successfully. Check if the target reported a SCSI error.
				 */
				if ((is->scsi_status & ISPFC_SS_SCSI_STATUS_MASK)) {
					SD_SET_DEV_STATUS(sd, is->scsi_status & ISPFC_SS_SCSI_STATUS_MASK);
					FC_DBG(1, ("%s: (%p) phys_dev_id 0x%x, handle %d, cdb %s, scsi_status 0x%x, retryCnt %d\n",
						   cb->debug_name, sd, dev->phys_dev_id, dev->dev_handle,
						   ispfc_scsi_cdb_to_string(sd->cdb),
						   sd->status,
						   sd->retryCnt));
					if (is->scsi_status & ISPFC_SS_SENSE_LEN_VALID) {
						FC_DBG(1, ("%s: (%p) phys_dev_id 0x%x, handle %d, cdb %s, sense %s, retryCnt %d\n",
							   cb->debug_name, sd, dev->phys_dev_id, dev->dev_handle,
							   ispfc_scsi_cdb_to_string(sd->cdb),
							   ispfc_scsi_sense_to_string(is->scsi_sense),
							   sd->retryCnt));

						/* Retriable event ? */
						switch (is->scsi_sense[2]) {
						case 0x2: /* SENSE_KEY_NOT_READY */
						case 0xb: /* SENSE_KEY_ABORTED_COMMAND */
						case 0x6: /* SENSE_KEY_UNIT_ATTENTION */
							if (is->scsi_sense[2] == 0x2 &&
							    is->scsi_sense[12] == 0x4 && 
							    is->scsi_sense[13] == 0x1) { 
								ispfc_device_pause(cb, dev, 2);
								++sd->retryCnt;
							} 
							if (is->scsi_sense[2] == 0x2 ||
							    is->scsi_sense[2] == 0xb)
								--sd->retryCnt;
							++retry;
							if (sd->retryCnt)
								break;
							
						default:
							sd->senseData = eee_ramAlloc(IOCB_SENSE_LEN);
							bzero(sd->senseData, IOCB_SENSE_LEN);
							SD_SET_SENSE(sd, is->scsi_sense, MIN(is->sense_data_len, IOCB_SENSE_LEN));
						}
					}
				}
				/* 
				 * Adjust data length if underrun.
				 */
				if (sd->eDesc && sd->expectDataLen && (is->scsi_status & ISPFC_SS_RESIDUAL_UNDER) && !retry) {
					if(sd->expectDataLen < is->residual) {
						FC_DBG(0, ("%s: ISPFC_CS_COMPLETE: Underrun (0x%x) on device 0x%x - SCSI opcode 0x%x
										scsi_status 0x%x expectDataLen = %d residual = %d\n",
									cb->debug_name, is->completion_status, dev->phys_dev_id, 
									sd->cdb[0], is->scsi_status, sd->expectDataLen, is->residual));
					}
					VERIFY(sd->expectDataLen >= is->residual);
					ispfc_edesc_set_len(sd->eDesc, sd->expectDataLen - is->residual);
				}
				if (sd->status & SD_DEV_DATA_IN_EXPECT) 
					sd->status |= SD_DEV_DATA_VALID;
				break;

			case ISPFC_CS_DMA_ERROR:
				FC_DBG(0, ("%s: DMA Error (0x%x) on device 0x%x - SCSI opcode 0x%x\n",
					   cb->debug_name, is->completion_status, dev->phys_dev_id, sd->cdb[0]));
				SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_ADAPTER_ERROR);
				break;

			case ISPFC_CS_TRANSPORT_ERROR:
			case ISPFC_CS_DATA_OVERRUN:
				FC_DBG(0, ("%s: Overrun/transport error (0x%x) on device 0x%x - SCSI opcode 0x%x\n",
					   cb->debug_name, is->completion_status, dev->phys_dev_id, sd->cdb[0]));
				/* ispfc_dump_response(cb, is); */
				SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_TRANSPORT_ERROR);
				--sd->retryCnt;
				++retry;
				break;

			case ISPFC_CS_RESET_OCCURRED:
			case ISPFC_CS_ABORTED:
				if (sd->state == SCSI_DESC_STATE_TIMED_OUT) {
					/* Command timed out; retry up to retryCnt limit */
					SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_EXCESSIVE_TIMEOUTS);
					++sd->current_to_retry_count;
					--sd->retryCnt;
					++retry;
				}
				else
				if (sd->state == SCSI_DESC_STATE_BUSY) {
					/* Command is a victim; retry but don't penalize retry counter */
					FC_DBG(1, ("%s: Victim abort (%p) phys_dev_id 0x%x, handle %d, cdb %s, scsi_status 0x%x, retryCnt %d\n",
						   cb->debug_name, sd, dev->phys_dev_id, dev->dev_handle,
						   ispfc_scsi_cdb_to_string(sd->cdb),
						   sd->status,
						   sd->retryCnt));
					SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_ABORTED);
					ispfc_send_marker(cb, dev, 0, MARKER_SYNC_TARGET);
					++retry;
				}
				else
				if (sd->state == SCSI_DESC_STATE_ABORTED) { 
					/* Command was aborted by the caller; return it. */
					FC_DBG(1, ("%s: Real abort (%p) phys_dev_id 0x%x, handle %d, cdb %s, scsi_status 0x%x, retryCnt %d\n",
						   cb->debug_name, sd, dev->phys_dev_id, dev->dev_handle,
						   ispfc_scsi_cdb_to_string(sd->cdb),
						   sd->status,
						   sd->retryCnt));
					SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_ABORTED);
				}
				else
					VERIFY(0);

				/*
				 * A RESET_OCCURRED completion implies
				 * a LIP-reset. While the ISP is
				 * logging in with all devices, all
				 * commands will be bounced
				 * back. Retrying too quickly will
				 * result in our quick retry
				 * exhaustion. We pause the device for
				 * a second or two. 
				 */
				if (is->completion_status == ISPFC_CS_RESET_OCCURRED)
					ispfc_device_pause(cb, dev, 2);
				break;

			case ISPFC_CS_TIMEOUT:
				/* We manage timeouts, so adapter should never timeout an IOCB */
				VERIFY(0);
				break;

			case ISPFC_CS_QUEUE_FULL:
				FC_DBG(0, ("%s: Queue full on device 0x%x.\n", 
					   cb->debug_name, dev->phys_dev_id));
				ispfc_device_pause(cb, dev, 2);
				SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_DEVICE_BUSY);
				--sd->retryCnt;
				++retry;
				break;

			case ISPFC_CS_DATA_UNDERRUN:
				/*
				 * Not all underrun conditions are
				 * created equal; an underrun due to
				 * too little being being transmitted
				 * by the target may be OK
				 * (e.g. INQUIRY, MODE SENSE, etc.)
				 * but an underrun due to a dropped frame is
				 * rarely OK. We treat the latter as a
				 * transport error. 
				 * 
				 * The 2300 spec is a little vague on
				 * how the adapter reports underruns
				 * when frames are dropped.  According
				 * to a trusted source:
				 * 
				 * If the residual value returned by
				 * target matches the residual
				 * expected by firmware, then
				 * "residual Under" or "residual Over"
				 * bit and "residual transfer length"
				 * fields in Status IOCB reflect
				 * values in FCP_RSP frame.
				 *
				 * If the residual value returned by target
				 * does not match the residual expected by
				 * firmware, then "residual Under" and
				 * "residual Over" bits are set to 0 and
				 * "residual transfer length" field reflects
				 * firmware's residual value.  
				 */
				if (!(is->state_flags & ISPFC_SF_XFER_DATA) ||
				    !(is->scsi_status & ISPFC_SS_RESIDUAL_UNDER)) {
					FC_DBG(0, ("%s: Underrun transport error on device 0x%x - cdb %s\n", 
						   cb->debug_name, dev->phys_dev_id, ispfc_scsi_cdb_to_string(sd->cdb)));
					SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_TRANSPORT_ERROR);
					--sd->retryCnt;
					++retry;
					cb->adap_transport_error_cnt++;
				}
				else { 
					FC_DBG(1, ("%s: (%p) Underrun on device 0x%x - cdb %s, residue %d.\n", 
						   cb->debug_name, sd, dev->phys_dev_id, ispfc_scsi_cdb_to_string(sd->cdb), is->residual));
					VERIFY(sd->expectDataLen >= is->residual);
					ispfc_edesc_set_len(sd->eDesc, sd->expectDataLen - is->residual);
					if (sd->status & SD_DEV_DATA_IN_EXPECT) 
						sd->status |= SD_DEV_DATA_VALID;
					cb->adap_underrun_cnt++; 
				}
				break;

			case ISPFC_CS_PORT_UNAVAIL:
			case ISPFC_CS_LOGGED_OUT:
			case ISPFC_CS_CONFIG_CHANGED:
				if (is->completion_status == ISPFC_CS_PORT_UNAVAIL) {
					FC_DBG(0, ("%s: Port 0x%x UNAVAILABLE \n", 
						   cb->debug_name, dev->phys_dev_id));
				}
				else 
				if (is->completion_status == ISPFC_CS_LOGGED_OUT) {
					FC_DBG(0, ("%s: Port 0x%x LOGGED OUT \n", 
						   cb->debug_name, dev->phys_dev_id));
				}
				/* Force a login with the device */
				ispfc_device_login(cb, dev);
				++retry;
				break;

			default:
				VERIFY(0);
			}

			ispfc_dequeue_desc_off_dev_busy_queue(cb, dev, sd);
			ispfc_dispose_of_scsi_desc(cb, dev, sd, retry);

			/* 
			 * If the busy queue has emptied, and 
			 * (a) device is quiesced, then unquiesce the device, or
			 * (b) device is in invalidate pending state, then invalidate the device.
			 */
			if (dev->scsi_desc_busy == NULL) {
				if (DEV_STATE(dev) == ISPFC_DS_STATE_QUIESCED) {
					SET_DEV_STATE(dev, ISPFC_DS_STATE_UP);
					FC_DBG(0, ("%s: Unquiescing device 0x%x\n", cb->debug_name, dev->phys_dev_id));
				}
				else 
				if (DEV_STATE(dev) == ISPFC_DS_STATE_INVALIDATE_PENDING) {
					ispfc_dev_t *		new_dev;
					uint32			new_phys_dev_id;
					channel_name_t		ch;
					int			rc;

					/*
					 * If we're asked to reinit the device,
					 * we need to create a new instance
					 * with the same physical ID.  
					 */
					if (dev->dev_state & ISPFC_DF_REINIT_REQUESTED) {
						new_phys_dev_id = dev->phys_dev_id;
						ch.channel_wwn.u.wwn = dev->channel_wwn.u.wwn;
						rc = ispfc_channel_lookup(&ch);
						VERIFY(rc == NFX_OK);
						ispfc_device_invalidate(cb, dev);

						new_dev = ispfc_device_lookup_by_phys_dev(cb, new_phys_dev_id);
						VERIFY(new_dev == NULL);
						new_dev = ispfc_device_new(cb, new_phys_dev_id, &ch);
						VERIFY(dev);
					}
					else 
						ispfc_device_invalidate(cb, dev);					
				}
			}

		}
	} /* while (cb->response_out != cb->prev_response_in) */

	if (response_count)
		ISPFC_REG_WRITE(cb, ISPFC_REG_RESPONSE_OUT, cb->response_out);
}

#define BYTES_PER_LINE 8

void 
ispfc_dump_response(ispfc_cb_t *cb, ispfc_iocb_status_t *is)
{
	byte *	p = (byte *)is;
	int	i;

	for (i = 0; i < sizeof(ispfc_iocb_status_t); ++i, ++p) {
		if ((i % BYTES_PER_LINE) == 0) {
			if (i > 0)
				P("\n");
			P("\t%p: ", p);
		}
		P("%02X ", *p);
	}
	P("\n");
}

void
ispfc_dump_cmd2(ispfc_cb_t *cb, ispfc_iocb_command2_t *cmd)
{
	int i;

	P("COMMAND IOCB: %p\n", cmd);
	P("  header.entry_type:     %02x\n", cmd->header.entry_type);
	P("  header.entry_cnt:      %02x\n", cmd->header.entry_cnt);
	P("  handle:                %08x\n", cmd->handle);
	P("  target:                %02x\n", cmd->target);
	P("  lun:                   %04x\n", cmd->lun);
	P("  control_flags:         %04x\n", cmd->control_flags);
	P("  timeout:               %04x\n", cmd->timeout);
	P("  seg_count:             %04x\n", cmd->seg_count);
	P("  CDB:                   ");
	for (i = 0; i < CMD_MAX_CDB; ++i)
		P("%02x ", cmd->cdb[i]);
	P("\n");
	P("  byte_count:            %08x\n", cmd->byte_count);
	for (i = 0; i < CMD_TYPE2_SG_CNT; ++i)
		P("  segments[%d]:           %08x %08x\n",
		  i, cmd->segments[i].buf, cmd->segments[i].len);
#if 0
	byte *	p = (byte *)cmd;

	for (i = 0; i < sizeof(ispfc_iocb_command2_t); ++i, ++p) {
		if ((i % BYTES_PER_LINE) == 0) {
			if (i > 0)
				P("\n");
			P("\t0x%x: ", p);
		}
		P("%02X ", *p);
	}
	P("\n");
#endif
}

void
ispfc_dump_cont0(ispfc_cb_t *cb, ispfc_iocb_cont0_t *ice)
{
	int i;

	P("CONTINUATION IOCB: %p\n", ice);
	P("  header.entry_type:     %02x\n", ice->header.entry_type);
	P("  header.entry_cnt:      %02x\n", ice->header.entry_cnt);
	for (i = 0; i < CONT_TYPE0_SG_CNT; ++i)
		P("  segments[%d]:           %08x %08x\n",
		  i, ice->segments[i].buf, ice->segments[i].len);
}

void
ispfc_dump_ct(ispfc_cb_t *cb, ispfc_iocb_ct_t *cmd)
{
	byte *	p = (byte *)cmd;
	int	i;

	for (i = 0; i < sizeof(ispfc_iocb_command2_t); ++i, ++p) {
		if ((i % BYTES_PER_LINE) == 0) {
			if (i > 0)
				P("\n");
			P("\t%p: ", p);
		}
		P("%02X ", *p);
	}
	P("\n");
}
