/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_mbox.c
 *
 * RCS:		$Id: ispfc_mbox.c,v 1.11 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver mailbox interface routines.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 5/24/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_mbox.h"

int32	ispfc_mbox_timeout_intr(int32, ispfc_cb_t *);

int 
ispfc_mbox_start(ispfc_cb_t *cb, void (*mbox_callback)(ispfc_cb_t *, void *, int32), void *mbox_cookie)
{
	mbox_cmd_sm_t *	mbox_sm = cb->mbox_sm;
	int		i;

	FC_DBG(2, ("%s: ispfc_mbox_start: %04x %04x %04x %04x %04x %04x %04x %04x\n", 
		   cb->debug_name, 
		   cb->mbox_in[0], cb->mbox_in[1], cb->mbox_in[2], cb->mbox_in[3], 
		   cb->mbox_in[4], cb->mbox_in[5], cb->mbox_in[6], cb->mbox_in[7]));

	/* Insure mailbox isn't busy */
	VERIFY(mbox_sm->mbox_state == ISPFC_MBOX_STATE_FREE);

	/*
         * HCCR_INTR still set at this point means the ISP is busy
         * processing a previous command ... which shouldn't really
         * happen. Just panic for now .....
	 */
        VERIFY((ISPFC_REG_READ(cb, ISPFC_REG_HCCR) & HCCR_INTR) == 0);

	/* Write the incoming mailbox registers */
        for (i = 0; i < cb->mbox_in_valid_count; ++i)
                ISPFC_REG_WRITE(cb, ISPFC_REG_MBOX(i), cb->mbox_in[i]);

	/* Hit the Host-to-RISC interrupt */
	ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_SET_HOST_INTR);

	/* Set mailbox state */
	mbox_sm->mbox_state = ISPFC_MBOX_STATE_BUSY;
	mbox_sm->mbox_callback = mbox_callback;
	mbox_sm->mbox_cookie = mbox_cookie;

	/*
	 * Start timeout timer going.
	 */
#if !defined(ISPFC_SIM)
	mbox_sm->mbox_timer_handle = eee_timerCreate(cb->mbox_timeout, EEE_TIMER_ONCE);
#else
	mbox_sm->mbox_timer_handle = eee_timerCreate(100, EEE_TIMER_ONCE);
#endif
	VERIFY(mbox_sm->mbox_timer_handle >= 0);
	eee_timerStart(mbox_sm->mbox_timer_handle, (void *)ispfc_mbox_timeout_intr,
		       cb, NULL, NULL, NULL);

	return(NFX_OK);
}

void
ispfc_mbox_done(ispfc_cb_t *cb)
{
	mbox_cmd_sm_t *	mbox_sm = cb->mbox_sm;
	int		i;
	int32		rc;

	/*
	 * We should only get here if a legitimate mailbox
	 * completion. If mailbox command isn't pending, just ignore
	 * the completion.
	 */
	if (mbox_sm->mbox_state == ISPFC_MBOX_STATE_FREE) {
		FC_DBG(2, ("%s: ispfc_mbox_done: Bogus completion; no mailbox command pending. Ignoring ....\n",
			   cb->debug_name));
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
		ISPFC_REG_WRITE(cb, ISPFC_REG_SEMAPHORE, 0);
		return;
	}

	if (ISPFC_REG_READ(cb, ISPFC_REG_HCCR) & HCCR_INTR) {
		FC_DBG(2, ("%s: ispfc_mbox_done: Bogus completion; command still in progress. Ignoring .... \n",
			   cb->debug_name));
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
		ISPFC_REG_WRITE(cb, ISPFC_REG_SEMAPHORE, 0);
		return;
	}

	/* We must have an out count of at least 1 */
	if (cb->mbox_out_valid_count == 0)
		++cb->mbox_out_valid_count;

	/* Read the outgoing mailbox registers */
	for (i = 0; i < cb->mbox_out_valid_count; ++i)
		cb->mbox_out[i] = ISPFC_REG_READ(cb, ISPFC_REG_MBOX(i));

	FC_DBG(2, ("%s: ispfc_mbox_done: %04x %04x %04x %04x %04x %04x %04x %04x\n", 
		   cb->debug_name, 
		   cb->mbox_out[0], cb->mbox_out[1], cb->mbox_out[2], cb->mbox_out[3], 
		   cb->mbox_out[4], cb->mbox_out[5], cb->mbox_out[6], cb->mbox_out[7]));

	rc = eee_timerStop(mbox_sm->mbox_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(mbox_sm->mbox_timer_handle);
	VERIFY(rc == NFX_OK);
	mbox_sm->mbox_timer_handle = 0;
	mbox_sm->mbox_state = ISPFC_MBOX_STATE_FREE;

	VERIFY(mbox_sm->mbox_callback);
	/* Mailbox command success ? */
	if (cb->mbox_out[0] == ISPFC_CS_MBOX_COMP) {
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
		ISPFC_REG_WRITE(cb, ISPFC_REG_SEMAPHORE, 0);
		(*mbox_sm->mbox_callback)(cb, mbox_sm->mbox_cookie, MBOX_CMD_STATUS_SUCCESS);
	}
	else {
		for (i = 0; i < 8; ++i) 
			cb->mbox_out[i] = ISPFC_REG_READ(cb, ISPFC_REG_MBOX(i));
		FC_DBG(0, ("%s: Mailbox failure: \n"
			   "   Input parameters:  0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n"
			   "   Current mailboxes: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n"
			   "   HCCR 0x%x, ISR 0x%x, H_TO_R_LOW 0x%x, H_TO_R_HI 0x%x\n"
			   "   REQ_IN 0x%x, REQ_OUT 0x%x, RESP_IN 0x%x, RESP_OUT 0x%x\n",
			   cb->debug_name,
			   cb->mbox_in[0], cb->mbox_in[1], cb->mbox_in[2], cb->mbox_in[3],
			   cb->mbox_in[4], cb->mbox_in[5], cb->mbox_in[6], cb->mbox_in[7],
			   cb->mbox_out[0], cb->mbox_out[1], cb->mbox_out[2], cb->mbox_out[3],
			   cb->mbox_out[4], cb->mbox_out[5], cb->mbox_out[6], cb->mbox_out[7],
			   ISPFC_REG_READ(cb, ISPFC_REG_HCCR),
			   ISPFC_REG_READ(cb, ISPFC_REG_INTSTAT),
			   ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_LOW),
			   ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_HI),
			   ISPFC_REG_READ(cb, ISPFC_REG_REQUEST_IN), ISPFC_REG_READ(cb, ISPFC_REG_REQUEST_OUT),
			   ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_IN), ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_OUT)));
		ISPFC_REG_WRITE(cb, ISPFC_REG_HCCR, HCCR_CMD_CLR_RISC_INTR);
		ISPFC_REG_WRITE(cb, ISPFC_REG_SEMAPHORE, 0);
		(*mbox_sm->mbox_callback)(cb, mbox_sm->mbox_cookie, MBOX_CMD_STATUS_FAILED);
	}
}

int32
ispfc_mbox_timeout_intr(int32 timer_handle, ispfc_cb_t *cb)
{
	mbox_cmd_sm_t *	mbox_sm = cb->mbox_sm;
	int		i;
	int32		rc;

	VERIFY(mbox_sm->mbox_state == ISPFC_MBOX_STATE_BUSY);
	VERIFY(mbox_sm->mbox_timer_handle == timer_handle);

	for (i = 0; i < 8; ++i) 
		cb->mbox_out[i] = ISPFC_REG_READ(cb, ISPFC_REG_MBOX(i));

	rc = eee_timerDestroy(mbox_sm->mbox_timer_handle);
	VERIFY(rc == NFX_OK);
	mbox_sm->mbox_timer_handle = 0;
	mbox_sm->mbox_state = ISPFC_MBOX_STATE_FREE;

#if !defined(ISPFC_SIM)
	FC_DBG(0, ("%s: Mailbox timeout: \n"
		   "   Input parameters:  0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n"
		   "   Current mailboxes: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n"
		   "   HCCR 0x%x, ISR 0x%x, H_TO_R_LOW 0x%x, H_TO_R_HI 0x%x\n"
		   "   REQ_IN 0x%x, REQ_OUT 0x%x, RESP_IN 0x%x, RESP_OUT 0x%x\n",
		   cb->debug_name,
		   cb->mbox_in[0], cb->mbox_in[1], cb->mbox_in[2], cb->mbox_in[3],
		   cb->mbox_in[4], cb->mbox_in[5], cb->mbox_in[6], cb->mbox_in[7],
		   cb->mbox_out[0], cb->mbox_out[1], cb->mbox_out[2], cb->mbox_out[3],
		   cb->mbox_out[4], cb->mbox_out[5], cb->mbox_out[6], cb->mbox_out[7],
		   ISPFC_REG_READ(cb, ISPFC_REG_HCCR),
		   ISPFC_REG_READ(cb, ISPFC_REG_INTSTAT),
		   ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_LOW),
		   ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_HI),
		   ISPFC_REG_READ(cb, ISPFC_REG_REQUEST_IN), ISPFC_REG_READ(cb, ISPFC_REG_REQUEST_OUT),
		   ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_IN), ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_OUT)));

	VERIFY(mbox_sm->mbox_callback);
	(*mbox_sm->mbox_callback)(cb, mbox_sm->mbox_cookie, MBOX_CMD_STATUS_TIMEOUT);
#else
	cb->mbox_out[0] = 0x4000;
	(*mbox_sm->mbox_callback)(cb, mbox_sm->mbox_cookie, MBOX_CMD_STATUS_SUCCESS);
#endif /* !defined(ISPFC_SIM) */

	return(NFX_OK);
}

int
ispfc_mbox_is_busy(ispfc_cb_t *cb)
{
	mbox_cmd_sm_t *	mbox_sm = cb->mbox_sm;

	return(mbox_sm->mbox_state == ISPFC_MBOX_STATE_BUSY);
}

void 
ispfc_mbox_abort(ispfc_cb_t *cb)
{
	mbox_cmd_sm_t *	mbox_sm = cb->mbox_sm;
	int32		rc;

	VERIFY(mbox_sm->mbox_state == ISPFC_MBOX_STATE_BUSY);

	FC_DBG(0, ("%s: Aborting in progress mailbox operation: \n"
		   "   Input parameters:  0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n"
		   "   Current mailboxes: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n"
		   "   HCCR 0x%x, ISR 0x%x, H_TO_R_LOW 0x%x, H_TO_R_HI 0x%x\n"
		   "   REQ_IN 0x%x, REQ_OUT 0x%x, RESP_IN 0x%x, RESP_OUT 0x%x\n",
		   cb->debug_name,
		   cb->mbox_in[0], cb->mbox_in[1], cb->mbox_in[2], cb->mbox_in[3],
		   cb->mbox_in[4], cb->mbox_in[5], cb->mbox_in[6], cb->mbox_in[7],
		   cb->mbox_out[0], cb->mbox_out[1], cb->mbox_out[2], cb->mbox_out[3],
		   cb->mbox_out[4], cb->mbox_out[5], cb->mbox_out[6], cb->mbox_out[7],
		   ISPFC_REG_READ(cb, ISPFC_REG_HCCR),
		   ISPFC_REG_READ(cb, ISPFC_REG_INTSTAT),
		   ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_LOW),
		   ISPFC_REG_READ(cb, ISPFC_REG_R_TO_H_HI),
		   ISPFC_REG_READ(cb, ISPFC_REG_REQUEST_IN), ISPFC_REG_READ(cb, ISPFC_REG_REQUEST_OUT),
		   ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_IN), ISPFC_REG_READ(cb, ISPFC_REG_RESPONSE_OUT)));

	rc = eee_timerStop(mbox_sm->mbox_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(mbox_sm->mbox_timer_handle);
	VERIFY(rc == NFX_OK);
	mbox_sm->mbox_timer_handle = 0;
	mbox_sm->mbox_state = ISPFC_MBOX_STATE_FREE;
}
