/*-----------------------------------------------------------------------------
 * Name:  ispfc_td.c
 *
 * RCS:  $Id: ispfc_td.c,v 1.8 2002/08/28 03:10:37 raster Exp $
 *
 * Description:  Create/action functions for FC topology discovery.
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/timer-api.h>
#include <exports/hal.h>
#include "ispfc_private.h"
#include "ispfc_ct.h"
#include "ispfc_td.h"
#include "ispfc_td_state.h"

void        fctd_free_all(fctd_t *);
void        fctd_traverse_all(fctd_t *);
void        fctd_ct_callback(ct_desc_t *);
void        fctd_giel_req(fctd_t *);
void        fctd_gieln_req(fctd_t *);
void        fctd_gpl_req(fctd_t *);
void        fctd_gppn_req(fctd_t *);
void        fctd_gapnl_req(fctd_t *);
void        fctd_get_loop_map_callback(wwn_t, hal_device_map_t *, int, fctd_t *);
portList_t *fctd_find_portname(fctd_t *, char *);
void        fctd_poll(int32, fctd_t *);
int32       fctd_timer_expired(int32, esm_handle_t);

fctd_t fctd_cb;
ispfc_ct_req_t fctd_req_buf;
ispfc_ct_resp_t fctd_rsp_buf;

/*-----------------------------------------------------------------------------
 * fctd_init
 *
 * Description:  FC topology discovery state machine instantiator and control
 *               block initialization routine. 
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
void
fctd_init(uint32 poll_period)
{
	int32			rc;
	fctd_t *		pcb = &fctd_cb;
	esm_handle_t		handle;
	eee_desc_t *		e_desc;
	eee_descBuf_t *		bd;

	pcb->slot_num = nfxMySlotId;
	pcb->port_num = nfxMyCPUId;
	sprintf(pcb->debug_name, "fctd:%d.%d",
		pcb->slot_num, pcb->port_num);

	rc = fctd_sm_create( &handle, pcb);               
	if (rc != NFX_OK) {
		FC_DBG(0, ("%s: fctd_instantiate: error: couldn't create state machine\n", pcb->debug_name));
		return;
	}
	pcb->sm_handle = handle;
	pcb->tmr_handle = eee_timerCreate(FCTD_TIMER_INTERVAL, EEE_TIMER_REPEAT);
	if (pcb->tmr_handle < 0) {
		FC_DBG(0, ("%s: fctd_instantiate: couldn't create timer handler\n", pcb->debug_name));
		return;
	}

	pcb->ie_count = 0;
	pcb->ie_index = 0;
	pcb->state = FCTD_IDLE_STATE;
	pcb->ctd.e_desc = eee_allocateDesc(EEE_SHARED_POOL, 0);
	VERIFY(pcb->ctd.e_desc);
	bzero(pcb->ctd.e_desc, sizeof(eee_desc_t));
	e_desc = pcb->ctd.e_desc;
	bd = e_desc->bd;
	bd[1].buf = (void *)&fctd_req_buf;
	bd[1].len = sizeof(ispfc_ct_req_t);
	bd[1].control = EDESC_BD_CMD_IN_USE;
	bd[2].buf = (void *)&fctd_rsp_buf;
	bd[2].len = sizeof(ispfc_ct_resp_t);
	bd[2].control = EDESC_BD_CMD_IN_USE;
	e_desc->hdr.control = 
		((2 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK) |
		((sizeof(ispfc_ct_req_t) + sizeof(ispfc_ct_resp_t)) & EDESC_TOTAL_LEN);

	/* Start poll period timer and register poll handler */
	pcb->poll_timer_handle = eee_timerCreate(poll_period, EEE_TIMER_REPEAT);
	VERIFY(pcb->poll_timer_handle >= 0);
	rc = eee_timerStart(pcb->poll_timer_handle, (void *)fctd_poll, pcb, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
}

void 
fctd_ct_callback(ct_desc_t *ctd)
{
	fctd_t *		pcb = ctd->handle;
	int32			rc;
	ispfc_ct_resp_t *	resp;
	esm_event_t		esmEvent;

	FC_DBG(1, ("%s: fctd_ct_callback: status 0x%x\n", pcb->debug_name, ctd->status));
	if (ctd->status != CT_STATUS_OK) {
		FC_DBG(1, ("%s: fctd_ct_callback: status 0x%x - send fail reply 1\n", pcb->debug_name, ctd->status));
		pcb->rsp_id = FCTD_SEND_FAIL_REPLY;
		esmEvent.event = FCTD_EVENT_PROC_MSG;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(pcb->sm_handle, &esmEvent);
	}
	else {
		resp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
		inplace_ntohs(&resp->header.response_code);
		ctd->status |= (resp->header.response_code << CT_STATUS_RESP_SHIFT) &
                        CT_STATUS_RESP_MASK;
		ctd->status |= CT_STATUS_RESP_VALID;
		if (resp->header.response_code != SNS_CT_FS_ACC) {
			FC_DBG(1, ("%s: fctd_ct_callback: CT response_code 0x%x reason_code 0x%x - send fail reply 2\n", 
				   pcb->debug_name, resp->header.response_code, resp->header.reason_code));
			pcb->rsp_id = FCTD_SEND_FAIL_REPLY;
			esmEvent.event = FCTD_EVENT_PROC_MSG;
			esmEvent.p1.u.p = NULL;
			esmEvent.p2.u.p = NULL;
			/* reject response is valid for GAPNL request */
			if (pcb->state == FCTD_SEND_GAPNL_REQ_STATE) {
				pcb->rsp_id = FCTD_SEND_GAPNL_REPLY;
				esmEvent.p2.u.p = (void *)0xffffffff;
			}
			rc = esm_deliverEvent(pcb->sm_handle, &esmEvent);
			VERIFY(rc == NFX_OK);
		}
		else {
			switch (pcb->state) {
			case FCTD_SEND_GIEL_REQ_STATE:
				FC_DBG(1, ("%s: fctd_ct_callback: Send GIEL reply\n", pcb->debug_name));
				pcb->rsp_id = FCTD_SEND_GIEL_REPLY;
				break;
			case FCTD_SEND_GIELN_REQ_STATE:
				FC_DBG(1, ("%s: fctd_ct_callback: Send GIELN reply\n", pcb->debug_name));
				pcb->rsp_id = FCTD_SEND_GIELN_REPLY;
				break;
			case FCTD_SEND_GPL_REQ_STATE:
				FC_DBG(1, ("%s: fctd_ct_callback: Send GPL reply\n", pcb->debug_name));
				pcb->rsp_id = FCTD_SEND_GPL_REPLY;
				break;
			case FCTD_SEND_GPPN_REQ_STATE:
				FC_DBG(1, ("%s: fctd_ct_callback: Send GPPN reply\n", pcb->debug_name));
				pcb->rsp_id = FCTD_SEND_GPPN_REPLY;
				break;
			case FCTD_SEND_GAPNL_REQ_STATE:
				FC_DBG(1, ("%s: fctd_ct_callback: Send GAPNL reply\n", pcb->debug_name));
				pcb->rsp_id = FCTD_SEND_GAPNL_REPLY;
				break;
			default:
				break;
			}
			esmEvent.event = FCTD_EVENT_PROC_MSG;
			esmEvent.p1.u.p = ctd->e_desc->bd[2].buf;
			esmEvent.p2.u.p = NULL;
			rc = esm_deliverEvent(pcb->sm_handle, &esmEvent);
			VERIFY(rc == NFX_OK);
		}
	}
}

/*-----------------------------------------------------------------------------
 * fctd_start_disc
 *
 * Description:  FC topology discovery state machine action to start discovery
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_start_disc( 
		esm_handle_t handle,
		int32        nextState,
		esm_event_t  thisEvent,
		void *	     userInfo)
{
	hal_dispatch_entry_t *	hde = hal_dt_entry_get();
	hal_adapter_info_t	info;
	fctd_t *		pcb = (fctd_t *)userInfo;
	int32			rc;
	esm_event_t		esmEvent;

	FC_DBG(1, ("%s: fctd_start_disc\n", pcb->debug_name));

	hde->hal_ioctl(HAL_IOCTL_FC_CONNECTION_STATE, &rc);
	if (rc != HAL_IOCTL_FC_LINK_STATE_LINK_UP) {
		esmEvent.event = FCTD_EVENT_START_DISC;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
   		rc = esm_deliverEvent(handle, &esmEvent);
		return(FCTD_STATE_IDLE);
   	} 
	/* Get self port name from FC-HAL */
	hde->hal_ioctl(HAL_IOCTL_COMMON_GET_ADAPTER_INFO, &info);
	inplace_htonll(&info.u.fc_data.fc_node_name);
	WWN_SET(&pcb->port_name, &info.u.fc_data.fc_node_name); 
	/* node name is port name in our case */
	FC_DBG(1, ("%s: My portname %s\n", pcb->debug_name, wwn_to_str(&pcb->port_name, NULL)));
	FC_DBG(1, ("%s: My interface type is %d\n", pcb->debug_name, info.interface_type));
	pcb->hal_interface_type = info.interface_type;

	if (pcb->hal_interface_type == HAL_IF_FC_L_PORT) {
		FC_DBG(1, ("%s: Acquiring loop map\n", pcb->debug_name));
		pcb->state = FCTD_SEND_GET_LOOP_MAP;
		hde->hal_ioctl(HAL_IOCTL_COMMON_GET_DEVICE_MAP, 
			       null_wwn, 
			       &pcb->loop_map, 
			       fctd_get_loop_map_callback, pcb);
	}
	else
	if (pcb->hal_interface_type == HAL_IF_FC_N_PORT) {
		FC_DBG(1, ("%s: Acquiring point-to-point peer identity.\n", pcb->debug_name));
		pcb->state = FCTD_SEND_GET_LOOP_MAP;
		hde->hal_ioctl(HAL_IOCTL_COMMON_GET_DEVICE_MAP, 
			       null_wwn, 
			       &pcb->loop_map, 
			       fctd_get_loop_map_callback, pcb);
	}
	else {
		/* Send GIEL request */
		pcb->state = FCTD_SEND_GIEL_REQ_STATE;
		pcb->ie_count = 0;
		pcb->ie_index = 0;
		fctd_giel_req(pcb);
	}

	return nextState;

} /* fctd_start_disc */

/*-----------------------------------------------------------------------------
 * fctd_cont_disc()
 *
 * Description:  FCTD state machine action to continue discovery procedure 
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_cont_disc( 
	       esm_handle_t handle,
	       int32        nextState,
	       esm_event_t  thisEvent,  /* state event */
	       void *	    userInfo)  /* control block */
{
	fctd_t *	pcb = (fctd_t *)userInfo;
	uint32		rc, abort = 0;
	esm_event_t	esmEvent;

	FC_DBG(1, ("%s: fctd_cont_disc: state = %d\n", pcb->debug_name, pcb->state));

	switch (pcb->state) {
	case FCTD_IDLE_STATE:
		/* idle or abort */
		abort = 1;
		FC_DBG(0, ("%s: fctd_cont_disc: In IDLE state.\n", pcb->debug_name));
		break;

	case FCTD_SEND_GIEL_REQ_STATE:
		pcb->state = FCTD_SEND_GIELN_REQ_STATE;
		pcb->ie_index = 0;
		fctd_gieln_req(pcb);
		rc = eee_timerStart(pcb->tmr_handle,(void *)fctd_timer_expired,
				    (void *)pcb->sm_handle, NULL, NULL, NULL); 
		if (rc != NFX_OK) {
			abort = 1; 
			FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
		}
		break;

	case FCTD_SEND_GIELN_REQ_STATE:
		pcb->ie_index++;
		if (pcb->ie_index < pcb->ie_count) {
			fctd_gieln_req(pcb);
			rc = eee_timerStart(pcb->tmr_handle,(void *)fctd_timer_expired,
					    (void *)pcb->sm_handle, NULL, NULL, NULL); 
			if (rc != NFX_OK) {
				abort = 1; 
				FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
			}
		}
		else {
			pcb->state = FCTD_SEND_GPL_REQ_STATE;
			pcb->ie_index = 0;
			fctd_gpl_req(pcb);
			rc = eee_timerStart(pcb->tmr_handle,(void *)fctd_timer_expired,
					    (void *)pcb->sm_handle, NULL, NULL, NULL); 
			if (rc != NFX_OK) {
				abort = 1; 
				FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
			}
		}
		break;

	case FCTD_SEND_GPL_REQ_STATE:
		pcb->ie_index++;
		if (pcb->ie_index < pcb->ie_count) {
			fctd_gpl_req(pcb);
			rc = eee_timerStart(pcb->tmr_handle,(void *)fctd_timer_expired,
					    (void *)pcb->sm_handle, NULL, NULL, NULL); 
			if (rc != NFX_OK) {
				abort = 1;
				FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
			}
		}
		else {
			pcb->ie_index = 0;
			pcb->ie_list[pcb->ie_index].top_list = pcb->ie_list[pcb->ie_index].port_list;
			VERIFY(pcb->ie_list[pcb->ie_index].top_list);
			pcb->state = FCTD_SEND_GPPN_REQ_STATE;
			fctd_gppn_req(pcb);
			rc = eee_timerStart(pcb->tmr_handle,
					    (void *)fctd_timer_expired,
					    (void *)pcb->sm_handle, NULL, NULL, NULL);
			if (rc != NFX_OK) {
				abort = 1;
				FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
			}
		}
		break;

	case FCTD_SEND_GPPN_REQ_STATE:
		/* get physical port number for each port on every IE list */
	        pcb->ie_list[pcb->ie_index].top_list = 
			pcb->ie_list[pcb->ie_index].top_list->next; 
		if (pcb->ie_list[pcb->ie_index].top_list == NULL) {
			pcb->ie_index++;
			if (pcb->ie_index >= pcb->ie_count) {
				pcb->ie_index = 0;
				pcb->ie_list[pcb->ie_index].top_list = pcb->ie_list[pcb->ie_index].port_list;
				VERIFY(pcb->ie_list[pcb->ie_index].top_list);
				pcb->state = FCTD_SEND_GAPNL_REQ_STATE;
				fctd_gapnl_req(pcb);
				rc = eee_timerStart(pcb->tmr_handle, (void *)fctd_timer_expired,
						    (void *)pcb->sm_handle, NULL, NULL, NULL); 
				if (rc != NFX_OK) {
					abort = 1;
					FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
				}
				break;
			}
			pcb->ie_list[pcb->ie_index].top_list = pcb->ie_list[pcb->ie_index].port_list;
		}
		fctd_gppn_req(pcb);
		rc = eee_timerStart(pcb->tmr_handle, (void *)fctd_timer_expired,
				    (void *)pcb->sm_handle, NULL, NULL, NULL); 
		if (rc != NFX_OK) {
			abort = 1;
			FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
		}
		break;

	case FCTD_SEND_GAPNL_REQ_STATE:
		/* get attached port list for each port on every IE list */
	        pcb->ie_list[pcb->ie_index].top_list = 
			pcb->ie_list[pcb->ie_index].top_list->next; 
		if (pcb->ie_list[pcb->ie_index].top_list == NULL) {
			pcb->ie_index++;
			if (pcb->ie_index >= pcb->ie_count) {
				pcb->state = FCTD_DISCOVERY_DONE_STATE;
				esmEvent.event = FCTD_EVENT_DISC_DONE;
				esmEvent.p1.u.p = NULL;
				esmEvent.p2.u.p = NULL;
				rc = esm_deliverEvent(handle, &esmEvent);
				if (rc != NFX_OK)
					FC_DBG(1, ("%s: fctd_cont_disc: Couldn't deliver DISC_DONE event\n", pcb->debug_name));
				break;
			}
			pcb->ie_list[pcb->ie_index].top_list = pcb->ie_list[pcb->ie_index].port_list;
		}
		fctd_gapnl_req(pcb);
		rc = eee_timerStart(pcb->tmr_handle, (void *)fctd_timer_expired,
				    (void *)pcb->sm_handle, NULL, NULL, NULL); 
		if (rc != NFX_OK) {
			abort = 1;
			FC_DBG(0, ("%s: fctd_cont_disc: unable to start timer\n", pcb->debug_name));
		}
		break;

	case FCTD_SEND_GET_LOOP_MAP:
		pcb->state = FCTD_DISCOVERY_DONE_STATE;
		esmEvent.event = FCTD_EVENT_DISC_DONE;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(handle, &esmEvent);
		if (rc != NFX_OK)
			FC_DBG(0, ("%s: fctd_cont_disc: Couldn't deliver DISC_DONE event\n", pcb->debug_name));
		break;

	case FCTD_DISCOVERY_DONE_STATE:
		break;

	default:
		FC_DBG(0, ("%s: fctd_cont_disc: Unknown state %d\n", pcb->debug_name, pcb->state));
		abort = 1;
	}

	if (abort) {
		esmEvent.event = FCTD_EVENT_ABORT_DISC;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(handle, &esmEvent);
		if (rc != NFX_OK)
			FC_DBG(1, ("%s: fctd_cont_disc: Couldn't deliver ABORT event\n", pcb->debug_name));
	}

	return nextState;

} /* fctd_cont_disc */

/*-----------------------------------------------------------------------------
 * fctd_proc_msg()
 *
 * Description:  FCTD state machine action to process message 
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_proc_msg( 
	      esm_handle_t handle,
	      int32        nextState,
	      esm_event_t  thisEvent,  /* state event */
	      void *	   userInfo)  /* control block */
{
	fctd_t *		pcb = (fctd_t *)userInfo;
	uchar8			msg_id = pcb->rsp_id;
	portList_t *		portlist;
	portList_t *		toplist;
	portList_t *		curlist;
	atPortList_t *		aportlist;
	atPortList_t *		atoplist;
	atPortList_t *		acurlist;
	uint32			ie_count;
	uint32			port_count;
	uint32			symbolic_name_len;
	ispfc_ct_resp_t *	rsp;
	int32			rc, i, abort = 0;
	esm_event_t		esmEvent;

	eee_timerStop(pcb->tmr_handle); /* timer may or may not be running */

	FC_DBG(1, ("%s: fctd_proc_msg: msg_id = %d\n", pcb->debug_name, msg_id));

	switch (msg_id) {
	case FCTD_SEND_GIEL_REPLY:
		/* 
		 *  After CT_IU header 
		 *  Number of IE entries (4 bytes)
		 *  IE name (8 bytes)
		 *  Reserved (3 bytes)
		 *  IE type (1 byte)
		 *  IE name ...
		 */
		if (pcb->state != FCTD_SEND_GIEL_REQ_STATE) {
			FC_DBG(0, ("%s: fctd_proc_msg: msg_id FCTD_SEND_GIEL_REPLY: unexpected state %d\n", 
				   pcb->debug_name, pcb->state));
			abort = 1;
			goto proc_done;
		}
		else {
			rsp = (ispfc_ct_resp_t *)thisEvent.p1.u.p;
			if (!rsp) {
				FC_DBG(0, ("%s: fctd_proc_msg: NULL parameter pointer\n", pcb->debug_name));
				abort = 1;
				goto proc_done;
			}
			inplace_ntohl(&rsp->u.GIEL.ie_count);
			FC_DBG(1, ("%s: fctd_proc_msg: GIEL reply: IE count = %d\n", pcb->debug_name, rsp->u.GIEL.ie_count));
			pcb->ie_count = rsp->u.GIEL.ie_count;
			if (pcb->ie_count > FCTD_MAX_IES) {
				FC_DBG(0, ("%s: fctd_proc_msg: max number of IEs exceeded -- %d\n", 
					   pcb->debug_name, pcb->ie_count));
				abort = 1;
				goto proc_done;
			} 
			for (ie_count = 0; ie_count < pcb->ie_count; ++ie_count) {
				FC_DBG(1, ("%s: fctd_proc_msg: ie[%d] name %s, type 0x%x\n", 
					   pcb->debug_name,
					   ie_count, 
					   wwn_to_str((wwn_t *)rsp->u.GIEL.ie[ie_count].ie_name, NULL),
					   rsp->u.GIEL.ie[ie_count].ie_type));
				WWN_SET(&pcb->ie_list[ie_count].node_name, rsp->u.GIEL.ie[ie_count].ie_name);
				pcb->ie_list[ie_count].node_type = rsp->u.GIEL.ie[ie_count].ie_type;
				pcb->ie_list[ie_count].num_ports = 0;
				pcb->ie_list[ie_count].top_list = NULL;
				pcb->ie_list[ie_count].port_list = NULL;
			}
		}
		break;

	case FCTD_SEND_GIELN_REPLY:
		/* 
		 *  After CT_IU header 
		 *  Symbolic name length (1 byte)
		 *  Symbolic name (up to 255 bytes)
		 */
		if (pcb->state != FCTD_SEND_GIELN_REQ_STATE) {
			FC_DBG(0, ("%s: fctd_proc_msg: msg_id FCTD_SEND_GIELN_REPLY: unexpected state %d\n", 
				   pcb->debug_name, pcb->state));
			abort = 1;
			goto proc_done;
		}
		else {
			rsp = (ispfc_ct_resp_t *)thisEvent.p1.u.p;
			if (!rsp) {
				FC_DBG(0, ("%s: fctd_proc_msg: NULL parameter pointer\n", pcb->debug_name));
				abort = 1;
				goto proc_done;
			}

			bzero(pcb->ie_list[pcb->ie_index].symbolic_name, FCTD_MAX_SYMBOLIC_NAME_LEN+1);
			symbolic_name_len = MIN(rsp->u.GIELN.ie_logical_name.len, FCTD_MAX_SYMBOLIC_NAME_LEN);
			bcopy(rsp->u.GIELN.ie_logical_name.name, 
			      pcb->ie_list[pcb->ie_index].symbolic_name, 
			      symbolic_name_len);
			FC_DBG(1, ("%s: fctd_proc_msg: GIELN reply: symbolic name '%s'\n", 
				   pcb->debug_name, pcb->ie_list[pcb->ie_index].symbolic_name));
		}
		break;

	case FCTD_SEND_GPL_REPLY:
		/*
		 *  After CT_IU header
		 *  Port count (4 bytes)
		 *  Port name (8 bytes)
		 *  Reserved (3 bytes)
		 *  Port type (1 byte)
		 *  ...
		 */
		if (pcb->state != FCTD_SEND_GPL_REQ_STATE) {
			FC_DBG(0, ("%s: fctd_proc_msg: msg_id FCTD_SEND_GPL_REPLY: unexpected state %d\n", 
				   pcb->debug_name, pcb->state));
			abort = 1;
			goto proc_done;
		}
		else {
			curlist = toplist = NULL;
			rsp = (ispfc_ct_resp_t *)thisEvent.p1.u.p;
			if (!rsp) {
				FC_DBG(0, ("%s: fctd_proc_msg: NULL parameter pointer\n", pcb->debug_name));
				abort = 1;
				goto proc_done;
			}
			inplace_ntohl(&rsp->u.GPL.port_count); 
			port_count = rsp->u.GPL.port_count; 
			if (port_count > FCTD_MAX_IE_PORTS) {
				FC_DBG(0, ("%s: fctd_proc_msg: max IE ports exceeded -- %d\n", 
					   pcb->debug_name, port_count));
				abort = 1;
				goto proc_done;
			}
			pcb->ie_list[pcb->ie_index].num_ports = port_count;
			FC_DBG(1, ("%s: fctd_proc_msg: GPL reply: Port count = %d\n", pcb->debug_name, port_count));
			for (i = 0; i < port_count; i++) {
				portlist = (portList_t *)eee_ramAlloc(sizeof(portList_t));
				VERIFY(portlist);
				bzero(portlist, sizeof(portList_t));
				FC_DBG(1, ("%s: fctd_proc_msg: port[%d] name %s, type 0x%x\n", 
					   pcb->debug_name,
					   i,
					   wwn_to_str((wwn_t *)rsp->u.GPL.port[i].port_name, NULL), 
					   rsp->u.GPL.port[i].port_type));
				WWN_SET(&portlist->port_name, rsp->u.GPL.port[i].port_name);
				portlist->port_type = rsp->u.GPL.port[i].port_type;
				if (toplist == NULL) {
					toplist = portlist;
					curlist = portlist;
				}
				else {
					curlist->next = portlist;
					curlist = portlist;
				}
			}
			pcb->ie_list[pcb->ie_index].port_list = toplist;
			pcb->ie_list[pcb->ie_index].top_list = toplist;
			pcb->ie_list[pcb->ie_index].port_list->atop_list = NULL;
			pcb->ie_list[pcb->ie_index].port_list->atport_list = NULL;
			pcb->ie_list[pcb->ie_index].port_list->num_atports = 0;
		}
		break;

	case FCTD_SEND_GPPN_REPLY:
		/*
		 *  After CT_IU header
		 *  Physical port number (4 bytes)
		 */
		if (pcb->state != FCTD_SEND_GPPN_REQ_STATE) {
			FC_DBG(0, ("%s: fctd_proc_msg: msg_id FCTD_SEND_GPPN_REPLY: unexpected state %d\n", 
				   pcb->debug_name, pcb->state));
			abort = 1;
			goto proc_done;
		}
		else {
			rsp = (ispfc_ct_resp_t *)thisEvent.p1.u.p;
			if (!rsp) {
				FC_DBG(0, ("%s: fctd_proc_msg: NULL parameter pointer\n", pcb->debug_name));
				abort = 1;
				goto proc_done;
			}
			inplace_ntohl(&rsp->u.GPPN.port_number); 
			FC_DBG(1, ("%s: fctd_proc_msg: GPPN reply: physical port number %d\n", 
				   pcb->debug_name, rsp->u.GPPN.port_number));
			pcb->ie_list[pcb->ie_index].top_list->physical_port_number = rsp->u.GPPN.port_number;
		}
		break;

	case FCTD_SEND_GAPNL_REPLY:
		/*
		 *  After CT_IU header
		 *  Number of ports (4 bytes)
		 *  Port name (8 bytes)
		 *  Reserved (2 bytes)
		 *  Port flags (1 byte)
		 *  Port type (1 byte)
		 *  ...
		 */
		if (pcb->state != FCTD_SEND_GAPNL_REQ_STATE) {
			FC_DBG(0, ("%s: fctd_proc_msg: msg_id FCTD_SEND_GAPNL_REPLY: unexpected state %d\n", 
				   pcb->debug_name, pcb->state));
			abort = 1;
			goto proc_done;
		}
		else {
			acurlist = atoplist = NULL;
			/* reject response is same as zero ports */
			if (thisEvent.p2.u.p == (void *)0xffffffff) { 
				port_count = 0;
			}
			else {
				rsp = (ispfc_ct_resp_t *)thisEvent.p1.u.p;
				if (!rsp) {
					FC_DBG(0, ("%s: fctd_proc_msg: NULL parameter pointer\n", pcb->debug_name));
					abort = 1;
					goto proc_done;
				}
				inplace_ntohl(&rsp->u.GAPNL.port_count); 
				port_count = rsp->u.GAPNL.port_count; 
				if (port_count > FCTD_MAX_ATTACHED_PORTS) {
					FC_DBG(0, ("%s: fctd_proc_msg: max attached ports exceeded -- %d\n", 
						   pcb->debug_name, port_count));
					abort = 1;
					goto proc_done;
				}
				if (port_count > 0) {
					for (i = 0; i < port_count; i++) {
						aportlist = (atPortList_t *)eee_ramAlloc(sizeof(atPortList_t));
						VERIFY(aportlist);
						bzero(aportlist, sizeof(atPortList_t));
						FC_DBG(1, ("%s: fctd_proc_msg: aport[%d] name %s, type 0x%x\n", 
							   pcb->debug_name,
							   i,
							   wwn_to_str((wwn_t *)rsp->u.GAPNL.attached_port[i].port_name, NULL),
							   rsp->u.GAPNL.attached_port[i].port_type));
						WWN_SET(&aportlist->port_name, rsp->u.GAPNL.attached_port[i].port_name);
						aportlist->port_type = rsp->u.GAPNL.attached_port[i].port_type;
						if (atoplist == NULL) {
							atoplist = aportlist;
							acurlist = aportlist;
						}
						else {
							(atPortList_t *)acurlist->next = aportlist;
							acurlist = aportlist;
						}
					}
				}
			}
			pcb->ie_list[pcb->ie_index].top_list->atport_list = atoplist;
			pcb->ie_list[pcb->ie_index].top_list->atop_list = atoplist;
			pcb->ie_list[pcb->ie_index].top_list->num_atports = port_count;
		}
		break;

	case FCTD_SEND_GET_LOOP_MAP_REPLY:
		FC_DBG(1, ("%s: fctd_proc_msg: Got loop map: hostid 0x%x, port_count %d\n", 
			   pcb->debug_name, pcb->loop_map.hostid, pcb->loop_map.port_count));
		break;

	case FCTD_SEND_FAIL_REPLY:
		FC_DBG(0, ("%s: fctd_proc_msg: FCTD_SEND_FAIL_REPLY\n", pcb->debug_name));
		abort = 1;
		break;

	case FCTD_SEND_CHANGE_NOTICE:
		fctd_free_all(pcb); /* free linked lists */
		esmEvent.event = FCTD_EVENT_START_DISC;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
    		rc = esm_deliverEvent(handle, &esmEvent);
    		if (rc != NFX_OK) {
			abort = 1;
			FC_DBG(0, ("%s: fctd_proc_msg: Couldn't deliver START_DISC event\n", pcb->debug_name));
		}
		break;

	default:
		abort = 1;
	}

 proc_done:
	if (abort) {
		esmEvent.event = FCTD_EVENT_ABORT_DISC;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(handle, &esmEvent);
		if (rc != NFX_OK)
			FC_DBG(1, ("%s: fctd_proc_msg: Couldn't deliver ABORT event\n", pcb->debug_name));
	}
	else {
		/* schedule to continue with discovery procedure */
		esmEvent.event = FCTD_EVENT_CONT_DISC;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(handle, &esmEvent);
		if (rc != NFX_OK)
			FC_DBG(1, ("%s: fctd_proc_msg: Couldn't deliver CONT_DISC event\n", pcb->debug_name));
	}
	return nextState;

} /* fctd_proc_msg */

/*-----------------------------------------------------------------------------
 * fctd_finish_disc()
 *
 * Description:  FCTD state machine action to send discovery message block 
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_finish_disc( 
		 esm_handle_t handle,
		 int32        nextState,
		 esm_event_t  thisEvent,  /* state event */
		 void *	      userInfo)  /* control block */
{
	fctd_t *		pcb = (fctd_t *)userInfo;
	hal_dispatch_entry_t *	hde = hal_dt_entry_get();
	hal_adapter_info_t	info;

	FC_DBG(1, ("%s: fctd_finish_disc \n", pcb->debug_name));

	/*	
	 *  Topology discovery information is complete at this point.
	 *  Format topolgy data and send it to Storage Manager.
	 */   

	/* save HAL state to check for potential fabric changes */
   	hde->hal_ioctl(HAL_IOCTL_COMMON_GET_ADAPTER_INFO, &info);
   	pcb->hal_state = info.u.fc_data.fc_state_change_generation_num;

#ifdef DEBUG
	if (ispfc_dbg_level >= 10)
		fctd_traverse_all(pcb);
#endif

	return nextState;

} /* fctd_finish_disc */

/*-----------------------------------------------------------------------------
 * fctd_abort_disc()
 *
 * Description:  FCTD state machine action to abort discovery 
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_abort_disc( 
		esm_handle_t handle,
		int32        nextState,
		esm_event_t  thisEvent,  /* state event */
		void *	     userInfo)  /* control block */
{
	fctd_t *	pcb = (fctd_t *)userInfo;

	FC_DBG(1, ("%s: fctd_abort_disc \n", pcb->debug_name));

	eee_timerStop(pcb->tmr_handle); /* timer may or may not be running */

	fctd_free_all(pcb); /* free all linked lists */

	pcb->state = FCTD_IDLE_STATE;

	return nextState;

} /* fctd_abort_disc */

/*-----------------------------------------------------------------------------
 * fctd_timer_expired
 *
 * Description: Timer handler callback function for timer expiration.  
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_timer_expired(int32 tmr_handler, esm_handle_t sm_handle)
{
	int32		rc;
	esm_event_t	esmEvent;

	esmEvent.event = FCTD_EVENT_TIMEOUT;
	esmEvent.p1.u.p = NULL;
	esmEvent.p2.u.p = NULL;
	rc = esm_deliverEvent(sm_handle, &esmEvent);
	return rc;
}

/*-----------------------------------------------------------------------------
 * fctd_ignore()
 *
 * Description:  FCTD state machine action to ignore event
 * 
 * Created By:  Tony Wong
 *
 * Date Created:  8/28/2001
 *-----------------------------------------------------------------------------
 */
int32
fctd_ignore(esm_handle_t handle, 
	    int32	 nextState,
	    esm_event_t	 thisEvent, 
	    void *	 userInfo) 
{
	fctd_t *	pcb = (fctd_t *)userInfo;

	FC_DBG(1, ("%s: fctd_ignore \n", pcb->debug_name));
	return nextState;
} 

int32 
fctd_create(esm_handle_t	handle, 
		  int32		nextState,
		  esm_event_t	thisEvent,
		  void *	userInfo)
{
	fctd_t *	pcb = (fctd_t *)userInfo;

	FC_DBG(1, ("%s: fctd_create \n", pcb->debug_name));
	return nextState;
}

int32 
fctd_bad_state(esm_handle_t	handle, 
		     int32		nextState,
                     esm_event_t	thisEvent,
                     void * 		userInfo)
{
	fctd_t *	pcb = (fctd_t *)userInfo;

	FC_DBG(1, ("%s: fctd_bad_state \n", pcb->debug_name));
	return nextState;
}

void 
fctd_free_all(fctd_t *pcb) 
{
	uint32		i;
	portList_t *	toplist;
	portList_t *	curlist;
	atPortList_t *	atoplist;
	atPortList_t *	acurlist;

	for (i = 0; i < pcb->ie_count; i++) {
		toplist = pcb->ie_list[i].port_list;
		while (toplist != NULL) {
			curlist = toplist;
			toplist = toplist->next;		
			atoplist = curlist->atport_list;
			while (atoplist != NULL) {
				acurlist = atoplist;
				atoplist = (atPortList_t *)atoplist->next;
				eee_ramDealloc(acurlist);
			}
			eee_ramDealloc(curlist);
		}
		pcb->ie_list[i].port_list = NULL;
		pcb->ie_list[i].top_list = NULL;
		pcb->ie_list[i].num_ports = 0;
	}
	pcb->ie_count = 0;
	pcb->ie_index = 0;

	bzero(&pcb->loop_map, sizeof(hal_device_map_t));
}

portList_t *
fctd_find_portname(fctd_t *pcb, char8 *port_name) /* return pointer if found */
{
	uint32 i;
	portList_t *toplist, *curlist;

	for (i = 0; i < pcb->ie_count; i++) {
		toplist = pcb->ie_list[i].port_list;
		while (toplist != NULL) {
			curlist = toplist;
			toplist = toplist->next;		
			if (WWN_EQUAL((char *)&curlist->port_name, port_name))
				return(curlist);
		}
	}
	return NULL;
}

void 
fctd_traverse_all(fctd_t *pcb) /* visit all linked lists */
{
	uint32		i, j, k;
	portList_t *	toplist;
	atPortList_t *	atoplist;

	char *fctd_porttype_to_str(uint32 porttype) 
		{
			switch(porttype) {
			case PORT_TYPE_N:
				return("N_PORT");
			case PORT_TYPE_NL:
				return("NL_PORT");
			case PORT_TYPE_F:
				return("F_PORT");
			case PORT_TYPE_FL:
				return("FL_PORT");
			case PORT_TYPE_E:
				return("E_PORT");
			case PORT_TYPE_B:
				return("B_PORT");
			default:
				return("UNKNOWN");
			}
		}

	if (!pcb) {
		FC_DBG(0, ("%s: fctd_traverse: null control block\n", pcb->debug_name));
		return;
	}
	FC_DBG(0, ("%s: FC Topology DUMP: Number of IEs: %d\n", pcb->debug_name, pcb->ie_count));
	for (i = 0; i < pcb->ie_count; i++) {
		FC_DBG(0, ("%s: IE[%d]: name: %s ('%s'), type %s, port count %d\n", 
			   pcb->debug_name,
			   i, 
			   wwn_to_str(&pcb->ie_list[i].node_name, NULL),
			   pcb->ie_list[i].symbolic_name,
			   fctd_porttype_to_str(pcb->ie_list[i].node_type),
			   pcb->ie_list[i].num_ports));
		for (toplist = pcb->ie_list[i].port_list, j = 0; toplist; toplist = toplist->next, ++j) {
			FC_DBG(0, ("%s: \tPort[%d]: name: %s, type %s, physical number %d, attached port count %d\n", 
				   pcb->debug_name,
				   j,
				   wwn_to_str(&toplist->port_name, NULL),
				   fctd_porttype_to_str(toplist->port_type),
				   toplist->physical_port_number,
				   toplist->num_atports));
			for (atoplist = toplist->atport_list, k = 0; atoplist; atoplist = (atPortList_t *)atoplist->next, ++k) {
				FC_DBG(0, ("%s: \t\tAttachedPort[%d]: name %s, type %s\n",
					   pcb->debug_name,
					   k,
					   wwn_to_str(&atoplist->port_name, NULL),
					   fctd_porttype_to_str(atoplist->port_type)));
			}
		}
	}
}

/* return topology info block */

int32
fctd_format_topology(char *msg, uint32 *len) 
{
	fctd_t *	pcb = &fctd_cb;
	portList_t *	port;
	atPortList_t *	aport;
	uint32		i, idx = 0;
	uint32		symbolic_name_len;
	wwn_t		swap_wwn;

	WWN_SET(&swap_wwn, &pcb->port_name);
	inplace_htonll(&swap_wwn);	
	WWN_SET(&msg[idx], &swap_wwn); idx += 8;					/* store self port name     - 8 bytes */
	if (pcb->hal_interface_type == HAL_IF_FC_L_PORT) {
		msg[idx++] = 0;								/* store format type        - 1 byte  */
		msg[idx++] = pcb->loop_map.hostid;					/* store host id            - 1 byte  */
		msg[idx++] = pcb->loop_map.port_count;					/* store loop port count    - 1 byte  */
		for (i = 0; i < pcb->loop_map.port_count; ++i)
			msg[idx++] = pcb->loop_map.port_map[i];				/* store loop port          - 1 byte  */
	} 
	else 
	if (pcb->hal_interface_type == HAL_IF_FC_N_PORT) {
		msg[idx++] = 1;								/* store format type        - 1 byte  */
		msg[idx++] = pcb->loop_map.hostid;					/* store host id            - 1 byte  */
		msg[idx++] = pcb->loop_map.port_count;					/* store loop port count    - 1 byte  */
		for (i = 0; i < pcb->loop_map.port_count; ++i)
			msg[idx++] = pcb->loop_map.port_map[i];				/* store loop port          - 1 byte  */
	} 

	else {
		msg[idx++] = 2;								/* store format type        - 1 byte  */		
		msg[idx++] = pcb->ie_count;						/* store IE count	    - 1 byte  */
		for (i = 0; i < pcb->ie_count; i++) {
			WWN_SET(&swap_wwn, &pcb->ie_list[i].node_name);
			inplace_htonll(&swap_wwn);
			WWN_SET(&msg[idx], &swap_wwn); idx += 8;			/* store IE nodename	    - 8 bytes */
			symbolic_name_len = strlen(pcb->ie_list[i].symbolic_name) + 1;
			msg[idx++] = symbolic_name_len;				        /* store IE symb name len   - 1 bytes */
			memcpy(&msg[idx], pcb->ie_list[i].symbolic_name,		/* store IE symb name       - n bytes */
			       symbolic_name_len); idx += symbolic_name_len;
			msg[idx++] = pcb->ie_list[i].node_type;				/* store IE type	    - 1 byte  */
			msg[idx++] = pcb->ie_list[i].num_ports;				/* store num ports	    - 1 byte  */
			for (port = pcb->ie_list[i].port_list; port; port = port->next) {
				WWN_SET(&swap_wwn, &port->port_name);
				inplace_htonll(&swap_wwn);			
				WWN_SET(&msg[idx], &swap_wwn); idx += 8;		/* store port name          - 8 bytes */
				msg[idx++] = port->physical_port_number;		/* store port physical num  - 1 byte  */
				msg[idx++] = port->port_type;				/* store port type          - 1 byte  */
				msg[idx++] = port->num_atports;				/* store num attached ports - 1 byte  */
				for (aport = port->atport_list; aport; aport = (atPortList_t *)aport->next) {
					WWN_SET(&swap_wwn, &aport->port_name);
					inplace_htonll(&swap_wwn); 
					WWN_SET(&msg[idx], &swap_wwn); idx += 8;	/* store attached port name - 8 bytes */
				}
			}
		}
	}
	*len = idx;
#if 0
	printf("topology block size = %d\n", idx+1); 
	for (i = 0; i < idx; i++) {
		if ((i % 16) == 0)
			printf("\n0x%04x: ");
		printf("%02x ", msg[i]);
	}
	printf("\n");
#endif
	return NFX_OK;
}


int32 
fctd_send_ct(ct_desc_t *ctd)
{
	hal_dispatch_entry_t 	*hde = hal_dt_entry_get();

	(*hde->hal_ct_send)(ctd);
	return(NFX_OK);
}

/* send GIEL request */
void 
fctd_giel_req(fctd_t *pcb)
{
	ct_desc_t *		ctd;
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	rsp;

	FC_DBG(1, ("%s: fctd_giel_req...\n", pcb->debug_name));

	ctd = &pcb->ctd;
	ctd->server_addr = CT_ADDRESS_MS;
	ctd->handle = pcb;
	ctd->callback = fctd_ct_callback;
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	rsp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
	bzero(rsp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GIEL;
	req->header.maximum_resp_size = sizeof(rsp->u.GIEL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);

	fctd_send_ct(ctd);
}

/* send GIELN request */
void
fctd_gieln_req(fctd_t *pcb)
{
	ct_desc_t *		ctd;
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	rsp;

	FC_DBG(1, ("%s: fctd_gieln_req...\n", pcb->debug_name));

	ctd = &pcb->ctd;
	ctd->server_addr = CT_ADDRESS_MS;
	ctd->handle = pcb;
	ctd->callback = fctd_ct_callback;
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	rsp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
	bzero(rsp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GIELN;
	req->header.maximum_resp_size = sizeof(rsp->u.GIELN)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);

	WWN_SET(&req->u.GIELN.ie_name, &pcb->ie_list[pcb->ie_index].node_name);

	fctd_send_ct(ctd);
}

/* Send GPL request */
void 
fctd_gpl_req(fctd_t *pcb)
{
	ct_desc_t *		ctd;
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	rsp;

	FC_DBG(1, ("%s: fctd_gpl_req...\n", pcb->debug_name));

	ctd = &pcb->ctd;
	ctd->server_addr = CT_ADDRESS_MS;
	ctd->handle = pcb;
	ctd->callback = fctd_ct_callback;
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	rsp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
	bzero(rsp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GPL;
	req->header.maximum_resp_size = sizeof(rsp->u.GPL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);

	WWN_SET(&req->u.GPL.ie_name, &pcb->ie_list[pcb->ie_index].node_name);

	fctd_send_ct(ctd);
}

/* send GPPN request */
void        
fctd_gppn_req(fctd_t *pcb)
{
	ct_desc_t *		ctd;
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	rsp;

	FC_DBG(1, ("%s: fctd_gppn_req...\n", pcb->debug_name));
	
	ctd = &pcb->ctd;
	ctd->server_addr = CT_ADDRESS_MS;
	ctd->handle = pcb;
	ctd->callback = fctd_ct_callback;
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	rsp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
	bzero(rsp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GPPN;
	req->header.maximum_resp_size = sizeof(rsp->u.GAPNL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);

	WWN_SET(&req->u.GPPN.port_name, &pcb->ie_list[pcb->ie_index].top_list->port_name);

	fctd_send_ct(ctd);

}

/* send GAPNL request */
void 
fctd_gapnl_req(fctd_t *pcb)
{
	ct_desc_t *		ctd;
	ispfc_ct_req_t *	req;
	ispfc_ct_resp_t *	rsp;

	FC_DBG(1, ("%s: fctd_gapnl_req...\n", pcb->debug_name));
	
	ctd = &pcb->ctd;
	ctd->server_addr = CT_ADDRESS_MS;
	ctd->handle = pcb;
	ctd->callback = fctd_ct_callback;
	req = (ispfc_ct_req_t *)ctd->e_desc->bd[1].buf;
	rsp = (ispfc_ct_resp_t *)ctd->e_desc->bd[2].buf;
	bzero(req, sizeof(ispfc_ct_req_t));
	bzero(rsp, sizeof(ispfc_ct_resp_t));

	req->header.revision = 1;
	req->header.fs_type = FS_TYPE_MS;
	req->header.fs_subtype = FS_SUB_TYPE_FABRIC_CONFIG_SERVICE;
	req->header.command_code = CT_MS_CMD_GAPNL;
	req->header.maximum_resp_size = sizeof(rsp->u.GAPNL)/sizeof(uint32);
	inplace_htons(&req->header.command_code);
	inplace_htons(&req->header.maximum_resp_size);

	WWN_SET(&req->u.GAPNL.port_name, &pcb->ie_list[pcb->ie_index].top_list->port_name);

	fctd_send_ct(ctd);
}

void
fctd_get_loop_map_callback(wwn_t channel_wwn, hal_device_map_t *map, int rc, fctd_t *pcb)
{
	esm_event_t	esmEvent;

	FC_DBG(1, ("%s: fctd_get_loop_map_callback: status 0x%x\n", pcb->debug_name, rc));

	if (rc != NFX_OK) {
		FC_DBG(1, ("%s: fctd_get_loop_map_callback: status 0x%x - send fail reply\n", 
			   pcb->debug_name, rc));
		pcb->rsp_id = FCTD_SEND_FAIL_REPLY;
		esmEvent.event = FCTD_EVENT_PROC_MSG;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(pcb->sm_handle, &esmEvent);
	}
	else {
		pcb->rsp_id = FCTD_SEND_GET_LOOP_MAP_REPLY;
		esmEvent.event = FCTD_EVENT_PROC_MSG;
		esmEvent.p1.u.p = NULL;
		esmEvent.p2.u.p = NULL;
		rc = esm_deliverEvent(pcb->sm_handle, &esmEvent);
		VERIFY(rc == NFX_OK);
	}
}

/* FCTD polling routine to check for fabric changes */
void 
fctd_poll(int32 timer_handle, fctd_t *pcb)
{
	hal_dispatch_entry_t *	hde = hal_dt_entry_get();
	ispfc_cb_t *		cb = hde->hal_handle;
	hal_adapter_info_t	info;
	uint32			connection_state;
	int32			rc;
	esm_event_t		esmEvent;

	if (pcb->state == FCTD_IDLE_STATE || pcb->state == FCTD_DISCOVERY_DONE_STATE) {

		rc = (*hde->hal_ioctl)(HAL_IOCTL_COMMON_GET_ADAPTER_INFO, &info);
		if (rc != NFX_OK)
			return;
		rc = (*hde->hal_ioctl)(HAL_IOCTL_FC_CONNECTION_STATE, &connection_state);
		if (rc != NFX_OK || connection_state != HAL_IOCTL_FC_LINK_STATE_LINK_UP)
			return;

		if (info.u.fc_data.fc_state_change_generation_num != pcb->hal_state) {
			pcb->rsp_id = FCTD_SEND_CHANGE_NOTICE;
			esmEvent.event = FCTD_EVENT_PROC_MSG;
			esmEvent.p1.u.p = NULL;
			esmEvent.p2.u.p = NULL;
			rc = esm_deliverEvent(pcb->sm_handle, &esmEvent);
			if (rc != NFX_OK)
				FC_DBG(0, ("%s: fctd_poll: unable to deliver state event\n",
					   cb->debug_name));
		}
	}
}
