/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        lock-api.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Byte range lock API functions.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:    11/23/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*-----------------------------------------------------------------
 * Name:        lock_initFileLockReq
 *
 * Description:
 *
 *     Allocate and initialize a new lock_fileLock_t structure.
 *
 * Arguments:    
 * 
 *     fileLockIdxp - Pointer to the index to the newly allocated
 *                    lock_fileLock_t structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_initFileLockReq(vs_mem_index_t	*fileLockIdxp)
{
	lock_fileLock_t	*fileLockp;

	if (fileLockIdxp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_initFileLockReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.initFileLockReq++;

	/*----------------------------------------
	 * Create a lock_fileLock_t data structure
	 *----------------------------------------
	 */
	fileLockp = lock_memAllocFileLock(fileLockIdxp);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_initFileLockReq: "
							"Failed to allocate a fileLock data structure."));
		return (NFX_ERR);
	}

	return (NFX_OK);
} /* lock_initFileLockReq() */

/*-----------------------------------------------------------------
 * Name:         lock_uninitFileLockReq
 *
 * Description:
 *
 *     Uninitialize an existing lock_fileLock_t structure.
 *
 * Arguments:    
 * 
 *     fileLockIdxp - Pointer to the index to a lock_fileLock_t structure.
 *                    If the data structure is uninit successfully, the
 *                    index becomes zero on return.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_uninitFileLockReq(vs_mem_index_t	*fileLockIdxp)
{
	int32			rc;
	lock_fileLock_t	*fileLockp;

	if (fileLockIdxp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_uninitFileLockReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.uninitFileLockReq++;

	/*---------------------------------
	 * Find the fileLock data structure
	 *---------------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(*fileLockIdxp);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_uninitFileLockReq: "
							"Couldn't find the lockInfo data structure."));
		return (NFX_ERR);
	}

	/*---------------------------------
	 * Free the fileLock data structure
	 *---------------------------------
	 */
	rc = lock_memFreeFileLock(fileLockIdxp, fileLockp);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_uninitFileLockReq: "
							"Failed to delete the lockInfo data structure."));
		return (NFX_ERR);
	}

	return (NFX_OK);
} /* lock_uninitFileLockReq() */

/*-----------------------------------------------------------------
 * Name:         lock_cifsLock
 *
 * Description:
 *
 *     Obtain a CIFS lock.
 *
 * Arguments:
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for byte range lock request.
 *
 * Return value:
 *
 *     LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or LOCK_STATE_BLOCKED -
 *                                                                  If succeed.
 *
 *     lock error code - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_cifsLock(lock_fileLock_t *fileLockp,
			  lock_request_t  *lockReqp)
{
	int32					result, rc;
	lock_cifsLockInfo_t		cifsLockInfo;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	if (lockReqp->lockType == LOCK_TYPE_EXCL)
		lock_globalCfgp->debugCounters.cifsExclByteRangeLockReq++;
	else if (lockReqp->lockType == LOCK_TYPE_SHAR)
		lock_globalCfgp->debugCounters.cifsSharedByteRangeLockReq++;

	/*-------------------------------------
	 * Init the cifsLockInfo data structure
	 *-------------------------------------
	 */
	rc = lock_initCifsLockInfo(lockReqp, &cifsLockInfo);
	if (rc == NFX_ERR)
		return (LOCK_ERR_GENERAL);

	LOCK_PRINTF(LOG_DEBUG, ("lock_cifsLock: "
							"offset(%lld) len(%lld)",
							cifsLockInfo.offset, cifsLockInfo.len));

	/*---------------------
	 * Process lock request
	 *---------------------
	 */
	result = lock_procLock(fileLockp,
						   (void *)&cifsLockInfo,
						   LOCK_OWNER_CIFS,
						   lockReqp->lockType,
						   lockReqp->timeout,
						   lockReqp->ulPrivate);

	return (result);
} /* lock_cifsLock() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmLock
 *
 * Description:
 *
 *     Obtain an NLM lock.
 *
 * Arguments:
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for byte range lock request.
 *
 * Return value:
 *
 *     LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or LOCK_STATE_BLOCKED -
 *                                                                  If succeed.
 *
 *     lock error code - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_nlmLock(lock_fileLock_t *fileLockp,
			 lock_request_t  *lockReqp)
{
	int32					result, rc;
	lock_nlmLockInfo_t		nlmLockInfo;

	if (lockReqp->len == 0)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_nlmLock: "
							"Byte range length should be greater than 0."));
		return (LOCK_ERR_INVALID_INPUT);
	}

	if ((eee_rtcCompare(&lockReqp->timeout, &lock_timeZero) != 0) &&
		(eee_rtcCompare(&lockReqp->timeout, &lock_timeIndefinite) != 0))
	{
		LOCK_PRINTF(LOG_ERR, ("lock_nlmLock: "
							"Timeout value should be either 0 or -1."));
		return (LOCK_ERR_INVALID_INPUT);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	if (lockReqp->lockType == LOCK_TYPE_EXCL)
		lock_globalCfgp->debugCounters.nlmExclByteRangeLockReq++;
	else if (lockReqp->lockType == LOCK_TYPE_SHAR)
		lock_globalCfgp->debugCounters.nlmSharedByteRangeLockReq++;

	/*------------------------------------
	 * Init the nlmLockInfo data structure
	 *------------------------------------
	 */
	rc = lock_initNlmLockInfo(lockReqp, &nlmLockInfo);
	if (rc == NFX_ERR)
		return (LOCK_ERR_GENERAL);

	LOCK_PRINTF(LOG_DEBUG, ("lock_nlmLock: "
								"offset(%lld) len(%lld))",
								nlmLockInfo.offset, nlmLockInfo.len));

	/*---------------------
	 * Process lock request
	 *---------------------
	 */
	result = lock_procLock(fileLockp,
						   (void *)&nlmLockInfo,
						   LOCK_OWNER_NLM,
						   lockReqp->lockType,
						   lockReqp->timeout,
						   lockReqp->ulPrivate);

	return (result);
} /* lock_nlmLock() */

/*-----------------------------------------------------------------
 * Name:         lock_byteRangeLockReq
 *
 * Description:
 *
 *     Lock a byte range.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for byte range lock request.
 *
 * Return value:
 *
 *     LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or LOCK_STATE_BLOCKED -
 *                                                                  If succeed.
 *
 *     lock error code - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_byteRangeLockReq(vs_mem_index_t  fileLockIdx,
				      lock_request_t  *lockReqp)
{
	int32			rc;
	lock_fileLock_t	*fileLockp;

	if (lockReqp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_byteRangeLockReq: NULL input pointer"));
		return (LOCK_ERR_GENERAL);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.byteRangeLockReq++;

	/*------------------------------------------
	 * Check if the locks already reached
	 * the per filer limit (or per file limit).
	 *------------------------------------------
	 */
	if (lock_globalCfgp->lockCount == LOCK_MAX_LOCKS)
	{
		/* Reached the limit */
		LOCK_PRINTF(LOG_ERR, ("lock_byteRangeLockReq: "
								"Locks reached per filer limit."));
		return (LOCK_ERR_REACH_LIMIT);
	}

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_byteRangeLockReq: "
								"Couldn't find fileLock data structure."));
		return (LOCK_ERR_GENERAL);
	}

	/*----------------------------------------
	 * Process the request based on owner type
	 *----------------------------------------
	 */
	switch (lockReqp->ownerType)
	{
		case LOCK_OWNER_CIFS:
			rc = lock_cifsLock(fileLockp, lockReqp);
			break;

		case LOCK_OWNER_NLM:
			rc = lock_nlmLock(fileLockp, lockReqp);
			break;

		default:
			LOCK_PRINTF(LOG_ERR, ("lock_byteRangeLockReq: "
								"Lock owner type (%d) is not supported.",
								lockReqp->ownerType));
			return (LOCK_ERR_UNSUPPORTED_OWNER_TYPE);
	}

	LOCK_PRINTF(LOG_DEBUG, ("lock_byteRangeLockReq: lock result: %s",
				(rc == LOCK_STATE_GRANTED) ?  "granted" :
	   					((rc == LOCK_STATE_BLOCKED) ? "blocked" : "denied")));

	return (rc);
} /* lock_byteRangeLockReq() */

/*-----------------------------------------------------------------
 * Name:         lock_cifsUnlockSingle
 *
 * Description:
 *
 *     Unlock a CIFS lock.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for unlock a single byte range request.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_cifsUnlockSingle(lock_fileLock_t *fileLockp,
					  lock_request_t  *lockReqp)
{
	int32					rc;
	lock_cifsLockInfo_t		cifsLockInfo;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.cifsUnlockSingleReq++;

	/*-------------------------------------
	 * Init the cifsLockInfo data structure
	 *-------------------------------------
	 */
	rc = lock_initCifsLockInfo(lockReqp, &cifsLockInfo);
	if (rc == NFX_ERR)
		return (NFX_ERR);

	LOCK_PRINTF(LOG_DEBUG, ("lock_cifsUnlockSingle: "
							"offset(%lld) len(%lld)",
							cifsLockInfo.offset, cifsLockInfo.len));

	/*-----------------------
	 * Process unlock request
	 *-----------------------
	 */
	rc = lock_procUnlockSingle(fileLockp,
							   (void *)&cifsLockInfo,
							   LOCK_OWNER_CIFS,
							   lockReqp->lockType);

	return (rc);
} /* lock_cifsUnlockSingle() */

/*-----------------------------------------------------------------
 * Name:         lock_nlmUnlockSingle
 *
 * Description:
 *
 *     Unlock an NLM lock.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for unlock a single byte range request.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_nlmUnlockSingle(lock_fileLock_t *fileLockp,
					 lock_request_t  *lockReqp)
{
	int32					rc;
	lock_nlmLockInfo_t		nlmLockInfo;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.nlmUnlockSingleReq++;

	/*------------------------------------
	 * Init the nlmLockInfo data structure
	 *------------------------------------
	 */
	rc = lock_initNlmLockInfo(lockReqp, &nlmLockInfo);
	if (rc == NFX_ERR)
		return (NFX_ERR);

	LOCK_PRINTF(LOG_DEBUG, ("lock_nlmUnlockSingle: "
								"offset(%lld) len(%lld))",
								nlmLockInfo.offset, nlmLockInfo.len));

	/*-----------------------
	 * Process unlock request
	 *-----------------------
	 */
	rc = lock_procUnlockSingle(fileLockp,
							   (void *)&nlmLockInfo,
							   LOCK_OWNER_NLM,
							   lockReqp->lockType);

	return (rc);
} /* lock_nlmUnlockSingle() */

/*-----------------------------------------------------------------
 * Name:         lock_unlockSingleReq
 *
 * Description:
 *
 *     Unlock a single byte range.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for unlock a single byte range request.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockSingleReq(vs_mem_index_t  fileLockIdx,
				     lock_request_t  *lockReqp)
{
	int32			rc;
	lock_fileLock_t	*fileLockp;

	if (lockReqp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_unlockSingleReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.unlockSingleReq++;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_unlockSingleReq: "
								"Couldn't find fileLock structrue."));
		return (NFX_ERR);
	}

	/*----------------------------------------
	 * Process the request based on owner type
	 *----------------------------------------
	 */
	switch (lockReqp->ownerType)
	{
		case LOCK_OWNER_CIFS:
			rc = lock_cifsUnlockSingle(fileLockp, lockReqp);
			break;

		case LOCK_OWNER_NLM:
			rc = lock_nlmUnlockSingle(fileLockp, lockReqp);
			break;

		default:
			LOCK_PRINTF(LOG_ERR, ("lock_unlockSingleReq: "
								"Lock owner type is not supported."));
			return (NFX_ERR);
	}

	LOCK_PRINTF(LOG_DEBUG, ("lock_unlockSingleReq: unlock result: %d", rc));

	return (rc);
} /* lock_unlockSingleReq() */

/*-----------------------------------------------------------------
 * Name:         lock_unlockAllByFidReq
 *
 * Description:
 *
 *     Release all locks held on an open file handle (CIFS only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     fid - File ID, representing an open file handle.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockAllByFidReq(vs_mem_index_t	fileLockIdx,
					   uint16			fid)
{
	int32	rc;
	uint32	key;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.unlockAllByFidReq++;

	/*-----------------------------------------
	 * Remove locks held on an open file handle
	 *-----------------------------------------
	 */
	key = LOCK_GENERATE_KEY(fid, 0);
	rc = lock_procUnlockAll(fileLockIdx, key, 0, LOCK_SEARCH_SAME_FID);

	return (rc);
} /* lock_unlockAllByFidReq() */

/*-----------------------------------------------------------------
 * Name:         lock_unlockAllByKeyReq
 *
 * Description:
 *
 *     Release all locks held on an open file handle that
 *     match a key (CIFS only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     fid - File ID, representing an open file handle.
 *
 *     pid - Process ID.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockAllByKeyReq(vs_mem_index_t	fileLockIdx,
					   uint16			fid,
					   uint16			pid)
{
	int32	rc;
	uint32	key;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.unlockAllByKeyReq++;

	/*--------------------------------
	 * Remove locks that match the key
	 *--------------------------------
	 */
	key = LOCK_GENERATE_KEY(fid, pid);
	rc = lock_procUnlockAll(fileLockIdx, key, 0, LOCK_SEARCH_SAME_KEY);

	return (rc);
} /* lock_unlockAllByKeyReq() */

/*-----------------------------------------------------------------
 * Name:         lock_unlockAllByIpAddrReq
 *
 * Description:
 *
 *     Release all locks held by a client (NLM only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     ipAddr - Client IP address.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/26/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockAllByIpAddrReq(vs_mem_index_t	fileLockIdx,
						  ip_index_t		ipAddr)
{
	int32			rc;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.unlockAllByIpAddrReq++;

	rc = lock_procUnlockAll(fileLockIdx, 0, ipAddr, LOCK_SEARCH_SAME_IPADDR);

	return (rc);
} /* lock_unlockAllByIpAddrReq() */

/*-----------------------------------------------------------------
 * Name:         lock_cancelLockReq
 *
 * Description:
 *
 *     Cancel a pending lock.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for cancel lock request.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 *     LOCK_ERR_NOT_FOUND - If the lock is not found.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_cancelLockReq(vs_mem_index_t	fileLockIdx,
				   lock_request_t	*lockReqp)
{
	int32						rc;
	lock_fileLock_t				*fileLockp;
	lock_cifsLockInfo_t			cifsLockInfo;
	lock_nlmLockInfo_t			nlmLockInfo;
	void						*lockInfop;
	lock_delPendLocksParam_t	param;

	if (lockReqp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_cancelLockReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.cancelLockReq++;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_cancelLockReq: "
								"Couldn't find fileLock structrue."));
		return (NFX_ERR);
	}

	/*-------------------------------------------
	 * Get lockInfo structure based on owner type
	 *-------------------------------------------
	 */
	switch (lockReqp->ownerType)
	{
		case LOCK_OWNER_CIFS:
		{
			/*----------------------------
			 * Update global debug counter
			 *----------------------------
			 */
			lock_globalCfgp->debugCounters.cifsCancelLockReq++;

			/*-------------------------------------
			 * Init the cifsLockInfo data structure
			 *-------------------------------------
			 */
			rc = lock_initCifsLockInfo(lockReqp, &cifsLockInfo);
			if (rc == NFX_ERR)
				return (NFX_ERR);

			lockInfop = (void *)&cifsLockInfo;

			LOCK_PRINTF(LOG_DEBUG, ("lock_cancelLockReq: "
									"offset(%lld) len(%lld)",
									cifsLockInfo.offset, cifsLockInfo.len));
			break;
		}

		case LOCK_OWNER_NLM:
		{
			/*----------------------------
			 * Update global debug counter
			 *----------------------------
			 */
			lock_globalCfgp->debugCounters.nlmCancelLockReq++;

			/*------------------------------------
			 * Init the nlmLockInfo data structure
			 *------------------------------------
			 */
			rc = lock_initNlmLockInfo(lockReqp, &nlmLockInfo);
			if (rc == NFX_ERR)
			{
				return (NFX_ERR);
			}
			lockInfop = (void *)&nlmLockInfo;

			LOCK_PRINTF(LOG_DEBUG, ("lock_cancelLockReq: "
								"offset(%lld) len(%lld))",
								nlmLockInfo.offset, nlmLockInfo.len));
			break;
		}

		default:
			LOCK_PRINTF(LOG_ERR, ("lock_cancelLockReq: "
								"Lock owner type (%d) is not supported.",
								lockReqp->ownerType));
			return (NFX_ERR);
	}

	/*---------------------------------
	 * Delete the pending lock if found
	 *---------------------------------
	 */
	param.type = LOCK_DELPENDLOCKS_TYPE_UNLOCKSINGLE;
	param.u.delSingle.lockInfop = lockInfop;
	param.u.delSingle.lockOwnerType = lockReqp->ownerType;
	param.u.delSingle.searchMask =
					(LOCK_SEARCH_SAME_RANGE | LOCK_SEARCH_SAME_OWNER);

	rc = lock_deletePendingLocksFromLocalList(fileLockp, param);

	return (rc);
} /*lock_cancelLockReq() */

/*-----------------------------------------------------------------
 * Name:         lock_testLockReq
 *
 * Description:
 *
 *     Test to see whether a given lock can be granted or not (NLM only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for test lock request.
 *
 *     testResultp - This is output. It contains the pointer to a structure
 *                   that contains testing result.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/27/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_testLockReq(vs_mem_index_t		fileLockIdx,
				 lock_request_t		*lockReqp,
				 lock_testResult_t	**testResultp)
{
	int32						rc, result, dummyResult;
	lock_fileLock_t				*fileLockp;
	lock_nlmLockInfo_t			nlmLockInfo;
	vs_mem_index_t				confLockIdx;
	lock_testResult_t			testResult;

	if (lockReqp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_testLockReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------
	 * Init result structure
	 *----------------------
	 */
	bzero(&testResult, sizeof(testResult));
	*testResultp = &testResult;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.testLockReq++;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_testLockReq: "
								"Couldn't find fileLock structrue."));
		return (NFX_ERR);
	}

	/*------------------------------------
	 * Init the nlmLockInfo data structure
	 *------------------------------------
	 */
	if (lockReqp->ownerType != LOCK_OWNER_NLM)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_testLockReq: "
								"Lock owner type (%d) is not supported.",
								lockReqp->ownerType));
		return (NFX_ERR);
	}

	rc = lock_initNlmLockInfo(lockReqp, &nlmLockInfo);
	if (rc != NFX_OK)
		return (NFX_ERR);

	LOCK_PRINTF(LOG_DEBUG, ("lock_testLockReq: "
								"offset(%lld) len(%lld))",
								nlmLockInfo.offset, nlmLockInfo.len));

	/*------------------------------------------------------------------
	 * Check conflict. If there is no conflict, the lock can be granted.
	 *------------------------------------------------------------------
	 */
	result = lock_checkConflict(fileLockp,
								(void *)&nlmLockInfo,
								LOCK_OWNER_NLM,
								lockReqp->lockType,
								FALSE,
								&confLockIdx,
								&dummyResult,
								&dummyResult);
	rc = lock_fillNlmTestResult(result, confLockIdx, &testResult);

	return (rc);
} /* lock_testLockReq() */

/*-----------------------------------------------------------------
 * Name:         lock_checkLockForWriteReq
 *
 * Description:
 *
 *     Called for write access to a range of bytes (CIFS only).
 *     Called by CIFS write to see if there is a conflicting mandatory lock.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for request of checking write access
 *                to a range of bytes.
 *
 * Return value:
 *
 *     NFX_OK - If write access is allowed.
 *
 *     NFX_ERR - If write access is denied.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkLockForWriteReq(vs_mem_index_t  fileLockIdx,
						  lock_request_t  *lockReqp)
{
	int32					rc;
	lock_fileLock_t			*fileLockp;
	lock_cifsLockInfo_t		cifsLockInfo;

	if (lockReqp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_checkLockForWriteReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.checkLockForWriteReq++;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_checkLockForWriteReq: "
								"Couldn't find fileLock structrue."));
		return (NFX_ERR);
	}

	/*------------------------
	 * Check zero-length write
	 *------------------------
	 */
	if (lockReqp->len == 0)
	{
		/* A zero-length write does not conflict with any locks. */
		return (NFX_OK);
	}

	/*-------------------------------------
	 * Init the cifsLockInfo data structure
	 *-------------------------------------
	 */
	rc = lock_initCifsLockInfo(lockReqp, &cifsLockInfo);
	if (rc == NFX_ERR)
		return (LOCK_ERR_GENERAL);

	LOCK_PRINTF(LOG_DEBUG, ("lock_lock_checkLockForWriteReq: "
							"offset(%lld) len(%lld)",
							cifsLockInfo.offset, cifsLockInfo.len));

	/*-----------------------------------
	 * Check if access is permitted or not
	 *-----------------------------------
	 */
	rc = lock_checkAccess(fileLockp,
						  &cifsLockInfo,
						  TRUE);
	return (rc);
} /* lock_checkLockForWriteReq() */

/*-----------------------------------------------------------------
 * Name:         lock_checkLockForReadReq
 *
 * Description:
 *
 *     Called for read access to a range of bytes (CIFS only).
 *     Called by CIFS read to see if there is a conflicting mandatory lock.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for request of checking read access
 *                to a range of bytes.
 *
 * Return value:
 *
 *     NFX_OK - If read access is allowed.
 *
 *     NFX_ERR - If read access is denied.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkLockForReadReq(vs_mem_index_t  fileLockIdx,
						 lock_request_t  *lockReqp)
{
	int32					rc;
	lock_fileLock_t			*fileLockp;
	lock_cifsLockInfo_t		cifsLockInfo;

	if (lockReqp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_checkLockForReadReq: NULL input pointer"));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.checkLockForReadReq++;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_checkLockForReadReq: "
								"Couldn't find fileLock structrue."));
		return (NFX_ERR);
	}

	/*-----------------------
	 * Check zero-length read
	 *-----------------------
	 */
	if (lockReqp->len == 0)
	{
		/* A zero-length read does not conflict with any locks. */
		return (NFX_OK);
	}

	/*-------------------------------------
	 * Init the cifsLockInfo data structure
	 *-------------------------------------
	 */
	rc = lock_initCifsLockInfo(lockReqp, &cifsLockInfo);
	if (rc == NFX_ERR)
		return (LOCK_ERR_GENERAL);

	LOCK_PRINTF(LOG_DEBUG, ("lock_lock_checkLockForReadReq: "
							"offset(%lld) len(%lld)",
							cifsLockInfo.offset, cifsLockInfo.len));

	/*-----------------------------------
	 * Check if access is permitted or not
	 *-----------------------------------
	 */
	rc = lock_checkAccess(fileLockp,
						  &cifsLockInfo,
						  FALSE);
	return (rc);
} /* lock_checkLockForReadReq() */

/*-----------------------------------------------------------------
 * Name:         lock_areThereCurrentFileLocksReq
 *
 * Description:
 *
 *     Check if there are any locks currently assigned to a file or not.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 * Return value:
 *
 *     TRUE - There are lock currently assigned to a file.
 *
 *     FALSE - No lock currently assigned to a file.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/27/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_areThereCurrentFileLocksReq(vs_mem_index_t fileLockIdx)
{
	lock_fileLock_t			*fileLockp;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.areThereCurrentFileLocksReq++;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_areThereCurrentFileLocksReq: "
								"Couldn't find fileLock structrue."));
		return (FALSE);
	}

	if (!VS_MEM_IDX_IS_NULL(fileLockp->exclLockTree))
		return (TRUE);


	if (!VS_MEM_IDX_IS_NULL(fileLockp->sharedLockTree))
		return (TRUE);

	if (!VS_MEM_IDX_IS_NULL(fileLockp->pendingLockList))
		return (TRUE);

	return (FALSE);
} /* lock_areThereCurrentFileLocksReq() */

/*-----------------------------------------------------------------
 * Name:         lock_getZeroTimeout
 *
 * Description:
 *
 *     Called by the CIFS or NLM module to get a 0 value timeout
 *     in real time clock form.
 *
 * Arguments:
 *
 *     rtcTimeout - Zero timeout in real time clock form.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
inline void
lock_getZeroTimeout(eeeRTC_t    *rtcTimeout)
{
    *rtcTimeout = lock_timeZero;
    return;
}

/*-----------------------------------------------------------------
 * Name:         lock_getIndefiniteTimeout
 *
 * Description:
 *
 *     Called by the CIFS or NLM module to get a timeout value
 *     in real time clock form to represent indefinite timeout.
 *
 * Arguments:    
 * 
 *     rtcTimeout - Indefinite timeout in real time clock form.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
inline void
lock_getIndefiniteTimeout(eeeRTC_t  *rtcTimeout)
{
    *rtcTimeout = lock_timeIndefinite;
    return;
}

