/* vi:ts=4
 *
 *-----------------------------------------------------------------
 *
 * Name:			lock-btree.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Binary tree functions.
 *              (The term "btree" used in the code means binary btree,
 *               not B-tree.)
 *
 * Created by:		Xiaobing Zhang
 *
 * Date Created:	10/29/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*-----------------------------------------------------------------
 * Name:         lock_btree_updateSubTreeCurr
 *
 * Description:
 *
 *     This routine is part of splay routine. It changes the current
 *     pointer to one of the subtrees, which are used during splay.
 *     It pushes the pointer down to the bottom left node for the 
 *     right subtree, or the bottom right node for the left subtree. 
 *
 * Arguments:    
 * 
 *     subTreeCurrp - Pointer to pointer to a node in the subtree.
 *
 *     subTreeCurrIdxp - Pointer to index to a node in the subtree.
 *
 *     dir - Left or right subtree.
 *
 * Return value:
 *
 *      None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
static inline void
lock_btree_updateSubTreeCurr(lock_btNode_t	**subTreeCurrp,
							 vs_mem_index_t	*subTreeCurrIdxp,
							 uint8			dir)
{
	lock_btNode_t	*nodep, *childNodep;
	vs_mem_index_t	nodeIdx;

	nodeIdx = *subTreeCurrIdxp;
	nodep = *subTreeCurrp;
	childNodep = (lock_btNode_t *)vs_mem_idxToPtr(nodep->child[dir]);
	while(childNodep)
	{
		nodeIdx = nodep->child[dir];
		nodep = childNodep;
		childNodep = (lock_btNode_t *)vs_mem_idxToPtr(nodep->child[dir]);
	}
	*subTreeCurrIdxp = nodeIdx;
	*subTreeCurrp = nodep;

} /*lock_btree_updateSubTreeCurr()*/

/*-----------------------------------------------------------------
 * Name:         lock_btree_splay
 *
 * Description:
 *
 *      Top-down splay.
 *
 * Arguments:    
 * 
 *     rootIdxp - Pointer to the index to the root of the tree.
 *
 *     inNodep - Pointer to the input node.
 *
 *     inLockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                   lock_nlmLockInfo_t structure. It contains lock
 *                   information, which we compare a tree node with.
 *
 *     inLockOwnerType - Type of lock owner.
 *                       Value is LOCK_OWNER_CIFS or LOCK_OWNER_NLM.
 *
 *     useNode - TRUE if inNodep is valid.
 *               FALSE if inLockInfop and inLockOwnerType are valid.
 *
 *
 *     searchMask - The criteria based on which we compare two nodes
 *                  or compare the lockInfo with a node in the tree.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 *     treeType - Exclusive or shared lock tree.
 *                Value is LOCK_TYPE_EXCL or LOCK_TYPE_SHAR.
 *
 *     rootNodep - This is output. It points to the new root of the tree.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
static int8
lock_btree_splay(vs_mem_index_t *rootIdxp,
				 lock_btNode_t  *inNodep,
				 void			*inLockInfop,
				 uint8			inLockOwnerType,
				 boolean		useNode,
				 uint16			searchMask,
				 int8			specialCompResult,
				 uint8			treeType,
				 lock_btNode_t	**rootNodep)
{
	vs_mem_index_t	currNodeIdx;      /*Current node in the tree*/
	lock_btNode_t	*currNodep;
	vs_mem_index_t	childNodeIdx;     /*currNode's child*/
	lock_btNode_t	*childNodep, *childNode2p;
	vs_mem_index_t	gchildNodeIdx;    /*currNode's grandchild*/
	lock_btNode_t	*gchildNodep, *gchildNode2p;

	lock_btNode_t	subTreeRoot[2];   /*Its children point to roots of future
									    left and right subtrees*/
	vs_mem_index_t	subTreeCurrIdx[2];
	lock_btNode_t	*subTreeCurrp[2]; /*Its children point to current nodes
										in the two subtrees*/
	int8			currComp;		  /*Comparison result of inNode/inLockInfo
										and currNode*/
	int8			childComp;		  /*Comparison result of inNode/inLockInfo
									   	and currNode's child*/
	uint8			childDir;		  /*Direction from currNode to its child*/
	uint8			gchildDir;		  /*Direction from currNode's child to its
									    grandchild*/

	boolean			zeroInput = FALSE;/*TRUE if inNodep is NULL or
									   	inLockInfop is NULL*/

	if (useNode == TRUE)
	{
		if (inNodep == NULL)
			zeroInput = TRUE;
	}
	else
	{
		if (inLockInfop == NULL)
			zeroInput = TRUE;
	}

	/*------------------------
	 * Get pointer to rootNode
	 *------------------------
	 */
	*rootNodep = (lock_btNode_t *)vs_mem_idxToPtr(*rootIdxp);

	/*------------------------------------------
	 * Init currNodep. Point to root of the tree
	 *------------------------------------------
	 */
	currNodep = *rootNodep;
	currNodeIdx = *rootIdxp;
	if (currNodep == NULL)
	{
		/*Empty tree*/
		return (1);		/*There is no match*/
	}
	
	/*--------------------------
	 * Create two empty subtrees
	 *--------------------------
	 */
	VS_MEM_IDX_SET_NULL(subTreeRoot[LOCK_BT_LEFT].child[LOCK_BT_RIGHT]);
	VS_MEM_IDX_SET_NULL(subTreeRoot[LOCK_BT_RIGHT].child[LOCK_BT_LEFT]);

	subTreeCurrp[LOCK_BT_LEFT] = &subTreeRoot[LOCK_BT_LEFT];
	subTreeCurrp[LOCK_BT_RIGHT] = &subTreeRoot[LOCK_BT_RIGHT];
	VS_MEM_IDX_SET_NULL(subTreeCurrIdx[LOCK_BT_LEFT]);
	VS_MEM_IDX_SET_NULL(subTreeCurrIdx[LOCK_BT_RIGHT]);

	/*---------------
	 * Split the tree
	 *---------------
	 */
	for (;;)
	{
		/*
		 * We are examining currNode.
		 * First find out which of currNode's children we should check next.
		 */
		if (zeroInput == TRUE)
			currComp = 1;
		else
			currComp = lock_compFunc(inNodep,
									 inLockInfop,
									 inLockOwnerType,
									 useNode,
									 currNodep,
									 searchMask,
									 specialCompResult,
									 treeType);

		childDir = (currComp < 0);
		childNodeIdx = currNodep->child[childDir];
		childNodep = (lock_btNode_t *)vs_mem_idxToPtr(childNodeIdx);
		if ((currComp == 0) || (childNodep == NULL))
			break;

		/*
		 * Now find currNode's grandchild
		 */
		if (zeroInput == TRUE)
			childComp = 1;
		else
			childComp = lock_compFunc(inNodep,
									  inLockInfop,
									  inLockOwnerType,
									  useNode,
									  childNodep,
									  searchMask,
									  specialCompResult,
									  treeType);

		gchildDir = (childComp < 0);
		gchildNodeIdx = childNodep->child[gchildDir];
		gchildNodep = (lock_btNode_t *)vs_mem_idxToPtr(gchildNodeIdx);

		if ((childComp == 0) || (gchildNodep == NULL))
		{
			/*
			 * If a match for inNode/inLockInfo is found (childComp == 0) or
			 * if there is no further node to go (gchildNodep == NULL), 
			 * then it is done.
			 * The node pointed by childNodep will be the root of the new tree.
			 * This is the point to exit from the loop.
			 */

			/*Break link between currNodep and childNodep*/
			VS_MEM_IDX_SET_NULL(currNodep->child[childDir]);

			/*Hang currNodep on subTreeRoot*/
			subTreeCurrp[!childDir]->child[childDir] = currNodeIdx;
			currNodep->parent = subTreeCurrIdx[!childDir];
			
			/*Push subTreeCurrp to bottom*/
			lock_btree_updateSubTreeCurr(&subTreeCurrp[!childDir],
										 &subTreeCurrIdx[!childDir],
						   				 childDir);

			/* Advance currNodep to childNodep*/
			currNodeIdx = childNodeIdx;
			currNodep = childNodep;
			currComp = childComp;
			break;		/*Proceed to tree reconstruction*/
		}
		else
		{
			/*
			 * Split up the tree
			 */
			if (childDir == gchildDir)
			{
				/*zig-zig*/
				currNodep->child[childDir] = childNodep->child[!childDir];
				gchildNode2p = (lock_btNode_t *)
								vs_mem_idxToPtr(childNodep->child[!childDir]);
				if (gchildNode2p != NULL)
					gchildNode2p->parent = currNodeIdx;

				childNodep->child[!childDir] = currNodeIdx;
				currNodep->parent = childNodeIdx;

				VS_MEM_IDX_SET_NULL(childNodep->child[childDir]);

				subTreeCurrp[!childDir]->child[childDir] = childNodeIdx;
				childNodep->parent = subTreeCurrIdx[!childDir];

				lock_btree_updateSubTreeCurr(&subTreeCurrp[!childDir],
										 	 &subTreeCurrIdx[!childDir],
											 childDir);
			}
			else
			{
				/*zig-zag*/
				VS_MEM_IDX_SET_NULL(currNodep->child[childDir]);
				subTreeCurrp[!childDir]->child[childDir] = currNodeIdx;
				currNodep->parent = subTreeCurrIdx[!childDir];

				lock_btree_updateSubTreeCurr(&subTreeCurrp[!childDir],
										 	 &subTreeCurrIdx[!childDir],
											 childDir);

				VS_MEM_IDX_SET_NULL(childNodep->child[gchildDir]);

				subTreeCurrp[!gchildDir]->child[gchildDir] = childNodeIdx;
				childNodep->parent = subTreeCurrIdx[!gchildDir];

				lock_btree_updateSubTreeCurr(&subTreeCurrp[!gchildDir],
										 	 &subTreeCurrIdx[!gchildDir],
							   				 gchildDir);
			}

			currNodeIdx = gchildNodeIdx;
			currNodep = gchildNodep;
		}
	} /*for(;;)*/

	/*
	 * Put it all together
	 */
	subTreeCurrp[LOCK_BT_LEFT]->child[LOCK_BT_RIGHT] =
									currNodep->child[LOCK_BT_LEFT];
	childNode2p = (lock_btNode_t *)
				vs_mem_idxToPtr(currNodep->child[LOCK_BT_LEFT]);
	if (childNode2p != NULL)
		childNode2p->parent = subTreeCurrIdx[LOCK_BT_LEFT];

	subTreeCurrp[LOCK_BT_RIGHT]->child[LOCK_BT_LEFT] =
									currNodep->child[LOCK_BT_RIGHT];
	childNode2p = (lock_btNode_t *)
				vs_mem_idxToPtr(currNodep->child[LOCK_BT_RIGHT]);
	if (childNode2p != NULL)
		childNode2p->parent = subTreeCurrIdx[LOCK_BT_RIGHT];

	currNodep->child[LOCK_BT_LEFT] = 
						subTreeRoot[LOCK_BT_LEFT].child[LOCK_BT_RIGHT];
	childNode2p =(lock_btNode_t *)
				vs_mem_idxToPtr(subTreeRoot[LOCK_BT_LEFT].child[LOCK_BT_RIGHT]);
	if (childNode2p != NULL)
		childNode2p->parent = currNodeIdx;

	currNodep->child[LOCK_BT_RIGHT] =
						subTreeRoot[LOCK_BT_RIGHT].child[LOCK_BT_LEFT];
	childNode2p =(lock_btNode_t *)
				vs_mem_idxToPtr(subTreeRoot[LOCK_BT_RIGHT].child[LOCK_BT_LEFT]);
	if (childNode2p != NULL)
		childNode2p->parent = currNodeIdx;

	*rootIdxp = currNodeIdx;
	*rootNodep = currNodep;
	VS_MEM_IDX_SET_NULL((*rootNodep)->parent);

	return (currComp);
} /*lock_btree_splay()*/

/*-----------------------------------------------------------------
 * Name:         lock_btree_insert
 *
 * Description:
 *
 *      Insert a node to a btree.
 *
 * Arguments:    
 * 
 *     rootIdxp - Pointer to the index to the root of the tree.
 *
 *     inNodeIdx - Index to the node to be inserted in the tree.
 *
 *     inNodep - Pointer to the node to be inserted in the tree.
 *
 *     searchMask - The criteria based on which we compare two nodes
 *                  in the tree.
 *
 *     treeType - Exclusive or shared lock tree.
 *                Value is LOCK_TYPE_EXCL or LOCK_TYPE_SHAR.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_btree_insert(vs_mem_index_t	*rootIdxp,
				  vs_mem_index_t	inNodeIdx,
				  lock_btNode_t		*inNodep,
				  uint16			searchMask,
				  uint8				treeType)
{
	int8			comp;
   	uint8			dir;
	lock_btNode_t	*rootNodep, *childNodep;

	comp = lock_btree_splay(rootIdxp,
							inNodep,
							NULL,
							LOCK_TYPE_DUMMY,
							TRUE,
							searchMask,
							LOCK_SPECIAL_COMP_RESULT,
							treeType,
							&rootNodep);
	if (comp == 0)
	{
		/*There is conflict*/
		return (NFX_ERR);
	}

	if (rootNodep != NULL)
	{
		/*Tree is not empty*/
		dir = (comp < 0);
		inNodep->child[dir] = rootNodep->child[dir];
		childNodep = (lock_btNode_t *)vs_mem_idxToPtr(rootNodep->child[dir]);
		if (childNodep != NULL)
			childNodep->parent = inNodeIdx;

		VS_MEM_IDX_SET_NULL(rootNodep->child[dir]);

		inNodep->child[!dir] = *rootIdxp;
		rootNodep->parent = inNodeIdx;
	}
	*rootIdxp = inNodeIdx;

	return (NFX_OK);
} /*lock_btree_insert()*/

/*-----------------------------------------------------------------
 * Name:         lock_btree_find
 *
 * Description:
 *
 *      Find a node in a btree based on criteria given in searchMask.
 *
 * Arguments:    
 * 
 *     rootIdxp - Pointer to the index to the root of the tree.
 *
 *     inNodep - Pointer to the node which we will look for in the tree.
 *
 *     inLockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                   lock_nlmLockInfo_t structure. It contains lock
 *                   information based on which we search for a node.
 *     inLockOwnerType - Type of lock owner.
 *                       Value is LOCK_OWNER_CIFS or LOCK_OWNER_NLM.
 *
 *     useNode - TRUE if inNodep is valid.
 *               FALSE if inLockInfop and inLockOwnerType are valid.
 *
 *     searchMask - The criteria based on which we search for
 *                  a matching node.
 *
 *     treeType - Exclusive or shared lock tree.
 *                Value is LOCK_TYPE_EXCL or LOCK_TYPE_SHAR.
 *
 * Return value:
 *
 *     Pointer to the found node or NULL if no matching node is found.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
lock_btNode_t *
lock_btree_find(vs_mem_index_t	*rootIdxp,
				lock_btNode_t	*inNodep,
				void			*inLockInfop,
				uint8			inLockOwnerType,
				boolean			useNode,
				uint16			searchMask,
				uint8			treeType)
{
	int8			comp, specialComp;
	boolean			specialSearch;
	lock_btNode_t	*rootNodep;


	/*
	 * Set "special search" flag
	 */
	if (searchMask & (LOCK_SEARCH_DIFF_RANGE |
					  LOCK_SEARCH_SAME_OWNER |
					  LOCK_SEARCH_DIFF_OWNER |
					  LOCK_SEARCH_MANDATORY_LOCK |
					  LOCK_SEARCH_NONZERO_LEN))
	{
		/* Special search means that if the two byte ranges overlap
		 * but the criteria is not satisfied, the comparison result
		 * is 1 instead of 0.
		 */
		specialSearch = TRUE;
	}
	else
		specialSearch = FALSE;


	comp = lock_btree_splay(rootIdxp,
							inNodep,
							inLockInfop,
							inLockOwnerType,
							useNode,
							searchMask,
							LOCK_SPECIAL_COMP_RESULT,	/*Direction towards
														  bigger nodes*/
							treeType,
							&rootNodep);

	if (comp && (rootNodep != NULL) && (specialSearch == TRUE))
	{
		/*
		 * At this point, we may miss some overlapping nodes that satisfy
		 * the criteira. These missed node's lock offsets are smaller than
		 * the input node. The following code is called to find these nodes
		 * if any.
		 */

		lock_btNode_t	*leftChildNodep;
		vs_mem_index_t	biggestSmallerChildIdx;
		lock_btNode_t	*biggestSmallerChildNodep = NULL;

		/*--------------------------
		 * Get the root's left child
		 *--------------------------
		 */
		leftChildNodep = NULL;
		if (!VS_MEM_IDX_IS_NULL(rootNodep->child[LOCK_BT_LEFT]))
		{
			leftChildNodep = (lock_btNode_t *)
								vs_mem_idxToPtr(rootNodep->child[LOCK_BT_LEFT]);
		}
		if (leftChildNodep == NULL)
		{
			/*
			 * Match not found
			 */
			return (NULL);
		}

		/*-------------------------------------------------------------
		 * Check the root's biggest child that is smaller than the root
		 * against the input node.
		 *--------------------------------------------------------------
		 */
		if (VS_MEM_IDX_IS_NULL(leftChildNodep->child[LOCK_BT_RIGHT]))
		{
			/* The root's left child has not right child.
			   Check the root's left child */
			biggestSmallerChildIdx = rootNodep->child[LOCK_BT_LEFT];
			biggestSmallerChildNodep = (lock_btNode_t *)
										vs_mem_idxToPtr(biggestSmallerChildIdx);
		}
		else
		{
			/* Check the root's left child's right most child */
			vs_mem_index_t	rightChildIdx;
			lock_btNode_t	*rightChildNodep;

			rightChildIdx = leftChildNodep->child[LOCK_BT_RIGHT];
			while (!VS_MEM_IDX_IS_NULL(rightChildIdx))
			{
				biggestSmallerChildIdx = rightChildIdx;
				rightChildNodep = (lock_btNode_t *)vs_mem_idxToPtr(rightChildIdx);
				rightChildIdx = rightChildNodep->child[LOCK_BT_RIGHT];
			}
			biggestSmallerChildNodep = rightChildNodep;
		}

		/*-----------------------------------------------------------------
		 * Compare the ranges of the biggestSmallerChild and the input node
		 *-----------------------------------------------------------------
		 */
		specialComp = lock_compFunc(inNodep,
									inLockInfop,
									inLockOwnerType,
									useNode,
									biggestSmallerChildNodep,
									LOCK_SEARCH_EXCL_BOTH_0LEN,	/*SearchMask*/
									LOCK_SPECIAL_COMP_RESULT,
									treeType);
		if (specialComp == 0)
		{
			/* The two byte ranges overlay. Do splay again. */
			comp = lock_btree_splay(rootIdxp,
									inNodep,
									inLockInfop,
									inLockOwnerType,
									useNode,
									searchMask,
									-1,		/*Direction towards smaller nodes*/
									treeType,
									&rootNodep);
		}
	} /* if(comp && ...) */

	if (comp)
	{
		/*
		 * Match not found
		 */
		return (NULL);
	}
	else
	{
		/*
		 * The found node becomes root
		 */
		return (rootNodep);
	}
} /*lock_btree_find()*/

/*-----------------------------------------------------------------
 * Name:         lock_btree_delete
 *
 * Description:
 *
 *      Delete a node from a btree.
 *
 * Arguments:    
 * 
 *     rootIdxp - Pointer to the index to the root of the tree.
 *
 *     inNodep - Pointer to the node to be deleted.
 *
 *     searchMask - The criteria based on which we search for
 *                  a matching node.
 *
 *     treeType - Exclusive or shared lock tree.
 *                Value is LOCK_TYPE_EXCL or LOCK_TYPE_SHAR.
 *
 *     delNodeIdxp - This is output. It points to the index to the
 *                   deleted node.
 *
 * Return value:
 *
 *     Pointer to the deleted node - If succeed.
 *
 *     NULL - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
lock_btNode_t *
lock_btree_delete(vs_mem_index_t	*rootIdxp,
				  lock_btNode_t		*inNodep,
				  uint16			searchMask,
				  uint8				treeType,
				  vs_mem_index_t	*delNodeIdxp)
{
	int8			comp;
	lock_btNode_t	*rootNodep;
	lock_btNode_t	*delNodep, *rchildNodep;
	vs_mem_index_t	rchildNodeIdx;

	comp = lock_btree_splay(rootIdxp,
							inNodep,
							NULL,
							LOCK_TYPE_DUMMY,
							TRUE,
							searchMask,
							LOCK_SPECIAL_COMP_RESULT,
							treeType,
							&rootNodep);
	if (comp)
	{
		/*Match not found*/
		delNodep = NULL;
	}
	else
	{
		*delNodeIdxp = *rootIdxp;
		delNodep = rootNodep;

		rchildNodeIdx = delNodep->child[LOCK_BT_RIGHT];
		rchildNodep = (lock_btNode_t *)vs_mem_idxToPtr(rchildNodeIdx);
		if (!VS_MEM_IDX_IS_NULL(delNodep->child[LOCK_BT_LEFT]))
		{
			/*delNodep has left child*/
			*rootIdxp = delNodep->child[LOCK_BT_LEFT];
			lock_btree_splay(rootIdxp,
							 NULL,
							 NULL,
							 LOCK_TYPE_DUMMY,
							 TRUE,
							 searchMask,
							 LOCK_SPECIAL_COMP_RESULT,
							 treeType,
							 &rootNodep);
			rootNodep->child[LOCK_BT_RIGHT] = rchildNodeIdx;
			if (rchildNodep)
				rchildNodep->parent = *rootIdxp;
		}
		else
		{
			*rootIdxp = rchildNodeIdx;
			rootNodep = (lock_btNode_t *)vs_mem_idxToPtr(*rootIdxp);
			if (rootNodep == NULL)
			{
				LOCK_PRINTF(LOG_DEBUG, ("lock_btree_delete: NULL pointer"));
			}
			else
				VS_MEM_IDX_SET_NULL(rootNodep->parent);
		}
	}

	return (delNodep);
} /*lock_btree_delete()*/

/*-----------------------------------------------------------------
 * Name:         lock_btree_deleteTree
 *
 * Description:
 *
 *      Delete a btree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     treeType - Exclusive or shared lock tree.
 *                Value is LOCK_TYPE_EXCL or LOCK_TYPE_SHAR.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_btree_deleteTree(lock_fileLock_t	*fileLockp,
					  uint8				treeType)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	lock_btNode_t	*rootNodep;

	switch (treeType)
	{
		case LOCK_TYPE_EXCL:
			rootIdxp = &(fileLockp->exclLockTree);
			break;

		case LOCK_TYPE_SHAR:
			rootIdxp = &(fileLockp->sharedLockTree);
			break;

		default:
			return (NFX_ERR);
	}

	while (!VS_MEM_IDX_IS_NULL(*rootIdxp))
	{
		rootNodep = (lock_btNode_t *)vs_mem_idxToPtr(*rootIdxp);
		if (rootNodep == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_btree_deleteTree: NULL pointer"));
			return (NFX_ERR);
		}

		/*Delete the delNode*/
		rc = lock_delBtNode(fileLockp,
							rootNodep,
							treeType,
							TRUE,
							TRUE);
		if (rc != NFX_OK)
			return (rc);
	}

	/*Tree is empty*/
	return (NFX_OK);
} /* lock_btree_deleteTree() */

/*-----------------------------------------------------------------
 * Name:         lock_btree_traverse_preorder
 *
 * Description:
 *
 *      Traverse a btree in preorder.
 *
 * Arguments:    
 * 
 *     rootIdx - Index to the root of the tree.
 *
 *     treeType - Type of the tree (exclusive or shared).
 *
 *     doFunc - Function to call on each node in tree.
 *
 *              The argument of the function:
 *              nodeIdx - Index to the node.
 *              level - The node is at this level in the tree.
 *              dir - The node is left or right child of its parent.
 *              count - The total number of nodes in the tree
 *                      that have been reached so far.
 *              treeType - Type of the tree (exclusive or shared).
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 10/29/2002
 *
 *-----------------------------------------------------------------
 */
void
lock_btree_traverse_preorder(
			vs_mem_index_t	rootIdx,
			uint8			treeType,
			void			(*doFunc)(vs_mem_index_t	nodeIdx,
									  int32				level,
									  int32				dir,
									  uint32			count,
									  uint8				treeType)
				   )
{
	vs_mem_index_t  currNodeIdx, parentNodeIdx;
	lock_btNode_t	*rootNodep, *currNodep, *parentNodep;
	int32			level = 1, dir = -1;
   	uint32			count = 0;

	if (VS_MEM_IDX_IS_NULL(rootIdx))
	{
		/*Empty tree*/
		return;
	}

	currNodeIdx = rootIdx;
	rootNodep = (lock_btNode_t *)vs_mem_idxToPtr(rootIdx);
	currNodep = rootNodep;
	if (currNodep == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_preorder 1: NULL pointer"));
		return;
	}

	while (1)
	{
		while (1)
		{
			/* Take the currNode as root of a sub-tree.
			 * Walk from the currNode to the sub-tree's bottom left node*/
			doFunc(currNodeIdx, level, dir, count++, treeType);
			if (!VS_MEM_IDX_IS_NULL(currNodep->child[LOCK_BT_LEFT]))
			{
				currNodeIdx = currNodep->child[LOCK_BT_LEFT];
				currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
				if (currNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_preorder 2: "
											"NULL pointer"));
					return;
				}
				level++;
				dir = LOCK_BT_LEFT;
			}
			else
				break;
		}

		if (!VS_MEM_IDX_IS_NULL(currNodep->child[LOCK_BT_RIGHT]))
		{
			/* Move to currNode's right child */
			currNodeIdx = currNodep->child[LOCK_BT_RIGHT];
			currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
			if (currNodep == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_preorder 3: "
										"NULL pointer"));
				return;
			}
			level++;
			dir = LOCK_BT_RIGHT;
			continue;
		}

lock_btree_traverse_preorder_goToParent:

		/* Exam parent node */
		parentNodeIdx = currNodep->parent;

		if (VS_MEM_IDX_IS_NULL(parentNodeIdx))
		{
			/* currNode is root. We are done.*/
			break;
		}
		else
		{
			parentNodep = (lock_btNode_t *)vs_mem_idxToPtr(parentNodeIdx);
			if (parentNodep == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_preorder 4: "
										"NULL pointer"));
				return;
			}
		}

		if (VS_MEM_IDX_IS_EQUAL(currNodeIdx, parentNodep->child[LOCK_BT_LEFT]))
		{
			if (!VS_MEM_IDX_IS_NULL(parentNodep->child[LOCK_BT_RIGHT]))
			{
				/* Move to parentNode's right child */
				currNodeIdx = parentNodep->child[LOCK_BT_RIGHT];
				currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
				if (currNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_preorder 5: "
											"NULL pointer"));
					return;
				}
				dir = LOCK_BT_RIGHT;
				continue;
			}
		}

		/* parentNode becomes currNode */
		currNodeIdx = parentNodeIdx;
		currNodep = parentNodep;
		level--;
		goto lock_btree_traverse_preorder_goToParent;
	}
} /* lock_btree_traverse_preorder() */

/*-----------------------------------------------------------------
 * Name:         lock_btree_traverse_inorder
 *
 * Description:
 *
 *      Traverse a btree in inorder.
 *
 * Arguments:    
 * 
 *     rootIdx - Index to the root of the tree.
 *
 *     treeType - Type of the tree (exclusive or shared).
 *
 *     doFunc - Function to call on each node in tree.
 *
 *              The argument of the function:
 *              nodeIdx - Index to the node.
 *              level - The node is at this level in the tree.
 *              dir - The node is left or right child of its parent.
 *              count - The total number of nodes in the tree
 *                      that have been reached so far.
 *              treeType - Type of the tree (exclusive or shared).
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/27/2002
 *
 *-----------------------------------------------------------------
 */
void
lock_btree_traverse_inorder(
			vs_mem_index_t	rootIdx,
			uint8			treeType,
			void 			(*doFunc)(vs_mem_index_t	nodeIdx,
						              int32				level,
									  int32				dir,
									  uint32			count,
									  uint8				treeType)
				   )
{
	vs_mem_index_t  currNodeIdx, parentNodeIdx;
	lock_btNode_t	*rootNodep, *currNodep, *parentNodep, *gparentNodep;
	int32			level = 1, dir = -1;
   	uint32			count = 0;

	if (VS_MEM_IDX_IS_NULL(rootIdx))
	{
		/*Empty tree*/
		return;
	}

	currNodeIdx = rootIdx;
	rootNodep = (lock_btNode_t *)vs_mem_idxToPtr(rootIdx);
	currNodep = rootNodep;
	if (currNodep == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_inorder 1: NULL pointer"));
		return;
	}

	while (1)
	{
		while (1)
		{
			/* Take the currNode as root of a sub-tree.
			 * Walk from the currNode to the sub-tree's bottom left node */
			if (!VS_MEM_IDX_IS_NULL(currNodep->child[LOCK_BT_LEFT]))
			{
				currNodeIdx = currNodep->child[LOCK_BT_LEFT];
				currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
				if (currNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_inorder 2: "
											"NULL pointer"));
					return;
				}
				level++;
				dir = LOCK_BT_LEFT;
			}
			else
			{
				/* Apply doFunc to bottom left node */
				doFunc(currNodeIdx, level, dir, count++, treeType);
				break;
			}
		}

		if (!VS_MEM_IDX_IS_NULL(currNodep->child[LOCK_BT_RIGHT]))
		{
			/* Move to currNode's right child */
			currNodeIdx = currNodep->child[LOCK_BT_RIGHT];
			currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
			if (currNodep == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_inorder 3: "
										"NULL pointer"));
				return;
			}
			level++;
			dir = LOCK_BT_RIGHT;
			continue;
		}

lock_btree_traverse_inorder_goToParent:

		/* Exam parent node */
		parentNodeIdx = currNodep->parent;

		if (VS_MEM_IDX_IS_NULL(parentNodeIdx))
		{
			/* currNode is root. We are done. */
			break;
		}
		else
		{
			parentNodep = (lock_btNode_t *)vs_mem_idxToPtr(parentNodeIdx);
			if (parentNodep == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_inorder 4: "
										"NULL pointer"));
				return;
			}
			level--;
		}

		if (VS_MEM_IDX_IS_EQUAL(currNodeIdx, parentNodep->child[LOCK_BT_LEFT]))
		{
			/* Apply doFunc to parent node */
			gparentNodep = (lock_btNode_t *)vs_mem_idxToPtr(parentNodep->parent);
			if (gparentNodep != NULL)
			{
				if (VS_MEM_IDX_IS_EQUAL(gparentNodep->child[LOCK_BT_LEFT],
									  parentNodeIdx))
					dir = LOCK_BT_LEFT;
				else
					dir = LOCK_BT_RIGHT;
			}
			else
				dir = -1;  /*root*/
			doFunc(parentNodeIdx, level, dir, count++, treeType);

			if (!VS_MEM_IDX_IS_NULL(parentNodep->child[LOCK_BT_RIGHT]))
			{
				/* Move to parentNode's right child */
				currNodeIdx = parentNodep->child[LOCK_BT_RIGHT];
				currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
				if (currNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_traverse_inorder 5: "
											"NULL pointer"));
					return;
				}
				level++;
				dir = LOCK_BT_RIGHT;
				continue;
			}
		}

		/* parentNode becomes currNode */
		currNodeIdx = parentNodeIdx;
		currNodep = parentNodep;
		goto lock_btree_traverse_inorder_goToParent;
	}
} /* lock_btree_traverse_inorder() */

/*-----------------------------------------------------------------
 * Name:         lock_btree_getNextNode
 *
 * Description:
 *
 *     Walk the tree in inorder, and return the tree nodes,
 *     one at a time.
 *
 * Arguments:    
 * 
 *     currNodeIdx - Index to the current node in the tree.
 *                   When the function is called the first time for a tree,
 *                   The value of this argument is the index to the root
 *                   of the tree.
 *
 *     firstTime - TURE if the function is called the first time for a tree.
 *                 FALSE otherwise.
 *
 *     nextNodeIdxp - This is output. Its value is index to the next
 *                    node in the tree.
 *  
 * Return value:
 *
 *     Pointer to next node or NULL if no more node in the tree.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/27/2002
 *
 *-----------------------------------------------------------------
 */
lock_btNode_t *
lock_btree_getNextNode(vs_mem_index_t	currNodeIdx,
					   boolean			firstTime,
					   vs_mem_index_t	*nextNodeIdxp)
{
	vs_mem_index_t	parentNodeIdx;
	lock_btNode_t	*currNodep, *parentNodep;
	boolean			toBotLeftNode = FALSE;

	/* Init nextNodeIdxp */
	VS_MEM_IDX_SET_NULL(*nextNodeIdxp);

	/* Validate input currNodeIdx */
	if (VS_MEM_IDX_IS_NULL(currNodeIdx))
	{
		return (NULL);
	}

	currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
	if (currNodep == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_btree_getNextNode 1: NULL pointer"));
		return (NULL);
	}

	if (firstTime == TRUE)
	{
		/* currNodeIdx is root of tree.
		 * Walk from the root to the tree's bottom left node */
		toBotLeftNode = TRUE;
	}

	while (1)
	{
		 /* Walk from the currNode to the bottom left node */
		if (toBotLeftNode == TRUE)
		{
			while (!VS_MEM_IDX_IS_NULL(currNodep->child[LOCK_BT_LEFT]))
			{
				currNodeIdx = currNodep->child[LOCK_BT_LEFT];
				currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
				if (currNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_getNextNode 2: "
											"NULL pointer"));
					return (NULL);
				}
			}
			*nextNodeIdxp = currNodeIdx;
			return (currNodep);
		}
		else
		{
			/* toBotLeftNode != TRUE */

			/* Exam currNode's right child */
			if (!VS_MEM_IDX_IS_NULL(currNodep->child[LOCK_BT_RIGHT]))
			{
				/* Move to currNode's right child */
				currNodeIdx = currNodep->child[LOCK_BT_RIGHT];
				currNodep = (lock_btNode_t *)vs_mem_idxToPtr(currNodeIdx);
				if (currNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_getNextNode 3: "
											"NULL pointer"));
					return (NULL);
				}
				toBotLeftNode = TRUE;
				continue;
			}

			/* Exam currNode's parent */
lock_btree_getNextNode_goToParent:

			parentNodeIdx = currNodep->parent;
			if (VS_MEM_IDX_IS_NULL(parentNodeIdx))
			{
				/* currNode is root. We are done. */
				return (NULL);
			}
			else
			{
				parentNodep = (lock_btNode_t *)vs_mem_idxToPtr(parentNodeIdx);
				if (parentNodep == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_btree_getNextNode 4: "
											"NULL pointer"));
					return (NULL);
				}
				if (VS_MEM_IDX_IS_EQUAL(currNodeIdx,
									  parentNodep->child[LOCK_BT_RIGHT]))
				{
					/* parentNode becomes currNode */
					currNodeIdx = parentNodeIdx;
					currNodep = parentNodep;
					goto lock_btree_getNextNode_goToParent;
				}
				else
				{
					*nextNodeIdxp = parentNodeIdx;
					return (parentNodep);
				}
			}
		}
	}

	return (NULL);
} /* lock_btree_getNextNode() */

