/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:		lock-cmd.c
 *                              
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Contain byte range lock UI command functions on TXRX only.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/21/2003
 *
 *-----------------------------------------------------------------
 */
#include "lock-incls.h"

int32 lock_cmdShow(int32 argc, char *argv[]);
int32 lock_cmdTrace(int32 argc, char *argv[]);
int32 lock_cmdHelpAll(void);
int32 lock_cmdHelpCmd (char *pCmd);
int32 lock_cmdClear(int32 argc, char *argv[]);

typedef struct
{
	int32	(*cmdFunc)(int32 argc, char *argv[]);
	char	*cmd;
	char	*helpTxt;
} lock_cmdTbl_t;


lock_cmdTbl_t lock_cmdTable[ ] =
{
	{lock_cmdShow,	"show",		"lock show stats"},
	{lock_cmdTrace,	"trace",	"lock trace debug|debug2|off"},
	{lock_cmdClear,	"clear",	"lock clear stats"}
};

#define LOCK_CMD_MAX	sizeof(lock_cmdTable)/sizeof(lock_cmdTbl_t)


/*-----------------------------------------------------------------
 * Name:		lock_cmdHelp
 *
 * Description: Display lock help-texts.
 *
 *              lock help [show | trace]
 *
 * Created by:	Xiaobing Zhang 
 *
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmdHelp(int32 argc, char *argv[])
{
	if (argc < 3)
		return (lock_cmdHelpAll());
	else
		return (lock_cmdHelpCmd(argv[2]));
} // lock_cmdHelp()

/*-----------------------------------------------------------------
 * Name:		lock_cmdHelpAll
 *
 * Description: Display help-texts for all lock commands.
 *
 * Created by:	Xiaobing Zhang 
 *
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmdHelpAll(void)
{
	int32	i;

	printf("Usage:");
	for (i=0; i < LOCK_CMD_MAX; i++)
	{
		printf("\t%s\n", lock_cmdTable[i].helpTxt);
	}
	return (0);
} // lock_cmdHelpAll()

/*-----------------------------------------------------------------
 * Name:		lock_cmdHelpCmd
 *
 * Description: Display help-texts for an individual lock command.
 *
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmdHelpCmd(char *pCmd)
{
	int32	i;

	for (i=0; i < LOCK_CMD_MAX; i++)
	{
		if (strequ(pCmd, lock_cmdTable[i].cmd))
		{
			printf("Usage:");
			printf("\t%s\n", lock_cmdTable[i].helpTxt);
			return 0;
		}
	}
	return (lock_cmdHelpAll());
} // lock_cmdHelpCmd()

/*-----------------------------------------------------------------
 * Name:		lock_cmdShow
 *
 * Description: Dispatch for lock show commands.
 *
 *              lock show stats
 * 
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmdShow(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		lock_cmdHelpCmd("show");
		return (0);
	}

	if ( !(strncmp(argv[2], "stats", 4)))
	{
		lock_showStats();
		return (NFX_OK);
	}
	return (lock_cmdHelpCmd("show"));
} // lock_cmdShow()

/*-----------------------------------------------------------------
 * Name:		lock_cmdTrace
 *
 * Description: Enables or disables lock tracing.
 *
 *              lock trace <on|off>
 * 
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmdTrace(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		lock_cmdHelpCmd("trace");
		return 0;
	}
	if (strequ(argv[2], "debug"))	
	{
		lock_traceEnable = 1;
		lock_debugLevel = LOG_DEBUG;
		printf("Lock debug trace has been turned on.\n");
	}
	if (strequ(argv[2], "debug"))	
	{
		lock_traceEnable = 1;
		lock_debugLevel = LOCK_DEBUG2;
		printf("Lock debug2 trace has been turned on.\n");
	}
	else if (strequ(argv[2], "off"))
	{
		lock_traceEnable = 0;
		printf("Lock trace has been turned off\n");
	}
	else
		lock_cmdHelpCmd("trace");

	return 0;
} // lock_cmdTrace()

/*-----------------------------------------------------------------
 * Name:		lock_cmdClear
 *
 * Description: Clear the stats and debug counters.
 *
 *              lock clear stats
 * 
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmdClear(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		lock_cmdHelpCmd("clear");
		return (0);
	}

	if ( !(strncmp(argv[2], "stats", 4)))
	{
		lock_clearStats();
		return (NFX_OK);
	}
	return (lock_cmdHelpCmd("clear"));
} // lock_cmdClear()

/*-----------------------------------------------------------------
 * Name:		lock_cmd
 *
 * Description: Main dispatch for lock commands.  Called from
 *              command table.
 *
 *				lock show stats
 *				lock trace debug|debug2|off
 *				lock show stats
 *
 * Created by : Xiaobing Zhang
 *
 * Date Created:	01/21/2003
 *-----------------------------------------------------------------
 */
int32 lock_cmd(int32 argc, char *argv[])
{
	int32	i;

	if (argc < 2)
	{
		return (lock_cmdHelp(argc, argv));
	}

	for (i=0; i < LOCK_CMD_MAX; i++)
	{
		if (strequ(argv[1], lock_cmdTable[i].cmd))
		{
			return ((*lock_cmdTable[i].cmdFunc)(argc, argv));
		}
	}
	return (lock_cmdHelp(argc, argv));
} // lock_cmd()

