/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         lock-deadlock.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Deadlock detection functions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*---------------------------------------------------------------------
 * Stack used during deadlock detection.
 * The stack stores a path in the dependency graph we are walk through.
 *---------------------------------------------------------------------
 */

/* This is the maximun number of nodes we can remember
 * along a path in the dependency graph.
 */
#define LOCK_DL_STACK_SIZE	64	/*TODO: The size can be changed*/

/* The stack. It stores pointers to structures of type lock_dependency_t. */
static lock_dependency_t *lock_dl_stack[LOCK_DL_STACK_SIZE];

/* Top of the stack */
static int16 lock_dl_top;

/*-----------------------------------------------------------------
 * Name:         lock_dl_stackInit
 *
 * Description:
 *
 *     Initialization the deadlock detection stack.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
inline void
lock_dl_stackInit(void)
{
	lock_dl_top = -1;

	return;
} /* lock_init() */

/*-----------------------------------------------------------------
 * Name:         lock_dl_stackPush
 *
 * Description:
 *
 *     Push an item to the deadlock detection stack.
 *
 * Arguments:
 *
 *     item - Index to a lock_dependencyList_t structure.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If the stack is full.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
inline int32
lock_dl_stackPush(lock_dependency_t	*item)
{
	if (lock_dl_top < LOCK_DL_STACK_SIZE)
	{
		/* The stack is not full yet */
		lock_dl_stack[++lock_dl_top] = item;
		return (NFX_OK);
	}

	return (NFX_ERR);
} /* lock_dl_stackPush() */

/*-----------------------------------------------------------------
 * Name:         lock_dl_stackPop
 *
 * Description:
 *
 *     Pop an item out of the deadlock detection stack.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     Pointer to a lock_dependencyList_t structure - If the stack is not empty.
 *
 *     NULL - If the stack is empty.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
inline lock_dependency_t *
lock_dl_stackPop(void)
{
	lock_dependency_t *item;

	if (lock_dl_top >= 0)
	{
		/* The stack is not empty */
		item = lock_dl_stack[lock_dl_top--];
		return (item);
	}

	return (NULL);
} /* lock_dl_stackPop() */

/*-----------------------------------------------------------------
 * Name:         lock_getDependencyListOfLockOwner
 *
 * Description:
 *
 *     Get the dependency list of the given lock owner.
 *
 * Arguments:
 *
 *     lockOwnerIdx - Index to the lock owner.
 *
 * Return Value:
 *
 *     Pointer to the dependency list - If there is such a list.
 *
 *     NULL - If the lock owner has not dependency list.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
lock_dependency_t *
lock_getDependencyListOfLockOwner(vs_mem_index_t lockOwnerIdx,
								  vs_mem_index_t *dependencyListIdxp)
{
	lock_dependency_t	*dependencyListp;
	uint8				lockOwnerType;
	void				*lockOwnerp;

	lockOwnerp = lock_getLockOwnerByIdx(lockOwnerIdx, &lockOwnerType);
	if (lockOwnerp == NULL)
		return (NULL);

	switch (lockOwnerType)
	{
		/*
		case LOCK_OWNER_CIFS:
			*dependencyListIdxp =
					((lock_cifsLockOwner_t *)lockOwnerp)->dependencyList;
			break;
		*/

		case LOCK_OWNER_NLM:
			*dependencyListIdxp =
					((lock_nlmLockOwner_t *)lockOwnerp)->dependencyList;
			break;

		default:
			VS_MEM_IDX_SET_NULL(*dependencyListIdxp);
			break;
	}

	dependencyListp = (lock_dependency_t *)vs_mem_idxToPtr(*dependencyListIdxp);

	return (dependencyListp);
} /* lock_getDependencyListOfLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_setDependencyListOfLockOwner
 *
 * Description:
 *
 *     Set the dependency list of the given lock owner to be the
 *     given dependency list.
 *
 * Arguments:
 *
 *     lockOwnerIdx - Index to the lock owner.
 *
 *     dependencyListIdx - Index to the dependency list.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_setDependencyListOfLockOwner(vs_mem_index_t lockOwnerIdx,
								  vs_mem_index_t dependencyListIdx)
{
	uint8				lockOwnerType;
	void				*lockOwnerp;

	lockOwnerp = lock_getLockOwnerByIdx(lockOwnerIdx, &lockOwnerType);
	if (lockOwnerp == NULL)
		return (NFX_ERR);

	switch (lockOwnerType)
	{
		/*
		case LOCK_OWNER_CIFS:
			((lock_cifsLockOwner_t *)lockOwnerp)->dependencyList = 
														dependencyListIdx;
			break;
		*/

		case LOCK_OWNER_NLM:
			((lock_nlmLockOwner_t *)lockOwnerp)->dependencyList =
				   										dependencyListIdx;
			break;

		default:
			return (NFX_ERR);
			break;
	}

	return (NFX_OK);
} /* lock_getDependencyListOfLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_detectDeadlock
 *
 * Description:
 *
 *     Deadlock detection. We use depth first search to detect a potential
 *     loop in the depenendcy graph.
 *
 * Arguments:
 *
 *     dependencyListp - A list of blocking owners, who own the locks that
 *                       block the lock requested by another owner, the
 *                       blocked owner. Any owner in this list won't be
 *                       the same as the blocked owner.
 *
 *     blockedOwnerIdx - Index to the owner who requests the lock that
 *                       is blocked.
 *
 *     The search starts from a blocking owner in the input blocking owner
 *     list. Check each owner in the dependency list of the blocking owner.
 *     If an owner in the dependency list is the same as the input blocked
 *     owner, then it can cause a deadlock if the requested lock is added
 *     to the pending list. This search continues until each owner in the
 *     dependency graph is checked or a deadlock is found.
 *
 * Return Value:
 *
 *     TRUE - There is a potential deadlock.
 *
 *     FALSE - There is no deadlock.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
boolean
lock_detectDeadlock(lock_dependency_t	*dependencyListp,
					vs_mem_index_t		blockedOwnerIdx)
{
	vs_mem_index_t		currNodeIdx;
	lock_dependency_t   *currNodep;

	if (VS_MEM_IDX_IS_NULL(blockedOwnerIdx))
		return (FALSE);

	/*---------------
	 * Init the stack
	 *---------------
	 */
	/*The stack stores a path in the dependency graph we are walk through.*/
	lock_dl_stackInit();

	currNodep = dependencyListp;
	while (1)
	{
		while (currNodep)
		{
			/* Check the current blocking owner against the blocked owner */
			if (lock_sameOwner(currNodep->blockingOwner, blockedOwnerIdx))
			{
				/* There is deadlock */
				return (TRUE);
			}

			/* Push the current node in stack */
			lock_dl_stackPush(currNodep);

			/* Get dependency list of the current blocking owner */
			currNodep =
				lock_getDependencyListOfLockOwner(currNodep->blockingOwner,
												  &currNodeIdx);	
		}

		/* Pop a node from the stack. This node has been checked. */
		currNodep = lock_dl_stackPop();
		if (currNodep == NULL)
		{
			/* No more node to check in the dependency graph */
			break;
		} /* while (currNodep) */

		/* Get next node in the dependency list */
		currNodep = (lock_dependency_t *)vs_mem_idxToPtr(currNodep->next);
	} /* while(1) */

	return (FALSE);
} /* lock_detectDeadlock() */

/*-----------------------------------------------------------------
 * Name:         lock_addOwnerToDpndListOfPendLock
 *
 * Description:
 *
 *     Update the dependency list of the given pending lock. Add a node
 *     or increment blockedLockcount in a node in the list.
 *
 * Arguments:
 *
 *     dependencyNodep - Pointer to the dependency list.
 *
 *     lockOwnerIdx - Index to a lock owner structure. This lock owner
 *                    is to be added in the dependency list.
 *
 *     dependencyListIdxp - This is output. Index to the new dependency list.
 *
 * Return Value:
 *
 *     Pointer to the head of new dependency list.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
lock_dependency_t *
lock_addOwnerToDpndListOfPendLock(lock_dependency_t	*dependencyListp,
								  vs_mem_index_t	lockOwnerIdx,
								  vs_mem_index_t	*dependencyListIdxp)
{
	vs_mem_index_t		dependencyIdx;
	lock_dependency_t	*dependencyp;

	if (dependencyListp == NULL)
	{
		/*------------------
		 * The list is empty
		 *------------------
		 */

		/* Allocate a new lock_dependency_t structure for the lock owner */
		dependencyp = lock_memAllocDependency(lockOwnerIdx,
											  &dependencyIdx);
		*dependencyListIdxp = dependencyIdx;
		return (dependencyp);
	}

	/*--------------------------------------------------
	 * Search the list to find a node for the lock owner
	 *--------------------------------------------------
	 */
	dependencyp = dependencyListp;
	while (dependencyp)
	{
		if (lock_sameOwner(dependencyp->blockingOwner, lockOwnerIdx))
		{
			/*--------------------------------
			 * Found the node.
			 * Increment the blockedLockCount.
			 *--------------------------------
			 */
			dependencyp->blockedLockCount++;
			return (dependencyListp);
		}

		/* Go to the next node in the list */
		dependencyp = (lock_dependency_t *)vs_mem_idxToPtr(dependencyp->next);
	}

	/*---------------------------------------------------------------
	 * Allocate a new lock_dependency_t structure for the lock owner,
	 * and insert it at the beginning of the list.
	 *---------------------------------------------------------------
	 */
	dependencyp = lock_memAllocDependency(lockOwnerIdx,
										  &dependencyIdx);
	if (dependencyp)
	{
		dependencyp->next = *dependencyListIdxp;
		*dependencyListIdxp = dependencyIdx;	
		dependencyListp = dependencyp;
	}

	return (dependencyListp);
} /* lock_addOwnerToDpndListOfPendLock() */

/*-----------------------------------------------------------------
 * Name:         lock_createDependencyListOfPendLock
 *
 * Description:
 *
 *     Create the dependency list of the given pending lock. The lock
 *     owners in the dependency list have locks that block the pending
 *     lock. This dependency list doesn't contain the owner of the
 *     pending lock.
 *
 * Arguments:
 *
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure. This is the pending lock.
 *
 *     lockType - The pending lock's type (LOCK_TYPE_EXCL or LOCK_TYPE_SHAR).
 *
 *     checkExclTree - TRUE if the pending lock has conflict with the locks
 *                     in the exclusive lock tree.
 *
 *     checkSharedTree - TRUE if the pending lock has conflict with the locks
 *                       in the sharedd lock tree.
 *
 *     dependencyNodeIdxp - This is output. Index to the new dependency list.
 *
 * Return Value:
 *
 *     Pointer to the dependency list - If succeed.
 *
 *     NULL - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
lock_dependency_t *
lock_createDependencyListOfPendLock(lock_fileLock_t	*fileLockp,
									void			*pendLockInfop,
									uint8			pendLockType,
									boolean			checkExclTree,
									boolean			checkSharedTree,
									vs_mem_index_t	*dependencyNodeIdxp)
{
	lock_dependency_t	*dependencyNodep = NULL;
	int32				protocol, locktype, treetype, owner, block;
	int32				result;
	uint8				ownerType, pendLockOwnerType = LOCK_OWNER_NLM;
	boolean				sameOwner;
	vs_mem_index_t		lockOwnerIdx;
	vs_mem_index_t		dummyLockInfoIdx;
	int8				comp;

	if ((fileLockp == NULL) || (pendLockInfop == NULL))
	{
		LOCK_PRINTF(LOG_ERR, ("lock_createDependencyListOfPendLock: "
								"NULL input pointer."));
		return (NULL);
	}

	VS_MEM_IDX_SET_NULL(*dependencyNodeIdxp);
	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);

	protocol = 1; /*LOCK_OWNER_NLM*/
	locktype = (pendLockType == LOCK_TYPE_EXCL) ? 0 : 1;
	block = 0;	/*yes, block*/

	if (checkExclTree == TRUE)
	{
		/*--------------------------
		 * Check exclusive lock tree
		 *--------------------------
		 */
		vs_mem_index_t		exLockIdx;
		lock_exLock_t		*exLockp;

		treetype = 0;   /*Exlcusive lock tree*/

		/*
		 * Get the first overlapping lock in the exclusive lock tree
		 */
		exLockp = lock_findExclLockByLockInfo(&(fileLockp->exclLockTree),
											  pendLockInfop,
											  pendLockOwnerType,
											  0);	/*searchMask*/

		/*
		 * Walk the tree to check other overlapping locks
		 */
		exLockIdx = fileLockp->exclLockTree;
		while (exLockp)
		{
			/* Check if the input pending lock conflicts with
			 * the current lock in the tree. */
			ownerType = lock_getOwnerTypeByLockInfoIdx(exLockp->info);	
			sameOwner = lock_haveSameOwner(dummyLockInfoIdx,
										   pendLockInfop,
										   exLockp->info,
										   pendLockOwnerType,
										   ownerType);
			if (sameOwner == TRUE)
				owner = 0;
			else
				owner = 1;

			result = lock_conflictRule[protocol]\
									  [locktype]\
									  [treetype]\
									  [owner]\
									  [block];
			if ((result == LOCK_STATE_BLOCKED) &&
				(ownerType == LOCK_OWNER_NLM))
			{
				/*
				 * Add the owner of the lock in the dependency list
				 */
				lockOwnerIdx = lock_getLockOwnerIdxByLockInfo(exLockp->info,
															  NULL,
															  ownerType);
				dependencyNodep =
					lock_addOwnerToDpndListOfPendLock(dependencyNodep,
													  lockOwnerIdx,
													  dependencyNodeIdxp);
			}

			/* Get next lock in the tree */
			exLockp = (lock_exLock_t *)lock_btree_getNextNode(exLockIdx,
															  FALSE,
															  &exLockIdx);

			if (exLockp != NULL)
			{
				/* Check if the lock overlaps with the pending lock */
				comp = lock_twoLockInfoCompFunc(dummyLockInfoIdx,
												pendLockInfop,
												pendLockOwnerType,
												exLockp->info,
												0,	/*searchMask*/
												LOCK_SPECIAL_COMP_RESULT);
				if (comp != 0)
				{
					/* Done */
					break;
				}
			}
		} /* while (exLockp) */
	} /* if (checkExclTree == TRUE) */

	if (checkSharedTree == TRUE)
	{
		/*-----------------------
		 * Check shared lock tree
		 *-----------------------
		 */
		boolean				found = FALSE, done = FALSE;
		vs_mem_index_t		shNodeIdx, shLockIdx;
		lock_shNode_t		*shNodep;
		lock_shLock_t		*shLockp;

		treetype = 1;   /*Shared lock tree*/

		/*
		 * Get the first overlapping node in the shared lock tree
		 */
		shNodep = lock_findSharedNodeByLockInfo(&(fileLockp->sharedLockTree),
												pendLockInfop,
												pendLockOwnerType,
												0);	/*searchMask*/

		/*
		 * Walk the tree to check other overlapping locks
		 */
		shNodeIdx = fileLockp->sharedLockTree;
		while (shNodep)
		{
			/*
			 * Check each lock in the linked list
			 */
			shLockIdx = shNodep->lockList;
			while (!VS_MEM_IDX_IS_NULL(shLockIdx))
			{
				shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
				if (shLockp == NULL)
				{
					LOCK_PRINTF(LOG_ERR,
								("lock_createDependencyListOfPendLock: "
								"Couldn't find shLock data structure."));
					break;
				}

				/* Check if the lock overlaps with the pending lock */
				comp = lock_twoLockInfoCompFunc(dummyLockInfoIdx,
												pendLockInfop,
												pendLockOwnerType,
												shLockp->info,
												0,	/*searchMask*/
												LOCK_SPECIAL_COMP_RESULT);
				if (comp != 0)
				{
					if (found == TRUE)
					{
						/* Done */
						done = TRUE;
						break;
					}
				}
				else
				{
	
					/* Check if the input pending lock conflicts with
					 * the lock in the list. */
					ownerType = lock_getOwnerTypeByLockInfoIdx(shLockp->info);	
					sameOwner = lock_haveSameOwner(dummyLockInfoIdx,
												   pendLockInfop,
												   shLockp->info,
												   pendLockOwnerType,
												   ownerType);
					if (sameOwner == TRUE)
						owner = 0;
					else
						owner = 1;

					result = lock_conflictRule[protocol]\
											  [locktype]\
											  [treetype]\
											  [owner]\
											  [block];
					if ((result == LOCK_STATE_BLOCKED) &&
						(ownerType == LOCK_OWNER_NLM))
					{
						/*
						 * Add the owner of the lock in the dependency list
						 */
						lockOwnerIdx =
							lock_getLockOwnerIdxByLockInfo(shLockp->info,
														   NULL,
														   ownerType);
						dependencyNodep = lock_addOwnerToDpndListOfPendLock(
														dependencyNodep,
														lockOwnerIdx,
														dependencyNodeIdxp);
					}
				}

				/* Get next shLock in the list */
				shLockIdx = shLockp->link;
			} /* while (!VS_MEM_IDX_IS_NULL(shLockIdx)) */

			if (done == TRUE)
				break;

			/* Get next exclusive lock in the tree */
			shNodep = (lock_shNode_t *)lock_btree_getNextNode(shNodeIdx,
															  FALSE,
															  &shNodeIdx);
		} /* while(shNodep) */
	} /* if (checkSharedTree == TRUE) */

	return (dependencyNodep);
} /* lock_createDependencyListOfPendLock() */

/*-----------------------------------------------------------------
 * Name:         lock_incDependencyListOfBlockedOwner
 *
 * Description:
 *
 *     Merge the dependency list of the given owner with the input
 *     dependency list.
 *
 * Arguments:
 *
 *     lockOwnerIdx - Index to a lock owner structure. This lock owner's
 *                    dependency list will be updated.
 *
 *     dpndListIdx - Index to the dependency list, each node of which
 *                   will be merged with the above lock owner's dependency
 *                   list.
 *
 *     dpndListp - Pointer to the dependency list.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_incDependencyListOfBlockedOwner(vs_mem_index_t		lockOwnerIdx,
									 vs_mem_index_t		inDpndListIdx,
									 lock_dependency_t	*inDpndListp)
{
	int32				rc;
	vs_mem_index_t		dpndListIdx, newDpndListIdx;
	lock_dependency_t	*dpndListp;
	vs_mem_index_t		inDpndNodeIdx, inNextDpndNodeIdx, dpndNodeIdx;
	lock_dependency_t	*inDpndNodep, *inNextDpndNodep, *dpndNodep;
	boolean				freeInNode;

	/*---------------------------------------------
	 * Get the dependency list of the blocked owner
	 *---------------------------------------------
	 */
	dpndListp = lock_getDependencyListOfLockOwner(lockOwnerIdx,
												  &dpndListIdx);	

	if (dpndListp == NULL)
	{
		/*---------------------------------------------------------------
		 * If the blocked owner doesn't have a dependency list,
		 * make the input dependency list to be the blocked owner's list.
		 *---------------------------------------------------------------
		 */
		rc = lock_setDependencyListOfLockOwner(lockOwnerIdx, inDpndListIdx);
		return (rc);
	}

	/*----------------------------------------------------------
	 * Init the head of new dependency list of the blocked owner
	 *----------------------------------------------------------
	 */
	newDpndListIdx = dpndListIdx;

	/*----------------------------------------------------
	 * Exam every node in the input dependency list
	 * to merge it to the blocked owner's dependency list.
	 *----------------------------------------------------
	 */
	inDpndNodeIdx = inDpndListIdx;
	inDpndNodep = inDpndListp;
	while (inDpndNodep)
	{
		freeInNode = FALSE;

		/*--------------------------------------------------------
		 * Check current node in the input dependency list against
		 * every node in the blocked owner's dependency list
		 *--------------------------------------------------------
		 */
		dpndNodeIdx = dpndListIdx;
		dpndNodep = dpndListp;
		while (dpndNodep)
		{
			if (lock_sameOwner(dpndNodep->blockingOwner,
							   inDpndNodep->blockingOwner))
			{
				/*-------------------------------------------
				 * Increment the blockedLockCount of dpndNode
				 *-------------------------------------------
				 */
				dpndNodep->blockedLockCount += inDpndNodep->blockedLockCount;
				freeInNode = TRUE;
				break;
			}

			/* Go to next node in the blocked owner's dependency list*/
			dpndNodeIdx = dpndNodep->next;
			dpndNodep = (lock_dependency_t *)vs_mem_idxToPtr(dpndNodeIdx);
		} /* while (dpndNodep) */

		/* Get next node in the input dependency list */
		inNextDpndNodeIdx = inDpndNodep->next;
		inNextDpndNodep = (lock_dependency_t *)vs_mem_idxToPtr(inNextDpndNodeIdx);

		/*-----------------------------------------------------
		 * Handle the current node in the input dependency list
		 *-----------------------------------------------------
		 */
		if (freeInNode == FALSE)
		{
			/* Insert the node in front of the list */
			inDpndNodep->next = newDpndListIdx;
			newDpndListIdx = inDpndNodeIdx;
		}
		else
		{
			rc = lock_memFreeDependency(inDpndNodeIdx);
			if (rc != NFX_OK)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_incDependencyListOfBlockedOwner: "
						"Failed to delete the node in the dependency list."));
			}
		}

		inDpndNodeIdx = inNextDpndNodeIdx;
		inDpndNodep = inNextDpndNodep;
	} /* while (inDpndNodep) */

	rc = lock_setDependencyListOfLockOwner(lockOwnerIdx, newDpndListIdx);

	return (rc);
} /* lock_incDependencyListOfBlockedOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_decDpndListOfOwnerByIdx
 *
 * Description:
 *
 *     Update the dependency list of the given lock owner. Delete a node
 *     or decrement blockedLockCount in a node in the list.
 *
 * Arguments:
 *
 *     blockedOwnerIdx - Index to a lock owner structure. Its dependency
 *                       list will be updated.
 *
 *     blockingOwnerIdx - Index to a lock owner structure. This lock owner
 *                        is to be deleted from the blocked owner's 
 *                        dependency list.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/08/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_decDpndListOfOwnerByIdx(vs_mem_index_t	blockedOwnerIdx,
							 vs_mem_index_t	blockingOwnerIdx)

{
	vs_mem_index_t		dependencyListIdx, dependencyIdx;
	lock_dependency_t	*dependencyListp, *dependencyp, *prevDependencyp;
	boolean				found;

	dependencyListp = lock_getDependencyListOfLockOwner(blockedOwnerIdx,
													    &dependencyListIdx);

	/*-----------------------------------------------------------
	 * Search the list to find a node for the blocking lock owner
	 *-----------------------------------------------------------
	 */
	found = FALSE;
	dependencyIdx = dependencyListIdx;
	dependencyp = dependencyListp;
	prevDependencyp = NULL;
	while (dependencyp)
	{
		if (lock_sameOwner(dependencyp->blockingOwner, blockingOwnerIdx))
		{
			/*--------------------------------
			 * Found the node.
			 * Decrement the blockedLockCount.
			 *--------------------------------
			 */
			found = TRUE;

			dependencyp->blockedLockCount--;
			if (dependencyp->blockedLockCount == 0)
			{
				/*----------------
				 * Delete the node
				 *----------------
				 */
				if (prevDependencyp == NULL)
				{
					/* Delete the first node in the list */
					lock_setDependencyListOfLockOwner(blockedOwnerIdx,
													  dependencyp->next);
				}
				else
				{
					prevDependencyp->next = dependencyp->next;
				}
				lock_memFreeDependency(dependencyIdx);
			}
			break;
		}

		/* Go to the next node in the list */
		prevDependencyp = dependencyp;
		dependencyIdx = dependencyp->next;
		dependencyp = (lock_dependency_t *)vs_mem_idxToPtr(dependencyIdx);
	} /* while () */

	if (found == TRUE)
		return (NFX_ERR);
	else
		return (NFX_OK);
} /* lock_decDpndListOfOwnerByIdx() */

/*-----------------------------------------------------------------
 * Name:         lock_decDependencyListOfBlockedOwners
 *
 * Description:
 *
 *     Update the dependency lists of the lock owners who own pending
 *     locks. These pending locks are blocked by the just deleted granted
 *     lock. This function is called after a granted lock is deleted.
 *
 * Arguments:
 *
 *     fileLockp - Index to a lock_fileLock_t structure. The dependency
 *                 lists of the owners who own some pending locks on this
 *                 file will be updated.
 *
 *     delLockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    data structure, which contains information of lock 
 *                    just deleted.
 *
 *     delLockType - The deleted lock's type (LOCK_TYPE_EXCL or LOCK_TYPE_SHAR).
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/08/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_decDependencyListOfBlockedOwners(lock_fileLock_t	*fileLockp,
									  void				*delLockInfop,
									  uint8				delLockType)
{
	int32				rc, rcFinal = NFX_OK;
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp;
	vs_mem_index_t		pendLockOwnerIdx, delLockOwnerIdx;
	uint8				pendLockOwnerType, delLockOwnerType = LOCK_OWNER_NLM;
	int32               protocol, locktype, treetype, owner, block;
	int32				result;
	vs_mem_index_t		dummyLockInfoIdx;
	int8				comp;
	boolean				sameOwner;

	if (fileLockp == NULL)
		return (NFX_ERR);

	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);

	treetype = (delLockType == LOCK_TYPE_EXCL) ? 0 : 1;
	block = 0;  /*yes, block*/

	/*------------------------------------------------------
	 * Go through the pending lock list to check
	 * which pending locks are blocked by this deleted lock.
	 *------------------------------------------------------
	 */
	pendLockIdx = fileLockp->pendingLockList;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		/* Get pending lock structure */
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		pendLockOwnerType = lock_getOwnerTypeByLockInfoIdx(pendLockp->info);
		if (pendLockOwnerType != LOCK_OWNER_NLM)
		{
			/* Go to next lock in the pending list */
			pendLockIdx = pendLockp->link;
			continue;
		}

		/*---------------------------------------------------------
		 * Check if the pending lock overlaps with the deleted lock
		 *---------------------------------------------------------
		 */
		comp = lock_twoLockInfoCompFunc(dummyLockInfoIdx,
										delLockInfop,
										delLockOwnerType,
										pendLockp->info,
										0, /*searchMask*/
										LOCK_SPECIAL_COMP_RESULT);
		if (comp == 0)
		{
			/* The two locks overlap */

			/*----------------------------------------------------------
			 * Check if the pending lock conflicts with the deleted lock
			 *----------------------------------------------------------
			 */
			sameOwner = lock_haveSameOwner(dummyLockInfoIdx,
										   delLockInfop,
										   pendLockp->info,
										   delLockOwnerType,
										   pendLockOwnerType);
			if (sameOwner == TRUE)
				owner = 0;
			else
				owner = 1;

			protocol = 1; /*LOCK_OWNER_NLM*/
			locktype = (pendLockp->type == LOCK_TYPE_EXCL) ? 0 : 1;

			result = lock_conflictRule[protocol]\
									  [locktype]\
									  [treetype]\
									  [owner]\
									  [block];
			if (result == LOCK_STATE_BLOCKED)
			{
				/* The two locks conflict */
				pendLockOwnerIdx = lock_getLockOwnerIdxByLockInfo(
														pendLockp->info,
														NULL,
														pendLockOwnerType);
				delLockOwnerIdx = lock_getLockOwnerIdxByLockInfo(
														dummyLockInfoIdx,
														delLockInfop,
														delLockOwnerType);

				/*-------------------------------------------------------
				 * Update the dependency list of the pending lock's owner
				 *-------------------------------------------------------
				 */
				rc = lock_decDpndListOfOwnerByIdx(pendLockOwnerIdx,
												  delLockOwnerIdx);
				if (rc != NFX_OK)
					rcFinal = rc;
			}
		}

		/* Go to next lock in the pending list */
		pendLockIdx = pendLockp->link;
	} /* while () */

	return (rcFinal);
} /* lock_decDependencyListOfBlockedOwners() */

/*-----------------------------------------------------------------
 * Name:         lock_delDpndListByOwnerId
 *
 * Description:
 *
 *     Update the dependency list of the given lock owner. Remove the
 *     nodes that represent the lock owners who match the given ID (fid,
 *     pid, or clientIpAddr).
 *
 * Arguments:
 *
 *     blockedOwnerIdx - Index to a lock owner structure. Its dependency
 *                       list will be updated.
 *
 *     clientIpAddr - Identify NLM lock owners.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/08/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_delDpndListByOwnerId(vs_mem_index_t	blockedOwnerIdx,
						  ip_index_t		clientIpAddr)

{
	vs_mem_index_t		dependencyListIdx, dependencyIdx, nextDependencyIdx;
	lock_dependency_t	*dependencyListp, *dependencyp, *prevDependencyp;
	boolean				matchResult;
	vs_mem_index_t		dummyLockInfoIdx;

	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);

	dependencyListp = lock_getDependencyListOfLockOwner(blockedOwnerIdx,
													    &dependencyListIdx);

	/*-----------------------------------------------------------
	 * Search the list to find nodes for the matching lock owners
	 *-----------------------------------------------------------
	 */
	dependencyIdx = dependencyListIdx;
	dependencyp = dependencyListp;
	prevDependencyp = NULL;
	while (dependencyp)
	{
		matchResult = lock_match(dummyLockInfoIdx,
								 dependencyp->blockingOwner,
								 FALSE, /*Use lockOwner*/
								 0,
								 clientIpAddr,
								 LOCK_SEARCH_SAME_IPADDR);

		if (matchResult == TRUE)
		{
			/*---------------------------
			 * Found the node. Delete it.
			 *---------------------------
			 */
			if (prevDependencyp == NULL)
			{
				/* Delete the first node in the list */
				lock_setDependencyListOfLockOwner(blockedOwnerIdx,
												  dependencyp->next);
			}
			else
			{
				prevDependencyp->next = dependencyp->next;
			}

			/* Go to the next node in the list */
			nextDependencyIdx = dependencyp->next;

			/* Free the memory */
			lock_memFreeDependency(dependencyIdx);
		}
		else
		{
			/* Go to the next node in the list */
			prevDependencyp = dependencyp;
			nextDependencyIdx = dependencyp->next;
		}

		dependencyIdx = nextDependencyIdx;
		if (VS_MEM_IDX_IS_NULL(dependencyIdx))
			dependencyp = NULL;
		else
			dependencyp = (lock_dependency_t *)vs_mem_idxToPtr(dependencyIdx);
	}

	return (NFX_OK);
} /* lock_delDpndListByOwnerId() */

/*-----------------------------------------------------------------
 * Name:         lock_delOwnersFmDependencyListOfBlockedOwner
 *
 * Description:
 *
 *     Update the dependency lists of the lock owners who own pending
 *     locks. This function is called after some granted locks, whose
 *     owners match input key or client IP address, are deleted. These
 *     matching owners will be deleted from the dependency lists.
 *
 * Arguments:
 *
 *     fileLockp - Index to a lock_fileLock_t structure. The dependency
 *                 lists of the owners who own some pending locks on this
 *                 file will be updated.
 *
 *     clientIpAddr - Identify NLM lock owners.
 *
 * Return Value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/08/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_delOwnersFmDependencyListOfBlockedOwners(lock_fileLock_t	*fileLockp,
											  ip_index_t		clientIpAddr)
{
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp;
	vs_mem_index_t		pendLockOwnerIdx;
	int8				pendLockOwnerType;

	if (fileLockp == NULL)
		return (NFX_ERR);

	/*---------------------------------
	 * Go through the pending lock list
	 *---------------------------------
	 */
	pendLockIdx = fileLockp->pendingLockList;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		/* Get the pending lock structure */
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		/* Get owner index of the pending lock */
		pendLockOwnerIdx = lock_getLockOwnerIdxByLockInfo(pendLockp->info,
														  NULL,
														  pendLockOwnerType);
		if (pendLockOwnerType == LOCK_OWNER_NLM)
			lock_delDpndListByOwnerId(pendLockOwnerIdx, clientIpAddr);

		/* Go to next lock in the pending list */
		pendLockIdx = pendLockp->link;
	} /* while (!VS_MEM_IDX_IS_NULL(pendLockIdx)) */

	return (NFX_OK);
} /* lock_delOwnersFmDependencyListOfBlockedOwners() */

