/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         lock-debug.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Byte range lock debugging and trace utility functions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

#define LOCK_LIST_SHARED			0
#define LOCK_LIST_PENDING_LOCAL		1
#define LOCK_LIST_PENDING_GLOBAL	2

#define LOCK_PRINT_BTREE_HEADER \
    { \
        printf("total level child lock\n"); \
        printf("----------------------\n"); \
    }

#define LOCK_PRINT_BTREE_INFO(_count_, _level_, _dirstr_) \
    {printf("%-6d%-6d%s", _count_, _level_, _dirstr_);}

/*-----------------------------------------------------------------
 * Name:         lock_showCifsLock
 *
 * Description:
 *
 *     Display a CIFS lock.
 *
 * Arguments:
 *
 *     cifsLockInfop - Pointer to cifsLockInfo structure.
 *
 *     lockTypeStr - "exclusive", "shared", or "unsupported".
 *
 *     indentation - A number of space characters used for displaying.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/13/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showCifsLock(lock_cifsLockInfo_t	*cifsLockInfop,
				  char					*lockTypeStr,
				  char					*indentation)
{
	uint8					lockOwnerType;
	lock_cifsLockOwner_t	*cifsLockOwnerp;

	cifsLockOwnerp = lock_getLockOwnerByIdx(cifsLockInfop->owner,
											&lockOwnerType);
	if (cifsLockOwnerp == NULL)
		return;

	printf("CIFS lock (%s) --\n", lockTypeStr);
	printf("%stid, fid, pid   : %d, %d, %d\n",
						indentation, cifsLockOwnerp->tid,
						cifsLockOwnerp->fid, cifsLockOwnerp->pid);
	printf("%soffset, len     : %lld, %lld\n",
						indentation, cifsLockInfop->offset, cifsLockInfop->len);

} /* lock_showCifsLock() */

/*-----------------------------------------------------------------
 * Name:         lock_showNlmLock
 *
 * Description:
 *
 *     Display a NLM lock.
 *
 * Arguments:
 *
 *     cifsLockInfop - Pointer to nlmLockInfo structure.
 *
 *     lockTypeStr - "exclusive", "shared", or "unsupported".
 *
 *     indentation - A number of space characters used for displaying.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/13/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showNlmLock(lock_nlmLockInfo_t	*nlmLockInfop,
				 char				*lockTypeStr,
				 char				*indentation)
{
	uint8					lockOwnerType;
	lock_nlmLockOwner_t		*nlmLockOwnerp;
	lock_nlmLockClient_t	*nlmLockClientp;
	char					name[1024];

	nlmLockOwnerp = lock_getLockOwnerByIdx(nlmLockInfop->owner,
										   &lockOwnerType);
	if (nlmLockOwnerp == NULL)
		return;

	nlmLockClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(nlmLockOwnerp->lockClient);
	if (nlmLockClientp == NULL)
		return;

	printf("NLM lock (%s) --\n", lockTypeStr);
	printf("%sclient IP addr, caller state : 0x%x, %d\n",
						indentation,
					   	nlmLockClientp->clientIpAddr,
					    nlmLockClientp->state);

	memcpy(name, nlmLockClientp->callerName, nlmLockClientp->callerNameLen);
	name[nlmLockClientp->callerNameLen] = 0;
	printf("%scaller name     : %s\n", indentation, name);

	memcpy(name, nlmLockOwnerp->ownerName, nlmLockOwnerp->ownerNameLen);
	name[nlmLockOwnerp->ownerNameLen] = 0;
	printf("%sowner pid, name : %d, %s\n", indentation, nlmLockOwnerp->pid, name);

	printf("%soffset, len     : %lld, %lld\n",
						indentation, nlmLockInfop->offset, nlmLockInfop->len);

} /* lock_showNlmLock() */

/*-----------------------------------------------------------------
 * Name:         lock_showLock
 *
 * Description:
 *
 *     Display a lock.
 *
 * Arguments:
 *
 *     lockInfoIdx - Index to lockInfo structure.
 *
 *     lockType - Type of the lock (exclusive or shared).
 *
 *     indentation - A number of space characters used for displaying.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/13/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showLock(vs_mem_index_t lockInfoIdx, uint8 lockType, char *indentation)
{
	char	lockTypeStr[15];
	uint8	lockOwnerType;
	void	*lockInfop;

	lockInfop = vs_mem_idxToPtr(lockInfoIdx);
	if (lockInfop == NULL)
		return;

	lockOwnerType = lock_getOwnerTypeByLockInfoIdx(lockInfoIdx);

	switch (lockType)
	{
		case LOCK_TYPE_EXCL:
			strcpy(lockTypeStr, "exclusive");
			break;

		case LOCK_TYPE_SHAR:
			strcpy(lockTypeStr, "shared");
			break;

		default:
			strcpy(lockTypeStr, "unsupported");
			break;
	}

	switch (lockOwnerType)
	{
		case LOCK_OWNER_CIFS:
			lock_showCifsLock((lock_cifsLockInfo_t *)lockInfop,
							  lockTypeStr,
							  indentation);
			break;

		case LOCK_OWNER_NLM:
			lock_showNlmLock((lock_nlmLockInfo_t *)lockInfop,
							 lockTypeStr,
							 indentation);
			break;

		default:
			break;
	}

} /* lock_showLock() */

/*-----------------------------------------------------------------
 * Name:         lock_showLockList
 *
 * Description:
 *
 *     Display a list of locks.
 *
 * Arguments:
 *
 *     listHeadIdx - Index to the head of the lock list.
 *
 *     listType - Shared lock list, pending lock list, or
 *                global pending timeout list.
 *
 *     indentation - A number of space characters used for displaying.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showLockList(vs_mem_index_t	listHeadIdx,
				  uint8				listType,
				  char				*indentation)
{
	vs_mem_index_t	lockIdx;
	void			*lockp;
	vs_mem_index_t	lockInfoIdx;
	uint8			lockType;
	char			indent[25];
	char			subindent[25];
	uint32			count = 0;

	if (VS_MEM_IDX_IS_NULL(listHeadIdx))
	{
		printf("Empty list!\n");
		return;
	}

	strcpy(indent, indentation);
	strcat(indent, "  ");

	strcpy(subindent, indent);
	strcat(subindent, "  ");

	lockIdx = listHeadIdx;
	while (!VS_MEM_IDX_IS_NULL(lockIdx))
	{
		lockp = vs_mem_idxToPtr(lockIdx);
		if (lockp == NULL)
			return;

		/* Get lockInfoIdx */
		if (listType != LOCK_LIST_SHARED)
		{
			/* The lock is pending lock */
			lockInfoIdx = ((lock_pendingLock_t *)lockp)->info;
			lockType = ((lock_pendingLock_t *)lockp)->type;
		}
		else
		{
			/* The lock is shared lock */
			lockInfoIdx = ((lock_shLock_t *)lockp)->info;
			lockType = LOCK_TYPE_SHAR;
		}

		/* Display lock */
		printf("%s", indent);
		lock_showLock(lockInfoIdx, lockType, subindent);

		if (listType != LOCK_LIST_SHARED)
		{
			/*The lock is a pending lock*/
			eeeRTC_t	rtcTimeout = ((lock_pendingLock_t *)lockp)->timeout;

			if (eee_rtcCompare(&rtcTimeout, &lock_timeZero) == 0)
				printf("%sTimeout: %s\n", subindent, "0");
			else if (eee_rtcCompare(&rtcTimeout, &lock_timeIndefinite) == 0)
				printf("%sTimeout: %s\n", subindent, "indefinite");
			else
				printf("%sTimeout: %s\n",
						subindent,
						eee_rtcStr(&(((lock_pendingLock_t *)lockp)->timeout))
				  );
		}
		printf("\n");

		count++;

		/* Go to next lock in the list */
		if (listType == LOCK_LIST_SHARED)
			lockIdx = ((lock_shLock_t *)lockp)->link;
		else if (listType == LOCK_LIST_PENDING_LOCAL)
			lockIdx = ((lock_pendingLock_t *)lockp)->link;
		else if (listType == LOCK_LIST_PENDING_GLOBAL)
			lockIdx = ((lock_pendingLock_t *)lockp)->timeoutLink;
		else
			break;
	}

	printf("%sTotal: %d\n", indentation, count);
} /* lock_showLockList() */

/*-----------------------------------------------------------------
 * Name:         lock_showSharedNode
 *
 * Description:
 *
 *     Display a shared lock node and the shared locks associated with it.
 *
 * Arguments:
 *
 *     sharedNodep - Pointer to the shNode struecture.
 *
 *     indentation - A number of space characters used for displaying.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/13/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showSharedNode(lock_shNode_t *sharedNodep, char *indentation)
{
	printf("Num of mandatory locks: %d\n",
					sharedNodep->lockCntAndHoley.mandatoryLockCnt);
	printf("%sNum of advisory locks: %d\n",
					indentation,
					sharedNodep->lockCntAndHoley.advisoryLockCnt);
	printf("%sHoley: %s\n",
					indentation,
				   	(sharedNodep->lockCntAndHoley.holey == 1) ?
				   	"TRUE": "FALSE");
	printf("%sEnd: %lld\n", indentation, sharedNodep->end);
	printf("%sLock list:\n\n", indentation);
	lock_showLockList(sharedNodep->lockList, LOCK_LIST_SHARED, indentation);

} /* lock_showSharedNode() */

/*-----------------------------------------------------------------
 * Name:         lock_showBtreeNode
 *
 * Description:
 *
 *     Display a node in an exclusive or shared lock tree.
 *
 * Arguments:
 *
 *     nodeIdx - Index to the root of the tree.
 *
 *     level - The node is at this level in the tree.
 *
 *     dir - The node is left or right child of its parent.
 *
 *     count - The total number of nodes in the tree
 *             that have been reached so far.
 *
 *     treeType - Type of the tree (exclusive or shared).
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/12/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showBtreeNode(vs_mem_index_t	nodeIdx,
				   int32			level,
				   int32			dir,
				   uint32			count,
				   uint8			treeType)
{
	lock_btNode_t		*nodep;
	char				dirstr[7] = "      ";
	char				indentation[21];

	if (dir == LOCK_BT_LEFT)
		strcpy(dirstr, "  L   ");
	if (dir == LOCK_BT_RIGHT)
		strcpy(dirstr, "  R   ");
	if (dir == -1)
		strcpy(dirstr, "root  ");

	LOCK_PRINT_BTREE_INFO(count, level, dirstr);

	/* Get tree node */
	nodep = (lock_btNode_t *)vs_mem_idxToPtr(nodeIdx);
	if (nodep == NULL)
		return;

	/* Display tree node */
	switch (treeType)
	{
		case LOCK_TYPE_EXCL:
			strcpy(indentation, "                    ");
			lock_showLock(((lock_exLock_t *)nodep)->info,
						  treeType,
						  indentation);
			break;

		case LOCK_TYPE_SHAR:
			strcpy(indentation, "                  ");
			lock_showSharedNode((lock_shNode_t *)nodep, indentation);
			break;

		default:
			break;
	}

	printf("\n");
} /* lock_showBtreeNode() */

/*-----------------------------------------------------------------
 * Name:         lock_showBtree
 *
 * Description:
 *
 *     Display an exclusive or shared lock tree.
 *
 * Arguments:
 *
 *     rootIdx - Index to the root of the tree.
 *
 *     treeType - Exclusive or shared lock tree.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/12/2003
 *
 *-----------------------------------------------------------------
 */
inline void
lock_showBtree(vs_mem_index_t rootIdx, uint8 treeType)
{
	if (VS_MEM_IDX_IS_NULL(rootIdx))
	{
		printf("Empty tree!\n\n");
		return;
	}

	LOCK_PRINT_BTREE_HEADER;
	lock_btree_traverse_preorder(rootIdx, treeType, lock_showBtreeNode);

} /* lock_showBtree() */

/*-----------------------------------------------------------------
 * Name:         lock_showFileLock
 *
 * Description:
 *
 *     Display the granted and pending locks on a given file.
 *
 * Arguments:
 *
 *     fileLockIdx - Index to a fileLock structure.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/12/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showFileLock(vs_mem_index_t fileLockIdx)
{
	lock_fileLock_t	*fileLockp;
	char			indentation[22] = "                     ";

	/* Get fileLock structure */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_showFileLock: NULL pointer to fileLock. "
								"Failed to show the locks on the file."));
		return;
	}

	printf("\n");
	printf("===================\n");
	printf("= Locks on a file =\n");
	printf("===================\n");
	
	printf("Lowest mandatory lock offset: %u (0x%x)\n",
					fileLockp->lowestMandatoryLockOffset,
					fileLockp->lowestMandatoryLockOffset);

	printf("Last granted lock: ");
	lock_showLock(lock_getLockInfoIdxByLockIdx(fileLockp->lastGrantedLock),
				  lock_getLockTypeByLockIdx(fileLockp->lastGrantedLock),
				  indentation);

	printf("Num of NLM locks: %d\n",
					fileLockp->nlmLockCount);
	printf("\n");

	printf("--------------------------------\n");
	printf("Exclusive lock tree of the file:\n");
	printf("--------------------------------\n\n");
	lock_showBtree(fileLockp->exclLockTree, LOCK_TYPE_EXCL);

	printf("-----------------------------\n");
	printf("Shared lock tree of the file:\n");
	printf("-----------------------------\n\n");
	lock_showBtree(fileLockp->sharedLockTree, LOCK_TYPE_SHAR);

	printf("------------------------------\n");
	printf("Pending lock list of the file:\n");
	printf("------------------------------\n\n");
	lock_showLockList(fileLockp->pendingLockList, LOCK_LIST_PENDING_LOCAL, "");
	printf("\n\n");

} /* lock_showFileLock() */

/*-----------------------------------------------------------------
 * Name:         lock_showLockCounters
 *
 * Description:
 *
 *     Display the current granted and pending locks in the filer.
 *
 * Arguments:
 *
 *     lockCntp - Pointer to the lock counters structure.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/11/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showLockCounters(lock_lockCounters_t *lockCntp)
{
	printf("Num of locks on the filer:               %d\n",
					lock_globalCfgp->lockCount);
	printf("Num of granted CIFS exclusive locks:     %d\n",
					lockCntp->grantedExclCifsLocks);
	printf("Num of granted NLM exclusive locks:      %d\n",
					lockCntp->grantedExclNlmLocks);
	printf("Num of granted CIFS shared locks:        %d\n",
					lockCntp->grantedSharedCifsLocks);
	printf("Num of granted NLM shared locks:         %d\n",
					lockCntp->grantedSharedNlmLocks);
	printf("Num of pending CIFS exclusive locks:     %d\n",
					lockCntp->pendingExclCifsLocks);
	printf("Num of pending NLM exclusive locks:      %d\n",
					lockCntp->pendingExclNlmLocks);
	printf("Num of pending CIFS shared locks:        %d\n",
					lockCntp->pendingSharedCifsLocks);
	printf("Num of pending NLM shared locks:         %d\n",
					lockCntp->pendingSharedNlmLocks);

} /* lock_showLockCounters() */

/*-----------------------------------------------------------------
 * Name:         lock_showDebugCounters
 *
 * Description:
 *
 *     Display lock debug counters.
 *
 * Arguments:
 *
 *     dbgCntp - Pointer to the debug counters structure.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/11/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showDebugCounters(lock_debugCounters_t *dbgCntp)
{
	printf("eee memory allocations/deallocations:\n\n");
	printf("Num of eeeAlloc calls:                   %lld\n",
					dbgCntp->eeeAlloc);
	printf("Num of eeeDealloc calls:                 %lld\n",
					dbgCntp->eeeDealloc);

	printf("\n\n");

	printf("Data structures allocations/deallocations:\n\n");
	printf("Num of CIFS owners allocated:            %lld\n",
					dbgCntp->allocCifsLockOwner);
	printf("Num of CIFS owners freed:                %lld\n",
					dbgCntp->freeCifsLockOwner);
	printf("Num of NLM RecOnVolCBs allocated:        %lld\n",
					dbgCntp->allocNlmRecOnVolCB);
	printf("Num of NLM RecOnVolCBs freed:            %lld\n",
					dbgCntp->freeNlmRecOnVolCB);
	printf("Num of NLM lock clients allocated:       %lld\n",
					dbgCntp->allocNlmLockClient);
	printf("Num of NLM lock clients freed:           %lld\n",
					dbgCntp->freeNlmLockClient);
	printf("Num of NLM lock owners allocated:        %lld\n",
					dbgCntp->allocNlmLockOwner);
	printf("Num of NLM lock owners freed:            %lld\n",
					dbgCntp->freeNlmLockOwner);
	printf("Num of dependency structs allocated:     %lld\n",
					dbgCntp->allocDependency);
	printf("Num of dependency structs freed:         %lld\n",
					dbgCntp->freeDependency);
	printf("Num of ownerTimeout structs allocated:   %lld\n",
					dbgCntp->allocOwnerTimeout);
	printf("Num of ownerTimeout structs freeed:      %lld\n",
					dbgCntp->freeOwnerTimeout);
	printf("Num of CIFS lockInfo allocated:          %lld\n",
					dbgCntp->allocCifsLockInfo);
	printf("Num of CIFS lockInfo freed:              %lld\n",
					dbgCntp->freeCifsLockInfo);
	printf("Num of NLM lockInfo allocated:           %lld\n",
					dbgCntp->allocNlmLockInfo);
	printf("Num of NLM lockInfo freed:               %lld\n",
					dbgCntp->freeNlmLockInfo);
	printf("Num of exclusive locks allocated:        %lld\n",
					dbgCntp->allocExLock);
	printf("Num of exclusive locks freed:            %lld\n",
					dbgCntp->freeExLock);
	printf("Num of shared locks allocated:           %lld\n",
					dbgCntp->allocShLock);
	printf("Num of shared locks freed:               %lld\n",
					dbgCntp->freeShLock);
	printf("Num of shared nodes allocated:           %lld\n",
					dbgCntp->allocShNode);
	printf("Num of shared nodes freed:               %lld\n",
					dbgCntp->freeShNode);
	printf("Num of pending locks allocated:          %lld\n",
					dbgCntp->allocPendingLock);
	printf("Num of pending locks freed:              %lld\n",
					dbgCntp->freePendingLock);
	printf("Num of fileLock structs allocated:       %lld\n",
					dbgCntp->allocFileLock);
	printf("Num of fileLock structs freed:           %lld\n",
					dbgCntp->freeFileLock);

	printf("\n\n");

	printf("Lock APIs:\n\n");
	printf("Num of initFileLockReq:                  %lld\n",
					dbgCntp->initFileLockReq);
	printf("Num of uninitFileLockReq:                %lld\n",
					dbgCntp->uninitFileLockReq);
	printf("Num of byteRangeLockReq:                 %lld\n",
					dbgCntp->byteRangeLockReq);
	printf("Num of cifsExclByteRangeLockReq:         %lld\n",
					dbgCntp->cifsExclByteRangeLockReq);
	printf("Num of cifsSharedByteRangeLockReq:       %lld\n",
					dbgCntp->cifsSharedByteRangeLockReq);
	printf("Num of nlmExclByteRangeLockReq:          %lld\n",
					dbgCntp->nlmExclByteRangeLockReq);
	printf("Num of nlmSharedByteRangeLockReq:        %lld\n",
					dbgCntp->nlmSharedByteRangeLockReq);
	printf("Num of unlockSingleReq:                  %lld\n",
					dbgCntp->unlockSingleReq);
	printf("Num of cifsUnlockSingleReq:              %lld\n",
					dbgCntp->cifsUnlockSingleReq);
	printf("Num of nlmUnlockSingleReq:               %lld\n",
					dbgCntp->nlmUnlockSingleReq);
	printf("Num of unlockAllByFidReq:                %lld\n",
					dbgCntp->unlockAllByFidReq);
	printf("Num of unlockAllByKeyReq:                %lld\n",
					dbgCntp->unlockAllByKeyReq);
	printf("Num of unlockAllByIpAddrReq:             %lld\n",
					dbgCntp->unlockAllByIpAddrReq);
	printf("Num of cancelLockReq:                    %lld\n",
					dbgCntp->cifsCancelLockReq);
	printf("Num of cifsCancelLockReq:                %lld\n",
					dbgCntp->cifsCancelLockReq);
	printf("Num of nlmCancelLockReq:                 %lld\n",
					dbgCntp->nlmCancelLockReq);
	printf("Num of testLockReq:                      %lld\n",
					dbgCntp->testLockReq);
	printf("Num of checkLockForWriteReq:             %lld\n",
					dbgCntp->checkLockForWriteReq);
	printf("Num of checkLockForReadReq:              %lld\n",
					dbgCntp->checkLockForReadReq);
	printf("Num of areThereCurrentFileLocksReq:      %lld\n",
					dbgCntp->areThereCurrentFileLocksReq);

	printf("\n\n");
	printf("Others:\n\n");
	printf("Num of CIFS lock owners:                 %lld\n",
					dbgCntp->cifsLockOwners);
	printf("Num of NLM lock clients:                 %lld\n",
					dbgCntp->nlmLockClients);
	printf("Num of NLM lock owners:                  %lld\n",
					dbgCntp->nlmLockOwners);
	printf("Num of deadlock conditions detected:     %lld\n",
					dbgCntp->detectedDeadlocks);
	printf("Num of \"holey\" nodes in shared trees:    %lld\n",
					dbgCntp->holeyNodes);

} /* lock_showDebugCounters() */

/*-----------------------------------------------------------------
 * Name:         lock_showErrCounters
 *
 * Description:
 *
 *     Display lock error counters.
 *
 * Arguments:
 *
 *     errCntp - Pointer to the error counters structure.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/11/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showErrCounters(lock_errCounters_t *errCntp)
{
	printf("Insufficient Memory:                     %lld\n",
					errCntp->insufficientMemory);

} /* lock_showErrCounters() */

/*-----------------------------------------------------------------
 * Name:         lock_showStats
 *
 * Description:
 *
 *     Display lock statistics.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_showStats(void)
{
	printf("\n");
	printf("-------------------\n");
	printf("Current Lock Counts\n");
	printf("-------------------\n");
	lock_showLockCounters(&lock_globalCfgp->lockCounters);

	printf("\n");
	printf("-------------------\n");
	printf("Lock Debug Counters\n");
	printf("-------------------\n");
	lock_showDebugCounters(&lock_globalCfgp->debugCounters);

	printf("\n");
	printf("-------------------\n");
	printf("Lock Error Counters\n");
	printf("-------------------\n");
	lock_showErrCounters(&lock_globalCfgp->errCounters);
	printf("\n");

} /* lock_showStats() */

/*-----------------------------------------------------------------
 * Name:         lock_clearStats
 *
 * Description:
 *
 *     Clear lock statistics.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_clearStats(void)
{
	bzero(&lock_globalCfgp->debugCounters, sizeof(lock_debugCounters_t));

	bzero(&lock_globalCfgp->errCounters, sizeof(lock_errCounters_t));

} /* lock_clearStats() */


