/* vi:ts=4
 *
 *-----------------------------------------------------------------
 *
 * Name:         lock-internal.h
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Byte range lock definitions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/11/2002
 *
 *-----------------------------------------------------------------
 */
#ifndef _LOCK_INTERNAL_H
#define _LOCK_INTERNAL_H

#include <nfx-incl.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#include <exports/vs-hash-api.h>
#include <exports/vs-mem-api.h>
#include <exports/nfs-api.h>
#include <exports/lock-api.h>

#define LOCK_MAX_LOCKS  4096    /*TODO: Maximum number of locks on a filer*/

#define LOCK_VSERVER_IDX_ANY 0  /*FIXME: TBD*/

#if 0
/*FIXME: Remove the followings when their real definitions are available*/
typedef uint32  ip_index_t; /*This data type is for both IPv4 and IPv6 address*/
/*FIXME: Remove the above when their real definitions are available*/
#endif


/*-----------
 * Lock owner
 *-----------
 */
typedef struct lock_ownerHeader {
    /*Type of the lock owner*/
    unsigned int ownerType : 2;

    /*Number of locks in either granted or pending state for the lock owner*/
    unsigned int lockCount : 29;
#define LOCK_MAX_COUNT  0x1FFFFFFF

	/* InList flag. It is 1 if the owner is in the owner timeout list */
	unsigned int inList    : 1;

    /* A lock owner structure immediately follows*/
} lock_ownerHeader_t;

/* The lock owner types imply lock types:
 *     An NLM lock is an advisory lock. 
 *     A CIFS lock is a mandatory lock. 
 *     An NLM lock is a persistent lock.
 */
#define LOCK_IS_ADVISORY(x)  (x == LOCK_OWNER_NLM)
#define LOCK_IS_MANDATORY(x) (x != LOCK_OWNER_NLM)
#define LOCK_IS_PERSIST(x)   (x == LOCK_OWNER_NLM)

/* CIFS lock owner */
typedef struct lock_cifsLockOwner {
    /*The header information*/
    lock_ownerHeader_t  header;

    /*The CIFS tree connection ID.  It identifies a client connection
	 * to a CIFS share.*/
    uint16              tid;

    /*The file ID generated by the CIFS server*/
    uint16              fid;

    /*The process ID supplied by the client*/
    uint16              pid;
} lock_cifsLockOwner_t;

/* The NLM lock owner */
/* NLM lock client and lock owner information will be stored in stable 
 * storage. The client and owner information will be stored in one or more
 * lock client record and lock owner record files respectively.
 */
typedef struct lock_nlmRecOnVolCB {
    evm_volId_t     volId;

   /*Index of the lock client/owner info
    *in the client/owner record file on this volume */
    uint32          idxInRecFile;

    union {
        /*For lock client, it is number of lock owners
         *that belong to the client*/
        uint32      ownerCount;

       /*For lock owner, it is number of locks in either granted or
        *pending state on a volume owned by the lock owner*/
        uint32      lockCount;
    } u;

    /*Next lock_nlmRecOnVolCB in the list*/
    vs_mem_index_t  next;           /*Type: lock_nlmRecOnVolCB_t*/
} lock_nlmRecOnVolCB_t;

typedef struct lock_nlmLockClient {
	struct {
        /* Number of owners that belong to the client */
        unsigned int ownerCount : 31;

	    /* InList flag. It is 1 if the client is in the owner timeout list */
	    unsigned int inList     : 1;
	} header;

    /*Client's Status Monitor state number*/
    int32       state;

    /*This lock client record is saved in .nlm_locks_client files
     *on these volumes*/
    vs_mem_index_t  volumeList;     /*Type: lock_nlmRecOnVolCB_t*/

    /*Client's IP address*/
    ip_index_t      clientIpAddr;

    /*The number of bytes in the client machine's name*/
    uint16          callerNameLen;

    /*Name of the client machine*/
    char            *callerName;
} lock_nlmLockClient_t;

typedef struct lock_nlmLockOwner {
    /*The header information*/
    lock_ownerHeader_t  header;

    /*The process ID supplied by the client*/
    int32               pid;

    /*This lock owner record is saved in .nlm_locks_owner files on
     *these volumes*/
    vs_mem_index_t      volumeList;      /*Type: lock_nlmRecOnVolCB_t*/

    /*Lock client information.  This is process-independent*/
    vs_mem_index_t      lockClient;      /*Type: lock_nlmLockClient_t*/

    /*This lock owner is blocked by the lock owners in the dependency list*/
    vs_mem_index_t      dependencyList;  /*Type: lock_dependency_t*/

    /*The number of bytes in the lock owner's name*/
    uint16              ownerNameLen;

    /*Name of the lock owner*/
    char                *ownerName;
} lock_nlmLockOwner_t;

typedef struct lock_dependency {
    vs_mem_index_t      next;            /*Type: lock_dependency_t*/
    vs_mem_index_t      blockingOwner;   /*Type: lock_cifsLockOwner or
                                                 lock_nlmLockOwner*/
    uint32              blockedLockCount; /*Number of locks being blocked by
                                            the locks owned by this owner*/
} lock_dependency_t;

/* Singly-linked list of CIFS owners, NLM clients, NLM owners. Every node
 * in the list represents a owner/client to be removed from the hash table
 * and deleted after the timeout.
 */
typedef struct lock_ownerTimeout {
    /*Next node in the list*/
    vs_mem_index_t      next;

    /*The time when this owner/client shoule be deleted*/
    eeeRTC_t            timeout;

    vs_mem_index_t      owner_client;    /*Type: lock_cifsLockOwner_t,
                                                 lock_nlmLockClient_t,
                                                 lock_nlmLockOwner_t*/
} lock_ownerTimeout_t;

/* CIFS lock info */
typedef struct lock_cifsLockInfo {
    vs_mem_index_t   owner;      /*Type: lock_ownerHeader_t
                                   The lock's owner information*/
    uint64           offset;     /*The starting offset of the byte range*/
    uint64           len;        /*The length of the byte range*/
} lock_cifsLockInfo_t;

/* NLM lock info */
/* NLM lock information will be stored in a file on stable storage. */
typedef struct lock_nlmLockInfo {
    vs_mem_index_t   owner;      /*Type: lock_ownerHeader_t
                                   The lock's owner information*/
    uint64           offset;     /*The starting offset of the byte range*/
    uint64           len;        /*The length of the byte range*/
    uint32           idxInRecFile;  /*Index of this lock info in the lock 
                                      record file .nlm_locks*/
    vs_mem_index_t   fcb;        /*Type: fcb_t*/
} lock_nlmLockInfo_t;

#define LOCK_GET_LOCK_RANGE_LENGTH(_offset_, _len_, _end_)                  \
    {                                                                       \
        if (_offset_ == _end_)                                              \
            _len_ = 0;                                                      \
        else                                                                \
            _len_ = _end_ - _offset_ + 1;                                   \
    }

#define LOCK_GET_LOCK_RANGE_END(_offset_, _len_, _end_)                     \
    {                                                                       \
        if (_len_ > 0)                                                      \
            _end_ = _offset_ + _len_ - 1;                                   \
        else                                                                \
            _end_ = _offset_;                                               \
    }

/*-----------------------
 * Granted exclusive lock
 *-----------------------
 */
/* Each of the lock_exLock_t structures is a node in the
 * granted exclusive lock tree (binary tree).
 */
typedef struct lock_exLock {
    /*binary tree links to parent, left and right children*/
    vs_mem_index_t   parent;     /*Type: lock_exLock_t*/
    vs_mem_index_t   child[2];   /*Type: lock_exLock_t*/

    /*The lock range information*/ 
    vs_mem_index_t   info;       /*Type: lock_cifsLockInfo_t or
                                         lock_nlmLockInfo_t*/
} lock_exLock_t;

/*----------------------------------
 * Granted shared lock and tree node
 *----------------------------------
 */
/* Granted shared lock */
typedef struct lock_shLock {
    /*A singly-linked list of shared lock*/
    vs_mem_index_t   link;       /*Type: lock_shLock_t*/

    /*The lock range information*/
    vs_mem_index_t   info;       /*Type: lock_cifsLockInfo_t or
                                         lock_nlmLockInfo_t*/
} lock_shLock_t;

/* Granted shared lock tree node */
/* A granted shared lock binary tree node contains a linked list of overlapping
 * shared locks. The characteristic of an entry in the linked list is that
 * the lock overlaps with at least one lock in the list. The linked list is
 * sorted by the starting offset of the lock.
 */
typedef struct lock_shNode {
    /*binary tree links to parent, left and right children*/
    vs_mem_index_t   parent;       /*Type: lock_shNode_t*/
    vs_mem_index_t   child[2];     /*Type: lock_shNode_t*/

    /*Head of a singly-linked list of locks under this node*/
    vs_mem_index_t   lockList;     /*Type: lock_shLock_t*/

    /*Last lock in the above list*/
    vs_mem_index_t   lockListTail; /*Type: lock_shLock_t*/

    struct {
        /*Number of mandatory locks in the list*/
        uint32       mandatoryLockCnt;

        /*Number of advisory locks in the list*/
        unsigned int advisoryLockCnt : 31;

        /*Holey flag. Flag whether this node is holey as
          a result of a failed allocation during a node split.*/
        unsigned int holey : 1;
    } lockCntAndHoley;

    /*Maximum byte offset covered by the locks of this node. The minimum
      offset is the starting offset of the first lock at this node.*/
    uint64           end;
} lock_shNode_t;

/* Increment mandatory or advisory lock count in a shared node */
#define LOCK_SHNODE_INC_COUNT(_nodep_, _ownerType_)							\
	{																		\
		if (LOCK_IS_MANDATORY(_ownerType_))									\
			(_nodep_)->lockCntAndHoley.mandatoryLockCnt++;					\
		else																\
			(_nodep_)->lockCntAndHoley.advisoryLockCnt++;					\
	}

/* Decrement mandatory or advisory lock count in a shared node */
#define LOCK_SHNODE_DEC_COUNT(_nodep_, _ownerType_)							\
	{																		\
		if (LOCK_IS_MANDATORY(_ownerType_))									\
			(_nodep_)->lockCntAndHoley.mandatoryLockCnt--;					\
		else																\
			(_nodep_)->lockCntAndHoley.advisoryLockCnt--;					\
	}

/* Increment manadatory or advisory lock count given the lockInfoIdx */
#define LOCK_INC_MAND_ADVI_COUNT(_lockInfoIdx_, _mandCnt_, _adviCnt_)		\
	{																		\
		if (LOCK_IS_MANDATORY(lock_getOwnerTypeByLockInfoIdx(_lockInfoIdx_)))\
			_mandCnt_++;													\
		else																\
			_adviCnt_++;													\
	}


/*-------------
 * Pending lock
 *-------------
 */
typedef struct lock_pendingLock {
    /*A singly-linked list of pending locks for a file. Sorted by request time*/
    vs_mem_index_t   link;          /*Type: lock_pendingLock_t*/

    /*The fileLock structure which the the pending lock belongs to*/
    vs_mem_index_t   fileLock;      /*Type: lock_fileLock_t*/

    /*The lock request structure passed from upper (protocol) layer.
      Pointer to context passed by the protocol layer. Will be used
      during callback. For NLM, it points to a nlm_request_t structure.*/
    vs_mem_index_t   ulPrivate;

    /*The lock range information*/
    vs_mem_index_t   info;          /*Type: lock_cifsLockInfo_t or
                                      lock_nlmLockInfo_t*/

    /*The time when this request times out in real time clock form.
      A value of infinite means indefinite, and indicates that
	  the structure is not in the global timeout lock list.*/
    eeeRTC_t         timeout;

    /*A singly-linked list of pending locks with non-zero timeouts.
      This is a global list. The locks are sorted by ascending timeout.*/
    vs_mem_index_t   timeoutLink;   /*Type: lock_pendingLock_t*/

    /*The lock type. 0 - exclusive, 1 - shared */
    uint8         type;
#define LOCK_TYPE_EXCLUSIVE 0
#define LOCK_TYPE_SHARED    1
#define LOCK_TYPE_DUMMY     2

} lock_pendingLock_t;

/*-----------
 * File locks
 *-----------
 */
/* This is the per-file lock information. Any file with byte range locks
 * held will have one of these structures.
 */
typedef struct lock_fileLock {
    /*Lowest existing mandatory lock starting offset*/
    uint32            lowestMandatoryLockOffset;
#define LOCK_MAX_LOWEST_MAND_OFFSET 0xFFFFFFFF

    /*Point to the root of the exclusive lock tree*/
    vs_mem_index_t    exclLockTree;     /*Type: lock_exLock_t*/

    /*Point to the root of the shared lock tree*/
    vs_mem_index_t    sharedLockTree;   /*Type: lock_shNode_t*/

    /*Point to first lock in the pending lock list*/
    vs_mem_index_t    pendingLockList;  /*Type: lock_pendingLock_t*/

    vs_mem_index_t    lastGrantedLock;  /*Type: lock_exLock_t or
                                                lock_shLock_t*/
    /* Number of NLM locks on this file */
    uint32            nlmLockCount;
} lock_fileLock_t;

/*---------
 * counters
 *---------
 */
typedef struct lock_lockCounters {
	uint32	grantedExclCifsLocks;      /*Current granted CIFS exclusive locks*/
	uint32	grantedExclNlmLocks;       /*Current granted NLM exclusive locks*/
	uint32	grantedSharedCifsLocks;    /*Current granted CIFS shared locks*/
	uint32	grantedSharedNlmLocks;     /*Current granted NLM shared locks*/
	uint32	pendingExclCifsLocks;      /*Current pending CIFS exclusive locks*/
	uint32	pendingExclNlmLocks;       /*Current pending NLM exclusive locks*/
	uint32	pendingSharedCifsLocks;    /*Current pending CIFS shared locks*/
	uint32	pendingSharedNlmLocks;     /*Current pending NLM shared locks*/
} lock_lockCounters_t;

/*Macros to increment or decrement the above lock counters*/
#define LOCK_INC_LOCK_COUNTERS(_ownerType_, _cifsCount_, _nlmCount_) \
    {                                                                \
        if (_ownerType_ == LOCK_OWNER_CIFS)                          \
            (_cifsCount_)++;                                         \
        else if (_ownerType_ == LOCK_OWNER_NLM)                      \
            (_nlmCount_)++;                                          \
    }

#define LOCK_DEC_LOCK_COUNTERS(_ownerType_, _cifsCount_, _nlmCount_) \
    {                                                                \
        if (_ownerType_ == LOCK_OWNER_CIFS)                          \
            (_cifsCount_)--;                                         \
        else if (_ownerType_ == LOCK_OWNER_NLM)                      \
            (_nlmCount_)--;                                          \
    }

typedef struct lock_debugCounters {
	uint64	eeeAlloc;
	uint64	eeeDealloc;

	uint64	allocCifsLockOwner;
	uint64	freeCifsLockOwner;
	uint64	allocNlmRecOnVolCB;
	uint64	freeNlmRecOnVolCB;
	uint64	allocNlmLockClient;
	uint64	freeNlmLockClient;
	uint64	allocNlmLockOwner;
	uint64	freeNlmLockOwner;
	uint64	allocDependency;
	uint64	freeDependency;
	uint64	allocOwnerTimeout;
	uint64	freeOwnerTimeout;
	uint64	allocCifsLockInfo;
	uint64	freeCifsLockInfo;
	uint64	allocNlmLockInfo;
	uint64	freeNlmLockInfo;
	uint64	allocExLock;
	uint64	freeExLock;
	uint64	allocShLock;
	uint64	freeShLock;
	uint64	allocShNode;
	uint64	freeShNode;
	uint64	allocPendingLock;
	uint64	freePendingLock;
	uint64	allocFileLock;
	uint64	freeFileLock;

	uint64	initFileLockReq;
	uint64	uninitFileLockReq;
	uint64	byteRangeLockReq;
	uint64	cifsExclByteRangeLockReq;
	uint64	cifsSharedByteRangeLockReq;
	uint64	nlmExclByteRangeLockReq;
	uint64	nlmSharedByteRangeLockReq;
	uint64	unlockSingleReq;
	uint64	cifsUnlockSingleReq;
	uint64	nlmUnlockSingleReq;
	uint64	unlockAllByFidReq;
	uint64	unlockAllByKeyReq;
	uint64	unlockAllByIpAddrReq;
	uint64	cancelLockReq;
	uint64	cifsCancelLockReq;
	uint64	nlmCancelLockReq;
	uint64	testLockReq;
	uint64	checkLockForWriteReq;
	uint64	checkLockForReadReq;
	uint64	areThereCurrentFileLocksReq;

	uint64	cifsLockOwners;		/* Number of CIFS lock owners */
	uint64	nlmLockClients;		/* Number of NLM lock clients */
	uint64	nlmLockOwners;		/* Number of NLM lock owners */
	uint64	detectedDeadlocks;	/* Number of deadlock conditions detected */
	uint64	holeyNodes;			/*Number of "holey" nodes in shared trees */
} lock_debugCounters_t;

typedef struct lock_errCounters {
	uint64	insufficientMemory;
	/*TODO: Add more error counters */
} lock_errCounters_t;

/*--------------
 * Global config
 *--------------
 */
typedef struct lock_globalCfg
{
    int32                   eee_lockAppId;

    uint32                  lockCount;      /*Number of locks on the filer*/
    lock_lockCounters_t     lockCounters;

    address_t               cifsOwnerHash;  /*Hash table of CIFS lock owners*/
    address_t               nlmClientHash;  /*Hash table of NLM lock clients*/
    address_t               nlmOwnerHash;   /*Hash table of NLM lock owners*/
#define LOCK_HASH_CIFS_OWNER_BUCKETS  256   /*TODO*/
#define LOCK_HASH_NLM_CLIENT_BUCKETS  256   /*TODO*/
#define LOCK_HASH_NLM_OWNER_BUCKETS   256   /*TODO*/

    vs_mem_index_t          pendingLockList;/*Pending lock timeout list*/
    eeeRTC_t                timeout;        /*The time when the first pending
                                              lock in the above list times out
                                              in real time clock form*/
    int32                   pendingLockTimeoutHandle;
    vs_mem_index_t          ownerTimeoutList; /*Owner timeout list*/
    int32                   ownerTimeoutHandle;
    lock_debugCounters_t    debugCounters;
    lock_errCounters_t      errCounters;
} lock_globalCfg_t;

/*-----------------
 * Global variables
 *-----------------
 */
extern lock_globalCfg_t     *lock_globalCfgp;
extern uint32               lock_traceEnable;
extern uint32               lock_debugLevel;
extern eeeRTC_t             lock_timeZero;
extern eeeRTC_t             lock_timeIndefinite;
extern eeeRTC_t             lock_timeTenMinutes;
extern eeeRTC_t             lock_timeFiveMinutes;

/*-------
 * Macros
 *-------
 */

/*TODO
#define LOCK_PRINTF(_lvl_, _par_)											\
	E_LOG(class_1, _lvl_, lock_globalCfgp->eee_lockAppId, 0, 0, 0, _par_);
*/
#define LOCK_DEBUG2 (LOG_DEBUG+1)
#define LOCK_PRINTF(_lvl_, _par_)											\
	if (lock_traceEnable && (_lvl_ <= lock_debugLevel))						\
		{printf _par_; printf("\n");}

#define LOCK_SHMEM_ALLOC(_p_, _vserver_, _type_, _size_)					\
	{																		\
		_p_ = (_type_ *) eee_ramAlloc(_size_);								\
		if (_p_ == NULL)													\
		{																	\
			LOCK_PRINTF(LOG_ERR, ("%s: eee_ramAlloc() failed", __FUNCTION__)) \
		}																	\
		else																\
		{																	\
			lock_globalCfgp->debugCounters.eeeAlloc++;						\
		}																	\
	}

#define LOCK_SHMEM_DEALLOC(_p_)												\
	{																		\
		if (_p_ != NULL)													\
		{																	\
			eee_ramDealloc(_p_);											\
			_p_ = NULL;														\
			lock_globalCfgp->debugCounters.eeeDealloc++;					\
		}																	\
	}

#define LOCK_PANIC  {*((char *)0) = 1;}

#define LOCK_GENERATE_KEY(_fid_, _pid_)  (((_pid_) << 16) | (_fid_))
#define LOCK_GET_FID_FROM_KEY(_key_)     ((_key_) & 0x00FF)

/*--------------------
 * Function Prototypes
 *--------------------
 */
void
lock_init(void);

int32
lock_checkPendingLockTimer(void);

int32
lock_pendingLockTimeoutHandler(int32    timer_handle,
                               void     *param1,
                               void     *param2,
                               void     *param3,
                               void     *param4);

int32
lock_insertOwnerClientToTimeoutList(vs_mem_index_t  ownerClientIdx);

int32
lock_ownerTimeoutHandler(int32    timer_handle,
                         void     *param1,
                         void     *param2,
                         void     *param3,
                         void     *param4);


#endif  /*_LOCK_INTERNAL_H*/

