/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        lock-utils.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Byte range lock utility functions.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:    11/25/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

extern void
nlm_procGrantedCallback(vs_mem_index_t private);

/*-----------------------------------------------------------------
 * Name:         lock_getCifsLockOwner
 *
 * Description:
 *
 *     Get the CIFS lock owner based on TID, FID and PID.
 *     If there is one in the CIFS lock owner hash table, return
 *     the pointer and index to lock_cifsLockOwner_t structure.
 *     Otherwise, create a new structure, and insert it in the hash table.
 *
 * Arguments:
 * 
 *     cifsOwnerInfop - Contains CIFS tid, fid and pid.
 *
 *     cifsLockOwnerIdxp - This is output. Index to the found or
 *                         created lock_cifsLockOwner_t structure.
 *
 * Return value:
 *
 *     Pointer to a lock_cifsLockOwner_t structure - If succeed.
 *
 *     NULL - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */ 
lock_cifsLockOwner_t *
lock_getCifsLockOwner(lock_requestCifsOwnerInfo_t	*cifsOwnerInfop,
					  vs_mem_index_t				*cifsLockOwnerIdxp)
{
	int32					rc;
	lock_cifsLockOwner_t	*cifsLockOwnerp, *foundCifsLockOwnerp;
	vs_mem_index_t			foundCifsLockOwnerIdx;

	if (cifsOwnerInfop == NULL)
		return (NULL);

	LOCK_PRINTF(LOG_DEBUG, ("lock_getCifsLockOwner: "
						"ownerInfop->tid: %d,
					    ownerInfop->fid: %d,
					   	ownerInfop->pid: %d",
						cifsOwnerInfop->tid,
						cifsOwnerInfop->fid,
					   	cifsOwnerInfop->pid));	

	/*--------------------------------------------
	 * Create a new CIFS lock owner data structure
	 *--------------------------------------------
	 */
	cifsLockOwnerp = lock_memAllocCifsLockOwner(cifsOwnerInfop->tid,
												cifsOwnerInfop->fid,
												cifsOwnerInfop->pid,
												cifsLockOwnerIdxp);
	if (cifsLockOwnerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_cifsGetLockOwner: "
						"Failed to allocate cifsLockOwner data structure."));
		return (NULL);
	}

	/*-------------------------------------------------
	 * Search for the CIFS lock owner in the hash table 
	 *-------------------------------------------------
	 */
	foundCifsLockOwnerp = lock_htLookupCifsLockOwner(cifsOwnerInfop->tid,
													 cifsOwnerInfop->fid,
													 cifsOwnerInfop->pid,
													 *cifsLockOwnerIdxp,
													 &foundCifsLockOwnerIdx);
	if (foundCifsLockOwnerp != NULL)
	{
		/* The CIFS lock owner already exists */
		LOCK_PRINTF(LOG_DEBUG, ("lock_getCifsLockOwner: "
							"Found the CIFS lock owner in the hash table."));

		/* Free the created new CIFS lock owner data structrue */
		lock_memFreeCifsLockOwner(*cifsLockOwnerIdxp);

		*cifsLockOwnerIdxp = foundCifsLockOwnerIdx;
		return (foundCifsLockOwnerp);
	}

	/*------------------------------------------------------------
	 * Insert the new CIFS lock owner data structure to hash table
	 *------------------------------------------------------------
	 */
	rc = lock_htInsertCifsLockOwner(cifsOwnerInfop->tid,
									cifsOwnerInfop->fid,
									cifsOwnerInfop->pid,
									*cifsLockOwnerIdxp);

	/*-----------------------------------------------------
	 * Put the lock owner structure in a list to be deleted
	 * a few minutes later.
	 *-----------------------------------------------------
	 */
	/* At this moment, the cifsLockOwnerp->lockCount is 0. In case no locks
	 * requests come from this owner in a few minutes, we need to delete this
	 * structure.
	 */
	rc = lock_insertOwnerClientToTimeoutList(*cifsLockOwnerIdxp);
	cifsLockOwnerp->header.inList = 1;

	return (cifsLockOwnerp);
} /* lock_getCifsLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_getNlmLockClient
 *
 * Description:
 *
 *     Get the NLM lock client based on client IP address.
 *     If there is one in the NLM lock client hash table, return
 *     the pointer and index to lock_nlmLockClient_t structure.
 *     Otherwise, create a new structure, and insert it in the hash table.
 *
 * Arguments:
 * 
 *     nlmOwnerInfop - Contains NLM lock owner information.
 *
 *     nlmLockClientIdxp - This is output. Index to the found or
 *                         created lock_nlmLockClient_t structure.
 *
 * Return value:
 *
 *     Pointer to a lock_nlmLockClient_t structure - If succeed.
 *
 *     NULL - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */ 
lock_nlmLockClient_t *
lock_getNlmLockClient(lock_requestNlmOwnerInfo_t	*nlmOwnerInfop,
					  vs_mem_index_t				*nlmLockClientIdxp)
{
	int32					rc;
	lock_nlmLockClient_t	*nlmLockClientp, *foundNlmLockClientp;
	vs_mem_index_t			foundNlmLockClientIdx;

	if (nlmOwnerInfop == NULL)
		return (NULL);

	/*--------------------------------------------
	 * Create a new NLM lock client data structure
	 *--------------------------------------------
	 */
	nlmLockClientp = lock_memAllocNlmLockClient(nlmOwnerInfop->clientIpAddr,
												nlmOwnerInfop->clientState,
												nlmOwnerInfop->pid,
												nlmOwnerInfop->callerNameLen,
												nlmOwnerInfop->callerName,
												nlmLockClientIdxp);
	if (nlmLockClientp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_nlmGetLockClient: "
						"Failed to allocate nlmLockClient data structure."));
		return (NULL);
	}

	/*-------------------------------------------------
	 * Search for the NLM lock client in the hash table 
	 *-------------------------------------------------
	 */
	foundNlmLockClientp = lock_htLookupNlmLockClient(
												nlmOwnerInfop->clientIpAddr,
												*nlmLockClientIdxp,
												&foundNlmLockClientIdx);
	if (foundNlmLockClientp != NULL)
	{
		/* The NLM lock client already exists */
		LOCK_PRINTF(LOG_DEBUG, ("lock_getNlmLockClient: "
						"Found the NLM lock client in the hash table."));

		/* Free the created new NLM lock client data structrue */
		lock_memFreeNlmLockClient(*nlmLockClientIdxp, nlmLockClientp);

		*nlmLockClientIdxp = foundNlmLockClientIdx;
		return (foundNlmLockClientp);
	}

	/*------------------------------------------------------------
	 * Insert the new NLM lock client data structure to hash table
	 *------------------------------------------------------------
	 */
	rc = lock_htInsertNlmLockClient(nlmOwnerInfop->clientIpAddr,
									*nlmLockClientIdxp);

	return (nlmLockClientp);
} /* lock_getNlmLockClient() */

/*-----------------------------------------------------------------
 * Name:         lock_getNlmLockOwner
 *
 * Description:
 *
 *     Get the NLM lock owner based on client IP address and PID.
 *     If there is one in the NLM lock owner hash table, return
 *     the pointer and index to lock_nlmLockOwner_t structure.
 *     Otherwise, create a new structure, and insert it in the hash table.
 *
 * Arguments:
 * 
 *     nlmOwnerInfop - Contains NLM lock owner information.
 *
 *     fcb - Index to the file control block. The lock owner
 *           owns lock(s) on this file.
 *
 *     nlmLockOwnerIdxp - This is output. Index to the found or
 *                        created lock_nlmLockOwner_t structure.
 *
 * Return value:
 *
 *     Pointer to a lock_nlmLockOwner_t structure - If succeed.
 *
 *     NULL - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */ 
lock_nlmLockOwner_t *
lock_getNlmLockOwner(lock_requestNlmOwnerInfo_t	*nlmOwnerInfop,
					 vs_mem_index_t				fcb,
					 vs_mem_index_t				*nlmLockOwnerIdxp)
{
	int32					rc;
	lock_nlmLockOwner_t		*nlmLockOwnerp, *foundNlmLockOwnerp;
	vs_mem_index_t			foundNlmLockOwnerIdx;
	lock_nlmLockClient_t	*nlmLockClientp;
	vs_mem_index_t			nlmLockClientIdx;

	if (nlmOwnerInfop == NULL)
		return (NULL);

	LOCK_PRINTF(LOG_DEBUG, ("lock_getNlmLockOwner: "
					"nlmOwnerInfop->clientIpAddr: 0x%x, nlmOwnerInfop->pid: %d",
					nlmOwnerInfop->clientIpAddr, nlmOwnerInfop->pid));	

	/*---------------------------------------
	 * Get the NLM lock client data structure
	 *---------------------------------------
	 */
	nlmLockClientp = lock_getNlmLockClient(nlmOwnerInfop,
									   	   &nlmLockClientIdx);
	if (nlmLockClientp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_getNlmLockOwner: "
								"Couldn't get NLM lock client."));
		return (NULL);
	}

	/*-------------------------------------------
	 * Create a new NLM lock owner data structure
	 *-------------------------------------------
	 */
	nlmLockOwnerp = lock_memAllocNlmLockOwner(nlmLockClientIdx,
											  nlmLockClientp,
											  nlmOwnerInfop->pid,
											  nlmOwnerInfop->ownerNameLen,
											  nlmOwnerInfop->ownerName,
											  fcb,
											  nlmLockOwnerIdxp);
	if (nlmLockOwnerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_nlmGetLockOwner: "
						"Failed to allocate nlmLockOwner data structure."));
		return (NULL);
	}

	/*------------------------------------------------
	 * Search for the NLM lock owner in the hash table 
	 *------------------------------------------------
	 */
	foundNlmLockOwnerp = lock_htLookupNlmLockOwner(nlmOwnerInfop->clientIpAddr,
												   nlmOwnerInfop->pid,
												   *nlmLockOwnerIdxp,
												   &foundNlmLockOwnerIdx);
	if (foundNlmLockOwnerp != NULL)
	{
		/* The NLM lock owner already exists */
		LOCK_PRINTF(LOG_DEBUG, ("lock_getNlmLockOwner: "
							"Found the NLM lock owner in the hash table."));

		/* Free the created new NLM lock owner data structrue */
		lock_memFreeNlmLockOwner(*nlmLockOwnerIdxp, nlmLockOwnerp);

		*nlmLockOwnerIdxp = foundNlmLockOwnerIdx;
		return (foundNlmLockOwnerp);
	}

	/*-----------------------------------------------------------
	 * Insert the new NLM lock owner data structure to hash table
	 *-----------------------------------------------------------
	 */
	 rc = lock_htInsertNlmLockOwner(nlmOwnerInfop->clientIpAddr,
									nlmOwnerInfop->pid,
									*nlmLockOwnerIdxp);

	/*-----------------------------------------------------
	 * Put the lock owner structure in a list to be deleted
	 * a few minutes later.
	 *-----------------------------------------------------
	 */
	/* At this moment, the nlmLockOwnerp->lockCount is 0. In case no locks
	 * requests come from this owner in a few minutes, we need to delete this
	 * structure.
	 */
	rc = lock_insertOwnerClientToTimeoutList(*nlmLockOwnerIdxp);
	nlmLockOwnerp->header.inList = 1;

	return (nlmLockOwnerp);
} /* lock_getNlmLockOwner() */

/*-----------------------------------------------------------------
 * Name:         lock_getLockOwnerTypeByIdx
 *
 * Description:
 *
 *     Get the CIFS or NLM lock owner type given its index.
 *
 * Arguments:
 * 
 *     lockOwnerIdx - Index to the lock_cifsLockOwner_t or
 *                    lock_nlmLockOwner_t structure.
 *
 * Return value:
 *
 *     The lock owner type.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */ 
inline uint8
lock_getLockOwnerTypeByIdx(vs_mem_index_t	lockOwnerIdx)
{
	uint8	ownerType;

	switch (VS_MEM_IDX_GET_DATA_TYPE(lockOwnerIdx))
	{
		case vs_mem_dt_cifsLockOwner:
			ownerType = LOCK_OWNER_CIFS;
			break;

		case vs_mem_dt_nlmLockOwner:
			ownerType = LOCK_OWNER_NLM;
			break;

		default:
			ownerType = LOCK_OWNER_DUMMY;
			break;
	}

	return (ownerType);
} /* lock_getLockOwnerByIdx() */

/*-----------------------------------------------------------------
 * Name:         lock_getLockOwnerByIdx
 *
 * Description:
 *
 *     Get the CIFS or NLM lock owner data structure given its index.
 *
 * Arguments:
 * 
 *     lockOwnerIdx - Index to the lock_cifsLockOwner_t or
 *                    lock_nlmLockOwner_t structure.
 *
 *     lockOwnerTypep - This is output. The lock owner type.
 *
 * Return value:
 *
 *     Pointer to the lock_cifsLockOwner_t or lock_nlmLockOwner_t structure -
 *                                                            If succeed.
 *
 *     NULL - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */ 
inline void *
lock_getLockOwnerByIdx(vs_mem_index_t	lockOwnerIdx,
					   uint8			*lockOwnerTypep)
{
	void *lockOwnerp;

	switch (VS_MEM_IDX_GET_DATA_TYPE(lockOwnerIdx))
	{
		case vs_mem_dt_cifsLockOwner:
			*lockOwnerTypep = LOCK_OWNER_CIFS;
			break;

		case vs_mem_dt_nlmLockOwner:
			*lockOwnerTypep = LOCK_OWNER_NLM;
			break;

		default:
			*lockOwnerTypep = LOCK_OWNER_DUMMY;
			break;
	}
	lockOwnerp = vs_mem_idxToPtr(lockOwnerIdx);

	return (lockOwnerp);
} /* lock_getLockOwnerByIdx() */

/*-----------------------------------------------------------------
 * Name:         lock_getOwnerTypeByLockInfoIdx
 *
 * Description:
 *
 *     Get lock owner type from lockInfo index.
 *
 * Arguments:
 * 
 *     lockInfoIdx - Index to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure.
 *
 * Return value:
 *
 *     Lock owner type.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */ 
inline uint8
lock_getOwnerTypeByLockInfoIdx(vs_mem_index_t lockInfoIdx)
{
	uint8	ownerType;

	switch (VS_MEM_IDX_GET_DATA_TYPE(lockInfoIdx))
	{
		case vs_mem_dt_cifsLockInfo:
			ownerType = LOCK_OWNER_CIFS;
			break;

		case vs_mem_dt_nlmLockInfo:
			ownerType = LOCK_OWNER_NLM;
			break;

		default:
			ownerType = LOCK_OWNER_DUMMY;
			break;
	}

	return (ownerType);
} /* lock_getOwnerTypeByLockInfoIdx() */

/*-----------------------------------------------------------------
 * Name:         lock_getLockOwnerIdxByLockInfo
 *
 * Description:
 *
 *     Get the index to CIFS or NLM lock owner structure given 
 *     lock info and lock owner type.
 *
 * Arguments:
 * 
 *     lockInfoIdx - Index to the lock_cifsLockInfo_t or
 *                   lock_nlmLockInfo_t structure.
 *
 *     lockInfop - Pointer to the lock_cifsLockInfo_t or
 *                 lock_nlmLockInfo_t structure.
 *
 *         At least one of the above agruments is valid.
 *
 *     lockOwnerType - The lock owner type. This may not be valid if
 *                     lockInfop is invalid.
 *
 * Return value:
 *
 *     Index to the lock_cifsLockInfo_t or lock_nlmLockInfo_t structure -
 *                                                            If succeed.
 *
 *     Zero value index - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */ 
vs_mem_index_t
lock_getLockOwnerIdxByLockInfo(vs_mem_index_t	lockInfoIdx,
							   void				*lockInfop,
							   uint8			lockOwnerType)
{
	vs_mem_index_t	lockOwnerIdx;

	VS_MEM_IDX_SET_NULL(lockOwnerIdx);

	if (VS_MEM_IDX_IS_NULL(lockInfoIdx) && (lockInfop == NULL))
		return (lockOwnerIdx);

	if (lockInfop == NULL)
	{
		lockInfop = lock_getLockInfo(lockInfoIdx, &lockOwnerType);
		if (lockInfop == NULL)
			return (lockOwnerIdx);
	}

	switch (lockOwnerType)
	{
		case LOCK_OWNER_CIFS:
			lockOwnerIdx = ((lock_cifsLockInfo_t *)lockInfop)->owner;
			break;

		case LOCK_OWNER_NLM:
			lockOwnerIdx = ((lock_nlmLockInfo_t *)lockInfop)->owner;
			break;

		default:
			break;
	}

	return (lockOwnerIdx);
} /* lock_getLockOwnerIdxByLockInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_initCifsLockInfo
 *
 * Description:
 *
 *     Initialize a lock_cifsLockInfo_t data structure, given lock request.
 *
 * Arguments:
 * 
 *     lockReqp - Pointer to data structure that contains lock information
 *                used to initialize cifsLockInfo structure.
 *
 *     cifsLockInfop - Pointer to the initialized lock_cifsLockInfo_t
 *                     structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/27/2002
 *
 *-----------------------------------------------------------------
 */ 
int32
lock_initCifsLockInfo(lock_request_t		*lockReqp,
					  lock_cifsLockInfo_t	*cifsLockInfop)
{
	vs_mem_index_t			cifsLockOwnerIdx;
	lock_cifsLockOwner_t	*cifsLockOwnerp;

	if ((lockReqp == NULL) || (cifsLockInfop == NULL))
		return (NFX_ERR);

	/*---------------
	 * Get lock owner
	 *---------------
	 */
	cifsLockOwnerp = lock_getCifsLockOwner(&(lockReqp->owner.cifsOwner),
										   &cifsLockOwnerIdx);
	if (cifsLockOwnerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_initCifsLockInfo: "
								"Couldn't get CIFS lock owner."));
		return (NFX_ERR);
	}

	/*--------------------------------------------
	 * Init the lock_cifsLockInfo_t data structure
	 *--------------------------------------------
	 */
	cifsLockInfop->owner = cifsLockOwnerIdx;
	cifsLockInfop->offset = lockReqp->offset;
	cifsLockInfop->len = lockReqp->len;

	return (NFX_OK);
} /* lock_initCifsLockInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_initNlmLockInfo
 *
 * Description:
 *
 *     Initialize a lock_nlmLockInfo_t data structure, given lock request.
 *
 * Arguments:
 * 
 *     lockReqp - Pointer to data structure that contains lock information
 *                used to initialize nlmLockInfo structure.
 *
 *     nlmLockInfop - Pointer to the initialized lock_nlmLockInfo_t
 *                    structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/27/2002
 *
 *-----------------------------------------------------------------
 */ 
int32
lock_initNlmLockInfo(lock_request_t		*lockReqp,
					 lock_nlmLockInfo_t	*nlmLockInfop)
{
	vs_mem_index_t			nlmLockOwnerIdx;
	lock_nlmLockOwner_t		*nlmLockOwnerp;

	if ((lockReqp == NULL) || (nlmLockInfop == NULL))
		return (NFX_ERR);

	/*---------------
	 * Get lock owner
	 *---------------
	 */
	nlmLockOwnerp = lock_getNlmLockOwner(&(lockReqp->owner.nlmOwner),
										 lockReqp->fcb,
										 &nlmLockOwnerIdx);
	if (nlmLockOwnerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_initNlmLockInfo: "
								"Couldn't get NLM lock owner."));
		return (NFX_ERR);
	}

	/*-------------------------------------------
	 * Init the lock_nlmLockInfo_t data structure
	 *-------------------------------------------
	 */
	nlmLockInfop->owner = nlmLockOwnerIdx;
	nlmLockInfop->offset = lockReqp->offset;
	nlmLockInfop->len = lockReqp->len;
	nlmLockInfop->fcb = lockReqp->fcb;
	nlmLockInfop->idxInRecFile = 0;

	return (NFX_OK);
} /* lock_initNlmLockInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_createLockInfo
 *
 * Description:
 *
 *     Allocate memory and initialize a lock_cifsLockInfo_t
 *     or lock_nlmLockInfo_t data structure.
 *
 * Arguments:
 * 
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure. The lock information it contains will
 *                 be copied to the newly created lockInfo structure.
 *
 *     lockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                     of the lock to be created.
 *
 *     lockInfoIdxp - This is output. It is pointer to the index to
 *                    the created lock_nlmLockInfo_t structure.
 *
 * Return value:
 *
 *     Pointer to the created lockInfo structure - If succeed.
 *
 *     NULL - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/27/2002
 *
 *-----------------------------------------------------------------
 */ 
void *
lock_createLockInfo(void			*lockInfop,
					uint8			lockOwnerType,
					vs_mem_index_t	*lockInfoIdxp)
{
	if (lockInfop == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_createLockInfo: Input NULL pointer."));
		return (NULL);
	}

	if (lockOwnerType == LOCK_OWNER_CIFS)
	{
		lock_cifsLockOwner_t	*cifsLockOwnerp;
		lock_cifsLockInfo_t		*cifsLockInfop;

		/*--------------------
		 * Get CIFS lock owner
		 *--------------------
		 */
		cifsLockInfop = (lock_cifsLockInfo_t *)lockInfop;
		cifsLockOwnerp = (lock_cifsLockOwner_t *)
								vs_mem_idxToPtr(cifsLockInfop->owner);
		if (cifsLockOwnerp == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_lock_createLockInfo: "
									"Couldn't find CIFS lock owner."));
			return (NULL);
		}

		/*--------------------------------------------
		 * Create a lock_cifsLockInfo_t data structure
		 *--------------------------------------------
		 */
		cifsLockInfop = lock_memAllocCifsLockInfo(cifsLockInfop->owner,
												  cifsLockOwnerp,
												  cifsLockInfop->offset,
												  cifsLockInfop->len,
												  lockInfoIdxp);
		if (cifsLockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_createLockInfo: "
						"Failed to allocate cifsLockInfo data structure."));
			return (NULL);
		}
		return (cifsLockInfop);
	}
	else if (lockOwnerType == LOCK_OWNER_NLM)
	{
		lock_nlmLockOwner_t		*nlmLockOwnerp;
		lock_nlmLockInfo_t		*nlmLockInfop;

		/*-------------------
		 * Get NLM lock owner
		 *-------------------
		 */
		nlmLockInfop = (lock_nlmLockInfo_t *)lockInfop;
		nlmLockOwnerp = (lock_nlmLockOwner_t *)
								vs_mem_idxToPtr(nlmLockInfop->owner);
		if (nlmLockOwnerp == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_createLockInfo: "
									"Couldn't find NLM lock owner."));
			return (NULL);
		}

		/*-------------------------------------------
		 * Create a lock_nlmLockInfo_t data structure
		 *-------------------------------------------
		 */
		nlmLockInfop = lock_memAllocNlmLockInfo(nlmLockInfop->owner,
												nlmLockOwnerp,
												nlmLockInfop->offset,
												nlmLockInfop->len,
												nlmLockInfop->fcb,
												lockInfoIdxp);
		if (nlmLockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_createLockInfo: "
						"Failed to allocate nlmLockInfo data structure."));
			return (NULL);
		}
		return (nlmLockInfop);
	}
	else
	{
		LOCK_PRINTF(LOG_ERR, ("lock_createLockInfo: Invalid lock owner type."));
		return (NULL);
	}
} /* lock_createLockInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_getLockInfoIdxByLockIdx
 *
 * Description:
 *
 *     Get index to a lockInfo data structure, given the index to
 *     the lock structure.
 *
 * Arguments:
 * 
 *     lockIdx - Index to a lock_exLock_t or lock_shLock_t structure.
 *
 * Return value:
 *
 *     Index to to lock_cifsLockInfo_t or lock_nlmLockInfo_t structure - 
 *                                                                 If succeed.
 *
 *     Zero valued index - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */ 
vs_mem_index_t
lock_getLockInfoIdxByLockIdx(vs_mem_index_t lockIdx)
{
	void			*lockp;
	uint8			lockType;
	vs_mem_index_t	lockInfoIdx;

	lockp = lock_getGrantedLock(lockIdx, &lockType);
	if (lockp == NULL)
	{
		VS_MEM_IDX_SET_NULL(lockInfoIdx);
		return(lockInfoIdx);
	}

	switch (lockType)
	{
		case LOCK_TYPE_EXCL:
			lockInfoIdx = ((lock_exLock_t *)lockp)->info;
			break;

		case LOCK_TYPE_SHAR:
			lockInfoIdx = ((lock_shLock_t *)lockp)->info;
			break;

		default:
			VS_MEM_IDX_SET_NULL(lockInfoIdx);
			break;
	}

	return (lockInfoIdx);
} /* lock_getLockInfoIdxByLockIdx() */

/*-----------------------------------------------------------------
 * Name:         lock_getLockInfo
 *
 * Description:
 *
 *     Get memory address of a lockInfo data structure, given
 *     the index to the data structure.
 *
 * Arguments:
 * 
 *     lockInfoIdx - Index to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure.
 *
 *     lockOwnerTypep - This is output. The type of owner (LOCK_OWNER_CIFS
 *                      or LOCK_OWNER_NLM) of the lock.
 *
 * Return value:
 *
 *     Pointer to lock_cifsLockInfo_t or lock_nlmLockInfo_t structure - 
 *                                                                 If succeed.
 *
 *     NULL - If fail.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */ 
void *
lock_getLockInfo(vs_mem_index_t	lockInfoIdx,
				 uint8			*lockOwnerTypep)
{
	void *lockInfop;

	/* Get lock owner type */
	*lockOwnerTypep = lock_getOwnerTypeByLockInfoIdx(lockInfoIdx);

	/* Get pointer to lockInfo structure */
	lockInfop = vs_mem_idxToPtr(lockInfoIdx);
	if (lockInfop == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_getLockInfo: "
					"Couldn't find lockInfo structure."));
		return (NULL);
	}

	return (lockInfop);
} /* lock_getLockInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_getLockTypeByLockIdx
 *
 * Description:
 *
 *     Get lock type (exclusive or shared) from lock index.
 *
 * Arguments:
 * 
 *     lockIdx - Index to a lock_exLock_t or lock_shLockInfo_t structure.
 *
 * Return value:
 *
 *     Lock type.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */ 
inline uint8
lock_getLockTypeByLockIdx(vs_mem_index_t lockIdx)
{
	uint8	lockType;

	switch (VS_MEM_IDX_GET_DATA_TYPE(lockIdx))
	{
		case vs_mem_dt_exLock:
			lockType = LOCK_TYPE_EXCLUSIVE;
			break;

		case vs_mem_dt_shLock:
			lockType = LOCK_TYPE_SHARED;
			break;

		default:
			lockType = LOCK_TYPE_DUMMY;
			break;
	}

	return (lockType);
} /* lock_getLockTypeByLockIdx() */

/*-----------------------------------------------------------------
 * Name:         lock_getGrantedLock
 *
 * Description:
 *
 *     Get memory address of an exLock or shLock data structure,
 *     given the index of the data structure.
 *
 * Arguments:
 * 
 *     lockIdx - Index to a lock_exLock_t or lock_shLockInfo_t structure.
 *
 *     lockTypep - This is output. The value is lock type.
 *
 * Return value:
 *
 *     Pointer to lock_exLock_t or lock_shLock_t structure - If succeed.
 *
 *     NULL - If failed.
 *  
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */ 
void *
lock_getGrantedLock(vs_mem_index_t	lockIdx,
					uint8			*lockTypep)
{
	void *lockp;

	if (lockTypep == NULL)
		return (NULL);

	/* Get lock type */
	*lockTypep = lock_getLockTypeByLockIdx(lockIdx);

	/* Get pointer to lock structure */
	lockp = vs_mem_idxToPtr(lockIdx);
	if (lockp == NULL)
	{
		LOCK_PRINTF(LOG_DEBUG, ("lock_getGrantedLock: "
					"Couldn't find lock structure."));
		return (NULL);
	}

	return (lockp);
} /* lock_getGrantedLock() */

/*-----------------------------------------------------------------
 * Name:        lock_getLockInformationFromLockInfo
 *
 * Description:
 *  
 *     Get lock owner type, lock offset, and lock length
 *     from a lock_cifsLockInfo_t or lock_nlmLockInfo_t structure.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure.
 *  
 *     lockInfop - Pointer to the lock_cifsLockInfo_t
 *                 or lock_nlmLockInfo_t structure.
 *  
 *     lockOwnerTypep - The value is the type of owner of the above lock.
 *                      If the input of this argument is invalid, then
 *                      it contains lock owner type on return.
 *
 *         As to the above three arguments, the caller either,
 *             - provides lockInfoIdx and
 *               leaves lockInfop and lockOwnerType invalid, or
 *             - provides lockInfop and lockOwnerType and
 *               leaves lockInfoIdx invalid.
 *  
 *     lockOffsetp - This is output. It contains byte range lock offset. 
 *  
 *     lockLenp - This is output. It contains byte range lock length. 
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_getLockInformationFromLockInfo(vs_mem_index_t	lockInfoIdx,
									void			*lockInfop,
									uint8			*lockOwnerTypep,
									uint64			*lockOffsetp,
									uint64			*lockLenp)
{
	int32	rc;

	if ((lockOwnerTypep == NULL) ||
		(lockOffsetp == NULL) ||
		(lockLenp == NULL))
		return (NFX_ERR);

	if (!lockInfop)
	{
		/*--------------------------------------------------
		 * Get lockInfo and lock owner type from lockInfoIdx
		 *--------------------------------------------------
		 */
		lockInfop = lock_getLockInfo(lockInfoIdx, lockOwnerTypep);
		if (lockInfop == NULL)
			return (NFX_ERR);
	}

	/*---------------------------
	 * Get lock offset and length 
	 *---------------------------
	 */
	if (*lockOwnerTypep == LOCK_OWNER_CIFS)
	{
		/* CIFS lock */
		lock_cifsLockInfo_t		*cifsLockInfop;

		cifsLockInfop = (lock_cifsLockInfo_t *)lockInfop;
		*lockOffsetp = cifsLockInfop->offset;
		*lockLenp = cifsLockInfop->len;

		rc = NFX_OK;
	}
	else if (*lockOwnerTypep == LOCK_OWNER_NLM)
	{
		/* NLM lock */
		lock_nlmLockInfo_t	*nlmLockInfop;

		nlmLockInfop = (lock_nlmLockInfo_t *)lockInfop;
		*lockOffsetp = nlmLockInfop->offset;
		*lockLenp = nlmLockInfop->len;

		rc = NFX_OK;
	}
	else
	{
		LOCK_PRINTF(LOG_ERR, ("lock_getLockInformationFromLockInfo: "
											"Invalid lock owner type."));
		rc = NFX_ERR;
	}

	return (rc);
} /* lock_getLockInformationFromLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_getLockInformationFromShNode
 *
 * Description:
 *  
 *     Get lock offset, and lock length from a shNode structure.
 *  
 * Arguments:    
 *      
 *     shNodep - Pointer to a shared node in shared lock tree.
 *  
 *     lockOffsetp - This is output. It contains byte range lock offset. 
 *  
 *     lockLenp - This is output. It contains byte range lock length. 
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_getLockInformationFromShNode(lock_shNode_t	*shNodep,
								  uint64		*lockOffsetp,
								  uint64		*lockLenp)
{
	int32			rc;
	lock_shLock_t	*shLockp;
	uint8			firstLockOwnerType;
	uint64			firstLockLen;

	if ((lockOffsetp == NULL) || (lockLenp == NULL))
		return (NFX_ERR);

	/*---------------------------
	 * Get first lock in the list
	 *---------------------------
	 */
	shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shNodep->lockList);
	if (shLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_getLockInformationFromShNode: "
							"Couldn't find first shLock in the list."));
		return (NFX_ERR);
	}

	/*----------------
	 * Get lock offset
	 *----------------
	 */
	rc = lock_getLockInformationFromLockInfo(shLockp->info,
											 NULL,
											 &firstLockOwnerType,
											 lockOffsetp,
											 &firstLockLen);
	if (rc != NFX_OK)
		return (rc);

	/*----------------
	 * Get lock length
	 *----------------
	 */
	if ((*lockOffsetp == shNodep->end) && (firstLockLen == 0))
	{
		/* The list contains only one zero-length lock */
		*lockLenp = 0;
	}
	else
		LOCK_GET_LOCK_RANGE_LENGTH(*lockOffsetp, *lockLenp, shNodep->end);

	return (NFX_OK);
} /* lock_getLockInformationFromShNode() */

/*-----------------------------------------------------------------
 * Name:        lock_matchKey
 *
 * Description:
 *  
 *     Check if a lock owner has the same key (pid+fid) as input.
 *  
 * Arguments:    
 *      
 *     lockOwnerIdx - Index to the lockOwner structure.
 *  
 *     inKey - Combination of fid and pid.
 *
 *     fidOnly - TRUE if only check fid,
 *               FALSE if check both fid and pid.
 *
 * Return value:
 *      
 *     TRUE - If the lock owner match the key.
 * 
 *     FALSE - If the lock owner doesn't match. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_matchKey(vs_mem_index_t	lockOwnerIdx,
			  uint32			inKey,
			  boolean			fidOnly)
{
	uint8					lockOwnerType;
	lock_cifsLockOwner_t	*ownerp;

	/*-----------------------------------------------------
	 * Verify that the lock's owner type is LOCK_OWNER_CIFS
	 *-----------------------------------------------------
	 */
	lockOwnerType = lock_getLockOwnerTypeByIdx(lockOwnerIdx);
	if (lockOwnerType != LOCK_OWNER_CIFS)
	{
		return (FALSE);
	}

	/*---------------
	 * Get lock owner
	 *---------------
	 */
	ownerp = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx);
	if (ownerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_matchKey: "
								"NULL pointer to owner"));
		return (FALSE);
	}

	/*----------------
	 * Check for match
	 *----------------
	 */
	if (fidOnly == TRUE)
	{
		/* Only check fid match */
		uint16	inFid = LOCK_GET_FID_FROM_KEY(inKey);

		LOCK_PRINTF(LOCK_DEBUG2, ("lock_matchKey: "
						"input fid: %d, ownerp->fid: %d", inFid, ownerp->fid));

		if (ownerp->fid == inFid)
			return (TRUE);
		else
			return (FALSE);
	}
	else
	{
		/* Check both fid and pid */
		uint32	key = LOCK_GENERATE_KEY(ownerp->fid, ownerp->pid);

		LOCK_PRINTF(LOCK_DEBUG2, ("lock_matchKey: "
						"input key: %d, ownerp->fid: %d", inKey, key));

		if (key == inKey)
			return (TRUE);
		else
			return (FALSE);
	}
} /* lock_matchKey() */

/*-----------------------------------------------------------------
 * Name:        lock_matchClientIpAddr
 *
 * Description:
 *  
 *     Check if a lock owner has the same client IP address as input.
 *  
 * Arguments:    
 *      
 *     lockOwnerIdx - Index to the lockOwner structure.
 *  
 *     clientIpAddr - Client IP address.
 *
 * Return value:
 *      
 *     TRUE - If the lock owner match the client IP address.
 * 
 *     FALSE - If the lock owner doesn't match. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_matchClientIpAddr(vs_mem_index_t	lockOwnerIdx,
			   		   ip_index_t		inClientIpAddr)
{
	uint8					lockOwnerType;
	lock_nlmLockOwner_t		*ownerp;
	lock_nlmLockClient_t	*clientp;

	/*----------------------------------------------------
	 * Verify that the lock's owner type is LOCK_OWNER_NLM
	 *----------------------------------------------------
	 */
	lockOwnerType = lock_getLockOwnerTypeByIdx(lockOwnerIdx);
	if (lockOwnerType != LOCK_OWNER_NLM)
	{
		return (FALSE);
	}

	/*-------------------------------
	 * Get lock client and lock owner
	 *-------------------------------
	 */
	ownerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx);
	if (ownerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_matchClientIpAddr: "
								"NULL pointer to owner"));
		return (FALSE);
	}
	clientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(ownerp->lockClient);
	if (clientp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_matchClientIpAddr: "
								"NULL pointer to client"));
		return (FALSE);
	}

	/*----------------
	 * Check for match
	 *----------------
	 */
	if (clientp->clientIpAddr == inClientIpAddr)
		return (TRUE);
	else
		return (FALSE);
} /* lock_matchClientIpAddr() */

/*-----------------------------------------------------------------
 * Name:        lock_match
 *
 * Description:
 *  
 *     Check if a lock owner has the same key (pid+fid) or
 *     client IP address as input.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to a lockInfo structure. Its lock owner
 *                   will be checked.
 *  
 *     lockOwnerIdx - Index to a lockOwner structure.
 *  
 *     useLockInfo - TRUE if lockInfoIdx is valid.
 *
 *     key - Combination of fid and pid.
 *
 *     clientIpAddr - Client IP address.
 *
 *     searchMask - The criteria based on which we check for a match.
 *
 * Return value:
 *      
 *     TRUE - If the lock match the key or client IP address.
 * 
 *     FALSE - If the lock doesn't match. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_match(vs_mem_index_t	lockInfoIdx,
		   vs_mem_index_t	lockOwnerIdx,
		   boolean			useLockInfo,
		   uint32			inKey,
		   ip_index_t		inClientIpAddr,
		   uint16			searchMask)
{
	boolean		fidOnly, matchResult;
	uint8		lockOwnerType;

	if (useLockInfo == TRUE)
	{
	 	/* Get index to the lock owner */
		lockOwnerIdx = lock_getLockOwnerIdxByLockInfo(lockInfoIdx,
													  NULL,
													  lockOwnerType);
	}

	if (searchMask != LOCK_SEARCH_SAME_IPADDR)
	{
		/* Match fid or key (CIFS) */
		if (searchMask == LOCK_SEARCH_SAME_FID)
			fidOnly = TRUE;
		else if (searchMask == LOCK_SEARCH_SAME_KEY)
			fidOnly = FALSE;
		else
		{
			/*Invalid searchMask*/
			LOCK_PRINTF(LOG_ERR, ("lock_match: Invalid searchMask (%d)",
																searchMask));
			return (FALSE);
		}

		matchResult = lock_matchKey(lockOwnerIdx,
									inKey,
								   	fidOnly);
	}
	else
	{
		matchResult = lock_matchClientIpAddr(lockOwnerIdx,
											 inClientIpAddr);
	}

	return (matchResult);
} /*lock_match() */

/*-----------------------------------------------------------------
 * Name:        lock_sameOwner
 *
 * Description:
 *  
 *     Check if two owners are same or not.
 *  
 * Arguments:    
 *      
 *     lockOwnerIdx1 - Index to the first lock owner.
 *
 *     lockOwnerIdx2 - Index to the second lock owner.
 *
 * Return value:
 *      
 *     TRUE - If two locks have the same owner.
 * 
 *     FALSE - If two locks have different owner.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_sameOwner(vs_mem_index_t	lockOwnerIdx1,
			   vs_mem_index_t	lockOwnerIdx2)
{
	uint8		lockOwnerType1, lockOwnerType2;

	/*----------------------
	 * Check lock owner type
	 *----------------------
	 */
	lockOwnerType1 = lock_getLockOwnerTypeByIdx(lockOwnerIdx1);
	lockOwnerType2 = lock_getLockOwnerTypeByIdx(lockOwnerIdx2);
	if (lockOwnerType1 != lockOwnerType2)
	{
		/* Locks are from different types of owners. */
		return (FALSE);
	}

	/*---------------------------------------
	 * Locks are from the same type of owners
	 *---------------------------------------
	 */
	if (lockOwnerType1 == LOCK_OWNER_CIFS)
	{
		/*------------
		 * CIFS owners
		 *------------
		 */
		lock_cifsLockOwner_t	*owner1, *owner2;

		/* Get owner of the first lock */
		owner1 = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx1);
		if (owner1 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: "
										"NULL pointer to CIFS owner1"));
			return (FALSE);
		}

		/* Get owner of the second lock */
		owner2 = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx2);
		if (owner2 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: "
										"NULL pointer to CIFS owner2"));
			return (FALSE);
		}

		LOCK_PRINTF(LOCK_DEBUG2, ("lock_sameOwner: "
						"CIFS tid1(%d) fid1(%d), pid1(%d), "
					   	"tid2(%d), fid2(%d), pid2(%d)",
						owner1->tid, owner1->fid, owner1->pid,
						owner2->tid, owner2->fid, owner2->pid));

		/* Compare tid, fid and pid of two owners */
		if ((owner1->tid == owner2->tid) &&
			(owner1->fid == owner2->fid) &&
			(owner1->pid == owner2->pid))
			return (TRUE);
		else
			return (FALSE);
	}
	else if (lockOwnerType1 == LOCK_OWNER_NLM)
	{
		/*-----------
		 * NLM owners
		 *-----------
		 */
		lock_nlmLockOwner_t		*owner1, *owner2;
		lock_nlmLockClient_t	*client1, *client2;

		/* Get owner of the first lock */
		owner1 = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx1);
		if (owner1 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: "
											"NULL pointer to NLM owner1"));
			return (FALSE);
		}
		client1 = (lock_nlmLockClient_t *)vs_mem_idxToPtr(owner1->lockClient);
		if (client1 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: "
											"NULL pointer to NLM client1"));
			return (FALSE);
		}

		/* Get owner of the second lock */
		owner2 = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx2);
		if (owner2 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: "
											"NULL pointer to NLM owner2"));
			return (FALSE);
		}
		client2 = (lock_nlmLockClient_t *)vs_mem_idxToPtr(owner2->lockClient);
		if (client2 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: "
											"NULL pointer to NLM client2"));
			return (FALSE);
		}

		LOCK_PRINTF(LOCK_DEBUG2, ("lock_sameOwner: "
						"NLM client1(0x%x), client2(0x%x), pid1(%d), pid2(%d)",
						client1->clientIpAddr, client2->clientIpAddr,
					   	owner1->pid, owner2->pid));


		/* Compare client IP address, pid, and owner name of two locks */
		if ((client1->clientIpAddr == client2->clientIpAddr) &&
			(owner1->pid == owner2->pid) &&
			(owner1->ownerNameLen == owner2->ownerNameLen) &&
			(memcmp(owner1->ownerName, owner2->ownerName, owner1->ownerNameLen) == 0))
			return (TRUE);
		else
			return (FALSE);	
	}
	else
	{
		LOCK_PRINTF(LOG_ERR, ("lock_sameOwner: Invalid lock owner type"));
		return (FALSE);
	}
} /* lock_sameOwner() */

/*-----------------------------------------------------------------
 * Name:        lock_haveSameOwner
 *
 * Description:
 *  
 *     Check if two locks have the same owner or not.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx1 - Index to the first lockInfo structure.
 *  
 *     lockInfop1 - Pointer to the first lockInfo structure.
 *  
 *         Only one of the above two paramenters is valid, not both.
 *  
 *     lockInfoIdx2 - Index to the second lockInfo structure.
 *  
 *     lockOwnerType1 - Owner type of the first lock.
 *
 *     lockOwnerType2 - Owner type of the second lock.
 *
 * Return value:
 *      
 *     TRUE - If two locks have the same owner.
 * 
 *     FALSE - If two locks have different owner.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_haveSameOwner(vs_mem_index_t	lockInfoIdx1,
				   void				*lockInfop1,
				   vs_mem_index_t	lockInfoIdx2,
				   uint8			lockOwnerType1,
				   uint8			lockOwnerType2)
{
	vs_mem_index_t	lockOwnerIdx1, lockOwnerIdx2;
	boolean			result;

	/*----------------------
	 * Check lock owner type
	 *----------------------
	 */
	if (lockOwnerType1 != lockOwnerType2)
	{
		/* Locks are from different types of owners. */
		return (FALSE);
	}

	/*---------------------------------------
	 * Locks are from the same type of owners
	 *---------------------------------------
	 */
	if (lockOwnerType1 == LOCK_OWNER_CIFS)
	{
		/*------------
		 * CIFS owners
		 *------------
		 */
		lock_cifsLockInfo_t		*lockInfo1, *lockInfo2;

		/* Get index to owner of the first lock */
		if (lockInfop1)
			lockInfo1 = (lock_cifsLockInfo_t *)lockInfop1;
		else
		{
			lockInfo1 = (lock_cifsLockInfo_t *)vs_mem_idxToPtr(lockInfoIdx1);
			if (lockInfo1 == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_haveSameOwner: "
										"NULL pointer to CIFS lockInfo1"));
				return (FALSE);
			}
		}
		lockOwnerIdx1 = lockInfo1->owner;

		/* Get owner of the second lock */
		lockInfo2 = (lock_cifsLockInfo_t *)vs_mem_idxToPtr(lockInfoIdx2);
		if (lockInfo2 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_haveSameOwner: "
										"NULL pointer to CIFS lockInfo2"));
			return (FALSE);
		}
		lockOwnerIdx2 = lockInfo2->owner;
	}
	else if (lockOwnerType1 == LOCK_OWNER_NLM)
	{
		/*-----------
		 * NLM owners
		 *-----------
		 */
		lock_nlmLockInfo_t		*lockInfo1, *lockInfo2;

		/* Get index to owner of the first lock */
		if (lockInfop1)
			lockInfo1 = (lock_nlmLockInfo_t *)lockInfop1;
		else
		{
			lockInfo1 = (lock_nlmLockInfo_t *)vs_mem_idxToPtr(lockInfoIdx1);
			if (lockInfo1 == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_haveSameOwner: "
											"NULL pointer to NLM lockInfo1"));
				return (FALSE);
			}
		}
		lockOwnerIdx1 = lockInfo1->owner;

		/* Get index to owner of the second lock */
		lockInfo2 = (lock_nlmLockInfo_t *)vs_mem_idxToPtr(lockInfoIdx2);
		if (lockInfo2 == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_haveSameOwner: "
											"NULL pointer to NLM lockInfo2"));
			return (FALSE);
		}
		lockOwnerIdx2 = lockInfo2->owner;
	}
	else
	{
		LOCK_PRINTF(LOG_ERR, ("lock_haveSameOwner: Invalid lock owner type"));
		return (FALSE);
	}

	result = lock_sameOwner(lockOwnerIdx1, lockOwnerIdx2);

	return (result);
} /* lock_haveSameOwner() */

/*-----------------------------------------------------------------
 * Name:        lock_findExclLockByLockInfo
 *
 * Description:
 *  
 *     Search for a lock in the exclusive lock tree given
 *     lockInfo and searchMask.
 *  
 * Arguments:    
 *      
 *     rootIdxp - Pointer to index to root of the exclusive lock tree.
 *  
 *     inLockInfop - Pointer to the lock_cifsLockInfo_t or
 *                   lock_nlmLockInfo_t structure. It contains
 *                   info of the lock to search for.
 *  
 *     inLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                       of the lock to search for.
 *  
 *     searchMask - The criteria based on which we search for the lock.
 *  
 * Return value:
 *      
 *     Point to the found exclusive lock - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */
inline lock_exLock_t *
lock_findExclLockByLockInfo(vs_mem_index_t	*rootIdxp,
							void			*inLockInfop,
							uint8			inLockOwnerType,
							uint16			searchMask)
{
	lock_exLock_t	*exLockp;

	exLockp = (lock_exLock_t *)lock_btree_find(rootIdxp,
											   NULL,
											   inLockInfop,
											   inLockOwnerType,
											   FALSE,
											   searchMask,
											   LOCK_TYPE_EXCL);
	return (exLockp);
} /* lock_findExclLockByLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_findExclLockByLock
 *
 * Description:
 *  
 *     Search for a lock in the exclusive lock tree given
 *     a lock and searchMask.
 *  
 * Arguments:    
 *      
 *     rootIdxp - Pointer to index to root of the exclusive lock tree.
 *  
 *     inLockp - Pointer to the lock_exLock_t structure. It contains
 *               info of the lock to search for.
 *  
 *     searchMask - The criteria based on which we search for the lock.
 *  
 * Return value:
 *      
 *     Point to the found exclusive lock - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */
inline lock_exLock_t *
lock_findExclLockByLock(vs_mem_index_t	*rootIdxp,
						lock_exLock_t	*inLockp,
						uint16			searchMask)
{
	lock_exLock_t	*exLockp;

	exLockp = (lock_exLock_t *)lock_btree_find(rootIdxp,
											   (lock_btNode_t *)inLockp,
											   NULL,
											   LOCK_OWNER_DUMMY,
											   TRUE,
											   searchMask,
											   LOCK_TYPE_EXCL);
	return (exLockp);
} /* lock_findExclLockByLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_findSharedNodeByLockInfo
 *
 * Description:
 *  
 *     Search for a node in the shared lock tree given
 *     lockInfo and searchMask.
 *  
 * Arguments:    
 *      
 *     rootIdxp - Pointer to index to root of the shared lock tree.
 *  
 *     inLockInfop - Pointer to the lock_cifsLockInfo_t or
 *                   lock_nlmLockInfo_t structure. We search for
 *                   a node whose lock range overlaps with that
 *                   of this lock.
 *  
 *     inLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                       of the input lock.
 *  
 *     searchMask - The criteria based on which we search for the node.
 *  
 * Return value:
 *      
 *     Point to the found shared node - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */
inline lock_shNode_t *
lock_findSharedNodeByLockInfo(vs_mem_index_t	*rootIdxp,
							  void				*inLockInfop,
							  uint8				inLockOwnerType,
							  uint16			searchMask)
{
	lock_shNode_t	*shNodep;

	shNodep = (lock_shNode_t *)lock_btree_find(rootIdxp,
											   NULL,
											   inLockInfop,
											   inLockOwnerType,
											   FALSE,
											   searchMask,
											   LOCK_TYPE_SHAR);
	return (shNodep);
} /* lock_findSharedNodeByLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_findSharedNodeByNode
 *
 * Description:
 *  
 *     Search for a node in the shared lock tree given
 *     a shNode and searchMask.
 *  
 * Arguments:    
 *      
 *     rootIdxp - Pointer to index to root of the shared lock tree.
 *  
 *     inShNodep - Pointer to a shared node. We search for a node whose
 *                 lock range overlaps with that of this node.
 *  
 *     searchMask - The criteria based on which we search for the node.
 *  
 * Return value:
 *      
 *     Point to the found shared node - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */
inline lock_shNode_t *
lock_findSharedNodeByNode(vs_mem_index_t	*rootIdxp,
						  lock_shNode_t		*inShNodep,
						  uint16			searchMask)
{
	lock_shNode_t	*shNodep;

	shNodep = (lock_shNode_t *)lock_btree_find(rootIdxp,
											   (lock_btNode_t *)inShNodep,
											   NULL,
											   LOCK_OWNER_DUMMY,
											   TRUE,
											   searchMask,
											   LOCK_TYPE_SHAR);
	return (shNodep);
} /* lock_findSharedNodeByNode() */

/*-----------------------------------------------------------------
 * Name:        lock_findSharedLockByLockInfo
 *
 * Description:
 *  
 *     Search for a shared lock in the shared lock tree given
 *     lockInfo and searchMask.
 *  
 * Arguments:    
 *      
 *     rootIdxp - Pointer to index to root of the shared lock tree.
 *  
 *     inLockInfop - Pointer to the lock_cifsLockInfo_t or
 *                   lock_nlmLockInfo_t structure. It contains
 *                   info of the lock to search for.
 *  
 *     inLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                       of the lock to search for.
 *  
 *     searchMask - The criteria based on which we search for the lock.
 *
 *     shNodepp - This is output. The value is pointer to the node
 *                the found shared lock belonged to.
 *
 *     prevShLockIdxp - This is output. The value is index to the lock
 *                      before the found one in the list.
 *
 * Return value:
 *      
 *     Point to the found shared lock - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/12/2002
 *
 *-----------------------------------------------------------------
 */
lock_shLock_t *
lock_findSharedLockByLockInfo(vs_mem_index_t	*rootIdxp,
							  void				*inLockInfop,
							  uint8				inLockOwnerType,
							  uint16			searchMask,
							  lock_shNode_t		**shNodepp,
							  vs_mem_index_t	*prevShLockIdxp)
{
	int8			comp;
	lock_shLock_t	*shLockp;

	*shNodepp = NULL;

	/* Find an overlapping shared node */
	*shNodepp = (lock_shNode_t *)lock_btree_find(rootIdxp,
												 NULL,
												 inLockInfop,
												 inLockOwnerType,
												 FALSE,
												 searchMask,
												 LOCK_TYPE_SHAR);
	if (*shNodepp == NULL)
		return (NULL);

	/* Search the linked list of the found shared node, to find the
	 * matching shared lock
	 */
	comp = lock_sharedLockInfoAndNodeCompFunc(inLockInfop,
											  inLockOwnerType,
											  *shNodepp,
											  searchMask,
											  LOCK_SPECIAL_COMP_RESULT,
											  &shLockp,
											  prevShLockIdxp);

	if (comp == 0)
	{
		/* Found the lock */
		return (shLockp);
	}
	else
	{
		return (NULL);
	}
} /* lock_findSharedLockByLockInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_mergeLockRanges
 *
 * Description:
 *
 *     Merge the two lock ranges.
 *
 * Arguments:    
 * 
 *     lockInfoIdxp1 - Pointer to index to lockInfo of first lock.
 *                     On return, the first lock contains merged 
 *                     lock range.
 *
 *     lockInfoIdxp2 - Pointer to index to lockInfo of second lock.
 *
 *     lockInfop2 - Pointer to lockInfo of second lock.
 *
 *         Either lockInfoIdxp2 or lockInfop2 is valid, not both.
 *
 *     lockOwnerType2 - Owner type of the second lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/11/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_mergeLockRanges(vs_mem_index_t *lockInfoIdxp1,
					 vs_mem_index_t *lockInfoIdxp2,
					 void			*lockInfop2,
					 uint8			lockOwnerType2)
{
	int32			rc, i;
	uint8			lockOwnerType1;
	vs_mem_index_t	*lockInfoIdxp[2];
	uint64			offset[2];
    uint64			end[2], newEnd;

	if (lockInfoIdxp1 == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_mergeLockRanges: "
						"NULL pointer of lockInfo 1."));
		return (NFX_ERR);
	}

	if ((lockInfoIdxp2 == NULL) && (lockInfop2 == NULL))
	{
		LOCK_PRINTF(LOG_ERR, ("lock_mergeLockRanges: "
						"NULL pointer of lockInfo 2."));
		return (NFX_ERR);
	}

	/*---------------------------------------------------
	 * Make sure two lock structures are of the same type
	 *---------------------------------------------------
	 */
	/* Get lockOwnerType1 */
	lockOwnerType1 = lock_getOwnerTypeByLockInfoIdx(*lockInfoIdxp1);

	/* Get lockOwnerType2, if not provided as an input */
	if (lockInfop2 == NULL)
		lockOwnerType2 = lock_getOwnerTypeByLockInfoIdx(*lockInfoIdxp2);

	if (lockOwnerType1 != lockOwnerType2)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_mergeLockRanges: "
						"Two lock structures should be of the same type."));
		return (NFX_ERR);
	}

	/*-------------------------------------------------------------------
	 * Get offsets and lengths of two locks, and then calculate new range
	 *-------------------------------------------------------------------
	 */
	lockInfoIdxp[0] = lockInfoIdxp1;
	lockInfoIdxp[1] = lockInfoIdxp2;

	if (lockOwnerType1 == LOCK_OWNER_CIFS)
	{
		/*-----------
		 * CIFS locks
		 *-----------
		 */
		lock_cifsLockInfo_t	*cifsLockInfop;

	 	/* Get offsets and lengths of two locks */
		for (i = 1; i >= 0; i--)
		{
			if ((i == 1) && (lockInfop2 != NULL))
				cifsLockInfop = (lock_cifsLockInfo_t *)lockInfop2;
			else
			{
				cifsLockInfop = (lock_cifsLockInfo_t *)
										vs_mem_idxToPtr(*lockInfoIdxp[i]);
				if (cifsLockInfop == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_mergeLockRanges: "
							"Couldn't find cifsLockInfo structure %d.", i+1));
					return (NFX_ERR);
				}
			}
			offset[i] =  cifsLockInfop->offset;
			LOCK_GET_LOCK_RANGE_END(offset[i], cifsLockInfop->len, end[i]);
		}

	 	/* Calculate new range, and store it in cifsLockInfo1 */
		cifsLockInfop->offset = (offset[0] <= offset[1]) ? offset[0] : offset[1];
		newEnd = (end[0] >= end[1]) ? end[0] : end[1];
		LOCK_GET_LOCK_RANGE_LENGTH(cifsLockInfop->offset, cifsLockInfop->len, newEnd);

		rc = NFX_OK;
	}
	else if (lockOwnerType1 == LOCK_OWNER_NLM)
	{
		/*----------
		 * NLM locks
		 *----------
		 */
		lock_nlmLockInfo_t  *nlmLockInfop;

	 	/* Get offsets and lengths of two locks */
		for (i = 1; i >= 0; i--)
		{
			if ((i == 1) && (lockInfop2 != NULL))
				nlmLockInfop = (lock_nlmLockInfo_t *)lockInfop2;
			else
			{
				nlmLockInfop = (lock_nlmLockInfo_t *)
										vs_mem_idxToPtr(*lockInfoIdxp[i]);
				if (nlmLockInfop == NULL)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_mergeLockRanges: "
							"Couldn't find nlmLockInfo structure %d.", i+1));
					return (NFX_ERR);
				}
			}
			offset[i] = nlmLockInfop->offset;
			LOCK_GET_LOCK_RANGE_END(offset[i], nlmLockInfop->len, end[i]);
		}

	 	/* Calculate new range, and store it in cifsLockInfo1 */
		nlmLockInfop->offset = (offset[0] <= offset[1]) ? offset[0] : offset[1];
		newEnd = (end[0] >= end[1]) ? end[0] : end[1];
		LOCK_GET_LOCK_RANGE_LENGTH(nlmLockInfop->offset, nlmLockInfop->len, newEnd);

		rc = NFX_OK;
	}
	else
	{

		LOCK_PRINTF(LOG_ERR, ("lock_mergeLockRanges: Invalid owner type."));
		rc = NFX_ERR;
	}

	return (rc);
} /* lock_mergeLockRanges() */

/*-----------------------------------------------------------------
 * Name:         lock_mergeExLocks
 *
 * Description:
 *
 *     Find all the locks in the exclusive lock tree that overlap with
 *     the lock given as input. Merge these locks into one.
 *     Before this function is called, we've already made sure that
 *     there is no conflict in the exclusive lock tree. That is to say,
 *     the overlapping locks found in this function should have same owner.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to lockInfo. We merge this lock with other
 *                 locks that overlap with it.
 *
 *     lockOwnerType - Owner type of the lock.
 *
 *     exLockIdxp - This is output. The value is index to the merged lock.
 *
 *     exLockpp - This is output. The value is pointer to the merged lock
 *                if succeed; NULL if failed.
 * Return value:
 *
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_mergeExLocks(lock_fileLock_t	*fileLockp,
				  void				*lockInfop,
				  uint8				lockOwnerType,
				  vs_mem_index_t	*exLockIdxp,
				  lock_exLock_t		**exLockpp)
{
	int32			rc, rcFinal = NFX_OK;
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask = 0;
	lock_exLock_t	*overlapExLockp;

	if ((lockInfop == NULL) || (exLockIdxp == NULL))
		return (NFX_ERR);

	VS_MEM_IDX_SET_NULL(*exLockIdxp);
	*exLockpp = NULL;

	rootIdxp = &(fileLockp->exclLockTree);

	/*---------------------------------------------
	 * Find a lock overlapping with the input lock.
	 * Merge its range with the input lock.
	 *---------------------------------------------
	 */
	searchMask = 0;
	*exLockpp = lock_findExclLockByLockInfo(rootIdxp,
											lockInfop,
											lockOwnerType,
											searchMask);
	if (*exLockpp == NULL)
		return (NFX_OK);

	*exLockIdxp = *rootIdxp;

	rc = lock_mergeLockRanges(&((*exLockpp)->info),
							  NULL,
							  lockInfop,
							  lockOwnerType);
	if (rc != NFX_OK)
		return (rc);

	/*--------------------------------------------------
	 * Find other overlapping locks and merge the ranges
	 *--------------------------------------------------
	 */
	searchMask = LOCK_SEARCH_DIFF_RANGE;
	while (1)
	{
		/*------------------------------------------------
		 * Find next lock overlapping with the merged lock
		 *------------------------------------------------
		 */ 
		overlapExLockp = lock_findExclLockByLock(rootIdxp,
												 *exLockpp,
												 searchMask);
		if (overlapExLockp == NULL)
		{
			/* No more overlapping lock */
			break;
		}

		/* The overlapping locks must have same owner */

		/*-----------------
		 * Merge the ranges
		 *-----------------
		 */
		rc = lock_mergeLockRanges(&((*exLockpp)->info),
								  &(overlapExLockp->info),
								  NULL,
								  lockOwnerType);
		if (rc != NFX_OK)
		{
			rcFinal = rc;
			break;
		}

		/*------------------------------------
		 * Delete the found lock from the tree
		 *------------------------------------
		 */
		rc = lock_delExclLock(fileLockp,
							  NULL,
							  LOCK_OWNER_DUMMY,
							  overlapExLockp);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_mergeExLocks: "
							"Couldn't delete the exclusive lock."));
			rcFinal = rc;
			break;
		}
	}

	if (lockOwnerType == LOCK_OWNER_NLM)
	{
		/*Update the NLM lock record on the stable storage if necessary*/
		rc = lock_nlmClusUpdateLock(((*exLockpp)->info),
									NULL,
									LOCK_STATE_GRANTED);
		/*TODO: What to do if rc is NFX_ERR? */

		if (rc != NFX_OK)
			rcFinal = rc;
	}

	return (rcFinal);
} /* lock_mergeExLocks() */

/*-----------------------------------------------------------------
 * Name:         lock_splitNlmLock
 *
 * Description:
 *
 *     In NLM, the byte range of a granted lock can be unlocked
 *     completely or partially. This function checks which of the
 *     two situations is true. If it is unlocked partially, the
 *     function updates the existing lock range and may add a new
 *     lock if the byte range of the original lock is split into
 *     two part.
 *     
 * Arguments:    
 * 
 *     rootIdxp - Pointer to index to root of the exclusive or
 *                shared lock tree.
 *
 *     grantedLockInfoIdx - Index to lock_nlmLockInfo_t structure.
 *                          This lock's byte range will be unlocked.
 *
 *     delLockInfop - Pointer to lock_nlmLockInfo_t structure.
 *                    It contains info of unlocked byte range.
 *
 *     lockType - Type of the lock (exclusive or shared);
 *
 *     deleteFlag - This is output. "1" if the lock is to be deleted
 *                  completely. "2" if the lock needed to be removed
 *                  and inserted again.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/09/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_splitNlmLock(lock_fileLock_t		*fileLockp,
				  vs_mem_index_t		grantedLockInfoIdx,
				  lock_nlmLockInfo_t	*delLockInfop,
				  uint8					lockType,
				  uint8					*deleteFlag)
{
	int32				rc;
	lock_nlmLockInfo_t	*grantedLockInfop;
	uint64				lowerOffset, upperOffset;
	uint64				lowerLen, upperLen;

	/*------------------------------------------------------------
	 * Check if the lock range will be split, removed, or modified
	 *------------------------------------------------------------
	 */
	grantedLockInfop = (lock_nlmLockInfo_t *)
									vs_mem_idxToPtr(grantedLockInfoIdx);
	if (grantedLockInfop == NULL)
		return (NFX_ERR);

	lock_checkUnlockedRange(grantedLockInfop,
							delLockInfop,
							&lowerOffset,
							&lowerLen,
							&upperOffset,
							&upperLen);

	if ((lowerLen != 0) && (upperLen != 0))
	{
		lock_nlmLockInfo_t	newNlmLockInfo;
		vs_mem_index_t		dummyLockInfoIdx;

		/*------------------------
		 * Split the lock into two
		 *------------------------
		 */
		*deleteFlag = LOCK_SPLIT_LOCK_FLAG_KEEP;

		/* Check if the locks already reached
		 * the per filer limit (or per file limit).
		 */
		if (lock_globalCfgp->lockCount == LOCK_MAX_LOCKS)
		{
			/* Reached the limit */
			LOCK_PRINTF(LOG_ERR, ("lock_splitNlmLock: "
									"Couldn't split the lock. "
									"Locks reached per filer limit."));
			return (NFX_ERR);
		}

		/* Modify the range of the existing lock */
		grantedLockInfop->offset = lowerOffset;
		grantedLockInfop->len = lowerLen;

		/* Add a new lock */
		newNlmLockInfo.owner = grantedLockInfop->owner;
		newNlmLockInfo.offset = upperOffset;
		newNlmLockInfo.len = upperLen;
		newNlmLockInfo.idxInRecFile = 0;
		newNlmLockInfo.fcb = grantedLockInfop->fcb;

		VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);
		rc = lock_addGrantedLock(fileLockp,
								 dummyLockInfoIdx,
								 &newNlmLockInfo,
								 LOCK_OWNER_NLM,
								 lockType);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_splitNlmLock: "
											"Failed to add a new lock."));
			return (NFX_ERR);
		}
	}
	else if ((lowerLen == 0) && (upperLen == 0))
	{
		/*----------------
		 * Delete the lock
		 *----------------
		 */
		/*Only set the flag. The caller will delete the lock*/
		*deleteFlag = LOCK_SPLIT_LOCK_FLAG_DELETE;
	}
	else
	{
		/*----------------------
		 * Modify the lock range
		 *----------------------
		 */
		*deleteFlag = LOCK_SPLIT_LOCK_FLAG_KEEP;

		if (upperLen == 0)
		{
			grantedLockInfop->offset = lowerOffset;
			grantedLockInfop->len = lowerLen;
		}
		else
		{
			grantedLockInfop->offset = upperOffset;
			grantedLockInfop->len = upperLen;

			if (lockType == LOCK_TYPE_SHAR)
				*deleteFlag = LOCK_SPLIT_LOCK_FLAG_DELETE_ADD;
		}
	}
	return (NFX_OK);
} /* lock_splitNlmLock() */

/*-----------------------------------------------------------------
 * Name:         lock_mergeShNodes
 *
 * Description:
 *
 *     Find all the nodes in the shared lock tree that
 *     overlap with a given node. Merge these nodes into one.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     shNodeIdx - Index to the lock_shNode_t data structure. 
 *                 We will check if the lock range of this node 
 *                 overlaps with any other nodes in the shared lock
 *                 tree. If yes, merge other nodes with this node.
 *
 *     shNodep - Pointer to the above lock_shNode_t data structure.
 *
 * Return value:
 *
 *     Pointer to the merged node.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
lock_shNode_t *
lock_mergeShNodes(lock_fileLock_t	*fileLockp,
				  lock_shNode_t		*shNodep)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask = LOCK_SEARCH_DIFF_RANGE;
   												/*Look for overlapping nodes
								 				 *except the input node itself*/
	lock_shNode_t	*overlapShNodep, *tempShNodep;
	uint64			shNodeOffset, overlapShNodeOffset;
	uint64			shNodeLen, overlapShNodeLen;
	lock_shLock_t	*tailShLockp;

	if (shNodep == NULL)
		return (NULL);

	rootIdxp = &(fileLockp->sharedLockTree);

	while (1)
	{
		overlapShNodep = lock_findSharedNodeByNode(rootIdxp,
												   shNodep,
												   searchMask);
		if (!overlapShNodep) 
		{
			/* No overlapping node found. Return. */
			return (shNodep);
		}

		/* Found an overlapping node.
		 * Note: the offset of lock range of any overlapping nodes
		 *       should be greater than that of input node.
		 */

		/*------------------------------------------
		 * Merge the found node with the input node.
		 * Result is stored in the input node.
		 *------------------------------------------
		 */
		/* Get the offsets of the two overlapping nodes */
		rc = lock_getLockInformationFromShNode(shNodep,
											   &shNodeOffset,
											   &shNodeLen);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_mergeShNodes: "
									"Failed to get shNodeOffset."));
			return (shNodep);
		}

		rc = lock_getLockInformationFromShNode(overlapShNodep,
											   &overlapShNodeOffset,
											   &overlapShNodeLen);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_mergeShNodes: "
									"Failed to get overlapShNodeOffset."));
			return (shNodep);
		}

		if (shNodeOffset > overlapShNodeOffset)
		{
			/* Let shNode be the node with smaller offset */
			tempShNodep = shNodep;
			shNodep = overlapShNodep;
			overlapShNodep = tempShNodep;
		}

		/* Get shLock at the end of list of the input node */
		tailShLockp = (lock_shLock_t *)vs_mem_idxToPtr(shNodep->lockListTail);
		if (tailShLockp == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_mergeShNodes: "
								"Failed to find shLock data structure."));
			return (shNodep);
		}

		/* Concatenate two linked lists */
		tailShLockp->link = overlapShNodep->lockList;

		/* Update shNode's fields */
		shNodep->lockListTail = overlapShNodep->lockListTail;

		shNodep->lockCntAndHoley.mandatoryLockCnt +=
			   	overlapShNodep->lockCntAndHoley.mandatoryLockCnt;

		shNodep->lockCntAndHoley.advisoryLockCnt +=
			   	overlapShNodep->lockCntAndHoley.advisoryLockCnt;

		if (overlapShNodep->lockCntAndHoley.holey)
			shNodep->lockCntAndHoley.holey = 1;

		if (shNodep->end < overlapShNodep->end)
			shNodep->end = overlapShNodep->end;

		/*------------------------------------
		 * Remove the found node from the tree
		 *------------------------------------
		 */
		rc = lock_delBtNode(fileLockp,
							(lock_btNode_t *)overlapShNodep,
							LOCK_TYPE_SHAR,
						   	TRUE,
							FALSE);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_mergeShNodes: "
					"Failed to remove the overlapping shNode data structure."));
		}
	} /*while(1)*/

	return (shNodep);
} /* lock_mergeShNodes() */

/*-----------------------------------------------------------------
 * Name:         lock_splitSharedNode
 *
 * Description:
 *
 *     Split a "holey" shared node.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     shNodeIdx - Index to a lock_shNode_t data structure, whose
 *                 lock range has hole(s). This node is to be split.
 *
 *     shNodep - Pointer to a lock_shNode_t data structure, whose
 *               lock range has hole(s). This node is to be split.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_splitSharedNode(lock_fileLock_t	*fileLockp,
					 vs_mem_index_t		shNodeIdx,
					 lock_shNode_t		*shNodep)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	uint32			mandatoryCnt, advisoryCnt;
	boolean			nodeSplit = FALSE;
	boolean			stillHoley = FALSE;

	boolean			firstShLock;
	vs_mem_index_t	shLockIdx, prevShLockIdx;
	lock_shLock_t	*shLockp, *prevShLockp;
	uint8			prevShLockOwnerType, shLockOwnerType;
	uint64			prevShLockInfoOffset, shLockInfoOffset;
	uint64			prevShLockInfoLen, shLockInfoLen;
	uint64			shLockInfoEnd, shNodeEnd;
	vs_mem_index_t	newShNodeIdx;
	lock_shNode_t	*newShNodep;


	if (shNodep == NULL)
		return (NFX_ERR);

	rootIdxp = &(fileLockp->sharedLockTree);

	/*-------------------------------------
	 * Remove the shared node from the tree
	 *-------------------------------------
	 */
	rc = lock_delBtNode(fileLockp,
				   		(lock_btNode_t *)shNodep,
						LOCK_TYPE_SHAR,
					   	FALSE,
						FALSE);
	if (rc != NFX_OK)
		return (NFX_ERR);

	/*---------------
	 * Split the node
	 *---------------
	 */
	while (1)
	{
		mandatoryCnt = 0;
		advisoryCnt = 0;
		shNodeEnd = 0;
		firstShLock = TRUE;
		prevShLockp = NULL;

		/*--------------------------------
		 * Go through the shared lock list 
		 *--------------------------------
		 */
		shLockIdx = shNodep->lockList;
		while (!VS_MEM_IDX_IS_NULL(shLockIdx))
		{
			/* Get current shLock structure in the list */
			shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
			if (shLockp == NULL)
			{
				goto lock_splitSharedNode_outOfLoop;
				rc = NFX_ERR;
			}

			/* Get shLock offset, length, and end */
			rc = lock_getLockInformationFromLockInfo(shLockp->info,
													 NULL,
													 &shLockOwnerType,
													 &shLockInfoOffset,
													 &shLockInfoLen);
			if (rc != NFX_OK)
			{
				goto lock_splitSharedNode_outOfLoop;
				rc = NFX_ERR;
			}
			LOCK_GET_LOCK_RANGE_END(shLockInfoOffset, shLockInfoLen, shLockInfoEnd);

			if (firstShLock == TRUE)
			{
				/*-----------------------------------
				 * Got the first lock in the lockList
				 *-----------------------------------
				 */
					
				firstShLock = FALSE;
				goto lock_splitSharedNode_nextLockInList;
			}
	
			/*-------------------------------------
			 * Compare prevShLock and shLock to see
			 * if there is a hole between them.
			 *------------------------------------- 
			 */
			{
				vs_mem_index_t	dummyLockInfoIdx;
				uint8			comp;

				/* Get prevShLock offset and length */
				rc = lock_getLockInformationFromLockInfo(prevShLockp->info,
														 NULL,
														 &prevShLockOwnerType,
														 &prevShLockInfoOffset,
														 &prevShLockInfoLen);
				if (rc != NFX_OK)
				{
					goto lock_splitSharedNode_outOfLoop;
					rc = NFX_ERR;
				}

				/* Compare two lock ranges */
				VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);
				comp = lock_lockInfoComp(dummyLockInfoIdx,
										 NULL,
										 dummyLockInfoIdx,
										 prevShLockOwnerType,
										 shLockOwnerType,
										 prevShLockInfoOffset,
										 shLockInfoOffset,
										 prevShLockInfoLen,
										 shLockInfoLen,
										 0,
										 LOCK_SPECIAL_COMP_RESULT);

				if (comp == 0)
				{
					/*----------------------------------
					 * Two lock ranges overlap - no hole
					 *----------------------------------
					 */
					goto lock_splitSharedNode_nextLockInList;
				}
			}

			/*-------------------------------------------------
			 * Two lock ranges do not overlap - there is a hole
			 * Try to split the node into two.
			 *-------------------------------------------------
			 */

			/* Get a new shNode */
			newShNodep = lock_memAllocShNode(shLockIdx,
											 shLockp,
											 &newShNodeIdx);
			if (newShNodep == NULL)
			{
				/*---------------------------------------
				 * Attempt to split a "holey" node failed
				 *---------------------------------------
				 */
				if (nodeSplit == FALSE)
				{
					/* Node is not split yet. Just exit the loop and
					 * put the original shNode back to tree */
					break;
				}

				stillHoley = TRUE;

				/*To finish corrent count of mandatory and advisory locks*/
				goto lock_splitSharedNode_nextLockInList;
			}

			/*--------------------------------
			 * Split the list into two nodes.
			 * Put the first node back to tree
			 *--------------------------------
			 */

			/* First node */
			VS_MEM_IDX_SET_NULL(prevShLockp->link);
			shNodep->end = shNodeEnd;
			shNodep->lockListTail = prevShLockIdx;	
			shNodep->lockCntAndHoley.mandatoryLockCnt = mandatoryCnt;
			shNodep->lockCntAndHoley.advisoryLockCnt = advisoryCnt;
			shNodep->lockCntAndHoley.holey = 0;

			rc = lock_btree_insert(rootIdxp,
								   shNodeIdx,
								   (lock_btNode_t *)shNodep,
								   0,
								   LOCK_TYPE_SHAR);
			if (rc != NFX_OK)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_lock_splitSharedNode: "
						"Failed to insert the shNode in shared lock tree."));

			}

			/* Second node */
			shNodeIdx = newShNodeIdx;
			shNodep = newShNodep;

			nodeSplit = TRUE;

			break;

lock_splitSharedNode_nextLockInList:
			/* Increment mandatory and advisory count */
			LOCK_INC_MAND_ADVI_COUNT(shLockp->info,
									 mandatoryCnt,
									 advisoryCnt);
			/* Update shNode "end" field */
			if (shNodeEnd < shLockInfoEnd)
				shNodeEnd = shLockInfoEnd;

			prevShLockIdx = shLockIdx;
			prevShLockp = shLockp;
			shLockIdx = shLockp->link;

		} /*while(!VS_MEM_IDX_IS_NULL(shLockIdx)) loop*/

		if (VS_MEM_IDX_IS_NULL(shLockIdx) || (nodeSplit == FALSE))
		{
			/* Already gone to the end of the list, or split failed*/
			break;
		}
	} /*while(1) loop*/

lock_splitSharedNode_outOfLoop:
	if (nodeSplit == TRUE)
	{
		/* The original node was split. Fill fields of new shNode. */
		shNodep->end = shNodeEnd;
		shNodep->lockListTail = prevShLockIdx;	
		shNodep->lockCntAndHoley.mandatoryLockCnt = mandatoryCnt;
		shNodep->lockCntAndHoley.advisoryLockCnt = advisoryCnt;
	}

	if (stillHoley == TRUE)
		shNodep->lockCntAndHoley.holey = 1;
	else
	{
		shNodep->lockCntAndHoley.holey = 0;

		/*----------------------
		 * Update global counter
		 *----------------------
		 */
		lock_globalCfgp->debugCounters.holeyNodes--;
	}

	/*------------------------
	 * Insert the node to tree
	 *------------------------
	 */
	rc = lock_btree_insert(rootIdxp,
						   shNodeIdx,
						   (lock_btNode_t *)shNodep,
						   0,
						   LOCK_TYPE_SHAR);
	return (rc);
} /* lock_splitSharedNode() */

/*-----------------------------------------------------------------
 * Name:         lock_addSharedLockToNode
 *
 * Description:
 *
 *     Add a lock to the shared lock tree.
 *     The shared lock list in a node is sorted by lock offset.
 *     If two locks have the same offset, one lock with smaller byte
 *     range comes first in the list.
 *
 * Arguments:    
 * 
 *     shNodep - Pointer to a lock_shNode_t data structure.
 *               The shared lock will be added to this node.
 *
 *     newShLockIdx - Index to the lock_shLock_t data structure
 *                    which is to be added to the node.
 *
 *     newShLockp - Pointer to the above lock_shLock_t data structure.
 *
 *     newLockInfoIdx - Index to the lock_cifsLockInfo_t or
 *                      lock_nlmLockInfo_t data structure which
 *                      is attached to the lock_shLock_t.
 *
 *     newLockInfop - Pointer to the above lockInfo data structure.
 *
 *     newLockOwnerType - Owner type of the lock to be added.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_addSharedLockToNode(lock_shNode_t	*shNodep,
						 vs_mem_index_t	newShLockIdx,
						 lock_shLock_t	*newShLockp,
						 vs_mem_index_t	newLockInfoIdx,
						 void			*newLockInfop,
						 uint8			newLockOwnerType)
{
	int32			rc;
	vs_mem_index_t	shLockIdx;
	lock_shLock_t	*shLockp, *prevShLockp = NULL;
	uint8			lockOwnerType;
	uint64			lockOffset, newLockOffset;
	uint64			lockLen, newLockLen;
	uint64			lockEnd, newLockEnd;
	boolean			done = FALSE;

	if ((shNodep == NULL) || (newShLockp == NULL))
		return (NFX_ERR);

	rc = lock_getLockInformationFromLockInfo(newLockInfoIdx,
											 newLockInfop,
											 &newLockOwnerType,
											 &newLockOffset,
											 &newLockLen);
	if (rc != NFX_OK)
		return (rc);
	LOCK_GET_LOCK_RANGE_END(newLockOffset, newLockLen, newLockEnd);

	shLockIdx = shNodep->lockList;
	while(!VS_MEM_IDX_IS_NULL(shLockIdx))
	{
		/* Get current shLock structure in the list*/
		shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
		if (shLockp == NULL)
			return (NFX_ERR);

		rc = lock_getLockInformationFromLockInfo(shLockp->info,
						                         NULL,
												 &lockOwnerType,
												 &lockOffset,
												 &lockLen);
		if (rc != NFX_OK)
			return (rc);
		LOCK_GET_LOCK_RANGE_END(lockOffset, lockLen, lockEnd);

		if ((newLockOffset < lockOffset) ||
			((newLockOffset == lockOffset) && (newLockEnd < lockEnd)))
		{
			/* Insert the new lock here before the current lock */
			if (prevShLockp == NULL)
			{
				/* Insert at the beginning of the list */
				newShLockp->link = shNodep->lockList;
				shNodep->lockList = newShLockIdx;
			}
			else
			{
				/* Insert in the middle of the list */
				prevShLockp->link = newShLockIdx;
				newShLockp->link = shLockIdx;
			}
			done = TRUE;
			break;
		}

		/* Go to next shLock in the list */
		prevShLockp = shLockp;
		shLockIdx = shLockp->link;
	}

	if (done == FALSE)
	{
		if (VS_MEM_IDX_IS_NULL(shNodep->lockList))
		{
			/* The list is empty. The lock becomes the only one in the list. */
			shNodep->lockList = newShLockIdx;
		}
		else
		{
			/* Append at the end of the list */
			shLockp->link = newShLockIdx;
		}
		shNodep->lockListTail = newShLockIdx;
	}
	
	/*Update "end" field of shNode*/
	if (shNodep->end < newLockEnd)
		shNodep->end = newLockEnd;

	/* Update mandatory/advisory lock count */
	LOCK_SHNODE_INC_COUNT(shNodep, newLockOwnerType);

	return (NFX_OK);
} /* lock_addSharedLockToNode() */

/*-----------------------------------------------------------------
 * Name:         lock_delSharedLockFromNode
 *
 * Description:
 *
 *     Delete a lock from the shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     shNodepp - The value is pointer to the shared node.
 *                A shared lock is to be deleted from this node.
 *                If the node no longer has any locks associated with
 *                it, the node will be removed.
 *
 *     prevShLockIdx - Index to a lock_shLock_t data structure.
 *                     This lock is in the linked list of the shared
 *                     node. Its next lock is the lock to be deleted.
 *
 *     shLockp - Pointer to the shared lock to be deleted.
 *
 *     lockOwnerType - Owner type of the lock to be deleted.
 *
 *     deleteLock - FALSE if just remove the lock from the node but not
 *                  to delete the lock structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/17/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delSharedLockFromNode(lock_fileLock_t	*fileLockp,
						   lock_shNode_t	**shNodepp,
						   vs_mem_index_t	prevShLockIdx,
						   lock_shLock_t	*shLockp,
						   uint8			lockOwnerType,
						   boolean			deleteLock)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	lock_shLock_t	*prevShLockp;
	vs_mem_index_t	shLockIdx, nextShLockIdx;

	rootIdxp = &(fileLockp->sharedLockTree);

	if ((*shNodepp == NULL) || (shLockp == NULL))
		return (NFX_ERR);

	if ((((*shNodepp)->lockCntAndHoley.mandatoryLockCnt +
		  (*shNodepp)->lockCntAndHoley.advisoryLockCnt) == 1) &&
		(deleteLock == TRUE))
	{
		/*----------------------------------------------------
		 * The lock to be deleted is the only one in the list.
		 * Delete the shared node
		 *----------------------------------------------------
		 */
		rc = lock_delBtNode(fileLockp,
							(lock_btNode_t *)(*shNodepp),
							LOCK_TYPE_SHAR,
							TRUE,
							TRUE);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_delSharedLockFromNode: "
							"Failed to free the shNode structure."));
			return (NFX_ERR);
		}
		else
		{
			*shNodepp = NULL;
			return (NFX_OK);
		}
	}

	/*-----------------------------------------
	 * Remove the lock from the linked list and
	 * delete the lock if deleteLock is TRUE
	 *-----------------------------------------
	 */
	nextShLockIdx = shLockp->link;

	prevShLockp = (lock_shLock_t *)vs_mem_idxToPtr(prevShLockIdx);
	if (prevShLockp == NULL)
	{
		/* The lock to be deleted is the head in the list */
		shLockIdx = (*shNodepp)->lockList;
	}
	else
		shLockIdx = prevShLockp->link;

	if (deleteLock == TRUE)
	{
		/*
		 * Delete the shLock
		 */
		rc = lock_memFreeShLock(shLockIdx, shLockp);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_delSharedLockFromNode: "
								"Failed to free the shLock structure."));
			return (NFX_ERR);
		}
	}
	else
	{
		VS_MEM_IDX_SET_NULL(shLockp->link);
	}

	/*
	 * Update the lock list
	 */
	if (prevShLockp == NULL)
	{
		/* Remove the first lock in the list */
		(*shNodepp)->lockList = nextShLockIdx;
	}
	else if (VS_MEM_IDX_IS_NULL(nextShLockIdx))
	{
		/* Remove the last lock in the list */
		(*shNodepp)->lockListTail = prevShLockIdx;
		VS_MEM_IDX_SET_NULL(prevShLockp->link);
	}
	else
	{
		/* Remove a lock in the middle of the list */
		prevShLockp->link = nextShLockIdx;
	}

	/*--------------------------------
	 * Update mandatory/advisory count
	 *--------------------------------
	 */
	LOCK_SHNODE_DEC_COUNT(*shNodepp, lockOwnerType);

	if (deleteLock == TRUE)
	{
		/*-----------------------
		 * Decrement nlmLockCount
		 *-----------------------
		 */
		if (lockOwnerType == LOCK_OWNER_NLM)
			lock_decNlmLockCntOfFile(fileLockp);

		/*----------------------------
		 * Update global lock counters
		 *----------------------------
		 */
		 LOCK_DEC_LOCK_COUNTERS(
						lockOwnerType,
						lock_globalCfgp->lockCounters.grantedSharedCifsLocks,
						lock_globalCfgp->lockCounters.grantedSharedNlmLocks);
	}

	return (NFX_OK);
} /* lock_delSharedLockFromNode() */

/*-----------------------------------------------------------------
 * Name:         lock_delBtNode
 *
 * Description:
 *
 *     Delete a node from the exclusive or shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     delNodep - Pointer to the tree node to be deleted.
 *
 *     treeType - Type of the tree (exclusive or shared).
 *
 *     freeNode - TRUE if the memory allocated for the deleted is
 *                     to be freed;
 *                FALSE if the memory is not to be freed.
 *
 *     freeShLock - This is meaningful for a shared node. 
 *                  TRUE if the associated shared locks are also to be freed.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delBtNode(lock_fileLock_t	*fileLockp,
			   lock_btNode_t	*delNodep,
			   uint8			treeType,
			   boolean			freeNode,
			   boolean			freeShLock)
{
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask;
	vs_mem_index_t	delNodeIdx;

	switch (treeType)
	{
		case LOCK_TYPE_EXCL:
			rootIdxp = &(fileLockp->exclLockTree);
			searchMask = (LOCK_SEARCH_SAME_RANGE | LOCK_SEARCH_SAME_OWNER);
			break;

		case LOCK_TYPE_SHAR:
			rootIdxp = &(fileLockp->sharedLockTree);
			searchMask = LOCK_SEARCH_SAME_RANGE;
			break;

		default:
			return (NFX_ERR);
	}

	/*------------------------------
	 * Delete the node from the tree
	 *------------------------------ 
	 */
	delNodep = lock_btree_delete(rootIdxp,
								 delNodep,
								 searchMask,
								 treeType,
								 &delNodeIdx);
	if (delNodep == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_delBtNode: "
								"Failed to delete the node from tree"));
		return (NFX_ERR);
	}
	else if (freeNode == TRUE)
	{
		/*-------------------------
		 * Free the data structures
		 *-------------------------
		 */
		int32	rc;

		switch (treeType)
		{
			case LOCK_TYPE_EXCL:
			{
				vs_mem_index_t	lockInfoIdx;
				uint8			lockOwnerType;

				/* Get lock owner type */
				lockInfoIdx = lock_getLockInfoIdxByLockIdx(delNodeIdx);
				lockOwnerType = lock_getOwnerTypeByLockInfoIdx(lockInfoIdx);

				rc = lock_memFreeExLock(delNodeIdx, (lock_exLock_t *)delNodep);
				if (rc == NFX_OK)
				{
					/*-----------------------
					 * Decrement nlmLockCount
					 *-----------------------
					 */
					if (lockOwnerType == LOCK_OWNER_NLM)
						lock_decNlmLockCntOfFile(fileLockp);

					/*----------------------------
					 * Update global lock counters
					 *----------------------------
					 */
					LOCK_DEC_LOCK_COUNTERS(
							lockOwnerType,
							lock_globalCfgp->lockCounters.grantedExclCifsLocks,
							lock_globalCfgp->lockCounters.grantedExclNlmLocks);
				}
			}
				break;

			case LOCK_TYPE_SHAR:
				rc = lock_memFreeShNode(fileLockp,
										delNodeIdx,
										(lock_shNode_t *)delNodep,
										freeShLock);
				break;

			default:
				rc = NFX_ERR;
		}
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_delBtNode: "
									"Failed to free the btNode structure"));
			return (NFX_ERR);
		}
	}

	return (NFX_OK);
} /* lock_delBtNode() */

/*-----------------------------------------------------------------
 * Name:         lock_insertPendLockToLocalList
 *
 * Description:
 *
 *     Insert the new pending lock to the file's pending lock list,
 *     sorted by request time.
 *
 * Arguments:    
 * 
 *     pendingListHeadp - Pointer to the head of pending lock list.
 *
 *     newPendLockIdx - Index to the new pending lock.
 *
 *     newPendLockp - Pointer to the new pending lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_insertPendLockToLocalList(vs_mem_index_t		*pendingListHeadp,
							   vs_mem_index_t		newPendLockIdx,
							   lock_pendingLock_t	*newPendLockp)
{
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp;

	if ((pendingListHeadp == NULL) || (newPendLockp == NULL))
		return (NFX_ERR);

	if (VS_MEM_IDX_IS_NULL(*pendingListHeadp))
	{
		/* List is empty. This lock becomes the only one in list. */
		*pendingListHeadp = newPendLockIdx;
		return (NFX_OK);
	}

	/* Append the new pending lock to the tail of the list */
	pendLockIdx = *pendingListHeadp;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		/* Get pendingLock structure */
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		pendLockIdx = pendLockp->link;
	}
	pendLockp->link = newPendLockIdx;

	return (NFX_OK);
} /* lock_insertPendLockToLocalList() */

/*-----------------------------------------------------------------
 * Name:         lock_deletePendingLocksFromLocalList
 *
 * Description:
 *
 *     Delete one or more pending locks from the pending lock list
 *     depending on different type (given as input), that determines
 *     which locks are to be deleted and what to do with the deleted locks.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     param - Information needed for deleting locks.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 *     LOCK_ERR_NOT_FOUND - If the lock to be deleted is not found.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/20/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_deletePendingLocksFromLocalList(
				lock_fileLock_t				*fileLockp,
				lock_delPendLocksParam_t	param)
{
	int32				rc, rcFinal = NFX_OK;
	uint8				type;
	vs_mem_index_t		pendLockIdx, newLockListHead;
	lock_pendingLock_t	*pendLockp, *prevPendLockp = NULL;
	void				*lockInfop;
	int8				lockOwnerType;
	boolean				deleteLock, freeLockInfo;
	boolean				singleLockDeleted = FALSE;
	vs_mem_index_t      dummyIdx;

	if (fileLockp == NULL)
		return (NFX_ERR);

	type = param.type;

	if (type == LOCK_DELPENDLOCKS_TYPE_GRANTPEND)
	{
		/*Do not free lockInfo structure, because the
		  lock is to be added to granted lock tree */
		freeLockInfo = FALSE;
	}
	else
	{
		/*Free lockInfo structure */
		freeLockInfo = TRUE;
	}

	VS_MEM_IDX_SET_NULL(dummyIdx);

	/*Init the newLockListHead to be the original pending lock list head*/
	pendLockIdx = fileLockp->pendingLockList;
	newLockListHead = pendLockIdx;

	/*---------------------------------
	 * Go through the pending lock list
	 *---------------------------------
	 */
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		/* Get pending lock structure */
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		/* Get pending lockInfo structure */
		lockInfop = lock_getLockInfo(pendLockp->info, &lockOwnerType);
		if (lockInfop == NULL)
			return (NFX_ERR);

		/*----------------------------------------------
		 * Do different checking based on different type
		 *----------------------------------------------
		 */
		switch (type)
		{
			case LOCK_DELPENDLOCKS_TYPE_UNLOCKALL:
			{
				/* Check if current pending lock matches the key or
				 * clientIpAddr */
				boolean				matchResult;

				matchResult = lock_match(pendLockp->info,
										 dummyIdx,
										 TRUE,	/*Use lockInfo*/
										 param.u.delAll.key,
										 param.u.delAll.clientIpAddr,
										 param.u.delAll.searchMask);
				deleteLock = matchResult;
			}
				break;

			case LOCK_DELPENDLOCKS_TYPE_UNLOCKSINGLE:
			{
				int8			comp;

				comp = lock_twoLockInfoCompFunc(dummyIdx,
												param.u.delSingle.lockInfop,
												param.u.delSingle.lockOwnerType,
												pendLockp->info,
												param.u.delSingle.searchMask,
												LOCK_SPECIAL_COMP_RESULT);
				if (comp == 0)
					deleteLock = TRUE;
				else
					deleteLock = FALSE;
			}
				break;

			case LOCK_DELPENDLOCKS_TYPE_GRANTPEND:
			{
				int32			conflictCheckResult, exclResult, sharResult;
				vs_mem_index_t	dummyLockIdx;

				conflictCheckResult = lock_checkConflict(fileLockp,
														 lockInfop,
														 lockOwnerType,
														 pendLockp->type,
														 TRUE,
														 &dummyLockIdx,
														 &exclResult,
														 &sharResult);
				if (conflictCheckResult == LOCK_STATE_GRANTED)
					deleteLock = TRUE;
				else
					deleteLock = FALSE;
			}
				break;

			case LOCK_DELPENDLOCKS_TYPE_TIMEOUT:
			{
				if (VS_MEM_IDX_IS_EQUAL(pendLockIdx, param.u.pendLockIdx))
					deleteLock = TRUE;
				else
					deleteLock = FALSE;
			}
				break;

			default:
				deleteLock = FALSE;
		} /*switch()*/

		if (deleteLock == TRUE)
		{
            vs_mem_index_t  nextPendLockIdx;

            nextPendLockIdx = pendLockp->link;

			/*----------------------------------------------------------
			 * Remove the pending lock from the file's pending lock list
			 *----------------------------------------------------------
			 */
			if (VS_MEM_IDX_IS_EQUAL(pendLockIdx, newLockListHead))
			{
				/* Remove the head of the list */
				newLockListHead = pendLockp->link;
                fileLockp->pendingLockList = newLockListHead;
			}
			else
			{
				/* Remove from middle of the list */
				prevPendLockp->link = pendLockp->link;
			}

			/*------------------------------------------------------
			 * Remove the pending lock from global pending lock list
			 *------------------------------------------------------
			 */
			if (type != LOCK_DELPENDLOCKS_TYPE_TIMEOUT)
			{
				/*If the lock is to be deleted due to timeout, the lock
				 *is already removed from the global pending list*/
				rc = lock_removePendLockFromGlobalList(pendLockIdx);
				if (rc != NFX_OK)
					return (rc);
			}

			if (type == LOCK_DELPENDLOCKS_TYPE_GRANTPEND)
			{
				/*------------------------------------------
				 * Add lock to exclusive or shared lock tree
				 *------------------------------------------
				 */
				if (lock_globalCfgp->lockCount < LOCK_MAX_LOCKS)
				{
					/* The number of total locks haven't reached the limit yet*/
					rc = lock_addGrantedLock(fileLockp,
											 pendLockp->info,
											 NULL,
											 lockOwnerType,
											 pendLockp->type);
					if (rc != NFX_OK)
					{
						freeLockInfo = TRUE;
						rcFinal = rc;
					}
					else
					{
						/*--------------------------------
						 * Notify the upper protocal layer
						 *--------------------------------
						 */
						if (lockOwnerType == LOCK_OWNER_CIFS)
						{
							/*TODO*/
						}
						else if (lockOwnerType == LOCK_OWNER_NLM)
							nlm_procGrantedCallback(pendLockp->ulPrivate);
					}
				}
			}

			/*--------------------------------------
			 * Delete the pendingLock data structure
			 *--------------------------------------
			 */
			rc = lock_memFreePendingLock(pendLockIdx,
										 pendLockp,
										 freeLockInfo);
			if (rc != NFX_OK)
				rcFinal = rc;
			else
			{
				/*----------------------------
				 * Update global lock counters
				 *----------------------------
				 */
				if (pendLockp->type == LOCK_TYPE_EXCL)
				{
					LOCK_DEC_LOCK_COUNTERS(
						lockOwnerType,
						lock_globalCfgp->lockCounters.pendingExclCifsLocks,
						lock_globalCfgp->lockCounters.pendingExclNlmLocks);
				}
				else if (pendLockp->type == LOCK_TYPE_SHAR)
				{
					LOCK_DEC_LOCK_COUNTERS(
						lockOwnerType,
						lock_globalCfgp->lockCounters.pendingSharedCifsLocks,
						lock_globalCfgp->lockCounters.pendingSharedNlmLocks);
				}
			}

			if (type == LOCK_DELPENDLOCKS_TYPE_UNLOCKSINGLE)
			{
				singleLockDeleted = TRUE;
				break;
			}

			if (type == LOCK_DELPENDLOCKS_TYPE_TIMEOUT)
				break;

            pendLockIdx = nextPendLockIdx;
		} /* if (deleteLock == TRUE) */
        else
        {
		    pendLockIdx = pendLockp->link;
		    prevPendLockp = pendLockp;
        }
	} /*while (!VS_MEM_IDX_IS_NULL(pendLockIdx))*/

	if ((type == LOCK_DELPENDLOCKS_TYPE_UNLOCKSINGLE) &&
		(singleLockDeleted == FALSE))
	{
		LOCK_PRINTF(LOG_INFO, ("lock_deletePendingLocksFromLocalList: "
						"The lock to be canceled is not found."));
		return (LOCK_ERR_NOT_FOUND);
	}

	/*----------------------------------------------------------
	 * After all the deleted pending locks are removed from
	 * the global pending lock list, restart timer if necessary.
	 *----------------------------------------------------------
	 */
	if (type != LOCK_DELPENDLOCKS_TYPE_TIMEOUT)
	{
		rc = lock_checkPendingLockTimer();
		if (rc != NFX_OK)
			rcFinal = rc;
	}

	return (rcFinal);
} /* lock_deletePendingLocksFromLocalList() */

/*-----------------------------------------------------------------
 * Name:         lock_insertPendLockToGlobalList
 *
 * Description:
 *
 *     Insert the new pending lock to the global pending lock list,
 *     sorted by ascending timeout.
 *
 * Arguments:    
 * 
 *     newPendLockIdx - Index to the new pending lock.
 *
 *     newPendLockp - Pointer to the new pending lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_insertPendLockToGlobalList(vs_mem_index_t		newPendLockIdx,
								lock_pendingLock_t	*newPendLockp)
{
	int32				rc;
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp, *prevPendLockp;
	boolean				done = FALSE;

	if (newPendLockp == NULL)
		return (NFX_ERR);

	if (VS_MEM_IDX_IS_NULL(lock_globalCfgp->pendingLockList))
	{
		/*---------------------------------------
		 * List is empty.
		 * This lock become the only one in list.
		 *---------------------------------------
		 */
		lock_globalCfgp->pendingLockList = newPendLockIdx;
	}
	else
	{
		/*-----------------------------------------
		 * List is not empty.
		 * Insert the new pending lock to the list.
		 *-----------------------------------------
		 */
		prevPendLockp = NULL;
		pendLockIdx = lock_globalCfgp->pendingLockList;
		while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
		{
			/* Get pendingLock structure */
			pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
			if (pendLockp == NULL)
				return (NFX_ERR);

			if (eee_rtcCompare(&newPendLockp->timeout,
							   &pendLockp->timeout) < 0)
			{
				if (prevPendLockp == NULL)
				{
					/* The new pending lock becomes head of the list */
					newPendLockp->timeoutLink = lock_globalCfgp->pendingLockList;
					lock_globalCfgp->pendingLockList = newPendLockIdx;
					done = TRUE;
					break;
				}
				else
				{
					/* Insert the new pending lock in front of the current one */
					newPendLockp->timeoutLink = pendLockIdx;
					prevPendLockp->timeoutLink = newPendLockIdx;
					done = TRUE;
					break;
				}
			}

			/* Go to next one in the list */
			prevPendLockp = pendLockp;
			pendLockIdx = pendLockp->timeoutLink;
		} /* while (!VS_MEM_IDX_IS_NULL(pendLockIdx)) */

		if (done == FALSE)
		{
			/* Append the new pending lock to the end of the list */
			pendLockp->timeoutLink = newPendLockIdx;
		}
	}

	/*-----------------------------------------------------------------
	 * Check to see if the global timeout timer needs to restart or not
	 *-----------------------------------------------------------------
	 */
	rc = lock_checkPendingLockTimer();
	
	return (rc);
} /* lock_insertPendLockToGlobalList() */

/*-----------------------------------------------------------------
 * Name:         lock_removePendLockFromGlobalList
 *
 * Description:
 *
 *     Remove a pending lock from the global pending lock list.
 *
 * Arguments:    
 * 
 *     delPendLockIdx - Index to the pending lock to be removed.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/25/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_removePendLockFromGlobalList(vs_mem_index_t	delPendLockIdx)
{
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp, *prevPendLockp = NULL;

	pendLockIdx = lock_globalCfgp->pendingLockList;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		/* Get pendingLock structure */
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		if (VS_MEM_IDX_IS_EQUAL(pendLockIdx, delPendLockIdx))
		{
			/* Found the pending lock. Remove it from the list */
			if (prevPendLockp == NULL)
			{
				/* The lock is the head in the list */
				lock_globalCfgp->pendingLockList = pendLockp->timeoutLink;
			}
			else
			{
				/* The lock is not the head in the list */
				prevPendLockp->timeoutLink = pendLockp->timeoutLink;
			}

			break;
		}

		/* Go to next one in the list */
		prevPendLockp = pendLockp;
		pendLockIdx = pendLockp->timeoutLink;
	}

	return (NFX_OK);
} /* lock_removePendLockFromGlobalList() */

/*-----------------------------------------------------------------
 * Name:         lock_deletePendingList
 *
 * Description:
 *
 *     Delete a pending lock list.
 *
 * Arguments:    
 * 
 *     pendingListp - Pointer to the head of pending lock list.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_deletePendingList(vs_mem_index_t *pendingListp)
{
	int32				rc, rcFinal;
	vs_mem_index_t		pendLockIdx, nextPendLockIdx;
	lock_pendingLock_t	*pendLockp;
	uint8				lockOwnerType;

	if (pendingListp == NULL)
		return (NFX_ERR);

	/*-----------------------------
	 * Delete each lock in the list
	 *-----------------------------
	 */
	pendLockIdx = *pendingListp;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

	 	/*---------------------------------------
		 * Modify the global timeout lock list to
		 * remove the links to the deleted locks
		 *---------------------------------------
		 */
		rc = lock_removePendLockFromGlobalList(pendLockIdx);
		if (rc != NFX_OK)
			rcFinal = rc;

		nextPendLockIdx = pendLockp->link;

		/* Get lock owner type */
		lockOwnerType = lock_getOwnerTypeByLockInfoIdx(pendLockp->info);

		rc = lock_memFreePendingLock(pendLockIdx, pendLockp, TRUE);
		if (rc != NFX_OK)
			rcFinal = rc;
		else
		{
			/*----------------------------
			 * Update global lock counters
			 *----------------------------
			 */
			if (pendLockp->type == LOCK_TYPE_EXCL)
			{
				LOCK_DEC_LOCK_COUNTERS(
						lockOwnerType,
						lock_globalCfgp->lockCounters.pendingExclCifsLocks,
						lock_globalCfgp->lockCounters.pendingExclNlmLocks);
			}
			else if (pendLockp->type == LOCK_TYPE_SHAR)
			{
				LOCK_DEC_LOCK_COUNTERS(
						lockOwnerType,
						lock_globalCfgp->lockCounters.pendingSharedCifsLocks,
						lock_globalCfgp->lockCounters.pendingSharedNlmLocks);
			}
		}
	
		pendLockIdx = nextPendLockIdx;
	}
	VS_MEM_IDX_SET_NULL(*pendingListp);

	/*----------------------------------------------------------
	 * After all the deleted pending locks are removed from
	 * the global pending lock list, restart timer if necessary.
	 *----------------------------------------------------------
	 */
	rc = lock_checkPendingLockTimer();
	if (rc != NFX_OK)
		rcFinal = rc;

	return (rc);
} /* lock_deletePendingList() */

/*-----------------------------------------------------------------
 * Name:         lock_updateShNodeEndAndHoley
 *
 * Description:
 *
 *     Update "end" and "holey" fields of a shared node.
 *
 * Arguments:    
 * 
 *     shNodep - Pointer to the shared lock node.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/20/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_updateShNodeEndAndHoley(lock_shNode_t	*shNodep)
{
	int32			rc;
	uint8			holey = 0;
	vs_mem_index_t	shLockIdx, dummyShLockIdx;
	lock_shLock_t	*shLockp, *dummyShLockp;
	void			*lockInfop;
	uint8			lockOwnerType;
	uint64			lockOffset, lockLen, lockEnd;
	boolean			firstShLock = TRUE;
	int8			comp;

	if (shNodep == NULL)
		return (NFX_ERR);

	shNodep->end = 0;

	/*---------------------------------------
	 * Go through the linked list of the node
	 *---------------------------------------
	 */
	shLockIdx = shNodep->lockList;
	while (!VS_MEM_IDX_IS_NULL(shLockIdx))
	{
		/* Get current shLock structure in the list */
		shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
		if (shLockp == NULL)
			return (NFX_ERR);

		/* Get lockInfo structure of the current lock */
		lockInfop = lock_getLockInfo(shLockp->info, &lockOwnerType);
		if (lockInfop == NULL)
			return (NFX_ERR);

		/* Get byte range offset, length, and end of the current lock */
		rc = lock_getLockInformationFromLockInfo(shLockp->info,
												 lockInfop,
												 &lockOwnerType,
												 &lockOffset,
												 &lockLen);
		if (rc != NFX_OK)
			return (NFX_ERR);

		LOCK_GET_LOCK_RANGE_END(lockOffset, lockLen, lockEnd);

		if (firstShLock == TRUE)
		{
			/*-----------------------------------
			 * Got the first lock in the lockList
			 *-----------------------------------
			 */
			firstShLock = FALSE;
			goto lock_updateShNodeEndAndHoley_nextLockInList;
		}

		/*-------------------------
		 * Check if there is a hole
		 *-------------------------
		 */
		if (holey == 0)
		{
			/* Check if there is a hole */
			comp = lock_sharedLockInfoAndNodeCompFunc(lockInfop,
													  lockOwnerType,
													  shNodep,
													  0,	/*searchMask*/
													  LOCK_SPECIAL_COMP_RESULT,
													  &dummyShLockp,
													  &dummyShLockIdx);
			if (comp != 0)
			{
				/* There is a hole */
				holey = 1;

				/*----------------------
				 * Update global counter
				 *----------------------
				 */
				lock_globalCfgp->debugCounters.holeyNodes++;
			}
		}

lock_updateShNodeEndAndHoley_nextLockInList:
		/*-------------------
		 * Update "end" field
		 *-------------------
		 */
		if (shNodep->end < lockEnd)
			shNodep->end = lockEnd;

		shLockIdx = shLockp->link;
	}

	shNodep->lockCntAndHoley.holey = holey;

	return (NFX_OK);
} /* lock_updateShNodeEndAndHoley() */

#if 0
/*-----------------------------------------------------------------
 * Name:         lock_updateLockListHead
 *
 * Description:
 *
 *     Update "lockListHead" field of every pending lock structure
 *     in a pending lock list, given the new head of the pending
 *     lock list.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     newLockListHead - Index to the new head.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/20/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_updateLockListHead(lock_fileLock_t	*fileLockp,
			   			vs_mem_index_t	newLockListHead)
{
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp;

	if (VS_MEM_IDX_IS_EQUAL(fileLockp->pendingLockList, newLockListHead))
	{
		/* The head is not changed, return */
		return (NFX_OK);
	}

	fileLockp->pendingLockList = newLockListHead;

	pendLockIdx = fileLockp->pendingLockList;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		pendLockp->lockListHead = newLockListHead;

		pendLockIdx = pendLockp->link;
	}

	return (NFX_OK);
} /* lock_updateLockListHead() */
#endif

/*-----------------------------------------------------------------
 * Name:         lock_grantPendingLocks
 *
 * Description:
 *
 *     Check to see if any pending locks of a file can be granted.
 *     If there is such a lock, move the lock from pending
 *     list to exclusive or shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/20/2002
 *
 *-----------------------------------------------------------------
 */
inline int32
lock_grantPendingLocks(lock_fileLock_t *fileLockp)
{
	int32						rc;
	lock_delPendLocksParam_t	param;

	param.type = LOCK_DELPENDLOCKS_TYPE_GRANTPEND;
	
	rc = lock_deletePendingLocksFromLocalList(fileLockp, param);

	return (rc);
} /* lock_grantPendingLocks() */

/*-----------------------------------------------------------------
 * Name:         lock_isMandatoryLock
 *
 * Description:
 *
 *     Check if a lock is mandatory lock.
 *     If yes, return the lock offset.
 *
 * Arguments:    
 * 
 *     lockInfoIdx - Index to a lock_cifsLockInfo_t or
 *                   lock_nlmLockInfo_t structure.
 *
 *     lockOffsetp - This is output. The value is mandatory lock offset.
 *
 * Return value:
 *
 *     TRUE - If lock is a mandatory lock.
 *
 *     FALSE - If lock is not a mandatory lock.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/29/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_isMandatoryLock(vs_mem_index_t	lockInfoIdx,
					 uint64			*lockOffsetp)
{
	int32			rc;
	void			*lockInfop;
	uint8			lockOwnerType;
	vs_mem_index_t	dummyLockInfoIdx;
	uint64			lockLen;

	lockInfop = lock_getLockInfo(lockInfoIdx, &lockOwnerType);
	if (lockInfop == NULL)
		return (FALSE);

	if (LOCK_IS_MANDATORY(lockOwnerType))
	{
		/* Get mandatory lock offset */
		VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);
		rc = lock_getLockInformationFromLockInfo(dummyLockInfoIdx,
												 lockInfop,
												 &lockOwnerType,
												 lockOffsetp,
												 &lockLen);
		if (rc != NFX_OK)
			return (FALSE);

		return (TRUE);
	}

	return (FALSE);
} /* lock_isMandatoryLock() */

/*-----------------------------------------------------------------
 * Name:         lock_fillNlmTestResultOwnerInfo
 *
 * Description:
 *
 *     Fill in the lock owner portion of the NLM lock test
 *     result structure.
 *
 * Arguments:    
 * 
 *     confOwnerIdx - Index to lock_cifsLockOwner_t or lock_nlmLockOwner_t
 *                    structure. This owner's info will be put in
 *                    the test result structure.
 *
 *     testResultp - Pointer to the test result structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_fillNlmTestResultOwnerInfo(vs_mem_index_t		confOwnerIdx,
			   					lock_testResult_t	*testResultp)
{
	int32					rc;
	uint8					confOwnerType;
	void					*confOwnerp;
	lock_cifsLockOwner_t	*cifsConfOwnerp;
	lock_nlmLockOwner_t		*nlmConfOwnerp;

	/* Get lockOwner */
	confOwnerp = lock_getLockOwnerByIdx(confOwnerIdx, &confOwnerType);
	if (confOwnerp == NULL)
		return (NFX_ERR);

	rc = NFX_OK;
	switch (confOwnerType)
	{
		case LOCK_OWNER_CIFS:
			cifsConfOwnerp = (lock_cifsLockOwner_t *)confOwnerp;

			/*TODO: Fake the NLM owner info.
			 *          caller name: CIFS client's domain name
			 *          lock owner: caller name + key
			 *          process ID: key
			 */
			break;

		case LOCK_OWNER_NLM:
			nlmConfOwnerp = (lock_nlmLockOwner_t *)confOwnerp;

			testResultp->pid = nlmConfOwnerp->pid;
			testResultp->ownerNameLen = nlmConfOwnerp->ownerNameLen;
			testResultp->ownerName = nlmConfOwnerp->ownerName;
			break;

		default:
			rc = NFX_ERR;
			break;
	}

	return (rc);
} /* lock_fillNlmTestResultOwnerInfo() */

/*-----------------------------------------------------------------
 * Name:         lock_fillNlmTestResult
 *
 * Description:
 *
 *     Fill in the test result structure. The structure contains
 *     a conflicting lock info in the response of NLM_TEST. 
 *
 * Arguments:    
 * 
 *     confResult - The result of conflict checking. The value is
 *                  LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or
 *                  LOCK_STATE_BLOCKED.
 *
 *     confLockIdx - Index to lock_exLock_t or lock_shLock_t
 *                   structure. This is the conflicting lock.
 *
 *     testResultp - Pointer to the test result structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/10/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_fillNlmTestResult(uint8				confResult,
					   vs_mem_index_t		confLockIdx,
			   		   lock_testResult_t	*testResultp)
{
	int32			rc;
	uint8			confLockType;
	vs_mem_index_t	confLockInfoIdx;
	uint8			confOwnerType;
	vs_mem_index_t	confOwnerIdx;

	if (confResult != LOCK_STATE_GRANTED)
	{
		/*--------------------------------
		 * There exists a conflicting lock
		 *--------------------------------
		 */
		testResultp->lockState = LOCK_STATE_DENIED;

		/*-----------------------------------------------
		 * Fill in the test result structure with
		 * conflicting lock info and the lock owner info.
		 *-----------------------------------------------
		 */
		/* Fill lockType */
		confLockType = lock_getLockTypeByLockIdx(confLockIdx);
		testResultp->lockType = confLockType;

		/* Fill pid, ownerNameLen, and ownerName */
		confLockInfoIdx = lock_getLockInfoIdxByLockIdx(confLockIdx);
		confOwnerIdx = lock_getLockOwnerIdxByLockInfo(confLockInfoIdx,
													  NULL,
													  confOwnerType);
		rc = lock_fillNlmTestResultOwnerInfo(confOwnerIdx, testResultp);
		if (rc != NFX_OK)
			return (NFX_ERR);

		/* Fill offset and len */
		rc = lock_getLockInformationFromLockInfo(confLockInfoIdx,
												 NULL,
												 &confOwnerType,
												 &(testResultp->offset),
												 &(testResultp->len));
	}	
	else
	{
		/*-----------------------------
		 * There is no conflicting lock
		 *-----------------------------
		 */
		testResultp->lockState = LOCK_STATE_GRANTED;
	}

	return (NFX_OK);
} /* lock_fillNlmTestResult() */

/*-----------------------------------------------------------------
 * Name:         lock_incNlmLockCntOfFile
 *
 * Description:
 *
 *     Increment the count of NLM locks on a file. If the first NLM
 *     lock on the file is granted, inform the CIFS module.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer lokc_fileLock_t structure.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/11/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_incNlmLockCntOfFile(lock_fileLock_t *fileLockp)
{
	fileLockp->nlmLockCount++;
	if (fileLockp->nlmLockCount == 1)
	{
		/*TODO: Inform the upper layer that the file has NLM locks on it*/
	}
} /* lock_incNlmLockCntOfFile */

/*-----------------------------------------------------------------
 * Name:         lock_decNlmLockCntOfFile
 *
 * Description:
 *
 *     Decrement the count of NLM locks on a file. If the last NLM
 *     lock on the file is released, inform the CIFS module.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer lokc_fileLock_t structure.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/11/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_decNlmLockCntOfFile(lock_fileLock_t	*fileLockp)
{
	fileLockp->nlmLockCount--;
	if (fileLockp->nlmLockCount == 1)
	{
		/*TODO: Inform the upper layer that the file has no NLM locks on it*/
	}
} /* lock_decNlmLockCntOfFile() */

