/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         lock.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Byte range lock initialization, and timer handling
 *               functions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/26/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*One minute in terms of milliseconds*/
#define LOCK_TIME_ONE_MINUTE    (SEC_PER_MIN * MSEC_PER_SEC)

/*Five minutes in terms of milliseconds*/
#define LOCK_TIME_FIVE_MINUTES  (5 * LOCK_TIME_ONE_MINUTE)

/*Ten minutes in terms of milliseconds*/
#define LOCK_TIME_TEN_MINUTES   (10 * LOCK_TIME_ONE_MINUTE)

uint32				lock_traceEnable = 0;
uint32				lock_debugLevel = LOG_DEBUG;
//uint32				lock_debugLevel = LOCK_DEBUG2;
lock_globalCfg_t	*lock_globalCfgp;
eeeRTC_t			lock_timeIndefinite;
eeeRTC_t			lock_timeZero;
eeeRTC_t			lock_timeTenMinutes;
eeeRTC_t			lock_timeFiveMinutes;

/**************************
 * Initialization functions
 **************************
 */

/*-----------------------------------------------------------------
 * Name:         lock_init
 *
 * Description:
 *
 *     Lock module initialization function on the TX/RX CPU.
 *
 * Arguments:
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/26/2002
 *
 *-----------------------------------------------------------------
 */
void
lock_init(void)
{
	/*------------------------------------------------
	 * Alloc & Initialize lock global config structure
	 *------------------------------------------------
	 */
	LOCK_SHMEM_ALLOC(lock_globalCfgp,
					 LOCK_VSERVER_IDX_ANY, /*FIXME: What value should it be?*/
					 lock_globalCfg_t,
					 sizeof(lock_globalCfg_t));
	if (lock_globalCfgp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_init: Insufficient memory!"));
		return;
	}
	bzero(lock_globalCfgp, sizeof(lock_globalCfg_t));

	/*-------------------------------------------------------------------
	 * Initialize lock conflict checking rule array and access rule array
	 *-------------------------------------------------------------------
	 */
	lock_conflictRuleInit();
	lock_accessRuleInit();

	/*---------------------------
	 * Initialize the hash tables
	 *---------------------------
	 */
	lock_globalCfgp->cifsOwnerHash = vs_hash_create(
												LOCK_HASH_CIFS_OWNER_BUCKETS,
												lock_cifsLockOwnerHashFunc,
												lock_cifsLockOwnerHashCompare,
												lock_cifsLockOwnerHashRelease);

	lock_globalCfgp->nlmClientHash = vs_hash_create(
												LOCK_HASH_NLM_CLIENT_BUCKETS,
												lock_nlmLockClientHashFunc,
												lock_nlmLockClientHashCompare,
												lock_nlmLockClientHashRelease);

	lock_globalCfgp->nlmOwnerHash = vs_hash_create(
												LOCK_HASH_NLM_OWNER_BUCKETS,
												lock_nlmLockOwnerHashFunc,
												lock_nlmLockOwnerHashCompare,
												lock_nlmLockOwnerHashRelease);

	/*-------------------------------
	 * Initialize the timeout handles
	 *-------------------------------
	 */
	lock_globalCfgp->pendingLockTimeoutHandle = TIMER_ERR_NOT_RUNNING;
	lock_globalCfgp->ownerTimeoutHandle = TIMER_ERR_NOT_RUNNING;

	/*--------------------------------------
	 * Initialize the various timeout values
	 *--------------------------------------
	 */
	lock_timeIndefinite = *(eee_rtcInf());
	lock_timeZero = *(eee_rtcZero());

	eee_msToRTC(LOCK_TIME_TEN_MINUTES, &lock_timeTenMinutes);
	eee_msToRTC(LOCK_TIME_FIVE_MINUTES, &lock_timeFiveMinutes);

    /*---------------------------
     * Set data size in VS module
     *---------------------------
     */
    vs_memSetDataSize(vs_mem_dt_cifsLockOwner, sizeof(lock_cifsLockOwner_t));
    vs_memSetDataSize(vs_mem_dt_nlmRecOnVolCB, sizeof(lock_nlmRecOnVolCB_t));
    vs_memSetDataSize(vs_mem_dt_nlmLockClient, sizeof(lock_nlmLockClient_t));
    vs_memSetDataSize(vs_mem_dt_nlmLockOwner, sizeof(lock_nlmLockOwner_t));
    vs_memSetDataSize(vs_mem_dt_cifsLockInfo, sizeof(lock_cifsLockInfo_t));
    vs_memSetDataSize(vs_mem_dt_nlmLockInfo, sizeof(lock_nlmLockInfo_t));
    vs_memSetDataSize(vs_mem_dt_exLock, sizeof(lock_exLock_t));
    vs_memSetDataSize(vs_mem_dt_shLock, sizeof(lock_shLock_t));
    vs_memSetDataSize(vs_mem_dt_shNode, sizeof(lock_shNode_t));
    vs_memSetDataSize(vs_mem_dt_pendingLock, sizeof(lock_pendingLock_t));
    vs_memSetDataSize(vs_mem_dt_fileLock, sizeof(lock_fileLock_t));
    vs_memSetDataSize(vs_mem_dt_lockDependency, sizeof(lock_dependency_t));
    vs_memSetDataSize(vs_mem_dt_lockOwnerTimeout, sizeof(lock_ownerTimeout_t));
	return;
} /* lock_init() */


/*************************
 * Timer related functions
 *************************
 */

/*-----------------------------------------------------------------
 * Name:         lock_createAndStartPendingLockTimer
 *
 * Description:
 *
 *     Create and start the global pending lock timer.
 *
 * Arguments:    
 * 
 *     future - A future time in real time clock form.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_createAndStartPendingLockTimer(eeeRTC_t	future)
{
	int32		rc;
	eeeRTC_t	now;
	eeeRTC_t	rtcDuration;
	uint64		msDuration;

   	/*
	 * Get current time
	 */
	eee_rtcStart(&now);

	/*
	 * Get duration between current time and future time
	 */
	rc = eee_rtcSub(&future, &now, &rtcDuration);
	if (rc != NFX_OK)
	{
		/*TODO:
		 * The global list has pending locks which should timeout
		 * in the past time. This shouldn't happen.
		 * There must be error in the timeout passed from upper layer!
		 */
		LOCK_PRINTF(LOG_ERR, ("lock_createAndStartPendingLockTimer: "
							"Invalid timeout value passed from upper layer!"));
		return (NFX_ERR);
	}

	/*
	 * Convert the duration to milliseconds
	 */
	msDuration = eee_rtcToMS(&rtcDuration);

	/*
	 * Create the timer
	 */
	lock_globalCfgp->pendingLockTimeoutHandle =
					            eee_timerCreate(msDuration, EEE_TIMER_ONCE);
	if (lock_globalCfgp->pendingLockTimeoutHandle ==
					   					TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_createAndStartPendingLockTimer: "
												"Failed to create timer!"));
		return (NFX_ERR);
	}

	/*
	 * Start the timer
	 */
	rc = eee_timerStart(lock_globalCfgp->pendingLockTimeoutHandle,
						(void *)lock_pendingLockTimeoutHandler,
						0, 0, 0, 0);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_createAndStartPendingLockTimer: "
												"Failed to start timer!"));
		return (NFX_ERR);
	}

	/*
	 * Remember the timeout value of current running timer
	 */
	lock_globalCfgp->timeout = future; 

	return (NFX_OK);
} /* lock_createAndStartPendingLockTimer() */
	
/*-----------------------------------------------------------------
 * Name:         lock_stopAndDestroyPendingLockTimer
 *
 * Description:
 *
 *     Stop and destroy the global pending lock timer.
 *
 * Arguments:    
 * 
 *     None.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_stopAndDestroyPendingLockTimer()
{
	int32	rc;

	rc = eee_timerStop(lock_globalCfgp->pendingLockTimeoutHandle);
	if (rc != NFX_OK)
		return (rc);

	rc = eee_timerDestroy(lock_globalCfgp->pendingLockTimeoutHandle);
	if (rc != NFX_OK)
		return (rc);

	lock_globalCfgp->timeout = lock_timeZero;
	lock_globalCfgp->pendingLockTimeoutHandle = TIMER_ERR_NOT_RUNNING;
		
	return (rc);
} /* lock_stopAndDestroyPendingLockTimer() */

/*-----------------------------------------------------------------
 * Name:         lock_checkPendingLockTimer
 *
 * Description:
 *
 *     This function is called after a pending lock is added to
 *     or some pending locks are removed from the global list.
 *     If the global timer is not started yet, start the timer.
 *     If the head of the global list changed and the new head
 *     will timeout sooner than the timer timeout, then restart
 *     the timer.
 *
 * Arguments:    
 * 
 *     None.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkPendingLockTimer(void)
{
	int32				rc = NFX_OK;
	lock_pendingLock_t	*globalPendLockHeadp;

	if (VS_MEM_IDX_IS_NULL(lock_globalCfgp->pendingLockList))
	{
		/*-------------------------
		 * The global list is empty
		 *-------------------------
		 */
		if (lock_globalCfgp->pendingLockTimeoutHandle != TIMER_ERR_NOT_RUNNING)
		{
			/*----------------------------
			 * The timer is running.
			 * Stop and destroy the timer.
			 *----------------------------
			 */
			rc = lock_stopAndDestroyPendingLockTimer();
		}
		return (rc);
	}

	/*-----------------------------
	 * The global list is not empty
	 *-----------------------------
	 */
	globalPendLockHeadp = (lock_pendingLock_t *)
							vs_mem_idxToPtr(lock_globalCfgp->pendingLockList);
	if (globalPendLockHeadp == NULL)
		return (NFX_ERR);

	if (lock_globalCfgp->pendingLockTimeoutHandle == TIMER_ERR_NOT_RUNNING)
	{
		/*------------------------------------
		 * The timer has not been started yet.
		 * Create and start the timer.
		 *------------------------------------
		 */
		rc = lock_createAndStartPendingLockTimer(globalPendLockHeadp->timeout);
		if (rc != NFX_OK)
			return (rc);
	}
	else if (eee_rtcCompare(&lock_globalCfgp->timeout,
						   	&globalPendLockHeadp->timeout) > 0)
	{
		/*---------------------------------------------
		 * New head of the pending list timeout sooner.
		 * Restart the timer with new timeout value.
		 *---------------------------------------------
		 */
		/* Stop and destroy the timer */
		rc = lock_stopAndDestroyPendingLockTimer();
		if (rc != NFX_OK)
			return (rc);

		 /* Create and start the timer */
		rc = lock_createAndStartPendingLockTimer(globalPendLockHeadp->timeout);
		if (rc != NFX_OK)
			return (rc);
	}

	return (NFX_OK);
} /* lock_checkPendingLockTimer() */

/*-----------------------------------------------------------------
 * Name:         lock_pendingLockTimeoutHandler
 *
 * Description:
 *
 *     This function is called when the global pending lock timer
 *     timeout. It removes all the locks, whose timeout value is
 *     current time, from the global pending lock list as well as
 *     from the files' local pending lock lists that these locks
 *     belong to.
 *
 * Arguments:    
 * 
 *     timer_handle - Global pending lock timer handle.
 *
 *     param1 - Not used.
 *
 *     param2 - Not used.
 *
 *     param3 - Not used.
 *
 *     param4 - Not used.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/27/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_pendingLockTimeoutHandler(int32	timer_handle, 
							   void		*param1,
							   void		*param2,
							   void		*param3,
							   void		*param4)
{
	int32						rc;
	vs_mem_index_t				pendLockIdx, nextPendLockIdx;
	lock_pendingLock_t			*pendLockp;
	lock_fileLock_t				*fileLockp;
	lock_delPendLocksParam_t	param;

	/*
	 * At this point, the timer has already been destroyed.
	 * Reset the global variable.
	 */
	lock_globalCfgp->timeout = lock_timeZero;
	lock_globalCfgp->pendingLockTimeoutHandle = TIMER_ERR_NOT_RUNNING;

	/*-------------------------------------------
	 * Go through the global pending lock list to
	 * remove all the locks that timeout.
	 *-------------------------------------------
	 */
	pendLockIdx = lock_globalCfgp->pendingLockList;
	while (!VS_MEM_IDX_IS_NULL(pendLockIdx))
	{
		/* Get pendingLock structure */
		pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
		if (pendLockp == NULL)
			return (NFX_ERR);

		nextPendLockIdx = pendLockp->timeoutLink;

		if (eee_rtcCompare(&lock_globalCfgp->timeout, &pendLockp->timeout) < 0)
		{
			/* Already removed all the locks that should timeout
			 * at present time*/
			break;
		}

		/*---------------------------------------------
		 * Remove the lock from local pending lock list
		 *---------------------------------------------
		 */
		fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(pendLockp->fileLock);
		if (fileLockp == NULL)
			return (NFX_ERR);

		param.type = LOCK_DELPENDLOCKS_TYPE_TIMEOUT;
		param.u.pendLockIdx = pendLockIdx;

		rc = lock_deletePendingLocksFromLocalList(fileLockp, param);
		if (rc != NFX_OK)
			return (NFX_ERR);

		/*----------------------------------------------------------------
		 * Notify the upper layer that the lock is canceled due to timeout
		 *----------------------------------------------------------------
		 */
		/*TODO*/

		/* Go to next one in the list */
		pendLockIdx = nextPendLockIdx;
	} /* while() */

	/*---------------------------------------------------
	 * Set the new header of the global pending lock list
	 *---------------------------------------------------
	 */
	lock_globalCfgp->pendingLockList = pendLockIdx;

	/*----------------------------------------------------------
	 * After all the timeout pending locks are removed from
	 * the global pending lock list, restart timer if necessary.
	 *----------------------------------------------------------
	 */
	rc = lock_checkPendingLockTimer();

	return (rc);
} /* lock_pendingLockTimeoutHandler() */

/*-----------------------------------------------------------------
 * Name:         lock_insertOwnerClientToTimeoutList
 *
 * Description:
 *
 *     Add a CIFS owner, an NLM client, or an NLM owner structure
 *     in the global timeout list. The owners/clients in the list
 *     no longer have locks associated. The structure will be deleted
 *     5 to 15 minutes later after it is put into the list. 
 *
 * Arguments:    
 * 
 *     ownerClientIdx - Index to a structure of type lock_cifsLockOwner_t,
 *                      lock_nlmLockClient_t, or lock_nlmLockOwner_t.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/23/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_insertOwnerClientToTimeoutList(vs_mem_index_t	ownerClientIdx)
{
	int32					rc;
	vs_mem_index_t			nodeIdx, newNodeIdx;
	lock_ownerTimeout_t		*nodep, *newNodep;

	/*----------------------------------------
	 * Allocate a new node and init the fields
	 *----------------------------------------
	 */
	newNodep = (lock_ownerTimeout_t *)lock_memAllocOwnerTimeout(ownerClientIdx,
																&newNodeIdx);
	if (newNodep == NULL)
		return (NFX_ERR);

	/*----------------------------------------------
	 * Add the node to the global owner timeout list
	 *----------------------------------------------
	 */
	if (VS_MEM_IDX_IS_NULL(lock_globalCfgp->ownerTimeoutList))
	{
		/*----------------------------------------
		 * List is empty.
		 * This node becomes the only one in list.
		 *----------------------------------------
		 */
		lock_globalCfgp->ownerTimeoutList = newNodeIdx;

		/*---------------------------
		 * Create and start the timer
		 *---------------------------
		 */

		/* Create a timer */
		lock_globalCfgp->ownerTimeoutHandle = eee_timerCreate(
													LOCK_TIME_TEN_MINUTES,
												   	EEE_TIMER_REPEAT);
		if (lock_globalCfgp->ownerTimeoutHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_insertOwnerClientToTimeoutList: "
												"Failed to create timer!"));
			return (NFX_ERR);
		}

		/* Start the timer */
		rc = eee_timerStart(lock_globalCfgp->ownerTimeoutHandle,
							(void *)lock_ownerTimeoutHandler,
							0, 0, 0, 0);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_insertOwnerClientToTimeoutList: "
												"Failed to start timer!"));
			return (NFX_ERR);
		}

		LOCK_PRINTF(LOG_DEBUG, ("lock_insertOwnerClientToTimeoutList: "
														"Timer started."));
		return (NFX_OK);
	}

	/*---------------------------------------------------------------
	 * List is not empty. Append the new node to the end of the list.
	 *---------------------------------------------------------------
	 */
	nodeIdx = lock_globalCfgp->ownerTimeoutList;
	while (!VS_MEM_IDX_IS_NULL(nodeIdx))
	{
		nodep = (lock_ownerTimeout_t *)vs_mem_idxToPtr(nodeIdx);
		if (nodep == NULL)
			return (NFX_ERR);

		nodeIdx = nodep->next;
	}
	nodep->next = newNodeIdx;

	return (NFX_OK);
} /* lock_insertOwnerClientToTimeoutList() */

/*-----------------------------------------------------------------
 * Name:         lock_ownerTimeoutHandler
 *
 * Description:
 *
 *     This function is called when the global timer for owner timeout
 *     list timeout. It removes the structures in the list, whose
 *     timeout value already expired or will expire within 5 minutes.
 *     Before deleting the structrue, check the lock/owner count in
 *     the structure. If the count is non-zero, do not delete the
 *     structure.
 *
 * Arguments:    
 * 
 *     timer_handle - Global owner timeout list timer handle.
 *
 *     param1 - Not used.
 *
 *     param2 - Not used.
 *
 *     param3 - Not used.
 *
 *     param4 - Not used.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_ownerTimeoutHandler(int32	timer_handle, 
						 void	*param1,
						 void	*param2,
						 void	*param3,
						 void	*param4)
{
	int32				rc;
	eeeRTC_t			future;
	vs_mem_index_t		ownerClientIdx;
	uint16				dataType;
	vs_mem_index_t		nodeIdx, nextNodeIdx;
	lock_ownerTimeout_t	*nodep;

	LOCK_PRINTF(LOG_DEBUG, ("lock_ownerTimeoutHandler: Enter."));

	/*
	 * Get future time - 5 minutes from now
	 */
	eee_rtcStart(&future);	/*Current time*/
	rc = eee_rtcIncr(&future, &lock_timeFiveMinutes);
	if (rc != NFX_OK)
		return (NFX_ERR);

	/*--------------------------------------------
	 * Go through the global owner timeout list to
	 * remove the nodes whose timeout value already
	 * expired or will expire within 5 minutes.
	 *--------------------------------------------
	 */
	nodeIdx = lock_globalCfgp->ownerTimeoutList;
	while (!VS_MEM_IDX_IS_NULL(nodeIdx))
	{
		/* Get ownerTimeout structure */
		nodep = (lock_ownerTimeout_t *)vs_mem_idxToPtr(nodeIdx);
		if (nodep == NULL)
			return (NFX_ERR);

		nextNodeIdx = nodep->next;

		if (eee_rtcCompare(&future, &nodep->timeout) < 0) 
		{
			/* Already removed all the locks that should be removed */
			break;
		}

		/*---------------------------------------------
		 * Remove the owner/client from the global list
		 *---------------------------------------------
		 */
		ownerClientIdx = nodep->owner_client;
		dataType = VS_MEM_IDX_GET_DATA_TYPE(ownerClientIdx);
		switch (dataType)
		{
			case vs_mem_dt_cifsLockOwner:
			{
				lock_cifsLockOwner_t	*lockOwnerp;

				lockOwnerp = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(ownerClientIdx);
				if (lockOwnerp != NULL)
                {
				    /* If the lock count is non-zero, do not delete it.
				     * Otherwise, remove it from the hash table, and
				     * free the memory. */
				    if (lockOwnerp->header.lockCount == 0)
				    {
					    rc = lock_htDeleteCifsLockOwner(lockOwnerp->tid,
													    lockOwnerp->fid,
													    lockOwnerp->pid,
													    ownerClientIdx);
					    lock_memFreeCifsLockOwner(ownerClientIdx);
				    }
					else
					{
						/* Remove it from the list */
						lockOwnerp->header.inList = 0;
					}
				}
			}
				break;

			case vs_mem_dt_nlmLockClient:
			{
				lock_nlmLockClient_t	*lockClientp;

				lockClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(ownerClientIdx);
				if (lockClientp != NULL)
                {
				    /* If the owner count is non-zero, do not delete it.
				     * Otherwise, remove it from the hash table, and
				     * free the memory. */
				    if (lockClientp->header.ownerCount == 0)
				    {
					    rc = lock_htDeleteNlmLockClient(lockClientp->clientIpAddr,
									                    ownerClientIdx);
					    lock_memFreeNlmLockClient(ownerClientIdx, lockClientp);
				    }
					else
					{
						/* Remove it from the list */
						lockClientp->header.inList = 0;
					}
				}
			}
				break;

			case vs_mem_dt_nlmLockOwner:
			{
				lock_nlmLockOwner_t	*lockOwnerp;

				lockOwnerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(ownerClientIdx);
				if (lockOwnerp != NULL)
                {
				    /* If the lock count is non-zero, do not delete it.
				     * Otherwise, remove it from the hash table, and
				     * free the memory. */
				    if (lockOwnerp->header.lockCount == 0)
				    {
					    rc = lock_htDeleteNlmLockOwner(lockOwnerp->lockClient,
												       lockOwnerp->pid,
												       ownerClientIdx);
					    lock_memFreeNlmLockOwner(ownerClientIdx, lockOwnerp);
				    }
					else
					{
						/* Remove it from the list */
						lockOwnerp->header.inList = 0;
					}
				}
			}
				break;

			default:
				break;
		} /* switch (dataType) */

        lock_memFreeOwnerTimeout(nodeIdx);

		/* Go to next one in the list */
		nodeIdx = nextNodeIdx;
	} /* while() */

	/*----------------------------------------------------
	 * Set the new header of the global owner timeout list
	 *----------------------------------------------------
	 */
	lock_globalCfgp->ownerTimeoutList = nodeIdx;

	if (VS_MEM_IDX_IS_NULL(lock_globalCfgp->ownerTimeoutList))
	{
		/* The list becomes empty. Stop and destroy the timer */
		rc = eee_timerStop(lock_globalCfgp->ownerTimeoutHandle);
		if (rc != NFX_OK)
			return (rc);

		rc = eee_timerDestroy(lock_globalCfgp->ownerTimeoutHandle);
		if (rc != NFX_OK)
			return (rc);

		LOCK_PRINTF(LOG_DEBUG, ("lock_ownerTimeoutHandler: Timer stopped."));
	}

	return (rc);
} /* lock_ownerTimeoutHandler() */

