/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: luc-cmd.c
 *
 * RCS:  $Id: luc-cmd.c,v 1.10 2002/08/01 17:56:37 sandrineb Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains routines to handle LUC UI Command.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 09/13/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/luc-api.h>
#include "luc-cam.h"

int32 luc_cmdHelp( int32 argc, char *argv[] );
int32 luc_cmdShow( int32 argc, char *argv[] );
int32 luc_simStart( int32 argc, char *argv[] );
int32 luc_cmdTrace( int32 argc, char *argv[] );
int32 luc_verify(int32 argc, char *argv[]);

extern int32 luc_sendintf(int32 argc, char *argv[]);

typedef struct
{
	int32	(*cmdFunc)( int32 argc, char *argv[] );
	char	*cmd;
	char	*helpTxt;
} luc_cmdTbl_t;

luc_cmdTbl_t	luc_cmdTable[ ] =
{
	{ luc_cmdShow, 	"show", 	"luc show ip|debug|stats" },
	{ luc_simStart, "start", 	"luc start <ipaddr> <nfs port>" },
	{ luc_cmdTrace, "trace", 	"luc trace all|off|fbd|rcd|tcd|trd|pkt" },
	{ luc_sendintf, "send",     "send local interface"},
	{ luc_verify,	"verify",   "luc verify 0/1"},
};

#define LUC_CMD_MAX		sizeof(luc_cmdTable)/sizeof(luc_cmdTbl_t)

/*-----------------------------------------------------------------
 * Name: luc_cmdHelp
 *
 * Description: Display help-texts for all LUC commands.
 *
 * Created by : Shaun Hui
 *  
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 luc_cmdHelpAll( void )
{
	int32 i;

	printf("Usage:");
	for (i=0; i < LUC_CMD_MAX; i++)
	{
		printf("\t%s\n", luc_cmdTable[i].helpTxt);
	}
	return 0;
}

/*-----------------------------------------------------------------
 * Name: luc_cmdHelpCmd
 *
 * Description: Display help-texts for an individual LUC command.
 *
 * Created by : Shaun Hui
 *  
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 luc_cmdHelpCmd( char *pCmd )
{
	int32 i;

	for (i=0; i < LUC_CMD_MAX; i++)
	{
		if (strequ(pCmd, luc_cmdTable[i].cmd))
		{
			printf("Usage:");
			printf("\t%s\n", luc_cmdTable[i].helpTxt);
			return 0;
		}
	}

	return (luc_cmdHelpAll( ));
}

/*-----------------------------------------------------------------
 * Name: luc_cmdHelp
 *
 * Description: Display LUC help-texts.  
 *
 *				luc help [show | start | trace]
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 luc_cmdHelp( int32 argc, char *argv[] )
{
	int32 i;

	if (argc < 3)
		return (luc_cmdHelpAll());
	else 
		return (luc_cmdHelpCmd( argv[2] ));

}

/*-----------------------------------------------------------------
 * Name: luc_cmdTrace
 *
 * Description: Enables or disables LUC tracing.
 *
 *				luc trace <on|off>
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 luc_cmdTrace( int32 argc, char *argv[] )
{

	if (argc < 3)
	{
		luc_cmdHelpCmd("trace");
		return 0;
	}

	if (strequ(argv[2], "all"))
	{
		luc_traceFlag = LUC_TRACE_ALL ;
	}
	else if (strequ(argv[2], "off"))
	{
		luc_traceFlag = 0;
	}
	else if (strequ(argv[2], "pkt"))
	{
		luc_traceFlag = LUC_TRACE_PKT ;
	}
	else if (strequ(argv[2], "trd"))
	{
		luc_traceFlag = LUC_TRACE_TRD ;
	}
	else if (strequ(argv[2], "tcd"))
	{
		luc_traceFlag = LUC_TRACE_TCD ;
	}
	else if (strequ(argv[2], "rcd"))
	{
		luc_traceFlag = LUC_TRACE_RCD ;
	}
	else if (strequ(argv[2], "fbd"))
	{
		luc_traceFlag = LUC_TRACE_FBD ;
	}
	else 
		luc_cmdHelpCmd("trace");
	
	return 0;
}

/*-----------------------------------------------------------------
 * Name: luc_cmdShow
 *
 * Description: Dispatch for LUC SHOW commands.  
 *
 *				luc show [fbr|rcr|tcr|trr]
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 luc_cmdShow( int32 argc, char *argv[] )
{

	if (argc < 3)
	{
		luc_cmdHelpCmd("show");
		return 0;
	}

	if (strequ(argv[2], "debug"))
	{
		return (luc_showDebug( ));
	}
#ifdef PKT_USE_TIME_STAMP
	else if (strequ(argv[2], "nfs"))
	{
		return (pkt_showNfsProcTime());	
	}
#endif
	else if (strequ(argv[2], "stats"))
	{
		return (luc_showStats(0));	
	}
	else if (strequ(argv[2], "ip"))
	{
		return (luc_showIpIfCfg());	
	}
	else
		return (luc_cmdHelpCmd("show"));
}

/*-----------------------------------------------------------------
 * Name: luc_cmd
 *
 * Description: Main dispatch for LUC commands.  Called from 
 *				command table.
 *
 *				luc addif <ipaddr> <nfsport> 
 *				luc show [fbr|rcr|tcr|trr]
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 luc_cmd( int32 argc, char *argv[] )
{
	int32	i;

	if (argc < 2)
	{
		return (luc_cmdHelp(argc, argv));
	}

	for (i=0; i < LUC_CMD_MAX; i++)
	{
		if (strequ(argv[1], luc_cmdTable[i].cmd))
		{
			return ((*luc_cmdTable[i].cmdFunc)(argc, argv));
		}
	}
	return (luc_cmdHelp(argc, argv));
}

int32 luc_verify(int32 argc, char* argv[])
{

	if (argc != 3)
		return 0;
	
	if (atoi(argv[2])) {
		printf("software verify enabled\n");
		pLucCfg->rxVerify =1;
	}
	else {
		printf("software verify disabled\n");
		pLucCfg->rxVerify = 0;
	}
	
	return 0;
}


