/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: luc-ipif.c
 *
 * RCS:  $Id: luc-ipif.c,v 1.4 2002/03/13 19:37:00 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: LUC code to add/delete IP interfaces. 
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 12/10/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>

#include <exports/pkt-api.h>
#include <exports/timer-api.h>
#include <exports/luc-api.h>

#define LUC_LINK_HASH_SHIFT 	6
#define LUC_MAX_IP_IF   		256						/* Max IP supported */
#define LUC_MAX_IP_HASH 		(LUC_MAX_IP_IF * 2)		/* Hash tbl size */


/*-------------------------------
 * IP Hash table entry.
 *-------------------------------
 */
typedef struct 
{
	luc_ipIf_t	*ipIfp;				/* ptr to ipIf_t */
	int32		count;				/* # of ip hashed into same idx */
} luc_ipHash_t;

/*-------------------------------
 * IP Config info.
 *-------------------------------
 */
typedef struct
{
	luc_ipHash_t	*pIpHashTbl;	/* Hash Tbl start */
	luc_ipIf_t		*pIpIfTbl;		/* Ip IF Tbl start */
	luc_ipIf_t		*pIpIfFree;		/* next available free entry */
	int32			freeCnt;		/* Num of Free entries */
} luc_ipIfCfg_t;


/*-------------------------------
 * Global IP Config info. ptr.
 *-------------------------------
 */
luc_ipIfCfg_t	*pLucIpIfCfg;

/*-----------------------------------------------------------------
 * Name:        luc_ipIfHash
 *
 * Description: Hashing function for IP Interface.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
static luc_ipHash_t *
luc_ipIfHash( short16 link, uint32 ipaddr, int32 *idx)
{
    *idx = ((ipaddr >> 16) ^ (ipaddr) + (link << LUC_LINK_HASH_SHIFT)) & 
			(LUC_MAX_IP_HASH -1);
	return (pLucIpIfCfg->pIpHashTbl + *idx);
}

/*-----------------------------------------------------------------
 * Name:        luc_ipIfAllocEntry
 *
 * Description: Allocate a free IP Interface for luc_ipIfTable.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
static luc_ipIf_t *
luc_ipIfAllocEntry( void )
{
    int32		i;
    luc_ipIf_t  *ipIfp;
    luc_ipIf_t  *newp;
    luc_ipIf_t  *ipIfTblEnd;

	/*-------------------------------
	 * Go thru entire table.
	 *-------------------------------
	 */
    newp       = NULL;
    ipIfp      = pLucIpIfCfg->pIpIfFree;
    ipIfTblEnd = pLucIpIfCfg->pIpIfTbl + LUC_MAX_IP_IF - 1;
    for (i=0; i < LUC_MAX_IP_IF; i++ )
    {
		/*-------------------------------
		 * find one that's not in use.
		 *-------------------------------
		 */
        if (!(ipIfp->inUse))
		{
			newp = ipIfp;
			break;
		}

		/*-------------------------------
		 * if end of table, wrap around.
		 *-------------------------------
		 */
		if (++ipIfp > ipIfTblEnd)
			ipIfp = pLucIpIfCfg->pIpIfTbl;
	}

	if (newp)
	{
		/*-------------------------------
		 * Mark it used & return.
		 *-------------------------------
		 */
        newp->inUse           = 1;
		pLucIpIfCfg->freeCnt -= 1;
		pLucIpIfCfg->pIpIfFree = (++ipIfp <= ipIfTblEnd) ?  
								ipIfp : pLucIpIfCfg->pIpIfTbl;
    } 
    return (newp);
}

/*-----------------------------------------------------------------
 * Name:        luc_ipIfFreeEntry
 *
 * Description: Return an IP Interface to luc_ipIfTable.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
static luc_ipIf_t *
luc_ipIfFreeEntry( luc_ipIf_t *ipIfp )
{

	/*-------------------------------
	 * Clear in use flag & next.
	 *-------------------------------
	 */
    ipIfp->inUse         = 0;
    ipIfp->next          = NULL;
	pLucIpIfCfg->freeCnt += 1;
   
}

/*-----------------------------------------------------------------
 * Name:        luc_ipIfFindEntry
 *
 * Description: Find an IP Interface given link & IP address.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
luc_ipIf_t *
luc_ipIfFindEntry( short16 link, uint32 ipaddr)
{
    luc_ipHash_t	*ipHashp;
    luc_ipIf_t  	*ipIfp;
    int32       	idx;

	/*-------------------------------
	 * Hash the link & ip.
	 *-------------------------------
	 */
    ipHashp = luc_ipIfHash(link, ipaddr, &idx);
	ipIfp   = ipHashp->ipIfp;
   
    while(ipIfp)
    {
		/*-------------------------------
		 * Make sure ip & link matches,
		 * in case of collision.
		 *-------------------------------
		 */
        if ((ipIfp->ipaddr == ipaddr) && 
			(ipIfp->link   == link))
            return (ipIfp);
        ipIfp = ipIfp->next;
    }

    return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        luc_ipIfAddEntry
 *
 * Description: Add an IP Interface given link & IP address.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
luc_ipIf_t *
luc_ipIfAddEntry( short16 link, uint32 ipaddr)
{
    luc_ipHash_t	*ipHashp;
    luc_ipIf_t  	*ipIfp;
    int32       	idx;

	/*-------------------------------
	 * Check for duplicate entry.
	 *-------------------------------
	 */
    if (luc_ipIfFindEntry(link, ipaddr))
		return NULL;
	
	/*-------------------------------
	 * Allocate a new entry.
	 *-------------------------------
	 */
	ipIfp = luc_ipIfAllocEntry( );
    if (!ipIfp)
        return (NULL);

	ipIfp->ipaddr = ipaddr;
	ipIfp->link   = link;

	/*-------------------------------
	 * Find out the hash table idx.
	 * Insert into head of chain.
	 *-------------------------------
	 */
    ipHashp = luc_ipIfHash(link, ipaddr, &idx);
	ipIfp->next     = ipHashp->ipIfp; 
	ipHashp->ipIfp  = ipIfp;
	ipHashp->count += 1;

	return (ipIfp);
}


/*-----------------------------------------------------------------
 * Name:        luc_ipIfDelEntry
 *
 * Description: Delete an IP Interface given link & IP address.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
int32
luc_ipIfDelEntry(short16 link, uint32 ipaddr)
{
    luc_ipHash_t  	*ipHashp;
    luc_ipIf_t  	*ipIfp;
    luc_ipIf_t  	*prevIfp;
    int32       	idx;

    ipHashp = luc_ipIfHash(link, ipaddr, &idx);
	ipIfp   = ipHashp->ipIfp;

	if (!ipIfp)
		return 0;

	/*-------------------------------
	 * Delete head of chain.
	 *-------------------------------
	 */
	if ((ipIfp->ipaddr == ipaddr) &&
		(ipIfp->link   == link))
	{
		ipHashp->ipIfp  = ipIfp->next;
		ipHashp->count -= 1;
		luc_ipIfFreeEntry( ipIfp );
		return 0;
	}

	/*-------------------------------
	 * Go thru chain.
	 *-------------------------------
	 */
	prevIfp = ipIfp;
    ipIfp   = ipIfp->next;
	while(ipIfp)
	{
        if ((ipIfp->ipaddr == ipaddr) &&
        	(ipIfp->link   == link))
		{
			ipHashp->count -= 1;
			prevIfp->next   = ipIfp->next;
			luc_ipIfFreeEntry( ipIfp );
            return 0;
		}
		prevIfp = ipIfp;
        ipIfp = ipIfp->next;
	}
    return 0;
}

/*-----------------------------------------------------------------
 * Name:        luc_initIpIf
 *
 * Description: Initialize IP IF related data structures.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */ 
void
luc_initIpIf( void )
{
	int32 			allocSize;
	luc_ipIfCfg_t	*pIpIfCfg;
	luc_ipHash_t	*pIpHashTbl;
	luc_ipIf_t		*pIpIfTbl;

	/*-------------------------------
	 * Allocate ip if config structure.
	 *-------------------------------
	 */
	allocSize = sizeof(luc_ipIfCfg_t);
	pIpIfCfg  = (luc_ipIfCfg_t *) eee_ramAlloc(allocSize);
	if (!pIpIfCfg)
		luc_panic();
	memset(pIpIfCfg, 0, allocSize);

	/*-------------------------------
	 * Allocate ip if hash table.
	 *-------------------------------
	 */
	allocSize  = sizeof(luc_ipHash_t) * LUC_MAX_IP_HASH;
	pIpHashTbl = (luc_ipHash_t *)  eee_ramAlloc(allocSize);
	if (!pIpHashTbl)
		luc_panic();
	memset(pIpHashTbl, 0, allocSize);

	/*-------------------------------
	 * Allocate ip if table.
	 *-------------------------------
	 */
	allocSize = sizeof(luc_ipIf_t) * LUC_MAX_IP_IF;
	pIpIfTbl  = (luc_ipIf_t *)  eee_ramAlloc(allocSize);
	if (!pIpIfTbl)
		luc_panic();
	memset(pIpIfTbl, 0, allocSize);

	pLucIpIfCfg             = pIpIfCfg;
	pLucIpIfCfg->pIpIfTbl   = pIpIfTbl;
	pLucIpIfCfg->pIpHashTbl = pIpHashTbl;
	pLucIpIfCfg->pIpIfFree  = pIpIfTbl;
	pLucIpIfCfg->freeCnt    = LUC_MAX_IP_IF;
}

/*-----------------------------------------------------------------
 * Name:        luc_showIpIfCfg
 *
 * Description: Shows IP Interface table & hash table entries.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 12/10/2001
 *-----------------------------------------------------------------
 */
int32
luc_showIpIfCfg( void )
{
	int32 			i;
	int32 			useCnt;
	int32 			count;
	int32 			hashCnt;
	luc_ipIf_t		*ipIfp;
	luc_ipHash_t	*ipHashp;

	useCnt   = (LUC_MAX_IP_IF - pLucIpIfCfg->freeCnt);
	hashCnt  = 0;
	ipHashp  = pLucIpIfCfg->pIpHashTbl;
	for (i=0; i < LUC_MAX_IP_HASH; i++, ipHashp++)
	{
		if (!ipHashp->ipIfp) 
			continue;

		count = ipHashp->count;
		ipIfp = ipHashp->ipIfp;
		hashCnt++;

		printf("hash[%3d]=%d :\t", i, count);
		while (ipIfp)
		{
			printf("ifp[%d]=%x ip=%x\n", ipIfp->link, ipIfp, ipIfp->ipaddr);
			ipIfp = ipIfp->next;
			printf("\t\t");
		}
		printf("\n");
	}
	if (!hashCnt)
		hashCnt = 1;

	printf("luc IP Tbl Start: %x\n", pLucIpIfCfg->pIpIfTbl);
	printf("luc IP Tbl End  : %x\n", pLucIpIfCfg->pIpIfTbl + LUC_MAX_IP_IF-1);
	printf("luc IP Hash Tbl : %x\n", pLucIpIfCfg->pIpHashTbl);
	printf("luc Ip IF Free  : %x\n", pLucIpIfCfg->pIpIfFree);
	printf("luc Ip Free Cnt : %d\n", pLucIpIfCfg->freeCnt);
	printf("luc Ip Use Cnt  : %d\n", useCnt);
	printf("luc Hash Ratio  : %d:1\n", LUC_MAX_IP_HASH/hashCnt);
	printf("luc Hash Chain  : %d (ave)\n", useCnt/hashCnt);

	return 0;
}
