/* vi:ts=4 
 *-----------------------------------------------------------------
 *
 * Name: luc-reassem.c
 *
 * RCS:  $Id: luc-reas.c,v 1.8 2002/10/22 01:30:38 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains IP re-assembly code. This in effect takes place
 *				of bsd reassembly. We need to perform IP reassembly before
 *				fastpath or slowpath is determined.
 *
 * Created by:	Su Wang
 *
 * Date Created: 03/06/2002
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/cm-api.h>
#include <exports/pkt-api.h>
#include <exports/luc-api.h>
#include <exports/tpl-api.h>
#include "exports/timer-api.h"

#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-netinet/ip_stats.h"
#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/ip.h"

#include "luc-cam.h"
#include "luc-reas.h"
#include "luc-pkthdr.h"

static  luc_srcq_t		*luc_reas_hash[LUC_REASHASH_SIZE];
static  int				luc_reas_timer;

luc_memstat_t			luc_memstat;

#define	LUC_REAS_ERR(a)		printf a
#define	LUC_REAS_PRINTF(a)  

/*
 *-----------------------------------------------------------------
 *
 * Name: luc_FreeReasQ
 *
 *
 * Description: free a given reassemble queue
 *			
 * 
 * Created by:	Su Wang
 *
 * Date Created: 03/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
luc_FreeReasQ(luc_reasq_t	*lucReasq, 
			  uint32	flag)		/* deallocate associated pkt desc */
{
	luc_srcq_t	*lucSrcq;
	luc_reasq_t	*tmpReasq;
	luc_frag_t	*tmpFrag, *lucFrag;
	
	
	/* decrement counter */
	lucSrcq = lucReasq->reas_bck;
	lucSrcq->src_count--;

	if ((lucSrcq->src_count < 0) || 
		(lucSrcq->src_count >=LUC_MAX_REAS_PER_SRC)) {
		LUC_REAS_ERR(("file %s, line %d, bad src (%lx) count (%d)\n", 
					  __FILE__, __LINE__,
					  (address_t)lucSrcq, lucSrcq->src_count));
		luc_memstat.fatal++;
	}
	
		
	/* remove reassembly queue from IP src node */
	if (lucReasq == lucSrcq->src_first) {/* first element */
		
		/* unlink */
		lucSrcq->src_first = lucReasq->reas_next;
		
		/* only node in queue */
		if (lucSrcq->src_last == lucReasq)
			lucSrcq->src_last = 0;
	}
	else {						/* lucReasq is not first element */
		for (tmpReasq = lucSrcq->src_first; tmpReasq; tmpReasq = tmpReasq->reas_next) 
			if (tmpReasq->reas_next == lucReasq) 
				break;

		/* no node found in src IP queue */
		if (!tmpReasq) {
			LUC_REAS_ERR(("file %s, line %d, reassem queue not found\n", 
							 __FILE__, __LINE__));
			luc_memstat.fatal++;
			return;
		}
		
		/* unlink node */
		tmpReasq->reas_next = lucReasq->reas_next;
		
		/* update src_last */
		if (lucSrcq->src_last == lucReasq)
			lucSrcq->src_last = tmpReasq;
			
	}

	/* free lucReasq and associated structure */
	for (lucFrag = lucReasq->reas_pkt; lucFrag; lucFrag = tmpFrag) {
		if (flag)				/* free pkt desc */
			PKT_FREE_PACKET(lucFrag->frag_pdesc);
		
		tmpFrag = lucFrag->frag_next;
		eee_ramDealloc(lucFrag);
		luc_memstat.frag_free++;
	}
	// LUC_REAS_PRINTF(("lucreas (%x) removed, src %x, first %x, count %d\n",
    //                    lucReasq, lucSrcq,
	//                    lucSrcq->src_first, lucSrcq->src_count)); 
	eee_ramDealloc(lucReasq);
	luc_memstat.reas_free++;
	return;
}

/*
 *-----------------------------------------------------------------
 *
 * Name: luc_GetReasQ
 *
 *
 * Description: search for ip reassemble queue, if not found create one
 *			
 * 
 * Created by:	Su Wang
 *
 * Date Created: 03/07/2002
 *
 *-----------------------------------------------------------------
 */
static void
luc_GetReasQ(uint32		srcAddr, 
			 uint32		destAddr, 
			 ushort16	ipId, 
			 luc_reasq_t	**lucReasqPtr)
{
	
	luc_srcq_t	*lucSrcq = 0, *lucSrcqPrev = 0;										  
	int			hash;											
	luc_reasq_t	*tmpReasq = 0;
	
																
	hash = LUC_HASH(srcAddr);										
	lucSrcq = luc_reas_hash[hash];								
																	
	
	/* while loop to find existing reassemble queue */
	*lucReasqPtr = 0;
	while (lucSrcq) {
		if (lucSrcq->src_srcip == srcAddr) {

			/* found the correct src, looking for correct reas */
			for (tmpReasq = lucSrcq->src_first; tmpReasq; tmpReasq = tmpReasq->reas_next) {
				if ((tmpReasq->reas_srcip == srcAddr) &&
					(tmpReasq->reas_destip == destAddr) &&
					(tmpReasq->reas_id ==ipId)) {
					*lucReasqPtr = tmpReasq; /* found */
					break;		/* for loop */
				}
			}
			break;				/* while loop */
		}
		else {
			/* save prev. src in case we want to add src later */
			lucSrcqPrev = lucSrcq; 
			lucSrcq = lucSrcq->src_next;
		}
	}

	/* reassemble queue found */
	if (*lucReasqPtr) {
//LUC_REAS_PRINTF(("lucreas (%x) found, src %x, dest %x, id %x\n",
//					 (*lucReasqPtr), srcAddr, destAddr, ipId));
		return;
	}

	/* corresponding src IP queue has not been created, create now */
	if (!lucSrcq) {
		lucSrcq = (luc_srcq_t*)eee_ramAlloc(sizeof(luc_srcq_t));
		if (!lucSrcq) {
			luc_memstat.alloc_err++;
//			LUC_REAS_PRINTF(("file %s, line %d, malloc failed\n",
//							 __FILE__, __LINE__));
			return;
		}
		luc_memstat.src_alloc++;
		
		/* fill in the content */
		lucSrcq->src_next = 0;
		lucSrcq->src_srcip = srcAddr;
		lucSrcq->src_count = 0;
		lucSrcq->src_first = 0;
		lucSrcq->src_last = 0;
		
		/* link to the hash bucket */
		if (lucSrcqPrev) {
			lucSrcq->src_next = lucSrcqPrev->src_next;
			lucSrcqPrev->src_next = lucSrcq;
		}
		else
			luc_reas_hash[hash] = lucSrcq; /* first queue in bucket */
	}

	/* create reassemble queue */
	tmpReasq = (luc_reasq_t*)eee_ramAlloc(sizeof(luc_reasq_t));
	if (!tmpReasq) {
		luc_memstat.alloc_err++;
//		LUC_REAS_PRINTF(("file %s, line %d, malloc failed\n",
//						 __FILE__, __LINE__));
		return;
	}
	luc_memstat.reas_alloc++;
	
	tmpReasq->reas_next = 0;
	tmpReasq->reas_srcip = srcAddr;
	tmpReasq->reas_destip = destAddr;
	tmpReasq->reas_id = ipId;
	tmpReasq->reas_bck = lucSrcq;
	tmpReasq->reas_ttl = LUC_REAS_TIME_OUT;
	tmpReasq->reas_count = 0;
	tmpReasq->reas_pkt = 0;
	
	
	
	/* make sure there is space */
	while (lucSrcq->src_count >=  LUC_MAX_REAS_PER_SRC) {
		luc_memstat.force_free++;
		LUC_REAS_PRINTF(("force to remove reassebly queue %x, count %d\n",
					  lucSrcq->src_first, lucSrcq->src_count));
		luc_FreeReasQ(lucSrcq->src_first, 1);
	}
	

	lucSrcq->src_count++;
	/* link reassembly queue to src IP node */
	if (lucSrcq->src_first == 0)/* first queue */
		lucSrcq->src_first = lucSrcq->src_last = tmpReasq;
	else {						/* insert at end */
		lucSrcq->src_last->reas_next = tmpReasq;
		lucSrcq->src_last = tmpReasq;
	}
	

	/* we are done */
	*lucReasqPtr = tmpReasq;
//	LUC_REAS_PRINTF(("lucreas (%x) created, src %x, dest %x, id %x\n",
//					 (*lucReasqPtr), srcAddr, destAddr, ipId));
//	LUC_REAS_PRINTF(("lucsrc (%x), first %x, count %d\n",
//                    lucSrcq, lucSrcq->src_first, lucSrcq->src_count));
//   LUC_REAS_PRINTF(("hash table is %x  %x\n",
//                    luc_reas_hash[0], luc_reas_hash[1]));   
	return;
}

/*
 *-----------------------------------------------------------------
 *
 * Name: luc_ReassemTimeOut
 *
 *
 * Description: check for queues need to be removed
 *			
 * 
 * Created by:	Su Wang
 *
 * Date Created: 03/07/2002
 *
 *-----------------------------------------------------------------
 */
static int32
luc_ReassemTimeOut(uint32 handle, 
				   void* parm1, 
				   void* parm2, 
				   void* parm3, 
				   void* parm4)
{
	int				i;
	luc_srcq_t		*lucSrcq, *tmpSrcq, *lucSrcqPrev;
	luc_reasq_t		*lucReasq, *tmpReasq;
	struct ipstat	*ipStats = &ipstat;
	
//		LUC_REAS_PRINTF(("timeout in\n"));
	for (i = 0; i< LUC_REASHASH_SIZE; i++) {
		lucSrcq = luc_reas_hash[i];
		
		if (lucSrcq == 0)
			continue;
		
		/* search for each src IP queue */
		for (lucSrcqPrev = NULL; lucSrcq; lucSrcq = tmpSrcq) {
//			LUC_REAS_PRINTF(("timeout: src %x, first %x\n", lucSrcq, lucSrcq->src_first));
			
			/* search for each reassemble queue */
			for (lucReasq = lucSrcq->src_first; lucReasq; lucReasq = tmpReasq) {

				/* remember next reassembly queue */
				tmpReasq = lucReasq->reas_next;
				
				if (--lucReasq->reas_ttl > 0)
					continue;
				
				/* time expired, need to remove this queue */
				LUC_REAS_PRINTF(("remove reas %x, reas next %x\n", lucReasq, lucReasq->reas_next));

				luc_memstat.timeout_free++;
			    luc_FreeReasQ(lucReasq, 1);
			}

			/* remember what is the next src */
			tmpSrcq = lucSrcq->src_next;

			/* ready to to remove src IP queue */
			if (lucSrcq->src_count == 0) {
				if ((lucSrcq->src_last != 0) ||
					(lucSrcq->src_first != 0)) {
					LUC_REAS_ERR(("file %s, line %d, bad queue last %lx, first %lx\n", 
								  __FILE__, __LINE__,
								  (address_t)lucSrcq->src_last,
								  (address_t)lucSrcq->src_first));
					luc_memstat.fatal++;
				}


				/* unlink src IP queue */
				if (luc_reas_hash[i] == lucSrcq)
					luc_reas_hash[i] = lucSrcq->src_next;
				else
					lucSrcqPrev->src_next = lucSrcq->src_next;
				
				
				/* remove src IP queue */
				eee_ramDealloc(lucSrcq);
				luc_memstat.src_free++;
			}
			else	{			/* src IP queue that stays */
				lucSrcqPrev = lucSrcq;
			}
		} /* for each src IP queue */
	} /* for each hash bucket */

//	LUC_REAS_PRINTF(("timeout out\n"));
	return NFX_OK;
}

/*
 *-----------------------------------------------------------------
 *
 * Name: luc_ReassemInit
 *
 *
 * Description: Initiate fast path ressembly
 *			
 * 
 * Created by:	Su Wang
 *
 * Date Created: 03/08/2002
 *
 *-----------------------------------------------------------------
 */
void
luc_ReassemInit()
{
	

	bzero(luc_reas_hash, sizeof(luc_srcq_t*)*LUC_REASHASH_SIZE);
	
	luc_reas_timer = eee_timerCreate(500, EEE_TIMER_REPEAT);	

	if (luc_reas_timer == TIMER_ERR_NO_TIMER_AVAILABLE)
		panic("luc_ReassemInit: create timer  failed\n");
	
	if (eee_timerStart(luc_reas_timer, luc_ReassemTimeOut, NULL,
					   NULL, NULL, NULL) != NFX_OK)
		panic("luc_ReassemInit: start timer  failed\n");
}

/*
 *-----------------------------------------------------------------
 *
 * Name: luc_Reassem
 *
 *
 * Description: function for ip reassemble. Also the path of resulting
 *				pkts is determined here.
 *			
 *				
 * hash table structure:
 * 
 * hash bucket ---------> luc_srcq_t -----------> luc_srcq_t
 *                           |
 *							 |
 *							 V
 *                        luc_reasq_t (queue for a single reassembling pkt)
 *							 |
 *							 |
 *							 V
 *						  luc_reasq_t 
 * 
 * Created by:	Su Wang
 *
 * Date Created: 03/06/2002
 *
 *-----------------------------------------------------------------
 */
pkt_desc_t*						/* ptr to resulting pkt chain, else zero */
luc_Reassem(pkt_desc_t	*pdesc, int link, void** pConn)
{
	uint32			srcAddr;
	uint32			destAddr;
	ushort16		ipId;
	luc_reasq_t		*lucReasq = 0;
	struct ip		*ipPtr;
	luc_frag_t		*lucFrag=0, *lucFragPrev, *lucFragNext, *tmpFrag;
	int				overlap;
	uint32			next;
	int32			camIdx;
	pkt_desc_t		*retDesc;
	luc_stats_t		*lucStats;
	struct ipstat	*ipStats;
	uint32			chkSum = 0;
	int				noChksum = 0;

	lucStats =  &pLucCfg->linkCfg[link].rxStats;
	ipStats  = &ipstat;

	if (!isCheetah()) {
		/* we need to do check sum here, remove these when hardware ready */
		if (tpl_cksumIp(NULL, pdesc)) {
			ipStats->ips_badsum++;
			goto dropfrag;
		}
	}
	
	/* 
	 * ***********************************************************
	 *  find the corresponding reassembly queue 
	 * ***********************************************************
	 */
	ipPtr	= (struct ip*)(pdesc->pIpHdr);
	srcAddr	= (uint32)(ipPtr->ip_src.s_addr);
	destAddr= (uint32)(ipPtr->ip_dst.s_addr);
	ipId	= (ushort16)(ipPtr->ip_id);


//	LUC_REAS_PRINTF(("fragment received, src %x, dest %x id %x\n", 
//							 srcAddr, destAddr, ipId));
					   
	luc_GetReasQ(srcAddr, destAddr, ipId, &lucReasq);
	
	/* reassembe queue not found, create */
	if (!lucReasq) {
		LUC_REAS_ERR(("file %s, line %d, reasemble queue not found\n", 
						 __FILE__, __LINE__));
		goto dropfrag;
	}

	
	/* 
	 * ***********************************************************
	 * perform reassembly
	 * ***********************************************************
	 */
	
	/* create new luc_frag */
	lucFrag = (luc_frag_t*)eee_ramAlloc(sizeof(luc_frag_t));
	if (!lucFrag) {
		luc_memstat.alloc_err++;
		goto dropfrag;
	}
	luc_memstat.frag_alloc++;

	lucFrag->frag_pdesc = pdesc;
	lucFrag->frag_len   = htons(ipPtr->ip_len) /* ip len exclude IP hdr */
		- (pdesc->pTplHdr - pdesc->pIpHdr);
	lucFrag->frag_off	  = (htons(ipPtr->ip_off)&IP_OFFMASK) << 3;
	lucFrag->frag_pHdr  = pdesc->pHdr;
	lucFrag->frag_next  = 0;
		
	
//	LUC_REAS_PRINTF(("pkt find location\n"));
	
	/* find a segment which begins after the current one */
	for (lucFragPrev = 0, lucFragNext = lucReasq->reas_pkt; lucFragNext; 
		 lucFragPrev = lucFragNext, lucFragNext = lucFragPrev->frag_next)
		if (lucFragNext->frag_off > lucFrag->frag_off)
			break;
	
//	LUC_REAS_PRINTF(("pkt chk pre overlapping, prev = %x\n", lucFragPrev));
	
	/*
	 * If there is a preceding segment, it may provide some of
	 * our data already.  If so, drop the data from the incoming
	 * segment.  If it provides all of our data, drop us, otherwise
	 * stick new segment in the proper place.
	 */
	if (lucFragPrev) {
		overlap = lucFragPrev->frag_off + lucFragPrev->frag_len - lucFrag->frag_off;
		
		/* overlap covers current pkt */
		if (overlap > 0) {
			if (overlap >= lucFrag->frag_len)
				goto dropfrag;
		
			/* this is not first fragment, pHdr must equal to pTplHdr */
			if (pdesc->pHdr != pdesc->pTplHdr) {
				LUC_REAS_ERR(("file %s, line %d, pHdr, pTplHdr not matching\n", 
								 __FILE__, __LINE__));
				goto dropfrag;
			}
			
			
			/* adjust over lapping portion, pHdr == pTplHdr */
			pdesc->pHdr += overlap;
			lucFrag->frag_off += overlap;
			lucFrag->frag_len -= overlap;
		}
		
		/* queue current pkt */
		lucFrag->frag_next = lucFragPrev->frag_next;
		lucFragPrev->frag_next = lucFrag;
//		LUC_REAS_PRINTF((" prev = %x, prev next = %x\n", lucFragPrev, lucFragPrev->frag_next));
//		LUC_REAS_PRINTF((" frag = %x, frag next = %x\n", lucFrag, lucFrag->frag_next));
	}
	else {						/* insert current pkt at begining of the queue */
		lucFrag->frag_next = lucReasq->reas_pkt;
		lucReasq->reas_pkt  = lucFrag;
//		LUC_REAS_PRINTF((" lucReasq_pkt = %x\n", lucReasq->reas_pkt));
//        LUC_REAS_PRINTF((" frag = %x, frag next = %x\n", lucFrag, lucFrag->frag_next));
		
	}
	
//	LUC_REAS_PRINTF(("pkt chk post overlapping, frag %x, next %x\n", lucFrag, lucFragNext));
	
	/*
	 * While we overlap succeeding segments trim them or,
	 * if they are completely covered, dequeue them.
	 */
	for (;lucFragNext && (lucFrag->frag_off + lucFrag->frag_len > lucFragNext->frag_off);
		 lucFragNext = tmpFrag) {
		
		overlap = (lucFrag->frag_off + lucFrag->frag_len) - lucFragNext->frag_off;

		/* overlapping only portion of next pkt */
		if (overlap < lucFragNext->frag_len) {
			lucFragNext->frag_len -= overlap;
			lucFragNext->frag_off += overlap;
			lucFragNext->frag_pdesc->pHdr += overlap;

			break;
		}

		/* overlapping entire next pkt, remove next pkt */
		tmpFrag = lucFragNext->frag_next;
		lucFrag->frag_next = tmpFrag;
		PKT_FREE_PACKET(lucFragNext->frag_pdesc);
		eee_ramDealloc(lucFragNext);
		luc_memstat.frag_free++;
	}

	// LUC_REAS_PRINTF(("pkt ready to check complete, \n"));
	
	/*
	 * Check for complete reassembly.
	 */
	next = 0;
	for (lucFragPrev = NULL, lucFragNext = lucReasq->reas_pkt; lucFragNext;
		 lucFragPrev = lucFragNext, lucFragNext = lucFragNext->frag_next) {
		if (lucFragNext->frag_off != next) /* waiting for more fragments */ 
			return 0;
		
		next += lucFragNext->frag_len;
	}
	
	/* Make sure the last packet didn't have the IP_MF flag */
	ipPtr = (struct ip*)lucFragPrev->frag_pdesc->pIpHdr;
	if (ntohs(ipPtr->ip_off) & IP_MF)
		return 0;

//	LUC_REAS_PRINTF(("pkt reassembly completed\n"));
	

	/*
	 * **************************************************************
	 * Determine which path to take and act accordingly
	 * **************************************************************
	 */
	pdesc = lucReasq->reas_pkt->frag_pdesc;
//	LUC_REAS_PRINTF(("find index with  src %x dest %x sport %d dport %d, prot %d\n",
//					 pdesc->laddr, pdesc->raddr, ntohs(pdesc->lport), ntohs(pdesc->rport), pdesc->tplProt));
	camIdx = luc_simCamFind(pdesc->tplProt, 
							pdesc->laddr, 
							pdesc->raddr, 
							pdesc->lport, 
							pdesc->rport);
//	LUC_REAS_PRINTF(("index %d found \n", camIdx));
	
 
	if (camIdx < 0) {			/* slow path prcessing */

		/* send each pkt to slowpath unchanged, because pkt relay to
		 * ssc need to be unchanged. The reassembly will be performed
		 * again for slowpath pkt. But this is bearable, since there are
		 * very few of these pkts
		 */
		
		for (lucFragNext = lucReasq->reas_pkt; lucFragNext; 
			 lucFragNext = lucFragNext->frag_next) {
		
			pdesc = lucFragNext->frag_pdesc;
			
			/* recover old pHdr */
			pdesc->pHdr = lucFragNext->frag_pHdr;

			/* give pkt to slowpath */
			luc_ifSlowPath(link, pdesc);
			lucStats->slowPktCnt++;
		}

		/* remove reassemble queue for this pkt */
		luc_FreeReasQ(lucReasq, 0);
		return 0;
	}
	
	/* pkts must be to fast path, chain them up */
	*pConn = (void*)luc_camGetUsrInfo(camIdx); 
	
	/* su, 06/11/02, see if we want an UDP/TCP
	 * cksum. cksum over udp is optional.
	 */
	pdesc = lucReasq->reas_pkt->frag_pdesc;
	if (pdesc->tplProt == TPL_PROT_UDP) {
		/* udp sum field is 6 bytes from udp hdr */
		if (*(ushort16*)(pdesc->pTplHdr + 6) == 0) {
			chkSum = 0xffff;	/* no chksum needed */
			noChksum = 1;
		}
	}
	else if (pdesc->tplProt == TPL_PROT_TCP) {
		/* Do not support fragmented tcp yet */
		noChksum = 1;
		chkSum = 0;
	}
		
	retDesc = 0;
	for (lucFragNext = lucReasq->reas_pkt; lucFragNext; 
		 lucFragNext = lucFragNext->frag_next) {

//		LUC_REAS_PRINTF(("lucFrag = %xin loop\n", lucFragNext));

		if (retDesc) {
			pdesc->pNext = lucFragNext->frag_pdesc;
			pdesc = pdesc->pNext;
		}
		else {
			retDesc = pdesc = lucFragNext->frag_pdesc; /* first in chain */
			pdesc->flags |= PKT_IP_FIRST_FRAG;
		}


		/* su, 06/11/02. Accumulate cksum. 
		 * Note if for any reason IP pkts
		 * has been trimmed in hardware
		 * supported cksum, the result will 
		 * not be accurate
		 */
		if (!noChksum) {
			ushort16	tmpSum;
			if (isCheetah())
				tmpSum = lucFragNext->frag_pdesc->chksum;
			else
				tmpSum = tpl_cksumTpl(*pConn, pdesc);	
			// printf("frag cksum is %x\n", tmpSum);
			chkSum += tmpSum;
		}
		
		
		lucStats->fastPktCnt++;
		ipStats->ips_fragments++;
		ipStats->ips_total++;
	}

	while (chkSum >> 16)
		chkSum = (chkSum & 0xFFFF) + (chkSum >> 16);

	if ((chkSum & 0xFFFF) != 0xffff) { /* chksum err detected */
		/* remove reassemble queue */
		// printf("cksum error sum %x\n", chkSum);
		luc_FreeReasQ(lucReasq, 0);
		lucStats->reaChkSumErr++;
		PKT_FREE_PACKET(retDesc);
		retDesc = 0;
	}
	else
	{
		ipStats->ips_reassembled++;
		luc_FreeReasQ(lucReasq, 0);
	}
	
	return retDesc;

	/*
	 * *************************************************************
	 * care taking misc. 
	 ***************************************************************
	 */
dropfrag:	
	if (lucFrag) {
		eee_ramDealloc(lucFrag);
		luc_memstat.frag_free++;
	}
	
	PKT_FREE_PACKET(pdesc);
	ipStats->ips_fragdropped++;
	return 0;
}


