/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc-api.c
 *
 * RCS:  $Id: malloc-api.c,v 1.2 2002/11/12 17:26:07 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions for ram allocation/deallocation
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/malloc-api.h>
#if defined(__mips__)
#include <sys/cache.h>
#endif 
#include "malloc.h"
#include <sys/machine-dep.h>
#include <exports/timer-api.h>

#ifdef MEM_HISTOGRAM_BUCKET
static eeeRTC_t memHistogramBucketsSizeRTC[MEM_HIST_NBR_BUCKETS];
static boolean	first = TRUE;
#endif

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocZero()
 *
 * Description:	Allocate a flat memory area and initialize it with 0's
 *
 *				This should only be used at intitialization time!
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocZero(
	uint32	num_bytes			/* size of buffer requested */
)
{
	uchar8	* bufptr;
	void	* ra;

	ra = (void *)get_ra();

	// printf("eee_ramAllocZero: %p\n", (void*)ra);

	bufptr = eee_ramAllocRA(num_bytes, ra);

	if (bufptr)
		bzero(bufptr, num_bytes);

//printf("eee_ramAllocZero: ra %p addr %p len %d\n", (void*)ra, bufptr, num_bytes);

	return bufptr;

}	/* eee_ramAllocZero() */

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocAligned()
 *
 * Description:	Allocate some memory on a cachelin boundary
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/18/01
 *
 *-----------------------------------------------------------------
 */

void *
eee_ramAllocAligned(
	uint32	num_bytes
)
{
    void *bufptr;    
	address_t ra = get_ra();
//	printf("eee_ramAllocAligned: from %p\n", (void*)ra);

    bufptr = eee_ramAllocRA(num_bytes, (void *)ra);
    
//printf("eee_ramAllocAligned: ra %p addr %p len %d\n", (void*)ra, bufptr, num_bytes);

	return bufptr;
}	/* eee_ramAllocAligned() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void *
eee_ramAllocAlignedZero(
	uint32	num_bytes
)
{
    address_t  cl_aligned_space;
	void	* ra;

	ra = (void *)get_ra();

    cl_aligned_space = (address_t)eee_ramAllocRA(
									num_bytes + MALLOC_ALIGNED_SIZE, ra);
	if (cl_aligned_space == NULL)
		return NULL;

    cl_aligned_space += MALLOC_ALIGNED_SIZE - 1;
    cl_aligned_space &= ~(MALLOC_ALIGNED_SIZE - 1);

	if (cl_aligned_space)
		bzero((void *)cl_aligned_space, num_bytes);

//printf("eee_ramAllocAlignedZero: ra %p addr %p len %d\n", (void*)ra, (void*)cl_aligned_space, num_bytes);

    return (void *)cl_aligned_space;
}	/* eee_ramAllocAlignedZero() */

/*-----------------------------------------------------------------
 * Name:	eee_ramDealloc()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDealloc(
	void	* ptr_in
)
{
	void        * ra = (void *)get_ra();

//printf("eee_ramDealloc:   ra %p addr %p \n", (void*)ra, ptr_in);

    return eee_ramDeallocRA(ptr_in, ra);
}

/*-----------------------------------------------------------------
 * Name:	eee_ramCopy()
 *
 * Description:	Used to copy data into a RAM buffer, allocated from the
 *				eee_ramAlloc() routine.  This function will prevent
 *				corruption of neighboring buffers, since it will do
 *				some address validation.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 * MALLOC_COPY_ERROR
						 */
eee_ramCopy(
	void	* buf_ptr,			/* destination address within buffer from
								 * eee_ramAlloc() call
								 */
	void	* src_addr,			/* source address from which to copy data */
	uint32	num_bytes			/* num bytes to copy from src to buf_ptr */
)
{

	/* for now, just use memcpy */
    memcpy(buf_ptr, src_addr, num_bytes);

//printf("eee_ramCopy: dest %p src %p len %d\n", buf_ptr, src_addr, num_bytes);

	return NFX_OK;

}	/* eee_ramCopy() */

