/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc-api.h
 *
 * RCS:  $Id: malloc-api.h,v 1.5 2002/11/22 22:03:40 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API definitions for ram allocation/deallocation routines
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _MALLOC_API_H
#define _MALLOC_API_H

#define FS_USE_MALLOC_SLAB 1

// #define MM_VALIDATE
// #define MM_VERIFY_ALLOCS
// #define MM_EXPENSIVE_VALIDATE

#if defined(SSC_MGMT)
typedef uint32 malloc_size_t;
#else
typedef uint64 malloc_size_t;
#endif

#include <exports/malloc-error.h>
#include <exports/queue-api.h>
#include <bqueue.h>

#define MM_NUM_BYTE_RANGES		157		/* 16+8+24+16+32+16+16+28+1 */

typedef struct malloc_stats {
	uint64		total_flex_size;         // Bytes, total available size
	uint64      total_flex_used;         // Bytes, currently used, with overhead
	uint64		total_stack_size;        // Bytes, available size, without overhead
	uint64      total_stack_used;        // Bytes, currently used, without overhead
	uint64		total_cont_size;
	uint64      total_cont_used;
	uint32		too_big;
	uint32		no_contig;
	uint32		no_stack;
	uint32		no_flex;
	uint32		not_enough_free_buffers;
    uint64      contHistogram[3];
	uint32		malloc_size[MM_NUM_BYTE_RANGES];
} mm_stats_t;

struct mm_slab
{
	SLIST_HEAD(, mm_bufctl) sl_buffers;
	TAILQ_ENTRY(mm_slab) sl_entry;
	unsigned int sl_inuse;
	void *sl_mem;
};

#define CACHE_INMEM_SLAB 1

struct mm_cache
{
	struct mm_slab *c_current;
	nfx_spinlock_t c_spin;
	unsigned int c_num;
	unsigned int c_slabsize;
	unsigned short c_flags;
	unsigned short c_align;
	unsigned int c_objsize;
	unsigned int c_offset;		/* Offset of bufctl in object */
	TAILQ_HEAD(mm_slab_queue, mm_slab) c_slabs;
	unsigned long c_calls;
	unsigned long c_maxinuse;
	unsigned long c_inuse;
	unsigned long c_numslabs;
	unsigned long c_maxnumslabs;
	CIRCLEQ_ENTRY(mm_cache) c_entry;
	void (*c_ctor)(void *obj, uint32 size);
	void (*c_dtor)(void *obj);
};


// for performance reasons, don't use the flex_tail structure
// This structure consumes 64 bytes, and mainly protect if the
// user writes more than the allocated space (tail-magic)
// Removing this could save 10% mem, which allows more caching

// define MM_USE_FLEX_TAIL  

#define MM_FLEX_MODULO          64           // modulo (bytes) for buckets

#define MM_FLEX_OVERHEAD        0            // overhead 0, in slab allocation

/*-----------------------------------------------------------------
 *
 * extern referenced vars
 *
 *-----------------------------------------------------------------
 */

extern mm_stats_t	mm_stats;
extern void *mm_pool_start;
extern void *mm_pool_end;
extern void *mm_pool_begin;
extern unsigned long mm_pool_inuse;


/*-----------------------------------------------------------------
 * Name:	eee_ramAlloc()
 *
 * Description:	Allocate a flat memory area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAlloc(
	uint32	num_bytes			/* size of buffer requested */
);

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocRA()
 *
 * Description:	Allocate a flat memory area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocRA(
	uint32	num_bytes,			/* size of buffer requested */
	void 	* ra
);

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocZero()
 *
 * Description:	Allocate a flat memory area and initialize it with 0's
 *
 *				This should only be used at intitialization time!
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocZero(
	uint32	num_bytes			/* size of buffer requested */
);

/*-----------------------------------------------------------------
 * Name:	eee_ramDealloc()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDealloc(
	void	* ptr_in
);

/*-----------------------------------------------------------------
 * Name:	eee_ramDeallocRA()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDeallocRA(
	void	* ptr_in,
    void    * ra
);

/*-----------------------------------------------------------------
 * Name:	eee_ramCopy()
 *
 * Description:	Used to copy data into a RAM buffer, allocated from the
 *				eee_ramAlloc() routine.  This function will prevent
 *				corruption of neighboring buffers, since it will do
 *				some address validation.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 * MALLOC_COPY_ERROR
						 */
eee_ramCopy(
	void	* buf_ptr,			/* destination address within buffer from
								 * eee_ramAlloc() call
								 */
	void	* src_addr,			/* source address from which to copy data */
	uint32	num_bytes			/* num bytes to copy from src to buf_ptr */
);

/*-----------------------------------------------------------------
 * Name:	mm_isMallocBuffer()
 *
 * Description:	Checks whether or not the given address is from a malloc pool
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isMallocBuffer(
	void	* addr
);

/*-----------------------------------------------------------------
 * Name:	mm_isSizeOkay()
 *
 * Description:	Checks malloc buffer to see if the len exceeds buffer length.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isSizeOkay(
	void	* addr,
	uint32	len,
	uint32	* buf_size
);

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocAligned()
 *
 * Description:	Allocate some memory on a cachelin boundary
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/18/01
 *
 *-----------------------------------------------------------------
 */
void *
eee_ramAllocAligned(
	uint32	num_bytes
    );

void *
eee_ramAllocAlignedZero(
	uint32	num_bytes
    );

/*-----------------------------------------------------------------
 *
 * Description: Reports the total/used memory sizes
 *
 *-----------------------------------------------------------------
 */
malloc_size_t
eee_ramTotalHeapSize();

malloc_size_t
eee_ramTotalHeapAvail();

malloc_size_t
eee_ramTotalFlexSize();

malloc_size_t
eee_ramTotalFlexAvail();

malloc_size_t
eee_ramTotalFlexUsed();


extern void cache_init(struct mm_cache *cache, int flags,
					   int size, int align,
					   int slabsize, void (*ctor)(void *, uint32),
					   void (*dtor)(void *));
extern struct mm_cache *cache_create(
	uint32 flags, uint32 objsize, uint32 align,
	uint32 slabsize, void (*ctor)(void *, uint32),
	void (*dtor)(void *));
extern void * cache_alloc(struct mm_cache *cache);
extern void cache_free(struct mm_cache *cache, void *obj);
extern int cleanup_caches(void);

#if defined(MM_VERIFY_ALLOCS)
extern void * cache_alloc_ra(struct mm_cache *cache, void *ra);
extern void cache_free_ra(struct mm_cache *cache, void *obj, void *ra);
#else
#define cache_alloc_ra(c, ra) cache_alloc(c)
#define cache_free_ra(c, o, ra) cache_free(c, o)
#endif


#endif	/* _MALLOC_API_H */
