/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc-api.c
 *
 * RCS:  $Id: malloc-api.c,v 1.40 2002/11/20 23:45:21 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions for ram allocation/deallocation
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/malloc-api.h>
#if defined(__mips__)
#include <sys/cache.h>
#endif 
#include "malloc.h"
#include <sys/machine-dep.h>
#include <exports/timer-api.h>

#ifdef MEM_HISTOGRAM_BUCKET
static eeeRTC_t memHistogramBucketsSizeRTC[MEM_HIST_NBR_BUCKETS];
static boolean	first = TRUE;
#endif

/* 
 * mm_flex_verify
 *
 * Verify the flex buffer data structure
 * Only used foir debugging
 *
 */
void
mm_flex_verify(void)
{
    uint32          nbr = 0 ;
    mm_flex_head    *head;
    uchar8          *buf;
#ifdef MM_USE_FLEX_TAIL
    mm_flex_tail    *tail;
#endif
    uint32          loop;
    mm_flex_t       *flex;
    mm_flex_desc_t  *mmFlex = mmBase.flex;

    flex = mmFlex->start;

    for (loop = 0 ; loop < (MM_FLEX_MAX >> MM_FLEX_MODULO_SHIFT); loop++)
    {
        nbr = 0;
        LIST_FOREACH(head, &(flex[loop].useList), entry)
            {
#ifdef MM_USE_FLEX_TAIL
                ASSERT(head->sig == MM_HEAD_SIG);
#endif
                ASSERT(head->free_ra == 0);
                ASSERT(head->malloc_ra);

                buf = (uchar8*)(head + 1);  // Skip the flex header 
#ifdef MM_USE_FLEX_TAIL
                tail = (mm_flex_tail*)(buf + head->length);
                ASSERT(tail->sig == MM_TAIL_SIG);
#endif
                nbr++;
			}
        ASSERT(nbr == flex[loop].histogram[MM_HIST_CUR]);

        nbr = 0;
        LIST_FOREACH(head, &(flex[loop].freeList), entry)
            {
#ifdef MM_USE_FLEX_TAIL
                ASSERT(head->sig == MM_HEAD_SIG);
#endif
                ASSERT(head->free_ra);
                ASSERT(head->malloc_ra);

                buf = (uchar8*)(head + 1);  // Skip the flex header 
#ifdef MM_USE_FLEX_TAIL
                tail = (mm_flex_tail*)(buf + head->length);
                ASSERT(tail->sig == MM_TAIL_SIG);
#endif
                nbr++;
			}
        ASSERT(nbr == flex[loop].free);
    }
    
};

#if MM_FLEX_MEM_PCT > 0
/* 
 * Name:	mm_flex_garbage_collector()
 *
 * Description:	Return flex entries to the pool
 *
 */

static void
mm_flex_garbage_collector(
	uint32	num_entries)           // target nbr of entries to return
{
    uchar8          *buf;
    uint32          loop;
    mm_flex_t       *flex;
    mm_flex_desc_t  *mmFlex = mmBase.flex;
    uint32          buf_index;

    // Must be hold by caller: NFX_SPIN_LOCK(&mmBase.flexSpin);

    for (loop = 0 ; loop < num_entries; loop++)
    {
        mm_flex_head    *head;
        void            *usedTail;
#ifdef MM_USE_FLEX_TAIL
        mm_flex_tail    *tail;
#endif

        // DLL_ITERATE_BACK(head, &(mmFlex->poolList), pool)

		head = SLIST_FIRST(&mmFlex->poolList);
        if (head->free_ra)
        {
            uint32 totalSize;
            
#ifdef MM_USE_FLEX_TAIL
            ASSERT(head->sig == MM_HEAD_SIG);
#endif
#if defined(MM_VALIDATE)
            ASSERT(head->malloc_ra);
#endif			

            buf = (uchar8*)(head + 1);  // Skip the flex header 

#ifdef MM_USE_FLEX_TAIL
            tail = (mm_flex_tail*)(buf + head->length);
            ASSERT(tail->sig == MM_TAIL_SIG);
            tail += 1;
            usedTail = tail;
#else
            usedTail = buf + head->length;
#endif

            // the next entry must match the start of the available pool
            ASSERT(usedTail == mmFlex->poolStart);
                
#ifdef MM_USE_FLEX_TAIL
            totalSize = head->length + sizeof (mm_flex_head) + sizeof(mm_flex_tail);
#else
            totalSize = head->length + sizeof (mm_flex_head);
#endif
            buf_index = (head->length >> MM_FLEX_MODULO_SHIFT) - 1;

#if defined(MM_VALIDATE)
            VERIFY((head->size >> MM_FLEX_MODULO_SHIFT) == buf_index);
            VERIFY((buf_index >= 0) && (head->size < MM_FLEX_MAX));
#endif			

            mmFlex->poolStart -= totalSize;
            mmFlex->poolAvail += totalSize;

            flex = mmFlex->start;

            LIST_REMOVE(head, entry);

            SLIST_REMOVE_HEAD(&(mmFlex->poolList), pool);

#ifdef MM_VALIDATE
			flex[buf_index].free--;
#endif
        }
        else
        {
            // this entry is not free
            // NFX_SPIN_UNLOCK(&mmBase.flexSpin);
            return ;
        }
    }
    //NFX_SPIN_UNLOCK(&mmBase.flexSpin);
};
#endif



/*-----------------------------------------------------------------
 * Name:	eee_ramAllocRigidRA()
 *
 * Description:	Allocate a flat memory area.
 *              from the continuous space
 *              This avoids holding op mem in the flex/stack space 
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocRigidRA(
	uint32	num_bytes,			/* size of buffer requested */
	void 	* ra
)
{
	uint32			num_bufs;
	mm_head_t		* head;
    mm_free_node_t	* np;
	mm_alloc_node_t	* anp;
	mm_bdesc_t		* bdesc;
	uint32			* buf_end_addr;

#ifdef MEM_HISTOGRAM_BUCKET
    eeeRTC_t        startTime, delta;
    uint32          bucket, loop;
#endif

#ifdef MEM_HISTOGRAM_BUCKET
	if (first) {
		first = FALSE;
		// Setup buckets in RTC form
		for (i=0; i<EEE_NBR_POLLING_BUCKETS; i++) {
			eee_usecToRTC((uint64)memHistogramBucketsSize[i],
						  &(memHistogramBucketsSizeRTC[i]));
		}
	}
#endif

#ifdef MEM_HISTOGRAM_BUCKET
	eee_rtcStart (&startTime);
#endif
    NFX_SPIN_LOCK(&mmBase.contSpin);

	mem_histogram_cont [MM_HIST_TOT] ++ ;
	mem_histogram_cont [MM_HIST_CUR] ++ ;
	if (mem_histogram_cont [MM_HIST_CUR] > mem_histogram_cont [MM_HIST_MAX])
        mem_histogram_cont [MM_HIST_MAX] = mem_histogram_cont [MM_HIST_CUR] ;
    // 	head->length = num_bytes;

	num_bufs = (num_bytes + MM_BUFSIZE - 1) / MM_BUFSIZE;

	if (num_bufs > mm_num_free) {
		mm_stats.not_enough_free_buffers++;
        malloc_showStats();
        malloc_showOwner();
        printf("Error in ramAlloc, not enough free buffers\n");
        NFX_SPIN_UNLOCK(&mmBase.contSpin);
		return NULL;
	}

	if (num_bufs > mm_max_contig) {
		mm_stats.no_contig++;
        malloc_showStats();
        malloc_showOwner();
        printf("Error in ramAlloc, not enough contig mem\n");
        NFX_SPIN_UNLOCK(&mmBase.contSpin);
		return NULL;
	}

	np = MM_FIND_FREE_SPACE(num_bufs);

	if (!np) {
		panic("Unable to allocate buffer\n");
        NFX_SPIN_UNLOCK(&mmBase.contSpin);
		return NULL;
	}

	bdesc = np->addr;

	if (np->size == num_bufs) {
		MM_REMOVE_FREE_NODE(np);
	}
	else {
		np->addr += num_bufs;
		np->size -= num_bufs;
		MM_CHANGE_FREE_SIZE_NODE(np);
	}

	anp = MM_GET_ALLOC_NODE();
	anp->size = num_bufs;
	anp->addr = bdesc;

	MM_INSERT_IN_ALLOC_LIST(anp);

	head = (mm_head_t *)bdesc;
	head->sig = MM_HEAD_SIG;
	head->length = num_bytes;
	head->malloc_ra = ra;
	head->free_ra = 0;
	buf_end_addr = (uint32 *)&(bdesc->buf[(num_bytes + 3) & ~3]);
	*buf_end_addr = ALLOC_BUF_END;

	mm_num_free -= num_bufs;
	mm_stats.total_cont_used += num_bufs * MM_BUFSIZE;

	MM_VALIDATE_ALLOC(bdesc, num_bufs);

	if ((address_t)bdesc->buf & 0xF)
		panic("misaligned malloc buffer, %p\n", bdesc->buf);

	CACHEOP(Hit_Writeback_Inv_S, head);
	CACHEOP(Hit_Writeback_Inv_S, buf_end_addr);

#ifdef MEM_HISTOGRAM_BUCKET
    {
        uint32 bucket, loop;
        
        eee_rtcDuration(&startTime, &delta);
        for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
        {
            if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
            {
                bucket = loop;
                break;
            }
        }
        if (loop >= MEM_HIST_NBR_BUCKETS)
        {
            bucket = MEM_HIST_NBR_BUCKETS - 1;
        }
        memHistogramBuckets[MEM_HIST_BUCKET_TYPE_CONT_ALLOC][bucket] += 1;
    }
#endif

    NFX_SPIN_UNLOCK(&mmBase.contSpin);

	return bdesc->buf;

}

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocRA()
 *
 * Description:	Allocate a flat memory area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocRA(
	uint32	num_bytes,			/* size of buffer requested */
	void 	* ra
)
{
	uchar8			* buf;
#ifdef MEM_HISTOGRAM
    uint32          histogram_size ;
#endif
#ifdef MEM_HISTOGRAM_BUCKET
    eeeRTC_t        startTime, delta;
    uint32          bucket, loop;
#endif

#ifdef MEM_HISTOGRAM_BUCKET
    eee_rtc(&startTime);
#endif

    // first try the flex buffers
#if MM_FLEX_MEM_PCT > 0
    if (num_bytes < MM_FLEX_MAX)
    {
        mm_flex_desc_t *mmFlex = mmBase.flex;
		mm_flex_head    *head;
        mm_flex_t      *flex;
	    uint32          buf_index;

        NFX_SPIN_LOCK(&mmBase.flexSpin);

        flex = mmFlex->start;

        buf_index = num_bytes >> MM_FLEX_MODULO_SHIFT;
        
        if ((head = LIST_FIRST(&flex[buf_index].freeList)))
        {
#ifdef MM_USE_FLEX_TAIL
            mm_flex_tail    *tail;
#endif
            // mm_flex_verify();
            
            // DLL_REMOVE_HEAD(&flex[buf_index].freeList, head, entry);  // fastest reuse
			LIST_REMOVE(head, entry);
            buf = (uchar8*)(head + 1);  // Skip the flex header 
#ifdef MM_USE_FLEX_TAIL
            tail = (mm_flex_tail*)(buf + head->length);
            ASSERT(head->sig == MM_HEAD_SIG);
            ASSERT(tail->sig == MM_TAIL_SIG);
#endif
#ifdef MM_VALIDATE
            head->size = num_bytes;
            ASSERT(head->length == (buf_index + 1) << MM_FLEX_MODULO_SHIFT);

            LIST_INSERT_HEAD(&(flex[buf_index].useList), head, entry);
#ifdef MM_USE_FLEX_TAIL
            tail->prev_malloc_ra = head->malloc_ra;
            tail->prev_free_ra = head->free_ra;
#else
            // head->prev_malloc_ra = head->malloc_ra;
            head->prev_free_ra = head->free_ra;
#endif
#endif			

#ifdef MM_USE_FLEX_TAIL
            mm_stats.total_flex_used += (head->length + sizeof(mm_flex_head)
										 + sizeof(mm_flex_tail));
#else
            mm_stats.total_flex_used += (head->length + sizeof(mm_flex_head));
#endif

#ifdef MM_VALIDATE
            head->malloc_ra = ra;
			flex[buf_index].free--;
#endif
			
            head->free_ra = 0;

#ifdef MEM_HISTOGRAM
            flex[buf_index].histogram[MM_HIST_CUR]++;
            if (flex[buf_index].histogram[MM_HIST_CUR] > flex[buf_index].histogram[MM_HIST_MAX])
                flex[buf_index].histogram[MM_HIST_MAX] = flex[buf_index].histogram[MM_HIST_CUR];
            flex[buf_index].histogram[MM_HIST_TOT] ++;
#endif

            // ?? CACHEOP(Hit_Writeback_Inv_S, _head);
            // ?? CACHEOP(Hit_Writeback_Inv_S, _tail);

#ifdef MEM_HISTOGRAM_BUCKET
            eee_rtcDuration (&startTime, &delta);
            for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
            {
                if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
                {
                    bucket = loop;
                    break;
                }
            }
            if (loop >= MEM_HIST_NBR_BUCKETS)
            {
                bucket = MEM_HIST_NBR_BUCKETS - 1;
            }
            memHistogramBuckets[MEM_HIST_BUCKET_TYPE_FLEX_ALLOC][bucket] += 1;
#endif

            NFX_SPIN_UNLOCK(&mmBase.flexSpin);

            return (buf);
        }
        else
        {
            uint32 totalSize;
            
#ifdef MM_USE_FLEX_TAIL
            totalSize = ((buf_index + 1) << MM_FLEX_MODULO_SHIFT) + sizeof (mm_flex_head) + sizeof(mm_flex_tail);
#else
            totalSize = ((buf_index + 1) << MM_FLEX_MODULO_SHIFT) + sizeof (mm_flex_head);
#endif
            if (mmFlex->poolAvail >= totalSize)
            {
                mm_flex_head    *head;
#ifdef MM_USE_FLEX_TAIL
                mm_flex_tail    *tail;
#endif
                head = (mm_flex_head*)mmFlex->poolStart;
                buf = (uchar8*)(head + 1);  // Skip the flex header 

                head->length = (buf_index + 1) << MM_FLEX_MODULO_SHIFT;

#ifdef MM_VALIDATE
#ifdef MM_USE_FLEX_TAIL
                head->sig = MM_HEAD_SIG;
#endif
                head->size = num_bytes ;
#endif				

#ifdef MM_USE_FLEX_TAIL
                tail = (mm_flex_tail*)(buf + head->length);
                tail->sig = MM_TAIL_SIG;
#endif

                mmFlex->poolStart += totalSize;
                mmFlex->poolAvail -= totalSize;

#ifdef MM_VALIDATE
                LIST_INSERT_HEAD(&flex[buf_index].useList, head, entry);
#endif				

				SLIST_INSERT_HEAD(&(mmFlex->poolList), head, pool);

#ifdef MM_USE_FLEX_TAIL
                mm_stats.total_flex_used += (head->length + sizeof(mm_flex_head) + sizeof(mm_flex_tail));
#else
                mm_stats.total_flex_used += (head->length + sizeof(mm_flex_head));
#endif

#ifdef MM_VALIDATE
#ifdef MM_USE_FLEX_TAIL
                tail->prev_malloc_ra = 0;
                tail->prev_free_ra = 0;
#else
                // head->prev_malloc_ra = 0;
                head->prev_free_ra = 0;
#endif
                head->malloc_ra = ra;
#endif				

                head->free_ra = 0;

                // ?? CACHEOP(Hit_Writeback_Inv_S, _head);
                // ?? CACHEOP(Hit_Writeback_Inv_S, _tail);

                // mm_flex_verify();

#ifdef MEM_HISTOGRAM
                flex[buf_index].histogram[MM_HIST_CUR]++;
                if (flex[buf_index].histogram[MM_HIST_CUR] > flex[buf_index].histogram[MM_HIST_MAX])
                    flex[buf_index].histogram[MM_HIST_MAX] = flex[buf_index].histogram[MM_HIST_CUR];
                flex[buf_index].histogram[MM_HIST_TOT] ++;
#endif

#ifdef MEM_HISTOGRAM_BUCKET
				eee_rtcDuration (&startTime, &delta);
                for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
                {
					if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
                    {
                        bucket = loop;
                        break;
                    }
                }
                if (loop >= MEM_HIST_NBR_BUCKETS)
                {
                    bucket = MEM_HIST_NBR_BUCKETS - 1;
                }
                memHistogramBuckets[MEM_HIST_BUCKET_TYPE_FLEX_ALLOC][bucket] += 1;
#endif

                NFX_SPIN_UNLOCK(&mmBase.flexSpin);

                return (buf);
            }
            else
            {    
                mm_stats.no_flex++;
                flex[buf_index].unavailable++;
                // scan (limited) the bigger size flexs ....
            }
        }

        NFX_SPIN_UNLOCK(&mmBase.flexSpin);
    }
#endif

    // printf ("eee_ramALloc request (cont) size == %d, ra = %p\n", num_bytes, ra);

//	MM_INC_BYTE_RANGE(num_bytes);

#ifdef MEM_HISTOGRAM
	if ((histogram_size = num_bytes / HISTOGRAM_MODULO) >= HISTOGRAM_MAX)
        histogram_size = HISTOGRAM_MAX - 1 ;
	mem_histogram [MM_HIST_TOT] [histogram_size]++ ;
	mem_histogram [MM_HIST_CUR] [histogram_size]++ ;
	if (mem_histogram [MM_HIST_CUR] [histogram_size] > mem_histogram [MM_HIST_MAX] [histogram_size])
        mem_histogram [MM_HIST_MAX] [histogram_size] = mem_histogram [MM_HIST_CUR] [histogram_size] ;
    // 	head->length = num_bytes;
#endif

#if MM_STACK_MEM_PCT > 0
	NFX_SPIN_LOCK(&mmBase.stackSpin);

	if ((buf = MM_STACK_ALLOC_BUF(num_bytes, ra))) 
	{
#ifdef MEM_HISTOGRAM_BUCKET
		uint32 bucket, loop;

		eee_rtcDuration (&startTime, &delta);
		for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
		{
			if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
			{
				bucket = loop;
				break;
			}
		}
		if (loop >= MEM_HIST_NBR_BUCKETS)
		{
			bucket = MEM_HIST_NBR_BUCKETS - 1;
		}
		memHistogramBuckets[MEM_HIST_BUCKET_TYPE_STACK_ALLOC][bucket] += 1;
#endif

		NFX_SPIN_UNLOCK(&mmBase.stackSpin);
		return buf;
	}
	NFX_SPIN_UNLOCK(&mmBase.stackSpin);
#endif

    return (eee_ramAllocRigidRA(num_bytes, ra));

}	/* eee_ramAlloc() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	2/21/02
 *
 *-----------------------------------------------------------------
 */
void *
eee_ramAlloc(
	uint32	num_bytes
)
{
	void	* ra;

	ra = (void *)get_ra();

	return eee_ramAllocRA(num_bytes, ra);
}

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocZero()
 *
 * Description:	Allocate a flat memory area and initialize it with 0's
 *
 *				This should only be used at intitialization time!
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocZero(
	uint32	num_bytes			/* size of buffer requested */
)
{
	uchar8	* bufptr;
	void	* ra;

	ra = (void *)get_ra();

	bufptr = eee_ramAllocRA(num_bytes, ra);

	if (bufptr)
		bzero(bufptr, num_bytes);

	return bufptr;

}	/* eee_ramAllocZero() */

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocAligned()
 *
 * Description:	Allocate some memory on a cachelin boundary
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/18/01
 *
 *-----------------------------------------------------------------
 */

void *
eee_ramAllocAligned(
	uint32	num_bytes
)
{
    address_t	cl_aligned_space;
	void	* ra;

	ra = (void *)get_ra();

    cl_aligned_space = (address_t)eee_ramAllocRA(
									num_bytes + MALLOC_ALIGNED_SIZE, ra);
	if (cl_aligned_space == NULL)
		return NULL;

    cl_aligned_space += MALLOC_ALIGNED_SIZE - 1;
    cl_aligned_space &= ~(MALLOC_ALIGNED_SIZE - 1);

    return (void *)cl_aligned_space;

}	/* eee_ramAllocAligned() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void *
eee_ramAllocAlignedZero(
	uint32	num_bytes
)
{
    address_t  cl_aligned_space;
	void	* ra;

	ra = (void *)get_ra();

    cl_aligned_space = (address_t)eee_ramAllocRA(
									num_bytes + MALLOC_ALIGNED_SIZE, ra);
	if (cl_aligned_space == NULL)
		return NULL;

    cl_aligned_space += MALLOC_ALIGNED_SIZE - 1;
    cl_aligned_space &= ~(MALLOC_ALIGNED_SIZE - 1);

	if (cl_aligned_space)
		bzero((void *)cl_aligned_space, num_bytes);

    return (void *)cl_aligned_space;

}	/* eee_ramAllocAlignedZero() */

/*-----------------------------------------------------------------
 * Name:	eee_ramDeallocRA()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDeallocRA(
	void	* ptr_in,
    void    * ra
)
{
	mm_bdesc_t	* bdesc = 0;
	uint32		size;
	uint32		* buf_end_addr;
	uchar8	    * bufptr = ptr_in;
#if MM_FLEX_MEM_PCT > 0
    mm_flex_desc_t *mmFlex = mmBase.flex;
#endif
#ifdef MEM_HISTOGRAM
    uint32      histogram_size ;
#endif
#ifdef MEM_HISTOGRAM_BUCKET
    eeeRTC_t	startTime, delta;
    uint32		bucket, loop;
#endif

#ifdef MEM_HISTOGRAM_BUCKET
	if (first) {
		first = FALSE;
		// Setup buckets in RTC form
		for (i=0; i<EEE_NBR_POLLING_BUCKETS; i++) {
			eee_usecToRTC((uint64)memHistogramBucketsSize[i],
						  &(memHistogramBucketsSizeRTC[i]));
		}
	}
#endif

#ifdef MEM_HISTOGRAM_BUCKET
	eee_rtcStart (&startTime);
#endif

	/*
	 * Did this buffer come from one of the flexible buffer pool ?
	 */
#if MM_FLEX_MEM_PCT > 0
    if (((address_t)bufptr >= (address_t)mmFlex->bufStart) && 
        ((address_t)bufptr < (address_t)mmFlex->poolStart))
    {
        mm_flex_head    *head;
        uchar8          *buf;
#ifdef MM_USE_FLEX_TAIL
        mm_flex_tail    *tail;
#endif
        mm_flex_t       *flex;
	    uint32      buf_index;

        NFX_SPIN_LOCK(&mmBase.flexSpin);

        // mm_flex_verify();

        flex = mmFlex->start;

        buf = bufptr;

        head = (mm_flex_head*)buf;
        head--;
#ifdef MM_USE_FLEX_TAIL
        tail = (mm_flex_tail*)(buf + head->length);
#endif        
        buf_index = (head->length >> MM_FLEX_MODULO_SHIFT) - 1;

#ifdef MM_USE_FLEX_TAIL
        ASSERT(head->sig == MM_HEAD_SIG);
        ASSERT(tail->sig == MM_TAIL_SIG);
#endif
#ifdef MM_VALIDATE
        VERIFY((head->size >> MM_FLEX_MODULO_SHIFT) == buf_index);
        VERIFY((buf_index >= 0) && (head->size < MM_FLEX_MAX));

        VERIFY(head->free_ra == 0);        // dup memfree

        LIST_REMOVE(head, entry);
		flex[buf_index].free++;
#endif		

		LIST_INSERT_HEAD(&(flex[buf_index].freeList), head, entry);

#ifdef MM_USE_FLEX_TAIL
        mm_stats.total_flex_used -= (head->length + sizeof(mm_flex_head) + sizeof(mm_flex_tail));
#else
        mm_stats.total_flex_used -= (head->length + sizeof(mm_flex_head));
#endif

        head->free_ra = ra;

        // find the mm_flex_head of the last buffer on the list
        // and if this is free, call the garbage collector
        if (SLIST_FIRST(&mmFlex->poolList)->free_ra)
        {
            mm_flex_garbage_collector(2);
        }


#ifdef MEM_HISTOGRAM
        flex[buf_index].histogram[MM_HIST_CUR]--;
#endif		

#ifdef MEM_HISTOGRAM_BUCKET
		eee_rtcDuration (&startTime, &delta);
        for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
        {
			if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
            {
                bucket = loop;
                break;
            }
        }
        if (loop >= MEM_HIST_NBR_BUCKETS)
        {
            bucket = MEM_HIST_NBR_BUCKETS - 1;
        }
        memHistogramBuckets[MEM_HIST_BUCKET_TYPE_FLEX_DEALLOC][bucket] += 1;
#endif

        NFX_SPIN_UNLOCK(&mmBase.flexSpin);

        return (NFX_OK);
	}
#endif
    
	/*
	 * Did this buffer come from one of the malloc stacks?
	 */
#if MM_STACK_MEM_PCT > 0
    {
		uint32		i;
        NFX_SPIN_LOCK(&mmBase.stackSpin);

        for (i=0; i<MM_NUM_STACK_SIZES; i++) 
        {
            if ((bufptr >= mm_stack_buf_base[i]) &&
                (bufptr < mm_stack_buf_end[i])) 
            {
#ifdef MEM_HISTOGRAM
                {
                    mm_bdesc_t	* bdesc;

                    bdesc = (mm_bdesc_t *)(mm_stack_buf_base[i] +
                                           (((bufptr - mm_stack_buf_base[i]) /
                                             mm_stack_bdesc_size[i]) *
                                            mm_stack_bdesc_size[i]));
                    if ((histogram_size = bdesc->head.length/HISTOGRAM_MODULO) 
                        >= HISTOGRAM_MAX)
                        histogram_size = HISTOGRAM_MAX - 1 ;
                    mem_histogram [MM_HIST_CUR] [histogram_size]-- ;
                }
#endif
                MM_STACK_FREE_BUF(bufptr, i, ra);

#ifdef MEM_HISTOGRAM_BUCKET
                {
					eee_rtcDuration (&startTime, &delta);
                    for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
                    {
						if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
                        {
                            bucket = loop;
                            break;
                        }
                    }
                    if (loop >= MEM_HIST_NBR_BUCKETS)
                    {
                        bucket = MEM_HIST_NBR_BUCKETS - 1;
                    }
                    memHistogramBuckets[MEM_HIST_BUCKET_TYPE_STACK_DEALLOC][bucket] += 1;
                }
#endif
                NFX_SPIN_UNLOCK(&mmBase.stackSpin);
                return NFX_OK;
            }
        }
        NFX_SPIN_UNLOCK(&mmBase.stackSpin);
    }
#endif   
	/*
	 * If not, then remove the buffer from the malloc lists
	 */

    NFX_SPIN_LOCK(&mmBase.contSpin);

    mm_stats.contHistogram[MM_HIST_CUR]--;

	MM_DELETE_BUFFER_FROM_ALLOC_LIST(bufptr, bdesc, size);

/*
	printf("ramDealloc, bdesc->buffer = %p, bufptr = %p, tail = %p\n",
					bdesc->buf, bufptr, &((mm_bdesc_t *)(bdesc+size))->tail); 
*/

#ifdef MEM_HISTOGRAM
    if ((histogram_size = bdesc->head.length/HISTOGRAM_MODULO) >= HISTOGRAM_MAX)
		histogram_size = HISTOGRAM_MAX - 1 ;
    mem_histogram [MM_HIST_CUR] [histogram_size]-- ;
#endif
    mem_histogram_cont [MM_HIST_CUR] -- ;

	bdesc->head.free_ra = ra;

	if ((bdesc->buf <= bufptr) &&
        (bufptr < (uchar8 *)&((mm_bdesc_t *)(bdesc+(size-1)))->tail)) {
		MM_INSERT_IN_FREE_LIST(bdesc, size);
	}
	else
	{
		printf("ramDealloc, bdesc->buffer = %p, bufptr = %p, tail = %p\n",
				bdesc->buf, bufptr, &((mm_bdesc_t *)(bdesc+(size-1)))->tail); 
		panic("Invalid buffer pointer\n");
    }

	mm_num_free += size;
	mm_stats.total_cont_used -= size * MM_BUFSIZE;

	buf_end_addr = (uint32 *)&(bdesc->buf[(bdesc->head.length + 3) & ~3]);
	if (*buf_end_addr != ALLOC_BUF_END)
		panic("ramDealloc, buf_end sig incorrect, bdesc at 0x%x\n", bdesc);

	MM_VALIDATE_FREE(bdesc, size);

#ifdef MEM_HISTOGRAM_BUCKET
    {
		eee_rtcDuration (&startTime, &delta);
        for (loop = 0; loop < MEM_HIST_NBR_BUCKETS; loop++)
        {
			if (eee_rtcCompare(&delta, &(memHistogramBucketsSizeRTC[loop])) <= 0)
            {
                bucket = loop;
                break;
            }
        }
        if (loop >= MEM_HIST_NBR_BUCKETS)
        {
            bucket = MEM_HIST_NBR_BUCKETS - 1;
        }
        memHistogramBuckets[MEM_HIST_BUCKET_TYPE_CONT_DEALLOC][bucket] += 1;
    }
#endif

    NFX_SPIN_UNLOCK(&mmBase.contSpin);

	return NFX_OK;

} /* eee_ramDeallocRA */

/*-----------------------------------------------------------------
 * Name:	eee_ramDealloc()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDealloc(
	void	* ptr_in
)
{
	void        * ra = (void *)get_ra();

    return eee_ramDeallocRA(ptr_in, ra);
}

/*-----------------------------------------------------------------
 * Name:	eee_ramCopy()
 *
 * Description:	Used to copy data into a RAM buffer, allocated from the
 *				eee_ramAlloc() routine.  This function will prevent
 *				corruption of neighboring buffers, since it will do
 *				some address validation.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 * MALLOC_COPY_ERROR
						 */
eee_ramCopy(
	void	* buf_ptr,			/* destination address within buffer from
								 * eee_ramAlloc() call
								 */
	void	* src_addr,			/* source address from which to copy data */
	uint32	num_bytes			/* num bytes to copy from src to buf_ptr */
)
{

	/* for now, just use memcpy */
    memcpy(buf_ptr, src_addr, num_bytes);

	return NFX_OK;

}	/* eee_ramCopy() */

/*-----------------------------------------------------------------
 * Name:	mm_isMallocBuffer()
 *
 * Description:	Checks whether or not the given address is from a flexible or stack malloc pool
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isMallocBuffer(
	void	* addr
)
{
	uint32	i;
    mm_flex_desc_t *mmFlex = mmBase.flex;

    // can be more precise (see dealloc)
    if ((mmBase.flexSize > 0) && 
        ((address_t)addr >= (address_t)mmFlex->bufStart) && 
        ((address_t)addr < (address_t)mmFlex->poolStart))
    {
		return TRUE;
    }
    

    if (mmBase.stackSize > 0)
    {
        for (i=0; i<MM_NUM_STACK_SIZES; i++) 
        {
            if (((address_t)addr >= (address_t)mm_stack_buf_base[i]) &&
                ((address_t)addr < (address_t)mm_stack_buf_end[i]))
                return TRUE;
        }
	}

	if (((address_t)addr >= mm_bdesc_base) && ((address_t)addr < mm_bdesc_end))
		return TRUE;

	return FALSE;

}	/* mm_isMallocBuffer() */

/*-----------------------------------------------------------------
 * Name:	mm_isSizeOkay()
 *
 * Description:	Checks malloc buffer to see if the len exceeds buffer length.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isSizeOkay(
	void	* addr,
	uint32	len,
	uint32	* buf_size
)
{
	uint32	i;
	mm_bdesc_t	* bdesc;
    mm_flex_desc_t *mmFlex = mmBase.flex;

    if ((mmBase.flexSize > 0) && 
        ((address_t)addr >= (address_t)mmFlex->bufStart) && 
        ((address_t)addr < (address_t)mmFlex->poolStart))
    {
        mm_flex_head    *head;
        uchar8          *buf;
#ifdef MM_USE_FLEX_TAIL
        mm_flex_tail    *tail;
#endif
        mm_flex_t       *flex;
        uint32          buf_index ;

        NFX_SPIN_LOCK(&mmBase.flexSpin);

        flex = mmFlex->start;

        buf = addr;

        head = (mm_flex_head*)buf;
        head--;
#ifdef MM_USE_FLEX_TAIL
        tail = (mm_flex_tail*)(buf + head->length);
#endif
        
        buf_index = (head->length >> MM_FLEX_MODULO_SHIFT) - 1;

#ifdef MM_VALIDATE
        ASSERT((head->size >= 0) && (head->size < MM_FLEX_MAX));
#endif		
#ifdef MM_USE_FLEX_TAIL
        ASSERT(head->sig == MM_HEAD_SIG);
        ASSERT(tail->sig == MM_TAIL_SIG);
#endif
        ASSERT(head->free_ra == 0);        // buf still in use (not yet freed)

#ifdef MM_VALIDATE
        if (len <= (head->size))
#else
		if (len < head->length)
#endif			
        {
            NFX_SPIN_UNLOCK(&mmBase.flexSpin);
            return TRUE;
        }
        else
        {
            NFX_SPIN_UNLOCK(&mmBase.flexSpin);
#ifdef MM_VALIDATE
			*buf_size = head->size;
#else
			*buf_size = head->length - 1;
#endif
            return FALSE;
        }
    }
    
    if (mmBase.stackSize > 0)
    {
        NFX_SPIN_LOCK(&mmBase.stackSpin);
        for (i=0; i<MM_NUM_STACK_SIZES; i++) 
        {
            if (((address_t)addr >= (address_t)mm_stack_buf_base[i]) &&
                ((address_t)addr < (address_t)mm_stack_buf_end[i])) 
            {
                if (len > mm_stack_buf_size[i]) 
                {
                    printf("mm_isSizeOkay stack error, len %d > buf size %d\n",
                           len, mm_stack_buf_size[i]); 
                    NFX_SPIN_UNLOCK(&mmBase.stackSpin);
                    return 0;
                }
                else
                {
                    NFX_SPIN_UNLOCK(&mmBase.stackSpin);
                    return 1;
                }
            }
        }
        NFX_SPIN_UNLOCK(&mmBase.stackSpin);
    }
    
    NFX_SPIN_LOCK(&mmBase.contSpin);

	MM_FIND_BUFFER_IN_ALLOC_LIST(addr, bdesc);

	if (!bdesc) {
		printf("mm_isSizeOkay error, %p not valid in alloc list\n", addr);
        NFX_SPIN_UNLOCK(&mmBase.contSpin);
		return 0;
	}
	if (len > bdesc->head.length - (((address_t)addr - (address_t)bdesc->buf))) {
		printf("mm_isSizeOkay error, %d > (%d - (%p - %p))\n",
								len, bdesc->head.length, addr, bdesc->buf);
        NFX_SPIN_UNLOCK(&mmBase.contSpin);
		return 0;
	}

    NFX_SPIN_UNLOCK(&mmBase.contSpin);
	return 1;

}	/* mm_isSizeOkay() */

/*-----------------------------------------------------------------
 *
 * Description: Reports the total/used memory sizes
 *
 *-----------------------------------------------------------------
 */
malloc_size_t
eee_ramTotalHeapSize()
{
    uint64 size;

    NFX_SPIN_LOCK(&mmBase.flexSpin);
    NFX_SPIN_LOCK(&mmBase.stackSpin);
    NFX_SPIN_LOCK(&mmBase.contSpin);
    size = mm_stats.total_flex_size + mm_stats.total_stack_size + mm_stats.total_cont_size;
    NFX_SPIN_UNLOCK(&mmBase.contSpin);
    NFX_SPIN_UNLOCK(&mmBase.stackSpin);
    NFX_SPIN_UNLOCK(&mmBase.flexSpin);
	return (size);
}

malloc_size_t
eee_ramTotalHeapAvail()
{
    uint64 size;
    
    NFX_SPIN_LOCK(&mmBase.flexSpin);
    NFX_SPIN_LOCK(&mmBase.stackSpin);
    NFX_SPIN_LOCK(&mmBase.contSpin);
    size = ((mm_stats.total_flex_size - mm_stats.total_flex_used) +
            (mm_stats.total_stack_size - mm_stats.total_stack_used) +
            (mm_stats.total_cont_size - mm_stats.total_cont_used));
    NFX_SPIN_UNLOCK(&mmBase.contSpin);
    NFX_SPIN_UNLOCK(&mmBase.stackSpin);
    NFX_SPIN_UNLOCK(&mmBase.flexSpin);
	return (size);
}

malloc_size_t
eee_ramTotalFlexSize()
{
    uint64 size;

    NFX_SPIN_LOCK(&mmBase.flexSpin);
    size = mm_stats.total_flex_size;
    NFX_SPIN_UNLOCK(&mmBase.flexSpin);
	return (size);
}

malloc_size_t
eee_ramTotalFlexAvail()
{
    uint64 size;
    
    NFX_SPIN_LOCK(&mmBase.flexSpin);
    size = (mm_stats.total_flex_size - mm_stats.total_flex_used);
    NFX_SPIN_UNLOCK(&mmBase.flexSpin);
	return (size);
}

malloc_size_t
eee_ramTotalFlexUsed()
{
    uint64 size;
    
    NFX_SPIN_LOCK(&mmBase.flexSpin);
    size = mm_stats.total_flex_used;
    NFX_SPIN_UNLOCK(&mmBase.flexSpin);
	return (size);
}
