/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc-api.h
 *
 * RCS:  $Id: malloc-api.h,v 1.10 2002/10/03 23:40:51 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API definitions for ram allocation/deallocation routines
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _MALLOC_API_H
#define _MALLOC_API_H

#if defined(SSC_MGMT)
typedef uint32 malloc_size_t;
#else
typedef uint64 malloc_size_t;
#endif

#include <exports/malloc-error.h>


// for performance reasons, don't use the flex_tail structure
// This structure consumes 64 bytes, and mainly protect if the
// user writes more than the allocated space (tail-magic)
// Removing this could save 10% mem, which allows more caching

// define MM_USE_FLEX_TAIL  

#define MM_FLEX_MODULO          64           // modulo (bytes) for buckets

#ifdef MM_USE_FLEX_TAIL
#define MM_FLEX_OVERHEAD 128
#else
#define MM_FLEX_OVERHEAD 64
#endif

/*-----------------------------------------------------------------
 * Name:	eee_ramAlloc()
 *
 * Description:	Allocate a flat memory area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAlloc(
	uint32	num_bytes			/* size of buffer requested */
);

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocRA()
 *
 * Description:	Allocate a flat memory area.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocRA(
	uint32	num_bytes,			/* size of buffer requested */
	void 	* ra
);

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocRigidRA()
 *
 * Description:	Allocate a flat memory area.
 *              from the continuous space
 *              This avoids holding op mem in the flex/stack space 
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocRigidRA(
	uint32	num_bytes,			/* size of buffer requested */
	void 	* ra
);


/*-----------------------------------------------------------------
 * Name:	eee_ramAllocZero()
 *
 * Description:	Allocate a flat memory area and initialize it with 0's
 *
 *				This should only be used at intitialization time!
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void *				/* returns pointer to buffer or NULL */
eee_ramAllocZero(
	uint32	num_bytes			/* size of buffer requested */
);

/*-----------------------------------------------------------------
 * Name:	eee_ramDealloc()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDealloc(
	void	* ptr_in
);

/*-----------------------------------------------------------------
 * Name:	eee_ramDeallocRA()
 *
 * Description:	Free a buffer back into the free pools
 *
 *				The buffer pointer parameter can be any address within
 *				the allocated buffer.  The dealloc code will be smart
 *				enough to figure out the 'real' base address of the
 *				alloc'd buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 */
eee_ramDeallocRA(
	void	* ptr_in,
    void    * ra
);

/*-----------------------------------------------------------------
 * Name:	eee_ramCopy()
 *
 * Description:	Used to copy data into a RAM buffer, allocated from the
 *				eee_ramAlloc() routine.  This function will prevent
 *				corruption of neighboring buffers, since it will do
 *				some address validation.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK
						 * MALLOC_INVALID_ADDR
						 * MALLOC_COPY_ERROR
						 */
eee_ramCopy(
	void	* buf_ptr,			/* destination address within buffer from
								 * eee_ramAlloc() call
								 */
	void	* src_addr,			/* source address from which to copy data */
	uint32	num_bytes			/* num bytes to copy from src to buf_ptr */
);

/*-----------------------------------------------------------------
 * Name:	mm_isMallocBuffer()
 *
 * Description:	Checks whether or not the given address is from a malloc pool
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isMallocBuffer(
	void	* addr
);

/*-----------------------------------------------------------------
 * Name:	mm_isSizeOkay()
 *
 * Description:	Checks malloc buffer to see if the len exceeds buffer length.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/25/01
 *
 *-----------------------------------------------------------------
 */
int32					/* 0 or 1 */
mm_isSizeOkay(
	void	* addr,
	uint32	len,
	uint32	* buf_size
);

/*-----------------------------------------------------------------
 * Name:	eee_ramAllocAligned()
 *
 * Description:	Allocate some memory on a cachelin boundary
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/18/01
 *
 *-----------------------------------------------------------------
 */
void *
eee_ramAllocAligned(
	uint32	num_bytes
    );

void *
eee_ramAllocAlignedZero(
	uint32	num_bytes
    );

/*-----------------------------------------------------------------
 *
 * Description: Reports the total/used memory sizes
 *
 *-----------------------------------------------------------------
 */
malloc_size_t
eee_ramTotalHeapSize();

malloc_size_t
eee_ramTotalHeapAvail();

malloc_size_t
eee_ramTotalFlexSize();

malloc_size_t
eee_ramTotalFlexAvail();

malloc_size_t
eee_ramTotalFlexUsed();


#endif	/* _MALLOC_API_H */
