/*
 *-----------------------------------------------------------------
 *
 * Name:	malloc.c
 *
 * RCS:  $Id: malloc.c,v 1.14 2002/11/20 23:45:21 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	internal functions for embedded memory allocator
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/malloc-api.h>
#include "malloc.h"

/*
 * stack structures
 */
uchar8	* mm_stack_buf_base[MM_NUM_STACK_SIZES];
uchar8	* mm_stack_buf_end[MM_NUM_STACK_SIZES];
uint32		mm_stack_size_ratio[MM_NUM_STACK_SIZES];
uint32		mm_stack_num_bufs[MM_NUM_STACK_SIZES];
uint32		mm_stack_buf_size[MM_NUM_STACK_SIZES];
uint32		mm_stack_bdesc_size[MM_NUM_STACK_SIZES];
mm_stack_t	mm_stack[MM_NUM_STACK_SIZES];

uint32		mm_num_bufs;		/* number of buffers in pool */
uint32		mm_max_contig;
uint32		mm_num_free;
mm_stats_t	mm_stats;

mm_free_node_t	mm_free_list;
mm_alloc_node_t	mm_alloc_list;
mm_free_node_stack_t	mm_free_node_stack;
mm_alloc_node_stack_t	mm_alloc_node_stack;
mm_bdesc_t		* mm_bdesc_list;
address_t		mm_bdesc_base;
address_t		mm_bdesc_end;

mm_free_node_t	* mm_free_stack_base;
mm_alloc_node_t	* mm_alloc_stack_base;
mm_free_node_t	* mm_free_node_list;
mm_alloc_node_t	* mm_alloc_node_list;

uint32			* mm_validate;

#ifdef MEM_HISTOGRAM
int32 mem_histogram [3][HISTOGRAM_MAX] ;
#endif
int32 mem_histogram_cont [3] ;

#ifdef MEM_HISTOGRAM_BUCKET
int32 memHistogramBucketsSize[MEM_HIST_NBR_BUCKETS] = { 10,100,1000,10000,100000,100000000};
int64 memHistogramBuckets [MEM_HIST_BUCKET_TYPES][MEM_HIST_NBR_BUCKETS];
#endif


mm_base_t mmBase ;


/*-----------------------------------------------------------------
 * Name:	mallocInit()
 *
 * Description:	Initialize memory manager structures for 'Malloc' and 'Free'
 *				functions.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/4/01
 *
 *-----------------------------------------------------------------
 */
void
mallocInit(
	uchar8	* base,
	uint32	num_bytes
)
{
	uint32			bytes_per_buf, i, loop;
	mm_bdesc_t		* bdesc;
	uchar8			* ptr;
	uint32			group_size;
	uint32			num_groups;
	malloc_size_t	num_bytes_left;
	malloc_size_t	used;
    mm_flex_t       *flex;
    mm_flex_desc_t  *mmFlex;
    malloc_size_t   flexSize;
    malloc_size_t   stackSize;

/*
	printf("mallocInit, base = %p, size = 0x%x bytes\n", base, num_bytes);
*/
	mm_stats.total_flex_size = 0;
	mm_stats.total_stack_size = 0;
	mm_stats.total_cont_size = 0;
    mm_stats.no_stack = 0;
	mm_stats.contHistogram[MM_HIST_MAX] = 0;
	mm_stats.contHistogram[MM_HIST_CUR] = 0;
	mm_stats.contHistogram[MM_HIST_TOT] = 0;

#ifdef MEM_HISTOGRAM
	bzero(mem_histogram, sizeof(mem_histogram));
#endif
	bzero(mem_histogram_cont, sizeof(mem_histogram_cont));

#ifdef MEM_HISTOGRAM_BUCKET
    bzero(memHistogramBuckets, sizeof(memHistogramBuckets));
#endif



	flexSize = (num_bytes / 100 * MM_FLEX_MEM_PCT) & ~0x3F;

	// remark stack size will be bigger, see further for updates
	stackSize = (num_bytes / 100 * MM_STACK_MEM_PCT) & ~0x3F;

    // assert ((MM_FLEX_MEM_PCT + MM_STACK_MEM_PCT) < 100);

    // general layout
    mmBase.flex = base;
    mmBase.flexSize = flexSize;
    mmBase.stack = base + flexSize;
    mmBase.stackSize = stackSize;

    NFX_SPIN_INIT(&mmBase.flexSpin, SPINLOCK_LEVEL_MALLOC);
    NFX_SPIN_INIT(&mmBase.stackSpin, SPINLOCK_LEVEL_MALLOC);
    NFX_SPIN_INIT(&mmBase.contSpin, SPINLOCK_LEVEL_MALLOC);

    if (flexSize > 0)
    {
        address_t  cl_aligned_space;

        // flexible buffer layout
        mmFlex = mmBase.flex;
        mmFlex->nbrFlexLists = MM_FLEX_MAX / MM_FLEX_MODULO + 1;
        mmFlex->start = mmBase.flex + sizeof (mm_flex_desc_t);
        cl_aligned_space = (address_t)mmFlex->start + sizeof (mm_flex_t) * mmFlex->nbrFlexLists;
        cl_aligned_space = ((cl_aligned_space + (MALLOC_ALIGNED_SIZE - 1)) & ~(MALLOC_ALIGNED_SIZE - 1)); 
        mmFlex->bufStart = (void *)cl_aligned_space;
        mmFlex->poolStart = (void *)cl_aligned_space;
        mmFlex->poolEnd = mmBase.stack - 1;
        mmFlex->poolAvail = (mmFlex->poolEnd - mmFlex->poolStart);
        mm_stats.total_flex_size += mmFlex->poolAvail;
        SLIST_INIT(&(mmFlex->poolList));
    
        flex = mmFlex->start;
        for (loop = 0; loop < mmFlex->nbrFlexLists ; loop++)
        {
            LIST_INIT(&(flex[loop].freeList));
            LIST_INIT(&(flex[loop].useList));
            flex[loop].free = 0;
            flex[loop].histogram[MM_HIST_MAX] = 0;
            flex[loop].histogram[MM_HIST_CUR] = 0;
            flex[loop].histogram[MM_HIST_TOT] = 0;
            flex[loop].unavailable = 0;
        }
    }

	ptr = mmBase.stack;

    if (stackSize > 0)
    {
        /*
         * Malloc Stacks
         */
        mm_stack_buf_size[0] = MM_STACK_SIZE_1;
        mm_stack_buf_size[1] = MM_STACK_SIZE_2;
        mm_stack_buf_size[2] = MM_STACK_SIZE_3;
        mm_stack_buf_size[3] = MM_STACK_SIZE_4;
        mm_stack_bdesc_size[0] = sizeof(mm_stack_bdesc_1_t);
        mm_stack_bdesc_size[1] = sizeof(mm_stack_bdesc_2_t);
        mm_stack_bdesc_size[2] = sizeof(mm_stack_bdesc_3_t);
        mm_stack_bdesc_size[3] = sizeof(mm_stack_bdesc_4_t);
        mm_stack_size_ratio[0] = MM_STACK_NUM_1_BUFS;
        mm_stack_size_ratio[1] = MM_STACK_NUM_2_BUFS;
        mm_stack_size_ratio[2] = MM_STACK_NUM_3_BUFS;
        mm_stack_size_ratio[3] = MM_STACK_NUM_4_BUFS;

        for (i=0, group_size=0; i<MM_NUM_STACK_SIZES; i++)
            group_size += (mm_stack_bdesc_size[i] * mm_stack_size_ratio[i]);

        num_groups = (num_bytes / 100 * MM_STACK_MEM_PCT) / group_size;

        for (i=0; i<MM_NUM_STACK_SIZES; i++)
            mm_stack_num_bufs[i] = num_groups * mm_stack_size_ratio[i];

        ptr = mmBase.stack;

        for (i=0; i<MM_NUM_STACK_SIZES; i++) 
        {
            mm_stack_buf_base[i] = ptr;
            ptr += (mm_stack_num_bufs[i] * mm_stack_bdesc_size[i]);
            mm_stack_buf_end[i] = ptr;
        }

        for (i=0; i<MM_NUM_STACK_SIZES; i++) 
        {
            uchar8		* bdesc;
            mm_head_t	* head;
            mm_tail_t	* tail;

            ptr += (mm_stack_num_bufs[i] * sizeof(void *));
            mm_stack[i].avail = 0;
            mm_stack[i].top = (mm_bdesc_t **)ptr;
            for (bdesc = mm_stack_buf_base[i]; bdesc < mm_stack_buf_end[i];
                 bdesc += mm_stack_bdesc_size[i]) 
            {
                head = (mm_head_t *)bdesc;
                tail = (mm_tail_t *)(((mm_bdesc_t *)bdesc)->buf +
                                     mm_stack_buf_size[i]);
                head->sig = 0x12344321;
                head->length = 0;
                head->malloc_ra = 0;
                head->free_ra = 0;
                tail->prev_malloc_ra = 0;
                tail->prev_free_ra = 0;
                tail->sig = 0xFFEDDEFF;
                MM_STACK_FREE_BUF(bdesc, i, 0);
                mm_stats.total_stack_size += mm_stack_buf_size[i];
            }
        }
    }
    
    mmBase.cont = ptr ;

	stackSize = mmBase.cont - mmBase.stack;
    mmBase.stackSize = stackSize;
    
	//used = 	stackSize + flexSize;
	//num_bytes_left = num_bytes - used;

	used = ptr - base;
	num_bytes_left = num_bytes - used;

	/*
	 * Malloc Lists
	 */
	bytes_per_buf =	sizeof(mm_bdesc_t) +		/* head, buffer[], tail */
					sizeof(mm_free_node_t) +	/* for entry in free list */
					sizeof(mm_alloc_node_t) +	/* for entry in alloc list */
					sizeof(address_t) +		/* node ptr in free node stack */
					sizeof(address_t) +		/* node ptr in alloc node stack */
					1;		/* for validate (this is OVERKILL, but easy to do */

	
	MM_NUM_BUFS = num_bytes_left / bytes_per_buf - 1;
	mm_num_free = MM_NUM_BUFS;
	mm_max_contig = MM_NUM_BUFS;
	mm_stats.total_cont_size += MM_NUM_BUFS * MM_BUFSIZE;

	/*
	 * Set aside space for free node stack
	 */
	mm_free_stack_base = (mm_free_node_t *)ptr;
	ptr += (MM_NUM_BUFS * sizeof(mm_free_node_t *));
	mm_free_node_stack.top = (mm_free_node_t **)ptr;
	mm_free_node_stack.avail = 0;

	/*
	 * Set aside space for alloc node stack
	 */
	mm_alloc_stack_base = (mm_alloc_node_t *)ptr;
	ptr += (MM_NUM_BUFS * sizeof(mm_alloc_node_t *));
	mm_alloc_node_stack.top = (mm_alloc_node_t **)ptr;
	mm_alloc_node_stack.avail = 0;

	/*
	 * Set aside space for free nodes
	 */
	mm_free_node_list = (mm_free_node_t *)ptr;
	ptr += (MM_NUM_BUFS * sizeof(mm_free_node_t));
	for (i=0; i<MM_NUM_BUFS; i++)
		MM_FREE_FREE_NODE(mm_free_node_list+i);

	/*
	 * Set aside space for alloc nodes
	 */
	mm_alloc_node_list = (mm_alloc_node_t *)ptr;
	ptr += (MM_NUM_BUFS * sizeof(mm_alloc_node_t));
	for (i=0; i<MM_NUM_BUFS; i++)
		MM_FREE_ALLOC_NODE(mm_alloc_node_list+i);

	/*
	 * Set aside space for validate array
 	 */
	mm_validate = (uint32 *)ptr;
	ptr += (MM_NUM_BUFS + 16);		/* for alignment */
	bzero(mm_validate, MM_NUM_BUFS);
	(address_t)ptr &= ~15;						/* alignment */

	/*
	 * The rest of the space is for the bdesc list
	 */
	mm_bdesc_list = (mm_bdesc_t *)ptr;
	mm_bdesc_base = (address_t)ptr;
	ptr += (MM_NUM_BUFS * sizeof(mm_bdesc_t));
	mm_bdesc_end = (address_t)ptr;

	if ((address_t)ptr > (address_t)base + num_bytes)
		panic("mallocInit, not enough space (%p > %p + 0x%x)\n",
													ptr, base, num_bytes);
/*
	else
		printf("mallocInit, used up to %p, base = %p, num_bytes = 0x%x (top = %p)\n", 
	            ptr, base, num_bytes, (address_t)base + num_bytes);
*/

	/*
	 * Initialize the list head nodes
	 */
	bzero(&mm_free_list, sizeof(mm_free_node_t));
	mm_free_list.Snext =
	mm_free_list.Sprev =
	mm_free_list.Anext =
	mm_free_list.Aprev = &mm_free_list;
	bzero(&mm_alloc_list, sizeof(mm_alloc_node_t));

	/*
	 * Initialize the bdesc list
	 */
	bdesc = mm_bdesc_list;
	for (i=0; i<MM_NUM_BUFS; i++, bdesc++) {
		bdesc->head.sig = MM_HEAD_SIG;
		bdesc->head.malloc_ra = 0;
		bdesc->tail.sig = MM_TAIL_SIG;
	}

	/*
	 * Add all buffers to free list
	 */
	MM_INSERT_IN_FREE_LIST(mm_bdesc_list, MM_NUM_BUFS);

	mm_stats.total_flex_used = 0;
	mm_stats.total_stack_used = 0;
	mm_stats.total_cont_used = 0;
    
}	/* mallocInit() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */

