/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        mnt-proc.c
 *
 * RCS:  $Id: mnt-proc.c,v 1.40 2002/12/06 01:37:25 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: 
 *      Contains MOUNT Procedures.
 * Created by: Joseph Vu 
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/nfs-api.h>

#define MNT_DBG_ON
extern uint32 trace_enable;
#define     MNT_DBG_PRINTF  if (trace_enable) printf

#define     MNT_CONN_DBG_PRINTF  


mnt_globalCfg_t	*mnt_globalCfgp;

/*
 *Externs
 */
extern int32   
nfs_addCBToList(nfp_list_t *nodeListp, void *CtrlBlkp, uint32 nodeType);
extern int32 
nfs_delFileSysCBFromList(nfp_fileSysCBPtr_t fileSysCBp, 
        nfp_clientCBPtr_t clientCBp); 
/*-----------------------------------------------------------------
 * Name:   	mnt_procRxNullReq 
 *
 * Description:
 * NULL procedure does no work.  It is made available to allow
 * server response testing and timing.
 * 
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procRxNullReq(mnt_request_t    *mntReqp )
{
	int32					error;			
	int32					txMsgLen;
	char                    *mnt_msgp;
	pkt_desc_t				*tplXmitPktDescp;
	char					*txMsgp;
	nfp_clientCBPtr_t		clientCBp;

	MNT_DBG_PRINTF ("mnt_procRxNullReq: Enter\n");
	mnt_globalCfgp->mnt_rxPktCounters.mnt_null++;
	
	/*
	 *Get start of message pointer
	 */
	mnt_msgp = mntReqp->procStart;
	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;
	error = NFSERR_SERVERFAULT;
	if ( mntReqp->rpcCallHdrp &&
		(mntReqp->rpcCallHdrp->progVer < MNT_VER3)) 
			error = NFSERR_ACCES;
 
	/*
	 *Send back response to the client
	 */
	tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);
	if (tplXmitPktDescp == NULL)
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (error);
	}
	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;
	/*
	 *Build RPC reply message to send to the client
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
					RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

    tplXmitPktDescp->hdrSize = txMsgLen;
	if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
	{
		/*
		 *Set request state to in-progress
		 */
		mntReqp->reqState = MNT_REQ_REPLIED;
		mnt_globalCfgp->mnt_xmtPktCounters.mnt_null++;
	}

	else
	{
		//free previous alloc TPL pkt.
		//Calling function shall mnt_freeAll();
		TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
        return (NFX_ERR);
	}

	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:	mnt_procRxMountReq 
 *
 * Description:
 *      Process  Mount request received from client.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procRxMountReq (mnt_request_t *mntReqp)
{
	int32					error;
	uint32					reqXid;
	uint32					procNum;
	int32					length;
	ushort16				pathLen;
	char					pathName[FS_PATH_MAX];
	char					*pathNamep;
	char					*mnt_msgp;
	nfp_clientCBPtr_t		clientCBp;
	nfp_fileSysCBPtr_t		fileSysCBp;

	/*
	 *Increment RX stat
	 */
	mnt_globalCfgp->mnt_rxPktCounters.mnt_addMnt++;
	reqXid = mntReqp->rpcCallHdrp->rpc_xid;
	procNum = mntReqp->rpcCallHdrp->rpc_proc;

	//Set default error code
	//NFS V3: NFSERR_SERVERFAULT
	//NFS V2: NFSERR_ACCES
	error = NFSERR_SERVERFAULT;
	if ( mntReqp->rpcCallHdrp &&
		(mntReqp->rpcCallHdrp->progVer < MNT_VER3))
		error = NFSERR_ACCES;
	
	/*
	 *Get start of message pointer
	 */
	mnt_msgp = mntReqp->procStart;
	/*
	 *Retrieve pathname
	 */
	pathLen = 0;
	memset(&pathName[0], 0, FS_PATH_MAX);
	length = nfs_extractString(mnt_msgp, &pathName[0], &pathLen, FS_PATH_MAX);
	if (length < 0)
	{
		nfs_errMsg(reqXid, procNum, length);
		return (length);
	}

	pathNamep = &pathName[0];
	/*
	 *Eat leading '/' 
	 */
	if (*pathNamep && *pathNamep == '/')
	{
		pathNamep += 1;	
		pathLen -= 1;
	}
	
	fileSysCBp = nfp_getFileSysByName (pathNamep, pathLen); 
	if (fileSysCBp == NULL) 
	{
		fileSysCBp = nfp_allocFileSysCB();
		if (fileSysCBp == NULL)
		{
			nfs_errMsg(reqXid, procNum, 0);
			return (error);
		}
		memcpy(&fileSysCBp->nfp_name[0], pathNamep, pathLen);
		fileSysCBp->nfp_nameLen = pathLen;
		fileSysCBp->nfp_flag |= NFP_SHARE_INIT_STATE;
	}

	fileSysCBp->atime = nfs_utilGetTimeOfDayInSecs();
	mntReqp->fileSysCBp =  (void *)fileSysCBp;

	/* Shaun 12/04/2002
	 *
	 * Send mount reply from cached file handle.
	 */ 
	if (fileSysCBp->nfp_flag & NFP_ENTRY_CACHE_FHANDLE)
	{
		mnt_pktCounters_t   *txPktStatp;
		pkt_desc_t          *tplXmitPktDescp;
		char 				*txMsgp;
		int32 				txMsgLen;
		int32 				length;

	    tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);
	    if (tplXmitPktDescp == NULL)
	    {
			nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
			mnt_freeAll(mntReqp);
			return (error);
		}

	    txMsgp = (char *)tplXmitPktDescp->pHdr;
	    txMsgLen = 0;
		/*
		 *Build RPC reply message to send to the client
		 */
		length    = nfs_msgRpcReply(txMsgp, 
									mntReqp->rpcCallHdrp->rpc_xid,
									RPC_REPLY_MSG_ACCEPTED, 
									RPC_REPLY_SUCCESS);
		txMsgp   += length;
		txMsgLen += length;

		txPktStatp = &mnt_globalCfgp->mnt_xmtPktCounters;

		txPktStatp->mnt_addMnt++;
	
		/* 
		 * Store Status 
		 */
		*(uint32 *)txMsgp = 0; 

		length    = sizeof(uint32);
		txMsgp   += length;
		txMsgLen += length;

		/* 
		 * Store root handle 
		 */
		length    = nfs_storeFhandle(mntReqp->rpcCallHdrp->progVer,
									 &fileSysCBp->nfp_fileHandle, txMsgp);
		txMsgp   += length;
		txMsgLen += length;

		/* 
		 * Store auth flavor
		 */
		length    = mnt_storeAuthFlavor(txMsgp);
		txMsgp   += length;
		txMsgLen += length;

		/*
		 * Set packet desc hdr size
		 */
		tplXmitPktDescp->hdrSize = txMsgLen;

		if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
		{
			/*
			 * Set request state to in-progress
		 	 */
			mntReqp->reqState =  MNT_REQ_REPLIED;
		}
		else
		{
			TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
			mnt_freeAll(mntReqp);
		}

		return (NFS_OK);
	}

	/*
	 *Add File System Control Block (fileSysCB) to the client 
	 *Control Block file system list
	 */	
	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;

	if (nfs_findFileSysCBNode (&clientCBp->nfp_fileSysList,fileSysCBp) == NULL)
	{
		//Add new entry
		if (nfs_addCBToList(&clientCBp->nfp_fileSysList, 
					(void *)fileSysCBp, NFP_FILESYSCB_NODE)  != NFX_OK)
		{
			nfs_errMsg(reqXid, procNum, 0);
			return (error);
		}
	}
	/*
	 *Call function in SSC to check client access permission 
	 *based on client IP address, UID/GID for the requested volume.
	 *The SSC module returns the volId for the filesystem if access is allow
	 */

	if (mnt_getMountdInfoReq(mntReqp, MNT_INFO_REQ) == NFX_ERR)
	{
		nfs_errMsg(reqXid, procNum, 0);
		return (error);
	}

	
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:        mnt_procRxShowMountReq
 *
 * Description:
 *	Returns a list of clients and the exported filesystems they have
 *	mounted.  
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procRxShowMountReq(mnt_request_t    *mntReqp	/*Mount request control block*/
)
{
	nfp_nodePtr_t		nodep;
	nfp_list_t 			*nodeListp;
	nfp_clientCBPtr_t	clientCBp;
	pkt_desc_t          *tplXmitPktDescp;
	char                *txMsgp;
	nfp_fileSysCBPtr_t	fileSysCBp;
	char				pathName[FS_PATH_MAX + 1];
	char				*pathNamep;
//	int32				count;
	int32				length;
	int32               txMsgLen;
//	int32				cli_count;
	int32				error;
	int32				clientCount;
	uint32				nodeCount;

	MNT_DBG_PRINTF ("mnt_procRxShowMountReq:  Enter\n");

	/*
	 *Increment RX stat
	 */
	mnt_globalCfgp->mnt_rxPktCounters.mnt_dump++;

	error = NFSERR_SERVERFAULT;
	if ( mntReqp->rpcCallHdrp &&
		(mntReqp->rpcCallHdrp->progVer < MNT_VER3))
		error = NFSERR_ACCES;


	/*
	 *Allocate packet Desc for transmission
	 */
	tplXmitPktDescp = TPL_ALLOC_PACKET((void *)mntReqp->tplConnp, 0);
	if (tplXmitPktDescp == NULL)
	{
		//panic ("mnt_procRxShowMountReq: Failure to alloc TPL packet!\n");
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (error);
	}

	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;

	/*
	 *Build RPC reply message to send to the client
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
				RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	clientCount = 0;	

	for (clientCBp= (nfp_clientCBPtr_t)nfp_globalCfgp->clientCbCfgList.listBasep;
		((clientCount < NFP_MAX_CLIENT) && (clientCBp != NULL)); 
		clientCount++, clientCBp++)
	{
		if ((nfp_validateClientCB(clientCBp) != NFX_OK) ||
			  (!(clientCBp->nfp_flag & NFP_ENTRY_INUSED)) ||
				(!clientCBp->nfp_ipAddr) )	
			continue;
		
		nodeListp = &clientCBp->nfp_fileSysList;
		nodeCount = nodeListp->nfp_reqCnt;
    	for (nodep = nodeListp->nfp_headp; 
				(nodep != NULL && nodeCount > 0);
        	nodep = nodep->nfp_nextp, nodeCount--)
    	{
        	fileSysCBp = (nfp_fileSysCBPtr_t)nodep->nfp_entryCBp;
			if ( (nfp_validateFileSysCB(fileSysCBp) != NFX_OK) ||
				(!(fileSysCBp->nfp_flag & NFP_ENTRY_INUSED )) )
				continue;

			//store entry value follows flag
			length = nfs_storeFieldPresentFLag(txMsgp, TRUE);
			//update msp pointer & length
			txMsgp += length;
			txMsgLen += length;

			//store client ip Address
			length = nfs_storeIpAddr(txMsgp, clientCBp->nfp_ipAddr);
			//update msp pointer & length
			txMsgp += length;
			txMsgLen += length;

			//store pathname in format /vol/testdir  
			memset(pathName, 0, sizeof(pathName));
			//Insert '/'
			pathNamep = pathName;
			sprintf(pathNamep, "%c",'/');
			pathNamep += 1;
			memcpy(pathNamep, fileSysCBp->nfp_name, 
							fileSysCBp->nfp_nameLen);
			//pathNameLen = fileSysCBp->nfp_nameLen + 1;
			MNT_DBG_PRINTF("%s: name: %s len: %d\n",
					__FUNCTION__,pathName, fileSysCBp->nfp_nameLen + 1);
		
			/*
		 	 *Store directory pathname
		 	 */
			length = nfs_storeString(txMsgp, pathName, 
						(fileSysCBp->nfp_nameLen + 1), FS_PATH_MAX);
			//update msp pointer & length
			txMsgp += length;
			txMsgLen += length;
		}
	}

	//Set value end of flag to FALSE
	length = nfs_storeFieldPresentFLag(txMsgp, FALSE);	
	txMsgp += length;
	txMsgLen += length;

	mnt_globalCfgp->mnt_xmtPktCounters.mnt_dump++;

	tplXmitPktDescp->hdrSize = txMsgLen;
	if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
	{
		/*
		 *Set request state to in-progress
		 */
		mntReqp->reqState = MNT_REQ_REPLIED;
		mnt_globalCfgp->mnt_xmtPktCounters.mnt_dump++;
	}
	else
	{
		//free previous alloc TPL pkt.
		//Calling function shall mnt_freeAll();
		TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
		return (NFX_ERR);
	}


	MNT_DBG_PRINTF ("mnt_procRxShowMountReq:  Exit\n");

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_procRxUnmountReq 
 *
 * Description:
 *	
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procRxUnmountReq (mnt_request_t    *mntReqp		/*Mount request control block*/
)
{
	int32					txMsgLen;
	uint32                  reqXid;
	uint32                  procNum;
	int32                   length;
	ushort16                pathLen;
	char                    pathName[FS_PATH_MAX];
	char                    *mnt_msgp;
	nfp_fileSysCBPtr_t		fileSysCBp;
	pkt_desc_t				*tplXmitPktDescp;
	char					*txMsgp;
	nfp_clientCBPtr_t		clientCBp;
	char					*pathNamep;

	//printf ("mnt_procRxUnmountReq Enter\n");
	mnt_globalCfgp->mnt_rxPktCounters.mnt_umnt++;
	reqXid = mntReqp->rpcCallHdrp->rpc_xid;
	procNum = mntReqp->rpcCallHdrp->rpc_proc;
	
	/*
	 *Get start of message pointer
	 */
	mnt_msgp = mntReqp->procStart;
	/*
	 *Retrieve pathname
	 */
	pathLen = 0;
	length = nfs_extractString(mnt_msgp, &pathName[0], &pathLen, FS_PATH_MAX);
	if (length < 0)
	{
		nfs_errMsg(reqXid, procNum, length);
		return (length);
	}
	//printf ("mnt_procRxUnmountReq 2:\n");
	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;

	/*
	 *Send back UMNT response to the client
	 */
	tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);

	if (tplXmitPktDescp != NULL)
	{
		txMsgp = (char *)tplXmitPktDescp->pHdr;
		txMsgLen = 0;
		/*
	 	 *Build RPC reply message to send to the client
	 	 */
		txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
					RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
		txMsgp += txMsgLen;

		mnt_globalCfgp->mnt_xmtPktCounters.mnt_umnt++;
    	tplXmitPktDescp->hdrSize = txMsgLen;
		if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
		{
			/*
		 	 *Set request state to in-progress
		 	 */
			mntReqp->reqState = MNT_REQ_REPLIED;
		}
		else
		{
			//free previous alloc TPL pkt.
			//Calling function shall mnt_freeAll();
			TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
			return (NFX_ERR);
		}
	}
	else
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
	}

	//printf("mnt_procRxUnmountReq 3: pathName: %s len: %d\n", &pathName[0], pathLen);
    pathNamep = &pathName[0];
    /*
     *Eat leading '/'
     */
    if (*pathNamep && *pathNamep == '/')
    {
        pathNamep += 1;
        pathLen -= 1;
    }

	fileSysCBp = nfp_getFileSysByName (pathNamep, pathLen); 
#ifdef SHAUN

	/*Do proper clean up when unmount
	 *-Remove the fileSys volume from the client list.
	 *-Removes all TPL Pconn connections 
	 */
	mnt_pntCleanUp(fileSysCBp, clientCBp);

	mnt_delTplPconn(clientCBp);

	nfs_clearCache();
#endif
	if (fileSysCBp)
		nfs_delFileSysCBFromList(fileSysCBp, clientCBp);
	MNT_DBG_PRINTF ("mnt_procRxUnmountReq Exit\n");

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_procRxUnmountAllReq 
 *
 * Description:
 *      
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procRxUnmountAllReq (mnt_request_t    *mntReqp	/*Mount request control block*/
)
{
	int32					error;			
	int32					txMsgLen;
	char                    *mnt_msgp;
	pkt_desc_t				*tplXmitPktDescp;
	char					*txMsgp;
	nfp_clientCBPtr_t		clientCBp;

	MNT_DBG_PRINTF ("mnt_procRxNullReq: Enter\n");
	mnt_globalCfgp->mnt_rxPktCounters.mnt_umntAll++;
	
	/*
	 *Get start of message pointer
	 */
	mnt_msgp = mntReqp->procStart;
	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;
	error = NFSERR_SERVERFAULT;
	if ( mntReqp->rpcCallHdrp &&
		(mntReqp->rpcCallHdrp->progVer < MNT_VER3)) 
			error = NFSERR_ACCES;
 
	/*
	 *Send back response to the client
	 */
	tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);
	if (tplXmitPktDescp == NULL)
	{
		//panic ("mnt_procRxNullReq: Failure to allocate TPL packet!\n");
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (error);
	}
	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;
	/*
	 *Build RPC reply message to send to the client
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
					RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

    tplXmitPktDescp->hdrSize = txMsgLen;
	if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
	{
		/*
		 *Set request state to in-progress
		 */
		mntReqp->reqState = MNT_REQ_REPLIED;
		mnt_globalCfgp->mnt_xmtPktCounters.mnt_umntAll++;
	}

	else
	{
		//free previous alloc TPL pkt.
		//Calling function shall mnt_freeAll();
		TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
        return (NFX_ERR);
	}

	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   mnt_storeExportList    	
 *
 * Description:
 *	Stores export list entry into message buffer.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        8/28/02
 *
 *-----------------------------------------------------------------
 */
int32 mnt_storeExportList(
	char *txMsgp, 
	nfp_fileSysCBPtr_t fileSysCBp,
	int32 includeClientSpecificExports)
{
	int32					length;
	int32					totalLen;
	static char				pathName[FS_PATH_MAX+1];
	char					*pathNamep;
	int32					pathNameLen;
	int32					i;
	uint32					clientAddr;


	if (fileSysCBp == NULL)
		return (0);	
	totalLen = 0;
	length = nfs_storeFieldPresentFLag(txMsgp, TRUE);
	txMsgp += length;
	totalLen += length;

	MNT_DBG_PRINTF ("mnt_storeExportList:name: %s len: %d\n", 
			fileSysCBp->nfp_name, fileSysCBp->nfp_nameLen);

	memset(pathName, 0, sizeof(pathName));
	//Insert '/'
	pathNamep = pathName;
	sprintf(pathNamep, "%c",'/');
	pathNamep += 1;
	memcpy(pathNamep, fileSysCBp->nfp_name, fileSysCBp->nfp_nameLen);
	pathNameLen = fileSysCBp->nfp_nameLen + 1;
	MNT_DBG_PRINTF("mnt_storeExportList 2: name: %s len: %d\n",
		pathName, pathNameLen);
		
	/*
	 *Store directory pathname
	 */
	length = nfs_storeString(txMsgp, pathName, 
					pathNameLen, FS_PATH_MAX);

	txMsgp += length;
	totalLen += length;
	if (includeClientSpecificExports)
	{
		for (i = 0; i < NFP_NFS_MAX_EXP_MASK_ENTRIES; i++)
		{
			clientAddr = fileSysCBp->clientSpecificExpMask[i].clientAddr;
			if (clientAddr)
			{
				MNT_DBG_PRINTF("mnt_storeExportList 3: export addr: %08x\n",clientAddr );
				//Set value field flag
				length = nfs_storeFieldPresentFLag(txMsgp, TRUE);
			txMsgp += length;
			totalLen += length;

			length = nfs_storeIpAddr(txMsgp, clientAddr);
			txMsgp += length;
			totalLen += length;
			}				
		}

	}
	/*
	 *Set group empty field
	 */
	length = nfs_storeFieldPresentFLag(txMsgp, FALSE);
	txMsgp += length;
	totalLen += length;

	return (totalLen);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_procExportListReq 
 *
 * Description:
 *	Return Export List.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procExportListReq (mnt_request_t    *mntReqp
)
{
	nfp_clientCBPtr_t		clientCBp;
	pkt_desc_t              *tplXmitPktDescp;
	nfp_fileSysCBPtr_t      fileSysCBp;
	char					*txMsgp;
	int32					txMsgLen;
	int32					length;
	int32           		error;
	int32					count;

	//Set default error code
	//Use NFSERR_SERVERFAULT for VER3
	//otherwise, use  NFSERR_ACCES
	error = NFSERR_SERVERFAULT;
	if ( mntReqp->rpcCallHdrp &&
		(mntReqp->rpcCallHdrp->progVer < MNT_VER3))
		error = NFSERR_ACCES;
	mnt_globalCfgp->mnt_rxPktCounters.mnt_export++;

	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;
	/*
	 *Allocate packet Desc for transmission
	 */
	tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);
	if (tplXmitPktDescp == NULL)
	{
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (error);
	}

	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;

	/*
	 *Build RPC reply message to send to the client
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
				RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	count = 0;

	for (fileSysCBp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
		((count < NFP_MAX_FILESYS) && (fileSysCBp != NULL));
		count++, fileSysCBp++)
	{
		//check if MNT point is in used
		if ( (!(fileSysCBp->nfp_flag & NFP_ENTRY_INUSED)) || 
			 (fileSysCBp->nfp_nameLen == 0)	)
			continue;

		if (fileSysCBp->defaultExpMask)
		{
			//Allow export for "all" on this filesystem
			length = mnt_storeExportList (txMsgp, fileSysCBp, 0);
			txMsgp += length;
			txMsgLen += length;
		}
		if (fileSysCBp->numOfClientSpecificMasks)
		{
			//Include specific group export on this filesystem
			length = mnt_storeExportList (txMsgp, fileSysCBp, 1);
			txMsgp += length;
			txMsgLen += length;
		}
	}

	length = nfs_storeFieldPresentFLag(txMsgp, FALSE);
	txMsgp += length;
	txMsgLen += length;

	mnt_globalCfgp->mnt_xmtPktCounters.mnt_export++;
	tplXmitPktDescp->hdrSize = txMsgLen;
	if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
	{
		/*
		 *Set request state to in-progress
		 */
		mntReqp->reqState = MNT_REQ_REPLIED;
		mnt_globalCfgp->mnt_xmtPktCounters.mnt_null++;
	}

	else
	{
		//free previous alloc TPL pkt.
		//Calling function shall mnt_freeAll();
		TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
        return (NFX_ERR);
	}
	return (NFX_OK);


}

int32
mnt_procExportAllReq (mnt_request_t    *mntReqp		/*Mount request control block*/
)
{


	return (NFX_OK);
}

int32
mnt_procPathconfReq (mnt_request_t    *mntReqp		/*Mount request control block*/
)
{



	return (NFX_OK);
}
