/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        mnt-tpl.c
 *
 * RCS:  $Id: mnt-tpl.c,v 1.39 2002/05/28 19:29:04 agile Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *  Contain MOUNT/Transport Layer Protocol (TPL) interface functions.
 *
 * Created by: Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/luc-api.h>
#include <exports/nfs-api.h>
extern uint32 trace_enable;
#define		MNT_DBG_PRINTF	if (trace_enable) printf

#define		MNT_CONN_DBG_PRINTF	
#define		MOUNT_PORT			2087 

/*
 *MOUNT procedure
 */
int32 (*mnt_procs[MNT_MAX_PROC]) (mnt_request_t    *mntReqp) =
{
    mnt_procRxNullReq,      	/*0: NULL*/
    mnt_procRxMountReq,     	/*1: Mount request*/
    mnt_procRxShowMountReq,     /*2: Showmount/DUMP: return mount entries*/
    mnt_procRxUnmountReq,       /*3: Unmount request*/
    mnt_procRxUnmountAllReq,    /*4: remove all mount entries*/
    mnt_procExportListReq,      /*5: request for Export list*/
    mnt_procExportAllReq,       /*6:  same as Export (procedure 5). Obsolete in ver 3*/
    mnt_procPathconfReq     	/*7: POSIX pathconf information*/
};

/*-----------------------------------------------------------------
 * Name:   	mnt_tplAddConnInd 
 *
 * Description:
 *	Add connection indication from the transport layer to
 *	adds a new connection.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_tplAddConnInd(tpl_bindCb_t	*pBind,		/*ptr to Bind CB */
				  void 			*connIdp,   /*ptr to newly added connection*/
				  void 			**appHdlpp, /*ptr to correspond app. handle*/
				  uint32 		raddr,      /*remote IP address*/
				  ushort16 		rport       /*remote port*/
)
{
	nfp_clientCBPtr_t   clientCBp;
	uint32          	connId;

#if defined(NFX_SMP) && defined(NFP_TXRX) && defined(SMP_TXRX_VERIFY)
	if (CM_IS_TXRX_NCPU)
		panic("%s: called in wrong cpu \n", __FUNCTION__);
#endif /* NFX_SMP */

	MNT_CONN_DBG_PRINTF("mnt_tplAddConnInd: Enter\n");
	connId = (uint32)connIdp;		

	/*
	 *Check if client control block exist 
	 */
	clientCBp = nfp_getClientCBbyAddr(raddr);
	if (clientCBp != NULL)
	{
		MNT_DBG_PRINTF("mnt_tplAddConnInd: Duplicate Conn. clientCBp:%p connId: %08x!\n",
					clientCBp, connIdp);
		/*
		 *Use same client control block
		 */
		*appHdlpp = clientCBp;
		return (NFX_OK);
	}

	/*
	 *Allocate Mount Client Control  Block
	 */
	clientCBp = nfp_allocClientCB();
	if (clientCBp == NULL)
		return (TPL_ERR_NO_RSRC);

	MNT_DBG_PRINTF("mnt_tplAddConnInd: raddr: %08x rport: %d\n", raddr, rport);
	/*
	 *Fill in handle upon return to TPL
	 */
	*appHdlpp = clientCBp;
	clientCBp->nfp_connId = connId;
	clientCBp->nfp_ipAddr = raddr;
	clientCBp->proto = NFP_PROTO_MNT;


	MNT_DBG_PRINTF("mnt_tplAddConnInd: Exit-clientCBp: %p connId: %08x\n", 
					clientCBp,clientCBp->nfp_connId);
	MNT_DBG_PRINTF("mnt_tplAddConnInd Exit raddr: %08x\n", clientCBp->nfp_ipAddr);

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:    mnt_tplDelConnInd
 *
 * Description:
 *  Delete connection indication from the transport layer to delete
 *  an existing connection.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_tplDelConnInd (void *appHdlp, void *connIdp)
{
	nfp_clientCBPtr_t   clientCBp;
	int32				index;

#if defined(NFX_SMP) && defined(NFP_TXRX) && defined(SMP_TXRX_VERIFY)
	if (CM_IS_TXRX_NCPU)
		panic("%s: called in wrong cpu \n", __FUNCTION__);
#endif /* NFX_SMP */

	MNT_DBG_PRINTF("mnt_tplDelConnInd: Enter\n");
	clientCBp= (nfp_clientCBPtr_t)appHdlp;
	if (nfp_validateClientCB(clientCBp) != NFX_OK)
	{
		panic("mnt_tplDelConnInd-Error:  bad client control block!\n");
	}

#ifdef	NOTUSED
	index = nfs_findPconnOnClientList(clientCBp, connIdp);
	if (index != NFX_ERR)
	{
		MNT_CONN_DBG_PRINTF("mnt_tplDelConnInd: del conn  %p\n", clientCBp->connId[index]);
		clientCBp->connId[index] = NULL;
	}
#endif	//NOTUSED

	MNT_DBG_PRINTF("mnt_tplDelConnInd: Exit\n");

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mnt_tplRcvCheckPkt 
 *
 * Description:
 *	Check received MOUNT packet.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_tplRcvCheckPkt(rpcCallHdr_t *rpcCallp)
{
	/*
	 *Check RPC version number
	 */
	if (rpcCallp->rpc_rpcVer != RPC_VER2)
	{
		MNT_DBG_PRINTF("mnt_tplRcvCheckPkt: RPC mistmatch Ver: %d\n", rpcCallp->rpc_rpcVer);		
		return (NFS_INT_RPC_MISMATCH);
	}

	/*
	 *Check program version number
	 */
	if ((rpcCallp->progVer <  MNT_VER1) ||
		(rpcCallp->progVer > MNT_VER3) )
	{
		MNT_DBG_PRINTF("mnt_tplRcvCheckPkt: program ver mistmatch %d\n", rpcCallp->progVer);		
//		return (NFS_RPC_PROG_MISMATCH);
		return (NFS_INT_RPC_PROG_MISMATCH);
	}
	/*
	 *Check procedure number
	 */
	if ( (rpcCallp->rpc_proc < MNT_MIN_PROC) ||
		(rpcCallp->rpc_proc > MNT_MAX_PROC) )
	{

		return (NFS_INT_RPC_PROC_UNAVAIL);
	}

    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mnt_tplRcvProcessPkt 
 *
 * Description:
 * 	Process packet received from the transport layer. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32		/* MOUNT error code*/
mnt_tplRcvProcessPkt(nfp_clientCBPtr_t clientCBp,
				rpcCallHdr_t	*pktRpcCallHdrp,
				pkt_desc_t *descp)
{
	uint32				rpcHdrSize;
	rpcCallHdr_t		*newRpcCallp;
	mnt_request_t		*mntReqp;
	nfp_requestCBPtr_t  reqCBp;
	int32				error;
	int32				ret;
	int32				nfs_outstandingReqs;
	int32				nfs_outstandingReqsSndToFs;

	MNT_DBG_PRINTF("mnt_tplRcvProcessPkt Enter\n");
	if (nfs_globalCfgp->nfs_maxReqsThreshold > 0)
	{
		//Check NFS receive request queue threshold
		//or outstanding request between NFS server
		//the file system
		//Drop received NFS request if file system is too
		//busy
		nfs_outstandingReqs =
			(NFP_MAX_REQUEST - nfp_globalCfgp->reqCbCfgList.freeEntriesCnt);
		nfs_outstandingReqsSndToFs =
			(nfs_globalCfgp->nfs_totalNumOfReqXmtToFs -
			nfs_globalCfgp->nfs_totalNumOfRespRcvdFromFs);

		if ( (nfs_outstandingReqs > nfs_globalCfgp->nfs_maxReqsThreshold) ||
			(nfs_outstandingReqsSndToFs > nfs_globalCfgp->nfs_maxReqsThreshold) )
		{
			mnt_globalCfgp->mnt_debugCounters.mnt_maxReqsDropPktErr++;
			return (NFX_ERR);
		}
    }

	newRpcCallp =  rpc_processAuth(pktRpcCallHdrp, &rpcHdrSize);
	if (newRpcCallp == NULL)
		return (NFX_ERR);
	/*
	 *For MOUNT, check if UID = 0 (Super-user)
	 */
#ifdef	NOTUSED
	if (newRpcCallp->authFlavor.authSys.userId != 0)
		return (NFX_ERR);
#endif
	mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt4++;
	/*
	 *Allocate request control block
	 */ 
	reqCBp = nfp_allocReqCB ();
	if (reqCBp == NULL)
	{
		if (newRpcCallp != NULL)
		{
			eee_ramDealloc(newRpcCallp);
		}

		return (NFX_ERR);
	}

	mnt_globalCfgp->mnt_debugCounters.mnt_allocReqCB++;
	mntReqp = &reqCBp->protocol.mnt_req;
	mntReqp->rpcCallHdrp = newRpcCallp;
	mntReqp->reqCBp = (void *)reqCBp;
	mntReqp->reqState = MNT_REQ_INPROG;
	mntReqp->clientCBp = (void *)clientCBp;
	mntReqp->tplConnp = descp->pTplConn;
	mntReqp->nfsReqp = NULL;
	clientCBp->nfp_localIpAddr = descp->laddr;	


#ifdef	NOTUSED
	/*
	 *Add reqCB to the Client Control Block
	 */
	if (nfs_addCBToList(&clientCBp->nfp_reqList, (void *)mntReqp, NFP_REQCB_NODE)  != NFX_OK)
		return (NFX_ERR);

#endif	/*NOTUSED*/

	MNT_DBG_PRINTF("mnt_tplRcvProcessPkt: clientCBp: %p client Addr: %08x local Addr: 0x%x\n",
				clientCBp, clientCBp->nfp_ipAddr, clientCBp->nfp_localIpAddr);

	mntReqp->procStart = (char *)pktRpcCallHdrp + rpcHdrSize;
	/*
	 *Call MOUNT protocol procedure
	 */
	error = (*(mnt_procs[pktRpcCallHdrp->rpc_proc]))(mntReqp);		
	if (error != NFX_OK)
		goto mnt_tplRcvProcessPktErr;
	return (NFX_OK);

mnt_tplRcvProcessPktErr:
	//nfp_deallocReqCB(reqCBp);
	mnt_freeAll(mntReqp);
	return (error);
}

/*
 *-------------------------------------------------------------------
 * Name:        mnt_tplRcvPktInd
 *
 * Description:
 *      Receive packet indication from the transport layer for processing.
 *
 * Created by:  Joseph Vu
 *
 * Date Create: 7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_tplRcvPktInd(void *appHdlp, 
			pkt_desc_t *descp)
{
	nfp_clientCBPtr_t	clientCBp;
	rpcCallHdr_t		*rpcCallp;
	int32				error;
	int32				rc;

#if defined(NFX_SMP) && defined(NFP_TXRX) && defined(SMP_TXRX_VERIFY)
	if (CM_IS_TXRX_NCPU)
		panic("%s: called in wrong cpu \n", __FUNCTION__);
#endif /* NFX_SMP */

	MNT_DBG_PRINTF("mnt_tplRcvPktInd Enter\n");
	if ((descp == NULL)	||
		(descp->pHdr == NULL) )
		return (TPL_FREE_PKT);

	/* 
	 * Perform Net to Host conversion.
	 */
    xdr_swap(descp->pHdr, (sizeof(rpcCallHdr_t)-sizeof(rpc_authSysInfo_t)), 0);

	error = NFX_OK;

	if (trace_enable != 0)
		mnt_rxDecodePkt(descp);

	rpcCallp = (rpcCallHdr_t *)descp->pHdr;

	clientCBp = (nfp_clientCBPtr_t)appHdlp;
	if (nfp_validateClientCB(clientCBp) != NFX_OK)
	{
		panic("mnt_tplRcvPktInd-Error:  bad client control block!\n");
	}

	rc = mnt_tplRcvCheckPkt(rpcCallp);
	if (rc != NFX_OK)
	{
		error = mnt_mapToNFSError(rc);
		goto mnt_rcvPktIndErr;
	}

	error = mnt_tplRcvProcessPkt(clientCBp, rpcCallp, descp);
	if (error == NFS_OK)
	{
		MNT_DBG_PRINTF("mnt_tplRcvPktInd 2: status OK. return TPL_FREE_PKT\n");
		return (TPL_FREE_PKT);
	}

mnt_rcvPktIndErr:

        /*
         *Handle error.  Return a error reply &
         *free all alloc memory
         *Call function nfs_msgSendErrResp() to send a error reply
         *pkt to client
         */
        //nfs_errMsg(rpcCallp->rpc_xid, rpcCallp->rpc_proc, error );
		/*
        nfs_msgSendErrResp(clientCBp, rpcCallp->rpc_xid, error);
		*/
		mnt_globalCfgp->mnt_debugCounters.mnt_rcvPktIndErr++;
        return (TPL_FREE_PKT);
}
/*-----------------------------------------------------------------
 * Name: mnt_tplXmtCmpInd
 *
 * Description:
 *  Transmit packet complete indication from the transport layer.
 *  
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_tplXmitCmpInd(void *appHdlp,         /*No application handle for MOUNT*/
            pkt_desc_t *descp)
{
	mnt_request_t			*mntReqp;

#if defined(NFX_SMP) && defined(NFP_TXRX) && defined(SMP_TXRX_VERIFY)
	if (CM_IS_TXRX_NCPU)
		panic("%s: called in wrong cpu \n", __FUNCTION__);
#endif /* NFX_SMP */

	MNT_DBG_PRINTF("mnt_tplXmitCmpInd 1: appHdlp: %08x descp: %08x\n", appHdlp,descp);
	mnt_globalCfgp->mnt_debugCounters.mnt_tplXmtPktCmpInd++;
	if (descp == NULL)
	{
		panic ("mnt_tplXmitCmpInd: bad descp!\n");
	}
	mntReqp = descp->appInfo;
	if ( (mntReqp == NULL) || 
			(nfp_validateMntReqCB(mntReqp) != NFX_OK ) )
	{
		/*
		 *Error:  No corresponding reqCB with the packet descriptor
		 */	
		panic ("mnt_tplXmitCmpInd: Bad request block!\n");	
//		return (TPL_FREE_PKT);
	}

	MNT_DBG_PRINTF("mnt_tplXmitCmpInd 2: Found mntReqp: %08x\n", mntReqp);
	/* 
	 * Shaun 03/25/02
	 * Fix for #273, workaround for #314
	 * Remove MOUNT related tpl connections.
	 */
	if ((mntReqp->rpcCallHdrp->rpc_proc == MNT_PROC_UMNT) ||
		(mntReqp->rpcCallHdrp->rpc_proc == MNT_PROC_MNT)  ||
		(mntReqp->rpcCallHdrp->rpc_proc == MNT_PROC_NULL))
	{
	
		/* 
		 * Shaun 09/21/01
		 * Delete TPL Connection after unmount.
		 * TPL will upcall MOUNT with mnt_tplDelConnInd().
		 * Return TPL_HOLD_PKT since tpl_delConnReq()
		 * will free packets back to PKT.
		 */
		if (descp->tplProt == TPL_PROT_UDP)
			tpl_delConnReq(descp->pTplConn);

//		mnt_freeAll(mntReqp);	
//		return (TPL_HOLD_PKT);	
	}
	mnt_freeAll(mntReqp);	
	return (TPL_FREE_PKT);	
}

/*-----------------------------------------------------------------
 * Name: mnt_tplBindReq
 *
 * Description:
 * Binds NFS to Transport Layer (UDP/TCP).
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_bindReq(uint32 ipaddr)
{
	int32			bindOpt;
	ushort16        protocol;
	tpl_bindCb_t    bindCB;
	tpl_bindCb_t    *bindCBp;
	tpl_bindCb_t    *tplRspBindp;

#if defined(NFX_SMP) && defined(NFP_TXRX) && defined(SMP_TXRX_VERIFY)
	if (CM_IS_TXRX_NCPU)
		panic("%s: called in wrong cpu \n", __FUNCTION__);
#endif /* NFX_SMP */

	protocol = TPL_PROT_UDP;
	while(1)
	{
		/*
		 * Bind MOUNT server on UDP or TCP
		 */
	    memset(&bindCB, 0, sizeof(tpl_bindCb_t));

		bindOpt = TPL_BIND_SERVER | TPL_BIND_USE_RPC | TPL_BIND_PKTCNT_SM;
		bindCBp = &bindCB;
		bindCBp->laddr         = ipaddr;
		bindCBp->lport         = htons(MOUNT_PORT);	
		bindCBp->tplProt       = protocol;
		bindCBp->ul_addConnInd = mnt_tplAddConnInd;
		bindCBp->ul_delConnInd = mnt_tplDelConnInd;
		bindCBp->ul_rcvPktInd  = mnt_tplRcvPktInd;
		bindCBp->ul_xmtCmpInd  = mnt_tplXmitCmpInd;

		if (tpl_bindReq(bindCBp, &tplRspBindp, bindOpt) != NFX_OK)
		{
			printf("mnt_bindReq: bind failed\n");
			return (NFX_ERR);
		}
		if (protocol == TPL_PROT_TCP)
			break;

		protocol = TPL_PROT_TCP;
	}

	return (NFX_OK);
}

