/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        mount.c
 *
 * RCS:  $Id: mount.c,v 1.63 2002/10/22 01:30:38 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *
 * Created by: Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/nfs-api.h>

extern uint32 trace_enable;
#define     MNT_DBG_PRINTF  if (trace_enable) printf
#define		MNT_CONN_DBG_PRINTF	


/*             
 *Extern declaration
 */ 
extern void mnt_msgSscRcv( eee_desc_t *eDescp );
extern rpcCallHdr_t *rpc_createHdrInfo(rpcCallHdr_t  *pktRpcCallHdrp);
extern uint64		mnt_rxAuthTypeErr;

/*--------------------------------------------------------------
 * Name:       	mnt_mapToNFSError	
 *
 * Description:
 *     	Map our internal error number to NFS error number. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *--------------------------------------------------------------
 */
int32 mnt_mapToNFSError(int32 internal_error)
{
	switch (internal_error)
	{
		case NFS_INT_RPC_MISMATCH:
			return (NFS_RPC_MISMATCH);

		case NFS_INT_RPC_AUTH_ERROR:
			return (NFS_RPC_AUTH_ERROR);

		case NFS_INT_RPC_PROG_UNAVAIL:
			return (NFS_RPC_PROG_UNAVAIL);

		case NFS_INT_RPC_PROG_MISMATCH:
			return (NFS_RPC_PROG_MISMATCH);

		case NFS_INT_RPC_PROC_UNAVAIL:
			return (NFS_RPC_PROC_UNAVAIL);

		case NFS_INT_RPC_PROG_BAD_ARGS:
			return (NFS_RPC_PROG_BAD_ARGS);

		case RPC_INT_PROG_SYS_ERR:
			return (RPC_PROG_SYS_ERR);

		default:
			return (NFX_ERR);
	}
	
}

/*-----------------------------------------------------------------
 * Name:       	mnt_getVolumeName 
 *
 * Description:
 * Extract volume name from the input pathname. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_getVolumeName(char *volNamep, int32 *volNameLen,
		char *pathNamep, int32 pathNameLen)
{
	int32	count = 0;
	
	while ((pathNameLen--) && (pathNamep) && (*pathNamep != '/') )
	{
		*volNamep++ = *pathNamep++;
		count++;
	}
	*volNameLen = count;
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_storeAuthFlavor 
 *
 * Description:
 *      Store authentication flavors support by our server.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */

int32   /*
         *length = number of bytes of the field
         */
mnt_storeAuthFlavor(char *msgp)
{
    uint32      *msg32p;

    msg32p = (uint32 *)msgp;
	*msg32p++ = 2;				/*2 flavors supported*/

    /*
     *We only support AUTH_NONE & AUTH_SYS  
     */
    *msg32p++ = NFS_RPC_AUTH_NONE;
    *msg32p++ = NFS_PRC_AUTH_SYS;

    return (12);
}


/*-----------------------------------------------------------------
 * Name:        mnt_init
 *
 * Description:
 *		MOUNT initialize function on the TX/RX CPU.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_init()
{
   
    MNT_DBG_PRINTF("mnt_init: Enter\n");

    mnt_globalCfgp = (mnt_globalCfg_t *)eee_ramAlloc(sizeof(mnt_globalCfg_t) );
    if (mnt_globalCfgp == NULL)
        panic ("mnt_init: Failure to alloc MOUNT NFS Global Cfg memory!\n");
    memset (mnt_globalCfgp, 0, sizeof(mnt_globalCfg_t));

	mnt_globalCfgp->evmAppId = eee_getApplicationId("evm");	
	if (mnt_globalCfgp->evmAppId == EEE_ERR_INVALID_APP_NAME)
		panic ("mnt_init: Failure to get EVM App Id!\n");

	printf("mnt_init: EVM EEE AppId: %d\n", mnt_globalCfgp->evmAppId);	

	mnt_globalCfgp->mountdAppId = eee_getApplicationId(MOUNTD_APP_NAME);	
	if (mnt_globalCfgp->mountdAppId == EEE_ERR_INVALID_APP_NAME)
		panic ("mnt_init: Failure to get MOUNTD App Id!\n");
	printf("mnt_init: MOUNTD AppId: %d\n", mnt_globalCfgp->mountdAppId);	

	//Register MNT Server EEE
	mnt_globalCfgp->mnt_srvrAppId = eee_allocateApplicationId(MNT_SRVR_APP_NAME);
	if (mnt_globalCfgp->mnt_srvrAppId == EEE_ERR_INVALID_APP_NAME)
	{
		printf("mnt_init: eee_allocateApplicationId() failed!\n");
		luc_panic();
	}

	if (eee_registerReceiveFunc(mnt_globalCfgp->mnt_srvrAppId, 
							mnt_msgRcv, NULL) != NFX_OK) 
	{
		printf("mnt_init: Failure to register received function for MNT Srvr!\n");
		luc_panic();
	}

	/* 
	 * Shaun: 03/13 Bind Mount server.
	 */
    if (mnt_bindReq( 0 ) != NFX_OK)
    {
        panic( "mnt_init: TPL BIND failed\n");
    }

	//create timer to get the NFS share entries from MOUND SSC
	mnt_globalCfgp->tmrId = eee_timerCreate( MNT_GET_SHARES_TIME, 
					EEE_TIMER_ONCE);

	if (mnt_globalCfgp->tmrId == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		printf("%s: Failure to create get NFS shares timer!\n",
					   	__FUNCTION__);
	}

	if (eee_timerStart(mnt_globalCfgp->tmrId, mnt_getNfsShareHandler, 
					0, 0, 0, 0) != NFX_OK)
	{
		printf("%s: Failure to start get NFS shares timer!\n",
					   	__FUNCTION__);
	}
	rpc_init();

    MNT_DBG_PRINTF("mnt_init: Exit\n");

    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:	mnt_findReqCBByDesc 
 *
 * Description:
 *  Find a Request Control Block with a correponding packet descriptor.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
mnt_request_t *mnt_findReqCBByDesc(
		nfp_clientCBPtr_t	clientCBp,    	/*Client control block*/
		pkt_desc_t *descp)   				/*Pkt descriptor*/
{
	mnt_request_t			*mntReqp;
	nfp_list_t              *nodeListp;	
	pkt_desc_t              *reqPktDescp;
	nfp_nodePtr_t           reqp;
	int32					reqCnt;

	nodeListp = &clientCBp->nfp_reqList;
	reqCnt = nodeListp->nfp_reqCnt;
	MNT_DBG_PRINTF("nfs_findReqCBByDesc 1: clientCBp: %p descp: %p reqCnt:%d\n",
					clientCBp, descp, reqCnt);

	for (reqp = nodeListp->nfp_headp; (reqp != NULL && reqCnt > 0);
			reqp = reqp->nfp_nextp, reqCnt--)
	{
		mntReqp = (mnt_request_t *)reqp->nfp_entryCBp;
		reqPktDescp = (pkt_desc_t *)mntReqp->pktDescp;
		if (reqPktDescp == descp)
		{
			/*
			 *return found request entry
			 */
			return (mntReqp);
		}
	}
    
    return (NULL);
}
/*-----------------------------------------------------------------
 * Name:   	mnt_findReqCBNode 
 *
 * Description:
 *  Find a Mount Request Control Block with a correponding 
 *	XID & Procedure number.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t mnt_findReqCBNode(nfp_list_t *reqListp, mnt_request_t *mntReqp)
{
	uint32 			count;
	mnt_request_t 	*entryReqp;
	nfp_nodePtr_t	nodep;

	count = reqListp->nfp_reqCnt;

	for (nodep = reqListp->nfp_headp; (nodep != NULL && count > 0);
		nodep = nodep->nfp_nextp, count--)
	{
		entryReqp = (mnt_request_t *)nodep->nfp_entryCBp;
		if ( (mntReqp->rpcCallHdrp == NULL) || (entryReqp->rpcCallHdrp == NULL))
			continue;
		if ((mntReqp->rpcCallHdrp->rpc_xid == entryReqp->rpcCallHdrp->rpc_xid) &&
			(mntReqp->rpcCallHdrp->rpc_proc == entryReqp->rpcCallHdrp->rpc_proc))
		{
			return (nodep);
		}
	}	
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name:   	mnt_findFileSysCB 
 * 
 * Description:
 *  Find a File System Control Block with a correponding
 *  filehandle.
 *  
 * Created by:  Joseph Vu
 * 
 * Date Created:        7/28/01
 * 
 *-----------------------------------------------------------------
 */
nfp_fileSysCBPtr_t mnt_findFileSysCB(
	nfp_clientCBPtr_t clientCBp,		/*MOUNT Client control block*/
	fs_Handle *fs_handlep)				/*filehandle*/
{
	nfp_nodePtr_t           nodep;
	int32                   nodeCnt;
	nfp_list_t              *nodeListp;
	nfp_fileSysCBPtr_t 		fileSysCBp;	

	if ((clientCBp == NULL) ||
	 	(nfp_validateClientCB(clientCBp) != NFX_OK) ||
		(fs_handlep == NULL))
		return (NULL);	
	nodeListp = &clientCBp->nfp_fileSysList;
	nodeCnt = nodeListp->nfp_reqCnt;
	MNT_DBG_PRINTF("mnt_findFileSysCB: nodeCnt: %d\n", nodeCnt);
	for (nodep = nodeListp->nfp_headp; (nodep != NULL && nodeCnt > 0);
		nodep = nodep->nfp_nextp, nodeCnt--)
	{
		if (nfp_validateNode(nodep) != NFX_OK)
			break;
		fileSysCBp = (nfp_fileSysCBPtr_t)nodep->nfp_entryCBp;
		if (nfp_validateFileSysCB(fileSysCBp) != NFX_OK)
			break;
		if (!(fileSysCBp->nfp_flag & NFP_ENTRY_INUSED))
			continue;
		if ((fileSysCBp->nfp_fileHandle.volId.whole_thing == 
							fs_handlep->volId.whole_thing) &&
			(fileSysCBp->nfp_fileHandle.mntPnt == 
				fs_handlep->mntPnt) )
			return (fileSysCBp);
    }

    return (NULL);
}

/*--------------------------------------------------------------------------
 * Name:  	mnt_delReqCBFromList 
 *
 * Description:
 *		Delete a MOUNT request control block from the client control block list.
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-------------------------------------------------------------------------
 */
int32 mnt_delReqCBFromList(mnt_request_t *mntReqp)
{
	nfp_nodePtr_t		delNodep;
	nfp_list_t          *reqListp;
	nfp_clientCBPtr_t   clientCBp;

	MNT_DBG_PRINTF("mnt_delReqCBFromList: Enter\n");

	clientCBp = mntReqp->clientCBp;
	if (clientCBp == NULL)
	{
		MNT_DBG_PRINTF("mnt_delReqCBFromList: No Client Control Block!\n");
		return (NFX_ERR);
	}

	reqListp = &clientCBp->nfp_reqList;

	delNodep = 	mnt_findReqCBNode(reqListp, mntReqp);

	MNT_DBG_PRINTF("mnt_delReqCBFromList: delNodep: %p\n", delNodep);
	return (nfs_delNode(reqListp, delNodep));

	MNT_DBG_PRINTF("mnt_delReqCBFromList: Exit\n");
}

/*-----------------------------------------------------------------
 * Name:        nfs_delFileSysCB
 *
 * Description:
 *      Delete and de-allocate all memory associate with this File
 *      System Control BLock (FileSysCB) list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
		 *NFX_OK or NFX_ERR
		 */
mnt_delFileSysCB(nfp_list_t *fileSysListp)
{
	nfp_nodePtr_t                   currp;
	nfp_fileSysCBPtr_t              fileSysCBp;
	uint32                          count;
  
	if (fileSysListp->nfp_headp == NULL)
		return (NFX_ERR);

	count = fileSysListp->nfp_reqCnt;

	for (currp = fileSysListp->nfp_headp; ((currp != NULL) && (count > 0));
		currp = currp->nfp_nextp, count--)
	{
		fileSysCBp = (nfp_fileSysCBPtr_t)currp->nfp_entryCBp;
		if  (fileSysCBp == NULL)
			continue;
		nfp_deallocFileSysCB(fileSysCBp);
	}
	return (NFX_OK);
}

/*
 *------------------------------------------------------------------------------
 * Function Name:      	mnt_freeAll 
 *
 * Description:
 *  This function called after received a transmit complete indicdation from
 *  Transport Layer (TPL) to de-allocate all previous allocated memory for this request.
 *
 * Created by:          Joseph Vu
 *
 * Created:             7/28/2000
 *
 *------------------------------------------------------------------------------
 */
int32
mnt_freeAll(mnt_request_t *mntReqp)
{
	MNT_DBG_PRINTF("mnt_freeAll: Enter\n");
	if (mntReqp == NULL)
		return (NFX_ERR);
	
	if (mntReqp->rpcCallHdrp != NULL)
	{
		if (eee_ramDealloc(mntReqp->rpcCallHdrp) != NFX_OK )
			MNT_DBG_PRINTF("mnt_freeAll: eee_ramDealloc failed!\n");			
	    /* 
	     * Shaun 03/14/02
	     * Clear rpcCallHdrp. 
	     */
		mntReqp->rpcCallHdrp = NULL;
		mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt4++;
	}
#ifdef	NOTUSED
	/*
	 *Delete request for Client control block list
	 */
	if (mnt_delReqCBFromList(mntReqp) != NFX_OK)
	{
		MNT_DBG_PRINTF("mnt_freeAll: mnt_delReqCBFromList failed!\n");
	}
#endif	/*NOTUSED*/
	/*
	 *De-alloc request control block
	 */
	if (nfp_deallocReqCB((nfp_requestCBPtr_t)mntReqp->reqCBp) != NFX_OK)
	{
		MNT_DBG_PRINTF("mnt_freeAll: nfp_deallocReqCB failed!\n");
	}

	if (mntReqp && mntReqp->tplConnp)
		mntReqp->tplConnp = NULL;	
			
	mnt_globalCfgp->mnt_debugCounters.mnt_freeReqCB++;
	MNT_DBG_PRINTF("mnt_freeAll: Exit\n");

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        mnt_delReqCB
 *
 * Description:
 *      Delete and de-allocate all memory associate with this Request
 *  Control BLock (RCB) list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32 mnt_delReqCB(nfp_list_t *reqListp)
{
	nfp_nodePtr_t                   currp;
	nfp_requestCBPtr_t              reqCBp;
	mnt_request_t					*mntReqp;
	int32							count;

	if (reqListp->nfp_headp == NULL)
		return (NFX_ERR);
	count = reqListp->nfp_reqCnt;

	for (currp = reqListp->nfp_headp; ((currp != NULL) && (count > 0));
			currp = currp->nfp_nextp, count--)
	{
		mntReqp = (mnt_request_t *)currp->nfp_entryCBp;
		if (mntReqp == NULL)
			continue;
		/*
		 *Free all memory corresponding to this request
		 */
		mnt_freeAll(mntReqp);
		reqCBp = (nfp_requestCBPtr_t)mntReqp->reqCBp;
		if (reqCBp == NULL)
			continue;
		nfp_deallocReqCB(reqCBp);
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_getMountdInfoReq	
 *
 * Description:
 *  Send mount info request to MOUNTD on SSC to check
 *  NFS shares if client access is allowed for a requested
 *  file system.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        8/03/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_getMountdInfoReq(mnt_request_t *mntReqp,
		int32 msgType)
{
	mnt_infoReq_t       *mntInfoReqp;
	nfp_fileSysCBPtr_t	fileSysCBp;
	nfp_clientCBPtr_t	clientCBp;
    uchar8              *buf;
    uint32              allocSize;
    uint32              destPort;
    uint32              srcPort;
	int32				rc;

	MNT_DBG_PRINTF("mnt_getMountdInfoReq Enter\n");
	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;
	fileSysCBp = (nfp_fileSysCBPtr_t)mntReqp->fileSysCBp;
	MNT_DBG_PRINTF("mnt_getMountdInfoReq 1: fileSysCBp: %p mntReqp: %p clientCBp: %p\n",
					fileSysCBp, mntReqp, clientCBp);
	if ( (fileSysCBp == NULL) || (mntReqp == NULL) )
		return (NFX_ERR);

	allocSize = sizeof(mnt_infoReq_t) + fileSysCBp->nfp_nameLen + 1;
	buf = (uchar8 * )eee_ramAlloc(allocSize);
	if (buf == NULL)
	{
		printf("mnt_getMountdInfoReq 2: Failure to alloc buffer!\n");
		return (NFX_ERR);
	}

	mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt2++;	
	memset(buf, 0, allocSize);
	mntInfoReqp = (mnt_infoReq_t *)buf;

	mntInfoReqp->msgHdr.msgType = msgType;
	mntInfoReqp->msgHdr.private = (uint64)((uintptr_t)mntReqp);

	MNT_DBG_PRINTF("mnt_getMountdInfoReq 3: private %08llx\n", 
					mntInfoReqp->msgHdr.private);
	mntInfoReqp->ipAddr = 0; 
	mntInfoReqp->localIpAddr = 0;
	mntInfoReqp->uid = 0;
	mntInfoReqp->gid = 0;
	if (clientCBp)
	{
		mntInfoReqp->ipAddr = clientCBp->nfp_ipAddr; 
		mntInfoReqp->localIpAddr = clientCBp->nfp_localIpAddr;
	}
	if (mntReqp->rpcCallHdrp)
	{
		mntInfoReqp->uid = mntReqp->rpcCallHdrp->authFlavor.authSys.userId;
		mntInfoReqp->gid = mntReqp->rpcCallHdrp->authFlavor.authSys.groupId;
	}	
	mntInfoReqp->nameLen = fileSysCBp->nfp_nameLen;
	memset(&mntInfoReqp->pathName[0], 0, sizeof(mntInfoReqp->pathName));
	memcpy(&mntInfoReqp->pathName[0], 
		&fileSysCBp->nfp_name[0], fileSysCBp->nfp_nameLen);

	/*
	printf("mnt_getMountdInfoReq:  pathname: %s Len: %d\n", 
			&mntInfoReqp->pathName[0], mntInfoReqp->nameLen);
	*/
					
	/*
	 *Send MOUNT info request to MOUNTD on SSC
	 */
	srcPort = PACK_UNICAST_PORT_ADDRESS(
										nfxMySlotId,
										nfxMyCPUId,
										mnt_globalCfgp->mnt_srvrAppId,
										EEE_MGMT,
										NFX_COS_HIGH,
										0);
	destPort = PACK_UNICAST_PORT_ADDRESS(
										0,
										0,         /*MOUNTD on SSC*/
										mnt_globalCfgp->mountdAppId,
										EEE_MGMT,
										NFX_COS_HIGH,
										0);

	MNT_DBG_PRINTF("MNT TXRX : src app_id (%d) src slot (%d) src cpu (%d)\n",
						EEE_GET_APP_ID(srcPort),
						EEE_GET_SLOT_NUM(srcPort),
						EEE_GET_DEST_CPU(srcPort));

	MNT_DBG_PRINTF("ipAddr: %08x\n", mntInfoReqp->ipAddr);
	rc = eee_sendMessage((uchar8 *)buf, allocSize, destPort, srcPort );

	eee_ramDealloc(buf);
	mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt2++;	
	if (rc != NFX_OK)
	{
		printf("MNT-ERR: Send req to MOUNTD, rc: %d\n", rc);
		return (NFX_ERR);
	}

	mnt_globalCfgp->mnt_debugCounters.mnt_xmtReqByNameToSsc++;
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   mnt_sendAdminRespToMountd     		
 *
 * Description:
 *	Send Admin response to MOUNTD on SSC.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        8/03/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_sendAdminRespToMountd(
		int32 msgType,			//message type
		char *pathNamep, 		//pathName
		int32 pathNameLen,		//pathname length in bytes
		fs_Handle	*fhandlep,	//!=NULL, contains filehandle
		int32 status,			//NFX_OK or NFX_ERR		
		uint32 srcPort,
		uint32 destPort
)
{
	mnt_infoReq_t       *mntInfoReqp;
	uchar8              *buf;
	uint32              allocSize;
	int32				rc;

	MNT_DBG_PRINTF("mnt_sendAdminRespToMountd: msgType: %d pathNamep:%s len:%d status:%d\n",
			msgType, pathNamep, pathNameLen, status);

	allocSize = sizeof(mnt_infoReq_t) + pathNameLen + 1;
	buf = (uchar8 * )eee_ramAlloc(allocSize);
	if (buf == NULL)
	{
		printf("mnt_sendAdminRespToMountd:Failure to alloc buffer!\n");
		return (NFX_ERR);
	}

	bzero(buf, allocSize);
	mntInfoReqp = (mnt_infoReq_t *)buf;

	mntInfoReqp->msgHdr.msgType = msgType;
	mntInfoReqp->msgHdr.private = 0;
	mntInfoReqp->msgHdr.status = status;
	mntInfoReqp->ipAddr = 0; 
	mntInfoReqp->localIpAddr = 0;
	mntInfoReqp->uid = 0;
	mntInfoReqp->gid = 0;
	mntInfoReqp->nameLen = pathNameLen;
	bzero(&mntInfoReqp->pathName[0], 
			sizeof(mntInfoReqp->pathName));
	memcpy(&mntInfoReqp->pathName[0], 
		pathNamep, pathNameLen);

	if (fhandlep != NULL)
	{
		/*
		 *Store update mount point filehandle
		 */
		mntInfoReqp->volId = fhandlep->volId;
		mntInfoReqp->inode = fhandlep->inum;
		mntInfoReqp->mntPnt = fhandlep->mntPnt;

	}
	MNT_DBG_PRINTF("mnt_sendAdminRespToMountd 3: volId: %llx inode: %llx mntPnt: %llx\n",
		mntInfoReqp->volId.whole_thing, mntInfoReqp->inode, mntInfoReqp->mntPnt);
					
	/*
	 *Send MOUNT info request to application
	 */
	rc = eee_sendMessage((uchar8 *)buf, allocSize, srcPort, destPort );
	
	//De-alloc allocate buffer
	eee_ramDealloc(buf);
	
	if (rc != NFX_OK)
	{
		printf("MNT-ERR: Send req to MOUNTD, rc: %d\n", rc);
		return (NFX_ERR);
	}

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_getVolId 
 *
 * Description:
 *		This function sends a request to the EVM on the SSC
 *		to retrieve the Volume Id for the input name.
 * Created by:  Joseph Vu
 *
 * Date Created:        8/03/01
 *
 *-----------------------------------------------------------------
 */
int32 mnt_getVolId(char *name, mnt_request_t  *mntReqp)
{
	uint32              srcPort;
	uint32              destPort;
	int32				rc;
	evm_MsgGetId_t 		evm_getIdMsg;

	MNT_DBG_PRINTF("mnt_getVolId: Enter\n");
	/*
	 *Build EVM Get Id message
	 */
	/*
	printf("mnt_getVolId: mntReqp: %08x\n", mntReqp);
	*/
	memset (&evm_getIdMsg, 0, sizeof(evm_getIdMsg)); 
	evm_getIdMsg.hdr.msgType = EVM_VOL_GET_ID;
	evm_getIdMsg.hdr.clientPrivate = (uint64)((uintptr_t)mntReqp);
	strcpy (evm_getIdMsg.name, name);

	srcPort = PACK_UNICAST_PORT_ADDRESS(
										nfxMySlotId,
										nfxMyCPUId,
										mnt_globalCfgp->mnt_srvrAppId,
										EEE_MGMT,
										NFX_COS_HIGH,
										0);
	destPort = PACK_UNICAST_PORT_ADDRESS(
										0,
										0,			/*SSC*/
										mnt_globalCfgp->evmAppId,
										EEE_MGMT,
										NFX_COS_HIGH,
										0);

	MNT_DBG_PRINTF("MNT TXRX : src app_id (%d) src slot (%d) src cpu (%d)\n",
                        EEE_GET_APP_ID(srcPort),
                        EEE_GET_SLOT_NUM(srcPort),
                        EEE_GET_DEST_CPU(srcPort));

	rc = eee_sendMessage((uchar8 *)&evm_getIdMsg, sizeof(evm_MsgGetId_t),
						destPort, srcPort);
	if (rc != NFX_OK)
	{
		printf("mnt_msgSendToFs: rc: %d\n", rc);
		return (NFX_ERR);
	}

	return (NFX_OK);
}
/*-----------------------------------------------------------------
 * Name:       	mnt_getExportOptMask	
 *
 * Description:
 * Get export options mask for the input 
 * client.  
 *
 * Created by:  Joseph Vu
 * Date Created:        5/03/02
 *
 *-----------------------------------------------------------------
 */

uint32 mnt_getExportOptMask(nfp_fileSysCBPtr_t fileSysCBp, uint32 inputAddr)
{
	nfp_nfsExport_t		*nfsExportp;
	int32				idx;
	uint32				defaultOptMask = 0;

	
	defaultOptMask = (MOUNTD_RW_OPT |MOUNTD_ROOT_SQUASH_OPT);

	if (fileSysCBp == NULL)
		return (defaultOptMask);

	defaultOptMask = fileSysCBp->defaultExpMask;
	MNT_DBG_PRINTF("mnt_getExportOptMask: inputAddr: %08x default mask: %08x total ExpMaskIdx: %d\n",
			inputAddr, defaultOptMask, fileSysCBp->clientExpMaskIdx);

	for (idx=0; ((idx < fileSysCBp->clientExpMaskIdx) && 
			(idx < NFP_NFS_MAX_EXP_MASK_ENTRIES)); idx++)
	{
		nfsExportp = &fileSysCBp->clientSpecificExpMask[idx];
		MNT_DBG_PRINTF("mnt_getExportOptMask: entry ip: %08x entry mask: %08x\n",
				nfsExportp->clientAddr, nfsExportp->exportOptionMask);
		if (nfsExportp->clientAddr == inputAddr)
		{
			if (nfsExportp->exportOptionMask == 0)
				return (defaultOptMask);
			else
				return (nfsExportp->exportOptionMask);
		}
	}

	//return default option mask
	return (defaultOptMask);
}

/*-----------------------------------------------------------------
 * Name:        mnt_isClientInCurrExpList
 *
 * Description:
 * Check if input client IP address exist on our current export list.
 *
 * Created by:  Joseph Vu
 * Date Created:        5/03/02
 *
 *-----------------------------------------------------------------
 */

int32 	//Index: Index of the client export entry
		//-1: Not Found
mnt_isClientInCurrExpList(nfp_fileSysCBPtr_t fileSysCBp, uint32 inputAddr)
{
    nfp_nfsExport_t     *nfsExportp;
    int32               idx;
//  uint32              defaultOptMask = 0;


    if (fileSysCBp == NULL)
        return (-1);


    for (idx=0; ((idx < fileSysCBp->clientExpMaskIdx) && 
            (idx < NFP_NFS_MAX_EXP_MASK_ENTRIES)); idx++)
    {
        nfsExportp = &fileSysCBp->clientSpecificExpMask[idx];
        if (nfsExportp->clientAddr == inputAddr)
        {
            return (idx);
        }
    }

    return (-1);
}


/*-----------------------------------------------------------------
 * Name:       	mnt_storeExportOpt	
 *
 * Description:
 * Get export options mask for the input 
 * client.  
 *
 * Created by:  Joseph Vu
 * Date Created:        5/03/02
 *
 *-----------------------------------------------------------------
 */
int32 mnt_storeExportOpt(
	nfp_fileSysCBPtr_t fileSysCBp, 
	mnt_infoEntry_t	*mtnInfoEntryp)
{
	int32		idx;
	int32		hostAddr;

	/*
	 *Set default export option: 
	 *rw: read-write 
	 *root_squash:  request from root clients (UID=0) are mapped to nobody user and
	 *group ID so they will only have privilleges associated with each other.
	 */
	fileSysCBp->protocol.nfp_nfsFileSysInfo.exportOptionMask = 
								(MOUNTD_RW_OPT | MOUNTD_ROOT_SQUASH_OPT);
	//fileSysCBp->defaultExpMask = (MOUNTD_RW_OPT |MOUNTD_ROOT_SQUASH_OPT);
	hostAddr = mtnInfoEntryp->hostAddr;

	if (mtnInfoEntryp->exportInfoLen)
	{
		nfp_nfsExport_t *nfsExportp;

		if (!hostAddr)
		{
			//store default export mask
			fileSysCBp->defaultExpMask =  mtnInfoEntryp->exportOptionMask;
			MNT_DBG_PRINTF(" mnt_storeExportOpt: "
							"fileSysCBp->defaultExpMask: 0x%x\n",
							fileSysCBp->defaultExpMask);

		}
		else
		{
			//Use client specific option mask
			if (fileSysCBp->clientExpMaskIdx >= NFP_NFS_MAX_EXP_MASK_ENTRIES)
			{
				//Use default option
				E_LOG (class_1, LOG_ERR, mnt_globalCfgp->mnt_srvrAppId, 0, 0, 0,
					("Failure to used client(IP: 0x%08x) share export otption.!",
		 			hostAddr));
				E_LOG (class_1, LOG_WARNING, mnt_globalCfgp->mnt_srvrAppId, 0, 0, 0,
					("Default share export Option will be used for client (IP: 0x%08x)",
		 			hostAddr));

				return (NFX_ERR);
			}
			//store client specific export mask	
			idx = mnt_isClientInCurrExpList(fileSysCBp, hostAddr);
			if (idx == -1)
			{
				idx = fileSysCBp->clientExpMaskIdx;
				fileSysCBp->clientExpMaskIdx++;
			}
			fileSysCBp->numOfClientSpecificMasks++;		
			nfsExportp = &fileSysCBp->clientSpecificExpMask[idx];
			nfsExportp->clientAddr = hostAddr;
			nfsExportp->exportOptionMask = mtnInfoEntryp->exportOptionMask;
			MNT_DBG_PRINTF("mnt_storeExportOpt: add entry for host %08x mask: %08x at idx: %d\n",
				nfsExportp->clientAddr, nfsExportp->exportOptionMask, idx);
		}

	}
	return NFX_OK;
}

#define MNT_EXPORT_STATE_CHANGE(_msgType_, _fileSysCBp_)					\
	{																		\
		if ((_msgType_ == MNT_EXPORT_ENABLED_REQ)||							\
			(_msgType_ == MNT_EXPORT_ENABLED_ALL_REQ))						\
			_fileSysCBp_->nfp_flag |= NFP_SHARE_ENABLED;					\
		else																\
			_fileSysCBp_->nfp_flag &= ~NFP_SHARE_ENABLED;					\
	}


/*-----------------------------------------------------------------
 * Name:        mnt_procExportStateChangeReq
 *
 * Description:
 *      Process share enable/disable request from MOUNTD on SSC.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        05/02/02
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procExportStateChangeReq(char *msgBuf, int32 msgType)
{
	mnt_exportStateChangeReq_t	*mntEscReqp;
	nfp_fileSysCB_t				*fileSysCBp, *currp;
	char						*pathNamep;
	uint32						nameLen;
	uint32						count;

	mntEscReqp = (mnt_exportStateChangeReq_t *)msgBuf;
	MNT_DBG_PRINTF("mnt_procExportStateChangeReq: "
					"nameLen: %d, pathName: %s, msgType 0x%x\n", 
					mntEscReqp->nameLen, mntEscReqp->pathName, msgType);

	if ((msgType == MNT_EXPORT_ENABLED_REQ) ||
		(msgType == MNT_EXPORT_DISABLED_REQ))
	{
		//Enable/disable one share
		if ((mntEscReqp->nameLen == 0) ||
			(strlen(mntEscReqp->pathName) == 0))
		{
			//Error
			MNT_DBG_PRINTF("mnt_procExportStateChangeReq: "
							"Invalid file system path name!\n");
			return (NFX_ERR);
		}

		pathNamep = mntEscReqp->pathName;
		nameLen = mntEscReqp->nameLen;
		if (*pathNamep == '/')
		{
			//Eat the leading "/"
			pathNamep++;
			nameLen--;
		}

		fileSysCBp = nfp_getFileSysByName(pathNamep, nameLen);
		if (fileSysCBp == NULL)
		{
			MNT_DBG_PRINTF("mnt_procExportStateChangeReq: "
							"File system (%s) not found.\n",
							mntEscReqp->pathName);
			return (NFX_ERR);
		}

		//Change flag
		MNT_EXPORT_STATE_CHANGE(msgType, fileSysCBp);
	}
	else
	{
		//Enable/disable all shares

		count = 0;
		for (currp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
			((count < NFP_MAX_FILESYS) && (currp != NULL)); count++, currp++)
		{
			if (!(currp->nfp_flag & NFP_ENTRY_INUSED))
				continue;

			//Change flag
			MNT_EXPORT_STATE_CHANGE(msgType, currp);
		}
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        mnt_updateOneNfsShare
 *
 * Description:
 *      Fill share info in file system cotrol block.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        05/08/02
 *
 *-----------------------------------------------------------------
 */
void
mnt_updateOneNfsShare(nfp_fileSysCB_t	*fileSysCBp,
					  mnt_infoEntry_t	*mtnInfoEntryp)
{
	memcpy(&fileSysCBp->nfp_name[0], 
			&mtnInfoEntryp->pathName[0], 
			mtnInfoEntryp->pathNameLen);
	fileSysCBp->nfp_nameLen = mtnInfoEntryp->pathNameLen;
	fileSysCBp->nfp_flag |= NFP_SHARE_INIT_STATE;
	fileSysCBp->nfp_fileHandle.volId = mtnInfoEntryp->volId;
	fileSysCBp->nfp_fileHandle.inum = mtnInfoEntryp->inode;
	fileSysCBp->nfp_fileHandle.mntPnt = mtnInfoEntryp->mntPnt;
	mnt_storeExportOpt(fileSysCBp, mtnInfoEntryp);
	MNT_DBG_PRINTF("%s: pathName: %s len: %d\n",
					__FUNCTION__,
					&fileSysCBp->nfp_name[0],
					fileSysCBp->nfp_nameLen);
	MNT_DBG_PRINTF("%s: volId: %llx inum: %llx mntPnt: %llx\n",
					__FUNCTION__, 
					fileSysCBp->nfp_fileHandle.volId.whole_thing,
					fileSysCBp->nfp_fileHandle.inum,
					fileSysCBp->nfp_fileHandle.mntPnt);
	return;
}

/*-----------------------------------------------------------------
 * Name:        mnt_procExportAddReq
 *
 * Description:
 *      Process share add request from MOUNTD on SSC.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        05/08/02
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procExportAddReq(char *msgBuf)
{
	mnt_exportAddReq_t	*mntEaReqp;
	mnt_infoEntry_t		*mntInfoEntryp;
	nfp_fileSysCB_t		*fileSysCBp;
	char				*pathNamep;
	uint32				nameLen;
	int32				idx;

	mntEaReqp = (mnt_exportAddReq_t *)msgBuf;
	mntInfoEntryp = &mntEaReqp->mntInfoEntry;
	MNT_DBG_PRINTF("mnt_procExportAddReq: "
					"nameLen: %d, pathName: %s, "
					"localIpAddr: 0x%x, clientIpAddr: 0x%x, "
					"exportOptionMask: 0x%x\n", 
					mntInfoEntryp->pathNameLen, mntInfoEntryp->pathName, 
					mntInfoEntryp->localIpAddr, mntInfoEntryp->hostAddr, 
					mntInfoEntryp->exportOptionMask);

	if ((mntInfoEntryp->pathName[0] == 0) || (mntInfoEntryp->pathNameLen == 0))
		return (NFX_ERR);

	pathNamep = mntInfoEntryp->pathName;
	nameLen = mntInfoEntryp->pathNameLen;
	if (*pathNamep == '/')
	{
		//Eat the leading "/"
		pathNamep++;
		nameLen--;
	}

	fileSysCBp = nfp_getFileSysByName(pathNamep, nameLen);
	if (fileSysCBp == NULL)
	{
		//MNT_DBG_PRINTF("mnt_procExportAddReq: "
		//				"File system (%s) not found.\n", pathNamep);
		fileSysCBp = nfp_allocFileSysCB();
		if (fileSysCBp == NULL)
		{
			return (NFX_ERR);
		}	
		mnt_updateOneNfsShare(fileSysCBp, mntInfoEntryp);
	}

	if (mntInfoEntryp->hostAddr)
	{
		idx = mnt_isClientInCurrExpList(fileSysCBp, mntInfoEntryp->hostAddr);
		if (idx == -1)
		{
			nfp_nfsExport_t     *nfsExportp;
			int32               idx;

			idx = fileSysCBp->clientExpMaskIdx;
			if (idx >= NFP_NFS_MAX_EXP_MASK_ENTRIES)
			{
                //Use default option
				E_LOG (class_1, LOG_ERR, 
				mnt_globalCfgp->mnt_srvrAppId, 0, 0, 0,
				("Failure to used client(IP: 0x%08x) share export otption.!",
                    mntInfoEntryp->hostAddr));
                E_LOG (class_1, LOG_WARNING, 
					mnt_globalCfgp->mnt_srvrAppId, 0, 0, 0,
                    ("Default share export Option will be used for client (IP: 0x%08x)",
                    mntInfoEntryp->hostAddr));


				return (NFX_ERR);
			}
			fileSysCBp->clientExpMaskIdx++;
            nfsExportp = &fileSysCBp->clientSpecificExpMask[idx];
            nfsExportp->clientAddr = mntInfoEntryp->hostAddr;
            nfsExportp->exportOptionMask = mntInfoEntryp->exportOptionMask;
			fileSysCBp->numOfClientSpecificMasks++;
			MNT_DBG_PRINTF("mnt_procExportAddReq: add export mask entry for host %08x mask %08x at idx: %d\n",
                 mntInfoEntryp->hostAddr,
                 nfsExportp->exportOptionMask, idx);

			return (NFX_OK);
		}	
	}
	mnt_storeExportOpt(fileSysCBp, mntInfoEntryp);
	fileSysCBp->nfp_flag |= NFP_SHARE_INIT_STATE;
	MNT_DBG_PRINTF("%s: exit: fileSysCBp: %p nfp_flag: %08x optMask: %08x\n",
		__FUNCTION__, fileSysCBp, fileSysCBp->nfp_flag,  fileSysCBp->defaultExpMask);
	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        mnt_procExportDeleteOneShare
 *
 * Description:
 *      Process share delete on one file system control block.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        05/13/02
 *
 *-----------------------------------------------------------------
 */
void
mnt_procExportDeleteOneShare(nfp_fileSysCB_t		*fileSysCBp, 
							mnt_exportDeleteReq_t	*mntEdReqp)
{
	uint32				clientIpAddr;
	uint32				idx;
	nfp_nfsExport_t		*nfsExportp;
	boolean				noClientMask;

	if (mntEdReqp->clientIpAddr)
	{
		//Clean client specific export option masks of the file system

		clientIpAddr = mntEdReqp->clientIpAddr;

		for (idx=0; ((idx < fileSysCBp->clientExpMaskIdx) &&
				(idx < NFP_NFS_MAX_EXP_MASK_ENTRIES)); idx++)
		{
			nfsExportp = &fileSysCBp->clientSpecificExpMask[idx];
			if (nfsExportp->clientAddr == clientIpAddr)
			{
				nfsExportp->exportOptionMask = 0;
				fileSysCBp->numOfClientSpecificMasks--;
				break;
			}
		}
	}
	else
	{
		//Clean default export option masks of the file system

		fileSysCBp->defaultExpMask = 0;
	}

	//If default export mask and all client specific export masks are cleared,
	//remove the fileSysCb
	noClientMask = TRUE;
	for (idx=0; (idx < NFP_NFS_MAX_EXP_MASK_ENTRIES); idx++)
	{
		nfsExportp = &fileSysCBp->clientSpecificExpMask[idx];
		if (nfsExportp->exportOptionMask != 0)
		{
			noClientMask = FALSE;
			break;
		}
	}
	MNT_DBG_PRINTF("%s: defaultExpMask: %08x noClientMask: %d\n", 
							__FUNCTION__, fileSysCBp->defaultExpMask, noClientMask);
	if ((fileSysCBp->defaultExpMask == 0) && (noClientMask == TRUE))
	{
		MNT_DBG_PRINTF("%s: delalloc fileSysCB: %p\n", __FUNCTION__, fileSysCBp);	
		nfp_deallocFileSysCB(fileSysCBp);
	}
	
	return;
}


/*-----------------------------------------------------------------
 * Name:        mnt_procExportDeleteReq
 *
 * Description:
 *      Process share delete request from MOUNTD on SSC.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        05/08/02
 *
 *-----------------------------------------------------------------
 */
int32
mnt_procExportDeleteReq(char *msgBuf)
{
	mnt_exportDeleteReq_t	*mntEdReqp;
	nfp_fileSysCB_t			*fileSysCBp, *currp;
	char					*pathNamep;
	uint32					nameLen;
	uint32					count, idx;


	mntEdReqp = (mnt_exportDeleteReq_t *)msgBuf;
	MNT_DBG_PRINTF("mnt_procExportDeleteReq: "
					"nameLen: %d, pathName: %s, 
					localIpAddr: 0x%x, clientIpAddr: 0x%x\n", 
					mntEdReqp->nameLen, mntEdReqp->pathName, 
					mntEdReqp->localIpAddr, mntEdReqp->clientIpAddr);

	if (mntEdReqp->msgHdr.msgType == MNT_EXPORT_DELETE_ALL_REQ)
	{
		//All the shares are deleted.
		//Clean all the export option masks of 
		//all the file system control blocks
		
		count = 0;
		for (currp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
		((count < NFP_MAX_FILESYS) && (currp != NULL)); count++, currp++)
		{
			/*
			currp->defaultExpMask = 0;
			for (idx=0; ((idx < currp->clientExpMaskIdx) &&
				(idx < NFP_NFS_MAX_EXP_MASK_ENTRIES)); idx++)
			{
				currp->clientSpecificExpMask[idx].exportOptionMask = 0;
			}
			*/

			//Remove fileSysCb
			nfp_deallocFileSysCB(currp);
		}
		return (NFX_OK);
	}

	if ((mntEdReqp->pathName[0] == 0) || (mntEdReqp->nameLen == 0))
		return (NFX_ERR);

	pathNamep = mntEdReqp->pathName;
	nameLen = mntEdReqp->nameLen;
	if (*pathNamep == '/')
	{
		//Eat the leading "/"
		pathNamep++;
		nameLen--;
	}

	fileSysCBp = nfp_getFileSysByName(pathNamep, nameLen);
	if (fileSysCBp == NULL)
	{
		MNT_DBG_PRINTF("mnt_procExportDeleteReq: "
						"File system (%s) not found.\n",
						mntEdReqp->pathName);
		return (NFX_ERR);
	}

	mnt_procExportDeleteOneShare(fileSysCBp, mntEdReqp);

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   mnt_updateNfsShares 
 *
 * Description:
 *  This function is call upon restarted of TXRX after retrieving
 *	NFS shares export from MOUNDD on SSC to
 *  set up appropriate File System Controls for all the NFS shares.
 *
 * Created by:  Joseph Vu 
 *
 * Date Created:        08/05/02
 *
 *-----------------------------------------------------------------
 */
int32
mnt_updateNfsShares(eee_desc_t *eDescp)
{
	nfp_fileSysCB_t				*fileSysCBp;
	mnt_infoEntry_t 			*mtnInfoEntryp;
	eee_descBuf_t				*bd;
	mnt_infoResp_t      		*mntInfoResp;
	int32						numOfEntries;
	int32						i;
	uint32						cookie;

	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
		&eDescp->bd[0] : &eDescp->bd[1];


	mntInfoResp = (mnt_infoResp_t  *)(bd->buf);

	cookie = (uint32)mntInfoResp->cookie;
	numOfEntries = mntInfoResp->numOfEntries;
	mtnInfoEntryp = &mntInfoResp->mtnInfoEntry[0];	

	MNT_DBG_PRINTF("%s: num of NFS shares: %d\n",
					__FUNCTION__, numOfEntries);
	
	if (numOfEntries <= 0)
	{
		return (NFX_OK);
	}

	for (i=0; i < numOfEntries; i++)
	{
		if ((mtnInfoEntryp->pathNameLen > 0) && 
			(mtnInfoEntryp->pathNameLen < MAX_NAME_LEN))
		{
			fileSysCBp = 
				nfp_getFileSysByName (&mtnInfoEntryp->pathName[0],
							mtnInfoEntryp->pathNameLen); 
			if (fileSysCBp == NULL) 
			{
				fileSysCBp = nfp_allocFileSysCB();
				if (fileSysCBp == NULL)
				{
					goto mnt_updateNfsSharesErr;
				}
				mnt_updateOneNfsShare(fileSysCBp, mtnInfoEntryp);
			}
			//update export option
			mnt_storeExportOpt(fileSysCBp, mtnInfoEntryp);
		}
		//advance to next entry
		MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);

	}

	mnt_getNfsShares(cookie);
	
	return (NFX_OK);

mnt_updateNfsSharesErr:

	return (NFX_ERR);
}


/*-----------------------------------------------------------------
 * Name:     mnt_getFileHandle  	 
 *
 * Description:
 *     Send a fs_MsgGetRootReq request to file system
 * to get filehandle for configured NFS share (mount point).
 *
 * Created by: 	Joseph Vu 
 *
 * Date Created:        07/02/02
 *
 *-----------------------------------------------------------------
 */
int32
mnt_getFileHandle(eee_desc_t *eDescp)
{
	eee_descBuf_t				*bd;
	nfp_fileSysCBPtr_t			fileSysCBp;
	nfp_requestCBPtr_t  		reqCBp;
	mnt_request_t    			*mntReqp;
	mnt_getFileHandleReq_t		*mnt_getFHandleReqp;
	char						*pathNamep;
	int32						rc;
	int32						pathLen;
	uint32                  	srcPort, destPort;

	reqCBp = NULL;
	fileSysCBp = NULL;
	mntReqp = NULL;
	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
		&eDescp->bd[0] : &eDescp->bd[1];
	mnt_getFHandleReqp = (mnt_getFileHandleReq_t *)(bd->buf);

	MNT_DBG_PRINTF("mnt_getFileHandle: "
					"nameLen: %d, pathName: %s volId: %llx\n", 
			mnt_getFHandleReqp->nameLen , 
			mnt_getFHandleReqp->pathName,
			mnt_getFHandleReqp->volId.whole_thing);

	/*Get application EEE source/dest port*/
	srcPort = eDescp->hdr.src_port;
	destPort = eDescp->hdr.dest_port;
	/*
	 *Allocate MNT request control block
	 */ 
	reqCBp = nfp_allocReqCB ();
	if (reqCBp == NULL)
	{
		//return (NFX_ERR);
		goto mnt_getFileHandleErr;
	}
	mnt_globalCfgp->mnt_debugCounters.mnt_allocAdminReqCB++;
	mntReqp = &reqCBp->protocol.mnt_req;
	mntReqp->reqCBp = (void *)reqCBp;
	mntReqp->reqState = MNT_REQ_GET_FHANDLE;
	mntReqp->srcPort = srcPort;
	mntReqp->destPort = destPort;
	mntReqp->clientCBp = NULL;
	mntReqp->rpcCallHdrp = NULL;
	mntReqp->tplConnp = NULL;
	mntReqp->nfsReqp = NULL;
	mntReqp->fileSysCBp = NULL;

	pathNamep = &mnt_getFHandleReqp->pathName[0];
	pathLen = mnt_getFHandleReqp->nameLen;
	/*
	 *Eat leading '/' 
	 */
	if (*pathNamep && *pathNamep == '/')
	{
		pathNamep += 1;	
		pathLen -= 1;
	}
	fileSysCBp = nfp_getFileSysByName (pathNamep, pathLen); 
	if (fileSysCBp == NULL) 
	{
		fileSysCBp = nfp_allocFileSysCB();
		if (fileSysCBp == NULL)
		{
			goto mnt_getFileHandleErr;
		}
		memcpy(&fileSysCBp->nfp_name[0], pathNamep, pathLen);
		fileSysCBp->nfp_nameLen = pathLen;
		fileSysCBp->nfp_flag |= NFP_SHARE_INIT_STATE;
	}
	mntReqp->fileSysCBp =  (void *)fileSysCBp;

	/*
	 *Set up fs_MsgGetRootReq message 
	 *to send to file system to get 
	 *Mount point filehandle 
	 */
	rc = mnt_msgMapMountReq(mntReqp, mnt_getFHandleReqp->volId);
	if (rc == NFX_ERR)
	{
		goto mnt_getFileHandleErr;
	}
	rc = mnt_msgSendToFs(mntReqp);
	if (rc == NFX_ERR)
	{
		goto mnt_getFileHandleErr;
	}
	return (NFX_OK);

mnt_getFileHandleErr:
	if (mntReqp && mntReqp->fileSysCBp)
		mntReqp->fileSysCBp = NULL;

	if (reqCBp)
	{
		nfp_deallocReqCB(reqCBp);
		mnt_globalCfgp->mnt_debugCounters.mnt_freeAdminReqCB++;
	}

	if (fileSysCBp)
		nfp_deallocFileSysCB(fileSysCBp);
	
	mnt_sendAdminRespToMountd(MNT_UPD_FHANDLE_REQ,
					mnt_getFHandleReqp->pathName,
					mnt_getFHandleReqp->nameLen,
					NULL,
					NFX_ERR,
					srcPort,
					destPort);
	return (NFX_ERR);
}

/*-----------------------------------------------------------------
 * Name:   mnt_sendFileHandleResp  
 *
 * Description:
 *	Send an update filehandle (for the request mount point)
 *	response to MOUNTD on SSC.
 *
 * Created by: 	Joseph Vu 
 *
 * Date Created:        07/02/02
 *
 *-----------------------------------------------------------------
 */
int32
mnt_sendFileHandleResp(eee_desc_t *eDescp)
{
	fs_MsgRespHdr   			*fsMsgRespHdrp;
	fs_MsgGetRootResp   		*fsGetRootRespMsgp;
	nfp_fileSysCBPtr_t			fileSysCBp;
	mnt_request_t    			*mntReqp;
	eee_descBuf_t				*bd;
	int32						rc;
	uint32						srcPort;
	uint32						destPort;


	MNT_DBG_PRINTF("mnt_SendFileHandleResp:enter\n");
	fileSysCBp = NULL;
	mntReqp = NULL;
	
	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
		&eDescp->bd[0] : &eDescp->bd[1];
	srcPort = destPort = 0;
	if (bd == NULL)
	{
		eee_freePacket(eDescp);
		goto mnt_sendFileHandleRespErr;	
	}
	fsMsgRespHdrp = (fs_MsgRespHdr   *) (bd->buf);
	mntReqp = (mnt_request_t *)fsMsgRespHdrp->clientPrivate;
	fileSysCBp = mntReqp->fileSysCBp;
	srcPort = mntReqp->srcPort;
	destPort = mntReqp->destPort;

	if (nfp_validateFileSysCB(fileSysCBp) != NFX_OK)
	{
		goto mnt_sendFileHandleRespErr;
	}

	if ((fsMsgRespHdrp->fsStatus == FS_ERR_NOT_MOUNTED) ||
    	(fsMsgRespHdrp->fsStatus == FS_ERR_NO_RESOURCE))
	{
		goto mnt_sendFileHandleRespErr;
		
	}
	fsGetRootRespMsgp = (fs_MsgGetRootResp *)bd->buf;

	/*
	 *Store update filehandle 
	 *in the file system control block
	 */
	fileSysCBp->nfp_fileHandle.volId = fsGetRootRespMsgp->dirHandle.volId;
	fileSysCBp->nfp_fileHandle.inum = fsGetRootRespMsgp->dirHandle.inum;
	fileSysCBp->nfp_fileHandle.mntPnt = fsGetRootRespMsgp->dirHandle.mntPnt;
	/*
	 *Send update Filehanldle To SSC
	 */
	rc = mnt_sendAdminRespToMountd(MNT_UPD_FHANDLE_REQ,
				&fileSysCBp->nfp_name[0],
				fileSysCBp->nfp_nameLen,
				&fileSysCBp->nfp_fileHandle,
				NFX_OK,
				srcPort,
				destPort);
	if (rc != NFX_OK)
	{
		goto mnt_sendFileHandleRespErr;
	}
	/*
	 *Free MNT request control block
	 */
	nfp_deallocReqCB(mntReqp->reqCBp);
	mnt_globalCfgp->mnt_debugCounters.mnt_freeAdminReqCB++;
	mntReqp->fileSysCBp = NULL;
	return (NFX_OK);	
	/*
	 *Free MNT req control Control block
	 */
mnt_sendFileHandleRespErr:
	if (fileSysCBp)
	{
		if (srcPort && destPort)
			mnt_sendAdminRespToMountd(MNT_UPD_FHANDLE_REQ,
					&fileSysCBp->nfp_name[0],
					fileSysCBp->nfp_nameLen,
					NULL,
					NFX_ERR, 
					srcPort,
					destPort);
		nfp_deallocFileSysCB(fileSysCBp);
	}
	if (mntReqp && mntReqp->fileSysCBp)
		mntReqp->fileSysCBp = NULL;

	if (mntReqp && mntReqp->reqCBp)
	{
		nfp_deallocReqCB(mntReqp->reqCBp);
		mnt_globalCfgp->mnt_debugCounters.mnt_freeAdminReqCB++;
		mntReqp->fileSysCBp = NULL;
	}
	
	return (NFX_ERR);
}

/*-----------------------------------------------------------------
 * Name:        mnt_getMountdReq
 *
 * Description:
 *      Receive request on FP-NIM from MOUNTD on SSC.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        05/02/02
 *
 *-----------------------------------------------------------------
 */
void mnt_getMountdReq(eee_desc_t *eDescp)
{
	eee_descBuf_t				*bd;
	mnt_msgHdr_t				*mntMsgHdrp;
	uint32						msgLen;
	int32						rc;

 	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
         &eDescp->bd[0] : &eDescp->bd[1];
	mntMsgHdrp = (mnt_msgHdr_t *)(bd->buf);
	msgLen = eDescp->hdr.control & EDESC_TOTAL_LEN;

	if ((mntMsgHdrp->msgType >= MNT_EXPORT_ENABLED_REQ) &&
		(mntMsgHdrp->msgType <= MNT_EXPORT_DISABLED_ALL_REQ))
	{
		//Share enable/disable request
		rc = mnt_procExportStateChangeReq(bd->buf, mntMsgHdrp->msgType);
	}
	else if (mntMsgHdrp->msgType == MNT_EXPORT_ADD_REQ)
	{
		//Share add request
		rc = mnt_procExportAddReq(bd->buf);
	}
	else if ((mntMsgHdrp->msgType == MNT_EXPORT_DELETE_REQ) ||
			 (mntMsgHdrp->msgType == MNT_EXPORT_DELETE_ALL_REQ)) 
	{
		//Share delete request
		rc = mnt_procExportDeleteReq(bd->buf);
	}
	else if (mntMsgHdrp->msgType == MNT_EXPORT_GET_FHANDLE_REQ)
	{
		MNT_DBG_PRINTF("mnt_getMountdReq: ***Rcv'd MNT_EXPORT_GET_FHANDLE_REQ\n");
		rc = mnt_getFileHandle(eDescp);	
	}
	else if (mntMsgHdrp->msgType == MNT_GET_SHARES_RESP)
	{
		rc = mnt_updateNfsShares(eDescp);	
	}
	eee_freePacket(eDescp);

	return;
}


/*-----------------------------------------------------------------
 * Name:       	mnt_msgSscRcv 
 *
 * Description:
 *		Receive function from SSC.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        8/03/01
 *
 *-----------------------------------------------------------------
 */
void mnt_msgSscRcv(eee_desc_t *eDescp )
{
	eee_descBuf_t		*bd;
	evm_MsgStat_t 		*evmStatMsgp;
	mnt_request_t  		*mntReqp;

	MNT_DBG_PRINTF("mnt_msgSscRcv: Enter\n");	
	bd = &eDescp->bd[1];
	evmStatMsgp = (evm_MsgStat_t       *)(bd->buf);

	if (evmStatMsgp->payload.lVolId.whole_thing < 0)
	{
		printf("MNT: Invalid VolId: %lld\n", evmStatMsgp->payload.lVolId.whole_thing);
		return;
	}
	mntReqp = ( mnt_request_t *)evmStatMsgp->hdr.clientPrivate;
	
	printf("mnt_msgSscRcv: volId: %lld\n", evmStatMsgp->payload.lVolId.whole_thing);
	MNT_DBG_PRINTF("mnt_msgSscRcv: mntReqp: %p\n", mntReqp);
	if (mnt_msgMapMountReq(mntReqp, evmStatMsgp->payload.lVolId) == NFX_ERR)
	{
		nfs_errMsg(mntReqp->rpcCallHdrp->rpc_xid, mntReqp->rpcCallHdrp->rpc_proc,
				NFSERR_MAP_FS_MSG);
		return;	
	}

	if (mnt_msgSendToFs(mntReqp) == NFX_ERR)
	{
		nfs_errMsg(mntReqp->rpcCallHdrp->rpc_xid, mntReqp->rpcCallHdrp->rpc_proc,
					NFSERR_SENDMSG_TO_FS);
		return;
	}

	return;
}
/*-----------------------------------------------------------------
 * Name:       mnt_sendMountdReqToSsc 
 *
 * Description:
 * Send a request to MOUNTD on SSC to retrieve information
 * on the volume based on the input filehandle (volume ID)
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        9/03/01
 *
 *-----------------------------------------------------------------
 */
int32 mnt_sendMountdReqToSsc ( mnt_request_t    *mntReqp, fs_Handle *fhandlep)
{
	nfp_fileSysCBPtr_t  fileSysCBp;
	nfp_clientCBPtr_t   clientCBp;
	mnt_infoReq_t       *mntInfoReqp;
	uchar8              *buf;
	uint32              allocSize;
	uint32              destPort;
	uint32              srcPort;
	int32               rc;

	MNT_DBG_PRINTF(" mnt_sendMountdReqToSsc enter\n");

	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;
	fileSysCBp = (nfp_fileSysCBPtr_t)mntReqp->fileSysCBp;
	MNT_DBG_PRINTF("mnt_getMountdInfoReq 2: fileSysCBp: %p mntReqp: %p clientCBp: %p\n",
					fileSysCBp, mntReqp, clientCBp);
	if ( (fileSysCBp == NULL) || (mntReqp == NULL) || (clientCBp == NULL) )
		return (NFX_ERR);
	
	allocSize = sizeof(mnt_infoReq_t);
	buf = (uchar8 * )eee_ramAlloc(allocSize);
	if (buf == NULL)
	{
		printf("mnt_getMountdInfoReq: Failure to alloc buffer!\n");
		return (NFX_ERR);
	}

	mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt3++;
	memset(buf, 0, allocSize);
	mntInfoReqp = (mnt_infoReq_t *)buf;
	
	mntInfoReqp->msgHdr.msgType = MNT_INFO_REQ_BY_ID;
	mntInfoReqp->msgHdr.private = (uint64)((uintptr_t)mntReqp);

	mntInfoReqp->ipAddr = clientCBp->nfp_ipAddr;
	mntInfoReqp->localIpAddr = clientCBp->nfp_localIpAddr;
	mntInfoReqp->uid = mntReqp->rpcCallHdrp->authFlavor.authSys.userId;
	mntInfoReqp->gid = mntReqp->rpcCallHdrp->authFlavor.authSys.groupId;
	mntInfoReqp->volId = fhandlep->volId; 
	mntInfoReqp->inode = fhandlep->inum;
	mntInfoReqp->mntPnt = fhandlep->mntPnt;
	/*
	 *Send MOUNT info request to MOUNTD on SSC
	 */
	srcPort = PACK_UNICAST_PORT_ADDRESS(
										nfxMySlotId,
										nfxMyCPUId,
										mnt_globalCfgp->mnt_srvrAppId,
										EEE_MGMT,
										NFX_COS_HIGH,
										0);
	destPort = PACK_UNICAST_PORT_ADDRESS(
										0,
										0,         /*MOUNTD on SSC*/
										mnt_globalCfgp->mountdAppId,
										EEE_MGMT,
										NFX_COS_HIGH,
										0);

	MNT_DBG_PRINTF("MNT TXRX : src app_id (%d) src slot (%d) src cpu (%d)\n",
						EEE_GET_APP_ID(srcPort),
						EEE_GET_SLOT_NUM(srcPort),
						EEE_GET_DEST_CPU(srcPort));

	MNT_DBG_PRINTF("mnt_sendMountdReqToSsc: volId: %llx inode: %llx mntPnt: %llx mntReqp: %08llx\n",
							mntInfoReqp->volId.whole_thing,
							mntInfoReqp->inode,
							mntInfoReqp->mntPnt,
							mntInfoReqp->msgHdr.private);
	MNT_DBG_PRINTF("ipAddr: %08x\n", mntInfoReqp->ipAddr);
	rc = eee_sendMessage((uchar8 *)buf, allocSize, destPort, srcPort );
	eee_ramDealloc(buf);

	mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt3++;
	if (rc != NFX_OK)
	{
		printf("MNT-ERR: Send req to MOUNTD, rc: %d\n", rc);
		return (NFX_ERR);
	}
	mnt_globalCfgp->mnt_debugCounters.mnt_xmtReqByIdToSsc++;

	return (NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:       	mnt_getMountdInfoReqById 
 *  
 * Description:
 *  We must have restarted or crashed and now received NFS request from
 *	the client.  We must retrieve MOUNTD (on the SSC) info request for this 
 *  client for the input volume (part of filehandle). 
 *
 * Created by:  Joseph Vu
 *  
 * Date Created:        8/03/01
 *
 *-----------------------------------------------------------------
 */ 
int32 mnt_getMountdInfoReqById(nfs_request_t *nfsReqp, fs_Handle *fhandlep)
{

	nfp_clientCBPtr_t	clientCBp;
	nfp_fileSysCBPtr_t	fileSysCBp;
	nfp_requestCBPtr_t  reqCBp;
	mnt_request_t       *mntReqp;
	rpcCallHdr_t        *newRpcCallp;

	MNT_DBG_PRINTF("mnt_getMountdInfoReqById: Enter\n");

	if ((nfsReqp == NULL) || (fhandlep == NULL) )
		return (NFX_ERR);
	newRpcCallp =  rpc_createHdrInfo(nfsReqp->rpcCallHdrp);
	if (newRpcCallp == NULL)
	{
		//panic ("mnt_getMountdInfoReqById: Failure to alloc RPC Header Info!\n");
		return (NFX_ERR);
	}	
	mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt4++;
	/*
	 *Allocate request control block
	 */
	reqCBp = nfp_allocReqCB ();
	if (reqCBp == NULL)
	{
		//panic ("mnt_getMountdInfoReqById: Failure to alloc MNT Reqest Control Blk!\n");
		if (newRpcCallp != NULL) 
		{
			eee_ramDealloc(newRpcCallp);
		}
		return (NFX_ERR);
	}
	mnt_globalCfgp->mnt_debugCounters.mnt_allocReqCB++;	
	clientCBp = nfsReqp->clientCBp;
	mntReqp = &reqCBp->protocol.mnt_req;
	mntReqp->rpcCallHdrp = newRpcCallp;
	mntReqp->reqCBp = (void *)reqCBp;
	mntReqp->reqState = MNT_REQ_INPROG;
	mntReqp->clientCBp = (void *)clientCBp;
	mntReqp->tplConnp = 0;
    
	MNT_DBG_PRINTF(" mnt_getMountdInfoReqById: mntReqp: %p clientCBp: %p tplConnp: %p\n",
						mntReqp, mntReqp->clientCBp, mntReqp->tplConnp);

	MNT_DBG_PRINTF(" mnt_getMountdInfoReqById: clientCBp: %p client Addr: %08x\n",
					clientCBp, clientCBp->nfp_ipAddr); 

	fileSysCBp = nfp_getFileSysCB(fhandlep);		
	if (fileSysCBp == NULL)
	{
		fileSysCBp = nfp_allocFileSysCB();
		if (fileSysCBp == NULL)
		{
			panic("mnt_getMountdInfoReqById: Failure to alloc file system control block!\n");
            return (NFX_ERR);
        }
		memcpy(&fileSysCBp->nfp_fileHandle, fhandlep, sizeof(fs_Handle));
		fileSysCBp->nfp_flag |= NFP_SHARE_INIT_STATE;
	}
	fileSysCBp->atime = nfs_utilGetTimeOfDayInSecs();
	mntReqp->fileSysCBp =  (void *)fileSysCBp;
	if (nfs_findFileSysCBNode (&clientCBp->nfp_fileSysList, fileSysCBp) == NULL)
	{
		if (nfs_addCBToList(&clientCBp->nfp_fileSysList,
				(void *)fileSysCBp, NFP_FILESYSCB_NODE)  != NFX_OK)
		{
			printf("mnt_getMountdInfoReqById: nfs_addCBToList file system control block failed\n");
			return (NFX_ERR);
		}
	}
	mntReqp->nfsReqp = nfsReqp;
	MNT_DBG_PRINTF("mnt_getMountdInfoReqById: mntReqp: %p nfsReqp: %p\n", mntReqp, nfsReqp);	
	return (mnt_sendMountdReqToSsc(mntReqp, fhandlep));
	
}

void mnt_displayStat(void)
{
    mnt_pktCounters_t   *rxPktStatp;
    mnt_pktCounters_t   *txPktStatp;
    mnt_statCounters_t  *errStatp;

    rxPktStatp = &mnt_globalCfgp->mnt_rxPktCounters;
    txPktStatp = &mnt_globalCfgp->mnt_xmtPktCounters;
    //total_nfsV2RxPkt =   nfs_totalPktCnt(rxPktStatp, NFS_VER2);

    printf("Client MOUNT call counters\n");
    printf("----------------------------\n");
    printf("null:       %lld\n", rxPktStatp->mnt_null);
    printf("mount:      %lld\n", rxPktStatp->mnt_addMnt);
    printf("umount:     %lld\n",rxPktStatp->mnt_umnt);
    printf("umountAll:  %lld\n",rxPktStatp->mnt_umntAll);
    printf("exportList: %lld\n",rxPktStatp->mnt_export);
    printf("exportAll:  %lld\n",rxPktStatp->mnt_exportAll);
    printf("pathConf:   %lld\n",rxPktStatp->mnt_pathConf);

    printf("Client MOUNT replies counters\n");
    printf("----------------------------\n");
    printf("null:       %lld\n", txPktStatp->mnt_null);
    printf("mount:      %lld\n", txPktStatp->mnt_addMnt);
    printf("umount:     %lld\n",txPktStatp->mnt_umnt);
    printf("umountAll:  %lld\n",txPktStatp->mnt_umntAll);
    printf("exportList: %lld\n",txPktStatp->mnt_export);
    printf("exportAll:  %lld\n",txPktStatp->mnt_exportAll);
    printf("pathConf:   %lld\n",txPktStatp->mnt_pathConf);


    errStatp = &mnt_globalCfgp->mnt_statCounters;
    printf("Client MOUNT error counters\n");
    printf("----------------------------\n");
    printf("permission errs:           %lld\n", errStatp->mnt_errPerm);
    printf("No Entry errs:             %lld\n", errStatp->mnt_errNoEnt);
    printf("I/O errs:                  %lld\n", errStatp->mnt_errIO);
    printf("access errs:               %lld\n", errStatp->mnt_errAccess);
    printf("not a directory errs:      %lld\n", errStatp->mnt_errNotDir);
    printf("invalid argument errs:     %lld\n", errStatp->mnt_errInval);
    printf("name too long errs:        %lld\n", errStatp->mnt_errNameTooLong);
    printf("oper not supported errs:   %lld\n", errStatp->mnt_errNotSupp);
    printf("failure on server errs:    %lld\n", errStatp->mnt_errSrvFault);
    printf("stale handle errs:         %lld\n", errStatp->mnt_errFhandle);


    printf("MOUNT debug counters\n");
    printf("-----------------------------------------------------------\n");
    printf("Drop TPL Rcvd Pkt:                                    %lld\n",
                mnt_globalCfgp->mnt_debugCounters.mnt_tplDropRcvPkt);
    printf("Alloc EEE buffers send to FS:                          %lld\n", 
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt1); 
    printf("Free EEE buffers send to FS:                           %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt1 );

    printf("Alloc EEE buffers send mount req (by name) to SSC:     %lld\n", 
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt2); 
    printf("Free EEE buffers send mount req (by name) to SSC:      %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt2 );
    printf("Alloc EEE buffers send mount req (by Vol Id)to SSC:    %lld\n", 
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt3); 
    printf("Free EEE buffers send mount req (by Vol Id) to SSC:    %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt3 );
    printf("Num of send mount req (by Pathname) to SSC:            %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_xmtReqByNameToSsc);
    printf("Num of rcv'd mount req (by PathName) from SSC:         %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_rxRespByNameFromSsc);
    printf("Num of send mount req (by Vol Id) to SSC:              %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_xmtReqByIdToSsc);
    printf("Num of rcv'd mount req (by Vol Id) from SSC:           %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_rxRespByIdFromSsc);
    printf("Alloc EEE buffers for RX RPC hdr:                      %lld\n", 
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt4); 
    printf("Free EEE buffers for RX PRC hdr:                       %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt4 );
	printf("Alloc NFP Req Control Blocks:                          %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_allocReqCB);
	printf("Free NFP Req Control Blocks:                           %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_freeReqCB);
	printf("Alloc Admin NFP Req Control Blocks:                    %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_allocAdminReqCB);
	printf("Free Admin NFP Req Control Blocks:                     %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_freeAdminReqCB);
	printf("MNT TPL XMT PKT:                                       %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_tplXmtPktSend);
	printf("MNT TPL XMT Complete:                                  %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_tplXmtPktCmpInd );
	printf("MNT TPL XMT PKT ERR:                                   %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_tplXmtSendErr );
	printf("MNT FS No Resource/Volume Err:                          %lld\n",
				mnt_globalCfgp->mnt_debugCounters.mnt_fsErrNoResource);
	printf("Num of MNT RPC AuthType Err:                           %lld\n",
				mnt_rxAuthTypeErr);


}

/*-----------------------------------------------------------------
 * Name:   mnt_getNfsShares    		
 *  
 * Description:
 *  Send a message to MOUNTD SSC to get 
 *  the status of the NFS shares. 
 *
 * Created by:  Joseph Vu
 *  
 * Date Created:        08/07/02
 *
 *-----------------------------------------------------------------
 */ 
void mnt_getNfsShares(uint32 cookie)
{

	uint32             	srcPort, destPort;
	mnt_infoReq_t   	mntInfoReq;

	memset(&mntInfoReq, 0, sizeof(mntInfoReq));
	mntInfoReq.cookie = cookie;

	mntInfoReq.msgHdr.msgType = MNT_GET_SHARES;

	// Set response buffer size
	mntInfoReq.respBufSize = MNT_MSG_RESP_BUF_MAX;

	/*
	 *Send MOUNT info request to MOUNTD on SSC
	 */
	srcPort = PACK_UNICAST_PORT_ADDRESS(
								nfxMySlotId,
								nfxMyCPUId,
								mnt_globalCfgp->mnt_srvrAppId,
								EEE_MGMT,
								NFX_COS_HIGH,
								0);
	destPort = PACK_UNICAST_PORT_ADDRESS(
								0,
								0,         /*MOUNTD on SSC*/
								mnt_globalCfgp->mountdAppId,
								EEE_MGMT,
								NFX_COS_HIGH,
								0);

	eee_sendMessage((uchar8 *)&mntInfoReq, 
					sizeof(mntInfoReq), destPort, srcPort );
	return;
}

/*-----------------------------------------------------------------
 * Name:  mnt_getNfsShareHandler	 
 *  
 * Description:
 *	Timer callback function to get the NFS shares from MOUNTD on SSC
 * Created by:  Joseph Vu
 *  
 * Date Created:        08/07/02
 *
 *-----------------------------------------------------------------
 */ 
int32
mnt_getNfsShareHandler( uint32 tmr,
              void  *info1,
              void  *info2,
              void  *info3,
              void  *info4 )
{
	MNT_DBG_PRINTF("%s: enter\n", __FUNCTION__);
	mnt_getNfsShares(0);
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:       	nfp_getFileSysMntPnt 
 *
 * Description: Retrieve a File System Control Block (FSCB) entry
 * for input mount point. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        8/20/02
 *
 *-----------------------------------------------------------------
 */
nfp_fileSysCBPtr_t  /*
                     *Return pointer to FSCB entry of NULL if entry
                     *not found
                     */
nfp_getFileSysMntPnt( fs_Handle    *nfp_fhandle        /*filehandle*/
)
{
    nfp_fileSysCBPtr_t  currp;
    uint32              count;
    uint64              searchId;
    uint64              searchInum;


    count = 0;
    searchId = nfp_fhandle->volId.whole_thing;
    searchInum = nfp_fhandle->inum;
    for (currp = (nfp_fileSysCBPtr_t)nfp_globalCfgp->fileSysCbCfgList.listBasep;
        ((count < NFP_MAX_FILESYS) && (currp != NULL)); count++, currp++)
    {
        if (!(currp->nfp_flag & NFP_ENTRY_INUSED))
            continue;
        if ( (currp->nfp_fileHandle.volId.whole_thing == searchId) &&
            (currp->nfp_fileHandle.mntPnt == nfp_fhandle->mntPnt) )
            return (currp);
    }
    return (NULL);

}

/*-----------------------------------------------------------------
 * Name: 	mnt_isFileSysDuplicate 
 *  
 * Description:
 *	Check if file System control block is already allocate
 *	for input file system.
 *  
 * Date Created:        08/07/02
 *
 *-----------------------------------------------------------------
 */ 
nfp_fileSysCBPtr_t mnt_isFileSysDuplicate(
	nfp_fileSysCBPtr_t fileSysCBp,
	mnt_infoEntry_t *mtnInfoEntryp)
{
	nfp_fileSysCBPtr_t		duplicateFileSysCBp;
	fs_Handle				fileHandle;	

	//Check input entry based on name

	if ( (mtnInfoEntryp == NULL) || (fileSysCBp == NULL) )
		return (NULL);
	MNT_DBG_PRINTF("%s: fileSysCBp: %p\n", __FUNCTION__, fileSysCBp);
	duplicateFileSysCBp = 
		nfp_getFileSysByName (&mtnInfoEntryp->pathName[0],
						mtnInfoEntryp->pathNameLen);

	if ( (duplicateFileSysCBp) && (duplicateFileSysCBp != fileSysCBp) )
		return (duplicateFileSysCBp);

	//Check input entry base volId & mntPnt Ids

	fileHandle.volId = mtnInfoEntryp->volId;
	fileHandle.mntPnt = mtnInfoEntryp->mntPnt;
	
	duplicateFileSysCBp = nfp_getFileSysMntPnt(&fileHandle);
	if ( (duplicateFileSysCBp) && (duplicateFileSysCBp != fileSysCBp) )
		return (duplicateFileSysCBp);

	return (NULL);
}

