/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        mountd-util.c
 *
 * RCS:  $Id: mountd-util.c,v 1.29 2002/09/19 22:50:18 joseph Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *		This file contains MOUNTD utility functions.
 *
 * Created by: Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
#if defined(LINUX_TEST) || defined(__OpenBSD__)
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <rpc/rpc.h>
#include <netdb.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>

#include <nfx-incl.h>
#include <nfx-defs.h>

#include <exports/evm-api.h> 
#include <exports/fs-api.h>
#include "rpc.h"
#include "mount.h"

#if defined(__OpenBSD__)
#include <exports/nfxnis-api.h>
#endif

#define NFS_PORT        2049
#define MNT_PORT        2087
#define NLM_PORT        2090
#define NSM_PORT		2092

//#include <exports/rpc.h>
//#include <exports/fs-api.h>
//#include <exports/mount.h>
//#include <exports/nfs.h>

#define MNT_DBG_PRINTF	if(mountd_trace_enable) printf

extern uint32 mountd_trace_enable;

/*-----------------------------------------------------------------
 * Name:  	mountd_isHostPresentInExpEntry 
 *  
 * Description:
 * 	Check if client or host is allow to access current export 
 *  entry. 
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int 	//found = 0: not found
		//found = 1: found
mountd_isHostPresentInExpEntry(
	mnt_infoReq_t *mnt_infoReqp,
	mnt_infoCB_t *mnt_InfoCBp,
	int uiFlag
)
{
	int         count;
    char        *exportInfop;
    int         ipAddr;
    char        infoStr[MNT_MAX_HOSTNAME_LEN];
    int         infoLen;
    int         foundHost;
    int         length;
    int         exportInfoLen;
    int         addr;

	foundHost = 0;	
    if ( (mnt_infoReqp == NULL) ||
        (mnt_InfoCBp == NULL) )
        return (foundHost);

    exportInfoLen = mnt_InfoCBp->mntInfoEntry.exportInfoLen;
    if (exportInfoLen == 0)
    {
        /*
         *No export specified, return OK
         */
        return (foundHost);
    }
    exportInfop = &mnt_InfoCBp->mntInfoEntry.pathName[0] +
                                mnt_InfoCBp->mntInfoEntry.pathNameLen;
    /*
     *locate host (by name or address) export info entry.
     *example:  IPaddr_String(options)
     *where,
     *IPAddr_String: Standard numbers-and-dots notation,
     * options is rw (read-write), ro (read-only),
     *root_squash or no_root_squash.
     */
	count = 0;
    ipAddr = mnt_infoReqp->ipAddr;
	
	MNT_DBG_PRINTF("mountd_isHostPresentInExpEntry 1: client IP: %08x\n", ipAddr);
	MNT_DBG_PRINTF("mountd_isHostPresentInExpEntry 2: pathname: %s len:%d export:%s len: %d\n", 
		&mnt_InfoCBp->mntInfoEntry.pathName[0], 
		mnt_InfoCBp->mntInfoEntry.pathNameLen,
		exportInfop, exportInfoLen);
	
 
	while ((*exportInfop != '\0') && (*exportInfop != '\n') &&
        (count < exportInfoLen) )
    {
        if (*exportInfop == ' ')
        {
            exportInfop++;
            continue;
        }
        if (*exportInfop == '(')
        {
            //Export entry does not contains host
            //allow all to access
            foundHost = 1;
            memset(infoStr, 0, sizeof(infoStr));
            mountd_storeHostPart(exportInfop,
                        exportInfoLen, &infoStr[0], &infoLen);
			if (!uiFlag)
			{
				return (foundHost);	
			}
            break;
        }
        /*
         *Extract host information.
         */
        memset(infoStr, 0, sizeof(infoStr));
        mountd_extractHost(infoStr, exportInfop, exportInfoLen);
        /*
         *Entry contains IP address
         */
        mountd_getIpAddr(&infoStr[0], &addr);

        if (ipAddr == ntohl(addr))
        {
            foundHost = 1;
            break;
        }
        /*
         *Get next host entry
         */
        length = mountd_getNextHost(exportInfop, exportInfoLen);
        exportInfop += length;
        count += length;

    }
				
	MNT_DBG_PRINTF("mountd_isHostPresentInExpEntry 3:  search client Addr: %08x curr entry clientaddr: %08x foundHost: %d\n",
		mnt_infoReqp->ipAddr, mnt_InfoCBp->mntInfoEntry.hostAddr, foundHost);
		
	if (!foundHost)
		return (foundHost);	
		
	MNT_DBG_PRINTF("mountd_isHostPresentInExpEntry 4:  search localIP: %08x Curr entry Ip: %08x \n", 
			ntohl(mnt_infoReqp->localIpAddr), 
			mnt_InfoCBp->mntInfoEntry.localIpAddr);
			

	if (mnt_InfoCBp->mntInfoEntry.localIpAddr && 
			(ntohl(mnt_infoReqp->localIpAddr) != mnt_InfoCBp->mntInfoEntry.localIpAddr) ) 
	{
		foundHost = 0;		
	}
	MNT_DBG_PRINTF("mountd_isHostPresentInExpEntry exit: foundHost: %d search localIP: %08x Curr entry Ip: %08x \n", 
			foundHost, mnt_infoReqp->localIpAddr, mnt_InfoCBp->mntInfoEntry.localIpAddr);

	return (foundHost);	

}

/*-----------------------------------------------------------------
 * Name:  	mountd_getAddrFromName 
 *
 * Description:
 *  Returns IP address based on input client name string. 
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_getAddrFromName(char *namep)
{
	struct in_addr	ip_addr;
	in_addr_t		host;

	//If string pointed by namep is an IP address, convert it to integer.
	if (inet_aton(namep, (struct in_addr*)&ip_addr) != 0)
	{
		//Valid IP address
		return(ip_addr.s_addr);
	}

	//String pointed by namep is a host name
	host = nfxnis_hostAddr("0.0.0.0", namep);
	if (!host)
	{
		printf("MOUNTD-ERR: mountd_getAddrFromName: nfxnis_hostAddr() failed\n");
		return (NFX_ERR);
	}
	return ((int)host);
}

/*-----------------------------------------------------------------
 * Name:   mountd_getIpAddr 
 *
 * Description:
 *	Convert an IP address string into 32 bits integer.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_getIpAddr( char *ipAddrStr, uint32 *addr)
{
	byte    ipAddr[4];
	char    *ip;
	char    *p;
	char    c;
	int32   i;

	ip = ipAddrStr;
	p = ip;
	c = *p;
	i = 0;

	while ( i < 4)
	{
		if ((c == '.') || (c == '\0'))
		{
			*p = '\0';
			ipAddr[i] = atoi(ip);
			ip = p+1;
			i++;
			if (i > 3)
				break;
		}
		p++;
		c = *p;
	}

	if (i < 4)
		return (NFX_ERR);
	*addr = (ipAddr[0] << 24) | (ipAddr[1] << 16) |
			(ipAddr[2] <<  8) | (ipAddr[3] );
	return (NFX_OK);
}

int mountd_extractIpaddr(char *inputStr, char *outputStr)
{
	int count;

	count = 0;
	while ((*inputStr) && (count < MNT_IP_ADDR_STRING_LEN) &&
			(*inputStr != '(') )
	{
		*outputStr++ = *inputStr++;
		count++;
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:  	mountd_extractHost 
 *
 * Description:
 *	Extract Host from export Info String.  Host could be a hostname or
 *	an IP address.
 *	Example of export Info string:
 * 		10.0.0.133(rw,no_root_squash) or
 * 		hostName(rw,no_root_squash)
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_extractHost(char *hostStr, char *exportInfop, int32 exportInfoLen)
{
	char    *exportp;
	int     count = 0;

	exportp = exportInfop;
	while ( (*exportp != '(') && (count < exportInfoLen) )
	{
		if ( (*exportp != ' ') || (*exportp != '\0') || 
			(*exportp != '\n')  ) 
			hostStr[count] = *exportp;
		exportp++;
		count++;
	}
	
	MNT_DBG_PRINTF("mountd_extractHost: host: %s len: %d\n", 
				hostStr, count);
	
	return (count);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_getNextHost 
 *
 * Description:
 *		Return length to next host in the export entry.
 *		example:
 *			hostName1(rw,no_root_squash) hostname2(ro) .... 
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_getNextHost(char *exportInfop, int exportInfoLen)
{
	char    *exportp;
	int     count = 0;

	exportp = exportInfop;
	while ( (*exportp != ')') && (count < exportInfoLen) )
	{
		exportp++;
		count++;
	}
	/*
	 *Skip pass ')'
	 */
	exportp++;
	count++;
	if ((*exportp == '\n') || (*exportp == '\0') )
	{
		return (count);
	}
	else
	{
		exportp++;
		count++;
	}
	return (count);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_removeSpaces 
 *
 * Description:
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
void mountd_removeSpaces(char *destp, char *sourcep, int length)
{
	int     count = 0;

	while ( *sourcep != ')' && count < length)
	{
		/*
		 *Skip spaces
		 */
		if (*sourcep == ' ')
		{
			count++;
			sourcep++;
			continue;
		}
		*destp++ = *sourcep++;
		count++;
	}
	*destp = ')';
}

/*-----------------------------------------------------------------
 * Name:   	mountd_extractOption 
 *
 * Description:
 * 	Extract option field in the current entry. 
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_extractOption(char *exportp, char *optionp)
{
	int     count;

	count = 0;
	while ( (count < MNT_EXPINFO_MAX_LEN) && (*exportp != '\0') && 
			(*exportp != '\n') && 
			(*exportp != ')') && (*exportp != ',') )
	{
		*optionp++ = *exportp++;
		count++;
	}
	if ( (*exportp == '\0') || (*exportp == '\n') )
		return (count);
	else
		return (count+1);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_checkOption 
 *
 * Description:
 * Check export options. It must be one the following options:
 * no_root_squash: 	Allows root users on client computers to 
 *					have root access.
 * root_squash:		Requests from root clients are mapped to 
 *					the nobody user and groupId only have privilleges
 *					associated with each other.
 * ro: 				read only access
 * rw: 				read write access
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_checkOption(char *optionStr)
{
	if (strcmp(optionStr, "ro") == 0)
	{
		return (MOUNTD_RO_OPT);
	}
	else if (strcmp(optionStr, "rw") == 0)
	{
		return (MOUNTD_RW_OPT);
	}
	else if (strcmp(optionStr, "no_root_squash") == 0)
	{
		return (MOUNTD_NO_ROOT_SQUASH_OPT);
	}
	else if (strcmp(optionStr, "root_squash") == 0)
	{
		return (MOUNTD_ROOT_SQUASH_OPT);
	}
	else
	{
		/*
		 *Invalid option
		 */
		return (0);
	}

}

/*-----------------------------------------------------------------
 * Name:   	mountd_processExportOptions 
 *
 * Description:
 * Process export options.
 * options:
 * no_root_squash:	Allows root users on client computers to have root access.
 * root_squash:  	Requests from root clients are mapped 
 *					to the nobody user and groupId
 *					only have privilleges associated with each other.
 * ro: 				read only access
 * rw: 				read write access
 * laddr: 			Check local interface address
 *		
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_processExportOptions(char *exportInfop, int length)
{
	char    *exportp;
	char    option[MNT_MAX_OPTION_LEN];
	int     count;
	int     option_len;
	int     optionMask;

	MNT_DBG_PRINTF("mountd_processExportOptions: export: %s len: %d\n", exportInfop, length);

	count = 0;
	exportp = exportInfop;
	while ((*exportp != '(') && (count < length) )
	{
		exportp++;
		count++;
	}
	exportp++;
	count++;

	if (count >= length)
		//Set default option read-only, root_squash
		optionMask = (MOUNTD_RW_OPT | MOUNTD_ROOT_SQUASH_OPT);
	else
		optionMask = 0;
	while ( (count < length) && 
			((*exportp != '\n') || ((*exportp != '\0')) ) )
	{
		memset(&option[0], 0, sizeof(option));
		option_len = mountd_extractOption(exportp, &option[0]);
		optionMask |= mountd_checkOption(&option[0]);
		count += option_len;
		exportp += option_len;
	}

	return (optionMask);
}

/*-----------------------------------------------------------------
 * Name:   mountd_isCharAlphaNumeric 
 *
 * Description:
 *  Check if input character is is alpha-numeric.
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_isCharAlphaNumeric(char *ch)
{

	if ( ((*ch >= '1') && (*ch <= '9')) ||
		( (*ch >= 'a') && (*ch <= 'z')) ||
		( (*ch >= 'A') && (*ch <= 'Z')) )
		return (1);

	return (0);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_isExportOptionPresent 
 *
 * Description:
 *	Check any export options is specified.
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_isExportOptionPresent(char *exportp, int length)
{
	int 	count;

	count = 0;
	
	while ( count < length) 
	{
		if ( (mountd_isCharAlphaNumeric(exportp)) || (*exportp == '(' )  )
			return (count);
		exportp++;
		count++;
	}
	return (0);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_parseClientNameField 
 *
 * Description:
 * 	Parse input client name field in the format of
 *  client:/pathname into its own buffer.
 * 
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int
mountd_parseClientNameField(char *nameFieldp, int nameFieldLen,
	char *clientNamep, int *clientNameLen,
	char *pathNamep, int *pathNameLen)
{
	char		*datap;		

	if ( (nameFieldp == NULL) || 
		(nameFieldLen == 0) )
		return (NFX_ERR);
	MNT_DBG_PRINTF("mountd_parseClientNameField 0: nameFieldp: %s len: %d\n",
			nameFieldp, nameFieldLen);

	datap = nameFieldp;

	*clientNameLen = 0;
	*pathNameLen = 0;
	if (*datap == '/')
	{
		datap++;
		*pathNameLen = nameFieldLen - 1;
		memcpy(pathNamep, datap, *pathNameLen);

    }
    else
    {
        //parse client & pathname
		while ( (*clientNameLen  < nameFieldLen) &&
                (*datap != ':') )
        {
            clientNamep[*clientNameLen] = *datap;
			*clientNameLen += 1;
            datap++;
        }
        datap++;		//skip ':'
		if (*datap++ != '/')
		{
			printf("MOUNTD-ERR: Format error! (client:/directory)\n");
			return (NFX_ERR);
		}
		//account for ':/'	
        *pathNameLen = nameFieldLen - *clientNameLen - 1 - 1;
		if ( (*pathNameLen <= 0) || (nameFieldLen == *clientNameLen))
			return (NFX_ERR);

        memcpy(pathNamep, datap, *pathNameLen);
    }

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:        mountd_getVolumeName
 *
 * Description:
 * Extract volume name from the input pathname.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int
mountd_getVolumeName(char *volNamep, int *volNameLen,
	char *pathNamep, int pathNameLen)
{
	int	count = 0;
   
	while ((pathNameLen--) && (pathNamep) && (*pathNamep != '/') &&
		  (*pathNamep != ' ') && (*pathNamep != '\t') )
	{
		*volNamep++ = *pathNamep++;
		count++;
	}
	*volNameLen = count;
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:mountd_portmapRegMountPort	 
 *
 * Description:
 *	Register all portmapper application MOUNT ports.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */

void mountd_portmapRegMountPort(void)
{

	int32               status;

	/*
	 *UNSET
	 */
	status = pmap_unset(MNT_PROG, MNT_VER1);
	if (status == 0)
    {
        MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET MOUNT V1!\n");
    }

	status = pmap_unset(MNT_PROG, MNT_VER2);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET MOUNT V2!\n");
	}

	/*
	 *Register for UDP
	 */
    status = pmap_set(MNT_PROG, MNT_VER1, IPPROTO_UDP, MNT_PORT);
    if (status == 0)
    {
        MNT_DBG_PRINTF("PORTMAP: Failure to register MOUNT V1 UDP on port %d\n",
                MNT_PORT);
    }

	status = pmap_set(MNT_PROG, MNT_VER2, IPPROTO_UDP, MNT_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register MOUNT V2 UDP on port %d\n",
				MNT_PORT);
	}

	/*
	 *Register services for V3
	 */
	/*
	 *UNSET
	 */
	status = pmap_unset(MNT_PROG, MNT_VER3);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET MOUNT V3!\n");
	}
	/*
	 *Register for UDP
	 */
	status = pmap_set(MNT_PROG, MNT_VER3, IPPROTO_UDP, MNT_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register MOUNT V3 UDP on port %d\n",
				MNT_PORT);
	}

	/*
	 *Register for TCP
	 */
	status = pmap_set(MNT_PROG, MNT_VER1, IPPROTO_TCP, MNT_PORT);
	if (status == 0)
    {
        MNT_DBG_PRINTF("PORTMAP: Failure to register MOUNT V1 TCP on port %d\n",
                MNT_PORT);
    }

	//NFS V2
	status = pmap_set(MNT_PROG, MNT_VER2, IPPROTO_TCP, MNT_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register MOUNT V2 TCP on port %d\n",
				MNT_PORT);
	}

	//NFS V3
	status = pmap_set(MNT_PROG, MNT_VER3, IPPROTO_TCP, MNT_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register MOUNT V3 TCP on port %d\n",
				MNT_PORT);
	}

	return;
}

/*-----------------------------------------------------------------
 * Name:mountd_portmapRegNfsPort	 
 *
 * Description:
 *	Register all portmapper application NFS ports.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */

int32 mountd_portmapRegNfsPort(void)
{
	int32               status;

	/*
	 *UNSET
	 */
	status = pmap_unset(NFS_PROG, NFS_VER2);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET MOUNT V2!\n");
	}

	status = pmap_set(NFS_PROG, NFS_VER2, IPPROTO_UDP, NFS_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NFS V2 UDP on port %d\n",
				NFS_PORT);
	}

	/*
	 *Register services for NFS V3
	 */
	/*
	 *UNSET
	 */
	status = pmap_unset(NFS_PROG, NFS_VER3);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET MOUNT V3!\n");
	}

	/*
	 *Register for UDP
	 */
	status = pmap_set(NFS_PROG, NFS_VER3, IPPROTO_UDP, NFS_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NFS V3 UDP on port %d\n",
				NFS_PORT);
	}

	status = pmap_set(NFS_PROG, NFS_VER2, IPPROTO_TCP, NFS_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NFS V2 TCP on port %d\n",
				NFS_PORT);
	}

	status = pmap_set(NFS_PROG, NFS_VER3, IPPROTO_TCP, NFS_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NFS V3 TCP on port %d\n",
				NFS_PORT);
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: mountd_portmapRegNlmPort	 
 *
 * Description:
 *	Register all portmapper application NLM ports.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */

int32 mountd_portmapRegNlmPort(void)
{
	int32		status;
	//Register for NLM
	status = pmap_unset(NLM_PROG, NLM_VER1);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET NLM V1!\n");
	}
	/*
	 *Register for UDP & TCP
	 */
	status = pmap_set(NLM_PROG, NLM_VER1, IPPROTO_UDP,
					NLM_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NLM V1 UDP on port %d\n",
			NLM_PORT);
	}

	status = pmap_set(NLM_PROG, NLM_VER1, IPPROTO_TCP,
					NLM_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NLM V1 TCP on port %d\n",
				NLM_PORT);
	}

	status = pmap_unset(NLM_PROG, NLM_VER3);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET NLM V3!\n");
	}
	/*
	 *Register for UDP & TCP
	 */
	status = pmap_set(NLM_PROG, NLM_VER3, IPPROTO_UDP,
					NLM_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NLM V3 UDP on port %d\n",
					NLM_PORT);
	}
   
	status = pmap_set(NLM_PROG, NLM_VER3, IPPROTO_TCP,
					NLM_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NLM V3 TCP on port %d\n",
				NLM_PORT);
	}


	status = pmap_unset(NLM_PROG, NLM_VER4);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET NLM V4!\n");
	}
	/*
	 *Register for UDP & TCP
	 */
    status = pmap_set(NLM_PROG, NLM_VER4, IPPROTO_UDP,
                    NLM_PORT);
    if (status == 0)
    {
        MNT_DBG_PRINTF("PORTMAP: Failure to register  NLM V1 UDP on port %d\n",
                NLM_PORT);
    }

	status = pmap_set(NLM_PROG, NLM_VER4, IPPROTO_TCP,
                    NLM_PORT);
    if (status == 0)
    {
        MNT_DBG_PRINTF("PORTMAP: Failure to register  NLM V1 TCP on port %d\n",
                NLM_PORT);
    }
	return (NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:  mountd_portmapRegNsmPort 	
 *
 * Description:
 *	Register all portmapper application NSM ports.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */

int32 mountd_portmapRegNsmPort(void)
{
	int32		status;
	//Register for NSM
	status = pmap_unset(NSM_PROG, NSM_VER1);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to PMAP_UNSET NSM V1!\n");
	}
	/*
	 *Register for UDP & TCP
	 */
	status = pmap_set(NSM_PROG, NSM_VER1, IPPROTO_UDP,
					NSM_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NSM V1 UDP on port %d\n",
			NSM_PORT);
	}

	status = pmap_set(NSM_PROG, NSM_VER1, IPPROTO_TCP,
					NSM_PORT);
	if (status == 0)
	{
		MNT_DBG_PRINTF("PORTMAP: Failure to register  NSM V1 TCP on port %d\n",
				NSM_PORT);
	}
	return (NFX_OK); 
}


/*-----------------------------------------------------------------
 * Name:   	mountd_portmap_register 
 *
 * Description:
 *	Register all portmapper application ports.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int32 mountd_portmap_register(void)
{

	//Register MOUNT port
	mountd_portmapRegMountPort();

	//Register NFS port
	mountd_portmapRegNfsPort();

	//Register NLM port
	mountd_portmapRegNlmPort();
	//MNT_DBG_PRINTF("PORTMAP: Register MOUNT/NFS/NLM Ports OK\n");

	//Register NSM port
	mountd_portmapRegNsmPort();

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:  	mountd_getLocalAddr 
 *
 * Description:
 *	Retrieves local address specified in export option field.
 *		
 * Created by:      Joseph Vu
 * Modified by:		Xiaobing Zhang
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_getLocalAddr(char *exportp, int *localAddr)
{
	int			count;
	int			optionLen;
	char		option[MNT_MAX_OPTION_LEN];
	char		*localAddrp;
	struct in_addr		addr;

	MNT_DBG_PRINTF("mountd_getLocalAddr 1: exportp: %s\n", exportp);
	*localAddr = 0;
	count = 0;

	/*
	 *Advance ptr to option field, where export line:
	 *filesystem	a.b.c.d(rw,no_root_squash,laddr=a.b.c.d)
	 *or            (rw,no_root_squash,laddr=a.b.c.d)
	 */
	while ((*exportp != '(') && (count < MNT_MAX_OPTION_LEN) )
	{
		exportp++;
		count++;
	}
	if (count >= MNT_MAX_OPTION_LEN)
		return (NFX_ERR);
	exportp++;
	count++;
	MNT_DBG_PRINTF("mountd_getLocalAddr 2: exportp: %s count: %d\n", exportp, count);

	while ( (count < MNT_MAX_OPTION_LEN) && (*exportp) && 
			(*exportp != ')') )
	{

		memset(&option[0], 0, sizeof(option));
		optionLen = mountd_extractOption(exportp, &option[0]);
		if (!optionLen)
			break;
		MNT_DBG_PRINTF("mountd_getLocalAddr 3: option: %s len: %d\n", option, optionLen);
		if (!(strncmp(&option[0], MNT_LADDR_OPT_STR,
				strlen(MNT_LADDR_OPT_STR))))
		{
			/*
			 *Found local address option
			 */
			localAddrp = &option[0] + strlen(MNT_LADDR_OPT_STR);
			MNT_DBG_PRINTF("mountd_getLocalAddr 4: localAddrp: %s\n", localAddrp);
			/*
			if (mountd_getIpAddr(localAddrp, &addr) == NFX_OK)
			{
				*localAddr = addr;
			}
			else
			{
				*localAddr = -1;
				return (NFX_ERR);
			}
			*/

			if (inet_aton(localAddrp, &addr) != 0)
			{
				*localAddr = addr.s_addr;
			}
			else
			{
				*localAddr = -1;
				return (NFX_ERR);
			}
			break;	
		}
		
        count += optionLen;
        exportp += optionLen;
    }
	MNT_DBG_PRINTF("mountd_getLocalAddr: exit: localAddr: %08x\n", *localAddr);
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_storeHostPart 
 *
 * Description:
 * Stores host part of the export info record
 * into new export buffer.  
 * Example:  IPAddr_Str(options)
 * where, IPAddr_Str is in standard dot notations.
 *   
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_storeHostPart(
	char *currExportp, 
	int currExportLen, 
	char *newExport,
	int *newLen
)
{
	int	count = 0;

	MNT_DBG_PRINTF("mountd_storeHostPart: currExportp: %s len: %d\n",
				currExportp, currExportLen);
	while ((*currExportp) && (count < currExportLen) &&
			(*currExportp != ')') )
	{
		newExport[count] = *currExportp;
		currExportp++;
		count++;	
	}
	//skip pass ')'
	newExport[count] = *currExportp++;
	count++;

	if ((*currExportp == '\n') || (*currExportp == '\0') )
	{
		*newLen += count;
		return (count);
	}
	newExport[count] = ' ';
	count++;
	*newLen += count;
	return (count);

}

/*-----------------------------------------------------------------
 * Name:   	mountd_storeHostPartByName 
 *
 * Description:
 * Stores host part of the export info record
 * into new export buffer, where host field is
 * specified by name.  We need to consult NIS
 * server to resolve hostname.
 * Example:  clientName(options)
 *
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_storeHostPartByName(
	char *currExportp, 		//curent export info record
	int currExportLen, 		//current export info len in bytes
	char *newExport,		//new parse export info record
	int *newLen,			//new export info len
	int *hostAddrp 			//return 32-bits IP address
)
{
	int					rc;
	int					localAddr;
	int					totalLen;
	int					length;
	int					ipAddrLen;
	char				*ipAddr;
	char				hostName[MNT_MAX_HOSTNAME_LEN];
	struct hostent		*host;
	struct in_addr		hostAddr; 

	MNT_DBG_PRINTF("mountd_storeHostPartByName 1:currExportp:%s len: %d\n", currExportp, currExportLen);

	//example:  currExportp:clientName(rw,laddr=10.2.2.200) or
	//a.b.c.d(rw,laddr=10.2.2.200)
	totalLen = 0;
	memset(hostName, 0, sizeof(hostName));	

	//extract hostname
	length = mountd_extractHost(hostName, 
			currExportp, MNT_MAX_HOSTNAME_LEN);

#if defined(__OpenBSD__)
	//Check if 'laddr=' option is specified
	rc = mountd_getLocalAddr(currExportp, &localAddr);
	MNT_DBG_PRINTF("mountd_storeHostPartByName 2: localAddr: %08x\n", localAddr);
	/*-JVU NOTUSED
	 *Check with Greg how this API is been used.
	if ((rc == NFX_OK) && localAddr)
	{
		//resolve name under NIS domain of that 
		//local interface (localAddr)
		hostAddr.s_addr = ntohl(localAddr);
		//ipAddr = inet_ntoa(hostAddr);	
		MNT_DBG_PRINTF("mountd_storeHostPartByName 2.5: ipAddr: %s hostName: %s\n", inet_ntoa(hostAddr), hostName);
		hostAddr.s_addr = nfxnis_hostAddr(inet_ntoa(hostAddr), hostName);
		MNT_DBG_PRINTF("mountd_storeHostPartByName 3:hostName:%s addr:0x%08x\n",
					hostName, ntohl(hostAddr.s_addr));
	}
	else
	
	*/
	{
		MNT_DBG_PRINTF("mountd_storeHostPartByName 2.6: ipAddr: %s hostName: %s\n", 
														inet_ntoa(hostAddr), hostName);
		MNT_DBG_PRINTF("mountd_storeHostPartByName  2.7: hostname: %s addr: %08x\n",
							hostName, ntohl(nfxnis_hostAddr(inet_ntoa(hostAddr), hostName)) );
		//Use default NIS server 
		host = gethostbyname(hostName);
		if (!host)
		{
			printf("MOUNTD-ERR: gethostbyname (%s)\n", hostName);
			return (NFX_ERR);
		}
		//hostAddr.s_addr = *((uint32 *)host->h_addr); 
		memcpy(&hostAddr.s_addr, host->h_addr, host->h_length);
		//printf("mountd_storeHostPartByName 3.5:hostName:%s addr:0x%08x\n",
		//				hostName, ntohl(hostAddr.s_addr));
	}
#else
	host = gethostbyname(hostName);
	if (!host)
	{
		printf("MOUNTD-ERR: gethostbyname (%s)\n", hostName);
		return (NFX_ERR);
	}

	//hostAddr.s_addr = *((uint32 *)host->h_addr); 
	memcpy(&hostAddr.s_addr, host->h_addr, host->h_length);

#endif
	//Set return Host addr to caller
	*hostAddrp = *((uint32*)host->h_addr);

	MNT_DBG_PRINTF("mountd_storeHostPartByName 5: host Addr: 0x%08x\n", 
			*hostAddrp);
	//convert IP address into ASCII			
	ipAddr = inet_ntoa(hostAddr);	

	MNT_DBG_PRINTF("hostName: %s addr: %s\n",
			hostName, ipAddr);	
	ipAddrLen = strlen(ipAddr);
	memcpy(newExport, ipAddr, ipAddrLen);
	newExport += ipAddrLen;
	*newLen += ipAddrLen;

	//advance pointer pass the client hostname or IP address string
	currExportp += length;
	currExportLen -= length;

	length +=  mountd_storeHostPart(currExportp,
				currExportLen, newExport, newLen);
	return (length);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_processHostPart 
 *
 * Description:
 * Process host or client part of the export info record (currExportp).
 * The host can be specified by IPAddr (in standard dot notation)
 * or hostname.  If host is specified by name we need to consult
 * NIS to revolve entry.  Return new data in input 
 * buffer (newExportp), its update length (newExportLen) and the
 * host 32 bits IP address (hostAddrp).
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int mountd_processHostPart(
	char *currExportp,			//curent export entry pointer
	int currExportLen,			//current export length
	char *newExportp,			//return parsed new export info 
	int *newExportLen,			//return update len of new export info
	int *hostAddrp				//return 32-bits IP address
)
{
	int			count;
	int			length;
	int			newLen;
	char		ipAddrStr[MNT_IP_ADDR_STRING_LEN];
	int			ipAddr;

	if ( (currExportp == NULL) || (newExportp == NULL) ||
		(currExportLen <= 0) )
		return (NFX_ERR);

	MNT_DBG_PRINTF("mountd_processHostPart: currExportp: %s len: %d\n",
			currExportp, currExportLen);

	//example: currExportp: clientName(rw,laddr=10.2.2.200) or
	//a.b.c.d(rw,laddr=10.2.2.200)

	*newExportLen = 0;
	newLen = 0;
	count = 0;
	ipAddr = 0;
	/*
	 *locate host (by name or address) export info entry.
	 *example:  hostname(options) or IPaddr(options)
	 */

	while ((*currExportp != '\0') && (*currExportp != '\n') && 
			(count < currExportLen) )
	{
		
		if (*currExportp == ' ')
		{
			/*skip spaces*/
			currExportp++;
			continue;
		}

		/*
		 *Extract host information.
		 *Host could be a host name or an IP address
		 */
		if (*currExportp == '(')
		{
			//No host specified, must be exported 'all'
			//for this entry
			MNT_DBG_PRINTF("mountd_processHostPart 2: all case\n");
			length = mountd_storeHostPart(currExportp, 
						currExportLen, newExportp, &newLen);			
			*newExportLen += newLen;
			break;
		}
		else if ((*currExportp >= '0') && (*currExportp <= '9'))
		{
			/*
			 *Entry contains IP address
			 */
			length = mountd_storeHostPart(currExportp, 
						currExportLen, newExportp, &newLen);			
			memset(&ipAddrStr[0], 0, sizeof(ipAddrStr));
			memcpy(&ipAddrStr[0], newExportp, newLen);
			MNT_DBG_PRINTF("mountd_processHostPart 3: ipAddrStr: %s\n", &ipAddrStr[0]);
			if (mountd_getIpAddr(&ipAddrStr[0], &ipAddr) == NFX_OK)
				*hostAddrp = ipAddr;			

		}
		else
		{
			/*
			 *Entry contains hostName
			 */
			MNT_DBG_PRINTF("mountd_processHostPart 4: entry contain host name\n");
			length =
					mountd_storeHostPartByName(currExportp, 
							currExportLen, newExportp, &newLen, &ipAddr);
			if (length == NFX_ERR)
			{
				printf("mountd_storeHostPartByName failed!\n");
				//return (count);
				return (NFX_ERR);
			}
			//Set return IP address
			*hostAddrp = ipAddr;			
		}
        /*
         *Get next host entry
         */
		newExportp += newLen;
		*newExportLen += newLen;
        currExportp += length;
        count += length;
		newLen = 0;

	}
	MNT_DBG_PRINTF("mountd_processHostPart exit:  hostAddr%08x\n", *hostAddrp);
	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_buildOptionString
 *
 * Description:
 * Given option mask, build the corresponding option string.
 *
 * Created by:      Xiaobing Zhang
 *
 * Date Created:    05/21/2002
 *
 *-----------------------------------------------------------------
 */
void mountd_buildOptionString(char *destStr, uint32 optionMask)
{
	char	*idx;

	if (optionMask & MOUNTD_RW_OPT)
	{
		strcat(destStr, MNT_RW_OPT_STR);
		strcat(destStr, ",");
	}

	if (optionMask & MOUNTD_RO_OPT)
	{
		strcat(destStr, MNT_RO_OPT_STR);
		strcat(destStr, ",");
	}

	if (optionMask & MOUNTD_ROOT_SQUASH_OPT)
	{
		strcat(destStr, MNT_ROOT_SQUASH_STR);
		strcat(destStr, ",");
	}

	if (optionMask & MOUNTD_NO_ROOT_SQUASH_OPT)
	{
		strcat(destStr, MNT_NO_ROOT_SQUASH_STR);
		strcat(destStr, ",");
	}

	idx = rindex(destStr, ',');
	*idx = 0;

	return;
}

/*-----------------------------------------------------------------
 * Name:       	  mountd_pathNameCmp
 *
 * Description:
 *		Compare two export path names.
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        08/13/2002
 *
 *-----------------------------------------------------------------
 */
int
mountd_pathNameCmp(char *name1, char *name2)
{
	int		rc;
	char	tmpName1[VOL_MAX_NAME], tmpName2[VOL_MAX_NAME];
	char	*p1, *p2;
	char	*tok1, *tok2;


	strcpy(tmpName1, name1);
	strcpy(tmpName2, name2);

	p1 = tmpName1;
	p2 = tmpName2;
	while(1)
	{
		// Compare one component in the paths
		tok1 = strsep(&p1, "/");
		tok2 = strsep(&p2, "/");

		rc = strcmp(tok1, tok2);
		if (rc != 0)
			return (rc);

		// Consider the case where either p1 or p2 is NULL
		if (p1 == NULL && p2 != NULL)
			return (-1);				// name1 < name2

		if (p1 != NULL && p2 == NULL)
			return (1);					// name1 > name2

		if (p1 == NULL && p2 == NULL)	// name1 == name2
			return (0);
	}
}

/*-----------------------------------------------------------------
 * Name:       	  mountd_ipAddrCmp
 *
 * Description:
 *		Compare two IP address strings.
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        08/13/2002
 *
 *-----------------------------------------------------------------
 */
int
mountd_ipAddrCmp(unsigned int ipAddr1, unsigned int ipAddr2)
{
	int				rc;
	struct in_addr	addr1, addr2;
	char			addrStr1[50], addrStr2[50];

	addr1.s_addr = ipAddr1;
	addr2.s_addr = ipAddr2;

	// Convert interger to string
	strcpy(addrStr1, inet_ntoa(addr1));
	strcpy(addrStr2, inet_ntoa(addr2));

	rc = strcmp(addrStr1, addrStr2);
	return (rc);
}


