/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        mountd.c
 *
 * RCS:  $Id: mountd.c,v 1.75 2002/10/02 22:13:15 joseph Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *	MOUNTD on SSC.
 * Created by: Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#include <nfx-incl.h>
#include <nfx-defs.h>

#define	MNT_DBG_PRINTF	if(mountd_trace_enable) printf

/*
#define	MOUNTD_DEBUG
*/


#define	LOG_ENABLED

#ifdef	LOG_ENABLED
#define	MOUNTD_ELOG		E_LOG
#else
#define	MOUNTD_ELOG
#endif

#define	MOUNTD_ALLOC		malloc
#define	MOUNTD_FREE			free

#include "exports/evm-api.h"
#include "exports/anpssc-api.h"
#include "exports/fs-api.h"
#include "rpc.h"
#include "mount.h" 
//#include <exports/nfs.h>
#include "exports/elog-api.h"

#include "exports/event-api.h"

#include "exports/ncm-api.h"



#define MOUNTD_STR					"MOUNTD"
#define	MOUNTD_OPEN_FILE_ERR_STR	"[%s]Failure to open file [%s] for reading\n"
#define MOUNTD_GET_VOLID_ERR_STR    "[%s]Failure to get volume ID for volume [%s]\n"
#define MOUNTD_REG_APP_ERR_STR  	"[%s]Failure to register for [%s]\n"
#define MOUNTD_GET_APPID_ERR_STR  	"[%s]Failure to get AppId for [%s]\n"
#define MOUNTD_RECV_ERR_STR			"[%s]Received error. Error from recvAgileMsg!\n"
#define MOUNTD_GETHOST_ERR_STR		"[%s]gethostbyaddr error for volume [%s]n"
#define MOUNTD_LOCATE_NAME_ERR_STR	"[%s]Failure to locate share dir [%s] entry\n"
#define MOUNTD_LOCATE_ID_ERR_STR	"[%s]Stale handle: volId [%llx] mntPnt [%llx]\n"
#define MOUNTD_BAD_VOLID_STR		"[%s]Bad volId for volume [%s]\n"
#define MOUNTD_PERM_DENIED_STR 		"[%s]Permission denied for client [x%08x] on pathname [%s]\n"
#define MOUNTD_INV_MSG_STR			"[%s]Invalid message from app [%s] Cpu [%d] Slot [%d]\n"
#define MOUNTD_SNDRSP_MSG_STR		"[%s]Send resp to app [%s] Cpu [%d] Slot [%d]\n"
#define MOUNTD_SENDRSP_ERR_STR  	"[%s]Error sending response!\n"
#define MOUNTD_DUP_EXPORT_STR		"[%s]Export [%s] is a duplicate entry\n"
#define MOUNTD_REPLACE_EXPORT_STR	"[%s]Export [%s] entry is been replaced\n"

#define MOUNTD_RCV_GETNAME_INFO_STR "[%s]Client IPAddr [%08x] pathname (%s)\n"
#define MOUNTD_EXPORT_DEL_INFO_STR 	"[%s]Export entry [%s] has been deleted\n"
#define MOUNTD_EXPORT_ADD_INFO_STR	"[%s]Export entry [%s] has been added\n"
#define MOUNTD_EXPORT_INFO_STR		"[%s]Received export request\n"
#define MOUNTD_UNEXPORT_INFO_STR	"[%s]Received unexport request\n"
#define MOUNTD_GET_VOLID_INFO_STR	"[%s]Received Id (%lld) for volume (%s) from EVM\n"
#define MOUNTD_EXPORT_DEL_ERR_STR	"[%s]Delete export entry [%s] not found!\n"
#define MOUNTD_INV_RX_IF_ERR_STR  	"[%s]Access Denied.  Rcvd on invalid interface [0x%08x]\n"
#define MOUNTD_EXPORT_ERR_STR		"[%s]Invalid export entry [%s]!\n"
#define MOUNTD_EXPORT_SEND_CCM_ERR_STR	"[%s]Failure to send export message to CCM rc [%d]\n"
#define MOUNTD_EXPORT_PERM_ERR_STR	"[%s]Invalid option permission [%s]!\n"
#define MOUNTD_EXPORT_OPT_ERR_STR	"[%s]Invalid export option [%s]!\n"
#define MOUNTD_FS_NOTREADY_STR		"[%s]File system [%s] is not in ONLINE state\n"

#define MOUNTD_ALL_ADDR		0xffffffff

mountd_globalInfo_t	mountd_globalInfo;
mountd_globalInfo_t	*mountd_globalInfop;
//ushort16			mnt_infoNextIndex = 0;

int					mnt_firstReq = 0;

uint32				mountd_trace_enable = 0;


/*-----------------------------------------------------------------------------
 * Define export entry search rules for adding a share.
 *
 * The following table illustrates the logic.
 *
 * o Columns "local A" and "client A" are local IP address and client IP address
 *   of the existing export entry.
 *
 * o Columns "local B" and "client B" are local IP address and client IP address
 *   of the export entry to be added.
 *
 *    "all" means all the interfaces or all the clients.
 *    "L1" and "L2" represents two different local interface.
 *    "C1" and "C2" represents two different client.
 *
 * o Column "option" specifies whether the export option of the existing
 *   entry and the new entry is same or different.
 *
 * o Column "result" shows whether the new entry is allowed to be added or not;
 *   if not, it also gives the reason, represented in D, G, S, C, M, where
 *       Y: Allowed.
 *       N: Disallowed.
 *       D: The same entry exists.
 *       G: There exists a more Generic entry.
 *       S: There exists a more Specific entry.
 *       C: There exists a Conflict entry.
 *       M: For simplicity, Multiple local addresses for a pathname 
 *          is not allowed.
 *
 * =============================================================================
 *  local A  client A  local B  client B   option     result
 *  -------  --------  -------  --------   ------     ------
 * 1.  all    all       all      all        same       N (D)
 * 2.  all    all       all      all        diff       N (C)
 * 3.  all    all       all      C2         same       N (G)
 * 4.  all    all       all      C2         diff       Y
 * 5.  all    all       L2       all        same       N (G)
 * 6.  all    all       L2       all        diff       N (M)
 * 7.  all    all       L2       C2         same       N (G)
 * 8.  all    all       L2       C2         diff       N (M)
 * 
 * 9.  all    C1        all      all        same       N (S)
 * 10. all    C1        all      all        diff       Y
 * 11. all    C1        all      C2         same       Y
 * 12. all    C1        all      C2         diff       Y
 * 13. all    C1        L2       all        same       N (M)
 * 14. all    C1        L2       all        diff       N (C)
 * 15. all    C1        L2       C2         same       N (M)
 * 16. all    C1        L2       C2         diff       N (M)
 * 
 * 17. all    C1        all      C1         same       N (D)
 * 18. all    C1        all      C1         diff       N (C)
 * 19. all    C1        L2       C1         same       N (G)
 * 20. all    C1        L2       C1         diff       N (M)
 * 
 * 21. L1     all       all      all        same       N (S)
 * 22. L1     all       all      all        diff       N (M)
 * 23. L1     all       all      C2         same       N (M)
 * 24. L1     all       all      C2         diff       N (C)
 * 25. L1     all       L2       all        same       N (M)
 * 26. L1     all       L2       all        diff       N (M)
 * 27. L1     all       L2       C2         same       N (M)
 * 28. L1     all       L2       C2         diff       N (M)
 * 
 * 29. L1     all       L1       all        same       N (D)
 * 30. L1     all       L1       all        diff       N (C)
 * 31. L1     all       L1       C2         same       N (G)
 * 32. L1     all       L1       C2         diff       Y
 * 
 * 33. L1     C1        all      all        same       N (S)
 * 34. L1     C1        all      all        diff       N (M)
 * 35. L1     C1        all      C2         same       N (M)
 * 36. L1     C1        all      C2         diff       N (M)
 * 37. L1     C1        L2       all        same       N (M)
 * 38. L1     C1        L2       all        diff       N (M)
 * 39. L1     C1        L2       C2         same       N (M)
 * 40. L1     C1        L2       C2         diff       N (M)
 * 
 * 41. L1     C1        all      C1         same       N (S)
 * 42. L1     C1        all      C1         diff       N (M)
 * 43. L1     C1        L1       all        same       N (S)
 * 44. L1     C1        L1       all        diff       Y
 * 
 * 45. L1     C1        L2       C1         same       N (M)
 * 46. L1     C1        L2       C1         diff       N (M)
 * 47. L1     C1        L1       C2         same       Y
 * 48. L1     C1        L1       C2         diff       Y
 * 
 * 49. L1     C1        L1       C1         same       N (D)
 * 50. L1     C1        L1       C1         diff       N (C)
 * ============================================================================
 *
 *
 * We implement the rule using a 5 dimension array of 2x2x3x3x2:
 *    array[localA][clientA][localB][clientB][option]
 * where,
 *     all            = 0
 *     L1             = 1
 *     C1             = 1
 *     L2             = 2
 *     C2             = 2
 *     same           = 0
 *     diff           = 1
 *     Y              = NFX_OK
 *     N              = NFX_ERR
 *     D              = (MOUNTD_ERR_DUPLICATE_EXPORT)
 *     G              = (MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND)
 *     S              = (MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND)
 *     C              = (MOUNTD_ERR_CONFLICT_EXPORT)
 *     M              = (MOUNTD_ERR_MULTIPLE_INTERFACE)
 *
 *  local A  client A  local B  client B  option       result
 *  -------  --------  -------  --------  ------       ------
 * 1.  0      0         0        0          0           N (D)
 * 2.  0      0         0        0          1           N (C)
 * 3.  0      0         0        2          0           N (G)
 * 4.  0      0         0        2          1           Y
 * 5.  0      0         2        0          0           N (G)
 * 6.  0      0         2        0          1           N (M)
 * 7.  0      0         2        2          0           N (G)
 * 8.  0      0         2        2          1           N (M)
 * 
 * 9.  0      1         0        0          0           N (S)
 * 10. 0      1         0        0          1           Y
 * 11. 0      1         0        2          0           Y
 * 12. 0      1         0        2          1           Y
 * 13. 0      1         2        0          0           N (M)
 * 14. 0      1         2        0          1           N (C)
 * 15. 0      1         2        2          0           N (M)
 * 16. 0      1         2        2          1           N (M)
 * 
 * 17. 0      1         0        1          0           N (D)
 * 18. 0      1         0        1          1           N (C)
 * 19. 0      1         2        1          0           N (G)
 * 20. 0      1         2        1          1           N (M)
 * 
 * 21. 1      0         0        0          0           N (S)
 * 22. 1      0         0        0          1           N (M)
 * 23. 1      0         0        2          0           N (M)
 * 24. 1      0         0        2          1           N (C)
 * 25. 1      0         2        0          0           N (M)
 * 26. 1      0         2        0          1           N (M)
 * 27. 1      0         2        2          0           N (M)
 * 28. 1      0         2        2          1           N (M)
 * 
 * 29. 1      0         1        0          0           N (D)
 * 30. 1      0         1        0          1           N (C)
 * 31. 1      0         1        2          0           N (G)
 * 32. 1      0         1        2          1           Y
 * 
 * 33. 1      1         0        0          0           N (S)
 * 34. 1      1         0        0          1           N (M)
 * 35. 1      1         0        2          0           N (M)
 * 36. 1      1         0        2          1           N (M)
 * 37. 1      1         2        0          0           N (M)
 * 38. 1      1         2        0          1           N (M)
 * 39. 1      1         2        2          0           N (M)
 * 40. 1      1         2        2          1           N (M)
 * 
 * 41. 1      1         0        1          0           N (S)
 * 42. 1      1         0        1          1           N (M)
 * 43. 1      1         1        0          0           N (S)
 * 44. 1      1         1        0          1           Y
 * 
 * 45. 1      1         2        1          0           N (M)
 * 46. 1      1         2        1          1           N (M)
 * 47. 1      1         1        2          0           Y
 * 48. 1      1         1        2          1           Y
 * 
 * 49. 1      1         1        1          0           N (D)
 * 50. 1      1         1        1          1           N (C)
 * ============================================================================
 *
 */

/* Definition of array size */
#define MOUNTD_RULEARRAY_S_LA		2
#define MOUNTD_RULEARRAY_S_CA		2
#define MOUNTD_RULEARRAY_S_LB		3
#define MOUNTD_RULEARRAY_S_CB		3
#define MOUNTD_RULEARRAY_S_OPT		2

char mountd_ruleStrLocalAddr[3][4] = {"all", "L1", "L2"};
char mountd_ruleStrClientAddr[3][4] = {"all", "C1", "C2"};
char mountd_ruleStrOption[2][5] = {"same", "diff"};

/* Defintion of array element structure */
typedef struct mountd_findRuleForExport
{
	int32	validFlag;		//Valid entry if set to 1
	int32	result;			//NFX_OK or NFX_ERR
	int32	reason;			//mountd error code
	short16	counter;		//Number of times the rule is hit 
	short16	changeCnt;		//Number of times the rule is hit since last display
	char	rrStr[4];		//Result and reason string
} mountd_findRuleForExport_t;

mountd_findRuleForExport_t 
	mountd_findRules[MOUNTD_RULEARRAY_S_LA]\
					[MOUNTD_RULEARRAY_S_CA]\
					[MOUNTD_RULEARRAY_S_LB]\
					[MOUNTD_RULEARRAY_S_CB]\
					[MOUNTD_RULEARRAY_S_OPT];

/*-----------------------------------------------------------------
 * Name:      	mountd_findRulesInit
 *
 * Description:
 *     	Initialize the find rules array
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/22/2002
 *
 *-----------------------------------------------------------------
 */
void mountd_findRulesInit()
{
	mountd_findRuleForExport_t Y, ND, NG, NS, NC, NM;

	memset(mountd_findRules, 0, sizeof(mountd_findRules));

	memset(&Y, 0, sizeof(mountd_findRuleForExport_t));
	memset(&ND, 0, sizeof(mountd_findRuleForExport_t));
	memset(&NG, 0, sizeof(mountd_findRuleForExport_t));
	memset(&NS, 0, sizeof(mountd_findRuleForExport_t));
	memset(&NC, 0, sizeof(mountd_findRuleForExport_t));
	memset(&NM, 0, sizeof(mountd_findRuleForExport_t));

	Y.validFlag = 1;
	Y.result = NFX_OK;
	Y.reason = 0;
	strcpy(Y.rrStr, "Y");

	ND.validFlag = 1;
	ND.result = NFX_ERR;
	ND.reason = MOUNTD_ERR_DUPLICATE_EXPORT;
	strcpy(ND.rrStr, "ND");

	NG.validFlag = 1;
	NG.result = NFX_ERR;
	NG.reason = MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND;
	strcpy(NG.rrStr, "NG");

	NS.validFlag = 1;
	NS.result = NFX_ERR;
	NS.reason = MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND;
	strcpy(NS.rrStr, "NS");

	NC.validFlag = 1;
	NC.result = NFX_ERR;
	NC.reason = MOUNTD_ERR_CONFLICT_EXPORT;
	strcpy(NC.rrStr, "NC");

	NM.validFlag = 1;
	NM.result = NFX_ERR;
	NM.reason = MOUNTD_ERR_MULTIPLE_INTERFACE;
	strcpy(NM.rrStr, "NM");


	/* 1*/	mountd_findRules[0][0][0][0][0] = ND;
	/* 2*/	mountd_findRules[0][0][0][0][1] = NC;
	/* 3*/	mountd_findRules[0][0][0][2][0] = NG;
	/* 4*/	mountd_findRules[0][0][0][2][1] = Y;
	/* 5*/	mountd_findRules[0][0][2][0][0] = NG;
	/* 6*/	mountd_findRules[0][0][2][0][1] = NM;
	/* 7*/	mountd_findRules[0][0][2][2][0] = NG;
	/* 8*/	mountd_findRules[0][0][2][2][1] = NM;
	/* 9*/	mountd_findRules[0][1][0][0][0] = NS;
	/*10*/	mountd_findRules[0][1][0][0][1] = Y;
	/*11*/	mountd_findRules[0][1][0][2][0] = Y;
	/*12*/	mountd_findRules[0][1][0][2][1] = Y;
	/*13*/	mountd_findRules[0][1][2][0][0] = NM;
	/*14*/	mountd_findRules[0][1][2][0][1] = NC;
	/*15*/	mountd_findRules[0][1][2][2][0] = NM;
	/*16*/	mountd_findRules[0][1][2][2][1] = NM;
	/*17*/	mountd_findRules[0][1][0][1][0] = ND;
	/*18*/	mountd_findRules[0][1][0][1][1] = NC;
	/*19*/	mountd_findRules[0][1][2][1][0] = NG;
	/*20*/	mountd_findRules[0][1][2][1][1] = NM;
	/*21*/	mountd_findRules[1][0][0][0][0] = NS;
	/*22*/	mountd_findRules[1][0][0][0][1] = NM;
	/*23*/	mountd_findRules[1][0][0][2][0] = NM;
	/*24*/	mountd_findRules[1][0][0][2][1] = NC;
	/*25*/	mountd_findRules[1][0][2][0][0] = NM;
	/*26*/	mountd_findRules[1][0][2][0][1] = NM;
	/*27*/	mountd_findRules[1][0][2][2][0] = NM;
	/*28*/	mountd_findRules[1][0][2][2][1] = NM;
	/*29*/	mountd_findRules[1][0][1][0][0] = ND;
	/*30*/	mountd_findRules[1][0][1][0][1] = NC;
	/*31*/	mountd_findRules[1][0][1][2][0] = NG;
	/*32*/	mountd_findRules[1][0][1][2][1] = Y;
	/*33*/	mountd_findRules[1][1][0][0][0] = NS;
	/*34*/	mountd_findRules[1][1][0][0][1] = NM;
	/*35*/	mountd_findRules[1][1][0][2][0] = NM;
	/*36*/	mountd_findRules[1][1][0][2][1] = NM;
	/*37*/	mountd_findRules[1][1][2][0][0] = NM;
	/*38*/	mountd_findRules[1][1][2][0][1] = NM;
	/*39*/	mountd_findRules[1][1][2][2][0] = NM;
	/*40*/	mountd_findRules[1][1][2][2][1] = NM;
	/*41*/	mountd_findRules[1][1][0][1][0] = NS;
	/*42*/	mountd_findRules[1][1][0][1][1] = NM;
	/*43*/	mountd_findRules[1][1][1][0][0] = NS;
	/*44*/	mountd_findRules[1][1][1][0][1] = Y;
	/*45*/	mountd_findRules[1][1][2][1][0] = NM;
	/*46*/	mountd_findRules[1][1][2][1][1] = NM;
	/*47*/	mountd_findRules[1][1][1][2][0] = Y;
	/*48*/	mountd_findRules[1][1][1][2][1] = Y;
	/*49*/	mountd_findRules[1][1][1][1][0] = ND;
	/*50*/	mountd_findRules[1][1][1][1][1] = NC;

	return;
}

/*-----------------------------------------------------------------
 * Name:      	mountd_displayRulesForExport
 *
 * Description:
 *     	Display the rules.
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/22/2002
 *
 *-----------------------------------------------------------------
 */
void mountd_displayRulesForExport()
{
	int i, j, k, l, m, count;

	printf("Rules for adding a share:\n");
	printf("===========================================================\n");
	printf("     LA    CA    LB    CB    OPT   RESULT   STATS   CHANGE\n");
	printf("    ----  ----  ----  ----  -----  ------   -----   ------\n");

	count = 0;

	for (i = 0; i < MOUNTD_RULEARRAY_S_LA; i++)
		for (j = 0; j < MOUNTD_RULEARRAY_S_CA; j++)
			for (k = 0; k < MOUNTD_RULEARRAY_S_LB; k++)
				for (l = 0; l < MOUNTD_RULEARRAY_S_CB; l++)
					for (m = 0; m < MOUNTD_RULEARRAY_S_OPT; m++)
	{
		if (!mountd_findRules[i][j][k][l][m].validFlag)
			continue;

		printf("%3d. ", count+1);
		printf("%-6s", mountd_ruleStrLocalAddr[i]);
		printf("%-6s", mountd_ruleStrClientAddr[j]);
		printf("%-6s", mountd_ruleStrLocalAddr[k]);
		printf("%-6s", mountd_ruleStrClientAddr[l]);
		printf("%-8s", mountd_ruleStrOption[m]);
		printf("%-4s", mountd_findRules[i][j][k][l][m].rrStr);
		printf("%7d", mountd_findRules[i][j][k][l][m].counter);
		printf("%8d\n", mountd_findRules[i][j][k][l][m].changeCnt);
		mountd_findRules[i][j][k][l][m].changeCnt = 0;
		count++;
	}

	return;
}

/*-----------------------------------------------------------------
 * Name:      	mountd_isReqVolExported 
 *
 * Description:
 *     	Check if input volume (by Volume ID) is currently been 
 *		exported. 
 *
 * Created by:          Joseph Vu
 *
 * Date Created:        03/20/02
 *
 *-----------------------------------------------------------------
 */
int
mountd_isReqVolExported(evm_volId_t searchVolId)
{
    mnt_infoCB_t    *currp;
    int             count;
    int             displayEntries;

    currp = mountd_globalInfop->mnt_infoHeadp;
    count = mountd_globalInfop->numOfEntries;
    displayEntries = 0;
    while ((currp != NULL) && (count > 0 ) )
    {
        if ( (currp->mntInfoEntry.volId.whole_thing != 0) &&
			 (currp->mntInfoEntry.volId.whole_thing == searchVolId.whole_thing) )
        {
			return 1;	
        }
        count--;    
        currp = currp->nextp;
    }

	return 0;
}

/*-----------------------------------------------------------------
 * Name:       	mountd_notifyCcm 
 * 
 * Description:
 *      Notify NCM/CCM if a volume (or directory) has been exported 
 *      or unexported. 
 *      
 * Created by:          Joseph Vu
 * 
 * Date Created:        03/20/02
 * 
 *-----------------------------------------------------------------
 */
void mountd_notifyCcm(
	mnt_infoEntry_t *mnt_infoEntryp, 		//export or unexport entry
	uint32 delflag							//true if delete or unexport 
)
{
	char					*pathNamep;
	char					*exportInfop;
	int						rc;
	nfx_eventMsg_t			export_evt;
	char					*expPathp;
	struct in_addr			in_hostAddr;

	memset (&export_evt, 0, sizeof(export_evt));

	export_evt.msgType = NFX_EVENT_POST;
	export_evt.eventType = NFX_EVENT_NFS;
	gethostname(export_evt.nodeName, sizeof(export_evt.nodeName));

	export_evt.param.exportEvent.volId = mnt_infoEntryp->volId;
	export_evt.param.exportEvent.expAddr = mnt_infoEntryp->localIpAddr;

	pathNamep = mnt_infoEntryp->pathName;
	expPathp = &export_evt.param.exportEvent.expPath[0];
	if (mnt_infoEntryp->hostAddr)
	{
		in_hostAddr.s_addr = mnt_infoEntryp->hostAddr;
		strcpy(expPathp, inet_ntoa(in_hostAddr));
		strcat(expPathp, ":");
	}
	strcat(expPathp, "/");
	strncat(expPathp, pathNamep, mnt_infoEntryp->pathNameLen);

	exportInfop = pathNamep + mnt_infoEntryp->pathNameLen;
	memcpy(export_evt.param.exportEvent.expOptions, exportInfop, 
	       mnt_infoEntryp->exportInfoLen);

	if (delflag) {
		export_evt.param.exportEvent.state = EVENT_STATE_REMOVE;
	}
	else {
		export_evt.param.exportEvent.state = EVENT_STATE_UP;
	}

	MNT_DBG_PRINTF("mountd_notifyCcm: volId: %lld export path: %s expOptions: %s delflag: %d\n",
			export_evt.param.exportEvent.volId.whole_thing, 
	        export_evt.param.exportEvent.expPath, 
			export_evt.param.exportEvent.expOptions, 
	        export_evt.param.exportEvent.state);

	//Post the event
	rc = sendAgileMsg(-1, &export_evt, sizeof(export_evt),
			          "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
			          NULL, NULL, 0);

	if (rc < 0) {
		printf ("mountd_notifyCcm:  MOUNT-ERR: rc: %d\n", rc);
		MOUNTD_ELOG (class_1, err_s, mountd_globalInfop->mountd_AppId, 0, 0, 0,
				(MOUNTD_EXPORT_SEND_CCM_ERR_STR, MOUNTD_STR, rc));
	}

	return;
}

/*-----------------------------------------------------------------
 * Name:       	mountd_notifyCcmAll 
 *  
 * Description: 
 *      Notify NCM/CCM if a volume (or directory) has been exported
 *      or unexported. 
 *      
 * Created by:          Joseph Vu
 *  
 * Date Created:        03/20/02
 *  
 *-----------------------------------------------------------------
 */
void mountd_notifyCcmAll(void)
{
    mnt_infoCB_t    *currp;
    int             count;
    int             displayEntries;

    currp = mountd_globalInfop->mnt_infoHeadp;
    count = mountd_globalInfop->numOfEntries;
    displayEntries = 0;
    while ((currp != NULL) && (count > 0 ) )
    {
        if (currp->mntInfoEntry.volId.whole_thing != 0)
        {
			mountd_notifyCcm(&currp->mntInfoEntry, 0);
        }
        count--;
        currp = currp->nextp;
    }
}



/*-----------------------------------------------------------------
 * Name:       mountd_findInfoEntryById 
 *
 * Description:
 * Retrieve MOUNTD info control block based on input entry filehandle.
 * Retrieve MOUNTD info control block based on input fileHandle,
 * client IP address, and local IP address from mount request.
 *
 *      Given local IP address (L) and client IP address (C) from
 *      mount request, we define 3 levels of matches against existing
 *      entries as follows,
 *
 *        ExistingEntry's    ExistingEntry's    matchLevel
 *          LocalAddr          ClientAddr 
 *        ---------------    ---------------    ----------
 *             all                 all              1
 *             all                 C                2
 *             L                   all              2
 *             L                   C                3
 *      We will return the entry found with highest matching level.

 *
 * Created by:          Joseph Vu
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */

mnt_infoCB_t *mountd_findInfoEntryById(mnt_infoReq_t *mnt_infoReqp)

{
	mnt_infoCB_t    *level1, *level2, *level3;
	mnt_infoCB_t    *currp;
	int32			count;
    uint32          localA, clientA;    //IP addresses for existing entry
    uint32          localB, clientB;    //IP addresses in mountd request


	//MNT_DBG_PRINTF("mountd_findInfoEntryById: enter\n");

    level1 = 0;
    level2 = 0;
    level3 = 0;

    localB = mnt_infoReqp->localIpAddr;
    clientB = mnt_infoReqp->ipAddr;
	
	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;
	
	MNT_DBG_PRINTF("Export Info Entries (%d):\n", count);
	while ((currp != NULL) && (count > 0 ) )
	{

		if ( (currp->mntInfoEntry.volId.whole_thing == 
					mnt_infoReqp->volId.whole_thing ) &&
		  	(currp->mntInfoEntry.mntPnt == mnt_infoReqp->mntPnt) )
		{
            localA = currp->mntInfoEntry.localIpAddr;
            clientA = currp->mntInfoEntry.hostAddr;

            if ((localA == 0) && (clientA == 0))
            {
                level1 = currp;
            }
            else if ((localA == 0) && (clientA == clientB))
            {
                level2 = currp;
            }
            else if ((localA == localB) && (clientA == 0))
            {
                level2 = currp;
            }
            else if ((localA == localB) && (clientA == clientB))
            {
                level3 = currp;
            }

            if (level3)
            {
                MNT_DBG_PRINTF("mountd_findInfoEntryById: level 3\n");
                return (level3);
            }

		}	
			
		count--;
		currp = currp->nextp;
	}
	//MNT_DBG_PRINTF("mountd_findInfoEntry: exit\n");
    if (level2)
    {
        MNT_DBG_PRINTF("mountd_findInfoEntryById: level 2\n");
        return (level2);
    }
    else if (level1)
    {
        MNT_DBG_PRINTF("mountd_findInfoEntryById: level 1\n");
        return (level1);
    }
    else
    {
        MNT_DBG_PRINTF("mountd_findInfoEntryById: Entry is not found\n");
        return (NULL);
    }

}

/*-----------------------------------------------------------------
 * Name:       	mountd_checkFindRulesForExport
 *
 * Description:
 *     	Check whether the two export entries (represented by input)
 *     	are duplicated, conflict, subset, or superset with each other.
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/18/2002
 *
 *-----------------------------------------------------------------
 */
int32 mountd_checkFindRulesForExport(int32 la, int32 ca, 
									int32 lb, int32 cb, int32 opt,
									int32 *reason)
{
	int32						rc;
	mountd_findRuleForExport_t	*entry;

	MNT_DBG_PRINTF("mountd_checkFindRulesForExport: "
					"la, ca, lb, cb, opt: %d %d %d %d %d\n",
					la, ca, lb, cb, opt);

	//Validate input
	if ((la < 0 || la >= MOUNTD_RULEARRAY_S_LA) ||
		(ca < 0 || lb >= MOUNTD_RULEARRAY_S_CA) ||
		(lb < 0 || lb >= MOUNTD_RULEARRAY_S_LB) ||
		(cb < 0 || lb >= MOUNTD_RULEARRAY_S_CB) ||
		(opt < 0 || opt >= MOUNTD_RULEARRAY_S_OPT)) 
	{
		*reason = MOUNTD_ERR_INTERNAL;
		rc = NFX_ERR;
	}

	//Obtain the entry
	entry = &mountd_findRules[la][ca][lb][cb][opt];
	entry->counter++;
	entry->changeCnt++;

	if (!entry->validFlag)
	{
		*reason = MOUNTD_ERR_INTERNAL;
		rc = NFX_ERR;
	}
	else
	{
		*reason = entry->reason;
		rc = entry->result;
	}

	MNT_DBG_PRINTF("mountd_checkFindRulesForExport: "
					"*reason: %d, rc: %d\n", *reason, rc);

	return (rc);
}

/*-----------------------------------------------------------------
 * Name:       	mountd_findInfoEntryForExport
 *
 * Description:
 *     	Retrieve MOUNTD info control block based on input pathname,
 *      client and local IP addresses, and export option mask.
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/18/2002
 *
 *-----------------------------------------------------------------
 */
mnt_infoCB_t *mountd_findInfoEntryForExport(char	*pathNamep,
											uint32	localIpAddr,
											uint32	hostIpAddr,
											uint32	optionMask,
											uint32	*status) //return err status
{
	mnt_infoCB_t    *currp;
	mnt_infoEntry_t	*mntInfoEntryp;
	int32           count;
	int32			pathNameLen;
	uint32			la, ca, lb, cb, opt; 
	int32			rc;

	MNT_DBG_PRINTF("mountd_findInfoEntryForExport: pathNamep: %s\n", pathNamep);
	MNT_DBG_PRINTF("mountd_findInfoEntryForExport: "
					"localIpAddr: 0x%x, hostIpAddr: 0x%x, optionMask: 0x%x\n",
					localIpAddr, hostIpAddr, optionMask);

	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;

	*status = NFX_OK;
	pathNameLen = strlen(pathNamep);

	while ((currp != NULL) && (count > 0 ))
	{
		if ( (currp->mntInfoEntry.pathNameLen == pathNameLen) &&
			 (!(strncmp(currp->mntInfoEntry.pathName,
					   	pathNamep, pathNameLen))) )
		{
			// Found the info entry with same pathname.

			// Now follow the rules described before definition of
			// mnt_infoEntrySearchRuleForExport

			mntInfoEntryp = &currp->mntInfoEntry;

			la = (mntInfoEntryp->localIpAddr) ? 1:0;
			ca = (mntInfoEntryp->hostAddr) ? 1:0;

			lb = (localIpAddr) ? 2:0;
			cb = (hostIpAddr) ? 2:0;

			if (la && lb &&
				(mntInfoEntryp->localIpAddr == localIpAddr))
				lb = 1;

			if (ca && cb &&
				(mntInfoEntryp->hostAddr == hostIpAddr))
				cb = 1;

			opt = (mntInfoEntryp->exportOptionMask != optionMask) ? 1:0;

			rc = mountd_checkFindRulesForExport(la, ca, lb, cb, opt, status);
			if (rc != NFX_OK)
			{
				//Found an entry or had an error
				return (currp);
			}
		}
		count--;
		currp = currp->nextp;
	}
	MNT_DBG_PRINTF("mountd_findInfoEntryForExport: Entry is not found\n");
	
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name:       	mountd_findInfoEntryForUnexport
 *
 * Description:
 *     	Retrieve MOUNTD info control block based on input pathname,
 *      client IP address, and local IP address.
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/18/2002
 *
 *-----------------------------------------------------------------
 */
mnt_infoCB_t *mountd_findInfoEntryForUnexport(char		*pathNamep,
											  uint32	localIpAddr,
											  uint32	hostIpAddr,
											  int32		*status)
{
	mnt_infoCB_t    *currp;
	int32           count;
	int32			pathNameLen;

	MNT_DBG_PRINTF("mountd_findInfoEntryForUnexport: "
					"pathNamep: %s, localIpAddr: 0x%x, hostIpAddr: 0x%x\n", 
					pathNamep, localIpAddr, hostIpAddr);

	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;

	pathNameLen = strlen(pathNamep);

	*status = MOUNTD_ERR_NO_EXPORT_FOUND;

	while ((currp != NULL) && (count > 0 ))
	{
		if ( (currp->mntInfoEntry.pathNameLen == pathNameLen) &&
			 (!(strncmp(currp->mntInfoEntry.pathName,
					   	pathNamep, pathNameLen))) )
		{
			if ((localIpAddr == MOUNTD_ALL_ADDR) &&
				(hostIpAddr == MOUNTD_ALL_ADDR))
			{
				return (currp);
			}
			else if (localIpAddr == MOUNTD_ALL_ADDR)
			{
				if (currp->mntInfoEntry.hostAddr == hostIpAddr)
					return (currp);
				else
				{
					/*Set error status*/
					if (hostIpAddr)
						*status = MOUNTD_ERR_EXPORT_INVALID_CLNT;
					else
						*status = MOUNTD_ERR_EXPORT_INVALID_0_CLNT;
				}
			}
			else if (hostIpAddr == MOUNTD_ALL_ADDR)
			{
				if (currp->mntInfoEntry.localIpAddr == localIpAddr)
					return (currp);
				else
				{
					/*Set error status*/
					if (localIpAddr)
						*status = MOUNTD_ERR_EXPORT_INVALID_LOCIP;
					else
						*status = MOUNTD_ERR_EXPORT_INVALID_0_LOCIP;
				}
			}
			else if ((currp->mntInfoEntry.localIpAddr == localIpAddr) &&
					 (currp->mntInfoEntry.hostAddr == hostIpAddr))
			{
				return (currp);
			}
			else
			{
				/*Set error status*/
				*status = MOUNTD_ERR_EXPORT_INVALID_CLNT_LOCIP;
			}
		}
		count--;
		currp = currp->nextp;
	}
	MNT_DBG_PRINTF("mountd_findInfoEntryForUnexport: Entry is not found\n");
	
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name:       	mountd_findInfoEntryForMountReq
 *
 * Description:
 *     	Retrieve MOUNTD info control block based on input pathname,
 *      client IP address, and local IP address from mount request.
 *
 *      Given local IP address (L) and client IP address (C) from
 *      mount request, we define 3 levels of matches against existing
 *      entries as follows,
 *
 *        ExistingEntry's    ExistingEntry's    matchLevel
 *          LocalAddr          ClientAddr 
 *        ---------------    ---------------    ----------
 *             all                 all              1
 *             all                 C                2
 *             L                   all              2
 *             L                   C                3
 *      We will return the entry found with highest matching level.
 *
 *
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/18/2002
 *
 *-----------------------------------------------------------------
 */
mnt_infoCB_t *mountd_findInfoEntryForMountReq(char		*pathNamep,
											  uint32	localIpAddr,
											  uint32	hostIpAddr)
{
	mnt_infoCB_t    *currp, *level1, *level2, *level3;
	int32           count;
	int32			pathNameLen;
	uint32			localA, clientA;	//IP addresses for existing entry
	uint32			localB, clientB; 	//IP addresses in mountd request

	MNT_DBG_PRINTF("mountd_findInfoEntryForMountReq: "
					"pathNamep: %s, localIpAddr: 0x%x, hostIpAddr: 0x%x\n", 
					pathNamep, localIpAddr, hostIpAddr);

	if ((localIpAddr == 0) || (hostIpAddr == 0))
	{
		MNT_DBG_PRINTF("mountd_findInfoEntryForMountReq: "
						"Invalid local (0x%x) or client (0x%x) IP address.\n",
						localIpAddr, hostIpAddr);
		return (NULL);
	}

	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;

	level1 = 0;
	level2 = 0;
	level3 = 0;

	localB = localIpAddr;
	clientB = hostIpAddr;

	pathNameLen = strlen(pathNamep);

	while ((currp != NULL) && (count > 0 ))
	{

		if ( (currp->mntInfoEntry.pathNameLen == pathNameLen) &&
			 (!(strncmp(currp->mntInfoEntry.pathName,
					   	pathNamep, pathNameLen))) )
		{
			localA = currp->mntInfoEntry.localIpAddr;
			clientA = currp->mntInfoEntry.hostAddr;

			if ((localA == 0) && (clientA == 0))
			{
				level1 = currp;
			}
			else if ((localA == 0) && (clientA == clientB))
			{
				level2 = currp;
			}
			else if ((localA == localB) && (clientA == 0))
			{
				level2 = currp;
			}
			else if ((localA == localB) && (clientA == clientB))
			{
				level3 = currp;
			}

			if (level3)
			{
				MNT_DBG_PRINTF("mountd_findInfoEntryForMountReq: levle 3\n");
				return (level3);
			}
		}
		count--;
		currp = currp->nextp;
	}
	
	if (level2)
	{
		MNT_DBG_PRINTF("mountd_findInfoEntryForMountReq: levle 2\n");
		return (level2);
	}
	else if (level1)
	{
		MNT_DBG_PRINTF("mountd_findInfoEntryForMountReq: levle 1\n");
		return (level1);
	}
	else
	{
		MNT_DBG_PRINTF("mountd_findInfoEntryForMountReq: Entry is not found\n");
		return (NULL);
	}
}

/*-----------------------------------------------------------------
 * Name:  mountd_isFileSysReady     		 
 *
 * Description:
 *	Check if filesystem is ready to be used.
 *	The file system is ready to be used if the state
 *	is ONLINE.	
 *  
 * Created by:          Joseph Vu
 *
 * Date Created:        07/03/02
 *
 *-----------------------------------------------------------------
 */

int32	//TRUE or FALSE
mountd_isFileSysReady(evm_volId_t volId)
{
	int32 			stat;
	uint32			state;	
	fs_MsgViewFsReq fsReq;
	fs_MsgViewFsResp *fsRsp;
	fs_FSInfo 		*fsInfo;
	char			respBuf[sizeof(fs_MsgViewFsResp) + sizeof(fs_FSInfo) ];
	uint32			respBufLen;
	struct timeval waitEfsTo = {60, 0};

	MNT_DBG_PRINTF("mountd_isFileSysReady 1: enter: volId: %llx\n", volId.whole_thing);
	//Init memory for req & response	
	respBufLen = sizeof(fs_MsgViewFsResp) + sizeof(fs_FSInfo) ;
	bzero(respBuf, respBufLen);
	bzero(&fsReq, sizeof(fsReq));
	fsRsp = (fs_MsgViewFsResp *)respBuf;

	//set up req
	fsReq.hdr.requestType = FS_REQ_VIEWFS;
	fsReq.hdr.len = sizeof(fsReq);
	fsReq.hdr.clientPrivate = getpid(); 
	fsReq.type = FS_VIEW_ONE_VOLUME;
	memcpy((void *) &(fsReq.volId), (const void *) &(volId),
			              sizeof(evm_volId_t));

	stat = sendAgileMsg (-1, &fsReq, sizeof(fsReq), 
	                     "efs", NFX_NFP_FP2, 1, MGMTPLANE,
		                     &waitEfsTo, fsRsp, respBufLen);

	if ((stat != respBufLen) || (fsRsp->nbrFs != 1))
	{
		MNT_DBG_PRINTF("mountd_isFileSysReady: Get file system state for volume %llx failed!\n", 
				volId.whole_thing);
		return FALSE;
	}

	fsInfo = &(fsRsp->u.fsInfo[0]);

	state = fsInfo->state;

	if ( (state == FS_VOL_STATE_ONLINE) ||
		(state == FS_VOL_STATE_ONLINE_RDONLY) ||
         (state == FS_VOL_STATE_BACKUP_RESTORE))
	{
		MNT_DBG_PRINTF("mountd_isFileSysReady: volId %llx is ready\n",
						volId.whole_thing);
		return (TRUE);
	}

	MNT_DBG_PRINTF("mountd_isFileSysReady: volId %llx is not ready. State: %d\n", 
				volId.whole_thing, state);

	return (FALSE);
}

/*-----------------------------------------------------------------
 * Name:   mountd_updateFileHandle    		 
 *
 * Description:
 *	Update configure NFS share directory (mount point) with it
 *  corresponding filehandle from the file system.
 *
 * Created by:          Joseph Vu
 *
 * Date Created:        07/03/02
 *
 *-----------------------------------------------------------------
 */

int32
mountd_updateFileHandle(mnt_infoReq_t	*mnt_infoReqp)
{
	int32           count;
	int32			pathNameLen;
	char			*pathNamep;
	mnt_infoCB_t    *currp;

	pathNamep = mnt_infoReqp->pathName;
	pathNameLen = mnt_infoReqp->nameLen;
	MNT_DBG_PRINTF("mountd_updateFileHandle: "
					"pathNamep: %s, len: %d\n", 
					pathNamep, pathNameLen);


	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;

	while ((currp != NULL) && (count > 0 ))
	{

		if ( (currp->mntInfoEntry.pathNameLen == pathNameLen) &&
			 (!(strncmp(currp->mntInfoEntry.pathName,
					   	pathNamep, pathNameLen))) )
		{
			currp->mntInfoEntry.volId = mnt_infoReqp->volId;
			currp->mntInfoEntry.inode = mnt_infoReqp->inode;
			currp->mntInfoEntry.mntPnt = mnt_infoReqp->mntPnt;
			currp->mntInfoEntry.state = MNT_SHARE_EXPORT_ENABLED;
			MNT_DBG_PRINTF("Enable: volId: %llx inode: %llx mntPnt: %llx name: %s\n",
			currp->mntInfoEntry.volId.whole_thing,
  			currp->mntInfoEntry.inode,currp->mntInfoEntry.mntPnt,
				currp->mntInfoEntry.pathName);
		}
		count--;
		currp = currp->nextp;
	}


	return (NFX_OK);	
}

 /*-----------------------------------------------------------------
 * Name:       	mountd_allocInfoCB 
 *
 * Description:
 *		Allocate export entry for input export info record.
 *
 * Created by:          Joseph Vu
 * Modified by:			Xiaobing Zhang
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
mnt_infoCB_t *mountd_allocInfoCB (int32 pathNameLen, int32 exportInfoLen)
{
	mnt_infoCB_t	*mnt_infoCBp;
	uint32			allocSize;

	//MNT_DBG_PRINTF("mountd_allocInfoCB: Enter\n");	
	/*
	 *TODO:  Need account for hostName Length???
	 */
	MNT_DBG_PRINTF("mountd_allocInfoCB: pathNameLen: %d, exportInfoLen: %d\n", 
					pathNameLen, exportInfoLen);
	allocSize = sizeof(mnt_infoCB_t) + pathNameLen + exportInfoLen + 2;

	mnt_infoCBp = (mnt_infoCB_t    *)MOUNTD_ALLOC(allocSize);
	if (mnt_infoCBp == NULL)
		return (NULL);

	memset(mnt_infoCBp, 0, allocSize);

	/*Assign cookie*/
	mountd_globalInfop->cookie++;
	if (mountd_globalInfop->cookie >= MOUNTD_MAX_COOKIE)
		mountd_globalInfop->cookie = 1;
	mnt_infoCBp->cookie = mountd_globalInfop->cookie;

	//MNT_DBG_PRINTF("mountd_allocInfoCB: Exit\n");	

	return (mnt_infoCBp);
}

/*-----------------------------------------------------------------
 * Name:       	mountd_delInfoAllCB  
 *
 * Description:
 *      Delete all export entries from the export list.
 * Created by:          Joseph Vu
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
int
mountd_delInfoAllCB (uint32 notifyCcm)
{
	mnt_infoCB_t    *currp;
	mnt_infoCB_t    *mntInfoCBp;
	int				count;

	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;

	while ((currp != NULL) && (count > 0 ) )
	{
		MOUNTD_ELOG(class_1, info_s, mountd_globalInfop->mountd_AppId,
				0, 0, 0,
				(MOUNTD_EXPORT_DEL_INFO_STR, MOUNTD_STR,
				currp->mntInfoEntry.pathName)); 

		//Free delete entry
		mntInfoCBp = currp;	
		if (mntInfoCBp->mntInfoEntry.volId.whole_thing != 0)
		{
			if (notifyCcm)
				mountd_notifyCcm(&mntInfoCBp->mntInfoEntry, 1);
			MOUNTD_FREE(mntInfoCBp);
		}
		currp = currp->nextp;
		mountd_globalInfop->numOfEntries -= 1;
		count--;
	}

	mountd_globalInfop->numOfEntries = 0;
	mountd_globalInfop->mnt_infoHeadp = NULL;
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mountd_delInfoCBByLocalIpAddr
 *
 * Description:
 *      Delete the export entries given their local IP address.
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/20/2002
 *
 *-----------------------------------------------------------------
 */
int
mountd_delInfoCBByLocalIpAddr (uint32 localIpAddr)
{
	mnt_infoCB_t    *currp;
	mnt_infoCB_t    *mntInfoCBp;
	int				count;

	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;

	while ((currp != NULL) && (count > 0 ) )
	{
		MOUNTD_ELOG(class_1, info_s, mountd_globalInfop->mountd_AppId,
				0, 0, 0,
				(MOUNTD_EXPORT_DEL_INFO_STR, MOUNTD_STR,
				currp->mntInfoEntry.pathName)); 

		//Free delete entry
		mntInfoCBp = currp;	
		if (mntInfoCBp->mntInfoEntry.localIpAddr == localIpAddr)
		{
			if (mntInfoCBp == mountd_globalInfop->mnt_infoHeadp)
			{
				mountd_globalInfop->mnt_infoHeadp = mntInfoCBp->nextp;
			}
			else
			{
				mntInfoCBp->prevp->nextp = mntInfoCBp->nextp;
				if (mntInfoCBp->nextp)
					mntInfoCBp->nextp->prevp = mntInfoCBp->prevp;
			}

			currp = currp->nextp;

			mountd_notifyCcm(&mntInfoCBp->mntInfoEntry, 1);
			MOUNTD_FREE(mntInfoCBp);
			mountd_globalInfop->numOfEntries -= 1;
		}
		else
			currp = currp->nextp;
		count--;
	}

	if (mountd_globalInfop->numOfEntries == 0)
		mountd_globalInfop->mnt_infoHeadp = NULL;
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:       mountd_delInfoCBByPointer 
 *
 * Description:
 *      Delete the export entrie.
 * Created by:          Xiaobing Zhang
 *
 * Date Created:        05/18/2002
 *
 *-----------------------------------------------------------------
 */
void
mountd_delInfoCBByPointer (mnt_infoCB_t *mnt_infoCBp)
{
	MNT_DBG_PRINTF("mountd_delInfoCBByPointer: mnt_infoCBp: %p\n", 
					mnt_infoCBp);

	if (mnt_infoCBp == mountd_globalInfop->mnt_infoHeadp)
	{
		mountd_globalInfop->mnt_infoHeadp = mnt_infoCBp->nextp;
	}
	else
	{
		mnt_infoCBp->prevp->nextp = mnt_infoCBp->nextp;
		if (mnt_infoCBp->nextp)
			mnt_infoCBp->nextp->prevp = mnt_infoCBp->prevp;
	}
	mountd_notifyCcm(&mnt_infoCBp->mntInfoEntry, 1);
	MOUNTD_FREE(mnt_infoCBp);
	mountd_globalInfop->numOfEntries -= 1;

	MNT_DBG_PRINTF("mountd_delInfoCBByPointer: "
					"mountd_globalInfop->numOfEntries: %d\n", 
					mountd_globalInfop->numOfEntries);

	return;
}

/*-----------------------------------------------------------------
 * Name:       	mountd_addInfoCB 
 *
 * Description:
 *		Add export info record to the export list.
 * Created by:          Joseph Vu
 * Modified by:			Xiaobing Zhang
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
int32 mountd_addInfoCB(mnt_infoCB_t *mnt_infoCBp, 
						mnt_infoReq_t *mnt_infoReqp,
						char *pathNamep, int32 pathNameLen,
						char *exportInfop, int32 exportInfoLen,
						uint32 optionMask)
{
	int32			rc;
	mnt_infoCB_t    *currp, *prevp;
	mnt_infoEntry_t	*mnt_infoEntryp;
	char			*newExportInfop;
	char			currPathName[VOL_MAX_NAME];

	MNT_DBG_PRINTF("mountd_addInfoCB: Enter\n");
	if ( (mnt_infoCBp == NULL) || (mnt_infoReqp == NULL) ||
		(pathNamep == NULL) || (exportInfop == NULL) )
		return (NFX_ERR);
	if (mountd_globalInfop->numOfEntries >= MOUNTD_MAX_EXPORTS)
	{

		printf("mountd_addInfoCB: Max Export entries reached (%d)!\n", 
						mountd_globalInfop->numOfEntries);		
		return (NFX_ERR);
	}
	
	MNT_DBG_PRINTF("mountd_addInfoCB: pathNamep: %s len: %d\n", 
					pathNamep, pathNameLen);
	MNT_DBG_PRINTF("mountd_addInfoCB: exportInfop: %s len: %d\n", 
					exportInfop, exportInfoLen);

	/*
	 * Fill in the fields of the export info entry
	 */
	mnt_infoEntryp = &mnt_infoCBp->mntInfoEntry;

	mnt_infoEntryp->volId.whole_thing = mnt_infoReqp->volId.whole_thing;
	mnt_infoEntryp->localIpAddr = mnt_infoReqp->localIpAddr;
	mnt_infoEntryp->exportOptionMask = optionMask;
	mnt_infoEntryp->hostAddr = mnt_infoReqp->ipAddr;

	mnt_infoEntryp->pathNameLen = pathNameLen; 
	memcpy (&mnt_infoEntryp->pathName[0], pathNamep, pathNameLen); 

	mnt_infoEntryp->exportInfoLen = exportInfoLen;
	newExportInfop = &mnt_infoEntryp->pathName[0] + mnt_infoEntryp->pathNameLen;
	memcpy(newExportInfop, exportInfop, exportInfoLen); 

//	mnt_infoEntryp->state = MNT_SHARE_EXPORT_ENABLED;
	//Init to disable state
	mnt_infoEntryp->state = MNT_SHARE_EXPORT_DISABLED;

	MNT_DBG_PRINTF("mountd_addInfoCB: add entry to the list\n");	
	/*
	 *Add new entry to the list
	 */
	if (mountd_globalInfop->mnt_infoHeadp == NULL)
	{
		/*
		 *Add to head of the list
		 */
		mountd_globalInfop->mnt_infoHeadp = mnt_infoCBp;
		mountd_globalInfop->numOfEntries++;
		goto mountd_addInfoCBdone;
	}
	/*
	 *Insert in the list or append to the end of the list
	 */
	currp = mountd_globalInfop->mnt_infoHeadp;	
	while(currp)
	{
		/*Get path name of mnt_infoCB pointed by currp*/
		memcpy(currPathName,
			   &currp->mntInfoEntry.pathName[0],
			   currp->mntInfoEntry.pathNameLen);
		currPathName[currp->mntInfoEntry.pathNameLen] = '\0';

		/*Compare path names*/
		rc = mountd_pathNameCmp(pathNamep, currPathName);
		if (rc == 0)
		{
			/*Two path names are same.*/
			/*Compare client IP addresses*/
			rc = mountd_ipAddrCmp(mnt_infoCBp->mntInfoEntry.hostAddr,
								  currp->mntInfoEntry.hostAddr);
		}

		if (rc < 0)
		{
			/*Insert mnt_infoCB to list in front of currp*/
			mnt_infoCBp->nextp = currp;
			mnt_infoCBp->prevp = currp->prevp;
			if (currp == mountd_globalInfop->mnt_infoHeadp)
				mountd_globalInfop->mnt_infoHeadp = mnt_infoCBp;
			else
				currp->prevp->nextp = mnt_infoCBp;
			currp->prevp = mnt_infoCBp;

			mountd_globalInfop->numOfEntries++;
			goto mountd_addInfoCBdone;
		}
		prevp = currp;
		currp = currp->nextp;
	}
	/*Append mnt_infoCB to the end of the list*/
	prevp->nextp = mnt_infoCBp;
	mnt_infoCBp->prevp = prevp;
	mountd_globalInfop->numOfEntries++;

mountd_addInfoCBdone:
	MOUNTD_ELOG(class_1, info_s, mountd_globalInfop->mountd_AppId,
			0, 0, 0,
			(MOUNTD_EXPORT_ADD_INFO_STR, MOUNTD_STR,
			pathNamep)); 
		
	MNT_DBG_PRINTF("mountd_addInfoCB: Exit\n");
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	 mountd_getId 
 *
 * Description:
 *		Get volume Id from EVM based for the input pathname.
 *
 * Created by:          Joseph Vu
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
evm_volId_t mountd_getId(char *pathNamep, int32 pathNameLen)
{
	char			volume[VOL_MAX_NAME];
	int				volumeLen;			
	evm_MsgGetId_t 	getVolIdmsg;
	char 			respbuff[1024];
	evm_MsgStat_t *respmsg = (evm_MsgStat_t *)respbuff;
	int 			stat;
	struct timeval      timeout = { 10, 0};

	memset(respbuff, 0, sizeof(respbuff));
	memset (&getVolIdmsg, 0, sizeof(getVolIdmsg)); 
	getVolIdmsg.hdr.msgType = EVM_VOL_GET_ID;

	getVolIdmsg.hdr.clientPrivate = 0;
	//
	//Get volume name to retrieve ID from EVM
	memset(&volume[0], 0, sizeof(volume));
	mountd_getVolumeName(&volume[0], &volumeLen,
			pathNamep, pathNameLen);

	//memcpy(getVolIdmsg.name, pathNamep, pathNameLen);
	memcpy(getVolIdmsg.name, &volume[0], volumeLen);

	stat = sendAgileMsg (-1, &getVolIdmsg, sizeof(getVolIdmsg),
						"evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout,
						respbuff, sizeof(respbuff));
	if (stat < 0)
	{
		respmsg->payload.lVolId.whole_thing = -1;
	}
	MOUNTD_ELOG(class_1, debug_s, mountd_globalInfop->mountd_AppId,
			0, 0, 0,
			(MOUNTD_GET_VOLID_INFO_STR, MOUNTD_STR,
			respmsg->payload.lVolId.whole_thing, 
			pathNamep)); 
	/*
	MNT_DBG_PRINTF("mountd_getId: volname: %s volId: %lld\n", 
				&volume[0], respmsg->payload.lVolId.whole_thing);
	*/
	return respmsg->payload.lVolId;
}

/*-----------------------------------------------------------------
 * Name:        	mountd_getvolIds 
 *
 * Description:
 *      Get volume Ids from EVM for all volumes specified in /exports file..
 *
 * Created by:          Joseph Vu
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
int32 mountd_getvolIds(void)
{
	mnt_infoCB_t    *currp;
	int32			count;

	currp = mountd_globalInfop->mnt_infoHeadp;
	count = mountd_globalInfop->numOfEntries;
	
	MNT_DBG_PRINTF("Export Info Entries (%d):\n", count);
	while ((currp != NULL) && (count > 0 ) )
	{
		currp->mntInfoEntry.volId = 
			mountd_getId(currp->mntInfoEntry.pathName,
						currp->mntInfoEntry.pathNameLen);
		if (currp->mntInfoEntry.volId.whole_thing <= 0)
		{
			MOUNTD_ELOG(class_1, err_s, mountd_globalInfop->mountd_AppId, 
				0, 0, 0,
				(MOUNTD_GET_VOLID_ERR_STR, MOUNTD_STR, 
				currp->mntInfoEntry.pathName));
		}
		count--;
		currp = currp->nextp;
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:	mountd_init 
 *
 * Description:
 *	Init. MOUNTD application runs on the SSC that got call at
 *	boot time to read the export information.  
 * Created by:          Joseph Vu
 *
 * Date Created:        07/10/01
 *
 *-----------------------------------------------------------------
 */
int32
mountd_init(void)
{
	int			mnt_fd;


	MNT_DBG_PRINTF("mountd_init Enter\n");		
		
	//Initialize find rule array
	mountd_findRulesInit();

	mountd_globalInfop = &mountd_globalInfo;
	mountd_globalInfop->mnt_infoHeadp = NULL;
	mountd_globalInfop->numOfEntries = 0;
	mountd_globalInfop->cookie = 0;

	mountd_delInfoAllCB (1);

	//mountd_exportAll();
#ifdef	NOTUSED
	/*
	 *Read exports file
	 */
	mountd_readExports();
#endif
	/*
	 *Register to receive messages on the SSC
	 */
	mountd_globalInfop->mnt_fd = regAgileApp(MOUNTD_APP_NAME);
	if (mnt_fd < 0)
	{
		MOUNTD_ELOG(class_1, err_s, 
			mountd_globalInfop->mountd_AppId, 0, 0, 0,
			(MOUNTD_REG_APP_ERR_STR, MOUNTD_STR, MOUNTD_APP_NAME)); 
		exit(1);
	}

	mountd_globalInfop->mountd_AppId = eee_getApplicationId(MOUNTD_APP_NAME); 
	if (mountd_globalInfop->mountd_AppId < 0)
	{
		printf ("mnt_init: Failure to get MOUNTD App Id!\n");
		MOUNTD_ELOG(class_1, err_s, 
			mountd_globalInfop->mountd_AppId, 0, 0, 0,
			(MOUNTD_GET_APPID_ERR_STR, MOUNTD_STR, MOUNTD_APP_NAME)); 
		exit(1);
	}
    //printf("mnt_init: MOUNTD AppId: %d\n", mountd_globalInfop->mountd_AppId);

	//printf("MOUNTD: Agile socket FD: %d\n",  mountd_globalInfop->mnt_fd);
	
#ifdef	MOUNTD_DEBUG
	mountd_cmdDisplayInfoList();	
#endif

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_setupRespMsg 
 *
 * Description:
 *	Set up response message to send back to requester.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int32 mountd_setupRespMsg(
	mnt_infoResp_t  *mnt_infoResp, 
	mnt_infoCB_t    *mnt_InfoCBp 
)
{
	mnt_infoEntry_t	*mtnInfoEntryp;
	char            *srcPathNamep;
	char            *srcExportInfop;
	char            *destPathnamep;
	char            *destExportInfop;
//	char            *hostNamep;
//	int32			hostNameLen;
	int32			respSize;

	mtnInfoEntryp = &mnt_infoResp->mtnInfoEntry[0];
	mtnInfoEntryp->exportOptionMask =
					mnt_InfoCBp->mntInfoEntry.exportOptionMask;
	MNT_DBG_PRINTF("mountd_setupRespMsg: exportmask: %08x\n", 
			mtnInfoEntryp->exportOptionMask);
	mtnInfoEntryp->volId = mnt_InfoCBp->mntInfoEntry.volId;
	mtnInfoEntryp->inode = mnt_InfoCBp->mntInfoEntry.inode;
	mtnInfoEntryp->mntPnt = mnt_InfoCBp->mntInfoEntry.mntPnt;
	mtnInfoEntryp->state = mnt_InfoCBp->mntInfoEntry.state;
	MNT_DBG_PRINTF("mountd_setupRespMsg:State: %s\n\n",
            (mtnInfoEntryp->state == MNT_SHARE_EXPORT_ENABLED)?
            "Enabled" : "Disabled");

	mtnInfoEntryp->localIpAddr = 
					mnt_InfoCBp->mntInfoEntry.localIpAddr;
	mtnInfoEntryp->hostAddr = mnt_InfoCBp->mntInfoEntry.hostAddr;
	mtnInfoEntryp->exportInfoLen =
					mnt_InfoCBp->mntInfoEntry.exportInfoLen;
	mtnInfoEntryp->pathNameLen =
					mnt_InfoCBp->mntInfoEntry.pathNameLen;
	srcPathNamep = &mnt_InfoCBp->mntInfoEntry.pathName[0];
	srcExportInfop = srcPathNamep + mnt_InfoCBp->mntInfoEntry.pathNameLen;
	destPathnamep = &mtnInfoEntryp->pathName[0];
	destExportInfop = destPathnamep + mnt_InfoCBp->mntInfoEntry.pathNameLen;
	/*
	 *copied volume name/id, export info and hostname
	 *into response buffer
	 */
	//hostNameLen = strlen(host->h_name);
	memcpy(destPathnamep, srcPathNamep,
			mtnInfoEntryp->pathNameLen);
	if (mtnInfoEntryp->exportInfoLen)
		memcpy(destExportInfop, srcExportInfop,
				mtnInfoEntryp->exportInfoLen);
	//hostNamep = destExportInfop +
	//				mnt_InfoCBp->mntInfoEntry.exportInfoLen;
	//mtnInfoEntryp->hostNameLen = hostNameLen;
	//memcpy(hostNamep, host->h_name, hostNameLen);
	respSize = sizeof(mnt_infoResp_t) + sizeof(mnt_infoEntry_t) + 
			mtnInfoEntryp->exportInfoLen +
			mtnInfoEntryp->pathNameLen +
			mtnInfoEntryp->hostNameLen + 3;
	
	MNT_DBG_PRINTF("Send Resp-pathName: %s len: %d\n",
			destPathnamep, mtnInfoEntryp->pathNameLen);
	MNT_DBG_PRINTF("Send Resp-exportInfo: %s len: %d\n",
			destExportInfop, mtnInfoEntryp->exportInfoLen);
	MNT_DBG_PRINTF("Send Resp-volId: 0x%llx\n",
			mtnInfoEntryp->volId.whole_thing);
	MNT_DBG_PRINTF("Send Resp-private: 0x%llx\n",
			mnt_infoResp->msgHdr.private);
	
	return (respSize);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_setupShowResp 
 *
 * Description:
 *  Set up response message to send to requester.
 *
 * Created by:      Joseph Vu
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int32 mountd_setupShowResp(mnt_infoReq_t *mnt_infoReqp,
							mnt_infoResp_t *mnt_infoResp,
							uint32 maxSize)
{
	char			*destPathnamep;
	char			*destExportInfop;
	mnt_infoEntry_t *mtnInfoEntryp, *tempp;
    char            *pathNamep;
    char            *exportInfop;
    mnt_infoCB_t    *currp;
    int             count;
    int             numOfEntries;
    char            pathName[VOL_MAX_NAME];
    char            exportInfo[MAX_NAME_LEN];
	uint32			respSize = 0;
	struct in_addr	in_hostAddr;
	evm_volId_t		volId;
	uint32			cookie;
	uint32			entrySize;
	boolean			start;


    currp = mountd_globalInfop->mnt_infoHeadp;
    count = mountd_globalInfop->numOfEntries;
	mtnInfoEntryp = &mnt_infoResp->mtnInfoEntry[0];
    numOfEntries = 0;
	respSize = sizeof(mnt_infoResp_t);

	/*Get volId of pathName in nfs show request*/
	if (mnt_infoReqp->nameLen) 
	{
		volId = mountd_getId(&mnt_infoReqp->pathName[0], mnt_infoReqp->nameLen);
		if (volId.whole_thing <= 0)
		{
			MOUNTD_ELOG(class_1, err_s,
					mountd_globalInfop->mountd_AppId,
					0, 0, 0,
					(MOUNTD_GET_VOLID_ERR_STR, MOUNTD_STR,
					&mnt_infoReqp->pathName[0]));
			mnt_infoResp->msgHdr.status = MOUNTD_ERR_VOL_NOT_FOUND;
			count = 0; //Don't go into the for loop below.
		}
		else
			mnt_infoReqp->volId.whole_thing = volId.whole_thing;
	}

	/*Get cookie from request*/
	cookie = (uint32)mnt_infoReqp->cookie;
	if (cookie == 0)
		start = TRUE;
	else
		start = FALSE;

	/*Initialize cookie in response*/
	mnt_infoResp->cookie = cookie;

	if (cookie == MOUNTD_MAX_COOKIE)
	{
		/*Already sent all the entries in the response messages*/
		count = 0; //Don't go into the for loop below.
	}

    for (; ((currp != NULL) && (count > 0)); count--, currp = currp->nextp)
    {
		if (start == FALSE)
		{
			if (cookie != currp->cookie)
			{
				/*The entry is already examed for last request*/
				continue;
			}
			else
				start = TRUE;
		}

        memset(&pathName[0], 0, sizeof(pathName));
        memset(&exportInfo[0], 0, sizeof(exportInfo));
        pathNamep = currp->mntInfoEntry.pathName;
        exportInfop = pathNamep + currp->mntInfoEntry.pathNameLen;
        memcpy(&pathName[0], pathNamep, currp->mntInfoEntry.pathNameLen);
		MNT_DBG_PRINTF("****mountd_setupShowResp 1: name: %s len: %d\n",
			pathNamep, currp->mntInfoEntry.pathNameLen);
		if (currp->mntInfoEntry.hostAddr)
		{
			in_hostAddr.s_addr = currp->mntInfoEntry.hostAddr;
			strcpy(&exportInfo[0], inet_ntoa(in_hostAddr));
		}
		strcat(&exportInfo[0], exportInfop);
		MNT_DBG_PRINTF("mountd_setupShowResp 2: entry volId: %llx inputVoldId:%llx\n",
			currp->mntInfoEntry.volId.whole_thing,
			mnt_infoReqp->volId.whole_thing);

        if (currp->mntInfoEntry.volId.whole_thing != 0)
        {

			if ((mnt_infoReqp->volId.whole_thing != 0) &&
				(mnt_infoReqp->volId.whole_thing != currp->mntInfoEntry.volId.whole_thing))
			{
				//Only look for entries that match volIn in request
				MNT_DBG_PRINTF("mountd_setupShowResp 3: skip volid does not match!\n");
				continue;
			}

			/*Get mntInfoEntry size*/
			entrySize = sizeof(mnt_infoEntry_t) +
						strlen(exportInfo) +
						currp->mntInfoEntry.pathNameLen +
						currp->mntInfoEntry.hostNameLen + 3;
			/*Check if this entry can be put into the buffer*/
			if ((respSize + entrySize + sizeof(uint64)) >= maxSize)
			{
				/*Exceed the buffer boundary*/
				MNT_DBG_PRINTF("mountd_setupShowResp 4: "
								"Enough entries for this response message!\n");
				break;
			}

            //printf("%s\n", &pathName[0]);
			
            //if (currp->mntInfoEntry.exportInfoLen)
                //printf("\t%s\n", &exportInfo[0]);
	
			mtnInfoEntryp->volId = currp->mntInfoEntry.volId;
			mtnInfoEntryp->inode = currp->mntInfoEntry.inode;
			mtnInfoEntryp->mntPnt = currp->mntInfoEntry.mntPnt;

			mtnInfoEntryp->localIpAddr = currp->mntInfoEntry.localIpAddr;
			mtnInfoEntryp->hostAddr = currp->mntInfoEntry.hostAddr;
			mtnInfoEntryp->exportOptionMask = currp->mntInfoEntry.exportOptionMask;
			mtnInfoEntryp->state = currp->mntInfoEntry.state;
			mtnInfoEntryp->exportInfoLen = strlen(exportInfo);
			mtnInfoEntryp->pathNameLen =
                    currp->mntInfoEntry.pathNameLen;

			destPathnamep = &mtnInfoEntryp->pathName[0];
			destExportInfop = destPathnamep + mtnInfoEntryp->pathNameLen;
			memcpy(destPathnamep, pathNamep,
            		mtnInfoEntryp->pathNameLen);
			if (mtnInfoEntryp->exportInfoLen)
				memcpy(destExportInfop, &exportInfo[0],
						mtnInfoEntryp->exportInfoLen);
			MNT_DBG_PRINTF("mountd_setupShowResp: name: %s len: %d state: %d\n",
				destPathnamep, mtnInfoEntryp->pathNameLen, mtnInfoEntryp->state);
			MNT_DBG_PRINTF("exportInfo: %s len: %d\n", 
				destExportInfop, mtnInfoEntryp->exportInfoLen);

            numOfEntries++;
			tempp = mtnInfoEntryp;
			MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);
			respSize += (uint32)((char *)mtnInfoEntryp - (char *)tempp);

			/*Set private field of resp msg hdr to be the cookie of next entry*/
		 	/*Response to next request may start from this entry*/
			if (currp->nextp != NULL)
				mnt_infoResp->cookie = currp->nextp->cookie;
			else
				/*Reached the end of the list*/
				mnt_infoResp->cookie = MOUNTD_MAX_COOKIE;
        }
    }

	mnt_infoResp->numOfEntries = numOfEntries;
	MNT_DBG_PRINTF("mountd_setupShowResp: respSize: %d numOfEntries: %d\n", respSize, mnt_infoResp->numOfEntries);
	return (respSize);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_processExportOptionString
 *
 * Description:
 *  Process an export option string from export request.
 *
 * Created by:      Xiaobing Zhang
 *
 * Date Created:    05/17/2002
 *
 *-----------------------------------------------------------------
 */
int32
mountd_processExportOptionString(mnt_infoReq_t	*mnt_infoReqp,
								char			*optionp,
								int32			*optionLenp,
								int32			*optionMaskp)
{
	char	*datap;
	char	tmpStr[MNT_MAX_OPTION_LEN + MNT_IP_ADDR_STRING_LEN + 8];
	char	optStr[MNT_MAX_OPTION_LEN];
	char	*addrp;
	int		tmpLen;

	memset(tmpStr, 0, sizeof(tmpStr));

	//Point to option string
	datap = &mnt_infoReqp->pathName[0] + mnt_infoReqp->nameLen;
	if (*datap == '(')
	{
		tmpLen = mnt_infoReqp->optionsLen;
		memcpy(tmpStr, datap, mnt_infoReqp->optionsLen);
	}
	else
	{
		tmpLen = mnt_infoReqp->optionsLen + 2;
		*tmpStr = '(';
		memcpy((tmpStr+1), datap, mnt_infoReqp->optionsLen);
		*(tmpStr + tmpLen - 1) = ')';
	}
	//Now option string has format of 
	//(rw,root_squash) or (rw,root_squash,laddr=a.b.c.d)

	strcpy(optStr, tmpStr);
	addrp = strstr(optStr, ",laddr=");
	if (addrp)
	{
		*addrp = ')';
		addrp++;
		*addrp = '\0';
	}
	//Now optStr contains string in format of "(rw,root_squash)"

	//Obtain export option mask
	*optionMaskp = mountd_processExportOptions(optStr, strlen(optStr));	

	//Validate export option
	if (!(*optionMaskp & MOUNTD_RW_OPT) &&
		!(*optionMaskp & MOUNTD_RO_OPT))
	{
		/*
		printf("MOUNTD-ERR: Invalid export option! (%s)\n", optStr);
		MOUNTD_ELOG (class_1, err_s,
					mountd_globalInfop->mountd_AppId, 0, 0, 0,
					(MOUNTD_EXPORT_OPT_ERR_STR,
					MOUNTD_STR, optStr));
		return (MOUNTD_ERR_INVALID_EXPORT_OPT);
		*/

		//Set to default "rw"
		*optionMaskp |= MOUNTD_RW_OPT;
	}

	//Check for conflicting permission
	if  ( ( (*optionMaskp & MOUNTD_RW_OPT) && 
			(*optionMaskp & MOUNTD_RO_OPT ) ) ||
		  ( (*optionMaskp & MOUNTD_NO_ROOT_SQUASH_OPT) && 
			(*optionMaskp & MOUNTD_ROOT_SQUASH_OPT) ) )
	{
		printf("MOUNTD-ERR: Conflicting in "
				"export option permission! (%s)\n", optStr);
		MOUNTD_ELOG (class_1, err_s,
					mountd_globalInfop->mountd_AppId, 0, 0, 0,
					(MOUNTD_EXPORT_PERM_ERR_STR,
					MOUNTD_STR, optStr));
		return (MOUNTD_ERR_CONFLICT_EXPORT_OPT);
	}

	if (!(*optionMaskp & MOUNTD_NO_ROOT_SQUASH_OPT) &&
		!(*optionMaskp & MOUNTD_ROOT_SQUASH_OPT))
	{
		//Add default option MOUNTD_ROOT_SQUASH_OPT
		*optionMaskp |= MOUNTD_ROOT_SQUASH_OPT;
	}

	//Construct the new option string
	*optionp = '(';
	mountd_buildOptionString(optionp, *optionMaskp);
	addrp = strstr(tmpStr, ",laddr=");
	if (addrp)
	{
		strcat(optionp, addrp);
	}
	else
	{
		strcat(optionp, ")");
	}
	*optionLenp = strlen(optionp);

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_processExportEntry 
 *
 * Description:
 *  Process an export entry.
 *
 * Created by:      Joseph Vu
 * Modified by:		Xiaobing Zhang
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int
mountd_processExportEntry(mnt_infoReq_t   *mnt_infoReqp, 
		int exportUpdate)
{
	mnt_infoCB_t	*mnt_infoCBp;
	char			pathName[VOL_MAX_NAME];
	char			client[MNT_MAX_HOSTNAME_LEN];
	char			option[MNT_MAX_OPTION_LEN + MNT_IP_ADDR_STRING_LEN + 8];
	int				pathNameLen, clientLen, optionLen;
	int32			optionMask;
	uint32			localIpAddr;
	int				status;
	evm_volId_t		volId;
	char			*addrp;

	//Def. 2260
	//Check for max num of shares 
	//has been reached
	if (mountd_globalInfop->numOfEntries >= MOUNTD_MAX_EXPORTS)
		return (MOUNTD_ERR_MAX_NUM_OF_SHARES_REACHED);
	

	MNT_DBG_PRINTF("mountd_processExportEntry: "
					"mnt_infoReqp->pathName: %s len: %d\n",
				&mnt_infoReqp->pathName[0], mnt_infoReqp->nameLen);

	// Input string is in one of these formats:
	//     /vol2rw,root_squash
	//     /vol2rw,root_squash,laddr=a.b.c.d
	//     pacman:/vol2rw,root_squash
	//     pacman:/vol2rw,root_squash,laddr=a.b.c.d
	//     /vol2(rw,root_squash,laddr=a.b.c.d)
	//     e.f.g.h:/vol2(rw,root_squash)
	//     e.f.g.h:/vol2(rw,root_squash,laddr=a.b.c.d)

	//-----------------------
	//Input string processing
	//-----------------------

	//Obtain client name/IP addr string and pathname string from
	//sting in format of client:/pathname
	memset (&pathName[0], 0, sizeof(pathName));
	memset (&client[0], 0, sizeof(client));
	pathNameLen = 0;
	clientLen = 0;
    status = mountd_parseClientNameField(&mnt_infoReqp->pathName[0],
                mnt_infoReqp->nameLen,
                &client[0], &clientLen,
                &pathName[0], &pathNameLen);
    if (status != NFX_OK)
        return (MOUNTD_ERR_INVALID_CLIENT_PATH);

	//Obtain client IP address from client string
	if (clientLen > 0)
	{
		mnt_infoReqp->ipAddr = mountd_getAddrFromName(client);
		if (mnt_infoReqp->ipAddr == NFX_ERR)
		{
			return (MOUNTD_ERR_NO_HOST_FOUND);
		}
	}
	MNT_DBG_PRINTF("mountd_processExportEntry 1: "
					"pathname:%s pathNameLen:%d client: %s clientLen: %d\n",
					&pathName[0], pathNameLen, &client[0], clientLen);

	//Process option string
	memset (&option[0], 0, sizeof(option));
	optionLen = 0;
	localIpAddr = 0;

	//Default option
	optionMask = MOUNTD_RW_OPT | MOUNTD_ROOT_SQUASH_OPT;	

	if (mnt_infoReqp->optionsLen > 0)
	{
		status = mountd_processExportOptionString(mnt_infoReqp,
													option,
													&optionLen,
													&optionMask);
		if (status != NFX_OK)
			return (status);
		//Now option string has format of 
		//(rw,root_squash) or (rw,root_squash,laddr=a.b.c.d)
 
		//Obtain local IP address from option string
		mountd_getLocalAddr(option, &localIpAddr);
		if (localIpAddr)
			mnt_infoReqp->localIpAddr = localIpAddr;

		//Add local IP address to option string if it is 
		//not in the string
		addrp = strstr(option, ",laddr=");
		if (!addrp)
		{
			struct in_addr	in_localAddr;

			in_localAddr.s_addr = mnt_infoReqp->localIpAddr;
			addrp = index(option, ')');
			strcpy(addrp, ",laddr=");
			strcat(option, inet_ntoa(in_localAddr));
			strcat(option, ")");
			optionLen = strlen(option);
		}
	}
	MNT_DBG_PRINTF("mountd_processExportEntry 2: "
					"option:%s optionLen:%d optionMask: 0x%x\n",
					&option[0], optionLen, optionMask);

	MNT_DBG_PRINTF("mountd_processExportEntry 3: "
					"mnt_infoReqp->ipAddr: 0x%x, "
					"mnt_infoReqp->localIpAddr: 0x%x\n",
					mnt_infoReqp->ipAddr, mnt_infoReqp->localIpAddr);

	//Get volume ID from EVM
	
	volId = mountd_getId(&pathName[0], pathNameLen);
	if (volId.whole_thing <= 0)
	{
		//printf("mountd_processExportEntry: Failure to getVolId!\n");
		MOUNTD_ELOG(class_1, err_s,
				mountd_globalInfop->mountd_AppId,
				0, 0, 0,
				(MOUNTD_BAD_VOLID_STR, MOUNTD_STR,
				&mnt_infoReqp->pathName[0]));
		return (MOUNTD_ERR_VOL_NOT_FOUND);
	}
	mnt_infoReqp->volId.whole_thing = volId.whole_thing;

	//---------------------------
	//Add export entry processing
	//---------------------------

	//Check if entry is already exported
	mnt_infoCBp = mountd_findInfoEntryForExport(pathName,
												mnt_infoReqp->localIpAddr,
												mnt_infoReqp->ipAddr,
												optionMask,
												&status);
	if (mnt_infoCBp != NULL)
	{
		///Found entry
		if (mnt_infoReqp->msgHdr.msgType == MNT_EXPORT_UPDATE)
		{
			//Replace current entry
			MOUNTD_ELOG(class_1, notice_s,
				mountd_globalInfop->mountd_AppId, 0, 0, 0,
				(MOUNTD_REPLACE_EXPORT_STR, MOUNTD_STR, &pathName[0]));
			mountd_delInfoCBByPointer (mnt_infoCBp);
		}
		else
			return (status);
	}

	//Allocate an export entry control block
	mnt_infoCBp = mountd_allocInfoCB(pathNameLen, optionLen);
	if (mnt_infoCBp == NULL)
	{
		printf ("MOUNTD-ERR: Couldn't alloc memory!\n");
		return (MOUNTD_ERR_INTERNAL);
	}

	//Add export entry to the export list
	status = mountd_addInfoCB(mnt_infoCBp, mnt_infoReqp,
							  pathName, pathNameLen, option, optionLen,
							  optionMask);

	//Notifies CCM
	mountd_notifyCcm(&mnt_infoCBp->mntInfoEntry, 0);
		
	//Notify NFS on TxRx
	mountd_nfsExportAddReq(&mnt_infoCBp->mntInfoEntry);

	//Check if the file system has been mounted (or ONLINE). 
	//If not, put share in "Disable" state and return error
	if (mountd_isFileSysReady(volId) != TRUE)
	{
		MOUNTD_ELOG(class_1, err_s,
				mountd_globalInfop->mountd_AppId,
				0, 0, 0,
				(MOUNTD_FS_NOTREADY_STR, MOUNTD_STR,
				&mnt_infoReqp->pathName[0]));
		mnt_infoCBp->mntInfoEntry.state = MNT_SHARE_EXPORT_DISABLED;
		return (MOUNTD_ERR_FILESYS_NOT_READY);
	}
	MNT_DBG_PRINTF("mountd_processExportEntry: File System has been mounted. NFS share state: %s\n", 
		(mnt_infoCBp->mntInfoEntry.state == MNT_SHARE_EXPORT_DISABLED)?
	   					"Disabled" : "Enabled");
	//Notify NFS server TxRx to get
	//mountpoint filehandle
	status = mountd_getExportFileHandle (&pathName[0], pathNameLen, volId);
	if (status != NFX_OK)
	{
		MNT_DBG_PRINTF("mountd_processExportEntry: Failure to send update fileHandle message to TXRX!\n");

		mnt_infoCBp->mntInfoEntry.state = MNT_SHARE_EXPORT_DISABLED;
		return (MOUNTD_ERR_INTERNAL);
	}
	MNT_DBG_PRINTF("mountd_processExportEntry: return from mountd_getExportFileHandle\n");
	if (mnt_infoCBp->mntInfoEntry.volId.whole_thing <= 0)
	{
		//Delete the entry from export list
		//mountd_delInfoCBByPointer(mnt_infoCBp);
		mnt_infoCBp->mntInfoEntry.state = MNT_SHARE_EXPORT_DISABLED;
		return (MOUNTD_ERR_PATHNAME_NOT_FOUND);
	}

	MNT_DBG_PRINTF("mountd_processExportEntry exit: pathName:%s len:%d state:%d\n",
				pathName, pathNameLen, mnt_infoCBp->mntInfoEntry.state);
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	mountd_processUnexportEntry 
 *
 * Description:
 *  Process an unexport entry.
 *
 * Created by:      Joseph Vu
 * Modified by:		Xiaobing Zhang
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int
mountd_processUnexportEntry(mnt_infoReq_t   *mnt_infoReqp)
{
	mnt_infoCB_t		*mnt_infoCBp;
	char				pathNameBuf[MAX_NAME_LEN];
	char				clientBuf[MAX_NAME_LEN];
	int					pathNameBufLen;
	int					clientBufLen;
	int					status;
	boolean				searchSingleEntry;
	uint32				count;

	if (mnt_infoReqp->nameLen == 0)
	{
		/*
		if (mnt_infoReqp->localIpAddr == 0)
		{
			//Unexport all entries
			mountd_delInfoAllCB (1);
		}
		else
		{
			//Unexport entries with matching local IP address
			mountd_delInfoCBByLocalIpAddr (mnt_infoReqp->localIpAddr);
		}
		mountd_nfsExportDeleteReq(NULL, mnt_infoReqp->localIpAddr, 0);
		*/

		//Unexport all entries
		mountd_delInfoAllCB (1);
		mountd_nfsExportDeleteReq(NULL, 0, 0);
		return (NFX_OK);
	}

	MNT_DBG_PRINTF("mountd_processUnexportEntry: "
					"mnt_infoReqp->pathName: %s len: %d\n",
				&mnt_infoReqp->pathName[0], mnt_infoReqp->nameLen);

	//-----------------------
	//Input string processing
	//-----------------------

	// Input string is in one of these formats:
	// 		/vol2
	// 		pacman:/vol2
	// 		10.0.0.218:/vol2
	// 		all:/vol2

	memset(pathNameBuf, 0, sizeof(pathNameBuf));
	memset(clientBuf, 0, sizeof(clientBuf));
	pathNameBufLen = 0;
	clientBufLen = 0;
	status = mountd_parseClientNameField(&mnt_infoReqp->pathName[0],
				mnt_infoReqp->nameLen,
				&clientBuf[0], &clientBufLen,
				&pathNameBuf[0], &pathNameBufLen);
	if (status != NFX_OK)
		return (MOUNTD_ERR_INVALID_CLIENT_PATH);

	MNT_DBG_PRINTF("mountd_processUnexportEntry: "
					"pathname:%s nameLen:%d client: %s clientBufLen: %d\n",
				&pathNameBuf[0], pathNameBufLen, &clientBuf[0], clientBufLen);

	//Obtain client IP address from client string
	if (clientBufLen > 0)
	{
		if (strcmp(clientBuf, "all") == 0)
		{
			mnt_infoReqp->ipAddr = MOUNTD_ALL_ADDR;
		}
		else
		{
			mnt_infoReqp->ipAddr = mountd_getAddrFromName(clientBuf);
			if (mnt_infoReqp->ipAddr == NFX_ERR)
			{
				return (MOUNTD_ERR_NO_HOST_FOUND);
			}
		}
	}
	else
		mnt_infoReqp->ipAddr = 0;

	//------------------------------
	//Delete export entry processing
	//------------------------------

	if ((mnt_infoReqp->localIpAddr == MOUNTD_ALL_ADDR) ||
		(mnt_infoReqp->ipAddr == MOUNTD_ALL_ADDR))
	{
		searchSingleEntry = FALSE;
	}
	else
		searchSingleEntry = TRUE;

	count = 0;
	while(1)
	{
		//Check if entry exists
		mnt_infoCBp = mountd_findInfoEntryForUnexport(pathNameBuf,
													  mnt_infoReqp->localIpAddr,
													  mnt_infoReqp->ipAddr,
													  &status);
		if (mnt_infoCBp == NULL)
		{
			if (count == 0)
			{
				MNT_DBG_PRINTF("mountd_processUnexportEntry: "
								"entry does not exist.\n");
				printf("MOUNTD-ERR: unexport entry does not exist!\n");
				return (status);
			}
			else
			{
				//Found all the entries
				break;
			}
		}

		//Notify TxRx that the entry is deleted
		mountd_nfsExportDeleteReq(pathNameBuf,
								  mnt_infoCBp->mntInfoEntry.localIpAddr,
								  mnt_infoCBp->mntInfoEntry.hostAddr);

		//Delete the entry from export list
		mountd_delInfoCBByPointer(mnt_infoCBp);

		count++;

		if (searchSingleEntry == TRUE)
			break;
	}

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:   	main 
 *
 * Description:
 *		MOUNTD main.
 * Created by:      Joseph Vu
 * Modified by:		Xiaobing Zhang
 *
 * Date Created:    08/9/01
 *
 *-----------------------------------------------------------------
 */
int main()
{
	int				byteCount;
	char			rcvMsgBuf[MNT_MSG_BUF_MAX];
	char			respMsgBuf[MNT_MSG_BUF_MAX];
	char			fromApp[256];
	uint32			fromCpu;
	uint32			fromSlot;
	int32			respSize, maxSize;
	mnt_infoResp_t  *mnt_infoResp;
	mnt_infoReq_t 	*mnt_infoReqp;
	mnt_infoEntry_t	*mtnInfoEntryp;


	//MNT_DBG_PRINTF("main: Enter\n");
//#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
	/*
	 *Register MOUNT & NFS port with PORTMAP 
	 */
	mountd_portmap_register();
//#endif
	/*
	 *Int MOUNTD.
	 *Reads /exports file
	 *register MOUNTD app.
	 */
	mountd_init();
	if (mountd_globalInfop->mnt_fd < 0)
	{
		//printf ("Unable to register for: %s\n", MOUNTD_APP_NAME);
		MOUNTD_ELOG(class_1, err_s, mountd_globalInfop->mountd_AppId, 
		0, 0, 0,
		(MOUNTD_REG_APP_ERR_STR, MOUNTD_STR, MOUNTD_APP_NAME));
		exit (1);
	}
	/*
	 * Tell MCP we're ready to go.
	 */
	kill (getppid(), SIGUSR1);

	/*
	 *Infinite loop to wait for new messages bound for this application
	 */
	while (1)
	{
		byteCount = recvAgileMsg (mountd_globalInfop->mnt_fd, 
							rcvMsgBuf, 
							sizeof(rcvMsgBuf), NULL,
							fromApp, &fromCpu, &fromSlot, 0);
		if (byteCount < 0)
		{
			MOUNTD_ELOG(class_1, err_s, mountd_globalInfop->mountd_AppId, 
					0, 0, 0,
					(MOUNTD_RECV_ERR_STR, MOUNTD_STR));
			exit(1);	
		}

		mnt_infoReqp = (mnt_infoReq_t *)rcvMsgBuf;	
		/*
		 *Initialize response buffer
		 */
		memset(respMsgBuf, 0, sizeof(respMsgBuf));
		mnt_infoResp = (mnt_infoResp_t *)respMsgBuf;
		mnt_infoResp->msgHdr.msgType = MNT_INFO_RESP;
		mnt_infoResp->msgHdr.private = mnt_infoReqp->msgHdr.private;
		mnt_infoResp->numOfEntries = 1;
		mnt_infoResp->msgHdr.status  = NFX_OK;
	    respSize = sizeof(mnt_infoResp_t) + sizeof(mnt_infoEntry_t);
		mtnInfoEntryp = &mnt_infoResp->mtnInfoEntry[0];

		/*Response message buffer max size
		 *Currently only used for MNT_EXPORT_SHOW and MNT_EXPORT_SHOWALL req*/
		if (mnt_infoReqp->respBufSize > 0)
			maxSize = MIN(mnt_infoReqp->respBufSize, sizeof(respMsgBuf));
		else
		{
			/* If request doesn't specify a buffer size*/ 
			maxSize = sizeof(respMsgBuf);
		}

		/*
		 *switch on message type
		 */
		switch (mnt_infoReqp->msgHdr.msgType)
		{
			case MNT_INFO_REQ:
			{
				mnt_infoCB_t  	*mnt_InfoCBp;
				evm_volId_t		volId;

				if (mnt_firstReq == 0)
				{
					/*
					 *Get volume Ids from EVM.
					 *Do it once only.
					 */
					mountd_getvolIds();
					mnt_firstReq = 1;
				}	

				MNT_DBG_PRINTF("MOUNTD-RX:client IP:%08x local IP:%08x\n",
					mnt_infoReqp->ipAddr, mnt_infoReqp->localIpAddr);
				mnt_infoReqp->pathName[mnt_infoReqp->nameLen] = '\0';
				MNT_DBG_PRINTF("MOUNTD RX: pathname: %s len: %d\n", 
					mnt_infoReqp->pathName,
					mnt_infoReqp->nameLen);

				MOUNTD_ELOG(class_1, debug_s, 
						mountd_globalInfop->mountd_AppId,
						0, 0, 0,
						(MOUNTD_RCV_GETNAME_INFO_STR, MOUNTD_STR,
						mnt_infoReqp->ipAddr, 
						&mnt_infoReqp->pathName[0])); 

				/*
				 *Get update VolId from EVM
				 */
				volId = 
					mountd_getId(&mnt_infoReqp->pathName[0],
								mnt_infoReqp->nameLen);
				if (volId.whole_thing <= 0)
				{
					printf("MOUNTD-ERR:  Bad VolId for volume %s\n",  
									&mnt_infoReqp->pathName[0]);
					MOUNTD_ELOG(class_1, err_s, 
						mountd_globalInfop->mountd_AppId, 
							0, 0, 0,
							(MOUNTD_BAD_VOLID_STR, MOUNTD_STR,
							&mnt_infoReqp->pathName[0])); 
					mtnInfoEntryp->volId.whole_thing = -1;
					mnt_infoResp->msgHdr.status = MOUNTD_ERR_BAD_VOL_ID;
					break;
				}

				/*
				 *Find Info entry with corresponding volume Name
				 */
				mnt_InfoCBp = 
					mountd_findInfoEntryForMountReq(
									mnt_infoReqp->pathName,
									mnt_infoReqp->localIpAddr,
									mnt_infoReqp->ipAddr);
				if (mnt_InfoCBp == NULL)
				{
					char	pathName[MAX_NAME_LEN];
					
					memset(pathName, 0, sizeof(pathName));
					memcpy(&pathName[0], mnt_infoReqp->pathName, 
								mnt_infoReqp->nameLen);
					MNT_DBG_PRINTF("MOUNTD-ERR: Failure to locate Export entry \n");
					MOUNTD_ELOG(class_1, info_s, 
						mountd_globalInfop->mountd_AppId,
						0, 0, 0,
						(MOUNTD_LOCATE_NAME_ERR_STR, MOUNTD_STR,
						 pathName));

					mtnInfoEntryp->volId.whole_thing = -1;
					mnt_infoResp->msgHdr.status = MOUNTD_ERR_NO_EXPORT_FOUND;
					break;
				}

				/*
				printf("MOUNTD RX: name: %s VolId: %llx\n", 
					mnt_InfoCBp->mntInfoEntry.pathName,
					mnt_InfoCBp->mntInfoEntry.volId.whole_thing);
				*/
				mnt_InfoCBp->mntInfoEntry.volId = volId;
				/*
				 *Send back info back to FP-NIM
				 *to check the export info
				 */
				respSize = mountd_setupRespMsg(mnt_infoResp, mnt_InfoCBp);

				break;
			}
			case MNT_INFO_REQ_BY_ID:
			{
				//struct in_addr  hostAddr;
				//uint32          hostNameLen;
				//struct hostent  *host;
				mnt_infoCB_t    *mnt_InfoCBp;


				if (mnt_firstReq == 0)
				{
					/*
					 *Get volume Ids from EVM.
					 *Do it once only.
					 */
					mountd_getvolIds();
					mnt_firstReq = 1;
				}	
				MNT_DBG_PRINTF("MOUNTD-RX 2: ipAddr:%08x volId: %lld\n",
					mnt_infoReqp->ipAddr, mnt_infoReqp->volId.whole_thing); 
			
				mnt_infoResp->msgHdr.msgType = MNT_INFO_RESP_BY_ID;
				/*
				 *Find Info entry with corresponding volume Name
				 */
				mnt_InfoCBp = mountd_findInfoEntryById(mnt_infoReqp); 
				if (mnt_InfoCBp == NULL)
				{

					MOUNTD_ELOG(class_1, info_s, 
						mountd_globalInfop->mountd_AppId,
						0, 0, 0,
						(MOUNTD_LOCATE_ID_ERR_STR, MOUNTD_STR,
						 mnt_infoReqp->volId.whole_thing,
					   	mnt_infoReqp->mntPnt));

					mtnInfoEntryp->volId.whole_thing = -1;
					mnt_infoResp->msgHdr.status = MOUNTD_ERR_BAD_VOL_ID;
					break;
				}

				respSize = mountd_setupRespMsg(mnt_infoResp, 
										mnt_InfoCBp);

				break;
			}
			case MNT_EXPORT_REQ:
			{
				MOUNTD_ELOG(class_1, debug_s, mountd_globalInfop->mountd_AppId,
						0, 0, 0,
						(MOUNTD_EXPORT_INFO_STR, MOUNTD_STR));
				mnt_infoResp->msgHdr.status = 
					mountd_processExportEntry(mnt_infoReqp, 
												MNT_EXPORT_REQ);
				break;	
			}
			case MNT_UNEXPORT_REQ:
			{
				MOUNTD_ELOG(class_1, debug_s, mountd_globalInfop->mountd_AppId,
						0, 0, 0,
						(MOUNTD_UNEXPORT_INFO_STR, MOUNTD_STR));
				

				mnt_infoResp->msgHdr.status = 
					mountd_processUnexportEntry(mnt_infoReqp);
				break;
			}
			case MNT_EXPORT_UPDATE:
			{
				mnt_infoResp->msgHdr.status = 
					mountd_processExportEntry(mnt_infoReqp, 
											MNT_EXPORT_UPDATE);
				break;
			}
			case MNT_EXPORT_SHOW:
			case MNT_EXPORT_SHOWALL: 
				respSize = mountd_setupShowResp(mnt_infoReqp, mnt_infoResp, maxSize);
				//mountd_cmdDisplayInfoList();
				
				break;	
			case MNT_GET_SHARES:
				mnt_infoResp->msgHdr.msgType = MNT_GET_SHARES_RESP;
				respSize = mountd_setupShowResp(mnt_infoReqp, mnt_infoResp, maxSize);
				break;
			case MNT_IS_VOL_EXPORTED:
			{
				int			exportFlag;
				exportFlag = mountd_isReqVolExported(mnt_infoReqp->volId);
				if (exportFlag)
				{
					mnt_infoResp->msgHdr.status = 1;
				}
				else
					mnt_infoResp->msgHdr.status = 0;
				
				break;
			}				
			case MNT_GET_VOLIDS:
				/*
				 *Get volume Ids from EVM.
				 *Do it once only.
				 */
				mountd_getvolIds();
				break;
			case MNT_PORTMAP_REG_PORTS:
				/*
	 			 *Register MOUNT & NFS port with PORTMAP 
	 			 */
				mountd_portmap_register();
				break;
			case MNT_EXPORT_ENABLED:
                mnt_infoResp->msgHdr.status =
                    mountd_enableExport(mnt_infoReqp);
				break;
			case MNT_EXPORT_ENABLED_ALL:
				mountd_enableAllExports();
				break;
			case MNT_EXPORT_DISABLED:
                mnt_infoResp->msgHdr.status =
					mountd_disableExport(mnt_infoReqp);
				break;
			case MNT_EXPORT_DISABLED_ALL:
				mountd_disableAllExports();
				break;

			case MNT_TRACE_ENABLE:
				mountd_trace_enable = 1;
				printf("mountd trace has been turned on.\n");
				break;
			case MNT_TRACE_DISABLE:
				mountd_trace_enable = 0;
				printf("mountd trace has been turned off.\n");
				break;

			case MNT_EXPRULE_DISPLAY:
				mountd_displayRulesForExport();
				break;

		   case NFX_EVENT_POST:
			   /* nothing to do for event */
			   break;
			
			default:
			{
				nfx_eventMsg_t *evt;

				//Handle events from event manager
				//Currently just handle CPU down event.
				//Volume events is handle by CCM

				evt = (nfx_eventMsg_t *) rcvMsgBuf;

				if ( (evt->msgType == NFX_EVENT_POST) &&
					(evt->eventType == NFX_EVENT_CPU) &&
					 (evt->param.cpuEvent.state == EVENT_STATE_DOWN) )
				{
					//defect 2169: 
					//delete all NFS shares 
					//CCM will callback MOUNTD 
					//upon restore config.

					mountd_delInfoAllCB (0);
					break;
				}
				/*
				 *Error case: 
				 */
				MOUNTD_ELOG(class_1, LOG_ERR, 
						mountd_globalInfop->mountd_AppId, 
						0, 0, 0,
						(MOUNTD_INV_MSG_STR, MOUNTD_STR,
						fromApp, fromCpu, fromSlot)); 

				break;
			}

		}

		/*
		 *Send a response back to the source 
		 */
			
		MNT_DBG_PRINTF("Send resp to App: %s Cpu: %d Slot: %d\n", 
				fromApp,fromCpu, fromSlot);
		/*		
		printf("Send resp to App: %s Cpu: %d Slot: %d\n", 
				fromApp,fromCpu, fromSlot);
		*/
		MOUNTD_ELOG(class_1, debug_s, mountd_globalInfop->mountd_AppId, 
				0, 0, 0,
				(MOUNTD_SNDRSP_MSG_STR, MOUNTD_STR,
				fromApp, fromCpu, fromSlot)); 
		if (sendAgileMsg (-1, respMsgBuf, respSize,
						fromApp, fromCpu, fromSlot, MGMTPLANE,
						NULL, NULL, 0) < 0) 
		{
			MOUNTD_ELOG(class_1, err_s, mountd_globalInfop->mountd_AppId, 
					0, 0, 0,
					(MOUNTD_SENDRSP_ERR_STR, MOUNTD_STR));
			exit(1);
		}


	}	/*while(1)*/
}
