/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-data.c
 *
 * RCS:  $Id: ndmp-data.c,v 1.38 2002/12/04 22:40:36 nitinj Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: Process all NDMP_DATA_xxx commands sent by the
 *              NDMP client.  These commands are ways for the client
 *              to
 *
 * Created by: Anh Vu
 *
 * Date Created:        10/25/01
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <nfx-incl.h>
#include <exports/elog-api.h> 
#include <exports/anpssc-api.h>
#include <exports/evm_com_defs.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"

#define NDMP_MAX_FLIST_MSG_SIZE 8192 // Max 8K packet.

extern ndmpAgileMsgEnt_t ndmpAgileMsgTbl;
int64 ndmp_getVolId(ndmpSess_t *sess, char *volname);

int32 ndmp_commProcEnv(ndmpSess_t *sess, int op_typ, char *bu_type,
                       uint32 env_len, ndmp_pval *env_val, ndmp_Err *error,
                       int *partial_bkup, char *bkupPath, int64 *volId,
                       int32 *tot_env_data_len);
ndmp_NFXReq_t *ndmp_commProcRecovReq(ndmpSess_t *sess, char *bu_type, 
		uint32 env_len, ndmp_pval *env_val, uint32 nlist_len, 
		void *nlist_val, ndmp_Err *error, int32 *msglen);
ndmp_NFXReq_t *ndmp_commProcBkupReq(ndmpSess_t *sess, char *bu_type, 
		uint32 env_len, ndmp_pval *env_val, ndmp_Err *error, int32 *msglen);

int32 fill_env(char *env_data, int32 env_len, ndmp_pval *env_val);
int32 fill_flist(char *flist_data, int32 nlist_len,
                 ndmp_name *nlist_val, char *cpath, int32 eliminateCPath);
int32 fill_flist_v2(char *flist_data, int32 nlist_len,
                 ndmpv2_name *nlist_val_v2, char *cpath, int32 eliminateCPath);
int32 get_tot_fillen(int32 i, char *orig_path, char *dest_dir,
		char *new_name, char *other_name,
		int32 *tot_org_fillen, int32 *tot_dst_fillen);
int32 get_cwd_v2(int32 nlist_len, ndmpv2_name *nlist_val_v2, char *cwd);
void get_common_path_v2(int32 nlist_len, ndmpv2_name *nlist_val_v2, char *path);
int32 get_cwd(int32 nlist_len, ndmp_name *nlist_val, char *cwd);
void get_common_path(int32 nlist_len, ndmp_name *nlist_val, char *path);

/*-----------------------------------------------------------------
 * Name: ndmp_procDataGetStateReq
 *
 * Description:     Process NDMP_DATA_GET_STATE req.  Pass it to NNIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataGetStateReq(ndmpSess_t *sess)
{

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataGetState:\n"));
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_DATA_GET_STATE, 0);
	return NFX_OK;

} /* ndmp_procDataGetState */
/*-----------------------------------------------------------------
 * Name: ndmp_procDataGetEnvReq
 *
 * Description:     Process NDMP_DATA_GET_STATE req.  Pass it to NNIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataGetEnvReq(ndmpSess_t *sess)
{

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataGetEnvReq:\n"));
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_DATA_GET_ENV, 0);
    return NFX_OK;

} /* ndmp_procDataGetEnvReq */

/*-----------------------------------------------------------------
 * Name: ndmp_getVolId
 *
 * Description:   Query evm for volId given volName. 
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int64 
ndmp_getVolId(ndmpSess_t *sess, char *volname)
{     
    char 				respbuff[1024];
    evm_MsgStat_t 		*respmsg = (evm_MsgStat_t *)respbuff;
	struct timeval      to; 
    evm_MsgGetId_t 		getmsg;

    /* Get volid from volname */
 
    to.tv_sec = 5;
    to.tv_usec = 0;
 
    memset (&getmsg, 0, sizeof(getmsg));    // Squeeky clean
    getmsg.hdr.msgType = EVM_VOL_GET_ID;
    getmsg.hdr.clientPrivate = sess->sessId;
    strcpy (getmsg.name, volname);
 
	sess->ndmp_ssc_chld_msgStats[EVM_VOL_GET_ID_REQ_SNT]++;
    if (sendAgileMsg (-1, &getmsg, sizeof(getmsg),
                         "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                         &to,
                         respbuff, sizeof(respbuff)) <= 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_getVolId:  Agile message sent to EVM_GET_VOLID failed\n"));
		return 0;
    }
	sess->ndmp_ssc_chld_msgStats[EVM_VOL_GET_ID_RSP_RCVD]++;
    return (respmsg->payload.lVolId.whole_thing);                    
}

/*-----------------------------------------------------------------
 * Name: 		  ndmp_commProcEnv
 *
 * Description:   Common function to process ENV variables.  Used
 *				  by both BKUP & RESTORE operations.
 *
 * Created by:    Anh Vu
 *
 * Date Created:  7/10/02
 *
 *-----------------------------------------------------------------
 */
int32 
ndmp_commProcEnv(ndmpSess_t *sess, int32 op_typ, char *bu_type, uint32 env_len, 
				ndmp_pval *env_val, ndmp_Err *error, int *partial_bkup, 
				char *bkupPath, int64 *volId, int32 *tot_env_data_len)
{     
	char		buf[200], *fs_name, *fs_ptr, *path_ptr, 
				fs_volname[MAX_VOL_NAME];
	Data_SSC_t  *data = &sess->Serv.Data;
	int32		i, name_len = 0, value_len = 0, nlen, vlen, 
				found_fs_name = FALSE, fs_name_len;
	time_t		t;
	struct 		tm *tm;


	/* First capture the time this ops starts for display later */
	time(&t);
	tm = localtime((const time_t *)&t);
	bzero(sess->start_time, sizeof(sess->start_time));
	sprintf(sess->start_time, "%02d:%02d:%02d %02d-%02d-%02d",
					tm->tm_hour, tm->tm_min, tm->tm_sec, tm->tm_mon+1, 
					tm->tm_mday, tm->tm_year+1900-2000);


	/* Assuming FILESYSTEM env will specify something like:
	 * /fs_volId/. or /fs_volId/anhv/dirone.  So we parse
     * the fs_volId to get the dumpdate for this FS.
	 */  
    /* Check for valid backup type */
    if (strcmp(bu_type, NDMP_EFX_BKUP_TYP_NAME))
    {
    	NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcEnv: invalid bkup type. %s\n", bu_type));
        *error = NDMP_ILLEGAL_ARGS_ERR;
		sprintf(buf, "Invalid backup type %s, expecting %s\n", 
				bu_type, NDMP_EFX_BKUP_TYP_NAME);
    	ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);

		return NFX_ERR;
    }

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcEnv:  Env Total = %d\n", env_len));
	for (i = 0; i< env_len; i++)
	{
		printf("[%d].name = %s\n", i, env_val[i].name);
		printf("[%d].value = %s\n", i, env_val[i].value);

	    if ( !found_fs_name && !strcmp("FILESYSTEM", env_val[i].name))
        {
			NDMP_SSC_ALLOC(fs_name, char, strlen(env_val[i].value)+1);
    		if (fs_name == NULL)
			{
				NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcEnv: Error mallocing fs_name.\n"));
   				*error = NDMP_UNDEFINED_ERR;
				return NFX_ERR;
			}
            strcpy(fs_name, env_val[i].value);
			fs_ptr = fs_name +1;				// scan past 1st '/' of '/volid/engineering/anh'

			// scan past '/volname', should pts to 2nd '/'
			path_ptr = strchr(fs_ptr, '/');	
        	fs_name_len = (path_ptr != NULL)? path_ptr - fs_ptr : strlen(fs_ptr);
			if (path_ptr && *(path_ptr+1) != '\0')
			{
				// it's a partial backup and not a full backup
				*partial_bkup = TRUE;
                memcpy(bkupPath, path_ptr, strlen(path_ptr));
                bkupPath[strlen(path_ptr)] = '\0';
			}
            else
                bkupPath[0] = '\0';
    		strncpy(fs_volname, fs_ptr, fs_name_len);   // get volId into fs_volid
    		fs_volname[fs_name_len] = '\0';

            found_fs_name = TRUE;
			NDMP_SSC_FREE(fs_name);
        }
		nlen = strlen(env_val[i].name);
		NDMP_INT_ALIGN(nlen);
		NDMP_DBG_PRINTF(D_INFO, ("PADDED [%d].name = %d\n", i, nlen));
		name_len += nlen;

		vlen = strlen(env_val[i].value);
		NDMP_INT_ALIGN(vlen);
		NDMP_DBG_PRINTF(D_INFO, ("PADDED [%d].value = %d\n", i, vlen));
		value_len += vlen;
	}
	/* Save this so we can append it to file history files sent to client later on */
	if (found_fs_name)
	{
		/* Get volid from volname */
		*volId = ndmp_getVolId(sess, fs_volname);
		if (*volId)
		{
			strcpy(data->fs_volname, fs_volname);
			data->fs_volId = *volId;
		}
	}
	if (op_typ == BKUP)
	{
		if ( !found_fs_name || !(*volId) )
		{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcEnv: invalid ORIG voldId. %llx\n", *volId));
        	*error = NDMP_ILLEGAL_ARGS_ERR;
			return NFX_ERR;
		}
	} 
    /* for recover, *volId may be gone */
	*tot_env_data_len = name_len + value_len + (env_len * 2 * (sizeof(int32)));

	return NFX_OK;

} /* ndmp_commProcEnv */

/*-----------------------------------------------------------------
 * Name: ndmp_commProcBkupReq
 *
 * Description:     Common function to process DATA_START_BKUP
 *					request.  Used by v2 and v3 DATA_START_BKUP
 *					processing routines.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
ndmp_NFXReq_t *
ndmp_commProcBkupReq(ndmpSess_t *sess, char *bu_type, uint32 env_len, 
					ndmp_pval *env_val, ndmp_Err *error, int32 *msglen)
{
    ndmp_NFXReq_t   	*ndmp_req = NULL;   /* relay client cmd to Mover on NNIM */
	ndmp_NFXBkupReq_t	*NFXbkreq;
	dmpdate_t			*ddrec = NULL;
	char8				*level_str = NULL, buf[200], *env_data;
	int32				tot_env_data_len = 0, i, level = 0, ret,  
						partial_bkup = FALSE;
	int64				volId;
	Data_SSC_t			*data = &sess->Serv.Data;
    char                bkupPath[1024];
//#ifdef FS_HIST_WAIT
    int                 appId;
//#endif


    NDMP_DBG_PRINTF(D_INFO, ("ndmp_commnt                 appId;
							//#endifProcBkupReq:\n"));
	
	*error = NDMP_NO_ERR;

	ret = ndmp_commProcEnv (sess,  BKUP, bu_type, env_len, env_val, error, 
					&partial_bkup, &bkupPath[0], &volId, &tot_env_data_len);
	if (ret == NFX_ERR)
	{
		return ndmp_req;
	}

    sess->partial_bkup = partial_bkup;

    NDMP_SSC_ALLOC(ddrec, dmpdate_t, sizeof(dmpdate_t));
	memset(ddrec, 0, sizeof(dmpdate_t));
	/* Get dumpdates */
	if ((level_str = ndmp_getEnvVar("LEVEL", env_val, env_len)) != NULL)
    {
		level = atoi(level_str);	
		if (level > 0)
		{
			/* Incremental bkup can only be done for a full backup, not for a partial backup,
			 * i.e. backup w/ a path such as '/vol1/engineering/anh' instead of '/vol1/'.
			 */
			if (partial_bkup)
			{
    			NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcBkupReq: Incremental bkup on a partial backup is not allowed, only on a full backup.\n"));
        		*error = NDMP_ILLEGAL_ARGS_ERR;
				sprintf(buf, "Incremental bkup on a partial backup is not allowed, only on a full backup.\n");
		        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
                NDMP_SSC_FREE(ddrec);
				return ndmp_req;
			}
			if (ndmp_getDumpDate(&ddrec, level, volId, data->fs_volname,
                                 sess->host_ipaddr.sin_addr.s_addr, buf) == NFX_ERR)	
			{	
				/* Can't read the file for some reason */
    			NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcBkupReq: Can't get dumpdate from file.\n"));
        		*error = NDMP_UNDEFINED_ERR;
				sprintf(buf, "Server internal error obtaining dumpdates.\n");
		        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
                NDMP_SSC_FREE(ddrec);
				return ndmp_req;
			}
            else if (!strlen(ddrec->snapshot))
                ndmp_sndLogMsg( sess, NDMP_LOG_WARNING, buf, 0);
		}
	} /* if there's no LEVEL env. var., treat it as level 0  & dumpdate is 0*/

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcBkupReq: LEVEL = %d\n", level));

	/* Forward NDMP client request to NNIM */
	/* Msg = NFXReq + dmpdate_t + NFXBkupReq + env_data */
	*msglen = sizeof(ndmp_NFXReq_t) + 
				(sizeof(dmpdate_t) - 1) +
				(sizeof(ndmp_NFXBkupReq_t) + tot_env_data_len -1);
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcBkupReq: msglen %d, tot_env_data_len %d\n", 
				*msglen, tot_env_data_len));

    NDMP_SSC_ALLOC(ndmp_req, ndmp_NFXReq_t, *msglen);
    if (ndmp_req == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcBkupReq: Error mallocing ndmp_req.\n"));
   		*error = NDMP_UNDEFINED_ERR;
		sprintf(buf, "Server internal malloc error.\n");
        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
        NDMP_SSC_FREE(ddrec);
		return ndmp_req;
	}

    memset((void*)ndmp_req, 0, *msglen);
    ndmp_req->hdr.mid        = NDMP_MID_DATA_START_BACKUP;
    ndmp_req->ndmp_mid       = sess->msgRcvd.hdr.msgId;
	memcpy(&ndmp_req->cl_hdr, &sess->msgRcvd.hdr, sizeof(ndmp_Hdr));
    ndmp_req->hdr.sessId     = sess->sessId;
#ifdef NFX_SMP
	ndmp_req->hdr.seqId     = sess->xmt2fp_seqId;
	sess->xmt2fp_seqId ++;
	ndmp_req->hdr.idx        = sess->sess_idx;
#endif

    ndmp_req->userInfo       = 0;

	// Fill in dumpdate first
    memcpy(ndmp_req->content, ddrec, sizeof(dmpdate_t));

	NFXbkreq = (ndmp_NFXBkupReq_t *) (ndmp_req->content + sizeof(dmpdate_t));
	NFXbkreq->fs_volId = volId;
	NFXbkreq->num_env = env_len;
	NFXbkreq->tot_env_data_len = tot_env_data_len;

	env_data = NFXbkreq->env_data;

	// Save on SSC side too
	data->env_len = env_len;					/* num of orig envs sent to us from client */
	data->srvr_env_len = env_len + 1;			/* +1 to include our own ENV var like the BKUPSIZ env */

    NDMP_SSC_ALLOC(data->env_lst, ndmp_pval, (data->srvr_env_len * sizeof(ndmp_pval)));
	if (data->env_lst == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcBkupReq: Error mallocing data->env_lst.\n"));
   		*error = NDMP_UNDEFINED_ERR;
		sprintf(buf, "Server internal malloc error.\n");
		        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
		NDMP_SSC_FREE(ndmp_req);
        NDMP_SSC_FREE(ddrec);
		return ndmp_req;
	}
	
	/* Copy to msg sent to DATA server on NNIM */
	fill_env(env_data, env_len, env_val);

	/* Copy to env data local on SSC, used for GET_ENV later on. */
	for (i = 0; i< env_len; i++)
    {   
        NDMP_SSC_ALLOC(data->env_lst[i].name,char,strlen(env_val[i].name)+1);
		NFX_ASSERT(data->env_lst[i].name != NULL);
		strcpy(data->env_lst[i].name, env_val[i].name);

        NDMP_SSC_ALLOC(data->env_lst[i].value,char,strlen(env_val[i].value)+1);
		NFX_ASSERT(data->env_lst[i].value != NULL);
		strcpy(data->env_lst[i].value, env_val[i].value);
    }

	/* Fill in our extra env var for BKUP_SIZ */
    NDMP_SSC_ALLOC(data->env_lst[i].name,char, strlen(BKUP_SIZ_ENV_STR)+1);
	NFX_ASSERT(data->env_lst[i].name != NULL);
	strcpy(data->env_lst[i].name, BKUP_SIZ_ENV_STR);

	sprintf(buf, "%d", 0);

    NDMP_SSC_ALLOC(data->env_lst[i].value,char, strlen(buf)+1);
	NFX_ASSERT(data->env_lst[i].value != NULL);
	strcpy(data->env_lst[i].value, buf);

	ndmp_req->len1 = sizeof(dmpdate_t);

	memcpy(&ndmp_req->cl_hdr, &sess->msgRcvd.hdr, sizeof(ndmp_Hdr));
    NDMP_SSC_FREE(ddrec);

//#ifdef FS_HIST_WAIT
	/* Connect to FP so we can send window size updates
	* to FP whenever FP sends a lot of File Histories and like
	* to get our receive queue availability.
	*/
	if ( (appId = eee_getApplicationId("efs")) == EEE_ERR_INVALID_APP_NAME)
	{
        printf("ndmp_CreateSess: eee_getApplicationId return invalid appId for 'efs'");
   		*error = NDMP_UNDEFINED_ERR;
        return ndmp_req;
	}	
	if (eee_connect(sess->Agilesock, NFX_SLOTTYPE_NFP, NFX_NFP_FP2, appId) == -1)	
						/* my socket, dest slot, dest cpu, dest app id */
	{
		printf("ndmp_CreateSess: eee_connect failed, error = %d (%s)\n", 
						errno,
						(errno==EBADF)?"EBADF": (errno==EFAULT)?"EFAULT":
						(errno==ENOTSOCK)?"ENOTSOCK": (errno==EISCONN)?"EISCONN":
						(errno==ECONNREFUSED)?"ECONNREFUSED": (errno==ETIMEDOUT)?"ETIMEDOUT":
						(errno==ENETUNREACH)?"ENETUNREACH":(errno==EADDRINUSE)?"EADDRINUSE":
						(errno==EINPROGRESS)?"EALREADY":(errno==EAGAIN)?"EAGAIN":
						(errno==EAFNOSUPPORT)?"EAFNOSUPPORT":(errno==EACCES)?"EACCES":
						(errno==EPERM)?"EPERM":"unknown");	
   		*error = NDMP_UNDEFINED_ERR;
        return ndmp_req;
	}
//#endif /* FS_HIST_WAIT */
	
	return (ndmp_req);

} /* ndmp_commProcBkupReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procDataStartBkupReq
 *
 * Description:     Process V3 NDMP_DATA_START_BKUP req.  Pass it to NNIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataStartBkupReq(ndmpSess_t *sess)
{
	ndmp_data_start_bkup_req	*cl_req = (ndmp_data_start_bkup_req *) sess->msgRcvd.content;
	ndmp_data_start_bkup_rsp	cl_rsp;
    ndmp_NFXReq_t   			*ndmp_req;   /* relay client cmd to Mover on NNIM */
	ndmp_Err					error;
	int32						msglen;
#if 0
	int32						i;
	char						*ptr;
#endif						

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataStartBkupReq:\n"));
	
	/* Formulate the BkupReq for NNIM side */
	ndmp_req = ndmp_commProcBkupReq(sess, cl_req->bu_type, 
								cl_req->env.env_len, cl_req->env.env_val, 
								&error, &msglen);

	/* Don't forward to FP if there's error */
	if (error == NDMP_NO_ERR)
	{
#if 0
    	ptr = (char *)ndmp_req;
    	for (i=0;i < msglen; i++)
    	{
        	if ((i % 16) == 0)
            	NDMP_DBG_PRINTF(D_DBG, ("\n[%04x]: ", i));
        	NDMP_DBG_PRINTF(D_DBG, ("%02x ", (uchar8)ptr[i]));
    	}
    	NDMP_DBG_PRINTF(D_DBG, ("\n"));
#endif
		/* Forward request to NNIM */
    	if (sendAgileMsg (sess->Agilesock, ndmp_req, msglen,
                              "ndmp",       /* NDMP on the N-NIM */
                               NFX_NFP_FP2, 1, MGMTPLANE, /* cpu, slot, plane */
                               NULL, NULL, 0) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataStartBkupReq:  Agile message send to NNIM to START_BKUP failed\n"));
        	error = NDMP_UNDEFINED_ERR;
    	}
		sess->ndmp_ssc_chld_msgStats[FP_DATA_START_BKUP_REQ_SNT]++;
	}
	NDMP_SSC_FREE(ndmp_req);

	/* Send response to client right away if there's any errors.
	 * Otherwise FP will send response to client later.
	 */
	if (error != NDMP_NO_ERR)
	{
		cl_rsp.error = error;
    	return (ndmp_rspClient(sess, &cl_rsp));
	}

	return NFX_OK;

} /* ndmp_procDataStartBkupReq */

/*-----------------------------------------------------------------
 * Name: ndmpv2_procDataStartBkupReq
 *
 * Description:		Process V2 NDMP_DATA_START_BKUP req.  Pass it to NNIM.
 *					The difference between v2 & v3 start_bkup is
 *					that in v2, the mover address is in this request
 *					for Remote bkup vs. in the DATA_CONNECT request.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmpv2_procDataStartBkupReq(ndmpSess_t *sess)
{
	ndmpv2_data_start_bkup_req	*cl_req = (ndmpv2_data_start_bkup_req *) sess->msgRcvd.content;
	ndmp_data_start_bkup_rsp	cl_rsp;
	int32               		addr_type = cl_req->addr.addr_type;
    ndmp_NFXReq_t   			*ndmp_req;   /* relay client cmd to Mover on NNIM */
	ndmp_NFXBkupReq_t			*NFXbkreq;
	ndmp_Err					error;
	int32						msglen;
						

    NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procDataStartBkupReq:\n"));
	
	/* First we look at the address of the mover to connect to */

    /* Handle some of the error handling here if we can */
    if ((addr_type != NDMP_ADDR_LOCAL) && (addr_type != NDMP_ADDR_TCP))
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procDataStartBkupReq: Invalid addr_type"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
						("ndmpv2_procDataStartBkupReq: Invalid addr_type"));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;

		return (ndmp_rspClient(sess, &cl_rsp));
    }

	/* Formulate the BkupReq for NNIM side */
	ndmp_req = ndmp_commProcBkupReq(sess, cl_req->bu_type, 
						cl_req->env.env_len, cl_req->env.env_val, 
						&error, &msglen);

	/* Forward to FP only if there's no error */
	if (error == NDMP_NO_ERR)
	{

		/* Set the data connection address type & info */
		NFXbkreq = (ndmp_NFXBkupReq_t *) 
				(ndmp_req->content + sizeof(dmpdate_t));
    	NFXbkreq->addr_typ = addr_type;

		/* If remote bkup, xfer the Mover address so Data server 
		 * can connect to 
		 */
		if (addr_type == NDMP_ADDR_TCP)
		{
    		NFXbkreq->ip_addr = cl_req->addr.ndmp_addr_u.tcp_addr.ip_addr;
    		NFXbkreq->port = cl_req->addr.ndmp_addr_u.tcp_addr.port;
		}

		/* Forward request to NNIM */
    	if (sendAgileMsg (sess->Agilesock, ndmp_req, msglen,
                              "ndmp",       /* NDMP on the N-NIM */
                               NFX_NFP_FP2, 1, MGMTPLANE, /* cpu, slot, plane */
                               NULL, NULL, 0) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procDataStartBkupReq:  Agile message send to NNIM to START_BKUP failed\n"));
        	error = NDMP_UNDEFINED_ERR;
    	}
		sess->ndmp_ssc_chld_msgStats[FP_DATA_START_BKUP_REQ_SNT]++;
	}

	NDMP_SSC_FREE(ndmp_req);

	/* Send response to client right away if there's any errors.
	 * Otherwise FP will send response to client later.
	 */
	if (error != NDMP_NO_ERR)
	{
		cl_rsp.error = error;
    	return (ndmp_rspClient(sess, &cl_rsp));
	}
	return NFX_OK;

} /* ndmpv2_procDataStartBkupReq */


/*-----------------------------------------------------------------
 * Name: 			fill_env 
 *
 * Description:     Fill in env vars for BKUP/RECOVER msg sent to FP.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	4/23/02
 *-----------------------------------------------------------------
 */
int32
fill_env(char *env_data, int32 env_len, ndmp_pval *env_val)
{
	int32 	i, name_len, value_len;
 
	NDMP_DBG_PRINTF(D_INFO, ("fill_env:  Begin\n"));
	for (i = 0; i< env_len; i++)
    {   
		name_len = strlen(env_val[i].name);
		memcpy(env_data, &name_len, 4);		// fill in length field
		env_data += 4;						// skip name_len field
		memcpy(env_data, env_val[i].name, name_len);
		NDMP_INT_ALIGN(name_len);
		env_data += name_len;				// position ptr to fill in value

		value_len = strlen(env_val[i].value);
		memcpy(env_data, &value_len, 4);	// fill in length field
		env_data += 4;						// skip value_len field
		memcpy(env_data, env_val[i].value, value_len);
		NDMP_INT_ALIGN(value_len);
		env_data += value_len;				// position ptr to fill in value
    }

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: 			fill_flist 
 *
 * Description:     Fill in file list for RECOVER msg sent to FP.
 *					For v3 and up.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	4/23/02
 *-----------------------------------------------------------------
 */
int32
fill_flist(char *flist_data, int32 nlist_len, ndmp_name *nlist_val,
           char *cpath, int32 eliminateCPath)
{
	int32 ofl, dfl, i = 0, j = 0;
    // int32 n = 0, ll = 0, reloc = 0, dstLen = 0;
    char  *d, *ostart;

	NDMP_DBG_PRINTF(D_INFO, ("fill_flist:  Begin\n"));
	NDMP_DBG_PRINTF(D_INFO, ("nlist_len = %d\n", nlist_len));
    NDMP_DBG_PRINTF(D_INFO, ("cpath = %s\n", cpath));
#if 0
    NDMP_DBG_PRINTF(D_INFO, ("cwd = %s\n", cwd));
    if (cpath)
    {
        n = strlen(cpath);
        for (i=0; i<n; i++)
            if (cpath[n-i-1] != '/')
                break;
        n -= i;
    }
    if (cwd)
    {
        /* if the CWD is more than the '/', we have a relocation */
        for (i=0; (!reloc) && i<strlen(cwd); i++)
            if (*(cwd+i) != '/') reloc = 1;
    }
    cwdlen = strlen(cwd);
#endif
    for (i = 0; i < nlist_len; i++)
    {
       	ofl = strlen(nlist_val[i].orig_path);
        ostart = &nlist_val[i].orig_path[0];
        /* remove the trailling '/' from the original path */
        for (j=0; j<ofl && (*(ostart+(ofl-j-1)) == '/'); j++)
            *(ostart+(ofl-j-1)) = '\0';
        ofl -= j;
        /* Since we expect a complete path, if there is not leading '/' in the
           origin, take it as root '/' */
        if (*ostart != '/')
             ofl++;
        dfl = ofl;
		NDMP_DBG_PRINTF(D_INFO, ("[%d].orig_fillen = %d %s\n", i, ofl, ostart));
       	memcpy(flist_data, &ofl, 4);	// fill in length field
       	flist_data += 4;      			// skip org_fillen field

        if (*ostart != '/')
        {
            /* We need the first '/' in the flist string */
            memcpy(flist_data, "/", 1);
       	    if ((ofl-1) > 0)
                memcpy(flist_data+1, ostart, ofl-1);
        }
        else
       	    memcpy(flist_data, ostart, ofl);

		NDMP_INT_ALIGN(ofl);
		NDMP_DBG_PRINTF(D_INFO, ("PADDED [%d].orig_fillen = %d %s\n", i,
                                 ofl, flist_data));
       	flist_data += ofl;  // position ptr to fill in dest file name

        d = ostart;
        if (eliminateCPath)
        {
            if (cpath)
            {
                d += strlen(cpath);
                dfl -= strlen(cpath);
            }
        }

        if (*d != '/')
            dfl++;

        memcpy(flist_data, &dfl, 4);
        flist_data += 4;

        if (*d != '/')
        {
            /* We need the first '/' in the flist string */
            memcpy(flist_data, "/", 1);
            if ((dfl-1) > 0)
                memcpy(flist_data+1, d, dfl-1);
        }
        else
            memcpy(flist_data, d, dfl);

		NDMP_INT_ALIGN(dfl);
		NDMP_DBG_PRINTF(D_INFO, ("PADDED [%d].dest_fillen = %d %s\n", i,
                                 dfl, flist_data));
       	flist_data += dfl;
#if 0
        dstlength = strlen(nlist_val[i].dest_dir);
        NDMP_DBG_PRINTF(D_INFO, ("[%d].dst = %d %s\n", i, dstlength,
                                 nlist_val[i].dest_dir));
        /* Remove the volname */
        c = strchr(&nlist_val[i].dest_dir[1], '/');
        volnamelen = c - &nlist_val[i].dest_dir[0];
        o = d = NULL;
        if (reloc)
        {
            /* Extract CWD from destination and adjust the destination for
               relocation */
            NDMP_DBG_PRINTF(D_INFO, ("Relocation\n"));
            d = c + cwdlen; // point beyond the CWD
            if (cpath)
            {
                /* This will be set only for Legato */
                o = &nlist_val[i].orig_path[n];
            }
            dst_fillen = ll = 0;
            if (dstlength > (volnamelen + cwdlen))
            {
                /* remove all, but one of the leading '/' */
                dst_fillen = strlen(d);
                for (; dst_fillen > 1 && (*(d+1) == '/'); dst_fillen--)
                    d++;
                if (dst_fillen <= 0)
                {
                    dst_fillen = 0;
                    d = NULL;
                }
            }
            else
                d = NULL;
            /* Compute the length of dest string */
            if (o)
                ll = strlen(o);
            dstLen = ll + dst_fillen;
            memcpy(flist_data, &dstLen, 4); // fill in length field
            flist_data += 4;
            /* Copy the destination string */
            if (d)
                memcpy(flist_data, d, dst_fillen);
            /* Copy the origin string */
            if (o)
                memcpy(flist_data+dst_fillen, o, ll);
            NDMP_DBG_PRINTF(D_INFO, ("[%d].dst_fillen = %d %s\n", i, dstLen, flist_data));
            NDMP_INT_ALIGN(dstLen);
            flist_data += dstLen;
        }
        else
        {
            /* We are not relocating, so just copy the cpath at the end */
            NDMP_DBG_PRINTF(D_INFO, ("not relocating\n"));
            d = c;
            dst_fillen = 0;
            if (d)
            {
                /* Remove the trailling '/' from the destinations */
                dst_fillen = strlen(d);
                for (;dst_fillen && (*(d+(dst_fillen-1)) == '/'); dst_fillen--);
            }
            if (cpath)
            {
                /* This will be set only for Legato */
                /* If we need to add the file part from origin, skip the '/' */
                for (j=0; j<strlen(nlist_val[i].orig_path); j++)
                    if (nlist_val[i].orig_path[j] != '/')
                    {
                        o = &nlist_val[i].orig_path[j];
                        break;
                    }
            }
            if (o)
                ll = strlen(o) + 1; // For a leading '/'
            dstLen = ll + dst_fillen;
            memcpy(flist_data, &dstLen, 4); // fill in length field
            flist_data += 4;
            /* Copy into flist, dest + '/' + orig */
            if (d)
                memcpy(flist_data, d, dst_fillen);
            if (o)
            {
                memcpy(flist_data+dst_fillen, "/", 1);
                memcpy(flist_data+dst_fillen+1, o, ll-1);
            }
            NDMP_DBG_PRINTF(D_INFO, ("[%d].dst_fillen = %d %s\n", i, dstLen, flist_data));
            if (dstLen)
                NDMP_INT_ALIGN(dstLen);
            flist_data += dstLen;
        }
#endif
	}
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: 			fill_flist_v2
 *
 * Description:     Fill in file list for RECOVER msg sent to FP.
 *                  For v2 only.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	4/23/02
 *-----------------------------------------------------------------
 */
int32
fill_flist_v2(char *flist_data, int32 nlist_len, ndmpv2_name *nlist_val_v2,
              char *cpath, int32 eliminateCPath)
{
	int32 ofl, dfl, i = 0, j = 0;
    char *d, *ostart;

	NDMP_DBG_PRINTF(D_INFO, ("fill_flist_v2:  Begin\n"));
#if 0
    if (cpath)
    {
        n = strlen(cpath);
        for (i=0; i<n; i++)
            if (cpath[n-i-1] != '/')
                break;
        n -= i;
    }
    if (cwd)
    {
        for (i=0; (!reloc) && i<strlen(cwd); i++)
            if (*(cwd+i) != '/') reloc = 1;
    }
    cwdLen = strlen(cwd);
#endif
    for (i = 0; i < nlist_len; i++)
    {
       	ofl = strlen(nlist_val_v2[i].orig_path);
        ostart = &nlist_val_v2[i].orig_path[0];
        for (j=0; j<ofl && (*(ostart+(ofl-j-1)) == '/'); j++)
            *(ostart+(ofl-j-1)) = '\0';
        ofl -= j;
        if (*ostart != '/')
             ofl++;
        dfl = ofl;
		NDMP_DBG_PRINTF(D_INFO, ("[%d].orig_fillen = %d %s\n", i, ofl, ostart));
       	memcpy(flist_data, &ofl, 4);	// fill in length field
       	flist_data += 4;      			// skip org_fillen field

        if (*ostart != '/')
        {
            memcpy(flist_data, "/", 1);
            if ((ofl-1) > 0)
       	    memcpy(flist_data+1, ostart, ofl-1);
        }
        else
       	    memcpy(flist_data, ostart, ofl);
		NDMP_INT_ALIGN(ofl);
		NDMP_DBG_PRINTF(D_INFO, ("PADDED [%d].orig_fillen = %d %s\n", i,
                                 ofl, flist_data));
       	flist_data += ofl;   // position ptr to fill in dest file name

        NDMP_DBG_PRINTF(D_INFO, ("[%d].dst = %s\n", i, nlist_val_v2[i].dest_dir));

        d = ostart;
        if (eliminateCPath)
        {
            if (cpath)
            {
                d += strlen(cpath);
                dfl -= strlen(cpath);
            }
        }

        if (*d != '/')
            dfl++;

        memcpy(flist_data, &dfl, 4);
        flist_data += 4;

        if (*d != '/')
        {
            /* We need the first '/' in the flist string */
            memcpy(flist_data, "/", 1);
            if ((dfl-1) > 0)
                memcpy(flist_data+1, d, dfl-1);
        }
        else
            memcpy(flist_data, d, dfl);

        NDMP_INT_ALIGN(dfl);
        NDMP_DBG_PRINTF(D_INFO, ("PADDED [%d].dest_fillen = %d %s\n", i,
                                 dfl, flist_data));
        flist_data += dfl;
#if 0
        dstlength = strlen(nlist_val_v2[i].dest_dir);
        c = strchr(&nlist_val_v2[i].dest_dir[1], '/');
        volnamelen = c - &nlist_val_v2[i].dest_dir[0];
        o = d = NULL;
        if (reloc)
        {
            /* Extract CWD from destination and adjust the destination for
               relocation */
            NDMP_DBG_PRINTF(D_INFO, ("Relocation\n"));
            d = c + cwdLen; // point beyond the CWD
            if (cpath)
            {
                /* This will be set only for Legato */
                o = &nlist_val_v2[i].orig_path[n];
            }
            dst_fillen = ll = 0;
            if (dstlength > (volnamelen + cwdLen))
            {
                dst_fillen = strlen(d);
                for (; dst_fillen > 1 && (*(d+1) == '/'); dst_fillen--)
                    d++;
                if (dst_fillen <= 0)
                {
                    dst_fillen = 0;
                    d = NULL;
                }
            }
            else
                d = NULL;
            if (o)
                ll = strlen(o);
            dstLen = ll + dst_fillen;
            memcpy(flist_data, &dstLen, 4); // fill in length field
            flist_data += 4;
            if (d)
                memcpy(flist_data, d, dst_fillen);
            if (o)
                memcpy(flist_data+dst_fillen, o, ll);
            NDMP_DBG_PRINTF(D_INFO, ("[%d].dst_fillen = %d %s\n", i, dstLen, flist_data));
            NDMP_INT_ALIGN(dstLen);
            flist_data += dstLen;
        }
        else
        {
            /* We are not relocating, so just copy the cpath at the end */
            NDMP_DBG_PRINTF(D_INFO, ("not relocating\n"));
            d = c;
            dst_fillen = 0;
            if (d)
            {
                dst_fillen = strlen(d);
                for (;dst_fillen && (*(d+(dst_fillen-1)) == '/'); dst_fillen--);
            }
            if (cpath)
            {
                /* This will be set only for Legato */
                for (j=0; j<strlen(nlist_val_v2[i].orig_path); j++)
                    if (nlist_val_v2[i].orig_path[j] != '/')
                    {
                        o = &nlist_val_v2[i].orig_path[j];
                        break;
                    }
            }
            if (o)
                ll = strlen(o) + 1;
            dstLen = ll + dst_fillen;
            memcpy(flist_data, &dstLen, 4); // fill in length field
            flist_data += 4;
            if (d)
                memcpy(flist_data, d, dst_fillen);
            if (o)
            {
                memcpy(flist_data+dst_fillen, "/", 1);
                memcpy(flist_data+dst_fillen+1, o, ll-1);
            }
            NDMP_DBG_PRINTF(D_INFO, ("[%d].dst_fillen = %d %s\n", i, dstLen, flist_data));
            if (dstLen)
                NDMP_INT_ALIGN(dstLen);
            flist_data += dstLen;
        }
#endif
	}
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:            get_cwd_v2
 *
 * Description:     Get the current working directory from the file list
 *                  For v2 only.
 *
 * Created by:      Nitin Jain
 *
 * Date Created:    10/31/02
 *-----------------------------------------------------------------
 */

int32
get_cwd_v2(int32 nlist_len, ndmpv2_name *nlist_val_v2, char *cwd)
{
    char *c, cpath[1024], *d;
    int32 len, pathFound = 0, i;

	NDMP_DBG_PRINTF(D_INFO, ("get_cwd_v2: Begin\n"));

    /* get the directory path from the orig. */
    /* Eg. orig = /A/B/C/f1, len = strlen(orig) - strlen('/f1') */
    c = strrchr(nlist_val_v2[0].orig_path, '/');
    if (c && c != nlist_val_v2[0].orig_path)
        len = c - nlist_val_v2[0].orig_path;
    else
        len = strlen(nlist_val_v2[0].orig_path);

    /* Remove leading '/' */
    d = nlist_val_v2[0].orig_path;
    for (; len && (*d == '/'); len--) d++;

    if (len)
    {
        strncpy(cpath, d, len);
        cpath[len] = '\0';
        NDMP_DBG_PRINTF(D_INFO, ("cpath : %s\n", cpath));
    }

    c = NULL;
    if (len)
        c = strstr(&nlist_val_v2[0].dest_dir[1], cpath);
    if (c)
    {
        /* This is typical case for veritas */
        /* Orig = /a/b/c; dest = /v1/a/b/c */
        d = strchr(&nlist_val_v2[0].dest_dir[1], '/');
        if (d && d != c)
        {
            strncpy(cwd, d, (c - d));
            pathFound = 1;
            len = strlen(cwd);
            for(i=1; i<len; i++)
            {
                if (*(cwd+(len-i)) == '/') *(cwd+(len-i)) = '\0';
                else break;
            }
        }
        else
            strcpy(cwd, "/");
    }
    else
    {
        /* This is typical case for Legato */
        /* Orig = /a/b/c; dest = /v1 */
        d = strchr(&nlist_val_v2[0].dest_dir[1], '/');
        if (d)
            strncpy(cwd, d, strlen(d));
        else
            strcpy(cwd, "/");
    }
    NDMP_DBG_PRINTF(D_INFO, ("CWD = %s\n", cwd));

	return pathFound;
}

/*-----------------------------------------------------------------
 * Name:            get_common_path_v2
 *
 * Description:     Between all the file list orig.'s, get the list
 *                  most common path. We may have remove this while
 *                  adding the rest of orig path to the destination.
 *                  Need to do this for Legato. It does not fill the
 *                  complete dest. path for the file list.
 *                  For v2 only.
 *
 * Created by:      Nitin Jain
 *
 * Date Created:    10/31/02
 *-----------------------------------------------------------------
 */

void
get_common_path_v2(int32 nlist_len, ndmpv2_name *nlist_val_v2, char *path)
{
	int32 j, i, n, m, len;
    char *c, cpath[1024];

	NDMP_DBG_PRINTF(D_INFO, ("get_common_path_v2: Begin\n"));

    bzero(cpath, sizeof(char)*1024);
    for (i = 0; i< nlist_len; i++)
    {
		NDMP_DBG_PRINTF(D_INFO, ("[%d] file list %s\n", i, nlist_val_v2[i].orig_path));
        if (i == 0)
        {
            c = strrchr(nlist_val_v2[i].orig_path, '/');
            if (c && c != nlist_val_v2[i].orig_path)
                len = c - nlist_val_v2[i].orig_path;
            else
                len = strlen(nlist_val_v2[i].orig_path);
            strncpy(cpath, nlist_val_v2[0].orig_path, len);
        }
        else
        {
            if (strcmp(cpath, nlist_val_v2[i].orig_path) != 0)
            {
                n = strlen(nlist_val_v2[i].orig_path);
                m = strlen(cpath);
                for (j = 0; j < (n>m)?m:n; j++)
                {
                    if (cpath[j] != nlist_val_v2[i].orig_path[j])
                    {
                        cpath[j] = '\0';
                        break;
                    }
                }
            }
        }
    }
    if (!nlist_len)
        strcpy(cpath, "/");
    if (cpath[0] != '/')
    {
        strcpy(path, "/");
        strcat(path, cpath);
    }
    else
        strcpy(path, cpath);
    NDMP_DBG_PRINTF(D_INFO, ("common path : %s\n", path));

	return;
}

/*-----------------------------------------------------------------
 * Name:            get_cwd
 *
 * Description:     Get the current working directory from the file list
 *                  For v3 only.
 *
 * Created by:      Nitin Jain
 *
 * Date Created:    10/31/02
 *-----------------------------------------------------------------
 */

int32
get_cwd(int32 nlist_len, ndmp_name *nlist_val, char *cwd)
{
    char *c, cpath[1024], *d;
    int32 len, pathFound = 0, i;

	NDMP_DBG_PRINTF(D_INFO, ("get_cwd: Begin\n"));

#if 0
    c = strrchr(nlist_val[0].orig_path, '/');
    if (c && c != nlist_val[0].orig_path)
        len = c - nlist_val[0].orig_path;
    else
#endif
        len = strlen(nlist_val[0].orig_path);

    d = nlist_val[0].orig_path;
    for (; len && (*d == '/'); len--) d++;

    if (len)
    {
        strncpy(cpath, d, len);
        cpath[len] = '\0';
        NDMP_DBG_PRINTF(D_INFO, ("cpath : %s\n", cpath));
    }
    
    c = NULL;
    if (len && strlen(nlist_val[0].dest_dir))
        c = strstr(&nlist_val[0].dest_dir[1], cpath);
    if (c)
    {
        d = strchr(&nlist_val[0].dest_dir[1], '/');
        if (d && d != c)
        {
            strncpy(cwd, d, (c - d));
            pathFound = 1;
            len = strlen(cwd);
            for(i=1; i<len; i++)
            {
                if (*(cwd+(len-i)) == '/') *(cwd+(len-i)) = '\0';
                else break;
            }
        }
        // else
        //     strcpy(cwd, "/");
    }
    else if (strlen(nlist_val[0].dest_dir))
    {
        d = strchr(&nlist_val[0].dest_dir[1], '/');
        if (d)
            strncpy(cwd, d, strlen(d));
        // else
        //     strcpy(cwd, "/");
    }
    NDMP_DBG_PRINTF(D_INFO, ("CWD = %s\n", cwd));

	return pathFound;
}

/*-----------------------------------------------------------------
 * Name:            get_common_path
 *
 * Description:     Between all the file list orig.'s, get the list
 *                  most common path. We may have remove this while
 *                  adding the rest of orig path to the destination.
 *                  Need to do this for Legato. It does not fill the
 *                  complete dest. path for the file list.
 *                  For v3 only.
 *
 * Created by:      Nitin Jain
 *
 * Date Created:    10/31/02
 *-----------------------------------------------------------------
 */

void
get_common_path(int32 nlist_len, ndmp_name *nlist_val, char *path)
{
	int32 j, i, n, m, len;
    char *c, cpath[1024];

	NDMP_DBG_PRINTF(D_INFO, ("get_common_path: Begin\n"));

    bzero(cpath, sizeof(char)*1024);
    for (i = 0; i< nlist_len; i++)
    {
		NDMP_DBG_PRINTF(D_INFO, ("[%d] file list %s\n", i, nlist_val[i].orig_path));
        if (i == 0)
        {
            c = strrchr(nlist_val[i].orig_path, '/');
            if (c && c != nlist_val[i].orig_path)
                len = c - nlist_val[i].orig_path;
            else
                len = strlen(nlist_val[i].orig_path);
            strncpy(cpath, nlist_val[0].orig_path, len);
        }
        else
        {
            if (strcmp(cpath, nlist_val[i].orig_path) != 0)
            {
                n = strlen(nlist_val[i].orig_path);
                m = strlen(cpath);
                for (j = 0; j < (n>m)?m:n; j++)
                {
                    if (cpath[j] != nlist_val[i].orig_path[j])
                    {
                        cpath[j] = '\0';
                        break;
                    }
                }
            }
        }
    }
    if (!nlist_len)
        strcpy(cpath, "/");
    if (cpath[0] != '/')
    {
        strcpy(path, "/");
        strcat(path, cpath);
    }
    else
        strcpy(path, cpath);
    NDMP_DBG_PRINTF(D_INFO, ("common path : %s\n", path));

	return;
}


/*-----------------------------------------------------------------
 * Name: 			get_tot_fillen
 *
 * Description:     
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	7/03/02
 *
 *-----------------------------------------------------------------
 */
int32 
get_tot_fillen(int32 i, char *orig_path, char *dest_dir,
				char *new_name, char *other_name,
				int32 *tot_org_fillen, int32 *tot_dst_fillen)
{
	int32	olen, dlen;

	NDMP_DBG_PRINTF(D_INFO, ("get_tot_fillen:  Begin\n"));

	NDMP_DBG_PRINTF(D_INFO, ("[%d].orig = %s\n", i, orig_path));
	NDMP_DBG_PRINTF(D_INFO, ("[%d].dest = %s\n", i, dest_dir));
	NDMP_DBG_PRINTF(D_INFO, ("[%d].new_name = %s\n", i, new_name));
	NDMP_DBG_PRINTF(D_INFO, ("[%d].other_name = %s\n", i, other_name));

	olen = strlen(orig_path);
	NDMP_INT_ALIGN(olen);
	*tot_org_fillen += olen;

	dlen = strlen(dest_dir);
	NDMP_INT_ALIGN(dlen);
	*tot_dst_fillen += (dlen + olen);

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: 			ndmp_commProcRecovReq
 *
 * Description:     Process NDMP_DATA_START_RECOVER req.  Pass it to NNIM.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	4/08/02
 *
 *-----------------------------------------------------------------
 */
ndmp_NFXReq_t *
ndmp_commProcRecovReq(ndmpSess_t *sess, char *bu_type, uint32 env_len, 
						ndmp_pval *env_val, uint32 nlist_len, void *nlist_v,
						ndmp_Err *error, int32 *msglen)
{

    ndmp_NFXReq_t   	*ndmp_req = NULL;   /* relay client cmd to Mover on NNIM */
	ndmp_NFXRecovReq_t	*NFXRecovReq = NULL;
	char8				*flist_data, found_dst_fs_name = FALSE,	*dar = NULL,
						dst_path[1025], *path_ptr, fs_dstvolname[MAX_VOL_NAME],	
						*fs_ptr, buf[200], cwd[1024], cpath[1024];
	int32				tot_env_data_len = 0, i, tot_flist_len, 
						fs_name_len, org_fillen, partialBkup = FALSE,
                        dst_fillen, ret, legatoVeritas, cwdLen;
	int64				volId = 0, dst_volId;
	Data_SSC_t			*data = &sess->Serv.Data;
	ndmpv2_name			*nlist_val_v2 = (ndmpv2_name *) nlist_v;
	ndmp_name			*nlist_val = (ndmp_name *) nlist_v;
    char                *d, bkupPath[1024], *fl;

    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_commProcRecovReq:\n"));

	*error = NDMP_NO_ERR;
	ret = ndmp_commProcEnv (sess,  RESTORE, bu_type, env_len, env_val, error, 
					&partialBkup, &bkupPath[0], &volId, &tot_env_data_len);
	if (ret == NFX_ERR)
		return NULL;

	NDMP_DBG_PRINTF(D_INFO, ("nlist: Total = %d\n", nlist_len));

	/* We don't support Direct Access option, send err back to NDMP client */

	if ((dar = ndmp_getEnvVar("DIRECT", env_val, env_len)) != NULL)
	{
		if ((*dar == 'y') || (*dar == 'Y') || (*dar == 'T') || (*dar == 't'))
       	{
       		NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcRecovReq: No DAR supported\n"));
       		*error = NDMP_NOT_SUPPORTED_ERR;
			sprintf(buf, "No DAR supported.\n");
	        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
			return NULL;
       	}
	}

	/* Find total len of all file list (in bytes)  */
	tot_flist_len = 0;
	org_fillen = dst_fillen = 0;

	if (sess->protoVer == NDMP_VER_2)
	{
		for (i = 0; i< nlist_len; i++)
		{
			get_tot_fillen(i, nlist_val_v2[i].orig_path, nlist_val_v2[i].dest_dir, 
								NULL, NULL,
								&org_fillen, &dst_fillen);
		}
	} else {
		for (i = 0; i< nlist_len; i++)
		{
			get_tot_fillen(i, nlist_val[i].orig_path, nlist_val[i].dest_dir, 
							nlist_val[i].new_name, nlist_val[i].other_name,
								&org_fillen, &dst_fillen);
		}
	}
	tot_flist_len = 2*org_fillen + dst_fillen + (nlist_len*2*(sizeof(int32)));
    NDMP_SSC_ALLOC(fl, char, tot_flist_len);
    if (!fl)
    {
       	NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcRecovReq: No mem.\n"));
       	*error = NDMP_NO_MEM_ERR;
		sprintf(buf, "No memory.\n");
	    ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
		return NULL;
    }

    bzero(cwd, sizeof(char)*1024);
	if (sess->protoVer == NDMP_VER_2)
        legatoVeritas = get_cwd_v2(nlist_len, nlist_val_v2, &cwd[0]);
    else
        legatoVeritas = get_cwd(nlist_len, nlist_val, &cwd[0]);
    cwdLen = strlen(cwd);

    if (partialBkup == TRUE)
    {
        for (i=0; i<cwdLen; i++)
            if (cwd[i] != '/')
                break;
        if (i == cwdLen)
        {
            /* Copy the CWD from bkupPath */
            cwdLen = strlen(bkupPath);
            printf("New CWD : %s\n", bkupPath);
        }
        else
            strcpy(bkupPath, cwd);
    }

	/* Forward NDMP client request to NNIM */
	*msglen = sizeof(ndmp_NFXReq_t) - 1 +
              (sizeof(ndmp_NFXRecovReq_t) + (cwdLen+4) + tot_env_data_len);
#if 0
	*msglen = sizeof(ndmp_NFXReq_t) - 1 +
              (sizeof(ndmp_NFXRecovReq_t) + (cwdLen+4) + tot_env_data_len +
               tot_flist_len - 1);
#endif

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcRecovReq: msglen fwd to FP %d\n", *msglen));

    NDMP_SSC_ALLOC(ndmp_req, ndmp_NFXReq_t, *msglen);
	if (ndmp_req == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcRecovReq: Err mallocing ndmp_req.\n"));
   		*error = NDMP_UNDEFINED_ERR;
		sprintf(buf, "Server internal Malloc error.\n");
        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
		return ndmp_req;
	}

    memset((void*)ndmp_req, 0, *msglen);
    ndmp_req->hdr.mid        = NDMP_MID_DATA_START_RECOVER;
    ndmp_req->ndmp_mid       = sess->msgRcvd.hdr.msgId;
	memcpy(&ndmp_req->cl_hdr, &sess->msgRcvd.hdr, sizeof(ndmp_Hdr));
    ndmp_req->hdr.sessId     = sess->sessId;
#ifdef NFX_SMP
	ndmp_req->hdr.seqId     = sess->xmt2fp_seqId;
	sess->xmt2fp_seqId ++;
	ndmp_req->hdr.idx        = sess->sess_idx;
#endif
    ndmp_req->userInfo       = 0;

	NFXRecovReq = (ndmp_NFXRecovReq_t *) ndmp_req->content;
	NFXRecovReq->num_env = env_len;
	NFXRecovReq->tot_env_data_len = tot_env_data_len;
	// NFXRecovReq->num_files = nlist_len;
	// NFXRecovReq->tot_file_name_len = tot_flist_len;
	NFXRecovReq->org_fsvolId = volId;

    d = NFXRecovReq->data;

	memcpy(d, &cwdLen, 4);
    d += 4;
    if (partialBkup == TRUE)
        memcpy(d, bkupPath, cwdLen);
    else
        memcpy(d, cwd, cwdLen);
	NDMP_INT_ALIGN(cwdLen);
    d += cwdLen;

	fill_env(d, env_len, env_val);

    d += tot_env_data_len;

    flist_data = d;

    bzero(cpath, sizeof(char)*1024);
    bzero(fl, tot_flist_len);
	if (sess->protoVer == NDMP_VER_2)
	{
        if (ndmp_restore_mode)
        {
            /* Eliminate common path */
            /* get common path in orig */
            get_common_path_v2(nlist_len, nlist_val_v2, &cpath[0]);
		    fill_flist_v2(fl, nlist_len, nlist_val_v2, cpath, 1);
#if 0
            if (!legatoVeritas)
            {
                /* For lagato, append the orig less the common path in dest */
		        fill_flist_v2(fl, nlist_len, nlist_val_v2, cpath, cwd);
            }
            else
		        fill_flist_v2(fl, nlist_len, nlist_val_v2, NULL, cwd);
#endif
        }
        else
		    fill_flist_v2(fl, nlist_len, nlist_val_v2, NULL, 0);

		/* Get the destination path, pick arbitrarily from first 
	 	 * of the nlist 
	 	 */
		strcpy(dst_path, nlist_val_v2[0].dest_dir);
   	} else {
        if (ndmp_restore_mode)
        {
            /* Eliminate common path */
            get_common_path(nlist_len, nlist_val, &cpath[0]);
		    fill_flist(fl, nlist_len, nlist_val, cpath, 1);
#if 0
            if (!legatoVeritas)
            {
                get_common_path(nlist_len, nlist_val, &cpath[0]);
		        fill_flist(fl, nlist_len, nlist_val, cpath, cwd);
            }
            else
		        fill_flist(fl, nlist_len, nlist_val, NULL, cwd);
#endif
        }
        else
		    fill_flist(fl, nlist_len, nlist_val, NULL, 0);
		strcpy(dst_path, nlist_val[0].dest_dir);
	}

	/* If dst_path is NULL, then we use volId of 'org' path
	 * which is from FILESYSTEM env.
	 */
	if ( !strlen(dst_path) )
	{
		dst_volId = volId;
		found_dst_fs_name = TRUE;
	} else {
		// scan past 'destination /volname', should pts to 2nd '/'
		fs_ptr = dst_path+1;	// skips the first '/' of '/vol1'
    	path_ptr = strchr(fs_ptr, '/');		//find the 2nd '/' if any
    	fs_name_len = (path_ptr != NULL)? path_ptr - fs_ptr : strlen(fs_ptr);
    	strncpy(fs_dstvolname, fs_ptr, fs_name_len);   // get volId into fs_volid
		fs_dstvolname[fs_name_len] = '\0';

    	dst_volId = ndmp_getVolId(sess, fs_dstvolname);
    	if (dst_volId)
    	{
    		strcpy(data->fs_dstvolname, fs_dstvolname);
        	data->fs_dstvolId = dst_volId;
			found_dst_fs_name = TRUE;
    	}
	}
    // printf ("Dest volume Id : %lld %lld\n", volId, dst_volId);

    if ((!found_dst_fs_name) || (!dst_volId))
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_commProcRecovReq: invalid DST voldId. %llx\n", dst_volId));
   		*error = NDMP_ILLEGAL_ARGS_ERR;
		sprintf(buf, "Invalid mount point.\n");
        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
        NDMP_SSC_FREE(ndmp_req);
		return NULL;
    }
	NFXRecovReq->dst_fsvolId = dst_volId;
    sess->nextblobPtr = fl;
    sess->fl_len   = tot_flist_len;
    sess->fl_num   = nlist_len;
    sess->fl_sent  = 0;
    sess->saveList = fl;
    // printf ("fl : Ptr %p\n", fl);

	return ndmp_req;

} /* ndmp_commProcRecovReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procDataStartRecovReq
 *
 * Description:     Process V3 NDMP_DATA_START_RECOVER req.  Pass it to NNIM.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:    10/25/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataStartRecovReq(ndmpSess_t *sess)
{

	ndmp_data_start_recover_req	*cl_req = ( ndmp_data_start_recover_req *) sess->msgRcvd.content;
	ndmp_data_start_recover_rsp	cl_rsp;
    ndmp_NFXReq_t   			*ndmp_req = NULL;   /* relay client cmd to Mover on NNIM */
	ndmp_Err					error = NDMP_NO_ERR;
	int32						msglen;


    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procDataStartRecovReq:\n"));


	/* Formulate the BkupReq for NNIM side */
	ndmp_req = ndmp_commProcRecovReq(sess, cl_req->bu_type,
					cl_req->env.env_len, cl_req->env.env_val, 
					cl_req->nlist.nlist_len, (void *)cl_req->nlist.nlist_val,
						&error, &msglen);

	/* Forward to FP only if there's no error */
	if (error == NDMP_NO_ERR)
	{
    	if (sendAgileMsg (sess->Agilesock, ndmp_req, msglen,
                              "ndmp",       /* NDMP on the N-NIM */
                               NFX_NFP_FP2, 1, MGMTPLANE, /* cpu, slot, plane */
                               NULL, NULL, 0) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataStartRecovReq:  Agile message send to NNIM to START_BKUP failed\n"));
        	error = NDMP_UNDEFINED_ERR;
    	}
		sess->ndmp_ssc_chld_msgStats[FP_DATA_START_RST_REQ_SNT]++;
	}
	NDMP_SSC_FREE(ndmp_req);

	/* Send response to client right away if there's any errors.
	 * Otherwise FP will send response to client later.
	 */
	if (error != NDMP_NO_ERR)
	{
		cl_rsp.error = error;
    	return (ndmp_rspClient(sess, &cl_rsp));
	}

	return NFX_OK;
} /* ndmp_procDataStartRecovReq */

/*-----------------------------------------------------------------
 * Name: 			ndmpv2_procDataStartRecovReq
 *
 * Description:     Process V2 NDMP_DATA_START_RECOVER req.  Pass it to NNIM.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:    4/9/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmpv2_procDataStartRecovReq(ndmpSess_t *sess)
{

	ndmpv2_data_start_recover_req	*cl_req = (ndmpv2_data_start_recover_req *) sess->msgRcvd.content;
	ndmp_data_start_recover_rsp	cl_rsp;
    ndmp_NFXReq_t   			*ndmp_req = NULL;   /* relay client cmd to Mover on NNIM */
	ndmp_NFXRecovReq_t			*NFXRecovReq = NULL;
	int32               		addr_type = cl_req->addr.addr_type;
	ndmp_Err					error = NDMP_NO_ERR;
	int32						msglen;

    NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procDataStartRecovReq:\n"));

	/* First we look at the address of the mover to connect to */

    /* Handle some of the error handling here if we can */
    if ((addr_type != NDMP_ADDR_LOCAL) && (addr_type != NDMP_ADDR_TCP))
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataListen: Invalid addr_type"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
						("ndmp_procMoverListen: Invalid addr_type"));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;

		return (ndmp_rspClient(sess, &cl_rsp));
    }

	/* Formulate the BkupReq for NNIM side */
	ndmp_req = ndmp_commProcRecovReq(sess, cl_req->bu_type,
						cl_req->env.env_len, cl_req->env.env_val, 
						cl_req->nlist.nlist_len, cl_req->nlist.nlist_val,
						&error, &msglen);

	/* Forward to FP only if there's no error */
	if (error == NDMP_NO_ERR)
	{
		NFXRecovReq = (ndmp_NFXRecovReq_t *) ndmp_req->content;
    	NFXRecovReq->addr_typ = addr_type;

		/* If remote bkup, xfer the Mover address so Data server 
		 * can connect to 
		 */
		if (addr_type == NDMP_ADDR_TCP)
		{
    		NFXRecovReq->ip_addr = cl_req->addr.ndmp_addr_u.tcp_addr.ip_addr;
    		NFXRecovReq->port = cl_req->addr.ndmp_addr_u.tcp_addr.port;
		}

		/* Forward request to NNIM */
    	if (sendAgileMsg (sess->Agilesock, ndmp_req, msglen,
                              "ndmp",       /* NDMP on the N-NIM */
                               NFX_NFP_FP2, 1, MGMTPLANE, /* cpu, slot, plane */
                               NULL, NULL, 0) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procDataStartRecovReq:  Agile message send to NNIM to START_BKUP failed\n"));
        	error = NDMP_UNDEFINED_ERR;
    	}
		sess->ndmp_ssc_chld_msgStats[FP_DATA_START_RST_REQ_SNT]++;
	}
	NDMP_SSC_FREE(ndmp_req);

	/* Send response to client right away if there's any errors.
	 * Otherwise FP will send response to client later.
	 */
	if (error != NDMP_NO_ERR)
	{
		cl_rsp.error = error;
    	return (ndmp_rspClient(sess, &cl_rsp));
	}
	return NFX_OK;

} /* ndmpv2_procDataStartRecovReq */


/*-----------------------------------------------------------------
 * Name: ndmp_procDataStopReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/25/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataStopReq(ndmpSess_t *sess)
{

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataStopReq.\n"));

    /* Forward DATA_STOP request to Mover on N_NIM */
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_DATA_STOP, 0);

	return NFX_OK;

} /* ndmp_procDataStopReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procDataAbortReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/25/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataAbortReq(ndmpSess_t *sess)
{
    char dumpLink[1024], dumpFileName[1024];
    int32 flen;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataAbort.\n"));

    /* Forward MOVER_ABORT request to Mover on N_NIM */
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_DATA_ABORT, 0);

    /*
     * Since we have already aborted, we can remove the entry from the
     * dumpdate file. On FP once we get the op's complete we will delete
     * the snapshot without forwarding it to SSC
     */
    bzero(dumpLink, sizeof(char)*1024);
    bzero(dumpFileName, sizeof(char)*1024);
    sprintf(dumpLink, "%s%lld.dmp", dumpDateDir, sess->Serv.Data.fs_volId);
    if (!((flen = readlink(dumpLink, dumpFileName, 1024)) == -1))
        ndmp_deleteDumpRecord(dumpFileName, sess->Serv.Data.dd.snapshot);

	return NFX_OK;
} /* ndmp_procDataAbortReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procDataListenReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:     10/24/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataListenReq(ndmpSess_t *sess)
{
    ndmp_data_listen_rsp   cl_rsp;
    ndmp_data_listen_req   *cl_req = (ndmp_data_listen_req *) sess->msgRcvd.content;


    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procDataListen: type:%s.\n",
          (cl_req->addr_type == NDMP_ADDR_LOCAL) ? "LOCAL" :
          (cl_req->addr_type == NDMP_ADDR_TCP) ? "TCP" : "undefined"));

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    cl_rsp.error = NDMP_NO_ERR;

    /* Handle some of the error handling here if we can */
	if ((cl_req->addr_type != NDMP_ADDR_LOCAL) && (cl_req->addr_type != NDMP_ADDR_TCP))
    {
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
						("ndmp_procDataListen: Invalid addr_type"));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
    }

    if (cl_rsp.error != NDMP_NO_ERR)
    {
		return (ndmp_rspClient(sess, &cl_rsp));
    }
    /* Forward DATA_LISTEN request to Data Server on N_NIM */
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_DATA_LISTEN, 0);
    return NFX_OK;

} /* ndmp_procDataListen */
/*-----------------------------------------------------------------
 * Name: ndmp_procDataConnReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataConnReq(ndmpSess_t *sess)
{
    ndmp_data_conn_rsp   cl_rsp;
    ndmp_data_conn_req   *cl_req = (ndmp_data_conn_req *) sess->msgRcvd.content;
	int32				addr_type = cl_req->addr.addr_type;

    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procDataConn: type:%s.\n",
          addr_type == NDMP_ADDR_LOCAL ? "LOCAL" :
          addr_type == NDMP_ADDR_TCP ? "TCP" : "undefined"));

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    /* Handle some of the error handling here if we can */
    if ((addr_type != NDMP_ADDR_LOCAL) && (addr_type != NDMP_ADDR_TCP))
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataListen: Invalid addr_type"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
						("ndmp_procMoverListen: Invalid addr_type"));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;

		return (ndmp_rspClient(sess, &cl_rsp));
    }
    /* Forward DATA_CONNECT request to Mover on N_NIM */
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_DATA_CONNECT, 0);

	return NFX_OK;
} /* ndmp_procDataConnReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procDataGetEnvRsp
 *
 * Description:	Process Data Get Env response sent from Data server 
 *				on NNIM to client.  
 *
 * Created by:  Anh Vu
 *
 * Date Created:        01/31/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDataGetEnvRsp(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{         
    ndmp_NFXRsp_t   *rsp = (ndmp_NFXRsp_t *)rcvbuf; 
	ndmp_Hdr		*cl_hdr = (ndmp_Hdr *) &rsp->cl_hdr;
	int32			i, j, rc = NFX_OK, offset_to_str;
	ndmp_data_get_env_rsp	ge_rsp;
	int32			error = rsp->userInfo;
	Data_SSC_t		*data = &sess->Serv.Data;
	ndmp_GetEnv_Rsp_t	*env_rsp = (ndmp_GetEnv_Rsp_t *) rsp->content;
	env_t			*env_t_ptr = (env_t *) env_rsp->content;

    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procDataGetEnvRsp: Begin.\n"));

	bzero(&ge_rsp, sizeof(ndmp_data_get_env_rsp)); 
	ge_rsp.error = error;

	if (error != NDMP_NO_ERR)
	{
		// send back response w/ error and send no content, i.e. no environ vars
		goto DataGetEnvSndRsp;
	}

	ge_rsp.env.env_len = data->srvr_env_len;

    NDMP_SSC_ALLOC(ge_rsp.env.env_val, ndmp_pval,
                                    (data->srvr_env_len * sizeof(ndmp_pval)));
    if (ge_rsp.env.env_val == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDataGetEnvRsp:  Malloc error env_val.\n"));
        return NFX_ERR;
	}

	/* Careful:  the below logic only works for 1 env update */
	offset_to_str = env_rsp->num_env_mod * sizeof(env_t);
	for (j = 0; (j< data->srvr_env_len); j++)
    {
        if (!strcmp(&env_rsp->content[offset_to_str], data->env_lst[j].name))
        {
			break;
        }
   	}
	printf("env_t_ptr->val_len = %d\n", env_t_ptr->val_len);
	printf("env_t_ptr->name_len = %d\n", env_t_ptr->name_len);
	NDMP_SSC_FREE(data->env_lst[j].value);
	NDMP_SSC_ALLOC(data->env_lst[j].value, char, env_t_ptr->val_len);
	NFX_ASSERT(data->env_lst[j].value != NULL);
	strcpy(data->env_lst[j].value, &env_rsp->content[offset_to_str + env_t_ptr->name_len + 1]);

	for (i = 0; i < data->srvr_env_len; i++)
	{
		ge_rsp.env.env_val[i].name = data->env_lst[i].name;
		ge_rsp.env.env_val[i].value = data->env_lst[i].value;
        printf("ndmp_procDataGetEnvRsp:  [%d].name = %s\n", i, data->env_lst[i].name);
	    printf("ndmp_procDataGetEnvRsp:  [%d].value = %s\n", i, data->env_lst[i].value);
	}

DataGetEnvSndRsp:

	if ((rc = ndmpXmtMsg(sess, NDMP_MSG_RSP, cl_hdr->msgId,
            cl_hdr->xmtSeq, NDMP_NO_ERR, (void *)&ge_rsp)) < 0)
    {
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,                ("ndmp_rspClient: error sending %d rsp.\n", cl_hdr->msgId));
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_rspClient: error sending %d rsp.\n", cl_hdr->msgId));
    }

	NDMP_SSC_FREE(ge_rsp.env.env_val);

	return (rc);
}
/*-----------------------------------------------------------------
 * Name: ndmp_procDM_Rsp
 *
 * Description: Process Data & Mover response sent from Data server
 *              or Mover server on NNIM to client.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        01/31/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procDM_Rsp(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXRsp_t   *rsp = (ndmp_NFXRsp_t *)rcvbuf;
	uint32          *msgTyp = (uint32 *)rcvbuf;
	ndmpAgileMsgEnt_t *msgent;
	

	if (ndmpXmtMsg(sess, NDMP_MSG_RSP, rsp->cl_hdr.msgId,
				rsp->cl_hdr.xmtSeq, NDMP_NO_ERR,
				(void *) rsp->content) < 0)
	{
		msgent = ndmpGetAgileMsgEnt(*msgTyp);

		NDMP_DBG_PRINTF(D_INFO, ("SessId %d - Error sending rsp for %s.\n", sess->sessId, msgent->msg_name));
		E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,                                ("Error sending %s rsp.", msgent->msg_name));
		return NFX_ERR;
	}
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procStartRecv_Rsp
 *
 * Description:
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 11/18/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procStartRecv_Rsp(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
    ndmp_NFXReq_t  *ndmp_req = NULL;
	ndmp_NFXSendNamesReq_t *NFXNameReq = NULL;
    ndmp_data_start_recover_rsp cl_rsp;
	int32          n, i, fnum, msgLen, flen;
    char           *d, *c, buf[1024];

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procStartRecv_Rsp:\n"));

    c = d = sess->nextblobPtr;
    flen = fnum = 0;
    i = sess->nextblobPtr - sess->saveList;
    // printf ("fl_len %d\n", sess->fl_len);
    // printf ("fl_num %d\n", sess->fl_num);
    // printf ("fl_sent %d\n", sess->fl_sent);
    // printf ("saveList %p\n", sess->saveList);
    // printf ("nextblob %p\n", sess->nextblobPtr);
    for (; (sess->fl_sent+fnum)<sess->fl_num &&
            (flen < (NDMP_MAX_FLIST_MSG_SIZE - 1024)) ;)
    {
        n = *((int32 *)d);
        d += 4; // Skip the orig file len
        // if (n)
            // printf ("orig [%d] : %s\n", n, d);
        // else
            // printf ("orig [%d] : \n", n);
        NDMP_INT_ALIGN(n);
        d += n; // Skip the orig file
        flen += (4 + n);
        n = *((int32 *)d);
        d += 4; // Skip the dest file len
        // if (n)
            // printf ("dest [%d] : %s\n", n, d);
        // else
            // printf ("dest [%d] : \n", n);
        NDMP_INT_ALIGN(n);
        d += n; // Skip the dest file
        flen += (4 + n);
        fnum++;
    }
    sess->nextblobPtr = d;
    sess->fl_sent += fnum;

    NDMP_DBG_PRINTF(D_INFO, ("flen [%d], fnum [%d]\n", flen, fnum));

	/* Forward NDMP client request to NNIM */
	msgLen = sizeof(ndmp_NFXReq_t) - 1 +
              (sizeof(ndmp_NFXSendNamesReq_t) + flen - 1);

    NDMP_SSC_ALLOC(ndmp_req, ndmp_NFXReq_t, msgLen);
	if (ndmp_req == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procStartRecv_Rsp: Err malloc ndmp_req.\n"));
		sprintf(buf, "Server internal Malloc error.\n");
        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
        NDMP_SSC_FREE(sess->saveList);
        sess->saveList = sess->nextblobPtr = NULL;
        sess->fl_len = sess->fl_num = sess->fl_sent = 0;
	    cl_rsp.error = NDMP_NO_MEM_ERR;
        return (ndmp_rspClient(sess, &cl_rsp));
	}

    memset((void*)ndmp_req, 0, msgLen);
    ndmp_req->hdr.mid        = NDMP_MID_DATA_SEND_NAME_REQ;
    ndmp_req->ndmp_mid       = NDMP_MID_DATA_SEND_NAME_REQ;
    ndmp_req->hdr.sessId     = sess->sessId;
#ifdef NFX_SMP
	ndmp_req->hdr.seqId     = sess->xmt2fp_seqId;
	sess->xmt2fp_seqId ++;
	ndmp_req->hdr.idx        = sess->sess_idx;
#endif
    ndmp_req->userInfo = 1;   // More send names to follow
    if (sess->fl_sent == sess->fl_num)
        ndmp_req->userInfo = 0;

	NFXNameReq = (ndmp_NFXSendNamesReq_t *) ndmp_req->content;
	NFXNameReq->num_files = fnum;
	NFXNameReq->tot_file_name_len = flen;

    memcpy(&NFXNameReq->data[0], c, (d-c));

    if (sendAgileMsg (sess->Agilesock, ndmp_req, msgLen,
                      "ndmp", NFX_NFP_FP2, 1, MGMTPLANE,
                      NULL, NULL, 0) < 0)  
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procStartRecv_Rsp: Names msg sent to NNIM failed\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
                ("ndmp_procStartRecv_Rsp: Cannot forward meg to FP\n"));
        sprintf(buf, "Internal error.\n");
        ndmp_sndLogMsg( sess, NDMP_LOG_ERROR, buf, 0);
	    cl_rsp.error = NDMP_UNDEFINED_ERR;
        ndmp_rspClient(sess, &cl_rsp);
    }
    sess->ndmp_ssc_chld_msgStats[FP_DATA_SEND_NAME_REQ_SNT]++;
    if (!ndmp_req->userInfo)
    {
        NDMP_SSC_FREE(sess->saveList);
        sess->saveList = sess->nextblobPtr = NULL;
        sess->fl_len = sess->fl_num = sess->fl_sent = 0;
        // printf ("Last name message\n");
    }
    NDMP_SSC_FREE(ndmp_req);
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procNotifDM_Rsp
 *
 * Description: Process Data & Mover Notification response sent 
 *				from Data server or Mover server on NNIM to client.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        01/31/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procNotifDM_Rsp(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXRsp_t   *rsp = (ndmp_NFXRsp_t *)rcvbuf;
	uint32          *msgTyp = (uint32 *)rcvbuf;
	ndmpAgileMsgEnt_t *msgent;
	int32	ndmp_mid;
	

	ndmp_mid = (*msgTyp == NDMP_MID_NOTIFY_DATA_READ)?
			NDMP_NOTIFY_DATA_READ: NDMP_NOTIFY_MOVER_PAUSED;

	if (ndmpXmtMsg(sess, NDMP_MSG_REQ, ndmp_mid, 0, NDMP_NO_ERR, (void *) rsp->content) < 0)
	{
		msgent = ndmpGetAgileMsgEnt(*msgTyp);

		NDMP_DBG_PRINTF(D_INFO, ("SessId %d - Error sending rsp for %s.\n", sess->sessId, msgent->msg_name));
		E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,                                ("Error sending %s rsp.", msgent->msg_name));
		return NFX_ERR;
	}
	return NFX_OK;
}


