/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-fs-api.h
 *
 * RCS:  $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: NDMP and FS message definitions (for dump/restore)
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/26/01
 *
 *-----------------------------------------------------------------
 */
#if !defined(NDMP_FS_API_H)
#define NDMP_FS_API_H

#include <exports/fs-ndmp-api.h>

/*
 *-------------------------------------------------
 * NDMP Message Types used between FS & NDMP server
 *-------------------------------------------------
 */
typedef enum {
    NDMP_MID_WRITE_DATA_REQ = NDMP_MSG_BASE+80,	// 9684 - write data to tape or data connection, bkup case
    NDMP_MID_READ_DATA_REQ,						// read data from tape or data connection. restore case
    NDMP_MID_OPS_COMPLETE_REQ,					// bkup/restore process is done w/ err/noerror
    NDMP_MID_SET_DUMPDATE_REQ,					// set dumpdate of when bkup begins
    NDMP_MID_SEND_LOG_REQ,						// FS asks NDMP server to send log to NDMP client.
    NDMP_MID_SEND_FH_DIR_REQ,					// FS asks NDMP server to send DIR File History to NDMP client.
    NDMP_MID_SEND_FH_NODE_REQ				    // FS asks NDMP server to send NODE File History to NDMP client.
} ndmp_fsMsgType_t;

#define NDMP_FS_WR_BUFSIZ		1024

/*------ NDMP msg header ------*/
typedef struct {
    ndmp_fsMsgType_t	mid;        	/* Message Id */
    uint32              clientCookie;	/* NDMP handle/cookie (sessid) */
} ndmp_fsMsgHdr_t;


/*------ DATA WRITE REQUEST -----*/
/* Used by FS to write a buffer of NDMP_FS_WR_BUFSIZ (1k) of data to 
 * tape device or data connection (for local/remote BACKUP respectively).
 */
typedef struct {
    ndmp_fsMsgHdr_t	hdr;            /* NDMP_MID_WRITE_DATA_REQ */
    void	        *cl_cookie;     /* Client's (FS) handle/cookie */ 
	uint64			ioCookie;		/* for i/o completions */
	char			*buf;			/* buffer of data to write, since we're 
									 * in the same board, ptr is ok. */
} ndmp_write_data_req_msg_t;

// For response to Rd/Wr i/o, we'll send FS IODone


/*------ DATA READ REQUEST -----*/

/* Used by FS to read a buffer (of size indicated 
 * to FS during fs_restore request) to read data from tape device or 
 * data connection.  This is for the local/remote RESTORE respectively.  
 */
typedef struct {
    ndmp_fsMsgHdr_t	hdr;            /* NDMP_MID_READ_DATA_REQ */
    void	        *cl_cookie;     /* Client's (FS) handle/cookie */ 
	uint64			ioCookie;	/* for i/o completions */
    int32           numIOs;
} ndmp_read_data_req_msg_t;


/*------ SET DUMPDATE REQUEST -----*/

/* Used by FS to set the time/date when backup begins. 
 * The dumpdate will be returned to FS as 'previousDumpDate' in
 * the fs_dump_req.  A previousDumpDate of 0 indicates there
 * was none.  This dumpdate is saved as ASCII in /usr/local/agile/dumpdates
 * file.  Format for each record per line is:
 *	1) FS name (represented as VolumId), 
 *  2) increment level (dump level), and
 *  3) time (provided by FS via this API, represented in seconds)
 */
typedef struct {
    ndmp_fsMsgHdr_t	hdr;            /* NDMP_MID_SET_DUMPDATE_REQ */
    void	        *cl_cookie;     /* Client's (FS) handle/cookie */ 
	evm_volId_t     file_sys_name;	/* Volume id */
	uint32			date;			/* time in secs (GMT time) */		
	int32			dump_level;		/* 0-9 */
	int32			snapshot_len;	/* len of the snapshot string */
	char			snapshot[FS_DUMP_MAX_SNAP_LEN+1];	/* id of the snapshot */
} ndmp_set_dumpdate_req_msg_t;


/*------ OPS COMPLETE REQUEST -----*/

/* Used by FS to notify NDMP that the operation is
 * either completed or something's wrong internally w/ FS.
 * status - FS_ERR_NONE -> success
 *		  - FS_ERR_IO   -> FS internal error.
 */
typedef struct {
    ndmp_fsMsgHdr_t	hdr;            /* NDMP_MID_OPS_COMPLETE_REQ */
    void	        *cl_cookie;     /* Client's (FS) handle/cookie */ 
	fs_Status		status;
} ndmp_ops_cmplt_req_msg_t;


/*--------- LOG MESSAGE TYPE --------*/

#define NDMP_LOG_MSG_TYP_INFO		1
#define NDMP_LOG_MSG_TYP_ERR		2
#define NDMP_LOG_MSG_TYP_WARNING	3
#define NDMP_LOG_MSG_TYP_DEBUG		4
#define NDMP_LOG_MSG_TYP_PANIC		5
#define NDMP_LOG_MSG_TYP_FILE		6	/* To indicate a file was recoved & its status */


/*------ SEND LOG REQUEST -----------*/
/* Used by FS to tell NDMP server to send log to NDMP client 
 */
typedef struct {
    ndmp_fsMsgHdr_t	hdr;            /* NDMP_MID_SEND_LOG_REQ */
    int32           log_msgTyp;		/* Type of log message defined above */
	fs_Status		status;     	/* status of bkup/restore ops: 
									 * FS_ERR_NONE		// successful
									 * FS_ERR_NXIO		// tmp i/o error
									 * FS_ERR_IO		// real i/o error
									 * FS_ERR_NOENT		// entry not found
									 * FS_ERR_NOSPACE	// no space to write a file
									 * FS_ERR_NODIR		// no path found 
									 */	
									/* The below is for NDMP_LOG_MSG_TYP_FILE: 
									 */
	fs_Status		file_recov_stat;/* May be these can be defined by FS-api.h as
									 * FS_ERR_NONE 	
									 * FS_ERR_NOENT 	//file not found
									 * FS_ERR_EXIST		//dst path exists already
									 * FS_ERR_INVAL 	//invalid arguments provided to FS 
									 * FS_ERR_IO		//catch all or io error 	
									 */	
	int32		file_name_len;		/* Length of file */
	int32		log_txt_len;		/* Length of log msg */
    void	        *cl_cookie;     /* Client's (FS) handle/cookie */ 
	char		blob[1];			/* Name of file not restored, only files w/ name
									 * that matches the name in recovery list.  Followed
									 * by log text.
									 */
} ndmp_send_log_req_msg_t;

typedef enum {
   NDMP_F_DIR = 0,
   NDMP_F_FIFO = 1,
   NDMP_F_CSPEC = 2,
   NDMP_F_BSPEC = 3,
   NDMP_F_REG = 4,
   NDMP_F_SLINK = 5,
   NDMP_F_SOCK = 6,
   NDMP_F_REGISTRY = 7,
   NDMP_F_OTHER = 8
} ndmp_filTyp ;

typedef struct
{
    uint32  mid;                    /* NFX NDMP internal msg id */
#ifdef NFX_SMP
#define FH_DIR_REQ      9633        
#define FH_NODE_REQ     9634         
#else
#define FH_DIR_REQ      9631        
#define FH_NODE_REQ     9632         
#endif
    int32   sessId;
    uint32  clientHdl;
    int     num_files;
    int32   tot_len;                // in bytes, of content below
    int32   dummy;
    char    content[1];             // 
                                    // itself, repeat for each file in num_files
} ndmp_NFX_FHReq_t;

typedef struct fs_DirHistory {
        uint64  parent;
        uint64  inum;
        uint16  len;
        char    name[1];
}fs_DirHistory_t;

#define FS_DIR_HISTORY_ADV(dhp)                                             \
do {                                                                        \
        char    *_tp = (char *) dhp;                                                 \
        _tp = (char *) (((address_t)_tp + sizeof (fs_DirHistory_t) + (dhp)->len + 7) & ~7);      \
        (dhp) = (fs_DirHistory_t *)_tp;                                     \
} while (0);

/*
 * Node History structure
 */

typedef struct fs_NodeHistory {
    uint64      inum;
    int32       file_Type;
    uint32      links;
    fs_Time     mtime;
    fs_Time     atime;
    fs_Time     ctime;
    uint32      group;
    uint32      owner;
    uint64      size;
    uint32      fattr;
    int32       pad;
} fs_NodeHistory_t; //64 bytes
#endif /* NDMP_FS_API_H */
