/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-scsi.c
 *
 * RCS:  $Id: ndmp-scsi.c,v 1.19 2002/10/15 02:59:43 anhv Exp $ 
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: Process all NDMP_SCSI_xxx commands sent by the
 *              NDMP client.  These commands are ways for the client
 *              to manipulate the media changer.  Whether
 *				to open, close, or execute CDBs.   The term SCSI device
 *				and Media changer regards to the same thing.  The driver
 *				(which supports both Scsi device and Media Changer device
 *				in the same module) calls it Media changer, we call SCSI 
 *				to conform to NDMP protocol spec.
 *
 * Created by: Anh Vu
 *
 * Date Created:        01/04/02
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <nfx-types.h>
#include <nfx-defs.h>
#include <exports/anpssc-api.h>
#include <exports/elog-api.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/tape-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"


typedef struct {
    int32       num_dev;
    scsi_id_t   scsi_id;
} Scsi_db;

extern int agileMsgBufSiz;

/*-----------------------------------------------------------------
 * Name: ndmp_closeScsiDev
 *
 * Description:	Send a close request to tape-driver and reset
 * 				our internal SCSI (Media Changer) device states.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_closeScsiDev(ndmpSess_t *sess)
{
    mc_close_req_msg_t  mc_req;
	SCSI_t             	*ScsiSrv = &(sess->Serv.SCSI);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_closeScsiDev:  sessId %d\n", sess->sessId));

	// if (ScsiSrv->hdl != 0)
	// {
		memset((void*)&mc_req, 0, sizeof(mc_close_req_msg_t));
    	mc_req.hdr.mid = MC_MID_CLOSE_REQ;
		strcpy(mc_req.devicename, ScsiSrv->devname);
        mc_req.client_tag = sess->sessId;
    	if (sendAgileMsg(sess->Agilesock, &mc_req, sizeof(mc_req),
                        "tape-driver", 0, 0, 0,  /* to which applic */
                        NULL, NULL, 0) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_closeScsiDev:  Agile message send to CLOSE SCSI device failed\n"));
        	return NFX_ERR;
    	}
		sess->ndmp_ssc_chld_msgStats[MC_CLOSE_REQ_SNT]++;
    	/* reset hdl */
		ScsiSrv->hdl = 0;
	// }

    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiOpenReq
 *
 * Description:	Client issue an OPEN Scsi request.  If the
 *				SCSI device is already opened for this session,
 *				we return error.  Otherwise, we send the driver
 *				the request.  
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/4/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiOpenReq(ndmpSess_t *sess)
{
    ndmp_scsi_open_req  *cl_req;
	ndmp_scsi_open_rsp  cl_rsp;
	mc_open_req_msg_t 	mc_req;

	SCSI_t				*ScsiSrv = &(sess->Serv.SCSI);
	Tape_t				*TapeSrv = &(sess->Serv.Tape);



	/* Look at client's request */
    cl_req = (ndmp_scsi_open_req *) sess->msgRcvd.content;

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp)); 
	
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_ScsiOpenReq: device: %s.\n", cl_req->device));

	
	/* If neither the Tape or Scsi device has been opened for this session, 
	 * then send open request to driver.  An NDMP connection may only have opened
	 * at most ONE device at any time. 
	 */
	if ( (!ScsiSrv->hdl) && (!TapeSrv->hdl) )
	{
		/* The client can give us either the ALIAS or Phys name.
		 * tape-driver can open on either.
		 */
		strcpy(ScsiSrv->devname, cl_req->device);
		memset((void*)&mc_req, 0, sizeof(mc_open_req_msg_t));
		mc_req.hdr.mid = MC_MID_OPEN_REQ;
		strcpy(mc_req.devicename, cl_req->device);
        mc_req.client_tag = sess->sessId;

		/* Query Scsi (Tape) driver */
		sess->ndmp_ssc_chld_msgStats[MC_OPEN_REQ_SNT]++;
		cl_rsp.error = NDMP_UNDEFINED_ERR;		// prep err code in case we have problem sending	
		return (sndMsgtoTapeDrv(sess, &mc_req, sizeof(mc_open_req_msg_t), &cl_rsp));

	} else
	{
	    /* Send Client a respond */
		cl_rsp.error = NDMP_DEVICE_OPENED_ERR;
		return (ndmp_rspClient(sess, &cl_rsp));
	}
	return NFX_OK;

} /* ndmp_procScsiOpenReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiCloseReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/04/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiCloseReq(ndmpSess_t *sess)
{
	mc_close_req_msg_t 		mc_req;
	ndmp_scsi_close_rsp  	cl_rsp;
	SCSI_t                  *ScsiSrv = &(sess->Serv.SCSI);
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiCloseReq: \n"));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;		
#if 0
	if (ndmp_chkDevOpen(sess, &cl_rsp, SCSI_DEV) != NFX_OK)
	{
        return NFX_ERR;
	}
#endif
	memset((void*)&mc_req, 0, sizeof(mc_close_req_msg_t));
    mc_req.hdr.mid = MC_MID_CLOSE_REQ;
    strcpy(mc_req.devicename, ScsiSrv->devname);
    mc_req.client_tag = sess->sessId;

	/* Query Scsi driver, reply to client if failed. */
	sess->ndmp_ssc_chld_msgStats[MC_CLOSE_REQ_SNT]++;
	cl_rsp.error = NDMP_UNDEFINED_ERR;
	return (sndMsgtoTapeDrv(sess, &mc_req, sizeof(mc_close_req_msg_t), &cl_rsp));

} /* ndmp_procScsiCloseReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiGetStateReq
 *
 * Description:	We reply w/ "not supported" response.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        3/20/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiGetStateReq(ndmpSess_t *sess)
{
	ndmp_scsi_get_state_rsp  	cl_rsp;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiGetStateReq: \n"));

	bzero(&cl_rsp, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_NOT_SUPPORTED_ERR;		
    ndmp_rspClient(sess, &cl_rsp);
    return NFX_OK;

} /* ndmp_procScsiGetStateReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiResetDevReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        3/20/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiResetDevReq(ndmpSess_t *sess)
{
	ndmp_scsi_reset_dev_rsp  	cl_rsp;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiResetDevReq: \n"));

	bzero(&cl_rsp, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_NOT_SUPPORTED_ERR;		
    ndmp_rspClient(sess, &cl_rsp);
    return NFX_OK;

} /* ndmp_procResetDevReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiResetBusReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        3/20/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiResetBusReq(ndmpSess_t *sess)
{
	ndmp_scsi_reset_bus_rsp  	cl_rsp;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiResetBusReq: \n"));

	bzero(&cl_rsp, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_NOT_SUPPORTED_ERR;		
    ndmp_rspClient(sess, &cl_rsp);
    return NFX_OK;

} /* ndmp_procResetBusReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiSetTargetReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        3/20/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiSetTargetReq(ndmpSess_t *sess)
{
	ndmp_scsi_set_target_rsp  	cl_rsp;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiSetTargetReq: \n"));

	bzero(&cl_rsp, sizeof(cl_rsp));
	
	/* Say we did it to please Veritas */
	cl_rsp.error = NDMP_NO_ERR;		
    ndmp_rspClient(sess, &cl_rsp);
    return NFX_OK;

} /* ndmp_procSetTargetReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiExecCdbReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiExecCdbReq(ndmpSess_t	*sess)
{
	ndmp_scsi_exec_cdb_req		*cl_req = (ndmp_exec_cdb_req*) sess->msgRcvd.content;
	ndmp_scsi_exec_cdb_rsp		cl_rsp;
	mc_passthru_req_msg_t		*mc_req;
	SCSI_t              		*ScsiSrv = &(sess->Serv.SCSI);
	int32						msglen, stat;


	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiExecCdbReq: \n"));

	bzero(&cl_rsp, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	if (ndmp_chkDevOpen(sess, &cl_rsp, SCSI_DEV) != NFX_OK)
        return NFX_ERR;

    if (cl_req->flags == NDMP_SCSI_DATA_OUT && cl_req->dataout.dataout_len == 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiExecCdbReq: Invalid dataout len\n"));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
        ndmp_rspClient(sess, &cl_rsp);
        return NFX_ERR;
    }

	msglen = sizeof(mc_passthru_req_msg_t) + cl_req->dataout.dataout_len;
    NDMP_SSC_ALLOC(mc_req, mc_passthru_req_msg_t, msglen);
    if (mc_req == NULL)
	{
	    /* Send Client an error respond */
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiExecCdbReq:  Malloc error\n"));
        cl_rsp.error = NDMP_UNDEFINED_ERR;
        ndmp_rspClient(sess, &cl_rsp);
        return NFX_ERR;
	}
	memset(mc_req, 0, msglen);

	mc_req->hdr.mid = MC_MID_PASSTHRU_REQ;
	mc_req->handle = ScsiSrv->hdl;
	mc_req->flags = (cl_req->flags & NDMP_SCSI_DATA_IN) ? MC_PASSTHRU_FLAGS_DATA_IN:
					(cl_req->flags & NDMP_SCSI_DATA_OUT) ? MC_PASSTHRU_FLAGS_DATA_OUT : 0;
	mc_req->timeout = cl_req->timeout;
	memcpy(&mc_req->cdb, cl_req->cdb.cdb_val, cl_req->cdb.cdb_len);


	/* If direction = IN, datalen is the expected no. of data bytes to read
	 * from SCSI device.  If SCSI return more than this, NDMP is to return 
 	 * just requested len.
	 *
	 * If direction = OUT, datalen is data to be xfered to SCSI device.
	 */
	mc_req->data_len = (cl_req->flags & NDMP_SCSI_DATA_IN) ? cl_req->datain_len:
						(cl_req->flags & NDMP_SCSI_DATA_OUT) ? 
						cl_req->dataout.dataout_len : 0;

	if (cl_req->flags & NDMP_SCSI_DATA_OUT)
	{
		memcpy(mc_req->data, cl_req->dataout.dataout_val, cl_req->dataout.dataout_len);
	}

	/* Query Scsi driver, reply to client if failed. */
	sess->ndmp_ssc_chld_msgStats[MC_PASSTHRU_REQ_SNT]++;
	cl_rsp.error = NDMP_UNDEFINED_ERR;
	stat = (sndMsgtoTapeDrv(sess, mc_req, msglen, &cl_rsp));

	/* Save the max data that we can return the client and a exec_cdb read */
	if (cl_req->flags & NDMP_SCSI_DATA_IN)
		ScsiSrv->datain_len = cl_req->datain_len;

	ScsiSrv->exec_cdb_dir_flags = cl_req->flags;

	if (mc_req)
        NDMP_SSC_FREE(mc_req);
	return stat;	

} /* ndmp_procScsiExecCdbReq */


/*********************** Process Responses from the Scsi Driver *******************/

/*-----------------------------------------------------------------
 * Name:  ndmp_procScsiGetDevLstRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */  
int32
ndmp_procScsiGetDevLstRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)  
{
	mc_get_devlist_rsp_msg_t 	*t_msg;
	ndmp_cfg_get_scsi_info_rsp  cl_rsp;	
	ndmp_dev_info               *ScsiLst;
	uint32						rc;
	int							i, j, oldsiz, minbufsiz, num_dev;
	ndmp_dev_capab 				*CapLst;		



	t_msg = (mc_get_devlist_rsp_msg_t *)rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiGetDevLstRsp:  ecode 0%d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));		

	switch (t_msg->ecode)
	{
		case TAPE_ECODE_SUCCESS: 
			num_dev = t_msg->dev_count;

			if (sess->in_len == agileMsgBufSiz)
    		{
        		/* double Agile receive buffer size & retry request */
				minbufsiz = sizeof(mc_dev_info_t) * num_dev;
				oldsiz = agileMsgBufSiz;
        		agileMsgBufSiz = ((minbufsiz/1024) * 1024) + 1024;

				NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiGetDevLstRsp:  min size needed %d.  AgileMsgBufSiz reset from %d to %d.  Resend request\n", 
					minbufsiz, oldsiz, agileMsgBufSiz));
        		/* resend request */
        		ndmp_procCfgGetScsiInfoReq(sess);
        		return NFX_OK;
    		}

			/* Create space for how many Scsi info entry we needed based
			 * on number of physical devices */
            NDMP_SSC_ALLOC(ScsiLst, ndmp_dev_info,
                                    (num_dev * sizeof(ndmp_dev_info)));
            if (ScsiLst == NULL)
            {
                /* Send Client an error respond */
                NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiGetDevLstRsp:  Malloc error ScsiLst.\n"));
                cl_rsp.error = NDMP_UNDEFINED_ERR;
				goto DevLstErr;
            }
			memset (ScsiLst, 0, (num_dev * sizeof(ndmp_dev_info)));

			/* Walk thru devlist given by Scsi driver */
			for (j=0; j < num_dev; j++)	
			{
				ScsiLst[j].model = t_msg->info[j].scsi_product_id;

				/* Create capability list, for media changer, we only 
				 * have one entry in capability list.
				 */
                NDMP_SSC_ALLOC(CapLst, ndmp_dev_capab, sizeof(ndmp_dev_capab));
				if (CapLst == NULL)
				{
                	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiGetDevLstRsp:  Malloc error CapLst\n"));
               		cl_rsp.error = NDMP_UNDEFINED_ERR;
					goto DevLstErr;
				}
				memset (CapLst, 0, sizeof(ndmp_dev_capab));
				ScsiLst[j].caplist.caplist_len = 1;
				ScsiLst[j].caplist.caplist_val = CapLst;

				/* Fill in capability list */
				CapLst->devname = t_msg->info[j].devicename;

			}
		    /* Put into response */
			cl_rsp.error = NDMP_NO_ERR;
    		cl_rsp.scsi_info.scsi_info_len = num_dev;
    		cl_rsp.scsi_info.scsi_info_val = ScsiLst;

			break;
		default: 
			cl_rsp.error = NDMP_UNDEFINED_ERR;
	}

DevLstErr:
    /* Send Client a respond */
	rc = ndmp_rspClient(sess, &cl_rsp);

	/* Free all memory we malloced */
    if (t_msg->ecode == TAPE_ECODE_SUCCESS)
    {
        /* for each physical device entry */
        for (i = 0; i < num_dev; i++)
        {
            /* for each logical device entry */
            if (ScsiLst[i].caplist.caplist_val)
                NDMP_SSC_FREE(ScsiLst[i].caplist.caplist_val);
        }
        if (ScsiLst)
            NDMP_SSC_FREE(ScsiLst);
    }

	/* Reset receive buffer size to its original size once we're 
	 * done processing this request successfully.
	 */
	if (agileMsgBufSiz != STDBUFSIZ)
	{
		agileMsgBufSiz = STDBUFSIZ;
	}
    return rc;

} /* ndmp_procScsiGetDevLstRsp */

/*-----------------------------------------------------------------
 * Name:  ndmp_procScsiOpenRsp
 *
 * Description:	Process the Scsi Open response from the Scsi driver.
 *				If we open the device ok, we need to send the driver
 *				a Get State cmd to find out if the device is 
 *				physically write protected and whether a Scsi is 
 *				currently loaded or not.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiOpenRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
	mc_open_rsp_msg_t 	*t_msg;
    ndmp_scsi_open_rsp  cl_rsp;
	SCSI_t    			*ScsiSrv = &(sess->Serv.SCSI); 



    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

	t_msg = (mc_open_rsp_msg_t *)rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiOpenRsp:  ecode 0%d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));		

	switch (t_msg->ecode)
	{
		case TAPE_ECODE_SUCCESS: 
			ScsiSrv->hdl = t_msg->handle; 	/* get device handle */
			//Radek will return these to me soon 3/15/02
			strcpy(ScsiSrv->devname, t_msg->devicename); /* alias name */
			strcpy(ScsiSrv->phys_devname, t_msg->physical_devicename); /* phys dev name */
			cl_rsp.error = NDMP_NO_ERR;
			break;

		case TAPE_ECODE_DEVICE_BUSY: 
			cl_rsp.error = NDMP_DEVICE_BUSY_ERR;
			break;

		case TAPE_ECODE_NO_DEVICE: 
			cl_rsp.error = NDMP_NO_DEVICE_ERR;
			break;
		default:
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiOpenRsp:  Undefined ECODE %d\n", t_msg->ecode));
			cl_rsp.error = NDMP_UNDEFINED_ERR;
	}

	/* This is error condition, no Scsi device is opened, we clean up here */
	if (cl_rsp.error != NDMP_NO_ERR)
	{
		ScsiSrv->hdl = 0;
	}

	/* Send Client a respond */
	return (ndmp_rspClient(sess, &cl_rsp));

}  /* ndmp_procScsiOpenRsp */


/*-----------------------------------------------------------------
 * Name:  ndmp_procScsiCloseRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/04/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiCloseRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    mc_close_rsp_msg_t 	 *t_msg;			// response from Scsi driver
    ndmp_scsi_close_rsp  cl_rsp;			// response to NDMP client
	SCSI_t    			 *ScsiSrv = &(sess->Serv.SCSI); 


    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    t_msg = (mc_close_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiCloseRsp:  ecode 0%d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));		

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
			ScsiSrv->hdl = 0;
 			cl_rsp.error = NDMP_NO_ERR;
			break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
			cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
			break;
		default: 
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiCloseRsp:  Undefined ECODE %d\n", t_msg->ecode));
			cl_rsp.error = NDMP_UNDEFINED_ERR;
	}

	/* Send Client a respond */
	return (ndmp_rspClient(sess, &cl_rsp));
}  

/*-----------------------------------------------------------------
 * Name: ndmp_procScsiPassThruRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        1/04/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procScsiPassThruRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    mc_passthru_rsp_msg_t    *t_msg;
    ndmp_scsi_exec_cdb_rsp   cl_rsp;
	SCSI_t             		*ScsiSrv = &(sess->Serv.SCSI);
	int32					len, rc;


	/* Clear out response */
    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
    
    t_msg = (mc_passthru_rsp_msg_t *) rcvbuf;
    
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procScsiPassThruRsp:  ecode 0%d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));		

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
			
            cl_rsp.error = NDMP_NO_ERR;
            cl_rsp.status = t_msg->scsi_status;

			/* Fill in data_in portion if any */
			if (ScsiSrv->exec_cdb_dir_flags & NDMP_SCSI_DATA_IN)
			{
				/* If length returned by tape-driver is greater than
				 * what the client asked for, then just give the client
				 * what he asks for.
				 */
				len = (t_msg->data_len > ScsiSrv->datain_len)?  ScsiSrv->datain_len: t_msg->data_len;
				cl_rsp.datain.datain_val = (len > 0)? t_msg->data: 0;
				cl_rsp.datain.datain_len = len;
			} else if (ScsiSrv->exec_cdb_dir_flags & NDMP_SCSI_DATA_OUT)
			{
				cl_rsp.dataout_len = t_msg->data_len;
			}

			/* Fill in extended sense */
			cl_rsp.ext_sense.ext_sense_val = t_msg->sense_data; 
			cl_rsp.ext_sense.ext_sense_len = NDMP_MAX_SENSE_DATA_LEN;
            break;

        case TAPE_ECODE_DEVICE_NOT_OPEN:
            cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
            break;

#if 0   // til tape-driver returns a TimeOut err to us
        case TAPE_ECODE_TIMEOUT_ERROR:
            cl_rsp.error = NDMP_TIMEOUT_ERR;
            break;
#endif

        case TAPE_ECODE_IO_ERROR:
        default:
            cl_rsp.error = NDMP_IO_ERR;
    }
	/* Send Client a respond */
    rc = (ndmp_rspClient(sess, &cl_rsp));
	
	// Reset these fields
	ScsiSrv->datain_len = 0;			
	ScsiSrv->exec_cdb_dir_flags = 0;

	return rc;

}  /* ndmp_procScsiPassThruRsp */
