/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-tape.c
 *
 * RCS:  $Id: ndmp-tape.c,v 1.38 2002/10/15 02:59:43 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: Process all NDMP_TAPE_xxx commands sent by the
 *              NDMP client.  These commands are ways for the client
 *              to manipulate the tape drives physically.  Whether
 *				to rewind, unload, move forward, backward, write
 *				file marks, etc.
 *
 * Created by: Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <nfx-types.h>
#include <nfx-defs.h>
#include <exports/anpssc-api.h>
#include <exports/elog-api.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/tape-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"

int32
ndmpv2_procTapeGetStateRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src);

typedef struct {
    int32       num_dev;
    scsi_id_t   scsi_id;
} tape_db;

extern int agileMsgBufSiz;

/*-----------------------------------------------------------------
 * Name:  ndmp_resetTapeDevState
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
void
ndmp_resetTapeDevState(ndmpSess_t *sess)
{
    Tape_t              *tapeSrv = &(sess->Serv.Tape);

    tapeSrv->hdl = 0;
    tapeSrv->mode = TAPE_MODE_NONE;
    tapeSrv->dev_opening = FALSE;
}

/*-----------------------------------------------------------------
 * Name: ndmp_closeTapeDev
 *
 * Description:	Send a close request to tape-driver and reset
 * 				our internal tape device states.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_closeTapeDev(ndmpSess_t *sess)
{
    tape_close_req_msg_t    tape_req;
	Tape_t              	*tapeSrv = &(sess->Serv.Tape);

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_closeTapeDev:  sessId %d\n", sess->sessId));

	// if (tapeSrv->hdl != 0)
	// {
		memset((void*)&tape_req, 0, sizeof(tape_close_req_msg_t));
    	tape_req.hdr.mid = TAPE_MID_CLOSE_REQ;
        tape_req.client_tag = sess->sessId;
    	strcpy(tape_req.devicename, tapeSrv->devname);
    	NDMP_DBG_PRINTF(D_INFO, ("ndmp_closeTapeDev:  Send tape CLOSE req\n"));
    	if (sendAgileMsg(sess->Agilesock, &tape_req, sizeof(tape_req),
                        "tape-driver", 0, 0, 0,  /* to which applic */
                        NULL, NULL, 0) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_closeTapeDev:  Agile message send to CLOSE tape failed\n"));
        	return NFX_ERR;
    	}
		sess->ndmp_ssc_chld_msgStats[TAPE_CLOSE_REQ_SNT]++;
	// }
    /* reset its mode, hdl, and dev_opening state */
    ndmp_resetTapeDevState(sess);

    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeOpenReq
 *
 * Description:	Client issue an OPEN request.  If either a tape
 *				or SCSI device is already opened for this session,
 *				we return error.  Otherwise, we send Tape-driver
 *				the request.  Once an OPEN rsp is received from
 *				the tape-driver, if the OPEN request is for R/W
 *				we have to do a GET_STATE req to tape-driver to
 *				see if the device is WRITE_PROTECTED physically.
 *				There's 2 types of WRITE-PROTECTIONs, logically
 *				& physically.  Logically, is set per NDMP client's OPEN
 *				request.  Physically, nothing we can do about but
 *				to obey.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeOpenReq(ndmpSess_t *sess)
{
    ndmp_tape_open_req  *cl_req;
	ndmp_tape_open_rsp  cl_rsp;
	tape_open_req_msg_t tape_req;

	Tape_t				*tapeSrv = &(sess->Serv.Tape);
	SCSI_t				*ScsiSrv = &(sess->Serv.SCSI);

	/* Look at client's request */
    cl_req = (ndmp_tape_open_req *) sess->msgRcvd.content;

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp)); 
	
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_TapeOpenReq: Sess: %d device: %s, mode: %s.\n",
            sess->sessId, cl_req->devname, 
			cl_req->mode == NDMP_TAPE_READ_MODE ? "RDONLY" : 
			cl_req->mode == NDMP_TAPE_RDWR_MODE ? "RW" : "undefined"));

	if ((cl_req->mode == NDMP_TAPE_READ_MODE) || (cl_req->mode == NDMP_TAPE_RDWR_MODE))
	{
	
		/* If no tape or SCSI device is opened for this session yet, 
		 * then send open request to tape driver.  An NDMP connection 
		 * may only have opened at most ONE device at any time. 
		 */
		if ((!tapeSrv->hdl) && (!ScsiSrv->hdl) )
		{
			tapeSrv->mode = (cl_req->mode == NDMP_TAPE_READ_MODE)? TAPE_MODE_RD_ONLY: TAPE_MODE_RD_WR;
			memset((void*)&tape_req, 0, sizeof(tape_open_req_msg_t));
			tape_req.hdr.mid = TAPE_MID_OPEN_REQ;
			strcpy(tape_req.devicename, cl_req->devname);
			strcpy(tapeSrv->devname, cl_req->devname);
            tape_req.client_tag = sess->sessId;

			/* Query tape driver */
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeOpenReq:  Set wait_tape_open_rsp to TRUE.\n"));
			sess->wait_tape_open_rsp = TRUE;
			cl_rsp.error = NDMP_UNDEFINED_ERR;		// prep err code in case we have problem sending	
			return (sndMsgtoTapeDrv(sess, &tape_req, sizeof(tape_req), &cl_rsp));

		} else {
			cl_rsp.error = NDMP_DEVICE_OPENED_ERR;
		}
	} else {
		cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
	}

    /* Send Client a respond */
	return (ndmp_rspClient(sess, &cl_rsp));
}

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeCloseReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeCloseReq(ndmpSess_t *sess)
{
	tape_close_req_msg_t 	tape_req;
	ndmp_tape_close_rsp  	cl_rsp;
	Tape_t                  *tapeSrv = &(sess->Serv.Tape);
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeCloseReq: Sess %d\n", sess->sessId));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
#if 0	// Cases where tape-driver dies before it can
    // give us the OPEN response along w/ the handle
    // and the client issues a close, we will go
    // ahead and forward this close to tape-driver
    // so device will get cleaned up properly.

	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;		
	if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
       return NFX_ERR;

#endif

	memset((void*)&tape_req, 0, sizeof(tape_close_req_msg_t));
    tape_req.hdr.mid = TAPE_MID_CLOSE_REQ;
    tape_req.client_tag = sess->sessId;
    strcpy(tape_req.devicename, tapeSrv->devname); 

	/* Query tape driver, reply to client if failed. */
	cl_rsp.error = NDMP_UNDEFINED_ERR;
	return (sndMsgtoTapeDrv(sess, &tape_req, sizeof(tape_req), &cl_rsp));

} /* ndmp_procTapeCloseReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procCloseCTagReq
 *
 * Description: Send close to tape driver for the client tag.
 *
 * Created by:  Nitin Jain
 *
 * Date Created:        12/30/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procCloseCTagReq(uint32 sessId, char *dev)
{
    tape_close_req_msg_t    tape_req;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeCloseCTagReq: Sess %d\n", sessId));

    memset((void*)&tape_req, 0, sizeof(tape_close_req_msg_t));
    tape_req.hdr.mid = TAPE_MID_ID_CLOSE_REQ;
    tape_req.client_tag = sessId;
    strcpy(tape_req.devicename, dev);
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeCloseCTagReq:  Send tape CLOSE req\n"));
    if (sendAgileMsg(-1, &tape_req, sizeof(tape_req),
                     "tape-driver", 0, 0, 0,  /* to which applic */
                     NULL, NULL, 0) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeCloseCTagReq: Agile message send to CLOSE tape failed\n"));
        return NFX_ERR;
    }

    return NFX_OK;
} /* ndmp_procTapeCloseReq */


/*-----------------------------------------------------------------
 * Name: ndmp_procTapeGetStateReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeGetStateReq(ndmpSess_t *sess)
{
	ndmp_tape_get_state_rsp			cl_rsp;
	ndmpv2_tape_get_state_rsp		cl_v2_rsp;
	tape_get_state_req_msg_t 		tape_req;
	void							*rsp;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmpdTapeGetState:\n"));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	if (sess->protoVer == NDMP_VER_2)
	{
		cl_v2_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	} else {
		cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	}
	rsp = (sess->protoVer == NDMP_VER_2)? 
				(void *)&cl_v2_rsp : (void*) &cl_rsp;
	if (ndmp_chkDevOpen(sess, rsp, TAPE_DEV) != NFX_OK)
		return NFX_ERR;

	tape_req.hdr.mid = TAPE_MID_GET_STATE_REQ;
	tape_req.handle = sess->Serv.Tape.hdl;

	/* Query tape driver, reply to client if failed.  */
	if (sess->protoVer == NDMP_VER_2)
	{
		cl_v2_rsp.error = NDMP_UNDEFINED_ERR;
	} else {
		cl_rsp.error = NDMP_UNDEFINED_ERR;
	}
	return (sndMsgtoTapeDrv(sess, &tape_req, sizeof(tape_req), rsp));

} /* ndmp_procTapeGetStateReq */
/*-----------------------------------------------------------------
 * Name: ndmp_sendTapeSpaceReq
 *
 * Description:		This func is used by the Mover on NNIM to space 
 *				the tape device.  It has no way to manipulate the
 *				tape directly from the NNIM so it has to go thru
 *				the SSC Tape Server.  
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_sendTapeSpaceReq(ndmpSess_t *sess, char* rcvbuf, ndmp_src_t *src)
{
	Tape_t                  *tapeSrv = &(sess->Serv.Tape);	
	tape_space_req_msg_t    tape_sp_req;
	ndmp_NFXReq_t			*req = (ndmp_NFXReq_t *)rcvbuf;
	ndmp_tape_mtio_req 		*ts_req = (ndmp_tape_mtio_req *)(req->content);
	ndmp_tape_mtio_op		tape_op	= ts_req->tape_op;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendSpaceReq: OPS-> %s, CNT-> %ld\n", 
          tape_op == NDMP_MTIO_FSF ? "Fwdspace-Filemark" :      // fwd space filemarks
          tape_op == NDMP_MTIO_BSF ? "Bkspace-Filemark" :       // bwd space filemarks
          tape_op == NDMP_MTIO_FSR ? "Fwdspace-Rec" :           // fwd space tape records
          tape_op == NDMP_MTIO_BSR ? "Bkspace-Rec" :            // bwd space tape records
          tape_op == NDMP_MTIO_REW ? "Rewind" :                 // rewind
          tape_op == NDMP_MTIO_EOF ? "Write-Filemarks" :        // write end-of-file marks
          tape_op == NDMP_MTIO_OFF ? "Unload" :                 // unload, eject tape
									"Undefined Ops",  ts_req->count));


#if 0
	if (tapeSrv->req_owner != NDMP_NO_ONE)
	{
		// error, Tape server is probably sending request to tape-driver
		// we back off.

		return NFX_ERR;
	}
	tapeSrv->req_owner = NDMP_MOVER_SERVER;
#endif
	tapeSrv->last_op_media_err = FALSE;

	switch(ts_req->tape_op)
	{
		case NDMP_MTIO_FSR:             // Forward space count of tape records
        case NDMP_MTIO_BSR:             // Backward space count of tape records	
            memset(&tape_sp_req, 0, sizeof(tape_sp_req));

            tape_sp_req.hdr.mid = TAPE_MID_SPACE_REQ;
            tape_sp_req.handle = tapeSrv->hdl;

            tape_sp_req.opcode = TAPE_SPACE_OPCODE_BLOCKS;
            tape_sp_req.count = (ts_req->tape_op == NDMP_MTIO_FSR)? ts_req->count : 0 - ts_req->count;
			break;
		default:
			// error
			return NFX_ERR;	
	}
    /* Query tape driver */
    if (sendAgileMsg(sess->Agilesock, &tape_sp_req, sizeof(tape_space_req_msg_t),
                        "tape-driver", 0, 0, 0,  /* to which applic */
                        NULL, NULL, 0) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("Agile message send failed\n"));
		// Notify NDMP Mover on NNIM that request failed
		// ....
        return NFX_ERR;
    }
	sess->ndmp_ssc_chld_msgStats[TAPE_SPACE_REQ_SNT]++;
    return NFX_OK;
	
}
	
/*-----------------------------------------------------------------
 * Name: ndmp_procTapeMtioReq
 *
 * Description:	Client asks us to space the tape around or write
 *				FileMarks onto the tape, unload, etc.  We in turn
 *				asks the tape-driver.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeMtioReq(ndmpSess_t *sess)
{
	ndmp_tape_mtio_req  	*cl_req	= (ndmp_tape_mtio_req *) sess->msgRcvd.content;
	ndmp_tape_mtio_rsp		cl_rsp;
    Tape_t                  *tapeSrv = &(sess->Serv.Tape);
	tape_space_req_msg_t    tape_sp_req;	
	tape_write_fm_req_msg_t tape_fm_req;
	int						rc;

	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeMtioReq:  OPS->%s,  CNT->%ld.\n",
		  cl_req->tape_op == NDMP_MTIO_FSF ? "Fwdspace-Filemark" :		// fwd space filemarks
		  cl_req->tape_op == NDMP_MTIO_BSF ? "Bkspace-Filemark" :		// bwd space filemarks
		  cl_req->tape_op == NDMP_MTIO_FSR ? "Fwdspace-Rec" :			// fwd space tape records
		  cl_req->tape_op == NDMP_MTIO_BSR ? "Bkspace-Rec" :			// bwd space tape records
		  cl_req->tape_op == NDMP_MTIO_REW ? "Rewind" :					// rewind
		  cl_req->tape_op == NDMP_MTIO_EOF ? "Write-Filemarks" :		// write end-of-file marks
		  cl_req->tape_op == NDMP_MTIO_OFF ? "Unload" :					// unload, eject tape
		  "illegal op", cl_req->count));

#if 0
	if (tapeSrv->req_owner != NDMP_NO_ONE)
    {
        // error, Mover server or another Tape space request is probably 
        // in progress, we back off.

        return NFX_ERR;
    }
#endif
	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */	
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
		return NFX_ERR;

	cl_rsp.error = NDMP_UNDEFINED_ERR;			// prep a default error

	tapeSrv->last_op_media_err = FALSE;
#if 0
	tapeSrv->req_owner = NDMP_TAPE_SERVER;		// specify who requested this tape request
#endif
	switch (cl_req->tape_op)
	{
		case NDMP_MTIO_FSF:				// Forward space count of filemarks	
		case NDMP_MTIO_BSF:				// Backward space count of filemarks
		case NDMP_MTIO_FSR:				// Forward space count of tape records
		case NDMP_MTIO_BSR:				// Backward space count of tape records
			memset(&tape_sp_req, 0, sizeof(tape_sp_req));

			tape_sp_req.hdr.mid = TAPE_MID_SPACE_REQ;
			tape_sp_req.handle = tapeSrv->hdl;
			
			tape_sp_req.opcode = ((cl_req->tape_op == NDMP_MTIO_FSF) || (cl_req->tape_op == NDMP_MTIO_BSF))?
					 TAPE_SPACE_OPCODE_FILEMARKS : TAPE_SPACE_OPCODE_BLOCKS;
			if ((cl_req->tape_op == NDMP_MTIO_FSF) || (cl_req->tape_op == NDMP_MTIO_FSR))
			{
				tape_sp_req.count = cl_req->count;
			} else if ((cl_req->tape_op == NDMP_MTIO_BSF) || (cl_req->tape_op == NDMP_MTIO_BSR))
			{
				tape_sp_req.count = 0 - cl_req->count;
			}
			/* Query tape driver, reply to client if failed. */
			if ((rc = sndMsgtoTapeDrv(sess, &tape_sp_req, sizeof(tape_sp_req), &cl_rsp)) == NFX_OK)
				sess->ndmp_ssc_chld_msgStats[TAPE_SPACE_REQ_SNT]++;
			return rc;

		case NDMP_MTIO_REW:				// Rewind tape
			memset((void*)&tape_sp_req, 0, sizeof(tape_sp_req));

			tape_sp_req.hdr.mid = TAPE_MID_SPACE_REQ;
			tape_sp_req.handle = tapeSrv->hdl;
			tape_sp_req.opcode	= TAPE_SPACE_OPCODE_REWIND;

			/* Query tape driver, reply to client if failed. */
			if ((rc = sndMsgtoTapeDrv(sess, &tape_sp_req, sizeof(tape_sp_req), &cl_rsp)) == NFX_OK)
				sess->ndmp_ssc_chld_msgStats[TAPE_SPACE_REQ_SNT]++;
			return rc;

		case NDMP_MTIO_EOF:				// Write end-of-file marks
            memset((void*)&tape_fm_req, 0, sizeof(tape_fm_req));

			/* Check for write-protected device, send err msg to client */
			if (tapeSrv->mode == TAPE_MODE_RD_ONLY)
			{
		    	/* Send Client a respond */
				cl_rsp.error = NDMP_WRITE_PROTECT_ERR;
				return (ndmp_rspClient(sess, &cl_rsp));
			}

            tape_fm_req.hdr.mid = TAPE_MID_WRITE_FM_REQ;
            tape_fm_req.handle = tapeSrv->hdl;
			tape_fm_req.count	= cl_req->count;	

			/* Query tape driver, reply to client if failed. */
			if ((rc = sndMsgtoTapeDrv(sess, &tape_fm_req, sizeof(tape_fm_req), &cl_rsp)) == NFX_OK)
				sess->ndmp_ssc_chld_msgStats[TAPE_WR_FM_REQ_SNT]++;
			return rc;

		case NDMP_MTIO_OFF:				// Unload, eject the tape
			{
			// Can't test w/ emulator, Radek did implement this one
			tape_load_req_msg_t    	tape_ld_req;
            memset((void*)&tape_ld_req, 0, sizeof(tape_ld_req));

            tape_ld_req.hdr.mid = TAPE_MID_LOAD_REQ;
            tape_ld_req.handle = tapeSrv->hdl;
			tape_ld_req.load_or_unload = 0;

			/* Query tape driver, reply to client if failed. */
			if ((rc = sndMsgtoTapeDrv(sess, &tape_ld_req, sizeof(tape_ld_req), &cl_rsp)) == NFX_OK)
				sess->ndmp_ssc_chld_msgStats[TAPE_LOAD_REQ_SNT]++;
			return rc;

			}
		default:
#if 0
			tapeSrv->req_owner = NDMP_NO_ONE;
#endif
		    /* Send Client a respond */
			cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
			return (ndmp_rspClient(sess, &cl_rsp));
	}
	return NFX_OK;	
}

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeWriteReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeWriteReq(ndmpSess_t *sess)
{
	ndmp_tape_write_req		*cl_req = (ndmp_tape_write_req*) sess->msgRcvd.content;
	ndmp_tape_write_rsp		cl_rsp;
	tape_write_req_msg_t	*tape_req;
	Tape_t              	*tapeSrv = &(sess->Serv.Tape);
	int32					msglen, stat;
	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeWriteReq: len: %d.\n",
		  cl_req->data_out.data_out_len));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
		return NFX_ERR;

	if (tapeSrv->mode == TAPE_MODE_RD_ONLY)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeWriteReq:  Write is not allowed on RDONLY tape device."));
		E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
					("Write is not allowed on RDONLY tape device."));

	    /* Send Client a error respond */
		cl_rsp.error = NDMP_PERMISSION_ERR;
		return (ndmp_rspClient(sess, &cl_rsp));
	}

	if (cl_req->data_out.data_out_len == 0)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeWriteReq:  Nothing's written, write len is 0."));

	    /* Send Client a respond */
		cl_rsp.error = NDMP_NO_ERR;
		return (ndmp_rspClient(sess, &cl_rsp));
	}

	tapeSrv->last_op_media_err = FALSE;

	msglen = sizeof(tape_write_req_msg_t) + cl_req->data_out.data_out_len - 1;
    NDMP_SSC_ALLOC(tape_req, tape_write_req_msg_t, msglen);
    if (tape_req == NULL)
	{
        /* Send Client an error respond */
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeWriteReq:  Malloc error\n"));
        cl_rsp.error = NDMP_UNDEFINED_ERR;
        ndmp_rspClient(sess, &cl_rsp);
        return NFX_ERR;
	}
	memset(tape_req, 0, msglen);	

	tape_req->hdr.mid = TAPE_MID_WRITE_REQ;
	tape_req->handle = tapeSrv->hdl;
	tape_req->bytecount = cl_req->data_out.data_out_len;
	memcpy(&tape_req->data, cl_req->data_out.data_out_val, cl_req->data_out.data_out_len); 

	/* Query tape driver, reply to client if failed. */
	cl_rsp.error = NDMP_UNDEFINED_ERR;
	stat = sndMsgtoTapeDrv(sess, tape_req, msglen, &cl_rsp);

	if (tape_req)
        NDMP_SSC_FREE(tape_req);

	/* Save it for tape-driver response processing */
	tapeSrv->tw_data_out_len = cl_req->data_out.data_out_len;

	return stat;

} /* ndmp_procTapeWriteReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeReadReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeReadReq(ndmpSess_t *sess)
{
	ndmp_tape_read_req		*cl_req = (ndmp_tape_read_req*) sess->msgRcvd.content;
	ndmp_tape_read_rsp		cl_rsp;
	tape_read_req_msg_t		tape_req;
	Tape_t    				*tapeSrv = &(sess->Serv.Tape); 
	int						oldsiz, minbufsiz;
	
	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeReadReq: len:%ld.\n", cl_req->count));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
        return NFX_ERR;

	/* Nothing to read, reply OK */
	if (cl_req->count == 0)
	{
	    /* Send Client a respond */
		cl_rsp.error = NDMP_NO_ERR;
		return (ndmp_rspClient(sess, &cl_rsp));
	}

	/* count exceeding max (0x7fffffff) ? */
	if (cl_req->count > NDMP_MAX_COUNT)
	{
	    /* Send Client a respond */
		cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
		return (ndmp_rspClient(sess, &cl_rsp));
	}
	
	/* See if our Agile recv buf can handle the 
	 * response from tape-driver if client asks
 	 * us to read too much.  200 is here is 
	 * leeway for hdrs.
	 */
	if ((cl_req->count + 200) > STDBUFSIZ)
	{
		oldsiz = agileMsgBufSiz;
		minbufsiz = cl_req->count + 200;	// 200 is for hdrs

        agileMsgBufSiz = (((minbufsiz/ ONE_K_SIZE) + 1) * ONE_K_SIZE) ;
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeReadReq:  min size needed %d.  AgileMsgBufSiz reset from %d to %d before send.\n", 
					minbufsiz, oldsiz, agileMsgBufSiz));
		
    }
	tapeSrv->last_op_media_err = FALSE;

	memset((void*)&tape_req, 0, sizeof(tape_read_req_msg_t));
    tape_req.hdr.mid = TAPE_MID_READ_REQ;
    tape_req.handle = tapeSrv->hdl;
    tape_req.bytecount = cl_req->count;

	/* Query tape driver, reply to client if failed. */
	cl_rsp.error = NDMP_UNDEFINED_ERR;

	return (sndMsgtoTapeDrv(sess, &tape_req, sizeof(tape_req), &cl_rsp));

} /* ndmp_procTapeReadReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeExecCdbReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        9/12/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeExecCdbReq(ndmpSess_t	*sess)
{
	ndmp_tape_exec_cdb_req		*cl_req = (ndmp_exec_cdb_req*) sess->msgRcvd.content;
	ndmp_tape_exec_cdb_rsp		cl_rsp;
	tape_passthru_req_msg_t		*tape_req;
	Tape_t              		*tapeSrv = &(sess->Serv.Tape);
	int32						msglen, stat;
	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeExecCdbReq: \n"));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
	/* See if device's open, reply to client if not */
	cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
	if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
        return NFX_ERR;

    if (cl_req->flags == NDMP_SCSI_DATA_OUT && cl_req->dataout.dataout_len == 0)    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeExecCdbReq: Invalid dataout len\n"));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
        ndmp_rspClient(sess, &cl_rsp);
        return NFX_ERR;
    }

	tapeSrv->last_op_media_err = FALSE;

	msglen = sizeof(tape_passthru_req_msg_t) + cl_req->dataout.dataout_len;
    NDMP_SSC_ALLOC(tape_req, tape_passthru_req_msg_t, msglen);
    if (tape_req == NULL)
	{
	    /* Send Client an error respond */
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeExecCdbReq:  Malloc error\n"));
        cl_rsp.error = NDMP_UNDEFINED_ERR;
        ndmp_rspClient(sess, &cl_rsp);
        return NFX_ERR;
	}
	memset(tape_req, 0, msglen);

	tape_req->hdr.mid = TAPE_MID_PASSTHRU_REQ;
	tape_req->handle = tapeSrv->hdl;
	tape_req->flags = (cl_req->flags & NDMP_SCSI_DATA_IN) ? TAPE_PASSTHRU_FLAGS_DATA_IN:
					(cl_req->flags & NDMP_SCSI_DATA_OUT)?TAPE_PASSTHRU_FLAGS_DATA_OUT: 0;
	tape_req->timeout = cl_req->timeout;
	memcpy(&tape_req->cdb, cl_req->cdb.cdb_val, cl_req->cdb.cdb_len);
	tape_req->data_len = (cl_req->flags & NDMP_SCSI_DATA_IN) ? cl_req->datain_len:
						(cl_req->flags & NDMP_SCSI_DATA_OUT)?cl_req->dataout.dataout_len:0;

	if (cl_req->flags & NDMP_SCSI_DATA_OUT)
	{
		memcpy(tape_req->data, cl_req->dataout.dataout_val, cl_req->dataout.dataout_len);
	}

	/* Query tape driver, reply to client if failed. */
	cl_rsp.error = NDMP_UNDEFINED_ERR;
	stat = (sndMsgtoTapeDrv(sess, tape_req, msglen, &cl_rsp));

	tapeSrv->datain_len = cl_req->datain_len;
	tapeSrv->exec_cdb_dir_flags = cl_req->flags;            

	if (tape_req)
        NDMP_SSC_FREE(tape_req);
	return stat;	
}


/*********************** Process Responses from the Tape Driver *******************/


/*-----------------------------------------------------------------
 * Name:  ndmp_procTapeGetDevLstRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */  
int32
ndmp_procTapeGetDevLstRsp(ndmpSess_t *sess, char8 *rcvbuf)
{
	tape_get_devlist_rsp_msg_t 	*t_msg;
	ndmp_cfg_get_tape_info_rsp  cl_rsp;	
	ndmp_dev_info               *tapeLst;
	ndmp_pval                   *pval;
	uint32						compression, rc;
	tape_db						*tl;
	int							prev_scsi_id, i, num_log_dev, devno, oldsiz, minbufsiz,
								j,		// denotes number of physical devices there are
								model_name_len;
	ndmp_dev_capab 				*CapLst;		
	char8						compression_str[] = "COMPRESSION";

	t_msg = (tape_get_devlist_rsp_msg_t *)rcvbuf;
		
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeGetDevLstRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

	switch (t_msg->ecode)
	{
		case TAPE_ECODE_SUCCESS: 
			num_log_dev = t_msg->dev_count;

			if (sess->in_len == agileMsgBufSiz)
    		{
        		/* double Agile receive buffer size & retry request */
				minbufsiz = sizeof(tape_dev_info_t) * num_log_dev;
				oldsiz = agileMsgBufSiz;
        		agileMsgBufSiz = ((minbufsiz/ONE_K_SIZE) +1) * ONE_K_SIZE;

				NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeGetDevLstRsp:  min size needed %d.  AgileMsgBufSiz reset from %d to %d.  Resend request\n", 
					minbufsiz, oldsiz, agileMsgBufSiz));

        		/* resend request */
        		ndmp_procCfgGetTapeInfoReq(sess);
        		return NFX_OK;
    		}

			// Temporarily group logical devices into a physical dev entry in this
			// array, include # of logical devices per physical device entry.
			// alloc max temp array 
            NDMP_SSC_ALLOC(tl, tape_db, (num_log_dev * sizeof(tape_db)));
            if (tl == NULL)
    		{
        		/* Send Client an error respond */
        		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeGetDevLstRsp:  Malloc error\n"));
        		cl_rsp.error = NDMP_UNDEFINED_ERR;
        		ndmp_rspClient(sess, &cl_rsp);
        		return NFX_ERR;		// always indic err for mallocing err
    		}

			memset ((tape_db *) tl, 0, num_log_dev * sizeof(tape_db));
			prev_scsi_id = 0;
			for (i=0, j=-1; i < num_log_dev; i++)	
			{
				if (t_msg->info[i].scsi_id != prev_scsi_id)
				{
					/* New tape_info entry */
					j++;
					tl[j].scsi_id = t_msg->info[i].scsi_id;	
					prev_scsi_id = t_msg->info[i].scsi_id;
				}
				tl[j].num_dev ++; 
			}	

			/* Create space for how many tape info entry we needed based
			 * on number of physical devices */
			prev_scsi_id = 0;
            NDMP_SSC_ALLOC(tapeLst, ndmp_dev_info,
                                    ((j+1) * sizeof(ndmp_dev_info)));
            if (tapeLst == NULL)
            {
                /* Send Client an error respond */
                NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeGetDevLstRsp:  Malloc error\n"));
                cl_rsp.error = NDMP_UNDEFINED_ERR;
                ndmp_rspClient(sess, &cl_rsp);
				if (tl)
                    NDMP_SSC_FREE(tl);
                return NFX_ERR;     // always indic err for mallocing err
            }

			memset (tapeLst, 0, ((j+1) * sizeof(ndmp_dev_info)));

			/* Walk thru devlist given by tape driver */
			for (i=0, j=-1; i < num_log_dev; i++)	
			{
				if (t_msg->info[i].scsi_id != prev_scsi_id)
				{
					/* We have a new tape_info entry */
					j++; 	
					prev_scsi_id = t_msg->info[i].scsi_id;
					model_name_len = sizeof(t_msg->info[i].scsi_product_id) +
                        sizeof(t_msg->info[i].scsi_vendor_id) + 2;
                                /* 1 for space in between & 1 for NULL */
                    NDMP_SSC_ALLOC(tapeLst[j].model, char, model_name_len);
                    if (tapeLst[j].model == NULL)
					{
                		/* Send Client an error respond */
                		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeGetDevLstRsp:  Malloc error\n"));
                		cl_rsp.error = NDMP_UNDEFINED_ERR;
                		ndmp_rspClient(sess, &cl_rsp);
                		if (tl)
                            NDMP_SSC_FREE(tl);
						if (tapeLst)
                            NDMP_SSC_FREE(tapeLst);
                		return NFX_ERR;     // always indic err for mallocing err
            		}
					strcpy(tapeLst[j].model, t_msg->info[i].scsi_vendor_id);
					strcat(tapeLst[j].model, " ");
					strcat(tapeLst[j].model, t_msg->info[i].scsi_product_id);

					/* Create capability list based on how many logical devices
					 * we have for this physical device.
					 */
                    NDMP_SSC_ALLOC(CapLst, ndmp_dev_capab,
                                   (tl[j].num_dev * sizeof(ndmp_dev_capab)));
					NFX_ASSERT(CapLst != NULL);
					memset (CapLst, 0, (tl[j].num_dev * sizeof(ndmp_dev_capab)));
					tapeLst[j].caplist.caplist_len = tl[j].num_dev;
					tapeLst[j].caplist.caplist_val = CapLst;
					devno = 0;	
				}

				/* Fill in capability list */
				CapLst[devno].devname = t_msg->info[i].devicename;

				if (t_msg->info[i].attributes & TAPE_ATTRIB_REWIND_ON_CLOSE) // rewind/unload attrib
                {
					CapLst[devno].attr |= NDMP_TAPE_ATTR_REWIND;
				}
				if (t_msg->info[i].attributes & TAPE_ATTRIB_UNLOAD_ON_CLOSE)
                {
					CapLst[devno].attr |= NDMP_TAPE_ATTR_UNLOAD;
				}
				compression = t_msg->info[i].attributes & TAPE_ATTRIB_DENSITY_MASK;
				if (compression)											// compression capability
				{
                    NDMP_SSC_ALLOC(pval, ndmp_pval, sizeof (ndmp_pval));
					NFX_ASSERT(pval != NULL);
					memset (pval, 0, sizeof (ndmp_pval));
					pval->name = compression_str;
					NDMP_SSC_ALLOC(pval->value, char, 2);
					NFX_ASSERT(pval->value != NULL);
					sprintf (pval->value, "%d", (compression >> 4) + 1);
					CapLst[devno].capab.capab_len = 1;
					CapLst[devno].capab.capab_val = pval;
				}
				devno ++;
			}
		    /* Put into response */
			cl_rsp.error = NDMP_NO_ERR;
    		cl_rsp.tape_info.tape_info_len = j+1;
    		cl_rsp.tape_info.tape_info_val = tapeLst;

			break;
		default: 
			cl_rsp.error = NDMP_UNDEFINED_ERR;
	}
    /* Send Client a respond */
	rc = ndmp_rspClient(sess, &cl_rsp);

	/* Free all memory we malloced */
	if (t_msg->ecode == TAPE_ECODE_SUCCESS)
	{
		ndmp_dev_info  	*p;
		ndmp_dev_capab 	*l;

		/* for each physical device entry */
		for (i = 0; i < cl_rsp.tape_info.tape_info_len; i++)
		{
			/* for each logical device entry */
			p = &cl_rsp.tape_info.tape_info_val[i];
            NDMP_SSC_FREE(l->capab.capab_val->value);
			for (j = 0; j < p->caplist.caplist_len; j++)
			{
				l = &(p->caplist.caplist_val[j]);
				if (l->capab.capab_len)
				{
                    NDMP_SSC_FREE(l->capab.capab_val->value);
                    NDMP_SSC_FREE(l->capab.capab_val);
				}
			}
            NDMP_SSC_FREE(p->caplist.caplist_val);
		}
        NDMP_SSC_FREE(cl_rsp.tape_info.tape_info_val);
	}
    NDMP_SSC_FREE(tl);

	/* Reset receive buffer size to its original size once we're 
	 * done processing this request successfully.
	 */
	if (agileMsgBufSiz != STDBUFSIZ)
	{
		agileMsgBufSiz = STDBUFSIZ;
	}
    return rc;

} /* ndmp_procTapeGetDevLstRsp */

/*-----------------------------------------------------------------
 * Name:  ndmp_procTapeOpenRsp
 *
 * Description:	Process the Tape Open response from the Tape driver.
 *				If we open the device ok, we need to send the driver
 *				a Get State cmd to find out if the device is 
 *				physically write protected and whether a tape is 
 *				currently loaded or not.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeOpenRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
	tape_open_rsp_msg_t *t_msg;
    ndmp_tape_open_rsp  cl_rsp;
	Tape_t    			*tapeSrv = &(sess->Serv.Tape); 

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

	t_msg = (tape_open_rsp_msg_t *)rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeOpenRsp:  Dev hdl 0x%x, ecode %d (%s). Set wait_tape_open_rsp FALSE\n", 
			t_msg->handle, t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

	sess->wait_tape_open_rsp = FALSE;

	if (sess->conn_await_close)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeOpenRsp:  Set conn_closed to TRUE, wait_tape_open_rsp and conn_await_close to FALSE\n"));
		sess->conn_await_close = FALSE;
		sess->conn_closed = TRUE;

		/* CloseSess() will close the tape device using handle 
		 * we receive here.
		 */
		return NFX_OK;
	}

	switch (t_msg->ecode)
	{
		case TAPE_ECODE_SUCCESS: 
			tapeSrv->hdl = t_msg->handle; 	/* get device handle */

			strcpy(tapeSrv->devname, t_msg->devicename);
			strcpy(tapeSrv->phys_devname, t_msg->physical_devicename);


			if (tapeSrv->mode == TAPE_MODE_RD_WR)
			{
				tape_get_state_req_msg_t 	tape_req;

				/* We keep track of soft write-protected mode, unless
				 * it's physically write-protected.
				 * Find out if it's physically write protected 
				 * (rdonly) 
				 */
				tape_req.hdr.mid = TAPE_MID_GET_STATE_REQ;
				tape_req.handle =  t_msg->handle;

				tapeSrv->dev_opening = TRUE;

				if (sendAgileMsg(sess->Agilesock, &tape_req, sizeof(tape_req),
                        "tape-driver", 0, 0, 0,  /* to which applic */
                        NULL, NULL, 0) < 0)     
				{
					cl_rsp.error = NDMP_UNDEFINED_ERR;
            		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeOpenRsp:  Agile message send failed.  Device may be left opened.\n"));
					break;
				}
				sess->ndmp_ssc_chld_msgStats[TAPE_GET_STATE_REQ_SNT]++;
				return NFX_OK;

			} else {
				tapeSrv->recCnt = 0; 			/* no. of records written */
				cl_rsp.error = NDMP_NO_ERR;
			}
			break;

		case TAPE_ECODE_NO_TAPE_LOADED:
			cl_rsp.error = NDMP_NO_TAPE_LOADED_ERR;
			break;
		case TAPE_ECODE_DEVICE_BUSY: 
			cl_rsp.error = NDMP_DEVICE_BUSY_ERR;
			break;
		case TAPE_ECODE_NO_DEVICE: 
			cl_rsp.error = NDMP_NO_DEVICE_ERR;
			break;
		case TAPE_ECODE_IO_ERROR: 
			cl_rsp.error = NDMP_IO_ERR;
			break;
		default:
			cl_rsp.error = NDMP_NO_DEVICE_ERR;
	}
	/* This is error condition, no tape device is opened, we clean up here */
	if (cl_rsp.error != NDMP_NO_ERR)
	{
		ndmp_resetTapeDevState(sess);
	}
	/* Send Client a respond */
	return (ndmp_rspClient(sess, &cl_rsp));

}  /* ndmp_procTapeOpenRsp */


/*-----------------------------------------------------------------
 * Name:  ndmp_procTapeCloseRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeCloseRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_close_rsp_msg_t *t_msg;		// response from tape driver
    ndmp_tape_close_rsp  cl_rsp;			// response to NDMP client


    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    t_msg = (tape_close_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeCloseRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
			ndmp_resetTapeDevState(sess);
 			cl_rsp.error = NDMP_NO_ERR;
			break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
			cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
			break;
		case TAPE_ECODE_IO_ERROR: 
		default: 
			cl_rsp.error = NDMP_IO_ERR;
	}

	/* Send Client a respond */
	return (ndmp_rspClient(sess, &cl_rsp));
}  
/*-----------------------------------------------------------------
 * Name:   ndmp_procTapeSpaceRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeSpaceRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_space_rsp_msg_t *t_msg;
    ndmp_tape_mtio_rsp  cl_rsp;
	Tape_t    			*tapeSrv = &(sess->Serv.Tape); 
#if 0
	ndmp_NFXRsp_t 		Mover_rsp;
#endif
	int32				error, rc;


    t_msg = (tape_space_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeSpaceRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

	error = NDMP_IO_ERR;
#if 0
	if (tapeSrv->req_owner == NDMP_MOVER_SERVER)
    {
		/* Prepare response to Mover */
    	memset((void*)&Mover_rsp, 0, sizeof(ndmp_NFXRsp_t));

    } else if (tapeSrv->req_owner == NDMP_TAPE_SERVER)
	{
		/* Prepare response to NDMP client */
    	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	}
#endif

   	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
		case TAPE_ECODE_FM_ENCOUNTERED:
		case TAPE_ECODE_EOD_ENCOUNTERED:
			// End of Data, means nada for WRITEs (you write, you don't care)
			// only for READs and FWD SPACE.
		case TAPE_ECODE_EOM_ENCOUNTERED: 
			// end of media, physical tape, means something for WRITEs
			// and READs and Forward Space.
            	error = NDMP_NO_ERR;
            break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
            	error = NDMP_DEV_NOT_OPEN_ERR;
            break;
		case TAPE_ECODE_MEDIUM_ERR: 
				tapeSrv->last_op_media_err = TRUE;
				// io error set for client response
			break;
        case TAPE_ECODE_IO_ERROR:
        default:
			// io error set for client response
    }
#if 0
	if (tapeSrv->req_owner == NDMP_TAPE_SERVER)
	{
#endif
		cl_rsp.error = error;
		cl_rsp.resid_count = t_msg->residue;

	    /* Send NDMP Client a respond */
    	rc =  ndmp_rspClient(sess, &cl_rsp);
#if 0
	} else if (tapeSrv->req_owner == NDMP_MOVER_SERVER)
	{
		Mover_rsp.userInfo = error;
	}
	tapeSrv->req_owner = NDMP_NO_ONE;		// specify who requested this tape request
#endif
    return rc;

}  
/*-----------------------------------------------------------------
 * Name:  ndmp_procTapeWrFMRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeWrFMRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_write_fm_rsp_msg_t *t_msg;
    ndmp_tape_mtio_rsp  	cl_rsp;
	Tape_t    				*tapeSrv = &(sess->Serv.Tape); 

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    t_msg = (tape_write_fm_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeWrFMRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

    cl_rsp.error = NDMP_IO_ERR;

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
            cl_rsp.error = NDMP_NO_ERR;
            break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
            cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
            break;
		case TAPE_ECODE_MEDIUM_ERR: 
			tapeSrv->last_op_media_err = TRUE;
			// io error set
			break;
        case TAPE_ECODE_IO_ERROR:
        default:
			// io error set
    }
    /* Send NDMP Client a respond */
    return (ndmp_rspClient(sess, &cl_rsp));

}   /* ndmp_procTapeWrFMRsp */

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeGetStateRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeGetStateRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{              
	tape_get_state_rsp_msg_t 	*t_msg;
    ndmp_tape_get_state_rsp  	cl_gstate_rsp;
	ndmp_tape_open_rsp  		cl_open_rsp;
	void						*cl_rsp;
	int32						error;
   	Tape_t    					*tapeSrv = &(sess->Serv.Tape);  
 
	if (sess->protoVer == NDMP_VER_2)
	{
		return ndmpv2_procTapeGetStateRsp(sess, rcvbuf, src);
	}
    t_msg = (tape_get_state_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeGetStateRsp:  Dev hdl 0x%x, ecode %d (%s).\n", 
			tapeSrv->hdl, t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

	memset((void*)&cl_open_rsp, 0, sizeof(ndmp_tape_open_rsp));
	memset((void*)&cl_gstate_rsp, 0, sizeof(ndmp_tape_get_state_rsp));

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
            if (tapeSrv->dev_opening == TRUE)
            {
				/* This get_state command was issued by us, after a Open cmds 
				 * to get further info about the device vs. the get_state 
				 * cmd coming directly from the NDMP client.
				 */
				if (tapeSrv->mode == TAPE_MODE_RD_WR)
				{
					/* Check if the tape is physically write-protected */
					if (t_msg->flags & TAPE_STATE_FLAGS_WRT_PROTECT)
					{
						error = NDMP_WRITE_PROTECT_ERR;
						break;
					}	
				}
				tapeSrv->recCnt = 0; 			/* no. of records written */
			} else {	
				/* Get_state here was issued by client, respond w/  */

				/* --- Fill in the INVALID mask ---*/
				if (t_msg->flags & TAPE_STATE_FLAGS_FILENO_VALID)
				{
					cl_gstate_rsp.file_num = t_msg->fileno;					// current file position
				} else {		
					cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_FILE_NUM_INVALID;
				}
				if (t_msg->flags & TAPE_STATE_FLAGS_SOFT_ERRORS_VALID)
				{
					cl_gstate_rsp.soft_errors = t_msg->soft_errors;
				} else {		
					cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_SOFT_ERRORS_INVALID;
				}
				if (t_msg->flags & TAPE_STATE_FLAGS_BLOCKSIZE_VALID)
				{
					cl_gstate_rsp.block_size = t_msg->blocksize;  			// same as tape block size
				} else {	
					cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_BLOCK_SIZE_INVALID;
				}
				if (t_msg->flags & TAPE_STATE_FLAGS_BLOCKNO_VALID)
				{
					cl_gstate_rsp.blockno = t_msg->blockno;					// current tape block no.
				} else {
					cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_BLOCKNO_INVALID;
				}
				if (t_msg->flags & TAPE_STATE_FLAGS_TOTAL_SPACE_VALID)
				{
					cl_gstate_rsp.total_space.low = t_msg->total_space;		// total space in bytes
					cl_gstate_rsp.total_space.high = t_msg->total_space >> 32;
				} else {		
					cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_TOTAL_SPACE_INVALID;
				}

				/* SPACE_REMAINING is always NOT supported.  We don't know what they
				 * mean, space remaining from what was written or from curr position of tape. 
				 */
				cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_SPACE_REMAIN_INVALID;

				/* PARTITION is always not supported */
				cl_gstate_rsp.invalid |= NDMP_TAPE_STATE_PARTITION_INVALID;

				/* --- Fill in the FLAG mask ---*/
				/* Is it NO REWIND upon device close */
				if (!(t_msg->flags & TAPE_STATE_FLAGS_REWIND_ON_CLOSE))
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_NOREWIND;

				/* is it UNLOAD upon close */
				if (t_msg->flags & TAPE_STATE_FLAGS_UNLOAD_ON_CLOSE)
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_UNLOAD;

				/* Is it WRITE-PROTECTED physically? */
				if (t_msg->flags & TAPE_STATE_FLAGS_WRT_PROTECT) 
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_WR_PROT;

				/* Was there a MEDIA_ERR w/ previous tape op? */
				if (tapeSrv->last_op_media_err == TRUE)
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_ERROR;
			}
			error = NDMP_NO_ERR;
			break;
		case TAPE_ECODE_IO_ERROR:
			error = NDMP_IO_ERR;
			break;
		case TAPE_ECODE_DEVICE_NOT_OPEN:
            error = (tapeSrv->dev_opening == TRUE) ? NDMP_IO_ERR: NDMP_DEV_NOT_OPEN_ERR;
			break;
		default:
			error = NDMP_IO_ERR;
	}
	/* If there was an error w/ the last Get_state cmd and that cmd was issued
	 * for the Open cmd, reset the opencmd state.
	 */
	if (tapeSrv->dev_opening == TRUE)
	{
		if (error != NDMP_NO_ERR)
		{
			/* close the half-opened device */
			ndmp_closeTapeDev(sess);
		}

		/* Tell client of our error */
		cl_open_rsp.error = error;
		cl_rsp = (void *) &cl_open_rsp;	
		tapeSrv->dev_opening = FALSE;	/* we're done w/ this state */

	} else {
		cl_gstate_rsp.error = error;
		cl_rsp = (void *) &cl_gstate_rsp;	
	}

    /* Send NDMP Client a respond */
    return (ndmp_rspClient(sess, cl_rsp));

} /* ndmp_procTapeGetStateRsp */

/*-----------------------------------------------------------------
 * Name: ndmp_procTapeLoadRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeLoadRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_load_rsp_msg_t 	*t_msg;
    ndmp_tape_mtio_rsp		cl_rsp;

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    t_msg = (tape_load_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeLoadRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
            cl_rsp.error = NDMP_NO_ERR;
            break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
            cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
            break;
        case TAPE_ECODE_IO_ERROR:
        default:
            cl_rsp.error = NDMP_IO_ERR;
    }
    /* Send NDMP Client a respond */
    return (ndmp_rspClient(sess, &cl_rsp));

}  /* ndmp_procTapeLoadRsp */
/*-----------------------------------------------------------------
 * Name: ndmp_procTapeReadRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeReadRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_read_rsp_msg_t     *t_msg;
    ndmp_tape_read_rsp      cl_rsp;
    Tape_t                  *tapeSrv = &(sess->Serv.Tape);
	int32					rc;


    t_msg = (tape_read_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeReadRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

	tapeSrv->last_op_media_err = FALSE;

    memset((void*)&cl_rsp, 0, sizeof(ndmp_tape_read_rsp));

    cl_rsp.error = NDMP_IO_ERR;

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
        case TAPE_ECODE_FM_ENCOUNTERED:
        case TAPE_ECODE_EOD_ENCOUNTERED:
			// EOD not necessarily means EOT.
        case TAPE_ECODE_ILI:
			/* Invalid length indicator, length asked for doesn't match
			 * length on tape, a residue will be returned.  Positive (+)
			 * residue means client has asked for more than what was
			 * there on tape.  Negative (-) residue means he asked for
			 * less than what was there on tape.
			 */
            cl_rsp.error = (t_msg->ecode == TAPE_ECODE_FM_ENCOUNTERED)? NDMP_EOF_ERR: NDMP_NO_ERR;
			cl_rsp.data_in.data_in_len = t_msg->bytecount;
			if (t_msg->bytecount > 0)
			{
                NDMP_SSC_ALLOC(cl_rsp.data_in.data_in_val, char,
                                                           t_msg->bytecount);
                if (cl_rsp.data_in.data_in_val != NULL)
				{
					memcpy(cl_rsp.data_in.data_in_val, t_msg->data, t_msg->bytecount);  
				} else {
					NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeReadRsp:  Malloc error\n"));
            		cl_rsp.error = NDMP_UNDEFINED_ERR;
				}
			}
            break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
            cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
            break;
        case TAPE_ECODE_MEDIUM_ERR:
			// something physically screwy, we have a problem if this
			// repeats
			tapeSrv->last_op_media_err = TRUE;
			break;	
        case TAPE_ECODE_IO_ERROR:
        case TAPE_ECODE_EOM_ENCOUNTERED:
        default:
    }
   	/* Send Client a respond */
    rc = (ndmp_rspClient(sess, &cl_rsp));

	if (cl_rsp.data_in.data_in_val)
        NDMP_SSC_FREE(cl_rsp.data_in.data_in_val);

	/* Reset receive buffer size to its original size once we're 
	 * done processing this request successfully.
	 */
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeReadRsp: agileMsgBufSiz = %d\n", agileMsgBufSiz));
	if (agileMsgBufSiz != STDBUFSIZ)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeReadRsp: Reset agileMsgBufSiz from %d to %d\n", agileMsgBufSiz, STDBUFSIZ));
		agileMsgBufSiz = STDBUFSIZ;
	}

    return rc;

}  /* ndmp_procTapeReadRsp */
/*-----------------------------------------------------------------
 * Name: ndmp_procTapeWriteRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapeWriteRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_write_rsp_msg_t    *t_msg;
    ndmp_tape_write_rsp      cl_rsp;
    Tape_t                  *tapeSrv = &(sess->Serv.Tape);

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    t_msg = (tape_write_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapeWriteRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
        case TAPE_ECODE_EOM_ENCOUNTERED:

			// EOM means tape-driver guaranteed to have written
			// all data, the data written may passed Logical End of Tape
			// but it would have been written.  So we return the
			// count of what got written here.  If tape-driver
			// couldn't write all that was requested to write, it'll
			// return an I/O error instead.

            cl_rsp.error = (t_msg->ecode == TAPE_ECODE_SUCCESS)? 
					NDMP_NO_ERR: NDMP_EOM_ERR;
            cl_rsp.count = tapeSrv->tw_data_out_len;
            break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
            cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
            break;
        case TAPE_ECODE_MEDIUM_ERR:
    		tapeSrv->last_op_media_err = TRUE;
        case TAPE_ECODE_DATA_SIZE_MISMATCH:
        case TAPE_ECODE_IO_ERROR:
        default:
            cl_rsp.error = NDMP_IO_ERR;
    }
    /* Send Client a respond */
    return (ndmp_rspClient(sess, &cl_rsp));

}  /* ndmp_procTapeWriteRsp */

/*-----------------------------------------------------------------
 * Name: ndmp_procTapePassThruRsp
 *
 * Description:	Response for the EXEC_CDB passthru request.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/01/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procTapePassThruRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{
    tape_passthru_rsp_msg_t  *t_msg;
    ndmp_tape_exec_cdb_rsp   cl_rsp;
    Tape_t                  *TapeSrv = &(sess->Serv.Tape);
    int32                   len, rc;


    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
    
    t_msg = (tape_passthru_rsp_msg_t *) rcvbuf;
    
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procTapePassThruRsp:  ecode %d (%s)\n", 
							t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
			
            cl_rsp.error = NDMP_NO_ERR;
			cl_rsp.status = t_msg->scsi_status;
			if (TapeSrv->exec_cdb_dir_flags & NDMP_SCSI_DATA_IN)
			{
				len = (t_msg->data_len > TapeSrv->datain_len)?  TapeSrv->datain_len: t_msg->data_len;
                if (len > 0)
					cl_rsp.datain.datain_val = t_msg->data;
				cl_rsp.datain.datain_len = len;
			} else if (TapeSrv->exec_cdb_dir_flags & NDMP_SCSI_DATA_OUT)
			{
				cl_rsp.dataout_len = t_msg->data_len;
			}
			/* Fill in extended sense */
			cl_rsp.ext_sense.ext_sense_val = t_msg->sense_data;
			cl_rsp.ext_sense.ext_sense_len = NDMP_MAX_SENSE_DATA_LEN;
            break;
        case TAPE_ECODE_DEVICE_NOT_OPEN:
            cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
            break;

#if 0	// til tape-driver returns a TimeOut err to us
        case TAPE_ECODE_TIMEOUT_ERROR:
            cl_rsp.error = NDMP_TIMEOUT_ERR;
            break;
#endif

        case TAPE_ECODE_IO_ERROR:
        default:
            cl_rsp.error = NDMP_IO_ERR;
    }
	/* Send Client a respond */
    rc = (ndmp_rspClient(sess, &cl_rsp));

	// reset these fields	
	TapeSrv->datain_len = 0;            
	TapeSrv->exec_cdb_dir_flags = 0;            

	return rc;

}  /* ndmp_procTapePassThruRsp */

/*===========================================
 * 			V2 specific Functions
 *===========================================
 */	
/*-----------------------------------------------------------------
 * Name: ndmpv2_procTapeGetStateRsp
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        04/01/02
 *
 *-----------------------------------------------------------------
 */
int32
ndmpv2_procTapeGetStateRsp(ndmpSess_t *sess, char8 *rcvbuf, ndmp_src_t *src)
{              
	tape_get_state_rsp_msg_t 	*t_msg;
    ndmpv2_tape_get_state_rsp  	cl_gstate_rsp;
	ndmp_tape_open_rsp  		cl_open_rsp;
	void						*cl_rsp;
	int32						error;
   	Tape_t    					*tapeSrv = &(sess->Serv.Tape);  
 
 
    t_msg = (tape_get_state_rsp_msg_t *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procTapeGetStateRsp:  Dev hdl 0x%x, ecode %d (%s).\n", 
			tapeSrv->hdl, t_msg->ecode, decode_tape_ecode(t_msg->ecode)));

	bzero((void*)&cl_open_rsp, sizeof(ndmp_tape_open_rsp));
	bzero((void*)&cl_gstate_rsp, sizeof(ndmpv2_tape_get_state_rsp));

    switch (t_msg->ecode)
    {
        case TAPE_ECODE_SUCCESS:
            if (tapeSrv->dev_opening == TRUE)
            {
				/* This get_state command was issued by us, after a Open cmds 
				 * to get further info about the device vs. the get_state 
				 * cmd coming directly from the NDMP client.
				 */
				if (tapeSrv->mode == TAPE_MODE_RD_WR)
				{
					/* Check if the tape is physically write-protected */
					if (t_msg->flags & TAPE_STATE_FLAGS_WRT_PROTECT)
					{
						error = NDMP_WRITE_PROTECT_ERR;
						break;
					}	
				}
				tapeSrv->recCnt = 0; 			/* no. of records written */
			} else {	
				/* Get_state here was issued by client, respond w/  */

				/* --- Fill in the gstate info base on mask ---*/
				if (t_msg->flags & TAPE_STATE_FLAGS_FILENO_VALID)
				{
					cl_gstate_rsp.file_num = t_msg->fileno;					// current file position
				} 
				if (t_msg->flags & TAPE_STATE_FLAGS_SOFT_ERRORS_VALID)
				{
					cl_gstate_rsp.soft_errors = t_msg->soft_errors;
				} 
				if (t_msg->flags & TAPE_STATE_FLAGS_BLOCKSIZE_VALID)
				{
					cl_gstate_rsp.block_size = t_msg->blocksize;  			// same as tape block size
				} 
				if (t_msg->flags & TAPE_STATE_FLAGS_BLOCKNO_VALID)
				{
					cl_gstate_rsp.blockno = t_msg->blockno;					// current tape block no.
				} 
				if (t_msg->flags & TAPE_STATE_FLAGS_TOTAL_SPACE_VALID)
				{
					cl_gstate_rsp.total_space.low = t_msg->total_space;		// total space in bytes
					cl_gstate_rsp.total_space.high = t_msg->total_space >> 32;
				} 

				/* SPACE_REMAINING is always NOT supported.  We don't know what they
				 * mean, space remaining from what was written or from curr position of tape. 
				 */

				/* PARTITION is always not supported */

				/* --- Fill in the FLAG mask ---*/
				/* Is it NO REWIND upon device close */
				if (!(t_msg->flags & TAPE_STATE_FLAGS_REWIND_ON_CLOSE))
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_NOREWIND;

				/* is it UNLOAD upon close */
				if (t_msg->flags & TAPE_STATE_FLAGS_UNLOAD_ON_CLOSE)
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_UNLOAD;

				/* Is it WRITE-PROTECTED logically or physically? */
				if (t_msg->flags & TAPE_STATE_FLAGS_WRT_PROTECT)
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_WR_PROT;

				/* Was there a MEDIA_ERR w/ previous tape op? */
				if (tapeSrv->last_op_media_err == TRUE)
					cl_gstate_rsp.flags |= NDMP_TAPE_STATE_ERROR;
			}
			error = NDMP_NO_ERR;
			break;
		case TAPE_ECODE_IO_ERROR:
			error = NDMP_IO_ERR;
			break;
		case TAPE_ECODE_DEVICE_NOT_OPEN:
            error = (tapeSrv->dev_opening == TRUE) ? NDMP_IO_ERR: NDMP_DEV_NOT_OPEN_ERR;
			break;
		default:
			error = NDMP_IO_ERR;
	}
	/* If there was an error w/ the last Get_state cmd and that cmd was issued
	 * for the Open cmd, reset the opencmd state.
	 */
	if (tapeSrv->dev_opening == TRUE)
	{
		if (error != NDMP_NO_ERR)
		{

			/* close the half-opened device */
			ndmp_closeTapeDev(sess);
		}

		/* Tell client of our error */
		cl_open_rsp.error = error;
		cl_rsp = (void *) &cl_open_rsp;	
		tapeSrv->dev_opening = FALSE;	/* we're done w/ this state */

	} else {
		cl_gstate_rsp.error = error;
		cl_rsp = (void *) &cl_gstate_rsp;	
	}

    /* Send NDMP Client a respond */
    return (ndmp_rspClient(sess, cl_rsp));

} /* ndmpv2_procTapeGetStateRsp */
