/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-ui.c
 *
 * RCS:  $Id: ndmp-ui.c,v 1.28 2002/12/04 22:40:36 nitinj Exp $ 
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:		Process UI msgs.
 *
 * Created by: Anh Vu
 *
 * Date Created:        02/11/02
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/elog-api.h>
#include <exports/anpssc-api.h>
#include <exports/opt-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"
#include "ndmp-ui-api.h"
#include "ndmp-fs-api.h"
#include <nfx-incl.h>

extern int ndmp_enabled;
void print_usage(char *command); 
sessId_entry_t *get_sess_tbl(void);
void ndmp_printMoverInfo(Mover_ui_t *mover);
void ndmp_printDataInfo(Data_ui_t *data);
void print_sess_detail(void *sess);
int print_sess_status(void *sess);
int print_sess_stats(void *rsp);
int print_stats(int *statistics);
int get_and_display_stat(int32 mid, int sessId, int (*print_func)(void *));
opt_getOptionRsp_t *ndmp_snd_option_req(char *req_str);
int ui_send_req(char *toapp, void *req, int req_len, void *resp_buf, 
				int resp_buf_len, int timeout);
int isargdigit(char *arg);
int ui_getSessId(char *sess_str, int32 *sessId);
int32 snd_req_to_all(int32 send_parent, int32 msglen, void *req,
                     int32 rsplen, char *rspbuf, int32 sessId);
char *decode_mvr_st(int32 state);
char *decode_mvr_hr(int32 reas);
char *decode_d_st(int32 state);
char *decode_d_hr(int32 reas);


#define FAIL_STR "NDMP request failed.\n" 
#define MIN_LOG_NUMDATE  1
#define MAX_LOG_NUMDATE  10

#define MAX_BUF_SIZ 8086
char recv_buf[MAX_BUF_SIZ];

/* We support the following:
 * ------------------------
 * - ndmp_ui disable
 * - ndmp_ui enable
 * - ndmp_ui show_active		(show all active sessions, displaying sessids)
 * - ndmp_ui delete_session <-a | -s id> (or <all|sessid>)
 * - ndmp_ui show_state  		(show whether enable or disable)
 * - ndmp_ui show_gstats  (Global Statistics for debug) 
 * - ndmp_ui clr_gstats  (Clear global Statistics for debug) 
 * - ndmp_ui show_stats  <-a | -s id> (or <all|sessid>) (Statistics per active session for debug) 
 * - ndmp_ui show_status  <-a | -s id> (or <all|sessid>) (brief status of session) 
 * - ndmp_ui show_detail  <-a | -s id> (or <all|sessid>) (detail status of session) 
 * - ndmp_ui set_version  <vernum>						(max version advertise to client) 
 * - ndmp_ui debug_level <num> <-a | -s id>
 * - ndmp_ui trace <ON|OFF> 
 */

ndmp_UIEnt_t ndmpUIMsgTbl[] = {

	// cmd name, arg_cnt, func, usage string

	{ "disable",	   	 0,  ndmp_UI_EnaDis,		"disable NDMP" },
	{ "enable",	      	 0,  ndmp_UI_EnaDis,		"enable NDMP" },
	{ "set_version",	 1,  ndmp_UI_SetVer,	 
	  	"set_version <num> - Set Highest version NDMP server will support (2/3)" },
	{ "show_state",	 0,  ndmp_UI_ShowState,	
	  	"show_state - show whether feature is enabled or disabled." },
	{ "show_active",	 0,  ndmp_UI_ShowActive,	
	  	"show_active - show all active session ids." },
	{ "show_gstats",	 0,  ndmp_UI_ShowParStats,		
	  	"show_gstats - Show NDMP global statistics" },
	{ "show_stats",	 1,  ndmp_UI_ShowStats,		
	  	"show_stats <all|sessid> - Show statistics of session" },
	{ "show_status",	 1,  ndmp_UI_ShowStatus,		
	  	"show_status <all|sessid> - Show status of session" },
	{ "show_detail",	 1,  ndmp_UI_ShowDetail,
	  	"show_detail <all|sessid> - Show states of session" },
	{ "show_auth_client",	 0,  ndmp_UI_ShowAuthClient,
	  	"show_auth_client - Show list of clients in Registry Manager" },
	{ "show_auth_server",	 0,  ndmp_UI_ShowAuthSrvr,
	  	"show_auth_server - Show server in Registry Manager" },
	{ "clr_gstats",	 0,  ndmp_UI_ClearParStats,		
	  	"clr_gstats - Clear NDMP global statistics" },
	{ "delete_session",	 1,  ndmp_UI_DelSess,	 
	  	"delete_session <all|sessid> - Kill a session" },
	{ "dbg",	 2,  ndmp_UI_SetDbgLevel,
	  	"dbg <level> <all|sessid> -- level = (0(OFF), 1(INFO), 2(DBG)" },
	{ "trace",	 2,  ndmp_UI_SetMsgTrace, "trace <ON|OFF> <all|sessid>" },
	{ "tracefile",	 1,  ndmp_UI_SetMsgTraceFile, "tracefile <numfile>" },
	{ "cleanup_sch", 2,  ndmp_schedule_clean_up, "cleanup_sch [-m MINUTE] [-h HOUR] [-d DATE] [-M MONTH] [-D DAY]" },
	{ "cleanup_sch_show", 0,  ndmp_schedule_clean_up_show, "cleanup_sch_show" },
	{ "cleanup_sch_del", 0,  ndmp_schedule_clean_up_del, "cleanup_sch_del" },
	{ "restore_path", 1,  ndmp_restore_path, "restore_path <ON|OFF>" },
	{ NULL,		      	 0,  NULL, 					NULL }
};

/*-----------------------------------------------------------------
 * Name: main
 *
 * Description:	
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int 
main(int argc, char **argv)
{
	int	i, found = 0;

	if (argc < 2)
		goto proc_error;

	for (i = 0; ndmpUIMsgTbl[i].cmd_name; ++i) 
	{
		if (!strcmp(ndmpUIMsgTbl[i].cmd_name, argv[1])) 
		{
            found = 1;
			if (argc-2 < ndmpUIMsgTbl[i].arg_cnt) 
			{
				goto proc_error;
			} else {
				return (ndmpUIMsgTbl[i].proc_func(argc-1, argv+1));
			}
		}
	}

proc_error:
    if (found)
        printf ("Usage : %s %s\n", argv[0], ndmpUIMsgTbl[i].usage);
    else
	    print_usage(argv[0]);
	exit(1);
}

/*-----------------------------------------------------------------
 * Name: print_usage
 *
 * Description:	
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
void
print_usage(char *command)
{
	int i;

	for (i = 0; ndmpUIMsgTbl[i].cmd_name; ++i) {
		fprintf(stderr, "%s %s %s\n",
			(i == 0 ? "Usage:" : "      "),
			command, ndmpUIMsgTbl[i].usage);
	}
}

/*-----------------------------------------------------------------
 * Name: ui_send_req
 *
 * Description: 	Send request to either the Parent NDMP server
 *					or the Child (the forked process) NDMP server.
 *					Wait for the response and put in resp_buf.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ui_send_req(char *toapp, void *req, int req_len, void *resp_buf, int resp_buf_len, int timeout)
{
    int     rc;
    struct timeval  to;

    to.tv_sec = timeout;
    to.tv_usec = 0;

    rc = sendAgileMsg(-1, req, req_len,
                        toapp, 0, 0, 0, &to,  /* to which applic, cpu, slot, plane, timeout */
                        resp_buf, resp_buf_len); 
	if (rc < 0) {       
        printf(FAIL_STR);
        NDMP_DBG_PRINTF(D_INFO, ("ui_send_req: Agile message send failed\n"));
        return NFX_ERR;
    }
    if (rc == 0) {
        printf(FAIL_STR);
        NDMP_DBG_PRINTF(D_INFO, ("ui_send_req: Agile message send timeout.\n"));
        return NFX_ERR;
    }
	return rc;
}
/*-----------------------------------------------------------------
 * Name: get_sess_tbl
 *
 * Description: Get sessid tbl.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        03/21/02
 *
 *-----------------------------------------------------------------
 */
sessId_entry_t *
get_sess_tbl()
{
    ndmp_NFXGenericUIReq_t 	req;
    ndmp_NFXGetTblRsp_t 	*rsp = (ndmp_NFXGetTblRsp_t*) recv_buf;
	int32					rc;

    bzero(&req, sizeof(ndmp_NFXGenericUIReq_t)); 
    req.mid = NDMP_MID_UI_GET_SESS_TBL;

    /* Query parent NDMP server */
    bzero(recv_buf, sizeof(recv_buf)); 

    NDMP_DBG_PRINTF(D_DBG, ("get_sess_tbl : Send message to ssc_ndmp\n"));
    /* Send to parent server to get table of sessions (children NDMP servers) */
    rc = ui_send_req("ssc_ndmp", &req, sizeof(ndmp_NFXGenericUIReq_t), rsp, sizeof(recv_buf), 30);
    if (rc <= 0) {
        printf(FAIL_STR);
        return NULL;     
    }
    return (&rsp->tbl[0]);

} /* get_sess_tbl */
/*-----------------------------------------------------------------
 * Name: isargdigit
 *
 * Description:  Is the argument w/ specified index a number?
 *
 *-----------------------------------------------------------------
 */
int
isargdigit(char *arg)
{
	int i = 0,  notNum = 0;

    while (i < strlen(arg) && !notNum)
        if (!isdigit(arg[i++])) notNum = 1;

    if (notNum)
    {
        printf(FAIL_STR);
        return 0;
    }
	return 1;
}

/*-----------------------------------------------------------------
 * Name: snd_req_to_all
 *
 * Description:  Send request to set a variable in parent
 *				NDMP server and/or in all or a particular child
 *				NDM process.
 *
 * Created by:  Anh Vu
 *
 * Date:		10/04/02
 *-----------------------------------------------------------------
 */
int32 
snd_req_to_all(int32 send_parent, int32 msglen, void *req,
               int32 rsplen, char *rspbuf, int32 sessId)
{

	ndmp_NFXGen2UIReq_t	 *request = (ndmp_NFXGen2UIReq_t *)req;
	ndmp_NFXGen2UIRsp_t	 *rsp = (ndmp_NFXGen2UIRsp_t *) rspbuf;
	sessId_entry_t		 *tbl;
	int 				 i, rc, found = FALSE, active_sess = FALSE;

	if (send_parent)
	{
		/* Send to parent server first so subsequent sessions will
	 	 * have the same variable set.
	 	 */
        bzero(rspbuf, rsplen);
    	rc = ui_send_req("ssc_ndmp", req, msglen, rspbuf, rsplen, 30);
    	if (rc <= 0) {
        	printf(FAIL_STR);
        	return NFX_ERR;
    	}
	}

	/* Get sessid table of children servers from parent server */
	if ((tbl = get_sess_tbl()) == NULL)
	{
		printf("No active sessions.\n");
		return NFX_ERR;
	}
	/* Now send to the particular child session */
    if (sessId != -1)
    {
		for (i= 0; (i < NDMP_MAX_SESS) && !found; i++)
		{
			if (tbl[i].sessId == sessId)
			{
				found = TRUE;
        		if (!tbl[i].in_use)
        		{
            		printf("Session %d is inactive, no action done.\n", sessId);
            		return NFX_ERR;
        		} else {
            		/* Query a child NDMP server */
            		request->sessId = sessId;	//use 'request' vs 'req' to set 
												//sessid for all types of messages
                    bzero(rspbuf, rsplen);
            		rc = ui_send_req(tbl[i].app_name, req,
                            msglen, rspbuf, rsplen, 30);
            		if ((rc <= 0)  || (rsp->stat == NFX_ERR))
					{       
                		printf(FAIL_STR);
						break;
            		} else {
						return NFX_OK;
					}
        		}
			}
		}
		if (!found)
			printf("Session %d entry not found, no action done.\n", sessId);
    }  else {
		/* Or send to all children who are active */
        for (i = 0; i < NDMP_MAX_SESS; i++)
        {
            if (tbl[i].in_use == TRUE)
            {
				active_sess = TRUE;
                /* Query all active NDMP children servers ourselves */
                request->sessId = tbl[i].sessId;
                bzero(rspbuf, rsplen);
                rc = ui_send_req(tbl[i].app_name, req,
                        msglen, rspbuf, rsplen, 30);
                if ((rc <= 0) || (rsp->stat == NFX_ERR)) 
				{  
                    printf("Request failed for session %d\n", tbl[i].sessId);
                } 
            }
        }
		if (!active_sess)
		{
			printf("There is no current active session.  New sessions will take effect.\n");	
		} else 
			return NFX_OK;
    }

	return NFX_ERR;
} /* snd_req_to_all */

/*-----------------------------------------------------------------
 * Name: ui_getSessId
 *
 * Description:		
 *
 * Created by:  Anh Vu
 *
 *-----------------------------------------------------------------
 */
int 
ui_getSessId(char *sess_str, int32 *sessId)
{
    if (!strcasecmp(sess_str, "all")) {
        *sessId = -1;
	} else if (!isargdigit(sess_str)) {
		return NFX_ERR;
	} else {
        *sessId = atoi(sess_str);
    }
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: 		ndmp_snd_option_req
 *
 * Description:	Send an option request and return a rsp.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        05/09/02
 *
 *-----------------------------------------------------------------
 */
opt_getOptionRsp_t *
ndmp_snd_option_req(char *req_str)
{
	char8               reqbuf[1024];
	opt_msgHdr_t		*opt_hdr, *rsp = (opt_msgHdr_t *) recv_buf;
	int32				rc, buflen;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_snd_option_req: \n"));

	bzero(reqbuf, sizeof(reqbuf));
	opt_hdr = (opt_msgHdr_t *) &reqbuf;
	opt_hdr->msgType = OPT_GET_OPTION_REQ;
	strcpy( reqbuf + sizeof(opt_msgHdr_t), req_str);
	buflen = sizeof(opt_msgHdr_t) + sizeof(opt_getOptionRsp_t);

    rc = ui_send_req("regm", reqbuf, buflen, rsp, sizeof(recv_buf), 30);
    if (rc <= 0) {
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_snd_option_req: Fail send regm requests for %s.\n",
				req_str));
        return NULL;     
    }
	return ((opt_getOptionRsp_t *)(rsp + 1));

}

/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowAuthClient
 *
 * Description:	Show authenticated client entries that are
 *				currently in Registry Manager.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        05/09/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowAuthClient(int argc, char **argv)
{
	char8               id[50];
	opt_getOptionRsp_t  *getrsp;
	int32				i;
	opt_entry_t         *entry;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_UI_ShowAuthClient: \n"));

    getrsp = ndmp_snd_option_req("options.ndmp.auth.client.*");
    if (getrsp == NULL) {
        printf(FAIL_STR);
        return NFX_ERR;     
    }
    if (getrsp->nbEntry < 1)
    {
        printf("There is no Client entry in the Registry Mgr\n");
		return NFX_OK;
    }
	printf("No.      Client Username\n");
	printf("==       ===============\n");

	for (i= 0 ; i < getrsp->nbEntry; i++)
	{
		entry = &getrsp->entry[i];
		strcpy(id, rindex((char *)entry->option, '.') + 1);
		printf("%2d       %s\n", i+1, id);
	}
	return NFX_OK;

} /* ndmp_UI_ShowAuthClient */

/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowAuthSrvr
 *
 * Description:	Show authenticated server entries that are
 *				currently in Registry Manager.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        05/09/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowAuthSrvr(int argc, char **argv)
{
	char8               id[50];
	opt_getOptionRsp_t  *getrsp;
	int32				i;
	opt_entry_t         *entry;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_UI_ShowAuthSrvr: \n"));

    getrsp = ndmp_snd_option_req("options.ndmp.auth.server.*");
    if (getrsp == NULL) {
        printf(FAIL_STR);
        return NFX_OK;     
    }
    if (getrsp->nbEntry < 1)
    {
        printf("There is no Server entry in the Registry Mgr\n");
		return NFX_OK;
    }
	printf("Srvr Username\n");
	printf("=============\n");

	for (i= 0 ; i < getrsp->nbEntry; i++)
	{
		entry = &getrsp->entry[i];
		strcpy(id, rindex((char *)entry->option, '.') + 1);
		printf("%s\n", id);
	}
	return NFX_OK;

} /* ndmp_UI_ShowAuthSrvr */
	
/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowState
 *
 * Description:	Show whether NDMP is Disable/Enable 
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowState(int argc, char **argv)
{
	ndmp_NFXGenericUIReq_t req;
	char 				rcvbuf[256];
	ndmp_NFXShowStateRsp_t *rsp = (ndmp_NFXShowStateRsp_t *)rcvbuf;
	int32				rc;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_UI_ShowState:  option = %s\n", argv[0]));

	req.mid = NDMP_MID_UI_SHOW_STATE;

    /* Query NDMP server */
	bzero(rcvbuf, sizeof(rcvbuf));

    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_ShowState : Send message to ssc_ndmp\n"));
	rc = ui_send_req("ssc_ndmp", &req, sizeof(ndmp_NFXGenericUIReq_t), rcvbuf, sizeof(rcvbuf), 30);
    if (rc <= 0) {   	
		printf("%s NDMP failed.\n", argv[0]);
		return NFX_ERR;
	}
	printf("NDMP Feature               : %s\n", (rsp->state == NDMP_ON)? "ENABLED": "DISABLED");
	printf("Highest version supported  : %d\n", rsp->version);
	printf("Trace                      : %s\n", rsp->trace ? "ON":"OFF");
	printf("Trace files retained       : %d\n", rsp->tracefile);
	printf("Restore full Path          : %s\n", (rsp->restoreMode)?"No":"Yes");

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_UI_SetVer
 *
 * Description:	Set highest version of NDMP to be supported by server.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        05/06/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_SetVer(int argc, char **argv)
{
	ndmp_NFXSetVerReq_t req;
	char 				rcvbuf[256];
	ndmp_NFXSetVerRsp_t *rsp = (ndmp_NFXSetVerRsp_t *)rcvbuf;
	int32				rc, version;


	if (!isargdigit(argv[1]))
		return -1;

	version = atoi(argv[1]);
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_UI_SetVer:  version = %d\n", version));

	req.mid = NDMP_MID_UI_SET_VERSION;
	req.version = version;

    /* Query NDMP server */
	bzero(rcvbuf, sizeof(rcvbuf));

	rc = ui_send_req("ssc_ndmp", &req, sizeof(ndmp_NFXSetVerReq_t), rcvbuf, sizeof(rcvbuf), 30);
    if (rc <= 0) {   	
		printf("Set NDMP version to %d failed.\n", version);
		return NFX_ERR;
	}
	if (rsp->stat == NFX_ERR)
		printf("Version %d is not supported by NFX NDMP server\n", version);
	else if (rsp->already_so)
		printf("Ndmp is already set to version %d.\n", version);
	else 
		printf("Highest NDMP version supported is now set to %d.\n", version);
	
	return NFX_OK;

} /* ndmp_UI_SetVer */

/*-----------------------------------------------------------------
 * Name: ndmp_UI_EnaDis
 *
 * Description:	Disable/Enable NDMP
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_EnaDis(int argc, char **argv)
{
	ndmp_NFXGenericUIReq_t req;
	char 				rcvbuf[256];
	ndmp_NFXDisEnaRsp_t *rsp = (ndmp_NFXDisEnaRsp_t *)rcvbuf;
	int32				rc;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_EnaDis:  option = %s\n", argv[0]));

	req.mid = (!strcmp(argv[0], "enable"))? NDMP_MID_UI_ENABLE_NDMP: 
				NDMP_MID_UI_DISABLE_NDMP;

    /* Query NDMP server */
	bzero(rcvbuf, sizeof(rcvbuf));

    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_EnaDis : Send message to ssc_ndmp\n"));
	rc = ui_send_req("ssc_ndmp", &req, sizeof(ndmp_NFXGenericUIReq_t), rcvbuf, sizeof(rcvbuf), 30);
    if (rc <= 0) {   	
		printf("%s NDMP failed.\n", argv[0]);
		return NFX_ERR;
	}

	if (rsp->mid == NDMP_MID_UI_ENABLE_NDMP)
	{
		if (rsp->already_so)
			printf("Ndmp is already enabled.\n");
		else {
			printf("Ndmp is now enabled.\n");
		}
	} else if (rsp->mid == NDMP_MID_UI_DISABLE_NDMP)
	{
		if (rsp->already_so)
			printf("Ndmp is already disabled.\n");
		else
		{
			if (rsp->running_sess)
			{
				printf("There are sessions still running, new requests will be rejected\n");
			}
			printf("Ndmp is now disabled.\n");
		}
	} else {
		printf("Invalid response %d received for cmd\n", rsp->mid);	
		printf("%s NDMP failed.\n", argv[0]);
	}
	
	return NFX_OK;
} /* ndmp_UI_EnaDis */

/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowActive
 *
 * Description:	Disable/Enable NDMP
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowActive(int argc, char **argv)
{
	int32				i;
	sessId_entry_t		*tbl;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_UI_showActive:\n"));

	if ((tbl = get_sess_tbl()) == NULL)
		printf(FAIL_STR);

	printf("---------------------------------------------------\n");
    printf("Current Active Session(s):\n");
    printf("=========================\n");
    printf("Idx    SessIds        Port   Host Address\n");
    printf("---    -------        ----   ---------------\n");
#if 0
	printf("SessIds     Host Address          AppName   InUse   pid\n");
    printf("-------     ------------          -------   -----   ---\n");
#endif
    for (i = 0; i < NDMP_MAX_SESS; i++)
    {
#if 0
        printf("%2d          %s              %s %s  %d\n", 
                i, inet_ntoa(tbl[i].from.sin_addr), tbl[i].app_name, tbl[i].in_use? "T":"F", tbl[i].pid);
#endif

        if (tbl[i].in_use == TRUE)
        {
            printf("%3d    %10d     %d   %s\n",
                i, tbl[i].sessId, ntohs(tbl[i].from.sin_port),
				inet_ntoa(tbl[i].from.sin_addr)); 
        }
    }
    printf("---------------------------------------------------\n");

	return NFX_OK;
} /* ndmp_UI_ShowActive */

/*-----------------------------------------------------------------
 * Name: print_sess_stats
 *
 * Description:     Show statistics of a session or all sessions.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
print_sess_stats(void *rsp_rcvd)
{
	ndmp_NFXShowChildStatsRsp_t *rsp = (ndmp_NFXShowChildStatsRsp_t *) rsp_rcvd;


	printf("Session statistics\n");
	printf("Sessid: %d\n", rsp->sessId);
	printf("------------------\n");
	print_stats(rsp->statistic);	
	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: print_stats
 *
 * Description:     Show statistics of a session or all sessions.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
print_stats(int *stats)
{
#if 0
    printf("UI_Show_Statistics_req_rcvd  : %d\n", stats[UI_SHOW_STATISTICS_REQ_RCVD]);
    printf("UI_Show_Status_req_rcvd      : %d\n", stats[UI_SHOW_STATUS_REQ_RCVD]);
    printf("UI_Show_Detail_req_rcvd      : %d\n", stats[UI_SHOW_DETAIL_REQ_RCVD]);
    printf("UI_Del_Sess_req_rcvd         : %d\n", stats[UI_DEL_SESS_REQ_RCVD]);
    printf("UI_Set_Dbg_req_rcvd          : %d\n", stats[UI_CHLD_SET_DBG_LEVEL_REQ_RCVD]);
    printf("UI_DM_GetState_req_rcvd      : %d\n\n", stats[UI_DM_GET_STATE_REQ_RCVD]);
#endif
    printf("FP_Mvr_Listen_req_snt        : %-5d\tFP_Mvr_Listen_rsp_rcvd      : %d\n", 
				stats[FP_MOVER_LISTEN_REQ_SNT], stats[FP_MOVER_LISTEN_RSP_RCVD]);	
    printf("FP_Mvr_Conn_req_snt          : %-5d\tFP_Mvr_Conn_rsp_rcvd        : %d\n", 
				stats[FP_MOVER_CONN_REQ_SNT], stats[FP_MOVER_CONN_RSP_RCVD]);	
    printf("FP_Mvr_Get_State_req_snt     : %-5d\tFP_Mvr_Get_State_rsp_rcvd   : %d\n", 
				stats[FP_MOVER_GET_STATE_REQ_SNT], stats[FP_MOVER_GET_STATE_RSP_RCVD]);;	
    printf("FP_Mvr_Cont_req_snt          : %-5d\tFP_Mvr_Cont_rsp_rcvd        : %d\n", 
				stats[FP_MOVER_CONT_REQ_SNT], stats[FP_MOVER_CONT_RSP_RCVD]);	
    printf("FP_Mvr_Abort_req_snt         : %-5d\tFP_Mvr_Abort_rsp_rcvd       : %d\n", 
				stats[FP_MOVER_ABORT_REQ_SNT], stats[FP_MOVER_ABORT_RSP_RCVD]);	
    printf("FP_Mvr_Close_req_snt         : %-5d\tFP_Mvr_Close_rsp_rcvd       : %d\n", 
				stats[FP_MOVER_CLOSE_REQ_SNT], stats[FP_MOVER_CLOSE_RSP_RCVD]);	
    printf("FP_Mvr_Stop_req_snt          : %-5d\tFP_Mvr_Stop_rsp_rcvd        : %d\n", 
				stats[FP_MOVER_STOP_REQ_SNT], stats[FP_MOVER_STOP_RSP_RCVD]);	
    printf("FP_Mvr_Set_Window_req_snt    : %-5d\tFP_Mvr_Set_Window_rsp_rcvd  : %d\n", 
				stats[FP_MOVER_SET_WDW_REQ_SNT], stats[FP_MOVER_SET_WDW_RSP_RCVD]);	
    printf("FP_Mvr_Set_RecSiz_req_snt    : %-5d\tFP_Mvr_Set_RecSiz_rsp_rcvd  : %d\n", 
				stats[FP_MOVER_SET_RECSIZ_REQ_SNT], stats[FP_MOVER_SET_RECSIZ_RSP_RCVD]);	
    printf("FP_Mvr_Read_req_snt          : %-5d\tFP_Mvr_Read_rsp_rcvd        : %d\n\n", 
				stats[FP_MOVER_READ_REQ_SNT], stats[FP_MOVER_READ_RSP_RCVD]);	

    printf("FP_Data_Listen_req_snt       : %-5d\tFP_Data_Listen_rsp_rcvd     : %d\n", 
				stats[FP_DATA_LISTEN_REQ_SNT], stats[FP_DATA_LISTEN_RSP_RCVD]);	
    printf("FP_Data_Conn_req_snt         : %-5d\tFP_Data_Conn_rsp_rcvd       : %d\n", 
				stats[FP_DATA_CONN_REQ_SNT], stats[FP_DATA_CONN_RSP_RCVD]);	
    printf("FP_Data_Start_Bkup_req_snt   : %-5d\tFP_Data_Start_Bkup_rsp_rcvd : %d\n", 
				stats[FP_DATA_START_BKUP_REQ_SNT], stats[FP_DATA_START_BKUP_RSP_RCVD]);	
    printf("FP_Data_Start_Rst_req_snt    : %-5d\tFP_Data_Start_Rst_rsp_rcvd  : %d\n", 
				stats[FP_DATA_START_RST_REQ_SNT], stats[FP_DATA_START_RST_RSP_RCVD]);	
    printf("FP_Data_Send_Name_req_snt    : %-5d\tFP_Data_Send_Name_rcvd      : %d\n", 
				stats[FP_DATA_SEND_NAME_REQ_SNT], stats[FP_DATA_SEND_NAME_RSP_RCVD]);	
    printf("FP_Data_Get_State_req_snt    : %-5d\tFP_Data_Get_State_rsp_rcvd  : %d\n", 
				stats[FP_DATA_GET_STATE_REQ_SNT], stats[FP_DATA_GET_STATE_RSP_RCVD]);	
    printf("FP_Data_Get_Env_req_snt      : %-5d\tFP_Data_Get_Env_rsp_rcvd    : %d\n", 
				stats[FP_DATA_GET_ENV_REQ_SNT], stats[FP_DATA_GET_ENV_RSP_RCVD]);	
    printf("FP_Data_Abort_req_snt        : %-5d\tFP_Data_Abort_rsp_rcvd      : %d\n", 
				stats[FP_DATA_ABORT_REQ_SNT], stats[FP_DATA_ABORT_RSP_RCVD]);	
    printf("FP_Data_Stop_req_snt         : %-5d\tFP_Data_Stop_rsp_rcvd       : %d\n\n", 
				stats[FP_DATA_STOP_REQ_SNT], stats[FP_DATA_STOP_RSP_RCVD]);	

    printf("FP_Notif_Data_Halted_req_rcvd: %-5d\tFP_Log_File_req_rcvd        : %d\n", 
				stats[FP_NOTIFY_DATA_HALTED_REQ_RCVD], stats[FP_LOG_FILE_REQ_RCVD]);
    printf("FP_Notif_Data_Rd_req_rcvd    : %-5d\tFP_Log_Msg_req_rcvd         : %d\n", 
				stats[FP_NOTIFY_DATA_RD_REQ_RCVD], stats[FP_LOG_MSG_REQ_RCVD]);
    printf("FP_Notif_Mvr_Halted_req_rcvd : %-5d\tFP_Fh_Add_File_Dir_req_rcvd     : %d\n", 
				stats[FP_NOTIFY_MOVER_HALTED_REQ_RCVD], stats[FP_FH_ADD_DIR_REQ_RCVD]);
    printf("                             :      \tFP_Fh_Add_File_Node_req_rcvd    : %d\n", 
				stats[FP_FH_ADD_NODE_REQ_RCVD]);
    printf("FP_Notif_Mvr_Paused_req_rcvd : %-5d\tFP_Set_DmpDate_req_rcvd     : %d\n", 
				stats[FP_NOTIFY_MOVER_PAUSED_REQ_RCVD], stats[FP_SET_DMPDATE_REQ_RCVD]);
	printf("FP_Upd_DmpDate_req_rcvd      : %d\n\n", stats[FP_UPD_ENV_REQ_RCVD]);

    printf("RM_Opt_Get_Option_req_snt    : %-5d\tRM_Opt_Get_Option_rsp_rcvd  : %d\n", 
					stats[RM_OPT_GET_OPTION_REQ_SNT], stats[RM_OPT_GET_OPTION_RSP_RCVD]);
    printf("RM_Opt_Get_Option_req_snt    : %-5d\tRM_Opt_Get_Option_rsp_rcvd  : %d\n", 
					stats[RM_OPT_GET_OPTION_REQ_SNT], stats[RM_OPT_GET_OPTION_RSP_RCVD]);
    printf("FS_Vol_FSStat_req_snt        : %-5d\tFS_Vol_FSStat_rsp_rcvd      : %d\n", 
					stats[FS_REQ_FSSTAT_VOL_REQ_SNT], stats[FS_REQ_FSSTAT_VOL_RSP_RCVD]);
    printf("EVM_Vol_Get_NameLst_req_snt  : %-5d\tEVM_Vol_Get_NameLst_rsp_rcvd: %d\n", 
					stats[EVM_VOL_GET_NAME_LST_REQ_SNT], stats[EVM_VOL_GET_NAME_LST_RSP_RCVD]);
    printf("EVM_Vol_Get_Id_req_snt       : %d\n\n", 
					stats[EVM_VOL_GET_ID_REQ_SNT]);

    printf("Tape_Open_req_snt            : %-5d\tTape_Open_rsp_rcvd          : %d\n", 
					stats[TAPE_OPEN_REQ_SNT], stats[TAPE_OPEN_RSP_RCVD]);
    printf("Tape_Close_req_snt           : %-5d\tTape_Close_rsp_rcvd         : %d\n", 
					stats[TAPE_CLOSE_REQ_SNT], stats[TAPE_CLOSE_RSP_RCVD]);
    printf("Tape_Space_req_snt           : %-5d\tTape_Space_rsp_rcvd         : %d\n", 
					stats[TAPE_SPACE_REQ_SNT], stats[TAPE_SPACE_RSP_RCVD]);
    printf("Tape_Wr_FM_req_snt           : %-5d\tTape_Wr_FM_rsp_rcvd         : %d\n", 
					stats[TAPE_WR_FM_REQ_SNT], stats[TAPE_WR_FM_RSP_RCVD]);
    printf("Tape_Get_State_req_snt       : %-5d\tTape_Get_State_rsp_rcvd     : %d\n", 
					stats[TAPE_GET_STATE_REQ_SNT], stats[TAPE_GET_STATE_RSP_RCVD]);
    printf("Tape_Load_req_snt            : %-5d\tTape_Load_rsp_rcvd          : %d\n", 
					stats[TAPE_LOAD_REQ_SNT], stats[TAPE_LOAD_RSP_RCVD]);
    printf("Tape_Rd_req_snt              : %-5d\tTape_Rd_rsp_rcvd            : %d\n", 
					stats[TAPE_RD_REQ_SNT], stats[TAPE_RD_RSP_RCVD]);
    printf("Tape_Wr_req_snt              : %-5d\tTape_Wr_rsp_rcvd            : %d\n", 
					stats[TAPE_WR_REQ_SNT], stats[TAPE_WR_RSP_RCVD]);
    printf("Tape_Passthru_req_snt        : %-5d\tTape_Passthru_rsp_rcvd      : %d\n", 
					stats[TAPE_PASSTHRU_REQ_SNT], stats[TAPE_PASSTHRU_RSP_RCVD]);
    printf("Tape_Get_DevLst_req_snt      : %-5d\tTape_Get_DevLst_rsp_rcvd    : %d\n", 
					stats[TAPE_GET_DEVLST_REQ_SNT], stats[TAPE_GET_DEVLST_RSP_RCVD]);

    printf("MC_Open_req_snt              : %-5d\tMC_Open_rsp_rcvd            : %d\n", 
					stats[MC_OPEN_REQ_SNT], stats[MC_OPEN_RSP_RCVD]);
    printf("MC_Close_req_snt             : %-5d\tMC_Close_rsp_rcvd           : %d\n", 
					stats[MC_CLOSE_REQ_SNT], stats[MC_CLOSE_RSP_RCVD]);
    printf("MC_Passthru_req_snt          : %-5d\tMC_Passthru_rsp_rcvd        : %d\n", 
					stats[MC_PASSTHRU_REQ_SNT], stats[MC_PASSTHRU_RSP_RCVD]);
    printf("MC_Get_DevLst_req_snt        : %-5d\tMc_Get_DevLst_rsp_rcvd      : %d\n", 
					stats[MC_GET_DEVLST_REQ_SNT], stats[MC_GET_DEVLST_RSP_RCVD]);

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: print_sess_status
 *
 * Description:     Show status of a session or all sessions.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
print_sess_status(void *sess_rcvd)
{
	ndmp_Session_t *sess = (ndmp_Session_t *) sess_rcvd;
    uint32         cur_st, d_op;
    Scsi_ui_t      *scsi = &sess->SCSI;
    Tape_ui_t      *tape = &sess->Tape;
    Data_ui_t      *data = &sess->Data;
    Mover_ui_t     *mover = &sess->Mover;

	/* print stats */
    printf("---------------------------------------------------\n");
    printf("Session Status:\n");
    printf("==============\n");
    printf("Session Id:     %d\n", sess->sessId);
    printf("Authorized:     %s\n", (sess->authed)? "Yes": "No" );
    printf("Version:        %d\n", (sess->protoVer != -1)? sess->protoVer: 0);
	printf("Client Address: %s\n", 
			inet_ntoa(sess->host_ipaddr.sin_addr));
	printf("Trace Flag:     %s\n", sess->trace?"ON":"OFF");
	printf("Time Start:     %s\n", sess->start_time);
    printf("Tape device:    %s\n", (tape->hdl)? "Open": "Closed");
    printf("Media Changer:  %s\n", (scsi->hdl)? "Open": "Closed");

    cur_st = mover->state;
    printf("Mover State:    %s\n", (cur_st == MOVER_ST_IDLE)? "Idle":
        (cur_st == MOVER_ST_LISTEN)? "Listen":
            (cur_st == MOVER_ST_ACTIVE)? "Active":
            (cur_st == MOVER_ST_PAUSED)? "Paused": "Halted");

    cur_st = data->state;
    printf("Data State:     %s\n", (cur_st == DATA_ST_IDLE)? "Idle":
        (cur_st == DATA_ST_LISTEN)? "Listen":
        (cur_st == DATA_ST_CONNECTED)? "Connected":
        (cur_st == DATA_ST_ACTIVE)? "Active": "Halted");

    d_op = data->op;
    printf("Data operation: %s\n", (d_op == DATA_OP_NOACTION)? "None":
        (d_op == DATA_OP_BACKUP)? "BACKUP": "RESTORE");
    printf("---------------------------------------------------\n");
	
	return NFX_OK;

} /* print_sess_status */


/*-----------------------------------------------------------------
 * Name: get_and_display_stat
 *
 * Description:		Obtains info and display NDMP session status as
 *					follows:
 *					1.  Send a request to parent NDMP server to get
 *						a tbl of children NDMP servers and their addresses.
 *					2.  Send the Show Stat request to each of the active
 *						children NDMP server.  Wait for the response.
 *					3.  The child server will respond w/ Sess_t which
 *						has info of all NDMP services such as tape, scsi,
 *						data, and mover, etc.
 *					4.  Display the necessary info.
 *
 *					This function is being used by both SHOW_STAT & 
 *					SHOW_DETAIL_STAT cmds.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
get_and_display_stat(int32 mid, int sessId, int (*print_func)(void *))
{
	ndmp_NFXShowReq_t 	c_req;
	char 				rspbuf[sizeof(ndmp_NFXShowChildRsp_t) + 10];
	ndmp_NFXShowChildRsp_t	*c_rsp = 
						(ndmp_NFXShowChildRsp_t *) rspbuf; //child response
	int32				rc, i, active_sess = FALSE, found = FALSE;
	sessId_entry_t		*tbl;


	if ((tbl = get_sess_tbl()) == NULL)
		printf(FAIL_STR);

	/* Send request to a specific child NDMP server */
	c_req.mid = mid;
	if (sessId != -1)
    {
		for (i = 0; i < NDMP_MAX_SESS; i++)
        {
			if (tbl[i].sessId == sessId)
			{
        		if (!tbl[i].in_use)
        		{
            		printf("Session %d is inactive, "
							"no stats available.\n", sessId);
            		return NFX_OK;
        		} else {
					/* Query a child NDMP server */
					c_req.sessId = sessId;
					rc = ui_send_req(tbl[i].app_name, &c_req, 
							sizeof(ndmp_NFXShowReq_t), rspbuf, 
							sizeof(rspbuf), 30);
    				if ((rc <= 0) || (c_rsp->stat == NFX_ERR)) 	
					{   	
						printf(FAIL_STR);
						return NFX_OK;
					}
					if (mid == NDMP_MID_UI_SHOW_STATISTICS)
						(print_func)((void *) rspbuf);
					else 
						(print_func)(&c_rsp->sess);
				}
				found = TRUE;
				break;
			}
		}
		if (!found)
			printf("Session %d entry not found,"
				" no stats available.\n", sessId);
	}  else {
		/* Send request to all active NDMP sessions */
		for (i = 0; i < NDMP_MAX_SESS; i++)
        {
            if (tbl[i].in_use == TRUE)
            {
				active_sess = TRUE;
                /* Query all active NDMP children servers ourselves */
				c_req.sessId = tbl[i].sessId;
				rc = ui_send_req(tbl[i].app_name, &c_req, 
						sizeof(ndmp_NFXShowReq_t), rspbuf, 
						sizeof(rspbuf), 30);
    			if ( (rc <= 0) || (c_rsp->stat == NFX_ERR) )
				{   	
					printf(FAIL_STR);
					return NFX_OK;
				}
				if (mid == NDMP_MID_UI_SHOW_STATISTICS)
					(print_func)((void *) rspbuf);
				else 
					(print_func)(&c_rsp->sess);
            }
		}
		if (!active_sess)
			printf("There is no current active session.\n");
	}
	return NFX_OK;
} /* get_and_display_stat */

char *decode_mvr_st(int32 state)
{
	switch(state)
	{
		case MOVER_ST_IDLE: return "Idle";
		case MOVER_ST_LISTEN: return "Listen";
		case MOVER_ST_ACTIVE: return "Active";
		case MOVER_ST_PAUSED: return "Paused";
		case MOVER_ST_HALTED: return "Halted";
		default: return "Unknown";
	}
}
char *decode_mvr_hr(int32 reas)
{
	switch(reas)
	{
			case MOVER_HALT_NA: return "NA";
			case MOVER_HALT_CONNECT_CLOSED: return "connect_close";
			case MOVER_HALT_ABORTED:  return "aborted";
			case MOVER_HALT_INTERNAL_ERROR: return "internal_err";
			case MOVER_HALT_CONNECT_ERROR: return "connect_err";
			default: return "Unknown";
	}
}
char *decode_d_st(int32 state)
{
	switch(state)
	{
		case DATA_ST_IDLE: return "Idle";
		case DATA_ST_LISTEN: return "Listen";
		case DATA_ST_ACTIVE: return "Active";
		case DATA_ST_CONNECTED: return "Connected";
		case DATA_ST_HALTED: return "Halted";
		default: return "Unknown";
	}
}
char *
decode_d_hr(int32 reas)
{
	switch(reas)
	{
		case DATA_HALT_NA: return "NA";
		case DATA_HALT_SUCCESSFUL: return "Successful";
		case DATA_HALT_ABORTED: return "Aborted";
		case DATA_HALT_INTERNAL_ERROR: return "Internal_err"; 
		case DATA_HALT_MEDIA_ERROR: return "Media_err"; 
		default:  return "Unknown";
	}
}

/*-----------------------------------------------------------------
 * Name: print_sess_detail
 *
 * Description:     Show status of a session or all sessions.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
void
print_sess_detail(void *sess_rcvd)
{
	ndmp_Session_t	*sess = (ndmp_Session_t *) sess_rcvd;
    Scsi_ui_t          *scsi = &sess->SCSI;
    Tape_ui_t          *tape = &sess->Tape;
    Data_ui_t      *data = &sess->Data;
    Mover_ui_t     *mover = &sess->Mover;
	int32			mcur_st, dcur_st, d_op;

	/* Formulate response */
	printf("---------------------------------------------------\n");
	printf("Session Detail Info:\n");
	printf("===================\n");
	printf("Session Id:               %d\n", sess->sessId);
	printf("Authorized:               %s\n", (sess->authed)? "Yes": "No" );
	printf("Version:                  %d\n", (sess->protoVer != -1)? sess->protoVer: 0);
	printf("Auth Type:                %s\n", (sess->authed)? 
			((sess->auth_typ == AUTHNONE)? "None" :
          	(sess->auth_typ == AUTHTXT) ? "Text" :
           	(sess->auth_typ == AUTHMD5) ? "MD5" : "Invalid") : "None");
	printf("Client Address:           %s\n", 
			inet_ntoa(sess->host_ipaddr.sin_addr));
	printf("Trace Flag:               %s\n", sess->trace?"ON":"OFF");
	printf("Trace File Retain:        %d (latest dates)\n", sess->trace_files);
	printf("Time Start:               %s\n", (sess->start_time==NULL)?"N/A":sess->start_time);

	printf("Tape state:               %s\n", (tape->hdl)? "Open": "Closed");
	printf("     alias:               %s\n", (tape->hdl)? tape->devname: "NA");
	printf("     phys name:	          %s\n", (tape->hdl)? tape->phys_devname: "NA");
	printf("     mode:                %s\n", (tape->mode == TP_MODE_RD_ONLY)? 
					"Read Only": (tape->mode == TP_MODE_RD_WR)? "Read Write": "None");
	printf("Media Changer state:      %s\n", (scsi->hdl)? "Open": "Closed");
	printf("     alias:               %s\n", (scsi->hdl)? scsi->devname: "NA");
	printf("     phys name:           %s\n", (scsi->hdl)? scsi->phys_devname: "NA");

	mcur_st = mover->state;
    printf("Mover state:              %s\n", decode_mvr_st(mcur_st));
	printf("     mode:                %s\n", 
			(mcur_st == MOVER_ST_IDLE)? "NA":  
			(mover->mode == MVR_MODE_WRITE)? "Write": 
			(mover->mode == MVR_MODE_READ)? "Read": "None");

    printf("     pause reason:        %s\n", decode_mvr_st(mover->pause_reason));
    printf("     halt reason:         %s\n", decode_mvr_hr(mover->halt_reason));
    printf("     recsiz:              %d\n", mover->recsiz);
    printf("     recnum:              %d\n", mover->recnum);
    printf("     data written:        %lld (Mb)\n", mover->data_written/(1024*1024));
    printf("     seek position:       %lld\n", mover->seek_pos);
    printf("     bytes left to read:  %lld\n", 	
								mover->bytes_left_to_read);
    printf("     window offset:       %lld\n", mover->window_offset);
    printf("     window len:          %lld\n", mover->window_len);
    printf("     addr type:           %s\n", 
			(mover->state == MOVER_ST_IDLE)? "NA": 
			(mover->addr_type == NDMP_ADDR_LOCAL)? "Local": "Tcp");                                     

	dcur_st = data->state;
	printf("Data state:               %s\n", decode_d_st(dcur_st));
	d_op = data->op; 
	printf("     operation:           %s\n", 
		(d_op == DATA_OP_NOACTION)? "None":
		(d_op == DATA_OP_BACKUP)? "Backup": "Restore");	
	printf("     halt reason:         %s\n", decode_d_hr(data->halt_reason));
	printf("     bytes processed:     %lld\n", data->bytes_processed);
    printf("     read offset:         %lld\n", data->read_offset);
    printf("     read len:            %lld\n", data->read_length);
    printf("     addr type:           %s\n", 
			(dcur_st == DATA_ST_IDLE)? "NA":
			(data->addr_type == ADDR_TYP_LOCAL)? "Local": "Tcp");                                     

	printf("---------------------------------------------------\n");

	return;
}

/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowStatus
 *
 * Description:		Show status of a session or all sessions.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowStatus(int argc, char **argv)
{
	int32	sessId;

	sessId = (!strcmp(argv[1], "all"))? -1: atoi(argv[1]);

	get_and_display_stat(NDMP_MID_UI_SHOW_STATUS, sessId, 
					(int (*)(void *)) (print_sess_status));	

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_UI_ClearParStats
 *
 * Description: Show statistics on parent NDMP server.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        03/21/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ClearParStats(int argc, char **argv)
{
    ndmp_NFXGenericUIReq_t 	req;
	int32	rc;

    bzero(&req, sizeof(ndmp_NFXGenericUIReq_t)); 
    req.mid = NDMP_MID_UI_CLEAR_PAR_STATS;

    /* Query parent NDMP server */
    bzero(recv_buf, sizeof(recv_buf)); 

    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_ShowParStats : Send message to ssc_ndmp\n"));
    /* Send to parent server to get table of sessions (children NDMP servers) */
	rc = sendAgileMsg(-1, &req, sizeof(ndmp_NFXGenericUIReq_t), 
			"ssc_ndmp", 0, 0, 0, NULL,  /* to which applic, cpu, slot, plane, timeout */
           NULL, 0);					/* rspbuf, rspbuf_len */

    if (rc < 0) {
        printf(FAIL_STR);
        return NULL;     
    }
	printf("Global stats cleared\n");
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowParStats
 *
 * Description: Show statistics on parent NDMP server.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        03/21/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowParStats(int argc, char **argv)
{
    ndmp_NFXGenericUIReq_t 	req;
    ndmp_NFXShowParStatsRsp_t 	*rsp = (ndmp_NFXShowParStatsRsp_t*) recv_buf;
	int32	rc, i;

    bzero(&req, sizeof(ndmp_NFXGenericUIReq_t)); 
    req.mid = NDMP_MID_UI_SHOW_PAR_STATS;

    /* Query parent NDMP server */
    bzero(recv_buf, sizeof(recv_buf)); 

    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_ShowParStats : Send message to ssc_ndmp\n"));
    /* Send to parent server to get table of sessions (children NDMP servers) */
    rc = ui_send_req("ssc_ndmp", &req, sizeof(ndmp_NFXGenericUIReq_t), rsp, sizeof(recv_buf), 30);
    if (rc <= 0) {
        printf(FAIL_STR);
        return NFX_ERR;     
    }
	i = MAX_SSC_CHLD_STAT;
	printf("Msg statistics between NDMP server and other processes:\n");
	printf("-------------------------------------------------------\n");
    printf("UI_Enable_req_rcvd      : %d\n", rsp->statistic[i+UI_ENABLE_REQ_RCVD]);
    printf("UI_Disable_req_rcvd     : %d\n", rsp->statistic[i+UI_DISABLE_REQ_RCVD]);
    printf("UI_Show_State_req_rcvd  : %d\n", rsp->statistic[i+UI_SHOW_STATE_REQ_RCVD]);
    printf("UI_Get_Sess_Tbl_req_rcvd: %d\n", rsp->statistic[i+UI_GET_SESSTBL_REQ_RCVD]);
    printf("UI_Set_Ver_Tbl_req_rcvd : %d\n", rsp->statistic[i+UI_SET_VER_REQ_RCVD]);
    printf("Chld_Rpt_req_rcvd       : %d\n", rsp->statistic[i+CHILD_RPT_REQ_RCVD]);
    printf("Event_Post_req_rcvd     : %d\n\n", rsp->statistic[i+EVENT_POST_REQ_RCVD]);
	print_stats(rsp->statistic);
    return (NFX_OK);

} /* ndmp_UI_ShowParStats */

/*-----------------------------------------------------------------
 * Name: 			ndmp_UI_ShowStats
 *
 * Description:		Show statistics of a child NDMP server, i.e. 
 *  				on a session or all sessions.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	07/29/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowStats(int argc, char **argv)
{
	int32	sessId;

	if (ui_getSessId(argv[1], &sessId) == NFX_ERR)
		return NFX_ERR;

	get_and_display_stat(NDMP_MID_UI_SHOW_STATISTICS, sessId, 
					(int (*)(void *)) (print_sess_stats));	

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_UI_ShowDetail
 *
 * Description:		Show detailed status of a session or all sessions.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_ShowDetail(int argc, char **argv)
{
	int32	sessId;

	if (ui_getSessId(argv[1], &sessId) == NFX_ERR)
		return NFX_ERR;

	get_and_display_stat(NDMP_MID_UI_SHOW_DETAIL, sessId, 
					(int (*)(void *)) (print_sess_detail));	

	return NFX_OK;
} /* ndmp_UI_ShowDetail */

/*-----------------------------------------------------------------
 * Name: ndmp_UI_DelSess
 *
 * Description:	Kill all or one session.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_DelSess(int argc, char **argv)
{
	ndmp_NFXDelSessReq_t 	c_req;
    char                	rspbuf[sizeof(ndmp_NFXDelSessRsp_t) + 10];
    int32               	sessId;


	if (ui_getSessId(argv[1], &sessId) == NFX_ERR)
		return NFX_ERR;

    c_req.mid = NDMP_MID_UI_DEL_SESS;
    strcpy(c_req.reason, "UI delete session");

	/* Send to parent server first so subsequent sessions will
	 * have the same debug level.
	 */
    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_DelSess: Send message to ssc_ndmp\n"));
	if (snd_req_to_all(FALSE, sizeof(ndmp_NFXDelSessReq_t), &c_req,
					   sizeof(rspbuf), rspbuf, sessId) == NFX_OK)
	{
		printf("Session(s) deleted.\n");
	}

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_UI_SetDbgLevel
 *
 * Description: Set debug level
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_UI_SetDbgLevel(int argc, char **argv)
{

	ndmp_NFXSetDebugReq_t 	req;
    char                	rspbuf[sizeof(ndmp_NFXSetDebugRsp_t) + 10];
    int32               	sessId;


    req.mid = NDMP_MID_UI_SET_DBG_LEVEL;
	if (!isargdigit(argv[1]))
		return -1;
	req.dbg_level = atoi(argv[1]);

	if (ui_getSessId(argv[2], &sessId) == NFX_ERR)
		return NFX_ERR;

	/* Send to parent server first so subsequent sessions will
	 * have the same debug level.
	 */
    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_SetDbgLevel : Send message to ssc_ndmp\n"));
	if (snd_req_to_all(TRUE, sizeof(ndmp_NFXSetDebugReq_t), &req,
					   sizeof(rspbuf), rspbuf, sessId) == NFX_OK)
	{
		printf("Debug level is set to %s for %s\n", argv[1], argv[2]);
	}

	return NFX_OK;

} /* ndmp_UI_SetDbgLevel */
/*
 *-----------------------------------------------------------------
 * Name: ndmp_UI_SetMsgTrace
 *
 * Description: Set trace message ON/OFF
 *
 * Created by:  Nitin Jain
 * Modified  :  Anh Vu
 *
 * Date Created: 08/20/02
 *		Mod    : 10/04/02 - enable this feature to be on/off during
 *				 bkup, i.e. for all or a particular active sesson.
 *-----------------------------------------------------------------
 */
int
ndmp_UI_SetMsgTrace(int argc, char **argv)
{
    ndmp_NFXSMsgTraceReq_t 	req;
    int32 					sessId;
    char 					rspbuf[sizeof(ndmp_NFXSMsgTraceRsp_t)+10];


    req.mid = NDMP_MID_UI_SET_TRACE_ON_OFF;
    req.flag = (strncasecmp(argv[1],"ON", 2) == 0)?1:0;

	if (ui_getSessId(argv[2], &sessId) == NFX_ERR)
        return NFX_ERR;

	/* Send to parent & active children servers so all can 
	 * have the same trace setting.
	 */
    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_SetMsgTrace : Send message to ssc_ndmp\n"));
	if (snd_req_to_all(TRUE, sizeof(ndmp_NFXSMsgTraceReq_t), &req,
                       sizeof(rspbuf), rspbuf, sessId) == NFX_OK)
	{
		printf("Trace message is now set to %s for %s\n", argv[1], argv[2]);
	}
    return NFX_OK;
}
/*
 *-----------------------------------------------------------------
 * Name: ndmp_UI_SetMsgTraceFile
 *
 * Description: Set how may trace log files to retain in system. 
 *
 * Created by:  Anh Vu
 *
 * Date Created: 10/11/02
 *-----------------------------------------------------------------
 */
int
ndmp_UI_SetMsgTraceFile(int argc, char **argv)
{
    ndmp_NFXSMsgTraceReq_t 	req;
    int32 					rc, numdate = 0;
    char 					rspbuf[sizeof(ndmp_NFXSMsgTraceRsp_t)+10];


    req.mid = NDMP_MID_UI_SET_TRACE_FILES;
	if (!isargdigit(argv[1]))
		return -1;
	numdate = atoi(argv[1]); 
	if ((numdate < MIN_LOG_NUMDATE) || (numdate > MAX_LOG_NUMDATE))
	{
		printf("NUM_DATES is out of range, valid range [%d-%d]\n", MIN_LOG_NUMDATE, MAX_LOG_NUMDATE); 
		return -1;
	}
	req.num_files =  numdate;

    /* Query parent NDMP server */
    bzero(rspbuf, sizeof(rspbuf)); 

    NDMP_DBG_PRINTF(D_DBG, ("ndmp_UI_SetMsgTraceFile : Send message to ssc_ndmp\n"));
    /* Send to parent server to get table of sessions (children NDMP servers) */
	rc = sendAgileMsg(-1, &req, sizeof(ndmp_NFXSMsgTraceReq_t), 
			"ssc_ndmp", 0, 0, 0, NULL,  /* to which applic, cpu, slot, plane, timeout */
           NULL, 0);					/* rspbuf, rspbuf_len */

    if (rc < 0) {
        printf(FAIL_STR);
        return NFX_ERR;     
    }
	printf("Number of trace files retain is set to %d\n", req.num_files);
	return NFX_OK;
}

/*
 *-----------------------------------------------------------------
 * Name: ndmp_schedule_clean_up
 *
 * Description: Schedule clean up for snapshots
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/10/02
 *-----------------------------------------------------------------
 */
#define NDMP_NFX_CRONTAB_FILE "/var/cron/tabs/root"
#define NDMP_CRONTAB_COMMAND "/usr/local/agile/etc/nscadmin"
#define NSCD_CMT_1 "# NSCD: snapshot cleanup. Do not modify the line manually. Use nfxsh."
int32 ndmp_schedule_clean_up(int argc, char **argv)
{
    char cronCmdLine[1024];
    char readCronCmdLine[1024];
    int32 idx, found = 0;
    FILE *fp, *fpTmp;
    int32 m = 0, h = 0, d = 0, M = 0, D = 0;
    char mm[10], hh[10], dd[10], MM[10], DD[10];

    if (strcasecmp(argv[0], "cleanup_sch") == 0)
    {
        argc--;
        argv++;
    }

    if (argc%2 != 0)
        return NFX_ERR;

    /* Parse the input parameters */
    bzero(mm, sizeof(char)*10);
    bzero(dd, sizeof(char)*10);
    bzero(dd, sizeof(char)*10);
    bzero(MM, sizeof(char)*10);
    bzero(DD, sizeof(char)*10);
    idx = 0;
    while (idx < argc)
    {
        if (strcmp(argv[idx], "-m") == 0)
        {
            m = 1;
            strcpy(mm, argv[idx+1]);
            // printf("found min. : %s\n", mm);
        }
        else if (strcmp(argv[idx], "-h") == 0)
        {
            h = 1;
            strcpy(hh, argv[idx+1]);
            // printf("found hour : %s\n", hh);
        }
        else if (strcmp(argv[idx], "-d") == 0)
        {
            d = 1;
            strcpy(dd, argv[idx+1]);
            // printf("found date : %s\n", dd);
        }
        else if (strcmp(argv[idx], "-M") == 0)
        {
            M = 1;
            strcpy(MM, argv[idx+1]);
            // printf("found month : %s\n", MM);
        }
        else if (strcmp(argv[idx], "-D") == 0)
        {
            D = 1;
            strcpy(DD, argv[idx+1]);
            // printf("found day : %s\n", DD);
        }
        else
            return NFX_ERR;
        idx += 2;
    }

    bzero(cronCmdLine, sizeof(char)*1024);
    if (m)
    {
        strcpy(cronCmdLine, mm);
        strcat(cronCmdLine, "	");
    }
    else
        strcpy(cronCmdLine, "*	");
    if (h)
    {
        strcat(cronCmdLine, hh);
        strcat(cronCmdLine, "	");
    }
    else
        strcat(cronCmdLine, "*	");
    if (d)
    {
        strcat(cronCmdLine, dd);
        strcat(cronCmdLine, "	");
    }
    else
        strcat(cronCmdLine, "*	");
    if (M)
    {
        strcat(cronCmdLine, MM);
        strcat(cronCmdLine, "	");
    }
    else
        strcat(cronCmdLine, "*	");
    if (D)
    {
        strcat(cronCmdLine, DD);
        strcat(cronCmdLine, "	");
    }
    else
        strcat(cronCmdLine, "*	");

    strcat(cronCmdLine, NDMP_CRONTAB_COMMAND);
    // printf ("New cron entry : %s\n", cronCmdLine);

    /* Delete the entry in crontab first */
    fp = fopen (NDMP_NFX_CRONTAB_FILE, "r");
    if (!fp)
    {
        printf ("Unable to open cron file\n");
        return NFX_ERR;
    }

    fpTmp = fopen("/tmp/.root.cron.tmp", "w");
    if (!fpTmp)
    {
        printf ("Unable to open tmp cron file\n");
        fclose(fp);
        return NFX_ERR;
    }

    bzero(readCronCmdLine, sizeof(char)*1024);
    while (fgets(readCronCmdLine, sizeof(char)*1024, fp) != NULL)
    {
        if (strstr(readCronCmdLine, NDMP_CRONTAB_COMMAND) != NULL)
        {
            found = 1;
            continue;
        }
        fputs(readCronCmdLine, fpTmp);
        bzero(readCronCmdLine, sizeof(char)*1024);
    }
    fclose(fp);
    fclose(fpTmp);
    system("mv /tmp/.root.cron.tmp "NDMP_NFX_CRONTAB_FILE);
    
    fp = fopen (NDMP_NFX_CRONTAB_FILE, "a");
    if (!fp)
    {
        printf ("Unable to open cron file\n");
        return NFX_ERR;
    }

    if (!found)
    {
        /*
         * Enter the comment line
         */
        fprintf(fp, "%s\n", NSCD_CMT_1);
    }

    fprintf(fp, "%s\n", cronCmdLine);
    fclose(fp);

    return NFX_OK;
}

/*
 *-----------------------------------------------------------------
 * Name: ndmp_schedule_clean_up_show
 *
 * Description: Show schedule clean up for snapshots
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/10/02
 *-----------------------------------------------------------------
 */
int32 ndmp_schedule_clean_up_show(int argc, char **argv)
{
    FILE *fp = NULL;
    char *tok = NULL, readCronCmdLine[1024];

    fp = fopen (NDMP_NFX_CRONTAB_FILE, "r");
    if (!fp)
    {
        printf ("Unable to open cron file\n");
        return NFX_ERR;
    }

    bzero(readCronCmdLine, sizeof(char)*1024);
    while (fgets(readCronCmdLine, sizeof(char)*1024, fp) != NULL)
    {
        if ((tok = strstr(readCronCmdLine, NDMP_CRONTAB_COMMAND)) != NULL)
        {
            printf ("Schedule clean up\n");
            printf ("=================\n");
            tok = strtok(readCronCmdLine, "	");
            printf ("Minute : %s\n", tok);
            tok = strtok(NULL, "	");
            printf ("Hour   : %s\n", tok);
            tok = strtok(NULL, "	");
            printf ("Date   : %s\n", tok);
            tok = strtok(NULL, "	");
            printf ("Month  : %s\n", tok);
            tok = strtok(NULL, "	");
            printf ("Day    : %s\n", tok);
            break;
        }
        bzero(readCronCmdLine, sizeof(char)*1024);
    }
    fclose(fp);

    return NFX_OK;
} 

/*
 *-----------------------------------------------------------------
 * Name: ndmp_schedule_clean_up_del
 *
 * Description: Delete schedule
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 10/10/02
 *-----------------------------------------------------------------
 */
int32 ndmp_schedule_clean_up_del(int argc, char **argv)
{
    FILE *fp, *fpTmp;
    char readCronCmdLine[1024];
    /* Delete the entry in crontab first */
    fp = fopen (NDMP_NFX_CRONTAB_FILE, "r");
    if (!fp)
    {
        printf ("Unable to open cron file\n");
        return NFX_ERR;
    }

    fpTmp = fopen("/tmp/.root.cron.tmp", "w");
    if (!fpTmp)
    {
        printf ("Unable to open tmp cron file\n");
        fclose(fp);
        return NFX_ERR;
    }

    bzero(readCronCmdLine, sizeof(char)*1024);
    while (fgets(readCronCmdLine, sizeof(char)*1024, fp) != NULL)
    {
        if (strstr(readCronCmdLine, NDMP_CRONTAB_COMMAND) != NULL)
            continue;
        if (strstr(readCronCmdLine, NSCD_CMT_1) != NULL)
            continue;
        fputs(readCronCmdLine, fpTmp);
        bzero(readCronCmdLine, sizeof(char)*1024);
    }
    fclose(fp);
    fclose(fpTmp);
    system("mv /tmp/.root.cron.tmp "NDMP_NFX_CRONTAB_FILE);

    return NFX_OK;
}

/*
 *-----------------------------------------------------------------
 * Name: ndmp_rem_common_path
 *
 * Description: Remove common path
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 12/01/02
 *-----------------------------------------------------------------
 */
int32 ndmp_restore_path(int argc, char **argv)
{
    ndmp_NFXSetRestoreModeReq_t req;
    char rspbuf[sizeof(ndmp_NFXSetRestoreModeReq_t)+1];

    req.mid = NDMP_MID_UI_SET_RESTORE_MODE;

    req.flag = 0;
    if (!strcasecmp(argv[1], "off"))
        req.flag = 1;

    if (snd_req_to_all(TRUE, sizeof(ndmp_NFXSetRestoreModeReq_t), &req,
                       sizeof(rspbuf), rspbuf, -1) == NFX_OK)
        printf("Restore mode is set %s.\n", (req.flag)?"to remove common path"
                                                   :"to restore complete path");

    return NFX_OK;
}
