/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-ds.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-ds.c
 *
 * Description	: Contains NetBIOS datagram service functions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 12/05/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ds.h"
#include "nb-util.h"

/*-----------------------------------------------------------------------------
 * Name			: nb_dsCreateDgmPkt
 *
 * Description	: Create and initialize a datagram service packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/07/2001
 *-----------------------------------------------------------------------------
 */
nb_dsPacket_t *	//Ptr to datagram service packet or NULL if failed
nb_dsCreateDgmPkt(nb_srvCB_t	*nbSrvCbP)		// Prt to NetBIOS server CB
{
	nb_dsPacket_t	*dsPktP;
	nb_dsHeader_t	*headerP;
	uint16			nodeType;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_dsCreateDgmPkt()"));

	/* Alloc memory for Datagram Service packet */
	NB_RAM_ALLOC_ZERO(dsPktP, nb_dsPacket_t, sizeof(nb_dsPacket_t));
	if (dsPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_dsCreateDgmPkt: "
					"Failed to alloc memory for datagram service packet"));
		return (NULL);
	}

	if (nbSrvCbP->currNodeType == NB_NS_NBFLAG_BNODE)
		nodeType = NB_DS_NODE_TYPE_B;
	else
		nodeType = NB_DS_NODE_TYPE_H;

	/* Set header */
	headerP = &dsPktP->header;

	headerP->flags.nodeType = nodeType;
    headerP->flags.first = TRUE;
	headerP->flags.more = FALSE;

	headerP->dgmID = nb_genXid();
	headerP->srcIP = nbSrvCbP->regCb.cifsSrvIpAddr;
	headerP->srcPort = NB_DS_UDP_PORT;
	headerP->offset = 0;
	headerP->dgmLen = NB_S_COMPRESSED_NAME * 2;
						// Init dgmLen to length of 2 compressed NetBIOS names

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_dsCreateDgmPkt()"));
	return(dsPktP);
} /* nb_dsCreateDgmPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_dsCreateAnncePkt
 *
 * Description	: Create and initialize a server announcement packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/07/2001
 *-----------------------------------------------------------------------------
 */
nb_dsPacket_t *	// Prt to datagram service packet
nb_dsCreateAnncePkt(nb_srvCB_t	*nbSrvCbP,		// Prt to NetBIOS server CB
					char		*destNameStr,	// Destination name
					uint32		userDataLen)	// Number of bytes of user data
{
	nb_dsPacket_t	*anncePktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_dsCreateAnncePkt()"));

	anncePktP = nb_dsCreateDgmPkt(nbSrvCbP);
	if (anncePktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_dsCreateAnncePkt: "
					"Failed to create datagram service packet"));
		return (NULL); 
	}

	/* Now the datagram length is length of user data plus
	 * length of 2 compressed NetBIOS names
	 */
	anncePktP->header.dgmLen += userDataLen;

	anncePktP->srcName = nbSrvCbP->regCb.cifsSrvName;
	anncePktP->srcName.type = NB_NS_NAMETYPE_WS;

	if (strlen(destNameStr) > NB_S_NAME)
	{
		/* Name length is greater than NB_S_NAME, truncated it */
		strcpy(anncePktP->destName.name, destNameStr);
		anncePktP->destName.name[NB_S_NAME] = '\0';
	}
	else
		strcpy(anncePktP->destName.name, destNameStr);
	anncePktP->destName.type = NB_NS_NAMETYPE_MASTBROWSER;

	anncePktP->dataLen = userDataLen;
	anncePktP->userData = NULL;		// We don't care in this case

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_dsCreateAnncePkt()"));
	return (anncePktP);
} /* nb_dsCreateAnncePkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_dsUpdateXid
 *
 * Description	: Update DGM_ID in a datagram packet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/20/2001
 *-----------------------------------------------------------------------------
 */
void
nb_dsUpdateXid(nb_dsPacket_t	*dsPktP,
			   uchar8			*dsXmtBufP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_dsUpdateXid()"));

	dsPktP->header.dgmID = nb_genXid();

	PUTBE2(dsXmtBufP, 2, dsPktP->header.dgmID);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_dsUpdateXid()"));
} /* nb_dsUpdateXid() */

/*-----------------------------------------------------------------------------
 * Name			: nb_dsUpdateNodeType
 *
 * Description	: Update node type in a datagram packet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/20/2001
 *-----------------------------------------------------------------------------
 */
void
nb_dsUpdateNodeType(nb_dsPacket_t	*dsPktP,
					uchar8			*dsXmtBufP,
					uint16			nodeType)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_dsUpdateNodeType()"));

	dsPktP->header.flags.nodeType = nodeType;

	/* Clear SNT bits in datagram packet header */
	*(dsXmtBufP + 1) &= 0xf3;
	
	*(dsXmtBufP + 1) |= ((nodeType &0x03) << 2);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_dsUpdateNodeType()"));
} /* nb_dsUpdateNodeType() */

/*-----------------------------------------------------------------------------
 * Name			: nb_dgmXmtPktReq
 *
 * Description	: Process a NetBIOS datagram send message request. 
 *				  Called by CIFS. Mainly used to send server announcement
 *				  to the master browser.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/06/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_dsXmtPktReq(nb_srvCB_t	*nbSrvCbP,		// Prt to NetBIOS server CB
				nb_name_t	destName,		// Destination name
				pkt_desc_t	*msgBufferP,	// Ptr to msg to be sent buffer
				uint32		smbMsgLen)		// SMB message length
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_dgmXmtPktReq()"));
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_dgmXmtPktReq()"));
	return (NFX_OK);
} /* nb_dgmXmtPktReq() */

