/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-tpl.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-tpl.c
 *
 * Description	: Contains  NetBIOS initialization and utility functions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/12/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ss.h"
#include "nb-ds.h"
#include "nb-tpl.h"
#include "nb-util.h"
#include "nb-msg.h"


/*-----------------------------------------------------------------------------
 *  Name		: nb_tplNsAddConnInd
 *
 * Description	: Add connection indication from the transport layer to
 *				  adds a new name service connection. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_tplNsAddConnInd(tpl_bindCb_t	*bindP,		// Ptr to Bind CB
				   void			*connHdlP,	// Ptr to newly added connection
				   void			**appHdlPP,	// Ptr to correspond app. handle
				   uint32		raddr,		// Remote IP address
				   ushort16		rport)		// Remote port
{
	nb_srvCB_t		*srvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplNsAddConnInd()"));

	srvCbP = (nb_srvCB_t *)bindP->usrInfo;

	/* Check if server control block is still valid */
	if (nb_isValidServer(srvCbP) == NFX_ERR)
	{
		/* Server is no longer in the server list */
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplNsAddConnInd: Server is no longer valid"));
		return (NFX_ERR);
	}

	if (srvCbP->srvNameState != NB_NAME_STATE_REGISTERED ||
		srvCbP->domNameState != NB_NAME_STATE_REGISTERED)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplNsAddConnInd: "
					"Server %s is not registered",
					srvCbP->regCb.cifsSrvName.name));
		return (NFX_ERR);
	}

	/* TODO - something more here? */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplNsAddConnInd()"));

	return (NFX_OK);
} /* nb_tplNsAddConnInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplNsDelConnInd
 *
 * Description	: Delete connection indication from the transport layer 
 * 				  to delete an existing name service connection.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_tplNsDelConnInd (void *appHdlP, void *connHdlP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplNsDelConnInd()"));
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplNsDelConnInd()"));

	return (NFX_OK);
} /* nb_tplNsDelConnInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplNsRcvPktInd
 *
 * Description	: Receive packet indication from the transport layer 
 * 				  for processing a name service packet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_tplNsRcvPktInd(void *appHdlP, pkt_desc_t *descP)
{
	int32				rc;
	nb_incomingMsgCB_t	inMsgCb;
	nb_srvCB_t			*srvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplNsRcvPktInd()"));

	if (appHdlP == NULL || descP == NULL || descP->pHdr == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplNsRcvPktInd: Invalid pointer"));
		return (TPL_FREE_PKT);
	}

	srvCbP = (nb_srvCB_t *)appHdlP;

	/* Check if server control block is still valid */
	if (nb_isValidServer(srvCbP) == NFX_ERR)
	{
		/* Server is no longer in the server list */
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplNsRcvPktInd: Server is no longer valid"));
		return (NFX_ERR);
	}

	inMsgCb.nbSrvCbP = (nb_srvCB_t *)appHdlP;
	inMsgCb.raddr = descP->raddr;
	inMsgCb.rport = descP->rport;

	NB_PRINT(NB_DBGLV_DBG, ("nb_tplNsRcvPktInd: "
					"Remote IP address = 0x%x, port = %d",
					descP->raddr, descP->rport));

	rc = nb_nsRcvMsg(&inMsgCb, descP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplNsRcvPktInd()"));

	return (rc);
} /* nb_tplNsRcvPktInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplNsXmtCmpInd
 *
 * Description	: Transmit packet complete indication from the transport layer
 * 				  for name service packet transmission.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_tplNsXmtCmpInd (void *appHdlP, pkt_desc_t *descP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplNsXmtCmpInd()"));
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplNsXmtCmpInd()"));

	return (NFX_OK);
} /* nb_tplNsXmtCmpInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplSsAddConnInd
 *
 * Description	: Add connection indication from the transport layer to
 *				  adds a new session service connection. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_tplSsAddConnInd(tpl_bindCb_t	*bindP,		// Ptr to Bind CB
				   void			*connHdlP,	// Ptr to newly added connection
				   void			**appHdlPP,	// Ptr to correspond app. handle
				   uint32		raddr,		// Remote IP address
				   ushort16		rport)		// Remote port
{
	nb_srvCB_t		*srvCbP;
	nb_ssnCB_t		*ssnCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplSsAddConnInd()"));

	srvCbP = (nb_srvCB_t *)bindP->usrInfo;

	/* Check if server control block is still valid */
	if (nb_isValidServer(srvCbP) == NFX_ERR)
	{
		/* Server is no longer in the server list */
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplSsAddConnInd: Server is no longer valid"));
		return (NFX_ERR);
	}

	if (srvCbP->srvNameState != NB_NAME_STATE_REGISTERED ||
		srvCbP->domNameState != NB_NAME_STATE_REGISTERED)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplSsAddConnInd: "
					"Server %s is not registered",
					srvCbP->regCb.cifsSrvName.name));
		return (NFX_ERR);
	}

	/*----------------------------------------
	 * Check if session control block exists.
	 * If not, create a session control block.
	 *----------------------------------------
	 */
	ssnCbP = nb_getSsnCb(srvCbP, raddr, rport);
	if (ssnCbP == NULL)
	{
		/* Allocate a session CB and
		 * add it to session CB list
		 */
		ssnCbP = nb_createSsnCb(srvCbP);
		if (ssnCbP == NULL)
			return (NFX_ERR);
	}

	ssnCbP->sessionService.tplBindCbP = bindP;
	ssnCbP->sessionService.tplConnCbP = connHdlP;
	// ssnCbP->sessionService.bsdSockP = svrCbP->sessionService.bsdSockP;
	ssnCbP->sessionService.mtu = 50;	//TODO. Set a constant value for now.
	ssnCbP->rIpAddr = raddr;
	ssnCbP->rPort = rport;


	/*----------------------------------
	 * Fill in handle upon return to TPL
	 *----------------------------------
	 */
	*appHdlPP = ssnCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplSsAddConnInd()"));

	return (NFX_OK);
} /* nb_tplSsAddConnInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplSsDelConnInd
 *
 * Description	: Delete connection indication from the transport layer 
 * 				  to delete an existing session service connection.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32 			// NFX_OK or NFX_ERR
nb_tplSsDelConnInd (void *appHdlP, void *connHdlP)
{
	nb_ssnCB_t	*ssnCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplSsDelConnInd()"));

	ssnCbP = (nb_ssnCB_t *)appHdlP;
	if (ssnCbP == NULL)
		return NFX_ERR;

	if (ssnCbP->state == NB_SS_ESTABLISHED)
	{
		/* Send indication to CIFS */
		ssnCbP->nbSrvCbP->regCb.ul_ssnEndInd(ssnCbP->cifsSsnCbP);
	}

	/* Delete session control block */
	nb_deleteSsnCb(ssnCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplSsDelConnInd()"));

	return (NFX_OK);
} /* nb_tplSsDelConnInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplSsRcvPktInd
 *
 * Description	: Receive packet indication from the transport layer 
 * 				  for processing a session service packet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_tplSsRcvPktInd(void *appHdlP, pkt_desc_t *descP)
{
	int32	rc;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplSsRcvPktInd()"));

	if (appHdlP == NULL || descP == NULL || descP->pHdr == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplSsRcvPktInd: Invalid pointer"));
		return (TPL_FREE_PKT);

	}

	rc = nb_ssRcvMsg((nb_ssnCB_t *)appHdlP, descP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplSsRcvPktInd()"));

	return (rc);
} /* nb_tplSsRcvPktInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplSsXmtCmpInd
 *
 * Description	: Transmit packet complete indication from the transport layer
 * 				  for session service packet transmission.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_tplSsXmtCmpInd (void *appHdlP, pkt_desc_t *descP)
{
	int32		rc = TPL_FREE_PKT;
	nb_ssnCB_t	*ssnCbP;
	nb_srvCB_t	*nbSrvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplSsXmtCmpInd()"));

	ssnCbP = (nb_ssnCB_t *)appHdlP;
	switch(ssnCbP->state)
	{
		case  NB_SS_INIT:
			/* We sent out a negative response */

			/* Close TCP connection */
			tpl_delConnReq(ssnCbP->sessionService.tplConnCbP);
			break;

		case NB_SS_IN_ESTABLISHING:
			/* We sent out a positive response */
			ssnCbP->state = NB_SS_ESTABLISHED;
			break;

		case NB_SS_OUT_ESTABLISHING:
			/* We sent out a session establishment request */
			break;

		case NB_SS_ESTABLISHED:
			/* We sent a message with user data */
			(ssnCbP->nbSrvCbP)->regCb.ul_ssnXmtCmpInd(ssnCbP->cifsSsnCbP,
													  descP->appInfo,
													  descP);
			rc = TPL_HOLD_PKT;
			break;

		default:
			break;
	}
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplSsXmtCmpInd()"));

	return (rc);
} /* nb_tplSsXmtCmpInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplDsAddConnInd
 *
 * Description	: Add connection indication from the transport layer to
 *				  adds a new datagram service connection. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_tplDsAddConnInd(tpl_bindCb_t	*bindP,		// Ptr to Bind CB
				   void			*connHdlP,	// Ptr to newly added connection
				   void			**appHdlPP,	// Ptr to correspond app. handle
				   uint32		raddr,		// Remote IP address
				   ushort16		rport)		// Remote port
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplDsAddConnInd()"));

	/* At present, we don't process any incoming datagram */
	/* TODO - Question: When to disconnect the connection? */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplDsAddConnInd()"));

	return (NFX_OK);
} /* nb_tplDsAddConnInd() */

#if defined(NFX_SMP)
/*-----------------------------------------------------------------------------
 *  Name		: nb_tplDsBindReqAck
 *
 * Description	: bind request ack.
 *
 * Created by	: Su Wang
 *
 * Date created	: 12/04/2002
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_tplDsBindReqAck(void	*pHandle, uint32 rc, tpl_bindCb_t pBind)
{
	if (rc != NFX_OK)
		panic("%s: tpl bind failed\n", __FUNCTION__);
	
	((nb_serviceCB_t*)pHandle)->tplBindCbP = pBind;
	
	return NFX_OK;
}

#endif /* NFX_SMP */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplDsDelConnInd
 *
 * Description	: Delete connection indication from the transport layer 
 * 				  to delete an existing datagram service connection.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_tplDsDelConnInd (void *appHdlP, void *connHdlP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplDsDelConnInd()"));
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplDsDelConnInd()"));

	return (NFX_OK);
} /* nb_tplDsDelConnInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplDsRcvPktInd
 *
 * Description	: Receive packet indication from the transport layer 
 * 				  for processing a datagram service packet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_tplDsRcvPktInd(void *appHdlP, pkt_desc_t *descP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplDsRcvPktInd()"));

	/* At present, we don't process any incoming datagram */
	/* Discard the packet */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplDsRcvPktInd()"));

	return (NFX_OK);
} /* nb_tplDsRcvPktInd() */

/*-----------------------------------------------------------------------------
 *  Name		: nb_tplDsXmtCmpInd
 *
 * Description	: Transmit packet complete indication from the transport layer
 * 				  for datagram service packet transmission.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_tplDsXmtCmpInd (void *appHdlP, pkt_desc_t *descP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplDsXmtCmpInd()"));
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplDsXmtCmpInd()"));

	return (NFX_OK);
} /* nb_tplDsXmtCmpInd() */


/*-----------------------------------------------------------------------------
 *  Name		: nb_tplBindReq
 *
 * Description	: Bind NetBIOS name service, session service, and 
 * 				  datagram service with TPL, in server or client mode.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or TPL error
nb_tplBindReq(nb_srvCB_t	*srvCbP,	// Prt to server control block
			  uint32		raddr,		// Remote IP addr if in client mode
			  uint16		rport,		// Remote port if in client mode
			  uint16		service,	// name, session, or datagram service
			  uint16		protocol,	// Protocol UDP/TCP
			  int32			bindMode)	// TPL_BIND_SERVER or TPL_BIND_CLIENT
{
	int32			rc;
	uint16			lport;
	tpl_bindCb_t	tplBindReqCb;
	tpl_bindCb_t	*tplBindRspCbP;
	nb_serviceCB_t	*serviceCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_tplBindReq()"));

	memset(&tplBindReqCb, 0, sizeof(tpl_bindCb_t));


	switch (service)
	{
		case NB_NS:
		{
			lport = NB_NS_UDP_PORT;

			tplBindReqCb.ul_addConnInd = nb_tplNsAddConnInd;
			tplBindReqCb.ul_delConnInd = nb_tplNsDelConnInd;
			tplBindReqCb.ul_rcvPktInd  = nb_tplNsRcvPktInd;
			tplBindReqCb.ul_xmtCmpInd  = nb_tplNsXmtCmpInd;

			serviceCbP = &srvCbP->nameService;
		}
			break;
		case NB_SS:
		{
			lport = NB_SS_TCP_PORT;

			tplBindReqCb.ul_addConnInd = nb_tplSsAddConnInd;
			tplBindReqCb.ul_delConnInd = nb_tplSsDelConnInd;
			tplBindReqCb.ul_rcvPktInd  = nb_tplSsRcvPktInd;
			tplBindReqCb.ul_xmtCmpInd  = nb_tplSsXmtCmpInd;

			serviceCbP = &srvCbP->sessionService;
		}
			break;
		case NB_DS:
		{
			lport = NB_DS_UDP_PORT;

			tplBindReqCb.ul_addConnInd = nb_tplDsAddConnInd;
			tplBindReqCb.ul_delConnInd = nb_tplDsDelConnInd;
			tplBindReqCb.ul_rcvPktInd  = nb_tplDsRcvPktInd;
			tplBindReqCb.ul_xmtCmpInd  = nb_tplDsXmtCmpInd;

			serviceCbP = &srvCbP->datagramService;
		}
			break;
		default:
			return (NFX_ERR);
	}
	tplBindReqCb.laddr = srvCbP->regCb.cifsSrvIpAddr;
	tplBindReqCb.raddr = raddr;
	tplBindReqCb.lport = lport;
	tplBindReqCb.rport = rport;
	tplBindReqCb.tplProt = protocol;
	tplBindReqCb.usrInfo = (void *)srvCbP;

#if defined(NFX_SMP)
	tplBindReqCb.ul_bindReqAck = nb_tplDsBindReqAck;
	rc = tpl_bindReq(&tplBindReqCb, serviceCbP, bindMode);
	
#else
	rc = tpl_bindReq(&tplBindReqCb, &tplBindRspCbP, bindMode);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_tplBindReq: Binding failure (rc = %d)", rc));
		return (rc);
	}

	serviceCbP->tplBindCbP = tplBindRspCbP;
	serviceCbP->bsdSockP = tplBindRspCbP->pBsdSock;
#endif /* NFX_SMP */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_tplBindReq()"));

	return (rc);
} /* nb_tplBindReq() */


/*-----------------------------------------------------------------------------
 * Name			: nb_xmtPktSP
 *
 * Description	: Transmit a packet to be sent via slow path.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
void
nb_xmtPktSP(void		*sock,		// Socket to send the message
			pkt_desc_t	*xmtPktP,	// Prt to packet descriptor 
									// containing message to be sent
			uint32		raddr,		// IP address in network byte order
			uint16		rport)		// Port in network byte order
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_xmtPktSP()"));

	if (xmtPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_xmtPktSP: NULL pointer to packet descriptor"));
		return;
	}

	NB_PRINT(NB_DBGLV_DBG, ("nb_xmtPktSP: "
				"Remote IP address = 0x%x, port = %d", raddr, rport));

#ifndef NBTEST
	bsd_SendSock(sock,
				 raddr,
				 rport,
				 xmtPktP->pHdr,
				 BSD_BUF_FLAT,
				 xmtPktP->hdrSize,
				 0);
#else
	printf("nb_xmtPktSP: Transmitting packet...\n");
	nb_displayEncodedMsg(xmtPktP->pHdr, xmtPktP->hdrSize);
#endif /* NBTEST */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_xmtPktSP()"));
	return;
} /* nb_xmtPktSP() */
