/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb.c
 *
 * Description	: Contains NetBIOS initialization and utility functions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ds.h"
#include "nb-tpl.h"
#include "nb-util.h"
#include "nb-test.h"

int32 nb_annceTimeout(int32 timerHdl, void* param1,
					  void* param2, void* param3, void* param4);

/* Global variables */
nb_globalCfg_t	*nb_globalCfgP;
uchar8			nb_debugLevel = NB_DBGLV_DBG2;
//uchar8			nb_debugLevel = NB_DBGLV_ERR;
int32			nb_cifsAppId;

uint32 nb_elog_map[6] = {
	0,
	LOG_ERR,						// 1, NB_DBGLV_ERR
	LOG_WARNING,					// 2, NB_DBGLV_WARN
	LOG_INFO,						// 3, NB_DBGLV_INFO
	LOG_INFO /*LOG_DEBUG*/,			// 4, NB_DBGLV_DBG
	LOG_INFO /*LOG_DEBUG*/			// 5, NB_DBGLV_DBG2
};

/*-----------------------------------------------------------------------------
 *  Name		: nb_init
 *
 * Description	: NetBIOS initialization function. Called from CIFS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/07/2001
 *----------------------------------------------------------------------------- */
void
nb_init(uint32	maxSrvCnt,		// Max server count
		uint32	maxSsnCnt,		// Max session count
		int32	cifsAppId)		// CIFS application ID registered with EEE
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_init()"));

	NB_PRINT(NB_DBGLV_DBG, ("nb_init: nfxMyCPUId = %d", nfxMyCPUId));

	nb_cifsAppId = cifsAppId;

	/* Initialize server and session control block data */
	NB_PRINT(NB_DBGLV_INFO, ("nb_init: initializing..."));
	NB_RAM_ALLOC_ZERO(nb_globalCfgP, nb_globalCfg_t, (sizeof(nb_globalCfg_t)));
	if (nb_globalCfgP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_init: "
					"Failed to alloc memory for NB global config."));
		return;
	}

	/* Init server CB list */
	nfp_objInit(&nb_globalCfgP->srvCbList,
				"nb_server",
				sizeof(nb_srvCB_t),
				maxSrvCnt);

	/* Init session CB list */
	nfp_objInit(&nb_globalCfgP->ssnCbList,
				"nb_session",
				sizeof(nb_ssnCB_t),
				maxSsnCnt);

	/* Init request CB list */
	nfp_objInit(&nb_globalCfgP->reqCbList,
				"nb_request",
				sizeof(nb_reqCB_t),
				NB_N_MAX_REQUEST);

	/* Init query CB list */
	nfp_objInit(&nb_globalCfgP->queryCbList,
				"nb_query",
				sizeof(nb_queryCB_t),
				NB_N_MAX_QUERY);

	/* Init incoming msg CB list */
	nfp_objInit(&nb_globalCfgP->inMsgCbList,
				"nb_inMsg",
				sizeof(nb_incomingMsgCB_t),
				NB_N_MAX_INMSG);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_init()"));

	return;
} /* nb_init() */

/*-----------------------------------------------------------------------------
 * Name			: nb_registerReq
 *
 * Description	: A CIFS server registers with NetBIOS module. Called by CIFS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/07/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_srvRegisterReq(nb_srvRegCB_t	*regCbP,	// Register request CB prt
				  nb_srvCB_t	**nbSrvCbP)	// NetBIOS srv CB ptr upon return
{
	int32			rc;
	uint16			reqType;
	nb_srvCB_t		*srvCbP = NULL;
	nb_reqCB_t		*reqCbP = NULL;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_srvRegisterReq()"));

	if (regCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: "
					"NULL pointer to register request control block"));
		return (NFX_ERR);
	}

	/*-------------------------------------------------- 
	 * Create a server control block for the CIFS server
	 *--------------------------------------------------
	 */
	srvCbP = nb_allocSrvCb();
	if (srvCbP == NULL)
	{
		/* Internal error */
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: : "
					"Couldn't allocate a server control block."));
		return (NFX_ERR);
	}
	memset(srvCbP, 0, sizeof(nb_srvCB_t));

	/*----------------------------------------
	 * Add the new server CB to server CB list
	 *----------------------------------------
	 */
	nb_addServerCb(&nb_globalCfgP->servers, srvCbP);

	/*------------------------------
	 * Init the server control block
	 *------------------------------
	 */
	memcpy(&srvCbP->regCb, regCbP, sizeof(nb_srvRegCB_t));
	if (srvCbP->regCb.winsSrvIpAddr)
	{
		/* WINS server is configured, then the node is h-node. */
		srvCbP->currNodeType = NB_NS_NBFLAG_HNODE;
	}
	else
	{
		/* WINS server is not configured, then the node is b-node */
		srvCbP->currNodeType = NB_NS_NBFLAG_BNODE;
	}
	srvCbP->srvNameTTL = NB_NS_TTL;
	srvCbP->srvNameState = NB_NAME_STATE_UNREGISTERED;
	srvCbP->domNameState = NB_NAME_STATE_UNREGISTERED;

	/*-------------------------------------
	 * Bind NetBIOS with TPL in server mode
	 *-------------------------------------
	 */
#ifndef NBTEST
	/* Bind name service */
	rc = nb_tplBindReq(srvCbP,				// Server CB
					   0, 0,
					   NB_NS,
					   TPL_PROT_UDP,
					   TPL_BIND_SERVER);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: "
					"Failed in binding name service"));
		goto nb_registerReqErr;
	}

	/* Bind session service */
	rc = nb_tplBindReq(srvCbP,				// Server CB
					   0, 0,
					   NB_SS,
					   TPL_PROT_TCP,
					   TPL_BIND_SERVER);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: "
					"Failed in binding session service"));
		goto nb_registerReqErr;
	}

	/* Bind datagram service */
	rc = nb_tplBindReq(srvCbP,				// Server CB
					   0, 0,
					   NB_DS,
					   TPL_PROT_UDP,
					   TPL_BIND_SERVER);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: : "
					"Failed in binding datagram service"));
		goto nb_registerReqErr;
	}
#endif /* NBTEST */

	/*------------------------------------------
	 * Register a CIFS server name.
	 * Assuming the cifs module is responsible
	 * for making sure the name length is <= 15.
	 *------------------------------------------
	 */
	reqType = NB_REQ_TYPE_NAME_REG;
	if (srvCbP->currNodeType != NB_NS_NBFLAG_BNODE)
		reqType |= NB_REQ_TO_WINS;

	srvCbP->regCb.cifsSrvName.type = NB_NS_NAMETYPE_SERVER;
	reqCbP = nb_nsRequest(srvCbP,
						  &srvCbP->regCb.cifsSrvName,
						  reqType,
						  FALSE);		// Unique name
	if (reqCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: "
					"Failed in sending NFX server name (%s) registration "
					"request", srvCbP->regCb.cifsSrvName.name));
		rc = NFX_ERR;
		goto nb_registerReqErr;
	}

	/*-----------------------------------
	 * Register a CIFS server domain name
	 *-----------------------------------
	 */
#ifdef NOTUSED
	srvCbP->regCb.cifsDomName.type = NB_NS_NAMETYPE_WS;
	reqCbP = nb_nsRequest(srvCbP,
						  &srvCbP->regCb.cifsDomName,
						  reqType,
						  TRUE);		// Group name
	if (reqCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_registerReq: "
					"Failed in sending domain name (%s) registration request",
					srvCbP->regCb.cifsDomName.name));
		rc = NFX_ERR;
		goto nb_registerReqErr;
	}
#else
	/* Set domain name state to be NB_NAME_STATE_REGISTERED for now */
	srvCbP->domNameState = NB_NAME_STATE_REGISTERED;
#endif /* NOTUSED */

	/*----------------------------------------
	 * Pass the pointer to server CB to caller
	 *----------------------------------------
	 */
	*nbSrvCbP = srvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_srvRegisterReq()"));

#ifdef NBTEST
	nbtest_setSrvCb();
#endif /* NBTEST */
	return (NFX_OK);

nb_registerReqErr:
	/* Remove server CB from server CB list and free it */
	nb_removeServerCb(&nb_globalCfgP->servers, srvCbP);
	nb_freeSrvCb(srvCbP);
	*nbSrvCbP = NULL;

	return (NFX_ERR);
} /* nb_srvRegisterReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_srvReleaseReq
 *
 * Description	: A CIFS server unregister with NetBIOS module.
 * 				  It releases name and disconnects all running sessions.
 *				  The function is called three times. It is first
 *				  called by CIFS module. Then, it is called by NetBIOS
 *				  name service function after name release request has
 *				  been sent out a number of times, or after all the
 *				  active sessions are disconnected. When called by 
 *				  NetBIOS name service, the callback function parameter 
 *				  of the function is NULL.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/07/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR 
nb_srvReleaseReq(nb_srvCB_t	*nbSrvCbP,			// NetBIOS srv CB ptr
				 int32 (*ul_releaseCfmInd) ()) 	// Callback function
{
	static boolean	firstTime = TRUE;	// Function is called the first time or not
	static int32	(*callback)();
	nb_reqCB_t		*reqCbP = NULL;
	nb_ssnCB_t		*ssnCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_srvReleaseReq()"));

	if (nbSrvCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_srvReleaseReq: "
					"NULL pointer to nb server control block"));
		return (NFX_ERR);
	}

	if (firstTime == TRUE)
	{
		uint16	reqType;

		if (ul_releaseCfmInd == NULL)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_srvReleaseReq: "
						"Invalid callback function"));
			return (NFX_ERR);
		}

		callback = ul_releaseCfmInd;

		/* Create a name release request to release CIFS server name,
		 * and send out the request.
		 */
		reqType = NB_REQ_TYPE_NAME_RELEASE;
		if (nbSrvCbP->currNodeType != NB_NS_NBFLAG_BNODE)
			reqType |= NB_REQ_TO_WINS;

		reqCbP = nb_nsRequest(nbSrvCbP,
							  &nbSrvCbP->regCb.cifsSrvName,
							  reqType,
							  FALSE);		// Unique name
		if (reqCbP == NULL)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_srvReleaseReq: "
						"Failed in sending name release request"));
			return(NFX_ERR);
		}
		
		/* Set domain name state to be NB_NAME_STATE_REGISTERED for now */
		nbSrvCbP->domNameState = NB_NAME_STATE_UNREGISTERED;

		/* Stop timers for refresh/registration requests */
		reqCbP = nbSrvCbP->reqCbP;
		while(reqCbP)
		{
			if (((reqCbP->reqType & ~NB_REQ_TO_WINS) == NB_REQ_TYPE_NAME_REG) ||
				((reqCbP->reqType & ~NB_REQ_TO_WINS) == NB_REQ_TYPE_NAME_REFRESH))
			{
				nb_timerStop(&reqCbP->reqTimerHdl);
			}
			reqCbP = reqCbP->nextP;
		}

		/* Disconnect and cleanup all the active sessions */
		ssnCbP = nbSrvCbP->ssnCbP;
		while(ssnCbP)
		{
			nb_ssnEndReq(ssnCbP);
			ssnCbP = ssnCbP->nextP;
		}

		firstTime = FALSE;
		return(NFX_OK);
	}

	/*---------------------------------------------
	 * After name has been released, or all session
	 * connections have been disconnected.
	 *---------------------------------------------
	 */
	if (nbSrvCbP->srvNameState == NB_NAME_STATE_UNREGISTERED &&
		nbSrvCbP->domNameState == NB_NAME_STATE_UNREGISTERED &&
		nbSrvCbP->ssnCbP == NULL)
	{
		/* Stop and destroy name list timer */
		nb_timerStop(&nbSrvCbP->nameListTimerHdl);
		nb_timerDestroy(&nbSrvCbP->nameListTimerHdl);

		/* Cleanup announcement CB, stop and destroy timer */
		nb_timerStop(&nbSrvCbP->annceCb.annceTimerHdl);
		nb_timerDestroy(&nbSrvCbP->annceCb.annceTimerHdl);

		nb_freeOutgoingPkt(NULL, nbSrvCbP->annceCb.xmtPktP);

		/* Cleanup all request CBs, stop and destroy timers */
		while(nbSrvCbP->reqCbP)
			nb_deleteReqCb(nbSrvCbP->reqCbP);

		/* Cleanup all query CBs */
		while(nbSrvCbP->queryCbP)
			nb_deleteQueryCb(nbSrvCbP->queryCbP);

#ifndef NBTEST
		/* Unbind with tpl */
		tpl_unbindReq(nbSrvCbP->nameService.tplBindCbP);
		tpl_unbindReq(nbSrvCbP->sessionService.tplBindCbP);
		tpl_unbindReq(nbSrvCbP->datagramService.tplBindCbP);
#endif /* NBTEST */

		/* Remove server CB from server CB list and free it */
		nb_removeServerCb(&nb_globalCfgP->servers, nbSrvCbP);
		nb_freeSrvCb(nbSrvCbP);

		(callback)();
	}
	else
	{
		/* Name release or session disconnect is not finished yet */
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_srvReleaseReq()"));
	return (NFX_OK);
} /* nb_srvReleaseReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_enableBrowserAnnouncement
 *
 * Description	: Called by CIFS to send server announcement
 *                to the master browser.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/07/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_enableBrowserAnnouncement(
					nb_srvCB_t	*nbSrvCbP,		// Prt to NetBIOS server CB
					char		*destName,		// Destination name
					pkt_desc_t	*msgBufferP,	// Ptr to msg to be sent buffer
					uint32		smbMsgLen) 		// SMB message length
{
	int32				rc;
	nb_announceCB_t		*annceCbP;
	nb_dsPacket_t		*anncePktP;
	pkt_desc_t			*xmtPktDescP;
	uint32				raddr;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_enableBrowserAnnouncement()"));

	if (nbSrvCbP == NULL || msgBufferP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_enableBrowserAnnouncement: "
					"Invalid pointer"));
		return (NFX_ERR);	
	}

	if (nbSrvCbP->srvNameState != NB_NAME_STATE_REGISTERED ||
		nbSrvCbP->domNameState != NB_NAME_STATE_REGISTERED)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_enableBrowserAnnouncement: "
					"Server %s is not registered", 
					nbSrvCbP->regCb.cifsSrvName.name));
		return (NFX_ERR);
	}

	/*-----------------------
	 * Initialize announce CB
	 *-----------------------
	 */
	annceCbP = &nbSrvCbP->annceCb;
	annceCbP->interval = NB_MIN_HOST_ANNCE_INTERVAL;
	annceCbP->xmtPktP = msgBufferP;

	/*---------------------------------------------------------
	 * Create a datagram data structure for server announcement
	 *---------------------------------------------------------
	 */
	anncePktP = nb_dsCreateAnncePkt(nbSrvCbP, destName, smbMsgLen);
	if (anncePktP == NULL)
		return (NFX_ERR);
	annceCbP->anncePktP = anncePktP;

	/*----------------------------------
	 * Prepare the packet to be sent out
	 *----------------------------------
	 */
	msgBufferP->hdrSize = 14 + anncePktP->header.dgmLen;
										// Length of datagram header is 14.
	xmtPktDescP = nb_createXmtPktSP((void *)anncePktP, NB_DS, msgBufferP);
										// xmtPktDescP = msgBufferP in this case

	/*------------------------------
	 * Broadcast the datagram packet
	 *------------------------------
	 */
	nb_xmtPktSP(nbSrvCbP->datagramService.bsdSockP,
				annceCbP->xmtPktP,
				nbSrvCbP->regCb.cifsSrvIpAddr | ~(nbSrvCbP->regCb.subnetMask),
				NB_DS_UDP_PORT);

	/*------------------------------------
	 * Create and start announcement timer
	 *------------------------------------
	 */
#ifndef FS_STANDALONE
	nbSrvCbP->annceCb.annceTimerHdl = 
			eee_timerCreate(nbSrvCbP->annceCb.interval, EEE_TIMER_REPEAT);
	if (nbSrvCbP->annceCb.annceTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_enableBrowserAnnouncement: "
					"Failed to create sever announcement timer"));
		return (NFX_ERR);
	}

	rc = eee_timerStart(nbSrvCbP->annceCb.annceTimerHdl,
						nb_annceTimeout,		// Timeout handler function
						(void *)nbSrvCbP,		// Ptr to server CB
						0, 0, 0);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_enableBrowserAnnouncement: "
					"Failed to start server announcement timer"));
		return (NFX_ERR);
	}
#endif //FS_STANDALONE

	nbSrvCbP->annceCb.count = 1;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_enableBrowserAnnouncement()"));
	return (NFX_OK);
} /* nb_enableBrowserAnnouncement() */

/*-----------------------------------------------------------------------------
 * Name			: nb_updateAnncePkt
 *
 * Description	: Update node type when CIFS server's node type switches
 * 				  between B-node and H-node.
 * 				  Increase DGM_ID when sending a new announcement.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/20/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_updateAnncePkt(nb_srvCB_t	*nbSrvCbP,
				  uint16		nodeType)
{
	nb_announceCB_t	*annceCbP;
	nb_dsPacket_t	*anncePktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_updateAnncePkt()"));

	if (nbSrvCbP == NULL)
		return;

	annceCbP = &nbSrvCbP->annceCb;
	anncePktP = annceCbP->anncePktP;

	/* Update DGM_ID */
	nb_dsUpdateXid(anncePktP, annceCbP->xmtPktP->pHdr);

	if (anncePktP->header.flags.nodeType != nodeType)
	{
		/* Change node type */
		nb_dsUpdateNodeType(anncePktP, annceCbP->xmtPktP->pHdr, nodeType);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_updateAnncePkt()"));
	return;
} /* nb_updateAnncePkt() */

#ifndef FS_STANDALONE
/*-----------------------------------------------------------------------------
 * Name			: nb_annceTimeout
 *
 * Description	: Server announcement timeout handler function.
 *				  If retry count is less than 3, send out the announcement 
 *				  every 1 minute. If retry count is greater than 3, send
 *				  out the announcement every 12 minutes.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/07/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_annceTimeout(int32 timerHdl, void* param1,
				void* param2, void* param3, void* param4)
{
	int32			rc;
	nb_srvCB_t      *nbSrvCbP;
	uint16			nodeType;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_annceTimeout()"));

	nbSrvCbP = (nb_srvCB_t *)param1;
	if (nbSrvCbP == NULL)
		return (NFX_ERR);

	if (nbSrvCbP->annceCb.annceTimerHdl != timerHdl)
		return (NFX_ERR);

	/* Increase DGM_ID in announcement datagram */
	if (nbSrvCbP->currNodeType == NB_NS_NBFLAG_BNODE)
		nodeType = NB_DS_NODE_TYPE_B;
	else
		nodeType = NB_DS_NODE_TYPE_H;
	nb_updateAnncePkt(nbSrvCbP, nodeType);

	/* Retransmit the announcement */
	nb_xmtPktSP(nbSrvCbP->datagramService.bsdSockP,
				nbSrvCbP->annceCb.xmtPktP,
				nbSrvCbP->regCb.cifsSrvIpAddr | ~(nbSrvCbP->regCb.subnetMask),
				NB_DS_UDP_PORT);

	/* Change announcement timer interval */
	if (nbSrvCbP->annceCb.count < 3)
	{
		/* Still transmit the announcement every minute */
	}
	else if (nbSrvCbP->annceCb.interval < NB_MAX_HOST_ANNCE_INTERVAL)
	{
		/* Stop and destroy the current timer */
		nb_timerStop(&nbSrvCbP->annceCb.annceTimerHdl);
		nb_timerDestroy(&nbSrvCbP->annceCb.annceTimerHdl);

		/* Change the retransmit interval, then create and start 
		 * a new timer with new interval.
		 */
		nbSrvCbP->annceCb.interval = NB_MAX_HOST_ANNCE_INTERVAL;
		nbSrvCbP->annceCb.annceTimerHdl = 
			eee_timerCreate(nbSrvCbP->annceCb.interval, EEE_TIMER_REPEAT);
		if (nbSrvCbP->annceCb.annceTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_annceTimeout: "
						"Failed to create sever announcement timer"));
			return (NFX_ERR);
		}

		rc = eee_timerStart(nbSrvCbP->annceCb.annceTimerHdl,
							nb_annceTimeout,		// Timeout handler function
							(void *)nbSrvCbP,		// Ptr to server CB
							0, 0, 0);
		if (rc != NFX_OK)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_annceTimeout: "
						"Failed to start server announcement timer"));
			return (NFX_ERR);
		}
	}

	nbSrvCbP->annceCb.count++;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_annceTimeout()"));
	return (NFX_OK);
} /* nb_annceTimeout() */
#endif //FS_STANDALONE

