/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfs-cmd.c
 *                              
 * RCS:  $Id: nfs-cmd.c,v 1.16 2002/10/30 00:49:00 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *  Contain NFS UI command functions on TXRX only.
 *
 * Created by: Joseph Vu
 *
 * Date Created:        9/28/01
 *
 *-----------------------------------------------------------------
 */
#ifdef	NFP_TXRX
#include <nfx-incl.h>
#include <exports/nfs-api.h>
#include "nfs-ssc-cmd.h"

#ifdef	LINUX_TEST	
#include <rpc/rpc.h>
#include <string.h>
#include <stdio.h>
#endif	//LINUX_TEST

extern int32   cacheEnable;

int32 nfs_cmdShow ( int32 argc, char *argv[] );
int32 nfs_cmdTrace ( int32 argc, char *argv[] );
int32 nfs_cmdCache( int32 argc, char *argv[] );
int32 nfs_cmdRegPorts( int32 argc, char *argv[] );
int32 nfs_cmdHelpAll(void);
int32 nfs_cmdHelpCmd (char *pCmd);
int32 nfs_cmdSetReqThresHold ( int32 argc, char *argv[] );

typedef struct
{
	int32   (*cmdFunc)( int32 argc, char *argv[] );
	char    *cmd;
	char    *helpTxt;
} nfs_cmdTbl_t;


nfs_cmdTbl_t    nfs_cmdTable[ ] =
{
	{ nfs_cmdShow,  "show",     "nfs show stats|mount|hash" },
	{ nfs_cmdTrace, "trace",    "nfs trace on|off" },
	{ nfs_cmdCache, "cache",    "nfs cache on|off|clear" },
	{ nfs_cmdRegPorts, "portmap",	"nfs portmap <nfsPort> <mntPort>" },
	{ nfs_cmdSetReqThresHold , "req", "nfs req threshold <value>" }
};

#define NFS_CMD_MAX	sizeof(nfs_cmdTable)/sizeof(nfs_cmdTbl_t)

/*-----------------------------------------------------------------
 * Name: nfs_cmdHelp
 *
 * Description: Display NFS help-texts.
 *
 *              nfs help [show | trace]
 *
 * Created by : Joseph Vu 
 *
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdHelp ( int32 argc, char *argv[] )
{


	if (argc < 3)
		return (nfs_cmdHelpAll());
	else
		return (nfs_cmdHelpCmd( argv[2] ));
}

/*-----------------------------------------------------------------
 * Name: nfs_cmdHelpAll
 *
 * Description: Display help-texts for all NFS commands.
 *
 * Created by : Joseph Vu 
 *
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdHelpAll(void)
{
	int32	i;

	printf("Usage:");
	for (i=0; i < NFS_CMD_MAX; i++)
	{
		printf("\t%s\n", nfs_cmdTable[i].helpTxt);
	}
	return (0);
}

/*-----------------------------------------------------------------
 * Name: nfs_cmdHelpCmd
 *
 * Description: Display help-texts for an individual NFS command.
 *
 * Created by : Joseph Vu 
 * 
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdHelpCmd (char *pCmd)
{
	int32	i;

	for (i=0; i < NFS_CMD_MAX; i++)
	{
		if (strequ(pCmd, nfs_cmdTable[i].cmd))
		{
			printf("Usage:");
			printf("\t%s\n", nfs_cmdTable[i].helpTxt);
			return 0;
		}
	}
	return (nfs_cmdHelpAll( ));

}

/*-----------------------------------------------------------------
 * Name: nfs_cmdShow
 *
 * Description: Dispatch for nfs show commands.
 *
 *              nfs show [stat]
 * 
 * Created by : Joseph Vu 
 * 
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdShow ( int32 argc, char *argv[] )
{
	if (argc < 3)
	{
		nfs_cmdHelpCmd("show");
		return (0);
	}

	if ( !(strncmp(argv[2], "stat", 4)))
	{
		nfs_displayStat();
		return (NFX_OK);
	}
	else if ( !(strncmp(argv[2], "mount", 4)))
	{
		nfs_displayMntInfo();
		return (NFX_OK);
	}
	else if ( !(strncmp(argv[2], "hash", 4)))
	{
		return(nfs_hashShow());
	}
	else if ( !(strncmp(argv[2], "cache", 5)))
	{
		return(nfs_cacheShowStats());
	}
	return (nfs_cmdHelpCmd("show"));
}


/*-----------------------------------------------------------------
 * Name: nfs_cmdTrace
 *
 * Description: Enables or disables NFS tracing.
 *
 *              nfs trace <on|off>
 * 
 * Created by : Joseph Vu 
 * 
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdTrace ( int32 argc, char *argv[] )
{
	if (argc < 3)
	{
		nfs_cmdHelpCmd("trace");
		return 0;
	}
	if (strequ(argv[2], "on"))	
	{
		trace_enable = 1;
		printf("NFS trace has been turned on.\n");
	}
	else if (strequ(argv[2], "off"))
	{
		trace_enable = 0;
		printf("NFS trace has been turned off\n");
	}
	else
		nfs_cmdHelpCmd("trace");

	return 0;
}

int32 nfs_cmdSetReqThresHold ( int32 argc, char *argv[] )
{
	int32		nfs_reqThresHold;

	if (argc != 4)
	{
		nfs_cmdHelpCmd("req");
		return 0;
	}

	if ((nfs_reqThresHold = atoi(argv[3])) < 0)
	{
		printf("ERR: Invalid value (%s)\n", argv[3]);
		return -1;
	}

	if (nfs_reqThresHold == 0)
	{
		nfs_globalCfgp->nfs_maxReqsThreshold = 0;
		printf("NFS/file system resources threshold check has been disabled\n");		
		return (0);
	}
	if ((nfs_reqThresHold < NFS_MIN_REQ_THRESHOLD) ||
		(nfs_reqThresHold > NFS_MAX_REQ_THRESHOLD) )
	{
		printf("ERR: Invalid value (%d).  valid threhold value range (%d-%d)\n",
				nfs_reqThresHold,NFS_MIN_REQ_THRESHOLD,NFS_MAX_REQ_THRESHOLD) ;
		printf("or enter '0' to disable checking\n");
		return 0;	
	}
	nfs_globalCfgp->nfs_maxReqsThreshold = nfs_reqThresHold;
	printf("Set NFS req threshold at %d\n", nfs_globalCfgp->nfs_maxReqsThreshold);

	return 0;	
}


/*-----------------------------------------------------------------
 * Name: nfs_cmdCache
 *
 * Description: Enables or disables NFS cache.
 *
 *              nfs cache <on|off>
 * 
 * Created by : Joseph Vu 
 * 
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdCache( int32 argc, char *argv[] )
{
	if (argc < 3)
	{
		nfs_cmdHelpCmd("cache");
		return 0;
	}
	if (strequ(argv[2], "on"))	
	{
		//DISABLE CACHE FOR NOW
		//NEED MORE MEMORY TO MAKE NFS CACHE
		//EFFECTIVE
		//cacheEnable = 1;
		cacheEnable = 0;		
		printf("NFS cache has been turned on.\n");
	}
	else if (strequ(argv[2], "off"))
	{
		cacheEnable = 0;
		printf("NFS cache has been turned off\n");
	}
	else if (strequ(argv[2], "clear"))
	{
		nfs_clearCache();
	}
	else
		nfs_cmdHelpCmd("cache");

	return 0;
}

/*-----------------------------------------------------------------
 * Name: 	nfs_cmdRegPorts 
 *
 * Description: Register MOUNT/NFS ports with port mapper
 *
 *              nfs portmap <nfsport> <mntport>
 * 
 * Created by : Joseph Vu 
 * 
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmdRegPorts( int32 argc, char *argv[] )
{
//	int32		status;
//	int32		port;
	ushort16	nfsPort;
	ushort16	mntPort;

	if (argc != 4)
	{
		nfs_cmdHelpCmd("portmap");
		return 0;
	}

#ifdef LINUX_TEST
	if ((port = atoi(argv[2])) < 0)
	{
		return -1;
	}
	nfsPort = (ushort16) port;

	if ((port = atoi(argv[3])) < 0)
	{
		return -1;
	}
	mntPort = (ushort16) port;

#else
	mntPort = (ushort16) 2087;
	nfsPort = (ushort16) 2088;
#endif

	/*
	 *UNSET
	 */
#ifdef LINUX_TEST
	status = pmap_unset(MNT_PROG, MNT_VER3);
	if (status == 0)
	{
		printf("PORTMAP: Failure to PMAP_UNSET MOUNT V3!\n");
	}

	status = pmap_unset(NFS_PROG, NFS_VER3);
	if (status == 0)
	{
		printf("PORTMAP: Failure to PMAP_UNSET MOUNT V3!\n");
	}

	status = pmap_set(MNT_PROG, MNT_VER3, IPPROTO_UDP, mntPort);
	if (status == 0)
	{
		printf("PORTMAP: Failure to register MOUNT V3 UDP on port %d\n",
                mntPort);
    }

    status = pmap_set(NFS_PROG, NFS_VER3, IPPROTO_UDP, nfsPort);
    if (status == 0)
    {
        printf("PORTMAP: Failure to register  NFS V3 UDP on port %d\n",
                nfsPort);
    }
	printf("Register mntport: %d nfsport:%d with portmapper\n", 
												mntPort, nfsPort);		
	printf("Remember to restart your portmapper &\n");
	printf("NFS deamons after you exit the\n");
	printf("Linux simulation environment.\n");
#endif
	return 0;
}


/*-----------------------------------------------------------------
 * Name: nfs_cmd
 *
 * Description: Main dispatch for NFS commands.  Called from
 *              command table.
 *
 *				nfs show statistics
 *				nfs trace on|off
 * Created by : Joseph Vu
 *
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */
int32 nfs_cmd ( int32 argc, char *argv[] )
{
	int32	i;

	if (argc < 2)
	{
		return (nfs_cmdHelp(argc, argv));
	}

	for (i=0; i < NFS_CMD_MAX; i++)
	{
		if (strequ(argv[1], nfs_cmdTable[i].cmd))
		{
			return ((*nfs_cmdTable[i].cmdFunc)(argc, argv));
		}
	}
	return (nfs_cmdHelp(argc, argv));
}

/*-----------------------------------------------------------------
 * Name:	nfs_rxSscCommand 
 *
 * Description: Received function to process commands received 
 * commands from SSC.
 *
 * Created by : Joseph Vu
 *
 * Date Created: 09/13/2001
 *-----------------------------------------------------------------
 */

void nfs_cmdRcv(eee_desc_t *eDescp )
{
	int32					rc;
	int32					msgLen;
	int32					sendMsgLen;
	uint32					srcPort, destPort;
	eee_descBuf_t			*bd;
	nfs_infoReq_t			*nfs_infoReqp;	
	nfs_statMsg_t  			*nfs_statMsgp;
	nfs_infoResp_t 			*nfs_infoRespMsgp;
	char					*msgBuf;
	
	//printf("nfs_rxSscCommand: Enter\n");
	srcPort = eDescp->hdr.src_port;
    bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
         &eDescp->bd[0] : &eDescp->bd[1];

	nfs_infoReqp = (nfs_infoReq_t *)(bd->buf);
	msgLen = eDescp->hdr.control & EDESC_TOTAL_LEN;
	srcPort = eDescp->hdr.src_port;
	destPort = eDescp->hdr.dest_port;


	msgBuf = eee_ramAlloc( sizeof(nfs_infoResp_t) );

	memset(msgBuf, 0, sizeof(nfs_infoResp_t));

//	printf("%s: Enter resp bufSize = %d\n", 
//		__FUNCTION__, sizeof(nfs_infoResp_t));

	nfs_infoRespMsgp = (nfs_infoResp_t *)msgBuf;

//	sendMsgLen = sizeof(nfs_infoResp_t) + 1;	 ????
	sendMsgLen = sizeof(nfs_infoResp_t) ;	

	switch (nfs_infoReqp->msgHdr.msgType)
	{
		case NFS_REQ_STAT_MSG:
		{
			uint32				clientAddr;
			nfs_clientInfo_t    *clientInfop;
			nfp_clientCBPtr_t   clientCBp;

			clientAddr = nfs_infoReqp->clientAddr;
			//printf("nfs_rxSscCommand: Stat ClientAddr: %08x\n", clientAddr);			
			nfs_infoRespMsgp->msgHdr.msgType = NFS_RESP_STAT_MSG;
			nfs_infoRespMsgp->msgHdr.private = 0;
			nfs_infoRespMsgp->msgHdr.status = NFX_OK;
			nfs_statMsgp = &nfs_infoRespMsgp->msg.stat;
			if (clientAddr)
			{
				//request statistic for a client
				clientCBp = nfp_getClientCBbyAddr(ntohl(clientAddr));
				if (clientCBp)
				{
					//printf("nfs_rxSscCommand 2: ClientAddr: %08x clientCBp: %p\n", clientAddr);
					clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;									
					memcpy(&nfs_statMsgp->nfs_v2XmtPktCounters, 
							&clientInfop->nfs_v2XmtPktCounters, 
								sizeof(nfs_pktCounters_t));
					memcpy(&nfs_statMsgp->nfs_v2RxPktCounters, 
							&clientInfop->nfs_v2RxPktCounters,
								sizeof(nfs_pktCounters_t));
					memcpy(&nfs_statMsgp->nfs_v3XmtPktCounters, 
							&clientInfop->nfs_v3XmtPktCounters, 
								sizeof(nfs_pktCounters_t));
					memcpy(&nfs_statMsgp->nfs_v3RxPktCounters, 
							&clientInfop->nfs_v3RxPktCounters,
								sizeof(nfs_pktCounters_t));
				}
			}
			else
			{
				//return stats received by the NFS server
				memcpy(&nfs_statMsgp->nfs_v2XmtPktCounters, 
					&nfs_globalCfgp->nfs_v2XmtPktCounters, 
					sizeof(nfs_pktCounters_t));

				memcpy(&nfs_statMsgp->nfs_v2RxPktCounters, 
					&nfs_globalCfgp->nfs_v2RxPktCounters,
					sizeof(nfs_pktCounters_t));

				memcpy(&nfs_statMsgp->nfs_v3XmtPktCounters, 
					&nfs_globalCfgp->nfs_v3XmtPktCounters, 
					sizeof(nfs_pktCounters_t));

				memcpy(&nfs_statMsgp->nfs_v3RxPktCounters, 
					&nfs_globalCfgp->nfs_v3RxPktCounters,
					sizeof(nfs_pktCounters_t));

			}

			nfs_statMsgp->nfs_v2RxPktCounters.totalUdpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalUdpCalls;
			nfs_statMsgp->nfs_v2RxPktCounters.totalBadUdpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalBadUdpCalls;
			nfs_statMsgp->nfs_v2RxPktCounters.totalTcpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalTcpCalls;
			nfs_statMsgp->nfs_v2RxPktCounters.totalBadTcpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalBadTcpCalls;

			nfs_statMsgp->nfs_v3RxPktCounters.totalUdpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalUdpCalls;
			nfs_statMsgp->nfs_v3RxPktCounters.totalBadUdpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalBadUdpCalls;
			nfs_statMsgp->nfs_v3RxPktCounters.totalTcpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalTcpCalls;
			nfs_statMsgp->nfs_v3RxPktCounters.totalBadTcpCalls =
				nfs_globalCfgp->nfs_rxPktCounters.totalBadTcpCalls;


			nfs_infoRespMsgp->count_per_usec = count_per_usec;
			pkt_getNfsProcTime(&nfs_infoRespMsgp->procTime); 
			break;
		}
		default:
            nfs_infoRespMsgp->msgHdr.msgType = -1;
            nfs_infoRespMsgp->msgHdr.private = -1;
            nfs_infoRespMsgp->msgHdr.status = NFX_ERR;
			break;

	}
	/*
	printf("nfs_rxSscCommand: src appId: %d src slot: %d src cpu: %d\n",
					EEE_GET_APP_ID(srcPort),
					EEE_GET_SLOT_NUM(srcPort),
					EEE_GET_DEST_CPU(srcPort) );

	printf("nfs_rxSscCommand: dest appId: %d dest slot: %d dest cpu: %d\n",
					EEE_GET_APP_ID(destPort),
					EEE_GET_SLOT_NUM(destPort),
					EEE_GET_DEST_CPU(destPort) );
	*/

	rc = eee_sendMessage(msgBuf, sendMsgLen, srcPort, destPort);
	if (rc != NFX_OK)
	{
		printf("eee_sendMessage failed (rc: %d)!\n",rc); 

	}
	/*
	printf("nfs_rxSscCommand: msgLen: %d dest: 0x%x src: 0x%x rc: %d\n", 
	       sendMsgLen, srcPort, destPort, rc);
	*/

	eee_ramDealloc( msgBuf );

	eee_freePacket(eDescp);
	return;

}

#endif	//NFP_TXRX
