/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfs-hash.c
 *
 * RCS:  $Id: nfs-hash.c,v 1.3 2002/10/01 00:37:34 shaun Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contain NFS hashing related routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 08/10/02
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX
    
#include <nfx-incl.h>
#include <exports/pkt-api.h> 
#include <exports/nfs-api.h> 
#include "nfs-hash.h"

#define NFS_HASH_XID( xid_key )	( ((xid_key >> 16) ^ (xid_key)) & \
								  (NFS_MAX_HASH_REQCB-1))

typedef struct
{
	int32	lookupCnt;
	int32	lookupHWM;
	int32	lookupErr;
	int32	deleteCnt;
	int32	deleteHWM;
	int32	deleteErr;
	int32	insertCnt;
	int32	insertErr;
} nfs_hashStats_t;

nfs_hashStats_t	nfs_hashStats = {0};

/*-----------------------------------------------------------------
 * Name:        nfs_hashCreateReqCb
 *  
 * Description: Create hash table for nfs reqCB.
 *              
 * Created by:  Shaun Hui
 *  
 * Date Created: 08/10/2002
 *-----------------------------------------------------------------
 */
void *
nfs_hashCreateReqCb( int32 tbl_size )
{
    int32   allocSize;
    void    *pTbl;

    allocSize = tbl_size * sizeof( void * );
    pTbl      = eee_ramAlloc( allocSize );

    printf("%s: pTbl=%p size=%d\n", __FUNCTION__, pTbl, allocSize);
    if (pTbl)
        memset(pTbl, 0, allocSize );

    return (pTbl);
}



/*-----------------------------------------------------------------
 * Name:        nfs_hashLookupReqCb
 *  
 * Description: Lookup Req CB based on XID.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 08/10/2002
 *-----------------------------------------------------------------
 */
void *
nfs_hashLookupReqCb( uint32 xid )
{
    int32           cnt;
    int32           hashIdx;
	void			*pHashTbl;
    nfs_request_t   *nfsReqp;
    nfs_request_t   **ppReq;
	nfs_hashStats_t	*pStats;

	pStats   = &nfs_hashStats;
	pHashTbl = nfs_globalCfgp->pReqCbHashTbl;

	/*-------------------------------
	 * hash xid as index into lookup table.
	 *-------------------------------
	 */
    hashIdx = NFS_HASH_XID( xid );
    ppReq   = (nfs_request_t **)(pHashTbl);
    nfsReqp = ppReq[hashIdx];

#ifdef NFS_HASH_DBG
    printf("%s: xid=%x hashIdx=%d ppReq=%p nfsReqp=%p\n",
            __FUNCTION__, xid, hashIdx, ppReq, nfsReqp);
#endif

    cnt = 0;

	/*-------------------------------
	 * Go thru chain, if collision.
	 *-------------------------------
	 */
    while(nfsReqp)
    {
		/*-------------------------------
		 * Update hi water mark for debugging.
		 *-------------------------------
		 */
        cnt++;
        if (pStats->lookupHWM < cnt)
            pStats->lookupHWM = cnt;

        if (nfsReqp->rpcCallHdrp)
        {
#ifdef NFS_HASH_DBG
            printf("%s: rpc_xid=%x nfsReqp=%p pNext=%p\n", 
                    __FUNCTION__, nfsReqp->rpcCallHdrp->rpc_xid,
                    nfsReqp, nfsReqp->pNext);
#endif
			/*-------------------------------
			 * Look for matching XID
			 *-------------------------------
			 */
            if (nfsReqp->rpcCallHdrp->rpc_xid == xid)
            {
				pStats->lookupCnt++;
                break;
            }
        }
		else
		{
				pStats->lookupErr++;
		}

        nfsReqp = nfsReqp->pNext;
    }

    return (nfsReqp);
}

/*-----------------------------------------------------------------
 * Name:        nfs_hashInsertReqCb
 *  
 * Description: Insert Req CB into hash table.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 08/10/2002
 *-----------------------------------------------------------------
 */
void 
nfs_hashInsertReqCb( uint32 xid, nfs_request_t *nfsReqp)
{
    int32           hashIdx;
	void			*pHashTbl;
    nfs_request_t   *hashReq;
    nfs_request_t   **ppReq;
	nfs_hashStats_t	*pStats;

	pStats   = &nfs_hashStats;
	pHashTbl = nfs_globalCfgp->pReqCbHashTbl;

	/*-------------------------------
	 * hash xid as index into lookup table.
	 *-------------------------------
	 */
    hashIdx = NFS_HASH_XID( xid );
    ppReq   = (nfs_request_t **)(pHashTbl);
    hashReq = ppReq[hashIdx];

#ifdef NFS_HASH_DBG
    printf("%s: xid=%x hashIdx=%d *ppReq=%p nfsReqp=%p\n",
            __FUNCTION__, xid, hashIdx, *ppReq, nfsReqp);
#endif

	/*-------------------------------
	 * Check if entry alreay exists.
	 *-------------------------------
	 */
    if (nfs_hashLookupReqCb( xid ))
    {
        printf("%s: duplicate entry: xid=%x\n", __FUNCTION__, xid);
		pStats->insertErr++;
        return;
    }

	pStats->insertCnt++;

	/*-------------------------------
	 * Insert at beginning of chain.
	 *-------------------------------
	 */
    ppReq[hashIdx] = nfsReqp;
    nfsReqp->pNext = hashReq;

}

/*-----------------------------------------------------------------
 * Name:        nfs_hashDeleteReqCb
 *  
 * Description: Delete Req CB from hash table.
 *
 * Created by:  Shaun Hui
 *  
 * Date Created: 08/10/2002
 *-----------------------------------------------------------------
 */
void *nfs_hashDeleteReqCb( uint32 xid )
{
    int32           cnt;
    int32           hashIdx;
    void			*pHashTbl;
    nfs_request_t   *nfsReqp;
    nfs_request_t   *pPrev;
    nfs_request_t   **ppReq;
	nfs_hashStats_t	*pStats;

	pStats   = &nfs_hashStats;
	pHashTbl = nfs_globalCfgp->pReqCbHashTbl;


	/*-------------------------------
	 * hash xid as index into lookup table.
	 *-------------------------------
	 */
    hashIdx = NFS_HASH_XID( xid );
    ppReq   = (nfs_request_t **)(pHashTbl);
    nfsReqp = ppReq[hashIdx];

#ifdef NFS_HASH_DBG
    printf("%s: xid=%x hashIdx=%d ppReq=%p nfsReqp=%p\n",
            __FUNCTION__, xid, hashIdx, ppReq, nfsReqp);
#endif

    cnt = 0;
    if ((!nfsReqp) || (nfsReqp->rpcCallHdrp == NULL))
    {
        pStats->deleteErr++;
        return NULL;
    }

	/*-------------------------------
	 * Delete entry from top of list.
	 *-------------------------------
	 */
    if (nfsReqp->rpcCallHdrp->rpc_xid == xid)
    {
        ppReq[hashIdx] = nfsReqp->pNext;
        nfsReqp->pNext = NULL;
		pStats->deleteCnt++;
    }
    else
    {
		/*-------------------------------
		 * Delete entry from chain.
		 *-------------------------------
		 */
        pPrev = nfsReqp;

        nfsReqp = nfsReqp->pNext;
        while(nfsReqp)
        {
			/*-------------------------------
			 * Update hi water mark for debugging.
			 *-------------------------------
			 */
			cnt++;
			if (pStats->deleteHWM < cnt)
				pStats->deleteHWM = cnt;

#ifdef NFS_HASH_DBG
            printf("%s: rpc_xid=%x nfsReqp=%p pNext=%p\n",
                    __FUNCTION__, nfsReqp->rpcCallHdrp->rpc_xid,
                    nfsReqp, nfsReqp->pNext);
#endif

            if ((nfsReqp->rpcCallHdrp != NULL) &&
                (nfsReqp->rpcCallHdrp->rpc_xid == xid))
            {
                pPrev->pNext = nfsReqp->pNext;
                nfsReqp->pNext = NULL;

				pStats->deleteCnt++;
                break;
            }
            pPrev   = nfsReqp;
            nfsReqp = nfsReqp->pNext;

        }
    }

    return (nfsReqp);
}


int32
nfs_hashShow( void )
{
	nfs_hashStats_t	*pStats;

	pStats = &nfs_hashStats;

	printf("nfs hash:\n");
	printf("lookup HWM=%d\n", pStats->lookupHWM);
	printf("delete HWM=%d\n", pStats->deleteHWM);
	printf("lookup cnt=%d\n", pStats->lookupCnt);
	printf("insert cnt=%d\n", pStats->insertCnt);
	printf("delete cnt=%d\n", pStats->deleteCnt);
	printf("lookup err=%d\n", pStats->lookupErr);
	printf("insert err=%d\n", pStats->insertErr);
	printf("delete err=%d\n", pStats->deleteErr);
	return 0;
}
#endif /* NFP_TXRX */
