/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfsv3-msg.c
 *
 * RCS:  $Id: nfsv3-msg.c,v 1.49 2002/10/30 00:52:20 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *      Contain NFS Version 3 and File System interface functions.
 *
 * Created by: Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
#ifdef  NFP_TXRX
#include <nfx-incl.h>
#include <exports/nfs-api.h>
#include <exports/malloc-api.h>
#include "nfsv3-msg.h"
#include "nfs-cache.h"

#define		NFS_DBG_COD
// #define	NFS_DBG_PRINTF_ON

extern uint32  trace_enable;

uint64	extern boottime;		/*TODO:  Get Boot time*/



typedef struct
{
	fs_Attr	*pAttr[ NFS_MAX_ATTR_IDX ];
}	nfs_attrArray_t;

uint32
nfs_findAttr( fs_AttrType		attrType,
			  fs_Attr 			*attrp,
			  int32				numAttrs,
			  nfs_attrArray_t 	*nfsAttrp)
{
	int32	attrIdx;
	int32	attrMap;


#ifdef NFS_DBG_PRINTF_ON
	printf("--------------------------------------------------\n");
	printf("%s: attr count %d\n", __FUNCTION__, numAttrs);
#endif

	/*
	 * When FS encounters err, numAttr is invalid. 
	 * Only attrType (part of response header) is valid.
	 */
	if (!attrType)
		return (0);

	attrMap = 0;
	while (numAttrs--)
	{
		if ( attrp->attrType <= FS_POST_TPAR_ATTRS) 
		{
			attrIdx  = nfs_attrIdxArray[attrp->attrType];
			attrMap |= attrp->attrType;
			nfsAttrp->pAttr[attrIdx] = attrp;
#ifdef NFS_DBG_PRINTF_ON
			printf("%s: map fs attr 0x%02x into nfs attr idx 0x%x\n",
				__FUNCTION__, attrp->attrType, attrIdx);
#endif
		}
		else
		{
			/*
			 * Encounter invalid attr.
			 * Return 0 attrMap so that no attr 
			 * processing will be done.
			 */
			return (0);
		}

		attrp++;
	}

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: attr map 0x%08x\n", __FUNCTION__, attrMap);
#endif

	return (attrMap);
}


/*-----------------------------------------------------------------
 * Name:		nfsv3_msgGetAttrResp 
 *
 * Description: Map file system Getattr response message to a NFS 
 *				reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
         *NFX_OK or NFX_ERR
         */
nfsv3_msgGetAttrResp( int32 		status, 	/*operation result status*/
					  eee_descBuf_t *bd,		/*file system resp. msg*/
					  char 			*txMsgp,	/*Pointer to tx msg buffer*/
					  int32 		*txMsgLen,  /*current length of message*/
					  nfs_request_t *nfsReqp  	/*reqCB*/
)
{
	fs_MsgGetAttrResp 		*fsGetAttrMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsGetAttrMsgp = (fs_MsgGetAttrResp *)bd->buf;

	/*
	 * Increment client counter.
	 */
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.getAttr++;
	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.getAttr++;

	/*
	 * Increment debug counters.
	 */
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.getAttr++;
	nfs_globalCfgp->nfs_xmtPktCounters.getAttr++;	
	nfs_globalCfgp->nfs_v3XmtPktCounters.getAttr++;

	/*
	 * Set status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	if (status != NFS_OK)
		return (NFX_OK);


#ifdef 	NFS_DBG_PRINTF_ON
	printf("%s: filesize: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, 
			fsGetAttrMsgp->attr.size, 
			fsGetAttrMsgp->attr.fsId, 
			fsGetAttrMsgp->attr.inode); 
#endif

	*txMsgLen += nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										&fsGetAttrMsgp->attr, 
										txMsgp);
	
#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	/* 
	 * Update attr cache.
	 */
	nfs_cacheUpdateAttr(&nfsReqp->fileHandle, &fsGetAttrMsgp->attr);

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	nfsv3_msgLookupResp	
 *
 * Description: Map File System Lookup response to a NFS V3 reply. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgLookupResp(int32 			status,		/*operation result status*/
					eee_descBuf_t 	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*Pointer to tx msg buffer*/
					int32 			*txMsgLen, 	/*current length of message*/
					nfs_request_t 	*nfsReqp )	/*reqCB*/ 
{
	fs_MsgLookupResp 		*fsLookupMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t		clientCBp;
	fs_Attr					*pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsLookupMsgp = (fs_MsgLookupResp *)bd->buf;

	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.lookUp++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.lookUp++;
	nfs_globalCfgp->nfs_xmtPktCounters.lookUp++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.lookUp++;

	length = 0;	

	/*
	 * Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	/*
	 * Find ALL attributes returned for FS.
	 */
	attrMap = nfs_findAttr( fsLookupMsgp->hdr.attrType,
							fsLookupMsgp->attr,
				  			fsLookupMsgp->numAttrs,
				  			&attrArray );

	/*
	 * Error occured, return
	 */
	if (status != NFS_OK)
	{
		/*
		 *Return dir post attributes
		 *Defect 775: Fixed logic to set "attributes 
		 *not available"
		 *when received no attributes from the file system
		 */
		if (attrMap & FS_POST_PAR_ATTRS)
		{
			pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];

			length = nfs_storeV3Attributes( nfsReqp->reqType,
											NFS_ATTR_AVAIL,
											pAttr, 
											txMsgp);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Not OK: Post Parent: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

			nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
    	}
		else
		{
			/*
			 * Set attr. available flag
			 */
			*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
			length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: Not OK: No Attr\n", __FUNCTION__);
#endif
		}

   		*txMsgLen += length;
		txMsgp    += length;

		return (NFX_ERR);
	}
	
	/*
	 *Store return filehandle from the file system for 
	 *search name object
	 *to destination response msg buffer
	 */
	length = nfs_storeV3Fhandle( &fsLookupMsgp->fileHandle, txMsgp);

	/*
 	 *Update current destination  buffer pointer & length
 	 */
	*txMsgLen += length;
	txMsgp    += length;

	/*
 	 *Store post-operation file attributes in to destination 
	 *response msg buffer 
 	 */

    /*
     *Get post-op object attributes
     */
	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);

		nfs_cacheUpdateAttr(&fsLookupMsgp->fileHandle, pAttr);
	}
	else
	{
#ifdef 	NFS_DBG_PRINTF
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);
	}

	/*
	 *update len & msg ptr.
	 */
 	*txMsgLen += length;
	txMsgp    += length;

	/*
	 *Get parent dir post attributes
	 */
	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
#ifdef 	NFS_DBG_PRINTF
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);
	}


   	*txMsgLen += length;
	txMsgp    += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif


	return (NFX_OK);		
}


/*-----------------------------------------------------------------
 * Name: 		nfsv3_msgSetAttrResp 
 *
 * Description: Map File System SetAttr response to a NFS reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgSetAttrResp(int32 			status,		/*operation result status*/
					 eee_descBuf_t	*bd,		/*file system resp. msg*/
					 char 			*txMsgp,	/*Pointer to tx msg buffer*/
					 int32 			*txMsgLen,	/*current length of message*/
					 nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgSetAttrResp 		*fsSetAttrMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	fs_Attr					*pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsSetAttrMsgp = (fs_MsgSetAttrResp *)bd->buf;

	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.setAttr++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.setAttr++;
	nfs_globalCfgp->nfs_xmtPktCounters.setAttr++;	
	nfs_globalCfgp->nfs_v3XmtPktCounters.setAttr++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp 	  += sizeof(int32);
	*txMsgLen += sizeof(int32);

	/*
	 * Find ALL attributes returned for FS.
	 */
	attrMap = nfs_findAttr( fsSetAttrMsgp->hdr.attrType,
							fsSetAttrMsgp->attr,
				  			fsSetAttrMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Get pre-operation attributes if available
	 */
	if (attrMap & FS_PRE_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE ];

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
	}
	else
	{
#ifdef 	NFS_DBG_PRINTF
		printf("%s: No Pre Attr\n", __FUNCTION__);
#endif
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);
	}

	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Store post-operation attributes if available
	 */
	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
#ifdef 	NFS_DBG_PRINTF
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);
	}

	*txMsgLen += length;
	txMsgp += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name: 		nfsv3_msgAccessResp 
 *
 * Description: Map File System Access response to a NFS V3 reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgAccessResp(int32 			status,		/*operation result status*/
					eee_descBuf_t	*bd,		/*file system resp. msg*/	
					char 			*txMsgp,	/*Pointer to tx msg buffer*/
					int32 			*txMsgLen, 	/*current length of message*/
					nfs_request_t 	*nfsReqp ) 	/*reqCB*/
{
	int32					length;
	fs_Attr					*pAttr;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	fs_MsgAccessResp 		*fsAccessMsgp;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsAccessMsgp = (fs_MsgAccessResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.access++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.access++;
	nfs_globalCfgp->nfs_xmtPktCounters.access++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.access++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	/*
	 * Find ALL attributes returned for FS.
	 */
	attrMap = nfs_findAttr( fsAccessMsgp->hdr.attrType,
				  			fsAccessMsgp->attr,
				  			fsAccessMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Store the post-op attributes of the object
	 */
	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
#ifdef 	NFS_DBG_PRINTF
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);
	}

	txMsgp += length;
	*txMsgLen += length;

	if (status != NFS_OK)
		return (NFX_OK);
	
	/*
	 *bit mask of access permission
	 */
	*(uint32 *)txMsgp = fsAccessMsgp->accessBits;
	*txMsgLen += 4;
	txMsgp += 4;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);	
}


/*-----------------------------------------------------------------
 * Name:		nfsv3_msgSymlinkResp 
 *
 * Description: Map File System Symlink response to a NFS reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgSymlinkResp(int32 			status,		/*operation result status*/
					 eee_descBuf_t	*bd,		/*file system resp. msg*/
					 char 			*txMsgp, 	/*pointer to tx Msg buffer*/
					 int32 			*txMsgLen, 	/*current length of message*/
					 nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	int32					idx;
	void					*pCache;
	fs_MsgSymLinkResp 		*fsSymlinkMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	fs_Attr                 *pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsSymlinkMsgp = (fs_MsgSymLinkResp *)bd->buf;

	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.symlink++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.symlink++;
	nfs_globalCfgp->nfs_xmtPktCounters.symlink++;	
	nfs_globalCfgp->nfs_v3XmtPktCounters.symlink++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsSymlinkMsgp->hdr.attrType,
				  			fsSymlinkMsgp->attr,
				  			fsSymlinkMsgp->numAttrs,
				  			&attrArray );

	if (status == NFS_OK)
	{	
		/*
		 * Set present flag.
		 */
		*(uint32 *)txMsgp = TRUE;

		txMsgp    += sizeof(int32);
		*txMsgLen += sizeof(int32);

		/*
	 	 *Store return filehandle from the file system 
	 	 */
		length = nfs_storeV3Fhandle(&fsSymlinkMsgp->fileHandle, txMsgp);

		/*
	 	 *Update current destination  buffer pointer & length
	 	 */
		txMsgp    += length;
		*txMsgLen += length;

		if (attrMap & FS_POST_ATTRS)
		{	
			pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];

			/*
	 	 	 *store attributes for the newly created symbolic link
	 	 	 */
			length = nfs_storeV3Attributes( nfsReqp->reqType, 
											NFS_ATTR_AVAIL,
											pAttr, 
											txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
            pCache = nfs_cacheFindSymlnk(&fsSymlinkMsgp->fileHandle, &idx );
            if ((!pCache) && (nfsReqp->nameLen > 0) && (nfsReqp->nameLen < 32))
            {
	            nfs_cacheAddSymlnk( &fsSymlinkMsgp->fileHandle,
	                                pAttr,
                                	idx,
                            		nfsReqp->fname,
                            		nfsReqp->nameLen);
            }

		}
		else
		{
			/*
			 * Set attr. available flag
			 */
			*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
			length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		}

		*txMsgLen += length;
		txMsgp += length;
	}

	/*
	 * Store pre-op (wcc_data) attributes for parent directory
	 */
	if (attrMap & FS_PRE_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];

		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

	/*
	 * Store post-op (wcc_data) attributes for parent directory
	 */
	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];

		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:      	nfsv3_msgHardLinkResp 
 *
 * Description: Map File System Link response to a NFS V3 reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgHardLinkResp(int32			status,		/*operation result status*/
					  eee_descBuf_t *bd,      	/*file system resp. msg*/
					  char 			*txMsgp,	/*pointer to tx msg buffer*/
					  int32 		*txMsgLen,	/*current length of message*/
					  nfs_request_t *nfsReqp )	/*reqCB*/
{
	fs_MsgHardLinkResp		*fsLinkRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	fs_Attr					*pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsLinkRespMsgp = (fs_MsgHardLinkResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.link++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.link++;
	nfs_globalCfgp->nfs_xmtPktCounters.link++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.link++;
	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp += 4;
	*txMsgLen += 4;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsLinkRespMsgp->hdr.attrType,
				  			fsLinkRespMsgp->attr,
				  			fsLinkRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Store post-op Attr
	 */
	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
		length = nfs_storeV3Attributes(	nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;


	/*
	 * Store pre-op (linkdir_wcc) 
	 * for parent dir for the link
	 */
	if ( attrMap & FS_PRE_PAR_ATTRS )
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];
		length = nfs_storeV3Attributes(nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->toFileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name: 		nfsv3_msgReadlinkResp 
 *
 * Description: Map File System Readlink response to a NFS V3 reply. 
 *
 * Created by:  Joseph Vu
 * 
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*
	 	 *NFX_OK or NFX_ERR
	 	 */
nfsv3_msgReadlinkResp(int32 		status,		/*operation result status*/
					  eee_descBuf_t *bd,		/*file system resp. msg*/
					  char 			*txMsgp, 	/*Pointer to tx msg buffer*/
					  int32 		*txMsgLen, 	/*length of msg buffer*/
					  nfs_request_t *nfsReqp )	/*req control block*/
{
	fs_MsgReadLinkResp 		*fsReadLinkMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	fs_Attr                 *pAttr;
	int32					length;
	ushort16				linkNameLen;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsReadLinkMsgp = (fs_MsgReadLinkResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.readLink++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.readLink++;
	nfs_globalCfgp->nfs_xmtPktCounters.readLink++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.readLink++;
	length = 0;

	/*
	 * Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp += 4;
	*txMsgLen += 4;


	/*
 	 * Get attributes returned from the symbolic link
 	 */
	if (fsReadLinkMsgp->hdr.attrType & FS_POST_ATTRS)
	{
		pAttr  = &fsReadLinkMsgp->attr;
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr,
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;


	if (status != NFS_OK)
		return (NFX_ERR);

	linkNameLen = fsReadLinkMsgp->nameLen;	
	if (linkNameLen <= 0)
		return (NFX_ERR);

	/*
	 *Get data associated with the symbolic link
	 */
	length = nfs_storeString(txMsgp, 
							fsReadLinkMsgp->linkName, 
							linkNameLen, 
							FS_PATH_MAX);
	*txMsgLen += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfsv3_msgCreateResp
 *
 * Description: Map File System Create response to a NFS V3 reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
         *NFS_OK or NFX_ERR
         */
nfsv3_msgCreateResp(int32 			status,		/*Operation result status*/
					eee_descBuf_t	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*Poiner to tx msg buffer*/
					int32 			*txMsgLen,	/*length of the message*/
					nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgCreateResp 		*fsCreateRespMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	fs_Attr                 *pAttr;
	int32                   length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsCreateRespMsgp = (fs_MsgCreateResp *)bd->buf;

	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.create++;	

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.create++;
	nfs_globalCfgp->nfs_xmtPktCounters.create++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.create++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	attrMap = nfs_findAttr( fsCreateRespMsgp->hdr.attrType,
				  			fsCreateRespMsgp->attr,
				  			fsCreateRespMsgp->numAttrs,
				  			&attrArray );

	if (status == NFX_OK)
	{

		/*
		 * Set present flag.
		 */
		*(uint32 *)txMsgp = TRUE;

		txMsgp    += sizeof(int32);
		*txMsgLen += sizeof(int32);

		/*
	 	 *Store return filehandle from file system 
	 	 *to the destination response msg buffer
	 	 */
		length = nfs_storeV3Fhandle(&fsCreateRespMsgp->fileHandle, txMsgp);

		/*
	 	 *Update current destination  buffer pointer & length
	 	 */
		*txMsgLen += length;
		txMsgp += length;

		/*
	 	 * Store the post object attributes
	 	 */
		if (attrMap & FS_POST_ATTRS)
		{
			pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
			length = nfs_storeV3Attributes(nfsReqp->reqType, 
											NFS_ATTR_AVAIL, 
											pAttr, 
											txMsgp);		
#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
				__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

			nfs_cacheUpdateAttr(&fsCreateRespMsgp->fileHandle, pAttr);
		}
		else
		{
			/*
			 * Set attr. available flag
			 */
			*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
			length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		}

		*txMsgLen += length;
		txMsgp += length;

	}	//if (status == NFX_OK)

	/*
	 * Store pre-op (wcc) attributes 
	 * for the parent directory
	 */
	if (attrMap & FS_PRE_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

	/*
	 * Store post-op (wcc) attributes 
	 * for the parent directory
	 */
	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;
	
#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return ((status != NFS_OK)	?  NFX_ERR : NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv3_msgRemoveResp 
 *
 * Description: Map File System Remove response to a NFS reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgRemoveResp(int32 			status,		/*Operation result status*/
					eee_descBuf_t	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*pointer to tx msg buffer*/
					int32 			*txMsgLen,	/*current length of message*/
					nfs_request_t	*nfsReqp )	/*reqCB*/
{
	fs_MsgRemoveResp 		*fsRemoveRespMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	fs_Attr                 *pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsRemoveRespMsgp = (fs_MsgRemoveResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.remove++;

	nfs_globalCfgp->nfs_v3XmtPktCounters.remove++;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.remove++;
	nfs_globalCfgp->nfs_xmtPktCounters.remove++;

	length = 0;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsRemoveRespMsgp->hdr.attrType,
				  			fsRemoveRespMsgp->attr,
				  			fsRemoveRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);


	/*
	 *Store attributes to return to client
	 */

	/*
	 *Store pre-op attributes from file system in 
	 *destination transmit msg buffer
	 */
	if (attrMap & FS_PRE_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Store post-op attributes from file system 
	 *in destination transmit msg buffer
	 */
	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

    if (attrMap & FS_POST_ATTRS)
    {
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
		nfs_cacheUpdateAttr(&fsRemoveRespMsgp->fileHandle, pAttr);
    }

	*txMsgLen += length;
	txMsgp += length; 

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return ((status != NFS_OK)	?  NFX_ERR : NFX_OK);
}       

/*-----------------------------------------------------------------
 * Name:      	nfsv3_msgRmdirResp 
 *
 * Description:
 *       Map File System RMDIR response to a NFS reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgRmdirResp( int32 			status,		/*Operation result status*/
					eee_descBuf_t	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*pointer to tx msg buffer*/
					int32 			*txMsgLen,	/*current length of message*/
					nfs_request_t	*nfsReqp )	/*reqCB*/
{
	fs_MsgRmdirResp 		*fsRmdirRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	fs_Attr                 *pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsRmdirRespMsgp = (fs_MsgRmdirResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.rmdir++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.rmdir++;
	nfs_globalCfgp->nfs_xmtPktCounters.rmdir++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.rmdir++;

	length = 0;


	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsRmdirRespMsgp->hdr.attrType,
				  			fsRmdirRespMsgp->attr,
				  			fsRmdirRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp += 4;
	*txMsgLen += 4;

	/*
	 *Store pre-op attributes from file system in 
	 *destination transmit msg buffer
	 */
	if (attrMap & FS_PRE_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Store post-op attributes from file system 
	 *in destination transmit msg buffer
	 */
	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length; 

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return ((status != NFS_OK)	?  NFX_ERR : NFX_OK);
}       
       

/*-----------------------------------------------------------------
 * Name:       	nfsv3_msgMkdirResp 
 *
 * Description: Map File System Mkdir response to a NFS reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgMkdirResp(int32 			status,		/*operation result status*/
					eee_descBuf_t	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*pointer to tx msg buffer*/
					int32 			*txMsgLen, 	/*current length of message*/
					nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgMkdirResp 		*fsMkdirRespMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t		clientCBp;
	fs_Attr                 *pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsMkdirRespMsgp = (fs_MsgMkdirResp *)bd->buf;

	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.mkdir++;

	nfs_globalCfgp->nfs_v3XmtPktCounters.mkdir++;
	nfs_globalCfgp->nfs_xmtPktCounters.mkdir++;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.mkdir++;

	length = 0;   

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsMkdirRespMsgp->hdr.attrType,
				  			fsMkdirRespMsgp->attr,
				  			fsMkdirRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	if  (status == NFS_OK)
	{
		/*
		 * Set field present flag.
		 */
		*(uint32 *)txMsgp = TRUE;
		*txMsgLen += sizeof(int32);
		txMsgp    += sizeof(int32);

		/*
	 	 *Store the return filehandle from the file system into the 
	 	 *destination transmit msg buffer
	 	 */
		length = nfs_storeV3Fhandle(&fsMkdirRespMsgp->dirHandle, txMsgp);
		*txMsgLen += length;
		txMsgp    += length;

		/*
	 	 *Store the attributes for newly created sub-dir,
		 *the object post-op attributes
	 	 */
		if (attrMap & FS_POST_ATTRS)
		{
			pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
			length = nfs_storeV3Attributes( nfsReqp->reqType, 
											NFS_ATTR_AVAIL, 
											pAttr, 
											txMsgp );
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
			nfs_cacheUpdateAttr(&fsMkdirRespMsgp->dirHandle, pAttr);
		}
		else
		{
			/*
			 * Set attr. available flag
			 */
			*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
			length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		}

		*txMsgLen += length;
		txMsgp    += length;

	}	//if  (status == NFS_OK)

	/*
	 *store the pre-op & post-op (wcc_data) attributes of the parent directory
	 */
	if (attrMap & FS_PRE_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}
	
	*txMsgLen += length;
	txMsgp    += length;

	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];
		length = nfs_storeV3Attributes(nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return ((status != NFS_OK)	?  NFX_ERR : NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:        nfsv3_msgReaddirResp
 *
 * Description: Map File System Readdir response to a NFS V3 reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*
		 *NFX_OK  or NFX_ERR
	 	 */
nfsv3_msgReaddirResp(int32 			status,		/*Operation result status*/
					 eee_descBuf_t 	*bd,		/*file system resp. msg*/
					 char 			*txMsgp,	/*Pointer to tx msg buffer*/
					 int32 			*txMsgLen,	/*current length of message*/
					 nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgReadDirResp 	*fsReaddirRespMsgp;
	nfs_clientInfo_t    *clientInfop;
	nfp_clientCBPtr_t	clientCBp;
	int32				msgLen;
	int32				length;
	int32				i;
	fs_DirEntry			*dirent;
	char8				*msgBufp;
	eee_desc_t 			*eDescp;
	static char			msgBuf[8192];

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	/*
	 *Retrieve EEE message into local buffer
	 */	
	
	eDescp = nfsReqp->edescp;
	msgLen =  (eDescp->hdr.control & EDESC_TOTAL_LEN);

	/*
	 *Check if message length fits in single descriptor
	 */
	if (msgLen < EEE_BUFSIZE_LARGE)
	{

		fsReaddirRespMsgp = (fs_MsgReadDirResp *)bd->buf;
	}
	else
	{

//		memset(msgBuf, 0, sizeof(msgBuf));
		msgBufp = &msgBuf[0];

		if (eee_copyPacketToLocalBuffer(msgBufp, 
										eDescp, 
										msgLen, 0) != msgLen)
		{
			/*copied failed*/	
			panic("%s: eee_copyPacketToLocalBuffer failed!\n", __FUNCTION__);
		}
		fsReaddirRespMsgp = (fs_MsgReadDirResp *)msgBufp;
	}
	
	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.readdir++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.readdir++;
	nfs_globalCfgp->nfs_xmtPktCounters.readdir++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.readdir++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	/*
 	 *Store post-op attributes of the directory
 	 *FS API is missing this field
 	 */
	length = nfs_storeV3Attributes( nfsReqp->reqType, 
									NFS_ATTR_AVAIL,
									&fsReaddirRespMsgp->dirAttr, 
									txMsgp);
	*txMsgLen += length;
	txMsgp += length;

	if (status != NFS_OK)
		return (NFX_OK);

	/*
 	 *Store the cookie verifier 
 	 */	
	length = nfs_storeLongLongValToBuf(txMsgp, 
				fsReaddirRespMsgp->cookieVerif);
	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Get number of directory entries
	 */	
#ifdef NFS_DBG_PRINTF_ON
	printf("%s: numOfEntries: %d\n", 
			__FUNCTION__, fsReaddirRespMsgp->numEntriesThis);
#endif

	dirent = &fsReaddirRespMsgp->dirEntry[0];
	for (i = 0; i < fsReaddirRespMsgp->numEntriesThis; i++)
	{
		/*
		 *Retrieve fileid, name, cookie
		 */
#ifdef NFS_DBG_PRINTF_ON
		printf("%s: i: %d inode: %llx cookie: %llx nameLen: %d name: %s\n",
				__FUNCTION__, i, 
				dirent->inode, dirent->cookie, 
				dirent->nameLen, dirent->name);
#endif

		length = nfs_storeDirEntry(NFS_VER3, dirent, txMsgp);
		FS_READDIR_ADV_ENTRY_PTR(dirent);
		txMsgp    += length;
		*txMsgLen += length;
	}

	/* 
	 * Store field present flag
	 */
	*(uint32 *)txMsgp = FALSE;
	*txMsgLen += sizeof(int32);
	txMsgp    += sizeof(int32);

	/*
	 *Store eof flag from file system into 
	 *destination message buffer
	 */
	*(uint32 *)txMsgp = fsReaddirRespMsgp->eof;
	*txMsgLen += sizeof(int32);

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);	

}

/*-----------------------------------------------------------------
 * Name:        nfsv3_msgReaddirplusResp
 *
 * Description: Map File System Readdirplus response to a NFS V3 reply.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*
		 *NFX_OK  or NFX_ERR
         */
nfsv3_msgReaddirplusResp(int32 			status,		/*Operation result status*/
						 eee_descBuf_t	*bd,		/*file system resp. msg*/
						 char 			*txMsgp, 	/*Pointer to tx msg buf*/
						 int32 			*txMsgLen,  /*current length of msg*/
						 nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgReadDirPlusResp 	*fsReaddirplusRespMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	int32					rc;
	int32					msgLen;
	int32                   length;
	int32                   i;
	fs_DirEntryPlus			*direntPlus;
	char8					*msgBufp;
	eee_desc_t 				*eDescp;
	static char8			msgBuf[8192];

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	/*
	 *Retrieve EEE message into local buffer
	 */	
	eDescp = nfsReqp->edescp;
	msgLen =  (eDescp->hdr.control & EDESC_TOTAL_LEN);

	/*
	 *Check if message length fits in single descriptor
	 */
	if (msgLen < EEE_BUFSIZE_LARGE)
	{
		fsReaddirplusRespMsgp = (fs_MsgReadDirPlusResp *)bd->buf;
	}
	else
	{

		memset(msgBuf, 0, sizeof(msgBuf));
		msgBufp = &msgBuf[0];

		rc = eee_copyPacketToLocalBuffer(msgBufp, eDescp, msgLen, 0);
		if (rc != msgLen)
		{
			/*copied failed*/	
			panic("%s: eee_copyPacketToLocalBuffer failed!\n",
						__FUNCTION__);
		}
		fsReaddirplusRespMsgp = (fs_MsgReadDirPlusResp *)msgBufp;
	}

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.readdirplus++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.readdirplus++;
	nfs_globalCfgp->nfs_xmtPktCounters.readdirplus++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.readdirplus++;

	length = 0;

	*(uint32 *)txMsgp = status;

	txMsgp += 4;
	*txMsgLen += 4;

	/*
	 *Store post-op directory attributes
	 */
	length = nfs_storeV3Attributes( nfsReqp->reqType, 
									NFS_ATTR_AVAIL,
									&fsReaddirplusRespMsgp->attr, 
									txMsgp);
	*txMsgLen += length;
	txMsgp    += length;

	if (status != NFS_OK)
		return (NFX_OK);

	/*
	 *Store the cookie verifier
	 */
	length = nfs_storeLongLongValToBuf(txMsgp,
					fsReaddirplusRespMsgp->cookieVerif);
	*txMsgLen += length;
	txMsgp    += length;

	/*
	 *Get number of directory entries
	 */
#ifdef NFS_DBG_PRINTF_ON
	printf("%s: numOfEntries: %d\n", 
		__FUNCTION__, fsReaddirplusRespMsgp->numEntriesThis);
#endif

	direntPlus = &fsReaddirplusRespMsgp->dirEntry[0];
	for (i = 0; i < fsReaddirplusRespMsgp->numEntriesThis; i++)
	{
		/*
		 *Retrieve fileid, name, cookie and entry attributes, 
		 *object filehandle
		 */
#ifdef NFS_DBG_PRINTF_ON
		printf("%s: %d inode: %llx cookie: %llx nameLen: %d name: %s\n",
				__FUNCTION__, i,
                direntPlus->dirEntry.inode,
                direntPlus->dirEntry.cookie,
                direntPlus->dirEntry.nameLen,
                direntPlus->dirEntry.name);
#endif

		length = nfs_storeDirEntryPlus(direntPlus, txMsgp);
		FS_READDIRPLUS_ADV_ENTRY_PTR(direntPlus);
		txMsgp    += length;
		*txMsgLen += length;
	}

	/* 
	 * Set field Present Flag
	 */
	*(uint32 *)txMsgp = FALSE;
	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);

	/*
	 *Store EOF flag from file system into 
	 *destination message buffer
	 */
	*(uint32 *)txMsgp = fsReaddirplusRespMsgp->eof;
	*txMsgLen += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);

}


/*-----------------------------------------------------------------
 * Name: 		nfsv3_msgMknodResp 
 *
 * Description: Map File System MKNOD response to a NFS V3 reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgMknodResp( int32 			status,		/*operation result status*/
					eee_descBuf_t 	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*Pointer to tx msg buffer*/
					int32 			*txMsgLen,	/*current length of message*/
					nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgMknodResp 		*fsMknodRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	fs_Attr                 *pAttr;
	int32                   length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsMknodRespMsgp = (fs_MsgMknodResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.mknod++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.mknod++;
	nfs_globalCfgp->nfs_xmtPktCounters.mknod++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.mknod++;

	length = 0;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsMknodRespMsgp->hdr.attrType,
				  			fsMknodRespMsgp->attr,
				  			fsMknodRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(int32);
	*txMsgLen += sizeof(int32);


	if (status == NFS_OK)
	{
		/*
	 	 *Store return filehandle for the newly 
	 	 *created special file into message buffer	
		 *Defect 1429: add filehandle present flag
	 	 */
		*(uint32 *)txMsgp = TRUE;
		*txMsgLen += sizeof(uint32);
		txMsgp    += sizeof(uint32);

		length = nfs_storeV3Fhandle(&fsMknodRespMsgp->fileHandle, txMsgp);
		*txMsgLen += length;
		txMsgp    += length;
	
		/*
	 	 *Store attributes for newly created 
	 	 *special file into message buffer
	 	 */
		if (attrMap & FS_POST_ATTRS)
		{
			pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
			length = nfs_storeV3Attributes( nfsReqp->reqType, 
											NFS_ATTR_AVAIL, 
											pAttr, 
											txMsgp );
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
		}
		else
		{
			/*
			 * Set attr. available flag
			 */
			*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
			length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
			printf("%s: No Post Attr\n", __FUNCTION__);
#endif
		}

		*txMsgLen += length;
		txMsgp    += length;

	}	//if (status == NFS_OK)

	/*
	 *Store parent dir pre-op/post-op (dir_wcc) 
	 *attributes into message buffer
	 */
	if (attrMap & FS_PRE_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;

	if (attrMap & FS_POST_PAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_PARENT ];
		length = nfs_storeV3Attributes(nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;


#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return ((status != NFS_OK) ?  NFX_ERR : NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:      	nfsv3_msgRenameResp	
 *
 * Description:
 * Map File System Rename response to a NFS V3 reply procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgRenameResp(int32 			status,		/*operation result status*/
					eee_descBuf_t 	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*pointer to tx msg buffer*/
					int32 			*txMsgLen,	/*current length of message*/
					nfs_request_t 	*nfsReqp )	/*reqCB*/
{
	fs_MsgRenameResp 		*fsRenameRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	fs_Attr                 *pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsRenameRespMsgp = (fs_MsgRenameResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.rename++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.rename++;
	nfs_globalCfgp->nfs_xmtPktCounters.rename++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.rename++;

	length = 0;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsRenameRespMsgp->hdr.attrType,
				  			fsRenameRespMsgp->attr,
				  			fsRenameRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;
	txMsgp    += sizeof(uint32);
	*txMsgLen += sizeof(uint32);

	/*
	 *Store the pre-op & post-op From directory (fromdir_wcc) attributes
	 */
	if (attrMap & FS_PRE_FPAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_FROM_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre From Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre From Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;
	
	if (attrMap & FS_POST_FPAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_FROM_PARENT ];
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post From Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post From Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Store the pre-op & post-op To directory (todir_wcc) attributes
	 */
	if (attrMap & FS_PRE_TPAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE_TO_PARENT ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre To Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre To Parent Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;

	if (attrMap & FS_POST_TPAR_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST_TO_PARENT ];
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post To Parent Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
		nfs_cacheUpdateAttr(&nfsReqp->toFileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post To Parent Attr\n", __FUNCTION__);
#endif
	}

    if (attrMap & FS_POST_ATTRS)
    {
        pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
		nfs_cacheUpdateAttr(&fsRenameRespMsgp->fileHandle, pAttr);
    }

	*txMsgLen += length;
	txMsgp    += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return ((status != NFS_OK) ?  NFX_ERR : NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:       	nfsv3_msgStatfsResp 
 *
 * Description: Map the file system control message to NFS V3 STATFS 
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgStatfsResp (int32 			status,		/*operation result status*/
					 eee_descBuf_t 	*bd,		/*file system resp. msg*/
					 char 			*txMsgp,	/*Ptr. to tx msg buffer*/
					 int32 			*txMsgLen,	/*current length of message*/
					 nfs_request_t 	*nfsReqp )	/*request control block*/
{
	fs_MsgStatResp 			*fsstatRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	int32					length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsstatRespMsgp = (fs_MsgStatResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.fsstat++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.fsstat++;
	nfs_globalCfgp->nfs_xmtPktCounters.fsstat++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.fsstat++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(uint32);
	*txMsgLen += sizeof(uint32);

	if (fsstatRespMsgp->hdr.attrType & FS_POST_ATTRS)
	{
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										&fsstatRespMsgp->attr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, fsstatRespMsgp->attr.size, 
			fsstatRespMsgp->attr.fsId, fsstatRespMsgp->attr.inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	if (status != NFS_OK)
	{
		return (NFX_ERR);
	}

	length = nfs_storeStatInfo(txMsgp, &fsstatRespMsgp->stat, NFS_VER3);
	*txMsgLen += length;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	nfsv3_msgPathconfResp 
 *
 * Description: Map the file system control message to NFS V3 PATHCONF 
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgPathconfResp ( int32 			status,		/*operation result status*/
						eee_descBuf_t 	*bd,		/*file system resp. msg*/
						char 			*txMsgp,	/*Pointer to tx msg buf*/
						int32 			*txMsgLen,	/*current length of msg*/
						nfs_request_t 	*nfsReqp )	/*request control block*/
{
	fs_MsgPathconfResp 		*fsPathconfRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	int32					length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsPathconfRespMsgp = (fs_MsgPathconfResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.pathconf++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.pathconf++;
	nfs_globalCfgp->nfs_xmtPktCounters.pathconf++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.pathconf++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(uint32);
	*txMsgLen += sizeof(uint32);

	if (fsPathconfRespMsgp->hdr.attrType & FS_POST_ATTRS)
	{
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										&fsPathconfRespMsgp->attr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, 
			fsPathconfRespMsgp->attr.size, 
			fsPathconfRespMsgp->attr.fsId, 
			fsPathconfRespMsgp->attr.inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	if (status != NFS_OK)
	{
		return (NFX_ERR);
	}

	/*
	 *fsPathconfRespMsgp->pathconf
	 */	
	length = nfs_storePathconf(txMsgp, &fsPathconfRespMsgp->pathconf);
	*txMsgLen += length;
	txMsgp += length;

	return (NFX_OK);	
}
	 
/*-----------------------------------------------------------------
 * Name:       	nfsv3_msgFSInfoResp 
 *
 * Description: Map the file system control message to NFS V3 FSINFO 
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgFSInfoResp(int32 			status,		/*operation result status*/
					eee_descBuf_t 	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*Pointer to tx msg buffer*/
					int32 			*txMsgLen,	/*current length of message*/
					nfs_request_t 	*nfsReqp )	/*request control block*/
{
    int32               	idx;
    void                	*pCache;
	fs_MsgFsinfoResp 		*fsInfoRespMsgp;
	nfp_clientCBPtr_t       clientCBp;
	nfs_clientInfo_t    	*clientInfop;
	int32					length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsInfoRespMsgp = (fs_MsgFsinfoResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.fsinfo++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.fsinfo++;
	nfs_globalCfgp->nfs_xmtPktCounters.fsinfo++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.fsinfo++;

	length = 0;

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;
	txMsgp    += sizeof(uint32);
	*txMsgLen += sizeof(uint32);

	if (fsInfoRespMsgp->hdr.attrType & FS_POST_ATTRS)
	{
		length = nfs_storeV3Attributes( nfsReqp->reqType, 
										NFS_ATTR_AVAIL,
										&fsInfoRespMsgp->attr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, 
			fsInfoRespMsgp->attr.size, 
			fsInfoRespMsgp->attr.fsId, 
			fsInfoRespMsgp->attr.inode); 
#endif

        pCache = nfs_cacheFindFsinfo(&nfsReqp->fileHandle, &idx );
        if (pCache == NULL)
        {
            nfs_cacheAddFsinfo( &nfsReqp->fileHandle,
                                &fsInfoRespMsgp->attr,
                                &fsInfoRespMsgp->fsinfo,
                                idx);
        }
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
	}
	
	*txMsgLen += length;
	txMsgp += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	if (status != NFS_OK)
	{
		return (NFX_ERR);
	}

	/*
	 *store filesystem fsinfo
	 */
	length = nfs_storeFsinfo(txMsgp, &fsInfoRespMsgp->fsinfo);
	*txMsgLen += length;

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:       	nfsv3_msgWriteResp 
 *
 * Description: Map File System Write response to a NFS V3 Write .
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgWriteResp (int32 			status,		/*operation result status*/
					eee_descBuf_t 	*bd,		/*file system resp. msg*/
					char 			*txMsgp,	/*Pointer to tx msg buffer*/
					int32 			*txMsgLen,	/*current length of message*/
					nfs_request_t 	*nfsReqp )	/*request control block*/
{
	fs_MsgWriteResp 		*fsWriteRespMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	fs_Attr                 *pAttr;
	uint64					verifier;
	uint32					bytesWrite;
	uint32					stableHow;
	int32                   length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsWriteRespMsgp = (fs_MsgWriteResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.write++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.write++;
	nfs_globalCfgp->nfs_xmtPktCounters.write++;	
	nfs_globalCfgp->nfs_v3XmtPktCounters.write++;

	length = 0;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsWriteRespMsgp->hdr.attrType,
				  			fsWriteRespMsgp->attr,
				  			fsWriteRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(uint32);
	*txMsgLen += sizeof(uint32);


	/*Store wcc_data
	 *Stores pre-op attributes if available
	 */
	if (attrMap & FS_PRE_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Attr\n", __FUNCTION__);
#endif
	}
	
	*txMsgLen += length;
	txMsgp    += length; 

	/*
	 *Store post-op attributes
	 */
	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
		length = nfs_storeV3Attributes(nfsReqp->reqType,
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	if (status != NFS_OK)
	{
		return (NFX_OK);
	}
	/*
	 *Fill in the COUNT 
	 *field from the file system (the number of 
	 *bytes actually get written to the file.
	 */
	//printf("nfsReqp->dataSize: %d\n", nfsReqp->dataSize);
	//bytesWrite = nfsReqp->dataSize;
	bytesWrite = fsWriteRespMsgp->count;
	length = nfs_storeLongValToBuf(txMsgp, bytesWrite);
	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Fill in commited level
	 */
	//stableHow = NFS_WRITE_FILESYNC;
	stableHow = fsWriteRespMsgp->stableHow;
	length = nfs_storeLongValToBuf(txMsgp, stableHow);
	*txMsgLen += length;
	txMsgp += length;

	/*
	 *Fill in the verifier field. 
	 *This is a cookie that the client can use to 
	 *determine whether NFS server has 
	 *change state between a call to WRITE and a 
	 *subsequent call to either WRITE or COMMIT
	 */
	verifier = 0;
	verifier = fsWriteRespMsgp->verifier;
	length = nfs_storeLongLongValToBuf(txMsgp, verifier);
	*txMsgLen += length;
	txMsgp += length;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       nfsv3_msgCommitResp 
 *
 * Description:
 * Map File System COMMIT response to a NFS V3 COMMIT procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgCommitResp (int32 			status,		/*operation result status*/
					 eee_descBuf_t	*bd,		/*file system resp. msg*/
					 char 			*txMsgp,	/*Pointer to tx msg buffer*/
					 int32 			*txMsgLen,	/*current length of message*/
					 nfs_request_t	*nfsReqp )	/*request control block*/
{
	fs_MsgCommitResp 		*fsCommitReqpMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	uint64					verifier;
	fs_Attr                 *pAttr;
	int32                   length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsCommitReqpMsgp = (fs_MsgCommitResp *)bd->buf;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.commit++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.commit++;
	nfs_globalCfgp->nfs_xmtPktCounters.commit++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.commit++;

	length = 0;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsCommitReqpMsgp->hdr.attrType,
				  			fsCommitReqpMsgp->attr,
				  			fsCommitReqpMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp += 4;
	*txMsgLen += 4;

	/*
	 *Stores pre-op attributes if available
	 */
	if (attrMap & FS_PRE_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_PRE ];
		length = nfs_storePreOpAttributes(pAttr, 
										  txMsgp, 
										  NFS_ATTR_AVAIL);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Pre Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Pre Attr\n", __FUNCTION__);
#endif
	}
	
	*txMsgLen += length;
	txMsgp    += length; 


	/*
	 *Store post-op attributes
	 */
	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
		length = nfs_storeV3Attributes(nfsReqp->reqType,
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	if (status != NFS_OK)
	{
		return (NFX_ERR);
	}

	/*
	 *fill in  verifier cookie.
	 */
	verifier = 0;
	verifier = fsCommitReqpMsgp->verifier;
	length = nfs_storeLongLongValToBuf(txMsgp, verifier);
	*txMsgLen += length;
	txMsgp += length;

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:	nfsv3_msgReadResp 
 *  
 * Description: Map File System Read response to a NFS V3 Read procedure.
 *
 * Created by:  Joseph Vu
 *
 * Date Created: 5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_msgReadResp(int32 			status,		/*operation result status*/
				  eee_descBuf_t 	*bd,		/*file system resp. msg*/
				  char 				*txMsgp,	/*Pointer to tx msg buffer*/
				  int32 			*txMsgLen,	/*current length of message*/
				  nfs_request_t 	*nfsReqp )	/*request control block*/
{
	fs_MsgReadResp 			*fsReadRespMsgp;
	nfs_clientInfo_t    	*clientInfop;
	nfp_clientCBPtr_t       clientCBp;
	pkt_desc_t				*tplXmitPktDescp;
	int32					codIdx;
	int32					pad;
	fs_Attr                 *pAttr;
	int32					length;
	uint32					attrMap;
	nfs_attrArray_t			attrArray;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	fsReadRespMsgp = (fs_MsgReadResp *)bd->buf;	

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3XmtPktCounters.read++;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsRxCounters.read++;
	nfs_globalCfgp->nfs_xmtPktCounters.read++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.read++;

	length = 0;

	/*
	 * Find ALL attributes returned from FS.
	 */
	attrMap = nfs_findAttr( fsReadRespMsgp->hdr.attrType,
				  			fsReadRespMsgp->attr,
				  			fsReadRespMsgp->numAttrs,
				  			&attrArray );

	/*
	 *Set operation status
	 */
	*(uint32 *)txMsgp = status;

	txMsgp    += sizeof(uint32);
	*txMsgLen += sizeof(uint32);

	if (attrMap & FS_POST_ATTRS)
	{
		pAttr  = attrArray.pAttr[ NFS_ATTR_IDX_POST ];
		length = nfs_storeV3Attributes(nfsReqp->reqType,
										NFS_ATTR_AVAIL,
										pAttr, 
										txMsgp);
#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: Post Attr: size: %lld fsId: %llx inode: %llx\n",
			__FUNCTION__, pAttr->size, pAttr->fsId, pAttr->inode); 
#endif

		nfs_cacheUpdateAttr(&nfsReqp->fileHandle, pAttr);
	}
	else
	{
		/*
		 * Set attr. available flag
		 */
		*(uint32 *)txMsgp = NFS_ATTR_NOT_AVAIL;
		length = sizeof(uint32);

#ifdef 	NFS_DBG_PRINTF_ON
		printf("%s: No Post Attr\n", __FUNCTION__);
#endif
	}

	*txMsgLen += length;
	txMsgp    += length;

	//Check for error status from the file system
	if (status != NFS_OK)
		return (NFX_ERR);

    *(uint32 *)txMsgp = fsReadRespMsgp->bytesRead;
	*txMsgLen += sizeof(uint32);
	txMsgp    += sizeof(uint32);

    *(uint32 *)txMsgp = fsReadRespMsgp->eof;
	*txMsgLen += sizeof(uint32);
	txMsgp    += sizeof(uint32);

    /* Account for the length in opaque < > data */

    *(uint32 *)txMsgp = fsReadRespMsgp->bytesRead;
	*txMsgLen += sizeof(uint32);

	/*
	 *save info fs Read Complete Indication
	 */	
	nfsReqp->readXmitCmpInd.fileOffset= fsReadRespMsgp->fileOffset;
	nfsReqp->readXmitCmpInd.bytesRead = fsReadRespMsgp->bytesRead;
	nfsReqp->readXmitCmpInd.fsPrivate = fsReadRespMsgp->fsPrivate;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_eeeRamAlloc2++;
	tplXmitPktDescp = nfsReqp->pktDescp;

	/*
	 * RPC requires data to be padded to 4 byte.
	 */
	pad = (fsReadRespMsgp->bytesRead & 0x03);
	if (pad)
		pad = 4 - pad;

	/*
	 *Fill in data part
	 */
	tplXmitPktDescp->padSize  = pad;
	tplXmitPktDescp->dataSize = pad +
				fsReadRespMsgp->bytesRead;	//Total number of bytes

	if (  ((fsReadRespMsgp->bytesRead <= 0) && 
		   (fsReadRespMsgp->nbrUsedCod != 0))   ||
			((fsReadRespMsgp->bytesRead > 0) &&
			(fsReadRespMsgp->nbrUsedCod > NFX_MAX_COD_MSG)) )
	{
		panic("%s: Incorrect # of CODs (%d) rcv'd from file system!\n", 
				__FUNCTION__, fsReadRespMsgp->nbrUsedCod);
	}

	tplXmitPktDescp->numOfCod = fsReadRespMsgp->nbrUsedCod;
	for (codIdx=0; codIdx < tplXmitPktDescp->numOfCod; codIdx++)
	{
		tplXmitPktDescp->codEntry[codIdx] = fsReadRespMsgp->cod[codIdx];
	}
	

#ifdef  NFS_DBG_COD
    { 
        nfx_cod_entry_t *tempCod;
        int32           cod_totalNumOfBytes;
        int32           codIdx;
        int32           dataSize;
        int32           numOfCod;

        numOfCod = fsReadRespMsgp->nbrUsedCod;
        dataSize = fsReadRespMsgp->bytesRead;
        tempCod = &fsReadRespMsgp->cod[0];;
        cod_totalNumOfBytes = 0;
        for (codIdx =0; codIdx < numOfCod; codIdx++)
        {
            cod_totalNumOfBytes +=  tempCod->length;
            tempCod++;
        }
        if (cod_totalNumOfBytes != dataSize)
        {
            tempCod = &fsReadRespMsgp->cod[0];;
            for (codIdx =0; codIdx < numOfCod; codIdx++, tempCod++)
            {
                printf("cod: %llx offset: %d length: %d\n",
                    tempCod->cod, tempCod->offset, tempCod->length);
            }
            panic("%s: bytesRead: %d numOfCod:%d cod_totalNumOfBytes: %d\n",
                    __FUNCTION__, dataSize, numOfCod, cod_totalNumOfBytes);
        }

   	} 
#endif  //NFS_DBG_COD

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif

	return (NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:nfsv3_msgErrResp
 *
 * Description:
 *  Received an invalid message type received from the file system,
 *  Set request state done to clean up request and return.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */

int32
nfsv3_msgErrResp( int32 		status,		/*operation result status*/
				  eee_descBuf_t *bd,		/*file system resp. msg*/
				  char 			*txMsgp,	/*Pointer to tx msg buffer*/
				  int32 		*txMsgLen,	/*current length of message*/
				  nfs_request_t *nfsReqp )	/*request control block*/
{
#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Enter\n", __FUNCTION__);
#endif

	nfs_globalCfgp->nfs_debugPktCounters.nfs_rxInvalidFSResp++;
 	nfsReqp->reqState = NFS_REQ_DONE;

#ifdef NFS_DBG_PRINTF_ON
	printf("%s: Exit status %d\n", __FUNCTION__, status);
#endif
	return (NFX_OK);
}

#endif  //NFP_TXRX
