/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfsv3-proc.c
 *
 * RCS:  $Id: nfsv3-proc.c,v 1.9 2002/10/30 00:53:21 shaun Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contains NFS V3 Procedures.
 *
 * Created by: 	Joseph Vu 
 *
 * Date Created:	5/28/01
 *
 * History:	8/27/02 - Shaun splitted nfs-proc.c into nfsv3-proc.c
 *					  nfsv2-proc.c
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX

#include <nfx-incl.h>
#include <exports/nfs-api.h>
#include <exports/luc-api.h>
#include "nfsv3-proc.h"
#include "nfs-cache.h"

#define	NFS_DBG_COD		1

#define	NFS_DBG_PRINTF	


#define NFS_V3_FHANDLE_LEN  (sizeof( fs_Handle ) + sizeof(uint32)) 

#define NTOHL( tmp )	(((tmp & 0xFF000000) >> 24) | 			\
						 ((tmp & 0x00FF0000) >> 8 ) | 			\
						 ((tmp & 0x0000FF00) << 8 ) | 			\
						 ((tmp & 0x000000FF) << 24))

pkt_desc_t *
nfs_tplAllocPkt(nfp_clientCBPtr_t   clientCBp,
                nfs_request_t       *nfsReqp,
                int32               usrDataLen);

/*-----------------------------------------------------------------
 * Name:		nfsv3_procRxNullReq 
 *
 * Description: Handle NFS V3 NULL Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxNullReq(nfs_request_t  *nfsReqp)
{
	uint32				txMsgLen;
	char				*txMsgp;
	nfp_clientCBPtr_t	clientCBp;
	pkt_desc_t 			*tplXmitPktDescp;
    nfs_clientInfo_t    *clientInfop;

	nfsReqp->reqType = NFS_NULL;

#ifdef NFS_USE_TIME_STAMP
    pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_REQ);
    pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_RSP);
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment NULL receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.null++;
	nfs_globalCfgp->nfs_v3RxPktCounters.null++;

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3RxPktCounters.null++;


	/*
	 *Alloc TPL packet to send
	 */
	tplXmitPktDescp = (pkt_desc_t *)nfs_tplAllocPkt(clientCBp, 
										nfsReqp, 0);

	if (tplXmitPktDescp == NULL)
	{
		//printf ("nfs_procRxNullReq: Failure to alloc TPL packet!\n");
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (NFSERR_SERVERFAULT);
	}


	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;
	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, 
								nfsReqp->rpcCallHdrp->rpc_xid,
                                RPC_REPLY_MSG_ACCEPTED, 
								RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	/*
	 *Set packet desc hdr size
	 */
	tplXmitPktDescp->hdrSize = txMsgLen;
	if (nfs_msgXmtPkt(tplXmitPktDescp, nfsReqp) != NFX_OK)
	{
		//Error:  set reqState to DONE so that
		//nfs_reqTimeoutHandler() will free nfsReqp
		nfsReqp->pktDescp = (void *)tplXmitPktDescp;
		nfsReqp->reqState = NFS_REQ_DONE;
	}

	nfs_globalCfgp->nfs_xmtPktCounters.null++;
	nfs_globalCfgp->nfs_v3XmtPktCounters.null++;

	clientInfop->nfs_v3XmtPktCounters.null++;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxGetAttrReq 
 *
 * Description: Handle NFS V3 Get Attribute Request.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxGetAttrReq(nfs_request_t *nfsReqp)
{
	int32				idx;
    void                *cachep;
    nfp_clientCBPtr_t   clientCBp;
    nfs_clientInfo_t    *clientInfop;
	fs_MsgGetAttrReq	*fsGetAttrReqp;

	nfsReqp->reqType = NFS_GETATTR;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment GetAttr receive statistic counter
	 */	
	nfs_globalCfgp->nfs_rxPktCounters.getAttr++;
	nfs_globalCfgp->nfs_v3RxPktCounters.getAttr++;

	/*
	 * Check if we can reply get attr 
	 * request from cache.
	 */
    cachep = nfs_cacheFindAttr( &nfsReqp->fileHandle, &idx);
    if (cachep)
    {
#ifdef PKT_USE_TIME_STAMP
        pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_REQ);
        pkt_setTimeStamp( nfsReqp->pTime, PKT_TIME_IDX_FS_RSP);
#endif
    
        return(nfs_cacheSendGetAttrResp( nfsReqp, cachep ));
    }


	/* Shaun 08/18/02
	 *
 	 * Allocate msg buff to FS.
 	 */
    nfsReqp->fileMsgLen= sizeof(fs_MsgGetAttrReq);
	nfsReqp->fileMsgp  = eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a Getattr procedure to File System Request message.
	 *Send request message to the File System 
	 *to retrieve the attributes 
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.getAttr++;

    /*
     *Fill in File System (FS) request header msg
     */
	fsGetAttrReqp = (fs_MsgGetAttrReq  *)nfsReqp->fileMsgp;
    nfs_msgBuildFSHdr(&fsGetAttrReqp->hdr, 
					FS_REQ_GET_ATTR, 
					FS_POST_ATTRS, 
					nfsReqp);

    /*
     *Fill in filehandle of the requested attributes
     */
    fsGetAttrReqp->fileHandle = nfsReqp->fileHandle;
   

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.getAttr++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}
	
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxSetAttrReq 
 *
 * Description: Handle NFS V3 Set Attribute Request.
 *
 *-----------------------------------------------------------------
 */
int32 	
nfsv3_procRxSetAttrReq(nfs_request_t *nfsReqp)
{
	rpcCallHdr_t		*rpcCallHdrp;
	nfp_clientCBPtr_t	clientCBp;
	uint32				guard;
	char                *nfs_msgp;
	int32				length;
	uint32				atimeType;
	uint32				mtimeType;
    fs_MsgSetAttrReq    *fsSetAttrReqp;

	/*
	 *Increment SetAttr Rx stat counter
	 */
	nfsReqp->reqType = NFS_SETATTR;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	rpcCallHdrp = nfsReqp->rpcCallHdrp;

	nfs_globalCfgp->nfs_rxPktCounters.setAttr++;
	nfs_globalCfgp->nfs_v3RxPktCounters.setAttr++;

	/* Shaun 08/18/02
	 *
 	 * Allocate msg buff to FS.
 	 */
    nfsReqp->fileMsgLen= sizeof(fs_MsgSetAttrReq);
	nfsReqp->fileMsgp  = eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	fsSetAttrReqp = (fs_MsgSetAttrReq *)nfsReqp->fileMsgp;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	/*
	 *Get filehandle length
	 */
	length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;

	length = nfs_extractSetattr(rpcCallHdrp->progVer,
								nfs_msgp, 
								&fsSetAttrReqp->attr, 
								&atimeType, 
								&mtimeType);
	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;


	length = nfs_extractLongVal(nfs_msgp, &guard);
	nfs_msgp += length;

    fsSetAttrReqp->checkCtime.secs  = 0;
    fsSetAttrReqp->checkCtime.nsecs = 0;

	if (guard)
	{
		length = nfs_extractLongVal(nfs_msgp, &fsSetAttrReqp->checkCtime.secs);
		nfs_msgp += length;
		length = nfs_extractLongVal(nfs_msgp, &fsSetAttrReqp->checkCtime.nsecs);
	}

	/*
	 *Map a SetAttr request to File System message
	 *Send request message to File System for processing
	 */
    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(  &fsSetAttrReqp->hdr,
                        FS_REQ_SET_ATTR,
                        (FS_PRE_ATTRS | FS_POST_ATTRS),
                        nfsReqp);
    /*
     *The filehandle of file/dir to set attributes
     */    
    fsSetAttrReqp->fileHandle = nfsReqp->fileHandle;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.setAttr++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}
	return (NFS_OK);

}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxLookupReq 
 *
 * Description: Handle NFS V3 Lookup requests
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxLookupReq (nfs_request_t  *nfsReqp)
{
    fs_MsgLookupReq     	*fsLookupReqp;
	nfp_clientCBPtr_t		clientCBp;
    nfs_clientInfo_t    	*clientInfop;
	char                    *nfs_msgp;
	int32					length;
	ushort16				nameLen;


	nfsReqp->reqType = NFS_LOOKUP;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.lookUp++;
	nfs_globalCfgp->nfs_v3RxPktCounters.lookUp++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	/* Shaun 08/18/02
	 *
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsLookupReqp = (fs_MsgLookupReq *)nfsReqp->fileMsgp;

	length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename
	 */
	length = nfs_extractString(nfs_msgp, 
								&fsLookupReqp->fileName[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.lookUp++;
    

	/*
	 *Map a Lookup request to File System message
	 *Send request message to File System for processing
	 */	

    nfs_msgBuildFSHdr(&fsLookupReqp->hdr, 
					FS_REQ_LOOKUP,
                    (FS_POST_PAR_ATTRS | FS_POST_ATTRS), 
					nfsReqp);
    /*
     *The handle of parent directory.
     */
    fsLookupReqp->parentHandle = nfsReqp->fileHandle;

    /*
     *Length of name in bytes
     *Name to lookup
     */
    fsLookupReqp->nameLen = nameLen;
    fsLookupReqp->fileName[nameLen] = '\0';

    nfsReqp->fileMsgLen = sizeof(fs_MsgLookupReq) + nameLen + 1;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.lookUp++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	} 
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxAccessReq 
 *
 * Description: Handles NFS V3 Access Calls.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxAccessReq(nfs_request_t *nfsReqp)
{
    int32               	idx;
    void                	*cachep;
    fs_MsgAccessReq         *fsAccessReqp;
	nfp_clientCBPtr_t       clientCBp;
    nfs_clientInfo_t    	*clientInfop;
	char                    *nfs_msgp;
	uint32					reqBits;

    nfsReqp->reqType = NFS_ACCESS;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Access receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.access++;
	nfs_globalCfgp->nfs_v3RxPktCounters.access++;

	/*
	 * Skip File handle from start of the Access request
	 */
	nfs_msgp = nfsReqp->procStart + NFS_V3_FHANDLE_LEN;

    /*
     *Retrieve accessbits
     */
    reqBits = NTOHL(*(uint32 *)nfs_msgp);

    cachep = nfs_cacheFindAttr( &nfsReqp->fileHandle, &idx);
    if (cachep)
    {
#ifdef PKT_USE_TIME_STAMP
        pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_REQ);
        pkt_setTimeStamp( nfsReqp->pTime, PKT_TIME_IDX_FS_RSP);
#endif

        return(nfs_cacheSendAccessResp( nfsReqp, cachep, reqBits ));
    }

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsAccessReqp = (fs_MsgAccessReq *)nfsReqp->fileMsgp;

	/*
	 *Retrieve accessbits
	 */
//	length = nfs_extractLongVal(nfs_msgp, &fsAccessReqp->accessBits);	

	fsAccessReqp->accessBits = NTOHL(*(uint32 *)nfs_msgp);

	/*
	 *Map a Access procedure to File System Request message.
	 *Send request message to the File System to retrieve the attributes
	 */


	clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v3RxPktCounters.access++;
    
    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsAccessReqp->hdr, FS_REQ_ACCESS,
                        FS_POST_ATTRS, nfsReqp);

    /*
     *Fill in filehandle of the object to check
     */
    fsAccessReqp->fileHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgAccessReq);

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.access++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}
	
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxReadLinkReq 
 *
 * Description: Handle NFS V3 Readlink Calls.
 *              Procedure READLINK reads the pathname stored in the 
 *              symbolic link. The pathname is an ASCII string that 
 *              is opague to the server.  
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxReadLinkReq(nfs_request_t *nfsReqp)
{
	int32				idx;
	void				*cachep;
    fs_MsgReadLinkReq   *fsReadlinkReqp;
	nfs_clientInfo_t    *clientInfop;
	nfp_clientCBPtr_t	clientCBp;
	char				*nfs_msgp;

	nfsReqp->reqType = NFS_READLINK;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Readlink receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.readLink++;
	nfs_globalCfgp->nfs_v3RxPktCounters.readLink++;

    cachep = nfs_cacheFindSymlnk(&nfsReqp->fileHandle, &idx );
    if (cachep)
    {
#ifdef PKT_USE_TIME_STAMP
        pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_REQ);
        pkt_setTimeStamp( nfsReqp->pTime, PKT_TIME_IDX_FS_RSP);
#endif

        return(nfs_cacheSendReadLinkResp( nfsReqp, cachep ));
    }

	/*
	 *Get start of the Readlink request
	 */
	nfs_msgp = nfsReqp->procStart;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    clientCBp   = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.readLink++;

    fsReadlinkReqp = (fs_MsgReadLinkReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header message
     */
    nfs_msgBuildFSHdr(&fsReadlinkReqp->hdr, 
					  FS_REQ_READLINK,
                      FS_POST_ATTRS, 
					  nfsReqp);

    /*
     *Filehandle of the requested attributes
     */
    fsReadlinkReqp->fileHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadLinkReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readLink++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;

	/*
	 *Send request to file system
	 */
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv3_procRxReadReq  
 *
 * Description: Handle NFS V3 Read Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxReadReq (nfs_request_t  *nfsReqp)
{
    fs_MsgReadReq       *readReqp;
    nfs_clientInfo_t    *clientInfop; 
	nfp_clientCBPtr_t   clientCBp;
	int32				length;
	char				*nfs_msgp;

		
	nfsReqp->reqType = NFS_READ;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.read++;
	nfs_globalCfgp->nfs_v3RxPktCounters.read++;


	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    readReqp = (fs_MsgReadReq *)nfsReqp->fileMsgp;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
                
    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Extract position offset within the file at which the read is to begin
	 */
	length = nfs_extractLongLongVal(nfs_msgp, &readReqp->fileOffset); 

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract number of bytes of data that are to be read
	 */
	length = nfs_extractLongVal(nfs_msgp, &readReqp->readLength); 

	/*
	 *Map a Read request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp   = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.read++;
    
    /*
     *Fill in File System request header message
     */
    nfs_msgBuildFSHdr(&readReqp->hdr, FS_REQ_READ,
                    FS_POST_ATTRS, nfsReqp);
    /*
     *Fill in filehandle of file to be read,
     *file offset and number of bytes to read
     */
    readReqp->fileHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.read++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:       	nfs_procRxWriteReq 
 *
 * Description: Handles NFS V3 Write Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxWriteReq(nfs_request_t  *nfsReqp)
{
    fs_MsgWriteReq      *writeReqp;
    nfs_clientInfo_t    *clientInfop;
	nfp_clientCBPtr_t	clientCBp;
	uint32			clientIpAddr;			
	int32           codIdx;
	int32			length;	
	int32			padSize;	
	uint64			fileOffset;
	uint32			numOfBytes;
	uint32			stable_how;
	char			*nfs_msgp;
	pkt_desc_t 		*descp;

	nfsReqp->reqType = NFS_WRITE;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	clientIpAddr = 0;
	clientCBp = nfsReqp->clientCBp;
	/*
	 *Increment Write receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.write++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.write++;



	/*
	 *Get start of the Write request
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;
	nfs_msgp += length;

	length = nfs_extractLongLongVal(nfs_msgp, &fileOffset);
	nfs_msgp += length;

	length = nfs_extractLongVal(nfs_msgp, &numOfBytes);
	nfs_msgp += length;

	/*
	 * stable how field
     */
	length = nfs_extractLongVal(nfs_msgp, &stable_how);
	nfs_msgp += length;

    /*
     * length field of opague data
     */
    length = nfs_extractLongVal(nfs_msgp, &numOfBytes);
    nfs_msgp += length;

    /*
     * Send write data to GMC.
     */

	descp = (pkt_desc_t      *)nfsReqp->rxPktDesc;
	if (!(descp->flags & PKT_DESC_INUSE))
	{
        nfs_globalCfgp->nfs_debugPktCounters.tpl_badDesc++;
		return (NFSERR_SERVERFAULT);
	}

	/* Shaun 09/12/02  Fix #2180
	 * Exclude NFS pad in data send to SAC.
	 */
	padSize = (4 - (numOfBytes & 0x03)) & 0x03;

#ifdef NFS_USE_TIME_STAMP
	/* Shaun 09/25/02
	 * Count RPC & NFS header & user data.
	 */
	pkt_setRxByteCount( (int32)(nfs_msgp - nfsReqp->procStart), NFS_WRITE);
	pkt_setRxByteCount( (int32)numOfBytes, NFS_WRITE);
#endif

#ifdef SBM_REV2
    nfsReqp->nbrUsedCod = luc_sndGmcData( nfsReqp->rxPktDesc, 
										  nfs_msgp, 
										  numOfBytes, 
										  fileOffset, 
										  padSize, 
										  &nfsReqp->cod[0]);
#else

    nfsReqp->nbrUsedCod = luc_sndGmcData( nfsReqp->rxPktDesc, 
										  nfs_msgp, 
										  numOfBytes, 
										  &nfsReqp->cod[0]);
#endif

	if (nfsReqp->nbrUsedCod == 0)
		return (NFSERR_SERVERFAULT);

#ifdef	NFS_DBG_COD

	if (nfs_dbgCod( nfsReqp, numOfBytes ) != NFX_OK)
		panic("%s: bad Cod\n", __FUNCTION__);

#endif	//NFS_DBG_COD

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a Write request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.write++;

    writeReqp = (fs_MsgWriteReq  *)nfsReqp->fileMsgp;

    memset(writeReqp, 0, sizeof(fs_MsgWriteReq));

    /*
     *Fill in File System header message
     */
    nfs_msgBuildFSHdr(&writeReqp->hdr, 
					FS_REQ_WRITE,
                    (FS_PRE_ATTRS | FS_POST_ATTRS),
					 nfsReqp);
    /*
     *Fill in the filehandle for the file to which
     *data to be written
     */
    writeReqp->fileHandle = nfsReqp->fileHandle;

    writeReqp->fileOffset = fileOffset;
    writeReqp->writeLength = numOfBytes;
    writeReqp->stableHow = stable_how;
    /*
     *Fill in COD for write buffer
     */
    for (codIdx =0; codIdx < nfsReqp->nbrUsedCod; codIdx++)
    {
        writeReqp->cod[codIdx].cod = nfsReqp->cod[codIdx].cod;
        writeReqp->cod[codIdx].offset = nfsReqp->cod[codIdx].offset;
        writeReqp->cod[codIdx].length = nfsReqp->cod[codIdx].length;
    }
    writeReqp->nbrUsedCod = nfsReqp->nbrUsedCod;

    nfsReqp->fileMsgLen = sizeof(fs_MsgWriteReq);

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.write++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxCreateReq 
 *
 * Description: Handle NFS V3 Create Requests
 *
 *-----------------------------------------------------------------
 */
int32 
nfsv3_procRxCreateReq (nfs_request_t  *nfsReqp)
{
    nfp_clientCBPtr_t       clientCBp;
    nfs_clientInfo_t        *clientInfop;
    fs_MsgCreateReq         *fsCreateReqp;

	char				*nfs_msgp;
	int32				length;
	int32				mode;
	uint32				atimeType;
	uint32				mtimeType;

	nfsReqp->reqType = NFS_CREATE;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsCreateReqp = (fs_MsgCreateReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.create++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.create++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename
	 */
	length = nfs_extractString(nfs_msgp, 
							&fsCreateReqp->fileName[0],
							&fsCreateReqp->nameLen,
							FS_NAME_MAX);
	if (length < 0)
	{
		eee_deallocateBuffer(nfsReqp->fileMsgp);

		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

    fsCreateReqp->fileName[fsCreateReqp->nameLen] = '\0';

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;


	/*
	 * Extract creation mode field.
	 */
	length = nfs_extractLongVal(nfs_msgp, &mode); 
	nfs_msgp += length;

	fsCreateReqp->flags = mode;
	switch (mode)
	{
		case 	FS_CREATE_UNCHECKED:
		case 	FS_CREATE_GUARDED:
				/*
				 *Get initial file attributes
				 */
				length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer, 
											nfs_msgp, 
        									&fsCreateReqp->u.attr, 
											&atimeType, 
											&mtimeType);
				break;
		case 	FS_CREATE_EXCLUSIVE:
				/*
				 *Create verifier
				 *Verifier is 64 bits, in fs-api.h verifer as
				 *32 bits uint32
				 */	
				length = nfs_extractLongLongVal(nfs_msgp, 
        										&fsCreateReqp->u.verifier); 
				break;					
		default:
				nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
				eee_deallocateBuffer(nfsReqp->fileMsgp);
				return (NFSERR_SERVERFAULT);
	}

	/*
	 *Map a NFS Create file request to file system control message
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.create++;



    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsCreateReqp->hdr, 
					FS_REQ_CREATE, 
					(FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS | FS_POST_ATTRS),
					nfsReqp);

    /*
     *Fill in the parentHandle
     */
    fsCreateReqp->parentHandle = nfsReqp->fileHandle;


    nfsReqp->fileMsgLen = sizeof(fs_MsgCreateReq) + fsCreateReqp->nameLen + 1;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.create++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}
	
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: 	nfsv3_procRxMkdirReq
 *
 * Description: Handle NFS V3 Mkdir requests.
 *
 *-----------------------------------------------------------------
 */
int32	
nfsv3_procRxMkdirReq(nfs_request_t  *nfsReqp)
{
    nfp_clientCBPtr_t   clientCBp;
    nfs_clientInfo_t    *clientInfop;
    fs_MsgMkdirReq      *fsMkdirReqp;
	char				*nfs_msgp;
	ushort16			nameLen;
	int32				length;
	uint32				atimeType;
	uint32				mtimeType;
	
	nfsReqp->reqType = NFS_MKDIR;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsMkdirReqp = (fs_MsgMkdirReq  *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.mkdir++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.mkdir++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;


    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Extract sub-directory name
	 */
	length = nfs_extractString(nfs_msgp, 
								&fsMkdirReqp->dirName[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

    fsMkdirReqp->nameLen = nameLen;
    fsMkdirReqp->dirName[nameLen] = '\0';

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

    /*
     * Extract the initial attributes of sub-dir
     */
	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer,
								nfs_msgp, 
								&fsMkdirReqp->dirAttr, 
								&atimeType, 
								&mtimeType);

	nfs_msgp += length;

	/*
	 *Map a NFS MKDIR request to File System message
	 *Send request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.mkdir++;
 

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsMkdirReqp->hdr, 
					FS_REQ_MKDIR,
    				(FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS | FS_POST_ATTRS),
					nfsReqp);
    /*
     *Fill in filehandle of parent dir
     */
    fsMkdirReqp->parentHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof (fs_MsgMkdirReq) + nameLen + 1;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.mkdir++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv3_procRxSymlinkReq 
 *
 * Description: Handle NFS V3 Symlink Requests.
 *
 *-----------------------------------------------------------------
 */
int32	
nfsv3_procRxSymlinkReq(nfs_request_t  *nfsReqp)
{
    fs_MsgSymLinkReq    *fsSymlinkReqp;
    char                *fsLinkNamep;
    char                *fsLinkToNamep;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;

	char				*nfs_msgp;
	int32				length;
	fs_Attr				fsAttr;
	char				linkName[FS_NAME_MAX];
	ushort16			linkNameLen;
	char				linkText[FS_PATH_MAX];
	ushort16			linkTextLen;
	uint32				atimeType;
	uint32				mtimeType;

	nfsReqp->reqType = NFS_SYMLINK;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Symlink Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.symlink++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.symlink++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get name string that is to be associated 
	 *with the created symbolic link 
	 */
//	memset(&linkName[0], 0, sizeof(linkName));
	length = nfs_extractString(	nfs_msgp,
								&linkName[0],
								&linkNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get initial attributes of the symbolic link
	 */
	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer, 
								nfs_msgp, 
								&fsAttr, 
								&atimeType, 
								&mtimeType);
	if (length == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}
	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;
	/*
	 *Get  string of symbolic data
	 */
	length = nfs_extractString( nfs_msgp, 
								&linkText[0], 
								&linkTextLen, 
								FS_PATH_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

    if (linkTextLen < 32)
    {
        memcpy(nfsReqp->fname, linkText, linkTextLen);
        nfsReqp->nameLen = linkTextLen;
    }
    else
        nfsReqp->nameLen = 0;


	/* Shaun 08/18/02
	 *
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a Symlink request to File System message
	 *Send request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.symlink++;


    fsSymlinkReqp = (fs_MsgSymLinkReq *)nfsReqp->fileMsgp;
//    memset(fsSymlinkReqp, 0, msgLen);

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsSymlinkReqp->hdr,
                    FS_REQ_SYMLINK, 
    				(FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS | FS_POST_ATTRS),
					nfsReqp);
    /*
     *Fill in filehandle of parent directory for linkName
     */
    fsSymlinkReqp->parentHandle = nfsReqp->fileHandle;

    /*
     *Set length in bytes
     */
    fsSymlinkReqp->linkNameLen = linkNameLen;
    fsSymlinkReqp->linkToNameLen = linkTextLen;

    /*
     *Fill in linkName & linkTest
     */
    fsLinkNamep = (char *)fsSymlinkReqp->linkName;
    fsLinkToNamep = fsLinkNamep + linkNameLen + 1;


    memcpy(fsLinkNamep, linkName, linkNameLen);
    fsSymlinkReqp->linkName[linkNameLen] = '\0';
    memcpy(fsLinkToNamep, linkText, linkTextLen);
    fsLinkToNamep[linkTextLen] = '\0';

    nfsReqp->fileMsgLen = sizeof(fs_MsgSymLinkReq) + 
							linkNameLen + 1 + linkTextLen + 1;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.symlink++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);

}

/*-----------------------------------------------------------------
 * Name: nfsv3_procRxMkNodReq 
 *
 * Description: Handle NFS V3 Mknod requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxMkNodReq(nfs_request_t  *nfsReqp)
{
	ushort16			nameLen;
	int32				length;
	uint32				ftype;
	uint32				atimeType;
	uint32				mtimeType;
	char				*nfs_msgp;
	nfp_clientCBPtr_t	clientCBp;
    nfs_clientInfo_t        *clientInfop;
    fs_MsgMknodReq          *fsMknodReqp;


	nfsReqp->reqType = NFS_MKNOD;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsMknodReqp = (fs_MsgMknodReq *)nfsReqp->fileMsgp;

	nfs_globalCfgp->nfs_rxPktCounters.mknod++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.mknod++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer 
	 */
	nfs_msgp += length;

	/*
	 *Extract the name with the associated with created special file
	 */
	length = nfs_extractString(nfs_msgp, 
								&fsMknodReqp->fileName[0], 
								&nameLen, 
								FS_NAME_MAX);

	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer(nfsReqp->fileMsgp);
		return (NFSERR_NAMETOL);
	}

	fsMknodReqp->nameLen = nameLen, 
    fsMknodReqp->fileName[nameLen] = '\0';

	nfs_msgp += length;

	/*
	 *Extract file type
	 */
	length = nfs_extractLongVal(nfs_msgp, &ftype);
	nfs_msgp += length;

	if ((ftype != FS_FTYPE_CHR) && ( ftype != FS_FTYPE_BLK) &&
		(ftype != FS_FTYPE_SOCK) && (ftype != FS_FTYPE_FIFO) )
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNoDev++;
		eee_deallocateBuffer(nfsReqp->fileMsgp);
		return (NFSERR_NODEV);
	}

	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer,
								nfs_msgp, 
								&fsMknodReqp->attr,
								&atimeType, 
								&mtimeType);
	nfs_msgp += length;

	if ((ftype == FS_FTYPE_CHR) || (ftype == FS_FTYPE_BLK) )
		nfs_extractDevNum(nfs_msgp, &fsMknodReqp->devNum);

	/*
	 *Map a NFS MKNOD request to a file system control message
	 */
	fsMknodReqp->attr.type = ftype;


    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.mknod++;


    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsMknodReqp->hdr,
            		FS_REQ_MKNOD,
    				(FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS | FS_POST_ATTRS),
            		nfsReqp);
    /*
     *Fill in the parentHandle of the directory
     *in which the file is to be created
     */
    fsMknodReqp->parentHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgMknodReq) + fsMknodReqp->nameLen + 1;  


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.mknod++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

#ifdef NOTUSED
/*-----------------------------------------------------------------
 * Name: nfs_procRxDelete 
 *
 * Description:
 *      Procedure REMOVE removes an entry from a directory.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_procRxDelete(nfs_request_t  *nfsReqp)
{
	int32					error;
	ushort16				nameLen;
	int32					rc;
	int32                   length;
	int32					nfs_version;
//	nfp_clientCBPtr_t       clientCBp;
	nfp_fileCBPtr_t         fileCBp;
	nfp_fileCBPtr_t         objFileCBp;
	char                    *nfs_msgp;
	char                    name[FS_NAME_MAX];
	rpcCallHdr_t			*rpcCallHdrp;
	nfs_fileInfo_t			*nfs_fileInfop;


	error = NFSERR_SERVERFAULT;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = nfs_getFhdlLen( nfs_version );

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename or sub-dir to be removed
	 */
	length = nfs_extractString(nfs_msgp,&name[0], &nameLen, FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (error);

	/*
	 *Map a Remove file request to File System message
	 *Send a request message to File System for processing
	 */
	if (nfs_version == NFS_VER2)	
	{
    	/*
     	 *Check to see if we have cache attributes for  
	 	 *parent directory
     	 */
		fileCBp = nfp_getFileCB (&nfsReqp->fileHandle);
		if (fileCBp != NULL)
		{
			nfs_fileInfop = &fileCBp->protocol.nfp_nfsFileInfo;
			if (cacheEnable  && (nfs_fileInfop->nfs_flag))
			{
				/*
			 	 *Parent directory has changed, clear cache flag
			 	 */
				nfs_fileInfop->nfs_flag = 0;
			}
		}

		if (nfsReqp->reqType == NFS_REMOVE)
			rc = nfsv2_msgRemoveReq(nfsReqp, &name[0], nameLen);
		else
			rc = nfsv2_msgRmdirReq(nfsReqp, &name[0], nameLen);
	}
	else
	{
		if (nfsReqp->reqType == NFS_REMOVE)
			rc = nfsv3_msgRemoveReq(nfsReqp, &name[0], nameLen);
		else
			rc = nfsv3_msgRmdirReq(nfsReqp, &name[0], nameLen);
	}

	if (rc == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (error);
	}
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (error);
	}

	return (NFS_OK);
}
#endif

/*-----------------------------------------------------------------
 * Name: nfs_procRxRemoveReq 
 *
 * Description:
 *	Procedure REMOVE removes an entry from a directory.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxRemoveReq(nfs_request_t  *nfsReqp)
{
	ushort16				nameLen;
	int32                   length;
	char                    *nfs_msgp;
    fs_MsgRemoveReq     *fsRemoveReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;



	nfsReqp->reqType = NFS_REMOVE;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsRemoveReqp = (fs_MsgRemoveReq *)nfsReqp->fileMsgp;

	/*
	 *Increment stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.remove++;;
	nfs_globalCfgp->nfs_v3RxPktCounters.remove++;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.remove++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename or sub-dir to be removed
	 */
	length = nfs_extractString(nfs_msgp,
								&fsRemoveReqp->name[0], 
								&nameLen,
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_NAMETOL);
	}

	fsRemoveReqp->nameLen = nameLen;
    fsRemoveReqp->name[nameLen] = '\0';

	/*
	 *Map a Remove file request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.remove++;


    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsRemoveReqp->hdr,
            		FS_REQ_REMOVE,
    				(FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS | FS_POST_ATTRS),
            		nfsReqp);

    /*
     *Fill in filehandle for the directory from which entry is to be removed
     */
    fsRemoveReqp->parentHandle = nfsReqp->fileHandle;

    /*
     *Length name in bytes
     */

#ifdef NOTUSED
    nfsReqp->namep = (uchar8 *)eee_ramAlloc(nameLen+1);
    if (nfsReqp->namep != NULL)
    {
        memcpy(nfsReqp->namep, &fsRemoveReqp->name[0], nameLen);
        nfsReqp->nameLen = nameLen;
        nfsReqp->namep[nameLen] = '\0';
    }
#endif

    nfsReqp->fileMsgLen = sizeof(fs_MsgRemoveReq) + nameLen + 1;



	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: 		nfsv3_procRxRmdirReq 
 *
 * Description: Handle NFS V3 Rmdir Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxRmdirReq(nfs_request_t  *nfsReqp)
{
    fs_MsgRmdirReq      *fsRmdirReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	ushort16				nameLen;
	int32                   length;
	char                    *nfs_msgp;

	nfsReqp->reqType = NFS_RMDIR;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsRmdirReqp = (fs_MsgRmdirReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.rmdir++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.rmdir++;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.rmdir++;


	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename or sub-dir to be removed
	 */
	length = nfs_extractString(nfs_msgp,
								&fsRmdirReqp->name[0],
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer(nfsReqp->fileMsgp);
		return (NFSERR_NAMETOL);
	}

    fsRmdirReqp->nameLen = nameLen;
    fsRmdirReqp->name[nameLen] = '\0';

	/*
	 *Map a Remove file request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.rmdir++;
   
    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsRmdirReqp->hdr,
            		FS_REQ_RMDIR,
		            (FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS),
       		     	nfsReqp);
    /*
     *Fill in filehandle for the directory from which entry is to be removed
     */
    fsRmdirReqp->parentHandle = nfsReqp->fileHandle;

#ifdef NOTUSED
    nfsReqp->namep = (uchar8 *)eee_ramAlloc(nameLen+1);
    if (nfsReqp->namep != NULL)
    {
        memcpy(nfsReqp->namep, &fsRmdirReqp->name[0], nameLen);
        nfsReqp->nameLen = nameLen;
        nfsReqp->namep[nameLen] = '\0';
    }
#endif

    nfsReqp->fileMsgLen = sizeof(fs_MsgRmdirReq) + nameLen + 1;;


	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv3_procRxRenameReq 
 *
 * Description: Handle NFS V3 Rename Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxRenameReq(nfs_request_t  *nfsReqp)
{
    fs_MsgRenameReq         *fsRenameReqp;
    nfs_clientInfo_t        *clientInfop;
    nfp_clientCBPtr_t       clientCBp;

    char                    *fsFromNamep;
    char                    *fsToNamep;
	char				*nfs_msgp;
	char				fromName[FS_NAME_MAX];
	ushort16			fromNameLen;
	char				toName[FS_NAME_MAX];
	ushort16			toNameLen;
	int32				length;


	nfsReqp->reqType = NFS_RENAME;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	nfs_globalCfgp->nfs_rxPktCounters.rename++;;
	nfs_globalCfgp->nfs_v3RxPktCounters.rename++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract filename to be renamed
	 */
	length = nfs_extractString( nfs_msgp,
								&fromName[0], 
								&fromNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}
	nfs_msgp += length;

	/*
	 *Get  filehandle of the destination (todir) directory
	 */
	length = nfs_extractFhandle(nfsReqp->rpcCallHdrp->progVer,
								nfs_msgp, 
								&nfsReqp->toFileHandle);
	if (length == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errStaleFhandle++;
		return (NFSERR_STALE);
	}
	nfs_msgp += length;

	/*
	 *Extract new name of the object (toname)
	 */
	length = nfs_extractString( nfs_msgp,
								&toName[0], 
								&toNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}


	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a rename request to file system control message
	 *Send a request to file system for proecessing
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.rename++;

    fsRenameReqp = (fs_MsgRenameReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsRenameReqp->hdr, 
					FS_REQ_RENAME,
    				(FS_PRE_FPAR_ATTRS | FS_POST_FPAR_ATTRS |
	                 FS_PRE_TPAR_ATTRS | FS_POST_TPAR_ATTRS |
	                 FS_POST_ATTRS),
                    nfsReqp);
    /*
     *Fill in the from directory filehandle
     */
    fsRenameReqp->fromDirHandle = nfsReqp->fileHandle;

    /*
     *Fill in the to directory filehandle
     */
    fsRenameReqp->toDirHandle = nfsReqp->toFileHandle;

    fsRenameReqp->fromNameLen = fromNameLen;
    fsRenameReqp->toNameLen = toNameLen;

    /*
     *Fill in fromName and toName
     */
    fsFromNamep = (char *)fsRenameReqp->fromName;
    fsToNamep = fsFromNamep + fromNameLen + 1;  /*account for NULL terminated*/

    memcpy(fsFromNamep, fromName, fromNameLen);
    fsRenameReqp->fromName[fromNameLen] = '\0';
    memcpy(fsToNamep, toName, toNameLen);
    fsToNamep[toNameLen] = '\0';

    nfsReqp->fileMsgLen = sizeof(fs_MsgRenameReq) + fromNameLen + 1
                                                  + toNameLen + 1 ;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.rename++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv3_procRxLinkReq 
 * 
 * Description: Handle NFS V3 Link Requests
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxLinkReq(nfs_request_t  *nfsReqp)
{
	char                    *nfs_msgp;
	int32                   length;
	ushort16                linkNameLen;
    fs_MsgHardLinkReq   *fsHardLinkReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;


	nfsReqp->reqType = NFS_LINK;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsHardLinkReqp = (fs_MsgHardLinkReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Link procedure Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.link++;
	nfs_globalCfgp->nfs_v3RxPktCounters.link++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get filehandle of for the directory in 
	 *which the link is to be created
	 */
	length = nfs_extractFhandle(nfsReqp->rpcCallHdrp->progVer, 
								nfs_msgp, 
								&nfsReqp->toFileHandle);
	if (length == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errStaleFhandle++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_STALE);
	}
	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get the name to be associated with the created link
	 */
	length = nfs_extractString(nfs_msgp,
								&fsHardLinkReqp->linkName[0],
								&linkNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_NAMETOL);
	}

    fsHardLinkReqp->linkName[linkNameLen] = '\0';
    fsHardLinkReqp->nameLen = linkNameLen;

	/*
	 *Map a NFS LINK request to to file system message.
	 *Send request message to file system
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.link++;


    /*
     *Fill in File System (FS) request header msg
     */

    nfs_msgBuildFSHdr(&fsHardLinkReqp->hdr, 
					FS_REQ_HARDLINK,
    				(FS_PRE_PAR_ATTRS | FS_POST_PAR_ATTRS | FS_POST_ATTRS),
					nfsReqp);

    /*
     *Fill in filehandle for file to link to
     */
    fsHardLinkReqp->fileHandle = nfsReqp->fileHandle;

    /*
     *Fill in filehandle for the parent dir for link
     */
    fsHardLinkReqp->parentDirHandle = nfsReqp->toFileHandle;

    /*
     *Fill in linkname
     */

    nfsReqp->fileMsgLen = sizeof(fs_MsgHardLinkReq) + linkNameLen + 1;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.link++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: 		nfsv3_procRxReaddirReq 
 *
 * Description: Handle NFS V3 Readdir requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxReaddirReq(nfs_request_t  *nfsReqp)
{
    nfp_clientCBPtr_t       clientCBp;
    nfs_clientInfo_t        *clientInfop;
    fs_MsgReadDirReq        *fsReaddirReqp;

	int32					length;
	uint64					cookie;
	uint64					cookieVerf;
	uint32					replyMaxSize;
	char					*nfs_msgp;

	nfsReqp->reqType = NFS_READDIR;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.readdir++;
	nfs_globalCfgp->nfs_v3RxPktCounters.readdir++;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readdir++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;


    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Get cookie & cookie verifier.
	 *This should be set 0 on first read request.  
	 *On subsequent requests,
	 *it should be value as returned by the server
	 */

	length = nfs_extractReaddirCookie(nfsReqp->rpcCallHdrp->progVer,
									  nfs_msgp, 
									  &cookie, 
									  &cookieVerf);
	if (length == NFX_ERR)
	{
    	return (NFSERR_BAD_COOKIE);
	}
			
	nfs_msgp += length;

	/*
	 *Get count field, the max size of the 
	 *reply arguments, in bytes.
	 */
	length = nfs_extractLongVal(nfs_msgp, &replyMaxSize);

	/*
	 *Map a NFS READDIR request to file system control message
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.readdir++;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    fsReaddirReqp = (fs_MsgReadDirReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsReaddirReqp->hdr, 
					FS_REQ_READDIR, 
    				FS_POST_ATTRS,
					nfsReqp);

    /*
     *Fill in filehandle of the directory
     */
    fsReaddirReqp->fileHandle = nfsReqp->fileHandle;

    /*
     *Fill in cookie & cookie verifier
     */
    fsReaddirReqp->cookie = cookie;
    fsReaddirReqp->cookieVerif =  cookieVerf;

    /*
     *Fill in count, max size of the reply arguments, in bytes
     */
    fsReaddirReqp->maxBytes = replyMaxSize;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadDirReq);


	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:	nfs_procRxReaddirPlusReq 
 *
 * Description: Handle NFS V3 ReadirPlus Requests
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxReaddirPlusReq(nfs_request_t  *nfsReqp)
{
    nfp_clientCBPtr_t       clientCBp;
    nfs_clientInfo_t        *clientInfop;
    fs_MsgReadDirPlusReq    *fsReaddirplusReqp;

	int32					length;
	uint64					cookie;
	uint64					cookieVerf;
	uint32					replyMaxSize;
	uint32					dirCount;
	char					*nfs_msgp;


	nfsReqp->reqType = NFS_READDIRPLUS;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.readdirplus++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.readdirplus++;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readdirplus++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V3_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Get cookie & cookie verifier.
	 *This should be set 0 on first read request.  
	 *On subsequent requests,
	 *it should be value as returned by the server
	 */

	length = nfs_extractReaddirCookie(nfsReqp->rpcCallHdrp->progVer,
									nfs_msgp, 
									&cookie, 
									&cookieVerf);

	if (length == NFX_ERR)
		return (NFSERR_BAD_COOKIE);
			
	nfs_msgp += length;
	/*
	 *Get max number of bytes of directory information returned.
	 */
	length = nfs_extractLongVal(nfs_msgp, &dirCount);
	nfs_msgp += length;

	/*
	 *Get count field, the max size of the 
	 *reply arguments, in bytes.
	 */
	length = nfs_extractLongVal(nfs_msgp, &replyMaxSize);

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);


    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.readdirplus++;

    fsReaddirplusReqp = (fs_MsgReadDirPlusReq *)nfsReqp->fileMsgp;

    /*
     *Fill in file system request message header
     */
    nfs_msgBuildFSHdr(&fsReaddirplusReqp->hdr, 
					FS_REQ_READDIRPLUS,
                    FS_POST_ATTRS, 
					nfsReqp);

    /*
     *Fill in filehandle of the directory
     */
    fsReaddirplusReqp->fileHandle = nfsReqp->fileHandle;

    /*
     *Fill in cookie & cookie verifier
     */
    fsReaddirplusReqp->cookie = cookie;
    fsReaddirplusReqp->cookieVerif =  cookieVerf;

    /*
     *Fill in dirCount (the max number of bytes of directory info returned)
     */
    fsReaddirplusReqp->maxDirBytes = dirCount;
    fsReaddirplusReqp->maxBytes = replyMaxSize;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadDirPlusReq);

	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);

}

/*-----------------------------------------------------------------
 * Name:       	nfsv3_procRxStatfsReq 
 *
 * Description: Handle NFS V3 Stats FS Requests.
 *
 *-----------------------------------------------------------------
 */
int32 
nfsv3_procRxStatfsReq(nfs_request_t  *nfsReqp)
{
	int32			length;
	char			*nfs_msgp;
    fs_MsgStatReq       *fsStatfsReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;


	nfsReqp->reqType = NFS_STATFS;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Rx Stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.statfs++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.fsstat++;


	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	length = NFS_V3_FHANDLE_LEN;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a STATFS request to file system 
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.fsstat++;

    fsStatfsReqp = (fs_MsgStatReq *)nfsReqp->fileMsgp;
    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsStatfsReqp->hdr,
                    FS_REQ_STAT, 
					FS_POST_ATTRS, 
					nfsReqp);
    /*
     *Fill in filehandle of the requested attributes
     */
    fsStatfsReqp->fsHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgStatReq);



	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.fsstat++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFS_OK);	
}

/*-----------------------------------------------------------------
 * Name:       	nfsv3_procRxFsinfoReq 
 *
 * Description: Handle NFS V3 FSInfo Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxFsinfoReq(nfs_request_t  *nfsReqp)
{
	int32				idx;
	void				*cachep;
    nfp_clientCBPtr_t   clientCBp;
    nfs_clientInfo_t    *clientInfop;
    fs_MsgFsinfoReq     *fsInfoReqp;

	int32				length;
	char				*nfs_msgp;

	nfsReqp->reqType = NFS_FSINFO;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Rx Stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.fsinfo++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.fsinfo++;

    cachep = nfs_cacheFindFsinfo( &nfsReqp->fileHandle, &idx );
    if (cachep) 
    {
#ifdef PKT_USE_TIME_STAMP
        pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_REQ);
        pkt_setTimeStamp( nfsReqp->pTime, PKT_TIME_IDX_FS_RSP);
#endif

        return(nfs_cacheSendFsinfoResp( nfsReqp, cachep ));
    }


	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	length = NFS_V3_FHANDLE_LEN;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a STATFS request to file system
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.fsinfo++;

    fsInfoReqp = (fs_MsgFsinfoReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System request header msg
     */
    nfs_msgBuildFSHdr(&fsInfoReqp->hdr, FS_REQ_FSINFO, FS_POST_ATTRS, nfsReqp);

    /*
     *Fill in the filehandle
     */
    fsInfoReqp->fsHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgFsinfoReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.fsinfo++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	nfs_procRxPathConfReq 
 *
 * Description:
 *      Procedure PATHCONF- Retrieve POSIX information.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxPathConfReq(nfs_request_t  *nfsReqp)
{
	int32           	length;
	char            	*nfs_msgp;
    nfp_clientCBPtr_t   clientCBp;
    nfs_clientInfo_t    *clientInfop;
    fs_MsgPathconfReq   *fsPathconfp;



	nfsReqp->reqType = NFS_PATHCONF;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/*
	 *Increment Rx Stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.pathconf++;	
	nfs_globalCfgp->nfs_v3RxPktCounters.pathconf++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	length = NFS_V3_FHANDLE_LEN;

	/* Shaun 08/18/02
	 *
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

	/*
	 *Map a PATHCONF request to file system
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.pathconf++;

    fsPathconfp = (fs_MsgPathconfReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System request header msg
     */
    nfs_msgBuildFSHdr(&fsPathconfp->hdr, 
					FS_REQ_PATHCONF,
                    FS_POST_ATTRS, 
					nfsReqp);
    /*
     *Fill in the filehandle
     */
    fsPathconfp->fsHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgFsinfoReq);

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.pathconf++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfsv3_procRxCommitReq
 * 
 * Description: Handle NFS V3 Commit Requests.
 * 
 *-----------------------------------------------------------------
 */
int32
nfsv3_procRxCommitReq(nfs_request_t  *nfsReqp)
{
    fs_MsgCommitReq     *commitReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	int32           	length;
	char            	*nfs_msgp;

	nfsReqp->reqType = NFS_COMMIT;

#ifdef NFS_USE_TIME_STAMP
	pkt_setRxByteCount(nfsReqp->procLen, nfsReqp->reqType);
#endif

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_SERVERFAULT);

    commitReqp = (fs_MsgCommitReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Commit receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.commit++;
	nfs_globalCfgp->nfs_v3RxPktCounters.commit++;

	/*
	 *Get start of the COMMIT request
	 */
	nfs_msgp = nfsReqp->procStart;

	/*
	 * Skip file handle
	 */
	length = NFS_V3_FHANDLE_LEN;
	nfs_msgp += length;

	/*
	 *Extract file offset
	 */
	length = nfs_extractLongLongVal(nfs_msgp, &commitReqp->offset);
	nfs_msgp += length;

	length = nfs_extractLongVal(nfs_msgp, &commitReqp->count);
	nfs_msgp += length;

	/*
	 *Map a Commit Req to File System message 
	 *Send a request message to File System
	 */	

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v3RxPktCounters.commit++;

    /*
     *Fill in File System header message
     */
    nfs_msgBuildFSHdr(&commitReqp->hdr, 
					FS_REQ_COMMIT,
    				FS_PRE_ATTRS | FS_POST_ATTRS,
					nfsReqp);
    /*
     *Fill in the filehandle for the file to which
     *data to be written
     */
    commitReqp->fsHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgCommitReq);

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.commit++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_SERVERFAULT);
	}

	return (NFX_OK);
}



#endif  //NFP_TXRX
