/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:		nlm-cmd.c
 *                              
 * RCS:  $Id: nlm-cmd.c,v 1.3 2002/09/24 21:40:22 joseph Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contain NLM UI command functions on TXRX only.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/19/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef	NFP_TXRX
#include <nfx-incl.h>
#include <exports/vs-mem-api.h>
#include <exports/lock-api.h>
#include <exports/nfs-api.h>
#include "nlm-prototypes.h"
#include <exports/libc.h>

//#define NLM_DEBUG
#define NLM_DBG_FUNC_PRINTF if (nlm_traceDebugEnable) printf

int32 nlm_cmdTest(int32 argc, char *argv[]);
int32 nlm_cmdShow(int32 argc, char *argv[]);
int32 nlm_cmdTrace(int32 argc, char *argv[]);
int32 nlm_cmdHelpAll(void);
int32 nlm_cmdHelpCmd (char *pCmd);
int32 nlm_cmdClear(int32 argc, char *argv[]);

typedef struct
{
	int32	(*cmdFunc)(int32 argc, char *argv[]);
	char	*cmd;
	char	*helpTxt;
} nlm_cmdTbl_t;


nlm_cmdTbl_t nlm_cmdTable[ ] =
{
#ifdef NLM_UNIT_TEST
	{nlm_cmdTest,	"test",		"nlm test init|add1|add2|add3|add4|add5|add6|add7|tst1|tst2|tst3|tst4|tst5|tst6|tst7|tst8|rm1|rm2|rml1|rml2|show"},
#endif //NLM_UNIT_TEST
	{nlm_cmdShow,	"show",		"nlm show stats|locks" },
	{nlm_cmdTrace,	"trace",	"nlm trace all|pkt|debug|off" },
	{nlm_cmdClear,	"clear",	"nlm clear stats|locks" }
};

#define NLM_CMD_MAX	sizeof(nlm_cmdTable)/sizeof(nlm_cmdTbl_t)

/*-----------------------------------------------------------------
 * Name:		nlm_cmdHelp
 *
 * Description: Display NLM help-texts.
 *
 *              nlm help [show | trace]
 *
 * Created by:	Xiaobing Zhang 
 *
 * Date Created:	07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdHelp(int32 argc, char *argv[])
{
	if (argc < 3)
		return (nlm_cmdHelpAll());
	else
		return (nlm_cmdHelpCmd(argv[2]));
}

/*-----------------------------------------------------------------
 * Name:		nlm_cmdHelpAll
 *
 * Description: Display help-texts for all NLM commands.
 *
 * Created by:	Xiaobing Zhang 
 *
 * Date Created:	07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdHelpAll(void)
{
	int32	i;

	printf("Usage:");
	for (i=0; i < NLM_CMD_MAX; i++)
	{
		printf("\t%s\n", nlm_cmdTable[i].helpTxt);
	}
	return (0);
}

/*-----------------------------------------------------------------
 * Name:		nlm_cmdHelpCmd
 *
 * Description: Display help-texts for an individual NLM command.
 *
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdHelpCmd(char *pCmd)
{
	int32	i;

	for (i=0; i < NLM_CMD_MAX; i++)
	{
		if (strequ(pCmd, nlm_cmdTable[i].cmd))
		{
			printf("Usage:");
			printf("\t%s\n", nlm_cmdTable[i].helpTxt);
			return 0;
		}
	}
	return (nlm_cmdHelpAll());
}

/*
 *Debug command to show current/pending locks
 */
void nlm_showLocks(void)
{
#ifdef NOT_USED
	nfp_fileCBPtr_t		fileCBp;
	int32				count;
	nfp_list_t			*nodeListp;

	count = 0;
	for (fileCBp= 
		(nfp_fileCBPtr_t)nfp_globalCfgp->fileCbCfgList.listBasep;
		((count < NFP_MAX_FILE) && 
		(fileCBp != NULL)); count++, fileCBp++)
	{
		nodeListp = &fileCBp->nfp_lockList;
		if (!(fileCBp->nfp_flag & NFP_ENTRY_INUSED))
			continue;
		nlm_showFileLocks(fileCBp);
	}
#else
    printf ("NLM does not store the locks anymore\n");
    return;
#endif
}

/*-----------------------------------------------------------------
 * Name:       nlm_cmdClear
 *
 * Description: Dispatch for nlm clear commands.
 *
 *              nlm clear <stat | lock>
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:    07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdClear(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		nlm_cmdHelpCmd("clear");
		return (0);
	}

	if ( !(strncmp(argv[2], "stat", 4)))
	{
		nlm_clearStat();
		return (NFX_OK);
	}
	if ( !(strncmp(argv[2], "lock", 4)))
	{
#ifdef NOT_USED
		nlm_clearLock(0);
#else
        printf ("Command not support\n");
#endif
		return (NFX_OK);
	}

	return (nlm_cmdHelpCmd("show"));

}

/*-----------------------------------------------------------------
 * Name:		nlm_cmdShow
 *
 * Description: Dispatch for nlm show commands.
 *
 *              nlm show [stat]
 * 
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdShow(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		nlm_cmdHelpCmd("show");
		return (0);
	}

	if ( !(strncmp(argv[2], "stat", 4)))
	{
		nlm_showStat();
		return (NFX_OK);
	}
	if ( !(strncmp(argv[2], "lock", 4)))
	{
		nlm_showLocks();
		return (NFX_OK);
	}

	return (nlm_cmdHelpCmd("show"));

}


/*-----------------------------------------------------------------
 * Name:		nlm_cmdTrace
 *
 * Description: Enables or disables NLM tracing.
 *
 *              nlm trace <on|off>
 * 
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdTrace(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		nlm_cmdHelpCmd("trace");
		return 0;
	}
	if (strequ(argv[2], "all"))	
	{
		nlm_traceEnable = 1;
		nlm_traceDebugEnable = 1;
//		printf("NLM trace has been turned on.\n");
	}
	else if (strequ(argv[2], "pkt"))
	{
		nlm_traceEnable = 0;
//		printf("NLM trace has been turned off\n");
	}
	else if (strequ(argv[2], "debug"))
	{
		nlm_traceDebugEnable = 1;
	}
	else if (strequ(argv[2], "off"))
	{
		nlm_traceEnable = 0;
		nlm_traceDebugEnable = 0;
	}
	else
		nlm_cmdHelpCmd("trace");

	return 0;
}

#ifdef NLM_UNIT_TEST
/*-----------------------------------------------------------------
 * Name:		nlm_cmdTest
 *
 * Description: Unit test of nlm data structre APIs.
 *
 * Created by:	Xiaobing Zhang
 * 
 * Date Created:	07/20/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmdTest(int32 argc, char *argv[])
{
	if (argc < 3)
	{
		nlm_cmdHelpCmd("test");
		return 0;
	}
	if (strequ(argv[2], "init"))
	{
		nlm_testLockInit();
		printf("Lock unit test init.\n");
	}
	if (strequ(argv[2], "add1"))
	{
		nlm_testAddLockToCurr(50, 11);
		printf("Add a lock to current list.\n");
	}
	if (strequ(argv[2], "add2"))
	{
		nlm_testAddLockToCurr(380, 11);
		printf("Add a lock to current list.\n");
	}
	if (strequ(argv[2], "add3"))
	{
		nlm_testAddLockToCurr(500, 11);
		printf("Add a lock to current list.\n");
	}
	if (strequ(argv[2], "add4"))
	{
		nlm_testAddLockToPend(50, 11);
		printf("Add a lock to pending list.\n");
	}
	if (strequ(argv[2], "add5"))
	{
		nlm_testAddLockToPend(55, 21);
		printf("Add a lock to pending list.\n");
	}
	if (strequ(argv[2], "add6"))
	{
		nlm_testAddLockToPend(385, 11);
		printf("Add a lock to pending list.\n");
	}
	if (strequ(argv[2], "add7"))
	{
		nlm_testAddLockToPend(505, 11);
		printf("Add a lock to pending list.\n");
	}
	else if (strequ(argv[2], "tst1"))
	{
		nlm_testTstCurrLock(0);
		printf("Test a lock in current list. No special check.\n");
	}
	else if (strequ(argv[2], "tst2"))
	{
		nlm_testTstCurrLock(NLM_SEARCH_OH | NLM_SEARCH_SVID);
		printf("Test a lock in current list. Match oh & svid.\n");
	}
	else if (strequ(argv[2], "tst3"))
	{
		nlm_testTstCurrLock(NLM_SEARCH_REGION_MATCH);
		printf("Test a lock in current list. Match lock region.\n");
	}
	else if (strequ(argv[2], "tst4"))
	{
		nlm_testTstCurrLock(NLM_SEARCH_EXCLUSIVE);
		printf("Test a lock in current list. Match exclusive.\n");
	}
	else if (strequ(argv[2], "tst5"))
	{
		nlm_testTstPendLock(50, 11);
		printf("Test a lock in pending list. \n");
	}
	else if (strequ(argv[2], "tst6"))
	{
		nlm_testTstPendLock(55, 21);
		printf("Test a lock in pending list. \n");
	}
	else if (strequ(argv[2], "tst7"))
	{
		nlm_testTstPendLock(385, 11);
		printf("Test a lock in pending list. \n");
	}
	else if (strequ(argv[2], "tst8"))
	{
		nlm_testTstPendLock(505, 11);
		printf("Test a lock in pending list. \n");
	}
	else if (strequ(argv[2], "rm1"))
	{
		nlm_testRmLock(380, 11);
		printf("Remove a lock.\n");
	}
	else if (strequ(argv[2], "rm2"))
	{
		nlm_testRmLock(55, 21);
		printf("Remove a lock.\n");
	}
	else if (strequ(argv[2], "rml1"))
	{
		nlm_testRmCurrLockList();
		printf("Remove current lock list.\n");
	}
	else if (strequ(argv[2], "rml2"))
	{
		nlm_testRmPendLockList(55, 21);
		printf("Remove pending lock list.\n");
	}
	else if (strequ(argv[2], "show"))
	{
		nlm_testShowLock();
		printf("Show the lock list.\n");
	}
	else
		nlm_cmdHelpCmd("test");

	return 0;
}
#endif //NLM_UNIT_TEST

/*-----------------------------------------------------------------
 * Name: nlm_cmd
 *
 * Description: Main dispatch for NLM commands.  Called from
 *              command table.
 *
 *				nlm show stats
 *				nlm trace on|off
 *
 * Created by : Xiaobing Zhang
 *
 * Date Created: 07/19/2002
 *-----------------------------------------------------------------
 */
int32 nlm_cmd(int32 argc, char *argv[])
{
	int32	i;

	if (argc < 2)
	{
		return (nlm_cmdHelp(argc, argv));
	}

	for (i=0; i < NLM_CMD_MAX; i++)
	{
		if (strequ(argv[1], nlm_cmdTable[i].cmd))
		{
			return ((*nlm_cmdTable[i].cmdFunc)(argc, argv));
		}
	}
	return (nlm_cmdHelp(argc, argv));
}

#endif	//NFP_TXRX
