/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:		nlm-msg.c
 *
 * RCS:  $Id: nlm-msg.c,v 1.5 2002/11/06 03:35:39 joseph Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contain NLM and File System interface functions.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef	NFP_TXRX
#include <nfx-incl.h>
#include <exports/vs-mem-api.h>
#include <exports/lock-api.h>
#include <exports/nfs-api.h>
#include "nlm-prototypes.h"


#define NLM_DBG_FUNC_PRINTF if (nlm_traceDebugEnable) printf

extern void luc_panic(void);
extern void pkt_freeTimeStamp(void  *pTime);
extern void nfs_updateAttr(nfp_fileCBPtr_t fileCBp, fs_Attr *fsAttrp, 
							uint32 nfs_version);


/*-----------------------------------------------------------------
 * Name:		nlm_msgBuildFSHdr 
 *
 * Description: Build a file system request message header for message
 *              that will be sent to file system.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
void
nlm_msgBuildFSHdr(fs_MsgReqHdr *fsMsgReqHdrp,	/*File System message req hdr*/
				  uint32		reqType, 		/*Req message type*/
				  uint32		attrFlag, 		/*request attributes flag*/
				  nlm_request_t	*nlmReqp)		/*reqCB*/
{
	int32				i;
	uint32				nlm_version;
	rpcCallHdr_t		*rpcCallHdrp;
	rpc_authSysInfo_t	*authSysInfop;

	authSysInfop = NULL;
	rpcCallHdrp = nlmReqp->rpcCallHdrp;
	if (rpcCallHdrp)
		authSysInfop = &rpcCallHdrp->authFlavor.authSys;
	nlm_version = rpcCallHdrp->progVer;
	/*
	 *Filled in File System (FS) Req header msg
	 */
	fsMsgReqHdrp->requestType = reqType;
	fsMsgReqHdrp->attrType = attrFlag;
	NLM_DBG_FUNC_PRINTF("nlm_msgBuildFSHdr: authSysInfop 0x%p UID: %d GID: %d\n",
			authSysInfop, authSysInfop->userId,  authSysInfop->groupId);
    if ( ( authSysInfop == NULL) ||
        (authSysInfop && (authSysInfop->authtype != PRC_AUTH_SYS)) )
    {
        /*
         *Authentication type is not AUTH_SYS.
         */
        fsMsgReqHdrp->uid = 0;
        fsMsgReqHdrp->gid = 0;
        fsMsgReqHdrp->numGids = 0;
        for (i = 0; (i < FS_REQ_GIDS_MAX); i++)
            fsMsgReqHdrp->gids[i] = 0;
    }
	else
	{		
		fsMsgReqHdrp->uid = authSysInfop->userId;
		fsMsgReqHdrp->gid = authSysInfop->groupId;

		fsMsgReqHdrp->numGids = authSysInfop->auxGidCount;
		for (i = 0; i < authSysInfop->auxGidCount; i++)
		{
			fsMsgReqHdrp->gids[i] = authSysInfop->auxGids[i];
		}
	}	
	fsMsgReqHdrp->proto = (nlm_version == NLM_VER4)? 
								FS_PROTO_NFS_V3: FS_PROTO_NFS_V2;
	fsMsgReqHdrp->clientPrivate = (uint64)nlmReqp;
	fsMsgReqHdrp->clientId = (uint32)nlmReqp->rpcCallHdrp->rpc_xid;
	NLM_DBG_FUNC_PRINTF("nlm_msgBuildFSHdr: "
					"clientPrivate-nlmReqp: 0x%p Xid: %d UID: %d GID: %d\n", 
					nlmReqp, fsMsgReqHdrp->clientId, fsMsgReqHdrp->uid, 
					fsMsgReqHdrp->gid);
	return;
}

/*-----------------------------------------------------------------
 * Name:		nlm_msgSendToFS
 *
 * Description: Send a control message to the file system for processing. 
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
int32	/*NFX_OK: No error. NFX_ERR: error occurred*/
nlm_msgSendToFs(nlm_request_t *nlmReqp)		/*current reqCB*/
{
	uint32				srcPort;
	uint32				destPort;
	int32				ret;
	boolean				err = FALSE;

	NLM_DBG_FUNC_PRINTF("%s: enter\n", __FUNCTION__);
	/*
	 *Set up EEE source port & dest port
	 *destPort = file system CPU
	 *srcPort = Tx CPU
	 */
	srcPort = nlm_globalCfgp->srcPort;
	destPort = nlm_globalCfgp->efsDstPort;

	/*
	 *Send message to file system
	 */
	ret = eee_sendMessage((uchar8 *)nlmReqp->fileMsgp, 
				nlmReqp->fileMsgLen, 
				destPort, srcPort);
	if (ret != NFX_OK)
	{
		NLM_DBG_FUNC_PRINTF("nlm_msgSendToFs: eee_sendMessage failed!\n");
		err = TRUE;
	}
	else
	{
		nlmReqp->reqState = NLM_REQ_SEND_TO_FS;
		//TODO
//		nlmReqp->timeSendToFsInSecs = nfs_utilGetTimeOfDayInSecs();
		nlm_globalCfgp->nlm_debugCounters.nlm_reqsXmtToFs++;
	}

	/*
	 *Free local buffer
	 */	
	if (eee_ramDealloc((uchar8 *)nlmReqp->fileMsgp) != NFX_OK )
	{
		NLM_DBG_FUNC_PRINTF("nlm_msgSendToFs: Failure to deallocate buf!\n");
		err = TRUE;
	}
	else
	{
		nlmReqp->fileMsgp = NULL;
		nlmReqp->fileMsgLen = 0;		
		nlm_globalCfgp->nlm_debugCounters.nlm_eeeDealloc++;
	}

	if (err == TRUE)
		return (NFX_ERR);
	else
		return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		nlm_msgFsGetAttrReq 
 *
 * Description: Prepare file system Getattr request control message.
 *              The purpose is to get size of the file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
int32		/*NFX_OK or NFX_ERR*/ 
nlm_msgFsGetAttrReq(nlm_request_t *nlmReqp)		/*Pointer to requestCB*/
{
	int32				msgLen;
	uchar8				*buf;
	fs_MsgGetAttrReq	*fsGetAttrReqp;
	nfp_fileCB_t		*fileCBp;

	NLM_DBG_FUNC_PRINTF("%s: enter\n", __FUNCTION__);
	if ( (!nlmReqp) || (!nlmReqp->fileCBp))
		return (NFX_ERR);

	/*
	 *Allocate buffer to send request to file system
	 */
	msgLen = sizeof(fs_MsgGetAttrReq);	
	buf = (uchar8 *)eee_ramAlloc(msgLen);
	if (buf == NULL)
		return (NFX_ERR);
	nlm_globalCfgp->nlm_debugCounters.nlm_eeeAlloc++;

	fsGetAttrReqp = (fs_MsgGetAttrReq  *)buf;	
	/*
	 *Fill in File System (FS) request header msg
	 */
	nlm_msgBuildFSHdr(&fsGetAttrReqp->hdr,
						FS_REQ_GET_ATTR, FS_POST_ATTRS, nlmReqp);
	/*
	 *Fill in filehandle of the requested attributes
	 */
	fileCBp = (nfp_fileCB_t *)nlmReqp->fileCBp;
	memcpy (&fsGetAttrReqp->fileHandle, &fileCBp->nfp_fileHandle, 
			sizeof (fs_Handle));
	
	/*
	 *Store our reqCB for FS to return in response message 
	 */
	nlmReqp->fileMsgp = (char *)buf;
	nlmReqp->fileMsgLen = msgLen;
	
	return (nlm_msgSendToFs(nlmReqp));
}

/*-----------------------------------------------------------------
 * Name:		nlm_msgFsGetAttrResp 
 *
 * Description: Process file system Getattr response message.
 *              The purpose is to get size of the file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
int32		/*NFX_OK or NFX_ERR*/ 
nlm_msgFsGetAttrResp(eee_descBuf_t	*bd,		/*Rile system resp. msg*/
					 nlm_request_t	*nlmReqp)	/*reqCB*/
{
	fs_MsgGetAttrResp		*fsGetAttrMsgp;
	fs_Attr					*fsAttrp;
	nfp_fileCB_t			*fileCBp;
	nfp_clientCB_t			*clientCBp;
	uint32					nlmVersion, nfsVersion;


	fsGetAttrMsgp = (fs_MsgGetAttrResp *)bd->buf;
	fsAttrp = &fsGetAttrMsgp->attr;
	NLM_DBG_FUNC_PRINTF("nlm_msgFsGetAttrResp: "
					"filesize: %lld fsId: %llx inode: %llx\n",
					 fsAttrp->size, fsAttrp->fsId, fsAttrp->inode);
	NLM_DBG_FUNC_PRINTF("nlm_msgFsGetAttrResp: "
					"filesize: %lld fsId: %llx inode: %llx\n",
					 fsAttrp->size, fsAttrp->fsId, fsAttrp->inode);

	clientCBp = (nfp_clientCB_t *)nlmReqp->clientCBp;
	nlmVersion = clientCBp->nfp_nlmClientInfo.nlm_version;
	nfsVersion = (nlmVersion == NLM_VER4)? NFS_VER3: NFS_VER2;	
	fileCBp = (nfp_fileCB_t *)nlmReqp->fileCBp;
	nfs_updateAttr(fileCBp, fsAttrp, nfsVersion);
    NLM_DBG_FUNC_PRINTF("nlm_msgFsGetAttrResp 2: "
        "filesize: %lld fsId: %llx inode: %llx\n",
         fileCBp->protocol.nfp_nfsFileInfo.fs_attr.size,
         fileCBp->protocol.nfp_nfsFileInfo.fs_attr.fsId,
         fileCBp->protocol.nfp_nfsFileInfo.fs_attr.inode);

    NLM_DBG_FUNC_PRINTF("nlm_msgFsGetAttrResp 2: "
        "filesize: %lld fsId: %llx inode: %llx\n",
         fileCBp->protocol.nfp_nfsFileInfo.fs_attr.size,
         fileCBp->protocol.nfp_nfsFileInfo.fs_attr.fsId,
         fileCBp->protocol.nfp_nfsFileInfo.fs_attr.inode);
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		nlm_msgGetFileSize
 *
 * Description: Get file size from file system Getattr response message.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
uint64		/*File size*/ 
nlm_msgGetFileSize(eee_descBuf_t	*bd)		/*Rile system resp. msg*/
{
	fs_MsgGetAttrResp		*fsGetAttrMsgp;
	fs_Attr					*fsAttrp;

	fsGetAttrMsgp = (fs_MsgGetAttrResp *)bd->buf;
	fsAttrp = &fsGetAttrMsgp->attr;

	return (fsAttrp->size);
}

/*-----------------------------------------------------------------
 * Name:	nlm_msgTestResp	
 *
 * Description: Process NLM_TEST response message after
 * received a response from the file system for the
 * size of the file.
 *
 * Created by:	Joseph Vu	
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */
int32 nlm_msgTestResp(int32 status, 
				nlm_request_t *nlmReqp)
{
	int32					rc;
	nlm_lockInfo_t			nlm_lockInfo;
	fs_Attr					*fsAttrp;
	nfp_fileCBPtr_t			fileCBp;	
#ifdef NOT_USED
	int32					listType;
	int32					searchMask;
	nlm_lockRec_t 			*lockRecp;
	nlm_lockRec_t 			*nextLockRecp;
#else
    int32                   error;
#endif

	rc = NFX_OK;
	fileCBp = nlmReqp->fileCBp;	
	if (nfp_validateFileCB(fileCBp) != NFX_OK)
	{
		//This should not happen.
		//Panic for now.  Do this during debugging
		panic("%s: Invalid FileCB: %p\n", __FUNCTION__ ,fileCBp);
		//return (NFX_ERR);

	}
	fsAttrp = &fileCBp->protocol.nfp_nfsFileInfo.fs_attr;

	//Retrieve offset & len from request
#ifdef NOT_USED
	nlm_lockInfo.lockRec.nlm_l_offset = 
			nlmReqp->nlm_lockInformation.lockRec.nlm_l_offset;
	nlm_lockInfo.lockRec.nlm_l_len = 
			nlmReqp->nlm_lockInformation.lockRec.nlm_l_len;

	//Calculate lock end region
	
	NLM_GET_L_REGION_END(nlm_lockInfo.lockRec.nlm_l_end, 
				nlm_lockInfo.lockRec.nlm_l_offset, 
				nlm_lockInfo.lockRec.nlm_l_len,
				nlm_lockInfo.lockRec.nlm_l_eof);			
#else
	nlm_lockInfo.nlm_l_offset = nlmReqp->nlm_lockInformation.nlm_l_offset;
	nlm_lockInfo.nlm_l_len = nlmReqp->nlm_lockInformation.nlm_l_len;

	//Calculate lock end region
	
	NLM_GET_L_REGION_END(nlm_lockInfo.nlm_l_end, nlm_lockInfo.nlm_l_offset, 
				nlm_lockInfo.nlm_l_len, nlm_lockInfo.nlm_l_eof);
#endif // NOT_USED

	//Check if request lock are conflict 
	//with existing lock(s)
#ifdef NOT_USED
	searchMask = 0;
	if (!nlmReqp->nlm_lockInformation.lockRec.nlm_exclusive)
		searchMask = NLM_SEARCH_EXCLUSIVE;	

	nlmReqp->lockRecp = NULL;		
	lockRecp = 	nlm_getConflictinglock(fileCBp,
									&nlm_lockInfo, 
									searchMask, 
									&nextLockRecp,
									&listType);
		
	if (lockRecp != NULL)
	{
		//conflicting lock, return DENIED 
		nlmReqp->lockRecp = lockRecp;
		rc = nlm_procTxSendErrResp(nlmReqp, NLM_ERR_DENIED,
				nlmReqp->procStart);
	}
	else
	{
		rc = nlm_procTxSendErrResp(nlmReqp, NLM_STAT_GRANTED,
			nlmReqp->procStart);
	}
#else // NOT_USED
    nlm_lockInfo.nlm_exclusive = nlmReqp->nlm_lockInformation.nlm_exclusive;
    error = nlm_procTestReq(fileCBp, &nlm_lockInfo, nlmReqp);
	rc = nlm_procTxSendErrResp(nlmReqp, error, nlmReqp->procStart);
#endif // NOT_USED
	nlm_globalCfgp->nlm_txPktCounters[NLM_TEST] += 1;

	return (rc);
}

/*-----------------------------------------------------------------
 * Name:	nlm_msgLockResp	
 *
 * Description: Process NLM_LOCK response message after
 * received a response from the file system for the
 * size of the file.
 *
 * Created by:	Joseph Vu	
 *
 * Date Created:	07/18/2002
 *
 *-----------------------------------------------------------------
 */

int32 
nlm_msgLockResp(int32 status, nlm_request_t *nlmReqp)
{
	int32					rc;
	char					*nlm_msgp;
	nlm_lockInfo_t			*nlm_lockInfop;
	fs_Attr					*fsAttrp;
//	nfp_lockCB_t 			*lockCBp;
	nfp_fileCBPtr_t			fileCBp;	
	static char				cookie[NLM_MAX_NAME];
	ushort16				cookieLen;
	int32					length;
#ifndef NOT_USED
    int32                   error, ret;
#endif


	rc = NFX_OK;
	fileCBp = nlmReqp->fileCBp;	
	if (nfp_validateFileCB(fileCBp) != NFX_OK)
	{
		//This should not happen.
		//Panic for now.  Do this during debugging
		panic("%s: Invalid FileCB: %p\n", 
						__FUNCTION__,fileCBp);
		//return (NFX_ERR);

	}
	nlm_lockInfop = &nlmReqp->nlm_lockInformation;
	fsAttrp = &fileCBp->protocol.nfp_nfsFileInfo.fs_attr;

	//Get start of NLM_TEST
	
	nlm_msgp = nlmReqp->procStart;


	//Get client's cookie 
	length = nfs_extractString(nlm_msgp, &cookie[0],
			(ushort16 *)&cookieLen, NLM_MAX_NAME);		
	if (length == NFX_ERR)
	{
		return (NFX_ERR);
	}
	cookie[cookieLen]  = '\0';		//NULL terminated

#ifdef NOT_USED
	rc = nlm_checkforConflictLock(nlmReqp,
				fileCBp, nlm_lockInfop, fsAttrp->size,
				cookie, cookieLen);
	if (rc != NFX_OK)
	{
		rc = nlm_procTxSendErrResp(nlmReqp, rc,
				nlmReqp->procStart);
		return (rc);
	}
	nlm_globalCfgp->nlm_txPktCounters[NLM_LOCK] += 1;
#else // NOT_USED
    nlmReqp->lockInfop = nlm_lockInfop;
    ret = nlm_procLockReq(nlm_lockInfop, nlmReqp);
    if (ret == LOCK_STATE_GRANTED)
    {
        /* The lock has been granted */
        error = NLM_STAT_GRANTED;
    }
    else if (ret == LOCK_STATE_DENIED)
    {
        /* The lock has been denied. This could happen for any # of reasons */
        error = (nlmReqp->block)?NLM_ERR_DEADLCK:NLM_ERR_DENIED;
    }
    else
    {
        /*
         * The lock has been put in the pending Q. We should get a call back.
         */
        error = NLM_ERR_BLOCKED;
    }
    if (nlmReqp->procNum == NLM_LOCK)
    {
        /*Response to synchronous request*/
        rc = nlm_procTxSendResp(nlmReqp,
                                error,
                                cookie, 
                                cookieLen);
	    nlm_globalCfgp->nlm_txPktCounters[NLM_LOCK] += 1;
    }
    else
    {
        /*Response to Asynchronous request*/
        rc = nlm_procTxSendASyncMsgResp(nlmReqp,
                                        error,
                                        cookie, 
                                        cookieLen);
	    nlm_globalCfgp->nlm_txPktCounters[NLM_LOCK_RES] += 1;
    }

    return rc;
#endif // NOT_USED

	return (NFX_OK);

}


/*----------------------------------------------------------------
 * Name:		nlm_msgRcv 
 *
 * Description: This function is registered with the EEE layer to
 *              receive all response messages. 
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	07/18/2002
 *                              
 *---------------------------------------------------------------
 */
void
nlm_msgRcv(eee_desc_t *eDescp)
{
	eee_descBuf_t		*bd;
	uint32				msgType;
	int32				msgLen;
	fs_MsgRespHdr		*fsMsgRespHdrp;
	nlm_request_t		*nlmReqp;
	rpcCallHdr_t		*rpcCallHdrp;
	int32				status;
//	uint64				fileSize;
	int32				rc;

	if (eDescp == NULL)
		return;

	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
						&eDescp->bd[0] : &eDescp->bd[1];

	if (bd == NULL)
	{
		eee_freePacket(eDescp);
		return;
	}

	nlm_globalCfgp->nlm_debugCounters.nlm_rspsRcvFmFs++;

	msgType = *(uint32 *)bd->buf;
	msgLen = (eDescp->hdr.control & EDESC_TOTAL_LEN);

	fsMsgRespHdrp = (fs_MsgRespHdr *) (bd->buf);

	nlmReqp = (nlm_request_t *)fsMsgRespHdrp->clientPrivate;
	if (nfp_validateNlmReqCB(nlmReqp) != NFX_OK )
	{
		//Invalid NLM request
		nlm_globalCfgp->nlm_debugCounters.nlm_rcvInvalidFSRespErr++;
		goto nlm_msgRcvExit;
	}

	if (msgType != FS_REQ_GET_ATTR)
	{
		//Only valid NLM responses from the file system is accepted
		nlm_globalCfgp->nlm_debugCounters.nlm_rcvInvalidFSRespErr++;
		goto nlm_msgRcvErr;
	}

	rpcCallHdrp = nlmReqp->rpcCallHdrp;

	//Validate request received from the file system
	if ( (rpcCallHdrp == NULL ) ||
		(rpcCallHdrp->rpc_xid != fsMsgRespHdrp->clientId) )
	{
		/* Set reqState to DONE so that
		 * nlm_reqTimeoutHandler() will free nlmReqp
		 * Also free eee descriptor.
		 */
		nlm_globalCfgp->nlm_debugCounters.nlm_rcvInvalidFSRespErr++;
		goto nlm_msgRcvErr;
	}

	/*Check FS response status*/
	status = fsMsgRespHdrp->fsStatus;
	NLM_DBG_FUNC_PRINTF("nlm_msgRcv: FS Status: %d\n", 
					fsMsgRespHdrp->fsStatus);

	if ((status == FS_ERR_NOT_MOUNTED) ||
		(status == FS_ERR_NO_RESOURCE))
	{
		nlm_globalCfgp->nlm_debugCounters.nlm_fsResourceErr++;
		goto nlm_msgRcvErr;
	}

	if (status != FS_ERR_NONE)
	{
		nlm_globalCfgp->nlm_debugCounters.nlm_fsErr++;
		goto nlm_msgRcvErr;	
	}

	/*Get file size*/
	nlm_msgFsGetAttrResp(bd, nlmReqp);
	//fileSize = nlm_msgGetFileSize(bd);

    switch (rpcCallHdrp->rpc_proc)
	{
		case NLM_TEST:
		case NLM_TEST_MSG:
			rc = nlm_msgTestResp(status, nlmReqp);
			break;
		case NLM_LOCK:
		case NLM_LOCK_MSG:
			rc = nlm_msgLockResp(status, nlmReqp);
			break;
		default:
			//Error case
			goto nlm_msgRcvErr;
	}

	if (rc == NFX_OK)
	{
		goto nlm_msgRcvExit;
	}

nlm_msgRcvErr:
	nlmReqp->reqState = NLM_REQ_DONE;
	nlm_freeAll(nlmReqp, TRUE);

nlm_msgRcvExit:
	eee_freePacket(eDescp);
	return;
}

/*----------------------------------------------------------------
 * Name:nlm_msgRpcCallHdr		
 *
 * Description: 
 * 	Stores RPC call header in the message.
 *	
 * Created by: Joseph Vu	
 *
 * Date Created:	07/18/2002
 *                              
 *---------------------------------------------------------------
 */
int32	//Return length in bytes of RPC call hdr
nlm_msgRpcCallHdr(
	char *msgp,
	uint32 cmdXid,
	uint32 rpcProgNum,
	uint32 rpcProgVer,
	uint32 rpcProcNum
)
{
	int32		hdrLen;
	uint32		*rpcCallHdrp;

	hdrLen = 0;
	rpcCallHdrp = (uint32 *)msgp;

	*rpcCallHdrp++ = cmdXid;
	*rpcCallHdrp++ = RPC_CALL;
	*rpcCallHdrp++ = RPC_VER2;
	*rpcCallHdrp++ = rpcProgNum;
	*rpcCallHdrp++ = rpcProgVer;
	*rpcCallHdrp++ = rpcProcNum;

	hdrLen = 24;

	//fill in credential field
	*rpcCallHdrp++ = RPC_AUTH_NONE;		//AUTH_NULL
	*rpcCallHdrp++ = 0;					//len=0
	hdrLen += 8;

	//fill in verifier field
	*rpcCallHdrp++ = RPC_AUTH_NONE;		//AUTH_NULL
	*rpcCallHdrp++ = 0;					//len=0
	hdrLen += 8;

	return (hdrLen);
}


#endif	//NFP_TXRX
