/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:			nsm-file.c
 *
 * RCS:  $Id: nsm-file.c,v 1.3 2002/10/18 02:52:43 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *	Contain file manipulation functions.
 *
 * Created by:		Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX

#include "nsm.h"


/*-----------------------------------------------------------------
 * Name:		nsm_fileOpen
 *
 * Description: Open file with given mode.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		// File descriptor or NFX_ERR
nsm_fileOpen(char *fullPathName, int32 mode, uint32 *count)
{
	int32   fd, rc;

	// Initialize read buffer or write buffer
	if (mode == NSM_FILE_OPENMODE_READ)
	{ 
		nsm_globalCfgp->readBuf.position = 0;
		nsm_globalCfgp->readBuf.size = 0;
	}

	if (mode == NSM_FILE_OPENMODE_WRITE)
	{ 
		nsm_globalCfgp->writeBuf.position = 0;
		nsm_globalCfgp->writeBuf.size = 0;
	}

	//NSM_PRINTF(LOG_DEBUG, ("nsm_fileOpen: fullPathName = %s", fullPathName));

	// Open file
	fd = file_open(fullPathName);

	if (fd != NFX_ERR)
	{
		// File exists. Read host count from file.
		if (mode == NSM_FILE_OPENMODE_READ)
		{
			// In read mode, read host count as well as host records
			rc = nsm_fileGetHostCount(fd, count);
			if (rc == NFX_ERR)
			{
				NSM_PRINTF(LOG_ERR, ("nsm_fileOpen: "
						"Read file (%s) failed", fullPathName));
				return (NFX_ERR);
			}
		}
		else
		{
			// In write mode, only read host count
			rc = file_read(fd, (char *)count, sizeof(uint32));
			if (rc == NFX_ERR)
			{
				NSM_PRINTF(LOG_ERR, ("nsm_fileOpen: "
						"Read count from file (%s) failed", fullPathName));
				return (NFX_ERR);
			}
		}
	}
	else
	{
		// File doesn't exists. Create one if mode is write
		if (mode == NSM_FILE_OPENMODE_WRITE)
		{
			fd = file_create(fullPathName, 0644);
			if (fd == NFX_ERR)
			{
				NSM_PRINTF(LOG_ERR, ("nsm_fileOpen: "
								"Create file (%s) failed", fullPathName));
				return (NFX_ERR);
			}
			*count = 0;

			// Save room for host count
			rc = file_write(fd, (char *)count, sizeof(uint32));
			if (rc == NFX_ERR)
			{
				NSM_PRINTF(LOG_ERR, ("nsm_fileOpen: File write failed"));
				return (NFX_ERR);
			}
		}
		else
		{
			NSM_PRINTF(LOG_ERR, ("nsm_fileOpen: "
							"Open file (%s) failed", fullPathName));
		}
	}

	return (fd);
} // nsm_fileOpen()

/*-----------------------------------------------------------------
 * Name:		nsm_fileClose
 *
 * Description: Close file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_fileClose(int32 fd, uint32 count)
{
	if (count)
	{
		// Put updated number of hosts at the beginning of file
		// The file must be opened in write mode
		int32	rc;

		rc = file_lseek(fd, FILE_SEEK_SET, 0);
		if (rc == NFX_ERR)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_fileClose: lseek failed"));
			return;
		}

		rc = file_write(fd, (char *)&count, sizeof(uint32));
		if (rc == NFX_ERR)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_fileClose: Write file failed"));
			return;
		}
	}

	file_close(fd);
} // nsm_fileClose()

/*-----------------------------------------------------------------
 * Name:		nsm_fileRead
 *
 * Description: Read a given number of bytes from file into buffer.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_fileRead(int32 fd, char *buf, int32 length)
{
	int32	rc;

	// Try to read 'length' bytes from file
	rc = file_read(fd, buf, length);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_fileRead: Read file failed"));
		return (NFX_ERR);
	}
	nsm_globalCfgp->readBuf.position = 0;
	nsm_globalCfgp->readBuf.size += rc;	//Number of bytes in the buffer

	/*
	if (nsm_traceEnable != 0)
	{
		printf("nsm_fileRead: rc = %d\n", rc);
		nsm_memDump(buf, rc);
	}
	*/

	return (NFX_OK);
} // nsm_fileRead()

/*-----------------------------------------------------------------
 * Name:		nsm_fileWrite
 *
 * Description: Flush the write buffer to file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_fileWrite(int32 fd, char *buf, int32 length)
{
	int32	rc;

	/*
	if (nsm_traceEnable != 0)
	{
		printf("nsm_fileWrite: length = %d\n", length);
		nsm_memDump(buf, length);
	}
	*/

	// Write 'length' bytes in buffer to file
	rc = file_write(fd, buf, length);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_fileWrite: Write file failed"));
		return (NFX_ERR);
	}
	nsm_globalCfgp->writeBuf.position = 0;

	return (NFX_OK);
} // nsm_fileWrite()

/*-----------------------------------------------------------------
 * Name:		nsm_fileGetHostCount
 *
 * Description: This function starts reading the lock hosts file.
 *              The first word in the file is the number of host
 *              records in the file. The function returns this number.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		//NFX_OK or NFX_ERR
nsm_fileGetHostCount(int32 fd, uint32 *countp)
{
	int32				rc;
	nsm_fileBuffer_t	*readBuf;

	readBuf = &nsm_globalCfgp->readBuf;

	//------------------------------------------------------
	// Read file. The first time the file is read after open
	//------------------------------------------------------
	rc = nsm_fileRead(fd, readBuf->buf, NSM_FILEBUF_SIZE);
	if (rc == NFX_ERR)
		return (NFX_ERR);

	if (readBuf->size == 0)
	{
		NSM_PRINTF(LOG_INFO, ("nsm_fileGetHostCount: File is empty"));
		return (NFX_ERR);
	}
	if (readBuf->size < sizeof(uint32))
	{
		NSM_PRINTF(LOG_INFO, ("nsm_fileGetHostCount: File is corrupted"));
		return (NFX_ERR);
	}

	//---------------
	// Get host count
	//---------------
	*countp = (uint32)*(readBuf->buf);
	readBuf->position += sizeof(uint32);

	return (NFX_OK);
} // nsm_fileGetHostCount()

/*-----------------------------------------------------------------
 * Name:		nsm_fileUpdateReadBuffer
 *
 * Description: Check read buffer. The buffer may contains an incomplete
 *              record, the other part of the record is in the file and
 *              hasn't been read to the buffer yet. If more bytes are 
 *              needed to read from file, do so.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_fileUpdateReadBuffer(int32 fd)
{
	int32				rc, length;
	nsm_fileBuffer_t	*readBuf = &nsm_globalCfgp->readBuf;
	char				*buf;

	if (readBuf->size < NSM_FILEBUF_SIZE)
	{
		// Reached end of file. No more record.
		return (NFX_ERR);
	}
	else
	{
		// Move the incomplete record at the end of the buffer
		// to the beginning of the buffer.
		length = readBuf->size - readBuf->position;	//Number of bytes to move
		memcpy(readBuf->buf, (readBuf->buf + readBuf->position), length);
		buf = readBuf->buf + length;
		readBuf->size = length;

		// Read more data from file, following the incomplete record
		rc = nsm_fileRead(fd, buf, (NSM_FILEBUF_SIZE - length));
		if (rc == NFX_ERR)
			return (NFX_ERR);

		if (readBuf->size == length)
		{
			// No byte is read to the buffer
			NSM_PRINTF(LOG_INFO, ("nsm_fileUpdateReadBuffer: File is corrupted"));
			return (NFX_ERR);
		}
	}

	return (NFX_OK);
} // nsm_fileUpdateReadBuffer()

/*-----------------------------------------------------------------
 * Name:		nsm_fileGetNextRecord
 *
 * Description: Get a host record stored in file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_fileGetNextRecord(int32 fd, nsm_fileRecord_t *recordp)
{
	int32				rc, length;
	nsm_fileBuffer_t	*readBuf;
	char				*currp;

	readBuf = &nsm_globalCfgp->readBuf;

	/*
	if (nsm_traceEnable != 0)
	{
		printf("nsm_fileGetNextRecord: size = %d\n",
						readBuf->size - readBuf->position);
		nsm_memDump(readBuf->buf + readBuf->position, 
						readBuf->size - readBuf->position);
	}
	*/

	//-------------------------
	// Get a record from buffer
	//-------------------------
	
	// Get IP address
	length = sizeof(uint32);
	if ((readBuf->position + length) > readBuf->size)
	{
		// Need to update read buffer
		rc = nsm_fileUpdateReadBuffer(fd);
		if (rc == NFX_ERR)
			return (NFX_ERR);
	}
	currp = readBuf->buf + readBuf->position;
	memcpy(&recordp->ipAddr, currp, length);
	readBuf->position += length;

	// Get name length 
	length = sizeof(uint32);
	if ((readBuf->position + length) > readBuf->size)
	{
		// Need to update read buffer
		rc = nsm_fileUpdateReadBuffer(fd);
		if (rc == NFX_ERR)
			return (NFX_ERR);
	}
	currp = readBuf->buf + readBuf->position;
	memcpy(&recordp->nameLen, currp, length);
	readBuf->position += length;

	// Get host name
	length = recordp->nameLen + 1;
	if ((readBuf->position + length) > readBuf->size)
	{
		// Need to update read buffer
		rc = nsm_fileUpdateReadBuffer(fd);
		if (rc == NFX_ERR)
			return (NFX_ERR);
	}
	currp = readBuf->buf + readBuf->position;
	memcpy(recordp->hostName, currp, length);
	readBuf->position += length;

	return (NFX_OK);
} // nsm_fileGetNextRecord()

/*-----------------------------------------------------------------
 * Name:		nsm_filePutRecord
 *
 * Description: Store a host record to file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/16/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_filePutRecord(int32 fd, nsm_fileRecord_t *recordp)
{
	int32				rc, length;
	nsm_fileBuffer_t	*writeBuf;
	char				*currp;

	writeBuf = &nsm_globalCfgp->writeBuf;

	// Calculate size of this record
	length = sizeof(recordp->ipAddr) +
			 sizeof(recordp->nameLen) +
			 strlen(recordp->hostName) + 1;

	if ((writeBuf->position + length) > NSM_FILEBUF_SIZE)
	{
		// Need to write the buffer to file
		rc = nsm_fileWrite(fd, writeBuf->buf, writeBuf->position);
		if (rc == NFX_ERR)
			return (NFX_ERR);
	}

	// Put the record in the write buffer

	// Put IP address
	currp = writeBuf->buf + writeBuf->position;
	length = sizeof(uint32);
	memcpy(currp, &recordp->ipAddr, length);
	writeBuf->position += length;

	// Put name length
	currp = writeBuf->buf + writeBuf->position;
	length = sizeof(uint32);
	memcpy(currp, &recordp->nameLen, length);
	writeBuf->position += length;
	
	// Put host name
	currp = writeBuf->buf + writeBuf->position;
	length = recordp->nameLen + 1;
	memcpy(currp, recordp->hostName, length);
	writeBuf->position += length;

	/*
	if (nsm_traceEnable != 0)
	{
		printf("nsm_filePutRecord: size = %d\n", writeBuf->position);
		nsm_memDump(writeBuf->buf, writeBuf->position);
	}
	*/

	return (NFX_OK);
} // nsm_filePutRecord()

/*-----------------------------------------------------------------
 * Name:		nsm_addHostToFile
 *
 * Description: Append a host record at the end of the file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/06/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_addHostToFile(char *volName, uint32 ipAddr, char *hostName)
{
	char				fullPathName[NSM_MAX_NAME + 1];
	int32				fd, rc;
	uint32				totalCount, size;
	nsm_fileRecord_t	record;

	// Get full path of lock hosts file on the volume
	// The format is volName:/filename
	strcpy(fullPathName, volName);
	strcat(fullPathName, ":/");
	strcat(fullPathName, NSM_LOCK_FILE_NAME);

	/*
	NSM_PRINTF(LOG_DEBUG, ("nsm_addHostToFile: "
			"volName = %s, hostName = %s, ipAddr = 0x%x, fullPathName = %s",
			volName, hostName, ipAddr, fullPathName));
	*/

	//----------------------------
	// Open/Create lock hosts file
	//----------------------------
	fd = nsm_fileOpen(fullPathName, NSM_FILE_OPENMODE_WRITE, &totalCount);
	if (fd == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_addHostToFile: "
								"Open/Create file (%s) failed", fullPathName));
		return (NFX_ERR);
	}
	else
	{
		//NSM_PRINTF(LOG_DEBUG, ("nsm_addHostToFile: totalCount = %d", totalCount));

		// Set offset to the end of file
		rc = file_lseek(fd, FILE_SEEK_END, 0);
		if (rc == NFX_ERR)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_addHostToFile: "
								"lseek file (%s) failed", fullPathName));
			return (NFX_ERR);
		}
	}

	//-----------------------------
	// Write the record to the file
	//-----------------------------
	record.ipAddr = ipAddr;
	record.nameLen = strlen(hostName);
	strcpy(record.hostName, hostName);
	size = sizeof(record.ipAddr) + sizeof(record.nameLen) + record.nameLen + 1;

	/*
	if (nsm_traceEnable != 0)
	{
		printf("nsm_addHostToFile: size = %d\n", size);
		nsm_memDump(&record, size);
	}
	*/

	//rc = file_write(fd, (char *)&record, size);
	rc = nsm_filePutRecord(fd, &record);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_addHostToFile: "
								"Put record to file (%s) failed", fullPathName));
		return (NFX_ERR);
	}

	// Need to write the buffer to file
	rc = nsm_fileWrite(fd,
						nsm_globalCfgp->writeBuf.buf,
						nsm_globalCfgp->writeBuf.position);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_addHostToFile: "
								"Write file (%s) failed", fullPathName));
		return (NFX_ERR);
	}

	totalCount++;
	nsm_fileClose(fd, totalCount);

	return (NFX_OK);
} // nsm_addHostToFile()

/*-----------------------------------------------------------------
 * Name:		nsm_removeHostFromFile
 *
 * Description: Remove a host record from the file.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/16/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_removeHostFromFile(char *volName, uint32 ipAddr, char *hostName)
{
	char				fullPathName[NSM_MAX_NAME + 1];
	char				fullPathNameTmp[NSM_MAX_NAME + 1];
	int32				rdFd, wrFd, rc;
	uint32				totalRdCount, totalWrCount;
	nsm_fileRecord_t	record;

	//------------------------------
	// Open lock hosts file for read
	//------------------------------

	// Get full path of lock hosts file on the volume
	// The format is volName:/filename
	strcpy(fullPathName, volName);
	strcat(fullPathName, ":/");
	strcat(fullPathName, NSM_LOCK_FILE_NAME);

	// Open the file
	rdFd = nsm_fileOpen(fullPathName, NSM_FILE_OPENMODE_READ, &totalRdCount);
	if (rdFd == NFX_ERR)
	{
		// File doesn't exist.
		NSM_PRINTF(LOG_ERR, ("nsm_removeHostFromFile: "
								"File (%s) doesn't exist", fullPathName));
		return (NFX_ERR);
	}
	NSM_PRINTF(LOG_DEBUG, ("nsm_removeHostFromFile: "
							"totalRdCount = %d", totalRdCount));

	//--------------------------------
	// Create temporary file for write
	//--------------------------------

	// Get full path of temporay lock hosts file
	strcpy(fullPathNameTmp, volName);
	strcat(fullPathNameTmp, ":/");
	strcat(fullPathNameTmp, NSM_LOCK_FILE_NAME_TMP);


	// If the temporary file exists, remove it first
	wrFd = file_open(fullPathNameTmp);
	if (wrFd != NFX_ERR)
	{
		rc = file_unlink(fullPathNameTmp);
		if (rc == NFX_ERR)
		{
			NSM_PRINTF(LOG_INFO, ("nsm_removeHostFromFile: "
								"Remove file (%s) failed", fullPathNameTmp));
			return (NFX_ERR);
		}
	}
	wrFd = nsm_fileOpen(fullPathNameTmp, NSM_FILE_OPENMODE_WRITE, &totalWrCount);
	if (wrFd == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_removeHostFromFile: "
								"Create file (%s) failed", fullPathNameTmp));
		return (NFX_ERR);
	}
	NSM_PRINTF(LOG_DEBUG, ("nsm_removeHostFromFile: "
							"totalWrCount = %d", totalWrCount));

	//----------------------------------------------------------
	// Loop through the hosts in the file to find the given host
	//----------------------------------------------------------
	while(nsm_fileGetNextRecord(rdFd, &record) == NFX_OK)
	{
		/*
		NSM_PRINTF(LOG_DEBUG, ("nsm_removeHostFromFile: "
			"record.ipAddr = 0x%x, record.nameLen = %d, record.hostName = %s",
			record.ipAddr, record.nameLen, record.hostName));
		*/

		if ((strcmp(record.hostName, hostName) == 0) &&
			(record.ipAddr == ipAddr))
		{
			// Found the record. Do nothing.
		}
		else
		{
			// Put the non-matching record back to file
			rc = nsm_filePutRecord(wrFd, &record);
			if (rc == NFX_ERR)
			{
				NSM_PRINTF(LOG_ERR, ("nsm_removeHostFromFile: "
						"Put record back to file (%s) failed", fullPathNameTmp));
				return (NFX_ERR);
			}
			totalWrCount++;
		}
	}

	if (totalWrCount != totalRdCount - 1)
	{
		// Something is wrong
		NSM_PRINTF(LOG_WARNING, ("nsm_removeHostFromFile: Host count Error"));
	}

	// Need to write the buffer to file
	rc = nsm_fileWrite(wrFd,
						nsm_globalCfgp->writeBuf.buf,
						nsm_globalCfgp->writeBuf.position);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_removeHostFromFile: "
								"Write file (%s) failed", fullPathNameTmp));
		return (NFX_ERR);
	}

	nsm_fileClose(rdFd, totalRdCount);
	nsm_fileClose(wrFd, totalWrCount);

	//--------------------------
	// Rename the temporary file
	//--------------------------
	rc = file_unlink(fullPathName);		//Remove the old file
	if (totalWrCount == 0)
	{
		// Remove the tmp file
		rc = file_unlink(fullPathNameTmp);
		return (rc);
	}
	rc = file_rename(fullPathNameTmp, fullPathName);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_removeHostFromFile: "
									"Rename file from %s to %s failed",
				   					fullPathName, fullPathNameTmp));
		return (NFX_ERR);
	}

	return (NFX_OK);
} // nsm_removeHostFromFile()

	
#endif	//NFP_TXRX
