/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:		nsm-msg.c
 *
 * RCS:  $Id: nsm-msg.c,v 1.3 2002/10/04 02:33:43 joseph Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contain interface functions between NSM and other modules.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef	NFP_TXRX

#include "nsm.h"

/*-----------------------------------------------------------------
 * Name:		nsm_msgGetVolCfgReq 
 *
 * Description: Send request to EVM to get a volume config given its ID.
 *              Input ID -1 mean we want the list of volumes.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/05/2002
 *
 *-----------------------------------------------------------------
 */
int32		//NFX_OK or NFX_ERR
nsm_msgGetVolCfgReq(evm_volId_t id)
{
	int32					rc;
	evm_MsgVolGetAttrib_t   evmGetAttrReq;

	// Build EVM Get Attr message
	memset(&evmGetAttrReq, 0, sizeof(evm_MsgVolGetAttrib_t));
	evmGetAttrReq.hdr.msgType = EVM_VOL_GET_ATTRIB;
	evmGetAttrReq.hdr.clientPrivate = nsm_globalCfgp->threadId;
	evmGetAttrReq.id = id;

	// Send messget to EVM
	rc = eee_sendMessage((uchar8 *)&evmGetAttrReq,
					 sizeof(evm_MsgVolGetAttrib_t),
					 nsm_globalCfgp->evmDstPort,
					 nsm_globalCfgp->mgmtSrcPort);
	if (rc != NFX_OK)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_msgGetVolCfgReq: "
								"Failed to send message to EVM!"));
		return (rc);
	}

	return (NFX_OK);
} // nsm_msgGetVolCfgReq()

/*-----------------------------------------------------------------
 * Name:		nsm_msgGetVolNameListReq
 *
 * Description: Send request to EVM to get a list of IDs and names
 *              of all existing volumes.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/05/2002
 *
 *-----------------------------------------------------------------
 */
int32		//NFX_OK or NFX_ERR
nsm_msgGetVolNameListReq()
{
	int32					rc;
	evm_MsgVolGetNameList_t	evmGetNameListReq;

	// Build EVM request message
	memset(&evmGetNameListReq, 0, sizeof(evm_MsgVolGetNameList_t));
	evmGetNameListReq.hdr.msgType = EVM_VOL_GET_NAME_LIST;
	evmGetNameListReq.hdr.clientPrivate = nsm_globalCfgp->threadId;

	// Send messget to EVM
	rc = eee_sendMessage((uchar8 *)&evmGetNameListReq,
					 sizeof(evm_MsgVolGetNameList_t),
					 nsm_globalCfgp->evmDstPort,
					 nsm_globalCfgp->mgmtSrcPort);
	if (rc != NFX_OK)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_msgGetVolNameListReq: "
								"Failed to send message to EVM!"));
		return (rc);
	}

	NSM_PRINTF(LOG_DEBUG, ("nsm_msgGetVolNameListReq: Message sent to EVM"));

	nsm_globalCfgp->nsm_debugCounters.nsm_reqsXmtToEvm++;

	return (NFX_OK);
} // nsm_msgGetVolNameListReq()

/*-----------------------------------------------------------------
 * Name:		nsm_msgViewFsReq
 *
 * Description: Send request to EFS to get file system state.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/05/2002
 *
 *-----------------------------------------------------------------
 */
int32		//NFX_OK or NFX_ERR
nsm_msgViewFsReq(evm_volId_t volId)
{
	int32					rc;
	fs_MsgViewFsReq			fsViewFsReq;

	// Build EFS request message
	memset(&fsViewFsReq, 0, sizeof(fs_MsgViewFsReq));
	fsViewFsReq.hdr.requestType = FS_REQ_VIEWFS;
	fsViewFsReq.hdr.len = sizeof(fs_MsgViewFsReq);
	fsViewFsReq.volId.whole_thing = volId.whole_thing;
	fsViewFsReq.type = FS_VIEW_ONE_VOLUME;
	fsViewFsReq.hdr.clientPrivate = nsm_globalCfgp->threadId;

	// Send messget to EFS
	rc = eee_sendMessage((uchar8 *)&fsViewFsReq,
					 sizeof(fs_MsgViewFsReq),
					 nsm_globalCfgp->efsDstPort,
					 nsm_globalCfgp->ctrlSrcPort);
	if (rc != NFX_OK)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_msgViewFsReq: "
								"Failed to send message to EFS!"));
		return (rc);
	}

	nsm_globalCfgp->nsm_debugCounters.nsm_reqsXmtToFs++;

	return (NFX_OK);
} // nsm_msgViewFsReq()

/*-----------------------------------------------------------------
 * Name:		nsm_msgToThread
 *
 * Description: Send request to thread to start/stop monitoring.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/05/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_msgToThread()
{
	int32					rc, size;
	nsm_msgToThreadCB_t		*msgToThreadCBp;
	nsm_event_t				*nsmEvtp;
	esm_event_t				evt;

	//----------------------------------------------
	// Get the first msgToThreadCB from waiting list
	//----------------------------------------------
	msgToThreadCBp = nsm_getMsgToThreadCB();
	if (msgToThreadCBp == NULL)
	{
		NSM_PRINTF(LOG_INFO, ("nsm_msgToThread: No message to send"));
		nsm_firstMsgToThread = TRUE;
		return (NFX_OK);
	}

	//----------------------------------------------------
	// Prepare a message to send as an event to the thread
	//----------------------------------------------------
	size = sizeof(nsm_event_t) + strlen(msgToThreadCBp->hostName) + 1;	
	NSM_SHMEM_ALLOC(nsmEvtp, nsm_event_t, size);
	if (nsmEvtp == NULL)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_msgToThread: Insufficient memory"));
		return (NFX_ERR);
	}
	//Question: Which module frees nsmEvtp allocation?
	//
	nsmEvtp->volId.whole_thing	= msgToThreadCBp->volId.whole_thing;
	nsmEvtp->nsm_clientCBp		= msgToThreadCBp->nsm_clientCBp;
	nsmEvtp->nlm_clientCBp		= msgToThreadCBp->nlm_clientCBp;
	nsmEvtp->nsm_requestCBp		= msgToThreadCBp->nsm_requestCBp;
	nsmEvtp->ipAddr				= msgToThreadCBp->hostIpAddr;
	nsmEvtp->nameLen			= strlen(msgToThreadCBp->hostName);
	strcpy(nsmEvtp->hostName, msgToThreadCBp->hostName);

	//-----------
	// Send event
	//-----------
	evt.event = msgToThreadCBp->msgType;
	evt.p1.u.p = nsmEvtp;
	evt.p2.u.p = NULL;

	rc = thread_event(nsm_globalCfgp->threadId, &evt);
	if (rc == NFX_ERR)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_msgToThread: "
							"Failed to send msg %d to thread %d\n",
							msgToThreadCBp->msgType, nsm_globalCfgp->threadId));
		nsm_globalCfgp->nsm_debugCounters.nsm_evtToThreadSendErr++;
	}
	else
	{
		NSM_PRINTF(LOG_DEBUG, ("nsm_msgToThread: "
							"Sent msg %d to thread %d\n",
							msgToThreadCBp->msgType, nsm_globalCfgp->threadId));
		nsm_globalCfgp->nsm_debugCounters.nsm_evtToThreadSend++;
	}

	//----------------------
	// Dealloc msgToThreadCB
	//----------------------
	NSM_SHMEM_DEALLOC(msgToThreadCBp);
	nsm_globalCfgp->nsm_debugCounters.freeMsgToThread++;
	nsm_firstMsgToThread = FALSE;

	return (rc);
} // nsm_msgToThread()

/*-----------------------------------------------------------------
 * Name:		nsm_msgFromThread
 *
 * Description: Send message from thread in response to start/stop 
 *              monitoring request.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/16/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_msgFromThread(evm_volId_t		volId,
				  nfp_clientCB_t	*nlm_clientCBp,	//NLM clientCB
				  nsm_requestCB_t	*nsm_requestCBp,
				  uint32			status,
				  int32				msgType)
{
	int32				rc;
	nsm_message_t		nsmMsg;

	// Build the  message
	memset(&nsmMsg, 0, sizeof(nsm_message_t));
	nsmMsg.msgType = msgType;
	nsmMsg.status = status;
	nsmMsg.nlm_clientCBp = nlm_clientCBp;
	nsmMsg.nsm_requestCBp = nsm_requestCBp;
	nsmMsg.volId.whole_thing = volId.whole_thing;
	nsmMsg.state = nsm_globalCfgp->state;

	// Send messget to NSM
	rc = eee_sendMessage((uchar8 *)&nsmMsg,
						 sizeof(nsm_message_t),
						 nsm_globalCfgp->ctrlSrcPort,
						 nsm_globalCfgp->ctrlSrcPort);
	if (rc != NFX_OK)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_msgFromThread: "
							"Failed to send msg %d from thread %d\n",
							msgType, esm_getCurrHandle()));
	}
	else
	{
		NSM_PRINTF(LOG_DEBUG, ("nsm_msgFromThread: "
							"Sent msg %d to thread %d\n",
							msgType, esm_getCurrHandle()));
		nsm_globalCfgp->nsm_debugCounters.nsm_msgFromThreadSend++;
	}

	return (rc);
} // nsm_msgFromThread()

/*-----------------------------------------------------------------
 * Name:		nsm_msgMonResp()
 *
 * Description: Process response message for monitor request
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/19/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_msgMonResp(nsm_message_t *nsmMsgp)
{
	evm_volId_t			volId;
	nfp_clientCB_t		*nlm_clientCBp;
	uint32				status;
	uint32				state;

	volId.whole_thing = nsmMsgp->volId.whole_thing;
	nlm_clientCBp = nsmMsgp->nlm_clientCBp;
	status = nsmMsgp->status;
	state = nsmMsgp->state;

	//Call NLM to inform it processing of monitor request is done
	nsm_globalCfgp->nlmCallbackFunc.monCmpInd(volId,
											  nlm_clientCBp,
											  status);

	return;
} // nsm_msgMonResp()

/*-----------------------------------------------------------------
 * Name:		nsm_msgUnmonResp()
 *
 * Description: Process response message for unmonitor request
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/19/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_msgUnmonResp(nsm_message_t *nsmMsgp)
{
	evm_volId_t			volId;
	nfp_clientCB_t		*nlm_clientCBp;
	uint32				status;
	uint32				state;

	volId.whole_thing = nsmMsgp->volId.whole_thing;
	nlm_clientCBp = nsmMsgp->nlm_clientCBp;
	status = nsmMsgp->status;
	state = nsmMsgp->state;

	//Call NLM to inform it processing of monitor request is done
	nsm_globalCfgp->nlmCallbackFunc.unmonCmpInd(volId,
												nlm_clientCBp,
												status);

	return;
} // nsm_msgUnmonResp()

/*-----------------------------------------------------------------
 * Name:		nsm_msgNotifyResp()
 *
 * Description: Process response message for notify request
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/19/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_msgNotifyResp(nsm_message_t *nsmMsgp)
{
	// Send NSM_NOTIFY response message to client
	nsm_procTxNotifyResp(nsmMsgp->nsm_requestCBp);

	return;
} // nsm_msgNotifyResp()


/*-----------------------------------------------------------------
 * Name:		nsm_msgRcv 
 *
 * Description: This function is registered with the EEE layer to
 *              receive all response messages.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/05/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_msgRcv(eee_desc_t *eDescp)
{
	eee_descBuf_t	*bd;
	uint32			msgType;
	nsm_message_t	*nsmMsgp;
	esm_event_t		evt;
	esm_handle_t	threadId = 0;
	boolean			qevent = TRUE;  // Call thread_event or not

	if (eDescp == NULL)
		return;

	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
								&(eDescp->bd[0]) : &(eDescp->bd[1]);

	if (bd == NULL)
	{
		eee_freePacket(eDescp);
		return;
	}

	msgType = *(uint32 *)bd->buf; 

	if ((msgType == NSM_MSG_MON_RESP) ||
		(msgType == NSM_MSG_UNMON_RESP) ||
		(msgType == NSM_MSG_NOTIFY_RESP))
	{
		NSM_PRINTF(LOG_INFO, ("nsm_msgRcv: msgType = %d", msgType));

		nsmMsgp = (nsm_message_t *)(bd->buf);

		nsm_globalCfgp->nsm_debugCounters.nsm_msgFromThreadRcvd++;

		// Received message from thread
		if (msgType == NSM_MSG_MON_RESP)
		{
			nsm_msgMonResp(nsmMsgp);
		}
		else if (msgType == NSM_MSG_UNMON_RESP)
		{
			nsm_msgUnmonResp(nsmMsgp);
		}
		else if (msgType == NSM_MSG_NOTIFY_RESP)
		{
			nsm_msgNotifyResp(nsmMsgp);
		}
		eee_freePacket(eDescp);

		// Send a message in the waiting list to thread
		nsm_msgToThread();
	}
	else
	{
		// Received message from other module.
		// Need to deliver event to thread

		evt.event = *(int32 *)bd->buf;

		if ((evt.event >= EVM_MSG_BASE) && (evt.event < EVM_DUMMY))
		{
			// Received from EVM
			threadId = (esm_handle_t)
						(((evm_msgHdr_t *)(bd->buf))->clientPrivate);
			nsm_globalCfgp->nsm_debugCounters.nsm_rspsRcvFmEvm++;
		}
		else if ((evt.event >= FS_MSG_BASE) && (evt.event < FS_REQ_MAX))
		{
			// Received from EFS
			threadId = (esm_handle_t)
						(((fs_MsgRespHdr *)(bd->buf))->clientPrivate);
			nsm_globalCfgp->nsm_debugCounters.nsm_rspsRcvFmFs++;
		}
		else
		{
			NSM_PRINTF(LOG_ERR, ("nsm_msgRcv: "
									"Invalid message type %d", evt.event));
			qevent = FALSE;
		}

		NSM_PRINTF(LOG_INFO, ("nsm_msgRcv: event = %d, threadId = %d", 
								evt.event, threadId));

		evt.p1.u.p = eDescp;
		evt.p2.u.p = bd->buf;
		if ((qevent == FALSE) || (thread_event(threadId, &evt) != NFX_OK))
			eee_freePacket(eDescp);
	}

	return;
}

#endif	//NFP_TXRX
