/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nsm-tpl.c
 *
 * RCS:  $Id: nsm-tpl.c,v 1.3 2002/10/17 01:38:56 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *	Contain Network Lock Manager (NSM)/Transport Layer Protocol (TPL)
 *	interface functions.
 *
 * Created by: Xiaobing Zhang
 *
 * Date Created:        08/20/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX

#include "nsm.h"

/*-----------------------------------------------------------------
 * Name:		nsm_tplAddConnInd	
 *
 * Description: Add connection indication from the transport layer to
 *	            add a new connection. 
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_tplAddConnInd(tpl_bindCb_t	*pBind,		/*ptr to Bind CB */
				  void			*connIdp,	/*ptr to newly added connection*/
				  void			**appHdlpp,	/*return ptr app. handle to TPL*/
				  uint32		raddr,		/*remote IP address*/
				  ushort16		rport)		/*remote port*/
{
	nsm_clientCB_t		*clientCBp;

	NSM_PRINTF(LOG_DEBUG, ("nsm_tplAddConnInd: "
							"connId: %p, raddr: %08x rport: %d",
							connIdp, raddr, rport));

	//-------------------------------------
	// Check if client control block exists 
	//-------------------------------------
	clientCBp = nsm_getClientCB(raddr);
	if (clientCBp == NULL)
	{
		//----------------------------------
		// Add a new NSM Client Control Block
		//----------------------------------
		clientCBp = nsm_addClientCB(raddr, NSM_MAX_NAME, NULL);
		if (clientCBp == NULL)
			return (TPL_ERR_NO_RSRC);
	}

	NSM_PRINTF(LOG_DEBUG, ("nsm_tplAddConnInd: clientCBp: %p", clientCBp));

	NSM_PRINTF(LOG_DEBUG, ("nsm_tplAddConnInd: "
						"htons(rport) = %d, clientCBp->hostPort = %d",
						htons(rport), clientCBp->hostPort));

	//if (htons(rport) == clientCBp->hostPort)
	{
		tpl_connCb_t		*pConn;

		pConn = (tpl_connCb_t *)connIdp;
		clientCBp->nsm_pBsdSock = pConn->pSock;
		NSM_PRINTF(LOG_DEBUG, ("nsm_tplAddConnInd: nsm_pBsdSock: %p",
							    clientCBp->nsm_pBsdSock));	
	}

	*appHdlpp = (void *)clientCBp;

	return (NFX_OK);
} // nsm_tplAddConnInd()

/*-----------------------------------------------------------------
 * Name:		nsm_tplDelConnInd	
 *
 * Description: Delete connection indication from the transport layer
 *              to delete an existing connection.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_tplDelConnInd(void *appHdlp, void *connIdp)
{
	NSM_PRINTF(LOG_INFO, ("nsm_tplDelConnInd: connection %p closed",
							connIdp));

	return (NFX_OK);
} // nsm_tplDelConnInd()

/*-----------------------------------------------------------------
 * Name:		nsm_tplRcvCheckRpcCallPkt 
 *
 * Description: Process NSM RPC call packet received from the 
 *              Transport Layer Protocol (TPL).
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Create:		08/21/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or internal error
nsm_tplRcvCheckRpcCallPkt(rpcCallHdr_t *rpcCallp)
{
	// Check RPC version number
	if (rpcCallp->rpc_rpcVer != RPC_VER2)
	{
		return (NFS_INT_RPC_MISMATCH);
	}

	// Check program version number
	if (rpcCallp->progVer != NSM_VER1)
	{
		return (NFS_INT_RPC_PROG_MISMATCH);
	}

	// Check procedure number
	if ((rpcCallp->rpc_proc != NSM_NULL) &&
		(rpcCallp->rpc_proc != NSM_STAT) &&
		(rpcCallp->rpc_proc != NSM_NOTIFY))
	{
		return (NFS_INT_RPC_PROC_UNAVAIL);
	}

	return (NFX_OK);
} // nsm_tplRcvCheckRpcCallPkt()

/*-----------------------------------------------------------------
 * Name:		nlm_tplRcvProcessPkt 
 *
 * Description: Process NSM packet received from the Transport Layer 
 *              Protocol (TPL).
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Create:		08/21/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_tplRcvProcessPkt(nsm_clientCB_t		*clientCBp,
					 rpcCallHdr_t		*rpcCallp,
					 pkt_desc_t			*descp)				
{
	int32					rc;
	int32					rpcHdrSize;
	nsm_requestCB_t			*reqCBp;
		
	rpcHdrSize = rpc_getHdrSize(rpcCallp);
	if (rpcHdrSize == NFX_ERR)
	{
		nsm_globalCfgp->nsm_debugCounters.nsm_rcvRpcHdrErr++;
		return (NFX_ERR);
	}

	// Increment rcv'd pkt counter
	nsm_globalCfgp->nsm_debugCounters.nsm_rxReqPktCnt[rpcCallp->rpc_proc]++;

	//-------------------------------
	// Allocate request control block
	//-------------------------------
	reqCBp = nsm_allocReqCB(clientCBp);
	if (reqCBp == NULL)
		return (NFX_ERR);

	reqCBp->rpcCallHdrp = rpcCallp;
	reqCBp->clientCBp = clientCBp;
	reqCBp->reqState = NSM_REQ_INPROG;
	reqCBp->tplConnp = descp->pTplConn;

	reqCBp->rxPktDescp = descp;
    NSM_PRINTF(LOG_DEUBG, ("nsm_tplRcvProcessPkt: "
							"reqCBp = %p, reqCBp->rpcCallHdrp = %p",
							reqCBp, reqCBp->rpcCallHdrp));

	//-----------------------------------------------------
	// Save procedure start of message ptr & message length
	//-----------------------------------------------------
	reqCBp->procStart = descp->pHdr + rpcHdrSize;
	reqCBp->procLen = descp->hdrSize - rpcHdrSize;

#ifdef NOTUSED
	//-----------------------------------
	// Add reqCB to client control block
	//-----------------------------------
    if (nfs_addCBToList(&clientCBp->clientReqList, 
				(void *)reqCBp, NFP_REQCB_NODE)  != NFX_OK)
    {
		NSM_PRINTF(LOG_ERR, ("nsm_tplRcvProcessPkt: "
							"Failure to add reqCb to client control block!"));
		NSM_SHMEM_DEALLOC(reqCBp);
		nsm_globalCfgp->nsm_debugCounters.nsm_freeReqCB++;
		return (NFX_ERR);
    }
#endif //NOTUSED

	//----------------------------
	// Call NSM protocol procedure
	//----------------------------
	rc = (*(nsm_procs[rpcCallp->rpc_proc]))(reqCBp);
	if (rc != NFX_OK)
	{
#ifdef NOTUSED
		// There is error
		if (nsm_procTxSendErrResp(reqCBp, rc) != NFX_OK)
		{
			nsm_globalCfgp->nsm_debugCounters.nsm_txRspPktCnt[rpcCallp->rpc_proc]++;
		}
#endif //NOTUSED

		// Remove reqCB
		reqCBp->reqState = NSM_REQ_DONE;
		reqCBp->rxPktDescp = NULL;	//rxPktDesc to be freed by TPL upon return
		nsm_delReqCB(reqCBp, FALSE);
	}

	return (rc);	
} // nsm_tplRcvProcessPkt()

/*-----------------------------------------------------------------
 * Name:		nsm_tplRcvPktInd 
 *
 * Description: Receive packet indication from the transport 
 *	            layer for processing.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Create:		08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32
nsm_tplRcvPktInd(void *appHdlp, pkt_desc_t *descp)
{
	int32				rc;
	rpcCallHdr_t		*rpcCallp;
	int32				rpcCallDirection;
	nsm_clientCB_t		*clientCBp;
#ifdef NOTUSED
	nsm_requestCB_t		*reqCBp;
#endif //NOTUSED

	NSM_PRINTF(LOG_DEBUG, ("nsm_tplRcvPktInd: "
				"appHdlp: %p descp: %p descp->pTplConn: %p, descp->pHdr: %p",
				appHdlp, descp, descp->pTplConn, descp->pHdr));

	if ((appHdlp == NULL) || (descp == NULL) || (descp->pHdr == NULL))
		return (TPL_FREE_PKT);

	if (!(descp->flags & PKT_DESC_INUSE))
	{
		panic("%s: bad pkt=%x flags=%x\n", __FUNCTION__,descp, descp->flags);
    }

	clientCBp = (nsm_clientCB_t *)appHdlp;
	NSM_PRINTF(LOG_DEBUG, ("nsm_tplRcvPktInd: "
							"clientCBp = %p, clientCBp->hostName = %s",
							clientCBp, clientCBp->hostName));

	rpcCallp = (rpcCallHdr_t *)descp->pHdr;
	rpcCallDirection = ntohl(rpcCallp->rpc_direction);

	NSM_PRINTF(LOG_DEBUG, ("nsm_tplRcvPktInd: RPC direction:  %s",
						(rpcCallDirection == RPC_CALL) ? "Call": "Reply"));

	if (rpcCallDirection == RPC_CALL)
	{
		//---------
		// RPC call
		//---------

		// Perform Net to Host conversion.
	    xdr_swap(descp->pHdr,
				 (sizeof(rpcCallHdr_t) - sizeof(rpc_authSysInfo_t)),
				 0);

		rc = nsm_tplRcvCheckRpcCallPkt(rpcCallp);
		if (rc != NFX_OK)
		{
			// Drop packet
			nsm_globalCfgp->nsm_debugCounters.nsm_tplDropRcvPkt++;
			return (TPL_FREE_PKT);
		}

		if (nsm_traceEnable != 0)
			nsm_decodeReqPkt(descp, NSM_DECODE_RX);

#ifdef NOTUSED
		// Check if request is in request list
		reqCBp = nsm_findReqCB(clientCBp,
								rpcCallp->rpc_xid,
								rpcCallp->rpc_proc);
		if (reqCBp != NULL)
		{
			// Found request in the list
			if (reqCBp->reqState == NSM_REQ_DONE)
			{
				// Handle retransmission
				rc = nsm_procTxPkt((pkt_desc_t *)reqCBp->txPktDescp, reqCBp);
				if (rc == NFX_OK)
				{
					reqCBp->reqState = NSM_REQ_REXMIT;
					nsm_globalCfgp->nsm_debugCounters.nsm_txRspPktCnt[rpcCallp->rpc_proc]++;
					NSM_PRINTF(LOG_INFO, ("nsm_tplRcvPktInd: rexmit Pkt!"));
				}
			}

			// Request in progress or request has been replied recently.
			// Ignore the request.

			return (TPL_FREE_PKT);
		}
		else
#endif //NOTUSED
		{
			rc = nsm_tplRcvProcessPkt(clientCBp, rpcCallp, descp);
			if (rc != NFX_OK)
			{
				NSM_PRINTF(LOG_ERR, ("nsm_tplRcvPktInd: "
										"Received Pkt Ind error!"));
				nsm_globalCfgp->nsm_debugCounters.nsm_rcvPktIndErr++;
				return (TPL_FREE_PKT);
			}
			nsm_globalCfgp->nsm_debugCounters.nsm_tplRxPktHold++;
			return (TPL_HOLD_PKT);
		}
	}
	else
	{
		//----------
		// RPC reply
		//----------
		uint32		replyXid;
		esm_event_t	evt;

		if (clientCBp->notifyReqp == NULL)
		{
			// This may happen when the last retry of notiryReq timeout
			// before this reply is received.
			NSM_PRINTF(LOG_ERR, ("nsm_tplRcvPktInd: "
									"clientCBp->notifyReqp is NULL"));
			nsm_globalCfgp->nsm_debugCounters.nsm_rcvNotifyLocateReqErr++;
			return (TPL_FREE_PKT);
		}

		NSM_PRINTF(LOG_DEBUG, ("nsm_tplRcvPktInd: clientCBp->notifyReqp = %p",
								clientCBp->notifyReqp));

		// Perform Net to Host conversion
		xdr_swap(descp->pHdr, descp->hdrSize, 0);
#ifdef	NOTUSED
		if (rc != NFX_OK)
		{
			// Drop packet
			nsm_globalCfgp->nsm_debugCounters.nsm_tplDropRcvPkt++;
			return (TPL_FREE_PKT);
		}
#endif	//NOTUSED

		// Check if it is reply to SM_NOTIFY request sent to the client
		replyXid = *(uint32 *)descp->pHdr;
		NSM_PRINTF(LOG_DEBUG, ("nsm_tplRcvPktInd: "
				"replyXid = %d, clientCBp->notifyReqp->txCmdInfo.cmdXid = %d",
			   	replyXid, clientCBp->notifyReqp->txCmdInfo.cmdXid));

		if (clientCBp->notifyReqp->txCmdInfo.cmdXid != replyXid)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_tplRcvPktInd: "
									"replyXid doesn't match reqXid"));
			nsm_globalCfgp->nsm_debugCounters.nsm_rcvNotifyLocateReqErr++;
			nsm_globalCfgp->nsm_debugCounters.nsm_tplDropRcvPkt++;
			return (TPL_FREE_PKT);
		}

		nsm_globalCfgp->nsm_debugCounters.nsm_rxNotifyRspPktCnt++;

		if (nsm_traceEnable != 0)
		{
			// Perform Host to Net conversion for notifyReq message RPC header
			xdr_swap((byte *)clientCBp->notifyReqp->rpcCallHdrp, 
							sizeof(rpcCallHdr_t), 0);
			nsm_decodeRspPkt(descp, clientCBp->notifyReqp);
		}

		// Send event to the thread that is waiting for the reply
		evt.event = NSM_MSG_NOTIFY_RESP;
		evt.p1.u.p = NULL;
		evt.p2.u.p = NULL;

		rc = thread_event(clientCBp->notifyThreadId, &evt);
		if (rc == NFX_ERR)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_tplRcvPktInd: "
							"Failed to send NSM_MSG_NOTIFY_RESP to thread %d\n",
							clientCBp->notifyThreadId));
		}
		else
		{
			NSM_PRINTF(LOG_DEBUG, ("nsm_tplRcvPktInd: "
							"Sent NSM_MSG_NOTIFY_RESP to thread %d\n",
							clientCBp->notifyThreadId));
		}

		return (TPL_FREE_PKT);
	}
} // nsm_tplRcvPktInd()

/*-----------------------------------------------------------------
 * Name:		nsm_tplXmtCmpInd       
 *
 * Description: Transmit packet complete indication from the transport layer.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32
nsm_tplXmtCmpInd(void *appHdlp, pkt_desc_t *descp)
{
	nsm_requestCB_t			*reqCBp;

	if ((descp == NULL) || (appHdlp == NULL))
	{
		panic ("%s: bad descp or app handle!\n", __FUNCTION__); 
	}
	NSM_PRINTF(LOG_DEBUG, ("nsm_tplXmtCmpInd: appHdlp: %p descp: %p",
							appHdlp, descp));
	
	nsm_globalCfgp->nsm_debugCounters.nsm_tplXmtPktCmpInd++;

	// Retrieve our NSM request from desc appInfo
	reqCBp = descp->appInfo;
	NSM_PRINTF(LOG_DEBUG, ("nsm_tplXmtCmpInd: reqCBp: %p", reqCBp));

	// Check for NULL procedure
	if ((reqCBp->rpcCallHdrp) && 
		(reqCBp->rpcCallHdrp->rpc_proc == NSM_NULL))
	{
		if (descp->tplProt == TPL_PROT_UDP)
			tpl_delConnReq(descp->pTplConn);
	}
	
#ifdef NOTUSED
	// Set reqState to Done.  We will hold the pkt for now
	// for the duplicate request cache.	
	// This pkt will be free later.
	reqCBp->reqState = NSM_REQ_DONE;

	return (TPL_HOLD_PKT);
#endif //NOTUSED

	// Remove reqCB
	reqCBp->reqState = NSM_REQ_DONE;
	reqCBp->txPktDescp = NULL;		//txPktDesc to be freed by TPL upon return
	nsm_delReqCB(reqCBp, FALSE);

	return (TPL_FREE_PKT);
} // nsm_tplXmtCmpInd()

#if defined(NFX_SMP)
/*-----------------------------------------------------------------
 * Name: nsm_tplBindReqAck
 *
 * Description:	bind request ack
 *	
 *
 * Created by:  Su Wang
 *
 * Date Created:        12/04/02
 *
 *-----------------------------------------------------------------
 */
int32 nsm_tplBindReqAck(void* pHandle, int32 rc, void *in_pBind)
{
	tpl_bindCb_t	*pBind = (tpl_bindCb_t*)in_pBind;

	if (rc != NFX_OK)
		panic("%s: tpl bind failed\n", __FUNCTION__);
	
	if (pBind->tplProt == TPL_PROT_UDP) {
		nsm_globalCfgp->nsm_pTplBind = pBind;
		nsm_globalCfgp->nsm_pBsdSock = pBind->pBsdSock;
	}
//	else
//		nsm_globalCfgp->nlm_pTplBindTcp = pBind;
	
	return NFX_OK;
}
#endif /* NFX_SMP */


/*-----------------------------------------------------------------
 * Name:		nsm_tplBindReq
 *
 * Description: Binds NSM to Transport Layer (UDP/TCP).  
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32 nsm_tplBindReq(uint32 ipaddr)
{
	int32			bindOpt;
	ushort16        protocol;
	tpl_bindCb_t    bindCB;
	tpl_bindCb_t    *bindCBp;
#if !defined(NFX_SMP)
	tpl_bindCb_t    *tplRspBindUdp;
	tpl_bindCb_t    *tplRspBindTcp;
#endif /* NFX_SMP */

	protocol = TPL_PROT_UDP;
	while (1)
	{
		//------------------------------
		// Bind NLM server on UDP or TCP
		//------------------------------
		memset(&bindCB, 0, sizeof(tpl_bindCb_t));

		bindOpt = TPL_BIND_SERVER | TPL_BIND_USE_RPC;
		bindCBp = &bindCB;
		bindCBp->laddr         = ipaddr;
		bindCBp->lport         = ntohs(NSM_PORT);
		bindCBp->tplProt       = protocol;
		bindCBp->ul_addConnInd = nsm_tplAddConnInd;
		bindCBp->ul_delConnInd = nsm_tplDelConnInd;
		bindCBp->ul_rcvPktInd  = nsm_tplRcvPktInd;
		bindCBp->ul_xmtCmpInd  = nsm_tplXmtCmpInd;

#if defined(NFX_SMP)
		bindCBp->ul_bindReqAck = nsm_tplBindReqAck;
		tpl_bindReq(bindCBp, NULL, bindOpt);
		
		if (protocol == TPL_PROT_TCP)
			break;
#else
		if (protocol == TPL_PROT_UDP)
		{
			if (tpl_bindReq(bindCBp, &tplRspBindUdp, bindOpt) != NFX_OK)
				panic("nsm_tplBindReq: bind failure (UDP)!\n");
			nsm_globalCfgp->nsm_pTplBind = tplRspBindUdp;
			nsm_globalCfgp->nsm_pBsdSock = tplRspBindUdp->pBsdSock;
		}
		else
		{
			if (tpl_bindReq(bindCBp, &tplRspBindTcp, bindOpt) != NFX_OK)
				panic("nsm_tplBindReq: bind failure (TCP)!\n");
			//TODO:
			break;	
		}
#endif /* NFX_SMP */
		protocol = TPL_PROT_TCP;

	}
	return (NFX_OK);
} // nsm_tplBindReq()

#endif  //NFP_TXRX
