/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:			nsm.c
 *
 * RCS:  $Id: nsm.c,v 1.7 2002/10/18 02:52:43 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *	Contain NSM utility and debugging functions.
 *
 * Created by:		Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX

#include "nsm.h"


static char *nsm_procName[] = {
                        "SM_NULL",
                        "SM_STAT",
                        "SM_MON",
                        "SM_UNMON",
                        "SM_UNMON_ALL",
                        "SM_SIM_CRASH",
                        "SM_NOTIFY" };

/*************************************
 * Registration function called by NLM
 *************************************
 */

/*-----------------------------------------------------------------
 * Name:		nsm_registerCallback
 *
 * Description: NLM registers callback functions for mon and unmom
 *              requests.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	09/03/2002
 *
 *-----------------------------------------------------------------
 */
void nsm_registerCallback(nsm_nlmCallbackCB_t	*callbackCBp)
{
	nsm_globalCfgp->nlmCallbackFunc.monCmpInd = 
										callbackCBp->monCmpInd;

	nsm_globalCfgp->nlmCallbackFunc.unmonCmpInd = 
										callbackCBp->unmonCmpInd;

	nsm_globalCfgp->nlmCallbackFunc.clientDownInd = 
										callbackCBp->clientDownInd;

	return;
}


/******************
 * Timeout handlers 
 ******************
 */

/*-----------------------------------------------------------------
 * Name:		nsm_reqTimeoutHandler
 *
 * Description: Periodic timeout function to delete oldest complete 
 *              request entry from the Duplicate request cache.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32
nsm_reqTimeoutHandler(	int32 timer_handle, 
						void* param1, void* param2,
						void* param3, void* param4)
{
	nfp_clientCB_t		*nfpClientCBp;
	nsm_clientCB_t		*nsmClientCBp;
	int32				count, iterations;
	nfp_node_t			*tailp;
	nsm_requestCB_t		*reqCBp;

	count = 0;
	for (nfpClientCBp = (nfp_clientCB_t *)nfp_globalCfgp->clientCbCfgList.listBasep;
		 ((count < NFP_MAX_CLIENT) && (nfpClientCBp != NULL));
		 count++, nfpClientCBp++)
	{
		// Search for nsmClientCB in hash table
		nsmClientCBp = nsm_getClientCB(nfpClientCBp->nfp_ipAddr);
		if (nsmClientCBp != NULL)
		{
			// Found nsmClientCB in hash table

			// Examine every node in the client request list
			iterations = nsmClientCBp->clientReqList.nfp_reqCnt;
			for (tailp = nsmClientCBp->clientReqList.nfp_tailp;
				 (tailp != NULL && iterations > 0);
				 tailp = tailp->nfp_prevp, iterations--)
			{
				if (nfp_validateNode(tailp) != NFX_OK)
					continue;

				reqCBp = (nsm_requestCB_t *)tailp->nfp_entryCBp;
				if (reqCBp->reqState == NSM_REQ_DONE)
				{
					// Remove reqCB if state is DONE
					nsm_delReqCB(reqCBp, TRUE);
				}
			}
		}
	}

	return (NFX_OK);
} // nsm_reqTimeoutHandler()


/*******************
 * Hashing functions
 *******************
 */

/*-----------------------------------------------------------------
 * Name:		nsm_clientHashFunc
 *
 * Description: Generic hashing function for hash table of clients.
 *              Called to return hash key.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
uint32		// hash value
nsm_clientHashFunc(void *key)			//key is pointer to IP address
{
	uint32	*hostIpAddr = (uint32 *)key;

	return (*hostIpAddr % NSM_HASH_CLI_BUCKETS);
} // nsm_clientHashFunc()

/*-----------------------------------------------------------------
 * Name:		nsm_clientHashCompare
 *
 * Description: Comparison function for hash table of clients.
 *              Called to check for a clientCB match.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32		// 0 - match; 1 - no match
nsm_clientHashCompare(nsm_clientCB_t *stored,		//Data stored in hasn
					  nsm_clientCB_t *compare)		//Data to compare
{
	if ((stored == NULL) || (compare == NULL))
	{
		// Shouldn't happen
		NSM_PRINTF(LOG_ERR, ("nsm_clientHashCompare: NULL pointer"));
		return (1);
	}

	if (stored->hostIpAddr == compare->hostIpAddr)
		return (0);
	else
		return (1);
} // nsm_clientHashCompare()

/*-----------------------------------------------------------------
 * Name:		nsm_clientRelease
 *
 * Description: Release function for hash table of clients.
 *              Called to destroy the entire hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_clientHashRelease(nsm_clientCB_t *clientCBp)
{
	// Nothing to do here
	return;
} // nsm_clientHashRelease()

/*-----------------------------------------------------------------
 * Name:		nsm_volumeHashFunc
 *
 * Description: Generic hashing function for hash table of volumes.
 *              Called to return hash key.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
uint32		// hash value
nsm_volumeHashFunc(void *key)			//key is pointer to volume ID
{
	evm_volId_t		*volId;

	volId = (evm_volId_t *)key;

	return ((uint32)(volId->whole_thing % NSM_HASH_VOL_BUCKETS));
} // nsm_volumeHashFunc()

/*-----------------------------------------------------------------
 * Name:		nsm_volumeHashCompare
 *
 * Description: Comparison function for hash table of volumes.
 *              Called to check for a volumeCB match.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32		// 0 - match; 1 - no match
nsm_volumeHashCompare(nsm_volumeCB_t *stored,		//Data stored in hasn
					  nsm_volumeCB_t *compare)		//Data to compare
{
	if ((stored == NULL) || (compare == NULL))
	{
		// Shouldn't happen
		NSM_PRINTF(LOG_ERR, ("nsm_volumeHashCompare: NULL pointer"));
		return (1);
	}

	if (stored->volId.whole_thing == compare->volId.whole_thing)
		return (0);
	else
		return (1);
} // nsm_volumeHashCompare()

/*-----------------------------------------------------------------
 * Name:		nsm_volumeRelease
 *
 * Description: Release function for hash table of volumes.
 *              Called to destroy the entire hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_volumeHashRelease(nsm_volumeCB_t *volumeCBp)
{
	// Nothing to do here
	return;
} // nsm_volumeHashRelease()


/*******************
 * Utility functions
 *******************
 */

/*-----------------------------------------------------------------
 * Name:		nsm_addClientCB
 *
 * Description: Allocate a Client Control Block and add it to hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/07/2002
 *
 *-----------------------------------------------------------------
 */
nsm_clientCB_t *
nsm_addClientCB(uint32	ipAddr,
				uint32	nameLen,
				char	*hostName)
{
	int32			rc, size;
	nsm_clientCB_t	*clientCBp;

	if (nameLen > NSM_MAX_NAME)
		nameLen = NSM_MAX_NAME;

	size = sizeof(nsm_clientCB_t) + nameLen + 1;
	NSM_SHMEM_ALLOC(clientCBp, nsm_clientCB_t, size);
	if (clientCBp == NULL)
	{
		return (NULL);
	}
	memset(clientCBp, 0, sizeof(nsm_clientCB_t));

	clientCBp->status = 0;
	clientCBp->hostIpAddr = ipAddr;
	clientCBp->nameLen = nameLen;
	if (hostName != NULL)
	{
		strncpy(clientCBp->hostName, hostName, nameLen);
		clientCBp->hostName[nameLen] = '\0';
	}

	//clientCBp->nsm_pBsdSock = nsm_globalCfgp->nsm_pBsdSock;
	//clientCBp->pmap_pBsdSock = nsm_globalCfgp->pmap_pBsdSock;

	rc = hash_insert(nsm_globalCfgp->nsm_clientHash, &ipAddr, clientCBp);
	if (rc != NFX_OK)
	{
		NSM_SHMEM_DEALLOC(clientCBp);
	}

	return (clientCBp);
} // nsm_addClientCB()

/*-----------------------------------------------------------------
 * Name:		nsm_delClientCB
 *
 * Description: Delete a Client Control Block, and remove it from
 *              the hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/07/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_delClientCB(nsm_clientCB_t *clientCBp)
{
	//TODO
	// Delete volume list
	
	// Delete client request list

	// Delete server reply list

	// Delete notify request CB

	// Dealloc host name buffer

	return;
} // nsm_delClientCB()

/*-----------------------------------------------------------------
 * Name:		nsm_getClientCB
 *
 * Description: Retrieve a Client Control Block from hash table
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/03/2002
 *
 *-----------------------------------------------------------------
 */
nsm_clientCB_t *	//Return pointer to the NSM Client Control Block or
					//NULL if entry is not found
nsm_getClientCB(uint32 clientIpAddr)
{
	nsm_clientCB_t	*currp, clientCB;
		                              
	clientCB.hostIpAddr = clientIpAddr;
	currp = (nsm_clientCB_t *)hash_lookup(nsm_globalCfgp->nsm_clientHash,
											(void *)&clientIpAddr,
											(void *)&clientCB);
	return (currp);
} // nsm_getClientCB()

/*-----------------------------------------------------------------
 * Name:		nsm_addVolumeCB
 *
 * Description: Allocate a Volume Control Block and add it to hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/07/2002
 *
 *-----------------------------------------------------------------
 */
nsm_volumeCB_t *
nsm_addVolumeCB(evm_volId_t	volId,
				char		*volName)
{
	int32			rc, size;
	nsm_volumeCB_t	*volumeCBp;

	size = sizeof(nsm_volumeCB_t) + strlen(volName) + 1;
	NSM_SHMEM_ALLOC(volumeCBp, nsm_volumeCB_t, size);
	if (volumeCBp == NULL)
	{
		return (NULL);
	}
	memset(volumeCBp, 0, sizeof(nsm_volumeCB_t));

	volumeCBp->volId.whole_thing = volId.whole_thing;
	volumeCBp->nameLen = strlen(volName);
	strcpy(volumeCBp->volName, volName);

	rc = hash_insert(nsm_globalCfgp->nsm_volumeHash, &volId, volumeCBp);
	if (rc != NFX_OK)
	{
		NSM_SHMEM_DEALLOC(volumeCBp);
	}

	return (volumeCBp);
} // nsm_addVolumeCB()

/*-----------------------------------------------------------------
 * Name:		nsm_getVolumeCB
 *
 * Description: Retrieve a Volume Control Block from the hash table 
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/03/2002
 *
 *-----------------------------------------------------------------
 */
nsm_volumeCB_t *	//Return pointer to the NSM Volume Control Block or
					//NULL if entry is not found
nsm_getVolumeCB(evm_volId_t volId)
{
	nsm_volumeCB_t	*currp, volumeCB;
		                              
	volumeCB.volId.whole_thing = volId.whole_thing;
	currp = (nsm_volumeCB_t *)hash_lookup(nsm_globalCfgp->nsm_volumeHash,
											(void *)&volId,
											(void *)&volumeCB);
	return (currp);
} // nsm_getVolumeCB()

/*-----------------------------------------------------------------
 * Name:		nsm_findVolumeCBNode
 *
 * Description: Find the node containing the volumeCB
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/08/2002
 *
 *-----------------------------------------------------------------
 */
nfp_node_t *
nsm_findVolumeCBNode(nfp_list_t *nodeListp, evm_volId_t volId)
{
	uint32			count;
	nfp_node_t		*nodep;
	nsm_volumeCB_t	*volCBp;

	count = nodeListp->nfp_reqCnt;
	for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
			nodep = nodep->nfp_nextp, count--)
	{
		volCBp = (nsm_volumeCB_t *)nodep->nfp_entryCBp;
		if (volCBp->volId.whole_thing == volId.whole_thing)
			return (nodep);
	}

	return (NULL);
} // nsm_findVolumeCBNode()

/*-----------------------------------------------------------------
 * Name:		nsm_allocReqCB
 *
 * Description: Allocate a Request Control Block.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
nsm_requestCB_t *
nsm_allocReqCB(nsm_clientCB_t *clientCBp)
{
	nsm_requestCB_t *reqCBp;

	NSM_SHMEM_ALLOC(reqCBp, nsm_requestCB_t, sizeof(nsm_requestCB_t));
	if (reqCBp == NULL)
	{
		return (NULL); 
	}
	memset(reqCBp, 0, sizeof(nsm_clientCB_t));

	reqCBp->reqState = NSM_REQ_INPROG;
	reqCBp->clientCBp = clientCBp;

	nsm_globalCfgp->nsm_debugCounters.nsm_allocReqCB++;

	return (reqCBp);
} // nsm_allocReqCB()

/*-----------------------------------------------------------------
 * Name:		nsm_findReqCB
 *
 * Description: Find a Request Control Block with a correponding 
 *              transaction ID & procedure number.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
nsm_requestCB_t *
nsm_findReqCB(nsm_clientCB_t	*clientCBp,		//Client control block
			  uint32			xid,			//Transaction ID
			  uint32			procNum)		//Procedure number
{
	nfp_list_t		*nodeListp;
	int32			count;
	nfp_nodePtr_t	reqp;
	nsm_requestCB_t	*nsmReqCBp;

	nodeListp = &clientCBp->clientReqList;
	count = nodeListp->nfp_reqCnt;
	NSM_PRINTF(LOG_DEBUG, ("nsm_findReqCB: count = %d", count));

	for (reqp = nodeListp->nfp_headp; (reqp != NULL && count > 0);
		 reqp = reqp->nfp_nextp, count--)
	{
		if (nfp_validateNode(reqp) != NFX_OK)
			continue;

		nsmReqCBp = (nsm_requestCB_t *)reqp->nfp_entryCBp;
		if ((nsmReqCBp->rpcCallHdrp) &&
			(nsmReqCBp->rpcCallHdrp->rpc_xid == xid) &&
			(nsmReqCBp->rpcCallHdrp->rpc_proc == procNum))
		{
			return (nsmReqCBp);
		}
	}

	return (NULL);
} // nsm_findReqCB()

/*-----------------------------------------------------------------
 * Name:		nsm_findReqCBNode
 *
 * Description: Find a node containing Request Control Block 
 *              with a correponding transaction ID & procedure number.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t
nsm_findReqCBNode(nfp_list_t		*nodeListp,		//Requeste node list
				  nsm_requestCB_t	*nsmReqCBp)	
{
	int32			count;
	nfp_nodePtr_t	reqp;
	nsm_requestCB_t	*reqCBp;

	nodeListp = &nsmReqCBp->clientCBp->clientReqList;
	count = nodeListp->nfp_reqCnt;
	NSM_PRINTF(LOG_INFO, ("nsm_findReqCB: count = %d", count));

	for (reqp = nodeListp->nfp_headp; (reqp != NULL && count > 0);
		 reqp = reqp->nfp_nextp, count--)
	{
		if (nfp_validateNode(reqp) != NFX_OK)
			continue;

		reqCBp = (nsm_requestCB_t *)reqp->nfp_entryCBp;
		if ((reqCBp->rpcCallHdrp) &&
			(reqCBp->rpcCallHdrp->rpc_xid == nsmReqCBp->rpcCallHdrp->rpc_xid) &&
			(reqCBp->rpcCallHdrp->rpc_proc == nsmReqCBp->rpcCallHdrp->rpc_proc))
		{
			return (reqp);
		}
	}

	return (NULL);
} // nsm_findReqCBNode()

/*-----------------------------------------------------------------
 * Name:		nsm_delReqCBNodeFromList
 *
 * Description: Delete a Request Control Block from client's request list.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_delReqCBNodeFromList(nsm_requestCB_t *reqCBp)
{
	nsm_clientCB_t		*clientCBp;
	nfp_list_t			*nsmReqListp;
	nfp_nodePtr_t		nodep;

	if (reqCBp == NULL)
		return (NFX_ERR);

	clientCBp = reqCBp->clientCBp;
	if (clientCBp == NULL)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_delReqCBFromList: No clientCB!"));
		return (NFX_ERR);
	}

	nsmReqListp = &clientCBp->clientReqList;

	//------------------------------
	// Delete node from request list
	//------------------------------
	nodep = nsm_findReqCBNode(nsmReqListp, reqCBp);
	if (nodep == NULL)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_delReqCBFromList: No node for reqCB!"));
		return (NFX_ERR);
	}	

	return (nfs_delNode(nsmReqListp, nodep));
} // nsm_delReqCBFromList()

/*-----------------------------------------------------------------
 * Name:		nsm_delReqCB
 *
 * Description: De-allocate all previous allocated memory for the request.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/21/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_delReqCB(nsm_requestCB_t *reqCBp, int32 rmFromList)
{
	int32		rc;
	void		*ra;

	if (reqCBp == NULL)
		return;

	ra = (void *)get_ra();
	if (reqCBp->reqState != NSM_REQ_DONE)
	{
		panic ("%s: Caller (%p) not in valid state!\n",
				__FUNCTION__, ra);
	}

	//---------------------------
	// Free TPL Rx request packet
	//---------------------------
	if ((reqCBp->rxPktDescp != NULL) &&
		(pkt_validateDesc(reqCBp->rxPktDescp , PKT_CHK_DUP) == NFX_OK))
	{
		rc = PKT_FREE_PACKET(reqCBp->rxPktDescp);
		if (rc != NFX_OK)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_delReqCB: Free TPL Rx req Pkt failed!"));
		}
		reqCBp->rxPktDescp = NULL;
		nsm_globalCfgp->nsm_debugCounters.nsm_tplRxPktFree++;
	}

	//----------------------------
	// Free TPL Tx response packet
	//----------------------------
	if ((reqCBp->txPktDescp != NULL) &&
		(pkt_validateDesc(reqCBp->txPktDescp , PKT_CHK_DUP) == NFX_OK))
	{
		rc = TPL_FREE_PACKET(reqCBp->tplConnp, reqCBp->txPktDescp);
		if (rc != NFX_OK)
		{
			NSM_PRINTF(LOG_ERR, ("nsm_delReqCB: Free TPL Tx req Pkt failed!"));
		}
		reqCBp->txPktDescp = NULL;
		nsm_globalCfgp->nsm_debugCounters.nsm_tplXmtPktFree++;
	}

	//------------------------------------------------------------
	// Delete NSM request control block from client's request list
	//------------------------------------------------------------
	if (rmFromList == TRUE)
	{
		rc = nsm_delReqCBNodeFromList(reqCBp);
		if (rc != NFX_OK)
		{
			// TODO: Handle error
			NSM_PRINTF(LOG_ERR, ("nsm_delReqCB: nsm_delReqCBFromList failed!"));
		}
	}

	//------------------------------
	// De-alloc rquest control block
	//------------------------------
	NSM_SHMEM_DEALLOC(reqCBp);
	nsm_globalCfgp->nsm_debugCounters.nsm_freeReqCB++;

	return;
} // nsm_delReqCB()

/*-----------------------------------------------------------------
 * Name:		nsm_addMsgToThreadCB
 *
 * Description: Allocate a msgToThreadCB and add it to waiting list.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/29/2002
 *
 *-----------------------------------------------------------------
 */
int32		// NFX_OK or NFX_ERR
nsm_addMsgToThreadCB(char				*hostName,
					 uint32				hostIpAddr,
					 evm_volId_t		volId,
					 nsm_clientCB_t		*nsm_clientCBp,
					 nfp_clientCB_t		*nlm_clientCBp, //NLM clientCB
					 nsm_requestCB_t	*nsm_requestCBp,
					 int32				msgType)
{
	int32					size;
	nsm_msgToThreadCB_t		*msgToThreadCBp;

	// Allocate memory
	size = sizeof(nsm_msgToThreadCB_t) + strlen(hostName) + 1;
	NSM_SHMEM_ALLOC(msgToThreadCBp, nsm_msgToThreadCB_t, size);
	if (msgToThreadCBp == NULL)
	{
		NSM_PRINTF(LOG_ERR, ("nsm_addMsgToThreadCB: Insufficient memory"));
		return (NFX_ERR);
	}
	nsm_globalCfgp->nsm_debugCounters.allocMsgToThread++;
	memset(msgToThreadCBp, 0, sizeof(nsm_msgToThreadCB_t));

	// Fill the fields
	msgToThreadCBp->msgType = msgType;
	msgToThreadCBp->nsm_clientCBp = nsm_clientCBp;
	msgToThreadCBp->nlm_clientCBp = nlm_clientCBp;
	msgToThreadCBp->nsm_requestCBp = nsm_requestCBp;
	msgToThreadCBp->volId.whole_thing = volId.whole_thing;
	msgToThreadCBp->hostIpAddr = hostIpAddr;
	strcpy(msgToThreadCBp->hostName, hostName);

	// Append the control block to waiting list
	if (nsm_globalCfgp->msgToThreadListHead == NULL)
	{
		// The list is empty
		nsm_globalCfgp->msgToThreadListHead = msgToThreadCBp;
		nsm_globalCfgp->msgToThreadListTail = msgToThreadCBp;
	}
	else
	{
		// Append to the end of the list
		msgToThreadCBp->prevp = nsm_globalCfgp->msgToThreadListTail;
		nsm_globalCfgp->msgToThreadListTail->nextp = msgToThreadCBp;
		nsm_globalCfgp->msgToThreadListTail = msgToThreadCBp;
	}

	return (NFX_OK);
} // nsm_addMsgToThreadCB()

/*-----------------------------------------------------------------
 * Name:		nsm_getMsgToThreadCB
 *
 * Description: Remove the first msgToThreadCB from waiting list.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/29/2002
 *
 *-----------------------------------------------------------------
 */
nsm_msgToThreadCB_t *
nsm_getMsgToThreadCB()
{
	nsm_msgToThreadCB_t		*msgToThreadCBp;

	if (nsm_globalCfgp->msgToThreadListHead == NULL)
	{
		// The list is empty
		return (NULL);
	}
	else
	{
		// Remove the first node in the list
		msgToThreadCBp = nsm_globalCfgp->msgToThreadListHead;
		nsm_globalCfgp->msgToThreadListHead = msgToThreadCBp->nextp;
	}

	return (msgToThreadCBp);
} // nsm_getMsgToThreadCB()


/*********************
 * Debugging functions
 *********************
 */

#define NSM_STR								"NSM"
#define NSM_BAD_RPC_HDR_ERR_STR				"[%s]Bad RPC Hdr\n"
#define NSM_BAD_RPC_MISMATCH_ERR_STR		"[%s]RPC mistmatch\n"
#define NSM_BAD_RPC_PROG_MISMATCH_ERR_STR	"[%s]RPC Prog mistmatch\n"
#define NSM_BAD_RPC_PROG_UNAVAIL_ERR_STR	"[%s]RPC Proc Unvail\n"

static char *nsm_procStrings[] =
{
	"SM_NULL",
	"SM_STAT",
	"SM_MON",
	"SM_UNMON",
	"SM_UNMON_ALL",
	"SM_SIMU_CRASH",
	"SM_NOTIFY",
};

/*-----------------------------------------------------------------
 * Name:		nsm_decodeRpcCallHdr
 *
 * Description: Print the RPC CALL header contents.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/24/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_decodeRpcCallHdr(rpcCallHdr_t *rpcCallHdrp)
{
	printf("\n\t----- RPC Header -----\n");

	printf("\tXID: 0x%08x\n", rpcCallHdrp->rpc_xid);
	printf("\tDirection: %s\n",
						(rpcCallHdrp->rpc_direction == 0)? "Call": "Reply");


	printf("\tRPC Version: %d \n", rpcCallHdrp->rpc_rpcVer);
	printf("\tProgram number: %d (%s)\n",
						rpcCallHdrp->rpc_prog,
						nfs_decodeGetProgNumStr(rpcCallHdrp->rpc_prog ));
	printf("\tProgram Version: %s\n", 
						(rpcCallHdrp->progVer == NSM_VER1)? "V1" : "Invalid");

	printf("\tProcedure number: %d (%s)\n",
						rpcCallHdrp->rpc_proc,
						nsm_procStrings[rpcCallHdrp->rpc_proc]);

	nfs_rxDecodeAuthSysHostName(rpcCallHdrp);

	return;
} // nsm_decodeRpcCallHdr()

/*-----------------------------------------------------------------
 * Name:		nsm_decodeStatProc
 *
 * Description: Decode SM_STAT request.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/24/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_decodeStatProc(char *msgp)
{
	int32			length;
	ushort16		bufLen;
	static char		buf[NSM_MAX_NAME + 1];

	bufLen = 0;
	memset(buf, 0, sizeof(buf));

	// Decode host name
	length = nfs_extractString(msgp, buf, &bufLen, NSM_MAX_NAME);
	if (length == NFX_ERR)
	{
		printf("%s: Failure to decode host name!\n", __FUNCTION__);
		return;
	}
	printf("\tmon_name = \t%s\n", buf);
	
	return;
} // nsm_decodeStatProc()

/*-----------------------------------------------------------------
 * Name:		nsm_decodeNotifyProc
 *
 * Description: Decode SM_NOTIFY request.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/24/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_decodeNotifyProc(char *msgp, int32 direction)
{
	int32			totalLen, length;
	ushort16		bufLen;
	static char		buf[NSM_MAX_NAME + 1];
	int32			state;

	totalLen = 0;
	bufLen = 0;
	memset(buf, 0, sizeof(buf));

	// Decode host name
	if (direction == NSM_DECODE_RX)
		length = nfs_extractString(msgp, buf, &bufLen, NSM_MAX_NAME);
	else
		length = nfs_txDecodeString(msgp, buf, &bufLen);
	if (length == NFX_ERR)
	{
		printf("%s: Failure to decode host name!\n", __FUNCTION__);
		return;
	}
	printf("\tmon_name = \t%s\n", buf);

	// Update message pointer
	totalLen += length;
	msgp += length;

	// Decode state field
	if (direction == NSM_DECODE_RX)
		length = nfs_extractLongVal(msgp, &state);
	else
		length = nfs_txExtractLongVal(msgp, &state);
	printf("\tState = %d\n", state);

	return;
} // nsm_decodeNotifyProc()
	
/*-----------------------------------------------------------------
 * Name:		nsm_decodeReqPkt
 *
 * Description: Decode NSM request packet.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/24/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_decodeReqPkt(pkt_desc_t *descp, int32 direction)
{
	rpcCallHdr_t	*rpcCallHdrp;
	uint32			rpcHdrSize;
	char			*msgp;

	if (descp == NULL)
	{
		printf(NSM_BAD_RPC_HDR_ERR_STR, NSM_STR);
		return;
	}

	rpcCallHdrp = (rpcCallHdr_t *)descp->pHdr;
	if (rpcCallHdrp == NULL)
	{
		printf(NSM_BAD_RPC_HDR_ERR_STR, NSM_STR);
		return;
	}

	nsm_decodeRpcCallHdr(rpcCallHdrp);

	printf("\t----- NSM packet -----\n");

	// Skip the RPC header
	rpcHdrSize = rpc_getHdrSize(rpcCallHdrp);
	msgp = (char *)rpcCallHdrp;
	msgp += rpcHdrSize;

	switch(rpcCallHdrp->rpc_proc)
	{
		case NSM_NULL:
			break;

		case NSM_STAT:
			nsm_decodeStatProc(msgp);
			break;

		case NSM_NOTIFY:
			nsm_decodeNotifyProc(msgp, direction);
			break;

		default:
			break;
	}
	printf("\n");

	return;
} // nsm_decodeReqPkt()

/*-----------------------------------------------------------------
 * Name:		nsm_decodeRpcReplyHdr
 *
 * Description: Print the RPC REPLY header contents.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/24/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_decodeRpcReplyHdr(rpcReplyHdr_t	*rpcReplyHdrp, rpcCallHdr_t *rpcCallHdrp)
{
	printf("\n\t----- RPC Header -----\n");

	printf("\tXID: 0x%08x\n", rpcReplyHdrp->rpc_xid);
	printf("\tDirection: %s\n",
						(rpcReplyHdrp->rpc_direction == 0)? "Call": "Reply");

	printf("\tProgram number: %d (%s)\n",
						rpcCallHdrp->rpc_prog,
						nfs_decodeGetProgNumStr(rpcCallHdrp->rpc_prog ));
	printf("\tProgram Version: %s\n",
						(rpcCallHdrp->progVer == NSM_VER1)? "V1" : "Invalid");
	printf("\tProcedure number: %d (%s)\n",
						rpcCallHdrp->rpc_proc,
						nsm_procStrings[rpcCallHdrp->rpc_proc]);

	switch (rpcReplyHdrp->rpc_status)
	{
		case RPC_REPLY_MSG_ACCEPTED:
		{
			printf("\tReply Status: %s\n",  "MSG_ACCEPTED");
			break;
		}
		case RPC_REPLY_MSG_DENIED:
			printf("\tReply Status: %s\n",  "MSG_DENIED");
			break;
	}

	return;
} // nsm_decodeRpcReplyHdr()
	
/*-----------------------------------------------------------------
 * Name:		nsm_decodeRspPkt
 *
 * Description: Decode NSM response packet.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/24/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_decodeRspPkt(pkt_desc_t *tplXmitPktDescp, nsm_requestCB_t *reqCBp)
{
	rpcCallHdr_t	*rpcCallHdrp;
	rpcReplyHdr_t	*rpcReplyHdrp;
	char			*msgp;
	int32			length;
	int32			res, state;

	if ((reqCBp == NULL) || (reqCBp->rpcCallHdrp == NULL))
		return;

	rpcCallHdrp = reqCBp->rpcCallHdrp;
	rpcReplyHdrp = (rpcReplyHdr_t *)tplXmitPktDescp->pHdr;

	nsm_decodeRpcReplyHdr(rpcReplyHdrp, rpcCallHdrp);

	if (rpcCallHdrp->rpc_proc != NSM_STAT)
	{
		// No arguments for SM_NULL & SM_NOTIFY procedures just return
	}
	else
	{
		// Decode SM_STAT reply that we are to send to client
		printf("\t----- NSM Reply packet -----\n");
		msgp = (char *)tplXmitPktDescp->pHdr;
		msgp += 24;

		length = nfs_txExtractLongVal(msgp, &res);
		msgp += length;
		length = nfs_txExtractLongVal(msgp, &state);

		printf("\tReturn Status: %s\n", (res == 0) ? "STAT_SUCC" : "STAT_FAIL");
		printf("\tState: %d\n", state);
	}

	printf("\n");
	return;
} // nsm_decodeRspPkt()

/*-----------------------------------------------------------------
 * Name:		nsm_totalPktCnt
 *
 * Description: Return total packets count on the NSM server.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
uint64
nsm_totalPktCnt(uint64 *pktCntp)
{
	uint64		totalPktCnt = 0;

	totalPktCnt = 	pktCntp[NSM_NULL] +
					pktCntp[NSM_STAT] +
					pktCntp[NSM_MON] +
					pktCntp[NSM_UNMON] +
					pktCntp[NSM_UNMON_ALL] +
					pktCntp[NSM_SIMU_CRASH] +
					pktCntp[NSM_NOTIFY];

	return (totalPktCnt);
} //nsm_totalPktCnt()


/*-----------------------------------------------------------------
 * Name:		nsm_showProcStat
 *
 * Description: Display NSM procedures statistics.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 * Modified:		09/25/2002		JVu 
 *
 *-----------------------------------------------------------------
 */
void
nsm_showProcStat(void)
{
	int32		proc;
	uint64		total_nsmRxPkt, total_nsmTxPkt;

	printf("NSM PROC                        NSM-RX  NSM-TX\n");
	printf("-------------------------------------------------\n");

	for (proc =NSM_NULL; proc <= NSM_NOTIFY; proc++)
	{
		printf("%12s:%21llu  ", nsm_procName[proc], 
			nsm_globalCfgp->nsm_debugCounters.nsm_rxReqPktCnt[proc]);
		printf(" %6llu ",
			nsm_globalCfgp->nsm_debugCounters.nsm_txRspPktCnt[proc]);
		printf("\n");
	}
	total_nsmRxPkt = 
		nsm_totalPktCnt(nsm_globalCfgp->nsm_debugCounters.nsm_rxReqPktCnt);
	total_nsmTxPkt = 
		nsm_totalPktCnt(nsm_globalCfgp->nsm_debugCounters.nsm_txRspPktCnt);
	printf("       Total:");
	printf("%21llu%9llu\n\n",total_nsmRxPkt, 
				total_nsmTxPkt);

	return;
} //nsm_showProcStat()


/*-----------------------------------------------------------------
 * Name:		nsm_showErrorCounters
 *
 * Description: Display NSM error counters.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_showErrorCounters(nsm_statCounters_t *errCntp)
{
	printf("NSM Succeed:                     %lld\n", errCntp->nsm_succ);
	printf("NSM Failed:                      %lld\n", errCntp->nsm_fail);
	printf("NSM Err Not Support:             %lld\n", errCntp->nsm_errNotSupp);
	printf("NSM Err Srvr Fault:              %lld\n", errCntp->nsm_errSrvFault);

	return;
} //nsm_showErrorCounters()

/*-----------------------------------------------------------------
 * Name:		nsm_showDebugCounters
 *
 * Description: Display NSM debug counters.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_showDebugCounters(nsm_debugCounters_t *dbgCntp)
{
	printf("TPL Xmt alloc pkts:              %lld\n",
										dbgCntp->nsm_tplXmtPktAlloc);
	printf("TPL Xmt free pkts:               %lld\n",
										dbgCntp->nsm_tplXmtPktFree);
	printf("TPL Xmt pkts:                    %lld\n",
										dbgCntp->nsm_tplXmtPktSend);
	printf("TPL Xmt pkts completed:          %lld\n",
										dbgCntp->nsm_tplXmtPktCmpInd);
	printf("TPL Rx pkts hold:                %lld\n",
										dbgCntp->nsm_tplRxPktHold);
	printf("TPL Rx pkts free:                %lld\n",
										dbgCntp->nsm_tplRxPktFree);
	printf("TPL Drop rcvd pkts:              %lld\n",
										dbgCntp->nsm_tplDropRcvPkt);
	printf("Alloc EEE buffers:               %lld\n",
										dbgCntp->nsm_eeeAlloc);
	printf("Free EEE buffers:                %lld\n",
										dbgCntp->nsm_eeeDealloc);
	printf("Alloc NSM req CBs:               %lld\n",
										dbgCntp->nsm_allocReqCB);
	printf("Free NSM req CBs:                %lld\n",
										dbgCntp->nsm_freeReqCB);
	printf("Alloc NSM Msg To Thread:         %lld\n",
										dbgCntp->allocMsgToThread);
	printf("Free NSM Msg To Thread:          %lld\n",
										dbgCntp->freeMsgToThread);
	printf("Num of requests sent to FS:      %lld\n",
										dbgCntp->nsm_reqsXmtToFs);
	printf("Num of responses from FS:        %lld\n",
										dbgCntp->nsm_rspsRcvFmFs);
	printf("Num of requests sent to EVM:     %lld\n",
										dbgCntp->nsm_reqsXmtToEvm);
	printf("Num of responses from EVM:       %lld\n",
										dbgCntp->nsm_rspsRcvFmEvm);
	printf("Num of events sent to thread:    %lld\n",
										dbgCntp->nsm_evtToThreadSend);
	printf("Num of events rcvd by thread:    %lld\n",
										dbgCntp->nsm_evtToThreadRcvd);
	printf("Num of events sent to thread Err:%lld\n",
										dbgCntp->nsm_evtToThreadSendErr);
	printf("Num of msgs sent from thread:    %lld\n",
										dbgCntp->nsm_msgFromThreadSend);
	printf("Num of msgs rcvd from thread:    %lld\n",
										dbgCntp->nsm_msgFromThreadRcvd);

	printf("TPL Xmt no pkt resources:        %lld\n",
										dbgCntp->nsm_tplXmtNoPktRsrcErr);
	printf("TPL Xmt pkt error:               %lld\n",
										dbgCntp->nsm_tplXmtPktSendErr);
	printf("TPL Rx bad desc:                 %lld\n",
										dbgCntp->nsm_tplBadDescErr);
	printf("TPL Rx invalid pkt:              %lld\n",
										dbgCntp->nsm_tplInvalidPktRcvdErr);
	printf("Rx pkt indication error:         %lld\n",
										dbgCntp->nsm_rcvPktIndErr);
	printf("Rx RPC header error:             %lld\n",
										dbgCntp->nsm_rcvRpcHdrErr);
	printf("Num of Getport locate Req Errs:  %lld\n",
										dbgCntp->nsm_rcvGetPortLocateReqErr);
	printf("Num of Notify Reply Errs:        %lld\n",
										dbgCntp->nsm_rcvNotifyLocateReqErr);

	return;
} //nsm_showDebugCounters()

/*-----------------------------------------------------------------
 * Name:		nsm_showStats
 *
 * Description: Display NSM statistics.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_showStats()
{

	//Show NSM procedure counters
	nsm_showProcStat();

	printf("Server NSM Notify call and reply counters\n");
	printf("-----------------------------------------\n");
	printf("Tx NSM_NOTIFY Req:               %lld\n",
					nsm_globalCfgp->nsm_debugCounters.nsm_txNotifyReqPktCnt);
	printf("Rx NSM_NOTIFY Rsp:               %lld\n",
					nsm_globalCfgp->nsm_debugCounters.nsm_rxNotifyRspPktCnt);
	printf("NSM_NOTIFY Req retries:          %lld\n",
					nsm_globalCfgp->nsm_debugCounters.nsm_pktNotifyRetries);
	printf("\n");

	printf("Server Portmap call and reply counters\n");
	printf("---------------------------------------\n");
	printf("Tx PORTMAP Req:                  %lld\n",
					nsm_globalCfgp->nsm_debugCounters.nsm_txGetPortReqPktCnt);
	printf("Rx PORTMAP Rsp:                  %lld\n",
					nsm_globalCfgp->nsm_debugCounters.nsm_rxGetPortRspPktCnt);
	printf("PORTMAP Req retries:             %lld\n",
					nsm_globalCfgp->nsm_debugCounters.nsm_pktGetPortRetries);
	printf("\n");

	printf("NSM Status counters\n");
	printf("-------------------\n");
	nsm_showErrorCounters(&nsm_globalCfgp->nsm_statCounters);
	printf("\n");

	printf("\nNSM debug counters\n");
	printf("------------------\n");
	nsm_showDebugCounters(&nsm_globalCfgp->nsm_debugCounters);
	printf("\n");

	return;
} // nsm_showStats()

/*-----------------------------------------------------------------
 * Name:		nsm_clearStats
 *
 * Description: Clear NSM statistics.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_clearStats()
{
	bzero(&nsm_globalCfgp->nsm_statCounters, sizeof(nsm_globalCfgp->nsm_statCounters));
	bzero(&nsm_globalCfgp->nsm_debugCounters, sizeof(nsm_globalCfgp->nsm_debugCounters));

	return;
} // nsm_clearStats()

#endif	//NFP_TXRX
