/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: fs.c
 *
 * RCS:  $Id: fs.c,v 1.5 2002/09/17 00:55:26 maximk Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: functions for making filesystems writable/readonly
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created: Aug 13 2002
 *
 *-----------------------------------------------------------------
 */


#undef DEBUG
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif


#include <sys/types.h>
#include <fstab.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include "opt-api.h"
#include <syslog.h>


/* Local filesystem type */
static const char *our_fstype = "ffs";

/* File for serializing the access */
static const char *lock_file = "/var/run/mount.lock";

/* File descriptor for serializing the access */
static int lock_fd = -1;


/*-----------------------------------------------------------------
 * Name :	lock()
 *
 * Description: obtain lock for remount
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/14/02
 *
 *-----------------------------------------------------------------
 */
static int
lock(void)
{
	/* Check if trying to do recursive lock */
	if (lock_fd != -1) abort();
	lock_fd = open(lock_file, O_RDONLY|O_EXLOCK);
	if (lock_fd < 0) {
		debug(("open %s failed: %s\n", lock_file, strerror(errno)));
	}
	return lock_fd >= 0;
}


/*-----------------------------------------------------------------
 * Name :	unlock()
 *
 * Description: release lock for remount
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/14/02
 *
 *-----------------------------------------------------------------
 */
static void
unlock(void)
{
	if (lock_fd >= 0) {
		int rv;
		rv = flock(lock_fd, LOCK_UN);
		if (rv != 0) {
			debug(("flock unlock failed: %s\n", strerror(errno)));
		}
		close(lock_fd);
		lock_fd = -1;
	}
}


/*-----------------------------------------------------------------
 * Name :	init_fstab()
 *
 * Description: Initialize fstab structure
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
static void
init_fstab(struct fstab *fs)
{
	fs->fs_file = fs->fs_spec = fs->fs_vfstype = fs->fs_mntops =
		fs->fs_type = NULL;
}


/*-----------------------------------------------------------------
 * Name :	free_fstab()
 *
 * Description: Free allocated fields
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
static void
free_fstab(struct fstab *fs)
{
	free(fs->fs_spec);
	free(fs->fs_file);
	free(fs->fs_vfstype);
	free(fs->fs_mntops);
	free(fs->fs_type);
	fs->fs_file = fs->fs_spec = fs->fs_vfstype = fs->fs_mntops =
		fs->fs_type = NULL;
}


/*-----------------------------------------------------------------
 * Name :	copy_fstab()
 *
 * Description: Copy fstab structures allocating strings for copy
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
static void
copy_fstab(struct fstab *to, struct fstab *from)
{
	free_fstab(to);
	to->fs_spec = strdup(from->fs_spec);
	to->fs_file = strdup(from->fs_file);
	to->fs_vfstype = strdup(from->fs_vfstype);
	to->fs_mntops = strdup(from->fs_mntops);
	to->fs_type = strdup(from->fs_type);
	to->fs_freq = from->fs_freq;
	to->fs_passno = from->fs_passno;
}


/*-----------------------------------------------------------------
 * Name :	find_matching_fs()
 *
 * Description: Find filesystem containing the specified pathname, return
 * true if found.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
static int 
find_matching_fs(const char *path, struct fstab *result)
{
	int found = 0;
	if (setfsent()) {
		int matched_chars = 0;
		struct fstab *fs;
		while ((fs = getfsent())) {
			if (!strncmp(fs->fs_vfstype, our_fstype, strlen(our_fstype))) {
				int len = strlen(fs->fs_file);
				if (!memcmp(fs->fs_file, path, len) && len > matched_chars) {
					matched_chars = len;
					copy_fstab(result, fs);
					found = 1;
				}
			}
		}
		endfsent();
	}
	return found;
}


/*-----------------------------------------------------------------
 * Name :	do_system()
 *
 * Description: system() with variable arguments like in printf
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
static int
do_system(const char *format, ...)
{
	va_list va;
	char command[256];
	va_start(va, format);
	vsprintf(command, format, va);
	va_end(va);
	return system(command);
}


/*-----------------------------------------------------------------
 * Name :	ensure_fs_writable()
 *
 * Description: Remount filesystem containing the file read-write
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
int
ensure_fs_writable(const char *path)
{
	int rv = -1;
	struct fstab fs;
	init_fstab(&fs);
	if (find_matching_fs(path, &fs)) {
		if (!strcmp(fs.fs_type, "ro")) {
			if (lock()) {
				rv = do_system("/sbin/mount -uw -osync %s", fs.fs_file);
			}
			else {
				debug(("Can't obtain lock: %s\n", strerror(errno)));
			}
		}
		else {
			/* Already writable */
			rv = 0;
		}
	}
	else {
		debug(("Could not find matching fs\n"));
	}
	free_fstab(&fs);
	return rv;
}


/*-----------------------------------------------------------------
 * Name :	restore_fs_permissions()
 *
 * Description: Remount the filesystem according to permissions
 * in /etc/fstab.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/13/02
 *
 *-----------------------------------------------------------------
 */
void
restore_fs_permissions(const char *path)
{
	struct fstab fs;
	init_fstab(&fs);
	if (find_matching_fs(path, &fs)) {
		if (!strcmp(fs.fs_type, "ro")) {
			int rv = do_system("/sbin/mount -ur -o %s %s", fs.fs_mntops,
							   fs.fs_file);
			unlock();
			if (rv != 0) {
				syslog(LOG_LOCAL0|LOG_ERR, 
					  "/sbin/mount -o %s -ur %s failed: %d\n",
					   fs.fs_mntops, fs.fs_file, rv);
			}
		}
	}
	else {
		debug(("Could not find matching fs\n"));
	}
	free_fstab(&fs);
}


/*-----------------------------------------------------------------
 * Name :	modify_readonly_file()
 *
 * Description: Make file writable and apply fun with arg to it, restore
 * afterwards.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/15/02
 *
 *-----------------------------------------------------------------
 */
int
modify_readonly_file(const char *path,
					 void *vfun, void *arg)
{
	int rv = -1;
	int (*fun)(const char *, void *) = vfun;
	if (ensure_fs_writable(path) == 0) {
		rv = fun(path, arg);
		restore_fs_permissions(path);
	}
	return rv;
}



/*-----------------------------------------------------------------
 * Name :	make_allfs_writable()
 *
 * Description: Make all filesystems in /etc/fstab writable
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/15/02
 *
 *-----------------------------------------------------------------
 */
int
make_allfs_writable()
{
	int error = 0;
	if (setfsent()) {
		struct fstab *fs;
		while ((fs = getfsent()) && !error) {
			if (!strcmp(fs->fs_vfstype, our_fstype)
				&& !strcmp(fs->fs_type, "ro")) {
				error = do_system("/sbin/mount -uw %s", fs->fs_file);
			}
		}
		endfsent();
	}
	return error;
}



/*-----------------------------------------------------------------
 * Name :	string_to_file()
 *
 * Description: Make file at path containing str.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/15/02
 *
 *-----------------------------------------------------------------
 */
int
string_to_file(const char *path, char *str)
{
	int rv = -1;
	FILE *f = fopen (path, "w");
	if (f) {
		rv = fprintf(f, "%s\n", str) == strlen(str) + 1 ? 0 : -1;
		fclose (f);
	}
	return rv;
}
