/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : opt-routines.c
 *
 * Description  : Contain all routines to handle to handle the options.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <db.h>
#include <string.h>

#include "opt.h"
#include "opt-api.h"
#include "exports/anpssc-api.h"
#include "exports/elog-api.h"

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Prototype of global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

#define MAX_OPTION_MSG_SIZE		4096

typedef struct
{
    char    buf[MAX_OPTION_MSG_SIZE];
    int32   len;
} opt_msgBuf_t;

struct timeval waitRegm = {10, 0};

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : getRegistryOption
 *
 * Description  : Can be called by any application to request the value
 *                of a registry option. This can be used ONLY obtain single
 * 				  value option requests.
 *
 * Created by   : Lan Chin/Pankaj Gupta
 *
 * Date Created : 
 *
 *------------------------------------------------------------------------------
 */

int32 getRegistryOption(char *option, char *value)
{
    opt_msgHdr_t        *hdr;
    opt_getOptionReq_t  *req;
    opt_getOptionRsp_t  *rsp;
    opt_msgBuf_t       	reqMsgBuf;
    opt_msgBuf_t       	rspMsgBuf;
	int32 				optionLen;

	if ((option == NULL) || (value == NULL))
		return NFX_ERR;

	optionLen = strlen(option);

    memset((void *) reqMsgBuf.buf, 0, sizeof(opt_msgBuf_t));
    memset((void *) rspMsgBuf.buf, 0, sizeof(opt_msgBuf_t));

	/*
	 * Prepare request
	 */
    hdr = (opt_msgHdr_t *) reqMsgBuf.buf;
    hdr->msgType = OPT_GET_OPTION_REQ;
    hdr->userInfo = 0; 
    req = (opt_getOptionReq_t *) (reqMsgBuf.buf + sizeof(opt_msgHdr_t));
    memcpy((void *) &req->option, (void *) option, optionLen);

    reqMsgBuf.len = sizeof(opt_msgHdr_t) + optionLen + 1;    
	rspMsgBuf.len = sizeof(opt_msgBuf_t);

	/*
	 * Send request
	 */
	rspMsgBuf.len = sendAgileMsg(-1, (void *) reqMsgBuf.buf, reqMsgBuf.len, 
							"regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                  			&waitRegm, (void *) rspMsgBuf.buf, 
				  			sizeof(opt_msgBuf_t));

	/*
	 * Process response
	 */
    if (rspMsgBuf.len <= 0)
        return NFX_ERR;

    hdr = (opt_msgHdr_t *) rspMsgBuf.buf;
    if (hdr->msgType != OPT_GET_OPTION_RSP)
        return NFX_ERR;

    rsp = (opt_getOptionRsp_t *) (((char*)hdr) + sizeof(opt_msgHdr_t));

	if (rsp->nbEntry != 1)
		return NFX_ERR;

	memcpy((char *) value, rsp->entry[0].value, 
							strlen(rsp->entry[0].value));

    return NFX_OK;

} /* evm_regGetOption */


/*------------------------------------------------------------------------------
 * Name         : setRegistryOption
 *
 * Description  : Can be called by any application to set the value
 *                of a single value specific registry option
 *
 * Created by   : Lan Chin/Pankaj Gupta
 *
 * Date Created : 
 *
 *------------------------------------------------------------------------------
 */

int32 setRegistryOption(char *option, char *value)
{
    opt_msgHdr_t        *hdr;
    opt_setOptionReq_t  *req;
    opt_setOptionRsp_t  *rsp;
    opt_msgBuf_t       	reqMsgBuf;
    opt_msgBuf_t       	rspMsgBuf;
	int32 				optionLen;
	int32				valueLen;

	if ((option == NULL) || (value == NULL))
		return NFX_ERR;

	optionLen = strlen(option);
	valueLen = strlen(value);

    memset((void *) &reqMsgBuf, 0, sizeof(opt_msgBuf_t));
    memset((void *) &rspMsgBuf, 0, sizeof(opt_msgBuf_t));

	/*
	 * Prepare request
	 */
    hdr = (opt_msgHdr_t *) reqMsgBuf.buf;
    hdr->msgType = OPT_SET_OPTION_REQ;
    hdr->userInfo = 0; 

    req = (opt_setOptionReq_t *) (reqMsgBuf.buf + sizeof(opt_msgHdr_t));
    memcpy((void *) req->entry.option, (void *) option, optionLen);
    memcpy((void *) req->entry.value, (void *) value, valueLen);
    req->del = 0;

    reqMsgBuf.len = sizeof(opt_msgHdr_t) + sizeof(opt_setOptionReq_t);
	rspMsgBuf.len = sizeof(opt_msgBuf_t);

	printf("setRegistryOption: %s %s\n", req->entry.option,
					req->entry.value);

	/*
	 * Send request
	 */
	rspMsgBuf.len = sendAgileMsg(-1, (void *) reqMsgBuf.buf, reqMsgBuf.len, 
							"regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                  			&waitRegm, (void *) rspMsgBuf.buf, 
				  			sizeof(opt_msgBuf_t));

	/*
	 * Process response
	 */
    if (rspMsgBuf.len <= 0)
        return NFX_ERR;

    hdr = (opt_msgHdr_t *) rspMsgBuf.buf;
	
    if (hdr->msgType != OPT_GET_OPTION_RSP)
        return NFX_ERR;

    rsp = (opt_setOptionRsp_t *) (((char*)hdr) + sizeof(opt_msgHdr_t));
    if (rsp->result != NFX_OK)
        return NFX_ERR;

    return NFX_OK;
}

/*----------------------------------------------------------------------------*/
