/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: opt-utils.c
 *
 * RCS:  $Id: opt_utils.c,v 1.2 2002/08/01 17:56:42 sandrineb Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Does the I/O for talking to registryMgr to get/store
 *              configuration key/data pairs.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 19 Mar 2002
 *
 *-----------------------------------------------------------------
 */
#undef DEBUG
#if defined (DEBUG)
#include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>

#include "exports/opt-api.h"
#include "exports/anpssc-api.h"

#define REGM_TO_SEC 5

static int mypid = -1; 	// Don't really care
/*-----------------------------------------------------------------
 * Name: get_cfg_val
 *
 * Description: Get configuration objects matching key
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Mar 2002
 *
 *-----------------------------------------------------------------
 */
#define BUF_SIZE	8192
extern char **
get_cfg_val (char *key)
{
	regmBigGetReq_t *req;
	regmBigGetRsp_t *rsp;
	int rc;
	struct timeval regmTo = {REGM_TO_SEC, 0};
	char **list;
	int i;
	int buf_size = BUF_SIZE;
	char *st_ptr, *end_ptr;

	i = sizeof(regmBigGetReq_t) + strlen (key);
	req = malloc(i);

	req->hdr.msgType = OPT_GET_BIGOPT_REQ;
	req->hdr.userInfo = mypid;
	strcpy (req->opt, key);
/*
	We want to get the whole thing not just part. If the returned byte
	count is the same size as our buffer, we assume it overflowed, make it
	bigger, and try again.
 */
	for(;;) {
		rsp = malloc (buf_size);

		rc = sendAgileMsg (-1, req, i,
		                   "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
		                   rsp, buf_size);

		if (rc != buf_size) {	// Get the whole thing (or none at all)?
			break;
		}

		free (rsp);
		buf_size += BUF_SIZE;
	}

	free (req);

	if (!rc) {
#if defined (DEBUG)
		fprintf (stderr, "Timeout waiting for response from regm\n");
#endif
		list =  NULL;
	}

	else if (!rsp->nbEntry) {
		list =  NULL;	// No matches
	}
/*
	Create and populate the list of key/data pairs
 */
	else {
		i = sizeof(char *) * 2 * (rsp->nbEntry + 1);
		list = malloc (i);
		memset (list, 0, i);

		end_ptr = (char *)rsp + buf_size;	// Don't pass this point
		st_ptr = rsp->opt_val;

		for (i = 0; i < rsp->nbEntry; i++) {
			list[2*i] = strdup (st_ptr);
			st_ptr += strlen (st_ptr) + 1;
			list[(2*i)+1] = strdup (st_ptr);
			st_ptr += strlen (st_ptr) + 1;
		}
	}
	free (rsp);
	return list;
}

/*-----------------------------------------------------------------
 * Name:free_cfg_val
 *
 * Description: Free the buffer space allocated via get_cfg_val
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Mar 2002
 *
 *-----------------------------------------------------------------
 */
extern void
free_cfg_val (char **list)
{
	char **str;

	if (!list) {
		return;
	}
	str = list;

	while (*str) {
		free (*str);
		str++;
	}
	free (list);
}

/*-----------------------------------------------------------------
 * Name: put_cfg_val, del_cfg_obj
 *
 * Description: Put/delete a config object in the configuration database
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Mar 2002
 *
 *-----------------------------------------------------------------
 */
static int
mod_cfg_val (char *obj[], int del)
{
	regmBigSetReq_t *req;
	regmSetRsp_t rsp;
	struct timeval regmTo = {REGM_TO_SEC, 0};
	int rc;
	int rqsize;
/*
	We're passed a key/data pair
 */
	if (!obj || !obj[0] || (!del && !obj[1])) {
		return 1;
	}

	rqsize = sizeof (regmBigSetReq_t) + strlen (obj[0]);
	if (!del) {
		rqsize += strlen (obj[1]) + 1;
	}
	req = malloc (rqsize);
	
	req->hdr.msgType = OPT_SET_BIGOPT_REQ;
	req->hdr.userInfo = mypid;
	req->del = del;
	strcpy (req->opt_val, obj[0]);
	if (!del) {
		strcpy (req->opt_val + strlen (obj[0]) + 1, obj[1]);
	}

	rc = sendAgileMsg (-1, req, rqsize,
	                   "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
	                   &rsp, sizeof(rsp));

	free (req);

	if ((rc != sizeof(rsp)) || (rsp.setRsp.result != NFX_OK)) {
#if defined (DEBUG)
		fprintf (stderr, "mod_cfg_val fail: rc %d result %d\n", 
		                  rc, rsp.setRsp.result);
#endif
		return 1;
	}
	else {
		return 0;
	}
}

extern int
put_cfg_val (char **objs)
{
	return mod_cfg_val (objs, 0);
}

extern int
del_cfg_obj (char **objs)
{
	return mod_cfg_val (objs, 1);
}

/*-----------------------------------------------------------------
 * Name: node_is_PCC
 *
 * Description: Determines if this node is a PCC/Standalone
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 Mar 2002
 *
 *-----------------------------------------------------------------
 */
extern int
node_is_PCC()
{
	char **type;
	int isPcc;
/*
	Is it a stand-alone node?
 */
	type = get_cfg_val("cluster.nodetype");
	if (!type) {
		return 1;
	}
	free_cfg_val (type);
/*
	Is it now the PCC?
 */
	type = get_cfg_val("node.type");

	isPcc = (type && !strcasecmp (type[1], "pcc"));

	free_cfg_val (type);

	return isPcc;
}

