/*
 *-----------------------------------------------------------------
 *
 * Name:	pci-api.c
 *
 * RCS:  $Id: pci-galileo.c,v 1.7 2002/03/23 17:52:19 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions for PCI operations
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/22/01
 *
 *-----------------------------------------------------------------
 */

#if !defined(LINUX_TEST) && !defined(SIBYTE)

#include <nfx-incl.h>
#include <exports/pci-api.h>
#include <exports/hwMap.h>
#ifdef R9000
#include <sys/gt64440r.h>
#else
#include <sys/gt64240r.h>
#endif
#include "pci.h"

/*-----------------------------------------------------------------
 * Name:	pci_gt_cfg_read()
 *
 * Description:	Generate a PCI read configuration cycle
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/22/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK if data was read successfully
							 * PCI_ERR_BAD_PARAM
							 * PCI_ERR_MASTER_ABORT
							 * PCI_ERR_TARGET_ABORT
							 * PCI_ERR_PARITY_ERROR
							 */
pci_gt_cfg_read(
	uint32	bus,
	uint32	dev,
	uint32	func,
	uint32	reg,
	uint32	* data
)
{
	uint32	* addr_reg, * data_reg;
	uint32	addr;
	int32	rc;
	uint32	cfg_bus;

/*
	printf("[pci_gt_cfg_read], bus %d, dev %d, reg 0x%x\n", bus, dev, reg);
*/

	if ((bus >= PCI_NUM_BUSSES) ||
		(dev >= PCI_NUM_DEVS) ||
		(func >= PCI_NUM_FUNCS) ||
		(reg >= PCI_NUM_REGS))
		return PCI_ERR_BAD_PARAM;

	/*
	 * (0 <= bus < 8) on PCI0
	 * (8 <= bus < 16) on PCI1
	 * (bus >= 16) invalid
	 */
	if (bus < 8) {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_0CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER);
		cfg_bus = bus;
	}
	else {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_1CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_1CONFIGURATION_DATA_VIRTUAL_REGISTER);
		cfg_bus = bus - 8;
	}

	addr = WORD_SWAP(	(cfg_bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(func << PCI_CFG_FUNC_SHIFT) |
						(reg) |
						PCI_CFG_ENABLE);

	*addr_reg = addr;
	*data = WORD_SWAP(*data_reg);

	delay_us(1000);

	/*
	 * check status of the controller, and clear errors, if any
	 */
	if (((bus != 0) && (bus != 8)) || (dev != 0)) {
		if ((rc = pci_gt_cfg_check_status(bus, 0)) != NFX_OK) {
			pci_gt_cfg_clear_status(bus, 0);
			return rc;
		}
	}

	return NFX_OK;

}	/* pci_gt_cfg_read() */

/*-----------------------------------------------------------------
 * Name:	pci_gt_cfg_write()
 *
 * Description:	Generate a PCI write configuration cycle
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/22/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK if data was written successfully
							 * PCI_ERR_BAD_PARAM
							 * PCI_ERR_MASTER_ABORT
							 * PCI_ERR_TARGET_ABORT
							 * PCI_ERR_PARITY_ERROR
							 */
pci_gt_cfg_write(
	uint32	bus,
	uint32	dev,
	uint32	func,
	uint32	reg,
	uint32	data
)
{
	uint32	* addr_reg, * data_reg;
	uint32	addr;
	int32	rc;
	uint32	cfg_bus;

/*
	printf("[pci_gt_cfg_write], bus %d, dev %d, reg 0x%x, data 0x%x\n",
														bus, dev, reg, data);
*/
	if ((bus >= PCI_NUM_BUSSES) ||
		(dev >= PCI_NUM_DEVS) ||
		(func >= PCI_NUM_FUNCS) ||
		(reg >= PCI_NUM_REGS))
		return PCI_ERR_BAD_PARAM;

	/*
	 * (0 <= bus < 8) on PCI0
	 * (8 <= bus < 16) on PCI1
	 * (bus >= 16) invalid
	 */
	if (bus < 8) {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_0CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER);
		cfg_bus = bus;
	}
	else {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_1CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_1CONFIGURATION_DATA_VIRTUAL_REGISTER);
		cfg_bus = bus - 8;
	}

	addr = WORD_SWAP(	(cfg_bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(func << PCI_CFG_FUNC_SHIFT) |
						(reg) | 
						PCI_CFG_ENABLE);

	*addr_reg = addr;
	*data_reg = WORD_SWAP(data);

	delay_us(1000);

	/*
	 * check status of the controller, and clear errors, if any
	 */
	if (((bus != 0) && (bus != 8)) || (dev != 0)) {
		if ((rc = pci_gt_cfg_check_status(bus, 0)) != NFX_OK) {
			pci_gt_cfg_clear_status(bus, 0);
			return rc;
		}
	}

	return NFX_OK;

}	/* pci_gt_cfg_write() */

/*-----------------------------------------------------------------
 * Name:	pci_gt_cfg_check_status()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/22/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * PCI_ERR_MASTER_ABORT
								 * PCI_ERR_TARGET_ABORT
								 * PCI_ERR_PARITY_ERROR
								 */
pci_gt_cfg_check_status(
	uint32	bus,
	uint32	dev
)
{
	uint32	* addr_reg, * data_reg;
	uint32	addr;
	uint32	stat;
	uint32	reg;
	uint32	cfg_bus;

#ifdef R9000
	dev = (bus < PCI1_PRI_BUS) ? PCI0_GT_DEV : PCI1_GT_DEV;
	bus = (bus < PCI1_PRI_BUS) ? PCI0_PRI_BUS : PCI1_PRI_BUS;
	reg = (bus < PCI1_PRI_BUS) ? PCI0_COMMAND_STATUS_REG : 
								 PCI1_COMMAND_STATUS_REG;
#else
	reg = PCI0_COMMAND_STATUS_REG;
#endif

	/*
	 * (0 <= bus < 8) on PCI0
	 * (8 <= bus < 16) on PCI1
	 * (bus >= 16) invalid
	 */
	if (bus < 8) {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_0CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER);
		cfg_bus = bus;
	}
	else {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_1CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_1CONFIGURATION_DATA_VIRTUAL_REGISTER);
		cfg_bus = bus - 8;
	}

	addr = WORD_SWAP(	(cfg_bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(0 << PCI_CFG_FUNC_SHIFT) |
						(reg) |
						PCI_CFG_ENABLE);

	*addr_reg = addr;
	stat = WORD_SWAP(*data_reg);
/*
	printf("gt_cfg_check, status = 0x%x\n", stat);
*/

	if (stat & PCI_CFG_STAT_MABORT) {
/*
		printf("!!!!!!!!!!!!!!MASTER ABORT, bus = %d, dev = %d\n", bus, dev);
*/
		return PCI_ERR_MASTER_ABORT;
	}
	if (stat & PCI_CFG_STAT_TABORT) {
/*
		printf("!!!!!!!!!!!!!!TARGET ABORT, bus = %d, dev = %d\n", bus, dev);
*/
		return PCI_ERR_TARGET_ABORT;
	}
	if (stat & PCI_CFG_STAT_PERR) {
/*
		printf("!!!!!!!!!!!!!!PARITY ERROR, bus = %d, dev = %d\n", bus, dev);
*/
		return PCI_ERR_PARITY_ERROR;
	}

	return NFX_OK;

}	/* pci_gt_cfg_check_status() */

/*-----------------------------------------------------------------
 * Name:	pci_gt_cfg_clear_status()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	6/22/01
 *
 *-----------------------------------------------------------------
 */
void
pci_gt_cfg_clear_status(
	uint32	bus,
	uint32	dev
)
{
	uint32	* addr_reg, * data_reg;
	uint32	addr, t;
	uint32	reg;

#ifdef XXXX
	dev = (bus < 8) ? 1 : 3;
	bus = (bus < 8) ? 0 : 8;
	reg = (bus < 8) ? PCI0_COMMAND_STATUS_REG : PCI1_COMMAND_STATUS_REG;
#endif

#ifdef R9000
	dev = (bus < PCI1_PRI_BUS) ? PCI0_GT_DEV : PCI1_GT_DEV;
	bus = (bus < PCI1_PRI_BUS) ? PCI0_PRI_BUS : PCI1_PRI_BUS;
	reg = (bus < PCI1_PRI_BUS) ? PCI0_COMMAND_STATUS_REG : 
								 PCI1_COMMAND_STATUS_REG;
#else
	reg = PCI0_COMMAND_STATUS_REG;
#endif

	/*
	 * (0 <= bus < 8) on PCI0
	 * (8 <= bus < 16) on PCI1
	 * (bus >= 16) invalid
	 */
	if (bus < 8) {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_0CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_0CONFIGURATION_DATA_VIRTUAL_REGISTER);
	}
	else {
		addr_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
												PCI_1CONFIGURATION_ADDRESS);
		data_reg = (uint32 *)PHYS_TO_K1(HW_GALILEO_BASE +
									PCI_1CONFIGURATION_DATA_VIRTUAL_REGISTER);
		bus -= 8;
	}

	addr = WORD_SWAP(	(bus << PCI_CFG_BUS_SHIFT) |
						(dev << PCI_CFG_DEV_SHIFT) |
						(0 << PCI_CFG_FUNC_SHIFT) |
						(reg) |
						PCI_CFG_ENABLE);

	*addr_reg = addr;
	t = WORD_SWAP(*data_reg);
	*data_reg = WORD_SWAP(t | 0xF9000000);

}	/* pci_gt_cfg_clear_status() */

#endif	/* !LINUX_TEST && !SIBYTE */

