/*
 *-----------------------------------------------------------------
 *
 * Name:	pci-sb1250.c
 *
 * RCS:  $Id: pci-sb1250.c,v 1.7 2002/01/19 00:16:26 lipikab Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions for PCI operations
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/30/01
 *
 *-----------------------------------------------------------------
 */

#if defined(SIBYTE)

#include <nfx-incl.h>
#include <exports/pci-api.h>
#include <sys/sb1250/sbmips.h>
#include "pci.h"

extern void hs_write32(uint64 addr, uint32 val);
extern uint32 hs_read32(uint64 addr);

/*-----------------------------------------------------------------
 * Name:	pci_sb1250_cfg_read()
 *
 * Description:	Generate a PCI read configuration cycle
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/30/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK if data was read successfully
							 * PCI_ERR_BAD_PARAM
							 * PCI_ERR_MASTER_ABORT
							 * PCI_ERR_TARGET_ABORT
							 * PCI_ERR_PARITY_ERROR
							 */
pci_sb1250_cfg_read(
	uint32	bus,
	uint32	dev,
	uint32	func,
	uint32	reg,
	uint32	* data
)
{
	uint64	addr;
	int32	rc;
	uint32	tag;
	uint32	d;

/*
	printf("[pci_sb1250_cfg_read], bus %d, dev %d, reg 0x%x = ", bus, dev, reg);
*/

	if ((bus >= PCI_NUM_BUSSES) ||
		(dev >= PCI_NUM_DEVS) ||
		(dev >= PCI_NUM_FUNCS) ||
		(reg >= PCI_NUM_REGS))
		return PCI_ERR_BAD_PARAM;

	tag =	(bus << PCI_CFG_BUS_SHIFT) |
			(dev << PCI_CFG_DEV_SHIFT) |
			(func << PCI_CFG_FUNC_SHIFT) |
			(reg);

	addr = 0x90000000fe000000 + tag;

	mips_wbflush();

	d = hs_read32(addr);

	mips_wbflush();

	* data = d;

/*
	printf("data = 0x%x\n", d);
*/

	/*
	 * check status of the controller, and clear errors, if any
	 */
	if ((bus == 0) && ((dev != 0) && (dev != 1))) {
		if ((rc = pci_sb1250_cfg_check_status(0, 0)) != NFX_OK) {
			pci_sb1250_cfg_clear_status(0, 0);
			return rc;
		}
	}
	else if (bus != 0) {
		if ((rc = pci_sb1250_cfg_check_status(0, 1)) != NFX_OK) {
			pci_sb1250_cfg_clear_status(0, 1);
			return rc;
		}
	}

	return NFX_OK;

}	/* pci_sb1250_cfg_read() */

/*-----------------------------------------------------------------
 * Name:	pci_sb1250_cfg_write()
 *
 * Description:	Generate a PCI write configuration cycle
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/30/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK if data was written successfully
							 * PCI_ERR_BAD_PARAM
							 * PCI_ERR_MASTER_ABORT
							 * PCI_ERR_TARGET_ABORT
							 * PCI_ERR_PARITY_ERROR
							 */
pci_sb1250_cfg_write(
	uint32	bus,
	uint32	dev,
	uint32	func,
	uint32	reg,
	uint32	data
)
{
	uint64	addr;
	int32	rc;
	uint32	tag;

/*
	printf("[pci_sb1250_cfg_write], bus %d, dev %d, reg 0x%x, data 0x%x\n",
														bus, dev, reg, data);
*/
	if ((bus >= PCI_NUM_BUSSES) ||
		(dev >= PCI_NUM_DEVS) ||
		(dev >= PCI_NUM_FUNCS) ||
		(reg >= PCI_NUM_REGS))
		return PCI_ERR_BAD_PARAM;

	tag =	(bus << PCI_CFG_BUS_SHIFT) |
			(dev << PCI_CFG_DEV_SHIFT) |
			(func << PCI_CFG_FUNC_SHIFT) |
			(reg);

	addr = 0x90000000fe000000 + tag;

	mips_wbflush();

	hs_write32(addr, data);

	mips_wbflush();

	/*
	 * check status of the controller, and clear errors, if any
	 */
	if (((bus != 0) && (bus != 1)) || (dev != 0)) {
		if ((rc = pci_sb1250_cfg_check_status(bus, dev)) != NFX_OK) {
			pci_sb1250_cfg_clear_status(0, 0);
			return rc;
		}
	}

	return NFX_OK;

}	/* pci_sb1250_cfg_write() */

/*-----------------------------------------------------------------
 * Name:	pci_sb1250_cfg_check_status()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/30/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * PCI_ERR_MASTER_ABORT
								 * PCI_ERR_TARGET_ABORT
								 * PCI_ERR_PARITY_ERROR
								 */
pci_sb1250_cfg_check_status(
	uint32	bus,
	uint32	dev
)
{
	uint64	addr;
	uint32	tag;
	uint32	stat;

	tag =	(bus << PCI_CFG_BUS_SHIFT) |
			(dev << PCI_CFG_DEV_SHIFT) |
			(0 << PCI_CFG_DEV_SHIFT) |
			4;

	addr = 0x90000000fe000000 + tag;

	mips_wbflush();

	stat = hs_read32(addr);

	mips_wbflush();

/*
	printf("sb1250_cfg_check, status = 0x%x\n", stat);
*/

	if (stat & PCI_CFG_STAT_MABORT) {
		printf("!!!!!!!!!!!!!!MASTER ABORT, bus = %d, dev = %d\n", bus, dev);
		return PCI_ERR_MASTER_ABORT;
	}
	if (stat & PCI_CFG_STAT_TABORT) {
		printf("!!!!!!!!!!!!!!TARGET ABORT, bus = %d, dev = %d\n", bus, dev);
		return PCI_ERR_TARGET_ABORT;
	}
	if (stat & PCI_CFG_STAT_SIG_TABORT) {
		printf("!!!!!!!!!!!!!SIG TARGET ABORT, bus = %d, dev = %d\n", bus, dev);
		return PCI_ERR_TARGET_ABORT;
	}
	if (stat & PCI_CFG_STAT_PERR) {
		printf("!!!!!!!!!!!!!!PARITY ERROR, bus = %d, dev = %d\n", bus, dev);
		return PCI_ERR_PARITY_ERROR;
	}

	return NFX_OK;

}	/* pci_sb1250_cfg_check_status() */

/*-----------------------------------------------------------------
 * Name:	pci_sb1250_cfg_clear_status()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/30/01
 *
 *-----------------------------------------------------------------
 */
void
pci_sb1250_cfg_clear_status(
	uint32	bus,
	uint32	dev
)
{
	uint64	addr;
	uint32	tag;
	uint32	stat;

	tag =	(bus << PCI_CFG_BUS_SHIFT) |
			(dev << PCI_CFG_DEV_SHIFT) |
			(0 << PCI_CFG_DEV_SHIFT) |
			4;

	addr = 0x90000000fe000000 + tag;

	mips_wbflush();

	stat = hs_read32(addr);

	mips_wbflush();

	stat |= 0xf9000000;

	hs_write32(addr,stat);

	mips_wbflush();

}	/* pci_sb1250_cfg_clear_status() */

#endif	/* SIBYTE */
