/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: pkt.c
 *
 * RCS:  $Id: pkt-free.c,v 1.14 2002/09/17 23:34:46 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains Packet Processing layer related routines
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 6/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/pkt-api.h>
#include <exports/tpl-api.h>

/*-----------------------------------------------------------------
 * Name:        pkt_validateDesc
 *
 * Description: Validate PKT Descriptor to be freed.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
int32
pkt_validateDesc( pkt_desc_t *pDesc, int32 chkDup)
{

	/*-------------------------------
	 * Check for valid desc. value.
	 *-------------------------------
	 */
    if ((pDesc == NULL)		 	  ||
		(pDesc < pPktCfg->descStart) ||
		(pDesc > pPktCfg->descEnd))
	{
		pPktCfg->stats.cntBadDescErr++;
        return PKT_ERR_BAD_DESC;
	}

	/*-------------------------------
	 * Check for duplicate free.
	 *-------------------------------
	 */
	if (chkDup && !(pDesc->flags & PKT_DESC_INUSE))
	{
		pPktCfg->stats.cntDupDescErr++;
		return PKT_ERR_DUP_DESC;	
	}

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:        pkt_freeDesc
 *
 * Description: Free a given PKT Descriptor. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
int32 
pkt_freeDesc(pkt_desc_t *pDesc, char *filename, int32 linenum)
{
	pkt_descRing_t	*pRing;
	int32			rc;

	pRing  = &pPktCfg->freeDescRing;

#ifdef PKT_USE_TIME_STAMP
        pkt_freeTimeStamp(pDesc->pTime);
        pDesc->pTime = NULL;
#endif

	/*-------------------------------
	 * Validate PKT descriptor.
	 *-------------------------------
	 */
#ifdef PKT_VALIDATE_ENABLE
    rc = pkt_validateDesc( pDesc, PKT_CHK_DUP );
    if (rc != NFX_OK)
        return rc;
#endif

	/*-------------------------------
	 * Check for free desc ring full.
	 *-------------------------------
	 */
	if (((pRing->wrAddr+1) == (pRing->rdAddr)) ||
		((pRing->rdAddr == pRing->startAddr) &&
		 (pRing->wrAddr == pRing->endAddr)))
	{
		pPktCfg->stats.cntDescRingFull++;
		return PKT_ERR_DESC_RING_FULL;
	}

	/*-------------------------------
	 * Clr desc flag & write free desc.
	 *-------------------------------
	 */
	pDesc->flags    = 0;
	pDesc->pNext    = 0;
	pDesc->dataSize = 0;
	pDesc->padSize  = 0;
	pDesc->freeFile = filename;
	pDesc->freeLine = linenum;
	*pRing->wrAddr  = pDesc;

	/*-------------------------------
	 * Update wr ptr & handle wrap.
	 *-------------------------------
	 */
	if (pRing->wrAddr == pRing->endAddr)
		pRing->wrAddr = pRing->startAddr;
	else
		pRing->wrAddr++; 

	pRing->avail++;

	pPktCfg->stats.cntFreeDesc++;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        pkt_freePkt
 *
 * Description: Free a given Packet (both desc & buffer).
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
int32	
pkt_freePkt(pkt_desc_t *pktDesc, char *filename, int32 linenum)
{
	pkt_desc_t	*pDesc;
	pkt_desc_t	*pNext;
	eee_desc_t	*eDescp;
	int32		rc;

	pDesc = pktDesc;

#if defined(NFX_SMP)
	if (!CM_IS_TXRX_NCPU) {
		return pkt_AcpuFreePkt(pDesc, filename, linenum);
	}
#endif /* NFX_SMP */

	do 
	{
		eDescp = pDesc->eDescp;
		pNext  = pDesc->pNext;

		rc = pkt_freeDesc(pDesc, filename, linenum);
		if (rc != NFX_OK)
		{
			pkt_logErr( filename, linenum, rc, pDesc);
			return rc;
		}

#ifdef	PROM
		if (!eDescp)
		{
			pkt_logErr( filename, linenum, rc, pDesc);
			return(PKT_ERR_BAD_DESC);
		}

		/*-------------------------------
		 * If free func idx is set, pkt_freePkt 
		 * is called from eee_freePacket as 
		 * part of Driver TX Complete. 
		 *
		 * DO NOT call eee_freePacket again.
		 *
		 * eDescp with be released by 
		 * eee_freePacket when we return.
		 *-------------------------------
		 */
		if (!(eDescp->hdr.control & EDESC_FREE_ROUTINE_INDEX))
		{
			/*-------------------------------
			 * free idx not set, MUST be a RX pkt.
			 * Call eee_freePacket to release it.
			 *-------------------------------
			 */
			pPktCfg->stats.cntFreeBuff++;
			eee_freePacket(eDescp);
		}
#else
		if (eDescp)
        {
    		eee_freePacket(eDescp);
    		pPktCfg->stats.cntFreeBuff++;
        }
#endif

		pPktCfg->stats.cntFreePkt++;

		pDesc = pNext;
	} while(pDesc);

	return (NFX_OK);
}
