/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: pkt.c
 *
 * RCS:  $Id: pkt-utils.c,v 1.24 2002/11/01 00:23:42 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains Packet Processing layer related routines
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 6/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/malloc-api.h>

/*-------------------------------
 * PKT error control block.
 *-------------------------------   
 */ 
typedef struct _pkt_errCb_t
{   
    char    *errMsg;
    int32   errCode;
    int32   action;
} pkt_errCb_t;

#define PKT_DO_PANIC        1
#define PKT_DO_NOTHING      0
 
static pkt_errCb_t errCb[ ]=
{
    {"Bad pkt buffer",   			PKT_ERR_BAD_BUFF,  		PKT_DO_PANIC},
    {"Bad pkt descriptor", 			PKT_ERR_BAD_DESC,  		PKT_DO_PANIC},
    {"Duplicate pkt buffer",		PKT_ERR_DUP_BUFF,  		PKT_DO_PANIC},
    {"Duplicate pkt descriptor",	PKT_ERR_DUP_DESC,  		PKT_DO_PANIC},
    {"No pkt buffer",				PKT_ERR_NO_BUFF,  		PKT_DO_PANIC},
    {"No pkt descriptor",			PKT_ERR_NO_DESC,  		PKT_DO_PANIC},
    {"Pkt buffer ring full",		PKT_ERR_BUFF_RING_FULL,	PKT_DO_PANIC},
    {"Pkt desc. ring full",			PKT_ERR_DESC_RING_FULL,	PKT_DO_PANIC},
    {"Unknown Error",               PKT_ERR_MAX,            PKT_DO_PANIC},
};


/*-----------------------------------------------------------------
 * Name:        pkt_showStats
 *
 * Description: Shows Pkt related statitics.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 11/01/01
 *-----------------------------------------------------------------
 */
void 
pkt_showStats( void )
{
	pkt_cfg_t		*pCfg;

	pCfg  = pPktCfg;


	printf("Pkt Stats \n");
	printf("---------------------------------\n");
	printf("  Bad Buff Err:    %12u\n", pCfg->stats.cntBadBuffErr);
	printf("  Bad Desc Err:    %12u\n", pCfg->stats.cntBadDescErr);
	printf("  Dup Buff Err:    %12u\n", pCfg->stats.cntDupBuffErr);
	printf("  Dup Desc Err:    %12u\n", pCfg->stats.cntDupDescErr);
	printf("  No  Buff Err:    %12u\n", pCfg->stats.cntNoBuffErr);
	printf("  No PDesc Err:    %12u\n", pCfg->stats.cntNoPktDescErr);
	printf("  No EDesc Err:    %12u\n", pCfg->stats.cntNoEeeDescErr);
	printf("  Ful Buff Err:    %12u\n", pCfg->stats.cntBuffRingFull);
	printf("  Ful Desc Err:    %12u\n", pCfg->stats.cntDescRingFull);
	printf("  Alloc Buff  :    %12u\n", pCfg->stats.cntAllocBuff);
	printf("  Alloc Desc  :    %12u\n", pCfg->stats.cntAllocDesc);
	printf("  Alloc Pkt   :    %12u\n", pCfg->stats.cntAllocPkt);
	printf("  Alloc Chain :    %12u\n", pCfg->stats.cntAllocChainPkt);
	printf("  Free  Buff  :    %12u\n", pCfg->stats.cntFreeBuff);
	printf("  Free  Desc  :    %12u\n", pCfg->stats.cntFreeDesc);
	printf("  Free  Pkt   :    %12u\n", pCfg->stats.cntFreePkt);
	printf("  Free  Chain :    %12u\n", pCfg->stats.cntFreeChainPkt);
	printf("  Desc  Avail :    %12u\n", pCfg->freeDescRing.avail);
	printf("  Desc  LWM   :    %12u\n", pCfg->freeDescRing.lowWaterMark);
	printf("  Cod  Len Err:    %12u\n", pCfg->stats.cntBadCodLen);
	printf("  Copy Len Err:    %12u\n", pCfg->stats.cntBadCopyLen);
//	printf("\tDesc MemSize:    %12u\n", sizeof(pkt_desc_t) *  PKT_MAX_DESC_CNT);

}

/*-----------------------------------------------------------------
 * Name:        pkt_showCfg
 *
 * Description: Shows Pkt Config Control data structure.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
void 
pkt_showCfg( void )
{
	pkt_cfg_t	*pCfg;

	pCfg = pPktCfg;

	printf("Pkt Config: 0x%p\n", pCfg);
	printf("---------------------------------\n");
	printf("\tDesc. Start=0x%p End=0x%p\n",
			pCfg->descStart, pCfg->descEnd);
	printf("\tBuff. Start=0x%p End=0x%p\n",
			pCfg->buffStart, pCfg->buffEnd);

	pkt_showStats( );
	pkt_showDescRing( &pCfg->freeDescRing);
	pkt_showBuffRing( &pCfg->freeBuffRing);
}

/*-----------------------------------------------------------------
 * Name:        pkt_showDescRing
 *
 * Description: Shows Pkt Descriptor Ring Control data structure.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
void 
pkt_showDescRing ( pkt_descRing_t *pRing )
{
	pkt_desc_t	**ptr;
	pkt_desc_t	*pDesc;
	int32		count;

	printf("Pkt Desc Ring :0x%p\n", pRing);
	printf("\tStart=0x%p End=0x%p Rd=0x%p Wr=0x%p Avail=%d\n",
		pRing->startAddr, pRing->endAddr, pRing->rdAddr, pRing->wrAddr,
		pRing->avail);
	printf("\n");

	count = 0;
	ptr = pRing->startAddr;

	/* su, 10/17/02 
	 * Ring can be empty 
	 */
	if (!ptr)
		return;

	while ( ptr <= pRing->endAddr)
	{
		pDesc = *(ptr);
		printf("\t[%03d] Entry=0x%p Desc=0x%p", count, ptr, pDesc);
		if (ptr == pRing->rdAddr)
			printf("<----Rd\n");
		else if (ptr == pRing->wrAddr)
			printf("<----Wr\n");
		else 
			printf("\n");

		ptr++;
		count++;
	}
	printf("\n");
}

/*-----------------------------------------------------------------
 * Name:        pkt_showBuffRing
 *
 * Description: Shows Pkt Buffer Ring Control data structure.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
void 
pkt_showBuffRing ( pkt_buffRing_t *pRing )
{
	char	**ptr;
	char	*pBuff;
	int32	count;

	printf("Pkt Buff Ring :0x%p\n", pRing);
	printf("\tStart=0x%p End=0x%p Rd=0x%p Wr=0x%p Avail=%d\n",
		pRing->startAddr, pRing->endAddr, pRing->rdAddr, pRing->wrAddr,
		pRing->avail);
	printf("\n");

	count = 0;
	ptr = pRing->startAddr;

	while ( ptr <= pRing->endAddr)
	{
		pBuff = *(ptr);
		printf("\t[%03d] Entry=0x%p Buff=0x%p", count, ptr, pBuff);
		if (ptr == pRing->rdAddr)
			printf("<----Rd\n");
		else if (ptr == pRing->wrAddr)
			printf("<----Wr\n");
		else 
			printf("\n");

		ptr++;
		count++;
	}
	printf("\n");
}

/*-----------------------------------------------------------------
 * Name:        pkt_showDesc
 *
 * Description: Shows a single Pkt Descriptor structure. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
void 
pkt_showDesc( pkt_desc_t *pDesc )
{
	printf("PKT DESC: %p\n", pDesc);
	printf("\tflags    = %04x\n", pDesc->flags);
	printf("\tlink     = %04x\n", pDesc->link);
	printf("\thdrSize  = %04x\n", pDesc->hdrSize);
	printf("\tdataSize = %04x\n", pDesc->dataSize);
	printf("\tpHdr     = %p\n", pDesc->pHdr);
	printf("\tpMac     = %p\n", pDesc->pMac);
	printf("\tcod      = %llx\n", pDesc->cod);
	printf("\tpTplConn = %p\n", pDesc->pTplConn);
	printf("\tpAppInfo = %p\n", pDesc->appInfo);
	printf("\tpNext    = %p\n", pDesc->pNext);
#ifdef NOTUSED
	printf("\thdrOffset[0] = %04x\n", pDesc->hdrOffset[0]);
	printf("\thdrOffset[1] = %04x\n", pDesc->hdrOffset[1]);
	printf("\thdrOffset[2] = %04x\n", pDesc->hdrOffset[2]);
	printf("\thdrOffset[3] = %04x\n", pDesc->hdrOffset[3]);
#endif
}

/*-----------------------------------------------------------------
 * Name:        pkt_logErr
 *
 * Description: Log Pkt Errors. Panic if PKT_DEBUG is defined.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 6/01/01
 *-----------------------------------------------------------------
 */
void
pkt_logErr( char *fileName, int32 lineNum, int32 errCode, pkt_desc_t *pDesc)
{
	pkt_errCb_t	*pErrCb;
	int32		msgIdx;
	int32		maxIdx;

    /*-------------------------------
     * Get msg Id relative to base.
     *-------------------------------
     */
    msgIdx = ((PKT_ERROR_BASE) - errCode);
    maxIdx = ((PKT_ERROR_BASE) - PKT_ERR_MAX);

    /*-------------------------------
     * Make sure err code is in range.
     *-------------------------------
     */
    if ((msgIdx > maxIdx) || (msgIdx < 0))
        msgIdx = maxIdx;

    /*-------------------------------
     * Get error CB.
     *-------------------------------
     */
    pErrCb = &errCb[msgIdx];

    printf("PKT Error (%d): %s!\n", errCode, pErrCb->errMsg); 
    printf("\tFile: %s Line: %d pDesc:%p\n", fileName, lineNum, pDesc );

	if (errCode == PKT_ERR_DUP_DESC)
	{
		printf("\tAllocated at: %s line:%d\n", 
				pDesc->allocFile, pDesc->allocLine);
		printf("\t1st Free at : %s line:%d\n", 
				pDesc->freeFile, pDesc->freeLine);
		printf("\t2nd Free at : %s line:%d\n", 
				fileName, lineNum);
	}

    if (pErrCb->action == PKT_DO_PANIC)
        PKT_PANIC()

}

/*-----------------------------------------------------------------
 * Name: pkt_AdjPkt
 *
 * Description: adjust pkt desc chain
 *
 * Created by:  Su Wang
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32                           /* NFX_OK, NFX_ERR */
pkt_AdjPkt(pkt_desc_t *pdesc,
           int offset           /* forward, - bckward */
           )
{
  pkt_desc_t *p;

  if (offset > 0)               /* adjust at start */
    {
      while (pdesc)
        {
          if (offset <= pdesc->hdrSize)
            {
              pdesc->pHdr += offset;
              pdesc->hdrSize -= offset;
              break;
            }
          else if (offset <= (pdesc->hdrSize + pdesc->dataSize) &&
                   offset > pdesc->hdrSize)
            {
              offset -= pdesc->hdrSize;
              pdesc->hdrSize = 0;
              pdesc->dataSize -= offset;
              pdesc->offset = offset;
              break;
            }
          else
            {
              offset -= pdesc->hdrSize + pdesc->dataSize;
			  pdesc->hdrSize = 0;
			  pdesc->dataSize = 0;
              pdesc = pdesc->pNext;
            }
        }
    }  
  else                          /* adjust at end */
    {                           /* will be slow because no bck ptr */
      int count = 0, sum = 0;

      offset = -offset;
      /* calculate total lenth of data in chain */
      for(p = pdesc; p; p = p->pNext)
        count += p->hdrSize + p->dataSize;

      count -= offset;
      if (count <= 0)             /* offset larger than data len */
        {
          for(p = pdesc; p; p = p->pNext)
            {
              p->hdrSize = 0;
              p->dataSize = 0;
            }
        }
      else
        {
          /* find the last valid pkt descr */
          for(p = pdesc; p; p = p->pNext)
            {
              sum += p->hdrSize + p->dataSize;
              if (sum>= count)
                break;
            }

          /* how much need to be removed at tail of this pkt desc */
          offset = sum - count;
          if (offset <=  p->dataSize)
            {
              p->dataSize -=offset;
              p->offset = -offset;
            }
          else
            {
              offset -= p->dataSize;
              p->dataSize = 0;
              p->hdrSize -= offset;
              p->pHdr += offset;
            }

          /* adjust reset of buffer */
          p = p->pNext;
          while (p)
            {
              p->dataSize = 0;
              p->hdrSize = 0;
            }
        }
    }

  return NFX_OK;
}        

#define PKT_MAX_OWNER_ENTRY		10

typedef struct
{
	char	*fileName;
	int32	lineNum;
	int32	pktCnt;
}	pkt_ownerStats_t;

pkt_ownerStats_t	pktOwnerTbl[PKT_MAX_OWNER_ENTRY];

/*-----------------------------------------------------------------
 * Name:        pkt_showInUse
 *
 * Description: Count & show PKTs that are in use & respective owner.
 *
 * Created by : Shaun Hui
 *
 *-----------------------------------------------------------------
 */
int32
pkt_showInUse( void )
{

	pkt_desc_t			*pDesc;
	pkt_ownerStats_t	*pOwner;
	int32				i;


    /*-------------------------------
     * Clear owner table.
     *-------------------------------
     */
	memset(pktOwnerTbl, 0, sizeof(pktOwnerTbl));

    /*-------------------------------
     * Go thru pkt ring.
     *-------------------------------
     */
	pDesc = pPktCfg->descStart;
	while (pDesc <= pPktCfg->descEnd)
	{
		if (pDesc->flags & PKT_DESC_INUSE)
		{
		    /*-------------------------------
		     * Go thru owner table.
			 * Look for empty or matching entry.
		     *-------------------------------
		     */
			pOwner = &pktOwnerTbl[0];
			for (i=0; i <PKT_MAX_OWNER_ENTRY; i++)
			{
				if ((pOwner->fileName == NULL) ||
					((pOwner->fileName == pDesc->allocFile) &&
					 (pOwner->lineNum  == pDesc->allocLine)))
				{
					pOwner->fileName = pDesc->allocFile;
					pOwner->lineNum  = pDesc->allocLine;
					pOwner->pktCnt++;
					break;
				}

				pOwner++;
			}
		}

		pDesc++;
	}
	
    /*-------------------------------
     * Display owner table.
     *-------------------------------
     */
	printf("\n");
	printf("Alloc File        Line #     Pkt Cnt\n");
	printf("------------------------------------\n");
	
	pOwner = &pktOwnerTbl[0];
	for (i=0; i <PKT_MAX_OWNER_ENTRY; i++)
	{
	    /*-------------------------------
	     * Skip empty entry.
	     *-------------------------------
	     */
		if (pOwner->fileName == NULL)
			break;

		printf("%-16s  %6d    %6d\n",
				pOwner->fileName, pOwner->lineNum, pOwner->pktCnt);
		pOwner++;
	}
	printf("\n");

	return 0;
}

#ifdef PKT_USE_TIME_STAMP
#include <exports/nfs-api.h>

#ifdef NOTUSED
typedef struct
{
	uint64		ops;
	uint64		numBytes;
	eeeRTC_t	rspHWM;
	uint32		rspTooLong;
	eeeRTC_t	duration[PKT_MAX_TIME_IDX];
} pkt_procTime_t;
#endif

nfs_procTime_t	nfsProcTime[NFS_COMMIT+1]={{0}};


void *pkt_allocTimeStamp( void )
{
	return (void *)eee_ramAlloc(sizeof (pkt_timeStamp_t));
}

void pkt_freeTimeStamp( void  *pTime )
{
	if (!pTime)
		return;

	eee_ramDealloc( pTime );
}


void pkt_copyTimeStamp( void  *pDstTime, void  *pSrcTime)
{
	memcpy(pDstTime, pSrcTime, sizeof(pkt_timeStamp_t));	
}

void pkt_setTimeStamp( pkt_timeStamp_t *pTime, int32 timeIdx )
{
	if ((timeIdx < NFS_TIME_IDX_LUC_RX) ||
		(timeIdx > NFS_TIME_IDX_LUC_TX_CMP))
		return;

	if (!pTime)
		return;

	eee_rtc( &(((pkt_timeStamp_t *)pTime)->timeVal[timeIdx]) );
}

inline void pkt_getNfsProcTime( char *bufp )
{
	extern uint64	ht_bytes_tx;

    nfsProcTime[0].ht_bytes_tx = ht_bytes_tx;
    ht_bytes_tx = 0 ;
    
	memcpy(bufp, &nfsProcTime[0], sizeof(nfsProcTime) );
    memset(nfsProcTime, 0, sizeof(nfsProcTime));
}

inline void pkt_setTxByteCount( int32 numBytes,	int32 proc )
{
	if ((proc < NFS_NULL) || (proc > NFS_COMMIT))
		return;

	nfsProcTime[proc].txBytes += numBytes;

}


inline void pkt_setRxByteCount( int32 numBytes,	int32 proc )
{
	if ((proc < NFS_NULL) || (proc > NFS_COMMIT))
		return;

	nfsProcTime[proc].rxBytes += numBytes;

}


static char *nfsProcName[] = {
						"NULL",
                        "GETATTR",
                        "SETATTR",
                        "LOOKUP",
                        "ACCESS",
                        "READLINK",
                        "READ",
                        "WRITE",
                        "CREATE",
                        "MKDIR",
                        "SYMLINK",
                        "MKNOD",
                        "REMOVE",
                        "RMDIR",
                        "RENAME",
                        "LINK",
                        "READDIR",
                        "READDIR+",
                        "STATFS",
                        "FSSTAT",
                        "FSINFO",
                        "PATHCONF",
                        "COMMIT" };


int32	nfsProcMaxRsp = 1000000; /* 1000 ms = 1 sec */

void
pkt_setNfsProcTime( pkt_timeStamp_t *pTime, int32 proc )
{
	int32	idx;
	eeeRTC_t diff;
	static boolean first = TRUE;
	static eeeRTC_t nfsProcMaxRspRTC;

	if (!pTime)
		return;

	if (first) {
		first = FALSE;
		eee_msToRTC(nfsProcMaxRsp, &nfsProcMaxRspRTC);
	}
	nfsProcTime[proc].ops++;
	diff = *(eee_rtcZero());

	for (idx = NFS_TIME_IDX_LUC_RX; idx < NFS_TIME_IDX_LUC_TX_CMP; idx++)
	{
		if (eee_rtcCompare(&(pTime->timeVal[idx]), &diff) == 0)
			continue;
		(void)eee_rtcSub(&(pTime->timeVal[idx+1]),
						 &(pTime->timeVal[idx]), &diff);

		if (idx == NFS_TIME_IDX_FS_REQ)
		{
			if (eee_rtcCompare(&diff, &nfsProcMaxRspRTC) > 0)
			{
				nfsProcTime[proc].rspTooLong++;
			}
			if (eee_rtcCompare(&diff, &(nfsProcTime[proc].rspHWM)) > 0)
				nfsProcTime[proc].rspHWM = diff;
		}
		eee_rtcIncr(&(nfsProcTime[proc].duration[idx]), &diff);
		diff = *(eee_rtcZero());
	}
}


void pkt_showNfsProcTime( void )
{
	int32 	proc;
    uint64  ave;
    int32   idx;

    printf("NFS PROC      NFS REQ  LUC-RX  TPL-RX  NFS-RX  FS-RSP  ");
	printf("NFS-TX  TPL-TX  LUC-TX  LONG-RSP  RSP-HWM ");
	printf("      RX BYTES       TX BYTES\n");
	printf("-------------------------------------------------------");
    printf("------------------------------------------");
    printf("-----------------------------\n");

    for (proc =0; proc <= NFS_COMMIT; proc++)
    {
    	printf("%9s:%10llu  ", nfsProcName[proc], nfsProcTime[proc].ops);

		for (idx = NFS_TIME_IDX_LUC_RX; idx < NFS_TIME_IDX_LUC_TX_CMP; idx++)
		{
        	ave = (nfsProcTime[proc].ops) ?
            	(eee_rtcToUsec(&(nfsProcTime[proc].duration[idx])) /
				 nfsProcTime[proc].ops): 0;

			printf(" %6llu ", ave);
		}
		printf(" %6u  ", nfsProcTime[proc].rspTooLong);
		printf(" %6llu ", eee_rtcToUsec(&(nfsProcTime[proc].rspHWM)));
		printf(" %15llu ", nfsProcTime[proc].rxBytes);
		printf(" %15llu ", nfsProcTime[proc].txBytes);
		printf("\n");
    }

    memset(nfsProcTime, 0, sizeof(nfsProcTime));
}

#endif /* PKT_USE_TIME_STAMP */
