

/*
 * Copyright (c) 1997 Marc Horowitz.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Marc Horowitz.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <nfx-defs.h>
#include <nfx-types.h>
#include "i82365reg.h"
#ifdef R9000
#include <exports/hwMap.h>
#include <sys/gt64440r.h>
#else
#include <sys/gt64240r.h>
#endif
#include <exports/libc.h>
#include <exports/pci-api.h>

#include "queue.h"

#include "pcmciavar.h"
#include "pcmciareg.h"

extern void *eee_ramAlloc(int);


extern int disk_debug_param;

#if defined(DEBUG)
#define debug(x) { if (disk_debug_param & DEBUG_PCMCIA) { printf x; }};
#else
#define debug(x) 
#endif

#define assert NFX_ASSERT

struct pcmcia pcmciasc;
struct pcmcia_socket pcmcia_sockets[2];



struct pcmcia *
pcmcia_alloc_sc()
{
	return &pcmciasc;
}


struct pcmcia *
pcmcia_sc()
{
	return &pcmciasc;
}


/*
 * pcmcia_valid_socket - return true if the socket id is valid
 */
static int
pcmcia_valid_socket(int sockid)
{
	return sockid == PCMCIA_SOCKID_A || sockid == PCMCIA_SOCKID_B;
}


struct pcmcia_socket *
pcmcia_socket(struct pcmcia *sc, int sockid)
{
	assert(pcmcia_valid_socket(sockid));
	return sc->sockets[sockid];
}


static struct pcmcia_socket *
pcmcia_alloc_socket(int sockid)
{
	assert(pcmcia_valid_socket(sockid));
	return &pcmcia_sockets[sockid];
}


static struct io_map_index_st {
	int	start_lsb;
	int	start_msb;
	int	stop_lsb;
	int	stop_msb;
	int	ioenable;
	int	ioctlmask;
	int	ioctlbits[3];		/* indexed by PCMCIA_WIDTH_* */
}               io_map_index[] = {
	{
		PCIC_IOADDR0_START_LSB,
		PCIC_IOADDR0_START_MSB,
		PCIC_IOADDR0_STOP_LSB,
		PCIC_IOADDR0_STOP_MSB,
		PCIC_ADDRWIN_ENABLE_IO0,
		PCIC_IOCTL_IO0_WAITSTATE | PCIC_IOCTL_IO0_ZEROWAIT |
		PCIC_IOCTL_IO0_IOCS16SRC_MASK | PCIC_IOCTL_IO0_DATASIZE_MASK,
		{
			PCIC_IOCTL_IO0_IOCS16SRC_CARD,
			PCIC_IOCTL_IO0_IOCS16SRC_DATASIZE |
			    PCIC_IOCTL_IO0_DATASIZE_8BIT,
			PCIC_IOCTL_IO0_IOCS16SRC_DATASIZE |
			    PCIC_IOCTL_IO0_DATASIZE_16BIT,
		},
	},
	{
		PCIC_IOADDR1_START_LSB,
		PCIC_IOADDR1_START_MSB,
		PCIC_IOADDR1_STOP_LSB,
		PCIC_IOADDR1_STOP_MSB,
		PCIC_ADDRWIN_ENABLE_IO1,
		PCIC_IOCTL_IO1_WAITSTATE | PCIC_IOCTL_IO1_ZEROWAIT |
		PCIC_IOCTL_IO1_IOCS16SRC_MASK | PCIC_IOCTL_IO1_DATASIZE_MASK,
		{
			PCIC_IOCTL_IO1_IOCS16SRC_CARD,
			PCIC_IOCTL_IO1_IOCS16SRC_DATASIZE |
			    PCIC_IOCTL_IO1_DATASIZE_8BIT,
			PCIC_IOCTL_IO1_IOCS16SRC_DATASIZE |
			    PCIC_IOCTL_IO1_DATASIZE_16BIT,
		},
	},
};


static struct mem_map_index_st {
	int	sysmem_start_lsb;
	int	sysmem_start_msb;
	int	sysmem_stop_lsb;
	int	sysmem_stop_msb;
	int	cardmem_lsb;
	int	cardmem_msb;
	int	memenable;
} mem_map_index[] = {
	{
		PCIC_SYSMEM_ADDR0_START_LSB,
		PCIC_SYSMEM_ADDR0_START_MSB,
		PCIC_SYSMEM_ADDR0_STOP_LSB,
		PCIC_SYSMEM_ADDR0_STOP_MSB,
		PCIC_CARDMEM_ADDR0_LSB,
		PCIC_CARDMEM_ADDR0_MSB,
		PCIC_ADDRWIN_ENABLE_MEM0,
	},
	{
		PCIC_SYSMEM_ADDR1_START_LSB,
		PCIC_SYSMEM_ADDR1_START_MSB,
		PCIC_SYSMEM_ADDR1_STOP_LSB,
		PCIC_SYSMEM_ADDR1_STOP_MSB,
		PCIC_CARDMEM_ADDR1_LSB,
		PCIC_CARDMEM_ADDR1_MSB,
		PCIC_ADDRWIN_ENABLE_MEM1,
	},
	{
		PCIC_SYSMEM_ADDR2_START_LSB,
		PCIC_SYSMEM_ADDR2_START_MSB,
		PCIC_SYSMEM_ADDR2_STOP_LSB,
		PCIC_SYSMEM_ADDR2_STOP_MSB,
		PCIC_CARDMEM_ADDR2_LSB,
		PCIC_CARDMEM_ADDR2_MSB,
		PCIC_ADDRWIN_ENABLE_MEM2,
	},
	{
		PCIC_SYSMEM_ADDR3_START_LSB,
		PCIC_SYSMEM_ADDR3_START_MSB,
		PCIC_SYSMEM_ADDR3_STOP_LSB,
		PCIC_SYSMEM_ADDR3_STOP_MSB,
		PCIC_CARDMEM_ADDR3_LSB,
		PCIC_CARDMEM_ADDR3_MSB,
		PCIC_ADDRWIN_ENABLE_MEM3,
	},
	{
		PCIC_SYSMEM_ADDR4_START_LSB,
		PCIC_SYSMEM_ADDR4_START_MSB,
		PCIC_SYSMEM_ADDR4_STOP_LSB,
		PCIC_SYSMEM_ADDR4_STOP_MSB,
		PCIC_CARDMEM_ADDR4_LSB,
		PCIC_CARDMEM_ADDR4_MSB,
		PCIC_ADDRWIN_ENABLE_MEM4,
	},
};


int pcmcia_initialized;
int intel_cf_part;
int ti_cf_part;


/*
 * pcmcia_valid_reg - return true if the register is valid
 */
static int
pcmcia_valid_reg(uint8 reg)
{
	return reg <= 0x3f;
}


/*
 * pcmcia_valid_index - return true if the register index is valid
 */
static int
pcmcia_valid_index(uint8 index)
{
	return pcmcia_valid_reg(index & ~PCIC_SOCKETB_INDEX) &&
		(index & 0x80) == 0;
}


/*
 * pcmcia_sokcname - return printable socket name
 */
char *
pcmcia_sockname(struct pcmcia_socket *sc)
{
	assert(pcmcia_valid_socket(sc->id));
	return sc->id == PCMCIA_SOCKID_A ? "SOCKET A" : "SOCKET B";
}


/*
 * pcmcia_write_index_reg - write the register index into index register.
 */
static void
pcmcia_write_index_reg(struct pcmcia *sc, uint8 index)
{
	assert(pcmcia_valid_index(index));
	*(uint8 *)sc->regs = index;
}


/*
 * pcmcia_write_data - write the data to the data register
 */
static void
pcmcia_write_data(struct pcmcia *sc, uint8 data)
{
	*(uint8 *)(sc->regs + PCIC_REG_DATA) = data;
}


/*
 * pcmcia_read_data - read the data from the data register
 */
static uint8
pcmcia_read_data(struct pcmcia *sc)
{
	return *(uint8 *)(sc->regs + PCIC_REG_DATA);
}


/*
 * pcmcia_sockid2idx - returns the index modifier for the socket
 */
static uint8
pcmcia_sockid2idx(int id)
{
	assert(pcmcia_valid_socket(id));
	return (id == PCMCIA_SOCKID_A) ? PCIC_SOCKETA_INDEX :
		PCIC_SOCKETB_INDEX;
}


/*
 * pcic_read - read from the given register
 */
static uint8
pcic_read(struct pcmcia_socket *sc, uint8 reg)
{
	assert(pcmcia_valid_socket(sc->id));
	assert(pcmcia_valid_reg(reg));

	pcmcia_write_index_reg(sc->pcmcia, pcmcia_sockid2idx(sc->id) | reg);
	return pcmcia_read_data(sc->pcmcia);
}


/*
 * pcic_write - write the value in the register
 */
static void
pcic_write(struct pcmcia_socket *sc, uint8 reg, uint8 data)
{
	assert(pcmcia_valid_socket(sc->id));
	assert(pcmcia_valid_reg(reg));

	pcmcia_write_index_reg(sc->pcmcia, pcmcia_sockid2idx(sc->id) | reg);
	pcmcia_write_data(sc->pcmcia, data);
}


/*
 * pcmcia_read_ext_reg - read a value from Cirrus extendsion register
 */
uint8
pcmcia_read_ext_reg(struct pcmcia_socket *sc, uint8 reg)
{
	pcic_write(sc, PCIC_CIRRUS_EXTENDED_INDEX, reg);
	return pcic_read(sc, PCIC_CIRRUS_EXTENDED_DATA);
}


/*
 * pcmcia_write_ext_reg - write the value into Cirrus extension register
 */
void
pcmcia_write_ext_reg(struct pcmcia_socket *sc, uint8 reg, uint8 data)
{
	pcic_write(sc, PCIC_CIRRUS_EXTENDED_INDEX, reg);
	pcic_write(sc, PCIC_CIRRUS_EXTENDED_DATA, reg);
}



/*
 * pcmcia_map_iowindow - setup the mappings for the I/O window
 */
static void
pcmcia_map_iowindow(struct pcmcia_socket *h, int window_id,
					struct pcmcia_window *window)
{
	uint8 reg;

	debug(("pcmcia_map_iowindow: sock %s win %d start %x size %x\n",
		   pcmcia_sockname(h), window_id, window->addr, window->size));

	pcic_write(h, io_map_index[window_id].start_lsb, window->addr & 0xff);
	pcic_write(h, io_map_index[window_id].start_msb,
	    (window->addr >> 8) & 0xff);

	pcic_write(h, io_map_index[window_id].stop_lsb,
	    (window->addr + window->size - 1) & 0xff);
	pcic_write(h, io_map_index[window_id].stop_msb,
	    ((window->addr + window->size - 1) >> 8) & 0xff);

	reg = pcic_read(h, PCIC_IOCTL);
	reg &= ~io_map_index[window_id].ioctlmask;
	reg |= io_map_index[window_id].ioctlbits[window->width];
	pcic_write(h, PCIC_IOCTL, reg);

	reg = pcic_read(h, PCIC_ADDRWIN_ENABLE);
	reg |= io_map_index[window_id].ioenable;
	pcic_write(h, PCIC_ADDRWIN_ENABLE, reg);
}


static void 
pcic_chip_do_mem_map(struct pcmcia_socket *h, int win)
{
	int reg;
	int kind = h->mem[win].kind & ~PCMCIA_WIDTH_MEM_MASK;
	int mem8 = 
	    (h->mem[win].kind & PCMCIA_WIDTH_MEM_MASK) == PCMCIA_WIDTH_MEM8
	    || (kind == PCMCIA_MEM_ATTR);

	debug(("pcic_chip_do_mem_map: addr=%x size=%x\n", h->mem[win].addr,
		   h->mem[win].size));
	
	pcic_write(h, mem_map_index[win].sysmem_start_lsb,
	    (h->mem[win].addr >> PCIC_SYSMEM_ADDRX_SHIFT) & 0xff);
	pcic_write(h, mem_map_index[win].sysmem_start_msb,
	    ((h->mem[win].addr >> (PCIC_SYSMEM_ADDRX_SHIFT + 8)) &
	    PCIC_SYSMEM_ADDRX_START_MSB_ADDR_MASK) |
	    (mem8 ? 0 : PCIC_SYSMEM_ADDRX_START_MSB_DATASIZE_16BIT));

	pcic_write(h, mem_map_index[win].sysmem_stop_lsb,
	    ((h->mem[win].addr + h->mem[win].size) >>
	    PCIC_SYSMEM_ADDRX_SHIFT) & 0xff);
	pcic_write(h, mem_map_index[win].sysmem_stop_msb,
	    (((h->mem[win].addr + h->mem[win].size) >>
	    (PCIC_SYSMEM_ADDRX_SHIFT + 8)) &
	    PCIC_SYSMEM_ADDRX_STOP_MSB_ADDR_MASK) |
	    PCIC_SYSMEM_ADDRX_STOP_MSB_WAIT2);

	pcic_write(h, mem_map_index[win].cardmem_lsb,
	    (h->mem[win].offset >> PCIC_CARDMEM_ADDRX_SHIFT) & 0xff);
	pcic_write(h, mem_map_index[win].cardmem_msb,
	    ((h->mem[win].offset >> (PCIC_CARDMEM_ADDRX_SHIFT + 8)) &
	    PCIC_CARDMEM_ADDRX_MSB_ADDR_MASK) |
	    ((kind == PCMCIA_MEM_ATTR) ?
	    PCIC_CARDMEM_ADDRX_MSB_REGACTIVE_ATTR : 0));

	reg = pcic_read(h, PCIC_ADDRWIN_ENABLE);
	reg |= (mem_map_index[win].memenable | PCIC_ADDRWIN_ENABLE_MEMCS16);
	pcic_write(h, PCIC_ADDRWIN_ENABLE, reg);


	debug(("pcic_chip_do_mem_map window %d: %02x%02x %02x%02x "
		   "%02x%02x\n", win, 
		   pcic_read(h, mem_map_index[win].sysmem_start_msb),
		   pcic_read(h, mem_map_index[win].sysmem_start_lsb),
		   pcic_read(h, mem_map_index[win].sysmem_stop_msb),
		   pcic_read(h, mem_map_index[win].sysmem_stop_lsb),
		   pcic_read(h, mem_map_index[win].cardmem_msb),
		   pcic_read(h, mem_map_index[win].cardmem_lsb)));
}


/*
 * pcmcia_map_io - setup the I/O mappings for the socket
 */
void
pcmcia_map_io(struct pcmcia_socket *sc)
{
	int window_id;

	for (window_id = 0; window_id < sc->niowindows; ++window_id) {
		pcmcia_map_iowindow(sc, window_id, &sc->iomap[window_id]);
	}
}


static uint32
galileo_get_pci0_iobase()
{
#ifdef R9000
	return (PHYS_TO_K1(HW_CF_CTRL_IO_BASE));
#else
	return 0xa0000000 +
		(*(uint32 *)(0xb4000000 + PCI_0I_O_LOW_DECODE_ADDRESS) << 20);
#endif
}



static uint32
galileo_get_pci0_mem2base()
{
#ifdef R9000
	return (PHYS_TO_K1(HW_CF_CTRL_MEM_BASE));
#else 
	return 0xa0000000 + (*(uint32 *)(0xb4000000 +
						  PCI_0MEMORY2_LOW_DECODE_ADDRESS) << 20);
#endif
}


/*
 * pcmcia_pci_find - find the pcmcia bridge. Returns NFX_OK if found
 */
static int
pcmcia_pci_find(uint32 *bus_ptr, uint32 *dev_ptr)
{
    uint32 bus = 0;
    int rc = NFX_ERR;
    uint32 dev;

    debug(("pcmcia_pci_find: "));
    //  New RoHS functions and registers differ 
    //  Flags for discovered parts...
    intel_cf_part = 0;
    ti_cf_part = 0;

    for (bus = 0; rc == NFX_ERR && bus < 16; ++bus) {
        for (dev = 0; rc == NFX_ERR && dev < 32; ++dev) {
            uint32 val;

            pci_cfg_read(bus, dev, 0, 0x0, &val);
            if ( (((val >> 16) & 0xffff) == PCMCIA_PCI_PRODUCT &&
                   (val & 0xffff) == PCMCIA_PCI_VENDOR) ) {
                intel_cf_part = 1;
                ti_cf_part = 0;
            } 
            if ( (((val >> 16) & 0xffff) == PCMCIA_PCI_PRODUCT_2 &&
                   (val & 0xffff) == PCMCIA_PCI_VENDOR_2)) {
                intel_cf_part = 0;
                ti_cf_part = 1;
            }

            if( intel_cf_part || ti_cf_part ) {
                debug(("found pcmcia bridge at bus %x dev %x: ", bus, dev));
                *bus_ptr = bus;
                *dev_ptr = dev;
                rc = NFX_OK;
            }
        }
    }

    debug(("rc = %d\n", rc));

    return rc;
}


		
/*
 * pcmcia_read_baseaddr - find out out I/O address
 */
static void
pcmcia_read_baseaddr(struct pcmcia *sc)
{
	uint32 bar;
	uint32 iobase;
	debug(("pcmcia_read_baseaddr:bus=%x dev=%x:", sc->bus, sc->dev));
	iobase = galileo_get_pci0_iobase();
	debug(("iobase=%x:", iobase));
        if( intel_cf_part ) { 
	    pci_cfg_read(sc->bus, sc->dev, 0, 0x10, &bar);
        }
        if( ti_cf_part ) {
	    pci_cfg_read(sc->bus, sc->dev, 0, 0x44, &bar);
        }
	debug(("bar=%x:", bar));
	sc->regs = (bar & ~0x1) + iobase;
	debug(("regs=%x\n", sc->regs));
}


/*
 * pcmcia_power_down - power down the socket
 */
static void
pcmcia_power_down(struct pcmcia_socket *sc)
{
	pcic_write(sc, PCIC_PWRCTL, 0);
}


/*
 * pcmcia_power_up - turn the power on for the socket
 */
static void
pcmcia_power_up(struct pcmcia_socket *sc)
{
	pcic_write(sc, PCIC_PWRCTL, PCIC_PWRCTL_DISABLE_RESETDRV |
			   PCIC_PWRCTL_PWR_ENABLE);
}


/*
 * pcmcia_output_enable - enable output to the socket
 */
static void
pcmcia_output_enable(struct pcmcia_socket *sc)
{
	pcic_write(sc, PCIC_PWRCTL, pcic_read(sc, PCIC_PWRCTL) |
			   PCIC_PWRCTL_OE);
}


/*
 * pcmcia_reset - reset the card
 */
static void
pcmcia_reset(struct pcmcia_socket *sc)
{
	pcic_write(sc, PCIC_INTR, 0);
}


/*
 * pcmcia_clear_reset - turn reset off
 */
static void
pcmcia_clear_reset(struct pcmcia_socket *sc)
{
	pcic_write(sc, PCIC_INTR, pcic_read(sc, PCIC_INTR) | PCIC_INTR_RESET |
		PCIC_INTR_CARDTYPE_IO);
}


/*
 * pcmcia_delay_ms - wait some milliseconds
 */
static void
pcmcia_delay_ms(int ms)
{
	delay_us(ms * 1000);
}

/*
 * pcmcia_socket_carddetech - return true if there is a card in the socket
 */
static int
pcmcia_socket_carddetect(struct pcmcia_socket *sc)
{
	return ((pcic_read(sc, PCIC_IF_STATUS)
			 & PCIC_IF_STATUS_CARDDETECT_MASK)
			 == PCIC_IF_STATUS_CARDDETECT_PRESENT);
}

/*
 * pcmcia_wait_ready - wait for the card to become ready
 */
static int
pcmcia_wait_ready(struct pcmcia_socket *sc)
{
	int i;
	int ready = 0;

	debug(("pcmcia_wait_read: status = %x\n",
		   pcic_read(sc, PCIC_IF_STATUS)));
        // Check to see if the card is installed in the socket (prevent wd time-out)
        if( pcmcia_socket_carddetect(sc)) {
	    for (i = 0; i < 10000 && !ready; i++) {
		ready = pcic_read(sc, PCIC_IF_STATUS) & PCIC_IF_STATUS_READY;
		pcmcia_delay_ms(1);
	    }

	    if (!ready) {
		printf("pcmcia_wait_ready: ready never happened, status = %02x\n",
			   pcic_read(sc, PCIC_IF_STATUS));
	    }
        }

	return ready;
}


/*
 * pcmcia_enable_socket - enable the socket. 
 */
void
pcmcia_enable_socket(struct pcmcia_socket *sc)
{
	debug(("pcmcia_enable_socket: %s\n", pcmcia_sockname(sc)));

	pcmcia_power_down(sc);
	pcmcia_delay_ms(300 + 100);

	debug(("pcmcia_enable_socket: power down done\n"));

	pcmcia_power_up(sc);
	pcmcia_delay_ms(100 + 20 + 200);

	debug(("pcmcia_enable_socket: power up done\n"));

	pcmcia_output_enable(sc);
	pcmcia_reset(sc);
	pcmcia_delay_ms(1);

	debug(("pcmcia_enable_socket: output enable, reset on\n"));

	pcmcia_clear_reset(sc);
	pcmcia_delay_ms(20);

	debug(("pcmcia_enable_socket; reset off\n"));

	if (pcmcia_wait_ready(sc)) {
		sc->state = pcmcia_socket_carddetect(sc) ?
			PCMCIA_SOCK_PRESENT : PCMCIA_SOCK_NOTPRESENT;
	}
	else {
		sc->state = PCMCIA_SOCK_ENABLED;
	}

	debug(("pcmcia_enable_socket: done\n"));
}


/*
 * pcmcia_device_present - return true if there is a card in the socket
 */
int
pcmcia_device_present(struct pcmcia_socket *sc)
{
	return sc->state == PCMCIA_SOCK_PRESENT;
}


/*
 * pcmcia_init_mem_window - initialize the memory window
 */
static void
pcmcia_init_mem_window(struct pcmcia_window *win, uint32 addr,
					   uint32 size, int kind)
{
	win->addr = addr;
	win->size = size;
	win->kind = kind;
}


/*
 * pcmcia_init_mem_map - initialize the memory map
 */
static void
pcmcia_init_mem_map(struct pcmcia_socket *sc)
{
	assert(pcmcia_valid_socket(sc->id));
	if (sc->id == PCMCIA_SOCKID_A) {
		sc->nmemwindows = 1;
		pcmcia_init_mem_window(&sc->mem[0], galileo_get_pci0_mem2base(),
							   4096, PCMCIA_MEM_ATTR);
		debug(("pcmcia_init_mem_map (%s): addr=%x\n",
			   pcmcia_sockname(sc), sc->mem[0].addr));
	}
	else if (sc->id == PCMCIA_SOCKID_B) {
		sc->nmemwindows = 1;
	 	pcmcia_init_mem_window(&sc->mem[0],
							   galileo_get_pci0_mem2base() + 4096,
 							   4096, PCMCIA_MEM_ATTR);
		debug(("pcmcia_init_mem_map (%s): addr=%x\n",
			   pcmcia_sockname(sc), sc->mem[0].addr));
	}
	pcic_write(sc, PCIC_ADDRWIN_ENABLE, 0);
}


static void
pcic_chip_mem_unmap(struct pcmcia_socket *h, int win)
{
	int reg;
	h->memalloc &= ~(1 << win);
	reg = pcic_read(h, PCIC_ADDRWIN_ENABLE);
	reg &= ~mem_map_index[win].memenable;
	pcic_write(h, PCIC_ADDRWIN_ENABLE, reg);
}


static int 
pcic_chip_mem_map(struct pcmcia_socket *h,
				  int kind, uint32 card_addr, uint32 size,
				  uint32 busaddr, int *offsetp, int *windowp)
{
	long card_offset;
	int i, win;

	win = -1;
	for (i = 0; i < (sizeof(mem_map_index) / sizeof(mem_map_index[0]));
	    i++) {
		if ((h->memalloc & (1 << i)) == 0) {
			win = i;
			h->memalloc |= (1 << i);
			break;
		}
	}

	if (win == -1)
		return (1);

	*windowp = win;

	/*
	 * Compute the address offset to the pcmcia address space for the
	 * pcic.  This is intentionally signed.  The masks and shifts below
	 * will cause TRT to happen in the pcic registers.  Deal with making
	 * sure the address is aligned, and return the alignment offset.
	 */

	*offsetp = card_addr % PCIC_MEM_ALIGN;
	card_addr -= *offsetp;

	debug(("pcic_chip_mem_map window %d bus %x+%x+%x at card addr "
		   "%x\n", win, (uint32) busaddr, (uint32) * offsetp, (uint32) size,
		   (uint32) card_addr));

	/*
	 * include the offset in the size, and decrement size by one, since
	 * the hw wants start/stop
	 */
	size += *offsetp - 1;

	card_offset = (((long) card_addr) - ((long) busaddr));

	h->mem[win].addr = busaddr;
	h->mem[win].size = size;
	h->mem[win].offset = card_offset;
	h->mem[win].kind = kind;

	pcic_chip_do_mem_map(h, win);

	return (0);
}


static char *
pcmcia_device_name(int dtype)
{
	switch (dtype) {
	case PCMCIA_DTYPE_NULL:
		return "null";
		break;
	case PCMCIA_DTYPE_ROM:
		return "rom";
		break;
	case PCMCIA_DTYPE_OTPROM:
		return "otprom";
		break;
	case PCMCIA_DTYPE_EPROM:
		return "eprom";
		break;
	case PCMCIA_DTYPE_EEPROM:
		return "eeprom";
		break;
	case PCMCIA_DTYPE_FLASH:
		return "flash";
		break;
	case PCMCIA_DTYPE_SRAM:
		return "sram";
		break;
	case PCMCIA_DTYPE_DRAM:
		return "dram";
		break;
	case PCMCIA_DTYPE_FUNCSPEC:
		return "funcspec";
		break;
	case PCMCIA_DTYPE_EXTEND:
		return "extend";
		break;
	default:
		return "reserved";
		break;
	}
}


static char *
pcmcia_device_speed(int speed)
{
	switch (speed) {
	case PCMCIA_DSPEED_NULL: 
		return "null";
		break;
	case PCMCIA_DSPEED_250NS:
		return "250ns";
		break;
	case PCMCIA_DSPEED_200NS:
		return "200ns";
		break;
	case PCMCIA_DSPEED_150NS:
		return "150ns";
		break;
	case PCMCIA_DSPEED_100NS:
		return "100ns";
		break;
	case PCMCIA_DSPEED_EXT:
		return "ext";
		break;
	default:
		return "reserved";
		break;
	}
}


static void
print_device_type(struct pcmcia_tuple *tuple)
{
	uint8 reg = pcmcia_tuple_read_1(tuple, 0);
	int dtype = reg & PCMCIA_DTYPE_MASK;
	int dspeed = reg & PCMCIA_DSPEED_MASK;
	debug(("dev type = %s speed = %s\n", pcmcia_device_name(dtype),
		   pcmcia_device_speed(dspeed)));
}


static void
pcmcia_print_info(struct pcmcia_socket *sc)
{
	int i;
	
	debug(("%s: CIS version ", pcmcia_sockname(sc)));
	if (sc->cis1_major == 4) {
		if (sc->cis1_minor == 0) {
			debug(("PCMCIA 1.0\n"));
		}
		else if (sc->cis1_minor == 1) {
			debug(("PCMCIA 2.0 or 2.1\n"));
		}
	} else if (sc->cis1_major >= 5) {
		debug(("PC Card Standard %d.%d\n", sc->cis1_major,
		    sc->cis1_minor));
	}
	else {
		debug(("unknown (major=%d, minor=%d)\n",
			   sc->cis1_major, sc->cis1_minor));
	}
	printf("%s: CIS info: ", pcmcia_sockname(sc));
	for (i = 0; i < 4; i++) {
		if (sc->cis1_info[i] == NULL)
			break;
		if (i)
			printf(", ");
		printf("%s", sc->cis1_info[i]);
	}
	printf("\n");
}


static void
parse_version_tuple(struct pcmcia_socket *sc, struct pcmcia_tuple *tuple)
{
	int start, count, ch, i;
	
	sc->cis1_major = pcmcia_tuple_read_1(tuple, 0);
	sc->cis1_minor = pcmcia_tuple_read_1(tuple, 1);
	for (count = 0, start = 0, i = 0;
		 (count < 4) && ((i + 4) < 256); i++) {
		ch = pcmcia_tuple_read_1(tuple, 2 + i);
		if (ch == 0xff)
			break;
		sc->cis1_info_buf[i] = ch;
		if (ch == 0) {
			sc->cis1_info[count] = sc->cis1_info_buf + start;
			start = i + 1;
			count++;
		}
	}
	pcmcia_print_info(sc);
}


static uint32
pcmcia_mem_alloc(uint32 size)
{
#if 0
	static uint32 last_alloced;
	uint32 rv;

	if (last_alloced == 0)
		last_alloced = galileo_get_pci0_mem2base();
	rv = last_alloced;
	last_alloced += (size + PCIC_MEM_PAGESIZE - 1) &
		~(PCIC_MEM_PAGESIZE - 1);
	return rv;
#else
	return galileo_get_pci0_mem2base();
#endif
}


static void
pcmcia_print_manfid(struct pcmcia_socket *sc)
{
	debug(("%s: manufacturer code 0x%x, product 0x%x\n",
		   pcmcia_sockname(sc), sc->manufacturer, sc->product));
}


static void
parse_manfid_tuple(struct pcmcia_socket *sc, struct pcmcia_tuple *tuple)
{
	if (tuple->length < 4) {
		debug(("CISTPL_MANFID too short %d\n", tuple->length));
	}
	else {
		sc->manufacturer = pcmcia_tuple_read_2(tuple, 0);
		sc->product = pcmcia_tuple_read_2(tuple, 2);
		pcmcia_print_manfid(sc);
	}
}


static struct pcmcia_function *
last_pf(struct pcmcia_socket *sc)
{
	struct pcmcia_function *pf = SIMPLEQ_FIRST(&sc->pf_head);
	while (pf && SIMPLEQ_NEXT(pf, pf_list))
		pf = SIMPLEQ_NEXT(pf, pf_list);
	return pf;
}


/* most of these are educated guesses */
static struct pcmcia_config_entry init_cfe = {
	-1, PCMCIA_CFE_RDYBSY_ACTIVE | PCMCIA_CFE_WP_ACTIVE |
	PCMCIA_CFE_BVD_ACTIVE, PCMCIA_IFTYPE_MEMORY,
};


static void
parse_config_tuple(struct pcmcia_socket *sc, struct pcmcia_tuple *tuple)
{
	if (tuple->length >= 5) {
		uint32 reg, rasz, rmsz, rfsz;
		int i;

		reg = pcmcia_tuple_read_1(tuple, 0);
		rasz = 1 + ((reg & PCMCIA_TPCC_RASZ_MASK) >>
					PCMCIA_TPCC_RASZ_SHIFT);
		rmsz = 1 + ((reg & PCMCIA_TPCC_RMSZ_MASK) >>
					PCMCIA_TPCC_RMSZ_SHIFT);
		rfsz = ((reg & PCMCIA_TPCC_RFSZ_MASK) >>
				PCMCIA_TPCC_RFSZ_SHIFT);

		if (tuple->length < (rasz + rmsz + rfsz)) {
			debug(("CISTPL_CONFIG (%d,%d,%d) too "
				   "short %d\n", rasz, rmsz, rfsz,
				   tuple->length));
		}
		else {
			struct pcmcia_function *pf = last_pf(sc);

			assert(pf != 0);
		
			pf->last_config_index =
				pcmcia_tuple_read_1(tuple, 1);

			pf->ccr_base = 0;
			for (i = 0; i < rasz; i++)
				pf->ccr_base |=
					((pcmcia_tuple_read_1(tuple, 2 + i)) <<
					 (i * 8));

			pf->ccr_mask = 0;
			for (i = 0; i < rmsz; i++)
				pf->ccr_mask |=
					((pcmcia_tuple_read_1(tuple,
										  2 + rasz + i)) << (i * 8));

			/* skip the reserved area and subtuples */

			/* reset the default cfe for each cfe list */
			pf->temp_cfe = init_cfe;
			pf->default_cfe = &pf->temp_cfe;

			debug(("ccr_base = %x ccr_mask = %x\n", pf->ccr_base,
				   pf->ccr_mask));
		}
	}
	else {
		debug(("CISTPL_CONFIG too short %d\n", tuple->length));
	}
}


static void
parse_cftable_tuple(struct pcmcia_socket *sc, struct pcmcia_tuple *tuple)
{
	if (tuple->length >=2) {
		int idx, i, j;
		uint32 reg, reg2;
		uint32 intface, def, num;
		uint32 power, timing, iospace, irq, memspace, misc;
		struct pcmcia_function *pf;
		struct pcmcia_config_entry *cfe;

		idx = 0;

		reg = pcmcia_tuple_read_1(tuple, idx);
		idx++;
		intface = reg & PCMCIA_TPCE_INDX_INTFACE;
		def = reg & PCMCIA_TPCE_INDX_DEFAULT;
		num = reg & PCMCIA_TPCE_INDX_NUM_MASK;

		pf = last_pf(sc);

		assert(pf != NULL);

			/*
			 * this is a little messy.  Some cards have only a
			 * cfentry with the default bit set.  So, as we go
			 * through the list, we add new indexes to the queue,
			 * and keep a pointer to the last one with the
			 * default bit set.  if we see a record with the same
			 * index, as the default, we stash the default and
			 * replace the queue entry. otherwise, we just add
			 * new entries to the queue, pointing the default ptr
			 * at them if the default bit is set.  if we get to
			 * the end with the default pointer pointing at a
			 * record which hasn't had a matching index, that's
			 * ok; it just becomes a cfentry like any other.
			 */

			/*
			 * if the index in the cis differs from the default
			 * cis, create new entry in the queue and start it
			 * with the current default
			 */
		if (num != pf->default_cfe->number) {
			cfe = (struct pcmcia_config_entry *) eee_ramAlloc(sizeof(*cfe));

			*cfe = *pf->default_cfe;

			SIMPLEQ_INSERT_TAIL(&pf->cfe_head, cfe, cfe_list);

			cfe->number = num;

			/*
			 * if the default bit is set in the cis, then
			 * point the new default at whatever is being
			 * filled in
			 */
			if (def)
				pf->default_cfe = cfe;
		} else {
			/*
				 * the cis index matches the default index,
				 * fill in the default cfentry.  It is
				 * assumed that the cfdefault index is in the
				 * queue.  For it to be otherwise, the cis
				 * index would have to be -1 (initial
				 * condition) which is not possible, or there
				 * would have to be a preceding cis entry
				 * which had the same cis index and had the
				 * default bit unset. Neither condition
				 * should happen.  If it does, this cfentry
				 * is lost (written into temp space), which
				 * is an acceptable failure mode.
				 */

			cfe = pf->default_cfe;

			/*
				 * if the cis entry does not have the default
				 * bit set, copy the default out of the way
				 * first.
				 */
			if (!def) {
				pf->temp_cfe = *pf->default_cfe;
				pf->default_cfe = &pf->temp_cfe;
			}
		}

		if (intface) {
			reg = pcmcia_tuple_read_1(tuple, idx);
			idx++;
			cfe->flags &= ~(PCMCIA_CFE_MWAIT_REQUIRED
							| PCMCIA_CFE_RDYBSY_ACTIVE
							| PCMCIA_CFE_WP_ACTIVE
							| PCMCIA_CFE_BVD_ACTIVE);
			if (reg & PCMCIA_TPCE_IF_MWAIT)
				cfe->flags |= PCMCIA_CFE_MWAIT_REQUIRED;
			if (reg & PCMCIA_TPCE_IF_RDYBSY)
				cfe->flags |= PCMCIA_CFE_RDYBSY_ACTIVE;
			if (reg & PCMCIA_TPCE_IF_WP)
				cfe->flags |= PCMCIA_CFE_WP_ACTIVE;
			if (reg & PCMCIA_TPCE_IF_BVD)
				cfe->flags |= PCMCIA_CFE_BVD_ACTIVE;
			cfe->iftype = reg & PCMCIA_TPCE_IF_IFTYPE;
		}
		reg = pcmcia_tuple_read_1(tuple, idx);
		idx++;

		power = reg & PCMCIA_TPCE_FS_POWER_MASK;
		timing = reg & PCMCIA_TPCE_FS_TIMING;
		iospace = reg & PCMCIA_TPCE_FS_IOSPACE;
		irq = reg & PCMCIA_TPCE_FS_IRQ;
		memspace = reg & PCMCIA_TPCE_FS_MEMSPACE_MASK;
		misc = reg & PCMCIA_TPCE_FS_MISC;

		if (power) {
			/* skip over power, don't save */
			/* for each parameter selection byte */
			for (i = 0; i < power; i++) {
				reg = pcmcia_tuple_read_1(tuple, idx);
				idx++;
				/* for each bit */
				for (j = 0; j < 7; j++) {
					/* if the bit is set */
					if ((reg >> j) & 0x01) {
						/* skip over bytes */
						do {
							reg2 = pcmcia_tuple_read_1(tuple, idx);
							idx++;
								/*
								 * until
								 * non-extensi
								 * on byte
								 */
						} while (reg2 & 0x80);
					}
				}
			}
		}
		if (timing) {
			/* skip over timing, don't save */
			reg = pcmcia_tuple_read_1(tuple, idx);
			idx++;

			if ((reg & PCMCIA_TPCE_TD_RESERVED_MASK) !=
				PCMCIA_TPCE_TD_RESERVED_MASK)
				idx++;
			if ((reg & PCMCIA_TPCE_TD_RDYBSY_MASK) !=
				PCMCIA_TPCE_TD_RDYBSY_MASK)
				idx++;
			if ((reg & PCMCIA_TPCE_TD_WAIT_MASK) !=
				PCMCIA_TPCE_TD_WAIT_MASK)
				idx++;
		}
		if (iospace) {
			if (tuple->length <= idx) {
				debug(("ran out of space before TPCE_IO\n"));

				return ;
			}

			reg = pcmcia_tuple_read_1(tuple, idx);
			idx++;

			cfe->flags &=
				~(PCMCIA_CFE_IO8 | PCMCIA_CFE_IO16);
			if (reg & PCMCIA_TPCE_IO_BUSWIDTH_8BIT)
				cfe->flags |= PCMCIA_CFE_IO8;
			if (reg & PCMCIA_TPCE_IO_BUSWIDTH_16BIT)
				cfe->flags |= PCMCIA_CFE_IO16;
			cfe->iomask =
				reg & PCMCIA_TPCE_IO_IOADDRLINES_MASK;

			if (reg & PCMCIA_TPCE_IO_HASRANGE) {
				reg = pcmcia_tuple_read_1(tuple, idx);
				idx++;

				cfe->num_iospace = 1 + (reg &
										PCMCIA_TPCE_IO_RANGE_COUNT);

				if (cfe->num_iospace >
					(sizeof(cfe->iospace) /
					 sizeof(cfe->iospace[0]))) {
					debug(("too many io "
							 "spaces %d",
							 cfe->num_iospace));
					
				}
				for (i = 0; i < cfe->num_iospace; i++) {
					switch (reg & PCMCIA_TPCE_IO_RANGE_ADDRSIZE_MASK) {
					case PCMCIA_TPCE_IO_RANGE_ADDRSIZE_ONE:
						cfe->iospace[i].start =
							pcmcia_tuple_read_1(tuple, idx);
						idx++;
						break;
					case PCMCIA_TPCE_IO_RANGE_ADDRSIZE_TWO:
						cfe->iospace[i].start =
							pcmcia_tuple_read_2(tuple, idx);
						idx += 2;
						break;
					case PCMCIA_TPCE_IO_RANGE_ADDRSIZE_FOUR:
						cfe->iospace[i].start =
							pcmcia_tuple_read_4(tuple, idx);
						idx += 4;
						break;
					}
					switch (reg &
							PCMCIA_TPCE_IO_RANGE_LENGTHSIZE_MASK) {
					case PCMCIA_TPCE_IO_RANGE_LENGTHSIZE_ONE:
						cfe->iospace[i].length =
							pcmcia_tuple_read_1(tuple, idx);
						idx++;
						break;
					case PCMCIA_TPCE_IO_RANGE_LENGTHSIZE_TWO:
						cfe->iospace[i].length =
							pcmcia_tuple_read_2(tuple, idx);
						idx += 2;
						break;
					case PCMCIA_TPCE_IO_RANGE_LENGTHSIZE_FOUR:
						cfe->iospace[i].length =
							pcmcia_tuple_read_4(tuple, idx);
						idx += 4;
						break;
					}
					cfe->iospace[i].length++;
				}
			} else {
				cfe->num_iospace = 1;
				cfe->iospace[0].start = 0;
				cfe->iospace[0].length =
					(1 << cfe->iomask);
			}
		}

		if (irq) {
			if (tuple->length <= idx) {
				debug(("ran out of space before TPCE_IR\n"));

				return;
			}

			reg = pcmcia_tuple_read_1(tuple, idx);
			idx++;

			cfe->flags &= ~(PCMCIA_CFE_IRQSHARE
							| PCMCIA_CFE_IRQPULSE
							| PCMCIA_CFE_IRQLEVEL);
			if (reg & PCMCIA_TPCE_IR_SHARE)
				cfe->flags |= PCMCIA_CFE_IRQSHARE;
			if (reg & PCMCIA_TPCE_IR_PULSE)
				cfe->flags |= PCMCIA_CFE_IRQPULSE;
			if (reg & PCMCIA_TPCE_IR_LEVEL)
				cfe->flags |= PCMCIA_CFE_IRQLEVEL;

			if (reg & PCMCIA_TPCE_IR_HASMASK) {
				/*
				 * it's legal to ignore the
				 * special-interrupt bits, so I will
				 */

				cfe->irqmask =
					pcmcia_tuple_read_2(tuple, idx);
				idx += 2;
			} else {
				cfe->irqmask =
					(1 << (reg & PCMCIA_TPCE_IR_IRQ));
			}
		}
		if (memspace) {
			if (tuple->length <= idx) {
				debug(("ran out of space before TPCE_MS\n"));
				return;
			}

			if (memspace == PCMCIA_TPCE_FS_MEMSPACE_NONE) {
				cfe->num_memspace = 0;
			} else if (memspace == PCMCIA_TPCE_FS_MEMSPACE_LENGTH) {
				cfe->num_memspace = 1;
				cfe->memspace[0].length = 256 *
					pcmcia_tuple_read_2(tuple, idx);
				idx += 2;
				cfe->memspace[0].cardaddr = 0;
				cfe->memspace[0].hostaddr = 0;
			} else if (memspace ==
					   PCMCIA_TPCE_FS_MEMSPACE_LENGTHADDR) {
				cfe->num_memspace = 1;
				cfe->memspace[0].length = 256 *
					pcmcia_tuple_read_2(tuple, idx);
				idx += 2;
				cfe->memspace[0].cardaddr = 256 *
					pcmcia_tuple_read_2(tuple, idx);
				idx += 2;
				cfe->memspace[0].hostaddr = cfe->memspace[0].cardaddr;
			} else {
				int lengthsize;
				int cardaddrsize;
				int hostaddrsize;

				reg = pcmcia_tuple_read_1(tuple, idx);
				idx++;

				cfe->num_memspace = (reg &
									 PCMCIA_TPCE_MS_COUNT) + 1;

				if (cfe->num_memspace >
					(sizeof(cfe->memspace) /
					 sizeof(cfe->memspace[0]))) {
					debug(("too many mem "
							 "spaces %d",
							 cfe->num_memspace));
				}
				lengthsize =
					((reg & PCMCIA_TPCE_MS_LENGTH_SIZE_MASK) >>
					 PCMCIA_TPCE_MS_LENGTH_SIZE_SHIFT);
				cardaddrsize =
					((reg & PCMCIA_TPCE_MS_CARDADDR_SIZE_MASK) >>
					 PCMCIA_TPCE_MS_CARDADDR_SIZE_SHIFT);
				hostaddrsize =
					(reg & PCMCIA_TPCE_MS_HOSTADDR) ? cardaddrsize : 0;

				if (lengthsize == 0) {
					debug(("cfe memspace "
							 "lengthsize == 0"));
					return;
				}
				for (i = 0; i < cfe->num_memspace; i++) {
					if (lengthsize) {
						cfe->memspace[i].length =
							256 * pcmcia_tuple_read_n(tuple, lengthsize,
													  idx);
						idx += lengthsize;
					} else {
						cfe->memspace[i].length = 0;
					}
					if (cfe->memspace[i].length == 0) {
						debug(("cfe->memspace[%d].length == 0",
								 i));
					}
					if (cardaddrsize) {
						cfe->memspace[i].cardaddr =
							256 * pcmcia_tuple_read_n(tuple, cardaddrsize,
													  idx);
						idx += cardaddrsize;
					} else {
						cfe->memspace[i].cardaddr = 0;
					}
					if (hostaddrsize) {
						cfe->memspace[i].hostaddr =
							256 * pcmcia_tuple_read_n(tuple, hostaddrsize,
													  idx);
						idx += hostaddrsize;
					} else {
						cfe->memspace[i].hostaddr = 0;
					}
				}
			}
		}
		if (misc) {
			if (tuple->length <= idx) {
				debug(("ran out of space before TPCE_MI\n"));
				return;
			}

			reg = pcmcia_tuple_read_1(tuple, idx);
			idx++;

			cfe->flags &= ~(PCMCIA_CFE_POWERDOWN
							| PCMCIA_CFE_READONLY
							| PCMCIA_CFE_AUDIO);
			if (reg & PCMCIA_TPCE_MI_PWRDOWN)
				cfe->flags |= PCMCIA_CFE_POWERDOWN;
			if (reg & PCMCIA_TPCE_MI_READONLY)
				cfe->flags |= PCMCIA_CFE_READONLY;
			if (reg & PCMCIA_TPCE_MI_AUDIO)
				cfe->flags |= PCMCIA_CFE_AUDIO;
			cfe->maxtwins = reg & PCMCIA_TPCE_MI_MAXTWINS;

			while (reg & PCMCIA_TPCE_MI_EXT) {
				reg = pcmcia_tuple_read_1(tuple, idx);
				idx++;
			}
		}
		/* skip all the subtuples */
	}
	else {
		debug(("CISTPL_CFTABLE_ENTRY too short %d\n", tuple->length));
	}
}


static struct pcmcia_function *
pcmcia_alloc_function(struct pcmcia_socket *sc)
{
	struct pcmcia_function *pf = eee_ramAlloc(sizeof(struct pcmcia_function));
	assert(pf != NULL);
	memset(pf, 0, sizeof(struct pcmcia_function));
	pf->socket = sc;
	SIMPLEQ_INIT(&pf->cfe_head);

	return pf;
}


static char *
pcmcia_function_name(int function)
{
	switch (function) {
	case PCMCIA_FUNCTION_UNSPEC:
		return("unspecified");
		break;
	case PCMCIA_FUNCTION_MULTIFUNCTION:
		return("multi-function");
		break;
	case PCMCIA_FUNCTION_MEMORY:
		return("memory");
		break;
	case PCMCIA_FUNCTION_SERIAL:
		return("serial port");
		break;
	case PCMCIA_FUNCTION_PARALLEL:
		return("parallel port");
		break;
	case PCMCIA_FUNCTION_DISK:
		return("fixed disk");
		break;
	case PCMCIA_FUNCTION_VIDEO:
		return("video adapter");
		break;
	case PCMCIA_FUNCTION_NETWORK:
		return("network adapter");
		break;
	case PCMCIA_FUNCTION_AIMS:
		return("auto incrementing mass storage");
		break;
	case PCMCIA_FUNCTION_SCSI:
		return("SCSI bridge");
		break;
	case PCMCIA_FUNCTION_SECURITY:
		return("Security services");
		break;
	case PCMCIA_FUNCTION_INSTRUMENT:
		return("Instrument");
		break;
	case PCMCIA_FUNCTION_IOBUS:
		return("Serial I/O Bus Adapter");
		break;
	default:
		return("unknown");
		break;
	}
}


static void
parse_funcid_tuple(struct pcmcia_socket *sc, struct pcmcia_tuple *tuple)
{
	if (tuple->length >= 2) {
		struct pcmcia_function *pf = pcmcia_alloc_function(sc);
		SIMPLEQ_INSERT_TAIL(&sc->pf_head, pf, pf_list);
		pf->function = pcmcia_tuple_read_1(tuple, 0);
		pf->number = sc->number_of_functions;
		++sc->number_of_functions;
		debug(("pf->function %d %s\n", pf->number,
			   pcmcia_function_name(pf->function)));
	}
	else {
		debug(("CISTPL_FUNCID too short %d\n", tuple->length));
	}
}


static void
parse_tuple(struct pcmcia_socket *sc, struct pcmcia_tuple *tuple)
{
	switch (tuple->code) {
	case PCMCIA_CISTPL_END:
		debug(("CISTPL_END\n"));
		break;
	case PCMCIA_CISTPL_VERS_1:
		debug(("CISTPL_VERS_1\n"));
		parse_version_tuple(sc, tuple);
		break;
	case PCMCIA_CISTPL_LONGLINK_MFC:
		debug(("CISTPL_LONGLINK_MFC\n"));
		break;
	case PCMCIA_CISTPL_MANFID:
		debug(("CISTPL_MANFID\n"));
		parse_manfid_tuple(sc, tuple);
		break;
	case PCMCIA_CISTPL_DEVICE:
	case PCMCIA_CISTPL_DEVICE_A:
		debug(("CISTPL_DEVICE\n"));
		print_device_type(tuple);
		break;
	case PCMCIA_CISTPL_CONFIG:
		debug(("CISTPL_CONFIG\n"));
		parse_config_tuple(sc, tuple);
		break;
	case PCMCIA_CISTPL_CFTABLE_ENTRY:
		debug(("CISTPL_CFTABLE_ENTRY\n"));
		parse_cftable_tuple(sc, tuple);
		break;
	case PCMCIA_CISTPL_FUNCID:
		debug(("CISTPL_FUNCID\n"));
		parse_funcid_tuple(sc, tuple);
		break;
	default:
		debug(("parse_tuple: %x not recognized\n", tuple->code));
		break;
	}
}


static int
cis_checksum_ok(struct pcmcia_tuple *tuple)
{
	return 0;
}


static int
tuple_valid_linktarget(struct pcmcia_tuple *tuple)
{
	return 0;
}

/*
 * pcmcia_read_cis - read card information structure
 */
static void
pcmcia_read_cis(struct pcmcia_socket *socksc)
{
	struct pcmcia_tuple tuple;
	uint32 longlink_addr = 0;
	uint32 longlink_present = 0;
	void *next_state = 0;
	int win = 0, offset = 0;
	
	debug(("pcmcia_read_cis: \n"));

	tuple.memh = pcmcia_mem_alloc(PCMCIA_CIS_SIZE);
	tuple.ptr = 0;
	tuple.mult = 2;
	tuple.length = 0;


	pcic_chip_mem_map(socksc, PCMCIA_MEM_ATTR, 0, PCMCIA_CIS_SIZE,
					  tuple.memh, &offset, &win);

 chain_start_label:
	debug(("Chain start:"));
	tuple.code = pcmcia_cis_read_1(&tuple, tuple.ptr);
	debug(("code=%02x:", tuple.code));
	tuple.length = pcmcia_cis_read_1(&tuple, tuple.ptr + 1);
	debug(("length=%02x:", tuple.length));
	switch (tuple.code) {
	case PCMCIA_CISTPL_END: next_state = &&end_tuple_label; break;
	case PCMCIA_CISTPL_NULL: next_state = &&null_tuple_label; break;
	case PCMCIA_CISTPL_LONGLINK_A: next_state = &&longlink_a_label; break;
	case PCMCIA_CISTPL_LONGLINK_C: next_state = &&longlink_c_label; break;
	case PCMCIA_CISTPL_NO_LINK: next_state = &&nolink_label; break;
	case PCMCIA_CISTPL_CHECKSUM: next_state = &&checksum_label; break;
	case PCMCIA_CISTPL_LONGLINK_MFC: next_state = &&longlink_mfc_label; break;
	default: next_state = &&parse_tuple_label; break;
		
	}
	goto *next_state;
 null_tuple_label:
	debug(("NULL tuple\n"));
	++tuple.ptr;
	goto chain_start_label;
 end_tuple_label:
	debug(("END tuple\n"));
	parse_tuple(socksc, &tuple);
	++tuple.ptr;
	tuple.length = pcmcia_cis_read_1(&tuple, tuple.ptr + 1);
	goto parse_tuple_label;
 longlink_a_label:
 longlink_c_label:
	debug(("LONGLINK tuple\n"));
	if (tuple.length < 4) {
		debug(("longlink length too short\n"));
		goto end_label;
	}
	longlink_addr = pcmcia_tuple_read_4(&tuple, 0);
	longlink_present = 1;
	debug(("longlink is %x\n", longlink_addr));
	goto next_chain_label;
 nolink_label:
	debug(("NOLINK tuple\n"));
	longlink_present = 0;
	goto next_chain_label;
 checksum_label:
	debug(("CHECKSUM tuple\n"));
	if (tuple.length < 5) {
		debug(("checksum length is too short\n"));
		goto end_label;
	}
	next_state = cis_checksum_ok(&tuple) ? &&next_chain_label : &&end_label;
	goto *next_state;
 longlink_mfc_label:
	debug(("LONGLINK MFS tuple\n"));
	if (tuple.length < 6) {
		debug(("longlink mfc too short %d\n", tuple.length));
		goto end_label;
	}
	if (((tuple.length - 1) % 5) != 0) {
		debug(("longlink mfs bogus length %d\n", tuple.length));
		goto end_label;
	}
	{
		int mfc_count = pcmcia_tuple_read_1(&tuple, 0);
		int i;
		
		for (i = 0; i < mfc_count; i++) {
		}
	}
	goto parse_tuple_label;
 parse_tuple_label:
	debug(("parse tuple\n"));
	parse_tuple(socksc, &tuple);
	goto next_tuple_label;
 next_tuple_label:
	debug(("next_tuple\n"));
	tuple.ptr += 2 + tuple.length;
	goto chain_start_label;
 next_chain_label:
	debug(("next chain\n"));
	if (longlink_present) {
#if 0
		pcmcia_mem_map(socksc, longlink_addr * 2);
		tuple.ptr = 0;
#endif		
	}
	if (!tuple_valid_linktarget(&tuple))
		goto end_label;
	goto chain_start_label;
 end_label:
	debug(("CIS read done\n"));
	pcic_chip_mem_unmap(socksc, win);
}


/*
 * pcmcia_init_io - initialize the I/O mappings
 */
static void
pcmcia_init_io(struct pcmcia_socket *sc)
{
	debug(("pcmcia_init_io: sock %s\n", pcmcia_sockname(sc)));
	
	switch (sc->id) {
	case PCMCIA_SOCKID_A:
		sc->niowindows = 1;
		sc->iomap[0].addr = galileo_get_pci0_iobase() + PCMCIA_IOWINDOW;
		sc->iomap[0].size = 0x10;
		break;
	case PCMCIA_SOCKID_B:
		sc->niowindows = 1;
		sc->iomap[0].addr = galileo_get_pci0_iobase() + PCMCIA_IOWINDOW +
			0x400;
		sc->iomap[0].size = 0x10;
		break;
	default:
		assert(0);
	}
}


/*
 * pcmcia_init_socket - initialize a socket
 */
static void
pcmcia_init_socket(struct pcmcia_socket *socksc, int id,
				   struct pcmcia *psc)
{
	assert(pcmcia_valid_socket(id));
	
	debug(("pcmcia_init_socket: %s\n", id == PCMCIA_SOCKID_A ? "A" : "B"));

	socksc->pcmcia = psc;
	socksc->id = id;
	socksc->state = PCMCIA_SOCK_UNITIALIZED;
	SIMPLEQ_INIT(&socksc->pf_head);
}


/*
 * pcmcia_enable_sockets - initialize the sockets
 */
static void
pcmcia_enable_sockets(struct pcmcia *sc)
{
	int i;

	debug(("pcmcia_enable_sockets\n"));

	for (i = 0; i < 2; ++i) {
		struct pcmcia_socket *socksc = pcmcia_socket(sc, i);
		pcmcia_enable_socket(socksc);
		if (pcmcia_device_present(socksc)) {
			pcmcia_init_mem_map(socksc);
			pcmcia_read_cis(socksc);
			pcmcia_init_io(socksc);
		}
	}
}


/*
 * pcmcia_alloc_sockets - allocate the socket structures
 */
static void
pcmcia_alloc_sockets(struct pcmcia *sc)
{
	int i;

	debug(("pcmcia_alloc_sockets\n"));

	for (i = 0; i < 2; ++i) {
		sc->sockets[i] = pcmcia_alloc_socket(i);
		pcmcia_init_socket(pcmcia_socket(sc, i), i, sc);
	}
}


/*
 * pcmcia_probe_self - read the pcmcia bridge identification registers
 */
static int
pcmcia_probe_self(struct pcmcia *sc)
{
	uint8 ident;

	debug(("pcmcia_probe_self: "));

	ident = pcic_read(pcmcia_socket(sc, PCMCIA_SOCKID_A), PCIC_IDENT);
	debug(("ident A=%x ", ident));
	ident = pcic_read(pcmcia_socket(sc, PCMCIA_SOCKID_B), PCIC_IDENT);
	debug(("ident B=%x\n", ident));

	return NFX_OK;
}


/*
 * pcmcia_socket_statename - return the printable state name
 */
static char *
pcmcia_socket_statename(int state)
{
	switch (state) {
	case PCMCIA_SOCK_PRESENT:
		return "CARD PRESENT";
	case PCMCIA_SOCK_ENABLED:
		return "ENABLED";
	case PCMCIA_SOCK_NOTPRESENT:
		return "CARD NOT PRESENT";
	case PCMCIA_SOCK_UNITIALIZED:
		return "UNITIALIZED";
	default :
		assert(0);
		return 0;
	}
}

/*
 * pcmcia_socket_print_status - 
 */
static void
pcmcia_socket_print_status(struct pcmcia_socket *sc)
{
	printf("socket %s: %s: ", pcmcia_sockname(sc),
		   pcmcia_socket_statename(sc->state));
}


/*
 * pcmcia_print_status - print the state of sockets
 */
static void
pcmcia_print_status(struct pcmcia *sc)
{
	if ((disk_debug_param & DEBUG_PCMCIA)) {
		int i = 0;

		printf("pcmcia_status: ");
		for (i = 0; i < 2; ++i) {
			pcmcia_socket_print_status(pcmcia_socket(sc, i));
		}
		printf("\n");
	}
}


/*
 * pcmcia_gt64240_init - additional inititalization for galileo
 */
// This is Bobcat PROM / Diags.. No need for Cheatah stuff
//static void
//pcmcia_gt64240_init()
//{
//	*(uint32*)(0xb4000000 + PCI_0MEMORY2_ADDRESS_REMAP) = 0x8;
//}

//#ifdef PROM_R9000
static void
pcmcia_gt64440_init()
{
        GT_WRITE(PCI_1MEMORY2_ADDRESS_REMAP, 0x00000080);
}
//#endif

// Make compiler happy
void pcic_wait_fifo(void);

void pcic_wait_fifo(void)
{
    // TI part RoHS doesn't have this register
    // intel_cf_part is set during init 
    if( intel_cf_part ) {
	while (!(pcic_read(&pcmcia_sockets[0], 0x17) & 0x80));
	while (!(pcic_read(&pcmcia_sockets[0], 0x17) & 0x80));
    }
}

/*
 * pcmcia_init - initialize pcmcia bridge 
 */
static int
pcmcia_init()
{
	uint32 dev;
	uint32 bus;
	int rc;

	debug(("pcmcia_init:\n"));

// This is Bobcat PROM / Diags.. No need for Cheatah stuff
// #ifdef R9000
	pcmcia_gt64440_init();
// #else
// 	pcmcia_gt64240_init();
// #endif

	if ((rc = pcmcia_pci_find(&bus, &dev)) == NFX_OK) {
		struct pcmcia *sc = pcmcia_alloc_sc();
		sc->dev = dev;
		sc->bus = bus;
		debug(("pcmcia_init: alloced sc: dev=%d\n", sc->dev));
		pcmcia_read_baseaddr(sc);
		debug(("pcmcia_init: read baseadddr: %x\n", sc->regs));
		pcmcia_alloc_sockets(sc);
		if ((rc = pcmcia_probe_self(sc)) == NFX_OK) {
			pcmcia_enable_sockets(sc);
		}
		pcmcia_print_status(sc);
		pcic_write(pcmcia_socket(sc, PCMCIA_SOCKID_A), 0x3c, 0x1);
	}
	

	return rc;
}

/*
 * pcmcia_init - initialize pcmcia cards if not initialized yet
 */
int
do_pcmcia_init()
{
	if (!pcmcia_initialized) {
		pcmcia_init();
		pcmcia_initialized = 1;
	}

	return NFX_OK;
}

/*****************************************************
* WTG  needed a diag function that returns 
*      socket id's.  status
*  Sort of a copy of the func pcmcia_enable_sockets()
*****************************************************/
// Make compiler happy
int diag_pcmcia_socket_check(void);

int
diag_pcmcia_socket_check(void)
{
        uint32 dev;
        uint32 bus;
        int rc;
        int i,devPresent;

#ifdef R9000
        pcmcia_gt64440_init();
#endif
        if ((rc = pcmcia_pci_find(&bus, &dev)) == NFX_OK) {
	    struct pcmcia *sc = pcmcia_alloc_sc();
            sc->dev = dev;
            sc->bus = bus;
            pcmcia_read_baseaddr(sc);
            pcmcia_alloc_sockets(sc);
            if ((rc = pcmcia_probe_self(sc)) == NFX_OK) {
                pcmcia_enable_sockets(sc);
            }
            pcic_write(pcmcia_socket(sc, PCMCIA_SOCKID_A), 0x3c, 0x1);

            // Init done now check for device present
            devPresent = 0; 
            for (i = 0; i < 2; ++i) {
                struct pcmcia_socket *socksc = pcmcia_socket(sc, i);
                pcmcia_enable_socket(socksc);
                if (pcmcia_device_present(socksc)) {
                   // if its there just set a 1 in devPresent
                   // bit 0 for Dev in socket A  
                   // bit 1 for Dev in Socket B
                   devPresent |= ((i) +1);
                }
            }
        }
        else {
            return( devPresent = 8);  // bridge error 
        }
        return(devPresent);
}


static int
pcmcia_ccr_read(pf, ccr)
	struct pcmcia_function *pf;
	int ccr;
{
	return *(uchar8 *)(pf->pf_ccrh + pf->pf_ccr_offset + ccr);
}

static void
pcmcia_ccr_write(pf, ccr, val)
	struct pcmcia_function *pf;
	int ccr;
	int val;
{
	if ((pf->ccr_mask) & (1 << (ccr / 2))) {
		*(uchar8 *)(pf->pf_ccrh + pf->pf_ccr_offset + ccr) = val;
	}
}


void
pcmcia_function_init(struct pcmcia_function *pf,
					 struct pcmcia_config_entry *cfe)
{
	pf->cfe = cfe;
}


int
pcmcia_function_enable(pf)
	struct pcmcia_function *pf;
{
	struct pcmcia_function *tmp = 0;
	int reg;

	if (pf->cfe == NULL)
		panic("pcmcia_function_enable: function not initialized");

#if 0
	/*
	 * Increase the reference count on the socket, enabling power, if
	 * necessary.
	 */
	if (pf->sc->sc_enabled_count++ == 0)
		pcmcia_chip_socket_enable(pf->sc->pct, pf->sc->pch);
	DPRINTF(("%s: ++enabled_count = %d\n", pf->sc->dev.dv_xname,
		 pf->sc->sc_enabled_count));
#endif	

#if 0
	if (pf->pf_flags & PFF_ENABLED) {
		/*
		 * Don't do anything if we're already enabled.
		 */
		debug(("%s: pcmcia_function_enable on enabled func\n"));
		return (0);
	}

	/* If there was no CIS don't mess with CCR */
	if (pf->pf_flags & PFF_FAKE)
		goto done;
#endif

#if 0
	/*
	 * It's possible for different functions' CCRs to be in the same
	 * underlying page.  Check for that.
	 */
	for (tmp = pf->sc->card.pf_head.sqh_first; tmp != NULL;
	    tmp = tmp->pf_list.sqe_next) {
		if ((tmp->pf_flags & PFF_ENABLED) &&
		    (pf->ccr_base >= (tmp->ccr_base - tmp->pf_ccr_offset)) &&
		    ((pf->ccr_base + PCMCIA_CCR_SIZE) <=
		     (tmp->ccr_base - tmp->pf_ccr_offset +
		      tmp->pf_ccr_realsize))) {
			pf->pf_ccrt = tmp->pf_ccrt;
			pf->pf_ccrh = tmp->pf_ccrh;
			pf->pf_ccr_realsize = tmp->pf_ccr_realsize;

			/*
			 * pf->pf_ccr_offset = (tmp->pf_ccr_offset -
			 * tmp->ccr_base) + pf->ccr_base;
			 */
			pf->pf_ccr_offset =
			    (tmp->pf_ccr_offset + pf->ccr_base) -
			    tmp->ccr_base;
			pf->pf_ccr_window = tmp->pf_ccr_window;
			break;
		}
	}

#endif	

	if (tmp == NULL) {
		pf->pf_ccrh = pcmcia_mem_alloc(PCMCIA_CCR_SIZE);
		pcic_chip_mem_map(pf->socket, PCMCIA_MEM_ATTR, pf->ccr_base,
						  PCMCIA_CCR_SIZE,
						  pf->pf_ccrh, &pf->pf_ccr_offset,
						  &pf->pf_ccr_window);
	}

	reg = (pf->cfe->number & PCMCIA_CCR_OPTION_CFINDEX);
	reg |= PCMCIA_CCR_OPTION_LEVIREQ;
#if 0
	if (pcmcia_mfc(pf->sc)) {
		reg |= (PCMCIA_CCR_OPTION_FUNC_ENABLE |
			PCMCIA_CCR_OPTION_ADDR_DECODE);
		if (pf->ih_fct)
			reg |= PCMCIA_CCR_OPTION_IREQ_ENABLE;

	}
#endif
	
	pcmcia_ccr_write(pf, PCMCIA_CCR_OPTION, reg);

	reg = 0;

	if ((pf->cfe->flags & PCMCIA_CFE_IO16) == 0)
		reg |= PCMCIA_CCR_STATUS_IOIS8;
	if (pf->cfe->flags & PCMCIA_CFE_AUDIO)
		reg |= PCMCIA_CCR_STATUS_AUDIO;
	pcmcia_ccr_write(pf, PCMCIA_CCR_STATUS, reg);

	pcmcia_ccr_write(pf, PCMCIA_CCR_SOCKETCOPY, 0);
	
#if 0
	if (pcmcia_mfc(pf->sc)) {
		long tmp, iosize;

		tmp = pf->pf_mfc_iomax - pf->pf_mfc_iobase;
		/* round up to nearest (2^n)-1 */
		for (iosize = 1; iosize < tmp; iosize <<= 1)
			;
		iosize--;

		pcmcia_ccr_write(pf, PCMCIA_CCR_IOBASE0,
				 pf->pf_mfc_iobase & 0xff);
		pcmcia_ccr_write(pf, PCMCIA_CCR_IOBASE1,
				 (pf->pf_mfc_iobase >> 8) & 0xff);
		pcmcia_ccr_write(pf, PCMCIA_CCR_IOBASE2, 0);
		pcmcia_ccr_write(pf, PCMCIA_CCR_IOBASE3, 0);

		pcmcia_ccr_write(pf, PCMCIA_CCR_IOSIZE, iosize);
	}
#endif	

#if defined(DEBUG)
	tmp = pf;
#if 0
	for (tmp = pf->sc->card.pf_head.sqh_first; tmp != NULL;
	     tmp = tmp->pf_list.sqe_next) {
#endif		
		debug(("function %d CCR at %d offset %x: "
		       "%x %x %x %x, %x %x %x %x, %x\n",
		       tmp->number,
		       tmp->pf_ccr_window, tmp->pf_ccr_offset,
		       pcmcia_ccr_read(tmp, 0x00),
		       pcmcia_ccr_read(tmp, 0x02),
		       pcmcia_ccr_read(tmp, 0x04),
		       pcmcia_ccr_read(tmp, 0x06),

		       pcmcia_ccr_read(tmp, 0x0A),
		       pcmcia_ccr_read(tmp, 0x0C), 
		       pcmcia_ccr_read(tmp, 0x0E),
		       pcmcia_ccr_read(tmp, 0x10),

		       pcmcia_ccr_read(tmp, 0x12)));
#if 0
	}
#endif
#endif

#if 0
 done:
	pf->pf_flags |= PFF_ENABLED;
#endif	
	delay_us(1000000);
	return (0);

#if 0
 bad:
	/*
	 * Decrement the reference count, and power down the socket, if
	 * necessary.
	 */
	if (--pf->sc->sc_enabled_count == 0)
		pcmcia_chip_socket_disable(pf->sc->pct, pf->sc->pch);
	debug(("%s: --enabled_count = %d\n", pf->sc->dev.dv_xname,
		 pf->sc->sc_enabled_count));
#endif	

	return (1);
}
