/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ipmd.c
 *
 * RCS:  $Id: ipmd.c,v 1.11 2002/02/04 19:11:41 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: ipm processing on ssc 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/11/2001
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h>
#if defined(SSC_x86) || defined(SSC_mips)
#include <neteee/eee.h>
#include <neteee/eee_if.h>
#endif /* SSC_x86 SSC_mips */
#include <sys/types.h>
#include <sys/shm.h> 
#include <sys/ipc.h>
#include <sys/sem.h>

#include <nfx-error.h>
#include <nfx-types.h>
#include <nfx-defs.h>
#include <exports/anpssc-api.h>
#include <exports/ppm-api.h>
#include <exports/ipmd-lib.h>
#include <exports/ipm-api.h>
#include "../sm-ipm/ipm.h"


/* msg process functions */
static void  ipmd_ProcessMsgFromSSC(int, char*, int, int, void*, int, void*);
static void  ipmd_ProcessMsg(int, int, void*, int, uint32, uint32, void*);	
static int32 ipmd_ProcRespIfMsg(ipm_mintf_t*, int, int);
static int32 ipmd_ProcStatMsg(int, void*);
static int32 ipmd_ProcNRespIfMsg(ipm_mintf_t*, int, int);
static int32 ipmd_ProcConnMsg(uint32, uint32, int, ipm_mconn_t*);

/* functions to update shared memory, the read part of */
/* func are in ipm-util.c because they need to shared with ipmd-lib*/
static int32 ipmd_InitShmem(void);
static int32 ipmd_InsConn(ipm_conn_t*,	uint32*, cmp_func);
static int32 ipmd_DelConn(ipm_conn_t*,	uint32*, cmp_func);


/* ipm stats */
ipm_stats_t		ipm_stats;

/* shmem ptr */
int				shmid;			/* shmem id */
void			*shmaddr;		/* shmem base addr */
ipm_shmroot_t	*rootaddr;		/* address of root struct */
FILE			*shm_fd;		/* file descriptor to create shmem */
int				sem_id;

/* on/off switch */
int				ipmd_active = 1;

void
ipmd_Init()
{
	
	if (ipmd_active == 1) {
  	/* register with ppm transport layer for the first time */
		if (ppmd_RegRcv(PPM_IPM, 0, ipmd_ProcessMsg, ipmd_ProcessMsgFromSSC) 
			!= NFX_OK) {
			printf("ipm: register failed\n");
			return;	
		}
	}

  /* reset stats counter */
  bzero(&ipm_stats, sizeof(ipm_stats));

  /* initiate shared mem */
  if (ipmd_InitShmem() != NFX_OK)
	{
	  printf("ipm: initate share mem failed\n");
	  ppmd_UnregRcv(PPM_IPM);
	  return;
	}
  
  ipmd_active = 1;

  return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcessMsg
 *
 * Description: process ipm messages from another card. 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/11/2001
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_ProcessMsg(int		gid,
			   int		action,
			   void		*msg,	
			   int		msgsize,
			   uint32	scpu,
			   uint32	sslot,
			   void		*parm	
			   )
{
  ipm_mhdr_t	*ipm_hdrp  = (ipm_mhdr_t*)msg;
  ipm_mintf_t	*ipm_intfp = (ipm_mintf_t*)(ipm_hdrp+1);
  void			*ipm_statp = (ipm_mintf_t*)(ipm_hdrp+1); 
  ipm_mchdr_t	*ipm_connhdrp = (ipm_mchdr_t*)(ipm_hdrp+1);
  int			mtype;

  if (!ipmd_active)
	return;

  /* sanity check */
  if (gid != PPM_IPM || (msgsize != ipm_hdrp->ipmh_size))
	{
	  IPM_LOGERR(IPM_ERR_RCV);
	  ipm_stats.err_rcv++;
	  return;
	}
  
  /* retrieve action */
  mtype = ipm_hdrp->ipmh_mtype;


  /* currently only IPM_MSG_INTF_ADD is sent by TXRX, */
  /* in the event of startup */
  switch(mtype)
	{
	case IPM_MSG_INTF_NEW:
	case IPM_MSG_INTF_RESP:
	  ipmd_ProcRespIfMsg(ipm_intfp, sslot, scpu);
	  ipm_stats.dev_rcv++;
	  break;

	  /* not used */
	case IPM_MSG_INTF_ERR:
	  ipmd_ProcNRespIfMsg(ipm_intfp, sslot, scpu);
	  ipm_stats.adm_rcv++;
	  break;

	case IPM_MSG_INTF_STATS:
	case IPM_MSG_IP_STATS:
	case IPM_MSG_ICMP_STATS:
	case IPM_MSG_TCP_STATS:
	case IPM_MSG_UDP_STATS:
	  if (ipmd_ProcStatMsg(mtype, ipm_statp) == NFX_OK)
		ipm_stats.info_rcv++;
	  else
		ipm_stats.err_rcv++;
	  break;

	case IPM_MSG_TCP_CONN:
	case IPM_MSG_UDP_CONN:
	  {
		ipm_mconn_t		*ipm_connp = (ipm_mconn_t*)(ipm_connhdrp+1);
		if (ipmd_ProcConnMsg(mtype, sslot, ipm_connhdrp->count, ipm_connp)
			== NFX_OK)
		  ipm_stats.info_rcv++;
		else
		  ipm_stats.err_rcv++; 
		break;
	  }
	  
	default:
	  ipm_stats.err_rcv++;
	}
  
  return;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcRespIfMsg
 *
 * Description: 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/11/2001
 *
 *-----------------------------------------------------------------
 */
static int32
ipmd_ProcRespIfMsg(ipm_mintf_t	*ipm_intfp,
				   int			sslot,
				   int			scpu)
{
  uchar8		ifname[24];
  ipm_if_t		ipm_if;
  int			count = 0, need=0;
  /* user interface */ 
  int			link;
  

  link = GET_ULINK_FROM_CPULINK(ipm_intfp->if_cpu, ipm_intfp->if_link);
  sprintf(ifname, "%s:%d/%d", bsd_ifnamebuf[ipm_intfp->if_dtype],
		  ipm_intfp->if_slot, link);
  
#if defined(SSC_x86) || defined(SSC_mips)  
  if (ipmd_GetIfByName(ifname, &ipm_if, sizeof(ipm_if), &count, &need)
	  == IPM_GET_ERR || 
	  count == 0)
	{
	  /* interface not found, add to kernel */
	  
         struct eee_attachreq req;
          int rc;
 
          req.addr.seee_slot = ipm_intfp->if_slot;
          req.addr.seee_cpu = ipm_intfp->if_cpu;
          req.link = link;
          strcpy(req.name, ifname);
          memcpy(req.ether_addr, ipm_intfp->if_mac, 6);
 printf("new interface %s detected \n", ifname);
          rc = ethereee_ifattach(&req);
 
          printf("rc = %d error=%s\n", rc, strerror(errno));
	  
	}
#endif /* SSC_x86 SSC_mips */
  return NFX_OK;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcNRespIfMsg
 *
 * Description: 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/11/2001
 *
 *-----------------------------------------------------------------
 */
static int32 ipmd_ProcNRespIfMsg(ipm_mintf_t *ipm_intfp, int sslot, int scpu)
{
  /* remove request records */
  return NFX_ERR;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_ProcessMsgFromSSC
 *
 * Description: process ipm messages from modules on SSC. 
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/11/2001
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_ProcessMsgFromSSC(int		mtype, 
					   char		*app,
					   int		gid,
					   int		action,
					   void		*msg,
					   int		msgsize,
					   void		*parm1
					   )
{
  ipm_mhdr_t	*ipm_hdrp = (ipm_mhdr_t*)msg;
  ipm_mintf_t	*ipm_intfp;
  int			slot;
  int			cpu;
  int			link;
  char			ifname[IPMD_MAX_STRING];
  ipm_if_t		ipm_if;
  int			count=0, need = 0;


  if ((!ipmd_active) && (mtype != IPM_MSG_RESTART))
	return;

  /* sanity check */
  if ((gid != PPM_IPM) ||
	  (msgsize != ipm_hdrp->ipmh_size))
	{
	  IPM_LOGERR(IPM_ERR_RCV);
	  ipm_stats.err_rcv++;
	  return;
	}

  switch (mtype)
	{
	case IPM_MSG_IP_ADD:
	case IPM_MSG_INTF_ENA:
	case IPM_MSG_INTF_DIS:
	case IPM_MSG_IP_DEL:
	  ipm_stats.adm_rcv++;
	  ipm_intfp = (ipm_mintf_t*)(ipm_hdrp + 1);
	  slot = ipm_intfp->if_slot;
	  cpu  = ipm_intfp->if_cpu;
	  link = ipm_intfp->if_link;

	  link =  GET_ULINK_FROM_CPULINK(cpu, link);
	  sprintf(ifname, "%s:%d/%d", bsd_ifnamebuf[ipm_intfp->if_dtype],
			  slot, link);

#if defined(SSC_x86) || defined(SSC_mips) /* linux does not work */
	  if (ipmd_GetIfByName(ifname, &ipm_if, sizeof(ipm_if), &count, &need)
		  == IPM_GET_ERR || 
		  count == 0)
		  /* perhaps we should send a request */
		  return;
#endif /* SSC_x86 SSC_mips */

	  /* we need to fill in the intf content */
	  ipm_intfp->if_flags = ipm_if.if_flags;
	  ipm_intfp->if_dhdrlen = 14; /* fixed value, */
	  ipm_intfp->if_daddrlen = 6;  /* we might have to get from kernel */
	  ipm_intfp->if_mtusize = ipm_if.if_mtusize;
	  memcpy(ipm_intfp->if_mac, ipm_if.if_mac, 6);
	  break;

	case IPM_MSG_ARP_ADD:
	case IPM_MSG_ARP_DEL:
	case IPM_MSG_RT_ADD:
	case IPM_MSG_RT_DEL:
	  ipm_stats.adm_rcv++;
	  break;

	case IPM_MSG_SHUTDOWN:
	  semctl(sem_id, 0, IPC_RMID);
	  shmdt(shmaddr);
	  shmctl(shmid, IPC_RMID, 0); /* remove shared memory and exit */
	  fclose(shm_fd);
	  ipmd_active = 0;
	  return;

	case IPM_MSG_RESTART:
	   printf("restart rcved \n");
	  ipmd_Init();
	  ipmd_active = 1;
	  return;

	default:
	  IPM_LOGERR(IPM_ERR_RCV);
	  ipm_stats.err_rcv++;
	  return;
	}

  /* relay msg to txrx, bcast */
  printf("relay msg from ssc to txrx \n");
  // bsd_ShowPkt(msg, msgsize);		
  if (ppmd_SendMsg(gid, action, 0, 0, msg, msgsize, 1) ==
	  NFX_OK)
	ipm_stats.adm_fwd++;
  else
	ipm_stats.err_fwd++;

  return;
}
	
/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcStatMsg
 *
 * Description: cpy rcved msg to shared memory
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
static int32
ipmd_ProcStatMsg(int mtype, void *msg)
{
  int			size;
  int			type, slot, link;
  ipm_shmcard_t	*cardaddr;
  void			*srcaddr, *destaddr;

  srcaddr = msg;

  switch(mtype)
	{
	case IPM_MSG_INTF_STATS:
	  {
		ipm_ifstat_t		*ifsp;

		size = sizeof(ipm_ifstat_t);
		ifsp = (ipm_ifstat_t*)msg;
		/* get interface */
		if ((ipmd_ParseIfName(ifsp->ifi_ifname, &type, &slot, &link) != NFX_OK) ||
			(type != IPM_INTF_TYPE_ETH))
		  return NFX_ERR;

		/* find the memory space responding to stats of this interface */
		cardaddr = IPM_SHMGET_CARD(shmaddr, slot);
		if (cardaddr->ifcnt <= link) /* invalid link num. */
		  return NFX_ERR;

		destaddr = IPM_SHMGET_ADDR(shmaddr, 
				   cardaddr->ifstatp[link]);
	  }
	  break;

	case IPM_MSG_IP_STATS:
	  {
		ipm_ipstat_t		*ipsp;

		size = sizeof(ipm_ipstat_t);
		ipsp = (ipm_ipstat_t*)msg;
		slot = ipsp->ips_card;

		cardaddr = IPM_SHMGET_CARD(shmaddr, slot);
		if (!cardaddr->ipstatp)
		  return NFX_ERR;

		destaddr = IPM_SHMGET_ADDR(shmaddr, 
								   cardaddr->ipstatp);
	  }
	  break;

	  
	case IPM_MSG_ICMP_STATS:
	  {
		ipm_icpstat_t		*icmpsp;

		size = sizeof(ipm_icpstat_t);
		icmpsp = (ipm_icpstat_t*)msg;
		slot = icmpsp->icps_card;

		cardaddr = IPM_SHMGET_CARD(shmaddr, slot);
		if (!cardaddr->icmpstatp)
		  return NFX_ERR;
		
		destaddr = IPM_SHMGET_ADDR(shmaddr, 
								   cardaddr->icmpstatp);
	  }
	  break;

	case IPM_MSG_TCP_STATS:
	  {
		ipm_tcpstat_t		*tcpsp;

		size = sizeof(ipm_tcpstat_t);
		tcpsp = (ipm_tcpstat_t*)msg;
		slot = tcpsp->tcps_card;

		cardaddr = IPM_SHMGET_CARD(shmaddr, slot);
		if (!cardaddr->tcpstatp)
		  return NFX_ERR;

		destaddr = IPM_SHMGET_ADDR(shmaddr, 
								   cardaddr->tcpstatp);
	  }
	  break;	  


	case IPM_MSG_UDP_STATS:
	  {
		ipm_udpstat_t		*udpsp;
		
		size = sizeof(ipm_udpstat_t);
		udpsp = (ipm_udpstat_t*)msg;
		slot = udpsp->udps_card;
		
		cardaddr = IPM_SHMGET_CARD(shmaddr, slot);
		if (!cardaddr->udpstatp)
		  return NFX_ERR;
		
		destaddr = IPM_SHMGET_ADDR(shmaddr, 
								   cardaddr->udpstatp);
	  }
	  break;

	default:
	  return NFX_ERR;
	}

  /* destaddr, srcaddr already setup */
  memcpy(destaddr, srcaddr, size);
  return NFX_OK;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcConnMsg
 *
 * Description: process connection update messages
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
static int32 
ipmd_ProcConnMsg(uint32			mtype, 
				 uint32			sslot,
				 int			count, 
				 ipm_mconn_t	*ipm_connp)
{
  uint32		connroot;
  int			i;
  uint32		*conns;

  /* which record we want to modify */
  if (mtype == IPM_MSG_UDP_CONN)
	{
	  connroot = rootaddr->udpconnp;
	  conns = &(rootaddr->udpcnt);
	}
  else
	{
	  connroot = rootaddr->tcpconnp;
	  conns = &(rootaddr->tcpcnt);
	}

  /* set semaphore */
  ipmd_SetSem((int)rootaddr->sem_id, IPM_SEM_CONN, 1);
  
  /* walk through each msg on list */
  for (i = 0; i < count; ipm_connp++, i++)
	{
	  if (ipm_connp->cmd == IPM_ADD)
		{
		  if (ipmd_InsConn(&(ipm_connp->conn), &connroot, ipmd_CmpConn)
			  == NFX_OK)
			(*conns)++;
		}
	  else
		{
		  if (ipmd_DelConn(&(ipm_connp->conn), &connroot, ipmd_CmpConn)
			  == NFX_OK)
			(*conns)--;
		}
	}

  /* update connroot as they might be changed */
    if (mtype == IPM_MSG_UDP_CONN)
	{
	  rootaddr->udpconnp = connroot;
	}
  else
	{
	  rootaddr->tcpconnp = connroot;
	}

    /* release semaphore */
  ipmd_SetSem((int)rootaddr->sem_id, IPM_SEM_CONN, 0);	
  return NFX_OK;
}

#define IPM_MAX_APPS			100
#define IPM_SHMEM_SIZE			1024*1024			
#define IPM_MAX_CONN_ONCHASSIS	4096*4

/*-----------------------------------------------------------------
 *
 * Name: ipmd_InitShmem
 *
 * Description: init shmem.
 * shmem structure:
 *		
 *      ipm_shmroot_t;
 *		ipm_shmcard_t * NFX_NUM_SLOTS_PER_CHASSIS;
 *      ipm_shmerr_t * IPM_MAX_APPS;
 *		
 *	for each valid card:
 *	    one or more ipm_ifstat_t
 *		ipm_ipstat_t
 *		ipm_icpstat_t
 *		ipm_tcpstat_t
 *		ipm_udpstat_t
 *		4000*ipm_conn;
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/11/2001
 *
 *-----------------------------------------------------------------
 */
static int32					/* NFX_OK, NFX_ERR */
ipmd_InitShmem()
{
  uint32		size = 0;		/* actul size */
  ipm_shmcard_t	*cardaddr;
  ipm_shmerr_t	*erraddr;
  uint32		cardsize = 0;
  int			i, connsize;
  ipm_shmconn_t *connaddr;

#if defined(LINUX_TEST)
#define SEM_R		0
#define SEM_A		0			/* linux dont support these */
#endif /* LINUX_TEST */

  /* initilize semaphore to shared mem */
  if (ipmd_GetSem(IPM_SHMEM_ID, IPM_SEM_MAX, 
				  IPC_CREAT|SEM_R|SEM_A|
				  (SEM_R|SEM_A)>>3|(SEM_R|SEM_A)>>6,
				  &sem_id) != NFX_OK)
	  return NFX_ERR;

  /* create file to determine shared memory */
  if ((shm_fd = fopen("/tmp/ipm_shmem123", "w+"))==0)
	  return NFX_ERR;
  
  if (ipmd_OpenShmem(IPM_SHMEM_ID, IPM_SHMEM_SIZE, 
					 IPC_CREAT|SHM_R|SHM_W, &shmaddr, &shmid) != NFX_OK)
	return NFX_ERR;


  
  /* initialize structures */
  bzero(shmaddr, IPM_SHMEM_SIZE);

  
  /* shmroot_t */
  rootaddr = (ipm_shmroot_t*)shmaddr;
  rootaddr->sem_id = (uint32)sem_id;
  size += ROUNDUP(sizeof(*rootaddr));
  rootaddr->cardp = size;
  size += (ROUNDUP(sizeof(*cardaddr)) * NFX_NUM_SLOTS_PER_CHASSIS);
  rootaddr->errp = size;		/* offset ptr is beyound ipm_shmcard_t */
  size += ROUNDUP(sizeof(*erraddr)) * IPM_MAX_APPS; 
								/* offset ptr is beyond ipm_shmerr_t */

  /* starting conn. stack at offset aligned to its size */
  connsize = ROUNDUP(sizeof(ipm_shmconn_t));
  if (size%connsize)
	  size = ((size/connsize)+1)*connsize;

  /* initialize free conn stack, which is shared by all cards */
  rootaddr->firstconnp = rootaddr->freeconnp = size;
  rootaddr->freeconncnt = IPM_MAX_CONN_ONCHASSIS;
  for (i=0, connaddr = (ipm_shmconn_t*)IPM_SHMGET_ADDR(shmaddr, size);
	   i < (IPM_MAX_CONN_ONCHASSIS - 1); i++, connaddr++)
	{
	  /* do not setup the last element */
	  size += connsize;
	  connaddr->next = size;
	}
  
  /* takecare of size for last node */
  size += connsize;
  rootaddr->lastconnp = size;
								/* offset ptr beyond free conn tbl, */

  

  /* save cardaddr and erraddr for local processing */
  cardaddr = (ipm_shmcard_t*)IPM_SHMGET_ADDR(shmaddr, rootaddr->cardp);
  erraddr  = (ipm_shmerr_t*)IPM_SHMGET_ADDR(shmaddr, rootaddr->errp);

  /* shmcard_t */
  /* assuming card configuration is static for now */

  
  cardaddr = (ipm_shmcard_t*)
	((char*)cardaddr + ROUNDUP(sizeof(ipm_shmcard_t)));	/* 0 reserved for ssc */
  cardaddr->card = 1;			/* first txrx */

  cardaddr->ifcnt = IPM_LINK_MAX;
  for (i = 0; i< IPM_LINK_MAX; i++)
	{
	  cardaddr->ifstatp[i] = size;
	  size += ROUNDUP(sizeof(ipm_ifstat_t));
	  cardsize += ROUNDUP(sizeof(ipm_ifstat_t));
	}
								/* offset ptr beyond ifstat */

  cardaddr->ipstatp = size;
  size += ROUNDUP(sizeof(ipm_ipstat_t));
  cardsize += ROUNDUP(sizeof(ipm_ipstat_t));
								/* offset ptr beyond ipstat */

  cardaddr->icmpstatp = size;
  size += ROUNDUP(sizeof(ipm_icpstat_t));
  cardsize += ROUNDUP(sizeof(ipm_icpstat_t));
								/* offset ptr beyond icmptat */

  cardaddr->tcpstatp = size;
  size += ROUNDUP(sizeof(ipm_tcpstat_t));
  cardsize += ROUNDUP(sizeof(ipm_tcpstat_t));
								/* offset ptr beyond tcptat */

  cardaddr->udpstatp = size;
  size += ROUNDUP(sizeof(ipm_udpstat_t));
  cardsize += ROUNDUP(sizeof(ipm_udpstat_t));
								/* offset ptr beyond udptat */
								/* ready for next valid card */  
  
  
  /* finish off */
  printf("shmid %d, total mem size: %d, size per card %d\n", 
		 shmid, size, cardsize);
  ipmd_PrintShmem((void*)shmaddr);
  if (size > IPM_SHMEM_SIZE)
	{							/* require more than allocated */
	  IPM_LOGERR(IPM_ERR_SHMEM);
	  shmctl(shmid, IPC_RMID, 0);
	  return NFX_ERR;
	}

  return NFX_OK;
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_InsConn
 *
 * Description: insert conn. to UPD/TCP conn table. 
 *
 * *connroot: the offset to start of shmaddr that represents
 *			  the said conn. table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
#define IPM_DUP	(NFX_ERR - 1)
int32							/* NFX_OK, NFX_ERR, IPM_DUP */
ipmd_InsConn(ipm_conn_t	*connp, 
			 uint32		*connroot, 
			 cmp_func	cmp)
{
  ipm_shmconn_t			*connaddr=0, *prev_connaddr=0, *new_connaddr; 
								/* absolute address */
  uint32				next, new_conn;
								/* offset to shmaddr */
  int					rc;
  

  /* find the place for insertion */
  for (next=*connroot; next != 0;/* end of table */
	   next = connaddr->next)
	{
	  connaddr = (ipm_shmconn_t*)IPM_SHMGET_ADDR(shmaddr, next);
	  rc = (*cmp)(connp, &(connaddr->conn));
	  
	  if (rc == 0)
		return IPM_DUP;			/* conn. already exist */
	  else if (rc < 0)
		break;					/* find the place */
	  prev_connaddr = connaddr;
	}

  /* allocate and cpy */
  IPM_SHMALLOC_CONN(new_conn)

  if (new_conn == 0)
	  return NFX_ERR;

  new_connaddr = (ipm_shmconn_t*)IPM_SHMGET_ADDR(shmaddr, new_conn);
  memcpy(&(new_connaddr->conn), connp, sizeof(ipm_conn_t));

  if (prev_connaddr == NULL)	/* inserting element is the first */
	{
	  new_connaddr->next = *connroot;
	  *connroot = new_conn;
	}
  else							/* insert in the middle */
	{
	  new_connaddr->next = prev_connaddr->next;
	  prev_connaddr->next = new_conn;
	}

  return NFX_OK;
} 


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DelConn
 *
 * Description: delete conn. from UPD/TCP conn table. 
 *
 * *connroot: the offset to start of shmaddr that represents
 *			  the said conn. table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_DelConn(ipm_conn_t	*connp, 
			 uint32		*connroot, 
			 cmp_func	cmp)
{
  ipm_shmconn_t *connaddr = 0, *prev_connaddr=0;
  uint32		next;
  int			rc;

  /* find the node to delete */
  for (next=*connroot; next != 0;/* end of table */
	   next = connaddr->next)
	{
	  connaddr = (ipm_shmconn_t*)IPM_SHMGET_ADDR(shmaddr, next);
	  rc = (*cmp)(connp, &(connaddr->conn));
	  
	  if (rc == 0)				/* node found */
		{
		  if (prev_connaddr == NULL) /* delete first node */ 
			*connroot = connaddr->next;
		  else					/* in the midle */
			prev_connaddr->next = connaddr->next; 

		  IPM_SHMFREE_CONN(next)
		  return NFX_OK;
		}
	  else if (rc < 0)			/* list in ascending order */
		break;

	  prev_connaddr = connaddr;
	}

  return NFX_ERR;
}

