/*
 *-----------------------------------------------------------------
 *
 * Name:	prof.c
 *
 * RCS:  $Id: prof.c,v 1.14 2002/12/03 21:59:02 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	System profiling
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	11/07/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/debug.h>
#include <exports/malloc-api.h>
#include <exports/esm-api.h>
#include <exports/file-api.h>
#include <exports/bsp-api.h>
#include "prof-thread.h"
#include "prof.h"
#include "gmon.h"

#ifdef NFX_PROFILING

char *prof_fileName = "vol1:gmon.out";
#define SAMPLE_SIZE (2500*1024)

static uint32* samples CPU_PRIVATE;

int32 prof_samplesIgnored = 0;
//int32 prof_hz = 10000;  // 100 usecs
//int32 prof_usecs = 100; // 100 usecs
int32 prof_hz = 100;
int32 prof_usecs = 10000;

#define PROF_FILENAME_MAX   128

void prof_int(uint32 sr, uint32 cause, address_t epc);

static int prof_timer_starts;
#if defined(NFX_SMP) && !defined(TIMER_TIC_INT)
static nfx_spinlock_t prof_timer_spin;
#endif

void
prof_init_common(void)
{
	NFX_SPIN_INIT(&prof_timer_spin, SPINLOCK_LEVEL_UNKNOWN);
	prof_timer_starts = 0;
}


void
prof_init()
{
    struct gmonparam    *p = &_gmonparam;
    uint32              textsize;
    extern uchar8       _start[], etext[];

    textsize = etext - _start;
    printf("Initializing profiler _start@%p _end@%p textsize %d.\n",
            _start, etext, textsize);

    /*
     * The samples array is as big as the text segment: the
     * PC is rounded to the nearest 4 bytes and represents a
     * 4 byte counter in the samples array.
     */
    samples = eee_ramAlloc(textsize);
    NFX_ASSERT(samples != NULL);
    bzero(samples, textsize);

    /*
     * Frompc is rounded to the nearest 4 bytes; so the froms array
     * has textsize/4 elements.
     */
    p->fromssize = (textsize / sizeof (uint32)) * sizeof *p->froms;
    p->froms = (void *) eee_ramAlloc(p->fromssize);
    bzero(p->froms, p->fromssize);

    /*
     * The heuristic in the BSD code is to allocate 50% of the text
     * space to arcs. It maxes out at 65536 arcs, though, since shorts
     * are used to address the arcs.
     *
     * XXX: move this magic number to a header file.
     */
    p->tossize = textsize / 2;
    if (p->tossize / sizeof *p->tos > 65536)
    	p->tossize = 65536 * sizeof *p->tos;
    p->tos = eee_ramAlloc(p->tossize);
    NFX_ASSERT(p->tos != NULL);
    bzero(p->tos, p->tossize);
    p->tolimit = p->tossize / sizeof *p->tos;

    p->lowpc = (uint32) &_start;
    p->highpc = (uint32) &etext;
    p->textsize = textsize;
    
    p->state = GMON_PROF_OFF;
}

/**
 * Increment the sample counter at the interrupt
 * tic interval.
 */
void
prof_sample(uint32 interrupt_pc)
{
    struct gmonparam *p = &_gmonparam;
    uint32 offset;
    
#ifndef TIMER_TIC_INT
	timer_int_clear(BSP_TIMER_PROFILER);
#endif

    if (p->state != GMON_PROF_ON)
    	return;

    if (interrupt_pc < p->lowpc ||
        interrupt_pc >= p->highpc)
    {
        prof_samplesIgnored++;
    	return;
    }

    offset = interrupt_pc - p->lowpc;
    if (!(offset < p->textsize))
    {
        printf("interrupt_pc %x lowpc %x offset %x  p->textsize %x\n",
            interrupt_pc, p->lowpc, offset, p->textsize);
    }
    NFX_ASSERT(offset < p->textsize);
    NFX_ASSERT((offset & 3) == 0);
    NFX_ASSERT(offset != 0);

    offset /= 4;
    if (offset > SAMPLE_SIZE - 1)
    {
        prof_samplesIgnored++;
    	return;
    }
        
    samples[offset]++;
}


/**
 * Stop profiling.
 */
int
prof_stop(int argc, char **argv)
{
    struct gmonparam *p = &_gmonparam;

	if (p->state != GMON_PROF_OFF) {
#ifndef TIMER_TIC_INT
		NFX_SPIN_LOCK(&prof_timer_spin);
		if (--prof_timer_starts == 0) {
			timer_stop(BSP_TIMER_PROFILER);
		}
		NFX_SPIN_UNLOCK(&prof_timer_spin);
#endif

		p->state = GMON_PROF_OFF;
	}
    return NFX_OK;
}


/**
 * Start profiling.
 */
int
prof_start(int argc, char **argv)
{
    struct gmonparam *p = &_gmonparam;

	if (p->state != GMON_PROF_ON) {
		uint32 sr;

		bzero(samples, p->textsize);
		bzero(p->froms, p->fromssize);
		bzero(p->tos, p->tossize);
		p->state = GMON_PROF_ON;

#ifndef TIMER_TIC_INT
		NFX_SPIN_LOCK(&prof_timer_spin);
		if (++prof_timer_starts == 1) {
			timer_start_us(BSP_TIMER_PROFILER, prof_usecs, prof_int);
		}
		sr = _mips_mfc0(C0_SR);
		sr |= (SR_IM0 << 5);
		_mips_mtc0(C0_SR, sr);
		NFX_SPIN_UNLOCK(&prof_timer_spin);
#endif
	}

    return NFX_OK;
}


struct prof_dump_cb
{
	char *pd_filename;
	struct gmonparam pd_param;
	void *pd_samples;
};

/**
 * Stop profiling and dump statistics to a gmon.out file.
 */
void
do_prof_dump(struct prof_dump_cb *cb)
{
    struct gmonparam *p = &cb->pd_param;
    struct gmonhdr gmonhdr;
    int i, fd, rc;

    gmonhdr.lpc = p->lowpc;
    gmonhdr.hpc = p->highpc;
    gmonhdr.ncnt = p->textsize + sizeof gmonhdr;
    gmonhdr.version = GMONVERSION;
    gmonhdr.profrate = prof_hz;
    
    printf("Writing profiler data to %s, this may take some time.\n",
            cb->pd_filename);

    fd = file_create(cb->pd_filename, 0644);
    if (fd < 0)
    {
    	printf("do_prof_dump: couldn't create %s\n", cb->pd_filename);
    	return;
    }

    rc = file_write(fd, (void *) &gmonhdr, sizeof gmonhdr);
    if (rc != sizeof gmonhdr)
    {
        printf("do_prof_dump: short write hdr (%ld/%d)\n",
            (unsigned long)(sizeof gmonhdr), rc);
        file_close(fd);
        return;
    }
    rc = file_write(fd, (void *) cb->pd_samples, p->textsize);
    if (rc != p->textsize)
    {
        printf("do_prof_dump: short write samples (%d/%d)\n",
                p->textsize, rc);
        file_close(fd);
        return;
    }

    for (i = 0; i < p->textsize / sizeof (uint32); i++)
    {
    	if (p->froms[i] != 0)
        {
    	    struct tostruct *top = &p->tos[p->froms[i]];
    	    struct rawarc arc;
    	    
    	    arc.raw_frompc = p->lowpc + (i * sizeof (uint32));
    	    arc.raw_selfpc = top->selfpc;
    	    arc.raw_count = top->count;

    	    rc = file_write(fd, (void *) &arc, sizeof arc);
            if (rc != sizeof arc)
            {
                printf("do_prof_dump: short write arc (%ld/%d)\n",
                        (unsigned long)(sizeof arc), rc);
                file_close(fd);
                return;
            }
    	}
    }
    
    file_close(fd);

    printf("Profiler data created at %s.\n",
            cb->pd_filename);
}


char *
strdup(char *str)
{
	char *newstr = eee_ramAlloc(strlen(str));
	if (newstr)
		strcpy(newstr, str);
	return newstr;
}

/**
 * Stop profiling and dump statistics to a gmon.out file.
 */
int
prof_dump(int argc, char **argv)
{
    esm_event_t     event;
    esm_handle_t    esmHandle;
	struct prof_dump_cb *cb;

    prof_stop(0, NULL);

	cb = eee_ramAlloc(sizeof(*cb));
	cb->pd_param = _gmonparam;
	cb->pd_samples = samples;

    if (argc > 1)
    {
        if (strlen(argv[1]) > PROF_FILENAME_MAX)
        {
            printf("prof_dump: fileName %s too long\n", argv[1]);
            return NFX_ERR;
        }
		cb->pd_filename = strdup(argv[1]);
    }
	else {
		cb->pd_filename = strdup(prof_fileName);
	}

    if(prof_thread_sm_create(&esmHandle, cb) != NFX_OK)
    {
        printf("prof_dump: error creating writer thread\n");
        return NFX_ERR;
    }

    event.event = PROF_THREAD_EVENT_START;
    if (esm_deliverEvent(esmHandle, &event) != NFX_OK)
    {
        printf("prof_dump: error starting writer thread\n");
        return NFX_ERR;
    }

    return NFX_OK;
}

/**
 * profiling interrupt
 */
void
prof_int(uint32 sr, uint32 cause, address_t epc)
{
	prof_sample((uint32)epc);
}

#else /* NFX_PROFILING */
void
prof_init()
{
}
#endif /* NFX_PROFILING */

/**
 * Thread create action function.  We must have this, but
 * we needn't do anything here.
 */
int32
prof_thread_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    return nextState;

}

/**
 * Another must have for the thread.
 */
int32
prof_thread_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         thisEvent,
    void                *userInfo)
{
    panic("prof_thread bad state\n");
    return nextState;
}

/**
 * Start action function for our dump thread. 
 * Do the work of writing out the file.  We need a
 * thread to interface with the file library.
 */
int32
prof_thread_start(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
#if defined(NFX_PROFILING)
    do_prof_dump(smCookie);
#endif
    return nextState;
}


