/*
 *-----------------------------------------------------------------
 *
 * Name:	from-bsd.c
 *
 * RCS:  $Id: from-bsd.c,v 1.5 2002/12/04 20:27:39 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	 Flash ROM routines for BSD kernel
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/10/2002
 *
 *-----------------------------------------------------------------
 */


#include "from-bsd.h"
#include <machine/ds1511.h>
#include <sys/conf.h>
#include <sys/types.h>
#include <sys/malloc.h>
#include <nfx-types.h>
#include "from-api.h"
#include <sys/amd29f040.h>
#include "prom-part-api.h"

#undef DEBUG
#define DEBUG
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif


static int prom_match(struct device *parent, void *cfdata, void *aux);
static void prom_attach(struct device *parent, struct device *self, void *aux);

struct prom_sc
{
	struct device dev;
	int opened;
};


struct cfattach prom_ca = {
	sizeof(struct prom_sc), prom_match, prom_attach
};


struct cfdriver prom_cd = {
	0, "prom", DV_DULL, 0, 0
};


cdev_decl(prom);


void
watchdog_stop(void)
{
	ds1511_disable_watchdog();
}


void
watchdog_restart(void)
{
	ds1511_enable_watchdog();
}


static int
prom_match(struct device *parent, void *cfdata, void *aux)
{
	struct confargs *ca = (struct confargs *)aux;
	static int from_inited;
	static int ndevs;
	if (!from_inited) {
		from_init();
		from_inited = 1;
	}
	return (strcmp(ca->ca_name, prom_cd.cd_name) == 0) &&
		from_device_present(ndevs++ == 0 ?
							FROM_CBX_PRIMARY : FROM_CBX_ALTERNATE);
}


static void
prom_attach(struct device *parent, struct device *self, void *aux)
{
	printf("\n");
}


static int
unit_prom(int unit)
{
	if (0 <= unit && unit <= 1) {
		if (unit == 1 && primary_is_rprom)
			return -1;
		return FROM_CBX_PRIMARY;
	}
	else if (2 <= unit && unit <= 3) {
		if (unit == 3 && !primary_is_rprom)
			return -1;
		return FROM_CBX_ALTERNATE;
	}
	else if (unit == 4) {
		if (from_device_present(FROM_CBX_ALTERNATE)) {
			return primary_is_rprom ? FROM_CBX_ALTERNATE : FROM_CBX_PRIMARY;
		}
		else
			return -1;
	}
	else {
		return -1;
	}
}


static int
unit_part_no(int unit)
{
	if (0 <= unit && unit <= 1)
		return unit;
	else if (2 <= unit && unit <= 3)
		return unit - 2;
	else
		return 0;
}


static struct prom_part_desc *
unit_part_desc(int unit)
{
	return &prom_get_layout(unit_prom(unit))[unit_part_no(unit)];
}


static int unit_offset(int unit)
{
	if (unit == 4) {
		if (primary_is_rprom) {
			return prom_get_layout(FROM_CBX_ALTERNATE)[0].part_offs;
		}
		else {
			return prom_get_layout(FROM_CBX_PRIMARY)[0].part_offs;
		}
	}
	else 
		return unit_part_desc(unit)->part_offs;
}


static int unit_size(int unit)
{
	if (unit == 4) {
		return 2 * 1024 * 1024;
	}
	else 
		return unit_part_desc(unit)->part_size;
}


int
promopen(dev_t dev, int flag, int mode, struct proc *p)
{
	int rc = ENXIO;
	int unit = minor(dev);
	int prom = unit_prom(unit);
	if (prom != -1) {
		debug(("promopen: %d %d\n", unit, prom));
		if (prom < prom_cd.cd_ndevs) {
			struct prom_sc *sc = prom_cd.cd_devs[prom];
			if (sc->opened) {
				rc = EBUSY;
			}
			else {
				sc->opened = 1;
				rc = 0;
				from_set_current(prom);
			}
		}
	}
	return rc;
}


int
promclose(dev_t dev, int flag, int mode, struct proc *p)
{
	int unit = minor(dev);
	int prom = unit_prom(unit);
	if (prom != -1) {
		struct prom_sc *sc = prom_cd.cd_devs[prom];
		debug(("promclose: %d %d\n", unit, prom));
		if (sc->opened) {
			sc->opened = 0;
		}
	}
	return 0;
}


static size_t
uio_cnt(struct uio *uio)
{
	int i;
	size_t sum = 0;
	for (i = 0; i < uio->uio_iovcnt; ++i) {
		sum += uio->uio_iov[i].iov_len;
	}
	return sum;
}


static int
valid_uio(struct uio *uio, int unit)
{
	int len = uio_cnt(uio);
	return uio->uio_offset % FROM_SECTORSIZE == 0
		&& len == FROM_SECTORSIZE
		&& uio->uio_offset + len <= unit_size(unit);
}


int
promwrite(dev_t dev, struct uio *uio, int flag)
{
	int rc = ENXIO;
	int unit = minor(dev);
	int prom = unit_prom(unit);
	if (prom != -1) {
		uchar8 *mem = malloc(FROM_SECTORSIZE, M_DEVBUF, 0);
		debug(("promwrite: %d(sz %d) at %d+%d for %d\n", unit,
			   unit_size(unit),
			   (int)uio->uio_offset, unit_offset(unit),
			   (int)uio_cnt(uio)));
		if (valid_uio(uio, unit)) {
			off_t offset = uio->uio_offset + unit_offset(unit);
			rc = uiomove(mem, FROM_SECTORSIZE, uio);
			if (rc == 0) {
				int s = splhigh();
				rc = from_programbytes(offset, mem, FROM_SECTORSIZE);
				if (rc != 0) {
					debug(("promwrite: from_programbytes failed\n"));
					rc = EIO;
				}
				splx(s);
			}
			else {
				debug(("promwrite: uiomove failed\n"));
			}
		}
		else {
			rc = EINVAL;
			debug(("promwrite: invalid offset or size\n"));
		}
		free(mem, M_DEVBUF);
	}
	return rc;
}


int
promread(dev_t dev, struct uio *uio, int flag)
{
	int rc = ENXIO;
	int unit = minor(dev);
	int prom = unit_prom(unit);
	if (prom != -1) {
		uchar8 *mem = malloc(FROM_SECTORSIZE, M_DEVBUF, 0);
		debug(("promread: %d(sz %d) at %d+%d for %d\n", unit,
			   unit_size(unit),
			   (int)uio->uio_offset, unit_offset(unit), (int)uio_cnt(uio)));
		if (valid_uio(uio, unit)) {
			off_t offset = uio->uio_offset + unit_offset(unit);
			rc = from_readbytes(offset, mem, FROM_SECTORSIZE);
			if (rc != 0) {
				debug(("promread: from_readbytes failed\n"));
				rc = EIO;
			}
			rc = uiomove(mem, FROM_SECTORSIZE, uio);
			if (rc != 0) {
				debug(("promread: uiomove failed\n"));
			}
		}
		else {
			rc = EINVAL;
			debug(("promread: invalid offset or size\n"));
		}
		free(mem, M_DEVBUF);
	}
	return rc;
}


int
promioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p)
{
	int unit = minor(dev);
	int rc = ENXIO;
	if (unit < prom_cd.cd_ndevs) {
		rc = EINVAL;
	}
	return rc;
}


/*
 * translate eee alloc calls from from.c
 */
void *
eee_ramAlloc(int size)
{
    return malloc(size, M_DEVBUF, 0);
}



void
eee_ramDealloc(void *buf)
{
    free(buf, M_DEVBUF);
}
