/*
 *-----------------------------------------------------------------
 *
 * Name:	from.c
 *
 * RCS:  $Id: from.c,v 1.21 2002/11/06 22:46:06 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	 Flash ROM routines
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/8/01
 *
 *-----------------------------------------------------------------
 */

#if defined(__OpenBSD__)
#include "from-bsd.h"
#include <nfx-types.h>
#include <nfx-defs.h>
#else
#include <nfx-incl.h>
#endif

#include "from-api.h"

#if !defined(__OpenBSD__)
#include <exports/malloc-api.h>
#include <sys/mips.h>
#include <exports/timer-api.h>
#endif

#include <exports/hwMap.h>
#include "prom-part-api.h"

#undef DEBUG_PRINT
#ifdef DEBUG_PRINT
#define DPRINT(_x)  printf _x
#else
#define DPRINT(_x)
#endif

int from_current;
int primary_is_rprom;
int from_probed_proms;

nfx_spinlock_t	prom_lock;

/*
 * flash rom device driver control block
 */
typedef struct {
    char *              name;               // description of chip
    TYPE_LEN            mfg_id;             // chip manufacturer Id
    TYPE_LEN            dev_id;             // chip device Id
    volatile TYPE_LEN * base_p;             // sw accessible adress to offset 0
    unsigned int        sector_size;        // sector size in bytes 
    unsigned int        num_sectors;        // number of sectors
    unsigned int        size;               // chip size in bytes
} FROM_CB_T, * FROM_CB_P;

FROM_CB_T from_cb[HW_MAX_PROM_NUM];


#if 0
#define SAFEROUTINE(name,protoargs,callargs)					\
int32 name protoargs											\
{																\
	int32	(*fn) protoargs;									\
	int32	ret;												\
	int32	codesize = (char8 *)name - (char8 *)_##name;		\
																\
	fn = eee_ramAlloc (codesize);								\
	if (!fn)													\
		return FROM_ERR_NOMEM;									\
	memcpy (fn, _##name, codesize);								\
	mips_clean_cache (fn, codesize);							\
																\
	ret = (*fn) callargs;										\
	eee_ramDealloc(fn);											\
	return ret;													\
}
#else
#define SAFEROUTINE(name,protoargs,callargs)					\
extern int32 name protoargs;									\
int32 name protoargs											\
{																\
	int32	ret;												\
	ret = _##name callargs;										\
	return ret;													\
}
#endif

#define PAUSE	{int32 _i; for (_i = 0; _i < 0x100; _i++); }

static void
fromunlock (volatile uchar8 *sec_base)
{
	*(TYPE_LEN *)(sec_base + CMD_UNLOCK_ADDR1) = CMD_DATA1;
	PAUSE;
	*(TYPE_LEN *)(sec_base + CMD_UNLOCK_ADDR2) = CMD_DATA2;
	PAUSE;
}

static void
fromcmd (volatile uchar8 *sec_base, TYPE_LEN cmd)
{
	*(TYPE_LEN *)(sec_base + CMD_UNLOCK_ADDR1) = CMD_DATA1;
	PAUSE;
	*(TYPE_LEN *)(sec_base + CMD_UNLOCK_ADDR2) = CMD_DATA2;
	PAUSE;
	*(TYPE_LEN *)(sec_base + CMD_ADDR3) = cmd;
	PAUSE;
}

static int32
frompoll (volatile TYPE_LEN *fp, TYPE_LEN expect)
{
	TYPE_LEN	poll;

	while (1) {
		poll = *fp;
		if (((poll ^ expect) & DQPOLL) == 0)
			break;
		if ((poll & DQTIMEEXCEEDED) == DQTIMEEXCEEDED) {
			poll = *fp;
			if (((poll ^ expect) & DQPOLL) == 0)
				break;
			*fp = CMD_DATA_RESET;
			(void) *fp;
			return FROM_ERR_FATAL;
		}
	}
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:	from_readId()
 *
 * Description: read manufacture and device id from chip
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/8/01
 *
 *-----------------------------------------------------------------
 */
static int32
_from_readId(TYPE_LEN * man, TYPE_LEN * dev)
{
	volatile TYPE_LEN *sec = FROM_ADDR(volatile TYPE_LEN *, 0);
	int32			prot;

	fromcmd ((uchar8 *)sec, CMD_DATA_AUTOSELECT);

	/* get manufacture and device id */
	if (man)
		*man = sec[0];
	if (dev)
		*dev = sec[1];

	/* get sector protect status */
	prot = *(volatile TYPE_LEN *)((TYPE_LEN *)sec + 2) & 1;

	sec[0] = CMD_DATA_RESET;

	(void) sec[0];

	return prot;
}

SAFEROUTINE(from_readId, (TYPE_LEN * man, TYPE_LEN * dev), (man, dev))

/*-----------------------------------------------------------------
 * Name:	from_erasedevice()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/8/01
 *
 *-----------------------------------------------------------------
 */
static int32
_from_erasedevice (void)
{
	volatile TYPE_LEN *sec = FROM_ADDR(volatile TYPE_LEN *, 0 /*offset*/);

	fromcmd ((uchar8 *)sec, CMD_DATA_ERASE);
	fromcmd ((uchar8 *)sec, CMD_DATA_ERASECHIP);

	return frompoll (sec, ~0);
}

SAFEROUTINE(from_erasedevice, (void), ())

/*-----------------------------------------------------------------
 * Name:	from_erasesector()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/8/01
 *
 *-----------------------------------------------------------------
 */
static int32
_from_erasesector (int32 sec_num)
{
	volatile TYPE_LEN *sp = 
                    FROM_ADDR(volatile TYPE_LEN *,(sec_num*FROM_SECTORSIZE));
	int32	rc;

	if (sec_num >= FROM_SECS)
		return FROM_ERR_FAIL;

	watchdog_stop();

	printf ("erase sector at %p ...", sp);
	fromcmd ((uchar8 *)sp, CMD_DATA_ERASE);
	fromunlock ((uchar8 *)sp);
	*sp = CMD_DATA_ERASESECT;

	rc = frompoll (sp, ~0);

	printf ("done\n");

	watchdog_restart();

	return rc;
}

SAFEROUTINE(from_erasesector, (int32 sec_num), (sec_num))

/*-----------------------------------------------------------------
 * Name:	from_programbytes()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/8/01
 *
 *-----------------------------------------------------------------
 */
/*
 *	We build a vector of these, each representing the data to be
 *	programmed into one sector.
 */
struct flashvec {
	uint32		sec;			/* sector num */
	uint32		offs; 			/* sector offset */
	void		* mem;			/* src pointer */
	uint32		num_bytes;
	uint32		erase;			/* unused */
	int32		scratch;		/* unused */
};
	
uchar8	secdata[FROM_SECTORSIZE];

static int32
_from_reallyprogrambytes (struct flashvec *v, int32 nv)
{
	volatile TYPE_LEN * fp;
	volatile TYPE_LEN * sp, * mp;
	int32			status = NFX_OK;
	uint32			j, o;
	int32			stat;

	while (nv-- != 0) {
		sp = FROM_ADDR(volatile TYPE_LEN *, (v->sec * FROM_SECTORSIZE));

		/*
		 * do we need to do a read and merge?
		 */
		if ((v->offs != 0) || (v->num_bytes != FROM_SECTORSIZE)) {
			volatile uchar8	* dst, * src;
			uint32			i;

			/*
			 * read in current data
			 */
			dst = secdata;
			src = (uchar8 *)sp;
			for (i=0; i<FROM_SECTORSIZE; i++)
				*dst++ = *src++;
			/*
			 * overwrite with new data
			 */
			dst = &(secdata[v->offs]);
			src = v->mem;
			for (i=0; i<v->num_bytes; i++)
				*dst++ = *src++;
			v->mem = secdata;
			v->num_bytes = FROM_SECTORSIZE;
			v->offs = 0;
		}

		/*
		 * erase sector
		 */
        DPRINT(("program/erase sector at %p ...", sp));
		fromcmd ((uchar8 *)sp, CMD_DATA_ERASE);
		fromunlock ((uchar8 *)sp);
		*sp = CMD_DATA_ERASESECT;
		stat = frompoll (sp, ~0);
		if (stat != NFX_OK)
			status = stat;

		/*
		 * program sector
		 */
#if !defined(_KERNEL)		
		DPRINT(("    program sector %d from %06x to %06x ...",
								v->sec, v->offs, v->offs + v->num_bytes - 1));
#endif		

		mp = (TYPE_LEN *)v->mem; 
#ifndef SSC_MGMT
		for (j = v->num_bytes, o = v->offs; j != 0; j--, o++, mp++) {
#else
		for (j = v->num_bytes, o = v->offs; j != 0; j-=2, o++, mp++) {
#endif
			/* otherwise program flash with sequential byte stream */
			fp = (volatile TYPE_LEN *)(sp + o);

			fromcmd ((uchar8 *)sp, CMD_DATA_PROGRAM);
			*fp = *mp;
			stat = frompoll (fp, *mp);
			if (stat != NFX_OK)
				status = stat;
		}
#if !defined(_KERNEL)
		printf ("%d ", v->sec);
#endif

		/* advance to next sector */
		v++;
	}
#if !defined(_KERNEL)
		printf ("done\n");
#endif

#if RAL_XXX
	/*
	 * Reboot?
	 */
	if (flags & FLASHDEV_PROG_REBOOT)
		((volatile void (*)(void))PA_TO_KVA1(BOOTPROM_BASE)) ();
#endif

	return status;
}

SAFEROUTINE(from_reallyprogrambytes,
		(struct flashvec *v, int32 nv),
		(v, nv))

int32
from_programbytes(int32 offset, TYPE_LEN * src_buf, uint32 len)
{
	struct flashvec	*   vecs;
	struct flashvec	*   v;
	uint32			    o, nb;
	int32			    nv = 0;
	int32			    ret;
	TYPE_LEN *          sp = src_buf;


    vecs = (struct flashvec *)eee_ramAlloc(FROM_SECS * sizeof(struct flashvec));
	if (vecs == NULL)
		return FROM_ERR_NOMEM;

    v = vecs;

	watchdog_stop();

#if !defined(_KERNEL)		
    printf("\nflash program 0x%x bytes to %p from %p\nsectors ... ",
            len, FROM_ADDR(void *, (offset)), src_buf);
#endif
	/* check limits */
	if (offset+len > FROM_SIZE) {
		ret = FROM_ERR_FAIL;
        goto exit;
	}

	if (!src_buf || len == 0) {
		ret = NFX_OK;
        goto exit;
    }

	/* construct the flash programming vector */
	for (o = offset, nb = len; nb != 0;) {
		uint32			sec_offset, sec_num;
		uint32			n;

		/* find starting sector */
		sec_offset = FROM_SECTOR_OFFSET(o);
		sec_num = FROM_SECTOR_NUM(o);

		/* how many bytes do we want to program in this sector */
		n = FROM_SECTORSIZE - sec_offset;
		if (n > nb)
			n = nb;

		/* add to programming vector */
		v->sec = sec_num;
		v->offs = sec_offset;
		v->num_bytes = n;
		v->mem = sp;
		v->scratch = 0;
		v->erase = 0;

        DPRINT(("program vector, sec = %d, offs = %d, mem = %p, len = %d\n",
                    v->sec, v->offs, v->mem, v->num_bytes));
		/* advance to next sector */
		v++;
		nv++;
		(uchar8 *)sp += n;
		o += n;
		nb -= n;
	}

#if 0
	/* wipe this area of the flash out of the data cache */
	mips_clean_dcache (PA_TO_KVA0(FROM_BASE + offset), len);
#endif

	/* now really do the programming operation */
	ret = from_reallyprogrambytes (vecs, nv);

exit:
	watchdog_restart();
    eee_ramDealloc(vecs);

	return ret;
}

/*-----------------------------------------------------------------
 * Name:	from_readbytes()
 *
 * Description:	Read bytes from Flash ROM to user supplied buffer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	10/8/01
 *
 *-----------------------------------------------------------------
 */
int32
from_readbytes(int32 offset, uchar8 * dest_buf, uint32 len)
{
	volatile uchar8	* sp;
	int32			i;

	/* check limits */
	if (offset+len > FROM_SIZE)
		return FROM_ERR_FAIL;

	sp = FROM_ADDR(volatile uchar8 *, offset);

	/* copy bytes to dest_buf array */
	for (i=0; i<len; i++)
		dest_buf[i] = sp[i];

	return NFX_OK;
}


/*-----------------------------------------------------------------
 * Name :	from_device_present()
 *
 * Description: Return true if the flash device is present
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	11/05/02
 *
 *-----------------------------------------------------------------
 */
int
from_device_present(int id)
{
	return from_cb[id].dev_id != 0xff;
}


/*
 * dump the flash device driver info for diag purposes
 */
void
from_print_info(void)
{
    int i;
    FROM_CB_P fcb_p;

    printf("Flash device: Running %s on %s\n",
            (primary_is_rprom)?"Recov Prom":"Std Prom",
#if   R7000
            (isOldFC())?"OldFC":"NewFC"
#elif SIBYTE
            (isCheetah())?"Cheetah":"Garfield"
#else
            " "
#endif
        );
    printf("from_current=%d\n", from_current);
    for (i=0, fcb_p=from_cb; i < HW_MAX_PROM_NUM; i++, fcb_p++) {
        printf("Device[%d]: %s\n", i, fcb_p->name);
        printf("\tIds: Manufacturer=0x%02x Device=0x%02x\n",
                fcb_p->mfg_id, fcb_p->dev_id);
        printf("\tbase_addr:\t%p\n", fcb_p->base_p);
        printf("\tsectors: num=%d * size=%d (0x%x)\n",
                fcb_p->num_sectors, fcb_p->sector_size, fcb_p->sector_size);
        printf("\tsize=%d (0x%x)\n", fcb_p->size, fcb_p->size);
    }
}

/*
 * number of flash chips on the system
 */
int
from_cfg_proms(void)
{
#if R7000
    if (isOldFC())
        return FROM_SINGLE_PROM_CFG;
    else
        return FROM_RECOV_PROM_CFG;
#elif SIBYTE
    if (isCheetah())
        return FROM_RECOV_PROM_CFG;
    else
        return FROM_SINGLE_PROM_CFG;
#else
    return FROM_SINGLE_PROM_CFG;
#endif
}

/*
 * initialize the flash driver
 */
void
from_init(void)
{
    int i;
    FROM_CB_P fcb_p;

    from_probed_proms = 0;
    for (i=0, fcb_p=from_cb; i < from_cfg_proms(); i++, fcb_p++) {
        if (i == FROM_CBX_PRIMARY)
            fcb_p->base_p = (volatile TYPE_LEN *)(PA_TO_KVA1(HW_PROM_BASE));
        else
            fcb_p->base_p = (volatile TYPE_LEN *)(PA_TO_KVA1(HW_BIG_BOOTROM_BASE));

        from_current = i;       // need to set for from_readId call
        from_readId(&fcb_p->mfg_id, &fcb_p->dev_id);

        switch (fcb_p->dev_id) {
            case DEV_29F040:
                fcb_p->num_sectors = AMD_29F040_MAX_SECTORS;
                fcb_p->name = AMD_29F040_STR;
                from_probed_proms++;
                break;

            case DEV_29LV040:
                fcb_p->num_sectors = AMD_29LV040_MAX_SECTORS;
                fcb_p->name = AMD_29LV040_STR;
                from_probed_proms++;
                break;

            case DEV_29LV017:
                fcb_p->num_sectors = AMD_29LV017_MAX_SECTORS;
                fcb_p->name = AMD_29LV017_STR;
                from_probed_proms++;
                break;

            case DEV_29LV033:
                fcb_p->num_sectors = AMD_29LV033_MAX_SECTORS;
                fcb_p->name = AMD_29LV033_STR;
                from_probed_proms++;
                break;

            case DEV_M29W040:
                fcb_p->num_sectors = ST_M29W040_MAX_SECTORS;
                fcb_p->name = ST_M29W040_STR;
                from_probed_proms++;
                break;
            default:
				fcb_p->dev_id = 0xff;
                break;
        }
        fcb_p->sector_size = FROM_SECTORSIZE;
        fcb_p->size = fcb_p->sector_size * fcb_p->num_sectors;
    }
    from_set_primary();
#ifndef SSC_MGMT
    if ((from_cb[FROM_CBX_PRIMARY].dev_id == DEV_29LV040)||
        (from_cb[FROM_CBX_PRIMARY].dev_id == DEV_M29W040))

        primary_is_rprom = 1;
    else
#endif
    {
        primary_is_rprom = 0;
    }
    prom_part_init();

	NFX_SPIN_INIT(&prom_lock, SPINLOCK_LEVEL_UNKNOWN);
}


address_t
from_get_base(void)
{
    return (address_t)from_cb[from_current].base_p;
}

uint32
from_get_size(void)
{
    return from_cb[from_current].size;
}

void
from_set_primary(void)
{
    from_current = FROM_CBX_PRIMARY;
}

void
from_set_alternate(void)
{
    if (from_probed_proms > 1)
        from_current = FROM_CBX_ALTERNATE;
}

void
from_set_current(int select)
{
    if (from_probed_proms > 1)
        from_current = select;
}

int
from_swap_base(void)
{
    int rv = from_current;

    if (from_probed_proms > 1) {
        if (from_current == FROM_CBX_PRIMARY)
            from_set_alternate();
        else
            from_set_primary();
    }
    return rv;
}

int
from_select_recov(void)
{
    int rv = from_current;

    if (from_probed_proms > 1) {
        if (primary_is_rprom) {
            from_set_primary();
        }
        else {
            from_set_alternate();
        }
    }
    return rv;
}

int
from_select_std(void)
{
    int rv = from_current;

    if (from_probed_proms > 1) {
        if (primary_is_rprom) {
            from_set_alternate();
        }
        else {
            from_set_primary();
        }
    }
    else {
        rv = FROM_SEL_INVALID;
    }
    return rv;
}
