/*
 *-----------------------------------------------------------------
 *
 * Name:	upgrade-app.c
 *
 * RCS:  $Id: upgrade-app.c,v 1.5 2002/08/22 06:05:40 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	 application for upgrading proms from BSD
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/2002
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/eee-api.h>
#include "../ssc-prom-upgrade/prom-upgrade-api.h"
#include "from-api.h"
#include <exports/prom-part-api.h>



#undef DEBUG
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif

typedef struct upgrade_cb
{
    char *  buf;
    char *  cur;
	int     max;
	int     len;
	int     err;
} upgrade_cb_t;


static upgrade_cb_t upgrade_cb;


static enum {
	WAIT_START,
	WAIT_DATA
} state;


/*-----------------------------------------------------------------
 * Name :	do_ack()
 *
 * Description: send acknowledgement message to the sender of the
 * descriptor descr with the message type msgType.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
do_ack(eee_descPtr_t descr, promUpgradeMsg_t msgType)
{
	promUpgradeHdr_t hdr;
	hdr.msgType = msgType;
	eee_sendMessage(&hdr, sizeof(hdr), descr->hdr.src_port, descr->hdr.dest_port);
}


/*-----------------------------------------------------------------
 * Name :	ack_start()
 *
 * Description: send aknowledgement for start message
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
ack_start(eee_descPtr_t descr)
{
	do_ack(descr, promUpgradeStartAck);
}


/*-----------------------------------------------------------------
 * Name :	ack_writeSuccess()
 *
 * Description: send message about success of the program operation
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
ack_writeSuccess(eee_descPtr_t descr)
{
	do_ack(descr, promUpgradeWriteSuccess);
}


/*-----------------------------------------------------------------
 * Name :	ack_writeFailure()
 *
 * Description: send message about failure of the program operation
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
ack_writeFailure(eee_descPtr_t descr)
{
	do_ack(descr, promUpgradeWriteFailure);
}


/*-----------------------------------------------------------------
 * Name :	ack_data()
 *
 * Description: acknowledge receiving a data message
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
ack_data(eee_descPtr_t descr)
{
	do_ack(descr, promUpgradeDataAck);
}


/*-----------------------------------------------------------------
 * Name :	ack_checksum()
 *
 * Description: send message about the result of checksum operation
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
ack_checksum(eee_descPtr_t descr, int msgType)
{
	do_ack(descr, msgType);
}


/*-----------------------------------------------------------------
 * Name :	eee_descrAdjust()
 *
 * Description: trim num_bytes from the head of the descriptor descr.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
eee_descrAdjust(eee_descPtr_t descr, int num_bytes)
{
	NFX_ASSERT(edesc_firstBufferIdx(descr) == 1);
	NFX_ASSERT(descr->bd[1].len > num_bytes);
	descr->bd[1].buf += num_bytes;
	descr->bd[1].len -= num_bytes;
	EEE_SET_PKT_LEN(descr, EEE_GET_PKT_LEN(descr) - num_bytes);
}



/*-----------------------------------------------------------------
 * Name :	do_start()
 *
 * Description: start the upgrade sequence
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
do_start(eee_descPtr_t descr, upgrade_cb_t *cb)
{
	debug(("> do_start\n"));
	ack_start(descr);
	eee_freePacket(descr);

	cb->buf = prom_upgrade_area;
    cb->max = PROM_UPGRADE_MAX_SIZE;
    cb->cur = cb->buf;
    cb->len = 0;
    cb->err = 0;
	debug(("< do_start\n"));
}


/*-----------------------------------------------------------------
 * Name :	add_data()
 *
 * Description: add a data packet
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
add_data(eee_descPtr_t descr, upgrade_cb_t *cb)
{
    int rv;
    int len;

	debug(("> add_data: len=%ld\n",
           EEE_GET_PKT_LEN(descr) - sizeof(promUpgradeHdr_t)));

	eee_descrAdjust(descr, sizeof(promUpgradeHdr_t));
	
    len = EEE_GET_PKT_LEN(descr);
    if ((len + cb->len) > cb->max) {
        debug(("err data over cb_max %d\n", cb->max));
        cb->err++;
    }
    if (cb->err) {
        goto add_data_exit;
    }
    rv = eee_copyPacketToLocalBuffer(cb->cur, descr, len, 0);
    if (rv == len) {
        cb->cur += len;
        cb->len += len;
	}
add_data_exit:
	ack_data(descr);
	eee_freePacket(descr);
	debug(("< add data cb_len %d curbuf %p\n", cb->len, cb->cur));
}


/*-----------------------------------------------------------------
 * Name :	do_write_prom()
 *
 * Description: program the prom
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
do_write_prom(upgrade_cb_t *cb)
{
	int error;

	NFX_SPIN_LOCK(&prom_lock);
    error = prom_part_upgrade(cb->buf, cb->len);
	NFX_SPIN_UNLOCK(&prom_lock);
    if (error != NFX_OK) {
        debug(("do_write_prom: prom_part_upgrade failed\n"));
    }
	return error;
}


/*-----------------------------------------------------------------
 * Name :	write_prom()
 *
 * Description: program the prom
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static
void write_prom(eee_descPtr_t descr, upgrade_cb_t *cb)
{
	debug(("> write_prom err %d len %d\n", cb->err, cb->len));
	if (cb->err == 0 && do_write_prom(cb) == NFX_OK) {
		ack_writeSuccess(descr);
	}
	else {
		ack_writeFailure(descr);
	}
	eee_freePacket(descr);
	cb->buf = NULL;
	debug(("< write_prom\n"));
}


/*-----------------------------------------------------------------
 * Name :	checksum()
 *
 * Description: checksum the data that has been received so far
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static
void checksum(eee_descPtr_t cksum_descr, upgrade_cb_t *cb)
{
	promUpgradeChecksumMsg_t *msg = (promUpgradeChecksumMsg_t *)
		cksum_descr->bd[edesc_firstBufferIdx(cksum_descr)].buf;
    uchar8 our_cksum;

	debug(("> checksum: %d total_len %d\n", msg->cksum, msg->total_len));
	NFX_ASSERT(cb->buf != NULL);

    /* 
     * possibly adjust the len as the last packet may have
     * been padded up to 23 bytes
     */

    if (msg->msg_verif == PROM_UPGRADE_MSG_VERIF(msg->cksum, msg->total_len)) {
        if (msg->total_len != cb->len) {
            if ((cb->len > msg->total_len) &&
                ((cb->len - msg->total_len) < PROM_UPGRADE_MIN_DATA_SIZE))
                cb->len = msg->total_len;
            else
                cb->err++;
        }
    }
    else {
        // the old upgrade dont have msg_verif size should be 512K
        if (cb->len != 0x80000) {
            cb->err++;
            debug(("msg_verif %08x dont match %08d\n", msg->msg_verif,
                        PROM_UPGRADE_MSG_VERIF(msg->cksum, msg->total_len)));
        }
    }

    /*
     * When supporting the 'old' prom format is no longer an issue
     * get rid of the checksum (though keeping the total size)
     * as there is much stronger crc32 over the data and partition
     * control block header.
     */
    our_cksum = 0;
    if (cb->err == 0) {
        int i;

        for (i = 0; i < cb->len; ++i) {
            our_cksum ^= cb->buf[i];
        }
        ack_checksum(cksum_descr, (our_cksum == msg->cksum) ?
                     promUpgradeChecksumSuccess : promUpgradeChecksumFail);
    }
    else {
        ack_checksum(cksum_descr, promUpgradeChecksumFail);
    }

	eee_freePacket(cksum_descr);
	debug(("< checksum: total_processed = %d our_cksum = %d\n",
           cb->len, our_cksum));
}


/*-----------------------------------------------------------------
 * Name :	receive_upgrade_message()
 *
 * Description: application receive function
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static void
receive_upgrade_message(eee_descPtr_t descr)
{
	promUpgradeHdr_t *hdr = (promUpgradeHdr_t *)descr->bd[edesc_firstBufferIdx(descr)].buf;
	switch (hdr->msgType) {
	case promUpgradeStart:
		do_start(descr, &upgrade_cb);
		state = WAIT_DATA;
		break;
	case promUpgradeData:
		if (state == WAIT_DATA) {
			add_data(descr, &upgrade_cb);
		}
		break;
	case promUpgradeChecksum:
		if (state == WAIT_DATA) {
			checksum(descr, &upgrade_cb);
		}
		break;
	case promUpgradePerformWrite:
		if (state == WAIT_DATA) {
			write_prom(descr, &upgrade_cb);
			state = WAIT_START;
		}
		break;
	default:
		debug(("receive_upgrade_message: Invalid message %d\n", hdr->msgType));
		break;
	}
}


/*-----------------------------------------------------------------
 * Name :	upgrade_init()
 *
 * Description: module initialization function
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
void
upgrade_init(void)
{
	int app_id = eee_getApplicationId("prom-upgrade");
	int rc = eee_registerReceiveFunc(app_id, receive_upgrade_message, NULL);
	debug(("> upgrade_init: register for app_id = %d\n", app_id));
	if (rc != NFX_OK) {
		panic("upgrade_init: can not register prom upgrade application\n");
	}
	debug(("< upgrade_init\n"));
}

