// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/**
 * Test code for doubly-linked list primitives.
 *
 * Paul Haahr <haahr@agilestorage.com>
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>

#include <nfx-types.h>
#include <nfx-defs.h>
#include <exports/queue-api.h>

#define assert VERIFY



//  B A S I C   D E F I N I T I O N S

#define LOOPS        10000
#define PERMUTATIONS 7

typedef struct Element Element;
typedef DLL_HDR(Element) Header;

struct Element {
    int payload;
    DLL_ENTRY(Element) list1;
    DLL_ENTRY(Element) list2;
};



//  U T I L I T I E S


#define ZERO(x)         bzero(x, sizeof *x)
#define NEW(type)       ((type *) alloc(sizeof (type)))
#define P(p)            printf("%s = %p\n", #p, p)


/**
 * Allocate zero-initialized memory.
 */
void *alloc(int size)
{
    void *p = malloc(size);
    assert(p != NULL);
    bzero(p, size);
    return p;
}


/**
 * Return true if the specified integer is odd.
 */
boolean isOdd(int n)
{
    return (n & 1) != 0;
}



//  T E S T S


/**
 * Simple insert and remove tests on a single element list.
 */
void test1(void)
{
    Element e;
    Element *p;
    Header header;

    DLL_INIT(&header);
    assert(DLL_IS_EMPTY(&header));

    ZERO(&e);
    assert(DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == NULL);
    
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == NULL);

    DLL_INSERT_HEAD(&header, &e, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));

    DLL_INSERT_TAIL(&header, &e, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));

    DLL_INSERT_HEAD(&header, &e, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));

    DLL_REMOVE(&header, &e, list1);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e, list1));
    assert(DLL_IS_ISOLATED(&e, list2));
}


/**
 * Insert and remove tests on two element lists.
 */
void test2(void)
{
    Element e1;
    Element e2;
    Element *p;
    Header header;

    DLL_INIT(&header);
    assert(DLL_IS_EMPTY(&header));

    ZERO(&e1);
    ZERO(&e2);
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));


    // head, head, head, head

    DLL_INSERT_HEAD(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // head, head, head, tail

    DLL_INSERT_HEAD(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    // head, head, e2, head

    DLL_INSERT_HEAD(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // head, head, e1, tail

    DLL_INSERT_HEAD(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));


    // head, tail, head, head

    DLL_INSERT_HEAD(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // head, tail, head, tail

    DLL_INSERT_HEAD(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    // head, tail, e2, head

    DLL_INSERT_HEAD(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // head, tail, e1, tail

    DLL_INSERT_HEAD(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));


    // tail, head, head, head

    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // tail, head, head, tail

    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    // tail, head, e2, head

    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // tail, head, e1, tail

    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_HEAD(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));


    // tail, tail, head, head

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // tail, tail, head, tail

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    // tail, tail, e2, head

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
    
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    // tail, tail, e1, tail

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(!DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE(&header, &e1, list1);
    assert(!DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(!DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));

    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));
    assert(DLL_IS_ISOLATED(&e1, list1));
    assert(DLL_IS_ISOLATED(&e2, list1));
    assert(DLL_IS_ISOLATED(&e1, list2));
    assert(DLL_IS_ISOLATED(&e2, list2));
}


/**
 * Check that the list, along the list1 chain, is well formed.
 */
void checkList1(Header *header)
{
    Element *p;
    
    DLL_ITERATE(p, header, list1) {
        if (p == header->head) {
            assert(p->list1.prev == DLL_NIL);
        }
        else {
            assert(p->list1.prev->list1.next == p);
        }
        if (p == header->tail) {
            assert(p->list1.next == DLL_NIL);
        }
        else {
            assert(p->list1.next->list1.prev == p);
        }
    }
}


/**
 * Test operations on three element lists, mostly moving to head and tail.
 */
void test3(void)
{
    Element e1;
    Element e2;
    Element e3;
    Element *p;
    Header header;

    DLL_INIT(&header);
    ZERO(&e1);
    ZERO(&e2);
    ZERO(&e3);

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));

    // single move-to-head tests

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_HEAD(&header, &e1, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_HEAD(&header, &e1, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_HEAD(&header, &e2, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_HEAD(&header, &e2, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_HEAD(&header, &e3, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_HEAD(&header, &e3, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    assert(DLL_IS_EMPTY(&header));

    // single move-to-tail tests

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_TAIL(&header, &e1, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_TAIL(&header, &e1, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_TAIL(&header, &e2, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_TAIL(&header, &e2, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_TAIL(&header, &e3, list1);
    checkList1(&header);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e1);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_HEAD(&header, p, list1);
    assert(p == &e3);
    assert(DLL_IS_EMPTY(&header));

    DLL_INSERT_TAIL(&header, &e1, list1);
    DLL_INSERT_TAIL(&header, &e2, list1);
    DLL_INSERT_TAIL(&header, &e3, list1);
    DLL_MOVE_TO_TAIL(&header, &e3, list1);
    checkList1(&header);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e3);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e2);
    DLL_REMOVE_TAIL(&header, p, list1);
    assert(p == &e1);
    assert(DLL_IS_EMPTY(&header));
}


/**
 * Iteration tests, more operations on longer lists, and operations using
 * both list chains.
 */
void testIterate(void)
{
    int i;
    Element *p;
    Element *prev;
    Element *next;
    Header header1;
    Header header2;

    DLL_INIT(&header1);
    DLL_INIT(&header2);

    // iterate over an empty list
    DLL_ITERATE(p, &header1, list1) {
        assert(FALSE);
    }

    // insert in-order on list1, reverse order on list2

    for (i = 0; i < LOOPS; i++) {
        p = NEW(Element);
        p->payload = i;
        DLL_INSERT_TAIL(&header1, p, list1);
        DLL_INSERT_HEAD(&header2, p, list2);
    }

    // check that list1 is in order

    i = 0;
    DLL_ITERATE(p, &header1, list1) {
        assert(p->payload == i);
        assert(!DLL_IS_ISOLATED(p, list1));
        assert(!DLL_IS_ISOLATED(p, list2));
        i++;
    }
    assert(i == LOOPS);

    // check that list2 is in reverse order

    i = 0;
    DLL_ITERATE_BACK(p, &header2, list2) {
        assert(p->payload == i);
        assert(!DLL_IS_ISOLATED(p, list1));
        assert(!DLL_IS_ISOLATED(p, list2));
        i++;
    }
    assert(i == LOOPS);

    // traverse list1, removing odd elements from list2

    DLL_ITERATE(p, &header1, list1) {
        if (isOdd(p->payload)) {
            DLL_REMOVE(&header2, p, list2);
        }
    }

    // check that list2 has the even elements

    i = 0;
    DLL_ITERATE_BACK(p, &header2, list2) {
        assert(p->payload == i);
        assert(!DLL_IS_ISOLATED(p, list1));
        assert(!DLL_IS_ISOLATED(p, list2));
        i += 2;
    }
    assert(i == LOOPS);

    // iterate along list1, copying the odd elements back to list2

    prev = NULL;
    DLL_ITERATE(p, &header1, list1) {
        if (isOdd(p->payload)) {
            assert(prev != NULL);
            assert(DLL_IS_ISOLATED(p, list2));
            DLL_INSERT_BEFORE(&header2, prev, p, list2);
        }
        else {
            assert(!DLL_IS_ISOLATED(p, list2));
            prev = p;
        }
    }

    // check that list2 is in reverse order

    i = 0;
    DLL_ITERATE_BACK(p, &header2, list2) {
        assert(p->payload == i);
        assert(!DLL_IS_ISOLATED(p, list1));
        assert(!DLL_IS_ISOLATED(p, list2));
        i++;
    }
    assert(i == LOOPS);

    // iterate along list2, moving everything to the head (reverse the list)

    DLL_ITERATE_NEXT(p, next, &header2, list2) {
        DLL_MOVE_TO_HEAD(&header2, p, list2);
    }

    // check that list2 is in forward order

    i = 0;
    DLL_ITERATE(p, &header2, list2) {
        assert(p->payload == i);
        assert(!DLL_IS_ISOLATED(p, list1));
        assert(!DLL_IS_ISOLATED(p, list2));
        i++;
    }
    assert(i == LOOPS);

    // iterate along list1 backwards, moving everything to the tail of list2,
    // thus leaving it in reverse order

    DLL_ITERATE_BACK(p, &header1, list1) {
        DLL_MOVE_TO_TAIL(&header2, p, list2);
    }

    // check that list2 is in reverse order

    i = 0;
    DLL_ITERATE_BACK(p, &header2, list2) {
        assert(p->payload == i);
        assert(!DLL_IS_ISOLATED(p, list1));
        assert(!DLL_IS_ISOLATED(p, list2));
        i++;
    }
    assert(i == LOOPS);
}


/**
 * Run a test, with elements permuted in the specified order.
 */
void runPermutedTest(int n, Element **elements, Element **permuted)
{
    int i;

    Element *p;
    Header header1;
    Header header2;

    DLL_INIT(&header1);
    DLL_INIT(&header2);
    
#if 0
    for (i = 0; i < n; i++) {
        printf(" %d", elements[i]->payload);
    }
    printf(" -");
    for (i = 0; i < n; i++) {
        printf(" %d", permuted[i]->payload);
    }
    printf("\n");
#endif

    for (i = 0; i < n; i++) {
        assert(DLL_IS_ISOLATED(elements[i], list1));
        assert(DLL_IS_ISOLATED(elements[i], list2));
    }

    for (i = 0; i < n; i++) {
        DLL_INSERT_HEAD(&header1, elements[i], list1);
        DLL_INSERT_HEAD(&header2, permuted[i], list2);
    }
    
    for (i = 0; i < n; i++) {
        assert(!DLL_IS_ISOLATED(elements[i], list1));
        assert(!DLL_IS_ISOLATED(elements[i], list2));
    }

    DLL_ITERATE(p, &header1, list1) {
        DLL_REMOVE(&header2, p, list2);
    }

    for (i = 0; i < n; i++) {
        assert(!DLL_IS_ISOLATED(elements[i], list1));
        assert(DLL_IS_ISOLATED(elements[i], list2));
    }

    for (i = 1; i < n; i++) {
        DLL_REMOVE(&header1, permuted[i], list1);
        DLL_INSERT_AFTER(&header1, permuted[i - 1], permuted[i], list1);
    }
    for (i = 0; i < n; i++) {
        DLL_REMOVE_HEAD(&header1, p, list1);
        assert(p == permuted[i]);
    }

    for (i = 0; i < n; i++) {
        DLL_INSERT_TAIL(&header1, elements[i], list1);
    }
    for (i = n - 2; i >= 0; i--) {
        DLL_REMOVE(&header1, permuted[i], list1);
        DLL_INSERT_BEFORE(&header1, permuted[i + 1], permuted[i], list1);
    }
    for (i = 0; i < n; i++) {
        DLL_REMOVE_HEAD(&header1, p, list1);
        assert(p == permuted[i]);
    }

    // WARNING - must leave all elements isolated

    for (i = 0; i < n; i++) {
        assert(DLL_IS_ISOLATED(elements[i], list1));
        assert(DLL_IS_ISOLATED(elements[i], list2));
    }
}


/**
 * Test permutations of n elements.
 */
void testPermutations(
    int pivot,
    int max,
    Element **original,
    Element **elements)
{
    if (pivot < max) {
        Element *permuted[PERMUTATIONS];
        int next = pivot + 1;
        int i;
        for (i = 0; i < max; i++)
            permuted[i] = elements[i];
        testPermutations(next, max, original, permuted);
        for (i = next; i < max; i++) {
            permuted[pivot] = elements[i];
            permuted[i] = elements[i - 1];
            testPermutations(next, max, original, permuted);
        }
    }
    else {
        runPermutedTest(max, original, elements);
    }
}


/**
 * Test permuted insertions.
 */
void testPermute(void)
{
    int i;
    Element *elements[PERMUTATIONS];
    for (i = 0; i < PERMUTATIONS; i++) {
        Element *e = NEW(Element);
        e->payload = i;
        elements[i] = e;
    }
    for (i = 0; i < PERMUTATIONS; i++) {
        testPermutations(0, i + 1, elements, elements);
    }
}


/**
 * Public entry point for testing.
 */
int main(int argc, char **argv)
{
    test1();
    test2();
    test3();
    testPermute();
    testIterate();

    printf("%s passed\n", argv[0]);
    return 0;
}
