/*
 *------------------------------------------------------------------------------
 *
 * Name:	raid-msg.c
 *
 * RCS:		$Id: raid-msg.c,v 1.22 2002/11/09 00:09:03 nitinb Exp $
 *
 * Copyright (C) 2001, 2002 Agile Storage Inc.
 *
 * Description:	RAID config manager message processing code.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 4/16/02
 *
 *------------------------------------------------------------------------------
 */

#include "raid-common.h"
#include "raid-api.h"
#include "raid-msg.h"
#include "../sm-scsi/scsi-dev-cmds.h"
#include <sys/socket.h>
#include <stdio.h>

extern uint32     raid_device_poll_count;

int		raid_rxif = -1;
struct timeval	raid_receive_poll_to;

void raid_receive_msg(raid_msgHdr_t *, int, char *, uint32, uint32);
void raid_proc_set_debug_level(raid_setDebugLevelReqMsg_t *, char *, uint32, uint32);
void raid_proc_get_list(raid_getListReqMsg_t *, char *, uint32, uint32);
void raid_proc_get_info(raid_getInfoReqMsg_t *, char *, uint32, uint32);
void raid_proc_open(raid_openReqMsg_t *, char *, uint32, uint32);
void raid_proc_close(raid_closeReqMsg_t *, char *, uint32, uint32);

void raid_proc_simple_datain_req(raid_msgHdr_t *, char *, uint32, uint32);
void raid_proc_simple_nodata_req(raid_msgHdr_t *, char *, uint32, uint32);
void raid_proc_lun_create_req(raid_createLunReqMsg_t *, char *, uint32, uint32);
void raid_proc_lun_create_simple_req(raid_createSimpleLunReqMsg_t *, char *, uint32, uint32);
void raid_proc_lun_delete_req(raid_deleteLunReqMsg_t *, char *, uint32, uint32);
void raid_proc_lun_init_req(raid_initLunReqMsg_t *, char *, uint32, uint32);
void raid_proc_lun_check_req(raid_checkLunReqMsg_t *, char *, uint32, uint32);
void raid_proc_lun_cache_req(raid_cacheLunReqMsg_t *, char *, uint32, uint32);
void raid_proc_ctrl_setparam_req(raid_setCtrlParamReqMsg_t *, char *, uint32, uint32);
void raid_proc_config_req(raid_ctrlConfigReqMsg_t *, char *, uint32, uint32);
void raid_proc_dev_add_notify(sdm_devInfoRspMsg_t *);
void raid_proc_phy_passthru_req(raid_phyPassThruReqMsg_t *, char *, uint32, uint32);
void raid_proc_update_drv_fw_req(raid_updateDrvFwReqMsg_t *, char *, uint32, uint32);

void raid_proc_sdm_open_rsp(sdm_openRspMsg_t *);
void raid_proc_sdm_info_rsp(sdm_devInfoRspMsg_t *);
void raid_proc_sdm_close_rsp(sdm_closeRspMsg_t *);
void raid_proc_sdm_relay_rsp(sdm_relayCdbRspMsg_t *);

void raid_proc_unimplemented(raid_msgHdr_t *, char *, uint32, uint32);

void raid_print_sense_data(uchar8 *);

void
raid_msg_init(void)
{
	raid_rxif = regAgileApp("raid-config");
	if (raid_rxif < 0 ) {
		perror("raid-config failed to register for messages");
		exit(1);
	}

	raid_receive_poll_to.tv_sec = 1;
	raid_receive_poll_to.tv_usec = 0;
}

void
raid_receive_poll(void)
{
	uint32		src_cpu, src_slot, src_plane ;
	char		src_app[32];
	static char	recv_msg[65536*4];
	int		recv_bytes;

	recv_bytes = recvAgileMsg(raid_rxif, 
				  recv_msg, 
				  sizeof(recv_msg), 
				  &raid_receive_poll_to, 
				  src_app, &src_cpu, &src_slot, &src_plane);
	if (recv_bytes < 0) {
		RAID_DBG(2, ("raid_receive_poll: msg receive error\n"));
	}
	else 
	if (recv_bytes) {
		RAID_DBG(2, ("raid_receive_poll: received msg: "
			     "%d bytes (mid %d) from app %s, cpu %d, slot %d\n",
			     recv_bytes, *((uint32 *)(recv_msg)), src_app, src_cpu, src_slot));
		raid_receive_msg((raid_msgHdr_t *)recv_msg, recv_bytes, src_app, src_cpu, src_slot);
	}
}

void 
raid_receive_msg(raid_msgHdr_t *recv_msg, int recv_len, char *app, uint32 cpu, uint32 slot)
{
	raid_msgHdr_t	resp;

	switch(recv_msg->mid) {
	case RAID_MID_SET_DEBUG_LEVEL_REQ:
		if (recv_len != sizeof(raid_setDebugLevelReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid SET_DEBUG_LEVEL_REQ message length %d\n", recv_len));
			goto error_response;
		}
		raid_proc_set_debug_level((raid_setDebugLevelReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_GET_LIST_REQ:
		if (recv_len != sizeof(raid_getListReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid GET_LIST_REQ message length %d\n", recv_len));
			goto error_response;
		}
		raid_proc_get_list((raid_getListReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_GET_INFO_REQ:
		if (recv_len != sizeof(raid_getInfoReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid GET_INFO_REQ message length %d\n", recv_len));
			goto error_response;
		}
		raid_proc_get_info((raid_getInfoReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_OPEN_REQ:
		if (recv_len != sizeof(raid_openReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid OPEN_REQ message length %d\n", recv_len));
			goto error_response;
		}
		raid_proc_open((raid_openReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_CLOSE_REQ:
		if (recv_len != sizeof(raid_closeReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid CLOSE_REQ message length %d\n", recv_len));
			goto error_response;
		}
		raid_proc_close((raid_closeReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case SDM_MID_DEV_OPEN_RSP:
		if (recv_len != sizeof(sdm_openRspMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid SDM_MID_DEV_OPEN_RSP message length %d -- dropping\n", recv_len));
			goto drop_message;
		}
		raid_proc_sdm_open_rsp((sdm_openRspMsg_t *)recv_msg);
		break;

	case SDM_MID_DEV_INFO_RSP:
		if (recv_len != sizeof(sdm_devInfoRspMsg_t)) {
			RAID_DBG(2, ("tape_receive_msg: Invalid SDM_MID_DEV_INFO_RSP message length %d -- dropping\n", recv_len));
			goto drop_message;
		}
		raid_proc_sdm_info_rsp((sdm_devInfoRspMsg_t *)recv_msg);
		break;

	case SDM_MID_DEV_CLOSE_RSP:
		if (recv_len != sizeof(sdm_closeRspMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid SDM_MID_DEV_CLOSE_RSP message length %d -- dropping\n", recv_len));
			goto drop_message;
		}
		raid_proc_sdm_close_rsp((sdm_closeRspMsg_t *)recv_msg);
		break;

	case SDM_MID_RELAY_CDB_RSP:
		if (recv_len < sizeof(sdm_relayCdbRspMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid SDM_MID_RELAY_CDB_RSP message length %d -- dropping\n", recv_len));
			goto drop_message;
		}
		raid_proc_sdm_relay_rsp((sdm_relayCdbRspMsg_t *)recv_msg);
		break;

	case RAID_MID_CTRL_INFO_REQ:
	case RAID_MID_CTRL_GETPARAM_REQ:
	case RAID_MID_CTRL_GETSTATE_REQ:
	case RAID_MID_PHYDEV_LIST_REQ:
	case RAID_MID_LOGDEV_LIST_REQ:
	case RAID_MID_CTRL_CLRCFG_REQ:
	case RAID_MID_LUN_BPLIST_REQ:
		if (recv_len != sizeof(raid_msgHdr_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_simple_datain_req((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_CTRL_RESET_REQ:
	case RAID_MID_CTRL_FLUCACHE_REQ:
		if (recv_len != sizeof(raid_msgHdr_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_simple_nodata_req((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_PHYDEV_INFO_REQ:
		if (recv_len != sizeof(raid_phyDevInfoReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_simple_datain_req((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_PHYDEV_STATS_REQ:
		if (recv_len != sizeof(raid_phyDevStatsReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_simple_datain_req((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_LOGDEV_INFO_REQ:
		if (recv_len != sizeof(raid_logDevInfoReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_simple_datain_req((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_CTRL_CFGDATA_REQ:
		if (recv_len != sizeof(raid_ctrlConfigReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_config_req((raid_ctrlConfigReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_LOGDEV_PHYLIST_REQ:
		if (recv_len != sizeof(raid_logDevPhyListReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_simple_datain_req((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_LUN_CREATE_REQ:
		if (recv_len < sizeof(raid_createLunReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_lun_create_req((raid_createLunReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_LUN_CREATE_SIMPLE_REQ:
		if (recv_len != sizeof(raid_createSimpleLunReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_lun_create_simple_req((raid_createSimpleLunReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_LUN_DELETE_REQ:
		if (recv_len != sizeof(raid_deleteLunReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_lun_delete_req((raid_deleteLunReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_LUN_INIT_REQ:
		if (recv_len != sizeof(raid_initLunReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_lun_init_req((raid_initLunReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_LUN_CHECK_REQ:
		if (recv_len != sizeof(raid_checkLunReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_lun_check_req((raid_checkLunReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_LUN_CACHE_REQ:
		if (recv_len != sizeof(raid_cacheLunReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_lun_cache_req((raid_cacheLunReqMsg_t *)recv_msg, app, cpu, slot);
		break;
		
	case RAID_MID_CTRL_SETPARAM_REQ:
		if (recv_len != sizeof(raid_setCtrlParamReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_ctrl_setparam_req((raid_setCtrlParamReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_PHY_PASSTHRU_REQ:
		if (recv_len < sizeof(raid_phyPassThruReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_phy_passthru_req((raid_phyPassThruReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_UPDATE_DRV_FW_REQ:
		if(recv_len != sizeof(raid_updateDrvFwReqMsg_t)) {
			RAID_DBG(2, ("raid_receive_msg: Invalid 0x%x message length %d\n", recv_msg->mid, recv_len));
			goto error_response;
		}
		raid_proc_update_drv_fw_req((raid_updateDrvFwReqMsg_t *)recv_msg, app, cpu, slot);
		break;

	case RAID_MID_CTRL_SETSTATE_REQ:
	case RAID_MID_CTRL_FWUPGD_REQ:
	case RAID_MID_CTRL_SETCLOCK_REQ:
	case RAID_MID_PHYDEV_STATE_REQ:
	case RAID_MID_PHYDEV_FWUPGD_REQ:
	case RAID_MID_PHYDEV_ID_REQ:
	case RAID_MID_LOGDEV_STATS_REQ:
	case RAID_MID_LOGDEV_ID_REQ:
	case RAID_MID_LUNMASK_INFO_REQ:
	case RAID_MID_LUNMASK_ADD_REQ:
	case RAID_MID_LUNMASK_DEL_REQ:
	case RAID_MID_LUN_CANCELBP_REQ:
	case RAID_MID_ENCLOSE_LIST_REQ:
	case RAID_MID_ENCLOSE_SCAN_REQ:
		RAID_DBG(2, ("raid_receive_msg: Unimplemented mid 0x%x (offset mid id %d)\n", 
			     recv_msg->mid, recv_msg->mid - RAID_MSG_BASE));
		raid_proc_unimplemented((raid_msgHdr_t *)recv_msg, app, cpu, slot);
		break;

	default:
		RAID_DBG(2, ("raid_receive_msg: Invalid mid 0x%x (offset mid id %d)\n", recv_msg->mid, recv_msg->mid - RAID_MSG_BASE));
		goto drop_message;
	}
	return;

 error_response:
	resp.mid = RAID_MID_ERROR_RSP;
	resp.cookie = recv_msg->cookie;
	raid_send_msg(&resp, sizeof(raid_msgHdr_t), app, cpu, slot);
	return;

 drop_message:
	return;
}

void 
raid_send_msg(void *send_msg, int send_len, char *app, uint32 cpu, uint32 slot)
{
	int rc;

	RAID_DBG(2, ("raid_send_msg: sending msg: "
		     "%d bytes to app %s, cpu %d, slot %d\n",
		     send_len, app, cpu, slot));
	
	rc = sendAgileMsg(raid_rxif,
			  send_msg, send_len, 
			  app, cpu, slot, 0, 
			  NULL, 
			  NULL, 0);
}

void 
raid_proc_set_debug_level(raid_setDebugLevelReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_setDebugLevelRspMsg_t	resp;

	RAID_DBG(2, ("raid_proc_set_debug_level: debug_level 0x%x, app %s, cpu %d, slot %d\n",
		     msg->debug_level, app, cpu, slot));

	if (msg->debug_level == (uint32)-1)
		raid_terminate();
	else
		raid_dbg_level = msg->debug_level;

	resp.header.mid = RAID_MID_SET_DEBUG_LEVEL_RSP;
	resp.header.cookie = msg->header.cookie;
	raid_send_msg(&resp, sizeof(raid_setDebugLevelRspMsg_t), app, cpu, slot);
}

void 
raid_proc_get_list(raid_getListReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_getListRspMsg_t *	resp;
	raid_ctrl_id_t *	id;
	int			alloced_size = 0;
	int			used_size = 0;
	raid_dev_t *		dev;
	raid_dev_t *		next;
	int			len;

	RAID_DBG(2, ("raid_proc_get_devlist: app %s, cpu %d, slot %d\n",
		     app, cpu, slot));
	alloced_size = 1024 * 16;
	resp = (raid_getListRspMsg_t *)malloc(alloced_size);
	resp->header.mid = RAID_MID_GET_LIST_RSP;
	resp->header.cookie = msg->header.cookie;
	resp->rCode = RAID_RCODE_SUCCESS;
	resp->numEntries = 0;

	id = resp->idList;
	used_size = sizeof(raid_getListRspMsg_t) - sizeof(raid_ctrl_id_t);

	if ((dev = raid_devs)) {
		do {
			next = dev->next;
			len = MIN(RAID_DEV_NAME_LEN, RAID_CTRL_ID_LEN);
			memcpy(id, dev->device_name, len);
			++resp->numEntries;
			++id;
			used_size += sizeof(raid_ctrl_id_t);
			dev = next;
		} while (dev != raid_devs);
	}

	raid_send_msg(resp, used_size, app, cpu, slot);
	return;
}

void
raid_proc_dev_add_notify(sdm_devInfoRspMsg_t *msg)
{
	raid_dev_t*	dev;
	
	dev = raid_device_lookup_by_id(msg->id);

	// device already exists, just update its name
	if(dev) {
		strncpy(dev->vendor_name, msg->st.vendor, sizeof(dev->vendor_name));
		strncpy(dev->model_name, msg->st.model, sizeof(dev->model_name));
		strncpy(dev->device_name, msg->st.name, sizeof(dev->device_name));
	} 

	// device doesn't exist, it will discovered when next time the timer pops
}

void 
raid_proc_get_info(raid_getInfoReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_getInfoRspMsg_t *	resp;
	raid_dev_t *		dev;

	RAID_DBG(2, ("raid_proc_get_info: app %s, cpu %d, slot %d\n",
		     app, cpu, slot));
	resp = (raid_getInfoRspMsg_t *)malloc(sizeof(raid_getInfoRspMsg_t));
	resp->header.mid = RAID_MID_GET_INFO_RSP;
	resp->header.cookie = msg->header.cookie;

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		resp->rCode = RAID_RCODE_SUCCESS;
		resp->scsi_id = dev->id;
		resp->wwn = dev->wwn;
		resp->lun = dev->lun;
		switch (dev->state) {
		case RAID_STATE_CLOSED:
		case RAID_STATE_SDM_OPEN_PENDING:
			resp->state = RAID_DEV_STATE_CLOSED;
			break;
		case RAID_STATE_OPEN:
			resp->state = RAID_DEV_STATE_OPEN_IDLE;
			break;
		case RAID_STATE_OPEN_ACTIVE:
			resp->state = RAID_DEV_STATE_OPEN_BUSY;
			break;
		case RAID_STATE_HOLDING_CLOSE:
		case RAID_STATE_SDM_CLOSE_PENDING:
		case RAID_STATE_DESTROY_PENDING:
		case RAID_STATE_DESTROY_PENDING_AWAITING_CLOSE:
			resp->state = RAID_DEV_STATE_CLOSING;
			break;
		default:
			resp->state = -1;
		}
	}
	else 
		resp->rCode = RAID_RCODE_NO_DEVICE;
	
	raid_send_msg(resp, sizeof(raid_getInfoRspMsg_t), app, cpu, slot);
	return;
}

void 
raid_proc_open(raid_openReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_openRspMsg_t   resp;
	raid_dev_t *        dev;
	raid_msg_context_t *context;

	RAID_DBG(2, ("raid_proc_open: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev == NULL) {
		/* Reject open -- no such physical device */
		goto open_failed_no_device;
	}
	
	VERIFY(strcmp(dev->device_name, msg->header.id) == 0);
	context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
	raid_sm_run(dev, context, RAID_EVENT_OPEN_REQ);

	return;

 open_failed_no_device:
	RAID_DBG(2, ("raid_proc_open: open failed: No device %s\n", msg->header.id));
	resp.header.mid = RAID_MID_OPEN_RSP;
	resp.header.cookie = msg->header.cookie;
	resp.rCode = RAID_RCODE_NO_DEVICE;
	raid_send_msg(&resp, sizeof(raid_openRspMsg_t), app, cpu, slot);
	return;
}

void 
raid_proc_close(raid_closeReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_closeRspMsg_t	resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;

	RAID_DBG(2, ("raid_proc_close: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev == NULL) {
		/* Reject close -- invalid device name */
		goto close_failed_invalid_name;
	}
	
	context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
	raid_sm_run(dev, context, RAID_EVENT_CLOSE_REQ);

	return;

 close_failed_invalid_name:
	RAID_DBG(2, ("raid_proc_close: close failed: Invalid dev ID %s\n", msg->header.id));
	resp.header.mid = RAID_MID_CLOSE_RSP;
	resp.header.cookie = msg->header.cookie;
	resp.rCode = RAID_RCODE_NO_DEVICE;
	raid_send_msg(&resp, sizeof(raid_closeRspMsg_t), app, cpu, slot);
	return;
}

void 
raid_proc_simple_datain_req(raid_msgHdr_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_genericRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	
	RAID_DBG(2, ("raid_proc_simple_datain_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->id, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_simple_datain_req: request failed: Invalid device name %s\n", msg->id));
		err_resp.header.mid = (uint32)msg->id + 1;
		err_resp.header.cookie = msg->cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_genericRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_simple_nodata_req(raid_msgHdr_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_genericRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	
	RAID_DBG(2, ("raid_proc_simple_nodata_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->id, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_simple_nodata_req: request failed: Invalid device name %s\n", msg->id));
		err_resp.header.mid = (uint32)msg->id + 1;
		err_resp.header.cookie = msg->cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_genericRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_lun_create_req(raid_createLunReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_createLunRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	int			i;
	
	RAID_DBG(2, ("raid_proc_lun_create_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));
	RAID_DBG(2, ("raid_proc_lun_create_req: raidLevel %d, sizeInMB %d, cacheEnabled %d, autoInit %d, stripeSizeInKB %d\n",
		     msg->lundef.raidLevel, 
		     msg->lundef.sizeInMB,
		     msg->lundef.cacheEnabled,
		     msg->lundef.autoInit,
		     msg->lundef.stripeSizeInKB));
	RAID_DBG(2, ("raid_proc_lun_create_req: numOfPhyDevices %d: ", 
		     msg->lundef.numOfPhyDevices));

	/* Check parameters */
	if ((msg->lundef.raidLevel != 0 &&
	     msg->lundef.raidLevel != 1 &&
	     msg->lundef.raidLevel != 3 &&
	     msg->lundef.raidLevel != 5) ||
	    (msg->lundef.raidLevel == 0 && 
	     msg->lundef.numOfPhyDevices < 2) ||
		(msg->lundef.raidLevel == 1 &&
		 msg->lundef.numOfPhyDevices != 2) ||
		(msg->lundef.raidLevel == 3 &&
		 msg->lundef.numOfPhyDevices < 3) ||
		(msg->lundef.raidLevel == 5 &&
		 msg->lundef.numOfPhyDevices < 3)) {
		/* Reject request -- invalid parameters */
		RAID_DBG(2, ("raid_proc_lun_create_req: request failed: Invalid parameters\n"));
		err_resp.header.mid = RAID_MID_LUN_CREATE_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_INVALID_PARAMETERS;
		raid_send_msg(&err_resp, sizeof(raid_createLunRspMsg_t), app, cpu, slot);
		return;
	}

#ifdef DEBUG
	if (raid_dbg_level >= 2) {
		for (i = 0; i < msg->lundef.numOfPhyDevices; ++i) {
			RAID_DBG(2, ("%d/%d ", 
				     msg->lundef.phyDevList[i].channelNumber,
				     msg->lundef.phyDevList[i].targetId));
		}
		RAID_DBG(2, ("\n"));
	}
#endif

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_create_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_LUN_CREATE_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_createLunRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_lun_create_simple_req(raid_createSimpleLunReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_createSimpleLunRspMsg_t	err_resp;
	raid_dev_t *			dev;
	raid_msg_context_t *		context;
	
	RAID_DBG(2, ("raid_proc_lun_create_simple_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));
	RAID_DBG(2, ("raid_proc_lun_create_simple_req: lunCount %d, raidLevel %d\n", 
		     msg->lunCount, msg->raidLevel));

	/* Check parameters */
	if (msg->lunCount > RAID_MAX_LUNS ||
		msg->lunCount <= 0 ||
	    (msg->raidLevel != 0 &&
	     msg->raidLevel != 3 &&
	     msg->raidLevel != 5)) {
		/* Reject request -- invalid parameters */
		RAID_DBG(2, ("raid_proc_lun_create_simple_req: request failed: Invalid parameters\n"));
		err_resp.header.mid = RAID_MID_LUN_CREATE_SIMPLE_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_INVALID_PARAMETERS;
		raid_send_msg(&err_resp, sizeof(raid_createSimpleLunRspMsg_t), app, cpu, slot);
		return;
	}

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_create_simple_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_LUN_CREATE_SIMPLE_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_createSimpleLunRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_lun_delete_req(raid_deleteLunReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_deleteLunRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	
	RAID_DBG(2, ("raid_proc_lun_delete_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));
	RAID_DBG(2, ("raid_proc_lun_delete_req: ID %d\n", msg->id));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_delete_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_LUN_DELETE_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_deleteLunRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_lun_init_req(raid_initLunReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_initLunRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	
	RAID_DBG(2, ("raid_proc_lun_init_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));
	RAID_DBG(2, ("raid_proc_lun_init_req: type %d, id %d\n", msg->type, msg->id));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_init_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_LUN_INIT_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_initLunRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_lun_check_req(raid_checkLunReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_checkLunRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	
	RAID_DBG(2, ("raid_proc_lun_check_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));
	RAID_DBG(2, ("raid_proc_lun_check_req: type %d, id %d\n", msg->type, msg->id));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_check_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_LUN_CHECK_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_checkLunRspMsg_t), app, cpu, slot);
		return;
	}
}

void
raid_proc_lun_cache_req(raid_cacheLunReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_cacheLunRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	
	RAID_DBG(2, ("raid_proc_lun_cache_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));
	RAID_DBG(2, ("raid_proc_lun_cache_req: wc_enabled %d, id %d\n", msg->wc_enabled, msg->id));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_cache_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_LUN_CACHE_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_cacheLunRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_ctrl_setparam_req(raid_setCtrlParamReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_setCtrlParamRspMsg_t err_resp;
	raid_dev_t *		  dev;
	raid_msg_context_t *	  context;
	
	RAID_DBG(2, ("raid_proc_setparam_req: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_ctrl_setparam_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_CTRL_SETPARAM_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_setCtrlParamRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_config_req(raid_ctrlConfigReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_ctrlConfigRspMsg_t	err_resp;
	raid_dev_t *		dev;
	raid_msg_context_t *	context;
	FILE *			f;

	RAID_DBG(2, ("raid_proc_config_req: dev ID %s, type %d, filename %s, app %s, cpu %d, slot %d\n",
		     msg->header.id, msg->type, msg->fileName, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		switch (msg->type) {
		case RAID_CONFIG_LOAD:
			if ((f = fopen(msg->fileName, "r")) < 0) {
				RAID_DBG(2, ("raid_proc_config_req: request failed: Cannot open file '%s' for reading.\n", 
					     msg->fileName));
				fclose(f);
				err_resp.header.mid = RAID_MID_CTRL_CFGDATA_RSP;
				err_resp.header.cookie = msg->header.cookie;
				err_resp.rCode = RAID_RCODE_CANNOT_OPEN_FILE;
				raid_send_msg(&err_resp, sizeof(raid_ctrlConfigRspMsg_t), app, cpu, slot);
				return;
			}
			break;

		case RAID_CONFIG_SAVE:
			if ((f = fopen(msg->fileName, "w")) < 0) {
				RAID_DBG(2, ("raid_proc_config_req: request failed: Cannot open file '%s' for writing.\n", 
					     msg->fileName));
				fclose(f);
				err_resp.header.mid = RAID_MID_CTRL_CFGDATA_RSP;
				err_resp.header.cookie = msg->header.cookie;
				err_resp.rCode = RAID_RCODE_CANNOT_OPEN_FILE;
				raid_send_msg(&err_resp, sizeof(raid_ctrlConfigRspMsg_t), app, cpu, slot);
				return;
			}
			break;
			
		default:
			err_resp.header.mid = RAID_MID_CTRL_CFGDATA_RSP;
			err_resp.header.cookie = msg->header.cookie;
			err_resp.rCode = RAID_RCODE_INVALID_PARAMETERS;
			raid_send_msg(&err_resp, sizeof(raid_ctrlConfigRspMsg_t), app, cpu, slot);
			return;
		}

		fclose(f);
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_config_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_CTRL_CFGDATA_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_ctrlConfigRspMsg_t), app, cpu, slot);
		return;
	}
}

void
raid_proc_phy_passthru_req(raid_phyPassThruReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_dev_t					*dev;
	raid_phyPassThruRspMsg_t	err_resp;
	raid_msg_context_t			*context = NULL;

	dev = raid_device_lookup_by_devname(msg->header.id);
	if (dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_phy_passthru_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_PHY_PASSTHRU_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_phyPassThruRspMsg_t), app, cpu, slot);
		return;
	}
}

void
raid_proc_update_drv_fw_req(raid_updateDrvFwReqMsg_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_dev_t					*dev;
	raid_updateDrvFwRspMsg_t	err_resp;
	raid_msg_context_t			*context = NULL;

	dev = raid_device_lookup_by_devname(msg->header.id);
	if(dev) {
		context = raid_save_context((raid_msgHdr_t *)msg, app, cpu, slot);
		raid_sm_run(dev, context, RAID_EVENT_MESSAGE_REQ);
		return;
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_update_drv_fw_req: request failed: Invalid device name %s\n", msg->header.id));
		err_resp.header.mid = RAID_MID_UPDATE_DRV_FW_RSP;
		err_resp.header.cookie = msg->header.cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_updateDrvFwRspMsg_t), app, cpu, slot);
		return;
	}
}

void 
raid_proc_sdm_open_rsp(sdm_openRspMsg_t *msg)
{
	raid_dev_t *	dev;

	dev = (raid_dev_t *)msg->hdr.chandle;
	if (msg->rCode == SDM_OK)
		raid_sm_run(dev, msg, RAID_EVENT_SDM_OPEN_DONE);
	else
	if (msg->rCode == SDM_TO_MANY_OPENS)
		raid_sm_run(dev, msg, RAID_EVENT_SDM_OPEN_ALREADY_OPEN);
	else
		raid_sm_run(dev, msg, RAID_EVENT_SDM_OPEN_FAILED);
}

void 
raid_proc_sdm_info_rsp(sdm_devInfoRspMsg_t *msg)
{
	raid_dev_t *	dev;

	dev = (raid_dev_t *)msg->hdr.chandle;

	if (msg->rc == SDM_OK) {
		// this case happens when sdm informs raid-cfgd of a name change in raid-device without
		// the raid-cfgd sending a devInfoReqMsg
		if(dev == NULL)
			raid_proc_dev_add_notify(msg);
		else
			raid_sm_run(dev, msg, RAID_EVENT_SDM_INFO_DONE);
	}
	else
		raid_sm_run(dev, msg, RAID_EVENT_SDM_INFO_FAILED);
}

void 
raid_proc_sdm_close_rsp(sdm_closeRspMsg_t *msg)
{
	raid_dev_t *	dev;

	dev = (raid_dev_t *)msg->hdr.chandle;
	if (msg->rCode == SDM_OK)
		raid_sm_run(dev, msg, RAID_EVENT_SDM_CLOSE_DONE);
	else
		raid_sm_run(dev, msg, RAID_EVENT_SDM_CLOSE_FAILED);
}

void 
raid_proc_sdm_relay_rsp(sdm_relayCdbRspMsg_t *msg)
{
	raid_dev_t *		dev;
	sdm_relayCdbReqMsg_t *	req;

	dev = (raid_dev_t *)msg->hdr.chandle;
	req = dev->pending_req_context;
	VERIFY(req);

	if (msg->eCode == SDM_FAIL && (msg->status & SD_DRV_STATUS_MASK) == 0)
		raid_sm_run(dev, msg, RAID_EVENT_SDM_RELAY_FAILED);
	else 
	if ((msg->status & SD_DEV_STATUS_MASK) == 0)
		raid_sm_run(dev, msg, RAID_EVENT_SDM_RELAY_DONE);
	else {
		RAID_DBG(2, ("raid_proc_sdm_relay_rsp: status 0x%08x\n", msg->status));
#ifdef DEBUG
		if (msg->status & SD_DEV_SENSE_VALID) {
			if (raid_dbg_level >= 2) {
				int i;

				raid_print_sense_data(msg->senseData);
				for (i = 0; i < 32; ++i) {
					if ((i % 8) == 0) {
						if (i > 0)
							printf("\n");
						printf("raid_proc_sdm_relay_rsp: sense ");
					}
					printf("%02x ", msg->senseData[i]);
				}
				printf("\n");
			}
		}
#endif /* DEBUG */
		raid_sm_run(dev, msg, RAID_EVENT_SDM_RELAY_SCSI_ERROR);
	}
}

void 
raid_proc_unimplemented(raid_msgHdr_t *msg, char *app, uint32 cpu, uint32 slot)
{
	raid_genericRspMsg_t	err_resp;
	raid_dev_t *		dev;
	
	RAID_DBG(2, ("raid_proc_unimplemented: dev ID %s, app %s, cpu %d, slot %d\n",
		     msg->id, app, cpu, slot));

	dev = raid_device_lookup_by_devname(msg->id);
	if (dev) {
		RAID_DBG(2, ("raid_proc_unimplemented: request failed: Unimplemented\n"));
		err_resp.header.mid = (uint32)msg->id + 1;
		err_resp.header.cookie = msg->cookie;
		err_resp.rCode = RAID_RCODE_NOT_IMPLEMENTED;
		raid_send_msg(&err_resp, sizeof(raid_genericRspMsg_t), app, cpu, slot);
	}
	else {
		/* Reject request -- invalid device name */
		RAID_DBG(2, ("raid_proc_lun_check_req: request failed: Invalid device name %s\n", msg->id));
		err_resp.header.mid = RAID_MID_LUN_CHECK_RSP;
		err_resp.header.cookie = msg->cookie;
		err_resp.rCode = RAID_RCODE_NO_DEVICE;
		raid_send_msg(&err_resp, sizeof(raid_genericRspMsg_t), app, cpu, slot);
	}
}


void
raid_cleanup(void)
{
	raid_close_all_devices();
}

void
raid_print_sense_data(uchar8 *data)
{
	scsi_sense_t	*sense = (scsi_sense_t *)data;
	uint32			info;

	printf("Error code = 0x%x\n", sense->respCode & 0x7F);
	printf("Segment number = %d\n", sense->segNumber);
	printf("Filemark %d EOM %d ILI %d Sense Key %d\n",
			sense->senseKey && SCSI_SENSE_FILEMARK_BIT ,
			sense->senseKey && SCSI_SENSE_EOM_BIT,
			sense->senseKey && SCSI_SENSE_ILI_BIT,
			sense->senseKey & SCSI_SENSE_KEY_MASK);

	memcpy((void *)&info, sense->info, sizeof(uint32));
	printf("Information = %d\n", ntohl(info));
	printf("Additional Sense 0x%x 0x%x\n", sense->asc, sense->ascQ);
}
