/*
 *------------------------------------------------------------------------------
 *
 * Name:	raid-parse-logpage.c
 *
 * RCS:		$Id: raid-parse-logpage.c,v 1.1 2002/10/23 18:37:03 nitinb Exp $
 *
 * Copyright (C) 2001, 2002 Agile Storage Inc.
 *
 * Description:	Processes log pages meaningfully.
 *
 * Created by:	Nitin Bahadur
 *
 * Date Created: 16 Oct, 2002
 *
 *------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <assert.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>

#include "raid-api.h"
#include "raid-parse-logpage.h"

static char*	raid_parse_get_pagename(uint32);
static char*	raid_parse_get_error_string(uint32);

static char*	raid_parse_get_pagename(uint32 page_code)
{
	switch(page_code) {
		case SCSI_LOGPAGE_CODE_SUPPORTED_PAGES:
			return "Supported log pages page";
		case SCSI_LOGPAGE_CODE_BUFFER_OVER_UNDER_RUN:
			return "Buffer over-run/under-run page";
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE:
			return "Error Counter page (write errors)";
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ:
			return "Error Counter page (read errors)";
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ_REV:
			return "Error Counter page (read reverse errors)";
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_VERIFY:
			return "Error Counter page (verify errors)";
		case SCSI_LOGPAGE_CODE_NON_MEDIUM_ERROR:
			return "Non-medium error page";
		case SCSI_LOGPAGE_CODE_LAST_N_ERRORS:
			return "Last n error events page";
		case SCSI_LOGPAGE_CODE_LAST_N_DEFERRED_ERRORS:
			return "Last n deferred errors or asynchronous events page";
		case SCSI_LOGPAGE_CODE_TEMPERATURE:
			return "Temperature page";
		case SCSI_LOGPAGE_CODE_START_STOP_CYCLE_COUNTER:
			return "Start-Stop cycle counter page";
		case SCSI_LOGPAGE_CODE_APPLICATION_CLIENT:
			return "Application client page";
		case SCSI_LOGPAGE_CODE_SELF_TEST_RESULTS:
			return "Self-test results page";
		default:
			if(page_code >= 0x30 && page_code <= 0x3E)
				return "Vendor-specific pages";
			else
				return "Reserved";
	}

	return "N/A";
}

void	raid_parse_logpage_list_rsp(uint32 dataLen, char *data)
{
	uint32	count;
	uint32	page_code;
	int		i;

	assert(data[0] == SCSI_LOGPAGE_CODE_SUPPORTED_PAGES);

	count = data[2] * 256 + data[3];

	if(!count) return;

	printf("\nPage Code	Page Name\n");
	printf("---------   ---------------------\n");
	for(i=0; i < count; i++) {
		page_code = data[4+i];
		printf("0x%-8x  %s\n", page_code, raid_parse_get_pagename(page_code));
	}
}

static char* raid_parse_get_error_string(uint32 error_code)
{
	switch(error_code) {
		case SCSI_ERROR_LOGPAGE_NO_DELAY_ERRORS:
			return "Errors corrected without substantial delay";
		case SCSI_ERROR_LOGPAGE_DELAY_ERRORS:
			return "Errors corrected with possible delays";
		case SCSI_ERROR_LOGPAGE_RETRIED_OPS:
			return "Total retried operations";
		case SCSI_ERROR_LOGPAGE_CORRECTED_ERRORS:
			return "Total errors corrected";
		case SCSI_ERROR_LOGPAGE_NUM_ECC_PROCESSED:
			return "Total times correction algorithm processed";
		case SCSI_ERROR_LOGPAGE_BYTES_PROCESSED:
			return "Total bytes processed";
		case SCSI_ERROR_LOGPAGE_UNCORRECTED_ERRORS:
			return "Total uncorrected errors";
		default:
			if(error_code >= SCSI_ERROR_LOGPAGE_FIRST_RESERVED && 
				error_code <= SCSI_ERROR_LOGPAGE_LAST_RESERVED)
				return "Reserved";
			else
				return "Vendor-specific";
	}

	return "N/A";
}

void	raid_parse_logpage_error_rsp(uint32 page_code, uint32 dataLen, char *data, int code)
{
	int		page_length;
	char*	param;
	int		param_len;
	uint32	param_code;
	uint32	val32;
	uint64	val64;
	uint64	val;
	char	du, ds, tsd, etc, tmc, lbin, lp;
	
	assert( data[0] == SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE ||
			data[0] == SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ ||
			data[0] == SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ_REV ||
			data[0] == SCSI_LOGPAGE_CODE_ERROR_COUNTER_VERIFY);
	
	if(code == ALL_PAGE_CODES) {
		printf("\n%s\n\n", raid_parse_get_pagename(page_code));
		printf("Error Type                                   Count           DU  DS  TSD  ETC  TMC  LBIN  LP\n");
		printf("-------------------------------------------  --------------  --  --  ---  ---  ---  ----  --\n");
	}

	page_length = data[2] * 256 + data[3];
	param = data + 4;

	while(page_length > 0) {
		param_code = param[0] * 256 + param[1];
		du = ( param[2] >> 7 ) & 0x1;
		ds = ( param[2] >> 6 ) & 0x1;
		tsd = ( param[2] >> 5 ) & 0x1;
		etc = ( param[2] >> 4 ) & 0x1;
		tmc = ( param[2] >> 2 ) & 0x3;
		lbin = ( param[2] >> 1 ) & 0x1;
		lp = param[2] & 0x1;
		param_len = param[3];
		switch(param_len) {
			case 1:
				val = param[4];
				break;
			case 2:
				val = param[4] * 256 + param[5];
				break;
			case 4:
				memcpy((void *)&val32, &param[4], 4);
				val = ntohl(val32);
				break;
			case 8:
				memcpy((void *)&val64, &param[4], 8);
				if(ntohl(16) != 16)
					SWAP8(&val64);
				val = val64;
				break;
		}
		if(code == ALL_PAGE_CODES) {
			printf("%-43s  %14lld  %2d  %2d  %3d  %3d  %3d  %4d  %d\n", 
					raid_parse_get_error_string(param_code), val,
					du, ds, tsd, etc, tmc, lbin, lp);	
		}
		else if(code == param_code)
			printf("%lld", val);

		param += 4 + param_len;
		page_length -= (4 + param_len);
	}
}

static char cause_string[][25] = { 
				"Undefined", 
				"Bus Busy", 
				"Transfer rate too slow", 
				"Reserved"
};

static char count_basis_string[][25] = {
				"Undefined",
				"Per command",
				"Per failed reconnect",
				"Per unit of time",
				"Reserved"
};

static char buffer_type_string[][12] = {
				"Under-Run",
				"Over-Run"
};

void	raid_parse_buffer_run_rsp(uint32 dataLen, char *data)
{
	int		page_length;
	char	*param;
	int		count_basis;
	int		cause;
	int		type;

	assert(data[0] = SCSI_LOGPAGE_CODE_BUFFER_OVER_UNDER_RUN);

	printf("Buffer Over-Run/Under-Run Page\n");
	printf("Count Basis          Cause                 Type\n");
	printf("-------------------  --------------------  ----\n");

	page_length = data[2] * 256 + data[3];
	param = data + 4;

	while(page_length > 0) {
		count_basis = ( param[1] >> 5 ) & 0x7;
		if(count_basis > 0x4) count_basis = 0x4;

		cause = ( param[1] >> 1) & 0xF;
		if(cause > 0x3) cause = 3;

		type = param[1] & 0x1;

		printf("%-22s  %s24  %s\n", 
				count_basis_string[count_basis],
				cause_string[cause], 
				buffer_type_string[type]);
				
		param += 2;
		page_length -= 2;
	}
}

void	raid_parse_logpage_temp_rsp(uint32 dataLen, char *data)
{
	int page_length;
	
	assert(data[0] == SCSI_LOGPAGE_CODE_TEMPERATURE);

	printf("\nTemperature Page\n");
	printf("Temperature = %d degree Celsius\n", data[9]);

	page_length = data[2] * 256 + data[3];

	if(page_length == 16 && data[10] == 0x01)
		printf("Reference Temperature = %d degree Celsius\n", data[15]);
}
