/*
 *------------------------------------------------------------------------------
 *
 * Name:	raid-passthru.c
 *
 * RCS:		$Id: raid-passthru.c,v 1.1 2002/10/23 18:37:03 nitinb Exp $
 *
 * Copyright (C) 2001, 2002 Agile Storage Inc.
 *
 * Description:	Sets up CDB for pass through SCSI commands
 *
 * Created by:	Nitin Bahadur
 *
 * Date Created: 16 Oct, 2002
 *
 *------------------------------------------------------------------------------
 */

#include "raid-common.h"
#include "raid-xyratex.h"
#include "../sm-scsi/scsi-dev-cmds.h"

#define RAID_PASSTHRU_MAX_DATA_IN_LEN	4096

int raid_xyratex_setup_cmd_cdb(raid_phyPassThruReqMsg_t *, sdm_relayCdbReqMsg_t *); 

// pass through command setup routines
static void raid_passthru_get_logpage(raid_phyPassThruReqMsg_t *, sdm_relayCdbReqMsg_t *);
static void raid_passthru_read_long(raid_phyPassThruReqMsg_t *, sdm_relayCdbReqMsg_t *);
static void raid_passthru_write_long(raid_phyPassThruReqMsg_t *, sdm_relayCdbReqMsg_t *);
static void raid_passthru_mode_sense(raid_phyPassThruReqMsg_t *, sdm_relayCdbReqMsg_t *);

// structure storing information about a pass through command
typedef struct {
	int	cmd;
	void	(*setup)(raid_phyPassThruReqMsg_t *, sdm_relayCdbReqMsg_t *); 
} raid_passthru_cmd;


// list of implemented pass through commands
static raid_passthru_cmd passthru_cmds[] = {
	{ RAID_PASSTHRU_PAGELIST,			raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_BUFFER_RUN_PAGE,	raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_WRITE_ERROR_PAGE,	raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_READ_ERROR_PAGE,	raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_REV_READ_ERROR_PAGE,raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_VERIFY_ERROR_PAGE,	raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_TEMP_PAGE,			raid_passthru_get_logpage		},
	{ RAID_PASSTHRU_READ_LONG,			raid_passthru_read_long			},
	{ RAID_PASSTHRU_WRITE_LONG,			raid_passthru_write_long		},
	{ RAID_PASSTHRU_MODE_SENSE,			raid_passthru_mode_sense		},
};


// set up the cdb for the command to be executed
int
raid_xyratex_setup_cmd_cdb(raid_phyPassThruReqMsg_t *raid_msg, 
			   sdm_relayCdbReqMsg_t *req_msg)
{
	int i;

	req_msg->dataLen = RAID_PASSTHRU_CDB_ALLOC_LEN;
	req_msg->data[0] = raid_msg->channelNumber;
	req_msg->data[1] = raid_msg->targetId;
	req_msg->data[4] = 1; // allocation len type (1 == bytes)

	for(i=0; i < sizeof(passthru_cmds)/sizeof(raid_passthru_cmd); i++) {
		if(raid_msg->cmd == passthru_cmds[i].cmd) {
			passthru_cmds[i].setup(raid_msg, req_msg);
			return RAID_RCODE_SUCCESS;
		}
	}
	return RAID_RCODE_INVALID_PARAMETERS;
}


void
raid_passthru_get_logpage(raid_phyPassThruReqMsg_t *raid_msg, 
			sdm_relayCdbReqMsg_t *req_msg)
{
	raid_msg->dataLen = RAID_PASSTHRU_MAX_DATA_IN_LEN;

	req_msg->data[2] = 10;	// pass thro. CDB Len
	req_msg->data[3] = 1;	// Data IN

	// start CDB of pass through cmd 
	req_msg->data[16] = SCSI_CMD_LOG_SENSE;											// byte 0
	switch(raid_msg->cmd) {
		case RAID_PASSTHRU_PAGELIST:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_SUPPORTED_PAGES;	
			break;
		case RAID_PASSTHRU_BUFFER_RUN_PAGE:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_BUFFER_OVER_UNDER_RUN;
			break;
		case RAID_PASSTHRU_WRITE_ERROR_PAGE:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE;
			break;
		case RAID_PASSTHRU_READ_ERROR_PAGE:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ;
			break;
		case RAID_PASSTHRU_REV_READ_ERROR_PAGE:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ_REV;
			break;
		case RAID_PASSTHRU_VERIFY_ERROR_PAGE:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_ERROR_COUNTER_VERIFY;
			break;
		case RAID_PASSTHRU_TEMP_PAGE:
			req_msg->data[18] = SCSI_LOGPAGE_CODE_TEMPERATURE;	
			break;
	}
	req_msg->data[18] |= (SCSI_LOG_PC_CURRENT_CUMULATIVE << SCSI_LOG_PC_OFFSET);
	req_msg->data[23] = (RAID_PASSTHRU_MAX_DATA_IN_LEN >> 8) & 0xFF;				// byte 7
	req_msg->data[24] = RAID_PASSTHRU_MAX_DATA_IN_LEN & 0xFF;						// byte 8
}

void
raid_passthru_read_long(raid_phyPassThruReqMsg_t *raid_msg, 
							sdm_relayCdbReqMsg_t *req_msg)
{
	req_msg->data[2] = 10;	// pass thro. CDB len
	req_msg->data[3] = 1;	// Data IN

	// start CDB of pass through cmd
	req_msg->data[16] = SCSI_CMD_READ_LONG;						// byte 0
	req_msg->data[17] = 0x2;									// byte 1
	req_msg->data[18] = ( raid_msg->lba >> 24 ) & 0xFF;			// byte 2
	req_msg->data[19] = ( raid_msg->lba >> 16 ) & 0xFF;			// byte 3
	req_msg->data[20] = ( raid_msg->lba >> 8 ) & 0xFF;			// byte 4
	req_msg->data[21] = raid_msg->lba & 0xFF;					// byte 5
	req_msg->data[23] = ( raid_msg->dataLen >> 8 ) & 0xFF;		// byte 7
	req_msg->data[24] = raid_msg->dataLen & 0xFF;				// byte 8
}

void
raid_passthru_write_long(raid_phyPassThruReqMsg_t *raid_msg, 
							sdm_relayCdbReqMsg_t *req_msg) 
{
	req_msg->data[2] = 10;	// pass thro. CDB len
	req_msg->data[3] = 0;	// Data OUT

	// start CDB of pass through cmd
	req_msg->data[16] = SCSI_CMD_WRITE_LONG;					// byte 0
	req_msg->data[18] = ( raid_msg->lba >> 24 ) & 0xFF;			// byte 2
	req_msg->data[19] = ( raid_msg->lba >> 16 ) & 0xFF;			// byte 3
	req_msg->data[20] = ( raid_msg->lba >> 8 ) & 0xFF;			// byte 4
	req_msg->data[21] = raid_msg->lba & 0xFF;					// byte 5
	req_msg->data[23] = ( raid_msg->dataLen >> 8 ) & 0xFF;		// byte 7
	req_msg->data[24] = raid_msg->dataLen & 0xFF;				// byte 8
}


void
raid_passthru_mode_sense(raid_phyPassThruReqMsg_t *raid_msg, 
			sdm_relayCdbReqMsg_t *req_msg)
{
	int page_code = raid_msg->data[0];

	raid_msg->dataLen = 255;

	req_msg->data[2] = 6;	// pass thro. CDB len
	req_msg->data[3] = 1;	// Data IN

	// start CDB of pass through cmd
	req_msg->data[16] = SCSI_CMD_MODE_SENSE_6;					// byte 0
	req_msg->data[17] = 0;										// byte 1	
	req_msg->data[18] = page_code;								// byte 2
	req_msg->data[19] = 0;										// byte 3
	req_msg->data[20] = 255;									// byte 4
	req_msg->data[21] = 0;										// byte 5
}
