/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: rcon-api.h
 *
 * RCS:  $Id: rcon-api.h,v 1.4 2002/12/05 06:45:31 tonyw Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains the definition of the external API to the remote
 *              console application of the NFX. There are two parts: The
 *              messages passed between the SSC and the rcon agent on the
 *              embedded systems; and the interface with the embedded system's
 *              UI.
 *           *** Currently only the former is described ***
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 19 Dec 2001
 *
 *-----------------------------------------------------------------
 */
#if !defined (__rcon_api_h__)
#define __rcon_api_h__

#include "nfx-incl.h"

/*
	These messages are exchanged between the rcon instances on the SSC
	and the embedded CPU UI's (one per CPU image). The protocol is simply:

		SSC instance of rcon sends RCON_CNTL(INIT_CONNECTION) to rcon app 
		on the embedded CPU it wishes to establish a console for. (Appropriate 
		timeout/retry)

		Embedded CPU instance of rcon will respond with RCON_STAT indicating:
		NFX_OK - Connection is established; or NFX_ERROR - if there is an
		established connection already (or any other error).

		Once the connection is established, the application id on the SSC
		will remain constant (read: not the same for each instance since we'll 
		be using a random port) for that session. Either end will exchange 
		RCON_DATA messages.

		The SSC instance will only send the RCON_DATA message containing a
		complete command line (with terminating "\n\0").

		Either end can terminate the connection by sending a 
		RCON_CNTL(DESTROY_CONNECTION) message to its peer.

	The message transport is unreliable and data could get dropped (especially
	if either end is really busy). It's console data after all.
 */

typedef enum {
	RCON_CNTL = RCONN_MSG_BASE,
	RCON_DATA,
	RCON_STAT
} rconMsg_t;

typedef struct {
	rconMsg_t	msgType;		// The message is...
	uint32		clientPrivate;	// Could eventually be useful
} rconHdr_t;

typedef enum {
	INIT_CONNECTION,
	DESTROY_CONNECTION
} rconCtl_t;

typedef struct {
	rconHdr_t 	hdr;			// RCON_CNTL
	rconCtl_t 	action;			// What are we doing
} rconCntlMsg_t;

typedef struct {
	rconHdr_t	hdr;			// RCON_DATA
	uint32		len;			// How many bytes follow
	char		buf[1];			// Will be of length 'len'
} rconDataMsg_t;

typedef struct {
	rconHdr_t	hdr;			// RCON_STAT
	uint32		stat;			// Is good, yes?
	char		errMsg[1];		// Optional indication of what happened
} rconStatMsg_t;

// Shall we describe the interface to the embedded UI here?

#define MAX_PID   20
#define RCON_MAX_USER_LEN   256
#define RCON_QUEUE_MAX		8
#define RCON_NUM_CPUS		4

extern void rcon_send_end_connection(uint32	slot, uint32 cpu);
extern void rconPoll(void);
extern int32	rcon_getchar(void);

extern uint32	rcon;
extern uint32	incoming_appid;
extern uint32	rcon_user_buf_len;

/*-----------------------------------------------------------------
 * Name:	rconUserBufService()
 *
 * Description:
 *
 * Created by:	eric choi
 *
 * Date Created:	02/07/02
 *
 *-----------------------------------------------------------------
 */
extern uint32  rconUserBufService(uchar8 * srcp, uint32 len);

#if defined (NFX_MOD_SSC)
// These puppies live in librcon for the pleasure of Christa
// get_from_rcon() returns the number of characters read from the embedded
// side. If 0, then secs_to_wait expired before anything was read.
// All others return 0 for success and -1 for error (with fprintf(stderr)
// telling why).

extern int establish_rcon_session (int slot, int cpu);
extern int send_to_rcon (char *a_string_to_send);
extern int get_from_rcon (char *a_place, int placesize, int secs_to_wait);
extern int shutdown_rcon_session (void);

#endif
#endif
