/*
 *  Copyright (c) 2001
 *  Broadcom Corporation, Irvine, CA
 *
 *  BROADCOM PROPRIETARY AND CONFIDENTIAL
 *  
 *  This software is furnished under a Broadcom license agreement
 *  and may only be used in accordance with and subject to the terms 
 *  and conditions of such license. 
 *
 * Author: Mark Vandevoorde
 */

#ifdef _PROFILING_

#include "sbprof.h"
#include "amelib.h"
#include "amerom.h"
#include "sb1250-include/sb1250_defs.h"
#include "sb1250-include/sb1250_regs.h"
#include "sb1250-include/sb1250_scd.h"
#include "sb1250-include/sbmips.h"

typedef unsigned short u_int16_t;

typedef u_int64_t tb_sample_t[6*256];

#define MAX_TB_SAMPLES (MAX_TBSAMPLE_BYTES/(sizeof(tb_sample_t)))

static tb_sample_t sbprof_tbbuf[MAX_TB_SAMPLES];
static int next_tb_sample;

/* must match source of pgather/pkludge */
typedef enum {
  DP_D_PCARCH,
  DP_D_STOPPED,
  DP_D_PID_IMAGE,
  DP_D_PID_DEAD,
  DP_D_CONTEXT,
  DP_D_EVENTS,
  DP_D_CONFIG,
  DP_D_LAST
} dp_driver_msg;
#define EVENT_HDR_MSG_SIZE 4         /* number of u_int8_t's per msg */
#define EVENT_MSG_SIZE 4             /* number of u_int8_t's per msg */
#define NEW_CONTEXT_MSG_SIZE 8      /* number of u_int8_t's per msg */
#define MAX_IMAGE_NAME 256           /* in 8-bit bytes */
#define NEW_PID_MSG_SIZE (28 + MAX_IMAGE_NAME)

/* The two definitions below must match those in pgather.c */
#define MAX_COUNTERS 4
#define MAX_SLOTS 8
#define MAX_CPUS 1

/* Tight loops can have skewed sample distributions when the sampling period
   is fixed.  If the loop and the performance counter interrupt handler
   have no dynamic stalls, one could even see all of the samples landing on a
   single instruction of the loop.

   A solution is to introduce a "ripple" to offset the fixed sampling period
   by a periodic wave.  Ripples are implemented here as a value in the range
   0..RIPPLE_SIZE-1 added to the sampling period.  RIPPLE_SIZE/2 is then
   subtracted to center the wave around the desired sampling period.

   A ripple value is generated by incrementing a previous ripple value by
   RIPPLE_GENERATOR using addition mod RIPPLE_SIZE.  Addition by
   RIPPLE_CONSTANT mod RIPPLE_SIZE must generate all values 0..RIPPLE_SIZE-1.

   If memory were infinite, one could keep the state of a previous ripple for
   each possible instruction PC.  This implementation approximates that by
   direct-mapping the PCs to RIPPLE_TABLE_SIZE distinct distinct ripples.
*/

#define RIPPLE_SIZE (1 << 12)
#define RIPPLE_TABLE_SIZE 1024
#define RIPPLE_GENERATOR (RIPPLE_SIZE/5)
static unsigned short ripple[RIPPLE_TABLE_SIZE];


typedef struct {
  u_int64_t total;	/* total number of interrupts */
  u_int64_t dropped;	/* total interrupts when buffers were full */
  u_int64_t value;	/* initial counter value when slot next entered */
  u_int32_t start_period_low;	/* counter val for starting a period */
  u_int8_t  start_period_high;
  u_int8_t  event;      /* pcarch-specific event_t */
  u_int8_t  hwcode;	/* pcarch-specific code for event */
  u_int8_t  flags;
#define ECS_INUSE   1   /* Is counter X slot actually used? */
#define ECS_USE_PTR 2   /* Use PTR or restart PC? */

} event_counter_state_t;

typedef struct {
  u_int32_t total;	 /* total for current run of slot */
  u_int32_t dropped;	 /* dropped for current run of slot */
  u_int64_t start_period;/* counter value to start a full period */
} active_counter_state_t;

typedef struct {
  /******* 32-byte boundary *********************************************/
  u_int32_t next;/* index into buf for next sample */
  int32_t   last_event; /* index of byte just past last event DP_D_EVENTS
			   message in buf that needs the number of events
			   filled in.  -1 means there is no such message and
			   that such a message must be entered before
			   adding event samples.
			 */
  u_int32_t threshold;  /* when does the current multiplexing slot expire? */
  u_int8_t slotid;      /* index of current multiplexing slot */
  u_int8_t nslots;	/* number of slots */
  u_int8_t use_ptr;     /* nonzero IFF counter 1 events get PC from PTR */
  u_int8_t pad[1];
  /******* 32-byte boundary *********************************************/
  active_counter_state_t cur_slot[MAX_COUNTERS];
  /******* 32-byte boundary *********************************************/
  event_counter_state_t slot[MAX_SLOTS][MAX_COUNTERS];
  /******* 32-byte boundary *********************************************/
  u_int8_t buf[MAX_SAMPLE_BYTES];	/* prolog plus samples */
} cpudata_t;

/* Hack to allocate 32-byte aligned memory for cpudata.
   We could call an allocator, but then we'd get an extra cache miss per
   interupt just to dereference the global pointer. */

static void alloc_space()
{
  __asm__ __volatile__ ("	.bss;"
			"	.align 5;" /* 2^5 == 32 byte-aligned */
			"	.globl _sbprof_cpudata;"
			"_sbprof_cpudata: .space %0;"
			"	.text;"
			: /* no output */
			: "i" (sizeof(cpudata_t) * MAX_CPUS));
  /* Silence compiler warning that this procedure is never used. */
  alloc_space();
}

extern cpudata_t _sbprof_cpudata[MAX_CPUS];

/* "declare" cpudata_t *cpudata; */
#define cpudata (_sbprof_cpudata[0])

static void slot_enter(cpudata_t *, u_int8_t, u_int32_t, u_int8_t);
static void slot_exit(cpudata_t *, u_int8_t, u_int64_t *);

/****************************************/
/* Perf Cntr Arch-specific Event Tables */
/****************************************/

#undef TRUE
#undef FALSE
#define TRUE (0 == 0)
#define FALSE (!TRUE)

#define BITVECTOR_CONST(x) ((u_int8_t) (x))

#define CNTRS_0123 BITVECTOR_CONST(0xf)
#define CNTRS_123  BITVECTOR_CONST(0xe)
#define CNTRS_1    BITVECTOR_CONST(0x2)

typedef struct {
  char     *name;
  u_int8_t event;
  u_int8_t hwcode;
  u_int8_t use_ptr;
  u_int8_t period;
  u_int8_t counters;
} sbprof_event_entry_t;

const sbprof_event_entry_t *event_lookup(const char *name,
					 const sbprof_event_entry_t *tab,
					 int num_events)
{
  int i;
  for (i = 0; i < num_events; i++) {
    if (!strcmp(name, tab[i].name)) return &tab[i];
  }
  return NULL;
}

static const sbprof_event_entry_t SB1_event_info[] =  {
  /* name   event_t   hwcode  PTR?  period counters */
  { "cycles",	0,	0x10,  FALSE, 17,  CNTRS_0123 },
  { "grad",     1,      0x04,  FALSE, 17,  CNTRS_123  },
  { "br",       2,      0x11,  FALSE, 17,  CNTRS_123  },
  { "ld",       3,      0x08,  FALSE, 17,  CNTRS_123  },
  { "rqlen",    4,      0x05,  FALSE, 17,  CNTRS_123  },
  { "urqlen",   5,      0x06,  FALSE, 17,  CNTRS_123  },
  { "ldhrq",    6,      0x0d,  FALSE, 17,  CNTRS_123  },
  { "pfnop",    7,      0x0e,  FALSE, 17,  CNTRS_123  },
  { "lshpf",    8,      0x0f,  FALSE, 17,  CNTRS_123  },
  { "st",       9,      0x09,  FALSE, 17,  CNTRS_123  },
  { "pref",     10,     0x0a,  FALSE, 17,  CNTRS_123  },
  { "biubsy",   11,     0x17,  FALSE, 17,  CNTRS_123  },
};

static const sbprof_event_entry_t SB1R2_event_info[] =  {
  /* name   event_t   hwcode  PTR?  period counters */
  { "cycles",   0,      16,   FALSE,  17,  CNTRS_0123 },
  /* Issue Breaks */
  { "ibreak",   1,      39,   TRUE,   19,  CNTRS_1    },
  /* Execution Counts and Slotting */
  { "L0",       2,      40,   TRUE,   19,  CNTRS_1    },
  { "L1",       3,      41,   TRUE,   19,  CNTRS_1    },
  { "E0",       4,      42,   TRUE,   19,  CNTRS_1    },
  { "E1",       5,      43,   TRUE,   19,  CNTRS_1    },
  /* Static and Dynamic Stalls */
  { "issue4",   6,      32,   FALSE,  17,  CNTRS_123  },
  { "noinst",   7,      33,   FALSE,  17,  CNTRS_123  },
  { "dfwait",   8,      34,   FALSE,  19,  CNTRS_123  },
  { "static",   9,      35,   FALSE,  19,  CNTRS_123  },
  { "nopipe",  10,      36,   FALSE,  19,  CNTRS_123  },
  { "ifill",   11,      37,   FALSE,  19,  CNTRS_123  }, /* usually right */
  { "dfill",   12,      38,   FALSE,  19,  CNTRS_123  }, /* broken */
  /* Pipeline Traps */
  { "mispr",   13,      47,   TRUE,   15,  CNTRS_1    },
  { "mboxrp",  14,      29,   FALSE,  15,  CNTRS_123  },
  { "dcfrp",   15,      28,   FALSE,  15,  CNTRS_123  },
  { "deprp",   16,      30,   FALSE,  15,  CNTRS_123  },
  { "fillrp",  17,      27,   FALSE,  15,  CNTRS_123  },
  { "replay",  18,      31,   FALSE,  15,  CNTRS_123  },
  /* Branch Information */
  { "br",      19,      44,   TRUE,   17,  CNTRS_1    },
  { "taken" ,  20,      45,   TRUE,   17,  CNTRS_1    },
  { "ptaken",  21,      46,   TRUE,   17,  CNTRS_1    },
  /* Cache Access */
  { "rqlen",   22,      1,    FALSE,  19,  CNTRS_123  },
  { "urqlen",  23,      2,    FALSE,  19,  CNTRS_123  },
  { "rdmiss",  24,      3,    FALSE,  17,  CNTRS_123  },
  { "dfsne",   25,      10,   FALSE,  17,  CNTRS_123  },
  { "dfs",     26,      11,   FALSE,  17,  CNTRS_123  },
  { "dfn",     27,      12,   FALSE,  17,  CNTRS_123  },
  { "if",      28,      15,   FALSE,  17,  CNTRS_123  },
  { "evict",   29,      13,   FALSE,  17,  CNTRS_123  },
  { "upgrd",   30,      7,    FALSE,  17,  CNTRS_123  },
  { "lshpf",   31,      6,    FALSE,  19,  CNTRS_123  },
  { "pfnop",   32,      5,    FALSE,  19,  CNTRS_123  },
  { "ldhrq",   33,      4,    FALSE,  19,  CNTRS_123  },
  /* BIU */
  { "zbabsy",  34,      17,   FALSE,  19,  CNTRS_123  },
  { "zbdbsy",  35,      18,   FALSE,  19,  CNTRS_123  },
  { "zbareq",  36,      19,   FALSE,  19,  CNTRS_123  },
  { "zbrdn",   37,      20,   FALSE,  19,  CNTRS_123  },
  { "biubsy",  38,      14,   FALSE,  19,  CNTRS_123  },
  /* Multiprocessor */
  { "scfail",  39,      26,   FALSE,  19,  CNTRS_123  },
  { "snphit",  40,      22,   FALSE,  19,  CNTRS_123  },
  { "snpqfl",  41,      23,   FALSE,  19,  CNTRS_123  },
  { "steal",   42,      24,   FALSE,  19,  CNTRS_123  },
  { "steald",  43,      25,   FALSE,  19,  CNTRS_123  },
  /* Instruction Counts */ 
  { "ld",      44,      8,    FALSE,  19,  CNTRS_123  },
  { "st",      45,      9,    FALSE,  19,  CNTRS_123  },
  { "grad",    46,      0,    FALSE,  19,  CNTRS_123  },
  /* Another BIU event */
  { "zbrd",    47,      21,   FALSE,  19,  CNTRS_123  }
};

/******************************************/
/* Perf Cntr Arch-specific Default Events */
/******************************************/

typedef const char *sbprof_default_slot[MAX_COUNTERS];

static const sbprof_default_slot SB1_default_slots[] = {
  {"cycles", "grad", "br", "ld" }
};

static const sbprof_default_slot SB1R2_default_slots[] = {
  {"cycles", "ibreak", "dfwait", "static" },
  {"cycles", "mispr",  "issue4", "replay" },
  {"cycles", "L0",     "nopipe", "noinst" },
  {"cycles", "L1",     "rqlen",  "rdmiss" },
  {"cycles", "E0",     "mboxrp", "dcfrp"  },
  {"cycles", "E1",     "deprp",  "fillrp" },
  {"cycles", "br",     "grad",   "rdmiss" }
};

typedef struct {
  const sbprof_default_slot * default_slots;
  int num_slots;
  const sbprof_event_entry_t *event_table;
  const int num_events;
} sbprof_pcarch_info;

static sbprof_pcarch_info pcarch_info[2] = {
  /* SB1 */
  {
    SB1_default_slots,
    sizeof(SB1_default_slots) / sizeof(sbprof_default_slot),
    SB1_event_info,
    sizeof(SB1_event_info) / sizeof (sbprof_event_entry_t)
  },
  /* SB1R2 */
  {
    SB1R2_default_slots,
    sizeof(SB1R2_default_slots) / sizeof(sbprof_default_slot),
    SB1R2_event_info,
    sizeof(SB1R2_event_info) / sizeof (sbprof_event_entry_t)
  }
};

/* Initializes cpudata.slot[s] by looking up the values for event_names in
 * table.  The periods in slots counters will be stored in log base 2.
 * Requires: table has num_events entries
 */
void parse_mux_slot(const sbprof_default_slot event_names, int s,
		    const sbprof_event_entry_t *table, int num_events)
{
  int i = 0;

  for (i = 0; i < MAX_COUNTERS; i++) {
    event_counter_state_t *ecs = &cpudata.slot[s][i];
    const sbprof_event_entry_t *ev =
            event_lookup(event_names[i], table, num_events);
    if (ev == NULL) {
      printf("sbprof fatal error: %s is not an event name\n", event_names[i]);
      while (1) ;
    }
    /* Check that event e can be counted on counter i */
    if (!(ev->counters & (1 << i))) {
      printf("sbprof fatal error: counter %d cannot count %s\n", i,
	     event_names[i]);
      while (1) ;
    }
    ecs->total = 0;
    ecs->dropped = 0;
    ecs->value = 0;
    ecs->start_period_low = 0;
    ecs->start_period_high = ev->period;
    ecs->event = ev->event;
    ecs->hwcode = ev->hwcode;
    ecs->flags = ECS_INUSE | (ev->use_ptr ? ECS_USE_PTR : 0);
  }
}

/* Converts the periods in slots[] from log base 2 to 64-bit uints.
   The period of event e will be reduced by the factor
     (# of times event e appears in a slot) / nslots */

static
void adjust_slot_periods(int nslots)
{
#define EVENT_MAX 256 /* exceed max event number */
  int occur[EVENT_MAX];
  int e, s, c;
  for (e = 0; e < EVENT_MAX; e++) occur[e] = 0;
  for (s = 0; s < nslots; s++) {
    for (c = 0; c < MAX_COUNTERS; c++) {
      e = cpudata.slot[s][c].event;
      occur[e]++;
    }
  }
  for (s = 0; s < nslots; s++) {
    for (c = 0; c < MAX_COUNTERS; c++) {
      int e = cpudata.slot[s][c].event;
      u_int64_t period = 1ULL << cpudata.slot[s][c].start_period_high;

      period *= occur[e];
      period += nslots/2;
      period /= nslots;
#define PERIOD_2_CNTVALUE(p) ((- (int64_t) (p)) & 0x000000ffffffffffULL)
      period = PERIOD_2_CNTVALUE(period);
      /* adjust period by 1/2 ripple so that the mean period will be
	 centered around the current value */
      period -= RIPPLE_SIZE / 2;
      e = cpudata.slot[s][c].event;
      cpudata.slot[s][c].start_period_low = (u_int32_t) period;
      cpudata.slot[s][c].start_period_high = (u_int8_t) (period >> 32);
    }
  }
}

#if 0
/* Align byte pointer "p" to next multiple of "n" bytes */
#define ALIGN(p,n) do { int _n = (n); \
                        int _x = (_n-1) & (int) (p); \
                        if (_x) (p) += _n - _x; } \
                   while (0)
#else
#define ALIGN(p,n) do { long _n = (n); \
                        long _x = (_n-1) & (long)(p); \
                        if (_x) (p) += _n - _x; } \
                   while (0)
#endif

/* Initialize the log and config events to monitor. */
void sbprof_config_log() {
  u_int32_t cpuid;
  u_int8_t pcarch = 0; /* SB1 */
  u_int64_t system_rev = *((u_int64_t *) PHYS_TO_K1(0x10020000));
  u_int8_t *p = cpudata.buf;
  sbprof_pcarch_info *pi = &pcarch_info[0];
  int s;

  __asm__ __volatile__ ("mfc0	%0, $15;" : "=r"(cpuid));

  if ((cpuid & 0xff) == 2) {
    /* A pass 2 cpu */
    pi = &pcarch_info[1];
    pcarch = 4; /* SB1R2 */
  }

  /* Enter pcarch info for pkludge */
  *p++ = 4; /* OS_AMECFE */
  *p++ = pcarch;
  *p++ = 1;
#if defined(MIPSEL)
  *p++ = TRUE;
#elif defined(MIPSEB)
  *p++ = FALSE;
#else
#error "unexpected endianness"
#endif
  *p++ = 0;
  *p++ = 1;
  *p++ = 2; /* major rev */
  *p++ = 0; /* minor rev */
  *(u_int64_t *) p = system_rev; p += 8;
  *(u_int32_t *) p = cpuid;      p += 4;
  p += 4; /* align to 8 bytes */

  /* set up mux slots */
  if (pi->num_slots > MAX_SLOTS) {
    printf("sbprof: excess slots ignored\n");
    pi->num_slots = MAX_SLOTS;
  }
  for (s = 0; s < pi->num_slots; s++) {
    parse_mux_slot(pi->default_slots[s], s, pi->event_table, pi->num_events);
  }

  /* Enter config info in log */
  *p++ = DP_D_CONFIG;
  *p++ = (u_int8_t) pi->num_slots;
  for (s = 0; s < pi->num_slots; s++) {
    int c;
    for (c = 0; c < MAX_COUNTERS; c++) {
      *p++ = cpudata.slot[s][c].event;
      *p++ = (u_int8_t) cpudata.slot[s][c].start_period_high;
    }
  }
  ALIGN(p, 8);
  cpudata.nslots = pi->num_slots;
  
  /* Tell pgather to expect -kf for kernel image */
  *p++ = DP_D_PID_IMAGE;
  p += 3;
  *(u_int32_t *) p = -1; p += 4;
  *(u_int64_t *) p = 1;  p += 8;
  *(u_int64_t *) p = 2;  p += 8;
  *(u_int32_t *) p = 3;  p += 4;
  memcpy((void *) p, (void *) "4", 2);
  p += MAX_IMAGE_NAME;
  
  cpudata.next =  p - &cpudata.buf[0];
  
  adjust_slot_periods(pi->num_slots);
}



static void unarm_counters(void);

/* Upon return, Status.IE is clear and interrupts are blocked */
static void disable_intr(void);

/* Upon return, Status.IE is set, but there may be a few
   cycles more delay before interrupts are truly enabled. */
static void enable_intr(void);

static __inline__
void disable_intr()
{
  /* Clear the IE bit of the status register */
  __asm__ __volatile__
    ( ".set push;"
      ".set noreorder;"
      ".set mips64;"
      "mfc0 $8, $12;"		/* t0 = STATUS */
      "subu $9, $0, 2;" 	/* t1 = 0xfffffffe */
      "and  $8, $8, $9;"	/* t0 = STATUS & ~1 */
      "mtc0 $8, $12;"		/* STATUS = t0 */
      "ssnop;"			/* Wait for cleared IE bit to take effect */
      "ssnop;"			/* 5 cycles may be excessive */
      "ssnop;"
      "ssnop;"
      "ssnop;"
      ".set pop;"
      : /* no outputs */
      : /* no inputs */
      : /* modifies: */ "$8", "$9"
      );
}

static __inline__
void enable_intr()
{
  /* Set the IE bit of the status register */
  __asm__ __volatile__
    ( ".set push;"
      ".set noreorder;"
      ".set mips64;"
      "mfc0 $8, $12;"		/* t0 = STATUS */
      "or   $8, $8, 1;"		/* t0 = STATUS | 1 */
      "mtc0 $8, $12;"		/* STATUS = t0 */
      ".set pop;"
      : /* no outputs */
      : /* no inputs */
      : /* modifies: */ "$8"
      );
}

/* Writes to cp0 regs happen at pipestage 8; interrupts at pipestage 4.
   3 or 4 cycles of delay are needed between writing a cp0 reg to disable
   interrupts or a particular counter before the write takes effect. */
#define PAUSE __asm__ __volatile__ \
        (".set push; .set mips64; ssnop; ssnop; ssnop; ssnop; .set pop;")


/* Puts the current count of cpu counter "counter_num" into "*dest" */
#define READ_COUNTER(counter_num,dest)    \
  __asm__ __volatile__ (".set push;"	  \
			".set mips64;"	  \
                        ".set noreorder;" \
			"dmfc0 $8,$25,%1;" \
			"sd    $8,0(%0);" \
                        ".set pop;"	  \
			:/* no output */ \
			: "r" (dest), "i" ((counter_num)*2+1) \
			:/* modifies */ "$8");

/* Puts "*val" into the current count of cpu counter "counter_num" */
#define WRITE_COUNTER(counter_num,val)    \
  __asm__ __volatile__ (".set push;"	  \
			".set mips64;"	  \
                        ".set noreorder;" \
			"ld    $4,0(%0);" \
			"dmtc0 $4,$25,%1;" \
                        ".set pop;"	  \
			:/* no output */  \
			: "r" (val), "i" ((counter_num)*2+1) \
			:/* modifies */ "$4");

/* Puts "val" into the current control reg of cpu counter "counter_num" */
#define WRITE_COUNTER_SEL(counter_num,val) \
  __asm__ __volatile__ (".set push;"	  \
			".set mips64;"	  \
			"mtc0 %0,$25,%1;"  \
                        ".set pop;"	  \
			:/* no output */:  "r" (val), "i" ((counter_num)*2));

#define ARM_COUNTER(counter_num, code, value) \
  /* disable counter; set value to *value; enable counter */ \
  /* When called with counters frozen, perhaps the disable is unnecessary. */ \
  WRITE_COUNTER_SEL(counter_num, 1 << 30); \
  WRITE_COUNTER(counter_num, value); \
  WRITE_COUNTER_SEL(counter_num, code)

/* Sets counter "counter_num" to monitor event "hwcode" starting from initial
   value "value" */
static __inline__
void arm_counter(u_int8_t counter_num, u_int32_t hwcode, u_int64_t value)
{
  u_int32_t code;
  hwcode &= 0x3f;
  code =
    (1 << 30) |		/* freeze enable */
    (hwcode << 5) |	/* event code */
    0x1f;		/* interrupt enable, count always */

  //printf("arm_counter %d 0x%d 0x%llx\n", counter_num, hwcode, value);
  switch(counter_num) {
    case 0: ARM_COUNTER(0, code, &value); break;
    case 1: ARM_COUNTER(1, code, &value); break;
    case 2: ARM_COUNTER(2, code, &value); break;
    case 3: ARM_COUNTER(3, code, &value); break;
    default: panic("arm_counter: unknown counter");
  }
}

#undef ARM_COUNTER

#define UNARM_COUNTER(counter_num) \
  WRITE_COUNTER_SEL(counter_num, 1 << 30); /* freeze enable */ \
  /* Make sure inactive counter doesn't look like it overflowed to prfcntintr \
     handler */ \
  WRITE_COUNTER(counter_num, &zero);

static __inline__
void unarm_counters(void)
{
  u_int64_t zero = 0;
  UNARM_COUNTER(0);
  UNARM_COUNTER(1);
  UNARM_COUNTER(2);
  UNARM_COUNTER(3);
  PAUSE;
}

#undef UNARM_COUNTER

static inline void record_sample(u_int32_t pc, u_int32_t ptr,
				 int counter_num, u_int8_t use_ptr)
{
  pc = ((counter_num == 1) && (use_ptr != 0)) ? ptr : pc;

  if (cpudata.next < sizeof(cpudata.buf) - 4) {
    *(u_int32_t *) &cpudata.buf[cpudata.next] = pc | counter_num;
    cpudata.next += 4;
  }
}

/* Called whenever a level 5 interrupt is raised.  Freeze-enable will freeze
   counters IFF a counter has overflowed. */
static void sbprof_cpu_intr(u_int64_t restartpc)
{
  int counter_num;
  u_int32_t pc = (u_int32_t) restartpc;
  u_int32_t ptr;
  u_int64_t msr[MAX_COUNTERS];
  cpudata_t *cpud = &cpudata;
  int cycles_tick;

  __asm__ __volatile__ (".set push;"
			".set mips64;"
			"mfc0 %0, $22,0;"
			".set pop;"
			: "=r"(ptr));
  READ_COUNTER(0, &msr[0]);
  READ_COUNTER(1, &msr[1]);
  READ_COUNTER(2, &msr[2]);
  READ_COUNTER(3, &msr[3]);

  ptr = (ptr & 1) ? ptr + 3 : ptr;

  cycles_tick = (msr[0] & (1ULL << 40)) ? 1 : 0;
  for (counter_num = 0; counter_num < 4; counter_num++) {
    if (msr[counter_num] & (1ULL << 40)) {
      u_int64_t next_period = cpud->cur_slot[counter_num].start_period;
      int index = (pc >> 3) % RIPPLE_TABLE_SIZE;

      record_sample(pc, ptr, counter_num, cpud->use_ptr);
      next_period += ripple[index];
      /* N.B. start_period was adjusted by 1/2 RIPPLE_SIZE in
	 ioctl SBPROF_START to center ripple around desired period */
      ripple[index] += RIPPLE_GENERATOR;
      ripple[index] &= RIPPLE_SIZE - 1;
      msr[counter_num] = next_period;
    }
  }

  if (cycles_tick) {
    cpud->threshold--;
    if ((cpud->nslots > 1) && !cpud->threshold) {
      cpud->threshold = 16;
      slot_exit(cpud, cpud->slotid, msr);
      slot_enter(cpud,
		 (cpud->slotid + 1) % cpud->nslots,
		 -1,
		 0);
      return;
    }
  }

  /* Case "no counter overflowed":
       The writes below will discard events that occurred during most of
       this routine.
     Case "some counter overflowed":
       For counters that overflowed, these writes unfreeze the counters and
       clear the interrupt request.  For counters that didn't overflow, these
       writes are NOPs since they write back the current value.
       Once the last overflowed counter is written, subsequent writes discard
       some events that might occur as the writes are happening.
  */

  WRITE_COUNTER(0, &msr[0]);
  WRITE_COUNTER(1, &msr[1]);
  WRITE_COUNTER(2, &msr[2]);
  WRITE_COUNTER(3, &msr[3]);
}

#define PERF_COUNT_INTR 0x8000   /* Mask for level 5 intr in status & cause */
#define TRACEBUF_INTR   0x4000   /* Mask for level 4 intr in status & cause */
#undef CAUSE_BD
#undef CAUSE_IV
#define CAUSE_BD  (1 << 31)  /* Mask for BD bit of cause */
#define CAUSE_IV  (1 << 23)  /* Mask for IV bit of & cause */
void sbprof_trapfunc(unsigned long type, regtype *gprs)
{
  if (type == TRAP_INTERRUPT) {
    u_int32_t status, cause;
    u_int64_t epc;

    __asm__ __volatile__ ("mfc0	%0, $12;" : "=r"(status));
    __asm__ __volatile__ ("mfc0	%0, $13;" : "=r"(cause));
    __asm__ __volatile__ ("dmfc0 %0, $14;" : "=r"(epc));
    if (cause & PERF_COUNT_INTR) {
      u_int64_t restart_pc = epc + (cause & CAUSE_BD ? 4 : 0);
    
      sbprof_cpu_intr(restart_pc);
      /* Knock down count/compare */
      __asm__ __volatile__ ("mfc0 $8, $9;"	// read count
			    "subu $8, $8, 1;"	// decrement
			    "mtc0 $8, $11;"	// write compare
			    ::: /* modifies */ "$8");
    }
    if (cause & TRACEBUF_INTR) {
      extern void sbprof_tb_intr();
      sbprof_tb_intr();
    }
  } else {
#if 0
    static _ame_trapfunc *real_trap = &_trap;
    (*real_trap)(type, gprs);
    while (1) ;
#else
	printf("what is this one???\n");
#endif
  }
}

/* If cpud's curbuf has an EVENT_HDR message lacking nevents, fills in nevents.
   (Unless nevents would be 0, in which case the EVENT_HDR message is
   removed from the buffer.)
   Sets last_event to -1.
*/
static void backfill_nevents(cpudata_t *cpud)
{
  if (cpud->last_event >= 0) {
    if (cpud->next == cpud->last_event) {
      cpud->next -= EVENT_HDR_MSG_SIZE;
    } else {
      int idx = cpud->last_event - 2;
      u_int16_t n = (cpud->next - cpud->last_event) / EVENT_MSG_SIZE;
      u_int16_t *p;
      p = (u_int16_t *) &cpud->buf[idx];
      *p = n;
    }
    cpud->last_event = -1;
  }
}

/***************************************************
 * Interrupt Invariant (holds whenver interrupts enabled)
 *  XOR
 *   o last_event == -1
 *   o cpudata.buf has a previous EVENT_HDR message
 *
 ***************************************************/

/* Requires: cpud->buf has room for event header and event
             p points to next buf entry
   Effect:   adds event header */
static inline
void enter_event_hdr(cpudata_t *cpud, u_int8_t *p)
{
  *p++ = DP_D_EVENTS;
  cpud->next += EVENT_HDR_MSG_SIZE;
  cpud->last_event = cpud->next;
}

/* Requires: cpud->last_event == -1
   Effect:   Enters a new_context message followed by an event_hdr message */
static inline
void enter_context(cpudata_t *cpud, u_int32_t pid, u_int8_t cpuid)
{
  u_int8_t *p = (u_int8_t *) &cpud->buf[cpud->next];

  *p++ = DP_D_CONTEXT;
  *p++ = cpuid;
  *p++ = cpud->slotid;
  p += 1; /* align to 4 bytes */
  *((u_int32_t *) p) = pid; p += 4;
  cpud->next += NEW_CONTEXT_MSG_SIZE;
  enter_event_hdr(cpud, p);
}  

/* Effect: If there's enough room in a single, non-full cpud buffer for a
           new_context message followed by a pc sample, enters the new_context
	   message + event_hdr message and returns 0; else returns 1.
 */
static inline
int report_new_context(cpudata_t *cpud, u_int32_t pid, u_int8_t cpuid)
{
  backfill_nevents(cpud);
  if (cpud->next > sizeof(cpudata.buf) - (NEW_CONTEXT_MSG_SIZE +
					  EVENT_HDR_MSG_SIZE +
					  EVENT_MSG_SIZE)) {
    return 1;
  }
  enter_context(cpud, pid, cpuid);
  return 0;
}

static __inline__ u_int64_t u64max(u_int64_t a, u_int64_t b) {
  return (a > b) ? a : b;
}

static void slot_enter(cpudata_t *cpup, u_int8_t slotid, u_int32_t pid,
		       u_int8_t cpuid)
{
  int i;
  event_counter_state_t *ecs = cpup->slot[slotid];

  /* XXX add prefetching, avoid branch? */
  for (i = 0; i < MAX_COUNTERS; i++) {
    /* Zero the per-slot-activation counts */
    cpup->cur_slot[i].total = 0;
    cpup->cur_slot[i].dropped = 0;
    if (ecs[i].flags & ECS_INUSE) {
      /* Compute value for re-arming counter */
      cpup->cur_slot[i].start_period =
	((u_int64_t) ecs[i].start_period_low) |
	(((u_int64_t) ecs[i].start_period_high) << 32);
      /* Set up counters with values from exit of last activation */
      arm_counter(i, ecs[i].hwcode, u64max(ecs[i].value,
					   cpup->cur_slot[i].start_period));
    }
  }
  cpup->slotid = slotid;
  cpup->use_ptr = ecs[1].flags & ECS_USE_PTR;
  report_new_context(cpup, pid, cpuid);
}

static void slot_exit(cpudata_t *cpup, u_int8_t slotid, u_int64_t *counter_val)
{
  int i;
  event_counter_state_t *ecs = cpup->slot[slotid];

  /* XXX add prefetching, avoid branch? */
  for (i = 0; i < MAX_COUNTERS; i++) {
    if (ecs[i].flags & ECS_INUSE) {
      /* Export the per-slot-activation counts */
      ecs[i].total += cpup->cur_slot[i].total;
      ecs[i].dropped += cpup->cur_slot[i].dropped;
      /* Remember where to start off next time */
      ecs[i].value = counter_val[i];
    }
  }
  unarm_counters();
}

/* Vector exceptions/interrupts to our handler */
_ame_trapfunc *_trap_func = &sbprof_trapfunc;

void
sbprof_nfx_intr(
	u_int32_t	sr,
	u_int32_t	cause,
	long	epc
)
{
	sbprof_trapfunc(TRAP_INTERRUPT, NULL);
}

void sbprof_start()
{
  int i;
	extern int32_t intr_registerHandler(u_int32_t ipl, void (* func)());
  
  if (sizeof(u_int64_t) != 8 ||
      sizeof(u_int32_t) != 4 ||
      sizeof(u_int8_t) != 1) {
    printf("sbprof_start: need to fix typedefs in sbprof.h!!\n");
    exit(1);
  }

  if (((int) &cpudata.buf) & 7) {
    printf("sbprof: fatal error -- buf not 8-byte aligned\n");
  }

  /* Initialize ripple start points to random values */
  for (i = 0; i < RIPPLE_TABLE_SIZE; i++) {
    static u_int64_t rand = 0;
    ripple[i] = rand & (RIPPLE_SIZE - 1);
    rand = (rand * 0x5deece66dULL + 0xb) & 0xffffffffffffULL;
  }
  
  sbprof_config_log();
  cpudata.threshold = 16;
  cpudata.slotid = 0;
  cpudata.last_event = -1;

  disable_intr();
  unarm_counters();
  slot_enter(&cpudata, 0, -1, 0);
  
#if 0
  /* Unmask I5/IM7 interrupts */
  __asm__ __volatile__ ("mfc0 $8, $12;"		// read status
			"ori  $8, $8, %0;"	// new bits
			"mtc0 $8, $12;"		// write status
			: /* no outputs */
			: "i"(PERF_COUNT_INTR)
			: /* modifies */ "$8");
  /* Set IV in cause to send interrupts to distinct vector */
  __asm__ __volatile__ ("mfc0 $8, $13;"		// read cause
			"or  $8, $8, %0;"	// new bits
			"mtc0 $8, $13;"		// write cause
			: /* no outputs */
			: "r"(CAUSE_IV)
			: /* modifies */ "$8");
#else
	intr_registerHandler(7,sbprof_nfx_intr);
#endif
  enable_intr();
}

void sbprof_stop()
{
  u_int64_t counter_val[MAX_COUNTERS];
    extern int32_t intr_unregisterHandler(u_int32_t);


  disable_intr();
  unarm_counters();
  enable_intr();
  /* Wait for any last interrupts */
  PAUSE;
  PAUSE;

  READ_COUNTER(0, &counter_val[0]);
  READ_COUNTER(1, &counter_val[1]);
  READ_COUNTER(2, &counter_val[2]);
  READ_COUNTER(3, &counter_val[3]);
  slot_exit(&cpudata, cpudata.slotid, counter_val);
  backfill_nevents(&cpudata);

	intr_unregisterHandler(7);

	{
		extern u_int64_t	sbprof_addr;
		extern u_int32_t	sbprof_len;
		sbprof_addr = (u_int64_t)&cpudata.buf;
		sbprof_len = cpudata.next;
	}
//  printf("\nsbprof dump vol1:xxx %p 0x%x\n", &cpudata.buf, cpudata.next);
}

/************************************************************************
 * Support for ZBbus sampling using the trace buffer
 *
 * We use the SCD performance counter interrupt, caused by a Zclk counter
 * overflow, to trigger the start of tracing.
 *
 * We set the trace buffer to sample everything and freeze on
 * overflow.
 *
 * We map the interrupt for trace_buffer_freeze to handle it on CPU 0.
 * 
 ************************************************************************/

#define SCD_PERFCNT_INT  14
#define TRACE_FREEZE_INT 15
#define SCD_PERFCNT_INT_MAP_0  (0x10020200 + SCD_PERFCNT_INT*8)
#define TRACE_FREEZE_INT_MAP_0 (0x10020200 + TRACE_FREEZE_INT*8)
//#define INTERRUPT_MASK_0 (A_IMR_REGISTER(0,R_IMR_INTERRUPT_MASK))
#define INTERRUPT_MASK_0  0x10020028
#define INTERRUPT_TRACE_0 0x10020038
#define A_PERF_CNT_CFG    0x100204c0
#define A_PERF_CNT_1      0x100204d8


#define V_SCD_TRACE_CFG_RESET       M_SCD_TRACE_CFG_RESET
#define V_SCD_TRACE_CFG_START_READ  M_SCD_TRACE_CFG_START_READ
#define V_SCD_TRACE_CFG_START	    M_SCD_TRACE_CFG_START	
#define V_SCD_TRACE_CFG_STOP	    M_SCD_TRACE_CFG_STOP	
#define V_SCD_TRACE_CFG_FREEZE	    M_SCD_TRACE_CFG_FREEZE	
#define V_SCD_TRACE_CFG_FREEZE_FULL M_SCD_TRACE_CFG_FREEZE_FULL
#define V_SCD_TRACE_CFG_DEBUG_FULL  M_SCD_TRACE_CFG_DEBUG_FULL
#define V_SCD_TRACE_CFG_FULL	    M_SCD_TRACE_CFG_FULL	
#define V_SCD_TRACE_CFG_CUR_ADDR(x) _SB_MAKEVALUE(x,S_SCD_TRACE_CFG_CUR_ADDR)

#define G_SCD_TRACE_CFG_CUR_ADDR(x) _SB_GETVALUE(x,S_SCD_TRACE_CFG_CUR_ADDR,M_SCD_TRACE_CFG_CUR_ADDR)


/* 100 trace buffer samples per second on a 500 Mhz 1250 */
#define TB_PERIOD 2500000ULL

void arm_tb()
{
  unsigned long long next = (1ULL << 40) - TB_PERIOD;
  /* Generate an SCD_PERFCNT interrupt in TB_PERIOD Zclks to trigger start
     of trace.
     XXX vary sampling period */
  SBWRITECSR(A_PERF_CNT_1, 0);
  /* Unfortunately, in Pass 2 we must clear all counters to knock down
     a previous interrupt request.  This means that bus profiling
     writes ALL of the SCD perf counter values. */
  SBWRITECSR(A_PERF_CNT_CFG,
	     M_SPC_CFG_ENABLE | // enable counting
	     M_SPC_CFG_CLEAR |	// clear all counters
	     V_SPC_CFG_SRC1(1));// counter 1 counts cycles
  SBWRITECSR(A_PERF_CNT_1, next);
  /* Reset the trace buffer */
  SBWRITECSR(A_SCD_TRACE_CFG, V_SCD_TRACE_CFG_RESET);
  SBWRITECSR(A_SCD_TRACE_CFG, V_SCD_TRACE_CFG_FREEZE_FULL);
}

void sbprof_tb_intr()
{
  int i;
  if (next_tb_sample < MAX_TB_SAMPLES) {
    /* XXX should use XKPHYS to make writes bypass L2 */
    unsigned long long *p = sbprof_tbbuf[next_tb_sample++];
    /* Read out trace */
    SBWRITECSR(A_SCD_TRACE_CFG, V_SCD_TRACE_CFG_START_READ);
    __asm__ __volatile__ ("sync" : : : "memory");
    /* Loop runs backwards because bundles are read out in reverse order */
    for (i = 256 * 6; i > 0; i -= 6) {
      // Subscripts decrease to put bundle in the order
      //   t0 lo, t0 hi, t1 lo, t1 hi, t2 lo, t2 hi
      p[i-1] = SBREADCSR(A_SCD_TRACE_READ);	// read t2 hi
      p[i-2] = SBREADCSR(A_SCD_TRACE_READ);	// read t2 lo
      p[i-3] = SBREADCSR(A_SCD_TRACE_READ);	// read t1 hi
      p[i-4] = SBREADCSR(A_SCD_TRACE_READ);	// read t1 lo
      p[i-5] = SBREADCSR(A_SCD_TRACE_READ);	// read t0 hi
      p[i-6] = SBREADCSR(A_SCD_TRACE_READ);	// read t0 lo
    }
    arm_tb();	// knock down current interrupt and get another one later
  } else {
    /* No more trace buffer samples */
    SBWRITECSR(A_SCD_TRACE_CFG, V_SCD_TRACE_CFG_RESET);
  }
}


void sbprof_zbprof_start()
{
  disable_intr();
  
  /* Map trace_freeze_int to TRACEBUF_INTR (I4/IM6) on CPU 0 */
  SBWRITECSR(TRACE_FREEZE_INT_MAP_0, 4);
  /* Map scd_perfcnt_int to (I3/IM5) on CPU 0, which will ignore it */
  SBWRITECSR(SCD_PERFCNT_INT_MAP_0, 3);
  /* Allow the scd_perfcnt_int to trigger trace events */
  SBWRITECSR(INTERRUPT_TRACE_0, (1ULL << SCD_PERFCNT_INT));
  
  /* Unmask the interrupts at SCD level */
  SBWRITECSR(INTERRUPT_MASK_0,
	     SBREADCSR(INTERRUPT_MASK_0)
	     & ~((1ULL << TRACE_FREEZE_INT) |
		 (1ULL << SCD_PERFCNT_INT)));

  /* Unmask only I4/IM6 interrupts at CPU level */
  __asm__ __volatile__ ("mfc0 $8, $12;"		// read status
			"ori  $8, $8, %0;"	// new bits
			"mtc0 $8, $12;"		// write status
			: /* no outputs */
			: "i"(TRACEBUF_INTR)
			: /* modifies */ "$8");

  /* Set IV in cause to send interrupts to distinct vector */
  __asm__ __volatile__ ("mfc0 $8, $13;"		// read cause
			"or  $8, $8, %0;"	// new bits
			"mtc0 $8, $13;"		// write cause
			: /* no outputs */
			: "r"(CAUSE_IV)
			: /* modifies */ "$8");

  /* Initialize address traps */
  SBWRITECSR(A_ADDR_TRAP_UP_0, 0);
  SBWRITECSR(A_ADDR_TRAP_UP_1, 0);
  SBWRITECSR(A_ADDR_TRAP_UP_2, 0);
  SBWRITECSR(A_ADDR_TRAP_UP_3, 0);
	      
  SBWRITECSR(A_ADDR_TRAP_DOWN_0, 0);
  SBWRITECSR(A_ADDR_TRAP_DOWN_1, 0);
  SBWRITECSR(A_ADDR_TRAP_DOWN_2, 0);
  SBWRITECSR(A_ADDR_TRAP_DOWN_3, 0);
	      
  SBWRITECSR(A_ADDR_TRAP_CFG_0, 0);
  SBWRITECSR(A_ADDR_TRAP_CFG_1, 0);
  SBWRITECSR(A_ADDR_TRAP_CFG_2, 0);
  SBWRITECSR(A_ADDR_TRAP_CFG_3, 0);

  /* Initialize Trace Event 0-7 */
  //                              when interrupt
  SBWRITECSR(A_SCD_TRACE_EVENT_0, (1 << 7));
  SBWRITECSR(A_SCD_TRACE_EVENT_1, 0);
  SBWRITECSR(A_SCD_TRACE_EVENT_2, 0);
  SBWRITECSR(A_SCD_TRACE_EVENT_3, 0);
  SBWRITECSR(A_SCD_TRACE_EVENT_4, 0);
  SBWRITECSR(A_SCD_TRACE_EVENT_5, 0);
  SBWRITECSR(A_SCD_TRACE_EVENT_6, 0);
  SBWRITECSR(A_SCD_TRACE_EVENT_7, 0);

  /* Initialize Trace Sequence 0-7 */
  //                                   Start on event 0 (interrupt)
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_0, (1 << 16)|0x0fff);
  //                        dsamp when d used | asamp when a used
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_1, (1 << 19)| (1 << 18)|0xffff);
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_2, 0);
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_3, 0);
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_4, 0);
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_5, 0);
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_6, 0);
  SBWRITECSR(A_SCD_TRACE_SEQUENCE_7, 0);

  arm_tb();
  
  //printf("start: trace_cfg is 0x%llx\n", SBREADCSR(A_SCD_TRACE_CFG));

  enable_intr();
}

void sbprof_zbprof_stop()
{
  disable_intr();

  /* mask the interrupt at SCD level */
  SBWRITECSR(INTERRUPT_MASK_0,
	     SBREADCSR(INTERRUPT_MASK_0) | (1ULL << TRACE_FREEZE_INT));

  enable_intr();
  /* Wait for any last interrupts */
  PAUSE;
  PAUSE;

  printf("ifconfig -auto eth0 ; save server_name:path/zb-samples %p 0x%lx\n",
	 &sbprof_tbbuf, next_tb_sample * sizeof(sbprof_tbbuf[0]));
}

#endif /* _PROFILING_ */

/* Map PLL_DIV to HZ assuming a 100Mhz external PLL */
#define Mhz * 1000000
#define UNDEF 0
static unsigned long long pll_div_to_hz[32] = {
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF,
  200 Mhz,
  250 Mhz,
  300 Mhz,
  350 Mhz,
  400 Mhz,
  450 Mhz,
  500 Mhz,
  550 Mhz,
  600 Mhz,
  650 Mhz,
  700 Mhz,
  750 Mhz,
  800 Mhz,
  850 Mhz,
  900 Mhz,
  950 Mhz,
  1000 Mhz,
  1050 Mhz,
  1100 Mhz,
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF,
  UNDEF
};
  

unsigned long long cpu_hz()
{
  unsigned long long system_config;
  int pll_div;
  __asm__ __volatile__ (".set push;"
                        ".set mips64;"
                        "la   $8, 0xb0020008;" /* SCD system_cfg */ \
                        "ld   %0, 0($8);"
                        ".set pop"
			: /* outputs */ "=r"(system_config)
			: /* inputs */
			: /* modifies */ "$8" );
  pll_div = (system_config >> 7) & 0x1f;
  return pll_div_to_hz[pll_div];
}
