/*
 *-----------------------------------------------------------------
 *
 * Name:	sbm-mgmt.c
 *
 * RCS:  $Id: sbm-mgmt.c,v 1.27 2002/11/01 19:30:12 rick Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description:	COD Manipulation commands
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/8/02
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>

#ifndef SBM_REV2
#error Compiled without SBM_REV2
#endif

#ifdef NFP_TXRX

#include <exports/eee-api.h>
#include <exports/sbm-api.h>
#include <exports/cm-api.h>
#ifndef LINUX_TEST
#include <exports/cpu-api.h>
#endif
#include <exports/malloc-api.h>
#include "sbm.h"
#ifdef USE_SAC_CHECKSUM
#include <exports/scsi-api.h>
#endif
#include <exports/dp-proxy.h>

/*-----------------------------------------------------------------
 * Name:	sbm_freeCOD()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/8/02
 *
 *-----------------------------------------------------------------
 */
int32
sbm_freeCOD(
	nfx_cod_t	cod,
	uint32		update_stats
)
{	
	eee_descPtr_t	edesc;
#ifdef USE_SAC_CHECKSUM
	uint32			calc;
#endif

	if (sbm_verifyIsCODInSac(cod, 1)) {

		edesc = sbm_getDescFromCOD(cod);

		if (!sbm_verifyIsPktOk(edesc))
			panic("sbm_freeCOD, verifyPkt failed, edesc = %p\n", edesc);

#ifdef USE_SAC_CHECKSUM
		if (edesc->hdr.attr != (calc = sbm_verifyCalcPktChecksum(edesc))) {
			SBM_MGMT_DEBUG(("chksum failed, stored = 0x%x, calc = 0x%x\n",
                                                        edesc->hdr.attr, calc));
			sbm_stats.chksum_verify_error++;
			sbm_stats.chksum_verify_error_total++;
			return;
		}
#endif
		sbm_verifyAssignPktAttribute(edesc, EEE_ATTR_SAC_COD_FREE_OK);

		sbm_stats.num_bufs -= sbm_getNumBufsInPkt(edesc);
		sbm_stats.tot_bytes -= EEE_GET_PKT_LEN(edesc);

		eee_freePacket(edesc);

		if (update_stats) {
			sbm_stats.cods_freed++;
			sbm_stats.cods_freed_total++;
			sbm_stats.cods_present--;
		}
	}
	else {
		SBM_MGMT_DEBUG(("cod not in SAC\n"));
        panic("freeCOD, cod not in SAC, 0x%016llx\n", cod);
		return SBM_ERR_INVALID_COD;
	}

	return NFX_OK;

}	/* sbm_freeCOD() */

/*-----------------------------------------------------------------
 * Name:	sbm_createCOD()
 *
 * Description:	The originator of the createCOD message had to allocate a
 *				TXRX descriptor, convert it to a cod, and then send that cod
 *				to the TXRX.  The SBM code will allocate the buffer, create
 *				the EEE packet, initialize the cod with the fill word, and
 *				'insert' it into the SAC.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/8/02
 *
 *-----------------------------------------------------------------
 */
int32
sbm_createCOD(
	nfx_cod_t	cod,
	uint32		len,
	uint32		fill,
	uint32		update_stats
)
{
	eee_descPtr_t	edesc;
	uchar8			* buf;
    uint32          save_len = len;
    uint32          nb = 0;
    uint32          blen;

	edesc = sbm_getDescFromCOD(cod);

	if (!sbm_verifyIsEdescInSac(edesc, 0)) {
		if (len > 8192)
			panic("createCod, invalid len, 0x%x\n", len);

        while (len) {
    		buf = eee_allocateBuffer(EEE_SHARED_POOL,
                                        EEE_BUF_LARGE,
    									NFX_COS_HIGH);
    		if (buf) {
                if (len < EEE_BUFSIZE_LARGE)
                    blen = len;
                else
                    blen = EEE_BUFSIZE_LARGE;
				if (fill == 0)
					bzero(buf, blen);
				else
	    			memset(buf, fill, blen);
    			edesc->bd[nb+1].buf = buf;
    			edesc->bd[nb+1].len = blen;
    			edesc->bd[nb+1].control = EDESC_BD_LARGE | EDESC_BD_CMD_IN_USE;
                nb++;
                len -= blen;
            }
    		else {
    			SBM_MGMT_DEBUG(("createCOD, cannot allocate large buffer\n"));
				sbm_stats.no_buf_error++;
				sbm_stats.no_buf_error_total++;
    			panic("createCOD, cannot allocate large buffer\n");
    		}
        }
   		edesc->hdr.control = (nb << EDESC_BUFCNT_SHIFT) | save_len;
   		edesc->hdr.offset = 0;
   		edesc->hdr.next = 0;
#ifdef SAC_USE_CHECKSUM
   		edesc->hdr.attr = sbm_verifyCalcPktChecksum(edesc);
#endif

   		sbm_verifyAssignPktAttribute(edesc, EEE_ATTR_SAC_COD_CREATE);
   		sbm_verifyAssignPktValidFreeAttr(edesc);

		sbm_stats.num_bufs += sbm_getNumBufsInPkt(edesc);
		sbm_stats.tot_bytes += EEE_GET_PKT_LEN(edesc);

   		if (update_stats) {
   			sbm_stats.cods_created++;
   			sbm_stats.cods_created_total++;
   			sbm_stats.cods_present++;
   		}
	}
	else {
		SBM_MGMT_DEBUG(("cod, 0x%llx, already exists in sac\n", cod));
		return SBM_ERR_INVALID_COD;
	}

	return NFX_OK;

}	/* sbm_createCOD() */


/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/7/02
 *
 *-----------------------------------------------------------------
 */
void
sbm_freeDesc(
	eee_descPtr_t	edesc
)
{
	sbm_verifyAssignDescAttribute(edesc, EEE_ATTR_SAC_COD_FREE_OK);
	eee_deallocateDesc(edesc);
}

/*-----------------------------------------------------------------
 * Name:	sbm_splitCOD()
 *
 * Description:	Split a src COD into a src COD and a dest COD
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/7/02
 *
 *-----------------------------------------------------------------
 */
int32
sbm_splitCOD(
	nfx_cod_t	src_cod,
	uint32		src_off,
	nfx_cod_t	dest_cod,
	uint32		update_stats
)
{
	eee_descBuf_t	* bd;
	uint32			num_bufs;
	uint32			nb, control;
	eee_descPtr_t	first_edesc;
	eee_descPtr_t	prev_edesc;
	eee_descPtr_t	src_edesc;
	eee_descPtr_t	dest_edesc;
	uint32			save_src_off, src_len;
	uint32			i;
	uchar8			* buf;
	uint32			new_bd_len;
    address_t       ra;
#ifdef USE_SAC_CHECKSUM
	uint32			calc;
#endif

    ra = get_ra();

	if (!sbm_verifyIsCODInSac(src_cod, 1)) {
		SBM_MGMT_DEBUG(("src_cod, 0x%llx, not in SAC\n", src_cod));
		return SBM_ERR_INVALID_COD;
	}
	if (sbm_verifyIsCODInSac(dest_cod, 0)) {
		SBM_MGMT_DEBUG(("dest_cod, 0x%llx, already in SAC\n", src_cod));
		return SBM_ERR_INVALID_COD;
	}

	src_edesc = sbm_getDescFromCOD(src_cod);
	dest_edesc = sbm_getDescFromCOD(dest_cod);
	src_len = EEE_GET_PKT_LEN(src_edesc);
	save_src_off = src_off;

#ifdef USE_SAC_CHECKSUM
	if (src_edesc->hdr.attr != (calc = sbm_verifyCalcPktChecksum(src_edesc))) {
		printf("src_cod chksum failed, stored = 0x%x, calc = 0x%x\n",
													src_edesc->hdr.attr, calc);
		sbm_stats.chksum_verify_error++;
		sbm_stats.chksum_verify_error_total++;
		return SBM_ERR_CHKSUM_FAILED;
	}
#endif

	if (!sbm_verifyIsPktOk(src_edesc))
		panic("sbm_splitCOD, verifyPkt failed, src_edesc = %p\n", src_edesc);

	if (src_off == 0)
		panic("src_off == 0, src_cod = 0x%llx, ra = %p\n", src_cod, ra);

	if (src_off >= src_len) {
		printf("src off err, off = 0x%x, pkt_len = 0x%x\n", src_off, src_len);
		return SBM_ERR_BAD_LEN;
	}

	/*
	 * Find bd where src_off < bd->len
	 */
	first_edesc = src_edesc;
	prev_edesc = NULL;
	while (1) {
		num_bufs = EEE_GET_NUM_BUFS(src_edesc);

		if (src_edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
			bd = &(src_edesc->bd[0]);
		else
			bd = &(src_edesc->bd[1]);

		while ((num_bufs) && (src_off)) {
			if (src_off >= bd->len) {
				src_off -= bd->len;
				bd++;
				num_bufs--;
			}
			else
				break;
		}
		if (num_bufs == 0) {
			prev_edesc = src_edesc;
			src_edesc = src_edesc->hdr.next;
		}
		else if (src_off < bd->len)
			break;
		else {
			prev_edesc = src_edesc;
			src_edesc = src_edesc->hdr.next;
		}
	}

	/*
	 * if src_off == 0, then the split line is on a buffer boundary.  We should
	 *				be able to just move pointers around.
	 * otherwise, we'll need to copy some bytes and then move pointers around.
	 */
	if (src_off == 0) {
		dest_edesc->hdr.next = src_edesc->hdr.next;

		/* copy src_edesc bd's to dest_edesc */
		for (i=0; i<num_bufs; i++) {
			dest_edesc->bd[i+1].buf = bd->buf;
			dest_edesc->bd[i+1].len = bd->len;
			dest_edesc->bd[i+1].control = bd->control;
			bd++;
		}
		/* update dest_edesc header fields */
		dest_edesc->hdr.control = (num_bufs << EDESC_BUFCNT_SHIFT) |
													(src_len - save_src_off);
		/* update src_edesc header fields */
		if (num_bufs == EEE_GET_NUM_BUFS(src_edesc)) {
			prev_edesc->hdr.next = NULL;
			control = first_edesc->hdr.control & ~EDESC_TOTAL_LEN;
			first_edesc->hdr.control = control | save_src_off;
			src_edesc->hdr.next = NULL;
			sbm_freeDesc(src_edesc);
		}
		else {
			src_edesc->hdr.next = NULL;
			if (prev_edesc) {
				nb = EEE_GET_NUM_BUFS(src_edesc) - num_bufs;
				control = src_edesc->hdr.control &
									~(EDESC_TOTAL_LEN | EDESC_BUFCNT_MASK);
				src_edesc->hdr.control = (nb << EDESC_BUFCNT_SHIFT) | control;
				control = first_edesc->hdr.control & ~EDESC_TOTAL_LEN;
				first_edesc->hdr.control = control | save_src_off;
			}
			else {
				nb = EEE_GET_NUM_BUFS(src_edesc) - num_bufs;
				control = first_edesc->hdr.control &
									~(EDESC_TOTAL_LEN | EDESC_BUFCNT_MASK);
				first_edesc->hdr.control = (nb << EDESC_BUFCNT_SHIFT) | control | save_src_off;
			}
		}
	}
	else {
		if (src_off > EEE_BUFSIZE_LARGE)
			panic("Invalid src_off, 0x%x, too big\n", src_off);

		buf = eee_allocateBuffer(EEE_TXRX_POOL, EEE_BUF_LARGE, NFX_COS_HIGH);
		if (!buf) {
			sbm_stats.no_buf_error++;
			sbm_stats.no_buf_error_total++;
			panic("Can not allocate buffer\n");
		}

		sbm_stats.num_bufs += 1;

		new_bd_len = bd->len - src_off;

		/*
		 * copy smaller number of bytes
		 */
		if (new_bd_len <= src_off) {
			memcpy(buf, bd->buf+src_off, new_bd_len);
			sbm_stats.bytes_copied_mgmt += new_bd_len;
			sbm_stats.bytes_copied_mgmt_total += new_bd_len;
			bd->len = src_off;
			dest_edesc->bd[1].buf = buf;
			dest_edesc->bd[1].len = new_bd_len;
			dest_edesc->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
		}
		else {
			memcpy(buf, bd->buf, src_off);
			sbm_stats.bytes_copied_mgmt += src_off;
			sbm_stats.bytes_copied_mgmt_total += src_off;
			dest_edesc->bd[1].buf = bd->buf + src_off;
			dest_edesc->bd[1].len = bd->len - src_off;
			dest_edesc->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
			bd->buf = buf;
			bd->len = src_off;
		}

		dest_edesc->hdr.next = src_edesc->hdr.next;
		src_edesc->hdr.next = NULL;

		bd++;
		for (i=1; i<num_bufs; i++) {
			dest_edesc->bd[i+1].buf = bd->buf;
			dest_edesc->bd[i+1].len = bd->len;
			dest_edesc->bd[i+1].control = bd->control;
			bd++;
		}

		dest_edesc->hdr.control = (num_bufs <<EDESC_BUFCNT_SHIFT) |
														(src_len - save_src_off);
		if (prev_edesc) {
			nb = (EEE_GET_NUM_BUFS(src_edesc) - num_bufs + 1);
			control = src_edesc->hdr.control &
									~(EDESC_TOTAL_LEN | EDESC_BUFCNT_MASK);
			src_edesc->hdr.control = (nb << EDESC_BUFCNT_SHIFT) | control;
			control = first_edesc->hdr.control & ~EDESC_TOTAL_LEN;
			first_edesc->hdr.control = control | save_src_off;
		}
		else {
			nb = (EEE_GET_NUM_BUFS(src_edesc) - num_bufs + 1);
			control = src_edesc->hdr.control &
									~(EDESC_TOTAL_LEN | EDESC_BUFCNT_MASK);
			first_edesc->hdr.control = (nb << EDESC_BUFCNT_SHIFT) | control | save_src_off;
		}
	}

	sbm_verifyAssignPktAttribute(first_edesc, EEE_ATTR_SAC_COD_SPLIT_SRC);
	sbm_verifyAssignPktAttribute(dest_edesc, EEE_ATTR_SAC_COD_SPLIT_DEST);
	sbm_verifyAssignPktValidFreeAttr(dest_edesc);

#ifdef USE_SAC_CHECKSUM
	first_edesc->hdr.attr = sbm_verifyCalcPktChecksum(first_edesc);
	dest_edesc->hdr.attr = sbm_verifyCalcPktChecksum(dest_edesc);
#endif

	if (update_stats) {
		sbm_stats.cods_present++;
		sbm_stats.cods_split++;
		sbm_stats.cods_split_total++;
	}

	return NFX_OK;

}	/* sbm_splitCOD() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *		Deals only with combining the current descriptor and the 'next'
 *		descriptor.  Subsequent calls will deal with more descriptors.
 *		Upon return, the current descriptor's bds will be filled in, the
 *		next descriptors remaining bds will be moved up, and all buf counts,
 *		pkt lens, etc will be properly calculated.
 *
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/28/02
 *
 *-----------------------------------------------------------------
 */
void
sbm_collapseCombineWithNext(
	eee_descPtr_t	ed,
	eee_descBuf_t	* bd,
	uint32			bd_left
)
{

	eee_descPtr_t	next_ed;
	eee_descBuf_t	* next_bd, * next_first_bd;
	uint32			num_bufs, next_num_bufs;

	num_bufs = EEE_GET_NUM_BUFS(ed);

	next_ed = ed->hdr.next;
	next_num_bufs = EEE_GET_NUM_BUFS(next_ed);
	next_bd = &(next_ed->bd[0]);
	if (!(next_ed->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER))
		next_bd++;
	next_first_bd = next_bd;

	if (next_num_bufs <= bd_left) {
		while (next_num_bufs--) {
			bd->control = next_bd->control;
			bd->len = next_bd->len;
			bd->buf = next_bd->buf;
			bd++; next_bd++;
			num_bufs++;
			sbm_stats.desc_collapse++;
			sbm_stats.desc_collapse_total++;
		}
		EEE_SET_NUM_BUFS(ed, num_bufs);
		ed->hdr.next = next_ed->hdr.next;
        sbm_verifyAssignDescAttribute(next_ed, EEE_ATTR_SAC_COD_FREE_OK);
		next_ed->hdr.next = NULL;
		eee_deallocateDesc(next_ed);
	}
	else {
		while (bd_left--) {
			bd->control = next_bd->control;
			bd->len = next_bd->len;
			bd->buf = next_bd->buf;
			bd++; next_bd++;
			num_bufs++;
			next_num_bufs--;
			sbm_stats.desc_collapse++;
			sbm_stats.desc_collapse_total++;
		}
		EEE_SET_NUM_BUFS(ed, num_bufs);
		num_bufs = 0;
		while (next_num_bufs--) {
			next_first_bd->control = next_bd->control;
			next_first_bd->len = next_bd->len;
			next_first_bd->buf = next_bd->buf;
			next_first_bd++; next_bd++;
			num_bufs++;
		}
		EEE_SET_NUM_BUFS(next_ed, num_bufs);
	}
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/30/02
 *
 *-----------------------------------------------------------------
 */
void
sbm_collapseCopyBufData(
	eee_descBuf_t	* prev_bd,
	eee_descBuf_t	* bd,
	eee_descBuf_t	* last_bd
)
{
	uint32	copylen, remain, bufsize;
	uchar8	* save;
	uchar8	* prev_buf_base;
	eee_descBuf_t	* tbd;

	bufsize = EEE_BUFSIZE_LARGE;
	copylen = bd->len;
	prev_buf_base = (uchar8 *)eee_getBufBase(prev_bd->buf);
	remain = bufsize - (prev_bd->buf - prev_buf_base) - prev_bd->len;

printf("collapseCopyBufData, bd->buf = %p\n", bd->buf);
printf("    copylen = 0x%x, remain = 0x%x\n", copylen, remain);
	if (copylen <= remain) {
		memcpy(prev_bd->buf+prev_bd->len, bd->buf, copylen);
		sbm_stats.bytes_copied_collapse += copylen;
		sbm_stats.bytes_copied_collapse_total += copylen;
	}
	else {
		save = (uchar8 *)eee_ramAlloc(prev_bd->len+copylen);
		if (!save)
			return;
		memcpy(save, prev_bd->buf, prev_bd->len);
		memcpy(save+prev_bd->len, bd->buf, copylen);
		memcpy(prev_buf_base, save, prev_bd->len+copylen);
		eee_ramDealloc(save);
		sbm_stats.bytes_copied_collapse += ((copylen + prev_bd->len) * 2);
		sbm_stats.bytes_copied_collapse_total += ((copylen + prev_bd->len) * 2);
	}
	prev_bd->len += copylen;
	sbm_verifyAssignBufAttribute(bd->buf, EEE_ATTR_SAC_COD_FREE_OK);
	eee_deallocateBuffer(bd->buf);
	sbm_stats.num_bufs--;
	tbd = bd;
	while (tbd != last_bd) {
		tbd->control = (tbd+1)->control;
		tbd->len = (tbd+1)->len;
		tbd->buf = (tbd+1)->buf;
		tbd++;
	}
}

/*-----------------------------------------------------------------
 * Name:	sbm_collapseCOD(cod)
 *
 * Description:	changes orientation of cod:
 *
 *				- so that the bd's in a descriptor are used more
 *					efficiently (instead of 1 bd per desc, and then
 *					chaining the descriptors),
 *
 *				- ensuring that the buffers are at least 1/2 full to
 *					fulfill the FP's buffer usage requirement.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	9/28/02
 *
 *-----------------------------------------------------------------
 */
void
sbm_collapseCOD(
	nfx_cod_t	cod
)
{
	eee_descPtr_t	ed;
	eee_descBuf_t	* bd, * prev_bd;
	uint32			bd_left;
	uint32			num_bufs, cod_num_bufs, cod_len, first, last, bufsize;

	ed = sbm_getDescFromCOD(cod);
	if (!sbm_verifyIsPktOk(ed))
		panic("pre-collapse, !OK, ed = %p\n", ed);
		
	/*
	 * consolidate bd's in chained descriptors into the least number of
	 * descriptors possible
	 */
	while (1) {
		if (!ed->hdr.next)
			break;
		num_bufs = EEE_GET_NUM_BUFS(ed);
		bd_left = EEE_MAX_BUFS - num_bufs;
		bd = &(ed->bd[0]);
		if (!(ed->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)) {
			bd++;
			bd_left--;
		}
		while (num_bufs--)
			bd++;
		if (bd_left == 0)
			ed = ed->hdr.next;
		else
			sbm_collapseCombineWithNext(ed, bd, bd_left);
	}

	ed = sbm_getDescFromCOD(cod);
	if (!sbm_verifyIsPktOk(ed))
		panic("post-collapseDesc, !OK, ed = %p\n", ed);

return;

	/*
	 * do appropriate buffer data copies so that the buffers are at
	 * least 1/2 full
	 */
	ed = sbm_getDescFromCOD(cod);
	cod_num_bufs = eee_getPktNumBufs(ed);
	if (cod_num_bufs == 1)
		return;
	cod_len = EEE_GET_PKT_LEN(ed);
	bufsize = EEE_BUFSIZE_LARGE;
	if (cod_len >= (cod_num_bufs * bufsize)/2)
		return;

	while (ed) {
		if (ed->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
			first = 0;
		else
			first = 1;
		num_bufs = EEE_GET_NUM_BUFS(ed);
		last = first + num_bufs - 1;
		for (bd = &(ed->bd[last]); bd > &(ed->bd[first]); bd--) {
			prev_bd = (bd-1);
			if ((bd->len + prev_bd->len) <= bufsize) {
				sbm_collapseCopyBufData(prev_bd, bd, &(ed->bd[last]));
				num_bufs--;
			}
		}
		EEE_SET_NUM_BUFS(ed, num_bufs);
		ed = ed->hdr.next;
	}

	ed = sbm_getDescFromCOD(cod);
	if (!sbm_verifyIsPktOk(ed))
		panic("post-collapseBuf, !OK, ed = %p\n", ed);
}

/*-----------------------------------------------------------------
 * Name:	sbm_concatCOD()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/7/02
 *
 *-----------------------------------------------------------------
 */
int32
sbm_concatCOD(
	nfx_cod_t	dest_cod,
	nfx_cod_t	src_cod,
	uint32		update_stats
)
{
	eee_descPtr_t	src_edesc;
	eee_descPtr_t	dest_edesc;
	eee_descPtr_t	last_edesc;
	uint32			total_len, src_len, dest_len;
	uint32			src_nbufs, dest_nbufs;
	uint32			control;
#ifdef USE_SAC_CHECKSUM
	uint32			calc;
#endif

	if (!sbm_verifyIsCODInSac(dest_cod, 1)) {
		SBM_MGMT_DEBUG(("dest_cod, 0x%llx, not in SAC\n", dest_cod));
		return SBM_ERR_INVALID_COD;
	}
	if (!sbm_verifyIsCODInSac(src_cod, 1)) {
		SBM_MGMT_DEBUG(("src_cod, 0x%llx, not in SAC\n", src_cod));
		return SBM_ERR_INVALID_COD;
	}
	src_edesc = sbm_getDescFromCOD(src_cod);
	dest_edesc = sbm_getDescFromCOD(dest_cod);

#ifdef USE_SAC_CHECKSUM
	if (dest_edesc->hdr.attr != (calc = sbm_verifyCalcPktChecksum(dest_edesc))) {
		printf("dest_cod chksum failed, stored = 0x%x, calc = 0x%x\n",
													dest_edesc->hdr.attr, calc);
		sbm_stats.chksum_verify_error++;
		sbm_stats.chksum_verify_error_total++;
		return SBM_ERR_CHKSUM_FAILED;
	}
	if (src_edesc->hdr.attr != (calc = sbm_verifyCalcPktChecksum(src_edesc))) {
		printf("src_cod chksum failed, stored = 0x%x, calc = 0x%x\n",
													src_edesc->hdr.attr, calc);
		sbm_stats.chksum_verify_error++;
		sbm_stats.chksum_verify_error_total++;
		return SBM_ERR_CHKSUM_FAILED;
	}
#endif

	if (!sbm_verifyIsPktOk(src_edesc))
		panic("sbm_concatCOD, verifyPkt failed, src_edesc = %p\n", src_edesc);
	if (!sbm_verifyIsPktOk(dest_edesc))
		panic("sbm_concatCOD, verifyPkt failed, dest_edesc = %p\n", dest_edesc);

	last_edesc = dest_edesc;
	while (last_edesc->hdr.next)
		last_edesc = last_edesc->hdr.next;

	src_len = EEE_GET_PKT_LEN(src_edesc);
	dest_len = EEE_GET_PKT_LEN(dest_edesc);

	src_nbufs = EEE_GET_NUM_BUFS(src_edesc);
	dest_nbufs = EEE_GET_NUM_BUFS(dest_edesc);

	control = dest_edesc->hdr.control & ~EDESC_TOTAL_LEN;
	total_len = src_len + dest_len;
	dest_edesc->hdr.control = control | total_len;

	src_edesc->hdr.control &= ~EDESC_TOTAL_LEN;	/* 0 the len of chained desc */

	last_edesc->hdr.next = src_edesc;

#ifdef USE_SAC_CHECKSUM
	dest_edesc->hdr.attr = sbm_verifyCalcPktChecksum(dest_edesc);
#endif

	if (update_stats) {
		sbm_stats.cods_concat++;
		sbm_stats.cods_concat_total++;
		sbm_stats.cods_present--;
	}

	if (update_stats)
		sbm_collapseCOD(dest_cod);

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:	sbm_swapCOD()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/7/02
 *
 *-----------------------------------------------------------------
 */
void
sbm_swapCOD(
	nfx_cod_t	c1,
	nfx_cod_t	c2
)
{
	eee_descPtr_t	ed1, ed2;
	eee_desc_t	tmp_ed;

	ed1 = sbm_getDescFromCOD(c1);
	ed2 = sbm_getDescFromCOD(c2);

	memcpy(&tmp_ed, ed1, sizeof(eee_desc_t));
	memcpy(ed1,ed2,sizeof(eee_desc_t));
	memcpy(ed2,&tmp_ed,sizeof(eee_desc_t));
}

/*-----------------------------------------------------------------
 * Name:	sbm_mergeCOD()
 *
 * Description:	Merge a src COD into a dest COD.  Src COD will be freed at
 *				end of this function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/8/02
 *
 *-----------------------------------------------------------------
 */
int32
sbm_mergeCOD(
	nfx_cod_t	src_cod,
	uint32		src_len,
	uint32		src_off,
	nfx_cod_t	dest_cod,
	uint32		dest_off,
	uint32		update_stats
)
{
	eee_descPtr_t	src_edesc, dest_edesc;
	int32			rc;
	uint32			src_pkt_len, dest_pkt_len;
	nfx_cod_t		trash_cod, rest_cod;
	eee_descPtr_t	trash_edesc, rest_edesc;
#ifdef USE_SAC_CHECKSUM
	uint32			calc;
#endif

	if (!sbm_verifyIsCODInSac(src_cod, 1)) {
		panic("src_cod, 0x%llx, not in SAC\n", src_cod);
		return SBM_ERR_INVALID_COD;
	}
	if (!sbm_verifyIsCODInSac(dest_cod, 1)) {
		panic("dest_cod, 0x%llx, not in SAC\n", dest_cod);
		return SBM_ERR_INVALID_COD;
	}

	src_edesc = sbm_getDescFromCOD(src_cod);
	dest_edesc = sbm_getDescFromCOD(dest_cod);

#ifdef USE_SAC_CHECKSUM
	if (src_edesc->hdr.attr != (calc = sbm_verifyCalcPktChecksum(src_edesc))) {
		printf("src_cod chksum failed, stored = 0x%x, calc = 0x%x\n",
													src_edesc->hdr.attr, calc);
		sbm_stats.chksum_verify_error++;
		sbm_stats.chksum_verify_error_total++;
		return SBM_ERR_CHKSUM_FAILED;
	}
	if (dest_edesc->hdr.attr != (calc = sbm_verifyCalcPktChecksum(dest_edesc))) {
		printf("dest_cod chksum failed, stored = 0x%x, calc = 0x%x\n",
													dest_edesc->hdr.attr, calc);
		sbm_stats.chksum_verify_error++;
		sbm_stats.chksum_verify_error_total++;
		return SBM_ERR_CHKSUM_FAILED;
	}
#endif

	if (!sbm_verifyIsPktOk(src_edesc))
		panic("sbm_mergeCOD, verifyPkt failed, src_edesc = %p\n", src_edesc);
	if (!sbm_verifyIsPktOk(dest_edesc))
		panic("sbm_mergeCOD, verifyPkt failed, dest_edesc = %p\n", dest_edesc);

	src_pkt_len = EEE_GET_PKT_LEN(src_edesc);
	dest_pkt_len = EEE_GET_PKT_LEN(dest_edesc);

	if (src_len+src_off > src_pkt_len) {
		printf("src len err, len = 0x%x, off = 0x%x, pkt_len = 0x%x\n",
											src_len, src_off, src_pkt_len);
		return SBM_ERR_BAD_LEN;
	}
	if (dest_off > dest_pkt_len+1) {
		printf("dest off err, off = 0x%x, pkt_len = 0x%x\n",
														dest_off, dest_pkt_len);
		return SBM_ERR_BAD_LEN;
	}

	/*
	 * First, make src_cod a standalone cod, based on src_off and src_len.
	 * There are 3 cases for converting the src_cod to a standalone cod:
	 *	1. the src_off == 0
	 *	2. the src_off+src_len == src_pkt_len
	 *	3. the src_off+src_len < src_pkt_len
	 *
	 * Then, there are 4 cases for merging:
	 *	1. the dest_off is at the end of the dest_cod, use concat.
	 *
	 *	2. the dest_off == 0.  If the entire cod is to be replaced, use
	 *		swapCOD and freeCOD, else use splitCOD, concatCOD, swapCOD, freeCOD.
	 *
	 *	3. the src_len overlaps the end of the dest_cod.  In this case, use
	 *		splitCOD, concatCOD, freeCOD.
	 *
	 *	4. the entire src_len fits within the dest_cod.  In this case, use
	 *		splitCOD, splitCOD, concatCOD, concatCOD, freeCOD.
	 */

	/*
	 * Make src_cod a complete cod, based on src_off and src_len
	 */
	if (src_off == 0) {
		if (src_len < src_pkt_len) {
			trash_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
			trash_cod = sbm_getCODFromDesc(trash_edesc, nfxMySlotId, nfxMyCPUId);
			if ((rc = sbm_splitCOD(src_cod, src_len, trash_cod, 0)) != NFX_OK) {
				printf("sbm_mergeCOD error, src convert type 1/split\n");
				return rc;
			}
			sbm_freeCOD(trash_cod, 0);
		}
	}
	else if (src_off+src_len == src_pkt_len) {
		trash_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
		trash_cod = sbm_getCODFromDesc(trash_edesc, nfxMySlotId, nfxMyCPUId);
		if ((rc = sbm_splitCOD(src_cod, src_off, trash_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, src convert type 2/split\n");
			return rc;
		}
		sbm_freeCOD(src_cod, 0);
		src_cod = trash_cod;
	}
	else {
		trash_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
		trash_cod = sbm_getCODFromDesc(trash_edesc, nfxMySlotId, nfxMyCPUId);
		if ((rc = sbm_splitCOD(src_cod, src_off, trash_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, src convert type 3/split 1\n");
			return rc;
		}
		sbm_freeCOD(src_cod, 0);
		src_cod = trash_cod;
		src_edesc = sbm_getDescFromCOD(src_cod);
		src_pkt_len = EEE_GET_PKT_LEN(src_edesc);

		trash_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
		trash_cod = sbm_getCODFromDesc(trash_edesc, nfxMySlotId, nfxMyCPUId);
		if ((rc = sbm_splitCOD(src_cod, src_len, trash_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, src convert type 3/split 2\n");
			return rc;
		}
		sbm_freeCOD(trash_cod, 0);
	}

	/*
	 * Now check the 4 cases for merging
	 */
	if (dest_off == dest_pkt_len) {
//		printf("merge at end of dest cod, use concat\n");
		if ((rc = sbm_concatCOD(dest_cod, src_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 1/concat\n");
			return rc;
		}
#ifdef USE_SAC_CHECKSUM
		dest_edesc->hdr.attr = sbm_verifyCalcPktChecksum(dest_edesc);
#endif
		if (update_stats) {
			sbm_stats.cods_merged++;
			sbm_stats.cods_merged_total++;
			sbm_stats.cods_present--;
		}
		sbm_collapseCOD(dest_cod);
		return NFX_OK;
	}

	if (dest_off == 0) {
		if (src_len < dest_pkt_len) {
			rest_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
			rest_cod = sbm_getCODFromDesc(rest_edesc, nfxMySlotId, nfxMyCPUId);
			if ((rc = sbm_splitCOD(dest_cod, src_len, rest_cod, 0)) != NFX_OK) {
				printf("sbm_mergeCOD error, type 2/split\n");
				return rc;
			}
			if ((rc = sbm_concatCOD(src_cod, rest_cod, 0)) != NFX_OK) {
				printf("sbm_mergeCOD error, type 2/concat\n");
				return rc;
			}
			sbm_swapCOD(dest_cod, src_cod);
			sbm_freeCOD(src_cod, 0);

			if (update_stats) {
				sbm_stats.cods_merged++;
				sbm_stats.cods_merged_total++;
				sbm_stats.cods_present--;
			}
			sbm_collapseCOD(dest_cod);
			return NFX_OK;
		}
		else if (src_len >= dest_pkt_len) {
            sbm_swapCOD(dest_cod, src_cod);
			sbm_freeCOD(src_cod, 0);
			if (update_stats) {
				sbm_stats.cods_merged++;
				sbm_stats.cods_merged_total++;
				sbm_stats.cods_present--;
			}
			sbm_collapseCOD(dest_cod);
			return NFX_OK;
        }
	}
	else if (dest_off + src_len >= dest_pkt_len) {
		trash_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
		trash_cod = sbm_getCODFromDesc(trash_edesc, nfxMySlotId, nfxMyCPUId);
		if ((rc = sbm_splitCOD(dest_cod, dest_off, trash_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 2/split\n");
			return rc;
		}
		if ((rc = sbm_concatCOD(dest_cod, src_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 2/concat\n");
			return rc;
		}
		sbm_freeCOD(trash_cod, 0);

#ifdef USE_SAC_CHECKSUM
		dest_edesc->hdr.attr = sbm_verifyCalcPktChecksum(dest_edesc);
#endif
		if (update_stats) {
			sbm_stats.cods_merged++;
			sbm_stats.cods_merged_total++;
			sbm_stats.cods_present--;
		}
		sbm_collapseCOD(dest_cod);
		return NFX_OK;
	}

	else {
		trash_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
		trash_cod = sbm_getCODFromDesc(trash_edesc, nfxMySlotId, nfxMyCPUId);
		rest_edesc = eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
		rest_cod = sbm_getCODFromDesc(rest_edesc, nfxMySlotId, nfxMyCPUId);
		if ((rc = sbm_splitCOD(dest_cod, dest_off, trash_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 3/split 1\n");
			return rc;
		}
		
		if ((rc = sbm_splitCOD(trash_cod, src_len, rest_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 3/split 2\n");
			return rc;
		}

		if ((rc = sbm_concatCOD(dest_cod, src_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 3/concat 1\n");
			return rc;
		}

		if ((rc = sbm_concatCOD(dest_cod, rest_cod, 0)) != NFX_OK) {
			printf("sbm_mergeCOD error, type 3/concat 2\n");
			return rc;
		}
		sbm_freeCOD(trash_cod, 0);

#ifdef USE_SAC_CHECKSUM
		dest_edesc->hdr.attr = sbm_verifyCalcPktChecksum(dest_edesc);
#endif
		if (update_stats) {
			sbm_stats.cods_merged++;
			sbm_stats.cods_merged_total++;
			sbm_stats.cods_present--;
		}
	}

	sbm_collapseCOD(dest_cod);
	return NFX_OK;

}	/* sbm_mergeCOD() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	8/7/02
 *
 *-----------------------------------------------------------------
 */
void
sbm_processReceivePackets(
	eee_descPtr_t	edesc
)
{
	eee_descBufPtr_t	bd;
	uint32				cmd;
	int32				rc;

	SBM_MGMT_DEBUG(("sbm_processReceivePackets, %p\n", edesc));

	if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
		bd = &(edesc->bd[0]);
	else
		bd = &(edesc->bd[1]);

	cmd = *(uint32 *)(bd->buf);

	switch (cmd) {

		case CMD_TO_MCPU_SEND_CONTROL_AND_DATA_TO_SNIM:
			{
				nfx_cod_t			cod;
				uint32				len;
				uint32				offset;
				uint32				tot_len;
				sbm_sendControlAndSACData_t	* msg;
				uchar8				* buf, * newbuf;
#ifdef USE_SAC_CHECKSUM
				scsi_writeReqMsg_t	* wp;
#endif

				SBM_TX_DEBUG(("SEND_DATA_TO_SNIM\n"));

                msg = (sbm_sendControlAndSACData_t *)bd->buf;

				cod = msg->COD;
				len = msg->length;
				offset = msg->offset;

				tot_len = msg->msg_len + len;
#ifndef LINUX_TEST
				if (sb1250_getRev() > SB1250_PASS1) {
					sbm_txCodToSnimWithDataMover(cod,
												len,
												offset,
												(uchar8 *)&(msg->msg),
												msg->msg_len,
												msg->dest_port,
												msg->source_port);
					sbm_stats.cods_tx_snim++;
					sbm_stats.cods_tx_snim_total++;
				}
				else {
#endif
					buf = newbuf = (uchar8 *)eee_ramAlloc(tot_len);
					if (buf == NULL) {
						printf("could not allocate memory for DATA_TO_SNIM\n");
						return;
					}
					memcpy(buf, &(msg->msg), msg->msg_len);
					sbm_stats.bytes_copied_tx += msg->msg_len;
					sbm_stats.bytes_copied_tx_total += msg->msg_len;
					buf += msg->msg_len;

					if (sbm_sac_tx_snim(cod, len, offset, buf) == NFX_OK) {

#ifdef USE_SAC_CHECKSUM
						wp = (scsi_writeReqMsg_t *)newbuf;
						wp->checkSum = (uint64)
									sbm_verifyCalcFlatBufferChecksum(buf,len);
#endif
						rc = dp_proxy_tx_flat_buffer(newbuf,
													tot_len,
													msg->dest_port,
													msg->source_port);
						if (rc != NFX_OK) {
							printf("dp_proxy returned error, rc = %x\n", rc);
							eee_ramDealloc(newbuf);
						}
					}
					else
						eee_ramDealloc(newbuf);
#ifndef LINUX_TEST
				}
#endif
			}

			break;

		case CMD_TO_MCPU_SEND_CONTROL_AND_DATA_TO_SRC:
			{
				nfx_cod_t			cod;
				uint32				len;
				uint32				offset;
				uint32				tot_len;
				sbm_sendControlAndSACData_t	* msg;
				uchar8				* buf, * newbuf;
				uint32				dp;

				SBM_TX_DEBUG(("SEND_DATA_TO_SRC\n"));

				msg = (sbm_sendControlAndSACData_t *)bd->buf;

				cod = msg->COD;
				len = msg->length;
				offset = msg->offset;

				tot_len = msg->msg_len + len;
				buf = newbuf = (uchar8 *)eee_ramAlloc(tot_len);
				if (buf == NULL) {
					printf("could not allocate memory for DATA_TO_SNIM\n");
					return;
				}
				memcpy(buf, &(msg->msg), msg->msg_len);
				sbm_stats.bytes_copied_tx += msg->msg_len;
				sbm_stats.bytes_copied_tx_total += msg->msg_len;
				buf += msg->msg_len;

				if (sbm_sac_tx_src(cod, len, offset, buf) == NFX_OK) {

					dp = edesc->hdr.src_port;
					rc = eee_sendMessage(newbuf,
										tot_len,
										dp,
										msg->source_port);
					eee_ramDealloc(newbuf);
				}
				else
					eee_ramDealloc(newbuf);
			}

			break;


		case CMD_TO_MCPU_MANIPULATE_FREE:
			{
				sbm_freeCOD_t		* msg;
				nfx_cod_t			cod;

				SBM_MGMT_DEBUG(("MANIPULATE_FREE, "));

				msg = (sbm_freeCOD_t *)bd->buf;
				cod = msg->COD;

				SBM_MGMT_DEBUG(("cod = 0x%llx\n", cod));

				if (sbm_freeCOD(cod, 1) != NFX_OK) {
					printf("sbm_free: Invalid COD handle\n");
				}

				break;
			}

		case CMD_TO_MCPU_MANIPULATE_MERGE:
			{
				sbm_mergeCOD_t		* msg;
				nfx_cod_t			src_cod, dest_cod;
				uint32				src_len, src_off, dest_off;

				SBM_MGMT_DEBUG(("MANIPULATE_MERGE, "));

				msg = (sbm_mergeCOD_t *)bd->buf;
				src_cod = msg->src_COD;
				src_len = msg->src_len;
				src_off = msg->src_offset;
				dest_cod = msg->dest_COD;
				dest_off = msg->dest_offset;

				SBM_MGMT_DEBUG(("src_cod = 0x%llx, len = 0x%x, offset = 0x%x\n",
													src_cod, src_len, src_off));

				SBM_MGMT_DEBUG(("                  "));
				SBM_MGMT_DEBUG(("dest_cod = 0x%llx, offset = 0x%x\n",
															dest_cod, dest_off));
				if (sbm_mergeCOD(	src_cod,
									src_len,
									src_off,
									dest_cod,
									dest_off,
									1) != NFX_OK) {
					printf("sbm_merge: Invalid COD handle\n");
				}

				break;
			}

		case CMD_TO_MCPU_MANIPULATE_CREATE:
			{
				sbm_createCOD_t		* msg;
				nfx_cod_t			cod;
				uint32				len, fill;

				SBM_MGMT_DEBUG(("MANIPULATE_CREATE, "));

				msg = (sbm_createCOD_t *)bd->buf;
				cod = msg->COD;
				len = msg->len;
				fill = msg->fill;

				SBM_MGMT_DEBUG(("cod = 0x%llx, len = 0x%x, fill = 0x%x\n",
																cod, len, fill));

				if (sbm_createCOD(	cod,
									len,
									fill,
									1) != NFX_OK) {
					panic("Could not create COD 0x%llx\n", cod);
				}

				break;
			}

		case CMD_TO_MCPU_MANIPULATE_SPLIT:
			{
				sbm_splitCOD_t		* msg;
				nfx_cod_t			src_cod, dest_cod;
				uint32				src_off;

				SBM_MGMT_DEBUG(("MANIPULATE_SPLIT, "));

				msg = (sbm_splitCOD_t *)bd->buf;
				src_cod = msg->src_COD;
				src_off = msg->src_offset;
				dest_cod = msg->dest_COD;

				SBM_MGMT_DEBUG(("src_cod = 0x%llx, off = 0x%x, dest_cod = 0x%llx\n",
													src_cod, src_off, dest_cod));

				if (sbm_splitCOD(	src_cod,
									src_off,
									dest_cod,
									1) != NFX_OK) {
					printf("Invalid split COD command\n");
				}

				break;
			}

		case CMD_TO_MCPU_MANIPULATE_CONCAT:
			{
				sbm_concatCOD_t		* msg;
				nfx_cod_t			src_cod, dest_cod;

				SBM_MGMT_DEBUG(("MANIPULATE_CONCAT, "));

				msg = (sbm_concatCOD_t *)bd->buf;

				dest_cod = msg->dest_COD;
				src_cod = msg->src_COD;

				SBM_MGMT_DEBUG(("dest_cod = 0x%llx, src_cod = 0x%llx\n",
															dest_cod, src_cod));
				if (sbm_concatCOD(dest_cod, src_cod, 1) != NFX_OK) {
					printf("Invalid concatenate COD command\n");
				}

				break;
			}
	
		case CMD_TO_MCPU_ADDRESS_OF_SAC_INFO_STRUCT:
			{
				sbm_infoAddrMsg_t	* msg;

				msg = (sbm_infoAddrMsg_t *)(bd->buf);

				sbm_sac_info = msg->addr;
//				printf("sbm_sac_info structure at %p\n", sbm_sac_info);

				sbm_mgmt_update_sac_info_struct(NULL, NULL);

				break;
			}

		default:
			panic("Invalid Message to MCPU, edesc at 0x%x, cmd = 0x%x\n",
															edesc, cmd);
	}

	eee_freePacket(edesc);

}	/* sbm_processReceivePackets() */


/*-----------------------------------------------------------------
 * Name:	sbm_processMcastReceivePackets(edesc)
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/5/01
 *
 *-----------------------------------------------------------------
 */
void
sbm_processMcastReceivePackets(
	eee_descPtr_t	edesc
)
{
	SBM_MGMT_DEBUG(("sbm_processMcastReceivePackets, 0x%x\n", edesc));

}	/* sbm_processMcastReceivePackets() */

#endif	/* NFP_TXRX */

