/*
 *-----------------------------------------------------------------
 *
 * Name:	sbm-queue.c
 *
 * RCS:  $Id: sbm-queue.c,v 1.10 2002/08/20 20:23:22 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Queueing routines for SAC Buffer Management
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/1/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>

#ifdef SBM_REV2
#error Compiled with SBM_REV2
#endif

#ifdef NFP_TXRX

#include <exports/sbm-api.h>
#include "sbm.h"

/*-----------------------------------------------------------------
 * Name:	sbm_freeHMTIndex(index)
 *
 * Description:	Adds HMT index onto free queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/1/01
 *
 *-----------------------------------------------------------------
 */
void						/* no return, if error detected, panic */
sbm_freeHMTIndex(
	uint32	index			/* index to add to free queue */
)
{
	sbm_hmtEntry_t	* hmt;

	if (index >= sbm.num_sac_blocks)
		panic("Invalid HMT index, 0x%x\n", index);

	hmt = sbm_hmt + index;

	if (!(hmt->bdl_index & HMT_ENTRY_IN_USE))
		panic("sbm_freeIFQIndex, HMT index 0x%x not in use\n", index);

	hmt->bdl_index = HMT_ENTRY_FREE;

	if (sbm_ifq->head == sbm_ifq->start)
		panic("sbm_freeIFQIndex, no available entry\n");

	*--sbm_ifq->head = index;

	sbm_stats->num_free_hmt_indicies++;

}	/* sbm_freeHMTIndex() */

/*-----------------------------------------------------------------
 * Name:	sbm_freeBDLIndex(index)
 *
 * Description:	Adds BDL index onto free queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/1/01
 *
 *-----------------------------------------------------------------
 */
void						/* no return, if error detected, panic */
sbm_freeBDLIndex(
	uint32	index			/* index to add to free queue */
)
{
	sbm_fbq_t		* fbq;
	sbm_bdlEntry_t	* bdl;

	if (index >= sbm.num_sac_blocks)
		panic("Invalid BDL index, 0x%x\n", index);

	bdl = sbm.bdl + index;

	if (!(bdl->next_bdl_index & BDL_ENTRY_IN_USE))
		panic("sbm_freeBDLIndex, freeing entry not in use, 0x%x\n", index);

	bdl->next_bdl_index = BDL_ENTRY_FREE;

	if (sbm_fbq->head == sbm_fbq->start)
		panic("sbm_freeBDLIndex, no available entry\n");

	*--sbm_fbq->head = index;

	sbm_stats->num_sac_blocks_in_use--;
	sbm_stats->num_sac_blocks_avail++;

}	/* sbm_freeBDLIndex() */

/*-----------------------------------------------------------------
 * Name:	sbm_getHMTIndex()
 *
 * Description:	Returns a free HMT index from the MCPU queue.
 *				Returns 0xFFFFFFFF if there is none available.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/26/01
 *
 *-----------------------------------------------------------------
 */
uint32
sbm_getHMTIndex()
{
	uint32	index;

	if (sbm_ifq->head == sbm_ifq->end) {
		sbm_stats->no_free_ifq++;
		return 0xFFFFFFFF;
	}

	sbm_stats->num_free_hmt_indicies--;

	index = *sbm_ifq->head;
	*sbm_ifq->head++ = 0;

	return (index);

}	/* sbm_getHMTIndex() */

void
sbm_ungetHMTIndex(
	uint32	index
)
{
	sbm_stats->num_free_hmt_indicies++;

	*--sbm_ifq->head = index;
}

/*-----------------------------------------------------------------
 * Name:	sbm_getBDLIndex()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	7/26/01
 *
 *-----------------------------------------------------------------
 */
uint32
sbm_getBDLIndex()
{
	uint32	index;

	if (sbm_fbq->head == sbm_fbq->end) {
		sbm_stats->no_free_fbq++;
		return 0xFFFFFFFF;
	}

	sbm_stats->num_sac_blocks_in_use++;
	sbm_stats->num_sac_blocks_avail--;

	index = *sbm_fbq->head;
	*sbm_fbq->head++ = 0;

	return (index);

}	/* sbm_getBDLIndex() */

void
sbm_ungetBDLIndex(
	uint32	index
)
{
	sbm_stats->num_sac_blocks_in_use--;
	sbm_stats->num_sac_blocks_avail++;

	*--sbm_fbq->head = index;
}

#endif	/* NFP_TXRX */

