/*  FileName: scsi-cmn-cmds.c
 *
 *  Description: Contains routines for issuing commons commands to SCSI devices
 *				 and processing responses.
 *
 *	Created:	12 February	2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"
#include <exports/malloc-api.h>

int scsi_initParseInqString(char *, int, char *, int);
extern int scsi_dispatchCmd( scsi_desc_t *sd, scsi_dev_t *pDev );

/*
 *-------------------------------------------------------
 * Function Name:	scsi_sendInquiry
 *
 * Description: Will send SCSI INQUERY command.
 *
 * Returns: 
 *
 *-------------------------------------------------------
 */
int
scsi_cmnRelayCDB(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32 len;
	uint32 free_packet = 1;
	uint32 rc = NFX_OK;
	unsigned char opCode;


	// Setup the scsi descriptor with some common values. These may be
	// overridden later depending on the opCode.
	if (pdev->type == SCSI_DEVTYPE_DISK) {
		if( sd->retryCnt == 0 )
			sd->retryCnt = 3;

		if( sd->timeOutCnt == 0 )
			sd->timeOutCnt = 5;
	}
	else {
		if( sd->retryCnt == 0 )
			sd->retryCnt = 1;

		if( sd->timeOutCnt == 0 )
			sd->timeOutCnt = 60;
	}

	if( sd->maxTimeToLive == 0 )
		sd->maxTimeToLive = 120;

	//-- Setup command data parameters.
	
	opCode = sd->cdb[0];
	switch( opCode )  
	{
		case SCSI_CMD_TEST_UNIT_READY:
		case SCSI_CMD_PREVENT_REMOVE:
			sd->maxTimeToLive = 120;
			sd->timeOutCnt = 5;
			break;

		case SCSI_CMD_REWIND_UNIT:
			sd->maxTimeToLive = sd->timeOutCnt = 4 * 60;
			break;

		case SCSI_CMD_LOAD_UNLOAD:
			sd->maxTimeToLive = sd->timeOutCnt = 16 * 60;
			break;

		case SCSI_CMD_SPACE:
			sd->maxTimeToLive = sd->timeOutCnt = 6 * 60 * 60; 
			break;

		case SCSI_CMD_WRITE_FILEMARKS:
			sd->maxTimeToLive = sd->timeOutCnt = 1 * 60 * 60; 
			break;

		case SCSI_CMD_ERASE:
			sd->maxTimeToLive = sd->timeOutCnt = 6 * 60 * 60; 
			break;

		case SCSI_CMD_REQUEST_SENSE:
			sd->status |= SD_DEV_DATA_IN_EXPECT; 
			sd->expectDataLen = 255;
			break;

		case SCSI_CMD_FORMAT_UNIT:
			sd->timeOutCnt    = 0x3FFFFFFF;			// Give it as long as it needs.
			break;

		case SCSI_CMD_READ_BLOCK_LIMITS:
			break;

		case SCSI_CMD_INQUIRY:
			sd->status |= SD_DEV_DATA_IN_EXPECT; 
			sd->expectDataLen = sd->cdb[4];
			break;

		case SCSI_CMD_MODE_SELECT_6:
			sd->status |= SD_DEV_DATA_OUT_EXPECT; 
			free_packet = 0;
			break;

		case SCSI_CMD_WRITE_6:
			sd->status |= SD_DEV_DATA_OUT_EXPECT; 
			if (pdev->type == SCSI_DEVTYPE_TAPE)
				sd->maxTimeToLive = sd->timeOutCnt = 60 * 60;
			free_packet = 0;
			break;

		case SCSI_CMD_MODE_SENSE_6:
			sd->status |= SD_DEV_DATA_IN_EXPECT; 
			sd->expectDataLen = sd->cdb[4];
			break;

		case SCSI_CMD_SEND_DIAGNOTIC:
			break;

		case SCSI_CMD_READ_CAPACITY:
			sd->expectDataLen = 0xFF;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;

		case SCSI_CMD_READ_6:
			len = (sd->cdb[2] << 16) | 
			    (sd->cdb[3] << 8) | 
			    (sd->cdb[4]);
			if (pdev->type == SCSI_DEVTYPE_DISK ||
				(pdev->type == SCSI_DEVTYPE_TAPE && (sd->cdb[1] & 0x01)))
				len *= pdev->blockSize;
            sd->expectDataLen = len;
            sd->status |= SD_DEV_DATA_IN_EXPECT;
			if (pdev->type == SCSI_DEVTYPE_TAPE)
				sd->maxTimeToLive = sd->timeOutCnt = 60 * 60;
			break;

		case SCSI_CMD_READ_POSITION:
			sd->expectDataLen = 0xFF;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;

	    case SCSI_CMD_LOG_SELECT:
		    sd->status |= SD_DEV_DATA_OUT_EXPECT;
			free_packet = 0;
			break;

	    case SCSI_CMD_LOG_SENSE:
		    sd->status |= SD_DEV_DATA_IN_EXPECT;
			sd->expectDataLen = (sd->cdb[7] << 8) | sd->cdb[8];
			break;

	    case SCSI_CMD_REPORT_DENSITY:
		    sd->status |= SD_DEV_DATA_IN_EXPECT;
			sd->expectDataLen = (sd->cdb[7] << 8) | sd->cdb[8];
			break;

    	case SCSI_CMD_RESERVE_6:
			len = ( (sd->cdb[3] << 8) | 
					(sd->cdb[4]) );
			sd->expectDataLen = len;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;

    	case SCSI_CMD_RELEASE_6:
            break;

		case SCSI_CMD_RESERVE_10:
			len = ( (sd->cdb[7] << 8) | 
					(sd->cdb[8]) );
			sd->expectDataLen = len;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;

		case SCSI_CMD_RELEASE_10:
			break;

		case SCSI_CMD_REPORT_LUNS:
			len = ( (sd->cdb[6] << 24) |
					(sd->cdb[7] << 16) |
					(sd->cdb[8] << 8)  |
					(sd->cdb[9]) );
			sd->expectDataLen = len;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;

	    case SCSI_CMD_PERSISTENT_RESERVE_IN:
			len = ( (sd->cdb[7] << 8) |
					(sd->cdb[8]) );
			sd->expectDataLen = len;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;
			
	    case SCSI_CMD_PERSISTENT_RESERVE_OUT:
			sd->status |= SD_DEV_DATA_OUT_EXPECT; 
			free_packet = 0;
			break;

	    case SCSI_CMD_INIT_ELEM_STATUS:
	    case SCSI_CMD_INIT_ELEM_STATUS_WITH_RANGE:
	    case SCSI_CMD_MOVE_MEDIUM:
	    case SCSI_CMD_POSITION_TO_ELEM: 
			sd->maxTimeToLive = sd->timeOutCnt = 16 * 60;
			break;

	    case SCSI_CMD_READ_ELEM_STATUS:
	    case SCSI_CMD_REPORT_VOL_ELEM_ADDRESS:
			len = ( (sd->cdb[7] << 16) |
					(sd->cdb[8] << 8)  |
					(sd->cdb[9]) );
			sd->expectDataLen = len;	
			sd->status |= SD_DEV_DATA_IN_EXPECT;
			break;

	    case SCSI_CMD_SEND_VOL_TAG:	
			sd->status |= SD_DEV_DATA_OUT_EXPECT; 
			free_packet = 0;
			break;

		default:		//-- All other commands are handled here.
			if( sd->status & SD_DEV_DATA_OUT_EXPECT )
				free_packet = 0;
	}
	
	if( free_packet ) {
		eee_freePacket(sd->eDesc);
		sd->eDesc = 0;
	}

	return rc;
} /* scsi_cmnRelayCDB */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_sendInquiry
 *
 * Description: Will send SCSI INQUERY command.
 *
 * Returns: 
 *
 *-------------------------------------------------------
 */
int
scsi_sendInquiry(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32 lun;
	uint32 rc = NFX_OK;

	lun = SCSI_PATH_GET_LUN(sd->path);
	/*
	 * Setup the scsi descriptor
	 */
	sd->control &= ~SD_OPCODE_MASK;
	sd->control |= SD_OPCODE_INQUIRY;
	sd->cHandle = 0;
	sd->pdev = pdev;
	sd->path = pdev->path;
	
	/*
	 * Form the SCSI INQUERY command.
	 */
	sd->cdb[0] = SCSI_CMD_INQUIRY;
	sd->cdb[1] = (lun << 5) & 0xe0;
	sd->cdb[4] = 0xFF;					/* Send back 256 bytes of data */	
	sd->expectDataLen = 0xFF;
	sd->status |= SD_DEV_DATA_IN_EXPECT; 
	sd->timeOutCnt = 5;
	sd->maxTimeToLive = 120;
	sd->retryCnt = 3;

	/*
	 * Send it down.. 
	 */
	rc = scsi_dispatchCmd(sd, pdev);

	return rc;
} /* scsi_sendInquiry */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_parseInqRsp
 *
 * Description: Will parse the response to the
 *				SCSI INQUERY command returned by 
 *  			the SCSI target device.
 *
 * Returns: SCSI_GENERIC_ERROR | SCSI_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_parseInqRsp( scsi_dev_t *pdev, scsi_desc_t *sd )
{
	uchar8 *rspBuf;
	uchar8 *mfgStr, *productIdStr;
	
	/*
	 * If device returns an inquiry error then assume 
	 * no LUN is  present so abort the whole process 
	 * and remove the device.
	 */
	if ((sd->status & SD_DRV_STATUS_MASK) || 
		(sd->status & SD_DEV_STATUS_MASK) != SCSI_DEV_STATUS_GOOD)
		return SCSI_ERR_GENERIC;

	/*
	 * Get pointer to buffer containing the inquiry data.
	 */
	rspBuf = sd->eDesc->bd[1].buf;

#if 0
	if( (rspBuf[0] >> 5) == 3 )		/* No Dev supported on this LUN */
		return SCSI_ERR_GENERIC;
#endif

	mfgStr   = &rspBuf[8];
	productIdStr = &rspBuf[16];

	scsi_initParseInqString(mfgStr, 8, pdev->vendor, 9);
	scsi_initParseInqString(productIdStr, 16, pdev->productId, 17);

	pdev->type = rspBuf[0] & 0x1F;
	
	/*
	 * Device is ready to go so set flag.
	 */
	pdev->flags.fDeviceReady = 1;

	return SCSI_OK;
} /* scsi_parseInqRsp */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_sendReportLun
 *
 * Description: Will send the SCSI REPORT LUN command to 
 *  			the SCSI device.
 *
 * Returns: 
 *
 *-------------------------------------------------------
 */
int
scsi_sendReportLun(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32 rc;
	uint32 allocLen;

	/*
	 * Setup the scsi descriptor
	 */
	sd->control &= ~SD_OPCODE_MASK;
	sd->control |= SD_OPCODE_RLUN;
	sd->cHandle = 0;
	sd->pdev = pdev;
	sd->path = pdev->path;
	
	/*
	 * Form the SCSI REPORT LUN command.
	 */
	sd->cdb[0]	= SCSI_CMD_REPORT_LUNS;

	allocLen   = (8 * SCSI_MAX_LUN_PER_DEV);	
	sd->cdb[6] = (allocLen >> 24) & 0xFF;		/* MSB */
	sd->cdb[7] = (allocLen >> 16) & 0xFF;
	sd->cdb[8] = (allocLen >> 8) & 0xFF;		
	sd->cdb[9] = allocLen & 0xFF;				/* LSB */

	sd->expectDataLen = allocLen;	
	sd->status |= SD_DEV_DATA_IN_EXPECT;

    sd->timeOutCnt = 5;
    sd->maxTimeToLive = 120;
    sd->retryCnt = 3;

	/*
	 * Send command to transport layer driver
	 */
	rc = scsi_dispatchCmd(sd, pdev);

	return rc;
} /* scsi_sendReportLun */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_parseReportLunRsp
 *
 * Description: Will parse the reponse to the SCSI 
 *				REPORT LUN command returned by the 
 *				SCSI target.
 * Returns: 
 *
 *-------------------------------------------------------
 */
int 
scsi_parseReportLunRsp(scsi_dev_t *pdev, 
					  scsi_desc_t *sd, 
					  int32 *lunCnt, 
					  int32 **luns)
{
	int32   i,j,tmp;
	uchar8 *buf;
	
	scsi_debug("[scsi_parseReportLunRsp] Enter\n", SCSI_DBL_3);
	/*
	 * Get pointer to buffer containing the inquiry data.
	 */
	buf = sd->eDesc->bd[1].buf;

	tmp = ((buf[0] << 24) |
		   (buf[1] << 16) |
		   (buf[2] <<  8) |
		   (buf[3]));
	
	
	*luns = (int32 *)eee_ramAlloc(sizeof(uint32) * (tmp/8)); 
	if( !*luns ) {
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "[scsi_parseReportLunRsp]");
		return NFX_ERR;
	}

	if( tmp )
		*lunCnt = tmp / 8;

	for(i=j=0; i < tmp; i+=8, j++) {
		(*luns)[j] = ((buf[8+i] << 24) |
				   (buf[9+i] << 16) |
				   (buf[10+i] <<  8) |
				   (buf[11+i]));
		(*luns)[j] >>= 16;
	}
	scsi_debug("[scsi_parseReportLunRsp] Leave\n", SCSI_DBL_3);
	return SCSI_OK;
} /* scsi_parseReportLunRsp */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_sendReadCapacity
 *
 * Description: Will send the SCSI READ CAPACITY command to 
 *  			the SCSI target LUN.
 *
 * Returns: 
 *
 *-------------------------------------------------------
 */
int
scsi_sendReadCapacity(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	uint32 rc;

	/*
	 * Setup the scsi descriptor
	 */
	sd->control &= ~SD_OPCODE_MASK;
	sd->control |= SD_OPCODE_RCAP;
	sd->cHandle = 0;
	sd->pdev = pdev;
	sd->path = pdev->path;
	
	/*
	 * Form the SCSI READ CAPACITY command.
	 */
	sd->cdb[0]	= SCSI_CMD_READ_CAPACITY;
	sd->expectDataLen = 0xFF;
	sd->status |= SD_DEV_DATA_IN_EXPECT;

    sd->timeOutCnt = 5;
    sd->maxTimeToLive = 120;
    sd->retryCnt = 3;

	/*
	 * Send command to transport layer driver
	 */
	rc = scsi_dispatchCmd(sd, pdev);

	return rc;
} /* scsi_sendReadCapacity */

/*
 *-------------------------------------------------------
 * Function Name:	scsi_parseReadCapacityRsp
 *
 * Description: Will parse the reponse to the SCSI 
 *				REPORT CAPACITY command returned by the 
 *				SCSI target.
 * Returns: 
 *
 *-------------------------------------------------------
 */
int 
scsi_parseReadCapacityRsp( scsi_dev_t *pdev, scsi_desc_t *sd )
{
	uchar8 *buf;
	
	/*
	 * Get pointer to buffer containing the inquiry data.
	 */
	buf = sd->eDesc->bd[1].buf;

	pdev->blocks = 1 + ((buf[0] << 24) |
						(buf[1] << 16) |
						(buf[2] <<  8) |
						(buf[3]));

	pdev->blockSize = (buf[4] << 24) |
		              (buf[5] << 16) |
		              (buf[6] <<  8) |
		               buf[7];
	return SCSI_OK;
} /* scsi_parseReadCapacityRsp */

//------------------------------------------------------------------------------
// Function Name:   scsi_initParseInqString
// 
// Description:     Inquiry string elements (VENDOR strings, PRODUCT strings, 
//                  and SN strings) may have leading, trailing and embedded 
//                  spaces. We want to strip leading and trailing spaces, and
//                  compress and replace by single '-' embedded spaces.
//
// String parse state machine:
//   state == 0 --> no non-space character found; filtering leading spaces.
//   state == 1 --> non-space character found.
//   state == 2 --> space character found after non-space character.
//
// Data Created:    April 23, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int 
scsi_initParseInqString(char *in_str, int in_strlen, char *out_str, int out_strlen)
{
	int state = 0;
	int i;

	memset(out_str, 0, sizeof(out_strlen));

	for (i = 0; in_strlen > 0 && i < out_strlen; --in_strlen, ++in_str) {
		switch (state) {
		case 0:
			if (*in_str != ' ') {
				out_str[i++] = *in_str;
				state = 1;
			}
			break;
						
		case 1:
			if (*in_str != ' ') {
				out_str[i++] = *in_str;
				state = 1;
			}
			else
				state = 2;
			break;

		case 2:
			if (*in_str != ' ') {
				out_str[i++] = '-';
				out_str[i++] = *in_str;
				state = 1;
			}
			else
				state = 2;
			break;
		}
	} /* for */
	return(NFX_OK);
}

