/*  File Name  :	scsi-cmds.h
 *
 *  Description:	Contains defines, macros and structures for managing 
 *					SCSI commands. 
 *
 *  Created    :    19 February 2001
 *
 *  Author	   :    Frank Salzmann
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */


#ifndef SCSI_CMD_H_INCLUDED

#define SCSI_CMD_H_INCLUDED

/*
 *==============================================================================
 * SCSI CDB OPCODES
 *==============================================================================
 */
#define SCSI_CMD_TEST_UNIT_READY			 0x00	/* Mandatory: All			*/
#define SCSI_CMD_REWIND_UNIT				 0x01	/* Mandatory: Tape          */
#define SCSI_CMD_REQUEST_SENSE				 0x03	/* Mandatory: All			*/
#define SCSI_CMD_FORMAT_UNIT				 0x04	/* Mandatory: Disk, Tape	*/
#define SCSI_CMD_READ_BLOCK_LIMITS			 0x05	/* Mandatory: Tape			*/
#define SCSI_CMD_INIT_ELEM_STATUS			 0x07	/* Mandatory: MC			*/
#define SCSI_CMD_READ_6						 0x08	/* Mandatory: Disk, Tape	*/
#define SCSI_CMD_WRITE_6					 0x0A	/* Mandatory: Disk, Tape	*/
#define SCSI_CMD_WRITE_FILEMARKS			 0x10	/* Mandatory: Tape          */
#define SCSI_CMD_SPACE						 0x11	/* Mandatory: Tape			*/
#define SCSI_CMD_INQUIRY					 0x12	/* Mandatory: All			*/
#define SCSI_CMD_MODE_SELECT_6				 0x15	/* Mandatory: Tape			*/
#define SCSI_CMD_RESERVE_6					 0x16	/* Mandatory: All           */
#define SCSI_CMD_RELEASE_6					 0x17	/* Mandatory: All           */
#define SCSI_CMD_ERASE						 0x19	/* Mandatory: Tape			*/
#define SCSI_CMD_MODE_SENSE_6				 0x1A	/* Mandatory: Tape			*/
#define SCSI_CMD_LOAD_UNLOAD				 0x1B	/* Mandatory: Tape			*/
#define SCSI_CMD_SEND_DIAGNOTIC				 0x1D	/* Mandatory: All			*/
#define SCSI_CMD_PREVENT_REMOVE				 0x1E	/* Mandatory: Tape			*/
#define SCSI_CMD_READ_CAPACITY				 0x25	/* Mandatory: Disk			*/
#define SCSI_CMD_READ_10					 0x28	/* Mandatory: Disk			*/
#define SCSI_CMD_WRITE_10					 0x2A	/* Mandatory: Disk			*/
#define SCSI_CMD_POSITION_TO_ELEM			 0x2B	/* Mandatory: MC			*/
#define SCSI_CMD_READ_POSITION				 0x34	/* Mandatory: Tape			*/
#define SCSI_CMD_READ_LONG					 0x3E	/* Optional : Disk			*/
#define SCSI_CMD_WRITE_LONG					 0x3F	/* Optional : Disk			*/
#define SCSI_CMD_REPORT_DENSITY				 0x44	/* Mandatory: Tape			*/
#define SCSI_CMD_LOG_SELECT					 0x4C	/* Mandatory: Tape			*/
#define SCSI_CMD_LOG_SENSE					 0x4D	/* Mandatory: Tape			*/
#define SCSI_CMD_RESERVE_10					 0x56	/* Mandatory: Disk, Tape 	*/
#define SCSI_CMD_RELEASE_10					 0x57	/* Mandatory: Disk, Tape	*/
#define SCSI_CMD_PERSISTENT_RESERVE_IN   	 0x5E	/* Mandatory: All           */
#define SCSI_CMD_PERSISTENT_RESERVE_OUT   	 0x5F	/* Mandatory: All           */
#define SCSI_CMD_REPORT_LUNS				 0xA0	/* Mandatory: Tape			*/
#define SCSI_CMD_MOVE_MEDIUM				 0xA5	/* Mandatory: MC			*/
#define SCSI_CMD_REPORT_VOL_ELEM_ADDRESS	 0xB5	/* Mandatory: MC			*/
#define SCSI_CMD_SEND_VOL_TAG				 0xB6	/* Mandatory: MC			*/
#define SCSI_CMD_READ_ELEM_STATUS			 0xB8	/* Mandatory: MC			*/
#define SCSI_CMD_INIT_ELEM_STATUS_WITH_RANGE 0xE7	/* Mandatory: MC			*/

/*
 *===================================================================
 * SCSI Device Status Response Codes
 *===================================================================
 */
#define SCSI_DEV_STATUS_GOOD				0x00
#define SCSI_DEV_STATUS_CHECK				0x02
#define	SCSI_DEV_STATUS_CONDITION_MET		0x04
#define SCSI_DEV_STATUS_BUSY				0x08
#define SCSI_DEV_STATUS_INTERMEDIATE		0x10
#define SCSI_DEV_STATUS_INTERMEDIATE_MET	0x14
#define SCSI_DEV_STATUS_RESERVE_CONFLICT	0x18
#define SCSI_DEV_STATUS_CMD_TERMINATED		0x22
#define SCSI_DEV_STATUS_TASK_SET_FULL		0x28
#define SCSI_DEV_STATUS_ACA_ACTIVE			0x30

/*
 *===================================================================
 * SCSI Cmd Table Template
 *===================================================================
 */
typedef struct _scsi_cmd_entry {
	int (*req_func)(struct _scsi_dev_entry *, scsi_desc_t*);
	int (*err_func)(struct _scsi_dev_entry *, scsi_desc_t*);
	char   *name;
} scsi_cmd_t;

/*
 *===================================================================
 * SCSI INQUIRY Cmd Response
 *===================================================================
 */
typedef struct _scsi_inquiry_response
{
	uchar8 devType;						/* Device type + qualifier		*/
	uchar8 rmb;							/* Removable Media Bit			*/
	uchar8 version;						/* ansi version, scsi-1,2,3		*/
	uchar8 ctrl0;
	uchar8 len;
	uchar8 resv1;
	uchar8 ctrl1;
	uchar8 ctrl2;
	uchar8 manufacturer[8];				/* ASCII manufacturer name 		*/
	uchar8 product[16];					/* ASCII product name			*/
	uchar8 revision[4];					/* ASCII product version string	*/
	uchar8 vendor1[20];
	uchar8 resv2[40];
} scsi_inquiry_response_t;

/*
 *===================================================================
 * SCSI SENSE_REQUEST response structure and defines
 *===================================================================
 */

/* ---------- Possible Response Keys ---------- */
#define SCSI_SENSE_KEY_NO_SENSE				0x0
#define SCSI_SENSE_KEY_RECOVERED_ERROR		0x1
#define SCSI_SENSE_KEY_NOT_READY			0x2
#define SCSI_SENSE_KEY_MEDIUM_ERROR			0x3	
#define SCSI_SENSE_KEY_HARDWARE_ERROR		0x4	
#define SCSI_SENSE_KEY_ILLEGAL_REQUEST		0x5	
#define SCSI_SENSE_KEY_UNIT_ATTENTION		0x6	
#define SCSI_SENSE_KEY_DATA_PROTECT			0x7	
#define SCSI_SENSE_KEY_BLANK_CHECK			0x8	
#define SCSI_SENSE_KEY_VENDOR_SPECIFIC		0x9	
#define SCSI_SENSE_KEY_COPY_ABORTED			0xA	
#define SCSI_SENSE_KEY_ABORTED_COMMAND		0xB	
#define SCSI_SENSE_KEY_OBSOLETE				0xC	
#define SCSI_SENSE_KEY_VOLUME_OVERFLOW		0xD	
#define SCSI_SENSE_KEY_MISCOMPARE			0xE	
#define SCSI_SENSE_KEY_RESERVED				0xF	

typedef struct _scsi_sense_data
{
	uchar8	respCode;				
#define SCSI_SENSE_VALID_BIT				0x80
#define SCSI_SENSE_RESPCODE_CURRENT_ERR		0x70
#define SCSI_SENSE_RESPCODE_DEFERRED_ERR	0x71
#define SCSI_SENSE_RESPCODE_VENDOR_SPECIFIC	0x7F
									   
	uchar8	segNumber;				/* Cur Seg Desc, COPY,COMPARE, CPY&VER 	*/
	uchar8	senseKey;				/* Sense Key + control bits				*/
#define SCSI_SENSE_KEY_MASK					0x0F
#define SCSI_SENSE_ILI_BIT					0x20
#define SCSI_SENSE_EOM_BIT					0x40
#define SCSI_SENSE_FILEMARK_BIT				0x80
    uchar8	 info[4];				/* information field				*/
	uchar8	 addSenseLen;			/* Additional sense len				*/
	uchar8	 csi[4];				/* cmd specific information			*/
	uchar8	 asc;					/* additional sense code			*/
	uchar8	 ascQ;					/* additional sense code qualifier	*/
	uchar8	 fieldUC;				/* field replaceable unit code		*/
	uchar8	 sksCtrl;				/* Sense Key specific control byte	*/
	uchar8	 sksMsb;				/* Sense Key specific field ptr MSB	*/
	uchar8	 sksLsb;				/* Sense Key specific filed ptr LSB	*/
	uchar8	 addSense;				/* Additional sense bytes			*/
} scsi_sense_t;

/*
 *===================================================================
 * SCSI MODE SENSE/SELECT structures and defines
 *===================================================================
 */

/* Page control field definitions */
#define SCSI_MPAGE_CONTROL_CURRENT		0x00
#define SCSI_MPAGE_CONTROL_CHANGEABLE	0x40
#define SCSI_MPAGE_CONTROL_DEFAULT		0x80
#define SCSI_MPAGE_CONTROL_SAVED		0xC0

/* Mode sense/select page codes */
#define	SCSI_MPAGE_ERROR_RECOVERY		0x01
#define	SCSI_MPAGE_CONN_CONTROL			0x02
#define	SCSI_MPAGE_DEVICE_FORMAT		0x03
#define	SCSI_MPAGE_DRIVE_GEOMETRY		0x04
#define	SCSI_MPAGE_CACHE_CONTROL		0x08
#define	SCSI_MPAGE_CONTROL_MODE			0x0a
#define	SCSI_MPAGE_NOTCH_PARTITION		0x0c
#define SCSI_MPAGE_COMPRESSION			0x0f
#define	SCSI_MPAGE_DEVICE_CONFIGURATION	0x10
#define SCSI_MPAGE_FC_CONTROL			0x19
#define	SCSI_MPAGE_RETURN_ALLPAGES		0x3f	/* sense only */

typedef struct {
    uchar8	   length;
    uchar8	   medium_type;
    uchar8	   flag;
    uchar8	   block_desc_len;
} scsi_mode_header_t;

typedef struct {
    uchar8	   density_code;
    uchar8	   blocks[3];
    uchar8	   reserved;
    uchar8	   block_size[3];
} scsi_mode_block_desc_t;

/*
 * page 0x0F: Tape data compression characteristics page
 */
#define SCSI_CM_ENABLE_COMPRESSION		0x80
#define SCSI_CM_COMPRESSION_CAPABLE		0x40

typedef struct {
    uchar8	compression_enab;
    uchar8	decompression_enab;
    uchar8	compression_algorithm[4];
    uchar8	decompression_algorithm[4];
    uchar8	reserved[4];
} scsi_mode_compress_mode_t;

#define	MODE_PAGE_CODE_SAVEABLE			0x80
typedef struct {
    uchar8	   code;
	uchar8	   len;
    union {
		scsi_mode_compress_mode_t	compress_mode;
		uchar8						data[1];
    } u;
} scsi_mode_page_desc_t;

/*
 *===================================================================
 * SCSI READ POSITION structures and defines
 *===================================================================
 */
#define SCSI_RPOS_CONTROL_LONG	0x02
#define SCSI_RPOS_CONTROL_TCLP	0x04

#define SCSI_RPOS_FLAGS_BOP		0x80
#define SCSI_RPOS_FLAGS_EOP		0x40
#define SCSI_RPOS_FLAGS_MPU		0x08
#define SCSI_RPOS_FLAGS_BPU		0x04

typedef struct {
	byte   flags;
	byte   resvd_1[3];
	uint32 partition_number;
	uint32 block_number[2];
	uint32 file_number[2];
	uint32 set_number[2];
} scsi_rpos_long_desc_t;

/*
 *===================================================================
 * SCSI SPACE structures and defines
 *===================================================================
 */
#define SCSI_SPACE_CODE_BLOCKS	0x00
#define SCSI_SPACE_CODE_FMS		0x01
#define SCSI_SPACE_CODE_EOD		0x03

/*
 *===================================================================
 * Prototypes for SCSI command related functions.
 *===================================================================
 */

#endif  /* SCSI_CMD_H_INCLUDED */
