/*  FileName: scsi.c
 *
 *  Description: Contains routines that provide the SCSI API to upper layer 
 *				 modules. As well as the SCSI statemachine code.
 *
 *	Created:	20 February	2001 *
 *  Author:		Frank Salzmann
 *
 *	$Id: scsi-msg.c,v 1.138 2002/11/23 00:30:31 nitinb Exp $ 
 *	
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"
#include "../sm-sdm/sdm-api.h"
#include <exports/hal.h>
#include <exports/eee-desc.h>

typedef struct {
	eee_desc_t *			i_edesc;
	scsi_ioctlRspMsg_t *	o_msg;
} scsi_ioctl_callback_t;
int scsi_procIoctlCallback(scsi_ioctl_callback_t *, uint32);

static int scsi_queue_tape_op(scsi_dev_t *, eee_desc_t *, uint64, uint32, uint32);
static void scsi_proc_tape_queue_ops(scsi_dev_t *);

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendMessage
 *
 * Description:		Queues up messages for delivery to SSC.	
 *
 * Created by:		Nitin Bahadur
 *
 * Data Created:	Nov 11, 2002
 *
 * Returns		:	None
 *------------------------------------------------------------------------------
 */
int32
scsi_sendMessage(uchar8 *msg, uint32 num_bytes, uint32 dest_port, uint32 source_port)
{
	if(NFX_SLOTTYPE_SSC != EEE_GET_SLOT_NUM(dest_port) ||
		NFX_SSC_UNIX != EEE_GET_DEST_CPU(dest_port)) {
		eee_sendMessage((void *)msg, num_bytes, dest_port, source_port);
		eee_ramDealloc(msg);
	}
	else {
		scsi_msg_queue_t *scsi_msg;

		VERIFY(msg);

		// don't bother queuing messages if SDM is not up
		if(!scsiCtrl->sscMsgs.sdm_id) {
			eee_ramDealloc(msg);
			return NFX_OK;
		}

		scsi_msg = (scsi_msg_queue_t *)eee_ramAlloc(sizeof(scsi_msg_queue_t));
		if( !scsi_msg) {
			scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
					  "MSG QUEUE ENTRY");
			return NFX_ERR;
		}
		scsi_msg->msg = msg;
		scsi_msg->num_bytes = num_bytes;
		scsi_msg->dest_port = dest_port;
		scsi_msg->source_port = source_port;
		scsi_msg->next = NULL;

		scsi_msg_queue_enqueue(scsi_msg);
	}

	return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_receiveMsg
 *
 * Description: 	This function is registered with the EEE layer to receive
 *					all messages sent to the scsi driver.	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 7, 2001
 *
 * Returns:         None
 *
 *------------------------------------------------------------------------------
 */
void 
scsi_receiveMsg( eee_desc_t *eDesc )
{
	uint32		  mid;
	scsi_msgHdr_t *msg; 
	eee_descBuf_t *bd;

	scsi_debug("[scsi_receiveMsg] Enter", SCSI_DBL_3);

	/*
	 * Point to request message portion of data buffer.
	 */
	if( !eDesc ) {
		scsi_debug("Null eDesc", SCSI_DBL_3);
		return;
	}
		
	bd = &eDesc->bd[1];
	msg = (scsi_msgHdr_t *)(bd->buf);

	mid = msg->mid;
	
	// if this msg. is from SDM and our sdm id is 0, then we assume that sdm has come up
	if(EEE_GET_SLOT_NUM(eDesc->hdr.src_port) == NFX_SLOTTYPE_SSC && 
		EEE_GET_DEST_CPU(eDesc->hdr.src_port) == NFX_SSC_UNIX &&
		scsiCtrl->sscMsgs.sdm_id == 0 && mid != EEE_MSG_WINDOW_UPDATE) {
		scsiCtrl->sscMsgs.sdm_id = 1;	// temporary value
		scsi_procNullMsgRsp(NULL, NULL); 
	}

	switch( mid )
	{
		case SCSI_MID_RELAY_REQ:
			scsi_procRelayReq(eDesc, msg);
			break;

		case SCSI_MID_READ_REQ:
			scsi_procReadReq(eDesc, msg);
			break;

		case SCSI_MID_WRITE_REQ:
			scsi_procWriteReq(eDesc, msg);
			break;

	    case SCSI_MID_ACK:
			scsi_procAck(eDesc, msg);
			break;

		case SCSI_MID_OPEN_REQ:
			scsiCtrl->stats.openReqCnt++;
			scsi_procOpenMsg(eDesc, (scsi_openReqMsg_t *)msg);
			break;

		case SCSI_MID_CLOSE_REQ:
			scsiCtrl->stats.closeReqCnt++;
			scsi_procCloseMsg(eDesc, msg);
			break;

		case SCSI_MID_SDM_ACK:
			scsi_procSDMAck(eDesc, msg);
			break;

		case SCSI_MID_SYNC_REQ:
			scsiCtrl->stats.syncReqCnt++;
			scsi_procSyncMsg(eDesc, msg);
			break;

		case SCSI_MID_DEV_RESET_REQ:
			scsi_procDevResetMsg(eDesc, msg);
			break;

	    case SCSI_MID_DEV_STATUS_REQ:
			scsiCtrl->stats.statusReqCnt++;
			scsi_procDevStatusMsg(eDesc, msg);			
			break;

		case SCSI_MID_IOCTL_REQ:
			scsiCtrl->stats.ioctlReqCnt++;
			scsi_procIoctlMsg(eDesc, msg);
			break;

		case SCSI_MID_GET_STATS_REQ:
			scsiCtrl->stats.statsReqCnt++;
			scsi_procGetStatsMsg(eDesc, msg);
			break;

		case SCSI_MID_DELETE_REQ:
			scsiCtrl->stats.deleteReqCnt++;
			scsi_procDelMsg(eDesc, msg);
			break;

		case SCSI_MID_ADD_REQ:
			scsiCtrl->stats.addReqCnt++;
			scsi_procAddMsg(eDesc, (scsi_addReqMsg_t *)msg);
			break;

		case SCSI_MID_DEV_CONTROL_REQ:
			scsi_procDevControlMsg(eDesc, msg);
			break;

		case SCSI_MID_DEL_DEV_REQ:
			scsi_procDelDevMsg(eDesc, msg);
			break;

		case SCSI_MID_CLR_STATS_REQ:
			scsi_procClearStatsMsg(eDesc, msg);
			break;

		case SCSI_MID_NULL_RSP:
			scsi_procNullMsgRsp(eDesc, msg);
			break;

		case EEE_MSG_WINDOW_UPDATE:
			scsi_procWindowUpdateMsg(eDesc);	
			break;

		case SCSI_MID_HUGEIO_REQ:
			scsi_procHugeioReq(eDesc, msg);
			break;

		default: /* Invalid message */
			scsiCtrl->stats.invalidMsgCnt++;
			SCSI_DBG(0, ("scsi_receiveMsg: Invalid Message: %x\n", mid));
			scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
							   msg->cHandle, SCSI_MSG_ECODE_INVALID_MSG);
			eee_freePacket(eDesc);
	}
	scsi_debug("[scsi_receiveMsg] Leave", SCSI_DBL_3);
	return;
} /* scsi_receiveMsg */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procOpenMsg
 *
 * Description: 	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 22, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_procOpenMsg(eee_desc_t *edesc, scsi_openReqMsg_t *msg) 
{
	uint32		rc;

	SCSI_DBG(1, ("procOpenMsg: path: %x\n", msg->hdr.path));

	rc = scsi_openDevice( msg->hdr.path );

	scsi_sendOpenRsp(msg->hdr.cHandle, rc, edesc->hdr.src_port); 
	eee_freePacket(edesc);
	return(NFX_OK);
} /* scsi_procOpenMsg */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procCloseMsg
 *
 * Description: 	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 11, 2001
 *
 * Returns:         NFX_OK 
 *
 *------------------------------------------------------------------------------
 */
#if SDM_DEV_INIT

int
scsi_procCloseMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg) 
{
	scsi_dev_t *pdev;
	
	scsi_debug("scsi_procCloseMsg", SCSI_DBL_3);

	pdev  = scsi_getDevice(msg->path);
	if( pdev ) {
		scsi_sendCloseRsp(msg->cHandle, 
						  SCSI_MSG_ECODE_OK, 
						  edesc->hdr.src_port); 
	}
	else {
		scsi_sendCmdErrRsp(edesc->hdr.src_port, 
						   msg->cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
	}
	eee_freePacket(edesc);
	return(NFX_OK);
} // scsi_procCloseMsg 

#else
int
scsi_procCloseMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg) 
{
	scsi_dev_t *pdev;
    esm_event_t esmEvent;
	
	scsi_debug("scsi_procCloseMsg", SCSI_DBL_3);

	pdev  = scsi_getDevice(msg->path);
	if( pdev ) {
		if( pdev->flags.fOpenMode ) {
            esmEvent.event = SCSI_DEVO_EVENT_CLOSE;
            esmEvent.p1.u.w = msg->cHandle;
            esmEvent.p2.u.w = edesc->hdr.src_port;
            esm_deliverEvent(pdev->smHandle, &esmEvent);
		}
		else {
			scsi_sendCloseRsp(msg->cHandle, 
							  SCSI_MSG_ECODE_OK, 
							  edesc->hdr.src_port); 
		}
		eee_freePacket(edesc);
	}
	else {
		scsi_sendCmdErrRsp(edesc->hdr.src_port, 
						   msg->cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
		eee_freePacket(edesc);
	}
	return(NFX_OK);
} /* scsi_procCloseMsg */
#endif // SDM_DEV_INIT

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procSDMAck
 *
 * Description: 	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 04, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_procSDMAck(eee_desc_t *eDesc, scsi_msgHdr_t *msg) 
{
#if 0
	int32      event;
	scsi_dev_t *pdev;
    esm_event_t esmEvent;

	pdev  = (scsi_dev_t *)scsi_getDevice(msg->path);
	if( pdev ) {
		
		if(pdev->flags.fInitMode)
			event = SCSI_DEVI_EVENT_DCS_ACK;
		else
			event = SCSI_DEVO_EVENT_ACK;

        esmEvent.event = event;
        esmEvent.p1.u.p = NULL;
        esmEvent.p2.u.p = NULL;
        esm_deliverEvent(pdev->smHandle, &esmEvent);
	}
	else {
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   msg->cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
	}
#endif
	eee_freePacket(eDesc);
	return(NFX_OK);
} /* scsi_procSDMAck */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procSyncMsg
 *
 * Description: 	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	January 18, 2002
 *
 * Returns:         NFX_OK 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_procSyncMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg) 
{
	uint32 i;
	scsi_dev_t *pdev;
	
	scsi_debug("scsi_procSyncMsg", SCSI_DBL_3);

	//-- Release the edesc containing the request message.
	eee_freePacket(edesc);

	for( i=0; i < SCSI_MAX_DEV_PER_PORT; i++ ) {
		if( (pdev = scsiDevList[i]) ) {
			if( pdev->flags.fInitMode )		// First see if init mode is set.
				scsi_sendNewDevMsg( pdev );
			else
			if( pdev->flags.fOpenMode )
				scsi_sendAddDevMsg( pdev );
		}
	}
	

	return(NFX_OK);
} // scsi_procSyncMsg 

int
scsi_procDevResetMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg)
{
	scsi_path_t					path;
	scsi_dev_t *				pDev;
	hal_dispatch_entry_t *		dte = hal_dt_entry_get();
	int							rc;
	scsi_devResetRspMsg_t *		o_msg;
	uint32						source_port = scsiCtrl->scsiDataAddr;

	path = scsi_swap32(msg->path);	
	pDev = (scsi_dev_t *)scsi_getDevice(path);
	if (pDev == (scsi_dev_t *)NULL) {
		/* Invalid device id, reject request */
		scsi_sendCmdErrRsp(edesc->hdr.src_port, 
						   msg->cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
		eee_freePacket(edesc);
		return(NFX_OK);
	}

	rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_RESET_DEVICE, pDev->llHandle);

	o_msg = (scsi_devResetRspMsg_t *)eee_ramAlloc(sizeof(scsi_devResetRspMsg_t));
	bzero(o_msg, sizeof(scsi_devResetRspMsg_t));
    o_msg->hdr.mid = scsi_swap32(SCSI_MID_DEV_RESET_RSP);
    o_msg->hdr.cHandle = msg->cHandle;
    o_msg->eCode = scsi_swap32(rc == NFX_OK ? 0 : SCSI_MSG_ECODE_IOCTL_FAILED);
	if (EEE_GET_DTYPE(edesc->hdr.src_port) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(edesc->hdr.src_port) << EDESC2_DATA_TYPE_SHIFT;
	}
    scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_devResetRspMsg_t), 
					edesc->hdr.src_port, source_port);

	eee_freePacket(edesc);

    return(NFX_OK);
}   /* scsi_procDevResetMsg */

int
scsi_procDevStatusMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg)
{
	scsi_dev_t *				pdev;
	scsi_devStatusRspMsg_t *	scsi_o_msg;
	uint32						source_port = scsiCtrl->scsiDataAddr;
	
	scsi_debug("scsi_procDevStatusMsg", SCSI_DBL_3);

	pdev = scsi_getDevice(msg->path);
	if (pdev) {
		scsi_o_msg = (scsi_devStatusRspMsg_t *)eee_ramAlloc(sizeof(scsi_devStatusRspMsg_t));
		memset(scsi_o_msg, 0, sizeof(scsi_devStatusRspMsg_t));

		scsi_o_msg->hdr.cHandle = msg->cHandle;
		scsi_o_msg->hdr.mid = SCSI_MID_DEV_STATUS_RSP;
		scsi_o_msg->scsi_id = pdev->id;
		if (pdev->type == SCSI_DEVTYPE_TAPE) {
			scsi_o_msg->eCode = SCSI_MSG_ECODE_OK;
			scsi_o_msg->tape_position_eom = pdev->eom;
			scsi_o_msg->tape_position_bom = pdev->bom;
			scsi_o_msg->tape_position_blockno = pdev->blockno;
			scsi_o_msg->tape_position_fileno = pdev->fileno;
		}
		else {
			scsi_o_msg->eCode = SCSI_MSG_ECODE_ILLEGAL_REQUEST;
		}

		if (EEE_GET_DTYPE(edesc->hdr.src_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(edesc->hdr.src_port) << EDESC2_DATA_TYPE_SHIFT;
		}
		scsi_sendMessage((uchar8 *)scsi_o_msg, 
						sizeof(scsi_devStatsRspMsg_t), 
						edesc->hdr.src_port, 
						source_port);
	}
	else {
		scsi_sendCmdErrRsp(edesc->hdr.src_port, 
						   msg->cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
	}
	eee_freePacket(edesc);
	return(NFX_OK);
}

int
scsi_procIoctlMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg)
{
	scsi_ioctlReqMsg_t *	i_msg = (scsi_ioctlReqMsg_t *)msg;
	hal_dispatch_entry_t *	dte = hal_dt_entry_get();
	int						rc;
	scsi_ioctlRspMsg_t *	o_msg = NULL;
	uint32					mlen = 0;
	uint32					ecode = SCSI_MSG_ECODE_OK;
	scsi_ioctl_callback_t * async_context = NULL;
	uint32					source_port = scsiCtrl->scsiDataAddr;

	switch (i_msg->code) {
	case HAL_IOCTL_COMMON_DRIVER_VERBOSE_LEVEL:
		/* hal_ioctl(HAL_IOCTL_COMMON_DRIVER_VERBOSE_LEVEL, uint32 verbose_level) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_DRIVER_VERBOSE_LEVEL, 
							   (uint32)i_msg->args[0]);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_SET_DRIVER_OPTIONS:
		/* hal_ioctl(HAL_IOCTL_COMMON_SET_DRIVER_OPTIONS, uint32 driver_options) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_SET_DRIVER_OPTIONS, 
							   (uint32)i_msg->args[0]);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_GET_INTERFACE_TYPE:
		/* hal_ioctl(HAL_IOCTL_COMMON_GET_INTERFACE_TYPE, uint32 *interface_type); */
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(uint32);
		o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_GET_INTERFACE_TYPE, 
							   (uint32 *)o_msg->data);
		if (rc == NFX_OK)
			o_msg->dataLen = sizeof(uint32);
		else
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_GET_ADAPTER_INFO:
		/* hal_ioctl(HAL_IOCTL_COMMON_GET_ADAPTER_INFO, hal_adapter_info_t *adapter_info) */
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(hal_adapter_info_t);
		o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_GET_ADAPTER_INFO, 
							   (hal_adapter_info_t *)o_msg->data);
		if (rc == NFX_OK)
			o_msg->dataLen = sizeof(hal_adapter_info_t);
		else
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_RESET_ADAPTER:
	case HAL_IOCTL_COMMON_DISABLE_ADAPTER:
	case HAL_IOCTL_COMMON_ENABLE_ADAPTER:
	case HAL_IOCTL_COMMON_RESCAN:
		/* hal_ioctl(HAL_IOCTL_COMMON_RESET_ADAPTER) */
		/* hal_ioctl(HAL_IOCTL_COMMON_DISABLE_ADAPTER) */
		/* hal_ioctl(HAL_IOCTL_COMMON_ENABLE_ADAPTER) */
		/* hal_ioctl(HAL_IOCTL_COMMON_RESCAN) */
		rc = (*dte->hal_ioctl)(i_msg->code);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_RESET_DEVICE:
		/* hal_ioctl(HAL_IOCTL_COMMON_RESET_DEVICE, hal_dev_handle_t hal_dev_handle) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_RESET_DEVICE,
							   (hal_dev_handle_t)i_msg->args[0]);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_DELAY_DEVICE:
		/* hal_ioctl(HAL_IOCTL_COMMON_DELAY_DEVICE, hal_dev_handle_t hal_dev_handle, uint32 pause_seconds) */
		if (i_msg->arg_count < 2) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_DELAY_DEVICE,
							   (hal_dev_handle_t)i_msg->args[0], 
							   (uint32)i_msg->args[1]);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_INVALIDATE_DEVICE:
		/* hal_ioctl(HAL_IOCTL_COMMON_INVALIDATE_DEVICE, hal_dev_handle_t hal_dev_handle) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_INVALIDATE_DEVICE,
							   (hal_dev_handle_t)i_msg->args[0]);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_GET_DEVICE_MAP:
		/* hal_ioctl(HAL_IOCTL_COMMON_GET_DEVICE_MAP, wwn_t channel_wwn, hal_device_map_t *device_map) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(hal_device_map_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_GET_DEVICE_MAP,
                               (uint64)i_msg->args[0], 
							   (hal_device_map_t *)o_msg->data);
		if (rc == NFX_OK)
			o_msg->dataLen = sizeof(hal_device_map_t);
		else
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_COMMON_DEV_LOOKUP:
		/* hal_ioctl(HAL_IOCTL_COMMON_DEV_LOOKUP, hal_dev_handle_t dev_handle, hal_dev_lookup_t *dev_lookup) */
		if (i_msg->arg_count < 1) {
            ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
            goto done;
        }
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(hal_dev_lookup_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_DEV_LOOKUP,
                               (hal_dev_handle_t)i_msg->args[0], 
							   (hal_dev_lookup_t *)o_msg->data);
		if (rc == NFX_OK)
			o_msg->dataLen = sizeof(hal_dev_lookup_t);
		else
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS:
		/* hal_ioctl(HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS, hal_fc_link_stats_t *adapter_link_stats) */
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(hal_fc_link_stats_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		async_context = (scsi_ioctl_callback_t *)eee_ramAlloc(sizeof(scsi_ioctl_callback_t));
		if (async_context == NULL) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		async_context->i_edesc = edesc;
		async_context->o_msg = o_msg;
		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS,
							   (hal_fc_link_stats_t *)&o_msg->data[0],
							   scsi_procIoctlCallback, async_context);
		if (rc != NFX_OK) {
			eee_ramDealloc(async_context);
			async_context = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		}
		break;

	case HAL_IOCTL_FC_CLR_ADAPTER_LINK_STATS:

		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_CLR_ADAPTER_LINK_STATS);

		ecode = ( rc == NFX_OK) ? SCSI_MSG_ECODE_OK : SCSI_MSG_ECODE_IOCTL_FAILED;

		break;

	case HAL_IOCTL_FC_GET_DEVICE_LINK_STATS:
		/* hal_ioctl(HAL_IOCTL_FC_GET_DEVICE_LINK_STATS, hal_dev_handle_t hal_dev_handle, hal_fc_link_stats_t *device_link_stats) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
            goto done;
        }
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(hal_fc_link_stats_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_GET_DEVICE_LINK_STATS,
							   (hal_dev_handle_t)i_msg->args[0],
							   (hal_fc_link_stats_t *)o_msg->data);
		if (rc == NFX_OK)
			o_msg->dataLen = sizeof(hal_fc_link_stats_t);
		else
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_FC_CLR_DEVICE_LINK_STATS:
		if(i_msg->arg_count < 1 ) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		break;

	case HAL_IOCTL_FC_LIP:
		/* hal_ioctl(HAL_IOCTL_FC_LIP, wwn_t channel_wwn, byte lip_byte3, byte lip_byte4) */
		if (i_msg->arg_count < 3) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		mlen = sizeof(scsi_ioctlRspMsg_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		async_context = (scsi_ioctl_callback_t *)eee_ramAlloc(sizeof(scsi_ioctl_callback_t));
		if (async_context == NULL) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		async_context->i_edesc = edesc;
		async_context->o_msg = o_msg;
		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_LIP,
                               (uint64)i_msg->args[0],
							   (uint32)i_msg->args[1],
							   (uint32)i_msg->args[2],
							   scsi_procIoctlCallback, async_context);
		if (rc != NFX_OK) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			eee_ramDealloc(async_context);
			async_context = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		}
		break;

	case HAL_IOCTL_FC_LOOP_PORT_BYPASS:
	case HAL_IOCTL_FC_LOOP_PORT_ENABLE:
		/* hal_ioctl(HAL_IOCTL_FC_LOOP_PORT_BYPASS, wwn_t dev_wwn) */
		/* hal_ioctl(HAL_IOCTL_FC_LOOP_PORT_ENABLE, wwn_t dev_wwn) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		mlen = sizeof(scsi_ioctlRspMsg_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		async_context = (scsi_ioctl_callback_t *)eee_ramAlloc(sizeof(scsi_ioctl_callback_t));
		if (async_context == NULL) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		async_context->i_edesc = edesc;
		async_context->o_msg = o_msg;
		rc = (*dte->hal_ioctl)(i_msg->code,
							   (uint64)i_msg->args[0],
							   scsi_procIoctlCallback, async_context);							   
		if (rc != NFX_OK) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			eee_ramDealloc(async_context);
			async_context = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		}
		break;

	case HAL_IOCTL_FC_LOOP_PORT_ENABLE_ALL:
		/* hal_ioctl(HAL_IOCTL_FC_LOOP_PORT_ENABLE_ALL, wwn_t channel_wwn) */
		if (i_msg->arg_count < 1) {
			ecode = SCSI_MSG_ECODE_INVALID_IOCTL_PARMS;
			goto done;
		}
		mlen = sizeof(scsi_ioctlRspMsg_t);
        o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		async_context = (scsi_ioctl_callback_t *)eee_ramAlloc(sizeof(scsi_ioctl_callback_t));
		if (async_context == NULL) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		async_context->i_edesc = edesc;
		async_context->o_msg = o_msg;
		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_LOOP_PORT_ENABLE_ALL,
                               (uint64)i_msg->args[0],
							   scsi_procIoctlCallback, async_context);							   
		if (rc != NFX_OK) {
			eee_ramDealloc(o_msg);
			o_msg = NULL;
			eee_ramDealloc(async_context);
			async_context = NULL;
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		}
		break;

	case HAL_IOCTL_FC_CONNECTION_STATE:
		/* hal_ioctl(HAL_IOCTL_FC_CONNECTION_STATE, uint32 *connection_state); */
		mlen = sizeof(scsi_ioctlRspMsg_t) + sizeof(uint32);
		o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_CONNECTION_STATE,
							   (uint32 *)o_msg->data);
		if (rc == NFX_OK)
			o_msg->dataLen = sizeof(uint32);
		else
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	case HAL_IOCTL_FC_TOPOLOGY:
		/* hal_ioctl(HAL_IOCTL_FC_TOPOLOGY, void *topology_buf, uint32 *topology_buf_len); */
		mlen = sizeof(scsi_ioctlRspMsg_t) + HAL_IOCTL_FC_TOPOLOGY_MAX_LEN;
		o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(mlen);
		if (o_msg == NULL) {
            ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
            goto done;
        }
		bzero(o_msg, mlen);
		o_msg->dataLen = HAL_IOCTL_FC_TOPOLOGY_MAX_LEN;
		rc = (*dte->hal_ioctl)(HAL_IOCTL_FC_TOPOLOGY,
							   (void *)o_msg->data,
							   (uint32 *)&o_msg->dataLen);
		if (rc != NFX_OK)
			ecode = SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	default:
		ecode = SCSI_MSG_ECODE_INVALID_IOCTL_CODE;
	}

 done:
	if (o_msg == NULL) {
		o_msg = (scsi_ioctlRspMsg_t *)eee_ramAlloc(sizeof(scsi_ioctlRspMsg_t));
		bzero(o_msg, sizeof(scsi_ioctlRspMsg_t));
	}

	if (async_context == NULL) {
		o_msg->hdr.mid = SCSI_MID_IOCTL_RSP;
		o_msg->hdr.cHandle = msg->cHandle;
		o_msg->eCode = ecode;
		if (EEE_GET_DTYPE(edesc->hdr.src_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(edesc->hdr.src_port) << EDESC2_DATA_TYPE_SHIFT;
		}

		scsi_sendMessage((uchar8 *)o_msg, (mlen ? mlen : sizeof(scsi_ioctlRspMsg_t)),
						edesc->hdr.src_port, source_port);

		eee_freePacket(edesc);
	}
	
    return(NFX_OK);
}   /* scsi_procIoctlMsg */

int 
scsi_procIoctlCallback(scsi_ioctl_callback_t *context, uint32 ioctl_status)
{
	eee_desc_t *		i_edesc = context->i_edesc;
	scsi_ioctlRspMsg_t *o_msg = context->o_msg;
	scsi_ioctlReqMsg_t *i_msg = (scsi_ioctlReqMsg_t *)i_edesc->bd[1].buf;
	uint32			    source_port = scsiCtrl->scsiDataAddr;

	switch(i_msg->code) {
	case HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS:
		if (ioctl_status == NFX_OK)
			o_msg->dataLen = sizeof(hal_fc_link_stats_t);
		o_msg->eCode = (ioctl_status == NFX_OK) ? SCSI_MSG_ECODE_OK : SCSI_MSG_ECODE_IOCTL_FAILED;
		break;

	default:
		o_msg->eCode = (ioctl_status == NFX_OK) ? SCSI_MSG_ECODE_OK : SCSI_MSG_ECODE_IOCTL_FAILED;
	}

	o_msg->hdr.mid = SCSI_MID_IOCTL_RSP;
	o_msg->hdr.cHandle = i_msg->hdr.cHandle;
	if (EEE_GET_DTYPE(i_edesc->hdr.src_port) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(i_edesc->hdr.src_port) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_ioctlRspMsg_t) + o_msg->dataLen,
					i_edesc->hdr.src_port, source_port);
	eee_freePacket(i_edesc);
	eee_ramDealloc(context);

	return(NFX_OK);
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procRelayCmd
 *
 * Description: 	Point of entry for processing SCSI Relay requests
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 7, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_procRelayReq(eee_desc_t *eDesc, scsi_msgHdr_t *msg)
{
	uint32				cHandle;
	uint32				pktlen;
	scsi_path_t			path;
	scsi_dev_t *		pDev;
	scsi_desc_t *		sd;	
	scsi_mp_desc_t *	smpd;
	uint32				transSeqId;
    esm_event_t			esmEvent;
	scsi_relayReqMsg_t *rmsg = (scsi_relayReqMsg_t *)msg;

#ifdef SCSI_TIMESTAMP
    eeeRTC_t			t;

	eee_rtcStart(&t);
#endif

	cHandle = msg->cHandle;

	/*
	 * Get pointer to SCSI Device Context Block
	 */
	path = msg->path;	
	pDev  = (scsi_dev_t *)scsi_getDevice(path);
	if( pDev == (scsi_dev_t *)NULL ) {

		/* Invalid device id, reject request */
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
		eee_freePacket(eDesc);
		return NFX_OK;
	}

	/* Need to create/update MP state machine? */

	SCSI_DBG(1, ("SCSI_MID_RELAY_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
				 "cdb[0] 0x%x, dataLen %d\n",
				 rmsg->hdr.cHandle,
				 rmsg->hdr.path,
				 rmsg->transSeqId,
				 rmsg->cdb[0],
				 rmsg->dataLen));

	eDesc->bd[1].len -= (sizeof(scsi_relayReqMsg_t)-1);
	eDesc->bd[1].buf += (sizeof(scsi_relayReqMsg_t)-1);
	pktlen = eDesc->hdr.control & EDESC_TOTAL_LEN;	
	pktlen -= (sizeof(scsi_relayReqMsg_t)-1);
	eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
	eDesc->hdr.control |= (pktlen & EDESC_TOTAL_LEN);

	if ( (transSeqId = rmsg->transSeqId) ) {
		if (SCSI_SEQUENCE_MAX(transSeqId) == 0) {
			/* Invalid SEQUENCE MAX, reject request */
			SCSI_DBG(0, ("scsi_procRelayReq: Invalid Sequence Max for RELAY request; rejecting.\n"));
			scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_MSG);
			eee_freePacket(eDesc);
			return NFX_OK;
		}
		if (SCSI_SEQUENCE_ID(transSeqId) == 0) {
			/* Invalid SEQUENCE ID, reject request */
			SCSI_DBG(0, ("scsi_procRelayReq: Invalid Sequence Id for RELAY request; rejecting.\n"));
			scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_MSG);
			eee_freePacket(eDesc);
			return NFX_OK;
		}
			
		/* First packet in multi-packet sequence ? */
		if (SCSI_SEQUENCE_ID(transSeqId) == 1) {
			smpd = scsi_MPEntryNew(pDev, cHandle);
			VERIFY(smpd);
			smpd->firstEDesc = smpd->lastEDesc = eDesc;
			smpd->transactionId = SCSI_TRANSACTION_ID(transSeqId);
			smpd->maxSequenceId = SCSI_SEQUENCE_MAX(transSeqId);
			smpd->lastSequenceId = SCSI_SEQUENCE_ID(transSeqId);
		}
		else {
			smpd = scsi_MPEntryLookup(pDev, cHandle);
			if (smpd == NULL) {
				SCSI_DBG(0, ("scsi_procRelayReq: Cannot find SMP entry; First packet dropped?; aborting.\n"));
				VERIFY(0); /* XXX need to finish up abort */
			}
			if (smpd->transactionId != SCSI_TRANSACTION_ID(transSeqId)) {
				SCSI_DBG(0, ("scsi_procRelayReq: Invalid transaction ID; aborting.\n"));
				VERIFY(0); /* XXX need to finish up abort */
			}
			if (SCSI_SEQUENCE_ID(transSeqId) - 1 != smpd->lastSequenceId) {
				SCSI_DBG(0, ("scsi_procRelayReq: Out of order packet received for RELAY request; aborting.\n"));
				VERIFY(0);	/* XXX need to finish up abort */
			}

			if (EEE_GET_NUM_BUFS(smpd->lastEDesc) == EEE_MAX_BUFS-1 && 
				(smpd->lastEDesc->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE)) {
				SCSI_DBG(0, ("scsi_procRelayReq: WRITE request too large; cannot chain thru last buffer; rejecting.\n"));
				scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_MSG);
				eee_freePacket(eDesc);
				return NFX_OK;
			}

			smpd->lastEDesc->hdr.control |= EDESC_COMMAND_CHAIN_LAST_BUFFER;
			smpd->lastEDesc->hdr.control &= ~EDESC_BD_CMD_IN_USE;
			smpd->lastEDesc->bd[EEE_MAX_BUFS-1].buf = (uchar8 *)eDesc;
			smpd->lastEDesc = eDesc;
			smpd->lastSequenceId = SCSI_SEQUENCE_ID(transSeqId);
		}

		/* Last packet in multi-packet sequence ? */
		if (smpd->lastSequenceId < smpd->maxSequenceId) {
			return NFX_OK;
		}
		else {
			eDesc = smpd->firstEDesc;
			scsi_MPEntryDestroy(smpd);
		}
	}

	/*
	 * Allocate SCSI descriptor
	 */
	sd = (scsi_desc_t *)scsi_allocDesc();
	if (sd == (scsi_desc_t *)NULL) {
		SCSI_DBG(0, ("scsi_procRelayReq: Failed to allocate SCSI desc.\n"));
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NO_RESOURCE);  
		eee_freePacket(eDesc);
		return NFX_OK;
	}
#ifdef SCSI_TIMESTAMP
	sd->cmdTS = t;
#endif

	sd->cHandle = cHandle;
	sd->path    = path;
#if !defined(SCSI_TEST) || SDM_DEV_INIT
	sd->rspDest = eDesc->hdr.src_port;
#else
	sd->rspDest = (EEE_USER_DATA << EDESC2_DATA_TYPE_SHIFT) & EDESC2_DATA_TYPE_MASK;
#endif
	sd->pdev    = pDev;

	sd->eDesc   = eDesc;
	sd->control = SD_OPCODE_RELAY;
	//-- Make sure the we talk to the corret LUN.
	SCSI_PATH_SET_LUN(sd->path, rmsg->lun); 

	//-- Setup data parameters.
	if( rmsg->dataInLen > 0 ) {
		sd->expectDataLen = rmsg->dataInLen;
		sd->status |= SD_DEV_DATA_IN_EXPECT;
	}
	else
	if( rmsg->dataLen > 0 ) {
		sd->status |= SD_DEV_DATA_OUT_EXPECT;
	}
			
	//-- Setup retry and timeout values
	sd->retryCnt      = rmsg->retryCnt;
	sd->timeOutCnt    = rmsg->timeOutCnt;
	sd->maxTimeToLive = rmsg->maxTimeToLive;

	memcpy(sd->cdb, (char *)(rmsg->cdb), 16); 
	pDev->relayReqCnt++;
	scsiCtrl->stats.relayCnt++;

	if (pDev->flags.fAllowChain == 0)
	    sd->control |= SD_FLAG_NOCOALLESCE;

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pDev->smHandle, &esmEvent);

	return(NFX_OK);
} /* scsi_procRelayReq */

static int
scsi_queue_tape_op(scsi_dev_t *pDev, eee_desc_t *eDesc, uint64 iocookie, uint32 transSeqId, uint32 reqType)
{
	scsi_tape_io_q_t	*scsi_msg;

	// if a new read sequence has been started, then we need to reset our state
	if(reqType == SCSI_MID_READ_REQ) {
		if(pDev->tape_iocookie >= iocookie) {
			pDev->tape_iocookie = iocookie + 1; // set it to next cookie to be recd.
			return 0;
		}
	}

	else if(reqType == SCSI_MID_WRITE_REQ) {
		if(SCSI_SEQUENCE_ID(transSeqId) == 1 && pDev->tape_iocookie >= iocookie) {
			pDev->tape_iocookie = iocookie;
			return 0;
		}
		else if(SCSI_SEQUENCE_ID(transSeqId) != 1 && pDev->tape_iocookie == iocookie) {
			// if last packet in sequence has been received, change cookie value
			if(SCSI_SEQUENCE_ID(transSeqId) == SCSI_SEQUENCE_MAX(transSeqId))
				pDev->tape_iocookie++;

			return 0;
		}
	}
	else return 0;

	SCSI_DBG(1, ("scsi_queue_tape_op: Queueing req. with iocookie %lld for 0x%llx:%d\n", 
				iocookie, pDev->wwn, pDev->lun));

	// we recd. iocookie greater than what we expected, so we are going to queue this msg
	scsi_msg = (scsi_tape_io_q_t*)eee_ramAlloc(sizeof(scsi_tape_io_q_t));
	if( !scsi_msg) {
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "MSG QUEUE ENTRY");
		return 0;
	}
	memset((void *)scsi_msg, 0, sizeof(scsi_tape_io_q_t));
	scsi_msg->msg = (void *)eDesc;
	scsi_msg->iocookie = iocookie;

	if(pDev->tape_io_q == NULL) {
		pDev->tape_io_q = (void *)scsi_msg;
	} 
	else {
		scsi_tape_io_q_t	*prev;
		scsi_tape_io_q_t 	*cur;

		if(scsi_msg->iocookie < ((scsi_tape_io_q_t *)pDev->tape_io_q)->iocookie) {
			scsi_msg->next = (scsi_tape_io_q_t *)pDev->tape_io_q;
			pDev->tape_io_q = (void *)scsi_msg;
		}
		else {
			prev = (scsi_tape_io_q_t *)pDev->tape_io_q;
			cur = ((scsi_tape_io_q_t *)pDev->tape_io_q)->next;
			while(cur) {
				if(scsi_msg->iocookie < cur->iocookie) {
					scsi_msg->next = cur;
					prev->next = scsi_msg;
					break;
				}
				prev = cur;
				cur = cur->next;
			}
			if(!cur) {
				prev->next = scsi_msg;
			}
		}
	}
	return 1;
}

static void
scsi_proc_tape_queue_ops(scsi_dev_t *pDev)
{
	scsi_tape_io_q_t	*scsi_msg;
	scsi_tape_io_q_t	*tmp_msg;
	eee_desc_t			*eDesc;
	eee_descBuf_t		*bd;
	scsi_msgHdr_t		*msg;
	
	while(pDev->tape_io_q) {
		scsi_msg = (scsi_tape_io_q_t *)pDev->tape_io_q;
		if(pDev->tape_iocookie == scsi_msg->iocookie) {
			eDesc = (eee_desc_t *)scsi_msg->msg;
			bd = &eDesc->bd[1];
			msg = (scsi_msgHdr_t *)(bd->buf);

			SCSI_DBG(0, ("scsi_proc_tape_queue_ops: Processing req. with iocookie %lld for 0x%llx:%d\n", 
							scsi_msg->iocookie, pDev->wwn, pDev->lun));

			switch(msg->mid) {
				case SCSI_MID_READ_REQ:
						scsi_procReadReq(eDesc, msg);
						break;
				case SCSI_MID_WRITE_REQ:
						scsi_procWriteReq(eDesc, msg);
						break;
			}
			tmp_msg = scsi_msg;
			pDev->tape_io_q = (void *)scsi_msg->next;
			eee_ramDealloc(tmp_msg);
		}	
		else
			break;
	}
}


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procReadReq
 *
 * Description: 	Point of entry for processing SCSI Read requests
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 7, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_procReadReq(eee_desc_t *eDesc, scsi_msgHdr_t *msg)
{
	uint32				cHandle;
	scsi_path_t			path;
	scsi_dev_t *		pDev;
	uint32				transSeqId;
	scsi_desc_t *		sd;	
    esm_event_t			esmEvent;
	scsi_readReqMsg_t *	rmsg = (scsi_readReqMsg_t *)msg;
#ifdef SCSI_TIMESTAMP
	eeeRTC_t			t;

	eee_rtcStart(&t);
#endif

	cHandle = msg->cHandle;

	/*
	 * Get pointer to SCSI Device Context Block
	 */
	path = msg->path;	
	pDev = (scsi_dev_t *)scsi_getDevice(path);
	if (pDev == NULL || !pDev->flags.fOpenMode || (pDev->type == SCSI_DEVTYPE_DISK && !pDev->blockSize)) {
		/* Invalid device id, reject request */
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
		eee_freePacket(eDesc);
		return NFX_OK;
	}

	SCSI_DBG(1, ("SCSI_MID_READ_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
				 "blkOffset %d, blkCnt %d\n",
				 rmsg->hdr.cHandle,
				 rmsg->hdr.path,
				 rmsg->transSeqId,
				 rmsg->blkOffset,
				 rmsg->blkCnt));

	if(pDev->type == SCSI_DEVTYPE_TAPE && rmsg->iocookie) {
		SCSI_DBG(1, ("SCSI_MID_READ_REQ: iocookie %lld expected %lld\n", 
					rmsg->iocookie, pDev->tape_iocookie));
	}

	/* If this is a tape read, then it might need to be queued up */
	if(pDev->type == SCSI_DEVTYPE_TAPE && scsi_queue_tape_op(pDev, eDesc, rmsg->iocookie, rmsg->transSeqId, SCSI_MID_READ_REQ))
		return 0;

	/* For now, we treat HI_PRI as equivalent to NO_REORDER */
	if (rmsg->attributes & SCSI_MSG_ATTR_HI_PRI)
		rmsg->attributes |= SCSI_MSG_ATTR_NO_REORDER;

#if 0 //-- FSDEBUG
	/* Kick open write gate if disk device and we are requested to do so */
	if (pDev->type == SCSI_DEVTYPE_DISK && 
		(rmsg->attributes & SCSI_MSG_ATTR_KICK_WRITE_GATE)) {
		scsi_write_gate_timer_restart(pDev);
	}
#endif

	if ( (transSeqId = rmsg->transSeqId) ) {
		if (SCSI_SEQUENCE_MAX(transSeqId))
			SCSI_DBG(0, ("scsi_procReadReq: Ignoring non-zero Sequence Max for READ request.\n"));
		if (SCSI_SEQUENCE_ID(transSeqId))
			SCSI_DBG(0, ("scsi_procReadReq: Ignoring non-zero Sequence ID for READ request.\n"));
	}

	/* Respond immediately to zero-length read */
	if (rmsg->blkCnt == 0) {
		scsi_readRspMsg_t *o_msg;
		uint32			   source_port = scsiCtrl->scsiDataAddr;
		uint32			   dest_port = eDesc->hdr.src_port;
		uint32             slot = EEE_GET_SLOT_NUM(source_port);

		pDev->readReqCnt++;

		if( slot == 1 ) {
			if( EEE_GET_DTYPE(dest_port) == EEE_SAC ) {
				scsiCtrl->stats.readUserReqCnt_fp++;
			}
			else {
				scsiCtrl->stats.readMetaReqCnt_fp++;
			}
		}
		else {
			scsiCtrl->stats.readMetaReqCnt_ssc++;
		}


		o_msg = (scsi_readRspMsg_t *)eee_ramAlloc(sizeof(scsi_readRspMsg_t));
		memset(o_msg, 0, sizeof(scsi_readRspMsg_t));

		o_msg->hdr.mid     = SCSI_MID_READ_RSP;
		o_msg->hdr.cHandle = rmsg->hdr.cHandle;
		o_msg->eCode       = SCSI_MSG_ECODE_OK;

		if (EEE_GET_DTYPE(dest_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(dest_port) << EDESC2_DATA_TYPE_SHIFT;
		}
		scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_readRspMsg_t), dest_port, source_port);
		eee_freePacket(eDesc);

		return(NFX_OK);
	}

	/* Verify that I/O does *not* span a checksum block. */
	if ((pDev->flags.fAllowBlock_CS) && !(rmsg->attributes & SCSI_MSG_ATTR_PHY_OP) &&
		rmsg->blkOffset / (BCSUM_BLOCKSIZE / BLOCKSIZE_512) != (rmsg->blkOffset + rmsg->blkCnt - 1) / (BCSUM_BLOCKSIZE / BLOCKSIZE_512)) {
		scsi_readRspMsg_t * o_msg;
		uint32				source_port = scsiCtrl->scsiDataAddr;
		uint32				dest_port = eDesc->hdr.src_port;
		uint32				slot = EEE_GET_SLOT_NUM(source_port);

		E_LOG (class_1, err_s, eee_getApplicationId("scsi"), 0, 0, 0, 
			   ("scsi: Invalid spanning request parameters on read request: scsi_id 0x%llx, blkOffset %d, blkCnt %d.",
				pDev->id,
				rmsg->blkOffset,
				rmsg->blkCnt));

		SCSI_DBG(0, ("SCSI_MID_READ_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
					 "scsi_id 0x%llx, blkOffset %d, blkCnt %d -- invalid spanning request parameters.\n",
					 rmsg->hdr.cHandle,
					 rmsg->hdr.path,
					 rmsg->transSeqId,
					 pDev->id,
					 rmsg->blkOffset,
					 rmsg->blkCnt));

		pDev->readReqCnt++;

		if( slot == 1 ) {
			if( EEE_GET_DTYPE(dest_port) == EEE_SAC ) {
				scsiCtrl->stats.readUserReqCnt_fp++;
			}
			else {
				scsiCtrl->stats.readMetaReqCnt_fp++;
			}
		}
		else {
			scsiCtrl->stats.readMetaReqCnt_ssc++;
		}		
		
		o_msg = (scsi_readRspMsg_t *)eee_ramAlloc(sizeof(scsi_readRspMsg_t));
		memset(o_msg, 0, sizeof(scsi_readRspMsg_t));

		o_msg->hdr.mid     = SCSI_MID_READ_RSP;
		o_msg->hdr.cHandle = rmsg->hdr.cHandle;
		o_msg->eCode       = SCSI_MSG_ECODE_ILLEGAL_REQUEST;
		o_msg->blkOffset   = rmsg->blkOffset;
		o_msg->blkCnt      = rmsg->blkCnt;

		if (EEE_GET_DTYPE(dest_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(dest_port) << EDESC2_DATA_TYPE_SHIFT;
		}
		scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_readRspMsg_t), dest_port, source_port);
		eee_freePacket(eDesc);
		return(NFX_OK);
	}

	/*
	 * Allocate SCSI descriptor
	 */
	sd = (scsi_desc_t *)scsi_allocDesc();
	if (sd == (scsi_desc_t *)NULL) {
		SCSI_DBG(0, ("scsi_procReadReq: Failed to allocate SCSI desc.\n"));
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NO_RESOURCE);  
		eee_freePacket(eDesc);
		return NFX_OK;
	}
#ifdef SCSI_TIMESTAMP
	sd->cmdTS = t;
#endif

	sd->cHandle = cHandle;
	sd->path    = path;
#if !defined(SCSI_TEST) || SDM_DEV_INIT
	sd->rspDest = eDesc->hdr.src_port;
#else
	sd->rspDest = (EEE_USER_DATA << EDESC2_DATA_TYPE_SHIFT) & EDESC2_DATA_TYPE_MASK;
#endif
	sd->pdev    = pDev;
	sd->eDesc   = 0;
	sd->control = SD_OPCODE_READ | 
		((rmsg->attributes & SCSI_MSG_ATTR_PHY_OP) ? SD_FLAG_PHY_OP : 0) | 
		((rmsg->attributes & SCSI_MSG_ATTR_NO_REORDER) ? SD_FLAG_NOREORDER : 0) |
		((rmsg->attributes & SCSI_MSG_ATTR_EXPRESS) ? SD_FLAG_EXPRESS : 0) |
		((rmsg->attributes & SCSI_MSG_ATTR_NO_GUAR_DELIVERY) ? SD_FLAG_NO_GUAR_DEL : 0) |
		((pDev->flags.fAllowEmbedded_CS) ? SD_FLAG_EMBEDDED_CS : 0) | 
		((pDev->flags.fAllowChain == 0) ? SD_FLAG_NOCOALLESCE : 0);
	sd->dataDest= rmsg->dataDest;

	/* if block cs is allowed and we don't want to do phy. read */
	if (pDev->flags.fAllowBlock_CS && !(sd->control & SD_FLAG_PHY_OP)) {
		sd->control |= SD_FLAG_BLOCK_CS;
		scsi_map_logical_to_physical(rmsg->blkOffset, rmsg->blkCnt, &sd->lba, &sd->lbCnt);
	}
	else {
		sd->lba = rmsg->blkOffset;
		sd->lbCnt = rmsg->blkCnt;
	}

	if (pDev->type == SCSI_DEVTYPE_DISK && !(sd->control & SD_FLAG_PHY_OP)) {
		/* 
		 * XXX Partitioning mapping will take place here. For now just
		 * offset. 
		 */
		sd->lba += pDev->offset;
	}

	pDev->readReqCnt++;
	pDev->readByteCnt += (pDev->blockSize * sd->lbCnt);

	if( EEE_GET_SLOT_NUM(sd->rspDest) == 1 ) {
		if( EEE_GET_DTYPE(sd->rspDest) == EEE_SAC ) {
			scsiCtrl->stats.readUserReqCnt_fp++;
		}
		else {
			scsiCtrl->stats.readMetaReqCnt_fp++;
		}
	}
	else {
		scsiCtrl->stats.readMetaReqCnt_ssc++;
	}

	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = pDev->id;
		scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_READ_REQ;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = cHandle;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

    //-- Look for loopback attribute.
    if(rmsg->attributes & SCSI_MSG_ATTR_LOOPBACK)   {
        eee_desc_t *ed;
        char       *lbuf;
        uint32 len;
 
        /*
         * Map back to logical address/count.
         */
        if (pDev->type == SCSI_DEVTYPE_DISK) {
            /*
             * XXX Partitioning mapping will take place here. For now
             * just offset.
             */
            sd->lba -= pDev->offset;
        }
        if (pDev->flags.fAllowBlock_CS) {
            scsi_map_physical_to_logical(sd->lba, sd->lbCnt, &sd->lba, &sd->lbCnt);
        }
 
        len = sd->lbCnt * 512;
 
            //-- Allocate Ultra Buffer
        lbuf =  eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_ULTRA,0);
        if( !lbuf ) {
            SCSI_DBG(0, ("scsi_procReadReq: No loopback buffer available. Dropping Request.\n"));
            return(NFX_ERR);
        }
 
        //-- Allocate edesc
        ed = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
        if( !ed ) {
            eee_deallocateBuffer(lbuf);
            SCSI_DBG(0, ("scsi_procReadReq: No loopback eee desc available. Dropping Request.\n"));
            return(NFX_ERR);
        }
 
        lbuf += 128;
        ed->bd[1].buf = lbuf;
        ed->bd[1].len = len;
        ed->hdr.src_port  = 0;
        ed->hdr.dest_port = 0;
        ed->hdr.control = ((len & EDESC_TOTAL_LEN) |
                          ((1<<EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK));
        sd->eDesc = ed;

        scsi_sendReadRspWithData( sd, SCSI_MSG_ECODE_OK, 0, 0, 0);

		//-- Release the read req packet.
		eee_freePacket(eDesc);
        scsi_deallocDesc(sd);
        return(NFX_OK);
    }

	//-- Release the read req packet.
	eee_freePacket(eDesc);

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pDev->smHandle, &esmEvent);

	/* If this is a tape read, then check if any queues messages can be processed */
	if(pDev->type == SCSI_DEVTYPE_TAPE)
		scsi_proc_tape_queue_ops(pDev);

	scsi_debug("[scsi_procDevCmd] Leave", SCSI_DBL_3);
	return(NFX_OK);
} /* scsi_procReadReq */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procWriteReq
 *
 * Description: 	Point of entry for processing SCSI Write requests
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 7, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_procWriteReq(eee_desc_t *eDesc, scsi_msgHdr_t *msg)
{
	uint32				cHandle;
	uint32				pktlen;
	uint32				packet_csum;
	uint32				total_csum;
	scsi_path_t			path;
	scsi_dev_t *		pDev;
	scsi_desc_t *		sd;	
	scsi_mp_desc_t *	smpd;
	uint32				transSeqId;
    esm_event_t			esmEvent;
	scsi_writeReqMsg_t *wmsg = (scsi_writeReqMsg_t *)msg;
    int32			    ttlblocks = 0, ttlpktlen = 0;
#ifdef SCSI_TIMESTAMP
	eeeRTC_t			t;

	eee_rtcStart(&t);
#endif

	scsi_debug("[scsi_procDevCmd] Enter", SCSI_DBL_3);
	cHandle = msg->cHandle;

	/*
	 * Get pointer to SCSI Device Context Block
	 */
	path = msg->path;	
	pDev = (scsi_dev_t *)scsi_getDevice(path);
	if (pDev == NULL || !pDev->flags.fOpenMode || (pDev->type == SCSI_DEVTYPE_DISK && !pDev->blockSize)) {
		/* Invalid device id, reject request */
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
		eee_freePacket(eDesc);
		return NFX_OK;
	}

	SCSI_DBG(1, ("SCSI_MID_WRITE_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
				 "blkOffset %d, blkCnt %d, checkSum 0x%x\n",
				 wmsg->hdr.cHandle,
				 wmsg->hdr.path,
				 wmsg->transSeqId,
				 wmsg->blkOffset,
				 wmsg->blkCnt,
				 wmsg->checkSum));

	if(pDev->type == SCSI_DEVTYPE_TAPE && wmsg->iocookie) {
		SCSI_DBG(1, ("SCSI_MID_WRITE_REQ: iocookie %lld expected %lld\n", 
					wmsg->iocookie, pDev->tape_iocookie));
	}

	/* If this is a tape write, then it might need to be queued up */
	if(pDev->type == SCSI_DEVTYPE_TAPE && scsi_queue_tape_op(pDev, eDesc, wmsg->iocookie, wmsg->transSeqId, SCSI_MID_WRITE_REQ))
		return 0;

	eDesc->bd[1].buf += sizeof(scsi_writeReqMsg_t);
	eDesc->bd[1].len -= sizeof(scsi_writeReqMsg_t);
	pktlen = eDesc->hdr.control & EDESC_TOTAL_LEN;	
	pktlen -= sizeof(scsi_writeReqMsg_t);
	eDesc->hdr.control &= ~EDESC_TOTAL_LEN;
	eDesc->hdr.control |= (pktlen & EDESC_TOTAL_LEN);

	/* For now, we treat HI_PRI as equivalent to NO_REORDER */
	if (wmsg->attributes & SCSI_MSG_ATTR_HI_PRI)
		wmsg->attributes |= SCSI_MSG_ATTR_NO_REORDER;
	
#if 0 //-- FSDEBUG
	/* Kick open write gate if disk device and we are requested to do so */
	if (pDev->type == SCSI_DEVTYPE_DISK && 
		(wmsg->attributes & SCSI_MSG_ATTR_KICK_WRITE_GATE)) {
		scsi_write_gate_timer_restart(pDev);
	}
#endif

#if 0 //-- Taking this all out for now: FS

#if 1
	/* XXX To enable honouring of write gate, comment out the next line */
	wmsg->attributes |= SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
#endif

	/* If disk device and write gate closed, reject message */
	if (pDev->type == SCSI_DEVTYPE_DISK && 
		!(wmsg->attributes & SCSI_MSG_ATTR_IGNORE_WRITE_GATE) && 
		pDev->write_gate_timer_running == 0) {
		scsi_writeRspMsg_t *o_msg;
		uint32				source_port = scsiCtrl->scsiDataAddr;
		uint32				dest_port = eDesc->hdr.src_port;

		E_LOG (class_1, err_s, eee_getApplicationId("scsi"), 0, 0, 0, 
			   ("scsi: Write rejected; write gate closed: scsi_id 0x%llx, blkOffset %d, blkCnt %d.",
				pDev->id,
				wmsg->blkOffset,
				wmsg->blkCnt));

		SCSI_DBG(0, ("SCSI_MID_WRITE_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
					 "scsi_id 0x%llx, blkOffset %d, blkCnt %d -- write rejected; write gate closed.\n",
					 wmsg->hdr.cHandle,
					 wmsg->hdr.path,
					 wmsg->transSeqId,
					 pDev->id,
					 wmsg->blkOffset,
					 wmsg->blkCnt));

		o_msg = (scsi_writeRspMsg_t *)eee_ramAlloc(sizeof(scsi_writeRspMsg_t));
		memset(o_msg, 0, sizeof(scsi_writeRspMsg_t));

		o_msg->hdr.mid     = SCSI_MID_WRITE_RSP;
		o_msg->hdr.cHandle = wmsg->hdr.cHandle;
		o_msg->eCode       = SCSI_MSG_ECODE_WRITE_GATE_CLOSED;
		o_msg->blkOffset   = wmsg->blkOffset;
		o_msg->blkCnt      = wmsg->blkCnt;

		if (EEE_GET_DTYPE(dest_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(dest_port) << EDESC2_DATA_TYPE_SHIFT;
		}
		scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_writeRspMsg_t), dest_port, source_port);
		eee_freePacket(eDesc);
		return(NFX_OK);
	}
#endif

	/* 
	 * Verify packet checksum, if valid
	 */
	if (pDev->flags.fLocalCSVerify && wmsg->checkSum != 0) {
		packet_csum = scsi_compute_checksum(eDesc);
		if (packet_csum != wmsg->checkSum) {
			pDev->writeCSECnt++;
			E_LOG (class_1, err_s, eee_getApplicationId("scsi"), 0, 0, 0, 
				   ("scsi: write data checksum verification failed: "
					"scsi_id 0x%llx, blkOffset %d, blkCnt %d -- Msg CheckSum: %x, Calculated: %x.",
					pDev->id, wmsg->blkOffset, wmsg->blkCnt,
					wmsg->checkSum, packet_csum));
			SCSI_DBG(0, ("SCSI write data checksum failed -- "
						 "Msg CheckSum: %x, Calculated: %x\n", 
						 wmsg->checkSum,
						 packet_csum));
			SCSI_DBG(0, ("SCSI_MID_WRITE_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
						 "blkOffset %d, blkCnt %d\n",
						 wmsg->hdr.cHandle,
						 wmsg->hdr.path,
						 wmsg->transSeqId,
						 wmsg->blkOffset,
						 wmsg->blkCnt));
			eee_descDisplay(eDesc);
			if( scsi_debug_drop == 1 ) {
				NFX_ASSERT(FALSE);
			}
			else {
				scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_DATA_CHECKSUM_ERR);
				eee_freePacket(eDesc);
				return(NFX_OK);
			}
		}
	}

	/* Verify that I/O does *not* span a checksum block. */
	if ((pDev->flags.fAllowBlock_CS) && !(wmsg->attributes & SCSI_MSG_ATTR_PHY_OP) &&
	    wmsg->blkOffset / (BCSUM_BLOCKSIZE / BLOCKSIZE_512) != (wmsg->blkOffset + wmsg->blkCnt - 1) / (BCSUM_BLOCKSIZE / BLOCKSIZE_512)) {
		scsi_writeRspMsg_t *o_msg;
		uint32				source_port = scsiCtrl->scsiDataAddr;
		uint32				dest_port = eDesc->hdr.src_port;

		E_LOG (class_1, err_s, eee_getApplicationId("scsi"), 0, 0, 0, 
			   ("scsi: Invalid spanning request parameters on write request: scsi_id 0x%llx, blkOffset %d, blkCnt %d.",
				pDev->id,
				wmsg->blkOffset,
				wmsg->blkCnt));

		SCSI_DBG(0, ("SCSI_MID_WRITE_REQ: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
					 "scsi_id 0x%llx, blkOffset %d, blkCnt %d -- invalid spanning request parameters.\n",
					 wmsg->hdr.cHandle,
					 wmsg->hdr.path,
					 wmsg->transSeqId,
					 pDev->id,
					 wmsg->blkOffset,
					 wmsg->blkCnt));

		pDev->writeReqCnt++;
		o_msg = (scsi_writeRspMsg_t *)eee_ramAlloc(sizeof(scsi_writeRspMsg_t));
		memset(o_msg, 0, sizeof(scsi_writeRspMsg_t));

		o_msg->hdr.mid     = SCSI_MID_WRITE_RSP;
		o_msg->hdr.cHandle = wmsg->hdr.cHandle;
		o_msg->eCode       = SCSI_MSG_ECODE_ILLEGAL_REQUEST;
		o_msg->blkOffset   = wmsg->blkOffset;
		o_msg->blkCnt      = wmsg->blkCnt;

		if (EEE_GET_DTYPE(dest_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(dest_port) << EDESC2_DATA_TYPE_SHIFT;
		}
		scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_writeRspMsg_t), dest_port, source_port);
		eee_freePacket(eDesc);
		return(NFX_OK);
	}

	/* 
	 * Invalidate checksum (to be written to disk) if (a) the LBA is not correctly
	 * aligned or (b) doesn't have the required length, but PADDING is not requested.
	 */
	if ((pDev->flags.fAllowBlock_CS) &&
		((wmsg->blkOffset % (BCSUM_BLOCKSIZE / BLOCKSIZE_512)) ||
	    ((wmsg->blkCnt != (BCSUM_BLOCKSIZE / BLOCKSIZE_512)) && !(wmsg->attributes & SCSI_MSG_ATTR_PAD_WRITE)))) {
		wmsg->checkSum = 0;
		pDev->writeSub8kCnt++;
	}		

	/* Need to create/update MP state machine? */
	if ( (transSeqId = wmsg->transSeqId) ) {
		if (SCSI_SEQUENCE_MAX(transSeqId) == 0) {
			/* Invalid SEQUENCE MAX, reject request */
			SCSI_DBG(0, ("scsi_procWriteReq: Invalid Sequence Max for WRITE request.\n"));
			if (wmsg->attributes & SCSI_MSG_ATTR_NO_GUAR_DELIVERY)
				scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_MSG);
			else {
				scsi_sendAck(eDesc->hdr.src_port, cHandle, 1);
				++pDev->nacksTransmitted;
			}
			eee_freePacket(eDesc);
			return(NFX_OK);
		}
		if (SCSI_SEQUENCE_ID(transSeqId) == 0) {
			/* Invalid SEQUENCE ID, reject request */
			SCSI_DBG(0, ("scsi_procWriteReq: Invalid Sequence Id for WRITE request.\n"));
			if (wmsg->attributes & SCSI_MSG_ATTR_NO_GUAR_DELIVERY)
				scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_MSG);
			else {
				scsi_sendAck(eDesc->hdr.src_port, cHandle, 1);
				++pDev->nacksTransmitted;
			}
			eee_freePacket(eDesc);
			return(NFX_OK);
		}
			
		/* First packet in multi-packet sequence ? */
		if (SCSI_SEQUENCE_ID(transSeqId) == 1) {
			smpd = scsi_MPEntryNew(pDev, cHandle);
			VERIFY(smpd);
			smpd->firstEDesc = smpd->lastEDesc = eDesc;
			smpd->transactionId = SCSI_TRANSACTION_ID(transSeqId);
			smpd->maxSequenceId = SCSI_SEQUENCE_MAX(transSeqId);
			smpd->lastSequenceId = SCSI_SEQUENCE_ID(transSeqId);
			smpd->accumulated_checksum = wmsg->checkSum;
            smpd->totalpktlen = pktlen;
            smpd->totalBlks   = wmsg->blkCnt;
			smpd->no_guaranteed_delivery = (wmsg->attributes & SCSI_MSG_ATTR_NO_GUAR_DELIVERY);
			scsi_start_exch_timer(smpd);
		}
		else {
			smpd = scsi_MPEntryLookup(pDev, cHandle);
			if (smpd == NULL) {
				SCSI_DBG(0, ("scsi_procWriteReq: Cannot find SMP entry; First packet dropped?.\n"));
				if (wmsg->attributes & SCSI_MSG_ATTR_NO_GUAR_DELIVERY)
					scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_TXN_SEQ);
				else {
					scsi_sendAck(eDesc->hdr.src_port, cHandle, 1);
					++pDev->nacksTransmitted;
				}
				eee_freePacket(eDesc);
                return(NFX_OK);
			}
			if (smpd->transactionId != SCSI_TRANSACTION_ID(transSeqId)) {
				SCSI_DBG(0, ("scsi_procWriteReq: Invalid transaction ID; got 0x%x, expected 0x%x.\n", 
							 SCSI_TRANSACTION_ID(transSeqId), smpd->transactionId));
				if (smpd->no_guaranteed_delivery) 
					scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_TXN_SEQ);
				else {
					scsi_sendAck(eDesc->hdr.src_port, cHandle, 1);
					++pDev->nacksTransmitted;
				}
				scsi_stop_exch_timer(smpd);
		        scsi_destroyDescChain(smpd->firstEDesc);
			    scsi_MPEntryDestroy(smpd);
                eee_freePacket(eDesc);
                return(NFX_OK);
			}
			if (SCSI_SEQUENCE_ID(transSeqId) - 1 != smpd->lastSequenceId) {
				SCSI_DBG(0, ("scsi_procWriteReq: Out of order packet received for WRITE request; got 0x%x, expected 0x%x.\n",
							 SCSI_SEQUENCE_ID(transSeqId), smpd->lastSequenceId + 1));
				if (smpd->no_guaranteed_delivery) 
					scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_TXN_SEQ);
				else {
					scsi_sendAck(eDesc->hdr.src_port, cHandle, 1);
					++pDev->nacksTransmitted;
				}
				scsi_stop_exch_timer(smpd);
		        scsi_destroyDescChain(smpd->firstEDesc);
			    scsi_MPEntryDestroy(smpd);
                eee_freePacket(eDesc);
                return(NFX_OK);
			}

			if (EEE_GET_NUM_BUFS(smpd->lastEDesc) == EEE_MAX_BUFS-1 && 
				(smpd->lastEDesc->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE)) {
				SCSI_DBG(0, ("scsi_procWriteReq: WRITE request too large; cannot chain thru last buffer.\n"));
				scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_MSG);
				scsi_stop_exch_timer(smpd);
		        scsi_destroyDescChain(smpd->firstEDesc);
			    scsi_MPEntryDestroy(smpd);
				return(NFX_OK);
			}

			smpd->lastEDesc->hdr.control |= EDESC_COMMAND_CHAIN_LAST_BUFFER;
			smpd->lastEDesc->hdr.control &= ~EDESC_BD_CMD_IN_USE;
			smpd->lastEDesc->bd[EEE_MAX_BUFS-1].buf = (uchar8 *)eDesc;
			smpd->lastEDesc = eDesc;
			smpd->lastSequenceId = SCSI_SEQUENCE_ID(transSeqId);
			smpd->accumulated_checksum = scsi_add_checksums(smpd->accumulated_checksum, wmsg->checkSum);
            smpd->totalpktlen += pktlen;
		}

		/* Last packet in multi-packet sequence ? */
		if (smpd->lastSequenceId < smpd->maxSequenceId) {
			return(NFX_OK);
		}
		else {
			if (!smpd->no_guaranteed_delivery)
				scsi_sendAck(eDesc->hdr.src_port, cHandle, 0); 
			scsi_stop_exch_timer(smpd);
			eDesc = smpd->firstEDesc;
			total_csum = smpd->accumulated_checksum;
            ttlpktlen = smpd->totalpktlen;
            ttlblocks = smpd->totalBlks;
			scsi_MPEntryDestroy(smpd);
		}
	}
	else {
		total_csum = wmsg->checkSum;
        ttlpktlen = pktlen;
        ttlblocks = wmsg->blkCnt;
	}

	if ((pDev->type == SCSI_DEVTYPE_DISK) && (ttlblocks*512 != ttlpktlen)) {
		SCSI_DBG(0, ("scsi_procWriteReq: Invalid size: blkcnt %d, pktlen %d\n", ttlblocks, ttlpktlen));
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, cHandle, SCSI_MSG_ECODE_INVALID_LEN);
		scsi_destroyDescChain(eDesc);
		return(NFX_OK);
    }

	/* If tape write, reject if at EOM */
	if (pDev->type == SCSI_DEVTYPE_TAPE && pDev->eom) {
		scsi_writeRspMsg_t *o_msg;
		uint32				source_port = scsiCtrl->scsiDataAddr;
		uint32				dest_port = eDesc->hdr.src_port;

		o_msg = (scsi_writeRspMsg_t *)eee_ramAlloc(sizeof(scsi_writeRspMsg_t));
		memset(o_msg, 0, sizeof(scsi_writeRspMsg_t));

		o_msg->hdr.mid     = SCSI_MID_WRITE_RSP;
		o_msg->hdr.cHandle = wmsg->hdr.cHandle;
		o_msg->eCode       = SCSI_MSG_ECODE_END_OF_MEDIUM;

		if (EEE_GET_DTYPE(dest_port) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(dest_port) << EDESC2_DATA_TYPE_SHIFT;
		}
		scsi_sendMessage((uchar8 *)o_msg, sizeof(scsi_writeRspMsg_t), dest_port, source_port);
		scsi_destroyDescChain(eDesc);
		return(NFX_OK);
	}
	
	/*
	 * Allocate SCSI descriptor
	 */
	sd = (scsi_desc_t *)scsi_allocDesc();
	if (sd == (scsi_desc_t *)NULL) {
		SCSI_DBG(0, ("scsi_procWriteReq: Failed to allocate SCSI desc.\n"));
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NO_RESOURCE);  
		eee_freePacket(eDesc);
		return NFX_OK;
	}
#ifdef SCSI_TIMESTAMP
	sd->cmdTS = t;
#endif

	sd->cHandle = cHandle;
	sd->path    = path;
#if !defined(SCSI_TEST) || SDM_DEV_INIT
	sd->rspDest = eDesc->hdr.src_port;
#else
	sd->rspDest = (EEE_USER_DATA << EDESC2_DATA_TYPE_SHIFT) & EDESC2_DATA_TYPE_MASK;
#endif
	sd->pdev    = pDev;
	sd->eDesc   = eDesc;
	sd->control = SD_OPCODE_WRITE | 
		((wmsg->attributes & SCSI_MSG_ATTR_PHY_OP) ? SD_FLAG_PHY_OP : 0) | 
		((wmsg->attributes & SCSI_MSG_ATTR_NO_REORDER) ? SD_FLAG_NOREORDER : 0) | 
		((wmsg->attributes & SCSI_MSG_ATTR_EXPRESS) ? SD_FLAG_EXPRESS : 0) |
		((pDev->flags.fAllowChain == 0) ? SD_FLAG_NOCOALLESCE : 0);

	if (pDev->flags.fAllowBlock_CS) {
		/* Do we need to pad this write */
		if ((wmsg->blkOffset + wmsg->blkCnt) % (BCSUM_BLOCKSIZE / BLOCKSIZE_512) &&
			(wmsg->attributes & SCSI_MSG_ATTR_PAD_WRITE)) {
			uint32 padding_blocks = 
					(BCSUM_BLOCKSIZE / BLOCKSIZE_512) - 
					(wmsg->blkOffset + wmsg->blkCnt) % (BCSUM_BLOCKSIZE / BLOCKSIZE_512);
			scsi_add_padding(sd->eDesc, padding_blocks);
			wmsg->blkCnt += padding_blocks;
			VERIFY(wmsg->blkCnt <= BCSUM_BLOCKSIZE / BLOCKSIZE_512);
			pDev->writeSub8kPaddedCnt++;
		}
		if(!(sd->control & SD_FLAG_PHY_OP)) {
			sd->control |= SD_FLAG_BLOCK_CS;
			scsi_insert_block_checksum(wmsg->blkOffset, wmsg->blkCnt, total_csum, sd->eDesc);
			scsi_map_logical_to_physical(wmsg->blkOffset, wmsg->blkCnt, &sd->lba, &sd->lbCnt);
		}
		else {
			sd->lba = wmsg->blkOffset;
			sd->lbCnt = wmsg->blkCnt;
		}
	}
	else {
		sd->lba = wmsg->blkOffset;
		sd->lbCnt = wmsg->blkCnt;
	}

	if (pDev->type == SCSI_DEVTYPE_DISK && !(sd->control & SD_FLAG_PHY_OP)) {
		/* 
		 * XXX Partitioning mapping will take place here. For now just
		 * offset. 
		 */
		sd->lba += pDev->offset;
	}
	
	if (pDev->flags.fAllowEmbedded_CS) {
		sd->control |= SD_FLAG_EMBEDDED_CS;
		scsi_insert_embedded_checksum(pDev, sd, total_csum);
	}

	pDev->writeReqCnt++;
	pDev->writeByteCnt += pktlen;

	if( EEE_GET_SLOT_NUM(sd->rspDest) == 0 ) {
		scsiCtrl->stats.writeUserReqCnt_ssc++;
		scsiCtrl->stats.writeUserBytes_ssc += pktlen;
	}
	else {
		scsiCtrl->stats.writeUserReqCnt_fp++;
		scsiCtrl->stats.writeUserBytes_fp += pktlen;
	}

	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = pDev->id;
		scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_WRITE_REQ;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = cHandle;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

	//-- Look for loopback attribute.
	if(wmsg->attributes & SCSI_MSG_ATTR_LOOPBACK) 	{
		/*
		 * Map back to logical address/count.
		 */
		if (pDev->type == SCSI_DEVTYPE_DISK && !(sd->control & SD_FLAG_PHY_OP)) {
			/* 
			 * XXX Partitioning mapping will take place here. For now
			 * just offset.  
			 */
			sd->lba -= pDev->offset;
		}
		if (pDev->flags.fAllowBlock_CS && !(sd->control & SD_FLAG_PHY_OP))
			scsi_map_physical_to_logical(sd->lba, sd->lbCnt, &sd->lba, &sd->lbCnt);

        pDev->writeRspCnt++;
        if( EEE_GET_SLOT_NUM(eDesc->hdr.src_port) == 0 )
            scsiCtrl->stats.writeUserRspCnt_ssc++;
        else
            scsiCtrl->stats.writeUserRspCnt_fp++;

		scsi_sendWriteRsp( sd, SCSI_MSG_ECODE_OK );			
		scsi_deallocDesc(sd);
		return(NFX_OK);
	}

	/* 
	 * Send event to device state machine. 	
	 */
    esmEvent.event = SCSI_DEVO_EVENT_REQ;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pDev->smHandle, &esmEvent);

	/* If this is a tape write, then check if any queues messages can be processed */
	if(pDev->type == SCSI_DEVTYPE_TAPE)
		scsi_proc_tape_queue_ops(pDev);

	return(NFX_OK);
} /* scsi_procWriteReq */

int 
scsi_procAck(eee_desc_t *eDesc, scsi_msgHdr_t *msg)
{
	scsi_ackMsg_t *		ack = (scsi_ackMsg_t *)msg;
	scsi_dev_t *		pDev;
	scsi_mp_desc_t *	smpd;
	scsi_desc_t *		sd;

	pDev = (scsi_dev_t *)scsi_getDevice(msg->path);
	if (pDev == NULL) {
		eee_freePacket(eDesc);
		return NFX_OK;
	}

	SCSI_DBG(1, ("SCSI_MID_ACK: path 0x%x, cHandle 0x%x, nack %d\n", 
				 msg->path, msg->cHandle, ack->nack));
	
	smpd = scsi_MPEntryLookup(pDev, msg->cHandle);
	if (smpd == NULL) {
		SCSI_DBG(0, ("scsi_procAck: Cannot find SMP entry; path 0x%x, cHandle 0x%x.\n",
					 msg->path, msg->cHandle));
		eee_freePacket(eDesc);
		return NFX_OK;
	}

	if (ack->nack) {
		/* NACK - schedule retransmission when ACK timer pops */
		++smpd->nack_received;
		++pDev->nacksReceived;
	}
	else {
		/* ACK -- free scsi desc and smpd context */

		sd = (scsi_desc_t *)smpd->private;

		/* Free scsi desc */
		VERIFY(sd->dataDest == 0);
		sd->eDesc = smpd->firstEDesc;
		sd->control &= ~SD_FLAG_NO_DEALLOC;
		scsi_deallocDesc(sd);

		/* Free smpd context */
		scsi_stop_ack_timer(smpd);
		smpd->private = NULL;
		scsi_MPEntryDestroy(smpd);
	}

	eee_freePacket(eDesc);

	return(NFX_OK);
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_pathErr
 *
 * Description: 	Will cause a path error notification message to be send to 
 *					the DCS for each LUN connected via a particular device
 *					path.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int32
scsi_pathErr(ushort16 llHandle, uint64 devWWN)
{
	int i;
	scsi_dev_t *pdev;

	for(i=0; i < SCSI_MAX_DEV_PER_PORT; i++) {
		if ( (pdev = scsiDevList[i]) ) {
			if( pdev->llHandle == llHandle )
				scsi_sendPathErrMsg(pdev);
		}
	}
	return(NFX_OK);
	scsi_sendPathErrMsg(pdev);
} /* scsi_pathErr */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendPathErrMsg
 *
 * Description: 	Send path error notification message to the SDM deamon on 
 * 					the SSC.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:		NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendPathErrMsg(scsi_dev_t * pdev)
{
	uint32			 	dest;
	sdm_pathErrMsg_t 	*msg;
	uint32				source_port = scsiCtrl->scsiDataAddr;

	msg = (sdm_pathErrMsg_t *)eee_ramAlloc(sizeof(sdm_pathErrMsg_t));
	memset(msg, 0, sizeof(sdm_pathErrMsg_t));

	msg->id       = pdev->id;
	msg->path     = pdev->path;
	msg->hdr.mid  = SDM_MID_PATH_ERR;

	/* TODO */
	dest = 0;    /* destination of the DCS running on the SSC */
	/* TODO */

	if (EEE_GET_DTYPE(dest) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(dest) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8 *)msg, sizeof(sdm_pathErrMsg_t), 
					dest, source_port);
	return( NFX_OK );
} /* scsi_sendPathErrMsg */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendRebalanceMsg
 *
 * Description: 	Send rebalance load message to the DCS process on 
 * 					the SSC.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendRebalanceMsg()
{
	uint32			 	dest;
	sdm_rebalanceMsg_t	*msg;
	uint32				source_port = scsiCtrl->scsiDataAddr;

	msg = (sdm_rebalanceMsg_t *)eee_ramAlloc(sizeof(sdm_rebalanceMsg_t));
	memset(msg, 0, sizeof(sdm_rebalanceMsg_t));

	/* Message Header */
	msg->hdr.mid  = SDM_MID_REBALANCE;

	/* TODO */
	dest = 0;    /* destination of the DCS running on the SSC */
	/* TODO */

	if (EEE_GET_DTYPE(dest) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(dest) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8 *)msg, sizeof(sdm_rebalanceMsg_t), 
					dest, source_port);
	return( NFX_OK );
} /* scsi_sendRebalanceMsg */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendCmdErrRsp
 *
 * Description: 	Sends a reject message back to the request originator and
 *					frees the request packet buffers and descriptors.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 06, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendCmdErrRsp( int32 rspAddr, uint32 cHandle, uint32 eCode )
{
	int32 rc;
	scsi_errRspMsg_t *msg;
	uint32			 source_port = scsiCtrl->scsiDataAddr;

	SCSI_DBG(0, ("scsi_sendCmdErrRsp: rspAddr 0x%x, cHandle 0x%x, eCode 0x%x\n", rspAddr, cHandle, eCode));

	/*
	 * Create reject message and send it.
	 */
	msg = (scsi_errRspMsg_t *)eee_ramAlloc(sizeof(scsi_errRspMsg_t));
	memset(msg, 0, sizeof(scsi_errRspMsg_t));

	msg->eCode 	     = eCode;
	msg->hdr.cHandle = cHandle;
	msg->hdr.mid 	 = SCSI_MID_ERR_RSP;

	if (EEE_GET_DTYPE(rspAddr) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(rspAddr) << EDESC2_DATA_TYPE_SHIFT;
	}
	rc = scsi_sendMessage((uchar8 *)msg, sizeof(scsi_errRspMsg_t), 
						 rspAddr, source_port);
												   
	return NFX_OK;
} /* scsi_sendCmdErrRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendCmdAbortedRsp
 *
 * Description: 	Sends a ABORTED response back to the request originator and
 *					frees the request packet buffers and descriptors.
 * 				
 * Created by:		Radek Aster
 *
 * Data Created:	Sept. 10, 2002
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendCmdAbortedRsp(int32 rspAddr, uint32 mid, uint32 cHandle, uint32 eCode)
{
	int32					rc;
	scsi_writeRspMsg_t *	write_msg;
	scsi_readRspMsg_t *		read_msg;
	scsi_errRspMsg_t *		err_msg;
	uint32					source_port = scsiCtrl->scsiDataAddr;
	void *					msg;
	uint32					msg_len;

	switch (mid) {
	case SCSI_MID_WRITE_RSP:
		write_msg = (scsi_writeRspMsg_t *)eee_ramAlloc(sizeof(scsi_writeRspMsg_t));
		memset(write_msg, 0, sizeof(scsi_writeRspMsg_t));
		write_msg->eCode = eCode;
		write_msg->hdr.cHandle = cHandle;
		write_msg->hdr.mid = mid;
		msg = write_msg;
		msg_len = sizeof(scsi_writeRspMsg_t);
		break;
		
	case SCSI_MID_READ_RSP:
		read_msg = (scsi_readRspMsg_t *)eee_ramAlloc(sizeof(scsi_readRspMsg_t));
		memset(read_msg, 0, sizeof(scsi_readRspMsg_t));
		read_msg->eCode = eCode;
		read_msg->hdr.cHandle = cHandle;
		read_msg->hdr.mid = mid;
		msg = read_msg;
		msg_len = sizeof(scsi_readRspMsg_t);
		break;
		
	default:
		err_msg = (scsi_errRspMsg_t *)eee_ramAlloc(sizeof(scsi_errRspMsg_t));
		memset(err_msg, 0, sizeof(scsi_errRspMsg_t));
		err_msg->eCode = eCode;
		err_msg->hdr.cHandle = cHandle;
		err_msg->hdr.mid = SCSI_MID_ERR_RSP;
		msg = err_msg;
		msg_len = sizeof(scsi_errRspMsg_t);
	}

	if (EEE_GET_DTYPE(rspAddr) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(rspAddr) << EDESC2_DATA_TYPE_SHIFT;
	}
	rc = scsi_sendMessage(msg, msg_len, rspAddr, source_port);
	SCSI_DBG(0, ("scsi_sendCmdAbortedRsp: mid 0x%x, cHandle 0x%x, eCode 0x%x\n", mid, cHandle, eCode));
	return NFX_OK;
} /* scsi_sendCmdAbortedRsp */

int 
scsi_sendAck(int32 rspAddr, uint32 cHandle, int nack)
{
	int32				rc;
	scsi_ackMsg_t *		ack;
	uint32				source_port = scsiCtrl->scsiDataAddr;

	SCSI_DBG(1, ("scsi_sendAck: rspAddr 0x%x, cHandle 0x%x, nack %d\n",
				 rspAddr, cHandle, nack));

	ack = (scsi_ackMsg_t *)eee_ramAlloc(sizeof(scsi_ackMsg_t));
	memset(ack, 0, sizeof(scsi_ackMsg_t));

	ack->hdr.cHandle = cHandle;
	ack->hdr.mid = SCSI_MID_ACK;
	ack->nack = nack;

	if (EEE_GET_DTYPE(rspAddr) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(rspAddr) << EDESC2_DATA_TYPE_SHIFT;
	}
	rc = scsi_sendMessage((uchar8 *)ack, sizeof(scsi_ackMsg_t), rspAddr, source_port);

	return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendOpenRsp
 *
 * Description: 	Send device open response message back to the DCS.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 06, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendOpenRsp( uint32 cHandle, uint32 eCode, uint32 dest )
{
	scsi_openRspMsg_t *msg;
	uint32			  source_port = scsiCtrl->scsiDataAddr;

	msg = (scsi_openRspMsg_t *)eee_ramAlloc(sizeof(scsi_openRspMsg_t));
	memset(msg, 0, sizeof(scsi_openRspMsg_t));

	msg->eCode 	     = eCode;
	msg->hdr.cHandle = cHandle;
	msg->hdr.mid 	 = SCSI_MID_OPEN_RSP;

	SCSI_DBG(1, ("scsi_sendOpenRsp: mid: %x, to addr: %x\n", msg->hdr.mid, dest));
#if !defined(SCSI_TEST) || SDM_DEV_INIT
	if (EEE_GET_DTYPE(dest) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(dest) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8 *)msg, sizeof(scsi_openRspMsg_t), 
					dest, source_port);
#else
	scsi_test_notify((void *)msg);
#endif
	return NFX_OK;
} /* scsi_sendOpenRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendCloseRsp
 *
 * Description: 	Send device close response message back to the DCS.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 06, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendCloseRsp( uint32 cHandle, uint32 eCode, uint32 dest )
{
	scsi_closeRspMsg_t *msg;
	uint32			   source_port = scsiCtrl->scsiDataAddr;

	msg = (scsi_closeRspMsg_t *)eee_ramAlloc(sizeof(scsi_closeRspMsg_t));
	memset(msg, 0, sizeof(scsi_closeRspMsg_t));

	msg->eCode 	     = eCode;
	msg->hdr.cHandle = cHandle;
	msg->hdr.mid 	 = SCSI_MID_CLOSE_RSP;

	if (EEE_GET_DTYPE(dest) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(dest) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8 *)msg, sizeof(scsi_closeRspMsg_t), 
					dest, source_port);
	return NFX_OK;
} /* scsi_sendCloseRsp */

//------------------------------------------------------------------------------ 
// Function Name:   scsi_sendReadRsp
//
// Description:     Send read data response message.
//
// Created by:      Frank Salzmann
//
// Data Created:    May 1, 2001
//
// Returns:         None 
//------------------------------------------------------------------------------ 
int
scsi_sendReadRsp( scsi_desc_t *sd, uint32 eCode, uint32 csum, uint64 codHandle)
{
	char              *buf;
	int                len;
	eee_desc_t        *edesc;
    scsi_readRspMsg_t *msg;
	uint32			  source_port = scsiCtrl->scsiDataAddr;
#ifdef SCSI_TIMESTAMP
	eeeRTC_t		  delta;
	eeeRTC_t		  sum;
#endif

	//-- Allocate Buffer
	buf =  eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_SMALL,0);
	if( !buf ) {
		return(NFX_ERR);  
	}

	//-- Offset into buffer to leave room for mac header
	buf += sizeof(enet_hdr_t) + sizeof(agile_hdr_t);

	//-- Fill in read response message.
    msg = (scsi_readRspMsg_t *)buf;

    msg->hdr.mid     = SCSI_MID_READ_RSP;
    msg->hdr.cHandle = sd->cHandle;
	msg->hdr.path    = 0;
	msg->transSeqId  = 0;
    msg->eCode       = eCode;
	msg->blkOffset   = sd->lba;
	msg->blkCnt      = sd->lbCnt;
	msg->dataLen     = 0;
	msg->queueDepth  = scsi_queue_len(sd->pdev->genio_q) + scsi_queue_len(sd->pdev->hiprio_q);
	msg->checkSum    = csum;
    msg->codHandle   = codHandle;

	//-- Allocate and initialize edesc
	edesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
	if( !edesc ) {
		eee_deallocateBuffer(buf);
		return(NFX_ERR);
	}
	edesc->bd[0].control = 0;		//-- Mark first buffer as unused.
	
	len = sizeof(scsi_readRspMsg_t);
	edesc->bd[1].buf = buf;
	edesc->bd[1].len = len;

	edesc->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;

	if (EEE_GET_DTYPE(sd->rspDest) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(sd->rspDest) << EDESC2_DATA_TYPE_SHIFT;
	}

	edesc->hdr.src_port  = source_port;
	edesc->hdr.dest_port = sd->rspDest;

	edesc->hdr.control = ((len & EDESC_TOTAL_LEN) | 
						  ((1<<EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK)); 
	
#ifdef SCSI_TIMESTAMP
    eee_rtcDuration(&sd->cmdTS, &delta);
    if (eee_rtcAdd(&delta, &scsiCtrl->stats.readUserTotalTime, &sum) == NFX_OK) {
        scsiCtrl->stats.readUserTotalTime = sum;
    }

    eee_rtcDuration(&sd->toFC, &delta);
    if (eee_rtcAdd(&delta, &scsiCtrl->stats.readUserFCTime, &sum) == NFX_OK) {
        scsiCtrl->stats.readUserFCTime = sum;
    }
#endif

	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = sd->pdev->id;
		scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_READ_RSP;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = sd->cHandle;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

#ifndef SCSI_TEST
	if(EEE_GET_SLOT_NUM(edesc->hdr.dest_port) == NFX_SLOTTYPE_SSC &&
		EEE_GET_DEST_CPU(edesc->hdr.dest_port) == NFX_SSC_UNIX) {
		scsiCtrl->sscMsgs.char_space -= SCSI_EEE_CHAR_SPACE(len);
		scsiCtrl->sscMsgs.mbuf_space -= SCSI_EEE_MBUF_SPACE(len);
	}
	eee_forwardPacket(edesc);
#else
    scsi_test_notify((void *)msg);
	eee_freePacket(edesc);
#endif

    return(NFX_OK);
}   /* scsi_sendReadRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendReadRspWithData
 *
 * Description: 	Send read data response message along with the data read..
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	July 30, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_sendReadRspWithData( scsi_desc_t *sd, uint32 eCode, uint32 csum, uint32 transSeqId, uint64* dataOffset )
{
	int32			  rc = NFX_OK;
	uint32			  mlen;
	char              *buf;
	eee_desc_t		  *edesc;
	eee_descBuf_t	  *bd;
	scsi_readRspMsg_t *msg;
	uint32             bufCnt, j, dataLen;
#ifdef SCSI_TIMESTAMP
	eeeRTC_t		  delta;
	eeeRTC_t		  sum;
	static boolean first = 1;
	static eeeRTC_t	  scsiRespBuckets[6];

	if (first) {
		first = 0;
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_1, &(scsiRespBuckets[0]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_2, &(scsiRespBuckets[1]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_3, &(scsiRespBuckets[2]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_4, &(scsiRespBuckets[3]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_5, &(scsiRespBuckets[4]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_6, &(scsiRespBuckets[5]));
	}
#endif

	edesc = sd->eDesc;
	if( !edesc ) {
		return(NFX_OK);
	}

	sd->eDesc = 0;			// Unlink the eee desc from the scsi desc because it 
	                        // will be freed by the E-Layer or the tx function.

#if 0
	if( edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER )
		bd = &edesc->bd[0];
	else
#endif

	//-- FC-DRIVER will never use bd[0]
	bd = &edesc->bd[1];

	// Prepend scsi read rsp msg header to first data buffer.
	buf  = bd->buf;
	mlen = edesc->hdr.control & EDESC_TOTAL_LEN;
	
	buf -= sizeof(scsi_readRspMsg_t);

	msg = (scsi_readRspMsg_t *)buf;
	msg->hdr.mid 	 = SCSI_MID_READ_RSP;
	msg->hdr.cHandle = sd->cHandle;
	msg->transSeqId  = transSeqId;
	msg->eCode 		 = eCode;
	msg->blkOffset   = sd->lba;
	msg->blkCnt      = sd->lbCnt;
	msg->dataLen     = mlen;
	msg->queueDepth  = scsi_queue_len(sd->pdev->genio_q) + scsi_queue_len(sd->pdev->hiprio_q);
	msg->checkSum    = csum;
	msg->codHandle	 = 0;
	if(dataOffset) {
		msg->dataOffset  = *dataOffset;
		*dataOffset = *dataOffset + mlen;
	} 
	else
		msg->dataOffset = 0;

	if( EEE_GET_SLOT_NUM(sd->rspDest) == 1 ) {

		if( EEE_GET_DTYPE(sd->rspDest) == EEE_SAC ) {
			scsiCtrl->stats.readUserRspCnt_fp++;
			scsiCtrl->stats.readUserBytes_fp += mlen;
		}
		else {
			scsiCtrl->stats.readMetaRspCnt_fp++;
			scsiCtrl->stats.readMetaBytes_fp += mlen;
		}
	}
	else {
		scsiCtrl->stats.readMetaRspCnt_ssc++;
		scsiCtrl->stats.readMetaBytes_ssc += mlen;
	}

	SCSI_DBG(1, ("SCSI_MID_READ_RSP: cHandle 0x%x, path 0x%x, transSeqId 0x%x, eCode 0x%x, csum 0x%x\n",
				 msg->hdr.cHandle,
				 msg->hdr.path,
				 msg->transSeqId,
				 msg->eCode,
				 msg->checkSum));

	bd->buf = buf;
	bd->len += sizeof(scsi_readRspMsg_t);
	mlen    += sizeof(scsi_readRspMsg_t);
	
	edesc->hdr.control &= ~EDESC_TOTAL_LEN;
	edesc->hdr.control |= (mlen & EDESC_TOTAL_LEN);

	edesc->hdr.dest_port = sd->rspDest;
	edesc->hdr.src_port  = PACK_UNICAST_PORT_ADDRESS( nfxMySlotId,
											 		  nfxMyCPUId,
											          scsiCtrl->appId,
													  EEE_GET_DTYPE(sd->rspDest),
											          NFX_COS_HIGH,
											          0);
	
#ifdef SCSI_TIMESTAMP
    eee_rtcDuration(&sd->cmdTS, &delta);
    if (eee_rtcAdd(&delta, &scsiCtrl->stats.readMetaTotalTime, &sum) == NFX_OK) {
        scsiCtrl->stats.readMetaTotalTime = sum;
    }

    if (eee_rtcCompare(&delta, &(scsiRespBuckets[0])) < 0) 
        ; // do nothing
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[1])) < 0)
        scsiCtrl->stats.totReadDataRspBuckets.moreThanBucket1++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[2])) < 0)
        scsiCtrl->stats.totReadDataRspBuckets.moreThanBucket2++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[3])) < 0)
        scsiCtrl->stats.totReadDataRspBuckets.moreThanBucket3++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[4])) < 0)
        scsiCtrl->stats.totReadDataRspBuckets.moreThanBucket4++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[5])) < 0)
        scsiCtrl->stats.totReadDataRspBuckets.moreThanBucket5++;
    else
        scsiCtrl->stats.totReadDataRspBuckets.moreThanBucket6++;

    if (eee_rtcCompare(&delta, &scsiCtrl->stats.totReadDataRspBuckets.maxTime) > 0)
        scsiCtrl->stats.totReadDataRspBuckets.maxTime = delta;

    if (eee_rtcAdd(&delta, &scsiCtrl->stats.totReadDataRspBuckets.sumTime, &sum) == NFX_OK) {
		scsiCtrl->stats.totReadDataRspBuckets.sumTime = sum;
	}
    scsiCtrl->stats.totReadDataRspBuckets.numTime++;


    eee_rtcDuration(&sd->toFC, &delta);
    if (eee_rtcAdd(&delta, &scsiCtrl->stats.readMetaFCTime, &sum) == NFX_OK) {
        scsiCtrl->stats.readMetaFCTime = sum;
    }

    if (eee_rtcCompare(&delta, &(scsiRespBuckets[0])) < 0) 
        ; // do nothing
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[1])) < 0)
        scsiCtrl->stats.fcReadDataRspBuckets.moreThanBucket1++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[2])) < 0)
        scsiCtrl->stats.fcReadDataRspBuckets.moreThanBucket2++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[3])) < 0)
        scsiCtrl->stats.fcReadDataRspBuckets.moreThanBucket3++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[4])) < 0)
        scsiCtrl->stats.fcReadDataRspBuckets.moreThanBucket4++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[5])) < 0)
        scsiCtrl->stats.fcReadDataRspBuckets.moreThanBucket5++;
    else
        scsiCtrl->stats.fcReadDataRspBuckets.moreThanBucket6++;

    if (eee_rtcCompare(&delta, &scsiCtrl->stats.fcReadDataRspBuckets.maxTime) > 0)
        scsiCtrl->stats.fcReadDataRspBuckets.maxTime = delta;

    if (eee_rtcAdd(&delta, &scsiCtrl->stats.fcReadDataRspBuckets.sumTime, &sum) == NFX_OK) {
		scsiCtrl->stats.fcReadDataRspBuckets.sumTime = sum;
	}

    scsiCtrl->stats.fcReadDataRspBuckets.numTime++;

#endif

	

	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = sd->pdev->id;
		scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_READ_RSP;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = sd->cHandle;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;

		//-- Sanity check length of read data being returned.
		if ((sd->pdev->type == SCSI_DEVTYPE_DISK) && (transSeqId == 0)) {
			bufCnt = (edesc->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;	
			dataLen = 0;
			for(j=0; j < bufCnt; j++, bd++) {
				dataLen += bd->len;
			}
			dataLen -= sizeof(scsi_readRspMsg_t);
			if( dataLen != (sd->lbCnt*512) ) {
				SCSI_DBG(0, ("Error: Read Response size mismatch in scsi_sendReadRspWithData"));
				SCSI_DBG(0, ("Should Have: %d, Got: %d\n", (sd->lbCnt * 512), dataLen));
				eee_descDisplay(edesc);
				VERIFY(FALSE);
			}
		}
	}

#ifndef SCSI_TEST	
	if(EEE_GET_SLOT_NUM(edesc->hdr.dest_port) == NFX_SLOTTYPE_SSC &&
		EEE_GET_DEST_CPU(edesc->hdr.dest_port) == NFX_SSC_UNIX) {
		scsiCtrl->sscMsgs.char_space -= SCSI_EEE_CHAR_SPACE(mlen);
		scsiCtrl->sscMsgs.mbuf_space -= SCSI_EEE_MBUF_SPACE(mlen);
	}
	rc = eee_forwardPacket(edesc);
	if (rc == EEE_DROP_PACKET ) {
		SCSI_DBG(0, ("scsi_sendReadRspWithData: Error: E-Layer dropped packet\n"));
		eee_freePacket(edesc);
	}
#else
	msg->codHandle = (uint32)edesc;
	scsi_test_notify((void *)msg);
	eee_freePacket(edesc);
#endif

	return(rc);
}	/* scsi_sendReadRspWithData */

//------------------------------------------------------------------------------
// Function Name:	scsi_sendWriteRsp
//
// Description: 	Send write data response message.
//
// Created by:		Frank Salzmann
//
// Data Created:	May 1, 2001
//
// Returns:         None 
//------------------------------------------------------------------------------
int 
scsi_sendWriteRsp( scsi_desc_t *sd, uint32 eCode )
{
	char              *buf;
	int                len;
	eee_desc_t        *edesc;
	scsi_writeRspMsg_t *msg;
	uint32			   source_port = scsiCtrl->scsiDataAddr;
#ifdef SCSI_TIMESTAMP
	eeeRTC_t		  delta;
	eeeRTC_t		  sum;
	static boolean first = 1;
	static eeeRTC_t	  scsiRespBuckets[6];

	if (first) {
		first = 0;
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_1, &(scsiRespBuckets[0]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_2, &(scsiRespBuckets[1]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_3, &(scsiRespBuckets[2]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_4, &(scsiRespBuckets[3]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_5, &(scsiRespBuckets[4]));
		eee_usecToRTC((uint64)SCSI_RSP_BUCKET_6, &(scsiRespBuckets[5]));
	}
#endif

	scsi_debug("scsi_sendWriteRsp", SCSI_DBL_3);

	//-- Allocate Buffer
	buf =  eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_SMALL,0);
	if( !buf ) {
		return(NFX_ERR);  
	}

	//-- Offset into buffer to leave room for mac header
	buf += sizeof(enet_hdr_t) + sizeof(agile_hdr_t);

	//-- Fill in write response message.
    msg = (scsi_writeRspMsg_t *)buf;

	msg->hdr.mid 	 = SCSI_MID_WRITE_RSP;
	msg->hdr.cHandle = sd->cHandle;
	msg->hdr.path    = 0;
	msg->transSeqId  = 0;
	msg->eCode 		 = eCode;
	msg->blkOffset   = sd->lba;
	msg->blkCnt      = sd->lbCnt;
	msg->queueDepth  = scsi_queue_len(sd->pdev->genio_q) + scsi_queue_len(sd->pdev->hiprio_q);
	
	SCSI_DBG(1, ("scsi_sendWriteRsp: cHandle 0x%x, lba 0x%x, lbCnt 0x%x, eCode 0x%x\n",
				 sd->cHandle, sd->lba, sd->lbCnt, eCode));

	//-- Allocate and initialize edesc
	edesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
	if( !edesc ) {
		eee_deallocateBuffer(buf);
		return(NFX_ERR);
	}
	edesc->bd[0].control = 0;		//-- Mark first buffer as unused.
	
	len = sizeof(scsi_writeRspMsg_t);
	edesc->bd[1].buf = buf;
	edesc->bd[1].len = len;
	edesc->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;

	if (EEE_GET_DTYPE(sd->rspDest) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(sd->rspDest) << EDESC2_DATA_TYPE_SHIFT;
	}

	edesc->hdr.src_port  = source_port;
	edesc->hdr.dest_port = sd->rspDest;

	edesc->hdr.control = ((len & EDESC_TOTAL_LEN) | 
						  ((1<<EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK)); 


#ifdef SCSI_TIMESTAMP
    eee_rtcDuration(&sd->cmdTS, &delta);
    if (eee_rtcAdd(&delta, &scsiCtrl->stats.writeUserTotalTime, &sum) == NFX_OK) {
        scsiCtrl->stats.writeUserTotalTime = sum;
    }

    if (eee_rtcCompare(&delta, &(scsiRespBuckets[0])) < 0)
        ; // do nothing
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[1])) < 0)
        scsiCtrl->stats.totWriteRspBuckets.moreThanBucket1++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[2])) < 0)
        scsiCtrl->stats.totWriteRspBuckets.moreThanBucket2++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[3])) < 0)
        scsiCtrl->stats.totWriteRspBuckets.moreThanBucket3++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[4])) < 0)
        scsiCtrl->stats.totWriteRspBuckets.moreThanBucket4++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[5])) < 0)
        scsiCtrl->stats.totWriteRspBuckets.moreThanBucket5++;
    else
        scsiCtrl->stats.totWriteRspBuckets.moreThanBucket6++;

    if (eee_rtcCompare(&delta, &scsiCtrl->stats.totWriteRspBuckets.maxTime) > 0)
        scsiCtrl->stats.totWriteRspBuckets.maxTime = delta;

    if (eee_rtcAdd(&delta, &scsiCtrl->stats.totWriteRspBuckets.sumTime, &sum) == NFX_OK) {
		scsiCtrl->stats.totWriteRspBuckets.sumTime = sum;
	}
    scsiCtrl->stats.totWriteRspBuckets.numTime++;

    eee_rtcDuration(&sd->toFC, &delta);
    if (eee_rtcAdd(&delta, &scsiCtrl->stats.writeUserFCTime, &sum) == NFX_OK) {
        scsiCtrl->stats.writeUserFCTime = sum;
    }

    if (eee_rtcCompare(&delta, &(scsiRespBuckets[0])) < 0) 
        ; // do nothing
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[1])) < 0)
        scsiCtrl->stats.fcWriteRspBuckets.moreThanBucket1++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[2])) < 0)
        scsiCtrl->stats.fcWriteRspBuckets.moreThanBucket2++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[3])) < 0)
        scsiCtrl->stats.fcWriteRspBuckets.moreThanBucket3++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[4])) < 0)
        scsiCtrl->stats.fcWriteRspBuckets.moreThanBucket4++;
    else if (eee_rtcCompare(&delta, &(scsiRespBuckets[5])) < 0)
        scsiCtrl->stats.fcWriteRspBuckets.moreThanBucket5++;
    else
        scsiCtrl->stats.fcWriteRspBuckets.moreThanBucket6++;

    if (eee_rtcCompare(&delta, &scsiCtrl->stats.fcWriteRspBuckets.maxTime) > 0)
        scsiCtrl->stats.fcWriteRspBuckets.maxTime = delta;

    if (eee_rtcAdd(&delta, &scsiCtrl->stats.fcWriteRspBuckets.sumTime, &sum) == NFX_OK) {
		scsiCtrl->stats.fcWriteRspBuckets.sumTime = sum;
	}
    scsiCtrl->stats.fcWriteRspBuckets.numTime++;

#endif

	//-- SCSI_HISTORY
	if( scsi_history_enable ) {
		scsiHistory[scsiHistoryCurEntry].id      = sd->pdev->id;
		scsiHistory[scsiHistoryCurEntry].op      = SCSI_MID_WRITE_RSP;
		scsiHistory[scsiHistoryCurEntry].lba     = sd->lba;
		scsiHistory[scsiHistoryCurEntry].lbCnt   = sd->lbCnt;
		scsiHistory[scsiHistoryCurEntry].cHandle = sd->cHandle;
		if( ++scsiHistoryCurEntry >= SCSI_MAX_HISTORY )
			scsiHistoryCurEntry = 0;
	}

#ifndef SCSI_TEST
	if(EEE_GET_SLOT_NUM(edesc->hdr.dest_port) == NFX_SLOTTYPE_SSC &&
		EEE_GET_DEST_CPU(edesc->hdr.dest_port) == NFX_SSC_UNIX) {
		scsiCtrl->sscMsgs.char_space -= SCSI_EEE_CHAR_SPACE(len);
		scsiCtrl->sscMsgs.mbuf_space -= SCSI_EEE_MBUF_SPACE(len);
	}
	eee_forwardPacket(edesc);
#else
    scsi_test_notify((void *)msg);
	eee_freePacket(edesc);
#endif

	return(NFX_OK);
}	// scsi_sendWriteRsp 

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendRelayRsp
 *
 * Description: 	Send device command response message.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 1, 2001
 *
 * Returns:         None 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_sendRelayRsp( scsi_desc_t *sd )
{
	uint32				dataLen;
	uint32			 	msize;
	uchar8			 	*buf;
	scsi_relayRspMsg_t 	*msg;
	eee_desc_t			*edesc;
	uint32				source_port = scsiCtrl->scsiDataAddr;
	int32				rc;
	uint32				seqId, seqMax, transId = 0;

	/* Get data len if present */
	if( (sd->status & SD_DEV_DATA_VALID) && (sd->eDesc) ) {
		edesc = sd->eDesc;
		if (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {				
			transId = scsi_getTransactionId();
			for (seqMax = 1; edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER; ++seqMax)
				edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;
			seqId = 1;
		}

		edesc = sd->eDesc;
		while (edesc) {
			dataLen = edesc->hdr.control & EDESC_TOTAL_LEN;
			msize = sizeof(scsi_relayRspMsg_t) + dataLen;
			buf = (uchar8 *)eee_ramAlloc(msize);
			if (!buf) {
				scsi_err(SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_1, "[scsi_sendRelayRsp]");
				return(NFX_ERR);
			}
			memset(buf, 0, sizeof(scsi_relayRspMsg_t));

			msg = (scsi_relayRspMsg_t *)buf;
			msg->transSeqId  = transId ? SCSI_MSG_TRANSSEQID_PACK(transId, seqMax, seqId) : 0;
			msg->status 	 = sd->status;
			msg->hdr.mid     = SCSI_MID_RELAY_RSP;
			msg->hdr.cHandle = sd->cHandle;
			msg->dataLen     = dataLen;

			/* Copy the CDB */
			memcpy(msg->cdb, sd->cdb, 16);

			/* Copy the Sense Data if available */
			if ((sd->status & SD_DEV_SENSE_VALID) && (sd->senseData))
				memcpy(msg->senseData, sd->senseData, 32);

			if (dataLen) {
				rc = eee_copyPacketToLocalBuffer(msg->data, edesc, dataLen, 0); 
				VERIFY(rc == dataLen);
			}

			if (EEE_GET_DTYPE(sd->rspDest) != EEE_USER_DATA) {
				source_port &= ~EDESC2_DATA_TYPE_MASK;
				source_port |= EEE_GET_DTYPE(sd->rspDest) << EDESC2_DATA_TYPE_SHIFT;
			}

			SCSI_DBG(1, ("SCSI_MID_RELAY_RSP: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
						 "cdb[0] 0x%x, dataLen %d\n",
						 msg->hdr.cHandle,
						 msg->hdr.path,
						 msg->transSeqId,
						 msg->cdb[0],
						 msg->dataLen));

			scsi_sendMessage(buf, msize, sd->rspDest, source_port);
			buf = NULL;

			if (edesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
				VERIFY(!(edesc->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE));
				edesc = (eee_desc_t *)edesc->bd[EEE_MAX_BUFS-1].buf;
			}
			else
				edesc = NULL;

			++seqId;

		} /* while */

	} /* if ( (sd->status & SD_DEV_DATA_VALID) && (sd->eDesc) ) */
	else {
		msize = sizeof(scsi_relayRspMsg_t);
		buf = (uchar8 *)eee_ramAlloc(msize);
		if (!buf) {
			scsi_err(SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_1, "[scsi_sendRelayRsp]");
			return(NFX_ERR);
		}
		memset(buf, 0, msize);

		msg = (scsi_relayRspMsg_t *)buf;
		msg->status 	 = sd->status;
		msg->hdr.mid     = SCSI_MID_RELAY_RSP;
		msg->hdr.cHandle = sd->cHandle;

		/* Copy the CDB */
		memcpy(msg->cdb, sd->cdb, 16);

		/* Copy the Sense Data if available */
		if ((sd->status & SD_DEV_SENSE_VALID) && (sd->senseData))
			memcpy(msg->senseData, sd->senseData, 32);

		if (EEE_GET_DTYPE(sd->rspDest) != EEE_USER_DATA) {
			source_port &= ~EDESC2_DATA_TYPE_MASK;
			source_port |= EEE_GET_DTYPE(sd->rspDest) << EDESC2_DATA_TYPE_SHIFT;
		}

		SCSI_DBG(1, ("SCSI_MID_RELAY_RSP: cHandle 0x%x, path 0x%x, transSeqId 0x%x, "
					 "cdb[0] 0x%x, dataLen %d\n",
					 msg->hdr.cHandle,
					 msg->hdr.path,
					 msg->transSeqId,
					 msg->cdb[0],
					 msg->dataLen));

		scsi_sendMessage(buf, msize, sd->rspDest, source_port);

#if defined(SCSI_TEST)
		scsi_test_notify((void *)buf);
#endif

	} /* if ! ( (sd->status & SD_DEV_DATA_VALID) && (sd->eDesc) ) */
	return(NFX_OK);
}	/* scsi_sendRelayRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendAddDevMsg
 *
 * Description: 	Sends new device & path information to the device cache 
 *					server on the SSC.	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_sendAddDevMsg(scsi_dev_t *pdev)
{
	uint32			 rc;
	uint32			 dest;
	uint32 			 sdm_deamon;
	sdm_addDevMsg_t *msg;
	uint32			 source_port = scsiCtrl->scsiDataAddr;
	char			 sdm_name[10];
	
	SCSI_DBG(1, ("scsi_sendAddDevMsg: Enter\n"));

	msg = (sdm_addDevMsg_t *)eee_ramAlloc(sizeof(sdm_addDevMsg_t));
#ifdef EVAL_BOARD
	msg = (sdm_addDevMsg_t *)((uint32)msg | 0x20000000);
#endif
	memset((uchar8 *)msg, 0, sizeof(sdm_addDevMsg_t));
	msg->hdr.mid  	 = SDM_MID_DEV_ADD;
	msg->hdr.chandle = 0;
	msg->path	     = pdev->path;
	msg->id	 	     = pdev->id;
	msg->wwn         = pdev->wwn;
	msg->chn_wwn     = pdev->chnWWN;
	msg->type        = pdev->type;
	msg->lbCnt       = pdev->blocks;
	msg->lbSize      = pdev->blockSize;
	msg->llHandle    = pdev->llHandle;
	msg->raidLevel   = pdev->raidLevel;

	if( pdev->flags.fOpenMode )
		msg->state	 = SCSI_DEV_STATE_OPEN;	 
	else
	if( pdev->flags.fCloseMode )
		msg->state	 = SCSI_DEV_STATE_CLOSED;	 
	else
		msg->state	 = SCSI_DEV_STATE_INIT;	 

	strcpy(msg->st.vendor, pdev->vendor);
	strcpy(msg->st.model, pdev->model);
	strcpy(msg->st.productId, pdev->productId);
	strncpy(msg->st.sn, pdev->sn, 32);
	strncpy(msg->st.name, pdev->name, SCSI_DEV_NAME_LEN);

	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	sdm_deamon = eee_getApplicationId(sdm_name);
	dest = PACK_UNICAST_PORT_ADDRESS(0,
									 0,
							 		 sdm_deamon,
							         EEE_MGMT,
							         NFX_COS_HIGH,
									 0);
	source_port &= ~EDESC2_DATA_TYPE_MASK;
	source_port |= EEE_MGMT << EDESC2_DATA_TYPE_SHIFT;
	rc = scsi_sendMessage((uchar8 *)msg, sizeof(sdm_addDevMsg_t), 
						 dest, source_port);
	if (rc != NFX_OK) {
		SCSI_DBG(0, ("scsi_sendAddDevMsg: scsi_sendMessage call failed: rc: %x\n", rc));
	}

#if defined(SCSI_TEST) && !SDM_DEV_INIT
	if (pdev->type == SCSI_DEVTYPE_DISK)
		scsi_test_new_device_notify(pdev->id, pdev->path, pdev->type, pdev->blocks, BLOCKSIZE_512);
	else
		scsi_test_new_device_notify(pdev->id, pdev->path, pdev->type, 0, 0);
#endif

	SCSI_DBG(1, ("scsi_sendAddDevMsg: Leave\n"));

	return(NFX_OK);
}	/* scsi_sendAddDevMsg */
/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendRemDevMsg
 *
 * Description: 	Sends message to the SDM to remove dev/path information
 * 					from the global device database on the SSC.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	July 26, 2001
 *
 * Returns:         NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_sendRemDevMsg(scsi_dev_t *pdev)
{
	uint32			 rc;
	uint32			 dest;
	uint32 			 sdm_deamon;
	sdm_remDevMsg_t *msg;
	uint32			 source_port = scsiCtrl->scsiDataAddr;
	char			 sdm_name[10];
	
	SCSI_DBG(1, ("scsi_sendRemDevMsg: Enter\n"));

	msg = (sdm_remDevMsg_t *)eee_ramAlloc(sizeof(sdm_remDevMsg_t));
#ifdef EVAL_BOARD
	msg = (sdm_remDevMsg_t *)((uint32)msg | 0x20000000);
#endif
	memset((uchar8 *)msg, 0, sizeof(sdm_remDevMsg_t));

	msg->hdr.mid  	 = SDM_MID_DEV_REMOVE;
	msg->hdr.chandle = 0;
	msg->path	     = pdev->path;
	msg->id	 	     = pdev->id;

	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	sdm_deamon = eee_getApplicationId(sdm_name);
	dest = PACK_UNICAST_PORT_ADDRESS(0,
									 0,
							         sdm_deamon,
							         EEE_MGMT,
							         NFX_COS_HIGH,
							         0);
	source_port &= ~EDESC2_DATA_TYPE_MASK;
	source_port |= EEE_MGMT << EDESC2_DATA_TYPE_SHIFT;
	rc = scsi_sendMessage((uchar8 *)msg, sizeof(sdm_remDevMsg_t), 
						 dest, source_port);
	if (rc != NFX_OK) {
		SCSI_DBG(0, ("scsi_sendRemDevMsg: scsi_sendMessage failed: rc %x\n", rc));
	}

#ifdef SCSI_TEST
	if (pdev->type == SCSI_DEVTYPE_DISK)
	    scsi_test_remove_device_notify(pdev->id, pdev->path);
#endif

	SCSI_DBG(1, ("scsi_sendRemDevMsg: Leave\n"));

	return(NFX_OK);
}	/* scsi_sendRemDevMsg */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendReadData
 *
 * Description:		Sends data read from a storage device onto its ultimate
 *					destination within the chassis. 
 *
 * Returns: NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendReadData( scsi_dev_t *pDev, scsi_desc_t *sd, uint64 cod )
{
	int32		rc;
	eee_desc_t	*edesc;

	edesc = sd->eDesc;
	scsi_addGmcHdr(edesc, cod);
	edesc->hdr.dest_port = sd->dataDest;
	edesc->hdr.src_port  = PACK_UNICAST_PORT_ADDRESS( nfxMySlotId,
											          nfxMyCPUId,
											          scsiCtrl->appId,
											          EEE_USER_DATA,
											          NFX_COS_HIGH,
											          0);

	if(EEE_GET_SLOT_NUM(edesc->hdr.dest_port) == NFX_SLOTTYPE_SSC &&
		EEE_GET_DEST_CPU(edesc->hdr.dest_port) == NFX_SSC_UNIX) {
		scsiCtrl->sscMsgs.char_space -= SCSI_EEE_CHAR_SPACE(edesc->hdr.control & EDESC_TOTAL_LEN);
		scsiCtrl->sscMsgs.mbuf_space -= SCSI_EEE_MBUF_SPACE(edesc->hdr.control & EDESC_TOTAL_LEN);
	}
	rc = eee_forwardPacket(edesc);
	if( rc == NFX_OK || rc == EEE_DROP_PACKET)
		sd->eDesc = 0;

	return(rc);
} /* scsi_sendReadData */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_addGmcHdr
 *
 * Description: 	Prepends GMC application header to user-data buffer.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 1, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_addGmcHdr(eee_desc_t *edesc, nfx_cod_t cod)
{
	int32			pktLen;
	int32			offset;
	int32			appId;
	gmc_app_hdr_t 	*gmcMsg;
	eee_descBuf_t	*bd;

	bd = &edesc->bd[1];
	
	// Backup buffer pointer to insert the gmc_app_hdr.
	offset = sizeof(gmc_app_hdr_t);
	bd->buf -= offset;
	bd->len += offset;

	pktLen = edesc->hdr.control & EDESC_TOTAL_LEN;

	appId = eee_getApplicationId("evm");
	gmcMsg = (gmc_app_hdr_t *)bd->buf;
#ifndef SBM_REV_2
    gmcMsg->control = scsi_swap32(appId); // This is who the MCPU will 
                                          // send notify msg to.
    gmcMsg->id.cod  = scsi_swap64(cod);
    gmcMsg->cod_len  = pktLen;
#endif

	pktLen += offset;
	edesc->hdr.control |= (pktLen & EDESC_TOTAL_LEN); 

	return(NFX_OK);
} /* scsi_addGmcHdr */

/*
 * scsi_getTransactionId - 
 *
 * The transaction ID should be unique across all SCSI command
 * originators. For now, since the only geneator of "large" I/Os will
 * be the SDM, we can limit the generation of the transaction ID
 * appropriately.  
 */
ushort16
scsi_getTransactionId(void)
{
	static ushort16 transactionId = 1;

	if (++transactionId == 0)
		transactionId = 1;
	return(transactionId);
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendSdmDelAllMsg
 *
 * Description: 	Send msg to sdm telling it to remove all device
 * 					path entries associated with this slot/port.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	January 30, 2002
 *					
 * Returns: 		SDM_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendSdmDelAllMsg() 
{
	uint32			    dest;
	uint32 			    sdm_deamon;
	sdm_delAllReqMsg_t *msg;
	uint32			    source_port = scsiCtrl->scsiDataAddr;
	char				sdm_name[10];
	
	SCSI_DBG(1, ("scsi_sendSdmDelAllMsg: Enter\n"));

	msg = (sdm_delAllReqMsg_t *)eee_ramAlloc(sizeof(sdm_delAllReqMsg_t));
	memset((uchar8 *)msg, 0, sizeof(sdm_delAllReqMsg_t));

	msg->hdr.mid  	 = SDM_MID_DEL_ALL_REQ;
	msg->hdr.chandle = 0;
	msg->slot	     = nfxMySlotId;
	msg->port	 	 = nfxMyCPUId;

	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	sdm_deamon = eee_getApplicationId(sdm_name);
	dest = PACK_UNICAST_PORT_ADDRESS(0,
									 0,
							         sdm_deamon,
							         EEE_MGMT,
							         NFX_COS_HIGH,
							         0);
	source_port &= ~EDESC2_DATA_TYPE_MASK;
	source_port |= EEE_MGMT << EDESC2_DATA_TYPE_SHIFT;
	scsi_sendMessage((uchar8 *)msg, sizeof(sdm_delAllReqMsg_t), 
						 dest, source_port);

	SCSI_DBG(1, ("scsi_sendSdmDelAllMsg: Leave\n"));

	return SDM_OK;
} // scsi_sendSdmDelAllMsg

//=======================================
// Temporary Functions For Byte Swapping.
//=======================================
typedef union _scsi_32 {
    uint32 w;
    char   b[4];
} scsi_32_t;

uint32
scsi_swap32(uint32 val)
{
#if 0
    int i, j;
    scsi_32_t in;
    scsi_32_t out;

#if NFX_SOCKET_BACKPLANE
    in.w = val;
    for(i=0, j=3; i < 4; i++, j--)
        out.b[i] = in.b[j];
    //printf("before: %llx, after: %llx\n", val, out.w);
    return(out.w);

#else
    return(val);
#endif
#endif 

    return(val);
} // scsi_swap32

typedef union _scsi_64 {
    uint64 w;
    char   b[8];
} scsi_64_t;

uint64
scsi_swap64(uint64 val)
{
#if 0
    int i, j;
    scsi_64_t in;
    scsi_64_t out;

#if NFX_SOCKET_BACKPLANE
    in.w = val;
    for(i=0, j=7; i < 8; i++, j--)
        out.b[i] = in.b[j];
    //printf("before: %llx, after: %llx\n", val, out.w);
    return(out.w);
#else
    return(val);
#endif
#endif
    return(val);
} // scsi_swap64

int
scsi_sendToSscApp( uchar8 *msg, uint32 msgSize, char *app)
{
	uint32 rc;
	uint32 dest;
	uint32 appId;
	uint32 source_port = scsiCtrl->scsiDataAddr;

	SCSI_DBG(1, ("scsi_sendToSscApp: Enter\n"));	

	appId = eee_getApplicationId(app);
	dest = PACK_UNICAST_PORT_ADDRESS(0,
									 0,
							 		 appId,
							         EEE_MGMT,
							         NFX_COS_HIGH,
									 0);

	source_port &= ~EDESC2_DATA_TYPE_MASK;
	source_port |= EEE_MGMT << EDESC2_DATA_TYPE_SHIFT;
	rc = scsi_sendMessage(msg, msgSize, dest, source_port);
	if (rc != NFX_OK) {
		SCSI_DBG(0, ("scsi_sendToSscApp: scsi_sendMessage call failed: rc: %x\n", rc));
	}

	SCSI_DBG(1, ("scsi_sendToSscApp: Leave\n"));	

	return NFX_OK;
} // scsi_sendToSscApp

int
scsi_procDelMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg)
{
	scsi_dev_t *pdev;
    esm_event_t esmEvent;
	
	pdev = scsi_getDevice(msg->path);
	if( pdev ) {
		esmEvent.event  = SCSI_DEVO_EVENT_DELETE;
		esmEvent.p1.u.w = msg->cHandle;
		esmEvent.p2.u.w = edesc->hdr.src_port;
		esm_deliverEvent(pdev->smHandle, &esmEvent);
	}
	else {
		scsi_sendCmdErrRsp(edesc->hdr.src_port, 
						   msg->cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
	}
	eee_freePacket(edesc);
	return(NFX_OK);
} // scsi_procDelMsg

int
scsi_procAddMsg(eee_desc_t *edesc, scsi_addReqMsg_t *i_msg)
{
	uint32            rc;
	scsi_dev_t       *pdev;
	scsi_addRspMsg_t *o_msg;
	char			 sdm_name[10];

	//printf("scsi_procAddMsg: \n");

	//-- Prepare response message to SDM.
	o_msg = (scsi_addRspMsg_t *)eee_ramAlloc(sizeof(scsi_addRspMsg_t));
	memset(o_msg, 0, sizeof(scsi_addRspMsg_t));
	o_msg->hdr.mid     = SCSI_MID_ADD_RSP;
	o_msg->hdr.cHandle = i_msg->hdr.cHandle;

	//printf("procAdd Lun: 0x%016llx:%d, llHandle: %d \n",
	//   i_msg->wwn, i_msg->lun, i_msg->llHandle);

	//-- See if the device already exists.
	if ((pdev = scsi_devCheckExist(i_msg->wwn, i_msg->lun, i_msg->llHandle))) {
		o_msg->eCode    = SCSI_MSG_ECODE_OK; 
	}
	else {	//-- Allocate a new device control block
		pdev = (scsi_dev_t *)scsi_allocDevice(i_msg->lun, i_msg->llHandle);
	}

	if( !pdev ) {
		o_msg->eCode       = SCSI_MSG_ECODE_NO_RESOURCE; 
		goto scsi_procAddMsg_sendRsp;
	}

	//-- Fill in device information, Anything not set has been set equal
	//-- to zero by the allocation function.
	pdev->wwn       = i_msg->wwn;
	pdev->lun       = i_msg->lun;
	pdev->id        = i_msg->id;
	pdev->chnWWN    = i_msg->chn_wwn;
	pdev->llHandle  = i_msg->llHandle;
	pdev->type      = i_msg->devType;
	pdev->blocks    = i_msg->blockCnt;
	pdev->blockSize = i_msg->blockSize;
	pdev->raidLevel = i_msg->raidLevel;
	strncpy(pdev->vendor, i_msg->vendor, 9);
	strncpy(pdev->model, i_msg->model, 16);	
	strncpy(pdev->productId, i_msg->productId, 16);	
	strncpy(pdev->sn, i_msg->sn, 32);
	strncpy(pdev->name, i_msg->name, SCSI_DEV_NAME_LEN);

	SCSI_DBG(1, ("scsi_procAddMsg: 0x%llx/%d, blocks %d, blockSize %d\n", 
				 pdev->wwn, pdev->lun, pdev->blocks, pdev->blockSize));

	if (pdev->type == SCSI_DEVTYPE_DISK) {
		pdev->flags.fAllowEmbedded_CS = (pdev->blockSize == BLOCKSIZE_520) ? 1 : 0;
		if (pdev->flags.fAllowEmbedded_CS)
			pdev->flags.fAllowBlock_CS = 0;
		else
			pdev->flags.fAllowBlock_CS = (i_msg->map.flags & SCSI_MAPPING_FLAGS_BCS_ENABLED) ? 1 : 0;
	}
	pdev->offset = i_msg->map.offset;
	pdev->partitionCount = i_msg->map.partitionCount;
	memcpy(&pdev->partitions, &i_msg->map.partitions, sizeof(i_msg->map.partitions));

	//-- Create device statemachine instance
	rc = scsi_openDevice(pdev->path);
	if( rc != SCSI_MSG_ECODE_OK ) {
		//-- State machine instance not created so..... 
		scsi_deallocDevice(pdev);        /* free dev ctrl block memory */
		scsi_err( SCSI_ERR_SM_CREATE | SCSI_ERR_LEVEL_0,   /* complain */
				  "func: scsi_procAddMsg");
		o_msg->eCode = SCSI_MSG_ECODE_NO_RESOURCE; 
	}
	else {
		o_msg->eCode    = SCSI_MSG_ECODE_OK; 
		o_msg->hdr.path = pdev->path;
		
		//-- Take device out of initMode.
		if(i_msg->sdmState == SDM_DEV_STATE_INIT)
			pdev->flags.fInitMode = 1;
		else 
			pdev->flags.fInitMode = 0;
	}

 scsi_procAddMsg_sendRsp:
	//-- Send response message back up to the SDM.
	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	scsi_sendToSscApp((char *)o_msg, sizeof(scsi_addRspMsg_t), sdm_name);

#if defined(SCSI_TEST) && SDM_DEV_INIT
	if (pdev->type == SCSI_DEVTYPE_DISK) {
		uint32 blocks = pdev->blocks - pdev->offset;
		if (pdev->flags.fAllowBlock_CS)
			blocks = blocks / (BCSUM_BLOCK_SPAN + 1) * BCSUM_BLOCK_SPAN;
		scsi_test_new_device_notify(pdev->id, pdev->path, pdev->type, blocks, BLOCKSIZE_512);
	}
	else
		scsi_test_new_device_notify(pdev->id, pdev->path, pdev->type, 0, 0);
#endif

    eee_freePacket(edesc);
	return NFX_OK;
} // scsi_procAddMsg

int 
scsi_procDevControlMsg(eee_desc_t *eDesc, scsi_msgHdr_t *msg)
{
	uint32						cHandle;
	scsi_path_t					path;
	scsi_dev_t *				pDev;
	scsi_controlReqMsg_t *		rmsg = (scsi_controlReqMsg_t *)msg;
	scsi_controlRspMsg_t		*rsp;
	uint32						resp_port = eDesc->hdr.src_port;
	uint32						source_port = scsiCtrl->scsiDataAddr;

	/*
	 * Get pointer to SCSI Device Context Block
	 */
	path = msg->path;	
	pDev = (scsi_dev_t *)scsi_getDevice(path);
	if (pDev == NULL) {
		/* Invalid device id, reject request */
		scsi_sendCmdErrRsp(eDesc->hdr.src_port, 
						   cHandle, SCSI_MSG_ECODE_NOSUCH_DEV);
		eee_freePacket(eDesc);
		return NFX_OK;
	}

	if (rmsg->control_enable_mask & SCSI_DEV_CONTROL_ENABLE_MASK_SORTING)
		pDev->flags.fAllowSort = (rmsg->control_mask & SCSI_DEV_CONTROL_ENABLE_SORTING) ? 1 : 0;

	if (rmsg->control_enable_mask & SCSI_DEV_CONTROL_ENABLE_MASK_COALLESCING)
		pDev->flags.fAllowChain = (rmsg->control_mask & SCSI_DEV_CONTROL_ENABLE_COALLESCING) ? 1 : 0;

	if (rmsg->control_enable_mask & SCSI_DEV_CONTROL_ENABLE_MASK_CS_VERIFY)
		pDev->flags.fLocalCSVerify = (rmsg->control_mask & SCSI_DEV_CONTROL_ENABLE_CS_VERIFY) ? 1 : 0;

	rsp = (scsi_controlRspMsg_t *)eee_ramAlloc(sizeof(scsi_controlRspMsg_t));
	memset(rsp, 0, sizeof(scsi_controlRspMsg_t));

	rsp->hdr.cHandle = rmsg->hdr.cHandle;
	rsp->hdr.mid = SCSI_MID_DEV_CONTROL_RSP;
	rsp->status = SCSI_MSG_ECODE_OK;

	if (EEE_GET_DTYPE(resp_port) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(resp_port) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8*)rsp, sizeof(scsi_controlRspMsg_t), resp_port, source_port);

	return(NFX_OK);
} /* scsi_procDevControlMsg */

int32
scsi_dumpSD( scsi_desc_t *sd ) 
{
    eeeGMT_t duration;

    if( !sd )
		return NFX_OK;

	printf("Desc Addr       :  %p\n", sd);
	printf("control         :  %x\n", sd->control);
	printf("pdev            :  %p\n", sd->pdev);
	printf("path            :  %x\n", sd->path);
	printf("halHandle       :  %x\n", sd->halHandle);
	printf("cHandle         :  %x\n", sd->cHandle);
	printf("status          :  %x\n", sd->status);
	printf("senseData       :  %p\n", sd->senseData);
	printf("expectDataLen   :  %x\n", sd->expectDataLen);
	printf("lba             :  %x\n", sd->lba);
	printf("lbCnt           :  %x\n", sd->lbCnt);
	printf("lbCntChained    :  %x\n", sd->lbCntChained);
	printf("dataDest        :  %x\n", sd->dataDest);
	printf("rspDest         :  %x\n", sd->rspDest);
	printf("tmrHandle       :  %x\n", sd->tmrHandle);
	printf("eDesc           :  %p\n", sd->eDesc);
	printf("chain           :  %p\n", sd->chain);
	printf("next            :  %p\n", sd->next);
	printf("prev            :  %p\n", sd->prev);
    printf("queued_timestamp:  %x\n",
           (uint32)eee_rtcToUsec(&sd->queued_timestamp) / USEC_PER_TICK);
    eee_rtcToTV(&sd->cmdTS, &duration);
    printf("cmdTS_sec       :  %x\n", duration.sec);
	printf("cmdTS_usec      :  %x\n", duration.usec);
    eee_rtcToTV(&sd->toFC, &duration);
	printf("toFC_sec        :  %x\n", duration.sec);
	printf("toFC_usec       :  %x\n", duration.usec);
    eee_rtcToTV(&sd->fromFC, &duration);
	printf("fromFC_sec      :  %x\n", duration.sec);
	printf("fromFC_usec     :  %x\n", duration.usec);
	printf("retryCnt        :  %x\n", sd->retryCnt);
	printf("timeOutCnt      :  %x\n", sd->timeOutCnt);
	printf("maxTimeToLive   :  %x\n", sd->maxTimeToLive);
	printf("state           :  %x\n", sd->state);
	
	return NFX_OK;
} // scsi_dumpSD


void
scsi_getstats (
    hal_fc_link_stats_t   *stats,
    uint32                isClear)
{
    scsi_stats_t *scsi_stats;

    scsi_stats = &scsiCtrl->stats;

    stats->userReadFp = scsi_stats->readUserReqCnt_fp;
    stats->userReadSsc = scsi_stats->readUserReqCnt_ssc;
    stats->metaReadFp = scsi_stats->readMetaReqCnt_fp;
    stats->metaReadSsc = scsi_stats->readMetaReqCnt_ssc;
    stats->writeFp = scsi_stats->writeUserReqCnt_fp;
    stats->writeSsc = scsi_stats->writeUserReqCnt_ssc;
    
    stats->totalTimeUserRead  = scsi_stats->readUserTotalTime;
    stats->totalTimeMetaRead  = scsi_stats->readMetaTotalTime;
    stats->totalTimeUserWrite = scsi_stats->writeUserTotalTime;
    stats->totalTimeMetaWrite = scsi_stats->writeMetaTotalTime;
    stats->fcTimeUserRead     = scsi_stats->readUserFCTime;
    stats->fcTimeMetaRead     = scsi_stats->readMetaFCTime;
    stats->fcTimeUserWrite    = scsi_stats->writeUserFCTime;
    stats->fcTimeMetaWrite    = scsi_stats->writeMetaFCTime;
    
    stats->readUserBytes_ssc    = scsi_stats->readUserBytes_ssc ;
    stats->readUserBytes_fp     = scsi_stats->readUserBytes_fp ;
    stats->writeUserBytes_ssc   = scsi_stats->writeUserBytes_ssc ;
    stats->writeUserBytes_fp    = scsi_stats->writeUserBytes_fp ;
    stats->readMetaBytes_ssc    = scsi_stats->readMetaBytes_ssc ;
    stats->readMetaBytes_fp     = scsi_stats->readMetaBytes_fp ;


#ifdef SCSI_TIMESTAMP
    stats->bucketSize1 = SCSI_RSP_BUCKET_1;
    stats->bucketSize2 = SCSI_RSP_BUCKET_2;
    stats->bucketSize3 = SCSI_RSP_BUCKET_3;
    stats->bucketSize4 = SCSI_RSP_BUCKET_4;
    stats->bucketSize5 = SCSI_RSP_BUCKET_5;
    stats->bucketSize6 = SCSI_RSP_BUCKET_6;
    
    memcpy(&stats->totReadDataRspBuckets, &scsiCtrl->stats.totReadDataRspBuckets, sizeof(scsi_timeTab_t));
    memcpy(&stats->totWriteRspBuckets,    &scsiCtrl->stats.totWriteRspBuckets, sizeof(scsi_timeTab_t));
    memcpy(&stats->fcReadDataRspBuckets,  &scsiCtrl->stats.fcReadDataRspBuckets, sizeof(scsi_timeTab_t));
    memcpy(&stats->fcWriteRspBuckets,     &scsiCtrl->stats.fcWriteRspBuckets, sizeof(scsi_timeTab_t));

#else
    stats->bucketSize1 = 0;  // this means that no buckets are available
#endif
    
    if (isClear)
    {
        scsi_ui_clearStats(0, NULL);
    }
}

// deletes a wwn:lun entry from scsi db when a lun gets deleted
int
scsi_procDelDevMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg)
{
	scsi_delDevReqMsg_t*	rmsg = (scsi_delDevReqMsg_t *)msg;
	scsi_dev_t 				*pdev;
	int						i;
    esm_event_t				esmEvent;
	
	for( i=0; i < SCSI_MAX_DEV_PER_PORT; i++ ) {
		pdev = scsiDevList[i];
		if(pdev && pdev->id == rmsg->id ) { 
			E_LOG(class_1, notice_s, eee_getApplicationId("scsi"), 0, 0, 0,
				  ("scsi: Deleting scsi_id %lld from scsi db\n",
				   pdev->id));

			if(pdev->lun != 0)	{ // do not delete lun 0 !!
				esmEvent.event  = SCSI_DEVO_EVENT_DELETE;
				esmEvent.p1.u.w = 0;
				esmEvent.p2.u.w = 0;
				esm_deliverEvent(pdev->smHandle, &esmEvent);
			}
			else
				pdev->flags.fInitMode = 1;
			break;
		}
	}

    eee_freePacket(edesc);

	return(NFX_OK);
} // scsi_procDelDevMsg

int
scsi_procClearStatsMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg) 
{
	scsi_clearStatsReqMsg_t*	rmsg = (scsi_clearStatsReqMsg_t *)msg;
	scsi_dev_t*					pdev;
	int							i;

	if(rmsg->type == SCSI_STATS_TYPE_SCSI) {
		bzero(&scsiCtrl->stats, sizeof(scsi_stats_t));
	}

	else if(rmsg->type == SCSI_STATS_TYPE_DEV || rmsg->type == SCSI_STATS_TYPE_DEVALL) {
		for(i=0; i < SCSI_MAX_DEV_PER_PORT; i++) {
			pdev = scsiDevList[i];
			if(pdev && (rmsg->type == SCSI_STATS_TYPE_DEVALL || pdev->id == rmsg->id)) {
				pdev->readReqCnt		 = 0;
				pdev->readByteCnt        = 0;
				pdev->readByteLast       = 0;
				pdev->readBytePerSecCur  = 0;
				pdev->readBytePerSecMax  = 0;
				pdev->readRspCnt         = 0;
				pdev->readRspLast        = 0;
				pdev->readIoPerSecCur    = 0;
				pdev->readIoPerSecMax    = 0;
				pdev->readCSECnt		 = 0;
				pdev->readBECnt			 = 0;
				pdev->writeSub8kCnt		 = 0;
				pdev->writeSub8kPaddedCnt= 0;
				pdev->writeReqCnt		 = 0;
				pdev->writeByteCnt       = 0;
				pdev->writeByteLast      = 0;
				pdev->writeBytePerSecCur = 0;
				pdev->writeBytePerSecMax = 0;
				pdev->writeRspCnt        = 0;
				pdev->writeRspLast       = 0;
				pdev->writeIoPerSecCur   = 0;
				pdev->writeIoPerSecMax   = 0;
				pdev->writeCSECnt		 = 0;
				pdev->relayReqCnt		 = 0;
				pdev->relayRspCnt		 = 0;
				pdev->devErrCnt			 = 0;
				memset((void *)pdev->r_chain_stats, 0, sizeof(pdev->r_chain_stats));
				memset((void *)pdev->w_chain_stats, 0, sizeof(pdev->w_chain_stats));
				if(rmsg->type == SCSI_STATS_TYPE_DEV) break;
			}
		}
	}

	eee_freePacket(edesc);

	return(NFX_OK);
} // scsi_procClearStatsMsg

int
scsi_procGetStatsMsg(eee_desc_t *edesc, scsi_msgHdr_t *msg) 
{
	scsi_getStatsReqMsg_t*	rmsg = (scsi_getStatsReqMsg_t *)msg;
	scsi_getStatsRspMsg_t	*rsp;
	scsi_dev_t*				pdev;
	uint32					resp_port = edesc->hdr.src_port;
	uint32					source_port = scsiCtrl->scsiDataAddr;
	int						i;

	rsp = (scsi_getStatsRspMsg_t *)eee_ramAlloc(sizeof(scsi_getStatsRspMsg_t));
	memset((void *)rsp, 0, sizeof(*rsp));
	rsp->hdr.mid = SCSI_MID_GET_STATS_RSP;
	rsp->hdr.cHandle = rmsg->hdr.cHandle;
	rsp->type  = rmsg->type;
	rsp->eCode = SCSI_MSG_ECODE_OK;

	if(rmsg->type == SCSI_STATS_TYPE_SCSI) {
		memcpy((void *)&rsp->stats.scsi_stats, &scsiCtrl->stats, sizeof(scsi_stats_t));
	}

	else {
		for(i=0; i < SCSI_MAX_DEV_PER_PORT; i++) {
			pdev = scsiDevList[i];
			if(pdev && pdev->id == rmsg->id) {
				rsp->stats.dev_stats.wwn = pdev->wwn;
				rsp->stats.dev_stats.lun = pdev->lun;
				rsp->stats.dev_stats.readReqCnt  = pdev->readReqCnt;
				rsp->stats.dev_stats.readByteCnt = pdev->readByteCnt;
				rsp->stats.dev_stats.readByteLast = pdev->readByteLast;
				rsp->stats.dev_stats.readBytePerSecCur = pdev->readBytePerSecCur;
				rsp->stats.dev_stats.readBytePerSecMax = pdev->readBytePerSecMax;
				rsp->stats.dev_stats.readRspCnt = pdev->readRspCnt;
				rsp->stats.dev_stats.readRspLast = pdev->readRspLast;
				rsp->stats.dev_stats.readIoPerSecCur = pdev->readIoPerSecCur;
				rsp->stats.dev_stats.readIoPerSecMax = pdev->readIoPerSecMax;
				rsp->stats.dev_stats.readCSECnt = pdev->readCSECnt;
				rsp->stats.dev_stats.readBECnt = pdev->readBECnt;
				rsp->stats.dev_stats.writeSub8kCnt = pdev->writeSub8kCnt;
				rsp->stats.dev_stats.writeSub8kPaddedCnt = pdev->writeSub8kPaddedCnt;
				rsp->stats.dev_stats.writeReqCnt = pdev->writeReqCnt;
				rsp->stats.dev_stats.writeByteCnt = pdev->writeByteCnt;
				rsp->stats.dev_stats.writeByteLast = pdev->writeByteLast;
				rsp->stats.dev_stats.writeBytePerSecCur = pdev->writeBytePerSecCur;
				rsp->stats.dev_stats.writeBytePerSecMax = pdev->writeBytePerSecMax;
				rsp->stats.dev_stats.writeRspCnt = pdev->writeRspCnt;
				rsp->stats.dev_stats.writeRspLast = pdev->writeRspLast;
				rsp->stats.dev_stats.writeIoPerSecCur = pdev->writeIoPerSecCur;
				rsp->stats.dev_stats.writeIoPerSecMax = pdev->writeIoPerSecMax;
				rsp->stats.dev_stats.writeCSECnt = pdev->writeCSECnt;
				rsp->stats.dev_stats.relayReqCnt = pdev->relayReqCnt;
				rsp->stats.dev_stats.relayRspCnt = pdev->relayRspCnt;
				rsp->stats.dev_stats.devErrCnt = pdev->devErrCnt;
				rsp->stats.dev_stats.genio_q_len = scsi_queue_len(pdev->genio_q),
				rsp->stats.dev_stats.hiprio_q_len = scsi_queue_len(pdev->hiprio_q),
				rsp->stats.dev_stats.other_q_len = scsi_queue_len(pdev->other_q),
				memcpy((void *)&rsp->stats.dev_stats.r_chain_stats, pdev->r_chain_stats, sizeof(pdev->r_chain_stats));
				memcpy((void *)&rsp->stats.dev_stats.w_chain_stats, pdev->w_chain_stats, sizeof(pdev->w_chain_stats));
				break;
			}
		} // end of for
		if(i == SCSI_MAX_DEV_PER_PORT)
			rsp->eCode = SCSI_MSG_ECODE_NOSUCH_DEV;
	}

	if (EEE_GET_DTYPE(resp_port) != EEE_USER_DATA) {
		source_port &= ~EDESC2_DATA_TYPE_MASK;
		source_port |= EEE_GET_DTYPE(resp_port) << EDESC2_DATA_TYPE_SHIFT;
	}
	scsi_sendMessage((uchar8*)rsp, sizeof(scsi_getStatsRspMsg_t), resp_port, source_port);

	eee_freePacket(edesc);

	return 0;
} // scsi_procGetStatsMsg

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procWindowUpdateMsg
 *
 * Description: 	Processes window update msg sent by SSC
 *
 * Created by:		Nitin Bahadur
 *
 * Data Created:	Nov 16, 2002
 *
 * Returns:         NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_procWindowUpdateMsg(eee_desc_t *edesc) 
{
	struct eee_window_update *wu;
	eee_descBuf_t *bd;

	bd = &edesc->bd[1];
	wu = (struct eee_window_update *)(bd->buf);
	
	if(wu->wu_msgtype == EEE_MSG_WINDOW_UPDATE) {
		if(scsiCtrl->sscMsgs.sdm_id != wu->wu_sockid) {	// change in SDM socket, reset window params
			scsiCtrl->sscMsgs.sdm_id = wu->wu_sockid;
			scsiCtrl->sscMsgs.char_space = wu->wu_char_space;
			scsiCtrl->sscMsgs.mbuf_space = wu->wu_mbuf_space;
		}
		else {	// update window params
			scsiCtrl->sscMsgs.char_space += SCSI_EEE_CHAR_SPACE(wu->wu_msg_length);
			scsiCtrl->sscMsgs.mbuf_space += SCSI_EEE_MBUF_SPACE(wu->wu_msg_length);
			if(scsiCtrl->sscMsgs.char_space > wu->wu_char_space) {
				printf("WARNING: Fixing inconsistency in SSC Window char space: old = %d new = %d\n",
						scsiCtrl->sscMsgs.char_space, wu->wu_char_space);
				scsiCtrl->sscMsgs.char_space = wu->wu_char_space;
			}
			if(scsiCtrl->sscMsgs.mbuf_space > wu->wu_mbuf_space) {
				printf("WARNING: Fixing inconsistency in SSC Window mbuf space: old = %d new = %d\n",
						scsiCtrl->sscMsgs.mbuf_space, wu->wu_mbuf_space);
				scsiCtrl->sscMsgs.mbuf_space = wu->wu_mbuf_space;
			}
		}
		//printf("SSC Window Update: sdm_id = %d char_space = %d mbuf_space = %d datalen=%d\n", 
		//		scsiCtrl->sscMsgs.sdm_id, scsiCtrl->sscMsgs.char_space, scsiCtrl->sscMsgs.mbuf_space, wu->wu_msg_length);
	}

	eee_freePacket(edesc);

	return NFX_OK;
} // scsi_procWindowUpdateMsg


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendNullMsg
 *
 * Description: 	Sends a null msg to SDM
 *
 * Created by:		Nitin Bahadur
 *
 * Data Created:	Nov 16, 2002
 *
 * Returns:         NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_sendNullMsg()
{
	static sdm_msgHdr_t msg;
	static char sdm_name[10];
	static uint32 source_port;
	static uint32 dest_port;
	static uint32 sdm_deamon;

	memset((void *)&msg, 0, sizeof(msg));
	msg.mid = SDM_MID_NULL_MSG;

	sprintf(sdm_name, "sdmfc%d", nfxMyCPUId); 
	sdm_deamon = eee_getApplicationId(sdm_name);
	source_port = PACK_UNICAST_PORT_ADDRESS( 	nfxMySlotId,
												nfxMyCPUId,
												scsiCtrl->appId,
												EEE_MGMT,
												NFX_COS_HIGH,
												0);
	dest_port = PACK_UNICAST_PORT_ADDRESS(	0,
											0,
											sdm_deamon,
											EEE_MGMT,
											NFX_COS_HIGH,
											0);
	
	// send null msg to sdm
	eee_sendMessage((void *)&msg, sizeof(msg), dest_port, source_port);	

	// if SDM is up, then change window size
	if(scsiCtrl->sscMsgs.sdm_id) {
		scsiCtrl->sscMsgs.char_space -= SCSI_EEE_CHAR_SPACE(sizeof(msg));
		scsiCtrl->sscMsgs.mbuf_space -= SCSI_EEE_MBUF_SPACE(sizeof(msg));
	}

	return NFX_OK;
} // scsi_sendNullMsg


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_procNullMsgRsp
 *
 * Description: 	Processes null msg response sent by SDM
 *
 * Created by:		Nitin Bahadur
 *
 * Data Created:	Nov 16, 2002
 *
 * Returns:         NFX_OK
 *
 *------------------------------------------------------------------------------
 */
int
scsi_procNullMsgRsp(eee_desc_t *edesc, scsi_msgHdr_t *msg) 
{
	/* destroy the polling timer */
	scsi_msg_queue_timer_destroy();

	/* create the timer for sending ordinary communication pkts */
	scsi_msg_queue_timer_create(SCSI_MSG_QUEUE_TIMER_PERIOD, scsi_msg_q_timer_handler);
	scsi_msg_queue_timer_restart();

	if(edesc) eee_freePacket(edesc);

	return NFX_OK;
} // scsi_procNullMsgRsp

int
scsi_procHugeioReq(eee_desc_t *edesc, scsi_msgHdr_t *msg)
{
	scsi_hugeioReqMsg_t*	req = (scsi_hugeioReqMsg_t *)msg;
	scsi_hugeioRspMsg_t		*rsp;
	scsi_dev_t				*pdev;
	int						op;
	int						eCode = 0;

	// get device structure
	pdev = scsi_ui_findDevice(req->wwn, req->lun);
	if(!pdev) {
		eCode = SCSI_MSG_ECODE_NOSUCH_DEV;
		goto cleanup;
	}

	if(req->size > (8*1024*1024)) {
		eCode = SCSI_MSG_ECODE_INVALID_LEN;
		goto cleanup;
	}

	// check if an op. is already going on
	if(pdev->hugeio_state) {
		eCode = SCSI_MSG_ECODE_ILLEGAL_REQUEST;
		goto cleanup;
	}

	if(req->count <= 0) {
		eCode = SCSI_MSG_ECODE_INVALID_LEN;
		goto cleanup;
	}

	if(req->op == SCSI_HUGEIO_READ)
		op = SD_OPCODE_READ;
	else if(req->op == SCSI_HUGEIO_WRITE)
		op = SD_OPCODE_WRITE;
	else {
		eCode = SCSI_MSG_ECODE_ILLEGAL_REQUEST;
		goto cleanup;

	}

	if(!scsi_start_huge_io(pdev, req->lba, req->size, req->count, op, scsi_huge_io_done_ssc, (void *)edesc))
		return NFX_OK;
	else
		eCode = SCSI_MSG_ECODE_NO_RESOURCE;

cleanup:

	rsp = (scsi_hugeioRspMsg_t *)eee_ramAlloc(sizeof(scsi_hugeioRspMsg_t));
	memset((void *)rsp, 0, sizeof(scsi_hugeioRspMsg_t));
	rsp->wwn = req->wwn;
	rsp->lun = req->lun;
	rsp->eCode = eCode;

	// send back error message
	scsi_sendMessage((uchar8*)rsp, sizeof(scsi_hugeioRspMsg_t), edesc->hdr.src_port, edesc->hdr.dest_port);

	eee_freePacket(edesc);

	return NFX_OK;
} // scsi_procHugeioReq
