/*
 *------------------------------------------------------------------------------
 *
 * Name:	scsiio.c
 *
 * RCS:		$Id: scsiio.c,v 1.14 2002/12/05 00:48:55 nitinb Exp $
 *
 * Copyright (C) 2001, 2002, Agile Storage Inc.
 *
 * Description:	Code for handling large SCSI IOs originated by FP processes.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 3/20/02
 *
 *------------------------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/malloc-api.h>
#include <exports/timer-api.h>
#include <exports/scsi-api.h>
#include <exports/sdm-agent-api.h>
#include <exports/cm-api.h>
#include <exports/debug.h>
#include "scsiio-api.h"
#include "scsiio.h"
#include "scsiio-debug.h"

scsiio_cb_t *scsiio_cb = NULL;

void
scsiio_init(void)
{
	int rc;

	SCSIIO_DBG(1, ("scsiio_init:\n"));

	scsiio_cb = (scsiio_cb_t *)eee_ramAlloc(sizeof(scsiio_cb_t));
	VERIFY(scsiio_cb);
	bzero(scsiio_cb, sizeof(scsiio_cb_t));
	
	scsiio_cb->scsiio_app_id = eee_allocateApplicationId("scsiio");
	VERIFY(scsiio_cb->scsiio_app_id >= 0);
	
	scsiio_cb->scsiio_data_addr = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, 
								nfxMyCPUId,
								scsiio_cb->scsiio_app_id,
								EEE_USER_DATA,
								NFX_COS_HIGH,
								0);
	NFX_SPIN_INIT(&scsiio_cb->scsiio_contextq_lock, SPINLOCK_LEVEL_UNKNOWN);

	rc = eee_registerReceiveFunc(scsiio_cb->scsiio_app_id,
				     scsiio_receive_msg,
				     NULL);
	VERIFY(rc == NFX_OK);
}

scsiio_context_t *
scsiio_allocate_context(void)
{
	int i, j;

	NFX_SPIN_LOCK(&scsiio_cb->scsiio_contextq_lock);
	for (i = scsiio_cb->next_context_id, j = 0; j < SCSIIO_MAX_CONTEXTS; ++j, i = (i+1) % SCSIIO_MAX_CONTEXTS)
		if (scsiio_cb->scsiio_contexts[i].context_id == 0) {
			SCSIIO_DBG(1, ("cpu %d:scsiio_allocate_context: --> %p (%d)\n", 
				       nfxMyCPUId, &scsiio_cb->scsiio_contexts[i], i));
			bzero(&scsiio_cb->scsiio_contexts[i], sizeof(scsiio_context_t));
			scsiio_cb->scsiio_contexts[i].context_id = i + 1;
			scsiio_cb->next_context_id = (i + 1) % SCSIIO_MAX_CONTEXTS;
			NFX_SPIN_UNLOCK(&scsiio_cb->scsiio_contextq_lock);
			scsiio_cb->scsiio_contexts[i].rsp_timer_handle = -1;
			scsiio_cb->scsiio_contexts[i].ack_timer_handle = -1;
			scsiio_cb->scsiio_contexts[i].exch_timer_handle = -1;
			return(&scsiio_cb->scsiio_contexts[i]);
		}
	NFX_SPIN_UNLOCK(&scsiio_cb->scsiio_contextq_lock);
	SCSIIO_DBG(0, ("cpu %d: scsiio_allocate_context: Failed\n", nfxMyCPUId));
	return(NULL);
}

void              
scsiio_deallocate_context(scsiio_context_t *context)
{
	SCSIIO_DBG(1, ("cpu %d: scsiio_deallocate_context: --> %p (%d)\n", 
		       nfxMyCPUId, context, context->context_id-1));

	if (context->wrt_buflist)
		eee_ramDealloc(context->wrt_buflist);

	VERIFY(context->rsp_timer_handle == -1);
	VERIFY(context->ack_timer_handle == -1);
	VERIFY(context->exch_timer_handle == -1);
	VERIFY(context->rd_edesc_first == NULL);
	VERIFY(context->rd_edesc_last == NULL);
	ATOMIC_AND_INT(&context->context_id, 0);
}

scsiio_context_t *
scsiio_lookup_context(uint32 context_id)
{
	scsiio_context_t *context;
	
	if (context_id && (context_id <= SCSIIO_MAX_CONTEXTS)) {
		context = &scsiio_cb->scsiio_contexts[context_id - 1];
		return((context->context_id && context->context_id == context_id ) ? context : NULL);
	}
	else
		return(NULL);
}

int 
scsiio_read(uint32 scsi_handle, char *buf, uint32 buflen, scsiio_callback_t callback, void *arg1, void *arg2)
{
	scsiio_context_t *	context;
	int32			rc;

	SCSIIO_DBG(1, ("cpu %d: scsiio_read: handle 0x%x, buf %p, buflen %d, callback %p, arg1 %p, arg2 %p\n",
		       nfxMyCPUId, scsi_handle, buf, buflen, callback, arg1, arg2));

	context = scsiio_allocate_context();
	VERIFY(context);

	rc = sdma_handleToPath(scsi_handle, &context->scsi_path, &context->scsi_port);
	if (rc == -1) {
		SCSIIO_DBG(0, ("cpu %d: scsiio_read: sdma_handleToPath failed.\n", nfxMyCPUId));
		scsiio_deallocate_context(context);
		return(NFX_ERR);
	}
		
	context->callback = callback;
	context->arg1 = arg1;
	context->arg2 = arg2;
	context->imm_free = 0;
	context->scsi_handle = scsi_handle;
	context->buf = context->current_buffer_pointer = buf;
	context->buflen = buflen;
	context->current_buffer_offset = 0;
	context->write = 0;
	context->disk_io = 0;
	NFX_SPIN_INIT(&context->context_lock, SPINLOCK_LEVEL_UNKNOWN);

	/* if data is going to take multiple sequences and user has not supplied a buffer
	 * then allocate a msg. queue */
	if(!buf && buflen > BLOCKSIZE_8192) {
		context->msg_q_size = buflen / BLOCKSIZE_8192 + ((buflen % BLOCKSIZE_8192) ? 1: 0) + 1; /* +1 since we don't use msg_q[0] */
		context->msg_q = (eee_desc_t **)eee_ramAlloc(sizeof(eee_desc_t *) * context->msg_q_size);
		VERIFY(context->msg_q);
		memset((void *)context->msg_q, 0, sizeof(eee_desc_t *) * context->msg_q_size);
	}

	scsiio_transmit_read_req(context);
	
	scsiio_start_rsp_timer(context);
	
	context->state = SCSIIO_CONTEXT_STATE_AWAITING_RSP;

	return(NFX_OK);
} /* scsiio_read() */

int 
scsiio_disk_read(uint32 scsi_handle, char *buf, uint32 lba, uint32 buflen, scsiio_callback_t callback, void *arg1, void *arg2)
{
	scsiio_context_t *	context;
	int32			rc;

	SCSIIO_DBG(1, ("cpu %d: scsiio_disk_read: handle 0x%x, buf %p, lba %d, buflen %d, callback %p, arg1 %p, arg2 %p\n",
		       nfxMyCPUId, scsi_handle, buf, lba, buflen, callback, arg1, arg2));

	context = scsiio_allocate_context();
	VERIFY(context);

	rc = sdma_handleToPath(scsi_handle, &context->scsi_path, &context->scsi_port);
	if (rc == -1) {
		SCSIIO_DBG(0, ("cpu %d: scsiio_disk_read: sdma_handleToPath failed.\n", nfxMyCPUId));
		scsiio_deallocate_context(context);
		return(NFX_ERR);
	}
		
	context->callback = callback;
	context->arg1 = arg1;
	context->arg2 = arg2;
	context->imm_free = 0;
	context->scsi_handle = scsi_handle;
	context->buf = context->current_buffer_pointer = buf;
	context->lba = lba;
	context->buflen = buflen;
	context->current_buffer_offset = 0;
	context->write = 0;
	context->disk_io = 1;
	NFX_SPIN_INIT(&context->context_lock, SPINLOCK_LEVEL_UNKNOWN);

	/* if data is going to take multiple sequences and user has not supplied a buffer
	 * then allocate a msg. queue */
	if(!buf && buflen > BLOCKSIZE_8192) {
		context->msg_q_size = buflen / BLOCKSIZE_8192 + ((buflen % BLOCKSIZE_8192) ? 1: 0) + 1; /* +1 since we don't use msg_q[0] */
		context->msg_q = (eee_desc_t **)eee_ramAlloc(sizeof(eee_desc_t *) * context->msg_q_size);
		VERIFY(context->msg_q);
		memset((void *)context->msg_q, 0, sizeof(eee_desc_t *) * context->msg_q_size);
	}

	scsiio_transmit_read_req(context);
	
	scsiio_start_rsp_timer(context);
	
	context->state = SCSIIO_CONTEXT_STATE_AWAITING_RSP;

	return(NFX_OK);
} /* scsiio_disk_read() */

int
scsiio_write(uint32 scsi_handle, char *buf, uint32 buflen, scsiio_callback_t callback, void *arg1, void *arg2, int imm_free)
{
	scsiio_context_t *	context;
	int32			i;
	int32			rc;

	SCSIIO_DBG(1, ("cpu %d: scsiio_write: handle 0x%x, buf %p, buflen %d, callback %p, arg1 %p, arg2 %p, imm_free %d\n",
		       nfxMyCPUId, scsi_handle, buf, buflen, callback, arg1, arg2, imm_free));

	context = scsiio_allocate_context();
	VERIFY(context);

	rc = sdma_handleToPath(scsi_handle, &context->scsi_path, &context->scsi_port);
	if (rc == -1) {
		SCSIIO_DBG(0, ("cpu %d: scsiio_write: sdma_handleToPath failed.\n", nfxMyCPUId));
		scsiio_deallocate_context(context);
		return(NFX_ERR);
	}
		
	context->callback = callback;
	context->arg1 = arg1;
	context->arg2 = arg2;
	context->imm_free = imm_free;
	context->scsi_handle = scsi_handle;
	context->buf = context->current_buffer_pointer = buf;
	context->buflen = buflen;
	context->current_buffer_offset = 0;
	context->write = 1;
	context->disk_io = 0;

	context->transaction_id = context->context_id;
	context->max_sequence_id = 0;

	context->wrt_buflist_len = buflen / SCSIIO_MAX_PACKET_DATA + 2;
	context->wrt_buflist = eee_ramAlloc(sizeof(eee_descBuf_t) * context->wrt_buflist_len);
	bzero(context->wrt_buflist, sizeof(eee_descBuf_t) * context->wrt_buflist_len);
	for (i = 0; context->current_buffer_offset < context->buflen; ++i) {
		context->wrt_buflist[i].buf = context->current_buffer_pointer;
		if ((address_t)context->current_buffer_pointer % 64)
			context->wrt_buflist[i].len = 64 - ((address_t)context->current_buffer_pointer % 64);
		else
			context->wrt_buflist[i].len = MIN(SCSIIO_MAX_PACKET_DATA, 
						      context->buflen - context->current_buffer_offset);
		context->current_buffer_offset += context->wrt_buflist[i].len;
		context->current_buffer_pointer += context->wrt_buflist[i].len;

		SCSIIO_DBG(1, ("cpu %d: scsiio_write: [%2d] buf %p, len %d\n", i, 
			       nfxMyCPUId, context->wrt_buflist[i].buf, context->wrt_buflist[i].len));

		++context->max_sequence_id;
	}

	scsiio_reset_write_context(context);
	scsiio_transmit_write_seq(context);
	scsiio_start_ack_timer(context);

	scsiio_start_rsp_timer(context);

	context->state = SCSIIO_CONTEXT_STATE_AWAITING_ACK;

	return(NFX_OK);
} /* scsiio_write() */

int
scsiio_disk_write(uint32 scsi_handle, char *buf, uint32 lba, uint32 buflen, scsiio_callback_t callback, void *arg1, void *arg2, int imm_free)
{
	scsiio_context_t *	context;
	int32			i;
	int32			rc;

	SCSIIO_DBG(1, ("cpu %d: scsiio_disk_write: handle 0x%x, buf %p, buflen %d, callback %p, arg1 %p, arg2 %p, imm_free %d\n",
		       nfxMyCPUId, scsi_handle, buf, buflen, callback, arg1, arg2, imm_free));

	context = scsiio_allocate_context();
	VERIFY(context);

	rc = sdma_handleToPath(scsi_handle, &context->scsi_path, &context->scsi_port);
	if (rc == -1) {
		SCSIIO_DBG(0, ("cpu %d: scsiio_disk_write: sdma_handleToPath failed.\n", nfxMyCPUId));
		scsiio_deallocate_context(context);
		return(NFX_ERR);
	}
		
	context->callback = callback;
	context->arg1 = arg1;
	context->arg2 = arg2;
	context->imm_free = imm_free;
	context->scsi_handle = scsi_handle;
	context->buf = context->current_buffer_pointer = buf;
	context->lba = lba;
	context->buflen = buflen;
	context->current_buffer_offset = 0;
	context->write = 1;
	context->disk_io = 1;

	context->transaction_id = context->context_id;
	context->max_sequence_id = 0;

	context->wrt_buflist_len = buflen / SCSIIO_MAX_PACKET_DATA + 2;
	context->wrt_buflist = eee_ramAlloc(sizeof(eee_descBuf_t) * context->wrt_buflist_len);
	bzero(context->wrt_buflist, sizeof(eee_descBuf_t) * context->wrt_buflist_len);
	for (i = 0; context->current_buffer_offset < context->buflen; ++i) {
		context->wrt_buflist[i].buf = context->current_buffer_pointer;
		if ((address_t)context->current_buffer_pointer % 64)
			context->wrt_buflist[i].len = 64 - ((address_t)context->current_buffer_pointer % 64);
		else
			context->wrt_buflist[i].len = MIN(SCSIIO_MAX_PACKET_DATA, 
						      context->buflen - context->current_buffer_offset);
		context->current_buffer_offset += context->wrt_buflist[i].len;
		context->current_buffer_pointer += context->wrt_buflist[i].len;

		SCSIIO_DBG(1, ("cpu %d: scsiio_disk_write: [%2d] buf %p, len %d\n", nfxMyCPUId, i, 
			       context->wrt_buflist[i].buf, context->wrt_buflist[i].len));

		++context->max_sequence_id;
	}

	scsiio_reset_write_context(context);
	scsiio_transmit_write_seq(context);
	scsiio_start_ack_timer(context);

	scsiio_start_rsp_timer(context);

	context->state = SCSIIO_CONTEXT_STATE_AWAITING_ACK;

	return(NFX_OK);
} /* scsiio_disk_write() */

void 
scsiio_receive_msg(eee_desc_t *eDesc)
{
	eee_descBuf_t *		bd;
	scsi_msgHdr_t *		hdr;
	uint32			mid;
	scsiio_context_t *	context;

	if (eDesc == NULL)
		return;

	/* XXX We assume there's only one buffer */
	bd = &eDesc->bd[1];
	hdr = (scsi_msgHdr_t *)(bd->buf);
	mid = hdr->mid;
	context = scsiio_lookup_context(hdr->cHandle);
	if (context == NULL) {
		SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: mid 0x%x; Invalid context_id/cHandle 0x%x -- dropping.\n",
			       nfxMyCPUId, mid, hdr->cHandle));
		eee_freePacket(eDesc);
		return;
	}

	switch(mid) {
	case SCSI_MID_WRITE_RSP:
		{
			scsi_writeRspMsg_t * rsp = (scsi_writeRspMsg_t *)hdr;

			SCSIIO_DBG(1, ("cpu %d: scsiio_receive_msg: SCSI_MID_WRITE_RSP context %p(%d), eCode 0x%x\n",
				       nfxMyCPUId, context, context->context_id, rsp->eCode));

			VERIFY(context->state == SCSIIO_CONTEXT_STATE_AWAITING_RSP);

			if (context->imm_free)
				eee_ramDealloc(context->buf);

			scsiio_stop_rsp_timer(context);
			
			(*context->callback)(rsp->eCode, 
					     context->buf, 
					     context->buflen, 
					     context->arg1, context->arg2);

			scsiio_deallocate_context(context);
			eee_freePacket(eDesc);
		}
		break;

	case SCSI_MID_READ_RSP:
		{
			scsi_readRspMsg_t * rsp = (scsi_readRspMsg_t *)hdr;
			uint32		    trans_seq_id;
			char *		    scsi_read_data;
			int32			seq_id;

			SCSIIO_DBG(1, ("cpu %d: scsiio_receive_msg: SCSI_MID_READ_RSP: context %p(%d): eCode 0x%x, trans_seq_id 0x%x\n",
				       nfxMyCPUId, context, context->context_id, rsp->eCode, rsp->transSeqId));

			/* Check if dataLen specified in response matches the actual data eee gives us */
			if(rsp->dataLen != (EEE_GET_PKT_LEN(eDesc) - sizeof(scsi_readRspMsg_t))) {
				SCSIIO_DBG(0, ("cpu %d: scsi_receive_msg: Expected %d bytes, got %ld bytes\n",
							nfxMyCPUId, rsp->dataLen, EEE_GET_PKT_LEN(eDesc) - sizeof(scsi_readRspMsg_t)));
				goto read_sequence_error;
			}

			/* Is this a multi-frame sequence */
			if ((trans_seq_id = rsp->transSeqId)) { 
				seq_id = SCSI_SEQUENCE_ID(trans_seq_id);

				if (SCSI_SEQUENCE_MAX(trans_seq_id) == 0) {
					/* Invalid SEQUENCE MAX, NACK it */
					SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: SCSI_MID_READ_RSP: "
						       "Invalid Sequence Max for READ response %d.\n", 
							   nfxMyCPUId, SCSI_SEQUENCE_MAX(trans_seq_id)));
					goto read_sequence_error;
				}
				if (seq_id == 0) {
					/* Invalid SEQUENCE ID, NACK it */
					SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: SCSI_MID_READ_RSP: "
						       "Invalid Sequence Id for READ response; expected non-zero, got 0.\n", nfxMyCPUId));
					goto read_sequence_error;
				}
			
				/* First packet in multi-packet sequence ? */
				if (seq_id == 1) {
					if(context->msg_q) 
						VERIFY((context->msg_q_size-1) >= context->max_sequence_id); /* -1 b'coz we allocated 1 extra entry */
					if(context->transaction_id) 
						VERIFY(context->transaction_id == SCSI_TRANSACTION_ID(trans_seq_id));
					context->transaction_id = SCSI_TRANSACTION_ID(trans_seq_id);
					context->max_sequence_id = SCSI_SEQUENCE_MAX(trans_seq_id);
					scsiio_start_exch_timer(context);
				}
				else {
					if (context->transaction_id != SCSI_TRANSACTION_ID(trans_seq_id)) {
						if(context->transaction_id)	{
							/* Invalid TRANSACTION ID, NACK it */
							SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: SCSI_MID_READ_RSP: "
								       "Invalid Transaction ID for READ response; got 0x%x, expected 0x%x.\n",
								       nfxMyCPUId, SCSI_TRANSACTION_ID(trans_seq_id), context->transaction_id));
							goto read_sequence_error;
						}
						else { /* transaction id not set, so set it here. This happens when seq > 1 is processed
								  before seq. # 1 */
							context->transaction_id = SCSI_TRANSACTION_ID(trans_seq_id);
							context->max_sequence_id = SCSI_SEQUENCE_MAX(trans_seq_id);
						}
					}
				}
				
				scsiio_cb->num_reads[nfxMyCPUId-2]++;	

				/* 
				 * Copy data into requestor's buffer
				 * if the caller provided a read
				 * buffer. Otherwise link onto the
				 * eDesc chain.
				 */
				if (context->buf) {
					scsi_read_data = (char *)rsp;
					scsi_read_data += sizeof(scsi_readRspMsg_t);
					bcopy(scsi_read_data, &context->buf[rsp->dataOffset], rsp->dataLen);
				}
				else {
					uint32	buflen;

					/* avoid duplicates */
					if(context->msg_q[seq_id]) return;

					context->msg_q[seq_id] = eDesc;

					/* Strip off SCSI header */
					VERIFY(EEE_GET_NUM_BUFS(eDesc) == 1);
					bd->buf += sizeof(scsi_readRspMsg_t);
					bd->len -= sizeof(scsi_readRspMsg_t);
					buflen = EEE_GET_PKT_LEN(eDesc);
					EEE_SET_PKT_LEN(eDesc, buflen - sizeof(scsi_readRspMsg_t));
					
				}

				NFX_SPIN_LOCK(&context->context_lock);
				context->current_sequence_id++;		/* increment number of pkts. recd. */
				context->current_buffer_offset += rsp->dataLen;

				/* Last packet in multi-packet sequence ? */
				/* If not last packet, then return */
				if (context->current_sequence_id < context->max_sequence_id) {
					NFX_SPIN_UNLOCK(&context->context_lock);
					if (context->buf)
						eee_freePacket(eDesc);					
					return;
				}

				NFX_SPIN_UNLOCK(&context->context_lock);

				/* All packets have been recd., so now send the msg up to requesting application */
				if(!context->buf) {     /* chain messages into eDesc chain */
					int     i;
					context->rd_edesc_first = context->rd_edesc_last = context->msg_q[1];
					for(i=2; i <= context->max_sequence_id; i++) {
						VERIFY(context->msg_q[i]);
						eee_chainDesc(context->rd_edesc_last, context->msg_q[i]);
                        context->rd_edesc_last = context->msg_q[i];
					}
				}

				scsiio_stop_rsp_timer(context);
				scsiio_stop_exch_timer(context);
				scsiio_transmit_ack(context, 0);

				(*context->callback)(rsp->eCode,
						     context->buf ? (void *)context->buf : context->rd_edesc_first,
						     context->current_buffer_offset, 
						     context->arg1, context->arg2);
				context->rd_edesc_first = context->rd_edesc_last = NULL;
				scsiio_deallocate_context(context);
				if (context->buf)
					eee_freePacket(eDesc);
				return;
			} /* if (trans_seq_id = rsp->transSeqId) */
			else {
				scsiio_cb->num_reads[nfxMyCPUId-2]++;	

				/*
				 * Copy data into requestor's buffer
				 * if the caller provided a read
				 * buffer. Otherwise return the eDesc.
				 */
				if (context->buf) {
					scsi_read_data = (char *)rsp;
					scsi_read_data += sizeof(scsi_readRspMsg_t);
					bcopy(scsi_read_data, context->current_buffer_pointer, rsp->dataLen);
					context->current_buffer_pointer += rsp->dataLen;
					context->current_buffer_offset += rsp->dataLen;
				}
				else {
					uint32  buflen;

					VERIFY(context->rd_edesc_first == NULL);
					context->rd_edesc_first = context->rd_edesc_last = eDesc;
					context->current_buffer_offset += rsp->dataLen;

					/* Strip off SCSI header */
					VERIFY(EEE_GET_NUM_BUFS(eDesc) == 1);
					bd->buf += sizeof(scsi_readRspMsg_t);
					bd->len -= sizeof(scsi_readRspMsg_t);
					buflen = EEE_GET_PKT_LEN(eDesc);
					EEE_SET_PKT_LEN(eDesc, buflen - sizeof(scsi_readRspMsg_t));
				}

				scsiio_stop_rsp_timer(context);
				
				(*context->callback)(rsp->eCode, 
						     context->buf ? (void *)context->buf : context->rd_edesc_first,
						     context->current_buffer_offset, 
						     context->arg1, context->arg2);
				context->rd_edesc_first = context->rd_edesc_last = NULL;
				scsiio_deallocate_context(context);
				if (context->buf)
					eee_freePacket(eDesc);
				return;
			}

		read_sequence_error:
			if (context->exch_timer_handle >= 0)
				scsiio_stop_exch_timer(context);
			scsiio_transmit_ack(context, 1);
			scsiio_reset_read_context(context);
			eee_freePacket(eDesc);
			return;
		}
		break;

	case SCSI_MID_ACK:
		{
			scsi_ackMsg_t * ack = (scsi_ackMsg_t *)hdr;

			SCSIIO_DBG(1, ("cpu %d: scsiio_receive_msg: SCSI_MID_ACK context %p(%d), nack %d\n", 
				       nfxMyCPUId, context, context->context_id, ack->nack));

			if (!context->write) {
				/* Don't expect a ACK when reading */
				SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: SCSI_MID_ACK context %p(%d); unexpected ACK for a read context.\n",
					       nfxMyCPUId, context, context->context_id));
				eee_freePacket(eDesc);
				return;
			}
			
			switch (context->state) {
			case SCSIIO_CONTEXT_STATE_AWAITING_ACK:
				if (ack->nack) {
					/* Scheduled a write sequence retransmit */
					++context->nack_received;
				}
				else {
					scsiio_stop_ack_timer(context);
					context->state = SCSIIO_CONTEXT_STATE_AWAITING_RSP;
				}
				eee_freePacket(eDesc);
				break;

			case SCSIIO_CONTEXT_STATE_AWAITING_RSP:
				/* Don't expect a while waiting for response */
				SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: SCSI_MID_ACK context %p(%d); unexpected ACK "
					       "SCSIIO_CONTEXT_STATE_AWAITING_RSP state.\n", 
					       nfxMyCPUId, context, context->context_id));
				eee_freePacket(eDesc);
				break;
				
			default:
				VERIFY(0);
			}
		}
		break;

	case SCSI_MID_ERR_RSP: 
		{
			scsi_errRspMsg_t * rsp = (scsi_errRspMsg_t *)hdr;

			SCSIIO_DBG(0, ("cpu %d: scsiio_receive_msg: SCSI_MID_ERR_RSP context %p(%d), eCode 0x%x\n",
				       nfxMyCPUId, context, context->context_id, rsp->eCode));

			if (context->write) {
				switch (context->state) {
				case SCSIIO_CONTEXT_STATE_AWAITING_ACK:
					scsiio_stop_ack_timer(context);
					break;

				case SCSIIO_CONTEXT_STATE_AWAITING_RSP:
					break;
					
				default:
				VERIFY(0);
				}
			}
			else {
				VERIFY(context->state == SCSIIO_CONTEXT_STATE_AWAITING_RSP);
				if (context->exch_timer_handle >= 0)
					scsiio_stop_exch_timer(context);
			}

			scsiio_stop_rsp_timer(context);
			
			if (context->imm_free)
				eee_ramDealloc(context->buf);

			(*context->callback)(rsp->eCode, 
					     context->buf, 
					     0, 
					     context->arg1, context->arg2);
			scsiio_free_read_chain(context);
			scsiio_deallocate_context(context);
			eee_freePacket(eDesc);
		}
		break;
		
	default: 
		printf("scsiio_receive_msg: Invalid Message: %x\n", mid);
		eee_freePacket(eDesc);
	}
	return;
} /* scsiio_receive_msg() */

void
scsiio_free_read_chain(scsiio_context_t *context)
{
	eee_desc_t	*ed;
	int			i;
	
    if (context->buf == NULL) {
		for(i=1; i < context->msg_q_size; i++) {
			ed = (eee_desc_t *)context->msg_q[i];
			context->msg_q[i] = NULL;
			if(ed) eee_freePacket(ed);
		}

        context->rd_edesc_first = NULL;
        context->rd_edesc_last = NULL;
    }
    return;
}

int32 
scsiio_transmit_read_req(scsiio_context_t *context)
{
	eee_desc_t *		ed;
	scsi_msgHdr_t *		scsi_hdr;
	scsi_readReqMsg_t *	scsi_read;
	int32			total_byte_count;
	int32			total_buf_count;
	int32			rc;

	ed = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
	VERIFY(ed);
	bzero(ed, sizeof(eee_desc_t));
	
	total_byte_count = sizeof(scsi_readReqMsg_t);
	total_buf_count = 1;
	scsi_hdr = (scsi_msgHdr_t *) eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_SMALL, 0);
	VERIFY(scsi_hdr);

	(address_t)scsi_hdr += (MALLOC_ALIGNED_SIZE*2);
	(address_t)scsi_hdr -= sizeof(scsi_readReqMsg_t);
	bzero(scsi_hdr, sizeof(scsi_readReqMsg_t));

	ed->bd[1].buf = (uchar8 *)scsi_hdr;
	ed->bd[1].len = total_byte_count;
	ed->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;
        
	scsi_hdr->mid = SCSI_MID_READ_REQ;
	scsi_hdr->cHandle = context->context_id;
	scsi_hdr->path = context->scsi_path;

	scsi_read = (scsi_readReqMsg_t *)scsi_hdr;
	scsi_read->blkCnt = context->buflen;

	if (context->disk_io) {
		scsi_read->blkOffset = context->lba;
		scsi_read->blkCnt /= SCSIIO_DISK_BLOCKSIZE;
	}

	ed->hdr.control = (total_buf_count << EDESC_BUFCNT_SHIFT) | total_byte_count;
	ed->hdr.dest_port = context->scsi_port;
	ed->hdr.src_port = scsiio_cb->scsiio_data_addr;

	SCSIIO_DBG(1, ("cpu %d: scsiio_transmit_read_req: cHandle 0x%x, path 0x%x, transSeqId 0x%x, blkCnt 0x%x\n",
		       nfxMyCPUId,
			   scsi_read->hdr.cHandle,
		       scsi_read->hdr.path,
		       scsi_read->transSeqId,
		       scsi_read->blkCnt));
#ifdef DEBUG
	if (scsiio_dbg_level >= 3)
		eee_descDisplay(ed);
#endif
	rc = eee_forwardPacket(ed);
	if (rc != EEE_FORWARDED_PACKET) {
		if (rc != EEE_DROP_PACKET)
			eee_freePacket(ed);
		SCSIIO_DBG(0, ("cpu %d: scsiio_read: eee_forwardPacket failed.\n", nfxMyCPUId));
		scsiio_deallocate_context(context);
		return(NFX_ERR);
	}
	return(NFX_OK);
}

int32 
scsiio_transmit_write_seq(scsiio_context_t *context)
{
	eee_desc_t *		ed;
	scsi_msgHdr_t *		scsi_hdr;
	scsi_writeReqMsg_t *	scsi_write;
	char *			scsi_write_data;
	int32			total_byte_count;
	int32			total_buf_count;
	int			i;
	int32			rc;

	for (i = 0; i < context->max_sequence_id; ++i) {
		ed = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
		VERIFY(ed);
		bzero(ed, sizeof(eee_desc_t));
	
		total_byte_count = sizeof(scsi_writeReqMsg_t);
		total_buf_count = 1;
		scsi_hdr = (scsi_msgHdr_t *) eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_SMALL, 0);
		VERIFY(scsi_hdr);

		(address_t)scsi_hdr += (MALLOC_ALIGNED_SIZE*2);
		(address_t)scsi_hdr -= sizeof(scsi_writeReqMsg_t);
		bzero(scsi_hdr, sizeof(scsi_writeReqMsg_t));

		ed->bd[1].buf = (uchar8 *)scsi_hdr;
		ed->bd[1].len = total_byte_count;
		ed->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;
        
		scsi_hdr->mid = SCSI_MID_WRITE_REQ;
		scsi_hdr->cHandle = context->context_id;
		scsi_hdr->path = context->scsi_path;

		scsi_write = (scsi_writeReqMsg_t *)scsi_hdr;
		scsi_write->transSeqId = SCSI_MSG_TRANSSEQID_PACK(context->transaction_id,
								  context->max_sequence_id,
								  context->current_sequence_id);
		scsi_write->blkCnt = context->buflen;

		if (context->disk_io) {
			scsi_write->blkOffset = context->lba;
			scsi_write->blkCnt /= SCSIIO_DISK_BLOCKSIZE;
		}

		if ((address_t)context->wrt_buflist[i].buf % 64) {
			VERIFY(i == 0);
			VERIFY(context->wrt_buflist[i].len < 64);

			scsi_write_data = (char *)scsi_write + sizeof(scsi_writeReqMsg_t);
			bcopy(context->wrt_buflist[i].buf, scsi_write_data, context->wrt_buflist[i].len);
			total_byte_count += context->wrt_buflist[i].len;
			ed->bd[1].len = total_byte_count;
		}
		else {
			ed->bd[2].buf = context->wrt_buflist[i].buf;
			ed->bd[2].len = context->wrt_buflist[i].len;
			ed->bd[2].control = EDESC_BD_CMD_NO_DEALLOC | EDESC_BD_CMD_IN_USE | EDESC_BD_ULTRA;
			total_byte_count += context->wrt_buflist[i].len;
			++total_buf_count;
		}

		ed->hdr.control = (total_buf_count << EDESC_BUFCNT_SHIFT) | total_byte_count;
		ed->hdr.dest_port = context->scsi_port;
		ed->hdr.src_port = scsiio_cb->scsiio_data_addr;

		SCSIIO_DBG(1, ("cpu %d: scsiio_write: cHandle 0x%x, path 0x%x, transSeqId 0x%x, blkCnt 0x%x\n",
				   nfxMyCPUId,
			       scsi_write->hdr.cHandle,
			       scsi_write->hdr.path,
			       scsi_write->transSeqId,
			       scsi_write->blkCnt));
#ifdef DEBUG
		if (scsiio_dbg_level >= 3)
			eee_descDisplay(ed);
#endif
		rc = eee_forwardPacket(ed);
		if (rc != EEE_FORWARDED_PACKET) {
			if (rc != EEE_DROP_PACKET)
				eee_freePacket(ed);
			SCSIIO_DBG(0, ("cpu %d: scsiio_write: eee_forwardPacket failed.\n", nfxMyCPUId));
			scsiio_deallocate_context(context);
			return(NFX_ERR);
		}

		scsiio_cb->num_writes[nfxMyCPUId-2]++;	
		++context->current_sequence_id;

	} /* for (i) */

	return(NFX_OK);
}

int32
scsiio_transmit_ack(scsiio_context_t *context, int nack)
{
	eee_desc_t *		ed;
	scsi_msgHdr_t *		scsi_hdr;
	scsi_ackMsg_t *		scsi_ack;
	int32                   total_byte_count;
        int32                   total_buf_count;
	int32			rc;

	SCSIIO_DBG(1, ("cpu %d: scsiio_transmit_ack: context %p(%d), nack %d\n",
		       nfxMyCPUId, context, context->context_id, nack));

	ed = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
	VERIFY(ed);
	bzero(ed, sizeof(eee_desc_t));

	total_byte_count = sizeof(scsi_ackMsg_t);
        total_buf_count = 1;
	scsi_hdr = (scsi_msgHdr_t *) eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_SMALL, 0);
	VERIFY(scsi_hdr);

	(address_t)scsi_hdr += (MALLOC_ALIGNED_SIZE*2);
	(address_t)scsi_hdr -= sizeof(scsi_ackMsg_t);
	bzero(scsi_hdr, sizeof(scsi_ackMsg_t));

	ed->bd[1].buf = (uchar8 *)scsi_hdr;
	ed->bd[1].len = total_byte_count;
	ed->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_SMALL;
        
	scsi_hdr->mid = SCSI_MID_ACK;
	scsi_hdr->cHandle = context->context_id;
	scsi_hdr->path = context->scsi_path;

	scsi_ack = (scsi_ackMsg_t *)scsi_hdr;
	scsi_ack->nack = nack;

	ed->hdr.control = (total_buf_count << EDESC_BUFCNT_SHIFT) | total_byte_count;
	ed->hdr.dest_port = context->scsi_port;
	ed->hdr.src_port = scsiio_cb->scsiio_data_addr;

	rc = eee_forwardPacket(ed);
	if (rc != EEE_FORWARDED_PACKET) {
		if (rc != EEE_DROP_PACKET)
			eee_freePacket(ed);
		SCSIIO_DBG(0, ("cpu %d: scsiio_transmit_ack: eee_forwardPacket failed.\n", nfxMyCPUId));
		scsiio_deallocate_context(context);
		return(NFX_ERR);
	}
	return(NFX_OK);
}

void
scsiio_reset_read_context(scsiio_context_t *context)
{
	context->current_buffer_pointer = context->buf;
	context->current_buffer_offset = 0;
	context->transaction_id = 0;
	context->current_sequence_id = 0;
	scsiio_free_read_chain(context);
	ATOMIC_AND_INT(&context->transaction_id, 0);
}

void
scsiio_reset_write_context(scsiio_context_t *context)
{
	context->current_sequence_id = 1;
}

int32 
scsiio_rsp_timeout_handler(int32 timer_handle, scsiio_context_t *context)
{
	int32		rc;

	VERIFY(context == scsiio_lookup_context(context->context_id));
	VERIFY(timer_handle == context->rsp_timer_handle);

	SCSIIO_DBG(0, ("cpu %d: scsiio_rsp_timeout_handler: context %p(%d)\n",
		       nfxMyCPUId, context, context->context_id));
	
	rc = eee_timerDestroy(context->rsp_timer_handle);
	VERIFY(rc == NFX_OK);
	context->rsp_timer_handle = -1;

	if (context->write) {
		if (context->imm_free)
			eee_ramDealloc(context->buf);
		if (context->ack_timer_handle >= 0)
			scsiio_stop_ack_timer(context);
	}
	else {
		if (context->exch_timer_handle >= 0)
			scsiio_stop_exch_timer(context);
	}

	(*context->callback)(SCSI_MSG_ECODE_TIMEOUT, 
			     context->buf, 
			     0, 
			     context->arg1, context->arg2);
	scsiio_free_read_chain(context);
	scsiio_deallocate_context(context);

	return(NFX_OK);
}

int32 
scsiio_ack_timeout_handler(int32 timer_handle, scsiio_context_t *context)
{
	int32		rc;

	VERIFY(context == scsiio_lookup_context(context->context_id));
	VERIFY(timer_handle == context->ack_timer_handle);

	SCSIIO_DBG(0, ("cpu %d: scsiio_ack_timeout_handler: context %p(%d)\n",
		       nfxMyCPUId, context, context->context_id));
	
	rc = eee_timerDestroy(context->ack_timer_handle);
	VERIFY(rc == NFX_OK);
	context->ack_timer_handle = -1;

	VERIFY(context->write);
	VERIFY(context->state == SCSIIO_CONTEXT_STATE_AWAITING_ACK);

	if (context->nack_received) {
		scsiio_reset_write_context(context);
		scsiio_transmit_write_seq(context);
		scsiio_start_ack_timer(context);
	}
	else {
		if (context->imm_free)
			eee_ramDealloc(context->buf);
	
		scsiio_stop_rsp_timer(context);

		(*context->callback)(SCSI_MSG_ECODE_TIMEOUT, 
				     context->buf, 
				     0, 
				     context->arg1, context->arg2);
		scsiio_free_read_chain(context);
		scsiio_deallocate_context(context);
	}

	return(NFX_OK);
}

int32 
scsiio_exch_timeout_handler(int32 timer_handle, scsiio_context_t *context)
{
	int32		rc;

	VERIFY(context == scsiio_lookup_context(context->context_id));
	VERIFY(timer_handle == context->exch_timer_handle);

	SCSIIO_DBG(0, ("cpu %d: scsiio_exch_timeout_handler: context %p(%d)\n",
		       nfxMyCPUId, context, context->context_id));
	
	rc = eee_timerDestroy(context->exch_timer_handle);
	VERIFY(rc == NFX_OK);
	context->exch_timer_handle = -1;

	VERIFY(!context->write);
	VERIFY(context->state == SCSIIO_CONTEXT_STATE_AWAITING_RSP);

	scsiio_transmit_ack(context, 1);
	scsiio_reset_read_context(context);
	
	return(NFX_OK);
}

void
scsiio_start_rsp_timer(scsiio_context_t *context)
{
	int32		rc;

	context->rsp_timer_handle = eee_timerCreate(context->disk_io ? 
						    SCSIIO_DISK_RSP_TIMEOUT : SCSIIO_RSP_TIMEOUT, 
						    EEE_TIMER_ONCE);
	VERIFY(context->rsp_timer_handle >= 0);
	rc = eee_timerStart(context->rsp_timer_handle, (void *)scsiio_rsp_timeout_handler,
			    context, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
}

void
scsiio_stop_rsp_timer(scsiio_context_t *context)
{
	int32		rc;

	rc = eee_timerStop(context->rsp_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(context->rsp_timer_handle);
	VERIFY(rc == NFX_OK);
	context->rsp_timer_handle = -1;
}

void
scsiio_start_ack_timer(scsiio_context_t *context)
{
	int32		rc;

	context->ack_timer_handle = eee_timerCreate(SCSIIO_ACK_TIMEOUT, EEE_TIMER_ONCE);
	VERIFY(context->ack_timer_handle >= 0);
	rc = eee_timerStart(context->ack_timer_handle, (void *)scsiio_ack_timeout_handler,
			    context, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
}

void
scsiio_stop_ack_timer(scsiio_context_t *context)
{
	int32		rc;

	rc = eee_timerStop(context->ack_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(context->ack_timer_handle);
	VERIFY(rc == NFX_OK);
	context->ack_timer_handle = -1;
}

void
scsiio_start_exch_timer(scsiio_context_t *context)
{
	int32		rc;

	context->exch_timer_handle = eee_timerCreate(SCSIIO_EXCH_TIMEOUT, EEE_TIMER_ONCE);
	VERIFY(context->exch_timer_handle >= 0);
	rc = eee_timerStart(context->exch_timer_handle, (void *)scsiio_exch_timeout_handler,
			    context, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
}

void
scsiio_stop_exch_timer(scsiio_context_t *context)
{
	int32		rc;

	rc = eee_timerStop(context->exch_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(context->exch_timer_handle);
	VERIFY(rc == NFX_OK);
	context->exch_timer_handle = -1;
}

int forever = 0;
int no_buf = 0;
uint32 lba;
uint32 datalen;
uint32 scsi_handle;
int read;

void
scsiio_cmd_done(int status, void *buf, uint32 len, void *arg1, void *arg2)
{
	int rc;

	printf("IO done: status = 0x%x, len %d\n", status, len);
	if (forever) {
		if (read) {
			rc = scsiio_read(scsi_handle, buf, datalen, scsiio_cmd_done, NULL, NULL);
			printf("READ initiated - rc - 0x%x\n", rc);
			return;
		}
		else {
			rc = scsiio_write(scsi_handle, buf, datalen, scsiio_cmd_done, NULL, NULL, 0);
			printf("WRITE initiated - rc - 0x%x\n", rc);
			return;
		}
	}
	else {
		if(no_buf) eee_freePacket((eee_desc_t *)buf);
		else eee_ramDealloc(buf);
		if(read) printf("READ complete\n");
		else printf("WRITE complete\n");
	}
}

int32
scsiio_cmd(int32 argc, char8 *argv[])
{
	int rc;
	char *buf = NULL;
	int do_disk = 0;
	int i, tmp_cnt;

	if (argc < 2) {
		printf("Usage: %s read <disk> <handle> <lba> <datalen> [no_buf] [loop]\n", argv[0]);
		printf("Usage: %s read <tape> <handle> <datalen> [no_buf] [loop]\n", argv[0]);
		printf("       %s write <disk> <handle> <lba> <datalen> [loop] \n", argv[0]);
		printf("       %s write <tape> <handle> <datalen> [loop] \n", argv[0]);
		printf("       %s stats show\n", argv[0]);
		printf("       %s stats clear\n", argv[0]);
		printf("       %s dbg_level <level>\n", argv[0]);
		return(NFX_OK);
	}

	if (strcmp(argv[1], "write") == 0)
		read = 0;
	else
	if (strcmp(argv[1], "read") == 0)
		read = 1;
	else
	if (strcmp(argv[1], "dbg_level") == 0) {
		scsiio_dbg_level = atol(argv[2]);
		return(NFX_OK);
	}
	else if(strcmp(argv[1], "stats") == 0) {
		if(strcmp(argv[2], "show") == 0) {
			printf("Cpu\tNum Reads    Num Writes\n");
			printf("---\t------------ -------------\n");
			for(i=0; i < 2; i++)
				printf("%d\t%-12d %-12d\n", i, scsiio_cb->num_reads[i], scsiio_cb->num_writes[i]);
			return(NFX_OK);
		}
		else if(strcmp(argv[2], "clear") == 0) {
			for(i=0; i < 2; i++) {
				scsiio_cb->num_reads[i] = 0;
				scsiio_cb->num_writes[i] = 0;
			}
			return(NFX_OK);
		}
		printf("Unrecognized command '%s'\n", argv[1]);
		return(NFX_OK);
	}
	else {
		printf("Unrecognized command '%s'\n", argv[1]);
		return(NFX_OK);
	}
	if(strcmp(argv[2], "disk") == 0)
		do_disk = 1;
	else if(strcmp(argv[2], "tape")) {
		printf("Unrecognized media type: %s\n", argv[2]);
		return(NFX_OK);
	}

	scsi_handle = atox(argv[3]);

	if(do_disk) {
		lba = atol(argv[4]);
		datalen = atol(argv[5]);
		if(datalen % BLOCKSIZE_512)
			datalen += (BLOCKSIZE_512 - datalen % BLOCKSIZE_512);
	}
	else
		datalen = atol(argv[4]);

	if(do_disk) tmp_cnt = 1;
	else tmp_cnt = 0;
	forever = 0;
	no_buf = 0;

	switch(argc + tmp_cnt) {
		case 8:
			if(!strcmp(argv[5+tmp_cnt], "loop")) forever = 1;
			if(!strcmp(argv[5+tmp_cnt], "no_buf")) no_buf = 1;
			else {
				printf("Unrecognized command\n");
				return(NFX_OK);
			}
			break;
		case 9:
			if(!strcmp(argv[5+tmp_cnt], "no_buf")) no_buf = 1;
			if(!strcmp(argv[6+tmp_cnt], "loop")) forever = 1;
			break;
	}
	if(forever)
		printf("Will loop forever\n");

	if(!no_buf) {
		buf = eee_ramAlloc(datalen);
		VERIFY(buf);
	}

	if (read) {
		if(do_disk)
			rc = scsiio_disk_read(scsi_handle, buf, 0, datalen, scsiio_cmd_done, NULL, NULL);
		else
			rc = scsiio_read(scsi_handle, buf, datalen, scsiio_cmd_done, NULL, NULL);
		printf("READ initiated - rc - 0x%x\n", rc);
		return(NFX_OK);
	}
	else {
		if(do_disk)
			rc = scsiio_disk_write(scsi_handle, buf, 0, datalen, scsiio_cmd_done, NULL, NULL, 0);
		else
			rc = scsiio_write(scsi_handle, buf, datalen, scsiio_cmd_done, NULL, NULL, 0);
		printf("WRITE initiated - rc - 0x%x\n", rc);
		return(NFX_OK);
	}
}

