/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: rw_api.c
 *
 * RCS:  
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Sends test read data to a device.
 *
 * Created by:	Frank Salzmann
 *
 * Date Created: July 28, 2001
 *
 *-----------------------------------------------------------------
 */
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/anpssc-api.h>
#include "sdm-api.h"
#include "sdm-db.h"
#include "sdm.h"
#include <exports/sbm-api.h>
#include "../sm-sbm/sbm.h"

#include "rw_api.h"

#ifndef SDM_MSG_ATTR_WRITE_SAME
	#define SDM_MSG_ATTR_WRITE_SAME			0x00000400L
#endif

char*	self;					// Our Name
char*	inputFile = NULL;		// Input file, if specified
char*	outputFile = NULL;		// Output file, if specified
char*	dataBuf = NULL;			// common data buffer

static void	disk_write_buf_precharge(char *buf, int buf_len, int level, u_char tag0, u_char tag1);
static int	get_num_blocks(uint64 wwn, uint32 lun);
static void	print_erase_status(int percentage);

/************** GENERAL FUNCTIONS ***************************/

void	panic(char8 *fmt, ...) {}

// General check of sendAgileMsg() returned status
static int bad_stat (int stat)
{
	if (stat < 0) {
		perror ("Send to sdm_cfgd failed");
		return 1;
	}
	else if (!stat) {
		fprintf (stderr, "Timed out waiting for response\n");
		return 1;
	}
	return 0;
}

uint32 sdm_ui_getSdmHandle( int64 wwn, int32 lun )
{
    int                       stat;
    uint32                    rc=0;
    char                	  respbuff[1024];
    sdm_getSdmHandleReqMsg_t  msg;
    sdm_getSdmHandleRspMsg_t *rsp;

	msg.hdr.mid = SDM_MID_GET_SDMHANDLE_REQ;
    msg.wwn      = wwn;
    msg.lun      = lun;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Msg Socket Create failed\n");
    }
    rsp = (sdm_getSdmHandleRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_GET_SDMHANDLE_RSP ) 
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
    else 
    if( rsp->rCode == SDM_FAIL ) 
		printf("\nDevice Not Open: 0x%016llx/%d\n\n", wwn, lun);
	else 
		rc = rsp->sdmHandle;

	return rc;
} // sdm_ui_getSdmHandle


/************** DISK READ FUNCTIONS ***************************/
int
disk_read(uint32 handle, uint32 lba, uint32 block_count, uint32 flags) 
{
	int					   dlen, rspLen;
	int 				   stat;
	int					   rc = 0;
	uint32				   dataAddr;
	uint32				   sacAppId;
	sdm_readDataReqMsg_t   msg;
	sdm_readDataRspMsg_t * rsp;

	dlen   = block_count * BLOCK_SIZE;
	rspLen = sizeof(sdm_readDataRspMsg_t) + dlen;
	if(flags & RW_FLAG_USE_BUFFER) {
		rsp	= (sdm_readDataRspMsg_t *)dataBuf;
	} 
	else {
		rsp    = (sdm_readDataRspMsg_t *)malloc( rspLen ); 
	}

	sacAppId = 12;
    dataAddr = PACK_UNICAST_PORT_ADDRESS(1,
                                         NFX_NFP_TXRX1,
                                         sacAppId,
                                         EEE_MGMT,
                                         NFX_COS_HIGH,
                                         0);
    
	memset(&msg, 0, sizeof(sdm_readDataReqMsg_t));
	msg.hdr.mid   = SDM_MID_READ_REQ;
	msg.handle    = handle;
	msg.blkOffset = lba;
	msg.blkCnt    = block_count;
	msg.blkSize   = BLOCK_SIZE;
	msg.attributes= ((flags & RW_FLAG_PHY_OP) ? SDM_MSG_ATTR_PHY_OP : 0) |
					((flags & RW_FLAG_KICK_WR_GATE) ? SDM_MSG_ATTR_KICK_WRITE_GATE : 0);
	msg.dataDest  = 0;
	
	stat = sendAgileMsg(-1, (char *)&msg, sizeof(sdm_readDataReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						(char*)rsp, rspLen);

	if( bad_stat(stat) ) {
		printf("Create failed\n");
		rc = 1;
	}

	if( rsp->hdr.mid != SDM_MID_READ_RSP ) {
		printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
		rc = 1;
	}
	else {
		if( rsp->eCode != SDM_OK ) {
			printf("DISK READ ERROR: Handle: 0x%x, eCode 0x%x\n", handle, rsp->eCode);
			rc = 1;
		}
		else {
			printf("\nDISK READ DONE: Bytes: %d, COD: %llx, LBA: %d\n\n", rsp->dataLen, rsp->codHandle, lba);
			rc = 0;
			if(flags & RW_FLAG_OUTPUT_FILE) 
				save_data((char *)rsp->data, rsp->dataLen);
			else 
				show_data((char *)rsp->data, rsp->dataLen );
		}
	}
	if(!(flags & RW_FLAG_USE_BUFFER)) free(rsp);
	return rc;
} /* disk_read */

int save_data(char *data, int len)
{
	int fd;
	mode_t mode = S_IRUSR | S_IRGRP | S_IROTH;

	if(len == 0) return 0;

	if(( fd = open(outputFile, O_WRONLY|O_CREAT, mode)) < 0) {
			fprintf(stderr, "open of %s failed: %s\n", outputFile, strerror(errno));
			return errno;
	}
	if(write(fd, data, len) != len) {
			fprintf(stderr, "write failed: %s\n", strerror(errno));
			close(fd);
			return errno;
	}	
	close(fd);
	printf("Data successfully saved to %s\n", outputFile);

	return 0;
}
	
int
show_data( char *data, int len )
{
	int i, j;
	char ch;

	for(i=0; i < len; i+=16) {
		printf("\n%04x:\t", i);
		for( j=i; j < (i+16); j++ )
			printf("%02x ", data[j] & 0xff); 
		printf("    ");
		for( j=i; j < (i+16); j++ ) {
			ch = data[j] & 0xff;
			if( (ch > 0x20) && (ch < 0x7F))
				printf("%c", ch); 
			else
			if( ch == 0 ) 
				printf("%c",'.'); 
			else
				printf("%c",'_'); 
		}

	}
	printf("\n");

	return 0;
} // show_data

typedef struct {
    uint8   enable;                     // 0 to disable automatic snapshots
    uint8   maxWeekly;                  // max weekly snapshots
    uint8   maxDaily;                   // max daily snapshots
    uint8   maxHourly;                  // max hourly snapshots
    uint32  hours;                      // bitmap of hours
                                        // bit 1-24, 24 midnight (bit 0 not used)
} read_snaptimes;

typedef struct {
    /** Snapshot version. */
    uint8 version;
    /** Snapshot being modified. */
    uint8 id;
    /** Phase-in-progress. */
    uint32 phase;
    read_snaptimes times;
} read_snapstate;

/**
 * The on-disk superblock.
 */
struct read_superblock
{
    uint32          magic;              // our magic number
    uint32          version;            // version for this filesystem
    uint32          flags;              // recovery state flags, etc.
    uint32          blkSize;            // num bytes for blks
    uint32          numRsvd;            // num blks reserved
    uint64          numBlks;            // total num blks in fs
    uint64          blksUsed;           // blocks in use
    uint64          refByteRecalcPageno;// page number of refbytes file being recounted after crash
    uint64          inodesUsed;         // inodes in use
    uint64          inodeRecalcPageno;  // page number of inode file being recounted after crash
    uint32          blockInodeRatio;    // ratio of blocks to inodes in fs
    uint32          mountTime;          // time of last mount
    uint32          mkfsTime;           // time of FS creation
    uint32          numFset;            // Number of fsets in this filesystem
    uint32          logDevType;         // int volid, ext vol, nvram, etc
                                        // default: Internal vol (reserve vol space)
    uint32          logSize;            //
    uint64     		logVolId;           // Volume id for ext vol logging
                                        //   only valid for FS_LOG_DEVTYPE_VOLID
    uint64          snapshots;          // bitmap of snapshots in use
    read_snapstate    snapState;          // in-progress snapshot state
 
    char            mountName[256];     // symbolic name fs mounted as
};

int
disk_read_super(uint32 handle, uint32 lba) 
{
	int					   dlen, rspLen;
	int 				   stat;
	int					   rc = 0;
	uint32                 block_count = 2;
	sdm_readDataReqMsg_t   msg;
	sdm_readDataRspMsg_t * rsp;
	struct read_superblock *sb;

	dlen   = block_count * BLOCK_SIZE;
	rspLen = sizeof(sdm_readDataRspMsg_t) + dlen;
	rsp    = (sdm_readDataRspMsg_t *)malloc( rspLen ); 

	memset(&msg, 0, sizeof(sdm_readDataReqMsg_t));
	msg.hdr.mid   = SDM_MID_READ_REQ;
	msg.handle    = handle;
	msg.blkOffset = lba;
	msg.blkCnt    = block_count;
	msg.blkSize   = BLOCK_SIZE;
	msg.dataDest  = 0;
	
	stat = sendAgileMsg(-1, (char *)&msg, sizeof(sdm_readDataReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						(char*)rsp, rspLen);

	if( bad_stat(stat) ) {
		printf("Create failed\n");
		rc = 1;
	}

	if( rsp->hdr.mid != SDM_MID_READ_RSP ) {
		printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
		rc = 1;
	}
	else {
		if( rsp->eCode == SDM_FAIL ) {
			printf("DISK READ ERROR: Handle: 0x%x\n", handle);
			rc = 1;
		}
		else {
			printf("\nSuper Block at LBA: %d\n\n", lba);
			rc = 0;
			sb = (struct read_superblock *)rsp->data;
			printf("magic	              : 0x%08x\n", sb->magic);
			printf("version               : 0x%08x\n", sb->version);
			printf("flags                 : 0x%08x\n", sb->flags);
			printf("blkSize               : %10d\n", sb->blkSize);
			printf("numRsvd               : %10d\n", sb->numRsvd);
			printf("numBlks               : %016lld\n", sb->numBlks);
			printf("blksUsed              : %016lld\n", sb->blksUsed);
			printf("refByteRecalcPageNo   : %016lld\n", sb->refByteRecalcPageno);
			printf("inodesUsed            : %016lld\n", sb->inodesUsed);
			printf("inodeREcalcPageNo     : %016lld\n", sb->inodeRecalcPageno);
			printf("blockInodeRatio       : %10d\n", sb->blockInodeRatio);
			printf("mountTime             : 0x%08x\n", sb->mountTime);
			printf("mkfsTime              : 0x%08x\n", sb->mkfsTime);
			printf("numFset               : %10d\n", sb->numFset);
			printf("logDevType            : 0x%08x\n", sb->logDevType);
			printf("logSize               : %10d\n", sb->logSize);
			printf("logVolId              : 0x%016llx\n", sb->logVolId);
			printf("snapshots             : 0x%016llx\n", sb->snapshots);
			printf("mountName             : %s\n",sb->mountName);
		}
	}
	free(rsp);
	return rc;
} /* disk_read_super */


/************** DISK WRITE FUNCTIONS ***************************/
int
disk_write(uint32 handle, uint32 lba, uint32 block_count, uint32 flags) 
{
	int					   i, dlen, mlen;
	int 				   stat;
	int					   rc = 0;
	char 				   respbuff[1024];
	sdm_writeDataReqMsg_t *msg;
	sdm_writeDataRspMsg_t *rsp;

	if(flags & RW_FLAG_WRITE_SAME) {
		//if(block_count < ERASE_BLK_CNT)
		//	dlen = BLOCK_SIZE * block_count;
		//else
			dlen = BLOCK_SIZE * ERASE_BLK_CNT;
	}
	else
		dlen = block_count * BLOCK_SIZE;
	mlen = sizeof(sdm_writeDataReqMsg_t) + dlen;

	if(flags & RW_FLAG_USE_BUFFER) {
		msg = (sdm_writeDataReqMsg_t *)dataBuf;
	}
	else {
		msg = (sdm_writeDataReqMsg_t *)malloc( mlen ); 
	}

	memset(msg, 0, mlen);
	msg->hdr.mid   = SDM_MID_WRITE_REQ;
	msg->handle    = handle;
	msg->blkOffset = lba;
	msg->blkCnt    = block_count;
	msg->blkSize   = BLOCK_SIZE;
	msg->attributes= SDM_MSG_ATTR_PAD_WRITE | SDM_MSG_ATTR_IGNORE_WRITE_GATE;
	msg->attributes |= ((flags & RW_FLAG_PHY_OP) ? SDM_MSG_ATTR_PHY_OP : 0);
	msg->attributes |= ((flags & RW_FLAG_KICK_WR_GATE) ? SDM_MSG_ATTR_KICK_WRITE_GATE : 0);
	msg->attributes |= ((flags & RW_FLAG_IGN_WR_GATE) ? SDM_MSG_ATTR_IGNORE_WRITE_GATE : 0);
	msg->attributes |= ((flags & RW_FLAG_WRITE_SAME) ? SDM_MSG_ATTR_WRITE_SAME : 0);
	
	if( flags & RW_FLAG_INPUT_FILE) {
		char ch;
		FILE *file;

		file = fopen(inputFile, "r");
		if( file == NULL ) {
			printf("Cannot open file: %s\n", inputFile);
			goto cleanup;
		}
		// Read contents of the file up to dlen. If contents of file are
		// less than dlen then pad out remaining space with zeros.
		i = 0;
		while( (ch = fgetc(file)) != EOF ) {
			msg->data[i] = ch;
			i++;
			if( i > dlen )
				break;
		}
		fclose(file);
		while( i < dlen ) {
			msg->data[i] = 0;
			i++;
		}
	}
	else if((flags & RW_FLAG_ERASE_DATA) && !(flags & RW_FLAG_USE_BUFFER)) {
			memset((void *)msg->data, 0, dlen);
	} 
	else {
			uint32 *		p = (uint32 *)msg->data;
			for (i=0; i < dlen; i += 4)
				*p++ = i;
	}

	stat = sendAgileMsg(-1, msg, mlen,
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));

	if( bad_stat(stat) ) {
		printf("Create failed\n");
		rc = 1;
		goto cleanup;
	}
	rsp = (sdm_writeDataRspMsg_t *)respbuff;

	if( rsp->hdr.mid != SDM_MID_WRITE_RSP ) {
		printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
		rc = 1;
	}
	else {
		if( rsp->eCode != SDM_OK ) {
			printf("DISK WRITE ERROR: Handle: 0x%x, eCode 0x%x\n", handle, rsp->eCode);
			rc = 1;
		}
		else {
			rc = 0;
		}
	}
	
cleanup:
	if(!(flags & RW_FLAG_USE_BUFFER)) free(msg);
	return rc;
} /* disk_write */

int
disk_write_from_sac(uint32 handle, uint64 cod, uint32 block_count) 
{
	int 				        stat;
	int					        rc;
	char 				        respbuff[1024];
	sdm_sacWriteReqMsg_t		msg;
	sdm_sacWriteRspMsg_t		*rsp;

	memset(&msg, 0, sizeof(sdm_sacWriteReqMsg_t));

	msg.hdr.mid     = SDM_MID_SAC_WRITE_REQ;
	msg.hdr.chandle = 0;
	msg.handle		= handle;
	msg.blkOffset   = 0;
	msg.blkCnt      = block_count;
	msg.cod			= cod;
	
	stat = sendAgileMsg(-1, &msg, sizeof(sdm_sacWriteReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));

	if( bad_stat(stat) ) {
		printf("Create failed\n");
		rc = 1;
	}
	rsp = (sdm_sacWriteRspMsg_t *)respbuff;

	if( rsp->hdr.mid != SDM_MID_WRITE_RSP ) {
		printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
		rc = 1;
	}
	else {
		if( rsp->eCode == SDM_FAIL ) {
			printf("DISK WRITE ERROR: Handle: 0x%x\n", handle);
			rc = 1;
		}
		else {
			printf("\nDISK WRITE DONE: Handle: 0x%x\n\n", handle);
			rc = 0;
		}
	}
	return rc;
} /* disk_write_from_sac */


/************** DISK WRITE VERIFY FUNCTIONS ***************************/

typedef struct _cmd_desc_ {
	int			index;
	uint			dev_handle;
	uint			dev_capacity;
	int			dev_sector_size;
	int			lba;
	int			lbc;
	sdm_writeDataReqMsg_t *	w_msg;
	int			w_msg_len;
	sdm_readDataReqMsg_t *	r_msg;
	int			r_msg_len;
	struct _cmd_desc_ *	next;
} cmd_desc_t;
cmd_desc_t *cd_head = NULL;


void
disk_write_buf_precharge(char *buf, int buf_len, int level, u_char tag0, u_char tag1)
{
	int		byte_count;
	uint32 *	p = (uint32 *)buf;

	if (level == 0) {
		for (byte_count = 0; byte_count < buf_len; byte_count += 4)
			*p++ = ((tag0 * 256 + tag1) << 16) + byte_count;
	}
	else {
		*p++ = tag0;
		*p++ = tag1;
		*p++ = tag0;
		*p++ = tag1;
	}
}

int
disk_wrv(uint handle, uint block_count, uint capacity, uint sector_size)
{
	int			i;
	cmd_desc_t *		cd;
	int			sock;
	sdm_readDataRspMsg_t *	resp;
	int			resp_len;
	int			stat;
	int			rc = 0;

	/* #define CHANDLE_HACK */

#ifdef CHANDLE_HACK
	cmd_desc_t *		only_cd;
#endif

	/* Allocate max sized resp */
	resp_len = sizeof(sdm_readDataReqMsg_t) + block_count * sector_size;
	resp = (sdm_readDataRspMsg_t *)malloc(resp_len);
					      
	/* Allocate command descriptors */
	for (i = 0; i < MAX_TAGS; ++i) {
		cd = malloc(sizeof(cmd_desc_t));
		cd->next = cd_head;
		cd_head = cd;
		cd->index = i;
		cd->dev_handle = handle;
		cd->dev_capacity = capacity;
		cd->dev_sector_size = sector_size;
		cd->lba = capacity / MAX_TAGS * cd->index;
		cd->lbc = block_count;
		cd->w_msg_len = sizeof(sdm_writeDataReqMsg_t) + cd->lbc * cd->dev_sector_size;
		cd->w_msg = (sdm_writeDataReqMsg_t *)malloc(cd->w_msg_len);
		cd->r_msg_len = sizeof(sdm_readDataReqMsg_t);
		cd->r_msg = (sdm_readDataReqMsg_t *)malloc(cd->r_msg_len);
	}

	/* Create socket for passing messages */
	sock = socket(PF_EEE, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("disk_wrv: socket failed");
		return(-1);
	}

	/* Fire off all the initial messages */
	while ((cd = cd_head)) {
		sdm_writeDataReqMsg_t *	msg;

		cd_head = cd_head->next;

#ifdef CHANDLE_HACK
		only_cd = cd;
		if (cd_head){
			printf("There can only be one .....\n");
			rc = 1;
			goto done;
		}
#endif

		bzero(cd->w_msg, cd->w_msg_len);
		/* Seed the write buffer */
		disk_write_buf_precharge(cd->w_msg->data, cd->lbc * cd->dev_sector_size, 
					 0, cd->dev_handle, cd->index);

		msg = cd->w_msg;
		msg->hdr.mid = SDM_MID_WRITE_REQ;
		msg->hdr.chandle = (uint32)cd;
		msg->handle = cd->dev_handle;
		msg->blkOffset = cd->lba;
		msg->blkCnt = cd->lbc;
		msg->blkSize = cd->dev_sector_size;
		msg->attributes = 0;

		printf("Writing LBA %d, LBC %d (chandle 0x%x)\n", 
		       msg->blkOffset, msg->blkCnt, msg->hdr.chandle);

		stat = sendAgileMsg(sock, cd->w_msg, cd->w_msg_len, "sdm", 0, 0, 0, NULL, NULL, 0);
		if (bad_stat(stat)) {
			printf("Message send failed\n");
			rc = 1;
			goto done;
		}
	}

	/* Now wait for responses and run state machine for each instance */
	while (1) {
		stat = recvAgileMsg (sock, resp, resp_len, NULL, NULL, 0, 0, 0);
		if (bad_stat(stat)) {
			printf("Message receive failed\n");
			rc = 1;
			goto done;
		}

		switch (resp->hdr.mid) {
		case SDM_MID_WRITE_RSP:
#ifndef CHANDLE_HACK
			cd = (cmd_desc_t *)resp->hdr.chandle;
#else
			cd = only_cd;
#endif
			if (resp->eCode == SDM_FAIL) {
				printf("write error\n");
				rc = 1;
				goto done;
			}

			/* Following completion of a write, Fire off a read */
			cd->r_msg->hdr.mid = SDM_MID_READ_REQ;
			cd->r_msg->hdr.chandle = (uint32)cd;
			cd->r_msg->handle = cd->dev_handle;
			cd->r_msg->blkOffset = cd->lba;
			cd->r_msg->blkCnt = cd->lbc;
			cd->r_msg->blkSize = cd->dev_sector_size;
			cd->r_msg->attributes = 0;

			printf("Reading LBA %d, LBC %d (chandle 0x%x)\n", 
			       cd->r_msg->blkOffset, cd->r_msg->blkCnt, cd->r_msg->hdr.chandle); 
			
			stat = sendAgileMsg(sock, cd->r_msg, cd->r_msg_len, "sdm", 0, 0, 0, NULL, NULL, 0);
			if (bad_stat(stat)) {
				printf("Message send failed\n");
				rc = 1;
				goto done;
			}
			break;

		case SDM_MID_READ_RSP:
#ifndef CHANDLE_HACK
			cd = (cmd_desc_t *)resp->hdr.chandle;
#else
			cd = only_cd;
#endif
			if (resp->eCode == SDM_FAIL) {
				printf("read error\n");
				rc = 1;
				goto done;
			}
			if (resp->dataLen != cd->r_msg->blkCnt * cd->r_msg->blkSize) {
				printf("Read data len invalid.\n");
				rc = 1;
				goto done;
			}
			
			/* compare read with write buffers */
			if (memcmp(cd->w_msg->data, resp->data, resp->dataLen)) {
				uint32 *p1 = (void *)cd->w_msg->data;
				uint32 *p2 = (void *)resp->data;
				int32  len = resp->dataLen;
					
				printf("Data miscompare: cd %p, dev_handle 0x%x, "
				       "write buf %p, "
				       "read buf %p.\n",
				       cd, cd->dev_handle,
				       cd->w_msg->data, resp->data);
				while (len) {
					if (*p1 != *p2) {
						printf("\t %p[offset 0x%04x] (0x%08x) != %p[offset 0x%04x] (0x%08x)\n",
						       p1, (uint32)p1 - (uint32)cd->w_msg->data, *p1, 
						       p2, (uint32)p2 - (uint32)resp->data, *p2);
					}
					++p1;
					++p2;
					len -= 4;
				}
				printf("\n\n");
			}

			/* Set new LBA */
			cd->lba += cd->lbc;
			if (cd->lba > ((cd->index + 1) * (cd->dev_capacity / MAX_TAGS)) - cd->lbc)
				cd->lba = cd->index * cd->dev_capacity / MAX_TAGS;

			/* Seed the write buffer */
			disk_write_buf_precharge(cd->w_msg->data, cd->lbc * cd->dev_sector_size, 
						 1, cd->lba, cd->lbc);
			
			/* .... and fire off the next write */
			cd->w_msg->hdr.mid = SDM_MID_WRITE_REQ;
			cd->w_msg->hdr.chandle = (uint32)cd;
			cd->w_msg->handle = cd->dev_handle;
			cd->w_msg->blkOffset = cd->lba;
			cd->w_msg->blkCnt = cd->lbc;
			cd->w_msg->blkSize = cd->dev_sector_size;
			cd->w_msg->attributes = 0;

			printf("Writing LBA %d, LBC %d, (chandle 0x%x)\n", 
			       cd->w_msg->blkOffset, cd->w_msg->blkCnt, cd->w_msg->hdr.chandle);
			
			stat = sendAgileMsg(sock, cd->w_msg, cd->w_msg_len, "sdm", 0, 0, 0, NULL, NULL, 0);
			if (bad_stat(stat)) {
				printf("Message send failed\n");
				rc = 1;
				goto done;
			}
			break;

		default:
			printf("Invalid response code: 0x%x\n", resp->hdr.mid);
			rc = 1;
			goto done;
		}
		if (resp->hdr.mid == SDM_MID_WRITE_RSP) {

		}
	}

 done:
 	close(sock);
	return(rc);
}


/******************** DISK COPY FUNCTIONS ***********************/
int
disk_copy(uint32 source_handle, uint32 source_lba, uint32 target_handle, uint32 target_lba, uint32 block_count)
{
	int		rc = 0;
	int		dlen, mlen;
	int		i;
	int		rw_blk_cnt;
	uint32	flags = RW_FLAG_USE_BUFFER;

	/* get size of buffer required */
	dlen = COPY_BLK_CNT * BLOCK_SIZE;
	if(sizeof(sdm_writeDataReqMsg_t) > sizeof(sdm_readDataRspMsg_t))
		mlen = sizeof(sdm_writeDataReqMsg_t) + dlen;
	else
		mlen = sizeof(sdm_readDataRspMsg_t) + dlen;

	/* allocate buffer */
	dataBuf = (char *)malloc(mlen);
	if(!dataBuf) {
			fprintf(stderr, "disk_copy: malloc failed: %s\n", strerror(errno));
			return errno;
	}
	
	/* do read and writes */
	for(i=0; i < block_count; i += COPY_BLK_CNT) {
		if(i+COPY_BLK_CNT > block_count) rw_blk_cnt = block_count - i;
		else rw_blk_cnt = COPY_BLK_CNT;
		
		rc = disk_read(source_handle, source_lba, rw_blk_cnt, flags);
		if(rc) break;
		rc = disk_write(target_handle, target_lba, rw_blk_cnt, flags);
		if(rc) break;
	}

	free(dataBuf);
	return rc;
}

/*********************** DISK ERASE FUNCTIONS ***********************/

// prints percentage erased
void print_erase_status(int percentage)
{
	fprintf(stdout, "\b\b\b\b\b");
	fprintf(stdout, "%3d%% ", percentage);
	fflush(stdout);
}

// returns the number of blocks for a wwn/lun
int get_num_blocks(uint64 wwn, uint32 lun)
{
	sdm_devInfoReqMsg_t	reqMsg;
	sdm_devInfoRspMsg_t	rspMsg;
	int					rc;

	reqMsg.hdr.mid = SDM_MID_DEV_INFO_REQ;

	/* create a scsi id */
	reqMsg.id = 0;
	//SCSI_ID_SET_WWN(reqMsg.id, wwn);
	//SCSI_ID_SET_LUN(reqMsg.id, lun);

	bzero(&rspMsg, sizeof(sdm_devInfoRspMsg_t));

	/* send req. to SDM and get response back */
	rc = sendAgileMsg(-1, 
				   	  &reqMsg, sizeof(sdm_devInfoReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &rspMsg, sizeof(sdm_devInfoRspMsg_t));

	if (rc < 0) {
		fprintf(stderr, "Failed to send msg. to SDM\n");
		return 0;
	}
	if (rspMsg.hdr.mid != SDM_MID_DEV_INFO_RSP ) {
		fprintf(stderr, "Invalid Response Received from SDM, rsp mid : %x\n",
					     rspMsg.hdr.mid);
		return 0;
	}
	if(rspMsg.rc == SDM_FAIL) {
		fprintf(stderr, "SDM Fail error: Device %lld:%d not found\n", wwn, lun);
		return 0;
	}

	return rspMsg.blockCnt;
}

int
disk_erase(uint64 wwn, uint32 lun, uint32 handle, uint32 lba, uint32 block_count, uint32 flags)
{
	int		rc = 0;
	int		mlen;
	int		eraseBlkCnt;
	int		blksPerErase = 0;	/* how many blocks per erase op. */
	float	percentPerErase = 0;
	float	percentDone = 0;
	int		end_block;
	int		total_blocks = 0;

	/* get number of blocks for this LUN */
	total_blocks = get_num_blocks(wwn, lun);
	if(!total_blocks) {
		fprintf(stderr, "Failed to get blk. count for %lld:%d\n", wwn, lun);
		return 1;
	}

	/* calculate last block to be written */
	if(block_count == 0) {
		end_block = total_blocks - 1;
		block_count = total_blocks - lba;
	}
	else {
		if(block_count < ERASE_BLK_CNT) {
			int padding = 16 - (block_count % 16);
			block_count += (padding % 16);
		}
		end_block = lba + block_count - 1;
		if(end_block >= total_blocks) {
			end_block = total_blocks - 1;
			block_count = total_blocks - lba;
		}
	}

	/* determine number of blocks to delete per iteration */
	while(percentPerErase < 1 ) {
		blksPerErase += 48000;
		percentPerErase = (float)blksPerErase * 100 / (float)block_count;
	}
	if(percentPerErase > 100) percentPerErase = 100;

	/* do the erase */
	/* use WRITE_SAME if write-same is supported, else simply use WRITE */
	flags |= RW_FLAG_WRITE_SAME;
	//if(block_count < ERASE_BLK_CNT) 
	//	mlen = sizeof(sdm_writeDataReqMsg_t) + BLOCK_SIZE * block_count;
	//else
		mlen = sizeof(sdm_writeDataReqMsg_t) + BLOCK_SIZE * ERASE_BLK_CNT;
	dataBuf = (char *)malloc(mlen);
	if(!dataBuf) {
		fprintf(stderr, "disk_erase: malloc failed: %s\n", strerror(errno));
		return errno;
	}
	memset((void *)dataBuf, 0, mlen);
	flags |= RW_FLAG_USE_BUFFER;

	do {
		if((lba+blksPerErase) > end_block) eraseBlkCnt = end_block - lba + 1;
		else eraseBlkCnt = blksPerErase;
		rc = disk_write(handle, lba, eraseBlkCnt, flags);
		if(rc) break;
		
		lba += eraseBlkCnt;
		percentDone += percentPerErase;
		if(percentDone > 100) percentDone = 100;
		print_erase_status((int)percentDone);
	} while(percentDone < 100 && lba <= end_block);

	/*do {
		if((lba+blksPerErase) > end_block) eraseBlkCnt = end_block - lba + 1;
		else eraseBlkCnt = blksPerErase;
		eraseBlkCnt = eraseBlkCnt - (lba % 16);		// align to 16 block boundary

		if(flags & RW_FLAG_WRITE_SAME) {
			if(block_count < ERASE_BLK_CNT) block_count = ERASE_BLK_CNT;
			rc = disk_write(handle, lba, block_count, flags);
			break;
		}
		else
			rc = disk_write(handle, lba, eraseBlkCnt, flags);
		if(rc) { 
			if(flags & RW_FLAG_USE_BUFFER) break;

			else {
				printf("Write Same Transaction Failed: error %d\n", rc);
				printf("Using sequential writes now\n");

				flags &= ~RW_FLAG_WRITE_SAME;
				blksPerErase = ERASE_BLK_CNT;
				percentPerErase = (float)blksPerErase * 100 / (float)block_count;

				mlen = sizeof(sdm_writeDataReqMsg_t) + BLOCK_SIZE*ERASE_BLK_CNT;
				dataBuf = (char *)malloc(mlen);
				if(!dataBuf) {
					fprintf(stderr, "disk_erase: malloc failed: %s\n", strerror(errno));
					return errno;
				}
				memset((void *)dataBuf, 0, mlen);
				flags |= RW_FLAG_USE_BUFFER;
				continue;
			}
		}

		lba += eraseBlkCnt;
		printf("lba=%d\n", lba);
		percentDone += percentPerErase;
		if(percentDone > 100) percentDone = 100;
		print_erase_status((int)percentDone);
	} while(percentDone < 100 && lba <= end_block);
	printf("\n");
	*/

	if(flags & RW_FLAG_USE_BUFFER) free(dataBuf);
	return rc;
}
