//vi:ts=4
//------------------------------------------------------------------------------
//  FileName   : sdm-init.c
//
//  Description: 
//
//	Created    : 06 March 2002
//
//  Author     : Frank Salzmann
//
//  RCS        : $Id: sdm-init.c,v 1.34 2002/11/13 23:36:55 nitinb Exp $
//
//  Copyright (c)  Agile Storage, Inc. 2002
//------------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#undef NULL

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>  
#include <exports/scsi-api.h>
#include <exports/elog-api.h>
//#include "../sm-scsi-test/scsi-test.h"

#include "sdm-api.h"
#include "sdm-db.h"
#include "sdm.h"
#include "sdm-init.h"
#include "sdm-conf.h"
#include "sdm-msg.h"
#include "../sm-scsi/scsi-dev-cmds.h"

//
//-- External Declarations
//
extern sdm_devSupport_t sdm_diskList[];
extern sdm_devSupport_t sdm_tapeList[];
extern sdm_devSupport_t sdm_changerList[];

static int 	sdm_initParseInqString(char *, int, char *, int);
static void	sdm_setDevrec(sdm_dir_t *, sdm_devDBRec_t *);

//------------------------------------------------------------------------------
// Function Name:   sdm_initDevSM
// 
// Description:     Will call the next action fuction based on the 
//					event sent, and the current state.
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_initDevSM( int         event, 
			   sdm_dir_t  *dir, 
			   void       *param)
{
	int i=0;
	int state;
	sdm_dist_t *dist = dir->dist;

	//printf("sdm_initDevSM\n");
	while( (state = dist[i].state) != 0 ) {
		if( state == dir->state ) {
			if( dist[i].event == event ) {
				if( dist[i].func ) {
					dir->nextEvt = dist[i].nextEvt;
					dist[i].func(dir, param);
				}
				else {
					printf("sdm_initDevSM: Invalid Event: %d\n", event);
				}
				return(SDM_OK);
			}
		}
		i++;
	}
	return SDM_FAIL;  //-- Should never get here.
} // sdm_initDevSM


//------------------------------------------------------------------------------
// Function Name:   sdm_getInitST
// 
// Description:     Will return device init state table.
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
sdm_dist_t *
sdm_getInitST( sdm_devSupport_t *dsup, char *vendor, char *productId, char *revision)
{
	int i;
	int len;

	//printf("sdm_getInitST\n");
	i=0;
	while( dsup[i].vendor != 0 ) {
		len = strlen(dsup[i].vendor);
		//-- Look for generic vendor support.
		if( !strncmp("*", dsup[i].vendor, len) ) {
				return dsup[i].dist;
		}
		if( !strncmp(vendor, dsup[i].vendor, len) ) {
			len = strlen(dsup[i].productId);
			//-- Look for generic productId support.
			if( !strncmp("*", dsup[i].productId, len) ) {
				return dsup[i].dist;
			}
			if( !strncmp(productId, dsup[i].productId, len) ) {
				len = strlen(dsup[i].revision);
				//-- Look for generic revision level
				if( !strncmp("*", dsup[i].revision, len)) {
					return dsup[i].dist;
				}
				if(!strncmp(revision, dsup[i].revision, len)) {
					return dsup[i].dist;
				}
			}
		}
		i++;
	}
	return (sdm_dist_t *)NULL;
} // sdm_getInitST

//------------------------------------------------------------------------------
// Function Name:   sdm_setInitST
// 
// Description:     Gets and then sets pointer to the device init statetable.
//					Based on vendor/productId and device type.
//
// Data Created:    April 10, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_setInitST( sdm_dir_t *dir )
{
	sdm_dist_t *		dist;
	int					i;

	//-- Get initialization state table.
	switch(dir->devType) {

		case SCSI_DEVTYPE_DISK:
			dist = sdm_getInitST(sdm_diskList, dir->st.vendor, dir->st.productId, dir->st.revision);
			break;

		case SCSI_DEVTYPE_TAPE:
			dist = sdm_getInitST(sdm_tapeList, dir->st.vendor, dir->st.productId, "*");
			
			if (dir->lun == 0) {
				/* 
				 * Probe for possibly CrossRoads attached device at LUNs > 0
				 */
				for (i = 1; i < SDM_DIE_MAX_LUN_PROBE; ++i)
					sdm_initNewLun(dir->wwn, i, dir->path, dir->llHandle);
			}
			break;

		case SCSI_DEVTYPE_MEDIA_CHANGER:
			dist = sdm_getInitST(sdm_changerList, dir->st.vendor, dir->st.productId, "*");
			break;

		//-- Device Types Not Supported 
		case SCSI_DEVTYPE_CDROM:
		case SCSI_DEVTYPE_WORM:
		case SCSI_DEVTYPE_OPTICAL_DISK:
		case SCSI_DEVTYPE_RAID:
		case SCSI_DEVTYPE_PRINTER:
		case SCSI_DEVTYPE_PROCESSOR:
		case SCSI_DEVTYPE_COMM_DEVICE:
		case SCSI_DEVTYPE_SES:
		case SCSI_DEVTYPE_UNKNOWN:
		default:	
			return SDM_FAIL;
	}

	dir->dist = dist;
	if( !dist ) {
		printf("Device Not Recognized: %s : %s\n", dir->st.vendor, dir->st.productId);
		return SDM_FAIL; 
	}

	return SDM_OK;
} // sdm_setInitST

//------------------------------------------------------------------------------
// Function Name:   sdm_initGetScsiId
// 
// Description:     Gets a SCSI_ID_T based on WWN, LUN and vendor information.
//
// Data Created:    March 16, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
scsi_id_t
sdm_initGetScsiId(uint64 wwn, uint32 lun, char *vendor) 
{
	scsi_id_t id = 0;

	if( strncmp(vendor, "HITACHI", 7) == 0 ) {
		wwn &= 0xFFFFFFFFFFFFFF00ll;
		SDM_INIT_ID_SET_WWN(id, wwn);
		SDM_INIT_ID_SET_LUN(id, lun);
	}
	else
	if( strncmp(vendor, "LSI", 3) == 0 ) {
		uint64 tmp;
		SDM_INIT_ID_SET_WWN(id, wwn);
		id &= 0x00000000FFFFFFFFll;
		tmp = lun;
		id |= (((tmp) << 32) & 0xFFFFFFFF00000000ll); 
	}
	else {
		SDM_INIT_ID_SET_WWN(id, wwn);
		SDM_INIT_ID_SET_LUN(id, lun);
	}
	return id;
} // sdm_initGetScsiId


//------------------------------------------------------------------------------
// Function Name:   sdm_initDevProcMsg
// 
// Description:     Interface for messages coming into the SDM reguarding
//					device initialization steps.
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_initDevProcMsg( sdm_dir_t *dir, void *msg ) 
{
	int rc;
	//printf("sdm_initDevProcMsg\n");

	if( dir->nextEvt == SDM_DIE_INIT_LUN ) {
		//-- Since we don't know yet what this device is exactly
		//-- we cannot call initDevSM because the dir->dist has
		//-- not yet been setup.
		rc = sdm_procInqRsp(dir, ((scsi_relayRspMsg_t *)msg)->data);
		if( rc == SDM_FAIL ) {
			if(dir) free(dir);
			return rc;	
		}
	}
	else
		sdm_initDevSM(dir->nextEvt, dir, msg );

	return SDM_OK;
} // sdm_initDevProcMsg

//------------------------------------------------------------------------------
// Function Name:   sdm_procNewDevMsg
//
// Description:     Handles new device discovered event as reported
//					by the SCSI application running on the SP_NIM card.   
//
// Created by:      Frank Salzmann
//
// Data Created:    March 07, 2002
//------------------------------------------------------------------------------
int
sdm_procNewDevMsg(sdm_msgDesc_t *mdesc)
{
	uint32           i, rc;
	sdm_dir_t       *dir;
	sdm_newDevMsg_t *i_msg;
	uint32           port, slot;

	i_msg = (sdm_newDevMsg_t *)mdesc->msg;

	port = SCSI_PATH_GET_PORT(i_msg->path);
	slot = SCSI_PATH_GET_PORT(i_msg->path);

	E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
		   ("SDM_INIT: New Device Path Reported By SCSI: 0x%016llx:0, port: %d", i_msg->wwn, port));

	
	// printf("sdm_procNewDevMsg: WWN 0x%llx\n, i_msg->wwn");

	dir = (sdm_dir_t *)malloc(sizeof(sdm_dir_t));
	if( !dir )
		return SDM_FAIL;

	memset(dir, 0, sizeof(sdm_dir_t));
	dir->id       = i_msg->id;
	dir->path     = i_msg->path;
	dir->wwn      = i_msg->wwn;
	dir->chn_wwn  = i_msg->chn_wwn;
	dir->llHandle = i_msg->llHandle;
	dir->lun      = 0;
	dir->state    = SDM_DIS_INIT;
	dir->nextEvt  = SDM_DIE_INIT_DEV;

	rc = sdm_procInqRsp( dir, i_msg->inqRsp);

	if( rc == SDM_FAIL ) {

		//-- LUN 0 is not a valid device. But there could still be other luns
		//-- attached to this WWN which are valid. So we need to go look for
		//-- them.
		
		for(i=1; i < SDM_DIE_MAX_LUN_PROBE; i++ ) {
			sdm_initNewLun(dir->wwn,i,dir->path, dir->llHandle);
		}

		//-- Device is not supported. Send msg to SCSI application 
		//-- to remove the device record.
		E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
			   ("SDM: Device Not Supported:  %s, 0x%016llx:0", 
				dir->st.productId, dir->wwn));
		//--sdm_sendScsiDelMsg(i_msg->path);
		free(dir);
		return rc;
	}

	return SDM_OK;
} // sdm_procNewDevMsg

//------------------------------------------------------------------------------
// Function Name:   sdm_procScsiAddRsp
//
// Description:     Handles SCSI response to the ADD_MSG.
//
// Created by:      Frank Salzmann
//
// Data Created:    March 20, 2002
//------------------------------------------------------------------------------
int
sdm_procScsiAddRsp(sdm_msgDesc_t *mdesc)
{
	uint32           rc;
	uint32           port;
	sdm_dir_t       *dir;
	scsi_addRspMsg_t *i_msg;

	//printf("sdm_procScsiAddRsp\n");
	i_msg = (scsi_addRspMsg_t *)mdesc->msg;

	dir = (sdm_dir_t *)i_msg->hdr.cHandle;

	//-- Overwrite with new path assigned by SCSI.
	dir->path = i_msg->hdr.path;

	if( i_msg->eCode != SCSI_MSG_ECODE_OK ) {
		//-- Device is not supported. Send msg to SCSI application 
		//-- to remove the device record.
		//printf("SCSI Add Failed\n");
		E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
			   ("SDM: SCSI Add Failed:  0x%016llx:%d", 
				dir->wwn, dir->lun));
		free(dir);
		return rc;
	}

	port = SCSI_PATH_GET_PORT(dir->path);
	E_LOG (class_1, debug_s, eee_getApplicationId("sdm"), 0, 0, 0,
		   ("SDM_INIT: SCSI Dev Add Rsp, Ok: 0x%016llx:0, port: %x", dir->wwn, port));

	//-- Call SCSI application to add device.
	//-- We will then create the SDM data base record
	//-- when a successful reply is received from the SCSI app.
	sdm_initCreateDevrec(dir);

	return SDM_OK;
} // sdm_procScsiAddRsp

//------------------------------------------------------------------------------
// Function Name:   sdm_initNewLun
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
sdm_dir_t *
sdm_initNewLun(uint64 wwn, uint32 lun, uint32 path, uint32 llHandle)
{
	sdm_dir_t *dir;

	dir = (sdm_dir_t *)malloc(sizeof(sdm_dir_t));
	if( !dir )
		return (sdm_dir_t *)NULL;

	memset(dir, 0, sizeof(sdm_dir_t));
	dir->id      = 0;
	dir->path    = path;
	dir->wwn     = wwn;
	dir->lun     = lun;
	dir->llHandle= llHandle;
	dir->state   = SDM_DIS_INIT;
	dir->nextEvt = SDM_DIE_INIT_LUN;

	//-- Send SCSI inquiry command to the lun. Because at this
	//-- point we don't know what it is.
	sdm_sendInqCmd(dir);

	return dir;
} // sdm_initNewLun


//------------------------------------------------------------------------------
// Function Name:   sdm_procInqRsp
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_procInqRsp(sdm_dir_t *dir, char *inqRsp) 
{
	uint32      rc;

	//-- Parse the SCSI device inquiry response.
	rc = sdm_parseInqRsp( dir->st.vendor, 
						  dir->st.productId, 
						  dir->st.model, 
						  dir->st.revision,
						  &dir->devType, 
						  inqRsp );

	if( rc == SDM_FAIL )
		return rc;

	//-- Get device initialization state table.
	rc = sdm_setInitST( dir );

	if( rc == SDM_FAIL )
		return rc;
	else 		
		sdm_initDevSM(dir->nextEvt, dir, inqRsp); //-- Initialize device.

	return SDM_OK;
} // sdm_procInqRsp


//------------------------------------------------------------------------------
// Function Name:   sdm_parseInqRsp
// 
// Description:     Will parse the response to the
//                  SCSI INQUERY command returned by
//                  the SCSI target device.
//
// Data Created:    March 15, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_parseInqRsp( uchar8 *vendor, 
				 uchar8 *productId, 
				 uchar8 *model, 
				 uchar8 *revision,
				 uint32 *devType, 
				 char *rspBuf )
{
    uchar8 *mfgStr, *productIdStr, *revisionStr;
	
	//printf("sdm_parseInqRsp\n");
 
    if( (rspBuf[0] >> 5) == 3 )     // No Dev supported on this LUN 
        return SDM_FAIL;
 
	//-- Extract the vendor string
    mfgStr   = &rspBuf[8];
	sdm_initParseInqString(mfgStr, 8, vendor, 9);

	//-- Extract the product Id string
    productIdStr = &rspBuf[16];
	sdm_initParseInqString(productIdStr, 16, productId, 17);

	//-- For now set the model string equal to the product id.
	memcpy(model, productId, 17);

	//printf("1. vendor: %s, productId: %s\n", vendor, productId);
    *devType = rspBuf[0] & 0x1F;
 
	revisionStr = &rspBuf[32];
	sdm_initParseInqString(revisionStr, 4, revision, 4);
	revision[4] = 0;

    return SDM_OK;
} // sdm_parseInqRsp 


//------------------------------------------------------------------------------
// Function Name:   sdm_sendInqCmd
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_sendInqCmd( sdm_dir_t *dir ) 
{
	uint32 mlen, destCpu, destSlot, tid; 
	scsi_relayReqMsg_t *scsi_o_msg;

	//printf("sdm_sendInqCmd\n");

	//-- Allocate an SDM transaction Id
	tid = sdm_allocRelayTransaction( NULL,
									 NULL,
									 sdm_initDevProcMsg,
									 (int)dir);

	//-- Allocate memory for the message.
	mlen = (sizeof(scsi_relayReqMsg_t) - 1 + 16);
	scsi_o_msg = (scsi_relayReqMsg_t *)malloc(mlen);
	memset(scsi_o_msg, 0, mlen);

	//-- Setup the Report Lun CDB.
	scsi_o_msg->cdb[0] = 0x12;  // SCSI_CMD_INQUIRY
    scsi_o_msg->cdb[4] = 255;

	//-- Setup the SCSI msg header.
	scsi_o_msg->hdr.mid     = SCSI_MID_RELAY_REQ;
	scsi_o_msg->hdr.cHandle = tid;
	scsi_o_msg->hdr.path    = dir->path;
	scsi_o_msg->lun         = dir->lun;
	scsi_o_msg->dataLen     = 0;

	//-- Send read command and data to the SCSI application.
	destCpu = SCSI_PATH_GET_PORT(dir->path);
	destSlot = SCSI_PATH_GET_SLOT(dir->path);
	sdm_sendMsg( (char *)scsi_o_msg, mlen, "scsi", destCpu, destSlot, 0, 0,0);
	free(scsi_o_msg);

	return SDM_OK;
} // sdm_sendInqCmd

//------------------------------------------------------------------------------
// Function Name:   sdm_sendSnReq
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_sendSnReq( sdm_dir_t *dir, void *notused ) 
{
	uint32 mlen, destCpu, destSlot, tid; 
	scsi_relayReqMsg_t *scsi_o_msg;

	//printf("sdm_sendSnReq\n");

	//-- Allocate an SDM transaction Id
	tid = sdm_allocRelayTransaction( NULL,
									 NULL,
									 sdm_initDevProcMsg,
									 (int)dir);

	//-- Allocate memory for the message.
	mlen = (sizeof(scsi_relayReqMsg_t) - 1 + 16);
	scsi_o_msg = (scsi_relayReqMsg_t *)malloc(mlen);
	memset(scsi_o_msg, 0, mlen);

	//-- Setup the Report Lun CDB.
	scsi_o_msg->cdb[0] = 0x12;  // SCSI_CMD_INQUIRY
	scsi_o_msg->cdb[1] = 0x1;	// Set EVPD Bit.
	scsi_o_msg->cdb[2] = 0x80;	// Request Unit Serial Number Page
    scsi_o_msg->cdb[4] = 255;

	//-- Setup the SCSI msg header.
	scsi_o_msg->hdr.mid     = SCSI_MID_RELAY_REQ;
	scsi_o_msg->hdr.cHandle = tid;
	scsi_o_msg->hdr.path    = dir->path;
	scsi_o_msg->lun         = dir->lun;
	scsi_o_msg->dataLen     = 0;

	//-- Send read command and data to the SCSI application.
	destCpu = SCSI_PATH_GET_PORT(dir->path);
	destSlot = SCSI_PATH_GET_SLOT(dir->path);
	sdm_sendMsg( (char *)scsi_o_msg, mlen, "scsi", destCpu, destSlot, 0, 0,0);
	free(scsi_o_msg);

	E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
			("SDM_INIT: Get Serial Num:   0x%016llx:%d", 
			 dir->wwn, dir->lun));
	return SDM_OK;
} // sdm_sendSnReq


//------------------------------------------------------------------------------
// Function Name:   sdm_sendRLunCmd
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_sendRLunCmd(sdm_dir_t *dir, void *notused) 
{
	uint32 mlen, allocLen, destCpu, destSlot, tid; 
	scsi_relayReqMsg_t *scsi_o_msg;

	//printf("sdm_sendRLunCmd\n");

	//-- Allocate an SDM transaction Id
	tid = sdm_allocRelayTransaction( NULL,
									 NULL,
									 sdm_initDevProcMsg,
									 (int)dir);

	//-- Allocate memory for the message.
	mlen = (sizeof(scsi_relayReqMsg_t) - 1 + 16);
	scsi_o_msg = (scsi_relayReqMsg_t *)malloc(mlen);
	memset(scsi_o_msg, 0, mlen);

	//-- Setup the Report Lun CDB.
	scsi_o_msg->cdb[0] = 0xA0;  // SCSI_CMD_REPORT_LUNS
	allocLen   = (8 * 256 + 4);		// Look for max 256 luns.
    scsi_o_msg->cdb[6] = (allocLen >> 24) & 0xFF;       /* MSB */
    scsi_o_msg->cdb[7] = (allocLen >> 16) & 0xFF;
    scsi_o_msg->cdb[8] = (allocLen >> 8) & 0xFF;
    scsi_o_msg->cdb[9] = allocLen & 0xFF;               /* LSB */

	//-- Setup the SCSI msg header.
	scsi_o_msg->hdr.mid     = SCSI_MID_RELAY_REQ;
	scsi_o_msg->hdr.cHandle = tid;
	scsi_o_msg->hdr.path    = dir->path;
	scsi_o_msg->lun         = dir->lun;
	scsi_o_msg->dataLen     = 0;

	//-- Send read command and data to the SCSI application.
	destCpu = SCSI_PATH_GET_PORT(dir->path);
	destSlot = SCSI_PATH_GET_SLOT(dir->path);
	sdm_sendMsg( (char *)scsi_o_msg, mlen, "scsi", destCpu, destSlot, 0, 0,0);
	free(scsi_o_msg);

	return SDM_OK;
} // sdm_sendRLunCmd

//------------------------------------------------------------------------------
// Function Name:   sdm_sendRCapCmd
// 
// Description:     
//
// Data Created:    March 19, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_sendRCapCmd(sdm_dir_t *dir, void *notused) 
{
	uint32 mlen, destCpu, destSlot, tid; 
	scsi_relayReqMsg_t *scsi_o_msg;

	//printf("sdm_sendRCapCmd\n");

	//-- Allocate an SDM transaction Id
	tid = sdm_allocRelayTransaction( NULL,
									 NULL,
									 sdm_initDevProcMsg,
									 (int)dir);

	//-- Allocate memory for the message.
	mlen = (sizeof(scsi_relayReqMsg_t) - 1 + 16);
	scsi_o_msg = (scsi_relayReqMsg_t *)malloc(mlen);
	memset(scsi_o_msg, 0, mlen);

	//-- Setup the Report Capacity CDB.
	scsi_o_msg->cdb[0] = 0x25;  // SCSI_CMD_REPORT_CAPACITY

	//-- Setup the SCSI msg header.
	scsi_o_msg->hdr.mid     = SCSI_MID_RELAY_REQ;
	scsi_o_msg->hdr.cHandle = tid;
	scsi_o_msg->hdr.path    = dir->path;
	scsi_o_msg->lun         = dir->lun;
	scsi_o_msg->dataLen     = 0;

	//-- Send read command and data to the SCSI application.
	destCpu = SCSI_PATH_GET_PORT(dir->path);
	destSlot = SCSI_PATH_GET_SLOT(dir->path);
	sdm_sendMsg( (char *)scsi_o_msg, mlen, "scsi", destCpu, destSlot, 0, 0,0);
	free(scsi_o_msg);

	return SDM_OK;
} // sdm_sendRCapCmd

//------------------------------------------------------------------------------
// Function Name:   sdm_setDevrec
// 
// Description:     Fills in a devrec structure with information from verify.
//
// Data Created:    Sept 26, 2002
//
//------------------------------------------------------------------------------
void
sdm_setDevrec(sdm_dir_t *dir, sdm_devDBRec_t *devRec)
{
	devRec->id      = dir->id;
	devRec->wwn     = dir->wwn; 
	devRec->devType = dir->devType;
	devRec->lun     = dir->lun;

	if( devRec->devType == SCSI_DEVTYPE_DISK ) {
		devRec->lbCnt     = dir->logicalBlockCnt;
		devRec->lbSize 	  = dir->logicalBlockSize;
		devRec->raidLevel = dir->raidLevel;
	}
	else { // TAPE
		devRec->lbCnt = 0;
		devRec->lbSize = 0;
	}

	memcpy((void *)&devRec->st, &dir->st, sizeof(sdm_dev_strings_t));
}



//------------------------------------------------------------------------------
// Function Name:   sdm_initCreateDevrec
// 
// Description:     Create a SCSI_ID_T based on WWN, LUN and vendor information.
//
// Data Created:    March 16, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
sdm_devDBRec_t *
sdm_initCreateDevrec(sdm_dir_t *dir)  
{
	uint32          i;
	uint32          pathfound;
	sdm_devDBRec_t *devRec;

	//-- Make sure device/path is allowed
	if( sdm_confVerify(dir->wwn, dir->lun, dir->path, dir->st.name) < 0 ) {
		//printf("SDM_MSG: Device 0x%016llx masked on slot %d, port %d.\n", 
		//	   id, SCSI_PATH_GET_SLOT(path), SCSI_PATH_GET_PORT(path));
		return (sdm_devDBRec_t *)NULL;
	}

	//-- Look for device in database.
	devRec = (sdm_devDBRec_t *)sdm_dbFindRec(sdmDevDB, dir->id);
	 
	//-- If found then just update the entry with the new path 
	//-- and channel information.
	if( devRec ) {

		//-- See if the path has already been recorded.
		pathfound = 0;
		for(i=0; i < devRec->pathCnt; i++) {
			if( devRec->pathList[i].path == dir->path ) {
				pathfound = 1;
				break; //-- Done.
			}
		}
		if( !pathfound ) {
			devRec->pathList[devRec->pathCnt].path     = dir->path;
			devRec->pathList[devRec->pathCnt].state    = SDM_PATH_STATE_AVAILABLE;
			devRec->pathList[devRec->pathCnt].wwn      = dir->wwn;
			devRec->pathList[devRec->pathCnt].lun      = dir->lun;
			devRec->pathList[devRec->pathCnt].llHandle = dir->llHandle;
			devRec->pathCnt++;

			E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
				("SDM_INIT: Add Dev Path To Dev DB:   0x%016llx:%d", 
				 dir->wwn, dir->lun));
		}

		// inform raid-cfgd if device state has changed
		if(devRec->devType == SCSI_DEVTYPE_DISK) {
			sdm_setDevrec(dir, devRec);

			if(dir->control & SDM_DIE_CTRL_INIT_ERROR) { // device has gone into init state
				if(devRec->handleCnt > 0)
					devRec->state = SDM_DEV_STATE_INIT_OPENED;
				else
					devRec->state = SDM_DEV_STATE_INIT;
				sdm_sendDevDelNotify("evm", devRec);	
				sdm_sendDevAddNotify("raid-config", devRec);
			}
			else {		// device probably came back into open/close state
				if(devRec->handleCnt > 0)
					devRec->state = SDM_DEV_STATE_OPENED;
				else 
					devRec->state = SDM_DEV_STATE_CLOSED;
				if(!(dir->control & SDM_DIE_CTRL_EVM_NO_NOTIFY))
					sdm_sendDevAddNotify("evm", devRec);
				sdm_sendDevAddNotify("raid-config", devRec);
			}
		}
	}
	else {

		//-- If not found then add a new record to the database.

		devRec = (sdm_devDBRec_t *)malloc(sizeof(sdm_devDBRec_t));
		if( !devRec ) {
			//printf("Error: Could not allocated device database record. ");
			//printf("Device not added to database.\n");
			return(NULL);
		}
		else  {

			memset(devRec, 0, sizeof(sdm_devDBRec_t));
			sdm_setDevrec(dir, devRec);

			if( dir->control & SDM_DIE_CTRL_INIT_ERROR )
				devRec->state = SDM_DEV_STATE_INIT;
			else
				devRec->state =  SDM_DEV_STATE_CLOSED;
				
			devRec->pathList[devRec->pathCnt].path     = dir->path;
			devRec->pathList[devRec->pathCnt].state    = SDM_PATH_STATE_AVAILABLE;
			devRec->pathList[devRec->pathCnt].wwn      = dir->wwn;
			devRec->pathList[devRec->pathCnt].lun      = dir->lun;
			devRec->pathList[devRec->pathCnt].llHandle = dir->llHandle;
			devRec->pathCnt++;

   			sdm_dbAddRec(sdmDevDB, devRec->id, devRec, sizeof(sdm_devDBRec_t),1);
			E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
				("SDM_INIT: Add Dev Rec To Dev DB:   0x%016llx:%d", 
				 dir->wwn, dir->lun));

			// Send new device notification msg to the evm.
			if(!(dir->control & SDM_DIE_CTRL_EVM_NO_NOTIFY) ) {
			    if( devRec->devType == SCSI_DEVTYPE_DISK)
				    sdm_sendDevAddNotify("evm", devRec);
                else
                    sdm_sendDevAddNotifySpm(devRec);
			}
		}
	}
	return  devRec;
} // sdm_initCreateDevRec

//------------------------------------------------------------------------------
// Function Name:   sdm_createDevName
// 
// Description:     
//
// Data Created:    March 27, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_createDevName(char *name, char *vendor, char *sn, uint32 lun) 
{ 
	char lunstr[8]; 

	//printf("************ LUN: %d\n", lun);
	sprintf(lunstr,"%d", lun);

	strcpy(name, vendor);
	strcat(name,"_");
	strcat(name, sn);
	strcat(name,"_");
	strcat(name, lunstr);
	
	return SDM_OK;
} // sdm_createDevName

//------------------------------------------------------------------------------
// Function Name:   sdm_procRLun
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_procRLun(sdm_dir_t *dir, scsi_relayRspMsg_t *i_msg ) 
{ 
	int32   i,j;
	uchar8 *buf;
	uint32 lun;
	uint32 lunCnt;
	scsi_sense_t *sense;

	//printf("sdm_procRLun\n");

	switch(i_msg->status & SD_DEV_STATUS_MASK)  
	{
		case SCSI_DEV_STATUS_CHECK:
		case SCSI_DEV_STATUS_CMD_TERMINATED:
			if( i_msg->status & SD_DEV_SENSE_VALID ) {

				sense = (scsi_sense_t *)i_msg->senseData;

				/* Look at sense data */
				switch( sense->senseKey )
				{
					case SCSI_SENSE_KEY_NOT_READY:
					case SCSI_SENSE_KEY_HARDWARE_ERROR:
						//-- Fatal Error unable to access device 
						break;

					case SCSI_SENSE_KEY_ILLEGAL_REQUEST:

						// Device does not support this command, so
						// probe for n LUN's and see what comes
						// back.
						for(i=1; i < 32; i++) 
							sdm_initNewLun(dir->wwn,i,dir->path, dir->llHandle);

						break;

					case SCSI_SENSE_KEY_NO_SENSE:
					case SCSI_SENSE_KEY_RECOVERED_ERROR:
					default:
						break;
				}
			}
			break;

		case SCSI_DEV_STATUS_GOOD:
		case SCSI_DEV_STATUS_CONDITION_MET:
			/* Parse the REPORT_LUN response. */
			/* For each LUN that is reported create SM instance. */

			//-- Get pointer to buffer containing the inquiry data.
			buf = i_msg->data;

			lunCnt = ((buf[0] << 24) |
					  (buf[1] << 16) |
					  (buf[2] <<  8) |
					  (buf[3]));

			//printf("lunCnt: %d\n", lunCnt);
			lunCnt /= 8;

			for(i=j=0; j < lunCnt; i+=8, j++ ) {

				//-- Extract LUN information from response buffer.
				lun = ((buf[8+i] << 24)  |
					   (buf[9+i] << 16)  |
					   (buf[10+i] <<  8) |
					   (buf[11+i]));

				lun >>= 16;
				//printf("procRLun: %d\n", lun);

				if( lun != dir->lun ) {
					sdm_initNewLun(dir->wwn, lun, dir->path, dir->llHandle);
				}
			}
			break;

		default:
	}

	//-- Perform next initialization step.
	sdm_initDevSM(dir->nextEvt, dir, NULL);  

	return SDM_OK;
} // sdm_procRLun


//------------------------------------------------------------------------------
// Function Name:   sdm_procRCap
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_procRCap(sdm_dir_t *dir, scsi_relayRspMsg_t *i_msg ) 
{ 
	uint32        fgoodRsp = 0;
	uint32        tmp;
	uchar8       *buf;
    scsi_sense_t *sense;
 
	//printf("sdm_procRCap\n");
 
    /*
     * Check command completion status.
     */
    switch( i_msg->status & SD_DEV_STATUS_MASK )
    {
        case SCSI_DEV_STATUS_GOOD:
        case SCSI_DEV_STATUS_CONDITION_MET:
			fgoodRsp = 1;
            break;

        case SCSI_DEV_STATUS_RESERVE_CONFLICT:
            /* Device is reserved by another initiator so
             * don't try to use it anymore.
             */
			fgoodRsp = 0;
            break;
 
        case SCSI_DEV_STATUS_CHECK:
        case SCSI_DEV_STATUS_CMD_TERMINATED:
        default:
            if( i_msg->status & SD_DEV_SENSE_VALID ) {
 
                sense = (scsi_sense_t *)i_msg->senseData;
 
                /* Look at sense key */
                switch( sense->senseKey )
                {
                    case SCSI_SENSE_KEY_RECOVERED_ERROR:
						fgoodRsp = 1;
						break;

                    case SCSI_SENSE_KEY_NOT_READY:
                    case SCSI_SENSE_KEY_HARDWARE_ERROR:
                    case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
                        /* Fatal Error unable to access device */
 
                    case SCSI_SENSE_KEY_NO_SENSE:
                    default:
						fgoodRsp = 0;
                }
            }
    }

	if( fgoodRsp ) {
		//-- Get pointer to buffer containing the read capacity data.
		buf = i_msg->data;

		// Parse the READ CAPACITY response. 
		tmp = 1 + ((buf[0] << 24) |
				   (buf[1] << 16) |
				   (buf[2] <<  8) |
				   (buf[3]));
		dir->physicalBlockCnt = tmp;
			
		dir->physicalBlockSize = ((buf[4] << 24) |
								  (buf[5] << 16) |
								  (buf[6] <<  8) |
								  (buf[7]));

		// Only advertise devices with "sane" blocksizes; ignore all others 
		if(dir->physicalBlockSize == 512 || dir->physicalBlockSize == 520) {

			dir->logicalBlockSize = 512;
#ifdef TURN_OFF_BLOCK_CHECKSUMMING
			/* XXX Radek For now assume all disks with 512BPS support block checksumming */
			if (dir->physicalBlockSize == 512)
				dir->logicalBlockCnt = dir->physicalBlockCnt/(BCSUM_BLOCK_SPAN+1)*BCSUM_BLOCK_SPAN - SCSI_MAPPING_OFFSET_BLOCKS;
			else
				dir->logicalBlockCnt = dir->physicalBlockCnt;
#else
			dir->logicalBlockCnt = dir->physicalBlockCnt;
#endif
			/*
			  printf("sdm_procRCap 0x%llx/%d: physical = %d/%d logical = %d/%d\n",
			  dir->wwn, dir->lun, 
			  dir->physicalBlockCnt, dir->physicalBlockSize,
			  dir->logicalBlockCnt, dir->logicalBlockSize);
			*/
		}
		else {
			dir->physicalBlockCnt  = 0;
			dir->logicalBlockSize  = 0;
			dir->physicalBlockSize = 0;
		}
	}
	else {
		dir->physicalBlockCnt  = 0;
		dir->logicalBlockSize  = 0;
		dir->physicalBlockSize = 0;
		dir->control |= (SDM_DIE_CTRL_EVM_NO_NOTIFY | SDM_DIE_CTRL_INIT_ERROR);	
	}

	//-- Perform next initialization step.
	sdm_initDevSM(dir->nextEvt, dir, NULL); 

	return SDM_OK;
} // sdm_procRCap

//------------------------------------------------------------------------------
// Function Name:   sdm_procSnRsp
// 
// Description:     
//
// Data Created:    March 27, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_procSnRsp(sdm_dir_t *dir, scsi_relayRspMsg_t *i_msg ) 
{ 
	int slen;
	uint32 fgoodRsp = 0;
	uchar8 *buf;
	scsi_sense_t *sense;

	E_LOG (class_1, notice_s, eee_getApplicationId("sdm"), 0, 0, 0,
		   ("SDM_INIT: Proc SN Rsp: 0x%016llx:0, path: %x", dir->wwn, dir->path));

	switch(i_msg->status & SD_DEV_STATUS_MASK)  
	{
		case SCSI_DEV_STATUS_GOOD:
		case SCSI_DEV_STATUS_CONDITION_MET:
			fgoodRsp = 1;
			break;

		case SCSI_DEV_STATUS_CHECK:
		case SCSI_DEV_STATUS_CMD_TERMINATED:
		default:
			if( i_msg->status & SD_DEV_SENSE_VALID ) {

				sense = (scsi_sense_t *)i_msg->senseData;

				/* Look at sense data */
				switch( sense->senseKey )
				{
					case SCSI_SENSE_KEY_RECOVERED_ERROR:
						fgoodRsp = 1;
						break;

					case SCSI_SENSE_KEY_NOT_READY:
					case SCSI_SENSE_KEY_HARDWARE_ERROR:
						//-- Fatal Error unable to access device 

					case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
						// Device does not support this command

					case SCSI_SENSE_KEY_NO_SENSE:
					default:
						fgoodRsp = 0;
						break;
				}
			}
	}

	if( fgoodRsp ) {
		// Parse serial number; remove leading spaces, replace
		// (and compact) inter-token spaces with dash ('-'), and
		// remove trailing spaces.
		slen = i_msg->data[3];
		buf = &i_msg->data[4];
		sdm_initParseInqString(buf, slen, dir->st.sn, sizeof(dir->st.sn));
	}
	else {
		dir->control |= (SDM_DIE_CTRL_EVM_NO_NOTIFY | SDM_DIE_CTRL_INIT_ERROR);
	}

	//-- Perform next initialization step.
	sdm_initDevSM(dir->nextEvt, dir, i_msg); 

	return SDM_OK;
} // sdm_procSnRsp

//------------------------------------------------------------------------------
// Function Name:   sdm_initFinish
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_initFinish(sdm_dir_t *dir, char *notused) 
{ 
	//printf("sdm_initFinish\n");

	//-- Create the device hitachi device name string.
	sdm_createDevName(dir->st.name, dir->st.vendor, dir->st.sn, dir->lun);

	//-- Create the real scsi id.
	dir->id = sdm_initGetScsiId(dir->wwn, dir->lun, dir->st.vendor);

	//-- Call SCSI application to add device.
	//-- We will then create the SDM data base record
	//-- when a successful reply is received from the SCSI app.
	sdm_sendScsiAddMsg(dir);

	return SDM_OK;
} // sdm_initFinish

//------------------------------------------------------------------------------
// Function Name:   sdm_initTimeout
// 
// Description:     
//
// Data Created:    March 18, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_initTimeout(sdm_dir_t *dir ) 
{ 
	//printf("sdm_initTimeout\n");

	return SDM_OK;
} // sdm_initTimeout


//------------------------------------------------------------------------------
// Function Name:   sdm_initParseInqString
// 
// Description:     Inquiry string elements (VENDOR strings, PRODUCT strings, 
//                  and SN strings) may have leading, trailing and embedded 
//                  spaces. We want to strip leading and trailing spaces, and
//                  compress and replace by single '-' embedded spaces.
//
// String parse state machine:
//   state == 0 --> no non-space character found; filtering leading spaces.
//   state == 1 --> non-space character found.
//   state == 2 --> space character found after non-space character.
//
// Data Created:    April 23, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int 
sdm_initParseInqString(char *in_str, int in_strlen, char *out_str, int out_strlen)
{
	int state = 0;
	int i;

	memset(out_str, 0, sizeof(out_strlen));

	for (i = 0; in_strlen > 0 && i < out_strlen; --in_strlen, ++in_str) {
		switch (state) {
		case 0:
			if (*in_str != ' ') {
				out_str[i++] = *in_str;
				state = 1;
			}
			break;
						
		case 1:
			if (*in_str != ' ') {
				out_str[i++] = *in_str;
				state = 1;
			}
			else
				state = 2;
			break;

		case 2:
			if (*in_str != ' ') {
				out_str[i++] = '-';
				out_str[i++] = *in_str;
				state = 1;
			}
			else
				state = 2;
			break;
		}
	} /* for */
	return(NFX_OK);
}

#if 0
//------------------------------------------------------------------------------
// Function Name:   sdm_sendTestUnitCmd
// 
// Description:     
//
// Data Created:    August 16, 2002
//
// Returns:         SDM_OK
//------------------------------------------------------------------------------
int
sdm_sendTestUnitCmd( sdm_dir_t *dir ) 
{
	uint32 mlen, destCpu, destSlot, tid; 
	scsi_relayReqMsg_t *scsi_o_msg;

	//printf("sdm_sendTestUnitCmd\n");

	//-- Allocate an SDM transaction Id
	tid = sdm_allocRelayTransaction( NULL,
									 NULL,
									 sdm_initDevProcMsg,
									 (int)dir);

	//-- Allocate memory for the message.
	mlen = (sizeof(scsi_relayReqMsg_t) - 1 + 16);
	scsi_o_msg = (scsi_relayReqMsg_t *)malloc(mlen);
	memset(scsi_o_msg, 0, mlen);

	//-- Setup the Test Unit Ready CDB.
	scsi_o_msg->cdb[0] = SCSI_CMD_TEST_UNIT_READY;  

	//-- Setup the SCSI msg header.
	scsi_o_msg->hdr.mid     = SCSI_MID_RELAY_REQ;
	scsi_o_msg->hdr.cHandle = tid;
	scsi_o_msg->hdr.path    = dir->path;
	scsi_o_msg->lun         = dir->lun;
	scsi_o_msg->dataLen     = 0;

	//-- Send read command and data to the SCSI application.
	destCpu = SCSI_PATH_GET_PORT(dir->path);
	destSlot = SCSI_PATH_GET_SLOT(dir->path);
	sdm_sendMsg( (char *)scsi_o_msg, mlen, "scsi", destCpu, destSlot, 0, 0,0);
	free(scsi_o_msg);

	return SDM_OK;
} // sdm_sendTestUnitCmd
#endif
