#if !defined(LINUX_TEST) && !defined(EVAL_BOARD) && !defined(SSC_MGMT) && !defined(SIBYTE)

#if defined(__OpenBSD__)
#include <sys/types.h>
#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <nfx-types.h>
#else
#include <nfx-incl.h>
#endif

#include <nfx-defs.h>
#include "gt64240-sdma.h"
#include "gt64240-uart-regs.h"
#include "gt64240-uart.h"

#if defined(DEBUG_UART)
#define debug(x) printf x
#else
#define debug(x)
#endif

int entering_debugger;

#define u_assert(x) { if (!entering_debugger && !(x)) {entering_debugger = 1; assert((x)); } }

/* Preallocated descriptors array, extra for alignment */
static struct gtsdma_desc descriptors[GTSDMA_NRX * GTSDMA_UNITS +
								GTSDMA_NTX * GTSDMA_UNITS + 4];
static struct gtsdma_desc *last_alloc;

/* Preallocate memory for rx/tx buffers, extra for alignment */
static char bufmemory[GTSDMA_NTX * GTSDMA_UNITS +
					 GTSDMA_NRX * GTSDMA_UNITS + 64];
static char *mem_last_alloc;

extern int gtuart_initialized;


static void *align(void *p, int align)
{
	if ((unsigned)p % align)
		p = (char *)p + align - (unsigned)p % align;
	assert((unsigned)p % align == 0);
	return p;
}

#define GTSDMA_DESC_ALIGN 32


static int
chain_nextidx(struct gtsdma_chain *chain, int idx)
{
	u_assert(idx < chain->length);
	
	++idx;
	if (idx == chain->length)
		idx = 0;

	u_assert(idx < chain->length);

	return idx;
}


static int
alloc_chain(struct gtsdma_chain *chain, unsigned length)
{
	if (!last_alloc)
		last_alloc = align(descriptors, GTSDMA_DESC_ALIGN);

	chain->length = length;
	chain->free = length;
	chain->first = 0;
	chain->last = 0;

	chain->desc = last_alloc;
	last_alloc += length;
	
	/* Use uncached memory to write into descriptors */
	CACHE_FLUSH(chain->desc, last_alloc);
	chain->desc = (struct gtsdma_desc *)K0_TO_K1(chain->desc);

	assert(last_alloc <=
		   (struct gtsdma_desc *)(char *)descriptors + sizeof(descriptors));

	return NFX_OK;
}


static int
alloc_memory(char **ptr, int length)
{
	if (!mem_last_alloc)
		mem_last_alloc = align(bufmemory, 32);

	*ptr = mem_last_alloc;
	mem_last_alloc += length;

	/* Use uncached memory for further accesses to buffer memory */
	CACHE_FLUSH(*ptr, mem_last_alloc);
	*ptr = (char *)K0_TO_K1(*ptr);
	
	assert(mem_last_alloc <= bufmemory + sizeof(bufmemory));

	return NFX_OK;
}


static void
init_txdescriptor(struct gtsdma_desc *d, char *txbuf,
				  struct gtsdma_desc *next)
{
	d->cmd_status = 0;
	d->bufsize_cnt = (1 << 16) | 1;
	d->bufptr = K1_TO_PHYS(txbuf);
	d->next = K1_TO_PHYS(next);
}


static void
init_txchain_descriptors(struct gtsdma_chain *chain)
{
	int i;
	int last;

	for (i = 0, last = chain->length - 1; i < last; ++i)
		init_txdescriptor(&chain->desc[i], &chain->mem[i],
						  &chain->desc[i + 1]);
	init_txdescriptor(&chain->desc[i], &chain->mem[i], &chain->desc[0]);
}


static int
init_txchain(struct gtsdma_chain *chain, unsigned length)
{
	int rc;
	
	if ((rc = alloc_chain(chain, length)) == NFX_OK &&
		(rc = alloc_memory(&chain->mem, length)) == NFX_OK) {
		init_txchain_descriptors(chain);
		chain->txcmd = GTSDMA_FIRST | GTSDMA_OWN | GTSDMA_EI | GTSDMA_LAST |
			GT_MPSC_UARTSDMA_TX_P;
	}

	return rc;
}


static void
init_rxdescriptor(struct gtsdma_desc *d, char *rxbuf,
				  struct gtsdma_desc *next)
{
	d->cmd_status = GTSDMA_OWN | GTSDMA_EI;
	d->bufsize_cnt = 1 << 16;
	d->bufptr = K1_TO_PHYS(rxbuf);
	d->next = K1_TO_PHYS(next);
}


static void
init_rxchain_descriptors(struct gtsdma_chain *chain)
{
	int i;
	int last;

	for (i = 0, last = chain->length - 1; i < last; ++i)
		init_rxdescriptor(&chain->desc[i], &chain->mem[i],
						  &chain->desc[i + 1]);
	init_rxdescriptor(&chain->desc[i], &chain->mem[i], &chain->desc[0]);
}


static int
init_rxchain(struct gtsdma_chain *chain, unsigned length)
{
	int rc;

	if ((rc = alloc_chain(chain, length)) == NFX_OK &&
		(rc = alloc_memory(&chain->mem, length)) == NFX_OK)
		init_rxchain_descriptors(chain);

	return rc;
}


static void
sdma_free_txchain(struct gtsdma_chain *chain)
{
	/* Should not be called for now */
	assert(0);
}


static void
sdma_free_rxchain(struct gtsdma_chain *chain)
{
	/* Should not be called for now */
	assert(0);
}


void
gtsdma_txabort(struct gtsdma *sc)
{
	/* write abort transmit bit and wait for it to clear */
	gtsdma_write_reg(sc->unit, GTSDMA_SDCM, GTSDMA_SDCM_AT);
	while (gtsdma_read_reg(sc->unit, GTSDMA_SDCM) & GTSDMA_SDCM_AT);
}


void
gtsdma_rxabort(struct gtsdma *sc)
{
	/* write abort receive bit and wait for it to clear */
	gtsdma_write_reg(sc->unit, GTSDMA_SDCM, GTSDMA_SDCM_AR);
	while (gtsdma_read_reg(sc->unit, GTSDMA_SDCM) & GTSDMA_SDCM_AR);
}


int
gtsdma_init(struct gtsdma *sc, int unit)
{
	int rc;

	sc->unit = unit;
	
	gtsdma_rxabort(sc);
	gtsdma_txabort(sc);

	if ((rc = init_txchain(&sc->tx, GTSDMA_NTX)) == NFX_OK
		&& (rc = init_rxchain(&sc->rx, GTSDMA_NRX)) == NFX_OK) {
		gtsdma_write_reg(unit, GTSDMA_SDC, GTSDMA_SDC_DEFAULT);

		/* Inilize the hardware descriptor pointers to point to the
		   first descriptors in chains */
		gtsdma_write_reg(unit, GTSDMA_SCRDP, K1_TO_PHYS(&sc->rx.desc[0]));
		gtsdma_write_reg(unit, GTSDMA_SCTDP, K1_TO_PHYS(&sc->tx.desc[0]));
		gtsdma_write_reg(unit, GTSDMA_SFTDP, K1_TO_PHYS(&sc->tx.desc[0]));

		/* Enable receiver */
		gtsdma_rxstart(sc);

		/* Enable interrupts */
		gtsdma_write_intreg(unit, GTSDMA_MASK, 0xffffffff);
	}

	if (rc != NFX_OK) {
		sdma_free_txchain(&sc->tx);
		sdma_free_rxchain(&sc->rx);
	}
	
	return rc;
}


void
gtsdma_rxstart(struct gtsdma *sc)
{
	gtsdma_write_reg(sc->unit, GTSDMA_SDCM, GTSDMA_SDCM_ERD);
}


void
gtsdma_txstart(struct gtsdma *sc)
{
	gtsdma_write_reg(sc->unit, GTSDMA_SDCM, GTSDMA_SDCM_TXD);
}


/*
 * gtsdma_desc - get the last descriptor in the chain
 */
static struct gtsdma_desc *
chain_getdesc(struct gtsdma_chain *ch)
{
	struct gtsdma_desc *d = 0;

	u_assert(ch->free);

	if (ch->free) {
		d = &ch->desc[ch->last];
		ch->last = chain_nextidx(ch, ch->last);
		--ch->free;
	}

	u_assert(ch->last != ch->first);
	u_assert((ch->first + ch->length - ch->free) % ch->length == ch->last);

	return d;
}


static void
gtsdma_tx_char(volatile struct gtsdma_desc *d, int c, uint32 cmd_status)
{
	*(char *)PHYS_TO_K1(d->bufptr) = c;
	d->cmd_status = cmd_status;
}

/*
 * gtsdma_tx - transmit some characters
 */
int
gtsdma_tx(struct gtsdma *sc, char *buf, int length)
{
	int rc = NFX_OK;
	uint32 cmd = sc->tx.txcmd;
	int start;

	if (length <= sc->tx.free) {
		start = length;
		while (length) {
			volatile struct gtsdma_desc *d = chain_getdesc(&sc->tx);

			gtsdma_tx_char(d, *buf, cmd);
			--length;
			++buf;
		}

		if (start)
			gtsdma_txstart(sc);
	}
	else
		rc = -1;

	return rc;
}


static void
do_error(struct gtsdma *sc)
{
	panic("");
}


static struct gtsdma_desc *
next_tx(struct gtsdma_chain *c)
{
	struct gtsdma_desc *d = &c->desc[c->first];
	if (c->first == c->last || (d->cmd_status & GTSDMA_OWN))
		d = 0;

	return d;
}


static void
tx_cleanup(struct gtsdma *sc, struct gtsdma_desc *d)
{
	struct gtsdma_chain *c = &sc->tx;

	/* Reinitialize cnt to 1 because hardware clears it */
	c->desc[c->first].bufsize_cnt = (1 << 16) | 1;
	c->first = chain_nextidx(c, c->first);
	++c->free;

	u_assert(c->free <= c->length);
	u_assert((c->first + c->length - c->free) % c->length == c->last);
}


static void
do_txeof(struct gtsdma *sc)
{
	struct gtsdma_desc *d;
	int processed = 0;

	while ((d = next_tx(&sc->tx))) {
		tx_cleanup(sc, d);
		++processed;
		u_assert(processed <= sc->tx.length)
	}
}


static struct gtsdma_desc *
next_rx(struct gtsdma_chain *c)
{
	struct gtsdma_desc *d = &c->desc[c->last];
	if (d->cmd_status & GTSDMA_OWN)
		d = 0;

	return d;
}


int gotchars ;


static void
rx_cleanup(struct gtsdma_chain *c, struct gtsdma_desc *d)
{
	/* Reinitialize count to 0 and return the descriptor to the device */
	//printf("rx cleanup\n");
	++gotchars;
	d->bufsize_cnt = (1 << 16);
	d->cmd_status = GTSDMA_OWN | GTSDMA_EI;
	c->last = chain_nextidx(c, c->last);
}


static void
do_rxeof(struct gtsdma *sc)
{
	struct gtsdma_desc *d;
	struct gtsdma_chain *c = &sc->rx;

	while ((d = next_rx(c)))
		rx_cleanup(c, d);
}


static int
intr_rxdone(uint32 status)
{
#if defined(GT_MPSC_GOOD_REGS)
	return status & GTSDMA_INT_RXBUF;
#else
	return 1;
#endif
}


static int
intr_txdone(uint32 status)
{
#if defined(GT_MPSC_GOOD_REGS)
	return status & GTSDMA_INT_TXBUF;
#else
	return 1;
#endif
}


static int
intr_error(uint32 status)
{
#if defined(GT_MPSC_GOOD_REGS)
	return status & (GTSDMA_INT_RXERR | GTSDMA_INT_TXERR);
#else
	return 0;
#endif
}


static void
do_intr(struct gtsdma *sc, uint32 status)
{
	if (intr_rxdone(status))
		do_rxeof(sc);
	if (intr_txdone(status))
		do_txeof(sc);
	if (intr_error(status))
		do_error(sc);
}

/*
 * gtsdma_intr - process interrupt
 */
void
gtsdma_intr(struct gtsdma *sc)
{
	uint32 status;

	status = gt_read_reg(GTSDMA_CAUSE);
	/* Clear the interrupts we are going to process */
#if 0
	gt_write_reg(GTSDMA_CAUSE, ~status);
#else
	gt_write_reg(GTSDMA_CAUSE, 0);
#endif
	do_intr(sc, status);
}


/*
 * gtsdma_rx - read some characters from rxmemory. Return the number
 * of characters read
 */
int
gtsdma_rx(struct gtsdma *sc, char *buf, int length)
{
	struct gtsdma_chain *c = &sc->rx;
	int cnt = 0;
	while (c->first != c->last && cnt < length) {
		*buf = c->mem[c->first];
		c->first = chain_nextidx(c, c->first);
		++cnt;
		++buf;
	}

	return cnt;
}


/*
 * gtsdma_tx_nointerrupt - output a character, do not generate an interrupt
 */
void
gtsdma_tx_nointerrupt(struct gtsdma *sc, int c)
{
	struct gtsdma_chain *ch = &sc->tx;
	volatile struct gtsdma_desc *d = chain_getdesc(ch);

	while (ch->free < 50)
		gtsdma_intr(0);

	/* Output char, do not set the interrupt bit */
	gtsdma_tx_char(d, c, sc->tx.txcmd & ~GTSDMA_EI);
	gtsdma_txstart(sc);
	while (d->cmd_status & GTSDMA_OWN);
}


void
gtsdma_print_status(struct gtsdma *sc)
{
	printf("rf=%d rl=%d tf=%d tl=%d\n", sc->rx.first, sc->rx.last,
		   sc->tx.first, sc->tx.last);
		   
}


#define DEBUG_CHAR 2

void
gtsdma_check_debugger(struct gtsdma *sc)
{
	int s = splhigh();
	int i, last = sc->rx.last;
	char *mem = sc->rx.mem;

	for (i = sc->rx.first; i < last; ++i)
		if (mem[i] == DEBUG_CHAR)
			Debugger();

	splx(s);
}

#endif
