#ifndef __GT64240_SDMA_H
#define __GT64240_SDMA_H

#include "gt64240.h"

/* SDMA descriptor */
struct gtsdma_desc {
	uint32 cmd_status;
	uint32 bufsize_cnt;
	uint32 bufptr;
	uint32 next;
};


struct gtsdma_chain {
	int length;					/* number of descriptors in chain */
	int free;					/* free descriptors */
	int first;					/* First output/input descriptor */
	int last;					/* Last output/input descriptor */
	struct gtsdma_desc *desc;	/* Array of descriptors */
	char *mem;					/* Rx buffer memory */
	uint32 txcmd;				/* default cmd_status bits */
};


struct gtsdma {
	int unit;
	struct gtsdma_chain rx;
	struct gtsdma_chain tx;
};


/* Number of SDMA units in the system */
#define GTSDMA_UNITS 2

#define GTSDMA_MAXBUFSIZE (64 * 1024)

/* SDMA command/status word bits, common for all modes */
#define GTSDMA_LAST (1<<16)	/* First descriptor in chain */
#define GTSDMA_FIRST (1<<17)	/* Last descriptor in chain */
#define GTSDMA_EI (1<<23)		/* Enable interrupt */
#define GTSDMA_AM (1<<30)		/* Auto Mode */
#define GTSDMA_OWN (1<<31)		/* Onwership */

/* SDMA Register Map */
#define GTSDMA_0 0x4000		/* SDMA 0 registers base */
#define GTSDMA_1 0x6000		/* SDMA 1 registers base */

#define GTSDMA_SDC 0x0			/* Configuration Register */
#define GTSDMA_SDCM 0x8		/* Command Register */
#define GTSDMA_SCRDP 0x810		/* Current Rx Descriptor Pointer */
#define GTSDMA_SCTDP 0xc10		/* Current Tx Descriptor Pointer */
#define GTSDMA_SFTDP 0xc14		/* First Tx Descriptor Pointer */

#define GTSDMA_CAUSE 0xb800	/* Interrupt Cause Register */
#define GTSDMA_MASK 0xb880		/* Interrupt Mask Register */

/* SDMA Cause and Mask register bits */
#define GTSDMA_INT_RXBUF (1<<0) /* Rx descriptor closed */
#define GTSDMA_INT_RXERR (1<<1) /* Rx error */
#define GTSDMA_INT_TXBUF (1<<2) /* Tx descriptor closed */
#define GTSDMA_INT_TXERR (1<<3) /* Tx error */
#define GTSDMA_INT_UNITOFFSET 8 /* Offset of bits */

/* SDMA Configuration Register bits */
#define GTSDMA_SDC_RFT (1<<0)	/* Receive FIFO threshold */
#define GTSDMA_SDC_SFM (1<<1)	/* Single Frame Mode */
#define GTSDMA_SDC_RC_OFFSET 2	/* Retransmit Count */
#define GTSDMA_SDC_BLMR (1<<6)	/* Big/Little Receive Mode */
#define GTSDMA_SDC_BLMT (1<<7)	/* Big/Little Transmit Mode */
#define GTSDMA_SDC_POVR (1<<8)	/* PCI override */
#define GTSDMA_SDC_RIFB (1<<9)	/* Interrupt on frame boundaries */
#define GTSDMA_SDC_BSZ_OFFSET 12 /* Burst Size */

#define GTSDMA_SDC_DEFAULT 0x30ff

/* SDMA Command Register bits */
#define GTSDMA_SDCM_ERD (1<<7)	/* Enable Rx */
#define GTSDMA_SDCM_AR (1<<15)	/* Abort Receive */
#define GTSDMA_SDCM_STD (1<<16) /* Stop Tx */
#define GTSDMA_SDCM_TXD (1<<23) /* Tx Demand */
#define GTSDMA_SDCM_AT (1<<31)	/* Abort Transmit */

static inline int
gtsdma_valid_unit(uint32 unit)
{
	return unit < GTSDMA_UNITS;
}


static inline int
gtsdma_valid_reg(uint32 reg)
{
	return reg == GTSDMA_SDC || reg == GTSDMA_SDCM || reg == GTSDMA_SCRDP
		|| reg == GTSDMA_SCTDP || reg == GTSDMA_SFTDP;
}


static inline int
gtsdma_valid_intreg(uint32 reg)
{
	return reg == GTSDMA_CAUSE || reg == GTSDMA_MASK;
}


static inline uint32
gtsdma_unit2base(uint32 unit)
{
	assert(gtsdma_valid_unit(unit));
	return 0x4000 + unit * 0x2000;
}


static inline uint32
gtsdma_read_reg(uint32 unit, uint32 reg)
{
	assert(gtsdma_valid_unit(unit));
	assert(gtsdma_valid_reg(reg));
	return GT_SWAP(GT64W(gtsdma_unit2base(unit) + reg));
}


static inline void
gtsdma_write_reg(uint32 unit, uint32 reg, uint32 val)
{
	assert(gtsdma_valid_unit(unit));
	assert(gtsdma_valid_reg(reg));
	GT_REG_WRITE(gtsdma_unit2base(unit) + reg, val);
}


static inline uint32
gtsdma_read_intreg(uint32 unit, uint32 reg)
{
	uint32 shift = unit * GTSDMA_INT_UNITOFFSET;

	assert(gtsdma_valid_unit(unit));
	assert(gtsdma_valid_intreg(reg));
	return (GT_SWAP(GT64W(reg)) & (0xff << shift)) >> shift;
}


static inline void
gtsdma_write_intreg(uint32 unit, uint32 reg, uint32 val)
{
	uint32 shift = unit * GTSDMA_INT_UNITOFFSET;
	uint32 v;

	assert(gtsdma_valid_unit(unit));
	assert(gtsdma_valid_intreg(reg));

	v = GT_SWAP(GT64W(reg));
	v &= ~(0xff << shift);
	v |= (val << shift);

	GT_REG_WRITE(reg, v);
}


void gtsdma_txabort(struct gtsdma *sc);
void gtsdma_rxabort(struct gtsdma *sc);
int gtsdma_init(struct gtsdma *sc, int unit);
void gtsdma_rxstart(struct gtsdma *sc);
void gtsdma_txstart(struct gtsdma *sc);
int gtsdma_tx(struct gtsdma *sc, char *buf, int length);
void gtsdma_intr(void *arg);
int gtsdma_rx(struct gtsdma *sc, char *buf, int length);


#endif
