/*
 *-----------------------------------------------------------------
 *
 * Name:	spinlock-test.c
 *
 * RCS:  $Id: spinlock-test.c,v 1.4 2002/12/04 00:47:18 rick Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	11/12/02
 *
 *-----------------------------------------------------------------
 */

#ifdef NFX_SMP

#ifdef NFX_SPINLOCK_DEBUG

#include <nfx-incl.h>
#include <exports/smp-api.h>

#define SPINLOCK_TEST_NUM_OPS	20000
#define SL_TEST_DEBUG(x)			

#define SPINLOCK_TEST_NUM_LOCKS			200
#define SPINLOCK_TEST_NUM_LOCKS_GET_MAX	100
#define SPINLOCK_TEST_NUM_LOCKS_TRY_MAX	150

#define SPINLOCK_TEST_FUNC_NUM	7
address_t	spinlock_test_func[SPINLOCK_TEST_FUNC_NUM];
/*
 * 0 = GETLOCK
 * 1 = UNLOCK
 * 2 = TRYLOCK
 * 3 = INC INT
 * 4 = DEC INT
 * 5 = INC DWORD
 * 6 = DEC DWORD
 */

nfx_spinlock_t	spinlock_test_lock[SPINLOCK_TEST_NUM_LOCKS];
uint32			spinlock_test_inuse[SPINLOCK_TEST_NUM_LOCKS]	CPU_PRIVATE;

typedef struct sl_test_stats {
	uint32	getlock;
	uint32	getlock_busy;
	uint32	getlock_not_busy;
	uint32	getlock_wait_time;

	uint32	unlock;

	uint32	trylock;
	uint32	trylock_passed;
	uint32	trylock_failed;

	uint32	inc_int;
	uint32	dec_int;
	uint32	inc_dword;
	uint32	dec_dword;
} sl_test_stats_t;

sl_test_stats_t	spinlock_test_totals;

uint32	spinlock_test_int_counter;
uint64	spinlock_test_dword_counter;

int32 spinlock_test_verify(uint32 which_core);

sl_test_stats_t	spinlock_test_core_stats[SMP_NUM_CPUS];

/*-----------------------------------------------------------------
 * Name:	spinlock_test
 *
 * Description:
 *			Do SPINLOCK_TEST_NUM_OPS random spinlock/atomic operations.
 *			Procedure: start test on altcpus, then start test on core0
 *				they will handshake via the sb1250 mailbox.
 *			At end of test, altcpus will send results to core0 for validation
 *			with test totals.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	11/12/02
 *
 *-----------------------------------------------------------------
 */
int32
spinlock_test(
	uint32	which_core
)
{
	uint32			num_locks_held = 0;
	sl_test_stats_t	* stats = &spinlock_test_core_stats[which_core];
	uint32			op;
	uint32			opFuncIndex;
	uint32			lnum;
	uint32			totry, toget, tounlock;
	uint32			i;
	uint32			step;

	bzero(stats, sizeof(sl_test_stats_t));

	for (i=0; i<SPINLOCK_TEST_NUM_LOCKS; i++)
		spinlock_test_inuse[i] = 0;

	if (which_core == 0) {
		bzero(&spinlock_test_totals, sizeof(sl_test_stats_t));
	
		for (i=0; i< SPINLOCK_TEST_NUM_LOCKS; i++) {
			NFX_SPIN_INIT_DEBUG(&spinlock_test_lock[i], SPINLOCK_LEVEL_DEBUG);
		}

		spinlock_test_int_counter = 0;
		spinlock_test_dword_counter = 0;
	}

	if (which_core == 0) {
		int32	rc;
		smp_remote_function_call(1, 0, (char8 *)spinlock_test);
		rc = smp_remote_function_wait_to_clear(1);
		if (rc) {
			smp_remote_function_clear(1);
			printf("Timed out waiting for core1 to begin spinlock test\n");
			return 0;
		}
		smp_remote_function_call(1, 0, (char8 *)0x11223344);
	}
	else {
		smp_rfunc_t	* rfunc;
		rfunc = smp_remote_function_poll(which_core);
		if (!rfunc) {
			printf("Timed out waiting for core0 to begin spinlock test\n");
			return 0;
		}
	}

	printf("Spinlock Test Start...\n");

	for (op = 0; op < SPINLOCK_TEST_NUM_OPS; op++) {
		opFuncIndex = random() % SPINLOCK_TEST_FUNC_NUM;
		switch (opFuncIndex) {
			case 0:			/* GETLOCK */
			{
				uint32	lockgot[SPINLOCK_TEST_NUM_LOCKS_GET_MAX];
				uint32	locksgot = 0;
				uint32	last_lnum = 0;
				SL_TEST_DEBUG(("   GETLOCK:"));
				toget = (random() % SPINLOCK_TEST_NUM_LOCKS_GET_MAX) + 1;
				for (i=0; i<toget; i++) {
					lnum = random() % SPINLOCK_TEST_NUM_LOCKS;
					if (lnum > last_lnum) {
					if (!spinlock_test_inuse[lnum]) {
						atomic_increment(&spinlock_test_totals.getlock, 4, 1);
						SL_TEST_DEBUG(("  %d", lnum));
						NFX_SPIN_LOCK(&spinlock_test_lock[lnum]);
						stats->getlock++;
						if (spinlock_test_lock[lnum].accumulated_wait_time)
							stats->getlock_busy++;
						else
							stats->getlock_not_busy++;
						spinlock_test_inuse[lnum] = 1;
						stats->getlock_wait_time +=
								spinlock_test_lock[lnum].accumulated_wait_time;
						spinlock_test_lock[lnum].accumulated_wait_time = 0;
						num_locks_held++;
						lockgot[locksgot++] = lnum;
						last_lnum = lnum;
					}
					}
				}
				SL_TEST_DEBUG(("\n"));
				for (i=0; i<locksgot; i++) {
					NFX_SPIN_UNLOCK(&spinlock_test_lock[lockgot[i]]);
					spinlock_test_inuse[lockgot[i]] = 0;
					num_locks_held--;
				}
				break;
			}

			case 1: 		/* UNLOCK */
				SL_TEST_DEBUG(("   UNLOCK:"));
				if (num_locks_held) {
					tounlock = (random() % num_locks_held) + 1;
					for (i=0; i<tounlock; i++) {
						do {
							lnum = random() % SPINLOCK_TEST_NUM_LOCKS;
						} while (!spinlock_test_inuse[lnum]);
						atomic_increment(&spinlock_test_totals.unlock, 4, 1);
						SL_TEST_DEBUG(("  %d", lnum));
						NFX_SPIN_UNLOCK(&spinlock_test_lock[lnum]);
						spinlock_test_inuse[lnum] = 0;
						num_locks_held--;
						stats->unlock++;
					}
				}
				SL_TEST_DEBUG(("\n"));
				break;

			case 2: 		/* TRYLOCK */
			{
				uint32	lockgot[SPINLOCK_TEST_NUM_LOCKS_TRY_MAX];
				uint32	locksgot = 0;
				uint32	last_lnum = 0;
				SL_TEST_DEBUG(("   TRYLOCK:"));
				totry = (random() % SPINLOCK_TEST_NUM_LOCKS_TRY_MAX) + 1;
				for (i=0; i<totry; i++) {
					lnum = random() % SPINLOCK_TEST_NUM_LOCKS;
					if (lnum > last_lnum) {
					if (!spinlock_test_inuse[lnum]) {
						atomic_increment(&spinlock_test_totals.trylock, 4, 1);
						SL_TEST_DEBUG(("  %d", lnum));
						stats->trylock++;
						if (NFX_SPIN_TRYLOCK(&spinlock_test_lock[lnum])) {
							spinlock_test_inuse[lnum] = 1;
							num_locks_held++;
							stats->trylock_passed++;
							lockgot[locksgot++] = lnum;
							last_lnum = lnum;
						}
						else
							stats->trylock_failed++;
					}
					}
				}
				SL_TEST_DEBUG(("\n"));
				for (i=0; i<locksgot; i++) {
					NFX_SPIN_UNLOCK(&spinlock_test_lock[lockgot[i]]);
					spinlock_test_inuse[lockgot[i]] = 0;
					num_locks_held--;
				}
				break;
			}

			case 3: 		/* INC INT */
				SL_TEST_DEBUG(("   INC INT:"));
				step = (random() % 5) + 1;
				SL_TEST_DEBUG(("%d", step));
				atomic_increment(&spinlock_test_totals.inc_int, 4, step);
				atomic_increment(&spinlock_test_int_counter, 4, step);
				stats->inc_int += step;
				SL_TEST_DEBUG(("\n"));
				break;

			case 4: 		/* DEC INT */
				SL_TEST_DEBUG(("   DEC INT:"));
				step = (random() % 5) + 1;
				SL_TEST_DEBUG(("%d", step));
				atomic_increment(&spinlock_test_totals.dec_int, 4, step);
				atomic_decrement(&spinlock_test_int_counter, 4, step);
				stats->dec_int += step;
				SL_TEST_DEBUG(("\n"));
				break;

			case 5: 		/* INC DWORD */
				SL_TEST_DEBUG(("   INC DWORD:"));
				step = (random() % 5) + 1;
				SL_TEST_DEBUG(("%d", step));
				atomic_increment(&spinlock_test_totals.inc_dword, 4, step);
				atomic_increment(&spinlock_test_dword_counter, 8, step);
				stats->inc_dword += step;
				SL_TEST_DEBUG(("\n"));
				break;

			case 6:			/* DEC DWORD */
				SL_TEST_DEBUG(("   DEC DWORD:"));
				step = (random() % 5) + 1;
				SL_TEST_DEBUG(("%d", step));
				atomic_increment(&spinlock_test_totals.dec_dword, 4, step);
				atomic_decrement(&spinlock_test_dword_counter, 8, step);
				stats->dec_dword += step;
				SL_TEST_DEBUG(("\n"));
				break;

			default:
				panic("Invalid opFuncIndex, %d\n", opFuncIndex);
		}
	}

	printf("Spinlock Test Stop...\n");

	for (i=0; i< SPINLOCK_TEST_NUM_LOCKS; i++) {
		if (spinlock_test_inuse[i]) {
			NFX_SPIN_UNLOCK(&spinlock_test_lock[i]);
			spinlock_test_inuse[i] = 0;
			num_locks_held--;
		}
	}

	printf("My Core Stats:\n");
	printf("--------------\n");
	printf("stats->getlock = 0x%x\n", stats->getlock);
	printf("stats->getlock_busy = 0x%x\n", stats->getlock_busy);
	printf("stats->getlock_not_busy = 0x%x\n", stats->getlock_not_busy);
	printf("stats->getlock_wait_time = 0x%x\n", stats->getlock_wait_time);

	printf("stats->unlock = 0x%x\n", stats->unlock);

	printf("stats->trylock = 0x%x\n", stats->trylock);
	printf("stats->trylock_passed = 0x%x\n", stats->trylock_passed);
	printf("stats->trylock_failed = 0x%x\n", stats->trylock_failed);

	printf("stats->inc_int = 0x%x\n", stats->inc_int);
	printf("stats->dec_int = 0x%x\n", stats->dec_int);
	printf("stats->inc_dword = 0x%x\n", stats->inc_dword);
	printf("stats->dec_dword = 0x%x\n", stats->dec_dword);

	return (spinlock_test_verify(which_core));
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	11/12/02
 *
 *-----------------------------------------------------------------
 */
int32
spinlock_test_verify(
	uint32	which_core
)
{
	if (which_core == 0) {
		uint32			i, j;
		sl_test_stats_t	* sptr;
		uint32			sum;
		uint64			dsum;
		sl_test_stats_t	totals;
		uint32			* iptr, * tptr = (uint32 *)&totals;
		uint32			failed = 0;

		/*
		 * Wait until individual cores tell us they're done with test
		 */
		for (i=1; i<SMP_NUM_CPUS; i++) {
			if (smp_remote_function_wait_to_clear(i)) {
				printf("Error waiting for core%d to finish test\n", i);
				failed = 1;
			}
		}

		for (j=0; j<sizeof(sl_test_stats_t)/sizeof(uint32); j++) {
			*(tptr+j) = 0;
			for (i=0; i<SMP_NUM_CPUS; i++) {
				iptr = (uint32 *)&spinlock_test_core_stats[i];
				*(tptr+j) += *(iptr+j);
			}
		}

		/*
		 * Verify summation of individual cores with the totals
		 */
		for (i=0, sum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			sum += sptr->getlock_busy + sptr->getlock_not_busy;
		}
		if (sum != spinlock_test_totals.getlock) {
			printf("Error in spinlock test GETLOCK verify\n");
			failed = 1;
		}

		for (i=0, sum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			sum += sptr->unlock;
		}
		if (sum != spinlock_test_totals.unlock) {
			printf("Error in spinlock test UNLOCK verify\n");
			failed = 1;
		}

		for (i=0, sum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			sum += sptr->trylock_passed + sptr->trylock_failed;
		}
		if (sum != spinlock_test_totals.trylock) {
			printf("Error in spinlock test TRYLOCK verify\n");
			failed = 1;
		}

		for (i=0, sum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			sum += sptr->inc_int;
		}
		if (sum != spinlock_test_totals.inc_int) {
			printf("Error in spinlock test INC_INT verify\n");
			failed = 1;
		}

		for (i=0, dsum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			dsum += sptr->inc_dword;
		}
		if (dsum != spinlock_test_totals.inc_dword) {
			printf("Error in spinlock test INC_DWORD verify\n");
			failed = 1;
		}

		for (i=0, sum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			sum += sptr->dec_int;
		}
		if (sum != spinlock_test_totals.dec_int) {
			printf("Error in spinlock test DEC_INT verify\n");
			failed = 1;
		}

		for (i=0, dsum=0; i<SMP_NUM_CPUS; i++) {
			sptr = &spinlock_test_core_stats[i];
			dsum += sptr->dec_dword;
		}
		if (dsum != spinlock_test_totals.dec_dword) {
			printf("Error in spinlock test DEC_DWORD verify\n");
			failed = 1;
		}

		if (spinlock_test_totals.inc_int - spinlock_test_totals.dec_int != spinlock_test_int_counter) {
			printf("Error in atomic int ops, inc_int = 0x%x, dec_int = 0x%x, counter = 0x%x\n", spinlock_test_totals.inc_int, spinlock_test_totals.dec_int, spinlock_test_int_counter);
			failed = 1;
		}

		if (spinlock_test_totals.inc_dword - spinlock_test_totals.dec_dword != (uint32)spinlock_test_dword_counter) {
			printf("Error in atomic dword ops, inc_dword = 0x%x, dec_dword = 0x%x, counter = 0x%llx\n", spinlock_test_totals.inc_dword, spinlock_test_totals.dec_dword, spinlock_test_dword_counter);
			failed = 1;
		}

		printf("\nTOTALS:");
		printf("\n--------------\n");

		printf("getlock           = 0x%x\n", totals.getlock);
		printf("getlock_busy      = 0x%x\n", totals.getlock_busy);
		printf("getlock_not_busy  = 0x%x\n", totals.getlock_not_busy);
		printf("getlock_wait_time = 0x%x\n", totals.getlock_wait_time);

		printf("unlock            = 0x%x\n", totals.unlock);

		printf("trylock           = 0x%x\n", totals.trylock);
		printf("trylock_passed    = 0x%x\n", totals.trylock_passed);
		printf("trylock_failed    = 0x%x\n", totals.trylock_failed);

		printf("inc_int           = 0x%x\n", totals.inc_int);
		printf("dec_int           = 0x%x\n", totals.dec_int);
		printf("inc_dword         = 0x%x\n", totals.inc_dword);
		printf("dec_dword         = 0x%x\n", totals.dec_dword);

		printf("int counter       = 0x%x\n", spinlock_test_int_counter);
		printf("dword counter     = 0x%llx\n", spinlock_test_dword_counter);

		if (failed)
			printf("\nSPINLOCK TEST FAILED\n");
		else
			printf("\nspinlock test passed\n");
	}
	else {
		/*
		 * Tell core0 we're done with test
		 */
		smp_remote_function_clear(which_core);
	}

	return 0;
}

#else


#endif /* NFX_SPINLOCK_DEBUG */

#endif	/* NFX_SMP */

