/*------------------------------------------------------------------------------
 * RCS:  $Id: spm-msg.c,v 1.33 2002/11/09 00:07:51 nitinb Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : spm-msg.c
 *
 * Description  : Entry module of the Storage Pool Manager (SPM).
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-FEB-200
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <sys/time.h>

#include "nfx-incl.h"
#include "spm.h"
#include <exports/ncm-api.h>
#include <exports/event-api.h>

/*
 * Global variables.
 */

/*
 * SPM application id.
 */

int32 spmAppId;
int32 spmDesc;
spm_state_t spmInitStat;
char spmNodeName[SPM_MAX_NODE_NAME_LEN];
spm_timer_list_t spmTimerList;
spm_msgTxn_t spm_txnIndex[SPM_MAX_PENDING_RSPS];

int32 spm_alloc_cnt;
int32 spm_free_cnt;
uint32 dbg_level;

extern int32 nfxDbNodes;

static struct timeval wait_to = {5, 0};
static struct timeval disc_to = {1, 0};

/*
 * Function prototypes.
 */
static void spm_initTxnIndex(void);

static int32 spm_allocateTxnIndex(void);

static int32 spm_freeTxnIndex(uint32 magicC);

static spm_msgTxn_t *spm_getTxnIndex(uint32 magicC);

#ifdef SPM_EA_AV_API_READY
static int32 spm_execPendingAllocReq(spm_evmRcvCB_t *spmCb,
                                     spm_MsgAllocLunReq_t *req);
#endif

static void spm_processRspMsg(char *buf);

static void spm_processMsg(void);

static int32 spm_addLun(spm_MsgAddLunReq_t *addReq, spm_dev_t devType);

static void spm_delLun(scsi_id_t sid, char *node);

static int32 spm_addLunDetailRec(scsi_id_t s_id, char *node,
                                 int32 checkState, int64 magicC);

static void spm_discoverLun(char *node);

static void spm_discoverTape(char *node);

static void spm_discoverChanger(char *node);

static spm_MsgAddLunReq_t *spm_getDevStatus(scsi_id_t sid, char *node);

static int32 spm_sendLabel_req(spm_msgTxn_t *txn, int32 reqType, char *cluster);

static void spm_askLunRescanOnNodes(scsi_id_t sid, char *node);

#ifdef SPM_EA_AV_API_READY
static int32 spm_sendAllocInd_msg(uint64 idCnt,
                                  scsi_id_t *scsiList,
                                  spm_MsgAllocLunReq_t *req,
                                  spm_allocStatus_t result);
#endif

static void spm_sendErrRspTO(int64 p1, int64 p2);

static int32 spm_procEvmLabelRsp(spm_msgTxn_t *txn, char *evmRsp, char *node);

static int32 spm_procSdmRsp(spm_msgTxn_t *txn, char *sdmRsp, char *node);

#ifdef SPM_EA_AV_API_READY
static int32 spm_procEaFsysGrowRsp(spm_msgTxn_t *txn, char *eaRsp, char *node);
#endif

static void spm_processPccEvent(spm_msgTxn_t *txn);
/*
 * Local functions.
 */

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_initTxnIndex
 *
 * Description  : Init Txn index
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 30-APR-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_initTxnIndex()
{
    int32 cnt;
    struct timeval stTime;
    uint32 baseCtr = 0;

    SPM_FCT_PRINT(("spm_initTxnIndex..."));
    bzero(&spm_txnIndex, sizeof(spm_msgTxn_t)*SPM_MAX_PENDING_RSPS);
    bzero(&stTime, sizeof(struct timeval));

    gettimeofday(&stTime, NULL);
    // Just get the multiple of SPM_MAX_PENDING_RSPS.
    baseCtr = stTime.tv_sec - stTime.tv_sec%SPM_MAX_PENDING_RSPS;
    for (cnt = 0; cnt < SPM_MAX_PENDING_RSPS; cnt++)
        spm_txnIndex[cnt].magicCookie = baseCtr + cnt;

    SPM_FCT_PRINT(("...spm_initTxnIndex"));
    return;
}

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_allocateTxnIndex
 *
 * Description  : Allocate Txn index
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 30-APR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_allocateTxnIndex()
{
    int32 cnt;

    for (cnt = 0; cnt < SPM_MAX_PENDING_RSPS; cnt++)
    {
        if (!spm_txnIndex[cnt].inUse)
        {
            spm_txnIndex[cnt].inUse = 1;
            spm_txnIndex[cnt].magicCookie += SPM_MAX_PENDING_RSPS;
            SPM_DBG_PRINT(("spm_allocateTxnIndex [%d] : %d", cnt,
                                   spm_txnIndex[cnt].magicCookie));
            return cnt;
        }
    }
    return -1;
}

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_freeTxnIndex
 *
 * Description  : Free Txn index
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 30-APR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_freeTxnIndex(uint32 magicC)
{
    int32 idx = (int32)(magicC%SPM_MAX_PENDING_RSPS);

    if (spm_txnIndex[idx].inUse && spm_txnIndex[idx].magicCookie == magicC)
    {
        SPM_DBG_PRINT(("spm_freeTxnIndex [%d] : %d from %d", idx,
                                   spm_txnIndex[idx].magicCookie, magicC));
        spm_txnIndex[idx].inUse = 0;
        (spm_txnIndex[idx].tmr)?spm_procTimerStop(spm_txnIndex[idx].tmr):0;
        if (spm_txnIndex[idx].sender)
            SPM_FREE(spm_txnIndex[idx].sender);
        if (spm_txnIndex[idx].msg)
            SPM_FREE(spm_txnIndex[idx].msg);
        spm_txnIndex[idx].msg = NULL;
        spm_txnIndex[idx].sender = NULL;
        spm_txnIndex[idx].callback = NULL;
        spm_txnIndex[idx].tmr = NULL;
        spm_txnIndex[idx].numReqs = 0;
        spm_txnIndex[idx].numRsps = 0;
        spm_txnIndex[idx].data1 = NULL;
        spm_txnIndex[idx].data2 = NULL;
        /*
         * Do not reset the magic cookie. It will be used as the base for
         * next allocation.
         */
    }
    else
        idx = 0;

    return idx;
}

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_getTxnIndex
 *
 * Description  : get Txn index
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 30-APR-2002
 *
 *------------------------------------------------------------------------------
 */

static spm_msgTxn_t *spm_getTxnIndex(uint32 magicC)
{
    int32 idx = (int32)(magicC%SPM_MAX_PENDING_RSPS);

    SPM_DBG_PRINT(("spm_getTxnIndex [%d] : %d from %d", idx,
                                   spm_txnIndex[idx].magicCookie, magicC));
    if (spm_txnIndex[idx].inUse && spm_txnIndex[idx].magicCookie == magicC)
        return (&spm_txnIndex[idx]);
    else
        return NULL;
}

#ifdef SPM_EA_AV_API_READY
/*  
 *------------------------------------------------------------------------------
 * Name         : spm_execPendingAllocReq
 *
 * Description  : Execute pending alloc request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 14-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_execPendingAllocReq(spm_evmRcvCB_t *spmCb,
                                     spm_MsgAllocLunReq_t *req)
{
    scsi_id_t *sList = NULL;
    int64     idCnt = 0;
    int32     retCode = 0, rc;
    uint64    allocSize = 0;

    SPM_FCT_PRINT(("spm_execPendingAllocReq..."));

    rc = spm_getFree_lunList(req->minSize, &req->attr,
                             req->hdr.node, &sList, &idCnt, &allocSize, 0);
    if (rc == SPM_ALLOC_OK && sList)
    {   
        SPM_DBG_PRINT(("Found luns......"));
        SPM_DBG_PRINT(("\tNumber : %lld", idCnt));
        SPM_DBG_PRINT(("\tLun ID : 0x%llx", sList[0]));
        /*  
         * Send indication to EVM for LUNs
         */
        rc = spm_sendAllocInd_msg(idCnt, sList, req, SPM_ALLOC_OK);
        if (rc == NFX_ERR)
            rc = spm_free_lunList(idCnt, sList);
        else
            retCode = 1;
    }
    else
    {
        SPM_DBG_PRINT(("Found luns...0"));
        rc = spm_sendAllocInd_msg(0, NULL, req, SPM_ALLOC_FAILED);
    }

    if (sList) SPM_FREE(sList);

    SPM_FCT_PRINT(("...spm_execPendingAllocReq"));
    return retCode;
}
#endif

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_sendErrRspTO
 *
 * Description  : Process Time Out for a pending request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 09-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_sendErrRspTO(int64 p1, int64 p2)
{
    spm_msgTxn_t *txn = spm_getTxnIndex(p1);
    spm_senderInfo_t *sender;
    int32 *mid;

    /* Only deal with cases where we start the timer */
    if (txn)
    {
        sender = txn->sender;
        mid = (int32 *)txn->msg;
        switch(*mid)
        {
            case SPM_MARK_LUN_AS_AGILE_REQ:
            case SPM_MARK_LUN_AS_OC_REQ:
            case SPM_UNMARK_LUN_AS_AGILE_REQ:
            {
                spm_MsgMarkAgileLunReq_t *req = txn->msg;
                spm_MsgMarkAgileLunRsp_t rsp;
                int32 nCnt;
                DB    *tDb;
                spm_scsiNode_t **snList = txn->data1;

                bzero(&rsp, sizeof(spm_MsgMarkAgileLunRsp_t));
                rsp.hdr.msgType = (req->hdr.msgType==SPM_MARK_LUN_AS_AGILE_REQ)
                                     ?SPM_MARK_LUN_AS_AGILE_RSP
                                     :(req->hdr.msgType==SPM_MARK_LUN_AS_OC_REQ)
                                          ?SPM_MARK_LUN_AS_OC_RSP
                                          :SPM_UNMARK_LUN_AS_AGILE_RSP;
                rsp.hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(rsp.hdr.node, req->hdr.node);
                rsp.status = SPM_NOK;

                sendAgileMsg(-1, &rsp, sizeof(spm_MsgMarkAgileLunRsp_t),
                             (req->hdr.spare)?"ea":sender->name,
                             sender->cpu, sender->slot, sender->plane,
                             NULL, NULL, 0);

                for (nCnt = 0; nCnt < txn->numReqs; nCnt++)
                {
                    tDb = snList[nCnt]->sid;
                    if (tDb) tDb->close(tDb);
                    SPM_FREE(snList[nCnt]);
                }
                SPM_FREE(snList);
            }
            break;

#ifdef SPM_EA_AV_API_READY
            case SPM_ALLOC_LUN_REQ:
            {
                spm_MsgAllocLunReq_t *req = (spm_MsgAllocLunReq_t *)txn->msg;
                spm_evmRcvCB_t *spmCb = (spm_evmRcvCB_t *)txn->data1;

                spm_sendAllocInd_msg(0, NULL, req, SPM_ALLOC_FAILED);
                if (spmCb)
                {
                    spm_unReserveLun(spmCb->lunReserved, spmCb->reservedSid);
                    if (spmCb->reservedSid) SPM_FREE(spmCb->reservedSid);
                    SPM_FREE(spmCb);
                }
            }
            break;
#endif

            case SPM_GET_DEV_DETAIL_REQ:
            {
                spm_MsgDevDetailReq_t *req = (spm_MsgDevDetailReq_t *)txn->msg;
                spm_MsgDevDetailRsp_t rsp;

                bzero(&rsp, sizeof(spm_MsgDevDetailRsp_t));
                rsp.hdr.msgType = SPM_GET_DEV_DETAIL_RSP;
                rsp.hdr.clientPrivate = req->hdr.clientPrivate;
                rsp.result = NFX_ERR;
                sendAgileMsg(-1, &rsp, sizeof(spm_MsgDevDetailRsp_t),
                             (req->hdr.spare)?"ea":sender->name,
                             sender->cpu, sender->slot, sender->plane,
                             NULL, NULL, 0);
            }
            break;
        }
        spm_freeTxnIndex(p1);
    }

    return;
}

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_addLun
 *
 * Description  : Process add LUN request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_addLun(spm_MsgAddLunReq_t *addReq, spm_dev_t devType)
{
    int32          poolId;
    char           err[256];
    spm_lunState_t state;

    SPM_FCT_PRINT(("spm_addLun..."));

    /*
     * Check is the nfxDb already has knowledge about this LUN.
     */
    if (!spm_checkLunInNfxDb(addReq->hdr.node, addReq->lunAttr.scsiId))
    {
        /*
         * This is the first time this node sees the LUN
         */
        spm_addLunToNfxDb(addReq->hdr.node, addReq->lunAttr.scsiId);

        switch(addReq->lunAttr.state)
        {
            case EVM_FREE_LUN :
                state = SPM_LUN_FREE;
                break;
            case EVM_USED_LUN :
                state = SPM_LUN_USED;
                break;
            case EVM_OC_FREE_LUN :
            case EVM_OC_USED_LUN :
                state = SPM_LUN_OC;
                break;
            case EVM_FOREIGN_LUN :
                state = SPM_LUN_FOREIGN;
                break;
        }
        /*
         * Check lunDb and scsiDb if we know anything about the LUN
         */
        /* Check with frank about device name */
        // if (spm_checkLunInLnSdb(NULL, addReq->lunAttr.scsiId))
        if (spm_checkLunInLnSdb(addReq->lunAttr.name, 0))
        {
            spm_lunDbCb_t    *lCfg = NULL;
            /*
             * We know about this LUN. It is seen by more than one node.
             */
            // spm_getScsiDb_rec(addReq->lunAttr.scsiId, &lCfg);
            spm_getLunDb_rec(addReq->lunAttr.name, &lCfg);
            NFX_ASSERT(lCfg != NULL)
            if (lCfg)
                poolId = lCfg->poolId;
            /*
             * Add lun to poolDb
             */
            if (spm_add_lunTo_poolDb(poolId, addReq->lunAttr.scsiId, 0,
                                  addReq->lunAttr.size, state, err) == NFX_ERR)
            {
                spm_delLunFromNfxDb(addReq->hdr.node, addReq->lunAttr.scsiId);
                return NFX_ERR;
            }
        }
        else
        {
            spm_poolAttr_t  poolAttr;
            spm_lunDbCb_t   *lCfg = NULL;
            char            ctrl[SDM_DEV_NAME_LEN+1];
            int32           i, done;

            /*
             * This is the first time we know about this LUN.
             */
            bzero (&poolAttr, sizeof(spm_poolAttr_t));
            poolAttr.mask = SPM_POOL_ATTR_ALL_MASK;
            poolAttr.raidLevel = addReq->lunAttr.raidLevel;
            // poolAttr.mask &= ~SPM_POOL_ATTR_MODEL_MASK; // Remove model mask
            poolAttr.type = devType;
            if (strlen(addReq->lunAttr.name))
            {
                bzero(ctrl, SDM_DEV_NAME_LEN+1);  // ctrl is null terminated
                for (i = 0, done = 0;
                     i < strlen(addReq->lunAttr.name) && done < 2;
                     i++)
                {
                    if (addReq->lunAttr.name[i] == '_')
                        done++;
                    if (done != 2)
                        ctrl[i] = addReq->lunAttr.name[i];
                }
                SPM_DBG_PRINT(("\tController name : %s", ctrl));
            }
            strcpy(poolAttr.ctrl, ctrl);
            strcpy(poolAttr.name, addReq->lunAttr.vendor);
            strcpy(poolAttr.model, addReq->lunAttr.model);
            poolId = spm_get_pool_by_attrs(&poolAttr, 0, 0, err);

            if (poolId < 0)
            {
                /*
                 * No matching pool. Create one
                 */
                if (spm_new_pool(&poolAttr, &poolId, err) == NFX_ERR)
                {
                    spm_delLunFromNfxDb(addReq->hdr.node,
                                        addReq->lunAttr.scsiId);
                    return NFX_ERR;
                }
            }

            /*
             * Add LUN to scsiDb and LunDb
             */

            if (spm_addLunToLnSdb(&addReq->lunAttr, poolId,
                                  addReq->hdr.node, err) == NFX_ERR)
            {
                spm_delLunFromNfxDb(addReq->hdr.node, addReq->lunAttr.scsiId);
                spm_del_pool(poolId);
                return NFX_ERR;
            }
            /*
             * Add lun to poolDb
             */
            if (spm_add_lunTo_poolDb(poolId, addReq->lunAttr.scsiId, 1,
                                  addReq->lunAttr.size, state, err) == NFX_ERR)
            {
                spm_delLunFromNfxDb(addReq->hdr.node, addReq->lunAttr.scsiId);
                spm_del_pool(poolId);
                spm_getScsiDb_rec(addReq->lunAttr.scsiId, &lCfg);
                spm_delLunFromLnSdb(lCfg, err); // lCfg will be freed
                return NFX_ERR;
            }
        }

        if (devType == SPM_DEV_DISK)
            spm_sendNewLunInd(&addReq->lunAttr, addReq->hdr.node);
    }
    else
    {
        /*
         * Since this is a duplicate message, we will take it as the most
         * upto date info
         */
        SPM_DBG_PRINT(("Update info for duplicate ADD message (%llx)",
                                                       addReq->lunAttr.scsiId));
        spm_updateLunLabelInd(&addReq->lunAttr, addReq->hdr.node, 1, devType);
        /* spm_updateLunLabelInd will send a new LUN indication to SRM */
    }

    SPM_FCT_PRINT(("...spm_addLun"));

    return NFX_OK;
}

/*  
 *------------------------------------------------------------------------------
 * Name         : spm_delLun
 *
 * Description  : Process add LUN request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_delLun(scsi_id_t sid, char *node)
{
    char          err[256];
    spm_lunDbCb_t *lCfg;

    SPM_FCT_PRINT(("spm_delLun..."));

    spm_delLunFromNfxDb(node, sid);

    spm_getScsiDb_rec(sid, &lCfg);
    if (lCfg)
    {
        if (!(spm_delLunFrm_poolDb(lCfg->poolId, sid, err) > 0))
        {
            spm_sendDelLunInd(lCfg->dev, node);
            /* Delete from lunDb and scsiDb */
            spm_delLunFromLnSdb(lCfg, err);
        }
    }

    SPM_FCT_PRINT(("...spm_delLun"));
    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_addLunDetailRec
 *
 * Description  : Get details for the LUN from EVM and add it to the DB. During
 *                INIT, we need to block and take care of things before we can
 *                process any other request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 03-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_addLunDetailRec(scsi_id_t s_id, char *node,
                                 int32 checkState, int64 cookie)
{
    ssize_t               ncmReqSize;
    ssize_t               ncmRspSize;
    ncmMsg_t              *ncmReq = NULL;
    ncmMsg_t              *ncmRsp = NULL;
    evm_MsgGetLunReq_t    *evmReq;
    evm_MsgGetLunRsp_t    *evmRsp;
    int32                 rc;

    SPM_FCT_PRINT(("spm_addLunDetailRec..."));
    SPM_DBG_PRINT(("\tSCSI-ID : 0x%llx", s_id));
    ncmReqSize = sizeof(ncmMsg_t) + sizeof(evm_MsgGetLunReq_t);
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);

    ncmRspSize = sizeof(ncmMsg_t) + sizeof(evm_MsgGetLunRsp_t)
                                  - sizeof(long long);
    SPM_ALLOC_ZERO(ncmRsp, ncmMsg_t, ncmRspSize);

    ncmReq->msgType = NCM_SEND;
    strcpy(ncmReq->nodeName, node);
    ncmReq->rqResp = 1;
    evmReq = (evm_MsgGetLunReq_t *)ncmReq->appMsg;
    evmReq->hdr.msgType = EVM_GET_LUN_REQ;
    evmReq->hdr.clientPrivate = cookie;
    evmReq->id = s_id;
    strcpy(ncmReq->appName, "evm");

    rc = sendAgileMsg(-1, ncmReq, ncmReqSize,
                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &disc_to, ncmRsp, ncmRspSize);

    if (rc < ncmRspSize && ncmRsp->msgType != NCM_RESP)
    {
        SPM_ERR_PRINT(("EVM could not return details for %llx",s_id));
        return 0;
    }

    evmRsp = (evm_MsgGetLunRsp_t *)ncmRsp->appMsg;
    if (evmRsp != NULL && evmRsp->hdr.msgType == EVM_GET_LUN_RSP)
    {
        if (checkState)
        {
            spm_lunDbCb_t  *lCfg = NULL;
            spm_lunState_t curSt, toSt;
            /*
             * We should know about this LUN. If the state has changed
             * issue a change of state
             */
            spm_getScsiDb_rec(evmRsp->id, &lCfg);
            if (lCfg)
            {
                curSt = spm_getLunState(lCfg->poolId, evmRsp->id);
                switch(evmRsp->state)
                {
                    case EVM_FREE_LUN :
                        toSt = SPM_LUN_FREE;
                        break;
                    case EVM_USED_LUN :
                        toSt = SPM_LUN_USED;
                        break;
                    case EVM_OC_FREE_LUN :
                    case EVM_OC_USED_LUN :
                        toSt = SPM_LUN_OC;
                        break;
                    case EVM_FOREIGN_LUN :
                        toSt = SPM_LUN_FOREIGN;
                        break;
                }
                if (toSt != curSt)
                {
                    spm_moveLunState(lCfg->poolId, evmRsp->id, curSt, toSt);
                    lCfg->lstate = evmRsp->state;
                    strcpy(lCfg->cluster, evmRsp->cluster);
                }
            }
        }
        else
        {
            /* Add LUN record */
            spm_MsgAddLunReq_t    spmReq;

            bzero(&spmReq, sizeof(spm_MsgAddLunReq_t));
            strcpy(spmReq.hdr.node, node); // Node where we requested info.
            spmReq.lunAttr.scsiId = evmRsp->id;
            spmReq.lunAttr.wwn = evmRsp->wwn;
            spmReq.lunAttr.lun = evmRsp->lun;
            spmReq.lunAttr.raidLevel = evmRsp->raidLevel;
            strcpy(spmReq.lunAttr.name, evmRsp->name);
            strcpy(spmReq.lunAttr.vendor, evmRsp->vendor);
            strcpy(spmReq.lunAttr.model, evmRsp->model);
            strcpy(spmReq.lunAttr.cluster, evmRsp->cluster);
            spmReq.lunAttr.size = evmRsp->size;
            spmReq.lunAttr.state = evmRsp->state;
            spmReq.lunAttr.devType = SCSI_DEVTYPE_DISK;
            spm_addLun(&spmReq, SPM_DEV_DISK);
        }
    }

    SPM_FREE(ncmReq);
    SPM_FREE(ncmRsp);
    SPM_FCT_PRINT(("...spm_addLunDetailRec"));

    return 1;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_discoverLun
 *
 * Description  : Get list of LUN's from SDM. During INIT, we need to block
 *                and take care of things before we can process any other
 *                request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 03-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_discoverLun(char *node)
{
    ssize_t               ncmReqSize;
    ssize_t               ncmRspSize;
    ncmMsg_t              *ncmReq = NULL;
    ncmMsg_t              *ncmRsp = NULL;
    sdm_devListReqMsg_t   *sdmReq = NULL;
    sdm_devListRspMsg_t   *sdmRsp = NULL;
    int32                 rc, cnt, numLuns = 0, evmRdy = 1;
    int64                 pid = getpid();

    SPM_FCT_PRINT(("spm_discoverLun..."));
    ncmReqSize = sizeof(ncmMsg_t) + sizeof(sdm_devListReqMsg_t);
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);

    ncmRspSize = sizeof(ncmMsg_t) + SPM_MAX_BUF_SIZE;
    SPM_ALLOC_ZERO(ncmRsp, ncmMsg_t, ncmRspSize);

    ncmReq->msgType = NCM_SEND;
    strcpy(ncmReq->nodeName, node);
    ncmReq->rqResp = 1;
    sdmReq = (sdm_devListReqMsg_t *)ncmReq->appMsg;
    sdmReq->hdr.mid = SDM_MID_DEV_LIST_REQ;
    sdmReq->hdr.chandle = pid;
    sdmReq->type = SDM_DEVLIST_DISK;
    sdmReq->port = SDM_DEVLIST_PORTALL;

    strcpy(ncmReq->appName, "sdm");
    rc = sendAgileMsg(-1, ncmReq, ncmReqSize,
                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &wait_to, ncmRsp, ncmRspSize);
    if (rc <= 0)
    {
        SPM_ERR_PRINT(("Timeout : No response from SDM"));
    }

    if (rc > 0)
    {
        sdmRsp = (sdm_devListRspMsg_t *)ncmRsp->appMsg;
        numLuns = sdmRsp->numEntries;
    }
    SPM_DBG_PRINT(("Number of SDM Lun's : %d", numLuns));
    for (cnt = 0; cnt < numLuns && sdmRsp != NULL && evmRdy; cnt++)
    {
        spm_askLunRescanOnNodes(sdmRsp->idList[cnt], "");
        evmRdy = spm_addLunDetailRec(sdmRsp->idList[cnt], node, 0, pid);
    }

    /*
     * if the response was not correct, the sdm module may not have
     * initialized. In which case, EVM will report new LUN's via ADD message
     */

    SPM_FCT_PRINT(("...spm_discoverLun"));
    SPM_FREE(ncmReq);
    SPM_FREE(ncmRsp);

    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_discoverTape
 *
 * Description  : Get list of LUN's from SDM. During INIT, we need to block
 *                and take care of things before we can process any other
 *                request.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 03-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_discoverTape(char *node)
{
    ssize_t               ncmReqSize;
    ssize_t               ncmRspSize;
    ncmMsg_t              *ncmReq = NULL;
    ncmMsg_t              *ncmRsp = NULL;
    sdm_devListReqMsg_t   *sdmReq = NULL;
    sdm_devListRspMsg_t   *sdmRsp = NULL;
    int32                 rc, cnt, numLuns = 0, evmRdy = 1;
    int64                 pid = getpid();
    spm_MsgAddLunReq_t    *spmReq;

    SPM_FCT_PRINT(("spm_discoverTape..."));
    ncmReqSize = sizeof(ncmMsg_t) + sizeof(sdm_devListReqMsg_t);
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);

    ncmRspSize = sizeof(ncmMsg_t) + SPM_MAX_BUF_SIZE;
    SPM_ALLOC_ZERO(ncmRsp, ncmMsg_t, ncmRspSize);

    ncmReq->msgType = NCM_SEND;
    strcpy(ncmReq->nodeName, node);
    ncmReq->rqResp = 1;
    sdmReq = (sdm_devListReqMsg_t *)ncmReq->appMsg;
    sdmReq->hdr.mid = SDM_MID_DEV_LIST_REQ;
    sdmReq->hdr.chandle = pid;
    sdmReq->type = SDM_DEVLIST_TAPE;
    sdmReq->port = SDM_DEVLIST_PORTALL;

    strcpy(ncmReq->appName, "sdm");
    rc = sendAgileMsg(-1, ncmReq, ncmReqSize,
                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &wait_to, ncmRsp, ncmRspSize);
    if (rc <= 0)
    {
        SPM_ERR_PRINT(("Timeout : No response from SDM"));
    }

    if (rc > 0)
    {
        sdmRsp = (sdm_devListRspMsg_t *)ncmRsp->appMsg;
        numLuns = sdmRsp->numEntries;
    }
    SPM_DBG_PRINT(("Number of SDM Lun's : %d", numLuns));
    for (cnt = 0; cnt < numLuns && sdmRsp != NULL && evmRdy; cnt++)
    {
        /*
         * Ask SDM to send details about each device.
         */
        spmReq = spm_getDevStatus(sdmRsp->idList[cnt], node);
        spmReq->lunAttr.devType = SCSI_DEVTYPE_TAPE;
        spm_addLun(spmReq, SPM_DEV_TAPE);
        SPM_FREE(spmReq);
    }

    SPM_FCT_PRINT(("...spm_discoverTape"));
    SPM_FREE(ncmReq);
    SPM_FREE(ncmRsp);

    return;
}

static void spm_discoverChanger(char *node)
{
    ssize_t               ncmReqSize;
    ssize_t               ncmRspSize;
    ncmMsg_t              *ncmReq = NULL;
    ncmMsg_t              *ncmRsp = NULL;
    sdm_devListReqMsg_t   *sdmReq = NULL;
    sdm_devListRspMsg_t   *sdmRsp = NULL;
    int32                 rc, cnt, numLuns = 0, evmRdy = 1;
    int64                 pid = getpid();
    spm_MsgAddLunReq_t    *spmReq;

    SPM_FCT_PRINT(("spm_discoverChanger..."));
    ncmReqSize = sizeof(ncmMsg_t) + sizeof(sdm_devListReqMsg_t);
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);

    ncmRspSize = sizeof(ncmMsg_t) + SPM_MAX_BUF_SIZE;
    SPM_ALLOC_ZERO(ncmRsp, ncmMsg_t, ncmRspSize);

    ncmReq->msgType = NCM_SEND;
    strcpy(ncmReq->nodeName, node);
    ncmReq->rqResp = 1;
    sdmReq = (sdm_devListReqMsg_t *)ncmReq->appMsg;
    sdmReq->hdr.mid = SDM_MID_DEV_LIST_REQ;
    sdmReq->hdr.chandle = pid;
    sdmReq->type = SDM_DEVLIST_STACKER;
    sdmReq->port = SDM_DEVLIST_PORTALL;

    strcpy(ncmReq->appName, "sdm");
    rc = sendAgileMsg(-1, ncmReq, ncmReqSize,
                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &wait_to, ncmRsp, ncmRspSize);
    if (rc <= 0)
    {
        SPM_ERR_PRINT(("Timeout : No response from SDM"));
    }

    if (rc > 0)
    {
        sdmRsp = (sdm_devListRspMsg_t *)ncmRsp->appMsg;
        numLuns = sdmRsp->numEntries;
    }
    SPM_DBG_PRINT(("Number of SDM Lun's : %d", numLuns));
    for (cnt = 0; cnt < numLuns && sdmRsp != NULL && evmRdy; cnt++)
    {
        /*
         * Ask SDM to send details about each device.
         */
        spmReq = spm_getDevStatus(sdmRsp->idList[cnt], node);
        spmReq->lunAttr.devType = SCSI_DEVTYPE_MEDIA_CHANGER;
        spm_addLun(spmReq, SPM_DEV_CHANGER);
        SPM_FREE(spmReq);
    }

    SPM_FCT_PRINT(("...spm_discoverChanger"));
    SPM_FREE(ncmReq);
    SPM_FREE(ncmRsp);

    return;
}


static spm_MsgAddLunReq_t *spm_getDevStatus(scsi_id_t sid, char *node)
{
    ssize_t               ncmReqSize;
    ssize_t               ncmRspSize;
    sdm_devStatusReqMsg_t *devStatusReq = NULL;
    sdm_devStatusRspMsg_t *devStatusRsp = NULL;
    ncmMsg_t              *ncmReq = NULL;
    ncmMsg_t              *ncmRsp = NULL;
    int32                 rc;
    int64                 pid = getpid();
    spm_MsgAddLunReq_t    *spmReq = NULL;

    SPM_FCT_PRINT(("spm_getDevStatus..."));

    ncmReqSize = sizeof(ncmMsg_t) + sizeof(sdm_devStatusReqMsg_t);
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);

    ncmRspSize = sizeof(ncmMsg_t) + sizeof(sdm_devStatusRspMsg_t);
    SPM_ALLOC_ZERO(ncmRsp, ncmMsg_t, ncmRspSize);

    ncmReq->msgType = NCM_SEND;
    strcpy(ncmReq->nodeName, node);
    ncmReq->rqResp = 1;
    devStatusReq = (sdm_devStatusReqMsg_t *)ncmReq->appMsg;
    devStatusReq->hdr.mid = SDM_MID_DEV_STATUS_REQ;
    devStatusReq->hdr.chandle = pid;
    devStatusReq->id = sid;

    strcpy(ncmReq->appName, "sdm");
    rc = sendAgileMsg(-1, ncmReq, ncmReqSize,
                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &wait_to, ncmRsp, ncmRspSize);
    if (rc <= 0)
        SPM_ERR_PRINT(("Timeout : No response from SDM"));
    else if (rc > 0)
    {
        devStatusRsp = (sdm_devStatusRspMsg_t *)ncmRsp->appMsg;
        if (devStatusRsp)
        {
            /* Add LUN record */
            SPM_ALLOC_ZERO(spmReq, spm_MsgAddLunReq_t,
                                   sizeof(spm_MsgAddLunReq_t));
            strcpy(spmReq->hdr.node, node); // Node where we requested info.
            spmReq->lunAttr.scsiId = sid;
            spmReq->lunAttr.wwn = devStatusRsp->paths[0].wwn;
            spmReq->lunAttr.lun = devStatusRsp->paths[0].lun;
            spmReq->lunAttr.raidLevel = devStatusRsp->raidLevel;
            strcpy(spmReq->lunAttr.name, devStatusRsp->st.name);
            strcpy(spmReq->lunAttr.vendor, devStatusRsp->st.vendor);
            strcpy(spmReq->lunAttr.model, devStatusRsp->st.model);
            strcpy(spmReq->lunAttr.cluster, currCluster);
            spmReq->lunAttr.size = 1;
            spmReq->lunAttr.state = EVM_FREE_LUN;
        }
    }

    SPM_FCT_PRINT(("...spm_getDevStatus"));

    return spmReq;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_sendLabel_req
 *
 * Description  : Send label/unlabel/outcluster request to EVM
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_sendLabel_req(spm_msgTxn_t *txn, int32 reqType,
                               char *cluster)
{
    ncmMsg_t               *ncmReq;
    int32                  ncmReqSize, ncmSize;
    evm_MsgLabelLunReq_t   *evmLReq;
    evm_MsgUnlabelLunReq_t *evmUnLReq;
    evm_MsgModLunClusterReq_t *evmModReq;
    int32                  rc;
    int32                  retCode = NFX_ERR;
    int32                  nCnt = 0, tCnt, stat;
    spm_scsiNode_t         **snList = txn->data1;
    DB                     *tDb;
    DBT                    key, rec;
    
    SPM_FCT_PRINT(("spm_sendLabel_req..."));

    SPM_DBG_PRINT(("EVM label request:"));
    if (reqType == EVM_LABEL_LUN_REQ)
        ncmSize = sizeof(ncmMsg_t) + sizeof(evm_MsgLabelLunReq_t);
    else if (reqType == EVM_MOD_LUN_CLUSTER_REQ)
        ncmSize = sizeof(ncmMsg_t) + sizeof(evm_MsgModLunClusterReq_t);
    else
        ncmSize = sizeof(ncmMsg_t) + sizeof(evm_MsgUnlabelLunRsp_t);

    for (; nCnt < txn->numReqs; nCnt++)
    {
        SPM_DBG_PRINT(("\t Node : %s", snList[nCnt]->node));
        ncmReqSize = ncmSize + sizeof(scsi_id_t)*(snList[nCnt]->sCnt - 1);

        SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->nodeName, snList[nCnt]->node);
        ncmReq->rqResp = 1;
        strcpy(ncmReq->appName, "evm");
        tDb = snList[nCnt]->sid;
        if (!tDb && retCode == NFX_ERR) return retCode;
        tCnt = 0;

        if (reqType == EVM_LABEL_LUN_REQ)
        {
            evmLReq = (evm_MsgLabelLunReq_t *)ncmReq->appMsg;
            evmLReq->hdr.msgType = EVM_LABEL_LUN_REQ;
            evmLReq->hdr.clientPrivate = txn->magicCookie;

            stat = tDb->seq(tDb, &key, &rec, R_FIRST);
            while(!stat)
            {
                memcpy(&evmLReq->id[tCnt], rec.data, sizeof(scsi_id_t));
                SPM_DBG_PRINT(("\t Scsi-Id : 0x%llx", evmLReq->id[tCnt]));
                stat = tDb->seq(tDb, &key, &rec, R_NEXT);
                tCnt++;
            }
            evmLReq->idCnt = snList[nCnt]->sCnt;
            if (cluster)
            {
                SPM_DBG_PRINT(("\t          OC : %s", cluster));
                strcpy(evmLReq->cluster, cluster);
            }
        }
        else if (reqType == EVM_MOD_LUN_CLUSTER_REQ)
        {
            evmModReq = (evm_MsgModLunClusterReq_t *)ncmReq->appMsg;
            evmModReq->hdr.msgType = EVM_MOD_LUN_CLUSTER_REQ;
            evmModReq->hdr.clientPrivate = txn->magicCookie;

            stat = tDb->seq(tDb, &key, &rec, R_FIRST);
            while(!stat)
            {
                memcpy(&evmModReq->id[tCnt], rec.data, sizeof(scsi_id_t));
                SPM_DBG_PRINT(("\t Scsi-Id : 0x%llx", evmModReq->id[tCnt]));
                stat = tDb->seq(tDb, &key, &rec, R_NEXT);
                tCnt++;
            }
            evmModReq->idCnt = snList[nCnt]->sCnt;
            if (cluster)
            {
                SPM_DBG_PRINT(("\t          OC : %s", cluster));
                strcpy(evmModReq->cluster, cluster);
            }
        }
        else
        {
            evmUnLReq = (evm_MsgUnlabelLunReq_t *)ncmReq->appMsg;
            evmUnLReq->hdr.msgType = EVM_UNLABEL_LUN_REQ;
            evmUnLReq->hdr.clientPrivate = txn->magicCookie;

            stat = tDb->seq(tDb, &key, &rec, R_FIRST);
            while(!stat)
            {
                memcpy(&evmUnLReq->id[tCnt], rec.data, sizeof(scsi_id_t));
                SPM_DBG_PRINT(("\t     Scsi-Id : 0x%llx", evmUnLReq->id[tCnt]));
                stat = tDb->seq(tDb, &key, &rec, R_NEXT);
                tCnt++;
            }
            evmUnLReq->idCnt = snList[nCnt]->sCnt;
        }
        rc = sendAgileMsg(spmDesc, ncmReq, ncmReqSize,
                          "ncm", NFX_SSC_UNIX, SSC_SLOT,
                          MGMTPLANE, NULL, NULL, 0);
        if (!(rc < 0)) retCode = NFX_OK;
        else SPM_ERR_PRINT(("Failed to send Label message to EVM"));
        SPM_FREE(ncmReq);
    }  // for

    SPM_DBG_PRINT(("...spm_sendLabel_req"));
    return retCode;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_askLunRescanOnNodes
 *
 * Description  : Send message to EVM for LUN rescan
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 04-JULY-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_askLunRescanOnNodes(scsi_id_t sid, char *node)
{
    ncmMsg_t              *ncmReq;
    evm_MsgRescanLunReq_t *req;
    int32                 ncmSize, cnt, rc;
    spm_nodeList_t        *nL = NULL;

    SPM_FCT_PRINT(("spm_askLunRescanOnNodes..."));
    SPM_DBG_PRINT(("\t ScsiId: 0x%llx", sid));

    ncmSize = sizeof(ncmMsg_t) + sizeof(evm_MsgRescanLunReq_t);
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmSize);

    ncmReq->msgType = NCM_SEND;
    ncmReq->rqResp = 0;
    strcpy(ncmReq->appName, "evm");
    req = (evm_MsgRescanLunReq_t *)ncmReq->appMsg;
    req->hdr.msgType = EVM_RESCAN_LUN_REQ;
    req->hdr.clientPrivate = getpid();
    req->id = sid;

    spm_getNfxDbNodeList(&nL);
    // if (nL && nL->cnt > 1)
    if (nL)
    {
        for (cnt = 0; cnt < nL->cnt; cnt++)
            if (strcasecmp(nL->node[cnt], node) != 0)
            {
                if (!strlen(node) || spm_checkLunInNfxDb(nL->node[cnt], sid))
                {
                    SPM_DBG_PRINT(("\t  Node : %s", nL->node[cnt]));
                    bzero(ncmReq->nodeName, MAXNAME);
                    strcpy(ncmReq->nodeName, nL->node[cnt]);
                    rc = sendAgileMsg(spmDesc, ncmReq, ncmSize, "ncm",
                                      NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                      NULL, NULL, 0);
                    if (rc < 0)
                        SPM_ERR_PRINT(("Error sending EVM_RESCAN_LUN_REQ to EVM"));
                }
            }
    }
    SPM_FREE(nL);
    SPM_FREE(ncmReq);

    SPM_FCT_PRINT(("...spm_askLunRescanOnNodes"));

    return;
}


#ifdef SPM_EA_AV_API_READY

/*
 *------------------------------------------------------------------------------
 * Name         : spm_sendAllocInd_msg
 *
 * Description  : Send message to EVM
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_sendAllocInd_msg(uint64 idCnt,
                                  scsi_id_t *scsiList,
                                  spm_MsgAllocLunReq_t *req,
                                  spm_allocStatus_t result)
{
    ncmMsg_t               *ncmReq;
    int32                  ncmReqSize, rc = NFX_OK;
    spm_MsgAllocLunRsp_t   *allocRsp;

    SPM_FCT_PRINT(("spm_sendAllocInd_msg..."));
    SPM_DBG_PRINT(("\tNode : %s", req->hdr.node));
    SPM_DBG_PRINT(("\tStat : %s", (result==SPM_ALLOC_FAILED)?"Failed":"Success"));
    SPM_DBG_PRINT(("\tCnt  : %lld", idCnt));

    ncmReqSize = sizeof(ncmMsg_t) + sizeof(spm_MsgAllocLunRsp_t);

    if (idCnt > 1)
        ncmReqSize += (sizeof(scsi_id_t)*(idCnt-1));
    SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmReqSize);

    ncmReq->msgType = NCM_SEND;
    strcpy(ncmReq->nodeName, req->hdr.node);
    ncmReq->rqResp = 0;
    strcpy(ncmReq->appName, "evm");

    allocRsp = (spm_MsgAllocLunRsp_t *)ncmReq->appMsg;

    allocRsp->hdr.msgType = SPM_ALLOC_LUN_RSP;
    allocRsp->hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(allocRsp->hdr.node, spmNodeName);
    allocRsp->volId = req->volId;

    allocRsp->result = result;
    allocRsp->nbLun = idCnt;
    if (idCnt > 1 && scsiList)
        memcpy(allocRsp->id, scsiList, sizeof(scsi_id_t)*idCnt);
    else if (scsiList)
        allocRsp->id[0] = scsiList[0];

    rc = sendAgileMsg(-1, ncmReq, ncmReqSize,
                      "ncm", NFX_SSC_UNIX, SSC_SLOT,
                      MGMTPLANE, NULL, NULL, 0);

    if (rc < 0)
    {
        SPM_ERR_PRINT(("Timeout : Failed to send Alloc Indication to evm"));
        rc = NFX_ERR;
    }

    SPM_DBG_PRINT(("...spm_sendAllocInd_msg"));
    return rc;
}
#endif

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getTotalPoolsMsg
 *
 * Description  : Get total # of pools
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getTotalPoolsMsg(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgTotalPoolReq_t *poolReq = (spm_MsgTotalPoolReq_t *)txn->msg;
    spm_senderInfo_t      *sender = txn->sender;
    spm_MsgTotalPoolRsp_t poolRsp;
    int32                 rc;

    SPM_FCT_PRINT(("spm_getTotalPoolsMsg..."));

    bzero(&poolRsp, sizeof(spm_MsgTotalPoolRsp_t));
    poolRsp.hdr.msgType = SPM_GET_TOTAL_POOLS_RSP;
    poolRsp.hdr.clientPrivate = poolReq->hdr.clientPrivate;
    strcpy(poolRsp.hdr.node, spmNodeName);
    poolRsp.num = spm_get_poolNum();

    rc = sendAgileMsg(-1, &poolRsp, sizeof(poolRsp),
                      (poolReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);

    SPM_DBG_PRINT(("\tNumer of pools : %d", poolRsp.num));
    SPM_FCT_PRINT(("...spm_getTotalPoolsMsg"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getPoolListMsg
 *
 * Description  : Get list of pools
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getPoolListMsg(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgPoolListReq_t *poolReq = (spm_MsgPoolListReq_t *)txn->msg;
    spm_senderInfo_t     *sender  = txn->sender;
    spm_MsgPoolListRsp_t *poolRsp = NULL;
    int32                idCnt, rc;
    spm_poolNode_t       *poolList = NULL;
    char                 *listMsg = NULL;
    int32                listMsgSize;

    SPM_FCT_PRINT(("spm_getPoolListMsg..."));

    idCnt = spm_get_pool_id_list(&poolList, err);
    SPM_DBG_PRINT(("Number of Pools : %d", idCnt));
    if (idCnt < 0)
        return -1;
    if (idCnt > 0)
    {
        SPM_DBG_PRINT((" First ID  : %s", (poolList+0)->ctrl));
        listMsgSize = sizeof(spm_MsgPoolListRsp_t) +
                      (sizeof(spm_poolNode_t)*(idCnt-1));
    }
    else
        listMsgSize = sizeof(spm_MsgPoolListRsp_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);
    poolRsp = (spm_MsgPoolListRsp_t *)listMsg;
    poolRsp->hdr.msgType = SPM_GET_POOL_LIST_RSP;
    poolRsp->hdr.clientPrivate = poolReq->hdr.clientPrivate;
    strcpy(poolRsp->hdr.node, spmNodeName);
    poolRsp->numEntries = idCnt;
    if (idCnt > 0)
        memcpy(poolRsp->poolList, poolList, sizeof(spm_poolNode_t)*idCnt);

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (poolReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_DBG_PRINT((" ID  : %s", poolRsp->poolList[0].ctrl));
    SPM_FREE(listMsg);
    SPM_FREE(poolList);

    SPM_FCT_PRINT(("...spm_getPoolListMsg"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getTotalLunMsg
 *
 * Description  : Get Total Luns
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getTotalLunMsg(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgTotalLunReq_t *lunReq = (spm_MsgTotalLunReq_t *)txn->msg;
    spm_senderInfo_t     *sender  = txn->sender;
    spm_MsgTotalLunRsp_t *lunRsp = NULL;
    char                 *listMsg = NULL;
    int32                listMsgSize;
    int32                lunCnt, rc;

    /*
     * When the name and spLun database are combined
     * get all info in one function call.
     *
     * Need for new function for CLI response.
     */
    SPM_FCT_PRINT(("spm_getTotalLunMsg..."));
    SPM_DBG_PRINT(("\t    Device : %s", (lunReq->diskOtherAll == 1)?"All":
                                  (lunReq->diskOtherAll == 0)?"Disk":"Others"));
    SPM_DBG_PRINT(("\t      CTRL : %s", lunReq->ctrl));
    SPM_DBG_PRINT(("\tRAID Level : %d", lunReq->rL));
    SPM_DBG_PRINT(("\t  LUN Type : %d", lunReq->lunType));

    lunCnt = spm_get_poolLunNum(lunReq->ctrl, lunReq->rL,
                                lunReq->lunType, lunReq->diskOtherAll, err);
    SPM_DBG_PRINT(("Number of match : %d", lunCnt));

    listMsgSize = sizeof(spm_MsgTotalLunRsp_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);

    lunRsp = (spm_MsgTotalLunRsp_t *)listMsg;
    lunRsp->hdr.msgType = SPM_GET_TOTAL_LUNS_RSP;
    lunRsp->hdr.clientPrivate = lunReq->hdr.clientPrivate;
    strcpy(lunRsp->hdr.node, spmNodeName);
    strcpy(lunRsp->ctrl, lunReq->ctrl);
    lunRsp->rL  = lunReq->rL;
    lunRsp->lunType = lunReq->lunType;
    lunRsp->numLun = lunCnt;

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (lunReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_FREE(listMsg);
    SPM_FCT_PRINT(("...spm_getTotalLunMsg"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getLunListMsg
 *
 * Description  : Get list Luns
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getLunListMsg(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgLunListReq_t *lunReq = (spm_MsgLunListReq_t *)txn->msg;
    spm_senderInfo_t    *sender  = txn->sender;
    spm_MsgLunListRsp_t *lunRsp = NULL;
    spm_poolLunList_t   *lunList = NULL;
    spm_poolNode_t      *poolNode = NULL;
    char                *listMsg = NULL, *node = NULL;
    int32               listMsgSize;
    int32               lunCnt, rc;

    /*
     * When the name and spLun database are combined
     * get all info in one function call.
     *
     * Need for new function for CLI response.
     */
    SPM_FCT_PRINT(("spm_getLunListMsg..."));
    SPM_DBG_PRINT(("\t    Device : %s", (lunReq->diskOtherAll == 1)?"All":
                                  (lunReq->diskOtherAll == 0)?"Disk":"Others"));
    SPM_DBG_PRINT(("\tMessage Type : %d", lunReq->hdr.msgType));
    SPM_DBG_PRINT(("\t        CTRL : %s", lunReq->ctrl));
    SPM_DBG_PRINT(("\t  RAID Level : %d", lunReq->rL));
    SPM_DBG_PRINT(("\t    LUN Type : %d", lunReq->lunType));
    SPM_DBG_PRINT(("\t    Last SID : 0x%llx", lunReq->lastSID));
    SPM_DBG_PRINT(("\t   Number Rq : %d", lunReq->numReq));
    node = (strlen(lunReq->tNode))?lunReq->tNode:NULL;
    node = (node && !strcasecmp(node, "local"))?spmNodeName:node;
    node = (node && !strcasecmp(node, "pcc"))?spmNodeName:node;
    SPM_DBG_PRINT(("\t   Req. Node : %s", (node)?node:"none"));

    lunCnt = spm_get_poolLunList(lunReq->ctrl, lunReq->rL, &lunList,
                                 lunReq->lunType, lunReq->lastSID,
                                 lunReq->numReq, lunReq->diskOtherAll,
                                 node, &poolNode, err);

    SPM_DBG_PRINT(("Number of match : %d", lunCnt));

    if (lunCnt < 0)
    {
        if (poolNode)
            SPM_FREE(poolNode);
        return -1;
    }
    if (lunCnt > 0)
        listMsgSize = sizeof(spm_MsgLunListRsp_t) +
                      (sizeof(spm_poolLunList_t)*(lunCnt-1));
    else
        listMsgSize = sizeof(spm_MsgLunListRsp_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);

    lunRsp = (spm_MsgLunListRsp_t *)listMsg;
    lunRsp->hdr.msgType = SPM_GET_LUN_LIST_RSP;
    lunRsp->hdr.clientPrivate = lunReq->hdr.clientPrivate;
    strcpy(lunRsp->hdr.node, spmNodeName);
    lunRsp->numLuns = lunCnt;
    if (lunCnt > 0)
    {
        strcpy(lunRsp->ctrl, poolNode->ctrl);
        lunRsp->rL  = poolNode->rL;
        strcpy(lunRsp->make, poolNode->make);
        // strcpy(lunRsp->model, poolNode->model);
        memcpy(lunRsp->lunList, lunList, sizeof(spm_poolLunList_t)*lunCnt);
    }

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (lunReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_FREE(listMsg);
    SPM_FREE(lunList);
    SPM_FREE(poolNode);
    SPM_FCT_PRINT(("...spm_getLunListMsg"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getStoragePoolList
 *
 * Description  : Get storage Pool list
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getStoragePoolList(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgStoragePoolListReq_t *poolReq =
                                (spm_MsgStoragePoolListReq_t *)txn->msg;
    spm_senderInfo_t            *sender  = txn->sender;
    spm_MsgStoragePoolListRsp_t *poolRsp = NULL;
    spm_storageAttrib_t         *poolList = NULL;
    char                        *listMsg = NULL;
    int32                       listMsgSize, idCnt, rc;

    SPM_FCT_PRINT(("spm_getStoragePoolList..."));

    idCnt = spm_get_pool_list_debug(&poolList, err);
    SPM_DBG_PRINT(("Number of Pools : %d", idCnt));
    if (idCnt < 0)
        return -1;
    if (idCnt > 0)
    {
        SPM_DBG_PRINT((" First ID  : %d", (poolList+0)->poolId));
        listMsgSize = sizeof(spm_MsgStoragePoolListRsp_t) +
                      (sizeof(spm_storageAttrib_t)*(idCnt-1));
    }
    else
        listMsgSize = sizeof(spm_MsgStoragePoolListRsp_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);
    poolRsp = (spm_MsgStoragePoolListRsp_t *)listMsg;
    poolRsp->hdr.msgType = SPM_GET_STORAGE_POOL_LIST_RSP;
    poolRsp->hdr.clientPrivate = poolReq->hdr.clientPrivate;
    strcpy(poolRsp->hdr.node, spmNodeName);
    poolRsp->num = idCnt;
    if (idCnt > 0)
        memcpy(poolRsp->strg, poolList, sizeof(spm_storageAttrib_t)*idCnt);

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (poolReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_FREE(listMsg);
    SPM_FREE(poolList);

    SPM_FCT_PRINT(("...spm_getStoragePoolList"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getStorageLunList
 *
 * Description  : Get storage Lun List
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getStorageLunList(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgStorageLunReq_t *lunReq = (spm_MsgStorageLunReq_t *)txn->msg;
    spm_senderInfo_t       *sender  = txn->sender;
    spm_MsgStorageLunRsp_t *lunRsp = NULL;
    int32                  idCnt;
    spm_lunsInPool_t       *lunList = NULL;
    char                   *listMsg = NULL;
    int32                  listMsgSize, rc;

    SPM_FCT_PRINT(("spm_getStorageLunList..."));
    SPM_DBG_PRINT(("\tPool Id : %d", lunReq->poolId));

    idCnt = spm_get_lun_list_debug(&lunList, SPM_LUN_ALL_LIST,
                                   lunReq->poolId, err);
    SPM_DBG_PRINT(("Number of Luns : %d", idCnt));

    if (idCnt < 0)
        return -1;

    if (idCnt > 0)
    {
        SPM_DBG_PRINT((" First ID  : 0x%llx", (lunList+0)->scsiId));
        listMsgSize = sizeof(spm_MsgStorageLunRsp_t) +
                      (sizeof(spm_lunsInPool_t)*(idCnt-1));
    }
    else
        listMsgSize = sizeof(spm_MsgStorageLunRsp_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);
    lunRsp = (spm_MsgStorageLunRsp_t *)listMsg;
    lunRsp->hdr.msgType = SPM_GET_STORAGE_LUN_LIST_RSP;
    lunRsp->hdr.clientPrivate = lunReq->hdr.clientPrivate;
    strcpy(lunRsp->hdr.node, spmNodeName);
    lunRsp->num = idCnt;
    if (idCnt > 0)
        memcpy(lunRsp->luns, lunList, sizeof(spm_lunsInPool_t)*idCnt);

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (lunReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_DBG_PRINT((" ID  : 0x%llx", lunRsp->luns[0].scsiId));
    SPM_FREE(listMsg);
    SPM_FREE(lunList);

    SPM_FCT_PRINT(("...spm_getStorageLunList"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getTotalCtrlsReq
 *
 * Description  : Get total list of controllers
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getTotalCtrlsReq(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgTotalCtrlsReq_t *req = (spm_MsgTotalCtrlsReq_t *)txn->msg;
    spm_senderInfo_t       *sender  = txn->sender;
    spm_MsgTotalCtrlsRsp_t rsp;
    int32                  idCnt;
    int32                  rc;

    SPM_FCT_PRINT(("spm_getTotalCtrlsReq..."));

    idCnt = spm_get_num_ctrls(err);
    SPM_DBG_PRINT(("Number of Controllers : %d", idCnt));

    if (idCnt < 0)
        return -1;

    bzero(&rsp, sizeof(spm_MsgTotalCtrlsRsp_t));
    rsp.hdr.msgType = SPM_GET_TOTAL_CTRLS_RSP;
    rsp.hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(rsp.hdr.node, spmNodeName);
    rsp.num = idCnt;

    rc = sendAgileMsg(-1, &rsp, sizeof(spm_MsgTotalCtrlsRsp_t),
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);

    SPM_FCT_PRINT(("...spm_getTotalCtrlsReq"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_getCtrlListReq
 *
 * Description  : Get ctrl list
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getCtrlListReq(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgCtrlListReq_t *req = (spm_MsgCtrlListReq_t *)txn->msg;
    spm_senderInfo_t     *sender  = txn->sender;
    spm_MsgCtrlListRsp_t *rsp = NULL;
    int32                ctrlCnt, cnt;
    char                 *cL = NULL;
    char                 *listMsg = NULL;
    int32                listMsgSize, rc, offset;

    SPM_FCT_PRINT(("spm_getCtrlListReq..."));
    SPM_DBG_PRINT(("\t Req #     : %d", req->numReq));
    SPM_DBG_PRINT(("\t last Ctrl : %s",
                             (strlen(req->lastCtrl)?req->lastCtrl:"(none)")));

    /* What to do with count numReq */
    ctrlCnt = spm_getCtrlList(&cL, req->lastCtrl, &offset, err);
    SPM_DBG_PRINT(("Num. Ctrl : %d", ctrlCnt));
    SPM_DBG_PRINT(("Offset    : %d", offset));

    if (ctrlCnt < 0)
        return -1;

    cnt = ((ctrlCnt - offset) > req->numReq) ? req->numReq : (ctrlCnt - offset);
    if (cnt > 0)
    {
        SPM_DBG_PRINT((" First CTRL  : %s", (cL+SDM_DEV_NAME_LEN*offset)));
        listMsgSize = sizeof(spm_MsgCtrlListRsp_t) + (SDM_DEV_NAME_LEN*(cnt-1));
    }
    else
        listMsgSize = sizeof(spm_MsgCtrlListRsp_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);
    rsp = (spm_MsgCtrlListRsp_t *)listMsg;
    rsp->hdr.msgType = SPM_GET_CTRL_LIST_RSP;
    rsp->hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(rsp->hdr.node, spmNodeName);
    rsp->numCtrls = cnt;
    if (cnt > 0)
        memcpy(rsp->ctrl, (cL+(SDM_DEV_NAME_LEN*offset)), SDM_DEV_NAME_LEN*cnt);

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_DBG_PRINT(("ctrl : %s", rsp->ctrl[0]));
    SPM_FREE(listMsg);
    SPM_FREE(cL);

    SPM_FCT_PRINT(("...spm_getCtrlListReq"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getCtrlNodeListReq
 *
 * Description  : Get node list for a controller
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getCtrlNodeListReq(spm_msgTxn_t *txn, char *err)
{
    /* Handle this request sync'ly, so don't store the transaction */
    spm_MsgCtrlNodeListReq_t *req = (spm_MsgCtrlNodeListReq_t *)txn->msg;
    spm_senderInfo_t         *sender  = txn->sender;
    spm_MsgCtrlNodeListRsp_t *rsp = NULL;
    int32                    idCnt;
    char                     *nL = NULL;
    char                     *listMsg = NULL;
    int32                    listMsgSize, rc;
    spm_dev_t                type;

    SPM_FCT_PRINT(("spm_getCtrlNodeListReq..."));
    SPM_DBG_PRINT(("\t Ctrl : %s", req->ctrl));

    idCnt = spm_get_ctrl_node_list(&nL, req->ctrl, &type, err);
    SPM_DBG_PRINT(("Num. nodes : %d", idCnt));

    if (idCnt < 0)
        return -1;

    if (idCnt > 0)
    {
        SPM_DBG_PRINT((" First node  : %s", nL));
        listMsgSize = sizeof(spm_MsgCtrlNodeListRsp_t) +
                      (SPM_MAX_NODE_NAME_LEN*(idCnt-1));
    }
    else
        listMsgSize = sizeof(spm_MsgCtrlNodeListReq_t);
    SPM_ALLOC_ZERO(listMsg, char, listMsgSize);
    rsp = (spm_MsgCtrlNodeListRsp_t *)listMsg;
    rsp->hdr.msgType = SPM_GET_CTRL_NODE_LIST_RSP;
    rsp->hdr.clientPrivate = req->hdr.clientPrivate;
    rsp->type = type;
    strcpy(rsp->hdr.node, spmNodeName);
    rsp->numNode = idCnt;
    if (strlen(req->ctrl)) strcpy(rsp->ctrl, req->ctrl);
    if (idCnt > 0)
        memcpy(rsp->nodes, nL, SPM_MAX_NODE_NAME_LEN*idCnt);

    rc = sendAgileMsg(-1, listMsg, listMsgSize,
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_DBG_PRINT((" Node : %s", rsp->nodes[0]));
    SPM_FREE(listMsg);
    SPM_FREE(nL);

    SPM_FCT_PRINT(("...spm_getCtrlNodeListReq"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getDevDetail
 *
 * Description  : Get node list for a controller
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getDevDetail(spm_msgTxn_t *txn, char *err)
{
    spm_MsgDevDetailReq_t *req = (spm_MsgDevDetailReq_t *)txn->msg;
    spm_MsgDevDetailRsp_t rsp;
    char                  *node = NULL;
    int32                 rc;
    sdm_devStatusReqMsg_t *sdmReq;
    ncmMsg_t              *ncmReq;
    int32                 ncmSize;
    int32                 txnIdx = 0;
    spm_lunDbCb_t         *lCfg = NULL;

    SPM_FCT_PRINT(("spm_getDevDetail..."));
    SPM_DBG_PRINT(("\t   Device : %s", req->devName));
    node = (strlen(req->tNode))?req->tNode:NULL;
    node = (node && !strcasecmp(node, "local"))?spmNodeName:node;
    node = (node && !strcasecmp(node, "pcc"))?spmNodeName:node;
    SPM_DBG_PRINT(("\tTgt. Node : %s", (node)?node:"none"));

    spm_getLunDb_rec(req->devName, &lCfg);
    if (!lCfg || !node)
    {
        bzero(&rsp, sizeof(spm_MsgDevDetailRsp_t));
        rsp.hdr.msgType = SPM_GET_DEV_DETAIL_RSP;
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        strcpy(rsp.hdr.node, spmNodeName);
        rsp.result = NFX_ERR;

        sendAgileMsg(-1, &rsp, sizeof(spm_MsgDevDetailRsp_t),
                     (req->hdr.spare)?"ea":txn->sender->name,
                     txn->sender->cpu, txn->sender->slot, txn->sender->plane,
                     NULL, NULL, 0);
        return NFX_OK;
    }
    if (!spm_checkLunInNfxDb(node, lCfg->sid))
    {
        bzero(&rsp, sizeof(spm_MsgDevDetailRsp_t));
        rsp.hdr.msgType = SPM_GET_DEV_DETAIL_RSP;
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        strcpy(rsp.hdr.node, spmNodeName);
        rsp.result = NFX_ERR;

        sendAgileMsg(-1, &rsp, sizeof(spm_MsgDevDetailRsp_t),
                     (req->hdr.spare)?"ea":txn->sender->name,
                     txn->sender->cpu, txn->sender->slot, txn->sender->plane,
                     NULL, NULL, 0);
        return NFX_OK;
    }

    if ((txnIdx = spm_allocateTxnIndex()) < 0)
        rc = NFX_ERR;
    else
    {
        spm_msgTxn_t          *tmpTxn = NULL;
        spm_MsgDevDetailReq_t *sReq = NULL;
        spm_senderInfo_t      *tSdr = NULL;

        SPM_ALLOC_ZERO(sReq, spm_MsgDevDetailReq_t,
                             sizeof(spm_MsgDevDetailReq_t));
        memcpy(sReq, req, sizeof(spm_MsgDevDetailReq_t));
        SPM_ALLOC_ZERO(tSdr, spm_senderInfo_t, sizeof(spm_senderInfo_t));
        memcpy(tSdr, txn->sender, sizeof(spm_senderInfo_t));

        tmpTxn = &spm_txnIndex[txnIdx];
        tmpTxn->msg = (void *)sReq;
        tmpTxn->sender  = tSdr;
        tmpTxn->numReqs = 1;
        tmpTxn->data1   = NULL;
        tmpTxn->data2   = NULL;

        ncmSize = sizeof(ncmMsg_t) + sizeof(sdm_devStatusReqMsg_t);
        SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmSize);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->nodeName, node);
        ncmReq->rqResp = 1;
        strcpy(ncmReq->appName, "sdm");

        sdmReq = (sdm_devStatusReqMsg_t *)ncmReq->appMsg;

        sdmReq->hdr.mid = SDM_MID_DEV_STATUS_REQ;
        sdmReq->hdr.chandle = tmpTxn->magicCookie;
        sdmReq->id = lCfg->sid;

        rc = sendAgileMsg(spmDesc, ncmReq, ncmSize, "ncm",
                          NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          NULL, NULL, 0);
        SPM_FREE(ncmReq);
        if (rc >= 0)
        {
            /* Start timer && assign callback */
            tmpTxn->callback = (int32(*)())spm_procSdmRsp;
            tmpTxn->tmr = spm_procTimerStart(10, spm_sendErrRspTO,
                                             tmpTxn->magicCookie, 0);
        }
    }

    SPM_FCT_PRINT(("...spm_getDevDetail"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getDevListOnPort
 *
 * Description  : Get node list for a controller
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getDevListOnPort(spm_msgTxn_t *txn, char *err)
{
    spm_MsgDevPortReq_t   *req = (spm_MsgDevPortReq_t *)txn->msg;
    char                  *node = NULL;
    int32                 rc;
    sdm_devListReqMsg_t   *sdmReq;
    ncmMsg_t              *ncmReq;
    int32                 ncmSize;
    int32                 txnIdx = 0;

    SPM_FCT_PRINT(("spm_getDevListOnPort..."));
    SPM_DBG_PRINT(("\t     Port : %d", req->port));
    SPM_DBG_PRINT(("\tNum. req. : %d", req->numReq));
    SPM_DBG_PRINT(("\tlast SID  : 0x%llx", req->lastSID));
    node = (strlen(req->tNode))?req->tNode:NULL;
    node = (node && !strcasecmp(node, "local"))?spmNodeName:node;
    node = (node && !strcasecmp(node, "pcc"))?spmNodeName:node;
    SPM_DBG_PRINT(("\tTgt. Node : %s", (node)?node:"none"));

    if (!node || !spm_checkNodeInNfxDb(node))
        return NFX_ERR;
    if ((txnIdx = spm_allocateTxnIndex()) < 0)
        rc = NFX_ERR;
    else
    {
        spm_msgTxn_t          *tmpTxn = NULL;
        spm_MsgDevPortReq_t   *sReq = NULL;
        spm_senderInfo_t      *tSdr = NULL;

        SPM_ALLOC_ZERO(sReq, spm_MsgDevPortReq_t, sizeof(spm_MsgDevPortReq_t));
        memcpy(sReq, req, sizeof(spm_MsgDevPortReq_t));
        SPM_ALLOC_ZERO(tSdr, spm_senderInfo_t, sizeof(spm_senderInfo_t));
        memcpy(tSdr, txn->sender, sizeof(spm_senderInfo_t));

        tmpTxn = &spm_txnIndex[txnIdx];
        tmpTxn->msg = (void *)sReq;
        tmpTxn->sender  = tSdr;
        tmpTxn->numReqs = 1;
        tmpTxn->data1   = NULL;
        tmpTxn->data2   = NULL;

        ncmSize = sizeof(ncmMsg_t) + sizeof(sdm_devListReqMsg_t);
        SPM_ALLOC_ZERO(ncmReq, ncmMsg_t, ncmSize);
        ncmReq->msgType = NCM_SEND;
        strcpy(ncmReq->nodeName, node);
        ncmReq->rqResp = 1;
        strcpy(ncmReq->appName, "sdm");

        sdmReq = (sdm_devListReqMsg_t *)ncmReq->appMsg;

        sdmReq->hdr.mid = SDM_MID_DEV_LIST_REQ;
        sdmReq->hdr.chandle = tmpTxn->magicCookie;
        sdmReq->type = SDM_DEVLIST_ALL;
        sdmReq->port = req->port;
        /* Send message to SDM. Ask for all devices on a port */

        rc = sendAgileMsg(spmDesc, ncmReq, ncmSize, "ncm",
                          NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          NULL, NULL, 0);
        SPM_FREE(ncmReq);
        if (rc >= 0)
        {
            /* Start timer && assign callback */
            tmpTxn->callback = (int32(*)())spm_procSdmRsp;
            tmpTxn->tmr = spm_procTimerStart(10, spm_sendErrRspTO,
                                             tmpTxn->magicCookie, 0);
        }
    }

    SPM_FCT_PRINT(("...spm_getDevListOnPort"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_labelLunCommand
 *
 * Description  : Mark the specified LUN as agile/OC/Foreign LUN
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_labelLunCommand(spm_msgTxn_t *txn, char *err)
{
    /* Handle this async'ly */
    spm_MsgMarkAgileLunReq_t *sReq, *req = (spm_MsgMarkAgileLunReq_t *)txn->msg;
    spm_senderInfo_t         *sender = txn->sender, *tSdr;
    spm_MsgMarkAgileLunRsp_t rsp;
    spm_lunDbCb_t            *lCfg;
    spm_ret_status_t         spmRc;
    int32                    state, rc = 0, cnt = 0, ndCnt = 0;
    spm_scsiNode_t           **snList;
    int32                    tCnt;
    DB                       *tDb;
    DBT                      key;
    DBT                      rec;

    SPM_FCT_PRINT(("spm_labelLunCommand..."));

    /* SPM_ALLOC_ZERO macro typecasts to a pointer */
    SPM_ALLOC_ZERO(snList, spm_scsiNode_t *, sizeof(spm_scsiNode_t *)*nfxDbNodes);
    if (req->devCnt)
    {
        /* We have a specific LUN(s) request */
        for (;cnt < req->devCnt; cnt++)
        {
            spmRc = SPM_NOK;
            if (!strlen(req->dev[cnt])) continue;
            SPM_DBG_PRINT(("\tDevice : %s", req->dev[cnt]));

            spm_getLunDb_rec(req->dev[cnt], &lCfg);
            if (!lCfg)
                spmRc = SPM_NOT_FND;
            else if (lCfg->devType != SPM_DEV_DISK)
                spmRc = SPM_NOT_DISK;
            else
            {
                spmRc = SPM_OK;
                state = spm_getLunState(lCfg->poolId, lCfg->sid);
                switch (state)
                {
                    case SPM_LUN_FREE:
                        if (req->hdr.msgType == SPM_UPDATE_CLUSTER_NAME_REQ)
                            break;
                        if (!req->force)
                        {
                            if (req->hdr.msgType == SPM_MARK_LUN_AS_AGILE_REQ)
                                spmRc = SPM_NOT_FOR;
                            if (req->hdr.msgType == SPM_MARK_LUN_AS_OC_REQ &&
                                !strcasecmp(req->cluster, currCluster))
                                continue;
                        }
                        break;
                    case SPM_LUN_USED:
                        if (req->hdr.msgType == SPM_UPDATE_CLUSTER_NAME_REQ)
                            break;
                        if (!req->force)
                            spmRc = SPM_NOT_FREE;
                        break;
                    case SPM_LUN_FOREIGN:
                        if (req->hdr.msgType == SPM_UPDATE_CLUSTER_NAME_REQ)
                        {
                            spmRc = SPM_NOT_FREE;
                            break;
                        }
                    case SPM_LUN_OC:
                        if (req->hdr.msgType == SPM_UPDATE_CLUSTER_NAME_REQ)
                            break;
                        if (!req->force)
                        {
                            if (req->hdr.msgType != SPM_MARK_LUN_AS_AGILE_REQ)
                                spmRc = SPM_NOT_FREE;
                            else
                            {
                            /*
                             * Make sure that the state of LUN has not changed.
                             * In case of foreign or out-cluster LUN's. Some
                             * other node may have claimed it after we updated
                             * our state.
                             */ 
                            spm_addLunDetailRec(lCfg->sid, lCfg->node, 1, 0);
                            /* Once again get the lCfg */
                            spm_getLunDb_rec(req->dev[cnt], &lCfg);
                            if (!lCfg)
                                spmRc = SPM_NOT_FND;
                            else if (lCfg->lstate == EVM_OC_USED_LUN)
                                spmRc = SPM_NOT_FREE;
                            }
                        }
                        break;
                    default :
                        spmRc = SPM_NOT_FND;
                        break;
                }
            }
            if (spmRc == SPM_OK)
            {
                for(tCnt = 0;
                   ndCnt && (tCnt < ndCnt) &&
                   ((snList[tCnt])?strcasecmp(snList[tCnt]->node,lCfg->node):1);
                   tCnt++);

                if (tCnt == ndCnt)
                {
                    SPM_ALLOC_ZERO(snList[ndCnt], spm_scsiNode_t,
                                                  sizeof(spm_scsiNode_t));
                    strcpy(snList[ndCnt]->node, lCfg->node);
                    tDb = dbopen(NULL, O_CREAT|O_TRUNC|O_RDWR,
                                 0666, DB_BTREE, NULL);
                    if (!tDb)
                    {
                        SPM_ERR_PRINT(("Error snList[%d].sid", ndCnt));
                        spmRc = SPM_NOK;
                    }
                    else
                    {
                        key.data = &lCfg->sid;
                        key.size = sizeof(lCfg->sid);
                        rec.data = &lCfg->sid;
                        rec.size = sizeof(lCfg->sid);
                        tDb->put(tDb, &key, &rec, R_NOOVERWRITE);
                        snList[ndCnt]->sCnt = 1;
                        snList[ndCnt++]->sid = tDb;
                    }
                }
                else
                {
                    tDb = snList[tCnt]->sid;
                    key.data = &lCfg->sid;
                    key.size = sizeof(lCfg->sid);
                    rec.data = &lCfg->sid;
                    rec.size = sizeof(lCfg->sid);
                    tDb->put(tDb, &key, &rec, R_NOOVERWRITE);
                    snList[tCnt]->sCnt++;
                }
            }
            else
                break;
        }  // for
    }
    else
        spmRc = SPM_NOK;

    if (spmRc == SPM_OK && ndCnt)
    {
        spm_msgTxn_t *tmpTxn = NULL;
        int32        txnIdx  = 0;

        if ((txnIdx = spm_allocateTxnIndex()) < 0)
            spmRc = SPM_NOK;
        else
        {
            tmpTxn = &spm_txnIndex[txnIdx];
            SPM_ALLOC_ZERO(sReq, spm_MsgMarkAgileLunReq_t,
                                 sizeof(spm_MsgMarkAgileLunReq_t));
            memcpy(sReq, req, sizeof(spm_MsgMarkAgileLunReq_t));
            SPM_ALLOC_ZERO(tSdr, spm_senderInfo_t, sizeof(spm_senderInfo_t));
            memcpy(tSdr, sender, sizeof(spm_senderInfo_t));

            tmpTxn->msg     = (void *)sReq;
            tmpTxn->sender  = tSdr;
            tmpTxn->numReqs = ndCnt;
            tmpTxn->data1   = (void *)snList;
            tmpTxn->data2   = (void *)SPM_OK;

            if (req->hdr.msgType == SPM_MARK_LUN_AS_AGILE_REQ)
                rc = spm_sendLabel_req(tmpTxn,EVM_LABEL_LUN_REQ,NULL);
            else if (req->hdr.msgType == SPM_MARK_LUN_AS_OC_REQ)
                rc = spm_sendLabel_req(tmpTxn,EVM_LABEL_LUN_REQ,req->cluster);
            else if (req->hdr.msgType == SPM_UPDATE_CLUSTER_NAME_REQ)
                rc = spm_sendLabel_req(tmpTxn,
                                       EVM_MOD_LUN_CLUSTER_REQ,req->cluster);
            else
                rc = spm_sendLabel_req(tmpTxn,EVM_UNLABEL_LUN_REQ,NULL);

            if (rc == NFX_ERR)
            {
                SPM_DBG_PRINT(("EVM Error"));
                sprintf(err, "EVM Error");
                spmRc = SPM_NOK;
                spm_freeTxnIndex(tmpTxn->magicCookie);
            }
        }
        /* Start timer && assign callback */
        tmpTxn->callback = (int32(*)())spm_procEvmLabelRsp;
        tmpTxn->tmr = spm_procTimerStart(10, spm_sendErrRspTO,
                                         tmpTxn->magicCookie, 0);
    }

    if (spmRc != SPM_OK || !ndCnt)
    {
        for (tCnt = 0; tCnt < nfxDbNodes; tCnt++)
        {
            if (!snList[tCnt]) continue;
            if (snList[tCnt]->sid)
                snList[tCnt]->sid->close(snList[tCnt]->sid);
            SPM_FREE(snList[tCnt]);
        }
        SPM_FREE(snList);


        /* Send response to sender that the request failed */
        bzero(&rsp, sizeof(spm_MsgMarkAgileLunRsp_t));

        rsp.hdr.msgType = (req->hdr.msgType == SPM_MARK_LUN_AS_AGILE_REQ)
                              ?SPM_MARK_LUN_AS_AGILE_RSP
                              :(req->hdr.msgType == SPM_MARK_LUN_AS_OC_REQ)
                                  ?SPM_MARK_LUN_AS_OC_RSP
                                  :SPM_UNMARK_LUN_AS_AGILE_RSP;
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        strcpy(rsp.hdr.node, spmNodeName);
        rsp.status = spmRc;

        rc = sendAgileMsg(-1, &rsp, sizeof(spm_MsgMarkAgileLunRsp_t),
                          (req->hdr.spare)?"ea":sender->name,
                          sender->cpu, sender->slot, sender->plane,
                          NULL, NULL, 0);
    }

    SPM_FCT_PRINT(("...spm_labelLunCommand"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_rescanLun
 *
 * Description  : Rescan all/one LUN
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_rescanLun(spm_msgTxn_t *txn, char *err)
{
    spm_MsgLunRescanReq_t  *spmReq = (spm_MsgLunRescanReq_t *)txn->msg;
    spm_senderInfo_t       *sender = txn->sender;
    spm_MsgLunRescanRsp_t  spmRsp;
    spm_lunDbCb_t          *lCfg;
    scsi_id_t              *sL;
    int32                  rc = NFX_OK, sCnt, cnt, found = 0;
    char                   *node = NULL;

    SPM_FCT_PRINT(("spm_rescanLun..."));

    /*
     * We don't want to handle rescan response async'y.
     * Because during the time we are waiting for EVM to respond
     * our database is in a transient state. We won't be able to
     * gurantee any requests!!! A nightmare for vol grow requests
     *
     * If we decide to change this behaviour, we need a locking mechanism
     * at the LUN level.
     */

    spmInitStat = SPM_NOT_PCC;
    if (strlen(spmReq->clNode))
    {
        node = (!strcasecmp(spmReq->clNode, "local"))?spmNodeName
                                                     :spmReq->clNode;
        if (spmReq->devCnt)
        {
            for (cnt = 0; cnt < spmReq->devCnt; cnt++)
            {
                spm_getLunDb_rec(spmReq->dev[cnt], &lCfg);
                if (lCfg && spm_checkLunInNfxDb(node, lCfg->sid))
                {
                    spm_askLunRescanOnNodes(lCfg->sid, "");
                    spm_addLunDetailRec(lCfg->sid, node, 1, txn->magicCookie);
                    found = 1;
                }
            }
        }
        else
        {
            if (spm_checkNodeInNfxDb(node))
            {
                /*
                 * Since we need to rescan LUN's on the node
                 * Let's first delete them
                 */
                spm_getLunListFromNode(node, &sL, &sCnt);
                if (sCnt)
                {
                    for (cnt = 0; cnt < sCnt; cnt++)
                        spm_delLun(sL[cnt], node);
                    SPM_FREE(sL);
                }
                spm_discoverLun(node);
                spm_discoverTape(node);
                spm_discoverChanger(node);
                found = 1;
            }
            else
                rc = NFX_ERR;
        }
    }
    else
    {
        /* Node not specified */
        if (spmReq->devCnt)
        {
            /* We have specific LUN's to re-scan */
            for (cnt = 0; cnt < spmReq->devCnt; cnt++)
            {
                spm_getLunDb_rec(spmReq->dev[cnt], &lCfg);
                if (lCfg)
                {
                    spm_askLunRescanOnNodes(lCfg->sid, "");
                    spm_addLunDetailRec(lCfg->sid, lCfg->node, 1,
                                                   txn->magicCookie);
                    found = 1;
                }
            }
        }
        else
        {
            /* Clean all databases */
            /* Re-register all LUN's in SPM */
            found = 1;
            clean_poolDb();
            clean_LnSDb();
            clean_nfxDb();
            /* Initialize all databases */
            rc = spm_init_db();
            if (rc != NFX_OK) SPM_DBG_PRINT(("SPM asserted for restarted!!!"));
            NFX_ASSERT(rc != NFX_ERR)
        }
    }
    spmInitStat = SPM_PCC_CONNECTED;

    /* Send response to sender only when all responses are received */
    bzero(&spmRsp, sizeof(spmRsp));
    spmRsp.hdr.msgType = SPM_RESCAN_LUN_RSP;
    spmRsp.hdr.clientPrivate = spmReq->hdr.clientPrivate;
    strcpy(spmRsp.hdr.node, spmNodeName);

    if (!found) rc = NFX_ERR;
    spmRsp.result = rc;
    rc = sendAgileMsg(-1, &spmRsp, sizeof(spm_MsgLunRescanRsp_t),
                      (spmReq->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);
    SPM_DBG_PRINT(("...spm_rescanLun"));

    return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_addLunReq
 *
 * Description  : add LUN request
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_addLunReq(spm_msgTxn_t *txn, char *err)
{
    spm_MsgAddLunReq_t *req = (spm_MsgAddLunReq_t *)txn->msg;
    spm_senderInfo_t   *sender = txn->sender;
    spm_MsgAddLunRsp_t rsp;
    int32              rc = NFX_OK;

    SPM_FCT_PRINT(("spm_addLunReq..."));
    SPM_DBG_PRINT(("\t   Scsi_id : 0x%llx", req->lunAttr.scsiId));
    SPM_DBG_PRINT(("\t       WWN : 0x%llx", req->lunAttr.wwn));
    SPM_DBG_PRINT(("\t       LUN : %d", req->lunAttr.lun));
    SPM_DBG_PRINT(("\tRAID Level : %d", req->lunAttr.raidLevel));
    SPM_DBG_PRINT(("\t      Size : %lld", req->lunAttr.size));
    SPM_DBG_PRINT(("\t     State : %d", req->lunAttr.state));
    SPM_DBG_PRINT(("\t  Dev name : %s", req->lunAttr.name));
    SPM_DBG_PRINT(("\tVendor Len : %d", strlen(req->lunAttr.vendor)));
    SPM_DBG_PRINT(("\t    Vendor : %s", req->lunAttr.vendor));
    SPM_DBG_PRINT(("\t Model Len : %d", strlen(req->lunAttr.model)));
    SPM_DBG_PRINT(("\t     Model : %s", req->lunAttr.model));
    SPM_DBG_PRINT(("\t      Type : %d", req->lunAttr.devType));

    bzero((void *)&rsp, sizeof(spm_MsgAddLunRsp_t));
    rsp.hdr.msgType = SPM_ADD_LUN_RSP;
    rsp.hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(rsp.hdr.node, spmNodeName);
    if (req->lunAttr.devType == SCSI_DEVTYPE_DISK)
        rsp.result = spm_addLun(req, SPM_DEV_DISK); // Also send ind. to SRM
    else if (req->lunAttr.devType == SCSI_DEVTYPE_MEDIA_CHANGER)
    {
        req->lunAttr.state = EVM_FREE_LUN;
        strcpy(req->lunAttr.cluster, currCluster);
        rsp.result = spm_addLun(req, SPM_DEV_CHANGER);
    }
    else if (req->lunAttr.devType == SCSI_DEVTYPE_TAPE)
    {
        req->lunAttr.state = EVM_FREE_LUN;
        strcpy(req->lunAttr.cluster, currCluster);
        rsp.result = spm_addLun(req, SPM_DEV_TAPE);
    }
    // else if (req->lunAttr.devType == SCSI_DEVTYPE_STACKER)
    //     rsp.result = spm_addLun(req, SPM_DEV_STACKER);
    else
        rsp.result = NFX_OK; // TODO what to send here ????

    rc = sendAgileMsg(-1, (void *)&rsp, sizeof(spm_MsgAddLunRsp_t),
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);

    SPM_DBG_PRINT(("...spm_addLunReq"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_delLunReq
 *
 * Description  : del LUN request
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_delLunReq(spm_msgTxn_t *txn, char *err)
{
    spm_MsgDelLunReq_t *req = (spm_MsgDelLunReq_t *)txn->msg;
    spm_senderInfo_t   *sender = txn->sender;
    spm_MsgDelLunRsp_t rsp;
    int32              rc;

    SPM_DBG_PRINT(("spm_delLunReq..."));
    SPM_DBG_PRINT(("\tSsci-Id : 0x%llx", req->id));

    bzero((void *) &rsp, sizeof(spm_MsgDelLunRsp_t));
    rsp.hdr.msgType = SPM_DEL_LUN_RSP;
    rsp.hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(rsp.hdr.node, spmNodeName);
    spm_delLun(req->id, req->hdr.node);   // We don't have a return code yet.
    rsp.result = NFX_OK;

    rc = sendAgileMsg(-1, (void *) &rsp, sizeof(spm_MsgDelLunRsp_t),
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);

    SPM_FCT_PRINT(("...spm_delLunReq"));

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_allocLunReq
 *
 * Description  : alloc LUN request
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_allocLunReq(spm_msgTxn_t *txn, char *err)
{
    spm_MsgAllocLunReq_t  *req = (spm_MsgAllocLunReq_t *)txn->msg;
    spm_senderInfo_t      *sender = txn->sender;
    spm_MsgAllocLunRsp_t  *rsp, *newRsp = NULL;
    int64                 idCnt, rspSize, rc;
    spm_allocStatus_t     spmRc;
    scsi_id_t             *scsiList = NULL;
    int32                 txnIdx;
    spm_msgTxn_t          *tmpTxn = NULL;
    uint64                allocSize = 0;
    int32                 n, i, done;
    char                  ctrl[SDM_DEV_NAME_LEN+1];

    SPM_FCT_PRINT(("spm_allocLunReq..."));

    if (!req->nbLun)
    {
        bzero(ctrl, SDM_DEV_NAME_LEN+1);
        n = strlen(req->attr.raidCtrl);
        for (i = 0, done = 0; i < n && done < 2; i++)
        {
            if (req->attr.raidCtrl[i] == '*')
                break;
            if (req->attr.raidCtrl[i] == '_')
                done++;
            if (done != 2)
                ctrl[i] = req->attr.raidCtrl[i];
        }
        bzero(req->attr.raidCtrl, SDM_DEV_NAME_LEN);
        strcpy(req->attr.raidCtrl, ctrl);
        SPM_DBG_PRINT(("\t    Min Size : %lld", req->minSize));
        SPM_DBG_PRINT(("\t  RAID Level : %d", req->attr.raidLevel));
        SPM_DBG_PRINT(("\tMin LUN Size : %d", req->attr.minLunSize));
        SPM_DBG_PRINT(("\tMax LUN Size : %d", req->attr.maxLunSize));
        SPM_DBG_PRINT(("\t        Flag : %d", req->attr.flags));
        SPM_DBG_PRINT(("\t        CTRL : %s", req->attr.raidCtrl));
        SPM_DBG_PRINT(("\t Num vendors : %d", req->attr.num));
        if (strlen(req->attr.vendorInfo[0].vendor))
            SPM_DBG_PRINT(("\t     Vendors : %s",
                         req->attr.vendorInfo[0].vendor));
        if (strlen(req->attr.vendorInfo[0].model))
            SPM_DBG_PRINT(("\t       Model : %s",
                         req->attr.vendorInfo[0].model));
    }
    else
    {
        SPM_DBG_PRINT(("\t    Num. Lun : %d", req->nbLun));
        SPM_DBG_PRINT(("\t      Lun #1 : %s", req->lunList[0]));
    }

    SPM_ALLOC_ZERO(rsp, spm_MsgAllocLunRsp_t, sizeof(spm_MsgAllocLunRsp_t));
    rsp->hdr.msgType = SPM_ALLOC_LUN_RSP;
    rsp->hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(rsp->hdr.node, spmNodeName);
    if ((txnIdx = spm_allocateTxnIndex()) < 0)
        spmRc = rsp->result = SPM_ALLOC_FAILED;
    else
    {
        tmpTxn = &spm_txnIndex[txnIdx];
        if (!req->nbLun)
            spmRc = spm_getFree_lunList(req->minSize, &req->attr, req->hdr.node,
                                        &scsiList, &idCnt, &allocSize,
                                        tmpTxn->magicCookie);
        else
            /*
             * Since we have been asked to allocate specific lun's don't send it
             * to SRM for any thing.
             */
            spmRc = spm_checkFreeLunList(req, &scsiList, &idCnt);
    }
    rspSize = sizeof(spm_MsgAllocLunRsp_t);
    rsp->volId = req->volId;
    if (spmRc == SPM_ALLOC_OK)
    {
        SPM_DBG_PRINT(("Found requested luns..."));
        SPM_DBG_PRINT(("\tNumber of luns : %lld", idCnt));
        SPM_DBG_PRINT(("\tSCS-I ID : 0x%llx", scsiList[0]));
        rsp->result = spmRc;
        rsp->nbLun  = idCnt;
        if (idCnt > 1)
        {
            rspSize += (sizeof(scsi_id_t)*(idCnt-1));
            SPM_ALLOC_ZERO(newRsp, spm_MsgAllocLunRsp_t, rspSize);
            memcpy(newRsp, rsp, sizeof(spm_MsgAllocLunRsp_t));
            memcpy(newRsp->id, scsiList, sizeof(scsi_id_t)*idCnt);
            SPM_FREE(rsp);
        }
        else
        {
            rsp->id[0] = scsiList[0];
            newRsp = rsp;
        }
        SPM_FREE(scsiList);
        if (tmpTxn) spm_freeTxnIndex(tmpTxn->magicCookie);
    }
#ifdef SPM_EA_AV_API_READY
    else if (spmRc == SPM_ALLOC_PENDING)
    {
        spm_evmRcvCB_t     *spmCb = NULL;
        ea_growFsysReq_t   eaReq;
        spm_MsgAllocLunReq_t *sReq;
        spm_senderInfo_t     *tSdr;

        SPM_DBG_PRINT(("Allocate lun request...Pending"));
        bzero(&eaReq, sizeof(ea_growFsysReq_t));
        rsp->result = spmRc;
        rsp->nbLun  = 0;
        newRsp = rsp;

        SPM_ALLOC_ZERO(sReq,spm_MsgAllocLunReq_t, sizeof(spm_MsgAllocLunReq_t));
        memcpy(sReq, req, sizeof(spm_MsgAllocLunReq_t));
        SPM_ALLOC_ZERO(tSdr,spm_senderInfo_t,sizeof(spm_senderInfo_t));
        memcpy(tSdr, sender, sizeof(spm_senderInfo_t));
        tmpTxn->msg     = (void *)sReq;
        tmpTxn->sender  = tSdr;
        tmpTxn->numReqs = 1;

        /* We are waiting for AV indication for FS grow */
        SPM_ALLOC_ZERO(spmCb, spm_evmRcvCB_t, sizeof(spm_evmRcvCB_t));
        spmCb->magic    = SPM_MAGIC_STRING;
        spmCb->rcvSize = 0;
        spmCb->reqSize = req->minSize - allocSize;
        spmCb->pendReq = 1;
        spmCb->smAllocated = 0;
        spmCb->smLunsRcv = 0;
        spmCb->lunReserved = idCnt;
        spmCb->reservedSid = (void *)scsiList;

        tmpTxn->data1 = spmCb;

        eaReq.hdr.msgType = EA_GROW_FSYS_REQ;
        eaReq.hdr.userInfo = tmpTxn->magicCookie;
        strcpy(eaReq.hdr.node, spmNodeName);
        strcpy(eaReq.node, req->hdr.node);
        eaReq.txnId = tmpTxn->magicCookie;
        eaReq.neededSize = (!req->minSize)?1:req->minSize - allocSize;
        memcpy(&eaReq.attr, &req->attr, sizeof(evm_strg_attributes_t));

        SPM_DBG_PRINT(("EA_GROW_FSYS_REQ :"));
        SPM_DBG_PRINT(("\tTxn-Id : 0x%llx", eaReq.txnId));
        SPM_DBG_PRINT(("\tSize   : %lld", eaReq.neededSize));
#ifndef JUST_FOR_TEST_DONT_SEND
        rc = sendAgileMsg(-1, &eaReq, sizeof(eaReq),
                          "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          NULL, NULL, 0);

        if (rc < 0)
        {
            SPM_ERR_PRINT(("Cannot send EA_GROW_FSYS_REQ to AV"));
            rsp->result = SPM_ALLOC_FAILED;
            SPM_FREE(sReq);
            SPM_FREE(tSdr);
            SPM_FREE(spmCb);
            spm_freeTxnIndex(tmpTxn->magicCookie);
        }
        tmpTxn->tmr = spm_procTimerStart(360, spm_sendErrRspTO,
                                             tmpTxn->magicCookie, 0);
#endif
        tmpTxn->callback = (int32(*)())spm_procEaFsysGrowRsp;
    }
#else
    else if (spmRc == SPM_ALLOC_PENDING)
    {
        SPM_DBG_PRINT(("Allocate lun request...failed"));
        rsp->result = SPM_ALLOC_FAILED;
        rsp->nbLun  = 0;
        newRsp = rsp;
        if (tmpTxn) spm_freeTxnIndex(tmpTxn->magicCookie);
        /*
         * Store the request message in a control block!!!
         */
    }
#endif
    else
    {
        SPM_DBG_PRINT(("Found requested luns...None"));
        rsp->result = spmRc;
        rsp->nbLun  = 0;
        newRsp = rsp;
        if (tmpTxn) spm_freeTxnIndex(tmpTxn->magicCookie);
    }
    rc = sendAgileMsg(-1, (void *)newRsp, rspSize,
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);

    SPM_FREE(newRsp);

    SPM_FCT_PRINT(("...spm_allocLunReq"));
    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_freeLunReq
 *
 * Description  : free LUN request
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_freeLunReq(spm_msgTxn_t *txn, char *err)
{
    spm_MsgFreeLunReq_t *req = (spm_MsgFreeLunReq_t *)txn->msg;
    spm_senderInfo_t    *sender = txn->sender;
    spm_MsgFreeLunRsp_t rsp;
    int32               rc;

    SPM_FCT_PRINT(("spm_freeLunReq..."));
    SPM_DBG_PRINT(("\tNumber of ID's : %d", req->nbLun));

    bzero((void *) &rsp, sizeof(spm_MsgFreeLunRsp_t));
    rsp.hdr.msgType = SPM_FREE_LUN_RSP;
    rsp.hdr.clientPrivate = req->hdr.clientPrivate;
    strcpy(rsp.hdr.node, spmNodeName);
    rc = spm_free_lunList(req->nbLun, req->id);
    rsp.result = rc;

    rc = sendAgileMsg(-1, (void *) &rsp, sizeof(spm_MsgFreeLunRsp_t),
                      (req->hdr.spare)?"ea":sender->name,
                      sender->cpu, sender->slot, sender->plane,
                      NULL, NULL, 0);

    SPM_FCT_PRINT(("...spm_freeLunReq"));
    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_updateLunLabelInd
 *
 * Description  : Update lun label indication from EVM
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 04-JULY-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_updateLunLabelInd(spm_lunAttr_t *attr, char *node, int32 checkAttr,
                           spm_dev_t devType)
{
    spm_lunDbCb_t *lCfg = NULL;
    int32         cState, nState;
    char          err[50];

    /*
     * Note : scsi-id cannot change and node in context cannot change
     */

    SPM_FCT_PRINT(("spm_updateLunLabelInd..."));

    if (!attr && attr->scsiId == 0)
        return;

    SPM_DBG_PRINT(("  Scsi-Id : 0x%llx", attr->scsiId));
    if (checkAttr)
    {
        SPM_DBG_PRINT(("   Device : %s", attr->name));
        SPM_DBG_PRINT(("   Vendor : %s", attr->vendor));
        SPM_DBG_PRINT(("   Raid L : %d", attr->raidLevel));
        SPM_DBG_PRINT(("   Model  : %s", attr->model));
        SPM_DBG_PRINT(("  Cluster : %s", attr->cluster));
        SPM_DBG_PRINT(("     Type : %d", attr->devType));
    }

    if (spm_checkLunInNfxDb(node, attr->scsiId))
    {
        /* We can get an update indication only for a known LUN */
        if (checkAttr)
        {
            if (attr->state == EVM_FREE_LUN)
                nState = SPM_LUN_FREE;
            else if (attr->state == EVM_USED_LUN)
                nState = SPM_LUN_USED;
            else if (attr->state == EVM_OC_FREE_LUN ||
                     attr->state == EVM_OC_USED_LUN)
                nState = SPM_LUN_OC;
            else
                nState = SPM_LUN_FOREIGN;
        }

        spm_getScsiDb_rec(attr->scsiId, &lCfg);
        if (lCfg)
        {
            spm_poolAttr_t  poolAttr;
            char            ctrl[SDM_DEV_NAME_LEN+1];
            int32           i, done, poolId;

            if (checkAttr)
            {
                /* Step 1 : Check if the pool has changed for LUN (poolDb) */
                bzero (&poolAttr, sizeof(spm_poolAttr_t));
                poolAttr.mask = SPM_POOL_ATTR_ALL_MASK;
                poolAttr.raidLevel = attr->raidLevel;
                // poolAttr.mask &= ~SPM_POOL_ATTR_MODEL_MASK; // Remove model mask
                poolAttr.type = devType;
                if (strlen(attr->name))
                {
                    bzero(ctrl, SDM_DEV_NAME_LEN+1);  // ctrl is null terminated
                    for (i = 0, done = 0;
                         i < strlen(attr->name) && done < 2;
                         i++)
                    {
                        if (attr->name[i] == '_')
                            done++;
                        if (done != 2)
                            ctrl[i] = attr->name[i];
                    }
                    SPM_DBG_PRINT(("\tController name : %s", ctrl));
                }
                strcpy(poolAttr.ctrl, ctrl);
                strcpy(poolAttr.name, attr->vendor);
                strcpy(poolAttr.model, attr->model);
                poolId = spm_get_pool_by_attrs(&poolAttr, 0, 0, err);

                if (poolId != lCfg->poolId)
                {
                    /*
                     * Pool attributes have changed. So change the pool
                     */
                    spm_delLunFrm_poolDb(lCfg->poolId, lCfg->sid, err);
                    if (poolId < 0)
                        /* Create pool */
                        spm_new_pool(&poolAttr, &poolId, err);
                    if (poolId >= 0)
                        spm_add_lunTo_poolDb(poolId, attr->scsiId, 1,
                                                     attr->size, nState, err);
                }

                /* Step 2 : Check the state of device */
                cState = spm_getLunState(poolId, lCfg->sid);
                (cState && (cState != nState))
                        ?spm_moveLunState(poolId, lCfg->sid, cState, nState):0;

                /* Step 3 : Other attributes may have changed */
                spm_delLunFromLnSdb(lCfg, err); /* Note : lCfg will be NULL. */
                spm_addLunToLnSdb(attr, poolId, node, err);

                /* Send new LUN indication to SRM */
                spm_sendNewLunInd(attr, node);
            }
            else
                /* Ask all other nodes that see this LUN to rescan */
                spm_askLunRescanOnNodes(attr->scsiId, node);
        }
    }

    SPM_FCT_PRINT(("...spm_updateLunLabelInd"));
    return;
}

#ifdef SPM_EA_AV_API_READY
/*
 *------------------------------------------------------------------------------
 * Name         : spm_eaGrowFsysInd
 *
 * Description  : EA grow fsys indication
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_eaGrowFsysInd(spm_msgTxn_t *txn, ea_growFsysInd_t *eaInd)
{
    spm_MsgAllocLunReq_t *req = (spm_MsgAllocLunReq_t *)txn->msg;
    spm_evmRcvCB_t   *spmCb = NULL;
    int32            cnt = 0, rc = 1;
    int32            state;
    spm_lunDbCb_t    *lCfg = NULL;

    SPM_FCT_PRINT(("spm_eaGrowFsysInd..."));
    SPM_DBG_PRINT(("\t    Node : %s", (strlen(eaInd->node)?eaInd->node
                                                          :"Local")));
    SPM_DBG_PRINT(("\t   CB-id : %lld", eaInd->txnId));
    SPM_DBG_PRINT(("\t# of LUN : %d", eaInd->num));

    spmCb = (spm_evmRcvCB_t *)txn->data1;
    if (!spmCb)
        return -1;
    if (spmCb->magic != SPM_MAGIC_STRING)
        /* We don't know what this response is for */
        return -1;

    // spmCb->smAllocated = eaInd->num;
    /*
     * Check for SCSI-Id's in lundb.
     */
    for (cnt = 0; cnt < eaInd->num; cnt++)
    {
        /* Add the LUN into spmAvDb */
        spm_getLunDb_rec(eaInd->devName[cnt], &lCfg);
        SPM_DBG_PRINT(("\tDevice[%d] : %s", cnt, eaInd->devName[cnt]));
        if (lCfg)
        {
            if (spm_checkLunInNfxDb(req->hdr.node, lCfg->sid))
            {
                state = spm_getLunState(lCfg->poolId, lCfg->sid);
                SPM_DBG_PRINT(("\tState : %d", state));
                switch (state)
                {
                    case SPM_LUN_USED :
                        SPM_DBG_PRINT(("AV allocated used LUN %s", lCfg->dev));
                        break;
                    case SPM_LUN_FOREIGN :
                        SPM_DBG_PRINT(("AV allocated foreign LUN %s", lCfg->dev));
                        break;
                    case SPM_LUN_OC :
                        SPM_DBG_PRINT(("AV allocated OC LUN %s", lCfg->dev));
                        break;
                    case SPM_LUN_FREE :
                        SPM_DBG_PRINT(("\tDevice : %s", lCfg->dev));
                        spmCb->rcvSize += lCfg->size;
                        // spmCb->smLunsRcv++;
                        break;
                    default :
                        break;
                }
            }
        }
    } // for

    /*
     * We should free the LUN's and devices irrespective of the result
     */
    spm_unReserveLun(spmCb->lunReserved, spmCb->reservedSid);
    spm_unReserveDevice(eaInd->num, eaInd->devName[0]);

    /* Zero byte request */
    if (spmCb->rcvSize && spmCb->rcvSize >= spmCb->reqSize)
    {
        if (spmCb->pendReq)
        {
            spm_execPendingAllocReq(spmCb, req);
            spmCb->pendReq = 0;
        }
    }
    else
        spm_sendAllocInd_msg(0, NULL, req, SPM_ALLOC_FAILED);

    if (spmCb)
    {
        if (spmCb->reservedSid) SPM_FREE(spmCb->reservedSid);
        SPM_FREE(spmCb);
    }
    SPM_FCT_PRINT(("...spm_eaGrowFsysInd"));
    return rc;
}
#endif

/*
 *------------------------------------------------------------------------------
 * Name         : spm_ncmNodeEvent
 *
 * Description  : NCM node indication
 *  
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_ncmNodeEvent(spm_msgTxn_t *txn, char *err)
{
    ncmnodeEvent_t  *newClNode = (ncmnodeEvent_t *)txn->msg;
    char *node = NULL;
    spm_state_t currState = spmInitStat;

    /*
     * This is just an event. No need to respond.
     */

    SPM_FCT_PRINT(("spm_ncmNodeEvent..."));
    SPM_DBG_PRINT(("\tNode : %s", newClNode->nodeName));
    SPM_DBG_PRINT(("\tState: %s",(newClNode->new)?"Add":"Delete"));

    node = (!strcasecmp(newClNode->nodeName, "local"))?spmNodeName
                                                      :newClNode->nodeName;

    if (newClNode->new)
    {
        /* We are adding a new node */
        switch (currState)
        {
            case SPM_NOT_PCC :
            {
                if (!strcasecmp(node, "pcc"))
                {
                    if (checkNodeIsPCC())
                    {
                        /*
                         * I am the PCC.
                         */
                        SPM_DBG_PRINT(("spm_ncmNodeEvent : should have received a PCC event first"));
                        if (spm_init_db() == NFX_ERR)
                            SPM_DBG_PRINT(("SPM init error"));
                        else
                        {
                            spmInitStat = SPM_PCC_CONNECTED;

                            /* Tell SRM that we have initialized */
                            spm_sendSPMInitInd();
                        }
                    }
                }
                else
                    SPM_ERR_PRINT(("spm_ncmNodeEvent : Non PCC should not get NCM_NODE_EVENT"));
            }
            break;

            case SPM_PCC_ELECTED :
            {
                /* We have been choosen as the PCC */
                if (!strcasecmp(node, "pcc"))
                {
                    /*
                     * Now we are connected to the PCC and will initialize
                     * our databases
                     */
                    SPM_DBG_PRINT(("spm_ncmNodeEvent : Got connected to PCC"));
                    if (spm_init_db() == NFX_ERR)
                            SPM_DBG_PRINT(("SPM init error"));
                    else
                    {
                        spmInitStat = SPM_PCC_CONNECTED;

                        /* Tell SRM that we have initialized */
                        spm_sendSPMInitInd();
                    }
                }
                else
                    SPM_ERR_PRINT(("spm_ncmNodeEvent : First connection has to be the PCC"));
            }
            break;

            case SPM_PCC_CONNECTED :
            {
                if (!strcasecmp(node, "pcc"))
                    /* We are the PCC */
                    SPM_ERR_PRINT(("spm_ncmNodeEvent : node already the PCC."));
                else
                {
                    SPM_DBG_PRINT(("spm_ncmNodeEvent : %s connected", node));
                    if (!spm_checkNodeInNfxDb(node))
                    {
                        spm_discoverLun(node);
                        spm_discoverTape(node);
                        spm_discoverChanger(node);
                    }
                    else
                        SPM_ERR_PRINT(("spm_ncmNodeEvent : duplicate connect message for %s", node));
                }
            }
            break;
        }
    }
    else
    {
        /* We are disconnected from a Node */
        switch (currState)
        {
            case SPM_NOT_PCC :
            {
                SPM_DBG_PRINT(("spm_ncmNodeEvent : PCC delete indication"));
            }
            break;

            case SPM_PCC_ELECTED :
            {
                SPM_ERR_PRINT(("spm_ncmNodeEvent : We are still not the PCC"));
            }
            break;

            case SPM_PCC_CONNECTED :
            {
                /* We are the PCC */
                if (!strcasecmp(node, "pcc"))
                    SPM_ERR_PRINT(("spm_ncmNodeEvent : PCC cannot be disconnected from PCC."));
                else
                {
                    /* Delete LUN's from this node */
                    scsi_id_t          *sList;
                    int32              sCnt = 0, cnt = 0;

                    SPM_DBG_PRINT(("spm_ncmNodeEvent : %s disconnected", node));
                    spm_getLunListFromNode(node, &sList, &sCnt);
                    if (sCnt > 0 && sList)
                    {
                        while (cnt < sCnt)
                            spm_delLun(sList[cnt++], node);
                        SPM_FREE(sList);
                    }
                }
            }
            break;

        }
    }

    SPM_FCT_PRINT(("...spm_ncmNodeEvent"));
    return 0;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_procEvmLabelRsp
 *
 * Description  : Process EVM label response
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 09-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_procEvmLabelRsp(spm_msgTxn_t *txn, char *evmRsp, char *node)
{
    int32                    *mType = (int32 *)evmRsp;
    spm_MsgMarkAgileLunReq_t *spmReq = (spm_MsgMarkAgileLunReq_t *)txn->msg;
    spm_MsgMarkAgileLunRsp_t spmRsp;
    spm_senderInfo_t         *sender = (spm_senderInfo_t *)txn->sender;
    spm_scsiNode_t           **snList = txn->data1;
    int32                    nCnt = 0, rc = 0;
    DB                       *tDb;
    DBT                      key, rec;
    scsi_id_t                sid;
    spm_lunDbCb_t            *lCfg;
    int32                    stat;
    spm_lunState_t           state, toState, lstate;
    int32                    mid;
    uint32                   vgIndex = 0;

    SPM_FCT_PRINT(("spm_procEvmLabelRsp..."));

    txn->numRsps++;
    mid = spmReq->hdr.msgType;
    if (txn->numRsps == txn->numReqs);
    {
        spm_procTimerStop(txn->tmr);
        txn->tmr = NULL;
        bzero(&spmRsp, sizeof(spm_MsgMarkAgileLunRsp_t));
        spmRsp.hdr.msgType = 
              (mid == SPM_MARK_LUN_AS_AGILE_REQ)?SPM_MARK_LUN_AS_AGILE_RSP:
              (mid == SPM_MARK_LUN_AS_OC_REQ)?SPM_MARK_LUN_AS_OC_RSP:
              (mid == SPM_UNMARK_LUN_AS_AGILE_REQ)?SPM_UNMARK_LUN_AS_AGILE_RSP:
                                                   SPM_UPDATE_CLUSTER_NAME_RSP;
        spmRsp.hdr.clientPrivate = spmReq->hdr.clientPrivate;
        strcpy(spmRsp.hdr.node, spmReq->hdr.node);
    }

    vgIndex = spmReq->vgReqIndex;
    if (mid == SPM_UNMARK_LUN_AS_AGILE_REQ)
    {
        toState = SPM_LUN_FOREIGN;
        lstate = EVM_FOREIGN_LUN;
    }
    else if (mid == SPM_MARK_LUN_AS_OC_REQ)
    {
        toState = SPM_LUN_OC;
        lstate = EVM_OC_FREE_LUN;
    }
    else if (mid == SPM_UPDATE_CLUSTER_NAME_REQ)
    {
        toState = SPM_LUN_OC;
        lstate = EVM_OC_USED_LUN;
    }
    else
    {
        toState = SPM_LUN_FREE;
        lstate = EVM_FREE_LUN;
    }

    switch (*mType)
    {
        case EVM_LABEL_LUN_RSP:
        {
            evm_MsgLabelLunRsp_t *rsp = NULL;
            rsp = (evm_MsgLabelLunRsp_t *)evmRsp;
            if (rsp->result != NFX_OK || (int32)txn->data2 == SPM_NOK)
            {
                txn->data2 = (void *)SPM_NOK;
                rc = SPM_NOK;
            }
            else
                rc = SPM_OK;
        }
        break;

        case EVM_UNLABEL_LUN_RSP:
        {
            evm_MsgUnlabelLunRsp_t *rsp = NULL;
            rsp = (evm_MsgUnlabelLunRsp_t *)evmRsp;
            if (rsp->result != NFX_OK || (int32)txn->data2 == SPM_NOK)
            {
                txn->data2 = (void *)SPM_NOK;
                rc = SPM_NOK;
            }
            else
                rc = SPM_OK;
        }

        case EVM_MOD_LUN_CLUSTER_RSP:
        {
            evm_MsgModLunClusterRsp_t *rsp = NULL;
            rsp = (evm_MsgModLunClusterRsp_t *)evmRsp;
            if (rsp->result != NFX_OK || (int32)txn->data2 == SPM_NOK)
            {
                txn->data2 = (void *)SPM_NOK;
                rc = SPM_NOK;
            }
            else
                rc = SPM_OK;
        }
    }

    if (rc == SPM_OK)
    {
        /* Change status of LUN's */
        for (; nCnt < txn->numReqs; nCnt++)
        {
            if (strcasecmp(snList[nCnt]->node, node))
                continue;

            tDb = snList[nCnt]->sid;
            if (tDb)
            {
                stat = tDb->seq(tDb, &key, &rec, R_FIRST);
                while (!stat)
                {
                    memcpy(&sid, rec.data, sizeof(sid));
                    spm_getScsiDb_rec(sid, &lCfg);
                    if (lCfg)
                    {
                        state = spm_getLunState(lCfg->poolId, sid);
                        if(!(mid == SPM_UPDATE_CLUSTER_NAME_REQ &&
                            (state == SPM_LUN_USED ||
                             state == SPM_LUN_FREE) &&
                             !strcasecmp(spmReq->cluster, currCluster)))
                        {
                            if (mid == SPM_UPDATE_CLUSTER_NAME_REQ)
                            {
                                if (lCfg->lstate == EVM_OC_FREE_LUN)
                                {
                                    if(!strcasecmp(spmReq->cluster,currCluster))
                                    {
                                        lstate = EVM_FREE_LUN;
                                        toState = SPM_LUN_FREE;
                                    }
                                    else
                                    {
                                        lstate = EVM_OC_FREE_LUN;
                                        toState = SPM_LUN_OC;
                                    }
                                }
                                else if (lCfg->lstate == EVM_OC_USED_LUN)
                                {
                                    if(!strcasecmp(spmReq->cluster,currCluster))
                                    {
                                        lstate = EVM_USED_LUN;
                                        toState = SPM_LUN_USED;
                                    }
                                }
                                else if (lCfg->lstate == EVM_FREE_LUN)
                                {
                                    lstate = EVM_OC_FREE_LUN;
                                    toState = SPM_LUN_OC;
                                }
                            }
                            spm_moveLunState(lCfg->poolId, sid, state, toState);
                            lCfg->lstate = lstate;
                            if (lstate == EVM_OC_FREE_LUN ||
                                lstate == EVM_OC_USED_LUN)
                                strcpy(lCfg->cluster, spmReq->cluster);
                            else if (lstate == EVM_FREE_LUN ||
                                     lstate == EVM_USED_LUN)
                                strcpy(lCfg->cluster, currCluster);
                            else
                                strcpy(lCfg->cluster, "");
                            if (vgIndex)
                                spm_reserveLun(sid, vgIndex);
                        }
                    }
                    stat = tDb->seq(tDb, &key, &rec, R_NEXT);
                }
            }
            tDb->close(tDb);
            SPM_FREE(snList[nCnt]);
            snList[nCnt] = NULL;
            break;
        }
    }

    if (txn->numRsps == txn->numReqs)
    {
        spmRsp.status = rc;
        sendAgileMsg(-1, &spmRsp, sizeof(spm_MsgMarkAgileLunRsp_t),
                     (spmReq->hdr.spare)?"ea":sender->name,
                     sender->cpu, sender->slot, sender->plane,
                     NULL, NULL, 0);

        /* All the pointers in snList have already been freed */
        SPM_FREE(snList);
        spm_freeTxnIndex(txn->magicCookie);
    }

    SPM_FCT_PRINT(("...spm_procEvmLabelRsp"));

    return 1;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_procSdmRsp
 *
 * Description  : Process SDM response
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 13-SEPT-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_procSdmRsp(spm_msgTxn_t *txn, char *sdmRsp, char *node)
{
    spm_MsgDevPortReq_t *req = (spm_MsgDevPortReq_t *)txn->msg;
    spm_senderInfo_t    *sender = (spm_senderInfo_t *)txn->sender;
    int32  *rcvMsgType = (int32 *)sdmRsp;

    SPM_FCT_PRINT(("spm_procSdmRsp..."));

    switch(*rcvMsgType)
    {
        case SDM_MID_DEV_LIST_RSP:
        {
            sdm_devListRspMsg_t   *rsp = (sdm_devListRspMsg_t *)sdmRsp;
            spm_MsgDevPortRsp_t   *spmRsp = NULL;
            int32                 spmRspSize = 0;

            spm_procTimerStop(txn->tmr);
            txn->tmr = NULL;

            SPM_DBG_PRINT(("\tSDM Count : %d", rsp->numEntries));
            spmRspSize = sizeof(spm_MsgDevPortRsp_t);
            if (rsp->numEntries > 0)
            {
                SPM_DBG_PRINT(("\tFirst entry : 0x%llx", rsp->idList[0]));
                spmRspSize += (sizeof(spm_poolLunList_t)*(rsp->numEntries-1));
            }
            SPM_ALLOC_ZERO(spmRsp, spm_MsgDevPortRsp_t, spmRspSize);
            spmRsp->hdr.msgType = SPM_GET_DEV_PORT_RSP;
            spmRsp->hdr.clientPrivate = req->hdr.clientPrivate;
            strcpy(spmRsp->hdr.node, spmNodeName);
            spmRsp->numLuns = spm_getSpmDevPortDetail(rsp->numEntries,
                                     &rsp->idList[0], req->numReq,
                                     req->lastSID, &spmRsp->lunList[0]);
            SPM_DBG_PRINT(("\tSPM Count : %d\n", spmRsp->numLuns));

            sendAgileMsg(-1, spmRsp, spmRspSize,
                         (req->hdr.spare)?"ea":sender->name,
                         sender->cpu, sender->slot, sender->plane,
                         NULL, NULL, 0);
            SPM_FREE(spmRsp);
        }
        break;

        case SDM_MID_DEV_STATUS_RSP:
        {
            sdm_devStatusRspMsg_t *rsp = (sdm_devStatusRspMsg_t *)sdmRsp;
            spm_MsgDevDetailRsp_t *spmRsp = NULL;
            int32                 spmRspSize = 0, i;
            spm_nodeList_t        *ndList = NULL;
            spm_lunDbCb_t         *lCfg = NULL;

            spm_procTimerStop(txn->tmr);
            txn->tmr = NULL;

            spm_getLunDb_rec (rsp->st.name, &lCfg);
            if (lCfg)
                spm_getNodeListForLun(lCfg->sid, &ndList);

            spmRspSize = sizeof(spm_MsgDevDetailRsp_t);
            if (ndList && ndList->cnt > 0)
                spmRspSize += SPM_MAX_NODE_NAME_LEN*(ndList->cnt-1);

            SPM_ALLOC_ZERO(spmRsp, spm_MsgDevDetailRsp_t, spmRspSize);
            spmRsp->hdr.msgType = SPM_GET_DEV_DETAIL_RSP;
            spmRsp->hdr.clientPrivate = req->hdr.clientPrivate;
            strcpy(spmRsp->hdr.node, spmNodeName);

            spmRsp->size = lCfg->size;
            spmRsp->state = spm_getLunState(lCfg->poolId, lCfg->sid);
            spmRsp->devType = lCfg->devType;
            if (rsp->state == SDM_DEV_STATE_OPENED)
                spmRsp->activeFlag = 1;
            spmRsp->active = rsp->activePath;
            spmRsp->rL = lCfg->rL;
            spmRsp->pathCnt = rsp->pathCnt;
            for (i = 0; i < rsp->pathCnt; i++)
            {
                spmRsp->path[i].path = rsp->paths[i].path;
                spmRsp->path[i].wwn  = rsp->paths[i].wwn;
                spmRsp->path[i].lun  = rsp->paths[i].lun;
            }
            spmRsp->result = NFX_OK;
            spmRsp->volid = lCfg->volId;
            strcpy(spmRsp->devName, rsp->st.name);
            strcpy(spmRsp->model, rsp->st.model);
            if (ndList)
            {
                spmRsp->nodeCnt = ndList->cnt;
                for (i = 0; ndList && i < ndList->cnt; i++)
                    strcpy(spmRsp->nodes[i], ndList->node[i]);
            }
            else
                spmRsp->result = NFX_ERR;

            sendAgileMsg(-1, spmRsp, spmRspSize,
                         (req->hdr.spare)?"ea":sender->name,
                         sender->cpu, sender->slot, sender->plane,
                         NULL, NULL, 0);
            SPM_FREE(spmRsp);
        }
        break;
    }
    spm_freeTxnIndex(txn->magicCookie);

    SPM_FCT_PRINT(("...spm_procSdmRsp"));

    return 1;
}

#ifdef SPM_EA_AV_API_READY
/*
 *------------------------------------------------------------------------------
 * Name         : spm_procEaFsysGrowRsp
 *
 * Description  : Process SRM/EA response
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 09-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

static int32 spm_procEaFsysGrowRsp(spm_msgTxn_t *txn, char *eaRsp, char *node)
{
    spm_MsgAllocLunReq_t *req = (spm_MsgAllocLunReq_t *)txn->msg;
    spm_evmRcvCB_t       *spmCb = (spm_evmRcvCB_t *)txn->data1;
    int32  *rcvMsgType = (int32 *)eaRsp;

    SPM_FCT_PRINT(("spm_procEaFsysGrowRsp..."));

    switch(*rcvMsgType)
    {
        case EA_GROW_FSYS_RSP:
        {
            ea_growFsysRsp_t *rsp = (ea_growFsysRsp_t *)eaRsp;
            if (rsp->result == NFX_ERR)
            {
                SPM_DBG_PRINT(("\t Result : NFX_ERR"));
                spm_procTimerStop(txn->tmr);
                txn->tmr = NULL;
                if (spmCb)
                {
                    spm_unReserveLun(spmCb->lunReserved, spmCb->reservedSid);
                    if (spmCb->reservedSid) SPM_FREE(spmCb->reservedSid);
                    SPM_FREE(spmCb);
                }
                spm_sendAllocInd_msg(0, NULL, req, SPM_ALLOC_FAILED);
                spm_freeTxnIndex(txn->magicCookie);
            }
            else
                SPM_DBG_PRINT(("\t Result : NFX_OK"));
        }
        break;

        case EA_GROW_FSYS_IND:
        {
            ea_growFsysInd_t *eaInd = (ea_growFsysInd_t *)eaRsp;
            spm_procTimerStop(txn->tmr);
            txn->tmr = NULL;
            /*
             * Send message to Alloc indication.
             */
            if (spm_eaGrowFsysInd(txn, eaInd) >= 0)
                spm_freeTxnIndex(txn->magicCookie);
        }
        break;
    }

    SPM_FCT_PRINT(("...spm_procEaFsysGrowRsp"));

    return 1;
}
#endif

/*
 *------------------------------------------------------------------------------
 * Name         : spm_processRspMsg
 *
 * Description  : SPM's RX response message.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 09-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_processRspMsg(char *buf)
{
    int32  *rcvMsgType = (int32 *)buf;
    spm_msgTxn_t *txn;
    char   *node;

    SPM_FCT_PRINT(("spm_processRspMsg..."));

    SPM_DBG_PRINT(("Receive message : %d", *rcvMsgType)); 

    switch (*rcvMsgType)
    {
        case NCM_RESP :
        {
            ncmMsg_t *nRsp = (ncmMsg_t *)buf;
            int32    *rcvType = (int32 *)nRsp->appMsg;

            switch (*rcvType)
            {
                case EVM_LABEL_LUN_RSP :
                {
                    evm_MsgLabelLunRsp_t *rsp = NULL;

                    rsp = (evm_MsgLabelLunRsp_t *)nRsp->appMsg;

                    SPM_DBG_PRINT(("EVM_LABEL_LUN_RSP..."));
                    txn = spm_getTxnIndex((uint32)rsp->hdr.clientPrivate);
                    node = (!strcasecmp(nRsp->nodeName, "local"))?spmNodeName
                                                        :nRsp->nodeName;
                    if (txn)
                        txn->callback(txn, (char *)rsp, node);
                }
                break;

                case EVM_UNLABEL_LUN_RSP :
                {
                    evm_MsgUnlabelLunRsp_t *rsp = NULL;

                    rsp = (evm_MsgUnlabelLunRsp_t *)nRsp->appMsg;

                    SPM_DBG_PRINT(("EVM_UNLABEL_LUN_RSP..."));
                    txn = spm_getTxnIndex((uint32)rsp->hdr.clientPrivate);
                    node = (!strcasecmp(nRsp->nodeName, "local"))?spmNodeName
                                                        :nRsp->nodeName;
                    if (txn)
                        txn->callback(txn, (char *)rsp, node);
                }
                break;

                case EVM_MOD_LUN_CLUSTER_RSP:
                {
                    evm_MsgModLunClusterRsp_t *rsp = NULL;

                    rsp = (evm_MsgModLunClusterRsp_t *)nRsp->appMsg;

                    SPM_DBG_PRINT(("EVM_MOD_LUN_CLUSTER_RSP..."));
                    txn = spm_getTxnIndex((uint32)rsp->hdr.clientPrivate);
                    node = (!strcasecmp(nRsp->nodeName, "local"))?spmNodeName
                                                        :nRsp->nodeName;
                    if (txn)
                        txn->callback(txn, (char *)rsp, node);
                }
                break;

                case EVM_GET_LUN_RSP :
                {
                    evm_MsgGetLunRsp_t *rsp = NULL;

                    rsp = (evm_MsgGetLunRsp_t *)nRsp->appMsg;
                    SPM_DBG_PRINT(("EVM_GET_LUN_RSP..."));
                    txn = spm_getTxnIndex((uint32)rsp->hdr.clientPrivate);
                    node = (!strcasecmp(nRsp->nodeName, "local"))?spmNodeName
                                                        :nRsp->nodeName;
                    if (txn)
                        txn->callback(txn, (char *)rsp, node);
                }
                break;

                case SDM_MID_DEV_LIST_RSP :
                {
                    sdm_devListRspMsg_t *rsp = NULL;

                    rsp = (sdm_devListRspMsg_t *)nRsp->appMsg;
                    SPM_DBG_PRINT(("SDM_MID_DEV_LIST_RSP..."));
                    txn = spm_getTxnIndex((uint32)rsp->hdr.chandle);
                    node = (!strcasecmp(nRsp->nodeName, "local"))?spmNodeName
                                                        :nRsp->nodeName;
                    if (txn)
                        txn->callback(txn, (char *)rsp, node);
                }
                break;

                case SDM_MID_DEV_STATUS_RSP :
                {
                    sdm_devStatusRspMsg_t *rsp = NULL;

                    rsp = (sdm_devStatusRspMsg_t *)nRsp->appMsg;
                    SPM_DBG_PRINT(("SDM_MID_DEV_STATUS_RSP..."));
                    txn = spm_getTxnIndex((uint32)rsp->hdr.chandle);
                    node = (!strcasecmp(nRsp->nodeName, "local"))?spmNodeName
                                                        :nRsp->nodeName;
                    if (txn)
                        txn->callback(txn, (char *)rsp, node);
                }
                break;
            }
        }
        break;

#ifdef SPM_EA_AV_API_READY
        case EA_GROW_FSYS_RSP:
        {
            ea_growFsysRsp_t *eaRsp = (ea_growFsysRsp_t *)buf;

            SPM_DBG_PRINT(("EA_GROW_FSYS_RSP..."));
            txn = spm_getTxnIndex(eaRsp->txnId);
            node = (!strcasecmp(eaRsp->hdr.node, "local"))?spmNodeName
                                                        :eaRsp->hdr.node;
            if (txn)
                txn->callback(txn, (char *)eaRsp, node);
        }
        break;

        case EA_GROW_FSYS_IND:
        {
            ea_growFsysInd_t *eaRsp = (ea_growFsysInd_t *)buf;

            SPM_DBG_PRINT(("EA_GROW_FSYS_IND..."));
            txn = spm_getTxnIndex(eaRsp->txnId);
            node = (!strcasecmp(eaRsp->hdr.node, "local"))?spmNodeName
                                                        :eaRsp->hdr.node;
            if (txn)
                txn->callback(txn, (char *)eaRsp, node);
        }
        break;
#endif

        case NCM_REJECT :
        {
        }
        break;
    }

    SPM_FCT_PRINT(("...spm_processRspMsg"));

    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_processPccEvent
 *
 * Description  : Process PCC event on the node.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 18-JULY-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_processPccEvent(spm_msgTxn_t *txn)
{
    nfx_eventMsg_t *req = (nfx_eventMsg_t *)txn->msg;

    SPM_FCT_PRINT(("spm_processPccEvent..."));

    if (!strcasecmp(req->nodeName, spmNodeName))
    {
        /* I am the PCC */
        if (spmInitStat == SPM_NOT_PCC)
        {
            if (req->param.nodeEvent.state == EVENT_STATE_UP)
            {
                SPM_DBG_PRINT(("%s is the PCC", req->nodeName));
                /* Just set the state to in-process. We may not be connected */
                spmInitStat = SPM_PCC_ELECTED;
            }
            else
                SPM_ERR_PRINT(("%s is not PCC. It should not get a PCC down event for itself", req->nodeName));
        }
        else
        {
            if (req->param.nodeEvent.state == EVENT_STATE_UP)
                SPM_ERR_PRINT(("spm_processPccEvent : Duplicate PCC event"));
            else
            {
                /* I was the PCC, but I just got a delete PCC event */
                SPM_DBG_PRINT(("%s is not PCC anymore", req->nodeName));
                spmInitStat = SPM_NOT_PCC;
                if (spmInitStat == SPM_PCC_CONNECTED)
                {
                    /* Clean all databases */
                    clean_poolDb();
                    clean_LnSDb();
                    clean_nfxDb();
                }
            }
        }
    }

    SPM_FCT_PRINT(("...spm_processPccEvent"));

    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_processMsg
 *
 * Description  : SPM's RX message and distributes the messages as it see fit.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

static void spm_processMsg(void)
{
    char            rtnmsg[SPM_MAX_BUF_SIZE]; // Sometimes it's not a MsgStat_t
    spm_MsgStat_t   *stat_msg = (spm_MsgStat_t *)rtnmsg;
    char *          rtnstr = &rtnmsg[sizeof(spm_MsgStat_t)];
    int32           rtnsize = 0;

    SPM_DBG_PRINT(("SPM: process incoming messages"));

    while(TRUE)
    {
        spm_senderInfo_t    sender;
        spm_msg_t           msg;
        int32               *rcvMsgType;
        spm_msgHdr_t        *msgHdr;
        struct timeval      rcv_to = {SPM_TIMER_VALUE, 0};

        /*
         * Wait for messages...
         */
        // SPM_DBG_PRINT(("==> End := SPM: message\n"));
        bzero(msg.buf, sizeof(msg.buf));
        spm_procTimerTick();
        msg.len = recvAgileMsg(spmDesc,
                               msg.buf,
                               sizeof(msg.buf),
                               &rcv_to,
                               sender.name,
                               &sender.cpu,
                               &sender.slot,
                               &sender.plane);
        NFX_ASSERT(msg.len >= 0);

        if (!msg.len) continue;
        rtnstr[0] = 0;
        msgHdr = (spm_msgHdr_t *)msg.buf;
        rcvMsgType = (int32 *)msg.buf;
        SPM_DBG_PRINT(("==> Begin := SPM: message %d [len %d] receive", *rcvMsgType, msg.len));

        if (spmInitStat != SPM_PCC_CONNECTED &&
            *rcvMsgType != SPM_DEBUG_LEVEL &&
            *rcvMsgType != NCM_NODE_EVENT &&
            *rcvMsgType != NCM_RESP &&
            *rcvMsgType != NCM_REJECT &&
            *rcvMsgType != NFX_EVENT_POST)
        {
            SPM_ERR_PRINT(("Not PCC : cannot process messages"));
            strcpy(rtnstr, "Node is not a PCC. Cannot process messages");
            stat_msg->stat = NFX_ERR;
            stat_msg->hdr.clientPrivate = msgHdr->clientPrivate;
            strcpy(stat_msg->hdr.node, spmNodeName);
            rtnsize = sizeof(spm_MsgStat_t);
            rtnsize += strlen(rtnstr) + 1; // message and its NULL

            if (sendAgileMsg (-1, rtnmsg, rtnsize,
                              (msgHdr->spare)?"ea":sender.name,
                              sender.cpu,
                              sender.slot,
                              sender.plane,
                              NULL, NULL, 0) < 0)
            {
                SPM_ERR_PRINT(("error sending response to %s", sender.name));
            }
            continue;
        }

        /*
         * Something has been received, check it.
         */
        switch(*rcvMsgType)
        {
            case SPM_GET_TOTAL_POOLS_REQ:
            {
                /* We don't have to wait for response from other components */
                spm_MsgTotalPoolReq_t *req = (spm_MsgTotalPoolReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_TOTAL_POOLS..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getTotalPoolsMsg(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_TOTAL_POOLS_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_POOL_LIST_REQ:
            {
                /* We don't have to wait for response from other components */
                spm_MsgPoolListReq_t *req = (spm_MsgPoolListReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_POOL_LIST..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getPoolListMsg(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_POOL_LIST_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_TOTAL_LUNS_REQ:
            {
                /* We don't have to wait for response from other components */
                spm_MsgTotalLunReq_t *req = (spm_MsgTotalLunReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_TOTAL_LUNS..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getTotalLunMsg(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_TOTAL_LUNS_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_LUN_LIST_REQ:
            {
                /* We don't have to wait for response from other components */
                spm_MsgLunListReq_t *req = (spm_MsgLunListReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_LUN_LIST..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getLunListMsg(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_LUN_LIST_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_STORAGE_POOL_LIST_REQ:
            {
                /* We don't have to wait for response from other components */
                spm_MsgStoragePoolListReq_t *req =
                                       (spm_MsgStoragePoolListReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_STORAGE_POOL_LIST Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getStoragePoolList(&rcvTxn, rtnstr);

                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_STORAGE_POOL_LIST_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_STORAGE_LUN_LIST_REQ:
            {
                /* We don't have to wait for response from other components */
                spm_MsgStorageLunReq_t *req = (spm_MsgStorageLunReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_STORAGE_LUN_LIST Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getStorageLunList(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_STORAGE_LUN_LIST_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_TOTAL_CTRLS_REQ:
            {
                spm_MsgTotalCtrlsReq_t *req = (spm_MsgTotalCtrlsReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_TOTAL_CTRLS_REQ Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getTotalCtrlsReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_TOTAL_CTRLS_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_CTRL_LIST_REQ:
            {
                spm_MsgCtrlListReq_t *req = (spm_MsgCtrlListReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_CTRL_LIST_REQ Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getCtrlListReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_CTRL_LIST_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_CTRL_NODE_LIST_REQ:
            {
                spm_MsgCtrlNodeListReq_t *req =
                                         (spm_MsgCtrlNodeListReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_CTRL_NODE_LIST_REQ Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getCtrlNodeListReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_CTRL_NODE_LIST_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_DEV_DETAIL_REQ:
            {
                spm_MsgDevDetailReq_t *req = (spm_MsgDevDetailReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_DEV_DETAIL_REQ Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getDevDetail(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_DEV_DETAIL_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_GET_DEV_PORT_REQ:
            {
                spm_MsgDevPortReq_t *req = (spm_MsgDevPortReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_GET_DEV_PORT_REQ Request..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_getDevListOnPort(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_GET_DEV_PORT_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_MARK_LUN_AS_AGILE_REQ:
            {
                /* We have to wait for EVM response */
                spm_MsgMarkAgileLunReq_t *req = NULL;
                spm_msgTxn_t   rcvTxn;

                req = (spm_MsgMarkAgileLunReq_t *)msg.buf;
                SPM_DBG_PRINT(("SPM_MARK_LUN_AS_AGILE_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_labelLunCommand(&rcvTxn, rtnstr);
                /* Remove !!! For Async handling, we need to wait for rsp */
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_MARK_LUN_AS_AGILE_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_MARK_LUN_AS_OC_REQ:
            {
                spm_MsgMarkAgileLunReq_t *req = NULL;
                spm_msgTxn_t   rcvTxn;

                req = (spm_MsgMarkAgileLunReq_t *)msg.buf;
                SPM_DBG_PRINT(("SPM_MARK_LUN_AS_OC_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_labelLunCommand(&rcvTxn, rtnstr);
                /* Remove !!! For Async handling, we need to wait for rsp */
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_MARK_LUN_AS_OC_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_UNMARK_LUN_AS_AGILE_REQ:
            {
                spm_MsgMarkAgileLunReq_t *req = NULL;
                spm_msgTxn_t   rcvTxn;

                req = (spm_MsgMarkAgileLunReq_t *)msg.buf;
                SPM_DBG_PRINT(("SPM_UNMARK_LUN_AS_AGILE_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_labelLunCommand(&rcvTxn, rtnstr);
                /* Remove !!! For Async handling, we need to wait for rsp */
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_UNMARK_LUN_AS_AGILE_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_RESCAN_LUN_REQ:
            {
                spm_MsgLunRescanReq_t  *req = (spm_MsgLunRescanReq_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("SPM_RESCAN_LUN_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_rescanLun(&rcvTxn, rtnstr);
                /* Remove !!! For Async handling, we need to wait for rsp */
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_RESCAN_LUN_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_ADD_LUN_REQ:
            {
                spm_MsgAddLunReq_t *req = (spm_MsgAddLunReq_t *)msg.buf;
                spm_msgTxn_t       rcvTxn;

                SPM_DBG_PRINT(("SPM_ADD_LUN_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_addLunReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_ADD_LUN_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_DEL_LUN_REQ:
            {
                spm_MsgDelLunReq_t *req = (spm_MsgDelLunReq_t *)msg.buf;
                spm_msgTxn_t       rcvTxn;

                SPM_DBG_PRINT(("SPM_DEL_LUN_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_delLunReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_DEL_LUN_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_ALLOC_LUN_REQ:
            {
                spm_MsgAllocLunReq_t *req = (spm_MsgAllocLunReq_t *)msg.buf;
                spm_msgTxn_t       rcvTxn;

                SPM_DBG_PRINT(("SPM_ALLOC_LUN_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_allocLunReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_ALLOC_LUN_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_FREE_LUN_REQ:
            {
                spm_MsgFreeLunReq_t *req = (spm_MsgFreeLunReq_t *)msg.buf;
                spm_msgTxn_t       rcvTxn;

                SPM_DBG_PRINT(("SPM_FREE_LUN_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_freeLunReq(&rcvTxn, rtnstr);
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_FREE_LUN_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

            case SPM_UPDATE_LUN_LABEL_IND:
            {
                /* No response for this message */
                spm_MsgUpdateLunLabelInd_t *req =
                                    (spm_MsgUpdateLunLabelInd_t *)msg.buf;

                SPM_DBG_PRINT(("SPM_UPDATE_LUN_LABEL_IND..."));

                spm_updateLunLabelInd(&req->attr, req->hdr.node, 0, SPM_DEV_DISK);
            }
            continue;

            case SPM_DEBUG_LEVEL:
            {
                /* No response for this message */
                spm_setDebugLevel_t *req = (spm_setDebugLevel_t *)msg.buf;

                SPM_DBG_PRINT(("SPM_DEBUG_LEVEL..."));
                SPM_DBG_PRINT(("\tLevel: %d -> %d", dbg_level, req->dbg_level));
                dbg_level = req->dbg_level;
            }
            continue;

            case SPM_UPDATE_CLUSTER_NAME_REQ:
            {
                spm_MsgMarkAgileLunReq_t *req = NULL;
                spm_msgTxn_t   rcvTxn;

                req = (spm_MsgMarkAgileLunReq_t *)msg.buf;
                SPM_DBG_PRINT(("SPM_UPDATE_CLUSTER_NAME_REQ..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                stat_msg->stat = spm_labelLunCommand(&rcvTxn, rtnstr);
                /* Remove !!! For Async handling, we need to wait for rsp */
                if (stat_msg->stat == 0)
                    continue;
                stat_msg->hdr.msgType = SPM_UPDATE_CLUSTER_NAME_RSP;
                stat_msg->hdr.clientPrivate = req->hdr.clientPrivate;
                strcpy(stat_msg->hdr.node, spmNodeName);
            }
            break;

#ifdef SPM_EA_AV_API_READY
            case EA_GROW_FSYS_RSP:
            {
                SPM_DBG_PRINT(("EA_GROW_FSYS_RSP..."));
                spm_processRspMsg(msg.buf);
            }
            continue;

            case EA_GROW_FSYS_IND:
            {
                SPM_DBG_PRINT(("EA_GROW_FSYS_IND..."));
                spm_processRspMsg(msg.buf);
            }
            continue;
#endif

            case NCM_NODE_EVENT :
            {
                /*
                 * This is just an event. No need to respond.
                 */
                ncmnodeEvent_t  *req = (ncmnodeEvent_t *)msg.buf;
                spm_msgTxn_t    rcvTxn;

                SPM_DBG_PRINT(("NCM_NODE_EVENT..."));
                bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                rcvTxn.msg = req;
                rcvTxn.sender = &sender;

                spm_ncmNodeEvent(&rcvTxn, rtnstr);
            }
            continue;

            case NCM_REJECT :
            {
                SPM_DBG_PRINT(("NCM_REJECT..."));
                spm_processRspMsg(msg.buf);
            }
            continue;

            case NCM_RESP :
            {
                SPM_DBG_PRINT(("NCM_RESP..."));
                spm_processRspMsg(msg.buf);
            }
            continue;

            case NFX_EVENT_POST :
            {
                nfx_eventMsg_t *req = (nfx_eventMsg_t *)msg.buf;
                spm_msgTxn_t   rcvTxn;

                SPM_DBG_PRINT(("NFX_EVENT_POST..."));
                if (req->eventType == NFX_EVENT_PCC)
                {
                    bzero(&rcvTxn, sizeof(spm_msgTxn_t));
                    rcvTxn.msg = req;
                    rcvTxn.sender = &sender;

                    spm_processPccEvent(&rcvTxn);
                }
            }
            continue;

            default:
            {
                SPM_ERR_PRINT(("SPM: Unknown message (%d)!", *rcvMsgType));
                stat_msg->stat = NFX_ERR;
                sprintf (rtnstr, "Invalid message %d", *rcvMsgType);
                break;
            } // default
        } // switch

        {
            /*
             * Return message to the sender of this message
             */
            rtnsize = sizeof(spm_MsgStat_t);
            if (stat_msg->stat != NFX_OK)
            {
                // Could be an error message
                rtnsize += strlen(rtnstr) + 1; // message and its NULL
            }

            if (sendAgileMsg (-1, rtnmsg, rtnsize,
                              (msgHdr->spare)?"ea":sender.name,
                              sender.cpu,
                              sender.slot,
                              sender.plane,
                              NULL, NULL, 0) < 0)
            {
                SPM_ERR_PRINT(("error sending response to %s", sender.name));
            }
        }
    } // while

    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_init_db
 *
 * Description  : init all DS in SPM
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 30-APR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_init_db()
{
    ncmRqNodeList_t       ndListReq;
    char                  buf[1500];
    ncmNodeList_t         *ndList;
    int32                 ndCnt, rc;

    SPM_FCT_PRINT(("spm_init_db..."));
    /*
     * Open up  databases.
     */
    if (init_poolDb())
    {
        SPM_ERR_PRINT(("SPM failed to init spool database %s",
                       strerror(errno)));
        return NFX_ERR;
    }
    
    if (init_LnSDb())
    {
        SPM_ERR_PRINT(("SPM failed to init lun databases %s",
                       strerror(errno)));
        return NFX_ERR;
    }

    if (init_nfxDb())
    {
        SPM_ERR_PRINT(("SPM failed to init nfx databases %s",
                       strerror(errno)));
        return NFX_ERR;
    }

    /*
     * Ask the SDM for the list of LUN's and then ask the EVM for the
     * details of each one of them.
     */
    bzero(&ndListReq, sizeof(ncmRqNodeList_t));
    bzero(buf, 1500);

    ndListReq.msgType = NCM_RQ_NODE_LIST;
    rc = sendAgileMsg(-1, &ndListReq, sizeof(ncmRqNodeList_t),
                      "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                      &wait_to, buf, 1500);
    if (rc < sizeof(ncmNodeList_t))
    {
        SPM_ERR_PRINT(("No nodes are available from NCM."));
        SPM_ERR_PRINT(("NCM should have the local node."));
    }
    else
    {
        ndList = (ncmNodeList_t *)buf;
        SPM_DBG_PRINT(("Num Nodes : %d", ndList->numConnNodes));

        /* Get list of cluster nodes and ask each one of them */
        for (ndCnt = 0; ndCnt < ndList->numConnNodes; ndCnt++)
        {
            SPM_DBG_PRINT(("Node Name : %s", ndList->nodeNames[ndCnt]));
            if(strcasecmp(ndList->nodeNames[ndCnt], "pcc") &&
               strcasecmp(ndList->nodeNames[ndCnt], spmNodeName))
            {
                if (!strcasecmp(ndList->nodeNames[ndCnt], "local"))
                {
                    spm_discoverLun(spmNodeName);
                    spm_discoverTape(spmNodeName);
                    spm_discoverChanger(spmNodeName);
                }
                else
                {
                    spm_discoverLun(ndList->nodeNames[ndCnt]);
                    spm_discoverTape(ndList->nodeNames[ndCnt]);
                    spm_discoverChanger(ndList->nodeNames[ndCnt]);
                }
            }
        }
    }

    SPM_FCT_PRINT(("...spm_init_db"));
    return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : main
 *
 * Description  : SPM's entry point.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

int
main(int        argc,
     char **    argv,
     char **    envp)
    
{
    int32  rc;

    dbg_level = SPM_DBG_LEVEL_NONE;
    if (argc > 2) dbg_level = atoi(argv[2]);

    SPM_FCT_PRINT(("SPM: > main()"));

    /*
     * Register to receive messages for EA.
     * Get its application id.
     */
    spm_alloc_cnt = 0;
    spm_free_cnt = 0;

    spmDesc = regAgileApp("spm");
    NFX_ASSERT(spmDesc >= 0);
    spmAppId = eee_getApplicationId("spm");

    bzero(&spmTimerList, sizeof(spm_timer_list_t));
    spm_initTxnIndex();

    /*
     * Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, spm_terminate);
    signal(SIGINT, spm_terminate);
    // signal(SIGALRM, spm_procTimerTick);

    /*
     * Get the name of the node where I am running.
     */
    if (gethostname(spmNodeName, SPM_MAX_NODE_NAME_LEN) != 0)
    {
        SPM_ERR_PRINT(("SPM: gethostname failed"));
        return NFX_ERR;
    }

    SPM_DBG_PRINT(("Host name : %s", spmNodeName));

    spm_regGetClusterName();

    SPM_DBG_PRINT(("Cluster name : %s", currCluster));

    spmInitStat = SPM_NOT_PCC;
    if (checkNodeIsPCC())
    {
        SPM_DBG_PRINT(("Node is PCC/Standalone"));
        rc = spm_init_db();
        if (rc == NFX_ERR)
            exit(1);
        spmInitStat = SPM_PCC_CONNECTED;

        /* Tell SRM that we have initialized */
        spm_sendSPMInitInd();
    }

    /*
     * Tell MCP we're ready to roll.
     */
    if (!(argc > 1 && !strcasecmp(argv[1], "standalone")))
        kill (getppid(), SIGUSR1);

    /*
     * Receive messages and process them.
     */
    spm_processMsg();
    
    SPM_ERR_PRINT(("Unexpected return from process_messages."));
    return 1; // Theory says we shouldn't get here
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_procTimerTick
 *
 * Description  : This handles the SIGALRM
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_procTimerTick()
{
    spm_timer_t  *prev = NULL;
    spm_timer_t  *curr = NULL;
    spm_timer_t  *tmrTO = NULL;
    struct timeval tv;

    // SPM_FCT_PRINT(("spm_procTimerTick..."));

    if (spmTimerList.numTimers)
    {
        bzero(&tv, sizeof(tv));
        if (gettimeofday(&tv, NULL))
        {
            SPM_DBG_PRINT(("Timer tick failed to get time of day"));
            return;
        }
        SPM_DBG_PRINT(("\tCurrent time : %ld", tv.tv_sec));
        curr = spmTimerList.start_ptr;
        while (curr)
        {
            SPM_DBG_PRINT(("\tExpire time : %lld", curr->timerExpire));
            if (curr->timerExpire <= tv.tv_sec)
            {
                /* Timer expired. Do the call back and remove the timer */
                tmrTO = curr;
                curr = (prev)?(prev->next = curr->next)
                             :(spmTimerList.start_ptr = curr->next);
                spmTimerList.numTimers--;
                (tmrTO->callback)?tmrTO->callback(tmrTO->param1, tmrTO->param2)
                                 :0;
                SPM_FREE(tmrTO);
            }
            else
            {
                prev = curr;
                curr = curr->next;
            }
        }
    }

    // if (spmTimerList.numTimers)
    //     alarm(SPM_TIMER_VALUE);

    // SPM_FCT_PRINT(("...spm_procTimerTick"));
    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_procTimerStart
 *
 * Description  : Start a timer
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

spm_timer_t *spm_procTimerStart(int32 secs, void (*cB)(),
                                int64 param1, int64 param2)
{
    spm_timer_t  *tmr = NULL;
    struct timeval tv;

    SPM_FCT_PRINT(("spm_procTimerStart..."));

    SPM_ALLOC_ZERO(tmr, spm_timer_t, sizeof(spm_timer_t));
    bzero(&tv, sizeof(tv));
    if (gettimeofday(&tv, NULL))
    {
        SPM_DBG_PRINT(("Timer start failed to get time of day"));
        return NULL;
    }
    // tmr->timerDuration = tmr->timeRemain = secs;
    tmr->timerExpire = tv.tv_sec + secs;
    SPM_DBG_PRINT(("\t  Duration : %d", secs));
    SPM_DBG_PRINT(("\tExpiration : %lld", tmr->timerExpire));
    tmr->callback = cB;
    tmr->param1 = param1;
    tmr->param2 = param2;

    (spmTimerList.numTimers == 0)?(spmTimerList.start_ptr = tmr)
                                 :(spmTimerList.end_ptr->next = tmr);
    spmTimerList.numTimers += 1;
    spmTimerList.end_ptr = tmr;
    // (spmTimerList.numTimers == 1)?alarm(SPM_TIMER_VALUE):0;

    SPM_FCT_PRINT(("...spm_procTimerStart"));
    return tmr;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_procTimerStop
 *
 * Description  : Stop timer
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 01-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_procTimerStop(spm_timer_t *tmr)
{
    spm_timer_t *curr = spmTimerList.start_ptr;
    spm_timer_t *prev = NULL;

    SPM_FCT_PRINT(("spm_procTimerStop..."));

    if (!tmr)
        return;

    while (curr)
    {
        if (curr == tmr)
        {
            (prev != NULL)?(prev->next = curr->next)
                          :(spmTimerList.start_ptr = curr->next);
            spmTimerList.end_ptr = (spmTimerList.end_ptr == curr)?prev
                                                :spmTimerList.end_ptr;
            spmTimerList.numTimers -= 1;
            SPM_FREE(curr);
            SPM_DBG_PRINT(("TimerStop: num remain %d", spmTimerList.numTimers));
            break;
        }
        prev = curr;
        curr = curr->next;
    }

    SPM_FCT_PRINT(("...spm_procTimerStop"));
    return;
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_terminate
 *
 * Description  : This handles the SIGTERM to ensure the config is
 *                synced other stuff.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_terminate(int sig)
{
    SPM_DBG_PRINT(("Terminating SPM..."));

    /*
     * Unregister SPM.
     */
    unRegAgileApp(spmDesc);

    exit(0);
}

/*----------------------------------------------------------------------------*/
