/*------------------------------------------------------------------------------
 * RCS:
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : spm-nfx.c
 *
 * Description  : This file contains the methods used to manipulate
 *                SPM's NFX database.
 *
 * Created by   : Nitin Jain
 * Date Created : 28-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "spm.h"

/*
 * Global variables.
 */

int32 nfxDbNodes;

DB *nfxDb;    // Index into spm_nfxDbCb_t and store all the SCSI-ID's

/*
 * Local function prototypes.
 */

/*
 * Global functions.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : spm_addLunToNfxDb
 *
 * Description  : Add this LUN to nfxDb
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 28-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_addLunToNfxDb(char *node, scsi_id_t sid)
{
    DBT   key;
    DBT   rec;
    spm_nfxDbCb_t  *nCfg;

    SPM_FCT_PRINT(("spm_addLunToNfxDb..."));
    SPM_DBG_PRINT(("\tSCSI : 0x%llx", sid));
    SPM_DBG_PRINT(("\tNode : %s", node));

    if (sid == 0)
        return;

    key.data = node;
    key.size = strlen(node);

    if (!(nfxDb->get(nfxDb, &key, &rec, 0)))
    {
        DB   *tmpDb = NULL;
        DBT  nkey;
        DBT  nrec;

        /* We know about this node */
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (!nCfg)
            return;
        if (nCfg->magic != SPM_MAGIC_STRING)
            return;
        tmpDb = nCfg->lunList;

        nkey.data = &sid;
        nkey.size = sizeof(sid);
        if (!(tmpDb->get(tmpDb, &nkey, &nrec, 0)))
        {
            /*
             * Should not get here if the call is after spm_checkLunInNfxDb
             */
            SPM_DBG_PRINT(("Duplicate entry"));
            return;
        }
        nrec.data = &sid;
        nrec.size = sizeof(sid);
        tmpDb->put(tmpDb, &nkey, &nrec, R_NOOVERWRITE);
        nCfg->numLuns++;
    }
    else
    {
        DBT  nkey;
        DBT  nrec;
        DB   *tmpDb = NULL;

        SPM_ALLOC_ZERO(nCfg, spm_nfxDbCb_t, sizeof(spm_nfxDbCb_t));

        nCfg->magic = SPM_MAGIC_STRING;
        strcpy(nCfg->node, node);
        nCfg->lunList = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR,
                                0666, DB_BTREE, NULL);
        if (!nCfg->lunList)
        {
            SPM_FREE(nCfg);
            NFX_ASSERT(nCfg->lunList != NULL);
        }

        /* Create a node entry in nfxDb */
        rec.data = &nCfg;  // Store the pointer
        rec.size = sizeof(nCfg);
        nfxDb->put(nfxDb, &key, &rec, R_NOOVERWRITE);

        /* Insert scsi-id in the node->lunlist */
        nkey.data = &sid;
        nkey.size = sizeof(sid);
        nrec.data = &sid;
        nrec.size = sizeof(sid);
        tmpDb = nCfg->lunList;
        if (!(tmpDb->get(tmpDb, &nkey, &nrec, 0)))
        {
            /*
             * Should not get here if the call is after spm_checkLunInNfxDb
             */
            SPM_DBG_PRINT(("Duplicate entry"));
            return;
        }
        tmpDb->put(tmpDb, &nkey, &nrec, R_NOOVERWRITE);
        nCfg->numLuns++;
        nfxDbNodes++;
    }

    SPM_FCT_PRINT(("...spm_addLunToNfxDb"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_checkLunInNfxDb
 *
 * Description  : Check if the node sees the LUN
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 28-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_checkLunInNfxDb(char *node, scsi_id_t sid)
{
    DBT   key;
    DBT   rec;
    spm_nfxDbCb_t  *nCfg = NULL;
    int32 found = 0;

    SPM_FCT_PRINT(("spm_checkLunInNfxDb..."));
    SPM_DBG_PRINT(("\tSCSI : 0x%llx", sid));
    SPM_DBG_PRINT(("\tNode : %s", node));

    if (sid == 0)
    {
        SPM_DBG_PRINT(("\tFound : %s", (found)?"Yes":"No"));
        return found;
    }

    key.data = node;
    key.size = strlen(node);

    if (!(nfxDb->get(nfxDb, &key, &rec, 0)))
    {
        DB   *tmpDb = NULL;
        DBT  nkey;
        DBT  nrec;

        /* We know about this node */
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (!nCfg)
        {
            SPM_DBG_PRINT(("\tFound : %s", (found)?"Yes":"No"));
            return found;
        }
        if (nCfg->magic != SPM_MAGIC_STRING)
        {
            SPM_DBG_PRINT(("\tFound : %s", (found)?"Yes":"No"));
            return found;
        }
        tmpDb = nCfg->lunList;

        nkey.data = &sid;
        nkey.size = sizeof(sid);
        if (!(tmpDb->get(tmpDb, &nkey, &nrec, 0)))
            found = 1;
    }

    SPM_DBG_PRINT(("\tFound : %s", (found)?"Yes":"No"));

    SPM_FCT_PRINT(("...spm_checkLunInNfxDb"));
    return found;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_checkNodeInNfxDb
 *
 * Description  : Check if we see the node
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-APR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_checkNodeInNfxDb(char *node)
{
    DBT   key;
    DBT   rec;
    int32 found = 0;
    
    SPM_FCT_PRINT(("spm_checkNodeInNfxDb..."));
    SPM_DBG_PRINT(("\tNode : %s", node));

    key.data = node;
    key.size = strlen(node);

    if (!(nfxDb->get(nfxDb, &key, &rec, 0)))
         found = 1;

    SPM_FCT_PRINT(("...spm_checkNodeInNfxDb"));
    return found;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_delLunFromNfxDb
 *
 * Description  : Delete LUN from nfxDb
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 28-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_delLunFromNfxDb(char *node, scsi_id_t sid)
{
    DBT   key;
    DBT   rec;
    spm_nfxDbCb_t  *nCfg;

    SPM_FCT_PRINT(("spm_delLunFromNfxDb..."));
    SPM_DBG_PRINT(("\tSCSI : 0x%llx", sid));
    SPM_DBG_PRINT(("\tNode : %s", node));

    if (sid == 0)
        return;

    key.data = node;
    key.size = strlen(node);

    if (!(nfxDb->get(nfxDb, &key, &rec, 0)))
    {
        DB   *tmpDb = NULL;
        DBT  nkey;

        /* We know about this node */
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (!nCfg)
            return;
        if (nCfg->magic != SPM_MAGIC_STRING)
            return;

        tmpDb = nCfg->lunList;

        /* Remove the LUN */
        nkey.data = &sid;
        nkey.size = sizeof(sid);
        tmpDb->del(tmpDb, &nkey, 0);
        nCfg->numLuns--;
        if (!nCfg->numLuns)
        {
            tmpDb->close(tmpDb);
            nfxDb->del(nfxDb, &key, 0);
            SPM_FREE(nCfg);
            nfxDbNodes--;
        }
    }

    SPM_FCT_PRINT(("...spm_delLunFromNfxDb"));
    return;
}

/*
 *-----------------------------------------------------------------
 * Name:        spm_getNfxDbNodeList
 *
 * Description: Get list of nodes in NFX database
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 28-APR-2002
 *
 *-----------------------------------------------------------------
 */

void spm_getNfxDbNodeList(spm_nodeList_t **nL)
{
    char            *lst;
    spm_nfxDbCb_t   *nCfg;
    DBT             key;
    DBT             rec;
    int32           stat;
    int32           cnt = 0;

    *nL = NULL;
    SPM_FCT_PRINT(("spm_getNfxDbNodeList..."));

    SPM_ALLOC_ZERO(lst, char, (sizeof(spm_nodeList_t) +
                         (SPM_MAX_NODE_NAME_LEN*(nfxDbNodes-1))));
    /*
     * Should be free'd by caller
     */
    *nL = (spm_nodeList_t *)lst;

    stat = nfxDb->seq(nfxDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (nCfg && nCfg->magic == SPM_MAGIC_STRING)
            strcpy((*nL)->node[cnt++], nCfg->node);
        stat = nfxDb->seq(nfxDb, &key, &rec, R_NEXT);
    }
    (*nL)->cnt = cnt;

    SPM_FCT_PRINT(("...spm_getNfxDbNodeList"));
    return;
}

/*
 *-----------------------------------------------------------------
 * Name:        spm_getLunListFromNode
 *
 * Description: Get list of LUNs for a node
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 28-APR-2002
 *
 *-----------------------------------------------------------------
 */

void spm_getLunListFromNode(char *node, scsi_id_t **sL, int32 *sCnt)
{
    DBT   key;
    DBT   rec;
    spm_nfxDbCb_t *nCfg; 
    int32 cnt = 0;
    
    SPM_FCT_PRINT(("spm_getLunListFromNode..."));
    SPM_DBG_PRINT(("\tNode : %s", node));

    *sL = NULL;
    *sCnt = 0;

    key.data = node;
    key.size = strlen(node);

    if (!(nfxDb->get(nfxDb, &key, &rec, 0)))
    {
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (nCfg && nCfg->magic == SPM_MAGIC_STRING)
        {
            int32   stat;

            SPM_ALLOC_ZERO(*sL, scsi_id_t, (sizeof(scsi_id_t)*nCfg->numLuns));
            stat = (nCfg->lunList)->seq((nCfg->lunList), &key, &rec, R_FIRST);
            while (!stat)
            {
                memcpy(((*sL)+cnt++), (scsi_id_t *)rec.data, sizeof(scsi_id_t));
                stat = (nCfg->lunList)->seq((nCfg->lunList),&key,&rec,R_NEXT);
            }
        }
    }
    *sCnt = cnt;

    SPM_FCT_PRINT(("...spm_getLunListFromNode"));
    return;
}


/*
 *-----------------------------------------------------------------
 * Name:        spm_getLunListFromNode
 *
 * Description: Get list of LUNs for a node
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 28-APR-2002
 *
 *-----------------------------------------------------------------
 */

void spm_getNodeListForLun(scsi_id_t sid, spm_nodeList_t **nL)
{
    DBT   key, nkey;
    DBT   rec, nrec;
    spm_nfxDbCb_t *nCfg; 
    int32 cnt = 0, stat;
    char            *lst;
    
    SPM_FCT_PRINT(("spm_getNodeListForLun..."));
    SPM_DBG_PRINT(("\tScsi-Id : 0x%llx", sid));

    *nL = NULL;
    SPM_ALLOC_ZERO(lst, char, (sizeof(spm_nodeList_t) +
                         (SPM_MAX_NODE_NAME_LEN*(nfxDbNodes-1))));
    /*
     * Should be free'd by caller
     */
    *nL = (spm_nodeList_t *)lst;

    stat = nfxDb->seq(nfxDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (nCfg && nCfg->magic == SPM_MAGIC_STRING)
        {
            DB   *tmpDb = NULL;

            tmpDb = nCfg->lunList;
            nkey.data = &sid;
            nkey.size = sizeof(sid);
            if (!(tmpDb->get(tmpDb, &nkey, &nrec, 0)))
                strcpy((*nL)->node[cnt++], nCfg->node);
        }

        stat = nfxDb->seq(nfxDb, &key, &rec, R_NEXT);
    }
    (*nL)->cnt = cnt;

    SPM_FCT_PRINT(("...spm_getNodeListForLun"));
    return;
}


/*
 *-----------------------------------------------------------------
 * Name: init_nfxDb
 *
 * Description: Open/Initialize the nfx database
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 28-MAR-2002
 *
 *-----------------------------------------------------------------
 */

int32 init_nfxDb(void)
{
    nfxDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
    if (!nfxDb )
    {
        SPM_ERR_PRINT(("Error opening lunDb %s", strerror(errno)));
        return NFX_ERR;
    }
    nfxDbNodes = 0;

    return NFX_OK;
}

/*
 *------------------------------------------------------------------------------
 * Name: clean_nfxDb
 *
 * Description: Clean nfx database
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 04-APR-2002
 *
 *------------------------------------------------------------------------------
 */

void clean_nfxDb(void)
{
    DBT key;
    DBT rec;
    spm_nfxDbCb_t *nCfg = NULL;
    int32  stat;

    stat = nfxDb->seq(nfxDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&nCfg, rec.data, sizeof(nCfg));
        if (nCfg && nCfg->magic == SPM_MAGIC_STRING)
        {
            (nCfg->lunList)->close(nCfg->lunList);
            SPM_FREE(nCfg);
        }
        stat = nfxDb->seq(nfxDb, &key, &rec, R_NEXT);
    }
    nfxDb->close(nfxDb);
    nfxDb = NULL;

    return;
}

/*----------------------------------------------------------------------------*/
