/*------------------------------------------------------------------------------
 * RCS:  $Id: spm.h,v 1.39 2002/10/17 18:53:58 nitinj Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : spm.h
 *
 * Description  : Storage Pool Manager (SPM) - common header file
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 15-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

#ifndef _SPM_H_
#define _SPM_H_

/*
 * Header files.
 */

#ifdef EFENCE
#include "ef-fence.h"
#endif // EFENCE


#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#include <sys/types.h>
#include <limits.h>
#include <db.h>

#include "exports/opt-api.h"
#include "exports/elog-api.h"
#include "spm-api.h"
#ifdef SPM_EA_AV_API_READY
#include <exports/ea-api.h>
#endif
#include <exports/sdm-api.h>

/*
 * Global variables.
 */

#define SPM_MAX_BUF_SIZE         8192
#define SPM_MAGIC_STRING         0x1234567890ll
#define SPM_TIMER_VALUE          5 /* 5 sec timer */
#define SPM_MAX_PENDING_RSPS     128

/*
 * SPM application id.
 */
extern int32 spmAppId;

extern char spmNodeName[SPM_MAX_NODE_NAME_LEN];
extern char currCluster[SPM_MAX_CLUSTER_NAME];

extern int32 spm_alloc_cnt;
extern int32 spm_free_cnt;
extern uint32 dbg_level;

/*
 * Macros.
 */

#ifdef EFENCE
void *ef_malloc(size_t);   // declaration
void ef_free(void *);   // declaration
void *ef_calloc(size_t, size_t);
#define malloc ef_malloc
#define calloc ef_calloc
#define free ef_free
#endif // EFENCE

#define SPM_DBG_PRINT(_par_) \
do {\
    if (dbg_level != SPM_DBG_LEVEL_NONE) \
    {\
        if (dbg_level == SPM_DBG_LEVEL_ELOG) \
            E_LOG(class_1, LOG_DEBUG, spmAppId, 0, 0, 0, _par_); \
        else if (dbg_level >= SPM_DBG_LEVEL_DBG) { printf _par_; printf ("\n"); } \
    }\
} while(0)
#define SPM_ERR_PRINT(_par_) \
do { \
    E_LOG(class_1, LOG_ERR, spmAppId, 0, 0, 0, _par_); \
    if (dbg_level != SPM_DBG_LEVEL_NONE) \
    {\
        if (dbg_level >= SPM_DBG_LEVEL_ERR) { printf _par_; printf ("\n"); } \
    }\
} while(0)
#define SPM_FCT_PRINT(_par_) \
do {\
    if (dbg_level != SPM_DBG_LEVEL_NONE) \
    {\
        if (dbg_level == SPM_DBG_LEVEL_ELOG) \
            E_LOG(class_1, LOG_DEBUG, spmAppId, 0, 0, 0, _par_); \
        else if (dbg_level >= SPM_DBG_LEVEL_FCT) { printf _par_; printf ("\n"); } \
    }\
} while(0)
#define SPM_MEM_PRINT(_par_) \
do {\
    if (dbg_level != SPM_DBG_LEVEL_NONE) \
    {\
        if (dbg_level == SPM_DBG_LEVEL_ELOG) \
            E_LOG(class_1, LOG_DEBUG, spmAppId, 0, 0, 0, _par_); \
        else if (dbg_level >= SPM_DBG_LEVEL_MEM) printf _par_; \
    }\
} while(0)

#define SPM_FREE(_p_)                               \
{                                                   \
    if (_p_ != NULL)                                \
    {                                               \
        spm_free_cnt++;                             \
        SPM_MEM_PRINT(("Free [%s:%-d] -> %d\n", __FILE__, __LINE__, spm_free_cnt)); \
        free(_p_);                                  \
        _p_ = NULL;                                 \
    }                                               \
}

#define SPM_ALLOC(_p_, _type_, _size_)              \
{                                                   \
    _p_ = (_type_ *) malloc(_size_);                \
    if (_p_ == NULL)                                \
    {                                               \
        SPM_MEM_PRINT(("malloc() failed\n"));       \
        NFX_ASSERT(TRUE);                           \
    }                                               \
    spm_alloc_cnt++;                                \
    SPM_MEM_PRINT(("Alloc [%s:%-d] -> %d\n", __FILE__, __LINE__, spm_alloc_cnt)); \
}

#define SPM_ALLOC_ZERO(_p_, _type_, _size_)         \
{                                                   \
    _p_ = (_type_ *) calloc(1, _size_);             \
    if (_p_ == NULL)                                \
    {                                               \
        SPM_MEM_PRINT(("calloc() failed\n"));       \
        NFX_ASSERT(TRUE);                           \
    }                                               \
    spm_alloc_cnt++;                                \
    SPM_MEM_PRINT(("Alloc [%s:%-d] -> %d\n", __FILE__, __LINE__, spm_alloc_cnt)); \
}

/*
 * Type definition.
 */

typedef enum {
    SPM_NOT_PCC, 
    SPM_PCC_ELECTED, 
    SPM_PCC_CONNECTED 
} spm_state_t;

/* Info needed to reply to a sender. */
typedef struct
{
    char    name[32];
    uint32  cpu;
    uint32  slot;
    uint32  plane;
} spm_senderInfo_t;

/* Timer related structures */
typedef struct _spm_timer_t
{
    uint64 timerExpire;
    // uint32 timerDuration;
    // uint32 timeRemain;
    void   (*callback)();
    int64  param1;
    int64  param2;
    struct _spm_timer_t *next;
} spm_timer_t;

typedef struct
{
    int32 numTimers;
    spm_timer_t *start_ptr;
    spm_timer_t *end_ptr;
} spm_timer_list_t;

typedef struct _spm_msgTxn_t spm_msgTxn_t;
struct _spm_msgTxn_t
{
    uint32           magicCookie;
    int32            inUse;
    void             *msg;    // Make a copy to store
    spm_senderInfo_t *sender; // Make a copy to store
    int32            (*callback)(spm_msgTxn_t *, char *, char *);
    spm_timer_t      *tmr;
    int32            numReqs;
    int32            numRsps;
    void             *data1;
    void             *data2;
};

#ifdef SPM_EA_AV_API_READY
typedef struct
{
    uint64               magic;
    uint32               smAllocated;   // # of LUN's from AV
    uint32               smLunsRcv;     // # of LUN's received from SDM
    int32                pendReq;       // 0 - executed, 1 - Pending
    uint64               rcvSize;
    uint64               reqSize;
    uint32               lunReserved;
    void                 *reservedSid;
} spm_evmRcvCB_t;
#endif

typedef struct
{
    char  node[SPM_MAX_NODE_NAME_LEN];
    int32 sCnt;
    DB    *sid;
} spm_scsiNode_t;

/*
 * Basic message structure.
 */
typedef struct
{
    char    buf[SPM_MAX_BUF_SIZE];
    uint32  len;
} spm_msg_t;

typedef struct
{
#define SPM_POOL_ATTR_ALL_MASK              0xFF
#define SPM_POOL_ATTR_RAID_MASK             0x01
#define SPM_POOL_ATTR_CTRL_MASK             0x02
#define SPM_POOL_ATTR_MAKE_MASK             0x04
#define SPM_POOL_ATTR_MODEL_MASK            0x08
#define SPM_POOL_ATTR_TYPE_MASK             0x10
    uint8        mask;       // Will be used to mask search attributes
    evm_raid_t   raidLevel;
    char         ctrl[SDM_DEV_NAME_LEN + 1];
    char         name[SDM_VENDOR_NAME_LEN + 1];
    char         model[SDM_MODEL_NAME_LEN + 1];
    spm_dev_t    type;
} spm_poolAttr_t;


/*
 * Control blocks for each database
 */

typedef struct
{
    scsi_id_t       sid;
    uchar8          nfxRefCnt;
    uint64          resId;
    uint64          size;
} spm_lunNodeInPool_t;

typedef struct
{
    uint64            magic;
    int32             id;         // Pool Id. key for poolDb.
    spm_poolAttr_t    attr;
    uint64            usedSize;   // MB
    uint32            usedLun;    // # of LUNs
    uint64            freeSize;   // MB
    uint32            freeLun;    // # of LUNs
    uint64            foreSize;   // MB
    uint32            foreLun;    // # of LUNs
    uint64            outcSize;   // MB
    uint32            outcLun;    // # of LUNs
    DB                *usedDb;    // Idx on scsi-id : spm_lunNodeInPool_t
    DB                *freeDb;    // Idx on scsi-id : spm_lunNodeInPool_t
    DB                *foreDb;    // Idx on scsi-id : spm_lunNodeInPool_t
    DB                *outcDb;    // Idx on scsi-id : spm_lunNodeInPool_t
} spm_poolDbCb_t;


typedef struct
{
    uint64            magic;
    uint64            wwn;
    uint16            lun;
    scsi_id_t         sid;  // Key for scsiDb
    uint64            size; // MB
    evm_raid_t        rL;
    evm_lunState_t    lstate;
    char              dev[SDM_DEV_NAME_LEN+1];  // Key for lunDb
    char              make[SDM_VENDOR_NAME_LEN+1];
    char              model[SDM_MODEL_NAME_LEN+1];
    int32             poolId; // Pool where this LUN belongs
    char              node[SPM_MAX_NODE_NAME_LEN]; // This is not the only
                                                   // node that sees this
                                                   // LUN !!! Beware
    char              cluster[SPM_MAX_CLUSTER_NAME]; // Only useful for OC.
    evm_volId_t       volId; // Will be set only for used
    spm_dev_t         devType;
} spm_lunDbCb_t;


typedef struct
{
    uint64            magic;
    char              node[SPM_MAX_NODE_NAME_LEN];  // Key for nfxDb
    uint32            numLuns;
    DB                *lunList;      // Index on scsi-id
} spm_nfxDbCb_t;

/* Set related data-struct */
typedef struct
{
    int32 id;
    spm_poolDbCb_t *cfg;
} spm_poolIdPtr_t;

typedef struct
{
    int32 num;
    spm_poolIdPtr_t *id;
} spm_poolSetNode_t;

typedef struct
{
    char data[SDM_DEV_NAME_LEN+1];
    int32 prevSetId;
    int32 setId;
    uint32 numLuns;
    uint64 size;
    spm_poolSetNode_t *s;
} spm_poolCmmNode_t;

typedef struct
{
    evm_raid_t rL;
    int32 prevSetId;
    int32 setId;
    uint32 numLuns;
    uint64 size;
    spm_poolSetNode_t *s;
} spm_poolCmmRLNode_t;

/*
 * Global function prototypes.
 */

// spm-reg.c

int32 spm_regGetClusterName(void);

int32 checkNodeIsPCC(void);


// spm-msg.c

int32 spm_getTotalPoolsMsg(spm_msgTxn_t *txn, char *err);

int32 spm_getPoolListMsg(spm_msgTxn_t *txn, char *err);

int32 spm_getTotalLunMsg(spm_msgTxn_t *txn, char *err);

int32 spm_getLunListMsg(spm_msgTxn_t *txn, char *err);

int32 spm_getStoragePoolList(spm_msgTxn_t *txn, char *err);

int32 spm_getStorageLunList(spm_msgTxn_t *txn, char *err);

int32 spm_getTotalCtrlsReq(spm_msgTxn_t *txn, char *err);

int32 spm_getCtrlListReq(spm_msgTxn_t *txn, char *err);

int32 spm_getCtrlNodeListReq(spm_msgTxn_t *txn, char *err);

int32 spm_getDevDetail(spm_msgTxn_t *txn, char *err);

int32 spm_getDevListOnPort(spm_msgTxn_t *txn, char *err);

int32 spm_labelLunCommand(spm_msgTxn_t *txn, char *err);

int32 spm_rescanLun(spm_msgTxn_t *txn, char *err);

int32 spm_addLunReq(spm_msgTxn_t *txn, char *err);

int32 spm_delLunReq(spm_msgTxn_t *txn, char *err);

int32 spm_allocLunReq(spm_msgTxn_t *txn, char *err);

int32 spm_freeLunReq(spm_msgTxn_t *txn, char *err);

void spm_updateLunLabelInd(spm_lunAttr_t *attr, char *node, int32 checkAttr,
                           spm_dev_t devType);

#ifdef SPM_EA_AV_API_READY
int32 spm_eaGrowFsysInd(spm_msgTxn_t *txn, ea_growFsysInd_t *eaInd);
#endif

int32 spm_ncmNodeEvent(spm_msgTxn_t *txn, char *err);

int32 spm_init_db(void);

void spm_procTimerTick(void);

spm_timer_t *spm_procTimerStart(int32 secs, void (*cB)(),
                                int64 param1, int64 param2);

void spm_procTimerStop(spm_timer_t *tmr);

void spm_terminate(int32 sig);



// spm-pool.c

int32 init_poolDb(void);

void clean_poolDb(void);

int32 spm_moveLunState(int32 poolId, scsi_id_t sid, spm_lunState_t fromState,
                        spm_lunState_t toState);

int32 spm_getLunState(int32 poolId, scsi_id_t sid);

int32 spm_get_pool_id_list(spm_poolNode_t **ids, char *err);

int32 spm_get_num_ctrls(char *err);

int32 spm_getCtrlList(char **ctrlList, char *lastCtrl, 
                      int32 *offset, char *err);

int32 spm_get_ctrl_node_list(char **nodeList, char *ctrl, spm_dev_t *type,
                             char *err);

int32 spm_get_pool_list_debug(spm_storageAttrib_t **poolList, char *err);

int32 spm_get_lun_list_debug(spm_lunsInPool_t **lunList, spm_lunType_t type,
                             int32 poolId, char *rtnstr);

int32 spm_compare_pool_attr(spm_poolAttr_t *poolAttr,
                            spm_poolDbCb_t *poolCfg);

int32 spm_get_pool_by_attrs(spm_poolAttr_t *attr, int32 lunsAvail,
                            int32 startId, char *err);

int32 spm_new_pool(spm_poolAttr_t *attrs, int32 *id, char *err);

void spm_del_pool(int32 poolId);

int32 spm_add_lunTo_poolDb(uint32 poolId, scsi_id_t scsiId, uchar8 addLun,
                           uint64 lunSize, spm_lunState_t state, char *err);

int32 spm_delLunFrm_poolDb(uint32 poolId, scsi_id_t sid, char *err);

int32 spm_getFree_lunList(uint64 reqSize, evm_strg_attributes_t *lunAttr,
                          char *node, scsi_id_t **scsiList, int64 *listCnt,
                          uint64 *allocSize, uint32 txnId);

int32 spm_checkFreeLunList(spm_MsgAllocLunReq_t *req,
                           scsi_id_t **scsiList, int64 *listCnt);

int32 spm_free_lunList(uint32 nbLun, scsi_id_t *scsiList);

uint32 spm_get_poolNum(void);

int32 spm_get_poolLunNum(char *ctrl, int32 rL, spm_lunType_t lType,
                         uint32 doa, char *err);

int32 spm_get_poolLunList(char *ctrl, int32 rL, spm_poolLunList_t **lunList,
                          spm_lunType_t lType, scsi_id_t s_sid, uint32 reqNum,
                          uint32 doa, char *tNode, spm_poolNode_t **poolNode,
                          char *err);

int32 spm_getSpmDevPortDetail(int32 idCnt, scsi_id_t *sList,
                              int32 numReq, scsi_id_t lastId,
                              spm_poolLunList_t *lunList);

void spm_unReserveLun(uint32 idCnt, scsi_id_t *list);

void spm_unReserveDevice(uint32 idCnt, char *list);

void spm_reserveLun(scsi_id_t sid, uint32 resId);


// spm-lun.c

int32 spm_addLunToLnSdb(spm_lunAttr_t *attr, int32 poolId,
                        char *node, char *err);

int32 spm_delLunFromLnSdb(spm_lunDbCb_t *lCfg, char *err);

int32 spm_checkLunInLnSdb(char *device, scsi_id_t sid);

void spm_getScsiDb_rec(scsi_id_t sid, spm_lunDbCb_t **lCfg);

void spm_getLunDb_rec(char *device, spm_lunDbCb_t **lCfg);

int32 spm_getPoolIdForLun(char *device, scsi_id_t sid);

int32 init_LnSDb(void);

void clean_LnSDb(void);


// spm-nfx.c

void spm_addLunToNfxDb(char *node, scsi_id_t sid);

int32 spm_checkLunInNfxDb(char *node, scsi_id_t sid);

int32 spm_checkNodeInNfxDb(char *node);

void spm_delLunFromNfxDb(char *node, scsi_id_t sid);

void spm_getNfxDbNodeList(spm_nodeList_t **nL);

void spm_getLunListFromNode(char *node, scsi_id_t **sL, int32 *sCnt);

void spm_getNodeListForLun(scsi_id_t sid, spm_nodeList_t **nL);

int32 init_nfxDb(void);

void clean_nfxDb(void);

// spm-srm.c

int32 spm_sendSPMInitInd(void);

int32 spm_sendNewLunInd(spm_lunAttr_t *attr, char *devNode);

int32 spm_sendDelLunInd(char *devName, char *devNode);

int32 spm_sendLunChangeStateInd(char *dev, spm_lunState_t frmState, 
                                           spm_lunState_t toState);

// spm-setUtil.c

int32 spm_createPoolSetsFromStrg(uint64 size, const evm_strg_attributes_t *const attr,
                                spm_poolCmmNode_t ***idSet, int32 *const idCnt);

void spm_destroyCmmNode(spm_poolCmmNode_t **cs, int32 num);

/*----------------------------------------------------------------------------*/
#endif // _SPM_H_
