/*------------------------------------------------------------------------------
 * RCS:  $Id: stats.c,v 1.3 2002/11/21 00:30:11 rick Exp $
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Name         : stats.c
 *
 * Description  : functions that are available to other modules
 *                eg init, EEE rcv message for stats agent, etc
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 10/22/02
 *
 *------------------------------------------------------------------------------
 */

#include "stats.h"
#include "stats-dbg.h"
#include "exports/timer-api.h"

//------------------------------------------------------------------------------
// Global variables.
//------------------------------------------------------------------------------

uint32			statsAgentPort;
int32			statsAppId;
nfx_spinlock_t	stats_lock;
#if 1
// this should be removed for formal release
int				collect_data = 0;
#endif

#if 0
static eeeGMT_t prevTime = {0, 0};
#endif


//------------------------------------------------------------------------------
// Global functions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name         : stats_rcvMsg
//
// Description  : Receives unicast messages sent to stats agent.
//
// Created by   : Shyh-Pei Yen
//
// Date Created : 10/22/02
//
//------------------------------------------------------------------------------

void
stats_rcvMsg(eee_descPtr_t edesc)
{
	uint32			msgLen = 0;
	char 			*msgBuf = NULL;
	uint32			*msgType;
	eee_descBuf_t	*bd = NULL;
	boolean			freeMsgBuf = TRUE;
	int32			rc, numBufs;
	uint32			srcPort, dstPort;

	numBufs = EEE_GET_NUM_BUFS(edesc);
	srcPort = edesc->hdr.src_port;
	dstPort = edesc->hdr.dest_port;
	if (numBufs == 1)
	{
		// Single buffer message, don't copy to flat buffer.
		if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
			bd = &(edesc->bd[0]);
		else
			bd = &(edesc->bd[1]);
		msgBuf = bd->buf;
		msgLen = bd->len;

		// Done with the descriptor.
		eee_deallocateDesc(edesc);
	}
	else
	{
		// Copy the multi-buffer message to a
		// contiguous malloc buffer and free the
		// received packet.
		msgLen = EEE_GET_PKT_LEN(edesc);
		msgBuf = (char *) eee_ramAlloc(msgLen);
		if (msgBuf == NULL)
		{
			stats_trace(LOG_ERR, "No memory, dropping packet.\n");
			eee_freePacket(edesc);
			return;
		}
		rc = eee_copyPacketToLocalBuffer(msgBuf, edesc, msgLen, 0);
		if (rc != msgLen)
		{
			stats_trace(LOG_ERR, "eee_copyPacketToLocal() failed.\n");
			eee_freePacket(edesc);
			return;
		}

		// Done with the packet.
		eee_freePacket(edesc);
	}

	msgType = (int *)msgBuf;

#if 0
	{
		uint64		diff;
		eeeGMT_t	currTime;

		if ((prevTime.sec == 0) && (prevTime.usec == 0))
		{
			eee_TimeGMT(&prevTime);
		}
		else
		{
			eee_TimeGMT(&currTime);
			diff = EEE_GMT_DELTA_MSEC(currTime, prevTime);
			if (diff >= 2000)
			{
				printf("diff = %lld (msg %d)\n", diff, *msgType);
			}
			prevTime = currTime;
		}
	}
#endif

#if 1
	// received message, turn on collecting flag...
	// this is for internal use only in case we don't want
	// to clear the hardware stats counters all the time
	// THESE CODE SHOULD BE REMOVED FOR FORMAL RELEASE
	collect_data = 1;
#endif

	switch (*msgType)
	{
	case STATS_GETSTATS:
		stats_trace(LOG_DEBUG, "STATS_GETSTATS");
		stats_getStats(msgBuf, msgLen, srcPort, dstPort);
		break;

	case STATS_RESETCOUNTERS:
		stats_trace(LOG_DEBUG, "STATS_RESETCOUNTERS");
		stats_resetCounters(msgBuf, msgLen, srcPort, dstPort);
		break;

	default:
		stats_trace(LOG_ERR, "Unknown msgType(%d) received.\n", *msgType);
		break;
	} // switch

	if (freeMsgBuf)
	{
		if (numBufs == 1)
			eee_deallocateBuffer(msgBuf);
		else
			eee_ramDealloc(msgBuf);
	}

	return;
}


//-----------------------------------------------------------------
// Name: stats_rcvMulticast
//
// Description: Receives multicast messages sent to stats agent.
//
// Created by:	Shyh-Pei Yen
//
// Date Created: 10/22/02
//
//-----------------------------------------------------------------

void
stats_rcvMulticast(eee_descPtr_t desc)
{
	return;
} /* stats_rcvMulticast */


/*------------------------------------------------------------------------------
 * Name         : stats_initModule
 *
 * Description  : Called by EEE to get stats agent ready to go
 *
 * Created by   : Shyh-Pei Yen
 *
 * Date Created : 14-AUG-2001
 *
 *------------------------------------------------------------------------------
 */

int32
stats_initModule(void)
{
	int32	statsTimer;
	int32   rc = NFX_OK;

	stats_trace(LOG_DEBUG, "> stats_initModule()\n");

	// Register our message handler.
	statsAppId = eee_allocateApplicationId("stats");
	if (statsAppId == EEE_ERR_INVALID_APP_NAME)
		return NFX_ERR;

	rc = eee_registerReceiveFunc(statsAppId, stats_rcvMsg, stats_rcvMulticast);
	if (rc != NFX_OK)
		return NFX_ERR;

	// Initialize the stats agent port address.
	statsAgentPort = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
							nfxMyCPUId, statsAppId,
							EEE_MGMT, NFX_COS_HIGH, PORT_BEST_FIT);
	NFX_SPIN_INIT(&stats_lock, SPINLOCK_LEVEL_STATS);

	init_statsCounters();
    statsTimer = eee_timerCreate(STATS_TICKS, EEE_TIMER_REPEAT);
    NFX_ASSERT(statsTimer != TIMER_ERR_NO_TIMER_AVAILABLE);
  
    rc = eee_timerStart(statsTimer, statsTimerProc, NULL, NULL, NULL, NULL);
    NFX_ASSERT(rc == NFX_OK);

	stats_trace(LOG_DEBUG, "< stats_initModule()\n");

	return NFX_OK;
}
