/*
 *------------------------------------------------------------------------------
 *
 * Name:	tape-device.c
 *
 * RCS:		$Id: tape-device.c,v 1.16 2002/11/01 19:29:15 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Tape driver device specific code.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 10/2/01
 *
 *------------------------------------------------------------------------------
 */

#include "tape-common.h"
#include "tape-device.h"
#include "../sm-scsi/scsi-dev-cmds.h"
#include <ctype.h>

/* 
 ********************************************************************************
 * tape_parse_devicename -- 
 *
 * Device names are ASCII strings of the form.
 * 
 * <{name}>_<{rewind-on-close-attribute}>_<{unload-on-close-attribute}>_<{compression_attribute}>
 * 
 * where {name} = device WWN or
 *                device alias
 *       {rewind-on-close-attribute} = R or NR (R is the default)
 *       {unload-on-close-attribute} = U or NU (NU is the default)
 *       {compression_attribute} = NC or C or HI or LO or MED (NC is the default, C is synonym for HI)
 *
 * On exit, attributes will be set to the parse values.
 *      and id will be set to the WWN of the device.
 *
 * The return code is NFX_OK, or NFX_ERR.
 ********************************************************************************
 */
int
tape_parse_devicename(char *devicename, scsi_id_t *id, uint32 *attributes)
{
#define FIELD_SEPARATOR '_'

	char *		p = devicename;
	char		tmp_buf[64];
	char *		tmp_p;
	int		density;
	tape_dev_t *	dev = NULL;
	int		alias;

	*attributes = 0;
	*id = 0;

	if (*p == 'N' || *p == 'n' || *p == 'R' || *p == 'r' || *p == 'U' || *p == 'u') {
		/* Parse as alias name */
		/* Parse rewind/unload attributes */
		while (*p == 'N' || *p == 'n' || *p == 'R' || *p == 'r' || *p == 'U' || *p == 'u') {
			/* Get next attribute token */
			tmp_p = tmp_buf;
			if (*p == 'N' || *p == 'n' || *p == 0) 
				*tmp_p++ = *p++;
			*tmp_p++ = *p++;
			*tmp_p = 0;

			/* Convert to lower case */
			tmp_p = tmp_buf;
			while (*tmp_p) {
				*tmp_p = tolower(*tmp_p);
				++tmp_p;
			}
			/* Parse attribute */
			if (strcmp(tmp_buf, "nr") == 0)
				*attributes &= ~TAPE_ATTRIB_REWIND_ON_CLOSE;
			else
			if (strcmp(tmp_buf, "r") == 0)
				*attributes |= TAPE_ATTRIB_REWIND_ON_CLOSE;
			else
			if (strcmp(tmp_buf, "nu") == 0)
				*attributes &= ~TAPE_ATTRIB_UNLOAD_ON_CLOSE;
			else
			if (strcmp(tmp_buf, "u") == 0)
				*attributes |= TAPE_ATTRIB_UNLOAD_ON_CLOSE;
		}

		/* Parse alias */
		tmp_p = tmp_buf;
		while (*p != 0 && isdigit(*p))
			*tmp_p++ = *p++;
		*tmp_p = 0;
		alias = atol(tmp_buf);
		dev = tape_device_lookup_by_alias(alias);
		if (dev == NULL)
			return(NFX_ERR);
			
		/* Parse density code */
		tmp_p = tmp_buf;
		while (*p != 0)
			*tmp_p++ = *p++;
		*tmp_p = 0;
		if ((density = tape_density_code_match(dev->cap, tmp_buf)) != NFX_ERR) {
			*attributes &= ~TAPE_ATTRIB_DENSITY_MASK;
			*attributes |= (density << TAPE_ATTRIB_DENSITY_SHIFT);
		}
		else
			return(NFX_ERR);
		*id = dev->id;
	}
	else {
		/* Parse as full physical name */
		/* Parse WWN */
		while (*p != 0 && *p != FIELD_SEPARATOR) {
			if (*p == 'x' || *p == 'X')
				;
			else
			if (isdigit(*p))
				*id = (*id << 4) | (*p - '0');
			else
			if (isxdigit(*p) && isupper(*p))
				*id = (*id << 4) | (*p - 'A' + 10);
			else
			if (isxdigit(*p) && islower(*p))
				*id = (*id << 4) | (*p - 'a' + 10);
			else
				return(NFX_ERR);
			++p;
		}
		if (*p == 0)
			return(NFX_OK);
		++p;

		dev = tape_device_lookup_by_id(*id);
		if (dev == NULL)
			return(NFX_ERR);

		while (*p != 0) {
			/* Get next token */
			tmp_p = tmp_buf;
			while (*p != 0 && *p != FIELD_SEPARATOR)
				*tmp_p++ = *p++;
			*tmp_p = 0;
			if (*p == FIELD_SEPARATOR)
				++p;
		
			/* Convert to lower case */
			tmp_p = tmp_buf;
			while (*tmp_p) {
				*tmp_p = tolower(*tmp_p);
				++tmp_p;
			}

			/* Parse attribute */
			if (strcmp(tmp_buf, "nr") == 0)
				*attributes &= ~TAPE_ATTRIB_REWIND_ON_CLOSE;
			else
			if (strcmp(tmp_buf, "r") == 0)
				*attributes |= TAPE_ATTRIB_REWIND_ON_CLOSE;
			else
			if (strcmp(tmp_buf, "nu") == 0)
				*attributes &= ~TAPE_ATTRIB_UNLOAD_ON_CLOSE;
			else
			if (strcmp(tmp_buf, "u") == 0)
				*attributes |= TAPE_ATTRIB_UNLOAD_ON_CLOSE;
			else 
			if ((density = tape_density_code_match(dev->cap, tmp_buf)) != NFX_ERR) {
				*attributes &= ~TAPE_ATTRIB_DENSITY_MASK;
				*attributes |= (density << TAPE_ATTRIB_DENSITY_SHIFT);
			}
			else
				return(NFX_ERR);
		}
	}
	return(NFX_OK);
}

/* 
 ********************************************************************************
 * mc_parse_devicename -- 
 *
 * Device names are ASCII strings of the form.
 * 
 * <{name}>
 * 
 * where {name} = device WWN or
 *                device alias
 *
 * On exit, id will be set to the WWN of the device.
 *
 * The return code is NFX_OK, or NFX_ERR.
 ********************************************************************************
 */
int
mc_parse_devicename(char *devicename, scsi_id_t *id)
{
	char *		p = devicename;
	tape_dev_t *	dev = NULL;
	int		alias;
	char		tmp_buf[64];
	char *		tmp_p;

#define MC_ALIAS_PREFIX "mc"

	*id = 0;

	if (*p == 'm' || *p == 'M') {
		/* Parse as alias name */
		/* Parse prefix */
		++p;
		if (*p != 'c' && *p != 'C')
			return(NFX_ERR);
		++p;

		/* Parse alias */
		tmp_p = tmp_buf;
		while (*p != 0 && isdigit(*p))
			*tmp_p++ = *p++;
		*tmp_p = '\0';
		if (strlen(tmp_buf) == 0)
			return(NFX_ERR);
		alias = atol(tmp_buf);
		dev = tape_device_lookup_by_alias(alias);
		if (dev == NULL)
			return(NFX_ERR);
		*id = dev->id;
	}
	else {
		/* Parse as full physical name */
		/* Parse WWN */
		while (*p != 0) {
			if (*p == 'x' || *p == 'X')
				;
			else
			if (isdigit(*p))
				*id = (*id << 4) | (*p - '0');
			else
			if (isxdigit(*p) && isupper(*p))
				*id = (*id << 4) | (*p - 'A' + 10);
			else
			if (isxdigit(*p) && islower(*p))
				*id = (*id << 4) | (*p - 'a' + 10);
			else
				return(NFX_ERR);
			++p;
		}
	}
	dev = tape_device_lookup_by_id(*id);
	if (dev == NULL)
		return(NFX_ERR);
	return(NFX_OK);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_tur_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_TEST_UNIT_READY;

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_mode_sense_compression_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_MODE_SENSE_6;
	req_msg->cdb[2] = SCSI_MPAGE_CONTROL_CURRENT | SCSI_MPAGE_COMPRESSION;
	req_msg->cdb[4] = 0xFF;

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_mode_select_compression_msg(tape_dev_t *dev, sdm_relayCdbRspMsg_t *mode_sense_resp)
{
	sdm_relayCdbReqMsg_t *	req_msg;
	scsi_mode_header_t *	mode_hdr;
	scsi_mode_block_desc_t *mode_block_desc;
	scsi_mode_page_desc_t * mode_page;
	int			data_len;
	int			density = (dev->open_attributes & TAPE_ATTRIB_DENSITY_MASK) >> TAPE_ATTRIB_DENSITY_SHIFT;
	int			compression;
	byte			comp_byte;

	data_len = sizeof(scsi_mode_header_t) + 
		sizeof(scsi_mode_block_desc_t) + 
		2 + sizeof(scsi_mode_compress_mode_t);

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t) + data_len);
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_MODE_SELECT_6;
	req_msg->cdb[1] = 0x10;		/* PF bit */
	req_msg->cdb[4] = data_len;

	req_msg->dataLen = data_len;

	/* Copy mode sense data into mode select buffer */
	bcopy(mode_sense_resp->data, req_msg->data, data_len);

	/* Set up mode header */
	mode_hdr = (scsi_mode_header_t *)&req_msg->data[0];
	mode_hdr->length = 0;				/* Reserved for mode select */
	mode_hdr->flag &= ~0x80;			/* WP bit reserved for mode select */

	/* Set up mode block descriptor */	
	mode_block_desc = (scsi_mode_block_desc_t *)(mode_hdr + 1);
	VERIFY(mode_hdr->block_desc_len == sizeof(scsi_mode_block_desc_t));
	if (dev->cap->densities[density].density_code != DEFAULT_DENSITY)
		mode_block_desc->density_code = dev->cap->densities[density].density_code;
	mode_block_desc->blocks[2] = 0;			/* Reserved for mode select */
	mode_block_desc->blocks[1] = 0;			/* Reserved for mode select */	       
	mode_block_desc->blocks[0] = 0;			/* Reserved for mode select */
	mode_block_desc->block_size[2] = 0;		/* Variable block mode by default */
	mode_block_desc->block_size[1] = 0;		/* Variable block mode by default */
	mode_block_desc->block_size[0] = 0;		/* Variable block mode by default */

	/* Set up page contents */
	mode_page = (scsi_mode_page_desc_t *)(mode_block_desc + 1);
	mode_page->code &= ~0xC0; /* Reserved for mode select */
	compression = dev->cap->densities[density].compression;
	comp_byte = mode_page->u.compress_mode.compression_enab;
	if (comp_byte & SCSI_CM_COMPRESSION_CAPABLE) {
		if (compression) {
			comp_byte |= SCSI_CM_ENABLE_COMPRESSION;
			mode_page->u.compress_mode.compression_algorithm[3] = compression;
		}
		else
			comp_byte &= ~SCSI_CM_ENABLE_COMPRESSION;
		mode_page->u.compress_mode.compression_enab = comp_byte;
	}

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_mode_select_dlteeprom_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;
	char *			data;
	int			data_len;

	data_len = 8 + 8 + 2 + 13;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t) + data_len);
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	req_msg->dataLen = data_len;
	
	req_msg->cdb[0] = 0x55;		/* SCSI_CMD_MODE_SELECT_10 */
	req_msg->cdb[1] = 0x10;		/* PF bit */
	req_msg->cdb[7] = 0;
	req_msg->cdb[8] = data_len;

	data = req_msg->data;

	/* Mode select (10) mode parameter header */
	data[0] = 0;
	data[1] = 0;
	data[2] = 0;
	data[3] = 0x10;
	data[4] = 0;
	data[5] = 0;
	data[6] = 0;
	data[7] = 8;

	/* Mode parameter block descriptor */
	data[8] = 0;
	data[9] = 0;
	data[10] = 0;
	data[11] = 0;
	data[12] = 0;
	data[13] = 0;
	data[14] = 0;
	data[15] = 0;

	data[16] = 0x3E;
	data[17] = 13;
	data[18+0 ] = 'S';
	data[18+1 ] = 'E';
	data[18+2 ] = 'T';
	data[18+3 ] = 'E';
	data[18+4 ] = 'O';
	data[18+5 ] = 'M';
	data[18+6 ] = 'A';
	data[18+7 ] = 'T';
	data[18+8 ] = 'E';
	data[18+9 ] = 'W';
	data[18+10] = ' ';
	data[18+11] = '1';
	data[18+12] = 0;

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_rewind_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_REWIND_UNIT;

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_unload_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_LOAD_UNLOAD;

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_load_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_LOAD_UNLOAD;
	req_msg->cdb[4] = 0x01;	/* load */

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_read_position_msg(tape_dev_t *dev)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;
	
	req_msg->cdb[0] = SCSI_CMD_READ_POSITION;
	req_msg->cdb[1] = SCSI_RPOS_CONTROL_LONG | SCSI_RPOS_CONTROL_TCLP;

	return(req_msg);

}

sdm_relayCdbReqMsg_t *
tape_build_scsi_space_msg(tape_dev_t *dev, tape_space_req_msg_t *msg)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	if (msg->opcode == TAPE_SPACE_OPCODE_REWIND)
		req_msg = tape_build_scsi_rewind_msg(dev);
	else {
		req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
		VERIFY(req_msg);
		bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

		req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
		req_msg->hdr.chandle = tape_device_to_chandle(dev);
		req_msg->handle = dev->open_handle;

		req_msg->cdb[0] = SCSI_CMD_SPACE;

		switch (msg->opcode) {
		case TAPE_SPACE_OPCODE_BLOCKS:
			req_msg->cdb[1] = SCSI_SPACE_CODE_BLOCKS; break;
		case TAPE_SPACE_OPCODE_FILEMARKS:
			req_msg->cdb[1] = SCSI_SPACE_CODE_FMS; break;
		case TAPE_SPACE_OPCODE_END_OF_DATA:
			req_msg->cdb[1] = SCSI_SPACE_CODE_EOD; break;
		default:
			VERIFY(0);
		}

		req_msg->cdb[2] = ((uint32)msg->count & 0x00FF0000) >> 16;
		req_msg->cdb[3] = ((uint32)msg->count & 0x0000FF00) >>  8;
		req_msg->cdb[4] = ((uint32)msg->count & 0x000000FF);
	}
	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_write_fms_msg(tape_dev_t *dev, tape_write_fm_req_msg_t *msg)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));
	
	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;

	req_msg->cdb[0] = SCSI_CMD_WRITE_FILEMARKS;

	req_msg->cdb[2] = ((uint32)msg->count & 0x00FF0000) >> 16;
	req_msg->cdb[3] = ((uint32)msg->count & 0x0000FF00) >>  8;
	req_msg->cdb[4] = ((uint32)msg->count & 0x000000FF);
	
	return(req_msg);
}

sdm_readDataReqMsg_t *
tape_build_scsi_read_msg(tape_dev_t *dev, tape_read_req_msg_t *msg)
{
	sdm_readDataReqMsg_t *	req_msg;

	req_msg = (sdm_readDataReqMsg_t *)malloc(sizeof(sdm_readDataReqMsg_t));
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_readDataReqMsg_t));

	req_msg->hdr.mid = SDM_MID_READ_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;

	req_msg->blkCnt = msg->bytecount;

	return(req_msg);
}

sdm_writeDataReqMsg_t *
tape_build_scsi_write_msg(tape_dev_t *dev, tape_write_req_msg_t *msg)
{
	sdm_writeDataReqMsg_t *	req_msg;

	req_msg = (sdm_writeDataReqMsg_t *)malloc(sizeof(sdm_writeDataReqMsg_t) + msg->bytecount);
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_writeDataReqMsg_t));
	
	req_msg->hdr.mid = SDM_MID_WRITE_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;

	req_msg->blkCnt = msg->bytecount;	
	bcopy(msg->data, req_msg->data, msg->bytecount);

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
tape_build_scsi_passthru_msg(tape_dev_t *dev, tape_passthru_req_msg_t *msg)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	if (msg->flags & TAPE_PASSTHRU_FLAGS_DATA_OUT)
		req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t) + msg->data_len);
	else
		req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
							 
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	/* Copy CDB */
	bcopy(msg->cdb, req_msg->cdb, sizeof(req_msg->cdb));

	/* Copy Data Out */
	if (msg->flags & TAPE_PASSTHRU_FLAGS_DATA_OUT) {
		bcopy(msg->data, req_msg->data, msg->data_len);
		req_msg->dataLen = msg->data_len;
	}
	
	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;

	return(req_msg);
}

sdm_relayCdbReqMsg_t *
mc_build_scsi_passthru_msg(tape_dev_t *dev, mc_passthru_req_msg_t *msg)
{
	sdm_relayCdbReqMsg_t *	req_msg;

	if (msg->flags & MC_PASSTHRU_FLAGS_DATA_OUT)
		req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t) + msg->data_len);
	else
		req_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t));
							 
	VERIFY(req_msg);
	bzero(req_msg, sizeof(sdm_relayCdbReqMsg_t));

	/* Copy CDB */
	bcopy(msg->cdb, req_msg->cdb, sizeof(req_msg->cdb));

	/* Copy Data Out */
	if (msg->flags & MC_PASSTHRU_FLAGS_DATA_OUT) {
		bcopy(msg->data, req_msg->data, msg->data_len);
		req_msg->dataLen = msg->data_len;
	}
	
	req_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	req_msg->hdr.chandle = tape_device_to_chandle(dev);
	req_msg->handle = dev->open_handle;

	return(req_msg);
}

/*
 * chandle structure is:
 *   bits 0  - 15 : device index
 *   bits 16 - 31 : process ID of tape driver
 */
uint32
tape_device_to_chandle(tape_dev_t *dev)
{
	uint32	chandle = 0;
	
	chandle = ((getpid() & 0xFFFF) << 16) | dev->index;
	return(chandle);
}

tape_dev_t *
tape_chandle_to_device(uint32 chandle)
{
	tape_dev_t *	dev;

	/* Is this out PID? */
	if (((chandle >> 16) & 0xFFFF) != (getpid() & 0xFFFF))
		return(NULL);
	
	dev = tape_device_lookup_by_index(chandle & 0xFFFF);
	return(dev);
}
