/*
 *------------------------------------------------------------------------------
 *
 * Name:	tape-sniffer.c
 *
 * RCS:		$Id: tape-sniffer.c,v 1.8 2002/09/11 02:59:29 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Tape sniffer application.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 10/19/01
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
//#include <string.h> - Isabelle
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/anpssc-api.h>
#include "tape-api.h"
#include "tape-debug.h"

#define MAX_IO_LEN	100000
#define SNIFF_IO_LEN	(MAX_IO_LEN - sizeof(tape_read_rsp_msg_t))
#define TAPE_SNIFFER_CLIENT_TAG 0x87654321

#define FAILED()	printf("FAILED\n");
#define SUCCEEDED()	printf("SUCCEEDED\n");

int interrupt = 0;

int send_tape_message(char *, int, char *, int *, int);
char *tape_ecode_to_string(uint32);
void tape_terminate_handler(int);
int tape_open(char *);
int tape_close(char *);
int tape_rewind(int);
int tape_read(int, int, int *, int *);

static void 
print_usage(char *command)
{
	fprintf(stderr, "Usage: %s [-d] <tape-devicename>\n", command);
}

int
tape_open(char *devicename)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	tape_open_req_msg_t *	open_req;
	tape_open_rsp_msg_t *	open_rsp;

	open_req = (tape_open_req_msg_t *)req_msg;
	open_req->hdr.mid = TAPE_MID_OPEN_REQ;
	open_req->hdr.cookie = 1;
	strcpy(open_req->devicename, devicename);
	open_req->client_tag = TAPE_SNIFFER_CLIENT_TAG;
			
	req_len = sizeof(tape_open_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 60);
	if (rc < 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) failed.\n", open_req->hdr.mid));
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) timed out.\n", open_req->hdr.mid));
		return(NFX_ERR);
	}

	open_rsp = (tape_open_rsp_msg_t *)rsp_msg;
	if (open_rsp->hdr.mid != TAPE_MID_OPEN_RSP) {
		TP_DBG(1, ("Invalid response mid; expected 0x%x got 0x%x\n",
			   TAPE_MID_OPEN_RSP, open_rsp->hdr.mid));
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_open_rsp_msg_t)) {
		TP_DBG(1, ("Invalid response length on mid 0x%x: expected %d got %d\n", 
			   open_req->hdr.mid, sizeof(tape_open_rsp_msg_t), rsp_len));
		return(NFX_ERR);
	}
	if (open_rsp->ecode != TAPE_ECODE_SUCCESS) {
		TP_DBG(1, ("tape open failed: ecode %s(0x%x)\n", 
			   tape_ecode_to_string(open_rsp->ecode), open_rsp->ecode));
		return(NFX_ERR);
	}

	return(open_rsp->handle);
}

int
tape_close(char *devicename)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	tape_close_req_msg_t *	close_req;
	tape_close_rsp_msg_t *	close_rsp;

	close_req = (tape_close_req_msg_t *)req_msg;
	close_req->hdr.mid = TAPE_MID_CLOSE_REQ;
	close_req->hdr.cookie = 1;
	strcpy(close_req->devicename, devicename);
	close_req->client_tag = TAPE_SNIFFER_CLIENT_TAG;
			
	req_len = sizeof(tape_close_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 60);
	if (rc < 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) failed.\n", close_req->hdr.mid));
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) timed out.\n", close_req->hdr.mid));
		return(NFX_ERR);
	}

	close_rsp = (tape_close_rsp_msg_t *)rsp_msg;
	if (close_rsp->hdr.mid != TAPE_MID_CLOSE_RSP) {
		TP_DBG(1, ("Invalid response mid; expected 0x%x got 0x%x\n",
			   TAPE_MID_CLOSE_RSP, close_rsp->hdr.mid));
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_close_rsp_msg_t)) {
		TP_DBG(1, ("Invalid response length on mid 0x%x: expected %d got %d\n", 
			   close_req->hdr.mid, sizeof(tape_close_rsp_msg_t), rsp_len));
		return(NFX_ERR);
	}
	if (close_rsp->ecode != TAPE_ECODE_SUCCESS) {
		TP_DBG(1, ("tape close failed: ecode %s(0x%x)\n", 
			   tape_ecode_to_string(close_rsp->ecode), close_rsp->ecode));
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
tape_rewind(int handle)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_space_req_msg_t *		req;
	tape_space_rsp_msg_t *		rsp;

	req = (tape_space_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_SPACE_REQ;
	req->hdr.cookie = 1;
	req->handle = handle;
	req->opcode = TAPE_SPACE_OPCODE_REWIND;

	req_len = sizeof(tape_space_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 60);
	if (rc < 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid));
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid));
		return(NFX_ERR);
	}

	rsp = (tape_space_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_SPACE_RSP) {
		TP_DBG(1, ("Invalid response mid; expected 0x%x got 0x%x\n",
			   TAPE_MID_SPACE_RSP, rsp->hdr.mid));
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_space_rsp_msg_t)) {
		TP_DBG(1, ("Invalid response length on mid 0x%x: expected %d got %d\n", 
			   req->hdr.mid, sizeof(tape_space_rsp_msg_t), rsp_len));
		return(NFX_ERR);
	}
	if (rsp->ecode != TAPE_ECODE_SUCCESS) {
		TP_DBG(1, ("tape rewind failed: ecode %s(0x%x)\n", 
			   tape_ecode_to_string(rsp->ecode), rsp->ecode));
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
tape_read(int handle, int bytecount, int *residue, int *ecode)
{
        int				rc;
	char 				req_msg[1024];
	char *				rsp_msg;
	int				req_len = 0;
	int				rsp_len = 0;
	tape_read_req_msg_t *		req;
	tape_read_rsp_msg_t *		rsp;

	req_len = sizeof(tape_read_req_msg_t);
	rsp_len = sizeof(tape_read_rsp_msg_t) + bytecount;

	if (rsp_len > MAX_IO_LEN) {
		TP_DBG(1, ("tape read failed: data length longer than %d.\n", MAX_IO_LEN - sizeof(tape_read_rsp_msg_t)));
		return(NFX_ERR);
	}

	rsp_msg = malloc(rsp_len);

	req = (tape_read_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_READ_REQ;
	req->hdr.cookie = 1;
	req->handle = handle;
	req->bytecount = bytecount;
	
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 60);
	if (rc < 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid));
		goto read_error;
	}
	else 
	if (rc == 0) {
		TP_DBG(1, ("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid));
		goto read_error;
	}

	rsp = (tape_read_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_READ_RSP) {
		TP_DBG(1, ("Invalid response mid; expected 0x%x got 0x%x\n",
			   TAPE_MID_READ_RSP, rsp->hdr.mid));
		goto read_error;
	}
	if (rsp_len < sizeof(tape_read_rsp_msg_t)) {
		TP_DBG(1, ("Invalid response length on mid 0x%x: expected at least %d got %d\n", 
			   req->hdr.mid, sizeof(tape_read_rsp_msg_t), rsp_len));
		goto read_error;
	}

	*ecode = rsp->ecode;
	*residue = rsp->residue;
		
	free(rsp_msg);
	return(NFX_OK);

 read_error: 
	free(rsp_msg);
	return(NFX_ERR);
}

int 
main(int argc, char **argv)
{
	int	i;
	char *	devicename = NULL;
	int	handle;
	int	rc;
	int	done = 0;
	int	blockno = 0;
	int	fileno = 0;
	int	request_len = SNIFF_IO_LEN;
	int	ecode;
	int	residue;

	if (argc < 2)
		goto arg_error;

	for (i = 1; i < argc; ++i) {
		if (argv[i][0] == '-')
			switch(argv[i][1]) {
			case 'd': 
				++tape_dbg_level;
				break;
			default:
				goto arg_error;
			}
		else
		if (devicename == NULL)
			devicename = argv[1];
		else
			goto arg_error;
	}

	if (devicename == NULL)
		goto arg_error;

	/* Register SIGTERM handler */
	signal(SIGKILL, tape_terminate_handler);
	signal(SIGINT, tape_terminate_handler);
	signal(SIGTERM, tape_terminate_handler);

	/*
	 ********************************************
	 * OPEN device
	 ********************************************
	 */
	printf("OPENING %s ... ", devicename);
	handle = tape_open(devicename);
	if (handle == NFX_ERR) {
		FAILED();
		exit(1);
	}
	else
		SUCCEEDED();

	/*
	 ********************************************
	 * REWIND device
	 ********************************************
	 */
	printf("REWINDING ... ");
	rc = tape_rewind(handle);
	if (rc == NFX_ERR) {
		FAILED();
		goto do_close;
	}
	else
		SUCCEEDED();

	/* 
	 ********************************************
	 * SNIFF the tape 
	 ********************************************
	 */
	while (!done && !interrupt) {
		rc = tape_read(handle, request_len, &residue, &ecode);
		if (rc == NFX_ERR) {
			printf("TAPE IO FAILED ...\n");
			++done;
		}
		switch(ecode) {
		case TAPE_ECODE_SUCCESS:
			residue = 0;
		case TAPE_ECODE_ILI:
			printf("Block %d: len %d\n", blockno++, request_len - residue);
			break;

		case TAPE_ECODE_FM_ENCOUNTERED:
			printf("Filemark %d\n", fileno++);
			blockno = 0;
			break;
		     
		case TAPE_ECODE_EOD_ENCOUNTERED:
		case TAPE_ECODE_EOM_ENCOUNTERED:
			printf("DONE\n");
			++done;
			break;

		default:
			printf("TAPE IO FAILED ...\n");
			++done;
		}
	}

 do_close:

	/*
	 ********************************************
	 * CLOSE device
	 ********************************************
	 */
	printf("CLOSING ... ");
	handle = tape_close(devicename);
	if (handle == NFX_ERR) {
		FAILED();
		exit(1);
	}
	else
		SUCCEEDED();

	exit(interrupt);

arg_error:
	print_usage(argv[0]);
	exit(1);
}

int
send_tape_message(char *req_buf, int req_len, char *resp_buf, int *resp_len, int timeout)
{
	int		stat;
	struct timeval	to;

	to.tv_sec = timeout;
	to.tv_usec = 0;
	
	stat = sendAgileMsg(-1, 
			    req_buf, req_len,
			    "tape-driver", 0, 0, 0, &to,
			    resp_buf, *resp_len);
	if (stat > 0)
		*resp_len = stat;
	return(stat);
}

char *
tape_ecode_to_string(uint32 ecode)
{
	switch(ecode) {
	case TAPE_ECODE_SUCCESS:		return("SUCCESS");
	case TAPE_ECODE_DEVICE_BUSY:		return("DEVICE_BUSY");
	case TAPE_ECODE_NO_DEVICE:		return("NO_DEVICE");
	case TAPE_ECODE_NO_TAPE_LOADED:		return("NO_TAPE_LOADED");
	case TAPE_ECODE_ILI:			return("ILI");
	case TAPE_ECODE_DATA_SIZE_MISMATCH:	return("DATA_SIZE_MISMATCH");
	case TAPE_ECODE_MEDIUM_ERR:		return("MEDIUM_ERR");
	case TAPE_ECODE_EOM_ENCOUNTERED:	return("EOM_ENCOUNTERED");
	case TAPE_ECODE_EOD_ENCOUNTERED:	return("EOD_ENCOUNTERED");
	case TAPE_ECODE_FM_ENCOUNTERED:		return("FM_ENCOUNTERED");
	case TAPE_ECODE_IO_ERROR:		return("IO_ERROR");
	case TAPE_ECODE_DEVICE_NOT_OPEN:	return("DEVICE_NOT_OPENED");
	default:				return("???");
	}
}


void 
tape_terminate_handler(int sig)
{
	TP_DBG(0, ("tape_terminate_handler: signal %d\n", sig));
	interrupt = 1;
}

