/*
 *------------------------------------------------------------------------------
 *
 * Name:	tape-ui.c
 *
 * RCS:		$Id: tape-ui.c,v 1.20 2002/11/07 23:20:57 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Tape driver UI.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 10/1/01
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
//#include <string.h> - Isabelle
#include <sys/types.h>
#include <sys/time.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/anpssc-api.h>
#include "tape-api.h"
#include "tape-common.h"

#define TAPE_CLIENT_TAG 0x12345678
#define MAX_IO_LEN 256000

int send_tape_message(char *, int, char *, int *, int);
char *tape_ecode_to_string(uint32);

int32 tape_ui_devname_to_handle(char *, uint32 *);

int tape_ui_devlist(int, char **, int, ...);
int tape_ui_open(int, char **, int, ...);
int tape_ui_close(int, char **, int, ...);
int tape_ui_devstate(int, char **, int, ...);
int tape_ui_space(int, char **, int, ...);
int tape_ui_write_fms(int, char **, int, ...);
int tape_ui_load(int, char **, int, ...);
int tape_ui_read(int, char **, int, ...);
int tape_ui_write(int, char **, int, ...);
int tape_ui_dbg_level(int, char **, int, ...);

typedef struct {
	char *	name;
	int	min_parm_cnt;	/* Min command-line parameter count */
	int	max_parm_cnt;	/* Max command-line parameter count */
	int	implicit_arg1;	/* Implicit argument, derived from command line parms */
	int	implicit_arg2;	/* Implicit argument, derived from command line parms */
	int (*func)(int, char **, int, ...);
	int	timeout;
	char *	usage;
} tape_ui_node_t;

tape_ui_node_t tape_ui_root[] = {
	{ "devlist",	      0,0,  0, 0,					tape_ui_devlist,	5,
	  "devlist" },
	{ "open",	      1,1,  0, 0,					tape_ui_open,		120,
	  "open <device-name>" },
	{ "close",	      1,1,  0, 0,					tape_ui_close,		120,
	  "close <device-name>" },
	{ "devstate",	      1,1,  0, 0,					tape_ui_devstate,	5,
	  "devstate <device-name>" },
	{ "fsr",	      2,2,  TAPE_SPACE_OPCODE_BLOCKS, 1,		tape_ui_space,		120,
	  "fsr <device-name> <count> -- forward space block(s)" },
	{ "bsr",	      2,2,  TAPE_SPACE_OPCODE_BLOCKS, -1,		tape_ui_space,		120,
	  "bsr <device-name> <count> -- backward space block(s)" },
	{ "fsf",	      2,2,  TAPE_SPACE_OPCODE_FILEMARKS, 1,		tape_ui_space,		120,
	  "fsf <device-name> <count> -- forward space filemark(s)" },
	{ "bsf",	      2,2,  TAPE_SPACE_OPCODE_FILEMARKS, -1,		tape_ui_space,		120,
	  "bsf <device-name> <count> -- backward space filemark(s)" },
	{ "eod",	      1,1,  TAPE_SPACE_OPCODE_END_OF_DATA, 0,		tape_ui_space,		120,
	  "eod <device-name> -- move forward to end-of-(recorded)data" },
	{ "rewind",	      1,1,  TAPE_SPACE_OPCODE_REWIND, 0,		tape_ui_space,		120,
	  "rewind <device-name> -- move backward to BOT" },
	{ "wfm",	      2,2,  0, 0,					tape_ui_write_fms,	120,
	  "wfm <device-name> <count> -- write filemark(s)" },
	{ "load",	      1,1,  1, 0,					tape_ui_load,		120,
	  "load <device-name> -- load device" },
	{ "unload",	      1,1,  0, 0,					tape_ui_load,		120,
	  "unload <device-name> -- unload device" },
	{ "read",	      2,3,  0, 0,					tape_ui_read,		120,
	  "read <device-name> [-b] <bytecount> -- read tape" },
	{ "write",	      3,3,  0, 0,					tape_ui_write,		120,
	  "write <device-name> <bytecount> <pattern> -- write tape" },
	{ "dbg_level",	      1,1,  0, 0,					tape_ui_dbg_level,	5,
	  "dbg_level <value> -- set driver debug level" },
	{ NULL,		      0,0,  0, 0,					NULL,			0,
	  NULL }
};
	
static void 
print_usage(char *command)
{
	int i;

	for (i = 0; tape_ui_root[i].name; ++i) {
		fprintf(stderr, "%s %s %s\n",
			(i == 0 ? "Usage:" : "      "),
			command, tape_ui_root[i].usage);
	}
}

int32
tape_ui_devname_to_handle(char *devname, uint32 *handle)
{
	int			    rc;
	char			    req_msg[1024];
	char			    rsp_msg[1024];
	int			    req_len = 0;
	int			    rsp_len = 0;
	tape_get_handle_req_msg_t * req;
	tape_get_handle_rsp_msg_t * rsp;

	req = (tape_get_handle_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_GET_HANDLE_REQ;
	req->hdr.cookie = 1;
	strcpy(req->devicename, devname);
			
	req_len = sizeof(tape_get_handle_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, 5);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (tape_get_handle_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_GET_HANDLE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_GET_HANDLE_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_get_handle_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, 
		       sizeof(tape_get_handle_rsp_msg_t), 
		       rsp_len);
		return(NFX_ERR);
	}
	if (rsp->ecode == TAPE_ECODE_SUCCESS)
		*handle = rsp->handle;
	return(rsp->ecode);
}

int
tape_ui_devlist(int argc, char **argv, int timeout, ...)
{
	int			     rc;
	char			     req_msg[1024];
	char			     rsp_msg[TAPE_MAX_PHYSICAL * 16 * sizeof(tape_dev_info_t) + sizeof(tape_get_devlist_rsp_msg_t)];
	int			     req_len = 0;
	int			     rsp_len = 0;
	tape_get_devlist_req_msg_t * devlist_req;
	tape_get_devlist_rsp_msg_t * devlist_rsp;
	int			     i;
	tape_dev_info_t *	     dev_info;

	devlist_req = (tape_get_devlist_req_msg_t *)req_msg;
	devlist_req->hdr.mid = TAPE_MID_GET_DEVLIST_REQ;
	devlist_req->hdr.cookie = 1;
			
	req_len = sizeof(tape_get_devlist_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", devlist_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", devlist_req->hdr.mid);
		return(NFX_ERR);
	}

	devlist_rsp = (tape_get_devlist_rsp_msg_t *)rsp_msg;
	if (devlist_rsp->hdr.mid != TAPE_MID_GET_DEVLIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_GET_DEVLIST_RSP, devlist_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_get_devlist_rsp_msg_t) + sizeof(tape_dev_info_t) * (devlist_rsp->dev_count - 1)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       devlist_req->hdr.mid, 
		       sizeof(tape_get_devlist_req_msg_t) + sizeof(tape_dev_info_t) * (devlist_rsp->dev_count - 1),
		       rsp_len);
		return(NFX_ERR);
	}
	if (devlist_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape get_devlist failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(devlist_rsp->ecode), devlist_rsp->ecode);
		return(NFX_ERR);
	}

	printf("%d devices found.\n", devlist_rsp->dev_count);

	for (i = 0, dev_info = devlist_rsp->info; i < devlist_rsp->dev_count; ++i, ++dev_info) {
		printf("[%2d]: devicename %s\n", i, dev_info->devicename);
		printf("      physical_devicename %s\n", dev_info->physical_devicename);
		printf("      vendor '%s', product '%s'\n", dev_info->scsi_vendor_id, dev_info->scsi_product_id);
		printf("      scsi ID 0x%016llx\n", dev_info->scsi_id);
		printf("      attributes (0x%08x) %s/%s/%s\n",
		       dev_info->attributes,
		       ((dev_info->attributes & TAPE_ATTRIB_REWIND_ON_CLOSE) ? "REWIND" : "NOREWIND"),
		       ((dev_info->attributes & TAPE_ATTRIB_UNLOAD_ON_CLOSE) ? "UNLOAD" : "NOUNLOAD"),
		       (((dev_info->attributes & TAPE_ATTRIB_DENSITY_MASK) == TAPE_ATTRIB_DENSITY_0) ?
			"DENSITY_0" :
			(((dev_info->attributes & TAPE_ATTRIB_DENSITY_MASK) == TAPE_ATTRIB_DENSITY_1) ?
			 "DENSITY_1" :
			 (((dev_info->attributes & TAPE_ATTRIB_DENSITY_MASK) == TAPE_ATTRIB_DENSITY_2) ?
			  "DENSITY_2" :
			  "DENSITY_3"))));
		printf("\n");
	} 
	return(NFX_OK);
}

int
tape_ui_open(int argc, char **argv, int timeout, ...)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	tape_open_req_msg_t *	open_req;
	tape_open_rsp_msg_t *	open_rsp;

	open_req = (tape_open_req_msg_t *)req_msg;
	open_req->hdr.mid = TAPE_MID_OPEN_REQ;
	open_req->hdr.cookie = 1;
	strcpy(open_req->devicename, argv[1]);
	open_req->client_tag = TAPE_CLIENT_TAG;
			
	req_len = sizeof(tape_open_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", open_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", open_req->hdr.mid);
		return(NFX_ERR);
	}

	open_rsp = (tape_open_rsp_msg_t *)rsp_msg;
	if (open_rsp->hdr.mid != TAPE_MID_OPEN_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_OPEN_RSP, open_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_open_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       open_req->hdr.mid, sizeof(tape_open_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (open_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape open failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(open_rsp->ecode), open_rsp->ecode);
		return(NFX_ERR);
	}

	printf("Handle               : 0x%x\n", open_rsp->handle);
	printf("SCSI ID              : %016llx\n", open_rsp->scsi_id);
	printf("WWN                  : %016llx\n", open_rsp->wwn);
	printf("LUN                  : %d\n", open_rsp->lun);
	printf("Device name          : %s\n", open_rsp->devicename);
	printf("Physical device name : %s\n", open_rsp->physical_devicename);
	printf("Client tag           : 0x%x\n", open_rsp->client_tag);
	return(NFX_OK);
}

int
tape_ui_close(int argc, char **argv, int timeout, ...)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	tape_close_req_msg_t *	close_req;
	tape_close_rsp_msg_t *	close_rsp;

	close_req = (tape_close_req_msg_t *)req_msg;
	close_req->hdr.mid = TAPE_MID_CLOSE_REQ;
	close_req->hdr.cookie = 1;
	strcpy(close_req->devicename, argv[1]);
	close_req->client_tag = TAPE_CLIENT_TAG;
			
	req_len = sizeof(tape_close_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", close_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", close_req->hdr.mid);
		return(NFX_ERR);
	}

	close_rsp = (tape_close_rsp_msg_t *)rsp_msg;
	if (close_rsp->hdr.mid != TAPE_MID_CLOSE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_CLOSE_RSP, close_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_close_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       close_req->hdr.mid, sizeof(tape_close_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (close_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape close failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(close_rsp->ecode), close_rsp->ecode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int
tape_ui_devstate(int argc, char **argv, int timeout, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_get_state_req_msg_t *	get_state_req;
	tape_get_state_rsp_msg_t *	get_state_rsp;

	get_state_req = (tape_get_state_req_msg_t *)req_msg;
	get_state_req->hdr.mid = TAPE_MID_GET_STATE_REQ;
	get_state_req->hdr.cookie = 1;

	rc = tape_ui_devname_to_handle(argv[1], &get_state_req->handle);
	if (rc == NFX_ERR)
		return(NFX_ERR);
	else
	if (rc == TAPE_ECODE_NO_DEVICE) {
		printf("tape %s failed: device %s does not exist\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else
	if (rc == TAPE_ECODE_DEVICE_BUSY) {
		printf("tape %s failed: device %s is busy\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else 
	if (rc == TAPE_ECODE_DEVICE_NOT_OPEN) {
		printf("tape %s failed: device %s is not open\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
			
	req_len = sizeof(tape_get_state_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", get_state_req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", get_state_req->hdr.mid);
		return(NFX_ERR);
	}

	get_state_rsp = (tape_get_state_rsp_msg_t *)rsp_msg;
	if (get_state_rsp->hdr.mid != TAPE_MID_GET_STATE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_GET_STATE_RSP, get_state_rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_get_state_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       get_state_req->hdr.mid, sizeof(tape_get_state_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (get_state_rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape getstate failed: ecode %s(0x%x)\n", 
		       tape_ecode_to_string(get_state_rsp->ecode), get_state_rsp->ecode);
		return(NFX_ERR);
	}

	printf("Flags: %s, %s, density %d: %s %s %s\n",
	       ((get_state_rsp->flags & TAPE_STATE_FLAGS_REWIND_ON_CLOSE) ? "REWIND" : "NO REWIND"),
	       ((get_state_rsp->flags & TAPE_STATE_FLAGS_UNLOAD_ON_CLOSE) ? "UNLOAD" : "NO UNLOAD"),
	       ((get_state_rsp->flags & TAPE_STATE_FLAGS_DENSITY_MASK) >> TAPE_STATE_FLAGS_DENSITY_SHIFT),
	       ((get_state_rsp->flags & TAPE_STATE_FLAGS_WRT_PROTECT) ? "WRTP" : ""),
	       ((get_state_rsp->flags & TAPE_STATE_FLAGS_BOM) ? "BOM" : ""),
	       ((get_state_rsp->flags & TAPE_STATE_FLAGS_EOM) ? "EOM" : ""));

	if (get_state_rsp->flags & TAPE_STATE_FLAGS_FILENO_VALID)
		printf("File number: %d\n", get_state_rsp->fileno);
	else
		printf("File number: INVALID\n");

	if (get_state_rsp->flags & TAPE_STATE_FLAGS_BLOCKNO_VALID)
		printf("Block number: %d\n", get_state_rsp->blockno);
	else
		printf("Block number: INVALID\n");

	if (get_state_rsp->flags & TAPE_STATE_FLAGS_BLOCKSIZE_VALID)
		printf("Fixed block size: %d\n", get_state_rsp->blocksize);
	else
		printf("Fixed block size: INVALID\n");

	if (get_state_rsp->flags & TAPE_STATE_FLAGS_TOTAL_SPACE_VALID)
		printf("Total space: %lld\n", get_state_rsp->total_space);
	else
		printf("Total space: INVALID\n");

	if (get_state_rsp->flags & TAPE_STATE_FLAGS_SPACE_REMAINING_VALID)
		printf("Space remaining: %lld\n", get_state_rsp->space_remaining);
	else
		printf("Space remaining: INVALID\n");

	if (get_state_rsp->flags & TAPE_STATE_FLAGS_SOFT_ERRORS_VALID)
		printf("Soft Errors: %d\n", get_state_rsp->soft_errors);
	else
		printf("Soft Errors: INVALID\n");

	return(NFX_OK);
}

int 
tape_ui_space(int argc, char **argv, int timeout, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_space_req_msg_t *		req;
	tape_space_rsp_msg_t *		rsp;
	va_list				ap;
	int32				arg1, arg2;
	char *				endptr;

	va_start(ap, timeout);
	arg1 = va_arg(ap, int32);
	arg2 = va_arg(ap, int32);
	va_end(ap);

	req = (tape_space_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_SPACE_REQ;
	req->hdr.cookie = 1;

	rc = tape_ui_devname_to_handle(argv[1], &req->handle);
	if (rc == NFX_ERR)
		return(NFX_ERR);
	else
	if (rc == TAPE_ECODE_NO_DEVICE) {
		printf("tape %s failed: device %s does not exist\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else
	if (rc == TAPE_ECODE_DEVICE_BUSY) {
		printf("tape %s failed: device %s is busy\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else 
	if (rc == TAPE_ECODE_DEVICE_NOT_OPEN) {
		printf("tape %s failed: device %s is not open\n", argv[0], argv[1]);
		return(NFX_ERR);
	}

	req->opcode = arg1;
	if (req->opcode == TAPE_SPACE_OPCODE_BLOCKS || req->opcode == TAPE_SPACE_OPCODE_FILEMARKS) {
		req->count = arg2 * strtoul(argv[2], &endptr, 0);
		if (endptr == argv[2] || *endptr != 0 || endptr - argv[2] > 8) {
			printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[2]);
			return(NFX_ERR);
		}
	}
	else
		req->count = 0;

	req_len = sizeof(tape_space_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (tape_space_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_SPACE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_SPACE_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_space_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, sizeof(tape_space_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (rsp->ecode == TAPE_ECODE_EOM_ENCOUNTERED) {
		printf("tape %s succeeded with EOM_ENCOUNTERED: residue %d\n",
		       argv[0], rsp->residue);
	}
	else
	if (rsp->ecode == TAPE_ECODE_FM_ENCOUNTERED) {
		printf("tape %s succeeded with FM_ENCOUNTERED: residue %d\n",
		       argv[0], rsp->residue);
	}
	else
	if (rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape %s failed: ecode %s(0x%x)\n", 
		       argv[0], tape_ecode_to_string(rsp->ecode), rsp->ecode);
		return(NFX_ERR);
	}

	if (rsp->residue == 0)
		printf("tape %s completed successfully.\n", argv[0]);
	else
		printf("tape %s completed successfully with a residue of %d.\n", 
		       argv[0], rsp->residue);
	return(NFX_OK);
}

int 
tape_ui_write_fms(int argc, char **argv, int timeout, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_write_fm_req_msg_t *	req;
	tape_write_fm_rsp_msg_t *	rsp;
	char *				endptr;
	
	req = (tape_write_fm_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_WRITE_FM_REQ;
	req->hdr.cookie = 1;

	rc = tape_ui_devname_to_handle(argv[1], &req->handle);
	if (rc == NFX_ERR)
		return(NFX_ERR);
	else
	if (rc == TAPE_ECODE_NO_DEVICE) {
		printf("tape %s failed: device %s does not exist\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else
	if (rc == TAPE_ECODE_DEVICE_BUSY) {
		printf("tape %s failed: device %s is busy\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else 
	if (rc == TAPE_ECODE_DEVICE_NOT_OPEN) {
		printf("tape %s failed: device %s is not open\n", argv[0], argv[1]);
		return(NFX_ERR);
	}

	req->count = strtoul(argv[2], &endptr, 0);
	if (endptr == argv[2] || *endptr != 0 || endptr - argv[2] > 8) {
		printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[2]);
		return(NFX_ERR);
	}

	req_len = sizeof(tape_write_fm_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (tape_write_fm_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_WRITE_FM_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_WRITE_FM_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_write_fm_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, sizeof(tape_write_fm_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape %s failed: ecode %s(0x%x)\n", 
		       argv[0], tape_ecode_to_string(rsp->ecode), rsp->ecode);
		return(NFX_ERR);
	}

	if (rsp->residue == 0)
		printf("tape %s completed successfully.\n", argv[0]);
	else
		printf("tape %s completed successfully with a residue of %d.\n", 
		       argv[0], rsp->residue);
	return(NFX_OK);
}

int 
tape_ui_load(int argc, char **argv, int timeout, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_load_req_msg_t *		req;
	tape_load_rsp_msg_t *		rsp;
	va_list				ap;
	int32				arg1;

	va_start(ap, timeout);
	arg1 = va_arg(ap, int32);
	va_end(ap);

	req = (tape_load_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_LOAD_REQ;
	req->hdr.cookie = 1;

	rc = tape_ui_devname_to_handle(argv[1], &req->handle);
	if (rc == NFX_ERR)
		return(NFX_ERR);
	else
	if (rc == TAPE_ECODE_NO_DEVICE) {
		printf("tape %s failed: device %s does not exist\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else
	if (rc == TAPE_ECODE_DEVICE_BUSY) {
		printf("tape %s failed: device %s is busy\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else 
	if (rc == TAPE_ECODE_DEVICE_NOT_OPEN) {
		printf("tape %s failed: device %s is not open\n", argv[0], argv[1]);
		return(NFX_ERR);
	}

	req->load_or_unload = arg1;

	req_len = sizeof(tape_load_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (tape_load_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_LOAD_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_LOAD_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_load_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, sizeof(tape_load_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	if (rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape %s failed: ecode %s(0x%x)\n", 
		       argv[0], tape_ecode_to_string(rsp->ecode), rsp->ecode);
		return(NFX_ERR);
	}
	return(NFX_OK);
}

int 
tape_ui_read(int argc, char **argv, int timeout, ...)
{
        int				rc;
	int				bytecount;
	char 				req_msg[1024];
	char *				rsp_msg;
	int				req_len = 0;
	int				rsp_len = 0;
	int				binary = 0;
	tape_read_req_msg_t *		req;
	tape_read_rsp_msg_t *		rsp;
	char *				endptr;

	if (strcmp(argv[2], "-b") == 0) {
		binary = 1;
		bytecount = strtoul(argv[3], &endptr, 0);
                if (endptr == argv[3] || *endptr != 0 || endptr - argv[3] > 8) {
                        printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[3]);
                        return(NFX_ERR);
                }
	}
	else {
		bytecount = strtoul(argv[2], &endptr, 0); 
                if (endptr == argv[2] || *endptr != 0 || endptr - argv[2] > 8) {
                        printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[2]);
                        return(NFX_ERR);
                }
	}

	req_len = sizeof(tape_read_req_msg_t);
	rsp_len = sizeof(tape_read_rsp_msg_t) + bytecount;

	if (rsp_len > MAX_IO_LEN) {
		printf("tape %s failed: data length longer than %d.\n", argv[0], MAX_IO_LEN - sizeof(tape_read_rsp_msg_t));
		return(NFX_ERR);
	}

	rsp_msg = malloc(rsp_len);

	req = (tape_read_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_READ_REQ;
	req->hdr.cookie = 1;

	rc = tape_ui_devname_to_handle(argv[1], &req->handle);
	if (rc == NFX_ERR)
		return(NFX_ERR);
	else
	if (rc == TAPE_ECODE_NO_DEVICE) {
		printf("tape %s failed: device %s does not exist\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else
	if (rc == TAPE_ECODE_DEVICE_BUSY) {
		printf("tape %s failed: device %s is busy\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else 
	if (rc == TAPE_ECODE_DEVICE_NOT_OPEN) {
		printf("tape %s failed: device %s is not open\n", argv[0], argv[1]);
		return(NFX_ERR);
	}

	req->bytecount = bytecount;
	
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		goto read_error;
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		goto read_error;
	}

	rsp = (tape_read_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_READ_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_READ_RSP, rsp->hdr.mid);
		goto read_error;
	}
	if (rsp_len < sizeof(tape_read_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected at least %d got %d\n", 
		       req->hdr.mid, sizeof(tape_read_rsp_msg_t), rsp_len);
		goto read_error;
	}
	if (rsp->ecode == TAPE_ECODE_ILI) {
		printf("tape %s succeeded with ILI: residue %d\n",
		       argv[0], rsp->residue);
	}
	else
	if (rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape %s failed: ecode %s(0x%x)\n", 
		       argv[0], tape_ecode_to_string(rsp->ecode), rsp->ecode);
		goto read_error;
	}

	if (rsp->bytecount > req->bytecount) {
		printf("tape %s failed: Received (%d) more data than asked for (%d).\n", 
		       argv[0], rsp->bytecount, req->bytecount);
		goto read_error;
	}
	else
	if (rsp->bytecount < req->bytecount) {
		printf("tape %s succeeded: Residue %d\n",
		       argv[0], req->bytecount - rsp->bytecount);
	}

	if (rsp->bytecount) {
		int i;

		if (!binary) {
			for (i = 0; i < rsp->bytecount; ++i) {
				if ((i % 16) == 0)
					printf("\n[0x%04x]: ", i);
				printf("%02x ", rsp->data[i]);
			}
		}
		else {
			for (i = 0; i < rsp->bytecount; ++i) {
				putchar(rsp->data[i]);
			}
		}
		
		printf("\n");
	}
		
	free(rsp_msg);
	return(NFX_OK);

 read_error: 
	free(rsp_msg);
	return(NFX_ERR);
}

int 
tape_ui_write(int argc, char **argv, int timeout, ...)
{
        int				rc;
	int				bytecount;
	byte				pattern;
	char *				req_msg;
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_write_req_msg_t *		req;
	tape_write_rsp_msg_t *		rsp;
	char *				endptr;

	bytecount = strtoul(argv[2], &endptr, 0);
	if (endptr == argv[2] || *endptr != 0 || endptr - argv[2] > 8) {
		printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[2]);
		return(NFX_ERR);
	}

	req_len = sizeof(tape_write_req_msg_t) - 1 + bytecount;
	rsp_len = sizeof(rsp_msg);

	if (req_len > MAX_IO_LEN) {
		printf("tape %s failed: data length longer than %d.\n", argv[0], MAX_IO_LEN - sizeof(tape_write_req_msg_t));
		return(NFX_ERR);
	}

	req_msg = malloc(req_len);
	req = (tape_write_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_WRITE_REQ;
	req->hdr.cookie = 1;

	rc = tape_ui_devname_to_handle(argv[1], &req->handle);
	if (rc == NFX_ERR)
		return(NFX_ERR);
	else
	if (rc == TAPE_ECODE_NO_DEVICE) {
		printf("tape %s failed: device %s does not exist\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else
	if (rc == TAPE_ECODE_DEVICE_BUSY) {
		printf("tape %s failed: device %s is busy\n", argv[0], argv[1]);
		return(NFX_ERR);
	}
	else 
	if (rc == TAPE_ECODE_DEVICE_NOT_OPEN) {
		printf("tape %s failed: device %s is not open\n", argv[0], argv[1]);
		return(NFX_ERR);
	}

	req->bytecount = bytecount;

	pattern = strtoul(argv[3], &endptr, 0);
	if (endptr == argv[3] || *endptr != 0 || endptr - argv[3] > 8) {
		printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[3]);
		return(NFX_ERR);
	}

	memset(req->data, pattern, bytecount);

	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		goto write_error;
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		goto write_error;
	}

	rsp = (tape_write_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_WRITE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_WRITE_RSP, rsp->hdr.mid);
		goto write_error;
	}
	if (rsp_len != sizeof(tape_write_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, sizeof(tape_write_rsp_msg_t), rsp_len);
		goto write_error;
	}
	if (rsp->ecode != TAPE_ECODE_SUCCESS) {
		printf("tape %s failed: ecode %s(0x%x)\n", 
		       argv[0], tape_ecode_to_string(rsp->ecode), rsp->ecode);
		goto write_error;
	}

	free(req_msg);
	return(NFX_OK);

 write_error: 
	free(req_msg);
	return(NFX_ERR);
}

int 
tape_ui_dbg_level(int argc, char **argv, int timeout, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	tape_set_debug_level_req_msg_t *req;
	tape_set_debug_level_rsp_msg_t *rsp;
	char *			        endptr;

	req = (tape_set_debug_level_req_msg_t *)req_msg;
	req->hdr.mid = TAPE_MID_SET_DEBUG_LEVEL_REQ;
	req->hdr.cookie = 1;
	req->debug_level = strtoul(argv[1], &endptr, 0);
	if (endptr == argv[1] || *endptr != 0 || endptr - argv[1] > 8) {
		printf("tape %s failed: Invalid argument '%s'\n", argv[0], argv[1]);
		return(NFX_ERR);
	}

			
	req_len = sizeof(tape_set_debug_level_req_msg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_tape_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_tape_message(mid 0x%x) failed.\n", req->hdr.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_tape_message(mid 0x%x) timed out.\n", req->hdr.mid);
		return(NFX_ERR);
	}

	rsp = (tape_set_debug_level_rsp_msg_t *)rsp_msg;
	if (rsp->hdr.mid != TAPE_MID_SET_DEBUG_LEVEL_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       TAPE_MID_SET_DEBUG_LEVEL_RSP, rsp->hdr.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(tape_set_debug_level_rsp_msg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->hdr.mid, sizeof(tape_set_debug_level_rsp_msg_t), rsp_len);
		return(NFX_ERR);
	}
	return(NFX_OK);
}

int 
main(int argc, char **argv)
{
	int	i;
	int	rc;

	if (argc < 2)
		goto arg_error;

	for (i = 0; tape_ui_root[i].name; ++i) {
		if (strcmp(tape_ui_root[i].name, argv[1]) == 0) {
			if (argc-2 < tape_ui_root[i].min_parm_cnt ||
			    argc-2 > tape_ui_root[i].max_parm_cnt) 
				goto arg_error;
			else {
				rc = tape_ui_root[i].func(argc-1, argv+1, 
							  tape_ui_root[i].timeout,
							  tape_ui_root[i].implicit_arg1, 
							  tape_ui_root[i].implicit_arg2);
				return(rc == NFX_OK ? 0 : 1);
			}
		}
	}

arg_error:
	print_usage(argv[0]);
	exit(1);
}

int
send_tape_message(char *req_buf, int req_len, char *resp_buf, int *resp_len, int timeout)
{
	int		stat;
	struct timeval	to;

	to.tv_sec = timeout;
	to.tv_usec = 0;

	stat = sendAgileMsg(-1,
			    req_buf, req_len,
			    "tape-driver", 0, 0, 0, &to,
			    resp_buf, *resp_len);
	if (stat > 0)
		*resp_len = stat;
	return(stat);
}

char *
tape_ecode_to_string(uint32 ecode)
{
	switch(ecode) {
	case TAPE_ECODE_SUCCESS:			return("SUCCESS");
	case TAPE_ECODE_DEVICE_BUSY:			return("DEVICE_BUSY");
	case TAPE_ECODE_NO_DEVICE:			return("NO_DEVICE");
	case TAPE_ECODE_NO_TAPE_LOADED:			return("NO_TAPE_LOADED");
	case TAPE_ECODE_ILI:				return("ILI");
	case TAPE_ECODE_DATA_SIZE_MISMATCH:		return("DATA_SIZE_MISMATCH");
	case TAPE_ECODE_MEDIUM_ERR:			return("MEDIUM_ERR");
	case TAPE_ECODE_EOM_ENCOUNTERED:		return("EOM_ENCOUNTERED");
	case TAPE_ECODE_EOD_ENCOUNTERED:		return("EOD_ENCOUNTERED");
	case TAPE_ECODE_FM_ENCOUNTERED:			return("FM_ENCOUNTERED");
	case TAPE_ECODE_IO_ERROR:			return("IO_ERROR");
	case TAPE_ECODE_DEVICE_NOT_OPEN:		return("DEVICE_NOT_OPENED");
	case TAPE_ECODE_DEVICE_NOT_OPEN_BY_CLIENT:	return("DEVICE_NOT_OPENED_BY_CLIENT");
	default:				return("???");
	}
}

