/*
 *-----------------------------------------------------------------
 *
 * Name:	membench.c
 *
 * RCS:  $Id: membench.c,v 1.11 2002/11/01 17:39:05 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Memory benchmarking tests
 *
 * Created by:	Bob Fozard
 *
 * Date Created:	8/14/2002
 *
 *-----------------------------------------------------------------
 */

#ifndef SSC_MGMT

#include <nfx-incl.h>
#include "exports/malloc-api.h"
#include "exports/timer-api.h"
#include "exports/eee-desc.h"
#include "sys/mips.h"


//
// TODO: - disable the watchdog?
//       - disable timer interrupts?
//       - run them from the prom?
//       - provide support for lcl/rmt shm?
//

extern void old_bzero(void *, __SIZE_TYPE__);

/**
 * Memory region types that can
 * be tested.
 */
enum
{
    MB_MEM_TYPE_HEAP,
    MB_MEM_TYPE_LOCAL_SHM,
    MB_MEM_TYPE_REMOTE_SHM,
};


/**
 * Options to the tests.
 */
enum
{
    MB_OPT_DO_UNROLLED  = 0x0001,
    MB_OPT_INIT_CACHE   = 0x0002,
    MB_OPT_CACHE_LINE   = 0x0004,
};


/**
 * Do this for the fp to provide an analyzer
 * trigger.
 */
#ifdef PROM_SIBYTE
#define TRIGGER_START    *(uint32 *)0xffffffffbf400088ll = 1;
#define TRIGGER_END      *(uint32 *)0xffffffffbf400088ll = 1;
#else
#define TRIGGER_START
#define TRIGGER_END
#endif


/**
 * Malloc a buffer of the specified type,
 * truncating numBytes if appropriate.
 */
static void *
mb_malloc(
    int32   type,
    int32   *numBytes)
{
    void    *retPtr = NULL;

    switch(type)
    {
    case MB_MEM_TYPE_HEAP:
        retPtr = eee_ramAlloc(*numBytes);
        break;
    case MB_MEM_TYPE_LOCAL_SHM:
        retPtr = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_ULTRA,
                                    NFX_COS_HIGH);
        if (*numBytes > EEE_BUFSIZE_ULTRA)
        {
            *numBytes = EEE_BUFSIZE_ULTRA;
            printf("truncating numBytes to %d (max shm buffer size)\n",
                EEE_BUFSIZE_ULTRA);
        }
        break;
    case MB_MEM_TYPE_REMOTE_SHM:
#ifdef NFP_FP
        retPtr = eee_allocateBuffer(EEE_TXRX_POOL, EEE_BUF_ULTRA,
                                    NFX_COS_HIGH);
#elif NFP_TXRX
        retPtr = eee_allocateBuffer(EEE_FP_POOL, EEE_BUF_ULTRA,
                                    NFX_COS_HIGH);
#else
        retPtr = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_ULTRA,
                                    NFX_COS_HIGH);
#endif
        if (*numBytes > EEE_BUFSIZE_ULTRA)
        {
            *numBytes = EEE_BUFSIZE_ULTRA;
            printf("truncating numBytes to %d (max shm buffer size)\n",
                EEE_BUFSIZE_ULTRA);
        }
        break;
    default:
        printf("mb_malloc bad memtype %d\n", type);
    }

    return retPtr;
}


/**
 * Free a buffer of the specified type.
 */
static void
mb_free(
    int32  type,
    void   *ptr)
{
    switch(type)
    {
    case MB_MEM_TYPE_HEAP:
        eee_ramDealloc(ptr);
        break;
    case MB_MEM_TYPE_LOCAL_SHM:
    case MB_MEM_TYPE_REMOTE_SHM:
        eee_deallocateBuffer(ptr);
        break;
    default:
        printf("mb_free bad memtype %d\n", type);
    }
}


/**
 * Present the results of a test.
 */
static void
mb_printResult(
    char    *callerStr,
    uint64  numBytes,
    uint64  numIter,
    uint64  usecs)
{
    printf("%s: %lld bytes transferred in %lld usecs\n",
           callerStr, numBytes * numIter, usecs);

	if (usecs > 0)
    printf("    %lld MegaBytes/sec throughput ",
            numBytes * numIter / usecs);

    if (usecs < 1000000)
    {
        printf("(elapsed time < 1 sec, result unreliable)");
    }
    printf("\n");
}


/**
 * Do bzero test.
 */
int32
mb_bzero(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options)
{
    char        *datap;
    eeeRTC_t    startTime, delta;
    char        *testName = "mb_bzero";
    int32       iter;

    if (numBytes == 0 ||
        numIter == 0)
    {
        printf("%s: bad params (numBytes: %d, numIter: %d)\n",
            testName, numBytes, numIter);
        return NFX_ERR;
    }

    datap = mb_malloc(memType, &numBytes);
    if (!datap)
    {
        printf("%s: malloc failed\n", testName);
        return NFX_ERR;
    }

    if (options & MB_OPT_INIT_CACHE)
    {
        bzero(datap, numBytes);
    }

    eee_rtc(&startTime);
    TRIGGER_START
    for (iter = 0; iter < numIter; iter++)
    {
        bzero(datap, numBytes);
    }
    TRIGGER_END
    eee_rtcDuration(&startTime, &delta);

    mb_printResult(testName, numBytes, numIter, eee_rtcToUsec(&delta));

    mb_free(memType, datap);

    return NFX_OK;
}


/**
 * Perform unrolled U64 write.
 */
static void
mb_write64unrolled(
    int32           numBytes,
    volatile uint64 *to)
{

#define WRITE_128_BYTES \
        to[0] = 1;      \
        to[1] = 1;      \
        to[2] = 1;      \
        to[3] = 1;      \
        to[4] = 1;      \
        to[5] = 1;      \
        to[6] = 1;      \
        to[7] = 1;      \
        to[8] = 1;      \
        to[9] = 1;      \
        to[10] = 1;     \
        to[11] = 1;     \
        to[12] = 1;     \
        to[13] = 1;     \
        to[14] = 1;     \
        to[15] = 1;     \
        to += 16;       \
        numBytes -= 128;

#define WRITE_UNROLLED_BYTES   512

    while (numBytes > WRITE_UNROLLED_BYTES)
    {
        WRITE_128_BYTES
        WRITE_128_BYTES
        WRITE_128_BYTES
        WRITE_128_BYTES
    }
#undef WRITE_128_BYTES
}


/**
 * Perform unrolled U32 write.
 */
static void
mb_write32unrolled(
    int32           numBytes,
    volatile uint32 *to)
{
#define WRITE_64_BYTES  \
        to[0] = 1;      \
        to[1] = 1;      \
        to[2] = 1;      \
        to[3] = 1;      \
        to[4] = 1;      \
        to[5] = 1;      \
        to[6] = 1;      \
        to[7] = 1;      \
        to[8] = 1;      \
        to[9] = 1;      \
        to[10] = 1;     \
        to[11] = 1;     \
        to[12] = 1;     \
        to[13] = 1;     \
        to[14] = 1;     \
        to[15] = 1;     \
        to += 16;       \
        numBytes -= 64;

    while (numBytes > WRITE_UNROLLED_BYTES)
    {
        WRITE_64_BYTES
        WRITE_64_BYTES
        WRITE_64_BYTES
        WRITE_64_BYTES
        WRITE_64_BYTES
        WRITE_64_BYTES
        WRITE_64_BYTES
        WRITE_64_BYTES
    }
#undef WRITE_64_BYTES
}


/**
 * Do write test.
 */
int32
mb_write(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options,
    int32   dataSize,
    char    *testName)
{
    char        *datap;
    eeeRTC_t    startTime, delta;
    int32       i;
    int32       iter;

    if (numBytes == 0 ||
        numIter == 0 ||
        numBytes % dataSize)
    {
        printf("%s: bad params (numBytes: %d, numIter: %d)\n",
            testName, numBytes, numIter);
        return NFX_ERR;
    }

    datap = mb_malloc(memType, &numBytes);
    if (!datap)
    {
        printf("%s: malloc failed\n", testName);
        return NFX_ERR;
    }

    if (options & MB_OPT_INIT_CACHE)
    {
        bzero(datap, numBytes);
    }

    if (options & MB_OPT_DO_UNROLLED)
    {
        if (numBytes % WRITE_UNROLLED_BYTES)
        {
            numBytes &= ~(WRITE_UNROLLED_BYTES - 1);
            printf("%s: unrolled test truncating to %d bytes\n",
                testName, numBytes);
        }

        if (dataSize == sizeof(uint32))
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_write32unrolled(numBytes, (volatile uint32 *) datap);
            }
        }
        else
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_write64unrolled(numBytes, (volatile uint64 *) datap);
            }
        }
    }
    else
    {
        if (dataSize == sizeof(uint32))
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                for (i = 0; i < numBytes; i += sizeof(uint32))
                {
                    *((volatile uint32 *)(datap + i)) = i;
                }
            }
        }
        else
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                for (i = 0; i < numBytes; i += sizeof(uint64))
                {
                    *((volatile uint64 *)(datap + i)) = i;
                }
            }
        }
    }
    TRIGGER_END
    eee_rtcDuration(&startTime, &delta);

    mb_printResult(testName, numBytes, numIter, eee_rtcToUsec(&delta));

    mb_free(memType, datap);

    return NFX_OK;
}


/**
 * Do U64 write test.
 */
int32
mb_write64(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options)
{
    return mb_write(numBytes, numIter, memType,
                    options, sizeof(uint64), "mb_write64");
}


/**
 * Do U32 write test.
 */
int32
mb_write32(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options)
{
    return mb_write(numBytes, numIter, memType,
                    options, sizeof(uint32), "mb_write32");
}


/**
 * Perform unrolled read by cache-line.
 */
static void
mb_readCacheLineUnrolled(
    int32           numBytes,
    volatile uint64 *from)
{

// hard-coded for 32-byte cache line size
#define READ_512_BYTES                  \
        __asm__ __volatile__ (          \
        "     .set noreorder  \n"       \
        "     ld $8,    0($5) \n"       \
        "     ld $9,   32($5) \n"       \
        "     ld $10,  64($5) \n"       \
        "     ld $11,  96($5) \n"       \
        "     ld $12, 128($5) \n"       \
        "     ld $13, 160($5) \n"       \
        "     ld $14, 192($5) \n"       \
        "     ld $15, 224($5) \n"       \
        "     ld $8,  256($5) \n"       \
        "     ld $9,  288($5) \n"       \
        "     ld $10, 320($5) \n"       \
        "     ld $11, 352($5) \n"       \
        "     ld $12, 384($5) \n"       \
        "     ld $13, 416($5) \n"       \
        "     ld $14, 448($5) \n"       \
        "     ld $15, 480($5) \n"       \
        "     daddiu  $5,$5,512 \n"     \
        "     addiu   %0,%0,-512 \n"    \
        "     .set reorder \n"          \
        : "=r" (numBytes)               \
        );

#define READ_CACHE_LINE_UNROLLED_BYTES     2048

    while (numBytes > READ_CACHE_LINE_UNROLLED_BYTES)
    {
        READ_512_BYTES
        READ_512_BYTES
        READ_512_BYTES
        READ_512_BYTES
    }
#undef READ_512_BYTES
}


/**
 * Perform unrolled U64 read.
 */
static void
mb_read64unrolled(
    int32           numBytes,
    volatile uint64 *from)
{
#define READ_128_BYTES      \
        (void) *(from+0);   \
        (void) *(from+1);   \
        (void) *(from+2);   \
        (void) *(from+3);   \
        (void) *(from+4);   \
        (void) *(from+5);   \
        (void) *(from+6);   \
        (void) *(from+7);   \
        (void) *(from+8);   \
        (void) *(from+9);   \
        (void) *(from+10);  \
        (void) *(from+11);  \
        (void) *(from+12);  \
        (void) *(from+13);  \
        (void) *(from+14);  \
        (void) *(from+15);  \
        from += 16;         \
        numBytes -= 128;

#define READ_UNROLLED_BYTES     512

    while (numBytes > READ_UNROLLED_BYTES)
    {
        READ_128_BYTES
        READ_128_BYTES
        READ_128_BYTES
        READ_128_BYTES
    }
#undef READ_128_BYTES
}


/**
 * Perform unrolled U32 read.
 */
static void
mb_read32unrolled(
    int32           numBytes,
    volatile uint32 *from)
{
#define READ_64_BYTES       \
        (void) *(from+0);   \
        (void) *(from+1);   \
        (void) *(from+2);   \
        (void) *(from+3);   \
        (void) *(from+4);   \
        (void) *(from+5);   \
        (void) *(from+6);   \
        (void) *(from+7);   \
        (void) *(from+8);   \
        (void) *(from+9);   \
        (void) *(from+10);  \
        (void) *(from+11);  \
        (void) *(from+12);  \
        (void) *(from+13);  \
        (void) *(from+14);  \
        (void) *(from+15);  \
        from += 16;         \
        numBytes -= 64;

    while (numBytes > READ_UNROLLED_BYTES)
    {
        READ_64_BYTES
        READ_64_BYTES
        READ_64_BYTES
        READ_64_BYTES
        READ_64_BYTES
        READ_64_BYTES
        READ_64_BYTES
        READ_64_BYTES
    }
#undef READ_64_BYTES
}


/**
 * Do U32/U64 read test.
 */
int32
mb_read(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options,
    int32   dataSize,
    char    *testName)
{
    char        *datap;
    eeeRTC_t    startTime, delta;
    int32       i;
    uint32      summ;
    int32       iter;

    if (numBytes == 0 ||
        numIter == 0 ||
        numBytes % dataSize)
    {
        printf("%s: bad params (numBytes: %d, numIter: %d)\n",
            testName, numBytes, numIter);
        return NFX_ERR;
    }

    datap = mb_malloc(memType, &numBytes);
    if (!datap)
    {
        printf("%s: malloc failed\n", testName);
        return NFX_ERR;
    }

    if (options & MB_OPT_INIT_CACHE)
    {
        bzero(datap, numBytes);
    }

    if (options & MB_OPT_DO_UNROLLED)
    {
        if (numBytes % READ_UNROLLED_BYTES)
        {
            numBytes &= ~(READ_UNROLLED_BYTES - 1);
            printf("%s: unrolled test truncating to %d bytes\n",
                testName, numBytes);
        }

        if (options & MB_OPT_CACHE_LINE)
        {
            if (numBytes % READ_CACHE_LINE_UNROLLED_BYTES)
            {
                numBytes &= ~(READ_CACHE_LINE_UNROLLED_BYTES - 1);
                printf("%s: unrolled test truncating to %d bytes\n",
                    testName, numBytes);
            }

            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_readCacheLineUnrolled(numBytes, (volatile uint64 *) datap);
            }
        }
        else if (dataSize == sizeof(uint32))
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_read32unrolled(numBytes, (volatile uint32 *) datap);
            }
        }
        else
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_read64unrolled(numBytes, (volatile uint64 *) datap);
            }
        }
    }
    else
    {
        if (dataSize == sizeof(uint32))
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                for (i = 0; i < numBytes; i += sizeof(uint32))
                {
                    summ += *((volatile uint32 *)(datap + i));
                }
            }
        }
        else
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                for (i = 0; i < numBytes; i += sizeof(uint64))
                {
                    summ += *((volatile uint64 *)(datap + i));
                }
            }
        }
    }
    TRIGGER_END
    eee_rtcDuration(&startTime, &delta);

    mb_printResult(testName, numBytes, numIter, eee_rtcToUsec(&delta));

    mb_free(memType, datap);

    return NFX_OK;
}


/**
 * Do U32 read test.
 */
int32
mb_read32(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options)
{
    return mb_read(numBytes, numIter, memType,
                   options, sizeof(uint32), "mb_read32");
}


/**
 * Do U64 read test.
 */
int32
mb_read64(
    int32   numBytes,
    int32   numIter,
    int32   memType,
    uint32  options)
{
    return mb_read(numBytes, numIter, memType,
                   options, sizeof(uint64), "mb_read64");
}



/**
 * Perform unrolled U64 copy.
 */
static void
mb_copy64unrolled(
    int32           numBytes,
    volatile uint64 *from,
    volatile uint64 *to)
{
#define COPY_128_BYTES      \
        to[0] = from[0];    \
        to[1] = from[1];    \
        to[2] = from[2];    \
        to[3] = from[3];    \
        to[4] = from[4];    \
        to[5] = from[5];    \
        to[6] = from[6];    \
        to[7] = from[7];    \
        to[8] = from[8];    \
        to[9] = from[9];    \
        to[10] = from[10];  \
        to[11] = from[11];  \
        to[12] = from[12];  \
        to[13] = from[13];  \
        to[14] = from[14];  \
        to[15] = from[15];  \
        to += 16;           \
        from += 16;         \
        numBytes -= 128;

#define COPY_UNROLLED_BYTES     512

    while (numBytes > COPY_UNROLLED_BYTES)
    {
        COPY_128_BYTES
        COPY_128_BYTES
        COPY_128_BYTES
        COPY_128_BYTES
    }
#undef COPY_128_BYTES
}


/**
 * Perform unrolled U32 copy.
 */
static void
mb_copy32unrolled(
    int32           numBytes,
    volatile uint32 *from,
    volatile uint32 *to)
{
#define COPY_64_BYTES       \
        to[0] = from[0];    \
        to[1] = from[1];    \
        to[2] = from[2];    \
        to[3] = from[3];    \
        to[4] = from[4];    \
        to[5] = from[5];    \
        to[6] = from[6];    \
        to[7] = from[7];    \
        to[8] = from[8];    \
        to[9] = from[9];    \
        to[10] = from[10];  \
        to[11] = from[11];  \
        to[12] = from[12];  \
        to[13] = from[13];  \
        to[14] = from[14];  \
        to[15] = from[15];  \
        to += 16;           \
        from += 16;         \
        numBytes -= 64;

    while (numBytes > COPY_UNROLLED_BYTES)
    {
        COPY_64_BYTES
        COPY_64_BYTES
        COPY_64_BYTES
        COPY_64_BYTES
        COPY_64_BYTES
        COPY_64_BYTES
        COPY_64_BYTES
        COPY_64_BYTES
    }
#undef COPY_64_BYTES
}


/**
 * Do U32/U64 copy test.
 */
int32
mb_copy(
    int32   numBytes,
    int32   numIter,
    int32   fromMemType,
    int32   toMemType,
    uint32  options,
    int32   dataSize,
    char    *testName)
{
    char        *from, *to;
    eeeRTC_t    startTime, delta;
    int32       iter;

    if (numBytes == 0 ||
        numIter == 0 ||
        numBytes % dataSize)
    {
        printf("%s: bad params (numBytes: %d, numIter: %d)\n",
            testName, numBytes, numIter);
        return NFX_ERR;
    }

    from = mb_malloc(fromMemType, &numBytes);
    if (!from)
    {
        printf("%s: malloc failed\n", testName);
        return NFX_ERR;
    }
    to = mb_malloc(toMemType, &numBytes);
    if (!to)
    {
        printf("%s: malloc failed\n", testName);
        mb_free(fromMemType, from);
        return NFX_ERR;
    }

    if (options & MB_OPT_INIT_CACHE)
    {
        bzero(from, numBytes);
        bzero(to, numBytes);
    }

    if (options & MB_OPT_DO_UNROLLED)
    {
        if (numBytes % COPY_UNROLLED_BYTES)
        {
            numBytes &= ~(COPY_UNROLLED_BYTES - 1);
            printf("%s: unrolled test truncating to %d bytes\n",
                testName, numBytes);
        }

        if (dataSize == sizeof(uint32))
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_copy32unrolled(numBytes, (volatile uint32 *) from,
                                            (volatile uint32 *) to);
            }
        }
        else
        {
            eee_rtc(&startTime);
            TRIGGER_START
            for (iter = 0; iter < numIter; iter++)
            {
                mb_copy64unrolled(numBytes, (volatile uint64 *) from,
                                            (volatile uint64 *) to);
            }
        }
    }
    else
    {
        printf("FYI, copy test optimized for platform (32/64 bit)\n");
        eee_rtc(&startTime);
        TRIGGER_START
        for (iter = 0; iter < numIter; iter++)
        {
            memcpy(to, from, numBytes);
        }
    }
    TRIGGER_END
    eee_rtcDuration(&startTime, &delta);

    mb_printResult(testName, numBytes, numIter, eee_rtcToUsec(&delta));

    mb_free(fromMemType, from);
    mb_free(toMemType, to);

    return NFX_OK;
}


/**
 * Do U32 copy test.
 */
int32
mb_copy32(
    int32   numBytes,
    int32   numIter,
    int32   fromMemType,
    int32   toMemType,
    uint32  options)
{
    return mb_copy(numBytes, numIter, fromMemType, toMemType,
                    options, sizeof(uint32), "mb_copy32");
}   

/**
 * Do U64 copy test.
 */
int32
mb_copy64(
    int32   numBytes,
    int32   numIter,
    int32   fromMemType,
    int32   toMemType,
    uint32  options)
{
    return mb_copy(numBytes, numIter, fromMemType, toMemType,
                   options, sizeof(uint64), "mb_copy64");
}   

typedef struct mb_CommonArgs
{
    int32   numBytes;
    int32   numIter;
    int32   memType;
    int32   options;
} mb_CommonArgs;


/**
 * Common parser for most tests.
 */
static int32
mb_parseCommon(int argc, char **argv, mb_CommonArgs *commonArgs)
{
    int32   i;

    if (argc < 4)
    {
        goto usage;
    }

    commonArgs->numBytes = atoi(argv[1]);
    commonArgs->numIter = atoi(argv[2]);
    if (!strcmp(argv[3], "heap"))
        commonArgs->memType = MB_MEM_TYPE_HEAP;
    else if (!strcmp(argv[3], "lcl"))
        commonArgs->memType = MB_MEM_TYPE_LOCAL_SHM;
    else if (!strcmp(argv[3], "rmt"))
        commonArgs->memType = MB_MEM_TYPE_REMOTE_SHM;
    else
        goto usage;

    commonArgs->options = 0;
    if (argc >= 5)
    {
        for (i = 0; i < argc - 4; i++)
        {
            if (!strcmp(argv[4 + i], "-u"))
                commonArgs->options |= MB_OPT_DO_UNROLLED;
            else if (!strcmp(argv[4 + i], "-c"))
                commonArgs->options |= MB_OPT_INIT_CACHE;
            else if (!strcmp(argv[4 + i], "-l"))
                commonArgs->options |= MB_OPT_CACHE_LINE;
            else {
                printf("unknown argument %s\n", argv[4 + i]);
                goto usage;
            }
        }
    }

    return NFX_OK;

usage:
    printf("Usage: %s <numBytes> <numIter> <memType> [-u] [-c] [-l]\n",
        argv[0]);
    printf("   use <memType>  heap   for malloc space\n");
    printf("   use <memType>  lcl    for local shm buffers\n");
    printf("   use <memType>  rmt    for rmt shm buffers\n");
    printf("   use -u to perform unrolled loops\n");
    printf("   use -c to init cache\n");
    printf("   use -l to touch by cache-line (for read tests only)\n");

    return NFX_ERR;
}


int32
mb_memBenchWrite32(int argc, char **argv)
{
    mb_CommonArgs   commonArgs;

    if (mb_parseCommon(argc, argv, &commonArgs) == NFX_OK)
    {
        return mb_write32(commonArgs.numBytes, commonArgs.numIter,
                          commonArgs.memType, commonArgs.options);
    }

    return -1;
}

int32
mb_memBenchWrite64(int argc, char **argv)
{
    mb_CommonArgs   commonArgs;

    if (mb_parseCommon(argc, argv, &commonArgs) == NFX_OK)
    {
        return mb_write64(commonArgs.numBytes, commonArgs.numIter,
                          commonArgs.memType, commonArgs.options);
    }

    return -1;
}

int32
mb_memBenchRead32(int argc, char **argv)
{
    mb_CommonArgs   commonArgs;

    if (mb_parseCommon(argc, argv, &commonArgs) == NFX_OK)
    {
        return mb_read32(commonArgs.numBytes, commonArgs.numIter,
                          commonArgs.memType, commonArgs.options);
    }

    return -1;
}

int32
mb_memBenchRead64(int argc, char **argv)
{
    mb_CommonArgs   commonArgs;

    if (mb_parseCommon(argc, argv, &commonArgs) == NFX_OK)
    {
        return mb_read64(commonArgs.numBytes, commonArgs.numIter,
                          commonArgs.memType, commonArgs.options);
    }

    return -1;
}

int32
mb_memBenchBzero(int argc, char **argv)
{
    mb_CommonArgs   commonArgs;

    if (mb_parseCommon(argc, argv, &commonArgs) == NFX_OK)
    {
        return mb_bzero(commonArgs.numBytes, commonArgs.numIter,
                        commonArgs.memType, commonArgs.options);
    }

    return -1;
}


/**
 * Structure for args to copy test.
 */
typedef struct mb_CopyArgs
{
    int32   numBytes;
    int32   numIter;
    int32   fromMemType;
    int32   toMemType;
    int32   options;
} mb_CopyArgs;


/**
 * Parser for copy tests.
 */
static int32
mb_parseCopy(int argc, char **argv, mb_CopyArgs *copyArgs)
{
    int32   i;

    if (argc < 5)
    {
        goto usage;
    }

    copyArgs->numBytes = atoi(argv[1]);
    copyArgs->numIter = atoi(argv[2]);
    if (!strcmp(argv[3], "heap"))
        copyArgs->fromMemType = MB_MEM_TYPE_HEAP;
    else if (!strcmp(argv[3], "lcl"))
        copyArgs->fromMemType = MB_MEM_TYPE_LOCAL_SHM;
    else if (!strcmp(argv[3], "rmt"))
        copyArgs->fromMemType = MB_MEM_TYPE_REMOTE_SHM;
    else
        goto usage;
    if (!strcmp(argv[4], "heap"))
        copyArgs->toMemType = MB_MEM_TYPE_HEAP;
    else if (!strcmp(argv[4], "lcl"))
        copyArgs->toMemType = MB_MEM_TYPE_LOCAL_SHM;
    else if (!strcmp(argv[4], "rmt"))
        copyArgs->toMemType = MB_MEM_TYPE_REMOTE_SHM;
    else
        goto usage;

    copyArgs->options = 0;
    if (argc >= 6)
    {
        for (i = 0; i < argc - 5; i++)
        {
            if (!strcmp(argv[5 + i], "-u"))
                copyArgs->options |= MB_OPT_DO_UNROLLED;
            else if (!strcmp(argv[5 + i], "-c"))
                copyArgs->options |= MB_OPT_INIT_CACHE;
            else {
                printf("unknown argument %s\n", argv[4 + i]);
                goto usage;
            }
        }
    }

    return NFX_OK;

usage:
    printf(
     "Usage: %s <numBytes> <numIter> <fromMemType> <toMemType> [-u] [-c]\n",
        argv[0]);
    printf("   use <memType>  heap   for malloc space\n");
    printf("   use <memType>  lcl    for local shm buffers\n");
    printf("   use <memType>  rmt    for rmt shm buffers\n");
    printf("   use -u to perform unrolled loops\n");
    printf("   use -c to init cache\n");

    return NFX_ERR;
}

int32
mb_memBenchCopy32(int argc, char **argv)
{
    mb_CopyArgs   copyArgs;

    if (mb_parseCopy(argc, argv, &copyArgs) == NFX_OK)
    {
        return mb_copy32(copyArgs.numBytes, copyArgs.numIter,
                         copyArgs.fromMemType, copyArgs.toMemType,
                         copyArgs.options);
    }

    return -1;
}

int32
mb_memBenchCopy64(int argc, char **argv)
{
    mb_CopyArgs   copyArgs;

    if (mb_parseCopy(argc, argv, &copyArgs) == NFX_OK)
    {
        return mb_copy64(copyArgs.numBytes, copyArgs.numIter,
                         copyArgs.fromMemType, copyArgs.toMemType,
                         copyArgs.options);
    }

    return -1;
}


int32
memcpy_cmd(
	int32	argc,
	char8	*argv[]
)
{
    eeeRTC_t    startTime, delta;
	uchar8		* src, * dst, * save_dst;
	uint32		numIter, numBytes, i;

	numBytes = atoi(argv[1]);
	numIter = atoi(argv[2]);

	src = (uchar8 *)0x1000000000;
	save_dst = dst = eee_ramAlloc(numBytes + 32);

#if 0
	oldmemcpy(dst, src, numBytes);

printf("-----------------------------------------------------------\n");
printf("memcpy_cmd, numIter = %d, numBytes = %d, src = %p, dst = %p\n", numIter, numBytes, src, dst);

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		oldmemcpy(dst, src, numBytes);
	eee_rtcDuration(&startTime, &delta);

    mb_printResult("oldmemcpy_double", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		fastmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("fast_memcpy_double", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		memcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("memcpy_double", numBytes, numIter, eee_rtcToUsec(&delta));

	if (numBytes == 4) {
	    eee_rtc(&startTime);
		for (i = 0; i < numIter; i++)
			memcpy4(dst, src, numBytes);
		eee_rtcDuration(&startTime, &delta);

		mb_printResult("memcpy4", numBytes, numIter, eee_rtcToUsec(&delta));
	}
	if (numBytes == 6) {
	    eee_rtc(&startTime);
		for (i = 0; i < numIter; i++)
			memcpy6(dst, src, numBytes);
		eee_rtcDuration(&startTime, &delta);

		mb_printResult("memcpy6", numBytes, numIter, eee_rtcToUsec(&delta));
	}
	if (numBytes == 40) {
	    eee_rtc(&startTime);
		for (i = 0; i < numIter; i++)
			memcpy40(dst, src, numBytes);
		eee_rtcDuration(&startTime, &delta);

		mb_printResult("memcpy40", numBytes, numIter, eee_rtcToUsec(&delta));
	}

#ifdef SIBYTE
    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		local_memcpy_ass(dst, src, numBytes/32);
	eee_rtcDuration(&startTime, &delta);

	mb_printResult("local_memcpy_ass", numBytes, numIter, eee_rtcToUsec(&delta));
#endif

	dst += 4;
printf("-----------------------------------------------------------\n");
printf("memcpy_cmd, numIter = %d, numBytes = %d, src = %p, dst = %p\n", numIter, numBytes, src, dst);

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		oldmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("oldmemcpy_word", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		fastmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("fast_memcpy_word", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		memcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("memcpy_word", numBytes, numIter, eee_rtcToUsec(&delta));

	if (numBytes == 6) {
	    eee_rtc(&startTime);
		for (i = 0; i < numIter; i++)
			memcpy6(dst, src, numBytes);
		eee_rtcDuration(&startTime, &delta);

		mb_printResult("memcpy6", numBytes, numIter, eee_rtcToUsec(&delta));
	}
	dst += 2;
printf("-----------------------------------------------------------\n");
printf("memcpy_cmd, numIter = %d, numBytes = %d, src = %p, dst = %p\n", numIter, numBytes, src, dst);

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		oldmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("oldmemcpy_short", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		fastmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("fast_memcpy_short", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		memcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("memcpy_short", numBytes, numIter, eee_rtcToUsec(&delta));

	if (numBytes == 6) {
	    eee_rtc(&startTime);
		for (i = 0; i < numIter; i++)
			memcpy6(dst, src, numBytes);
		eee_rtcDuration(&startTime, &delta);

		mb_printResult("memcpy6", numBytes, numIter, eee_rtcToUsec(&delta));
	}
	dst += 1;
printf("-----------------------------------------------------------\n");
printf("memcpy_cmd, numIter = %d, numBytes = %d, src = %p, dst = %p\n", numIter, numBytes, src, dst);

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		oldmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("oldmemcpy_byte", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		fastmemcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("fast_memcpy_byte", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		memcpy(dst, src, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("memcpy_byte", numBytes, numIter, eee_rtcToUsec(&delta));

	if (numBytes == 6) {
	    eee_rtc(&startTime);
		for (i = 0; i < numIter; i++)
			memcpy6(dst, src, numBytes);
		eee_rtcDuration(&startTime, &delta);

		mb_printResult("memcpy6", numBytes, numIter, eee_rtcToUsec(&delta));
	}
#endif

#if 1

	old_bzero(dst, numBytes);

	printf("bzero, numIter = %d, numBytes = %d, dst = %p\n",
														numIter, numBytes, dst);

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		old_bzero(dst, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("old_bzero", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		bzero(dst, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("new_bzero", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		memset(dst, 0, numBytes);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("memset0", numBytes, numIter, eee_rtcToUsec(&delta));
	
#endif

#if 0
	
	numBytes = 128;

#if defined(NFP_TXRX)
	src =  (uchar8 *)eee_allocateDesc(EEE_FP_POOL, NFX_COS_HIGH);
#elif defined(NFP_FP)
	src =  (uchar8 *)eee_allocateDesc(EEE_TXRX_POOL, NFX_COS_HIGH);
#endif

	printf("ht_memcpy, numIter = %d, numBytes = %d, dst = %p, src = %p\n",
													numIter, numBytes, dst, src);

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		ht_memcpy_fast(dst, src, 4);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("ht_memcpy_fast", numBytes, numIter, eee_rtcToUsec(&delta));

    eee_rtc(&startTime);
	for (i = 0; i < numIter; i++)
		ht_memcpy_fast_with_prefetch(dst, src, 4);
    eee_rtcDuration(&startTime, &delta);

    mb_printResult("ht_memcpy_fast_with_prefetch", numBytes, numIter, eee_rtcToUsec(&delta));

	eee_deallocateDesc(src);
#endif

	eee_ramDealloc(save_dst);

	return 0;
}

#endif //  SSC_MGMT
