/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tftp-cmd.c
 *
 * RCS:  $Id: tftp-cmd.c,v 1.1.1.1 2001/10/12 03:52:49 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains TFTP command UI routines.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 09/25/2001
 *
 *-----------------------------------------------------------------
 */
#include    <nfx-incl.h>

#include    <exports/pkt-api.h>
#include    <exports/tftp-api.h>
#include    "tftp.h"


tftp_cb_t	*TftpCbPtr;
extern ushort16 nfsPort;

int32 tftp_cmdGet(  int32 argc, char *argv[] );
int32 tftp_cmdHelp( int32 argc, char *argv[] );
int32 tftp_cmdMode( int32 argc, char *argv[] );
int32 tftp_cmdShow( int32 argc, char *argv[] );

typedef struct
{
	int32	(*cmdFunc)(int32 argc, char *argv[]);
	char	*cmd;
	char	*helpTxt;
} tftp_cmdTbl_t;

static tftp_cmdTbl_t	tftp_cmdTable[ ] =
{
	{tftp_cmdGet,	"get",	"tftp get <ipaddr> <file>"},
	{tftp_cmdMode,	"mode",	"tftp mode ascii|bin"},
	{tftp_cmdShow,	"show", "tftp show cfg|stats" },
};

#define TFTP_CMD_MAX	sizeof(tftp_cmdTable)/sizeof(tftp_cmdTbl_t)

/*-----------------------------------------------------------------
 * Name:        tftp_cmdDataInd
 *
 * Description: Stub data rcv routine.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmdDataInd( char *pData, int32 size )
{
/*
    luc_trace("tftp_cmdDataInd", LUC_TRACE_PKT, 0, pData, size);
*/
	return 0;
}

/*-----------------------------------------------------------------
 * Name:        tftp_cmdEventInd
 *
 * Description: Stub event indication routine.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 	tftp_cmdEventInd(void *tftpCb, int32 event) 
{
	switch (event)
	{
		case	TFTP_EVT_TIMEOUT:
				printf("TFTP Timeout.\n");
				break;

		case	TFTP_EVT_ERROR:
				printf("TFTP Error.\n");
				break;

		case	TFTP_EVT_EOF:
				printf("TFTP transfer completed.\n");
				break;

		default:
				printf("Unknown TFTP Event.\n");
				break;
	}

	tftp_close( tftpCb );
	return 0;
}

/*-----------------------------------------------------------------
 * Name: 		tftp_cmdHelp
 *
 * Description: Display help-texts for all TFTP commands.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 09/25/2001
 *-----------------------------------------------------------------
 */
int32 tftp_cmdHelpAll( void )
{
    int32 i;

    printf("Usage:");
    for (i=0; i < TFTP_CMD_MAX; i++)
    {
        printf("\t%s\n", tftp_cmdTable[i].helpTxt);
    }
    return 0;
}

/*-----------------------------------------------------------------
 * Name: 		tftp_cmdHelpCmd
 *
 * Description: Display help-texts for an individual TFTP command.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 09/25/2001
 *-----------------------------------------------------------------
 */
int32 tftp_cmdHelpCmd( char *pCmd )
{
    int32 i;

    for (i=0; i < TFTP_CMD_MAX; i++)
    {
        if (strequ(pCmd, tftp_cmdTable[i].cmd))
        {
            printf("Usage:");
            printf("\t%s\n", tftp_cmdTable[i].helpTxt);
            return 0;
        }
    }

    return (tftp_cmdHelpAll( ));
}

/*-----------------------------------------------------------------
 * Name: 		tftp_cmdHelp
 *
 * Description: Display TFTP help-texts.
 *              tftp help [show | get | mode]
 * 
 * Created by : Shaun Hui
 * 
 * Date Created: 09/25/2001
 *-----------------------------------------------------------------
 */
int32 tftp_cmdHelp( int32 argc, char *argv[] )
{
    int32 i;

    if (argc < 3)
        return (tftp_cmdHelpAll());
    else
        return (tftp_cmdHelpCmd( argv[2] ));

}

/*-----------------------------------------------------------------
 * Name:        tftp_cmdShowCmd
 *
 * Description: Show TFTP Config
 *				tftp show cfg
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmdShowCfg( int32 argc, char *argv[] )
{
	tftp_cb_t	*tftpCb;

	tftpCb = TftpCbPtr;

	printf("TFTP Config \n");
	printf("-------------------------------------\n");
	printf("Mode      : %s\n", (tftpCb->mode == TFTP_MODE_ASCII) ? 
								"ascii" : "bin");
	printf("File Name : %s\n", tftpCb->filename);
	printf("Local IP  : %d.%d.%d.%d ", 
						((tftpCb->laddr >> 24) & 0xFF),
						((tftpCb->laddr >> 16) & 0xFF),
						((tftpCb->laddr >>  8) & 0xFF),
						((tftpCb->laddr)       & 0xFF));
	printf("Local Port  : %d\n", tftpCb->lport );
	printf("Server IP : %d.%d.%d.%d ", 
						((tftpCb->raddr >> 24) & 0xFF),
						((tftpCb->raddr >> 16) & 0xFF),
						((tftpCb->raddr >>  8) & 0xFF),
						((tftpCb->raddr)       & 0xFF));
	printf("Server Port : %d\n", tftpCb->rport );
	return 0;
}

/*-----------------------------------------------------------------
 * Name:        tftp_cmdShowStats
 *
 * Description: Show TFTP Statistics
 *				tftp show stats
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmdShowStats( int32 argc, char *argv[] )
{
	tftp_stats_t	*pStats;

	pStats = &TftpCbPtr->stats;

	printf("TFTP Stats \n");
	printf("-------------------------------------\n");
	printf("Rx RRQ    : %10d\n", pStats->rcvRrqCnt);
	printf("Rx WRQ    : %10d\n", pStats->rcvWrqCnt);
	printf("Rx Data   : %10d\n", pStats->rcvDataCnt);
	printf("Rx Ack    : %10d\n", pStats->rcvAckCnt);
	printf("Rx Err    : %10d\n", pStats->rcvErrCnt);
	printf("Tx RRQ    : %10d\n", pStats->xmtRrqCnt);
	printf("Tx WRQ    : %10d\n", pStats->xmtWrqCnt);
	printf("Tx Data   : %10d\n", pStats->xmtDataCnt);
	printf("Tx Ack    : %10d\n", pStats->xmtAckCnt);
	printf("Tx Err    : %10d\n", pStats->xmtErrCnt);
	printf("Tx Retry  : %10d\n", pStats->xmtRetryCnt);
	printf("Tx No Rsrc: %10d\n", pStats->xmtNoRsrcCnt);
	return 0;
}

/*-----------------------------------------------------------------
 * Name:        tftp_cmdShow
 *
 * Description: Handle TFTP show commands.
 *				tftp show  cfg | stats
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmdShow( int32 argc, char *argv[] )
{
	if (argc != 3)
        return (tftp_cmdHelpCmd("show"));

	if (strequ(argv[2], "cfg"))
	{
		return (tftp_cmdShowCfg( argc, argv ));
	}
	else if (strequ(argv[2], "stats"))
	{
		return (tftp_cmdShowStats( argc, argv ));
	}
	else
		return (tftp_cmdHelpCmd("show"));
}

/*-----------------------------------------------------------------
 * Name:        tftp_cmdMode
 *
 * Description: Set TFTP transfer mode to ASCII | BINARY
 *				tftp mode <ascii | bin>
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmdMode( int32 argc, char *argv[] )
{
	if (argc != 3)
        return (tftp_cmdHelpCmd("mode"));

	if (strequ(argv[2], "ascii"))
		TftpCbPtr->mode = TFTP_MODE_ASCII;
	else if (strequ(argv[2], "bin"))
		TftpCbPtr->mode = TFTP_MODE_BIN;
	else
		return (tftp_cmdHelpCmd("mode"));

}

/*-----------------------------------------------------------------
 * Name:        tftp_cmdGet
 *
 * Description: tftp get <ipaddr> <file>
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmdGet( int32 argc, char *argv[] )
{
	uint32	srvrIp;
	char	*filename;
	void	*fd;

	if (argc != 4)
        return (tftp_cmdHelpCmd("get"));

	if (getIpAddr( argv[2], &srvrIp ) < 0)
        return (tftp_cmdHelpCmd("get"));

	filename = argv[3];

	fd = tftp_open(filename, srvrIp,  tftp_cmdDataInd, tftp_cmdEventInd);

	if (!fd)
		return -1;
}

/*-----------------------------------------------------------------
 * Name:        tftp_cmd
 *
 * Description: Main dispatch for TFTP commands. Called from 
 *				command table.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
int32 tftp_cmd( int32 argc, char *argv[])
{
    int32   i;

    if (argc < 2)
    {
        return (tftp_cmdHelp(argc, argv));
    }

    for (i=0; i < TFTP_CMD_MAX; i++)
    {
        if (strequ(argv[1], tftp_cmdTable[i].cmd))
        {
            return ((*tftp_cmdTable[i].cmdFunc)(argc, argv));
        }
    }
    return (tftp_cmdHelp(argc, argv));

}

/*-----------------------------------------------------------------
 * Name:        tftp_logErr
 *
 * Description: Display TFTP error messages & crash if needed.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/25/01
 *-----------------------------------------------------------------
 */
void
tftp_logErr( char *fileName, int32 lineNum, int32 errCode, void *info )
{
    char    *errStr;
    int32   do_panic;

    do_panic = 0;
    switch (errCode)
    {
        case    TFTP_ERR_NOT_INIT:
                errStr = "TFTP not initialized";
                do_panic = 1;
                break;

        case    TFTP_ERR_NO_RAM_RSRC:
                errStr = "No Memory Resource";
                do_panic = 1;
                break;

        case    TFTP_ERR_NO_TIMER:
                errStr = "No Timer Resource";
                do_panic = 1;
                break;

        case    TFTP_ERR_NO_FUNC:
                errStr = "No callback function(s)";
                do_panic = 1;
                break;

        case    TFTP_ERR_BIND_FAIL:
                errStr = "TPL binding failed";
                do_panic = 1;
                break;

		default:
                errStr = "Unknown Error";
                do_panic = 1;
                break;
	}

    printf("TFTP Error (%d): %s! File: %s Line: %d Info:%x\n",
            errCode, errStr, fileName, lineNum, info);

    if (do_panic)
        TFTP_PANIC()

}

