/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tftp-rx.c
 *
 * RCS:  $Id: tftp-rcv.c,v 1.2 2002/06/26 19:51:08 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains TFTP protocol receive related routines.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 09/15/2001
 *
 *-----------------------------------------------------------------
 */
#include	<nfx-incl.h>

#include	<exports/pkt-api.h>
#include	<exports/timer-api.h>
#include	<exports/tftp-api.h>
#include	"tftp.h"

static char spinChar[] = {'-', '\\', '|', '/'};

/*-------------------------------
 * Tftp opcode error messages.
 *-------------------------------
 */
static char *tftpErrMsg[ ] =
{
	"Undefine error",
	"File not found",
	"Permission denied",
	"Disk full",
	"Illegal TFTP operation",
	"Unknown transfer ID",
	"File already exists",
	"No such user"
};

static void
tftp_spin( ushort16 spin )
{
	printf("%c\r", spinChar[ spin & 0x03 ]);
}

/*-----------------------------------------------------------------
 * Name:        tftp_rcvErr
 *
 * Description: Handle TFTP Error packet (OPCODE 5)
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
static void
tftp_rcvErr( tftp_cb_t *tftpCb, pkt_desc_t *pDesc )
{
	tftp_hdr_t	*tftpHdr;
	char		*errMsg;
	ushort16	errCode;

	tftpHdr = (tftp_hdr_t *)pDesc->pHdr;
	errCode = tftpHdr->th_code & TFTP_ERRCODE_MASK;
	errMsg	= NULL;

	/*-------------------------------
	 * Check for ErrMsg string.
	 *-------------------------------
	 */
	if ( tftpHdr->th_msg[0] )
		errMsg  = &tftpHdr->th_msg[0];
	else
		errMsg = tftpErrMsg[errCode];

	printf("\nTFTP Error (%02d): %s\n\n", errCode, errMsg);

	tftpCb->flags |= TFTP_FLAG_ERR;

	/*-------------------------------
	 * Upcall app. about error.
	 *-------------------------------
	 */
	(*tftpCb->eventInd)( tftpCb, TFTP_EVT_ERROR );
}

/*-----------------------------------------------------------------
 * Name:        tftp_rcvRrq
 *
 * Description: Handle TFTP RRQ packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
static void
tftp_rcvRrq( tftp_cb_t *tftpCb, pkt_desc_t *pDesc )
{
	return; /* NOT SUPPORTED */
}

/*-----------------------------------------------------------------
 * Name:        tftp_rcvWrq
 *
 * Description: Handle TFTP WRQ packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
static void
tftp_rcvWrq( tftp_cb_t *tftpCb, pkt_desc_t *pDesc )
{
	return; /* NOT SUPPORTED */
}

/*-----------------------------------------------------------------
 * Name:        tftp_rcvAck
 *
 * Description: Handle TFTP Ack packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
static void
tftp_rcvAck( tftp_cb_t *tftpCb, pkt_desc_t *pDesc )
{
	return; /* NOT SUPPORTED */
}

/*-----------------------------------------------------------------
 * Name:        tftp_rcvData
 *
 * Description: Handle TFTP Data packet (OPCODE 3)
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
static void
tftp_rcvData( tftp_cb_t *tftpCb, pkt_desc_t *pDesc )
{
	tftp_hdr_t	*tftpHdr;
	int32		eof;

	tftpHdr = (tftp_hdr_t *)pDesc->pHdr;

	/*-------------------------------
	 * Display "spin" char.
	 *-------------------------------
	 */
	tftp_spin( ntohs(tftpHdr->th_block) );

	/*-------------------------------
	 * Make sure block # is correct.
	 *-------------------------------
	 */
	if ( ntohs(tftpHdr->th_block) != (tftpCb->block+1))
		return;

	/*-------------------------------
	 * Incr. block #. Clear retry count 
	 * & stop timer. 
	 *-------------------------------
	 */
	tftpCb->block++;
	tftpCb->retry = 0;
	tftp_stopTimer( tftpCb );

	/*-------------------------------
	 * Exclude opcode & block #
	 *-------------------------------
	 */
	pDesc->pHdr    += TFTP_ACK_SIZE;
	pDesc->hdrSize -= TFTP_ACK_SIZE;

	/*-------------------------------
	 * Upcall app. with data.
	 *-------------------------------
	 */
	if ((*tftpCb->dataInd)( pDesc->pHdr, pDesc->hdrSize) != 0)
    {
		tftp_sendError( tftpCb);
    }
	else
    {
        /*-------------------------------
         * Ack. block rcvd. 
         *-------------------------------
         */
        tftp_sendAck( tftpCb );
    }

	/*-------------------------------
	 * Set EOF flag if data is less
	 * than TFTP segment size (512).
	 * Let Tftp Timer upcall app.
	 *-------------------------------
	 */
	if (pDesc->hdrSize < TFTP_SEG_SIZE) 
	{
		tftpCb->flags |= TFTP_FLAG_EOF;
		return;
	}
	
}

/*-----------------------------------------------------------------
 * Name:        tftp_rcvInd
 *
 * Description: Upcalled by lower layer to handle packet received.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
int32	/* PKT_FREE_PKT */
tftp_rcvInd(void *pHdl, pkt_desc_t *pDesc)
{

	tftp_cb_t	*tftpCb;
	tftp_hdr_t	*tftpHdr;

	tftpHdr = (tftp_hdr_t *)((pkt_desc_t *)pDesc)->pHdr;
	tftpCb  = pHdl;
	
	/*-------------------------------
	 * Check tftp opcode.
	 *-------------------------------
	 */
	switch(ntohs(tftpHdr->th_opcode))
	{
		case	TFTP_OP_RRQ:
				tftpCb->stats.rcvRrqCnt++;
				tftp_rcvRrq( tftpCb, pDesc );
				break;

		case	TFTP_OP_WRQ:
				tftpCb->stats.rcvWrqCnt++;
				break;

		case	TFTP_OP_DATA:
				tftpCb->stats.rcvDataCnt++;
				tftp_rcvData( tftpCb, pDesc);
				break;
				
		case	TFTP_OP_ACK:
				tftpCb->stats.rcvAckCnt++;
				tftp_rcvAck( tftpCb, pDesc );
				break;

		case	TFTP_OP_ERROR:
				tftpCb->stats.rcvErrCnt++;
				tftp_rcvErr( tftpCb, pDesc );
				break;
	}

	return PKT_FREE_PKT;
}

