/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tftp-send.c
 *
 * RCS:  $Id: tftp-send.c,v 1.5 2002/10/22 01:30:42 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains TFTP protocol send relatedroutines.
 *
 * Created by:  Shaun Hui
 *
 * Date Created: 09/15/2001
 *
 *-----------------------------------------------------------------
 */
#include    <nfx-incl.h>

#include    <exports/pkt-api.h>
#include    <exports/tpl-api.h>
#include    "../sm-tpl/tpl.h"
#include    <exports/timer-api.h>
#include    <exports/tftp-api.h>
#include    <exports/bsdsock-api.h>
#include    "tftp.h"


/*-----------------------------------------------------------------
 * Name:        tftp_timeout
 *
 * Description: Handle TFTP receive timeout.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
static int32
tftp_timeout( uint32 tmr,
              void  *info1,
              void  *info2,
              void  *info3,
              void  *info4 )
{
    tftp_cb_t   *tftpCb;

    tftpCb = info1;

    /*-------------------------------
     * Upcall app. if EOF.
     *-------------------------------
     */
	if (tftpCb->flags & TFTP_FLAG_EOF)
	{
		(*tftpCb->eventInd)( info1, TFTP_EVT_EOF);
		return 0;
	}

    /*-------------------------------
     * Upcall app. if error.
     *-------------------------------
     */
	if (tftpCb->flags & TFTP_FLAG_ERR)
	{
	    (*tftpCb->eventInd)( info1, TFTP_EVT_ERROR );
		return 0;
	}

    /*-------------------------------
     * Upcall app. if max consecutive
	 * retries failed.
     * Retry cleared when data is rcvd.
     *-------------------------------
     */
    if (++tftpCb->retry > TFTP_MAX_RETRY)
    {
        (*tftpCb->eventInd)( info1, TFTP_EVT_TIMEOUT);
        return 0;
    }

    /*-------------------------------
     * Re-xmt previous RRQ or ACK.
     *-------------------------------
     */
	if (tftpCb->block == 0)
    	tftp_sendRrq( tftpCb );
	else
    	tftp_sendAck( tftpCb );

    tftpCb->stats.xmtRetryCnt++;

    return 0;
}


/*-----------------------------------------------------------------
 * Name:        tftp_senRrq
 *
 * Description: Send TFTP Read Request packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
void 
tftp_sendRrq(tftp_cb_t *tftpCb )
{
	tftp_hdr_t	*tftpHdr;
	pkt_desc_t	*pDesc;
	char		*modeStr;
	char		*cp;

	tftp_startTimer( tftpCb );

	pDesc = PKT_ALLOC_PACKET(1);
	if (!pDesc)
	{
		tftpCb->stats.xmtNoRsrcCnt++;
		return;
	}
	tftpCb->stats.xmtRrqCnt++;
	
	/*-------------------------------
	 * Setup TFTP Opcode.
	 *-------------------------------
	 */
	tftpHdr = (tftp_hdr_t *)pDesc->pHdr;
	tftpHdr->th_opcode = htons(TFTP_OP_RRQ);
	cp = tftpHdr->th_stuff;

	/*-------------------------------
	 * Copy filename & NULL terminate.
	 *-------------------------------
	 */
	strcpy(cp, tftpCb->filename);
	cp += strlen(tftpCb->filename);
	*cp++ = '\0';

	/*-------------------------------
	 * Copy mode string & NULL terminate.
	 *-------------------------------
	 */
	modeStr = (tftpCb->mode == TFTP_MODE_ASCII) ? "netascii" : "octet";
	strcpy(cp, modeStr);
	cp += strlen(modeStr);
	*cp++ = '\0';

	/*-------------------------------
	 * Setup size of xmt packet 
	 *-------------------------------
	 */
	pDesc->hdrSize = (cp - pDesc->pHdr);

	/*-------------------------------
	 * Send RRQ pkt thru slow path.
	 *-------------------------------
	 */
	bsd_SendSock(tftpCb->pBsdSock, 
				htonl(tftpCb->raddr), 
				htons(tftpCb->rport), 
				pDesc->pHdr, 
				BSD_BUF_FLAT, 
				pDesc->hdrSize, 
				0);

	PKT_FREE_PACKET(pDesc);
	
}

/*-----------------------------------------------------------------
 * Name:        tftp_senAck
 *
 * Description: Send TFTP Ack packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/15/01
 *-----------------------------------------------------------------
 */
void
tftp_sendAck( tftp_cb_t *tftpCb)
{
	tftp_hdr_t	*tftpHdr;
	pkt_desc_t	*pDesc;
	char		*cp;

	tftp_startTimer( tftpCb );

/*
	pDesc = tpl_allocPkt( tftpCb->pTplConn, 1);
*/
	pDesc = PKT_ALLOC_PACKET(1);
	if (!pDesc)
	{
		tftpCb->stats.xmtNoRsrcCnt++;
		return;
	}

	tftpCb->stats.xmtAckCnt++;
	
	/*-------------------------------
	 * Setup TFTP Opcode & block ID.
	 *-------------------------------
	 */
	tftpHdr = (tftp_hdr_t *)pDesc->pHdr;
	tftpHdr->th_opcode = htons(TFTP_OP_ACK);
	tftpHdr->th_block  = htons(tftpCb->block);

	/*-------------------------------
	 * Setup size of xmt packet 
	 *-------------------------------
	 */
	pDesc->hdrSize = TFTP_ACK_SIZE;

	/*-------------------------------
	 * Send ACK pkt thru fast path.
	 * TPL CONN  has already been created
	 * at this point.
	 *-------------------------------
	tpl_xmtPkt( tftpCb->pTplConn, pDesc );
	 */

	/*-------------------------------
	 * Send RRQ pkt thru slow path.
	 *-------------------------------
	 */
	bsd_SendSock(((tpl_connCb_t *)(tftpCb->pTplConn))->pSock, 
				0, 0,
				pDesc->pHdr, 
				BSD_BUF_FLAT, 
				pDesc->hdrSize, 
				0);

	PKT_FREE_PACKET(pDesc);
	
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	3/19/02
 *
 *-----------------------------------------------------------------
 */
void
tftp_sendError( tftp_cb_t *tftpCb)
{
	tftp_hdr_t	*tftpHdr;
	pkt_desc_t	*pDesc;
	char		*cp;

    //printf("sendError\n");
	tftp_startTimer( tftpCb );

/*
	pDesc = tpl_allocPkt( tftpCb->pTplConn, 1);
*/
	pDesc = PKT_ALLOC_PACKET(1);
	if (!pDesc)
	{
		tftpCb->stats.xmtNoRsrcCnt++;
		return;
	}

	tftpCb->stats.xmtAckCnt++;
	
	/*-------------------------------
	 * Setup TFTP Opcode & block ID.
	 *-------------------------------
	 */
	tftpHdr = (tftp_hdr_t *)pDesc->pHdr;
	tftpHdr->th_opcode = htons(TFTP_OP_ERROR);
	tftpHdr->th_code  = htons(3);		/* Disk Full */
	strcpy(tftpHdr->th_data, "err");

	/*-------------------------------
	 * Setup size of xmt packet 
     * opcode=2
     * reasoncode=2
     * errorstring=4
     * ============
     * total 8
	 *-------------------------------
	 */
	pDesc->hdrSize = 8;

	/*-------------------------------
	 * Send ACK pkt thru fast path.
	 * TPL CONN  has already been created
	 * at this point.
	 *-------------------------------
	tpl_xmtPkt( tftpCb->pTplConn, pDesc );
	 */

	/*-------------------------------
	 * Send ERR pkt thru slow path.
	 *-------------------------------
	 */
	bsd_SendSock(((tpl_connCb_t *)(tftpCb->pTplConn))->pSock, 
				0, 0,
				pDesc->pHdr, 
				BSD_BUF_FLAT, 
				pDesc->hdrSize, 
				0);

	PKT_FREE_PACKET(pDesc);
	
}

tftp_stopTimer( tftp_cb_t *tftpCb )
{
    eee_timerStop(tftpCb->tmr_id);
}

tftp_startTimer( tftp_cb_t *tftpCb )
{
    eee_timerStart(tftpCb->tmr_id, tftp_timeout, tftpCb, 0, 0, 0);
}

