/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	timer-tic.c
 *
 * RCS:  $Id: timer-tic.c,v 1.40 2002/11/27 22:11:14 tonyw Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Functions that manage the actual timer tic from the system
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <sys/tx3927.h>
#include <exports/timer-api.h>
#include <exports/eee-api.h>
#include <exports/except-api.h>
#include <exports/malloc-api.h>
#include <exports/cm-api.h>
#if defined(NFP_FP)
#include <exports/thread-api.h>
#include <exports/esm-api.h>
#endif

#include <bqueue.h>

#ifdef LINUX_TEST
#include <sys/time.h>
#include "unistd.h"
#endif

#ifdef NFX_PROFILING
extern void prof_sample(uint32 pc);
#endif

#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
#include <sys/mips.h>
#endif

#ifdef TX3927
uint32 count_per_msec = 1;
uint32 count_per_usec = 0;
uint32 count_per_sec = 1;
#else
uint64 count_per_msec = 1;
uint64 count_per_usec = 0;
uint64 count_per_sec = 1;
#endif
eeeGMT_t bootGMTTime = {0,0};
eeeGMT_t sscGMTTime = {0,0};
eeeRTC_t sscRTC = {0};
// This is the last value of the real time clock.
eeeRTC_t eeeLastRTC = {0};
#ifndef TIMER_USE_SB1250_TIMER
// This is the value of the real time clock before the last wrap.
eeeRTC_t eeeLastRTCPreWrap = {0};
#endif

static uint32 cpu_count_per_tic;

#ifdef NFX_SMP
nfx_spinlock_t	rtc_spin_lock;
#endif

#if defined(NFX_PROFILING) && defined(TIMER_TIC_INT) && (defined(EVAL_BOARD) || defined(AGILE_HARDWARE))

#ifndef TX3927
static uint32 next_compare;
static int32  prev_gmt_count;
#endif

/*-----------------------------------------------------------------
 * Name: timer_prof_int
 *
 * Description:  This is an interrupt routine that provides profiling.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/02
 *
 *-----------------------------------------------------------------
 */
void
timer_prof_int(uint32 sr, uint32 cause, address_t epc)
{
	prof_sample((uint32)epc);

#ifndef TX3927
	prev_gmt_count = _mips_mfc0(C0_COMPARE);
	next_compare += cpu_count_per_tic;
	_mips_mtc0(C0_COMPARE, next_compare);
#else
	{
		tx3927_tmr_t	* tmr = (tx3927_tmr_t *)TX3927_TMR0_BASE;
		tmr->tisr = 0;		/* negate interrupt */
	}
#endif

}
#endif /* NFX_PROFILING && TIMER_TIC_INT && (EVAL_BOARD || AGILE_HARDWARE) */ 

/*-----------------------------------------------------------------
 * Name:	timer_ticInit()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void
timer_ticInit()
{
    static uint32 timer_ticInited = 0;
#if defined(EVAL_BOARD) || defined(AGILE_HARDWARE)
	uint32	cpu_countfreq;
#if defined(NFX_PROFILING) && defined(TIMER_TIC_INT) && !defined(TX3927)
	uint32	count;
#endif /* NFX_PROFILING && TIMER_TIC_INT && !TX3927 */

	if (!timer_ticInited) {

        uint32 cpu_pipefreq;

#if defined(SIBYTE)
		extern uint32 get_busclock();
		cpu_pipefreq = get_busclock() * 2 * 1000000;
		cpu_countfreq = cpu_pipefreq / 2;
#elif defined(R7000)
		extern uint32 get_pipefreq();
		cpu_pipefreq = get_pipefreq();
		cpu_countfreq = cpu_pipefreq / 2;
#elif defined(TX3927)
		cpu_pipefreq = 100 * 1000000;
		cpu_countfreq = (cpu_pipefreq / 4) / 2;
#else
		panic("timer_ticInit: unsupported timer platform\n");
#endif
#ifdef TIMER_USE_SB1250_TIMER
		count_per_usec = 1;
#else
		count_per_usec = cpu_countfreq / 1000000;
#endif
		count_per_msec = count_per_usec * 1000;
		count_per_sec = count_per_msec * 1000;
		cpu_count_per_tic = count_per_usec * TIMER_TIC_MS * 1000;

		timer_ticInited = 1;
		//printf("[timer_ticInit] cpu_count_per_tic (%d ms) = %d\n", TIMER_TIC_MS, cpu_count_per_tic);


#if defined(NFX_PROFILING) && defined(TIMER_TIC_INT)

		// Start the interrupt handler for profiling.

#ifndef TX3927
		count = _mips_mfc0(C0_COUNT);
		next_compare = count + cpu_count_per_tic;
		_mips_mtc0(C0_COMPARE, next_compare);

		intr_registerHandler(7, timer_prof_int);
#else /* TX3927 */
		{
			tx3927_tmr_t		* tmr;

			tmr = (tx3927_tmr_t *)TX3927_TMR0_BASE;
			tmr->tcr = 0x60;
			tmr->cpra = 12500 * TIMER_TIC_MS;
			tmr->itmr = 0x8001;		/* interrupt enable | TZCE */
			tmr->ccdr = 0;

			intr_registerHandler(0x2d, timer_prof_int);

			tmr->tcr = 0xC0;		/* enable */
		}

#endif /* TX3927 */
#endif /* NFX_PROFILING && TIMER_TIC_INT */ 		
#ifdef TX3927
    *(volatile uint32 *)TX3927_TMTISR0 = 0;
    *(volatile uint32 *)TX3927_TMPGMR0 = 0;
    *(volatile uint32 *)TX3927_TMCPRB0 = 0xffffff;
    *(volatile uint32 *)TX3927_TMCPRA0 = 0xffffff;
	*(volatile uint32 *)TX3927_TMCCDR0 = 0;
	*(volatile uint32 *)TX3927_TMTCR0 = 0xc1;
#endif
		eee_rtcInit();
	}
#endif	/* EVAL_BOARD || AGILE_HARDWARE */

}	/* timer_ticInit() */

#if defined(NFP_FP)
struct waiting_thread
{
	esm_handle_t wt_handle;					/* waiting thread */
	SLIST_ENTRY(waiting_thread) wt_list;	 /* next in list */
};

/* List of threads waiting to receive the time */
static SLIST_HEAD (, waiting_thread) wait_time_threads;
/* lock to protect the list */
#if defined(NFX_SMP)
nfx_spinlock_t wait_time_lock;
#endif


/*-----------------------------------------------------------------
 * Name :	eee_TimeGMT_wait_valid()
 *
 * Description: Wait until we receive valid time from SSC. 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/26/02
 *
 *-----------------------------------------------------------------
 */
void
eee_TimeGMT_wait_valid()
{
	if (!eee_rtcSSCTimeInitted()) {
		struct waiting_thread *elm = eee_ramAlloc(sizeof(*elm));
		if (elm == NULL) {
			panic("eee_TimeGMT_wait_valid: out of memory\n");
		}
		elm->wt_handle = esm_getCurrHandle();
		NFX_SPIN_LOCK(&wait_time_lock);
		if (!eee_rtcSSCTimeInitted()) {
			/* Add ourselves to list and go to sleep */
			SLIST_INSERT_HEAD(&wait_time_threads, elm, wt_list);
			NFX_SPIN_UNLOCK(&wait_time_lock);
			thread_yield(NULL, 0);
			NFX_ASSERT(eee_rtcSSCTimeInitted());
		}
		else {
			NFX_SPIN_UNLOCK(&wait_time_lock);
			eee_ramDealloc(elm);
		}
	}
}


/*-----------------------------------------------------------------
 * Name :	notify_time_valid()
 *
 * Description: Notify all threads waiting for the time to become
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/26/02
 *
 *-----------------------------------------------------------------
 */
static void
notify_time_valid()
{
	while (!SLIST_EMPTY(&wait_time_threads)) {
		esm_event_t event;
		struct waiting_thread *elm;

		NFX_SPIN_LOCK(&wait_time_lock);
		elm = SLIST_FIRST(&wait_time_threads);
		SLIST_REMOVE_HEAD(&wait_time_threads, wt_list);
		NFX_SPIN_UNLOCK(&wait_time_lock);
		
		thread_event(elm->wt_handle, &event);
		eee_ramDealloc(elm);
	}
}


/*-----------------------------------------------------------------
 * Name :	get_time_thread()
 *
 * Description: thread to obtain the current time from SSC. Send the
 * packets to timekeeper application until it replies
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/26/02
 *
 *-----------------------------------------------------------------
 */
static void
get_time_thread(esm_handle_t handle, void *userInfo)
{
	uint32 tk_app = eee_getApplicationId("timekeeper");
	uint32 dp, sp;
	int dummy;
	dp = PACK_UNICAST_PORT_ADDRESS(0, 0, tk_app,
								   EEE_MGMT, NFX_COS_HIGH, 0);
	sp = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId,
								   tk_app, EEE_MGMT,
								   NFX_COS_HIGH, 0);
	while (!eee_rtcSSCTimeInitted()) {
		eee_sendMessage(&dummy, sizeof(dummy), dp, sp);
		/* Keep trying every second */
		thread_yield(NULL, 1);
	}
}


/*-----------------------------------------------------------------
 * Name :	eee_timerInitGMTTime()
 *
 * Description: Get the current time from SSC
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/26/02
 *
 *-----------------------------------------------------------------
 */
void
eee_timerInitGMTTime(eee_descPtr_t desc)
{
    esm_handle_t handle;
	NFX_SPIN_INIT(&wait_time_lock, SPINLOCK_LEVEL_WAIT_TIMER);
	SLIST_INIT(&wait_time_threads);
	handle = thread_create(NULL);
	thread_start(&handle, get_time_thread);
}
#endif


void 
eee_keepGMTtime (eee_descPtr_t desc)
{
	eeeGMT_t time;

    eee_copyPacketToLocalBuffer (&time, desc, sizeof(time), 1);

    if (!eee_rtcSSCTimeInitted())
    {
        // Initialize the real time clock and sscGMTTime
        eee_rtcSSCTimeInit(time);

		// Remember the time we booted.
		bootGMTTime = sscGMTTime;
#if defined(NFP_FP)
		notify_time_valid();
#endif
		
    } else {

		// Reinitialize the real time clock and sscGMTTime
        eee_rtcSSCTimeInit(time);
	}

	eee_freePacket (desc);
}

/*-----------------------------------------------------------------
 * Name :	eee_timerGetCurrTic()
 *
 * Description: This returns the real time clock in 10ms units.
 *
 *              Callers should try to use the RTC routines directly
 *              to avoid the performance hit of this routine.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
uint32
eee_timerGetCurrTic()
{
    eeeRTC_t clock;

    eee_rtc(&clock);
	return (uint32)(clock.val / cpu_count_per_tic);

}	/* eee_timerGetCurrTic() */

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
void
eee_timerShowTic()
{

    eeeRTC_t clock;

    eee_rtc(&clock);
#ifdef TX3927
    printf("showtic: low = %u, high = %u\n", clock.u.low, clock.u.high);
    printf("seconds tick: %u\n", (clock.u.low / count_per_sec));
#else
	printf("Current %lld ms tick count = %lld\n", count_per_usec, eee_rtcToMS(&clock));
#endif

}	/* eee_timerShowTic() */

#ifdef LINUX_TEST
/*-----------------------------------------------------------------
 * Name: eee_timerLinuxAdjustCurrTicCount()
 *
 * Description:	Polling function which should update the current tick
 *				counter based on how long the process has been running
 *
 *   This is obsoleted by the new timer API(s).
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
void
eee_timerLinuxAdjustCurrTicCount(
	uint32	* cb,
	uint32	tref
)
{
    // This is obsolete
    return;

}	/* eee_timerLinuxAdjustCurrTicCount() */
#endif	/* LINUX_TEST */

