/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	timer.c
 *
 * RCS:  $Id: timer.c,v 1.9 2002/11/22 21:11:30 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Internal functions for timer module
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/queue-api.h>
#include <exports/timer-api.h>
#include <exports/eee-api.h>
#include "timer.h"

nfx_spinlock_t wait_time_lock;


nfx_spinlock_t timer_spin;


#if defined(NFX_SMP) && defined(NFP_TXRX)
/* txrx acpu need a separate timer queue */
struct timer_queue timersAcpu;
#endif /* NFX_SMP */

struct timer_queue timers;
timer_block_t			* tmr_free_blk;

timer_block_t			tmr_block[TIMER_NUM_TIMER_BLOCKS];

extern void eee_timerLinuxAdjustCurrTicCount();

uint32 max_timers_per_loop = 10;
extern void eee_keepGMTtime (eee_descPtr_t desc);	// In timer-tic.c

/*-----------------------------------------------------------------
 * Name:	eee_timerInit()
 *
 * Description:	Initialize timer blocks
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
void
eee_timerInit()
{
	int	i;
	int32	rc;

    NFX_SPIN_INIT(&wait_time_lock, SPINLOCK_LEVEL_WAIT_TIMER);
    
	memset(tmr_block, 0, TIMER_NUM_TIMER_BLOCKS * sizeof(timer_block_t));

	/*
	 * Set up free lists
	 */
	tmr_free_blk = &(tmr_block[0]);
	for (i=0; i<TIMER_NUM_TIMER_BLOCKS-1; i++) {
		tmr_block[i].usr_p1 = &(tmr_block[i+1]);
		tmr_block[i].tmr_handle = i;
		tmr_block[i].tb_flags = TIMER_STATE_FREE|TIMER_STATE_STOPPED;
	}
	tmr_block[i].usr_p1 = NULL;
	tmr_block[i].tmr_handle = i;


	TAILQ_INIT(&timers);
#if defined(NFP_TXRX) && defined(NFX_SMP)
	TAILQ_INIT(&timersAcpu);
#endif /* NFX_SMP */

	NFX_SPIN_INIT(&timer_spin, SPINLOCK_LEVEL_WAIT_TIMER);

	rc = eee_registerPollingFunc(	EEE_HIGH_PRIORITY,
									eee_processTimers,
									(void *)((uintptr_t)max_timers_per_loop));
	if (rc < 0)
		panic("error in EEE polling registration\n");
/*
	Register the timekeeper
 */
	rc = eee_allocateApplicationId ("timekeeper");
	if (rc == EEE_ERR_INVALID_APP_NAME)
		panic("error getting timekeeper app id\n");

	rc = eee_registerReceiveFunc( 	rc,
									eee_keepGMTtime,
									eee_keepGMTtime);
	if (rc < 0)
		panic("error registering for timekeeper messages\n");

}	/* eee_timerInit() */


void
timer_block_add(struct timer_block *new)
{
	struct timer_block *tb;

#if defined(NFP_TXRX) && defined(NFX_SMP)
	struct timer_queue	*timerPtr;

	if (CM_IS_TXRX_NCPU)
		timerPtr = &timers;
	else
		timerPtr = &timersAcpu;
	
	if (new->tb_flags & TIMER_STATE_ONQUEUE) {
		TAILQ_REMOVE(timerPtr, new, tb_entry);
	}
	
	new->tb_flags |= (TIMER_STATE_ONQUEUE | TIMER_STATE_RUNNING);
	
	TAILQ_FOREACH(tb, timerPtr, tb_entry) {
		if (eee_rtcCompare(&new->tb_expiry_time, &tb->tb_expiry_time) < 0) {
			TAILQ_INSERT_BEFORE(tb, new, tb_entry);
			return;
		}
	}
	TAILQ_INSERT_TAIL(timerPtr, new, tb_entry);	
#else
	if (new->tb_flags & TIMER_STATE_ONQUEUE) {
		TAILQ_REMOVE(&timers, new, tb_entry);
	}
	
	new->tb_flags |= (TIMER_STATE_ONQUEUE | TIMER_STATE_RUNNING);
	
	TAILQ_FOREACH(tb, &timers, tb_entry) {
		if (eee_rtcCompare(&new->tb_expiry_time, &tb->tb_expiry_time) < 0) {
			TAILQ_INSERT_BEFORE(tb, new, tb_entry);
			return;
		}
	}
	TAILQ_INSERT_TAIL(&timers, new, tb_entry);
#endif /* NFX_SMP */
}

void
timer_block_stop(struct timer_block *tb)
{
	tb->tb_flags &= ~TIMER_STATE_RUNNING;
	tb->tb_expiry_time = *(eee_rtcZero());
}


void
timer_block_del(struct timer_block *tb)
{
#if defined(NFP_TXRX) && defined(NFX_SMP)
	struct timer_queue	*timerPtr;

	if (CM_IS_TXRX_NCPU)
		timerPtr = &timers;
	else
		timerPtr = &timersAcpu;

	if (tb->tb_flags & TIMER_STATE_ONQUEUE) {
		TAILQ_REMOVE(timerPtr, tb, tb_entry);
		tb->tb_flags &= ~TIMER_STATE_ONQUEUE;
	}
	timer_block_stop(tb);
#else
	if (tb->tb_flags & TIMER_STATE_ONQUEUE) {
		TAILQ_REMOVE(&timers, tb, tb_entry);
		tb->tb_flags &= ~TIMER_STATE_ONQUEUE;
	}
	timer_block_stop(tb);
#endif /* NFX_SMP */
}

/*-----------------------------------------------------------------
 * Name:	eee_processTimers()
 *
 * Description:	Timer polling function during which the timer blocks will
 *				be checked to see which has expired.  If a timer has expired,
 *				the callback function will be called.  If the timer is a
 *				repeating timer, it will be restarted.
 *
 *				This function will be registered with the EEE layer as a
 *				polling function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
void
eee_processTimers(
	void		* max_timers,		/* maximum timers to expire in a single
									 * iteration of this function
									 */
	uint32 		tref				/* from EEE layer */
)
{
	eeeRTC_t		curr_time;
	timer_block_t	* tmr;
	int cnt = 0;
#if defined(NFX_SMP) && defined(NFP_TXRX)
	struct timer_queue	*timerPtr;
#endif /* NFX_SMP */

	eee_rtc(&curr_time);

	NFX_SPIN_LOCK(&timer_spin);

#if defined(NFX_SMP) && defined(NFP_TXRX)
	if (CM_IS_TXRX_NCPU)
		timerPtr = &timers;
	else
		timerPtr = &timersAcpu;

	while ((tmr = TAILQ_FIRST(timerPtr)) != NULL
		   && eee_rtcCompare(&tmr->tb_expiry_time, &curr_time) <= 0
		   && ++cnt <= tref) {
		TAILQ_REMOVE(timerPtr, tmr, tb_entry);
		tmr->tb_flags &= ~TIMER_STATE_ONQUEUE;
		if (tmr->type == EEE_TIMER_ONCE) {
			timer_block_stop(tmr);
		}
		else {
			eee_rtcAdd(&curr_time, &tmr->duration, &tmr->tb_expiry_time);
			timer_block_add(tmr);
		}
		NFX_SPIN_UNLOCK(&timer_spin);
		tmr->callback_func(	tmr->tmr_handle,
							tmr->usr_p1,
							tmr->usr_p2,
							tmr->usr_p3,
							tmr->usr_p4);
		NFX_SPIN_LOCK(&timer_spin);
	}
#else
	while ((tmr = TAILQ_FIRST(&timers)) != NULL
		   && eee_rtcCompare(&tmr->tb_expiry_time, &curr_time) <= 0
		   && ++cnt <= tref) {
		TAILQ_REMOVE(&timers, tmr, tb_entry);
		tmr->tb_flags &= ~TIMER_STATE_ONQUEUE;
		if (tmr->type == EEE_TIMER_ONCE) {
			timer_block_stop(tmr);
		}
		else {
			eee_rtcAdd(&curr_time, &tmr->duration, &tmr->tb_expiry_time);
			timer_block_add(tmr);
		}
		NFX_SPIN_UNLOCK(&timer_spin);
		tmr->callback_func(	tmr->tmr_handle,
							tmr->usr_p1,
							tmr->usr_p2,
							tmr->usr_p3,
							tmr->usr_p4);
		NFX_SPIN_LOCK(&timer_spin);
	}
#endif /* NFX_SMP */

	NFX_SPIN_UNLOCK(&timer_spin);

}	/* eee_processTimers() */

