/*
 *-----------------------------------------------------------------
 *
 * Name:	timer.h
 *
 * RCS:  $Id: timer.h,v 1.8 2002/11/22 21:11:30 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Internal functions for timers
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _TIMER_H
#define _TIMER_H

#include <exports/queue-api.h>
#include <exports/timer-api.h>
#include <bqueue.h>

#define TIMER_NUM_TIMER_BLOCKS		1024
#define TIMER_NUM_EXPIRY_BLOCKS		1024

typedef struct timer_block {
	TAILQ_ENTRY(timer_block) tb_entry;
	eeeRTC_t			duration;
	eeeRTC_t			tb_expiry_time;
	uint32				tb_flags;
	uint32				type;
	address_t			owner;
	uint32				tmr_handle;
	timer_callback_func	callback_func;
	void				* usr_p1;
	void				* usr_p2;
	void				* usr_p3;
	void				* usr_p4;
} timer_block_t;


TAILQ_HEAD(timer_queue, timer_block);
extern struct timer_queue timers;
extern timer_block_t	* tmr_free_blk;
extern nfx_spinlock_t timer_spin;

extern timer_block_t	tmr_block[];

#define TIMER_STATE_ALLOC		0x1
#define TIMER_STATE_RUNNING		0x2
#define TIMER_STATE_ONQUEUE		0x4
#define TIMER_STATE_STOPPED     0x0
#define TIMER_STATE_FREE		0x0

/*-----------------------------------------------------------------
 * Name:	eee_processTimers()
 *
 * Description:	Timer polling function during which the timer blocks will
 *				be checked to see which has expired.  If a timer has expired,
 *				the callback function will be called.  If the timer is a
 *				repeating timer, it will be restarted.
 *
 *				This function will be registered with the EEE layer as a
 *				polling function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
extern void
eee_processTimers(
	void		* max_timers,		/* maximum timers to expire in a single
									 * iteration of this function
									 */
	uint32 		tref				/* from EEE layer */
);

/*-----------------------------------------------------------------
 * Name:	eee_timerSetRunningState()
 *
 * Description:	Error checking on tmr_handle must already have been done,
 *				since no checking will be done here.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_timerSetRunningState(
	int	tmr_handle,
	uint32	tmr_running_state
)
{
	tmr_block[tmr_handle].tb_flags &= ~TIMER_STATE_RUNNING;
	tmr_block[tmr_handle].tb_flags |= tmr_running_state;
}	/* eee_timerSetRunningState() */

/*-----------------------------------------------------------------
 * Name:	eee_timerGetRunningState()
 *
 * Description:	Error checking on tmr_handle must already have been done,
 *				since no checking will be done here.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
static inline uint32
eee_timerGetRunningState(
	int	tmr_handle
)
{
	return tmr_block[tmr_handle].tb_flags & TIMER_STATE_RUNNING;
}	/* eee_timerGetRunningState() */

/*-----------------------------------------------------------------
 * Name:	eee_timerSetAllocState()
 *
 * Description:	Error checking on tmr_handle must already have been done,
 *				since no checking will be done here.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_timerSetAllocState(
	int	tmr_handle,
	uint32	tmr_alloc_state
)
{
	tmr_block[tmr_handle].tb_flags &= ~TIMER_STATE_ALLOC;
	tmr_block[tmr_handle].tb_flags |= tmr_alloc_state;
}	/* eee_timerSetAllocState() */

/*-----------------------------------------------------------------
 * Name:	eee_timerGetAllocState()
 *
 * Description:	Error checking on tmr_handle must already have been done,
 *				since no checking will be done here.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
static inline uint32
eee_timerGetAllocState(
	int	tmr_handle
)
{
	return tmr_block[tmr_handle].tb_flags & TIMER_STATE_ALLOC;
}	/* eee_timerGetAllocState() */

/*-----------------------------------------------------------------
 * Name:	eee_timerSetExpTime()
 *
 * Description:	sets expiration time in array.  This array will be used
 *				in the timerStop() function, which needs to get the
 *				expiration time to find the timer handle in the
 *				timers queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_timerSetExpTime(
	int		tmr_handle,
	eeeRTC_t	*exp_time
)
{
	tmr_block[tmr_handle].tb_expiry_time =
		exp_time ? *exp_time : *(eee_rtcZero());
}	/* eee_timerSetExpTime() */

/*-----------------------------------------------------------------
 * Name:	eee_timerGetExpTime()
 *
 * Description:	returns the expiration time for the given timer handle.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/29/01
 *
 *-----------------------------------------------------------------
 */
static inline eeeRTC_t
eee_timerGetExpTime(
	int	tmr_handle
)
{
	return tmr_block[tmr_handle].tb_expiry_time;
}	/* eee_timerGetExpTime() */


void timer_block_add(struct timer_block *tb);
void timer_block_del(struct timer_block *tb);
void timer_block_stop(struct timer_block *tb);

#endif	/* _TIMER_H */

