/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-bind.c
 *
 * RCS:  $Id: tpl-bind.c,v 1.13 2002/03/13 22:48:52 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer interface binding related routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 07/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/tpl-api.h>
#include "tpl.h"
#include <exports/bsd-api.h>
#include <exports/bsdsock-api.h>



/*-----------------------------------------------------------------
 * Name: 		tpl_findBindCb
 *
 * Description: Returns Bind CB that matches protocol & port number.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 07/01/2001
 *
 *-----------------------------------------------------------------
 */
tpl_bindCb_t * 
tpl_findBindCb( ushort16 prot, ushort16 lport )
{
	tpl_bindCb_t	*pBind;
	int32			i;

	printf("tpl_findBindCb prot=%d lport=%d\n", prot, lport);
	/*-------------------------------
	 * Find Bind CB that matches
	 * IP, port & protocol.
	 *-------------------------------
	 */
	pBind = pTplCfg->bindTblBase;
	for (i=0; i < pTplCfg->maxBindCb; i++)
	{
		if ((pBind->lport   != TPL_ANY_PORT)	&& 
			(pBind->lport   == lport) 			&& 
            (pBind->tplProt == prot))
			return(pBind);

		pBind++;
	}
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name: tpl_allocBindCb
 *
 * Description: Allocates a Bind CB. Returns NULL if none available.
 *
 *-----------------------------------------------------------------
 */
static tpl_bindCb_t *
tpl_allocBindCb( void )
{
	tpl_bindCb_t	*pBind;
	int32			i;

	printf("tpl_allocBindCb\n");
	pBind = pTplCfg->bindTblBase;
	for (i=0; i < pTplCfg->maxBindCb; i++)
	{
		if (!(pBind->lport) 		&& !(pBind->tplProt) && 
			!(pBind->ul_addConnInd) && !(pBind->ul_delConnInd) && 
			!(pBind->ul_rcvPktInd)  && !(pBind->ul_xmtCmpInd)) 
			return(pBind);

		pBind++;
	}
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name: tpl_freeBindCb
 *
 * Description: Frees a Bind CB. 
 *
 *-----------------------------------------------------------------
 */
static void
tpl_freeBindCb( tpl_bindCb_t *pBind )
{
	memset(pBind, 0, sizeof(tpl_bindCb_t));
}

/*-----------------------------------------------------------------
 * Name: tpl_validateBindReq
 *
 * Description: Validate user inputs to bind request 
 *
 *-----------------------------------------------------------------
 */
static int32
tpl_validateBindReq( tpl_bindCb_t *pBind, int32 bindMode)
{
	switch (bindMode)
	{
		/*-------------------------------
		 * For client bind, laddr, raddr 
		 * & rport are REQUIRED.
		 *-------------------------------
		 */
		case	TPL_BIND_CLIENT:
				if (pBind->laddr == 0)
					return (TPL_ERR_NO_LADDR );

				if (pBind->raddr == 0) 
					return (TPL_ERR_NO_RADDR );

				if (pBind->rport == 0)
					return (TPL_ERR_NO_RPORT );
				break;

		/*-------------------------------
		 * For server bind, raddr * & rport 
		 * MUST NOT be set.
		 *-------------------------------
		 */
		case	TPL_BIND_SERVER:
/*
				if (pBind->laddr == 0)
					return (TPL_ERR_NO_LADDR );
*/
				if (pBind->lport == 0)
					return (TPL_ERR_NO_LPORT );

				if (pBind->raddr != 0) 
					return (TPL_ERR_RADDR_IN_SRVR );

				if (pBind->rport != 0)
					return( TPL_ERR_RPORT_IN_SRVR );
				break;

		default:
				return (TPL_ERR_BAD_MODE );
	}

	/*-------------------------------
	 * Accepts TCP & UDP only.
	 *-------------------------------
	 */
	if ((pBind->tplProt != TPL_PROT_UDP) && 
		(pBind->tplProt != TPL_PROT_TCP))
	{
		return (TPL_ERR_BAD_PROT );
	}

	/*-------------------------------
	 * Checks for upcalls are set.
	 *-------------------------------
	 */
	if ( !(pBind->ul_addConnInd) || !(pBind->ul_delConnInd) ||
		 !(pBind->ul_rcvPktInd)  || !(pBind->ul_xmtCmpInd) )
	{
		return (TPL_ERR_BAD_FUNC );
	}

	return NFX_OK;
}


/*-----------------------------------------------------------------
 * Name:        tpl_bindReq
 *
 * Description: Binds upper layer to Transport layer. Also binds
 *				TPL to BSD layer for slow-path.
 *
 *              The following table shows the fields required 
 *              in the tpl_bindCb for Server or Client bind request.
 *
 *                          SERVER          CLIENT
 *              ------------------------------------
 *              laddr       Required        Required
 *              lport       Required        Optional
 *              raddr       No              Required
 *              rport       No              Required
 *              addConnInd  Required        Required
 *              delConnInd  Required        Required
 *              rcvPktInd   Required        Required
 *              xmtCmpInd   Required        Required
 *
 *              NOTE: laddr, lport, raddr, rport are expected to
 *                    be in NETWORK ORDER (Big Endian)
 * Created by : Shaun Hui
 *
 * Date Created: 7/01/01
 *-----------------------------------------------------------------
 */
#if defined(NFX_SMP)
int32
tpl_bindReqCore(tpl_bindCb_t *pReq , 	/* ptr to Bind Req Cb */
				tpl_bindCb_t **pRsp,	/* ptr to Bind Response Cb ptr */
				int32		 bindOpt)	/* see tpl-api.h for options. */
#else
int32
tpl_bindReq(tpl_bindCb_t *pReq , 	/* ptr to Bind Req Cb */
			tpl_bindCb_t **pRsp,	/* ptr to Bind Response Cb ptr */
			int32		 bindOpt)	/* see tpl-api.h for options. */
#endif /* NFX_SMP */
{
	tpl_bindCb_t	*pBind;
	void			*pSock;
	int32			protocol;
	int32			bindType;
	int32			rc;
	ushort16		flags;
	bsd_sarg_t		sockArg;

	pBind = pReq;
	flags = pBind->bindFlags;

	/*-------------------------------
	 * Check user input.
	 *-------------------------------
	 */
	rc = tpl_validateBindReq( pReq, (bindOpt & TPL_BIND_CLIENT) );
	if (rc != NFX_OK)
	{
		TPL_LOGERR( rc, pBind );
		return (rc);
	}

	/*-------------------------------
	 * Checks for duplication.
	 *-------------------------------
	 */
	if (tpl_findBindCb( pBind->tplProt, pBind->lport ) != NULL)
	{
		TPL_LOGERR( TPL_ERR_DUP_BIND, pBind );
		return (TPL_ERR_DUP_BIND);
	}

	/*-------------------------------
	 * Alloc tpl_bindCb.
	 *-------------------------------
	 */
	pBind = tpl_allocBindCb();
	if (pBind == NULL)
	{
		TPL_LOGERR( TPL_ERR_NO_BIND_RSRC, pBind );
		return (TPL_ERR_NO_BIND_RSRC);
	}

	/*-------------------------------
	 * Copy request into new bindCb.
	 * Return real bindCb in pRsp.
	 *-------------------------------
	 */
	memcpy(pBind, pReq, sizeof(tpl_bindCb_t));
	pBind->pConnList = NULL;
	pBind->pBsdSock  = NULL;
	
	/*-------------------------------
	 * Setup BSD socket arg for binding.
	 *-------------------------------
	 */
	memset( &sockArg, 0, sizeof(bsd_sarg_t));

	sockArg.laddr   = pBind->laddr;		/* Local  IP */
	sockArg.faddr   = pBind->raddr;		/* Remote IP */
	sockArg.lport   = pBind->lport;		/* Local  Port */
	sockArg.fport   = pBind->rport;		/* Remote Port */
	sockArg.handle  = pBind;			/* Handle */
	sockArg.bufType = BSD_BUF_DESCR;	/* Expects pkt desc */
	sockArg.cb      = tpl_bsdConnInd;	/* Callback for bsd conn event */
	sockArg.db      = tpl_bsdRcvInd;	/* Callback for bsd data event */
	sockArg.flags   = BSD_SO_FASTPATH;

	if (bindOpt & TPL_BIND_CLIENT)
		pBind->bindFlags |= TPL_FLAG_CLIENT;

	if (bindOpt & TPL_BIND_USE_RPC)
		pBind->bindFlags |= TPL_FLAG_USE_RPC;

	if (bindOpt & TPL_BIND_PKTCNT_SM)
		pBind->bindFlags |= TPL_FLAG_PKTCNT_SM;

	/*-------------------------------
	 * Setup bsd bind protocol & type.
	 *-------------------------------
	 */
	bindType = (bindOpt & TPL_BIND_CLIENT) ? 
				BSD_SO_CLIENT : BSD_SO_SERVER;

	protocol = (pBind->tplProt == TPL_PROT_UDP) ? 
				BSD_IPPROTO_UDP : BSD_IPPROTO_TCP;

	/*-------------------------------
	 * Bind with FreeBSD Code.
	 *-------------------------------
	 */
	pSock = bsd_OpenSock(protocol, bindType, &sockArg);

	if (!pSock)
	{
		TPL_LOGERR( TPL_ERR_NO_BIND_RSRC, pBind );
		return (TPL_ERR_NO_BIND_RSRC);
	}

	pBind->pBsdSock = pSock;
	*pRsp = pBind;

	pTplCfg->tplStats.cntBindReq++;
	pTplCfg->numBindCb++;

	return (NFX_OK);
}

#if defined(NFX_SMP)
/*-----------------------------------------------------------------
 * Name:        tpl_bindReq
 *
 * Description:bind request on ACPU
 *
 * Created by : Su Wangg
 *
 * Date Created: 12/02/02
 *-----------------------------------------------------------------
 */
int32
tpl_bindReq(tpl_bindCb_t *bindReq, void* handle, ushort16 bindOpt)
{
	tpl_bindCb_t	myBindReq;
	
	memcpy(&myBindReq, bindReq, sizeof(tpl_bindCb_t));
	
	/* use flags field for options, and pBsdSock field 
	 * for handle
	 */
	myBindReq.pBsdSock = handle;
	myBindReq.bindFlags = bindOpt;
	
	/* send to NCPU */
	return tpl_ipcSendEvntMsg(TPL_EVNT_BIND_REQ, &myBindReq, 0);
}
#endif /* NFX_SMP */

/*-----------------------------------------------------------------
 * Name:        tpl_unbindReq
 *
 * Description: Unbinds upper layer from Transport layer. 
 *				Deletes all connections associated with bindCB.
 *				Also removes TPL from BSD layer for slow-path.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 12/12/01
 *-----------------------------------------------------------------
 */
#if defined(NFX_SMP)
int32
tpl_unbindReqCore( tpl_bindCb_t *pBind ) 
#else
int32
tpl_unbindReq( tpl_bindCb_t *pBind ) 	/* ptr to Bind Cb for unbind */
#endif /* NFX_SMP */
{
    /*-------------------------------
     * Check if bindCb is valid.
     *-------------------------------
     */
    if (tpl_validateBindCb( pBind ) != NFX_OK)
        return ( TPL_ERR_NOT_BIND );

    /*-------------------------------
     * Removes associated connections.
	 * pConnList modified in delConnReq().
     *-------------------------------
     */
	while (pBind->pConnList)
#if defined(NFX_SMP)
		tpl_delConnReqCore(pBind->pConnList);
#else
		tpl_delConnReq(pBind->pConnList);
#endif /* NFX_SMP */
	
	/*-------------------------------
	 * Close assoc. BSD sock.
	 *-------------------------------
	 */
	if (pBind->pBsdSock)
		bsd_CloseSock(pBind->pBsdSock);

	/*-------------------------------
	 * Deallocate bindCb. 
	 *-------------------------------
	 */
	tpl_freeBindCb( pBind );

	pTplCfg->tplStats.cntUnbindReq++;
	pTplCfg->numBindCb--;

	return NFX_OK;
}

#if defined(NFX_SMP)
/*-----------------------------------------------------------------
 * Name:        tpl_unbindReq
 *
 * Description: unbind request on ACPU
 *
 * Created by : Su Wangg
 *
 * Date Created: 12/02/02
 *-----------------------------------------------------------------
 */
int32
tpl_unbindReq(tpl_bindCb_t *pBind)
{
	/* send to NCPU */
	return tpl_ipcSendEvntMsg(TPL_EVNT_UNBIND_REQ, pBind, 0);
}
#endif /* NFX_SMP */
