/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-cmd.c
 *
 * RCS:  $Id: tpl-cmd.c,v 1.4 2002/10/17 01:38:56 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains routines to handle TPL UI Command.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 11/10/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/luc-api.h>
#include <exports/tpl-api.h>
#include "tpl.h"

int32 tpl_cmdHelp( int32 argc, char *argv[] );
int32 tpl_cmdShow( int32 argc, char *argv[] );
int32 tpl_cmdLoop( int32 argc, char *argv[] );
int32 tpl_cmdClear( int32 argc, char *argv[] );
int32 tpl_cmdUnbind( int32 argc, char *argv[] );

typedef struct
{
	int32	(*cmdFunc)( int32 argc, char *argv[] );
	char	*cmd;
	char	*helpTxt;
} tpl_cmdTbl_t;

tpl_cmdTbl_t	tpl_cmdTable[ ] =
{
	{ tpl_cmdUnbind, "unbind", 	"tpl unbind" },
	{ tpl_cmdShow, "show", 	"tpl show bind|cfg|conn|pkt|stats|ipc" },
	{ tpl_cmdLoop, "loop", 	"tpl loop <pkt len> <pkt count>" },
	{ tpl_cmdClear, "clear","tpl clear stats" }
};

#define TPL_CMD_MAX		sizeof(tpl_cmdTable)/sizeof(tpl_cmdTbl_t)

/*-----------------------------------------------------------------
 * Name: 		tpl_cmdUnbind
 *
 * Description: Command to test TPL unbind. 
 *
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmdUnbind( int32 argc, char *argv[] )
{
	ushort16 lport;

	lport = (argc == 3) ?  lport = atoi(argv[argc-1]) : TPL_ANY_PORT;

	tpl_unbindPort( lport );

	return 0;
}

/*-----------------------------------------------------------------
 * Name: 		tpl_cmdLoop
 *
 * Description: Command to do TPL layer loopback. 
 *
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmdLoop( int32 argc, char *argv[] )
{
    int32   pktCnt;
    int32   pktLen;

    if (argc != 4)
        return -1;

    if ((pktLen = atoi(argv[2])) < 0)
        return -1;

    if ((pktCnt = atoi(argv[3])) < 0)
        return -1;

    if (pktLen < 64)
    {
        pktLen = 64;
        printf("pkt len < 64 bytes, use 64 bytes instead.\n");
    }

    tpl_loopTestStart( pktLen, pktCnt );
    return 0;
}

/*-----------------------------------------------------------------
 * Name: tpl_cmdHelp
 *
 * Description: Display help-texts for all TPL commands.
 *
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmdHelpAll( void )
{
	int32 i;

	printf("Usage:");
	for (i=0; i < TPL_CMD_MAX; i++)
	{
		printf("\t%s\n", tpl_cmdTable[i].helpTxt);
	}
	return 0;
}

/*-----------------------------------------------------------------
 * Name: tpl_cmdHelpCmd
 *
 * Description: Display help-texts for an individual TPL command.
 *
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmdHelpCmd( char *pCmd )
{
	int32 i;

	for (i=0; i < TPL_CMD_MAX; i++)
	{
		if (strequ(pCmd, tpl_cmdTable[i].cmd))
		{
			printf("Usage:");
			printf("\t%s\n", tpl_cmdTable[i].helpTxt);
			return 0;
		}
	}

	return (tpl_cmdHelpAll( ));
}

/*-----------------------------------------------------------------
 * Name: tpl_cmdHelp
 *
 * Description: Display TPL help-texts.  
 *
 *				tpl help [show | start | trace]
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmdHelp( int32 argc, char *argv[] )
{

	if (argc < 3)
		return (tpl_cmdHelpAll());
	else 
		return (tpl_cmdHelpCmd( argv[2] ));

}

/*-----------------------------------------------------------------
 * Name: tpl_cmdShow
 *
 * Description: Dispatch for TPL SHOW commands.  
 *
 *				tpl show stats
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmdShow( int32 argc, char *argv[] )
{

	ushort16 	lport;

	lport = (argc == 4) ?  lport = atoi(argv[argc-1]) : TPL_ANY_PORT;

	if (argc < 3)
	{
		tpl_cmdHelpCmd("show");
		return 0;
	}

	if (strequ(argv[2], "bind"))
	{
		tpl_showBind( lport );
		
	}
	else if (strequ(argv[2], "cfg"))
	{
		tpl_showCfg( );	
	}
	else if (strequ(argv[2], "conn"))
	{
		tpl_showConn( lport );	
	}
	else if (strequ(argv[2], "pkt"))
	{
		tpl_showPkt( lport );	
	}
	else if (strequ(argv[2], "stats"))
	{
		tpl_showStats( );	
	}
#if defined(NFX_SMP)
	else if (strequ(argv[2], "ipc"))
		tpl_ipcShowQueues();
#endif /* NFX_SMP */
	else
		return (tpl_cmdHelpCmd("show"));

	return 0;
}

int32 tpl_cmdClear( int32 argc, char *argv[] )
{
	tpl_clearStats();
	return 0;
}

/*-----------------------------------------------------------------
 * Name: tpl_cmd
 *
 * Description: Main dispatch for TPL commands.  Called from 
 *				command table.
 *
 *				tpl loop <pkt len> <pkt count>
 *				tpl show stats
 *  
 * Created by : Shaun Hui
 *  
 * Date Created: 11/10/2001
 *-----------------------------------------------------------------
 */
int32 tpl_cmd( int32 argc, char *argv[] )
{
	int32	i;

	if (argc < 2)
	{
		return (tpl_cmdHelp(argc, argv));
	}

	for (i=0; i < TPL_CMD_MAX; i++)
	{
		if (strequ(argv[1], tpl_cmdTable[i].cmd))
		{
			return ((*tpl_cmdTable[i].cmdFunc)(argc, argv));
		}
	}
	return (tpl_cmdHelp(argc, argv));
}

