/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-rcv.c
 *
 * RCS:  $Id: tpl-rcv.c,v 1.29 2002/10/24 17:58:00 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer receive related routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 07/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>

#include <exports/pkt-api.h>
#include <exports/luc-api.h>
#include <exports/tpl-api.h>
#include <exports/bsdnet-api.h>
#include "tpl.h"
#include "../sm-bsd-api/bsd-vif.h"

//#define TPL_DBG_RPC_RASM	1

/*-------------------------------
 * Shaun 6/26/02
 * Increasd pullup size from 256
 * to 512. Fixed bug #1281.
 *-------------------------------
 */
#define TPL_RPC_PULLUP_SIZE		512

static int32
tpl_rcvRpcPkt(tpl_connCb_t *pConn, pkt_desc_t *pDesc );

/*-----------------------------------------------------------------
 * Name:  		tpl_rcvPkt
 *
 * Description: Receive function to be called by lower layer (LUC) 
 *				for receiving a "fast-path" packet. 
 *				pHdr should point beyound UDP/TCP header to RPC header.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
int32 		/* TPL_FREE_PKT or TPL_HOLD_PKT */
tpl_rcvPkt( void 		*pConnId, 	/* TPL Connection CB */
			pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_bindCb_t	*pBind;
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;

#ifdef PKT_USE_TIME_STAMP
	pkt_setTimeStamp(pDesc->pTime, PKT_TIME_IDX_TPL_RX);
#endif

	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;

	/*-------------------------------
	 * Make sure connCb is still vaild.
	 *-------------------------------
	 */
	if (!pConn || ((pConn->flags & TPL_CONN_INUSE) == 0))
	{
		pStats->cntBadConnErr++;
		return (TPL_FREE_PKT);
	}

	/*-------------------------------
	 * Make sure bindCb is still vaild.
	 *-------------------------------
	 */
	pBind = pConn->pBindCb;
	if ((!pBind) || (!pBind->ul_rcvPktInd))
	{
		pStats->cntBadBindErr++;
		return (TPL_FREE_PKT);
	}

	/* su, 06/27.
	 * if packet from slow path no need
     * for chksum.
	 */
	if (!(pDesc->flags & PKT_FAST_PATH))
		goto pkt_rcv;

	/* su, 07/12/02, hardware support chksum */
	/* in prom, traffic come from bckplane
	 */
#if !defined(PROM)
	if (isCheetah())
		goto pkt_rcv;
#endif

	/*-------------------------------
	 * Check IP header checksum.
	 *-------------------------------
	 */
	if (tpl_cksumIp(pConn, pDesc) != 0)
	{
		pStats->cntIpCsumErr++;
		return (TPL_FREE_PKT);
	}
	
	/*-------------------------------
	 * Check TCP/UDP checksum.
	 *-------------------------------
	 */
	/* su, 06/11/02, chk for PKT_IP_FRAG_FISRT
	 * if set, pdesc is start of seqence of 
	 * fragments. Cksum is already done.
	 */
	if (!(pDesc->flags &  PKT_IP_FIRST_FRAG)) {
		if (tpl_cksumTpl(pConn, pDesc) != 0)
		{
				if (pConn->prot == TPL_PROT_UDP) 
					pStats->cntUdpCsumErr++;
				else
					pStats->cntTcpCsumErr++;
				return (TPL_FREE_PKT);
		}
	}
	

pkt_rcv:
	pStats->cntRcvPkt++;

	/*-------------------------------
	 * Save pConn to be used later
	 * by application.
	 *-------------------------------
	 */
	pDesc->pTplConn = pConn;

#if defined(LINK_AGG)
	/* chk to see if input link has changed */
	if (pConn->linkNum != pDesc->link) {
		pConn->linkNum = pDesc->link;
		memcpy(&pConn->srcMac[0], 
			   luc_getMacAddr(pDesc->link), 6);
		bsd_setLinkId(pConn->pInpCb, pDesc->link);
		pStats->cntConnLinkChge ++;		
	}
#endif /* LINK_AGG */

	/*-------------------------------
	 * Handle TCP packets.
	 *-------------------------------
	 */
	if (pConn->prot == TPL_PROT_TCP)
	{
	  
		if (pDesc->flags & PKT_FAST_PATH)
		{
			int			ret;
			pkt_desc_t	*retp;

			/*-------------------------------
			 * Call tcp_inputC to update 
			 * tcp ack, seq & state etc.
			 *-------------------------------
			 */
			ret = tcp_inputC(pDesc, (int)(pDesc->pTplHdr - pDesc->pIpHdr), 
							 pConn, pConn->pInpCb, (pDesc->pIpHdr),
							 &retp, 0);
			
			switch (ret)
			{
				case	BSD_FAST_ERR:			/* Rcv err pkt */
				case	BSD_FAST_ACK:			/* Rcv TCP Ack */
						return TPL_FREE_PKT;

				case	BSD_FAST_QUEUED:		/* Out of Seq Seg */
						return TPL_HOLD_PKT;

				default: 						/* BSD_FAST_FWD fallthrough */
						pDesc = retp;
						break;
			}

		}

		/*-------------------------------
		 * Handle RPC marker & reassembly.
		 *-------------------------------
		 */
		if (pConn->flags & TPL_CONN_USE_RPC)
		{
            pkt_desc_t  *tmpDesc;

			/*-------------------------------
			 * Shaun 06/11/02 Bug #1095
			 *
			 * FreeBSD client sends a pkt with
			 * BAD TCP checksum which causes
			 * re-transmission. As a result,
			 * tcp_inputC gives us a PKT CHAIN 
			 * instead of individual pkt.
			 * Old code did not handle PKT CHAIN
			 * which cause usr data to be "corrupted".
			 *-------------------------------
			 */
            while( pDesc )
            {

				/*-------------------------------
				 * Save next pkt and removed it
				 * from pDesc.
				 *-------------------------------
				 */
                tmpDesc = pDesc->pNext;
                pDesc->pNext = NULL;

				/*-------------------------------
				 * Perfrom RPC receive.
				 *-------------------------------
				 */
    			if (tpl_rcvRpcPkt(pConn, pDesc) == TPL_FREE_PKT)
                {
				    PKT_FREE_PACKET( pDesc );
                }

                pDesc = tmpDesc;
            }

   			pStats->cntRasmDepth  = 0;
			return TPL_HOLD_PKT;
		}

	}

	
	/*-------------------------------
	 * Upcall app. for rcv pkt.
	 *-------------------------------
	 */
#if defined(NFX_SMP)
	{
		tpl_ipcq_t	*pktRxQueuePtr = &(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_RX]);
		
		if (TPL_IPCQ_FULL(pktRxQueuePtr))
			panic("%s: ipc pkt rx queue full\n", __FUNCTION__);
		
		TPL_IPCQ_WRITE(pktRxQueuePtr, pDesc);

		return TPL_HOLD_PKT;
	}
#else
	return ((pBind->ul_rcvPktInd)(pConn->pAppHdl, pDesc ));
#endif /* NFX_SMP */
	
}

/*-----------------------------------------------------------------
 * Name:        tpl_bsdRcvInd
 *
 * Description: Receive function to be called by BSD layer 
 *              for receiving a "slow-path" packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
void
tpl_bsdRcvInd(  void        *sock,
                void        *pDesc,
                uint32      saddr,  
                ushort16    sport, 
                uint32      len,
                void        *pConn )
{
	/*-------------------------------
	 * A tpl_addConnInd should have 
	 * been done before this point.
	 *
	 * Upcall tpl_rcvPkt to handle pkt.
	 * TPL binding to BSD specifies 
	 * BSD_BUF_DESCR. 
	 *-------------------------------
	 */
    if (tpl_rcvPkt( pConn, pDesc ) == PKT_FREE_PKT)
	{
		PKT_FREE_PACKET( pDesc );
	}
    return;
}

static void 
tpl_flushRpcRasm( tpl_connCb_t *pConn )
{
	/*-------------------------------
	 * Free pkt in reassem chain. 
	 *-------------------------------
	 */
	if (pConn->rpcRasmList)
		PKT_FREE_PACKET(pConn->rpcRasmList);
					
	/*-------------------------------
	 * Clear reassem list & state.
	 *-------------------------------
	 */
	pConn->rpcRasmList = NULL;
	pConn->rpcRasmCurr = NULL;
	pConn->rpcRasmLen  = 0;
	pConn->rpcRasmCnt  = 0;
	pConn->rpcPullUp   = 0;

	pConn->flags &= ~TPL_CONN_RPC_REASM;

}


static pkt_desc_t	*
tpl_splitPkt( pkt_desc_t *pDesc ,int32 rpcLen)
{
    tpl_stats_t 	*pStats;
    eee_descBuf_t   *bd;
    eee_desc_t      *eDescp;
	pkt_desc_t		*newPkt;
	int32			hdrSize;
	int32			remain;


#ifdef TPL_DBG_RPC_RASM
    printf("\n-----------------------------------------\n");
    printf("%s: pDesc=%p hdrSize=%d rpcLen=%d\n",
            __FUNCTION__, pDesc, pDesc->hdrSize, rpcLen);
#endif

    pStats  = &pTplCfg->tplStats;
	hdrSize = pDesc->hdrSize;
	remain  = rpcLen;
		
	pDesc->hdrSize = remain;

	/*-------------------------------
	 * Pkt has remaining RPC data &
	 * start of new RPC.
	 * Split pkt into 2:
	 * old pDesc has remaining data,
	 * newPkt has start of new RPC.
	 *-------------------------------
	 */
    newPkt = PKT_ALLOC_PACKET(1);
    if (!newPkt)
	{
		/*-------------------------------
		 * No more pkt resource.
		 *-------------------------------
		 */
        pStats->cntNoPktRsrc++;
		return NULL;
	}


	/*-------------------------------
	 * Copy new RPC data.
	 *-------------------------------
	 */
    memcpy(newPkt->pHdr, pDesc->pHdr+remain, hdrSize-remain);

	/*-------------------------------
	 * Copy desc info from old pkt.
	 *-------------------------------
	 */
    newPkt->laddr   = pDesc->laddr;
    newPkt->raddr   = pDesc->raddr;
    newPkt->pTplConn= pDesc->pTplConn;
    newPkt->flags   = pDesc->flags;
    newPkt->hdrSize = (hdrSize-remain);
    newPkt->pktSize = (hdrSize-remain);
    newPkt->pMac    = newPkt->pHdr;

	/*-------------------------------
	 * Setup eee pkt size & bd len
	 * for new pkt.
	 *-------------------------------
	 */
    eDescp = newPkt->eDescp;
	bd     = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
           	 &(eDescp->bd[0]) : &(eDescp->bd[1]);
	bd->len = hdrSize-remain;

    eDescp->hdr.control = (newPkt->pktSize & EDESC_TOTAL_LEN) |
                          (( 1 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);


	/*-------------------------------
	 * Shaun 08/30/02 bug #1983
	 * Need to adjust pktSize of 
	 * remaining data after split
	 * on old pkt.
	 *-------------------------------
	 */
	pDesc->pktSize -= (hdrSize-remain);

    eDescp = pDesc->eDescp;
	bd     = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
           	 &(eDescp->bd[0]) : &(eDescp->bd[1]);
	bd->len = pDesc->pktSize;

    eDescp->hdr.control = (pDesc->pktSize & EDESC_TOTAL_LEN) |
                          (( 1 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);


#ifdef TPL_DBG_RPC_RASM
	printf("%s: Split: old pDesc=%p pktSize=%d hdrSize=%d\n",
		__FUNCTION__, pDesc, pDesc->pktSize, pDesc->hdrSize );

	printf("%s: Split: new pDesc=%p pktSize=%d hdrSize=%d\n",
		__FUNCTION__, newPkt, newPkt->pktSize, newPkt->hdrSize);
#endif

	/*-------------------------------
	 * Return new Pkt for start of RPC.
	 *-------------------------------
	 */
	return newPkt;
}

/*-----------------------------------------------------------------
 * Name:  		tpl_startRpcRasm
 *
 * Description: Start RPC reassembly
 *
 * Created by : Shaun Hui
 *
 *-----------------------------------------------------------------
 */
static int32 
tpl_startRpcRasm(tpl_connCb_t 	*pConn, 
				 pkt_desc_t 	*pDesc ,
				 pkt_desc_t 	**newDesc )
{

	tpl_stats_t		*pStats;
	pkt_desc_t		*newPkt;
	tpl_bindCb_t	*pBind;
	int32			rc;
	uint32 			rpcLen;
	uint32 			rpcMarker;

	pStats = &pTplCfg->tplStats;
	pBind  = pConn->pBindCb;

	/*-------------------------------
	 * Shaun 10/05/02 #2310
	 * Drop pkt with start of RPC on
	 * non word-aligned boundary.
	 *-------------------------------
	 */
	if ((address_t)(pDesc->pHdr) & 0x03)
	{
		pTplCfg->tplStats.cntRpcRasmErr++;
		PKT_FREE_PACKET( pDesc );

		return TPL_HOLD_PKT;
	}

    /*-------------------------------
     * Get RPC len from marker.
     *-------------------------------
     */
    rpcMarker = *(uint32 *)pDesc->pHdr;
    rpcMarker = ntohl(rpcMarker);
    rpcLen    = rpcMarker & 0x7FFFFFFF;

    /*-------------------------------
     * Last RPC fragment bit set.
     *-------------------------------
     */
    if ((rpcMarker & 0xFFF00000) != 0x80000000)
    {
		/*-------------------------------
		 * Not in reassem., incr. err cnt,
		 * free fragments.
		 *-------------------------------
		 */
		pTplCfg->tplStats.cntRpcRasmErr++;

		PKT_FREE_PACKET( pDesc );

		return TPL_HOLD_PKT;
	}


#ifdef TPL_DBG_RPC_RASM
    printf("\n-----------------------------------------\n");
    printf("%s: pDesc=%p hdrSize=%d rpcLen=%d\n",
            __FUNCTION__, pDesc, pDesc->hdrSize, rpcLen);
#endif


	/*-------------------------------
	 * If RPC len > 40K, drop pkt.
	 *-------------------------------
	 */
	if (rpcLen > (40 * 1024))
	{
		pStats->cntRpcLenErr++;

		PKT_FREE_PACKET( pDesc );

		return TPL_HOLD_PKT;
	}

	/*-------------------------------
	 * Count max. depth of startRpc
	 *-------------------------------
	 */
	pStats->cntRasmDepth++;
	if (pStats->cntRasmDepth > pStats->cntMaxRasmDepth)
		pStats->cntMaxRasmDepth = pStats->cntRasmDepth;

	/*-------------------------------
	 * Exclude RPC Marker from hdr.
	 *-------------------------------
	 */
	pDesc->pHdr    += 4;
	pDesc->hdrSize -= 4;

	/*-------------------------------
	 * Single fragment, no reassembly.
	 *-------------------------------
	 */
	if (rpcLen < pDesc->hdrSize)
	{

		/*-------------------------------
		 * Pkt has remaining RPC data &
		 * start of new RPC.
		 * Split pkt into 2:
		 * old pDesc has remaining data,
		 * newPkt has start of new RPC.
		 *-------------------------------
		 */
        newPkt = tpl_splitPkt(pDesc, rpcLen );
        if (!newPkt)
		{
			/*-------------------------------
			 * No more pkt resource.
			 * Flush RPC Reassembly list.
			 *-------------------------------
			 */
           	tpl_flushRpcRasm(pConn);
			return TPL_HOLD_PKT;
		}

		pStats->cntRpcSplit1++;

		/*-------------------------------
		 * Return new Pkt for start of RPC.
		 *-------------------------------
		 */
		*newDesc = newPkt;

		/*-------------------------------
		 * Upcall app. for rcv pkt.
		 *-------------------------------
		 */
		pStats->cntRpcUpcall1++;
#if defined(NFX_SMP)
	{
		tpl_ipcq_t	*pktRxQueuePtr = &(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_RX]);

		
		if (TPL_IPCQ_FULL(pktRxQueuePtr))
			panic("%s: ipc pkt rx queue full\n", __FUNCTION__);
		
		TPL_IPCQ_WRITE(pktRxQueuePtr, pDesc);

		rc =  TPL_HOLD_PKT;
	}
#else
		rc =  (pBind->ul_rcvPktInd)(pConn->pAppHdl, pDesc );
#endif /* NFX_SMP */

		if (rc == TPL_FREE_PKT)
			PKT_FREE_PACKET( pDesc );


		return TPL_HOLD_PKT;
	}
	else if (rpcLen > pDesc->hdrSize )
	{

		/*-------------------------------
		 * More pkts needed. Conn. is in
		 * reassem. state.
		 *-------------------------------
		 */
		pConn->flags |= TPL_CONN_RPC_REASM;

		/*-------------------------------
		 * Reassemble pkt fragments in list.
		 *-------------------------------
		 */
		pConn->rpcRasmList = pDesc;
		pConn->rpcRasmCurr = pDesc;
		pConn->rpcRasmLen  = rpcLen - pDesc->hdrSize ;

		/*-------------------------------
		 * Shaun 6/26/02
		 * Increasd pullup size from 256
		 * to 512. Fixed bug #1281.
		 *-------------------------------
		 */
		pConn->rpcPullUp   = (pDesc->hdrSize < TPL_RPC_PULLUP_SIZE) ? 1: 0 ;

		pConn->rpcRasmCnt++;

		*newDesc = NULL;

		if (pConn->rpcRasmCnt > pStats->cntMaxRpcDepth)
			pStats->cntMaxRpcDepth = pConn->rpcRasmCnt;

#ifdef TPL_DBG_RPC_RASM
		printf("%s: Queue: old pDesc=%p pktSize=%d hdrSize=%d\n",
                __FUNCTION__, pDesc, pDesc->pktSize, pDesc->hdrSize );


		printf("**** %s: List=%x Curr=%x Len=%d hdrSize=%d\n", 
				__FUNCTION__,
				pConn->rpcRasmList, 
				pConn->rpcRasmCurr, 
				pConn->rpcRasmLen,
				pDesc->hdrSize);
#endif


		return TPL_HOLD_PKT;
	}
	else if (rpcLen == pDesc->hdrSize)
	{

#ifdef TPL_DBG_RPC_RASM
    	printf("%s: start2: old pDesc=%p pktSize=%d hdrSize=%d\n",
                __FUNCTION__, pDesc, pDesc->pktSize, pDesc->hdrSize );
#endif

		/*-------------------------------
		 * Upcall app. for rcv pkt.
		 *-------------------------------
		 */
		pStats->cntRpcUpcall2++;
#if defined(NFX_SMP)
	{
		tpl_ipcq_t	*pktRxQueuePtr = &(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_RX]);

		if (TPL_IPCQ_FULL(pktRxQueuePtr))
			panic("%s: ipc pkt rx queue full\n", __FUNCTION__);
		
		TPL_IPCQ_WRITE(pktRxQueuePtr, pDesc);

		rc =  TPL_HOLD_PKT;
	}
#else
		rc =  (pBind->ul_rcvPktInd)(pConn->pAppHdl, pDesc );
#endif /* NFX_SMP */
		if (rc == TPL_FREE_PKT)
			PKT_FREE_PACKET( pDesc );

		*newDesc = NULL;

		return TPL_HOLD_PKT;
	}

	return TPL_HOLD_PKT;
}


/*-----------------------------------------------------------------
 * Name:  		tpl_rpcQueueRasm
 *
 * Description: Queue PKT on RPC reassembly queue.
 *
 * Created by : Shaun Hui
 *
 *-----------------------------------------------------------------
 */
static int32
tpl_rpcQueueRasm(tpl_connCb_t 	*pConn, 
				 pkt_desc_t 	*pDesc, 
				 pkt_desc_t 	**newDesc)
{
	tpl_bindCb_t	*pBind;
	tpl_stats_t		*pStats;
	pkt_desc_t		*newPkt;
	int32			freePkt;
	int32			remain;
	int32			hdrSize;
	int32			hdrOffset;
	int32			rc;

	*newDesc = NULL;

	hdrOffset = 0;
	freePkt = 0;
	remain  = pConn->rpcRasmLen;
	hdrSize = pDesc->hdrSize;
	pStats  = &pTplCfg->tplStats;

#ifdef TPL_DBG_RPC_RASM
    printf("\n-----------------------------------------\n");
    printf("%s: pDesc=%p hdrSize=%d rpcLen=%d\n",
            __FUNCTION__, pDesc, pDesc->hdrSize, remain);
#endif

	/*-------------------------------
	 * Check if we need to copy part
	 * or all of next packet.
	 *-------------------------------
	 */
	if (pConn->rpcPullUp)
	{
	    eee_descBuf_t   *bd;
	    eee_desc_t      *eDescp;
		pkt_desc_t 		*dstDesc;
		int32			copyLen;

		/*-------------------------------
		 * RasmCurr has prev. pkt desc.
		 * that needed to have more rpc data
		 * copied (pull-up).
		 *-------------------------------
		 */
		dstDesc = pConn->rpcRasmCurr;

		/*-------------------------------
		 * Copy remaining rpc data if
		 * within this pkt, else copy
		 * up to PULLUP_SIZE.
		 *-------------------------------
		 */
		if (remain <= hdrSize )
			copyLen = remain; 
		else
			copyLen = TPL_RPC_PULLUP_SIZE - dstDesc->hdrSize;

		pStats->cntRpcPullup++;

		/*-------------------------------
		 * Don't copy more than what's in pkt.
		 *-------------------------------
		 */
		if (copyLen > pDesc->hdrSize) 
		{
			copyLen = pDesc->hdrSize;
		} 

		/*-------------------------------
		 * Copy RPC data.
		 *-------------------------------
		 */
		memcpy(dstDesc->pHdr+dstDesc->hdrSize, pDesc->pHdr, copyLen);

		/*-------------------------------
		 * Adjust hdr & pkt size by copylen.
		 *-------------------------------
		 */
		dstDesc->hdrSize += copyLen;
		dstDesc->pktSize += copyLen;

		/*-------------------------------
		 * Adjust eee desc pkt & bd len.
		 *-------------------------------
		 */
        eDescp = dstDesc->eDescp;
        bd     = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
                     &(eDescp->bd[0]) : &(eDescp->bd[1]);
        bd->len += copyLen;
        eDescp->hdr.control = (dstDesc->pktSize & EDESC_TOTAL_LEN) |
                            (( 1 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);


		/*-------------------------------
		 * Su, 05/01/02. 
		 * need to take mac/ip/tcp hdr into 
		 * account when move partial buffer
		 *-------------------------------
		 */
		hdrOffset = (address_t)pDesc->pHdr - (address_t)pDesc->pMac;

		/*-------------------------------
		 * Adjust mac & hdr ptr to exclude 
		 * copied data.
		 *-------------------------------
		 */
		pDesc->pHdr += copyLen;
		pDesc->pMac  = pDesc->pHdr;

		/*-------------------------------
		 * Adjust hdr & pkt size to exclude 
		 * copied data & hdr offset.
		 *-------------------------------
		 */
		pDesc->hdrSize -= copyLen;
        pDesc->pktSize -= copyLen + hdrOffset;

		/*-------------------------------
		 * Free pkt if all of it is copied.
		 *-------------------------------
		 */
		if (pDesc->hdrSize <= 0)
		{

#ifdef TPL_DBG_RPC_RASM
printf("**** %s PullUp  Free Desc=%x \n", __FUNCTION__, pDesc);
#endif
			PKT_FREE_PACKET(pDesc);
		
			/*-------------------------------
			 * Use prev pkt as curr pkt.
			 *-------------------------------
			 */
			pDesc   = dstDesc;
			hdrSize = dstDesc->hdrSize;
			freePkt = 1;

			*newDesc = NULL;
		}
		else
		{
			/*-------------------------------
			 * Adjust eee desc pkt & bd len.
			 * for pulled-up pkt.
			 *-------------------------------
			 */
	        eDescp = pDesc->eDescp;
	        bd     = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
                     &(eDescp->bd[0]) : &(eDescp->bd[1]);
	        bd->len -= (copyLen + hdrOffset);
	        eDescp->hdr.control = (pDesc->pktSize & EDESC_TOTAL_LEN) |
                            (( 1 << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

#ifdef TPL_DBG_RPC_RASM
printf("**** %s PullUp 2: dst: Desc=%x hdrSize=%d\n", 
			__FUNCTION__,
			dstDesc, dstDesc->hdrSize);
printf("**** %s PullUp 2: src: Desc=%x buf=%x hdrSize=%d copyLen=%d\n",
			__FUNCTION__,
			pDesc, pDesc->pHdr, pDesc->hdrSize, copyLen);
#endif

			hdrSize -= copyLen;

			*newDesc = NULL;
		}

#ifdef TPL_DBG_RPC_RASM
        printf("%s: PullUp: new pDesc=%p pktSize=%d hdrSize=%d\n",
                __FUNCTION__, pDesc, pDesc->pktSize, pDesc->hdrSize );
#endif


		pConn->rpcPullUp = 0;
		pConn->rpcRasmLen -= copyLen;

		if (pConn->rpcRasmLen <= 0)
		{
#ifdef TPL_DBG_RPC_RASM
    		printf("%s: Upcall Ind 1: pDesc=%p hdrSize=%d rpcLen=%d\n",
		            __FUNCTION__, 
					pConn->rpcRasmList, 
					pConn->rpcRasmList->hdrSize, 
					pConn->rpcRasmLen);
#endif

	        /*-------------------------------
	         * Upcall app. for rcv pkt.
	         *-------------------------------
	         */
			
			pStats->cntRpcUpcall3++;
	        pBind = pConn->pBindCb;
#if defined(NFX_SMP)
	{
		tpl_ipcq_t	*pktRxQueuePtr = &(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_RX]);
		
		if (TPL_IPCQ_FULL(pktRxQueuePtr))
			panic("%s: ipc pkt rx queue full\n", __FUNCTION__);
		
		TPL_IPCQ_WRITE(pktRxQueuePtr, pConn->rpcRasmList);

		rc = TPL_HOLD_PKT;
	}
#else
	        rc = (pBind->ul_rcvPktInd)(pConn->pAppHdl, pConn->rpcRasmList);
#endif /* NFX_SMP */

	        if (rc != TPL_FREE_PKT)
	            pConn->rpcRasmList = NULL;

	        /*-------------------------------
	         * Flush Rasm chain.
	         *-------------------------------
	         */
			tpl_flushRpcRasm(pConn);

			
	        /*-------------------------------
	         * Start RPC reassembly with 
			 * remaining pkt.
	         *-------------------------------
	         */
			if (!freePkt)
				*newDesc = pDesc;

			return TPL_HOLD_PKT;
		}
	}


	if (!freePkt)
	{
		/*-------------------------------
		 * Chain pkt fragment in curr ptr.
		 *-------------------------------
		 */
		pConn->rpcRasmCurr->pNext = pDesc;
		pConn->rpcRasmCurr		 = pDesc;

		/*-------------------------------
		 * Reduce reassem. len by amount of
		 * data rcvd. which has no RPC marker
		 *-------------------------------
		 */
		pConn->rpcRasmLen -= pDesc->hdrSize; 

		pConn->rpcRasmCnt++;

		/*-------------------------------
		 * Count max. # pkts in RPC q.
		 *-------------------------------
		 */
		if (pConn->rpcRasmCnt > pStats->cntMaxRpcDepth)
			pStats->cntMaxRpcDepth = pConn->rpcRasmCnt;
	}

	/*-------------------------------
	 * Reassembly done if no more 
	 * data remaining.
	 *-------------------------------
	 */
	if (pConn->rpcRasmLen <= 0)
	{

        if (remain < hdrSize)
        {

			/*-------------------------------
			 * Curr pkt has data remaining & 
			 * new RPC.  Need to split pkt.
			 * old pDesc has remaining data,
			 * newPkt has start of new RPC.
			 *-------------------------------
			 */
            newPkt = tpl_splitPkt(pDesc, remain );
            if (!newPkt)
			{
				/*-------------------------------
				 * No more pkt resource.
				 * Flush RPC Reassembly list.
				 *-------------------------------
				 */
            	tpl_flushRpcRasm(pConn);
				return TPL_HOLD_PKT;
			}

			pStats->cntRpcSplit2++;

			*newDesc = newPkt;

        }

#ifdef TPL_DBG_RPC_RASM
   		printf("%s: Upcall Ind 2: pDesc=%p hdrSize=%d rpcLen=%d\n",
	            __FUNCTION__, 
				pConn->rpcRasmList, 
				pConn->rpcRasmList->hdrSize, 
				pConn->rpcRasmLen);
#endif

        /*-------------------------------
         * Upcall app. for rcv pkt.
         *-------------------------------
         */
		pStats->cntRpcUpcall4++;
        pBind = pConn->pBindCb;
#if defined(NFX_SMP)
   {
		tpl_ipcq_t	*pktRxQueuePtr = &(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_RX]);
		
		if (TPL_IPCQ_FULL(pktRxQueuePtr))
			panic("%s: ipc pkt rx queue full\n", __FUNCTION__);
		
		TPL_IPCQ_WRITE(pktRxQueuePtr, pConn->rpcRasmList);

		rc = TPL_HOLD_PKT;
   }
#else
        rc = (pBind->ul_rcvPktInd)(pConn->pAppHdl, pConn->rpcRasmList);
#endif /* NFX_SMP */

		/*-------------------------------
		 * Shaun 5/9/02
		 * Bug & fix found by Pankaj.
		 *
		 * Flush the RPC RASM queue since we have handed
		 * the packet to application. If application does
		 * not want the packet freed then set rpcRasmList
		 * to NULL before flush to avoid duplicate free
		 *-------------------------------
		 */
        if (rc != TPL_FREE_PKT)
            pConn->rpcRasmList = NULL;

		tpl_flushRpcRasm(pConn);
	}

	/*-------------------------------
	 * Don't free pkt in reassem chain. 
	 * Free by upcall or flush later.
	 *-------------------------------
	 */
	return TPL_HOLD_PKT;
}

/*-----------------------------------------------------------------
 * Name:  		tpl_rcvRpcPkt
 *
 * Description: Receive function to handle RPC packet.
 *
 * Created by : Shaun Hui
 *
 *-----------------------------------------------------------------
 */
static int32
tpl_rcvRpcPkt(tpl_connCb_t *pConn, pkt_desc_t *pDesc )
{
	pkt_desc_t	*newPkt;
	int32		rc;

	while( pDesc ) 
	{
		newPkt = NULL;

		/*-------------------------------
		 * Conn. not in RPC reassem. state.
		 *-------------------------------
		 */
		if (!(pConn->flags & TPL_CONN_RPC_REASM))
		{
			rc = tpl_startRpcRasm(pConn, pDesc, &newPkt );
		}
		else 	
		{

			/*-------------------------------
		 	 * Conn. in RPC reassem. state.
			 * Do RPC reassembly.
			 *-------------------------------
			 */
			rc = tpl_rpcQueueRasm(pConn, pDesc, &newPkt);
		}

		pDesc = newPkt;

#ifdef TPL_DBG_RPC_RASM
        if (pDesc)
            printf("%s: rc=%d pDesc=%p hdrSize=%d rpcLen=%d\n",
                __FUNCTION__, rc, pDesc, pDesc->hdrSize, rpcLen);
#endif

	}

	return TPL_HOLD_PKT;
}


