/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-utils.c
 *
 * RCS:  $Id: tpl-utils.c,v 1.27 2002/10/29 01:21:34 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer utility routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 08/20/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/pkt-api.h>
#include <exports/tpl-api.h>
#include "tpl.h"
#include <exports/bsd-api.h>

/*-------------------------------
 * TPL error control block.
 *-------------------------------
 */
typedef struct _tpl_errCb_t
{
	char	*errMsg;
	int32	errCode;
	int32	action;
} tpl_errCb_t;

#define TPL_DO_PANIC	1
#define TPL_DO_NOTHING	0

#define TPL_CKSUM_COMPLETE  0
#define TPL_CKSUM_PARTIAL   1

static tpl_errCb_t errCb[ ]=
{
	{"No Shared Memory Resource",	TPL_ERR_NO_SHMEM_RSRC,	TPL_DO_PANIC},
	{"No Bind Resource",			TPL_ERR_NO_BIND_RSRC,	TPL_DO_PANIC},
	{"No CAM Resource",				TPL_ERR_NO_CAM_RSRC,	TPL_DO_NOTHING},
	{"No Pkt Resource",				TPL_ERR_NO_PKT_RSRC,	TPL_DO_PANIC},
	{"No Resource",					TPL_ERR_NO_RSRC,		TPL_DO_PANIC},
	{"Duplicate Connection",		TPL_ERR_DUP_CONN,		TPL_DO_PANIC},
	{"Bad Handle",					TPL_ERR_BAD_HDL,		TPL_DO_PANIC},
	{"Invalid  Protocol",			TPL_ERR_BAD_PROT,		TPL_DO_PANIC},
	{"No Indication Function",		TPL_ERR_BAD_FUNC,		TPL_DO_PANIC},
	{"Invalid Bind CB",				TPL_ERR_BAD_BIND,		TPL_DO_PANIC},
	{"Invalid Conn CB",				TPL_ERR_BAD_CONN,		TPL_DO_NOTHING},
	{"Invalid BSD Event",			TPL_ERR_BAD_EVT,		TPL_DO_PANIC},
	{"Invalid Pkt",					TPL_ERR_BAD_PKT,		TPL_DO_PANIC},
	{"Duplicate Bind",				TPL_ERR_DUP_BIND,		TPL_DO_PANIC},
	{"No Binding found",			TPL_ERR_NOT_BIND,		TPL_DO_PANIC},
	{"Upper layer add failed",		TPL_ERR_UL_ADD_FAIL,	TPL_DO_PANIC},
	{"Upper layer delet failed",	TPL_ERR_UL_DEL_FAIL,	TPL_DO_PANIC},
	{"Free PKT ring full",			TPL_ERR_RING_FULL,		TPL_DO_PANIC},
	{"Failed to send",				TPL_ERR_XMT_FAIL,		TPL_DO_PANIC},
	{"Missing local address",		TPL_ERR_NO_LADDR,		TPL_DO_PANIC},
	{"Missing remote address",		TPL_ERR_NO_RADDR,		TPL_DO_PANIC},
	{"Missing local port",			TPL_ERR_NO_LPORT,		TPL_DO_PANIC},
	{"Missing remote port",			TPL_ERR_NO_RPORT,		TPL_DO_PANIC},
	{"Remote address in server",	TPL_ERR_RADDR_IN_SRVR,	TPL_DO_PANIC},
	{"Remote port  in server",		TPL_ERR_RPORT_IN_SRVR,	TPL_DO_PANIC},
	{"Bad binding mode",			TPL_ERR_BAD_MODE,		TPL_DO_PANIC},
	{"Bad link ID ",				TPL_ERR_BAD_LINK,		TPL_DO_PANIC},
	{"Unknown Error",				TPL_ERR_MAX,			TPL_DO_PANIC},
};

/*-----------------------------------------------------------------
 * Name:  		tpl_logErr
 *
 * Description: Log TPL Errors. Panic if TPL_DEBUG is defined.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
void
tpl_logErr( char *fileName, int32 lineNum, int32 errCode, void *info )
{
	tpl_errCb_t	*pErrCb;
	int32	msgIdx;
	int32	maxIdx;

	/*-------------------------------
	 * Get msg Id relative to base.
	 *-------------------------------
	 */
    msgIdx = ((TPL_ERROR_BASE) - errCode);
    maxIdx = ((TPL_ERROR_BASE) - TPL_ERR_MAX);

	/*-------------------------------
	 * Make sure err code is in range.
	 *-------------------------------
	 */
	if ((msgIdx > maxIdx) || (msgIdx < 0))
		msgIdx = maxIdx;

	/*-------------------------------
	 * Get error CB.
	 *-------------------------------
	 */
	pErrCb = &errCb[msgIdx];

	if (pErrCb->action == TPL_DO_PANIC)
	{
    	printf("TPL Error (%d): %s! File: %s Line: %d Info:%lx\n",
	            errCode, pErrCb->errMsg, fileName, lineNum, (address_t)info);
   
    	TPL_PANIC()
	}
}
 		
/*-----------------------------------------------------------------
 * Name:        tpl_unbindPort
 *
 * Description: Removes bind control blocks of given local port. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 12/12/01
 *-----------------------------------------------------------------
 */
void tpl_unbindPort( ushort16 lport )
{
	int32			i;
	tpl_bindCb_t	*pBind;

    pBind = pTplCfg->bindTblBase;
    for (i=0; i < pTplCfg->maxBindCb; i++)
    {
        if (pBind->lport || pBind->rport)
		{
			if ((lport == ntohs(pBind->lport)) || 
				(lport == TPL_ANY_PORT))
				tpl_unbindReq(pBind);
		}
        pBind++;
    }
}

/*-----------------------------------------------------------------
 * Name:        tpl_showCfg
 *
 * Description: Show TPL config block contents.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
void 
tpl_showCfg( void )
{
    tpl_cfg_t *pCfg;

    pCfg = pTplCfg;

    printf("TPL CFG: %lx\n", (address_t)pCfg);
    printf("----------------------------------------------------\n");
    printf("\tBind CB Base= 0x%lx\n", (address_t)pCfg->bindTblBase);
    printf("\tConn CB Base= 0x%lx\n", (address_t)pCfg->connTblBase);
    printf("\tNum Bind CB = %8d\n", pCfg->numBindCb);
    printf("\tNum Conn CB = %8d\n", pCfg->numConnCb);
    printf("\tMax Bind CB = %8d\n", pCfg->maxBindCb);
    printf("\tMax Conn CB = %8d\n", pCfg->maxConnCb);
    printf("\tMax Conn Pkt= %8d\n", pCfg->maxPkt);
	printf("\tNum. Pend Q = %8d\n", pCfg->numPend);
	printf("\tPend Q start=%lx\n",  (address_t)pCfg->tplPend);
}

/*-----------------------------------------------------------------
 * Name:        tpl_showStats
 *
 * Description: Show TPL config block contents.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
void 
tpl_showStats( void )
{
    tpl_stats_t *pStats;

    pStats = &pTplCfg->tplStats;

    printf("TPL Stats: \n" );
    printf("----------------------------------------------------\n");
    printf("  Bad Bind Err:    %12u\n", pStats->cntBadBindErr);
    printf("  Bad Conn Err:    %12u\n", pStats->cntBadConnErr);
    printf("  Bad Pkt  Err:    %12u\n", pStats->cntBadPktErr);
    printf("  No  CAM Rsrc:    %12u\n", pStats->cntNoCamRsrc);
    printf("  No  Pkt Rsrc:    %12u\n", pStats->cntNoPktRsrc);
    printf("  Bind Req    :    %12u\n", pStats->cntBindReq);
    printf("  Unbind Req  :    %12u\n", pStats->cntUnbindReq);
    printf("  Add Conn Ind:    %12u\n", pStats->cntAddConnInd);
    printf("  Del Conn Ind:    %12u\n", pStats->cntDelConnInd);
    printf("  Del Conn Req:    %12u\n", pStats->cntDelConnReq);
    printf("  Alloc Pkt   :    %12u\n", pStats->cntAllocPkt);
	printf("  Alloc OBd   :    %12u\n", pStats->cntOutConnAlloc);
	printf("  HW  Outbd   :    %12u\n", pStats->cntOutHWater);
    printf("  Free  Pkt   :    %12u\n", pStats->cntFreePkt);
    printf("  Rcv   Pkt   :    %12u\n", pStats->cntRcvPkt);
    printf("  Xmt   Pkt   :    %12u\n", pStats->cntXmtPkt);
    printf("  Xmt   Cmp   :    %12u\n", pStats->cntXmtCmp);
    printf("  Xmt Up Call :    %12u\n", pStats->cntXmtUpCall);
	printf("  Xmt Dwn Call:    %12u\n", pStats->cntXmtDownCall);
    printf("  Ip csum Err :    %12u\n", pStats->cntIpCsumErr);
    printf("  Udp csum Err:    %12u\n", pStats->cntUdpCsumErr);
    printf("  Tcp csum Err:    %12u\n", pStats->cntTcpCsumErr);
    printf("  Ref Cnt     :    %12u\n", pStats->cntRefCnt);
    printf("  In Tcp TxQ  :    %12u\n", pStats->cntInTcpTxQ);
    printf("  In Tcp Xmt  :    %12u\n", pStats->cntInTcpXmt);
    printf("  RPC Len Err :    %12u\n", pStats->cntRpcLenErr);
    printf("  RPC Rasm Err:    %12u\n", pStats->cntRpcRasmErr);
    printf("  RPC Upcall1 :    %12u\n", pStats->cntRpcUpcall1);
    printf("  RPC Upcall2 :    %12u\n", pStats->cntRpcUpcall2);
    printf("  RPC Upcall3 :    %12u\n", pStats->cntRpcUpcall3);
    printf("  RPC Upcall4 :    %12u\n", pStats->cntRpcUpcall4);
    printf("  RPC Split1  :    %12u\n", pStats->cntRpcSplit1);
    printf("  RPC Split2  :    %12u\n", pStats->cntRpcSplit2);
    printf("  RPC Pullup  :    %12u\n", pStats->cntRpcPullup);
    printf("  RPC Max Cnt :    %12u\n", pStats->cntMaxRpcDepth);
    printf("  RPC Max Rasm:    %12u\n", pStats->cntMaxRasmDepth);
	printf("  MAC Pkt Drop:	   %12u\n", pStats->cntXmtNoMacErr);
	printf("  MAC Queue   :	   %12u\n", pStats->cntXmtMacQueue);
	printf("  MAC Uncalled:	   %12u\n", pStats->cntXmtMacUncall);
	printf("  MAC Pkt Drop:	   %12u\n", pStats->cntXmtNoMacErr);
	printf("  Dup.Pkt Err :	   %12u\n", pStats->cntXmtDup);
	printf("  Bad Pkt Tpl :	   %12u\n", pStats->cntXmtNoTplUsr);
	printf("  Dup.Tcp Err :	   %12u\n", pStats->cntXmtTcpDup);
	printf("  Tcp pkt Err :	   %12u\n", pStats->cntXmtTcpErr);
	printf("  Tcp Queue   :	   %12u\n", pStats->cntXmtTcpQueue);
	printf("  Conn. lnkChg:    %12u\n", pStats->cntConnLinkChge);
	printf("  pkt lnkChg  :    %12u\n", pStats->cntPktLinkChge);
}

/*-----------------------------------------------------------------
 * Name:        tpl_showBindCb
 *
 * Description: Show given bind control block contents.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
static void 
tpl_showBindCb( tpl_bindCb_t *pBind )
{
    printf("\nBind CB=0x%lx protocol:%d mode: %s\n", (address_t)pBind, pBind->tplProt,
			((pBind->bindFlags & TPL_FLAG_CLIENT) ? "Client" : "Server") );
    printf("----------------------------------------------\n");
    printf("laddr=0x%x lport=%d\n", ntohl(pBind->laddr), ntohs(pBind->lport));
    printf("raddr=0x%x rport=%d\n", ntohl(pBind->raddr), ntohs(pBind->rport));
    printf("usrInfo       = 0x%lx\n", (address_t)pBind->usrInfo);
    printf("pConnList     = 0x%lx\t", (address_t)pBind->pConnList);
    printf("num of Conn   = %d\n",   pBind->numConnCb);
    printf("pBsdSock      = 0x%lx\n", (address_t)pBind->pBsdSock);
    printf("ul_addConnInd = 0x%lx\n", (address_t)pBind->ul_addConnInd);
    printf("ul_delConnInd = 0x%lx\n", (address_t)pBind->ul_delConnInd);
    printf("ul_rcvPktInd  = 0x%lx\n", (address_t)pBind->ul_rcvPktInd);
    printf("ul_xmtCmpInd  = 0x%lx\n", (address_t)pBind->ul_xmtCmpInd);

}

/*-----------------------------------------------------------------
 * Name:        tpl_showBind
 *
 * Description: Show all bind control blocks of given local port. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 12/12/01
 *-----------------------------------------------------------------
 */
void tpl_showBind( ushort16 lport )
{
	int32			i;
	tpl_bindCb_t	*pBind;

    pBind = pTplCfg->bindTblBase;
    for (i=0; i < pTplCfg->maxBindCb; i++)
    {
        if (pBind->lport || pBind->rport)
		{
			if ((lport == ntohs(pBind->lport)) || 
				(lport == TPL_ANY_PORT))
				tpl_showBindCb(pBind);
		}
        pBind++;
    }
}

/*-----------------------------------------------------------------
 * Name:        tpl_showConnCb
 *
 * Description: Show given connection control block contents.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
static void 
tpl_showConnCb( tpl_connCb_t *pConn )
{

    printf("TPL CONN CB=0x%lx protocol=%d\n", (address_t)pConn, pConn->prot );
    printf("----------------------------------------------\n");
    printf("laddr=0x%x lport=%d\n", ntohl(pConn->laddr), ntohs(pConn->lport));
    printf("raddr=0x%x rport=%d\n", ntohl(pConn->raddr), ntohs(pConn->rport));
    printf("inpcb=0x%lx\n",  (address_t)pConn->pInpCb );
    printf("tcpcb=0x%lx\n",  (address_t)pConn->pTcpCb );
    printf("bindCb=0x%lx\n", (address_t)pConn->pBindCb );
    printf("appHdl=0x%lx\n", (address_t)pConn->pAppHdl );
	printf("src mac=%s\n",   bsdPrintMac((char*)pConn->srcMac));
	printf("dst mac=%s\n",   bsdPrintMac((char*)pConn->dstMac));
	printf("link=%d\n", pConn->linkNum);

    pkt_showDescRing( &pConn->txPktRing );
}

/*-----------------------------------------------------------------
 * Name:        tpl_showConn
 *
 * Description: Show all connection control blocks of given local port. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 12/12/01
 *-----------------------------------------------------------------
 */
void tpl_showConn( ushort16 lport )
{
	int32			i;
	tpl_bindCb_t	*pBind;
	tpl_connCb_t	*pConn;

    pBind = pTplCfg->bindTblBase;
    for (i=0; i < pTplCfg->maxBindCb; i++)
    {
        if (pBind->lport || pBind->rport)
		{
			if ((lport == ntohs(pBind->lport)) || 
				(lport == TPL_ANY_PORT))
			{
				pConn = pBind->pConnList;
				while(pConn)
				{
					tpl_showConnCb(pConn);
					pConn = pConn->pNext;
				}
			}
		}
        pBind++;
    }
}

/*-----------------------------------------------------------------
 * Name:        tpl_showConnPkt
 *
 * Description: Show all pkts of connCB. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 12/12/01
 *-----------------------------------------------------------------
 */
static void
tpl_showConnPkt( tpl_connCb_t *pConn )
{
	pkt_descRing_t	*pRing;
	pkt_desc_t 		**pCurr;

	pRing = &pConn->txPktRing;
	pCurr = pRing->startAddr;

	while (pCurr <= pRing->endAddr)
	{
    	pkt_showDesc( *pCurr );
		pCurr++;
	}
	
}

/*-----------------------------------------------------------------
 * Name:        tpl_showPkt
 *
 * Description: Show all pkts of connCB of given local port. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 12/12/01
 *-----------------------------------------------------------------
 */
void tpl_showPkt( ushort16 lport )
{
	int32			i;
	tpl_bindCb_t	*pBind;
	tpl_connCb_t	*pConn;

    pBind = pTplCfg->bindTblBase;
    for (i=0; i < pTplCfg->maxBindCb; i++)
    {
        if (pBind->lport || pBind->rport)
		{
			if ((lport == ntohs(pBind->lport)) || 
				(lport == TPL_ANY_PORT))
			{
				pConn = pBind->pConnList;
				while(pConn)
				{
					tpl_showConnPkt(pConn);
					pConn = pConn->pNext;
				}
			}
		}
        pBind++;
    }
}

void tpl_clearStats( void )
{
	memset(&pTplCfg->tplStats, 0, sizeof(tpl_stats_t));
}

/*-----------------------------------------------------------------
 * Name:        tpl_validateConnCb
 *
 * Description: Validate a given connection control block.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/10/01
 *-----------------------------------------------------------------
 */
int32 tpl_validateConnCb( void *pConnCb )
{
	tpl_connCb_t	*pConn;

	pConn = pConnCb;

	/*-------------------------------
	 * Check for valid range & flag.
	 *-------------------------------
	 */
	if ((pConn < pTplCfg->connTblBase ) ||
		(pConn > pTplCfg->connTblEnd )	||
		((pConn->flags & TPL_CONN_INUSE) == 0))
	{
		pTplCfg->tplStats.cntBadConnErr++;
		return TPL_ERR_BAD_CONN;
	}
	
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        tpl_validateBindCb
 *
 * Description: Validate a given Binding control block.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/10/01
 *-----------------------------------------------------------------
 */
int32 tpl_validateBindCb( tpl_bindCb_t *pBind )
{

	/*-------------------------------
	 * Check for valid range. 
	 *-------------------------------
	 */
	if ((pBind < pTplCfg->bindTblBase ) ||
		(pBind > pTplCfg->bindTblEnd ))	
	{
		pTplCfg->tplStats.cntBadBindErr++;
		return TPL_ERR_BAD_BIND;
	}

	/*-------------------------------
	 * Check if upcalls missing.
	 *-------------------------------
	 */
	if ((!pBind->ul_delConnInd) ||
		(!pBind->ul_addConnInd) ||
		(!pBind->ul_rcvPktInd) ||
		(!pBind->ul_xmtCmpInd)) 
	{
		pTplCfg->tplStats.cntBadBindErr++;
		return TPL_ERR_NOT_BIND;
	}
	
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:        tpl_cksum
 *
 * Description: Performs 16 bit IP/TCP/UDP checksum.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 10/25/01
 *-----------------------------------------------------------------
 */
// #define tpl_cksum tpl_cksum_ushort16
#define tpl_cksum tpl_cksum_uint64

ushort16
tpl_cksum_ushort16(ushort16 *pData, int32 len, int32 sum, int32 mode)
{

	/*-------------------------------
	 * Sum Data.
	 *-------------------------------
	 */
	while (len > 1)
	{
		sum += ntohs(*pData++);
		if (sum & 0x80000000)
			sum = (sum & 0xFFFF) + (sum >> 16);
		len -= 2;
	}

	if (len)
	{
		/*-------------------------------
		 * Odd byte. Pad last byte with 0.
		 *-------------------------------
		 */
		*((char *)pData+1) = 0;
		sum += ntohs(*pData);
	}

	while (sum >> 16)
		sum = (sum & 0xFFFF) + (sum >> 16);

    return ((mode == TPL_CKSUM_PARTIAL) ? sum : (~sum & 0xFFFF));
}

/*-----------------------------------------------------------------
 * Name:        tpl_cksum_uin64
 *
 * Description: Performs 16 bit IP/TCP/UDP checksum.
 *              Same as above, but uses 64 bit capabilities of CPU
 *-----------------------------------------------------------------
 */

ushort16
tpl_cksum_uint64(ushort16 *pData, int32 len, int32 sum, int32 mode)
{
    int alignLen;
	int64 sum_loc = sum;
    int64 local;
    int64 *pDataLoc = (int64*)pData;

	/*-------------------------------
	 * Sum Data.
	 *-------------------------------
	 */
    alignLen = ((address_t)pData & 0x7);
    if (alignLen)
    {
        alignLen = MIN(len, 8 - alignLen);
        while (alignLen > 1)
        {
            sum_loc += ((ntohs(*pData++)) & 0xffff);
            alignLen -= 2;
            len -=2;
        }
        if (len == 1)
        {
            ushort16 pad = 0;
            ((char*)&pad)[0] = *((char *)pData);
            sum_loc += ntohs(pad);
            len -= 1;
        }
    }
    
    
    pDataLoc = (int64*)pData;
	while (len > 7)
	{
        local = *pDataLoc++;
        // ntohs (4 times ntoh - short over 64 bits) 
		local = (((local & 0xff00ff00ff00ff00ll) >> 8) | ((local & 0x00ff00ff00ff00ffll) << 8));
		sum_loc += (((local      ) & 0xffffffff) +
                    ((local >> 32) & 0xffffffff));

		len -= 8;
	}
    pData = (ushort16 *)pDataLoc;
    
	while (len > 1)
	{
		sum_loc += ntohs(*pData++);
		len -= 2;
	}

	if (len)
	{
		/*-------------------------------
		 * Odd byte. Pad last byte with 0.
		 *-------------------------------
		 */
        ushort16 pad = 0;
        ((char*)&pad)[0] = *((char *)pData);
		sum_loc += ntohs(pad);
	}

	while (sum_loc >> 16)
    {
		sum_loc = (sum_loc & 0xFFFF) + (sum_loc >> 16);
    }
    
    sum = sum_loc;

    if (mode != TPL_CKSUM_PARTIAL)
    {
        sum = (~sum & 0xFFFF);
    }
    
    return sum;
}


/*-----------------------------------------------------------------
 * Name:        tpl_cksumTpl
 *
 * Description: Performs Transport layer (TCP/UDP) checksum.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 10/25/01
 *-----------------------------------------------------------------
 */
ushort16
tpl_cksumTpl(tpl_connCb_t *pConn, pkt_desc_t *pDesc)
{
    eee_desc_t      *eDescp;
    eee_descBuf_t   *bd;
    int32           bufCnt;
	int32			len;
	int32			sum;
	ushort16		*ip;
	ushort16		*pData;
	int32			mode = TPL_CKSUM_COMPLETE;

	/* su, 06/11/02, cksum for fragments too */
#if 0
	/*-------------------------------
	 * Cksum already done in slowpath
	 * (bsd code) for IP fragments. 
	 *-------------------------------
	 */
	if (pDesc->flags & PKT_IP_FRAGMENT)
		return 0;
#endif
	
	len  = pDesc->hdrSize;
    pData= (ushort16 *)pDesc->pTplHdr;

	/* su, 06/11/02, calculate for pseudo hdr sum, if single pkt or 
	 * first fragment
	 */
	if (!(pDesc->flags & PKT_IP_FRAGMENT) ||
		(pDesc->flags & PKT_IP_FIRST_FRAG))	{
		
		/* include hdr */
		len += (pConn->prot == TPL_PROT_UDP) ? PKT_UDP_HDR_SIZE : PKT_TCP_HDR_SIZE;
		

		/*-------------------------------
		 * Sum Pseudo header Dst IP.
		 *-------------------------------
		 */
		ip   = (ushort16 *)&(pConn->raddr);
		sum  = ntohs(*ip++);
		sum += ntohs(*ip);
		
		/*-------------------------------
		 * Sum Pseudo header Src IP.
		 *-------------------------------
		 */
		ip   = (ushort16 *)&(pConn->laddr);
		sum += ntohs(*ip++);
		sum += ntohs(*ip);
		
		/*-------------------------------
		 * Sum Pseudo header protocol & len.
		 *-------------------------------
		 */
		sum += pConn->prot;

		/* su, 06/12/02, can't use len for fragments
		 * chksum
		 */
		if (pDesc->flags & PKT_IP_FIRST_FRAG) {
			ushort16	udpLen;

			udpLen = *(ushort16*)(pDesc->pTplHdr + 4);
			sum += ntohs(udpLen);
		}
		else
			sum += len;
	}
	else
		sum = 0;


	/* su, 06/11/02,
	 * if pkt is fragment, return partial cksum 
	 */
	if ((pDesc->flags & PKT_IP_FRAGMENT))
		mode = TPL_CKSUM_PARTIAL;

	/* Su, tmp printf */
//	if (pDesc->flags & PKT_IP_FRAGMENT) {
//		printf("ip %lx, tpl %lx, hdr %lx, flag %x, proto %d hdrsize %d, sum %x\n",
//			   pDesc->pIpHdr, pDesc->pTplHdr, pDesc->pHdr, 
//			   pDesc->flags, pDesc->tplProt, len, sum);
//	}


	/*-------------------------------
	 * Get buffer count.
	 *-------------------------------
	 */
	eDescp = pDesc->eDescp;
	bufCnt = (eDescp->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;

    /*-------------------------------
     * Handle single buffer pkt.
     *-------------------------------
     */
    if (bufCnt == 1)
        return (tpl_cksum(pData, len, sum, mode));

    /*-------------------------------
     * Handle multi-buffer pkt.
     *-------------------------------
     */
	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
        &(eDescp->bd[0]) : &(eDescp->bd[1]);

    len = bd->len - ((uchar8 *)pDesc->pTplHdr - bd->buf);
    while( bufCnt-- )
    {
	    /*-------------------------------
	     * Partial cksum on EEE buff.
	     *-------------------------------
	     */
        sum = tpl_cksum(pData, len, sum, TPL_CKSUM_PARTIAL);

        bd++;
        pData = (ushort16 *)bd->buf;
        len   = bd->len ;
    }

	/* su, 06/11/02,
	 * if pkt is fragment, return partial cksum 
	 */
	return ((mode == TPL_CKSUM_PARTIAL)?sum:(~sum & 0xFFFF));
}

/*-----------------------------------------------------------------
 * Name:        tpl_cksumIp
 *
 * Description: Performs IP header checksum.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 10/25/01
 *-----------------------------------------------------------------
 */
ushort16
tpl_cksumIp(tpl_connCb_t *pConn, pkt_desc_t *pDesc)
{
	int32		len;
	ushort16	*pData;

	/* su, 06/11/02, chksum for all pkts 
	 * and fragments
	 */
#if 0
	/*-------------------------------
	 * Cksum already done in slowpath
	 * (bsd code) for IP fragments. 
	 *-------------------------------
	 */
	if (pDesc->flags & PKT_IP_FRAGMENT)
		return 0;
#endif /* 0 */

    len  = (pDesc->pTplHdr - pDesc->pIpHdr);
    pData= (ushort16 *)pDesc->pIpHdr;

	return (tpl_cksum(pData, len, 0, TPL_CKSUM_COMPLETE));
}
