/*------------------------------------------------------------------------------
 * RCS:  $Id: utf-api.h,v 1.1.1.1 2002/04/12 01:43:44 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : utf-api.c
 *
 * Description  : Unicode Transformation Format (UTF).
 *                API.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-APR-2002
 *
 *------------------------------------------------------------------------------
 */
#ifndef _UTF_API_H_
#define _UTF_API_H_

#ifdef FS_STANDALONE
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "../sm-fs/fs-includes.h"
#else
#include "nfx-incl.h"
#endif // FS_STANDALONE

/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef uint32  utf32;  // UTF32 = UCS4
typedef uint16  utf16;	// UTF16 = UCS2
typedef byte    utf8;	// UTF8

/* Some fundamental constants */
#define UNI_REPLACEMENT_CHAR    (utf32) 0x0000FFFD
#define UNI_MAX_BMP             (utf32) 0x0000FFFF
#define UNI_MAX_UTF16           (utf32) 0x0010FFFF
#define UNI_MAX_UTF32           (utf32) 0x7FFFFFFF

/*
  The value "sourceIllegal" is returned from some routines if the input
  sequence is malformed.  When "sourceIllegal" is returned, the source
  value will point to the illegal value that caused the problem. E.g.,
  in UTF-8 when a sequence is malformed, it points to the start of the
  malformed sequence.
*/
typedef enum
{
	UTF_CONVERSION_OK,      /* conversion successful */
	UTF_SOURCE_EXHAUSTED,	/* partial character in source, but hit end */
	UTF_TARGET_EXHAUSTED,	/* insuff. room in target for conversion */
	UTF_SOURCE_ILLEGAL      /* source sequence is illegal/malformed */
} utf_conversionResult_t;

/*
  These conversion functions take a ConversionFlags argument. When this
  flag is set to strict, both irregular sequences and isolated surrogates
  will cause an error.  When the flag is set to lenient, both irregular
  sequences and isolated surrogates are converted.

  Whether the flag is strict or lenient, all illegal sequences will cause
  an error return. This includes sequences such as: <F4 90 80 80>, <C0 80>,
  or <A0> in UTF-8, and values above 0x10FFFF in UTF-32. Conformant code
  must check for illegal sequences.

  When the flag is set to lenient, characters over 0x10FFFF are converted
  to the replacement character; otherwise (when the flag is set to strict)
  they constitute an error.
*/
typedef enum
{
	UTF_STRICT_CONVERSION = 0,
	UTF_LENIENT_CONVERSION
} utf_conversionFlags_t;

/*------------------------------------------------------------------------------
 * Functions prototypes.
 *------------------------------------------------------------------------------
 */

/* Conversions between UTF32, UTF-16, and UTF-8.
 * ---------------------------------------------
 * defined in utf-convert.c
 */

utf_conversionResult_t
utf_convertUTF32toUTF16(utf32** sourceStart,
                        const utf32* sourceEnd, 
                        utf16** targetStart,
                        const utf16* targetEnd,
                        utf_conversionFlags_t flags);

utf_conversionResult_t
utf_convertUTF16toUTF32(utf16** sourceStart,
                        utf16* sourceEnd, 
                        utf32** targetStart,
                        const utf32* targetEnd,
                        utf_conversionFlags_t flags);

utf_conversionResult_t
utf_convertUTF16toUTF8(utf16** sourceStart,
                       const utf16* sourceEnd, 
                       utf8** targetStart,
                       const utf8* targetEnd,
                       utf_conversionFlags_t flags);
		
utf_conversionResult_t
utf_convertUTF8toUTF16(utf8** sourceStart,
                       utf8* sourceEnd, 
                       utf16** targetStart,
                       const utf16* targetEnd,
                       utf_conversionFlags_t flags);

utf_conversionResult_t                              // output
utf_convertUTF32toUTF8(utf32 *      sourceStart,    // input
                       uint32       sourceLen,      // input
                       utf8 *       targetStart,    // input/output
                       uint32 *     targetLen,      // input/output
                       utf_conversionFlags_t flags);// input

		
utf_conversionResult_t                          // output
utf_convertUTF8toUTF32(utf8 *   sourceStart,    // input
                       uint32   sourceLen,      // input
                       utf32 *  targetStart,    // input/output
                       uint32 * targetLen);     // input/output

boolean
utf_isLegalUTF8Sequence(utf8 *  source,
                        utf8 *  sourceEnd);

/* Case Mapping.
 * -------------
 * defined in utf-case.c
 */

utf32
utf_toUpper(utf32   c);

utf32
utf_toLower(utf32   c);

void
utf_UTF32strToUpper(const utf32 *   source,
                    const int       sourceLen,
                    utf32 *         target);

boolean
utf_isAsciiString(utf8 *    source,
                  int       len);

void
utf_UTF8strToUpper(utf8 *   source,     // input
                   uint32   sourceLen,  // input
                   utf8 **  target,     // output
                   uint32 * targetLen); // output

/*----------------------------------------------------------------------------*/
#endif // _UTF_API_H_
