/*------------------------------------------------------------------------------
 * RCS:  $Id: utf-convert.c,v 1.1.1.1 2002/04/12 01:43:44 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : utf-convert.c
 *
 * Description  : Unicode Transformation Format (UTF).
 *                Conversions between UTF32, UTF-16, and UTF-8.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-APR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "utf-api.h"

/*------------------------------------------------------------------------------
 * Constants.
 *------------------------------------------------------------------------------
 */

static const int halfShift	= 10; /* used for shifting by 10 bits */

static const utf32 halfBase	= 0x0010000UL;
static const utf32 halfMask	= 0x3FFUL;

#define UNI_SUR_HIGH_START	(utf32)0xD800
#define UNI_SUR_HIGH_END	(utf32)0xDBFF
#define UNI_SUR_LOW_START	(utf32)0xDC00
#define UNI_SUR_LOW_END		(utf32)0xDFFF

/*------------------------------------------------------------------------------
 * Local functions prototypes.
 *------------------------------------------------------------------------------
 */

static boolean
utf_isLegalUTF8(utf8 *  source,
                int     length);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*
 * Utility routine to tell whether a sequence of bytes is legal UTF-8.
 * This must be called with the length pre-determined by the first byte.
 * If not calling this from ConvertUTF8to*, then the length can be set by:
 *	length = trailingBytesForUTF8[*source]+1;
 * and the sequence is illegal right away if there aren't that many bytes
 * available.
 * If presented with a length > 4, this returns false.  The Unicode
 * definition of UTF-8 goes up to 4-byte sequences.
 */

static boolean
utf_isLegalUTF8(utf8 *  source,
                int     length)
    
{
	utf8    a;
	utf8 *  srcptr = source+length;
    
	switch (length)
    {
	default: return FALSE;
        /* Everything else falls through when "true"... */
	case 4: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return FALSE;
	case 3: if ((a = (*--srcptr)) < 0x80 || a > 0xBF) return FALSE;
	case 2: if ((a = (*--srcptr)) > 0xBF) return FALSE;
		switch (*source)
        {
		    /* no fall-through in this inner switch */
        case 0xE0: if (a < 0xA0) return FALSE; break;
        case 0xF0: if (a < 0x90) return FALSE; break;
        case 0xF4: if (a > 0x8F) return FALSE; break;
        default:  if (a < 0x80) return FALSE;
		}
    case 1:
        if (*source >= 0x80 && *source < 0xC2) return FALSE;
        if (*source > 0xF4) return FALSE;
	}
    
	return TRUE;
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

utf_conversionResult_t
utf_convertUTF32toUTF16(utf32 **                    sourceStart,
                        const utf32 *               sourceEnd, 
                        utf16 **                    targetStart,
                        const utf16 *               targetEnd,
                        const utf_conversionFlags_t flags)
    
{
	utf_conversionResult_t  result = UTF_CONVERSION_OK;
	utf32 *                 source = *sourceStart;
	utf16 *                 target = *targetStart;
    
	while (source < sourceEnd)
    {
		utf32 ch;
		if (target >= targetEnd)
        {
			result = UTF_TARGET_EXHAUSTED;
            break;
		}
		ch = *source++;
		if (ch <= UNI_MAX_BMP)
        {
            /* Target is a character <= 0xFFFF */
			if ((flags == UTF_STRICT_CONVERSION) &&
                (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_LOW_END))
            {
				--source; /* return to the illegal value itself */
				result = UTF_SOURCE_ILLEGAL;
				break;
			}
            else
            {
			    *target++ = ch;	/* normal case */
			}
		}
        else if (ch > UNI_MAX_UTF16)
        {
			if (flags == UTF_STRICT_CONVERSION)
            {
				result = UTF_SOURCE_ILLEGAL;
			}
            else
            {
				*target++ = UNI_REPLACEMENT_CHAR;
			}
		}
        else
        {
			/* target is a character in range 0xFFFF - 0x10FFFF. */
			if (target + 1 >= targetEnd)
            {
				result = UTF_TARGET_EXHAUSTED;
                break;
			}
			ch -= halfBase;
			*target++ = (ch >> halfShift) + UNI_SUR_HIGH_START;
			*target++ = (ch & halfMask) + UNI_SUR_LOW_START;
		}
	}
	//*sourceStart = source;
	//*targetStart = target;
	return(result);
}

/*----------------------------------------------------------------------------*/

utf_conversionResult_t
utf_convertUTF16toUTF32(utf16 **                    sourceStart,
                        utf16 *                     sourceEnd, 
                        utf32 **                    targetStart,
                        const utf32 *               targetEnd,
                        const utf_conversionFlags_t flags)

{
	utf_conversionResult_t  result = UTF_CONVERSION_OK;
	utf16 *                 source = *sourceStart;
	utf32 *                 target = *targetStart;
	utf32                   ch, ch2;
    
	while (source < sourceEnd)
    {
		ch = *source++;
		if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END &&
            source < sourceEnd)
        {
			ch2 = *source;
			if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END)
            {
				ch = ((ch - UNI_SUR_HIGH_START) << halfShift)
					+ (ch2 - UNI_SUR_LOW_START) + halfBase;
				++source;
			}
            else if (flags == UTF_STRICT_CONVERSION)
            {
                /* it's an unpaired high surrogate */
				--source; /* return to the illegal value itself */
				result = UTF_SOURCE_ILLEGAL;
				break;
			}
		}
        else if ((flags == UTF_STRICT_CONVERSION) &&
                 (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END))
        {
			/* an unpaired low surrogate */
			--source; /* return to the illegal value itself */
			result = UTF_SOURCE_ILLEGAL;
			break;
		}
		if (target >= targetEnd)
        {
			result = UTF_TARGET_EXHAUSTED;
            break;
		}
		*target++ = ch;
	}
	//*sourceStart = source;
	//*targetStart = target;
	return result;
}

/*----------------------------------------------------------------------------*/

/*
 * Index into the table below with the first byte of a UTF-8 sequence to
 * get the number of trailing bytes that are supposed to follow it.
 */
static const char trailingBytesForUTF8[256] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

/*
 * Magic values subtracted from a buffer value during UTF8 conversion.
 * This table contains as many values as there might be trailing bytes
 * in a UTF-8 sequence.
 */
static const utf32 offsetsFromUTF8[6] = {
    0x00000000UL, 0x00003080UL, 0x000E2080UL, 
	0x03C82080UL, 0xFA082080UL, 0x82082080UL };

/*
 * Once the bits are split out into bytes of UTF-8, this is a mask OR-ed
 * into the first byte, depending on how many bytes follow.  There are
 * as many entries in this table as there are UTF-8 sequence types.
 * (I.e., one byte sequence, two byte... six byte sequence.)
 */
static const utf8 firstByteMark[7] = {
    0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC };

/*----------------------------------------------------------------------------*/

/* The interface converts a whole buffer to avoid function-call overhead.
 * Constants have been gathered. Loops & conditionals have been removed as
 * much as possible for efficiency, in favor of drop-through switches.
 * (See "Note A" at the bottom of the file for equivalent code.)
 * If your compiler supports it, the "isLegalUTF8" call can be turned
 * into an inline function.
 */

/*----------------------------------------------------------------------------*/

utf_conversionResult_t
utf_convertUTF16toUTF8(utf16 **                 sourceStart,
                       const utf16 *            sourceEnd, 
                       utf8 **                  targetStart,
                       const utf8 *             targetEnd,
                       utf_conversionFlags_t    flags)

{
	utf_conversionResult_t  result = UTF_CONVERSION_OK;
	utf16 *                 source = *sourceStart;
	utf8 *                  target = *targetStart;
    
	while (source < sourceEnd)
    {
		utf32 ch;
		unsigned short bytesToWrite = 0;
		const utf32 byteMask = 0xBF;
		const utf32 byteMark = 0x80; 
		ch = *source++;
		/* If we have a surrogate pair, convert to UTF32 first. */
		if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END &&
            source < sourceEnd)
        {
			utf32 ch2 = *source;
			if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END)
            {
				ch = ((ch - UNI_SUR_HIGH_START) << halfShift)
					+ (ch2 - UNI_SUR_LOW_START) + halfBase;
				++source;
			}
            else if (flags == UTF_STRICT_CONVERSION)
            {
                /* it's an unpaired high surrogate */
				--source; /* return to the illegal value itself */
				result = UTF_SOURCE_ILLEGAL;
				break;
			}
		}
        else if ((flags == UTF_STRICT_CONVERSION) &&
                 (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END))
        {
			--source; /* return to the illegal value itself */
			result = UTF_SOURCE_ILLEGAL;
			break;
		}
		/* Figure out how many bytes the result will require */
		if (ch < (utf32)0x80)
        {
            bytesToWrite = 1;
		}
        else if (ch < (utf32)0x800)
        {
            bytesToWrite = 2;
		}
        else if (ch < (utf32)0x10000)
        {
            bytesToWrite = 3;
		}
        else if (ch < (utf32)0x200000)
        {
            bytesToWrite = 4;
		}
        else
        {
            bytesToWrite = 2;
            ch = UNI_REPLACEMENT_CHAR;
		}

		target += bytesToWrite;
		if (target > targetEnd)
        {
			target -= bytesToWrite; result = UTF_TARGET_EXHAUSTED;
            break;
		}
		switch (bytesToWrite)
        {
            /* note: everything falls through. */
			case 4:	*--target = (ch | byteMark) & byteMask; ch >>= 6;
			case 3:	*--target = (ch | byteMark) & byteMask; ch >>= 6;
			case 2:	*--target = (ch | byteMark) & byteMask; ch >>= 6;
			case 1:	*--target =  ch | firstByteMark[bytesToWrite];
		}
		target += bytesToWrite;
	}
	//*sourceStart = source;
	//*targetStart = target;
	return result;
}

/*----------------------------------------------------------------------------*/

/*
 * Exported function to return whether a UTF-8 sequence is legal or not.
 * This is not used here; it's just exported.
 */

boolean
utf_isLegalUTF8Sequence(utf8 *  source,
                        utf8 *  sourceEnd)
    
{
	int length = trailingBytesForUTF8[*source]+1;
    
	if (source+length > sourceEnd)
    {
	    return FALSE;
	}
    
	return utf_isLegalUTF8(source, length);
}

/*----------------------------------------------------------------------------*/

utf_conversionResult_t
utf_convertUTF8toUTF16(utf8 **                      sourceStart,
                       utf8 *                       sourceEnd, 
                       utf16 **                     targetStart,
                       const utf16 *                targetEnd,
                       const utf_conversionFlags_t  flags)

{
	utf_conversionResult_t result = UTF_CONVERSION_OK;
	utf8 *                  source = *sourceStart;
	utf16 *                 target = *targetStart;
    
	while (source < sourceEnd)
    {
		utf32 ch = 0;
		unsigned short extraBytesToRead = trailingBytesForUTF8[*source];
		if (source + extraBytesToRead >= sourceEnd)
        {
			result = UTF_SOURCE_EXHAUSTED;
            break;
		}
		/* Do this check whether lenient or strict */
		if (!utf_isLegalUTF8(source, extraBytesToRead+1))
        {
			result = UTF_SOURCE_ILLEGAL;
			break;
		}
		/*
		 * The cases all fall through. See "Note A" below.
		 */
		switch (extraBytesToRead)
        {
			case 3:	ch += *source++; ch <<= 6;
			case 2:	ch += *source++; ch <<= 6;
			case 1:	ch += *source++; ch <<= 6;
			case 0:	ch += *source++;
		}
		ch -= offsetsFromUTF8[extraBytesToRead];

		if (target >= targetEnd)
        {
			result = UTF_TARGET_EXHAUSTED; break;
		}
		if (ch <= UNI_MAX_BMP)
        {
            /* Target is a character <= 0xFFFF */
			if ((flags == UTF_STRICT_CONVERSION) &&
                (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_LOW_END))
            {
				--source; /* return to the illegal value itself */
				result = UTF_SOURCE_ILLEGAL;
				break;
			}
            else
            {
			    *target++ = ch;	/* normal case */
			}
		}
        else if (ch > UNI_MAX_UTF16)
        {
			if (flags == UTF_STRICT_CONVERSION)
            {
				result = UTF_SOURCE_ILLEGAL;
				source -= extraBytesToRead; /* return to the start */
			}
            else
            {
				*target++ = UNI_REPLACEMENT_CHAR;
			}
		}
        else
        {
			/* target is a character in range 0xFFFF - 0x10FFFF. */
			if (target + 1 >= targetEnd)
            {
				result = UTF_TARGET_EXHAUSTED; break;
			}
			ch -= halfBase;
			*target++ = (ch >> halfShift) + UNI_SUR_HIGH_START;
			*target++ = (ch & halfMask) + UNI_SUR_LOW_START;
		}
	}
	//*sourceStart = source;
	//*targetStart = target;
	return result;
}

/*----------------------------------------------------------------------------*/

utf_conversionResult_t                              // output
utf_convertUTF32toUTF8(utf32 *      sourceStart,    // input
                       uint32       sourceLen,      // input
                       utf8 *       targetStart,    // input/output
                       uint32 *     targetLen,      // input/output
                       utf_conversionFlags_t flags) // input

{
	utf_conversionResult_t result = UTF_CONVERSION_OK;
	utf32 * source = sourceStart;
    utf32 * sourceEnd = sourceStart + sourceLen;
	utf8 *  target = targetStart;
    utf8 *  targetEnd = targetStart + *targetLen;
    uint32  targetNbChar = 0;
    
	while (source < sourceEnd)
    {
		utf32 ch;
		unsigned short bytesToWrite = 0;
		const utf32 byteMask = 0xBF;
		const utf32 byteMark = 0x80;
        
		ch = *source++;
		/* surrogates of any stripe are not legal UTF32 characters */
		if (flags == UTF_STRICT_CONVERSION )
        {
			if ((ch >= UNI_SUR_HIGH_START) && (ch <= UNI_SUR_LOW_END))
            {
				--source; /* return to the illegal value itself */
				result = UTF_SOURCE_ILLEGAL;
				break;
			}
		}
		/* Figure out how many bytes the result will require */
		if (ch < (utf32)0x80)
        {
            bytesToWrite = 1;
		}
        else if (ch < (utf32)0x800)
        {
            bytesToWrite = 2;
		}
        else if (ch < (utf32)0x10000)
        {
            bytesToWrite = 3;
		}
        else if (ch < (utf32)0x200000)
        {
            bytesToWrite = 4;
		}
        else
        {
            bytesToWrite = 2;
            ch = UNI_REPLACEMENT_CHAR;
		}
		
		target += bytesToWrite;
		if (target > targetEnd)
        {
			target -= bytesToWrite; result = UTF_TARGET_EXHAUSTED;
            break;
		}
		switch (bytesToWrite)
        {	/* note: everything falls through. */
			case 4:	*--target = (ch | byteMark) & byteMask; ch >>= 6;
			case 3:	*--target = (ch | byteMark) & byteMask; ch >>= 6;
			case 2:	*--target = (ch | byteMark) & byteMask; ch >>= 6;
			case 1:	*--target =  ch | firstByteMark[bytesToWrite];
		}
		target += bytesToWrite;
        targetNbChar++;
	}
    *targetLen = targetNbChar;
    
	return(result);
}

/*----------------------------------------------------------------------------*/

utf_conversionResult_t                          // output
utf_convertUTF8toUTF32(utf8 *   sourceStart,    // input
                       uint32   sourceLen,      // input
                       utf32 *  targetStart,    // input/output
                       uint32 * targetLen)      // input/output

{
	utf_conversionResult_t result = UTF_CONVERSION_OK;
	utf8 *  source = sourceStart;
    utf8 *  sourceEnd = sourceStart + sourceLen;
	utf32 * target = targetStart;
    utf32 * targetEnd = targetStart + *targetLen;
    uint32  targetNbChar = 0;
    
	while (source < sourceEnd)
    {
		utf32   ch = 0;
		uint16  extraBytesToRead = trailingBytesForUTF8[*source];
        
		if (source + extraBytesToRead >= sourceEnd)
        {
            result = UTF_SOURCE_EXHAUSTED;
            break;
		}
		/* Do this check whether lenient or strict */
		if (!utf_isLegalUTF8(source, extraBytesToRead+1))
        {
			result = UTF_SOURCE_ILLEGAL;
			break;
		}
		/*
		 * The cases all fall through. See "Note A" below.
		 */
		switch (extraBytesToRead)
        {
			case 3:	ch += *source++; ch <<= 6;
			case 2:	ch += *source++; ch <<= 6;
			case 1:	ch += *source++; ch <<= 6;
			case 0:	ch += *source++;
		}
		ch -= offsetsFromUTF8[extraBytesToRead];

		if (target >= targetEnd)
        {
			result = UTF_TARGET_EXHAUSTED;
            break;
		}
		if (ch <= UNI_MAX_UTF32)
        {
			*target++ = ch;
		}
        else if (ch > UNI_MAX_UTF32)
        {
			*target++ = UNI_REPLACEMENT_CHAR;
		}
        else
        {
			if (target + 1 >= targetEnd)
            {
				result = UTF_TARGET_EXHAUSTED;
                break;
			}
			ch -= halfBase;
			*target++ = (ch >> halfShift) + UNI_SUR_HIGH_START;
			*target++ = (ch & halfMask) + UNI_SUR_LOW_START;
		}
        targetNbChar++;
	}
    *targetLen = targetNbChar;
    
	return(result);
}

/*------------------------------------------------------------------------------

	Note A.
	The fall-through switches in UTF-8 reading code save a
	temp variable, some decrements & conditionals.  The switches
	are equivalent to the following loop:
		{
			int tmpBytesToRead = extraBytesToRead+1;
			do {
				ch += *source++;
				--tmpBytesToRead;
				if (tmpBytesToRead) ch <<= 6;
			} while (tmpBytesToRead > 0);
		}
	In UTF-8 writing code, the switches on "bytesToWrite" are
	similarly unrolled loops.

------------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
