/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         vs-mem.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Virtual server memory management functions.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/09/2002
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
// #include <exports/lock-internal.h>
#include "vs-mem.h"
#include <exports/elog-api.h>

uint32		vs_debugLevel = LOG_ERR;

/*FIXME: The virtual server memory management should be provided by
 *       the virtual server memory manager.
 *       Before it becomes available, the lock module temporarily uses
 *       the memory management scheme implemented in this file.
 */

/* Size of each data structure type */
#if 0
int vs_mem_dataSize[vs_mem_dt_max] = {
    sizeof(lock_cifsLockOwner_t),
    sizeof(lock_nlmRecOnVolCB_t),
    sizeof(lock_nlmLockClient_t),
    sizeof(lock_nlmLockOwner_t),
    sizeof(lock_cifsLockInfo_t),
    sizeof(lock_nlmLockInfo_t),
    sizeof(lock_exLock_t),
    sizeof(lock_shLock_t),
    sizeof(lock_shNode_t),
    sizeof(lock_pendingLock_t),
    sizeof(lock_fileLock_t),
	sizeof(lock_dependency_t),
	sizeof(lock_ownerTimeout_t),
    /*Note: The above data structures are defined in lock-internal.h"*/
	sizeof(nlm_request_t),
    /*Note: The above data structures are defined in nlm.h"*/
};
#endif

/* Virtual server table. It is made up of memSrvrCBs for memory management.*/
vs_mem_srvrCB_t vs_mem_srvrTable[VS_MEM_MAX_SRVR_IDX];


/*-----------------------------------------------------------------
 * Name:         vs_mem_init
 *
 * Description:
 *
 *     Initialize virtual server table.
 *
 * Arguments:    
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void
vs_mem_init(void)
{
	uint32				i, j;
	vs_mem_srvrCB_t		*memSrvrCBp;
	vs_mem_memCBList_t	*memCBListp;

	/*---------------------------------------------
	 * Initialize each item in virtual server table
	 *---------------------------------------------
	 */
	for (i = 0; i < VS_MEM_MAX_SRVR_IDX; i++)
	{
		memSrvrCBp = &vs_mem_srvrTable[i];

		/*---------------------
		 * Initialize server Id
		 *---------------------
		 */
		/*FIXME*/
		memSrvrCBp->srvrIdx = 0;

		/*----------------------------------------
		 * Initialize each item in memCBList table
		 *----------------------------------------
		 */
		for (j = 0; j < vs_mem_dt_max; j++)
		{
			memCBListp = &(memSrvrCBp->memCBListTable[j]);
			memCBListp->highestIdx = 0;
			// memCBListp->dataSize = vs_mem_dataSize[j];
			memCBListp->dataSize = 0;
			memCBListp->usedCount = 0;
			memCBListp->freeCount = 0;
			memCBListp->firstMemCBp = NULL;
		}
	}
} /* vs_mem_init() */

/*-----------------------------------------------------------------
 * Name:         vs_mem_initChunk
 *
 * Description:
 *     Initialize each memory unit in a memory chunk.
 *
 * Arguments:    
 *
 *     firstMemUnitp - Pointer to the first memory unit in the memory chunk.
 *
 *     lowestIdx - The lowest index to the memory units in the memory chunk.
 *
 *     highestIdx - The highest index to the memory units in the memory chunk.
 *
 *     dataSize - Size of the data structure type.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
static inline void
vs_mem_initChunk(vs_mem_unitHead_t	*firstMemUnitp,
				 uint32				lowestIdx,
				 uint32				highestIdx,
				 uint32				dataSize)
{
	uint32				i;
	vs_mem_unitHead_t	*memUnitp;

	memUnitp = firstMemUnitp;
	for (i = lowestIdx; i <= highestIdx; i++)
	{
		/* Init admin fields */
		memUnitp->admin.used = VS_MEM_UNIT_NOTUSED;

		/* Set nextFreeIdx */
		if (i < highestIdx)
			memUnitp->nextFreeIdx = i + 1;
		else
			memUnitp->nextFreeIdx = 0;

		/* Go to next memUnit */
		memUnitp = (vs_mem_unitHead_t *)
				((char *)(memUnitp) + VS_MEM_UNIT_HEAD_SIZE + dataSize);
	}
} /* vs_mem_initChunk() */

/*-----------------------------------------------------------------
 * Name:         vs_mem_addChunk
 *
 * Description:
 *
 *     Allocate a new chunk of memory from the memory manager, and attach
 *     it to a memory control block under a virtual server.
 *
 * Arguments:    
 *
 *     srvrIdx - Virtual server index.
 *
 *     memCBListp - Pointer to memory CB list.
 *
 * Return Value:
 *
 *    Pointer to memCB which the new memory chunk is attached to - If succeed.
 *
 *    NULL - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void *
vs_mem_addChunk(uint32				srvrIdx,
				vs_mem_memCBList_t	*memCBListp)
{
	vs_mem_memCB_t		*currMemCBp, *prevMemCBp = NULL;
	boolean				reuse = FALSE;
	uint32				count;
	uint32				memSize;
	vs_mem_unitHead_t	*firstMemUnitp;

	/*-----------------------------------------------
	 * In the memCBList, look for a memCB that we can 
	 * attach a new chunk of memory to it.
	 *-----------------------------------------------
	 */
	currMemCBp = memCBListp->firstMemCBp;
	while (currMemCBp)
	{
		if ((currMemCBp->usedCount == 0) && (currMemCBp->freeCount == 0))
		{
			/* No memory is attached to this memCB. We can reuse it. */
			reuse = TRUE;
			break;
		}
		else
		{
			prevMemCBp = currMemCBp;
			currMemCBp = currMemCBp->nextMemCBp;
		}
	}

	if (currMemCBp == NULL)
	{
		/* Either there is no memCB in the memCBList or
		 * all memCB's in the memCBList have memory attached to them.
		 */

		uint32	availCount;

		/*--------------------------------------
		 * Check if there is room for more units 
		 *--------------------------------------
		 */
		availCount = VS_MEM_MAX_INDEX - memCBListp->usedCount;
		if (!availCount)
		{
			VS_MEM_PRINTF(LOG_INFO, ("vs_mem_addChunk: "
										"Reached maximum number of units."));
			return (NULL);
		}

		/*-------------------------------------------
		 * We need to allocate a new memCB
		 * This memory is virtual server independent.
		 *-------------------------------------------
		 */
		VS_MEM_SHMEM_ALLOC(currMemCBp,
						   VS_MEM_SRVR_ID_ANY, /*FIXME: What value should be?*/
						   vs_mem_memCB_t,
						   sizeof(vs_mem_memCB_t));
		if (currMemCBp == NULL)
			return (NULL);

		bzero(currMemCBp, sizeof(vs_mem_memCB_t));

		/*---------------------------------
		 * Add the new memCB into memCBList
		 *---------------------------------
		 */
		if (memCBListp->firstMemCBp == NULL)
		{
			/* Become the firstMemCB in the list*/
			memCBListp->firstMemCBp = currMemCBp;

			/* Set the lowestIdx */
			currMemCBp->lowestIdx = 1;
		}
		else
		{
			/* Become the tail of the list*/
			prevMemCBp->nextMemCBp = currMemCBp;

			/* Set the lowestIdx */
			currMemCBp->lowestIdx = prevMemCBp->highestIdx + 1;
		}
		/* Set the highestIdx, assuming we can allocate 
		 * up to VS_MEM_MAX_UNITS_IN_MEMCB uints */
		count = (availCount < VS_MEM_MAX_UNITS_IN_MEMCB) ?
					availCount : VS_MEM_MAX_UNITS_IN_MEMCB;
		currMemCBp->highestIdx = currMemCBp->lowestIdx + count - 1;

		currMemCBp->nextMemCBp = NULL;

		if (memCBListp->highestIdx < currMemCBp->highestIdx)
			memCBListp->highestIdx = currMemCBp->highestIdx;
	}

	/*---------------------------------------------------
	 * Allocate a chunk of memory from the memory manager
	 *---------------------------------------------------
	 */
	/* Determine the number of units to allocate */
	count = currMemCBp->highestIdx - currMemCBp->lowestIdx + 1;
	memSize = (VS_MEM_UNIT_HEAD_SIZE + memCBListp->dataSize) * count;
	VS_MEM_SHMEM_ALLOC(firstMemUnitp, srvrIdx, vs_mem_unitHead_t, memSize);
	if (firstMemUnitp == NULL)
	{
		/*TODO: Improvement: Try agian. Allocate half size of memory?*/ 
		return (NULL);
	}

	/*--------------------------------
	 * Initialize the new memory chunk
	 *--------------------------------
	 */
	vs_mem_initChunk(firstMemUnitp,
					 currMemCBp->lowestIdx,
					 currMemCBp->highestIdx,
					 memCBListp->dataSize);

	/*-------------------------
	 * Initialize the currMemCB
	 *-------------------------
	 */
	currMemCBp->usedCount = 0;
	currMemCBp->freeCount = count;
	currMemCBp->freeQHeadIdx = currMemCBp->lowestIdx;
	currMemCBp->freeQTailIdx = currMemCBp->highestIdx;
	currMemCBp->firstMemUnitp = firstMemUnitp;

	/*---------------------
	 * Update the memCBList
	 *---------------------
	 */
	memCBListp->freeCount += count;

	return (currMemCBp);
} /* vs_mem_addChunk() */

/*-----------------------------------------------------------------
 * Name:         vs_mem_alloc
 *
 * Description:
 *
 *     Allocate a free memory unit of a given data type.
 *
 * Arguments:    
 *
 *     srvrIdx - Virtual server index.
 *
 *     dataType - Data type index.
 *
 *     index - This is an output. It it the index to the allocated
 *             memory unit.
 *
 * Return Value:
 *
 *    Pointer to the allocated data structure - If succeed.
 *
 *    NULL - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void *
vs_mem_alloc(vs_mem_srvr_idx_t	srvrIdx,
			 vs_mem_dataType_t	dataType,
			 vs_mem_index_t		*index)
{
	vs_mem_srvrCB_t		*memSrvrCBp;
	vs_mem_memCBList_t	*memCBListp;
	vs_mem_memCB_t		*memCBp;
	uint32				memUnitSize;
	vs_mem_unitHead_t	*freeMemUnitp;
	void				*freeDataStructp = NULL;

	if (index == NULL)
		return (NULL);

	/*---------------------------
	 * Init return value of index 
	 *-------------------------
	 */
	index->srvrIdx = srvrIdx;
	index->dataType = dataType;
	index->idx = 0;

	/*--------------------------
	 * Validate input parameters
	 *--------------------------
	 */
	if ((srvrIdx >= VS_MEM_MAX_SRVR_IDX) ||
		(dataType >= vs_mem_dt_max))
	{
		VS_MEM_PRINTF(LOG_ERR, ("vs_mem_alloc: Invalid input parameter."));
		return (NULL);
	}

	/*---------------------------------------------------------
	 * Look for the memCB based on virtual server and data type
	 *---------------------------------------------------------
	 */
	memSrvrCBp = &vs_mem_srvrTable[srvrIdx];

	memCBListp = &(memSrvrCBp->memCBListTable[dataType]);
	memUnitSize = VS_MEM_UNIT_HEAD_SIZE + memCBListp->dataSize;

	if (memCBListp->freeCount == 0)
	{
		/* All allocated memory are used up.
		 * Get more memory from memory manager
		 */
		memCBp = vs_mem_addChunk(memSrvrCBp->srvrIdx, memCBListp);
	}
	else
	{
		memCBp = memCBListp->firstMemCBp;
	}

	/*------------------------
	 * Look for an unused unit
	 *------------------------
	 */
	while (memCBp)
	{
		if (memCBp->freeCount == 0)
		{
			/* Search for a free unit in the next memory trunk */
			memCBp = memCBp->nextMemCBp;
			continue;
		}

		/* Found an unused unit in this memory trunk */
		freeMemUnitp = (vs_mem_unitHead_t *)
					((char *)(memCBp->firstMemUnitp) + 
					(memCBp->freeQHeadIdx - memCBp->lowestIdx) * memUnitSize);

		freeDataStructp = 
					(void *)((char *)freeMemUnitp + VS_MEM_UNIT_HEAD_SIZE);

		index->idx = memCBp->freeQHeadIdx;

		memCBp->usedCount++;
		memCBp->freeCount--;
		memCBListp->usedCount++;
		memCBListp->freeCount--;

		memCBp->freeQHeadIdx = freeMemUnitp->nextFreeIdx;
		if (memCBp->freeCount == 0)
			memCBp->freeQTailIdx = 0;

		/*Initialize the allocated memory unit*/
		freeMemUnitp->admin.used = VS_MEM_UNIT_USED;
		freeMemUnitp->nextFreeIdx = 0;

		break;
	}

	return (freeDataStructp);
} /* vs_mem_alloc() */

/*-----------------------------------------------------------------
 * Name:         vs_mem_free
 *
 * Description:
 *
 *     Free a memory unit.
 *
 * Arguments:    
 *
 *     index - The index to the memory unit to be freed.
 *
 * Return Value:
 *
 *    NFX_OK - If succeed.
 *
 *    NFX_ERR - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
vs_mem_free(vs_mem_index_t	index)
{
	uint32				srvrIdx;
	vs_mem_dataType_t	dataType;
	uint32				idx;
	vs_mem_srvrCB_t		*memSrvrCBp;
	vs_mem_memCBList_t	*memCBListp;
	uint32				memUnitSize;
	boolean				delList = FALSE;
	vs_mem_memCB_t		*memCBp, *prevMemCBp;
	vs_mem_unitHead_t	*memUnitp, *freeQTailp;

	srvrIdx = index.srvrIdx;
	dataType = index.dataType;
	idx = index.idx;

	/*--------------------------
	 * Validate input parameters
	 *--------------------------
	 */
	if ((srvrIdx >= VS_MEM_MAX_SRVR_IDX) ||
		(dataType >= vs_mem_dt_max) ||
		(idx == 0))
	{
		VS_MEM_PRINTF(LOG_ERR, ("vs_mem_free: Invalid input parameter."));
		return (NFX_ERR);
	}

	/*---------------------------------------------------------
	 * Look for the memCB based on virtual server and data type
	 *---------------------------------------------------------
	 */
	memSrvrCBp = &vs_mem_srvrTable[srvrIdx];

	memCBListp = &(memSrvrCBp->memCBListTable[dataType]);
	memUnitSize = VS_MEM_UNIT_HEAD_SIZE + memCBListp->dataSize;

	/*Check number of used units*/
	if (memCBListp->usedCount == 0)
	{
		VS_MEM_PRINTF(LOG_ERR, ("vs_mem_free: No memory unit is being used."));
		return (NFX_ERR);
	}

	/*Validate index in input*/
	if (memCBListp->highestIdx < idx)
	{
		VS_MEM_PRINTF(LOG_ERR, ("vs_mem_free: "
							"Invalid memory unit index in input parameter."));
		return (NFX_ERR);
	}

	if (memCBListp->usedCount == 1)
	{
		/*Free all the memCBs in the list */
		delList = TRUE;
	}

	memCBp = memCBListp->firstMemCBp;
	while (memCBp)
	{
		if (memCBp->highestIdx < idx)
		{
			/*Go to check next memCB*/
			prevMemCBp = memCBp;
			memCBp = memCBp->nextMemCBp;

			if (delList == TRUE)
			{
				/*Dealloc this memCB*/
				VS_MEM_SHMEM_DEALLOC(prevMemCBp);
			}
			continue;
		}

		/*----------------
		 * Found the memCB
		 *----------------
		 */
		if (memCBp->usedCount == 1)
		{
			/* All the units in the memory chunk are unused.
			 * Free the memory chunk*/
			VS_MEM_SHMEM_DEALLOC(memCBp->firstMemUnitp);
			memCBListp->usedCount -= 1;
			if (memCBListp->freeCount > memCBp->freeCount)
				memCBListp->freeCount -= (memCBp->freeCount + 1);
			else
				memCBListp->freeCount = 0;
			memCBp->usedCount = 0;
			memCBp->freeCount = 0;
		}
		else
		{
			/*Add the unit to free queue*/
			/*Locate address of the memory unit to be freed*/
			memUnitp = (vs_mem_unitHead_t *)
					((char *)(memCBp->firstMemUnitp) + 
					(idx - memCBp->lowestIdx) * memUnitSize);

			/* Mark the unit to be unused */
			memUnitp->admin.used = VS_MEM_UNIT_NOTUSED;

			memCBp->usedCount--;
			memCBp->freeCount++;
			memCBListp->usedCount--;
			memCBListp->freeCount++;

			/*Add the unit to the free queue*/
			if (memCBp->freeCount == 1)
			{
				/*The unit is the only one in the queue*/
				memCBp->freeQHeadIdx = idx;
				memCBp->freeQTailIdx = idx;
			}
			else
			{
				/*Locate address of the tail unit in the free queue*/
				freeQTailp = (vs_mem_unitHead_t *)
					((char *)(memCBp->firstMemUnitp) + 
					(memCBp->freeQTailIdx - memCBp->lowestIdx) * memUnitSize);

				/*Append the unit to end of the queue*/
				freeQTailp->nextFreeIdx = idx;
				memCBp->freeQTailIdx = idx;
			}
		}

		break;
	}

	/* Continue to dealloc the memCB if necessary */
	if (delList == TRUE)
	{
		while (memCBp)
		{
			prevMemCBp = memCBp;
			memCBp = memCBp->nextMemCBp;
			VS_MEM_SHMEM_DEALLOC(prevMemCBp);
		}

		memCBListp->highestIdx = 0;
		memCBListp->usedCount = 0;
		memCBListp->freeCount = 0;
		memCBListp->firstMemCBp = NULL;
	}

	return (NFX_OK);
} /* vs_mem_free() */

/*-----------------------------------------------------------------
 * Name:         vs_mem_idxToPtr
 *
 * Description:
 *
 *     Find a memory unit.
 *
 * Arguments:    
 *
 *     index - The index to the memory unit to be found.
 *
 * Return Value:
 *
 *    Pointer to the data structure - If found.
 *
 *    NULL - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void *
vs_mem_idxToPtr(vs_mem_index_t	index)
{
	uint32				srvrIdx;
	vs_mem_dataType_t	dataType;
	uint32				idx;
	vs_mem_srvrCB_t		*memSrvrCBp;
	vs_mem_memCBList_t	*memCBListp;
	uint32				memUnitSize;
	vs_mem_memCB_t		*memCBp;
	vs_mem_unitHead_t	*memUnitp;
	void				*dataStructp;

	srvrIdx = index.srvrIdx;
	dataType = index.dataType;
	idx = index.idx;

	/*--------------------------
	 * Validate input parameters
	 *--------------------------
	 */
	if ((srvrIdx >= VS_MEM_MAX_SRVR_IDX) ||
		(dataType >= vs_mem_dt_max))
	{
		VS_MEM_PRINTF(LOG_ERR, ("vs_mem_idxToPtr: Invalid input parameter."));
		return (NULL);
	}

	if (idx == 0)
		return (NULL);

	/*---------------------------------------------------------
	 * Look for the memCB based on virtual server and data type
	 *---------------------------------------------------------
	 */
	memSrvrCBp = &vs_mem_srvrTable[srvrIdx];

	memCBListp = &(memSrvrCBp->memCBListTable[dataType]);
	memUnitSize = VS_MEM_UNIT_HEAD_SIZE + memCBListp->dataSize;

	/*Validate memory unit index in input*/
	if (memCBListp->highestIdx < idx)
	{
		//VS_MEM_PRINTF(LOG_ERR, ("vs_mem_idxToPtr: "
		VS_MEM_PRINTF(LOG_DEBUG, ("vs_mem_idxToPtr: "
							"Invalid memory unit index in input parameter."));
		return (NULL);
	}

	memCBp = memCBListp->firstMemCBp;
	while (memCBp)
	{
		if (memCBp->highestIdx < idx)
		{
			/*Go to check next memCB*/
			memCBp = memCBp->nextMemCBp;
			continue;
		}

		/*----------------
		 * Found the memCB
		 *----------------
		 */
		if (memCBp->firstMemUnitp == NULL)
		{
			//VS_MEM_PRINTF(LOG_CRIT, ("vs_mem_idxToPtr: "
			VS_MEM_PRINTF(LOG_DEBUG, ("vs_mem_idxToPtr: "
									"memCB->firstMemUnitp Shouldn't be NULL!"));
		}

		/*Locate address of the found unit*/
		memUnitp = (vs_mem_unitHead_t *)
					((char *)(memCBp->firstMemUnitp) + 
					(idx - memCBp->lowestIdx) * memUnitSize);
		if (memUnitp->admin.used != VS_MEM_UNIT_USED)
		{
			//VS_MEM_PRINTF(LOG_ERR, ("vs_mem_idxToPtr: "
			VS_MEM_PRINTF(LOG_DEBUG, ("vs_mem_idxToPtr: "
									"Memory unit is unused."));
			return (NULL);
		}

		dataStructp = (void *)((char *)memUnitp + VS_MEM_UNIT_HEAD_SIZE);
		break;
	}

	return (dataStructp);
} /* vs_mem_idxToPtr() */

/*-----------------------------------------------------------------
 * Name:         vs_mem_ptrToIdx
 *
 * Description:
 *
 *     Obtain the index from a given data structure pointer.
 *
 * Arguments:    
 *
 *     srvrIdx - Virtual server index.
 *
 *     dataType - Data type index.
 *
 *     dataStruct - Pointer to the data structure.
 *
 * Return Value:
 *
 *    A valid index to the given data structure - If found.
 *
 *    A zero index - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/20/2002
 *
 *-----------------------------------------------------------------
 */
vs_mem_index_t
vs_mem_ptrToIdx(vs_mem_srvr_idx_t	srvrIdx,
				vs_mem_dataType_t	dataType,
				void				*dataStructp)
{
	vs_mem_index_t		dataStructIdx;

	/*TODO*/
	dataStructIdx.srvrIdx = srvrIdx;
	dataStructIdx.dataType = dataType;
	dataStructIdx.idx = 1;
	
	return (dataStructIdx);
} /* vs_mem_ptrToIdx() */

/*-----------------------------------------------------------------
 * Name:         vs_memSetDataSize
 *
 * Description:
 *
 *     Initialize data size.
 *
 * Arguments:    
 *
 *     None.
 *
 * Return Value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/14/2002
 *
 *-----------------------------------------------------------------
 */
void
vs_memSetDataSize(uint32 type, uint32 size)
{
	uint32				i;
	vs_mem_srvrCB_t		*memSrvrCBp;

    if (type >= vs_mem_dt_max)
        return;

	for (i = 0; i < VS_MEM_MAX_SRVR_IDX; i++)
	{
		memSrvrCBp = &vs_mem_srvrTable[i];

		/*----------------------------------------
		 * Initialize data size item in memCBList table
		 *----------------------------------------
		 */
        memSrvrCBp->memCBListTable[type].dataSize = size;
	}

    return;
} /* vs_memSetDataSize() */

