/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:         vs-mem.h
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description:  Virtual server memory management data structures.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/09/2002
 *
 *-----------------------------------------------------------------
 */

#ifndef _VS_MEM_H
#define _VS_MEM_H

#include <exports/vs-mem-api.h>

/*FIXME: The virtual server memory management should be provided by the 
 *       virtual server memory manager.
 *       Before it becomes available, the lock module temporarily uses
 *       the memory management scheme described as follows.
 */

/* All the instances of the data structures that contain persistent data are
   organized in a memory table. Every data structure instance is identified
   by an index. From this index, we can derive the address where the data
   structure instance is stored in the memory. The organization of the memory
   management is illustrated in the following diagram.


        Virtual server table
           +------------+   memory control blocks of one data type
  srvrCB   |            |   +-------+    +-------+
    1      |memCBList 0 +-->| memCB +--->| memCB +--->
           |            |   +---+---+    +-------+
           |            |       |    memory control blocks of another data type
           |            |       |     +-------+    +-------+
           |memCBList 1 +-------|---->| memCB +--->| memCB +--->
           | :          |       |     +-------+    +-------+
           +------------+       |
  srvrCB   |            |       |
    2      |            |       |
           +------------+       |
           | :          |       |
           | :          |       |
           +------------+       |
  srvrCB   |memCBList   |       |
    255    |        255 |       |
           +------------+       |
                                |
             +------------------+
             |
             |          memory chunk
             +------->|===============================|<-------+
            memUnit 1 | admin fields | next free unit +--------|---+
                      |-------------------------------|        |   |
   head of            | data structure                |        |   |
   free queue-------> |===============================|        |   |
            memUnit 2 | admin fields | next free unit +-----+  |   |
                      |-------------------------------|     |  |   |
                      | data structure                |     |  |   |
                      |===============================|     |  |   |
            memUnit 3 | admin fields | next free unit +     |  |   |
                      |-------------------------------|     |  |   |
                      | data structure                |     |  |   |
                      |===============================|<----+  |   |
            memUnit 4 | admin fields | next free unit +--------+   |
                      |-------------------------------|            |
                      | data structure                |            |
                      |===============================|            |
            memUnit 5 | admin fields | next free unit +            |
                      |-------------------------------|            |
   tail of            | data structure                |            |
   free queue-------->|===============================|<-----------+
            memUnit 6 | admin fields | next free unit +            
                      |-------------------------------|
                      | data structure                |
                      |===============================|
                      |  :                            |
                      |  :                            |
                      |===============================|


      Several chunks of memory are pre-allocated for the different types of
    data structures, including
        lock_cifsLockOwner_t,
        lock_nlmRecOnVolCB_t,
        lock_nlmLockClient_t,
        lock_nlmLockOwner_t,
        lock_cifsLockInfo_t,
        lock_nlmLockInfo_t,
        lock_exLock_t,
        lock_shLock_t,
        lock_shNode_t,
        lock_pendingLock_t,
        lock_fileLock_t.

      A memory chunk for any type of data structures is made up of a number
    of memory units, which are controlled by a memory control block, which in
    turn are organized by a virtual server table.

      At any time, some of the memory units may be used and some may be free.
    The free units are chained in a free queue. We maintain two pointers to
    the head and the tail of the free queue. When allocating a free unit for
    use, we get one from the head of the free queue. When a unit is to be put
    back to the free queue, we append it to the end of the queue.
    
      Each memory unit contains the data structure itself and some
    admnistrative fields, such as
        - Flag indicating whether the unit is used or not,
        - Pointers to file name and line number where the unit is allocated
          for use,
        - Pointer to next free unit in the free queue,

      Admnistrative variables for each type of queues may include,
        - Flag indicating whether the data structures should be moved during
          switch-over from one filer to another,
        - Available free units,
        - Low water mark.

	TODO:
	    To accelerate searching a memory unit, we may change the orgnization
		of memCB as follows.

		+---------+                   +---------+
		| memCB   +------------------>| memCB   +------>
		|         |                   |         |
		|         +-+                 |         |
        +---------+ |                 +---------+
		            |
		            |    mem chunk table
		            +-->+-------------+
					    | mem chunk 1 +---> memory chunk
						+-------------+
					    | mem chunk 2 +---> memory chunk
						+-------------+
						| :           |
						+-------------+
					    | mem chunk N +---> memory chunk
						+-------------+
*/


/*------------------------------
 * Memory unit in a memory chunk
 *------------------------------
 */
#define VS_MEM_UNIT_NOTUSED        0
#define VS_MEM_UNIT_USED           1
#define VS_MEM_MAX_UNITS_IN_MEMCB  128  /*Max # of units in a memory chunk*/

typedef struct vs_mem_admin {
    uint32              used;         /*The memory unit is in use or not*/

    /*Other fields may be added if necessary*/
} vs_mem_admin_t;

typedef struct vs_mem_unitHead {
    vs_mem_admin_t      admin;        /*Admin info*/
    uint32              nextFreeIdx;  /*Index of next free unit in the memory
                                        chunk*/

    /*A data structure immediately follows*/
} vs_mem_unitHead_t;

#define VS_MEM_UNIT_HEAD_SIZE sizeof(vs_mem_unitHead_t)

/*---------------------------------------------------------------------
 * Memory control block. It manages a chunk of memory of one data type.
 *---------------------------------------------------------------------
 */
typedef struct vs_mem_memCB {
    uint32              lowestIdx;    /*Lowest unit index in the memory chunk*/
    uint32              highestIdx;   /*Highest unit index in the memory chunk*/
    uint32              usedCount;    /*Number of used memory units*/
    uint32              freeCount;    /*Number of free memory units*/
    uint32              freeQHeadIdx; /*Head of free memory unit queue*/
    uint32              freeQTailIdx; /*Tail of free memory unit queue*/
    vs_mem_unitHead_t   *firstMemUnitp; /*Point to the first memory unit*/
    struct vs_mem_memCB *nextMemCBp;  /*Point to next memCB in the list*/
} vs_mem_memCB_t;

/*------------------------------------------------------------------------
 * Memory control block list. It manages a list of memCB of one data type.
 *------------------------------------------------------------------------
 */
typedef struct vs_mem_memCBList {
    uint32              flag;
#define VS_MEM_DT_MOVE 0x0001;        /*Move the memory for this data type
                                        during switch-over*/

    uint32              highestIdx;   /*Highest unit index of memory chunks*/
    uint32              dataSize;     /*Size of data structure*/
    uint32              usedCount;    /*Total number of used memory units*/
    uint32              freeCount;    /*Total number of free memory units*/
    vs_mem_memCB_t      *firstMemCBp; /*Point to the first memCB in the list*/
} vs_mem_memCBList_t;

/*----------------------------------------------------------------------------
 * Virtual server control block.
 * It manages all memory allocated for data structures under a virtual server.
 *----------------------------------------------------------------------------
 */
typedef struct vs_mem_srvrCB {
    uint32              srvrIdx;
    vs_mem_memCBList_t  memCBListTable[vs_mem_dt_max];
} vs_mem_srvrCB_t;

extern uint32 vs_debugLevel;

/*-------
 * Macros
 *-------
 */
/*FIXME: TBD. Use elog?*/
#define VS_MEM_PRINTF(_lvl_, _par_)                                         \
    if (_lvl_ <= vs_debugLevel)                                             \
        {printf _par_; printf("\n");}

/* Shared memory alloc */
#define VS_MEM_SHMEM_ALLOC(_p_, _srvr_, _type_, _size_)                     \
    {                                                                       \
        _p_ = (_type_ *) eee_ramAlloc(_size_);                              \
        if (_p_ == NULL)                                                    \
        {                                                                   \
            VS_MEM_PRINTF(LOG_ERR, ("%s: eee_ramAlloc() failed", __FUNCTION__)) \
        }                                                                   \
    }

/* Shared memory dealloc */
#define VS_MEM_SHMEM_DEALLOC(_p_)                                           \
    {                                                                       \
        if (_p_ != NULL)                                                    \
        {                                                                   \
            eee_ramDealloc(_p_);                                            \
            _p_ = NULL;                                                     \
        }                                                                   \
    }


/*--------------------
 * Function Prototypes
 *--------------------
 */

#endif /* _VS_MEM_H */

