/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ccm-lvfailed.c
 *
 * RCS:  $Id: ccm-lvfailed.c,v 1.14 2002/09/12 15:52:24 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Contains all the bits and pieces relating to tracking
 *              failed volumes.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 23 May 2002
 *
 *-----------------------------------------------------------------
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

#include "exports/evm-api.h"

#include "exports/anpssc-api.h"
#include "exports/ncm-api.h"
#include "exports/ccm-api.h"
#include "exports/rpc.h"    // Because mount.h needs it that's why
#include "exports/ea-api.h"
#include "exports/mount.h"
#include "exports/elog-api.h"

#include "ccm-loc.h"

static fail_list_t *failed_vols = NULL;

/*
	We have two classes of failed volumes:
	1) On failed_vols, these have gone offline due to EVM or storage failures.
	   We attempt to put these back online as indicated below.
	These LVs will have fail_type set recoverable_failure.
	2) On perm failure, these volumes have suffered terminal failure either:
	   a) due to conflicts in their management area.
	   b) Multiple failures within a certain time frame.
	These LVs will have fail_type set permanent_failure.
	In either case, the NFS sharing will be removed to stop the file system
	from receiving requests on that volume.

	LVs on failed_vols will attempt to be put back online when:
	   a) The file system posts FSYSABORT without a corresponding EVM error.
	   b) After EVM indicates the it is visible again.
 */
#define WAIT_FOR_FSYSABORT	40	// File system aborts I/O 30 secs after lv
								// fail. Wait an extra 10 secs.
#define WAIT_FOR_EVMFAIL	2	// EVM should fail the LV before the file
								// system aborts it but just in case there's
								// some message reseq, we'll briefly hold off
#define RETRY_HOLDOFF		60	// Wait for things to settle down first.
#define MAX_FAILURES		10	// Allow this many failures within
#define FAILURE_WINDOW		600	// this period before permanent failure

static void
add_to_list (fail_list_t **list, fail_list_t *vol)
{
	if (*list) {
		(*list)->prev = vol;
	}
	vol->next = *list;
	*list = vol;
}

static void
remove_from_list (fail_list_t **list, fail_list_t *vol)
{
	if (vol->prev) {
		vol->prev->next = vol->next;
	}
	else {
		*list = vol->next;
	}
	if (vol->next) {
		vol->next->prev = vol->prev;
	}
}

static fail_list_t *
add_failed_to_list (evm_volId_t vol)
{
	fail_list_t *deadfred;

	deadfred = malloc(sizeof(fail_list_t));
	memset (deadfred, 0, sizeof(fail_list_t));

	deadfred->lv = vol;
	deadfred->first_fail = time (NULL);
	add_to_list (&failed_vols, deadfred);

	return deadfred;
}

static char *
find_lv_failed_node (evm_volId_t vol)
{
	char **clus_lv;
	char *lv_key;
	char *node = NULL;

	asprintf (&lv_key, "cluster.lv.%016llx", vol.whole_thing);

	clus_lv = get_cfg_val (lv_key);
	free (lv_key);

	if (clus_lv) {
		node = strdup (index (clus_lv[1], ':') + 1);
		free_cfg_val (clus_lv);
	}

	return node;
}

extern int
have_failed_lvs (void)
{
	return failed_vols != NULL;
}

static fail_list_t *
lv_on_failed_list (evm_volId_t vol)
{
	fail_list_t *look;

	for (look = failed_vols; look; look = look->next) {
		if (vol.whole_thing == look->lv.whole_thing) {
			break;
		}
	}

	return look;
}

extern int
lv_has_failed (evm_volId_t vol)
{
	fail_list_t *look = lv_on_failed_list (vol);

	return look && (look->fail_type != not_failed);
}

extern int
lv_can_be_online (evm_volId_t vol, char *node)
{
	fail_list_t *it;
	int ok = 0;
	char **lv_obj;
	char *lv_failed;

/*
 * Make sure it's OK to put this LV in service on this node
 */
	asprintf (&lv_failed, "cluster.lv.%016llx", vol.whole_thing);
	lv_obj = get_cfg_val (lv_failed);
	free (lv_failed);

	if (lv_obj) {
		asprintf (&lv_failed, "failed:%s", node);
		ok = !strcasecmp (lv_obj[1], lv_failed);
		free (lv_failed);
		free_cfg_val (lv_obj);
/*
 * If it had failed on this node, it can only be put back online once
 * the timer has expired.
 */
		if (ok) {
			it = lv_on_failed_list (vol);

			if (it) {
				time_t now = time (NULL);

				it->evm_OK = 1;		// Mark the fact that EVM says its ready
				ok = now > it->can_continue;
			}
		}
	}

	return ok;
}

extern void
lv_going_online (evm_volId_t vol)
{
	fail_list_t *it;

	it = lv_on_failed_list (vol);

	if (it) {
		it->can_continue = 0;
		it->evm_OK = 1;
		it->fail_type = not_failed;	
		it->got_fsysabort = 0;
	}
}

extern void
remove_failed_lv (evm_volId_t vol)
{
	fail_list_t *it;

	it = lv_on_failed_list (vol);

	if (it) {
		remove_from_list (&failed_vols, it);
		free (it);
	}
}

extern void
lv_just_failed (nfx_eventMsg_t *evt)
{
	char **cfg_obj;
	char *key;
	char *nxt;
	char *path;
	char *laddr;
	uint64 msg_buf[256];
	ncmMsg_t *ncm_msg = (ncmMsg_t *)msg_buf;
	mnt_infoReq_t *umnt_msg = (mnt_infoReq_t *)ncm_msg->appMsg;
	int sz;
	fail_list_t *failed_lv;
	int perm_dead;
/*
	If it has just failed, keep track of it so we can later restore it
 */
	if (!(failed_lv = lv_on_failed_list(evt->param.volEvent.volId))) {
		E_LOG (class_7, info_s, myappid, 0, 0, 0,
		       ("New vol fail reported for %016llx on %s", 
		        evt->param.volEvent.volId.whole_thing, evt->nodeName));
		        
		failed_lv = add_failed_to_list (evt->param.volEvent.volId);
	}
	else {
		E_LOG (class_7, info_s, myappid, 0, 0, 0,
		       ("Another (%d) vol fail reported for %016llx on %s", 
		        failed_lv->failure_count,
		        evt->param.volEvent.volId.whole_thing, evt->nodeName));
	}
	failed_lv->failure_count++;
	failed_lv->can_continue = time (NULL);
/*
	In case we get an EVM failure AFTER the FSYSABORT, we
	can attempt to restart at the next visible event. Otherwise we give
	the other (EVM or FS) one the opportunity to chime in before allowing 
	the visible event to restart the LV.
 */
	if (evt->param.volEvent.event != EVENT_VOL_FSYSABORT) {
		if (!failed_lv->got_fsysabort) {
			failed_lv->can_continue += WAIT_FOR_FSYSABORT;
		}
		failed_lv->evm_OK = 0;
	}
	else {
		failed_lv->can_continue += WAIT_FOR_EVMFAIL;
		failed_lv->got_fsysabort = 1;
		if (failed_lv->fail_type == not_failed) {
			failed_lv->evm_OK = 1;	// EVM still thinks the volume good
		}
	}
/*
	Permanent failures are just that: permanent
	Notice the one-shot latch.
 */
	perm_dead = ((evt->param.volEvent.event == EVENT_VOL_CONFLICT) ||
	             (evt->param.volEvent.event == EVENT_VOL_CORRUPT));
	if ((failed_lv->fail_type != permanent_failure) && !perm_dead) {
		failed_lv->fail_type = recoverable_failure;
	}
	else {
		failed_lv->fail_type = permanent_failure;
	}

/*
	Now, let's mark it as failed and remove it's export(s)
 */
	cfg_obj = malloc (3 * sizeof(char *));
	asprintf (&cfg_obj[0], "cluster.lv.%016llx", 
	          evt->param.volEvent.volId.whole_thing);
	asprintf (&cfg_obj[1], "failed:%s", evt->nodeName);
	cfg_obj[2] = NULL;
	put_cfg_val (cfg_obj);
	sync_SCC (cfg_obj);
	free_cfg_val (cfg_obj);

	asprintf (&key, "cluster.nfs.%016llx", 
	          evt->param.volEvent.volId.whole_thing);

	cfg_obj = get_cfg_val (key);
	free (key);

	if (cfg_obj) { 		// Was exported
		ncm_msg->msgType = NCM_SEND;
		ncm_msg->rqResp = 0;
		strcpy (ncm_msg->appName, "mountd");
		strcpy (ncm_msg->nodeName, evt->nodeName);

		umnt_msg->msgHdr.msgType = MNT_UNEXPORT_REQ;
		umnt_msg->optionsLen = 0;

		for (nxt = cfg_obj[1]; nxt;) {
			path = strsep (&nxt, ";");
			laddr = strsep (&path, ":");
			*index (path, '|') = '\0';

			sz = (char *)umnt_msg - (char *)ncm_msg + 
			     sizeof (mnt_infoReq_t) + strlen (path) + 1;

			strcpy (umnt_msg->pathName, path);
			umnt_msg->nameLen = strlen (path);

			umnt_msg->localIpAddr = strtoul (laddr, NULL, 16);

			E_LOG (class_7, debug_s, myappid, 0, 0, 0,
			       ("Removed export: %s on %s", path, evt->nodeName));

			sendAgileMsg (-1, ncm_msg, sz,
			              "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
			              NULL, NULL, 0);
		}

		free_cfg_val (cfg_obj);
	}
}

extern void
check_failed_lvs (void)
{
	fail_list_t *look;
	time_t now = time (NULL);

	for (look = failed_vols; look;) {
/*
	Permanent failures are just that. They cannot be automatically restarted.
 */
		if (look->fail_type == recoverable_failure) {
/*
		Check if we got a visible event we could not act on AND
		if we've waited long enough for the FSYSABORT 
		after EVM failed the volume OR
		we waited long enough after getting the FSYSABORT for EVM 
		to tell us of the failed volume
		Both cases we had set can_continue to the time after which we'll
		attempt to put the LV back in service.
 */
			if (now >= look->can_continue) {
				if (look->evm_OK) {
					char *on_node = find_lv_failed_node (look->lv);
					if (on_node) {
						fake_vis_on_new_node (on_node, look->lv);
						free (on_node);
						look->evm_OK = 0;
					}
				}
			}
		}
/*
		Check to see if this one has exceeded its failure quota
 */
		if ((now - look->first_fail) <= FAILURE_WINDOW) {
			if (look->failure_count > MAX_FAILURES) {
				look->fail_type = permanent_failure;
			}
		}

		else if (look->fail_type == not_failed) {
			// Window expired so remove it from the list
			fail_list_t *whack = look;
			remove_from_list (&failed_vols, look);
			look = look->next;	// Gotta do this prior to returning
			free (whack);		// the heap
			continue;			// Bypass the statement below
		}
		look = look->next;	// Obviously unable to be put in for() statement
	}
}

/*
 * For some reason, EA failed to allow this LV to be put online and mounted.
 * We don't know why so we'll put it back in a recoverable state and try
 * a little later (once EVM again says it's OK).
 */
extern void
lv_failed_movin (evm_volId_t lv)
{
	fail_list_t *it = lv_on_failed_list (lv);

	if (it) {
		it->fail_type = recoverable_failure;
		it->evm_OK = 0;
		it->can_continue = time (NULL) + RETRY_HOLDOFF;
	}
}
