/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ccm-take.c
 *
 * RCS:  $Id: ccm-take.c,v 1.7 2002/10/04 02:17:06 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Contains the node take over stuff.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 10 Apr 2002
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "nfx-incl.h"

#include "exports/anpssc-api.h"
#include "exports/evm-api.h"
#include "exports/elog-api.h"

#include "ncm-api.h"
#include "ccm-api.h"
#include "ccm-loc.h"

/*
	Each node in the process of takeover will have one of these on the list
 */

typedef struct ip_list {
	char			*if_name_info;
	uint32			ip_addr;
	struct ip_list	*next;
} ip_list_t;

typedef struct vol_list {
	evm_volId_t		vol_id;
	char			*nfs_opts;
	struct vol_list	*next;
} vol_list_t;

#define TO_NODE_MAGIC 0xbabebeef
typedef struct to_node {
	uint32			magic;
	char			*new_node;
	char			*old_node;
	ip_list_t		*ip_addrs;
	vol_list_t		*vol_services;
	char			*rq_app;		// Who requested the takeover
	int				rq_cpu;			// so we can send them status
	int				rq_slot;		// when takeover is complete
	int				rq_plane;
	uint64			client;
	struct to_node	*prev;
	struct to_node	*next;
} to_node_t;

static __inline__ void free_to_node (to_node_t *n) 
{
	ip_list_t *i;
	vol_list_t *v;

	if (n->new_node) {
		free (n->new_node);
	}
	if (n->old_node) {
		free (n->old_node);
	}
	if (n->rq_app) {
		free (n->rq_app);
	}
	while (n->ip_addrs) {
		i = n->ip_addrs;
		n->ip_addrs = n->ip_addrs->next;
		free (i);
	}
	while (n->vol_services) {
		v = n->vol_services;
		n->vol_services = n->vol_services->next;
		if (v->nfs_opts) {
			free (v->nfs_opts);
		}
		free (v);
	}

	free (n);
}

extern void
fake_vis_on_new_node (char *new, evm_volId_t lv)
{
	char **names;
	char **look;
	char lvid_str[17];

	uint64 buf[256];
	ncmMsg_t *to_pcc = (ncmMsg_t *)buf;
	nfx_eventMsg_t *vis_evt = (nfx_eventMsg_t *)(to_pcc->appMsg);
	int sz = (char *)vis_evt - (char *)to_pcc + sizeof(nfx_eventMsg_t);

	to_pcc->msgType = NCM_SEND;
	strcpy (to_pcc->appName, "ccm");
	strcpy (to_pcc->nodeName, "PCC");
	to_pcc->rqResp = 0;

	vis_evt->msgType = NFX_EVENT_POST;
	strcpy (vis_evt->nodeName, new);
	vis_evt->needAck = 0;
	vis_evt->status = NFX_EVENT_STATUS_OK;
	vis_evt->eventScope = NFX_EVENT_GLOBAL;
	vis_evt->eventType = NFX_EVENT_VOLUME;
	vis_evt->param.volEvent.event = EVENT_VOL_CREATE;
	vis_evt->param.volEvent.volId = lv;
	vis_evt->param.volEvent.state = EVENT_STATE_DOWN;
/*
	The visible event also has the volume name in it. Since we're faking
	this one, the real one MUST already be in the config database. Use it.
	If it's not there, we're really screwed. No way to recover.
 */
	sprintf (lvid_str, "%016llx", lv.whole_thing);
	names = get_cfg_val ("lvname.*");

	strcpy (vis_evt->param.volEvent.volName, lvid_str); // Make it something
	                                                    // in case of disaster

	for (look = names; *look; look +=2) {
		if (!strcmp (look[1], lvid_str)) {
			strcpy (vis_evt->param.volEvent.volName, 
			        look[0] + strlen ("lvname."));
			break;
		}
	}
	free_cfg_val (names);

	sendAgileMsg (-1, buf, sz,
	              "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
}

static to_node_t *the_list = NULL;	// These nodes in the process of take over

extern void
record_ip_taken (uint32 addr, char *old_node, char *new_node, char *port_info)
{
	char *cfg_obj[3];
	struct in_addr ip_addr;

// Record taken IP addresses as:
//              cluster.taken.<old_node>.ip.<addr>=<new_node>:<port>/<mask>

	asprintf (&cfg_obj[0], "cluster.taken.%s.ip.%08x", old_node, addr);
	asprintf (&cfg_obj[1], "%s:%s", new_node, port_info);
	cfg_obj[2] = NULL;

	put_cfg_val (cfg_obj);

	free (cfg_obj[0]);

// Rewrite the cluster.ip.<addr>=<new_node>:<port>/<mask>

	ip_addr.s_addr = addr;
	asprintf (&cfg_obj[0], "cluster.ip.%s", inet_ntoa (ip_addr));

	put_cfg_val (cfg_obj);

	free (cfg_obj[0]);
	free (cfg_obj[1]);
}

extern void
record_lv_taken (evm_volId_t id, char *old_node, char *new_node)
{
	char *cfg_obj[3];

// Record taken LVs as cluster.taken.<old_node>.lv.<vol_id>=<new_node>

	asprintf (&cfg_obj[0], "cluster.taken.%s.lv.%016llx", old_node, 
	          id.whole_thing);
	cfg_obj[1] = new_node;
	cfg_obj[2] = NULL;

	put_cfg_val (cfg_obj);

	free (cfg_obj[0]);

// Rewrite cluster.lv.<vol_id>=online:<new-node>

	asprintf (&cfg_obj[0], "cluster.lv.%016llx", id.whole_thing);
	asprintf (&cfg_obj[1], "online:%s", new_node);

	put_cfg_val (cfg_obj);

	free (cfg_obj[0]);
	free (cfg_obj[1]);
}

extern void
complete_takeover (ccmShutdown_Msg_t *msg)
{
	to_node_t *node;
	ip_list_t *this_ip;
	vol_list_t *this_lv;
	ccmRsp_Msg_t compl_rsp;
/*
	Make sure this is a rightous shutdown response to our command.
 */
	node = (to_node_t *)(uint32)(msg->hdr.clientPrivate);
	if (node->magic != TO_NODE_MAGIC) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("complete_takeover: invalid msg\n"));
		return;
	}

	if (node->prev) {						// Pull it off the list first
		node->prev->next = node->next;
	}
	else {
		the_list = node->next;
	}
	if (node->next) {
		node->next->prev = node->prev;
	}
/*
	The old node is shutdown. Change the old node's cluster configuration
	to show that it had those services (in order that they might be restored,
	someday).
 */
	for (this_ip = node->ip_addrs; this_ip; this_ip = this_ip->next) {

		record_ip_taken (this_ip->ip_addr, node->old_node, node->new_node, 
		                 index(this_ip->if_name_info, ':') + 1);
	}

	for (this_lv = node->vol_services; this_lv; this_lv = this_lv->next) {

		record_lv_taken (this_lv->vol_id, node->old_node, node->new_node);
/*
	Send the new node fake LV Visible events for the old node's LVs as if they
	just became visible on the new node to trigger their export et.al. on
	the new node. (They were already visible there or we wouldn't have
	gotten this far.)
 */
		fake_vis_on_new_node (node->new_node, this_lv->vol_id);
	}
/*
	Beauty! We've completed the takeover so respond to whomever started
	this whole mess.
 */
	compl_rsp.hdr.msgType = ccmStatRsp;
	compl_rsp.hdr.clientPrivate = node->client;
	compl_rsp.status = NFX_OK;	// Complete

	sendAgileMsg (-1, &compl_rsp, sizeof(compl_rsp),
	              node->rq_app, node->rq_cpu, node->rq_slot, node->rq_plane,
	              NULL, NULL, 0);

	E_LOG (class_1, debug_s, myappid, 0, 0, 0,
	       ("Takeover of %s completed by %s", node->old_node, node->new_node));
/*
 * We need to take this node off the "shutting down" list since we overloaded
 * (it's doing the same thing really) the "take" part of the takeover onto
 * the shutdown method.
 */
	update_pcc_cfg ("", "", node->old_node);

	free_to_node (node);
	
	return;
}

/*
	Get a list of LVs that 'node' is exporting and their NFS export opts
 */
static vol_list_t *
get_exp_vols (char *node, char **cfg)
{
	vol_list_t *vols = NULL;
	vol_list_t **vp;
	char **look;
	char *match_this;
	int sz;

	sz = strlen("slave.") + strlen(node) + 1;
	asprintf (&match_this, "%s%s.", "slave.", node);

	vp = &vols;
	for (look = cfg; *look; look += 2) {
/*
	'node' was serving an LV if the operational state was not just visible
 */
		if (!strncasecmp (look[0], match_this, sz)) {
			if (strcmp (look[1], "visible")) {
				*vp = malloc (sizeof(vol_list_t));
				(*vp)->vol_id.whole_thing = strtoq (look[0]+sz, NULL, 16);
				(*vp)->nfs_opts = strdup (look[1]);
				(*vp)->next = NULL;
				vp = &((*vp)->next);
			}
		}
	}
	
	free (match_this);
	return vols;
}

static ip_list_t *
get_ip_addrs (char *node, char **cfg)
{
	ip_list_t *ips = NULL;
	ip_list_t **pip;
	char **look;
	char *match_node;
	int sz;
	int tsz;

	sz = strlen (node) + 1;
	match_node = malloc (sz + 1);
	sprintf (match_node, "%s:", node);

	tsz = strlen ("cluster.ip.");
/*
	The IP addresses assigned to a node will show up in the "cluster.ip.*"
	objects. The value is "node:port".
 */
	pip = &ips;
	for (look = cfg; *look; look += 2) {
		if (!strncasecmp (look[0], "cluster.ip.", tsz)) {
			if (!strncasecmp (look[1], match_node, sz)) {
				*pip = malloc (sizeof(ip_list_t));
				(*pip)->if_name_info = strdup (look[1]);
				(*pip)->ip_addr = inet_addr (look[0]+tsz);
				(*pip)->next = NULL;
				pip = &((*pip)->next);
			}
		}
	}

	return ips;
}

static int
new_node_unable (char *node, char **cfg, vol_list_t *vols)
{
	vol_list_t *this;
	char **look;
	char *lvid;
	char *key;
	int key_sz;

	key_sz = strlen("slave.") + strlen (node) + 17;	// trailing '.LV
	key = malloc (key_sz + 1);	// and \0
	sprintf (key, "slave.%s.", node);
	lvid = key + strlen(key);
/*
	As we look through each volume served by the old node,
	we need to make sure the new node can see it.
 */
	for (this = vols; this; this = this->next) {
		sprintf (lvid, "%016llx", this->vol_id.whole_thing);

		for (look = cfg; *look; look +=2) {
			if (!strcasecmp (look[0], key)) {
				if (!strcasecmp (look[1], "visible")) {
					break;
				}
			}
		}

		if (!(*look)) {	// Didn't find that one visible
			free (key);
			return 1;
		}
	}

	free (key);
	return 0;
}

extern void
takeover_node_services (int sock, char *app, int CPU, int Slot,
int Plane, ccmTakeOverRq_t *msg)
{
	char **admin_state;
	char **oper_state;
	vol_list_t *vol_list;
	to_node_t *new_to = NULL;
	char *err_str;
	uint64 buf[256];
	ccmRsp_Msg_t *rsp = (ccmRsp_Msg_t *)buf;
	ccmShutdown_Msg_t *shut_msg = (ccmShutdown_Msg_t *)buf;
	int return_status = NFX_OK;

	admin_state = get_cfg_val ("cluster.*");
	oper_state = get_cfg_val ("slave.*");

	if (admin_state && oper_state) {
/*
	Collect the list of cluster services the shutting down node is assigned.
 */
		err_str = NULL;	// No error so far
		vol_list = get_exp_vols (msg->old_node, oper_state);
		if (vol_list) {	// Stuff to do

			new_to = malloc (sizeof(to_node_t));
			memset (new_to, 0, sizeof(to_node_t));

			new_to->magic = TO_NODE_MAGIC;
			new_to->rq_app = strdup (app);
			new_to->rq_cpu = CPU;
			new_to->rq_slot = Slot;
			new_to->rq_plane = Plane;
			new_to->client = msg->hdr.clientPrivate;

			new_to->old_node = strdup (msg->old_node);
			new_to->new_node = strdup (msg->new_node);

			new_to->vol_services = vol_list;
			new_to->ip_addrs = get_ip_addrs (msg->old_node, admin_state);
/*
	Can the new node handle the old nodes LVs? (Are they visible?)
 */
			if (new_node_unable (new_to->new_node, oper_state, vol_list)) {
				free_to_node (new_to);
				new_to = NULL;
				err_str = "New node unable to serve all volumes";
				return_status = NFX_ERR;
			}
		}
		else {
			err_str = "No volumes served by old node";
		}

		free_cfg_val (admin_state);
		free_cfg_val (oper_state);
	}
	else {
		err_str = "Node not assigned cluster services";
	}
/*
	If we added a new takeover, then send a shutdown request for the old node
 */
	if (new_to) {	// Link it up and send us shutdown rq
		new_to->next = the_list;
		if (the_list) {
			the_list->prev = new_to;
		}
		the_list = new_to;

		shut_msg->hdr.msgType = ccmShutDown;
		shut_msg->hdr.clientPrivate = (int)new_to;
		shut_msg->done = 0;
		strcpy (shut_msg->fromNode, msg->old_node);

		sendAgileMsg (sock, shut_msg, sizeof (ccmShutdown_Msg_t),
		              "ccm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
		              NULL, NULL, 0);

		E_LOG (class_4, debug_s, myappid, 0, 0, 0,
		       ("Initiated shutdown of %s for takeover by %s", msg->old_node,
		        msg->new_node));
	}
	else {	// Something might not be quite right
		int sz;

		rsp->hdr.msgType = ccmStatRsp;
		rsp->hdr.clientPrivate = msg->hdr.clientPrivate;
		rsp->status = return_status;
		sz = sizeof (ccmRsp_Msg_t);

		if (err_str) {
			strcpy ((char *)rsp + sz, err_str);
			sz += strlen (err_str) + 1;
		}

		sendAgileMsg (-1, rsp, sz, app, CPU, Slot, Plane, NULL, NULL, 0);

		E_LOG (class_4, alert_s, myappid, 0, 0, 0,
		       ("Takeover request unable to execute from %s to %s because %s",
		        msg->old_node, msg->new_node, err_str));
	}
	
	return;
}
