/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ncm.c
 *
 * RCS:  $Id: ncm.c,v 1.32 2002/10/08 22:45:37 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Node Connection Manager. Maintains intra-cluster node
 *              connections via TCP.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Feb 2002
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include <neteee/eee.h>

#include "exports/anpssc-api.h"
#include "exports/elog-api.h"
#include "exports/event-api.h"

#include "ncm-api.h"
#include "ncm-loc.h"

static ncmLink_t	*nodes;	// Node(s) we're connected to

static int appSock;		// Our registered socket
static int lisSock;		// What we listen to

static char myName[256];		// What our hostname is
uint32 myappid;

#define MAXFD(x, max) (((x) + 1) > max) ? ((x) + 1) : max

/*
	These guys are here to tell folks of the comings and goings of nodes
 */
static void
del_node_event (ncmLink_t *n)
{
	ncmnodeEvent_t evt_msg;

	evt_msg.msgType = NCM_NODE_EVENT;
	evt_msg.new = 0;
	strcpy (evt_msg.nodeName, n->nodeName);

	sendAgileMsg (-1, &evt_msg, sizeof(evt_msg),
	              "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
}

static void
new_node_event (ncmLink_t *n)
{
	ncmnodeEvent_t evt_msg;

	evt_msg.msgType = NCM_NODE_EVENT;
	evt_msg.new = 1;
	strcpy (evt_msg.nodeName, n->nodeName);

	sendAgileMsg (-1, &evt_msg, sizeof(evt_msg),
	              "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
	sendAgileMsg (-1, &evt_msg, sizeof(evt_msg),
	              "ccm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
}

/*
	This is how we acknowledge and resend messages over the ncm links
 */

typedef struct dNP {
	time_t			died;
	char			*name;
	struct ncmAck	*tbs;	// Pending messages - to be sent
	struct dNP		*next;
} deadNode_t;

#define MAXMSGBUF	65535
#define PEND_HOW_LONG 60	// Longer than a minute and the node probably died
static deadNode_t *deadNodeswPending = NULL;

static void
grim_reaper (void)
{
	struct ncmAck *msgs;
	deadNode_t *adieu, *prev;
	time_t now = time (NULL);
/*
	Look at all the nodes that have died and discard those dead for too long.
	It shouldn't take but a couple of seconds for a node to reconnect via its
	redundant management link.
 */
	for (prev = NULL, adieu = deadNodeswPending; adieu;) {
		if ((now - adieu->died) > PEND_HOW_LONG) {
			deadNode_t *nxtdn = adieu->next;
			if (prev) {
				prev->next = nxtdn;
			}
			else {
				deadNodeswPending = nxtdn;
			}

			for (msgs = adieu->tbs; msgs;) {
				struct ncmAck *nxt = msgs->next;

				free (msgs->msg);
				free (msgs);
				msgs = nxt;
			}

			free (adieu->name);
			free (adieu);
			adieu = nxtdn;
		}
		else {
			prev = adieu;
			adieu = adieu->next;
		}
	}
}

static void
enqueue_for_ack (ncmLink_t *node, ncmMsg_t *msg, int size)
{
	struct ncmAck *anew1;
/*
	Track this message. It'll get removed and free()ed when It's acknowledged.
	We don't timeout/retry because that's what TCP does for us. This list
	is here in case we have to reconnect on our redundant link.
 */
	anew1 = malloc (sizeof(struct ncmAck));
	anew1->msgSeq = msg->rqResp & SEQ_MASK;
	anew1->msg = msg;
	anew1->msgSize = size;
	anew1->next = node->needsAck;
	node->needsAck = anew1;
}

static void
got_ack (ncmLink_t *node)
{
	int seq;
	struct ncmAck *look, *prev;
/*
	We've gotten an ACK, read the sequence number of the message being ACKed
	and remove it from the pending list returning its heap.
 */
	read (node->s, &seq, sizeof(seq));

	for (prev = NULL, look = node->needsAck; look; look = look->next) {
		if (look->msgSeq == seq) {
			break;
		}
		prev = look;
	}

	if (look) {
		if (prev) {
			prev->next = look->next;
		}
		else {
			node->needsAck = look->next;
		}

		free (look->msg);
		free (look);
	}
	else {
		E_LOG (class_1, warning_s, myappid, 0, 0, 0,
		       ("Extra ACK from %s", node->nodeName));
	}
}

static void
ack_this (int sock, int seq)
{
	int ackmsg[2];
/*
	Just got this message so send ACK to the other end
*/
	ackmsg[0] = FRAMETAG | ACKMAGIC;
	ackmsg[1] = seq;

	write (sock, ackmsg, sizeof (ackmsg));
}

/* Resend any messages on node->needsAck since we just reconnected */

static void
mk_next_seq (ncmLink_t *node, ncmMsg_t *msg)
{
	node->ackSeq = (node->ackSeq + 1) & SEQ_MASK;
	msg->rqResp &= RESPONSE_MASK;
	msg->rqResp |= node->ackSeq;
}

static void
resend_unacked (ncmLink_t *node)
{
	struct ncmAck *snd;

	for (snd = node->needsAck; snd; snd = snd->next) {
		mk_next_seq (node, snd->msg);
		snd->msgSeq = snd->msg->rqResp & SEQ_MASK;

		write (node->s, snd->msg, snd->msgSize);
	}
}

static void
save_unacked (ncmLink_t *node)
{
	deadNode_t *ick;
/*
	Link died. If it had unACKed messages, save them for a bit in case we
	can connect via the redundant link.
 */
	if (node->needsAck) {
		ick = malloc (sizeof(deadNode_t));
		ick->died = time (NULL);
		ick->name = strdup (node->nodeName);
		ick->tbs = node->needsAck;
		ick->next = deadNodeswPending;
		deadNodeswPending = ick;
	}
}

/*
 * New connection. Look for node->nodeName on deadNodeswPending and
 * resend any messages enqueued there
 */
static void
resend_pending (ncmLink_t *node)
{
	deadNode_t *look, *prev;
	
	for (prev = NULL, look = deadNodeswPending; look; look = look->next) {
		if (!strcasecmp (node->nodeName, look->name)) {
			break;
		}
		prev = look;
	}

	if (look) {
		if (prev) {
			prev->next = look->next;
		}
		else {
			deadNodeswPending = look->next;
		}
		node->needsAck = look->tbs;
		free (look->name);
		free (look);

		resend_unacked (node);
	}
}

/*-----------------------------------------------------------------
 * Name:clearNodeConn
 *
 * Description: Clear this nodes outstanding responses and close its socket
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static void
clearNodeConn (ncmLink_t *n, fd_set *fds)
{
	ncmRsp_t *l;

	close (n->s);
	while (n->outstanding) {
		l = n->outstanding;
		close (n->outstanding->rspSock);
		if (fds) {
			FD_CLR (n->outstanding->rspSock, fds);
		}
		n->outstanding = n->outstanding->next;
		free(l);
	}
}

/*-----------------------------------------------------------------
 * Name:connToPCC
 *
 * Description: Connect to our cluster's PCC via the given IP address
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int
connToPCC (in_addr_t pccAddr)
{
	int s;
	struct sockaddr_in pcc;
	struct protoent *proto;
	struct servent *serv;

	if (!pccAddr) { // Really gotta have an address
		return -1;
	}

	proto = getprotobyname("tcp");
	if (!proto) {
		perror ("connToPCC: getprotobyname failed");
		return -1;
	}

	s = socket (AF_INET, SOCK_STREAM, proto->p_proto);

	if (s < 0) {
		perror ("connToPCC: socket failed");
	}
	else {
		serv = getservbyname ("ncm", "udp");
		if (!serv) {
			perror ("connToPCC: getservbyname failed");
			close (s);
			return -1;
		}

		memset (&pcc, 0, sizeof(pcc));
		pcc.sin_family = AF_INET;
		pcc.sin_port = serv->s_port;
		pcc.sin_addr.s_addr = pccAddr;

		if (connect (s, (struct sockaddr *)&pcc, sizeof(pcc))) {
			close (s);
			s = -1;
		}
	}
	return s;
}

static int
establish_pcc_conn (ncmLink_t *pcc, fd_set *fds, int *maxfds)
{
	struct in_addr pcc_addr;

	pcc->lastTime = time (NULL);
	pcc->pingSent = 0;

	pcc->usingIF = (pcc->usingIF + 1) % 2;	// Try the other one now
	pcc->s = connToPCC (pcc->lnkAddr[pcc->usingIF]);

	if (pcc->s < 0) {
		pcc->usingIF = (pcc->usingIF + 1) % 2;	// Last try
		pcc->s = connToPCC (pcc->lnkAddr[pcc->usingIF]);
	}

	if (pcc->s < 0) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0, ("unable to connect to PCC"));
		return 0;
	}
	else {
		FD_SET (pcc->s, fds);
		*maxfds = MAXFD (pcc->s, *maxfds);

		pcc_addr.s_addr = pcc->lnkAddr[pcc->usingIF];
		E_LOG(class_1, info_s, myappid, 0, 0, 0,
		      ("Connect to PCC at %s", inet_ntoa (pcc_addr)));
		return 1;
	}

}

/*
	Node no longer a cluster member so disconnect it
 */
static void
revoke_connection (char *node_name, fd_set *fds)
{
	ncmLink_t *look;
	ncmLink_t *l;

	for (l = NULL, look = nodes; look; l = look, look = look->next) {
		if (!strcasecmp (look->nodeName, node_name)) {
			break;
		}
	}
/*
 * We NEVER want to tear down the local connections. This MIGHT happen when
 * the PCC connection fails before we see the PCC down event. This only
 * came about after we changed it to restore the local_rx connection back
 * to PCC when we lose the PCC.
 */
	if (look && (look->lnkAddr[0] != inet_addr ("127.0.0.1"))) {
		FD_CLR (look->s, fds);
		clearNodeConn (look, fds);
		del_node_event (look);
		if (l) {
			l->next = look->next;
		}
		else {
			nodes = look->next;
		}
		free (look->nodeName);
		free (look);
	}
}

static void
restore_local_to_PCC (void)
{
	ncmLink_t *look;

	for (look = nodes; look; look = look->next) {
		if (!strcasecmp ("local_rx", look->nodeName)) {
			free (look->nodeName);
			look->nodeName = strdup ("PCC");
			break;
		}
	}
}

static void
replace_pcc (nfx_eventMsg_t *msg, fd_set *fds, int *maxfds)
{
	ncmLink_t *look;
	ncmLink_t *new_node = NULL;

	E_LOG (class_1, debug_s, myappid, 0, 0, 0,
	       ("replace_pcc: %s @ %08x %08x state %d", 
	        msg->nodeName,
			msg->param.nodeEvent.mgmt_ip[0], msg->param.nodeEvent.mgmt_ip[1],
	        msg->param.nodeEvent.state));
/*
	It could be our PCC failed and LNM is telling us of a new one or
	we're replacing the original receive side of our local loopback.
	We need to create another node for the "real" PCC.
 */
	if (msg->param.nodeEvent.state == EVENT_STATE_UP) {

		for (look = nodes; look; look = look->next) {
			if (!strcasecmp ("PCC", look->nodeName)) {
				break;
			}
		}

		if (!look || (look->lnkAddr[0] == inet_addr ("127.0.0.1"))) {
			new_node = malloc (sizeof(ncmLink_t));
			memset (new_node, 0, sizeof(ncmLink_t));
			new_node->nodeName = strdup ("PCC");
			new_node->next = nodes;
			nodes = new_node;

			if (look) {
				free (look->nodeName);
				look->nodeName = strdup ("local_rx");
			}
			look = new_node;
		}
		else {
			// Just replace our existing PCC
			FD_CLR (look->s, fds);
			clearNodeConn (look, fds);
			del_node_event (look);
		}

		memcpy (look->lnkAddr, msg->param.nodeEvent.mgmt_ip, 
		        sizeof(look->lnkAddr));

		look->usingIF = 1; 	// Start at the first one

		if (establish_pcc_conn (look, fds, maxfds)) {
			new_node_event (look);
		}
		else { // Restore local_rx because we really can't connect
			if (new_node) {
				nodes = new_node->next;
				free (new_node->nodeName);
				free (new_node);
			}
			restore_local_to_PCC();
		}
	}
	else {	// He's dead, Jim.
		revoke_connection ("PCC", fds);
		restore_local_to_PCC();
	}
}

/*
	Nodes that are currently members of our cluster
 */
typedef struct clusMemb {
	char			*name;
	in_addr_t		mgmt_ip[2];
	struct clusMemb	*prev;
	struct clusMemb	*next;
} clusMemb_t;

static clusMemb_t *memb_list = NULL;

static __inline__ void
free_memb (clusMemb_t *memb)
{
	if (memb) {
		if (memb->name) {
			free (memb->name);
		}
		free (memb);
	}
}

/*
	Got a message to either add or delete a node from our cluster
 */
static int
check_membership (nfx_eventMsg_t *msg, fd_set *fds)
{
	if (msg->param.nodeEvent.state == EVENT_STATE_DOWN) {
		clusMemb_t *look;
		for (look = memb_list; look; look = look->next) {
			if (!strcasecmp (msg->nodeName, look->name)) {
				break;
			}
		}

		if (!look) {
			return 1;
		}
		else {
			E_LOG (class_1, debug_s, myappid, 0, 0, 0,
			       ("Removing node: %s @ %08x %08x", look->name, 
			        look->mgmt_ip[0], look->mgmt_ip[1]));

			revoke_connection (msg->nodeName, fds);

			if (!look->prev) {
				memb_list = look->next;
			}
			else {
				look->prev->next = look->next;
			}
			if (look->next) {
				look->next->prev = look->prev;
			}
			free_memb (look);
		}
	}
	else {
		clusMemb_t *newMemb;

		newMemb = malloc (sizeof (clusMemb_t));
		newMemb->name = strdup (msg->nodeName);
		memcpy (newMemb->mgmt_ip, msg->param.nodeEvent.mgmt_ip, 
		        sizeof(newMemb->mgmt_ip));
		newMemb->next = memb_list;
		newMemb->prev = NULL;
		if (memb_list) {
			memb_list->prev = newMemb;
		}
		memb_list = newMemb;
		E_LOG (class_1, debug_s, myappid, 0, 0, 0,
		       ("Added new node: %s @ %08x %08x", newMemb->name, 
		        newMemb->mgmt_ip[0], newMemb->mgmt_ip[1]));
	}
	return 0;
}

static void
send_node_list (char *to_whom)
{
	uint64 buf[1024];
	ncmNodeList_t *node_list_msg = (ncmNodeList_t *)buf;
	ncmLink_t *look;
	int sz;

	node_list_msg->msgType = NCM_NODE_LIST;
	node_list_msg->numConnNodes = 0;

	for (look = nodes; look; look = look->next) {
		if (strcmp (look->nodeName, "local_rx")) {	// ignore this one
			strcpy (node_list_msg->nodeNames[node_list_msg->numConnNodes],
			        look->nodeName);
			node_list_msg->numConnNodes++;
		}
	}

	sz = sizeof(ncmNodeList_t) + ((node_list_msg->numConnNodes - 1) * MAXNAME);
	sendAgileMsg (-1, node_list_msg, sz,
	              to_whom, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
}

/*-----------------------------------------------------------------
 * Name: handleANP
 *
 * Description: Process datagram from ANPSSC network
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int
handleANP (int s, fd_set *fds, int *nfds)
{
	char app[MAXNAME];
	ncmMsg_t *lnkMsg;
	nfx_eventMsg_t *evt;
	int sndSize;
	int rc;
	ncmLink_t *look;
	int dealloc_here = 1;
/*
	Get the message off our registered socket
	We AssUMe these messages are ALWAYS from SSC Unix applications and
	don't track where they come from beyond application name.
 */
	lnkMsg = malloc (MAXMSGBUF);
	evt = (nfx_eventMsg_t *)lnkMsg;

	sndSize = recvAgileMsg (s, lnkMsg, MAXMSGBUF, NULL, app, NULL, NULL, NULL);
/*
	Is it an event or relay
 */
	switch (lnkMsg->msgType) {
		case NCM_SEND:
			lnkMsg->cpu = NFX_SSC_UNIX;
			lnkMsg->slot = SSC_SLOT;

		case NCM_SEND_EMB:
/*
		Our "real" name is an alias for the local node.
 */
			if (!strcasecmp (myName, lnkMsg->nodeName)) {
				strcpy (lnkMsg->nodeName, "local");
			}
/*
		If we're not connected to the destination node, reject the SEND
 */
			for (look = nodes; 
			     look && strcasecmp (look->nodeName, lnkMsg->nodeName);
			     look = look->next);

			if (!look) {	// No such node so send reject message
				ncmReject_t badDest;

				badDest.msgType = NCM_REJECT;
				badDest.reason = INVALID_DEST_NODE;
				sendAgileMsg (-1, &badDest, sizeof(badDest), 
				              app, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 
				              NULL, NULL, 0);
				E_LOG (class_1, debug_s, myappid, 0, 0, 0,
				       ("ncm reject: invalid destination node %s", 
				        lnkMsg->nodeName));
				rc = 1;
			}
			else {
/*
 *-----------------------------------------------------------------------
	FUTURE EXPANSION GOES HERE
	Log the message's destination node with the raw message
 *-----------------------------------------------------------------------
 */
#if defined (VERBOSE_DEBUG)
				E_LOG (class_6, debug_s, myappid, 0, 0, 0,
				       ("ncm send to %s at %s len %d from %s", lnkMsg->appName,
				        lnkMsg->nodeName, sndSize, app));
#endif
				         
/*
	Do the overloads for size and source application name
 */
				lnkMsg->msgType = FRAMETAG | (sndSize & COUNTMASK);
				strncpy (lnkMsg->SOURCEAPP, app, sizeof(lnkMsg->SOURCEAPP));
				if (lnkMsg->rqResp) {
					lnkMsg->rqResp = EXPECT_REPONSE;
				}
/*
	Add this to the list of messages sent. We'll get an ACK from the other end
	and then we can drop it. If we switch links, we'll use this list to
	resend un-ACKed messages.
 */
				mk_next_seq (look, lnkMsg);
				enqueue_for_ack (look, lnkMsg, sndSize);
				dealloc_here = 0;	// These message buffers get free()ed
									// after they're ACKed
/*
	Send it on its way
 */
				rc = write (look->s, lnkMsg, sndSize);

				if (rc != sndSize) {	// Detect bad write
					E_LOG (class_1, debug_s, myappid, 0, 0, 0,
					       ("Link write error. want %d sent %d", sndSize, rc));
				}
			}
			break;

		case NFX_EVENT_POST:
			switch (evt->eventType) {
				case NFX_EVENT_PCC:
					replace_pcc (evt, fds, nfds);
					break;

				case NFX_EVENT_SCC:
					//replace_scc (evt, fds, nfds);
					break;

				case NFX_EVENT_NODE:
					rc = 0;
					if (check_membership (evt, fds)) {
						E_LOG (class_1, debug_s, myappid, 0, 0, 0,
						       ("cluster membership err. Node %s", 
						        evt->nodeName));
						rc = 1;
					}
					break;
				default:
					E_LOG (class_1, err_s, myappid, 0, 0, 0,
					       ("Invalid event received %d\n", evt->eventType));
			}
			break;

		case NCM_RQ_NODE_LIST:
			send_node_list (app);
			break;

		default:
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
			       ("got invalid message id %d from %s\n",
			        lnkMsg->msgType, app));
	}
	if (dealloc_here) {
		free (lnkMsg);
	}
	return rc;
}

/*-----------------------------------------------------------------
 * Name: relayResponse
 *
 * Description: A local response needs to be relayed back to its
 *              sender on the other node.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Feb 2002
 *
 *-----------------------------------------------------------------
 */
#define LNKHDRSIZE (sizeof(ncmMsg_t) - sizeof(long long))
static int
relayResponse (ncmRsp_t *r)
{
	ncmMsg_t *lnkMsg;
	int sndSize;
	int rc;

	lnkMsg = malloc (MAXMSGBUF);
	memset (lnkMsg, 0, sizeof(ncmMsg_t));
/*
	Get the response off our ANP socket (datagram).
 */
	sndSize = recv (r->rspSock, lnkMsg->appMsg, MAXMSGBUF - LNKHDRSIZE, 0);
	close (r->rspSock);		// One shot deal
	if (sndSize <= 0) {
		// This is not a good thing
		return 1;
	}
#if defined (VERBOSE_DEBUG)
	E_LOG (class_6, debug_s, myappid, 0, 0, 0,
	       ("response size %d rem %s loc %s", sndSize, r->remApp, r->locApp));
#endif
/*
	Figure out where it needs to go and send it on its way
 */
	sndSize += LNKHDRSIZE;
	lnkMsg->msgType = FRAMETAG | sndSize;
	lnkMsg->rqResp = IS_RESPONSE;
	strncpy (lnkMsg->appName, r->remApp, sizeof(lnkMsg->appName));
	strncpy (lnkMsg->SOURCEAPP, r->locApp, sizeof(lnkMsg->SOURCEAPP));

	free (r->remApp);
	free (r->locApp);

	mk_next_seq (r->link, lnkMsg);
	enqueue_for_ack (r->link, lnkMsg, sndSize);
	rc = write (r->link->s, lnkMsg, sndSize);
/*
	Caller will deallocte *r
 */
	return (rc != sndSize);
}

/*-----------------------------------------------------------------
 * Name: handleIO
 *
 * Description: Got something from another node. What to do?
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int
handleIO (ncmLink_t *l, fd_set *set, int *mx)
{
	static long long buf[(COUNTMASK+1)/sizeof(long long)];	
	// long long to ensure 8 byte alignment and big enough for max msg
	ncmMsg_t *lnkMsg = (ncmMsg_t *)buf;
	int count;
	int rc;
	int sendFd;
	ncmRsp_t *t;
	int nrd;
	int n2rd;
	void *msg;
	int msgsize;
	char *locApp;
	char *rdhere;
/*
	First thing, make sure we're synced up by getting the frame delimter
	and message length.
 */
	rc = read (l->s, &(lnkMsg->msgType), sizeof(lnkMsg->msgType));
	if (rc != sizeof(lnkMsg->msgType)) {
		E_LOG (class_1, debug_s, myappid, 0, 0, 0,
		       ("link %s read error %d: %s", l->nodeName, rc, 
		        strerror(errno)));
		return -1;
	}

	if ((lnkMsg->msgType & TAGMASK) != FRAMETAG) {
		// Well now, this sucks
		E_LOG (class_1, notice_s, myappid, 0, 0, 0,
		       ("Lost sync"));
		// TODO - how to resync?
		return 1;
	}
	count = lnkMsg->msgType & COUNTMASK;
/*
	Is this a "real" message or a ping/ping response? Handle the pings
	right here.
 */
	if (count == PINGMAGIC) {
		int32 pingMsg = FRAMETAG | RESPMAGIC;
		// Return the response
		rc = write (l->s, &pingMsg, sizeof (pingMsg));
		return 0;
	}
	else if (count == RESPMAGIC) {
		return 0;
	}
	else if (count == ACKMAGIC) {
		got_ack (l);
		return 0;
	}
/*
	It's a real message, let's get the whole thing and see where it wants
	to go.
 */
	n2rd = count - sizeof(lnkMsg->msgType);
	rdhere = lnkMsg->appName;

	while (n2rd) {
		nrd = read (l->s, rdhere, n2rd);
		n2rd -= nrd;
		rdhere += nrd;
	}

	ack_this (l->s, lnkMsg->rqResp & SEQ_MASK);
/*
	Do they expect a response?
 */
	if ((lnkMsg->rqResp & RESPONSE_MASK) == EXPECT_REPONSE) {
		struct sockaddr_eee us;

		t = malloc (sizeof(ncmRsp_t));

		sendFd = socket (PF_EEE, SOCK_DGRAM, 0);

		us.seee_family = AF_EEE;
		us.seee_bplane = EEE_CONTROLBPLANE;
		us.seee_slot = SSC_SLOT;
		us.seee_cpu = NFX_SSC_UNIX;
		us.seee_app = 0;
		rc = bind (sendFd, (struct sockaddr *)&us, sizeof(us));

		t->rspSock = sendFd;
		t->link = l;
		t->remApp = strdup (lnkMsg->SOURCEAPP);
		t->locApp = strdup (lnkMsg->appName);
		t->next = l->outstanding;
		l->outstanding = t;
		t->sent = time (NULL);
/*
	Don't forget to update the select mask
 */
		FD_SET (sendFd, set);
		*mx = ((sendFd + 1) > *mx) ? (sendFd + 1) : *mx;
	}
	else {
		sendFd = -1;
	}
/*
 *-----------------------------------------------------------------------
	FUTURE EXPANSION GOES HERE
	Log the message's source node with the raw message
 *-----------------------------------------------------------------------
 */

/*
	In the event it's a response coming back from an app on the distant
	node, we need to send the entire link packet.
 */
	locApp = strdup (lnkMsg->appName);

	if ((lnkMsg->rqResp & RESPONSE_MASK) == IS_RESPONSE) {
		msg = lnkMsg;
		msgsize = count;
		strcpy (lnkMsg->appName, lnkMsg->SOURCEAPP);
		strcpy (lnkMsg->nodeName, l->nodeName);
		lnkMsg->rqResp = 0;
		lnkMsg->msgType = NCM_RESP;
	}
	else {
		msg = lnkMsg->appMsg;
		msgsize = count - LNKHDRSIZE;
	}

#if defined (VERBOSE_DEBUG)
	E_LOG (class_6, debug_s, myappid, 0, 0, 0,
	       ("IO from %s size %d to %s/C%d/S%d", l->nodeName,
	        msgsize, locApp, lnkMsg->cpu, lnkMsg->slot));
#endif

/*
	Fire it off to the application.
 */
	rc = sendAgileMsg (sendFd, msg, msgsize,
	     locApp, lnkMsg->cpu, lnkMsg->slot, MGMTPLANE,
	     NULL, NULL, 0);
	free (locApp);

	return rc;
}

/*-----------------------------------------------------------------
 * Name: nodeIsOK
 *
 * Description: Is this connecting node one we are allowed to listen to?
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int
nodeIsOK (ncmLink_t *n)
{
	clusMemb_t *look;
	int using_if;
/*
	Is it me on localhost?
 */
	if (n->lnkAddr[0] == inet_addr ("127.0.0.1")) {
		n->nodeName = strdup ("local");
		return 1;
	}
/*
	Trundle down the list of legal member nodes and see if its one of them
 */
	for (look = memb_list; look; look = look->next) {

		if (look->mgmt_ip[0] == n->lnkAddr[0]) {
			using_if = 0;
		}
		else if (look->mgmt_ip[1] == n->lnkAddr[0]) {
			using_if = 1;
		}
		else {
			continue;		// Not this one
		}
/*
	Found it!
 */
		n->nodeName = strdup (look->name);
		n->lnkAddr[1] = look->mgmt_ip[(using_if + 1) % 2];
		n->usingIF = 0;
		return 1;
	}

	return 0;
}

/*-----------------------------------------------------------------
 * Name: tryToAccept
 *
 * Description: Accept a connection creating its descriptor
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static ncmLink_t *
tryToAccept (int s)
{
	int newSock;
	struct sockaddr_in who;
	socklen_t whsz;
	ncmLink_t *node;

	whsz = sizeof(who);
	newSock = accept (s, (struct sockaddr *)&who, &whsz);

	if (newSock < 0) {
		perror ("tryToAccept: accept error");
		return NULL;
	}

	node = malloc (sizeof(ncmLink_t));

	if (node) {
		memset (node, 0, sizeof(ncmLink_t));

		node->s = newSock;
		node->lnkAddr[0] = who.sin_addr.s_addr;

		if (!nodeIsOK (node)) {
			shutdown (node->s, SHUT_RDWR);
			close (node->s);
			free (node);
			node = NULL;
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
			       ("Invalid connection from %s", inet_ntoa(who.sin_addr)));
		}
		else {
			E_LOG (class_1, debug_s, myappid, 0, 0, 0,
			       ("Accepted connection from %s@%s", node->nodeName,
			        inet_ntoa(who.sin_addr)));
		}
	}

	return node;
}

/*-----------------------------------------------------------------
 * Name: setUpListen
 *
 * Description: Listen for connections from our slaves (and us).
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 Feb 2002
 *
 *-----------------------------------------------------------------
 */
#define RESTART_WAIT 1
#define RESTART_TIMES 60
static int
setUpListen(int is_restart)
{
	struct protoent *proto;
	int tcpProto;
	struct servent *serv;
	int lisSock;
	struct sockaddr_in laddr;
	int try_count;
	int rc;
/*
	We listen for connections on the ncm port. We only accept those
	those connections if we're the PCC and the connecting node is known
	to be a slave in our cluster or it's us via localhost.
 */
	proto = getprotobyname("tcp");
	if (!proto) {
		perror ("setUpListen: getprotobyname failed");
		return -1;
	}
	tcpProto = proto->p_proto;

	lisSock = socket (AF_INET, SOCK_STREAM, tcpProto);
	if (lisSock < 0) {
		perror ("setUpListen: socket failed");
		return -1;
	}

	memset (&laddr, 0, sizeof(laddr));

	serv = getservbyname ("ncm", "udp");
	if (!serv) {
		perror ("setUpListen: getservbyname failed");
		return -1;
	}

	laddr.sin_port = serv->s_port;
	laddr.sin_family = AF_INET;
	laddr.sin_addr.s_addr = INADDR_ANY;
/*
	So as not to delay system restart while debugging, enable the immediate
	reuse of ncm's local address.
 */
	try_count = 1;
	if (setsockopt (lisSock, SOL_SOCKET, SO_REUSEADDR, 
	                &try_count, sizeof(try_count))) {
		perror ("setsockopt SO_REUSEADDR failed");
	}
/*
	If this is a restart, our listening port will be busy so we'll wait
	before declaring an error. (That's if the setsockopt() fails.)
 */
	try_count = 0;
	do {
		rc = bind (lisSock, (struct sockaddr *)&laddr, sizeof(laddr));
		if (is_restart) {
			try_count++;
			sleep (RESTART_WAIT);
		}
	} while (is_restart && (rc < 0) && (errno == EADDRINUSE) &&
	         (try_count < RESTART_TIMES));

	if (rc < 0) {
		perror ("setUpListen: bind failed");
		close (lisSock);
		lisSock = -1;
	}
	else {
		if (listen (lisSock, 5)) {
			perror ("setUpListen: listen failed");
			close (lisSock);
			lisSock = -1;
		}
	}

	return lisSock;
}

static ncmLink_t *
attempt_reconnect (ncmLink_t *danode, fd_set *fds, int *mxfd)
{
	ncmLink_t *bef;

	FD_CLR (danode->s, fds);
	clearNodeConn (danode, fds);
/*
	If the connection to our PCC choked, let's try to reconnect.
	Failing both addresses, we'll drop it entirely. Not a good thing
 */
	if (!strcasecmp (danode->nodeName, "PCC")) {
		if (establish_pcc_conn (danode, fds, mxfd)) {
			resend_unacked (danode);
			return NULL;	// Successful reconnect
		}
		else {	// We need to restore "local_rx" as PCC
			restore_local_to_PCC();
		}
	}

	save_unacked (danode);
/*
	Couldn't, or shouldn't, reconnect so remove it from the connected node list
 */
	del_node_event (danode);

	if (nodes == danode) {	// Obvious to the casual observer
		static ncmLink_t faker;	// Hidden from scope it's in the bss
		faker.next = nodes = danode->next;
		bef = &faker;		// Avert your eyes. It's ugly.
	}
	else {
		for (bef = nodes; bef->next != danode; bef = bef->next){}
		bef->next = danode->next;
	}

	free (danode->nodeName);
	free (danode);
/*
	We return the address of the node that preceeded the one we just deleted
	so the loop where we were called can proceed, sort of, normally
 */
	return bef;
}

/*-----------------------------------------------------------------
 * Name: main
 *
 * Description: Where it all happens
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Feb 2002
 *
 *-----------------------------------------------------------------
 */
extern int
main (int argc, char *argv[])
{
	fd_set ourFds;
	fd_set rdFds;
	int maxFd;
	int numFds;
	time_t now;
	ncmRsp_t **last;
	int rc;
	int pingMsg = (FRAMETAG | PINGMAGIC);
	struct timeval selTo;
	ncmLink_t *look;
	ncmRsp_t *rsp;
/*
	Register ourselves to receive all the node-directed messages
 */
	myappid = eee_getApplicationId ("ncm");
	appSock = regAgileApp ("ncm");
	if (appSock < 0) {
		perror ("Error registering ncm app");
		exit(1);
	}
/*
	Set up the bits we need for select() on all our open sockets.
 */
	FD_ZERO (&ourFds);
	FD_SET (appSock, &ourFds);
	maxFd = appSock + 1;
/*
	Let's listen for connections from our slaves
 */
	lisSock = setUpListen(argc > 1);
	if (lisSock < 0) {
		E_LOG (class_1, emerg_s, myappid, 0, 0, 0,
		       ("NCM: unable to listen for connections"));
		exit (1);
	}
	FD_SET (lisSock, &ourFds);
	maxFd = MAXFD (lisSock, maxFd);
/*
	Create our initial node descriptor and connect to ourself. 
 */
	nodes = malloc (sizeof(ncmLink_t));
	if (!nodes) {
		perror ("NCM: unable to allocate initial node descriptor");
		exit (1);
	}
	memset (nodes, 0, sizeof(ncmLink_t));

	nodes->lnkAddr[0] = inet_addr("127.0.0.1");
	nodes->lnkAddr[1] = 0;
	nodes->nodeName = strdup ("PCC");
	nodes->s = connToPCC (nodes->lnkAddr[0]);
	if (nodes->s < 0) {
		perror ("NCM: unable to make initial PCC connection");
		exit (1);
	}
	FD_SET (nodes->s, &ourFds);
	maxFd = MAXFD (nodes->s, maxFd);
/*
	Record who we are
 */
	gethostname (myName, sizeof(myName));
/*
	Tell MCP we're ready to go
 */
	kill (getppid(), SIGUSR1);
/*
	Wait for something to do
 */
	while (1) {
		rdFds = ourFds;
		selTo.tv_sec = PING_FREQ;
		selTo.tv_usec = 0;
		numFds = select (maxFd, &rdFds, NULL, NULL, &selTo);
		now = time (NULL);
/*
	If there's a local message coming in, get it and forward
 */
		if (FD_ISSET (appSock, &rdFds)) {
			numFds--;
/*
	We'll either be relaying messages or getting node events
 */
			if (handleANP (appSock, &ourFds, &maxFd)) {
				// A bad thing happened or we got a message to an unknown node
			}
		}
/*
	Is this a new connection request?
	(We do this AFTER handleANP() in case there is a race between LNM accepting
	 a node into the cluster, sending us the message, and the member node
	 connecting to us. We need to get the membership message processed before
	 the connection request comes in.)
 */
		if (FD_ISSET (lisSock, &rdFds)) {
			ncmLink_t *newNode;

			numFds--;
			newNode = tryToAccept (lisSock);
			if (newNode) {
				newNode->next = nodes;	// Push it on list
				nodes = newNode;
				FD_SET (newNode->s, &ourFds);
				maxFd = MAXFD (newNode->s, maxFd);
				new_node_event (newNode);
				resend_pending (newNode);
			}
		}
/*
	Look through the list of node connections and forward those messages and
	see if we need to test the link. Also checks for time outs and the like.
 */
		for (look = nodes; look; look = look->next) {
/*
	First see if this is an outstanding response we need to route back
 */
			for (last = &(look->outstanding), rsp = look->outstanding; rsp;) {
				if (FD_ISSET (rsp->rspSock, &rdFds)) {
					numFds--;
					FD_CLR (rsp->rspSock, &ourFds); // It's a one-shot socket
					if (relayResponse (rsp)) {
						E_LOG (class_1, notice_s, myappid, 0, 0, 0,
						       ("relayResponse failed"));
						// Bad News
					}
					*last = rsp->next;
					free (rsp);
					rsp = *last;
				}
/*
	When we tire of waiting for the response, drop it.
 */
				else if ((rsp->sent + RESPONSE_WAIT) < now) {
					*last = rsp->next;
					E_LOG (class_1, warning_s, myappid, 0, 0, 0,
					       ("timed out response wait rem %s loc %s",
					        rsp->remApp, rsp->locApp));
					FD_CLR (rsp->rspSock, &ourFds);
					close (rsp->rspSock);
					free (rsp);
					rsp = *last;
				}
				else {	// Take a look at any more pending on this node's port
					last = &(rsp->next);
					rsp = rsp->next;
				}
			}
/*
	Did something come in from another node?
 */
			if (FD_ISSET (look->s, &rdFds)) {
				numFds--;
				look->pingSent = 0;		// Doesn't much matter now
				look->lastTime = now;
				if (handleIO (look, &ourFds, &maxFd) < 0) {
					ncmLink_t *go_again;

					go_again = attempt_reconnect(look, &ourFds, &maxFd);

					if (go_again) {
						look = go_again;
					}
				}
			}
			else {		// See if we need to check this link's status
				if ((look->lastTime + PING_FREQ) < now) {
					if (look->pingSent) {
						ncmLink_t *go_again;

						E_LOG (class_1, err_s, myappid, 0, 0, 0,
						       ("ping wait timeout on link %s",
						        look->nodeName));

						go_again = attempt_reconnect(look, &ourFds, &maxFd);

						if (go_again) {
							look = go_again;
						}
					}
					else {
						// Send ping
						look->pingSent = 1;
						look->lastTime = now;
						rc = write (look->s, &pingMsg, sizeof(pingMsg));
						if (rc != sizeof(pingMsg)) {
						E_LOG (class_1, err_s, myappid, 0, 0, 0,
						       ("ping send fail on link %s",
						        look->nodeName));
							// Problem sending ping
						}
					}
				}
			}
		}
/*
	Make sure we serviced all the fds from select
 */
		if (numFds != 0) {
			E_LOG (class_1, warning_s, myappid, 0, 0, 0,
			       ("%d fds left over", numFds));
		}
/*
	Reap any dead node message ACK pending list that have been there too long
 */
		grim_reaper ();
	}
}
