/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: licman.c
 *
 * RCS:  $Id: licman.c,v 1.7 2002/08/28 22:10:16 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: NFX license manager interface to applications
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 4 Jan 2002
 *
 *-----------------------------------------------------------------
 */

#include <sys/types.h>
#include <fcntl.h>
#include <limits.h>
#include <db.h>
#include <unistd.h>
#include <sys/time.h>

#include "licman-api.h"
#include "licrec.h"

static int
valid_for_this_host (u_int32_t prodId, int sw_rev, u_int64_t *key)
{
	u_int32_t thisHostsId;
	u_int8_t sum;
	u_int32_t prod_part;
	u_int32_t sw_prod;
	u_int32_t host_part;
	u_int32_t ser_part;
	u_int32_t exp_date;
	struct timeval now;
	int i;
/*
	Get our host id (aka chassis serial id). Only need lower 20 bits.
 */
	thisHostsId = (u_int32_t)gethostid() & 0xfffff;
/*
	Check the key's validity
 */
	for (sum = i = 0; i < sizeof (u_int64_t); i++) {
		sum ^= ((u_int8_t *)key)[i];
	}
	if (sum != 0xff) {
		return 0;
	}
/*
	Pull it apart and check product and host id
 */
	prod_part = (u_int32_t)((*key >> 40) & 0xffff);
	sw_prod = prod_part & 0xff;
	if (sw_prod != prodId) {
		return 0;
	}

	host_part = (u_int32_t)((*key >> 20) & 0xfffff);
	ser_part = (u_int32_t)(*key & 0xfffff);
	if (sw_rev > (ser_part >> 12)) {
		return 0;
	}

	exp_date = (ser_part & 0xfff) << 20;
	gettimeofday (&now, NULL);
	if ((u_int32_t)(now.tv_sec) > exp_date) {
		return 0;
	}

/*
	Avert your eyes because this part is secret
 */
	host_part ^= (~ser_part & 0xfffff) ^ prod_part;

	if (host_part != thisHostsId) {
		return 0;
	}
	else {
		return 1;
	}
}

/*-----------------------------------------------------------------
 * Name: licman_is_valid
 *
 * Description: Returns validity of product's license on local machine
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 4 Jan 2002
 *
 *-----------------------------------------------------------------
 */
extern int 
licman_is_valid (int prodId, int sw_rev)
{
	DB *lic_db;
	DBT keyThang;
	DBT dataThang;
	int rc;
	key_rec_t rec;
/*
	Is it a vaild prodId?
 */
	if ((prodId <= 0) || (prodId >= LIC_INVALID)) {
		return 0;
	}
/*
	Open license database and retrieve key for prodId
 */
	lic_db = dbopen (LICDBFILE, O_RDONLY, 0, DB_BTREE,
	                 NULL);

	if (!lic_db) {
		return 0;
	}

	keyThang.data = &prodId;
	keyThang.size = sizeof (prodId);

	rc = 1;
	if (lic_db->get (lic_db, &keyThang, &dataThang, 0)) {
		rc = 0;
	}

	if (rc && (dataThang.size != sizeof (rec))) {
		rc = 0;
	}
/*
	Make sure it's valid and bound to this host
 */
	if (rc) {
		memcpy (&rec, dataThang.data, sizeof(rec));
		rc = valid_for_this_host (prodId, sw_rev, &rec.lic_key);
	}

	lic_db->close (lic_db);	// Only now is it OK to discard dataThang
	return rc;
}
