/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	lnm-main.c
 *
 * RCS:  $Id: lnm-main.c,v 1.15 2002/09/27 01:00:55 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Clustering initialization routines
 *
 * Created by:	Pankaj Gupta 
 *
 * Date Created:	11/05/01
 *
 *-----------------------------------------------------------------
 */


#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()


#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <ifaddrs.h>

#include "nfx-incl.h"
#include "exports/anpssc-api.h"
#include "exports/lnm-api.h"
#include "exports/ncm-api.h"
#include "exports/evm-api.h"
#include "exports/elog-api.h"
#include "exports/event-api.h"
#include "lnm-local.h"


/*------------------------------------------------------------------------------
 * Prototypes of functions.
 *------------------------------------------------------------------------------
 */
void                lnm_processEvent(nfx_eventMsg_t *eventMsg);

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

#define 			CLUSTER_SERVICE_NAME      	"cluster"
#define				MAXBINDTRIES				10
#define 			MAX_MSG_SIZE				1500
#define				SCC_TAKEOVER_HOLDOFF		30

int     			numIntervalCompleted = 0;
int     			clusterAppFd;
int 				clusterSockFd;
int					clusterSockPort = 6000;
char				myClusterName[LNM_MAX_CLUSTER_NAME];
int					clusterDebugLevel = 1;


int					myNodeIdx;
cluster_nodeInfo_t 	*myClusterNode;
cluster_nodeInfo_t 	*myClusterPCC;
time_t				PCC_failed_time;
cluster_nodeInfo_t 	*myClusterSCC;
cluster_nodeInfo_t	clusterNodeInfoArray[MAX_NODE_PER_CLUSTER];

uint32 lnmDebugFlag = 0;

int nfxMyCPUId, nfxMyChassisId, nfxMySlotId;
int myAppId;

time_t now;
time_t node_timeout;

/*------------------------------------------------------------------------------
 * Name         : lnm_initCommunication()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
static
int lnm_initCommunication()
{
	int one = 1;
	int	bindtries;
	int	boundyet=0;
	struct sockaddr_in my_addr;    /* my address information */

	/*
	 * Sets the procedure to follow in case of request for 
	 * program termination or CTRL-C interrupt.
	 */
	signal(SIGTERM, lnm_stop);
	signal(SIGINT, lnm_stop);

	/*
	 * Register to receive messages for cluster
	 */
	clusterAppFd = regAgileApp(CLUSTER_SERVICE_NAME);
	assert(clusterAppFd != 0);

	/*
	 * Open socket for transmitting your messages
	 */
	if ((clusterSockFd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
	{
		E_LOG(class_1, err_s, myAppId, 0, 0, 0,
		      ("Error getting socket: %s", strerror(errno)));
		return(-1);
   	}

	/* 
	 * Warn that we're going to broadcast 
	 */
	if (setsockopt(clusterSockFd, SOL_SOCKET, SO_BROADCAST, 
	               &one, sizeof(one)) < 0)
	{
		E_LOG(class_1, err_s, myAppId, 0, 0, 0,
		      ("Error setting socket option SO_BROADCAST: %s", 
		       strerror(errno)));
		close(clusterSockFd);
		return(-1);
	}

	if (setsockopt(clusterSockFd, SOL_SOCKET, SO_REUSEADDR, 
	               &one, sizeof(one)) < 0)
	{
		/* 
		 * Ignore it.  It will almost always be OK anyway. 
		 */ 
		 E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
		       ("Error setting option SO_REUSEADDR", strerror(errno)));
	}        

	/* 
	 * Try binding a few times before giving up 
	 * Sometimes a process with it open is exiting 
	 * right now
	 */

	bzero(&(my_addr), sizeof(my_addr));	/* zero my address struct */
	my_addr.sin_family = AF_INET;		/* host byte order */
	my_addr.sin_port = htons(clusterSockPort + myClusterNode->nodePortOffset);												/* short, network byte order */
	my_addr.sin_addr.s_addr = INADDR_ANY;	/* auto-fill with my IP */

	for(bindtries=0; !boundyet && bindtries < MAXBINDTRIES; ++bindtries) 
	{
		if (bind(clusterSockFd, (struct sockaddr *)&my_addr,	
					sizeof(struct sockaddr)) < 0) 
		{
			E_LOG(class_1, err_s, myAppId, 0, 0, __LINE__,
			      ("Error binding socket. Retrying: %s", strerror(errno)));
			sleep(1);
		}
		else
		{
			boundyet = 1;
		}
	}

	if (!boundyet) 
	{
		if (errno == EADDRINUSE) 
		{
			E_LOG(class_1, err_s, myAppId, 0, 0, 0,
			      ("Someone already listening on port"));
		}
		else
		{
			E_LOG(class_1, err_s, myAppId, 0, 0, 0,
			      ("Unable to bind socket. Giving up %s", strerror(errno)));
		}
		close(clusterSockFd);
		return(-1);
	}

	return(0);
}

/*------------------------------------------------------------------------------
 * Name         : cluster_postEvent()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
int cluster_postEvent(cluster_nodeInfo_t* nodeEntry, int eventId, 
                      int eventState, void *ips, int ipsz)
{
	nfx_eventMsg_t eventMsg;

	/*
	 * Just post the event we do not want to wait for 
	 * response
	 */
	eventMsg.msgType = NFX_EVENT_POST;
	eventMsg.eventType = eventId;
	eventMsg.eventScope = NFX_EVENT_LOCAL;
	eventMsg.needAck = 0;
	eventMsg.reason[0] = '\0';

	/*
	 * Set event specific paramaters
	 */
	strcpy(eventMsg.nodeName, nodeEntry->nodeName);
	eventMsg.param.nodeEvent.state = eventState;

    /*
	 * and if there's a payload to the event, send it too
	 */
	if (ips) {
		memcpy (eventMsg.param.nodeEvent.mgmt_ip, ips, ipsz);
	}

	sendAgileMsg(-1, &eventMsg, sizeof(nfx_eventMsg_t),"eventd",
				NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				0, 0, 0);
	return 0;
}

/*------------------------------------------------------------------------------
 * Name         : lnm_processEvent()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void lnm_processEvent(nfx_eventMsg_t *eventMsg)
{
	/*
	 * Just process the incoming event
	 */
	switch(eventMsg->eventType)
	{
		case NFX_EVENT_NODE:
		printf("lnm_processEvent Rcvd NFX_EVENT_NODE\n");
		break;

		case NFX_EVENT_PCC:
		printf("lnm_processEvent Rcvd NFX_EVENT_PCC\n");
		break;

		case NFX_EVENT_SCC:
		printf("lnm_processEvent Rcvd NFX_EVENT_SCC\n");
		break;

		default:
		printf("lnm_processEvent Rcvd unhandled event %d\n",
					eventMsg->eventType);
	}
}

/*------------------------------------------------------------------------------
 * Name         : lnm_start()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
static
int lnm_start()
{
	int	rc;
	int	maxFd;    // maximum file descriptor (select optimization)
	fd_set	fdMap;       // set of file descriptors we're interested in
	static	struct timeval	timeout;
	time_t next_time;
	char **nodeTimeoutCfg;

	/*
	 * Setup desciptors to monitor
	 */
	FD_ZERO(&fdMap);                     // reset fdMap membership
	FD_SET(clusterSockFd, &fdMap);    
	FD_SET(clusterAppFd, &fdMap);    
	maxFd = MAX(clusterSockFd, clusterAppFd);
#if THIS_IS_STANDALONE
	FD_SET(fileno(stdin), &fdMap);    
	maxFd = MAX(maxFd, fileno(stdin));
#endif
	maxFd++;

	/*
	 * See if there is a node timeout specified (disabled when set to 0)
	 * We'll ensure it's not < 0 nor < CLUSTER_TIMER_SECONDS
	 */
	nodeTimeoutCfg = get_cfg_val ("cluster.nodefailtime");
	if (nodeTimeoutCfg) {
		node_timeout = strtol (nodeTimeoutCfg[1], NULL, 10);
		if (node_timeout && (node_timeout < CLUSTER_TIMER_SECONDS)) {
			node_timeout = CLUSTER_TIMER_SECONDS;
		}
		free_cfg_val (nodeTimeoutCfg);
	}
	else {
		node_timeout = CLUST_NODE_FAIL_TIMEOUT;	// Default
	}

	/*
	 * Set timeout to CLUSTER_TIMER_SECONDS
	 * We track time elapsed between each loop
	 * and use it for transmission of periodic
	 * messages that may need to be transmitted 
	 * by this node
	 */
	timeout.tv_sec = CLUSTER_TIMER_SECONDS;
	timeout.tv_usec = 0;

    E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
	      ("Ready for communication on port %d\n", clusterSockPort +
									myClusterNode->nodePortOffset));

	/*
	 * Monitor your descriptors and process messages
	 */
	next_time = time (NULL) + CLUSTER_TIMER_SECONDS;
	while(1)
	{
		fd_set readFdMap = fdMap;  /* copy flags (structure copy) */

		if ((rc = select(maxFd, &readFdMap, NULL, NULL, &timeout)) < 0)
		{
			E_LOG(class_1, err_s, myAppId, 0, 0, 0,
			      ("select() returned with error: %s", strerror(errno)));
			exit(1);
		}

		now = time (NULL);	// At the tone, the time will be...

		/*
		 * Check if we have any thing to process
		 * and process it
		 */
		if (FD_ISSET(clusterSockFd, &readFdMap))
		{
			struct 	sockaddr_in  their_addr; 
			char	buf[MAX_MSG_SIZE];
			int		addr_len = sizeof(struct sockaddr);
			int 	numbytes;
	
        	/* 
			 * Read the message
         	 */
			addr_len = sizeof(their_addr);
			if ((numbytes = recvfrom(clusterSockFd, buf, sizeof(buf), 0,   
		 			(struct sockaddr *)&their_addr, &addr_len)) == -1) 
			{
				E_LOG(class_1, err_s, myAppId, 0, 0, __LINE__,
				      ("Error receiving from socket"));
			}
			else 
			{
				cluster_rxControlMsg((cluster_controlMsg_t *)buf, numbytes, 
							their_addr.sin_addr.s_addr);
			}
		}

		if (FD_ISSET(clusterAppFd, &readFdMap))
		{
			char	buf[MAX_MSG_SIZE];
			char	sendername[256];
			int 	numbytes;
	
        	/* 
			 * Read the message
         	 */
			numbytes = recvAgileMsg(clusterAppFd, buf, sizeof(buf),
							NULL, sendername, 0, 0, 0);
			if (numbytes > 0)
			{
				cluster_rxAgileMsg(buf, numbytes, sendername);
			}
			else
			{
				E_LOG(class_1, err_s, myAppId, 0, 0, __LINE__,
				      ("Error recv'ing AgileMsg"));
			}
		}

#if THIS_IS_STANDALONE

		if (FD_ISSET(fileno(stdin), &readFdMap))
		{
			char line[512];

			if (fgets(line, 512, stdin) == NULL)
			{
				printf("error\n");
			}
			else
			{
				int argc = 0;
				char * argv[10];
				char * linetmp;
				char name[10] = "cluster";

				argv[0] = name;

				linetmp = (char *) malloc(512);
				strcpy(linetmp, line);
				for (argc = 1; strcmp(linetmp, "") && argc < 10; argc++)
				{
					argv[argc] = linetmp;
					if (index(argv[argc], ' ') != NULL)
					{
						linetmp =  index(linetmp, ' ') + 1;
						*index(argv[argc], ' ') = 0;
					}
					else if (index(argv[argc], '\n') != NULL)
					{
						linetmp = index(linetmp, '\n');
						*index(argv[argc], '\n') = 0;
					}

				} // for

				cluster_processCli(argc, argv);
			}
		}
#endif

		/*
		 * Every CLUSTER_TIMER_SECONDS we check to see what the all the
		 * nodes status are.
		 */ 
		if (next_time <= now)
		{
			numIntervalCompleted++;
			lnm_processTimer();
			next_time = now + CLUSTER_TIMER_SECONDS;
		}
	}


    exit(0);
}

/*------------------------------------------------------------------------------
 * Name         : lnm_stop()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void lnm_stop(int sig)
{
    /* 
	 * Close sockets.
     */
    close(clusterSockFd);
	unRegAgileApp(clusterAppFd);
    
    exit(0);
}

/*------------------------------------------------------------------------------
 * Name         : lnm_processTimer()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void lnm_processTimer()
{
	int idx;
	cluster_nodeInfo_t   *nodeEntry;

	/*
	 * Update your own node
	 */
	cluster_updateNode(myClusterNode, NULL);

	/*
	 * Check your state and take appropriate action
	 */
	if (myClusterNode->nodeState == CN_STATE_LISTEN)
	{
		/*
	 	 * If you are in CN_STATE_LISTEN and not heard from PCC
	 	 * for LISTEN_STATE_INTERVAL go ahead and take up role of PCC
	 	 * else keep transmitting Find PCC request
	 	 */
		if (numIntervalCompleted <= LISTEN_STATE_INTERVAL)
		{
			cluster_sendFindPCCReq();
		}
		else
		{
			/*
			 * Waited long enough not heard from PCC
			 * go ahead and become the PCC
			 */
			cluster_changeNodeState(myClusterNode, CN_STATE_PCC);
		}
	}
	else if (myClusterNode->nodeState == CN_STATE_FIND_PCC)
	{
	 	/*
	 	 * If you are waiting to find PCC just keep 
	 	 * transmitting Find PCC request
	 	 */
		cluster_sendFindPCCReq();
	}
	else if (cluster_nodeIsUp(myClusterNode))
	{
		cluster_sendNodeHello(myClusterNode->nodeBroadcast);
	}

	/*
	 * Only PCC and SCC go beyond here
	 */
	if ((myClusterNode->nodeState != CN_STATE_PCC) &&
			(myClusterNode->nodeState != CN_STATE_SCC))
	{
		return;
	}

	/*
	 * Nothing to do if you are not CC enabled
	 */
	if (myClusterNode->CCMode != CC_mode)
		return;

	/*
	 * SCC/PCC Monitor state of other nodes
	 * Update age for all nodes and check for nodes
	 * with age greater than timeout period
	 */
	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if (!(nodeEntry->useFlag & NODE_ENTRY_ADDED))
			continue;

		if ((nodeEntry->nodeState != CN_STATE_DOWN) &&
			node_timeout && (now >= nodeEntry->deadWhen))
		{
			cluster_changeNodeState(nodeEntry, CN_STATE_DOWN);
		}
	}

	/*
	 * If you are SCC and you can't see the PCC
	 * takeover role of PCC once you've waited SCC_TAKEOVER_HOLDOFF seconds
	 */
	if (myClusterNode->nodeState == CN_STATE_SCC)
	{
		if (!myClusterPCC && 
		    ((PCC_failed_time + SCC_TAKEOVER_HOLDOFF) < time (NULL)))
		{
			cluster_changeNodeState(myClusterNode, CN_STATE_PCC);
		}
	}

	/*
	 * If you are PCC and do not have an assigned
	 * SCC look for a suitable candidate and assign 
	 * him SCC
	 */
	if (myClusterNode->nodeState == CN_STATE_PCC)
	{
		cluster_nodeInfo_t	*nodeEntry = NULL;

		if (myClusterSCC == NULL)
		{
			nodeEntry = cluster_findSccCandidate();
			if (nodeEntry)
				cluster_sendAssignSCCReq(nodeEntry);
		}
	}
}

static ccmode_t
clusterControllerMode(void)
{
	regmGetReq_t req;
	uint64 buf[1024];
	regmGetRsp_t *rsp = (regmGetRsp_t *)buf;
	int rc;
	struct timeval regmTo = {10, 0};
	ccmode_t my_mode;

	req.hdr.msgType = OPT_GET_OPTION_REQ;
	req.hdr.userInfo = getpid();
	strcpy (req.getReq.option, "cluster.nodetype");

	rc = sendAgileMsg (-1, &req, sizeof(req),
	                   "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
	                   buf, sizeof(buf));

	my_mode = standalone_mode;
	if (!rc) {
		E_LOG (class_1, alert_s, myAppId, 0, 0, __LINE__,
		       ("Timeout waiting for response from regm"));
	}
	else if (rsp->getRsp.nbEntry) {
		if (!strcasecmp (rsp->getRsp.entry[0].value, "pcc") ||
		    !strcasecmp (rsp->getRsp.entry[0].value, "cc"))
		{
			my_mode = CC_mode;
		}
		else
		{
			my_mode = member_mode;
		}
	}
	return my_mode;
}

static int
clusterPreferredPCCMode(void)
{
	regmGetReq_t req;
	uint64 buf[1024];
	regmGetRsp_t *rsp = (regmGetRsp_t *)buf;
	int rc;
	struct timeval regmTo = {10, 0};

	req.hdr.msgType = OPT_GET_OPTION_REQ;
	req.hdr.userInfo = getpid();
	strcpy (req.getReq.option, "cluster.nodetype");

	rc = sendAgileMsg (-1, &req, sizeof(req),
	                   "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
	                   buf, sizeof(buf));

	if (!rc) {
		E_LOG (class_1, alert_s, myAppId, 0, 0, __LINE__,
		       ("Timeout waiting for response from regm"));
	}
	else if (!rsp->getRsp.nbEntry) {
		rc = 0;
	}
	else {
		rc = !strcasecmp (rsp->getRsp.entry[0].value, "pcc");
	}
	return rc;
}

static void
get_cluster_name (char *name)
{
	regmGetReq_t req;
	uint64 buf[1024];
	regmGetRsp_t *rsp = (regmGetRsp_t *)buf;
	int rc;
	struct timeval regmTo = {10, 0};

	req.hdr.msgType = OPT_GET_OPTION_REQ;
	req.hdr.userInfo = getpid();
	strcpy (req.getReq.option, "cluster.name");

	rc = sendAgileMsg (-1, &req, sizeof(req),
	                   "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
	                   buf, sizeof(buf));

	if (!rc) {
		E_LOG (class_1, alert_s, myAppId, 0, 0, __LINE__,
		       ("Timeout waiting for response from regm"));
		strcpy (name, "");
	}
	else if (!rsp->getRsp.nbEntry) {
		strcpy (name, "");
	}
	else {
		strcpy (name, rsp->getRsp.entry[0].value);
	}
}

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
int main(int argc, char **argv)
{
	int		nodePortOffset = 0;
	struct in_addr myaddr[2];
	struct servent *service;
	struct ifaddrs *if_info, *lookie;

	/*
	 * Setup the necessary things here 
	 * These values should be either set 
	 * by configuration or obtained 
	 * from other modules as needed
	 */

	myAppId = eee_getApplicationId(CLUSTER_SERVICE_NAME);
	nfxMyCPUId = NFX_SSC_UNIX;
	nfxMySlotId = 0;

	/*
	 * Get from the configuration file the cluster name. An empty string
	 * indicates we're a single node (no cluster) and we operate as a PCC
	 * but without the advertisement.
	 */

	get_cluster_name(myClusterName);

	nfxMyChassisId = gethostid();

#ifdef SIMULATE_TWO_NODES
	nodePortOffset = atoi(argv[2]);
	if ((nodePortOffset < 0) || (nodePortOffset > 2))
	{
		perror("Cluster node port offset range is 0 to 2");
		return(-1);
	}
#else
	nodePortOffset = 0;
#endif

	memset(clusterNodeInfoArray, 
		sizeof(cluster_nodeInfo_t)*(MAX_NODE_PER_CLUSTER), 0);

	/*
	 * Allocate, Initialize and add entry
	 * for our own record
	 */

	myClusterNode = cluster_allocateNode();

	gethostname(myClusterNode->nodeName, sizeof(myClusterNode->nodeName));
	get_ip_addr(myaddr, myClusterNode->nodeName);

	E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
	      ("Cluster SSC IP addr for node %s is %08x & %08x\n", 
			myClusterNode->nodeName, myaddr[0].s_addr, 
	                    myaddr[1].s_addr));

	myClusterNode->nodeState = CN_STATE_INIT;

	myClusterNode->CCMode = clusterControllerMode();
	myClusterNode->preferredPCCMode = clusterPreferredPCCMode();
	myClusterNode->clusterState = 1;
	myClusterNode->unmanagedOrphanAdoptMode = 1;
	myClusterNode->nodeTimeoutForcedTakeoverMode = 1;

	myClusterNode->deadWhen = time (NULL) + CLUST_NODE_FAIL_TIMEOUT;
	myClusterNode->nodePortOffset = nodePortOffset;
	myClusterNode->nodeChassisId = nfxMyChassisId;

	/*
	 * Record our addrs and get the broadcast addresses for our interfaces
	 *
	 */
	myClusterNode->nodeIpa[0] = myaddr[0].s_addr;
	myClusterNode->nodeIpa[1] = myaddr[1].s_addr;

	if (!getifaddrs(&if_info)) 
	{
		for (lookie = if_info; lookie; lookie = lookie->ifa_next)
		{
			if (lookie->ifa_addr->sa_family == AF_INET)
			{
				int i;
				struct sockaddr_in *net_addr;
				for (i = 0; i < 2; i++)
				{
					net_addr = (struct sockaddr_in *)(lookie->ifa_addr);
					if (myaddr[i].s_addr && 
					 (myaddr[i].s_addr == net_addr->sin_addr.s_addr))
					{
						myClusterNode->nodeBroadcast[i] = 
						            ((struct sockaddr_in *)
						             (lookie->ifa_broadaddr))->sin_addr.s_addr;
					}
				}
			}
		}
		freeifaddrs (if_info);
		E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
		      ("addr/broadcast: %x/%x %x/%x", myClusterNode->nodeIpa[0],
		      myClusterNode->nodeBroadcast[0], myClusterNode->nodeIpa[1],
		      myClusterNode->nodeBroadcast[1]));
	}
	else
	{
		E_LOG(class_1, err_s, myAppId, 0, 0, 0,
		      ("Error getting if addr info: %s", strerror(errno)));
	}

	/*
	 * Add the entry now
	 */
	cluster_addNode(myClusterNode);

	if (myClusterNode->CCMode == CC_mode)
		cluster_changeNodeState(myClusterNode, CN_STATE_LISTEN);
	else if (myClusterNode->CCMode == member_mode)
		cluster_changeNodeState(myClusterNode, CN_STATE_FIND_PCC);
	else
		cluster_changeNodeState(myClusterNode, CN_STATE_DOWN);


	/* 
	 * Find the port on which to run services
	 * If our service name is in /etc/services, 
	 * then use it 
	 */
	if ((service = getservbyname(CLUSTER_SERVICE_NAME, "udp")) != NULL) 
		clusterSockPort = ntohs(service->s_port);

	/*
	 * Setup the socket for communication
	 */
	if (lnm_initCommunication() < 0)
	{
		E_LOG(class_1, err_s, myAppId, 0, 0, 0,
		      ("Error initializing cluster communication"));
		return(-1);
	}
	/*
	 * Tell MCP we're ready to roll
	 */
	kill (getppid(), SIGUSR1);

	if (lnm_start() < 0)
	{
		E_LOG(class_1, err_s, myAppId, 0, 0, 0,
		      ("Error starting cluster communication"));
		return(-1);
	}

	return 0;
}
