/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	lnm-rx.c
 *
 * RCS:  $Id: lnm-rx.c,v 1.12 2002/10/09 18:52:17 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Process received LNM messages
 *
 * Created by:	Pankaj Gupta 
 *
 * Date Created:	11/05/01
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()

#include <fcntl.h>
#include <netinet/in.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>

#include "nfx-incl.h"

#include "exports/lnm-api.h"
#include "exports/anpssc-api.h"
#include "exports/opt-api.h"
#include "exports/elog-api.h"
#include "exports/event-api.h"
#include "lnm-local.h"

#undef DO_VERBOSE
#if defined(DO_VERBOSE)
#define VERBOSE_DEBUG(x) x
#else
#define VERBOSE_DEBUG(x)
#endif

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */
void                lnm_processEvent(nfx_eventMsg_t *eventMsg);


/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

extern 	char				myClusterName[LNM_MAX_CLUSTER_NAME];
extern  cluster_nodeInfo_t 	*myClusterNode;
extern	int					clusterDebugLevel;

/*------------------------------------------------------------------------------
 * Name         : cluster_rcvInitHello()
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/22/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_rcvInitHello(cluster_controlMsg_t *ccMsg, int msgLen)
{
	if (ccMsg->srcChassisId == nfxMyChassisId)
		return;

	VERBOSE_DEBUG(
	E_LOG(class_7, debug_s, myAppId, 0, 0, 0,
	      ("Handling cluster_rcvInitHello from %08x %08x to %08x",
	        ccMsg->srcIpa[0], ccMsg->srcIpa[1], ccMsg->dstIpa));)

}

/*------------------------------------------------------------------------------
 * Name         : cluster_rcvNodeHello()
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/22/01
 *
 *------------------------------------------------------------------------------
 */
static void 
cluster_rcvNodeHello(cluster_controlMsg_t *ccMsg, int msgLen)
{
	cluster_nodeInfo_t   *nodeEntry;

	if (ccMsg->srcChassisId == nfxMyChassisId)
		return;

	VERBOSE_DEBUG(
	E_LOG(class_7, debug_s, myAppId, 0, 0, 0,
	      ("Handling cluster_rcvNodeHello from %08x %08x to %08x",
	       ccMsg->srcIpa[0], ccMsg->srcIpa[1], ccMsg->dstIpa));)

	if (ccMsg->msg.nodeHelloMsg.nodeState == CN_STATE_PCC)
	{
		if (myClusterNode->nodeState == CN_STATE_FIND_PCC)
		{
			if (myClusterNode->CCMode == CC_mode)
				cluster_changeNodeState(myClusterNode, CN_STATE_CC);
			else
				cluster_changeNodeState(myClusterNode, CN_STATE_NON_CC);
		}
		else if (myClusterNode->nodeState == CN_STATE_LISTEN)
		{
			cluster_sendFindPCCReq();
			cluster_changeNodeState(myClusterNode, CN_STATE_FIND_PCC);
		}
		else if (myClusterNode->nodeState == CN_STATE_PCC)
		{
			/*
			 * PCC got message from another PCC
			 * Conflict case should not NORMALLY come here
			 * We get here if a node becomes "separated" from the cluster
			 * and elects itself PCC. When it comes back into the cluster,
			 * (i.e. the idiot who unplugged the network cable puts it back)
			 * it needs to negotiate with the now-elected PCC.
			 * The winner will be the node configured to be preferred PCC
			 * or the lowest chassis id. Someone has to win.
		 	 */
			E_LOG(class_1, alert_s, myAppId, 0, 0, 0,
			      ("Node %d and %d are both PCC",
					ccMsg->srcChassisId, nfxMyChassisId));

			if (ccMsg->msg.nodeHelloMsg.preferredPCCMode == 
			    myClusterNode->preferredPCCMode) {

				if (ccMsg->srcChassisId < nfxMyChassisId) {
					cluster_changeNodeState(myClusterNode, CN_STATE_LISTEN);
				}
			}
			else {
				if (!myClusterNode->preferredPCCMode) {
					cluster_changeNodeState(myClusterNode, CN_STATE_LISTEN);
				}
			}
		}
	}

	nodeEntry = cluster_findNode(ccMsg->srcChassisId);
	if (!nodeEntry)
	{
		nodeEntry = cluster_allocateNode();
		nodeEntry->nodeState = CN_STATE_INIT;
	    nodeEntry->nodeIpa[0] = ccMsg->srcIpa[0];
	    nodeEntry->nodeIpa[1] = ccMsg->srcIpa[1];
	    nodeEntry->nodeChassisId = ccMsg->srcChassisId;
		strncpy (nodeEntry->nodeName, ccMsg->msg.nodeHelloMsg.nodeName,
		         sizeof (nodeEntry->nodeName));
		cluster_addNode(nodeEntry);
	}

	if (nodeEntry->nodeState != ccMsg->msg.nodeHelloMsg.nodeState)
	{
	/*
	 * Do not transition the other node to PCC until it acknowledges our
	 * findPCC request.
	 */
		if ((ccMsg->msg.nodeHelloMsg.nodeState != CN_STATE_PCC) ||
		    (ccMsg->dstIpa == myClusterNode->nodeIpa[0]) ||
		    (ccMsg->dstIpa == myClusterNode->nodeIpa[1]))
		cluster_changeNodeState(nodeEntry, 
		                        ccMsg->msg.nodeHelloMsg.nodeState);
	}

	cluster_updateNode(nodeEntry, ccMsg);
}

/*------------------------------------------------------------------------------
 * Name         : cluster_rcvAssignSCCReq()
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/22/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_rcvAssignSCCReq(cluster_controlMsg_t *ccMsg, int msgLen)
{
	if (ccMsg->srcChassisId == nfxMyChassisId)
		return;

	VERBOSE_DEBUG(
	E_LOG(class_7, debug_s, myAppId, 0, 0, 0,
	      ("Handling cluster_rcvAssignSCCReq von %d", ccMsg->srcChassisId));)

	if ((myClusterNode->nodeState == CN_STATE_CC) &&
		!memcmp(ccMsg->msg.assignSCCReq.sccIpa, myClusterNode->nodeIpa,
	                sizeof (myClusterNode->nodeIpa)))
	{
		/*
		 * Node in CC state is being requested to become
		 * SCC by the PCC
		 */
		 cluster_changeNodeState(myClusterNode, CN_STATE_SCC);
	}

	/*
	 * Update info from PCC TBD
	 */
	myClusterNode->clusterState = ccMsg->msg.assignSCCReq.clusterState;
	myClusterNode->unmanagedOrphanAdoptMode = 
				ccMsg->msg.assignSCCReq.unmanagedOrphanAdoptMode;
	myClusterNode->nodeTimeoutForcedTakeoverMode = 
				ccMsg->msg.assignSCCReq.nodeTimeoutForcedTakeoverMode;
	
}

/*------------------------------------------------------------------------------
 * Name         : cluster_rcvFindPCCReq()
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/22/01
 *
 *------------------------------------------------------------------------------
 */
static void 
cluster_rcvFindPCCReq(cluster_controlMsg_t *ccMsg, int msgLen)
{
	if (ccMsg->srcChassisId == nfxMyChassisId)
		return;

	VERBOSE_DEBUG(
	E_LOG(class_7, debug_s, myAppId, 0, 0, 0,
	      ("Handling cluster_rcvFindPCCReq src %08x %08x to %08x",
	       ccMsg->srcIpa[0], ccMsg->srcIpa[1], ccMsg->dstIpa));)

	if (myClusterNode->nodeState == CN_STATE_LISTEN)
	{
		/*
		 * CC enabled node in listen state 
		 * Non CC enabled node starts in CN_STATE_FIND_PCC state
		 */
		if (ccMsg->msg.findPCCReq.preferredPCCMode)
		{
			/*
			 * Other node is preferred PCC. Change to 
			 * CN_STATE_FIND_PCC state if you are not the
			 * one to become PCC
			 */
			if (!myClusterNode->preferredPCCMode)
			{
				/*
				 * You are not configured as preferred PCC
				 */
				cluster_changeNodeState(myClusterNode, CN_STATE_FIND_PCC);
			}
			else if (myClusterNode->preferredPCCMode &&
					(ccMsg->srcChassisId < nfxMyChassisId))
			{
				/*
				 * You are also configured as preferred PCC
				 * (Invalid config  !! But give up if your
				 * chassis id is higher than the other node
				 */
				E_LOG(class_1, alert_s, myAppId, 0, 0, 0,
				      ("Node %d and %d are both configured as preferred PCC",
						ccMsg->srcChassisId, nfxMyChassisId));

				cluster_changeNodeState(myClusterNode, CN_STATE_FIND_PCC);
			}
		}
	}
	else if (myClusterNode->nodeState == CN_STATE_PCC)
	{
		/*
		 * Make sure we've added this node before we tell it we're PCC
		 * otherwise it won't be authorized to connect.
		 */
		cluster_nodeInfo_t *nodeEntry =  cluster_findNode(ccMsg->srcChassisId);
		if (!nodeEntry)
		{
			nodeEntry = cluster_allocateNode();
			nodeEntry->nodeState = CN_STATE_INIT;
	    	nodeEntry->nodeIpa[0] = ccMsg->srcIpa[0];
	    	nodeEntry->nodeIpa[1] = ccMsg->srcIpa[1];
	    	nodeEntry->nodeChassisId = ccMsg->srcChassisId;
			strncpy (nodeEntry->nodeName, ccMsg->msg.findPCCReq.myName,
			         sizeof (nodeEntry->nodeName));
			cluster_addNode(nodeEntry);
		}
		cluster_updateNode(nodeEntry, ccMsg);
		cluster_changeNodeState(nodeEntry, ccMsg->msg.findPCCReq.nodeState);

		/*
		 * Respond to the requesting node with PCC info
		 */
		 cluster_sendNodeHello(ccMsg->srcIpa);

	}
}

/*------------------------------------------------------------------------------
 * Name         : cluster_rxControlMsg()
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_rxControlMsg(cluster_controlMsg_t* msg, int msgLen, int srcIpa)
{
	cluster_controlMsg_t *ccMsg = (cluster_controlMsg_t *)msg;

	if (!clusterDebugLevel)
	E_LOG(class_7, debug_s, myAppId, 0, 0, __LINE__,
	      ("CC Msg (%d, %d) Src: %d 0x%x:%x Dst: 0x%x fm: %x", 
			msgLen, ccMsg->msgType, ccMsg->srcChassisId, 
			ccMsg->srcIpa[0],ccMsg->srcIpa[1], 
            ccMsg->dstIpa, srcIpa));

	if (strcmp(ccMsg->clusterName, myClusterName) != 0)
		return;

	switch (ccMsg->msgType)
	{
		case CLUSTER_INIT_HELLO_MSG:
			cluster_rcvInitHello(ccMsg, msgLen);
			break;

		case CLUSTER_NODE_HELLO_MSG:
			cluster_rcvNodeHello(ccMsg, msgLen);
			break;

		case CLUSTER_ASSIGN_SCC_REQ:
			cluster_rcvAssignSCCReq(ccMsg, msgLen);
			break;

		case CLUSTER_FIND_PCC_REQ:
			cluster_rcvFindPCCReq(ccMsg, msgLen);
			break;

		default:
			E_LOG(class_1, err_s, myAppId, 0, 0, __LINE__, 
			      ("cluster_rxControlMsg unknown message type %d",
							ccMsg->msgType));
			return;
	}
}

static void
reset_node_fail_times (void)
{
	int i;
	cluster_nodeInfo_t *nodeEntry = clusterNodeInfoArray;

	for (i = 0; i < MAX_NODE_PER_CLUSTER; i++, nodeEntry++)
	{
		if ((nodeEntry->useFlag & NODE_ENTRY_ADDED)) {
			nodeEntry->deadWhen = now + node_timeout;
		}
	}
}

/*------------------------------------------------------------------------------
 * Name         : cluster_rxAgileMsg()
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_rxAgileMsg(char *rcvdBuf, int msgLen, char *sender)
{
	cluster_msgType_t *msgType = (cluster_msgType_t *)rcvdBuf;

	switch (*msgType) // First word has to be message type
	{
		case NFX_EVENT_POST:
			lnm_processEvent((nfx_eventMsg_t *)rcvdBuf);
			break;

		case CLUSTER_NODE_TIMEOUT_MSG:
			node_timeout = ((lnm_newTimeoutMsg_t *)rcvdBuf)->newTimeout;
			if (node_timeout && (node_timeout < CLUSTER_TIMER_SECONDS)) {
				node_timeout = CLUSTER_TIMER_SECONDS;
			}
			reset_node_fail_times();
			break;

		default:
			E_LOG(class_1, err_s, myAppId, 0, 0, __LINE__,
			      ("cluster_rxAgileMsg unknown message type"));
			return;
	}
}

