/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: nfxnis-api.h
 *
 * RCS:  $Id: nfxnis-api.h,v 1.6 2002/04/03 01:38:46 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This is the way to retrieve name service in a multiple
 *              domain/multiple service environment.
 *
 *              In this instance, NIS stands for name information service
 *              since we allow both YP and DNS.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Nov 2001
 *
 *-----------------------------------------------------------------
 */
#if !defined(__nfxnis_api_h__)
#define __nfxnis_api_h__

#include <sys/types.h>	// Need definition for in_addr_t

// Returns the IP address (in network order) of a given "hostName"
// accessing an NFX export via address "exportAddr".
// The relationship between NIS domains and exported network addresses
// is a part of the NFX configuration.
//
// Returns 0 upon failure.

extern in_addr_t nfxnis_hostAddr (char *exportAddr, char *hostName);

// Returns a malloc()ed string containing the map entry for:
// passwd.byname in the nfxnis_userinfo() case,
// group.byname in the nfxnis_groupinfo() case
// from the NIS server assigned to exportAddr.
//
// Returns NULL upon failure.

extern char * nfxnis_userinfo (char *exportAddr, char *userName);
extern char * nfxnis_groupinfo (char *exportAddr, char *groupName);


/*-----------------------------------------------------------------*/

typedef enum {
	dns_name,
	nis_name
} nfxnis_serv_t;

// How to configure nameservice:
//
// You can add/del a service for a network.
//
// Given the service type (NIS(aka YP) or DNS),
// The network it's valid for (these are in network order)
// Use this server and the domain name.
//
// E.g. If you wanted this entry to be valid for ALL networks (you can have
//      only one), network = 0; netmask = 0; will match.
//
//      For NIS, the domain name is that which the given server serves 
//      (is running ypserv).
//
//      For DNS, the domain is the suffix for all nfxnis_hostAddr() requests.
//      For instance, hostname="foo" and the domain was set to "bar.com" will
//      resolve "foo.bar.com".
//
// Returns 0 upon success. Non-zero for failure.
// Upon failure for NIS only, errno will be set to EEXIST if a server has 
// already been set for the specified network since there can be only 
// one server per domain. 
// nfxnis_delservice() it first if you want to replace said service.
// System error encountered will return non-zero and errno set accordingly.
//
// For DNS served networks, only the first call to nfxnis_addservice() affects
// the domain name. The subsequent calls will add servers to the list up to
// some predetermined limit (oh, let's say 3 like a resolv.conf). Again,
// EEXIST will be used to indicate this overflow.

#define NFX_MAX_DNS_SERVERS_PER_DOMAIN 3

extern int nfxnis_addservice (nfxnis_serv_t type, 
                              in_addr_t network, in_addr_t netmask, 
                              in_addr_t serv_addr, char *domain);

extern int nfxnis_delservice (in_addr_t network, in_addr_t netmask);

// So you want to know what's stored in the configuration file?
// You can ask by network and mask. (remember a mask of 0 gets you everything.)

typedef struct {
	nfxnis_serv_t	type;		// DNS or YP
	in_addr_t		net;		// What network this was configured for 
	in_addr_t		mask;		// Goes with above (did I mention net order?)
	char			*domain;	// What's the domain
	int				num_srvrs;	// How many servers (only 1 for YP/NIS)
	in_addr_t		srvrs[NFX_MAX_DNS_SERVERS_PER_DOMAIN];// Address (net order)
} nfxnis_cfg_t;

// Input addresses are network order.
//
// It will return you a pointer to a malloc()ed, NULL terminated, list of
// descriptors with the info matching the specified network/mask.
//
// Just a reminder, the caller is responsible for free()ing this memory
// when they're done with it. Free domain string first, then the 
// nfxnis_cfg_t(s) and finally the list/array.

extern nfxnis_cfg_t **nfxnis_get_cfg (in_addr_t network, in_addr_t netmask);

// *** Helpful Hint: If you need to format these addresses into strings for
// display, look at the man page for inet_ntoa(). ***

// All of this lovely configuration info gets stored in the configuration
// file via the registry manager. We use the objects:
//
//   nameservice.nnnnnnnn.mmmmmmmm=(dns|yp):domain:server_addr[:moreservers]
//
// where:
//   nnnnnnnn - is the network-order network addr in hex
//   mmmmmmmm - is the network-order mask in hex

#define NAMESERVICE "nameservice."

#endif
