/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: nfxnis.c
 *
 * RCS:  $Id: nfxnis.c,v 1.9 2002/06/27 22:20:44 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains the library used to obtain multi-NIS domain name
 *              service.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Nov 2001
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <rpc/rpc.h>
#include <rpcsvc/ypclnt.h>
#include <rpcsvc/yp_prot.h>


#include "nfxnis-api.h"
#include "nfxnis-loc.h"

/*-----------------------------------------------------------------
 * Name: upd_cfg
 *
 * Description: Update the internal configuration data structure when
 *              the file changes.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Nov 2001
 *
 *-----------------------------------------------------------------
 */
struct domainRelation {
	struct in_addr theNet;		// The network for this domain
	in_addr_t mask;				// the net mask
	char domainName[80];		// The domain name
	struct in_addr NISserver;	// and who is serving it
	struct domainRelation *next;
};
static struct domainRelation *cfgLst = NULL;

/*
	The reason for using both stream and fd is that the stat system call
	need a file descriptor while the "easy" way of reading files uses
	streams. Figured it would be faster to NOT re-open the file everytime
	we needed stat.
 */
static FILE *cfgFile = NULL;	// Gets opened on the first call and stays
static int cfgfd;
static struct stat its_stat;

static void
upd_cfg()
{
	struct stat stat_now;
	struct domainRelation *x;
	struct domainRelation **nextPtr;
	char cfgLine[256];
	char *net;
	char *dName;
	char *server;
	char *Line;
	int bits;
	int whtSpc;
/*
	If the cfg file isn't open, open it and get initial stat
 */
	if (!cfgFile) {
		cfgfd = open (NFXNISCFG, O_RDONLY|O_SHLOCK, 0);
		if (cfgfd < 0) { // Trouble to beat all trouble
			perror ("NFXNIS: Unable to open NIS domain cfg file");
			exit (1);
		}
		cfgFile = fdopen (cfgfd, "r");
		fstat (cfgfd, &its_stat);	// Initial stat
	}
/*
	Otherwise, we need to see if the file has changed since last time
	we looked at it. We used advisory locks to avoid reading while it
	might be getting updated. Readers use shared locks and the writers
	use exclusive locks.
 */
	else {
		if (fstat (cfgfd, &stat_now)) {	// This is wierd
			perror ("NFXNIS: Oddly got error asking for stat");
			exit (1);
		}
/*
	Will need to re-read the cfg file if it has been removed or modified
 */
		if (!stat_now.st_nlink || 
		    memcmp (&stat_now.st_mtimespec, &its_stat.st_mtimespec, 
		            sizeof (struct timespec))) {

			close (cfgfd);
			fclose (cfgFile);
			cfgFile = NULL;

			while (cfgLst) {	// Toss old cfg
				x = cfgLst->next;
				free (cfgLst);
				cfgLst = x;
			}

			upd_cfg();			// Dontcha just love recursion?
		}
		return;
	}
/*
	Now we read the cfg file and populate the address/NIS domain relationship
	list.
 */
	nextPtr = &cfgLst;
	while (fgets (cfgLine, sizeof (cfgLine), cfgFile)) {
/*
	Ignore comment lines
 */
		if (cfgLine[0] =='#') {
			continue;
		}

		Line = cfgLine;
/*
	This line has (should have) three fields separated by white space that
	looks like:

		A.B.C.D[/Net] NISDomain A.B.C.D

        ^ For this network
                      ^ This NIS domain
                                ^ Is served by this host (running ypserv)
 */
		whtSpc = strspn(Line, " \t");	// Ignore leading white space
		Line += whtSpc;
		net = strsep (&Line, " \t");
		if (!net) {
			fprintf (stderr, "NFXNIS: Ignoring bogus line in cfg file\n");
			continue;
		}
		whtSpc = strspn(Line, " \t");	// Ignore leading white space
		Line += whtSpc;
		dName = strsep (&Line, " \t");
		if (!dName) {
			fprintf (stderr, "NFXNIS: Ignoring bogus line in cfg file\n");
			continue;
		}
		whtSpc = strspn(Line, " \t");	// Ignore leading white space
		Line += whtSpc;
		server = strsep (&Line, " \t\n");
		if (!server) {
			fprintf (stderr, "NFXNIS: Ignoring bogus line in cfg file\n");
			continue;
		}
/*
	All there, let's see if they're valid
 */
		x = (struct domainRelation *)malloc (sizeof (struct domainRelation));
		if (!x) {
			fprintf (stderr, "NFXNIS: failed to get memory for list\n");
			exit (1);
		}
		memset (x, 0, sizeof (struct domainRelation));

		bits = inet_net_pton (AF_INET, net, &(x->theNet), sizeof (x->theNet));
		if (bits == -1) {
			fprintf (stderr, "NFXNIS: Invalid dest net specifier %s\n",net);
			free (x);
			continue;
		}
		x->mask = (1 << bits) - 1;

		strncpy (x->domainName, dName, sizeof (x->domainName));

		if (!inet_aton (server, &(x->NISserver))) {
			fprintf (stderr, "NFXNIS: Invalid NIS server address for %s\n",
			         dName);
			free (x);
			continue;
		}
/*
	Passed muster so link it up
 */
		*nextPtr = x;
		nextPtr = &x->next;
	}

	if (ferror (cfgFile)) {
		perror ("NFXNIS: Error while reading cfg file");
		clearerr (cfgFile);
	}


	flock (cfgfd, LOCK_UN); // That's all folks
	rewind (cfgFile);
	return;
}

/*-----------------------------------------------------------------
 * Name: get_NISdomain
 *
 * Description: Given an address we export, find its NIS domain in
 *              the configuration.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Nov 2001
 *
 *-----------------------------------------------------------------
 */
static int
get_NISdomain (char *addr, char *domain, int allowdefault)
{
	struct in_addr realAddr;
	struct domainRelation *look;	
	struct domainRelation *def;	

	if (inet_aton (addr, &realAddr)) {
		for (look = cfgLst, def = NULL; look; look = look->next) {
			if (!look->mask) {
				if (!allowdefault) {
					continue;
				}
				def = look;
			}
			else if ((look->mask & realAddr.s_addr) == 
			    (look->mask & look->theNet.s_addr)) {
				strcpy (domain, look->domainName);
				return 0;
			}
		}

		if (def) {
			strcpy (domain, def->domainName);
			return 0;
		}
	}

	return 1;
}

/*-----------------------------------------------------------------
 * Name: resolveNISname
 *
 * Description: Resolves host name returning its IP address given
 *              a specific NIS domain.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Nov 2001
 *
 *-----------------------------------------------------------------
 */
static in_addr_t
resolveNISname (char *name, char *domain)
{
	char *data;
	int datalen;
	struct in_addr addr;
	in_addr_t ret;

	ret = 0;
	if (!yp_match (domain, "hosts.byname", name, strlen (name),
	               &data, &datalen)) {
		strtok (data, " ");
		
		if (inet_aton (data, &addr)) {
			ret = addr.s_addr;
		}
		free (data);
	}
	return ret;
}

static char *
resolveUserName (char *name, char *domain)
{
	char *data;
	int datalen;

	if (yp_match (domain, "passwd.byname", name, strlen (name),
	               &data, &datalen)) {
		data = NULL;
	}
	return data;
}

static char *
resolveGroupName (char *name, char *domain)
{
	char *data;
	int datalen;

	if (yp_match (domain, "group.byname", name, strlen (name),
	               &data, &datalen)) {
		data = NULL;
	}
	return data;
}

/*-----------------------------------------------------------------
 * Name: nfxnis_hostAddr
 *
 * Description: Returns the IP address of "hostname" given that host is
 *              accessible through "exportAddr" using NIS.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Nov 2001
 *
 *-----------------------------------------------------------------
 */
typedef enum {
	HOSTNAME,
	USERNAME,
	GROUPNAME
} lookup_t;

static int 
nfxnis_lookup (lookup_t lType, char *exportAddr, char *hostName, 
               void *rtn_data)
{
	struct in_addr ourAddr;
	char NISdomain[256];
	int rc;
/*
	Is this a valid IP address they've given us?
 */
	if (!inet_aton (exportAddr, &ourAddr)) {
		return 1;	// Not even close to a good address
	}
/*
	Make sure we have the latest configuration
 */
	upd_cfg();
/*
	Give me the NIS domain name valid for that exported address
	The first time, we disallow the default case in case we need to give
	DNS a chance at it.
	If DNS is the default, it'll get resolved there. If NIS is the default,
	the DNS resoltion will fail and we need to give NIS another crack
	at it.
 */
	rc = 1;
	switch (lType) {
		case HOSTNAME:
			if ((rc = get_NISdomain (exportAddr, NISdomain, 0)) != 0) {

				if ((rc = nfxnis_resolve_via_DNS (exportAddr, hostName, 
						rtn_data)) != 0) {

					if ((rc = get_NISdomain(exportAddr, NISdomain, 1)) == 0) {
						*(in_addr_t *)rtn_data = 
						                 resolveNISname (hostName, NISdomain);
					}
				}
			}
			else {
				*(in_addr_t *)rtn_data = resolveNISname (hostName, NISdomain);
			}
			break;
		case USERNAME:
			if (!(rc = get_NISdomain (exportAddr, NISdomain, 1))) {
				*(char **)rtn_data = resolveUserName (hostName, NISdomain);
			}
			break;
		case GROUPNAME:
			if (!(rc = get_NISdomain (exportAddr, NISdomain, 1))) {
				*(char **)rtn_data = resolveGroupName (hostName, NISdomain);
			}
	}
	return rc;
}

in_addr_t 
nfxnis_hostAddr (char *exportAddr, char *hostName)
{
	in_addr_t theirAddr;
	
	if (nfxnis_lookup (HOSTNAME, exportAddr, hostName, &theirAddr)) {
		theirAddr = 0;
	}

	return theirAddr;
}

char * 
nfxnis_userinfo (char *exportAddr, char *userName)
{
	char *uinfo;

	if (nfxnis_lookup (USERNAME, exportAddr, userName, &uinfo)) {
		uinfo = NULL;
	}

	return uinfo;
}

char * 
nfxnis_groupinfo (char *exportAddr, char *groupName)
{
	char *ginfo;

	if (nfxnis_lookup (GROUPNAME, exportAddr, groupName, &ginfo)) {
		ginfo = NULL;
	}

	return ginfo;
}
