/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_flash.c,v 1.8 2002/11/02 22:31:51 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created: 06/25/2002
 *
 *-----------------------------------------------------------------
 */

/*
 * Copyright (c) 1983, 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include <fstab.h>
#include <sys/disklabel.h>
#include <util.h>

#include "exports/opt-api.h"

extern int noisy_make_allfs_writable();

static char *primary_disk = "/dev/wd0";
static char *secondary_disk = "/dev/wd1";
static char *secondary_device = "wd1";
char *secondary_mount = "/mnt1";

/* List of files to copy for "copy all" command */
static char *all_files = ".cshrc .profile 4.4BSD.dist altroot bin bsd dev etc home mnt root sbin ssc-mgmt.bin stand tmp usr var version";

/* List of system configuration files, to copy by "copy config" command" */
char *sys_config_files = "usr/local/agile/conf/agile.conf etc/hosts etc/mygate etc/hostname.sc1 etc/hostname.sc2 etc/master.passwd etc/passwd etc/pwd.db etc/spwd.db etc/myname";


static struct {
	uint64_t size_min;
	uint64_t size_max;
	char *name;
} disknames[] = {
	{ 240000000, 350000000, "FLASH256"},
	{ 900000000, 1100000000, "FLASH1024"},
	{ 0, 0, NULL }
};

extern int mount_secondary(vector v);
extern vector make_secondary_fstabs(void);
extern void u_mount_secondary(vector v);
extern void free_fstab_vector(vector v);
extern int do_copy_files(char *fromdir, char *todir, char *files);

/*-----------------------------------------------------------------
 * Name :	disk_name()
 *
 * Description: Determine disk name in disk tab according to the disk
 * size.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
static char *
disk_name(size_t disk_blocks, size_t block_size)
{
	uint64_t size = disk_blocks * block_size;
	int i;
	for (i = 0; disknames[i].name != NULL; ++i) {
		if (disknames[i].size_min <= size
			&& size <= disknames[i].size_max) {
			return disknames[i].name;
		}
	}
	return NULL;
}


/*
 * mkpath -- create directories.  
 *	path     - path
 *	mode     - file mode of terminal directory
 *	dir_mode - file mode of intermediate directories
 */
static int
mkpath(char *path, mode_t mode, mode_t dir_mode)
{
	struct stat sb;
	register char *slash;
	int done = 0;

	slash = path;

	while (!done) {
		slash += strspn(slash, "/");
		slash += strcspn(slash, "/");

		done = (*slash == '\0');
		*slash = '\0';

		if (stat(path, &sb)) {
			if (errno != ENOENT ||
			    (mkdir(path, done ? mode : dir_mode) &&
			    errno != EEXIST)) {
				return (-1);
			}
		} else if (!S_ISDIR(sb.st_mode)) {
			return (-1);
		}

		*slash = '/';
	}

	return (0);
}


static int
mkpath1(char *path, mode_t mode)
{
	return mkpath(path, mode, mode);
}


/*-----------------------------------------------------------------
 * Name :	free_fstab_vector()
 *
 * Description: Free vector of malloced struct fstab *
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
void
free_fstab_vector(vector v)
{
	int i;
	for (i = 0; i < vector_max(v); ++i) {
		struct fstab *fs = vector_lookup_index(v, i);
		zfree(MTYPE_VECTOR, fs->fs_spec);
		zfree(MTYPE_VECTOR, fs->fs_file);
		zfree(MTYPE_VECTOR, fs);
	}
	vector_free(v);
}


/*-----------------------------------------------------------------
 * Name :	get_device_names()
 *
 * Description: find out which disk we booted from and setup device names
 * variables.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	11/01/02
 *
 *-----------------------------------------------------------------
 */
static void
get_device_names(void)
{
	struct stat st;
	char *rootdevice;
	stat("/", &st);
	rootdevice = devname(st.st_dev, S_IFBLK);
	if (!strcmp(rootdevice, "wd1a")) {
		primary_disk = "/dev/wd1";
		secondary_disk = "/dev/wd0";
		secondary_device = "wd0";
	}
	else {
		primary_disk = "/dev/wd0";
		secondary_disk = "/dev/wd1";
		secondary_device = "wd1";
	}
}


/*-----------------------------------------------------------------
 * Name :	make_secondary_fstabs()
 *
 * Description: Create fstab structures to mount the secondary
 * disk.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
vector
make_secondary_fstabs(void)
{
	struct fstab *fs;
	vector v = vector_init(6);

	get_device_names();
	
	setfsent();
	while ((fs = getfsent()) != NULL) {
		if (!strncmp(fs->fs_vfstype, "ffs", strlen("ffs"))
			&& strstr(fs->fs_spec, primary_disk) == fs->fs_spec) {
			struct fstab *fscopy = zmalloc(MTYPE_VECTOR, sizeof(*fscopy));
			char spec[MAXPATHLEN];
			char mount[MAXPATHLEN];
			sprintf(spec, "%s%s", secondary_disk,
					fs->fs_spec + strlen(primary_disk));
			fscopy->fs_spec = zstrdup(MTYPE_VECTOR, spec);
			sprintf(mount, "%s%s", secondary_mount, fs->fs_file);
			fscopy->fs_file = zstrdup(MTYPE_VECTOR, mount);
			vector_push(v, fscopy);
		}
	}
	endfsent();
	
	return v;
}


/*-----------------------------------------------------------------
 * Name :	ensure_dir_exists()
 *
 * Description: make sure directory exists on a possibly read-only
 * partition.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	09/12/02
 *
 *-----------------------------------------------------------------
 */
int
ensure_dir_exists(const char *path, int mode)
{
	struct stat statbuf;
	int rc = stat(path, &statbuf);
	if (rc != 0) {
		rc = modify_readonly_file(path, mkpath1, (void *)mode);
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	mount_secondary()
 *
 * Description: Mount the secondary disk
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
int
mount_secondary(vector v)
{
	int i, error = 0;
	for (i = 0; !error && i < vector_max(v); ++i) {
		struct fstab *fs = vector_lookup_index(v, i);
		char path[MAXPATHLEN];
		strcpy(path, fs->fs_file);
		if (ensure_dir_exists(path, 0777) == 0) {
			if (!error) {
				if (execute_command("/sbin/mount", 2, fs->fs_spec,
									fs->fs_file, NULL) != 0) {
					fprintf(stderr, "Can't mount %s on %s\n",
							fs->fs_spec, fs->fs_file);
					error = 1;
				}
			}
		}
		else {
			error = 1;
			printf("Can't create directory %s\n", path);
		}
	}

	return error;
}


/*-----------------------------------------------------------------
 * Name :	u_mount_secondary()
 *
 * Description: Unmount the secondary disk
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
void
u_mount_secondary(vector v)
{
	int i;
	/* Umount the directories in reverse order */
	for (i = vector_max(v) - 1; i >= 0; --i) {
		struct fstab *fs = vector_lookup_index(v, i);
		char command[256];
		sprintf(command, "/sbin/umount %s >/dev/null 2>/dev/null",
				fs->fs_file);
		system(command);
	}
}


/*-----------------------------------------------------------------
 * Name :	disk_size()
 *
 * Description: Determine the device size
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
static int
disk_size(char *device, size_t *disk_blocks,
					size_t *block_size)
{
	int error = 1;
	char *specname;
	int f;
	
	become_root();
	if ((f = opendev(device, O_RDONLY, OPENDEV_PART, &specname)) > 0) {
		struct disklabel label;
		if (ioctl(f, DIOCGPDINFO, &label) == 0) {
			error = 0;
			*disk_blocks = label.d_secperunit;
			*block_size = label.d_secsize;
		}
		else {
			fprintf(stderr, "Can not read disk label: %s\n",
					strerror(errno));
		}
		close(f);
	}
	else {
		fprintf(stderr, "Can not open %s: %s\n", specname,
				strerror(errno));
	}
	unbecome_root();
	return error;
}


/*-----------------------------------------------------------------
 * Name :	do_disklabel()
 *
 * Description: Create a disk label on the device
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
static int
do_disklabel(char *device_name, char *label_name)
{
	int rc;
	printf("Creating the disk label...");
	fflush(stdout);
	become_root();
	rc = execute_command("/sbin/disklabel", 3,
						   "-w", device_name, label_name, NULL);
	unbecome_root();
	printf("done\n");
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	make_filesystems()
 *
 * Description: Create filesystems for the devices specified by fs_spec
 * field in vector of struct fstab*.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *----------------------------------------------------------------- */
static int
make_filesystems(vector v)
{
	int i;
	int error = 0;
	become_root();
	for (i = 0; error == 0 && i < vector_max(v); ++i) {
		char cmd[MAXPATHLEN];
		struct fstab *fs = vector_lookup_index(v, i);
		printf("Making filesystem on %s...", fs->fs_spec);
		fflush(stdout);
		sprintf(cmd, "/sbin/newfs %s 2>/dev/null 1>/dev/null\n",
				fs->fs_spec);
		error = system(cmd);
		printf(error ? "failed\n" : "done\n");
	}
	unbecome_root();
	return error;
}


/*-----------------------------------------------------------------
 * Name :	do_copy_files()
 *
 * Description: Copy the files from fromdir to todir.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
int
do_copy_files(char *fromdir, char *todir, char *files)
{
	char cmd[1024];
	sprintf(cmd, "(cd %s; /bin/tar -cf - %s 2>/dev/null) | (cd %s; tar -xvpf - 2>/dev/null)",
			fromdir, files, todir);
	return system(cmd);
}


/*-----------------------------------------------------------------
 * Name :	do_copy_init()
 *
 * Description: Label and create the filesystems on the secondary disk
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
int
do_copy_init(int argc, char **argv)
{
	size_t disk_blocks;
	size_t block_size;
	int error = CMD_ERR_FAILED;
	vector v = make_secondary_fstabs();
	
	if (disk_size(secondary_device, &disk_blocks, &block_size) == 0) {
		char *name;
		name = disk_name(disk_blocks, block_size);
		if (name != NULL) {
			error = do_disklabel(secondary_device, name);
			if (!error) {
				error = make_filesystems(v);
			}
		}
		else {
			printf("Disk size %lld bytes is not supported\n",
				   (uint64_t)disk_blocks * (uint64_t)block_size);
		}
	}

	free_fstab_vector(v);
	
	return error;
	
}


/*-----------------------------------------------------------------
 * Name :	copy_to_secondary()
 *
 * Description: Recursively copy the files specified by the argument
 * files to the secondary disk. If a file is a directory, all files
 * below it will be copied.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
static int
copy_to_secondary(char *files)
{
	int error;
	vector v = make_secondary_fstabs();

	become_root();

	if ((error = mount_secondary(v)) == 0) {
		printf("Mounted secondary disk\n");
		printf("Starting to copy the files to the secondary disk\n");
		error = do_copy_files("/", secondary_mount, files);
	}

	u_mount_secondary(v);

	unbecome_root();

	free_fstab_vector(v);
			
	return error == 0 ? CMD_SUCCESS : CMD_ERR_FAILED;
}


/*-----------------------------------------------------------------
 * Name :	do_copy_all()
 *
 * Description: Copy the system from primary to secondary disk.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
int
do_copy_all(int argc, char **argv)
{
	return copy_to_secondary(all_files);
}


/*-----------------------------------------------------------------
 * Name :	do_copy_config
 *
 * Description: Copy the system configuration files from primary to
 * secondary disk.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/25/02
 *
 *-----------------------------------------------------------------
 */
int
do_copy_config(int argc, char **argv)
{
	return copy_to_secondary(sys_config_files);
}

static int
y_or_n(void)
{
	int first, ch;
	first = ch = getchar();
	while (ch != '\n' && ch != EOF)
		ch = getchar();
	return (first == 'y' || first == 'Y');
}

		
/*-----------------------------------------------------------------
 * Name :	do_restore_config()
 *
 * Description: Restore the system configuration files from secondary
 * disk.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/26/02
 *
 *-----------------------------------------------------------------
 */
int
do_restore_config(int argc, char **argv)
{
	int error = 0;
	vector v = make_secondary_fstabs();

	printf("The changes in the configuration since the last "
		   "save will be lost\n"
		   "Do you wish to proceed (y/n)? ");
	fflush(stdout);
	if (y_or_n()) {
		become_root();

		if ((error = mount_secondary(v)) == 0) {
			printf("Mounted secondary disk\n");
			shutdown_all();
			if (noisy_make_allfs_writable() == 0) {
				printf("Starting to copy the files to the primary disk\n");
				error = do_copy_files(secondary_mount, "/", sys_config_files);
				reboot_all();
			}
		}

		u_mount_secondary(v);

		unbecome_root();
	}

	free_fstab_vector(v);
			
	return error;
}


/*-----------------------------------------------------------------
 * Name :	do_reset_config()
 *
 * Description: Reset the system configuration to factory defaults
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/06/02
 *
 *-----------------------------------------------------------------
 */
int
do_reset_config(int argc, char **argv)
{
	printf("All system configuration will be lost.\n");
	printf("Do you wish to proceed (y/n)? ");
	fflush(stdout);
	if (y_or_n()) {
		become_root();
		
		shutdown_all();

		if (noisy_make_allfs_writable() == 0) {
		
			(void)unlink("/etc/hostname.sc1");
			(void)unlink("/etc/hostname.sc2");
			(void)unlink("/etc/myname");
			(void)unlink("/etc/mygate");
			(void)unlink("/etc/pwd.db");
			(void)unlink("/etc/spwd.db");
			(void)unlink("/etc/passwd");
			(void)unlink("/usr/local/agile/conf/agile.conf");
			(void)unlink("/usr/local/agile/conf/agile.conf.bak");

			system("/bin/cp /etc/hosts.org /etc/hosts");
		
			system("/bin/cp /etc/master.passwd.org /etc/master.passwd");
			chmod("/etc/master.passwd", 0600);

			(void)unlink("/etc/agile.installed");

			reboot_all();
		}

		unbecome_root();
	}
	return 0;
}
