/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 * RCS          : $Id: cmd_lvol.c,v 1.50 2002/11/07 01:14:46 rick Exp $
 *
 * Name: cmd_lvol.c
 *
 * Description  : Diag commands for the logical volumes
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 05-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include "nfx-incl.h"
#include "exports/anpssc-api.h"
#include "exports/ea-api.h"
#include "exports/timer-api.h"

/* Declare globals used by eee_rtc calls.  The actual values will be returned
 * as part of the responses.
 */
uint64 count_per_usec = 1;
uint64 count_per_msec = 1;

#define LVOL_COMMAND_STR  "Logical volume diagnostic commands.\n"

/* Wait for EVM to respond.
 * We'll wait 60 seconds.
 */
static struct timeval cfgd_to = {60, 0};
static struct timeval long_cfgd_to = {180, 0};

/*------------------------------------------------------------------------------
 * Types definition.
 *------------------------------------------------------------------------------
 */

typedef struct
{
    char    buf[EVM_MAX_BUF_SIZE];
    uint32  len;
} msg_t;

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static boolean
onlyDigit(char *    str) ;

static
int bad_stat (int stat);

static
int bad_resp (int stat, char *resp);

static
int get_cfg (evm_volId_t            id,
             evm_MsgVolCreate_t *   cfg);

extern evm_volId_t nfxsh_get_vol_id(char *   name);

static evm_volId_t get_newVolId(void);

static
void display_cfg(evm_VolConfig_t *  cfg,
                 int32              state);

uint64 mbyte2block_64bit(uint64 mb);
uint64 block2mbyte_64bit(uint64 blk);
uint32 mbyte2block_32bit(uint32 mb);
uint32 block2mbyte_32bit(uint32 blk);

static void
displayStats(evm_stats_t * stats);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

uint64 mbyte2block_64bit(uint64 mb)
{
    uint64 blk = 0;

    NFX_ASSERT(EVM_LBLK_SIZE == 512);
    if ((mb == 0) || (mb >= ((uint64)1 << 53)))
    {
        printf("input is out of range (0 < X < 2^53)\n");
        return(0);
    }
    blk = mb * EVM_BLK_PER_MBYTES;

    return(blk);
}

uint64 block2mbyte_64bit(uint64 blk)
{
    uint64 mb = 0;
    
    mb = blk / EVM_BLK_PER_MBYTES;

    return(mb);
}

uint32 mbyte2block_32bit(uint32 mb)
{
    uint32 blk = 0;

    NFX_ASSERT(EVM_LBLK_SIZE == 512);
    if ((mb == 0) || (mb >= ((uint32)1 << 21)))
    {
        printf("input is out of range (0 < X < 2^21)\n");
        return(0);
    }
    blk = mb * EVM_BLK_PER_MBYTES;

    return(blk);
}

uint32 block2mbyte_32bit(uint32 blk)
{
    uint32 mb = 0;
    
    mb = blk / EVM_BLK_PER_MBYTES;

    return(mb);
}

/*------------------------------------------------------------------------------
 * Name         : onlyDigit
 *
 * Description  : Return TRUE if the string contains only digits.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 20-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static boolean
onlyDigit(char *    str) 
{
    int32  cnt;

    for (cnt=0; cnt < strlen(str); cnt++)
    {
        if (!isdigit(str[cnt]))
        {
            return(FALSE);
        }
    } // for
    return(TRUE);
}

/*------------------------------------------------------------------------------
 * Name         : bad_stat
 *
 * Description  : General check of sendAgileMsg() returned status
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

static int
bad_stat(int stat)
    
{
    if (stat < 0)
    {
        perror("Send to EVM failed.");
        return 1;
    }
    else if (!stat)
    {
        printf("Timed out waiting for response.\n");
        return 1;
    }
    return 0;
}

/*------------------------------------------------------------------------------
 * Name         : bad_resp
 *
 * Description  : General check of response from evm_cfgd
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

static
int bad_resp(int stat, char *resp)

{
    evm_MsgStat_t * stat_msg = (evm_MsgStat_t *)    resp;
    char *          rtnstr = &resp[sizeof(evm_MsgStat_t)];

    if (stat < sizeof(evm_MsgStat_t))
    {
        printf("Response message truncated.\n");
        return 1;
    }
#ifdef ISABELLE // code to delete
    /* Check the returned status to make sure it's from us and the status
     * is OK.
     */
    else if (stat_msg->hdr.clientPrivate != getpid())
    {
        printf("Response wasn't from our message.\n");
        return 1;
    }
#endif // ISABELLE

    /* If we got an error, they might have hidden a string at the end of
     * the response.
     */
    else if (stat_msg->stat != NFX_OK)
    {
        printf("Response indicates message was rejected.\n");
        if ((stat > sizeof(evm_MsgStat_t)) && *rtnstr)
        {
            printf("%s\n", rtnstr);
        }
        return 1;
    }
    return 0;
}

/*------------------------------------------------------------------------------
 * Name         : get_cfg
 *
 * Description  : Get a LV config given its id
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *-----------------------------------------------------------------
 */

static
int get_cfg (evm_volId_t            id,
             evm_MsgVolCreate_t *   cfg)

{
    evm_MsgVolGetAttrib_t   attrrq;
    int                     stat;

    memset(&attrrq, 0, sizeof(evm_MsgVolGetAttrib_t));
    attrrq.hdr.msgType = EVM_VOL_GET_ATTRIB;
    attrrq.hdr.clientPrivate = getpid();
    attrrq.id = id;

    stat = sendAgileMsg(-1, &attrrq, sizeof(attrrq),
                        "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to,
                        cfg, sizeof(evm_MsgVolCreate_t));
    /* Did we get what we expected?
     */
    if (bad_stat (stat))
    {
        return 1;
    }
    else if ((stat == sizeof(evm_MsgVolCreate_t)) &&
             (cfg->hdr.msgType == EVM_VOL_CONFIG))
    {
        return 0;
    }
    else
    {
        bad_resp (stat, (char *)cfg);
        return 1;
    }
}

/*------------------------------------------------------------------------------
 * Name         : nfxsh_get_vol_id
 *
 * Description  : Given a name, get the LV id
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

evm_volId_t
nfxsh_get_vol_id(char *   name)

{
    evm_MsgGetId_t  getReq;
    msg_t           msg;
    evm_MsgStat_t * getRsp = (evm_MsgStat_t *) msg.buf;

    bzero((void *) &getReq, sizeof(evm_MsgGetId_t));
    bzero((void *) &msg, sizeof(msg));

    getReq.hdr.msgType = EVM_VOL_GET_ID;
    getReq.hdr.clientPrivate = getpid();
    strcpy (getReq.name, name);

    msg.len = sendAgileMsg (-1, &getReq, sizeof(evm_MsgGetId_t),
                            "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &cfgd_to,
                            msg.buf, sizeof(msg.buf));

    if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
    {
        getRsp->payload.lVolId.whole_thing = -1;
    }

    return(getRsp->payload.lVolId);
}

/*------------------------------------------------------------------------------
 * Name         : get_newVolId
 *
 * Description  : Get a new volume id
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-JUL-2002
 *
 *------------------------------------------------------------------------------
 */

static evm_volId_t
get_newVolId(void)

{
    evm_MsgGetNewVolId_t    getReq;
    msg_t                   msg;
    evm_MsgStat_t *         getRsp = (evm_MsgStat_t *) msg.buf;

    bzero((void *) &getReq, sizeof(evm_MsgGetNewVolId_t));
    bzero((void *) &msg, sizeof(msg));

    getReq.hdr.msgType = EVM_VOL_GET_NEWVOLID;
    getReq.hdr.clientPrivate = getpid();

    msg.len = sendAgileMsg(-1, &getReq, sizeof(evm_MsgGetNewVolId_t),
                           "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &cfgd_to, msg.buf, sizeof(msg.buf));

    if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
    {
        getRsp->payload.lVolId.whole_thing = -1;
    }

    return(getRsp->payload.lVolId);
}

/*------------------------------------------------------------------------------
 * Name         : display_cfg
 *
 * Description  : Display a single LV config
 *
 * Created by   : Greg Seibert
 *
 * Date Created : 20-JUN-2001
 *
 *------------------------------------------------------------------------------
 */

static void
display_cfg(evm_VolConfig_t *  cfg,
            int32              state)

{
    evm_MsgGetStrg_t    getReq;
    msg_t               msg;
    evm_MsgLVStrg_t *   getRsp = (evm_MsgLVStrg_t *) msg.buf;
    int                 i;

    printf("\nLogical volume name: %s, id: 0x%llx\n",
           cfg->name, cfg->id.whole_thing);
     printf("  current size (MB): %lld, minautogw (MB): %lld\n",
            block2mbyte_64bit(cfg->currSize),
            block2mbyte_64bit(cfg->minAutoGrow));
    printf("  hard quota (MB): %lld, soft quota (MB): %d\n",
           block2mbyte_64bit(cfg->maxSize),
           cfg->softQuota);
    printf("  nim: %d, online: %s\n",
           cfg->onNIM, cfg->onLine?"Yes":"No");

    bzero((void *) &getReq, sizeof(evm_MsgGetStrg_t));
    bzero((void *) &msg, sizeof(msg));

    getReq.hdr.msgType = EVM_GET_LV_STRG;
    getReq.hdr.clientPrivate = getpid();
    getReq.id = cfg->id;

    msg.len = sendAgileMsg(-1, &getReq, sizeof(evm_MsgGetStrg_t),
                           "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &cfgd_to, msg.buf, sizeof(msg.buf));

    if (!bad_stat(msg.len))
    {
        if (getRsp->hdr.msgType == EVM_LV_STRG_CFG)
        {
            for (i = 0; i < getRsp->nbStrg; i++)
            {
                printf("  - lun #%d\n", i + 1);
                printf("    scsi id: 0x%llx, size (MB): %lld\n",
                       getRsp->strg[i].au.strgId,
                       block2mbyte_64bit(getRsp->strg[i].au.auSize));
            }
        }
        else
        {
            printf("Could not get LV's storage config.\n");
            bad_resp(msg.len, msg.buf);
        }
    }
    if (state == EVM_VOL_NOT_READY)
    {
        printf("  *** volume missing LUNs, will not be put online ***\n");
    }
    return;
}

/*------------------------------------------------------------------------------
 * Name         : displayStats
 *
 * Description  : Displays the EVM server statistics.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 28-JUL-2002
 *------------------------------------------------------------------------------
 */

static void
displayStats(evm_stats_t * stats)
    
{
    int slot, cpu;
    
	/* Set the value of the rtc to usec conversion factor as
	 * returned by the server.
	 */
	count_per_usec = stats->count_per_usec;
	count_per_msec = count_per_usec * MSEC_PER_SEC;

    printf("\n");
    printf("Number of calls to evm_IOReq ................... %lld\n",
           stats->nIOReqs);
    printf("Number evm_IOReq user callbacks ................ %lld\n",
           stats->nIOCompletes);
    printf("Meta read:\n");
    printf("     Number of requests ........................ %lld\n",
           stats->nMReads);
    printf("     Number of retries ......................... %lld\n",
           stats->nMReadsRetry);
    printf("     Number of responses ....................... %lld\n",
           stats->nScsiMReadRsps);
    printf("     Number of bytes ........................... %lld\n",
           stats->nBlkMRead * EVM_LBLK_SIZE);
    printf("Meta write:\n");
    printf("     Number of requests ........................ %lld\n",
           stats->nMWrites);
    printf("     Number of retries ......................... %lld\n",
           stats->nMWritesRetry);
    printf("     Number of responses ....................... %lld\n",
           stats->nScsiMWriteRsps);
    printf("     Number of bytes ........................... %lld\n",
           stats->nBlkMWrite * EVM_LBLK_SIZE);
    printf("User read:\n");
    printf("     Number of requests ........................ %lld\n",
           stats->nUReads);
    printf("     Number of retries ......................... %lld\n",
           stats->nUReadsRetry);
    printf("     Number of responses ....................... %lld\n",
           stats->nScsiUReadRsps);
    printf("     Number of bytes ........................... %lld\n",
           stats->nBlkURead * EVM_LBLK_SIZE);
    printf("User write:\n");
    printf("     Number of requests ........................ %lld\n",
           stats->nUWrites);
    printf("     Number of retries ......................... %lld\n",
           stats->nUWritesRetry);
    printf("     Number of responses ....................... %lld\n",
           stats->nScsiUWriteRsps);
    printf("     Number of bytes ........................... %lld\n",
           stats->nBlkUWrite * EVM_LBLK_SIZE);
    printf("\n");
    printf("Total number of read operations ................ %lld\n",
           stats->nScsiMReadRsps +
           stats->nScsiUReadRsps);
    printf("Total number of bytes Read ..................... %lld\n",
           stats->nBlkMRead * EVM_LBLK_SIZE + 
           stats->nBlkURead * EVM_LBLK_SIZE);
    printf("Total number of write operations ............... %lld\n",
           stats->nMWrites +
           stats->nMWritesRetry +
           stats->nUWrites +
           stats->nUWritesRetry);
    printf("Total number of bytes Written .................. %lld\n",
           stats->nBlkUWrite * EVM_LBLK_SIZE +
           stats->nBlkMWrite * EVM_LBLK_SIZE);    
    printf("\n");
    printf("Number evm_IOReq rejected ...................... %lld\n",
           stats->nIORejected);
    printf("Number of SCSI device errors ................... %lld\n",
           stats->nScsiDevErrs);
    printf("Number of errors ............................... %lld\n",
           stats->nErrors);
    printf("Number of IO Timeouts .......................... %lld\n",
           stats->nTimeouts);
    printf("Number of EEE Tx Queue Full .................... %lld\n",
           stats->nEEETxQueueFull);
    printf("Number of stale IO responses ................... %lld\n",
           stats->nStales);
    printf("Number of bad packet length .................... %lld\n",
           stats->nBadPacketLength);
    printf("\n");
    printf("Number of memory allocations ................... %lld\n",
           stats->nAllocs);
    printf("Number of memory deallocations ................. %lld\n",
           stats->nDeallocs);
    printf("Number of timer allocations .................... %lld\n",
           stats->nTimerCreates);
    printf("Number of timer deallocations .................. %lld\n",
           stats->nTimerDestroys);
    printf("\n");
    printf("Meta I/O queue:\n");
    printf("     Number of free elements ................... %d\n",
           stats->qCnt.metaFree);
    printf("     Number of waiting physical requests ....... %d (got %d)\n",
           stats->qCnt.metaPioWait, stats->nMetaMaxPioWait);
    printf("     Number of pending physical requests ....... %d (got %d)\n",
           stats->qCnt.metaPioPend, stats->nMetaMaxPioPend);
    printf("User I/O queue:\n");
    printf("     Number of free elements ................... %d\n",
           stats->qCnt.userFree);
    printf("     Number of waiting physical requests ....... %d (got %d)\n",
           stats->qCnt.userPioWait, stats->nUserMaxPioWait);
    printf("     Number of pending physical requests ....... %d (got %d)\n",
           stats->qCnt.userPioPend, stats->nUserMaxPioPend);
    printf("Total I/O queues (meta + user):\n");
    printf("     Max. number of waiting physical requests .. %d\n",
           stats->nMaxPioWait);
    printf("     Max. number of pending physical requests .. %d\n",
           stats->nMaxPioPend);
    printf("\n");
#ifdef DEBUG
    printf("Queue depth .................................... %lld (got %lld)\n",
           (stats->numQueueDepth == 0)? 0 :
           (stats->sumQueueDepth/stats->numQueueDepth),
           stats->maxQueueDepth);
    printf("\n");
#endif // DEBUG
    printf("IO response time (in msec):\n");
    printf("more than  %7d    %7d    %7d    %7d    "
           "%7d    %7d        maxTime    avgTime    \n",
           EVM_AMSEC, EVM_BMSEC, EVM_CMSEC, EVM_DMSEC, EVM_EMSEC, EVM_FMSEC);
    printf("-----------------------------------------------------------"
           "-------------------------------------------\n");
    printf("META READ .... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10.2f\n",
           stats->MReadTime.moreThanAmsec,
           stats->MReadTime.moreThanBmsec,
           stats->MReadTime.moreThanCmsec,
           stats->MReadTime.moreThanDmsec,
           stats->MReadTime.moreThanEmsec,
           stats->MReadTime.moreThanFmsec,
           eee_rtcToMS(&stats->MReadTime.maxTime),
           (stats->MReadTime.numTime == 0)? 0:
		   eee_rtcDivMS(&stats->MReadTime.sumTime, stats->MReadTime.numTime));
    printf("META WRITE ... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10.2f\n",
           stats->MWriteTime.moreThanAmsec,
           stats->MWriteTime.moreThanBmsec,
           stats->MWriteTime.moreThanCmsec,
           stats->MWriteTime.moreThanDmsec,
           stats->MWriteTime.moreThanEmsec,
           stats->MWriteTime.moreThanFmsec,
           eee_rtcToMS(&stats->MWriteTime.maxTime),
           (stats->MWriteTime.numTime == 0)? 0:
           eee_rtcDivMS(&stats->MWriteTime.sumTime, stats->MWriteTime.numTime));
    printf("USER READ .... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10.2f\n",
           stats->UReadTime.moreThanAmsec,
           stats->UReadTime.moreThanBmsec,
           stats->UReadTime.moreThanCmsec,
           stats->UReadTime.moreThanDmsec,
           stats->UReadTime.moreThanEmsec,
           stats->UReadTime.moreThanFmsec,
           eee_rtcToMS(&stats->UReadTime.maxTime),
           (stats->UReadTime.numTime == 0)? 0:
           eee_rtcDivMS(&stats->UReadTime.sumTime, stats->UReadTime.numTime));
    printf("USER WRITE ... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10.2f\n",
           stats->UWriteTime.moreThanAmsec,
           stats->UWriteTime.moreThanBmsec,
           stats->UWriteTime.moreThanCmsec,
           stats->UWriteTime.moreThanDmsec,
           stats->UWriteTime.moreThanEmsec,
           stats->UWriteTime.moreThanFmsec,
           eee_rtcToMS(&stats->UWriteTime.maxTime),
           (stats->UWriteTime.numTime == 0)? 0:
           eee_rtcDivMS(&stats->UWriteTime.sumTime, stats->UWriteTime.numTime));
    printf("\n");
    for (slot = 0; slot < 1; slot++)
    {
        for (cpu = 0; cpu < 4; cpu++)
        {
            printf("Number of IOs exchanged with FC-%d cpu %d\n", slot, cpu);
            printf("\tmeta read req/rsp .................... %lld / %lld\n",
                   stats->ioPerFc[slot][cpu].metaReadReq,
                   stats->ioPerFc[slot][cpu].metaReadRsp);
            printf("\tmeta write req/rsp ................... %lld / %lld\n",
                   stats->ioPerFc[slot][cpu].metaWriteReq,
                   stats->ioPerFc[slot][cpu].metaWriteRsp);
            printf("\tuser read req/rsp .................... %lld / %lld\n",
                   stats->ioPerFc[slot][cpu].userReadReq,
                   stats->ioPerFc[slot][cpu].userReadRsp);
            printf("\tuser write req/rsp ................... %lld / %lld\n",
                   stats->ioPerFc[slot][cpu].userWriteReq,
                   stats->ioPerFc[slot][cpu].userWriteRsp);
        } // for (cpu)
    } // for (slot)
    printf("\n");
 }

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : lvol_show
 *
 * Description  : Show volume configuration.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 30-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_show,
      lvol_show_cmd,
      "lvol show [VOLNAME]",
      "\nShow logical volume(s) configuration.\n")

{
    evm_MsgVolGetAttrib_t   attrrq;
    evm_MsgVolList_t        vol_list;
    evm_MsgVolCreate_t      cfg;
    evm_volId_t             id;
    int                     stat;

    if (argc == 0)  // Show all volumes
    {
        /* First, get a list of ALL volume ids
         */
        memset(&attrrq, 0, sizeof(evm_MsgVolGetAttrib_t));
        attrrq.hdr.msgType = EVM_VOL_GET_ATTRIB;
        attrrq.hdr.clientPrivate = getpid();
        attrrq.id.whole_thing = -1;

        stat = sendAgileMsg (-1, &attrrq, sizeof(attrrq),
                             "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                             &cfgd_to,
                             &vol_list, sizeof(vol_list));
        /* Did we get what we expected?
         */
        if (!bad_stat (stat) && (stat == sizeof(evm_MsgVolList_t)) &&
            (vol_list.hdr.msgType == EVM_VOL_LIST))
        {
            int num;

            /* Get and display configuration for each volume.
             */
            for (num = 0; num < vol_list.num; num++)
            {
                if (get_cfg(vol_list.ids[num], &cfg) != NFX_OK)
                {
                    printf("Error getting cfg for %llx.\n",
                           vol_list.ids[num].whole_thing);
                    return(CMD_ERR_FAILED);
                }
                else
                {
                    display_cfg (&(cfg.config), cfg.state);
                }
            } // for
        }
        else
        {
            printf("Failed to get list of vol ids.\n");
            return(CMD_ERR_FAILED);
        }
    }
    else if (argc == 1) // Show one specific volume
    {
        /* Get and display configuration for the volume.
         */
        id = nfxsh_get_vol_id(argv[0]);
        if (id.whole_thing == -1)
        {
            printf("Error getting id for %s\n", optarg);
            return(CMD_ERR_FAILED);
        }
        else
        {
            if (get_cfg (id, &cfg) != NFX_OK)
            {
                printf("Error getting cfg for %s\n", optarg);
                return(CMD_ERR_FAILED);
            }
            else
            {
                display_cfg (&(cfg.config), cfg.state);
            }
        }
    }
    else
    {
        printf("Invalid options!\n");
        return(CMD_ERR_FAILED);
    }

    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_create
 *
 * Description  : Create a  volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_create,
      lvol_create_cmd,
      "lvol create VOLNAME [-o] [-L LUN] [-n NIM] "
      "[-q HARDQUOTA] [-s SOFTQUOTA] [-g MINAUTOGROW] "
      "[-r RAIDLEVEL] [-l MINLUNSIZE] [-u MAXLUNSIZE] "
      "[-d MAKE] [-m MODEL] [-R RAIDCONTROLLER] [-t VOLTYPE]",
      "\nCreate a volume.\n")

{
    evm_MsgVolCreate_t  createReq;
    msg_t               msg;
    evm_MsgStat_t *     createRsp = (evm_MsgStat_t *) msg.buf;
    int                 c;
    extern int          optind;

    bzero((void *) &createReq, sizeof(evm_MsgVolCreate_t));
    bzero((void *) &msg, sizeof(msg));

    createReq.hdr.msgType = EVM_VOL_CREATE;
    createReq.hdr.clientPrivate = NFX_STRG_MAGIC;

#ifdef NEWVOLID
    createReq.config.id = get_newVolId();
    if (createReq.config.id.whole_thing == -1)
    {
        printf("Cannot get a new volume id.\n");
        return(CMD_ERR_FAILED);
    }
#else
    createReq.config.id.whole_thing = 0;
#endif
    strncpy(createReq.config.name, argv[0], VOL_MAX_NAME);
    createReq.config.type = EVM_TYPE_FSYS;
    createReq.config.maxSize = 0;
    createReq.config.softQuota = 0;
    createReq.config.minAutoGrow = 0;
    createReq.config.onNIM = 1;
    createReq.config.onLine = FALSE;
    createReq.config.attrs.raidLevel = EVM_RAID_ANY;
    createReq.config.attrs.minLunSize = EVM_DONT_CARE;
    createReq.config.attrs.maxLunSize = EVM_DONT_CARE;
    createReq.config.attrs.flags = EVM_SAME_VENDOR | EVM_SAME_MODEL;
    strcpy(createReq.config.attrs.raidCtrl, "");
    createReq.config.attrs.num = 0;
    createReq.nbLun = 0;

#ifdef ISABELLE // debug only
    {
        int i;

        for (i = 0; i < argc; i++)
        {
            printf("argv[%d] = %s\n", i, argv[i]);
        } // for
    }
#endif // ISABELLE

    optind = 0;
    while ((c = getopt(argc, argv, "on:q:s:g:r:l:u:d:m:R:L:t:")) != -1)
    {
        switch (c)
        {
        case 'o':
            createReq.config.onLine = TRUE;
            printf("online\n");
            break;

        case 'n':
            createReq.config.onNIM = atoi(optarg);
            printf("nim = %d\n", createReq.config.onNIM);
            break;

        case 'q':
            if (onlyDigit(optarg) == FALSE)
            {
                printf ("invalid value for hardquota\n");
                return(CMD_ERR_FAILED);
            }
            createReq.config.maxSize =
                mbyte2block_64bit(strtoq(optarg, NULL, 0));
            if (createReq.config.maxSize == 0)
            {
                return(CMD_ERR_FAILED);
            }
            printf("hardquota (512 blocks) = %lld\n", createReq.config.maxSize);
            break;

        case 's':
            createReq.config.softQuota = atoi(optarg);
            if ((onlyDigit(optarg) == FALSE) ||
                (createReq.config.softQuota < 0) ||
                (createReq.config.softQuota > 100))
            {
                printf ("invalid value for softquota\n");
                return(CMD_ERR_FAILED);
            }
            printf("softQuota = %d\n", createReq.config.softQuota);
            break;

        case 'g':
            if (onlyDigit(optarg) == FALSE)
            {
                printf ("invalid value for minAutoGrow\n");
                return(CMD_ERR_FAILED);
            }
            createReq.config.minAutoGrow =
                mbyte2block_64bit(strtoq(optarg, NULL, 0));
            if (createReq.config.minAutoGrow == 0)
            {
                return(CMD_ERR_FAILED);
            }
            printf("minAutoGrow (512 blocks) = %lld\n",
                   createReq.config.minAutoGrow);
            break;

        case 'r':
            if (strcasecmp(optarg, "Raid-0") == 0)
                createReq.config.attrs.raidLevel = EVM_RAID0;
            else if (strcasecmp(optarg, "Raid-1") == 0)
                createReq.config.attrs.raidLevel= EVM_RAID1;
            else if (strcasecmp(optarg, "Raid-01") == 0)
                createReq.config.attrs.raidLevel= EVM_RAID01;
            else if (strcasecmp(optarg, "Raid-3") == 0)
                createReq.config.attrs.raidLevel= EVM_RAID3;
            else if (strcasecmp(optarg, "Raid-5") == 0)
                createReq.config.attrs.raidLevel= EVM_RAID5;
            else if (strcasecmp(optarg, "Plain") == 0)
                createReq.config.attrs.raidLevel= EVM_RAID_NONE;
            else if (strcasecmp(optarg, "Any") == 0)
                createReq.config.attrs.raidLevel= EVM_RAID_ANY;
            else
            {
                printf("invalid value for RAID level\n");
                return(CMD_ERR_FAILED);
            }
            printf("raidLevel = %d\n", createReq.config.attrs.raidLevel);
            break;

        case 'l':
            if (onlyDigit(optarg) == FALSE)
            {
                printf ("invalid value for minLunSize\n");
                return(CMD_ERR_FAILED);
            }
            createReq.config.attrs.minLunSize = mbyte2block_32bit(atoi(optarg));
            if (createReq.config.attrs.minLunSize == 0)
            if (onlyDigit(optarg) == FALSE)
            {
                return(CMD_ERR_FAILED);
            }
            printf("minLunSize (512 blocks) = %d\n",
                   createReq.config.attrs.minLunSize);
            break;

        case 'u':
            if (onlyDigit(optarg) == FALSE)
            {
                printf ("invalid value for maxLunSize\n");
                return(CMD_ERR_FAILED);
            }
            createReq.config.attrs.maxLunSize = mbyte2block_32bit(atoi(optarg));
            if (createReq.config.attrs.maxLunSize == 0)
            if (onlyDigit(optarg) == FALSE)
            {
                return(CMD_ERR_FAILED);
            }
            printf("maxLunSize (512 blocks) = %d\n",
                   createReq.config.attrs.maxLunSize);
            break;

        case 'd':
            createReq.config.attrs.num = 1;
            strcpy(createReq.config.attrs.vendorInfo[0].vendor, optarg);
            strcpy(createReq.config.attrs.vendorInfo[0].model, "");
            printf("vendor = %s\n",
                   createReq.config.attrs.vendorInfo[0].vendor);
            break;

        case 'm':
            strcpy(createReq.config.attrs.vendorInfo[0].model, optarg);
            printf("model = %s\n", createReq.config.attrs.vendorInfo[0].model);
            break;

        case 'R':
            strcpy(createReq.config.attrs.raidCtrl, optarg);
            printf("raidCtrl = %s\n", createReq.config.attrs.raidCtrl);
            break;

        case 'L':
            createReq.nbLun = 1;
            strcpy(createReq.lunList[0], optarg);
            printf("lun = %s\n", createReq.lunList[0]);
            break;

        case 't':
            if (!strcmp(optarg, "corevol"))
                createReq.config.type = EVM_TYPE_COREVOL;
            printf("type = %d\n", createReq.config.type);
            break;

        default:
            printf("Invalid options!\n");
            return(CMD_ERR_FAILED);
        } // switch
    } // while

    if ((createReq.config.maxSize != 0) &&
        (createReq.config.minAutoGrow == 0))
    {
        printf("Cannot specify hardQuota without specifying minAutoGrow.\n");
        return(CMD_ERR_FAILED);
    }
    
    if ((createReq.nbLun != 0) && (createReq.config.minAutoGrow != 0))
    {
        printf("Cannot request autogrow if you specify which LUN to use.\n");
        return(CMD_ERR_FAILED);
    }
    
    /*  Send this request to evmcfgd and wait for it to tell us it's LVolid
     */
    msg.len = sendAgileMsg (-1, &createReq, sizeof(evm_MsgVolCreate_t),
                            "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &cfgd_to,
                            msg.buf, sizeof(msg.buf));

    if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
    {
        printf("Create failed.\n");
        return(CMD_ERR_FAILED);
    }

    printf("Created logical volume '%s' / 0x%llx.\n",
           argv[0], createRsp->payload.lVolId.whole_thing);

    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_delete
 *
 * Description  : Destroy a logical volume
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_delete,
      lvol_delete_cmd,
      "lvol delete VOLNAME",
      "\nDelete a volume.\n")

{
    evm_MsgVolDestroy_t delReq;
    evm_volId_t         id;
    msg_t               msg;

    id = nfxsh_get_vol_id(argv[0]);
    if (id.whole_thing != -1)
    {
        bzero(&delReq, sizeof(evm_MsgVolDestroy_t));

        delReq.hdr.msgType = EVM_VOL_DESTROY;
        delReq.hdr.clientPrivate = getpid();
        delReq.id = id;

        msg.len = sendAgileMsg(-1, &delReq, sizeof(evm_MsgVolDestroy_t),
                               "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                               &cfgd_to,
                               msg.buf, sizeof(msg.buf));

        if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
        {
            printf("Destroy failed.\n");
            return(CMD_ERR_FAILED);
        }
    }
    else
    {
        return(CMD_ERR_FAILED);
    }

    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_options
 *
 * Description  : Set options on the logical volumes.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_modify,
      lvol_modify_cmd,
      "lvol modify VOLNAME [-p NEWVOLNAME] "
      "[-q HARDQUOTA]  [-S SOFTQUOTA] [-g MINAUTOGROW]",
      "\nModify logical volume parameters.\n")

{
    evm_volId_t         id;
    evm_MsgVolCreate_t  cfg;
    evm_MsgSetOpt_t     optionMsg;
    evm_MsgVolRename_t  renameMsg;
    msg_t               msg;
    int                 c;
    extern int          optind;
    boolean             renameFlag = FALSE;
    boolean             optionFlag = FALSE;

    bzero((void *) &msg, sizeof(msg));

    /* After we get the id of the specified LV, we get the current option
     * config to start with.
     */
    id = nfxsh_get_vol_id(argv[0]);
    if (id.whole_thing != -1)
    {
        if (!get_cfg(id, &cfg))
        {
            bzero(&optionMsg, sizeof(optionMsg));
            optionMsg.hdr.msgType = EVM_SET_LV_OPTS;
            optionMsg.hdr.clientPrivate = getpid();
            optionMsg.id = id;
            optionMsg.maxSize = cfg.config.maxSize;
            optionMsg.softQuota = cfg.config.softQuota;
            optionMsg.minAutoGrow = cfg.config.minAutoGrow;
            optionMsg.type = cfg.config.type;

            bzero(&renameMsg, sizeof(renameMsg));
            renameMsg.hdr.msgType = EVM_VOL_RENAME;
            renameMsg.hdr.clientPrivate = getpid();
            renameMsg.id = id;

            /* Modify the current option's values based upon what the arguments
             * to this command were.
             */
            optind = 0;
            while ((c = getopt(argc, argv, "p:q:S:g:")) != -1)
            {
                switch (c)
                {
                case 'p':
                    renameFlag = TRUE;
                    strcpy(renameMsg.newName, optarg);
                    break;

                case 'q':
                    optionFlag = TRUE;
                    if (onlyDigit(optarg) == FALSE)
                    {
                        printf ("invalid value for hardquota\n");
                        return(CMD_ERR_FAILED);
                    }
                    optionMsg.maxSize =
                        mbyte2block_64bit(strtoq(optarg, NULL, 0));
                    if (optionMsg.maxSize == 0)
                    {
                        return(CMD_ERR_FAILED);
                    }
                    printf("hardquota (512 blocks) = %lld\n",
                           optionMsg.maxSize);
                    break;

                case 'S':
                    optionFlag = TRUE;
                    optionMsg.softQuota = atoi(optarg);
                    if ((onlyDigit(optarg) == FALSE) ||
                        (optionMsg.softQuota < 0) ||
                        (optionMsg.softQuota > 100))
                    {
                        printf ("invalid value for softquota\n");
                        return(CMD_ERR_FAILED);
                    }
                    printf("softQuota = %d\n", optionMsg.softQuota);
                    break;

                case 'g':
                    optionFlag = TRUE;
                    if (onlyDigit(optarg) == FALSE)
                    {
                        printf ("invalid value for hardquota\n");
                        return(CMD_ERR_FAILED);
                    }
                    optionMsg.minAutoGrow =
                        mbyte2block_64bit(strtoq(optarg, NULL, 0));
                    if (optionMsg.minAutoGrow == 0)
                    {
                        return(CMD_ERR_FAILED);
                    }
                    printf("minAutoGrow (512 blocks) = %lld\n",
                           optionMsg.minAutoGrow);
                    break;

                default:
                    printf("Invalid option!\n");
                } // switch
            } // while

            if ((optionMsg.maxSize != 0) && (optionMsg.minAutoGrow == 0))
            {
                printf("Cannot specify hardQuota without specifying "
                       "minAutoGrow.\n");
                return(CMD_ERR_FAILED);
            }
            
            if (optionFlag == TRUE)
            {
                msg.len = sendAgileMsg(-1, &optionMsg, sizeof(optionMsg),
                                       "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                       &cfgd_to,
                                       msg.buf, sizeof(msg.buf));
                if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
                {
                    printf("Option set failed\n");
                    return(CMD_ERR_FAILED);
                }
            }
            if (renameFlag == TRUE)
            {
                msg.len = sendAgileMsg(-1, &renameMsg, sizeof(renameMsg),
                                       "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                       &cfgd_to,
                                       msg.buf, sizeof(msg.buf));
                if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
                {
                    printf("Rename set failed\n");
                    return(CMD_ERR_FAILED);
                }
            }
        }
        else
        {
            return(CMD_ERR_FAILED);
        }
    }
    else
    {
        return(CMD_ERR_FAILED);
    }

    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_online
 *
 * Description  : Put a volume on-line
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_online,
      lvol_online_cmd,
      "lvol online VOLNAME [-n NIM] [-K]",
      "\nPut a logical volume on-line.\n")

{
    evm_MsgVolOnline_t  onReq;
    evm_volId_t         id;
    msg_t               msg;
    int                 c;
    extern int          optind;

    id = nfxsh_get_vol_id (argv[0]);
    if (id.whole_thing != -1)
    {
        bzero((void *) &onReq, sizeof(evm_MsgVolOnline_t));
        bzero((void *) &msg, sizeof(msg));

        onReq.hdr.msgType = EVM_VOL_ONLINE;
        onReq.hdr.clientPrivate = getpid();
        onReq.id = id;
        onReq.nim = 1;          // R1 limitation
        onReq.checkFlag = TRUE; // check if sbdy is already managing the volume
                                // before turning volume online

#ifdef ISABELLE // debug only
        {
            int i;

            for (i = 0; i < argc; i++)
            {
                printf("argv[%d] = %s\n", i, argv[i]);
            } // for
        }
#endif // ISABELLE

        optind = 0;
        while ((c = getopt(argc, argv, "n:K")) != -1)
        {
            switch (c)
            {
            case 'n':
                onReq.nim = atoi(optarg);
                printf("nim = %d\n", onReq.nim);
                break;

            case 'K':
                onReq.checkFlag = FALSE;
                printf("no check requested\n");
                break;

            default:
                printf("Invalid options!\n");
                return(CMD_ERR_FAILED);
            } // switch
        } // while

        msg.len = sendAgileMsg(-1, &onReq, sizeof(evm_MsgVolOnline_t),
                               "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                               &long_cfgd_to, msg.buf, sizeof(msg.buf));

        if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
        {
            printf("Online failed.\n");
            return(CMD_ERR_FAILED);
        }
    }
    else
    {
        return (CMD_ERR_FAILED);
    }

    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_grow
 *
 * Description  : Add explicit storage to logical volumes
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_grow,
      lvol_grow_cmd,
      "lvol grow VOLNAME LUN",
      "\nGrow manually a logical volume\n")

{
    evm_volId_t         id;
    evm_MsgVolCreate_t  cfg;
    evm_MsgVolStrg_t    volstrgReq;
    msg_t               msg;

    /* And the lvol id is:
     */
    id = nfxsh_get_vol_id (argv[0]);
    if (id.whole_thing < 0)
    {
        return(CMD_ERR_FAILED);
    }

    /* First we need to get the LV's config to figure out where to put the
     * storage allocations (mirrors before stripes in the 1+0 case). Also
     * can count allocation to ensure entire stripe/mirror group is specifed.
     */
    if (get_cfg(id, &cfg))
    {
        return(CMD_ERR_FAILED);
    }

    /* Parse the user's command and form the message.
     */
    bzero((void *) &volstrgReq, sizeof(evm_MsgVolStrg_t));
    volstrgReq.hdr.msgType = EVM_VOL_GROW;
    volstrgReq.hdr.clientPrivate = getpid();
    volstrgReq.id = id;
    volstrgReq.nbLun = 1;
    strcpy(volstrgReq.lunList[0], argv[1]);
    printf("lun = %s\n", volstrgReq.lunList[0]);

    /* We made it through all that so on appearance, the allocated storage
     * seems to match the logical volume.
     */
    bzero((void *) &msg, sizeof(msg));
    msg.len = sendAgileMsg(-1, &volstrgReq, sizeof(evm_MsgVolStrg_t),
                           "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &cfgd_to, msg.buf, sizeof(msg.buf));

    if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
    {
        printf("Storage addition failed\n");
        return(CMD_ERR_FAILED);
    }

    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_offline
 *
 * Description  : Take a volume off-line
 *
 * Created by   : Isabelle
 *
 * Date Created : 03-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_offline,
      lvol_offline_cmd,
      "lvol offline VOLNAME",
      "\nTake a logical volume offline.\n")

{
    evm_MsgVolOffline_t offReq;
    evm_volId_t         id;
    msg_t               msg;

    id = nfxsh_get_vol_id (argv[0]);
    if (id.whole_thing != -1)
    {
        bzero((void *) &offReq, sizeof(evm_MsgVolOffline_t));
        bzero((void *) &msg, sizeof(msg));

        offReq.hdr.msgType = EVM_VOL_OFFLINE;
        offReq.hdr.clientPrivate = getpid();
        offReq.id = id;
        offReq.nim = 0; // Don't care right now

        msg.len = sendAgileMsg(-1, &offReq, sizeof(offReq),
                               "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                               &cfgd_to,
                               msg.buf, sizeof(msg.buf));

        if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
        {
            printf("Offline failed\n");
            return(CMD_ERR_FAILED);
        }
    }
    else
    {
        return (CMD_ERR_FAILED);
    }

    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : lvol_takeover
 *
 * Description  : Takeover a volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 25-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(lvol_takeover,
      lvol_takeover_cmd,
      "lvol takeover VOLNAME [-n NIM] [-T MAXTAKEOVERTIME]",
      "\nTakeover a logical volume.\n")

{
    evm_volId_t             volId;
    evm_MsgVolTakeover_t    req;
    msg_t                   msg;
    //evm_MsgStat_t *         rsp = (evm_MsgStat_t *) msg.buf;
    int                     c;
    extern int              optind;

    volId = nfxsh_get_vol_id(argv[0]);
    if (volId.whole_thing != -1)
    {
        bzero((void *) &req, sizeof(evm_MsgVolTakeover_t));
        bzero((void *) &msg, sizeof(msg));

        req.hdr.msgType = EVM_VOL_TAKEOVER;
        req.hdr.clientPrivate = (uint64) getpid();
        req.id = volId;
        req.nim = 1;    // R1 limitation
        req.delay = 0;  // forever

        optind = 0;
        while ((c = getopt(argc, argv, "n:T:")) != -1)
        {
            switch (c)
            {
            case 'n':
                req.nim = atoi(optarg);
                printf("nim = %d\n", req.nim);
                break;

            case 'T':
                req.delay = atoi(optarg);
                printf("max takeover time = %d\n", req.delay);
                break;

            default:
                printf("Invalid options!\n");
                return(CMD_ERR_FAILED);
            } // switch
        } // while

        msg.len = sendAgileMsg(-1, &req, sizeof(evm_MsgVolTakeover_t),
                               "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                               &cfgd_to, msg.buf, sizeof(msg.buf));

        if (bad_stat(msg.len) || bad_resp(msg.len, msg.buf))
        {
            printf("Takeover failed.\n");
            return(CMD_ERR_FAILED);
        }
    }
    else
    {
        return(CMD_ERR_FAILED);
    }
    
    printf("Done\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : evm_stats
 *
 * Description  : Show EVM server statistics.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 28-JUL-2002
 *
 *------------------------------------------------------------------------------
 */

DEFUN(evm_stats,
      evm_stats_cmd,
      "evm show statistic [-r]",
      "\nShow EVM server statistics.\n")

{
    evm_MsgStatReq_t    statReq;
    evm_MsgStatRsp_t    statRsp;
    int                 msgLen;

    statReq.hdr.msgType = EVM_STAT_REQ;
    statReq.hdr.clientPrivate = getpid();
    if (argc == 0)
    {
        statReq.reset = FALSE;
    }
    else  if ((argc == 1) && (!strcmp(argv[0], "-r")))
    {
        statReq.reset = TRUE;
        printf("Reset statistics after reading them.\n");            
    }
    else
    {
        printf("Invalid options!\n");
        return(CMD_ERR_FAILED);
    }
    
    msgLen = sendAgileMsg(-1, &statReq, sizeof(evm_MsgStatReq_t),
                          "evm", NFX_NFP_FP2, 1, MGMTPLANE,
                          &cfgd_to, &statRsp, sizeof(evm_MsgStatRsp_t));
    if (bad_stat(msgLen))
    {
        printf("Get EVM server statistics.\n");
        return(CMD_ERR_FAILED);
    }

    displayStats(&(statRsp.stats));
    
    printf("Done.\n");
    return(CMD_SUCCESS);
}

/*------------------------------------------------------------------------------
 * Name         : install_lvol
 *
 * Description  : 
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 05-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
install_lvol(priv_t         myPriv,
             enum node_type node)
    
{
    switch (myPriv)
    {
    case diagPriv:
        install_element(node, &lvol_show_cmd);
        install_element(node, &lvol_create_cmd);
        install_element(node, &lvol_delete_cmd);
        install_element(node, &lvol_modify_cmd);
        install_element(node, &lvol_grow_cmd);
        install_element(node, &lvol_online_cmd);
        install_element(node, &lvol_offline_cmd);
        install_element(node, &lvol_takeover_cmd);
        install_element(node, &evm_stats_cmd);
    case fullPriv:
    case changePriv:
    case addPriv:
    case viewPriv:
    } // switch
    
    return;
}

/*----------------------------------------------------------------------------*/
