/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: cmd_spool
 *
 * RCS:  $Id: cmd_spool.c,v 1.32 2002/11/07 00:11:33 nitinj Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include "exports/spm-api.h"
#include "exports/ea-api.h"

/*
 * Constants
 */

#define SPOOL_CMD_STR   "Lun show command.\n"
#define MAX_CLI_DISPLAY_LINE 14
#define SPM_MAX_RSP_MSG_SIZE 8192
#define BYTES_IN_MB 1048576
#define LUN_LIST_RECS 14
// #define SIMULATE_EA_AV_SM_INTERFACE
#define PRINT_SPM_LUN_CLUSTER_INFO

static struct timeval cfgd_to = {60, 0};

char *raidString[] =    
{   
    "Raid-0", // EVM_RAID0 = 0,      // stripping
    "Raid-1", // EVM_RAID1 = 1,      // mirror
    "Raid-01", // EVM_RAID01 = 2,     // stripe + mirror
    "Raid-3", // EVM_RAID3 = 3,      // raid 3
    "",
    "Raid-5", // EVM_RAID5 = 5,      // raid 5
    "Plain", // EVM_RAID_NONE = 6,  // no raid level, ie plain
    "Any", // EVM_RAID_ANY = 0xffff
    "None"
};

char *spmString[] = 
{
    "",
    "Free", // SPM_LUN_FREE    1
    "Used", // SPM_LUN_USED    2
    "Foreign", // SPM_LUN_FOREIGN 3
    "Out-of-Cluster" // SPM_LUN_OC      4
};

char *spmDevTypeSting[] =
{
    "Disk",
    "Tape",
    "Changer",
    "Stacker",
    "Misc"
};


/*
 * Local function declaration
 */

static
int bad_stat (int stat);

// static
// int bad_resp (int stat, char *resp);

/*
 * Local function
 */

static int
bad_stat(int stat)

{
    if (stat < 0)
    {
        perror("Send to SPM failed.");
        return 1;
    }
    else if (!stat)
    {
        printf("Timed out waiting for response.\n");
        return 1;
    }
    return 0;
}


static void
display_pool_list(spm_MsgPoolListRsp_t *poolRsp)
{
    int32    cnt;
    int32    lineCnt = 2, rdIdx;
    char     c;
    char     ctrl[SDM_DEV_NAME_LEN+1];
    uint64 uSize, fSize, oSize, nSize;
    evm_raid_t rL;
    int32    devIdx = 0;

    if (poolRsp->numEntries <= 0)
        return;

    printf("Controller                            Type      Raid    Used (MB)         ");
    printf("Free (MB)         Out Cluster (MB)  Foreign (MB)\n");
    printf("-----------------------------------------------------------------");
    printf("---------------------------------------------------------\n");
    for(cnt = 0; cnt < poolRsp->numEntries; cnt++, lineCnt++)
    {
       bzero(ctrl, SDM_DEV_NAME_LEN+1);
       strcpy(ctrl, poolRsp->poolList[cnt].ctrl);
       strcat(ctrl, "_*");
       uSize = (uint64)poolRsp->poolList[cnt].usedSize*(uint64)EVM_LBLK_SIZE;
       fSize = (uint64)poolRsp->poolList[cnt].freeSize*(uint64)EVM_LBLK_SIZE;
       oSize = (uint64)poolRsp->poolList[cnt].ocSize*(uint64)EVM_LBLK_SIZE;
       nSize = (uint64)poolRsp->poolList[cnt].foreignSize*(uint64)EVM_LBLK_SIZE;
       rL = poolRsp->poolList[cnt].rL;
       rdIdx = 8;
       if (rL == EVM_RAID_ANY) rdIdx = 7;
       else if (rL >= EVM_RAID0 && rL <= EVM_RAID_NONE) rdIdx = (int32)rL;
       devIdx = poolRsp->poolList[cnt].type;
       if (poolRsp->poolList[cnt].type > SPM_DEV_MISC) devIdx = SPM_DEV_MISC;
       if (devIdx == SPM_DEV_DISK)
       {
            printf("%-36s  %-8s  %-6s  %-16.2f  %-16.2f  %-16.2f  %-.2f\n",
                           ctrl, spmDevTypeSting[devIdx], raidString[rdIdx],
                           (double)uSize/(double)BYTES_IN_MB,
                           (double)fSize/(double)BYTES_IN_MB,
                           (double)oSize/(double)BYTES_IN_MB,
                           (double)nSize/(double)BYTES_IN_MB);
       }
       else
       {
            printf("%-36s  %-8s  %-6s  %-16s  %-16s  %-16s  %-s\n",
                           ctrl, spmDevTypeSting[devIdx], "-",
                           "-", "-", "-", "-");
       }
       if (lineCnt == MAX_CLI_DISPLAY_LINE)
       {
            printf("---more---[q:Quit][Enter] ");
            c = getchar();
            if (c == 'q' || c == 'Q')
            {
                getchar();
                return;
            }
            else if (cnt < (poolRsp->numEntries - 1))
                lineCnt = 2;
       }
    }
    return;
}


static uint64
display_lun_list_disk(spm_MsgLunListRsp_t *lunRsp, int head)
{
    int32    cnt;
    uint64   size, lastLun;
    int32    rdIdx;
    char     st[20], *lclst;
    evm_raid_t rL;

    if (lunRsp->numLuns <= 0)
        return -1;

    if (head)
    {
        printf("Device Name                                                             Model             Raid    Size (MB)    State       Cluster\n");
        printf("---------------------------------------------------------------------------------------------------------------------------------------\n");
    }

    for (cnt = 0; cnt < lunRsp->numLuns; cnt++)
    {
        bzero(st, 20);
        if (lunRsp->lunList[cnt].state == EVM_FREE_LUN)
            strcpy(st, "free");
        else if (lunRsp->lunList[cnt].state == EVM_USED_LUN)
            strcpy(st, "used");
        else if (lunRsp->lunList[cnt].state == EVM_FOREIGN_LUN)
            strcpy(st, "foreign");
        else
            strcpy(st, "outCluster");
        lclst = lunRsp->lunList[cnt].lunCluster;
        size = (uint64)lunRsp->lunList[cnt].size * (uint64)EVM_LBLK_SIZE;
        rL = lunRsp->lunList[cnt].rL;
        rdIdx = 8;
        if (rL == EVM_RAID_ANY) rdIdx = 7;
        else if (rL >= EVM_RAID0 && rL <= EVM_RAID_NONE) rdIdx = (int32)rL;

        printf("%-70s  %-16s  %-6s  %-11.2f  %-10s  %-s\n",
                                              lunRsp->lunList[cnt].dev,
                                              lunRsp->lunList[cnt].model,
                                              raidString[rdIdx],
                                              (double)size/(double)BYTES_IN_MB,
                                              st,
                                              (strlen(lclst))?lclst:"n/a");
        lastLun = lunRsp->lunList[cnt].sid;
    }
    return lastLun;
}

static uint64
display_lun_list_tape(spm_MsgLunListRsp_t *lunRsp, int head)
{
    int32    cnt;
    uint64   lastLun;
    uint32   devIdx;

    if (lunRsp->numLuns <= 0)
        return -1;

    if (head)
    {
        printf("Device Name                                                             Model             Type\n");
        printf("--------------------------------------------------------------------------------------------------\n");
    }

    for (cnt = 0; cnt < lunRsp->numLuns; cnt++)
    {
        devIdx = lunRsp->lunList[cnt].type;
        if (lunRsp->lunList[cnt].type > SPM_DEV_MISC) devIdx = SPM_DEV_MISC;
        printf("%-70s  %-16s  %-s\n", lunRsp->lunList[cnt].dev, lunRsp->lunList[cnt].model, spmDevTypeSting[devIdx]);
        lastLun = lunRsp->lunList[cnt].sid;
    }
    return lastLun;
}

static uint64
display_lun_list_all(spm_MsgLunListRsp_t *lunRsp, int head)
{
    int32    cnt;
    uint64   size, lastLun;
    int32    rdIdx;
    char     st[20], *lclst;
    evm_raid_t rL;
    uint32   devIdx;

    if (lunRsp->numLuns <= 0)
        return -1;

    if (head)
    {
        printf("Device Name                                                             Type      Raid    Size (MB)    State       Cluster\n");
        printf("-------------------------------------------------------------------------------------------------------------------------------\n");
    }

    for (cnt = 0; cnt < lunRsp->numLuns; cnt++)
    {
        bzero(st, 20);
        if (lunRsp->lunList[cnt].state == EVM_FREE_LUN)
            strcpy(st, "free");
        else if (lunRsp->lunList[cnt].state == EVM_USED_LUN)
            strcpy(st, "used");
        else if (lunRsp->lunList[cnt].state == EVM_FOREIGN_LUN)
            strcpy(st, "foreign");
        else
            strcpy(st, "outCluster");
        lclst = lunRsp->lunList[cnt].lunCluster;
        size = (uint64)lunRsp->lunList[cnt].size * (uint64)EVM_LBLK_SIZE;
        rL = lunRsp->lunList[cnt].rL;
        rdIdx = 8;
        if (rL == EVM_RAID_ANY) rdIdx = 7;
        else if (rL >= EVM_RAID0 && rL <= EVM_RAID_NONE) rdIdx = (int32)rL;

        devIdx = lunRsp->lunList[cnt].type;
        if (lunRsp->lunList[cnt].type > SPM_DEV_MISC) devIdx = SPM_DEV_MISC;

        if (lunRsp->lunList[cnt].type == SPM_DEV_DISK)
        {
            printf("%-70s  %-8s  %-6s  %-11.2f  %-10s  %-s\n",
                                     lunRsp->lunList[cnt].dev,
                                     spmDevTypeSting[devIdx], raidString[rdIdx],
                                     (double)size/(double)BYTES_IN_MB,
                                     st,
                                     (strlen(lclst))?lclst:"n/a");
        }
        else
        {
            printf("%-70s  %-8s  %-6s  %-11s  %-10s  %-s\n",
                                     lunRsp->lunList[cnt].dev,
                                     spmDevTypeSting[devIdx],
                                     "-", "-", "-", "-");
        }
        lastLun = lunRsp->lunList[cnt].sid;
    }
    return lastLun;
}

static int rescan_lun_req(char *name, char *node)
{
    spm_MsgLunRescanReq_t req;
    spm_MsgLunRescanRsp_t rsp;
    int32                 stat, rc;

    bzero(&req, sizeof(req));
    bzero(&rsp, sizeof(rsp));
    req.hdr.msgType = SPM_RESCAN_LUN_REQ;
    req.hdr.clientPrivate = getpid();
    strcpy(req.hdr.node, "PCC");

    if (name)
    {
        req.devCnt = 1;
        strcpy(req.dev[0], name);
    }
    if (node)
        strcpy(req.clNode, node);

    stat = sendAgileMsg(-1, &req, sizeof(spm_MsgLunRescanReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, &rsp, sizeof(spm_MsgLunRescanRsp_t));
    if (!bad_stat(stat) && (stat == sizeof(spm_MsgLunRescanRsp_t)) &&
        rsp.result == NFX_OK)
        rc = 1;
    else
        rc = 0;

    return rc;
}

/*
 *------------------------------------------------------------------------------
 * Name         : lun_list
 *
 * Description  : Display the list of LUN's that are visible to the cluster.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 08-FEB-2002
 *                27-FEB-2002
 *------------------------------------------------------------------------------
 */

DEFUN(lun_show,
      lun_show_cmd,
      "lun show",
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      SPOOL_CMD_STR)
{
    spm_MsgTotalPoolReq_t poolNumReq;
    spm_MsgPoolListReq_t  poolReq;
    char                  resp[SPM_MAX_RSP_MSG_SIZE], *errStr;
    spm_MsgTotalPoolRsp_t *poolNumRsp;
    spm_MsgPoolListRsp_t  *poolRsp;
    int32                 stat;
    int32                 numPools = 0, rspSize;

    bzero (&poolNumReq, sizeof(spm_MsgTotalPoolReq_t));
    poolNumReq.hdr.msgType = SPM_GET_TOTAL_POOLS_REQ;
    poolNumReq.hdr.clientPrivate = getpid();
    strcpy(poolNumReq.hdr.node,""); // myNfxId;

    stat = sendAgileMsg(-1, &poolNumReq, sizeof(spm_MsgTotalPoolReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, resp, SPM_MAX_RSP_MSG_SIZE);

    poolNumRsp = (spm_MsgTotalPoolRsp_t *)resp;
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgTotalPoolRsp_t)) &&
        (poolNumRsp->hdr.msgType == SPM_GET_TOTAL_POOLS_RSP))
        numPools = poolNumRsp->num;
    else
    {
        printf("Failed to get number of pools\n");
        if (stat > sizeof(spm_MsgStat_t))
        {
            errStr = (resp + sizeof(spm_MsgStat_t));
            if (errStr) printf ("%s\n", errStr);
        }
        return CMD_SUCCESS;
    }

    bzero (&poolReq, sizeof(spm_MsgPoolListReq_t));
    poolReq.hdr.msgType = SPM_GET_POOL_LIST_REQ;
    poolReq.hdr.clientPrivate = getpid();
    strcpy(poolReq.hdr.node,""); // myNfxId;

    rspSize = sizeof(spm_MsgPoolListRsp_t)+
              (sizeof(spm_poolNode_t)*(numPools-1));
    poolRsp = (spm_MsgPoolListRsp_t *)malloc(rspSize);
    NFX_ASSERT(poolRsp != NULL);
    bzero(poolRsp, rspSize);

    stat = sendAgileMsg(-1, &poolReq, sizeof(spm_MsgPoolListReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, poolRsp, rspSize);

    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgPoolListReq_t)) &&
        (poolRsp->hdr.msgType == SPM_GET_POOL_LIST_RSP))
    {
        printf ("\nStorage availability report.\n");
        printf ("============================\n\n");
        display_pool_list(poolRsp);
    }
    else
        printf("Failed to get list of vol ids\n");
    free(poolRsp);
    printf("Done.\n");
    return CMD_SUCCESS;
}

/*
 *------------------------------------------------------------------------------
 * Name         : lun_list_disk
 *
 * Description  : Display the list of LUN's that are visible to the cluster.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *------------------------------------------------------------------------------
 */
#define USG_D_STR_1  "Usage: lun show disk [-c CONTROLLER] [-r RAID_LEVEL] [-t STATE] [-n NODE]\n"
#define USG_D_STR_2  "       -c CONTROLLER : Controller string or wildcard (*)\n"
#define USG_D_STR_3  "       -r RAID_LEVEL : Raid level (Raid-0, Raid-1, Raid-01, Raid-3, Raid-5, Plain, Any)\n"
#define USG_D_STR_4  "       -t STATE : Device state (all, used, free, foreign, outCluster)\n"
#define USG_D_STR_5  "       -n NODE : cluster node name\n"

#define D_OPT "[-c CONTROLLER] [-r RAID_LEVEL] [-t STATE] [-n NODE]\n"

DEFUN(lun_show_disk,
      lun_show_disk_cmd,
      "lun show disk [OPT] [OPT] [OPT] [OPT] [OPT] [OPT] [OPT] [OPT]",
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      "Disk devices\n"
      D_OPT      // "Controller string or wildcard(*)\n"
      D_OPT      // "Controller string or wildcard(*)\n"
      D_OPT      // "Raid level(Raid-0, Raid-1, Raid-01, Raid-3, Raid-5, Any)\n"
      D_OPT      // "Raid level(Raid-0, Raid-1, Raid-01, Raid-3, Raid-5, Any)\n"
      D_OPT      // "Device state (all, used, free, foreign, outCluster)\n"
      D_OPT      // "Device state (all, used, free, foreign, outCluster)\n"
      D_OPT      // "Node name\n"
      D_OPT)     // "Node name\n"
{
    spm_MsgLunListReq_t   lunReq;
    spm_MsgTotalLunReq_t  numReq;
    char                  lresp[SPM_MAX_RSP_MSG_SIZE], *errStr;
    spm_MsgLunListRsp_t   *lunRsp;
    spm_MsgTotalLunRsp_t  *numRsp;
    int32                 stat;
    int32                 lCnt = 0, currCnt = 0;
    int                   c;
    extern int            optind;
    uint64 getNext = 0;
    int32 next = 0, i, done, head = 1, n;
    char  ctrl[SDM_DEV_NAME_LEN+1];
    char  node[SPM_MAX_NODE_NAME_LEN+1];

    bzero (&lunReq, sizeof(spm_MsgLunListReq_t));
    bzero (&numReq, sizeof(spm_MsgTotalLunReq_t));

    lunReq.hdr.msgType = SPM_GET_LUN_LIST_REQ;
    numReq.hdr.msgType = SPM_GET_TOTAL_LUNS_REQ;
    lunReq.hdr.clientPrivate = getpid();
    numReq.hdr.clientPrivate = getpid();
    strcpy(lunReq.hdr.node,"local"); // myNfxId;
    strcpy(numReq.hdr.node,"local"); // myNfxId;
    bzero(ctrl, SDM_DEV_NAME_LEN+1); // default
    bzero(node, SPM_MAX_NODE_NAME_LEN+1); // default
    lunReq.rL = EVM_RAID_ANY; // default
    numReq.rL = EVM_RAID_ANY; // default
    lunReq.lunType = SPM_LUN_ALL_LIST;
    numReq.lunType = SPM_LUN_ALL_LIST;
 
    if (argc > 0)
    {
        optind = 0;
        while ((c = getopt (argc+1, argv-1, "c:r:t:n:")) != -1)
        {
            switch (c)
            {
                case 'c':
                    n = strlen(optarg);
                    for (i = 0, done = 0; i < n && done < 2; i++)
                    {
                        if (optarg[i] == '*')
                            break;
                        if (optarg[i] == '_')
                            done++;
                        if (done != 2)
                            ctrl[i] = optarg[i];
                    }
                    strcpy(lunReq.ctrl, ctrl);
                    strcpy(numReq.ctrl, ctrl);
                    break;
                case 'r':
                    if (strcasecmp(optarg, "Raid-0") == 0)
                        lunReq.rL = EVM_RAID0;
                    else if (strcasecmp(optarg, "Raid-1") == 0)
                        lunReq.rL = EVM_RAID1;
                    else if (strcasecmp(optarg, "Raid-01") == 0)
                        lunReq.rL = EVM_RAID01;
                    else if (strcasecmp(optarg, "Raid-3") == 0)
                        lunReq.rL = EVM_RAID3;
                    else if (strcasecmp(optarg, "Raid-5") == 0)
                        lunReq.rL = EVM_RAID5;
                    else if (strcasecmp(optarg, "Plain") == 0)
                        lunReq.rL = EVM_RAID_NONE;
                    else if (strcasecmp(optarg, "Any") == 0)
                        lunReq.rL = EVM_RAID_ANY;
                    else
                    {
                         printf("%s%s%s", USG_D_STR_1,USG_D_STR_2,USG_D_STR_3);
                         printf ("%s%s", USG_D_STR_4, USG_D_STR_5);
                         return CMD_SUCCESS;
                    }
                    numReq.rL = lunReq.rL;
                    break;
                case 't':
                    if (strcasecmp(optarg, "used") == 0)
                        lunReq.lunType = SPM_LUN_USED_LIST;
                    else if (strcasecmp(optarg, "free") == 0)
                        lunReq.lunType = SPM_LUN_FREE_LIST;
                    else if (strcasecmp(optarg, "outCluster") == 0)
                        lunReq.lunType = SPM_LUN_OUTCLUSTER_LIST;
                    else if (strcasecmp(optarg, "foreign") == 0)
                        lunReq.lunType = SPM_LUN_FOREIGN_LIST;
                    else if (strcasecmp(optarg, "all") == 0)
                        lunReq.lunType = SPM_LUN_ALL_LIST;
                    else
                    {
                        printf("%s%s%s", USG_D_STR_1, USG_D_STR_2, USG_D_STR_3);
                        printf ("%s%s", USG_D_STR_4, USG_D_STR_5);
                        return CMD_SUCCESS;
                    }
                    numReq.lunType = lunReq.lunType;
                    break;
                case 'n':
                    strcpy(node, optarg);
                    strcpy(lunReq.tNode, node);
                    break;
                default:
                    printf ("%s%s%s", USG_D_STR_1, USG_D_STR_2, USG_D_STR_3);
                    printf ("%s%s", USG_D_STR_4, USG_D_STR_5);
                    return CMD_SUCCESS;
            }
        }

        if (optind != argc+1)
        {
            printf ("%s%s%s", USG_D_STR_1, USG_D_STR_2, USG_D_STR_3);
            printf ("%s%s", USG_D_STR_4, USG_D_STR_5);
            return CMD_SUCCESS;
        }
    }

    /*
     * Get number of LUN's
     */

    stat = sendAgileMsg(-1, &numReq, sizeof(spm_MsgTotalLunReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);

    numRsp = (spm_MsgTotalLunRsp_t *)lresp;
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgTotalLunRsp_t)) &&
        (numRsp->hdr.msgType == SPM_GET_TOTAL_LUNS_RSP))
        lCnt = numRsp->numLun;
    else
    {
        printf("Failed to get response from SPM\n");
        if (stat > sizeof(spm_MsgStat_t))
        {
            errStr = (lresp + sizeof(spm_MsgStat_t));
            if (errStr) printf ("%s\n", errStr);
        }
        return CMD_SUCCESS;
    }
    bzero(lresp, SPM_MAX_RSP_MSG_SIZE);

    lunReq.lastSID = 0;
    lunReq.numReq  = LUN_LIST_RECS;
    n = strlen(node);
    printf ("\nDisk devices known to %-s.\n", (n)?node:"PCC");
    printf ("=======================");
    n = (n)?n:3;
    for (i = 0; i < n; i++) printf ("=");
    printf("\n\n");
    while(next != -1 && lCnt > 0)
    {
        stat = sendAgileMsg(-1, &lunReq, sizeof(spm_MsgLunListReq_t),
                            "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);
        lunRsp = (spm_MsgLunListRsp_t *)lresp;
        if (!bad_stat(stat) &&
            (stat >= sizeof(spm_MsgLunListRsp_t)) &&
            (lunRsp->hdr.msgType == SPM_GET_LUN_LIST_RSP))
        {
            if (lunRsp->numLuns > 0)
            {
                if (currCnt >= LUN_LIST_RECS)
                {
                    printf("---more---[q:Quit][Enter] ");
                    c = getchar();
                    if (c == 'q' || c == 'Q')
                    {
                        getchar();
                        return CMD_SUCCESS;
                    }
                }
                getNext = display_lun_list_disk(lunRsp, head);
                if (getNext == -1) next = -1;
                head = 0;
            }
            else
                next = -1;
        }
        else
        {
            printf("Failed to get list of devices\n");
            next = -1;
        }
        currCnt += lunRsp->numLuns;
        if (currCnt >= lCnt)
            next = -1;
        lunReq.lastSID = getNext;
        bzero(lresp, SPM_MAX_RSP_MSG_SIZE);
    }
    printf("Done.\n");
    return CMD_SUCCESS;
}

/*
 *------------------------------------------------------------------------------
 * Name         : lun_list_tape
 *
 * Description  : Display the list of LUN's that are visible to the cluster.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *------------------------------------------------------------------------------
 */

#define USG_T_STR_1  "Usage: lun show tape [-c CONTROLLER] [-n NODE]\n"
#define USG_T_STR_2  "       -c CONTROLLER : Controller string or wildcard (*)\n"
#define USG_T_STR_3  "       -n NODE : cluster node name\n"

#define T_OPT "[-c CONTROLLER] [-n NODE]\n"

DEFUN(lun_show_tape,
      lun_show_tape_cmd,
      "lun show tape [OPT] [OPT] [OPT] [OPT]",
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      "Tape devices\n"
      T_OPT    // "Controller string or wildcard(*)\n"
      T_OPT    // "Controller string or wildcard(*)\n"
      T_OPT    // "Node name\n"
      T_OPT)   // "Node name\n"
{
    spm_MsgLunListReq_t   lunReq;
    spm_MsgTotalLunReq_t  numReq;
    char                  lresp[SPM_MAX_RSP_MSG_SIZE], *errStr;
    spm_MsgLunListRsp_t   *lunRsp;
    spm_MsgTotalLunRsp_t  *numRsp;
    int32                 stat;
    int32                 lCnt = 0, currCnt = 0;
    int                   c;
    extern int            optind;
    uint64 getNext = 0;
    int32 next = 0, i, done, head = 1, n;
    char  ctrl[SDM_DEV_NAME_LEN+1];
    char  node[SPM_MAX_NODE_NAME_LEN+1];

    bzero (&lunReq, sizeof(spm_MsgLunListReq_t));
    bzero (&numReq, sizeof(spm_MsgTotalLunReq_t));

    lunReq.hdr.msgType = SPM_GET_LUN_LIST_REQ;
    numReq.hdr.msgType = SPM_GET_TOTAL_LUNS_REQ;
    lunReq.hdr.clientPrivate = getpid();
    numReq.hdr.clientPrivate = getpid();
    strcpy(lunReq.hdr.node,"local"); // myNfxId;
    strcpy(numReq.hdr.node,"local"); // myNfxId;
    lunReq.diskOtherAll = 2;
    numReq.diskOtherAll = 2;
    bzero(ctrl, SDM_DEV_NAME_LEN+1);
    bzero(node, SPM_MAX_NODE_NAME_LEN+1);
    lunReq.rL = EVM_RAID_ANY; // default
    numReq.rL = EVM_RAID_ANY; // default
    lunReq.lunType = SPM_LUN_ALL_LIST;
    numReq.lunType = SPM_LUN_ALL_LIST;
 
    if (argc > 0)
    {
        optind = 0;
        while ((c = getopt (argc+1, argv-1, "c:n:")) != -1)
        {
            switch (c)
            {
                case 'c':
                    n = strlen(optarg);
                    for (i = 0, done = 0; i < n && done < 2; i++)
                    {
                        if (optarg[i] == '*')
                            break;
                        if (optarg[i] == '_')
                            done++;
                        if (done != 2)
                            ctrl[i] = optarg[i];
                    }
                    strcpy(lunReq.ctrl, ctrl);
                    strcpy(numReq.ctrl, ctrl);
                    break;
                case 'n':
                    strcpy(node, optarg);
                    strcpy(lunReq.tNode, node);
                    break;
                default:
                    printf ("%s%s%s", USG_T_STR_1, USG_T_STR_2, USG_T_STR_3);
                    return CMD_SUCCESS;
            }
        }

        if (optind != argc+1)
        {
            printf ("%s%s%s", USG_T_STR_1, USG_T_STR_2, USG_T_STR_3);
            return CMD_SUCCESS;
        }
    }

    /*
     * Get number of LUN's
     */

    stat = sendAgileMsg(-1, &numReq, sizeof(spm_MsgTotalLunReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);

    numRsp = (spm_MsgTotalLunRsp_t *)lresp;
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgTotalLunRsp_t)) &&
        (numRsp->hdr.msgType == SPM_GET_TOTAL_LUNS_RSP))
        lCnt = numRsp->numLun;
    else
    {
        printf("Failed to get response from SPM\n");
        if (stat > sizeof(spm_MsgStat_t))
        {
            errStr = (lresp + sizeof(spm_MsgStat_t));
            if (errStr) printf ("%s\n", errStr);
        }
        return CMD_SUCCESS;
    }
    bzero(lresp, SPM_MAX_RSP_MSG_SIZE);

    lunReq.lastSID = 0;
    lunReq.numReq   = LUN_LIST_RECS;
    n = strlen(node);
    printf ("\nTape devices known to %-s.\n", (n)?node:"PCC");
    printf ("=======================");
    n = (n)?n:3;
    for (i = 0; i < n; i++) printf ("=");
    printf("\n\n");
    while(next != -1 && lCnt > 0)
    {
        stat = sendAgileMsg(-1, &lunReq, sizeof(spm_MsgLunListReq_t),
                            "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);
        lunRsp = (spm_MsgLunListRsp_t *)lresp;
        if (!bad_stat(stat) &&
            (stat >= sizeof(spm_MsgLunListRsp_t)) &&
            (lunRsp->hdr.msgType == SPM_GET_LUN_LIST_RSP))
        {
            if (lunRsp->numLuns > 0)
            {
                if (currCnt >= LUN_LIST_RECS)
                {
                    printf("---more---[q:Quit][Enter] ");
                    c = getchar();
                    if (c == 'q' || c == 'Q')
                    {
                        getchar();
                        return CMD_SUCCESS;
                    }
                }
                getNext = display_lun_list_tape(lunRsp, head);
                if (getNext == -1) next = -1;
                head = 0;
            }
            else
                next = -1;
        }
        else
        {
            printf("Failed to get list of devices\n");
            next = -1;
        }
        currCnt += lunRsp->numLuns;
        if (currCnt >= lCnt)
            next = -1;
        lunReq.lastSID = getNext;
        bzero(lresp, SPM_MAX_RSP_MSG_SIZE);
    }
    printf("Done.\n");
    return CMD_SUCCESS;
}

/*
 *------------------------------------------------------------------------------
 * Name         : lun_list_all
 *
 * Description  : Display the list of LUN's that are visible to the cluster.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *------------------------------------------------------------------------------
 */

#define USG_A_STR_1  "Usage: lun show all [-n NODE]\n"
#define USG_A_STR_2  "       -n NODE : cluster node name\n"

#define OPT_NODE "[-n NODE]\n"

DEFUN(lun_show_all,
      lun_show_all_cmd,
      "lun show all [OPT] [OPT]",
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      "All devices\n"
      OPT_NODE   // "Node name\n"
      OPT_NODE)  // "Node name\n"
{
    spm_MsgLunListReq_t   lunReq;
    spm_MsgTotalLunReq_t  numReq;
    char                  lresp[SPM_MAX_RSP_MSG_SIZE], *errStr;
    spm_MsgLunListRsp_t   *lunRsp;
    spm_MsgTotalLunRsp_t  *numRsp;
    int32                 stat;
    int32                 lCnt = 0, currCnt = 0;
    int                   c;
    extern int            optind;
    uint64 getNext = 0;
    int32 next = 0, head = 1, n, i;
    char  node[SPM_MAX_NODE_NAME_LEN+1];

    bzero (&lunReq, sizeof(spm_MsgLunListReq_t));
    bzero (&numReq, sizeof(spm_MsgTotalLunReq_t));
    bzero(node, SPM_MAX_NODE_NAME_LEN+1);

    lunReq.hdr.msgType = SPM_GET_LUN_LIST_REQ;
    numReq.hdr.msgType = SPM_GET_TOTAL_LUNS_REQ;
    lunReq.hdr.clientPrivate = getpid();
    numReq.hdr.clientPrivate = getpid();
    strcpy(lunReq.hdr.node,"local"); // myNfxId;
    strcpy(numReq.hdr.node,"local"); // myNfxId;
    lunReq.diskOtherAll = 1;
    numReq.diskOtherAll = 1;
    lunReq.rL = EVM_RAID_ANY; // default
    numReq.rL = EVM_RAID_ANY; // default
    lunReq.lunType = SPM_LUN_ALL_LIST;
    numReq.lunType = SPM_LUN_ALL_LIST;

    if (argc > 0)
    {
        optind = 0;
        while ((c = getopt (argc+1, argv-1, "n:")) != -1)
        {
            switch (c)
            {
                case 'n':
                    strcpy(node, optarg);
                    strcpy(lunReq.tNode, node);
                    break;
                default:
                    printf ("%s%s", USG_A_STR_1, USG_A_STR_2);
                    return CMD_SUCCESS;
            }
        }

        if (optind != argc+1)
        {
            printf ("%s%s", USG_A_STR_1, USG_A_STR_2);
            return CMD_SUCCESS;
        }
    }

    /*
     * Get number of LUN's
     */

    stat = sendAgileMsg(-1, &numReq, sizeof(spm_MsgTotalLunReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);

    numRsp = (spm_MsgTotalLunRsp_t *)lresp;
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgTotalLunRsp_t)) &&
        (numRsp->hdr.msgType == SPM_GET_TOTAL_LUNS_RSP))
        lCnt = numRsp->numLun;
    else
    {
        printf("Failed to get response from SPM\n");
        if (stat > sizeof(spm_MsgStat_t))
        {
            errStr = (lresp + sizeof(spm_MsgStat_t));
            if (errStr) printf ("%s\n", errStr);
        }
        return CMD_SUCCESS;
    }
    bzero(lresp, SPM_MAX_RSP_MSG_SIZE);

    lunReq.lastSID = 0;
    lunReq.numReq = LUN_LIST_RECS;
    n = strlen(node);
    printf ("\nAll devices known to %-s.\n", (n)?node:"PCC");
    printf ("======================");
    n = (n)?n:3;
    for (i = 0; i < n; i++) printf ("=");
    printf("\n\n");
    while(next != -1 && lCnt > 0)
    {
        stat = sendAgileMsg(-1, &lunReq, sizeof(spm_MsgLunListReq_t),
                            "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);
        lunRsp = (spm_MsgLunListRsp_t *)lresp;
        if (!bad_stat(stat) &&
            (stat >= sizeof(spm_MsgLunListRsp_t)) &&
            (lunRsp->hdr.msgType == SPM_GET_LUN_LIST_RSP))
        {
            if (lunRsp->numLuns > 0)
            {
                if (currCnt >= LUN_LIST_RECS)
                {
                    printf("---more---[q:Quit][Enter] ");
                    c = getchar();
                    if (c == 'q' || c == 'Q')
                    {
                        getchar();
                        return CMD_SUCCESS;
                    }
                }
                getNext = display_lun_list_all(lunRsp, head);
                if (getNext == -1) next = -1;
                head = 0;
            }
            else
                next = -1;
        }
        else
        {
            printf("Failed to get list of devices\n");
            next = -1;
        }
        currCnt += lunRsp->numLuns;
        if (currCnt >= lCnt)
            next = -1;
        lunReq.lastSID = getNext;
        bzero(lresp, SPM_MAX_RSP_MSG_SIZE);
    }
    printf("Done.\n");
    return CMD_SUCCESS;
}

/*
 *------------------------------------------------------------------------------
 * Name         : lun_label
 *
 * Description  : Convert a foreign or out-cluster lun to an Agile LUN.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-FEB-2002
 *
 *------------------------------------------------------------------------------
 */

#define L_OPT "[-f] [-c CLUSTERNAME]\n"

DEFUN(lun_label,
      lun_label_cmd,
      "lun label DEVICE_NAME [OPT] [OPT] [OPT]",
      SPOOL_CMD_STR
      "Put agile label on LUN.\n"
      "- mark a specific DEVICE_NAME\n"
      L_OPT    // "Force label command\n"
      L_OPT    // "Cluster Name\n"
      L_OPT)   // "Cluster Name\n"
{
    spm_MsgMarkAgileLunReq_t   req;
    spm_MsgMarkAgileLunRsp_t   rsp;
    int                        stat, force = 0;
    int                        ans, c, cnt;
    pid_t                      twid;

    bzero(&req, sizeof(spm_MsgMarkAgileLunReq_t));
    req.hdr.msgType = SPM_MARK_LUN_AS_AGILE_REQ;
    req.hdr.clientPrivate = getpid();

    if (argc <= 4)
    {
        if (strlen(argv[0]))
        {
                req.devCnt = 1;
                strcpy(req.dev[0], argv[0]);
        }
        else
        {
            printf ("Usage :\n\tlun label DEVICE_NAME [-f] [-c CLUSTERNAME]\n");
            return CMD_SUCCESS;
        }

        if (argc >= 2)
        {
            cnt = 1;
            while (cnt < argc)
            {
                if (!strcasecmp(argv[cnt], "-c"))
                {
                    if (strlen(argv[++cnt]))
                    {
                        strcpy(req.cluster, argv[cnt]);
                        req.hdr.msgType = SPM_MARK_LUN_AS_OC_REQ;
                    }
                    else
                    {
                        printf ("Usage :\n\tlun label DEVICE_NAME [-f] [-c CLUSTERNAME]\n");
                        return CMD_SUCCESS;
                    }
                }
                if (!strcasecmp(argv[cnt], "-f"))
                {
                    printf ("WARNING: This device may be part of a volume owned on another node.\nThe -f option does not guarantee consistency with all nodes in the network.\n");
                    req.force = 1;
                }
                cnt++;
            }
        }
    }
    else
        return CMD_SUCCESS;

    if (!force)
    {
        printf ("Are you sure ? [Y/N] : ");
        ans = getchar();
        c = ans;
        while (c != '\n')
            c = getchar();
        if (!(ans == 'y' || ans == 'Y'))
        {
            printf ("Request cancled\n");
            return CMD_SUCCESS;
        }
    }

    twid = start_twiddle ();
    stat = sendAgileMsg(-1, &req, sizeof(spm_MsgMarkAgileLunReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, &rsp, sizeof(spm_MsgMarkAgileLunRsp_t));
    stop_twiddle (twid);
    if (!bad_stat(stat) && (stat == sizeof(spm_MsgMarkAgileLunRsp_t)) &&
        (rsp.status == SPM_OK &&
        (rsp.hdr.msgType == SPM_MARK_LUN_AS_AGILE_RSP ||
         rsp.hdr.msgType == SPM_MARK_LUN_AS_OC_RSP)))
        printf("Done\n");
    else
    {
        if (rsp.status == SPM_NOT_FND)
            printf ("Error : %s not found\n", req.dev[0]);
        else if (rsp.status == SPM_NOT_DISK)
            printf ("Error : %s is not a Disk\n", req.dev[0]);
        else if (rsp.status == SPM_NOT_FREE &&
                 req.hdr.msgType == SPM_MARK_LUN_AS_AGILE_REQ)
        {
            printf ("Error : %s is part of a volume\n", req.dev[0]);
        }
        else if (rsp.status == SPM_NOT_FREE &&
                 req.hdr.msgType == SPM_MARK_LUN_AS_OC_REQ)
        {
            printf ("Error : %s cannot be assigned to cluster\n", req.dev[0]);
            printf ("\tOnly FREE device can be assigned to different cluster\n");
        }
        else if (rsp.status == SPM_NOT_FOR)
        {
            printf ("Error : %s is not foreign\n", req.dev[0]);
            printf ("\tOnly foreign device can be labeled\n");
        }
        else
            printf("Error : failed to write agile label\n");
    }

    return CMD_SUCCESS;
}


DEFUN(lun_unlabel,
      lun_unlabel_cmd,
      "lun unlabel DEVICE_NAME",
      SPOOL_CMD_STR
      "Release agile LUN.\n"
      "specify DEVICE_NAME\n")
{
    spm_MsgMarkAgileLunReq_t   *req;
    spm_MsgMarkAgileLunRsp_t   rsp;
    int                        stat;
    int                        ans, c, reqSize, i;
    pid_t                      twid;

    if (argc >= 1)
    {
        reqSize = sizeof(spm_MsgMarkAgileLunReq_t)+(SDM_DEV_NAME_LEN+1)*argc;
        req = (spm_MsgMarkAgileLunReq_t *)malloc(reqSize);
        bzero(req, reqSize);
        req->hdr.msgType = SPM_UNMARK_LUN_AS_AGILE_REQ;
        req->hdr.clientPrivate = getpid();
        req->devCnt = argc;
        for (i = 0; i < argc; i++)
            if (strlen(argv[i]))
                strcpy(req->dev[i], argv[i]);
    }
    else
    {
        printf ("Usage :\n\tlun unlabel DEVICE_NAME\n");
        return CMD_SUCCESS;
    }

    printf ("Are you sure ? [Y/N] : ");
    ans = getchar();
    c = ans;
    while (c != '\n')
        c = getchar();
    if (!(ans == 'y' || ans == 'Y'))
    {
        printf ("Request cancled\n");
        return CMD_SUCCESS;
    }

    twid = start_twiddle ();
    stat = sendAgileMsg(-1, req, reqSize,
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, &rsp, sizeof(spm_MsgMarkAgileLunRsp_t));
    stop_twiddle (twid);
    free(req);
    if (!bad_stat(stat) && (stat == sizeof(spm_MsgMarkAgileLunRsp_t)) &&
        (rsp.hdr.msgType == SPM_UNMARK_LUN_AS_AGILE_RSP && rsp.status == SPM_OK))
        printf("Done\n");
    else
    {
        if (rsp.status == SPM_NOT_FND)
            printf ("Error : %s not found\n", argv[0]);
        else if (rsp.status == SPM_NOT_FREE)
        {
            printf ("Error : Cannot unmark %s\n", argv[0]);
            printf ("        Make sure %s is free\n", argv[0]);
        }
        else if (rsp.status == SPM_NOT_FOR)
            printf("Error : Cannot unmark outCluster device\n");
        else
            printf("Error : failed to unmake device\n");
    }

    return CMD_SUCCESS;
}

DEFUN(lun_mod_cluster,
      lun_mod_cluster_cmd,
      "lun modify cluster DEVICE_NAME CLUSTER_NAME",
      SPOOL_CMD_STR
      "Modify\n"
      "Cluster\n"
      "specify DEVICE_NAME\n"
      "specify CLUSTER_NAME\n")
{
    spm_MsgMarkAgileLunReq_t   *req;
    spm_MsgMarkAgileLunRsp_t   rsp;
    int32                      stat;
    int                        ans, c, reqSize;

    if (argc >= 1)
    {
        reqSize = sizeof(spm_MsgMarkAgileLunReq_t)+(SDM_DEV_NAME_LEN+1)*argc;
        req = (spm_MsgMarkAgileLunReq_t *)malloc(reqSize);
        bzero(req, reqSize);
        req->hdr.msgType = SPM_UPDATE_CLUSTER_NAME_REQ;
        req->hdr.clientPrivate = getpid();
        req->devCnt = 1;
        if (strlen(argv[0]))
                strcpy(req->dev[0], argv[0]);
        if (strlen(argv[1]))
                strcpy(req->cluster, argv[1]);
    }
    else
    {
        printf ("Usage :\n\tlun modify cluster DEVICE_NAME CLUSTER_NAME\n");
        return CMD_SUCCESS;
    }

    printf ("Warning : You may LOOSE the volume associated with device\n");
    printf ("          If you want to continue, make sure you modify all the devices associated with the volume\n");
    printf ("Are you sure ? [Y/N] : ");
    ans = getchar();
    c = ans;
    while (c != '\n')
        c = getchar();
    if (!(ans == 'y' || ans == 'Y'))
    {
        printf ("Request cancled\n");
        return CMD_SUCCESS;
    }

    bzero(&rsp, sizeof(rsp));
    stat = sendAgileMsg(-1, req, reqSize,
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, &rsp, sizeof(spm_MsgMarkAgileLunRsp_t));
    free(req);
    if (!bad_stat(stat) && (stat == sizeof(spm_MsgMarkAgileLunRsp_t)) &&
        (rsp.hdr.msgType == SPM_UPDATE_CLUSTER_NAME_RSP && rsp.status == SPM_OK))
        printf("Done\n");
    else
    {
        if (rsp.status == SPM_NOT_FND)
            printf ("Device not found\n");
        else if (rsp.status == SPM_NOT_DISK)
            printf ("Error : %s is not a Disk\n", argv[0]);
        else if (rsp.status == SPM_NOT_FREE ||
                 rsp.status == SPM_NOT_FOR)
            printf ("Error : Cannot modify %s\n", argv[0]);
        else
            printf("Error : failed to modify cluster\n");
    }

    return CMD_SUCCESS;
}

DEFUN(lun_rescan,
      lun_rescan_cmd,
      "lun rescan (DEVICE_NAME|all) [OPT] [OPT]",
      SPOOL_CMD_STR
      "Re-scan the LUN label\n"
      "specify DEVICE_NAME\n"
      "ALL devices\n"
      OPT_NODE
      OPT_NODE)
{
    char *name = NULL, *node = NULL;
    int  rc;
    pid_t twid;

    if (strcasecmp(argv[0], "all"))
        name = argv[0];
    if (argc == 3)
        node = argv[2];

    // printf ("This may take a while.\n");

    if (argc > 0)
    {
        twid = start_twiddle ();
        rc = rescan_lun_req(name, node);
        stop_twiddle (twid);
        if (rc)
            printf ("Done\n");
        else
            printf ("Could not rescan device(s) on node.\n");
    }

    return CMD_SUCCESS;
}

DEFUN(lun_topology,
      lun_topology_cmd,
      "lun topology",
      SPOOL_CMD_STR
      "show storage topology\n")
{
    char  rsp[SPM_MAX_RSP_MSG_SIZE];
    char  lrsp[SPM_MAX_RSP_MSG_SIZE];
    spm_MsgTotalCtrlsReq_t  ttlCtrlNumReq;
    spm_MsgTotalCtrlsRsp_t  *ttlCtrlNumRsp;
    int i, ctrls, j, ctrlNames, stat, k, devIdx;
    spm_MsgCtrlListReq_t ctrlListReq;
    spm_MsgCtrlListRsp_t *ctrlListRsp;
    spm_MsgCtrlNodeListReq_t cnListReq;
    spm_MsgCtrlNodeListRsp_t *cnListRsp;
    char lastCtrl[SDM_DEV_NAME_LEN];

    bzero(&ttlCtrlNumReq, sizeof(ttlCtrlNumReq));

    ttlCtrlNumReq.hdr.msgType = SPM_GET_TOTAL_CTRLS_REQ;
    ttlCtrlNumReq.hdr.clientPrivate = getpid();

    stat = sendAgileMsg(-1, &ttlCtrlNumReq, sizeof(ttlCtrlNumReq),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, rsp, SPM_MAX_RSP_MSG_SIZE);
    if (!bad_stat(stat) && (stat == sizeof(spm_MsgTotalCtrlsRsp_t)))
    {
        ttlCtrlNumRsp = (spm_MsgTotalCtrlsRsp_t *)rsp;
        ctrls = ttlCtrlNumRsp->num;
        // printf ("Total controllers : %d\n", ttlCtrlNumRsp->num);

        printf ("\nStorage topology report.\n");
        printf ("========================\n\n");
        printf ("Controllers                              Type     Nodes\n");
        printf ("-----------------------------------------------------------------------\n");
        bzero(lastCtrl, SDM_DEV_NAME_LEN);
        for (i = 0; i < ctrls; i += LUN_LIST_RECS)
        {
            bzero(&ctrlListReq, sizeof(ctrlListReq));
            ctrlListReq.hdr.msgType = SPM_GET_CTRL_LIST_REQ;
            ctrlListReq.hdr.clientPrivate = getpid();
            ctrlListReq.numReq = LUN_LIST_RECS;
            strcpy(ctrlListReq.lastCtrl, lastCtrl);
            bzero(rsp, SPM_MAX_RSP_MSG_SIZE);
            stat = sendAgileMsg(-1, &ctrlListReq, sizeof(ctrlListReq),
                                "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                &cfgd_to, rsp, SPM_MAX_RSP_MSG_SIZE);
            if (!bad_stat(stat) && (stat >= sizeof(spm_MsgCtrlListRsp_t)))
            {
                bzero(lastCtrl, SDM_DEV_NAME_LEN);
                ctrlListRsp = (spm_MsgCtrlListRsp_t *)rsp;
                ctrlNames = ctrlListRsp->numCtrls;
                // printf ("Total controller name : %d\n", ctrlNames);
                for (j = 0; j < ctrlNames; j++)
                {
                    // printf (" Name [%d] : %s\n", j+1, ctrlListRsp->ctrl[j]);
                    bzero(&cnListReq, sizeof(cnListReq));
                    cnListReq.hdr.msgType = SPM_GET_CTRL_NODE_LIST_REQ;
                    cnListReq.hdr.clientPrivate = getpid();
                    strcpy(cnListReq.ctrl, ctrlListRsp->ctrl[j]);
                    bzero(lrsp, SPM_MAX_RSP_MSG_SIZE);
                    stat = sendAgileMsg(-1, &cnListReq, sizeof(cnListReq),
                                     "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                     &cfgd_to, lrsp, SPM_MAX_RSP_MSG_SIZE);
                    if (!bad_stat(stat) &&
                                   (stat >= sizeof(spm_MsgCtrlNodeListRsp_t)))
                    {
                        cnListRsp = (spm_MsgCtrlNodeListRsp_t *)lrsp;
                        devIdx = cnListRsp->type;
                        if (cnListRsp->type > SPM_DEV_MISC)
                            devIdx = SPM_DEV_MISC;
                        printf ("%-39s  %-8s ",
                              ctrlListRsp->ctrl[j], spmDevTypeSting[devIdx] );
                        for (k = 0; k < cnListRsp->numNode; k++)
                            printf ("%s ", cnListRsp->nodes[k]);
                        printf ("\n");
                    }
                }
                strcpy(lastCtrl, ctrlListRsp->ctrl[ctrlNames-1]);
            }
        }
        printf ("Done\n");
    }

    return CMD_SUCCESS;
}

/*
 *------------------------------------------------------------------------------
 * Name         : lun_list_detail
 *
 * Description  : Display the list of LUN's that are visible to the cluster.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *------------------------------------------------------------------------------
 */

#define USG_DTL_1  "Usage: lun show {device DEV_NAME | port PORT} [-n NODE]\n"
#define USG_DTL_2  "       device DEV_NAME : Device Name\n"
#define USG_DTL_3  "       port PORT : Port number (sp.0 - sp.3)\n"
#define USG_DTL_4  "       -n NODE : cluster node name\n"

DEFUN(lun_show_detail_dev,
      lun_show_detail_dev_cmd,
      "lun show device DEVICE_NAME [OPT] [OPT]",
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      "Device detail\n"
      "Device Name\n"
      OPT_NODE
      OPT_NODE)
{
    spm_MsgDevDetailReq_t req;
    char                  lresp[SPM_MAX_RSP_MSG_SIZE], *errStr;
    spm_MsgDevDetailRsp_t *rsp;
    int32                 stat, rdIdx;
    int                   c;
    extern int            optind;
    int32 n, i;
    char  node[SPM_MAX_NODE_NAME_LEN+1], portStr[10];

    bzero (&req, sizeof(spm_MsgLunListReq_t));
    bzero(node, SPM_MAX_NODE_NAME_LEN+1);
    bzero(portStr, sizeof(char)*10);

    req.hdr.msgType = SPM_GET_DEV_DETAIL_REQ;
    req.hdr.clientPrivate = getpid();
    strcpy(req.hdr.node,"local"); // myNfxId;
    strcpy(req.devName, argv[0]);
    strcpy(req.tNode, "local");

    if (argc > 0)
    {
        optind = 0;
        while ((c = getopt (argc, argv, "n:")) != -1)
        {
            switch (c)
            {
                case 'n':
                    strcpy(node, optarg);
                    strcpy(req.tNode, node);
                    break;
                default:
                    printf ("%s%s", USG_DTL_1, USG_DTL_2);
                    printf ("%s%s", USG_DTL_3, USG_DTL_4);
                    return CMD_SUCCESS;
            }
        }

        if (optind != argc)
        {
            printf ("%s%s", USG_DTL_1, USG_DTL_2);
            printf ("%s%s", USG_DTL_3, USG_DTL_4);
            return CMD_SUCCESS;
        }
    }

    /*
     */

    stat = sendAgileMsg(-1, &req, sizeof(spm_MsgDevDetailReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);

    rsp = (spm_MsgDevDetailRsp_t *)lresp;
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgDevDetailRsp_t)) &&
        (rsp->hdr.msgType == SPM_GET_DEV_DETAIL_RSP))
    {
        n = strlen(node);
        printf ("\nDevice detail from %-s.\n", (n)?node:"PCC");
        printf ("====================");
        n = (n)?n:3;
        for (i = 0; i < n; i++) printf ("=");
        printf("\n\n");
        if (rsp->result != NFX_OK)
        {
            printf ("Device not found\n");
            printf("Done.\n");
            return CMD_SUCCESS;
        }
        printf ("Name                    : %s\n", rsp->devName);
        // printf ("WWN                     : 0x%llx\n", rsp->wwn);
        // printf ("LUN                     : %d\n", rsp->lun);
        printf ("Model                   : %s\n", rsp->model);
        printf ("Type                    : %s\n",
                                   (rsp->devType==SPM_DEV_DISK)?"Disk":
                                   (rsp->devType==SPM_DEV_TAPE)?"Tape":
                                   (rsp->devType==SPM_DEV_CHANGER)?"Changer":
                                   (rsp->devType==SPM_DEV_STACKER)?"Stacker":
                                                                   "Unknown");

        if (rsp->devType == SPM_DEV_DISK)
        {
            rdIdx = 8;
            if (rsp->rL == EVM_RAID_ANY) rdIdx = 7;
            else if (rsp->rL >= EVM_RAID0 && rsp->rL <= EVM_RAID_NONE) rdIdx = (int32)rsp->rL;

            printf ("Size (MB)               : %-16.2f\n",
               (double)(rsp->size*(uint64)EVM_LBLK_SIZE)/(double)BYTES_IN_MB);
            printf ("Raid Level              : %s\n", raidString[rdIdx]);
        }
        else
        {
            printf ("Size (MB)               : -\n");
            printf ("Raid Level              : NA\n");
        }

        if (rsp->devType == SPM_DEV_DISK)
        {
            printf ("State                   : %s\n",
                                     (rsp->state == SPM_LUN_USED)?"Used":
                                     (rsp->state == SPM_LUN_FREE)?"Free":
                                     (rsp->state == SPM_LUN_FOREIGN)?"Foreign":
                                                             "Out-of-Cluster");
            // printf ("Volume                  : %s\n", volName);
        }
        else
        {
            printf ("State                   : -\n");
            // printf ("Volume                  : -\n");
        }
        printf ("Nodes                   : ");
        for (i = 0; i < rsp->nodeCnt; i++)
            printf ("%-s", rsp->nodes[i]);
        printf ("\n");
        if (rsp->activeFlag)
        {
            sprintf (portStr, "sp.%-d", SCSI_PATH_GET_PORT(rsp->active));
            printf ("Active path             : %-s\n", portStr);
        }
        else
            printf ("Active path             : \n");
        printf ("Path list               : Port  WWN                 LUN\n");
        for (i = 0; i < rsp->pathCnt; i++)
        {
            bzero(portStr, sizeof(char)*10);
            sprintf (portStr, "sp.%-d", SCSI_PATH_GET_PORT(rsp->path[i].path));
            printf ("                          %-4s  0x%-16llx  %-d\n",
                    portStr, rsp->path[i].wwn, rsp->path[i].lun);
        }
    }
    else
    {
        printf("Failed to get response from SPM\n");
        if (stat > sizeof(spm_MsgStat_t))
        {
            errStr = (lresp + sizeof(spm_MsgStat_t));
            if (errStr) printf ("%s\n", errStr);
        }
    }
    printf("Done.\n");
    return CMD_SUCCESS;
}


/*
 *------------------------------------------------------------------------------
 * Name         : lun_list_detail
 *
 * Description  : Display the list of LUN's that are visible to the cluster.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 10-SEPT-2002
 *------------------------------------------------------------------------------
 */

DEFUN(lun_show_detail_port,
      lun_show_detail_port_cmd,
      "lun show port PORT [OPT] [OPT]",
      SPOOL_CMD_STR
      SPOOL_CMD_STR
      "Device detail\n"
      "Port Number (sp.0 - sp.3)\n"
      OPT_NODE
      OPT_NODE)
{
    spm_MsgDevPortReq_t   lunReq;
    spm_MsgTotalLunReq_t  numReq;
    char                  lresp[SPM_MAX_RSP_MSG_SIZE], *errStr;
    spm_MsgDevPortRsp_t   *lunRsp;
    spm_MsgTotalLunRsp_t  *numRsp;
    int32                 stat;
    int32                 lCnt = 0, currCnt = 0;
    int                   c, port;
    extern int            optind;
    uint64 getNext = 0;
    int32 next = 0, head = 1, n, i;
    char  node[SPM_MAX_NODE_NAME_LEN+1];

    if (!strcasecmp(argv[0], "sp.0"))
        port = 0;
    else if (!strcasecmp(argv[0], "sp.1"))
        port = 1;
    else if (!strcasecmp(argv[0], "sp.2"))
        port = 2;
    else if (!strcasecmp(argv[0], "sp.3"))
        port = 3;
    else
    {
        printf ("%s%s", USG_DTL_1, USG_DTL_2);
        printf ("%s%s", USG_DTL_3, USG_DTL_4);
        return CMD_SUCCESS;
    }
    bzero (&lunReq, sizeof(spm_MsgDevPortReq_t));
    bzero (&numReq, sizeof(spm_MsgTotalLunReq_t));
    bzero(node, SPM_MAX_NODE_NAME_LEN+1);

    lunReq.hdr.msgType = SPM_GET_DEV_PORT_REQ;
    numReq.hdr.msgType = SPM_GET_TOTAL_LUNS_REQ;
    lunReq.hdr.clientPrivate = getpid();
    numReq.hdr.clientPrivate = getpid();
    strcpy(lunReq.hdr.node,"local"); // myNfxId;
    strcpy(numReq.hdr.node,"local"); // myNfxId;
    numReq.diskOtherAll = 1;
    numReq.rL = EVM_RAID_ANY; // default
    numReq.lunType = SPM_LUN_ALL_LIST;
    strcpy(lunReq.tNode, "local");
    lunReq.port = port;

    if (argc > 0)
    {
        optind = 0;
        while ((c = getopt (argc, argv, "n:")) != -1)
        {
            switch (c)
            {
                case 'n':
                    strcpy(node, optarg);
                    strcpy(lunReq.tNode, node);
                    break;
                default:
                    printf ("%s%s", USG_DTL_1, USG_DTL_2);
                    printf ("%s%s", USG_DTL_3, USG_DTL_4);
                    return CMD_SUCCESS;
            }
        }

        if (optind != argc)
        {
            printf ("%s%s", USG_DTL_1, USG_DTL_2);
            printf ("%s%s", USG_DTL_3, USG_DTL_4);
            return CMD_SUCCESS;
        }
    }

    /*
     * Get number of LUN's
     */

    stat = sendAgileMsg(-1, &numReq, sizeof(spm_MsgTotalLunReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);

    numRsp = (spm_MsgTotalLunRsp_t *)lresp;
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgTotalLunRsp_t)) &&
        (numRsp->hdr.msgType == SPM_GET_TOTAL_LUNS_RSP))
        lCnt = numRsp->numLun;
    else
    {
        printf("Failed to get response from SPM\n");
        if (stat > sizeof(spm_MsgStat_t))
        {
            errStr = (lresp + sizeof(spm_MsgStat_t));
            if (errStr) printf ("%s\n", errStr);
        }
        return CMD_SUCCESS;
    }
    bzero(lresp, SPM_MAX_RSP_MSG_SIZE);

    lunReq.lastSID = 0;
    lunReq.numReq  = LUN_LIST_RECS;
    n = strlen(node);
    printf ("\nDevices detail from %-s on port sp.%-2d\n", (n)?node:"PCC", port);
    printf ("=================================");
    n = (n)?n:3;
    for (i = 0; i < n; i++) printf ("=");
    printf("\n\n");
    while(next != -1 && lCnt > 0)
    {
        stat = sendAgileMsg(-1, &lunReq, sizeof(spm_MsgDevPortReq_t),
                            "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                            &cfgd_to, lresp, SPM_MAX_RSP_MSG_SIZE);
        lunRsp = (spm_MsgDevPortRsp_t *)lresp;
        if (!bad_stat(stat) &&
            (stat >= sizeof(spm_MsgDevPortRsp_t)) &&
            (lunRsp->hdr.msgType == SPM_GET_DEV_PORT_RSP))
        {
            if (lunRsp->numLuns > 0)
            {
                if (currCnt >= LUN_LIST_RECS)
                {
                    printf("---more---[q:Quit][Enter] ");
                    c = getchar();
                    if (c == 'q' || c == 'Q')
                    {
                        getchar();
                        return CMD_SUCCESS;
                    }
                }
                getNext = display_lun_list_all(lunRsp, head);
                if (getNext == -1) next = -1;
                head = 0;
            }
            else
                next = -1;
        }
        else
        {
            printf("Failed to get list of devices\n");
            next = -1;
        }
        currCnt += lunRsp->numLuns;
        if (currCnt >= lCnt)
            next = -1;
        lunReq.lastSID = getNext;
        bzero(lresp, SPM_MAX_RSP_MSG_SIZE);
    }
    printf("Done.\n");
    return CMD_SUCCESS;
}


DEFUN(spool_list,
      spool_list_cmd,
      "spool list",
      SPOOL_CMD_STR
      "List the storage pools in SPM\n")
{
    spm_MsgStoragePoolListReq_t  poolReq;
    char  rsp[SPM_MAX_RSP_MSG_SIZE];
    spm_MsgStoragePoolListRsp_t  *poolRsp;
    int32                        cnt = 0, stat, rdIdx;
    evm_raid_t rL;

    poolReq.hdr.msgType = SPM_GET_STORAGE_POOL_LIST_REQ;
    poolReq.hdr.clientPrivate = getpid();

    stat = sendAgileMsg(-1, &poolReq, sizeof(spm_MsgStoragePoolListReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, rsp, SPM_MAX_RSP_MSG_SIZE);
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgStoragePoolListRsp_t)))
    {
        poolRsp = (spm_MsgStoragePoolListRsp_t *)rsp;
        if (poolRsp->hdr.msgType == SPM_GET_STORAGE_POOL_LIST_RSP &&
            poolRsp->num > 0)
        {
            printf("Id        Raid    Controller                            Model             Used  Free  Fore  Out-C\n");
            printf("-----------------------------------------------------------------------------------------------------\n");
            for (cnt = 0; cnt < poolRsp->num; cnt++)
            {
                printf ("%-8d  ", poolRsp->strg[cnt].poolId);

                rL = poolRsp->strg[cnt].raidLevel;
                rdIdx = 8;
                if (rL == EVM_RAID_ANY) rdIdx = 7;
                else if (rL >= EVM_RAID0 && rL <= EVM_RAID_NONE)
                          rdIdx = (int32)rL;
                printf ("%-8s", raidString[rdIdx]);
                printf ("%-36s  ", poolRsp->strg[cnt].ctrl);
                printf ("%-16s  ", (strlen(poolRsp->strg[cnt].model))?
                                 poolRsp->strg[cnt].model:"No Model");
                printf ("%-4d  ", poolRsp->strg[cnt].usedLun);
                printf ("%-4d  ", poolRsp->strg[cnt].freeLun);
                printf ("%-4d  ", poolRsp->strg[cnt].foreLun);
                printf ("%-4d\n", poolRsp->strg[cnt].outcLun);
            }
            printf ("Done\n");
        }
    }
    else
         printf("SPM error\n");

    return CMD_SUCCESS;
}


DEFUN(spool_lun_list,
      spool_lun_list_cmd,
      "spool lun list POOL-ID",
      SPOOL_CMD_STR
      "List the LUNs in a pool\n")
{
    spm_MsgStorageLunReq_t  lunReq;
    char  rsp[SPM_MAX_RSP_MSG_SIZE];
    spm_MsgStorageLunRsp_t  *lunRsp;
    int32                   cnt = 0, stat;

    bzero(&lunReq, sizeof(spm_MsgStorageLunReq_t));
    lunReq.hdr.msgType = SPM_GET_STORAGE_LUN_LIST_REQ;
    lunReq.hdr.clientPrivate = getpid();
    if (argc == 1)
        lunReq.poolId = atoi(argv[0]);
    else
    {
        printf("Usage :\n\tspool lun list <pool-id>\n");
        return (CMD_ERR_FAILED);
    }
    // printf ("Pool Id : %d\n", atoi(argv[0]));

    stat = sendAgileMsg(-1, &lunReq, sizeof(spm_MsgStorageLunReq_t),
                        "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                        &cfgd_to, rsp, SPM_MAX_RSP_MSG_SIZE);
    if (!bad_stat(stat) && (stat >= sizeof(spm_MsgStorageLunRsp_t)))
    {
        lunRsp = (spm_MsgStorageLunRsp_t *)rsp;
        if (lunRsp->hdr.msgType == SPM_GET_STORAGE_LUN_LIST_RSP &&
            lunRsp->num > 0)
        {
            double mbSize = 0;

            printf("List of LUN's in pool : %d\n\n", lunReq.poolId);
            printf("SCSI-ID\t\t\tLun Size (MB)   Lun State\tNFX ref cnt\n");
            printf("-------------------------------------------------------------------\n");
            for (cnt = 0; cnt < lunRsp->num; cnt++)
            {
                mbSize = (double)(lunRsp->luns[cnt].lunSize*(uint64)EVM_LBLK_SIZE)/(double)BYTES_IN_MB;
                printf ("0x%-22llx", lunRsp->luns[cnt].scsiId);
                printf ("%-13.2f", mbSize);
                printf ("   %-16s", spmString[lunRsp->luns[cnt].lunState]);
                printf ("%-d\n", lunRsp->luns[cnt].nfxRefCnt);
            }
            printf ("Done\n");
        }
    }
    else
         printf("SPM error\n");

    return CMD_SUCCESS;
}


DEFUN(spool_dbg_level,
      spool_dbg_level_cmd,
      "spool debug (none | elog | function | error | debug | memory)",
      SPOOL_CMD_STR
      "level for SPM\n")
{
    spm_setDebugLevel_t req;

    bzero(&req, sizeof(spm_setDebugLevel_t));
    req.hdr.msgType = SPM_DEBUG_LEVEL;
    req.hdr.clientPrivate = 0;
    if (argc == 1)
    {
        if (!strcasecmp(argv[0], "none"))
            req.dbg_level = SPM_DBG_LEVEL_NONE;
        else if (!strcasecmp(argv[0], "function"))
            req.dbg_level = SPM_DBG_LEVEL_FCT;
        else if (!strcasecmp(argv[0], "error"))
            req.dbg_level = SPM_DBG_LEVEL_ERR;
        else if (!strcasecmp(argv[0], "debug"))
            req.dbg_level = SPM_DBG_LEVEL_DBG;
        else if (!strcasecmp(argv[0], "memory"))
            req.dbg_level = SPM_DBG_LEVEL_MEM;
        else
            req.dbg_level = SPM_DBG_LEVEL_ELOG;

        sendAgileMsg(-1, &req, sizeof(spm_setDebugLevel_t),
                 "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                 NULL, NULL, 0);

        return CMD_SUCCESS;
    }
    else
        printf ("Usage :\n\tspool debug (none | elog | function | error | debug | memory)\n");

    return CMD_ERR_FAILED;
}


#ifdef SIMULATE_EA_AV_SM_INTERFACE
DEFUN(spool_lun_alloc,
      spool_lun_alloc_cmd,
      "spool lun alloc TXN-ID NAME",
      SPOOL_CMD_STR
      "Allocate response from AV\n")
{
    ea_growFsysInd_t  eaInd;

    bzero(&eaInd, sizeof(ea_growFsysInd_t));
    eaInd.hdr.msgType = EA_GROW_FSYS_IND;
    eaInd.hdr.userInfo = getpid();
    eaInd.txnId = strtoq(argv[0], NULL, 0);
    eaInd.num = 1;
    strcpy(eaInd.devName[0], argv[1]);

    sendAgileMsg(-1, &eaInd, sizeof(ea_growFsysInd_t),
                 "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                 NULL, NULL, 0);

    return CMD_SUCCESS;
}


DEFUN(spool_lun_add,
      spool_lun_add_cmd,
      "spool lun add NAME WWN LUN SIZE",
      SPOOL_CMD_STR
      ""
      "Add LUN from SDM\n"
      "Enter WWN\n"
      "Enter LUN number\n"
      "Enter Size\n")
{
    spm_MsgAddLunReq_t addReq;

    bzero(&addReq, sizeof(spm_MsgAddLunReq_t));
    addReq.hdr.msgType = SPM_ADD_LUN_REQ;
    gethostname(addReq.hdr.node, 256);
    addReq.lunAttr.scsiId = 0x13241234123412;
    strcpy(addReq.lunAttr.name, argv[0]);
    addReq.lunAttr.wwn = strtoq(argv[1], NULL, 0);
    addReq.lunAttr.lun = strtoq(argv[2], NULL, 0);
    addReq.lunAttr.size = strtoq(argv[3], NULL, 0);
    addReq.lunAttr.state = EVM_OC_USED_LUN;
    addReq.lunAttr.raidLevel = EVM_RAID_NONE;
    strcpy(addReq.lunAttr.vendor, "HITACHI");
    strcpy(addReq.lunAttr.model, "OPEN-9");

    sendAgileMsg(-1, &addReq, sizeof(spm_MsgAddLunReq_t),
                 "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                 NULL, NULL, 0);

    return CMD_SUCCESS;
}


DEFUN(spool_lun_del,
      spool_lun_del_cmd,
      "spool lun del SCSIID",
      SPOOL_CMD_STR
      ""
      "Delete LUN from SDM\n"
      "Enter scsi-id\n")
{
    spm_MsgDelLunReq_t delReq;

    bzero(&delReq, sizeof(delReq));
    delReq.hdr.msgType = SPM_DEL_LUN_REQ;
    gethostname(delReq.hdr.node, 256);
    delReq.id = strtoq(argv[0], NULL, 0);

    sendAgileMsg(-1, &delReq, sizeof(delReq),
                 "spm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                 NULL, NULL, 0);

    return CMD_SUCCESS;
}

#endif


void
install_spool (priv_t myPriv, enum node_type node)
{
/*
	Organizing this switch statment such that the higher priority commands
	are first allows us to fall through and pick up the lower priority
	ones at no extra cost.
 */
	switch (myPriv) {
		case diagPriv:
#ifdef SIMULATE_EA_AV_SM_INTERFACE
            install_element(node, &spool_lun_alloc_cmd);
            install_element(node, &spool_lun_add_cmd);
            install_element(node, &spool_lun_del_cmd);
#endif
            install_element(node, &spool_dbg_level_cmd);
            install_element(node, &spool_list_cmd);
            install_element(node, &spool_lun_list_cmd);
            install_element(node, &lun_rescan_cmd);
            install_element(node, &lun_mod_cluster_cmd);
		case fullPriv:
		case changePriv:
            install_element(node, &lun_label_cmd);
            install_element(node, &lun_unlabel_cmd);
		case addPriv:
		case viewPriv:
            install_element(node, &lun_show_cmd);
            install_element(node, &lun_show_tape_cmd);
            install_element(node, &lun_show_disk_cmd);
            install_element(node, &lun_show_all_cmd);
            install_element(node, &lun_topology_cmd);
            install_element(node, &lun_show_detail_dev_cmd);
            install_element(node, &lun_show_detail_port_cmd);
/*
	These commands are available in all modes
 */
	}
}
