
/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: cmd_upgrade
 *
 * RCS:  $Id: cmd_upgrade.c,v 1.14 2002/11/22 20:20:16 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created: 04/29/2002
 *
 *-----------------------------------------------------------------
 */

/* Some parts are: */
/*
 * Copyright (c) 1987, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */



#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include "vector.h"

#include <dirent.h>

#include "nfx-defs.h"
#include "exports/cm-api.h"
#include "exports/anpssc-api.h"
#include "exports/opt-api.h"

#include "exports/nfxnis-api.h"


#define ALLOW_PROM_UPGRADE_FAIL 1

extern int mount_secondary(vector v);
extern vector make_secondary_fstabs(void);
extern void u_mount_secondary(vector v);
extern void free_fstab_vector(vector v);
extern void do_copy_files(char *fromdir, char *todir, char *files);

struct descriptor
{
	char *name;
	char *type;
	char *board;
	char *patch_name;
};


struct descriptor default_descriptor = {
	NULL,
	"app",
	"ssc",
	NULL
};


char *board_types[] = {
	"sp",
	"txrx",
	"fp",
	"ssc",
	"toshiba",
	NULL
};


char *file_types[] = {
	"prom",
	"bigprom",
	"runtime",
	"daemon",
	"app",
	"descr",
	"version",
	"kernel",
	"config",
	"services",
	NULL
};


static char *mount_dir = "/mnt";
static char *patch_dir_name = NULL;
static char *local_dest = "/mnt2";
static char *descriptor_file = "/usr/local/agile/etc/descr";
static char *mtree_file_name = "/etc/mtree/4.4BSD.dist";
static char ftpprefix[] = "ftp://";
static char *patch_location;
static int upgrade_secondary;

extern char *secondary_mount;
extern char *sys_config_files;

struct version
{
	int major;
	int minor;
	int patch;
};


static void print_vector(vector v, void (*print_fn)(void*)) __attribute((unused));
static void print_descriptor_vector(vector v) __attribute__ ((unused));
static void print_string_vector(vector v) __attribute__ ((unused));


typedef int (*upgrade_method_t)(vector descriptors,
								vector files, char *(*truename_func)(char *),
								char *mtree_files);

typedef char* (*truename_func_t)(char *file);

truename_func_t cur_truename_func;

extern int noisy_make_allfs_writable(void);


/*-----------------------------------------------------------------
 * Name :	find_str()
 *
 * Description: Find a string in array of strings
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static char *
find_str(char **array, char *str)
{
	char **ptr;
	for (ptr = &array[0]; *ptr != NULL && strcmp(*ptr, str); ++ptr) {
		/* */
	}
	return *ptr;
}


/*-----------------------------------------------------------------
 * Name :	vector_free_all()
 *
 * Description: Free the vector and call free() for all the elements
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
vector_free_all(vector v)
{
	int i;

	if (v != NULL) {
		for (i = 0; i < vector_max(v); ++i) {
			zfree(MTYPE_VECTOR, vector_lookup_index(v, i));
		}
		vector_free(v);
	}
}



/*-----------------------------------------------------------------
 * Name :	read_descriptor()
 *
 * Description: Read the descriptor file and return the vector of
 * struct descriptor.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static vector
read_descriptor(const char *filename)
{
	FILE *f = fopen(filename, "r");
	vector v = NULL;
	if (f != NULL) {
		v = vector_init(200);
		while (!feof(f)) {
			size_t size;
			char *str = fgetln(f, &size);
			if (str != NULL) {
				struct descriptor *descr = zmalloc(MTYPE_VECTOR,
												   sizeof(struct descriptor));
				char type[100], board[100], name[4096];
				int rc = sscanf(str, "%s %s %s\n", type, board, name);
				assert(rc == 3);
				descr->name = zstrdup(MTYPE_VECTOR, name);
				descr->patch_name = NULL;
				descr->type = find_str(file_types, type);
				assert(descr->type != NULL);
				descr->board = find_str(board_types, board);
				assert(descr->board != NULL);
				vector_push(v, descr);
			}
		}
		fclose(f);
	}
	return v;
}


/*-----------------------------------------------------------------
 * Name :	print-vector()
 *
 * Description: print the vector elements using function print_fn
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
print_vector(vector v, void (*print_fn)(void *))
{
	int i;
	for (i = 0; i < vector_max(v); ++i) {
		(*print_fn)(vector_lookup_index(v, i));
	}
}


/*-----------------------------------------------------------------
 * Name :	print_descriptor()
 *
 * Description: print function for descriptors
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
print_descriptor(void *vd)
{
	struct descriptor *d = vd;
	printf("%s %s %s %s\n", d->type, d->board, d->name,
		   d->patch_name ? d->patch_name : "(null)");
}


/*-----------------------------------------------------------------
 * Name :	print_descriptor_vector()
 *
 * Description: print vector of struct descriptor
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
print_descriptor_vector(vector v) 
{
	print_vector(v, print_descriptor);
}


/*-----------------------------------------------------------------
 * Name :	print_string_vector()
 *
 * Description: print vector of strings
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
print_string_vector(vector v)
{
	print_vector(v, (void (*)(void *))puts);
}


/*-----------------------------------------------------------------
 * Name :	parse_version_str()
 *
 * Description: parse a string into struct version
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
parse_version_str(char *str, struct version *version)
{
	int rc = sscanf(str, "%d.%d.%d",
					&version->major, &version->minor, &version->patch);
	if (rc != 3) {
		printf("invalid version string %s", str);
		rc = CMD_ERR_FAILED;
	}
	else {
		rc = CMD_SUCCESS;
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	compare_versions()
 *
 * Description: compare two versions, return value is < 0 if the first is
 * lower, 0 if equal, > 0 if the first is greater
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
compare_versions(struct version *v1, struct version *v2)
{
	int major_diff, minor_diff;
	if ((major_diff = v1->major - v2->major) != 0)
		return major_diff;
	else if ((minor_diff = v1->minor - v2->minor) != 0)
		return minor_diff;
	else return v1->patch - v2->patch;
}


/*-----------------------------------------------------------------
 * Name :	valid_slot_cpu()
 *
 * Description: Return true if slot/cpu combination is valid
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
valid_slot_cpu(int slot, int cpu)
{
	switch (slot) {
	case 0 :
		return 0 <= cpu && cpu < 2;
	case 1:
		return 0 <= cpu && cpu < 4;
	case 2:
		return 0 <= cpu && cpu < 4;
	default:
		return 0;
	}
	
}


/*-----------------------------------------------------------------
 * Name :	read_version()
 *
 * Description: Read the version from the file named filename
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
read_version(const char *filename, struct version *version)
{
	FILE *f = fopen(filename, "r");
	char *s;
	int ok = CMD_ERR_FAILED;

	if (f != NULL) {
		size_t size;
		s = fgetln(f, &size);
		if (s != NULL && s[size - 1] == '\n') {
			s[size - 1] = 0;
			ok = parse_version_str(s, version);
		}
		fclose(f);
	}

	return ok;
}


/*-----------------------------------------------------------------
 * Name :	is_config()
 *
 * Description: Return true if the descriptor type is 'config'
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
is_config(struct descriptor *descr)
{
	return !strcmp(descr->type, "config");
}


/*-----------------------------------------------------------------
 * Name :	is_version()
 *
 * Description: Return true if the descriptor type is 'version'
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
is_version(struct descriptor *descr)
{
	return !strcmp(descr->type, "version");
}


/*-----------------------------------------------------------------
 * Name :	is_prom()
 *
 * Description: Return true if the descriptor type is prom
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
is_prom(struct descriptor *descr)
{
	return !strcmp(descr->type, "prom");
}


/*-----------------------------------------------------------------
 * Name :	is_bigprom()
 *
 * Description: Return true if the descriptor type is bigprom
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
is_bigprom(struct descriptor *descr)
{
	return !strcmp(descr->type, "bigprom");
}


/*-----------------------------------------------------------------
 * Name :	print_version()
 *
 * Description: 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
print_version(struct version *version)
{
	printf("%d.%d.%d\n", version->major, version->minor, version->patch);
}


/*-----------------------------------------------------------------
 * Name :	convert_rc()
 *
 * Description: Convert unix return code to nfxsh values
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	11/01/02
 *
 *-----------------------------------------------------------------
 */
static inline int
convert_rc(int rc)
{
	return rc == 0 ? CMD_SUCCESS : CMD_ERR_FAILED;
}


/*-----------------------------------------------------------------
 * Name :	is_ftp_upgrade()
 *
 * Description: return true if upgrading using ftp
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	11/01/02
 *
 *-----------------------------------------------------------------
 */
static int
is_ftp_upgrade(char *from)
{
	return !strncmp(from, ftpprefix, strlen(ftpprefix));
}


/*-----------------------------------------------------------------
 * Name :	umount_patch()
 *
 * Description: unmount patch directory if it is mounted
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
umount_patch()
{
	char command_str[4096];

	sprintf(command_str, "/sbin/umount %s 2>/dev/null", mount_dir);
	(void)system(command_str);
}


/*-----------------------------------------------------------------
 * Name :	mount_patch()
 *
 * Description: mount the patch directory 'from' on mount_dir
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
mount_patch(char *from)
{
	char command_str[4096];

	sprintf(command_str, "/sbin/umount -f %s 2>/dev/null", mount_dir);
	(void)system(command_str);

	if (is_ftp_upgrade(from)) {
		int rc;

		printf("Downloading upgrade distribution.\n");
		sprintf(command_str, "/sbin/mount_mfs -s 300000 swap %s", mount_dir);
		rc = convert_rc(system(command_str));
		if (rc == CMD_SUCCESS) {
			sprintf(command_str, "/usr/bin/ftp -o - -m %s | (cd %s; tar xzpf -)",
					from, mount_dir);
			rc = convert_rc(system(command_str));
			patch_dir_name = mount_dir;
			if (rc == CMD_SUCCESS) {
				printf("Done.\n");
			}
			else {
				umount_patch();
				printf("Download failed.\n");
			}
		}
		return rc;
	}
	else {
		sprintf(command_str, "/sbin/umount %s 2>/dev/null", mount_dir);
		(void)system(command_str);

		sprintf(command_str, "/sbin/mount -t nfs %s %s", from, mount_dir);
		patch_dir_name = mount_dir;
	
		return convert_rc(system(command_str));
	}
}


/*-----------------------------------------------------------------
 * Name :	get_files_list()
 *
 * Description: read the directory into the vector v
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
get_files_list(DIR *dir, vector v)
{
	struct dirent *dp;
	while ((dp = readdir(dir)) != NULL) {
		if (strcmp(dp->d_name, ".") && strcmp(dp->d_name, "..")) {
			vector_push(v, zstrdup(MTYPE_VECTOR, dp->d_name));
		}
	}
	return CMD_SUCCESS;
}


/*-----------------------------------------------------------------
 * Name :	get_all_patch_dirs()
 *
 * Description: get the list of patch directories
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
get_all_patch_dirs(vector *vector_ptr)
{
	DIR *patch_dir = opendir(patch_dir_name);
	int rc = CMD_ERR_FAILED;
	if (patch_dir != NULL) {
		vector dirs = vector_init(10);
		get_files_list(patch_dir, dirs);
		vector_sort(dirs, (int (*)(const void*, const void*))strcmp);
		*vector_ptr = dirs;
		rc = CMD_SUCCESS;
		closedir(patch_dir);
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	get_patch_version()
 *
 * Description: determine the version of the patch
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
get_patch_version(struct version *version)
{
	vector dirs;
	int rc = get_all_patch_dirs(&dirs);
	if (rc == CMD_SUCCESS) {
		if (vector_max(dirs) != 0) {
			rc = parse_version_str(vector_lookup_index(dirs,
													   vector_max(dirs) - 1),
								   version);
		}
		else {
			rc = CMD_ERR_FAILED;
		}
		vector_free_all(dirs);
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	get_upgrade_version()
 *
 * Description: determine the version of system upgrade
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
get_upgrade_version(struct version *version)
{
	char version_file[MAXPATHLEN + 1];
	strcpy(version_file, patch_dir_name);
	strcat(version_file, "/version");
	return read_version(version_file, version);
}

/*-----------------------------------------------------------------
 * Name :	zero_version()
 *
 * Description: Fill version with zeroes
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	07/03/02
 *
 *-----------------------------------------------------------------
 */
static void
zero_version(struct version *version)
{
	version->major = version->minor = version->patch = 0;
}


/*-----------------------------------------------------------------
 * Name :	get_org_upgrade_version()
 *
 * Description: read original version of upgrade
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	07/03/02
 *
 *-----------------------------------------------------------------
 */
static void
get_org_upgrade_version(struct version *version)
{
	char version_file[MAXPATHLEN + 1];
	strcpy(version_file, patch_dir_name);
	strcat(version_file, "/version.org");
	if (read_version(version_file, version) != 0) {
		zero_version(version);
	}
}


/*-----------------------------------------------------------------
 * Name :	is_dotdir()
 *
 * Description: return true if the name is "." or "..".
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
is_dotdir(char *name)
{
	return (name[0] == '.' && name[1] == '\0')
		|| (name[0] == '.' && name[1] == '.' && name[2] == '\0');
}


/*-----------------------------------------------------------------
 * Name :	find_files()
 *
 * Description: Find the list of the files in the directory
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
find_files(const char *dir_name, vector files)
{
	int rc = CMD_ERR_FAILED;
	DIR *dir = opendir(dir_name);
	if (dir != NULL) {
		struct dirent *dp;
		int error = 0;
		while (!error && (dp = readdir(dir)) != NULL) {
			if (!is_dotdir(dp->d_name) && strcmp(dp->d_name, ".snapshot")) {
				struct stat stbuf;
				char *fname = zmalloc(MTYPE_VECTOR,
									  strlen(dir_name) + 2 + dp->d_reclen);
				sprintf(fname, "%s/%s", dir_name, dp->d_name);
				if (stat(fname, &stbuf) < 0) {
					fprintf(stderr, "Error: can't stat %s: %s\n", fname,
							strerror(errno));
					error = 1;
				}
				else {
					if (S_ISREG(stbuf.st_mode)) {
						vector_push(files, zstrdup(MTYPE_VECTOR, fname));
					}
					else if (S_ISDIR(stbuf.st_mode)) {
						if (find_files(fname, files) != CMD_SUCCESS) {
							error = 1;
						}
					}
					zfree(MTYPE_VECTOR, fname);
				}
			}
		}
		closedir(dir);
		rc = error ? CMD_ERR_FAILED : CMD_SUCCESS;
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	get_local_version()
 *
 * Description: determine the current system version
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
get_local_version(struct version *version)
{
	char version_file[MAXPATHLEN + 1];
	strcpy(version_file, patch_dir_name);
	strcat(version_file, "/version");
	if (read_version(cur_truename_func(version_file),
					 version) != CMD_SUCCESS) {
		version->major = 0;
		version->minor = 0;
		version->patch = 0;
	}
	return CMD_SUCCESS;
}


/*-----------------------------------------------------------------
 * Name :	get_root_version()
 *
 * Description: get system version from root directory
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	01/10/03
 *
 *-----------------------------------------------------------------
 */
static int
get_root_version(struct version *version)
{
	if (read_version("/version", version) != CMD_SUCCESS) {
		version->major = 0;
		version->minor = 0;
		version->patch = 0;
	}
	return CMD_SUCCESS;
}


/*-----------------------------------------------------------------
 * Name :	compare_descr_file()
 *
 * Description: compare the descriptors using the file names
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
compare_descr_file(void *descr, void *fname)
{
	return !strcmp(((struct descriptor *)descr)->name, fname);
}


/*-----------------------------------------------------------------
 * Name :	get_file_descriptor()
 *
 * Description: find the descriptor for the file in the descriptor
 * vector dv.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static struct descriptor *
get_file_descriptor(vector dv, char *fname)
{
	struct descriptor *descr = vector_find(dv, compare_descr_file, fname);
	if (descr == NULL)
		descr = &default_descriptor;
	return descr;
}


/*-----------------------------------------------------------------
 * Name :	compare_file_contents()
 *
 * Description: compare the file contents
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
compare_file_contents(int from_fd, int to_fd, char *src_name, int from_len, char *dst_name,
		int to_len, int *result)
{
	caddr_t p1, p2;
	size_t length, remainder;
	off_t from_off, to_off;
	int dfound = 0;
	int rc;

	if (from_len != to_len) {
		*result = 1;
		return CMD_SUCCESS;
	}

	/*
	 * Compare the two files being careful not to mmap
	 * more than 8M at a time.
	 */
	rc = CMD_SUCCESS;
	from_off = to_off = (off_t)0;
	remainder = from_len;
	do {
		length = MIN(remainder, 8 * 1048576);
		remainder -= length;
		p1 = MAP_FAILED;
		p2 = MAP_FAILED;

		if ((p1 = mmap(NULL, length, PROT_READ, MAP_PRIVATE,
					   from_fd, from_off)) == MAP_FAILED) {
			rc = CMD_ERR_FAILED;
		}
		if ((p2 = mmap(NULL, length, PROT_READ, MAP_PRIVATE,
					   to_fd, to_off)) == MAP_FAILED) {
			rc = CMD_ERR_FAILED;
		}
		if (rc == CMD_SUCCESS) {
			dfound = memcmp(p1, p2, length);

			from_off += length;
			to_off += length;
		}

		if (p1 != MAP_FAILED)
			(void) munmap(p1, length);
		if (p2 != MAP_FAILED)
			(void) munmap(p2, length);
	} while (rc == CMD_SUCCESS && !dfound && remainder > 0);

	*result = dfound;
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	file_exists()
 *
 * Description: return true if the file exists
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
file_exists(char *file)
{
	struct stat statbuf;
	return stat(file, &statbuf) == 0;
}


/*-----------------------------------------------------------------
 * Name :	upgrade_truename()
 *
 * Description: return the installation name of a file in upgrade
 * directory.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static char*
upgrade_truename(char *file)
{
	return strchr(file + 1, '/');
}


static char *
secondary_upgrade_truename(char *file)
{
	static char buf[MAXPATHLEN];
	strcpy(buf, secondary_mount);
	strcat(buf, strchr(file + 1, '/'));
	return buf;
}


/*-----------------------------------------------------------------
 * Name :	shutdown_system()
 *
 * Description: tell the system daemons to shutdown
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
shutdown_system()
{
	printf("Shutting down the system software...");
	fflush(stdout);
	if (file_exists("/var/run/mcp.pid")) {
		system("/usr/local/agile/bin/sysdown");
	}
	printf("done\n");
}


/*-----------------------------------------------------------------
 * Name :	kill_daemon()
 *
 * Description: kill the daemon, pid is in pid_file and the daemon
 * is supposed to remove the file when exiting on SIGTERM
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
void
kill_daemon(char *pid_file)
{
	FILE *f;
	int pid;
	int time_to_sleep = 10;

	f = fopen(pid_file, "r");
	if (f) {
		fscanf(f, "%d\n", &pid);
		fclose(f);
		kill(pid, SIGTERM);
		while (time_to_sleep && file_exists(pid_file)) {
			sleep(1);
			--time_to_sleep;
		}
		if (time_to_sleep == 0) {
			kill(pid, SIGKILL);
			unlink(pid_file);
		}
	}
}


/*-----------------------------------------------------------------
 * Name :	shutdown_inetd()
 *
 * Description: shutdown the inetd
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
shutdown_inetd(void)
{
	printf("Shutting down the inetd...");
	fflush(stdout);
	kill_daemon("/var/run/inetd.pid");
	printf("done\n");
}


/*-----------------------------------------------------------------
 * Name :	shutdown_mcp()
 *
 * Description: shutdown the mcp
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
shutdown_mcp(void)
{
	printf("Shutting down the MCP...");
	fflush(stdout);
	kill_daemon("/var/run/mcp.pid");
	printf("done\n");
}


/*-----------------------------------------------------------------
 * Name :	upgrade_files()
 *
 * Description: install the files in the vector files. 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_files(vector files, char *(*truename_func)(char *))
{
	int i;
	int error = 0;
	for (i = 0; error == 0 && i < vector_max(files); ++i) {
		char *src_file = vector_lookup_index(files, i);
		char *dst_file = (*truename_func)(src_file);
		struct stat src_st;
		printf("Installing %s...", dst_file);
		fflush(stdout);
		if (lstat(src_file, &src_st) == 0) {
			if (S_ISREG(src_st.st_mode)) {
				mode_t mode = src_st.st_mode & 07777;
				uid_t owner = src_st.st_uid;
				gid_t group = src_st.st_gid;
				char install_command[1024];
				sprintf(install_command,
						"/usr/bin/install -S -o %d -g %d -m 0%o %s %s",
						owner, group, mode, src_file, dst_file);
				fflush(stdout);
				error = system(install_command);
			}
			else if (S_ISLNK(src_st.st_mode)) {
				char link_contents[MAXPATHLEN];
				int nbytes;
				nbytes = readlink(src_file, link_contents,
								 sizeof(link_contents));
				if (nbytes > 0) {
					link_contents[nbytes] = '\0';
					symlink(link_contents, dst_file);
				}
				else {
					error = 1;
				}
			}
			else if (S_ISBLK(src_st.st_mode) || S_ISCHR(src_st.st_mode)) {
				error = mknod(dst_file, src_st.st_mode, src_st.st_dev);
			}
			else {
				printf("unsupported file type\n");
				error = 1;
			}
		}
		else {
			error = 1;
		}
		if (error)
			printf("failed\n");
		else
			printf("done\n");
	}
	return error;
}


/*-----------------------------------------------------------------
 * Name :	update_devices_if_needed()
 *
 * Description: Rerun /dev/MAKEDEV if it was modified
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
update_devices_if_needed(vector dv, vector files_list, char *(*truename_func)(char *))
{
	int i;
	int rc = 0;
	for (i = 0; i < vector_max(files_list); ++i) {
		char *dst_file = (*truename_func)(vector_lookup_index(files_list, i));
		if (!strcmp(dst_file, "/dev/MAKEDEV")) {
			printf("Updating devices...");
			fflush(stdout);
			rc = system("cd /dev && ./MAKEDEV all >/dev/null 2>/dev/null");
			printf("done\n");
			break;
		}
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	upgrade_prom_slotcpu()
 *
 * Description: upgrade prom for the specified slot/cpu
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/26/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_prom_slotcpu(int slot, int cpu, char *file)
{
	char cmd[1024];
	assert(valid_slot_cpu(slot, cpu));
	sprintf(cmd, "/usr/local/agile/bin/prom-upgrade %d %d %s", slot, cpu,
			file ? file : "");
	return system(cmd);
}


/*-----------------------------------------------------------------
 * Name :	upgrade_prom()
 *
 * Description: Upgrade the prom on the given board. Returns 0 if prom
 * upgrade successfully for all cpus.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_prom(char *board, char *file)
{
	int slot, cpu_min, cpu_max, cpu, error = 0;
	if (!strcmp(board, "toshiba")) {
		slot = 0;
		cpu_min = 1;
		cpu_max = 1;
	}
	else if (!strcmp(board, "ssc")) {
		slot = 0;
		cpu_min = 0;
		cpu_max = 0;
	}
	else if (!strcmp(board, "txrx")) {
		slot = 1;
		cpu_min = 0;
		cpu_max = 0;
	}
	else if (!strcmp(board, "fp")) {
		slot = 1;
		cpu_min = 3;
		cpu_max = 3;
	}
	else if (!strcmp(board, "sp")) {
		slot = 2;
		cpu_min = 0;
		cpu_max = 3;
	}
	else {
		return 1;
	}
	for (cpu = cpu_min; cpu <= cpu_max; ++cpu) {
		if (upgrade_prom_slotcpu(slot, cpu, file) != 0) {
#if ALLOW_PROM_UPGRADE_FAIL
			error = 0;
#else
			printf("Failed to upgrade the prom for slot %d cpu %d\n",
				   slot, cpu);
			printf("After the system comes up after reboot, verify that the slot/cpu\n");
			printf("is in UP state and execute the following command:\n");
			printf("system promupgrade %d %d\n", slot, cpu);
			error = 1;
#endif			
		}
	}
	return error;
}


/*-----------------------------------------------------------------
 * Name :	upgrade_proms_if_needed()
 *
 * Description: Upgrade the proms for the boards which had their
 * PROMs modified.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_proms_if_needed(vector dv, vector files_list, char *(*truename_func)(char *))
{
	int i;
	int rc = 0;
	for (i = 0; i < vector_max(files_list); ++i) {
		char *dst_file = (*truename_func)(vector_lookup_index(files_list, i));
		char *root_dst_file = upgrade_truename(
			vector_lookup_index(files_list, i));
		struct descriptor *descr = get_file_descriptor(dv, root_dst_file);
		if (is_prom(descr) || is_bigprom(descr)) {
			int j;
			for (j = 0; j < vector_max(dv); ++j) {
				struct descriptor *descr = vector_lookup_index(dv, j);
				if (!strcmp(descr->name, root_dst_file) && is_bigprom(descr)) {
					rc |= upgrade_prom(descr->board, dst_file);
				}
			}
			for (j = 0; j < vector_max(dv); ++j) {
				struct descriptor *descr = vector_lookup_index(dv, j);
				if (!strcmp(descr->name, root_dst_file) && is_prom(descr)) {
					rc |= upgrade_prom(descr->board, dst_file);
				}
			}
		}
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	reboot_boards()
 *
 * Description: reboot slots 1 and 2
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
reboot_boards(vector files_list, char *(*truname_func)(char *))
{
	chassisdMsgRebootReq_t msg;
	printf("Rebooting boards...");
	fflush(stdout);
	msg.hdr.mid = CHASSISD_REBOOT_REQ;
	msg.reboot_type = CM_REBOOT_TYPE_SLOT;
	msg.slot = 1;
	msg.slice = 0;
	sendAgileMsg(-1, &msg, sizeof(msg), "cm", 1, 0, 1, NULL, NULL, 0);
	msg.slot = 2;
	sendAgileMsg(-1, &msg, sizeof(msg), "cm", 1, 0, 1, NULL, NULL, 0);
	printf("done\n");
	return 0;
}


/*-----------------------------------------------------------------
 * Name :	reboot_cold()
 *
 * Description: Do a cold reboot of the system
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static void
reboot_cold()
{
	char command[256];
	printf("Rebooting the system...\n");
	/* if upgrading to secondary, tell reboot to switch the disks */
	sprintf(command, "/sbin/reboot -c %s\n", upgrade_secondary ? "-s" : "");
	system(command);
}


/*-----------------------------------------------------------------
 * Name :	upgrade_directories()
 *
 * Description: Update directories according to new dist file.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/26/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_directories(char *mtree_file)
{
	char cmd[MAXPATHLEN];
	int rc;
	printf("Updating directories...");
	fflush(stdout);
	sprintf(cmd, "/usr/sbin/mtree -qdUef %s -p %s", mtree_file,
			upgrade_secondary ? secondary_mount : "/");
	rc = system(cmd);
	printf("done\n");
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	shutdown_all()
 *
 * Description: Shutdown the system software and kill the daemons so
 * we can proceed with upgrade in peace.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/26/02
 *
 *-----------------------------------------------------------------
 */
void
shutdown_all(void)
{
	shutdown_system();
	shutdown_mcp();
	shutdown_inetd();
}


/*-----------------------------------------------------------------
 * Name :	reboot_all()
 *
 * Description: reset FP and SP and reboot SSC.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/26/02
 *
 *-----------------------------------------------------------------
 */
void
reboot_all(void)
{
	reboot_boards(NULL, NULL);
	reboot_cold();
}


/*-----------------------------------------------------------------
 * Name :	noisy_make_allfs_writable
 *
 * Description: Remount local filesystems writable making some noise
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	08/15/02
 *
 *-----------------------------------------------------------------
 */
int
noisy_make_allfs_writable(void)
{
	int rv;
	printf("Remounting the local filesystem read-write...");
	fflush(stdout);
	rv = make_allfs_writable();
	if (rv == 0)
		printf("done\n");
	else
		printf("failed\n");
	return rv;
}



/*-----------------------------------------------------------------
 * Name :	complete_upgrade()
 *
 * Description: The worst-case upgrade - shutdown and restart everything
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
complete_upgrade(vector dv, vector files_list, char *(*truename_func)(char *),
				 char *mtree_file)
{
	int rc = 0;

	if (!upgrade_secondary)
		shutdown_all();

	if ((upgrade_secondary || (rc = noisy_make_allfs_writable()) == 0)
		&& (rc = upgrade_directories(mtree_file)) == 0
		&& (rc = upgrade_files(files_list, truename_func)) == 0
		&& (rc = update_devices_if_needed(dv, files_list, truename_func)) == 0
		&& (rc = upgrade_proms_if_needed(dv, files_list,
										 truename_func) == 0)) {
		if (upgrade_secondary) {
			char cmd[256];
			sync();
			shutdown_all();

			printf("Updating configuration files\n");
			
			do_copy_files("/", secondary_mount, sys_config_files);
			sprintf(cmd, "/usr/sbin/pwd_mkdb -d %s/etc %s/etc/master.passwd",
					secondary_mount, secondary_mount);
			system(cmd);
		}				
		
		reboot_boards(files_list, truename_func);
		reboot_cold();
	}
	
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	determine_upgrade_method()
 *
 * Description: Find out the upgrade procedure for files in the files_list.
 * dv is the descriptors vector.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static upgrade_method_t
determine_upgrade_method(vector dv, vector files_list)
{
	return complete_upgrade;
}


/*-----------------------------------------------------------------
 * Name :	needs_upgrade()
 *
 * Description: Stores true in the location pointed to by the argument
 * need_upgrade_ptr if the file needs to be upgraded.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
needs_upgrade(vector dv, char *src_file,
			  char *dst_file, int *need_upgrade_ptr)
{
	struct descriptor *descr;
	int need_upgrade = 0;
	int rc = CMD_SUCCESS;

	printf("checking %s...", dst_file);
	fflush(stdout);
	descr = get_file_descriptor(dv, upgrade_truename(src_file));
	if (is_config(descr)) {
		printf("user data");
	}
	else if (is_version(descr)) {
		printf("version file");
		need_upgrade = 1;
	}
	else {
		struct stat src_st, dst_st;
		if (lstat(src_file, &src_st) == 0) {
			if (lstat(dst_file, &dst_st) < 0) {
				printf("new file");
				need_upgrade = 1;
			}
			else {
				if (S_ISREG(src_st.st_mode) && S_ISREG(dst_st.st_mode)) {
					int src_fd = -1, dst_fd = -1;
					if ((src_fd = open(src_file, O_RDONLY)) >= 0
						&& (dst_fd = open(dst_file, O_RDONLY)) >= 0) {
						int cmp_result;
						if ((rc = compare_file_contents(src_fd, dst_fd, src_file, src_st.st_size,
														dst_file, dst_st.st_size, &cmp_result)) == CMD_SUCCESS) {
							if (cmp_result) {
								need_upgrade = 1;
								printf("needs upgrade");
							}
							else {
								printf("up-to-date");
							}
						}
						else {
							printf("failed to compare\n");
						}
					}
					close(dst_fd);
					close(src_fd);
				}
				else if (S_ISLNK(src_st.st_mode)) {
					printf("symbolic link");
				}
				else {
					printf("special file");
				}
			}
		}
		else {
			rc = CMD_ERR_FAILED;
		}
	}
	printf("\n");
	*need_upgrade_ptr = need_upgrade;
	return rc;
}
						 


/*-----------------------------------------------------------------
 * Name :	make_files_to_upgrade_list()
 *
 * Description: Make the list of the files which need to be upgraded
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
make_files_to_upgrade_list(
	vector dv,
	vector files_to_upgrade,
	vector files)
{
	int rc = CMD_SUCCESS;
	int i;
	for (i = 0; rc == CMD_SUCCESS && i < vector_max(files); ++i) {
		char *src_file = vector_lookup_index(files, i);
		char *dst_file = cur_truename_func(src_file);
		int need_upgrade;
		rc = needs_upgrade(dv, src_file, dst_file,  &need_upgrade);
		if (need_upgrade)
			vector_push(files_to_upgrade, src_file);
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	read_upgrade()
 *
 * Description: Read the descriptor in the upgrade directory
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static vector
read_upgrade_descriptor()
{
	char filename[MAXPATHLEN];
	sprintf(filename, "%s%s", patch_dir_name, descriptor_file);
	return read_descriptor(filename);
}


/*-----------------------------------------------------------------
 * Name :	get_file_inode()
 *
 * Description: Return the inode of the file filename or -1 if can not stat
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/28/02
 *
 *-----------------------------------------------------------------
 */
static void *
get_file_inode(void *filename)
{
	struct stat st;
	ino_t rv = -1;
	if (lstat(filename, &st) == 0) {
		rv = st.st_ino;
	}
	return (void *)rv;
}


/*-----------------------------------------------------------------
 * Name :	copy_local_truename()
 *
 * Description: Return the name of temporary copy of the file.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	09/13/02
 *
 *-----------------------------------------------------------------
 */
static char *
copy_local_truename(char *file)
{
	static char name[MAXPATHLEN];
	strcpy(name, local_dest);
	strcat(name, file + strlen(patch_dir_name));
	return name;
}


/*-----------------------------------------------------------------
 * Name :	copy_files_to_local()
 *
 * Description: Copy files to local disk and return new files locations.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	09/13/02
 *
 *-----------------------------------------------------------------
 */
static int
copy_files_to_local(vector files, vector *new_files)
{
	int rc = CMD_ERR_FAILED;
	char *tmp = zstrdup(MTYPE_TMP, local_dest);

	if (ensure_dir_exists(tmp, 0777) == 0) {
		char command[256];
		sprintf(command, "mount_mfs -s 100000 swap %s", local_dest);
		if (system(command) == 0) {
			/* Create directories in temporary copy */
			sprintf(command, "/usr/sbin/mtree -qdUef "
					"%s%s -p %s >/dev/null 2>/dev/null",
					mount_dir, mtree_file_name, local_dest);
			if (system(command) == 0) {
				/* Copy mtree file to temporary location because will
				   need it later to make directories in the real tree */
				sprintf(command, "/usr/bin/install %s%s %s%s",
						mount_dir, mtree_file_name,
						local_dest, mtree_file_name);
				if (system(command) == 0) {
					if (upgrade_files(files, copy_local_truename) == 0) {
						int i;
						*new_files = vector_init(vector_max(files));
						for (i = 0; i < vector_max(files); ++i) {
							char *f = vector_lookup_index(files, i);
							char *newf;
							asprintf(&newf, "%s%s", local_dest,
									 cur_truename_func(f));
							vector_push(*new_files, newf);
						}
						rc = CMD_SUCCESS;
					}
					else {
						printf("Please retry upgrade using a management"
							   "interface.\n");
					}
				}
			}
			else {
				printf("Can't update local directory tree\n");
			}
			if (rc != CMD_SUCCESS)
				execute_command("/sbin/umount", 1, local_dest, NULL);
		}
		else {
			printf("Can't mount filesystem for local copy\n");
		}
	}
	else {
		printf("Can't create directory %s\n", local_dest);
	}
	zfree(MTYPE_TMP, tmp);
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	is_fp_addr()
 *
 * Description: Return true if the patch is accessed through fp
 * interface.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	11/01/02
 *
 *-----------------------------------------------------------------
 */
static int
is_fp_addr(char *patch_location)
{
	char *column = strchr(patch_location, ':');
	in_addr_t host_addr, our_addr;
	int rc;
	char iface[256];
	*column = 0;
	host_addr = nfxnis_hostAddr("0.0.0.0", patch_location);
	rc = get_our_addr(host_addr, iface, &our_addr);
	return !(rc == 0 && (!strcmp(iface, "sc1") || !strcmp(iface, "sc2")));
}

/*-----------------------------------------------------------------
 * Name :	need_local_copy()
 *
 * Description: Return true if need to copy files to local disk before
 * upgrading - if using fp interface to upgrade.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	09/13/02
 *
 *-----------------------------------------------------------------
 */
static int
need_local_copy()
{
	return !is_ftp_upgrade(patch_location) && is_fp_addr(patch_location);
}


/*-----------------------------------------------------------------
 * Name :	perform_upgrade()
 *
 * Description: upgrade a system from a complete distribution
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
perform_upgrade(void)
{
	vector files = NULL;
	vector files_no_dups = NULL;
	vector dv;
	int rc = CMD_ERR_FAILED;
	
	printf("Reading the system descriptor...");
	fflush(stdout);
	dv = read_upgrade_descriptor();
	if (dv != NULL) {
		printf("done\n");
		files = vector_init(200);
		printf("Reading the upgrade directory...");
		fflush(stdout);
		rc = find_files(patch_dir_name, files);
		if (rc == CMD_SUCCESS) {
			vector files_to_upgrade;
			printf("done\n");
			printf("Making upgrade files list\n");
			files_no_dups = vector_remove_duplicates(files, get_file_inode);
			vector_free(files);
			files_to_upgrade = vector_init(200);
			rc = make_files_to_upgrade_list(dv, files_to_upgrade, files_no_dups);
			if (rc == CMD_SUCCESS) {
				int make_local_copy = need_local_copy();
				printf("The following files will be upgraded:\n");
				print_string_vector(files_to_upgrade);
				if (make_local_copy) {
					vector new_files;
					printf("Copying files to local disk\n");
					rc = copy_files_to_local(files_to_upgrade, &new_files);
					if (rc == CMD_SUCCESS) {
						execute_command("/sbin/umount", 1, mount_dir, NULL);
						files_to_upgrade = new_files;
					}
				}
				if (rc == CMD_SUCCESS) {
					char mtree_file[256];
					sprintf(mtree_file, "%s%s",
							make_local_copy ? local_dest : mount_dir,
							mtree_file_name);
					rc = (*determine_upgrade_method(dv, files_to_upgrade))
						(dv, files_to_upgrade, cur_truename_func,
						 mtree_file);
				}
			}
			else {
				printf("Failed to get the list of files to upgrade\n");
			}
			vector_free_all(files_no_dups);
		}
		else {
			printf("failed\n");
		}
		vector_free(files);
		vector_free_all(dv);
	}
	else {
		printf("failed\n");
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	parse_args()
 *
 * Description: parse command line arguments
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	11/01/02
 *
 *-----------------------------------------------------------------
 */
static int
parse_args(int argc, char **argv)
{
	int rc = CMD_ERR_FAILED;
	upgrade_secondary = 0;

	if (argc == 1) {
		patch_location = argv[0];
		rc = CMD_SUCCESS;
	}
	else if (argc == 2) {
		if (!strcmp(argv[0], "-s")) {
			patch_location = argv[1];
			upgrade_secondary = 1;
			rc = CMD_SUCCESS;
		}
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	do_upgrade()
 *
 * Description: make system upgrade
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
int
do_upgrade(int argc, char **argv)
{
	struct version current_version, upgrade_version;
	struct version org_upgrade_version;
	int rc = CMD_SUCCESS;
	vector secondary_fstabs = NULL;

	rc = parse_args(argc, argv);
	if (rc != CMD_SUCCESS)
		return rc;
	
	become_root();
	if (upgrade_secondary) {
		printf("Mounting secondary disk...");
		fflush(stdout);
		secondary_fstabs = make_secondary_fstabs();
		u_mount_secondary(secondary_fstabs);
		rc = convert_rc(mount_secondary(secondary_fstabs));
		printf("%s\n", rc == CMD_SUCCESS ? "done." : "failed.");
		cur_truename_func = secondary_upgrade_truename;
	}
	else {
		cur_truename_func = upgrade_truename;
	}

	if (rc == CMD_SUCCESS) {
		rc = mount_patch(patch_location);
		if (rc == CMD_SUCCESS) {
			printf("Mounted the upgrade directory\n");
			if ((rc = get_upgrade_version(&upgrade_version)) == CMD_SUCCESS) {
				get_org_upgrade_version(&org_upgrade_version);
				if ((rc = get_local_version(&current_version)) == CMD_SUCCESS) {
					printf("Current system version: %d.%d.%d "
						   "upgrade version: %d.%d.%d\n",
						   current_version.major, current_version.minor,
						   current_version.patch,
						   upgrade_version.major, upgrade_version.minor,
						   upgrade_version.patch);
					if (compare_versions(&current_version, &upgrade_version) >= 0) {
						printf("The system is up-to-date\n");
					}
					else {
						if (compare_versions(&current_version,
											 &org_upgrade_version) >= 0) {
							rc = perform_upgrade();
						}
						else {
							printf("Start version %d.%d.%d is newer "
								   "than current, can not upgrade\n",
								   org_upgrade_version.major,
								   org_upgrade_version.minor,
								   org_upgrade_version.patch);
							rc = CMD_ERR_FAILED;
						}
					}
				}
			}
			else {
				printf("failed to determine the upgrade version\n");
			}
			umount_patch();
		}
	}

	if (upgrade_secondary) {
		u_mount_secondary(secondary_fstabs);
		free_fstab_vector(secondary_fstabs);
	}
	
	unbecome_root();
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	show_local_version()
 *
 * Description: Show the system version
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
show_local_version(void)
{
	struct version current_version;
	int rc = get_root_version(&current_version);
	if (rc != CMD_SUCCESS) {
		fprintf(stderr, "Can't determine current version\n");
	}
	else {
		print_version(&current_version);
	}
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	show_remote_version()
 *
 * Description: Show the version of the patch 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
static int
show_remote_version(char *patch_path)
{
	struct version version;
	int rc;
	
	become_root();
	rc = mount_patch(patch_path);
	if (rc == CMD_SUCCESS) {
		if ((rc = get_patch_version(&version)) == CMD_SUCCESS) {
			print_version(&version);
		}
		umount_patch();
	}
	else {
		fprintf(stderr, "Can not mount the patch directory\n");
	}
	unbecome_root();
	
	return rc;
}


/*-----------------------------------------------------------------
 * Name :	do_show_version()
 *
 * Description: show the local version or a patch version
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/14/02
 *
 *-----------------------------------------------------------------
 */
int
do_show_version(int argc, char **argv)
{
	if (argc == 0) {
		return show_local_version();
	}
	else {
		assert(argc == 1);
		return show_remote_version(argv[0]);
	}
}


/*-----------------------------------------------------------------
 * Name :	do_prom_upgrade()
 *
 * Description: prom upgrade command 
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/26/02
 *
 *-----------------------------------------------------------------
 */
int
do_prom_upgrade(int argc, char **argv)
{
	int rc = CMD_ERR_FAILED;
	if (argc == 2) {
		int slot = atoi(argv[0]);
		int cpu = atoi(argv[1]);
		if (valid_slot_cpu(slot, cpu)) {
			rc = upgrade_prom_slotcpu(slot, cpu, NULL);
		}
	}
	return rc;
}
