/*
 *-----------------------------------------------------------------
 *
 * Name:	prom-upgrade.c
 *
 * RCS:  $Id: prom-upgrade.c,v 1.7 2002/12/04 20:09:40 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Prom upgrade utility for BSD
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created: 09/05/2002
 *
 *-----------------------------------------------------------------
 */


#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <neteee/eee.h>
#include <sys/socket.h>

#include "nfx-incl.h"
#include "exports/ncm-api.h"
#include "exports/evm-api.h"
#include "exports/ccm-api.h"
#include "exports/anpssc-api.h"
#include "exports/elog-api.h"

#include "prom-upgrade-api.h"

#undef DEBUG
#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif

/*XXX use defs in amd29f040.h*/
#define SECTOR_SIZE 0x10000
#define NSECTORS 8

static int upgrade_secondary;
static int upgrade_diag;
static int upgrade_large;

/*-----------------------------------------------------------------
 * Name :	valid_slot_cpu()
 *
 * Description: Return true if slot/cpu combination is valid
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
valid_slot_cpu(int slot, int cpu)
{
	switch (slot) {
	case NFX_SLOTTYPE_SSC:
		return 0 <= cpu && cpu < NFX_SSC_NUM_CPUS;
	case NFX_SLOTTYPE_NFP:
		return 0 <= cpu && cpu < NFX_NFP_NUM_CPUS;
	case NFX_SLOTTYPE_FC:
		return 0 <= cpu && cpu < NFX_FCNIM_NUM_CPUS;
	default:
		return 0;
	}
	
}


/*-----------------------------------------------------------------
 * Name :	usage()
 *
 * Description: display usage message
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/10/02
 *
 *-----------------------------------------------------------------
 */
static void
usage(void)
{
	printf("Usage: prom-upgrade #slot #cpu [<prom-file>]\n");
}


/*-----------------------------------------------------------------
 * Name :	prom_location()
 *
 * Description: return the location of the file containing the PROM for
 * the give slot/cpu combination.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static char *
prom_location(int slot, int cpu)
{
	assert(valid_slot_cpu(slot, cpu));
	
	switch (slot) {
	case NFX_SLOTTYPE_SSC:
		if (cpu == 0)
			return "/usr/local/agile/Images/r7k.bin";
		else if (cpu == 1)
			return "/usr/local/agile/Images/toshiba.bin";
	case NFX_SLOTTYPE_NFP:
		return "/usr/local/agile/Images/sibyte.bin";
	case NFX_SLOTTYPE_FC:
		return "/usr/local/agile/Images/r7k.bin";
	default:
		return NULL;
	}
}



typedef int (*upgrade_func_t)(int slot, int cpu, char *prom_file);


static int
ssc_nsectors(void)
{
	return upgrade_large ? (2 * 1024 * 1024) / SECTOR_SIZE :
		(512 * 1024) / SECTOR_SIZE;
}


/*-----------------------------------------------------------------
 * Name :	compare_proms()
 *
 * Description: Compare the contents of the prom files src_fd and dst_fd
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/10/02
 *
 *-----------------------------------------------------------------
 */
static int
compare_proms(int src_fd, int dst_fd, int *differ_ptr)
{
	int i;
	int differ = 0;
	int error = 0;
	char src_buf[SECTOR_SIZE];
	char dst_buf[SECTOR_SIZE];
	
	lseek(src_fd, SEEK_SET, 0);
	lseek(dst_fd, SEEK_SET, 0);
	for (i = 0; !error && !differ && i < ssc_nsectors(); ++i) {
		int nbytes;
		nbytes = read(src_fd, src_buf, SECTOR_SIZE);
		if (nbytes != SECTOR_SIZE) {
			if (nbytes < 0) {
				perror("read");
				error = 1;
			}
			else {
				printf("premature end of the source file, i=%d, nbytes=%d\n",
					   i, nbytes);
				error = 1;
			}
		}
		nbytes = read(dst_fd, dst_buf, SECTOR_SIZE);
		if (nbytes != SECTOR_SIZE) {
			if (nbytes < 0) {
				perror("write");
				error = 1;
			}
			else {
				printf("premature end of the destination file, i=%d, nbytes=%d\n",
					   i, nbytes);
				error = 1;
			}
		}
		differ = memcmp(src_buf, dst_buf, SECTOR_SIZE);
	}
	if (!error) {
		*differ_ptr = differ;
	}
	return error;
}


/*-----------------------------------------------------------------
 * Name :	ssc_write_prom()
 *
 * Description: write the prom from src_fd to dst_fd
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/10/02
 *
 *-----------------------------------------------------------------
 */
static int
ssc_write_prom(int src_fd, int dst_fd)
{
	char src_buf[SECTOR_SIZE];
	int i;
	int error = 0;
	
	lseek(src_fd, SEEK_SET, 0);
	lseek(dst_fd, SEEK_SET, 0);
	for (i = 0; !error && i < ssc_nsectors(); ++i) {
		int nbytes = read(src_fd, src_buf, SECTOR_SIZE);
		if (nbytes != SECTOR_SIZE) {
			if (nbytes < 0) {
				perror("read");
				error = 1;
			}
			else {
				printf("premature end of the source file\n");
				error = 1;
			}
		}
		if (!error) {
			printf("    program sector %d...", i);
			fflush(stdout);
			nbytes = write(dst_fd, src_buf, SECTOR_SIZE);
			if (nbytes != SECTOR_SIZE) {
				printf("failed to write to the PROM sector %d\n", i);
				perror("write");
				error = 1;
			}
			else {
				printf("done\n");
			}
		}
	}
	return error;
}


static char *
prom_device_part(void)
{
	if (upgrade_large) {
		return "c";
	}
	else {
		return upgrade_secondary ? "b" : "a";
	}
}


static char *
prom_device_unit(void)
{
	if (upgrade_large) {
		return "";
	}
	else {
		return upgrade_diag ? "1" : "0";
	}
}


/*-----------------------------------------------------------------
 * Name :	upgrade_ssc_prom()
 *
 * Description: Upgrade the local prom
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_ssc_prom(int slot, int cpu, char *prom_file)
{
	int src_fd = -1;
	int dst_fd = -1;
	int error = 0;
	int differ;
	char prom_device_name[20];

	printf("Upgrading SSC PROM from %s\n", prom_file);

	if (!error && (src_fd = open(prom_file, O_RDONLY)) < 0) {
		perror("open(source)");
		error = 1;
	}
	sprintf(prom_device_name, "/dev/prom%s%s", prom_device_part(),
			prom_device_unit());
	debug(("open prom dev %s\n", prom_device_name));
	if (!error && (dst_fd = open(prom_device_name, O_RDWR)) < 0) {
		perror("open(dest)");
		error = 1;
	}

	error = compare_proms(src_fd, dst_fd, &differ);

	if (!error) {
		if (differ) {
			error = ssc_write_prom(src_fd, dst_fd);
			if (!error) {
				error = compare_proms(src_fd, dst_fd, &differ);
				if (differ) {
					printf("prom-upgrade: verify failed\n");
				error = 1;
				}
			}
		}
		else {
			printf("The contents of PROM is identical with the source file\n");
		}
	}

	if (src_fd != -1)
		close(src_fd);
	if (dst_fd != -1) 
		close(dst_fd);
   
	return error;
}


/*-----------------------------------------------------------------
 * Name :	sendWriteMessage()
 *
 * Description: start writing the remote prom and wait for the response.
 * The timeout is based on the amount data to program the prom flash memory
 * It takes about 1~3 sec to program 1 64K block. So allow for
 * 5 x transfer-size/64k secs, or least 1 min.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
sendWriteMessage(int s, int slot, int cpu, int len)
{
	promUpgradeHdr_t hdr;
	promUpgradeHdr_t resp;
	struct timeval tv;

    tv.tv_sec  = MAX(60, (len/SECTOR_SIZE) * 5);
    tv.tv_usec = 0;
	hdr.msgType = promUpgradePerformWrite;
	return sendAgileMsg(s, &hdr, sizeof(hdr), "prom-upgrade", cpu, slot, 0,
						&tv, &resp, sizeof(resp)) == sizeof(resp)
		&& resp.msgType == promUpgradeWriteSuccess;
}


/*-----------------------------------------------------------------
 * Name :	sendStartMessage()
 *
 * Description: Send a message beginning the prom upgrade sequence
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
sendStartMessage(int s, int slot, int cpu)
{
	promUpgradeHdr_t hdr;
	promUpgradeHdr_t resp;
	struct timeval tv = { 10, 0};
	hdr.msgType = promUpgradeStart;
	return sendAgileMsg(s, &hdr, sizeof(hdr), "prom-upgrade", cpu, slot, 0,
						&tv, &resp, sizeof(resp)) == sizeof(resp)
		&& resp.msgType == promUpgradeStartAck;
		
}


/*-----------------------------------------------------------------
 * Name :	sendChecksum()
 *
 * Description: Send a checksum request
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
sendChecksum(int s, int slot, int cpu, uchar8 cksum, uint32 sum_len)
{
		
	promUpgradeChecksumMsg_t msg;
	promUpgradeHdr_t resp;
	struct timeval tv = { 10, 0};
	msg.hdr.msgType = promUpgradeChecksum;
	msg.cksum = cksum;
	msg.total_len = sum_len;
	msg.msg_verif = PROM_UPGRADE_MSG_VERIF(sum_len, cksum);
	return sendAgileMsg(s, &msg, sizeof(msg), "prom-upgrade", cpu, slot, 0,
						&tv, &resp, sizeof(resp)) == sizeof(resp)
		&& resp.msgType == promUpgradeChecksumSuccess;
}


/*-----------------------------------------------------------------
 * Name :	upgrade_remote_prom()
 *
 * Description: Upgrade remote prom on the slot/cpu from the file prom_file.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
upgrade_remote_prom(int slot, int cpu, char *prom_file)
{
	int fd = -1;
	struct stat statbuf;
	int s = -1;
	int error = 0;

	printf("Upgrading PROM for slot %d cpu %d from %s\n",
		   slot, cpu, prom_file);

	s = socket(PF_EEE, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		error = 1;
	}
	if (error == 0 && bindEEESocket(s, slot, cpu, 0)) {
		perror("bindEEESocket\n");
		error = 1;
	}

	if (error == 0 && (fd = open(prom_file, O_RDONLY)) < 0) {
		perror("open");
		error = 1;
	}
	if (error == 0 && fstat(fd, &statbuf) < 0) {
		perror("stat");
		error = 1;
	}

	if (error == 0 && sendStartMessage(s, slot, cpu)) {
		int remain, send_bytes;
		uchar8 cksum = 0;
		remain = statbuf.st_size;
		debug(("Sending file length %d\n", remain));
		error = 0;
		while (remain > 0 && !error) {
			promUpgradeDataMsg_t msg;
			promUpgradeHdr_t resp;
			struct timeval tv = { 10, 0};
			send_bytes = MIN(remain, PROM_UPGRADE_MAX_DATA_SIZE);
			msg.hdr.msgType = promUpgradeData;
			if (read(fd, msg.data, send_bytes) == send_bytes) {
				int i;
				for (i = 0; i < send_bytes; ++i) {
					cksum ^= msg.data[i];
				}
				debug(("Send data len = %d, remain = %d\n", send_bytes, remain));
				if (sendAgileMsg(s, &msg, sizeof(promUpgradeHdr_t) + send_bytes, "prom-upgrade",
								 cpu, slot, 0, &tv, &resp, sizeof(resp)) == sizeof(resp)
					&& resp.msgType == promUpgradeDataAck) {
					remain -= send_bytes;
				}
				else {
					error = 1;
				}
			}
			else {
				error = 1;
			}
		}
		if (error == 0) {
			if (!sendChecksum(s, slot, cpu, cksum, statbuf.st_size)) {
				printf("checksum failed\n");
				error = 1;
			}
		}
		if (error == 0) {
			printf("File sent, started programming...");
			fflush(stdout);
			if (!sendWriteMessage(s, slot, cpu, statbuf.st_size)) {
				printf("failed\n");
				error = 1;
			}
			else {
				printf("done\n");
			}
		}
	}
	else {
		printf("Timed out waiting for response\n");
		error = 1;
	}

	if (fd != -1)
		close(fd);
	if (s != -1)
		close(s);

	return error;
}



/*-----------------------------------------------------------------
 * Name :	prom_upgrade_func()
 *
 * Description: Return the function for upgrading the prom on slot/cpu
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static upgrade_func_t
prom_upgrade_func(int slot, int cpu)
{
	assert(valid_slot_cpu(slot, cpu));
	switch (slot) {
	case NFX_SLOTTYPE_SSC:
		if (cpu == 0)
			return upgrade_ssc_prom;
		else if (cpu == 1)
			return upgrade_remote_prom;
	case NFX_SLOTTYPE_NFP:
	case NFX_SLOTTYPE_FC:
		return upgrade_remote_prom;
	default:
		return NULL;
	}
}


/*-----------------------------------------------------------------
 * Name :	do_prom_upgrade()
 *
 * Description: Perform the prom upgrade for slot/cpu
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	05/09/02
 *
 *-----------------------------------------------------------------
 */
static int
do_prom_upgrade(int slot, int cpu, char * prom_file)
{
	return (*(prom_upgrade_func(slot, cpu)))(slot, cpu, prom_file);
}



int
main(int argc, char **argv)
{
	int     slot, cpu;
    char *  prom_file;
	int c;

	while ((c = getopt(argc, argv, "sdl")) != -1) {
		switch (c) {
		case 's':
			upgrade_secondary = 1;
			break;
		case 'd':
			upgrade_diag = 1;
			break;
		case 'l':
			upgrade_large = 1;
			break;
		case '?':
		default:
			usage();
			exit(1);
		}
	}
	
	argc -= optind;
	argv += optind;

	if (argc < 2 || argc > 3) {
		usage();
		exit(1);
	}

	slot = atoi(argv[0]);
	cpu = atoi(argv[1]);

	(void)mknod("/dev/proma0", 0660 | S_IFCHR, makedev(54, 0));
	(void)mknod("/dev/proma1", 0660 | S_IFCHR, makedev(54, 1));
	(void)mknod("/dev/promb0", 0660 | S_IFCHR, makedev(54, 2));
	(void)mknod("/dev/promb1", 0660 | S_IFCHR, makedev(54, 3));
	(void)mknod("/dev/promc", 0660 | S_IFCHR, makedev(54, 4));

	if (!valid_slot_cpu(slot, cpu)) {
		usage();
		exit(1);
	}
    if (argc == 3)
        prom_file = argv[2];
    else
        prom_file = prom_location(slot, cpu);

	return do_prom_upgrade(slot, cpu, prom_file);
}
