/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: System Timekeeper
 *
 * RCS:  $Id: timekeeper.c,v 1.5 2002/10/08 03:16:19 johnh Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This program periodically sends time tags to all the
 *              embedded systems.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 15 Mar 2002
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>

#include "nfx-defs.h"
#include "exports/eee-error.h"
#include "exports/anpssc-api.h"
#include "exports/timer-api.h"
#include "exports/elog-api.h"

typedef struct {
	uint32	slot;
	uint32	cpu;
} time_client_t;

// The list of embedded systems we're going to update with time
// THIS SHOULD BE DYNMAICALLY CONFIGURED or use a broadcast message

static time_client_t clients[] = {
	{NFX_SLOTTYPE_NFP, NFX_NFP_TXRX1},
	{NFX_SLOTTYPE_NFP, NFX_NFP_FP2},
	{NFX_SLOTTYPE_FC, NFX_FCNIM_CPUID3},    // cpuid3 is fc slice0
	{NFX_SLOTTYPE_FC, NFX_FCNIM_CPUID2},    // cpuid2 is fc slice1
	{NFX_SLOTTYPE_FC, NFX_FCNIM_CPUID1},    // cpuid1 is fc slice2
	{NFX_SLOTTYPE_FC, NFX_FCNIM_CPUID0}     // cpuid0 is fc slice3
};

static int num_clients = sizeof (clients) / sizeof (time_client_t);

#define SYNC_INTERVAL 100		// seconds between messages


static void
get_new_time(eeeGMT_t *new_time)
{
	struct timeval curr_time;
	gettimeofday(&curr_time, NULL);
	new_time->sec = (int32)curr_time.tv_sec;		// Probably not
	new_time->usec = (int32)curr_time.tv_usec;	// necessary
}


static void
send_everybody(int sock)
{
	int i;
	eeeGMT_t new_time;
	get_new_time(&new_time);
	for (i = 0; i < num_clients; i++) {
		sendAgileMsg (sock, &new_time, sizeof(new_time),
					  "timekeeper", clients[i].cpu,
					  clients[i].slot, MGMTPLANE, NULL, NULL, 0);
	}
}


extern int
main(int argc, char *argv[])
{
	int32 our_app_id = eee_getApplicationId ("timekeeper");
	int sock;
	
	if (our_app_id == EEE_ERR_INVALID_APP_NAME) {
		fprintf (stderr, "timekeeper application unknown\n");
		kill (getppid(), SIGTERM);	// This should stop things dead
		exit (1);
	}

	sock = regAgileApp ("timekeeper");
	if (sock < 0) {
		fprintf (stderr, "Error registering timekeeper\n");
		kill (getppid(), SIGTERM);	// This should stop things dead
		exit (1);
	}

	kill (getppid(), SIGUSR1);		// Tell MCP all is OK

	send_everybody(sock);

// Forever, get current time and send it to ALL the embedded CPUs
	while (1) {
		struct timeval to = { SYNC_INTERVAL, 0 };
		int rv;
		char buf[100];
		char app_name[100];
		uint32_t cpu, plane, slot;
		
		rv = recvAgileMsg(sock, buf, sizeof(buf),
						  &to, app_name, &cpu, &slot, &plane);
		if (rv < 0) {
			E_LOG(class_1, err_s, eee_getApplicationId("timekeeper"),
				  0, 0, 0, ("recvAgileMsg failed: %m"));
			exit(1);
		}
		else if (rv > 0) {
			/* Got a message, send current time in reply */
			eeeGMT_t new_time;
			get_new_time(&new_time);
			if (sendAgileMsg(sock, &new_time, sizeof(new_time),
							 app_name, cpu, slot, plane,
							 NULL, NULL, NULL) < 0) {
				E_LOG(class_1, err_s, eee_getApplicationId("timekeeper"),
					  0, 0, 0, ("sendAgileMsg failed: %m"));
				exit(1);
			}
		}
		else if (rv == 0) {
			/* Send current time to everybody */
			send_everybody(sock);
		}
	}
}
