#include <sys/types.h>
#include <sys/param.h>
#include <sys/fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <nfx-defs.h>
#include <nfx-types.h>
#include <exports/anpssc-api.h>
#include <exports/evm-api.h>
#include <exports/scsi-api.h>
#include <exports/scsi-msg.h>
#include <sys/socket.h>


static int
getVolumeId(int s, char *volname, evm_volId_t *vol_id)
{
	evm_MsgGetId_t msg;
	evm_MsgStat_t resp;
	int rv;
	int error = 0;

	msg.hdr.msgType = EVM_VOL_GET_ID;
	msg.hdr.len = sizeof(msg);
	msg.hdr.clientPrivate = EVM_VOL_GET_ID;
	memset(msg.name, 0, sizeof(msg.name));
	strcpy(msg.name, "crash");
	rv = sendAgileMsg(s, &msg, sizeof(msg), "evm", 0, 0, 0, 0, &resp, sizeof(resp));
	if (rv < 0) {
		fprintf(stderr, "getVolumeId: sendAgileMessage failed\n");
		error = 1;
	}
	if (!error && (resp.hdr.msgType != EVM_MSG_STAT || resp.hdr.clientPrivate != EVM_VOL_GET_ID)) {
		fprintf(stderr, "getVolumeId: invalid response message\n");
		error = 1;
	}
	if (!error && resp.stat != NFX_OK) {
		fprintf(stderr, "getVolumeId: EVM_VOL_GET_ID failed\n");
		error = 1;
	}
	if (!error) {
		*vol_id = resp.payload.lVolId;
	}
	return error;
}


static int
getStorageId(int s, evm_volId_t vol_id, uint32_t *handle, uint64_t *startBlock)
{
	evm_MsgGetStrg_t msg;
	evm_MsgLVStrg_t resp;
	int rv;
	int error = 0;
	msg.hdr.msgType = EVM_GET_LV_STRG;
	msg.hdr.len = sizeof(msg);
	msg.hdr.clientPrivate = EVM_GET_LV_STRG;
	msg.id = vol_id;
	rv = sendAgileMsg(s, &msg, sizeof(msg), "evm", 0, 0, 0, 0, &resp, sizeof(resp));
	if (rv < 0) {
		fprintf(stderr, "getStorageId: sendAgileMsg failed\n");
		error = 1;
	}
	if (!error && (resp.hdr.msgType != EVM_LV_STRG_CFG ||
				   resp.hdr.clientPrivate != EVM_GET_LV_STRG)) {
		fprintf(stderr, "getStorageId: invalid response message\n");
		error = 1;
	}
	if (!error && resp.nbStrg == 1) {
		evm_strgConfig_t *cfg = &resp.strg[0].au;
		*handle = cfg->openHandle;
		*startBlock = cfg->startBlk;
	}
	return error;
}


static void
usage(void)
{
	fprintf(stderr, "Usage: volread <volume-name> <lba> <count>\n");
}

int
main(int argc, char **argv)
{
	char *volname;
	uint64_t block_count;
	uint64_t block_offset;
	int s;
	evm_volId_t vol_id;
	uint32_t handle;
	uint64_t startBlock;
	sdm_readDataReqMsg_t readreq;
	uint64_t block;
	int done;
	int outstanding_ios;
	int max_outstanding_ios;
	int min_outstanding_ios;
	int val;

	if (argc < 4) {
		usage();
		return -1;
	}
	
	volname = argv[1];
	block_offset = atoi(argv[2]);
	block_count = atoi(argv[3]);

	fprintf(stderr, "volname=%s block_offset=%lld block_count=%lld\n",
		   volname, block_offset, block_count);

	s = socket(PF_EEE, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket\n");
		exit(1);
	}

	if (!bindEEESocket(s, 0, 0, 0)) {
		perror("bindEEESocket\n");
		exit(1);
	}

	val = 256000;
	if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, &val, sizeof(val)) < 0) {
		perror("setsockopt");
		exit(1);
	}

	if (getVolumeId(s, volname, &vol_id) != NFX_OK) {
		perror("getVolumeId");
		exit(1);
	}

	fprintf(stderr, "got volume id %llx\n", vol_id.whole_thing);

	if (getStorageId(s, vol_id, &handle, &startBlock) != NFX_OK) {
		perror("getStorageId");
		exit(1);
	}

	fprintf(stderr, "storage handle = %x start_block = %llx\n", handle, startBlock);

	block = startBlock + block_offset;
	done = 0;
	outstanding_ios = 0;
	max_outstanding_ios = 120;
	min_outstanding_ios = 80;
	while (!done) {
		int rv;
		readreq.hdr.mid = SDM_MID_READ_REQ;
		readreq.hdr.chandle = block;
		readreq.handle = handle;
		readreq.blkOffset = block;
		readreq.blkCnt = 2;
		readreq.blkSize = 512;
		readreq.dataDest = 0;
		rv = sendAgileMsg(s, &readreq, sizeof(readreq), "sdm", 0, 0, 0, 0, 0, 0);
		if (rv < 0) {
			fprintf(stderr, "sendAgileMsg failed\n");
		}
		++outstanding_ios;
		block += 2;

		if (outstanding_ios >= max_outstanding_ios
			|| block == startBlock + block_offset + block_count) {
			int needios = block == startBlock + block_offset + block_count ? 0 : min_outstanding_ios;
			while (outstanding_ios > needios) {
				char rspbuf[4096];
				rv = recvAgileMsg(s, rspbuf, sizeof(rspbuf), 0, 0, 0, 0, 0);
				if (rv > 0) {
					sdm_readDataRspMsg_t *rsp = (sdm_readDataRspMsg_t *)rspbuf;
					if (rsp->hdr.mid != SDM_MID_READ_RSP) {
						fprintf(stderr, "invalid message received\n");
					}
					if (rsp->eCode != NFX_OK) {
						fprintf(stderr, "read failed\n");
					}
					lseek(1, (rsp->hdr.chandle - startBlock) * 512, SEEK_SET);
					write(1, rsp->data, rsp->dataLen);
				}
				else {
					fprintf(stderr, "recv message failed\n");
				}
				--outstanding_ios;
			}
			if (block == startBlock + block_offset + block_count)
				done = 1;
		}
	}

	return 0;
}
